# /etc/profile.ksh
# System wide environment and startup programs
# Functions and aliases go in $HOME/.profile.ksh
#[ -e /etc/D ] && echo "/etc/profile.ksh: \$-='$-'" 1>&2

_ETC_PROFILE_KSH=1

PATH="/bin:/usr/bin"

[ -d /usr/openwin ] && export OPENWINHOME=/usr/openwin
[ -d /opt/kde ] && export KDEDIR=/opt/kde

umask 022
if [ -x /usr/bin/id ]; then
  [ `id -gn` = `id -un` ] && [ `id -u` != 0 ] && umask 002
  [ -z "$UID" ] && UID=`id -u`
  [ -z "$USER" ] && USER=`id -un`
fi

_p="" # has to have a trailing ':'!
[ -d $HOME/bin ]        && _p=${_p}$HOME/bin:
[ "$UID" = 0 ]          && _p=${_p}/sbin:/usr/sbin:

_q=$PATH # must *not* have a trailing ':'!
[ -d /opt/bin ]         && _q=${_q}:/opt/bin
[ -d /usr/local/bin ]   && _q=${_q}:/usr/local/bin
[ -d /usr/TeX/bin ]     && _q=${_q}:/usr/TeX/bin
[ -d /usr/X11R6/bin ]   && _q=${_q}:/usr/X11R6/bin
[ -n "$KDEDIR" ]&&[ -d $KDEDIR/bin ]            && _q=${_q}:$KDEDIR/bin
[ -n "$OPENWINHOME" ]&&[ -d $OPENWINHOME/bin ]  && _q=${_q}:$OPENWINHOME/bin

# if paranoid, remove trailing colon ...
PATH=${_p}${_q}:
unset _p _q

[ -z "$LOGNAME" ] && LOGNAME=$USER

export PATH USER LOGNAME

OS=${OS:-`uname -s`}
ARCH=${ARCH:-`uname -m`}
HOSTNAME=`/bin/hostname`
MAIL="/var/spool/mail/$USER"
MAILPATH=$MAIL:/etc/motd

export OS ARCH HOSTNAME MAIL MAILPATH


case "$_INIT_" in
*env*) ;;
*)  # do these once
  _INIT_="$_INIT_"env
  export _INIT_

  if [ -t 1 ]; then
    # we are interactive
    TTY=`tty`
    TTY=`basename $TTY`
    if [ -f /etc/organization ]; then
      ORGANIZATION="`cat /etc/organization`"
      COPYRIGHT="Copyright (c) `date +19%y` $ORGANIZATION"
    fi

    case $UID in
    0) PS1S='# ';;
    esac
    PS1S=${PS1S:-'$ '}
    PROMPT="<$LOGNAME@$HOSTNAME:!>$PS1S"
  else
    TTY=none
  fi

  export ORGANIZATION COPYRIGHT TTY

  # test if we are a Korn shell
  if [ -n "KSH_VERSION" ]; then
    SHELL=/bin/ksh
    ENV=${HOME%/}/.kshrc
    if [ ! -f $ENV ]; then
      ENV=/etc/ksh.kshrc
    fi
    HISTFILE=${HOME%/}/.ksh_hist
    HISTSIZE=100
    PROMPT="<$LOGNAME@$HOSTNAME:!>$PS1S"
    CDPATH=.
    set -o emacs
  else
    SHELL=/bin/sh
  fi
  PS1=$PROMPT

  export SHELL PS1 PROMPT CDPATH HISTFILE HISTSIZE ENV
;;
esac
