/* This file is part of the KDE project
   Copyright (C) 1998, 1999 Simon Hausmann <hausmann@kde.org>
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
 
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.
 
   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#pragma interface

#ifndef __kactivator_h__
#define __kactivator_h__

#include "kded.h"
#include <qobject.h>
#include <qstring.h>

class QStringList;
class KServiceEntry;
class KServiceFactory;

/**
 * Abstract interface
 */
class KActivator
{
public:

  /**
   * Will register service at the IMR
   */
  virtual bool registerService( const QString &name, const QString &activationMode,
                                const QStringList &repoIds, const QString &exec ) = 0L;

  /**
   * Will unregister service at the IMR
   */  
  virtual bool unregisterService( const QString &name ) = 0L;

  /**
   * Will create a virtual reference for a server listed in the IMR
   */
  virtual CORBA::Object_ptr activateService( const QString &name, const QString &repoId, const QString &tag = QString::null ) = 0L;

};

class KLocalActivator : public QObject, virtual public POA_PortableServer::ServantActivator, public KActivator
{
  Q_OBJECT
public:
  KLocalActivator( CORBA::ORB_ptr orb, PortableServer::POA_ptr poa, KServiceFactory *serviceFactory );
  virtual ~KLocalActivator() {};
  
  virtual bool registerService( const QString &name, const QString &activationMode,
                                const QStringList &repoIds, const QString &exec );
  
  virtual bool unregisterService( const QString &name );

  virtual CORBA::Object_ptr activateService( const QString &name, const QString &repoId, const QString &tag = QString::null );

  virtual PortableServer::Servant incarnate( const PortableServer::ObjectId &,
				     PortableServer::POA_ptr );
				     
  virtual void etherealize( const PortableServer::ObjectId & oid,
		    PortableServer::POA_ptr poa,
		    PortableServer::Servant serv,
		    CORBA::Boolean cleanup_in_progress,
		    CORBA::Boolean remaining_activations);

private slots:
  void regEntryAdded( KServiceEntry *entry );
  void regEntryRemoved( KServiceEntry *entry );

private:
  CORBA::Object_ptr activate( const char *_implreponame, const char *_repoid, const char *_tag );

  CORBA::ORB_var m_vORB;
  CORBA::ImplRepository_var m_vIMR;
  PortableServer::POA_var m_vPOA;  
};

class KRemoteActivator : public KActivator
{
public:
  KRemoteActivator( KDED::Activator_ptr activator );
  virtual ~KRemoteActivator() {}
  
  virtual bool registerService( const QString &name, const QString &activationMode,
                                const QStringList &repoIds, const QString &exec );

  virtual bool unregisterService( const QString &name );

  virtual CORBA::Object_ptr activateService( const QString &name, const QString &repoId, const QString &tag = QString::null );
  
private:
  KDED::Activator_var m_vActivator;
};

#endif
