/* This file is part of the KDE project
   Copyright (C) 1999 Steffen Hansen <hansen@kde.org>
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
 
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.
 
   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#pragma interface

#ifndef KDED_IMPL_H
#define KDED_IMPL_H

#include "kded.h"

class ServiceStarter;
class KLocalActivator;
class KLocalTrader;
class KLocalNaming;
class KServiceFactory;

class KDED_Impl : virtual public POA_KDED::InitialReferences {
public:
     KDED_Impl( CORBA::ORB_ptr);
     virtual ~KDED_Impl();

     KDED::ObjectIdList*     list_initial_services();
     CORBA::Object_ptr       resolve_initial_references( const char *id);
     KDED::Trader_ptr        ktradingService();
     KDED::Activator_ptr     kactivatorService();
     KDED::Naming_ptr        knamingService();
protected:
     CORBA::ORB_var                orb;
     KDED::ObjectIdList            ids;
};

class KTrader_Impl : virtual public POA_KDED::Trader
{
public:
  KTrader_Impl();
  virtual ~KTrader_Impl();
  
  KDED::Trader::StreamedOfferListData *fastQuery( const CORBA::WChar *serviceType,
                                                  const CORBA::WChar *constraint,
				                  const CORBA::WChar *preferences );

  KDED::Trader::OfferList *query( const CORBA::WChar *serviceType,
				  const CORBA::WChar *constraint,
				  const CORBA::WChar *preferences );

  KDED::Trader::StreamedOfferListData *fastListServices();
  
  KDED::Trader::OfferList *listServices();

  KDED::Trader::StreamedOfferListData *fastServiceByName( const CORBA::WChar *name );
  
  KDED::Trader::CService *serviceByName( const CORBA::WChar *name );

  KLocalTrader *localTrader() const { return m_pLocalTrader; }

  static KTrader_Impl *self() { return s_pTraderImpl; }

private:
  KLocalTrader *m_pLocalTrader;
  static KTrader_Impl *s_pTraderImpl;
};

class KActivator_Impl : virtual public POA_KDED::Activator
{
public:
  KActivator_Impl( CORBA::ORB_ptr orb, PortableServer::POA_ptr poa, KServiceFactory *factory );
  virtual ~KActivator_Impl();

  CORBA::Boolean registerService( const char *name, const char *activationMode,
                                  const KDED::Activator::StringList &repoIds, const char *exec );
  
  CORBA::Boolean unregisterService( const char *name );
  
  CORBA::Object_ptr activateService( const char *name, const char *repoId, const char *tag );

  static KActivator_Impl *self() { return s_pActivatorImpl; }

  KLocalActivator *localActivator() { return m_pLocalActivator; }
  
private:
  KLocalActivator *m_pLocalActivator;
  static KActivator_Impl *s_pActivatorImpl;
};

class KNaming_Impl : virtual public POA_KDED::Naming
{
public:
  KNaming_Impl();
  virtual ~KNaming_Impl();
  
  CORBA::Object_ptr resolve( const CORBA::WChar *name );
  
  CORBA::Boolean bind( const CORBA::WChar *name, CORBA::Object_ptr obj );
  
  CORBA::Boolean unbind( const CORBA::WChar *name );
  
  KDED::Naming::NamingList *list();
  
  static KNaming_Impl *self() { return s_pNamingImpl; }
  
private:
  KLocalNaming *m_pLocalNaming;
  static KNaming_Impl *s_pNamingImpl;
};

#endif /* KDED_IMPL_H */
/*
 * Local variables:
 * mode: c++
 * c-file-style: "k&r"
 * End:
 */
