/* This file is part of the KDE project
   Copyright (C) 1999 Steffen Hansen <hansen@kde.org>
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
 
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.
 
   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#pragma interface

#ifndef KDED_INSTANCE_H
#define KDED_INSTANCE_H

#include <qstring.h>
#include <kded.h>

// Forward declarations:
class KTrader;
class KActivator;
class KNaming;
class KRemoteTrader;
class KRemoteActivator;
class KRemoteNaming;
class KService;
/* Forward declaretions with namespaces _require_ namespace support 
 * in the compiler, sigh.
 */
/*
namespace CORBA {
  class Object;
  class ORB;
  typedef Object* Object_ptr;
  typedef ORB* ORB_ptr;
};
*/


/**
 * Create an instance of this class if you want to use the kded. 
 * Only one instance of KdedInstance may exist per application.
 */
class KdedInstance {
public:
     /**
      * Pass the commandline argc/argv to the KdedInstance constructor.
      * The following options are understood by kded and will be
      * removed from the commandline before the constructor returns.
      *
      * -kdedior ior:  Set reference to KDED::InitialServices
      * -kdedlocal  :  Make local trader
      * -kdedremote :  If kded is needed, it will be search on the 
      *                root window of the display. If it isn't found, 
      *                a new kded will be started. This is the default
      *                behavior, and the option is only provided to
      *                override -kdedlocal.
      *
      */
     KdedInstance( int &argc, char **argv, CORBA::ORB_ptr _orb_ptr);

     virtual ~KdedInstance();
     
     /**
      * Returns true if we are succesfully connected to a kded.
      * If it returns false, it is OK to try to create a new 
      * KdedInstance.
      */
     bool isKdedRunning() { return m_is_kded_running;}
     
     /**
      * Returns true if we are running a local kded.
      */
     bool isLocalKded() { return m_use_local_kded;}
     
     /**
      * Returns a stringified IOR to kded. Will be QString::null
      * if an error has occured.
      */
     QString kdedIor() { return m_kdedior;}

     /**
      * Returns a pointer to a KTrader.
      *
      * (Note: <b>Never</b> delete this pointer)
      */
     KTrader* ktrader();
     
     /**
      * Returns a point to a KActivator
      *
      * (Note: <b>Never</b> delete this pointer)
      */
     KActivator *kactivator();

     /**
      * Returns a point to a KNaming
      *
      * (Note: <b>Never</b> delete this pointer)
      */
     KNaming *knaming();

     static KdedInstance *self() { return s_self; }

protected:
     bool                 initializeKded();

     bool                 m_use_local_kded;
     bool                 m_is_kded_running;
     QString              m_kdedior;
     CORBA::ORB_ptr       m_orb_ptr;

     static KdedInstance *s_self;
     
     static KRemoteTrader *s_pRemoteTrader;
     static KRemoteActivator *s_pRemoteActivator;
     static KRemoteNaming *s_pRemoteNaming;
};

#endif

/*
 * Local variables:
 * mode: c++
 * c-file-style: "k&r"
 * End:
 */
