/* This file is part of the KDE project
   Copyright (C) 1998, 1999 Simon Hausmann <hausmann@kde.org>
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
 
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.
 
   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#pragma interface

#ifndef __knaming_h__
#define __knaming_h__

#include "kded.h"

#include <qstring.h>
#include <qvaluelist.h>
#include <qmap.h>

class KNaming
{
public:

  struct NamingEntry
  {
    QString name;
    CORBA::Object_var obj;
  };
  typedef QValueList<NamingEntry> NamingList;

  /**
   * Tries to resolve the given name to a bound object reference
   * and returns nil if there's no such binding.
   */
  virtual CORBA::Object_ptr resolve( const QString &name ) = 0L;
  
  /**
   * Binds the specified name with the specified object in the global naming
   * service.
   */
  virtual bool bind( const QString &name, CORBA::Object_ptr obj ) = 0L;

  /**
   * Removes the object binding for the specified name.
   */
  virtual bool unbind( const QString &name ) = 0L;
  
  /**
   * Lists all registered entries in the naming service.
   */
  virtual NamingList list() = 0L;
};

class KLocalNaming : public KNaming
{
public:
  KLocalNaming();
  virtual ~KLocalNaming();

  virtual CORBA::Object_ptr resolve( const QString &name );
  
  virtual bool bind( const QString &name, CORBA::Object_ptr obj );

  virtual bool unbind( const QString &name );

  virtual NamingList list();

private:
  QMap<QString,CORBA::Object_var> m_mapNaming;
};

class KRemoteNaming : public KNaming
{
public:
  KRemoteNaming( KDED::Naming_ptr naming );
  virtual ~KRemoteNaming();

  virtual CORBA::Object_ptr resolve( const QString &name );
  
  virtual bool bind( const QString &name, CORBA::Object_ptr obj );

  virtual bool unbind( const QString &name );
  
  virtual NamingList list();
  
private:
  KDED::Naming_var m_vNaming;
};

#endif

