/* This file is part of the KDE project
   Copyright (C) 1998, 1999 Torben Weis <weis@kde.org>
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
 
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.
 
   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/     

#ifndef __kom__plugin_h__
#define __kom__plugin_h__

#include "kom.h"
#include "komComponent.h"

/**
 *  class KOMPluginFactory.
 *
 *  You have to implement these on your own!
 *  The following methods have to be implemented:
 *
 *  @li virtual ::KOM::Plugin_ptr create( Component_ptr core );
 *
 *  @short class KOMPluginFactory.
 */
class KOMPluginFactory : virtual public KOMComponentFactory,
			 virtual public KOM::PluginFactory_skel
{

public:

};

/**
 *  Plugins are a kind of counterpart to aggregates. They do not extend the functionality
 *  of a component by providing new interfaces, but instead usually change the behaviour
 *  of it, by
 *
 *  @li doing things like installing  event filters to the object, etc.
 *  @li providing special plugin interfaces
 *
 *  @short Changes the implementetion of the functionality of a component
 *
 *  @see KOMAggregate
 */
class KOMPlugin : virtual public KOMComponent,
		  virtual public KOM::Plugin_skel
{
public:

  /**
   *  Destructor.
   */
  virtual ~KOMPlugin();

  // C++
  /**
   *  This method cleans up memory when the object will be deleted. 
   */
  virtual void cleanUp();

  // IDL
  /**
   *  Retrieves the parent of the plugin.
   *
   *  @return A pointer to the parent KOM::Component object.
   */
  virtual KOM::Component_ptr parent();
  
protected:

  /**
   *  Constructor.
   *
   *  @param _com Pointer to a KOM::Component object that will be the parent of the plugin.
   */
  KOMPlugin( KOM::Component_ptr _com );
  
  KOM::Component_var m_vParent;
};

#endif
