// -*- c++ -*-
/*
 *  Qt support for MICO
 *  Copyright (C) 1997 Lars Doelle
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Send comments and/or bug reports to:
 *                 mico@informatik.uni-frankfurt.de
 */

#ifndef __op_app_h__
#define __op_app_h__

/*
 * XXX ugly hack to make QApplication::quit_now accessible
 * to OPApplication
 */
#define private protected
#include <qapplication.h>
#undef private

#include <kom.h>
#include <komApplication.h>
#include "openparts.h"

#include <CORBA.h>

#define opapp     OPApplication::getOPApplication()
#define opapp_orb OPApplication::getOPApplication()->orb()
#define opapp_boa OPApplication::getOPApplication()->boa()

// Is this needed ? QTDispatcher is declared in komApplication.h.
class QTDispatcher;

/**
 *  The OPApplication class.
 */
class OPApplication : public KOMApplication
{
  friend QTDispatcher;

  Q_OBJECT

public:

  /**
   *  Constructor.
   *
   *  @param _argc     Number of the command-line arguments
   *  @param _argv     Array of char* filled with the command-line argument strings.
   *  @param _rAppName Name of the application.
   */
  OPApplication( int& _argc, char** _argv, const QString& _rAppName = QString::null );

  /**
   *  Destruktor
   */
  ~OPApplication();

  /**
   *  Retrieves a pointer to an OPApplication object. In an application only one instance
   *  of this class can exist.
   *
   *  @return A pointer to the object instance.
   */
  static OPApplication* getOPApplication()
  { return s_pOPApplication; }

protected:

  /**
   *  Filters all events sent by the X-server.
   *
   *  @param _event The event that will be processed.
   */
  virtual bool x11EventFilter( XEvent* _event );
  
  static OPApplication* s_pOPApplication;
};

/**
 *  The OPApplicationIf class.
 */
class OPApplicationIf : virtual public KOMApplicationIf,
			virtual public OpenParts::Application_skel
{
public:

  /**
   *  Constructor.
   */
  OPApplicationIf();

  /**
   *  Constructor.
   */
  OPApplicationIf( const CORBA::BOA::ReferenceData &refdata );

  /**
   *  Constructor.
   */
  OPApplicationIf( CORBA::Object_ptr _obj );

  /**
   *  CreateDoc.
   *
   *  @return Document-pointer.
   */
  OpenParts::Document_ptr createDoc();
 
  /**
   *  CreatePart.
   *
   *  @return Part-pointer.
   */
  OpenParts::Part_ptr createPart();
 
  /**
   *
   */
  OpenParts::MainWindow_ptr createWindow();
};

#endif // __mico_qt_h__

