#ifndef __OPcolorbar_h__
#define __OPcolorbar_h__

#include <ktoolbar.h>
#include <qbutton.h>
#include <qbrush.h>

#include <qintdict.h>

class OPColorBarIf;

/**
 *  The ColorField class.
 */
class ColorField : public QButton
{
  Q_OBJECT

public:

  /**
   *  Constructor.
   *
   *  @param _id ID  ID of the color field.
   *  @param _brush  Brush for drawing the color field.
   *  @param _parent Parent of the widget.
   *  @param _name   Name of the widget.
   */
  ColorField( int _id, const QBrush& _brush, QWidget* _parent = 0, const char* _name = 0 );

  /**
   *  Highlights the color field.
   *
   *  @param _flag Indicates the color field is highlighted or not.
   */
  void highlight( bool _flag );

  /**
   *  Retrieves the brush of the color field.
   *
   *  @return The brush of the color field.
   */
  const QBrush& brush () const
  { return m_brush; }

protected:

  /**
   *  Draws a color button.
   *
   *  @param _painter The painter object that shall draw the button.
   */
  void drawButton( QPainter* _painter );

  /**
   *  Handles mouse release events.
   *
   *  @param _event The event that shall be handled.
   */
  void mouseReleaseEvent( QMouseEvent* _event );
    
signals:

  /**
   *  This signal will be emited if a foreground color is selected.
   */
  void fgColorSelected( int id );

  /**
   *  This signal will be emited if a background color is selected.
   */
  void bgColorSelected( int id );

private:

  bool m_highlighted;
  int m_id;
  QBrush m_brush;
};

/**
 *  The OPColorBar class.
 */
class OPColorBar : public KToolBar
{
  Q_OBJECT

public:

  /**
   *  Constructor.
   *
   *  @param _parent    Parent of the widget.
   *  @param _name      Name of the widget.
   *  @param _item_size Item size.
   */
  OPColorBar(QWidget* _parent = 0L, const char* _name = 0L, int _item_size = -1 );

  /**
   *  Destructor.
   */
  ~OPColorBar();
  
  /**
   *  Retrieves the interface of the object.
   */
  OPColorBarIf* interface ();

  /**
   *  Inserts a color button into the color bar.
   *
   *  @param _id    .
   *  @param _red   .
   *  @param _green .
   *  @param _blue  .
   *  @param _file  .
   */
  void insertColorButton( int _id, short _red, short _green, short _blue, bool _fill );

  /**
   *  Retrieves the color of a color field with a given ID.
   *
   *  @param _id    ID of the color field.
   *  @param _red   Red value of the color.
   *  @param _green Green value of the color.
   *  @param _blue  Blue value of the color.
   *  @param _fill  Filled value of the color.
   */
  bool getColor( int _id, short& _red, short& _green, short& _blue, unsigned char& _fill );

signals:

  /**
   *  This signal will be emited if a foreground color is selected.
   */
  void fgColorSelected (int id);

  /**
   *  This signal will be emited if a background color is selected.
   */
  void bgColorSelected (int id);

public slots:

  /**
   *  Handles color bar if foreground color had chenged.
   */
  void slotFgColorChanged (int id);

  /**
   *  Handles color bar if background color had chenged.
   */
  void slotBgColorChanged (int id);

protected:

  OPColorBarIf* m_pInterface;
  QIntDict<ColorField> fields;
};

#endif

