/* This file is part of the KDE project
   Copyright (C) 1998, 1999 Torben Weis <weis@kde.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#ifndef __op_frame_h__
#define __op_frame_h__

#include <kom.h>
#include "openparts.h"
#include "opPart.h"

#include <qwidget.h>
#include <qxembed.h>

/**
 *  The OPFrame class.
 */
class OPFrame : public QXEmbed
{

  Q_OBJECT

public:

  /**
   *  Constructor (standard QWidget constructor)
   */
  OPFrame( QWidget* _parent = 0L, const char* _name = 0L );

  /**
   *  The Destructor calls @ref #detach automatically.
   */
  ~OPFrame();

  /**
   *  If 'attach' is called twice, then the first control attached
   *  becomes detached automatically. The controls reference becomes
   *  increased.
   */
  virtual bool attach( OpenParts::Part_ptr _part );

  /**
   *  If 'attachLocal' is called twice, then the first control attached
   *  becomes detached automatically. The controls reference becomes
   *  increased. This function is the optimized version of @ref #attach
   *  which only works if the part and the frame are running in the same
   *  process and if the widget of the part is a direct child of the
   *  frame.
   *
   *  @param _part Part.
   */
  virtual bool attachLocal( OPPartIf* _part );
  
  /**
   * The controls reference becomes decreased.
   */
  virtual void detach();

  /**
   *  Retrieves the part attached to this frame.
   *
   *  @return The part attached to this frame
   */
  virtual OpenParts::Part_ptr part();

protected:

  /**
   *  ResizeEvent.
   *
   *  @param _event Resize-event.
   */
  void resizeEvent( QResizeEvent* _event );

  OPPart_ref m_rPart;
  OPPartIf* m_pPart;
};

#endif

