/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: am_bld.c
   Applix ASC version: 41
   Date: Thu Feb 18 21:12:28 1999
*/

#include <am_bld.h>
#include <signal.h>
#include <sys/stat.h>
#if AXMS
#include <process.h>
#endif
#if !NO_VARARGS
#include <varargs.h>
#endif

typedef struct pretend_info {
    struct pretend_info *next;
    long time;
    char filename[257];
} pretend_info;

char *ModsToBuild[1000];    /* module names to build */
char ModsToForce[1000]; /* modules to force build */
int BuildModsCnt;   /* how many to build */
int CreateDependencies; /* TRUE if should depend scan on compilation */
node *CurrentOutputNode = NULL;
extern char *DirSlash;
extern int CalledSetDir;   /* Sanity check on rules */
extern ID CallUid;
extern int forceBuild;
int AllowDeletes;       /* hack to keep bad output files */
int CmdTokens;      /* number of tokens in last shcmd */
int ForceAllow = FALSE; /* overrides the allow flag in rules */
int NoReportErrMsg; /* Used to briefly keep messages out of report file */
int CProtosGenerated;   /* Set by compilation if protos created */
int CProtosCreated; /* only TRUE if there wasn't already a proto file at all */
int Savedpretend;		/* Remember "noproto" around */

pretend_info *PretendList;  /* files "touched" during pretend */
long time();


extern char ThisCommand[];
extern int KeepGarbage; /* TRUE if should keep broken output files */
extern int NumCallsByIdAllocated;
extern int JustCleanAll;
extern int DoingReport;
extern int TouchBuild;
extern int PretendBuild;
extern int NarrowMode;
extern int LocAutoCache;    /* FALSE if shouldn't put checkdirs in files list*/
extern int CprotoForcing;       /* TRUE if ModsToForce set by -fcp */
extern int RebuildProtoOnAllCompiles;   /* if FALSE, just on srcfile edits */

#define bld_err(body) {ErrorMsg body; goto bld_failed;}


/*************************************************************************
function:  DoBuild - Build any/all components
*/
DoBuild ()
{
    char *name;
    int ix, jx, code;
    node *nodeP;
    ID id;
    int error_count = TotalErrors;
    int cproto_deps_built = FALSE;

	if ( BuildModsCnt==1 && streq(ModsToBuild[0],"nothing") )
		return;

again:      /* only comes here if CProtosGenerated */
	if ( PretendBuild ) {
	    TELL(("Starting prototyping phase...\n"));
	}
	else {
	    TELL(("Starting build...\n"));
	}

    BuildPhase = TRUE;
/*  if ( GetVar("AOS_GEN_DEPENDS") ) */
        CreateDependencies = TRUE;

    if ( JustCleanAll )
        DeleteAll();

    else {
		/* build cproto.tag and then do pretend build (proto type
		   checking) and finally do the actual build of the
                   specified targets */

		if (!cproto_deps_built) {
			VERBOSE(("Building dependencies of cproto.tag\n"));
			Savedpretend = PretendBuild;	   /*19Aug98 rjd Save around BuildNode*/
			PretendBuild = FALSE;
			nodeP = LookupNode("cproto.tag");
			if (!nodeP->ruleid) {
				aborter("Missing rule for %s, exiting\n",nodeP->name);
			}
			BuildNode(NodeIdLooked,TRUE,CprotoForcing);
			VERBOSE(("Done building dependencies of cproto.tag\n"));
			cproto_deps_built = TRUE;
			PretendBuild = Savedpretend;		/* Restore noproto setting	  */
		}
		
		if ( BuildModsCnt || NumAppsToBuild ) {
			if ( BuildModsCnt ) {
				for ( ix = 0; ix < BuildModsCnt; ix++ ) {
					name = ModsToBuild[ix];
					TRACE(("To build: %s\n",name));
					nodeP = LookupNode(name);
					if ( !nodeP->ruleid ) {
						if ( !PretendBuild )
							ErrorMsg("No such output module - %s",nodeP->name);
						continue;
					}
					if ( ModsToForce[ix] ) {
						StatNode(nodeP);
						if ( nodeP->exists )
							nodeP->dtmod = 2;
					}
					BuildNode(NodeIdLooked,TRUE,CprotoForcing);
				}
			}
			if ( NumAppsToBuild ) {
				for ( ix = 0; ix < NumAppsToBuild; ix++ ) {
					code = GetCIdFromName(AppNamesToBuild[ix]);
					if ( code < 0 )
						continue;
					for (id = NextType(0,TYPE_NODE);id;id = NextType(id,TYPE_NODE)){
						nodeP = NodePtr(id);
						if ( OUTPUT_NODE(nodeP) && (NodeCompId(nodeP) == code)  )
							BuildNode(id,FALSE,FALSE);
					}
				}
			}
		}
		else {
			if ( !PretendBuild )
				ErrorMsg("No build target specified");
		}
	}

    if ( PretendBuild ) {
        PretendBuild = FALSE;
        PretendList = NULL;
        if ( CProtosCreated )
            LocAutoCache = TRUE;    /* captures new proto files */
        CleanCache();
        TotalErrors = error_count;
        goto again;
    }
}

#define BN_BUSTED 0
#define BN_OKAY 1
#define BN_NEEDS_BUILT 2

/*************************************************************************
function:  BuildNode - Build a node, may recurse
*/
BuildNode (id,force_allow,force_cprotos)
ID id;
int force_allow;/* if FALSE suppress optional (e.g lint) on targetless makes */
int force_cprotos; /* If true, does cproto stuff even if on exclusion list */
{
    node *nodeP;
    int retval;

    nodeP = NodePtr(id);
    if ( nodeP->resolving ) {
        ErrorMsg("circular on module %s",nodeP->name);
        nodeP->busted = TRUE;
        nodeP->up_to_date = TRUE;
        return(BN_BUSTED);
    }
    nodeP->resolving = TRUE;
    retval = BuildIt(id,force_allow);
    NodePtr(id)->resolving = FALSE;
    return(retval);
}

/*************************************************************************
function: BuildIt - The node builder
*/
BuildIt(id,force_allow,force_cprotos)
ID id;
int force_allow;/* if FALSE suppress optional (e.g lint) on targetless makes */
int force_cprotos; /* If true, does cproto stuff even if on exclusion list */
{
#define INPUTS_LOOP 1
#define DEPENDS_LOOP 2
#define DO_INPUTS (looptype == INPUTS_LOOP)

    long my_dtmod;
    node *nodeP, *childP;
    ID child_id;
    int ix;
    char *dirname;
    int new_dir;
    int buildit = FALSE;
    int errors = FALSE;
    int gave_reason = FALSE;
    int same_file();
    int incount, depcount;
    int input_changed = FALSE;
    char filename[STRPATH+1];
	char innames[200];
	char outnames[200];
	char *rulename, *toolname;

    nodeP = NodePtr(id);
    LOGGER(("\nBuildNode: %s",nodeP->name));
    if ( !nodeP->ruleid ) {
        nodeP->busted = TRUE;
        nodeP->up_to_date = TRUE;
		if ( !PretendBuild )
	        ErrorMsg("don't know how to build %s",NodeName(nodeP));
        return(BN_BUSTED);
    }
    if ( !RuleArray[nodeP->ruleid].allow && !force_allow )
        return(BN_OKAY);    /* doesn't set the up_to_date, good! */
    if ( !nodeP->dtmod )
        StatNode(nodeP);
    my_dtmod = nodeP->dtmod;

    if ( nodeP->up_to_date )
        return(BN_OKAY);
    NodeFilename(nodeP,filename,TRUE);

do_deps_again:
    if ( RuleArray[nodeP->ruleid].dep_srch && !nodeP->deps_done ) {
        FindMyDeps(id);
        nodeP = NodePtr(id);
    }
    incount = NodeInCount(nodeP);
    depcount = NodeDepCount(nodeP);
    for ( ix = 0; ix < (incount + depcount); ix++ ) {
retry_this:
        childP = NodePtr(child_id = NodeInId(nodeP,ix));
        TRACE(("child %s dtmod %ld exists %d ruleid %d\n",childP->name,
                childP->dtmod,childP->exists,childP->ruleid));
        if ( childP->busted || (!childP->dtmod && !StatNode(childP)) ) {
            errors = TRUE;
            continue;
        }
        if ( childP->up_to_date && (childP->dtmod < my_dtmod) )
            continue;
        if ( (ix >= incount) && !childP->exists && !childP->ruleid )
            continue;   /* depends don't need to exist */
        if ( childP->ruleid || !childP->exists ) {
            if ( !BuildNode(child_id,TRUE,FALSE) )
                goto retry_this;
            childP = NodePtr(child_id);
            nodeP = NodePtr(id);
        }
        if ( RuleArray[nodeP->ruleid].dep_srch && !nodeP->deps_done && !PretendBuild )
            goto do_deps_again; /* this node created as side effect of child */
        if ( (childP->dtmod > my_dtmod) || !nodeP->exists ) {
            if ( ix < incount )
                input_changed = TRUE;
            buildit = TRUE;
        }
        buildit |= nodeP->refers_to_moved;
        if ( buildit && Verbose && !gave_reason ) {
            if ( !nodeP->exists ) {
                VERBOSE(("%s needs to be created\n",nodeP->name));
            }
            else if ( nodeP->refers_to_moved ) {
                VERBOSE(("%s refers to a moved file\n",nodeP->name));
            }
            else if ( (nodeP->ruleid != DUMMY_RULE) &&
                      (nodeP->ruleid != TRUE_NOP_RULE) ) {
                VERBOSE(("%s is newer than %s\n",childP->name,nodeP->name));
            }
            gave_reason = TRUE;
        }
    }
    if ( !buildit ) {
        ID *callsP = AddrOfCallVals(nodeP);
        for ( ix = 0; ix < nodeP->callcount; ix++,callsP++ ) {
#if 0
printf("sub %s time %d mine %d\n",CallsById[*callsP]->name,CallsById[*callsP]->time,my_dtmod);
#endif
            if ( *callsP >= CallUid )
                {
                printf("ERROR - *callsP = %d, CallUid = %d\n",*callsP,CallUid );
                printf("   node being process = %s\n", nodeP->name );
                printf("   nodeP->callcount = %d\n", nodeP->callcount );
                printf("   ix = %d\n", ix );
                printf("   exiting\n");
                exit(1);
                }

            if ( CallsById[*callsP]->time >= my_dtmod ) {
                if ( Verbose && !gave_reason ) {
                    VERBOSE(("%s prototype newer than %s\n",CallsById[*callsP]->name,nodeP->name));
                }
                buildit = TRUE;
                break;
            }
        }
    }
    if ( errors && (nodeP->ruleid != TRUE_NOP_RULE) && !nodeP->has_siblings ){
		if ( !PretendBuild )
	        wrn_msg("%s not built because of subordinate errors",nodeP->name);
        nodeP->busted = TRUE;
        nodeP->up_to_date = TRUE;
        if ( !KeepGarbage && (nodeP->ruleid != DUMMY_RULE) 
			&& DeleteNodeFile(nodeP) && !PretendBuild )
            wrn_msg("%s deleted",nodeP->name);
        return(BN_BUSTED);
    }
    if ( !buildit ) {
        nodeP->up_to_date = TRUE;
        return(BN_OKAY);
    }
    if ( nodeP->ruleid == DUMMY_RULE ) {
        long newest;
        newest = nodeP->dtmod;
        for ( ix = 0; ix < (incount + depcount); ix++ )
            newest = MAX(newest,NodePtr(child_id = NodeInId(nodeP,ix))->dtmod);
        nodeP->dtmod = newest;
        nodeP->up_to_date = TRUE;
        nodeP->deps_done = FALSE;
        FindMyDeps(id);
        return(BN_OKAY);
    }
    if ( nodeP->ruleid == TRUE_NOP_RULE ) {  /* true nop for multi-outputs */
        if ( TouchBuild ) {
            ReStatNode(nodeP);
            if ( nodeP->exists )
                TouchFile(filename);
            else {
                nodeP->exists = TRUE;
                nodeP->dtmod = 10;
                nodeP->up_to_date = TRUE;
            }
        }
        else
            ReStatNode(nodeP);
        if ( !nodeP->exists ) {
            nodeP->busted = TRUE;
            nodeP->up_to_date = TRUE;
			if ( !PretendBuild )
	            ErrorMsg("file %s did not come into existance",filename);
            return(BN_BUSTED);
        }
        return(BN_OKAY);
    }
    CurrentOutputNode = nodeP;
    CalledSetDir = FALSE;
    AllowDeletes = TRUE;
	if ( RuleArray[nodeP->ruleid].func != same_file ) {
		if ( !TouchBuild && !PretendBuild
			&& (nodeP->ruleid != TRUE_NOP_RULE) && !nodeP->has_siblings )
			DeleteNodeFile(nodeP);
	
		rulename = RuleArray[nodeP->ruleid].name;
		if ( nodeP->has_siblings )
			sprintf(outnames,"group %s...",nodeP->name);
		else
			strcpy(outnames,nodeP->name);
		
		if ( RuleArray[nodeP->ruleid].tool_based ) {
			toolname = NodeInVal(nodeP,0)->name;
			if ( incount == 2 )
				sprintf(innames," %s",NodeInVal(nodeP,1)->name);
			else
				sprintf(innames,"");
			rulename = toolname;
		}
		else {
			if ( incount == 1 )
				sprintf(innames," %s",NodeInVal(nodeP,0)->name);
			else
				strcpy(innames,"");
		}
	
		if ( !PretendBuild && NarrowMode ) {
		    CMPTELL(("%s%s into %s\n",rulename,innames,outnames));
		}
		else if ( !PretendBuild ) {
		    if ( strlen(NodeWithStr(nodeP)) )
		        CMPTELL(("%s%s into %s with %s\n",rulename,innames,filename,NodeWithStr(nodeP)))
		    else
		        CMPTELL(("%s%s into %s with %s\n",rulename,innames,filename))
		}
	}
    nodeP->refers_to_moved = FALSE;
    if ( RuleArray[nodeP->ruleid].dep_srch ) {
        nodeP->deps_done = FALSE;
        FindMyDeps(id);
        nodeP = NodePtr(id);
    }
    CmdTokens = 0;      /* safety for non-procs */
    CreateNodeDir(nodeP);
    if ( TouchBuild )
        ix = TouchFile(filename);
    else if ( PretendBuild ) {
        ix = TouchFile(filename);
        strcpy(filename, NodeInVal(nodeP,0)->name);
        if ( (input_changed || RebuildProtoOnAllCompiles || forceBuild) &&
             (streq("compile",RuleArray[nodeP->ruleid].name) ||
              streq("shlib_compile",RuleArray[nodeP->ruleid].name) ||
              (streq("shlib_cpp_compile", RuleArray[nodeP->ruleid].name) &&
               streq(".c", filename + strlen(filename) - 2)) ||
              streq("nodebug_compile",RuleArray[nodeP->ruleid].name) ||
              streq("ansi_compile",RuleArray[nodeP->ruleid].name))
             && (CprotoForcing || !NodeInVal(nodeP,0)->cproto_excluded) ) {
			char name[STRPATH+1];
	    	node* cnP= NodeInVal(nodeP,0);
			if (cnP->ruleid > 0 &&
				(streq("elf_compile",RuleArray[cnP->ruleid].name)
				 || streq("elf_portable",RuleArray[cnP->ruleid].name)))
			{
				/* for elf generated c files exclude from cproto */
				cnP->cproto_excluded = TRUE;
			}
            else {
			    CMPTELL(("prototyping %s for %s\n",
                      NodeInVal(nodeP,0)->name,nodeP->name));
                GenCProtoFiles(id);
                nodeP = NodePtr(id);
            }
        }
    }
    else
        ix = (*RuleArray[nodeP->ruleid].func)(nodeP);
    CurrentOutputNode = NULL;
    assert(CalledSetDir);   /* verify rule set the output directory */
    if ( ix != 0 ) {
        char err_buf[256];
        nodeP->busted = TRUE;
        nodeP->up_to_date = TRUE;
        if ( DoingReport )
            DoReport(nodeP,ThisCommand,ix);
        if ( !Verbose && !DoingReport )
            wrn_msg("*** cmd: %s",ThisCommand);
        strcpy(err_buf,"Error code %d");
        if ( !KeepGarbage && AllowDeletes 
			&& (nodeP->ruleid != TRUE_NOP_RULE) && !nodeP->has_siblings
			&& DeleteNodeFile(nodeP) )
            strcat(err_buf,", %s deleted");
        NoReportErrMsg = TRUE;
        ErrorMsg(err_buf,ix,nodeP->name);
        NoReportErrMsg = FALSE;
        goto bld_failed;
    }
    if ( DoingReport ) {
        unlink(StdErrFile);
        unlink(StdOutFile);
    }
    nodeP->up_to_date = TRUE;
    ReStatNode(nodeP);
    if ( !nodeP->exists && !PretendBuild )
        bld_err(("Module %s not created",nodeP->name));
    return(BN_OKAY);

bld_failed:
    if ( DoingReport ) {
        unlink(StdErrFile);
        unlink(StdOutFile);
    }
    return(FALSE);
}


#if SYS&VMSSYS
extern int childStatus, childTerm, lookPid;    /* process procing */
extern char **AMenviron;
#include <ssdef.h>
#include <stsdef.h>
#define IO_NORMAL SS$_NORMAL | STS$M_INHIB_MSG

static void handle_ast_term (child_info)
struct {
    int pid;
    int completion_status;
}   *child_info;
{
    if ((*child_info).pid == lookPid) {
        childStatus= (*child_info).completion_status;
        /* Fake this out for VMS. See acpp code for
            VMS termination */
        if (childStatus == IO_NORMAL)
            childStatus = 0;
        childTerm = TRUE;
    }
    free (child_info);
}
#endif

/* The following has been pulled from shcmd.c: */

/* This procedure starts a child process, waits for its termination and
returns the exit code with which the process terminated.  If the child
is never started, aborts, terminates due to a signal, or otherwise dies
without executing an exit() statement, this procedure returns 0177. */

int DefineForeignCommand;

int  proc ( cmd, vms_define_foreign )
char *cmd;
int vms_define_foreign;
{
    register int rtnv;

    DefineForeignCommand = vms_define_foreign;

    rtnv = shcmd( cmd ) & 0177777; /*only 16 bits are defined*/
    if( (rtnv & 0377) == 0 ) /* then child exited and exit code is
                            in upper byte. */
        return( rtnv >> 8 );

    /* If it is 0177 then either the child stopped (whatever that means) or
    the shcmd() routine failed and explicitly returned this value.  Don't do
    any further interpretation.
    If not 0177 then, the child terminated due to a signal; treat this as a
    serious error. */

    return( 0177 );
}

int shcmd (cmd)
char *cmd;
{
    return(zshcmd(cmd,(char*)NULL,FALSE));
}

#define DOS_SLASH '\\'
#define UNIX_SLASH '/'

int zshcmd(cmd,outfile,emit_stderr)
char *cmd;
char *outfile;
int emit_stderr;
/*
 * Issue a shell command and return its status
 * 127 is returned if the command couldn't be exec'ed.
 * Double quoted args are passed as a single string after
 * the quotes are removed to make sh -c "string" work
 */
{
    register int num_args = 0;
    register char *p;
    char *alced;
    int status;
    int (*sfunc)();
    char *argv[2000];
    char command_name[256];
    int jpid, ipid;
    int toUnix=FALSE;

    p = cmd = alced = AllcStr(cmd);
    CleanWhiteSpace(cmd);
    if ( CommandLog )
        VerboseLog("%s\n",cmd);
    while ( *p ) {
        if ( *p == ' ' ) {
            while ( *p == ' ' )
                *p++ = 0;
#if AXMS
/* on NT, c-shell scripts need to be started as "csh -c foo.csh"
 * and perl scripts need to be started as "perl foo.pl"
 */
            if ( num_args == 0 ) {
                if ( strstr(cmd, ".bsh") ) {
                    argv[num_args++] = "sh";
					toUnix = TRUE;
                }
                else if ( strstr(cmd, ".csh") ) {
                    argv[0] = "csh";
                    argv[1] = "-c";
                    num_args = 2;
					toUnix = TRUE;
                }
                else if ( strstr(cmd, ".pl") ) {
                    argv[num_args++] = "perl";
					toUnix = TRUE;
                }    
            }
#endif
            argv[num_args++] = cmd;
            cmd = p;
        }
        else if (toUnix && *p == DOS_SLASH)
			*p++ = UNIX_SLASH;
		else
            p++;
    }
    argv[num_args++] = cmd;
    argv[num_args] = 0;
    CmdTokens = num_args;
	if ( argv[0][0] != *DirSlash && argv[0][1] != ':' ) {
    	SelectProg(command_name,argv[0]);
	}
	else {
		strcpy(command_name, argv[0]);
	}
    argv[0] = command_name;
    TRACE(("command: %s\n",command_name));
    if ( !StdOutFile[0] )
        GetTemp("AMSO",StdOutFile);
    if ( !StdErrFile[0] )
        GetTemp("AMSE",StdErrFile);
#if SYS&VMSSYS
    childStatus = 0;
    childTerm = 0;
    lookPid = vms_fork_exec (command_name, "./", argv, AMenviron,
                    handle_ast_term);
    if (lookPid == -1) {
        childTerm = 1;
        childStatus = 1;
    }
    while (childTerm == 0) {
        sleep (1);
    }

    return (childStatus);
#else

#if AXMS
    status = spawnvp(P_WAIT,command_name,argv);
    return(status);
#else

    if ((ipid = FORK()) == 0) {
        if ( outfile || DoingReport ) {
            if ( DoingReport ) {
                freopen(StdOutFile,"a",stdout); /* 'a' for muliple proc rules */
                freopen(StdErrFile,"a",stderr);
            }
            else {
                freopen(StdOutFile,"w",stdout);
                freopen(StdErrFile,"w",stderr);
            }
        }
        execvp(command_name, argv);
        ErrorMsg("%s not found",argv[0]);
        _exit(127);
    }

    /* wait for exec'ed process to finish */
    do {
        jpid = wait(&status);
    } while (jpid != ipid && jpid > 0);

#endif  /* not MS */
#endif  /* not VMS */
    if ( outfile ) {
        int dst, out, err;
        int got;
        char buf[8096];

        VERBOSE(("creating stdout redirection file %s\n",outfile));
        dst = creat(outfile,0666);
        if ( dst == -1 ) {
            ErrorMsg("could not create %s",outfile);
            goto wrap_up;
        }
        err = open(StdErrFile,0);
        fflush(stderr);
        while ( got = read(err,buf,8096) ) {
            if ( emit_stderr ) {
                write(fileno(stderr),buf,got);
                status = 3;
            }
            write(dst,buf,got);
        }
        close(err);
        out = open(StdOutFile,0);
        while ( got = read(out,buf,8096) )
            write(dst,buf,got);
        close(out);
        close(dst);
wrap_up:
        unlink(StdErrFile);
        unlink(StdOutFile);
    }
    else if ( DoingReport && ((status &0xFFFF) == 0) ) {
        int got, out, err;    /* SAFETY NET: report stdout/err on successes */
        char buf[8096];

        err = open(StdErrFile,0);
        fflush(stderr);
        while ( got = read(err,buf,8096) ) write(fileno(stderr),buf,got);
        close(err);
        out = open(StdOutFile,0);
        while ( got = read(out,buf,8096) ) write(fileno(stdout),buf,got);
        close(out);
    }

    free(alced);
    return (status);
}

/*
 * function:  VpShcmd - Do shell command, getting arg vector rather than string
*/
VpShcmd (argv)
char *argv[];       /* null term'd vector */
{
    int rtnv;
    int (*sfunc)();
    char command_name[256];
    int jpid, ipid;
    int ix;

#if 0
    VERBOSE(("Starting cpp..\n"));
    for ( ix = 0; argv[ix]; ix++ ) {
        VERBOSE(("  %s\n",argv[ix]));
    }
#endif
    SelectProg(command_name,argv[0]);
    argv[0] = command_name;
    TRACE(("command: %s\n",command_name));
#if SYS&VMSSYS
    childStatus = 0;
    childTerm = 0;
    lookPid = vms_fork_exec (command_name, "./", argv, AMenviron,
                    handle_ast_term);
    if (lookPid == -1) {
        childTerm = 1;
        childStatus = 1;
    }

    while (childTerm == 0) {
        sleep (1);
    }

    return (childStatus);
#else

#if AXMS
/* on NT, c-shell scripts need to be started as "csh -c foo.csh"
 * and perl scripts need to be started as "perl foo.pl"
 */
	{ 
	char cmd[1000];
    if ( strstr(command_name, ".csh") )
        sprintf(cmd, "csh -c %s", command_name);
    else if ( strstr(command_name, ".pl") )
        sprintf(cmd, "perl %s", command_name);
    else
        strcpy(cmd, command_name);
    rtnv = spawnvp(P_WAIT,cmd,argv);
    return(rtnv);
	}
#else

    if ((ipid = FORK()) == 0) {
        execvp(command_name, argv);
        ErrorMsg("%s not found",argv[0]);
        _exit(127);
    }

    /* wait for exec'ed process to finish */
    do {
        jpid = wait(&rtnv);
    } while (jpid != ipid && jpid > 0);

    rtnv &= 0177777; /*only 16 bits are defined*/

    if( (rtnv & 0377) == 0 ) /* then child exited and exit code is
                            in upper byte. */
        return( rtnv >> 8 );

    /* If it is 0177 then either the child stopped (whatever that means) or
    the shcmd() routine failed and explicitly returned this value.  Don't do
    any further interpretation.
    If not 0177 then, the child terminated due to a signal; treat this as a
    serious error. */

    return( 0177 );
#endif  /* not MS */
#endif  /* not VMS */
}

/*************************************************************************
function: PipeProc - Fork, redirect output and proc
*/
PipeProc (cmd,outfile)
char *cmd;      /* to exec */
char *outfile;  /* redirection file */
{
    register int rtnv;

    rtnv = zshcmd( cmd,outfile,FALSE ) & 0177777; /*only 16 bits are defined*/
    if( (rtnv & 0377) == 0 )
        return( rtnv >> 8 );
    return( 0177 );
}

#if NO_VARARGS
/*
 * function:  wrn_msg - Send warning msg
*/
wrn_msg(a,b,c,d,e,f,g,h,i,j,k)
char *a;
int b,c,d,e,f,g,h,i,j,k;
{
    fprintf(stdout,a,b,c,d,e,f,g,h,i,j,k);
    fprintf(stdout,"\n");
}
#else
int wrn_msg(fmt, va_alist)
  char *fmt;
  va_dcl
{
  va_list args;

  va_start(args);
  vfprintf(stdout, fmt, args);
  va_end(args);
  fprintf(stdout,"\n");
  fflush(stdout);
 }
#endif

/*************************************************************************
function: UpgradeToPretendTime - Adjust status buffer if file Pretend touched
*/
UpgradeToPretendTime (name,status)
char *name;
struct stat *status;
{
    pretend_info *p;

    for ( p = PretendList; p; p = p->next ) {
        if ( streq(p->filename,name) )
            status->st_mtime = p->time;
    }
}

/*************************************************************************
function: TouchFile - Update datestamp on a file
*/
TouchFile (filename)
char *filename;
{
    int fd;
	int ret;
    FILE *fp;
    char buf[2];
    pretend_info *p;

    CalledSetDir = TRUE;

    if ( PretendBuild ) {       /* note for later dtmod check */
        p = (pretend_info*)malloc(sizeof(pretend_info));
        p->next = PretendList;
        p->time = time(0L);
        strcpy(p->filename,filename);
        PretendList = p;
        return(0);
    }
    else {
		ret = utime(filename,NULL);
		if ( ret == -1 ) {
	        fd = open(filename,O_RDWR+O_CREAT,0);
	        if ( fd == -1 )
	            return(1);
	        if ( 1 != read(fd,buf,1) )
	            return(1);
	        if ( -1 == lseek(fd,0L,0) )
	            return(1);
	        write(fd,buf,1);
	        close(fd);
	        return(0);
		}
        return(0);
    }
}

