/* Amake cache dumper and opensrc tree generator */

/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: am_cdump.c
   Applix ASC version: 34
   Date: Thu May 20 21:13:16 1999
*/

#include <am_bld.h>
#include <signal.h>
#if !NO_VARARGS
#include <varargs.h>
#endif
#include <time.h>

char *strdup(), *strrchr(), *strstr(), *strchr();
char *readFileIntoBuf();
extern char *MyBuildTree;
extern char ZZtemp[STRPATH+1];
static void addLine();
extern char *RenameDirs[]; /* opensrc renames */
extern int NumRenameDirs;

static FILE *DumpFp;

/*************************************************************************
function: DumpMemory - Dump the state of the object space
*/
DumpMemory (filename)
char *filename;
{
    DumpFp = fopen(filename,"w");
    if ( DumpFp == NULL ) {
        ErrorMsg("could not open %s",filename);
        return;
    }
    TELL(("Starting Dump to %s\n",filename));
    DumpCml();
    DumpCtrlFiles();
    DumpComps();
    DumpDirs();
    DumpNodes();
    fclose(DumpFp);
}

#if 0
#define PUT(body) (printf body, printf("\n"))
#else
#define PUT(body) Dumper body
#endif

#if NO_VARARGS
/*************************************************************************
function: Dumper - File io
*/
Dumper (str,a,b,c,d,e,f)
char *str;
int a,b,c,d,e,f;
{
    fprintf(DumpFp,str,a,b,c,d,e,f);
    fprintf(DumpFp,"\n");
}
#else
 int
 Dumper(fmt, va_alist)
 char *fmt;
 va_dcl
 {
	 va_list args;

	 va_start(args);
	 vfprintf(DumpFp, fmt, args);
	 va_end(args);
	 fprintf(DumpFp,"\n");
	 fflush(DumpFp);
 }
#endif

/*************************************************************************
function: DumpCml - Dump the command line -D's
*/
DumpCml ()
{
    char *p;
    ID id;

    PUT((""));
    PUT(("******************* COMMAND LINE ************************"));
    PUT((""));
    for ( id = NextType(0,TYPE_CMLVAR); id; id = NextType(id,TYPE_CMLVAR) ) {
        p = CmlVarPtr(id);
        PUT(("-D%s",p));
    }
}

/*************************************************************************
function: DumpCtrlFiles - Dump the names of the control files
*/
DumpCtrlFiles ()
{
    ctrlfile_info *p;
    ID id;

    PUT((""));
    PUT(("******************* CONTROL FILES ***********************"));
    PUT((""));
    for ( id = NextType(0,TYPE_CTRLFILE); id; id = NextType(id,TYPE_CTRLFILE)){
        p = CtrlFilePtr(id);
        PUT(("%s",p->name));
    }
}

/*************************************************************************
function: DumpComps - Dump component info
*/
DumpComps ()
{
    comp_info *p;
    ID id;

    PUT((""));
    PUT(("******************* COMPONENTS **************************"));
    PUT((""));
    for ( id = NextType(0,TYPE_COMP); id; id = NextType(id,TYPE_COMP)){
        p = CompPtr(id);
        PUT(("cname:        %s",p->cname));
        PUT(("default dir:  %s",p->dflt ? DirPtr(p->dflt)->dname : "NONE" ));
        PUT((""));
    }
}

/*************************************************************************
function: DumpNodes - Dump all nodes
*/
DumpNodes ()
{
	node *p;
	ID id;
	char name[STRPATH+1];
	int ix;
	ID compid;

	PUT((""));
	PUT(("******************* MODULES *****************************"));
	PUT((""));
	for ( id = NextType(0,TYPE_NODE); id; id = NextType(id,TYPE_NODE)){
		p = NodePtr(id);
		compid = NodeCompId(p);
		NodeFilename(p,name,TRUE);
		PUT(("name:	 %s (%d)",name,id));
		PUT(("comp:	 %s",compid ? GetCompStr(compid) :"global"));
		if ( p->ruleid ) {
				PUT(("rule:	 %s",RuleArray[p->ruleid].name));
			if ( p->ruleid && strlen(NodeWithStr(p)) )
				PUT(("with:	 %s",NodeWithStr(p)));
			for ( ix = 0; ix < NodeInCount(p); ix++ ) {
				NodeFilename(NodeInVal(p,ix),name,TRUE);
				PUT(("input:	%s",name));
			}
			for ( ix = 0; ix < NodeDepCount(p); ix++ ) {
				NodeFilename(NodeDepVal(p,ix),name,TRUE);
				PUT(("depends:  %s",name));
			}
			for ( ix = 0; ix < p->callcount; ix++ ) {
				PUT(("calls:  %s",CallsById[NodeCallVal(p,ix)]->name));
			}
		}
		PUT((""));
	}
}

/*************************************************************************
function: DumpDirs - Dump dir objects
*/
DumpDirs ()
{
    dir_info *p;
    ID id;
    int ix;

    PUT((""));
    PUT(("******************* DIRECTORIES USED ********************"));
    PUT((""));
    for ( id = NextType(0,TYPE_DIR); id; id = NextType(id,TYPE_DIR) ) {
        p = DirPtr(id);
        PUT(("dname:    %s",p->dname));
        PUT(("owner:    %s",p->owner ? GetCompStr(p->owner) : "global"));
        if ( p->is_locdir )
            PUT(("is a locdir"));
        if ( p->is_checkdir ) {
            PUT(("is a checkdir"));
            PUT(("priority:     %d",p->priority));
        }
        PUT((""));
    }
}

extern ID NodeIdLooked;            /* last node found/created by lookupnode */
extern ID DirNameMap[MAX_LOGICAL_DIRS];
extern char *LogicalDirNames[MAX_LOGICAL_DIRS];
extern int NumLogNames;

static FILE *srcfp, *amkfp;
static int genAmkCompId;
static void GenAmkNode(ID id, char *my_reason);
static void GenAmk (ID id);
static void putInPublicDir (node *nodeP, char *publicdir);
static char *genDepName(node *nodeP);
static int genWithName(node *nodeP);
static char last_reason[300];
static int numWithStrs;
static char *withStrs[1000];
static char srcfile[STRPATH+1], amkfile[STRPATH+1];
static SkippedFiles;

typedef struct deplist_t {
	int count;
	char *name;
	ID ids[1];
} deplist_t;

static int numDepLists;
static deplist_t *depLists[1000];

/*************************************************************************
function: CopyIfChanged - Update dst file if different from src file
Each arg is a full pathname.
*/
void CopyIfChanged (char *srcname, char *dstname) {
	char *fname, *dirname;
	char buf[STRPATH+1];

	if ( !FilesIdentical(srcname,dstname) ) {
		fname = strrchr(dstname,'/') + 1;
		strcpy(buf,dstname);
		dirname = strrchr(buf,'/');
		*dirname = 0;
		dirname = strrchr(buf,'/') + 1;
		printf("\t%s\tcopying %s\n",dirname,fname);
		AmCopyFile(srcname,dstname);
	}
}

/*************************************************************************
function: GenPublicFiles - Gen amk fiels for each passed node
*/
void GenPublicFiles (int count, char **names) {

/* go back and restore the code that reads the nodes, and writes *\
\* the .defs and .amk files (for future ports)                   */
}

