/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: am_deps.c
   Applix ASC version: 35
   Date: Thu Feb 18 21:12:28 1999
*/

#include <am_bld.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <string.h>

#define MAX_DEPS 5000
#define MAX_INCNAMES 30000

extern int SkipGeneratingProtos;
extern int NeedRebuiltProtos;	/* TRUE if should rebuild all_protos.dat */
extern char *CProtoMacro;   /* macro name for cproto */
extern char *DirSlash, *MyBuildTree;
extern int VerifyIncludes;
extern char *Pstreq();
extern char *GetCProtoDirName();
extern int PretendBuild;
void SaveCallFile();
void ReadCallFile();
void ReBuildMasterProtos();
char *readFileIntoBuf();
static ID FindIncDeps();
static void FindCallDeps();
char *AllocCallInPool();
static long lastFileTime;   /* hack set by readFileIntoBuf(); */
int DeferSavingCallFile;
ID NodeDirId();
static int FindIncludes();

/*************************************************************************
function:  FindMyDeps - Find includes files of inputs to a node
Returns T/F for circ detection.
*/
void FindMyDeps (id)
ID id;              /* output node */
{
    char name[STRPATH+1];
    node *nodeP, *inputP;
    int ix,jx;
    char **depends[MAX_DEPS];
    int num_depends = 0;
    static char namebuf[MAX_INCNAMES];
    int incount;
    int             code;
    char *fileBuf;
    char specialName[STRPATH+1];

    nodeP = NodePtr(id);
    if ( nodeP->deps_done ) {
        if ( PretendBuild || nodeP->call_deps_done )
            return;
    }
    if ( nodeP->dep_searching )     /* circ terminator */
        return;
    if ( !RuleArray[nodeP->ruleid].dep_srch )
        return;

    incount = NodeInCount(nodeP);
    for ( ix = 0; ix < incount; ix++ ) {
        NodeFilename(inputP = NodeInVal(nodeP,ix),name,TRUE);
        fileBuf = readFileIntoBuf(name);
        if ( !fileBuf )
            continue;
        strcpy(specialName,inputP->name);
        specialName[strlen(specialName)-1] = 'p';
        TRACE(("Finding dependencies in %s...",inputP->name));
        nodeP->dep_searching = TRUE;
        nodeP->callcount = 0;
        num_depends = FindIncludes(name,fileBuf,depends,namebuf,specialName);
        TRACE(("%d\n",num_depends));
        for ( jx = 0; jx < num_depends; jx++ )
            AddDepend(id,FindIncDeps(id,depends[jx]));
        nodeP = NodePtr(id);
        if ( !PretendBuild ) {
            FindCallDeps(id);
            nodeP = NodePtr(id);
            nodeP->call_deps_done = TRUE;
        }
        free(fileBuf);
        nodeP->dep_searching = FALSE;
    }
    nodeP->deps_done = TRUE;
    return;
}

/*************************************************************************
function: readFileIntoBuf - Read entire file into malloc'd memory
Returns null ternminate string or NULL on IO failures
*/
char *readFileIntoBuf(name)
char *name;
{
    struct stat  status;
    int          code,fd = -1;
    char *fileBuf = NULL;

#if AXMS
    fd = open(name, (O_RDONLY | O_BINARY), _S_IREAD);
#else
    fd = open(name, 0, 0);
#endif
    if ( fd == -1 )
        goto done;
    code = fstat(fd, &status);
    if (code == -1)
        goto done;
    lastFileTime = status.st_mtime;
    fileBuf = (char*)malloc(status.st_size+1);
    if ( !fileBuf )
        goto done;
    code = read(fd, fileBuf, status.st_size);
    if (code != status.st_size) {
        free(fileBuf);
        fileBuf = NULL;
        goto done;
    }
    fileBuf[status.st_size] = 0;    /* null term for easy searching */
done:
    if ( fd != -1 )
        close(fd);
    return(fileBuf);
}

int NumCalls;
int NumCallsAllocated;
int NumCallsAlloced;
callInfo **Calls;   /* Ptr to table of callInfo ptrs */
callInfo **CallsById;   /* callInfo.id links to calls info */
int NumCallsByIdAllocated;
ID CallUid = 1;     /* next avaibale cproto id number */
int ReplacedCallValue = FALSE;  /* for tracking state of file vs. memory */
int AddedCallValue = FALSE; /* for tracking state of file vs. memory */

/***************************************************************************
function: cmpCall 
*/
static cmpCall (w1, w2)
callInfo **w1;
callInfo **w2;
{
	int val;
	val = strcmp((*w1)->name,(*w2)->name);
	if ( val == 0 && (*w1)->fromFile && (*w2)->fromFile )
		val = strcmp((*w1)->fromFile,(*w2)->fromFile);
    return(val);
}

callInfo **LastCallComp;

/***************************************************************************
function: cmpNoteCall 
*/
static int cmpNoteCall (w1, w2)
callInfo **w1;
callInfo **w2;
{
	int val;
    LastCallComp = w2;
	val = strcmp((*w1)->name,(*w2)->name);
	if ( val == 0 && (*w1)->fromFile && (*w2)->fromFile )
		val = strcmp((*w1)->fromFile,(*w2)->fromFile);
    return(val);
}

/*************************************************************************
function: AddCallProto - Add a prototype
*/
static callInfo *AddCallProto (name,protoString,time,fromPath)
char *name;
char *protoString;
unsigned int time;
char *fromPath;
{
    callInfo key, **callPP, *callP;
    callInfo *keyP = &key;
    int insertAt, ix;
	char *fromFile;
	char *p;

	fromFile = strrchr(fromPath,'/');
	if ( !fromFile )
		fromFile = strrchr(fromPath,'\\');
	if ( !fromFile )
		fromFile = fromPath;
	else
		fromFile++;

    if ( NumCalls != 0 ) {
        key.name = name;
		key.fromFile = fromFile;
        callPP = (callInfo**)bsearch((char*)&keyP,(char*)Calls,NumCalls,sizeof(callInfo*),cmpNoteCall);
        if ( callPP ) {
            callP = *callPP;
            if ( !streq(callP->protoString,protoString) ) {
                callP->protoString = AllocCallInPool(protoString);
				callP->fromFile = AllocCallInPool(fromFile);
                callP->time = time;
                ReplacedCallValue = TRUE;
            }
            return(callP);
        }
    }
    if ( LastCallComp ) {   /* null only on first name */
        insertAt = LastCallComp - Calls;
        if ( insertAt < 0 )
            insertAt = 0;
        if ( strcmp(Calls[insertAt]->name,name) == 0 ) {
			if ( strcmp(Calls[insertAt]->fromFile,fromFile) < 0 )
	            insertAt++;     /* make sure we shift name greater! */
		}
        else if ( strcmp(Calls[insertAt]->name,name) < 0 )
            insertAt++;     /* make sure we shift name greater! */
    }
    else
        insertAt = 0;       /* only for first name */
    NumCalls++;
    if ( NumCalls > NumCallsAllocated ) {
        NumCallsAllocated = NumCalls*2;
		if ( !Calls ) Calls = malloc(10);	/* for realloc below */
        Calls = (callInfo**)realloc(Calls,NumCallsAllocated*sizeof(callInfo*));
    }
    callP = (callInfo*)calloc(1,sizeof(callInfo));
    callP->name = AllocCallInPool(name);
    callP->protoString = AllocCallInPool(protoString);
    callP->time = time;
	callP->fromFile = AllocCallInPool(fromFile);
    callP->id = CallUid++;
    for ( ix = NumCalls-1; ix > insertAt; ix-- )
        Calls[ix] = Calls[ix-1];
    Calls[insertAt] = callP;
    if ( CallUid > NumCallsByIdAllocated ) {
        NumCallsByIdAllocated = CallUid*2;
		if ( !CallsById ) CallsById = malloc(10);	/* for realloc below */
        CallsById = (callInfo**)realloc(CallsById,NumCallsByIdAllocated*sizeof(callInfo*));
    }
    CallsById[callP->id] = callP;
    AddedCallValue = TRUE;
/* printf("added: %s -> %s\n",callP->name,callP->protoString); */
    return(callP);
}

/*************************************************************************
function: AllocCallInPool - Dense word allocator
*/
char *AllocCallInPool (str)
char *str;
{
    int size;
    static char *CallPool;
    static int LeftInPool;


    size = strlen(str)+1;
    if ( size > LeftInPool ) {
        CallPool = malloc(64000);
        LeftInPool = 64000;
    }
    LeftInPool -= size;
    strcpy(CallPool,str);
    CallPool += size;
    return(CallPool-size);
}

#define MAX_CALLS 10000

/*************************************************************************
function: FindCallDeps - Find all function calls
*/
static void FindCallDeps (id)
    ID id;
{
    char *fileBuf, *curTok;  /* the entire file */
    node *nodeP;
    register char *tokEnds, *p;
    ID names[MAX_CALLS], nameid;
    callInfo *ptrs[MAX_CALLS];
    int num_calls = 0;
    callInfo key, **callPP, *callP;
    callInfo *keyP = &key;
    int ix, size;
    node *src;
    char srcfile[STRPATH], filename[STRPATH];
    ID *depsP;

    nodeP = NodePtr(id);

    src = NodeInVal(nodeP,0);
    NodeFilename(src,srcfile,TRUE);
    sprintf(filename,"%s%s%s",GetCProtoDirName(),DirSlash,GetComponent(nodeP));
    CreateDir(filename);
    sprintf(filename+strlen(filename),"%s%s",DirSlash,nodeP->name);
    filename[strlen(filename)-1] = 'r';
    fileBuf = readFileIntoBuf(filename);
    if ( !fileBuf )
		return;

	for ( curTok = fileBuf; *curTok; curTok = tokEnds+1 ) {
		p = tokEnds = strchr(curTok,'\n');
		if ( !tokEnds ) break;
        while ( *p < ' ' )
            {
            *p = 0;
            p--;
            }
       
        key.name = curTok;
		key.fromFile = NULL;
        callPP = (callInfo**)bsearch((char*)&keyP,(char*)Calls,
                NumCalls,sizeof(callInfo*),cmpCall);
        if ( callPP ) {
            callP = *callPP;
            nameid = callP->id;
            if ( !callP->inCurrentFile ) {
                ptrs[num_calls] = callP;
                names[num_calls++] = nameid;
                callP->inCurrentFile = TRUE;
                bld_assert(num_calls<MAX_CALLS,("Too many calls in %s",src->name));
            }
        }
    }
	free(fileBuf);
    if ( num_calls == 0 )
        return;

    nodeP->callcount = num_calls;
    size = sizeof(node)+(nodeP->incount+nodeP->depcount+num_calls-1)*sizeof(ID);
    ResizeObj(id,size);
    nodeP = NodePtr(id);
    depsP = AddrOfCallVals(nodeP);
    memcpy(depsP,names,num_calls*sizeof(ID));

    for ( ix = 0; ix < num_calls; ix++ )
        ptrs[ix]->inCurrentFile = FALSE;
}

/*************************************************************************
function: FindIncDeps - Find dependencies in an include file
*/
static ID FindIncDeps (parent,incname)
ID parent;
char *incname;
{
    ID id, compid;
    char name[STRPATH+1];
    node *nodeP;
    int ix,jx;
    char **depends[MAX_DEPS];
    int num_depends = 0;
    char *namebuf;
    node *NodeCvtToRuled();
    char *fileBuf;

    id = LookupIdForMe(incname,compid = NodeCompId(NodePtr(parent)));
    nodeP = NodePtr(id);
    if ( nodeP->deps_done )
        return(id);
    if ( nodeP->dep_searching )     /* circ terminator */
        return(0);

    NodeFilename(nodeP,name,TRUE);
    fileBuf = readFileIntoBuf(name);
    if ( !fileBuf ) {
        nodeP->deps_done = TRUE;
        return(id);
    }
    TRACE(("Finding dependencies in %s...",nodeP->name));
    nodeP->dep_searching = TRUE;
    namebuf = malloc(MAX_INCNAMES);
    num_depends = FindIncludes(name,fileBuf,depends,namebuf,"all_applix.p");
    free(fileBuf);
    TRACE(("%d\n",num_depends));
    if ( num_depends && !nodeP->ruleid )  /* dummy up a rule */
        nodeP = NodeCvtToRuled(id,DUMMY_RULE,compid);
    for ( jx = 0; jx < num_depends; jx++ )
        AddDepend(id,FindIncDeps(id,depends[jx]));
    nodeP = NodePtr(id);
    nodeP->dep_searching = FALSE;
    nodeP->deps_done = TRUE;
    free(namebuf);
    return(id);
}

#define TYPE_ELF_FILE 1
#define TYPE_AB_FILE 2
#define TYPE_C_FILE 3

/*************************************************************************
function: FindIncludes - Search buffer for #include string, ret count
Manual buffer scanning is done rather than fgets because the ALTOS
fgets is such a disaster.
*/
static int FindIncludes (filename,fileBuf,names,namebuf,specialName)
char *filename; /* for error msgs only */
char *fileBuf;  /* the entire file */
char **names;
char *namebuf;
char *specialName;  /* special file prototypes */
{
    register char *p;
    char *name;
    int count = 0;
    char *line;
    int namelen = 0;
	char *suffix;
	int fileType;

	suffix = strrchr(filename,'.');
	if ( !suffix )
		fileType = TYPE_C_FILE;
	else if ( streq(".els",suffix) || streq(".am",suffix) || streq(".elf",suffix) )
		fileType = TYPE_ELF_FILE;
	else if ( streq(".ab",suffix) )
		fileType = TYPE_AB_FILE;
	else
		fileType = TYPE_C_FILE;

    name = namebuf;
    while ( *fileBuf ) {
        line = fileBuf;
		switch ( fileType ) {
		case TYPE_C_FILE:
	        if ( *line != '#' )
	            goto nexter;
	        p = line+1;
			break;
		case TYPE_ELF_FILE:
	        if ( *line == '#' )
		        p = line+1;
			else
				p = line;
			break;
		case TYPE_AB_FILE:
	        if ( *line != '"' )
				goto nexter;
	        if ( line[1] == '#' )
		        p = line+2;
			else
				p = line+1;
			break;
		}
        while ( (*p == ' ') || (*p == TAB) ) p++;
        if ( p = Pstreq(p,"include") ) {
            while ( (*p == ' ') || (*p == TAB) ) p++;
            if ( (*p != '"') && (*p != '<') ) goto nexter;
            p++;
            bld_assert((name+100)<namebuf+MAX_INCNAMES,("Too many include files in %s",filename));
            namelen = 0;
            while ( *p && (*p != '"') && (*p != '>') && (*p != '\n') ) {
				if ( *p == '/' || *p == '\\' )	/* ignore partila paths */
					goto nexter;
                name[namelen++] = *p++;
			}
            if ( (*p != '"') && (*p != '>') ) goto nexter;
            name[namelen++] = 0;
            if ( streq(name,specialName) || (strlen(name) >= MAX_NODENAME) )
                goto nexter;
            assert(count < MAX_DEPS);
            names[count++] = name;
            name += namelen;
        }
nexter: while ( *fileBuf && *fileBuf != '\n' )
            fileBuf++;
        if ( *fileBuf )
            fileBuf++;      /* get beyond '\n' */
    }
    return(count);
}

/*************************************************************************
function: GenProtosForMe - Create amake_protes file
*/
void GenProtosForMe (nodeP)
node *nodeP;
{
    char includer[1000];
    int ix;
    ID *callsP;
    FILE *fp;
    char sproto[1000], *fileBuf;
    char specialName[STRPATH+1];
    node *inputP;
    char *p;

    inputP = NodeInVal(nodeP,0);
    strcpy(specialName,inputP->name);
    specialName[strlen(specialName)-1] = 'p';
    sprintf(includer,"%s/%s",DefaultDirName(nodeP),specialName);
    fp = fopen(includer,"w");
	if (!fp) {
		fprintf( stderr, "Cannot open file %s, errno=%d\n", includer, errno);
		fflush(stderr);
		exit(44);
	}

    sprintf(sproto,"%s%s%s%s%s",GetCProtoDirName(),DirSlash,
                GetComponent(nodeP),DirSlash,nodeP->name);
    sproto[strlen(sproto)-1] = 's';
    fileBuf = readFileIntoBuf(sproto);
    if ( fileBuf ) {
        fwrite(fileBuf,strlen(fileBuf),1,fp);
        free(fileBuf);
    }

    callsP = AddrOfCallVals(nodeP);
    for ( ix = 0; ix < nodeP->callcount; ix++,callsP++ )
        {
        p = strstr( CallsById[*callsP]->protoString, "dllexport" );
        if ( p != (char *)NULL )
            {
            /* we do not put these into the prototype file, since this */
            /* file may actaully be importin the function              */
            continue;
            }
        fprintf(fp,"%s\n",CallsById[*callsP]->protoString);
        }
    fclose(fp);
}

/*************************************************************************
function: UpdateProtoTables - Updates protos based on new cproto file
*/
void UpdateProtoTables (filename)
char *filename;
{
    char *fileBuf, *dataBuf, *newlineAt, *protoAt, *p;
    char *start_proto;
    char name[300];

    dataBuf = fileBuf = readFileIntoBuf(filename);
	if ( !dataBuf ) {
		ErrorMsg("Could not read %s",filename);
		return;
	}

    while ( *dataBuf ) {
        p = newlineAt = strchr(dataBuf,'\n');
        bld_assert(newlineAt,("Could not read %s",filename));
        while ( *p < ' ' )
           {
           *p = 0;
           p--;
           }
        if ( 0 == strncmp("static ",dataBuf,7)
			|| 0 == strncmp("/* ",dataBuf,3)
			|| strstr(dataBuf,"static\n")
			|| strstr(dataBuf,"static\t")
			|| strstr(dataBuf,"static ") ) {
            dataBuf = newlineAt + 1;
            continue;
        }

        start_proto = dataBuf;
#if AXMS
        p = strstr( dataBuf, "dllexport" );
        if ( p != (char *)NULL )
            {
            /* skip past the leading "__declspec ( dllexport )" so we do not */
            /* find this opening paren as the start of the func name          */
            start_proto = p + 10;
            }
        p = strstr( dataBuf, "dllimport" );
        if ( (p != (char *)NULL) && (p > start_proto) )
            start_proto = p + 10;
#endif

        if ( CProtoMacro && CProtoMacro[0] ) {
            protoAt = strstr(start_proto,CProtoMacro);
            bld_assert(protoAt,("Could not find %s in %s",CProtoMacro,filename));
            protoAt -= 2;       /* point to last char of name */
        }
        else {
            protoAt = strchr(start_proto,'(');
            bld_assert(protoAt,("Could not find ( in %s",filename));
            protoAt--;      /* point to last char of name */
        }
        for ( p = protoAt; (*p != ' ') && (*p != '*'); p-- );
        p++;        /* point to first char of name */
        memcpy(name,p,protoAt-p+1);
        name[protoAt-p+1] = 0;
		if ( strlen(name) > 1 )	/* int (*func(int arg))(); has a problem... */
	        AddCallProto(name,dataBuf,lastFileTime,filename);
        dataBuf = newlineAt + 1;
    }
    free(fileBuf);
    if ( ReplacedCallValue || AddedCallValue ) {
		if ( !DeferSavingCallFile )
	        SaveCallFile();
        ReplacedCallValue = FALSE;
        AddedCallValue = FALSE;
    }
}

/*************************************************************************
function: SaveCallFile - Write all prototypes info to file
*/
void SaveCallFile ()
{
    char filename[1000];
    char tempname[1000];
    FILE *fp;
    int ix;
    callInfo *p;

    sprintf(tempname,"%s/protos.new",GetCProtoDirName());
    fp = fopen(tempname,"w");
    bld_assert(fp,("Could not write %s",tempname));
    fprintf(fp,"%06d\n",NumCalls);
    for ( ix = 0; ix < NumCalls; ix++ ) {
        p = Calls[ix];
        fprintf(fp,"%04X %X %s %s %s\n",p->id,p->time,p->name,p->fromFile,p->protoString);
    }
    bld_assert(!ferror(fp),("Could not write %s",tempname));
    bld_assert(EOF != fclose(fp),("Could not write %s",tempname));
    sprintf(filename,"%s/protos.dct",GetCProtoDirName());
    Arename(tempname,filename);
}

int HEX(a)
int a;
{
    return((a > '9') ? (a-'A'+10) : (a-'0'));
}
#if 0
#define HEX(a) ((a) > '9') ? ((unsigned int)((a-'A'+10))) : ((unsigned int)((a-'0')))
#endif

/*************************************************************************
function: ReadCallFile - Read all prototypes info from file to memory
*/
void ReadCallFile ()
{
    char filename[1000];
    char *fileBuf, *newlineAt, *protoAt, *p;
    char name[300];
    callInfo *allcallsP, *callP;
    int callnum;
    int ix;
    int fileCalls;

	if ( NeedRebuiltProtos ) {
		ReBuildMasterProtos();
		return;
	}
    sprintf(filename,"%s/protos.dct",GetCProtoDirName());
    TELL(("Reading %s...\n",filename));
    fileBuf = readFileIntoBuf(filename);
    if ( !fileBuf ) {
        Calls = (callInfo**)malloc(1);
        CallsById = (callInfo**)malloc(1);
        return;
    }
    bld_assert(fileBuf,("Could not read %s",filename));
    newlineAt = strchr(fileBuf,'\n');
    bld_assert(newlineAt,("Could not read %s",filename));
    fileCalls = atoi(fileBuf);
    fileBuf = newlineAt + 1;
    Calls = (callInfo**)malloc(fileCalls*sizeof(callInfo*));
    allcallsP = (callInfo*)calloc(1,sizeof(callInfo)*fileCalls);
    callnum = 0;
    while ( *fileBuf ) {
        p = newlineAt = strchr(fileBuf,'\n');
        bld_assert(newlineAt,("Could not read %s",filename));
        while ( *p < ' ' )
           {
           *p = 0;
           p--;
           }
        Calls[callnum] = callP = allcallsP+callnum;
        callP->id = (HEX(fileBuf[0]) << 12) +
                    (HEX(fileBuf[1]) << 8) +
                    (HEX(fileBuf[2]) << 4) +
                    HEX(fileBuf[3]);
        if ( callP->id >= CallUid )
            CallUid = callP->id + 1;
        callP->time = (HEX(fileBuf[5]) << 28) +
                        (HEX(fileBuf[6]) << 24) +
                        (HEX(fileBuf[7]) << 20) +
                        (HEX(fileBuf[8]) << 16) +
                        (HEX(fileBuf[9]) << 12) +
                        (HEX(fileBuf[10]) << 8) +
                        (HEX(fileBuf[11]) << 4) +
                        HEX(fileBuf[12]);
        callP->name = fileBuf+14;
        p = strchr(callP->name,' ');
        *p = 0;
		callP->fromFile = p+1;
        p = strchr(callP->fromFile,' ');
        *p = 0;
        callP->protoString = p+1;
        callnum++;
        fileBuf = newlineAt + 1;
    }
    NumCalls = callnum;
    CallsById = (callInfo**)malloc((CallUid+1)*sizeof(callInfo*));
    for ( ix = 0; ix < NumCalls; ix++ ) {
        callP = Calls[ix];
        CallsById[callP->id] = callP;
    }
}

/*************************************************************************
function: ResetDeps - Clera all deps_done flags
*/
void ResetDeps ()
{
    register node *nodeP;
    register ID id;

    for ( id = NextType(0,TYPE_NODE); id ;id = NextType(id,TYPE_NODE) ) {
        nodeP = NodePtr(id);
		nodeP->deps_done = FALSE;
    }
}

/*************************************************************************
function: ReBuildMasterProtos - Produce protos.dct from the .s files
*/
void ReBuildMasterProtos ()
{
	int priority = 1;
	char the_dir[STRPATH+1], name[STRPATH+1], *afdir, curdir[STRPATH+1];
	char fullname[STRPATH+1];
	int code, maxlevels = 99;

    sprintf(fullname,"%s/protos.dct",GetCProtoDirName());
	unlink(fullname);
	TELL(("Generating %s...\n",fullname));
    CreateDir(GetCProtoDirName());
	DeferSavingCallFile = TRUE;
	afdir = GetCProtoDirName();
	strcpy(curdir,afdir);
/*	TELL(("Searching %s\n",afdir)); */

	/* reset state */
	ResetDeps();

	while ( TRUE ) {
		code = AFindFiles(afdir,maxlevels,the_dir,name);
		afdir = NULL;
		if ( code == 1 )
			break;
		else if ( code == 2 ) {
			TELL(("could not stat %s\n",name));
			continue;
		}
		assert((code == 0) || (code == 3));
		if ( !streq(the_dir,curdir) ) {
/*			TELL(("Searching %s\n",the_dir)); */
			strcpy(curdir,the_dir);
			if ( code == 3 )
				continue;
		}
		if ( name[strlen(name)-1] != 's' )
			continue;
		sprintf(fullname,"%s/%s",the_dir,name);
		UpdateProtoTables(fullname);
	}
	DeferSavingCallFile = FALSE;
	SaveCallFile();
}

/*************************************************************************
function: VfyProtoableIncludes - Creat includes for prototyper if not already there
*/
void VfyProtoableIncludes (id,dir)
ID id;
char *dir;
{
	node *nodeP = NodePtr(id);
	register ID *arrP = AddrOfDepVals(nodeP);
	int depcount = NodeDepCount(nodeP);
	register int ix;
	ID child;
	char filename[500];

    for ( ix = 0; ix < depcount; ix++ ) {
		child = *arrP++;
        nodeP = NodePtr(child);
		StatNode(nodeP);
        if ( nodeP->ruleid == DUMMY_RULE )
			VfyProtoableIncludes(child,dir);
        if ( 0 == NodeDirId(nodeP) )
            continue;
#if 0 
/* Not needed any more, now all required header files should be
   built before proto type checking by adding to dependencies list
   of cproto.tag */
		if ( !nodeP->exists || nodeP->dtmod == 10 ) {	/* magic 10 in am_bld.c */
		    sprintf(filename,"%s%s%s",dir,DirSlash,nodeP->name);
			if ( -1 == access(filename,0) ) {
				fclose(fopen(filename,"w"));
			    TELL(("Created placeholder for %s\n",nodeP->name));
			}
		}
#endif
    }
}

/*************************************************************************
function: CheckForProtoConflicts - Produce error message if there is a proto error
*/
void CheckForProtoConflicts ()
{
	int dex = 0, first_dex;
	callInfo dummy;
    callInfo *prev = &dummy;

    if ( SkipGeneratingProtos )
        return;

	dummy.name = "not a possible name";
	while ( dex < NumCalls ) {
		if ( streq(prev->name,Calls[dex]->name) ) {
			first_dex = dex-1;
			do {
                if ( !PrototypesAreSame( prev->protoString,
                                         Calls[dex]->protoString ) ) {
					ErrorMsg("\n       Prototype conflict on '%s'",Calls[dex]->name);
					for ( dex = first_dex;
							dex < NumCalls && streq(Calls[first_dex]->name,Calls[dex]->name);
							dex++ ) {
						fprintf(ErrStream,"%s: %s\n",Calls[dex]->fromFile,Calls[dex]->protoString);
					}
				}
				else
					dex++;
				prev = Calls[dex-1];
			}
			while ( dex < NumCalls && streq(prev->name,Calls[dex]->name) );
		}
		else
			dex++;
		prev = Calls[dex-1];
	}
}

/*************************************************************************
function: Whencer - Report inputs
*/
void Whencer (nodeP,leader,recurse)
node *nodeP; char *leader; int recurse;
{
	char filename[STRPATH];
	char *newleader;
	node *childP;
	int ix, incount;

	if ( !nodeP->ruleid ) {
		printf("%s: not an output module\n",nodeP->name);
		return;
	}
    incount = NodeInCount(nodeP);
    for ( ix = 0; ix < incount; ix++ ) {
        childP = NodePtr(NodeInId(nodeP,ix));
		NodeFilename(childP,filename,TRUE);
		printf("%s%s\n",leader,filename);
		if ( recurse && childP->ruleid ) {
			newleader = malloc(strlen(leader)+3);
			strcpy(newleader,leader);
			strcat(newleader,"  ");
			Whencer(childP,newleader,TRUE);
		}
	}
}


/* Remove the variable names to leave just the types in the prototype */
/* ie. convert "void d_rect(RectFata *data)" to "void d_rect(RectFata *)" */
void remove_variable_names( char *proto_str )
{
    char *p1, *p2;

    p1 = proto_str;

    /* look for the opening paren */
    while ( (*p1 != '(') && (*p1 != '\0') ) 
        {  p1++; }

    if ( *p1 == '\0' )
       return;

    /* now go through the arguments */
    p1++;        /* just after opening paren */
    p2 = p1;

    while ( (*p1 != ')') && (*p1 != '\0') )
       {
       /* get to start of type */
       while ( (*p1 == ' ') || (*p1 == ',') )
           {  p1++; }

 
       /* look for end of type word */
       while ( (*p1 != ' ') && (*p1 != ')') && (*p1 != ',') )
           { p1++; }
       if ( *p1 == ')' )
           /* all done */
           return;
       
       if ( *p1 == ',' )
           {
           /* end of type, on to the next */
           p1++;
           continue;
           }
       
       p1++;     /* character after the blank at end type word */
       if ( *p1 == '*' )
          p1++;

       /* should now be pointing to the variable name, if any */
       if ( *p1 == ',' )
          {
          p1++;
          continue;
          }

       /* look for the end of the variable name */
       p2 = p1;
       while ( (*p2 != ' ') && (*p2 != ')') && (*p2 != ',') )
           { p2++; }

       /* clip out the variable name */
       strcpy( p1, p2 );
       }
    return;
}


#define DLL_EXPORT_STR "__declspec ( dllexport ) "

PrototypesAreSame( proto1, proto2 )
char *proto1, *proto2;
{
    char p1_str[500], p2_str[500];

    
    /* on NT, we allow some functions to be declared as DLL_EXPORT,     */
    /* while the same function name elsewhere (AxDispatch() for example */
    /* may be local, and not DLL_EXPORT                                 */
    if ( strncmp( DLL_EXPORT_STR, proto1, strlen(DLL_EXPORT_STR) ) == 0 )
        proto1 = proto1 + strlen(DLL_EXPORT_STR);
    if ( strncmp( DLL_EXPORT_STR, proto2, strlen(DLL_EXPORT_STR) ) == 0 )
        proto2 = proto2 + strlen(DLL_EXPORT_STR);

    strcpy( p1_str, proto1 );
    strcpy( p2_str, proto2 );

    /* now go through the string and remove any variable names in the */
    /* prototype definitions, so we are left just with the types, to avoid */
    /* un-necessary complaints based on a simnple string coparison         */
    remove_variable_names( p1_str );
    remove_variable_names( p2_str );

    if ( strcmp( p1_str, p2_str ) == 0 )
       return( TRUE );    /* the same */
    else
       return( FALSE );   /* different */
}

