/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: am_main.c
   Applix ASC version: 89.42
   Date: Tue Jun  1 21:12:14 1999
*/

#define FOR_MAIN

#include <errno.h>
#include <signal.h>
#include <ctype.h>

#if AXMS
#include <string.h>
#include <io.h>
#include <sys\stat.h>
#else
#include <sys/stat.h>
#endif

#include <am_bld.h>

#define MAX_TAPES 20

#define SZ_FILE_CACHE 32700
#define MATCH_PSUEDO "^match"
#define KEY_INTO "into"
#define KEY_WITH "with"
#define KEY_DEPEND "dep"

#define DR_CHAR '.'
#define DR_LOCATE ".locate "
#define DR_BUILD ".buildfile "
#define DR_MAPFILE ".mapfile "
#define DR_DEFCOMP ".defcomp "
#define DR_TEMPDIR ".tempdir "
#define DR_TELL ".tell "
#define DR_FRCLOC ".relocate "
#define DR_ERROR ".error "
#define DR_OUTDIR ".outdir "
#define DR_CHECKDIR ".checkdir "
#define DR_LOCDIR ".locdir "
#define DR_ALTDIR ".altdir "
#define DR_1CHECKDIR ".check1dir "
#define DR_1LOCDIR ".loc1dir "
#define DR_RULEDIR ".ruledir "
#define DR_START_TAPE ".start_tape "
#define DR_COMPONENT ".component "
#define DR_EXPORTVAR ".export "
#define DR_EXTERN ".extern "
#define DR_GLOBAL ".global "
#define DR_DEFDIR ".defdir "
#define DR_OBJDIR ".objdir "
#define DR_TAPE ".tape "
#define DR_ENDTAPE ".endtape "
#define DR_DEFTAPE ".deftape "
#define DR_CPROTO_EXCLUDE ".cproto_exclude "
#define DR_STRICT_EXCLUDE ".strict_exclude "
#define DR_CPYRIGHT_EXCLUDE ".cpyright_exclude "
#define DR_CPYRIGHT_NOWARN ".cpyright_nowarn "
#define DR_OPENSRC_EXCLUDE ".opensrc_exclude "
#define DR_OPENSRC_RENAME ".opensrc_rename "
#define DR_OPENSRC_BUILTFILE ".opensrc_builtfile "
#define DR_SHLIB_DIR ".shared_lib_dir "
#define DR_STRICT_COMPILES ".strict_compiles "

#define MAX_DIRECTIVE_LEN 1000      /* includes options */

#define MF_USED 1       /* Macro Currently in use */
#define MF_PERM 2       /* Macro is a permanent macro */

#if HP800V3
#define LIB_PATH "SHLIB_PATH"
#else
#if RIOS
#define LIB_PATH "LIBPATH"
#else
#if AXMS
#define LIB_PATH "PATH"
#else
#define LIB_PATH "LD_LIBRARY_PATH"
#endif
#endif
#endif

#if AXMS
    char *DirSlash = "\\";
	char *DirSep = ";";
#else
    char *DirSlash = "/";
	char *DirSep = ":";
#endif

char MacroDfltFlag = MF_USED | MF_PERM;     /* creation mask */
char *MacroNames[MAX_VARS];   /* macro vars */
char *MacroValues[MAX_VARS];  /* macro var values */
char MacroFlags[MAX_VARS];
int NumMacros = 0;      /* high water */
int RunningNT3_1;

char *getenv(), *strchr(), *strrchr(), *strdup();
char StartDir[STRPATH+1];     /* dir building from */
char VarsFile[STRPATH+1];
int LocAutoCache = FALSE;    /* FALSE if shouldn't put checkdirs in files list*/
int TellVars = FALSE;   /* TRUE verbose all #defines */
int Quiet;      /* Error messages only */
int Verbose;        /* log commands while doing them */
int CommandLog;        /* log commands while doing them */
int Trace;          /* tell about state */
int TraceNodes;     /* tell about node states */
int Logging;        /* more telling */
int Debug;          /* most telling */
int VerifyIncludes = FALSE; /* tell of include file non-existance */
int Turbo = FALSE;	/* Debug/Optimized build */
int Opensrc = FALSE;
int InputLine;          /* input line number */
char InputFilename[STRPATH+1];	/* file matching LineNumber */
int forceBuild = FALSE;
char *ErrorBuf = "";        /* Buffer containing error info */
char *TempDir;  /* directory to put tempfiles */
int DoingReport;    /* send errors to a report file */
FILE *ReportFp;     /* File to send errors to */
int TotalErrors = 0;
ID *HashArray;
int StatingWorld;   /* Doing world-wide stat of all nodes */
char StdOutFile[STRPATH+1];
char StdErrFile[STRPATH+1];  /* for reports & outfile commands */
char ReportFile[STRPATH+1];
int BuildPhase;     /* TRUE while building (dummy for amcache) */
int NarrowMode = TRUE;
int JustDump;
int DoingProfileBuild = FALSE;
int StrictCompiles;	/* if TRUE, use the c++ rather tha C compiler */

#define MAX_RENAMES 200
char *RenameDirs[MAX_RENAMES]; /* opensrc renames */
int NumRenameDirs;

char *AppVars[MAX_APPVARS];
int NumAppVars = 4;         /* acpp -T<defs>/-Q -I<varsdir> -I<vd2> ... */
int ForceLocating = FALSE;/* TRUE while processing a .relocate vars directive */
ID ForceDir;        /* Current ".locate" dir while ForLocateing */
char ZZtemp[STRPATH+1], ZZtempout[STRPATH+1];
char LineBuf[MAX_LINELEN];  /* the line being read */
char *BuildFiles[MAX_BUILD_FILES];
char *OutDirs[MAX_OUTDIRS]; /* Dirs allowed to contain output files */
char *CprotoExcludeFile;    /* Filename of Cproto exclusion list */
char *StrictExcludeFile;    /* Filename of Struct compile exclusion list */
char *CpyrightExcludeFile;/* Filename of files to not touch cpyright */
char *CpyrightNoWarnFile;/* excess vopy tags verified to be ignorable */
char *OpensrcExcludeFile;/* Filename of files to NEVER copy to opensrc dirs */
char *OpensrcBuiltFile;/* output file to include rather than base sources */
int CprotoForcing;      /* TRUE if ModsToForce set by -fcp */
int NumOutDirs = 0;
int NumBuildFiles;
ID AltOutDirId;     /* directory for bypassed output */
ID *RuleDirs;
int DoMemoryDump;   /* TRUE if should write an ascii dump */
char *DumpFilename; /* file to dump to */
int NumAppsToBuild = 0;     /* How many -app apps */
char *AppNamesToBuild[100]; /* Their names */
int JustCleanAll;       /* TRUE on the -clean option */
int KeepGarbage = FALSE;    /* TRUE if should keep broken output files */
int FindAllDeps = FALSE;    /* set by -alldeps */
int NodesMoved = FALSE; /* TRUE if any locdir/checkdir nodes have moved */
int UsingCache = TRUE;     /* TRUE if cache is valid */
int CacheOK = FALSE;       /* do not bother to rebuild the cache */
int LimitedDebug = FALSE;  /* use the file /ax/debug_list to determine */
                           /* what to build with the -g switch         */
int NeedRebuiltProtos;  /* TRUE if should rebuild all_protos.dat */
int RebuildProtoOnAllCompiles;  /* if FALSE, just on srcfile edits */
int SkipGeneratingProtos = FALSE;   /* fast way for quicker re-build */

char *CmlVars[50];
int NumCmlVars;
int ReCheckDirs = FALSE;    /* set by CleanCache is a locdir/checdir changed */
int TouchBuild;       /* Just for testing */
int PretendBuild = TRUE;       /* First 'pretend' pass for proto generation */
char *CProtoMacro; /* = "AMKP_";    /* macro name for cproto */
char *RS6000Log;    /* For module name tracking for IBM RS6000 port */
char *LogicalDirNames[MAX_LOGICAL_DIRS];
ID DirNameMap[MAX_LOGICAL_DIRS];
int NumLogNames = 0;
ID ObjDir;      /* default objects directory */
int TellCmds;
char *FileCache;
int NumToGrep;     /* number of dirs to grep */
char *ToGrep[100];  /* pathname segments to grep */
int NumToFind;     /* number of files to find */
char *ToFind[1000];  /* module names to make amk files for */
int NumToAmkFile;     /* number of files to make amk files for */
char *ToAmkFile[1000];  /* module names to find */
char *ToWhence; /* module to report inputs for */
int ReportStrictFiles; /* TEMP */
int WhenceRecursive = FALSE;    /* whether to go recusrice ToWhence */
int n_ctl_files = 0;
char **ctl_file = NULL;

/* Structures to hold information from VERSION files */
typedef struct {
	char major_version[16];		/* Integer part of version number */
	char minor_version[16];		/* Fractional part of version number */
	char project_name[256];		/* Name of project */
} version_line;
typedef struct {
	version_line project;		/* This project */
	version_line base;			/* Version info re base SHELF 'project' uses */
	int nSubProjects;			/* Number of subprojects below */
	version_line subProjects[1];/* Growable array of subproject info */
} version_info;

int BaseMajorVersion=0, BaseMinorVersion=0;
char *MyBuildTree;

extern char StartDir[STRPATH+1];     /* dir building from */
extern int LocAutoCache;
extern int VerifyIncludes;
extern int TellVars;
extern int NodeCreated;
extern char *NodeName();
extern int DirObjCreated;
extern int CompObjCreated;
void CrtSpecials ();
void GetSpecials ();
int Cstreq ();
void ReadCommandLine();
version_info *ReadVersionInfo();
int CleanCache ();
int  LoadCmlVars ();
void AddAppVar ();
void ReadVarsFile ();
void ReadSavedVars ();
int DoLine();
void CleanWhiteSpace();
void Usage();
void ExpandVars();
int DoAssignment();
void AddVar ();
void ClipVars ();
void ExportVar ();
int AppSub ();
void AddCtrlFile ();
void NoteVars ();
void DoDirective ();
void CrtForced ();
void DoLocators ();
void CompareVersions ();
void FixupMoved ();
void AddLocs ();
void VerifyOutdirs ();
void DoFindAllDeps ();
int true_nop();
void TellLog();
int ReadMemory();
void aborter();
void DumpMemory();
void GrepFiles(), SetSysVars();
char **AmListFiles();
void FindFiles(), NoteDeltaDirs();
void SetDir();
void CreateDir();
static void ReadExclusionFiles();
static void ReadBuildFiles();
void WriteMemory();
void DeleteAll();
void DoBuild();
void DoReportStrictFiles(); /* TEMP */
void ErrorMsg();
void FlushMemory();
ID NextType();
void AlcAStr();
void VerboseLog();
void ResizeObj();
void SetLibDirs ();
void ReadABuildFile();
/* forward refs */
char *LocateVar();              /* Get value of a variable */
char *GetVar();                 /* Get value of a variable or NULL */
char *NeedVar();                /* Get value of a variable or "" */
char CacheFile[STRPATH+1];
char *WhereAmI();
void FindMyDeps();
int AddRuleDeps();
void line_error();
void wrn_msg();
int DoRule();
int NodeFilename();
int NodeDepCount();
int NodeInCount();
ID NodeCompId();
void TraceLog();
int AFindFiles();
int FilterToTokArray();
ID LookupComp();
int VpShcmd();
int TokenizeString();
void MyFree();
void LogLog();
int SizeObj();
char *readFileIntoBuf(), *strstr();

/*************************************************************************
function: GetMachineName - Return machine name based on amake build switch
*/
char *GetMachineName ()
{
#if SUN4S               /* Solaris 2.X+ only */
    return("SUN4S");
#elif SIEMENS
    return("SIEMENS");
#elif SUN4
    return("SUN4");
#elif SUNIS
    return("SUNIS");
#elif PMAX
    return("PMAX");
#elif NT_X86
    return("NT_X86");
#elif NT_MIPS
    return("NT_MIPS");
#elif NT_ALPHA
    return("NT_ALPHA");
#elif NT_PPC
    return("NT_PPC");
#elif HP800V3 /* Note: Must come *before* HP800. */
    return("HP800V3");
#elif HP800
    return("HP800");
#elif RIOS
    return("RIOS");
#elif ALPHA
	return("ALPHA");
#elif SGI
	return("SGI");
#elif M_88K
	return("M_88K");
#elif AVIION
	return("AVIION");
#elif UNIXWARE
	return("UNIXWARE");
#elif SCOUNIX
	return("SCOUNIX");
#elif NCR
	return("NCR");
#elif NEC_EWS
	return("NEC_EWS");
#elif LINUX && !LIBC5 /* so that amake.amk won't drop into this case */
	return("LINUX");
#elif LINUX_LIBC5
	return("LINUX_LIBC5");
#else
No_machine_specified;_enter_one_here_(am_main.c).
#endif
}

/*************************************************************************
function: main - main
*/
int main(argc,argv)
int argc;
char *argv[];
{
    char *p;
    char aos_buf[300];
    char machine_buf[300];
    char ctl[STRPATH+1];
    struct stat ctl_buf;

if (getenv("CATCH_AMAKE")) {
	static int spin = 1;
	while (spin);
}

#if !AXMS && !ALPHA
    am_set_sigs();
#endif

#if AXMS
    /* redirect stderr to stdout since none of the Windows utilities know */
    /*   how to redirect stderr to capture error/warning output from */
    /*   compiles, etc.                                              */
    dup2( 1, 2 );

    /* See what version we are running, since the compiler/linker name */
    /* changes between NT3.1 and NT3.5.  Look for the old one          */
    RunningNT3_1 = TRUE; 
    aos_buf[0] = 0;
    _searchenv( "LIB.EXE", "path", aos_buf );
    if ( strlen(aos_buf) == 0 )
       _searchenv( "LIB.EXE", "PATH", aos_buf );
    if ( strlen(aos_buf) != 0 )
       /* found the "new" linker */
       RunningNT3_1 = FALSE;
#endif

/* Get amake's directory... should be at top level of build tree */

	MyBuildTree = getenv("AOS_LOCATION");
	if ( !MyBuildTree ) /* use amake's locations to get the root of the tree */
       MyBuildTree = WhereAmI(argv[0]);

	/* If there is a pgms tree, this is not opensrc mode ... */

	sprintf(ctl,"%s%spgms",MyBuildTree,DirSlash);
	Opensrc = !GetSimpleDtmod(ctl);

    TempDir = TEMPNAME; /* just till tempdir directive read */
    ReadCommandLine(argc,argv);
    TELL(("Amake...\n"));

	SetSysVars();

    AMakePath(CacheFile,TempDir,"amake.dat");
    if ( UsingCache ) UsingCache = ReadMemory(CacheFile);
    if ( JustDump ) {
        bld_assert(UsingCache,("no cache to dump"));
        GetSpecials();
        ReadCallFile();
        DumpMemory(DumpFilename);
        exit(EXIT_SUCCESS);
    }
    else if ( NumToGrep ) {
        bld_assert(UsingCache,("no cache to use as grep basis"));
        GetSpecials();
        GrepFiles(ToGrep[0],NumToGrep-1,&ToGrep[1]);
        exit(EXIT_SUCCESS);
    }
    else if ( ToWhence ) {
        bld_assert(UsingCache,("no cache to use as search basis"));
        GetSpecials();
        Whencer(LookupNode(ToWhence),"",WhenceRecursive);
        exit(EXIT_SUCCESS);
    }
    else if ( NumToFind ) {
        bld_assert(UsingCache,("no cache to use as search basis"));
        GetSpecials();
        FindFiles(NumToFind,ToFind);
        exit(EXIT_SUCCESS);
    }
    if ( !UsingCache ) {
        TELL(("Will create cache\n"));
        CrtSpecials();
    }
    if ( !LoadCmlVars() && !CacheOK ) {     /* FALSE on cml mismatch */
        UsingCache = FALSE;
        CrtSpecials();
        LoadCmlVars();
    }
    if ( UsingCache && !CleanCache() && !CacheOK ) {
        UsingCache = FALSE;
        CrtSpecials();
        LoadCmlVars();
    }
    GetSpecials();
    if ( !getcwd(StartDir, STRPATH+1) ) {
        printf("could not get current directory!\n");
        exit(1);
    }
    SetDir(StartDir);
    CreateDir(TempDir);
    if ( !UsingCache ) {
        ReadVarsFile();
        DoLocators();
        ReadBuildFiles();
    }
    else if ( ReCheckDirs )
        DoLocators();
    if ( UsingCache ) ReadSavedVars();
    if ( !UsingCache || ReCheckDirs ) {
        VerifyOutdirs();
        FixupMoved();
        WriteMemory(CacheFile);
    }
    SetDir(StartDir);
	SetLibDirs();
    if ( JustCleanAll ) {
        DeleteAll();
        exit(EXIT_SUCCESS);
    }
	if ( NumToAmkFile ) {
		TotalErrors = 0;
		GenPublicFiles(NumToAmkFile,ToAmkFile);
/*       exit(EXIT_SUCCESS); */
        exit(TotalErrors ? EXIT_FAILURE : EXIT_SUCCESS);
	}
    ReadCallFile();
if ( ReportStrictFiles ) /* TEMP */
DoReportStrictFiles(); /* TEMP */
    DoBuild();
    if ( FindAllDeps ) DoFindAllDeps();
    WriteMemory(CacheFile);
    CheckForProtoConflicts();
    if ( DoMemoryDump ) {
        SetDir(StartDir);
        DumpMemory(DumpFilename);
    }
    if ( TotalErrors ) {
        extern int NoReportErrMsg;
        NoReportErrMsg = TRUE;
        ErrorMsg("TOTAL ERRORS: %d",TotalErrors);
        if ( DoingReport )
            ErrorMsg("MODULE ERROR REPORTS in %s",ReportFile);
        NoReportErrMsg = FALSE;
        exit(EXIT_FAILURE);
    }
    exit(EXIT_SUCCESS);
}

/*************************************************************************
function: PutVar - Set a control var
*/
void PutVar (char *name, char *value) {
	char str[2*STRPATH];

	if ( value ) {
	    sprintf(str,"%s=%s",name,value);
	    CmlVars[NumCmlVars++] = strdup(str);
	}
	else
	    CmlVars[NumCmlVars++] = strdup(name);
}

/*************************************************************************
function: SetSysVars - Set basic control variables
Programatically set top level vars.  This is needed because of the
peculaiarity that the opensrc control files are derived from the Applix
control files, yet the opensrc control files have a different naming convention.

How command line, environment vars are handled should really be redone!!!!
*/
void SetSysVars ()
{
	char path[STRPATH+1];
	char axobj[STRPATH+1];

	PutVar("AOS",MyBuildTree);
	PutVar(GetMachineName(),NULL);
	PutVar("AOS_MACHINE",GetMachineName());
	if ( !GetVar("AOS_PGMS") ) {
		sprintf(path,"%s%s%s",MyBuildTree,DirSlash,Opensrc?"src":"pgms");
		PutVar("AOS_PGMS",path);
	}
	if ( !GetVar("AOS_EXEC") ) {
		sprintf(path,"%s%s%s",MyBuildTree,DirSlash,Opensrc?"exec":"axexec");
		PutVar("AOS_EXEC",path);
	}
	if ( !GetVar("AOS_AXOBJ") ) {
		sprintf(axobj,"%s%s%s",MyBuildTree,DirSlash,Opensrc?"obj":"axobj");
		PutVar("AOS_AXOBJ",axobj);
	}
	else
		strcpy(axobj,GetVar("AOS_AXOBJ"));

	if ( !GetVar("AOS_OBJECTS") ) {
		sprintf(path,"%s%s%s",axobj,DirSlash,
			Opensrc ? (Turbo?"turbo":"debug"):(Turbo?"axdata":"daxdata"));
		PutVar("AOS_OBJECTS",path);
	}
	if ( !GetVar("AOS_CPROTO_DIR") ) {
		sprintf(path,"%s%sproto",axobj,DirSlash);
		PutVar("AOS_CPROTO_DIR",path);
	}
	if ( !GetVar("AOS_LOCAL") ) {
		sprintf(path,"%s%s%s",MyBuildTree,DirSlash,Opensrc?"lsrc":"lpgms");
		PutVar("AOS_LOCAL",path);
	}
}

/*************************************************************************
function: CrtSpecials - Create node hash & aos_vars
*/
void CrtSpecials ()
{
    FlushMemory();
    assert(HASH_ARR_ID == AllocObj((AM_HASH_PRIME+1)*sizeof(ID),TYPE_ID));
    assert(RULE_ARR_ID == AllocObj(sizeof(ID)*NUM_RULES,TYPE_ID));
}
/*************************************************************************
function: GetSpecials - Get specials into memory
*/
void GetSpecials ()
{
    ID id;
    dir_info *dirP;

    HashArray = IdPtr(HASH_ARR_ID);
    RuleDirs = IdPtr(RULE_ARR_ID);
    NumOutDirs = 0;
    for ( id = NextType(0,TYPE_DIR);id; id = NextType(id,TYPE_DIR) ) {
        dirP = DirPtr(id);
        if ( dirP->is_altdir )
            AltOutDirId = id;
        else if ( dirP->is_outdir )
            OutDirs[NumOutDirs++] = dirP->dname;
    }
}

/*************************************************************************
function: Cstreq - Perform caseless string compare
*/
int Cstreq (a,b)
register char *a,*b;
{
    register char ca, cb;

    if ( !a || !b )
        return(FALSE);

    while ( *a ) {
        ca = *a++; cb = *b++;
        if ( islower(ca) ) ca = toupper(ca);
        if ( islower(cb) ) cb = toupper(cb);
        if ( ca != cb )
            return(FALSE);
    }
    if ( *b == 0 )
        return(TRUE);
    return(FALSE);
}

char *MoreArgs[] = {
    "-Dignore_me"
};

/*
 * function:  ReadCommandLine - Process command line args into global vars
*/
void ReadCommandLine(argc,passed_argv)
int argc;
char *passed_argv[];
{
    char **argv;
    int ix, jx;
    char *p;
    int more_args;
    int skip_arg = 0;
    int turbo_value;
    static char turbo_switch[15];
    static char cacheDir[STRPATH+1], tmpDir[STRPATH+1], objDir[STRPATH+1];

/* Insert standard args before all... unless still used on MS-Windpos, obsolete */

    argv = (char**)malloc( (argc+1)*sizeof(char*) + sizeof(MoreArgs) );
    more_args = sizeof(MoreArgs) / sizeof(char*);
    for ( ix = 1, jx = 0; jx < more_args; jx++,ix++ )
        argv[ix] = MoreArgs[jx];
    for ( jx = 1; jx < argc; jx++, ix++ )
        argv[ix] = passed_argv[jx];
    argv[0] = passed_argv[0];
    argc = argc + more_args;
    argv[argc] = NULL;

/* Allow turbo flag to obtain from environmemt */

    turbo_value = 0;
    if ( getenv("AOS_TURBO") != (char *)NULL )
       turbo_value = atoi( getenv("AOS_TURBO") );
    sprintf( turbo_switch, "TURBO=%d", turbo_value );


    AMakePath(objDir,MyBuildTree,Opensrc ? "obj" : "axobj");
    AMakePath(cacheDir,objDir,"caches");
	CmlVars[NumCmlVars++] = "AMAKE_V2=1";	/* obsolete, needed? */
    CmlVars[NumCmlVars++] = "ATM_ED=1 ";	/* obsolete, needed? */
    CmlVars[NumCmlVars++] = "ATM_CL=0";	/* obsolete, needed? */
    CmlVars[NumCmlVars++] = "ATM_GE=1";	/* obsolete, needed? */
    CmlVars[NumCmlVars++] = "ATM_DC=1";	/* obsolete, needed? */
    CmlVars[NumCmlVars++] = "ATM_SS=1";	/* obsolete, needed? */
    CmlVars[NumCmlVars++] = "ELFRT_ONLY=0";	/* obsolete, needed? */
    CmlVars[NumCmlVars++] = "ATM_PF=1";	/* obsolete, needed? */


/* now process all args */

    for ( ix = 1; ix < argc; ix++ ) {
        p = argv[ix];

        if ( Cstreq(p,"-cplus") )
            CmlVars[NumCmlVars++] = "CPLUS_ENABLED=1";
        else if ( Cstreq(p,"-icache") )
            LocAutoCache = TRUE;
        else if ( Cstreq(p,"-limited") )
            LimitedDebug = TRUE;
        else if ( Cstreq(p,"-cache") )
            UsingCache = FALSE;
        else if ( Cstreq(p,"-allproto") || Cstreq(p,"-allprotos") ) {
            NeedRebuiltProtos = TRUE;
            RebuildProtoOnAllCompiles = TRUE;
        }
else if ( Cstreq(p,"-report_strict") )/* TEMP */
ReportStrictFiles = TRUE;/* TEMP */
        else if ( Cstreq(p,"-noproto") )
            PretendBuild = FALSE;
        else if ( Cstreq(p,"-doproto") )
            RebuildProtoOnAllCompiles = TRUE;
        else if ( Cstreq(p,"-protos_ok") )
            SkipGeneratingProtos = TRUE;
        else if ( Cstreq(p,"-cacheok") )
            CacheOK = TRUE;
        else if ( Cstreq(p,"-Vincs") )
            VerifyIncludes = TRUE;
        else if ( Cstreq(p,"-narrow") )
            NarrowMode = TRUE;
        else if ( Cstreq(p,"-wide") )
            NarrowMode = FALSE;
        else if ( Cstreq(p,"-Vvars") )
            TellVars = TRUE;
        else if ( Cstreq(p,"-Vcmds") )
            TellCmds = TRUE;
        else if ( Cstreq(p,"-alldeps") )
            FindAllDeps = TRUE;
        else if ( Cstreq(p,"-dump") ) {
            DoMemoryDump = TRUE;
            DumpFilename = argv[++ix];
        }
        else if ( Cstreq(p,"-dumponly") ) {
            JustDump = TRUE;
            DumpFilename = argv[++ix];
        }
        else if ( Cstreq(p,"-report") )
            DoingReport = TRUE;
        else if ( Cstreq(p,"-nodelete") )
            KeepGarbage = TRUE;
        else if ( Cstreq(p,"-clean") )
            JustCleanAll = TRUE;
        else if ( Cstreq(p,"-touchonly") )
            TouchBuild = TRUE;
        else if ( Cstreq(p,"-fcp") ) {
            CprotoForcing = TRUE;
            forceBuild = TRUE;
        }
        else if ( Cstreq(p,"-cpmacro") ) {
            PretendBuild = TRUE;    /* now the default */
/*            CProtoMacro = "AMKP_";  /* default */
        }
        else if ( Cstreq(p,"-cpmacroname") ) {
            PretendBuild = TRUE;    /* now the default */
            CProtoMacro = argv[++ix];
        }
        else if ( Cstreq(p,"-rs6000log") )
            RS6000Log = argv[++ix];
        else if ( Cstreq(p,"-commands") ) {
			CommandLog = TRUE;
		}
        else if ( Cstreq(p,"-trace") ) {
            Trace = TRUE;
            Verbose = TRUE;
			CommandLog = TRUE;
        }
        else if ( Cstreq(p,"-logging") ) {
            Logging = TRUE;
            Verbose = TRUE;
            Trace = TRUE;
			CommandLog = TRUE;
        }
        else if ( Cstreq(p,"-debug") ) {
            Debug = TRUE;
            Logging = TRUE;
            Verbose = TRUE;
            Trace = TRUE;
			CommandLog = TRUE;
        }
        else if ( Cstreq(p,"-app") ) {
            for ( ix++; ix < argc; ix++ ) {
                p = argv[ix];
                if ( *p == '-' ) { ix--; break; }
                assert(NumAppsToBuild < 100);
                AppNamesToBuild[NumAppsToBuild++] = p;
            }
            continue;
        }
        else if ( Cstreq(p,"-grep") ) {
            for ( ix++; ix < argc; ix++ ) {
                p = argv[ix];
                if ( *p == '-' ) { ix--; break; }
                assert(NumToGrep < 100);
                ToGrep[NumToGrep++] = p;
            }
            continue;
        }
        else if ( Cstreq(p,"-whence") ) {
            Quiet = TRUE;
            ToWhence = argv[++ix];
        }
        else if ( Cstreq(p,"-rwhence") ) {
            Quiet = TRUE;
            ToWhence = argv[++ix];
            WhenceRecursive = TRUE;
        }
        else if ( Cstreq(p,"-ctl") ) {
			strcpy(VarsFile,argv[++ix]);
		}
        else if ( Cstreq(p,"-gen_public") ) {
            for ( ix++; ix < argc; ix++ ) {
                p = argv[ix];
                if ( *p == '-' ) { ix--; break; }
	            assert(NumToAmkFile < 1000);
	            ToAmkFile[NumToAmkFile++] = p;
			}
			UsingCache = FALSE;
            continue;
        }
        else if ( Cstreq(p,"-find") ) {
            Quiet = TRUE;
            for ( ix++; ix < argc; ix++ ) {
                p = argv[ix];
                if ( *p == '-' ) { ix--; break; }
                assert(NumToFind < 1000);
                ToFind[NumToFind++] = p;
            }
            continue;
        }
        else if ( streq(p,"-allow") ) {
            char *name;
            int jx;
            ix++; 
            p = argv[ix];
            if ( *p == '-' ) 
                { 
                ix--; 
                break; 
                }
            for ( jx = 0; *(name = RuleArray[jx].name); jx++ ) {
                if ( 0 == strcmp(p,name) ) {
                    RuleArray[jx].allow = TRUE;
                    break;
                }
            }
            if ( !*(name) ) {
                ErrorMsg("unknown rule to allow - %s",p);
                exit(3);
            }
            continue;
        }
        else if ( Cstreq(p,"-shared") ) {
            CmlVars[NumCmlVars++] = "BUILD_SHLIBS=1";
        }
        else if ( Cstreq(p,"-static") || Cstreq(p, "-noshared") ) {
            CmlVars[NumCmlVars++] = "BUILD_SHLIBS=0";
        }
        else if ( streq(p,"-") ) {
            forceBuild = FALSE;
            continue;
        }
	    else if ( Cstreq(p,"-o") ) {
#if AXMS
	        TempDir = AMakePath(tmpDir,cacheDir,"turbo_al");
#else
	        TempDir = AMakePath(tmpDir,cacheDir,"turbo_all");
#endif
	        CmlVars[NumCmlVars++] = "BLD_FONTS=0";
	        CmlVars[NumCmlVars++] = "TURBO=1";
			Turbo = TRUE;
	    }
	    else if ( Cstreq(p,"-orel") ) {
#if AXMS
	        TempDir = AMakePath(tmpDir,cacheDir,"turbo_al");
#else
	        TempDir = AMakePath(tmpDir,cacheDir,"turbo_all");
#endif
	        CmlVars[NumCmlVars++] = "BLD_FONTS=1";
	        CmlVars[NumCmlVars++] = "START_ANLM=1";
	        CmlVars[NumCmlVars++] = "TURBO=1";
	        CmlVars[NumCmlVars++] = "RELEASE_VERSION=1";
			Turbo = TRUE;
		}
	    else if ( Cstreq(p,"-rel") ) {
	        TempDir = AMakePath(tmpDir,cacheDir,"dbx_all");
            CmlVars[NumCmlVars++] = "BLD_FONTS=1";
            CmlVars[NumCmlVars++] = "START_ANLM=1";
            CmlVars[NumCmlVars++] = "TURBO=0";
            CmlVars[NumCmlVars++] = "RELEASE_VERSION=1";
	    }

        else if ( *p == '-' ) {
            p++;
            switch ( *p ) {
            case 'D':
                CmlVars[NumCmlVars++] = argv[ix]+2;
            CASE 'T':
            case 't':
                TempDir = argv[ix]+2;
            CASE 'Q':
            case 'q':
                Quiet = TRUE;
            CASE 'F':
            case 'f':
                forceBuild = TRUE;
            CASE 'V':
                Verbose = TRUE;
          		CommandLog = TRUE;
            CASE 'v':
                Verbose = TRUE;
         		CommandLog = TRUE;
            DEFAULT:
                ErrorMsg("unknown switch number %d  <%s>",ix,argv[ix]);
                Usage();
                exit(1);
            }
        }
        else {
            if ( forceBuild )
                ModsToForce[BuildModsCnt] = TRUE;
            ModsToBuild[BuildModsCnt++] = p;
        }
    }
	if ( !TempDir || streq(TempDir,TEMPNAME) ) {
#if PORTING
	    /* force use of meta option */
	    ErrorMsg("Must use a Meta switch ");
	    Usage();
	    exit(1);
#endif
        TempDir = AMakePath(tmpDir,cacheDir,"dbx_all");
        CmlVars[NumCmlVars++] = "BLD_FONTS=0";
        CmlVars[NumCmlVars++] = turbo_switch;
	}
}

/*************************************************************************
function: CleanCache - Verify validity & clean old data from cache
*/
int CleanCache ()
{
    ID id;
    register node* nodeP;
    ctrlfile_info *cP;
    dir_info *dirP;
    long dt;

    for ( id = NextType(0,TYPE_CTRLFILE);id;id = NextType(id,TYPE_CTRLFILE) ){
        cP = CtrlFilePtr(id);
        dt = GetSimpleDtmod(cP->name);
        if ( dt && (cP->dtmod != dt) ) {
            TELL(("%s changed, creating new cache\n",cP->name));
            return(FALSE);
        }
    }

/* ERROR on recursive checks: subordnate can change with paret change */

    if ( LocAutoCache ) {
        for ( id = NextType(0,TYPE_DIR); id; id = NextType(id,TYPE_DIR) ) {
            dirP = DirPtr(id);
            if ( !dirP->is_locdir && !dirP->is_checkdir ) continue;
            if ( dirP->dtmod != GetSimpleDtmod(dirP->dname) ) {
                TELL(("%s changed, will re-search it\n",dirP->dname));
                dirP->loc_visited = FALSE;
                ReCheckDirs = TRUE;
            }
#if AXMS
            if ( (stricmp(dirP->dname, "\\AX\\LPGMS") == 0) ||
                 (stricmp( (dirP->dname+2), "\\AX\\LPGMS") == 0) )
                {
                /* the test withe "+2" is in case there is a drive specifier */
                TELL(("%s changed, will re-search it\n",dirP->dname));
                dirP->loc_visited = FALSE;
                ReCheckDirs = TRUE;
                }
#endif

        }
    }
    for ( id = NextType(0,TYPE_NODE); id; id = NextType(id,TYPE_NODE)) {
        nodeP = NodePtr(id);
        nodeP->up_to_date = FALSE;
        nodeP->busted = FALSE;
        nodeP->exists = FALSE;
        nodeP->dtmod = 0;
        nodeP->haveProtoed = FALSE;
    }
    return(TRUE);
}

/*************************************************************************
function: LoadCmlVars - Load/Verify command line variables
*/
int LoadCmlVars ()
{
    char temp[100], *p;
    char *newvars[100];
    char *oldvars[100];
    int ix, jx;
    ID id;
    int oldcount = 0;

    if ( !UsingCache ) {
        for ( ix = 0; ix < NumCmlVars; ix++ ) {
            strcpy(temp,CmlVars[ix]);
            VERBOSE(("Command line definer %s\n",temp));
            AlcAStr(temp,TYPE_CMLVAR);
            for ( p = temp; *p ; p++ )
                if ( *p == '=' )
                    break;
            if ( *p == 0 ) {
                AddAppVar((FILE *) NULL, temp,"1");
                AddVar(temp,"1");
                if ( TellVars )
                    VERBOSE(("#define %s 1\n",temp));
            }
            else {
                *p = 0;
                AddAppVar((FILE *) NULL, temp,p+1);
                AddVar(temp,p+1);
                if ( TellVars )
                    VERBOSE(("#define %s %s\n",temp,p+1));
            }
        }
        return(TRUE);
    }
    for ( ix = 0; ix < NumCmlVars; ix++ )
        newvars[ix] = CmlVars[ix];
    for ( id = NextType(0,TYPE_CMLVAR); id; id = NextType(id,TYPE_CMLVAR) )
        oldvars[oldcount++] = CmlVarPtr(id);
    if ( oldcount != NumCmlVars ) {
        TELL(("Command line def count changed, creating new cache\n"));
        UsingCache = FALSE;
        return(FALSE);
    }
    for ( ix = 0; ix < oldcount; ix++ ) {
        for ( jx = 0; jx < oldcount; jx++ )
            {
#if AXMS
            if ( Cstreq(oldvars[jx],newvars[ix]) )
#else
            if ( streq(oldvars[jx],newvars[ix]) ) 
#endif
               break;
            }

        if ( jx == oldcount ) {
            TELL(("Command line defines changed, creating new cache\n"));
            UsingCache = FALSE;
            return(FALSE);
        }
        oldvars[jx] = "";   /* to keep matches from happening */
    }
    return(TRUE);
}

/*************************************************************************
function:  AddAppVar - Add Var to acpp command vector
Also adds it as a macro var so we can easily verify existance of required
settings.
*/
void AddAppVar (fp,name,val)
char *name;
char *val;
FILE *fp;
{
    int len;
    char *p;

    assert(NumAppVars != MAX_APPVARS);

    if (fp == NULL) {
        len = strlen(name) + strlen(val) + 4;
        AppVars[NumAppVars++] = p = malloc(len);
        strcpy(p,"-D");
        strcat(p,name);
        strcat(p,"=");
        strcat(p,val);
    }
    else {
        fprintf (fp, "#define %s %s\n", name, val);
    }
}

/*************************************************************************
function:  ReadVarsFile - Read the alis.vars file
*/
void ReadVarsFile ()
{
    int fd;
    int junk, ix;
    static char temp3[256];
    static char inc_srch[256];
    static char inc_srch2[256];
    char ZZdefs[STRPATH+1];     /* #defines in the vars file */
    char ZZvars[STRPATH+1];     /* post-processed vars file */
    char *p;
    int init = TRUE;
	char *name = VarsFile;
    char version[STRPATH+1];     /* Base version file */
	version_info *info;
	FILE *fp;

	if (Opensrc) {
		/* Read the version file for the SHELF base code */
		sprintf(version,"%s%sversion",MyBuildTree,DirSlash);
		/* if version doesn't exist, look for dversion */
		if ((fp = fopen(version, "r")) == NULL) {
			sprintf(version,"%s%sdversion",MyBuildTree,DirSlash);
		} else {
			fclose(fp);
		}
		info = ReadVersionInfo( version);
		BaseMajorVersion = atoi(info->project.major_version);
		BaseMinorVersion = atoi(info->project.minor_version);

		/* Get the version info from the .ctl file for the current project */
		NoteDeltaDirs(0,0,"");
	}

	if ( !name[0] ) {	/* Look in default places for .ctl file */
		/* Look for all.ctl in pgms/axbuild and lpgms */
	    sprintf(name,"%s%slpgms%sall.ctl",MyBuildTree,DirSlash,DirSlash);
		if ( GetSimpleDtmod(name) ) goto gotctl;
	    sprintf(name,"%s%spgms%saxbuild%sall.ctl",MyBuildTree,DirSlash,DirSlash,DirSlash);
		if ( GetSimpleDtmod(name) ) goto gotctl;
		/* Look for opensrc.ctl in src/axbuild */
		sprintf(name,"%s%slsrc%sopensrc.ctl",MyBuildTree,DirSlash,DirSlash);
		if ( GetSimpleDtmod(name) ) goto gotctl;
		sprintf(name,"%s%ssrc%saxbuild%sopensrc.ctl",MyBuildTree,DirSlash,DirSlash,DirSlash);
	}
	if ( !GetSimpleDtmod(name) ) {
		printf("%s: No control file found.\n",name);
		exit(1);
	}
gotctl:
    AMakePath(ZZvars,TempDir,"zzvars");
    AMakePath(ZZdefs,TempDir,"zzdefs");
    AMakePath(ZZtemp,TempDir,"zztemp");
    unlink(ZZtemp);
    AMakePath(ZZtempout,TempDir,"zztempo");
    sprintf(inc_srch,"-I%s",name);
    p = FindEntryName(name);
    inc_srch[p-name+2] = 0;
	if ( Opensrc )
		sprintf(inc_srch2,"-I%s%ssrc%saxbuild",MyBuildTree,DirSlash,DirSlash);
	else
		sprintf(inc_srch2,"-I%s%spgms%saxbuild",MyBuildTree,DirSlash,DirSlash);
	AppVars[3] = inc_srch2;
    AppVars[2] = inc_srch;
    sprintf(temp3,"-T%s",ZZdefs);
    AppVars[1] = temp3;
    if ( 0 != AppSub(name,ZZvars) ) /* use preprocessor */
        aborter("partially processed vars file %s",name);
    fd = open(ZZvars,0,0);
    bld_assert(fd != -1,("Could not open file '%s'",ZZvars));
    InputLine = 0;
	strcpy(InputFilename,ZZvars);
    ErrorBuf = LineBuf;
    while ( DoLine(init,fd,LineBuf,&junk) ) {
        init = FALSE;
        if ( LineBuf[0] == DR_CHAR ) {
            ForceLocating = FALSE;
            DoDirective(LineBuf);
        }
        else if ( ForceLocating )
            CrtForced(LineBuf);
        else if ( !DoAssignment(LineBuf) )
            ErrorMsg("syntax error");
    }
    close(fd);

	/* read local ctl files */
	for ( ix=0; ix<n_ctl_files; ix++ )
	{
		fd = open(ctl_file[ix], 0, 0);
		if ( fd != (-1) )
		{
			InputLine = 0;
			strcpy(InputFilename, ctl_file[ix]);
			ErrorBuf = LineBuf;
			init = TRUE;
			while ( DoLine(init, fd, LineBuf, &junk) )
			{
				if ( init )
					init = FALSE;
				if ( LineBuf[0] == DR_CHAR )
				{
					ForceLocating = FALSE;
					DoDirective(LineBuf);
				}
				else if ( ForceLocating )
					CrtForced(LineBuf);
				else if ( !DoAssignment(LineBuf) )
					ErrorMsg("syntax error");
			}
			close(fd);
		}
	}

    NoteVars(ZZdefs);
    AppVars[1] = "-Q";
    InputLine = 0;

/* save permanent vars */

{
    int cnt = 0;
    ID *idp;
    extern ID VarsIdArray;

    for ( ix = 0; ix < NumMacros; ix++ ) {
        if ( (MacroFlags[ix] & MF_USED) && (MacroFlags[ix] & MF_PERM) )
            cnt++;
    }
    VarsIdArray = AllocObj(sizeof(ID)*cnt*2,TYPE_ID);
    idp = IdPtr(VarsIdArray);
    for ( ix = 0; ix < NumMacros; ix++ ) {
        if ( (MacroFlags[ix] & MF_USED) && (MacroFlags[ix] & MF_PERM) ) {
            *idp++ = ObjStr(MacroNames[ix]);
            *idp++ = ObjStr(MacroValues[ix]);
        }
    }
}
}

/*************************************************************************
function: ReadSavedVars - Load the permanent vars
*/
void ReadSavedVars ()
{
    int ix, cnt = 0;
    ID *idp;
    extern ID VarsIdArray;

    cnt = SizeObj(VarsIdArray)/(sizeof(ID)*2);
    idp = IdPtr(VarsIdArray);
    for ( ix = 0; ix < cnt; ix++ )
        AddVar(StrPtr(idp[2*ix]),StrPtr(idp[2*ix+1]));
}

/*************************************************************************
function:  DoLine - Get a build/locate file line
*/
int DoLine(init,fd,buf,lenP)
int init;
int fd;
char *buf;
int *lenP;
{
    char *line;
    int len;
    char temp[MAX_LINELEN];
    extern int UsingCaches;
    char *readaline();

    if ( FileCache == NULL )
        FileCache = malloc(SZ_FILE_CACHE);

    line = buf;
    *line = 0;
read_some:
    if ( !readaline(fd,line,MAX_LINELEN,FileCache,SZ_FILE_CACHE,init) )
        return(FALSE);

    init = FALSE;
    InputLine++;

    if ( *line == COMMENT_CHAR )
        goto read_some;

    CleanWhiteSpace(line);

    if ( *line == '#' ) {
        char *p;
        char file[STRPATH+1];
        int len = 0;
        *(line+5) = 0;
        assert(streq(line+1,"line"));
        p = line+6;
        while ( *p ) {
            if ( *p == ' ' ) {
                while ( (*p != '"') && *p ) p++;
                assert(*p);
                p++;
                while ( (*p != '"') && *p ) file[len++] = *p++;
                assert(*p);
                file[len] = 0;
                if ( !streq(file,ZZtempout) )
                    AddCtrlFile(file);
                *p = 0;
            }
            p++;
        }
        InputLine = atoi(line+6);
        goto read_some;
    }

    if ( *line == 0 )               /* blank line */
        goto read_some;

    ExpandVars(line,temp,&len);
    if ( *temp == 0 )               /* expanded to blank line */
        goto read_some;
    strcpy(line,temp);

    if ( line[len-1] == CONTINUE_CHAR ) {   /* loop on continueations */
        line = line + len - 1;
        goto read_some;
    }

    *lenP = len;
    return(TRUE);
}

/*************************************************************************
function: readaline - Read a line from file, return status
*/
char *readaline (fd,buf,max_len,cache,cache_size,init)
int fd;
register char *buf;
register int max_len;
register char *cache;
int cache_size;
int init;
{
    register char *wrtP;        /* ptr into passed buffer */
    static int chars_left;      /* how many left in cache */
    static int charpos;         /* where in cache to read from */
    register char the_char;

    if ( init ) {
        chars_left = 0;
        charpos = 0;
    }
    wrtP = buf;
    while ( TRUE ) {
        if ( chars_left ) {
            chars_left--;
            the_char = cache[charpos++];
            if ( the_char == '\n' ) {
do_nl:
                *wrtP = 0;
                return(buf);
            }
            else if ( (wrtP-buf) < (max_len-1) )
                *wrtP++ = the_char;
        }
        else {
            chars_left = read(fd,cache,cache_size);
            charpos = 0;
            if ( chars_left == 0 ) {
                if ( wrtP != buf )
                    return(buf);
                return(NULL);
            }
            else if ( chars_left < 0 ) {
                chars_left = 0;
                return(NULL);
            }
        }
    }
}

/*************************************************************************
function:  CleanWhiteSpace - Get rid of extra spaces, tabs ...
*/
void CleanWhiteSpace(passed_line)
char *passed_line;
{
    register char *line;
    register int got_space = FALSE;
    register char *p;

    p = line = passed_line;

    while ( (*line == ' ') || (*line == TAB) )  /* get past leading spaces */
        line++;

    while ( *line ) {
        if ( (*line == ' ') || (*line == TAB) ) {
            if ( !got_space ) {
                *p++ = ' ';
                got_space = TRUE;
            }
            line++;
        }
        else if ( (p != passed_line) && (*line == COMMENT_CHAR) )
            break;
        else {
            *p++ = *line++;
            got_space = FALSE;
        }
    }
    *p = 0;
    if ( p == passed_line )
        return;

    /* trim trailing newline and/or space */

    if ( *(p-1) == '\n' )
        *(--p) = 0;
    if ( p == passed_line )
        return;
    if ( *(p-1) == ' ' )
        *(p-1) = 0;
}

/*************************************************************************
function:  ExpandVars - Expand bld vars
*/
void ExpandVars(line,resultP,lenP)
register char *line;
register char *resultP;
int *lenP;
{
    char varbuf[MAX_VARNAME+2];
    register char *src, *dst, ch;
    int ix;
    char *start;
    int re_space_out = FALSE;

    start = resultP;
    while ( *line ) {
        if ( (*line == VAR_FLAG) && ( *(line+1) == '(' ) ) {
            dst = varbuf;
            src = line+2;
            for ( ix = 0; ix < MAX_VARNAME+1; ix++ ) {
                ch = *dst++ = *src++;
                if ( (0 == ch) || (' ' == ch) )
                    goto punt;
                if ( ')' == ch ) {
                    line = src;
                    *(dst-1) = 0;
                    src = GetVar(varbuf);
                    if ( src ) {
                        LOGGER(("%s expanded to '%s'",varbuf,src));
                        strcpy(resultP,src);
                        resultP += strlen(src);
                    }
                    re_space_out = TRUE;
                    goto didit;
                }
            }
        }
punt:
        *resultP++ = *line++;
didit:  ;
    }
    *resultP++ = 0;
    *lenP = resultP - start - 1;
    if ( re_space_out )
        CleanWhiteSpace(start);
}

/*************************************************************************
function:  DoAssignment - Parse for assingment statement
*/
int DoAssignment(line)
char *line;
{
    register char ch, *p;
    int ix;

    p = line;
    for ( ix = 0; ix < MAX_VARNAME+2; ix++, p++ ) {
        ch = *p;
        if ( ch == '=' )
            break;
        else if ( (ch == ' ') || (ch == 0) )
            return(FALSE);
    }
    if ( ix == MAX_VARNAME+2 )
        return(FALSE);

    if ( *(p-1) == ' ' )    /* Null term name area */
        *(p-1) = 0;
    else
        *p = 0;

    p++;
    if ( *p == ' ' )    /* trim lead space from value */
        *p++ = 0;

    AddVar(line,p);

    return(TRUE);
}

/*************************************************************************
function:  AddVar - Add a variable
*/
void AddVar (name,value)
register char *name, *value;
{
    register int ix;
    char *ExpandMetasAndMalc();

    for ( ix = 0; ix < NumMacros; ix++ ) {
        if ( !(MacroFlags[ix] & MF_USED) )
            continue;
        if ( streq(name,MacroNames[ix]) ) {
            free(MacroNames[ix]);
            free(MacroValues[ix]);
            goto got_ix;
        }
    }
    for ( ix = 0; MacroFlags[ix] & MF_USED; ix++ );
got_ix:
    assert(ix < MAX_VARS);
    NumMacros = MAX(ix+1,NumMacros);
    MacroNames[ix] = AllcStr(name);
    MacroFlags[ix] = MacroDfltFlag;
    if ( !value ) value = "";
    MacroValues[ix] = ExpandMetasAndMalc(value);
    LOGGER(("AddVar: added var %s val %s pos %d",name,value,ix));
}

/*************************************************************************
function: ClipVars - Get rid of non-permannt vars
*/
void ClipVars ()
{
    int ix;

    for ( ix = 0; ix < NumMacros; ix++ ) {
        if ( (MacroFlags[ix] & MF_USED) && !(MacroFlags[ix] & MF_PERM) ) {
            free(MacroNames[ix]);
            free(MacroValues[ix]);
            MacroFlags[ix] = 0;
            continue;
        }
    }
}

/*************************************************************************
function: ExportVar - Make a var permanent
*/
void ExportVar (name)
char *name;
{
    int ix;

    for ( ix = 0; ix < NumMacros; ix++ ) {
        if ( !(MacroFlags[ix] & MF_USED) )
            continue;
        if ( streq(name,MacroNames[ix]) ) {
            MacroFlags[ix] |= MF_PERM;
            return;
        }
    }
    ErrorMsg("no such macro - %s",name);
}

/*************************************************************************
function: ExpandMetasAndMalc - return malcd of meta expanded string
*/
char *ExpandMetasAndMalc (str)
char *str;
{
    char *p;
    int num_toks;
    int bytes;
    char *pfx, *sfx;
    char *to_pfx, *to_sfx, *to_add, *head, *px;
    char *to_free, *to_ret;
    char *dst;
    int ix, jx, len;

#define PFX_STR "^prefix"
#define SFX_STR "^suffix"
#define RESFX_STR "^resuffix"
#define MERGE_STR "^merge_for_unique"

    p = malloc(len = strlen(str)+1);
    if ( *str == '^' ) {
        num_toks = TokenizeString(str,p);
        if ( num_toks < 3 ) {
            str = "";
            goto copy;
        }
        if ( streq(p,MERGE_STR) ) {
            to_free = p;
            dst = to_ret = malloc(len+1);
            p += strlen(p)+1;       /* get past the merge word */
            to_add = head = p;
            num_toks -= 1;
            for ( ix = 0; ix < num_toks; ix++ ) {
                px = head;
                for ( jx = 0; jx < ix; jx++ ) {
                    if ( streq(px,to_add) )
                        goto next_merge;
                    px += strlen(px)+1;
                }
                strcpy(dst,to_add);
                strcat(dst," ");
                dst += strlen(dst);
next_merge:;
                to_add += strlen(to_add)+1;
            }
            *(dst-1) = 0;
            free(to_free);
            return(to_ret);
        }
        else if ( streq(p,PFX_STR) ) {
            to_free = p;
            pfx = p + strlen(p) + 1;
            to_pfx = pfx + strlen(pfx) + 1;
            num_toks -= 2;
            bytes = strlen(str)-strlen(PFX_STR)-strlen(pfx);
            bytes += num_toks * strlen(pfx);
            dst = p = malloc(bytes+1);
            for ( ix = 0; ix < num_toks; ix++, to_pfx += strlen(to_pfx)+1 ) {
                strcpy(dst,pfx);
                strcat(dst,to_pfx);
                strcat(dst," ");
                dst += strlen(dst);
            }
            *(dst-1) = 0;
            free(to_free);
            return(p);
        }
        else if ( streq(p,SFX_STR) ) {
            to_free = p;
            sfx = p + strlen(p) + 1;
            to_sfx = sfx + strlen(sfx) + 1;
            num_toks -= 2;
            bytes = strlen(str)-strlen(SFX_STR)-strlen(sfx);
            bytes += num_toks * strlen(sfx);
            dst = p = malloc(bytes+1);
            for ( ix = 0; ix < num_toks; ix++, to_sfx += strlen(to_sfx)+1 ) {
                strcpy(dst,to_sfx);
                strcat(dst,sfx);
                strcat(dst," ");
                dst += strlen(dst);
            }
            *(dst-1) = 0;
            free(to_free);
            return(p);
        }
        else if ( streq(p,RESFX_STR) ) {
            to_free = p;
            sfx = p + strlen(p) + 1;
            to_sfx = sfx + strlen(sfx) + 1;
            num_toks -= 2;
            bytes = strlen(str)-strlen(RESFX_STR)-strlen(sfx);
            bytes += num_toks * strlen(sfx);
            dst = p = malloc(bytes+1);
            for ( ix = 0; ix < num_toks; ix++, to_sfx += strlen(to_sfx)+1 ) {
                strcpy(dst,to_sfx);
                for ( jx = strlen(to_sfx); jx != 0; jx-- )
                    if ( to_sfx[jx] == '.' )
                        break;
                if ( jx )
                    dst[jx] = 0;
                strcat(dst,sfx);
                strcat(dst," ");
                dst += strlen(dst);
            }
            *(dst-1) = 0;
            free(to_free);
            return(p);
        }
        else {
            strcpy(p,str);
            return(p);
        }
    }
    else {
copy:
        strcpy(p,str);
        return(p);
    }
}

/*************************************************************************
function:  GetVar - Get value of a variable or NULL
*/
char *GetVar(name)
char *name;
{
    register int ix;
    char* val;
    char* getenv();
    char* env_var;
    char  buf[MAX_VARNAME+10];
	int len;

    if(strncmp(name,"AOS_",4) != 0) { 
        /* If the var-name doesn't have AOS_ prefix, add the prefix for
           the env variable name to avoid clash with some other env variable name */
        sprintf(buf,"AOS_%s",name);
        env_var = buf;
    }
    else
        env_var = name;
    if((val=getenv(env_var)) != NULL)
        return(val);
    for ( ix = 0; ix < NumMacros; ix++ ) {
        if ( !(MacroFlags[ix] & MF_USED) )
            continue;
        if ( streq(name,MacroNames[ix]) )
            return(MacroValues[ix]);
    }
	len = strlen(name);
	for ( ix = 0; ix < NumCmlVars; ix++ ) {
		if ( streq(CmlVars[ix],name) )
			return("1");
		else if ( strncmp(CmlVars[ix],name,len) == 0 && CmlVars[ix][len] == '=' ) {
			return(CmlVars[ix]+len+1);
		}
	}

    return(NULL);
}

/*************************************************************************
function: NeedFullVar - Fully resolve a var
*/
char *NeedFullVar (char *name) {
	static char result[10000];	/* plenty, tested in a lazy manner */
	char buf[10000], tok[10000]; /* plenty, tested in a lazy manner */
	char *p, *bufP, *newTok;
	int toklen;

	p = GetVar(name);
	if ( !p )
		return("");
	bufP = buf;
	while ( *p ) {
		toklen = 0;
		while ( *p && !strchr(" \t\n/\\,.()*&^:|",*p) )
			tok[toklen++] = *p++;
		tok[toklen] = 0;
        bld_assert(bufP-buf < sizeof(buf)/2,("expansion too long",buf));
		newTok = NeedFullVar(tok);
		if ( newTok[0] )
			strcpy(bufP,newTok);
		else
			strcpy(bufP,tok);
		bufP += strlen(bufP);
		if ( *p )
			*bufP++ = *p++;
	}
	*bufP = 0;
	strcpy(result,buf);
	return(result);
}

/*************************************************************************
function: NeedVar - Just like GetVar but returns "" on not found
*/
char *NeedVar (name)
char *name;
{
    if ( NULL == (name = GetVar(name)) )
        return("");
    return(name);
}

/*************************************************************************
function:  AppSub - Kick off acpp
*/
int AppSub (infile,outfile)
char *infile, *outfile;
{
    FILE *tempin, *tempout;
    char *p;
    char dirname[STRPATH+1];
    char acpp[STRPATH+1];
    int ret;

    sprintf(acpp,"%s/acpp",MyBuildTree);
    AppVars[0] = acpp;

/* Linux needs a way to distinguish between different hardware architectures
   during processing of .amk files, but acpp doesn't figure this out
   automatically and send the appropriate defines as does the system cc */
#if __linux__
#if __i386__
	AddAppVar(NULL,"__i386__","1");
#elif __alpha__
	AddAppVar(NULL,"__alpha__","1");
#elif __sparc__
	AddAppVar(NULL,"__sparc__","1");
#elif __powerpc__
	AddAppVar(NULL,"__powerpc__","1");
#endif
#endif
	
    AddCtrlFile(infile);
    tempin = fopen (ZZtemp, "r");
    unlink (ZZtempout);
    tempout = fopen (ZZtempout, "w+");
    bld_assert(tempout,("Could not create file '%s'",ZZtempout));
    if ( tempin ) {
        while ( fgets(LineBuf,MAX_LINELEN,tempin) ) {
            fputs (LineBuf, tempout);
        }
        fclose (tempin);
    }
    fprintf(tempout,"#include \"%s\"\n",infile);
    fclose (tempout);
    sprintf(dirname,"-I%s",infile);
    p = FindEntryName(dirname);
    *(p-1) = 0;
    AppVars[NumAppVars] = dirname;
    AppVars[NumAppVars+1] = ZZtempout;
    AppVars[NumAppVars+2] = outfile;

    ret = VpShcmd(AppVars);
    if ( ret != 0 )
      {
      char msg[1200];
      int i;
      ErrorMsg("acpp command failed, command:");
      msg[0] = '\0';
      for ( i=0; i<NumAppVars+2; i++ )
         {
         strcat( msg, AppVars[i] );
         strcat( msg, " " );
         }
      ErrorMsg("%s\n", msg );
      }
    return( ret );
}

/*************************************************************************
function: AddCtrlFile - Uniquely add name of a control file
*/
void AddCtrlFile (name)
char *name;
{
    register ID id;
    ctrlfile_info *p;

    for ( id = NextType(0,TYPE_CTRLFILE); id ; id = NextType(id,TYPE_CTRLFILE)){
        if ( streq(name,CtrlFilePtr(id)->name) )
            return;
    }
    id = AllocObj(sizeof(ctrlfile_info),TYPE_CTRLFILE);
    p = CtrlFilePtr(id);
    strcpy(p->name,name);
    p->dtmod = GetSimpleDtmod(name);
}

/*************************************************************************
function:  NoteVars - Note acpp defines in my vars
*/
void NoteVars (filename)
char *filename;
{
    char *p, *line;
    FILE *fp;
    FILE *tempfp;
    int len;

    fp = fopen(filename,"r");
    unlink (ZZtemp);
    tempfp = fopen (ZZtemp, "w");
    while ( fgets(LineBuf,MAX_LINELEN,fp) ) {
        p = line = LineBuf;
        len = strlen(p);
        if ( p[len-1] == '\n' )
            p[--len] = 0;
        while ( *p && (*p != ' ') ) p++;
        if ( *p == ' ' ) {
            *p++ = 0;
            while ( *p == ' ' ) p++;
        }
        AddAppVar(tempfp,line,p);
        AddVar(line,p);
        if ( TellVars )
            VERBOSE(("#define %s %s\n",line,p));
    }
    fclose (tempfp);
    fclose(fp);
}

/*************************************************************************
function: AddLibPath - Add a library path to LIB_PATH
*/
void AddLibPath (char *name) {

	char all[10000];
	char *old = getenv(LIB_PATH);
	if ( old && old[0] )
		sprintf(all,"%s=%s%s%s",LIB_PATH,name,DirSep,old);
	else
		sprintf(all,"%s=%s",LIB_PATH,name);
	putenv(strdup(all));
}

/*************************************************************************
function: SetLibDirs - Add shared-lib-tagged dors to LIBRARY_PATH
*/
void SetLibDirs () {
	int id;
	dir_info *dirP;

    for ( id = NextType(0,TYPE_DIR); id; id = NextType(id,TYPE_DIR) ) {
        dirP = DirPtr(id);
		if ( dirP->is_shlib_dir )
			AddLibPath(dirP->dname);
	}
}

/*************************************************************************
function:  DoDirective - Perform a 'dot' directive
*/
void DoDirective (str)
register char *str;
{
    char *name;
    char *the_dir;
    char *p;
    dir_info *dirP;
    char *Pstreq();
    ID id;
    char tokens[MAX_DIRECTIVE_LEN+1];
    char *tokP;
    int ix, count;

    if ( p = Pstreq(str,DR_DEFCOMP) ) {
        comp_info *cp;
        dir_info *dp;
        name = str = p;
        while ( *str ) {
            if ( *str == ' ' ) {
                *str = 0;
                the_dir = ++str;
                break;
            }
            str++;
        }
        bld_assert(*str,("defcomp syntax error"));
        id = LookupComp(name);
        bld_assert(CompObjCreated,("component %s already .decomp'ed",name));
        cp = CompPtr(id);
        cp->person = 0;
        cp->dflt = LookupDir(the_dir);
        dp = DirPtr(cp->dflt);
        bld_assert((dp->owner == 0) || (dp->owner == id),
                        ("directory %s already has owner",the_dir));
        dp->owner = id;
    }
    else if ( p = Pstreq(str,DR_CPROTO_EXCLUDE) ) {
        bld_assert(!CprotoExcludeFile,
        ("Already have a cproto exclude file: %s vs. %s",CprotoExcludeFile,p));
        CprotoExcludeFile = AllcStr(p);
    }
    else if ( p = Pstreq(str,DR_STRICT_EXCLUDE) ) {
        bld_assert(!StrictExcludeFile,
        ("Already have a strict exclude file: %s vs. %s",StrictExcludeFile,p));
        StrictExcludeFile = AllcStr(p);
    }
    else if ( p = Pstreq(str,DR_STRICT_COMPILES) ) {
        StrictCompiles = atoi(p);
    }
    else if ( p = Pstreq(str,DR_CPYRIGHT_EXCLUDE) ) {
        bld_assert(!CpyrightExcludeFile,
        ("Already have a cpyright exclude file: %s vs. %s",CpyrightExcludeFile,p));
        CpyrightExcludeFile = AllcStr(p);
    }
    else if ( p = Pstreq(str,DR_CPYRIGHT_NOWARN) ) {
        bld_assert(!CpyrightNoWarnFile,
        ("Already have a cpyright no-warn file: %s vs. %s",CpyrightNoWarnFile,p));
        CpyrightNoWarnFile = AllcStr(p);
    }
    else if ( p = Pstreq(str,DR_OPENSRC_BUILTFILE) ) {
        bld_assert(!OpensrcBuiltFile,
        ("Already have an opensrc built file: %s vs. %s",OpensrcBuiltFile,p));
        OpensrcBuiltFile = AllcStr(p);
    }
    else if ( p = Pstreq(str,DR_OPENSRC_EXCLUDE) ) {
        bld_assert(!OpensrcExcludeFile,
        ("Already have an opensrc exclude file: %s vs. %s",OpensrcExcludeFile,p));
        OpensrcExcludeFile = AllcStr(p);
    }
    else if ( p = Pstreq(str,DR_OPENSRC_RENAME) ) {
        bld_assert(NumRenameDirs < MAX_RENAMES,("too many .opensrc_renames"));
        RenameDirs[NumRenameDirs++] = p = AllcStr(p);
        bld_assert(p,("renamed must include 'old new' names."));
		p = strchr(p,' ');
		*p = 0;
	}
    else if ( p = Pstreq(str,DR_BUILD) ) {
        p = AllcStr(p);
        assert(NumBuildFiles < MAX_BUILD_FILES);
        BuildFiles[NumBuildFiles++] = p;
    }
    else if ( p = Pstreq(str, ".chmod_dir") )
       {
       count = TokenizeString(p,tokens);
       bld_assert( count == 2 , (".chmod_dir takes 2 arguments"));
       p = tokens + strlen(tokens)+1;
       AddDirToChmodList( tokens, p );
       }

    else if ( p = Pstreq(str,DR_TELL) ) {
        TELL(("%s\n",p));
    }
    else if ( p = Pstreq(str,DR_ERROR) ) {
        ErrorMsg("%s",p);
    }
    else if ( p = Pstreq(str,DR_TEMPDIR) ) {
        TempDir = p;
        CreateDir(TempDir);
    }
    else if ( (p = Pstreq(str,DR_FRCLOC)) || (p = Pstreq(str,DR_LOCATE)) ) {
        ForceDir = LookupDir(p);
        ForceLocating = TRUE;
    }
    else if ( p = Pstreq(str,DR_OUTDIR) ) {
        bld_assert(NumOutDirs < MAX_OUTDIRS,("too many .outdirs"));
        OutDirs[NumOutDirs++] = (dirP = DirPtr(LookupDir(p)))->dname;
        dirP->is_outdir = TRUE;
    }
    else if ( p = Pstreq(str,DR_ALTDIR) ) {
        AltOutDirId = id = LookupDir(p);
        DirPtr(id)->is_altdir = TRUE;
    }
    else if ( p = Pstreq(str,DR_DEFDIR) ) {
        char *namep;
        count = TokenizeString(p,tokens);
        bld_assert((count == 2) && (NumLogNames < MAX_LOGICAL_DIRS),
            ("too many .defdirs"));
        LogicalDirNames[NumLogNames] = namep = AllcStr(tokens);
        p = tokens + strlen(tokens)+1;
        DirNameMap[NumLogNames++] = id = LookupDir(p);
        dirP = DirPtr(id);
        if ( !dirP->is_defdir )
            dirP->defdirname = ObjStr(namep);
        dirP->is_defdir = TRUE;
    }
    else if ( p = Pstreq(str,DR_SHLIB_DIR) ) {
		DirPtr(LookupDir(p))->is_shlib_dir = TRUE;
	}
    else if ( p = Pstreq(str,DR_CHECKDIR) ) {
        count = TokenizeString(p,tokens);
        if ( !((count >= 1) && (count < MAX_DIR_USERS+2)) ) {
            ErrorMsg(".checkdir has no directory value, check #defines");
            return;
        }
        dirP = DirPtr(id = LookupDir(tokens));
        dirP->recursive = TRUE;
        dirP->is_checkdir = TRUE;
    }
    else if ( p = Pstreq(str,DR_LOCDIR) ) {
        count = TokenizeString(p,tokens);
        if ( !((count >= 1) && (count < MAX_DIR_USERS+2)) ) {
            ErrorMsg(".locdir has no directory value, check #defines");
            return;
        }
        dirP = DirPtr(id = LookupDir(tokens));
        DirPtr(id)->recursive = TRUE;
        dirP->is_locdir = TRUE;
    }
    else if ( p = Pstreq(str,DR_1CHECKDIR) ) {
        count = TokenizeString(p,tokens);
        if ( !((count >= 1) && (count < MAX_DIR_USERS+2)) ) {
            ErrorMsg(".check1dir has no directory value, check #defines");
            return;
        }
        dirP = DirPtr(id = LookupDir(tokens));
        dirP->is_checkdir = TRUE;
    }
    else if ( p = Pstreq(str,DR_1LOCDIR) ) {
        count = TokenizeString(p,tokens);
        if ( !((count >= 1) && (count < MAX_DIR_USERS+2)) ) {
            ErrorMsg(".loc1dir has no directory value, check #defines");
            return;
        }
        dirP = DirPtr(id = LookupDir(tokens));
        dirP->is_locdir = TRUE;
    }
    else if ( p = Pstreq(str,DR_OBJDIR) ) {
        bld_assert(ObjDir == 0,(".objdir already set"));
        ObjDir = LookupDir(p);
    }
    else if ( p = Pstreq(str,DR_RULEDIR) ) {
        dispatch_rule *ruleP;
        extern dispatch_rule *LocateRule();
        count = TokenizeString(p,tokP = tokens);
        bld_assert(count == 2,("badly formatted .ruledir"));
        ruleP = LocateRule(tokens);
        bld_assert(ruleP,("no such rule: %s",tokens));
        tokP += strlen(tokP)+1;
        assert(NUM_RULES > ruleP-RuleArray);
        RuleDirs[ruleP-RuleArray] = id = LookupDir(tokP);
        TELL(("%s rules go to %s\n",ruleP->name,DirPtr(id)->dname));
    }
    else
        ErrorMsg("undefined command %s",str);
}

/*************************************************************************
function: CrtForced - Create nodes of relocated modules
*/
void CrtForced (str)
char *str;      /* string of name tokens */
{
    char *tokens[2000];
    int count, ix;
    node *nodeP;

    count = FilterToTokArray(str,tokens);
    for ( ix = 0; ix < count; ix++ ) {
        nodeP = LookupNode(tokens[ix]);
        nodeP->relocated = TRUE;
        nodeP->is_global = TRUE;
        nodeP->dirid = ForceDir;
    }
}

static void
add_ctl_file(char *dir, char *project_name)
{
	char cfile[STRPATH+1], **ptr;
	int sz;

	sprintf(cfile, "%s/%s.ctl", dir, project_name);
	if ( n_ctl_files%10 == 0 )
	{
		sz = sizeof(char *)*(10+n_ctl_files);
		ptr = (char **)realloc(ctl_file, sz);
		if ( ptr == NULL )
		{
			ErrorMsg("error generating .ctl file list");
			return;
		}
		ctl_file = ptr;
	}
	ctl_file[n_ctl_files] = (char *)malloc(strlen(cfile)+1);
	if ( ctl_file[n_ctl_files] == NULL )
		ErrorMsg("error generating .ctl file list");
	else
	{
		strcpy(ctl_file[n_ctl_files], cfile);
		n_ctl_files++;
	}
}

/*************************************************************************
function: NoteDeltaDirs - Recursively process version files
Changed mind - does NOT fully recurse.  Just recurse one level
to verify version numbers.
*/
void NoteDeltaDirs (int wantMajorVersion, int wantMinorVersion, char *dirStem) {
	char vfile[STRPATH+1],dirname[STRPATH+1];
	char *p, *line;
	dir_info *dirP;
	char lineBuf[MAX_LINELEN];
	int version = 0;
	FILE *fp;
	version_info *info=NULL;
	int ix;

	if ( wantMajorVersion == 0 ) {
		dirStem = NeedVar("AOS_DELTA_BUILD");
		if ( !dirStem[0] )
			return;
	}
	sprintf(dirname,"%s/delta/%s",MyBuildTree,dirStem);
	add_ctl_file(dirname, dirStem);
	sprintf(vfile,"%s/%s.ctl",dirname, dirStem);
	info = ReadVersionInfo( vfile);

	/* If we don't have the specified version of the SHELF base,
		give a warning */
	if ( atoi(info->base.major_version) < BaseMajorVersion )
		wrn_msg("%s: Specified base version %d.%d is lower than actual base %d.%d",
					dirStem, atoi(info->base.major_version), 
					atoi(info->base.minor_version),
					BaseMajorVersion, BaseMinorVersion);
	else if ( atoi(info->base.major_version) > BaseMajorVersion )
		wrn_msg("%s: Specified base version %d.%d is higher than actual base %d.%d",
					dirStem, atoi(info->base.major_version), 
					atoi(info->base.minor_version),
					BaseMajorVersion, BaseMinorVersion);
	else if ( atoi(info->base.minor_version) < BaseMinorVersion )
		wrn_msg("%s: Specified base version %d.%d is lower than actual base %d.%d",
					dirStem, atoi(info->base.major_version), 
					atoi(info->base.minor_version),
					BaseMajorVersion, BaseMinorVersion);
	else if ( atoi(info->base.minor_version) > BaseMinorVersion )
		wrn_msg("%s: Specified base version %d.%d is higher than actual base %d.%d",
					dirStem, atoi(info->base.major_version), 
					atoi(info->base.minor_version),
					BaseMajorVersion, BaseMinorVersion);

	/* Warn about version mis-match (unless caller specifies
		version zero, which means "don't care") */
	if (wantMajorVersion) {
		if ( atoi(info->project.major_version) < wantMajorVersion )
			wrn_msg("%s: version %d.%d is lower than specified %d.%d",
					dirStem, atoi(info->project.major_version), 
					atoi(info->project.minor_version),
					wantMajorVersion, wantMinorVersion);
		else if ( atoi(info->project.major_version) > wantMajorVersion )
			wrn_msg("%s: version %d.%d is higher than specified %d.%d",
					dirStem, atoi(info->project.major_version), 
					atoi(info->project.minor_version),
					wantMajorVersion, wantMinorVersion);
		else if ( atoi(info->project.minor_version) < wantMinorVersion )
			wrn_msg("%s: version %d.%d is lower than specified %d.%d",
					dirStem, atoi(info->project.major_version), 
					atoi(info->project.minor_version),
					wantMajorVersion, wantMinorVersion);
		else if ( atoi(info->project.minor_version) > wantMinorVersion )
			wrn_msg("%s: version %d.%d is higher than specified %d.%d",
					dirStem, atoi(info->project.major_version), 
					atoi(info->project.minor_version),
					wantMajorVersion, wantMinorVersion);
	}

	/* First time thru, caller passes version zero, because we don't
		know what version number the main project will be. This signals
		the main project, and we can recurse one level */
	else {
		for (ix=0; ix<info->nSubProjects; ix++)
			NoteDeltaDirs(atoi(info->subProjects[ix].major_version),
								atoi(info->subProjects[ix].minor_version),
								info->subProjects[ix].project_name);
	}
	Opensrc = TRUE;	/* only opensrc build supports delta chains */

	dirP = DirPtr(LookupDir(dirname));
	dirP->recursive = TRUE;
	dirP->is_checkdir = TRUE;

	if ( !VarsFile[0] ) {
		sprintf(vfile,"%s/opensrc.ctl",dirname,dirStem);
		if ( GetSimpleDtmod(vfile) )
			strcpy(VarsFile,vfile);
	}
}

/*************************************************************************
function: ReadVersionInfo - Extract info from .ctl file

In SHELF 1.3 and before, this was a separate VERSION file.
After 1.3, this became the project.ctl file.
*/
version_info *
ReadVersionInfo (char *vfile) {
	char *p, *line;
	dir_info *dirP;
	char lineBuf[MAX_LINELEN];
	int first_line=0;
	int major_version=0, minor_version=0;
	int major_uses=0, minor_uses=0;
	char project[MAX_LINELEN];
	FILE *fp;
	version_info *info;
	version_line *version_array=NULL;
	int nProjects=0;
	int ix, jx;

	fp = fopen(vfile,"r");
	bld_assert(fp,("Could not open file '%s'",vfile));

	while ( fgets(lineBuf,MAX_LINELEN,fp) ) {	
		line = lineBuf;
		CleanWhiteSpace(line);
		/* Some lines obviously aren't the version lines */
		if ( *line == 0 || *line == '#' || *line == '.' )
			continue;

		/* The project version should be indicated in this format: */
				/* VERSION x.y */
		/* Yes, the information is placed in a comment in the file.
		   Note x is the major version number, and y is an OPTIONAL
		   minor version number (if no y, the period is optional too).
		   Keep looking until we find this line.
		*/
		if ( !first_line ) {
			sscanf( line, "/* VERSION %d.%d */", 
								&major_version, &minor_version);
			if (major_version > 0) {
				info = (version_info*)malloc( sizeof(version_info));
				sprintf( info->project.major_version, "%d", major_version);
				sprintf( info->project.minor_version, "%d", minor_version);
				first_line = 1;
			}
			continue;
		}

		/* Subsequent lines should be of the format: */
				/* USES Project-name x.y */
		/* Build up an array of this info (we need to extract the base
		   info and then sort backwards).
		*/
		project[0] = '\0';
		major_uses = minor_uses = 0;
		sscanf( line, "/* USES %s %d.%d */", 
						&project, &major_uses, &minor_uses);
		if (major_uses > 0) {
			bld_assert(project[0],
					("%s: %s: Missing or bad project name",vfile,line));
			bld_assert(major_uses > 0,
					("%s: %s: Missing or bad version number",vfile,line));

			/* If this is the base code indicator, it gets stored specially */
			if (!strcmp( project, "Base")) {
				sprintf( info->base.major_version, "%d", major_uses);
				sprintf( info->base.minor_version, "%d", minor_uses);
			}

			/* Else it is one of the subprojects */
			else {
				if (!nProjects)
					version_array = 
							(version_line*)malloc( sizeof(version_line));
				else
					version_array = (version_line*)realloc( version_array,
							(nProjects+1) * sizeof(version_line));
				strcpy( version_array[nProjects].project_name, project);
				sprintf( version_array[nProjects].major_version, "%d", 
														major_uses);
				sprintf( version_array[nProjects].minor_version, "%d", 
														minor_uses);
				++nProjects;
			}
		}
	} /* while */

	/* If we didn't find the version number, complain */
	bld_assert(major_version > 0,
					("%s: Missing or bad version number",vfile));

	/* Copy the subprojects into the info array, reversing the order.
		The subprojects specified LAST in the .ctl file have lower
		priority when files conflict (aside from base code, which is
		lowest priority of all). They will be put FIRST in the info
		array, so that the "file finding" code will overwrite files
		found there with files found in projects later in the info list
		(i.e. specified earlier in the .ctl file) */
	if (nProjects > 1)
		info = (version_info*)realloc( info, sizeof(version_info) + 
							((nProjects-1) * sizeof(version_line)));

	for (ix=0; ix<nProjects; ix++) {
		jx = nProjects - ix - 1;
		strcpy( info->subProjects[ix].project_name, 
								version_array[jx].project_name);
		strcpy( info->subProjects[ix].major_version, 
								version_array[jx].major_version);
		strcpy( info->subProjects[ix].minor_version, 
								version_array[jx].minor_version);
	}
	info->nSubProjects = nProjects;
	if (version_array)
		free( version_array);
	fclose(fp);
	return( info);
}


/*************************************************************************
function:  DoLocators - Read the locator files
This relies on directory order in ObjTable matching the vars file.
*/
void DoLocators ()
{
	int doing_locs = TRUE;  /* start with locdirs, then checkdirs */
    int priority = 1;
    char the_dir[STRPATH+1], name[STRPATH+1], *afdir;
    dir_info *dirP, *topP;
    node *nodeP;
    ID top_id, dir_id;
    int code, maxlevels;

/* Perform the locdirs, then the checkdirs */

locloop:
    top_id = 0;
    while ( TRUE ) {
        if ( !(top_id = NextType(top_id,TYPE_DIR)) ) break;
        afdir = (topP = dirP = DirPtr(top_id))->dname;
        if ( (doing_locs && !dirP->is_locdir) ||
             (!doing_locs && !dirP->is_checkdir) ||
             (dirP->loc_visited) ) continue;
        dirP->loc_visited = TRUE;
        dirP->dtmod = GetSimpleDtmod(dirP->dname);
        TELL(("Searching %s\n",afdir));
        maxlevels = (dirP->recursive ? 99 : 0);
        if ( !doing_locs ) dirP->priority = priority++;
        dir_id = top_id;
        while ( TRUE ) {
            code = AFindFiles(afdir,maxlevels,the_dir,name);
            afdir = NULL;
            if ( code == 1 ) goto locdone;
            else if ( code == 2 ) {
                TELL(("could not stat %s\n",name));
                continue;
            }
            assert((code == 0) || (code == 3));
            if ( !streq(the_dir,dirP->dname) ) {
                dirP = DirPtr(dir_id = LookupDir(the_dir));
                if ( !dirP->loc_visited ) {
                    VERBOSE(("Searching %s\n",the_dir));
                    dirP->dtmod = GetSimpleDtmod(dirP->dname);
                    dirP->is_checkdir = !(dirP->is_locdir = doing_locs);
                    if ( !doing_locs ) dirP->priority = priority++;
                    dirP->owner = topP->owner;
                    dirP->loc_visited = TRUE;
                }
                if ( code == 3 )
                    continue;
            }
            nodeP = LookupForMe(name,dirP->owner);
            if ( nodeP->ruleid )
                continue;
            if ( nodeP->relocated && !streq(NodeDirName(nodeP),the_dir) ) {
                VERBOSE(("    %s already .locate'd in %s\n",
                            name,NodeDirName(nodeP)));
                nodeP->conflict = TRUE;
                continue;
            }
            TRACE(("   found %s in %s\n",name,the_dir));
            if ( !nodeP->located && nodeP->dirid && (nodeP->dirid != dir_id) )
                nodeP->moved = NodesMoved = TRUE;
            if ( nodeP->located && doing_locs && (nodeP->dirid != dir_id) &&
                 (NodeCompId(nodeP) == dirP->owner) ) {
                if ( !(streq("makefile",name)) )
                  TELL(("    %s already in %s\n",name,NodeDirName(nodeP)));
                nodeP->conflict = TRUE;
            }
			if ( !doing_locs ) {
				if ( nodeP->dirid && Opensrc )
					CompareVersions(nodeP->dirid,dir_id,nodeP->name);
			}
            nodeP->dirid = dir_id;
            nodeP->located = TRUE;
        }
locdone: ;
    }
    if ( doing_locs ) {
        doing_locs = FALSE;
        goto locloop;
    }
}

/*************************************************************************
function: FixupMoved - Anybody referencing a moved file needs built
*/
void FixupMoved ()
{
    register node *nodeP;
    register ID id, x;
    register int i;
    int incount, depcount;

    for ( id = NextType(0,TYPE_NODE); id; id = NextType(id,TYPE_NODE) ) {
        if ( !(nodeP = NodePtr(id))->moved ) continue;
        TELL(("%s has been moved\n",nodeP->name));
        for ( x = NextType(0,TYPE_NODE); x; x = NextType(x,TYPE_NODE) ) {
            nodeP = NodePtr(id);
            incount = NodeInCount(nodeP);
            depcount = NodeDepCount(nodeP);
            for ( i = 0; i < (incount + depcount); i++ ) {
                if ( NodeInId(nodeP,i) == id )
                    nodeP->refers_to_moved = TRUE;
            }
        }
    }
}

#define ASC_REVTAG "Applix ASC version: "

/*************************************************************************
function: CompareVersions - Compared checkdir revstamps against locdir revstamps
*/
void CompareVersions (old_dir, new_dir, name)
ID old_dir, new_dir;
char *name;
{
	char srcfile[STRPATH+1], prevfile[STRPATH+1];
	char *srcImage, *prevImage;
	char *srcVersion, *prevVersion;
	char *p, *q;

	if (streq(name, "VERSION"))
		return;
	AMakePath(srcfile,DirPtr(new_dir)->dname,name);
	srcImage = readFileIntoBuf(srcfile);
	p = strstr(srcImage,ASC_REVTAG);
	if ( !p ) {
		ErrorMsg("%s: missing Applix ASC tag",srcfile);
		free(srcImage);
		return;
	}
	srcVersion = p + strlen(ASC_REVTAG);
	q = strchr(srcVersion, '\n');
	*q = '\0';

	AMakePath(prevfile,DirPtr(old_dir)->dname,name);
	prevImage = readFileIntoBuf(prevfile);
	p = strstr(prevImage,ASC_REVTAG);
	if ( !p ) {
		ErrorMsg("%s: missing Applix ASC tag",prevfile);
		free(prevImage);
		free(srcImage);
		return;
	}
	prevVersion = p + strlen(ASC_REVTAG);
	q = strchr(prevVersion, '\n');
	*q = '\0';
	if ( !streq(srcVersion,prevVersion) ) {
		ErrorMsg("%s versions don't match",name);
		printf("    File %s: %s\n",srcfile,srcVersion);
		printf("    Overrides %s: %s\n",prevfile,prevVersion);
	}
	free(prevImage);
	free(srcImage);
}

/*************************************************************************
function: ReadExclusionFiles - Process cproto/strict compile lists
*/
static void ReadExclusionFiles ()
{
    char *p;
    FILE *fp;
    char *filename;
    int len;
    char *exclude;
    node *nodeP;
    int group;
    char name[256];
	char *names[6];

	names[0] = CprotoExcludeFile;
	names[1] = StrictExcludeFile;
	names[2] = CpyrightExcludeFile;
	names[3] = OpensrcExcludeFile;
	names[4] = CpyrightNoWarnFile;
	names[5] = OpensrcBuiltFile;
    for ( group = 0; group < 6; group++ ) {
        filename = names[group];
        if ( !filename )    /* no exclusions */
            continue;
        bld_assert(NodeFilename(LookupNode(filename),name,TRUE),
                ("Could not locate file '%s'",filename));
        fp = fopen(name,"r");
        bld_assert(fp,("Could not open file '%s'",name));
        VERBOSE(("Reading %s...\n",name));
        while ( fgets(LineBuf,MAX_LINELEN,fp) ) {
            p = LineBuf;
            len = strlen(p);
            if ( p[len-1] == '\n' )
                p[--len] = 0;
            while ( *p && (*p == ' ' || *p == '\t') ) p++;
            if ( *p == '#' || *p == 0 )
                continue;
            exclude = p;
            while ( *p && (*p != ' ') && (*p != '\t') ) p++;
            *p = 0;
            nodeP = LookupNode(exclude);
			switch ( group ) {
			case 0: nodeP->cproto_excluded = TRUE; break;
            case 1: nodeP->strict_excluded = TRUE; break;
            case 2: nodeP->cpyright_excluded = TRUE; break;
            case 3: nodeP->opensrc_excluded = TRUE; break;
            case 4: nodeP->cpyright_nowarn = TRUE; break;
            case 5: nodeP->opensrc_builtfile = TRUE; break;
			}
        }
        fclose(fp);
    }
}

/*************************************************************************
function:  ReadBuildFiles - Preprocess & read the .amk files
*/
static void ReadBuildFiles ()
{
    int ix;
    char tempfile[256];
    int passcount = 0;
    char name[STRPATH+1];
    ID id;

TELL(("Reading amk files...\n"));
for ( passcount = 0; passcount < 2; passcount++ ) {
    for ( ix = 0; ix < NumBuildFiles; ix++ ) {
        AMakePath(tempfile,TempDir,FindEntryName(BuildFiles[ix]));
        strcpy(name,FindEntryName(BuildFiles[ix]));
        if ( streq(name,BuildFiles[ix]) ) {     /* no path */
            if ( !NodeFilename(LookupNode(name),name,TRUE) ) {
                ErrorMsg("Cannot locate '%s'",name);
                continue;
            }
        }
        else
            strcpy(name,BuildFiles[ix]);   /* use the hardcoded path */
        if ( passcount == 0 ) {
            VERBOSE(("Reading (%d) %s...\n",passcount+1,name));
            if ( 0 != AppSub(name,tempfile) )
                aborter("partially processed build file %s",tempfile);
        }
        else {
            VERBOSE(("Reading (%d) %s...\n",passcount+1,tempfile));
        }
        MacroDfltFlag = MF_USED;
        ReadABuildFile(tempfile,passcount);
        ClipVars();
    }
}
BuildPhase = TRUE;     /* supresses line numbers in error msgs */
ReadExclusionFiles();
}

/*************************************************************************
function:  ReadABuildFile - Read the build file
Uses the locate data unless the file has a hardcoded path
*/
void ReadABuildFile (name,passcount)
char *name;
int passcount;
{
    int fd;
    int junk;
    ID CompId = 0;
    char *p;
    ID id;
    int ix, count, doing_locs;
    int init = TRUE;
    char *toks[5000];

    fd = open(name,0,0);
    bld_assert(fd != -1,("Could not open file '%s'",name));

    doing_locs = (passcount == 0);
    InputLine = 0;
	strcpy(InputFilename,name);
    ErrorBuf = LineBuf;
    while ( DoLine(init,fd,LineBuf,&junk) ) {
        init = FALSE;
        if ( LineBuf[0] == DR_CHAR ) {
            if ( p = Pstreq(LineBuf,DR_COMPONENT) ) {
                CompId = LookupComp(p);
                if ( !CompPtr(CompId)->dflt ) {
                    if ( !ObjDir )
                        ErrorMsg(".objdir not set, no such component - %s",p);
                    else {
                        char dirname[STRPATH+1];
                        AMakePath(dirname,DirPtr(ObjDir)->dname,p);
                        CompPtr(CompId)->dflt = id = LookupDir(dirname);
                        DirPtr(id)->owner = CompId;
                    }
                }
            }
            else if ( p = Pstreq(LineBuf,DR_EXPORTVAR) ) {
                if ( doing_locs ) continue;
                ExportVar(p);
            }
            else if ( p = Pstreq(LineBuf,DR_TELL) ) {
                if ( !doing_locs ) continue;
                TELL(("%s\n",p));
            }
            else if ( p = Pstreq(LineBuf,DR_LOCATE) ) {
                if ( !doing_locs ) continue;
                count = FilterToTokArray(p,toks);
                if ( count < 1 ) {
                    ErrorMsg("badly formatted .locate line");
                    continue;
                }
                if ( !CompId && (count >= 2) ) {
                    ErrorMsg(".locate preceedes .component - %s",toks[1]);
                    continue;
                }
                for ( ix = 0; ix < NumLogNames; ix++ ) {
                    if ( streq(toks[0],LogicalDirNames[ix]) )
                        break;
                }
                if ( ix == NumLogNames )
                    ErrorMsg("not a known dir - %s",p);
                else
                    AddLocs(&toks[1],count-1,DirNameMap[ix],CompId);
            }
            else if ( p = Pstreq(LineBuf,DR_GLOBAL) ) {
                node *nodeP;
                if ( !doing_locs ) continue;
                count = FilterToTokArray(p,toks);
                for ( ix = 0; ix < count; ix++ ) {
                    nodeP = LookupForMe(toks[ix],CompId);
                    nodeP->is_global = TRUE;
                    nodeP->comp_id = CompId;
                }
            }
            else if ( p = Pstreq(LineBuf,DR_EXTERN) ) { /* unsuppoprted!! */
                if ( !doing_locs ) continue;
                count = FilterToTokArray(p,toks);
                for ( ix = 0; ix < count; ix++ )
                    LookupNode(toks[ix])->is_global = TRUE;
            }
            else {
                ErrorMsg("bad directive");
            }
        }
        else if ( DoAssignment(LineBuf) );
        else if ( !doing_locs && DoRule(CompId,LineBuf) );
        else if ( !doing_locs )
            ErrorMsg("syntax error");
    }
    close(fd);
}

/*************************************************************************
function: AddLocs - Add some .amk locators
*/
void AddLocs (tokens,num_toks,dir,comp)
char *tokens[];
int num_toks;
ID dir;
ID comp;
{
    int ix;
    node *nodeP;

    for ( ix = 0; ix < num_toks; ix++ ) {
        nodeP = LookupOwnedNode(tokens[ix],comp);
        if ( (nodeP->relocated) && (dir != nodeP->dirid) )
            wrn_msg("%s already .located",nodeP->name);
        else {
            nodeP->relocated = TRUE;
            nodeP->is_global = TRUE;
            nodeP->dirid = dir;
            nodeP->comp_id = comp;
        }
    }
}

/*************************************************************************
function:  DoRule - Process line as a rule statement
*/
int DoRule (CompId,line)
int CompId;
char *line;
{
#define rule_error(body) {line_error body;return(TRUE);}
#define MAX_RULE_TOKS 4000

    dispatch_rule *ruleP;   /* ptr to rule object to use */
    ID with_id, id;
    int RuleId;
    int inputs_count;
    int outputs_count;
    int depends_count;
    char **inputsP;    /* array of inputs */
    char **outputsP;   /* array of outputs */
    char **dependsP;   /* array of dependencies */
    char **CreateMatchArray();
    register char *p;
    int ix;
    char withstr[MAX_RULE_TOKS+1];
    char *tokens[MAX_RULE_TOKS+1];
    int num_tokens, cur_tok;
    int user_deps;
    extern dispatch_rule RuleArray[];
    extern dispatch_rule *LocateRule();


    LOGGER(("\nDoRule: %s\n",line));

/* Convert to a token stream */

    num_tokens = FilterToTokArray(line,tokens);

/* Clip empty 'with' or 'dep' */

    if ( streq(tokens[num_tokens-1],KEY_WITH) ||
         streq(tokens[num_tokens-1],KEY_DEPEND) )
            tokens[--num_tokens] = NULL;

/* Extract and locate rule */

    if ( num_tokens < 2 )
        rule_error(("rule has no arguments, syntax error"));
    ruleP = LocateRule(tokens[0]);
    if ( ruleP == NULL )
        rule_error(("unknown rule, %s",tokens[0]));
    RuleId = ruleP - RuleArray;

/* Count how many input nodes */

    for ( cur_tok = 1; cur_tok < num_tokens; cur_tok++ )
        if ( streq(tokens[cur_tok], KEY_INTO) )
            goto got_into_word;
    rule_error(("rule has no output"));
got_into_word:
    if ( cur_tok == 1 )
        rule_error(("rule has no inputs"));

    inputsP = tokens + 1;
    inputs_count = cur_tok - 1;
    if ( (inputs_count != 1) && (!ruleP->multi_inputs) )
        rule_error(("rule accepts only one input"));

/* Count how many outputs */

    for ( outputsP = &tokens[++cur_tok]; cur_tok < num_tokens; cur_tok++ ) {
        p = tokens[cur_tok];
        if ( streq(p,KEY_WITH) || streq(p,KEY_DEPEND) )
            break;
    }
    outputs_count = cur_tok - inputs_count - 2;
    if ( ruleP->matched_io && streq(MATCH_PSUEDO,*outputsP) ) {
        outputsP = CreateMatchArray(ruleP,inputsP,inputs_count);
        outputs_count = inputs_count;
    }

    if ( ruleP->matched_io && (inputs_count != outputs_count) )
            rule_error(("inputs count does not match outputs count %/%d",
                inputs_count,outputs_count));

    if ( (outputs_count != 1) && (!ruleP->multi_outputs) )
        rule_error(("rule accepts only one output"));

/* note user's 'with' string */

    withstr[0] = 0;
    if ( (cur_tok < num_tokens) && streq(tokens[cur_tok],KEY_WITH) ) {
        if ( !ruleP->ok_with )
            rule_error(("rule does not accept 'with' arguments"));
        for ( cur_tok++, p = withstr; cur_tok < num_tokens; cur_tok++ ) {
            if ( streq(tokens[cur_tok],KEY_DEPEND) )
                break;
            strcpy(p,tokens[cur_tok]);
            p += strlen(p);
            *p++ = ' ';
        }
        if ( p > withstr )  /* get rid of trailing space */
            *(p-1) = 0;
    }

/* Isolate the 'dep' list */

    if ( (cur_tok < num_tokens) && streq(tokens[cur_tok],KEY_DEPEND) ) {
        dependsP = tokens + cur_tok + 1;
        depends_count = num_tokens - (cur_tok + 1);
        user_deps = TRUE;
    }
    else {
        dependsP = tokens + num_tokens;
        depends_count = 0;
        user_deps = FALSE;
    }
    depends_count += AddRuleDeps(ruleP,tokens+num_tokens);
    with_id = ObjStr(withstr);
 
/* Fill in the node based on data collected.
   Note that matched multi-outputs are split into multiple fill-ins here.
   Other rules with multi-outputs do a dependent split.
*/
    if ( ruleP->matched_io ) {
        for ( ix = 0; ix < outputs_count; ix++ )
            FillInNode(outputsP[ix],CompId,RuleId,1,inputsP+ix,
                        depends_count,dependsP,with_id, user_deps);
    }
    else if ( outputs_count != 1 ) {
        id = FillInNode(outputsP[0],CompId,RuleId,inputs_count,inputsP,
                    depends_count,dependsP, with_id, user_deps);
        NodePtr(id)->has_siblings = TRUE;
        for ( ix = 1; ix < outputs_count; ix++ )
            FillInNode(outputsP[ix],CompId,TRUE_NOP_RULE,1,
                    outputsP,0,(char**)0,0, user_deps);
    }
    else
        FillInNode(outputsP[0],CompId,RuleId,inputs_count,inputsP,
                    depends_count,dependsP, with_id, user_deps);
    return(TRUE);
}

/*************************************************************************
function:  CreateMatchArray - Temp match rule scheme
*/
char **CreateMatchArray (ruleP,inputs,count)
dispatch_rule *ruleP;
char **inputs;
int count;
{
#define MAX_MATCHES 5000
    int ix;
    static char *new[MAX_MATCHES];
    static char toks[MAX_MATCHES*10];
    char *p = toks;
    char *endtoks = toks + sizeof(toks);


    assert(streq(ruleP->name,"compile"));
    assert(count < MAX_MATCHES);
    for ( ix = 0; ix < count; ix++ ) {
        assert((p + strlen(inputs[ix]) + 1) < endtoks);
        strcpy(p,inputs[ix]);
        p[strlen(p)-2] = '\0';
        strcat (p, OEXT);
        new[ix] = p;
        p += strlen(p)+1;
    }
    return(new);
}

/*************************************************************************
function: AddRuleDeps - Add implicit rule dependencies
Returns how many were added
*/
int AddRuleDeps (ruleP,toks)
register dispatch_rule *ruleP;
register char **toks;
{
    register int ix;
	char *cmd;

    if ( !ruleP->dep_cmd )
        return(0);
    for ( ix = 0; ruleP->dep_cmd[ix]; ix++ ) {
		cmd = ruleP->dep_cmd[ix];
		if ( *cmd == '*' )
			cmd = FindDefine(cmd+1);
		bld_assert(cmd && cmd[0],("%s not defined for rule %s",ruleP->dep_cmd[ix],ruleP->name));
        toks[ix] = cmd;
	}
    return(ix);
}

/*************************************************************************
function: VerifyOutdirs - Make sure outputs go into proper dirs
*/
void VerifyOutdirs ()
{
    register int jx;
    register char *name;
    char *NodeDirName();
    register node *nodeP;
    ID id;

    for ( id = NextType(0,TYPE_NODE); id; id = NextType(id,TYPE_NODE) ) {
        nodeP = NodePtr(id);
        if ( !OUTPUT_NODE(nodeP) || !nodeP->dirid ) continue;
        name = NodeDirName(nodeP);
        for ( jx = 0; jx < NumOutDirs; jx++ ) {
            if ( Pstreq(name,OutDirs[jx]) )
                goto got_it;
        }
        if ( Cstreq(nodeP->name,"amake") || Cstreq(nodeP->name,"acpp") || Cstreq(nodeP->name,"acproto") )
            goto got_it;
#if 0
		/* Some other stuff that can be at the top level */
        if ( Cstreq(nodeP->name,"README.txt") || Cstreq(nodeP->name,"Test-Shelf.bsh") )
            goto got_it;
#endif
        bld_assert(AltOutDirId,("%s not in valid outdir %s", nodeP->name,name));
        nodeP->dirid = AltOutDirId;
        VERBOSE(("re-directing %s to %s\n",
                  nodeP->name,DirPtr(AltOutDirId)->dname));
        nodeP->relocated = TRUE;
        nodeP->is_global = TRUE;
got_it:;
    }
}

/*************************************************************************
function: DoFindAllDeps - Find all dependencies
*/
void DoFindAllDeps ()
{
    ID id;
    TELL(("Performing full dependency search...\n"));
    for  ( id = NextType(0,TYPE_NODE); id; id = NextType(id,TYPE_NODE) )
        FindMyDeps(id);
}

/*************************************************************************
function: FillInNode - Fill output node with its info
*/
ID FillInNode (name,compid,rule_id,inputs_count,inputsP,depends_count,dependsP,
            withstr, user_deps)
char *name;
ID compid;
int rule_id;
int inputs_count;
char **inputsP;
int depends_count;
char **dependsP;
ID withstr;
int user_deps;
{
    register node *nodeP;
    int ix, size;
    ID id, *ip;
    extern ID *RuleDirs;

    if ( RuleDirs[rule_id] )
        nodeP = LookupForMe(name,compid);
    else
        nodeP = LookupOwnedNode(name,compid);
    assert(rule_id);
    assert(inputs_count+depends_count < 4550);
    assert(nodeP->callcount==0);
    id = NodeIdLooked;
    if ( nodeP->ruleid ) {
        if ( (nodeP->ruleid == rule_id) &&
             streq(StrPtr(withstr),NodeWithStr(nodeP)) )
            goto wrap_up;
        else {
            ErrorMsg("rules already exist for '%s'",nodeP->name);
            goto wrap_up;
        }
    }
    nodeP->ruleid = rule_id;
    size = sizeof(node) + (inputs_count+depends_count-1)*sizeof(ID);
    bld_assert(strlen(nodeP->name) < MAX_NODENAME,("name too long - %s",nodeP->name));
    ResizeObj(id,size);
    nodeP = NodePtr(id);
    nodeP->withstr = withstr;
    nodeP->incount = inputs_count;
    nodeP->depcount = depends_count;
    nodeP->comp_id = compid;
    nodeP->user_deps = user_deps;
    if ( RuleDirs[rule_id] && !nodeP->relocated ) {
        if ( nodeP->dirid = RuleDirs[nodeP->ruleid] ) {
            nodeP->relocated = TRUE;
            nodeP->is_global = TRUE;
            wrn_msg("WARNING: %s relies on .ruledir safety net!",nodeP->name);
        }
    }
    if ( !nodeP->dirid )
        nodeP->dirid = CompPtr(compid)->dflt;
    ip = nodeP->inputs;
    for ( ix = 0; ix < inputs_count; ix++ )
        *ip++ = LookupIdForMe(inputsP[ix],compid);
    for ( ix = 0; ix < depends_count; ix++ )
        *ip++ = LookupIdForMe(dependsP[ix],compid);

wrap_up:
    return(id);
}

int true_nop(){return(0);}


/*************************************************************************
function: AMakePath - Join directory and Entry names into a pathname
Result is a passed buffer ptr, this returns same ptr (like strcpy).
On Unix, this just cats with a "/" separator.
*/
char *AMakePath (result,dirname,entryname)
char *result;      /* resulting full pathname */
char *dirname;     /* orginal directory name */
char *entryname;   /* file entry name */
{
    if (dirname == NULL)
        dirname =".";
    sprintf(result,"%s%s%s",dirname,DirSlash,entryname);
    return(result);
}

/*************************************************************************
function: Arename - Rename a file
This returns TRUE on success, FALSE on failure.
*/
#ifdef Arename
#undef Arename
#endif
 

int Arename (src,dst)
register char *src;     /* source filename */
register char *dst;     /* destination filename */
{
    char acpy_cmd[300];

#if AXMS
    /* on NT, we need to make sure the destination file doesn't exist. */
    remove(dst);
#endif

    /** First we try the quick and simple approach **/
    if ( rename( src, dst ) == 0 )
       /* all is well */
       return( TRUE );
 
    if ( errno == EXDEV )
       {
       /* trying to move across devices, do a copy */
       sprintf(acpy_cmd,"cp %s %s 2>/dev/null", src,  dst );
       if ( system(acpy_cmd) )
           return( FALSE );
 
       /* delete the source */
       unlink( src );
       }
    else
        {
        printf("rename failed! (rename %s %s, errno=%d)\n", src, dst, errno);
        return FALSE;
        }
}


void Usage()
{
#if PORTING
printf("\n\ Summary of major command line options:    (processed by ReadCommandLine) \n\
\n\
    -prof                  Meta: builds all with profiling \n\
    -rel                   Meta: builds dbx with version information  \n\
    -orel                  Meta: builds optimized with version information  \n\
\n\
    -cache                 Rebuild cache \n\
    -cacheok               Treat the cache as if it is ok, do not rebuild \n\
    -icache                Add stuff in lpgms to cache \n\
    -limited               Use /ax/debug_list to set usage of -g \n\
    -f <names>             Force rebuild of files <names> \n\
    -touchonly             Don't build, just update timestamps \n\
\n\
    -grep s1 s2            Find occurrences string s1 in files whose pathname has s2 \n\
    -find <name>           Report pathname of module \n\
    -whence <name>         List inputs used to produce a given output file\n\
    -rwhence <name>        Recursively list full inputs tree\n\
\n\
    -commands              Emit command lines to stdout\n\
    -report                Produce 'lpr' report file rather than stdout \n\
    -dumponly <pathname>   Produce human meaningful dump of amake cache \n\
\n\
    -allprotos             Regenerate protos.dct fresh from all the .s files\n\
    -doproto               Produce new .s files even for dependency compiles\n\
    -noproto               Don't run prototyper on compilations\n\
\n\
NOTE: The 'meta' arg must be the first arg.  You can use only one 'meta' arg. \n\
If there are no meta args, build all dbx is assumed.\n\
\n\
");
#else
printf("\n Summary of major command line options:    (processed by ReadCommandLine) \n\
\n\
    -o                     Meta: builds all optimized   \n\
\n\
    -cache                 Rebuild cache    \n\
    -cacheok               Treat the cache as if it is ok, do not rebuild \n\
    -icache                Add stuff in lpgms to cache  \n\
    -limited               Use /ax/debug_list to set usage of -g \n\
    -f <names>             Force rebuild of files <names>   \n\
    -touchonly             Don't build, just update timestamps \n\
\n\
    -grep s1 s2            Find occurrences string s1 in files whose pathname has s2 \n");

printf(
"    -find <name>           Report pathname of module \n\
    -whence <name>         List inputs used to produce a given output file\n\
    -rwhence <name>        Recursively list full inputs tree\n\
    -dumponly <pathname>   Produce human meaningful dump of amake cache \n\
    -v                     Verbose build\n\
\n\
    -report                Produce 'lpr' report file rather than stdout \n\
    -allprotos             Regenerate protos.dct fresh from all the .s files\n\
    -doproto               Produce new .s files even for dependency compiles\n\
    -noproto               Don't run prototyper on compilations\n\
\n\
NOTE: The 'meta' arg must be the first arg.  You can use only one 'meta' arg. \n\
If there are no meta args, build all dbx is assumed.\n\
\n\
");
#endif
}




void DoReportStrictFiles(){ /* TEMP */
	int id;
	node *nodeP;
    int (*func)();      /* the function to call */

	printf("Strict Compiles:\n");
	for ( id = NextType(0,TYPE_NODE); id ;id = NextType(id,TYPE_NODE) ) {
		nodeP = NodePtr(id);
		if ( nodeP->strict_excluded || !OUTPUT_NODE(nodeP) )
			continue;
		func = RuleArray[nodeP->ruleid].func;
		if ((func == ansi_compile)
			|| (func == mkshared_cc)
			|| (func == mkshared_cc_mt)
			|| (func == noopt_compile)
			|| (func == dbg_compile)
			|| (func == dbg_compile_mt)
			|| (func == nodbg_compile) )
		    CMPTELL(("%s\n",nodeP->name));
	}
	printf("End Strict Compiles:\n");
	exit(0);
}


