/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: am_mem.c
   Applix ASC version: 9
   Date: Sun Jan 24 21:13:14 1999
*/

/* Boom-like memory manager for amake. */

#define MAX_OBJECTS 64000

#define AMAKE_REVNUM 8

#include <am_bld.h>

typedef struct {
    long revision;
    long tbl_size;
    long vars_id;       /* ID of #defines array */
    long resvd[10];
} file_header;

typedef union {
    char *ch;
    long num;
    node *ndp;
    dir_info *dir;
    comp_info *cmp;
    ID *id;
    ctrlfile_info *cfile;
} am_ptr;

#define FREE_TYPE 0

am_ptr ObjTable[MAX_OBJECTS];
char ObjTypes[MAX_OBJECTS];
unsigned short ObjSizes[MAX_OBJECTS];
long NextFree = 0;
ID NumUsed = 1;
ID VarsIdArray;   /* #defines */

#define TYPEMATCH(type,id) assert(ObjTypes[id]==type)
#define TYPECHECK(type) assert((type >= 1) && (type < TYPE_UNUSED));
#define IDCHECK(type,id) assert(id && (id < NumUsed)); TYPECHECK(type)

static ReadSome();

/*************************************************************************
function: NodePtr - Return node
*/
node *NodePtr (id)
register ID id;
{
    IDCHECK(TYPE_NODE,id);
    TYPEMATCH(TYPE_NODE,id);
    return(ObjTable[id].ndp);
}

/*************************************************************************
function: IdPtr - Return ID ptr
*/
ID *IdPtr (id)
register ID id;
{
    IDCHECK(TYPE_ID,id);
    TYPEMATCH(TYPE_ID,id);
    return(ObjTable[id].id);
}

/*************************************************************************
function: StrPtr - Retunr a string ptr
*/
char *StrPtr (id)
register ID id;
{
    IDCHECK(TYPE_STR,id);
    TYPEMATCH(TYPE_STR,id);
    return(ObjTable[id].ch);
}

/*************************************************************************
function: DirPtr - Return a directory objetc
*/
dir_info *DirPtr (id)
register ID id;
{
    IDCHECK(TYPE_DIR,id);
    return(ObjTable[id].dir);
}

/*************************************************************************
function: CompPtr - Return a component object ptr
*/
comp_info *CompPtr (id)
register ID id;
{
    IDCHECK(TYPE_COMP,id);
    TYPEMATCH(TYPE_COMP,id);
    return(ObjTable[id].cmp);
}

/*************************************************************************
function: CmlVarPtr - Return a command line varibale ptr
*/
char *CmlVarPtr (id)
register ID id;
{
    IDCHECK(TYPE_CMLVAR,id);
    TYPEMATCH(TYPE_CMLVAR,id);
    return(ObjTable[id].ch);
}

/*************************************************************************
function: CtrlFilePtr - Return name of a control file
*/
ctrlfile_info *CtrlFilePtr (id)
register ID id;
{
    IDCHECK(TYPE_CTRLFILE,id);
    TYPEMATCH(TYPE_CTRLFILE,id);
    return(ObjTable[id].cfile);
}

/*************************************************************************
function: AllocObj - Allocate an object
*/
ID AllocObj (size,type)
unsigned int size;
int type;
{
    ID id;

    TYPECHECK(type);
    LOGGER(("allocObj: size %d type %d",size,type));
    if ( NextFree ) {
        id = NextFree;
        NextFree = ObjTable[id].num;
    }
    else {
        id = NumUsed++;
        assert(id < MAX_OBJECTS);
    }
    memset(ObjTable[id].ch = malloc(size),0,size);
    ObjTypes[id] = type;
    ObjSizes[id] = size;
    return(id);
}

/*************************************************************************
function: ResizeObj - Chnage the size of an object
*/
ResizeObj (id,newsize)
ID id;
unsigned newsize;
{
    register int oldsize;
    register char *p;

    oldsize = ObjSizes[id];
    ObjTable[id].ch = p = realloc(ObjTable[id].ch,newsize);
    ObjSizes[id] = newsize;
    if ( oldsize < newsize )
        memset(p+oldsize,0,newsize-oldsize);
}

/*************************************************************************
function: SizeObj - return the current size of an object
*/
SizeObj (id)
ID id;
{
    return(ObjSizes[id]);
}

/*************************************************************************
function: AlcAStr - Allocate a 'string', setting type
*/
ID AlcAStr (p,type)
char *p;
int type;
{
    register ID id;

    id = AllocObj(strlen(p)+1,type);
    strcpy(ObjTable[id].ch,p);
    return(id);
}

/*************************************************************************
function: ObjStr - Allocate a string
*/
ID ObjStr (p)
char *p;
{
    register ID id;

    id = AllocObj(strlen(p)+1,TYPE_STR);
    strcpy(ObjTable[id].ch,p);
    return(id);
}
/*************************************************************************
function: NextType - Get next object of a given type
*/
ID NextType (id,type)
{
    if ( id != 0 ) {
        IDCHECK(type,id);
    }
    else {
        TYPECHECK(type);
    }
    for ( id++; id < NumUsed; id++ ) {
        if ( ObjTypes[id] == type )
            return(id);
    }
    return(0);
}

/*************************************************************************
function: ReadMemory - Read object file into memory
*/
ReadMemory (name)
char *name;
{
    FILE *fp;
    ID ix;
    file_header hdr;

#if AXMS
    fp = fopen(name,"rb");
#else
    fp = fopen(name,"r");
#endif
    if ( fp == NULL )
        return(FALSE);
    assert(ReadSome(fp,&hdr,(long)sizeof(file_header)));
    if ( hdr.revision != AMAKE_REVNUM ) {
        fclose(fp);
        return(FALSE);
    }
    TELL(("Reading %s\n",name));
    assert(ReadSome(fp,ObjTypes,(long)(NumUsed = hdr.tbl_size)));
    assert(ReadSome(fp,ObjSizes,(long)(hdr.tbl_size * sizeof(short))));
    VarsIdArray = hdr.vars_id;
    NextFree = 0;
    for ( ix = 0; ix < NumUsed; ix++ ) {
        if ( 0 == ObjTypes[ix] ) {
            ObjTable[ix].num = NextFree;
            NextFree = ix;
        }
        else {
            ObjTable[ix].ch = malloc(ObjSizes[ix]);
            assert(ReadSome(fp,ObjTable[ix].ch,(long)ObjSizes[ix]));
        }
    }
    fclose(fp);
    return(TRUE);
}

/*************************************************************************
function: FlushMemory - Toss memory space
*/
FlushMemory ()
{
    ID ix;

    NextFree = 0;
    for ( ix = 0; ix < NumUsed; ix++ ) {
        if ( 0 != ObjTypes[ix] )
            free(ObjTable[ix].ch);
        ObjTypes[ix] = 0;
    }
    NumUsed = 1;
}

/*************************************************************************
function: WriteMemory - Write memory to object file
*/
WriteMemory (name)
char *name;
{
    FILE *fp;
    ID ix;
    file_header hdr;
    char tempfile[STRPATH];

    GetTemp("AMCA",tempfile);
#if AXMS
    fp = fopen(tempfile,"wb");
#else
    fp = fopen(tempfile,"w");
#endif
    if ( fp == NULL )
        return(FALSE);
    memset(&hdr,0,sizeof(hdr));
    hdr.tbl_size = NumUsed;
    hdr.revision = AMAKE_REVNUM;
    hdr.vars_id = VarsIdArray;
    assert(WriteSome(fp,&hdr,(long)sizeof(file_header)));
    assert(WriteSome(fp,ObjTypes,(long)hdr.tbl_size));
    assert(WriteSome(fp,ObjSizes,(long)(hdr.tbl_size * sizeof(short))));
    TELL(("Writing %s\n",name));
    for ( ix = 0; ix < hdr.tbl_size; ix++ ) {
        if ( 0 != ObjTypes[ix] )
            assert(WriteSome(fp,ObjTable[ix].ch,(long)ObjSizes[ix]));
    }
    fclose(fp);
    unlink(name);
    if ( !Arename(tempfile,name) ) {
        ErrorMsg("could not create %s",name);
        unlink(tempfile);
        return(FALSE);
    }
    return(TRUE);
}

/*************************************************************************
function: ReadSome - A 'read' that can do more than 32K on AT
returns T/F
*/
static ReadSome (fp,p,cnt)
FILE *fp;
char *p;
long cnt;
{
    int c;
#if AT
    if ( cnt > 32767 )
        abort();
#endif
    c = cnt;
    return(1 == fread(p,c,1,fp));
}

/*************************************************************************
function: WriteSome - A 'write' that can do more than 32K on AT
returns T/F
*/
WriteSome (fp,p,cnt)
FILE *fp;
char *p;
long cnt;
{
    int c;
#if AT
    if ( cnt > 32767 )
        abort();
#endif
    c = cnt;
    return(1 == fwrite(p,c,1,fp));
}

