/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: cpr_system.h
   Applix ASC version: 4
   Date: Sun Jan 24 21:13:17 1999
*/
/* $Id: cpr_system.h,v 4.2 1994/10/13 17:26:04 cthuang Exp $
 *
 * cproto configuration and system dependencies
 */

#ifndef SYSTEM_H
#define SYSTEM_H
 
#if !defined(TRUE) || (TRUE != 1)
#undef  TRUE
#define TRUE    (1)
#endif

#if !defined(FALSE) || (FALSE != 0)
#undef  FALSE
#define FALSE   (0)
#endif

/* Borland C++ for MS-DOS predefines __MSDOS__ */
#ifdef __MSDOS__
#ifndef MSDOS
#define MSDOS
#endif
#endif

/* Borland C++ for OS/2 predefines __OS2__ */
#ifdef __OS2__
#ifndef OS2
#define OS2
#endif
#endif

#if FOR_AMAKE && HP800
#define CONST
#else
#define CONST const
#endif

#if FOR_AMAKE   /* disable protos, breaks on HPUX builds, -zane 15-jul-96 */
#define ARGS(p) ()
#else
/* don't use continuation-lines -- breaks on VAXC */
#if defined(__STDC__) || defined(__cplusplus) || defined(MSDOS) || defined(OS2) || defined(vms)
#define ARGS(p) p
#else
#define ARGS(p) ()
#endif
#endif

/* Turbo C preprocessor */
#ifdef __TURBOC__
#define YY_READ_BUF_SIZE 256    /* patch */
#define HAVE_TMPFILE 1
#define HAVE_GETOPT_H 1 /* use the one from porting-directory */
#include <io.h>     /* declares 'read()' for flex */
#endif

#ifdef TURBO_CPP
#define CPP "cpp -P-"
#else
#ifdef OS2
#define HAVE_POPEN_PROTOTYPE 1
#define popen _popen
#define pclose _pclose
#endif
#endif

/* Microsoft C preprocessor */
#ifdef M_I86
#define CPP "cl /E /nologo"
#define HAVE_TMPFILE 1
#endif

/* Vax C */
#ifdef vms
#define CPP "cc /preprocess_only=%s"
#define unlink remove
#define HAVE_TMPFILE 1
#define HAVE_GETOPT_H 1
#define USE_flex 1
#endif

/* Set configuration parameters for systems on which we cannot run autoconf.
 * (Assumes Posix unless overridden).
 */
#ifndef HAVE_STDLIB_H
#define HAVE_STDLIB_H 1
#endif

#ifndef HAVE_STRING_H
#define HAVE_STRING_H 1
#endif

#ifndef HAVE_STRSTR
#define HAVE_STRSTR 1
#endif

#ifndef HAVE_GETOPT_H
#define HAVE_GETOPT_H 0
#endif 

/* Default C preprocessor on UNIX systems */
#ifndef MSDOS
#ifndef CPP
#define CPP "/lib/cpp"
#endif
#endif

/* maximum include file nesting */
#ifndef MAX_INC_DEPTH
#define MAX_INC_DEPTH 15
#endif

/* maximum number of include directories */
#ifndef MAX_INC_DIR
#define MAX_INC_DIR 25
#endif

/* maximum text buffer size */
#ifndef MAX_TEXT_SIZE
#define MAX_TEXT_SIZE 256
#endif

#if HAVE_STDLIB_H
#include <stdlib.h>
#else
extern char *malloc  ARGS((size_t n));
extern char *realloc ARGS((char *p, size_t n));
#endif

/* Declare argument for exit() function */
#ifdef vms
#include <stsdef.h>
#define EXIT_SUCCESS    (STS$M_INHIB_MSG | STS$K_SUCCESS)
#define EXIT_FAILURE    (STS$M_INHIB_MSG | STS$K_ERROR)
#else
#ifndef EXIT_SUCCESS
#define EXIT_SUCCESS    (0)
#endif
#ifndef EXIT_FAILURE
#define EXIT_FAILURE    (1)
#endif
#endif /* vms */

#if HAVE_UNISTD_H
#if !MSDOS
#include <unistd.h>
#endif
#endif

#if STDC_HEADERS || HAVE_STRING_H
#  include <string.h>
/* An ANSI string.h and pre-ANSI memory.h might conflict.  */
#  if !STDC_HEADERS && HAVE_MEMORY_H
#    include <memory.h>
#  endif /* not STDC_HEADERS and HAVE_MEMORY_H */
#else /* not STDC_HEADERS and not HAVE_STRING_H */
#  include <strings.h>
#  define strchr index
#  define strrchr rindex
/* memory.h and strings.h conflict on some systems.  */
#endif /* not STDC_HEADERS and not HAVE_STRING_H */

extern char *getenv  ARGS((const char *v));
extern char *strstr  ARGS((const char *s, const char *p));

/*
 * The DOALLOC symbol controls whether we compile in the simple memory tests
 * in 'trace.c' (use dbmalloc for more rigorous testing).
 */
#ifdef DOALLOC
extern char *doalloc(char *,unsigned);
extern void dofree(char *);

#define malloc(n)    doalloc(0,n)
#define realloc(p,n) doalloc(p,n)
#define free(n) dofree(n)
#endif /* DOALLOC */

/*
 * Lint libraries are useful only on systems that are likely to have lint.
 * The OPT_LINTLIBRARY symbol controls whether we compile in the lint library
 * support.
 */
#ifndef OPT_LINTLIBRARY
# ifdef unix
#  define OPT_LINTLIBRARY 1
# endif
#endif

#if BISON_HAS_YYTNAME || YACC_HAS_YYTOKS || YACC_HAS_YYTOKS_2 || YACC_HAS_YYNAME
# define YYDEBUG 1
#endif

#if HAVE_LIBDBMALLOC
#include <dbmalloc.h>
#endif

#endif /* SYSTEM_H */
