/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: cproto.h
   Applix ASC version: 9
   Date: Mon Feb  1 21:13:55 1999
*/
/* $Id: cproto.h,v 4.4 1995/08/24 01:03:48 cthuang Exp $
 *
 * Declarations for C function prototype generator
 */

/* The following config group was setup for SUN4/SUN4S.. may need to change */
#define STDC_HEADERS    1
#define HAVE_STDLIB_H   1
#define HAVE_UNISTD_H   1
#define HAVE_STRING_H   1
#define HAVE_MALLOC_H   1
#define HAVE_STRSTR 1
#define HAVE_TMPFILE    1
#define HAVE_LINK   1
#define HAVE_UNLINK 1
#define HAVE_MKTEMP 1
#define YACC_HAS_YYTOKS 1

#if AXMS
#define CPP "cl -E -nologo"
#elif (SUN4 && !SUN4S)
#define CPP "acc -E"
#elif HP800V3
/* The following arguments are from "cc -E -Aa -v foo.c" (piyush).
 * The -H has been added for asc2wp to fix "too many defines" error (zane). */
#define CPP "/lib/cpp.ansi -H500000 -D__hp9000s700 -D__hp9000s800 -D__hppa -D__hpux -D__unix -D_PA_RISC1_1 -A -I /usr/include"
#elif RIOS
#define CPP "xlc -E"
#else
#define CPP "cc -E"
#endif

#include "cpr_system.h"

#if HAVE_LIBDBMALLOC || defined(DOALLOC)
#undef  NO_LEAKS
#define NO_LEAKS 1
#endif

#ifdef  lint
#define ALLOC(cast) (cast *)0
#else
#define ALLOC(cast) (cast *)xmalloc(sizeof(cast))
#endif

/* Useful constants (mainly to avoid problems balancing parentheses...) */
#define PAREN_L       '('
#define PAREN_R       ')'
#define SQUARE_L      '['
#define SQUARE_R      ']'
#define CURL_L        '{'
#define CURL_R        '}'
#define COMMENT_BEGIN "/*"
#define COMMENT_END   "*/"

/* Boolean type */
typedef char boolean;

/* Source file text */
typedef struct text {
    char text[MAX_TEXT_SIZE];   /* source text */
    long begin;         /* offset in temporary file */
} Text;

/* This is a list of function parameters. */
typedef struct parameter_list {
    struct parameter *first;    /* pointer to first parameter in list */
    struct parameter *last; /* pointer to last parameter in list */  
    long begin_comment;     /* begin offset of comment */
    long end_comment;       /* end offset of comment */
    char *comment;      /* comment at start of parameter list */
} ParameterList;

/* Declaration specifier flags */
#define DS_NONE     0   /* default */
#define DS_EXTERN   1   /* contains "extern" specifier */
#define DS_STATIC   2   /* contains "static" specifier */
#define DS_CHAR     4   /* contains "char" type specifier */
#define DS_SHORT    8   /* contains "short" type specifier */
#define DS_FLOAT    16  /* contains "float" type specifier */
#define DS_JUNK     32  /* we're not interested in this declaration */

/* This structure stores information about a declaration specifier. */
typedef struct decl_spec {
    unsigned short flags;   /* flags defined above */
    char *text;         /* source text */
    long begin;         /* offset in temporary file */
} DeclSpec;

/* Styles of function definitions */
#define FUNC_NONE       0   /* not a function definition */
#define FUNC_TRADITIONAL    1   /* traditional style */
#define FUNC_ANSI       2   /* ANSI style */
#define FUNC_BOTH       3   /* both styles */
typedef int FuncDefStyle;

/* This structure stores information about a declarator. */
typedef struct declarator {
    char *name;             /* name of variable or function */
    char *text;             /* source text */
    long begin;             /* offset in temporary file */
    long begin_comment;         /* begin offset of comment */
    long end_comment;           /* end offset of comment */
    FuncDefStyle func_def;      /* style of function definition */
    ParameterList params;       /* function parameters */
    boolean pointer;            /* TRUE if it declares a pointer */
    struct declarator *head;        /* head function declarator */
    struct declarator *func_stack;  /* stack of function declarators */
    struct declarator *next;        /* next declarator in list */
} Declarator;

/* This is a list of declarators. */
typedef struct declarator_list {
    Declarator *first;      /* pointer to first declarator in list */
    Declarator *last;       /* pointer to last declarator in list */  
} DeclaratorList;

/* This structure stores information about a function parameter. */
typedef struct parameter {
    struct parameter *next; /* next parameter in list */
    DeclSpec decl_spec;
    Declarator *declarator;
    char *comment;      /* comment following the parameter */
} Parameter;

/* parser stack entry type */
typedef union {
    Text text;
    DeclSpec decl_spec;
    Parameter *parameter;
    ParameterList param_list;
    Declarator *declarator;
    DeclaratorList decl_list;
} YYSTYPE;

/* Prototype styles */
#if OPT_LINTLIBRARY
#define PROTO_LINTLIBRARY   -1  /* form lint-library source */
#endif
#define PROTO_NONE      0   /* do not output any prototypes */
#define PROTO_TRADITIONAL   1   /* comment out parameters */
#define PROTO_ABSTRACT      2   /* comment out parameter names */
#define PROTO_ANSI      3   /* ANSI C prototype */
typedef int PrototypeStyle;

#define lintLibrary() (proto_style == PROTO_LINTLIBRARY)

/* The role of a function declarator */
#define FUNC_OTHER  0   /* miscellaneous declaration */
#define FUNC_PROTO  1   /* prototype */
#define FUNC_DEF    2   /* function definition */
typedef int FuncDeclRole;

/* Prototype/function definition output formats */
#define FMT_OTHER       0   /* miscellaneous */
#define FMT_PROTO       1   /* prototype */
#define FMT_FUNC        2   /* function definition */
#define FMT_FUNC_COMMENT    3   /* func. def. with parameter comments */
typedef int FuncFormatType;

/* select scope of declarations to output */
#define SCOPE_STATIC    1   /* only output declarations with local scope */
#define SCOPE_EXTERN    2   /* only output declarations with global scope */
#define SCOPE_ALL   3   /* output all declarations */
typedef int Scope;

/* Prototype/function definition output format */
typedef struct func_format {
    char *decl_spec_prefix; /* output before declaration specifier */
    char *declarator_prefix;    /* output before declarator name */
    char *declarator_suffix;    /* output before '(' of parameter list */
    char *first_param_prefix;   /* output before first parameter */
    char *middle_param_prefix;  /* output before each subsequent parameter */
    char *last_param_suffix;    /* output after last parameter */
} FuncFormat;

/* Program options */
extern boolean extern_out;
extern Scope scope_out;
#if OPT_LINTLIBRARY
extern boolean types_out;
extern boolean lint_shadowed;
#endif
extern boolean variables_out;
extern boolean promote_param;
extern PrototypeStyle proto_style;
extern FuncDefStyle func_style;
extern boolean proto_macro;
extern boolean define_macro;
extern char *macro_name;
extern boolean proto_comments;
extern boolean file_comments;
extern boolean quiet;
extern char *func_directive;
extern int num_inc_dir;
extern char *inc_dir[];
extern FuncFormat fmt[4];

/* Global declarations */
extern char *progname;
extern int varargs_num;     /* supports varargs-comment */
extern char *varargs_str;   /* additional info, such as PRINTFLIKEnn */
extern int extern_in;       /* supports "LINT_EXTERNnn" */
extern int exitlike_func;   /* supports noreturn-attribute */
extern int in_include;      /* current include-level */
extern int debug_trace;
extern char base_file[];

/* cproto.c */
#ifdef NO_LEAKS
extern char *xMalloc        ARGS((unsigned n, char *f, int l));
extern char *xStrdup        ARGS((char *s,    char *f, int l));
#define xmalloc(n)          xMalloc(n, __FILE__, __LINE__)
#define xstrdup(s)          xStrdup(s, __FILE__, __LINE__)
#else
extern char *xmalloc        ARGS((unsigned n));
extern char *xstrdup        ARGS((char *src));
#endif
#if MSDOS
extern int put_error       ARGS((void));
#else
extern void put_error       ARGS((void));
#endif
extern int is_path_sep      ARGS((int ch));
extern char *trim_path_sep  ARGS((char *s));

/* cpr_lintlibs.c */
#if OPT_LINTLIBRARY
extern void put_string      ARGS((FILE *outf, char *s));
extern void put_char        ARGS((FILE *outf, int c));
extern void put_newline     ARGS((FILE *outf));
extern void put_blankline   ARGS((FILE *outf));
extern void put_padded      ARGS((FILE *outf, char *s));
extern void fmt_library     ARGS((int code));
extern void begin_tracking  ARGS((void));
extern int already_declared ARGS((char *name));
extern void track_in        ARGS((void));
extern int want_typedef     ARGS((void));
extern void begin_typedef   ARGS((void));
extern void copy_typedef    ARGS((char *s));
extern void end_typedef     ARGS((void));
extern void imply_typedef   ARGS((char *s));
extern char *implied_typedef ARGS((void));
extern void indent          ARGS((FILE *outf));
extern int lint_ellipsis    ARGS((Parameter *p));
extern void ellipsis_varargs ARGS((Declarator *d));
extern char *supply_parm    ARGS((int count));
extern int is_actual_func   ARGS((Declarator *d));
extern void put_body        ARGS((FILE *outf, DeclSpec *decl_spec, Declarator *declarator));
# ifdef NO_LEAKS
extern void free_lintlibs   ARGS((void));
# endif
#else
#define put_string(fp,S)    fputs(S, fp)
#define put_char(fp,C)      fputc(C, fp)
#define put_padded(fp,S)    fprintf(fp, "%s ", S)
#define put_body(fp,s,d)    put_string(fp,";\n")
#define track_in()
#define begin_typedef()
#define copy_typedef()
#define end_typedef()
#define imply_typedef(s)
#define implied_typedef()   ((char *)0)
#endif

#if OPT_LINTLIBRARY
extern void flush_varargs   ARGS((void));
#else
#define flush_varargs() /* nothing */
#endif

/* cpr_strkey.c */
extern char *strkey         ARGS((char *src, char *key));
extern void strcut          ARGS((char *src, char *key));

/* cpr_grammar.y */
extern boolean is_typedef_name ARGS((char *name));
extern char *cur_file_name  ARGS((void));
extern unsigned cur_line_num ARGS((void));
extern FILE *cur_tmp_file   ARGS((void));
extern void cur_file_changed ARGS((void));
extern long cur_begin_comment ARGS((void));
extern char *cur_text       ARGS((void));
extern void pop_file        ARGS((void));
extern void init_parser     ARGS((void));
extern void process_file    ARGS((FILE *infile, char *name));
#ifdef NO_LEAKS
extern void free_parser     ARGS((void));
#endif
