/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: bldrextbase.hh
   Applix ASC version: 5
   Date: Mon Nov 16 14:27:53 1998
*/
#ifndef _BLDREXTBASE_HH_
#define _BLDREXTBASE_HH_

#include <stdlib.h>
#include <elfapi.hh>
#include <bldrextinfobase.hh>

/** 
** Root Builder to External Object Class 

** All methods declared in this file are pure-virtual. The method described
** here are the minimum set that Builder requires to be able to support an
** external object model. Since all these methods are pure-virtual they have
** to be over-ridden by the class that implements them even if they are not
** required or supported by the object model. 

***************** The Methods ****************************
** ~BuilderExternalObjectRoot() - every implementation must have a destructor 
** createObject         - create a external object
** invokeMethod         - invoke a method on the external object
** listOfObjects        - Return a list of objects available to the user.
** infoObject           - Set/Get an information object that describes the actual object
** newInfoObject        - Create a new information object
** objectPtr            - return a pointer to the actual object.
** registerForCallbacks - Make sure the external object model can call into builder
** app_terminated       - If the object model needs to clean up when the app terminates.
****************************************************/

class BuilderExternalObjectRoot {
// --------------------------------------------------------------------------
//                  P U R E    V I R T U A L     M E T H O D S
// --------------------------------------------------------------------------
	public:
		virtual ~BuilderExternalObjectRoot() {}; 
		virtual void *createObject(
			elfData objectInfo,
			int *status,
			elfData addInfo
				) = 0;
		virtual elfData invokeMethod(
			char *mName,
			elfData args,
			int *status,
			elfData addInfo = 0
				) = 0;
		virtual elfData listOfObjects(
			elfData arg1,
			elfData arg2,
			elfData addInfo=0
				) = 0;
		virtual bldrextinfobase *infoObject(void) = 0;
		virtual void infoObject(bldrextinfobase *info) = 0;
		virtual bldrextinfobase *newInfoObject(void) = 0;
		virtual void* objectPtr (void) = 0;
		virtual void registerForCallbacks(int, int, void *) = 0;
		virtual void app_terminated(void) = 0;

};

typedef BuilderExternalObjectRoot bldrextobjectbase;


#endif 	// _BLDREXTBASE_HH_
