/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: ctobldr.c
   Applix ASC version: 4
   Date: Wed Feb 24 21:14:51 1999
*/
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <axsocket.h>
#include <elfapi.hh>
#include <axlist.hh>

/** Functions in this file */
elfData AxCallBuilderObjectMethod(int, char *, elfData, int *, char **);
int AxInitializeBuilderObject(char *, char *, char **);
int AxTerminateBuilderObject(int);


#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif

#define NXLT(str) str

#define BLDR_CORBA_SERVICE_ID_ NXLT("bldr_corba_service")
#define BLDR_CORBA_SERVER_MACRO_ NXLT("service_corba_request$")

/** Macro that handles all corba requests */
static char *BldrCorbaSrvcMacro = BLDR_CORBA_SERVER_MACRO_;

/** Service Name registered with axnet */
static char *BldrServiceName = BLDR_CORBA_SERVICE_ID_;

/** Function that compares the elements of BldrObjectsList */
static int CompareObjectListTypes(void *, void *);

/** List of Bldr Objects */
static axlist BldrObjectsList(CompareObjectListTypes);

/** An element of the list of Builder Objects */
typedef struct ObjectListElem {
	int id;			/** Unique identifier */
	int channel;	/** Open channel to axnet */
	char *host;		/** The host on which */
	char *bldrName; /** see buildlib.els register_builder_object@()*/
} ObjectListElem;

static int
CompareObjectListTypes(void *elem1, void *elem2)
{
	ObjectListElem *e1 = (ObjectListElem *)elem1;
	ObjectListElem *e2 = (ObjectListElem *)elem2;
	
	if (!e1->host || !e2->host) {
		/** Find by id */
		if (e1->id == e2->id)
			/** same object */
			return(0);
		else
			/** different objects */
			return(1);
	}
	else {
		/** find by host and bldrName */
		// if (e1->id == e2->id && 
		if (0 == strcmp(e1->host, e2->host) &&
			0 == strcmp(e1->bldrName, e2->bldrName) )
			/** Same objects */
			return(0);
		else
			/** different objects */
			return(1);
	}
	
	return(1);
}

int
AxInitializeBuilderObject(char *bldrObjectName, char *host, char **str)
{
	static int stat_id = 0;
	int code;
	char fullServiceName[300];
	char *obj;
	ObjectListElem *elem = new ObjectListElem;
	ObjectListElem *elem2 = 0;
	int channel = -1;
	
	/** first see if this object is already in list */
	elem->host = new char[strlen(host)+1];
	strcpy(elem->host, host);
	elem->bldrName = new char[strlen(bldrObjectName)+1];
	strcpy(elem->bldrName, bldrObjectName);
	elem2 = (ObjectListElem *)BldrObjectsList.find(elem);
	if(elem2) {
		/** Object in list - clean up and return the org object id */
		delete [] elem->bldrName;
		delete [] elem->host;
		delete elem;
		return(elem2->id);
/*
		else { 
			elem->id = stat_id++;
			elem->channel = elem2->channel;
			BldrObjectsList.add(elem);
		}
*/
	}


	/** New object - need to create a new connection */
	sprintf(fullServiceName, "%s:%s", getenv("USER"), BldrServiceName);
	channel = AxNetElfServiceConnect(host, fullServiceName, &code, str, &obj);
	if ( code != 0 ) {
		/** Connection not established - clean up and return */
		delete [] elem->bldrName;
		delete [] elem->host;
		delete elem;
		return(-1);
	}

	/** Get a new id, add this to the BldrObjectsList and return the id */
	elem->id = stat_id++;
	elem->channel = channel;
	BldrObjectsList.add(elem);
	return(elem->id);
}
	
/** Note: closeChannel and destroyObject are not used now - these could
 ** be removed for now. The use would be to close the Builder application 
 ** on the Server Side but there has to be a way to make sure that no
 ** other objects in that application are being used as servers or that no
 ** other client is still talking to this builder object
 ** int AxTerminateBuilderObject(int id, int closeChannel, int destroyObject)
**/
int 
AxTerminateBuilderObject(int id)
{
	ObjectListElem tmp;
	tmp.id = id;
	tmp.host = 0;

	/** Find BldrObject based on id */
	ObjectListElem *elem = (ObjectListElem *)BldrObjectsList.find(&tmp);

	if (!elem)
		return(-1);

	/** Let the object know we are going away */
	int error = 0;
	char *str;
	AxCallBuilderObjectMethod(id, "terminate_object", 0, &error, &str);

	int objChannel = 0;
	objChannel = elem->channel;

	/** Free up memory */
	BldrObjectsList.unlink(elem);
	delete [] elem->bldrName;
	delete [] elem->host;
	delete elem;

	/** Close connection */
	AxClientDisConnect(objChannel);
	return(0);
}

elfData
AxCallBuilderObjectMethod(int id, char *methodName, elfData argv, int *error, char **str)
{
	ObjectListElem tmp;
	tmp.id = id;
	tmp.host = 0;
	/** Find object by id */
	ObjectListElem *elem = (ObjectListElem *)BldrObjectsList.find(&tmp);
	
	if (!elem) {
		/** No elem - invalid id */
		*error = -1;
		*str = "Invalid Object Id";
		return(0);
	}

	/** Package into elfdata - 
		[0] = methodName,
		[1] = BuilderObjectName,
		[3] = Array of parameters.
	**/
	char *obj;
	elfData response, Args;
	Args = AxMakeArray(0);
	Args = AxAddStrToArray(Args,0,(mbchar *)methodName);
	Args = AxAddStrToArray(Args,1,(mbchar *)elem->bldrName);
	Args = AxAddArrayToArray(Args,2,argv);

	/** make call - this is handled by the 
     ** BldrCorbaSrvcMacro (service_corba_request@ see buildlib.els ) 
	**/
	response = AxNetElfChannelCall(elem->channel, BldrCorbaSrvcMacro, Args, error, str, &obj);

	/** Free memory, check for error and return */
	AxFreeData(Args);
	if ( *error != 0 )
		return(0);
	return(response);
}

