/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axnet.c
   Applix ASC version: 125
   Date: Mon May 24 21:14:25 1999
*/
#include <revstamp.h>
#include <stdio.h>
#include <stdlib.h>
#include <atypes.h>
#include <sys/stat.h>
#include <axtime.h>

#if LINUX
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#endif

#if RIOS
#include <sys/select.h>
#endif

#include <stdarg.h>

#if !AXMS
#include <sys/wait.h>
#else
#include <winmain.h>
#include <winmap.h>
#define S_ISREG( mode_bits ) (mode_bits & _S_IFREG)
#define S_ISDIR( mode_bits ) (mode_bits & _S_IFDIR)
#endif

#include <axsocket.h>
#include <errno.h>
#include <signal.h>
#include <setjmp.h>
#include <apwd.h>
#include <elfnet.h>
#include <elfapi.h>
#include <axnet.h>
#define THIM_STATIC 1
#include <aos.h>


/* stamp executable with version and build machine */
static char *my_program_version = ASC_WHAT_STR;
static char *my_build_machine = ASC_WHAT_MACHINE2;


#define NXLT(str) str   /*  Keep xlt scans sane by NXLT'g where necess */
#define XLT(str) str

static FILE *netlogfp = NULL;
static int AxNetLogging = FALSE;
#define NETLOG(m) AxNetLog m

typedef struct service_desc {
    struct service_desc *next;
    char *name;         /* logical name */
    char *pathname;     /* to executable */
    char *socket;       /* to local machine conection file */
    int axnetStarts;    /* TRUE if started by axnet rather than client */
    int isRunning;      /* only relavent if axnetStarts */
    int isPrivate;      /* whether it shows up on a service list */
} service_desc;

static service_desc *Services;
static char ExecutablesDir[STRPATH+1];  /* where I am */
static int AxNetAllowProc = TRUE;

extern char AxMyProgramName[];

#if AXMS
#define STRIEQ( a, b )  (stricmp(a,b) == 0)

char *AxnetServiceName = "axnet";

extern void service_main( DWORD argc, LPSTR *argv );

/* thread argument struct for axnetEstablishRemoteConnection on NT. */
typedef struct remote_conn_t {
    int clientChannel;
    int socknum;
    char *host;
    char *socket;
    char *service_name;
} remote_conn_t;
#endif

extern int AxSvrSockFd;
extern int AxSvrAltSockFd;
extern int AxnetNoResponse; /* Just used for fork/execing new tasks by axnet */
extern int AxPreventLowCodesFromClient; /* keep client from sending <1 funcId */

/********** information from axnet configuration file */
struct axnetConfigLine {
    int command;
#define SETENV_FOR 1
#define ARG_FOR 2
#define LIB_FOR 3
#define PATH_FOR 4
#define UID_FOR 5
    char *arg1; /* first arg */
    char *arg2; /* second arg */
};
typedef struct axnetConfig {
    time_t last_modified; /* time last modified, when file was last read */
    int linecount;
    struct axnetConfigLine line[1];
} axnetConfigInfo;
axnetConfigInfo *axnetConfig;

char axnetConfigFile[STRPATH+1];

/* the UID_FOR feature in the axnet configuration file introduces the
spectre of axnet starting processes with any uid. Axnet can be used
to signal these processes. For security we normally do the signaling
as the requesting user. However in this case the requestor has started
a process under another user, but will still want to signal that
process. We do not want to allow axnet to signal just any process.
It can signal processes owned by the requestor, or also processes
in the following list of spawned processes where the requestor
uid is not the same as the process uid. These processes can be
signalled by the requestor. If there happen to be too many
processes for the following list, the worst that can happen
is that the ones that don't make it to the list cannot
be signalled. */
#define MAX_SPAWN 64
typedef struct {
    pid_t pid; /* spawned process */
    uid_t requestor; /* uid of requestor (part of request message) */
    uid_t actual; /* actual uid for process (as specified in axnet.cnf file) */
} spawnItem;
static spawnItem spawn[MAX_SPAWN];

#include <axnet.p>

main (argc,argv)
int argc;
char *argv[];
{
#if AX_X11                            /*  {  */
    elfmain(argc, argv);
#else                                 /*  }  {  */
    /* on NT we need to put a bunch of special stuff here on startup */
    /* to get axnet working as a service                             */
    int i;
    OSVERSIONINFO osversion;
    int Running_Windows95 = FALSE;
    int ret;
    int start_service = TRUE;
    char *account_name, *clear_password;

    osversion.dwOSVersionInfoSize = sizeof( OSVERSIONINFO );
    ret = MS_GetVersionEx( &osversion );
    if ( ret )
       {
       if ( osversion.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS )
           {
           Running_Windows95 = TRUE;
           start_service = FALSE;
           }
       }

    account_name = clear_password = NULL;
    for ( i=1; i<argc; i++ )
       {
       if ( STRIEQ( "-account", argv[i] ) )
          {
          account_name = argv[i+1];
          i++;
          }
       else if ( STRIEQ( "-password", argv[i] ) )
          {
          clear_password = argv[i+1];
          i++;
          }
       else if ( STRIEQ( "-install", argv[i] ) )
          {
          AddNtService(AxnetServiceName, account_name, clear_password);
          exit(0);
          }
       else if ( STRIEQ("-start", argv[i]) )
          {
          axmsInstallAndStartNtServiceAccount(AxnetServiceName, FALSE,
                                              account_name, clear_password );
          exit(0);  
          }
       else if ( STRIEQ( "-quiet", argv[i] ) )
          {
          /* Force an install, start the service, don't notify user. */
          axmsInstallAndStartNtServiceAccount(AxnetServiceName, TRUE,
                                              account_name, clear_password );
          exit(0);
          }
       else if ( STRIEQ( "-remove", argv[i] ) )
          {
          RemoveProgramAsNtService(AxnetServiceName, FALSE);
          exit(0);
          }
       else if ( STRIEQ("-entsvr", argv[i]) )
           start_service = FALSE;
       else if ( STRIEQ("-x", argv[i]) )
           start_service = FALSE;
       }


    if ( start_service )
        { 
        /* no arguments specified, assume that we are being started by */
        /* the system service manager. Here we pass control, via the   */
        /* the creaetion of a new thread, to the service manager libraries */
        AxInitService(AxnetServiceName);
        /* we don't call elfmain in this case because the service will
         * start a separate thread for it (see AxStartServiceThread).
         */
        return;
        }
    else if ( argv[1] && strcmp("-x", argv[1]) == 0 )
        {
        /* a dummy switch to indicate that we are auto-starting from axmain */
        /* for a particular user. Clear flags for lower layers              */
        argc = 1;
        }

    /* now go do the real axnet work */
    elfmain(argc, argv);    /* never returns */
 
    exit( 0 );
#endif      /* } AXMS  */
}


/*************************************************************************
function: axnetChildTerminated - Get here ONLY from SIGCHLD signal
This is here just to keep the children from going into the <defunct> state.

NOTE: The call to wait3(2) is inside a loop to ensure that all dead children
go away (i.e. don't sit in <defunct> state).

A single wait3(2) does not suffice in the scenerio where a bunch of 'axnet
launched' process get killed at the same time -- 'axnetChildTerminated ()'
will miss some!!  Possible explanation for why we see tons of zombie
processes owned by axnet on 'sparcy2'.  This fix cures the problem on
'sparcy2', where several axnet-launched 'elfinfmx' servers run...
*/
void axnetChildTerminated ()
{
#if !AXMS                 /*  { */
    int status = 0;
    int pid;

    while (TRUE) {

#if SUN4S || SUNIS || SGI || UNIXWARE || NCR
        pid = waitpid((pid_t)-1, &status, WNOHANG);
#else
/*      pid = wait3(&status, WNOHANG, NULL); */
        /* Use waitpid for all platforms */
        pid = waitpid((pid_t)-1, &status, WNOHANG);
#endif

        signal(SIGCHLD,
               (void (*)())axnetChildTerminated); /* re-install handler */
        if (pid <= 0)
            break; /* no more processes to wait for */
        else {
            NETLOG(("child, pid %d terminated\n", pid));
            removeSpawn(pid); /* remove from our list */
        }
    }
    signal(SIGCHLD, (void (*)())axnetChildTerminated); /* re-install handler */
#endif                        /* }  */
}

/*************************************************************************
function: addSpawn - add to list of spawned processed where funny UID business
has been done
*/
void addSpawn(pid_t pid, uid_t requestorUID, uid_t actualUID)
{
    register int ix;
    register spawnItem *spawnP;
    pid_t cpid;

    /* add new spawn */
    spawnP = getSpawnItem(0);
    if (spawnP) {
        spawnP->pid = pid;
        spawnP->requestor = requestorUID;
        spawnP->actual = actualUID;
    }
}

/*************************************************************************
function: getSpawnItem - get item from list of spawned processes
*/
spawnItem *getSpawnItem(pid)
pid_t pid;
{
    register int ix;
    for(ix=0; ix<MAX_SPAWN; ix++) {
        if (spawn[ix].pid == pid) {
            return(&(spawn[ix]));
        }
    }
    return((spawnItem *)0);
}

/*************************************************************************
function: removeSpawn - remove process from list 
*/
void removeSpawn(pid)
pid_t pid;
{
    register spawnItem *spawnP;
    spawnP = getSpawnItem(pid);
    if (spawnP) {
        spawnP->pid = 0;
        spawnP->requestor = 0;
        spawnP->actual = 0;
    }
}

/***********************************************************************
function: clientPipeError - write on client fd failed with EPIPE.
shut it down and continue.
*/
void
clientPipeError(fd)
int fd;
{
    NETLOG(("axnet: write fd %d to client failed with EPIPE. continuing...\n",
            fd));
    return;
}

/***********************************************************************
function: axnetInterrupted - Handle a SIGINT.  Simply log it to logfile.
*/
void
axnetInterrupted()
{
    NETLOG(("axnet: received INTERRUPT\n"));
    signal(SIGINT, (void (*)())axnetInterrupted); /* re-install handler */
}

/***********************************************************************
function: axnetToggleLogging - Handle a SIGHUP.  Toggle value on
`AxNetLogging' to allow/disallow logging.
This allows logging to be turned on/off while axnet is running.
*/
void
axnetToggleLogging()
{
#if CLIX300
void axnetToggleLogging();
#endif
#if !AXMS
    if (AxNetLogging) {
        NETLOG(("SIGHUP -- stop logging\n"));
        AxNetLogging = FALSE;
    } else {
        AxNetLogging = TRUE;
        NETLOG(("SIGHUP -- start logging\n"));
    }
    signal(SIGHUP, (void (*)())axnetToggleLogging); /* re-install handler */
#endif
}

/***********************************************************************
function: axnetGracefulTerminate - Do graceful program termination.
*/
void
axnetGracefulTerminate()
{
    char *pipe;

    NETLOG(("SIGTERM -- graceful termination\n"));
    pipe = getenv("AXNETPIPE");
    if ( !pipe || !pipe[0] )
        pipe = NXLT("/tmp/.axnetipc");

    /* FIXME, our RPC toolkit must have a user callable function that
       shutdown() and close()s sockets, unlinks files... */
    (void) unlink(pipe);
    exit(0);
}

/*************************************************************************
function: AxDispatch - Call customer specific C code
*/
elfData AxDispatch (funcId, funcData, quitNow)
int funcId;
elfData funcData;
int32 *quitNow;
{
    void axnetToggleLogging();
    void axnetChildTerminated();
    void axnetInterrupted();
    elfData rData = NULL;       /* return data, presumed no-return */
    struct servent *serverInfo;
    char *pipe;
    int port_number;
    int ix;
#if AX_X11
    char *p;
#endif

    switch ( funcId ) {

    case -3:    /* get port from /etc/services */
        serverInfo = getservbyname(NXLT("axnet"),NXLT("tcp"));
        if ( !serverInfo ) {
#if AXMS
            ix = WSAGetLastError();
            MS_MessageBox(NULL, "Unable to find 'axnet' in 'services'",
                                               "AXNET Error",MB_ICONHAND+MB_OK);
#else
            fprintf(stderr,XLT("Error, 'axnet' not a 'tcp' service in the '/etc/services' file.\n"));
#endif
            exit(1);
        }
        else {
            rData = AxMakeArray(2);
            port_number = ntohs( serverInfo->s_port );
            AxAddIntToArray(rData,0,port_number);
            pipe = getenv("AXNETPIPE");
            if ( !pipe || !pipe[0] )
                pipe = NXLT("/tmp/.axnetipc");
            AxAddStrToArray(rData,1,pipe);
        }
        break;

    case 0:
        AxPreventLowCodesFromClient = TRUE;
        if (getenv("AXNETLOG"))
            AxNetLogging = TRUE;
        else
            AxNetLogging = FALSE;
#if !AXMS
        for (ix = 1; ix < NSIG; ix++)
            signal(ix, SIG_IGN);
        signal(SIGCHLD,(void (*)())axnetChildTerminated);
        signal(SIGHUP, (void (*)())axnetToggleLogging);
        signal(SIGTERM, (void (*)()) axnetGracefulTerminate);
        signal(SIGINT, (void (*)())axnetInterrupted);
        NETLOG(("Determine executables dir ..."));
        p = WhereAmI(AxStrFromArray(funcData,0));
        strcpy(ExecutablesDir,p);
#else
        axnetGetExecLoc(ExecutablesDir);
#endif
        NETLOG((" <%s>\n", ExecutablesDir));
        AxSetProgramName(NXLT("axnet"));
        AxSetBrokenPipeCB(clientPipeError);

        AxNetAllowProc = TRUE;
        for (ix = 0; ix < AxArraySize(funcData); ix++) {
            if (0 == strcmp(NXLT("-restricted"), AxStrFromArray(funcData, ix)))
                AxNetAllowProc = FALSE;
        }
        break;

    case AXNET_GET_SOCKET: {
        char socket[STRPATH+1];
        NETLOG(("AxDispatch: GET_SOCKET\n"));
        SelectSocketName(socket);
        rData = AxMakeStrData(-1,socket);
        break;
    }

    case AXNET_REGISTER: {
        char *name, *socket;
        int isPrivate;
        NETLOG(("AxDispatch: REGISTER\n"));
        AxDecomposeArray(funcData,NXLT("ssi"),&name,&socket,&isPrivate);
        axnetAddService(name,socket,isPrivate);
        break;
    }

    case AXNET_UNREGISTER:
        NETLOG(("AxDispatch: UNREGISTER\n"));
        axnetDropService(AxStrFromDataPtr(funcData));
        break;

    case AXNET_DROP_SOCKET:
        NETLOG(("AxDispatch: DROP_SOCKET\n"));
        axnetDropSocket(AxStrFromDataPtr(funcData));
        break;

    case AXNET_LIST_SERVICES:
        NETLOG(("AxDispatch: LIST_SERVICES\n"));
        rData = axnetListServices();
        break;

    case AXNET_FIND_SERVICE:
        NETLOG(("AxDispatch: FIND_SERVICE\n"));
        rData = axnetFindService(funcData);
        break;

    case AXNET_PROC:
        NETLOG(("AxDispatch: PROC\n"));
        if (!AxNetAllowProc) {
            NETLOG(("    WILL NOT ALLOW PROC\n"));
        } else
            axnetProc(funcData);
        break;

    case AXNET_PROC_SLAVE:
        NETLOG(("AxDispatch: PROC_SLAVE\n"));
        if (!AxNetAllowProc) {
            NETLOG(("    WILL NOT ALLOW PROC_SLAVE\n"));
        } else
            axnetProcSlave(funcData);
        AxnetNoResponse = TRUE;
        break;

    case AXNET_SIGNAL_PROCESS:
        NETLOG(("AxDispatch: AXNET_SIGNAL_PROCESS\n"));
        axnetSignalProcess(funcData);
        break;

    case AXNET_READ_ODBC_INI:
        NETLOG(("AxDispatch: AXNET_READ_ODBC_INI\n"));
        rData = AxNetRead_ODBC_INI(funcData);
        break;

    default:
        NETLOG((XLT("Unknown command code in AxDispatch: %d\n"), funcId));
        *quitNow = 1;
        break;
    }
    return(rData);
}

static service_desc *axnetLookupService(name)
char *name;
{
    service_desc *info;

    for ( info = Services; info; info = info->next ) {
        if ( 0 == strcmp(info->name,name) )
            return(info);
    }
    return(NULL);
}

/*************************************************************************
function: axnetListServices - Return array of service names
*/
elfData axnetListServices ()
{
    service_desc *p;
    elfData ret;
    int cnt = 0;

    ret = AxMakeArray(0);
    for ( p = Services; p; p = p->next ) {
        if ( !p->isPrivate )
            ret = AxAddStrToArray(ret,cnt++,p->name);
    }
    return(ret);
}

/*************************************************************************
function: axnetProc - Start a process, can only be used for axnet servers!
*/
static void axnetProc (funcData)
elfData funcData;
{
    elfData argvP;
    int errval, size, ix, uid;
    char cmd[5000], *arg, *cmdP;
    char *progname = NULL;
    char pathname[STRPATH+1];
    char *serviceName;
    char socket[STRPATH+1];
    int isPrivate = TRUE;
    char *p;
#if AX_X11
    mbchar *username;
    struct passwd *pwP;
#endif

    uid = AxIntFromArray(funcData,0);
    argvP = AxArrayFromArray(funcData,1);

#if AX_X11
    if ( uid == -1 )
       {
       /* from a Windows/NT machine.  Get the passed username and look */
       /* up the user-id here on the unix machine                      */
       username = AxStrFromArray(funcData,2);
       if ( username != (mbchar *)NULL )
          {
          pwP = getpwnam( username );
          if ( pwP != (struct passwd *)NULL )
             uid = pwP->pw_uid;
          }
       }
#endif

    size = AxArraySize(argvP);
    progname = AxStrFromArray(argvP,0);
    if ( *progname == '*' ) {
        strcpy(pathname,AxMyProgramName);
        p = strrchr(pathname,'/');
        if ( !p )
            p = strrchr(pathname,'\\');
        if ( !p )
            progname = progname + 1;    /* lets hope its in path */
        else {
#if AXMS
            strcat(pathname,"\\");
#else
            strcat(pathname,"/");
#endif
            *(p+1) = 0;
            strcat(pathname,progname+1);
            progname = pathname;
        }
    }
    serviceName = AxStrFromArray(argvP,1);
    SelectSocketName(socket);
    sprintf(cmd,"%s %s -fork -service",progname,socket);
    cmdP = cmd + strlen(cmd);
    for ( ix = 2; ix < size; ix++ ) {
        arg = AxStrFromArray(argvP,ix);
        if ( cmdP-cmd+strlen(arg) > 4900 )
            AxError(99,NXLT("Command line too long"),NULL);
        *cmdP++ = ' ';
        strcpy(cmdP,arg);
        cmdP += strlen(cmdP);
    }
    errval = AxCallUidProcess(uid,cmd);
    if ( errval != 0 )
        AxError(99,ErrnoTextFromUnix(errval),progname);
    axnetAddService(serviceName,socket,isPrivate);
}

/*************************************************************************
function: axnetProcSlave - Start a process, can only be used for axnet servers!
*/
static void axnetProcSlave (funcData)
elfData funcData;
{
    elfData argvP;
    int size, ix, uid;
    char *progname = NULL;
    char *p;
    char proc_as_user[30];

#if AX_X11
    int save_errno;
    int errval;
    char errstr[STRPATH * 2];
    struct passwd *pwP;
    char channelStr[10];
    int child, channel;
    char pathname[STRPATH+1];
    char *argv[1000];
    int argc;
    uid_t overrideUID;
#endif

#if AXMS
    char *encrypted_password;
    char command_line[2048];
    int ret;
    char path_buff[MAX_PATH+1];
    char *filenameP;
    SOCKET curr_channel, new_channel;
    HANDLE process_h;
#endif

    uid = AxIntFromArray(funcData,0);
    argvP = AxArrayFromArray(funcData,1);

    size = AxArraySize(argvP);
    progname = AxStrFromArray(argvP,0);
    p = AxStrFromArray(funcData,2);
    strcpy( proc_as_user, p );


#if AX_X11                   /*  {  */
/***** Deutche Bank fix for leak #2119554 *** --Kiran ****
Read axnet.cnf to find if uidFor is specified, if so don't
call getpwnam. ***/

    axnetGetConfig(); /* get latest version of axnet.cnf */

    /* we need to determine if the configuration file
    will cause the child process to be started with some
    uid other than the uid arg to this function. If so, we
    keep a list of such processes. */
    overrideUID = configFileUID(uid,progname);


    if ( uid == -1 && !overrideUID )
       {
       /* from a Windows/NT machine.  Get the passed username and look */
       /* up the user-id here on the unix machine                      */
       NETLOG(("Getting uid for user <%s>\n", proc_as_user ));
       pwP = getpwnam( proc_as_user );
       if ( pwP != (struct passwd *)NULL )
          uid = pwP->pw_uid;
       }

/* prepare a command line for task to start */
    argc = 0;
    argv[argc++] = progname; /* may be changed in axnetConfigForExec */
    argv[argc++] = "-fork";
    argv[argc++] = "-service";
    argv[argc++] = "-channel";
    sprintf(channelStr,"%d",AxChannel());
    argv[argc++] = channelStr;
    for ( ix = 1; (ix < size) && (argc < 500); ix++ )
        argv[argc++] = AxStrFromArray(argvP,ix);
    argv[argc] = NULL; /* args may be added in axnetConfigForExec */

    NETLOG(("Program args are:\n"));
    for ( ix=0; ix<500; ix++ )
        {
        if ( argv[ix] == NULL )
           break;
        NETLOG(("   <%s>\n", argv[ix]));
        }

    /*----------------------------------------------------*/
    /* Fork and exec process, passing the -channel option */
    /*----------------------------------------------------*/
    NETLOG(("forking ..."));
    child = fork();
    if ( child == -1 )
        AxError(errno,ErrnoTextFromUnix(errno),NXLT("fork in axnet"));
    if ( child ) {          /* parent closes channel and continues */
        if (overrideUID)
            addSpawn(child, uid, overrideUID); /* remember this */
        AxCloseChannel();
        return;
    }
    close(AxSvrSockFd);     /* and child closes socket */

    axnetConfigForExec(uid,
            progname,
            argc,argv,1000); /* additional args may be added */

    NETLOG(("Child: program args after axnetConfigForExec() are:\n"));
    for ( ix=0; ix<500; ix++ )
        {
        if ( argv[ix] == NULL )
           break;
        NETLOG(("   <%s>\n", argv[ix]));
        }

    execvp(argv[0], argv);
    save_errno = errno;

    sprintf(errstr, NXLT("exec: %s: %s"), argv[0],
            ErrnoTextFromUnix(save_errno));
    NETLOG(("*** AXNET_PROC_SLAVE: child execvp fails: %s\n", errstr));
    _exit(save_errno);

#else                                   /*  } {  */

   /** Windows NT  **/

   if ( *progname == '*' )
      {
      /* command is to be run from the same directory as axnet */
      sprintf( path_buff, "%s\\%s", ExecutablesDir, (progname+1) );
      }
   else
      {
      /* look for the executable on the path */
      /* the input command is a path, or a relative path.  Since */
      /* MS_CreateProcess will not search the path, we must do it now */
      /* first we look for a ".EXE, then a ".BAT", then a ".COM"      */
      ret = MS_SearchPath( (LPCTSTR)NULL, progname, (LPCTSTR)".EXE",
                               sizeof(path_buff), path_buff, &filenameP );
      if ( !ret )
          ret = MS_SearchPath( (LPCTSTR)NULL, progname, (LPCTSTR)".BAT",
                               sizeof(path_buff), path_buff, &filenameP );
      if ( !ret )
          ret = MS_SearchPath( (LPCTSTR)NULL, progname, (LPCTSTR)".COM",
                               sizeof(path_buff), path_buff, &filenameP );
      if ( !ret )
          /* just go with what we have, will probably fail but ... */
          strcpy( path_buff, progname );
      }

    /* now duplicate the socket to make sure that it is an inheritable */
    /* item  (on Win95 sockets do not seem to be inheritable by default) */
    /* The channel handle is passed on the command line as an integer */
    /* argument, and turned back into a socket descriptor in the child */
    /* (see "elfmain()" in elrcp3.c)                                   */
    curr_channel = AxChannel();
    process_h = MS_GetCurrentProcess();
    ret = MS_DuplicateHandle( process_h, (HANDLE)curr_channel, process_h,
                (HANDLE*)&new_channel, 0, TRUE, DUPLICATE_SAME_ACCESS );
    if ( ret == 0 )
        {
        /* do not know why it would fail, but try to use the original */
        ret = GetLastError();
        new_channel = curr_channel;
        }

    sprintf( command_line, "%s -fork -service -channel %d",
                                               path_buff, new_channel );


    /* add on any other arguments to the command line */
    for ( ix = 1; ix < size; ix++ )
       {
       p = AxStrFromArray(argvP,ix);
       strcat( command_line, " " );
       strcat( command_line, p );
       }

    encrypted_password = AxStrFromArray(funcData,3);
    StartWinProcAsUser( command_line, proc_as_user, encrypted_password );
    return;
#endif                                  /*  }  */
}

/***********************************************************************
function: axnetSignalProcess - Send a signal to a process.
*/
static elfData
axnetSignalProcess(funcData)
elfData funcData;
{
#if AX_X11
    pid_t pid;
    int signum;
    uid_t uid=0;
    char errstr[STRPATH+32];
    int err,uidArg;
    mbchar *proc_as_user;
    struct passwd *pwP;
    char *username;
    spawnItem *spawnP;

    pid = (pid_t) AxIntFromArray(funcData,0);
    signum = (int) AxIntFromArray(funcData, 1);
        /* convert signal from generic ax constant to platform-specific */
    signum = portSpecificSignal(signum);
    if( signum == AXNOSIG )
        AxError(99, XLT("Attempt to send illegal signal"),"");
    uidArg = AxIntFromArray(funcData, 2);

    if ( uidArg == -1 )
       {
       /* from a Windows/NT machine.  Get the passed username and look */
       /* up the user-id here on the unix machine                      */
       username = AxStrFromArray(funcData,3);
       if ( username != (mbchar *)NULL )
          {
          pwP = getpwnam( username );
          if ( pwP != (struct passwd *)NULL )
             uid = pwP->pw_uid;
          }
       }
    else
        uid = (uid_t) uidArg;

    /* if the process we want to signal was started for user uid, but
    was started with another uid, the latter uid is the one to 
    switch to before sending the signal */
    spawnP = getSpawnItem(pid);
    if (spawnP) {
        if (spawnP->requestor != uid) {
            sprintf(errstr, NXLT("User %d can not signal pid %d"), uid, pid);
            AxError(99, errstr, "");
            return(NULL);
        }
        uid = spawnP->actual; /* this is the uid of the process */
    }

    if (pid < 2) {
        /* Note pids -1, 0, 1 have special meanings.  Sending signals
           to them cause disaster.  See man page on kill(2). */
        sprintf(errstr, NXLT("Will not deliver signal to pid %d"), pid);
        AxError(99, errstr, "");
        return(NULL);
    }

    err = deliversig(signum, pid, uid);
    if (err == 0)
        return(NULL);
    /* if there is failure, and pid is our child, attempt to send
    signal on our own behalf */

    sprintf(errstr, NXLT("Signal to pid %d as user %d failed: %s"),
            pid, uid,
            ErrnoTextFromUnix(err));
    AxError(err, errstr, "");
#else
    printf("deliversig: Port Me!\n");
#endif

    return(NULL);
}

#if AX_X11
/***********************************************************************
function: deliversig - Deliver a signal 'sig' to 'pid' on behalf of
user 'uid'.  Do a fork(), have the child setuid to 'uid' and call kill(2).
The child exits with 0 on success, 'errno' on failure.  The parent
catches the child's exit code and returns 0 on success, 'errno' on failure.
*/
static int
deliversig(sig, pid, uid)
int sig;
pid_t pid;
uid_t uid; /* on behalf of 'uid' */
{
    int err;
    pid_t child;
    int code;
    int status = 0;
    struct sigaction ignore, saveintr, savequit;
    sigset_t childmask, savemask;

    NETLOG(("send signal %d to process %d as user %d\n",sig,pid,uid));

    /* we can try to kill our own children directly. */

    if (uid == getuid()) {
        if (waitpid(pid, &status, WNOHANG) == -1) {
            if (errno == ECHILD)
                return(1); /* pid is not our child, or doesn't exist */
            else
                return(1); /* some other error, also forget it. */
        }
        return(kill(pid, sig));
    }

    /* become given uid to send signal */

    ignore.sa_handler = SIG_IGN;
    sigemptyset(&ignore.sa_mask);
    ignore.sa_flags = 0;
    if (sigaction(SIGINT, &ignore, &saveintr) < 0)
        return(1);
    if (sigaction(SIGQUIT, &ignore, &savequit) < 0)
        return(1);
    sigemptyset(&childmask);
    sigaddset(&childmask, SIGCHLD);
    if (sigprocmask(SIG_BLOCK, &childmask, &savemask) < 0)
        return(1);

    child = fork();
    if (child == -1)
        return(errno);

    if (!child) { /* child */
        if (uid != getuid())
            if ((err = setuid(uid)) < 0)
                _exit(errno);
        err = kill(pid, sig);
        if (err < 0)
            _exit(errno);
        _exit(0);
    }

    waitpid(child, &status, 0);

    if (sigaction(SIGINT, &saveintr, NULL) < 0)
        return(1);
    if (sigaction(SIGQUIT, &savequit, NULL) < 0)
        return(1);
    if (sigprocmask(SIG_SETMASK, &savemask, NULL) < 0)
        return(1);

    code = WEXITSTATUS(status);
    return(code);
}
#endif /* AX_X11 */

/***********************************************************************
function: axnetProxy - Function as a pipeline to real server
*/
void axnetProxy (host,socket)
char *host, *socket;
{
    service_desc info;

    info.next = NULL;
    info.name = "proxy-server";
    info.pathname = "";
    info.socket = NULL;
    info.isRunning = TRUE;
    info.isPrivate = FALSE;
    NETLOG(("Establishing proxy connection to host <%s>\n", host));
    axnetEstablishRemoteConnection(&info,host,atoi(socket));
    NETLOG(("done proxy connection\n"));
}

/*************************************************************************
function: axnetFindService - Find a named service
If there is no entry, return datum 0.
If the connection is remote (re-use axnet channel) return NULL.
If the connection is local, return socket string
*/
static elfData axnetFindService (find)
elfData find;
{
    int isRemote;       /* true if cleint is calling from a another machine */
    service_desc *info;
    char *serviceName;

/* Find entry for service */

    serviceName = AxStrFromArray(find,0);
    if ( 0 == strncmp(NXLT("@proxy@"),serviceName,7) ) {
        char buf[100];
        int ix = 0;
        strcpy(buf,serviceName+7);
        while ( buf[ix] && (buf[ix] != '@') ) ix++;
        if ( !buf[ix] )
            AxError(99,serviceName,NXLT("badly formed proxy"));
        buf[ix] = 0;
        axnetProxy(buf,&buf[ix+1]);
        return(NULL);
    }
    isRemote = AxBoolFromArray(find,1);
    info = axnetLookupService(serviceName);

    if ( !info ) {                  /* no server running */
        return(AxMakeIntData(0));
    }
    else if ( !isRemote ) {         /* local, return socket name */
        return(AxMakeStrData(-1,info->socket));
    }
    else {                          /* remote */
        axnetEstablishRemoteConnection(info,NULL,0);
        return(NULL);
    }
}

/*************************************************************************
function: axnetEstablishRemoteConnection - Fork for connection server
*/
static elfData axnetEstablishRemoteConnection (info,host,socknum)
service_desc *info;
char *host;     /* NULL or hostname to proxy to */
int socknum;    /* 0 or proxy host's port number */
{
#if AX_X11    /*  {  */
    int ix;
    int child, channel;
    char *socket;
    elfData request;
    int clientChannel;
    int code;
    char *str, *obj;
    int serverChannel;
    elfData rP;
    static fd_set read_fds, wtfds, nofds;
    int events;
    int32 nsize;
    char *inBuf, *outBuf;
    int processed;
    int size, cnt;


/* If got here, client is remote.  Fork, connect and return channel */

    child = fork();
    if ( child == -1 )
        AxError(errno,ErrnoTextFromUnix(errno),NXLT("fork in axnet"));

    clientChannel = AxChannel();
#if 1
    if ( child ) {      /* parent closes channel and continues */
        AxCloseChannel();
        return(NULL);
    }
    else                    /* and child closes socket */
        close(AxSvrSockFd);
#else
    if ( child == 0 ) {     /* child closes channel and continues as axnet */
        AxCloseChannel();
        return(NULL);
    }
    else                    /* and parent closes socket */
        close(AxSvrSockFd);
#endif

/* Connect to target service here on this machine */

    socket = info->socket;
    channel = AxClientConnect(socknum,socket,host,&code,&str,&obj);
    if ( !host && ((code == ECONNREFUSED) || (code == ENOENT)) ) {
        int axnetChannel;   /* axnet may have listed a server that's gone */
        elfData response;

        axnetChannel = AxNetConnect(NULL,&code,&str,&obj);  /* tell axnet to drop service if was there */
        AxNetDropServiceBySocket(axnetChannel,socket,&code,&str,&obj);
        if ( code != 0 ) {
            AxClientDisConnect(axnetChannel);
            AxError(code,str,obj);
        }
        request = AxMakeArray(2);
        request = AxAddStrToArray(request,0,info->name);
        AxSendMessageToServer(axnetChannel,AXNET_FIND_SERVICE,request,&code, &str, &obj);
        if ( code != 0 ) {
            AxClientDisConnect(axnetChannel);
            AxError(code,str,obj);
        }
        response = AxGetResponseFromServer(axnetChannel,&code,&str,&obj);
        AxClientDisConnect(axnetChannel);
        if ( code != 0 )
            AxError(code,str,obj);
        channel = AxClientConnect(0,AxStrFromDataPtr(response),NULL,&code,&str,&obj);
    }
    if ( code != 0 )
        AxError(code,str,NXLT("remote connection to service"));

    serverChannel = channel;

/* Respond to client that it should continue to use its current channel */

    rP = AxMakeArray(2);
    AxAddIntToArray(rP,0,0);
    AxAddArrayToArray(rP,1,NULL);
    size = AxMemWrite(rP,0,&outBuf);
    cnt = write(clientChannel,outBuf,size);
    if ( cnt != size )
        exit(0);

/*  Read from serverChannel, write to clientChannel.
    Read from clientChannel, write to serverChannel
*/

    while ( TRUE ) {

        FD_SET(clientChannel,&read_fds);
        FD_SET(serverChannel,&read_fds);
        events = select(FD_SETSIZE, &read_fds, &wtfds, &nofds, NULL);
        if ( events <= 0 )  /* error occurred */
            break;

        if ( FD_ISSET(clientChannel,&read_fds) ) {
            cnt = read(clientChannel,&nsize,sizeof(int32));
            if ( cnt <= 0 )
                break;
            size = ntohl(nsize);
            inBuf = (char*)AxMalloc(size);
            processed = 0;
prdr:       cnt = read(clientChannel,inBuf+processed,size-processed);
            if ( cnt <= 0 )
                break;
            if ( cnt != size-processed ) {
                processed += cnt;
                goto prdr;
            }
            cnt = write(serverChannel,&nsize,sizeof(int32));
            if ( cnt < sizeof(int32) )
                break;
            cnt = write(serverChannel,inBuf,size);
            if ( cnt != size )
                break;
            AxFree(inBuf);
        }
        if ( FD_ISSET(serverChannel,&read_fds) ) {
            cnt = read(serverChannel,&nsize,sizeof(int32));
            if ( cnt <= 0 )
                break;
            size = ntohl(nsize);
            inBuf = (char*)AxMalloc(size);
            processed = 0;
crdr:       cnt = read(serverChannel,inBuf+processed,size-processed);
            if ( cnt <= 0 )
                break;
            if ( cnt != size-processed ) {
                processed += cnt;
                goto crdr;
            }
            cnt = write(clientChannel,&nsize,sizeof(int32));
            if ( cnt < sizeof(int32) )
                break;
            cnt = write(clientChannel,inBuf,size);
            if ( cnt != size )
                break;
            AxFree(inBuf);
        }
    }
    shutdown(serverChannel,2);
    shutdown(clientChannel,2);
    close(serverChannel);
    close(clientChannel);
    _exit(0);
#else    /*  }  {  */
/*** for Windows NT ***/
/* On NT we start a thread rather than fork-ing. The thread does the same thing
 * the fork'd child would do, which is to act as a proxy between the client
 * axmain and the server axmain, while the "parent" axnet can continue
 * to service other requests.            -donato
 */
    int clientChannel;
    DWORD thread_id;
    HANDLE hthread;
    remote_conn_t *thread_args;

    clientChannel = AxChannel();

    thread_args = malloc(sizeof(remote_conn_t));
    thread_args->clientChannel = clientChannel;
    thread_args->socknum = socknum;
    if ( host )
        thread_args->host = strcpy(malloc(strlen(host)+1), host);
    else
        thread_args->host = NULL;
    thread_args->socket = strcpy(malloc(strlen(info->socket)+1),info->socket);
    thread_args->service_name = strcpy(malloc(strlen(info->name)+1),info->name);

    hthread = MS_CreateThread((LPSECURITY_ATTRIBUTES)NULL, (DWORD)0,
                                (LPTHREAD_START_ROUTINE)RemoteConnectionThread,
                                (void*)thread_args, (DWORD)0, &thread_id);
    MS_CloseHandle(hthread);  /* this doesn't stop the thread, just
                               * invalidates the handle, which we don't need.
                               */
    /* On Unix, the parent would close SvrSockChannel. On NT we can't because
     * the thread has the same handle, not a clone of it. So instead we simply
     * invalidate it (set it to -1) here, while the thread still has the
     * valid handle.    -donato
     */
    AxInvalidateChannel();
    return( (elfData)NULL );
#endif   /*   }  AX_X11  */
}

/*************************************************************************
function: axnetStartService - Fire up process, return connection info
*/
static service_desc *axnetStartService (pathname,serviceName,cmdOptions,doProc)
char *pathname;
char *serviceName;
elfData cmdOptions;
int doProc;     /* TRUE if should actually proc process */
{
    char cmd[10000];
    char socket[30];
    service_desc *info;
    int errval;
    int ix;
    int size;

    SelectSocketName(socket);
    info = (service_desc*)AxMalloc(sizeof(service_desc));
    memset(info,0,sizeof(service_desc));
    if ( serviceName && serviceName[0] )
        info->name = strcpy(AxMalloc(strlen(serviceName)+1),serviceName);
    else
        info->name = strcpy(AxMalloc(strlen(socket)+1),socket);
    info->socket = strcpy(AxMalloc(strlen(socket)+1),socket);
    info->next = Services;
    Services = info;
    if ( doProc ) {
        sprintf(cmd,"%s %s -fork",pathname,socket);
        size = AxArraySize(cmdOptions);
        for ( ix = 0; ix < size; ix++ ) {
            strcat(cmd," ");
            strcat(cmd,AxStrFromArray(cmdOptions,ix));
        }
        errval = AxCallProcess(cmd);
        if ( errval != 0 )
            AxError(errval,ErrnoTextFromUnix(errval),pathname);
        info->isRunning = TRUE;
    }
    return(info);
}

/*************************************************************************
function: axnetDropSocket - Remove service by socket name
*/
void axnetDropSocket (name)
char *name;
{
    service_desc *info, *prev = NULL;

again:
    for ( info = Services; info; info = info->next ) {
        if ( 0 == strcmp(info->socket,name) ) {
            if ( prev )
                prev->next = info->next;
            else
                Services = info->next;
            free(info->name);
            if ( info->pathname )
                free(info->pathname);
            free(info->socket);
            free(info);
            goto again;     /* get all services on this socket */
        }
        prev = info;
    }
}

/*************************************************************************
function: axnetDropService - Remove service by service name
*/
void axnetDropService (name)
char *name;
{
    service_desc *info, *prev = NULL;

    for ( info = Services; info; info = info->next ) {
        if ( 0 == strcmp(info->name,name) ) {
            if ( prev )
                prev->next = info->next;
            else
                Services = info->next;
            free(info->name);
            if ( info->pathname )
                free(info->pathname);
            free(info->socket);
            free(info);
            return;
        }
        prev = info;
    }
}

/*************************************************************************
function: axnetAddService - Create new service
*/
void axnetAddService (name, socket, isPrivate)
char *name, *socket;
int isPrivate;
{
    service_desc *p;

    p = axnetLookupService(name);
    if ( p ) {
        axnetDropService(name);
    }
    p = (service_desc*)AxMalloc(sizeof(service_desc));
    memset(p,0,sizeof(service_desc));
    p->name = strcpy(AxMalloc(strlen(name)+1),name);
    p->socket = strcpy(AxMalloc(strlen(socket)+1),socket);
    p->next = Services;
    Services = p;
}

/*************************************************************************
function: SelectSocketName - Trivial unque name mechanism for child servers
The verifes there is no existant file left over from a prior session of axnet.
*/
static void SelectSocketName (socket)
char *socket;
{
#if AX_X11
    static int uid = 0;
    int ix;

    for ( ix = 0; ix < 10000; ix++ ) {
        sprintf(socket,"/tmp/.axsc%d",++uid);
        if ( (-1 == access(socket,0)) && (errno == ENOENT) )
            return; /* will always happen unless there are old leftover axnet files */
    }
    printf("Unable to create %s\n",socket); /* will NEVER happen */
    exit(1);
#else
    /* on NT we use a port number instead of a named pipe. */
    sprintf(socket, "%d", axmsCreateAndBindToSocket());
#endif
}

/***********************************************************************
function: AxNetLog - Log axnet messages to "/tmp/axnetlog"
*/
void
AxNetLog (char* fmt,...)
{
    va_list argp;

    if (!AxNetLogging)
        return;
    if (!netlogfp)
#if AXMS
        netlogfp = fopen("c:\\temp\\axnetlog", "w");
#else
        netlogfp = fopen("/tmp/axnetlog", "w");
#endif
    if (!netlogfp)
        return;
    va_start(argp, fmt);
    vfprintf(netlogfp, fmt, argp);
    va_end(argp);
    fflush(netlogfp);
    return;
}


#if AX_X11    /* { */
/*************************************************************************
function: axnetConfigForExec - process commands in config file related to
    executing this named executable. This sets environ variables and
    add args to argv. Call after fork and before exec.
*/
void axnetConfigForExec(uid, execNameArg,argc,argv,argvSize)
int uid;    /* change to this uid before exec, unless there is an
            override uid in the config file. */
char *execNameArg; /* name, with possible * prefix. If it has a * prefix,
                prepend path. in any case, resultant execname is
                copied to argv[0]. */
int argc;   /* count of args in array already. */
char *argv[]; /* (in,out) array of args. has args added when done.
                argv[0] is set to pathname to execute. */
int argvSize;   /* size of argv */
{
    register struct axnetConfigLine *lineP;
    char *getenv();
    char *p;
    int prependPath;
    static char execName[STRPATH+1];
    static char execPath[STRPATH+1];
    struct stat statbuf;
    int uidval;


    /* if there is a "*" prefix, prepend a path to name. */
    prependPath = (execNameArg[0] == '*');
    execNameForConfig(execName,execNameArg); /* get lookup name */

    if (prependPath) {
        /* axnet and configuration file pathname is default path for
        executable */
        strcpy(execPath,AxMyProgramName);
#if AXMS
        p = strrchr(execPath,'\\');
#else
        p = strrchr(execPath,'/');
#endif
        if ( p )
            *(p) = 0;
        else
            strcpy(execPath,NXLT(".")); /* prog name has no path */

        /* get configuration file info */
#if AXMS
        strcat(execPath,"\\");
#else
        strcat(execPath,"/");
#endif
        argv[0] = strcat(execPath,execName); /* default exec path */
    }
    else {
        argv[0] = execNameArg; /* use exec name as is, by default */
    }

    if (axnetConfig) {
        int lx;

        /* do relevent commands from config file */
        for (lx=0; lx<axnetConfig->linecount; lx++) {

            lineP = &axnetConfig->line[lx];
            if (!lineP->arg1 || strcmp(lineP->arg1, execName) != 0)
                continue; /* not relevent */
            switch(lineP->command) {
            case SETENV_FOR: /* set environ variable */
                if (lineP->arg2)
                    putenv(lineP->arg2);
                break;
            case ARG_FOR: /* arg for exec */
                if (lineP->arg2)
                    if (argc < argvSize)
                        argv[argc++] = lineP->arg2;
                break;
            case LIB_FOR: /* extend LD_LIBRARY_PATH, or LIBPATH or whatever */
                if (lineP->arg2) {
                    char *libPath, *newLibPath;
                    int len;
#if RIOS
#define SHARED_LIB_ENV_VAR "LIBPATH" /* AIX */
#elif HP800V3
#define SHARED_LIB_ENV_VAR "SHLIB_PATH" /* HP-UX */
#else
#define SHARED_LIB_ENV_VAR "LD_LIBRARY_PATH" /* SUN */
#endif

                    len = strlen(lineP->arg2) + 32;
                    libPath = getenv(SHARED_LIB_ENV_VAR);
                    if (libPath != NULL) { /* append old pathlist to new */
                        newLibPath = (char *)AxMalloc(len+strlen(libPath));
                        sprintf(newLibPath,"%s=%s:%s",
                                SHARED_LIB_ENV_VAR,
                                lineP->arg2,
                                libPath);
                    }
                    else {
                        newLibPath = (char *)AxMalloc(len);
                        sprintf(newLibPath,"%s=%s",
                                SHARED_LIB_ENV_VAR,
                                lineP->arg2);
                    }
                    putenv(newLibPath);
                }
                break;
            case PATH_FOR:
                /* if execNameArg begins with "*":
                if pathname given here is to a directory,
                then replace the * with the pathname prefix given here.
                if pathname here names a file, then use that as
                file to execute, replacing entire execNameArg. */
                if (!prependPath) break;
                if (lineP->arg2) {
                    /* check if file has changed since last time it was read */
                    if (stat(lineP->arg2,&statbuf) == 0) {
                        if (S_ISDIR(statbuf.st_mode)) {
                            strcpy(execPath,lineP->arg2);
#if AXMS
                            strcat(execPath,"\\");
#else
                            strcat(execPath,"/");
#endif
                            argv[0] = strcat(execPath,execName);
                            
                        }
                        else if (S_ISREG(statbuf.st_mode)) {
                            argv[0] = lineP->arg2;
                        }
                    }
                }
                break;
            case UID_FOR:
                /* this is the uid for this executable to run under */
                if (lineP->arg2) {
                    uidval = atoi(lineP->arg2);
                    if (uidval > 0) uid = uidval;
                }
            }
        }
    }

    
    /* args may have been added above */

    if (argc < argvSize)
        argv[argc] = NULL;
    else
        argv[argvSize] = NULL;

    /* set uid */
    if ( uid != getuid() ) {
        char errstr[100];
        if ( 0 != setuid(uid) ) {
            sprintf(errstr,NXLT("setuid to %d"),uid);
            AxError(errno,ErrnoTextFromUnix(errno),errstr);
        }
    }

    axnetSetLib(WhereAmI(argv[0]));
    
    if (AxNetLogging) {
        int ix;
        NETLOG(("as user %d, exec %s with args:\n",uid,argv[0]));
        for (ix=1; ix<argc; ix++) {
            NETLOG(("  %s\n",argv[ix]));
        }
    }

}
#endif    /* } AX_X11 */


/*************************************************************************
function: execNameForConfig - construct name of executable as used
to look up configuration file entries
*/
void execNameForConfig(lookupName, execNameArg)
char *lookupName; /* out: name to use for lookup */
char *execNameArg; /* name, with possible * prefix. If it has a * prefix,
                prepend path. in any case, resultant execname is
                copied to argv[0]. */
{
    char *p;

    if (execNameArg[0] == '*') {
        execNameArg++;
    }
    /* if execName is not a simple filename make a simple file name,
    which is needed to scan the config file for relevent entries. */
    p = strrchr(execNameArg,'/');
    if ( !p )
        p = strrchr(execNameArg,'\\');
    if (p)
        strcpy(lookupName,p+1);
    else
        strcpy(lookupName,execNameArg);
}

/*************************************************************************
function: configFileUID  - if non-zero is unix uid to use when
running given exec, and is different from requestor's uid.
*/
uid_t configFileUID(uid_t uid, char* execNameArg)
 /* uid_t uid;  * in: requestor's uid *
char *execNameArg; * in: name, with possible * prefix. */
{
    register struct axnetConfigLine *lineP;
    char execName[STRPATH+1];
    uid_t overrideUID = 0;

    execNameForConfig(execName,execNameArg); /* get lookup name */

    /* look for uid override for this exec name */
    if (axnetConfig) {
        int lx;

        /* do relevent commands from config file */
        for (lx=0; lx<axnetConfig->linecount; lx++) {

            lineP = &axnetConfig->line[lx];
            if (lineP->command != UID_FOR
                    || !lineP->arg1
                    || strcmp(lineP->arg1, execName) != 0)
                continue; /* not relevent */

            /* this is the uid for this executable to run under */
            if (lineP->arg2) {
                overrideUID = atoi(lineP->arg2);
            }
        }
    }

    return ( (overrideUID != uid) ? overrideUID : 0 );
    
}

/*************************************************************************
function: axnetGetConfig 

axnet configuration file. Used to control what happens when axnet exec's a
process. Read whenever axnetProcSlave() is called. contains lines of ascii
text. When file is read, date last modified is stored, file is 
re-read if file has been modified.

File location:
 file is named axnet.cnf and is in same directory as axnet.

File Format:

<command> <arg1> <arg2> <NL>

 In the following, <exec> is the filename, not full pathname,
of the file to be executed.

<command> is
    setenvFor  <exec> <setenvArg>
            <setenvArg> is string of the form name=val.
            added to environment of process <exec> to be executed.
    libFor <exec> <lib path>
            <lib path> is prepended to LD_LIBRARY_PATH or SHLIB_PATH or other
            environ variable for shared libs.
    argFor  <exec> <val>
            <val> is added to command line for <exec>. forexample,
            argFor foo -l -O name
    pathFor <exec> <path>
            <path> is path for <exec>. If path is a file, execute it, otherwise
            execute <path>/<exec>. This command is only done when the
            name of the the executable was given to axnet with a "*" prefix.
            That is, if executable was "*foo", do this, running
            <path>/foo. If executable name
            was given as "foo" do not do the pathFor command.
    uidFor  <exec> <unix uid>
            run <exec> as user with id <unix uid>

 Lines starting with "#" are comments.

*/
void axnetGetConfig()
{
#define LINEBUFSIZ 1000
    register struct axnetConfigLine *lineP;
    register axnetConfigInfo *aCIP;
    char linebuf[LINEBUFSIZ+1];
    static char seps[] = " \t";
    char *tok, *p;
    FILE *fp;
    int lx,commandCode;
    struct stat statbuf;

    /* first time, make configuration file name */
    if (axnetConfigFile[0] == '\0') {
        strcpy(axnetConfigFile,AxMyProgramName);
#if AXMS
        p = strrchr(axnetConfigFile,'\\');
#else
        p = strrchr(axnetConfigFile,'/');
#endif
        if ( p ) {
            *(p) = 0;
        }
        else
            strcpy(axnetConfigFile,NXLT(".")); /* prog name has no path */

#if AXMS
        strcat(axnetConfigFile,"\\axnet.cnf");
#else
        strcat(axnetConfigFile,"/axnet.cnf");
#endif
    }

    /* check if file has changed since last time it was read */
    if (stat(axnetConfigFile,&statbuf) == 0) {
        if (axnetConfig && statbuf.st_mtime == axnetConfig->last_modified) {
            return; /* file already processed */
        }
    }
    else if (errno != ENOENT && errno != ENOTDIR) {
        /* error accessing file */
        AxError(errno,ErrnoTextFromUnix(errno),axnetConfigFile);
    }
    else {
        /* file does not exist */
        return;
    }


    /* open file */
    fp = fopen(axnetConfigFile,"r");
    if (!fp) {
        if (errno != ENOENT && errno != ENOTDIR) 
            AxError(errno,ErrnoTextFromUnix(errno),axnetConfigFile);
        return;
    }

    /* free old */
    if (axnetConfig) {
        for (lx=0; lx<axnetConfig->linecount; lx++) {
            if (axnetConfig->line[lx].arg1)
                AxFree(axnetConfig->line[lx].arg1);
            if (axnetConfig->line[lx].arg2)
                AxFree(axnetConfig->line[lx].arg2);
        }
        AxFree((void*)axnetConfig);
        axnetConfig = 0;
    }

    /* allocate new */
    aCIP = (axnetConfigInfo *)AxMalloc(sizeof(axnetConfigInfo));
    memset(aCIP,0,sizeof(axnetConfigInfo));
    aCIP->last_modified = statbuf.st_mtime; /* time last modified */

    /* parse lines */
    while (fgets(linebuf,LINEBUFSIZ,fp)) {
        lx = strlen(linebuf)-1;
        if (lx >= 0 && linebuf[lx] == '\n') { /* strip newline */
            linebuf[lx] = 0;
        }
        
        /* get command code */
        tok = strtok(linebuf,seps);
        if (!tok) continue; /* skip line */
        if (tok[0] == '#') continue; /* comment line */

        if (strcmp(tok,NXLT("setenvFor"))==0) {
            commandCode = SETENV_FOR;
        }
        else if (strcmp(tok,NXLT("argFor"))==0) {
            commandCode = ARG_FOR;
        }
        else if (strcmp(tok,NXLT("libFor"))==0) {
            commandCode = LIB_FOR;
        }
        else if (strcmp(tok,NXLT("pathFor"))==0) {
            commandCode = PATH_FOR;
        }
        else if (strcmp(tok,NXLT("uidFor"))==0) {
            commandCode = UID_FOR;
        }
        else
            commandCode = 0; /* not known */

        /* get args for command */
        switch (commandCode) {
        case PATH_FOR:
        case ARG_FOR:
        case SETENV_FOR:
        case LIB_FOR:
        case UID_FOR:
            lx = aCIP->linecount; /* new entry */
            aCIP = (axnetConfigInfo *)
                    AxRealloc((void*)aCIP, sizeof(axnetConfigInfo)
                            + (lx * sizeof (struct axnetConfigLine)));
            lineP = &aCIP->line[lx];
            memset(lineP,0,sizeof(struct axnetConfigLine));
            aCIP->linecount++;
            lineP->command = commandCode;

            tok = strtok(0,seps); /* <exec> */
            if (!tok) break;
            lineP->arg1 = strcpy(AxMalloc(strlen(tok)+1),tok);
            tok = strtok(0,""); /* rest of line */
            if (!tok) break;
            lineP->arg2 = strcpy(AxMalloc(strlen(tok)+1),tok);
            break;
        }
    }
    fclose(fp);

    axnetConfig = aCIP; /* set global */

}



/***************************************************************************
function: AxNetRead_ODBC_INI - Read the relevant .odbc.ini file and send back
the ODBC Data Sources part of the file.
*/
elfData
AxNetRead_ODBC_INI (fData)
elfData fData;
{
#if AXMS                     /*  {  */
    /* The real way to do this on Windows is to read the registry   */
    /* I am replacing all of the .INI file stuff below, since ODBC.INI */
    /* on windows often has a lot of non-data source stuff written to it */
    /* by some of the Windows tools (why, I do not know!)                */
    /*  December 11, 1997   - jon                                        */
    HKEY reg_key, detail_key;
    char key_name[400], class_string[400], description[400];
    char detail_name[400];
    int size, size_class;
    FILETIME file_time;
    int ix, ret, ret2;
    int type_code;
    elfData retData;

    retData = AxMakeArray(0);

    /* first open a key to where the dta sources are listed */
    ret = RegOpenKeyEx( HKEY_LOCAL_MACHINE, 
                "SOFTWARE\\ODBC\\ODBC.INI\\ODBC Data Sources",
                0, KEY_READ, &reg_key );
    if ( ret != ERROR_SUCCESS )
         {
         /* some error, no sources available */
         return( retData );
         }

    
    /* now enumerate all of the data source names */
    ix = 0;
    size = sizeof(key_name);
    size_class = sizeof(class_string);
    ret = RegEnumValue( reg_key, ix, key_name, &size, 0, NULL, NULL, NULL );
    while ( ret == ERROR_SUCCESS )
        {
        /* for the datasource, get its description */
        sprintf( detail_name, "SOFTWARE\\ODBC\\ODBC.INI\\%s", key_name );
        strcat( key_name, NXLT("=") );
        ret2 = RegOpenKeyEx( HKEY_LOCAL_MACHINE, detail_name,
                                                     0, KEY_READ, &detail_key );
        if ( ret2 == ERROR_SUCCESS )
            {
            size = sizeof(description);
            ret2 = RegQueryValueEx( detail_key, "Description",
                                             0, NULL, description, &size );
            if ( (ret2 == ERROR_SUCCESS) && (size > 1) )
                {
                strcat( key_name, description );
                }
            RegCloseKey( detail_key );
            }
        retData = AxAddStrToArray(retData, ix, key_name);

        /* on to the next entry */
        size = sizeof(key_name);
        size_class = sizeof(class_string);
        ix++;
        ret = RegEnumValue( reg_key, ix, key_name, &size, 0, NULL, NULL, NULL );
        }

    RegCloseKey( reg_key );
    return( retData );


#else              /*  }   { */
#define LINEBUFSIZ 1000
    int ix;
    int len;
    int uid;
    int envlen;
    int datasources_only;
    int started_datasources;
    char *p;
    char *envname;
    char errstr[256];
    char ini_file[STRPATH];
    char lineBuf[LINEBUFSIZ+1];
    FILE *fp;
    elfData retData;
    struct axnetConfigLine *lineP;
#if AX_X11                   /*  {  */
    char *username;
    struct passwd *pw_entry;
#else                            /*  } AXMS {  */
    int size;
    char filename[MAX_PATH+1];
#endif                           /*  }  */

    /* Which odbc.ini to use?
     * 
     * 1. Pathname to a odbc.ini passed as an arg to this function.
     * 2. Environment variable ODBCINI set using setenvFor in axnet.cnf.
     * 3. Environment variable ODBCINI set in the shell.
     * 4.
     *    On UNIX:
     *        a) .odbc.ini in home dir of userid set using uidFor in axnet.cnf.
     *        b) .odbc.ini in passed user's home directory.
     *    On NT:
     *        ODBC.INI in NT system directory (e.g., C:\winnt35).
     */

    uid = 0;
    ini_file[0] = 0;

    p = (char *)AxStrFromArray(fData, 1);
    if ( p && p[0] )
        strcpy(ini_file, p);
    datasources_only = AxBoolFromArray(fData, 2);

    if ( !ini_file[0] ) {
        envname = NXLT("ODBCINI");
        envlen = strlen(envname);
        /* first get ODBCINI from current env */
        p = getenv(envname);
        if ( p && p[0] )
            strcpy(ini_file, p);

        /* get ODBCINI env from axnet.cnf
         * (this will override ODBCINI env set in shell) */
        axnetGetConfig(); /* get latest version of axnet.cnf */
        if ( axnetConfig ) {
            for ( ix = 0; ix < axnetConfig->linecount; ix++ ) {
                lineP = &axnetConfig->line[ix];
                if ( !lineP->arg1 || (strcmp(lineP->arg1, NXLT("elfodbc")) != 0) )
                    continue; /* not relevent */
                if ( !lineP->arg2 )
                    continue;
                switch ( lineP->command ) {
                case SETENV_FOR:
                    p = strchr(lineP->arg2, '=');
                    if ( p ) {
                        if ( strncmp(lineP->arg2, envname, 
                                            (envlen * sizeof(char))) == 0 ) {
                            /* Driver Manager will use this .odbc.ini */
                            strcpy(ini_file, p+1);
                        }
                    }
                    break;
                case UID_FOR:
                    /* this is the uid for this executable to run under */
                    uid = atoi(lineP->arg2);
                    break;
                default:
                    break;
                }
            }
        }
    }

    if ( !ini_file[0] ) {
#if AX_X11                   /*  {  */
        if ( uid > 0 ) {
            /* elfodbc will be launched as user with this userid;
             * get .odbc.ini from this user's home directory */
            pw_entry = getpwuid(uid);
            if ( pw_entry == (struct passwd *)NULL ) {
                sprintf(errstr, NXLT("%d"), uid);
                AxError(99, XLT("No such uid"), errstr);
            }
        }
        else {
            /* get .odbc.ini from passed user's home directory */
            username = (char *)AxStrFromArray(fData, 0);
            pw_entry = getpwnam(username);
            if ( pw_entry == (struct passwd *)NULL )
                AxError(99, XLT("No such user"), username);
        }
        sprintf(ini_file, NXLT("%s/.odbc.ini"), pw_entry->pw_dir);
#else                            /*  } AXMS {  */
        size = MS_GetSystemDirectory(filename, MAX_PATH);
        if ( !size )  /* failed */
            AxError(99, XLT("Error getting system directory"), (char *)NULL);
        sprintf(ini_file, "%s\\..\\odbc.ini", filename);
#endif                           /*  }  */
    }

    fp = fopen(ini_file, NXLT("r"));
    if ( !fp ) {
        sprintf(errstr, XLT("Cannot read file \"%s\""), ini_file);
        AxError(errno, errstr, NXLT("axnet"));
    }

    ix = 0;
    started_datasources = FALSE;
    retData = AxMakeArray(0);

    /* Read the file one line at a time */

    while ( fgets(lineBuf, LINEBUFSIZ, fp) ) {
        len = strlen(lineBuf);
        if ( (len > 0) && (lineBuf[len-1] == '\n') )
            lineBuf[len-1] = 0;
        if ( datasources_only ) {
            if ( lineBuf[0] == '[' ) {
                if ( started_datasources )
                    break;
#if AX_X11
                if ( strcmp(NXLT("[ODBC Data Sources]"), lineBuf) == 0 ) 
#else
                if ( strcmp(NXLT("[ODBC 32 bit Data Sources]"), lineBuf) == 0 )
#endif
                    {
                    started_datasources = TRUE;
                    continue;  /* skip this line */
                    }
            }
            if ( !started_datasources )
                continue;
            if ( !lineBuf[0] )
                continue;  /* skip empty line when reading datasources */
        }
        retData = AxAddStrToArray(retData, ix++, lineBuf);
    }

    fclose(fp);
    return(retData);
#endif                               /*  }  */
}



void axnetSetLib(WhereIAm)
char *WhereIAm;
{
    char *newlibpath, *p;
    char here[STRPATH+100];
 
#if HP800V3
#define VARNAME "SHLIB_PATH"
#else
#if RIOS
#define VARNAME "LIBPATH"
#else
#define VARNAME "LD_LIBRARY_PATH"
#endif
#endif
 
    p = getenv(VARNAME);
    sprintf(here,"%s=%s/axshlib:%s",VARNAME,WhereIAm,WhereIAm);
    if ( p ) {
        newlibpath = AxMalloc(strlen(here)+1 + strlen(p)+1);
        sprintf(newlibpath,"%s:%s",here,p);
    }
    else {
        newlibpath = AxMalloc(strlen(here)+1);
        strcpy(newlibpath,here);
    }
    putenv(newlibpath);
}





#if AXMS          /*  {  */

int 
axnetGetExecLoc(char *ExecLocation)
{
    int ret;
    char *p;
    int retry = 0;
    HMODULE hxll;

    *ExecLocation = '\0';
    hxll = MS_GetModuleHandle( NULL );
    if ( hxll == (HMODULE)NULL )
       return( 0 );

    ret = MS_GetModuleFileName( hxll, ExecLocation, MAX_PATH );
    if ( ret <= 0 )
       /* failure fallback */
       return( 0 );

    if ( retry > 1 )
       /* could not fnd a path */
       return( 0 );

    p = ExecLocation + strlen(ExecLocation) - 1;
    if ( strlen(ExecLocation) > 4 )
       {
       if ( *(p-3) == '.' )
          *(p-3) = '\0';      /* clip off any ".EXE" or .DLL */
       }

    while ( (p >=ExecLocation) && (*p != '\\') )
       { p--; }

    /* step on last slash to leave just the location as long as this is */
    /* not the root                                                     */

    /* clip off the executable name to leave the install directory */
    if ( *(p-1) == ':' )
       /* exec is at the root level */
       *(p+1) = '\0';       /* leave "D:\" */
    else
       *p = '\0';           /* leave D:\xxx\yyy   */
    return( TRUE );
}



void StartWinProcAsUser( char *command_line, 
                         char *proc_as_user, 
                         char *encrypted_password )
{
    int error_code;
    char local_password[30];
    HANDLE user_token;
    PVOID err_str;
    DWORD buf_len;
    int ret;
    char *error_reason;
    PROCESS_INFORMATION process_info;
    STARTUPINFO startup;
    char *p;
    int simple_proc;
    char running_as_user[30];

    /* if the passed username matches the name we are running as, then */
    /* this copy of axnet was started by the user, and we can do a simple */
    /* "CreateProcess" without needing a password, rather than the more  */
    /* complex case where we were started as a system service, and will  */
    /* need a real password.                                             */
    simple_proc = FALSE;
    ret = sizeof(running_as_user) - 1;
    running_as_user[0] = '\0';
    MS_GetUserName( running_as_user, &ret );

    local_password[0] = '\0';
    if ( STRIEQ( running_as_user, proc_as_user ) )
       {
       /* we are starting a process for the same username under which */
       /* we are running, so we will not require the password         */
       simple_proc = TRUE;
       }
    else
       {
       /* On NT we expect to be passed the user's password as part of the */
       /* message, since it is required to create a sub-process as that user */
       if ( (encrypted_password != (char *)NULL) && (*encrypted_password != '\0') )
           GetClearPassword( encrypted_password, local_password );
       else
           local_password[0] = '\0';
       }

    NETLOG( ("runing as user <%s>   proc as user <%s>   clear passwd = <%s>\n",
           running_as_user, proc_as_user, local_password)  );

    memset(&startup, 0, sizeof(STARTUPINFO)); 
    startup.cb = sizeof(STARTUPINFO);

    memset(&process_info, 0, sizeof(PROCESS_INFORMATION));

    /* set the wierd Windows global so things get passed along */
    _fileinfo = TRUE;

    ret = 0;
    if ( simple_proc )
        {
        /* we are running as the user, so we do not need to impersonate */
        /*  another username                                            */
        NETLOG( ("Doing simple proc\n") );
        ret = MS_CreateProcess( NULL, command_line,
                           NULL,     /* process security attributes */
                           NULL,     /* thread security attributes */
                           TRUE,     /* process inherits handles   */
                           (DWORD)0, /* creation flags */
                           (LPVOID)0, /* address of new environment block */
                           (LPCSTR)NULL, /* address of current directory */
                           &startup,
                           &process_info );
        user_token = NULL;
        }
    else
        {
        /* we need to start the sub-process as another user.  First we */
        /* get a handle to that user's context, then we start the process */
        NETLOG( ("Procing as user\n") );
        ret = MSNT_LogonUser(proc_as_user,
	                         (LPTSTR)NULL, /* search local account database,
	                                        * then trusted domains.
	                                        */
	                         local_password, 
	                         LOGON32_LOGON_INTERACTIVE,
	                         LOGON32_PROVIDER_DEFAULT,
                             &user_token);
        if ( !ret )
           {
           error_code = MS_GetLastError();
           error_reason = NXLT("LogonUser in axnet");
           goto post_error;
           }

        /* turns out we need to specify this window station and desktop
         * if the process we create makes any USER32 calls.  But only
         * administrator has access to it, by default.  So we need to
         * give the user the appropriate rights. (see winsecur.c)   -donato
         */
        startup.lpDesktop = "WinSta0\\Default";
        axmsGrantDefaultWindowStationAccessToUser(proc_as_user);

        ret = MSNT_CreateProcessAsUser( user_token,
                           NULL, command_line,
                           NULL,     /* process security attributes */
                           NULL,     /* thread security attributes */
                           TRUE,     /* process inherits handles   */
                           (DWORD)0, /* creation flags */
                           (LPVOID)0, /* address of new environment block */
                           (LPCSTR)NULL, /* address of current directory */
                           &startup,
                           &process_info );
        }

    if ( !ret )
         {
         /* error */
         error_code = MS_GetLastError();
         error_reason = NXLT("CreateProcess in axnet");
         }
 
post_error:
    if ( !ret )
        {
        /* make sure we print the error message to stderr. */
        buf_len = MS_FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
                             FORMAT_MESSAGE_FROM_SYSTEM,
                             NULL, error_code,
                             0, /* let system decide language! */
                            (LPTSTR)&err_str, 0, NULL);
        p = (char *)err_str;
        p = p  + strlen(p);
        while ( (*p < ' ') && (p >= (char *)err_str) )
          {
          *p = '\0';     /* strip CR/NL put on by Microsoft function */
          p--;
          }
        AxError(error_code,err_str,error_reason);
 
        /* need to free the string allocated by FormatMessage */
        MS_LocalFree(err_str);
        }
    else
        {
        /* successfully started process. we don't need these handles. */
        if ( user_token )
            MS_CloseHandle(user_token);
        MS_CloseHandle(process_info.hThread);
        MS_CloseHandle(process_info.hProcess);
        }

    /* parent closes channel and continues */
    AxCloseChannel();
    return;
}


/* struct to pass standard argc/argv to threads. */
typedef struct thread_args_t {
    int argc;
    char **argv;
} thread_args_t;

DWORD WINAPI
AxStartServiceThread(void *argsP)
{
    thread_args_t *thread_args = (thread_args_t*)argsP;
    elfmain(thread_args->argc, thread_args->argv);
    return 0;
}


char *TrueMalloc(size)
int size;
{
    return( AxMalloc(size) );
}

char *TrueRealloc(p, size)
void *p;
int size;
{
    if (p == (char *) NULL)
        return ( AxMalloc(size) );
    else
        return( AxRealloc( p, size ) );
}


DWORD WINAPI
RemoteConnectionThread(void *argsP)
{
    int channel, serverChannel, clientChannel;
    int code;
    char *str, *obj;
    elfData rP;
    fd_set read_fds, wtfds, nofds;
    int events;
    int32 nsize;
    char *inBuf, *outBuf;
    int processed, written;
    int size, cnt;
    unsigned long lsize;
    int socknum;
    char *host, *socket, *service_name;
    remote_conn_t *thread_args = (remote_conn_t*)argsP;


/* Connect to target service here on this machine */

    clientChannel = thread_args->clientChannel;
    socknum = thread_args->socknum;
    service_name = thread_args->service_name;
    if ( socknum == 0 )
        {
        socknum = atoi(thread_args->socket);
        socket = NULL;
        }
    else
        {
        socket = thread_args->socket;
        }
    if ( thread_args->host )
        host = thread_args->host;
    else
        host = NULL;

    channel = AxClientConnect(socknum, socket, host, &code, &str, &obj);
    if ( !host && ((code == WSAECONNREFUSED) || (code == ENOENT)) )
        {
        int axnetChannel;   /* axnet may have listed a server that's gone */
        elfData response, request;

        /* tell axnet to drop service if was there */
        axnetChannel = AxNetConnect(NULL,&code,&str,&obj);
        AxNetDropServiceBySocket(axnetChannel, socket, &code, &str, &obj);
        if ( code != 0 )
            {
            AxClientDisConnect(axnetChannel);
            AxError(code,str,obj);
            }
        request = AxMakeArray(2);
        request = AxAddStrToArray(request, 0, service_name);
        AxSendMessageToServer(axnetChannel,AXNET_FIND_SERVICE,request,&code, &str, &obj);
        if ( code != 0 )
            {
            AxClientDisConnect(axnetChannel);
            AxError(code,str,obj);
            }
        response = AxGetResponseFromServer(axnetChannel,&code,&str,&obj);
        AxClientDisConnect(axnetChannel);
        if ( code != 0 )
            AxError(code,str,obj);
        channel = AxClientConnect(socknum,AxStrFromDataPtr(response),NULL,&code,&str,&obj);
        }
    if ( code != 0 )
        AxError(code,str,NXLT("remote connection to service"));

    serverChannel = channel;

/* Respond to client that it should continue to use its current channel */
    rP = AxMakeArray(2);
    AxAddIntToArray(rP,0,0);
    AxAddArrayToArray(rP,1,NULL);
    size = AxMemWrite(rP,0,&outBuf);
    cnt = send(clientChannel,outBuf,size,0);
    if ( cnt != size )
        return 0;    /* implicitly exits the thread. */

    FD_ZERO(&read_fds);
    FD_ZERO(&wtfds);
    FD_ZERO(&nofds);

/*
    Read from clientChannel (remote axmain),
    write to serverChannel (local axmain).
    Read from serverChannel, write to clientChannel.
*/

    while (TRUE)
        {
        FD_SET(clientChannel,&read_fds);
        FD_SET(serverChannel,&read_fds);
        events = select(FD_SETSIZE, &read_fds, &wtfds, &nofds, NULL);
        if ( events <= 0 )  /* error occurred. */
            break;

        if ( FD_ISSET(clientChannel,&read_fds) )
            {
            processed=0;
rmore0:     cnt = recv(clientChannel, (char*)&nsize+processed,
                                    sizeof(int32)-processed, 0);
            if ( cnt <= 0 )
                break;    /* some error, or client went away. */
            if ( cnt != sizeof(int32)-processed )
                {
                processed += cnt;
                goto rmore0;    /* need to read more. */
                }

            lsize = ntohl(nsize);
            inBuf = (char*)AxMalloc(lsize);
            processed = 0;
rmore1:     cnt = recv(clientChannel,inBuf+processed,lsize-processed,0);
            if ( cnt <= 0 )
                break;  /* some error */
            if ( cnt != lsize-processed )
                {
                processed += cnt;
                goto rmore1;
                }

            written = 0;
wmore0:     cnt = send(serverChannel, (char*)&nsize+written,
                                    sizeof(int32)-written, 0);
            if ( cnt <= 0 )
                break;    /* some error */
            if ( cnt != sizeof(int32)-written )
                {
                written += cnt;
                goto wmore0;    /* need to write more. */
                }

            written = 0;
wmore1:     cnt = send(serverChannel,inBuf+written,lsize-written,0);
            if ( cnt <= 0 )
                break;    /* some error */
            if ( cnt != lsize-written )
                {
                written += cnt;
                goto wmore1;
                }

            AxFree(inBuf);
            }


        if ( FD_ISSET(serverChannel,&read_fds) )
            {
            processed = 0;
rmore2:     cnt = recv(serverChannel, (char*)&nsize+processed,
                                    sizeof(int32)-processed, 0);
            if ( cnt <= 0 )
                break;    /* some error */
            if ( cnt != sizeof(int32)-processed )
                {
                processed += cnt;
                goto rmore2;
                }

            lsize = ntohl(nsize);
            inBuf = (char*)AxMalloc(lsize);
            processed = 0;
rmore3:      cnt = recv(serverChannel, inBuf+processed, lsize-processed, 0);
            if ( cnt <= 0 )
                break;
            if ( cnt != lsize-processed )
                {
                processed += cnt;
                goto rmore3;
                }

            written = 0;
wmore2:      cnt = send(clientChannel, (char*)&nsize+written,
                                    sizeof(int32)-written, 0);
            if ( cnt <= 0 )
                break;    /* some error */
            if ( cnt != sizeof(int32)-written )
                {
                written += cnt;
                goto wmore2;
                }

            written = 0;
wmore3:      cnt = send(clientChannel, inBuf+written, lsize-written, 0);
            if ( cnt <= 0 )
                break;    /* some error */
            if ( cnt != lsize-written )
                {
                written += cnt;
                goto wmore3;
                }

            AxFree(inBuf);
            }
        }

    /* cleanup and exit. */
    if ( host )
        free(thread_args->host);
    free(thread_args->socket);
    free(thread_args->service_name);
    free(thread_args);

    shutdown(serverChannel,2);
    shutdown(clientChannel,2);
    closesocket(serverChannel);
    closesocket(clientChannel);
    ExitThread(0);    /* explicitly exit the thread. */
    return 0;
}

/***
    This function creates a socket and binds to it, then returns the port
    number. This is used to deal with Unix named pipes on NT. The "named
    pipe" becomes a string holding a port number (see SelectSocketName()).
***/
static int
axmsCreateAndBindToSocket(void)
{
    int ret, size, socket_handle;
    SOCKET sockH;
    struct sockaddr_in in_serv_addr; 
    int trueVal = TRUE;

    /* create the socket handle */
    sockH = socket(PF_INET, SOCK_STREAM, 0);
    if ( sockH == INVALID_SOCKET )
        {
        ret = WSAGetLastError();
        return 0;
        }

    /* Allow the child to bind to this socket, since we may still */
    /* appear to be using it. THis must be done BEFORE bind call  */
    ret = setsockopt(sockH, SOL_SOCKET, SO_REUSEADDR,
               (char *)(&trueVal), sizeof(trueVal));
    if ( ret == SOCKET_ERROR )
        return 0;

    /* bind to it to get a port number */
    memset(&in_serv_addr, 0, sizeof(in_serv_addr));
    in_serv_addr.sin_family = AF_INET;
    in_serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    in_serv_addr.sin_port = 0;
    ret = bind(sockH, (struct sockaddr*)&in_serv_addr, sizeof(in_serv_addr));
    if ( ret == SOCKET_ERROR )
        {
        ret = WSAGetLastError();
        return 0;
        }

    size = sizeof(in_serv_addr);
    ret = getsockname(sockH, (struct sockaddr*)&in_serv_addr, &size);
    if ( ret == SOCKET_ERROR )
        {
        ret = WSAGetLastError();
        return 0;
        }

    socket_handle = (int)ntohs(in_serv_addr.sin_port);
    /* close the original socket */
    closesocket(sockH);
    return socket_handle;
}

#endif      /* } AXMS  */

