/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: winsecur.c
   Applix ASC version: 5
   Date: Mon Nov 16 13:53:27 1998
*/
#include <stdio.h>
#include <windows.h>
#include <winmap.h>


static FILE *seclogfp = NULL;
#if WIN_DEBUG
static int axmsSecurityLogging = TRUE;
#else
static int axmsSecurityLogging = FALSE;
#endif
#define SECLOG(m) axmsSecurityLog m


/********************* move these to winmap.h !!! ***************************/
#define MS_GetThreadDesktop                 GetThreadDesktop
#define MS_SetThreadDesktop                 SetThreadDesktop
#define MS_GetCurrentThreadId               GetCurrentThreadId

#define MSNT_LookupAccountName              LookupAccountName
#define MSNT_OpenWindowStation              OpenWindowStation
#define MSNT_GetProcessWindowStation        GetProcessWindowStation
#define MSNT_SetProcessWindowStation        SetProcessWindowStation
#define MSNT_OpenDesktop                    OpenDesktop
#define MSNT_CloseDesktop                   CloseDesktop
#define MSNT_CloseWindowStation             CloseWindowStation

#define MSNT_GetUserObjectSecurity          GetUserObjectSecurity
#define MSNT_InitializeSecurityDescriptor   InitializeSecurityDescriptor
#define MSNT_GetSecurityDescriptorDacl      GetSecurityDescriptorDacl
#define MSNT_GetAclInformation              GetAclInformation
#define MSNT_InitializeAcl                  InitializeAcl
#define MSNT_GetAce                         GetAce
#define MSNT_AddAce                         AddAce
#define MSNT_EqualSid                       EqualSid
#define MSNT_AddAccessAllowedAce            AddAccessAllowedAce
#define MSNT_SetSecurityDescriptorDacl      SetSecurityDescriptorDacl
#define MSNT_SetUserObjectSecurity          SetUserObjectSecurity
/****************************************************************************/



/*****************************************************************************
function: axmsSecurityLog - Log axnet messages to "/tmp/winsecur.log"
*/
static void
axmsSecurityLog(char *fmt, ...)
{
    va_list args;
    char tmp_path[MAX_PATH+1];

    if ( !axmsSecurityLogging )
        return;

    if ( !seclogfp )
        {
        MS_GetTempPath(MAX_PATH+1, tmp_path);
        strcat(tmp_path, "winsecur.log");
        seclogfp = fopen(tmp_path, "w");
        }

    if ( !seclogfp )
        return;

    va_start(args, fmt);
    vfprintf(seclogfp, fmt, args);
    va_end(args);
    fflush(seclogfp);
}
/****************************************************************************/



void
axmsGrantAccessToUserObject(PSID user_sidP, HANDLE hObj)
{
    PSECURITY_DESCRIPTOR    objSD;
    DWORD                   dwSDLengthNeeded;
    PACL                    pACL;
    BOOL                    bDaclPresent;
    BOOL                    bDaclDefaulted;
    ACL_SIZE_INFORMATION    AclInfo;
    PACL                    pNewACL;
    DWORD                   dwNewACLSize;
    PSECURITY_DESCRIPTOR    newSD;
    PVOID                   pTempAce;
    UINT                    CurrentAceIndex;
    SECURITY_INFORMATION    sec_info;
    DWORD                   sd_size;
    BOOL                    ret;

    objSD = newSD = NULL;
    pNewACL = NULL;

    /* 1:  Get the security descriptor of the object we are granting
     *     access to.
     */
    sd_size = 0;
    sec_info = (SECURITY_INFORMATION)DACL_SECURITY_INFORMATION;
    /* first call to get the SD size. */
    ret = MSNT_GetUserObjectSecurity(hObj, &sec_info, objSD,
                                    sd_size,
                                    (LPDWORD)&dwSDLengthNeeded);
    if ( dwSDLengthNeeded > 0 )
        {
        /* now really get it */
        objSD = (PSECURITY_DESCRIPTOR)malloc(dwSDLengthNeeded);
        ret = MSNT_GetUserObjectSecurity(hObj, &sec_info, objSD,
                                        dwSDLengthNeeded,
                                        (LPDWORD)&dwSDLengthNeeded);
        }
    if ( !ret )
        {
        SECLOG(("Error %d:GetUserObjectSecurity\n", (int)MS_GetLastError()));
        goto cleanup;
        }

    /* 2:  Get the discretionary ACL for the object. */
    if ( !MSNT_GetSecurityDescriptorDacl(objSD, &bDaclPresent, &pACL,
                                                    &bDaclDefaulted) )
        {
        SECLOG(("Error %d:GetSecurityDescriptorDacl\n",(int)MS_GetLastError()));
        goto cleanup;
        }

    /* 3:  Get object ACL size information */
    if ( !MSNT_GetAclInformation(pACL, &AclInfo, sizeof(ACL_SIZE_INFORMATION),
                                                        AclSizeInformation) )
        {
        SECLOG(("Error %d:GetAclInformation\n",MS_GetLastError()));
        goto cleanup;
        }

    /* 4:  Compute size needed for the new ACL */
    dwNewACLSize = AclInfo.AclBytesInUse +
                  sizeof(ACCESS_ALLOWED_ACE) +
                  GetLengthSid(user_sidP) - sizeof(DWORD);

    /* 5:  Allocate memory for new ACL */
    pNewACL = (PACL)malloc(dwNewACLSize);

    /* 6:  Initialize the new ACL */
    if ( !MSNT_InitializeAcl(pNewACL, dwNewACLSize, ACL_REVISION) )
        {
        SECLOG(("Error %d:InitializeAcl\n",(int)MS_GetLastError()));
        goto cleanup;
        }

    /* 7:  If DACL is present, copy it to a new DACL */
    if ( bDaclPresent )
        {
        /*  7A:  Copy the object's ACEs to our new ACL */
        if ( AclInfo.AceCount )
            {
            for ( CurrentAceIndex = 0; CurrentAceIndex < AclInfo.AceCount;
                                                        CurrentAceIndex++ )
                {
                if ( !MSNT_GetAce(pACL,CurrentAceIndex,&pTempAce) )
                    {
                    SECLOG(("Error %d: GetAce\n",(int)MS_GetLastError()));
                    goto cleanup;
                    }

                if ( MSNT_EqualSid(
                        (PSID)&((ACCESS_ALLOWED_ACE*)pTempAce)->SidStart,
                         user_sidP) )
                    {
                    SECLOG(("User already has rights. Continuing...\n"));
					continue;
                    }

                /* Add the ACE to the new ACL */
                if ( !MSNT_AddAce(pNewACL, ACL_REVISION, MAXDWORD, pTempAce,
                                        ((PACE_HEADER)pTempAce)->AceSize) )
                    {
                    SECLOG(("Error %d:AddAce\n",(int)MS_GetLastError()));
                    goto cleanup;
                    }
                }
            }
        }

    /* 8:  Add the access-allowed ACE to the new DACL */
    if ( !MSNT_AddAccessAllowedAce(pNewACL, ACL_REVISION,
                SPECIFIC_RIGHTS_ALL | STANDARD_RIGHTS_ALL |
                GENERIC_ALL | GENERIC_EXECUTE | GENERIC_WRITE | GENERIC_READ,
                user_sidP) )
        {
        SECLOG(("Error %d:AddAccessAllowedAce",(int)MS_GetLastError()));
        goto cleanup;
        }

    /* 9:  Initialize a new object SD */
    newSD = (PSECURITY_DESCRIPTOR)malloc(
								SECURITY_DESCRIPTOR_MIN_LENGTH + dwNewACLSize);
    if ( !MSNT_InitializeSecurityDescriptor(newSD,
                            SECURITY_DESCRIPTOR_REVISION) )
        {
        SECLOG(("Error %d:InitializeSecurityDescriptor\n",(int)MS_GetLastError()));
        goto cleanup;
        }

    /* 10:  Set our new DACL to the new object SD */
    if ( !MSNT_SetSecurityDescriptorDacl(newSD, TRUE, pNewACL, FALSE) )
        {
        SECLOG(("Error %d:SetSecurityDescriptorDacl",(int)MS_GetLastError()));
        goto cleanup;
        }

    /* 11:  Set the new SD to the user object */
    sec_info = (SECURITY_INFORMATION)DACL_SECURITY_INFORMATION;
    if ( !MSNT_SetUserObjectSecurity(hObj, &sec_info, newSD) )
        {
        SECLOG(("Error %d:SetUserObjectSecurity\n",(int)MS_GetLastError()));
        goto cleanup;
        }

cleanup:
    if ( objSD )
        free(objSD);
    if ( newSD )
        free(newSD);
    if ( pNewACL )
        free(pNewACL);
    return;
}


void
axmsGrantDefaultWindowStationAccessToUser(char *user)
{
    PSID    user_sidP;
    UCHAR   psnuType[2048];
    UCHAR   lpszDomain[2048];
    DWORD   dwDomainLength = 250;
    DWORD   dwSIDBufSize=1024;
    DWORD   err;
    HANDLE  hObj;
    HANDLE  hWinSta=NULL, hDesk=NULL;
    HANDLE  hWinStaSave, hDeskSave;

    user_sidP = (PSID)malloc(1024);
    while ( !MSNT_LookupAccountName((LPSTR)NULL, user, user_sidP,
                                &dwSIDBufSize, lpszDomain,
                                &dwDomainLength, (PSID_NAME_USE)psnuType) )
        {
        err = MS_GetLastError();
        if ( err == ERROR_INSUFFICIENT_BUFFER )
            user_sidP = (PSID)realloc(user_sidP, dwSIDBufSize);
        else
            {
            SECLOG(("Error %d:LookupAccountName\n",(int)err));
            free(user_sidP);
            return;
            }
        }

    /* the following USER32 call is needed just to make the connection to
     * the window station & desktop.
     */
    MS_GetDesktopWindow();
    /* save the window station & desktop of the service. */
    hWinStaSave = MSNT_GetProcessWindowStation();
    hDeskSave = MS_GetThreadDesktop(MS_GetCurrentThreadId());

    /* Set the "WinSta0\Default" window station for this process.
     * We do this because we can't pass the handle we get from
     * OpenWindowStation to GetUserObjectSecurity.  We must instead
     * use the handle from GetProcessWindowStation, which by default
     * won't be "WinSta0", so we go through this open/set/get business.
     */
    hWinSta = MSNT_OpenWindowStation("WinSta0", FALSE, MAXIMUM_ALLOWED);
    if ( hWinSta )
        {
        if ( !MSNT_SetProcessWindowStation(hWinSta) )
            SECLOG(("Could not set WinSta,err=%d\n",(int)MS_GetLastError()));
        }
    else
        SECLOG(("Could not open windowstation,err=%d\n",(int)MS_GetLastError()));

    hDesk = MSNT_OpenDesktop("Default", 0, FALSE, MAXIMUM_ALLOWED);
    if ( hDesk )
        {
        if ( !MS_SetThreadDesktop(hDesk) )
            SECLOG(("Could not set desktop,err=%d\n",(int)MS_GetLastError()));
        }
    else
        SECLOG(("Could not open desktop,err=%d\n",(int)MS_GetLastError()));

    /* we can now retrieve the newly set window station & desktop
     * and grant the user access to them.
     */
    hObj = MSNT_GetProcessWindowStation();
    axmsGrantAccessToUserObject(user_sidP, hObj);

    hObj = MS_GetThreadDesktop(GetCurrentThreadId());
    axmsGrantAccessToUserObject(user_sidP, hObj);

    /* restore window station & desktop */
    MS_SetThreadDesktop(hDeskSave);
    MSNT_SetProcessWindowStation(hWinStaSave);
    MSNT_CloseDesktop(hDesk);
    MSNT_CloseWindowStation(hWinSta);
    free(user_sidP);
}

