/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: elbspell.c
   Applix ASC version: 11
   Date: Mon Mar 22 21:14:09 1999
*/
#include <stdlib.h>
#include <elbltn.h>
#include <aspeller.h>       /* Combined Speller/Proximity include file */
#include <getstr7.h>    /* Need for Speller error codes */
#include <wchars.h>

static char NameOfThisSourceFile[] = __FILE__ ; /* for the assert code */

#ifdef ASSERT
#undef ASSERT
#endif
#if !TURBO_BUILD            /* i.e. if building daxdata... */
#define ASSERT(x) if ( !(x) ) WPAsserter(NameOfThisSourceFile,__LINE__)
#else
#define ASSERT(x)
#endif
#ifdef VERIFY
#undef VERIFY
#endif
#define VERIFY(x) if ( !(x) ) WPAsserter(NameOfThisSourceFile,__LINE__)

#ifndef MBstrcmp
#define MBstrcmp strcmp
#endif

#define ESCAPE_CHAR '\\' /* to escape hyphenation points in user dict words */
#define HYPHEN_POINT_CHAR '-' /* use "\-" to indicate hyphen points in user
                                dictionary word */

#define MAXNDICTS   8       /*this must match number in elbspell.c*/
			  /* Maximum number of dictionary files allowed */
#define NODICT      -1      /* Invalid dictionary index */

typedef struct {
    char      phonFile[STRPATH];      /* Language env file (xxphon.env) */
    char      clxFile[STRPATH];       /* Corelex file (.clx) */
    char      lexFile[STRPATH];       /* Linguibase file (.uh) */
    char      detFile[STRPATH];       /* "Detect" file (.env) */
    char      hyFile[STRPATH];        /* Hyphenation data file (.hy) */
    char      thesFile[STRPATH];      /* Thesaurus file (.syn) */
} ed_spell_files;

typedef struct {
    char          file[STRPATH];      /* Name of file */
    int             code;               /* Code # associated w/dictionary */
    bool            write;              /* TRUE -> we can alter this file */
    CLAM            *clamptr;           /* Proximity pointer */
} ed_dict;

/* State variables */
static bool         started=FALSE;      /* TRUE -> Proximity is up & running */
static int          currLang=NOLANG;    /* Code for current language */
static char         currLangName[32];   /* Name of current language */
static bool         thesOpen=FALSE;     /* TRUE -> thesaurus is open */
static bool         doingHyphen=FALSE;  /* TRUE -> doing hyphenation */
static int          nDictFiles=0;       /* Number of open dictionary files */

/* Essential Proximity pointers */
static CACHE        *stopCache=NULL;        /* Proximity bad-word cache */
static CACHE        *normalCache=NULL;      /* Proximity good-word cache */
static CORELEX      *corelexPtr=NULL;       /* Proximity Core Lexicon pointer */
static LEX          *lexPtr=NULL;           /* Proximity Linguibase pointer */
static LEX          *thesPtr=NULL;          /* Proximity Thesaurus pointer */

/* Filenames for Proximity files */
static ed_spell_files   spellFiles;         /* All files except for dicts */
static ed_dict      dictFile[MAXNDICTS];    /* Dictionary ("clam") files and
                                                related state variables */

/* Some Proximity functions */
#if 0
extern int Prox_Quflags();
extern int Prox_Qulen();
extern int Prox_Quwlist();
extern int Prox_Quwflags();
extern UCHAR *Prox_Quword();
extern UCHAR *Prox_Quinfo();
extern UCHAR *Prox_Qulookup();
extern char *prx_langprf();
#endif

/* Stuff for Speller */
extern int CSpellPid;       /* pid of speller, if it has been sucessfully
                 * proc-ed */

axstatic short ElfSpellChan = -1;   /* communication channel to speller */
static unsigned int spinbits();
/* Error conditions */
char        SpErrorObject[STRPATH+1];       /* Object error relates to */
char        SpErrorOverrideStr[MAX_PROX_WORD+STRLONG+1];  /* Override error text */

/* Translation tables: Aster*x-to-Proximity and Proximity-to-Aster*x */
char ProxToAx[256] = PROX2AX1 PROX2AX2 PROX2AX3 PROX2AX4;
char AxToProx[256] = AX2PROX1 AX2PROX2 AX2PROX3 AX2PROX4;

/* Private functions */
static void     spGetLangName AXP_((int full, int langCode, char *langName));
static typeErr  spSpellerStart AXP_((int language, char **dictionaries));
static typeErr  spGetThesaurusMeanings AXP_((mbhar *word,int start,int numWant, int *numGotP,mbchar **meanings));
static typeErr  spGetThesaurusWords AXP_((int meaningCode,char **wordArray[MAXLISTS]));
static typeErr  spOpenThesaurus AXP_((int language));
static mbchar   *spSplitWord AXP_((mbchar *word,bool proxWord,mbchar **leftOvers));
static typeErr  spAddDictionaries AXP_((char **dictionaries));
static typeErr  spAddDictionary AXP_((char *dfile,int code));
static typeErr  spRemoveAllDictionaries AXP_((void));
static int      spDictListIndex AXP_((char *dfile));
static void     spXlateString  AXP_(( mbchar *string, mbchar *table));
static void     spRemoveSlashes AXP_((char *str));
static void     spSpellFilenamesAvailable AXP_((int language,ed_spell_files      *files));
static void     spSpellFilenames AXP_((int language,char *prefix,              ed_spell_files *files));
static void     spSpellFilenamesInit AXP_((ed_spell_files *files));
static typeErr  spAccessFile AXP_((char *file,int accessFlag));
static mbchar   *spEscapeHyphPoints AXP_((mbchar *word));
static typeErr  spUnescapeHyphPoints AXP_((char *outword, char *inword));

/* Extern for a Proximity fcn */
extern UCHAR    **correct( );



/*** Notes:

    - Callers to these routines should include aspeller.h. Note this will
        #include the necessary Proximity include files. These files contain
        certain #defines that the user will need.

    - Initial code does not use caches (in-memory lists of properly and
        improperly-spelled words). These can be added for better performance.

    - All Proximity functions are commented with "Px".

***/

/************************** Speller misc queries *********************** {*/

/*-----------------------------------------------------------------------
    Clear out the error-related strings. This should be the first thing
    in any public function.
------------------------------------------------------------------------*/
void     
spClearErrors ( )
{
    MBstrcpy( SpErrorObject, "");
    MBstrcpy( SpErrorOverrideStr, "");
}

/*-----------------------------------------------------------------------
    Return whether the necessary Proximity files exist for a given
    language. If not, the speller will not come up properly in that
    language.

    This function only checks for the essential files. Hyphenation and
    Thesaurus files (as well as all dictionary files) are not considered
    essential.
------------------------------------------------------------------------*/
bool            
SpIsLanguageOK ( language )
int                 language;           /* Desired language (use Prox code) */
{
    ed_spell_files  files;              /* Structure to hold filenames */

    /* Verify requested language is acceptable */
    if ((language < ENGLISH) || (language > MAXLANG)) {
        ASSERT( 0);         /* Shouldn't happen during development */
        return( FALSE);     /* Unknown languages aren't supported */
    }

    spSpellFilenamesAvailable( language, &files);
    return( *files.phonFile && *files.clxFile && *files.lexFile && 
                                                        *files.detFile);
}

/*-----------------------------------------------------------------------
    Return current language code (or NOLANG if speller is not up)
------------------------------------------------------------------------*/
int             
SpGetCurrentLanguage ( )
{
    if (!started) return( NOLANG);
    ASSERT( currLang == getsclang());       /* Development sanity check */
    return( currLang);
}

/*-----------------------------------------------------------------------
    Given the name of the language (e.g. "English"), return the proper
    Proximity code.

    The legal Proximity codes run from 1 to MAXLANG (16 as of 3/92).
    The test is case insensitive. NOLANG is returned if a bad name is
    passed.
------------------------------------------------------------------------*/
int             
SpGetLanguageCode ( langStr )
char              *langStr;               /* Name of language */
{
    /* !!0 Use wpstrcasecmp() when available */
    if (!strcasecmp( langStr, XLT("english"))) return ENGLISH;
    if (!strcasecmp( langStr, XLT("german"))) return GERMAN;
    if (!strcasecmp( langStr, XLT("french"))) return FRENCH;
    if (!strcasecmp( langStr, XLT("spanish"))) return SPANISH;
    if (!strcasecmp( langStr, XLT("italian"))) return ITALIAN;
    if (!strcasecmp( langStr, XLT("british"))) return BRITISH;
    if (!strcasecmp( langStr, XLT("swedish"))) return SWEDISH;
    if (!strcasecmp( langStr, XLT("danish"))) return DANISH;
    if (!strcasecmp( langStr, XLT("norwegian"))) return NORWEGIAN;
    if (!strcasecmp( langStr, XLT("dutch"))) return DUTCH;
    if (!strcasecmp( langStr, XLT("portuguese"))) return PORTUGUESE;
    if (!strcasecmp( langStr, XLT("brazilian"))) return BRAZILIAN;
    if (!strcasecmp( langStr, XLT("french-canadian"))) return CANADIAN_FRENCH;
    if (!strcasecmp( langStr, XLT("swiss-german"))) return SWISS_GERMAN;
    if (!strcasecmp( langStr, XLT("nynorsk"))) return NYNORSK;
    if (!strcasecmp( langStr, XLT("finnish"))) return FINNISH;

	/* The passed language string is supposed to be XLT'd, but this is
		not documented. Allow untranslated strings to be passed as well */
    if (!strcasecmp( langStr, NXLT("english"))) return ENGLISH;
    if (!strcasecmp( langStr, NXLT("german"))) return GERMAN;
    if (!strcasecmp( langStr, NXLT("french"))) return FRENCH;
    if (!strcasecmp( langStr, NXLT("spanish"))) return SPANISH;
    if (!strcasecmp( langStr, NXLT("italian"))) return ITALIAN;
    if (!strcasecmp( langStr, NXLT("british"))) return BRITISH;
    if (!strcasecmp( langStr, NXLT("swedish"))) return SWEDISH;
    if (!strcasecmp( langStr, NXLT("danish"))) return DANISH;
    if (!strcasecmp( langStr, NXLT("norwegian"))) return NORWEGIAN;
    if (!strcasecmp( langStr, NXLT("dutch"))) return DUTCH;
    if (!strcasecmp( langStr, NXLT("portuguese"))) return PORTUGUESE;
    if (!strcasecmp( langStr, NXLT("brazilian"))) return BRAZILIAN;
    if (!strcasecmp( langStr, NXLT("french-canadian"))) return CANADIAN_FRENCH;
    if (!strcasecmp( langStr, NXLT("swiss-german"))) return SWISS_GERMAN;
    if (!strcasecmp( langStr, NXLT("nynorsk"))) return NYNORSK;
    if (!strcasecmp( langStr, NXLT("finnish"))) return FINNISH;
		
    return( NOLANG);
}

/*-----------------------------------------------------------------------
    Given the Proximity code for the language, output the proper name
    (e.g. "English") or the 3-letter abbreviation used as part of
    language data file names.

    The caller should allocate a buffer of at least size 32 to hold the
    full returned name or 4 for the file infix.

    The legal Proximity codes run from 1 to MAXLANG (16 as of 3/92).
    It is NOT an error to pass an illegal code; the function simply
    returns an empty string in this case.
------------------------------------------------------------------------*/
void            
SpGetLanguageName ( langCode, langName )
int                 langCode;               /* Code for language */
char                *langName;              /* OUT: Name of language */
{
    spGetLangName(TRUE, langCode, langName);
}

void            
SpGetLanguageFileInfix ( langCode, infix )
int                 langCode;               /* Code for language */
char                *infix;             /* OUT: infix (3 letters) */
{
    spGetLangName(FALSE, langCode, infix);
}

static void     
spGetLangName ( full, langCode, langName )
int                 full;                   /* TRUE for full name */
int                 langCode;               /* Code for language */
char                *langName;              /* OUT: Name of language */
{
    char *XLT();    /* ??0 why is this needed? */
    switch( langCode) {
        case ENGLISH:
            MBstrcpy( langName, (full)? XLT("English"):NXLT("eng")); return;
        case GERMAN:
            MBstrcpy( langName, (full)? XLT("German"):NXLT("grm")); return;
        case FRENCH:
            MBstrcpy( langName, (full)? XLT("French"):NXLT("frn")); return;
        case SPANISH:
            MBstrcpy( langName, (full)? XLT("Spanish"):NXLT("spn")); return;
        case ITALIAN:
            MBstrcpy( langName, (full)? XLT("Italian"):NXLT("itl")); return;
        case BRITISH:
            MBstrcpy( langName, (full)? XLT("British"):NXLT("brt")); return;
        case SWEDISH:
            MBstrcpy( langName, (full)? XLT("Swedish"):NXLT("swd")); return;
        case DANISH:
            MBstrcpy( langName, (full)? XLT("Danish"):NXLT("dan")); return;
        case NORWEGIAN:
            MBstrcpy( langName, (full)? XLT("Norwegian"):NXLT("nrw")); return;
        case DUTCH:
            MBstrcpy( langName, (full)? XLT("Dutch"):NXLT("dut")); return;
        case PORTUGUESE:
            MBstrcpy( langName, (full)? XLT("Portuguese"):NXLT("prt")); return;
        case BRAZILIAN:
            MBstrcpy( langName, (full)? XLT("Brazilian"):NXLT("brz")); return;
        case CANADIAN_FRENCH:
            MBstrcpy( langName, (full)? XLT("French-Canadian"):NXLT("cfr")); return;
        case SWISS_GERMAN:
            MBstrcpy( langName, (full)? XLT("Swiss-German"):NXLT("sgr")); return;
        case NYNORSK:
            MBstrcpy( langName, (full)? XLT("Nynorsk"):NXLT("nyn")); return;
        case FINNISH:
            MBstrcpy( langName, (full)? XLT("Finnish"):NXLT("fin")); return;
        default:
            MBstrcpy( langName, ""); return;
    }
}

/* }***********************************************************************/

typeErr  
spSetState ( language, dictionaries, force, dictSubsetOk )
int                 language;           /* Proximity code for desired language*/
char                **dictionaries;     /* NULL-terminated array of user
                                            dictionary filenames to use */
bool                force;              /* TRUE -> If Proximity is up, it
                                            MUST be shut down COMPLETELY and 
                                            then brought back up */
bool                dictSubsetOk;       /* TRUE -> It is OK if the desired
                                            set of dictionaries is a subset
                                            of the loaded set of dictionries. */
{
    typeErr         status;             /* Speller return code */
    int             i;                  /* Loop counter */

    /* Verify requested language is acceptable */
    /* Higher-level function should intercept an unknown language name;
        this check is here just in case. */
    if ((language < ENGLISH) || (language > MAXLANG)) {
        sprintf( SpErrorObject, NXLT("%d"), language);
        return( SPELL_BADCODE);
    }

    /* If speller is up... */
    if (started) {

        if (force) goto shutdown;       /* Unconditional shutdown */

        /* Can return immediately if it is in the correct state */
        if (currLang == language) {

            /* Make sure the dictionaries we need are a subset of the ones
                we have. */
            /* We use this logic when called from the "add dict word" or
                "delete dict word" functions, which only specify the
                dictionary to be altered. In this case we do not want to
                shutdown and restart if there are other dictionaries being
                used, because they won't interfere. */
            if (dictSubsetOk) {
                for (i=0; dictionaries && dictionaries[i]; ++i) {
                    if (spDictListIndex ( dictionaries[i]) == NODICT) 
                        goto shutdown;      /* Didn't find dictionary */
                }
            }

            /* Else compare desired and current set of dictionaries for
                an EXACT match */
            else {
                for (i=0; i<nDictFiles; i++) {
                    if (!dictionaries || !dictionaries[i] || 
                                (i != spDictListIndex ( dictionaries[i]))) 
                        goto shutdown;
                }
                for (i=0; dictionaries && dictionaries[i]; ++i) 
                    ;                                   /* count passed dicts */
                if (nDictFiles != i) {
                    goto shutdown;
                }
            }

            return( AOK);       /* Already in desired state */
        }

shutdown:
        /* Else shut it down */
        /* !!1 Note shutdown is technically unnecessary if we are in the
            right language and there is enough room to add another dictionary.
            But watch out if the ordering of the dictionaries in the
            searchpath becomes important... */
        /* SpCloseSpeller(), being a public function, clears out the error
            strings, but we don't care since we wouldn't have gotten to
            this point if there had been an error. */
        status = SpCloseSpeller( );
        if (status != AOK) return( status);
    }

    /* Start it up */
    status = spSpellerStart( language, dictionaries);
    return( status);
}

/*-----------------------------------------------------------------------
    Private function to put the speller into a particular "state".

    A state consists of a language and an ordered set of user dictionaries.

    - If the speller is already up in the desired state, do nothing.
    - If the speller is not up, bring it up in the desired state.
    - If the speller is already up in a different state, bring
        it down and up again in the new state.

    Every public speller function that uses the Proximity services will
    specifiy the state desired for the operation via this function.
------------------------------------------------------------------------*/
/******************* Speller setup & terminate functions *************** {*/

/*-----------------------------------------------------------------------
    Private function to start up speller

    Note: If we get partway through but can't complete the startup,
    shut everything down and return an error. Perhaps the caller can
    retry in some other language or without the offending dictionary...
------------------------------------------------------------------------*/
static typeErr  
spSpellerStart ( language, dictionaries)
int                 language;           /* Desired language */
char              **dictionaries;     /* NULL-terminated array of user
                                              dictionary filenames to use */
{
    ed_spell_files  local;              /* Local filenames */
    typeErr         status;             /* Speller return code */

    currLang = language;
    SpGetLanguageName ( language, currLangName );

    started = TRUE;     /* SpCloseSpeller will turn this back off if problem */

    /* Generate the filenames that we will need for this language and
        look for the files in all the appropriate places. If any of
        the necessary files cannot be located or opened, throw an error
        immediately and exit */
    spSpellFilenamesAvailable( language, &spellFiles);

    /* Generate the "local" filenames to use in "file does not exist"
        error messages. */
    spSpellFilenames( language, "", &local);

    /* Open the basic "Language Environment" file if it exists and can
        be read. This file defines the basic characteristics of the 
        current language (e.g. phonetics, hyphenation, punctutation). */
    if (!strlen( spellFiles.phonFile)) {
        SpCloseSpeller( );              /* Shut down the speller */
        MBstrcpy( SpErrorObject, local.phonFile);
        return( FEENOENT);              /* No such file or directory */
    }
    if ((status = spAccessFile( spellFiles.phonFile, R_OK)) != AOK) {
        SpCloseSpeller( );              /* Shut down the speller */
        return( status);                /* No read permission */
    }
    if (phonread( spellFiles.phonFile, language, 0) == PRX_ERROR) {     /* Px */
        SpCloseSpeller( );              /* Shut down the speller */
        MBstrcpy( SpErrorObject, spellFiles.phonFile);
        return( PROX_BAD_FILE);         /* Prox says bad file (?) */
    }

    /* Set the active language appropriately */
    if (setlanguage( language) == PRX_ERROR) {                          /* Px */
        SpCloseSpeller( );              /* Shut down the speller */
        MBstrcpy( SpErrorObject, currLangName);
        return( PROX_BAD_LANG_ENV);
    }

#if 0       /* No 'gencmpd' in Proximity 6.1 - D Miller 2/94 */
    /* Language-specific setups */
    /* DHM 3/92: I copied this blindly from the Aster*x 2.0 code (in
        dc/sources/spell/utilities.c); I have no idea if it is still needed. */
    switch( language) {
        case SWEDISH: 
        case DANISH: 
        case NORWEGIAN: 
        case NYNORSK: 
        case FINNISH:
            /** proximity currently has a bug that these languages do **/ 
            /** not correctly set the compound word handling for spell **/ 
            /** correction.  This code is as per Proximity source code **/ 
            /** patch date March 20, 1991                              **/ 
            gencmpd( TRUE, TRUE, TRUE );                                /* Px */
            break; 
        default: 
            break;
    }
#endif

    /* Open the Spelling Error Detection System file if it exists and can
        be read. If a word is not found in the wordlists, this facility
        will make a guess at what is wrong with the word. */
    if (!strlen( spellFiles.detFile)) {
        SpCloseSpeller( );              /* Shut down the speller */
        MBstrcpy( SpErrorObject, local.detFile);
        return( FEENOENT);              /* No such file or directory */
    }
    if ((status = spAccessFile( spellFiles.detFile, R_OK)) != AOK) {
        SpCloseSpeller( );              /* Shut down the speller */
        return( status);                /* No read permission */
    }
    if (envread( spellFiles.detFile, language, 0) == PRX_ERROR) {       /* Px */
        SpCloseSpeller( );              /* Shut down the speller */
        MBstrcpy( SpErrorObject, spellFiles.detFile);
        return( PROX_BAD_FILE);         /* Prox says bad file (?) */
    }

    /* set compound correction for those languages which need it */
    setcpdcorr( language, TRUE );                                       /* Px */

    /*** Open the word lists (that help Proximity decide whether a word
        is legal or not) in the recommended order:
        - Stop cache (a super-override for all forms of a given word)
        - Normal cache (quick access to document-specific or common words)
        - Core lexicon (commonly-used words for the current language)
        - Linguibase ("complete" set of words for current language)
        - Dictionaries (disk-based, customized files of special words)

        Note that when a word is being checked, Proximity searches the
        lists in the above order. ***/

    /* Open the stop cache for use in detection of spelling errors. */
    stopCache = (CACHE *)cshopen( IN_STOPCACHE, SP_DETECT | SP_STOP, 4000, -1);  /* Px */
    if (!stopCache) {
        SpCloseSpeller( );              /* Shut down the speller */
        sprintf( SpErrorOverrideStr, XLT("Proximity problem opening stop cache"));
        return( PROX_BAD_CACHE);
    }

    /* Open the normal cache for use in detection of spelling errors. */
    normalCache = (CACHE *)cshopen( IN_CACHE, SP_DETECT, 4000, -1);              /* Px */
    if (!normalCache) {
        SpCloseSpeller( );              /* Shut down the speller */
        sprintf( SpErrorOverrideStr, XLT("Proximity problem opening cache"));
        return( PROX_BAD_CACHE);
    }

    /* Open the core lexicon for rapid detection and correction of
        spelling errors */
    if (!strlen( spellFiles.clxFile)) {
        SpCloseSpeller( );              /* Shut down the speller */
        MBstrcpy( SpErrorObject, local.clxFile);
        return( FEENOENT);              /* No such file or directory */
    }
    if ((status = spAccessFile( spellFiles.clxFile, R_OK)) != AOK) {
        SpCloseSpeller( );              /* Shut down the speller */
        return( status);                /* No read permission */
    }
    corelexPtr = (CORELEX *)clxread( spellFiles.clxFile, language, 0, NULL);      /* Px */
    if (!corelexPtr) {
        SpCloseSpeller( );              /* Shut down the speller */
        MBstrcpy( SpErrorObject, spellFiles.clxFile);
        return( PROX_BAD_FILE);         /* /* Prox says bad file (?) */
    }
    corelexPtr = (CORELEX *)clxopen( IN_CORELEX, SP_DETECT | SP_CORRECT, corelexPtr);            /* Px */
    if (!corelexPtr) {
        SpCloseSpeller( );              /* Shut down the speller */
        return( PROX_BAD_CLXOPEN);      /* Couldn't put clx in searchpath */
    }

    /* Open the main lexicon for detection and correction of spelling 
        errors and hyphenation points. */
    if (!strlen( spellFiles.lexFile)) {
        SpCloseSpeller( );              /* Shut down the speller */
        MBstrcpy( SpErrorObject, local.lexFile);
        return( FEENOENT);              /* No such file or directory */
    }
    if ((status = spAccessFile( spellFiles.lexFile, R_OK)) != AOK) {
        SpCloseSpeller( );              /* Shut down the speller */
        return( status);                /* No read permission */
    }
    lexPtr = (LEX *)lexopen( IN_LINGUIBASE, SP_DETECT | SP_CORRECT | SP_HYPHEN,spellFiles.lexFile, 0);             /* Px */
    if (!lexPtr) {
        SpCloseSpeller( );              /* Shut down the speller */
        MBstrcpy( SpErrorObject, spellFiles.lexFile);
        return( PROX_BAD_FILE);         /* /* Prox says bad file (?) */
    }

    /* Add the dictionaries now */
    if ((status = spAddDictionaries( dictionaries)) != AOK) {
        /* Speller was already shut down by spAddDictionaries() */
        return( status);                /* Missing or corrupt dictionary (?) */
    }

    return( AOK);
}


/*-----------------------------------------------------------------------
    Public function to completely shut down the speller.
------------------------------------------------------------------------*/
typeErr         
SpCloseSpeller ( )
{
    typeErr     status1=AOK;            /* Speller return code */
    typeErr     status2=AOK;            /* Speller return code */

    spClearErrors();                    /* Clear out errors */

    if (!started) return( AOK);         /* Redundant closings are OK */

    /* Close the thesaurus */
    if (thesOpen) {
        status1 = SpCloseThesaurus();
        thesOpen = FALSE;
    }

    /* Shut down hyphenation */
    if (doingHyphen) {
        hyclose();                                                      /* Px */
        doingHyphen = FALSE;
    }

    /* Take everything off the search path */
    if (stopCache) {
        cshclose( stopCache);                                           /* Px */
        stopCache = NULL;
    }
    if (normalCache) {
        cshclose( normalCache);                                         /* Px */
        normalCache = NULL;
    }
    if (corelexPtr) {
		/* note sequence: clxread, clxopen, clxclose, clxfree */
        clxclose( corelexPtr);                                          /* Px */
        clxfree( corelexPtr);											/* Px */
        corelexPtr = NULL;
    }
    if (lexPtr) {
        lexclose( lexPtr);                                              /* Px */
        lexPtr = NULL;
    }

    /* Remove dictionaries from search path */
    status2 = spRemoveAllDictionaries( );

    /* Shut down the language environment */
    envclose( currLang);                                                /* Px */
    phonclose( currLang);                                               /* Px */

    /* Free up Proximity's memory */
    memfree();                                                          /* Px */

    started = FALSE;
    currLang = NOLANG;
    strcpy( currLangName, "");

    /* Return the "last" error; it will match the SpErrorObject */
    if (status2 != AOK) return( status2);
    if (status1 != AOK) return( status1);
    return( AOK);
}
/* }***********************************************************************/


/*************** Spelling error detect & correct functions  ************ {*/


/*-----------------------------------------------------------------------
    Retrieve a set of alternate spellings for the passed word.

    The caller states how many alternate spellings he wants. The number of 
    words actually being returned is output, and an array is filled 
    with pointers to those words (words are allocated locally and should
    be freed by the caller). The caller should declare the array as 
    "char  *words[# words wanted]".

    A flag is also output that indicates whether the passed word is
    on the list; i.e. that indicates that the passed word is a legal word.

    The number of words being reported may be set to zero, or fewer
    than the caller requested.

    The Proximity doc says the alternate spellings will be ordered from
    most to least likely (i.e. of what the user wanted). If the passed
    word is in a dictionary as the "misspell" part of a misspell/correct
    pair, the correction is returned at or near the top of the list.
------------------------------------------------------------------------*/
typeErr         
SpListValidSpellings( in_word, language, dictionaries, 
                                            maxGet, gottenP, words, inList)
mbchar              *in_word;               /* Word in question. */
int                 language;           /* Language we are working in
                                            (i.e. appropriate for 'word') */
char                **dictionaries;     /* NULL-terminated array of user
                                            dictionaries that SHOULD be on
                                            the search path. */
int                 maxGet;             /* Maximum # of alternatives wanted. */
int                 *gottenP;           /* OUT: Number of alts returned. */
mbchar              *words[];           /* OUT: Caller's word array */
bool                *inList;            /* OUT: TRUE -> passed word is in list*/
{
    typeErr         tError;             /* Subroutine error code */
    char            **rankptr;          /* List of alternate words */
    int             i;                  /* Loop counter */
    mbchar          word[SPELL_LONG_WORD+1];

    spClearErrors();                    /* Clear out errors */
    *gottenP = 0;
    *inList = 0;

#if BUILD_WIDE
    if ( MBnum_wide_chars_n( in_word, MBstrlen(in_word) ) != 0 )
        /* do not spell check a wide char word */
        return( AOK );
#endif

    /* Enforce word length */
    if ( MBstrlen(in_word) >= MAX_PROX_WORD) {
        MBstrcpy( SpErrorObject, in_word );
        return( SPELL_LONG_WORD);       /* Text string too long */
    }

    /* Set the speller state to the proper language, and cause the specified
        dictionary to be placed on the search path if it is not already 
        there. */
    /* Note we insist that the set of dictionaries that the caller specified
        EXAXTLY matches the set of dictionaries that are currently on the 
        search path. */
    if ((tError = spSetState( language, dictionaries, FALSE, FALSE )) != AOK) {
        return( tError);                /* oh horrors! */
    }

    /* make a copy */
    MBstrcpy( word, in_word );

    /* Convert the word to the Proximity character set */
    spXlateString( word, AxToProx);

    rankptr = (char **) correct( (UCHAR *)word );                       /* Px */

    *inList = FALSE;
    *gottenP = 0;
    for (i=0; rankptr[i] != NULL; ++i) {
        if (i+1 > maxGet) break;
        spRemoveSlashes( (char *) rankptr[i]);

        /* !!0 use wpstrcmp when available */
        if (!*inList && !MBstrcmp( word, rankptr[i])) 
              *inList = TRUE;
        words[i] = (mbchar *)malloc( (MBstrlen(rankptr[i])+1) * sizeof(mbchar));
        MBstrcpy( words[i], rankptr[i]);

        /* Convert the word to the Aster*x character set */
        spXlateString( words[i], ProxToAx);

        ++*gottenP;
    }
    return( AOK);
}


/*-----------------------------------------------------------------------
    Determine whether a word is spelled correctly or not.

    By "spelled correctly" we mean the word is found in one of the word
    lists in the current search path, and that it is not flagged as
    misspelled.

    This function does not attempt to classify the type of error (if
    any), nor does it examine the word in the context of the surrounding
    document. There are other, more complex functions available for this.
------------------------------------------------------------------------*/
typeErr         
SpWordSpelledCorrectly ( in_word, language, dictionaries, spellOK) 
mbchar               *in_word;               /* The word in question */
int                 language;           /* Language we are working in
                                            (i.e. appropriate for 'word') */
char                **dictionaries;     /* NULL-terminated array of user
                                            dictionaries that SHOULD be on
                                            the search path. */
bool                *spellOK;           /* OUT: TRUE if word found in
                                            search path (i.e. spelled OK) */
{
    typeErr         tError;             /* Subroutine error code */
    int             retcode;            /* Proximity return code */
    mbchar          *word;

    spClearErrors();                    /* Clear out errors */

#if BUILD_WIDE
    if ( MBnum_wide_chars_n( in_word, MBstrlen(in_word) ) != 0 )
        /* do not spell check a wide char word */
        return( AOK );
#endif

    /* Enforce word length */
    if ( MBstrlen(in_word) >= MAX_PROX_WORD) {
        MBstrcpy(SpErrorObject, in_word);
        return( SPELL_LONG_WORD);       /* Text string too long */
    }

    /* Set the speller state to the proper language, and cause the specified
        dictionary to be placed on the search path if it is not already 
        there. */
    /* Note we insist that the set of dictionaries that the caller specified
        EXAXTLY matches the set of dictionaries that are currently on the 
        search path. */
    if ((tError = spSetState( language, dictionaries, FALSE, FALSE )) != AOK) {
        return( tError);                /* oh horrors! */
    }

    /* make a narrow copy */
    word = (mbchar *)malloc( (MBstrlen(in_word)+1) * sizeof(mbchar) );
    MBstrcpy(word, in_word);

    /* Convert the word to the Proximity character set */
    spXlateString( word, AxToProx);

    retcode = isword( (UCHAR *)word );                                  /* Px */
    if (retcode == IW_NOTFOUND) 
        *spellOK = FALSE;
    else 
        *spellOK = TRUE;


    free( word );
    return( AOK);
}


/*-----------------------------------------------------------------------
    Determine whether a word is spelled correctly or not; output a
    variety of Proximity information about the word.

    For use by hardcore hackers or desperate debuggers who need to get 
    down and dirty with Proximity. The caller will need to be able
    to interpret Proximity flags and codes to make any sense of this.

    The caller should allocate three character arrays, each of length
    MAX_PROX_WORD+1, as indicated by the parameter list below.

    In general, if "iswordRet" is not set to IW_NOTFOUND, the word
    was found on the search path. 
------------------------------------------------------------------------*/
typeErr         
SpWordProximityData( in_word, language, dictionaries, 
                                query, qflags, qlen, qlookup, qinfo, 
                                wlist, wflags, iswordRet)
mbchar               *in_word;               /* The word in question */
int                 language;           /* Language we are working in
                                            (i.e. appropriate for 'word') */
mbchar              **dictionaries;     /* NULL-terminated array of user
                                            dictionaries that SHOULD be on
                                            the search path. */
mbchar              query[MAX_PROX_WORD+1];   /* OUT: Query as received by Prox; 
                                            converted form of 'word' */
int                 *qflags;            /* OUT: flags of doflagged 'query' */
int                 *qlen;              /* OUT: length of doflagged 'query' */
mbchar              qlookup[MAX_PROX_WORD+1]; /* OUT: Doflagged version of 'query'; 
                                            true size is 'qlen' */
mbchar              qinfo[MAX_PROX_WORD+1];   /* OUT: Char-by-char flags of 'qlookup';
                                            true size is 'qlen' */
int                 *wlist;             /* OUT: Code number of word list the
                                            word was found in, if it was found.
                                            Either a constant like IN_CACHE,
                                            or a user dictionary code as was
                                            specified to EdAddDictionary(). */
int                 *wflags;            /* OUT: Flags associated with the word
                                            in the word list the word was
                                            found in, if word was found. */
int                 *iswordRet;         /* OUT: Proximity return code from 
                                            the isword() call */
{
    typeErr         tError;             /* Subroutine error code */
    int             i;                  /* Loop counter */
    mbchar          *word;

    spClearErrors();                    /* Clear out errors */

#if BUILD_WIDE
    if ( MBnum_wide_chars_n( in_word, MBstrlen(in_word) ) != 0 )
        /* do not spell check a wide char word */
        return( AOK );
#endif

    /* Enforce word length */
    if ( MBstrlen(in_word) >= MAX_PROX_WORD) {
        MBstrcpy(SpErrorObject,in_word);
        return( SPELL_LONG_WORD);       /* Text string too long */
    }

    /* Set the speller state to the proper language, and cause the specified
        dictionares to be placed on the search path if they are not already 
        there. */
    /* Note we insist that the set of dictionaries that the caller specified
        EXAXTLY matches the set of dictionaries that are currently on the 
        search path. */
    if ((tError = spSetState( language, dictionaries, FALSE, FALSE )) != AOK) {
        return( tError);                /* oh horrors! */
    }

    /* make a narrow copy */
    word = (mbchar *)malloc( (MBstrlen(in_word)+1) * sizeof(mbchar) );
    MBstrcpy(word, in_word);

    /* Convert the word to the Proximity character set */
    spXlateString( word, AxToProx);

    *iswordRet = isword( (UCHAR *)word );							/* Px */
    VERIFY( Prox_Qulen() <= MAX_PROX_WORD);							/* Px */
    MBstrcpy( query, (mbchar *) Prox_Quword());						/* Px */

    *qflags = Prox_Quflags();										/* Px */
    *qlen = Prox_Qulen();											/* Px */
    MBstrcpy( qlookup, (mbchar *)Prox_Qulookup());					/* Px */
    MBstrncpy( qinfo, (mbchar *)Prox_Quinfo(), *qlen-1);			/* Px */
    *wlist = Prox_Quwlist();										/* Px */
    *wflags = Prox_Quwflags();										/* Px */


    free( word );
    return( AOK);
}


/*-----------------------------------------------------------------------
    This function controls the low-level spellchecking.

    It accepts the "latest" character from the caller and presents
    it to Proximity, to see if it will trigger a reportable error.
    Note that it is the caller's responsibility the initialize
    Proximity token building.
------------------------------------------------------------------------*/
typeErr  
spCheckChar ( currCharInt, canReport, mask, tokP, tillDone,
                            tokCode, errP, tokStartPos, tokEndPos)
wchar               currCharInt;        /* Next character in string being
                                            spellchecked */
bool                canReport;          /* TRUE -> error at this location can
                                            be reported. */
sp_error_mask       *mask;              /* Errors caller wants reported */
TOKEN               *tokP;              /* IN/OUT: Ptr to Proximity token */
bool                tillDone;           /* TRUE -> even when a reportable
                                            error is found, keep calling
                                            tokdesc() till TOK_DONE */
int                 *tokCode;           /* OUT: Proximity's comment on latest
                                            token; TOK_MADE means a possible
                                            error has been found */
int                 *errP;              /* OUT: AOK if no errors found; else
                                            a PROX_TOK_ code */
int                 *tokStartPos;       /* OUT: Offset in current token that
                                            begins the error */
int                 *tokEndPos;         /* OUT: Offset in current token that
                                            ends the error */

{
    int             proxChar;           /* Proximity wants this as an int */
    register int    tokval;             /* Code for token description */

    *errP = AOK;                    /* No error so far */
    proxChar = ((unsigned) currCharInt) & 0xFF;     /* !!0 OK for new char scheme? */

    /* Add the passed character to the token under construction */
    *tokCode = tokmake( proxChar, 0, tokP);                         /* Px */

    /* Unless Proximity is ready to rule on the token, return the
        token value to the caller */
    if (*tokCode != TOK_MADE) {
        return( AOK);
    }

    /* If we are here, we must have received a TOK_MADE response to the
        tokmake() call. This means Proximity has isolated a token and
        is prepared to rule on it. */

    /* Analyze the current token (made above) */
    tokval = tokdesc( tokP);                                        /* Px */

    /* If Proximity tells us immediately we are "done", without any other
        codes, that means the word is valid. Move beyond the text in this
        token and continue processing. */
    if (tokval == TOK_DONE) {
        return( AOK);
    }

    /* If we are here, Proximity has some information about this token.
        We will have to decide if this represents a "misspelling" or
        not. If the caller is not yet interested in errors (at this
        offset), we consider it not to be an error so we call tokskip()
        to terminate the analysis of this token (we should receive 
        TOK_DONE on the next cycle).

        Note that Proximity may have more than one comment (or "error") 
        about a given token. For example, if the token "programmer's" 
        was contstructed, a tokdesc call would first return the TOK_POSSESS 
        code for the entire token (means "a word in possessive form"), and 
        then the TOK_SAPOS code for the "'s" portion (means "the 's part 
        of a possessive word"). As illustrated, not all token descriptions 
        will be considered "errors"; some just identify types of words.

        The major design decision here is, given a token, what to do if
        Proximity returns a tokdesc call that the user is NOT interested
        in marking as misspelled. We can either:
            1. Keep analyzing the token till we receive a tokdesc that
                the user IS interested in seeing and report it (or
                we get the TOK_DONE code), OR
            2. Declare the token to be "OK" and call tokskip(), which
                should generate the TOK_DONE code at the next cycle
                and halt the analysis.
    */

    /* Comments on Method 1:
    
       Note that method 1 as above is somewhat inefficient. Since
        the "error mask" allows the caller to have ANY tokdesc() code
        reported, we have to analyze every token "to exhaustion".
        In the above example, even if the passed error mask does not
        want TOK_POSSESS codes reported, we have to keep analyzing
        the token since the caller may want TOK_SAPOS codes reported.
        There is no way for the caller to state "I don't want TOK_POSSESS
        _or any of its subcodes_ reported". Although this method does
        allow the caller maximum flexibility, it is somewhat slower
        when processing possessives, prefix/suffix words, and
        hyphenated words (for "normal" words there should be no
        difference). I have no idea whether this inefficiency will
        be noticable in "human time". 

       In addition, method 1 causes "false positives": If the user is
        not interested in seeing Roman numerals marked as misspelled,
        we will first get the TOK_ROMAN code and then (for most cases)
        the generic TOK_MISP code. The user would have rather skipped
        the token alltogether.
    */

    /* Comments on Method 2:

       Note that method 2 as above could hide spelling errors. For
        example, no one will be interested in marking possessives as
        misspelled. Suppose we have some misspelled possessive word
        (e.g. mskblt's). If Proximity first gives us a TOK_POSSESS,
        which the user does not want reported, we will conclude the
        word is OK and continue. This is an error.
    */

    /* Prior to 10/93, we had implemented method 1. As of 10/93, we
        will implement a modified method 1, where finding certain
        codes that the user does not want reported will halt the
        analysis, but others will allow it to continue. See the
        actual code below for more details. - D Miller 10/5/93 */

    do {

        /* If an error is detected (should not happen), clear the error
            and continue with the next character */
        if (tokval == TOK_ERROR) {
            ASSERT( 0);         /* Catch during development */
            tokinit(tokP, DET_RST);                                 /* Px */
            return( AOK);
        }

        /* If we are not yet at the offset where error reports are
            desired, try to get through this token as quickly as
            possible without reporting anything */
        /* Also, if we have already discovered a reportable error
            but the caller wants us to "exhaust" the token (so it
            can continue with the next character), do the same. */
        if (!canReport || ((*errP != AOK) && tillDone)) {
            tokskip(tokP);                                          /* Px */
        }

        /* If the user wanted this error reported, we can return it
            immediately; but if the caller requests that the token be
            exhausted, we must store the error and continue. */
        else {
            if (SpGetErrMaskValue( mask, tokval)) {
                *errP = tokval;
                *tokStartPos = pub_tokspos(tokP);
                *tokEndPos = pub_tokepos(tokP);
                if (!tillDone) return( AOK);
            }

            /* If the user did not want this error reported, we MAY
                halt further analysis of this token. */
            /* There is no point further analyzing any tokens of the
                following type, assuming the user does not want these
                types reported:
                    Dates
                    Non-alphabetic text
                    Numbers
                    Outline markers
                    Roman numerals
                    Times
            */
            else {
                if ((tokval == TOK_DATE) || (tokval == TOK_NONALPH) ||
                            (tokval == TOK_NUMBER) || (tokval == TOK_OUTLN) ||
                            (tokval == TOK_ROMAN) || (tokval == TOK_TIME)) {
                    tokskip(tokP);                                  /* Px */
                }
            }
        }
    } while ((tokval=tokdesc(tokP)) != TOK_DONE);                   /* Px */

    /* Either we did not get a reportable error, or we did get one but
        had to exhaust the token */
    return( AOK);

}

/*-----------------------------------------------------------------------
    Return the first error occuring in the passed string.

    The caller will pass a structure containing the Proximity spellcheck
    return codes that he cares about. This function outputs the code for
    the first error (that the caller cares about) that we find, plus the
    starting and ending offsets (with respect to the passed string) of the 
    "scope" of the error. Note that the ending offset indicates the
    offset of the last character within the problem area; e.g. if the
    string "See tha cat" is passed, the misspelling area starts at offset
    4 and ends at offset 6.

    The caller can delay the reporting of errors until some arbitrary 
    offset in the passed string. The entire string is fed to the Proximity
    routines (to provide context), but errors are ignored until the specifed
    point. Ideally, the caller will pass at least one complete word before 
    the word at which error reporting is desired so that "repeated word" 
    and "non-capitalized start of sentence" errors can be detected.

    If the caller requests error reporting beginning at offset zero, it
    will be assumed that offset zero represents the start of a sentence.

    If no errors can be reported, the error code will be set to zero (an
    unused value).
------------------------------------------------------------------------*/
typeErr         
SpCheckString ( in_string, startOffset, language, dictionaries,
                            mask, errP, errStartP, errEndP, scanEndP)
wchar               *in_string;         /* String to spellcheck. */
int                 startOffset;        /* Start reporting errors here */
int                 language;           /* Language we are working in
                                            (i.e. appropriate for 'string') */
mbchar              **dictionaries;     /* NULL-terminated array of user
                                            dictionaries that SHOULD be on
                                            the search path. */
sp_error_mask       *mask;              /* Errors caller wants reported */
int                 *errP;              /* OUT: AOK if no errors found; else
                                            a PROX_TOK_ code */
int                 *errStartP;         /* OUT: First character in "string"
                                            that is part of the error */
int                 *errEndP;           /* OUT: Last character in "string"
                                            that is part of the error */
int                 *scanEndP;          /* OUT: Last character in "string"
                                            that was examined in order to
                                            generate reported error. Caller
                                            should request reporting AFTER this 
                                            point on next call to avoid getting
                                            the same error */
{
    typeErr         tError;             /* Subroutine error code */
    TOKEN          	*tokP;              /* Proximity token */
    bool            done=FALSE;         /* TRUE -> done w/passed string */
    register int    currLoc;            /* Offset being examined */
    register char   currChar;           /* Current character being examined */
    bool            canReport;          /* TRUE -> can report error here */
    int             tokCode;            /* tokmake code from subroutine */
    int             tokStartPos;        /* Offset in current token that
                                            begins the error */
    int             tokEndPos;          /* Offset in current token that
                                            ends the error */
    mbchar *string;

    spClearErrors();                    /* Clear out errors */

    *errP = AOK;

#if BUILD_WIDE
    if ( Wnum_wide_chars( in_string ) != 0 )
        /* do not spell check a wide char word */
        return( AOK );
#endif
    /* Set the speller state to the proper language, and cause the specified
        dictionaries to be placed on the search path if they are not already 
        there. */
    /* Note we insist that the set of dictionaries that the caller specified
        EXAXTLY matches the set of dictionaries that are currently on the 
        search path. */
    if ((tError = spSetState( language, dictionaries, FALSE, FALSE )) != AOK) {
        return( tError);                /* oh horrors! */
    }

    /* make a narrow copy */
    string = (mbchar *)malloc( (Wstrlen(in_string)+1) * sizeof(wchar) );
    WToMbStr( in_string, string );

    /* Convert the string to the Proximity character set */
    spXlateString( string, AxToProx);

    /* Initialize the token-construction facility. */
    /* The following call causes Proximity to immediately expect the start 
        of a sentence. By convention, if the caller requests error reporting
        starting at offset zero, offset zero should represent the start
        of a sentence (or paragraph, etc). If the caller requests error
        reporting starting at some offset, we don't really care what
        Proximity expects since we won't report any errors associated with
        offset zero. So this is OK... I think... */
	tokP = (TOKEN *)pub_token();
    if (!tokinit( tokP, DET_NEW)) {                 /* Px */
        return( PROX_BAD_TOKEN);
    }

    /* !!1 If necessary, do a "tokinit( tokP, DET_CLR)" at this point to
        make Proximity NOT expect the start of a sentence. */

    *errStartP = 0;
    canReport = FALSE;
    for( currLoc=0; !done; ++currLoc) {

        /* Extract character. Note we DO want to pass the terminating
            NULL through to force Proximity to rule on the last token */
        currChar = *(string+currLoc);
        if (currChar == '\0') done = TRUE;

        /* If we are not yet reporting errors, see if we can start here. */
        if (!canReport && (currLoc >= startOffset)) canReport = TRUE;

        tError = spCheckChar( (int)currChar, canReport, mask, tokP, FALSE,
                    &tokCode, errP, &tokStartPos, &tokEndPos);

        /* If Proximity wants more characters (before ruling on the token)
            keep passing 'em through. */
        if (tokCode == TOK_MORE) {
            continue;
        }

        /* Else if token is still empty, keep passing characters. Note we
            consider the token has not yet begun (in terms of where to
            start highlighting if an eventual misspelling is detected). */
        else if (tokCode == TOK_NULL) {
            *errStartP = currLoc+1;
            continue;
        } 

        /* Else if an error is detected (should not happen), clear the error
            and continue with the next character */
        else if (tokCode == TOK_ERROR) {
            ASSERT( 0);         /* Catch during development */
            tokinit(tokP, DET_RST);                                     /* Px */
            *errStartP = currLoc+1;
            continue;
        }

        /* If we are here, we must have received a TOK_MADE response to the
            tokmake() call. This means Proximity has isolated a token and
            has either returned an error or has decided the token is OK
            (if we are not reporting errors yet, it will always be the latter
            case). */

        /* Found an error we want to report */
        if (*errP > 0) {
            *errStartP += tokStartPos;
            *errEndP = *errStartP + (tokEndPos - tokStartPos - 1);
            *scanEndP = currLoc;
            tokinit(tokP, DET_RST); /* !!1 Necessary? */        /* Px */
            free( string );
            return( AOK);
        } 

        /* Else no error to report; we will start (to build) a new token */
        *errStartP = currLoc+1;

    } /* for each character */

    /* All done; no errors */
    free( string );
    return( AOK);
}


/* }***********************************************************************/


/************************** Hyphenation functions  ********************* {*/


/*-----------------------------------------------------------------------
    Return whether the necessary Proximity files exist to perform
    hyphenation in a given language. 

    This function checks for the files needed to bring the speller up
    in the language, as well as the specific files needed for hyphenation.
------------------------------------------------------------------------*/
bool            
SpIsHyphenationOK ( language )
int                 language;           /* Desired language (use Prox code) */
{
    ed_spell_files  files;              /* Structure to hold filenames */

    /* Verify requested language is acceptable */
    if ((language < ENGLISH) || (language > MAXLANG)) {
        ASSERT( 0);         /* Shouldn't happen during development */
        return( FALSE);     /* Unknown languages aren't supported */
    }

    spSpellFilenamesAvailable( language, &files);
    return( *files.phonFile && *files.clxFile && *files.lexFile && 
                                *files.detFile && *files.hyFile);
}

/*-----------------------------------------------------------------------
    Output a set of ranked hyphenation points for the passed word.

    Caller should set 'how' to HYPH_DICT if the hyphenation points
    are to come from the Proximity data files (or user dictionaries),
    HYPH_CALC if the hyphenation points are to be calculated, or
    HYPH_DICT_CALC to look first in the data files and calculate
    the points if the word is not found. The returned 'howDid'
    will be set to HYPH_DICT or HYPH_CALC depending on how the
    hyphenation points were generated; this should only be a surprise
    if 'how' was HYPH_DICT_CALC. If the word could not be hyphenated
    (this should only happen when 'how' is HYPH_DICT, since the
    calculation SHOULD be able to handle any word), 'howDid' is
    set to HYPH_NONE.

    Caller should pass 'hyPoints' as an array of size MAX_PROX_WORD+1 (or,
    really, the same size as 'word') to hold the hyphenation points
    for 'word'. For every character in 'word', the corresponding
    character in 'hyPoints' will hold one of the following values
    in its lowest 4 bits:
        0 if not a hyphenation point
        HY_NORM if a normal hyphenation point FOLLOWS the character
        HY_PREF if a preferred hyphenation point FOLLOWS the character
        HY_UNPREF if an unpreferred hyphenation point FOLLOWS the character

    If the caller does not wish the hyphenation points to be ranked,
    'rank' should be set FALSE and the lowest 4 bits of 'hyPoints' will 
    only contain values 0 and HY_NORM.

    Please see the Proximity documentation (and the Proximity file
    sahyphen.h) for a more complete description of the various hyphenation
    point values. Specifically, please note that there may be other values
    in positions above the lowest 4 bits.
------------------------------------------------------------------------*/
typeErr         
SpHyphenateWord ( in_word, language, dictionaries, rank, 
                                            how, howDid, hyPoints)
mbchar               *in_word;               /* Word to hyphenate. */
int                 language;           /* Language we are working in
                                            (i.e. appropriate for 'word') */
mbchar              **dictionaries;     /* NULL-terminated array of user
                                            dictionaries that SHOULD be on
                                            the search path. */
bool                rank;               /* TRUE -> rank the hyphenation
                                            points. */
int                 how;                /* How to generate hyph points 
                                            (a HYPH_XXXX value) */
int                 *howDid;            /* OUT: How the hyph points were
                                            generated (a HYPH_XXXX value) */
mbchar              hyPoints[MAX_PROX_WORD+1];/* OUT: Hyphenation points for word */
{
    typeErr         status;             /* Subroutine retcode */
    int             proxErr;            /* Proximity error code */
    ed_spell_files  local;              /* Local filenames */
    mbchar          *word;

    spClearErrors();                    /* Clear out errors */

#if BUILD_WIDE
    if ( MBnum_wide_chars_n( in_word, MBstrlen(in_word) ) != 0 )
        /* do not spell check a wide char word */
        return( AOK );
#endif


    /* Enforce word length */
    if ( MBstrlen(in_word) >= MAX_PROX_WORD) {
        MBstrcpy(SpErrorObject, in_word);
        return( SPELL_LONG_WORD);       /* Text string too long */
    }

    /* Set the speller state to the proper language, and cause the specified
        dictionaries to be placed on the search path if they are not already 
        there. */
    /* Note we insist that the set of dictionaries that are currently on the 
        search path EXACTLY matches the set of dictionaries that the caller 
        specifies. */
    if ((status = spSetState( language, dictionaries, FALSE, FALSE )) != AOK) {
        return( status);
    }

    /* Also read in the hyphenation data file if necessary */
    if (!doingHyphen) {
        spSpellFilenames( language, "", &local);
        if (!MBstrlen( spellFiles.hyFile)) {
            MBstrcpy( SpErrorObject, local.hyFile);
            return( FEENOENT);              /* No such file or directory */
        }
        if ((status = spAccessFile( spellFiles.hyFile, R_OK)) != AOK) {
            return( status);                /* No read permission */
        }
        proxErr = hyread(spellFiles.hyFile, 0, FALSE);                  /* Px */
        if (proxErr == PRX_ERROR ) {
            MBstrcpy( SpErrorObject, spellFiles.hyFile);
            return( PROX_BAD_FILE);         /* /* Prox says bad file (?) */
        }
        doingHyphen = TRUE;
    }

    /* make a narrow copy */
    word = (mbchar *)malloc( (MBstrlen(in_word)+1) * sizeof(mbchar) );
    MBstrcpy(word, in_word);

    /* Convert the word to the Proximity character set */
    spXlateString( word, AxToProx);

    switch( how) {

        /* A silly call */
        case HYPH_NONE:
            {   int     len, i;
                len = MBstrlen( word);
                for (i=0; i<len; i++) hyPoints[i] = 0;
                *howDid = HYPH_NONE;
                ASSERT( 0);         /* Shouldn't happen during development */
                free( word );
                return( AOK);
            }

        /* Attempt to find the hyphenation points in the dictionary */
        case HYPH_DICT:
        case HYPH_DICT_CALC:
            proxErr = vhyword( (UCHAR *)word, (UCHAR *)hyPoints );      /* Px */
            if (proxErr == PRX_ERROR) {
                MBstrcpy( SpErrorObject, word);
                free( word );
                return( PROX_BAD_HYPH);
            }

            /* Found it in dictionary. */
            if (proxErr != IW_NOTFOUND) {
                *howDid = HYPH_DICT;
                break;
            }

            /* Not in dictionary; if only looking there, we failed */
            if (how == HYPH_DICT) {
                *howDid = HYPH_NONE;
                free( word );
                return( AOK);
            }

            /* Not in dictionary, so caller wants us to calculate the points. 
                Purposely fall through to case HYPH_CALC */

        case HYPH_CALC:
            proxErr = hycalc( (UCHAR *)word, (UCHAR *)hyPoints );        /* Px */
            if (proxErr == PRX_ERROR) {
                MBstrcpy(SpErrorObject, in_word);
                return( PROX_CALC_HYPH);
            }
            *howDid = HYPH_CALC;
            break;

        default:
            sprintf( SpErrorObject, NXLT("%d"), how);
            ASSERT( 0);         /* Shouldn't happen during development */
            return( SPELL_BAD_HYPH_CODE);

    }
    if (!rank) 
       {
       free( word );
       return( AOK);
       }

    /* Rank the hyphenation points into 3 categories */
    proxErr = vhyrank( (UCHAR *)word, (UCHAR *)hyPoints );              /* Px */
    if (proxErr == PRX_ERROR) {
        MBstrcpy(SpErrorObject, in_word);
        return( PROX_RANK_HYPH);
    }

    free( word );
    return( AOK);
}

/* }***********************************************************************/

/************************** Thesaurus functions  *********************** {*/

/*-----------------------------------------------------------------------
    Return whether the necessary Proximity files exist to perform
    thesaurus queries in a given language. 

    This function checks for the files needed to bring the speller up
    in the language, as well as the specific files needed by the
    thesaurus.
------------------------------------------------------------------------*/
bool            
SpIsThesaurusOK ( language )
int                 language;           /* Desired language (use Prox code) */
{
    ed_spell_files  files;              /* Structure to hold filenames */

    /* Verify requested language is acceptable */
    if ((language < ENGLISH) || (language > MAXLANG)) {
        ASSERT( 0);         /* Shouldn't happen during development */
        return( FALSE);     /* Unknown languages aren't supported */
    }

    spSpellFilenamesAvailable( language, &files);
    return( *files.phonFile && *files.clxFile && *files.lexFile && 
                                *files.detFile && *files.thesFile);
}

/*-----------------------------------------------------------------------
    Retrieve information about the data the thesaurus can provide in a 
    given language.

    The caller will need to be able to decode the Proximity ST_XXXXXX
    flage (e.g. ST_HASSYN) that are returned.

    If the speller cannot be brought up and the thesaurus opened in
    the desired language, an error will be issued. If the caller wishes
    to avoid this error, SpIsThesaurusOK() should be called first.
------------------------------------------------------------------------*/
typeErr         
SpWhatThesaurusData ( language, thesDataP)
int                 language;           /* Language for which we want thesaurus
                                            information. */
int                 *thesDataP;         /* OUT: Proximity flags for thesaurus */
{
    typeErr         status;             /* Subroutine retcode */

    spClearErrors();                    /* Clear out errors */

    /* Set the language and bring up the thesaurus if necessary */
    if ((status = spOpenThesaurus( language)) != AOK) return( status);

    *thesDataP = thstatus();                                            /* Px */
    return( AOK);
}

/*-----------------------------------------------------------------------
    Output the complete set of thesaurus definitions for the passed word.

    A thesaurus definition consists of a "meaning" (a brief dictionary
    definition of the word); plus synonyms, antonyms, etc. for each meaning.
    Note a single word can have multiple meanings; e.g. "point" can mean
    1) noun: a location, 2) verb: to indicate, etc.

    This function (or subfunctions below here) allocate the memory for
    the meanings and wordlists. The caller should...

    Some foreign languages have no thesaurus support, others only have
    synonyms.
------------------------------------------------------------------------*/
typeErr         
SpGetThesaurus (in_word, language, numMeaningsP, meanings, words)
mbchar              *in_word;           /* Word to find thesaurus defs for. */
int                 language;           /* Language we are working in */
int                 *numMeaningsP;      /* OUT: # of meanings being returned. */
mbchar              *meanings[MAXMEAN+1];   /* OUT: Meaning strings */
mbchar              **words[][MAXLISTS];    /* OUT: word lists (each meaning 
                                            has MAXLISTS arrays of words) */
{
    typeErr         status;             /* Subroutine retcode */
    int             i;                  /* Loop counter */
	static int		maxexp=-1;			/* Proximity MAXEXP */

    spClearErrors();                    /* Clear out errors */

    /* Enforce word length */
    if ( MBstrlen( in_word ) >= (unsigned)MAX_PROX_WORD) {
        MBstrcpy( SpErrorObject, in_word );
        return( SPELL_LONG_WORD);       /* Text string too long */
    }

    /* Set the language and bring up the thesaurus if necessary */
    if ((status = spOpenThesaurus( language)) != AOK) return( status);

    /*** Get the meanings ***/
    /* Make room for the meanings */
	if (maxexp < 0)
		maxexp = pub_MAXEXP();                                          /* Px */
    for (i=0; i<=MAXMEAN; i++) {
        meanings[i] = (mbchar *) malloc( (maxexp + 1) * sizeof( mbchar));
        VERIFY( meanings[i]);/* !!0 Throw out-of-memory err instead! */
    }

    /* Proximity "meaning numbers" are 1-based */
    /* Also note translations between Aster*x and Proximity character set 
        (for input word and output meanings) is done in the subroutine. */
    status = spGetThesaurusMeanings( in_word, 1, MAXMEAN+1, numMeaningsP, 
                                                                meanings);
    if (status != AOK) 
        goto cleanup;

    /* Proximity claims no word has more than MAXMEAN meanings. We tried to
        get MAXMEAN+1 meanings. If we ever succeed, MAXMEAN is defined
        incorrectly. Let's try to catch this in development; ignore the
        problem in the released product. */
    ASSERT( *numMeaningsP < MAXMEAN+1 );

    /* Free unused memory */
    for (i=MAXMEAN; i>=*numMeaningsP; --i) {
        free( meanings[i]);
        meanings[i] = NULL;
    }

    /*** Get the words for each meaning ***/

    for (i=0; i<*numMeaningsP; i++) {
        /* Proximity "meaning numbers" are 1-based */
        /* Also note translation to Aster*x character set is done in
            the subroutine. */
        status = spGetThesaurusWords ( i+1, words[i]);
        if (status != AOK)
          goto cleanup;
    }
    status = AOK;

cleanup:
    return( AOK);
}

/*-----------------------------------------------------------------------
    Output a set of thesaurus meanings for the passed word.
    The caller needs to pass an array ('meanings') large enough to
    hold 'numWant' strings, each of size MAXEXP+1.
------------------------------------------------------------------------*/
static typeErr 
spGetThesaurusMeanings (in_word, start, numWant, numGotP, meanings)
mbchar              *in_word;           /* Word to find thesaurus defs for. */
int                 start;              /* First meaning number we want; should
                                            be 1 at first call for 'word' */
int                 numWant;            /* Number of meanings we want to get */
int                 *numGotP;           /* OUT: Number of meanings we got. If 
                                            less than we wanted, all done. */
mbchar              **meanings;         /* OUT: Array is filled w/meaning
                                            strings */
{
    int             i;                  /* Loop counter */
    mbchar          *word;
    int             len_in_string;

    len_in_string = MBstrlen( in_word );
#if BUILD_WIDE
    if ( MBnum_wide_chars_n( in_word, len_in_string ) != 0 )
        {
        /* do not spell check a wide char word */
        *numGotP = 0;
        return( AOK );
        }
#endif

    if (!thesOpen) VERIFY(0);   /* Disaster; "can't happen" */

    /* Enforce word length */
    if ( len_in_string >= MAX_PROX_WORD) {
        MBstrcpy( SpErrorObject, in_word );
        return( SPELL_LONG_WORD);       /* Text string too long */
    }

    /* make a copy so we can convert character sets */
    word = (mbchar *)malloc( (len_in_string+1) * sizeof(mbchar) );
    MBstrcpy( word, in_word );

    /* Convert the word to the Proximity character set */
    spXlateString( word, AxToProx);

    *numGotP = thesword( (UCHAR *)word, start, numWant, (UCHAR **)meanings);/* Px */

    if (*numGotP == PRX_ERROR) {
        /* Convert the word BACK to the Aster*x character set so we
            can give a good error. */
        MBstrcpy(SpErrorObject, in_word);
        sprintf( SpErrorOverrideStr, 
    XLT("Proximity problem in thesaurus accessing %d meanings starting at meaning %d for word"),
                start, numWant, SpErrorObject);
        return( PROX_THES_MEANINGS);
    }


    /* Convert the meaning strings to the Aster*x character set */
    for (i=0; i<*numGotP; ++i) {
        spXlateString( meanings[i], ProxToAx);
    }

    free( word );
    return( AOK);
}

/*-----------------------------------------------------------------------
    Output a set of synonyms, antonumns, etc. for the passed meaning 

    Has to be used in synch with spGetThesaurusWords(). That is, the
    caller must first call spGetThesaurusWords() to get a set of
    meanings for a given word before calling this functions to get
    the set of words associated with one of those meanings. This
    rule is due to the method in which Proximity caches thesaurus
    data.
------------------------------------------------------------------------*/
static typeErr  
spGetThesaurusWords ( meaningCode, wordArray)
int                 meaningCode;            /* Meaning number from thesword() */
mbchar              **wordArray[MAXLISTS];  /* OUT: Array of MAXLISTS arrays of
                                                words */
{
    mbchar          *c;                 /* Misc char ptr */
    mbchar          **wordPtr;          /* Ptr to array of words */
    mbchar          *currWord;          /* Current word being examined. */
    int             i,j;                /* Loop counters */
    mbchar          **myWordList;       /* Private copy of Prox wordlist */
    mbchar          *leftOvers;         /* Multiple-words pointer */
    bool            proxWord;           /* TRUE -> word being split was alloc'd
                                            by Proximity */

    if (!thesOpen) VERIFY(0);   /* Disaster; "can't happen" */

    /* Here is the structure of the "wordArray" returned by Proximity:

        "wordArray" is a pointer to an array of MAXLISTS pointers (for 
        synonyms, comparisons, "see also's", contrasts, and antonyms).
        Proximity's constant MAXLISTS is 5, of course.

        Each of the 5 pointers points to a wordlist. A wordlist is a
        null-terminated array of pointers; each pointer points to
        an allocated string representing a word. */

    c = (mbchar *) thesinfo( meaningCode, (UCHAR ***)wordArray );        /* Px */

    /* Certain "words" are actually groups of words; they must be split up
        into individual words. For example:
            "drew on/draw on"       becomes "drew on" and "draw on"
            "get/got or gotten"     becomes "get" and "got" and "gotten" */

    /* We do this as follows: For each of the five Proximity wordlists, 
        we create a private wordlist. For each word in a given wordlist,
        if it is a single word, our private wordlist will point to it.
        If it is a group of words, we split it up into single words,
        each of which gets an individual pointer (so our wordlist will
        become longer than Proximity's). Then we free Proximity's wordlist
        (but NOT the pointers in that list) and reset the pointer in the
        5-member array to point to our wordlist. */

    /* For each of the 5 wordlists */
    for (i=0,j=0; i<MAXLISTS; i++,j=0) {
        myWordList = (mbchar **) malloc( sizeof(char *)); 
        wordPtr = wordArray[i];

        /* For each word in the wordlist */
        while (*wordPtr) {
            currWord = *wordPtr;
            leftOvers = *wordPtr;       /* Any non-null value will do */
            proxWord = TRUE;

            /* Convert the word to the Aster*x character set */
            spXlateString( currWord, ProxToAx);

            /* For each word that the Proximity word splits into (usually
                just one, but sometimes more) */
            while( leftOvers) {

                /* Make space for another pointer in private wordlist */
                myWordList = (mbchar **) realloc( myWordList, 
                                        (j+1) * sizeof(char *)); 
                VERIFY( myWordList);/* !!0 Throw out-of-memory err instead! */

                /* Stuff a pointer to the word into the private wordlist */
                myWordList[j++] = spSplitWord( currWord, proxWord, &leftOvers);
                proxWord = FALSE;
                currWord = leftOvers;
            }
            ++wordPtr;          /* Next Proximity word */
        }

        /* Put a NULL pointer at the end of the list */
        myWordList = (mbchar **) realloc( myWordList, (j+1) * sizeof(mbchar *)); 
        VERIFY( myWordList);/* !!0 Throw out-of-memory err instead! */
        myWordList[j] = NULL;

        /* Replace the Proximity wordlist with the private one */
#if 0
        nzfree( wordArray[i]);                                          /* Px */
#endif
        wordArray[i] = myWordList;
    }
    return( AOK);
}

/*-----------------------------------------------------------------------
    Open the thesaurus in the indicated language, if necessary.

    It is not an error if this function is called and the Speller
    happens to already be in the right language, and the thesaurus
    is open.
------------------------------------------------------------------------*/
static typeErr  
spOpenThesaurus ( language)
int                 language;           /* Desired language */
{
    char            *dictionaries[1];   /* Empty list of dictionaries */
    typeErr         status;             /* Subroutine retcode */
    ed_spell_files  local;              /* Local filenames */
    int             proxErr;            /* Proximity retcode */

    /* Set the speller state to the proper language */
    /* Note we don't care what dictionaries are on the search path during 
        this operation. */
    dictionaries[0] = NULL;
    if ((status = spSetState( language, dictionaries, FALSE, TRUE )) != AOK) {
        return( status);
    }

    /* Bring up the thesaurus if necessary */
    if (!thesOpen) {

        /* Verify the thesaurus file is available */
        spSpellFilenames( language, "", &local);
        if (!strlen( spellFiles.thesFile)) {
            MBstrcpy( SpErrorObject, local.thesFile);
            return( FEENOENT);              /* No such file or directory */
        }
        if ((status = spAccessFile( spellFiles.thesFile, R_OK)) != AOK) {
            return( status);                /* No read permission */
        }

        /* The thesaurus file is opened as a Linguibase. However it is
            not placed on the searchpath, because we are only using it
            in the context of the thesaurus (i.e. it represents the list
            of words for which we have thesaurus definitions). */
        thesPtr = (LEX *)lexalloc(spellFiles.thesFile, 0);                     /* Px */
        if (!thesPtr ) {
            MBstrcpy( SpErrorObject, spellFiles.thesFile);
            return( PROX_BAD_FILE);         /* Prox says bad file (?) */
        }

        /* Now read the thesaurus-specific information (which is hiding
            after the normal Linguibase information) into memory. */
        proxErr = thesopen( thesPtr);                                   /* Px */
        if (!thesPtr) {         /* 0 -> failure, positive -> success */
            MBstrcpy( SpErrorObject, spellFiles.thesFile);
            return( PROX_BAD_FILE);         /* Prox says bad file (?) */
        }

        /* Set the language-specific Proximity function pointers */
        setths( language );                                             /* Px */

        /* Request return of all possible thesaurus information including
            all parts of speech (adjectives, nouns, etc.), all word list
            types (synonyms, antonyms, etc.). Also put the returned
            words into the same case, tense, etc. as the passed word. */
        thescntrl( TH_ALLPOS | TH_ALLLIST | TH_INFALL );                /* Px */

        thesOpen = TRUE;
    }
    return( AOK);
}

/*-----------------------------------------------------------------------
    Close up the thesaurus, if open.
------------------------------------------------------------------------*/
typeErr         
SpCloseThesaurus ( )
{
    /* If not open, no work to do */
    if (!thesOpen) {
        ASSERT( 0);             /* Shouldn't be here in development */
        return( AOK);
    }

    thesclose();                                                        /* Px */
    if (!thesPtr) {
        ASSERT( 0);             /* Shouldn't be here in development */
        return( AOK);           /* Shouldn't happen but we can recover */
    }
    lexfree( thesPtr);                                                  /* Px */
    thesPtr = NULL;
    thesOpen = FALSE;
    return( AOK);
}

/*------------------------------- spSplitWord --------------------------
    Given a word from the Proximity Thesaurus, split it into multiple
    words if necessary and return them individually. For example,
        "drew on/draw on"       becomes "drew on" and "draw on"
        "get/got or gotten"     becomes "get" and "got" and "gotten"

    The caller passes a pointer to the word. If the word is a single
    word, this function simply mallocs sufficient memory, copies the
    word in and returns its pointer (and 'leftOvers' is set NULL). 
    However, if the passed word contains multiple words, the first 
    word is extracted into newly-malloc'ed memory and returned. 
    The remainder of the word is extracted (also into newly-malloc'ed 
    memory) and output as 'leftOvers'. The caller should call this
    function again, passing 'leftOvers' as the word in question, until
    'leftOvers' is set NULL. 
-------------------------------------------------------------------------*/
static mbchar   *
spSplitWord ( word, proxWord, leftOvers)
mbchar              *word;          /* Current word */
bool                proxWord;       /* TRUE -> 'word' was allocated by
                                        Proximity */
mbchar              **leftOvers;    /* OUT: Pointer to remainder of word */
{
    int             len;            /* Overall word(s) length */
    int             i;              /* Loop counter */
    mbchar          *newWord;       /* Word to return */
    int             sepLen=0;       /* Number of characters spearating one
                                        word from another; 1 if "word/word2", 
                                        4 if "word or word2" */

    /* !!1 Note: this function only frees words that it previously 
        allocated.  It does NOT free words allocated by Proximity; 
        this seemed to lead to problems. I have no idea whether this 
        will lead to a memory leak, or whether Proximity cleans up 
        after itself on subsequent Thesaurus calls. */

    *leftOvers = NULL;
    len = MBstrlen( word);
    i = 0;
    newWord = word;
    while (i<len) {

        /* If we see a slash, return the preceeding word and skip over 
            the slash */
        if (word[i] == '/') {
            sepLen = 1;
            break;
        }
        /* Else if we see the separation string, do the same thing */
        /* !!1 Is " or " used in every language? If not, see ThesOrWord[]
            and ThesOrLen[] in dc/spell/thesaurus.c in Aster*x 2.0 for
            a per-language-specific strategy */
        /* !!0 Use wpMBstrncmp() when available */
        else if ((i + 4 < len) && !MBstrncmp( &(word[i]), NXLT(" or "), 4)) {
            sepLen = 4;
            break;
        }
        ++i;
    }

    /* Single word can survive */
    if (!sepLen) {
        newWord = (mbchar *) malloc( (len + 1) * sizeof(mbchar));
        VERIFY( newWord);/* !!0 Throw out-of-memory err instead, please! */
        MBstrcpy( newWord, word);
        if (!proxWord) 
            free( word);
        return( newWord);
    }

    /* Else isolate first word... */
    newWord = (mbchar *) malloc( (i+1) * sizeof(mbchar));
    VERIFY( newWord);/* !!0 Throw out-of-memory err instead, please! */
    /* !!0 use wpstrncpy() when available */
    MBstrncpy( newWord, word, i);
	newWord[i] = '\0';

    /* ...and rest of word beyond separation characters */
    *leftOvers = (mbchar *) malloc( (len - i - sepLen + 1) * sizeof(mbchar));
    VERIFY( *leftOvers);/* !!0 Throw out-of-memory err instead, please! */
    MBstrcpy( *leftOvers, word + i + sepLen);
    if (!proxWord) free( word);
    return( newWord);
}

/* }***********************************************************************/

/************************** Dictionary functions *********************** {*/

/*-----------------------------------------------------------------------
    Return whether we can add another dictionary file
------------------------------------------------------------------------*/
bool            
SpCanAddAnotherDictionary ( )
{
    if (nDictFiles < MAXNDICTS) return( TRUE);
    return( FALSE);
}

/*-----------------------------------------------------------------------
    Add the passed dictionaries to the search path.

    If a dictionary does not exist, we shut down the speller and
    return an error. We attempt to open the dictionary in read/write 
    mode; if that is not possible, it is opened in read-only mode. 
    If too many dictionaries are specified, we will return an error.

    This function will not create new dictionaries. Instead, the caller
    SpCreateDictionary() to create new, empty dictionaries.

    The dictionaries currently on the search path are NOT removed. If
    desired, this can be done the caller before calling this function
    (via spRemoveAllDictionaries).

    Assumption: The rest of the search path has been specified; this is
    the last operation before the speller is considered to be "up".
------------------------------------------------------------------------*/
static typeErr  
spAddDictionaries ( dictionaries)
char                **dictionaries;     /* NULL-terminated array of user
                                            dictionary filenames to use */
{
    int             i;                  /* Loop counter */
    typeErr         status;             /* Subroutine retcode */

    /* New document - initialize WP global properly */
    if (!dictionaries) {
        return( AOK);
    }

    for (i=0; dictionaries[i]; ++i) {

        /* !!1 If the caller is ever interested in WHICH dictionary the
            word was found in, the caller should probably pass a code
            for each dictionary. Using the array index for the code is
            a bit lame, because if the caller ever gets the ability to
            add and remove dictionaries on the fly, the caller will soon
            lose track of what (speller) index number corresponds to
            which dictionary. */

        status = spAddDictionary( dictionaries[i], i);
        if (status != AOK) {
            SpCloseSpeller( );              /* Shut down the speller */
            MBstrcpy( SpErrorObject, dictionaries[i]);
            return( status);
        }
    }
    return( AOK);
}


/*-----------------------------------------------------------------------
    Attempt to add another dictionary to the search path.

    See comments for spAddDictionaries() above.
------------------------------------------------------------------------*/
static typeErr  
spAddDictionary ( dfile, code)
char                *dfile;             /* Filename of dictionary */
int                 code;               /* Code number for dictionary */
{
    int             ix;                 /* File index */
    bool            writeFlag;          /* TRUE -> open it read/write */
    int             flags;              /* Proximity flags */
    CLAM            *clamptr;           /* Proximity dictionary pointer */
    typeErr         errcode;

    ASSERT( code < IN_STOPCACHE);       /* Don't collide w/reserved #'s */

	/* In debugging, catch empty string in dictionary list.
		In released code, just ignore it. */
	ASSERT( *dfile);		/* Our sole caller won't pass NULL */
	if (!*dfile)
		return( AOK);		/* Yeah, sure, I addded your 'dictionary'! ;-) */

    /* While debugging, catch cases where dictionary is already open.
        In released code, this is harmless so let it go. */
    ix = spDictListIndex ( dfile);
    ASSERT( ix == NODICT);
    if (ix != NODICT) return( AOK);

    if (!SpCanAddAnotherDictionary() ) {
        return( SPELL_TOO_MANY_DICTS);
    }

    /* If we can't read the dictionary, report the error */
    if ((errcode = spAccessFile( dfile, R_OK)) != AOK)
        return( errcode);

    /* If we have read/write access, attempt to open it as such. */
    writeFlag = FALSE;
    if ((errcode = spAccessFile( dfile, W_OK)) == AOK) writeFlag = TRUE;

    /* Tell Proximity about the dictionary */
    flags = SP_DETECT | SP_CORRECT | SP_HYPHEN;
    if (!writeFlag) flags = flags | CL_READONLY;
    clamptr = (CLAM *)clamopen( code, flags, dfile);                            /* Px */
    if (!clamptr) {
        MBstrcpy( SpErrorObject, dfile);
        return( PROX_BAD_DICT);
    }

    /* If we attempted to open a dictionary read/write but got it as
        read-only, note this fact but don't throw an error till the
        user tries to alter the dictionary */
    if (writeFlag && pub_cl_rdonly( clamptr)) writeFlag = FALSE;        /* Px */

    /* Put dictionary at the end of the list */
    MBstrcpy( dictFile[nDictFiles].file, dfile);
    dictFile[nDictFiles].code = code;
    dictFile[nDictFiles].write = writeFlag;
    dictFile[nDictFiles].clamptr = clamptr;
    ++nDictFiles;
    return( AOK);

}


/*-----------------------------------------------------------------------
    Return the language code for a dictionary file.

    Returns zero if the file does not exist, or is corrupt, etc.
------------------------------------------------------------------------*/
int             
SpDictionaryLanguage ( dict)
char                *dict;              /* Dictionary in question */
{
    typeErr     err;
    int         lang;

    err = spAccessFile( dict, R_OK);
    if (err != AOK)
        return( 0);

    lang = clamlanguage( dict);
    return( lang);
}


/*-----------------------------------------------------------------------
    Creates a new, empty dictionary.

    If the dictionary already exists, this is reported as an error.  
    If the dictionary cannot be created, this is reported as an error.

    When this function exits, the new dictionary is NOT on the
    search path. The caller can issue some other spelling command
    that includes this dictionary in the "state" information,
    however.

    It is a Proximity requirement that Proximity be up and running
    (i.e. in a some language environment) when this action occurs.
    Thus the caller should specify some legal language (for which
    the language data files exist). Proximity is placed in that
    language if it is not already there.
------------------------------------------------------------------------*/
typeErr         
SpCreateDictionary ( language, dict)
int                 language;           /* Some legal language */
char                *dict;              /* Dictionary to create */
{
    char            *noDicts[1];        /* Empty list of dictionaries */
    int             flags;              /* Proximity flags */
    CLAM            *clamptr;           /* Proximity dictionary CLAM pointer */
    int             proxErr;            /* Proximity error */
    typeErr         errcode;

    spClearErrors();                    /* Clear out errors */

    /* Set the speller state to the proper language if it is not
        already there. Ignore dictionaries completely, for now... */
    if (!started || (currLang != language)) {
        noDicts[0] = NULL;
        if ((errcode = spSetState( language, noDicts, FALSE, TRUE )) != AOK) {
            return( errcode);
        }
    }

    if (!SpCanAddAnotherDictionary() ) {

        /* !!1 Well, to be nice, we could temporarily remove the last
            dictionary from the search list and restore it at the end... */

        return( SPELL_TOO_MANY_DICTS);
    }

    /* If dictionary already exists, that's an error */
    if ((errcode = spAccessFile( dict, F_OK)) == AOK) {
        MBstrcpy( SpErrorObject, dict);
        return( FEEEXIST);
    }

    /* Prove that we can open the file for writing */
    if ((errcode = spAccessFile( dict, W_OK)) != AOK) {
        return( errcode);
    }

    /* Have Proximity create the dictionary. */
    flags = CL_INIT;
    clamptr = (CLAM *)clamopen( 0, flags, dict);                                /* Px */
    if (!clamptr) {
        MBstrcpy( SpErrorObject, dict);
        return( PROX_NO_CREATE_DICT);       /* who knows why? */
    }

    /* Now close it (note it is NOT in the current list of dictionaries) */
    proxErr = clamclose( clamptr);                                      /* Px */
    if (proxErr != PRX_OKAY) {
        MBstrcpy( SpErrorObject, dict);
        return( PROX_BAD_DICT_CLOSE);
    }
    return( AOK);
}

/*-----------------------------------------------------------------------
    Remove ALL the dictionaries from the search path.

    Assumption: Global array of open dictionaries is packed with
    no gaps.

    Assumption: Subroutine spRemoveDictionary() decrements global
    "number of dictionaries" variable.
------------------------------------------------------------------------*/
static typeErr  
spRemoveAllDictionaries ( )
{
    typeErr         status;             /* Subroutine return code */

    while( nDictFiles) {
        status = SpRemoveDictionary( dictFile[nDictFiles-1].file);
        if (status != AOK) return( status);
    }
    return( AOK);
}

/*-----------------------------------------------------------------------
    Remove a dictionary from the search path.
------------------------------------------------------------------------*/
typeErr         
SpRemoveDictionary ( dfile)
mbchar              *dfile;             /* Filename of dictionary */
{
    int             ix=0;               /* File index */
    int             i;                  /* Loop counter */

    /* Verify that the file is being used */
    /* If not, ignore for general release; catch this during development */
    ix = spDictListIndex ( dfile);
    ASSERT( ix != NODICT);
    if (ix == NODICT) return( AOK);

    i = clamclose( dictFile[ix].clamptr);                               /* Px */
    if (i != PRX_OKAY) {
        MBstrcpy( SpErrorObject, dfile);
        return( PROX_BAD_DICT_CLOSE);
    }

    /* Tighten up the dictionary list */
    for( i=ix+1; i<nDictFiles; ++i) {
        MBstrcpy( dictFile[i-1].file, dictFile[i].file);
        dictFile[i-1].code = dictFile[i].code;
        dictFile[i-1].write = dictFile[i].write;
        dictFile[i-1].clamptr = dictFile[i].clamptr;
    }
    --nDictFiles;
    MBstrcpy( dictFile[nDictFiles].file, "");
    dictFile[nDictFiles].code = 0;
    dictFile[nDictFiles].write = 0;
    dictFile[nDictFiles].clamptr = NULL;
    return( AOK);
}


/*-----------------------------------------------------------------------
    Add a word or a misspelling pair to a dictionary.

    A pair of words is taken as input. If the second word is NULL or empty,
    the first is taken as a "good" word representing a legal spelling.
    If the second word contains a string, the first word is considered
    to represent a misspelling and the the second word the associated 
    correction. If the user requests a list of alternative spellings for 
    the misspelled word, the correction as stored will be reported as one 
    of the alternatives (at or near the top of the list).

    If we do not have or cannot obtain write access to the dictionary,
    an error is thrown.

    Proximity says: "Only one correction can be associated with a given
    misspelling. If a pair is added which has a misspelling that is used
    in another pair, the second pair will replace the first."
------------------------------------------------------------------------*/
typeErr         
SpAddWord ( in_wordP, in_correct, language, dfile)
mbchar              *in_wordP;              /* Word to add. */
                                        /* If 'correct' is NULL or empty,
                                            this is a "good" word. But if
                                            a 'correct' word was specified,
                                            this is the misspelled word. */
mbchar              *in_correct;            /* Associated correction to add 
                                            (if specified, makes 'wordP' the
                                            misspelled word) */
int                 language;           /* Language we are working in 
                                            (i.e. appropriate for 'wordP') */
mbchar              *dfile;             /* Filename of dictionary */
{
    mbchar          *dictionaries[2];   /* 1-member list of dicts */
    typeErr         tError;             /* Subroutine err code */
    int             ix;                 /* Which dictionary is this? */
    int             status;             /* Proximity status code */
    int             flags;              /* Proximity word flag */
    mbchar          flagWord[LONGWORD]; /* Flagged word */
    mbchar          word[MAX_PROX_WORD+1];    /* unescaped word to add */
    mbchar          *wordP, *correction;
    int             ret_val;
    int             in_word_len;

    spClearErrors();                    /* Clear out errors */

    /* Set the speller state to the proper language, and cause the specified
        dictionary to be placed on the search path if it is not already 
        there. */
    /* Note we don't care what other dictionaries are on the search path
        during this operation. */
    dictionaries[0] = dfile;
    dictionaries[1] = NULL;
    if ((tError = spSetState( language, dictionaries, FALSE, TRUE )) != AOK) {
        return( tError);                /* oh horrors! */
    }

    /* Verify dictionary is open and writable */
    /* If we weren't able to open it writable, we can't proceed. */
    ix = spDictListIndex ( dfile);
    ASSERT( ix != NODICT);      /* Impossible; should have gotten error above */
    if( (ix == NODICT) || !dictFile[ix].write) {
        MBstrcpy( SpErrorObject, dfile);
        return( PROX_DICT_WRITE);
    }

    /* make a copy of the word */
    in_word_len = MBstrlen( in_wordP );
    wordP = (mbchar *)malloc( (in_word_len+1) * sizeof(mbchar) );
    MBstrcpy( wordP, in_wordP );

    if (in_correct && (in_correct[0] != '\0'))
       {
       /* make a  copy */
       correction = (mbchar *)malloc( (MBstrlen(in_correct)+1) * sizeof(mbchar) );
       MBstrcpy( correction, in_correct );
       }
    else
      correction = (mbchar *)NULL;

    if (correction && (correction[0] != '\0')) { /* misspelling + correction */

        /* un-escape escape sequences, checking for errors */
        if (tError = spUnescapeHyphPoints(word,correction)) return(tError);

        /* Enforce word length */
        if ( in_word_len >= MAX_PROX_WORD) {
            MBstrcpy( SpErrorObject, in_wordP );
            return( SPELL_LONG_WORD);       /* Text string too long */
        }

        /* Convert the word, and correct word to the Proximity character set */
        spXlateString( word, AxToProx);
        spXlateString( wordP, AxToProx);

        status = clampair( (UCHAR *)wordP, (UCHAR *)word, dictFile[ix].clamptr);/* Px */

        if (status != PRX_OKAY) {
            MBstrcpy( SpErrorObject, dfile);
            sprintf( SpErrorOverrideStr, 
    XLT("Proximity problem adding misspell '%s' correction '%s' to dictionary"),
                wordP, correction);
            ret_val = PROX_ADD_DICT_PAIR;
            goto cleanup;
        }
    }
    else { /* correct word only */

        /* un-escape escape sequences, checking for errors */
        if (tError = spUnescapeHyphPoints(word,wordP)) return(tError);

        /* Convert the word to the Proximity character set */
        spXlateString( word, AxToProx);

        flags = doflags( (UCHAR *)word, (UCHAR *)flagWord, MAX_PROX_WORD ); /* Px */
        if (flags == PRX_ERROR) {
            MBstrcpy( SpErrorObject, in_wordP );
            ret_val = PROX_FLAG_WORD;
            goto cleanup;
        }

        flags = flags & 0xff;                   /* Take low 8 bits only */
        status = clamadd( (UCHAR *)flagWord, flags, dictFile[ix].clamptr ); /* Px */
        if (status == PRX_ERROR) {
            MBstrcpy( SpErrorObject, dfile);

            sprintf( SpErrorOverrideStr, 
                XLT("Proximity problem adding word '%s' to dictionary"), wordP);
            ret_val = PROX_ADD_DICT_WORD;
            goto cleanup;
        }
    }
    ret_val = AOK;

cleanup:
    free( wordP );
    if ( correction != (mbchar *)NULL )
        free( correction );

    return( ret_val );
}


/*-----------------------------------------------------------------------
    Remove a word or misspelling pair from a dictionary

    A pair of words is taken as input. If the second word is NULL or empty,
    the first is taken as a "good" word representing a legal spelling;
    this word is removed from the indicated dictionary. If the second word 
    contains a string, the first word is considered to represent a misspelling 
    and the the second word the associated correction; this misspelling/
    correction pair is removed from the dictionary.

    If we do not have or cannot obtain write access to the dictionary,
    an error is thrown.

    Proximity says: "To delete a misspelling pair, call the clampair()
    with _correction_ set to a null string ("", not NULL)."
------------------------------------------------------------------------*/
typeErr         
SpRemoveWord ( in_wordP, in_correct, language, dfile)
mbchar               *in_wordP;              /* Word to remove. */
                                        /* If 'correct' is NULL or empty,
                                            this is a "good" word. But if
                                            a 'correct' word was specified,
                                            this is the misspelled word. */
mbchar               *in_correct;            /* Associated correction to delete 
                                            (if specified, makes 'word' the
                                            misspelled word) */
int                 language;           /* Language we are working in 
                                            (i.e. appropriate for 'word') */
mbchar              *dfile;             /* Filename of dictionary */
{
    int             ix;                 /* Which dictionary is this? */
    int             status;             /* Proximity status code */
    mbchar          *dictionaries[2];   /* 1-member list of dicts */
    typeErr         tError;             /* Subroutine err code */
    int             flags;              /* Proximity word flag */
    mbchar          flagWord[LONGWORD]; /* Flagged word */
    mbchar          word[MAX_PROX_WORD+1];    /* unescaped word to remove */
    mbchar          *wordP;
    mbchar          *correction;
    int             ret_val;

    spClearErrors();                    /* Clear out errors */

    /* Set the speller state to the proper language, and cause the specified
       dictionary to be placed on the search path if it is not already there. */
    /* Note we don't care what other dictionaries are on the search path
        during this operation. */
    dictionaries[0] = dfile;
    dictionaries[1] = NULL;
    if ((tError = spSetState( language, dictionaries, FALSE, TRUE )) != AOK) {
        return( tError);                /* oh horrors! */
    }

    /* Verify dictionary is open and writable */
    /* If we weren't able to open it writable, we can't proceed. */
    ix = spDictListIndex ( dfile);
    VERIFY( ix != NODICT);      /* Impossible; should have gotten error above */
    if( !dictFile[ix].write) {
        MBstrcpy( SpErrorObject, dfile);
        return( PROX_DICT_WRITE);
    }

    /* make a narrow copy of the word */
    wordP = (mbchar *)malloc( (MBstrlen(in_wordP)+1) * sizeof(mbchar) );
    MBstrcpy(wordP, in_wordP);

    if (in_correct && (in_correct[0] != '\0'))
       {
       /* make a narrow copy */
       correction = (mbchar *)malloc((MBstrlen(in_correct)+1) * sizeof(mbchar));
       WToMbStr( in_correct, correction );
       }
    else
      correction = (mbchar *)NULL;

    if (correction && (correction[0] != '\0')) {

        /* un-escape escape sequences, checking for errors */
        if (tError = spUnescapeHyphPoints(word,correction)) return(tError);

        /* Convert the word, and correct word to the Proximity character set */
        spXlateString( word, AxToProx);

        status = clampair( (UCHAR *)word, (UCHAR *)"", dictFile[ix].clamptr);/* Px */
        if (status != PRX_OKAY) {
            MBstrcpy( SpErrorObject, dfile);
            sprintf( SpErrorOverrideStr, 
    XLT("Proximity problem removing misspell '%s' correction '%s' from dictionary"),
                wordP, correction);
            ret_val = PROX_DEL_DICT_PAIR;
            goto cleanup;
        }
    }
    else {
        /* un-escape escape sequences, checking for errors */
        if (tError = spUnescapeHyphPoints(word,wordP)) return(tError);

        /* Convert the word to the Proximity character set */
        spXlateString( word, AxToProx);

        flags = doflags( (UCHAR *)word, (UCHAR *)flagWord, MAX_PROX_WORD ); /* Px */
        if (flags == PRX_ERROR) {
            WToMbStr( in_wordP, SpErrorObject );
            ret_val = PROX_FLAG_WORD;
            goto cleanup;
        }
        flags = 0;                  /* Implies deletion of word */

        status = clamadd( (UCHAR *)flagWord, flags, dictFile[ix].clamptr ); /* Px */
        if (status == PRX_ERROR) {
            MBstrcpy( SpErrorObject, dfile);

            sprintf( SpErrorOverrideStr, 
                XLT("Proximity problem removing word '%s' from dictionary"), wordP);
            ret_val = PROX_DEL_DICT_WORD;
            goto cleanup;
        }
    }
    ret_val = AOK;
cleanup:
    free( wordP );
    if ( correction != (mbchar *)NULL )
        free( correction );

    return( ret_val );
}


/*-----------------------------------------------------------------------
    Retrieve (another) set of entries from indicated dictionary.

    The caller states how many entries he wants. The number of entries 
    actually being returned is output, and an array is filled with 
    those entries. The caller should declare the array as 
    "mbchar  *words[# words wanted]" (see below).

    An "entry" from the dictionary either consists of a single word
    (representing what we claim is a legal word) or a "misspelling
    pair", representing a misspelled word and its preferred correct
    spelling. In the latter case, we use the format "misspell;correct"
    to return the pair to the caller.

    If the number of entries being reported is set to zero, or fewer
    than the caller requested, there are no more entries in the dictionary.

    Assumption: If the caller says he wants "maxGet" entries, the passed
    array can hold at least that number of entries. This function allocates
    the memory for the words; the passed array contains pointers to
    the strings. The caller should free the strings.
------------------------------------------------------------------------*/
typeErr         
SpGetDictWords ( language, dfile, firstTime, 
                                                    maxGet, gottenP, words)
int                 language;           /* Language we are working in 
                                            (i.e. appropriate for 'word') */
                                        /* !!0 is language needed? */
mbchar              *dfile;             /* Filename of dictionary */
bool                firstTime;          /* TRUE -> start at beginning of dict */
int                 maxGet;             /* Maximum number of words wanted. */
int                 *gottenP;           /* OUT: Number of words being returned*/
mbchar              **words;            /* OUT: Caller's word array */
{
    mbchar          *dictionaries[2];   /* 1-member list of dicts */
    typeErr         tError;             /* Subroutine err code */
    int             ix;                 /* Which dictionary is this? */
    bool            firstTimeThru;      /* TRUE -> start at beginning */
    int             i;                  /* Loop counter */
    int             status;             /* Proximity status code */
    mbchar          flaggedWord[ LONGWORD]; /* Latest retrieved word */
    int             flag;               /* Flags associated w/"flaggedWord" */
    mbchar          unflaggedWord[ LONGWORD];   /* Latest unflagged word */
    mbchar          correction[ LONGWORD]; /* Correct word of misspell pair */

    spClearErrors();                    /* Clear out errors */

    /* Set the speller state to the proper language, and cause the specified
        dictionary to be placed on the search path if it is not already 
        there. */
    /* Note we don't care what other dictionaries are on the search path
        during this operation. */
    dictionaries[0] = dfile;
    dictionaries[1] = NULL;
    if ((tError = spSetState( language, dictionaries, FALSE, TRUE )) != AOK) {
        return( tError);                /* oh horrors! */
    }

    /* Verify dictionary is open */
    ix = spDictListIndex ( dfile);
    VERIFY( ix != NODICT);      /* Impossible; should have gotten error above */

    firstTimeThru = firstTime;
    *gottenP = 0;
    for (i=0; i<maxGet; ++i) {
        status = clamnext( (UCHAR *)flaggedWord, &flag, firstTimeThru, 
                                                dictFile[ix].clamptr);  /* Px */
        if (status == PRX_ERROR) {
            MBstrcpy( SpErrorObject, dfile);
            return( PRX_READ_DICT);
        }

        else if ((status == CL_PAIR) || (status == PRX_OKAY)) {

            /* To support misspell pairs, look for ';' in the flaggedWord 
                that separates good/bad words, and call the undoflags for 
                the misspell part only. Also the caller should be prepared 
                to handle the resulting pair... */
            if (status == CL_PAIR) {
                mbchar          *semi;          /* Find the semicolon */
                semi = MBstrchr( (char *)flaggedWord, (int) ';');
                ASSERT( semi);                  /* Really should be there... */
                if (!semi) status = PRX_OKAY;   /* ...but handle if not */
                else {
                    MBstrcpy( correction, semi);       /* ";correction" */
                    *semi = '\0';
                }
            }

            /*  "Undo" the flagging operation to retrieve original word */
            undoflags( (UCHAR *)flaggedWord, flag, (UCHAR *)unflaggedWord );/* Px */

            /* To support misspell pairs, graft the correction back
                onto the undoflagged misspell word (separate with ';'
                just like Proximity does; caller will have to scan for
                this to dectect a misspell pair) */
            if (status == CL_PAIR) {
                MBstrcat( unflaggedWord, correction);  /* "misspell;correction" */
            }

            /* Convert the word to the Aster*x character set */
            spXlateString( unflaggedWord, ProxToAx);

            /* returns a word with escaped hyphenation points in
            malloc'ed memory. */
            words[i] = spEscapeHyphPoints(unflaggedWord);

            ++*gottenP;
        }

        /* No more words in dictionary */
        else if (status == CL_EOF) {
            break;
        }

        /* No longer the first time thru */
        firstTimeThru = FALSE;
    }

    return( AOK);
}



/*-----------------------------------------------------------------------
    Return the index for the passed dictionary, or NODICT if the
    dictionary is not open.
------------------------------------------------------------------------*/
static int      
spDictListIndex ( dfile)
char                *dfile;             /* Filename of dictionary */
{
    int             i;                  /* Loop counter */

    for( i=0; i<nDictFiles; ++i) {
        /* !!0 use wpstrcmp when available */
        if (!strcmp( dictFile[i].file, dfile)) {
            return( i);
        }
    }
    return( NODICT);
}

/* }***********************************************************************/

/******************** Word and string-related utilities **************** {*/


/*-----------------------------------------------------------------------
    Translate strings between the Aster*x and Proximity character sets.

    The string is converted in-place.

    Aster*x uses all 8 bits (i.e. 1-255) of the "ISO Latin-1" character 
    set, while Proximity uses the lower 7 bits properly (i.e. "standard
    ASCII") but has their own codes (between 128 and 255) for special 
    characters. See the Proximity manual for the chart.
------------------------------------------------------------------------*/
static void     
spXlateString ( string, table)
mbchar              *string;                /* String to translate */
mbchar              *table;                 /* xlation table to use */
{
    register mbchar *c;                     /* Current character */

    /* Sanity checks */
    ASSERT( string);            /* Choke on NULL during development... */
    if (!string) return;        /* ...but handle gracefully otherwise */

    c = string;
    while ( *c != '\0' ) {
        *c = table[(*c) & 0xFF];
        c++;
    }
}

/*-----------------------------------------------------------------------
    Translate character between the Aster*x and Proximity character sets;
    return the translated character.

    See spXlateString() above for more details
------------------------------------------------------------------------*/
int  
spXlateChar ( theChar, table)
int                 theChar;                /* Character to translate */
mbchar              *table;                 /* xlation table to use */
{
    /* Sanity checks */
    ASSERT( theChar);           /* Choke on NULL during development */

    return( table[ theChar & 0xFF]);
}

/*-----------------------------------------------------------------------
    Remove slashes from the given string (sometimes Proximity returns
    strings containing extraneous slashes; e.g. in German)
------------------------------------------------------------------------*/
static void     
spRemoveSlashes ( str)
char              *str;                   /* String in question */
{
    char          *c;

    /* !!0 Use wpstrrchr() when available */
    while((c = strrchr(str,'/')) != NULL) MBstrcpy(c,c+1);
}

/* }***********************************************************************/

/**************************** Generic utilities ************************ {*/

/*-----------------------------------------------------------------------
    Generate the full pathnames for the Proximity files for the passed
    language. 

    For each file, look first at the user's home Asterx directory and then in 
    the primary and secondary public directories for the desired files. 
    Output the pathname for the first matching file actually found. 
    Output an empty string if the file is not available.

    The names are placed in the passed ed_spell_files structure.
------------------------------------------------------------------------*/
static void     
spSpellFilenamesAvailable ( language, files)
int                 language;           /* Language in question */
ed_spell_files      *files;             /* OUT: Will contain filenames */
{
    ed_spell_files      tryFiles[3];    /* Filenames for local, primary,
                                            and 2ndary locations */
    char                *homedir;       /* User's home Asterx directory */
    char                *aosdir;        /* In effect, /ax/axexec */

    /* Initialize the return names */
    spSpellFilenamesInit( files);

    /* Generate the filenames */
    spSpellFilenamesInit( &tryFiles[0]);
    homedir = (char *) AGetAxHomeDir();
    spSpellFilenames( language, homedir, &tryFiles[0]);     /* Home dir */
    spSpellFilenamesInit( &tryFiles[1]);
    aosdir = (char *) ThimApplixDir;
    spSpellFilenames( language, aosdir, &tryFiles[1]);      /* primary */
    spSpellFilenamesInit( &tryFiles[2]);
    spSpellFilenames( language, ThimLangDir, &tryFiles[2]); /* secondary */

    /* Test them in order and record the first file that actually exists */
    if (spAccessFile( tryFiles[0].phonFile, F_OK) == AOK) 
        MBstrcpy( files->phonFile, tryFiles[0].phonFile);
    else if (spAccessFile( tryFiles[1].phonFile, F_OK) == AOK) 
        MBstrcpy( files->phonFile, tryFiles[1].phonFile);
    else if (spAccessFile( tryFiles[2].phonFile, F_OK) == AOK) 
        MBstrcpy( files->phonFile, tryFiles[2].phonFile);

    if ( spAccessFile( tryFiles[0].clxFile, F_OK) == AOK) 
        MBstrcpy( files->clxFile, tryFiles[0].clxFile);
    else if ( spAccessFile( tryFiles[1].clxFile, F_OK) == AOK) 
        MBstrcpy( files->clxFile, tryFiles[1].clxFile);
    else if ( spAccessFile( tryFiles[2].clxFile, F_OK) == AOK) 
        MBstrcpy( files->clxFile, tryFiles[2].clxFile);

    if ( spAccessFile( tryFiles[0].lexFile, F_OK) == AOK) 
        MBstrcpy( files->lexFile, tryFiles[0].lexFile);
    else if ( spAccessFile( tryFiles[1].lexFile, F_OK) == AOK) 
        MBstrcpy( files->lexFile, tryFiles[1].lexFile);
    else if ( spAccessFile( tryFiles[2].lexFile, F_OK) == AOK) 
        MBstrcpy( files->lexFile, tryFiles[2].lexFile);

    if ( spAccessFile( tryFiles[0].detFile, F_OK) == AOK) 
        MBstrcpy( files->detFile, tryFiles[0].detFile);
    else if ( spAccessFile( tryFiles[1].detFile, F_OK) == AOK) 
        MBstrcpy( files->detFile, tryFiles[1].detFile);
    else if ( spAccessFile( tryFiles[2].detFile, F_OK) == AOK) 
        MBstrcpy( files->detFile, tryFiles[2].detFile);

    if ( spAccessFile( tryFiles[0].hyFile, F_OK) == AOK) 
        MBstrcpy( files->hyFile, tryFiles[0].hyFile);
    else if ( spAccessFile( tryFiles[1].hyFile, F_OK) == AOK) 
        MBstrcpy( files->hyFile, tryFiles[1].hyFile);
    else if ( spAccessFile( tryFiles[2].hyFile, F_OK) == AOK) 
        MBstrcpy( files->hyFile, tryFiles[2].hyFile);

    if ( spAccessFile( tryFiles[0].thesFile, F_OK) == AOK) 
        MBstrcpy( files->thesFile, tryFiles[0].thesFile);
    else if ( spAccessFile( tryFiles[1].thesFile, F_OK) == AOK) 
        MBstrcpy( files->thesFile, tryFiles[1].thesFile);
    else if ( spAccessFile( tryFiles[2].thesFile, F_OK) == AOK) 
        MBstrcpy( files->thesFile, tryFiles[2].thesFile);
}

/*-----------------------------------------------------------------------
    Given the directory we expect the files to be in, generate the
    full pathnames for the Proximity files for the proposed language.

    For local files, pass the empty string (not a NULL) for the prefix.

    The names are placed in the passed ed_spell_files structure.
------------------------------------------------------------------------*/
static void     
spSpellFilenames ( language, prefix, files)
int                 language;           /* Language in question */
char                *prefix;            /* Directory for files (NOT NULL) */
ed_spell_files      *files;             /* OUT: Will contain filenames */
{
    int             i;
    char            slash[2];           /* Pathname separator */
	mbchar			reformed[2];		/* Indicates reformed spelling file */
	mbchar			*langName;			/* 3-letter Proximity language code */

    VERIFY( prefix);

    /* Append a slash to the end of a non-empty prefix */
    MBstrcpy( slash, "");
    i = strlen( prefix);
    if (i && (prefix[i-1] != '/')) MBstrcpy( slash, NXLT("/"));

	/* Proximity 6.1d and later supports several 'reformed' dictionaries
		(currently dut, grm, sgr). We allow the user to choose the old
		or reformed versions via profile vars. The naming convention
		for 'reformed' data files is as follows: If an old data file
		is 'xxx.ext', the reformed version is 'xxxr.ext' */
	reformed[0] = '\0';						/* Default to old */
	switch( language) {
        case DUTCH:
    		if (UPnumVal(PROF_WP_SPELL_NEW_DUT))
				strcpy( reformed, NXLT("r"));
				break;
        case GERMAN:
    		if (UPnumVal(PROF_WP_SPELL_NEW_GRM))
				strcpy( reformed, NXLT("r"));
				break;
        case SWISS_GERMAN:
    		if (UPnumVal(PROF_WP_SPELL_NEW_SGR))
				strcpy( reformed, NXLT("r"));
				break;
		default:
			break;
	}

	langName = (mbchar *)prx_langprf(language);				/* Px */
    sprintf( files->phonFile, NXLT("%s%s%s%s%s%s"), prefix, slash,
				langName, PHON_SUFFIX1, reformed, PHON_SUFFIX2);
    sprintf( files->clxFile, NXLT("%s%s%s%s%s"), prefix, slash,
				langName, reformed, CORELEX_SUFFIX);
    sprintf( files->lexFile, NXLT("%s%s%s%s%s"), prefix, slash,
				langName, reformed, LEX_SUFFIX);
    sprintf( files->detFile, NXLT("%s%s%s%s"), prefix, slash,
				langName, DETECT_SUFFIX);  /*no reform version*/
    sprintf( files->hyFile, NXLT("%s%s%s%s%s"), prefix, slash,
				langName, reformed, HY_SUFFIX);
    sprintf( files->thesFile, NXLT("%s%s%s%s%s"), prefix, slash,
				langName, reformed, SYNONYM_SUFFIX);
}

/*-----------------------------------------------------------------------
    Initialize filenames in file structure to empty string
------------------------------------------------------------------------*/
static void     
spSpellFilenamesInit ( files)
ed_spell_files      *files;             /* OUT: Will contain filenames */
{
    MBstrcpy( files->phonFile, "");
    MBstrcpy( files->clxFile, "");
    MBstrcpy( files->lexFile, "");
    MBstrcpy( files->detFile, "");
    MBstrcpy( files->hyFile, "");
    MBstrcpy( files->thesFile, "");
}

/*-----------------------------------------------------------------------
    Verify that we can either read or write the passed file, or that
    it exists.

    If so, return AOK. If not, set the global "error object" to the
    filename and return the appropriate error code.

    Note: if the caller requests write access, we return AOK if the file
    exists and is writable, or if it doesn't exist but its directory
    is writable.
------------------------------------------------------------------------*/
static typeErr  
spAccessFile ( file, accessFlag)
char                *file;              /* File in question */
int                 accessFlag;         /* F_OK, R_OK, or W_OK */
{
    typeErr         tError;             /* Get it? */

    VERIFY( file && (file[0] != '\0'));
    ASSERT( accessFlag == F_OK || accessFlag == R_OK || accessFlag == W_OK);

    /* Return AOK if we can access file as we expect */
    switch( accessFlag) {
        case W_OK:
            /* If file exists, see if we have write access. */
            if (!access( file, F_OK)) {
                if (!access( file, W_OK)) return( AOK);
                MBstrcpy( SpErrorObject, file);
                break;
            }
            /* Else prove directory exists and we have write access to IT. */
            else {
                char      *dir;
                int         len;
                char      *c;

                /* Extract directory name */
                len = strlen( file);
                dir = (char*)malloc( (len < 2 ? 3 : len + 1) * sizeof( char));
                VERIFY( dir);/* !!0 Throw out-of-memory err instead, please! */
                strcpy( dir, file);
                c = dir + len - 1;
                while ((*c != '/') && (c != dir)) --c;
                if (*c == '/') *c = '\0';
                else MBstrcpy( dir, NXLT("./"));

                /* OK if directory exists and is writable by us */
                if (!access( dir, W_OK)) return( AOK);
                MBstrcpy( SpErrorObject, dir);
                break;
            }

        default:
            if (!access( file, accessFlag)) return( AOK);
            MBstrcpy( SpErrorObject, file);
            break;
    }

    /* Return the appropriate error */
    tError = ErrnoErr();
    return( tError);
}

/*-----------------------------------------------------------------------
    Initialize the passed "spellcheck error report" mask; all values set
    to "don't report".
------------------------------------------------------------------------*/
void            
SpInitErrMask ( mask)
sp_error_mask       *mask;              /* OUT: Contains Initialized members */
{
    mask->unused0 = 0x0; mask->unused1 = 0x0; mask->unused2 = 0x0;
    mask->unused3 = 0x0; mask->unused4 = 0x0; mask->repeat = 0x0;
    mask->unused6 = 0x0; mask->number = 0x0; mask->date = 0x0;
    mask->time = 0x0; mask->nonAlph = 0x0; mask->misspell = 0x0;
    mask->startSent = 0x0; mask->roman = 0x0; mask->punctPre = 0x0;
    mask->punctSuff = 0x0; mask->outline = 0x0; mask->wordPS = 0x0;
    mask->prefix = 0x0; mask->wordWithPS = 0x0; mask->suffix = 0x0;
    mask->parens = 0x0; mask->badParens = 0x0; mask->needParens = 0x0;
    mask->goodCmpdPart = 0x0; mask->cmpd = 0x0; mask->badCmpd = 0x0;
    mask->goodCmpd = 0x0; mask->slash = 0x0; mask->unneededHyphen = 0x0;
    mask->contract = 0x0; mask->apos = 0x0; mask->misspellApos = 0x0;
    mask->hyphen = 0x0; mask->extraHyphen = 0x0; mask->hasHyphen = 0x0;
    mask->misspellHyphen = 0x0; mask->hasPosses = 0x0; mask->misspellPosses = 0x0;
    mask->aposPrefix = 0x0; mask->badAposPrefix = 0x0; mask->posses = 0x0;
    mask->badPosses = 0x0; mask->linkage = 0x0; mask->hyNumber = 0x0;
}

/*-----------------------------------------------------------------------
    Set the passed "spellcheck error report" mask to profile values
------------------------------------------------------------------------*/
void            
SpSetErrMaskProfile ( mask)
sp_error_mask       *mask;              /* OUT: All members are set */
{
    SpSetErrMaskValue(mask, 5, UPnumVal(PROF_WP_SPELL_TOK5));   /* repeat */
    SpSetErrMaskValue(mask, 7, UPnumVal(PROF_WP_SPELL_TOK7));   /* number */
    SpSetErrMaskValue(mask, 8, UPnumVal(PROF_WP_SPELL_TOK8));   /* date */
    SpSetErrMaskValue(mask, 9, UPnumVal(PROF_WP_SPELL_TOK9));   /* time */
    SpSetErrMaskValue(mask, 10, UPnumVal(PROF_WP_SPELL_TOK10)); /* alpha */
    SpSetErrMaskValue(mask, 11, UPnumVal(PROF_WP_SPELL_TOK11)); /* misspl */
    SpSetErrMaskValue(mask, 12, UPnumVal(PROF_WP_SPELL_TOK12)); /* Cap */
    SpSetErrMaskValue(mask, 13, UPnumVal(PROF_WP_SPELL_TOK13)); /* roman */
    SpSetErrMaskValue(mask, 14, UPnumVal(PROF_WP_SPELL_TOK14)); /* .pref */
    SpSetErrMaskValue(mask, 15, UPnumVal(PROF_WP_SPELL_TOK15)); /* .suff */
    SpSetErrMaskValue(mask, 16, UPnumVal(PROF_WP_SPELL_TOK16)); /* outlne */
    SpSetErrMaskValue(mask, 17, UPnumVal(PROF_WP_SPELL_TOK17)); /* presuf */
    SpSetErrMaskValue(mask, 22, UPnumVal(PROF_WP_SPELL_TOK22)); /* (bad) */
    SpSetErrMaskValue(mask, 23, UPnumVal(PROF_WP_SPELL_TOK23)); /* (need) */
    SpSetErrMaskValue(mask, 26, UPnumVal(PROF_WP_SPELL_TOK26)); /* cmpnd */
    SpSetErrMaskValue(mask, 28, UPnumVal(PROF_WP_SPELL_TOK28)); /* slash */
    SpSetErrMaskValue(mask, 34, UPnumVal(PROF_WP_SPELL_TOK34)); /* hyphen */
    SpSetErrMaskValue(mask, 42, UPnumVal(PROF_WP_SPELL_TOK42)); /* bad 's */
    SpSetErrMaskValue(mask, 44, UPnumVal(PROF_WP_SPELL_TOK44)); /* hyph num */
}

/*-----------------------------------------------------------------------
    Set a value in the passed "spellcheck error report" mask.
------------------------------------------------------------------------*/
void            
SpSetErrMaskValue ( mask, code, value)
sp_error_mask       *mask;              /* OUT: Desired member is set */
int                 code;               /* Code (0-63) of member to set */
bool                value;              /* TRUE -> report as misspelling. */
{
    bool            boolVal;            /* Legal mask value: 0x0 or 0x1 */

    if (value) boolVal = 0x1;
    else boolVal = 0x0;

    switch( code) {
        case 0: mask->unused0 = boolVal; break;
        case 1: mask->unused1 = boolVal; break;
        case 2: mask->unused2 = boolVal; break;
        case 3: mask->unused3 = boolVal; break;
        case 4: mask->unused4 = boolVal; break;
        case 5: mask->repeat = boolVal; break;
        case 6: mask->unused6 = boolVal; break;
        case 7: mask->number = boolVal; break;
        case 8: mask->date = boolVal; break;
        case 9: mask->time = boolVal; break;
        case 10: mask->nonAlph = boolVal; break;
        case 11: mask->misspell = boolVal; break;
        case 12: mask->startSent = boolVal; break;
        case 13: mask->roman = boolVal; break;
        case 14: mask->punctPre = boolVal; break;
        case 15: mask->punctSuff = boolVal; break;
        case 16: mask->outline = boolVal; break;
        case 17: mask->wordPS = boolVal; break;
        case 18: mask->prefix = boolVal; break;
        case 19: mask->wordWithPS = boolVal; break;
        case 20: mask->suffix = boolVal; break;
        case 21: mask->parens = boolVal; break;
        case 22: mask->badParens = boolVal; break;
        case 23: mask->needParens = boolVal; break;
        case 24: mask->goodCmpdPart = boolVal; break;
        case 25: mask->cmpd = boolVal; break;
        case 26: mask->badCmpd = boolVal; break;
        case 27: mask->goodCmpd = boolVal; break;
        case 28: mask->slash = boolVal; break;
        case 29: mask->unneededHyphen = boolVal; break;
        case 30: mask->contract = boolVal; break;
        case 31: mask->apos = boolVal; break;
        case 32: mask->misspellApos = boolVal; break;
        case 33: mask->hyphen = boolVal; break;
        case 34: mask->extraHyphen = boolVal; break;
        case 35: mask->hasHyphen = boolVal; break;
        case 36: mask->misspellHyphen = boolVal; break;
        case 37: mask->hasPosses = boolVal; break;
        case 38: mask->misspellPosses = boolVal; break;
        case 39: mask->aposPrefix = boolVal; break;
        case 40: mask->badAposPrefix = boolVal; break;
        case 41: mask->posses = boolVal; break;
        case 42: mask->badPosses = boolVal; break;
        case 43: mask->linkage = boolVal; break;
        case 44: mask->hyNumber = boolVal; break;
        default: ASSERT(0);
    }
}

/*-----------------------------------------------------------------------
    Return a value from the passed "spellcheck error report" mask.
------------------------------------------------------------------------*/
bool            
SpGetErrMaskValue ( mask, code)
sp_error_mask       *mask;              /* Error reporting structure */
int                 code;               /* Code (0-63) of member to get */
{

    switch( code) {
        case 0: return( mask->unused0);
        case 1: return( mask->unused1);
        case 2: return( mask->unused2);
        case 3: return( mask->unused3);
        case 4: return( mask->unused4);
        case 5: return( mask->repeat);
        case 6: return( mask->unused6);
        case 7: return( mask->number);
        case 8: return( mask->date);
        case 9: return( mask->time);
        case 10: return( mask->nonAlph);
        case 11: return( mask->misspell);
        case 12: return( mask->startSent);
        case 13: return( mask->roman);
        case 14: return( mask->punctPre);
        case 15: return( mask->punctSuff);
        case 16: return( mask->outline);
        case 17: return( mask->wordPS);
        case 18: return( mask->prefix);
        case 19: return( mask->wordWithPS);
        case 20: return( mask->suffix);
        case 21: return( mask->parens);
        case 22: return( mask->badParens);
        case 23: return( mask->needParens);
        case 24: return( mask->goodCmpdPart);
        case 25: return( mask->cmpd);
        case 26: return( mask->badCmpd);
        case 27: return( mask->goodCmpd);
        case 28: return( mask->slash);
        case 29: return( mask->unneededHyphen);
        case 30: return( mask->contract);
        case 31: return( mask->apos);
        case 32: return( mask->misspellApos);
        case 33: return( mask->hyphen);
        case 34: return( mask->extraHyphen);
        case 35: return( mask->hasHyphen);
        case 36: return( mask->misspellHyphen);
        case 37: return( mask->hasPosses);
        case 38: return( mask->misspellPosses);
        case 39: return( mask->aposPrefix);
        case 40: return( mask->badAposPrefix);
        case 41: return( mask->posses);
        case 42: return( mask->badPosses);
        case 43: return( mask->linkage);
        case 44: return( mask->hyNumber);
        default: ASSERT(0);
    }
	return(0);  /* should never get here */
}

/*-----------------------------------------------------------------------
    Handle "\-" escaped hyphenation points, and other "\"-escaped chars.
    Checks for word bigger than MAX_PROX_WORD.
------------------------------------------------------------------------*/
static typeErr 
spUnescapeHyphPoints (outstr,word)
char *outstr; /* out: string with hyphenation codes.
                buffer at least MAX_PROX_WORD. */
char *word; /* in: string with escaped hyphenation codes */
{
    register int ii;
    register char *instr;

    for (ii=1, instr=word; *instr; ii++, instr++, outstr++) {
        if (ii >= MAX_PROX_WORD) {
            MBstrcpy( SpErrorObject, word);
            return( SPELL_LONG_WORD);       /* Text string too long */
        }
        if(*instr == ESCAPE_CHAR) {
            instr++;
            if(*instr == HYPHEN_POINT_CHAR) {
                *outstr = HY_NORM;  /* Hi, Norm! Proximity hyphenation code */
                continue;
            }
        }
        *outstr = *instr;
    }
    *outstr = 0;
    return AOK;
}

/*-----------------------------------------------------------------------
    Make "\" escaped hyphenation points from word with proximity hyphenation
    codes in it.
------------------------------------------------------------------------*/
static mbchar *
spEscapeHyphPoints(word)
mbchar *word;
{
    int len;
    mbchar *wP, *unwP, *unword;

    /* how long will word be? */
    len = 0;
    for (wP=word; *wP; wP++) {
        if (*wP == HY_NORM || *wP == HY_PREF || *wP == HY_UNPREF
                || *wP == ESCAPE_CHAR) {
            len += 2;
        }
        else {
            len++;
        }
    }

    /* allocate space for word */
    unword = (mbchar *)malloc(sizeof(char) * len + 1);

    /* copy and escape hyph points */
    for (wP=word,unwP=unword; *wP; wP++) {
        if (*wP == HY_NORM || *wP == HY_PREF || *wP == HY_UNPREF) {
            *unwP++ = ESCAPE_CHAR;
            *unwP++ = HYPHEN_POINT_CHAR;
        }
        else if (*wP == ESCAPE_CHAR) {
            *unwP++ = ESCAPE_CHAR;
            *unwP++ = ESCAPE_CHAR;
        }
        else {
            *unwP++ = *wP;
        }
    }
    *unwP = 0;

    return unword;
}


/*-----------------------------------------------------------------------
    Get pathname for user dictionary, given language code.
------------------------------------------------------------------------*/
char *
SpUserDictFile(langCode)
int langCode;   /* language code */
{
    char *file;
    char infix[4];

    file = (char*)malloc(STRPATH+1);
    SpGetLanguageFileInfix(langCode,infix);
    sprintf(file, NXLT("%s/ax_%s.dct"), AGetAxHomeDir(), infix);

    return file;
}

/* }***********************************************************************/


/*************************************************************************
function: ElfbTextDictAddWord

    Add a word (or a misspell/correction pair) to the indicated dictionary.
*/
#ifdef AX_USE_PROTOS
void ElfbTextDictAddWord(
    char *word,     /* Word */
    char *correct,      /* If not null or empty, 'word'->misspell and
    * this is correction. */
    int language,       /* Work in this language */
    char *dict      /* Name of dictionary upon which to operate */ )
#else
void ElfbTextDictAddWord(word, correct, language, dict)
    char *word;     /* Word */
    char *correct;      /* If not null or empty, 'word'->misspell and
                 * this is correction. */
    int language;       /* Work in this language */
    char *dict;     /* Name of dictionary upon which to operate */
#endif
{
#if !FOR_TESTER
#if !STUB_PROX
    typeErr status;     /* Subroutine retcode */

    XmPutBusyCursor(NULL);
    status = SpAddWord(word, correct, language, dict);
    if (status != AOK)
        ElfStrAbort(status, SpErrorOverrideStr, SpErrorObject);
    return;
#endif
#endif
}

/*************************************************************************
function: ElfbTextDictGetWords

    Get some or all of the words from the indictated dictionary
*/
#ifdef AX_USE_PROTOS
ElfbTextDictGetWords(
    int language,       /* Work in this language */
    char *dict,     /* Name of dictionary from which to get words */
    AX_BOOLEAN firstTime,  /* Should be TRUE on 1st request, FALSE for
    * subsequent "more words" requests */
    int maxGet      /* Maximum number of words desired */ )
#else
ElfbTextDictGetWords(language, dict, firstTime, maxGet)
    int language;       /* Work in this language */
    char *dict;     /* Name of dictionary from which to get words */
    AX_BOOLEAN firstTime;  /* Should be TRUE on 1st request, FALSE for
                 * subsequent "more words" requests */
    int maxGet;     /* Maximum number of words desired */
#endif
{
#if !FOR_TESTER
#if !STUB_PROX
    char **words;       /* Words in dictionary */
    typeErr status;     /* Subroutine retcode */
    int gotten;     /* Number of words returned */
    int inList;     /* TRUE -> passed word was in list */
    register data_ptr aP, waP;
    register int ix;

    /* Dictionary words are returned in a single array as follows: 0:
     * Number of words being returned 1: Array of those words */

    XmPutBusyCursor(NULL);

    /* Allocate memory for array of dictionary words. */
    /* The function we call insists that we allocate LONGWORD+2 characters
     * for each word. */
    words = (char **) malloc(maxGet * sizeof(char *));
    /* !!0 throw error for out-of-memory */

    /* Convert ELF booleans to C booleans */
    if (firstTime == -1)
        firstTime = TRUE;
    else
        firstTime = FALSE;

    status = SpGetDictWords(language, dict, firstTime, maxGet, &gotten, words);
    if (status != AOK)
        ElfStrAbort(status, SpErrorOverrideStr, SpErrorObject);

    aP = MkArray(2);
    ElfAddNumToArray(aP, 0, gotten);
    waP = MkArray(gotten);
    for (ix = 0; ix < gotten; ++ix)
        {
        waP = ElfAddStrToArray(waP, ix, words[ix]);
        }
    aP = WriteArray(aP, 1, waP);
    ElfRetData = aP;    /* Alis func generic return value */

    /* Free memory */
    for (ix = 0; ix < gotten; ++ix)
        free(words[ix]);
    free(words);

    return;
#endif
#endif
}

/*************************************************************************
function: ElfbTextDictRemoveWord

    Remove a word (or a misspell/correction pair) from the indicated dictionary.
*/
#ifdef AX_USE_PROTOS
ElfbTextDictRemoveWord(
    char *word,     /* Word */
    char *correct,      /* If not null or empty, 'word'->misspell and
    * this is correction. */
    int language,       /* Work in this language */
    char *dict      /* Name of dictionary upon which to operate */ )
#else
ElfbTextDictRemoveWord(word, correct, language, dict)
    char *word;     /* Word */
    char *correct;      /* If not null or empty, 'word'->misspell and
                 * this is correction. */
    int language;       /* Work in this language */
    char *dict;     /* Name of dictionary upon which to operate */
#endif
{
#if !FOR_TESTER
#if !STUB_PROX
    typeErr status;     /* Subroutine retcode */

    XmPutBusyCursor(NULL);
    status = SpRemoveWord(word, correct, language, dict);
    if (status != AOK)
        ElfStrAbort(status, SpErrorOverrideStr, SpErrorObject);
    return;
#endif
#endif
}

/*************************************************************************
function: ElfbTextDictCanAdd
*/
ElfbTextDictCanAdd()
{
#if !FOR_TESTER
#if !STUB_PROX
    AX_BOOLEAN status;     /* Subroutine retcode */

    status = SpCanAddAnotherDictionary();
    if (status)
        return (-1);    /* "TRUE" in ELF */
    return (0);     /* "FALSE" in ELF */
#endif
#endif
}

/*************************************************************************
function: ElfbTextDictClose

    Close a dictionary
*/
#ifdef AX_USE_PROTOS
void ElfbTextDictClose( char *dict      /* Dictionary to close */ )
#else
void ElfbTextDictClose(dict)
    char *dict;     /* Dictionary to close */
#endif
{
#if !FOR_TESTER
#if !STUB_PROX
    typeErr status;     /* Subroutine retcode */

    status = SpRemoveDictionary(dict);
    if (status != AOK)
        ElfStrAbort(status, SpErrorOverrideStr, SpErrorObject);
    return;
#endif
#endif
}

/*************************************************************************
function: ElfbTextDictCreate

    Create a new, empty dictionary
*/
#ifdef AX_USE_PROTOS
void ElfbTextDictCreate(
    int language,       /* Work in this language */
    char *dict      /* Dictionary to create */ )
#else
void ElfbTextDictCreate(language, dict)
    int language;       /* Work in this language */
    char *dict;     /* Dictionary to create */
#endif
{
#if !FOR_TESTER
#if !STUB_PROX
    typeErr status;     /* Subroutine retcode */

    XmPutBusyCursor(NULL);
    status = SpCreateDictionary(language, dict);
    if (status != AOK)
        ElfStrAbort(status, SpErrorOverrideStr, SpErrorObject);
    return;
#endif
#endif
}

/*************************************************************************
function: ElfbTextDictLanguage

    Return the laguage of a dictionary
*/
#ifdef AX_USE_PROTOS
int ElfbTextDictLanguage(
    char *dict      /* Dictionary in question */ )
#else
int ElfbTextDictLanguage( dict)
    char *dict;     /* Dictionary in question */
#endif
{
#if !FOR_TESTER
#if !STUB_PROX
    int langCode;       /* Subroutine return */

    /* Return the integer code for the language. Note most other speller
     * functions will need this code. The constant "NOLANG" (0) is returned
     * if the file does not exist or is corrupt. */
    langCode = SpDictionaryLanguage( dict);
    return (langCode);
#endif
#endif
}

/*************************************************************************
function: ElfbTextGetLanguageCode
*/
#ifdef AX_USE_PROTOS
int ElfbTextGetLanguageCode( char *langStr       /* "English", "French", etc. */ )
#else
int ElfbTextGetLanguageCode(langStr)
    char *langStr;      /* "English", "French", etc. */
#endif
{
#if !FOR_TESTER
#if !STUB_PROX
    int langCode;       /* Subroutine return */

    /* Return the integer code for the language. Note most other speller
     * functions will need this code. The constant "NOLANG" (0) is returned
     * if the string is not recognized. */
    /* The fact that the Speller recognizes the language name does not mean
     * it will be supported... */
    langCode = SpGetLanguageCode(langStr);
    return (langCode);
#endif
#endif
}

/*************************************************************************
function: ElfbTextGetLanguageName
*/
#ifdef AX_USE_PROTOS
void ElfbTextGetLanguageName( int langCode        /* Integer code for language */ )
#else
void ElfbTextGetLanguageName(langCode)
    int langCode;       /* Integer code for language */
#endif
{
#if !FOR_TESTER
#if !STUB_PROX
    char langName[32];  /* Name of language */

    /* Return the corresponding language name for the passed integer
     * language code. The empty string is returned if the integer is out of
     * bounds. */
    /* The fact that the Speller recognizes the language name does not mean
     * it will be supported... */
    SpGetLanguageName(langCode, langName);
    ElfStrRetter(langName);
    return;
#endif
#endif
}

/*************************************************************************
function: ElfbTextHyphenate
*/
#ifdef AX_USE_PROTOS
ElfbTextHyphenate(
    char *word,     /* Find hyphenation points for this word */
    int language,       /* Work in this language */
    char **dictionaries,    /* NULL-terminated array of dictionaries */
    AX_BOOLEAN rank,       /* Should hyphenation points be ranked? */
    int how         /* How to generate hyph points: HYPH_XXX flag */ )
#else
ElfbTextHyphenate(word, language, dictionaries, rank, how)
    char *word;     /* Find hyphenation points for this word */
    int language;       /* Work in this language */
    char **dictionaries;    /* NULL-terminated array of dictionaries */
    AX_BOOLEAN rank;       /* Should hyphenation points be ranked? */
    int how;        /* How to generate hyph points: HYPH_XXX flag */
#endif
{
#if !FOR_TESTER
#if !STUB_PROX
    int nDicts;     /* # of dictionaries */
    char **dicts;       /* Array of dictionary names */
    typeErr status;     /* Subroutine retcode */
    int howDid;     /* How hyph points were generated */
    char hyPoints[MAX_PROX_WORD + 1]; /* Hyph points */
    int len;        /* Length of passed word */
    register data_ptr aP, haP;
    register int ix;

    /* Hyphenation info is returned in a single array as follows: 0: How
     * the hyphenation points were generated (a HYPH_XXXX value) 1:
     * Hyphenation points for the word: an array of integers, one for each
     * character in the word, describing hyphenation point following the
     * character (if any). Please refer to the Proximity documentation (and
     * to Proximity file sahyphen.h) for a description of these integers. */

    XmPutBusyCursor(NULL);
    nDicts = ElfbArraySize(dictionaries);
    dicts = (char **) malloc((nDicts + 1) * sizeof(char *));
    for (ix = 0; ix < nDicts; ++ix)
        {
        dicts[ix] = StrFromArray(dictionaries, ix);
        if (!*(dicts[ix])) dicts[ix] = NULL;        /* NULL term'd array */
        }
    dicts[nDicts] = NULL;

    status = SpHyphenateWord(word, language, dicts, rank, how,
                 &howDid, hyPoints);
    if (status != AOK)
        ElfStrAbort(status, SpErrorOverrideStr, SpErrorObject);
    aP = MkArray(2);
    ElfAddNumToArray(aP, 0, howDid);
    len = strlen(word);
    haP = MkArray(len);
    for (ix = 0; ix < len; ++ix)
        {
        haP = ElfAddNumToArray(haP, ix, (int) hyPoints[ix]);
        }
    aP = WriteArray(aP, 1, haP);
    ElfRetData = aP;    /* Alis func generic return value */

	free( dicts);
#endif
#endif
}

/*************************************************************************
function: ElfbTextSpelledCorrectly
*/
#ifdef AX_USE_PROTOS
int ElfbTextSpelledCorrectly(
    char *word,     /* See if this word is spelled correctly */
    int language,       /* Work in this language */
    char **dictionaries     /* NULL-terminated array of dictionaries */ )
#else
int ElfbTextSpelledCorrectly(word, language, dictionaries)
    char *word;     /* See if this word is spelled correctly */
    int language;       /* Work in this language */
    char **dictionaries;    /* NULL-terminated array of dictionaries */
#endif
{
#if !FOR_TESTER
#if !STUB_PROX
    int nDicts;     /* # of dictionaries */
    char **dicts;       /* Array of dictionary names */
    typeErr status;     /* Subroutine retcode */
    int ok;         /* TRUE -> found word in Proximity wordlists */
    register int ix;

    XmPutBusyCursor(NULL);
    nDicts = ElfbArraySize(dictionaries);
    dicts = (char **) malloc((nDicts + 1) * sizeof(char *));
    for (ix = 0; ix < nDicts; ++ix)
        {
        dicts[ix] = StrFromArray(dictionaries, ix);
        if (!*(dicts[ix])) dicts[ix] = NULL;        /* NULL term'd array */
        }
    dicts[nDicts] = NULL;

    status = SpWordSpelledCorrectly(word, language, dicts, &ok);
	free( dicts);
    if (status != AOK)
        ElfStrAbort(status, SpErrorOverrideStr, SpErrorObject);
    if (ok)
        return (-1);    /* "TRUE" in ELF */
    return (0);     /* "FALSE" in ELF */
#endif
#endif
}

/*************************************************************************
function: ElfbTextSpellerClose
*/
void ElfbTextSpellerClose()
{
#if !FOR_TESTER
#if !STUB_PROX
    typeErr status;     /* Subroutine retcode */

    XmPutBusyCursor(NULL);
    status = SpCloseSpeller();
    if (status != AOK)
        ElfStrAbort(status, SpErrorOverrideStr, SpErrorObject);
    return;
#endif
#endif
}

/*************************************************************************
function: ElfbTextSpellerCurrLanguage
*/
int ElfbTextSpellerCurrLanguage()
{
#if !FOR_TESTER
#if !STUB_PROX
    int currLang;       /* Subroutine retcode */

    /* Note: "NOLANG" (0) is returned if speller not started. Legal
     * language values are currently 1-16. */
    currLang = SpGetCurrentLanguage();
    return (currLang);
#endif
#endif
}

/*************************************************************************
function: ElfbTextSpellerHyphenationOK
*/
#ifdef AX_USE_PROTOS
ElfbTextSpellerHyphenationOK( int language        /* Language code (1-16) in question */ )
#else
ElfbTextSpellerHyphenationOK(language)
    int language;       /* Language code (1-16) in question */
#endif
{
#if !FOR_TESTER
#if !STUB_PROX
    AX_BOOLEAN ok;     /* Subroutine retcode */

    XmPutBusyCursor(NULL);
    ok = SpIsHyphenationOK(language);
    if (ok)
        return (-1);    /* "TRUE" in ELF */
    return (0);     /* "FALSE" in ELF */
#endif
#endif
}

/*************************************************************************
function: ElfbTextSpellerLanguageOK
*/
#ifdef AX_USE_PROTOS
int ElfbTextSpellerLanguageOK( int language        /* Language code (1-16) in question */ )
#else
int ElfbTextSpellerLanguageOK(language)
    int language;       /* Language code (1-16) in question */
#endif
{
#if !FOR_TESTER
#if !STUB_PROX
    AX_BOOLEAN ok;     /* Subroutine retcode */

    XmPutBusyCursor(NULL);
    ok = SpIsLanguageOK(language);
    if (ok)
        return (-1);    /* "TRUE" in ELF */
    return (0);     /* "FALSE" in ELF */
#endif
#endif
}

/*************************************************************************
function: ElfbTextSpellerThesaurusOK
*/
#ifdef AX_USE_PROTOS
ElfbTextSpellerThesaurusOK( int language        /* Language code (1-16) in question */ )
#else
ElfbTextSpellerThesaurusOK(language)
    int language;       /* Language code (1-16) in question */
#endif
{
#if !FOR_TESTER
#if !STUB_PROX
    AX_BOOLEAN ok;     /* Subroutine retcode */

    XmPutBusyCursor(NULL);
    ok = SpIsThesaurusOK(language);
    if (ok)
        return (-1);    /* "TRUE" in ELF */
    return (0);     /* "FALSE" in ELF */
#endif
#endif
}

/*************************************************************************
function: ElfbTextSpellingAlternatives

    Return a list of suggested spelling for the (possibly misspelled)
    passed word.
*/
#ifdef AX_USE_PROTOS
ElfbTextSpellingAlternatives(
    char *word,     /* Find alternatives for this word */
    int language,       /* Work in this language */
    char **dictionaries,    /* NULL-terminated array of dictionaries */
    int maxGet      /* Maximum number of alternatives to retrieve */ )
#else
ElfbTextSpellingAlternatives(word, language, dictionaries, maxGet)
    char *word;     /* Find alternatives for this word */
    int language;       /* Work in this language */
    char **dictionaries;    /* NULL-terminated array of dictionaries */
    int maxGet;     /* Maximum number of alternatives to retrieve */
#endif
{
#if !FOR_TESTER
#if !STUB_PROX
    int nDicts;     /* # of dictionaries */
    char **dicts;       /* Array of dictionary names */
    int gotten;     /* Number of alternatives returned */
    char **words;       /* Alternatives returned */
    int inList;     /* TRUE -> passed word was in list */
    typeErr status;     /* Subroutine retcode */
    register data_ptr aP, waP;
    register int ix;

    /* Spelling alternatives info is returned in a single array as follows:
     * 0: # of meanings returned 1: TRUE if passed word is in list. FALSE
     * if not. 2: Spelling alternatives, in the following array: a: First
     * alternative b: Second alternative ...etc... */

    XmPutBusyCursor(NULL);
    nDicts = ElfbArraySize(dictionaries);
    dicts = (char **) malloc((nDicts + 1) * sizeof(char *));
    for (ix = 0; ix < nDicts; ++ix)
        {
        dicts[ix] = StrFromArray(dictionaries, ix);
        if (!*(dicts[ix])) dicts[ix] = NULL;        /* NULL term'd array */
        }
    dicts[nDicts] = NULL;

    /* Allocate memory for array of alternatives. */
    words = (char **) malloc(maxGet * sizeof(char *));
    /* !!0 throw error for out-of-memory */

    status = SpListValidSpellings(word, language, dicts, maxGet,
                      &gotten, words, &inList);
    if (status != AOK)
        ElfStrAbort(status, SpErrorOverrideStr, SpErrorObject);
    aP = MkArray(3);
    ElfAddNumToArray(aP, 0, gotten);
    if (inList)
        ElfAddNumToArray(aP, 1, -1);
    else
        ElfAddNumToArray(aP, 1, 0);
    waP = MkArray(gotten);
    for (ix = 0; ix < gotten; ++ix)
        {
        waP = ElfAddStrToArray(waP, ix, words[ix]);
        }
    aP = WriteArray(aP, 2, waP);
    ElfRetData = aP;    /* Alis func generic return value */

    /* Free memory */
    for (ix = 0; ix < gotten; ++ix)
        free(words[ix]);
    free(words);
	free( dicts);
    return;
#endif
#endif
}

/*************************************************************************
function: ElfbTextSpellingErrors

    Return information about the first spelling error found in the passed
    string.
*/
#ifdef AX_USE_PROTOS
ElfbTextSpellingErrors(
    char *string,       /* Find 1st error in here */
    int startOff,       /* Start reporting errors at this offset */
    int language,       /* Work in this language */
    char **dictionaries,    /* NULL-terminated array of dictionaries */
    int *misspells      /* Array of speller codes that we are */
                        /* interested in hearing about (running from */
                        /* approx 5-44). If array is completely empty, */
                        /* use profile vars instead. */ 
    )
#else
ElfbTextSpellingErrors(string, startOff, language, dictionaries, misspells)
    char *string;       /* Find 1st error in here */
    int startOff;       /* Start reporting errors at this offset */
    int language;       /* Work in this language */
    char **dictionaries;    /* NULL-terminated array of dictionaries */
    int *misspells;     /* Array of speller codes that we are
                 * interested in hearing about (running from
                 * approx 5-44). If array is completely empty,
                 * use profile vars instead. */
#endif
{
#if !FOR_TESTER
#if !STUB_PROX
    int nDicts;     /* # of dictionaries */
    char **dicts;       /* Array of dictionary names */
    int nMispCodes;     /* # of misspelling codes */
    sp_error_mask mask; /* Structure for misspell codes */
    typeErr status;     /* Subroutine retcode */
    int errCode;        /* Spelling error code reported */
    int errStart;       /* Offset where error starts */
    int errEnd;     /* Offset (inclusive) where error ends */
    int scanEnd;        /* Offset of last character Speller analyzed */
    register int ix, jx;
    register data_ptr aP;
    wchar *wstrP, *wtempP;
    int found_narrow;

    /* Spelling error info is returned in a single array as follows: 0:
     * Error code (zero means no error found). 1: Starting character offset
     * of error 2: Ending character offset of error 3: Last character
     * Speller examined to generate this error */

    XmPutBusyCursor(NULL);
    nDicts = ElfbArraySize(dictionaries);
    dicts = (char **) malloc((nDicts + 1) * sizeof(char *));
    for (ix = 0; ix < nDicts; ++ix)
        {
        dicts[ix] = StrFromArray(dictionaries, ix);
        if (!*(dicts[ix])) dicts[ix] = NULL;        /* NULL term'd array */
        }
    dicts[nDicts] = NULL;

    SpInitErrMask(&mask);
    nMispCodes = ElfbArraySize(misspells);
    if (!nMispCodes)
        {
        SpSetErrMaskProfile( &mask);
        }
    else
        for (ix = 0; ix < nMispCodes; ++ix)
            {
            jx = IntFromArray(misspells, ix);
            SpSetErrMaskValue(&mask, jx, TRUE);
            }

#if BUILD_WIDE
    wstrP = MbToWOneStr( string );
    wtempP = wstrP;
    found_narrow = FALSE;
    while ( *wtempP != (wchar)0 )
       {
       /* convert every wide char to a blank so they get ignored */
       if ( IS_WIDE_CHAR( *wtempP ) || IS_NARROW_KANA( *wtempP ) )
         *wtempP = ' ';
       else
         found_narrow = TRUE;
       wtempP++;
       }
    if ( !found_narrow )
       {
       /* all wide chars, skip it all */
       errStart = errEnd = scanEnd = 0;
       errCode = AOK;
       goto finish_up;
       }
#else
    wstrP = (wchar *)string;
#endif

    status = SpCheckString(wstrP, startOff, language, dicts, &mask,
                   &errCode, &errStart, &errEnd, &scanEnd);
    if (status != AOK)
        ElfStrAbort(status, SpErrorOverrideStr, SpErrorObject);

finish_up:
    aP = MkArray(4);
    ElfAddNumToArray(aP, 0, errCode);
    ElfAddNumToArray(aP, 1, errStart);
    ElfAddNumToArray(aP, 2, errEnd);
    ElfAddNumToArray(aP, 3, scanEnd);
    ElfRetData = aP;    /* Alis func generic return value */
	free( dicts);
    return;
#endif
#endif
}

/*************************************************************************
function: ElfbTextSpellingInfo
*/
ElfbTextSpellingInfo()
{
#if !FOR_TESTER
#if !STUB_PROX
    int nDicts;     /* # of dictionaries */
    char **dicts;       /* Array of dictionary names */
    typeErr status;     /* Subroutine retcode */
    register int ix;

    /******************
    nDicts = ElfbArraySize(dictionaries);
    dicts = (char**) malloc((nDicts+1) * sizeof(char*));
    for (ix=0; ix<nDicts; ++ix)     {
        dicts[ix] = StrFromArray(dictionaries,ix);
        if (!*(dicts[ix])) dicts[ix] = NULL;
        }
    dicts[nDicts] = NULL;
    *******************/
#endif
#endif
}

/*************************************************************************
function: ElfbTextThesaurus
*/
#ifdef AX_USE_PROTOS
void ElfbTextThesaurus(
    char *theWord,
    int language  )
#else
void ElfbTextThesaurus(theWord, language)
    char *theWord;
    int language;
#endif
{
#if !FOR_TESTER
#if !STUB_PROX
    register data_ptr aP, maP, waP;
    register int ix, jx, kx, cnt = 0;
    int nMeanings;      /* Number of meanings found for word */
    char *meanings[MAXMEAN + 1];    /* Array of ptrs to meaning strings */
    char **words[MAXMEAN + 1][MAXLISTS];    /* For each meaning, an array
                         * of wordlists (synonyms,
                         * antonyms...) each wordlist
                         * == array of ptrs to words */
    char **wordPtr;     /* Current word */
    typeErr status;     /* Subroutine retcode */

    /* Thesaurus information is returned in a single array as follows: 0: #
     * of meanings returned 1: Data for first meaning, in the following
     * array: a: First meaning string b: List (i.e. subarray) of synonyms
     * for 1st meaning c: List of compared words for 1st meaning d: List of
     * related words for 1st meaning e: List of contrasted words for 1st
     * meaning f: List of antonyms for 1st meaning 2: Data for second
     * meaning, in the following array: ...etc... */

    XmPutBusyCursor(NULL);
    status = SpGetThesaurus(theWord, language, &nMeanings, meanings, words);
    if (status != AOK)
        ElfStrAbort(status, SpErrorOverrideStr, SpErrorObject);

    aP = MkArray(nMeanings + 1);
    ElfAddNumToArray(aP, 0, nMeanings);
    for (ix = 0; ix < nMeanings; ix++)
        {
        maP = MkArray(MAXLISTS + 1);
        ElfAddStrToArray(maP, 0, meanings[ix]);
        for (jx = 0; jx < MAXLISTS; jx++)
            {
            wordPtr = words[ix][jx];
            waP = MkArray(0);
            if (wordPtr)
                {
                kx = 0;
                while (*wordPtr)
                    {
                    waP = ElfAddStrToArray(waP, kx++, *wordPtr++);
                    }
                }
            maP = WriteArray(maP, jx + 1, waP);
            }
        aP = WriteArray(aP, ix + 1, maP);
        }
    ElfRetData = aP;    /* Alis func generic return value */
    return;
#endif
#endif
}

/*************************************************************************
function: ElfbTextThesaurusClose
*/
void ElfbTextThesaurusClose()
{
#if !FOR_TESTER
#if !STUB_PROX
    typeErr status;     /* Subroutine retcode */

    status = SpCloseThesaurus();
    if (status != AOK)
        ElfStrAbort(status, SpErrorOverrideStr, SpErrorObject);
    return;
#endif
#endif
}

/*************************************************************************
function: ElfbTextThesaurusInfo
*/
#ifdef AX_USE_PROTOS
ElfbTextThesaurusInfo( int language        /* Work in this language */ )
#else
ElfbTextThesaurusInfo(language)
    int language;       /* Work in this language */
#endif
{
#if !FOR_TESTER
#if !STUB_PROX
    int data;       /* A bunch of flags pertaining to Thesaurus */
    typeErr status;     /* Subroutine retcode */

    status = SpWhatThesaurusData(language, &data);
    if (status != AOK)
        ElfStrAbort(status, SpErrorOverrideStr, SpErrorObject);
    return (data);
#endif
#endif
}


