/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: elcomp.h
   Applix ASC version: 19
   Date: Tue Feb  2 15:21:13 1999
*/
#ifndef ELFCOMP_INCLUDED
#define ELFCOMP_INCLUDED 1

#include <stdio.h>
#include <aos.h>

#include <elf.h>

extern DLL_IMPORT char **ElfCompSrcArrayPtr;	/* if not NULL, fetch source from this array */
extern DLL_IMPORT char **ElfCompCodeBufferPtr;	/* if not NULL, write code buffer to this addr */
extern DLL_IMPORT int *ElfCompSizeCodeBuffer;	/* set by compiler if ElfCompCodeBufferPtr */

#ifdef assert
#undef assert
#endif
#define assert(x) if ( !(x) ) Asserter(__FILE__,__LINE__)

char *ElfcAlloc();


#if DEBUG
#define LOGGING 1
extern DLL_IMPORT char *LogMod;
#define LOGGER(body) { if ( Logging ) {Logger body;}}
#define VLOGGER(body) { if ( Logging ) {vLogger body;}}
#define ALOG(body) ALog body ; LOGGER(body)
#define TRACE(body) { if ( Tracing ) {Logger body;}}
#define ASSY(body) WtAs body
#else
#define LOGGER(body)
#define VLOGGER(body)
#define ALOG(body)
#define ASSY(body)
#endif

#define MAX_LINELEN 4000 /* length of a source line - same as ed'd USERLEN */
#define MAX_ERRS 20     /* number of errs before compilation aborted */

typedef struct data_id {        /* variables, functions, labels */
    shrt type;          /* data type */
    long locate;         /* frame offset value */
    shrt loaded;        /* whether value in code frame yet (num & str only) */
    int srcLine;
    int formatDim;	/* array levels of format */
    int moduleName;   /* NULL or for externs, name of module item must be in */
    unsigned is_builtin : 1,	/* matches name of builtin function */
    is_funckey : 1,	/* matches name of function key */
    blt_checked : 1,	/* flag saying whether tested against bltins */
    need_link : 1,	/* TRUE is bltin but still needs dyn link info */
    isObject : 1,	/* For VARS, if TRUE declared as an object */
	isPublic : 1,	/* for objvars */
	isPrivate : 1,	/* for object methods */
    bltinpos : 10;	/* pos in table above */
    struct format_info *format;	/* null or suggested format */
    char name[1];       /* varying extent, null terminated */
} data_id;

typedef struct {          /* compiler's number */
    shrt type;              /* data type */
    long locate;            /* 0 or where in code frame */
    shrt loaded;            /* whether value in code frame yet */
    double value;
} data_comp_number;

typedef struct {          /* compiler's small number */
    shrt type;              /* data type */
    long locate;            /* 0 or where in code frame */
    shrt loaded;            /* whether value in code frame yet */
    int value;
} data_comp_integer;

typedef struct {          /* compiler's string */
    shrt type;              /* data type */
    long locate;            /* 0 or where in code frame */
    shrt loaded;            /* whether value in code frame yet */
    char text[1];       /* varying extent, null terminatd */
} data_comp_string;

typedef union _data_ {
    shrt type;		/* data type */
    long locate;		/* 0 or where in code frame */
    shrt loaded;		/* whether value in code frame yet */
    data_id id;
    data_comp_number number;
    data_comp_integer integer;
    data_comp_string string;
} data_;

/************** COMPILER TYPEDEFS ****************************/
#define MAX_ELF_SELECTS 20
#define MAX_WHILES 20 /* max while loop nesting */
#define MAX_IFS 20  /* max if-then-else nesting */

typedef struct {
    shrt id;
    shrt label;
    shrt next_label;
    shrt end_label;
    shrt step_reg;
    shrt stop_reg;
} for_loop_info;

typedef struct {
    shrt label;
    shrt end_label;
} while_loop_info;

typedef struct select_item {	/* list of items for select list */
	struct select_item *next;
	data_id *item;
	shrt code_label;
} select_item;

typedef struct {
	shrt opcode_label;
    shrt end_label;
    shrt data_label;
	shrt dflt_label;
    shrt did_dflt;
	shrt last_case_ended_pc;
	shrt is_numbers, is_strings;
	select_item *items;	
} select_info;

typedef struct {
    shrt num_for_loops;
    for_loop_info for_loops[MAX_FORS];
    shrt num_while_loops;
    while_loop_info while_loops[MAX_WHILES];
    shrt num_selects;
    select_info selects[MAX_ELF_SELECTS];
    shrt cur_label;          /* for automatic labels */
} comp_loop_info;

/*

Types of tokens:

Simple: operators, punctionation, keywords
Word:   identifiers
String: quoted strings
Number: Numbers

*/

#define MAX_TOKENS 1000
#define MAX_FORMATS 100

typedef struct {
    shrt type;
    shrt code;
} tok_simple;

typedef struct {
    shrt type;
    char name[1];
} tok_id;

typedef struct {
    shrt type;
    char str[1];	/* varying extent */
} tok_str;

typedef struct {
    shrt type;
    double val;
} tok_number;

typedef struct {
    shrt type;
    int val;
} tok_integer;

typedef union {
    tok_simple *simple;
    tok_id *id;
    tok_str *str;
    tok_number *num;
    tok_integer *small;
} tok_ptr;

extern DLL_IMPORT tok_ptr *ElfTokArray;
extern DLL_IMPORT int TokPos;

#define TKT_EMPTY   0   /* of simple */
#define TKT_SIMPLE  1
#define TKT_ID      2
#define TKT_STRING  3
#define TKT_NUMBER  4
#define TKT_INTEGER 5
#define TKT_EOF     6   /* of simple */

/* Parse tree number fudges */

#define SYM_FUDGE 1000
#define NODE_FUDGE 10000        /* start of nodes */

#define ELFC_VERIFY(x,str) if ( !(x) ) {ErrOut(str);}

typedef struct def_info {        /* format of #defines */
    struct def_info *next;
    char *name;
    shrt num_toks;
    tok_ptr *toks;      /* array of replacement tokens */
} def_info;
extern DLL_IMPORT def_info **Defines;   /* heads of defines list */

typedef struct format_info {	/* description of array format */
	char *myname;				/* name of this format */
	struct format_info *next;	/* next in chain */
	int used;				/* TRUE if used by a var for the debugger */
	char *fileName;	/* name of file containing this format */
	char *names[MAX_FORMATS];	/* names of elements */
	char *formats[MAX_FORMATS];	/* null or subformat ptr */
	shrt formatDim[MAX_FORMATS];		/* array levels of format */
} format_info;

extern DLL_IMPORT int Logging;
extern DLL_IMPORT int Tracing;
extern DLL_IMPORT int NumErrs;         /* number of compilation errors */


#endif      /*  ELFCOMP_INCLUDED   */
