/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: eldata.c
   Applix ASC version: 134
   Date: Tue Feb  2 15:21:13 1999
*/
#if X386_NEW || AXMS || HP800 || AVIION || LINUX
#include <stdlib.h>
#endif

#if AXMS
#include <fcntl.h>
#endif

#include <sys/types.h>
#include <axsocket.h>
#include <math.h>
#include <aos.h>
#include <elf_env.h>
#include <eldata.h>
#include <getstr4.h>
#include <getstr5.h>

typedef struct {
	int size;	/* bytes allocated */
	int cnt;	/* bytes in use */
	char *buf;	/* allocated buffer (AxMemWrite CALLER MUST FREE) */
} wrtbuf_t;

typedef struct {
	int cnt;	/* bytes in use */
	char *buf;	/* allocated buffer */
} rdbuf_t;

#include <eldata.p>



#if ELF_COMM_STAND_ALONE  /*  {  ----------------------  */
#   define THIMpid 0
#   define XLT( str ) str
#   define CHECK(test,code,str) if ( !(test) ) AxError(code,str,NULL);
#   define AxTaskPoolAllocate(task,size) AxMalloc(size)
#   define AxTaskPoolRealloc(task,p,size,oldsize) AxRealloc(p,size)
#   define AxTaskPoolFree(task,p,size) AxFree(p)
#   define ConvertUSDotsToEuropeComma(s)

#if USE_GLOBAL_ALLOC       /*  {  */
    /* true on NT where the 'C' library "malloc/free" does NOT re-use */
    /* free'd memory in the interests of speed, how dumb - jon        */
#   define TaskAlloc(task,size) (char *)GlobalAlloc( GMEM_FIXED, size )
#   define TaskFree(task, ptr) GlobalFree( ptr )
#   define TaskRealloc( task, ptr, new_size ) (char *)GlobalReAlloc( ptr, new_size, GMEM_MOVEABLE )
#else                       /*  }  {  */
#   define TaskAlloc(t,n) malloc((n))
#   define TaskFree(t,p) free((p))
#   define TaskRealloc(t,p,n) realloc(p, n)
#endif                      /* }  */

#else             /* } ------------------- { */

#   define CHECK(test,code,str) if ( !(test) ) AxElfAbort(code,NULL,str);
char *TaskAlloc(), *TaskRealloc();

#endif           /* } ------------------- */




#define MEM_BUF_DLFT_SIZE 1000

#include <java_syms.h>

double Astrtod AXP_((char* str, char **dst));

/* we don't need the dll_export specifier anymore, because we list
 * all the exports in a defintion file (.def)
 */
#undef DLL_EXPORT
#define DLL_EXPORT

DLL_EXPORT void AxTaskFreeArrayButNotContents();


/*************************************************************************
function: AxMemWrite - Write Elf array to memory buffer, return buffer address
After processing the returned buffer, the CALLER MUST FREE IT.
*/
int AxMemWrite (data,makeRoom,buf)
data_ptr data;          /* Elf data array */
int makeRoom;           /* leave this amount available */
mbchar **buf;             /* returned buf addr */
{
    return(AxMemWritePacked(data,makeRoom,buf,FALSE));
}


/*************************************************************************
function: AxMemWritePacked - Write Elf array to memory buffer, return buffer address
After processing the returned buffer, the CALLER MUST FREE IT.
*/
int AxMemWritePacked (data,makeRoom,buf,packedElfIO)
data_ptr data;          /* Elf data array */
int makeRoom;           /* leave this amount available */
mbchar **buf;             /* returned buf addr */
int packedElfIO;        /* If true, use tighter algorithm */
{
    int size;
	wrtbuf_t membuf;

    membuf.cnt = sizeof(int) + makeRoom;
    membuf.size = MEM_BUF_DLFT_SIZE;
    membuf.buf = AxMalloc(membuf.size);
    AxAddSome(&membuf,data,packedElfIO);
    size = membuf.cnt - sizeof(int);
    size = htonl(size);
    memcpy(membuf.buf+makeRoom,&size,sizeof(size));
    *buf = membuf.buf;
    return(membuf.cnt);
}

/*************************************************************************
function: AxMemRead - Decompose mem buffer to Elf array
After processing the returned data, the CALLER MUST CALL AxFreeData(aP)
*/
data_ptr AxMemRead (bufP)
mbchar *bufP;
{
    data_ptr dp;
	rdbuf_t membuf;

	membuf.buf = bufP;
    membuf.cnt = sizeof(int);    /* skip over size */
    dp = ReadSomeElfData(&membuf);
    return(dp);
}

/*************************************************************************
function: AxRPCRead - Read message portion itself
*/
data_ptr AxRPCRead (bufP)
mbchar *bufP;
{
    data_ptr dp;
	rdbuf_t membuf;

    membuf.buf = bufP;
    membuf.cnt = 0;
    dp = ReadSomeElfData(&membuf);
    return(dp);
}


/*************************************************************************
************* CREATE DATA ************************************************
*************************************************************************/

rleNulls(memP,ap,dexP,arrsize)
wrtbuf_t *memP;
data_ptr *ap; /* ap => &val[*dexP] */
int *dexP;
int arrsize;
{
    int cnt;
    int dex = *dexP;

    for ( cnt = 0; (dex < arrsize) && !ap[cnt]; cnt++, dex++ )
        /* empty */ ; 
    if(cnt < 3)
        return(FALSE);
    while ( cnt > 127 ) {
        add_byte(memP,SOCK_NULLS);
        add_byte(memP,127);
        cnt -= 127;
    }
    add_byte(memP,SOCK_NULLS);
    add_byte(memP,cnt);
    *dexP = dex;
    return(TRUE);
}

/*************************************************************************
function: AxAddSome - Recursive adder
*/
static void AxAddSome (wrtbuf_t *memP, data_ptr dP, int packedElfIO) {
    mbchar *str;
    int templong;
    int dex;
    int arrsize;
    mbchar buf[100];
    double tempnum;

    if ( !dP ) {
        if ( packedElfIO )
            add_byte(memP,SOCK_NULL);
        else
            add_long(memP,SOCK_NULL);
    }
    else if ( IS_INTEGER(dP) ) {
        templong = GET_INTEGER(dP);
        if ( packedElfIO ) {
            if ( (templong >= 0) && (templong <= 15) )
                add_byte(memP,SOCK_I0+templong);
            else if ( templong == -1 ) add_byte(memP,SOCK_INEG1);
            else if ( templong == 127 ) add_byte(memP,SOCK_I127);
            else if ( templong == 128 ) add_byte(memP,SOCK_I128);
            else if ( templong == 256 ) add_byte(memP,SOCK_I256);
            else if ( templong == 255 ) add_byte(memP,SOCK_I255);
            else if ( (templong >= -128) && (templong < 128) ) {
                add_byte(memP,SOCK_BYTE);
                add_byte(memP,templong);
            }
            else if ( (templong >= -32768) && (templong < 32768) ) {
                add_byte(memP,SOCK_WORD);
                add_byte(memP,templong >> 8);
                add_byte(memP,templong & 0x00FF);
            }
            else {
                add_byte(memP,SOCK_INT);
                add_long(memP,templong);
            }
        }
        else {
            add_long(memP,SOCK_INT);
            add_long(memP,templong);
        }
    }
    else if ( AxIsNumber(dP) ) {
        tempnum = AxFloatFromDataPtr(dP);
        sprintf((char *)buf,"%.15g",tempnum);
        if ( packedElfIO ) {
            add_byte(memP,SOCK_FLOAT_SHORT);
            add_short_string(memP,buf);
        }
        else {
            add_long(memP,SOCK_FLOAT);
            add_c_string(memP,buf);
        }
    }
    else if ( AxIsString(dP) ) {
        str = AxStrPtrFromDataPtr(dP);
        if ( packedElfIO ) {
            int len = 0;
#if (0 && BUILD_WIDE)
			/* convert str to unicode_str, then output unicode_str */
			wchar *wstr = elfMbStrToUnicodeStr(str,&len);
			add_byte(memP,SOCK_STRING_UNICODE);
			add_unicode_string(memP,len,wstr);
#else
			len = strlen(str);
            if (0 == len)
                add_byte(memP,SOCK_STR_EMPTY);
            else if ((1 == len) && (' ' == str[0]))
                add_byte(memP,SOCK_STR_SPACE);
			else if (1 == len) {
                add_byte(memP,SOCK_STR_CHAR);
                add_byte(memP,str[0]);
			}
            else if (len < 256) {
                add_byte(memP,SOCK_STR_SMALL);
                add_short_string(memP,str);
            }
            else {
                add_byte(memP,SOCK_STRING);
                add_c_string(memP,str);
            }
#endif	/* BUILD_WIDE */
        }
        else {
            add_long(memP,SOCK_STRING);
            add_c_string(memP,str);
        }
    }
    else if ( AxIsArray(dP) ) {
        arrsize = AxArraySize(dP);  /* can throw error */
        if ( packedElfIO ) {
			if (arrsize <= 4) {
				add_byte(memP,SOCK_ARR_0+arrsize);
				for (dex = 0; dex < arrsize; ++dex)
					AxAddSome(memP,AxArrayElement(dP,dex),packedElfIO);
			}
            else {
                data_ptr val;
				if (arrsize < 256) {
					add_byte(memP,SOCK_ARR_SMALL);
					add_byte(memP,arrsize);
				}
				else {
					add_byte(memP,SOCK_ARRAY);
					add_long(memP,arrsize);
				}
                for ( dex = 0; dex < arrsize; ) {
					val = AxArrayElement(dP,dex);
					if ( val || !rleNulls(memP,&(AX(dP)->vals[dex]),&dex,arrsize) ) {
						AxAddSome(memP,val,packedElfIO);
						dex++;
					}
				}
			}
		}
        else {
            add_long(memP,SOCK_ARRAY);
            add_long(memP,arrsize);
            for ( dex = 0; dex < arrsize; dex++ )
                AxAddSome(memP,AxArrayElement(dP,dex),packedElfIO);
        }
    }
    else if ( AxIsBinary(dP) ) {
        int rleSize;
        unsigned char *rleBuf;
        if ( packedElfIO && RLEpacked(dP,&rleBuf,&rleSize) ) {
            /* We are allowed to use a run-length-compression, and */
            /* compressing the data will save space, so, use it    */
            add_byte(memP,SOCK_BINARY_RLE);
            templong = AxBinarySize(dP);
            add_long(memP,templong);                /* the original size */
            add_long(memP,rleSize);                 /* the compressed size */
            add_bytes(memP,(char*)rleBuf,rleSize);
            AxFree((mbchar*)rleBuf);
        }
        else {
            /* no compresion allowed, or data would not get smaller */
            add_long(memP,SOCK_BINARY);
            templong = AxBinarySize(dP);
            add_long(memP,templong);
            str = (mbchar *)((data_binary*)dP)->bytes;
            add_bytes(memP,str,templong);
        }
    }
#if !ELF_COMM_STAND_ALONE
    else if ( AxIsObject(dP) ) {
		if (packedElfIO)
			add_byte(memP,SOCK_OBJECT);
		else
			add_long(memP,SOCK_OBJECT);
        if ( OX(dP)->objUid == 0 && OX(dP)->objHostname == NULL )
            ElfObjPtr(dP);
        add_long(memP,OX(dP)->objDex);
        add_long(memP,OX(dP)->objUid);
    }
#endif
}

RLEpacked(dP,rleBuf,rleSize)
data_ptr dP;
unsigned char **rleBuf;
int *rleSize;
{
    register unsigned char *src = (unsigned char*)(((data_binary*)dP)->bytes);
    register unsigned char *dst, *outbuf;
    register unsigned char ch;
    register int total = ((data_binary*)dP)->size;
    register int ix = 0, cnt;

    dst = outbuf = (unsigned char*)AxMalloc(total*2);
    while ( ix < total ) {
        for ( ch = *src, cnt = 0;
                (*src == ch) && (cnt < 127) && (ix < total);
                src++,cnt++,ix++ );
        *dst++ = cnt;
        *dst++ = ch;
    }
    if ( dst-outbuf > total ) {
        AxFree((mbchar*)outbuf);
        return(FALSE);
    }
    *rleBuf = outbuf;
    *rleSize = dst-outbuf;
    return(TRUE);
}




/* Uncompress a run-length encoded block of bytes.  The format is one byte */
/* as a count of the number of the floowing byte which occur               */
void UncompressRLE( unsigned char *compress_buf, unsigned char *out_buff,
                                    int compressed_size, int orig_size )
{
    unsigned char *inP, *outP;
    unsigned int char_count, char_val;
    unsigned char *endP;

    endP = compress_buf + compressed_size;
    outP = out_buff;
    inP = compress_buf;

    while ( inP < endP )
        {
        char_count = *inP;
        inP++;
        char_val = *inP;
        inP++;
        memset( outP, char_val, char_count );
        outP = outP + char_count;
        }

    if ( (outP - out_buff) != orig_size )
        outP = (unsigned char *)NULL;       /* place for debug check */
    return;
}







static int add_bytes(wrtbuf_t *memP, mbchar *p, int cnt) {
    if ( memP->cnt + cnt > memP->size ) {
        do memP->size *= 2;
            while ( memP->cnt + cnt > memP->size );
        memP->buf = AxRealloc(memP->buf,memP->size);
    }
    memcpy(memP->buf+memP->cnt,p,cnt);
    memP->cnt += cnt;
}

static add_byte(wrtbuf_t *memP, mbchar b) {
    add_bytes(memP,&b,1);
}

static add_long(wrtbuf_t *memP, int l) {
    l = htonl(l);
    return(add_bytes(memP,(char *)&l,sizeof(int)));  /*  type fixing - lhunt */
}

static add_unicode_string (wrtbuf_t *memP, int len, wchar *buf) {
	int ix = 0;
	add_long(memP,len);
	for (ix = 0; ix < len; ++ix) {
		unsigned short sh = (unsigned short)htons(buf[ix]);
		add_bytes(memP,(char*)&sh,sizeof(short));
	}
}

static add_unicode_c_string (wrtbuf_t *memP, char *p) {
	int ix = 0;
    int len = strlen(p);
    add_long(memP,len);
	for (ix = 0; ix < len; ++ix) {
		add_byte(memP,0);
		add_byte(memP,p[ix]);
	}
}

static add_short_unicode_c_string (wrtbuf_t *memP, char *p) {
	int ix = 0;
    int len = strlen(p);
    add_byte(memP,len);
	for (ix = 0; ix < len; ++ix) {
		add_byte(memP,0);
		add_byte(memP,p[ix]);
	}
}

static add_c_string(wrtbuf_t *memP, mbchar *p) {
    int len;

    len = strlen(p) + 1;
    add_long(memP,len);
    add_bytes(memP,p,len);
}

static add_short_string(wrtbuf_t *memP, mbchar *p) {
    int len = strlen(p);
    add_byte(memP,len);
    add_bytes(memP,p,len);
}

static read_long(rdbuf_t *memP, int *l) {
    int buf;

    memcpy(&buf,memP->buf+memP->cnt,sizeof(int));
    memP->cnt += sizeof(int);
    *l = ntohl(buf);
}


static read_bytes(rdbuf_t *memP, mbchar *p, int cnt ) {
    memcpy(p,memP->buf+memP->cnt,cnt);
    memP->cnt += cnt;
}

static int read_byte (rdbuf_t *memP) {
	int rb = memP->buf[memP->cnt++];
    return (rb & 0xFF);
}

/*************************************************************************
function: ReadSomeElfData - Recursive Read array into elf
*/
static data_ptr ReadSomeElfData (rdbuf_t *memP)
{
    int type;
    int templong;
    int orig_size, compressed_size;
    char str[100];
    char *buf, *mbP;
    data_ptr dP;
    int ix;

	type = read_byte(memP);

	if (type == 0)
	{
		read_byte(memP);
		read_byte(memP);
		type = read_byte(memP);
	}

    switch ( type ) {
        case SOCK_INT:
            read_long(memP,&templong);
            return(AxMakeFloatData((double)templong));
            break;

		case SOCK_STR_SMALL_UNICODE:
			templong = read_byte(memP);
			buf = (char*)AxMalloc(templong*2);
			read_bytes(memP,buf,templong*2);
#if BUILD_WIDE
			mbP = elfUnicodeStrToMbStr(buf,&templong);
            dP = AxMakeStrData(templong,"");
			memcpy(SX(dP)->text,mbP,templong);
#else
            dP = AxMakeStrData(templong,"");
			for (ix = 0; ix < templong; ++ix) {
				SX(dP)->text[ix] = buf[ix*2+1];
			}
#endif
			SX(dP)->text[templong] = 0;
			AxFree(buf);
            return(dP);
			break;

		case SOCK_STRING_UNICODE:
            read_long(memP,&templong);
			buf = (char*)AxMalloc(templong*2);
			read_bytes(memP,buf,templong*2);
#if BUILD_WIDE
			mbP = elfUnicodeStrToMbStr(buf,&templong);
            dP = AxMakeStrData(templong,"");
			memcpy(SX(dP)->text,mbP,templong);
#else
            dP = AxMakeStrData(templong,"");
			for (ix = 0; ix < templong; ++ix) {
				SX(dP)->text[ix] = buf[ix*2+1];
			}
#endif
			SX(dP)->text[templong] = 0;
			AxFree(buf);
            return(dP);
			break;

        case SOCK_STRING:
            read_long(memP,&templong);
            dP = AxMakeStrData(templong-1,"");
            read_bytes(memP,SX(dP)->text,templong);
            return(dP);
            break;

        case SOCK_FLOAT:
            read_long(memP,&templong);
            read_bytes(memP,str,templong);
            return(AxMakeFloatData(atof(str)));
            break;

        case SOCK_BINARY:
            read_long(memP,&templong);
            dP = AxMakeBinaryData(templong, (char *)NULL);
            buf = ((data_binary*)dP)->bytes;
            read_bytes(memP,buf,templong);
            return(dP);
            break;

        case SOCK_BINARY_RLE:
            read_long(memP, &orig_size );
            read_long(memP, &compressed_size );

            buf = AxMalloc( compressed_size );
            read_bytes(memP, buf, compressed_size );
            dP = AxMakeBinaryData( orig_size, (char *)NULL);
            UncompressRLE( (unsigned char *)buf, 
                           (unsigned char *)(((data_binary*)dP)->bytes), 
                           compressed_size, orig_size );
			AxFree(buf);
            return(dP);
            break;

#if !ELF_COMM_STAND_ALONE
        case SOCK_OBJECT: {
            int dex,uid;
            
            read_long(memP,&dex);
            read_long(memP,&uid);
            dP = ELF_CREATE_OBJECT(dex,uid,NULL);
            return(dP);
			break;
        }
#endif

        case SOCK_ARRAY:
            read_long(memP,&templong);
            dP = AxMakeArray(templong);
            for ( ix = 0; ix < templong; ix++ )
                dP = AxWriteArray(dP,ix,ReadSomeElfData(memP));
            return(dP);
            break;
	/**
	 * Packed ELF IO data types
	 */
		case SOCK_I0:
		case SOCK_I1:
		case SOCK_I2:
		case SOCK_I3:
		case SOCK_I4:
		case SOCK_I5:
		case SOCK_I6:
		case SOCK_I7:
		case SOCK_I8:
		case SOCK_I9:
		case SOCK_I10:
		case SOCK_I11:
		case SOCK_I12:
		case SOCK_I13:
		case SOCK_I14:
		case SOCK_I15:
			return AxMakeIntData(type-SOCK_I0);
			break;
		case SOCK_INEG1:
			return AxMakeIntData(-1);
			break;
		case SOCK_I127:
			return AxMakeIntData(127);
			break;
		case SOCK_I128:
			return AxMakeIntData(128);
			break;
		case SOCK_I255:
			return AxMakeIntData(255);
			break;
		case SOCK_I256:
			return AxMakeIntData(256);
			break;
		case SOCK_BYTE:	/* signed byte */
			templong = read_byte(memP);
			if (templong > 127)
				templong -= 256;
			return AxMakeIntData(templong);
			break;
		case SOCK_WORD:	/* signed word */
			templong = read_byte(memP);
			templong = (templong << 8) | read_byte(memP);
			if (templong > 32767)
				templong -= 65536;
			return AxMakeIntData(templong);
			break;
		case SOCK_STR_EMPTY:	/* empty string */
			return AxMakeStrData(-1,"");
			break;
		case SOCK_STR_SPACE:	/* space string */
			return AxMakeStrData(-1," ");
			break;
		case SOCK_STR_SMALL:	/* small string, length < 128 */
			templong = read_byte(memP);
            dP = AxMakeStrData(templong,"");
            read_bytes(memP,SX(dP)->text,templong);
			SX(dP)->text[templong] = 0;
            return dP;
			break;
		case SOCK_STR_CHAR:		/* one char string */
            dP = AxMakeStrData(1,"");
            read_bytes(memP,SX(dP)->text,1);
			SX(dP)->text[1] = 0;
			return dP;
			break;
		case SOCK_FLOAT_SHORT:	/* short float */
			templong = read_byte(memP);
            read_bytes(memP,str,templong);
			str[templong+1] = 0;
            return AxMakeFloatData(atof(str));
			break;
		case SOCK_ARR_0:	/* zero size array */
		case SOCK_ARR_1:	/* array of size 1 */
		case SOCK_ARR_2:	/* array of size 2 */
		case SOCK_ARR_3:	/* array of size 3 */
		case SOCK_ARR_4:	/* array of size 4 */
		case SOCK_ARR_SMALL:	/* small array, length < 128 */
			templong = (type == SOCK_ARR_SMALL) ? read_byte(memP) : (type-SOCK_ARR_0);
            dP = AxMakeArray(templong);
            for (ix = 0; ix < templong; ++ix)
                dP = AxWriteArray(dP,ix,ReadSomeElfData(memP));
            return dP;
			break;

		case SOCK_NULLS:
			memP->buf[memP->cnt]--;
			if ( memP->buf[memP->cnt] == 0 )
				memP->cnt++;
			else
				memP->cnt--;
			return(0);
			break;

        case SOCK_NULL:	/* null datum */
        default:
            return(0);
    }
}

/*************************************************************************
function: AxMakeArray - Create array with initial size against THIMpid
*/
DLL_EXPORT data_ptr 
AxMakeArray (elements)
register int elements;
{
    return(AxTaskMakeArray(THIMpid, elements));
}

/*************************************************************************
function: AxTaskMakeArray - Create array with initial size - any task
*/
DLL_EXPORT data_ptr 
AxTaskMakeArray (pid, elements)
register int pid;
register int elements;
{
    register data_array *p;
    register unsigned int size;

    size = sizeof(data_array) + (sizeof(data_ptr) * (elements - 1));
    p = (data_array*)AxTaskPoolAllocate(pid,size);
    MEMSET(p,0,size);
    p->type = TYPE_ARRAY;
    p->size = elements;
    return((data_ptr)p);
}

/*************************************************************************
function: AxMakeStrData - Allocate a string object against THIMpid
*/
DLL_EXPORT data_ptr 
AxMakeStrData (len,str)
register int len;       /* length of string or -1 */
register mbchar *str;     /* null terminated intial string */
{
    return( AxTaskMakeStrData(THIMpid,len,str));
}

/*************************************************************************
function: AxTaskMakeStrData - Allocate a string object - any task
*/
DLL_EXPORT data_ptr 
AxTaskMakeStrData (pid,len,str)
register int pid;       /* Owner of memory - THIMpid style number */
register int len;       /* length of string or -1 */
register mbchar *str;     /* null terminated intial string */
{
    register data_string *p;
    register unsigned int ix;

    if ( !str )
        str = (mbchar *)"";
    if ( len == -1 ) {
        len = MBstrlen(str);
        p = (data_string *)AxTaskPoolAllocate(pid,sizeof(data_string)+len);
        MBstrcpy(p->text,str);
    }
    else {
        p = (data_string *)AxTaskPoolAllocate(pid,sizeof(data_string)+len);
        for ( ix = 0; (ix < len) && *str; ix++ )
            p->text[ix] = str[ix];
        p->text[len] = 0;
    }
    p->type = TYPE_STRING;
    return((data_ptr)p);
}


mbchar *WnToMbOneStr();

/*************************************************************************
function: AxMakeWideStrData - Allocate a string object against THIMpid
*/
DLL_EXPORT data_ptr 
AxMakeWideStrData (w_len,str)
register int w_len;       /* length of string or -1 */
register wchar *str;     /* null terminated wide-character string */
{
    return( AxTaskMakeWideStrData(THIMpid,w_len,str));
}

/*************************************************************************
function: AxTaskMakeWideStrData - Allocate a string object - any task
NOTE: This function differs in model against AxMakeStrData.  In the
latter, if the len is not -1, that DICTATES the buffer size, even
though the string may be smaller.  The function just uses len
as a convenience.
*/
DLL_EXPORT data_ptr 
AxTaskMakeWideStrData (pid,w_len,str)
register int pid;       /* Owner of memory - THIMpid style number */
register int w_len;       /* length of string or -1 */
register wchar *str;     /* null terminated wide-character string */
{
    mbchar *buff;
    data_ptr retP;

    if ( str == (wchar *)NULL )
        {
        retP = AxTaskMakeStrData(pid,-1,(mbchar *)NULL);
        return( retP );
        }

    if ( w_len == -1 )
        w_len = Wstrlen(str);

    buff = WnToMbOneStr(str, w_len);
    retP = AxTaskMakeStrData(pid,-1,buff);
    return( retP );
}


/***************************************************************************
function: AxMakeObjectData -  create data_object object with dex/uid
*/
DLL_EXPORT data_ptr
AxMakeObjectData (objdex, objuid)
    int objdex, objuid;
{
    return(AxTaskMakeObjectData(THIMpid,objdex, objuid));
}

/***************************************************************************
function: AxTaskMakeObjectData -  create data_object object with dex/uid
*/
DLL_EXPORT data_ptr
AxTaskMakeObjectData (pid,objdex, objuid)
    register int pid;       /* Owner of memory - THIMpid style number */
    int objdex, objuid;
{
    data_object *p;
    
    p = (data_object *)AxTaskPoolAllocate(pid, sizeof(data_object));
    memset(p,0,sizeof(data_object));
    p->type = TYPE_OBJECT;
    p->objHostname = NULL;
    p->objDex = objdex;
    p->objUid = objuid;
    return((data_ptr)p);
}


/*************************************************************************
function: AxMakeBinaryData - Create binary object with initial size
*/
DLL_EXPORT data_ptr 
AxMakeBinaryData (elements,dataBytes)
register int elements;
mbchar *dataBytes;    /* may be NULL */
{
    register data_binary *p;
    register unsigned int size;

    size = sizeof(data_binary) + elements - 1;
    p = (data_binary*)AxMalloc(size);
    p->type = TYPE_BINARY;
    p->size = elements;
    if (dataBytes) {
        memcpy(p->bytes, dataBytes, elements);
    }
    else {
        memset(p->bytes,0,elements);
    }
    return((data_ptr)p);
}






/*************************************************************************
function: AxMakeFloatData - Allocate a float object against THIMpid
*/
DLL_EXPORT data_ptr 
AxMakeFloatData (val)
double val;
{
    return( AxTaskMakeFloatData(THIMpid,val));
}

/*************************************************************************
function: AxTaskMakeFloatData - Allocate a float object - any task
*/
DLL_EXPORT data_ptr 
AxTaskMakeFloatData (pid,val)
register int pid;       /* Owner of memory - THIMpid style number */
double val;
{
    register data_number *p;
    register int num;

    if ( C_DBL_IS_ELF_INT(val,num) ) {
        return((data_ptr)MK_INTEGER(num));
    }
    p = (data_number*)AxTaskPoolAllocate(pid,sizeof(data_number));
    p->type = TYPE_NUMBER;
    p->value = val;
    return((data_ptr)p);
}

/*************************************************************************
function: AxMakeIntData - Allocate a number against THIMpid
*/
DLL_EXPORT data_ptr 
AxMakeIntData (num)
register int num;
{
    return( AxTaskMakeIntData(THIMpid,num));
}

/*************************************************************************
function: AxMakeBoolData - Allocate a number against THIMpid
*/
DLL_EXPORT data_ptr 
AxMakeBoolData (num)
register int num;
{
    return( AxTaskMakeIntData(THIMpid,num ? -1 : 0));
}

/*************************************************************************
function: AxTaskMakeIntData - Allocate a number - any task
*/
DLL_EXPORT data_ptr 
AxTaskMakeIntData (pid,num)
register int pid;       /* Owner of memory - THIMpid style number */
register int num;
{
    register data_number *p;

    if ( C_INT_IS_ELF_INT(num) )
        p = (data_number*)MK_INTEGER(num);
    else {
        p = (data_number*)AxTaskPoolAllocate(pid,sizeof(data_number));
        p->type = TYPE_NUMBER;
        p->value = num;
    }
    return((data_ptr)p);
}

/*************************************************************************
function: AxWriteArray - Write element, may increase size of array
*/
DLL_EXPORT data_ptr 
AxWriteArray (p,dex,new_val)
register data_ptr p;
register int dex;
register data_ptr new_val;
{
    return(AxTaskWriteArray(THIMpid,p,dex,new_val));
}

/*************************************************************************
function: AxOverwriteArray - Copy new value into array slot, without freeing old
Like AxTaskWriteArray, but does not free old.
Cannot 
*/
DLL_EXPORT void
AxOverwriteArray (p,dex,new_val)
register data_ptr p;
register int dex;
register data_ptr new_val;
{
    CHECK(AxIsArray(p),ERR_NOTARRAY,XLT("AxWriteArray arg is not an array"));
    ((data_array*)p)->vals[dex] = new_val;
}

/*************************************************************************
function: AxTaskWriteArray - Write element, may increase size of array
*/
DLL_EXPORT data_ptr 
AxTaskWriteArray (pid,p,dex,new_val)
register int pid;       /* Owner of memory - THIMpid style number */
register data_ptr p;
register int dex;
register data_ptr new_val;
{
    register int old_bytes,new_bytes;

    CHECK(AxIsArray(p),ERR_NOTARRAY,XLT("AxWriteArray arg is not an array"));
    old_bytes = sizeof(data_array)+(((data_array*)p)->size-1)*sizeof(data_ptr);
    new_bytes = sizeof(data_array)+dex*sizeof(data_ptr);
    if ( new_bytes > old_bytes ) {
        p = (data_ptr)AxTaskPoolRealloc(pid,
                (char*)XX(p),(unsigned)new_bytes,old_bytes);
        MEMSET(((char*)p)+old_bytes,0,new_bytes-old_bytes);
        ((data_array*)p)->size = dex+1;
    }
    AxTaskFreeData(pid,((data_array*)p)->vals[dex]);    /* in case there was one there */
    ((data_array*)p)->vals[dex] = new_val;
    return((data_ptr)p);
}

/*************************************************************************
function: AxTruncArray - Decrease size of an array against THIMpid
*/
DLL_EXPORT data_ptr 
AxTruncArray (p,newsize)
register data_ptr p;
register int newsize;
{
    return(AxTaskTruncArray(THIMpid,p,newsize));
}

/*************************************************************************
function: AxTaskTruncArray - Decrease size of an array - any task
*/
DLL_EXPORT data_ptr 
AxTaskTruncArray (pid,p,newsize)
register int pid;       /* Owner of memory - THIMpid style number */
register data_ptr p;
register int newsize;
{
    register int ix, new_bytes, old_bytes;

    CHECK(newsize >= 0,ERR_ILLVAL,XLT("AxTruncArray passed a negative number"));
    if ( AX(p)->size <= newsize )
        return(p);
    for ( ix = newsize; ix < AX(p)->size; ix++ )
        AxTaskFreeData(pid,AX(p)->vals[ix]);
    new_bytes = sizeof(data_array)+(newsize-1)*sizeof(data_ptr);
    old_bytes = sizeof(data_array)+(AX(p)->size-1)*sizeof(data_ptr);
    p = (data_ptr)AxTaskPoolRealloc(pid,
            (char*)XX(p),(unsigned)new_bytes,old_bytes);
    AX(p)->size = newsize;
    return(p);
}

/*************************************************************************
function: AxAppendArray - WriteArray elements of one array to another
New array references SAME (not copies of) elements in old arrays.
Accepts NULL at first array.
*/
DLL_EXPORT data_ptr 
AxAppendArray (p1,p2)
register data_ptr p1, p2;
{
    return(AxTaskAppendArray(THIMpid,p1,p2));
}

/*************************************************************************
function: AxTaskAppendArray - WriteArray elements of one array to another
*/
DLL_EXPORT data_ptr 
AxTaskAppendArray (pid,p1,p2)
register int pid;       /* Owner of memory - THIMpid style number */
register data_ptr p1, p2;
{
    register data_ptr rP;
    register int p2size, p1size;

    CHECK(!p1 || AxIsArray(p1),ERR_NOTARRAY,XLT("AxAppendArray arg is not an array"));
    CHECK(AxIsArray(p2),ERR_NOTARRAY,XLT("AxAppendArray arg is not an array"));
    p1size = p1 ? AX(p1)->size : 0;
    p2size = AX(p2)->size;
    rP = AxTaskMakeArray(pid,p1size + p2size);
    memcpy(&(AX(rP)->vals[0]),AX(p1)->vals,p1size*sizeof(data_ptr));
    memcpy(&(AX(rP)->vals[p1size]),AX(p2)->vals,p2size*sizeof(data_ptr));
    return((data_ptr)rP);
}

/*************************************************************************
function: AxAddStrToArray - Put string into array as string object
*/
DLL_EXPORT data_ptr 
AxAddStrToArray (ap,dex,str)
data_ptr ap;
int dex;
mbchar *str;
{
    return(AxTaskAddStrToArray(THIMpid,ap,dex,str));
}

/*************************************************************************
function: AxTaskAddStrToArray - Put string into array as string object
*/
DLL_EXPORT data_ptr 
AxTaskAddStrToArray (pid,ap,dex,str)
register int pid;       /* Owner of memory - THIMpid style number */
data_ptr ap;
int dex;
mbchar *str;
{
    return(AxTaskWriteArray(pid,ap,dex,AxTaskMakeStrData(pid,-1,str)));
}

/*************************************************************************
function: AxAddWideStrToArray - Put a wide-character  string into array 
          as string object
*/
DLL_EXPORT data_ptr 
AxAddWideStrToArray (ap,dex,str)
data_ptr ap;
int dex;
wchar *str;
{
    return(AxTaskAddWideStrToArray(THIMpid,ap,dex,str));
}

/*************************************************************************
function: AxTaskAddWideStrToArray - Put a wide-character  string into array 
          as string object
*/
DLL_EXPORT data_ptr 
AxTaskAddWideStrToArray (pid,ap,dex,str)
data_ptr ap;
int dex;
wchar *str;
{
    return(AxTaskWriteArray(pid,ap,dex,AxTaskMakeWideStrData(pid,-1,str)));
}



/*************************************************************************
function: AxAddBinaryToArray - Put binary data into array against THIMpid
*/
DLL_EXPORT data_ptr 
AxAddBinaryToArray (ap,dex,dataSize,binaryData)
data_ptr ap;
int dex;
int dataSize;
mbchar *binaryData;
{
    return(AxTaskAddBinaryToArray(THIMpid,ap,dex,dataSize,binaryData));
}

/*************************************************************************
function: AxTaskAddBinaryToArray - Put binary data into array - any task
*/
DLL_EXPORT data_ptr 
AxTaskAddBinaryToArray (pid,ap,dex,dataSize,binaryData)
register int pid;       /* Owner of memory - THIMpid style number */
data_ptr ap;
int dex;
int dataSize;
mbchar *binaryData;
{
    return(AxTaskWriteArray(pid,ap,dex,AxMakeBinaryData(dataSize,binaryData)));
}


/*************************************************************************
function: AxAddIntToArray - Put integer into array against THIMpid
*/
DLL_EXPORT data_ptr 
AxAddIntToArray (ap,dex,n)
data_ptr ap;
int dex;
int n;
{
    return(AxTaskAddIntToArray(THIMpid,ap,dex,n));
}

/*************************************************************************
function: AxTaskAddIntToArray - Put integer into array - any task
*/
DLL_EXPORT data_ptr 
AxTaskAddIntToArray (pid,ap,dex,n)
register int pid;       /* Owner of memory - THIMpid style number */
data_ptr ap;
int dex;
int n;
{
    return(AxTaskWriteArray(pid,ap,dex,AxTaskMakeIntData(pid,n)));
}

/*************************************************************************
function: AxAddBoolToArray - Put integer into array (ELF TRUE/FALSE)
*/
DLL_EXPORT data_ptr 
AxAddBoolToArray (ap,dex,n)
data_ptr ap;
int dex;
int n;
{
    return(AxTaskAddBoolToArray(THIMpid,ap,dex,n));
}

/*************************************************************************
function: AxTaskAddBoolToArray - Put integer into array (ELF TRUE/FALSE)
*/
DLL_EXPORT data_ptr 
AxTaskAddBoolToArray (pid,ap,dex,n)
register int pid;       /* Owner of memory - THIMpid style number */
data_ptr ap;
int dex;
int n;
{
    return(AxTaskWriteArray(pid,ap,dex,AxTaskMakeIntData(pid,n ? -1 : 0)));
}

/*************************************************************************
function: AxAddFloatToArray - Put number into array against THIMpid
*/
DLL_EXPORT data_ptr 
AxAddFloatToArray (ap,dex,n)
data_ptr ap;
int dex;
double n;
{
    return(AxTaskAddFloatToArray(THIMpid,ap,dex,n));
}

/*************************************************************************
function: AxTaskAddFloatToArray - Put number into array - any task
*/
DLL_EXPORT data_ptr 
AxTaskAddFloatToArray (pid,ap,dex,n)
register int pid;       /* Owner of memory - THIMpid style number */
data_ptr ap;
int dex;
double n;
{
    return(AxTaskWriteArray(pid,ap,dex,AxTaskMakeFloatData(pid,n)));
}

/*************************************************************************
function: AxAddArrayToArray - Put number into array against THIMpid
*/
DLL_EXPORT data_ptr 
AxAddArrayToArray (ap,dex,n)
data_ptr ap;
int dex;
data_ptr n;
{
    return(AxTaskAddArrayToArray(THIMpid,ap,dex,n));
}

/*************************************************************************
function: AxTaskAddArrayToArray - Put number into array - any task
*/
DLL_EXPORT data_ptr 
AxTaskAddArrayToArray (pid,ap,dex,n)
register int pid;       /* Owner of memory - THIMpid style number */
data_ptr ap;
int dex;
data_ptr n;
{
    return((data_ptr)AxTaskWriteArray(pid,ap,dex,n));
}

/*************************************************************************
function: AxAddDataToArray - Put number into array against THIMpid
*/
DLL_EXPORT data_ptr 
AxAddDataToArray (ap,dex,n)
data_ptr ap;
int dex;
data_ptr n;
{
    return(AxTaskAddDataToArray(THIMpid,ap,dex,n));
}

/*************************************************************************
function: AxTaskAddDataToArray - Put number into array - any task
*/
DLL_EXPORT data_ptr 
AxTaskAddDataToArray (pid,ap,dex,n)
data_ptr ap;
int dex;
data_ptr n;
{
    return((data_ptr)AxTaskWriteArray(pid,ap,dex,n));
}


/*************************************************************************
function: AxArraySize - Return size of an array
*/
DLL_EXPORT int AxArraySize (arg)
register data_ptr arg;
{
    if ( !arg )
        return(0);
    CHECK(AxIsArray(arg),ERR_NOTARRAY,XLT("AxArraySize arg is not an array"));
    return(((data_array*)(arg))->size);
}

/*************************************************************************
function: AxArrayElement - Fetch element (0 if beyond array end)
*/
DLL_EXPORT data_ptr AxArrayElement (ap,dex)
register data_ptr ap;
register int dex;
{
    if ( !ap )
        return(0);
    CHECK(AxIsArray(ap),ERR_NOTARRAY,XLT("AxArrayElement arg is not an array"));
    if ( ((data_array *)ap)->size <= (unsigned)dex )
        return(0);
    return(((data_array *)ap)->vals[dex]);
}

/*************************************************************************
function: AxArrayDataAddr - Return address of first datum
*/
DLL_EXPORT data_ptr *AxArrayDataAddr (ap)
register data_ptr ap;
{
    if ( !ap )
        return(NULL);
    CHECK(AxIsArray(ap),ERR_NOTARRAY,XLT("AxArrayElement arg is not an array"));
    return(((data_array *)ap)->vals);
}

/*************************************************************************
function: AxIsArray - Return whether datum is an array
*/
DLL_EXPORT int AxIsArray (p)
register data_ptr p;
{    
    return(p && !IS_INTEGER(p) && (p->type == TYPE_ARRAY));
}

/*************************************************************************
function: AxCanBeString - Return whether datum is a null/number/string
*/
DLL_EXPORT int AxCanBeString (p)
register data_ptr p;
{
	return(!p || IS_INTEGER(p) || (p->type == TYPE_STRING) || (p->type == TYPE_NUMBER));
}

/*************************************************************************
function: AxArrayFromArray - Return element
*/
DLL_EXPORT data_ptr AxArrayFromArray (ap,dex)
register data_ptr ap;
register int dex;
{
    return(AxArrayElement(ap,dex));
}

/*************************************************************************
function: AxIntFromArray - FOR NOW
*/
DLL_EXPORT int AxIntFromArray (ap,dex)
register data_ptr ap;
register int dex;
{
    ap = AxArrayElement(ap,dex);
    return(AxIntFromDataPtr(ap));
}

/*************************************************************************
function: AxBoolFromArray - FOR NOW
*/
DLL_EXPORT int AxBoolFromArray (ap,dex)
register data_ptr ap;
register int dex;
{
    ap = AxArrayElement(ap,dex);
    return(AxIntFromDataPtr(ap) ? TRUE : FALSE );
}

/*************************************************************************
function: AxIntFromDataPtr - Very similar to the one in elintr.c!
*/
DLL_EXPORT int AxIntFromDataPtr (dP)
register data_ptr dP;
{
    double dval;

    if ( IS_INTEGER(dP) )
        return( GET_INTEGER(dP) );
    dP = (data_ptr)((long)dP & ~FAULT_BIT);/* allow pops of read-only values */
    if ( !XX(dP) )
        return(0);
    dval = AxFloatFromDataPtr(dP);
    if ( dval >= 0.0 )
        dval += 0.5;
    else
        dval -= 0.5;
    return(dval);
}

/*************************************************************************
function: AxBoolFromDataPtr - FOR NOW
*/
DLL_EXPORT int AxBoolFromDataPtr (ap)
register data_ptr ap;
{
    return(AxIntFromDataPtr(ap) ? TRUE : FALSE);
}

/*************************************************************************
function: AxFloatFromArray - fetch a float
*/
DLL_EXPORT double AxFloatFromArray (ap,dex)
register data_ptr ap;
register int dex;
{
    ap = AxArrayElement(ap,dex);
    return(AxFloatFromDataPtr(ap));
}

/*************************************************************************
function: AxFloatFromDataPtr - Fetch a float
*/
DLL_EXPORT double AxFloatFromDataPtr (dP)
register data_ptr dP;
{
    double retval;
    register char *p;
    char *dst;

    if ( IS_INTEGER(dP) )
        return( GET_INTEGER(dP) );
    dP = (data_ptr)((long)dP & ~FAULT_BIT);/* allow pops of read-only values */
    if ( !XX(dP) )
        return(0.0);
    if ( XX(dP)->type == TYPE_NUMBER )
        return(NX(dP)->value);
    if ( XX(dP)->type == TYPE_STRING ) {
        p = SX(dP)->text;   /* convert spaces to 0.0 */
        if ( !(*p) )
            return(0.0);
        retval = Astrtod(p,&dst);
        while ( (*dst != 0) && isspace(*dst) ) dst++;
        if ( *dst != 0 ) {
            CHECK(FALSE,ERR_NOTNUMBER,XLT("AxFloatFromDataPtr cannot convert value"));
        }
        return(retval);
    }
    CHECK(FALSE,ERR_NOTNUMBER,XLT("AxFloatFromDataPtr cannot convert value"));
}
/*************************************************************************
function: AxIsFloatable - Return TRUE if dP is readable by AxFloatFromDataPtr
datum
*/
DLL_EXPORT int AxIsFloatable(dP)
register data_ptr dP;
{
    register char *p;
    double retval;
    char *dst;

    if ( IS_INTEGER(dP) )
        return(TRUE);

    /* allow pops of read-only values */
    dP = (data_ptr)((long)dP & ~FAULT_BIT);
    if ( !XX(dP) )
        return(TRUE);

    if ( XX(dP)->type == TYPE_NUMBER )
        return(TRUE);

    if ( XX(dP)->type == TYPE_STRING ) {
        p = SX(dP)->text;   /* convert spaces to 0.0 */
        if ( !(*p) )
            return(TRUE);
          Astrtod(p,&dst);
        while ( (*dst != 0) && isspace(*dst) ) dst++;
        if ( *dst == 0 )
            return(TRUE);
    }
    return(BFALSE);
}


/*************************************************************************
function: AxStrPtrFromArray - Fetch string
*/
DLL_EXPORT mbchar *AxStrPtrFromArray (ap,dex)
register data_ptr ap;
register int dex;
{
    return(AxStrPtrFromDataPtr(AxArrayElement(ap,dex)));
}


/*************************************************************************
function: AxStrFromArray - Fetch string
*/
DLL_EXPORT mbchar *AxStrFromArray (ap,dex)
register data_ptr ap;
register int dex;
{
    return((mbchar *)AxStrFromDataPtr(AxArrayElement(ap,dex)));
}


/***************************************************************************
function: AxBinaryFromDataPtr - get binary data, returns data size
*/
DLL_EXPORT int AxBinaryFromDataPtr(ap, binP)
register data_ptr ap;
mbchar **binP; /* out: points to binary data */
{
    int size;
    size = AxBinarySize(ap);
    *binP = ((data_binary*)ap)->bytes;
    return size;
}


/***************************************************************************
function: AxBinaryFromArray - get binary data array element
*/
DLL_EXPORT int AxBinaryFromArray(ap, dex, binP)
register data_ptr ap;
register int dex;
mbchar **binP; /* out: points to binary data */
{
    return (AxBinaryFromDataPtr(AxArrayElement(ap,dex),binP));
}


/*************************************************************************
function: AxStrPtrFromDataPtr - Fetch string
*/
DLL_EXPORT mbchar *AxStrPtrFromDataPtr (ap)
register data_ptr ap;
{
    static char buf[30];

    if ( !ap )
        return((mbchar *)"");
    if ( IS_INTEGER(ap) ) {
        sprintf(buf,"%d",GET_INTEGER(ap));
        return( (mbchar *)buf );
    }
    ap = (data_ptr)((long)ap & ~FAULT_BIT);/* allow pops of read-only values */
    if ( ap->type == TYPE_STRING )
        return((mbchar *)((data_string*)ap)->text);
    if ( ap->type == TYPE_NUMBER ) {
        double val;
        int valint;
        val = ((data_number*)ap)->value;
        valint = val;
        if ( valint == val )
            sprintf(buf,"%d",valint);
        else {
            sprintf(buf,"%.15g",val);
            ConvertUSDotsToEuropeComma(buf);
        }
        return( (mbchar *)buf );
    }
    CHECK(FALSE,ERR_NOTSTRING,XLT("AxStrFromDataPtr cannot convert value"));
}


#if !NOLIB_THIMBLE


/*************************************************************************
function: AxWideStrPtrFromDataPtr - Fetch  wide-character string
*/
DLL_EXPORT wchar *AxWideStrPtrFromDataPtr (ap)
register data_ptr ap;
{
    mbchar *orig_string;
    wchar *ret_str;
    int len;

    orig_string = AxStrPtrFromDataPtr( ap );
    len = MBstrlen(orig_string) + 1;       /* max number of chars (not bytes) */
    ret_str = (wchar *)get_static_buffer( len * sizeof(wchar) );

    MbToWStr( orig_string, ret_str );
    return( ret_str );
}



/*************************************************************************
function: AxWideStrPtrFromArray - Fetch string
*/
DLL_EXPORT wchar *AxWideStrPtrFromArray (ap,dex)
register data_ptr ap;
register int dex;
{
    mbchar *orig_string;
    wchar *ret_str;
    int len;

    orig_string = AxStrPtrFromDataPtr( AxArrayElement(ap,dex) );
    len = MBstrlen(orig_string) + 1;       /* max number of chars (not bytes) */
    ret_str = (wchar *)get_static_buffer( len * sizeof(wchar) );

    MbToWStr( orig_string, ret_str );
    return( ret_str );
}
#endif


/*************************************************************************
function: AxFreeArrayButNotContents - Free just the array wrapper
*/
DLL_EXPORT void 
AxFreeArrayButNotContents (p)
register data_ptr p;
{
    AxTaskFreeArrayButNotContents(THIMpid,p);
}

/*************************************************************************
function: AxTaskFreeArrayButNotContents - Free just the array wrapper
*/
DLL_EXPORT void 
AxTaskFreeArrayButNotContents (pid,p)
register int pid;       /* Owner of memory - THIMpid style number */
register data_ptr p;
{
    if ( !p || ( ((long)p) & FAULT_BIT) )
        return;
    AxTaskPoolFree(pid,(char*)p,
             sizeof(data_array)+(AX(p)->size-1)*sizeof(data_ptr));
}

/*************************************************************************
function: AxFreeData - Free a datum (recurse thru arrays) from THIMpid
*/
DLL_EXPORT void 
AxFreeData (p)
register data_ptr p;
{
    AxTaskFreeData(THIMpid,p);
}

/*************************************************************************
function: AxTaskFreeData - Free a datum (recurse thru arrays) - any task
*/
DLL_EXPORT void 
AxTaskFreeData (pid,p)
register int pid;       /* Owner of memory - THIMpid style number */
register data_ptr p;
{
    register unsigned int ix;

    if ( !p || ( ((long)p) & FAULT_BIT) )
        return;
    switch ( p->type ) {
    case TYPE_ARRAY:
        for ( ix = 0; ix < ((data_array*)p)->size; ix++ )
            AxTaskFreeData(pid,((data_array*)p)->vals[ix]);
        AxTaskPoolFree(pid,(char*)p,
                sizeof(data_array)+(AX(p)->size-1)*sizeof(data_ptr));
        break;
    case TYPE_NUMBER:
        AxTaskPoolFree(pid,(char*)p,sizeof(data_number));
        break;
    case TYPE_STRING:
        AxTaskPoolFree(pid,(char*)p,
                        sizeof(data_string)+strlen(SX(p)->text));
        break;
#if !ELF_COMM_STAND_ALONE
    case TYPE_OBJECT:
        ElfObjDecRef((data_object*)p);
        AxTaskPoolFree(pid, (char *)p, sizeof(data_object));     
        break;
#endif
    case TYPE_ENTRY:
        CHECK(FALSE,ERR_ILLVAL,XLT("Attempted to AxTaskFreeData an entry block"));
        break;
    default:
        TaskFree(0,p);     
    }
}

/*************************************************************************
function: AxIsInt - Return Basic TRUE if arg is an integer datum
*/
DLL_EXPORT int AxIsInt (arg)
register data_ptr arg;
{
    if ( IS_INTEGER(arg) )
        return(TRUE);
    else
        return(BFALSE);
}

/*************************************************************************
function: AxIsNumber - Return Basic TRUE if arg is a number (int or float) 
datum
*/
DLL_EXPORT int AxIsNumber (arg)
register data_ptr arg;
{
    if ( IS_INTEGER(arg) )
        return(TRUE);
    else if ( XX(arg) && (NX(arg)->type == TYPE_NUMBER) )
        return(TRUE);
    else
        return(BFALSE);
}

/*************************************************************************
function: AxIsFloat - Return Basic TRUE if arg is a floating point datum
*/
DLL_EXPORT int AxIsFloat (arg)
register data_ptr arg;
{
    if (AxIsNumber( arg)) {
        if ( IS_INTEGER(arg) )
            return(FALSE);
        return(TRUE);
    }
    else
        return(BFALSE);
}

/*************************************************************************
function: AxIsString - Return Basic TRUE if arg is a string datum
*/
DLL_EXPORT int AxIsString (arg)
register data_ptr arg;
{
    if ( XX(arg) && !IS_INTEGER(arg) && (SX(arg)->type == TYPE_STRING) )
        return(TRUE);
    else
        return(BFALSE);
}

/*************************************************************************
function: AxIsBinary - Return Basic TRUE if arg is a binary object
*/
DLL_EXPORT int AxIsBinary (arg)
register data_ptr arg;
{
    if ( XX(arg) && !IS_INTEGER(arg) && (SX(arg)->type == TYPE_BINARY) )
        return(TRUE);
    else
        return(BFALSE);
}

/*************************************************************************
function: AxIsObject - Return Basic TRUE if arg is an object
*/
DLL_EXPORT int AxIsObject (arg)
register data_ptr arg;
{
    if ( arg && !IS_INTEGER(arg) && (arg->type == TYPE_OBJECT) )
        return(TRUE);
    else
        return(BFALSE);
}

/*************************************************************************
function: AxBinarySize - Return size of an array
*/
DLL_EXPORT int AxBinarySize (arg)
register data_ptr arg;
{
    data_binary *bp;

    if ( !XX(arg) )
        return(0);
    CHECK(!IS_INTEGER(arg) && AX(arg)->type == TYPE_BINARY,ERR_ILLVAL,
        NXLT("AxBinarySize not passed a binary object"));
    bp = (data_binary*)arg;
    return(bp->size);
}
/*************************************************************************
function: AxBinaryBytesPtr - Return address of binary block
*/
DLL_EXPORT char *AxBinaryBytesPtr (arg)
register data_ptr arg;
{
    AxBinarySize(arg);  /* just to get the assertion */
    return(BX(arg)->bytes);
}

/*************************************************************************
function: AxFree - Release passed memptr
*/
DLL_EXPORT void AxFree (p)
mbchar *p;
{
#if NOLIB_THIMBLE
    /* built without the standard Aster*x memory handlers */
    if ( p )
        TaskFree(0,p);
#else
    TaskFree(0,p);
#endif
}

static int axHaveMallocHandler = FALSE;
static jmp_buf axMallocHandler;

/*************************************************************************
function: AxMalloc - Safe allocator
*/
DLL_EXPORT char *AxMalloc (n)
int n;
{
    char *p;

    p = (char*)TaskAlloc(0,n);   /* never fails in axmain */
    if ( !p ) {
        if ( axHaveMallocHandler )
            longjmp(axMallocHandler,1);
        else
            CHECK(FALSE,0,XLT("Out of memory in AxMalloc"));
    }
    return(p);
}

/*************************************************************************
function: AxRealloc - Safe allocator
*/
DLL_EXPORT char *AxRealloc (p,n)
mbchar *p;
int n;
{
#if NOLIB_THIMBLE
    /* built without the standard Aster*x memory handlers */
    if ( p )
        p = (char*)TaskRealloc(0,p,n);
    else
        p = (char*)TaskAlloc(0,n);
#else
    p = (char*)TaskRealloc(0,p,n);    /* never fails in axmain */
#endif

    if ( !p ) {
        if ( axHaveMallocHandler )
            longjmp(axMallocHandler,1);
        else
            CHECK(FALSE,0,XLT("Out of memory in AxRealloc"));
    }

    return(p);
}

/*************************************************************************
function: AxMallocAborter - Set abort handler
*/
DLL_EXPORT void AxMallocAborter (abortBuf)
jmp_buf abortBuf;
{
    axHaveMallocHandler = TRUE;
    memcpy(axMallocHandler,abortBuf,sizeof(jmp_buf));
}

int axErrorCode;
mbchar *axErrorString;
mbchar *axErrorObject;
jmp_buf axErrorHandler;

/*************************************************************************
function: AxError - Throw an error back to Aster*x
*/
DLL_EXPORT void AxError (code, str, object)
int code;
mbchar *str, *object;
{
   
    int ix, total;
    mbchar *charP;
    mbchar error_message[2000];

    axErrorCode = code;
    if ( axErrorString ) TaskFree(0,axErrorString);
    if ( axErrorObject ) TaskFree(0,axErrorObject);
    if ( str )
        axErrorString = MBstrcpy(TaskAlloc(0,strlen(str)+1),str);
    else
        axErrorString = MBstrcpy(TaskAlloc(0,2),"");
    if ( object )
        axErrorObject = MBstrcpy(TaskAlloc(0,strlen(object)+1),object);
    else
        axErrorObject = MBstrcpy(TaskAlloc(0,2),"");

    /* if there is a valid long jump buffer set up, ie. something in the global buffer */
    /* is non-zero  indicating that it has been set, it is OK to... LONGJMP OUT!!! */
    charP = (mbchar *)(&axErrorHandler);
    for ( ix = 0; ix < sizeof(axErrorHandler); ix++ ) {
        if ( *charP != 0 )
            longjmp(axErrorHandler,1);
        charP++;
    }
#if !ELF_COMM_STAND_ALONE
    ElfStrAbort(code,str,object);
#endif

    if ( str && str[0] ) {
        if ( object && object[0] )
            sprintf((char *)error_message,"AxError: %s: %s\n",object,str);
        else
            sprintf((char *)error_message,"AxError: %s\n",str);
    }
    else {
        if ( object && object[0] )
            sprintf((char *)error_message,"AxError: %s: Error %d\n",object,code);
        else
            sprintf((char *)error_message,"AxError: Error %d\n",code);
    }
#if AX_X11
    fprintf(stderr,"%s\n",error_message );
#else
#ifndef MS_MessageBox
#define MS_MessageBox MessageBox
#endif   

    MS_MessageBox((HWND)NULL, error_message, "Enterprise Internal Error", 
                                                                     MB_OK );

#endif


    exit(code);
}

/*************************************************************************
function: fullTrim - Trim leading and trailing spaces
*/
static fullTrim (dst,src)
register char *dst, *src;
{
    register char *p = dst;

    /* trim leading spaces */

    while ( (*src == ' ') || (*src == '\t') ) src++;
    if ( src != dst )
        MBstrcpy(dst,src);

    /* trim trailing spaces */

    for ( src = dst + strlen(dst) - 1; src > dst; src-- )
        if ( (*src != ' ') && (*src != '\t') )
            break;
    *(src+1) = 0;
}

#if !ELF_COMM_STAND_ALONE    /*   {   */

/*************************************************************************
function: AxAddDblToArray - Put number into array
!!0 TEMP... replaced by AxAdFloatToArray
*/
DLL_EXPORT data_ptr AxAddDblToArray (ap,dex,n)
data_ptr ap;
int dex;
double n;
{
    return(AxWriteArray(ap,dex,AxMakeFloatData(n)));
}

/*************************************************************************
function: AxMakeNumData - Allocate a number
!!0 TEMP... replaced by AxMakeIntData()
*/
DLL_EXPORT data_ptr AxMakeNumData (num)
register int num;
{
    return(AxMakeIntData(num));
}

/*************************************************************************
function: AxDblFromArray - fetch a float
*/
DLL_EXPORT double AxDblFromArray (ap,dex)
register data_ptr ap;
register int dex;
{
    ap = AxArrayElement(ap,dex);
    return(AxFloatFromDataPtr(ap));
}

#endif     /*  }   !ELF_COMM_STAND_ALONE   */


/*************************************************************************
function: AxBuildArray - Produce array from printf style format
*/
DLL_EXPORT data_ptr AxBuildArray (va_alist)
va_dcl /* where to write results to */
{
    va_list ptr;
    register char *format;   /* format string describing decomposition mode */
    register char *p;
    register int dex = 0;
    data_ptr aP;

    va_start(ptr);
    format = va_arg(ptr,char *);

    aP = AxMakeArray(0);

/* copy in the data */

    for ( p = format; *p; p++, dex++ ) {
        switch ( *p ) {
        case 's':
        case 'c':
            aP = AxAddStrToArray(aP,dex,va_arg(ptr,char*));
            break;
        case 'i':
            aP = AxAddIntToArray(aP,dex,va_arg(ptr,int));
            break;
        case 'b':
            aP = AxAddIntToArray(aP,dex,va_arg(ptr,int) ? -1 : 0);
            break;
        case 'd':
            aP = AxAddFloatToArray(aP,dex,va_arg(ptr,double));
            break;
        case 'f':
            aP = AxAddFloatToArray(aP,dex,va_arg(ptr,float));
            break;
        case 'v':
            aP = AxWriteArray(aP,dex,va_arg(ptr,data_ptr));
            break;          
        default:
            CHECK(FALSE,ERR_ILLVAL,XLT("Illegal format passed to AxBuildArray"));
        }
    }
    return(aP);
}

/*************************************************************************
function: AxDecomposeArray - Read ints/strings/floats from passed array
*/
DLL_EXPORT void AxDecomposeArray (va_alist)
va_dcl /* where to write results to */
{
    va_list ptr;
    register char *format;   /* format string describing decomposition mode */
    register char *p;
    register int dex = 0;
    data_ptr aP;
    mbchar *charP;
    mbchar **strP;
    int *intP;
    int len;
    double *doubleP;
    float *floatP;
    data_ptr *dataP;

    va_start(ptr);

    aP = va_arg(ptr,data_ptr);
    format = va_arg(ptr,char *);

/* copy in the data */

    for ( p = format; *p; p++, dex++ ) {
        switch ( *p ) {
        case 's':           /* addr of char ptr */
            strP = (mbchar **)va_arg(ptr,char**);
            *strP = AxStrFromArray(aP,dex);
            break;
        case 'c':           /* addr of char buffer, next arg is buffer size */
            charP = (mbchar *)va_arg(ptr,char*);
            len = va_arg(ptr,int);
            MBstrncpy(charP,AxStrFromArray(aP,dex),len);
            charP[len-1] = 0;
            break;
        case 'i':           /* addr of int */
            intP = va_arg(ptr,int*);
            *intP = AxIntFromArray(aP,dex);
            break;
        case 'b':           /* addr of int */
            intP = va_arg(ptr,int*);
            *intP = AxIntFromArray(aP,dex) ? 1 : 0;
            break;
        case 'd':           /* addr of double */
            doubleP = va_arg(ptr,double*);
            *doubleP = AxFloatFromArray(aP,dex);
            break;
        case 'f':           /* addr of float */
            floatP = va_arg(ptr,float*);
            *floatP = AxFloatFromArray(aP,dex);
            break;
        case 'v':           /* addr of double */
            dataP = va_arg(ptr,data_ptr*);
            *dataP = AxCopyData(AxArrayElement(aP,dex));
            break;
        default:
            CHECK(FALSE,ERR_ILLVAL,
                            XLT("Illegal format passed to AxDecomposeArray"));
        }
    }
}


/*************************************************************************
function: AxCopyData - Create a copy of a datum against THIMpid
*/
DLL_EXPORT data_ptr 
AxCopyData (p)
register data_ptr p;
{
    return(AxTaskCopyData(THIMpid,p));
}

/*************************************************************************
function: AxTaskCopyData - Create a copy of a datum - any task
*/
DLL_EXPORT data_ptr 
AxTaskCopyData (pid,p)
register int pid;       /* Owner of memory - THIMpid style number */
register data_ptr p;
{
    register data_ptr *px, *pends, *pdst;
    register data_ptr new;
    register data_binary *newbin, *oldbin;

    if ( IS_INTEGER(p) || !p )
        return(p);
    p = (data_ptr)((long)p & ~FAULT_BIT);
    if ( p->type == TYPE_NUMBER ) {
        new = (data_ptr)AxTaskPoolAllocate(pid,sizeof(data_number));
        *NX(new) = *NX(p);
        return(new);
    }
    else if ( p->type == TYPE_STRING )
        return(AxTaskMakeStrData(pid,-1,SX(p)->text));
    else if ( p->type == TYPE_ARRAY ) {
        new = AxTaskMakeArray(pid,AX(p)->size);
        pends = (px = AX(p)->vals) + AX(p)->size;
        pdst = AX(new)->vals;
        while ( px < pends ) {  /* fastest possible array copy? */
            p = *px++;
            if ( IS_INTEGER(p) || !XX(p) )
                *pdst++ = p;
            else
                *pdst++ = AxTaskCopyData(pid,p);
        }
        return(new);
    }
    else if ( p->type == TYPE_BINARY ) {
        oldbin = (data_binary*)p;
        newbin = (data_binary*)AxMakeBinaryData((int)oldbin->size, oldbin->bytes);      /*  type checking - lhunt */
        return((data_ptr)newbin);
    }
#if !ELF_COMM_STAND_ALONE
    else if ( p->type == TYPE_OBJECT ) {
        if (OX(p)->objHostname)
            return(ELF_TASK_CREATE_REMOTE_OBJECT(pid,OX(p)->objDex,OX(p)->objUid,OX(p)->objHostname));
        else
            return(ELF_TASK_CREATE_OBJECT(pid,OX(p)->objDex,OX(p)->objUid,NULL));
    }
#endif
    else if ( p->type == TYPE_ENTRY ) {
        CHECK(p->type != TYPE_ENTRY,ERR_ILLVAL,
                           XLT("Attempted to AxFreeData an entry block"));
    }
    else
        CHECK(FALSE,ERR_ILLVAL,XLT("AxTaskCopyData of non-Elf data"));
}


/** The following routine is a bit of a hack.  Normally the caller to
**  functions such as "AxStrPtrFromDataPtr()" should not expect that the
**  character pointer which is returned will be 'stable' over time.  It is
**  the caller's responsibility to copy the data before making any other
**  calls to AxStrPtrFromDataPtr() or its relatives.  HOWEVER!, many places
**  in the code 'know' that the returned pointer is to a stable elf data
**  structure and therefore assume that the returned pointer remains valid.
**  In the wide character case we need to expand strings to their wide form,
**  so that even if the underlying ELF object is stable, the widened version
**  of the same string would not be.  Sooo....
**  Here we keep a number of static buffers, and use them in a round-robin
**  fashion, so that programmers who have made assumptions get away with it
**  The buffers are malloc'ed to a high water mark length, as they are used.
**/

#define NUM_STATIC_BUFFS 50
static int next_static_buf = -1;
static int static_buf_len[NUM_STATIC_BUFFS];
static mbchar *static_buf_ptrs[NUM_STATIC_BUFFS];

mbchar *get_static_buffer( len )
int len;        /* length in bytes of desired buffer */
{

    register int i;
    register mbchar *ret_ptr;

    if ( next_static_buf == -1 )
       {
       /** initialize */
       next_static_buf = 0;
       for ( i=0; i<NUM_STATIC_BUFFS; i++ )
           {
           static_buf_len[i] = -1;
           static_buf_ptrs[i] = (mbchar *)NULL;
           }
       }

    if ( len == 0 )
       len = 1;    /* do not try to malloc zero bytes */

    if ( static_buf_len[next_static_buf] < len )
       {
       /** need to allocate some more space */
       if( static_buf_ptrs[next_static_buf] != (mbchar *)NULL )
#if NOLIB_THIMBLE
          {
          /* built without the standard Aster*x memory handlers */
          AxFree( (char *)static_buf_ptrs[next_static_buf] );
          }
       static_buf_ptrs[next_static_buf] = (mbchar *)AxMalloc( len );
#else
          {
          TrueFree( (char *)static_buf_ptrs[next_static_buf] );
          }
       static_buf_ptrs[next_static_buf] = (mbchar *)TrueMalloc( len );
#endif
       static_buf_len[next_static_buf] = len;
       }

    ret_ptr = static_buf_ptrs[next_static_buf];
    next_static_buf = (next_static_buf+1) % NUM_STATIC_BUFFS;

    return( ret_ptr );
}


typedef struct tempStr_t {
    struct tempStr_t *next;
    int8 str[24];
} tempStr_t;
static tempStr_t *axTempStrs;

/*************************************************************************
function: AxStrFromDataPtr - Get a string from a data ptr - do conversions
*/
DLL_EXPORT mbchar *AxStrFromDataPtr (dP)
register data_ptr dP;
{
    tempStr_t *tempStr;

    if ( !dP )
        return((mbchar *)"");
    if ( IS_INTEGER(dP) ) {
#if ELF_COMM_STAND_ALONE
        tempStr = (tempStr_t*)AxMalloc(sizeof(tempStr_t));
        tempStr->next = axTempStrs;
        axTempStrs = tempStr;
#else
        tempStr = (tempStr_t*)get_static_buffer(sizeof(tempStr_t));
#endif
        sprintf(tempStr->str,"%d",GET_INTEGER(dP));
        return( (mbchar *)(tempStr->str) );
    }
    dP = (data_ptr)((long)dP & ~FAULT_BIT);/* allow pops of read-only values */
    if ( XX(dP)->type == TYPE_STRING )
        return((mbchar *)(SX(dP)->text));
    if ( XX(dP)->type == TYPE_NUMBER ) {
        double val;
        int valint;
#if ELF_COMM_STAND_ALONE
        tempStr = (tempStr_t*)AxMalloc(sizeof(tempStr_t));
        tempStr->next = axTempStrs;
        axTempStrs = tempStr;
#else
        tempStr = (tempStr_t*)get_static_buffer(sizeof(tempStr_t));
#endif
        val = NX(dP)->value;
        valint = val;
        if ( valint == val )
            sprintf(tempStr->str,"%d",valint);
        else {
            sprintf(tempStr->str,"%.15g",val);
            ConvertUSDotsToEuropeComma(tempStr->str);
        }
        return((mbchar *)(tempStr->str));
    }

    /* Fell through, not a string or convertable type */
    CHECK(FALSE,ERR_NOTSTRING,XLT("AxStrFromDataPtr cannot convert value"));
    return( (mbchar *)NULL );
}

static data_ptr *ListWalkerArr;
static int NumListWalkerElements;
static int ListWalkerIdx;
static void AxCreateListFromElfData();
/*
*   AxListWalkerInit -  initalize the list walker array
*       decomposes an arbitary shaped  array into a list of data of pointers
*   returns the total number of elements in the list
*/
DLL_EXPORT int AxListWalkerInit(aP)
    data_ptr aP;
{
    if(ListWalkerArr != NULL)   
        TaskFree(0, ListWalkerArr);

    /* init all vars */
    ListWalkerArr = NULL;
    ListWalkerIdx = 0;
    NumListWalkerElements = 0;

    /* create the list */
    AxCreateListFromElfData(aP, &ListWalkerArr, &NumListWalkerElements);    

    /* return the total number of elements */
    return(NumListWalkerElements);
}

/*
*   AxCreateListFromElfData - create a list of data pointers given
*       a arbitary shaped array
*/
static void AxCreateListFromElfData(aP, listP, num)
    data_ptr aP, **listP;
    int *num;
{
    int ix;
    
    if(NULL == aP)
        return;
    if(AxIsArray(aP))   {
        for(ix=0; ix<AxArraySize(aP); ix++)
            AxCreateListFromElfData(AxArrayElement(aP, ix), listP, num);
        return;
    }
    if( *listP )    {
        *listP = (data_ptr *)TaskRealloc( 0, 
                        *listP, ((*num)+1) *sizeof(data_ptr));
    }
    else
        *listP = (data_ptr *)TaskAlloc( 0, ((*num)+1) *sizeof(data_ptr));
    (*listP)[*num] = aP;
    (*num)++;   
}


/*
*   AxListWalkerNext - returns the next argument from the list created
*       using AxListWalkerInit
*/
DLL_EXPORT data_ptr AxListWalkerNext()  
{
    if(ListWalkerIdx >= NumListWalkerElements)
        return(NULL);
    return(ListWalkerArr[ListWalkerIdx++]);
}




#if ELF_COMM_STAND_ALONE
/*************************************************************************
function: AxFreeTempData - Release data malloced by StrFrom calls
*/
void AxFreeTempData ()
{
    tempStr_t *p, *next;

    for ( p = axTempStrs; p; p = next ) {
        next = p->next;
        AxFree((char*)p);
    }
    axTempStrs = NULL;
}
#endif

#define BLK 60		  /* used for reading/writing binary from elf data files */
#define BL3 20		  /* used for reading/writing binary from elf data files */
#define BLT 80		  /* used for reading/writing binary from elf data files */

/*************************************************************************
function: hex6_to_bin - Convert hex6 string back to BLK bytes
*/
static hex6_to_bin(char *result, char *inbyte)
{
    register int    ix, jx;
    register unsigned int ch;

    for (ix = 0; ix < BLK; ix++)
        result[ix] = (inbyte[ix] - ' ') & 0x3F;
    for (jx = 0; jx < BLK; ix++) {
        ch = (inbyte[ix] - ' ') & 0xFF;
        result[jx++] |= (ch << 2) & 0xC0;
        result[jx++] |= (ch << 4) & 0xC0;
        result[jx++] |= (ch << 6) & 0xC0;
    }
}

extern data_ptr AxReadElfDatumFromBuf (char *buf);

/*************************************************************************
function: AxReadElfFile - Read file, return datum
Returns null datum if file could not be read.  Note that this
would also return null datum if file just contains a null datum.
This must NOT be called within axmain, would unlock file!
*/
data_ptr AxReadElfFile (char *filename) {
    struct stat  status;
    int          code,fd = -1;
    char *fileBuf = NULL;
	data_ptr aP = NULL;

#if AXMS
    fd = open(filename, (_O_RDONLY | _O_BINARY), _S_IREAD);
#else
    fd = open(filename, 0, 0);
#endif
    if ( fd == -1 )
        goto done;
    code = fstat(fd, &status);
    if (code == -1)
        goto done;
    fileBuf = (char*)AxMalloc(status.st_size+1);
    if ( !fileBuf )
        goto done;
    code = read(fd, fileBuf, status.st_size);
    if (code != status.st_size)
        goto done;
    fileBuf[status.st_size] = 0;    /* null term for easy searching */
	aP = AxReadElfDatumFromBuf(fileBuf);
done:
    if ( fd != -1 )
        close(fd);
	if ( fileBuf )
		AxFree(fileBuf);
    return(aP);
}

static data_ptr readOneFromBuf(char **bufP);

/*************************************************************************
function: AxReadElfDatumFromBuf - Read an elf datum from passed 'file' buffer
The buffer MUST be newline followed by NULL terminated.
Leaks memory if the file is truncated on array/binary datums.
*/
data_ptr AxReadElfDatumFromBuf (char *buf)
{
	char *bufP = buf;
	return(readOneFromBuf(&bufP));
}
static data_ptr readOneFromBuf(char **bufP) { /* IN/OUT address of buffer */
	char *buf;
	data_ptr aP;
	int code;
	char *newlineP;
    register int    ix, cnt;

again:
	buf = *bufP;
	if ( !buf[0] )
		return(NULL);
	newlineP = strchr(buf,'\n');	/* null term this line */
	if ( !newlineP )
		return(NULL);
	*newlineP = 0;
	*bufP = newlineP + 1;			/* setup for next line */
	code = *buf++;

	switch ( code ) {
	case ':':	/* comment, skip to end of line */
	case '*':	/* comment, skip to end of line */
		goto again;
	case '#':
		aP = MK_INTEGER(atoi(buf));
		break;
    case '@':
        cnt = atoi(buf);
        aP = AxMakeArray(cnt);
        for (ix = 0; ix < cnt; ix++)
            AxWriteArray(aP, ix, readOneFromBuf(bufP));
        break;
	case '"':	   /* older style left in for compatibility */
		aP = AxMakeStrData(-1,buf);
		break;
	case '\'': {
		register char *src = buf;
		register char *dst = buf;
		while ( *src ) {
			if ( *src == '\\' ) {
				src++;
				if ( *src == 'n' ) {
					*dst++ = '\n';
					src++;
				}
				else
					*dst++ = *src++;
			}
			else
				*dst++ = *src++;
		}
		*dst++ = 0;
		aP = AxMakeStrData(-1,buf);
		break;
	}
	case '.':
		aP = AxMakeFloatData(atof(buf));
		break;
	case '%':
	{
		int size;
		char blkbuf[BLK];
		char *stuff;

		size = atoi(buf);
		aP = AxMakeBinaryData(size,NULL);
		AxBinaryFromDataPtr(aP, &stuff);
		while (size > 0) {
			buf = *bufP;
			if ( !buf[0] )
				return(NULL);
			newlineP = strchr(buf,'\n');	/* null term this line */
			if ( !newlineP )
				return(NULL);
			*newlineP = 0;
			*bufP = newlineP + 1;			/* setup for next line */
			if ( strlen(buf) != BLT )
				return(NULL);
			hex6_to_bin(blkbuf, buf);
			memcpy(stuff, blkbuf, MIN(size, BLK));
			stuff += BLK;
			size -= BLK;
		}
		break;
	}
	case '-':
	default:
		aP = NULL;
		break;	  /* null datum */
	}		   /* treat anything else as null datum */
	return(aP);
}

/*************************************************************************
function: Astrtod - Applix strtod allows commas as well as dots
*/
double Astrtod (str,pp)
mbchar *str;
mbchar **pp;
{
    extern char *strchr();
    char *p;
    char buf[STRLONG];
    double v;
    extern double strtod();

    if ( !strchr(str,',') )     /* quick & easy test & convert */
        return(strtod(str,pp));
    else {                          /* oh, well, slow */
        if ( strlen(str) >= STRLONG ) {
            *pp = str;
            return(0.0);
        }
        MBstrcpy(buf,str);
        for ( p = buf; *p; p++ ) {
            if ( *p == ',' )
                *p = '.';
        }
        v = strtod(buf,pp);
        *pp = str + (*pp - buf);
        return(v);
    }
}

#if !ELF_COMM_STAND_ALONE
ConvertUSDotsToEuropeComma(s)
mbchar *s;
{
    static int called;
    static int convert;

    if ( !called )
        convert = UPnumVal(PROF_NUMBERS_);
    if ( convert ) {
        while ( *s ) {
            if ( *s == '.' )
                *s = ',';
            s++;
        }
    }
}
#endif


typedef struct      {
    data_ptr arrayP;
    data_ptr sheetP;
    data_ptr rowP;
    data_ptr colP;
    int numSheets;
    int numRows;
    int numCols;
    int freeCount;  /* num wrappers to free */
}AxElfSSArgDesc;

/*
*   AxInitCube - initialize the argument desc
*/
DLL_EXPORT void *AxInitCube(aP, numSheets, numRows, numCols)
    data_ptr aP;
    int *numSheets, *numRows, *numCols;
{
    int numDimensions;
    AxElfSSArgDesc *aArgP;  
    data_ptr sheetP, rowP, colP;
    int freeCount;

    /* get the dimensions of the passed array */
    numDimensions = axGetArrayDim(aP);
    if(numDimensions > 3)
        return(NULL);   /* can't handle more than three dimensions */

    /* allocate the book keeping structure */
    aArgP = (AxElfSSArgDesc *) AxMalloc(sizeof(AxElfSSArgDesc));
    memset(aArgP, '\0', sizeof(AxElfSSArgDesc));

    /* put wrappers around the passed array so that the final array is */
    /* a 3-d array and keep track of how many layers of these arrays */
    /* we need to free later on. freeCount keep tracks of that */
    switch(numDimensions)   {
        case 0:
            freeCount = 3;
            colP = AxMakeArray(1);
            AxWriteArray(colP, 0, aP);
            rowP = AxMakeArray(1);
            AxAddArrayToArray(rowP, 0, colP);
            sheetP = AxMakeArray(1);
            AxAddArrayToArray(sheetP, 0, rowP);
            break;
        case 1:
            freeCount = 2;
            rowP = AxMakeArray(1);
            AxAddArrayToArray(rowP, 0, aP);
            sheetP = AxMakeArray(1);
            AxAddArrayToArray(sheetP, 0, rowP);
            colP = aP;
            break;
        case 2:
            freeCount = 1;
            sheetP = AxMakeArray(1);
            AxAddArrayToArray(sheetP, 0, aP);
            rowP = aP;
            colP = AxArrayElement(rowP, 0);
            break;
        case 3:
            freeCount = 0;
            sheetP = aP;
            rowP = AxArrayElement(sheetP, 0);
            colP = AxArrayElement(rowP, 0);
            break;
    }

    /* set up the handle to be passed back */
    aArgP->arrayP = aP;
    aArgP->numSheets = *numSheets = AxArraySize(sheetP);            
    aArgP->numRows = *numRows = AxArraySize(rowP);          
    aArgP->numCols = *numCols = AxArraySize(colP);          
    aArgP->sheetP = sheetP;
    aArgP->rowP = rowP;
    aArgP->colP = colP;
    aArgP->freeCount = freeCount;
    return(aArgP);
}


/*
*   AxFreeCube - free the cube data structure and wrapper arrays
*/
DLL_EXPORT void AxFreeCube(in_aArgP)
    void *in_aArgP;  
{
    AxElfSSArgDesc *aArgP=(AxElfSSArgDesc *)in_aArgP;  
    if(0 == aArgP->freeCount)   
        goto free_it;   
    AxFreeArrayButNotContents(aArgP->sheetP);
    if(1 == aArgP->freeCount)       
        goto free_it;
    AxFreeArrayButNotContents(aArgP->rowP);
    if(2 == aArgP->freeCount)
        goto free_it;
    AxFreeArrayButNotContents(aArgP->colP);
free_it:
    AxFree((char*)aArgP);
}

/*
*   AxCubeElement - return the element at the given indices
*/
DLL_EXPORT data_ptr AxCubeElement(in_aArgP, sheetDex, rowDex, colDex)
    void *in_aArgP;
    int sheetDex, rowDex, colDex;
{
    data_ptr rowP, colP, itemP;
    AxElfSSArgDesc *aArgP=(AxElfSSArgDesc *)in_aArgP;  

    if(!aArgP)
        return(NULL); 
    if(sheetDex >= aArgP->numSheets)
        return(NULL);
    rowP = ((data_array*)(aArgP->sheetP))->vals[sheetDex];
    if(!AxIsArray(rowP))    {  
        if((rowDex == 0) && (colDex == 0))
            return(rowP);
        return(NULL);
    }
    if(rowDex >= aArgP->numRows)
        return(NULL);
    colP = ((data_array*)rowP)->vals[rowDex];
    if(!AxIsArray(colP))    {  
        if(colDex == 0)
            return(colP);
        return(NULL);
    }
    if(colDex >= aArgP->numCols)
        return(NULL);
    itemP = ((data_array*)colP)->vals[colDex];
    return(itemP);
}

/*
*   axGetArrayDim - returns the dimensions
*/
axGetArrayDim(aP)
    data_ptr aP;
{
    return(axDoGetArrayDim(aP, 0));
}

axDoGetArrayDim(aP, level)
    data_ptr aP;
    int level;
{
    data_ptr nextP;
    if(!AxIsArray(aP))
        return(level);
    level++;    
    nextP = AxArrayElement(aP, 0);
    level = axDoGetArrayDim(nextP, level); 
    return(level);
}

/*============================================================================*/
/*============================================================================*/
/* Special API to SS from addins */
/*============================================================================*/
static char *AxSSNaStr = "NA";
static char *AxSSErrorStr = "ERROR";

DLL_EXPORT data_ptr AxSSMakeNaData()
{
    data_ptr dP;
    dP = AxMakeBinaryData(strlen(AxSSNaStr)+1, AxSSNaStr);
    return(dP);
}

DLL_EXPORT data_ptr AxSSMakeErrorData()
{
    data_ptr dP;
    dP = AxMakeBinaryData(strlen(AxSSErrorStr)+1, AxSSErrorStr);
    return(dP);
}

DLL_EXPORT int AxSSIsNa(dP)
    data_ptr dP;
{
    char *p;
    if(!AxIsBinary(dP))
        return(FALSE);
    p = AxBinaryBytesPtr(dP);
    if(0 == strcmp(p, AxSSNaStr))
        return(TRUE);
    return(FALSE);
}

DLL_EXPORT int AxSSIsError(dP)
    data_ptr dP;
{
    char *p;
    if(!AxIsBinary(dP))
        return(FALSE);
    p = AxBinaryBytesPtr(dP);
    if(0 == strcmp(p, AxSSErrorStr))
        return(TRUE);
    return(FALSE);
}

/*============================================================================*/



#if BUILD_WIDE      /*  {  */
static wchar *elfMbStrToUnicodeStr (mbchar *bufP, int *ret_len) {
    int len = MBstrlen(bufP);
    wchar *ret_str = (wchar *)get_static_buffer((len+1)*sizeof(wchar));
    *ret_len = AxMbStrToUnicodeStr(bufP,len,ret_str);
    return ret_str;
}

static mbchar *elfUnicodeStrToMbStr (char *bufP, int *wlen) {
	wchar *wstr = (wchar *)get_static_buffer(*wlen);
	int len = 2*(*wlen);
	mbchar *mbstr = (mbchar *)get_static_buffer(len+1);
	if (0xCAFEBABE != ntohl(0xCAFEBABE)) {
		/* network and host byte order differ, so swap the bytes */
		int ix = 0;
		for (ix = 0; ix < len; ix += 2) {
			mbchar t = bufP[ix];
			bufP[ix] = bufP[ix+1];
			bufP[ix+1] = t;
		}
	}
	*wlen = AxUnicodeStrToMbStr((wchar*)bufP,*wlen,mbstr);
	return mbstr;
}
#endif            /*  }  BUILD_WIDE */

#if !AX_NT
/*************************************************************************
function: AxDumpForcerNeverCalled - stub so we always get eldump.o!
This exists so that any program that uses elf data always gets eldump.o.
This way, when run from a debugger, you never get "AxPrintfDatum" not defined.
*/
void AxDumpForcerNeverCalled ()
{
	AxPrintfDatum(NULL);
	AxFileDatum(NULL,NULL);
	AxTempFileDatum(NULL);
}
#endif

