/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: eldump.c
   Applix ASC version: 4
   Date: Tue Feb  2 15:21:49 1999
*/
#include <stdio.h>
#include "elfapi.h"

extern void AxDumpDatum();
extern void AxPrintfDatum();
static char *tabs();
static void dumper();

/*************************************************************************
function: AxPrintfDatum - Put datum to stdout
*/
void AxPrintfDatum (datum)
elfData datum;
{
	AxDumpDatum(datum,stdout);
}

/*************************************************************************
function: AxDumpDatum - Log contents of a datum to a file in a "pretty" manner
*/
void AxDumpDatum (datum,fp)
elfData datum;
FILE *fp;
{
	dumper(datum,fp,0);
	fflush(fp);
}

/*************************************************************************
function: AxFileDatum - Log contents to passed filename
*/
void AxFileDatum (char *filename, elfData datum)
{
	FILE *fp;

	fp = fopen(filename,"w");
	if ( !fp ) {
		fprintf(stderr,"Could not open dumpfile %s\n",filename);
		return;
	}
	AxDumpDatum(datum,fp);
	fclose(fp);
}

/*************************************************************************
function: AxTempFileDatum - Create a tempfile, log contents, return filename
*/
char *AxTempFileDatum (elfData datum)
{
	char *filename;

	filename = tmpnam(NULL);
	AxFileDatum(filename,datum);
	return(filename);
}

/*************************************************************************
function: dumper - Dump datum
*/
static void dumper (datum, fp, arrayNesting)
elfData datum;
FILE *fp;
int arrayNesting;
{
	int ix;

	if (!fp)
		fp = stdout;

	if ( !datum ) {
		fprintf(fp, "%sNULL\n",tabs(arrayNesting));
	}
	else if ( AxIsNumber(datum) ) {
		fprintf(fp, "%snumber: %.15g\n",tabs(arrayNesting),
				AxFloatFromDataPtr(datum));
	}
	else if ( AxIsString(datum) ) {
		fprintf(fp, "%sstring: %s\n",tabs(arrayNesting),
				AxStrFromDataPtr(datum));
	}
	else if ( AxIsArray(datum) ) {
		tabs(arrayNesting);
		fprintf(fp, "%sarray of %d\n",tabs(arrayNesting),AxArraySize(datum));
		for ( ix = 0; ix < AxArraySize(datum); ix++ )
			dumper(AxArrayElement(datum,ix),fp,arrayNesting+1);
	}
	else if ( AxIsBinary(datum) ) {
		fprintf(fp, "%sbinary: %d bytes\n",tabs(arrayNesting),
				AxBinarySize(datum));
	}
	else {
		fprintf(fp, "%sUNKNOWN DATA TYPE\n",tabs(arrayNesting));
	}
}

/*************************************************************************
function: tabs - Mindless routine to get N tabs
*/
static char *tabs(n)
int n;
{
	static char buf[101];
	register int ix;

	for ( ix = 0; (ix < 100) && (ix < n); ix++ )
		buf[ix] = '\t';
	buf[ix] = 0;
	return(buf);
}


