/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: elf.h
   Applix ASC version: 101
   Date: Sat Feb 20 21:13:13 1999
*/
#ifndef ELF_LOADED

#ifdef data_ptr /* happens if elfapi.h included before elf.h */
#undef data_ptr
#endif

#if DEBUG == 0
/* Using the added layer is a big performance loose! */
#define ElfMalloc malloc
#define ElfFree free
#define ElfRealloc realloc
#endif

#ifndef FOR_TESTER
#define FOR_TESTER 0
#endif
#ifndef FOR_EDITOR
#define FOR_EDITOR 0
#endif

#define MAX_DIMS 10      /* max array dimensions */
#define MAX_GVARS 2000  /* size of TaskFrame */
#define MAX_QPUSH 100   /* biggest quick push literal number OP_PUSH_NUM.
                        range is -1 to 100.  This must be small becuse the
                        opcode decode in elf_intr does a sign extend */
#define MAX_FORS 8 /* max for-loop nesting */
#define bint int      /* Elf Basic integer... was short in Alis days */

#ifndef MEMSET
#define MEMSET(a,b,c) memset((char*)(a),(int)(b),(int)(c))
#endif

#ifndef MEMCPY
#define MEMCPY(a,b,c) memcpy((char*)(a),(char*)(b),(int)(c))
#endif

#define ELFIDSTR 60 /* func/var names must be 30<ELFIDSTR<STRLONG */

#define MAX_STRLEN (unsigned)4000
#define ELF_HASH_PRIME 1009 /* Hash basis (must be a prime number) */

#define CASE break; case
#define DEFAULT break; default

#define BTRUE -1        /* Basic TRUE */
#define BFALSE 0        /* Basic FALSE */
#define BMAXINT 0x1FFFFFFF      /* 28 bits plus 1 sign bit */

/* Type of datums (matches table order in elcode.c, !!0 reorder) */

#define TYPE_UNKNOWN 0   /* type not yet determined */
#define TYPE_NUMBER 1    /* (comp/intr) numeric */
#define TYPE_STRING 2    /* (comp/intr) text */
#define TYPE_A_INTEGER 3 /* OBSOLETE */
#define TYPE_U_FUNC 4    /* (intr) shared libray based function */
#define TYPE_C_FUNC 5    /* (intr) builtin function */
#define TYPE_C_MACRO 6   /* recordable C call */
#define TYPE_R_FUNC 7    /* (intr) rpc based function */
#define TYPE_ARRAY 8     /* (intr) array */
#define TYPE_LOCVAR 9    /* (comp) variable data type */
#define TYPE_ARGVAR 10    /* (comp) argument data type */
#define TYPE_GLOVAR 11   /* (comp) global var */
#define TYPE_EXTVAR 12   /* (comp) extern varibale */
#define TYPE_QFUNC 13    /* (comp) non-recorded function data type */
#define TYPE_MACRO 14    /* (comp) recordable function */
#define TYPE_UIFUNC 15   /* (comp) non-recordable top-level function */
#define TYPE_EXTFUNC 16  /* (comp) remote function data type */
#define TYPE_NRFUNC 17   /* (comp) functions invokable by command key */
#define TYPE_LABEL 18    /* (comp) label data type */
#define TYPE_INTEGER 19  /* (comp) label data type */
#define TYPE_ENTRY 20    /* (intr) entry variable block */
#define TYPE_OBJECT 21   /* (intr/comp) object */
#define TYPE_BINARY 22  /* (intr) binary data */
#define TYPE_SETOBJ 23  /* write method */
#define TYPE_GETOBJ 24  /* read method */
#define TYPE_OBJVAR 25   /* (comp) object local data var */
#define TYPE_SYSVAR 26   /* (comp) sysvar */

#define ELF_OBJ "elfx"  /* name of object file segment */
#define ELF_ONEDOTOBJ "1.elfx"  /* name of object file segment */

#ifdef lint
#define register        /* bogus complaints about register ptr unions */
#endif
#ifdef FOR_DBX
#define register    /* clean access to all vars during debug */
#endif

#include <setjmp.h> /* needed for abortBuf in elf_info */

/************** GENERAL TYPEDEFS **************************/

typedef unsigned short shrt;

/************** DATA OBJECT STRUCTURES *************************/

#define XX(x) ((data_any*)(x))
#define NX(x) ((data_number*)(x))
#define SX(x) ((data_string*)(x))
#define AX(x) ((data_array*)(x))
#define EX(x) ((data_entry*)(x))
#define OX(x) ((data_object*)(x))
#define FX(x) ((fkey_info*)(x))
#define BX(x) ((data_binary*)(x))
#define LX(x) ((long)(x))

/* Macros to manipulate data_ptr embedded integers */

#define MK_INTEGER(x) ((data_ptr)(BITS_INTEGER | ((x) << SHFT_INTEGER)))
#define IS_INTEGER(dp) ((((long)(dp)) & BITS_INTEGER) == BITS_INTEGER)
#define GET_INTEGER(dp) (((long)(dp)) >> SHFT_INTEGER)
#define FAULT_BIT 0x0001
#define BITS_INTEGER 0x0003 /* fault bit plus next bit */
#define INTEGER_BIT 0x0002  /* integer BITS minus the FAULT bit */
#define SHFT_INTEGER 2
#define C_INT_IS_ELF_INT(n) ( ((n) >= -536870912) && ((n) <= 536870911) )
#define C_DBL_IS_ELF_INT(d,n) ( C_INT_IS_ELF_INT(d) && ((n = d) == d) )

typedef struct data_any *data_ptr;

typedef struct data_any {             /* just a template */
    shrt type;              /* data type */
    double foo;
} data_any;

typedef struct data_number {          /* numbers */
    shrt type;              /* data type */
    double value;

} data_number;

typedef struct data_string {          /* strings */
    shrt type;              /* data type */
    char text[1];       /* varying extent, null terminatd */
} data_string;

typedef struct data_array {           /* arrays */
    shrt type;              /* data type */
    int size;          /* current number of elements */
    data_ptr vals[1];   /* varying, NULLs or ptrs to elements */
} data_array;

typedef struct data_binary {           /* bytes */
    shrt type;              /* data type */
    int size;          /* current number of elements */
    char bytes[1];   /* varying */
} data_binary;

typedef struct data_entry {         /* entry var */
    shrt type;              /* data type */
    struct {    /* MUST BE JUST LIKE ext_link STRUCTURE */
        shrt *addr;        /* address of gsym_info/OP_ENTER/bltin_desc/fkey_info */
        struct package_info *packP; /* package containig addr */
        int hashval;    /* hash value of name */
        char *nameP;
        char *moduleName;
    } linkage;
} data_entry;

typedef struct data_object {          /* object reference */
    shrt type;              /* data type */
    int objDex;     /* index in object table for object access */
    int objUid;     /* object unique id (detects refs to destroyed objects) */
    char *objHostname;      /* null for local objects or to get to remote server */
    data_ptr dynBinder; /* array of parentage used if objUid is zero */
    struct data_object *nextUnsnapped;  /* NULL unless objUid is zero */
    struct data_object *prevUnsnapped;  /* NULL unless objUid is zero */
    char *unsnappedAppName; /* for best performance in DecObjPtr */
} data_object;

typedef struct elf_obj_grp {
    struct elf_obj_grp *next;       /* NULL if last */
    int hashval;                    /* hash of name */
    int static_info;                /* if true save to disk */
    data_ptr value;
    char name[1];       /* varying extent */
} elf_obj_grp;

typedef struct elf_objvar_item {    /* Objvar hash table element */
    struct elf_obj_grp *obj_grp;
    struct elf_objvar_item *next;
} elf_objvar_item;

typedef struct cpp_method_info_t {  /* parallel defs in "elapi.hpp" */
    char *methodName;
    data_ptr (*memberFunc)();
    char *helpStr;
    int ftype;
    int hashval;
} cpp_method_info_t;

typedef struct cpp_class_info_t {   /* info about C++ object */
    char *className;            /* name of the class */
    void *modelObj;             /* model object */
    char *libName;              /* shared library pathname */
    cpp_method_info_t *methods; /* null terminated member func descriptions */
} cpp_class_info_t;

typedef struct cpp_obj_info_t {
    void *objP;                 /* object instance */
    data_ptr classP;            /* link to class info */
} cpp_obj_info_t;

typedef struct elf_object {
    int uid;                    /* object creation counter */
    int dex;                    /* pos in ObjectTable */
	unsigned int isTopLevelApp : 3, /* 0-4 If True, is THE top level object of app */
			javaDlgMapped : 1,		/* TRUE if in javaDlgMap table */
			hasBeenRenamed : 1;	/* if true, needs slow destroy code */
    int task_uid;               /* task id of this application */
    char *hostname;             /* non-null if its created by a remote client */
    char *appName;              /* Application this object is a member of */
    struct elf_object *inheritFromP;    /* inheritance parent (not obj->parent) */
    struct elf_object *parentP; /* containing object */
    data_ptr *children;         /* contained objects */
    int numChildren;            /* count of contained objects */
    elf_obj_grp *localData;     /* state of local objvars */
    data_ptr *models;           /* User defined models used in this application */
    int numModels;              /* number of models loaded */
    data_ptr *libobjs;          /* refs to library objects */
    int numLibobjs;             /* number of library objects */
    struct package_info *packP; /* methods associated with this object */
    unsigned short *hashtable;          /* optional table for object searches */
    struct gfunc_info **hashpool;       /* optional table for object searches */
    int hashpoolcount;                  /* # shrts in hashpool */
    cpp_class_info_t *cpp_class_info;   /* C++ class description */
    cpp_obj_info_t *cpp_obj_info;       /* C++ object info */
    void* ext_obj;              /* Stores the ext_obj identifier */
    char nameTag[ELFIDSTR+1];
} elf_object;

typedef struct bltin_desc {        /* built-in function */
    shrt type;
    int symid;          /* symtab id */
    data_ptr (*func)();
} bltin_desc;

typedef struct data_user_func {        /* shared library based func */
    shrt type;
    char *category;
    char *alias;
    char *nameInCcode;
    char *argInfo;
    int callMode;   /* callModes defined in elapi.h */
    char *libname;
    data_ptr (*func)();
} data_user_func;


#ifndef CONST
#define CONST 
#endif

typedef struct gvar_info {        /* global var info */
    int value;      /* offset in task frame */
    int hashval;    /* hash value of name */
    CONST char *name;       /* reference into hash table */
    CONST char *formatP;    /* NULL or name of format */
} gvar_info;

typedef struct gfunc_info {      /* global function info */
    shrt *value;    /* ptr to function entry block */
    int srcLine;    /* source code line number of func */
    int hashval;    /* hash value of name */
    CONST char *name;       /* reference into hash table */
} gfunc_info;

typedef struct ext_link {        /* extern link entry in code frame */
    union { /* address of gsym_info/OP_ENTER/bltin_desc/fkey_info */
      int dex;
      shrt *ptr;
    } value;
    struct package_info *packP; /* package containig addr */
    int hashval;    /* hash value of name */
    char *nameP;
    char *moduleName;   /* NULL or name of module to snap to */
} ext_link;         /* See ref in data_entry struct before changing this! */

typedef struct ext_link_for_elccc { /* SAME MEMORY AS ext_link!!!! */
    shrt *ptr;
    struct package_info *packP;
    int hashval;
    CONST char *nameP;
    CONST char *moduleName;
} ext_link_for_elccc;


/************** INTERPRETER TYPEDEFS **************************/

typedef struct stack_frame {    /* function description */
    struct stack_frame *caller; /* who called me */
    shrt *return_addr;          /* where to return to */
    shrt *pc;                   /* PC before call/break */
    struct package_info *packP; /* package containing code for current PC */
    elf_object *inObj;          /* NULL or object whose method we are in */
    shrt num_args_passed;       /* how many args he actually passed */
    data_ptr *locvars;          /* array of locals ptrs */
    data_ptr **argvars;         /* array of arg val ptrs */
    data_ptr objVals;   /* NULL or objects to operate on */
    int currentObjVal;  /* where is vals array we are currently processing */
    shrt *objLoopPC;    /* when object looping, reset PC to this value */
    data_ptr **loopArgs;    /* copy of arg ptrs for re-push when looping */
    int numlocs;                /* number of local vars here */
    int numargs;                /* number of arg vars declared */
    int numregs;                /* number of register vars used */
    shrt functype;              /* TYPE_UIFUNC, TYPE_QFUNC, TYPE_MACRO... */
    shrt varnames;              /* index into varnames list */
    shrt *funcstarts;               /* addr of func descriptor */
    shrt *error_catcher;        /* NULL or error catcher in code frame */
    data_ptr **catcher_sp;       /* SP to restore at catcher time */
    shrt *error_addr;           /* addr error occured */
    bint error_code;            /* code this handler is working on */
    shrt retloc;                /* 0 or where to put return value */
    shrt curline;               /* source line number */
    unsigned int time_in;       /* microsecond entry time */
    unsigned recorded : 1,      /* true if recodring occured at this stack level */
            error_caught : 1,   /* RESET by 'ON ERROR' */
            stop_here : 1;  /* prevents over-returns on calls from C code */
} stack_frame;

#if VAX
    double to_align_good;   /* forces good alignment */
#endif

typedef struct package_info {   /* Elf package descriptor */
    shrt *code_starts;  /* where executable code begins */
    shrt *code_ends;    /* EXCLUSIVE code endpoint */
    gvar_info **gvar_list;  /* null ptr termed array */
    gfunc_info **gfun_list; /* null ptr termed array */
    ext_link **ext_list;    /* null ptr termed array of external vars/funcs */
    data_ptr *literals_table;         /* table of dataptr for literals */
    gvar_info **ovar_list;  /* null ptr termed array (just used by debugger) */
    gvar_info **svar_list;  /* null ptr termed array (just used by debugger) */
    char *varNames;     /* names of local & arg vars */
    data_ptr formats;   /* elf info about formats */
    char *moduleName;   /* produced from source_name */
    char *source_name;  /* name of source file/doc */
    int src_map;                /* TRUE if there is source info */
    int dynamic_load;     /* TRUE if this was LoadImaged rather than ELF_PRELOAD */
    int task_uid;       /* non-zero if should be de-install on task exit */
    int num_literals;                   /* how many in table */
} package_info;

/* internal breakpoint data */

typedef struct {
    package_info *packP;        /* package containing breakpointed code */
    int line;                   /* line number */
    data_ptr obj;               /* NULL or breakpoint object */
    int dbg_uid;                /* Debugger task to which this breakpoint belongs to */
} breakpoint_t;

typedef struct {
    int dbg_uid;                /* Debugger task */
    int debuggee_uid;           /* task under debug */
    unsigned is_stopped : 1,    /* is currently stopped due to brekpt/error */
            is_stepping : 1,    /* is currently stepping */
            is_nexting  : 1,    /* is currently nexting */
            is_error_break: 1;  /* is catching errors enabled */
    breakpoint_t exbrk;         /* execute to here brekpoint */
    stack_frame *nexterFrame;
    stack_frame *breakFrame;
    data_ptr *breakGlobals;
} debugger_info;

typedef struct elf_str {        /* user generated "getStr" string */
    struct elf_str *next;
    typeErr num;
    char *text;
} elf_str;

typedef struct garbage_info {   /* per-instruction temp mallocs */
    struct garbage_info *next;  /* next in chain */
    char data[1];
} garbage_info;

/************** THE TASK TYPEDEF **************************/

typedef struct elf_info {
    int status;          /* State of elf interpreter */
    shrt *pc;           /* Program counter */
    data_ptr **sp;       /* Stack Pointer */
    shrt errorLine;          /* line of instruction that threw error */
    package_info *errorPackage; /* Package id of where error started */
    stack_frame *curFrame;  /* Frame associated with currenbt routine */
    jmp_buf abortBuf;    /* Branches out of error condition */
    int recursion;       /* Current CallSub/Return recursion level */
    data_ptr *taskFrame;  /* gvar ptrs */
    bint ErrorCode;
    char *ErrorStr; /* goes along with ErrorCode */
    char *ErrorObject;  /* goes along with ErrorCode */
    char *ErrorFile;    /* goes along with ErrorCode */
    long ErrorLine; /* goes along with ErrorCode */
    char ErrorFunc[ELFIDSTR+1]; /* goes along with ErrorCode */
    elf_object *ErrorElfObj; /* goes along with ErrorCode */
    int RunJanitoring;  /* Check for garbage, or whether hit breakpoint */
    garbage_info *GarbageList;   /* head of per-inst garbage */
    char *functionCall; /* Name of Elfb function just called (or NULL) */
    int noErrorBox;     /* If TRUE, caller handles errors */
    int inited;     /* TRUE if ElfRun'd in this task */
    data_ptr sqlStruct; /* SQL generated by the OP_SQL opcodes */
} elf_info;

extern DLL_IMPORT elf_info Elf;
extern DLL_IMPORT elf_str *ElfStrings;     /* Head of user "getStr" string  objects list */
extern DLL_IMPORT bltin_desc ElfCoreFuncs[];
extern DLL_IMPORT int ElfIamInterpreter;   /* TRUE in ATM, false otherwise */
extern DLL_IMPORT int ElfLanguage;     /* language for symbol table lookups */

typeErr StrGet();       /* (n,p) getStr interface */
data_ptr MkNumData();
data_ptr WriteArray();
data_ptr ReadArray();
elf_object *ElfObjPtr();
data_any *CopyData();
data_ptr MkStrData();
data_ptr MkArray();
#define ELF_LOADED 1

#if !FOR_TESTER
void taskAbort();
void AxLogger();
void THIMexit();
#endif
#endif

extern DLL_IMPORT short THIMpid;

#define ELF_DATA_FREE(x)  AxTaskFreeElfData(THIMpid,x)
#define ElfDataFree(x) ELF_DATA_FREE(x) /* FIX ALL SOURCE !!0 */

#define ELF_DATA_COPY(x)  AxTaskCopyElfData(THIMpid,x)
#define ElfDataCopy(x) ELF_DATA_COPY(x) /* FIX ALL SOURCE !!0 */

#define ELF_CREATE_STRING(str) AxTaskCreateElfString(THIMpid,str)
#define MkStrData(len,str) ELF_CREATE_STRING(str)   /* FIX ALL SOURCE !!0 */

#define ELF_CREATE_ARRAY(size) AxTaskCreateElfArray(THIMpid,size)
#define MkArray(size) ELF_CREATE_ARRAY(size)    /* FIX ALL SOURCE !!0 */

#define ELF_CREATE_BINARY(size) AxTaskCreateElfBinary(THIMpid,size)
#define MkBinary(size) ELF_CREATE_BINARY(size)  /* FIX ALL SOURCE !!0 */

#define ELF_CREATE_ENTRY(size) AxTaskCreateElfEntry(THIMpid,size)
#define MkEntry(size) ELF_CREATE_ENTRY(size)    /* FIX ALL SOURCE !!0 */

#define ELF_CREATE_NUMBER(val) AxTaskCreateElfNumber(THIMpid,val)
#define MkNumData(val) ELF_CREATE_NUMBER(val)   /* FIX ALL SOURCE !!0 */

#define ELF_CREATE_INTEGER(intval) AxMakeIntData(intval)

#define ELF_CREATE_SMALL_INTEGER(int_30_bits) MK_INTEGER(int_30_bits)

#define ELF_CREATE_OBJECT(d,u,b) AxTaskCreateElfObject(THIMpid,d,u,b)
#define ELF_TASK_CREATE_OBJECT(p,d,u,b) AxTaskCreateElfObject(p,d,u,b)
#define MkObject(d,u,b) ELF_CREATE_OBJECT(d,u,b)    /* FIX ALL SOURCE !!0 */

#define ELF_CREATE_REMOTE_OBJECT(d,u,h) AxTaskCreateRemoteElfObject(THIMpid,d,u,h)
#define ELF_TASK_CREATE_REMOTE_OBJECT(p,d,u,h) AxTaskCreateRemoteElfObject(p,d,u,h)


data_ptr AxTaskCreateElfEntry AXP_((int task, char *name));
data_ptr AxTaskCreateElfObject AXP_((int task,int dex, int uid, data_ptr bl));
data_ptr AxTaskCreateElfNumber AXP_((int task,double val));
data_ptr AxTaskCreateElfString AXP_((int task, char *str));
data_ptr AxTaskCreateElfArray AXP_((int task,int cnt));
data_ptr AxTaskCreateRemoteElfObject AXP_((int task,int dex,int uid,char *hostname));
data_ptr AxTaskCreateElfBinary AXP_((int task,int cnt));

#if 0
data_ptr AxTaskCopyElfData AXP_((int task, data_ptr aP));
void AxTaskFreeElfData AXP_((int task, data_ptr aP));
data_ptr AxTaskCreateSizedElfString AXP_((int task, int len, char *str));
data_ptr AxMakeIntData AXP_((int num));
#endif


