/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: elio.c
   Applix ASC version: 100
   Date: Wed May  5 21:13:58 1999
*/
#if AXMS
#include <errnohack.h>
#include <io.h>
#endif

#include <math.h>
#include <aos.h>
#include <afile.h>
#include <doctypes.h>
#include <elf_env.h>
#include <eldata.h>
#include <getstr10.h>
#include <getstr4.h>
#include <elhelper.h>
#include <thimble.h>
#if !ELIO_STUB
#include <elio.p>
#endif
#define G_PREC   15  /* precesion for %g format */

extern data_ptr RecordedEvents;     /* Array of events that gets built up */
extern breakpoint_t ElfBreakData[]; /* breakpoint info */
extern int ElfNumBreakpoints;   /* number of breakpoints active */
extern debugger_info ElfDebugger[]; /* debuggers info */

typeErr ErrnoErr();

#if ELIO_STUB /* Stubs to allow linking to external programs (like elfpdb) */
#undef malloc
#undef realloc
#undef free
extern double AxFloatFromDataPtr();
#define XLT(x) (x)
#define NumFromDataPtr AxFloatFromDataPtr
#define TrueFree free
#define TrueMalloc malloc
#define AxTaskCreateElfArray(t,n) AxMakeArray(n)
#define IntFromDataPtr AxIntFromDataPtr
#define ParsePath AxParsePath
#define StrFromDataPtr AxStrFromDataPtr
#define WriteArray AxWriteArray
#define ReadArray AxArrayElement
#define AxTaskCreateElfNumber(t,n) AxMakeFloatData(n)
#define AxTaskCreateElfString(t,s) AxMakeStrData(-1,s)
#define TrueRealloc realloc
#define AxTaskCreateElfBinary(t,p) AxMakeBinaryData(p)
void taskAbort(char *msg) { abort(); }
data_ptr ElfRetData;
short THIMpid;
data_ptr AxTaskCreateElfObject(register int task, register int dex, register int uid, register data_ptr binderList){taskAbort("AxTaskCreateElfObject called");}
data_ptr AxTaskCreateElfEntry(register int task, register char *name){taskAbort("AxTaskCreateElfEntry caled");}
mbchar *AFileEnderFromInfo(AFileInfo *info_ptr, char *header){taskAbort("AFileEnderFromInfo called");}
void AGetDefaultFileInfo(AFileInfo *info_ptr){taskAbort("AGetDefaultFileInfo called");}
void AFileHeaderFromInfo(AFileInfo *info_ptr, char *header){taskAbort("AFileHeaderFromInfo called");}
short IsValidPathFilename(register char *file_path) {return(AOK);}
void TaskFree(register int pid, register void *inP) {free(inP);}
char *TaskAlloc(register int pid, unsigned int size){return(malloc(size));}
char *TaskRealloc(int pid, void *inP, unsigned int newsize) {return(realloc(inP,newsize));}
int ElfbFileWritable(char *name) {return(0);}
void ElfStrAbort(int code, char *str, char *prepend) { abort(); }  /* !!0 TEMP */
int ElfAbort(int code) { abort(); }     /* !!0 TEMP */
int FSFilterDocToCurRev(char *infile, char *outfile){taskAbort("FSFilterDocToCurRev called");}
short AFileInfoFromHeader(register char *headerLine, register AFileInfo *info_ptr){ return FENoHeader; }
void AxTaskFreeElfData(register int task, register data_ptr aP){taskAbort("AxTaskFreeElfData called");}
ElfbDebugFuncs (data_ptr dP1,data_ptr dP2,data_ptr dP3,data_ptr dP4,data_ptr dP5,data_ptr dP6,data_ptr dP7){return(0);}
#endif

#if FOR_TESTER
int AFTaskLockFile(char *pathname, int pid, int *force_read_only, int retry)
{
    return(AOK);
}
FILE *AFTaskfopen(int pid, char *pathname, char *mode_str)
{
    return(fopen(pathname, mode_str));
}
int AFfclose(char *pathname, int read_only, FILE *fp)
{
    return(fclose(fp));
}
int AFfcloseNoTruncate(char *pathname, int read_only, FILE *fp)
{
    return(fclose(fp));
}
void AFTaskUnlockFile(char *pathname, int pid, int read_only)
{
    return;
}
int AFGetAccessInfo(char *pathname, int *read_only, int *group, int *other, int *exists)
{
    *read_only = FALSE;
    *group = 0;
    *other = 0;
    *exists = FALSE;
    return(AOK);
}
#endif

/*************************************************************************
function: ElfbIsNull - Return Basic TRUE if arg is null
*/
ElfbIsNull(arg)
    register data_ptr arg;
{
    return (XX(arg) ? BFALSE : BTRUE);
}

/*************************************************************************
function: ElfbIsNumber - Return Basic TRUE if arg is a number datum
*/
ElfbIsNumber(arg)
    register data_ptr arg;
{
    if (IS_INTEGER(arg))
        return (BTRUE);
    else if (XX(arg) && (NX(arg)->type == TYPE_NUMBER))
        return (BTRUE);
    else
        return (BFALSE);
}

/*************************************************************************
function: ElfbIsString - Return Basic TRUE if arg is a string datum
*/
ElfbIsString(arg)
    register data_ptr arg;
{
    if (XX(arg) && !IS_INTEGER(arg) && (SX(arg)->type == TYPE_STRING))
        return (BTRUE);
    else
        return (BFALSE);
}

/*************************************************************************
function: ElfbIsArray - Return Basic TRUE if arg is a Array datum
*/
ElfbIsArray(arg)
    register data_ptr arg;
{
    if (XX(arg) && !IS_INTEGER(arg) && (AX(arg)->type == TYPE_ARRAY))
        return (BTRUE);
    else
        return (BFALSE);
}

/*************************************************************************
function: ElfbArraySize - Return size of an array
*/
ElfbArraySize(arg)
    register data_ptr arg;
{
    if (!XX(arg))
        return (0);
    VERIFY(!IS_INTEGER(arg) && AX(arg)->type == TYPE_ARRAY, ERR_NOTARRAY);
    return (AX(arg)->size);
}

/*************************************************************************
function: ElfbIsBinary - Return Basic TRUE if arg is a binary object
*/
ElfbIsBinary(arg)
    register data_ptr arg;
{
    if (XX(arg) && !IS_INTEGER(arg) && (SX(arg)->type == TYPE_BINARY))
        return (BTRUE);
    else
        return (BFALSE);
}

/*************************************************************************
function: ElfbBinarySize - Return size of an array
*/
ElfbBinarySize(arg)
    register data_ptr arg;
{
    data_binary    *bp;

    if (!XX(arg))
        return (0);
    VERIFY(!IS_INTEGER(arg) && AX(arg)->type == TYPE_BINARY, ERR_ILLVAL);   /* ??1 */
    bp = (data_binary *) arg;
    return (bp->size);
}

/*************************************************************************
function: ElfbIsObject - Return true if dataptr is an object
*/
int ElfbIsObject (p)
data_ptr p;
{
    if ( p && !IS_INTEGER(p) && (p->type == TYPE_OBJECT) )
        return(BTRUE);
    else
        return(FALSE);
}

/*************************************************************************
function: ElfbIsEntry - Return whether passed object is an entry var
*/
ElfbIsEntry(dp)
    data_ptr        dp;
{
    if (dp && (dp->type == TYPE_ENTRY))
        return (BTRUE);
    else
        return (FALSE);
}

/***********************************************************************
function: ElfAddBinaryToArray - Write a binary handle to array
*/
data_ptr
ElfAddBinaryToArray(ap, ix, elements, data_bytes)
data_ptr ap;
int ix;
int elements;
char *data_bytes;   /* may be NULL */
{
    register data_binary *bP;

    bP = (data_binary*)ELF_CREATE_BINARY(elements);
    if (data_bytes) {
        memcpy(bP->bytes, data_bytes, elements);
    }
    else {
        memset(bP->bytes, 0, elements);
    }
    return(WriteArray(ap, ix, (data_ptr) bP));
}

/*************************************************************************
function: ElfAddNumToArray - One step write a number to array
*/
data_ptr 
ElfAddNumToArray(ap, ix, n)
    data_ptr        ap;
    int             ix;
    int             n;
{
    return (WriteArray(ap, ix, MK_INTEGER(n)));
}

/*************************************************************************
function: ElfAddDblToArray - One step write a float number to array
*/
data_ptr 
ElfAddDblToArray(ap, ix, n)
    data_ptr        ap;
    int             ix;
    double          n;
{
    return (WriteArray(ap, ix, ELF_CREATE_NUMBER(n)));
}

/*************************************************************************
function: ElfAddStrToArray - One step write a string to array
*/
data_ptr 
ElfAddStrToArray(ap, ix, str)
    data_ptr        ap;
    int             ix;
    char           *str;
{
    return (WriteArray(ap, ix, ELF_CREATE_STRING(str)));
}

/*************************************************************************
function: LongFromArray - Retunr a long integer from an array pos */

LongFromArray(pp, dex)
    register data_ptr pp;
    register int    dex;
{
    register data_array *p = (data_array *) pp;

    if ( !p )
        return(0);
    VERIFY(!IS_INTEGER(p) && (p->type == TYPE_ARRAY), ERR_NOTARRAY);
    if (p->size <= (unsigned) dex)
        return (0);
    return (IntFromDataPtr(p->vals[dex]));
}

/*************************************************************************
function: IntFromArray - Return an integer from an array pos
*/
IntFromArray(pp, dex)
    register data_ptr pp;
    register int    dex;
{
    register data_array *p = (data_array *) pp;

    if ( !p )
        return(0);
    VERIFY(!IS_INTEGER(p) && (p->type == TYPE_ARRAY), ERR_NOTARRAY);
    if (p->size <= (unsigned) dex)
        return (0);
    return (IntFromDataPtr(p->vals[dex]));
}

/*************************************************************************
function: DblFromArray - Return a float from an array pos
*/
double 
DblFromArray(pp, dex)
    register data_ptr pp;
    register int    dex;
{
    double          NumFromDataPtr();
    register data_array *p = (data_array *) pp;

    if ( !p )
        return(0.0);
    VERIFY(!IS_INTEGER(p) && (p->type == TYPE_ARRAY), ERR_NOTARRAY);
    if (p->size <= (unsigned) dex)
        return (0);
    return (NumFromDataPtr(p->vals[dex]));
}

/*************************************************************************
function: AnyFromArray - Return an element from an array pos
*/
data_ptr 
AnyFromArray(pp, dex)
    register data_ptr pp;
    register int    dex;
{
    register data_array *p = (data_array *) pp;

    if ( !p )
        return(0);
    VERIFY(!IS_INTEGER(p) && (p->type == TYPE_ARRAY), ERR_NOTARRAY);
    if (p->size <= (unsigned) dex)
        return (0);
    return (p->vals[dex]);
}

/*************************************************************************
function: StrFromArray - Return a string ptr from an array
*/
char           *
StrFromArray(pp, dex)
    register data_ptr pp;
    register int    dex;
{
    register data_array *p = (data_array *) pp;

    if ( !p )
        return("");
    VERIFY(!IS_INTEGER(p) && (p->type == TYPE_ARRAY), ERR_NOTARRAY);
    if (p->size <= (unsigned) dex)
        return ("");
    return (StrFromDataPtr(p->vals[dex]));
}

#if !ELIO_STUB
/*************************************************************************
function: ElfbReadFile - Read an ascii file into an array of strings
*/
ElfbReadFile(filename)
    char           *filename;
{
    register FILE  *fp;
    register int    ix, len, line_count=0;
    register data_ptr aP;
    char            *buf, junk2[1001];
    int read_only, pid, err;


    /* open the file read only */
    read_only = TRUE;
    pid = THIMpid;
    err = AFTaskLockFile(filename, pid, &read_only, FALSE);
    if ( err != AOK )
        ElfStrAbort(err, NULL, filename);
    fp = (FILE *)AFTaskfopen(pid, filename, READ_STR); /* r */
    if ( !fp )
    {
        AFTaskUnlockFile(filename, pid, read_only);
        ElfStrAbort(ErrnoErr(), NULL, filename);
    }

    /* read the file */
    aP = ELF_CREATE_ARRAY(0);
#if FOR_TESTER /* add line count */
    while( fgets(junk2, 1000, fp) )
        line_count++;
    fflush(fp);
    AFfclose(filename, read_only, fp);
    fp = (FILE *)AFTaskfopen(pid, filename, READ_STR); /* r */
    if ( !fp )
    {
        AFTaskUnlockFile(filename, pid, read_only);
        ElfStrAbort(ErrnoErr(), NULL, filename);
    }
    aP = ElfAddStrToArray(aP, line_count, " " );
#endif
    for (ix = 0; buf = AxMFgets(fp); ix++)
        aP = ElfAddStrToArray(aP, ix, buf);

    /* done */
    AFfclose(filename, read_only, fp);
    AFTaskUnlockFile(filename, pid, read_only);
    ElfRetData = aP;
}
#endif

#define BLK 60          /* used for reading/writing binary from elf
                 * data files */
#define BL3 20          /* used for reading/writing binary from elf
                 * data files */
#define BLT 80          /* used for reading/writing binary from elf
                 * data files */

/*************************************************************************
function: bin_to_hex6 - Convert BLK bytes to a hex6 string
Result str is BLK+BL3+1 bytes.  Last char is a newline
*/
bin_to_hex6(result, inbyte)
    register unsigned char *result, *inbyte;
{
    register int    ix, jx;

    for (ix = 0; ix < BLK; ix++)
        result[ix] = (inbyte[ix] & 0x003F) + ' ';
    for (jx = 0; jx < BLK; ix++) {
        result[ix] = ((inbyte[jx++] & (unsigned) 0xC0) >> 2);
        result[ix] |= ((inbyte[jx++] & (unsigned) 0xC0) >> 4);
        result[ix] |= ((inbyte[jx++] & (unsigned) 0xC0) >> 6);
        result[ix] += ' ';
    }
    result[ix++] = '\n';
    result[ix] = 0;
}

/*************************************************************************
function: hex6_to_bin - Convert hex6 string back to BLK bytes
*/
hex6_to_bin(result, inbyte)
    register char  *result, *inbyte;
{
    register int    ix, jx;
    register unsigned int ch;

    for (ix = 0; ix < BLK; ix++)
        result[ix] = (inbyte[ix] - ' ') & 0x3F;
    for (jx = 0; jx < BLK; ix++) {
        ch = (inbyte[ix] - ' ') & 0xFF;
        result[jx++] |= (ch << 2) & 0xC0;
        result[jx++] |= (ch << 4) & 0xC0;
        result[jx++] |= (ch << 6) & 0xC0;
    }
}

static char *
AxMBufGetsNoCR (bufP, bufPosP, size, stripCR)
char *bufP;
int *bufPosP;
int size;
int stripCR;
{
    char *buf = AxMBufGets(bufP, bufPosP, size);

    if (buf && stripCR) {
        int len = strlen(buf);
        if (len > 0 && buf[len-1] == 0x0D)          /* CR - '\r' */
            buf[len-1] = 0;
    }
    return(buf);
}

/*************************************************************************
function: ElfReadOneTaskDatum - Read array from file, return created array object and,optionally, ** comments from file.
*/
static data_ptr 
ElfReadOneTaskDatum(line, bufP, bufPosP, bufSize, commentArrayP, task, stripCR)
    char           *line;   /* 1st line (maybe only line) of object */
    char           *bufP;
    int            *bufPosP;
    int             bufSize;/* total size of bufP */
    data_ptr        *commentArrayP; /* array, already allocated.  */
                    /* If not null, return comments in this array */
    int             task;   /* task id */
    int             stripCR;    /* if TRUE, strip CR at end */
{
    register data_ptr aP = NULL;
    register int    ix, cnt;
    data_ptr comments = NULL;
    double          atof(); /* not in some math.h files! */
    char            *linebuf;
    data_ptr results,commentLines;

again:
    if (bufSize == 0)
        goto returnHere;
    switch (line[0]) {

    case ':':   /* comment */
ignore_comment:
        if (!(linebuf = AxMBufGetsNoCR(bufP, bufPosP, bufSize,stripCR)) ) {
            TaskFree(task,bufP);
            ElfStrAbort(ERR_ILLVAL, XLT("Not Elf data"), NULL);
        }
        line = linebuf;
        goto again;

    case '@':
        cnt = atoi(line + 1);
        aP = ELF_CREATE_ARRAY(cnt);
        for (ix = 0; ix < cnt; ix++) {
            if ( !(line = AxMBufGetsNoCR(bufP, bufPosP, bufSize,stripCR)) ) {
                TaskFree(task,bufP);
                ElfStrAbort(ERR_ILLVAL, XLT("Not Elf data"), NULL);
            }
            WriteArray(aP, ix, ElfReadOneTaskDatum(line, bufP, bufPosP,
                                            bufSize, commentArrayP, task,stripCR));
        }
        break;
    case '#':
        aP = MK_INTEGER(atoi(line + 1));
        break;
    case '"':       /* older style left in for compatibility */
        aP = ELF_CREATE_STRING(line + 1);
        break;
    case '\'': {
        register char *src = line+1;
        register char *dst;
        char *buf;

        buf = malloc(strlen(src)*2+10);
        dst = buf;
        while ( *src ) {
            if ( *src == '\\' ) {
                src++;
                if ( *src == 'n' ) {
                    *dst++ = '\n';
                    src++;
                }
                else
                    *dst++ = *src++;
            }
            else
                *dst++ = *src++;
        }
        *dst++ = 0;
        aP = ELF_CREATE_STRING(buf);
        free(buf);
        break;
    }
    case '.':
        aP = ELF_CREATE_NUMBER(atof(line + 1));
        break;
    case '!':
        aP = ELF_CREATE_ENTRY(line + 1);
        break;
#if !ELIO_STUB
    case '&':{
            data_ptr dP;
            char           *elfHashName();

            cnt = atoi(line + 1);
            dP = ELF_CREATE_ARRAY(cnt);
            if ( !(linebuf = AxMBufGetsNoCR(bufP, bufPosP, bufSize,stripCR)) ) {
                TaskFree(task,bufP);
                ElfStrAbort(ERR_ILLVAL, XLT("Not Elf data"), NULL);
            }
            ElfAddStrToArray(dP, 0, linebuf + 1);
            while (--cnt > 0) {
                if ( !(linebuf = AxMBufGetsNoCR(bufP, bufPosP, bufSize,stripCR)) )
                    ElfStrAbort(ERR_ILLVAL, XLT("Not Elf data"), NULL);
                ElfAddStrToArray(dP, cnt, linebuf + 1);
            }
            aP = ELF_CREATE_OBJECT(0, 0, dP);
            ELF_DATA_FREE(dP);
            break;
        }
#endif
    case '*':   /* asterx file standard line. treat as comment. */

        if (  commentArrayP ) {
            if (*(line+1) == '*') {
                /* this is a "**" line that should be saved. */
                /* save comment line in comment array */
                *commentArrayP = ElfAddStrToArray(*commentArrayP,
                                        ElfbArraySize(*commentArrayP),
                                         line+2);
            }
            goto ignore_comment;
        }
        /* ??? fall thru ??? */

    case '%':{
            int             size;
            char            *inbuf;
            char            blkbuf[BLK];    /* needed because
                             * hex6_to_bin always
                             * does BLK bytes */
            char           *stuff;


            size = atoi(line + 1);
            aP = ELF_CREATE_BINARY(size);
            stuff = ((data_binary *) aP)->bytes;
            while (size > 0) {
                if ( !(inbuf = AxMBufGetsNoCR(bufP, bufPosP, bufSize,stripCR)) ) {
                    TaskFree(task,bufP);
                    ElfStrAbort(ERR_ILLVAL, XLT("Not Elf data"), NULL);
                }
                if ( strlen(inbuf) != BLT ) {
                    TaskFree(task,bufP);
                    ElfStrAbort(ERR_ILLVAL, XLT("Not Elf data"), NULL);
                }
                hex6_to_bin(blkbuf, inbuf);
                memcpy(stuff, blkbuf, MIN(size, BLK));
                stuff += BLK;
                size -= BLK;
            }
        }
    case '-':
        break;      /* null datum */
    default:
        TaskFree(task,bufP);
        ElfStrAbort(ERR_ILLVAL, XLT("Not Elf data"), NULL);
    }           /* treat anything else as null datum */

returnHere:
    /* return data and possibly comments. */
    return (aP);
}

/*************************************************************************
function: ConvertAFileInfoToELf - convert AFileInfo to Elf array  afile_info format
*/
static data_ptr ConvertAFileInfoToElf(info)
AFileInfo *info; /* in */
{
    register data_ptr dataP;

    dataP = ELF_CREATE_ARRAY(6);
    dataP = ElfAddNumToArray(dataP,0,info->encoding);
    dataP = ElfAddNumToArray(dataP,1,info->version);
    dataP = ElfAddNumToArray(dataP,2,info->docType);
    dataP = ElfAddNumToArray(dataP,3,info->old_version);
    dataP = ElfAddNumToArray(dataP,4,info->min_version);
    if (info->content) {
        dataP = ElfAddStrToArray(dataP,5,info->content);
    }
    return dataP;
}

/***************************************************************************
function: ElfReadDataAndInfoFromBuffer - read data array given elf data file in buffer
*/
data_ptr ElfReadDataAndInfoFromBuffer(task,bufP, size, file_infoP, commentsP)
char *bufP; /* contains data to read. Will be freed by this function. */
int size;   /* of buffer */
data_ptr *file_infoP;
data_ptr *commentsP;
{
    data_ptr        dataP, comments;
    int             bufPos = 0;
    char            *buf;
    AFileInfo       fileInfo;
    typeErr         err;
    int             stripCR = FALSE;
    int             len;


    /* read Applix standard header, if present. Temporarily null-terminate
    the first line in the buffer. */
    if ( !(buf = AxMBufGetsNoCR(bufP, &bufPos, size, stripCR)) ) {
        free(bufP);
        ElfStrAbort(ERR_ILLVAL, XLT("Not Elf data"), NULL);
    }

    len = strlen(buf);
    if (len > 0 && buf[len-1] == 0x0D) {        /* CR - '\r' */
        buf[len-1] = 0;
        stripCR = TRUE;
    }
    
    err = AFileInfoFromHeader(buf,&fileInfo);

    /* is header OK? */
    if (err == AOK || err == FETooNew ) {
        comments = ELF_CREATE_ARRAY(0);
        if ( !(buf = AxMBufGetsNoCR(bufP, &bufPos, size, stripCR)) ) { /* get next line */
            free(bufP);
            ElfStrAbort(ERR_ILLVAL, XLT("Not Elf data"), NULL);
        }
        if (file_infoP) {
            *file_infoP = ConvertAFileInfoToElf(&fileInfo);
        }
    }
    else if (err == FENoHeader) {
        comments = NULL; /* no Applix header */
        if (file_infoP)
            *file_infoP = NULL;
    }
    else {
        ElfStrAbort(err,NULL,NULL); /* error in header */
    }

    /* read data and comments from file */

    dataP = ElfReadOneTaskDatum(buf,
                                bufP,
                                &bufPos,
                                size,
                                ((comments) ? &comments : NULL),
                                task,
                                stripCR);
    free(bufP);
    if (commentsP) {
        *commentsP = comments;
    }
    else if (comments) {
        ELF_DATA_FREE(comments);
    }

    return dataP; /* return data */
}


/***************************************************************************
function: ElfReadDataFromBuffer - read data array given elf data file in buffer
*/
data_ptr ElfReadDataFromBuffer(task,bufP, size)
char *bufP; /* contains data to read. Will be freed by this function. */
int size;   /* of buffer */
{
    return ElfReadDataAndInfoFromBuffer(task,bufP,size, NULL, NULL);
}

/*************************************************************************
function: ElfbReadDataFile - Read a file into an Elf datum
The file is an ascii file that is almost human readable.  Each line
contains exactly one value.

Array:      @cnt
String:     "str
Number:     .number (number in g9 format)
Integer:    #value
Null Datum: -
*/
ElfbReadDataFile(filename)
    char           *filename;   /* file to read */
{
    int             bufPos = 0;
    char           *bufP;
    int             size;

    bufP = AxReadFileIntoBuf(filename, &size);
    SVERIFY(bufP, ErrnoErr(), filename);

    ElfRetData = ElfReadDataFromBuffer(THIMpid, bufP, size);
    return;
}


/********** WriteDateFile buffering code *****************/

static int      elfWrtBufSize, elfWrtBufPos;
static char    *elfWrtBuf;

static 
elfWrtBufStr(str)
    register char  *str;
{
    register int    len = strlen(str);

    if (elfWrtBufPos + len >= elfWrtBufSize)
        extendBuf(elfWrtBufPos + len);
    if ( str != (elfWrtBuf + elfWrtBufPos) )
        memcpy(elfWrtBuf + elfWrtBufPos, str, len);
    elfWrtBufPos += len;
}
static 
extendBuf(need)
int need;
{
    elfWrtBufSize = 2 * need;
    elfWrtBuf = TrueRealloc(elfWrtBuf, elfWrtBufSize);
}
static 
elfInitWrtBuf()
{
    elfDestroyWrtBuf();
    elfWrtBufSize = 2000;
    elfWrtBuf = TrueMalloc(elfWrtBufSize);
}
static 
elfDestroyWrtBuf()
{               /* optional call */
    if (elfWrtBuf)
        TrueFree(elfWrtBuf);
    elfWrtBufSize = 0;
    elfWrtBuf = NULL;
    elfWrtBufPos = 0;
}

static char *elfNeedBufSpace(cnt)
int cnt;
{
    if ( elfWrtBufSize <= elfWrtBufPos+cnt )
        extendBuf(elfWrtBufPos+cnt);
    return(&elfWrtBuf[elfWrtBufPos]);
}
#define EBPUTC(c)   if (elfWrtBufSize == elfWrtBufPos) \
                        extendBuf(elfWrtBufSize+1); \
                    elWrtBuf[elfWrtBufPos++] = c;
#define EBPUTS(s)   elfWrtBufStr(s)

/********** end of WriteDateFile buffering code *****************/

/*************************************************************************
function: ElfbWriteDataFile - Write an Elf data to a file
The file is an ascii file that is almost human readable.  Each line
contains exactly one value.
Array:      @cnt
String:     "str
Number:     .number (number in g9 format)
Integer:    #value
Null Datum: -
*/

int ElfbWriteDataFile(filenameP, aP, moreStuff)
data_ptr        filenameP;  /* file to create */
data_ptr        aP;
data_ptr    moreStuff;
{
    return(AxElfbWriteDataFile(filenameP, aP, moreStuff, FALSE));
}

/*************************************************************************
function: ConvertElfToAfileInfo - convert ELF afile_info format to AfileInfo
*/
static void ConvertElfToAFileInfo(elfdata, info)
data_ptr elfdata; /* in: elf array in afile_info format */
AFileInfo *info; /* out */
{
    data_ptr content;

    AGetDefaultFileInfo(info);
    info->encoding = (0 != IntFromArray(elfdata,0));
    info->version = IntFromArray(elfdata,1);
    info->docType = IntFromArray(elfdata,2);
    info->old_version = IntFromArray(elfdata,3);
    info->min_version = IntFromArray(elfdata,4);
    content = AnyFromArray(elfdata,5);
    if (ElfbIsString(content))
        info->content = StrFromDataPtr(content);
}

/*************************************************************************
function: AxFilePermissions - returns default group and world access for
new files (as 2-element array). For unix, gets umask. optional arg.
*/
data_ptr AxFilePermissions(filename)
data_ptr filename; /* name of existing file (optional) */
{
    int mask,code,group,all ;
    data_ptr perms;
    mbchar *str;
    struct stat     lstat;

    code = -1;

    /* get permissions of existing file */
    if (filename) {
        str = (mbchar *)StrFromDataPtr(filename);
        if ( MBstrlen(str)) {
            if (0 == Astat( (char *)str, &lstat)) {
                code = lstat.st_mode;
            }
        }
    }
        
    /* factor in the umask, for new files */
    if (code == -1) {
        code = 0666;
        mask = umask(0); umask(mask);
        code = code ^ mask; /* xor */
    }

    perms = ELF_CREATE_ARRAY(2);

    group = ((code & 060) == 060)
        ? 2
        : ((code & 040) == 040)
            ? 1
            : 0;
    perms = ElfAddNumToArray(perms, 0, group);

    all = ((code & 06) == 06)
        ? 2
        : ((code & 04) == 04)
            ? 1
            : 0;
    perms = ElfAddNumToArray(perms, 1, all);

    return(perms);
}

/*** This routine is called either fromthe ElfbWriteDataFile() which 
sets the last argument return_err to tbe false which means that elf is calling
this routine and it should use the ElfStrAbort() rather then returning false.
if this function has last argument set to true that means on error it should
return back false rather then calling ElfStrAbort(). This function was renamed
from ElfWriteDataFile() since now that function acts as a wrapper for this 
new named function -vikas ***/
/** fix for #10393 **/
int
AxElfbWriteDataFile(filenameP, aP, moreStuff, return_err)
data_ptr filenameP;  /* file to create */
data_ptr aP;         /* elf data to write */
data_ptr moreStuff;  /* a write_data_format@ array. See fileinf_.am. */
int      return_err; /* (T) return err or (F) abort on err */
{
    int       exists, all_access, grp_access, read_only, pid;
    int       err, ix, num_comments, mode, buf_len;
    mbchar    header[ALINELENGTH+2];
    char      dst[STRPATH + 1];
    char      backup[STRPATH + 1];
    char      junk[STRPATH + 1];
    char      *filename;
    char      *buf;
    data_ptr  comments;
    data_ptr  header_info;
    data_ptr  permissions;
    FILE      *fp, *backup_fp;
    AFileInfo fileInfo;
    long      pos;

    /* get the name of the file */
    filename = StrFromDataPtr(filenameP);
    AxSysDosUnixPath(filename);
    err = IsValidPathFilename(filename);
    if ( err != AOK )
        if (return_err) return(err); else ElfStrAbort(err,NULL,filename);
    strcpy(dst, filename);

    /* check permissions and lock the file for read/write */
    read_only = FALSE;
    err = AFGetAccessInfo(dst, &read_only,&grp_access,&all_access, &exists);
    if ( err != AOK )
        if (return_err) return(err); else ElfStrAbort(err,NULL,filename);
    if ( read_only == TRUE )
    {
        errno = EACCES;
        err = ErrnoErr();
        if (return_err) return(err); else ElfStrAbort(err,NULL,filename);
    }
    pid = THIMpid;
    err = AFTaskLockFile(filename, pid, &read_only, FALSE);
    if ( err != AOK )
        if (return_err) return(err); else ElfStrAbort(err,NULL,filename);
    if ( exists == FALSE )
        fp = (FILE *)AFTaskfopen(pid, filename, WRITE_STR); /* w */
    else
        fp = (FILE *)AFTaskfopen(pid, filename, READ_TRUNC_STR); /* r+ */
    if ( !fp )
    {
        err = ErrnoErr();
        AFTaskUnlockFile(filename, pid, read_only);
        if (return_err) return(err); else ElfStrAbort(err,NULL,filename);
    }

    /* create a backup file */
    if ( exists == FALSE )
        buf = (char *)0;
    else
    {
        ParsePath(dst, backup, junk);
        strcat(backup, NXLT("/tmXXXXXX"));
        AxSysDosUnixPath(backup);
        mktemp(backup);
        if ( backup[0] == (char)0 )
        {
            errno = EIO;
            err = ErrnoErr();
            AFfcloseNoTruncate(filename, read_only, fp);
            AFTaskUnlockFile(filename, pid, read_only);
            if (return_err) return(err); else ElfStrAbort(err,NULL,filename);
        }
        backup_fp = (FILE *)fopen(backup, WRITE_STR); /* w */
        if ( !backup_fp )
        {
            err = ErrnoErr();
            AFfcloseNoTruncate(filename, read_only, fp);
            AFTaskUnlockFile(filename, pid, read_only);
			if (return_err) return(err); else ElfStrAbort(err,NULL,backup);
        }
        pos = ftell(fp);
        fseek(fp, 0L, 2);
        buf_len = ftell(fp);
        if ( buf_len <= 0 ) 
        {
            buf = (char *)0;
            fclose(backup_fp);
        }
        else
        {
            buf = (char *)TaskAlloc(pid, buf_len+1);
            fseek(fp, 0L, 0);
            if ( fread(buf, buf_len, 1, fp) != 1 )
            {
                err = FEErrno+errno;
                fclose(backup_fp);
            }
            else /* if read is succesfull */
            {
                /* write the buffer to disk */
                if ( fwrite(buf, buf_len, 1, backup_fp) != 1 )
                {
                    err = FEErrno+errno;
                    fclose(backup_fp);
                }
                else
                {
                    if ( fclose(backup_fp) != 0 )
                       err = FEErrno+errno;
                    else
                       err = AOK;
                }
            }

            if ( err != AOK )
            {
                AFfcloseNoTruncate(filename, read_only, fp);
                AFTaskUnlockFile(filename, pid, read_only);
                if (return_err) return(err); else ElfStrAbort(err,NULL,filename);
            }
        }
        fseek(fp, pos, 0);
    }

    /* get some important file info */
    memset((char *)&fileInfo, 0, sizeof(AFileInfo));
    if ( moreStuff )
    {
        /* set filetype defaults */
        AGetDefaultFileInfo(&fileInfo);
        fileInfo.docType = AX_DOC_TYPE_ASCII; /* no header by default */

        /* get any comments */
        ElfbArraySize(moreStuff);   /* just to get err if not an array */
        comments = ReadArray(AX(moreStuff),0);
        num_comments = ElfbArraySize(comments);

        /* get access info */
        grp_access = IntFromArray(moreStuff,1);
        all_access = IntFromArray(moreStuff,2);

        /* get header info */
        header_info = AnyFromArray(moreStuff, 3);
        if (ElfbIsArray(header_info)) /* may set type and force header */
            ConvertElfToAFileInfo(header_info,&fileInfo);
    }
    else
    {
        /* no comments */
        comments = NULL;
        num_comments = 0;

        /* get access info */
        permissions = AxFilePermissions(filenameP);
        if ( ElfbIsArray(permissions) )
        {
            grp_access = IntFromArray(permissions,0);
            all_access = IntFromArray(permissions,1);
            AxFreeData(permissions);
        }

        /* init header info */
        fileInfo.docType = AX_DOC_TYPE_ASCII; /* no header by default */
    }

    /* write the data to a buffer */
    elfInitWrtBuf();
    if ( fileInfo.docType != AX_DOC_TYPE_ASCII ) /* standard file header */
    {
        AFileHeaderFromInfo( &fileInfo, (char *)header );
        MBstrcat(header, NXLT("\n"));
        EBPUTS(header);
    }
    for ( ix=0; ix<num_comments; ix++ ) /* comments */
    {
        sprintf(junk, "** %s\n", StrFromArray(comments,ix));
        EBPUTS(junk);
    }
    ElfWriteOneDatum(aP);
    if ( fileInfo.docType != AX_DOC_TYPE_ASCII ) /* file ender */
    {
        AFileEnderFromInfo(&fileInfo, (char *)header);
        MBstrcat(header, NXLT("\n"));
        EBPUTS(header);
    }

    /* write new stuff to original file */
    if ( fwrite(elfWrtBuf, elfWrtBufPos, 1, fp) != 1 )
    {
        err = ErrnoErr();

        /* restore backup on error */
        fseek(fp, 0L, 0);
        if ( buf_len <= 0 || buf == (char *)0 )
            unlink(backup);
        else if ( fwrite(buf, buf_len, 1, fp) == 1 )
            unlink(backup);
#if AXMS
        chsize(fileno(fp), ftell(fp));
#else
        ftruncate(fileno(fp), ftell(fp));
#endif

        AFfclose(filename, read_only, fp);
    }
    else
    {
        unlink(backup);
        if ( AFfclose(filename, read_only, fp) == EOF )
            err = ErrnoErr();
        else
        {
            err = AOK;
            mode = AxChmodValueFromCodes(dst, grp_access, all_access);
            chmod(dst,mode);
        }
    }

    /* done */
    AFTaskUnlockFile(filename, pid, read_only);
    elfDestroyWrtBuf();
    if ( buf )
        TaskFree(pid, buf);
    if ( err != AOK )
        if (return_err) return(err); else ElfStrAbort(err,NULL,filename);
    else
        return(AOK);
}

/*************************************************************************
function: ElfWriteOneDatum - Write datum/array to file
*/
static 
ElfWriteOneDatum(datum)
    data_ptr        datum;
{
    char            *buf = elfNeedBufSpace(200);
    int intval;

    if (!XX(datum)) {
        EBPUTS("-\n");
    } else if (IS_INTEGER(datum)) {
        sprintf(buf, "#%d\n", GET_INTEGER(datum));
        EBPUTS(buf);
    } else if (ElfbIsString(datum)) {
        register char *src, *dst;
        src = StrFromDataPtr(datum);
        buf = elfNeedBufSpace(strlen(src)*2+10);
        if ( !strchr(src,'\n') ) {
            sprintf(buf, "\"%s\n", src);
            EBPUTS(buf);
        }
        else {
            dst = buf;
            *dst++ = '\'';
            while ( *src ) {
                if ( *src == '\n' ) {
                    src++;
                    *dst++ = '\\';
                    *dst++ = 'n';
                }
                else if ( *src == '\\' ) {
                    *dst++ = *src++;
                    *dst++ = '\\';
                }
                else
                    *dst++ = *src++;
            }
            *dst++ = '\n';
            *dst++ = 0;
            EBPUTS(buf);
        }
    } else if (ElfbIsArray(datum)) {
        int             count, ix;

        count = AX(datum)->size;
        sprintf(buf, "@%d\n", count);
        EBPUTS(buf);
        for (ix = 0; ix < count; ix++)
            ElfWriteOneDatum(AX(datum)->vals[ix]);
    } else if (ElfbIsNumber(datum)) {
        sprintf(buf, ".%.16g\n", NumFromDataPtr(datum));
        EBPUTS(buf);
    } else if (ElfbIsEntry(datum)) {
        sprintf(buf, "!\n", EX(datum)->linkage.nameP);
        EBPUTS(buf);
#if !ELIO_STUB
    } else if (ElfbIsObject(datum)) {
#define MAX_OBJ_PARENTS 100
        elf_object     *parents[MAX_OBJ_PARENTS], *oP;
        int             numParents = 0;

        for (oP = ElfObjPtr(datum); oP; oP = oP->parentP) {
            parents[numParents++] = oP;
        }
        sprintf(buf, "&%d\n", numParents);
        EBPUTS(buf);
        while (numParents > 0) {
            numParents--;
            buf = elfNeedBufSpace(200);
            sprintf(buf, "'%s\n", parents[numParents]->nameTag);
            EBPUTS(buf);
        }
#endif
    } else if (ElfbIsBinary(datum)) {
        long            size;
        unsigned char   binbuf[BLK];
        unsigned char   outbuf[BLT+1];
        char           *stuff;

        size = BX(datum)->size;
        stuff = BX(datum)->bytes;
        sprintf((char *)binbuf, "%c%d\n",'%',size);
        EBPUTS(binbuf);
        while (size > 0) {
            memcpy(binbuf, stuff, MIN(size, BLK));
            bin_to_hex6(outbuf, binbuf);
            EBPUTS(outbuf);
            stuff += BLK;
            size -= BLK;
        }
    } else {
        EBPUTS("-\n");  /* entry pointers dumped as null datums */
    }
}

#if !FOR_TESTER
/*************************************************************************
function: ElfWriteObjectDataFile - Write Object Vars to file
*/
void ElfWriteObjectDataFile (filename, objP)
char *filename;
data_ptr objP;
{
    return;
}

/*************************************************************************
function: ElfReadObjectDataFile - Process object file and return object
*/
elf_object *ElfReadObjectDataFile(filename)
char *filename; /* file to read */
{
}

/*************************************************************************
function: ElfbWriteAppFile - Write all objects associated with an App to file
*/
ElfbWriteAppFile(filename, appP)
char           *filename;
data_ptr        appP;
{
}

#define ODEX_nameTag 0
#define ODEX_appName 1
#define ODEX_isTopLevelApp 2
#define ODEX_inheritFrom 3
#define ODEX_parent 4
#define ODEX_sourceLines 5
#define ODEX_sourceFileName 6
#define ODEX_localData 7
#define ODEX_children 8

/*************************************************************************
function: WriteOneObjectContent - Dump object to data file
Represents the content as an Elf array and dumps the array.
*/
static 
WriteOneObjectContent(oP)
    elf_object     *oP;
{
}

/*************************************************************************
function: ElfbReadAppFile - Process object file and return app object
*/
ElfbReadAppFile(filename)
char           *filename;   /* file to read */
{
}

#endif /* !FOR_TESTER */


#if !ELIO_STUB
/*************************************************************************
function: ElfArrayThatDescribesDatum - Returns array of strings
The array of strings are suitable for macro recording and the Elf debugger
*/
data_ptr ElfArrayThatDescribesDatum (name,argP,formatName,formatDim,formatInfo)
char *name;
data_ptr argP;
char *formatName;
int formatDim;
data_ptr formatInfo;
{
    char Buf[20000];
    data_ptr resultP;
    data_ptr elfDescribeArray();
    char *p;
    int ix, size=0, jx, did_work;
    char buf[ELFIDSTR+100+1];
    elf_object *obj;
    char *str, *str2;
    char namebuf[1024];

    resultP = MkArray(0);

#define FAULTED(p) ( LX(p) & FAULT_BIT )
#define UNFAULT(p) ((data_ptr)((long)p & ~FAULT_BIT))

    namebuf[0] = 0;
    if (name)
        sprintf(namebuf, "%s = ", name);

    if ( !IS_INTEGER(argP) && FAULTED(argP) )
        argP = UNFAULT(argP);

    if ( !argP ) {
        sprintf(Buf,NXLT("%snull"),namebuf);
        resultP = ElfAddStrToArray(resultP,AX(resultP)->size,Buf);
    }
    else if ( ElfbIsNumber(argP) ) {
        sprintf(Buf,NXLT("%s%.*g"),namebuf,G_PREC,NumFromDataPtr(argP));
        resultP = ElfAddStrToArray(resultP,AX(resultP)->size,Buf);
    }
    else if ( ElfbIsString(argP) ) {
        str = ElfQuoter(StrFromDataPtr(argP));
        sprintf(Buf,NXLT("%s%s"),namebuf,str);
        TaskFree(0,str);
        resultP = ElfAddStrToArray(resultP,AX(resultP)->size,Buf);
    }
#if !FOR_TESTER
    else if ( ElfbIsObject(argP) ) {
        if (OX(argP)->objHostname)
            sprintf(Buf,XLT("%sRemote Object"), namebuf);
        else {
            if (obj = ElfObjPtrIfExists(OX(argP),TRUE))
                sprintf(Buf,NXLT("%sObject %s, inheritsFrom %s"),namebuf,
                  obj->nameTag,
                  (obj->inheritFromP)?obj->inheritFromP->nameTag:NXLT("<null>"));
            else
                sprintf(Buf,NXLT("%sNon-existant object"),namebuf);
        }
        resultP = ElfAddStrToArray(resultP,AX(resultP)->size,Buf);
    }
#endif
    else if ( ElfbIsBinary(argP) ) {
        sprintf(Buf,XLT("%sBinary data, size = %d bytes"),namebuf,BX(argP)->size);
        resultP = ElfAddStrToArray(resultP,AX(resultP)->size,Buf);
    }
    else if ( ElfbIsArray(argP) ) {
        resultP = elfDescribeArray(0,name,argP,Buf,Buf,formatName,
                                   formatDim,formatInfo,resultP);
    }
    else {
        sprintf(Buf,XLT("%sDatatype cannot be displayed"),namebuf);
        resultP = ElfAddStrToArray(resultP,AX(resultP)->size,Buf);
    }

/* convert XXX[a][b] into XXX[a,b] for better readability */

    if (resultP)
    	  size = AX(resultP)->size;
    for ( ix = 0; ix < size; ix++ ) {
        str = StrFromArray(resultP,ix);
        if ( strlen(str) >= sizeof(buf) )   /* cannot happen */
            continue;
        strcpy(buf,str);
        did_work = FALSE;
        for ( jx = 0; buf[jx+1]; jx++ ) {
            if(jx && buf[jx] == ']' && buf[jx-1]=='\\' &&
                     buf[jx+1] == '\\' && buf[jx+2] == '[') {
                buf[jx-1] = ',';
                strcpy(buf+jx,buf+jx+3);
                did_work = TRUE;
            }
            else
              if ( jx && (buf[jx] == '[' || buf[jx] == ']')
                  && (buf[jx-1] == '\\') ) {
                strcpy(buf+jx-1,buf+jx);
                did_work = TRUE;
            }
        }
        if ( did_work )
            ElfAddStrToArray(resultP,ix,buf);
    }
    return(resultP);
}

/*************************************************************************
function: elfDescribeArray - Recursive dump of array assignment
*/
static data_ptr elfDescribeArray (level,name,dataP,buf,linebuf,
                    formatName,formatDim,formatInfo,resultP)
int level;
char *name;
data_ptr dataP;
char *buf, *linebuf;
char *formatName;
int formatDim;
data_ptr formatInfo;
data_ptr resultP;
{
    char *p,*pp, *dst;
    int eval_var=FALSE;
    int ix, jx, size;
    char *elementName;
    char *subFormat;
    int subDim;
    data_ptr cP, argDatum, array_str, subStrDatum, tmpP;
    char *str;
    char *elfGetFormatElementName();
    char *dot=NULL, *bracket=NULL, *member_dot, *member_bracket;
    char *bracket_close;
    char tmpName[ELFIDSTR+1],tmp_varname[ELFIDSTR+1];
    char member[ELFIDSTR+1], sub_varname[ELFIDSTR+1];
    int nChars=0, dot_idx=0, bracket_idx=0, bracket_close_idx=0;
    int array_idx=-1, eq_idx=0;
    data_ptr cmd1, cmd2, cmd3, varStr_array, subDatum, subDatum2;
    char *var_str;

    if ( !dataP )   /* ignore nulls */
        return;

    /* Initialize all char arrays to avoid garbage */
    tmpName[0] = 0;
    tmp_varname[0] = 0;
    member[0] = 0;
    sub_varname[0] = 0;

    argDatum = ELF_CREATE_ARRAY(1);
    argDatum = ElfAddStrToArray (argDatum, 0, name);
    argDatum = ElfAddStrToArray (argDatum, 1, ".");
    tmpP = (data_ptr)AxfStrIndex(argDatum);
    dot_idx = IntFromDataPtr(tmpP);
    argDatum = ElfAddStrToArray (argDatum, 1, "[");
    tmpP = (data_ptr)AxfStrIndex(argDatum);
    bracket_idx = IntFromDataPtr (tmpP);

	if (name) {
    	dot = strchr(name, '.');
    	bracket = strchr(name, '[');
	}
	else
		dot = bracket = 0;

    if (dot_idx > 0) {
	  if (dot_idx < bracket_idx || bracket_idx == 0) {
		/* there is a bracket after the dot */
		goto dot_first;
	  }
	  else if (bracket_idx > 0) {
		/* there is a bracket before the dot */
		goto bracket_first;
	  }
dot_first:
	  nChars = strlen(name)-strlen(dot);
	  strncpy(tmpName, name, nChars);
	  tmpName[nChars] = 0;
	  dot++;
	  strcpy(member, dot);
	  strcpy(sub_varname, dot);
    	  argDatum = ElfAddStrToArray (argDatum, 0, dot);
    	  argDatum = ElfAddStrToArray (argDatum, 1, ".");
    	  tmpP = (data_ptr)AxfStrIndex(argDatum);
    	  dot_idx = IntFromDataPtr(tmpP);
    	  argDatum = ElfAddStrToArray (argDatum, 1, "[");
    	  tmpP = (data_ptr)AxfStrIndex(argDatum);
    	  bracket_idx = IntFromDataPtr (tmpP);
    	  member_dot = strchr(dot, '.');
    	  member_bracket = strchr(dot, '[');
    	  if (dot_idx > 0) {
		  if (dot_idx < bracket_idx) {	/* dot before bracket */
	       	nChars = strlen(dot)-strlen(member_dot);
	  	  	strncpy(member, dot, nChars);
	  	  	member[nChars] = 0;
		  }
		  else
		  	goto bracket_next;
	  }
	  else if (bracket_idx>0) {
bracket_next:
		nChars = strlen(dot) - strlen(member_bracket);
		strncpy(member, dot, nChars);
		member[nChars] = 0;
	  }
    }
    else if (bracket_idx > 0) {

bracket_first:

	  nChars = strlen(name)-strlen(bracket);
	  strncpy(tmpName, name, nChars);
	  tmpName[nChars] = 0;
	  bracket_close = strchr(bracket, ']');
	  if (bracket_close)
		bracket_close++;
	  strcpy(member, bracket_close);
	  strcpy(sub_varname, bracket_close);

       argDatum = ElfAddStrToArray (argDatum, 1, "]");
       tmpP = (data_ptr)AxfStrIndex(argDatum);
       bracket_close_idx = IntFromDataPtr(tmpP);
	  subStrDatum = ELF_CREATE_ARRAY(1);
	  subStrDatum = ElfAddStrToArray (subStrDatum, 0, name);
	  subStrDatum = ElfAddNumToArray (subStrDatum, 1, bracket_idx+1);
	  subStrDatum = ElfAddNumToArray (subStrDatum, 2, bracket_close_idx-bracket_idx-1);
	  array_str = (data_ptr) AxfSubstr (subStrDatum);

	  if (ElfbIsString (array_str))
	  {	
		pp = StrFromDataPtr(array_str);
		for (ix=0; ix<strlen(pp); ix++)
			if (!isdigit(pp[ix])) {
				eval_var = TRUE;
				break;
			}

		if (eval_var == TRUE)
		{
	  /* array index is a variable, go get the value */
			cmd1 = (data_ptr)MK_INTEGER(14);	/* PRINT_EXPRESSION */
			cmd2 = array_str;				/* variable name */
			cmd3 = (data_ptr)MK_INTEGER(0);	/* level */
			ElfbDebugFuncs (cmd1,cmd2,cmd3,NULL,NULL,NULL,NULL);
			if (ElfbArraySize(ElfRetData) != 2)
				return (NULL);
			varStr_array = AnyFromArray(ElfRetData, 1);
			var_str = StrFromDataPtr(varStr_array);
			subDatum2 = ELF_CREATE_ARRAY(1);
			subDatum2 = ElfAddStrToArray (subDatum2,0,var_str);
			subDatum2 = ElfAddStrToArray (subDatum2,1,"=");
			tmpP = (data_ptr)AxfStrIndex(subDatum2);
			eq_idx = IntFromDataPtr(tmpP);

	     	subDatum = ELF_CREATE_ARRAY(1);
			subDatum = ElfAddStrToArray(subDatum, 0, var_str);
			subDatum = ElfAddNumToArray(subDatum, 1, eq_idx+1);
			array_str = (data_ptr)AxfSubstr (subDatum);
	  		array_idx = IntFromDataPtr(array_str);
		}
		else
	  		array_idx = IntFromDataPtr(array_str);
	  }
	  else
	  	array_idx = IntFromDataPtr(array_str);

	  if (array_idx < 0)
		return;		/* How did this happen? */
	  argDatum = ElfAddStrToArray (argDatum, 0, bracket_close);
	  argDatum = ElfAddStrToArray (argDatum, 1, ".");
	  tmpP = (data_ptr)AxfStrIndex(argDatum);
	  dot_idx = IntFromDataPtr(tmpP);
	  argDatum = ElfAddStrToArray (argDatum, 1, "[");
	  tmpP = (data_ptr)AxfStrIndex(argDatum);
	  bracket_idx = IntFromDataPtr (tmpP);
	  member_dot = strchr(bracket_close, '.');
	  member_bracket = strchr(bracket_close, '[');
	  if (dot_idx > 0 && dot_idx > bracket_idx) {
		nChars = strlen(bracket_close)-strlen(member_dot);
		strncpy(member, bracket_close,nChars);
		member[nChars] = 0;
	  }
	  else if (bracket_idx>0) {
		nChars = strlen(bracket_close)-strlen(member_bracket);
		strncpy(member, bracket_close,nChars);
		member[nChars] = 0;
	  }
    }
    else
	  strcpy (tmpName, (name ? name : ""));

    size = AX(dataP)->size;
    for ( ix = 0; ix < size; ix++ ) {
        p = buf;
        cP = ReadArray(AX(dataP),ix);
        if ( !cP )      /* ignore nulls in an array */
            continue;
        subFormat = "";
        if ( buf == linebuf && tmpName) {
            strcpy(p,tmpName);
            p += strlen(p);
        }
        subFormat = "";
        if ( formatName && formatName[0] ) {
            if ( formatDim > 0 ) {
                subFormat = formatName;
                subDim = formatDim-1;
                goto as_array;
            }
            elementName = elfGetFormatElementName(&subFormat,&subDim,ix,
                                formatName,formatInfo);

            if ( elementName && elementName[0] )
		  {		/* user didn't ask */
		  	 if (strlen(member)>0 && strcmp (elementName, member)) 
			 	continue;
                sprintf(p,NXLT(".%s"),elementName);
		  }
            else
		  {
                sprintf(p,NXLT("\\[%d\\]"),ix);
		  }
            p += strlen(p);
            if ( ElfbIsArray(cP) ) {
                resultP = elfDescribeArray(level+1,sub_varname,cP,p,linebuf,
                                    subFormat,subDim,formatInfo,resultP);
                continue;
            }
            else {
                elfBottomItem(cP,p);
                resultP = ElfAddStrToArray(resultP,AX(resultP)->size,linebuf);
            }
        }
        else {
as_array:   
		if (array_idx >= 0 && ix != array_idx)
			  continue;
		  sprintf(p,NXLT("\\[%d\\]"),ix);
            p += strlen(p);
            if ( ElfbIsArray(cP) ) {
                resultP = elfDescribeArray(level+1,sub_varname,cP,p,
							linebuf, subFormat,subDim,
							formatInfo, resultP);
                continue;
            }
            else {
                elfBottomItem(cP,p);
                resultP = ElfAddStrToArray(resultP,AX(resultP)->size,linebuf);
            }
        }
    }
    return(resultP);
}

static elfBottomItem(cP,p)
data_ptr cP;
char *p;
{
    char *str, *str2;
    elf_object *obj;

    if ( !cP )
        strcpy(p,NXLT(" = null"));
    else if ( ElfbIsNumber(cP) )
        sprintf(p,NXLT(" = %.*g"),G_PREC,NumFromDataPtr(cP));
    else if ( ElfbIsString(cP) ) {
        str = StrFromDataPtr(cP);
        str2 = ElfQuoter(str);
        sprintf(p,NXLT(" = %s"),str2);
        TaskFree(0,str2);
    }
#if !FOR_TESTER
    else if ( ElfbIsObject(cP) ) {
        obj = ElfObjPtrIfExists(OX(cP),TRUE);
        if ( obj )
            sprintf(p,NXLT(" = Object %s, inheritsFrom %s"),
              obj->nameTag,
              (obj->inheritFromP)?obj->inheritFromP->nameTag:NXLT("<null>"));
        else
            sprintf(p,NXLT(" = Non-existant object"));
    }
#endif
    else if ( ElfbIsBinary(cP) ) {
        sprintf(p,XLT(" = Binary data, size = %d bytes"), BX(cP)->size);
    }
    else {
        sprintf(p,XLT(" = Datatype cannot be displayed"));
    }
}

/*************************************************************************
function: ElfQuoter - Given a string, return a string with quotes around it
Handles embedded quotes, newlines and backslashes consistant with the language
*/
char *ElfQuoter (src)
register char *src;
{
    char *buf, *dst;

    dst = buf = TaskAlloc(0,strlen(src)*2+3);
    *dst++ = '"';
    while ( *src ) {
        switch ( *src ) {
        case '\\':
            *dst++ = '\\';
            *dst++ = '\\';
            break;
        case '\n':
            *dst++ = '\\';
            *dst++ = 'n';
            break;
        case '"':
            *dst++ = '\\';
            *dst++ = '"';
            break;
        default:
            *dst++ = *src;
        }
        src++;
    }
    *dst++ = '"';
    *dst++ = 0;
    return(buf);
}
#if !ELIO_STUB
/*************************************************************************
function: ElfbDebugFuncs - Stuff to support the debugger...
*/
ElfbDebugFuncs (dP1,dP2,dP3,dP4,dP5,dP6,dP7)
data_ptr dP1,dP2,dP3,dP4,dP5,dP6,dP7;
{
    int cmd;
    int px;
    package_info *packP;
    char *formatName;
    int fdim;

    cmd = IntFromDataPtr(dP1);

    switch ( cmd ) {

    case 1:         /* get array info */
        formatName = StrFromDataPtr(dP4);
        for ( fdim = 0; formatName[fdim] == '['; fdim++ );
        formatName += fdim;
        ElfRetData = ElfArrayThatDescribesDatum(
                        StrFromDataPtr(dP2),    /* name */
                        dP3,                    /* datum */
                        formatName,             /* formatName */
                        fdim,
                        dP5);                   /* formatInfo */
        break;

    case 2:         /* get formats in package */
        packP = FindPackageByName(StrFromDataPtr(dP2));
        if ( !packP )
            ElfRetData = NULL;
        else
            ElfRetData = ElfDataCopy(packP->formats);
        break;

#if !FOR_TESTER
    case 3:
        ElfRecord(ELFREC_EMIT_CHARS);
        ElfRetData = ElfDataCopy(RecordedEvents);
        break;

    case 4:             /* attach debugger */
        ElfAttachDebugger();
        break;

    case 5:             /* detach debugger */
        ElfDetachDebugger();
        break;

    case 6:             /* set break point */
        ElfRetData = MK_INTEGER(ElfSetBreakPoint(StrFromDataPtr(dP2),IntFromDataPtr(dP3),dP4));
        break;

    case 7: {
        int task, subcmd, pid;
        char msg_buf[512];
        
        task = IntFromDataPtr(dP2);
        subcmd = IntFromDataPtr(dP3);

        if (!ElfDebugResponse(task, subcmd)) {
            pid = ThimPidFromUID(task);
            if (!pid) {
                strcpy(msg_buf,XLT("Application already completed or hasn't been started yet."));
            }
            else if (TaskInfo[pid].rsp || !TaskInfo[pid].ipc_response_waiting) {
                strcpy(msg_buf, XLT("Application not listening now."));
            }
            ElfRetData = ELF_CREATE_STRING(msg_buf);
        }
        break;
    }

    case 8: {   /* get info about breakpoint */
        char funcname[ELFIDSTR+1];
        int line;
        int level;
        int ok;
        char src[STRPATH+1];
        data_ptr aP, oP;
        elf_object *inObj, *srcObj;

        level = IntFromDataPtr(dP2);
        ok = ElfWhereBreakFrame(funcname,&line,src,&inObj,&srcObj,level);
        if ( !ok )
            return;
        aP = MkArray(5);
        aP = ElfAddStrToArray(aP,0,funcname);
        aP = ElfAddNumToArray(aP,1,line);
        aP = ElfAddStrToArray(aP,2,src);
        if (inObj) {
            oP = ELF_CREATE_OBJECT(inObj->dex,inObj->uid,NULL);
            aP = WriteArray(aP,3,oP);
        }
        if (srcObj) {
            oP = ELF_CREATE_OBJECT(srcObj->dex,srcObj->uid,NULL);
            aP = WriteArray(aP,4,oP);
        }
        ElfRetData = aP;
        break;
    }
    case 9: {
        char *str;

        str = ElfQuoter(StrFromDataPtr(dP2));
        ElfRetData = MkStrData(-1,str);
        TaskFree(0,str);
        break;
    }
    case 10: {
        AxmHourglassTaskWindows(TaskInfo[AxInTask].uid);
        break;
    }
    case 11: {  /* clear break point */
        ElfRetData = MK_INTEGER(ElfClrBreakPoint(StrFromDataPtr(dP2),IntFromDataPtr(dP3),dP4));
        break;
    }
    case 12: {  /* clear all break points */
        ElfClrAllBreaks();
        break;
    }
#endif
    case 13: { /* list of breakpoints */
        data_ptr aP, oP, bP;
        int ix, jx;
        
        aP = ELF_CREATE_ARRAY(0);
        for (jx = 0, ix = 0; ix < ElfNumBreakpoints; ix++) {
            if (ElfBreakData[ix].dbg_uid == TaskInfo[THIMpid].uid) {
                bP = ELF_CREATE_ARRAY(3);
                bP = ElfAddStrToArray(bP,0,ElfBreakData[ix].packP->source_name);
                bP = ElfAddNumToArray(bP,1,ElfBreakData[ix].line);
                bP = WriteArray(bP,2,ElfDataCopy(ElfBreakData[ix].obj));
                aP = WriteArray(aP, jx, bP);
                jx++;
            }
        }
        ElfRetData = aP;
        break;
    }

    case 14: {                   /* print a variable */
        char *p;
        char varname[ELFIDSTR+1];
        char *dot=NULL, *bracket=NULL;
        char tmp1[ELFIDSTR+1];
        char funcname[2000], packname[STRPATH+1];
        int line, level, ok;
        data_ptr varP,formatInfo;
        elf_object *inObj, *srcObj;
        char buf[2000], tmp_var[2000];
        int ix;
        data_ptr aP,bP;
        typeErr err;
	   int nChars=0, dot_idx=0, bracket_idx=0, bracket_close_idx=0;
	   data_ptr cP, argDatum, tmpP;
        
        strcpy(tmp1, StrFromDataPtr(dP2));
	   argDatum = ELF_CREATE_ARRAY(1);
	   argDatum = ElfAddStrToArray (argDatum, 0, tmp1);
	   argDatum = ElfAddStrToArray (argDatum, 1, ".");
	   tmpP = (data_ptr)AxfStrIndex(argDatum);
	   dot_idx = IntFromDataPtr(tmpP);
	   argDatum = ElfAddStrToArray (argDatum, 1, "[");
	   tmpP = (data_ptr)AxfStrIndex(argDatum);
	   bracket_idx = IntFromDataPtr (tmpP);

	   dot = strchr (tmp1, '.');
	   bracket = strchr (tmp1, '[');

	   if (dot_idx > 0) {
		 if (dot_idx < bracket_idx || bracket_idx == 0)
			goto dot_first;
		 else if (bracket_idx > 0)
			goto bracket_first;
		 else
			goto default_action;
dot_first:
		 nChars = strlen(tmp1) - strlen(dot);
		 strncpy(varname, tmp1, nChars);
		 varname[nChars] = 0;
		 dot++;
	   }
	   else if (bracket_idx > 0) {
bracket_first:
		 nChars = strlen(tmp1)- strlen(bracket);
		 strncpy (varname, tmp1, nChars);
		 varname[nChars] = 0;
	   }		
	   else
default_action:
		 strcpy (varname, tmp1);

        for (p = tmp1; *p; p++)
            *p = toupper(*p);
        for (p = varname; *p; p++)
            *p = toupper(*p);
        level = 0;
        if (dP3)
            level = IntFromDataPtr(dP3);
        ok = ElfWhereBreakFrame(funcname,&line,packname,&inObj,&srcObj,level);
        if (!ok)
            return;
        if (funcname[0] == '?') {
            sprintf(buf,"get method %s in object ``%s''",funcname+1,inObj->nameTag);
            strcpy(funcname,buf);
        }
        else if (funcname[0] == '!') {
            sprintf(buf,"set method %s in object ``%s''",funcname+1,inObj->nameTag);
            strcpy(funcname,buf);
        }
        else {
            sprintf(buf,"macro %s",funcname);
            strcpy(funcname,buf);
        }
        aP = ELF_CREATE_ARRAY(0);
        err = ElfFetchVar(&varP,&formatName,&formatInfo,varname,level);
        if (err != AOK ) {
            sprintf(buf,XLT ("No variable '%s' in %s"),varname,funcname);
            ElfRetData = ElfAddStrToArray(aP,0,buf);
            return;
        }
	   strcpy (varname, tmp1);
        if ( formatName[0] )
            sprintf(buf,XLT ("*print %s(%s) in %s"),varname,formatName,funcname);
        else
            sprintf(buf,XLT ("*print %s in %s"),varname,funcname);

        aP = ElfAddStrToArray(aP,0,buf);
        for (fdim = 0; formatName[fdim] == '['; fdim++);
        formatName += fdim;
        bP = ElfArrayThatDescribesDatum(varname,varP,formatName,fdim,formatInfo);
        for (ix = 0; ix < AX(bP)->size; ++ix)
            aP = WriteArray(aP,ix+1,AnyFromArray(bP,ix));
        ElfRetData = aP;
        return;
    }

    case 15: {          /* set/get ElfErrorBreak status */
        int subcmd;
        int status;
        
        subcmd = 0;
        if (dP2)
            subcmd = IntFromDataPtr(dP2);
        if (subcmd == 0)        /* get ElfErrorBreak status */
            ElfRetData = MK_INTEGER((ElfGetErrorBreak()) ? (BTRUE) : (BFALSE));
        else if (subcmd == 1) { /* set ElfErrorBreak status */
            if (dP3)
                ElfSetErrorBreak(((IntFromDataPtr(dP3)==BFALSE) ? FALSE : TRUE));
        }
        return;
    }
        
    case 16: {
        char *p;
        char varname[ELFIDSTR+1];
        char funcname[2000], packname[STRPATH+1];
        int line;
        data_ptr varP,formatInfo;
        elf_object *inObj, *srcObj;
        int vx, ix;
        data_ptr aP, retP;
        
        if (!ElfbIsArray(dP2))
            return;
        retP = ELF_CREATE_ARRAY(0);
        for (vx = 0; vx < AX(dP2)->size; ++vx) {
            strcpy(varname, StrFromArray(dP2,vx));
            for (p = varname; *p; p++)
                *p = toupper(*p);
            if (!ElfWhereBreakFrame(funcname,&line,packname,&inObj,&srcObj,0))
                continue;
            if (ElfFetchVar(&varP,&formatName,&formatInfo,varname,0) != AOK ) {
                aP = ELF_CREATE_ARRAY(0);
                aP = ElfAddStrToArray(aP,0,XLT("No such variable"));
            }
            else {
                for (fdim = 0; formatName[fdim] == '['; fdim++)
                    ;
                formatName += fdim;
                aP = ElfArrayThatDescribesDatum(NULL,varP,formatName,
                                                fdim,formatInfo);
            }
            retP = WriteArray(retP, vx, aP);
        }
        ElfRetData = retP;
        return;
    }

    case 17:    /* is debugger stopped */
        ElfRetData = MK_INTEGER((ElfDbgGetStopStatus() ? BTRUE : BFALSE));
        return;

    case 18:    /* execute up to here */
        ElfRetData = ElfDbgExecuteToHere(IntFromDataPtr(dP2),   /* task */
                                         StrFromDataPtr(dP3),   /* packname */
                                         IntFromDataPtr(dP4));  /* lineno */
        return;

    case 19:    /* stop in func, [<packname> | <object>] */
        ElfRetData = ElfDbgStopInFunc(dP2, dP3);
        return;

    default:
        ElfAbort(ERR_ILLVAL);
    }
}
#endif
/*************************************************************************
function: elfGetFormatElementName - Get info about format element
*/
static char *elfGetFormatElementName (subName,subDimP,dex,formatName,formatInfo)
char **subName;     /* returned subformat name or "" */
int *subDimP;       /* returned subformat dimensions */
int dex;
char *formatName;
data_ptr formatInfo;
{
    data_ptr formP, elementP;
    int ix, size;
    int subDim;
    char *name;

    size = ElfbArraySize(formatInfo);
    for ( ix = 0; ix < size; ix++ ) {
        formP = ReadArray(AX(formatInfo),ix);
        if ( streq(formatName,StrFromArray(formP,0)) )
            goto got_format;
    }
    return(NULL);

got_format:

    dex++;      /* 1st element is format Name */
    elementP = ReadArray(AX(formP),dex);
    if ( ElfbIsArray(elementP) ) {
        name = StrFromArray(elementP,1);
        for ( subDim = 0; name[subDim] == '['; subDim++ );
        *subDimP = subDim;
        *subName = name + subDim;
        return(StrFromArray(elementP,0));
    }
    else {
        *subName = "";
        *subDimP = 0;
        return(StrFromDataPtr(elementP));
    }
}
#endif /* !ELIO_STUB */
/*
*   ElfWriteDataToBuf - write the elf data ptr to a buf.
*       Caller should use the 'data' and immediately call ElfDestroyData
*           to free the data.
*/
ElfWriteDataToBuf(aP, data, size)
data_ptr aP;    /* IN, the data ptr containing data */
char **data;    /* the address of the buffer written to */
int *size;      /* the size of the buffer */
{
    elfInitWrtBuf();
    ElfWriteOneDatum(aP);
    *data = elfWrtBuf;
    *size = elfWrtBufPos;
}

/*
*   free the elfWrtBuf and reset elfWrtBufPos
*/
ElfDestroyDataBuf()
{
    elfDestroyWrtBuf();
}

/*************************************************************************
function: AxChmodValueFromCodes - Return a "chmod" value
Gets execute permissions from file itself
*/
int AxChmodValueFromCodes (pathname,grpAccess,allAccess)
char *pathname;
int grpAccess, allAccess;
{
    int mode = 0600;
    struct stat status;

    if ( grpAccess == 1 )       /* read only */
        mode |= 040;
    else if ( grpAccess == 2 )      /* writable */
        mode |= 060;
    if ( allAccess == 1 )       /* read only */
        mode |= 04;
    else if ( allAccess == 2 )      /* writable */
        mode |= 06;

/* Add in Execute permission if there were such */

    if ( 0 == stat(pathname,&status) )
        mode |= status.st_mode & 0111;
    return(mode);
}

#if !FOR_TESTER
/***************************************************************************
function: AxfReadElfFile - read a elf data file.
returns a two element array in read_elf_format@ :
first element is a write_data_format@ containing afile_info from header (null
	if no header), comments, and permissions.
second element is the elf data array.

If file has a header, and is not current version, a filter is run.
*/
data_ptr AxfReadElfFile(argDatum)
data_ptr argDatum;
{
    mbchar *filename = StrFromArray(argDatum,0);
	int noFiltering = IntFromArray(argDatum,1); /* supress conversion */
    char currentVersionName[STRPATH+1];
    int             bufPos = 0;
    char           *bufP;
    int             size;
    data_ptr retval = NULL;
    data_ptr data, write_data_format, comments, access, file_info;
    AFileInfo fileInfo;
	int converted;
 
	/* get file permissions */
	access = AxFilePermissions(AxArrayElement(argDatum,0));

	/* filter to current rev unless noFiltering flag is set */
	if (noFiltering) {
		converted = FALSE;
		strcpy(currentVersionName,filename);
	}
	else {
		converted = FSFilterDocToCurRev(filename,currentVersionName);
	}
    bufP = AxReadFileIntoBuf(currentVersionName, &size);
    SVERIFY(bufP, ErrnoErr(), filename);
	if (converted)
		AxDeleteTempfile(currentVersionName);
 
	/* read file into data array */
    data = ElfReadDataAndInfoFromBuffer(THIMpid, bufP, size,
                                            &file_info, &comments);
	/* construct return array */
    retval = AxMakeArray(2);
	write_data_format = AxMakeArray(4);
    write_data_format = AxAddArrayToArray(write_data_format, 0, comments);
    write_data_format =
		AxAddDataToArray(write_data_format, 1, AxArrayElement(access, 0));
    write_data_format =
		AxAddDataToArray(write_data_format, 2, AxArrayElement(access, 0));
    write_data_format = AxAddArrayToArray(write_data_format, 3, file_info);

    retval = AxAddArrayToArray(retval, 0, write_data_format);
    retval = AxAddArrayToArray(retval, 1, data);
	return retval;
}
#endif
