/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: ellex.c
   Applix ASC version: 52
   Date: Tue Feb  2 15:21:09 1999
*/
#if 1           /* Yacc foobar, see elf_syntax.y for more info */
int yylval;
#endif

#define FOR_LEX

#if BLD_SOLARIS2_0 || LINUX
#include <unistd.h>
#endif

#include <ctype.h>
#if !NO_VARARGS
#include <varargs.h>
#endif
#include <elcomp.h>
#include <eltokens.h>
#include <style.h>      /* for parsenum */
#include <getstr4.h>
#include <fsrts.h>
#include <afile.h>
#include <elf_env.h>
#include <ellex.p>

int ElfcLastFuncKeyAtLine;  /* hack for object code buster-upper...records last
                                occurance of macro/function/uimacro... */

extern char *ElfcSrcFileName;		/* Needed for lexer's FILE@@@ function */
extern int IsElfTokSep[256];

extern char *ElfOpChars;   /* gotten from string file ELFOPCHARS */
/* ". ! , : ; ( ) [ ] += -= >= <= <> = ^ - * / \\ >>= <<= >> << ++ > < + { }  "; */

axstatic int ElfOpTokens[] = {
TK_DOT, TK_ENTRY, TK_COMMA, TK_COLON, TK_DOLLAR, TK_HOOK, TK_SEMICO, TK_OPEN, TK_CLOSE, TK_OBRAK, TK_CBRAK,
TK_ADDTO, TK_SUBTO, TK_GEQ, TK_LEQ,
TK_NEQ, TK_EQUALS, TK_EXP, TK_SUB, TK_MULT, TK_DIV, TK_IDIV, TK_SGEQ, TK_SLEQ,
TK_SGRTR, TK_SLESS, TK_SADD, TK_GRTR, TK_LESS, TK_ADD, TK_BEGINSET, TK_ENDSET
};

extern DocInfoType ElfcIncDocInfo;  /* include file doc info struct */

#define WithLine(x) ((((long)LineCount)<<16)|(x))
#define TAB 9

static char *TokStarts;
axglobal int LineCount;              /* line counter */

axglobal def_info **Defines;   /* heads of defines list */
axglobal tok_ptr *ElfTokArray;
axglobal int TokPos = 0xFFFF;
axglobal int DoingComment = FALSE;
axglobal int DoingInclude;
static int HaveReadFirstLineOfIncludeFile;
axglobal FILE *FpInclude;
axglobal char *ElfcLineBuf;      /* ElfcAlloc'ed fgets buffer */
extern char *DirSlash;
extern data_ptr ElfCompSrcArrayP;

/* The include file name is captured as "LastInclude", which is never reset
because the lexer tends to be slightly ahead of the parser.  This means that
customers MUST put their formats in include files. */
axglobal char ElfcLastInclude[STRPATH+1];

static int TokType; /* value returned by yylex */
static int TokVal;  /* yylval set by yylex */

/*************************************************************************
function: yylex - Called by parser until it returns a -1
*/
yylex ()
{
    register tok_ptr tp;

    while ( (TokPos == 0xFFFF) ||
            (ElfTokArray[TokPos].simple->type == TKT_EMPTY) )
        ElfcReadLine();

    tp = ElfTokArray[TokPos++];

    switch ( tp.simple->type ) {
    case TKT_SIMPLE:
        yylval = TokVal = WithLine(0);
        return(TokType = tp.simple->code);
    case TKT_STRING:
        yylval = TokVal = WithLine(AddStrConst(tp.str->str));
        return(TokType = TK_STRVAL);
    case TKT_NUMBER:
        yylval = TokVal = WithLine(AddNumConst(tp.num->val));
        return(TokType = TK_NUMVAL);
    case TKT_INTEGER:
        yylval = TokVal = WithLine(AddIntConst(tp.small->val));
        return(TokType = TK_NUMVAL);
    case TKT_ID:
        yylval = TokVal = WithLine(AddIdentifier(tp.id->name));
        return(TokType = TK_ID);
    case TKT_EOF:
        TokPos = 0xFFFF;
        return(TokType = -1);
    default:
        abort();return(0);
    }
}

/*************************************************************************
function: ElfcReadLine - Read, tokenize, do defines, next input line
*/
ElfcReadLine()
{
    int num_toks, total_toks;
    int ix;
    register tok_ptr tp;
    tok_ptr my_array[MAX_TOKENS+1];
    int did_a_replace = FALSE;
    int loop_count = 0; /* recursion counter while replacing */
    def_info *dP;

/* get a line from the input device */

    TokPos = 0;
    if ( DoingInclude ) {
        if ( NULL == fgets(ElfcLineBuf,MAX_LINELEN,FpInclude) ) {
closei:     DoingInclude = FALSE;
            fclose(FpInclude);
			FpInclude = NULL;
            ElfTokArray[0].simple = (tok_simple*)ElfcAlloc(sizeof(tok_simple));
            ElfTokArray[0].simple->type = TKT_EMPTY;
            return;
        }
        if ( !HaveReadFirstLineOfIncludeFile ) {
            if ( !strncmp(NXLT("*START"),ElfcLineBuf,6) || 
                 !strncmp("227879",ElfcLineBuf,6) ) {
                TokErr(XLT("Include files must be saved as ASCII (not formatted) using 'File -> Save As...'"));
                goto closei;
            }
			if (strlen(ElfcLineBuf) == MAX_LINELEN-1)  {
                TokErr(XLT("line too long"));
                goto closei;
			}
            HaveReadFirstLineOfIncludeFile = TRUE;
        }
    }
    else if ( ElfCompSrcArrayP ) {  /* Memory array mode */
        data_ptr dP;
        dP = ReadArray(AX(ElfCompSrcArrayP),LineCount++);
        if ( !dP ) {
            ElfTokArray[0].simple = (tok_simple*)ElfcAlloc(sizeof(tok_simple));
            ElfTokArray[0].simple->type = TKT_EOF;
            return;
        }
        else {
			char * str;
			str = StrFromDataPtr(dP);
			if (strlen(str) > MAX_LINELEN)
                TokErr(XLT("line too long"));
			else
				strcpy(ElfcLineBuf,str);
        }
    }
    else {          /* File object mode */
        abort();
    }

/* tokenize directly into the final array */

    ElfcTokenizeLine(ElfcLineBuf,ElfTokArray,&num_toks);

/* see if there are any identifiers matching replace list */

retry_replaces:
    for ( ix = 0; ix < num_toks; ix++ ) {
        tp = ElfTokArray[ix];
        if ( tp.id->type == TKT_ID ) {
            for ( dP = Defines[hash(tp.id->name)]; dP; dP = dP->next ) {
                if ( streq(dP->name,tp.id->name) ) {
                    if ( (ix == 1)
                            && (ElfTokArray[0].simple->type == TKT_SIMPLE)
                            && (ElfTokArray[0].simple->code == TK_DEFINE) )
                        continue; /* don't replace foo in #define foo ... */
                    else
                        goto do_replaces;
                }
            }
        }
    }

/* no, just put in line termination and check for control info */

    ElfTokArray[num_toks].simple = (tok_simple*)ElfcAlloc(sizeof(tok_simple));
    ElfTokArray[num_toks].simple->type = TKT_EMPTY;
    goto check_control;

/* yes, copy array and do replaces into ElfTokArray */

do_replaces:
    MEMCPY(my_array,ElfTokArray,num_toks*sizeof(tok_ptr));
    total_toks = 0;
    for ( ix = 0; ix < num_toks; ix++ ) {
        tp = my_array[ix];
        if ( tp.id->type != TKT_ID ) {
            ElfTokArray[total_toks++] = tp;
            continue;
        }
        for ( dP = Defines[hash(tp.id->name)]; dP; dP = dP->next ) {
            if ( !streq(dP->name,tp.id->name) )
                continue;
            if ( dP->num_toks + total_toks > MAX_TOKENS-4 ) {
                TokErr(XLT("line too long"));
                ElfTokArray[total_toks].simple = (tok_simple*)
                    ElfcAlloc(sizeof(tok_simple));
                ElfTokArray[total_toks].simple->type = TKT_EMPTY;
                return;
            }
            else {
                MEMCPY(ElfTokArray+total_toks,dP->toks,
                    dP->num_toks * sizeof(tok_ptr));
                total_toks += dP->num_toks;
                did_a_replace = TRUE;
                goto did_load;
            }
        }
        ElfTokArray[total_toks++] = tp;
did_load:;
    }

    if ( !did_a_replace )
        return;
    if ( loop_count++ > 20 ) {
        TokErr(XLT("replace depth too big"));
        ElfTokArray[total_toks].str = (tok_str*)ElfcAlloc(sizeof(tok_simple));
        ElfTokArray[total_toks].simple->type = TKT_EMPTY;
        return;
    }
    num_toks = total_toks;
    did_a_replace = FALSE;
    goto retry_replaces;

/* check for defines and includes */

check_control:
{ int type;
  char *iname;
  char path[STRPATH+1];
  int ix;
  int hashval;
  data_ptr searchList;
  int searchListCount;

    tp = ElfTokArray[0];
    if ( tp.simple->type != TKT_SIMPLE )
        return;
    type = tp.simple->code;
    if ( (type == TK_DEFINE) && (num_toks > 2) ) {
        dP = (def_info*)ElfcAlloc(sizeof(def_info));
        hashval = hash(ElfTokArray[1].str->str);
        dP->next = Defines[hashval];
        Defines[hashval] = dP;
        iname = ElfTokArray[1].str->str;
        dP->name = ElfcStrAlloc(iname);
        dP->toks = (tok_ptr*)ElfcAlloc((num_toks-2)*sizeof(tok_ptr));
        MEMCPY(dP->toks,&ElfTokArray[2],
            sizeof(tok_ptr) * (dP->num_toks = num_toks-2));
        ElfTokArray[0].simple->type = TKT_EMPTY;
    }
    else if ( (type == TK_INCLUDE) && (num_toks == 2) &&
               (ElfTokArray[1].simple->type == TKT_STRING) &&
                !DoingInclude && strlen(iname = ElfTokArray[1].str->str) ) {
        AxSysDosUnixPath( iname );

		if ( 0 == access(iname,R_OK) ) {	/* see if its available in current dir */
            strcpy(path,iname);
			goto read_it;
		}

        if ( iname[0] == *DirSlash || iname[0] == '.' )
            {
            /* a fully qualified pathname */
            if ( 0 != Aaccess(iname,R_OK) )
                goto ferr;
            strcpy(path,iname);
read_it:
            if ( FpInclude = fopen(path,"r") ) {
                DoingInclude = TRUE;
                HaveReadFirstLineOfIncludeFile = FALSE;
                strcpy(ElfcLastInclude,path);
                ElfTokArray[0].simple->type = TKT_EMPTY;
                return;
            }
        }
srchlister:
        ElfbGetSysVar(NXLT("elf_search_list@"));
        searchList = ElfRetData;
        searchListCount = ElfbArraySize(searchList);
        if ( 0 == searchListCount ) {
            ElfSetStandardSearchList();
            goto srchlister;
        }
        for ( ix = 0; ix < searchListCount; ix++ ) {
            AxMakePath(path,StrFromArray(searchList,ix),iname);
            if ( 0 == access(path,R_OK) ) {
                ElfDataFree(searchList);
                goto read_it;
            }
        }
        ElfDataFree(searchList);
ferr:   TokErr(XLT("Cannot find %s"),iname);
        ElfTokArray[0].simple->type = TKT_EMPTY;
        return;
    }
    else if ( type == TK_SEARCH ) {
        ;
    }
}
}

/*************************************************************************
function: ElfcTokenizeLine - Read, tokenize, next input line
*/
ElfcTokenizeLine(LineBuf,TokArray,num_toks)
char *LineBuf;
tok_ptr TokArray[];
int *num_toks;
{
#define TOK_ERR(body) {TokErr body; goto token_no_good;}

    register char *p, *pl,*pk;
    register int ix;
    register int NumToks = 0;
    register char *TokP = NULL, *EndP = NULL;
    int small;

    TokP = LineBuf;
    EndP = TokP + strlen(LineBuf);

get_again:

    TokStarts = TokP;
    p = TokP++;

    /* see if have finished this line */

    if (*p == 0)
        goto done_line;
    if ( NumToks > MAX_TOKENS-4 ) TOK_ERR((XLT("line too long")));

    if ( DoingComment ) {
        if ( (*p == '*') && (*(p+1) == '/') ) {
            DoingComment = FALSE;
            TokP++;
        }
        goto get_again;
    }

    /* ignore spaces, tabs, embedded newlines */
    /* And hard (non-breaking) spaces: 0xa0 in Applixware text strings */

    if ( (*p == ' ') || (*p == '\t') || (*p == '\n') || (*p == (char)0xa0) )
        goto get_again;

    /* check for C style comments */

    if ( (*p == '/') && (*(p+1) == '*') ) {
        DoingComment = TRUE;
        TokP++;
        goto get_again;
    }

    /* check for Basic style comments */

    if ( *p == '\'' ) {
        TokP = EndP;
        goto get_again;
    }

    /* special case .123 so . not regarded separator */

    if ( (*p == '.') && isdigit(*(p+1)) )
        goto not_op;

    /* check for operator/punctuation */

    for ( pl = ElfOpChars, ix = 0; *pl; ix++ ) {
        register int toklen;
        pk = p;
        for ( toklen = 0; *pl; toklen++ ) {
            if ( *pl++ != *pk++ ) {
                while ( *pl++ );
                goto failed;
            }
        }
        TokP += toklen - 1;
        TokArray[NumToks].simple = (tok_simple*)ElfcAlloc(sizeof(tok_simple));
        TokArray[NumToks].simple->type = TKT_SIMPLE;
        TokArray[NumToks++].simple->code = ElfOpTokens[ix];
        goto get_again;
failed:;
    }
not_op:

    /* check for string literal */

    if ( *p == '"' ) {
        char *p, *start, *dst;
        char buf[5003];     /* matches a buffer in AddStrCont() */
        int len = 0;
 
        start = TokP-1;
        dst = buf;
        for ( p = start + 1; p != EndP; p++ ) {
            if ( ++len > 5000 )     /* matches a buffer in AddStrCont() */
                TOK_ERR((XLT("string too long")))
            if ( *p == '\\' ) {
                p++;
                if ( *p == 'n' )
                    *dst++ = '\n';
                else if ( *p == 't' )
                    *dst++ = '\t';
                else if ( *p == '(' ) {
                    char *px;
                    char nbuf[10];
                    px = strchr(p,')');
                    if ( !px || (px - p > 4) )
                        goto slash_punt;
                    p++;
                    memcpy(nbuf,p,px-p);
                    nbuf[px-p] = 0;
                    *dst++ = atoi(nbuf);
                    p = px;
                }
                else if ( (*p == '\\') || (*p == '"') )
                    *dst++ = *p;
                else {
slash_punt:         *dst++ = '\\';
                    *dst++ = *p;
                }
            }
            else if ( *p == '\n' )
                TOK_ERR((XLT("missing end quote")))
            else if ( *p == '"' ) {
                *dst = 0;
                TokArray[NumToks].str = 
                    (tok_str*)ElfcAlloc(sizeof(tok_str)+strlen(buf));
                TokArray[NumToks].str->type = TKT_STRING;
                strcpy(TokArray[NumToks++].str->str,buf);
                TokP = p + 1;
                goto get_again;
            }
            else
                *dst++ = *p;
        }
        TOK_ERR((XLT("missing end quote")))
    }

    /* isolate token into a buffer */

    else {
        char *start, *end;
        int type;
        char buf[ELFIDSTR+1];
        register char *p, *dst;
        double value;
 
        start = TokP-1;
        for ( end = start+1; end != EndP; end++ ) {
#if BUILD_ORIENT
            /* Ashisuto add to correspond 8bit chars.   95.10.06 */
            extern int  ElfcAllowSQL;
            if (ElfcAllowSQL && (*end & 0x80))
                continue;
#endif
            if ( IsElfTokSep[((unsigned)(*end)) & 0xFF] )
                break;
        }
 
        for ( p = start, dst = buf; p < end; p++,dst++ ) {
            if ( dst-buf >= ELFIDSTR ) {
                 *dst = 0;
                 TOK_ERR((XLT("token too long %s"),buf));
            }
            *dst = *p;
        }
        *dst = 0;
        TokP = end;

        if ( buf[0] == '&' && (strlen(buf) > 2)) {  /* hex/octal/binary */
            p = start;
        }

        /* Detect literal numbers */

        if ( isdigit(buf[0]) || ((buf[0] == '.') && isdigit(buf[1])) ) {
            register int got_e = FALSE, got_dot = FALSE;
            int xval;

            /* Detect HEX and OCTAL literal numbers */

            if ( (buf[0] == '0') && strchr("xdo",buf[1]) ) {
                for ( p = start, dst = buf; p < EndP; p++, dst++ ) {
                    if ( IsElfTokSep[((unsigned)(*p)) & 0xFF] )
                        break;
                }
                *dst = 0;
                switch ( buf[1] ) {
                case 'x':
                    if ( 1 == sscanf(buf+2,"%x",&xval) ) {
                        value = xval;
                        goto got_num;
                    }
                    break;
                case 'o':
                    if ( 1 == sscanf(buf+2,"%o",&xval) ) {
                        value = xval;
                        goto got_num;
                    }
                    break;
                }
            }

            /* produce a token for ParseNum() */

            for ( p = start, dst = buf; p < EndP; p++, dst++ ) {
                *dst = *p;
                if ( (*p == 'e') || (*p == 'E') ) {
                    if ( got_e )
                        goto num_no_good;
                    else
                        got_e = TRUE;
                }
				else if (*p == '.') {
					if (got_dot || got_e) {
						goto num_no_good;
					}
					else
						got_dot = TRUE;
				}
                else if ( ((*p == '+') || (*p == '-')) && (got_e == 1) )
                    got_e = 2;
                else if (IsElfTokSep[((unsigned)(*p)) & 0xFF] )
                    break;
                else if ( !strchr("0123456789",*p) )
                    goto num_no_good;
            }
            TokP = p;   /* override prior toke-bounds definition (for '.') */
            *dst = 0;
            value = atof(buf);
got_num:    if ( C_DBL_IS_ELF_INT(value,small) && !strchr(buf,'.') ) {
                TokArray[NumToks].small = (tok_integer*)
                                ElfcAlloc(sizeof(tok_integer));
                TokArray[NumToks].small->type = TKT_INTEGER;
                TokArray[NumToks++].small->val = small;
            }
            else {
                TokArray[NumToks].num = (tok_number*)
                                     ElfcAlloc(sizeof(tok_number));
                TokArray[NumToks].num->type = TKT_NUMBER;
                TokArray[NumToks++].num->val = value;
            }
            goto get_again;
num_no_good: 
			while(p < EndP && (*p == '.' || 
				  !IsElfTokSep[((unsigned)(*p)) & 0xFF]))
				*dst++ = *p++;
			*dst = 0;       /* fall thru */
        }

        for ( p = buf; *p; p++ )        /* convert token to upcase */
            {
#if BUILD_ORIENT
            /* Ashisuto added to correspond 8bit chars.   95.10.06 */
            if (!(*p&0x80))
#endif
            *p = toupper(*p);
            }

        if ( streq(buf,XLT("TIME@@@")) ) {
            TokArray[NumToks].num = (tok_number*)
                ElfcAlloc(sizeof(tok_number));
            TokArray[NumToks].num->type = TKT_NUMBER;
            TokArray[NumToks++].num->val = (double)time(0L);
            goto get_again;
        }
		else if ( streq(buf,NXLT("FILE@@@")) ) {
			p = ElfcSrcFileName;
			if ( !p )
				p = "???";
            TokArray[NumToks].str = 
                    (tok_str*)ElfcAlloc(sizeof(tok_str)+strlen(p));
            TokArray[NumToks].str->type = TKT_STRING;
            strcpy(TokArray[NumToks++].str->str,p);
            goto get_again;
		}

        if ( (buf[0] == '#') && (type = FindKeyword(buf)) ) {
do_key:     TokArray[NumToks].simple = (tok_simple*)ElfcAlloc(sizeof(tok_simple));
            TokArray[NumToks].simple->type = TKT_SIMPLE;
            TokArray[NumToks++].simple->code = type;
            if ( (type == TK_UIPROC) || (type == TK_NRPROC) ||
                 (type == TK_PROC) || (type == TK_FUNC) ||
                 (type == TK_OLDSET) || (type == TK_OLDGET) ||
                 (type == TK_SET) || (type == TK_GET) )
                ElfcLastFuncKeyAtLine = LineCount;
            goto get_again;
        }
#if BUILD_ORIENT
        /* Ashisuto added to allow multi-byte field name in data-base tables.
         * the function IsOrientalSQLIdentifier() is defined in elstrs.c
         *  95.10.06
         */
        if (IsIdentifier(buf) || IsOrientalSQLIdentifier(buf)) 
#else
        if ( IsIdentifier(buf) ) 
#endif
            {
            if ( type = FindKeyword(buf) )
                goto do_key;
            TokArray[NumToks].id = (tok_id*)ElfcAlloc(sizeof(tok_id)+strlen(buf));
            TokArray[NumToks].id->type = TKT_ID;
            strcpy(TokArray[NumToks++].id->name,buf);
            goto get_again;
        }
        TOK_ERR((XLT("illegal token - %s"),buf))
    }

token_no_good:
    TokArray[NumToks].simple = (tok_simple*)ElfcAlloc(sizeof(tok_simple));
    TokArray[NumToks].simple->type = TKT_SIMPLE;
    TokArray[NumToks++].simple->code = TK_LEXERROR;

done_line:
    *num_toks = NumToks;
    return(0);
}


/*
 * function:  WindToEndOfLine - Advane lexor to end of line
*/
WindToEndOfLine ()
{
    TokPos = 0xFFFF;
}

/****************************************************************************
function:  yyerror - Error messager
?? should be able to describe token syntax error detected at
*/
int yyerror(a, va_alist)
char *a;
va_dcl
{
    va_list args;

    va_start(args);
    vyyerror(a, args);
    va_end(args);
}
static int vyyerror (a, args)
char *a;
va_list args;
{
    char buf[256], buf2[256];
    data_id *GetSymP();

    if ( streq(a,NXLT("syntax error")) ) {
        a = XLT("syntax error");
    } else if ( streq(a,NXLT("yacc stack overflow")) ) {
        a = XLT("code block too large/complex");
    }
    vsprintf(buf,a,args);
    buf2[0] = 0;
    switch ( TokType ) {
    case TK_STRVAL:
    {
        register data_comp_string *symP;
        symP = (data_comp_string*)GetSymP(TokVal);
        strncpy(buf2,symP->text,50);
        buf2[50] = 0;
    }
    CASE TK_NUMVAL:
    {
        register data_comp_number *symP;
        symP = (data_comp_number*)GetSymP(TokVal);
        if ( symP->type == TYPE_INTEGER ) {
            register data_comp_integer *isymP;
            isymP = (data_comp_integer*)symP;
            sprintf(buf2,"%d",isymP->value);
        } else sprintf(buf2,"%g",symP->value);
    }
    CASE TK_ID:
        strcpy(buf2,GetSymP(TokVal)->name);
    DEFAULT:
    {
        int ix;
        char *p;
        char *ElfKeywordFromTokType();
            p = ElfKeywordFromTokType(TokType);
        if ( p )
            strcpy(buf2,p);
        p = ElfOpChars;
        for ( ix = 0; ix < sizeof(ElfOpTokens)/sizeof(int); ix++ ) {
            if ( ElfOpTokens[ix] == TokType )
            strcpy(buf2,p);
            p += strlen(p)+1;
        }
    }
    }   /* end of switch */
    if ( buf2[0] != 0 )
        sprintf(buf+strlen(buf)," at or near %s",buf2);
    err_msg(LineCount,"%s",buf);
}

/****************************************************************************
function:   TokErr - Error messager
*/
int TokErr(str, va_alist)
char *str;
va_dcl
{
    va_list args;

    va_start(args);
    vTokErr(str, args);
    va_end(args);
}
static int vTokErr(str, args)
char *str;
va_list args;
{
    verr_msg(LineCount,str,args);
}

/*************************************************************************
function: ElfcStrAlloc - Alloc, strcpy & return ptr
*/
char *ElfcStrAlloc (iname)
register char *iname;
{
    register char *p;
    char *strcpy();

    return(strcpy(p = ElfcAlloc(strlen(iname)+1),iname));
}

