/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: elmain.c
   Applix ASC version: 27
   Date: Wed Jun 16 21:11:26 1999
*/
#include <errno.h>

#include <varargs.h>
#include <elcomp.h>
#include <elf_strs.h>
#include <fsrts.h>
#include <getstr10.h>
#include <elf_env.h>
#include <elmain.p>

data_ptr ElfCompSrcArrayP;	/* passed onto lexer */
axstatic jmp_buf aborter;        /* general aborter */
axstatic int Aborting;   /* safety check to eliminate abort recursion */
axglobal int ElfcNumErrs = 0;        /* error counter */
extern int ElfAllowAbort;
axstatic int (*ElfPostError)(); /* callback functions */
data_ptr ElfCompErrorsP;	/* array to write errors to */
char *ElfcSrcFileName;		/* Needed for lexer's FILE@@@ function */
comp_loop_info CG;	/* code generator info (only one compilation at a time) */
int ElfInitedStrs;	/* flag saying whether ElfInitStrs has run */
char *ElfKeyWords;      /* gotten from string file ELFKx */
char *ElfRsvdWords;     /* gotten from string file ELFRx */
char *ElfSQLWords;		/* SQL Key words */
char *ElfOracleWords;		/* SQL Key words */
char *ElfObjectWords;	/* Object mode key words */
char *ElfOpChars;   /* gotten from string file ELFOPCHARS */
/* ". ! , : ( ) [ ] += -= >= <= <> = ^ - * / \\ >>= <<= >> << ++ > < + { }  "; */

int ElfQsrchCalled;	/* used by elf_lex */
int ElfLanguage = 1;		/* language for symbol table lookups */
char **ElfProtected;    /* Elf names accesable only is sys mode */
int ElfNumProtected;
extern char *ThimUserElfDir;

/*************************************************************************
function: ElfCompileArray - Compile passed array of source strs
If the return value[0] is not an array, retval is error strings
*/
data_ptr ElfCompileArray (srcname,srcP)
char *srcname;	/* for debugger access */
data_ptr srcP;	/* IN, ptr to NULL termed array of source lines */
{
	data_ptr	ElfDoCompileArray();

	ElfCompErrorsP = MkArray(0);
	return(ElfDoCompileArray(srcname,srcP));
}

data_ptr ElfDoCompileArray (srcname,srcP)
char *srcname;	/* for debugger access */
data_ptr srcP;	/* IN, ptr to NULL termed array of source lines */
{
    char *DoCompile();
    extern int LineCount;
	data_ptr dP;
	data_ptr WriteCodeArray();

	ElfcSrcFileName = srcname;
    ElfInitStrs();
    ElfcNumErrs = 0;
    LineCount = 0;
	Aborting = FALSE;
	ElfCompSrcArrayP = srcP;
    if ( setjmp(aborter) ) {
		ElfcSrcFileName = NULL;
        if ( !Aborting )
            ErrOut(XLT("compilation terminated"));
        MemClean();
        return(ElfCompErrorsP);
    }
    InitCompiler();
    while ( 1 == yyparse() )
        WindToEndOfLine();
	ElfcSrcFileName = NULL;
    if ( ElfcNumErrs ) {
        LOGGER(("***total errors = %d",ElfcNumErrs));
        MemClean();
        return(ElfCompErrorsP);
    }
	dP = WriteCodeArray(srcname);
    MemClean();
	ElfDataFree(ElfCompErrorsP);
	return(dP);
}

/*************************************************************************
function:  Logger - Perform logging
*/
int Logger(fmt, va_alist)
char *fmt;
va_dcl
{
	va_list args;

	va_start(args);
	vLogger(fmt, args);
	va_end(args);
}

static int vLogger(fmt, args)
char *fmt;
va_list args;
{
	static FILE *LogFile = NULL;

	if (!LogFile)
		LogFile = fopen("zzelfc.log","w");

	if (*fmt)
		vfprintf(LogFile, fmt, args);
	else fprintf(LogFile, "Logger: internal error!");
	fprintf(LogFile,"\n");
	fflush(LogFile);
}

/************* ERROR SUPPORT **************************************/

/*************************************************************************
function: ElfCompAddrs - Set callback ptrs from app
*/
ElfCompAddrs (poster)
int (*poster)();
{
	ElfPostError = poster;
}

/*************************************************************************
function: CodeError - Send a code generator error message
 */
int CodeError(str, va_alist)
char *str;
va_dcl
{
	va_list args;

	va_start(args);
	vCodeError(str, args);
	va_end(args);
}
static int vCodeError(str, args)
char *str;
va_list args;
{
	extern int CGcurLine;	   /* current line code generator working on */
	verr_msg(CGcurLine,str,args);
}


/****************************************************************************
function:  err_msg - Send an error message
*/
int err_msg(line, fmt, va_alist)
int line;
char *fmt;
va_dcl
{
	va_list args;

	va_start(args);
	verr_msg(line, fmt, args);
	va_end(args);
}

#if defined(__GNUC__) && defined(__PPC__)
#define VA_ASSIGN(VA1,VA2) memcpy((VA1),(VA2),sizeof(va_list))
#else
#define VA_ASSIGN(VA1,VA2) (VA1) = (VA2)
#endif

int verr_msg(line, fmt, args)
int line;
char *fmt;
va_list args;
{
	char buf[400];
	char buf2[400];
	char **errP;
	va_list argsl;

#if (LINUX && __powerpc__)
	memcpy(argsl,args,sizeof(va_list));
#else
	argsl = args;
#endif

	VLOGGER((fmt, argsl));

	ElfcNumErrs++;
	if ( ElfCompErrorsP ) {
		if (*fmt)
			vsprintf(buf,fmt,args);
		else sprintf(buf, "err_msg: internal error!\n");
		sprintf(buf2,"%d: %s",line+1,buf);
		ElfCompErrorsP = ElfAddStrToArray(ElfCompErrorsP,ElfcNumErrs-1,buf2);
	}
	else (*ElfPostError)(line, fmt, args);

	if ( !Aborting && (ElfcNumErrs >= MAX_ERRS) ) {
		ErrOut(XLT("too many errors, compilation aborted"));
	}
}

/****************************************************************************
 * function:  ErrOut - Error out of compilation
*/
ErrOut (str)
char *str;
{
    Aborting = TRUE;
    err_msg(MAXINT,str);
    if ( ElfAllowAbort )
		exit(1);
    longjmp(aborter,10);
}

/*************************************************************************
function: Asserter - Assert
*/
Asserter (file,line)
char *file;
int line;
{
    Aborting = TRUE;
    err_msg(MAXINT,XLT("assertion failed %s line %d"),file,line);
    if ( ElfAllowAbort )
		exit(1);
    longjmp(aborter,10);
}

/************* MEMORY MANAGER ******************************************/
/*
   ?? This scheme doesn't free anything while the compiler is running.  All Frees occur when the compiler finishes/aborts.  This is not great since it allows a very high peak of memory usage.
*/

axstatic char **AlcList;  /* head of allocation list */

/*************************************************************************
function: ElfcAlloc - Allocate some memory, add to AlcList
Allocates excess to include AlcList thread before the returned ptr.
Always allocates to an 8-byte boundary.
*/
char *ElfcAlloc(size)
unsigned size;
{
    int size_8byte;
    char **p;

    size_8byte = size + 8;
    p = (char**)calloc(1,size_8byte);
    *p = (char*)AlcList;
    AlcList = p;
    return((char*)p + 8 );

}
/*************************************************************************
function: MemClean - Release all allocate memory
*/
MemClean()
{
    register char **p, **next;
    extern FILE* FpInclude;
	extern int AxElfrtMode;

    for ( p = AlcList; p; p = next ) {
        next = (char**)*p;
        free(p);
    }
    AlcList = NULL;
    if ( FpInclude ) {
#if FOR_TESTER || TRUE	/* !!0 Issue with R2 macro editor header strs!!! */
        fclose(FpInclude);
#else
		CloseFileObject(FpInclude);
		FSClsDoc(&ElfcIncDocInfo,FALSE);
#endif
		FpInclude = NULL;
	}
}

/*************************************************************************
function: ElfSetStandardSearchList - User/Cust/Sys dirs
Added directories with language-dependent files to the search path. - lhunt
*/
ElfSetStandardSearchList ()
{
#if !FOR_TESTER
	char path[STRPATH+1];
	char *dir, slash;
	char *UPneedByCode();
	extern char *ThimApplixDir;		/* Where the THIMBLE executable lies */
	extern char *AxLangDir;			/* Where lang-dependent THIMBLE files lie */
	extern char *ThimCustomDir;		/* Where customizeable files are kept */
	extern char *AxLocalLangDir;	/* Where lang-depend. customizeables lie */
	extern char *AxLangDemosDir;	/* Where lang-depend. sample macros lie */
	data_ptr searchList;

	searchList = MkArray(0);
#if AXMS
	slash = '\\';
#else
	slash = '/';
#endif
	/* Look at ThimUserElfDir instead of reading the profile value
	   PROF_CMDS_DIR_.  The user's command directory is
	   ThimUserElfDir, which is set in `SetThimbleDirs ()'. */
	/* dir = UPneedByCode(PROF_CMDS_DIR_); */
	dir = ThimUserElfDir;

	if ( dir[0] )
		searchList = ElfAddStrToArray(searchList,ElfbArraySize(searchList),dir);

	/*	After clearing axlocal for customer only, search path needs to include:
	 *	ThimCustomDir - to point to axlocal
	 *	AxLocalLangDir - to point to any customized lang dirs in axlocal
	 *	AxLangDemosDir - point to the moved sample macros directory
	 */
	sprintf(path,"%s",ThimCustomDir);	/*axlocal*/
	searchList = ElfAddStrToArray(searchList,ElfbArraySize(searchList),path);

	sprintf(path,"%s",AxLocalLangDir);	/*	axlocal + lang */
	searchList = ElfAddStrToArray(searchList,ElfbArraySize(searchList),path);

	sprintf(path,"%s",AxLangDemosDir, slash);	/*	sample macros dir */
	searchList = ElfAddStrToArray(searchList,ElfbArraySize(searchList),path);

								/*	Should never find anything here, but - */
	sprintf(path,"%s",AxLangDir);	/*	ThimApplixDir+lang-dir */
	searchList = ElfAddStrToArray(searchList,ElfbArraySize(searchList),path);

	sprintf(path,"%s%celf",ThimApplixDir, slash);	/* ThimApplixDir + elf */
	searchList = ElfAddStrToArray(searchList,ElfbArraySize(searchList),path);

	sprintf(path,"%s%cprojects",ThimApplixDir, slash);	/* " + projects */
	searchList = ElfAddStrToArray(searchList,ElfbArraySize(searchList),path);

	ElfbSetSysVar("elf_search_list@",searchList);
#endif
}


