/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: elmsgs.c
   Applix ASC version: 7
   Date: Tue Feb  2 15:21:26 1999
*/
#include <aos.h>
#include <thimble.h>
#include <getstr4.h>
#include <paripc.h>
#include <scac.h>
#include <eldata.h>
#include <elmsgs.p>

/* ??x This runtime should be part of linkage logic */

#define MAX_ELF_MES 900					/* must be <= MAX_ELF_DATA */
#define streq(a,b) (!strcmp(a,b))

extern short THIMpid;


/*************************************************************************
function: ElfReadIpc - Read the IPC message into memory
*/
ElfReadIpc (int absorbeMessage)
{
	long size, senderType, senderTask;
	long ljunk;
	int code;
	char *msgP;

	TaskInfo[THIMpid].ElfRetCode = 1;	/* paranoid - assumption of success */
	code = AxGetMessage(-1,&msgP,&size,&senderType,
					&senderTask, &ljunk); /* will always succeed */
	if ( ((senderType != APELFM) && (senderType != APELFASYNC))
				|| (code != IPC_OK) ) {
		char buf[1000];
		sprintf(buf,"bad message code %d sender %d",code,senderType);
		taskAbort(buf);
	}
	THIMfree(TaskInfo[THIMpid].ElfIpcMessage);
	TaskInfo[THIMpid].ElfIpcMessage = TaskAlloc(THIMpid,size);
	TaskInfo[THIMpid].ElfIpcElfArray = NULL;	/* in case prior owner crashed before responding */
	memcpy(TaskInfo[THIMpid].ElfIpcMessage,msgP,size);
	TaskInfo[THIMpid].ElfSenderPid = senderTask;
	if ( absorbeMessage )
		AxMessageAbsorbed();
    return(*((int*)TaskInfo[THIMpid].ElfIpcMessage));	/* code is first int */
}

/*************************************************************************
function: ElfNoteIpc - Note the message into TaskInfo[THIMpid].ElfIpcMessage
*/
ElfNoteIpc (len,sender,type)
short *len,*sender,*type;
{
	abort();
}

/*************************************************************************
function: ElfGetMsg - Extract & return elements of a message
This function assumes that all pointers are of the same size in memory.
It is machine specific based on how the stack is packed.
Assumes data storage allowed on int boundaries for all datatypes.
*/
typeErr ElfGetMsg (va_alist)
va_dcl /* where to write results to */
{
    va_list ptr;
    register char *format;   /* format string describing decomposition method */
    register char *dataP;
    register char *p;
    register int argdex = 0;
	data_ptr aP;

	va_start(ptr);
	format = va_arg(ptr,char *);

	TaskInfo[THIMpid].ElfRetCode = 1;	/* paranoid - assumption of success */
	TaskInfo[THIMpid].ElfIpcElfArray = aP = AxMemRead(TaskInfo[THIMpid].ElfIpcMessage+sizeof(int));
    for ( p = format; *p; p++, argdex++ ) {
        switch ( *p ) {
        case 'I':
			*(va_arg(ptr,int *)) = AxIntFromArray(aP,argdex);
            break;
        case 'N':
			*(va_arg(ptr,double *)) = AxDblFromArray(aP,argdex);
            dataP += sizeof(double);
            break;
        case 'S':
			*(va_arg(ptr,char **)) = AxStrPtrFromArray(aP,argdex);
            break;
        case 'L':
			*(va_arg(ptr,long *))  = AxIntFromArray(aP,argdex);
            break;
        default:
			taskAbort("ElfGetMsg: unknown datum type");
        }
    }
    return(AOK);
}

/*************************************************************************
function: ElfHasMsgData - Return TRUE/FALSE
*/
int ElfHasMsgData ()
{
	return(TaskInfo[THIMpid].ElfIpcMessage ? TRUE : FALSE);
}

/*************************************************************************
function: ElfGetMsgData - Return data array
*/
data_ptr ElfGetMsgData ()
{
	TaskInfo[THIMpid].ElfRetCode = 1;	/* paranoid - assumption of success */
	return(AxMemRead(TaskInfo[THIMpid].ElfIpcMessage+sizeof(int)));
}

/*************************************************************************
function: ElfRespond - Send and Elf ipc response
This function assumes that all pointers are of the same size in memory.
It is machine specific based on how the stack is packed.
Assumes data storage allowed on int boundaries for all datatypes.
*/
LINTCOMMENT(VARARGS2)
typeErr ElfRespond(va_alist)
va_dcl /* where to write results to */
{
    va_list ptr;
    register char *format;   /* format string describing decomposition mode */
    register char *p;
    register int dex = 0;
    register int size;
	char *msg;
	data_ptr aP;

	if ( !TaskInfo[THIMpid].ElfSenderPid || (TaskInfo[THIMpid].ElfSenderPid == THIMpid) )
		return;

	va_start(ptr);
    format = va_arg(ptr,char *);

	aP = AxMakeArray(strlen(format));

/* copy in the data */

	for ( p = format; *p; p++, dex++ ) {
        switch ( *p ) {
        case 'I':
            AxAddIntToArray(aP,dex,*(va_arg(ptr,int*)));
            break;
        case 'N':
            AxAddDblToArray(aP,dex,*(va_arg(ptr,double*)));
            break;
        case 'S':
			AxAddStrToArray(aP,dex,*(va_arg(ptr,char**)));
            break;
        case 'C':
			AxAddStrToArray(aP,dex,va_arg(ptr,char*));
            break;
        case 'L': {
			long l;
			l = *(va_arg(ptr,long*));
			if ( l == (l & 0x3FFFFFFF) )
	            AxAddIntToArray(aP,dex,l);
			else
	            AxAddDblToArray(aP,dex,(double)l);
            break;
	        }
		default:
			abort();
		}
    }
	size = AxMemWrite(aP,sizeof(int),&msg);		/* pack into a single buffer */
	memcpy(msg,&TaskInfo[THIMpid].ElfRetCode,sizeof(int));
	IPCsendMes(IPC_RES,TaskInfo[THIMpid].ElfSenderPid,msg,size);
	AxFreeData(aP);
	if ( TaskInfo[THIMpid].ElfIpcElfArray ) {
		AxFreeData(TaskInfo[THIMpid].ElfIpcElfArray);
		TaskInfo[THIMpid].ElfIpcElfArray = NULL;
	}
	THIMfree(msg);
    return(AOK);
}

/*************************************************************************
function: ElfReturnData - Return array of info
*/
typeErr ElfReturnData (aP)
data_ptr aP;
{
	int size;
	char *msg;

	if ( !TaskInfo[THIMpid].ElfSenderPid || (TaskInfo[THIMpid].ElfSenderPid == THIMpid) ) {
		AxFreeData(aP);
		return;
	}
	size = AxMemWrite(aP,sizeof(int),&msg);		/* pack into a single buffer */
	memcpy(msg,&TaskInfo[THIMpid].ElfRetCode,sizeof(int));
	IPCsendMes(IPC_RES,TaskInfo[THIMpid].ElfSenderPid,msg,size);
	AxFreeData(aP);
	THIMfree(msg);
	if ( TaskInfo[THIMpid].ElfIpcElfArray ) {
		AxFreeData(TaskInfo[THIMpid].ElfIpcElfArray);
		TaskInfo[THIMpid].ElfIpcElfArray = NULL;
	}
    return(AOK);
}

/*************************************************************************
function: ElfRetErr - Return an error code to Elf
*/
typeErr ElfRetErr (code)
typeErr code;
{
	int intcode = code;

	TaskInfo[THIMpid].ElfRetCode = 0;	/* failure flag */
	ElfRespond("ICC",&intcode,"","");
	TaskInfo[THIMpid].ElfRetCode = 1;
	return(AOK);
}

/*************************************************************************
function: ElfRet - Return error or OK ipc to Elf
*/
typeErr ElfRet (code)
register typeErr code;
{
    if ( code == AOK )
        return(ElfRespond(""));
    else
        return(ElfRetErr(code));
}

/*************************************************************************
function: ElfRetStrErr - Return error & override string & error object
*/
typeErr ElfRetStrErr (code,str,object)
int code;
char *str;
char *object;
{
	int intcode = code;

	TaskInfo[THIMpid].ElfRetCode = 0;	/* failure flag */
	if ( !str ) str = "";
	if ( !object ) object = ""; 
	ElfRespond("ICC",&intcode,str,object);
	TaskInfo[THIMpid].ElfRetCode = 1;
	return(AOK);
}

