/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: elobj.c
   Applix ASC version: 142
   Date: Tue Feb  2 15:20:41 1999
*/
#include <sys/types.h>
#include <axsocket.h>
#include <stdio.h>
#include <aos.h>
#include <elf_env.h>
#include <getstr4.h>
#include <objinfo.h>
#include <thimble.h>
#include <adlgpoke.h>

#define ENABLE_CORBA 1
 
#if ENABLE_CORBA
#define USE_DATA_PTR
#include <bldrextint.hh>
#undef USE_DATA_PTR
#endif


#if BUILD_ORIENT
#include <wchars.h>
#endif

#if HP800V3		/* What a lame Hack! need this in aos.h instead !!! */
#define ACONST
#else
#define ACONST const
#endif

extern elf_obj_grp *ElfFindObjChangeMethod();
extern elf_obj_grp *ElfFindObjQueryMethod();
extern elf_obj_grp *ElfFindObjLocalData();
extern elf_obj_grp *ElfVfyObjLocalData();
extern elf_obj_grp *ElfObjData();
extern char *elfHashNameAndVal();
extern char *elfHashName();
extern package_info *ElfLoadPackage();
extern elf_object *ElfReadObjectDataFile();
extern void ElfFleshOutMethodHash();
static void ElfUnFleshOutMethodHash();
static elf_object *elfCreateObject();
static elf_obj_grp *getLocalData();
static elf_obj_grp *crtLocalData();
static void elfCreateCppObject();
static void elfDestroyCppObject();
static data_ptr elfCreateObjectFromArray();

static void objAddClassToAppModels();
static data_ptr objectStructureFromObject();
static data_ptr objGetObjvar(data_ptr, char*);
static data_ptr objChildren();
static data_ptr objLocalStaticDatums();
static data_ptr objInheritFrom();
static data_ptr elfGetCppClassNames();
static data_ptr elfGetAllCppClassObjects();
static data_ptr elfObjGetCppFuncsList();
static data_ptr elfObjGetCppHelpStr();
static data_ptr findChildrenInheritingFrom();
static data_ptr findParentInheritingFrom();
static void elfCreateCorbaObject(data_ptr, data_ptr);
static void elfCreateCorbaClass(data_ptr, data_ptr);
static void elfSetCorbaInvocationMode(data_ptr, data_ptr);
static data_ptr elfGetDeferredRequestResult(data_ptr, data_ptr);
static data_ptr elfHelpOnExtObjMethod(data_ptr, data_ptr, data_ptr);
static data_ptr elfExtStructsAsFormats(data_ptr, data_ptr, data_ptr);
static void elfExtAppTerminated(data_ptr);
static void elfRegisterObjectForCallbacks(data_ptr, data_ptr, data_ptr);
static int elfObjectType(data_ptr);
static void AddObjectToDlgMap();
static char *compilationErrors(data_ptr);

/* alternate external object model */
static void elfCreateExternalObject(data_ptr,data_ptr,data_ptr);

extern char *AxTaskPoolAllocate();

extern void AxDestroyCppObject(void *);
extern void *AxCreateCppObject(void *);
extern data_ptr AxCallCppMemberFunc();
extern char **AxQueryCppInterface();
extern char **AxQueryCppInterfaceHelpStrs();
extern char *TrueMalloc();

#define OBJECT_HASH_PRIME	1777	/* a prime number */
#define OBJVAR_HASHTABLE_SIZE (ELF_HASH_PRIME+ELF_HASH_PRIME)

#define QUICK_STR_FROM_ARRAY(aP,dex)	(SX(AX((aP))->vals[dex])->text)
#define TRUEALCSTR(s)  strcpy(TrueMalloc(strlen(s)+1),s)

#define OBJ_DEXIX(d,x)	((((x) & 0xFFFF) << 16) | (d))
#define OBJ_DEX(dx)		((dx) & 0xFFFF)
#define OBJ_IX(dx)		(((dx) >> 16) & 0xFFFF)

#define OBJVAR_TYPE_STATIC		(0x01)
#define OBJVAR_TYPE_PUBLIC		(0x02)
#define OBJVAR_TYPE_PRIVATE		(0x04)
#define OBJVAR_TYPE_PROTECTED	(0x08)

#define OBJVAR_TYPE_IS_STATIC(f)		((f) & OBJVAR_TYPE_STATIC)
#define OBJVAR_TYPE_IS_PUBLIC(f)		((f) & OBJVAR_TYPE_PUBLIC)
#define OBJVAR_TYPE_IS_PRIVATE(f)		((f) & OBJVAR_TYPE_PRIVATE)
#define OBJVAR_TYPE_IS_PROTECTED(f)		((f) & OBJVAR_TYPE_PROTECTED)

#define IS_APP_IN_CONSTRUCTION(o)		((o)->isTopLevelApp == 2)
#define IS_APP_IN_READ(o)				((o)->isTopLevelApp == 3)
#define IS_CLASS_IN_READ(o)				((o)->isTopLevelApp == 4)


extern data_ptr NullDataPtr;
extern char *DirSlash;
		
typedef struct object_hashtbl_entry_t {
	elf_object **object_table;
	int table_size;
} object_hashtbl_entry_t;

/* private data */

/* Mapping table to allow for caching of dialogs in Anyware client
 When a dialog is displayed, the builder-side widget objects are
 mapped to simple numeric widget ids in the dialog and a per-instance
 unique dialog box id.  This way, object tuples don't need to go to/from
 java client - this makes cached dialogs possible.
*/
typedef struct {
	elf_object *oP;
	int dlgId;		/* uid of dialog during this session - changes each session */
	int widgetId;	/* id of widget within dialog - constant for all sessions */
} javaDlgMap_t;
static javaDlgMap_t *javaDlgMap;	/* alloced array of entries */
static int javaDlgNumMapped;	/* how many entries in table */
static int javaDlgNumAlloced;	/* how many entries in table */
static int javaDlgMapHaveSlot;	/* if TRUE, at least one slot has been freed */

static elf_object **AppObjTable;	/* Table of pointers to Application objects */
static int SizeAppObjTable = 0;		/* size of Application object table */
static elf_object *ApplixBuilderApp;	/* arcade application object */

static elf_object **CppClassTable;		/* Table of C++ class info objects */
static int SizeCppClassTable = 0;		/* size of C++ class info objects table */

static int UidCounter;				/* for object creation */
/* table of objects */
static object_hashtbl_entry_t ObjectHashTable[OBJECT_HASH_PRIME+1];

static data_object *elfUnsnappedObjRefs;	/* head of list */

#include <elobj.p>

/***************************************************************************
function: objHash - fast and reliable hash function (only additions and subtractions)
*/
static int
objHash (name)
register char *name;
{
	register int h = 0;
	
	if (!name)
		return(0);
	while (*name) {
		h = h + h + (((int)*name) & 0xFF);
		while (h >= OBJECT_HASH_PRIME)
			h -= OBJECT_HASH_PRIME;
		name++;
	}
	return(h);
}

/***************************************************************************
function: getObjectFromTable - returns address of elf_object in object hash table
*/
static elf_object *
getObjectFromTable (dex_ix)
int dex_ix;
{
	int dex, ix;
	
	dex = OBJ_DEX(dex_ix);
	ix = OBJ_IX(dex_ix);
	if (ObjectHashTable[dex].table_size <= ix)
		return(NULL);
	return(ObjectHashTable[dex].object_table[ix]);
}

/*************************************************************************
function: addObjectToHashTable - Find a slot, extend bin-table if needed
This could be done a lot faster by keeping a freelist, but there seems
to be no major need... at this time.
*/
static int
addObjectToHashTable (oP)
elf_object *oP;
{
	register int ix;
	register elf_object **table;
	int dex, size;

	dex = objHash(oP->nameTag);
	table = ObjectHashTable[dex].object_table;
	size = ObjectHashTable[dex].table_size;
	for (ix = 0; ix < size; ix++) {
		if (!table[ix]) {		/* found a free slot */
			table[ix] = oP;
			goto done;
		}
	}
	size++;
	table = (elf_object **)TrueRealloc(table, (size * sizeof(elf_object *)));
	table[ix] = oP;
	ObjectHashTable[dex].object_table = table;
	ObjectHashTable[dex].table_size = size;
done:	
	oP->dex = OBJ_DEXIX(dex,ix);
	return(oP->dex);
}

/*************************************************************************
function: ElfObjPtr - Given object reference, return ptr to object
*/
elf_object *
ElfObjPtr (passedP)
data_ptr passedP;
{
	register data_object *p = OX(passedP);
	register elf_object *oP;

	VERIFY(p && !IS_INTEGER(p) && (p->type == TYPE_OBJECT),ERR_NOTOBJECT);
	if (p->objHostname)
		tempAbort("Remote object de-reference");
	if (p->objUid == 0)
		SnapObject(p,TRUE);
	VERIFY(oP = getObjectFromTable(p->objDex),ERR_OBJECTGONE);
	VERIFY(oP->uid == p->objUid,ERR_OBJECTGONE);
	return(oP);
}

/*************************************************************************
function: ElfObjPtrIfExists - Given object reference, return ptr to object
Same as ElfObjPtr, but does not throw error on non-existant objs...
Just returns a NULL
*/
elf_object *
ElfObjPtrIfExists (p,snapIt)
register data_object *p;
{
	register elf_object *oP;

	VERIFY(p && !IS_INTEGER(p) && (p->type == TYPE_OBJECT),ERR_NOTOBJECT);
	if (p->objHostname)
		return(NULL);
	if (p->objUid == 0) {
		if (!snapIt)
			return(NULL);
		SnapObject(p,FALSE);
		if (p->objUid == 0)
			return(NULL);
	}
	oP = getObjectFromTable(p->objDex);
	if (!oP || (oP->uid != p->objUid))
		return(NULL);
	return(oP);
}

/*************************************************************************
function: ElfObjPtrNoSnap - Given object reference, return ptr to object
*/
elf_object *
ElfObjPtrNoSnap (p)
register data_ptr p;
{
	register elf_object *oP;

	VERIFY(p && !IS_INTEGER(p) && (p->type == TYPE_OBJECT),ERR_NOTOBJECT);
	if (OX(p)->objHostname)
		return(NULL);
	if (OX(p)->objUid == 0)
		return(NULL);
	VERIFY(oP = getObjectFromTable(OX(p)->objDex),ERR_OBJECTGONE);
	VERIFY(oP->uid == OX(p)->objUid,ERR_OBJECTGONE);
	return(oP);
}


/***************************************************************************
function: FreeupExitedObjApps - free object applications that are exited
*/
FreeupExitedObjApps ()
{
	register int ix;
	register elf_object *oP;

	for (ix = 0; ix < SizeAppObjTable; ix++) {
		if (oP = AppObjTable[ix]) {
			if (oP->task_uid == -1) {			/* application exited */
				AppObjTable[ix] = NULL;
				elfDestroyAppObjects(oP);
			}
		}
	}
	return;
}


/***************************************************************************
function: addObjectToAppObjTable - add object to table of top level objects.
All application (top level) objects should be found here.
*/
static void
addObjectToAppObjTable (oP)
elf_object *oP;
{
	register int ix;
	register int free_dex;
	
	if (!oP || !oP->isTopLevelApp)
		return;

	free_dex = -1;
	for (ix = 0; ix < SizeAppObjTable; ++ix) {
		if (AppObjTable[ix] == oP)	/* alreay registered in Application Table */
			return;
		if (!AppObjTable[ix] && free_dex < 0)
			free_dex = ix;
	}
	if (free_dex >= 0)
		AppObjTable[free_dex] = oP;
	else {		/* increase table size */
		SizeAppObjTable += 32;
		AppObjTable = (elf_object **) TrueRealloc(AppObjTable,
										  SizeAppObjTable*sizeof(elf_object*));
		memset(AppObjTable+SizeAppObjTable-32, 0, 32*sizeof(elf_object*));
		AppObjTable[ix] = oP;
	}
}

static void
deleteObjectFromAppObjTable (oP)
register elf_object *oP;
{
	register int i;
	
	for (i = 0; i < SizeAppObjTable; ++i) {
		if (AppObjTable[i] == oP) {
			AppObjTable[i] = NULL;
			return;
		}
	}
	/* should error ?? */
}

/*************************************************************************
function: ElfbObjectExists - Return whether ptr pts to a non-detroyed object
*/
int
ElfbObjectExists (p)
register data_ptr p;
{
	register elf_object *oP;
	register int dex, ix;

	if (ElfbIsObject(p)) {
		dex = OBJ_DEX(OX(p)->objDex);
		ix = OBJ_IX(OX(p)->objDex);

		oP = NULL;
		if ((dex < OBJECT_HASH_PRIME) && (ix < ObjectHashTable[dex].table_size))
			oP = ObjectHashTable[dex].object_table[ix];

		if (oP && (oP->uid == OX(p)->objUid))
			return(BTRUE);
	}
	return(BFALSE);
}
/*************************************************************************
function: ElfbDataType - Return data type
*/
int
ElfbDataType (p)
data_ptr p;
{
	if (!p)
		return(0);
	if (IS_INTEGER(p))
		return(TYPE_NUMBER);
	return(p->type);
}

/*************************************************************************
function: elfLookupObject - Find object by name
*/
static elf_object *
elfLookupObject (name)
register char *name;
{
	register int dex, ix;
	register elf_object *oP;

	dex = objHash(name);
	for (ix = 0; ix < ObjectHashTable[dex].table_size; ++ix) {
		oP = ObjectHashTable[dex].object_table[ix];
		if (oP && oP->nameTag[0] && streq(oP->nameTag,name))
			return(oP);
	}
	return(NULL);
}

static elf_object *
elfLookupModel (name)
register char *name;
{
	register int ix;
	register char *appname;

	appname = elfHashName(name);
	for (ix = 0; ix < SizeAppObjTable; ix++) {
		if (AppObjTable[ix] && (AppObjTable[ix]->appName == appname) &&
			IS_CLASS_IN_READ(AppObjTable[ix]))
			return(AppObjTable[ix]);
	}
	return(elfLookupObject(name));
}

/***************************************************************************
function: findObjectWithPackname - scan entire object table for object
with given pack name
*/
static elf_object *
findObjectWithPackname (name)
char *name;
{
	register int dex, ix;
	register elf_object *oP, **table;
	
	for (dex = 0; dex < OBJECT_HASH_PRIME+1; ++dex) {
		table = ObjectHashTable[dex].object_table;
		for (ix = 0; ix < ObjectHashTable[dex].table_size; ++ix) {
			if ((oP = table[ix]) && oP->packP
				&& streq(oP->packP->source_name,name))
				return(oP);
		}
	}
	return(NULL);
}

/***************************************************************************
function: elfbCreateRemoteObject - create remote object and local
								   reference for remote object
*/
static data_ptr
elfbCreateRemoteObject (hostname, dex, uid, name)
char *hostname;
int dex, uid;
char *name;
{
	data_ptr send_msg, msg_args;
	data_ptr retval, inh;
	data_ptr hostnameP;
	int retCode;
	
	/* create format object_server_request$ (send message packet) */
	send_msg = ELF_CREATE_ARRAY(4);
	send_msg = ElfAddNumToArray(send_msg, 0, OBJSVR_CREATE_OBJECT_);
	msg_args = ELF_CREATE_ARRAY(2);	/* create msg.args array */
	inh = ELF_CREATE_ARRAY(2);
	inh = ElfAddNumToArray(inh, 0, dex);
	inh = ElfAddNumToArray(inh, 1, uid);
	msg_args = WriteArray(msg_args, 0, inh);
	msg_args = ElfAddStrToArray(msg_args, 1, name);
	msg_args = ElfAddStrToArray(msg_args, 2, hostname);
	send_msg = WriteArray(send_msg, 3, msg_args);

	hostnameP = AxTaskCreateElfString(0,hostname);
	retCode = ElfCallByName(NXLT("BUILDER_AXNET_RPC$"),&retval,2,
							hostnameP,send_msg);
	if (retCode != 0)
		retval = NULL;
	ELF_DATA_FREE(send_msg);
	AxTaskFreeElfData(0,hostnameP);
	VERIFY(retval && !IS_INTEGER(retval) && (retval->type == TYPE_ARRAY),
		   ERR_NOTARRAY);
	dex = IntFromArray(retval,0);
	uid = IntFromArray(retval,1);
	ELF_DATA_FREE(retval);
	ElfRetData = ELF_CREATE_REMOTE_OBJECT(dex,uid,hostname);
	return(ElfRetData);
}	

/*************************************************************************
function: ElfbCreateObject - Create an object
*/
void
ElfbCreateObject (inherit,typeName,hostname)
register data_ptr inherit;	/* name of object to inherirt from */
register data_ptr typeName;		/* user defined class name or "" */
register data_ptr hostname;
{
	register elf_object *oP;
	register int dex, uid;

	register elf_object *inheritFromP;
	register char* cpp_class_name;

	inheritFromP = NULL;
	if (ElfbIsString(inherit)) 
		inheritFromP = elfLookupModel(StrFromDataPtr(inherit));
	else if (inherit)
		inheritFromP = ElfObjPtr(inherit);

	if (inheritFromP && inheritFromP->cpp_class_info) {
		cpp_class_name = inheritFromP->cpp_class_info->className;
		elfCreateCppObject(ELF_CREATE_STRING("BaseClass"), 
						   ELF_CREATE_STRING(cpp_class_name), typeName, TRUE);
		return;
	}

#if 0
	if (inheritFromP && inheritFromP->ext_obj) {
		/* elfCreateCorbaObject(inherit,typeName);*/
		elfCreateExternalObject(inherit,typeName,hostname);
		return;
    }
#endif

		
	if (inherit && ElfbIsObject(inherit) && OX(inherit)->objHostname) {
		/* mimic remote_obj_create@(hostname, inheritParent, name) */
		elfbCreateRemoteObject(OX(inherit)->objHostname, OX(inherit)->objDex,
							   OX(inherit)->objUid, StrFromDataPtr(typeName));
		/* ElfFleshOutMethodHash(ElfObjPtr(inherit)); */
		return;
	}
	oP = elfCreateObject(StrFromDataPtr(typeName));
	if (inheritFromP) {
		oP->inheritFromP = inheritFromP;
		ElfFleshOutMethodHash(oP->inheritFromP);
	}
	else if (inherit) {
		ElfStrAbort(ERR_NOTOBJECT,NULL,StrFromDataPtr(inherit));
	}

	if (hostname || (oP->inheritFromP && oP->inheritFromP->hostname)) {
		if (hostname) {
			oP->hostname = elfHashName(StrFromDataPtr(hostname));
		}
		else if (oP->inheritFromP) {
			oP->hostname = oP->inheritFromP->hostname;
			ElfFleshOutMethodHash(oP->inheritFromP);
		}
	}
	ElfRetData = ELF_CREATE_OBJECT(oP->dex,oP->uid,NULL);
}

static elf_object *
elfCreateObject (nameTag)
register char *nameTag;
{
	register elf_object *oP;

	oP = (elf_object*)TrueCalloc(1,sizeof(elf_object));
	strcpy(oP->nameTag,nameTag);
	oP->uid = ++UidCounter;
	oP->hostname = NULL;
	addObjectToHashTable(oP);

	return(oP);
}

static elf_obj_grp *
elfCrtSubObjectGroup(name)
register char *name;
{
	register elf_obj_grp *mP;
	register int len;
	register char *p;

	len = strlen(name);
	if (len > ELFIDSTR)
		len = ELFIDSTR;

	mP = (elf_obj_grp*)TrueCalloc(1,sizeof(elf_obj_grp)+len);
	strncpy(mP->name,name,len);
	mP->name[len] = 0;
	for (p = mP->name; *p; p++)
		*p = toupper(*p);
	mP->hashval = hash(mP->name);
	return(mP);
}

/***************************************************************************
function: objUnloadPackage - unload safely current objects package
*/
static void
objUnloadPackage (oP)
register elf_object *oP;
{
	register int ix;
	register stack_frame *fp;
	package_info *FindPackage();
	
	if (!oP->packP)
		return;

	ElfUnFleshOutMethodHash(oP);
	ElfUnloadPackage(oP->packP);
	oP->packP = NULL;
}

/*************************************************************************
function: ElfbObjLoadImage - load compiled code into object
*/
ElfbObjLoadImage (dP,codeP,filenameP)
data_ptr dP;
data_ptr codeP;
data_ptr filenameP;	/* "" or name of file */
{
	elf_object *oP;
	package_info *packP;
	package_info *FindPackageByModuleName();

	oP = ElfObjPtr(dP);

	if (!codeP) {	/* connect to pre-installed package... Applix ONLY */
		if (!filenameP)
			objUnloadPackage(oP);
		else {
			packP = FindPackageByModuleName(StrFromDataPtr(filenameP));
			if (!packP) /* !!0 PUT BACK || packP->dynamic_load) */
				ElfAbort(ERR_ILLVAL);

			objUnloadPackage(oP);
			oP->packP = packP;
			ElfFleshOutMethodHash(oP->inheritFromP);
		}
		return;
	}

	if (!ElfbIsArray(codeP) || !ElfbIsArray(ReadArray(AX(codeP),0)))
		ElfAbort(ERR_ILLVAL);

	objUnloadPackage(oP);

	/*
	 * do not slave to this task, this will get unloaded
	 * when object is destroyed.
	 */
	oP->packP = ElfLoadPackage(codeP,FALSE);
	ElfFleshOutMethodHash(oP->inheritFromP);
}

/*************************************************************************
function: ElfFindAppObjByName - Return app object given its name
*/
static elf_object *
ElfFindAppObjByName (name, uid)
register char *name;	/* assumed to be hashed */
int uid;
{
	register int ix;

	for (ix = 0; ix < SizeAppObjTable; ix++) {
		if (AppObjTable[ix] && (AppObjTable[ix]->appName == name)) {
			if (!uid || (uid && AppObjTable[ix]->task_uid &&
						 uid == AppObjTable[ix]->task_uid))
				return(AppObjTable[ix]);
		}
	}
	return(NULL);
}

static elf_object *
ElfFindAppObjInRead (name)
register char *name;	/* assumed to be hashed */
{
	register int ix;

	for (ix = 0; ix < SizeAppObjTable; ix++) {
		if (AppObjTable[ix] && (AppObjTable[ix]->appName == name) &&
			IS_APP_IN_READ(AppObjTable[ix]))
			return(AppObjTable[ix]);
	}
	return(ElfFindAppObjByName(name,0));
}

/*************************************************************************
function: ElfFindObjInObjByName - Find child in object by name
While children of objects ever get snapped?
*/
elf_object *
ElfFindObjInObjByName (parent,childName)
register elf_object *parent;
register char *childName;
{
	register int dex,ix;
	register elf_object *oP, **table;

	/* Quick check thru parent's children for this obj */
	for (ix = 0; ix < parent->numChildren; ix++) {
		if (parent->children[ix]) {	/* can go to 0 during elfDestroyObject */
			oP = ElfObjPtrNoSnap(parent->children[ix]);
			if (oP && streq(childName,oP->nameTag))
					return(oP);
		}
	}

	/* Scan whole object table for object with matching parent */
	/* !!0 very expensive */
	for (dex = 0; dex < OBJECT_HASH_PRIME+1; ++dex) {
		table = ObjectHashTable[dex].object_table;
		for (ix = 0; ix < ObjectHashTable[dex].table_size; ++ix) {
			if ((oP = table[ix]) && streq(oP->nameTag,childName) &&
				oP->parentP && (oP->parentP == parent))
			return(oP);
		}
	}
	return(NULL);
}

/*************************************************************************
function: ObjChildOfApp - Return TRUE if liniage points back to app object
*/
int
ObjChildOfApp (oP,appName)
register elf_object *oP;
register char *appName;	/* hashed version */
{
	while (oP) {
		if (oP->appName != appName)
			return(FALSE);
		if (oP->isTopLevelApp)
			return(TRUE);
		oP = oP->parentP;
	}
	return(FALSE);	/* should never get here */
}

static
data_ptr getAllChildren(dP,aP)
data_ptr dP, aP;
{
	elf_object *oP;
	int ix;

	oP = ElfObjPtr(dP);
	for (ix = 0; ix < oP->numChildren; ix++) {
		aP = WriteArray(aP,AX(aP)->size,ElfDataCopy(oP->children[ix]));
		aP = getAllChildren(oP->children[ix],aP);
	}
	return(aP);
}


/***************************************************************************
function: elfDestroyRemoteObjects - destroy objects created on this hostname
by a client
*/
static
elfDestroyRemoteObjects (hostname)
char *hostname;
{
	register int dex, ix;
	register elf_object *oP, **table;
	register char *host;
	
	if (!hostname)
		return;
	host = elfHashName(hostname);
	for (dex = 0; dex < OBJECT_HASH_PRIME+1; ++dex) {
		table = ObjectHashTable[dex].object_table;
		for (ix = 0; ix < ObjectHashTable[dex].table_size; ++ix) {
			if ((oP = table[ix]) && oP->hostname == host)
				elfDestroyObject(oP);
		}
	}
}

static data_ptr
elfObjGetFuncsList (oP, lowercase)
elf_object *oP;
{
	register int ix,jx;
	register gfunc_info *sP;
	data_ptr aP;
	char func_name[128];

	if (!oP)
		return(NULL);

	/* For C++ Classes return methodNamesList */
	if (oP->cpp_class_info) 
		return(elfObjGetCppFuncsList(oP, lowercase));

#if ENABLE_CORBA
	if (oP->ext_obj)
		return(bldrExtObjMethodList(
					oP->ext_obj,
					NULL  ));
#endif

	if (!oP->packP)
		return(NULL);

	aP = ELF_CREATE_ARRAY(0);
	for (ix = 0; sP = oP->packP->gfun_list[ix]; ix++) {
		if (!lowercase)
			aP = ElfAddStrToArray(aP, ix, sP->name);
		else {
			strcpy(func_name,sP->name);
			for (jx=0; func_name[jx]; ++jx)
				func_name[jx] = Atolower((unsigned int)func_name[jx]);
			aP = ElfAddStrToArray(aP, ix, func_name);
		}
	}

	return(aP);
}

static char *valid_chars = NXLT("_+-.~");
static int uid_file_counter = 0;

static data_ptr
elfObjMakeUidFilename (name)
char *name;
{
	register char *s, *f;
	char filename[512];
	
	if (!name)
		return(NULL);
	f = filename;
	for (s = name; *s; ++s)
		{
#if BUILD_ORIENT
        if (IS_MULTI_BYTE(s))
            {
            *f++ = *s++;
            *f++ = *s;
            }
        else
#endif
		*f++ = (isalnum(*s) || strchr(valid_chars,*s)) ? *s: '_';
		}
	sprintf(f,"-%d",uid_file_counter);
	uid_file_counter++;
	return(ELF_CREATE_STRING(filename));
}


/***************************************************************************
function: elfObjInitObjvars - reset all local objvars that are declared in
the object's package and do it recursively for its children.
*/
static void
elfObjInitObjvars (oP)
elf_object *oP;
{
	register elf_obj_grp *p;
	register int ix;
	char *name;
	package_info *packP;

	/* reset all objvars that are declared in this package */
	if (packP = oP->packP) {
		for (ix = 0; packP->ovar_list[ix]; ++ix) {
			name = packP->ovar_list[ix]->name;
			if (p = getLocalData(oP, name, hash(name))) {
				if (!OBJVAR_TYPE_IS_STATIC(p->static_info) && p->value) {
					AxTaskFreeElfData(0,p->value);
					p->value = NULL;
				}
			}
		}
	}
	for (ix = 0; ix < oP->numChildren; ++ix)
		elfObjInitObjvars(ElfObjPtr(oP->children[ix]));
}

#define OBJINFO_CREATE_NEW_OBJ_FROM_OBJ_ 61
/*************************************************************************
function: ElfbObjInfo - Return array the describes object's contents
Jack of all trades.
*/
ElfbObjInfo (cmdP,dP,arg1,arg2,arg3)
data_ptr cmdP, dP, arg1, arg2, arg3;
{
	register int dex, uid, ix, jx, kx;
	register elf_object *oP, *o2;
	register char *name, *s;
	register data_ptr aP,next,item;
	register elf_obj_grp *p, *q;
	register int hashval;
	data_object *dO;
	char *host;
	data_ptr ObjectCreateFromStructure();
	data_ptr copyObject();
	int status = 0;

	ElfRetData = NULL;
	switch (IntFromDataPtr(cmdP)) {
/*
	case OBJINFO_CREATE_NEW_OBJ_FROM_OBJ_:
		ElfRetData = copyObject(dP, StrFromDataPtr(arg1), IntFromDataPtr(arg2),
								IntFromDataPtr(arg3));
		break;
*/
	case OBJINFO_CREATE_OBJ_FROM_STRUCT_:
		ElfRetData = ObjectCreateFromStructure(dP, IntFromDataPtr(arg1),
								  arg2, arg3);
		break;

	case OBJINFO_IS_SAME_DATA_:
		ElfRetData = MK_INTEGER(ElfSameData(dP, arg1) ? BTRUE: BFALSE);
		return;
		
	case OBJINFO_INIT_OBJVARS_:
		if (!dP)
			return;
		elfObjInitObjvars(ElfObjPtr(dP));
		return;

	case OBJINFO_MAKE_UID_FILENAME_:
		if (!dP)
			return;
		ElfRetData = elfObjMakeUidFilename(StrFromDataPtr(dP));
		return;

	case OBJINFO_GET_FUNCS_LIST_:
		ElfRetData = elfObjGetFuncsList(ElfObjPtr(dP),IntFromDataPtr(arg1));
		return;

	case OBJINFO_SET_BUILDER_APP_:
		oP = NULL;
		if (dP)
			oP = ElfObjPtr(dP);
		ApplixBuilderApp = oP;
		break;

	case OBJINFO_GET_BUILDER_APP_:
		if (ApplixBuilderApp)
			ElfRetData = ELF_CREATE_OBJECT(ApplixBuilderApp->dex,
										   ApplixBuilderApp->uid,NULL);
		break;

	case OBJINFO_DESTROY_REMOTE_OBJECTS_:
		elfDestroyRemoteObjects(StrFromDataPtr(dP)); /* hostname */
		break;

	case OBJINFO_CREATE_REF_FROM_DEXUID_:  /* create ref from dex-uid pair */
		host = StrFromDataPtr(dP);		/* hostname */
		dex = IntFromDataPtr(arg1);		/* dex */
		uid = IntFromDataPtr(arg2);		/* uid */
		/* create reference to remote object */
		ElfRetData = ELF_CREATE_REMOTE_OBJECT(dex,uid,host);
		return;

	case OBJINFO_GET_DEXUID_FROM_REF_:  /* get dex-uid from reference */
		dO = OX(dP);
		VERIFY(dO && !IS_INTEGER(dO) && (dO->type == TYPE_OBJECT),ERR_NOTOBJECT);
		aP = ELF_CREATE_ARRAY(0);
		aP = ElfAddNumToArray(aP,0,dO->objDex);
		aP = ElfAddNumToArray(aP,1,dO->objUid);
		ElfRetData = aP;
		return;
		
	case OBJINFO_GET_DEXUID_FROM_NAME_:  /* get dex-uid from name */
		name = StrFromDataPtr(dP);
		oP = elfLookupObject(name);
		if (oP) {
			aP = ELF_CREATE_ARRAY(2);
			aP = ElfAddNumToArray(aP,0,oP->dex);
			aP = ElfAddNumToArray(aP,1,oP->uid);
			ElfRetData = aP;
		}
		return;

	case OBJINFO_GET_OBJ_FROM_DEXUID_:  /* get object from dex-uid pair */
		dex = IntFromDataPtr(dP);
		uid = IntFromDataPtr(arg1);
		oP = getObjectFromTable(dex);
		if (oP && oP->dex == dex && oP->uid == uid)
			ElfRetData = ELF_CREATE_OBJECT(dex,uid,NULL);
		break;
		
	case OBJINFO_OBJ_FROM_PACK_NAME_:
		name = StrFromDataPtr(dP);
		oP = findObjectWithPackname(name);
		if (oP)
			ElfRetData = ELF_CREATE_OBJECT(oP->dex,oP->uid,NULL);
		return;

	case OBJINFO_GET_PACKAGE_:	/* get package name str for object */
		oP = ElfObjPtr(dP);
		if (oP->packP)
			ElfRetData = ELF_CREATE_STRING(oP->packP->source_name);
		return;

	case OBJINFO_READ_VARS_:	/* Create object from vars file */
		oP = ElfReadObjectDataFile(StrFromDataPtr(dP));
		if (arg1) {
			oP->appName = elfHashName(oP->nameTag);
			oP->isTopLevelApp = TRUE;
			addObjectToAppObjTable(oP);
		}
		ElfRetData = ELF_CREATE_OBJECT(oP->dex,oP->uid,NULL);
		break;

	case OBJINFO_WRITE_VARS_:	/* Write object's vars to file */
		ElfWriteObjectDataFile(StrFromDataPtr(dP), arg1);
		break;

	case OBJINFO_CREATE_APP_:	/* Create top level APP */
		/* dP is new obj name, arg1 is obj to inherit from */
		name = StrFromDataPtr(dP);
		name = elfHashName(name);	/* get one true copy of name */
		/* !!0 Verify unique name!! */
		ElfbCreateObject(arg1,dP,0);	/* inh, appname */
		oP = ElfObjPtr(ElfRetData);	/* set by ElfbCreateObject() */
		oP->appName = name;
		oP->isTopLevelApp = TRUE; /* Top Leve Application + in Construction */
		addObjectToAppObjTable(oP);
		return;

	case OBJINFO_GETAPP_:		/* Given appname, get app object */
		name = StrFromDataPtr(dP);
		name = elfHashName(name);	/* get one true copy of name */
		if (oP = ElfFindAppObjByName(name, 0)) {
			ElfRetData = ELF_CREATE_OBJECT(oP->dex,oP->uid,NULL);
			return;
		}
		ElfStrAbort(ERR_NOTOBJECT,NULL,name);

	case OBJINFO_APPLIST_:		/* Get list of apps */
		name = StrFromDataPtr(dP);
		dex = 0;
		aP = ELF_CREATE_ARRAY(0);
		if (!name || !name[0]) {
			for (ix = 0; ix < SizeAppObjTable; ix++) {
				if (oP = AppObjTable[ix])
					aP = WriteArray(aP, dex++,
									ELF_CREATE_OBJECT(oP->dex,oP->uid,NULL));
			}
		}
		else {
			name = elfHashName(name);
			for (ix = 0; ix < SizeAppObjTable; ix++) {
				if ((oP = AppObjTable[ix]) && oP->appName == name)
					aP = WriteArray(aP, dex++,
									ELF_CREATE_OBJECT(oP->dex, oP->uid, NULL));
			}
		}
		ElfRetData = aP;
		return;

	case OBJINFO_GETNAME_:		/* Return name */
		oP = ElfObjPtr(dP);
		ElfRetData = ELF_CREATE_STRING(oP->nameTag);
		return;

	case OBJINFO_GET_APPNAME_:
		oP = ElfObjPtr(dP);
		SVERIFY(oP->appName,ERR_ILLVAL,oP->nameTag);	/* !!0 better code */
		ElfRetData = ELF_CREATE_STRING(oP->appName);
		return;

	case OBJINFO_GET_INHERIT_:		/* Return obj this inherits from*/
		oP = ElfObjPtr(dP);
		if (!oP->inheritFromP)
			return;
		ElfRetData = ELF_CREATE_OBJECT(oP->inheritFromP->dex,oP->inheritFromP->uid,NULL);
		return;

	case OBJINFO_SET_INHERIT_:		/* set obj inherits from */
		oP = ElfObjPtr(dP);
		if (ElfbIsString(arg1)) {
			oP->inheritFromP = elfLookupModel(StrFromDataPtr(arg1));
			if (oP->inheritFromP == NULL)
				ElfStrAbort(ERR_NOTOBJECT,NULL,StrFromDataPtr(arg1));

			if (oP->ext_obj)
				if (oP->inheritFromP->ext_obj)
					bldrSetExtInheritance(oP->ext_obj, oP->inheritFromP->ext_obj, NULL);	
					
			ElfFleshOutMethodHash(oP->inheritFromP);
		}
		else if (arg1) {
			oP->inheritFromP = ElfObjPtr(arg1);

			if (oP->ext_obj)
				if (oP->inheritFromP->ext_obj)
					bldrSetExtInheritance(oP->ext_obj, oP->inheritFromP->ext_obj, NULL);	
					
			ElfFleshOutMethodHash(oP->inheritFromP);
		}
		return;

	case OBJINFO_GET_DATA_NAMES_:
		oP = ElfObjPtr(dP);
		aP = ELF_CREATE_ARRAY(0);
		for (p = oP->localData, ix = 0; p; p = p->next, ix++)
			aP = ElfAddStrToArray(aP,ix,p->name);
		ElfRetData = aP;
		return;

	case OBJINFO_GET_DATUM_:		/* local datum, do not look at inheritance */
		oP = ElfObjPtr(dP);
		name = StrFromDataPtr(arg1);
		if (p = getLocalData(oP, name, hash(name)))
			ElfRetData = ElfDataCopy(p->value);
		return;

	case OBJINFO_SET_DATUM_:
		oP = ElfObjPtr(dP);
		name = StrFromDataPtr(arg1);
		p = getLocalData(oP, name, hash(name));
		if (p && p->value) {
			AxTaskFreeElfData(0,p->value);
			p->value = AxTaskCopyElfData(0,arg2);
		}
		else
			crtLocalData(oP,name,arg2);
		return;

	case OBJINFO_LOCAL_STATIC_DATUMS_:
		if (!dP)
			return;
		oP = ElfObjPtr(dP);
		o2 = (ElfbIsObject(arg1)) ? ElfObjPtr(arg1) : NULL;
		aP = ELF_CREATE_ARRAY(0);
		for (ix = 0, p = oP->localData; p; p = p->next) {
			if (OBJVAR_TYPE_IS_STATIC(p->static_info) && p->value) {
				if (o2 && (q = getLocalData(o2,p->name,p->hashval))) {
					if (!ElfSameData(p->value, q->value)) {
						item = ELF_CREATE_ARRAY(2);
						item = ElfAddStrToArray(item,0,p->name);
						item = WriteArray(item,1,ELF_DATA_COPY(p->value));
						aP = WriteArray(aP, ix, item);
						ix++;
					}
				}
				else {
					item = ELF_CREATE_ARRAY(2);
					item = ElfAddStrToArray(item,0,p->name);
					item = WriteArray(item,1,ELF_DATA_COPY(p->value));
					aP = WriteArray(aP, ix, item);
					ix++;
				}
			}
		}
		ElfRetData = aP;
		return;
		
	case OBJINFO_SET_LOCAL_DATUMS_:
		oP = ElfObjPtr(dP);
		if (!arg1 || !ElfbIsArray(arg1))
			return;
		for (ix = AX(arg1)->size-1; ix >= 0; --ix) {
			aP = AnyFromArray(arg1,ix);
			name = StrFromArray(aP,0);
			p = ElfObjData(oP, name, hash(name));
			AxTaskFreeElfData(0,p->value);
			p->value = AxTaskCopyElfData(0,AnyFromArray(aP,1));
		}
		return;

	case OBJINFO_SET_STATIC_VARNAMES_:
		oP = ElfObjPtr(dP);
		jx = ElfbArraySize(arg1);
		for (ix = 0; ix < jx; ++ix) {
			name = StrFromArray(arg1, ix);
			for (s=name; *s; ++s)
				*s = toupper(*s);
			p = ElfObjData(oP, name, hash(name));
			if (p)
				p->static_info |= OBJVAR_TYPE_STATIC;
		}
		return;

	case OBJINFO_GET_STATIC_VARNAMES_:
		aP = ELF_CREATE_ARRAY(0);
		for (oP = ElfObjPtr(dP); oP; oP = oP->inheritFromP) {
			for (p = oP->localData; p; p = p->next) {
				if (OBJVAR_TYPE_IS_STATIC(p->static_info)) {
					/* do not duplicate */
					for (ix = 0; ix < ElfbArraySize(aP); ++ix) {
						if (streq(StrFromArray(aP,ix), p->name))
							break;
					}
					aP = ElfAddStrToArray(aP,ix,p->name);
				}
			}
		}
		ElfRetData = aP;
		return;

	case OBJINFO_SET_PUBLIC_VARNAMES_:
		oP = ElfObjPtr(dP);
		jx = ElfbArraySize(arg1);
		for (ix = 0; ix < jx; ++ix) {
			name = StrFromArray(arg1, ix);
			for (s=name; *s; ++s)
				*s = toupper(*s);
			hashval = hash(name);
			p = ElfObjData(oP, name, hashval);
			if (p) {
				p->static_info |= OBJVAR_TYPE_PUBLIC;
			}
		}
		if ( jx > 0 ) {
			if ( oP->hashtable )
				ElfUnFleshOutMethodHash(oP);/* clear current publics list */
			ElfFleshOutMethodHash(oP);		/* create publics list */
		}
		return;

	case OBJINFO_JAVADLG_SET_MAP_:
		oP = ElfObjPtr(dP);
		AddObjectToDlgMap(IntFromDataPtr(arg1), IntFromDataPtr(arg2), oP);
		return;

	case OBJINFO_JAVADLG_GET_MAP_: {
		int dlgId = IntFromDataPtr(arg1);
		int wId = IntFromDataPtr(arg2);
		for ( ix = 0; ix < javaDlgNumMapped; ix++ ) {
			if ( javaDlgMap[ix].oP
				&& (javaDlgMap[ix].dlgId == dlgId)
				&& (javaDlgMap[ix].widgetId == wId) ) {
				oP = javaDlgMap[ix].oP;
				ElfRetData = ELF_CREATE_OBJECT(oP->dex,oP->uid,NULL);
				return;
			}
		}
		return;
	}
	case OBJINFO_SET_APPNAME_:
		oP = ElfObjPtr(dP);
		oP->appName = elfHashName(StrFromDataPtr(arg1));
		if (arg2) {
			oP->isTopLevelApp = IntFromDataPtr(arg2);
			addObjectToAppObjTable(oP);
		}
		return;

	case OBJINFO_GET_PARENT_:
		oP = ElfObjPtr(dP);
		if (!oP->parentP)
			return;
		ElfRetData = ELF_CREATE_OBJECT(oP->parentP->dex,oP->parentP->uid,NULL);
		return;

	case OBJINFO_SET_PARENT_:
		oP = ElfObjPtr(dP);
		if (arg1)
			oP->parentP = ElfObjPtr(arg1);
		else
			oP->parentP = NULL;
		return;

	case OBJINFO_ADD_CHILD_:
		oP = ElfObjPtr(dP);
		ElfObjPtr(arg1)->appName = oP->appName;
		oP->numChildren++;
		oP->children = (data_ptr*)TrueRealloc(oP->children,
							(oP->numChildren * sizeof(data_ptr)));
		oP->children[oP->numChildren-1] = AxTaskCopyElfData(0,arg1);
		return;

   case OBJINFO_INSERT_CHILD_:
		oP = ElfObjPtr(dP);
		if (!arg1)		/* no object to insert */
			return;
		ix = 0;
		if (arg2) {
			o2 = ElfObjPtr(arg2);
			if (o2->parentP != oP)		/* wrong parent */
				return;
			for (ix = 0; ix < oP->numChildren; ++ix) {
				if (ElfObjPtr(oP->children[ix]) == o2)
					break;
			}
		}
		oP->numChildren++;
		oP->children = (data_ptr*)TrueRealloc(oP->children,
									  (oP->numChildren * sizeof(data_ptr)));
		bcopy(&oP->children[ix], &oP->children[ix+1], 
			  (oP->numChildren-ix-1) * sizeof(data_ptr));
		oP->children[ix] = AxTaskCopyElfData(0, arg1);
		ElfObjPtr(arg1)->parentP = oP;
		break;

	case OBJINFO_GET_CHILD_:
		oP = ElfObjPtr(dP);
		name = StrFromDataPtr(arg1);
		for (ix = 0; ix < oP->numChildren; ix++) {
			o2 = ElfObjPtr(oP->children[ix]);
			if (streq(o2->nameTag,name)) {
				ElfRetData = ELF_CREATE_OBJECT(o2->dex,o2->uid,NULL);
				return;
			}
		}
		return;

	case OBJINFO_GET_CHILDREN_:
		oP = ElfObjPtr(dP);
		aP = ELF_CREATE_ARRAY(0);
		for (jx = ix = 0; ix < oP->numChildren; ix++) {
			if (oP->uid == 1) {	/* root object */
				o2 = ElfObjPtr(oP->children[ix]);
				if (strrchr(o2->nameTag, '$'))
					continue;
				else {
					/* check the alias name */
					item = objGetObjvar(oP->children[ix], NXLT("ALIAS$$"));
					name = StrFromDataPtr(item);
					if (strrchr(name, '$'))
						continue;
				}
			}
			aP = WriteArray(aP,jx,ElfDataCopy(oP->children[ix]));
			jx++;
		}
		ElfRetData = aP;
		return;

	case OBJINFO_GET_CHILDREN_REVORDER_:
		oP = ElfObjPtr(dP);
		aP = ELF_CREATE_ARRAY(0);
		for (jx = 0,ix = oP->numChildren-1; ix >= 0; --ix) {
			if (oP->uid == 1) {	/* root object */
				o2 = ElfObjPtr(oP->children[ix]);
				if (strrchr(o2->nameTag, '$'))
					continue;
				else {
					/* check the alias name */
					item = objGetObjvar(oP->children[ix], NXLT("ALIAS$$"));
					name = StrFromDataPtr(item);
					if (strrchr(name, '$'))
						continue;
				}
			}
			aP = WriteArray(aP,jx,ElfDataCopy(oP->children[ix]));
			jx++;
		}
		ElfRetData = aP;
		return;

	case OBJINFO_GET_ALL_CHILDREN_:
		aP = ELF_CREATE_ARRAY(0);
		ElfRetData = getAllChildren(dP,aP);
		return;

	case OBJINFO_DEL_CHILDREN_:
		oP = ElfObjPtr(dP);
		for (ix = 0; ix < oP->numChildren; ix++) {
			elfDestroyAppObjects(ElfObjPtr(oP->children[ix]));
			AxTaskFreeElfData(0,oP->children[ix]);
		}
		oP->numChildren = 0;
		TrueFree(oP->children);
		oP->children = NULL;
		return;

	case OBJINFO_DEL_CHILD_:
		oP = ElfObjPtr(dP);
		if (!arg1)
			return;
		uid = ElfObjPtr(arg1)->uid;
		for (ix = 0; ix < oP->numChildren; ix++) {
			o2 = ElfObjPtr(oP->children[ix]);
			if (o2->uid == uid) {
				o2->parentP = NULL;
				AxTaskFreeElfData(0,oP->children[ix]);
				for (jx = ix+1; jx < oP->numChildren; jx++)
					oP->children[jx-1] = oP->children[jx];
				oP->numChildren--;
				oP->children = (data_ptr*)TrueRealloc(oP->children,
							(oP->numChildren * sizeof(data_ptr)));
				return;
			}
		}
		return;

	case OBJINFO_REPLACE_CHILD_:
		oP = ElfObjPtr(dP);
		if (!arg1)
			return;
		uid = ElfObjPtr(arg1)->uid;
		for (ix = 0; ix < oP->numChildren; ix++) {
			o2 = ElfObjPtr(oP->children[ix]);
			if (o2->uid == uid) {
				o2->parentP = NULL;
				AxTaskFreeElfData(0,oP->children[ix]);
				if (arg2) {
					elf_object *o3 = ElfObjPtr(arg2);
					o3->parentP = oP;
					oP->children[ix] = AxTaskCopyElfData(0,arg2);
					/* transfer all children from o2 to o3 */
					o3->numChildren = o2->numChildren;
					o3->children = o2->children;
					for (jx = 0; jx < o3->numChildren; ++jx)
						ElfObjPtr(o3->children[jx])->parentP = o3;
					o2->numChildren = 0;
					o2->children = NULL;
				}
				else {
					for (jx = ix+1; jx < oP->numChildren; jx++)
						oP->children[jx-1] = oP->children[jx];
					oP->numChildren--;
					oP->children = (data_ptr*)TrueRealloc(oP->children,
										  (oP->numChildren * sizeof(data_ptr)));
				}
				return;
			}
		}
		return;

	case OBJINFO_DESTROY_:
		if (!dP)
			return;
		oP = ElfObjPtr(dP);
		if (o2 = oP->parentP) {	/* delete from parent's children list */
			for (ix = 0; ix < o2->numChildren; ix++) {
				if (oP == ElfObjPtr(o2->children[ix])) {
					AxTaskFreeElfData(0,o2->children[ix]);
					for (jx = ix+1; jx < o2->numChildren; jx++)
						o2->children[jx-1] = o2->children[jx];
					o2->numChildren--;
					o2->children = (data_ptr*)TrueRealloc(o2->children,
										  (o2->numChildren * sizeof(data_ptr)));
					break;
				}
			}
		}
		elfDestroyAppObjects(oP);
		return;

	case OBJINFO_KILL_APP_:
		oP = ElfObjPtr(dP);
		if (!oP->isTopLevelApp)
			return;
		deleteObjectFromAppObjTable(oP);
		elfDestroyAppObjects(oP);
		return;

	case OBJINFO_SNAP_OBJECTS_:
		oP = ElfObjPtr(dP);
		elfSnapAppObjects(oP,IntFromDataPtr(arg1));
		if (IS_APP_IN_READ(oP))
			oP->isTopLevelApp = TRUE;
		for (ix = 0; ix < oP->numModels; ix++)
			deleteObjectFromAppObjTable(ElfObjPtr(oP->models[ix]));
		return;

	case OBJINFO_SET_APP_TASK_UID_:
		oP = ElfObjPtr(dP);
		if (!arg1)
			oP->task_uid = TaskInfo[AxInTask].uid;
		else
			oP->task_uid = IntFromDataPtr(arg1);
		return;

	case OBJINFO_GET_APP_TASK_UID_:
		oP = ElfObjPtr(dP);
		ElfRetData = MK_INTEGER(oP->task_uid);
		return;

	case OBJINFO_ADD_CLASS_:
		oP = ElfObjPtr(dP);
		o2 = ElfObjPtr(arg1);
		o2->appName = oP->appName;
		deleteObjectFromAppObjTable(o2);
		oP->numModels++;
		oP->models = (data_ptr*)TrueRealloc(oP->models,
											(oP->numModels * sizeof(data_ptr)));
		oP->models[oP->numModels-1] = AxTaskCopyElfData(0,arg1);
		return;

	case OBJINFO_GET_CLASS_:
		oP = ElfObjPtr(dP);
		name = StrFromDataPtr(arg1);
		for (ix = 0; ix < oP->numModels; ix++) {
			o2 = ElfObjPtr(oP->models[ix]);
			if (streq(o2->nameTag,name)) {
				ElfRetData = ELF_CREATE_OBJECT(o2->dex,o2->uid,NULL);
				return;
			}
		}
		return;

	case OBJINFO_GET_CLASSES_:
		oP = ElfObjPtr(dP);
		aP = ELF_CREATE_ARRAY(oP->numModels);
		for (ix = 0; ix < oP->numModels; ix++)
			WriteArray(aP,ix,ElfDataCopy(oP->models[ix]));
		ElfRetData = aP;
		return;

	case OBJINFO_DEL_CLASSES_:
		oP = ElfObjPtr(dP);
		for (ix = 0; ix < oP->numModels; ix++) {
			deleteObjectFromAppObjTable(ElfObjPtr(oP->models[ix]));
			elfDestroyAppObjects(ElfObjPtr(oP->models[ix]));
			AxTaskFreeElfData(0,oP->models[ix]);
		}
		oP->numModels = 0;
		TrueFree(oP->models);
		oP->models = NULL;
		return;

	case OBJINFO_DEL_CLASS_:
		oP = ElfObjPtr(dP);
		if (!arg1)
			return;
		uid = ElfObjPtr(arg1)->uid;
		for (ix = 0; ix < oP->numModels; ix++) {
			o2 = ElfObjPtr(oP->models[ix]);
			if (o2->uid == uid) {
				deleteObjectFromAppObjTable(ElfObjPtr(oP->models[ix]));
				ElfObjPtr(oP->models[ix])->parentP = NULL;
				AxTaskFreeElfData(0,oP->models[ix]);
				for (jx = ix+1; jx < oP->numModels; jx++)
					oP->models[jx-1] = oP->models[jx];
				oP->numModels--;
				oP->models = (data_ptr*)TrueRealloc(oP->models,
							(oP->numModels * sizeof(data_ptr)));
				return;
			}
		}
		return;

	case OBJINFO_LIBRARY_OBJECT_:
		oP = ElfObjPtr(dP);
		switch (IntFromDataPtr(arg1)) {
		case 0:	/* get lib object */
			name = StrFromDataPtr(arg2);
			if (!name)
				return;
			for (ix = 0; ix < oP->numLibobjs; ix++) {
				o2 = ElfObjPtr(oP->libobjs[ix]);
				if (streq(o2->nameTag, name)) {
					ElfRetData = ELF_CREATE_OBJECT(o2->dex,o2->uid,NULL);
					return;
				}
			}
			return;

		case 1: /* add lib object */
			if (!arg2)
				return;
			o2 = ElfObjPtr(arg2);
			o2->appName = oP->appName;
			oP->numLibobjs++;
			oP->libobjs = (data_ptr*)TrueRealloc(oP->libobjs,
										 (oP->numLibobjs * sizeof(data_ptr)));
			oP->libobjs[oP->numLibobjs-1] = AxTaskCopyElfData(0,arg2);
			return;
			
		case 2: /* delete lib object */
			if (!arg2)
				return;
			uid = ElfObjPtr(arg2)->uid;
			for (ix = 0; ix < oP->numLibobjs; ix++) {
				o2 = ElfObjPtr(oP->libobjs[ix]);
				if (o2->uid == uid) {
					ElfObjPtr(oP->libobjs[ix])->parentP = NULL;
					AxTaskFreeElfData(0,oP->libobjs[ix]);
					for (jx = ix+1; jx < oP->numLibobjs; jx++)
						oP->libobjs[jx-1] = oP->libobjs[jx];
					oP->numLibobjs--;
					oP->libobjs = (data_ptr*)TrueRealloc(oP->libobjs,
											(oP->numLibobjs * sizeof(data_ptr)));
					return;
				}
			}
			return;

		case 3: /* delete all lib objects */
			for (ix = 0; ix < oP->numLibobjs; ix++) {
				deleteObjectFromAppObjTable(ElfObjPtr(oP->libobjs[ix]));
				elfDestroyAppObjects(ElfObjPtr(oP->libobjs[ix]));
				AxTaskFreeElfData(0,oP->libobjs[ix]);
			}
			oP->numLibobjs = 0;
			TrueFree(oP->libobjs);
			oP->libobjs = NULL;
			return;

		case 4: /* get all lib objects */
			if (!oP->numLibobjs)
				return;
			aP = ELF_CREATE_ARRAY(0);
			for (ix = 0; ix < oP->numLibobjs; ix++)
				aP = WriteArray(aP,ix,ElfDataCopy(oP->libobjs[ix]));
			ElfRetData = aP;
			return;

		default :
			return;
		}
		
	case OBJINFO_MOVE_KID_:
		oP = ElfObjPtr(dP);
		dex = IntFromDataPtr(arg1);
		if (dex == 0 || !oP->parentP) {
			return;
		}
		for (ix = 0; ix < oP->parentP->numChildren; ++ix) {
			if (ElfObjPtr(oP->parentP->children[ix]) == oP) {
				jx = ix + dex;
				break;
			}
		}
		if (ix >= oP->parentP->numChildren)
			ElfStrAbort(ERR_ILLVAL,
						XLT("Object not found in its parent's children list"),
						oP->nameTag);
		while (jx < 0)
			jx += oP->parentP->numChildren;
		while (jx >= oP->parentP->numChildren)
			jx -= oP->parentP->numChildren;
		/* swap children jx and ix */
		next = oP->parentP->children[jx];
		oP->parentP->children[jx] = oP->parentP->children[ix];
		oP->parentP->children[ix] = next;
		ElfRetData = ELF_CREATE_INTEGER(jx);	/* return new location */
		return;

	case OBJINFO_GET_NEXT_SIBLING_:
		oP = ElfObjPtr(dP);
		jx = IntFromDataPtr(arg1);
		if (!oP->parentP)
			return;
		for (ix = 0; ix < oP->parentP->numChildren; ++ix)
			if (ElfObjPtr(oP->parentP->children[ix]) == oP)
				break;
		if (ix >= oP->parentP->numChildren)
			ElfStrAbort(ERR_ILLVAL,
						XLT("Object not found in its parent's children list"),
						oP->nameTag);
		ix += jx;
		if (ix < 0)
			ix += oP->parentP->numChildren;
		else if (ix >= oP->parentP->numChildren)
			ix -= oP->parentP->numChildren;
		if (ix < 0 || ix >= oP->parentP->numChildren)
			return;
		ElfRetData = ELF_DATA_COPY(oP->parentP->children[ix]);
		return;
		
	case OBJINFO_EXCHANGE_KIDS_:
		oP = ElfObjPtr(dP);
		o2 = ElfObjPtr(arg1);
		/*
		 * make sure oP and o2 are siblings
		 */
		if (!oP->parentP || oP->parentP != o2->parentP)
			return;
		for (ix = 0; ix < oP->parentP->numChildren; ++ix) {
			if (ElfObjPtr(oP->parentP->children[ix]) == oP)
				break;
		}
		if (ix >= oP->parentP->numChildren)
			ElfStrAbort(ERR_ILLVAL,
						XLT("Object not found in its parent's children list"),
						oP->nameTag);
		for (jx = 0; jx < oP->parentP->numChildren; ++jx) {
			if (ElfObjPtr(oP->parentP->children[jx]) == o2)
				break;
		}
		if (jx >= oP->parentP->numChildren)
			ElfStrAbort(ERR_ILLVAL,
						XLT("Object not found in its parent's children list"),
						o2->nameTag);
		/* swap children jx and ix */
		next = oP->parentP->children[jx];
		oP->parentP->children[jx] = oP->parentP->children[ix];
		oP->parentP->children[ix] = next;
		ElfRetData = ELF_CREATE_INTEGER(jx);	/* return new location */
		return;

	case OBJINFO_RENAME_:
		oP = ElfObjPtr(dP);
		name = StrFromDataPtr(arg1);
		if (oP->parentP) {
			for (ix = 0; ix < oP->parentP->numChildren; ++ix) {
				if (ElfObjPtr(oP->parentP->children[ix]) == oP) {
					AxTaskFreeElfData(0,oP->parentP->children[ix]);
					oP->parentP->children[ix] = NULL;
					break;
				}
			}
		}

		if (name && *name) {
			oP->hasBeenRenamed = TRUE;
			strcpy(oP->nameTag,name);
		}

		/* rehash and add to hash table */
		/* keep the old reference */
		addObjectToHashTable(oP);
		if (oP->parentP && ix < oP->parentP->numChildren)
			oP->parentP->children[ix] =
				AxTaskCreateElfObject(0,oP->dex,oP->uid,NULL);
		ElfRetData = ELF_CREATE_OBJECT(oP->dex,oP->uid,NULL);
		return;

	case OBJINFO_CREATE_CPP_OBJECT_:
		switch (IntFromDataPtr(arg3)) {
            case 0:	/* create the c++ object */
                elfCreateCppObject(dP,arg1,arg2, FALSE);   /* ELFclass, C++class, name */
                break;
            case 1:	/* return all C++ classes in the library */
                ElfRetData = elfGetCppClassNames(StrFromDataPtr(arg1)); /* libname*/
                break;
			case 2:	/* Get all the Class objects in the library */
				ElfRetData = elfGetAllCppClassObjects(dP, StrFromDataPtr(arg1));
				break;
			case 3: /* Return the quick help strs */
				ElfRetData = elfObjGetCppHelpStr(ElfObjPtr(dP), StrFromDataPtr(arg1));
				break;
            default:
                elfCreateCppObject(dP,arg1,arg2, FALSE);
        }
        break;

	case OBJINFO_PROCESS_CORBA_OBJECT_ :
		switch (IntFromDataPtr(arg3)) {
			case 0:	/* create the corba object - inheritfrom, name*/
				elfCreateExternalObject(dP,arg1,arg2);
				break;
			case 1:	/* Create Class - class, {objName,serverName,hostName} */
				elfCreateCorbaClass(arg1, arg2);
				break;
			case 2: /* set the invocation mode on the corba object */
				elfSetCorbaInvocationMode(dP,arg1);
				break;
			case 3:	/* get the result of the deferred invocation */
				ElfRetData = elfGetDeferredRequestResult(dP, arg1);
				break;
			case 4:	/* get the list of objects registered with the Naming Service */
#if 0
				ElfRetData = (data_ptr)AxGetNSObjectList(dP, arg1);
				break;
			case 5: /* convert information from the IR into a elf array */
				ElfRetData = (data_ptr)AxSaveCorbaInfo(arg1);
				break;
			case 6:
				AxCreateCorbaInfo(arg1);
				break;
#endif
			case 7: /* Return ELF_TRUE (-1) if the library is libaxcorba.so */
				ElfRetData = (data_ptr)bldrUseAxORB();
				break;
			case 8: /* Open socket and be ready for a callback event */
				elfRegisterObjectForCallbacks(dP, arg1, arg2);
				break;
			case 9: /* Terminate event for an external object */
				elfExtAppTerminated(dP);
				break;
			default:
				ElfRetData = NULL;
		}
		break;
	case OBJINFO_REPOSITORY_REQUEST_ :
		switch (IntFromDataPtr(arg3)) {
			case 0:
				/* arg1 = hostname, arg2 = objectname */
				ElfRetData = (data_ptr)bldrExtListOfObjects(arg1,arg2, NULL);
				break;
			case 1:
				/* arg1 = objectname */
				ElfRetData = (data_ptr)bldrExtObjMethodList(NULL,arg1);
				break;
			case 2:
				ElfRetData = (data_ptr)bldrExtAdditionalInfo(arg1, arg2);
				break;
			case 3:
				ElfRetData = (data_ptr)elfHelpOnExtObjMethod(dP,arg1,arg2);
				break;
			case 4:
				ElfRetData = (data_ptr)bldrExtHelpOnAllMethods(arg1);
				break;
			case 5:	/* get the idl file name from the IR for the given object */
				ElfRetData = (data_ptr)bldrGetInterfaceName(arg1,arg2,&status);
				if (status == -1) {
        			ElfStrAbort(-1,AxStrFromDataPtr(ElfRetData),0);
					ElfRetData = NULL;
    			}
				break;
			case 6: /* get formats for the method editor */
				ElfRetData = (data_ptr)elfExtStructsAsFormats(dP,arg1,arg2);
				break;
			case 7: /* Call dummy function to load the library - for debugging */
				ElfRetData = (data_ptr)bldrDummyCallToLoadLibrary();
				break;
			default:
				ElfRetData = NULL;
		}
		break;
    case OBJINFO_FIND_OBJECTS_:
        oP = ElfObjPtr(dP); /* find starting from this object */
        jx = AxIntFromDataPtr(arg1);    /* find objects with class_uid = jx*/
        ix = AxIntFromDataPtr(arg2);    /* 1 = find children */
        if (1 == ix) {
            /* find all children inheriting from name */
            ElfRetData = findChildrenInheritingFrom(dP, jx);
        }
        else {
            /* find all ancestors inheriting from name */
            ElfRetData = findParentInheritingFrom(dP, jx);
        }
        return;
		break;

	case OBJINFO_OBJECT_TYPE_:
		ElfRetData = AxMakeIntData(elfObjectType(dP));
		return;

 
    }
}
 


enum {NOT_AN_OBJECT=0, NORMAL_OBJECT, CORBA_OBJECT, CPP_OBJECT};
static int
elfObjectType(dP)
data_ptr dP;

{
	elf_object *oP = ElfObjPtr(dP);	
	int ret = NORMAL_OBJECT;

	if (ElfbObjectExists(dP) == BFALSE)
		ret = NOT_AN_OBJECT;
	else if (oP->ext_obj)
		ret = CORBA_OBJECT;
	else if (oP->cpp_obj_info)
		ret = CPP_OBJECT;
	
	return(ret);
}

static data_ptr findParentInheritingFrom(dP, class_uid)
data_ptr dP;
int class_uid;
{
	return(0);
}
/***************************************************************************
function: findChildrenInheritingFrom - find all children that inherit from
name
*/
 
static data_ptr
findChildrenInheritingFrom(dP, class_uid)
data_ptr dP;
int class_uid;
{
    register int ix, dex;
    data_ptr retP, kid_uidP, all_children;
    register int n_children;
    elf_object *oP;
 
    retP = AxMakeArray(0);
 
    dex = 0;
    all_children = AxMakeArray(0);
    all_children = getAllChildren(dP, all_children);
    n_children = AxArraySize(all_children);
    for (ix=0;ix<n_children;ix++) {
        kid_uidP = objGetObjvar((data_ptr)AnyFromArray(all_children,ix),NXLT("CLASS_UID$$"));
        if (class_uid == AxIntFromDataPtr(kid_uidP)) {
            oP = ElfObjPtr(AnyFromArray(all_children,ix));
            retP = WriteArray(retP, dex, ELF_CREATE_OBJECT(oP->dex,oP->uid,NULL)
);
            dex++;
        }
    }
 
    return(retP);
}


/***************************************************************************
function: elfDestroyAppObjects - destroy all objects with appName = name
*/
static
elfDestroyAppObjects (oP)
elf_object *oP;
{
	register int ix;
	
	for (ix = 0; ix < oP->numChildren; ++ix) {
		elfDestroyAppObjects(ElfObjPtr(oP->children[ix]));
		AxTaskFreeElfData(0,oP->children[ix]);
	}
	if (oP->numChildren)
		TrueFree(oP->children);
	oP->numChildren = 0;
	oP->children = NULL;
	elfDestroyObject(oP);
}

/*************************************************************************
function: AddObjectToDlgMap - Add Anyware widget to dialog cache map
*/
static
void AddObjectToDlgMap (int dlgId, int widgetId, elf_object *oP) {
	int ix = javaDlgNumMapped;

	if ( javaDlgMapHaveSlot ) {
		for ( ix = 0; ix < javaDlgNumMapped; ix++ )
			if ( !javaDlgMap[ix].oP )
				break;
	}
	if ( ix == javaDlgNumMapped ) {
		if ( javaDlgNumMapped == javaDlgNumAlloced ) {
			if ( !javaDlgMap )
				javaDlgMap = (javaDlgMap_t*) TrueCalloc(
								javaDlgNumAlloced=100,sizeof(javaDlgMap_t));
			else {
				javaDlgMap = (javaDlgMap_t*) TrueRealloc(javaDlgMap,
								javaDlgNumAlloced*2*sizeof(javaDlgMap_t));
				memset(&javaDlgMap[javaDlgNumAlloced],0,
							javaDlgNumAlloced*sizeof(javaDlgMap_t));
				javaDlgNumAlloced *= 2;
			}
		}
		javaDlgMapHaveSlot = FALSE;
		javaDlgNumMapped++;
	}
	javaDlgMap[ix].oP = oP;
	javaDlgMap[ix].dlgId = dlgId;
	javaDlgMap[ix].widgetId = widgetId;
}

/*************************************************************************
function: elfDestroyObject - Destroy object & its children
Will hang if an object is a child of itself!!0
*/
static
elfDestroyObject (oP)
elf_object *oP;
{
	register int dex, ix;
	elf_obj_grp *p, *next;

	if (oP->uid == 1)
		tempAbort("Attempted to destroy root object");

	if (oP->cpp_obj_info) {
		elfDestroyCppObject(oP);	/* call destructor to free the instance */
		return;
	}

	if (oP->ext_obj) {
		bldrExtDeleteObject(oP->ext_obj);	
	}

	for (ix = 0; ix < oP->numChildren; ix++) {
		ElfObjPtr(oP->children[ix])->parentP = NULL;
		AxTaskFreeElfData(0,oP->children[ix]);
	}
	for (p = oP->localData; p; p = next) {
		next = p->next;
		AxTaskFreeElfData(0,p->value);
		TrueFree(p);
	}
	oP->localData = NULL;
	
	if (oP->isTopLevelApp)
		deleteObjectFromAppObjTable(oP);
		
	for (ix = 0; ix < oP->numModels; ix++) {
		deleteObjectFromAppObjTable(ElfObjPtr(oP->models[ix]));
		elfDestroyAppObjects(ElfObjPtr(oP->models[ix]));
		AxTaskFreeElfData(0,oP->models[ix]);
	}
	oP->numModels = 0;
	TrueFree(oP->models);
	oP->models = NULL;
	
	for (ix = 0; ix < oP->numLibobjs; ix++) {
		elfDestroyAppObjects(ElfObjPtr(oP->libobjs[ix]));
		AxTaskFreeElfData(0,oP->libobjs[ix]);
	}
	oP->numLibobjs = 0;
	TrueFree(oP->libobjs);
	oP->libobjs = NULL;
	
	if (oP->packP)
		objUnloadPackage(oP);

	if ( !oP->hasBeenRenamed ) {
		dex = OBJ_DEX(oP->dex);
		ix = OBJ_IX(oP->dex);
		ObjectHashTable[dex].object_table[ix] = NULL;
	}
	else {	/* find older references, too */
		for (dex = 0; dex < OBJECT_HASH_PRIME+1; ++dex) {
			for (ix = 0; ix < ObjectHashTable[dex].table_size; ++ix) {
				if (oP == ObjectHashTable[dex].object_table[ix])
					ObjectHashTable[dex].object_table[ix] = NULL;
			}
		}
	}
	if ( oP->javaDlgMapped ) {
		for ( ix = 0; ix < javaDlgNumMapped; ix++ ) {
			if ( javaDlgMap[ix].oP == oP ) {
				javaDlgMap[ix].oP = NULL;
				javaDlgMapHaveSlot = TRUE;
			}
		}
	}

	TrueFree(oP);
}

void ElfbObjCrtChangeMethod(){abort();}
void ElfbObjCrtQueryMethod(){abort();}

/*************************************************************************
function: ElfbObjCrtLocalData - Add data
*/
void
ElfbObjCrtLocalData (obj,name,initialValue)
data_ptr obj, name, initialValue;
{
	crtLocalData(ElfObjPtr(obj),StrFromDataPtr(name),initialValue);
}

static elf_obj_grp *
getLocalData(oP,name,hashval)
elf_object *oP;
char *name;
int hashval;
{
	int ix;
	elf_objvar_item *bin;
	elf_obj_grp *p;
	
	for (p = oP->localData; p; p = p->next)
		if (p->hashval == hashval && streq(p->name,name))
			return(p);
	return(NULL);
}
			
static elf_obj_grp *
crtLocalData(oP,name,initialValue)
elf_object *oP;
char *name;
data_ptr initialValue;
{
	register int ix;
	register elf_objvar_item *bin, *new_item;
	elf_obj_grp *mP;

	mP = elfCrtSubObjectGroup(name);
	mP->value = AxTaskCopyElfData(0,initialValue);
	mP->next = oP->localData;
	oP->localData = mP;

	return(mP);
}

/***************************************************************************
function: ElfGetObjData - find local data in obj or in its inherited parents.
return NullDataPtr if does not exist.
*/
data_ptr *
ElfGetObjData (obj, name, hashval)
register elf_object *obj;
register char *name;
register int hashval;
{
	register elf_obj_grp *p;

	p = ElfFindObjLocalData(obj, name, hashval);
	if (p)
		return(&p->value);
	else
		return(&NullDataPtr);
}

/*************************************************************************
function: ElfObjData - Find local data in obj, create if not there
If the datum is inherited, it is snapshotted locally.
If the datum does not exist at all, created locally
*/
elf_obj_grp *
ElfObjData (obj, name, hashval)
register elf_object *obj;
register char *name;
register int hashval;
{
	register elf_obj_grp *p, *n;

/* Look for datum within */

	p = getLocalData(obj,name,hashval);
	if (p)
		return(p);

/* No, check inheritence, copy from inheriter */

	p = ElfFindObjLocalData(obj->inheritFromP, name, hashval);
	if (p) {
		n = crtLocalData(obj,name,p->value);
		n->static_info = p->static_info;
		return(n);
	}

/* No, create here now */

	return(crtLocalData(obj,name,NULL));
}

/***************************************************************************
function: AxFreeApplicationObject - kill application objects at
task-termination time
*/
AxFreeApplicationObject (pid)
int pid;
{
	register int i;
	int task_uid;
	
	task_uid = TaskInfo[pid].uid;
	for (i = 0; i < SizeAppObjTable; ++i) {
		if (AppObjTable[i] && (AppObjTable[i]->task_uid == task_uid))
			elfDestroyAppObjects(AppObjTable[i]);
	}
}


/***************************************************************************
function: ElfAddUnsnappedObjRefToList - add unsnapped obj ref to list
*/
void
ElfAddUnsnappedObjRefToList (dP)
data_object *dP;
{
	if (!dP->dynBinder)
		tempAbort(NXLT("no binder for dynamic obj"));
	dP->unsnappedAppName = elfHashName(StrFromArray(dP->dynBinder, 0));
	dP->prevUnsnapped = NULL;
	dP->nextUnsnapped = elfUnsnappedObjRefs;
	elfUnsnappedObjRefs = dP;
	if (dP->nextUnsnapped)
		dP->nextUnsnapped->prevUnsnapped = dP;
}

/*************************************************************************
function: ElfDecoupleUnsnappedObjPtrs - Called during task termination
*/
void
ElfDecoupleUnsnappedObjPtrs (pid)
register int pid;		/* THIMpid */
{
	register int ix;
	register elf_object *oP;
	int task_uid;
	
	task_uid = TaskInfo[pid].uid;
	for (ix = 0; ix < SizeAppObjTable; ++ix) {
		if ((oP = AppObjTable[ix]) && (oP->task_uid == task_uid)) {
			if (!IS_APP_IN_CONSTRUCTION(oP))
				oP->task_uid = -1;
			return;
		}
	}
}


/*************************************************************************
function: ElfFindObjLocalData - Search object & its antecendants
Must be blazingly fast!
*/
elf_obj_grp *
ElfFindObjLocalData (obj, name, hashval)
register elf_object *obj;
register char *name;
register int hashval;
{
	register elf_obj_grp *p;

	while (obj) {
		p = getLocalData(obj,name,hashval);
		if (p)
			return(p);
		obj = obj->inheritFromP;
	}
	return(NULL);
}

elf_obj_grp *
ElfFindObjPublicData (obj, name, hashval)
register elf_object *obj;
register char *name;
register int hashval;
{
	register elf_obj_grp *p;

	if (p = getLocalData(obj,name,hashval)) {
		if (OBJVAR_TYPE_IS_PUBLIC(p->static_info))
			return(p);
	}
	return(NULL);
}

/*************************************************************************
function: ElfObjDecRef - Decrement object refcount, free is goes to zero
Verifies there are no unsnapped references to the object before destroying
it.
This is ONLY called by the ELF_DATA_FREE right before freeing an object ptr,
therefore, this unthreads the object ptr from the unsnapped list.
*/
void
ElfObjDecRef (dP)
data_object *dP;
{
	if (dP->objUid == 0 && dP->objHostname == NULL) {
		if (dP->prevUnsnapped)
			dP->prevUnsnapped->nextUnsnapped = dP->nextUnsnapped;
		else
			elfUnsnappedObjRefs = dP->nextUnsnapped;
		if (dP->nextUnsnapped)
			dP->nextUnsnapped->prevUnsnapped = dP->prevUnsnapped;
		dP->nextUnsnapped = NULL;
		dP->prevUnsnapped = NULL;
	}
	ElfObjPtrIfExists(dP,FALSE);
}


/*************************************************************************
function: SnapObject - Load dex & uid from dynBinder info
The object's refcount is bumped at snap time because it connot be located
at ELF_CREATE_OBJECT(0,0,ap) time.
*/
static
SnapObject (p,errorOut)
register data_object *p;
register int errorOut;
{
	/* register data_object *p = OX(passedP); */
	register data_ptr aP;
	register elf_object *oP;
	register int cnt;
	register char *name;

	aP = p->dynBinder;
	cnt = ElfbArraySize(aP);

	name = p->unsnappedAppName;
	oP = ElfFindAppObjInRead(name);
	if (!oP) {
#if 0
		ElfbCreateObject(ELF_CREATE_STRING(name),ELF_CREATE_STRING("BaseClass"));
		oP = ElfObjPtr(ElfRetData);
		oP->appName = elfHashName(StrFromDataPtr(arg1));
		oP->isTopLevelApp = 4;
		addObjectToAppObjTable(oP);
#else
		if (errorOut)
			ElfStrAbort(ERR_ILLVAL,
						XLT("Reference to non-existant application"),name);
		else
			return;
#endif
	}

	while (--cnt > 0) {
		name = QUICK_STR_FROM_ARRAY(aP,cnt);
		oP = ElfFindObjInObjByName(oP,name);
		if (!oP) {
			if (errorOut)
				ElfStrAbort(ERR_ILLVAL,
							XLT("Reference to non-existant object"),name);
			else
				return;
		}
	}
	p->objDex = oP->dex;
	p->objUid = oP->uid;
	if (p->prevUnsnapped)
		p->prevUnsnapped->nextUnsnapped = p->nextUnsnapped;
	else 
		elfUnsnappedObjRefs = p->nextUnsnapped; 
	if (p->nextUnsnapped)
		p->nextUnsnapped->prevUnsnapped = p->prevUnsnapped;
	p->nextUnsnapped = NULL;
	p->prevUnsnapped = NULL;
}

/***************************************************************************
function: elfSnapAppObjects - Snap any unsnapped references of this application.
*/
static int
elfSnapAppObjects (app, hard_snap)
elf_object *app;
int hard_snap;
{
	register data_object *p, *next;
	register data_ptr dP, aP;
	register elf_object *oP;
	register int cnt;
	register char *name;
	
	for (p = elfUnsnappedObjRefs; p; p = next) {
		next = p->nextUnsnapped;
		/* snap object */
		if (streq(app->nameTag,p->unsnappedAppName))
			oP = app;
		else
			oP = ElfFindAppObjInRead(p->unsnappedAppName);
		if (!oP) {
			if (!hard_snap)
				continue;
			ElfStrAbort(ERR_ILLVAL,
						XLT("Reference to non-existant Application"),
						p->unsnappedAppName);
		}
		dP = p->dynBinder;
		cnt = ElfbArraySize(dP);
		while (--cnt > 0) {
			name = QUICK_STR_FROM_ARRAY(dP,cnt);
			oP = ElfFindObjInObjByName(oP,name);
			if (!oP) {
				if (!hard_snap)
					goto nextp;
				ElfStrAbort(ERR_ILLVAL,
							XLT("Reference to non-existant object"),name);
			}
		}
		p->objDex = oP->dex;
		p->objUid = oP->uid;
		if (p->prevUnsnapped)
			p->prevUnsnapped->nextUnsnapped = p->nextUnsnapped;
		else
			elfUnsnappedObjRefs = p->nextUnsnapped;
		if (p->nextUnsnapped)
			p->nextUnsnapped->prevUnsnapped = p->prevUnsnapped;
		p->nextUnsnapped = NULL;
		p->prevUnsnapped = NULL;
	nextp:;
	}
	/* at this point elfUnsnappedObjRefs should be null */
	return;
}

/***************************************************************************
function: DeReferenceServerObjects - create remote object references
for objects recieved on wire from server
*/
data_ptr
DeReferenceServerObjects (dP, hostname)
	data_ptr        dP;
	char            *hostname;
{
	register int    i;
	register data_ptr aP;
	int             size;

	if (ElfbIsArray(dP)) {
		size = ElfbArraySize(dP);
		for (i = 0; i < size; ++i) {
			aP = (data_ptr) AnyFromArray(dP, i);
			if (ElfbIsArray(aP))
				aP = DeReferenceServerObjects(aP, hostname);
			else if (ElfbIsObject(aP)) {
				OX(aP)->objHostname = hostname;
			}
		}
	}
	else if (ElfbIsObject(dP)) {
		OX(dP)->objHostname = hostname;
	}
	return (dP);
}

/*************************************************************************
function: ChannelDeReferenceServerObjects - Like DeRef... but encodes channel info
*/
data_ptr ChannelDeReferenceServerObjects (dP, channel)
	data_ptr        dP;
	int channel;
{
	char hostname[30];
	char *name;

	sprintf(hostname,":%d",channel);
	name = elfHashName(hostname);
	return(DeReferenceServerObjects(dP, name));
}

#if 0                           /* pgms/arcade/oui.els */
format obj_svr_msg$
	comcode,
	dex,
	uid,
	args
#endif

#ifndef ELFC_OGET
#define ELFC_OGET -8	/* Reserved for Builder->C interface */
#define ELFC_OSET -9	/* Reserved for Builder->C interface */
#endif

/***************************************************************************
function: elfCallRemoteObjMethod - run set/get method on remote object
*/
data_ptr
elfCallRemoteObjMethod(p, name, args_passed, query, broadcast)
	data_object    *p;          /* reference of remote object */
	char           *name;       /* method name */
	int             args_passed;/* number of args pushed on stack */
	int             query;      /* 0 if set method, else get method */
	int             broadcast;  /* TRUE if broadcast method */
{
	register int ix;
	data_ptr send_msg, msg_args,hostnameP;
	data_ptr retval, PopDatum();
	char *newname;
	int retCode;

	newname = name;
	if (broadcast) {
		newname = (char *) malloc(strlen(name) + 2);
		newname[0] = '*';       /* broadcast method identifier */
		strcpy(newname + 1, name);
	}

	if ( p->objHostname[0] == ':' ) {	/* leaks on thrown error... */
		int channel;
		channel = atoi(p->objHostname+1);		/* get past colon */
		send_msg = ELF_CREATE_ARRAY(3);	/* object, method, vals */
		WriteArray(send_msg,0,ELF_CREATE_OBJECT(p->objDex,p->objUid,NULL));
		ElfAddStrToArray(send_msg,1,newname);
		msg_args = ELF_CREATE_ARRAY(args_passed);
		for (ix = 0; ix < args_passed; ++ix)
			WriteArray(msg_args, ix, ELF_DATA_COPY(PopDatum()));
		WriteArray(send_msg,2,msg_args);
		ElfbSockRPCCall(channel, query ? ELFC_OGET : ELFC_OSET, send_msg);
		retval = ElfRetData;
	}
	else {
		/* create format object_server_request$ (send message packet) */
		send_msg = ELF_CREATE_ARRAY(4);
		if (query)
			send_msg = ElfAddNumToArray(send_msg, 0, OBJSVR_GET_METHOD_);
		else
			send_msg = ElfAddNumToArray(send_msg, 0, OBJSVR_SET_METHOD_);
		send_msg = ElfAddNumToArray(send_msg, 1, p->objDex);
		send_msg = ElfAddNumToArray(send_msg, 2, p->objUid);
		/* create msg.args array */
		msg_args = ELF_CREATE_ARRAY(args_passed+1);
		msg_args = ElfAddStrToArray(msg_args, 0, newname); /* name, without ?/! */
		for (ix = 0; ix < args_passed; ++ix)
			msg_args = WriteArray(msg_args, ix + 1, ELF_DATA_COPY(PopDatum()));
		send_msg = WriteArray(send_msg, 3, msg_args);
		ElfRetData = NULL;
		hostnameP = AxTaskCreateElfString(0,p->objHostname);
		retCode = ElfCallByName(NXLT("BUILDER_AXNET_RPC$"),&retval,
								2,hostnameP,send_msg);
		if (retCode != 0)
			retval = NULL;
							
		AxTaskFreeElfData(0,hostnameP);
	}
	ELF_DATA_FREE(send_msg);
	if (name != newname)
		free(newname);
	if (query) {
		retval = DeReferenceServerObjects(retval, p->objHostname);
		return (retval);
	}
	if (retval)
		ELF_DATA_FREE(retval);

	return (NULL);
}

static int
compSubArrayElem(arg1, arg2)
ACONST void *arg1, *arg2;
{
	data_ptr *e1, *e2;
	e1 = (data_ptr *)arg1;
	e2 = (data_ptr *)arg2;
	if (ElfbArraySize(*e1) != ElfbArraySize(*e2))
		tempAbort("Elements are of different size");
	return(IntFromArray(*e1,1)-IntFromArray(*e2,1));
}

static data_ptr 
elfFuncsInPackage(packP)
register package_info *packP;
{
	register int ix;
	register gfunc_info *sP;
	data_ptr aP, subArray;
	int cnt = 0;

	aP = ELF_CREATE_ARRAY(0);
	for (ix = 0; sP = packP->gfun_list[ix]; ix++) {
		subArray = ELF_CREATE_ARRAY(2);
		ElfAddStrToArray(subArray, 0, sP->name);
		ElfAddNumToArray(subArray, 1, sP->srcLine);	/* 0 - based */
		aP = WriteArray(aP, cnt++, subArray);
	}
	/* sort on line numbers */
	if (cnt)
		qsort(AX(aP)->vals,cnt,sizeof(data_ptr),compSubArrayElem);

	return(aP);
}



/*************************************************************************
function: ElfbFuncsInObject - Return list of funcs & their source line numbers
*/
ElfbFuncsInObject(dP)
data_ptr        dP;
{
	register elf_object *oP;

	oP = ElfObjPtr(dP);
	if (oP && oP->packP)
		ElfRetData = elfFuncsInPackage(oP->packP);
}

static shrt DynamicPublic_val = TYPE_OBJVAR;
static shrt *DynamicPublic_ptr = &DynamicPublic_val;

/*************************************************************************
function: ElfFleshOutMethodHash - Add Quick-hash if not already there
*/
void ElfFleshOutMethodHash (oP)
register elf_object *oP;
{
	register int hashcodes, ix, methods, hashval, next, poolpos, head;
	gfunc_info *sP;
	elf_obj_grp *vP;
	int publics = 0;
	int hashpoolcount;
	struct sp_pool_t {
		int count;
		gfunc_info **list;
	} sppool[ELF_HASH_PRIME], *s;

#define MAX_HASH_SHRT 65500

	if (!oP)
		return;
	if (oP->hashtable)
		return;
	if (!oP->packP)
		return;

	memset(sppool, 0, ELF_HASH_PRIME * sizeof(struct sp_pool_t));
	oP->hashtable = (unsigned short*)TrueCalloc(ELF_HASH_PRIME,
												sizeof(unsigned short));

	/* count how many methods there are plus number of hash vals */
	hashcodes = 0;
	head = MAX_HASH_SHRT;
	for (methods = 0; sP = oP->packP->gfun_list[methods]; methods++) {
		hashval = sP->hashval;
		s = &sppool[hashval];
		if (s->count == 0) {
			oP->hashtable[hashval] = head;
			head = hashval;
			hashcodes++;
		}
	/* add gfun_info node to the pool list */
		s->count++;
		s->list = (gfunc_info**)realloc(s->list,s->count*sizeof(gfunc_info*));
		s->list[s->count-1] = sP;
	}

	for (vP = oP->localData; vP; vP = vP->next) {
		if ( OBJVAR_TYPE_IS_PUBLIC(vP->static_info) ) {
			publics++;

			hashval = vP->hashval;
			s = &sppool[hashval];
			if (s->count == 0) {
				oP->hashtable[hashval] = head;
				head = hashval;
				hashcodes++;
			}

		/* create gfun_info node for this public var */
			sP = (gfunc_info*)AxTaskPoolAllocate(0,sizeof(gfunc_info));
			sP->value = DynamicPublic_ptr;
			sP->srcLine = 0;
			sP->hashval = hashval;
			sP->name = vP->name;

		/* add gfun_info node to the pool list */
			s->count++;
			s->list = (gfunc_info**)realloc(s->list,s->count*sizeof(gfunc_info*));
			s->list[s->count-1] = sP;
		}
	}

	hashpoolcount = methods + publics + hashcodes + 1;
	if (hashpoolcount >= MAX_HASH_SHRT)
		tempAbort("Object too complex");

	oP->hashpool = (gfunc_info**)TrueMalloc(hashpoolcount*sizeof(gfunc_info*));
	oP->hashpoolcount = hashpoolcount;
	poolpos = 0;
	oP->hashpool[poolpos++] = 0;		/* so first one will miss */

	/* load the table */
	for (hashval = head; hashval != MAX_HASH_SHRT; hashval = next) {
		next = oP->hashtable[hashval];
		oP->hashtable[hashval] = poolpos;
		s = &sppool[hashval];
		for (ix = 0; ix < s->count; ix++)
			oP->hashpool[poolpos++] = s->list[ix];
		free(s->list);
		oP->hashpool[poolpos++] = NULL;		/* null-term hash list */
	}
	if ( poolpos != hashpoolcount )
		abort();
}

/*************************************************************************
function: ElfUnFleshOutMethodHash - Remove quick-hash if there is one
*/
static void
ElfUnFleshOutMethodHash(oP)
elf_object *oP;
{
	register int ix;

	if (!oP)
		return;
	if (!oP->hashtable)
		return;
	for ( ix = oP->hashpoolcount-1; ix >= 0; ix-- ) {
		if ( oP->hashpool[ix] && oP->hashpool[ix]->value == DynamicPublic_ptr )
			AxTaskPoolFree(0,(char *)oP->hashpool[ix],sizeof(gfunc_info));
	}
	TrueFree(oP->hashtable);
	oP->hashtable = NULL;
	TrueFree(oP->hashpool);
	oP->hashpool = NULL;
}

/**************************************************************************
function: elfGetCppClass - Returns the class with the given name
*/

/**************************************************************************
function: elfGetCppClasses - Returns all the c++ classes
*/
static data_ptr
elfGetAllCppClassObjects(app, libname)
data_ptr app;
char *libname;
{
	register int ix, dex;
	data_ptr retP;
	cpp_class_info_t *cinfoP;
	register elf_object *oP, *o2;

	retP = ELF_CREATE_ARRAY(0);
	/* if app is NULL, look in the AppObjTable for the class */
	if (!app) {
		dex = 0;
		for (ix=0;ix < SizeAppObjTable; ++ix) {
			if ((oP = AppObjTable[ix]) && (cinfoP = oP->cpp_class_info))  {
				if (streq(cinfoP->libName, libname)) {
					retP = WriteArray(retP, dex, ELF_CREATE_OBJECT(oP->dex,oP->uid,NULL));
					dex = dex + 1;
				}
			}
		}
	}
	/* if app is not null look in app's model list */
	else {
		o2 = ElfObjPtr(app);
		dex=0;
		for (ix=0;ix<o2->numModels;ix++) {
			oP = ElfObjPtr(o2->models[ix]);
			if (oP && (cinfoP = oP->cpp_class_info)) {	
				if (streq(cinfoP->libName, libname)) {
					retP = WriteArray(retP, dex, ELF_CREATE_OBJECT(oP->dex,oP->uid,NULL));
					dex = dex+1;
				}
			}
		}
	}
		
	return (retP);
}


/**************************************************************************
function: elfObjGetCppHelpStr - Returns a help str for the given methodName 
*/
static data_ptr 
elfObjGetCppHelpStr(oP, methodName)
elf_object *oP;
char *methodName;
{
	data_ptr ap;
	register cpp_method_info_t *mP;

	if (!oP->cpp_class_info)
		return(NULL);
	
	for (mP = oP->cpp_class_info->methods;mP->methodName;++mP) {
		if (streqi(methodName, mP->methodName))
			if (!streqi(mP->helpStr, ""))
				return(ELF_CREATE_STRING(mP->helpStr));
			else
				return(NULL);
	}
	return(NULL);
}
	

/**************************************************************************
function: elfObjGetCppFuncsList - Returns a list of methods in the class
*/
static data_ptr
elfObjGetCppFuncsList(oP, lowercase)
elf_object *oP;
int lowercase;
{
	data_ptr aP;
	register cpp_method_info_t *mP;
	register int ix, jx;
	char func_name[128];

	func_name[0] = '?';		/* Make all C++ methods as "Get" Methods */
    aP = ELF_CREATE_ARRAY(0);
    for (ix = 0,mP = oP->cpp_class_info->methods;mP->methodName;++mP, ix++) {
		strcpy(&func_name[1], mP->methodName);
        if (!lowercase)
            aP = ElfAddStrToArray(aP, ix, func_name);
        else {
            for (jx=0; func_name[jx]; ++jx)
                func_name[jx] = Atolower((unsigned int)func_name[jx]);
            aP = ElfAddStrToArray(aP, ix, func_name);
        }
    }
	return(aP);
}

/**************************************************************************
function: elfGetCppClassNames - Returns a list of C++ classes.
*/
static data_ptr
elfGetCppClassNames(libname)
char *libname;
{
    register int ix, dex;
    data_ptr retP;
	cpp_class_info_t *cinfoP;
	register elf_object *oP;
 
    retP = ELF_CREATE_ARRAY(0);
    dex = 0;
    for (ix=0;ix < SizeAppObjTable; ++ix) {
        if ((oP = AppObjTable[ix]) && (cinfoP = oP->cpp_class_info))  {
			if (streq(cinfoP->libName, libname)) {
            	retP = ElfAddStrToArray(retP, dex, cinfoP->className);
            	dex = dex + 1;
			}
		}
    }
    return (retP);
}
 

/***************************************************************************
function: addObjectToCppClassTable - Add C++ class info object to the table.
*/
static void
addObjectToCppClassTable (cPP)
elf_object *cPP;
{
	register int ix;
	register int free_dex;
	
	if (!cPP || !cPP->cpp_class_info)
		return;

	free_dex = -1;
	for (ix = 0; ix < SizeCppClassTable; ++ix) {
		if (CppClassTable[ix] == cPP)	/* alreay registered in the Table */
			return;
		if (!CppClassTable[ix] && free_dex < 0)
			free_dex = ix;
	}
	if (free_dex >= 0)
		CppClassTable[free_dex] = cPP;
	else {		/* increase table size */
		SizeCppClassTable++;
		CppClassTable = (elf_object **) TrueRealloc(CppClassTable,
									  SizeCppClassTable*sizeof(elf_object*));
		CppClassTable[ix] = cPP;
	}
}

void 
ElfBuilderLoadCppClassInfo(libname, className, modelObj)
char *libname;
char *className;
void *modelObj;
{
	register int ix, cnt;	
	register char *p;
	elf_object *cP, *cpp_class;
	cpp_class_info_t *cinfoP;
	cpp_method_info_t *mP;
	char **names, **method_names, **help_strs, **strs;

	for (ix=0; ix < SizeAppObjTable; ++ix) {
		if ((cP = AppObjTable[ix]) && cP->cpp_class_info) {
			if (streq(className, cP->cpp_class_info->className))
				break;
		}
	}
	if (ix < SizeAppObjTable) {
		cpp_class = AppObjTable[ix];
		cinfoP = cpp_class->cpp_class_info;
		for (mP = cinfoP->methods; mP && mP->methodName; ++mP) {
			TrueFree(mP->methodName);
			TrueFree(mP->helpStr);
		}	
		if (cinfoP->modelObj != modelObj)
			AxDestroyCppObject(cinfoP->modelObj);
		TrueFree(cinfoP->methods);
        TrueFree(cinfoP->className);
        TrueFree(cinfoP);
        cpp_class->cpp_class_info = NULL;
    }
	else {
		ElfbCreateObject(ELF_CREATE_STRING("BaseClass"), 
						 ELF_CREATE_STRING(className), NULL);
		cpp_class = ElfObjPtr(ElfRetData);
	}
	    /* create class info */
    cinfoP = (cpp_class_info_t *)TrueMalloc(sizeof(cpp_class_info_t));
 
    cinfoP->className = TRUEALCSTR(className);
    cinfoP->libName = libname;      /* perm string */
    cinfoP->modelObj = modelObj;
 
    method_names = AxQueryCppInterface(&cinfoP->modelObj);
	help_strs 	 = AxQueryCppInterfaceHelpStrs (&cinfoP->modelObj);
 
    /* count method_info_table size */
    for (cnt = 0; method_names[cnt]; ++cnt)
        ;
 
    cinfoP->methods = (cpp_method_info_t *)TrueCalloc(cnt+1,
                                      sizeof(cpp_method_info_t));
    strs  = help_strs;
    names = method_names;
    for (mP = cinfoP->methods; *names; ++mP, ++names, ++strs) {
        mP->methodName = TRUEALCSTR(*names);
        for (p = mP->methodName; *p; ++p)       /* convert to uppercase */
            *p = toupper(*p);
        mP->hashval = hash(mP->methodName);
        mP->ftype = TYPE_U_FUNC;
		mP->helpStr = TRUEALCSTR(*strs);
    }
 
    TrueFree(method_names);     /* was alloc'd in AxQueryCppInterface() */
    TrueFree(help_strs);     	/* was alloc'd in AxQueryCppInterfaceHelpStrs() */
    cpp_class->cpp_class_info = cinfoP;

	cpp_class->appName = elfHashName(cpp_class->nameTag);
	cpp_class->isTopLevelApp = 4;
 
    /* Add this cpp object to its table */
    addObjectToAppObjTable(cpp_class);
}



/***************************************************************************
function: ElfObjLoadCppClassInfo - Create an elf_object to hold C++ class
information and keep this info object in a special table for faster lookup
If a class already exist in that special table, old info will be replaced
with and new info.
*/
void
ElfObjLoadCppClassInfo (libname, className, modelObj)
char *libname;
char *className;
void *modelObj;
{
	register int ix, cnt;
	register char *p;
	elf_object *cP, *cpp_class;
	cpp_class_info_t *cinfoP;
	cpp_method_info_t *mP;
	char **names, **method_names;

	data_ptr load_class_into_builder;
	
	/* Some way to tell if the install_c_library is being done from Builder */
	ElfbGetSysVar(NXLT("LOAD_CLASS_INTO_BUILDER$$"));
	load_class_into_builder = ElfRetData;
	if ( -1 == IntFromDataPtr(load_class_into_builder)) {     /* Elf True = -1 */
		ElfBuilderLoadCppClassInfo(libname, className, modelObj);
		return;
	}

	/* see if this "classname" already exist */
	for (ix = 0; ix < SizeCppClassTable; ++ix) {
		if ((cP = CppClassTable[ix]) && cP->cpp_class_info) {
			if (streq(className, cP->cpp_class_info->className))
				break;
		}
	}
	if (ix < SizeCppClassTable)	{	/* found in the table */
		cpp_class = CppClassTable[ix];
		/* delete old info */
		cinfoP = cpp_class->cpp_class_info;
		for (mP = cinfoP->methods; mP && mP->methodName; ++mP) {
			TrueFree(mP->methodName);
			TrueFree(mP->helpStr);
		}
		if (cinfoP->modelObj != modelObj)
			AxDestroyCppObject(cinfoP->modelObj);
		TrueFree(cinfoP->methods);
		TrueFree(cinfoP->className);
		TrueFree(cinfoP);
		cpp_class->cpp_class_info = NULL;
	}
	else	/* create an object to hold class info */
		cpp_class = elfCreateObject(className);

	/* create class info */
	cinfoP = (cpp_class_info_t *)TrueMalloc(sizeof(cpp_class_info_t));

	cinfoP->className = TRUEALCSTR(className);
	cinfoP->libName = libname;		/* perm string */
	cinfoP->modelObj = modelObj;

	method_names = AxQueryCppInterface(&cinfoP->modelObj);

	/* count method_info_table size */
	for (cnt = 0; method_names[cnt]; ++cnt)
		;

	cinfoP->methods = (cpp_method_info_t *)TrueCalloc(cnt+1,
									  sizeof(cpp_method_info_t));
	names = method_names;
	for (mP = cinfoP->methods; *names; ++mP, ++names) {
		mP->methodName = TRUEALCSTR(*names);
		for (p = mP->methodName; *p; ++p)		/* convert to uppercase */
			*p = toupper(*p);
		mP->hashval = hash(mP->methodName);
		mP->ftype = TYPE_U_FUNC;
	}

	TrueFree(method_names);		/* was alloc'd in AxQueryCppInterface() */
	cpp_class->cpp_class_info = cinfoP;
	
	cpp_class->inheritFromP = elfLookupModel(NXLT("BaseClass"));
	/* Add this cpp object to its table */
	addObjectToCppClassTable(cpp_class);
}


/***************************************************************************
function: ElfObjUnloadCppClasses - Unload C++ class info objects from memory
*/
void
ElfObjUnloadCppClasses (libname)
char *libname;
{
	register int ix;
	elf_object *cP;
	cpp_class_info_t *cinfoP;
	cpp_method_info_t *mP;
	
	for (ix = 0; ix < SizeCppClassTable; ++ix) {
		if ((cP = CppClassTable[ix]) && cP->cpp_class_info) {
			cinfoP = cP->cpp_class_info;
			if (cinfoP->libName != libname)	/* not this library */
				continue;
#if 0
			for (mP = cinfoP->methods; mP && mP->methodName; ++mP) {
				TrueFree(mP->methodName);
				TrueFree(mP->helpStr);
			}
			AxDestroyCppObject(cinfoP->modelObj);
			TrueFree(cinfoP->methods);
			TrueFree(cinfoP->className);
			TrueFree(cinfoP);
			cP->cpp_class_info = NULL;
			elfDestroyObject(cP);
			CppClassTable[ix] = NULL;
#endif
		}
	}
}

/***************************************************************************
function: elfCreateCppObject - instantiate a C++ object
*/
static void
elfCreateCppObject (inheritFrom, classnameP, objnameP, forBuilder)
data_ptr inheritFrom;	/* ELF class name or object */
data_ptr classnameP;	/* C++ class name */
data_ptr objnameP;		/* object name */
int forBuilder;
{
	elf_object *oP, *cP;
	cpp_class_info_t *cinfoP;
	register int ix;
	char *class_name;
	
	ElfbCreateObject(inheritFrom,objnameP,NULL);	/* inh, objname,hostname */
	oP = ElfObjPtr(ElfRetData);	/* set by ElfbCreateObject() */

	/* lookup for C++ class definition */

	class_name = StrFromDataPtr(classnameP);
	if (forBuilder)
		cP = elfLookupModel(class_name);
	else {
		for (ix = 0; ix < SizeCppClassTable; ++ix) {
			cP = CppClassTable[ix];
			if (!(cinfoP = cP->cpp_class_info))
				continue;
			if (streq(cinfoP->className,class_name))
				break;
		}
		if (ix == SizeCppClassTable)
			ElfStrAbort(ERR_ILLVAL,XLT("No such C++ class found"),class_name);
	}


	cinfoP = cP->cpp_class_info;
	if(!cP || !cinfoP)
		ElfStrAbort(ERR_ILLVAL,XLT("No such C++ class found"),class_name);

	oP->inheritFromP = cP;
	oP->cpp_obj_info = (cpp_obj_info_t *)TrueMalloc(sizeof(cpp_obj_info_t));
	oP->cpp_obj_info->classP = AxTaskCreateElfObject(0,cP->dex,cP->uid,NULL);
	oP->cpp_obj_info->objP = AxCreateCppObject(cinfoP->modelObj);

	
	return;		/* ElfRetData is already set */
}

/***************************************************************************
function: elfDestroyCppObject - Destroy a C++ object instance.
*/
static void
elfDestroyCppObject (oP)
elf_object *oP;
{
	elf_object *cP = NULL;
	cpp_class_info_t *cinfoP = NULL;
	
	if (oP->cpp_obj_info) {
		if (oP->cpp_obj_info->classP)
			cP = ElfObjPtr(oP->cpp_obj_info->classP);
		if (cP && cP->cpp_class_info && oP->cpp_obj_info->objP) {
			if (cinfoP = cP->cpp_class_info)
				AxDestroyCppObject(oP->cpp_obj_info->objP);
		}
		AxTaskFreeElfData(0, oP->cpp_obj_info->classP);
		TrueFree(oP->cpp_obj_info);
		oP->cpp_obj_info = NULL;
	}
	elfDestroyObject(oP);
}

/***************************************************************************
function: ElfCallCppMember - call a C++ member function thru dispatcher.
*/
data_ptr
ElfCallCppMember (oP, name, hashval, num_args)
elf_object *oP;
char *name;
int hashval;
int num_args;
{
	register int ix;
	data_ptr args, retP;
	cpp_class_info_t *cinfoP;
	cpp_method_info_t *mP;
	elf_object *cP;
	data_ptr PopDatum();
	
	/* pop args from stack and package as array */
	args = ELF_CREATE_ARRAY(num_args);
	for (ix = 0; ix < num_args; ++ix)
		args = WriteArray(args, ix, ELF_DATA_COPY(PopDatum()));
	retP = NULL;
	while (oP) {
		if (oP->cpp_obj_info) {
			if (!oP->cpp_obj_info->classP)
				ElfStrAbort(ERR_ILLVAL, XLT("C++ class not found"), name);
			cP = ElfObjPtr(oP->cpp_obj_info->classP);
			cinfoP = cP->cpp_class_info;
			if (!cinfoP)
				ElfStrAbort(ERR_ILLVAL, XLT("C++ class not found"), name);
			/* find the method interface node */
			for (mP = cinfoP->methods; mP && mP->methodName; ++mP) {
				if (hashval == mP->hashval && streq(name, mP->methodName)) {
					retP = AxCallCppMemberFunc(cinfoP->modelObj, &oP->cpp_obj_info->objP, 
											   mP-cinfoP->methods, args);
					goto done;
				}
			}
		}
		oP = oP->inheritFromP;
	}
 done:
	ELF_DATA_FREE(args);
	return(retP);
}


#if 0

format object_format@
	name,
	inherits_from,
	source,					/* source code - null if turbo version */
	exec_image,				/* null - exec image if turbo version */
	format arrayof object_format@ children,
	format arrayof object_local_format@ locals,
	class_libraries,		/* list of localized and links of class libraries used */
	resources,				/* list of localized and links of resource files */
	object_library_pathname,/* if it is a library object, its library name */
	lib_obj_name		    /* if it is a library object, its name in library */

#endif

#define NAME_dex		0
#define INHERIT_dex		1
#define SOURCE_dex		2
#define IMAGE_dex		3
#define CHILDINFO_dex	4
#define LOCALS_dex		5
#define CLASSLIBS_dex	6
#define RESOURCES_dex	7
#define OBJLIBPATH_dex	8
#define LIBOBJNAME_dex	9

static data_ptr
ObjectCreateFromStructure (objinfo, is_top_level, libobjs, flags)
data_ptr objinfo;
int is_top_level;
data_ptr libobjs;
data_ptr flags;
{
	data_ptr name;
	data_ptr inherits_from;
	data_ptr source;
	data_ptr exec_image;
	data_ptr children;
	data_ptr locals;
	data_ptr class_libraries;
	data_ptr resources;
	char *objlib_path;
	char *libobj_name;
	data_ptr theObj;
	char filename[STRPATH+1];
	char libpath[STRPATH+1];
	char libfile[STRPATH+1];
	register int ix, numKids;
	data_ptr source_fileinfo;
	data_ptr copyObject();
	data_ptr ext_file;

	int is_turbo, dontError;
	static data_ptr read_errors = NULL;
	data_ptr errors;
	char *inheritStr, *nameStr, errStr[512];
	elf_object *oP;
	elf_object *appObj = 0;
	static data_ptr classList;
	static int classIx = 0;
	
	data_ptr corba_obj_info;

	is_turbo = AxIntFromArray(flags,0);
	dontError = AxIntFromArray(flags,1);

#define formatMember(dex) AnyFromArray(objinfo,dex)
#define setObjvar(obj,name,val) do { data_ptr str = ELF_CREATE_STRING(name);\
		ElfbObjInfo(MK_INTEGER(OBJINFO_SET_DATUM_),obj,str,val); \
		ELF_DATA_FREE(str); } while (0)
#define getObjvar(obj,name) do { data_ptr str = ELF_CREATE_STRING(name);\
		ElfbObjInfo(MK_INTEGER(OBJINFO_GET_DATUM_),obj,str); \
		ELF_DATA_FREE(str); } while (0)
	
	name = formatMember(NAME_dex);
	inherits_from = formatMember(INHERIT_dex);
	source = formatMember(SOURCE_dex);
	exec_image = formatMember(IMAGE_dex);
	children = formatMember(CHILDINFO_dex);
	locals = formatMember(LOCALS_dex);
	class_libraries = formatMember(CLASSLIBS_dex);
	resources = formatMember(RESOURCES_dex); 
	objlib_path = StrFromDataPtr(formatMember(OBJLIBPATH_dex));
	libobj_name = StrFromDataPtr(formatMember(LIBOBJNAME_dex));

	/* check if its a library object */
	if (*objlib_path) {
		char *t;

		/* first make sure we have the right dir slashes. */
        AxSysDosUnixPath(objlib_path);
		strcpy(libpath,objlib_path);
		if (t = strrchr(objlib_path, *DirSlash))
			strcpy(libpath,t+1);
		if (!libobjs) {
			if (dontError) {
				/** this library object did not get created	*/ 
                if (AxIsArray(name))
                    nameStr = AxStrFromArray(name,0);
                else
                    nameStr = StrFromDataPtr(name);
				errors = AxMakeArray(2);
                sprintf(errStr,"%s - %s",nameStr, XLT("Library object not found."));
                errors = AxAddStrToArray(errors, 0, errStr);
                read_errors = AxAppendArray(read_errors,errors);
				return(NULL);
			}
			else
				ElfStrAbort(-1,"No such library object",libobj_name);
		}
		else {
			for (ix = 0; ix < AX(libobjs)->size; ++ix) {
				data_ptr libobj, libpathinfo;

				libobj = AnyFromArray(libobjs,ix);
				if (0 == strcmp(ElfObjPtr(libobj)->nameTag, libobj_name)) {
					libpathinfo = objGetObjvar(libobj,"LIBRARY_OBJECT_INFO$$");
					strcpy(libfile, StrFromArray(libpathinfo,2));
					if (t = strrchr(libfile,*DirSlash))
						strcpy(libfile,t+1);
					if (0 == strcmp(libpath,libfile)) {
						theObj = copyObject(libobj,StrFromDataPtr(name),flags,TRUE);
						break;
					}
				}
			}
			if (ix == AX(libobjs)->size) {
				if (dontError) {
					errors = AxMakeArray(3);
					sprintf(errStr,"%s - %s",libobj_name, XLT("Library object not found."));
					errors = AxAddStrToArray(errors, 0, errStr);
					errStr[0] = '\0';
					sprintf(errStr,"%s - %s",libpath, XLT("Probably missing."));
					errors = AxAddStrToArray(errors, 0, errStr);
					read_errors = AxAppendArray(read_errors,errors);
					return(NULL);
				}
				else
					ElfStrAbort(-1,"No such library object",libobj_name);
			}
		}
	}
	else {
		if (dontError) {
			/* objects with non-existant classes are made to inherit from
			   the BaseClass. This prevents throwing an error later */
			if (ElfbIsString(inherits_from)) {
				inheritStr = StrFromDataPtr(inherits_from);
				oP = elfLookupModel(inheritStr);
			}
			else {
				oP = getObjectFromTable(OX(inherits_from)->objDex);
				inheritStr = OX(inherits_from)->unsnappedAppName;
				if (!oP)
					oP = ElfFindAppObjInRead(inheritStr);
			}
			if (!oP) {
				errors = AxMakeArray(3);
        		sprintf(errStr,"%s -  %s",inheritStr, XLT("No such class."));
        		errors = AxAddStrToArray(errors, 0, errStr);
        		errStr[0] = '\0';
				if (AxIsArray(name))
					nameStr = AxStrFromArray(name,0);
				else
					nameStr = StrFromDataPtr(name);
        		sprintf(errStr,"%s %s", nameStr,XLT("now inherits from BaseClass."));
				errors = AxAddStrToArray(errors, 1, errStr);
				read_errors = AxAppendArray(read_errors,errors);
        		ElfbCreateObject(ELF_CREATE_STRING("BaseClass"),ELF_CREATE_STRING(inheritStr), NULL);
        		oP = ElfObjPtr(ElfRetData);
        		oP->appName = elfHashName(inheritStr);
        		oP->isTopLevelApp = 4;
				classList = AxAddArrayToArray(classList,classIx,ElfRetData);
				classIx++;
        		addObjectToAppObjTable(oP);
			}           
		}	/* end dontError */
		
		ElfbCreateObject(inherits_from,name,NULL);
		theObj = ElfRetData;
			 
		if (is_top_level)  {
			read_errors = AxMakeArray(0);
			classIx = 0;
			classList = AxMakeArray(0);
			objSetAppName(theObj,name,3); /* in read */
		}
		if (ElfbIsArray(source) && AX(source)->size > 0) {
			if (is_turbo)
				loadMethodsFile(theObj,source,NULL); 
			objSetObjvar(theObj, "SOURCE_ARRAY$$", source);
		}
		else if ( (BTRUE == ElfbIsArray(AnyFromArray(exec_image,0))) &&
				  (BTRUE == ElfbIsBinary(AnyFromArray(exec_image,5))) ) {
			data_ptr fileinfo, new_image;
			
			new_image = ELF_DATA_COPY(exec_image);
			WriteArray(AnyFromArray(new_image,0),1,
					   elfObjMakeUidFilename(StrFromDataPtr(name)));
			loadExecImage(theObj,new_image);
			objSetObjvar(theObj, "EXEC_IMAGE$$", new_image);
			objSetObjvar(theObj,"EXTERNAL_SOURCE_FILE$$", NULL);
			ELF_DATA_FREE(new_image);
		}
		
		numKids = ElfbArraySize(children);
		for (ix = 0; ix < numKids; ++ix) {
			data_ptr kid;
			
			kid = ObjectCreateFromStructure(AnyFromArray(children,ix),
											FALSE, libobjs, flags);
			if (kid) {
				objAddChild(theObj,kid);
				ElfObjPtr(kid)->parentP = ElfObjPtr(theObj);
				ELF_DATA_FREE(kid);
			}
		}
	}

	/* local variables */
	filename[0] = '\0';
	objSetLocalDatums(theObj,locals);
	corba_obj_info = objGetObjvar(theObj,"FULL_CORBA_NAME$$");
	if (corba_obj_info) {
		int status = 0;
		int use_file = 0;
		elf_object *classP;
		classP = ElfObjPtr(theObj)->inheritFromP;
		use_file = AxIntFromArray(corba_obj_info,3);
		if (use_file == BTRUE) {
			strcpy(filename, StrFromArray(corba_obj_info,2));	
			ElfbReadFile(filename);
			corba_obj_info = AxAddDataToArray (corba_obj_info, 0, ElfRetData);
		}
		if (!classP->ext_obj) {
			data_ptr corbaClassName, tmp, objectInfo;
			corbaClassName  = AnyFromArray(corba_obj_info, 4);
			classP->ext_obj = (void *)bldrExtCreateClassObj(
										corbaClassName, corba_obj_info, NULL, &status);
		}
		ElfObjPtr(theObj)->ext_obj = bldrExtCreateObject(corba_obj_info,
											classP->ext_obj,&status, NULL);
	}

	filename[0] = '\0';
	source_fileinfo = objGetObjvar(theObj,"EXTERNAL_SOURCE_FILE$$");
	strcpy(filename, StrFromArray(source_fileinfo,1));
	strcat(filename, StrFromArray(source_fileinfo,0));
	if (filename[0] && !ElfbIsArray(exec_image)) {
		ElfbReadFile(filename);
        source = ElfRetData;
		if (is_turbo)
			loadMethodsFile(theObj,source,filename);
		objSetObjvar(theObj, "EXTERNAL_SOURCE_FILE$$", source_fileinfo);
		objSetObjvar(theObj,"SOURCE_ARRAY$$",source);
		ELF_DATA_FREE(source);

	}
	if (is_turbo)
		objSetObjvar(theObj,"SOURCE_ARRAY$$",NULL);

	if (is_top_level && read_errors) {
		/** when we get to the app object move all dummy classes to the models list */
		appObj = ElfObjPtr(theObj);
		for (ix=0;ix<classIx;ix++) 
			objAddClassToAppModels(appObj,AnyFromArray(classList,ix));
		objSetObjvar(theObj,"READ_ERRORS$$",read_errors);
		ELF_DATA_FREE(read_errors);
		read_errors = 0;
		classIx = 0;
		classList = 0;
	}

	return(theObj);
}

static data_ptr
elfObjectCreateFromStructure (args)
data_ptr args;
{
	data_ptr info;
	int is_top_level;
	data_ptr libobjs, flags;

	info = AnyFromArray(args, 0);
	is_top_level = IntFromArray(args, 1);
	libobjs = AnyFromArray(args, 2);
	flags = AnyFromArray(args,3);

	return(ObjectCreateFromStructure(info,is_top_level,libobjs,flags));
}


static
loadExecImage (theObj, image)
data_ptr theObj, image;
{
		ElfbObjLoadImage(theObj,image,NULL);
}

static int
isExecImage(image)
data_ptr image;
{
	if (!ElfbIsArray(image) || 
	    !ElfbIsArray(ReadArray(AX(image),0)) ||	
		!ElfbIsBinary(AnyFromArray(image,5)) )
		return(FALSE);
	else
		return(TRUE);
}

static
loadMethodsFile (theObj, source, file)
data_ptr theObj;
data_ptr source;
char *file;
{
	data_ptr image, arg;
	elf_object *oP;
	char *tmpfile;
	data_ptr src;

	tmpfile = (char *)NULL;
	src = NULL;

   	oP = ElfObjPtr(theObj);

	tmpfile = StrFromDataPtr(elfObjMakeUidFilename(oP->nameTag));
	src = source;

	if (src && tmpfile) {
		arg = AxMakeArray(2);
		arg = AxAddStrToArray(arg, 0, tmpfile);
		arg = AxAddArrayToArray(arg, 1, src);
		image = (data_ptr)AxfCompileArray(arg);
		if (!isExecImage(image)) {
			char *errStr = 0;
			errStr = compilationErrors(image);
			ELF_DATA_FREE(image);
        	ElfStrAbort(ERR_NOCOMPILE, errStr, tmpfile);
		}
		loadExecImage(theObj, image);
		objSetObjvar(theObj, "EXEC_IMAGE$$", image);
		ELF_DATA_FREE(image);
	}
}

static char *
compilationErrors(image)
data_ptr image;
{
	char *err, *ret;
	register int n, ix, sz, dex;

	err = 0;
	ret = (char *)malloc(sizeof(char)*1024);
	memset(ret, 0, 1024);
	n = AxArraySize(image);
	sz = 0;
	dex = 0;
	for (ix=0;ix<n;ix++) {
		err = AxStrFromArray(image, ix);
		sz = strlen(err)+2;
		if (dex+sz > 1024)
			break;
		sprintf(&ret[dex], "%s\n",err);	
		dex += sz;
	}
	return(ret);
}

static elf_object *
doForkObject (sP, name)
elf_object *sP;
char *name;
{
	register int ix;
	elf_object *oP, *cP;
	elf_obj_grp *s, *p;
	
	oP = elfCreateObject(name ? name : sP->nameTag);
	oP->inheritFromP = sP;

	if ((oP->numChildren = sP->numChildren) > 0)
		oP->children = (data_ptr*)TrueMalloc(oP->numChildren * sizeof(data_ptr));
	for (ix = 0; ix < oP->numChildren; ++ix) {
		cP = doForkObject(ElfObjPtr(sP->children[ix]), NULL);
		cP->parentP = oP;
		oP->children[ix] = AxTaskCreateElfObject(0,cP->dex,cP->uid,NULL);
	}
	return(oP);
}


static data_ptr
forkObject (srcObj,dstName)
data_ptr srcObj;
data_ptr dstName;
{
	char *name;
	elf_object *sP, *oP;
	
	sP = ElfObjPtr(srcObj);
	name = StrFromDataPtr(dstName);
	if (!name || !*name)
		name = NULL;
	oP = doForkObject(sP, name);
	return(ELF_CREATE_OBJECT(oP->dex,oP->uid,NULL));
}

/* TAKE THIS OUT
static data_ptr
copyObject (srcObj,dstName,is_turbo,soft_snap)
{
	data_ptr dstObj;
	int retcode;

	retcode = ElfCallByName("COPY_OBJECT$", &dstObj, 4, srcObj, dstName,
							MK_INTEGER(is_turbo), BTRUE);
	if (retcode != 0) {
		ELF_DATA_FREE(dstObj);
		return(NullDataPtr);
	}
	return(dstObj);
}
*/

static data_ptr
copyObject (srcObj,dstName,flags,soft_snap)
data_ptr srcObj;
char *dstName;
data_ptr flags;
int soft_snap;
{

	data_ptr info;
	data_ptr dstObj;
	char *d_name;
	data_ptr fastCopyObject();
	int is_turbo = AxIntFromArray(flags,0);

	d_name = (char *)NULL;

	if (!ElfbIsObject(srcObj)) 
		return (NULL);
	
	if (0 == strcmp(dstName, ""))
		d_name = ElfObjPtr(srcObj)->nameTag;
	else
		d_name = dstName;

	info = objectStructureFromObject(srcObj, is_turbo, soft_snap);
	info = WriteArray(info, NAME_dex, ELF_CREATE_STRING(d_name));	
	dstObj = ObjectCreateFromStructure(info, FALSE, NULL, flags);	
/* 	dstObj = fastCopyObject(srcObj,is_turbo,TRUE); */
	objSnapObjects(dstObj, soft_snap);
	return(dstObj);
}

static data_ptr
fastCopyObject(srcObj,is_turbo,is_topLevel)
data_ptr srcObj;
int is_turbo, is_topLevel;
{
	data_ptr name;
	data_ptr inherits_from;
	data_ptr source;
	data_ptr exec_image;
	data_ptr kids;
	data_ptr theObj;
	char filename[STRPATH+1];
	register int ix, num_kids;
	data_ptr source_fileinfo;
	data_ptr locals;

	name = ELF_CREATE_STRING(ElfObjPtr(srcObj)->nameTag);
	inherits_from = objInheritFrom(srcObj);
	ElfbCreateObject(inherits_from, name, NULL);
	theObj = ElfRetData;
	if (is_topLevel)
		objSetAppName(theObj, name, 3);	/* in read */
	source = objGetObjvar(srcObj, NXLT("SOURCE_ARRAY$$"));
	exec_image = objGetObjvar(srcObj, NXLT("EXEC_IMAGE$$"));
	if (ElfbIsArray(source) && AX(source)->size>0) {
		if (is_turbo)
			loadMethodsFile(theObj, source, NULL);
		objSetObjvar(theObj, "SOURCE_ARRAY$$", source);
	}
	else if ( (BTRUE == ElfbIsArray(AnyFromArray(exec_image, 0))) &&
			  (BTRUE == ElfbIsBinary(AnyFromArray(exec_image,5))) ) {
		data_ptr fileinfo, new_image;
		
		new_image = ELF_DATA_COPY(exec_image);
		WriteArray(AnyFromArray(new_image,0),1,
					elfObjMakeUidFilename(StrFromDataPtr(name)));
		loadExecImage(theObj,new_image);
		objSetObjvar(theObj,"EXEC_IMAGE$$", new_image);
		objSetObjvar(theObj, "EXTERNAL_SOURCE_FILE$$", NULL);
		ELF_DATA_FREE(new_image);
	}

	kids = objChildren(srcObj);
	num_kids = ElfbArraySize(kids);
	for(ix=0;ix<num_kids;++ix) {
		data_ptr kid;
		
		kid = fastCopyObject(AnyFromArray(kids, ix),is_turbo,FALSE);
		objAddChild(theObj, kid);
		ElfObjPtr(kid)->parentP = ElfObjPtr(theObj);
		ELF_DATA_FREE(kid);
    }

	filename[0] = '\0';
	locals = objLocalStaticDatums(srcObj, NULL);
	objSetLocalDatums(theObj, locals);
	if (!ElfbIsArray(exec_image)) {
		source_fileinfo = objGetObjvar(theObj, "EXTERNAL_SOURCE_FILE$$");
		strcpy(filename, StrFromArray(source_fileinfo, 1));
		strcat(filename, StrFromArray(source_fileinfo, 0));
		if (filename[0]) {
			ElfbReadFile(filename);
			source = ElfRetData;
			if (is_turbo)
				loadMethodsFile(theObj, source, filename);
			objSetObjvar(theObj, "EXTERNAL_SOURCE_FILE$$", source_fileinfo);
			objSetObjvar(theObj, "SOURCE_ARRAY$$", source);
			ELF_DATA_FREE(source);
		}
	}
	if (is_turbo)
		objSetObjvar(theObj, "SOURCE_ARRAY$$", NULL);

	return(theObj);
}	
	
static data_ptr
objectStructureFromObject(theObj, turbo, soft_snap)
data_ptr theObj;
int turbo, soft_snap;
{

	data_ptr rval;
	data_ptr source;
	data_ptr kidInfo, exec_image, kids, kid;
	data_ptr builder_turbo, tmp;
	int ix;

#define setMember(dex, val) rval = WriteArray(rval, dex, val)

	rval = ELF_CREATE_ARRAY(0);
	kidInfo = ELF_CREATE_ARRAY(0);
	
	setMember(NAME_dex, ELF_CREATE_STRING(ElfObjPtr(theObj)->nameTag));
	setMember(INHERIT_dex, objInheritFrom(theObj));
	source = objGetObjvar(theObj, NXLT("SOURCE_ARRAY$$"));
	if (turbo || !ElfbIsArray(source) || (AX(source)->size == 0)) {
		exec_image = objGetObjvar(theObj, NXLT("EXEC_IMAGE$$"));
		ElfbGetSysVar(NXLT("bld_builder_turbo$$"));
		builder_turbo = ElfRetData;
		if ( -1 == IntFromDataPtr(builder_turbo)) {		/* Elf True = -1 */
			tmp = AnyFromArray(exec_image, 0);
			tmp = WriteArray(tmp, 2, NULL);
			exec_image = WriteArray(exec_image, 0, tmp);
		}
		setMember(IMAGE_dex, exec_image);
	}
	else {
		if (!objGetObjvar(theObj, NXLT("EXTERNAL_SOURCE_FILE$$")))
			setMember(SOURCE_dex, source);
	}

	kids = objChildren(theObj);
	ix = 0;
	kid = AnyFromArray(kids, ix);
	while (kid) {
		kidInfo = WriteArray(kidInfo, ix, objectStructureFromObject(kid, turbo, soft_snap));
		ix = ix+1;
		kid = AnyFromArray(kids, ix);
	}
	
	setMember(CHILDINFO_dex, kidInfo);

	setMember(LOCALS_dex, objLocalStaticDatums(theObj, NULL));

	return(rval);
}

static data_ptr
elfCreateObjectFromArray(dp , turbo)
data_ptr dp;
int turbo;
{

}


static 
objSetAppName(obj, passedName, is_top_level)
data_ptr obj;
data_ptr passedName;
int is_top_level;
{
	register elf_object *oP;
	char *name = StrFromDataPtr(passedName);
	
	oP = ElfObjPtr(obj);
	oP->appName = elfHashName(name);
	if (is_top_level) {
		oP->isTopLevelApp = is_top_level;
		addObjectToAppObjTable(oP);
	}
	return;
}

static void
objAddClassToAppModels(appObj,classObj)
elf_object *appObj;
data_ptr classObj;
{
	elf_object *classP = ElfObjPtr(classObj);
	classP->appName = appObj->appName;
	appObj->numModels++;
	appObj->models = (data_ptr*)TrueRealloc(appObj->models,
									(appObj->numModels * sizeof(data_ptr))) ;
	appObj->models[appObj->numModels-1] = AxTaskCopyElfData(0,classObj);
	return;
}


static
objSetObjvar(obj, name, val)
data_ptr obj;
char *name;
data_ptr val;
{
	register elf_object *oP;
	register elf_obj_grp *p;

	oP = ElfObjPtr(obj);
	p = getLocalData(oP, name, hash(name));
	if (p && p->value) {
		AxTaskFreeElfData(0,p->value);
		p->value = AxTaskCopyElfData(0,val);	
	}
	else
		crtLocalData(oP, name, val);
	return;
}


static data_ptr
objGetObjvar(obj, name)
data_ptr obj;
char *name;
{
	register elf_object *oP;
	data_ptr rval;
	register elf_obj_grp *p;

	rval = NULL;

	oP = ElfObjPtr(obj);
	p = getLocalData(oP, name, hash(name));
	if (p) {
		rval = ElfDataCopy(p->value);
	}

	return(rval);
}

static 
objAddChild(obj, kid)
data_ptr obj, kid;
{
	register elf_object *oP;

	oP = ElfObjPtr(obj);
	ElfObjPtr(kid)->appName = oP->appName;
	oP->numChildren++;
	oP->children = (data_ptr*)TrueRealloc(oP->children,
						(oP->numChildren * sizeof(data_ptr)));
	oP->children[oP->numChildren-1] = AxTaskCopyElfData(0,kid);
	
	return;
}

static 
objSetLocalDatums(obj, locals)
data_ptr obj, locals;
{
	register int ix;
	register char *name;
	register data_ptr aP;
	register elf_obj_grp *p;
	register elf_object *oP;

	oP = ElfObjPtr(obj);
	if (!locals || !ElfbIsArray(locals))
		return;
	for (ix = AX(locals)->size-1; ix >= 0; --ix) {
		aP = AnyFromArray(locals,ix);
		name = StrFromArray(aP,0);
		p = ElfObjData(oP, name, hash(name));
		AxTaskFreeElfData(0,p->value);
		p->value = AxTaskCopyElfData(0,AnyFromArray(aP,1));
	}
	return;
}

static data_ptr
objInheritFrom(obj)
data_ptr obj;
{
	register elf_object *oP;
	register elf_object *parentP;
	data_ptr rval;

	rval = NULL;

	oP = ElfObjPtr(obj);
	if(oP->inheritFromP) {
		parentP = oP->inheritFromP;
		rval = ELF_CREATE_OBJECT(parentP->dex, parentP->uid, NULL);
	}

	return(rval);
	
}

static data_ptr
objChildren(obj)
data_ptr obj;
{
	register int ix, jx;
	register elf_object *oP, *childP;
	data_ptr rval;

	rval = NULL;

	oP = ElfObjPtr(obj);
	rval = ELF_CREATE_ARRAY(0);
	for (jx = ix = 0; ix < oP->numChildren; ix++) {
		if (oP->uid == 1) { /* root object */
			childP = ElfObjPtr(oP->children[ix]);
			if (strrchr(childP->nameTag, '$'))
				continue;
		}
		rval = WriteArray(rval,jx,ElfDataCopy(oP->children[ix]));
		jx++;
	}

	return(rval);
}


static data_ptr
objLocalStaticDatums(obj, arg1)
data_ptr obj;
data_ptr arg1;
{
	register int ix;
	register elf_obj_grp *p, *q;
	register elf_object *oP, *o2;
	register data_ptr item;
	data_ptr rval;

	rval = NULL;
	oP = ElfObjPtr(obj);
	o2 = (ElfbIsObject(arg1)) ? ElfObjPtr(arg1) : NULL;
	rval = ELF_CREATE_ARRAY(0);
	for (ix = 0, p = oP->localData; p; p = p->next) {
		if (OBJVAR_TYPE_IS_STATIC(p->static_info) && p->value) {
			if (o2 && (q = getLocalData(o2,p->name,p->hashval))) {
				if (!ElfSameData(p->value, q->value)) {
					item = ELF_CREATE_ARRAY(2);
					item = ElfAddStrToArray(item,0,p->name);
					item = WriteArray(item,1,ELF_DATA_COPY(p->value));
					rval = WriteArray(rval, ix, item);
					ix++;
				}
			}
			else {
				item = ELF_CREATE_ARRAY(2);
				item = ElfAddStrToArray(item,0,p->name);
				item = WriteArray(item,1,ELF_DATA_COPY(p->value));
				rval = WriteArray(rval, ix, item);
				ix++;
			}
		}
	}
	return(rval);
}

static 
objSnapObjects(obj, snap_type)	
data_ptr obj;
int snap_type;
{

	register elf_object *oP;
	register int ix;
	
	oP = ElfObjPtr(obj);
	elfSnapAppObjects(oP,snap_type);
	if (IS_APP_IN_READ(oP))
		oP->isTopLevelApp = TRUE;
	for (ix = 0; ix < oP->numModels; ix++)
		deleteObjectFromAppObjTable(ElfObjPtr(oP->models[ix]));

	return;
}

/*
static data_ptr
createLibraryObjectFromStructure(libs, in_apptable, turbo, snaptype)
data_ptr libs;
int in_apptable;
int turbo;
int snapType;
{
	int ix;
	data_ptr lib, libObjs, content, ollDocinfo;
	char *filename, *tmpfile, *pathname;

	ix = 0;
	lib = AnyFromArray(libs, ix);
	
	while(lib) {
		content = member(CONTENT_dex, lib);
		filename = StrFromDataPtr(member(FILENAME_dex, lib));
		if (!content) {
			pathname = AxfFindElfFile(filename); *** CHECK THIS ***
			if (!pathname) {
				continue;
				*** ERROR ****
			}
			ollDocinfo = objectLibraryReadDoc(pathname, in_apptable, turbo, snapType);
		}
		else {
			if ElfbIsArray(content) {
				ollDocinfo = objectLibraryProcessInfo(content, in_apptable, turbo, snapType);
			}
			else ElfbIsBinary(content) {	
				tmpfile = AxfCreateTempfile();
				ElfbWriteBinaryFile(tmpfile, content);
				ollDocinfo = objectLibraryReadDoc(tmpfile, in_apptable, turbo, snapType);
				ElfbDelete(tmpfile);
		}
		
		ix = ix+1;
		lib = AnyFromArray(libs, ix);
		member(OBJTABLE_dex, ollDocinfo);
		libObjs = AxAddArrayToArray(libObjs, libObjTable);
	}

	return(libObjs);
}	
		 
*/		


/**** CORBA RELATED FUNCTIONS ****/

static elf_object *
elfFindCorbaModel(name)
data_ptr name;
{
	register ix;
	register char *appname;

	appname = elfHashName(StrFromDataPtr(name));	
	for(ix=0;ix<SizeAppObjTable;ix++) {
		if (AppObjTable[ix] && (AppObjTable[ix]->appName == appname) &&
			IS_CLASS_IN_READ(AppObjTable[ix]) && AppObjTable[ix]->ext_obj)
			return(AppObjTable[ix]);
	}
	return(NULL);
}

static void
elfCreateExternalObject(inheritFromP,objectNameP,bldrName)
data_ptr inheritFromP;
data_ptr objectNameP;
data_ptr bldrName;
{
	elf_object *oP, *cP;
	data_ptr builderName;	
	int status;
	char *err;

	ElfbCreateObject(ELF_CREATE_STRING("BaseClass"),bldrName,NULL);
    oP = ElfObjPtr(ElfRetData);
 
    if (ElfbIsString(inheritFromP))
        cP = elfLookupModel(StrFromDataPtr(inheritFromP));
    else
        cP = ElfObjPtr(inheritFromP);
    oP->inheritFromP = cP;

    status = 0;
    /* Now to actually create the corba object */
	oP->ext_obj = (void *)bldrExtCreateObject(	objectNameP,
												cP->ext_obj,
												&status,
												NULL
											  	);
	if (status == -1) {
		err = (char *)oP->ext_obj;
		oP->ext_obj = 0;
		ElfStrAbort(99,err,NULL);
	}
	return;
}

static void
elfCreateCorbaClass(classP, classInfo)
data_ptr classP;
data_ptr classInfo;
{
	elf_object *class;
	char *objectName, *serverName, *hostName;
	data_ptr className;
	int status = 0;

	class = ElfObjPtr(classP);
	if (AxIsArray(classInfo)) {
		/* ORBIX SPECIFIC - classInfo contains array of information
		   className, host, implementation, marker, IRHost, IRImplementation, IRMarker
		*/
		className = AxArrayElement(classInfo,0);
		class->ext_obj = (void *)bldrExtCreateClassObj(className, NULL, classInfo, &status);
		/* bldrUpdateClassInfo(class->ext_obj, name); */
	}
	else { 
		/* classInfo = className */
		class->ext_obj = (void *)bldrExtCreateClassObj(classInfo, NULL, NULL, &status);
	}
	return;
}	

data_ptr
ElfCallCorbaMethod(oP,mName,num_args)
elf_object *oP;
char *mName;
int num_args;
{
	data_ptr args, retP;
	register int ix;
	int status = 0;
	data_ptr PopDatum();

	args = ELF_CREATE_ARRAY(num_args);
	for (ix = 0; ix < num_args; ++ix)
		args = WriteArray(args, ix, ELF_DATA_COPY(PopDatum()));

	retP = bldrInvokeMethodOnExtObj( oP->ext_obj,
									mName,
									args,
									&status);
	ELF_DATA_FREE(args);
	if (status == -1) {
		ElfStrAbort(-1,AxStrFromDataPtr(retP),mName);	
	}
	else if (status == 1) {		/* Object Reference */

		/** THIS IS BEING DONE IN bldraxorb.cc now - 
		 ** NEEDED HERE FOR IONA SUPPORT 
		**/

		/* Create a Object and a Pseudo Class */
		elf_object *classP, *objP;
		char *interface = 0, *classStr = 0, *obj_name = 0;
		char buf[512];
		data_ptr className, objName, class_ptr;
		static int objId = 1;
		buf[0] = '\0';

		interface = AxStrPtrFromArray(retP,0);
		classStr  = strcat(interface, "Class");
		sprintf(buf,"%s_%d",interface,objId);
		objName	  = AxMakeStrData(-1, buf);
		classP = elfLookupModel(classStr);
		if (classP) {
			char *tmp;
			ElfbCreateObject(AxMakeStrData(-1, "CorbaClass"), objName, NULL);
			objP = ElfObjPtr(ElfRetData);
			objP->inheritFromP = classP;
			AxBinaryFromArray(retP,2,&tmp);
			memcpy(&objP->ext_obj, tmp, sizeof(tmp));
			return(ElfRetData);
		}
		else {
			className = AxMakeStrData(-1, classStr);
			return(NULL);
		}	
	}
	return(retP); 
}

/** Set up this elf-object to receive call backs from a server */
static void
elfRegisterObjectForCallbacks(objP, classP, addInfo)
data_ptr objP;
data_ptr classP;
data_ptr  addInfo;
{
	elf_object *oP = ElfObjPtr(objP);
	elf_object *cP = 0;
	char objId[16];

	if (oP->ext_obj) {
		/** Should error and return ?? */
		bldrExtDeleteObject(oP->ext_obj);
	}

	if (ElfbIsObject(classP)) {
		cP = ElfObjPtr(classP);
	}

	else {
		ElfStrAbort(99, XLT("Class should be an object."), oP->nameTag);
	}

	oP->ext_obj = bldrExtRegisterCallBack(oP->dex, oP->uid, NULL);
	bldrSetExtInheritance(oP->ext_obj, cP->ext_obj, NULL);
}

	
data_ptr
bldrExtElfObjectFromExtObj (info, extObj)
data_ptr info;
void *extObj;
{
   	elf_object *classP, *objP;
	char *interface = 0, *classStr = 0, *obj_name = 0;
	char buf[512];
	data_ptr className, objName, class_ptr;
	static int objId = 1;
	buf[0] = '\0';

	interface = AxStrPtrFromArray(info,0);
	sprintf(buf,"%sObj%d",interface,objId);
	objName   = AxMakeStrData(-1, buf);
	classStr  = strcat(interface, "Class");
	classP = elfLookupModel(classStr);
	if (classP) {
		char *tmp;
		ElfbCreateObject(AxMakeStrData(-1, "CorbaClass"), objName, NULL);
		objP = ElfObjPtr(ElfRetData);
		objP->inheritFromP = classP;
		objP->ext_obj = extObj;
		return(ElfRetData);
	}
	else {
		className = AxMakeStrData(-1, classStr);
		return(NULL);
        }
}

static void 
elfSetCorbaInvocationMode(dP, arg1)
data_ptr dP;
data_ptr arg1;
{
	elf_object *oP;
	int mode;

	oP = ElfObjPtr(dP);
	mode = IntFromDataPtr(arg1);
	bldrExtSetInvocationMode(oP->ext_obj, mode,NULL);
}

static data_ptr 
elfGetDeferredRequestResult(dP, arg1)
data_ptr dP;
data_ptr arg1;
{
	data_ptr retP;
	elf_object *oP;
	int status;

	status = 0;
	oP = ElfObjPtr(dP);
	retP = bldrDeferredInvocationResult(oP->ext_obj, arg1, &status,NULL);
	if (status == -1)
		ElfStrAbort(-1,AxStrFromDataPtr(retP),"");
	else if (status == 1) {
		/* 1. Create a Corba Class. Set CorbaInfo ptr in the corba_info member
		   2. Create a Corba Object - set the class info and object id.
		*/	
		/* retP is an array of the following
			1. Object Id - integer
			2. Corba Info ptr for the Class stuff.
		*/
	
		status = 0;	
	}
	return(retP);
}

void * 
AxCorbaObjectId(dP)
data_ptr dP;
{
	elf_object *oP = ElfObjPtr(dP);
	return((void *)oP->ext_obj);
}

static data_ptr
elfHelpOnExtObjMethod(dP, methodName, addInfo)
data_ptr dP, methodName, addInfo;
{
	data_ptr ret;
	elf_object *oP = ElfObjPtr(dP);
	ret = bldrExtHelpOnMethod(oP->ext_obj, methodName, addInfo);
	return(ret);
}

static data_ptr
elfExtStructsAsFormats(dP, arg1, addInfo)
data_ptr dP, arg1, addInfo;
{
	data_ptr ret;
	elf_object *oP = ElfObjPtr(dP);
	
	ret = bldrExtStructsAsFormats(oP->ext_obj,arg1,addInfo);
	return(ret);
}


int
elfExtMethodExists(ext_objP, name)
void *ext_objP;
char *name;
{
	return(bldrExtMethodExists(ext_objP, name));
}

/** THIS POKE CODE SHOULD MATCH ONE IN builder.elh */
#define POKE_CORBA_CALLBACK_ 12019
data_ptr
elfCallMethodOnObject(dex, uid, methodName, args)
int dex;
int uid;
char *methodName;
data_ptr args;
{
	data_ptr aP, callbackData;
	data_ptr objDataPtr;
	int pid;
	elf_object *oP = 0;
	elf_object *app = 0;

	/** Find the object from the dex,uid pair */
	oP = getObjectFromTable(dex);
	if (!oP || (oP->uid != uid))
		return;		/* should error back to the corba caller */
	
	/** Create and fill in the callback methodName and methodArgs */
	callbackData = ELF_CREATE_ARRAY(2);
	callbackData = WriteArray(callbackData,0,AxMakeStrData(-1, methodName));
	callbackData = WriteArray(callbackData,1,args);

	/** Fill in the Callback Object and callbackData */
	objDataPtr = ELF_CREATE_OBJECT(oP->dex,oP->uid,NULL);
	aP = ELF_CREATE_ARRAY(3);
	aP = WriteArray(aP,0,ElfDataCopy(objDataPtr));
	aP = WriteArray(aP,1,AxMakeIntData(0));				/** reply expected */
	aP = WriteArray(aP,2,callbackData);

	/** Find the task-id of the object */	
	/** The task id is stored on the app object */
	app = oP;
	while(app && !app->isTopLevelApp)
		app = app->parentP;

	if (!app)
		app = oP;
	pid = ThimPidFromUID(app->task_uid);

	/** Send the poke */
	SendElfData(pid,POKE_CORBA_CALLBACK_,aP);
	ELF_DATA_FREE(aP);

	/** return the object on which the poke was sent */
	return(objDataPtr);
}

/** terminate_event equivalent for external objects */
static void
elfExtAppTerminated(dP)
data_ptr dP;
{
	elf_object *oP = ElfObjPtr(dP);
	if (!oP || !oP->ext_obj)
		return;

	bldrExtAppTerminated(oP->ext_obj);
	return;
}
