/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: elrecord.c
   Applix ASC version: 36
   Date: Tue Feb  2 15:21:22 1999
*/
#include <varargs.h>
#include <axtime.h>
#if AXMS
#include <axtimev.h>
#endif
#include <elf_env.h>
#include <elhelper.h>
#include <adocio.h>
#include <fsrts.h>
#include <elrecarg.h>

typedef struct ep_info {
    int top_level;      /* if TRUE, part of top-level report */
    shrt *funcstarts;   /* actual funcstarts from stack_frame */
    char *source_name;  /* from package */
    int numCalls;       /* total number of calls */
    int totalTime;      /* total number of microseconds */
    int subs;   /* points to first of functions called by this function */
    int next;   /* next in hash-list or subs-list */
} ep_info;

typedef int (*QSORT_FUNCPTR)(const void*, const void*);

#include <elrecord.p>

#define MAX_CHARS 50
#define MAX_RECORD_DEX 8000 /* number of events we are willing to record */

data_ptr RecordedEvents;        /* Array of events that gets built up */

extern bint ElfInLearnMode;

char *ElfSymbolNameFromId();
extern int ElfLanguage;

static char *AlcStr(p)
char *p;
{
    return(strcpy(malloc(strlen(p)+1),p));
}

static char *Name(n)
int n;
{
    return(ElfSymbolNameFromId(n,ElfLanguage));
}

extern int ElfStdoutRecord; /* If TRUE, sends stuff to stdout while recording */
data_ptr TaskWriteArray();

#define RECSTRLEN 50    /* chars to record into a single call */

/*************************************************************************
function: ElfRecord - Record an event
*/
ElfRecord (va_alist)
va_dcl
{
    va_list ptr;
    int type;
    int ix, jx;
    int dex;
    data_ptr eventP;
    static int CharsInBuf;
    static wchar Chars[RECSTRLEN+1];
    int the_char;
    mbchar buf[1000];
    int int1, int2, int3;
    char *str;
    static int priorEvent = 0;

    /* fetch type */
    va_start(ptr);
    type = va_arg(ptr,int);
    if ( !ElfInLearnMode && type != ELFREC_START && type != ELFREC_EMIT_CHARS )
        return;

    if ( type == ELFREC_WCHAR ) {
        the_char = va_arg(ptr,int);
        if ( CharsInBuf == RECSTRLEN )
            ElfRecord(ELFREC_EMIT_CHARS);
        Chars[CharsInBuf++] = (wchar) the_char;
    }
    else if ( (CharsInBuf != 0) && (type != ELFREC_EMIT_CHARS) )
        ElfRecord(ELFREC_EMIT_CHARS);
    else if ( (type == ELFREC_EMIT_CHARS) && (CharsInBuf == 0) )
        return;

    if ( type == ELFREC_END ) {
        ElfInLearnMode = FALSE;
        AxTaskFreeElfData(0,RecordedEvents);
        RecordedEvents = NULL;
    }
    else if ( type == ELFREC_START ) {
        ElfInLearnMode = BTRUE;
        AxTaskFreeElfData(0,RecordedEvents);
        RecordedEvents = NULL;
        return;
    }

    if ( !RecordedEvents ) {
        RecordedEvents = AxTaskCreateElfArray(0,0);
        priorEvent = 0;
    }

    dex = AX(RecordedEvents)->size;
    if ( dex >= MAX_RECORD_DEX ) {
        SuperMessage(XLT("Macro Recording is too long, select \"Record Macro\" to complete it."));
        return;
    }

    eventP = AxTaskCreateElfArray(0,1);
    eventP = TaskWriteArray(0,eventP,0,MK_INTEGER(type));

    switch ( type ) {
    case ELFREC_EMIT_CHARS:
        Chars[CharsInBuf] = 0;
        str = WToMbOneStr(Chars);
        CharsInBuf = 0;
        eventP = TaskWriteArray(0,eventP,1,AxTaskCreateElfString(0,str));
        RecordedEvents = TaskWriteArray(0,RecordedEvents,AX(RecordedEvents)->size,eventP);
        if ( ElfStdoutRecord )
            printf("TYPE@(\"%s\")\n",str);
        break;

    case ELFREC_VCALL:
        str = va_arg(ptr,char*);
        eventP = TaskWriteArray(0,eventP,1,AxTaskCreateElfString(0,str));
        RecordedEvents = TaskWriteArray(0,RecordedEvents,AX(RecordedEvents)->size,eventP);
        if ( ElfStdoutRecord )
            printf("%s\n",str);
        break;

    case ELFREC_WIN:
        str = va_arg(ptr,char*);
        /* ignore this if last action was select/create of same window title */
        if ( (priorEvent == ELFREC_WIN) || (priorEvent == ELFREC_CRTWIN) ) {
            data_ptr priorP;
            priorP = ReadArray(AX(RecordedEvents),AX(RecordedEvents)->size-1);
            if ( streq(StrFromArray(priorP,1),str) ) {
                AxTaskFreeElfData(0,eventP);
                break;
            }
        }
        eventP = TaskWriteArray(0,eventP,1,AxTaskCreateElfString(0,str));
        RecordedEvents = TaskWriteArray(0,RecordedEvents,AX(RecordedEvents)->size,eventP);
        if ( ElfStdoutRecord )
            printf("SELECT_WINDOW_BY_NAME@(%s)\n",str);
        break;

    case ELFREC_CRTWIN:
        str = va_arg(ptr,char*);
        eventP = TaskWriteArray(0,eventP,1,AxTaskCreateElfString(0,str));
        RecordedEvents = TaskWriteArray(0,RecordedEvents,AX(RecordedEvents)->size,eventP);
        if ( ElfStdoutRecord )
            printf("'\n' Window Created: %s\n'\n",str);
        break;

    case ELFREC_FCALL: {
        char *name;
        int arg_count;
        data_ptr *args;
        char **arg_names;
        char **format_names;
        int func_args;
        data_ptr argInfo, allArgInfo;
        package_info *packP;
        data_ptr value, aname, aformatname;

        name = va_arg(ptr,char*);               /* name */
        arg_count = va_arg(ptr,int);            /* arg count */
        args = va_arg(ptr,data_ptr*);           /* arg data */
        func_args = va_arg(ptr,int);            /* func arg count */
        arg_names = va_arg(ptr,char**);         /* func arg names */
        format_names = va_arg(ptr,char**);      /* func arg formats */
        packP = va_arg(ptr,package_info*);      /* package holding format desc */

        if ( ElfStdoutRecord ) {
            if ( arg_count == 0 )
                printf("%s",name);
            else
                printf("%s(",name);
        }

        eventP = TaskWriteArray(0,eventP,1,AxTaskCreateElfString(0,name));
        if ( packP )
            eventP = TaskWriteArray(0,eventP,2,
                        AxTaskCreateElfString(0,packP->source_name));
        allArgInfo = AxTaskCreateElfArray(0,0);
        for ( ix = 0; ix < arg_count; ix++ ) {
            argInfo = AxTaskCreateElfArray(0,1);
            value = AxTaskCopyElfData(0,args[ix]);
            if ( ElfStdoutRecord ) {
                if ( !value ) {
                    printf("null");
                }
                else if ( IS_INTEGER(value) ) {
                    printf("%d",IntFromDataPtr(value));
                }
                else {
                    switch ( value->type ) {
                    case TYPE_INTEGER:
                        printf("%d",IntFromDataPtr(value));
                        break;
                    case TYPE_NUMBER:
                        printf("%.15g",NumFromDataPtr(value));
                        break;
                    case TYPE_STRING:
                        printf("\"%s\"",StrFromDataPtr(value));
                        break;
                    case TYPE_ARRAY:
                        printf("array:%d",ElfbArraySize(value));
                        break;
                    case TYPE_BINARY:
                        printf("binary:%d",ElfbBinarySize(value));
                        break;
                    case TYPE_OBJECT:
                        printf("object");
                        break;
                    case TYPE_ENTRY:
                        printf("entry");
                        break;
                    default:
                        printf("???");
                        break;
                    }
                }
                if ( ix != arg_count-1 )
                    printf(",");
            }
            argInfo = TaskWriteArray(0,argInfo,0,value);
            if ( packP && (ix < func_args) ) {
                aname = AxTaskCreateElfString(0,arg_names[ix]);
                argInfo = TaskWriteArray(0,argInfo,1,aname);
                aformatname = AxTaskCreateElfString(0,format_names[ix]);
                argInfo = TaskWriteArray(0,argInfo,2,aformatname);
            }
            else {
                char tmp[20];
                sprintf(tmp,NXLT("ARG%d"),ix);
                aname = AxTaskCreateElfString(0,tmp);
                argInfo = TaskWriteArray(0,argInfo,1,aname);
                aformatname = AxTaskCreateElfString(0,"");
                argInfo = TaskWriteArray(0,argInfo,2,aformatname);
            }
            allArgInfo = TaskWriteArray(0,allArgInfo,ix,argInfo);
        }
        eventP = TaskWriteArray(0,eventP,3,allArgInfo);
        RecordedEvents = TaskWriteArray(0,RecordedEvents,AX(RecordedEvents)->size,eventP);
        if ( ElfStdoutRecord ) {
            if ( arg_count == 0 )
                printf("\n",name);
            else
                printf(")\n",name);
        }
        break;
    }

    }
    priorEvent = type;
}

#define EP_HASH_SZ 65536
extern int ElfProfiling;

ep_info *ProfBlocks;
int NumProfileBlocks;
int AllocedProfileBlocks;
short *ElfProfileHash;
static struct timeval StartTime, EndTime;

/*************************************************************************
function: ElfProfileStart - Enable profile counting
*/
void ElfProfileStart ()
{
    int ix;
    struct timezone cur_timezone;

    if ( ElfProfiling ) {
        free(ProfBlocks);
        free(ElfProfileHash);
    }
    NumProfileBlocks = 1;   /* 0'th never used */
    ElfProfileHash = (short*)calloc(EP_HASH_SZ,sizeof(short));
    AllocedProfileBlocks = 1000;
    ProfBlocks = (ep_info*)calloc(AllocedProfileBlocks,sizeof(ep_info));
    ElfProfiling = TRUE;
    gettimeofday(&StartTime, &cur_timezone);
    StartTime.tv_sec++;
}

/*************************************************************************
function: ElfProfileStop - Disable profile counting
*/
void ElfProfileStop ()
{
    struct timezone cur_timezone;

    if ( !ElfProfiling )
        return;
    gettimeofday(&EndTime, &cur_timezone);
    ElfProfiling = FALSE;
}

static int ep_cmp(const ep_info *p1, const ep_info *p2) {   /* reverse sorter */
    if ( p1->totalTime > p2->totalTime )
        return(-1);
    if ( p2->totalTime > p1->totalTime )
        return(1);
    return(0);
}

/*************************************************************************
function: ElfProfileReport - Create output file, reset counters
*/
void ElfProfileReport (char *filename)
{
    FILE *fp;
    int hash, dex, ix, ep_cnt, subcnt;
    ep_info *sorted, *subsorted;

    ElfProfileStop();
    fp = fopen(filename, "w");
    SVERIFY(fp, ErrnoErr(), filename);
    fprintf(fp,"Total Time: %d milliseconds\n",millis(microsecDiff(&EndTime,&StartTime)));

    fprintf(fp,"millisec    #  name\n");
    fprintf(fp,"========   ==  ====\n");
    fprintf(fp,"\n");
    fprintf(fp,"subs:          millisec    #  sub-name\n");
    fprintf(fp,"               --------   --  --------\n");

#if 0
    for ( dex = 0; dex < NumProfileBlocks; dex++ ) {
        if ( !ProfBlocks[dex].top_level )
            continue;
        postLine(fp,&ProfBlocks[dex],"");
        for ( ix = ProfBlocks[dex].subs; ix; ix = ProfBlocks[ix].next )
            postLine(fp,&ProfBlocks[ix],"                       ");
    }
#endif
    for ( ep_cnt = 0,ix = 1; ix < NumProfileBlocks; ix++ ) {
        if ( ProfBlocks[ix].top_level )
            ep_cnt++;
    }
    sorted = (ep_info*)malloc(ep_cnt*sizeof(ep_info));
    for ( ep_cnt = 0,ix = 1; ix < NumProfileBlocks; ix++ ) {
        if ( ProfBlocks[ix].top_level )
            sorted[ep_cnt++] = ProfBlocks[ix];
    }

    qsort(sorted,ep_cnt,sizeof(ep_info),(QSORT_FUNCPTR)ep_cmp);
    for ( dex = 0; dex < ep_cnt; dex++ ) {
        postLine(fp,&sorted[dex],"");
        if ( !sorted[dex].subs )
            continue;
        for ( subcnt = 0, ix = sorted[dex].subs; ix; ix = ProfBlocks[ix].next )
            subcnt++;
        subsorted = (ep_info*)malloc(subcnt*sizeof(ep_info));
        for ( subcnt = 0, ix = sorted[dex].subs; ix; ix = ProfBlocks[ix].next )
            subsorted[subcnt++] = ProfBlocks[ix];
        qsort(subsorted,subcnt,sizeof(ep_info),(QSORT_FUNCPTR)ep_cmp);
        for ( ix = 0; ix < subcnt; ix++ )
            postLine(fp,&subsorted[ix],"               ");
        free(subsorted);
    }
    free(sorted);
    fclose(fp);
    ElfProfileStart();
}

static void postLine(FILE *fp, ep_info *ep, char *leader) {
    char funcName[100];
    char rawFuncName[100];
    char *filename;

    if ( ep->totalTime < 0 )    /* ignore stuff started before profile_start@ */
        return;

    FetchNameOfFunction(rawFuncName, ep->funcstarts);
    if ( rawFuncName[0] == '!' )
        sprintf(funcName,"Set %s",&rawFuncName[1]);
    else if ( rawFuncName[0] == '?' )
        sprintf(funcName,"Get %s",&rawFuncName[1]);
    else
        strcpy(funcName,rawFuncName);
    filename = ep->source_name;
    if ( !filename ) {
        if ( ep->numCalls == 0 )
            fprintf(fp,"%s               %s\n",leader,funcName);
        else
            fprintf(fp,"%s%7d%6d  %s\n",leader,millis(ep->totalTime),
                ep->numCalls,funcName);
    }
    else {
        if ( ep->numCalls == 0 )
            fprintf(fp,"%s               %s in %s\n",leader,funcName,baseof(filename));
        else
            fprintf(fp,"%s%7d%6d  %s in %s\n",leader,millis(ep->totalTime),
                ep->numCalls,funcName,baseof(filename));
    }
}

static int millis(int mic) {
    return((mic+500)/1000);
}

static char *baseof(char *str) {
    char *p;
    p = strrchr(str,'/');
    if ( !p )
        p = strrchr(str,'\\');
    if ( p )
        return(p+1);
    else
        return(str);
}

/*************************************************************************
function: ElfProfileNote - Note a call
*/
void ElfProfileNote (stack_frame *fp)
{
    struct timeval current_time;
    struct timezone cur_timezone;

    /* don't track non-visible calls */
    if ( fp->packP ) {
        if ( !fp->packP->src_map )
            return;
    }
    else {
        if ( !fp->caller || !fp->caller->packP || !fp->caller->packP->src_map )
            return;
    }
    gettimeofday(&current_time, &cur_timezone);
    fp->time_in = microsecDiff(&current_time,&StartTime);
}

/*************************************************************************
function: ElfProfileExit - Perform routine exit point work
*/
void ElfProfileExit (stack_frame *fp)
{
    struct timeval current_time, timeIn;
    struct timezone cur_timezone;
    int timer, dex, cdex, ix;

    /* don't track non-visible calls */
    if ( !fp->time_in )
        return;
    if ( fp->packP ) {
        if ( !fp->packP->src_map )
            return;
    }
    else {
        if ( !fp->caller || !fp->caller->packP || !fp->caller->packP->src_map )
            return;
    }
    gettimeofday(&current_time, &cur_timezone);
    timeIn = StartTime;
    timeIn.tv_usec += fp->time_in;
    timer = microsecDiff(&current_time,&timeIn);

    /* update statistics for this call */

    dex = FindInfoBlock(fp->funcstarts,fp->packP);
    ProfBlocks[dex].numCalls++;
    ProfBlocks[dex].totalTime += timer;

    /* update caller's info */

    if ( fp->caller && fp->caller->funcstarts ) {
        cdex = FindInfoBlock(fp->caller->funcstarts,fp->caller->packP);
        for ( ix = ProfBlocks[cdex].subs; ix; ix = ProfBlocks[ix].next ) {
            if ( ProfBlocks[ix].funcstarts == fp->funcstarts )
                break;
        }
        if ( !ix ) {
            ix = AllocNewBlock(fp->funcstarts,fp->packP,FALSE);
            ProfBlocks[ix].next = ProfBlocks[cdex].subs;
            ProfBlocks[cdex].subs = ix;
        }
        ProfBlocks[ix].numCalls++;
        ProfBlocks[ix].totalTime += timer;
    }
}

/*************************************************************************
function: FindInfoBlock - find info block in ProfBlocks, create if does not exist
*/
static int FindInfoBlock (shrt *funcstarts,struct package_info *packP)
{
    long hash;
    int dex;

    hash = (int)(funcstarts);
    hash = (hash >> 2) & 0x0000FFFF;
    for ( dex = ElfProfileHash[hash]; dex; dex = ProfBlocks[dex].next){
        if ( ProfBlocks[dex].funcstarts == funcstarts )
            break;
    }
    if ( !dex ) {
        dex = AllocNewBlock(funcstarts,packP,TRUE);
        ProfBlocks[dex].next = ElfProfileHash[hash];
        ElfProfileHash[hash] = dex;
    }
    return(dex);
}

static int AllocNewBlock(shrt *funcstarts,struct package_info *packP,int top) {
    int dex;

    if ( NumProfileBlocks == AllocedProfileBlocks ) {
        AllocedProfileBlocks *= 2;
        ProfBlocks = (ep_info*)realloc(ProfBlocks,
                            AllocedProfileBlocks*sizeof(ep_info));
    }
    dex = NumProfileBlocks++;
    memset(&ProfBlocks[dex],0,sizeof(ep_info));
    ProfBlocks[dex].funcstarts = funcstarts;
    ProfBlocks[dex].source_name = (packP ? packP->source_name : NULL);
    ProfBlocks[dex].top_level = top;
    return(dex);
}

/*************************************************************************
function: microsecDiff - Subtract second time arg from first
Returns difference in microseconds
*/
int microsecDiff (t1P, t2P)
struct timeval *t1P, *t2P;
{
    return(((t1P->tv_sec - t2P->tv_sec)) * 1000000 + 
           (t1P->tv_usec - t2P->tv_usec));
}

