/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: elrpc1.c
   Applix ASC version: 98
   Date: Tue Feb  2 15:21:18 1999
*/
#include <stdlib.h>    /* for sys_errlist */
#include <sys/types.h>
#include <axtime.h>
#include <axsocket.h>
#include <errno.h>
#if AX_X11
#if (SUN4 && !SUN4S) || LINUX || HP800 || PMAX
/* SUN OS, et. al. don't have select.h */
#else
#include <sys/select.h>
#endif
#endif
#include <setjmp.h>
#include <aos.h>
#include <elfapi.h>
#include <axnet.h>
#include <getstr10.h>
#include <elrpc1.p>
#if !ELF_COMM_STAND_ALONE
#define THIM_STATIC 1
#include <thim_gen.h>
#endif


#define int32 int       /* change for 286 class machines */

extern int axErrorCode;
extern char *axErrorString;
extern char *axErrorObject;
extern jmp_buf axErrorHandler;

#if PMAX
#define EPROTO 71
#endif

#if AXMS && !ELF_COMM_STAND_ALONE
extern HWND WXDummyWnd;
#endif


int AxRPCNonBlockingIO; /* If TRUE, use AxFdTimedReadBytes() */
int AxnetNoResponse = FALSE;/* Just used for fork/execing new tasks by axnet */
int AxPreventLowCodesFromClient;       /* keep client from sending <1 funcId */
int AxTimeoutOnReads = 0; /* set >0 for num seconds to timeout a client */

void axSendIt();
char *ErrnoTextFromUnix();


#if ELF_COMM_STAND_ALONE    /* To keep from doing FreeTempData in axmain */
#ifndef streq
#define streq(a,b) !strcmp(a,b)
#define TaskAlloc(t,n) malloc((n))
#define TaskFree(t,p) free((p))
#define TaskRealloc(t,p,n) realloc(p, n)
#endif

/*************************************************************************
function: AxNetConnect - Connect to axnet server, return channel
*/
int xxAxNetConnect (hostNameIn,code,str,obj)
char *hostNameIn;
int *code;
char **str, **obj;
{
    static int called = FALSE;
    static char *pipeEnv = NULL;
    char *pipeFile;
	static char serviceName[20];

    if ( !called ) {
        pipeEnv = getenv("AXNETPIPE");
        called = TRUE;
		sprintf(serviceName,"%d",AxServicePort("axnet"));
    }
#if AX_X11
    pipeFile = "/tmp/.axnetipc";
#else
    /* Windows uses the port num. */
    pipeFile = NULL;
#endif
    return(
        NetConnect(hostNameIn,pipeFile,serviceName,pipeEnv,code,str,obj));
}

void (*AxBrokenPipeCB)(/* int fd */);

elfData AxDfltDispatch();
elfData AxDfltDispatchMsg();

/*************************************************************************
function: AxSelectWithTimeout - Returns >1 on success, -1 error, 0 timeout
*/
int AxSelectWithTimeout(callerfd,timeout)
int callerfd;
int timeout;
{
    struct timeval downtime;
    fd_set selectfd;
    int readyfds;

    downtime.tv_sec = timeout;
    downtime.tv_usec = timeout;
    FD_ZERO(&selectfd);
    FD_SET(callerfd,&selectfd);
    readyfds = select(FD_SETSIZE,&selectfd,NULL,NULL,&downtime);

    return(readyfds);
}

/*************************************************************************
function: AxSockTop - Get message, convert to array, call dispatcher
Returns 0 on success, -1 and errno I/O error, 1 on shutdown, 2 on no client
*/
int AxSockTop (fd)
int fd;     /* socket channel to get message from and write response to */
{
    return(AxMessageTop(fd,fd));
}

/*************************************************************************
function: AxSockTop - Get message, convert to array, call dispatcher
Returns 0 on success, -1 and errno I/O error, 1 on shutdown, 2 on no client
*/
int AxMessageTop(rfd,wfd)
int rfd;        /* channel to get message from */
int wfd;        /* channel to write response to */
{
    elfData aP,     /* passed to function */
            fP,     /* returned from function */
            rP,     /* sent back to Aster*x */
            errP;   /* on error, sent back to Aster*x */
    char *inBuf;    /* holds the message in from the socket */
    char *outBuf;   /* holds the message on the way out */
    int32 size;     /* size of socket message */
    int32 cnt;      /* counter */
    int32 quitNow = 0;  /* if set by Dispatcher, walk out */
    int32 processed = 0;
    int cmdCode;
    elfData subP;
    int seen_write_error = FALSE;
    int got = 0;
    int readyfds;
#if AXMS
    int is_svr_ch = FALSE;
    if ( AxIsSvrChannel(wfd) )
        is_svr_ch = TRUE;
#endif

    /* get message size, allocate buffer for message, read message */
    while ( got < sizeof(int32) ) {
        if ( AxRPCNonBlockingIO )
            cnt = AxFdTimedReadBytes(rfd,((char*)&size)+got, 
                    sizeof(int32)-got,-1,TRUE);
        else {

            if( AxTimeoutOnReads && rfd != -1 ){
try_again:
              readyfds = AxSelectWithTimeout(rfd,AxTimeoutOnReads);
              if( readyfds == -1 ){
                if( errno == EINTR )
                  goto try_again;
                return(-1);
              } else if( readyfds == 0 )
                return(2);
            }

            cnt = AxReadSomeSocketData( rfd, ((char*)&size)+got, 
                    sizeof(int32)-got );
        }
        if ( cnt == 0 )
            return(2);
        if ( cnt == -1 )
            return(cnt);
        got += cnt; 
    }   

    size = ntohl(size);
    inBuf = (char*)AxMalloc(size);
rd: if ( AxRPCNonBlockingIO )
        cnt = AxFdTimedReadBytes(rfd,inBuf+processed,size-processed,-1,TRUE);
    else
        cnt = AxReadSomeSocketData( rfd,inBuf+processed,size-processed);
    if ( cnt <= 0 )
        return(-1);
    if ( cnt != size-processed ) {
        processed += cnt;
        goto rd;
    }

/* convert to ELF data */

    aP = AxRPCRead(inBuf);

/* call customer supplied dispatcher */

    if ( setjmp(axErrorHandler) ) {
        rP = AxMakeArray(2);
        AxAddIntToArray(rP,0,1);
        errP = AxMakeArray(3);
        AxAddIntToArray(errP,0,axErrorCode);
        AxAddStrToArray(errP,1,axErrorString);
        AxAddStrToArray(errP,2,axErrorObject);
        AxAddArrayToArray(rP,1,errP);
    }
    else {
        cmdCode = AxIntFromArray(aP,0);
        subP = AxArrayFromArray(aP,1);
        if ( (cmdCode < 1) && AxPreventLowCodesFromClient )
            AxError(1,NXLT("Client sending invalid cmdCode to AxDispatch"),NULL);
        fP = AxDispatch(cmdCode,subP,&quitNow);
        if ( !fP && (cmdCode < 0) )
            fP = AxDfltDispatch(cmdCode,subP);

        /* respond to Aster*x */

        rP = AxMakeArray(2);
        AxAddIntToArray(rP,0,0);
        AxAddArrayToArray(rP,1,fP);
    }
    size = AxMemWrite(rP,0,&outBuf);
    if ( !AxnetNoResponse ) {
        if ( AxRPCNonBlockingIO )
            cnt = AxFdSocketWriter(wfd,outBuf,size);
        else {
#if AX_X11
            cnt = write(wfd,outBuf,size);
            if ((cnt < 0) && (errno == EPIPE)) {
                seen_write_error = TRUE;
                if (AxBrokenPipeCB) {
                    (void) (*AxBrokenPipeCB)(wfd);
                    /* This handler returned -- meaning this
                       server wants to keep running.  Set 'cnt' to 'size'
                       so that a 0 will be returned by this function.
                       FIXME, deal with other kinds of errors too! */
                    cnt = size;
                }
            }
#else
            /* Windows/NT  */
            /* if we came in here with a wfd that was the SvrSockChannel
             * (is_svr_ch) and SvrSockChannel has become invalid, we want
             * to make wfd invalid (this is because axnet may have started
             * a thread that uses SvrSockChannel and invalidated it for
             * non-thread use. On UNIX, a fork occurs and the parent
             * actually closes the channel, which we can't do on NT).  -donato
             */
            if ( is_svr_ch && !AxIsSvrChannel(wfd) )
                wfd = INVALID_SOCKET;
            cnt = send( wfd, outBuf, size, 0 );
            if ( cnt == SOCKET_ERROR ) {
                errno = WSAGetLastError();
                if ( errno == WSAENOTSOCK && AxBrokenPipeCB ) {
                    (void) (*AxBrokenPipeCB)(wfd);
                    cnt = size;
                }
                else
                    cnt = -1;
            }
#endif
        }
    }
    else
        {
        /* no response required, set things for error check below */
        cnt = size;
        }

    AxnetNoResponse = FALSE;

/* Release arrays and temporarily allocated data */

    AxFreeData(aP);
    AxFreeData(rP);
    AxFreeTempData();
    AxFree(inBuf);
    AxFree(outBuf);
    if ( cnt != size )  /* from the write */
        return(-1);
    return(quitNow);
}

elfData AxDfltDispatchMsg(chanP, cmdCode, subP)
axio_channel_t *chanP;
int cmdCode;
elfData subP;
{
	switch (cmdCode)
	{
		case ELFC_PACK_DATA:
			chanP->axio_flags.io_packed = 1;
			return(AxMakeIntData(1));
		default:
			return(AxDfltDispatch(cmdCode, subP));
	}
}

elfData AxDfltDispatch(cmdCode,subP)
int cmdCode;
elfData subP;
{
    elfData retval = NULL;
    int code; /* int on bsd, pid_t on sys V. */
    char *str;
 
        switch ( cmdCode ) {
        case ELFC_GET_PID:
    case ELFC_HS3:
#if AX_X11
                code = getpid();
#else
                code = GetCurrentProcessId();
#endif
                retval = AxMakeArray(0);
                retval = AxAddIntToArray(retval, 0, cmdCode);
                retval = AxAddIntToArray(retval, 1, code);
                break;
         }
         return( retval);
}
#endif /* ELF_COMM_STAND_ALONE To keep from doing FreeTempData in axmain */

/*************************************************************************
function: AxSendMessageToServer - Send data to a server
*/
void AxSendMessageToServer (channel,cmdcode,fP,code,str,obj)
int channel;
int cmdcode;
elfData fP;
int *code;
char **str, **obj;
{
    axSendIt(channel,AxMakeIntData(cmdcode),fP,code,str,obj);
}
void axSendIt (channel,cmdcode,fP,code,str,obj)
int channel;
elfData cmdcode;
elfData fP;
int *code;
char **str, **obj;
{
    elfData rP;
    int size, cnt;
    char *outBuf;   /* holds the message on the way out */

    rP = AxMakeArray(2);
#if !ELF_COMM_STAND_ALONE
    cmdcode = AxCopyData(cmdcode);/* prevents runtime from freeing caller data */
    fP = AxCopyData(fP);
#endif
    AxAddArrayToArray(rP,0,cmdcode);
    AxAddArrayToArray(rP,1,fP);
    size = AxMemWrite(rP,0,&outBuf);
    if ( AxRPCNonBlockingIO )
        cnt = AxFdSocketWriter(channel,outBuf,size);
    else
        {
#if AX_X11
        cnt = write(channel,outBuf,size);
#else
        /* Windows/NT  */
        cnt = send( channel, outBuf, size, 0 );
        if ( cnt == SOCKET_ERROR )
           {
           errno = WSAGetLastError();
           cnt = -1;
           }
#endif
        }

    AxFree(outBuf);
    AxFreeData(rP);
#if ELF_COMM_STAND_ALONE    /* To keep from doing FreeTempData in axmain */
    AxFreeTempData();
#endif
    if ( cnt != size ) { /* from the write */
        *code = ErrnoErr();
        *str = ErrnoTextFromUnix(errno);
        *obj = NXLT("AxSendMessageToServer");
    }
    else
        *code = 0;
}

/*************************************************************************
function: AxGetResponseFromServer - Get incoming message
*/
elfData AxGetResponseFromServer (channel,codeP,strP,objP)
int channel;
int *codeP;             /* OUT: 0 or eror code */
char **strP, **objP;    /* OUT: only if error */
{
    elfData aP;      /* passed to function */
    char *inBuf;    /* holds the message in from the socket */
    int32 size;     /* size of socket message */
    int32 cnt;      /* counter */
    elfData subP;
    int code;
    int processed = 0;
    static char *errStr, *objStr;
    int got = 0;

    /* get message size, allocate buffer for message, read message */
    while ( got < sizeof(int32) ) {
        if ( AxRPCNonBlockingIO )
            cnt = AxFdTimedReadBytes(channel,((char*)&size)+got, 
                    sizeof(int32)-got,-1,TRUE);
        else
            cnt = AxReadSomeSocketData( channel, ((char*)&size)+got, 
                    sizeof(int32)-got );

        if ( cnt == 0 ) {
            *codeP = FEEPIPE;
            *objP = NXLT("server");
            *strP = NXLT("connection closed");
            return(NULL);
        }

        if ( cnt == -1 ) {
            *codeP = ErrnoErr();
            *objP = NXLT("server");
            *strP = ErrnoTextFromUnix(errno);
            return(NULL);
        }
        got += cnt; 
    }   

    size = ntohl(size);
    inBuf = (char*)AxMalloc(size);
rd: if ( AxRPCNonBlockingIO )
        cnt = AxFdTimedReadBytes(channel,inBuf+processed,size-processed,-1,TRUE);
    else
        cnt = AxReadSomeSocketData(channel,inBuf+processed,size-processed);
    if ( cnt <= 0 ) {
        *codeP = ErrnoErr();
        *objP = NXLT("server");
        *strP = ErrnoTextFromUnix(errno);
        return(NULL);
    }
    if ( cnt != size-processed ) {
        processed += cnt;
        goto rd;
    }

/* convert to ELF data */

    aP = AxRPCRead(inBuf);

    AxFree( inBuf );       /* all done with this */
    inBuf = (char *)NULL;

    code = AxIntFromArray(aP,0);
    if ( code == 0 ) {
        *codeP = 0;
        *strP = NULL;
        *objP = NULL;
        subP = AxArrayFromArray(aP,1);
        AxFreeArrayButNotContents(aP);
        return(subP);
    }
    else {
        /* release any prior error message buffers */
        if ( errStr ) { TaskFree(0,errStr); errStr = NULL; }
        if ( objStr ) { TaskFree(0,objStr); objStr = NULL; }
        subP = AxArrayFromArray(aP,1);
        *codeP = AxIntFromArray(subP,0);
        errStr = AxStrFromArray(subP,1);
        errStr = strcpy(TaskAlloc(0,strlen(errStr)+1),errStr);
        *strP = errStr;
        objStr = AxStrFromArray(subP,2);
        objStr = strcpy(TaskAlloc(0,strlen(objStr)+1),objStr);
        *objP = objStr;
        AxFreeData(aP);
        return(NULL);
    }
}

/*************************************************************************
function: ErrnoTextFromUnix - Read Unix errno text
*/
char *ErrnoTextFromUnix (err)
int err;
{
    static char *buf;
#if !LINUX
    extern char *sys_errlist[];
    extern int sys_nerr;
#endif

    if ( err < sys_nerr )
        return((char*)sys_errlist[err]); /* cast gets rid of 'const' warning */
    if ( !buf )
        buf = malloc(300);
    sprintf(buf,"err %d",err);
    return(buf);
}

char *WhereAmI();

static char ExecFile[STRPATH+1];

/*************************************************************************
function: AxNetSetServerPath - Load static var with pathname to axnet server
*/
void AxNetSetServerPath (path)
char *path;
{
    strcpy(ExecFile,path);
}

/*************************************************************************
function: AxServicePort - Fetch port number from /etc/services
*/
int AxServicePort (char *serviceName) {
    struct servent *serverInfo;
	int port;

	port = atoi(serviceName);	/* all numeric services don't lookup */
	if ( port != 0 )			/* just use name as port number */
		return(port);
	serverInfo = getservbyname(serviceName,NXLT("tcp"));
	if ( !serverInfo )
		return(-1);
	port = ntohs(serverInfo->s_port);
	return(port);
}

/*************************************************************************
function: NetConnect - Connect to axnet/javanet server, return channel
*/
int NetConnect (hostNameIn,pipeFile,serviceName,pipeEnv,code,str,obj)
char *hostNameIn;
char *pipeFile;
char *serviceName;
char *pipeEnv;
int *code;
char **str, **obj;
{
    char *hostName = hostNameIn;
    int channel;
    char buf[STRPATH+100];
    int retryCount = 0; /* Handle simultanious server startup */
    char *pipe;
    int port = 0;
    static char *ThisHostName;
    char *p;

    pipe = pipeEnv;

/* Treat connects to this hostname same as local */

    if ( hostName && hostName[0] ) {
        if ( !ThisHostName ) {
            if ( 0 == gethostname(buf,STRPATH) ) {
                ThisHostName = malloc(strlen(buf)+1);
                strcpy(ThisHostName,buf);
            }
        }
        if ( ThisHostName && streq(hostName,ThisHostName) )
            hostName = NULL;
    }

    if ( !pipe && (!hostName || !hostName[0]) && pipeFile )
        pipe = pipeFile;

retry:
    if ( pipe )
        channel = AxClientConnect(0,pipe,NULL,code,str,obj);
    else {
		port = AxServicePort(serviceName);
		if ( port == -1 ) {
            *str = NXLT("not a tcp service in /etc/services");
            *obj = serviceName;
            *code = ENOENT;
            return(-1);
        }
        channel = AxClientConnect(port,NULL,hostName,code,str,obj);
    }
    if ( *code != 0 ) {
#if AX_X11                                /*  {  */
        if ( ((*code == FEECONNREFUSED)
                    || (*code == FEEPERM)
                    || (*code == FEEPROTO)
                    || (*code == FEENOENT))
            && ExecFile[0]
            && (!hostName || !hostName[0]) ) {
            if ( pipe )
                sprintf(buf,"%s %s -fork",ExecFile,pipe);
            else
                sprintf(buf,"%s %d -fork",ExecFile,port);
            *code = AxCallProcess(buf);
            if ( *code != 0 ) {
                retryCount++;
                if ( retryCount == 1 )
                    goto retry;
                *str = ErrnoTextFromUnix(*code);
                *obj = ExecFile;
                return(-1);
            }
            goto retry;
        }
#else                   /*  }  { */
        /* Window NT  */
#if !ELF_COMM_STAND_ALONE
        if ( (retryCount == 0) && (*code == WSAECONNREFUSED) &&
             (strcmp(serviceName,"axnet") == 0) &&
             (hostName == (char *)NULL || (hostName && !hostName[0])) )
           {
           /* request was for axnet on this machine, which failed with a */
           /* "connection refused".  THis means that axnet is not running */
           /* try to start it                                             */
           char axnet_cmd[500], *p;
           extern int ThimApplixDir;

           sprintf(axnet_cmd, "\"%s\\axnet.exe\" -x -fork", ThimApplixDir);
           AxRunProgram( axnet_cmd, NULL, NULL, NULL, FALSE );
           Sleep( 500 );          /* sleep 1/2 second */
           retryCount++;
           goto retry;
           }
#endif
#endif               /*  }  */
           
    }
    return(channel);
}

/*************************************************************************
function: AxNetConnect - Connect to axnet server, return channel
*/
int AxNetConnect (hostNameIn,code,str,obj)
char *hostNameIn;
int *code;
char **str, **obj;
{
    static int called = FALSE;
    static char *pipeEnv = NULL;
    char *pipeFile;
	static char serviceName[20];

    if ( !called ) {
        pipeEnv = getenv("AXNETPIPE");
        called = TRUE;
		sprintf(serviceName,"%d",AxServicePort("axnet"));
    }
#if AX_X11
    pipeFile = "/tmp/.axnetipc";
#else
    /* Windows uses the port num. */
    pipeFile = NULL;
#endif
    return(
        NetConnect(hostNameIn,pipeFile,serviceName,pipeEnv,code,str,obj));
}

/*************************************************************************
function: JavaNetConnect - Connect to javanet server, return channel
*/
int JavaNetConnect (hostNameIn,code,str,obj)
char *hostNameIn;
int *code;
char **str, **obj;
{
    return(NetConnect(hostNameIn,NULL,"aagtwy",NULL,code,str,obj));
}


/*************************************************************************
function: AxNetDropServiceBySocket - Tell axnet to unregister a service
*/
void AxNetDropServiceBySocket (channel,sockname,code,str,obj)
int channel;
char *sockname;
int *code;
char **str, **obj;
{
    elfData request;

    request = AxMakeStrData(-1,sockname);
    AxSendMessageToServer(channel,AXNET_DROP_SOCKET,request,code,str,obj);
    if ( *code != 0 )
        return;
    AxGetResponseFromServer(channel,code,str,obj);
}

/*************************************************************************
function: AxNetServices - Return elf array of service names
Note that the returned array is unstable - will be freed at next call.
*/
elfData AxNetServices (hostName,qualifier,code,str,obj)
char *hostName;
char *qualifier;
int *code;
char **str, **obj;
{
    int channel;
    elfData request, response;

    channel = AxNetConnect(hostName,code,str,obj);
    if ( *code != 0 ) return(NULL);
    request = AxMakeStrData(-1,qualifier);
    AxSendMessageToServer(channel,AXNET_LIST_SERVICES,request,code,str,obj);
    if ( *code != 0 ) return(NULL);
    response = AxGetResponseFromServer(channel,code,str,obj);
    if ( *code != 0 ) return(0);
    AxClientDisConnect(channel);
    return(response);
}

/*************************************************************************
function: AxNetCall - Send message to Elf server
*/
elfData AxNetCall (hostName,serviceName, serviceArgs, passwd, code,str,obj)
char *hostName, *serviceName;
elfData serviceArgs;
char *passwd;
int *code;
char **str, **obj;
{
    int channel;
    elfData request, response;
    char sockName[STRPATH+1];

/* find axmain */

    channel = AxNetConnect(hostName,code,str,obj);
    if ( *code != 0 )
        return(NULL);
    request = AxMakeArray(2);
    request = AxAddStrToArray(request,0,serviceName);
    request = AxAddIntToArray(request,1, hostName != NULL);
    AxSendMessageToServer(channel,AXNET_FIND_SERVICE,request,code,str,obj);
    if ( *code != 0 ) return(NULL);
    response = AxGetResponseFromServer(channel,code,str,obj);
    if ( *code != 0 ) {
        AxClientDisConnect(channel);
        return(NULL);
    }

/* send message to axmain */

    if ( !response ) {
        channel = channel;
    }
    else {
        strcpy(sockName,AxStrFromArray(response,0));
        AxClientDisConnect(channel);
        channel = AxClientConnect(0,sockName,hostName,code,str,obj);/* connect to axmain */
        if ( *code != 0 ) return(NULL);
    }
    axSendIt(channel,AxMakeStrData(-1,serviceName),serviceArgs,code,str,obj);
    if ( *code != 0 ) return(NULL);
    response = AxGetResponseFromServer(channel,code,str,obj);
    AxClientDisConnect(channel);
    return(response);
}

/*************************************************************************
function: AxNetProc - Start a server --- UnPublished!!!
*/
void AxNetProc (hostName,argvP,code,str,obj)
char *hostName;
elfData argvP;
int *code;
char **str, **obj;
{
    int channel;
    elfData request, response;

    channel = AxNetConnect(hostName,code,str,obj);
    if ( *code != 0 )
        return;
    request = AxCopyData(argvP);
    AxSendMessageToServer(channel,AXNET_PROC,request,code,str,obj);
    if ( *code != 0 ) return;
    response = AxGetResponseFromServer(channel,code,str,obj);
    AxClientDisConnect(channel);
}

/*************************************************************************
function: AxNetProcSlave - Start a slaved server
*/
int AxNetProcSlave (hostName,argvP,code,str,obj)
char *hostName;
elfData argvP;
int *code;
char **str, **obj;
{
    int channel;
    elfData request, response;

    channel = AxNetConnect(hostName,code,str,obj);
    if ( *code != 0 )
        return(-1);
    request = AxCopyData(argvP);
    AxSendMessageToServer(channel,AXNET_PROC_SLAVE,request,code,str,obj);
    if ( *code != 0 ) return;
    response = AxGetResponseFromServer(channel,code,str,obj);
    if ( *code != 0 ) {
        AxClientDisConnect(channel);
        return(-1);
    }
    return(channel);
}



 
void BadExit( exit_code )
int exit_code;
{
#if AXMS
    /* make sure that we clean up any socket stuff */
    WSACleanup();
#endif
 
    exit(exit_code);
}





int AxReadSomeSocketData( fd, buffP, size_buf )
int fd;
char *buffP;
int size_buf;
{
    int cnt;
    int err_code;

#if AX_X11      /*  {  */
try_again:
    cnt = read( fd, buffP, size_buf );
    if (cnt < 0) {
        /* Note: On SIEMENS/SINIX, the above read returns -1 with
           errno is set to one of below when the client end does a
           shut-down.  All other platforms return 0.
           Therefore, sense the condition of the other side (client)
           doing a shut-down, and return a 0.
           The calling code treats a 0 return from here to mean
           'client went away'.  -1 is returned for all other error
           conditions. */
        if (errno == ECONNRESET || errno == ECONNABORTED)
            cnt = 0;
        if (errno == EWOULDBLOCK)
            goto try_again;
        if (errno == EINTR)
            goto try_again;
    }
#else           /*  }  {  */

#if !ELF_COMM_STAND_ALONE
    /* turn off notifier while we are reading */
    WSAAsyncSelect( fd, WXDummyWnd, 0, 0 );
#endif

    err_code = 0;

#if ELFAPI_WIND   /* { */
    /* we must process windows messages, rather than wait on a 'recv' */
    /* This is required by the ODBC gateway, since the MS ODBC driver */
    /* creates a background, top-level window, and unless we process  */
    /* messages for that window, other things in the system (like a DDE */
    /* client, will hang waiting for an ACK/NACK on the DDE initial     */
    /* broadcast message                                                */
    cnt = ReadSockethandleMessages( fd, buffP, size_buf, &err_code );

#else    /*  }  {  */

    cnt = recv( fd, buffP, size_buf, 0 );
    if ( cnt == SOCKET_ERROR )
       {
       err_code = WSAGetLastError();
       }
#endif   /* }  */

    if ( err_code != 0 )
       {
       if ( (err_code == WSAENETDOWN) || (err_code == WSAESHUTDOWN) ||
            (err_code == WSAECONNABORTED) || (err_code == WSAECONNRESET)||
            (err_code == WSAENOTSOCK) )
           {
           /* The calling code treats a 0 return from here to mean     */
           /* 'client went away'.  -1 is returned for other error conditions.*/
           cnt = 0;
           }
       else
           /* more serious error */
           cnt = -1;
       }

#if !ELF_COMM_STAND_ALONE
    /* now turn the notifier back on.  If there is more data in the */
    /* socket, this will generate another message, so we will get back */
    /* here.                                              */
    SetSocketAsyncCallback( fd );     /* restore async messages */
#endif
#endif           /*  }  */

    return( cnt );
}




#if ELFAPI_WIND     /*  {  */

/* This code will handle windows messages for this thread, and will read */
/* from the socket only when there is data available.  This is done so   */
/* that the windows which belong to this thread get servcied.            */
/* This setup is required for the NT ODBC driver, which creates a background */
/*   top level window.   If messages to that window are not serviced, and    */
/*   system broadcasts which wait for an ACK/NACK (like a DDE client coming  */
/*   up) will hang, because the window message is not handled.  In the case  */
/*   of the ODBC driver, the background window may not be created the first  */
/*   thime this gets created, which accound for the "look for a window" code */
/*   below.   - jon                                                          */

#define SOCKET_PSEUDO_EVENT (WM_USER + 5)

#define MS_GetMessage GetMessage
#define MS_GetCurrentThreadId GetCurrentThreadId
#define MS_GetCurrentProcessId GetCurrentProcessId


LRESULT CALLBACK wnd_proc (HWND hwnd, UINT message, UINT wParam, LONG lParam)
{
    return DefWindowProc (hwnd, message, wParam, lParam) ;
}


extern HANDLE WinMain_hInstance;     /* see "winmain.h"   */
char *class_title = "ENT_ELF_ASYNC_HOOK";

int ReadSockethandleMessages( int fd, char *buffP, int size_buf, int *err_code )
{
    static HWND thread_window = NULL;
    int our_pid, our_thread, cnt;
    int thread_id, process_id;
    HWND winD;
    MSG msg;
    int ret;
    long flags;
    unsigned long available;
    static int dbg;
    int event_type, event_error;
    WNDCLASS    wndclass;


    *err_code = 0;

    if ( thread_window == NULL )
        {
        /* we do not yet have a window to send message to, see if this */
        /* thread owns a window.  We only do this so many times by the way */
        /* figuring that if we read 100 things from the socket, and a      */
        /* background window still has not been created, then it won't ever */
        /* be created and we can default to a blocking 'recv' call.        */
        

        our_pid = MS_GetCurrentProcessId();
        our_thread = MS_GetCurrentThreadId();

#if 0
        if ( !dbg )
           {
           char out_line[300];
           sprintf( out_line, "catch %d", our_pid );
           MessageBox((HWND)NULL, out_line, out_line, MB_OK );
           dbg = 1;
           }
#endif

        /* create a basic frame window */
        memset( &wndclass, 0, sizeof(wndclass) );
        wndclass.style         = CS_HREDRAW | CS_VREDRAW;
        wndclass.lpfnWndProc   = wnd_proc;
        wndclass.cbClsExtra    = 0 ;
        wndclass.cbWndExtra    = 0 ;
        wndclass.hInstance     = WinMain_hInstance;
        wndclass.hIcon         = NULL ;
        wndclass.hbrBackground = GetStockObject (WHITE_BRUSH);
        wndclass.lpszClassName = class_title;
    
        RegisterClass (&wndclass) ;
    
        /* create the container simple window */
        thread_window = CreateWindow( class_title, class_title,
                   (WS_OVERLAPPED | WS_CAPTION | WS_MINIMIZEBOX |
                                       WS_SYSMENU | WS_MAXIMIZEBOX),
                   CW_USEDEFAULT, CW_USEDEFAULT,
                   20, 20, 
                   NULL, NULL, WinMain_hInstance, NULL);

        if ( thread_window == (HWND)NULL )
            {
            ret = GetLastError();
            }
        else
            {
            /* set this window as the target for the socket messages */
            flags = (FD_READ | FD_CLOSE);
            ret = WSAAsyncSelect( fd, thread_window,
                                             SOCKET_PSEUDO_EVENT, flags );
            if (ret == SOCKET_ERROR )
                {
                ret = WSAGetLastError();
                thread_window = NULL;
                }
            }
        }

    if ( thread_window == NULL )
        {
        /* there is no background window, so we can do a blocking read */
        cnt = recv( fd, buffP, size_buf, 0 );
        if ( cnt == SOCKET_ERROR )
           {
           *err_code = WSAGetLastError();
           }
        return( cnt );
        }

    /** If we get here, it means that we are to run a message queue, and */
    /** read socket data only when informed that there is something available*/
    while ( ret = MS_GetMessage( &msg, NULL, 0, 0) )
        {
        if ( ret == -1 )
            {
            /* some error, like an invalid window handle */
            break;           /* will exit */
            }

        if ( msg.message != SOCKET_PSEUDO_EVENT )
            {
            TranslateMessage (&msg) ;
            DispatchMessage (&msg) ;
            continue;
            }

        event_type = WSAGETSELECTEVENT( msg.lParam );
        event_error = WSAGETSELECTERROR( msg.lParam );
            
        if ( event_type == FD_CLOSE  )
            {
            cnt = 0;
            *err_code = WSAECONNRESET;
            return(0);
            }
        if ( event_type != FD_READ )
            {
            /* should not happen, only a place to put a breakpoint */
            cnt = 0;
            }

        /* we have been informed that there is data to be read, read it  */
        cnt = recv( fd, buffP, size_buf, 0 );
        if ( cnt == SOCKET_ERROR )
            {
            ret = WSAGetLastError();
            if ( ret == WSAEWOULDBLOCK )
                /* the PSEUDO_SOCKET message was spurrious */
                continue;
            if ( (ret == WSAENETDOWN) || (ret == WSAESHUTDOWN) ||
                (ret == WSAEBADF) || (ret == WSAENOTSOCK) ||
                (ret == WSAECONNABORTED) || (ret == WSAECONNRESET) )
                {
                /* socket gone */
                cnt = 0;
                *err_code = WSAECONNRESET;
                return(0);
                }
            }
        /* successful read */
        return(cnt);
        }

    /* treat a null response from the GetMessage as "exit" */
    *err_code = WSAECONNRESET;
    return(0);
}

#endif              /*  }  */
