/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: elrpc3.c
   Applix ASC version: 32
   Date: Tue Feb  2 15:21:28 1999
*/
#include <sys/types.h>
#include <sys/stat.h>
#include <axsocket.h>
#include <errno.h>
#include <signal.h>

#define THIM_STATIC 1    /** do not redefine such things as "exit()"  */
#include <aos.h>
#include <axnet.h>
#include <elfapi.h>

#ifndef SIGCLD
#define SIGCLD SIGCHLD
#endif

#if AXMS
#include <winmain.h>    /* re-map "main()"  */
#endif


#if !LINUX
extern char *sys_errlist[];	/* Unix errro strings */
#endif

extern void AxErrnoError();

void SocketMain();
void BadExit();

static char **AxArgv0;
static char **AxArgv1;
char AxMyProgramName[STRPATH+1];
void (*AxDisconnectHandler)();
extern void AxIOElfSvrAccept();

int AxMultiClient; /* If TRUE, support multiple simultanious clients */
int AxThreadPerClient; /* If TRUE, create a new thread on accept() */
int AxExitMethod; /* 0:exit(0), 1:_exit(0), 2:kill(pid,9) */

/*************************************************************************
function: main - Top of server
*/


elfmain(argc, argv)
int argc;
char *argv[];
{
        jmp_buf dieGraceFully;
	int forkIt = 0;
	int keepOpen = 1;
	int ix;
	int quitNow = FALSE;
	int isUnix;
	int sockfd, altsockfd;
	char sockname[300],altnamebuf[300],*altname = NULL;
	elfData argArray;
	elfData axnetP, initP;
	int isPrivate;
	char *axnetServiceName = NULL;
	int passedServiceName = FALSE;
	int pipeMode = FALSE;
	int pipeFd;
	int code;

#if AXMS
    int ret;
    int trueVal = TRUE;
    WSADATA wsadata;
    struct sockaddr_in in_serv_addr;

    ret = WSAStartup(0x0101, &wsadata);
    if ( ret != 0 )
       {
       printf( "WSAStartup error = %d\n", ret );
       exit(1);
       }
#endif

#if AX_X11
    if ( !access( "/tmp/debug_elrpc3", 0 ) )
       {
       /* pause till we get a hold with dbx */
       code = 99;
       while ( code >= 99 )
         {
         code++;
         sleep(2);
         }
       }
#endif

    sockfd = altsockfd = 0;

	strcpy(AxMyProgramName,argv[0]);
	AxArgv0 = &argv[0];
	AxArgv1 = &argv[1];

	sockname[0] = 0;
    if (argc >= 2)
        /* old code assumes that the first argument is the socket name */
        strcpy(sockname, argv[1]);

	for ( ix = 1; ix < argc; ix++ ) {
        if ( 0 == strcmp(argv[ix], "-sockname") )
            {
            /* socket name is the PRECEEDING argument (see "rpc_connector$") */
            strcpy(sockname,argv[ix-1]);
            }
#if AXMS           /*  {  */
        else if (0 == strcmp(argv[ix], "-sockhandle") )
            {
            /* we were passed a port number from our parent */
            /* so we use it to create the socket here       */
            /* The real work to bind to this socket happens below the call */
            /* to AxOpenDualServer() below                                */
            strcpy( sockname, argv[ix+1] );
            }
#endif           /*  }  AXMS  */
		else if ( 0 == strcmp(argv[ix], "-slave") )
			keepOpen = 0;
		else if ( 0 == strcmp(argv[ix], "-fork") )
			forkIt = 1;
		else if ( 0 == strcmp(argv[ix], "-service") )
			passedServiceName = TRUE;
		else if ( 0 == strcmp(argv[ix],"-channel") ) {
			pipeMode = TRUE;
			pipeFd = atoi(argv[++ix]);
		}
	}

/* set and catch out of memory condition */

	if ( setjmp(dieGraceFully) ) {
		printf("Out of memory\n");
		BadExit(1);
	}
	AxMallocAborter(dieGraceFully);

/* Note command line into arg array for subsequent AxDispatchCalls */

	argArray = AxMakeArray(argc);
	for ( ix = 0; ix < argc; ix++ )
		AxAddStrToArray(argArray,ix,argv[ix]);

/* See if consumer wants to get socket from /etc/services */

	if ( !pipeMode && !passedServiceName && !sockname[0] ) {
		initP = AxDispatch(-3,argArray,&quitNow);
		if ( initP ) {
			if ( AxIsArray(initP) ) {	/* axnet only */
				altname = altnamebuf;
				strcpy(sockname,AxStrFromArray(initP,0));
 				strcpy(altname,AxStrFromArray(initP,1));
			}
			else
				strcpy(sockname,AxStrFromDataPtr(initP));
		}
	}

/* See if consumer wants to bind into axnet */

	if ( !pipeMode && !passedServiceName ) {
		axnetP = AxDispatch(-2,argArray,&quitNow);
		axnetServiceName = AxStrFromArray(axnetP,0);
		isPrivate = AxIntFromArray(axnetP,1);
	}

/* If yes, register with axnet and get channel */

	if ( !pipeMode && axnetServiceName && axnetServiceName[0] ) {
		int channel;
		char *str, *obj;
		elfData response;

		channel = AxNetConnect(NULL,&code,&str,&obj);
		if ( code != 0 ) {
			char buf[300];
err:		printf("%s: %s\n",obj,str);
			BadExit(1);
		}

		AxSendMessageToServer(channel,AXNET_GET_SOCKET,NULL,&code,&str,&obj);
		if ( code != 0 )
			goto err;

		response = AxGetResponseFromServer(channel,&code,&str,&obj);
		strcpy(sockname,AxStrPtrFromDataPtr(response));
		AxClientDisConnect(channel);
		AxNetRegisterService(sockname,axnetServiceName,TRUE,"",&code, &str, &obj);
		if ( code != 0 )
			goto err;
	}
	else if ( !pipeMode && !sockname[0] && (sockfd == 0) ) {
		printf("Usage:\n");
		printf("\t %s <socket> [-slave]\n",argv[0]);
		BadExit(1);
	}

	if ( !pipeMode ) {
		AxOpenDualServer(sockname,altname,forkIt,&sockfd,&altsockfd,FALSE);
	}

/* Let customer code know server has started */

	AxDispatch(0,argArray,&quitNow);

	AxFreeData(argArray);

/* recieve messages, call customer dispatcher */

	if ( pipeMode ) {
		elfData rP;
		int size, cnt;
		char *outBuf;

		rP = AxMakeArray(2);
		AxAddIntToArray(rP,0,0);
		AxAddArrayToArray(rP,1,NULL);
		size = AxMemWrite(rP,0,&outBuf);
#if AX_X11
		cnt = write(pipeFd,outBuf,size);	/* Unblock client */
#else
		cnt = send(pipeFd,outBuf,size,0);
#endif
		if ( cnt != size )
			BadExit(1);

		do { code = AxSockTop(pipeFd); }
		while ( code == 0 );
		if ( code == -1 )
			AxErrnoError("io");
	}
	else {
		isUnix = (sockname[0] == '/');
		if ( AxMultiClient ) {
			AxIORegisterService(sockfd,AxIOElfSvrAccept);
			if ( altsockfd )
				AxIORegisterService(altsockfd,AxIOElfSvrAccept);
			AxIOSelect(NULL);
		}
		else
			AxServerDualMainLoop(sockfd,altsockfd,isUnix,keepOpen);
	}

/* if got here, channel closed or customer set the quitNow */

	if ( AxDisconnectHandler )
		(*AxDisconnectHandler)();
	if ( isUnix && !pipeMode )
		unlink(sockname);
	if ( altname )
		unlink(altname);

#if AXMS
    WSACleanup();
#endif
	switch ( AxExitMethod ) {
	case 2:
#if !AXMS
		kill(getpid(),9);
#endif
	case 1:
		_exit(0);
	default:
		exit(0);
	}
}

/*************************************************************************
function: AxSetDisconnectHandler - Defines a shutdown callback
Called by customer code.
*/
void AxSetDisconnectHandler (handler)
void (*handler)();
{
	AxDisconnectHandler = handler;
}

/*************************************************************************
function: AxSocketMain - Alternate acess
*/
void AxSocketMain (sockname,forkIt,keepOpen)
char *sockname;
int forkIt, keepOpen;
{
	int sockfd;
	int isUnix;

	isUnix = (sockname[0] == '/');

/* Allow child programs to start & finish */

/*	signal (SIGCLD, SIG_IGN); */

/* bind to socket */

	sockfd = AxOpenServer(sockname,forkIt);

/* recieve messages, call customer dispatcher */

	isUnix = (sockname[0] == '/');
	AxServerMainLoop(sockfd,isUnix,keepOpen);

/* if got here, channel closed or customer set the quitNow */

	if ( isUnix )
		unlink(sockname);
	exit(0);
}

/*************************************************************************
function: AxSetProgramName - Just for fun with ps command
Probably not portable!
*/
void AxSetProgramName (str)
char *str;
{
#if 0
	static char buf[51];

	buf[0] = '(';
	strncpy(buf+1,str,48);
	buf[49] = 0;
	strcat(buf,")");
	*AxArgv0 = buf;
	*AxArgv1 = NULL;
#endif
}
