/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: elsql.c
   Applix ASC version: 20
   Date: Tue Feb  2 15:20:51 1999
*/
#define COMPILER_OPCODES 1	/* to get opcodes in compiler form */

#include <elcomp.h>
#include <eltokens.h>
#include <elnodes.h>
#include <elcodes.h>
#include <sqlops.h>
#include <getstr4.h>

extern node_ptr *NodeArray;	/* malloced array to put nodes */
extern data_id *GetSymP();
extern data_id *NeedVariable();
extern ndata GetListMemberAndAttr();

/*************************************************************************
function: ElfcSQLOp - Add a SQL opcode & subcode
*/
void
ElfcSQLOp(subcode)
	int             subcode;
{
	AddGenericInst(OP_SQL);
	AddWord(subcode);
}

/*************************************************************************
function: ElfcSQLPushToken - Push a string literal as a SQL token
*/
void
ElfcSQLPushToken(str)
	char           *str;
{
	AddEaInstruction(OP_PUSH_LIT, GetSymP(AddStrConst(str)));
	ElfcSQLOp(SQLOP_TOKEN);
}

/*************************************************************************
function: EmitSQLPrepare - Send a PREPARE stmt to gateway
*/
EmitSQLPrepare (stmtId,sqlStmt)
register ndata stmtId,sqlStmt;
{
    register node_ptr np;
    register data_id *symP;
	register int type;
	
    symP = NeedVariable(stmtId);
	ElfcSQLOp(SQLOP_START);
	np = NODEPTR(sqlStmt);
	type = np.any->type;
	switch ( type ) {
	case ND_SELECT:
		emit_select_expression(sqlStmt);
		break;
	case ND_UPDATE:
		emit_update_statement(np.any->data[0],np.any->data[1],np.any->data[2]);
		break;
	default:
		assert(FALSE);
	}
	AddEaInstruction(OP_PUSH_LIT, GetSymP(AddStrConst(NXLT("sql_prepare_esql$"))));
	ElfcSQLOp(SQLOP_PUSH);
	AddWord(GetRegister());
	AddEaInstruction(OP_POPTO,symP);
}

/*************************************************************************
function: EmitSQLUnPrepare - Send a UNPREPARE stmt to gateway
*/
EmitSQLUnPrepare (stmtId)
register ndata stmtId;
{
	NeedVariable(stmtId);
	EmitExpression(stmtId);
	ElfcSQLOp(SQLOP_UNPREPARE);
}

/*************************************************************************
function: EmitSQLExecute - Execute prepared statement
*/
EmitSQLExecute (stmtId,exprs)
register ndata stmtId;
register ndata exprs;
{
	int bind_count = 0;

	if ( exprs ) {
        bind_count = SizeList(exprs);
		EmitExpression(exprs);
	}
	EmitExpression(stmtId);
	ElfcSQLOp(SQLOP_EXECUTE);
	AddWord(bind_count);
}

/*************************************************************************
function: EmitSQLDeclare - Send a DECLARE stmt to gateway
Creates a cursor mapped to passed stmt id
*/
EmitSQLDeclare (cursorId, stmtId)
register ndata cursorId, stmtId;
{
    register node_ptr np;
    register data_id *symP;
	
    symP = NeedVariable(cursorId);
	EmitExpression(stmtId);
	ElfcSQLOp(SQLOP_DCLCURSOR);
	AddWord(GetRegister());
	AddEaInstruction(OP_POPTO,symP);
}

/*************************************************************************
function: EmitSQLOpen - Send a cursor OPEN stmt to gateway
*/
EmitSQLOpen (cursorId,exprs)
register ndata cursorId;
register ndata exprs;
{
	int bind_count = 0;

	NeedVariable(cursorId);
	if ( exprs ) {
        bind_count = SizeList(exprs);
		EmitExpression(exprs);
	}
	EmitExpression(cursorId);
	ElfcSQLOp(SQLOP_OPENCURSOR);
	AddWord(bind_count);
}

/*************************************************************************
function: EmitSQLClose - Send a cursor Close stmt to gateway
*/
EmitSQLClose (cursorId)
register ndata cursorId;
{
	NeedVariable(cursorId);
	EmitExpression(cursorId);
	ElfcSQLOp(SQLOP_CLOSECURSOR);
}

/*************************************************************************
function: EmitSQLUpdate - Send an UPDATE statement
*/
EmitSQLUpdate (table,colsets,search)
register ndata table, colsets, search;
{
	ElfcSQLOp(SQLOP_START);
	emit_update_statement(table,colsets,search);
	AddEaInstruction(OP_PUSH_LIT, GetSymP(AddStrConst(NXLT("sql_esql$"))));
	ElfcSQLOp(SQLOP_PUSH);
	AddWord(GetRegister());
}

/*************************************************************************
function: EmitSQLFetch - Send a FETCH directive to gateway
*/
EmitSQLFetch (data)
register ndata data;
{
    register node_ptr np;
	register ndata cursorId, count;

	np = NODEPTR(data);
	cursorId = np.any->data[0];
	count = np.any->data[1];

	EmitExpression(cursorId);
	if ( !count )
		ElfcSQLOp(SQLOP_FETCHCURSOR);
	else {
		EmitExpression(count);
		ElfcSQLOp(SQLOP_ARRAYFETCHCURSOR);
	}
	AddWord(GetRegister());
}

/*************************************************************************
function: EmitSelect - Start a SQL select expression
*/
EmitSelect(data)
register ndata data;
{
	ElfcSQLOp(SQLOP_START);
	emit_select_expression(data);
	AddEaInstruction(OP_PUSH_LIT, GetSymP(AddStrConst(NXLT("sql_select_esql$"))));
	ElfcSQLOp(SQLOP_PUSH);
	AddWord(GetRegister());
}

emit_update_statement(table,colsets,search)
register ndata table, colsets, search;
{
	int             ix, cnt;
	register node_ptr np;

	ElfcSQLPushToken(NXLT("update"));
	emit_itemname(table);
	ElfcSQLPushToken(NXLT("set"));

	np = NODEPTR(colsets);
	cnt = SizeList(colsets);
	for (ix = 0; ix < cnt; ix++) {
		emit_colsetter(GetListMember(colsets, ix));
		if (ix != cnt - 1)
			ElfcSQLPushToken(",");
	}
	ElfcSQLPushToken(NXLT("where"));
	emit_search_condition(search);
}

emit_colsetter(data)
register ndata data;
{
	register node_ptr np;

	np = NODEPTR(data);
	emit_itemname(np.any->data[0]);
	ElfcSQLPushToken("=");
	emit_sql_expr(np.any->data[1]);
}

emit_select_expression(data)
register ndata data;
{
	register ndata  set_qualifier, select_list, table_expressions;

	register node_ptr np;

	ElfcSQLPushToken(NXLT("select"));

	np = NODEPTR(data);
	set_qualifier = np.any->data[0];
	select_list = np.any->data[1];
	table_expressions = np.any->data[2];

	switch (set_qualifier) {
	case TK_ALL:
		ElfcSQLPushToken(NXLT("all"));
		break;
	case TK_DISTINCT:
		ElfcSQLPushToken(NXLT("distinct"));
		break;
	}

	emit_select_list(select_list);
	emit_table_expressions(table_expressions);
}

emit_select_list(data)
	register ndata  data;
{
	if (data == TK_MULT)
		ElfcSQLPushToken("*");
	else
		emit_column_list(data);
}

emit_column_list(data)
	register ndata  data;
{
	int             ix, cnt;

	cnt = SizeList(data);
	for (ix = 0; ix < cnt; ix++) {
		emit_sql_expr(GetListMember(data, ix));
		if (ix != cnt - 1)
			ElfcSQLPushToken(",");
	}
}

emit_bindable(data)
	register ndata  data;
{
	register node_ptr np;
	register data_id *symP;

	if (IS_NODE(data) && (NodeType(data) == ND_BINDER)) {
		np = NODEPTR(data);
		if (np.any->data[0] == TK_SQL) {
			if ( np.any->data[1] == TK_HOOK ) {
				ElfcSQLOp(SQLOP_BIND_TOKEN);
			}
			else {
				symP = NeedVariable(np.any->data[1]);
				ElfcSQLOp(SQLOP_EA_REPL_TOKEN);
				AddEaCanMerge(symP);
			}
		}
		else {
			if ( np.any->data[1] == TK_HOOK ) {
				ElfcSQLOp(SQLOP_BIND_DATUM);
			}
			else {
				symP = NeedVariable(np.any->data[1]);
				ElfcSQLOp(SQLOP_EA_REPL_DATUM);
				AddEaCanMerge(symP);
			}
		}
	} else
		ElfcSQLPushToken(GetSymP(data)->name);
}

emit_table_expressions(data)
	register ndata  data;
{
	int             ix, cnt;
	register node_ptr np;

	np = NODEPTR(data);
	cnt = SizeList(data);
	for (ix = 0; ix < cnt; ix++) {
		emit_table_expression(GetListMember(data, ix));
	}
}

emit_table_expression(data)
	register ndata  data;
{
	register node_ptr np;

	np = NODEPTR(data);
	switch ( np.any->type ) {
	case ND_FROMTAB:
		ElfcSQLPushToken(NXLT("from"));
		emit_table_names(np.any->data[0]);
		break;
	case ND_WHERE:
		ElfcSQLPushToken(NXLT("where"));
		emit_search_condition(np.any->data[0]);
		break;
	case ND_CONNECT:
		ElfcSQLPushToken(NXLT("connect by"));
		emit_search_condition(np.any->data[0]);
		if ( np.any->data[1] ) {
			ElfcSQLPushToken(NXLT("start with"));
			emit_search_condition(np.any->data[1]);
		}
		break;
	case ND_ORDERBY:
		ElfcSQLPushToken(NXLT("order by"));
		emit_order_names(np.any->data[0]);
		break;
	case ND_GROUPBY:
		ElfcSQLPushToken(NXLT("group by"));
		emit_table_names(np.any->data[0]);
		break;
	case ND_UNION:
		ElfcSQLPushToken(NXLT("union"));
		emit_select_expression(np.any->data[0]);
		break;
	case ND_INTERSECT:
		ElfcSQLPushToken(NXLT("intersect"));
		emit_select_expression(np.any->data[0]);
		break;
	case ND_MINUS:
		ElfcSQLPushToken(NXLT("minus"));
		emit_select_expression(np.any->data[0]);
		break;
	default:
		assert(FALSE);
	}
}

emit_order_names(data)
	register ndata  data;
{
	int             ix, cnt;
	register node_ptr np;
	ndata desc, datum;

	np = NODEPTR(data);
	cnt = SizeList(data);
	for (ix = 0; ix < cnt; ix++) {
		datum = GetListMemberAndAttr(data, ix, &desc);
		emit_order_name(datum);
		if ( desc == TK_DESC )
			ElfcSQLPushToken(NXLT("desc"));
		if (ix != cnt - 1)
			ElfcSQLPushToken(",");
	}
}

emit_order_name(data)
	register ndata  data;
{
	register node_ptr np;

	emit_itemname(data);
}

emit_table_names(data)
	register ndata  data;
{
	int             ix, cnt;
	register node_ptr np;

	np = NODEPTR(data);
	cnt = SizeList(data);
	for (ix = 0; ix < cnt; ix++) {
		emit_itemname(GetListMember(data, ix));
		if (ix != cnt - 1)
			ElfcSQLPushToken(",");
	}
}

emit_itemname(data)
	register ndata  data;
{
	int             ix, cnt;
	register node_ptr np;

	np = NODEPTR(data);
	cnt = SizeList(data);
	for (ix = 0; ix < cnt; ix++) {
		emit_bindable(GetListMember(data, ix));
		if (ix != cnt - 1)
			ElfcSQLPushToken(".");
	}
}

emit_search_condition(data)
register ndata  data;
{
	register node_ptr np;

	if ( IS_NODE(data) && NodeType(data) == ND_SQLOR ) {
		np = NODEPTR(data);
		emit_search_condition(np.any->data[0]);
		ElfcSQLPushToken(NXLT("or"));
		emit_boolean_term(np.any->data[1]);
	}
	else {
		emit_boolean_term(data);
	}
}

emit_boolean_term(data)
register ndata  data;
{
	register node_ptr np;

	if ( IS_NODE(data) && NodeType(data) == ND_SQLAND ) {
		np = NODEPTR(data);
		emit_boolean_term(np.any->data[0]);
		ElfcSQLPushToken(NXLT("and"));
		emit_boolean_factor(np.any->data[1]);
	}
	else {
		emit_boolean_factor(data);
	}
}

emit_boolean_factor(data)
register ndata  data;
{
	register node_ptr np;

	if ( IS_NODE(data) && NodeType(data) == ND_SQLNOT ) {
		np = NODEPTR(data);
		ElfcSQLPushToken(NXLT("not"));
		emit_boolean_test(np.any->data[0]);
	}
	else {
		emit_boolean_test(data);
	}
}

emit_boolean_test(data)
register ndata  data;
{
	emit_boolean_primary(data);
}

emit_boolean_primary(data)
register ndata  data;
{
	register node_ptr np;

	if ( IS_NODE(data) && NodeType(data) == ND_SQLPAREN ) {
		np = NODEPTR(data);
		ElfcSQLPushToken("(");
		emit_search_condition(np.any->data[0]);
		ElfcSQLPushToken(")");
	}
	else {
		emit_predicate(data);
	}
}

emit_sql_exprs(data)
register ndata  data;
{
	int             ix, cnt;

	cnt = SizeList(data);
	for (ix = 0; ix < cnt; ix++) {
		emit_sql_expr(GetListMember(data, ix));
		if (ix != cnt - 1)
			ElfcSQLPushToken(",");
	}
}

emit_predicate(data)
register ndata  data;
{
	register node_ptr np;

	np = NODEPTR(data);
	switch ( np.any->data[0] ) {
	case TK_IN:
		emit_sql_expr(np.any->data[2]);
		if ( np.any->data[1] )
			ElfcSQLPushToken(NXLT("not in ("));
		else
			ElfcSQLPushToken(NXLT("in ("));
		emit_sql_exprs(np.any->data[3]);
		ElfcSQLPushToken(")");
		break;
	case TK_BETWEEN:
		emit_sql_expr(np.any->data[2]);
		if ( np.any->data[1] )
			ElfcSQLPushToken(NXLT("not between"));
		else
			ElfcSQLPushToken(NXLT("between"));
		emit_sql_expr(np.any->data[3]);
		ElfcSQLPushToken(NXLT("and"));
		emit_sql_expr(np.any->data[4]);
		break;
	case TK_LESS:
		emit_sql_expr(np.any->data[1]);
		ElfcSQLPushToken("<");
		emit_sql_expr(np.any->data[2]);
		break;
	case TK_GRTR:
		emit_sql_expr(np.any->data[1]);
		ElfcSQLPushToken(">");
		emit_sql_expr(np.any->data[2]);
		break;
	case TK_LEQ:
		emit_sql_expr(np.any->data[1]);
		ElfcSQLPushToken("<=");
		emit_sql_expr(np.any->data[2]);
		break;
	case TK_GEQ:
		emit_sql_expr(np.any->data[1]);
		ElfcSQLPushToken(">=");
		emit_sql_expr(np.any->data[2]);
		break;
	case TK_EQUALS:
		emit_sql_expr(np.any->data[1]);
		ElfcSQLPushToken("=");
		emit_sql_expr(np.any->data[2]);
		break;
	case TK_NEQ:
		emit_sql_expr(np.any->data[1]);
		ElfcSQLPushToken("<>");
		emit_sql_expr(np.any->data[2]);
		break;
	default:
		assert(FALSE);
	}
}

emit_sql_expr(data)
register ndata  data;
{
	register node_ptr np;

	if ( !IS_NODE(data) ) {
		EmitExpression(data);
		ElfcSQLOp(SQLOP_DATUM);
		return;
	}

	np = NODEPTR(data);

	switch ( np.any->type ) {

	case ND_USERTAB:
		emit_itemname(data);
		return;

	case ND_SELECT:
		emit_select_expression(data);
		return;

	case ND_COLSTAR:
		emit_itemname(np.any->data[0]);
		ElfcSQLPushToken(".*");
		return;

	case ND_SQLFUNC:
		ElfcSQLPushToken(GetSymP(np.any->data[0])->name);
		ElfcSQLPushToken("(");
		if ( np.any->data[1] == TK_ALL )
			ElfcSQLPushToken(NXLT("all"));
		else if ( np.any->data[1] == TK_DISTINCT )
			ElfcSQLPushToken(NXLT("distinct"));
		if ( np.any->data[2] == TK_MULT )
			ElfcSQLPushToken("*");
		else
			emit_sql_exprs(np.any->data[2]);
		ElfcSQLPushToken(")");
		return;
	}

	switch ( np.any->data[0] ) {
	case TK_SUB:
		if ( np.any->data[1] ) {
			emit_sql_expr(np.any->data[1]);
			ElfcSQLPushToken("-");
			emit_sql_expr(np.any->data[2]);
		}
		else {
			ElfcSQLPushToken("-");
			emit_sql_expr(np.any->data[2]);
		}
		break;
	case TK_ADD:
		emit_sql_expr(np.any->data[1]);
		ElfcSQLPushToken("+");
		emit_sql_expr(np.any->data[2]);
		break;
	case TK_EXP:
		emit_sql_expr(np.any->data[1]);
		ElfcSQLPushToken("^");
		emit_sql_expr(np.any->data[2]);
		break;
	case TK_MULT:
		emit_sql_expr(np.any->data[1]);
		ElfcSQLPushToken("*");
		emit_sql_expr(np.any->data[2]);
		break;
	case TK_DIV:
		emit_sql_expr(np.any->data[1]);
		ElfcSQLPushToken("/");
		emit_sql_expr(np.any->data[2]);
		break;
	case TK_REM:
		emit_sql_expr(np.any->data[1]);
		ElfcSQLPushToken(NXLT("rem"));		/* !!1 remainders? */
		emit_sql_expr(np.any->data[2]);
		break;
	default:
		assert(FALSE);
	}
}

