/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: elssfunc.c
   Applix ASC version: 9
   Date: Tue Feb  2 15:22:13 1999
*/
#include <errno.h>
#include <aos.h>
#include <eldata.h>
#include <arghdr.h>
#include <ssfunc.h>
#include <getstr4.h>
#include <elssfunc.p>

data_ptr AxfSSFuncs(argDatum)
	data_ptr argDatum;
{ 
    data_ptr cmdCodeP = AnyFromArgArray(argDatum,0),
            arg1 = AnyFromArgArray(argDatum,1),
            arg2 = AnyFromArgArray(argDatum,2),
            arg3 = AnyFromArgArray(argDatum,3),
            arg4 = AnyFromArgArray(argDatum,4),
            arg5 = AnyFromArgArray(argDatum,5),
            arg6 = AnyFromArgArray(argDatum,6);
    int cmdCode = IntFromDataPtr(cmdCodeP);
	int isError;
	double 	value, hour, sec, min, day, month, year, ret_val;
	double 	num_hours, num_seconds, num_minutes, num_days, num_months,
			num_years, dateTime, x, y;
	int fmt, fullYear;
	data_ptr retP = NULL;
	char *str, *obj;
	int  code = ERR_ILLVAL;
	char dateTimeStr[STRPATH+1];

	isError = FALSE;
    errno = 0;
	switch(cmdCode)	{
		case 1:		/* Today */
			value = ssToday(&isError);	
			retP = AxMakeFloatData(value);
			break;
		case 2:		/* Now */
			value = ssNow(&isError);	
			retP = AxMakeFloatData(value);
			break;
		case 3:		/* Year */
			dateTime = AxFloatFromDataPtr(arg1);
			value = ssYear(dateTime, &isError);
			if( isError)	{
				obj = NXLT("Year@");
				break;
			}
			retP = AxMakeFloatData(value);
			break;
		case 4:		/* Month */
			dateTime = AxFloatFromDataPtr(arg1);
			value = ssMonth(dateTime, &isError);
			if( isError)	{
				obj = NXLT("Month@");
				break;
			}
			retP = AxMakeFloatData(value);
			break;
		case 5:		/* Day */
			dateTime = AxFloatFromDataPtr(arg1);
			value = ssDay(dateTime, &isError);
			if( isError)	{
				obj = NXLT("Day@");
				break;
			}
			retP = AxMakeFloatData(value);
			break;
		case 6:		/*  Hour */
			dateTime = AxFloatFromDataPtr(arg1);
			value = ssHour(dateTime, &isError);
			if( isError)	{
				obj = NXLT("Hour@");
				break;
			}
			retP = AxMakeFloatData(value);
			break;
		case 7:		/* Minute */
			dateTime = AxFloatFromDataPtr(arg1);
			value = ssMinute(dateTime, &isError);
			if( isError)	{
				obj = NXLT("Minute@");
				break;
			}
			retP = AxMakeFloatData(value);
			break;
		case 8:		/* Second */	
			dateTime = AxFloatFromDataPtr(arg1);
			value = ssSecond(dateTime, &isError);
			if( isError)	{
				obj = NXLT("Second@");
				break;
			}
			retP = AxMakeFloatData(value);
			break;
		case 9:		/* Date */
			year = AxFloatFromDataPtr(arg1);
			month = AxFloatFromDataPtr(arg2);
			day = AxFloatFromDataPtr(arg3);
			value = ssDate(year, month, day, &isError);
			if( isError)	{
				obj = NXLT("Date@");
				break;
			}
			retP = AxMakeFloatData(value);
			break;
		case 10:	/* Time */	
			hour = AxFloatFromDataPtr(arg1);
			min = AxFloatFromDataPtr(arg2);
			sec = AxFloatFromDataPtr(arg3);
			value = ssTime(hour, min, sec, &isError);
			if( isError)	{
				obj = NXLT("Time@");
				break;
			}
			retP = AxMakeFloatData(value);
			break;
		case 11:	/* AddDays */
			dateTime = AxFloatFromDataPtr(arg1);
			num_days = AxFloatFromDataPtr(arg2);
			value = ssAddDays(dateTime, num_days, &isError);
			if( isError)	{
				obj = NXLT("AddDays@");
				break;
			}
			retP = AxMakeFloatData(value);
			break;
		case 12:	/* AddMonths */
			dateTime = AxFloatFromDataPtr(arg1);
			num_months = AxFloatFromDataPtr(arg2);
			value = ssAddMonths(dateTime, num_months, &isError);
			if( isError)	{
				obj = NXLT("AddMonths@");
				break;
			}
			retP = AxMakeFloatData(value);
			break;	
		case 13:	/* AddYears */
			dateTime = AxFloatFromDataPtr(arg1);
			num_years = AxFloatFromDataPtr(arg2);
			value = ssAddYears(dateTime, num_years, &isError);
			if( isError)	{
				obj = NXLT("AddYears@");
				break;
			}
			retP = AxMakeFloatData(value);
			break;
		case 14:	/* AddHours */
			dateTime = AxFloatFromDataPtr(arg1);
			num_hours = AxFloatFromDataPtr(arg2);
			value = ssAddHours(dateTime, num_hours, &isError);
			if( isError)	{
				obj = NXLT("AddHours@");
				break;
			}
			retP = AxMakeFloatData(value);
			break;
		case 15:	/* AddMinutes */
			dateTime = AxFloatFromDataPtr(arg1);
			num_minutes = AxFloatFromDataPtr(arg2);
			value = ssAddMinutes(dateTime, num_minutes, &isError);
			if( isError)	{
				obj = NXLT("AddMinutes@@");
				break;
			}
			retP = AxMakeFloatData(value);
			break;
		case 16:	/* AddSeconds */
			dateTime = AxFloatFromDataPtr(arg1);
			num_seconds = AxFloatFromDataPtr(arg2);
			value = ssAddSeconds(dateTime, num_seconds, &isError);
			if( isError)	{
				obj = NXLT("AddSeconds@");
				break;
			}
			retP = AxMakeFloatData(value);
			break;
		case 17:
			dateTime = AxFloatFromDataPtr(arg1);
			fmt = AxIntFromDataPtr(arg2);
			fullYear = AxIntFromDataPtr(arg3);
			ssDateStr(dateTime, fmt, fullYear, dateTimeStr, &isError);	
			if( isError)	{
				obj = NXLT("datestr@");
				break;
			}
			retP = AxMakeStrData(-1, dateTimeStr);
			break;
		case 18:
			dateTime = AxFloatFromDataPtr(arg1);
			fmt = AxIntFromDataPtr(arg2);
			ssTimeStr(dateTime, fmt, dateTimeStr, &isError);	
			if( isError)	{
				obj = NXLT("timestr@");
				break;
			}
			retP = AxMakeStrData(-1, dateTimeStr);
			break;
		case 19:
			value = AxFloatFromDataPtr(arg1);
			ret_val = ssExp(value, &isError);
			if(isError)	{
				obj = NXLT("exp@");
				break;
			}
			retP = AxMakeFloatData(ret_val);
			break;
		case 20:
			x = AxFloatFromDataPtr(arg1);
			y = AxFloatFromDataPtr(arg2);
			ret_val = ssPower(x, y, &isError);
			if(isError)	{
				obj = NXLT("power@");
				break;
			}
			retP = AxMakeFloatData(ret_val);
			break;
		case 21:
			x = AxFloatFromDataPtr(arg1);
			y = AxFloatFromDataPtr(arg2);
			ret_val = ssRound(x, y, &isError);
			if(isError)	{
				obj = NXLT("round@");
				break;
			}
			retP = AxMakeFloatData(ret_val);
			break;
		case 22:
			ret_val = ssPi();
			retP = AxMakeFloatData(ret_val);
			break;
		case 23:		/* YearAbs */
			dateTime = AxFloatFromDataPtr(arg1);
			value = ssYearAbs(dateTime, &isError);
			if( isError)	{
				obj = NXLT("YearAbs@");
				break;
			}
			retP = AxMakeFloatData(value);
			break;
	}
	if(isError)	{
    	ElfStrAbort(code, XLT("Error in function"), obj);
		return(NULL);
	}
	return(retP);
}

AxSSCalcGetCellRange(tssP, col1, row1, sheet1, col2, row2, sheet2)
	void *tssP;
	int *col1, *row1, *sheet1, *col2, *row2, *sheet2;
{
	ElfSSCalcGetCellRange(tssP, col1, row1, sheet1, col2, row2, sheet2) ;
}
