/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: attr.h
   Applix ASC version: 2
   Date: Mon Nov 16 13:56:04 1998
*/
#ifdef COMPILE_COMMENTS

How the bits are crammed into their words:

====================================
CHARACTER ATTRIBUTES:	t->patr - 2bits avail
  +---+---+---+---+---+---+
  | 4 | 4 | 2 | 2 | 2 | 2 |
  +-|-+-|-+-|-+-|-+-|-+-|-+
    |   |   |   |   |   V
    |   |   |   |   V   reserved (0)
    |   |   |   V   italics
    |   |   V   bold
    |   V   underline
    V   size
    font
====================================
TEXT ATTRIBUTES:	t->attr	- 9bits avail
  +---+---+---+---+---+---+---+
  |(2)| 1 | 2 | 2 |(1)| 2 |(6)|
  +-|-+-|-+-|-+-|-+-|-+-|-+-|-+
    |   |   |   |   |   |   V
    |   |   |   |   |   V   reserved (0)
    |   |   |   |   V   writing-mode	(new in V2)
    |   |   |   V   reserved (0)
    |   |   V   vertical-jutification	(not used)
    |   V   horizontal-justification
    V   reverse
    reserved (0)
====================================
LINE ATTRIBUTES:	t->patr - 3bits avail
  +---+---+---+---+---+---+---+
  |(2)| 1 | 5 | 1 | 1 | 3 | 3 |
  +-|-+-|-+-|-+-|-+-|-+-|-+-|-+
    |   |   |   |   |   |   V
    |   |   |   |   |   V   thickness
    |   |   |   |   V   symbol		(marker symbol)
    |   |   |   V   arrowheaded		(line vis if 0)
    |   |   V   curved
    |   V   pattern
    V   reverse				(reverse in V1)
    reserved (0)			(was old V1 color)

arrowhead is like the sign bit of the symbol field:
arrowhead = 0:
	symbol = 0:	normal line
	symbol = 1-5:	markers & line
	symbol = 6:	reserved (becomes normal line)
	symbol = 7:	arrows on both ends
arrowhead = 1:
	symbol = 0:	normal line
	symbol = 1-5:	markers only
	symbol = 6:	left arrowhead
	symbol = 7:	right arrowhead

reverse here is V1 position; we move this bit, if set,
to the V2 position & clear it on startup, for compatibility.
====================================
FILL ATTRIBUTES:	t->attr	- 5bits avail
  +---+---+---+---+---+---+---+
  |(2)| 1 | 5 | 2 | 1 | 2 |(3)|
  +-|-+-|-+-|-+-|-+-|-+-|-+-|-+
    |   |   |   |   |   |   V
    |   |   |   |   |   V   reserved (0)
    |   |   |   |   V   arrow-type
    |   |   |   V   closed
    |   |   V   writing-mode
    |   V   pattern
    V   reverse				(reverse in V2)
    reserved (0)
====================================
COLOR ATTRIBUTES:	t->color - 8bits avail
  +---+---+---+
  |(8)| 4 | 4 |
  +-|-+-|-+-|-+
    |   |   V
    |   V   foreground color (xor-ed w/ 1 for compatibility)
    V   background color
    reserved (0)
====================================

#endif	/* COMMENTS */

/****************************************************************************\
\****************************************************************************/

/* char attrs */
#define FONT_S	12
#define SIZE_S	8
#define ULIN_S	6
#define BOLD_S	4
#define ITAL_S	2

#define FONT_M	0xF000
#define SIZE_M	0x0F00
#define ULIN_M	0x00C0
#define BOLD_M	0x0030
#define ITAL_M	0x000C

/* text attrs */
#define REVRS_S	13
#define HJUST_S	11
#define VJUST_S	9

#define REVRS_M	0x2000
#define HJUST_M	0x1800
#define VJUST_M	0x0600

/* line attrs, following REVRS... */
#define PATN_S	8
#define CURV_S	7
#define AROW_S	6
#define SYMB_S	3
#define LSIZ_S	0

#define PATN_M	0x1F00
#define CURV_M	0x0080
#define AROW_M	0x0040
#define SYMB_M	0x0038
#define LSIZ_M	0x0007

/* poly attrs following REVRS, PATN... */
#define WMODE_S	6
#define CLOSE_S	5
#define ARTYP_S	3

#define WMODE_M	0x00C0
#define CLOSE_M	0x0020
#define ARTYP_M	0x0018

/* color attrs */
#define BCOLR_S	4
#define FCOLR_S	0

#define	BCOLR_M	0x00F0
#define	FCOLR_M	0x000F

/* magic macros */
#define G_A(v, n, M, S)	((unsigned short) ((v) & M) >> S)
#define S_A(v, n, M, S)	(v) = (v) & ~M | ((n) << S) & M

#define FONT(op, v, n)	op(v, n, FONT_M, FONT_S)
#define SIZE(op, v, n)	op(v, n, SIZE_M, SIZE_S)
#define ULIN(op, v, n)	op(v, n, ULIN_M, ULIN_S)
#define BOLD(op, v, n)	op(v, n, BOLD_M, BOLD_S)
#define ITAL(op, v, n)	op(v, n, ITAL_M, ITAL_S)
#define REVRS(op, v, n)	op(v, n, REVRS_M, REVRS_S)
#define HJUST(op, v, n)	op(v, n, HJUST_M, HJUST_S)
#define VJUST(op, v, n)	op(v, n, VJUST_M, VJUST_S)
#define PATN(op, v, n)	op(v, n, PATN_M, PATN_S)
#define CURV(op, v, n)	op(v, n, CURV_M, CURV_S)
#define AROW(op, v, n)	op(v, n, AROW_M, AROW_S)
#define SYMB(op, v, n)	op(v, n, SYMB_M, SYMB_S)
#define LSIZ(op, v, n)	op(v, n, LSIZ_M, LSIZ_S)
#define WMODE(op, v, n)	op(v, n, WMODE_M, WMODE_S)
#define CLOSE(op, v, n)	op(v, n, CLOSE_M, CLOSE_S)
#define ARTYP(op, v, n)	op(v, n, ARTYP_M, ARTYP_S)
#define BCOLR(op, v, n) op(v, n, BCOLR_M, BCOLR_S)
#define FCOLR(op, v, n) op(v, n, FCOLR_M, FCOLR_S)

#define GET_ATTR(v, A)		A(G_A, v, 0)
#define SET_ATTR(v, n, A)	A(S_A, v, n)

/**************************************************/

	/* Marker types; if negative, then w/o lines */
#define MRKR_NONE	0
#define MRKR_DOT	1
#define MRKR_CROSS	2
#define MRKR_STAR	3
#define MRKR_CIRCLE	4
#define MRKR_EX		5

	/* Arrowheads */
#define RESERVED_ARROW	6
#define BOTH_ARROW	7
#define FIRST_ARROW	(-6)
#define LAST_ARROW	(-7)

	/* Arrow types */
#define ARROW_1   0
#define ARROW_2   1
#define ARROW_3   2
#define ARROW_4   3
#define ARROW_5   4
#define ARROW_6   5
#define ARROW_7   6
#define ARROW_8   7
#define ARROW_9   8
#define ARROW_10  9
#define ARROW_11 10
#define ARROW_12 11
#define ARROW_13 12
#define ARROW_14 13

	/* Misc. */
#define GET_MARKER(p, n) \
	{ if (((n) = GET_ATTR(p, SYMB)) && GET_ATTR(p, AROW)) (n) = -(n); }

#define SET_MARKER(p, n) \
	{ SET_ATTR(p, n < 0, AROW); SET_ATTR(p, ABS(n), SYMB); }


/* 1.20 font stuff */
#ifdef SLAN_FONTS		/* should have been set up as 1/0, not null */
#define DO_ALL_FONTS 1
#else
#if THIMBLE
#define DO_ALL_FONTS 1
#endif
#endif
#if DO_ALL_FONTS
#define MAX_FONTS	13
#else
#define MAX_FONTS	8
#endif /* SLAN_FONTS */
#define MAX_SIZES	8
#define MAX_ULINS	4

