/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: bg.h
   Applix ASC version: 104
   Date: Sat Jan 23 21:14:31 1999
*/

/* three D */
typedef struct pnt32 {
    int x;
    int y;
    int z;
    int split;
} pnt32;
typedef struct poly32 {
    int n_pnts;
    int n_ctls;
    pnt32 *pnt;
    pnt32 *ctl;
} poly32;
typedef struct bg_edge {
    int ix;
    int fg;
    int split;
    double min_v;
    double max_v;
    int n_pnts;
    pnt32 *pnt;
} bg_edge;
typedef struct bg_edge_list {
    int n_edges;
    bg_edge *edge;
} bg_edge_list;

/* chart tokens */
typedef struct bg_token_type {
	char *str;
	int id;
} bg_token_type;

/* chart type */
#define BG_PERCENT    0
#define BG_CAT_VAL    1
#define BG_VAL_VAL    2
#define BG_HISTO      3
#define BG_SPIDER     4
#define BG_STR_VAL    5
#define BG_SURFACE    6
#define BG_CENT_VAL   7
#define BG_NEG_STRATA 8
#define BG_VAL_STRATA 9
#define BG_HISTO_HYT 10

extern bg_token_type bg_chart_types[];

/* data types */
#define BG_LINE            0
#define BG_CURVE           1
#define BG_BAR             2
#define BG_LBAR            3
#define BG_CBAR            4
#define BG_BUBBLE          5
#define BG_HLCO1           6
#define BG_STRATA          7
#define BG_PIE             8
#define BG_HLCO2           9
#define BG_STACKED_BAR    10
#define BG_STACKED_CBAR   11
#define BG_STACKED_STRATA 12
#define BG_HILO1          13
#define BG_HILO2          14
#define BG_TREND          15
#define BG_HLCO3          16
#define BG_HILO3          17
#define BG_CENT_BAR       18
#define BG_CENT_STRATA    19
extern bg_token_type bg_types[];

/* axis types */
#define X_AXIS 0
#define Y_AXIS 1
#define Z_AXIS 2
#define CAT_AXIS  0
#define VAL_AXIS  1
#define LOG_AXIS  2
#define HST_AXIS  3
#define STR_AXIS  4
#define VLOG_AXIS 5
#define CENT_AXIS 6

/* data point display flags */
#define BG_PLOT 0
#define BG_SUBS 1
#define BG_SPAN 2
#define BG_GAP  3
#define BG_NULL 4

/* the parts that make up a chart */
#define FRAME_BACKGROUND  0
#define FRAME_MIN_X_GRID  1
#define FRAME_MAJ_X_GRID  2
#define FRAME_MIN_Y_GRID  3
#define FRAME_MAJ_Y_GRID  4
#define FRAME_MIN_Z_GRID  5
#define FRAME_MAJ_Z_GRID  6
#define FRAME_BG_BORDERS  7
#define FRAME_BG_X_AXIS   8
#define FRAME_BG_Y_AXIS   9
#define FRAME_BG_Z_AXIS  10
#define FRAME_DATA_3D    11
#define FRAME_DATAGROUPS 12
#define FRAME_DATALABELS 13
#define FRAME_FG_BORDERS 14
#define FRAME_FG_X_AXIS  15
#define FRAME_FG_Y_AXIS  16
#define FRAME_FG_Z_AXIS  17
#define FRAME_TITLE      18
#define FRAME_SUBTITLE   19
#define FRAME_FOOTER     20
#define FRAME_LEGEND     21
#define FRAME_ERROR      22

/* the attribute types */
#define DEFAULT_ATTR           0
#define MIN_X_GRID_ATTR        1
#define MAJ_X_GRID_ATTR        2
#define MIN_Y_GRID_ATTR        3
#define MAJ_Y_GRID_ATTR        4
#define MIN_Z_GRID_ATTR        5
#define MAJ_Z_GRID_ATTR        6
#define X_AXIS_LINE_ATTR       7
#define X_AXIS_TIK_ATTR        8
#define X_AXIS_LABEL_ATTR      9
#define X_AXIS_TIK_LABEL_ATTR 10
#define Y_AXIS_LINE_ATTR      11
#define Y_AXIS_TIK_ATTR       12
#define Y_AXIS_LABEL_ATTR     13
#define Y_AXIS_TIK_LABEL_ATTR 14
#define Z_AXIS_LINE_ATTR      15
#define Z_AXIS_TIK_ATTR       16
#define Z_AXIS_LABEL_ATTR     17
#define Z_AXIS_TIK_LABEL_ATTR 18
#define DATAGROUP_ATTR        19
#define DATALABEL_ATTR        20
#define TITLE_ATTR            21
#define SUBTITLE_ATTR         22
#define FOOTER_ATTR           23
#define LEG_BOX_ATTR          24
#define LEG_TITLE_ATTR        25
#define LEG_LABEL_ATTR        26

/* Trend line types */
#define BG_TREND_LINEAR		 0
#define BG_TREND_LOG10		 1
#define BG_TREND_POWER	 	 2
#define BG_TREND_EXPONENT	 3
#define BG_TREND_POLY	 	 4
#define BG_TREND_MOVAVG		 5

/* Data Group Label Types */
#define BG_LAB_NONE                 0
#define BG_LAB_VALUE                1
#define BG_LAB_PERCENT              2
#define BG_LAB_STRING               3
#define BG_LAB_VALUE_PERCENT        4
#define BG_LAB_VALUE_STRING         5
#define BG_LAB_PERCENT_STRING       6


#define NULL_FMT(flag, null_flag) ((flag == BG_NULL) ? null_flag : flag)

typedef struct projection_volume {
	int max_x;
	int min_x;
	int max_y;
	int min_y;
	int max_z;
	int min_z;
	double sc_x;
	double sc_y;
	int dx;
	int dy;
} projection_volume;

typedef struct chart_trend_format {
	int    type;
	int    order;
	int    n_correlations;
	double r_squared;
	double correlations[13];
} chart_trend_format;

typedef struct chart_3D_effect_type {
	int enabled;
	int yaw;
	int pitch;
	int depth;
	int projection;
	int right_angles;
} chart_3D_effect_type;

typedef struct chart_data_point_type {
	/* data point flags */
	unsigned int display_flag : 3,
	             local_x      : 1,
	             local_y      : 1;

	/* the data label */
	int label_x_off;
	int label_y_off;
	char *label_str; /* used only if label type is string   */

	/* the data (some types require more than 1 - eg. bubbles need 2) */
	int nx, ny, nz;
	double x[5];
	double y[5];
	double z[5];
	double sum;
} chart_data_point_type;

typedef struct chart_data_group_type {
	/* run-time flag - has this group been drawn yet? */
	unsigned int
	 drawn    : 1,
	 new      : 1,

	/* if this flag is set, don't draw this group */
	disabled : 1,

	/* type of data this group represents */
	type     : 5,

	/*  label info: 
	 *  none, value, percent, string, value & percent, value & string 
	 *  percent & string 
	 */
	data_label_type : 3,     
	label_align     : 2;     /* left, center, right                   */
	
	/* position of this group when the chart was drawn (or -1 if new) */
	int old_ix;

	struct thing_attributes label_attr;
	struct number_format label_frmt;
	int label_x_off;
	int label_y_off;

	/* the axes this data will be plotted against */
	int x_axis; 			/* id of axis in list 			      */
	int y_axis; 			/* id of axis in list 			      */
	int z_axis; 			/* id of axis in list 			      */

	/* data point info */
	struct thing_attributes attr;
	int stack_id;            /* stack id for stacked bars             */
	char *part;              /* part name for custom bars and markers */

	/* one for each data point */
	struct chart_data_point_type *dat;
	int n;
	
	/* data storage for RT charts */
	int n_bufs;
	int enable_bufs;
	
	/* trend data type info */
	struct chart_trend_format trend_info;

	/* legend name */
	char *legend_str;

	/* group name (ELF address) */
	char name[STRPATH];
} chart_data_group_type;

typedef struct chart_old_data_group_type {
	unsigned int drawn: 1;
	int n;
	int type;
	int stack_id;
	int stack_ax;
} chart_old_data_group_type;

/* data types for constructing chart axes */
typedef struct chart_axis_type {
	unsigned int
	 type              : 3, /* cat, val, log10val, histo, strata */
	 hidden            : 1, /* true or false */
	 inverted          : 1, /* true or false */	
	 auto_max          : 1, /* true or false */
	 auto_min          : 1, /* true or false */
	 auto_bas          : 1, /* true or false */
	 auto_n_major_tiks : 1, /* true or false */
	 auto_n_minor_tiks : 1, /* true or false */
	 major_tik_format  : 2, /* none, inside, outside, both */
	 minor_tik_format  : 2, /* none, inside, outside, both */
	 pos               : 3, /* left,right, bot,top, v_val, h_val */
	 label_align       : 2, /* left, center, right */
	 tik_align         : 2, /* left, center, right */
	 invalid_tiks      : 1, /* if TRUE, don't save bounders on reformat */
	 tight_label       : 1, /* if TRUE, move label with a floating axis */
	 tight_tiks        : 1, /* if TRUE, move tiks labels with a floating axis */
	 no_space          : 1, /* if TRUE, don't adjust chart extent by margin */
	 draw_behind       : 1, /* if TRUE, draw axis in BG slot (2D charts) */
	 use_tik_attr      : 1; /* if TRUE, use tik attr for tik marks */

	/* info on bars */
	double bar_margin;  /* 0 to 500% */
	double bar_overlap; /* -100% to 100% */
	int n_simple_bars;
	int n_stacked_bars;
	int *stacks;

	/* size of axis (used to position the next axis of the same type) */
	int x_offset;
	int y_offset;

	/* margin between tik and tik label */
	int tik_margin;

	/* axis position value */
	double pos_val;

	/* major tik info */
	int major_sz;
	int n_maj_steps;
	double maj_delta;
	double maj_step;

	/* minor tik info */
	int minor_sz;
	int n_min_steps;
	double min_delta;
	double min_step;

	/* generic tik placement info */
	double first_delta;

	/* tik formatting info (for catagory axes) */
	int filter_labels; /* all (0) or use display flags */
	char **tik_str;
	int   *display;
	int n;
	int c_tiks[2];  /* to display X ticks from Y Tick labels */
	
	/* tik formatting info (for value axes) */
	struct number_format tik_format;

	/* value axis min and max */
	double min_val;
	double max_val;
	double bas_val;

	/* axis attribute info */
	struct thing_attributes attr;           /* attributes for axis line  */
	struct thing_attributes tik_attr;       /* attributes for tik line   */
	struct thing_attributes label_attr;     /* attributes for axis label */
	struct thing_attributes tik_label_attr; /* attributes for tik labels */

	/* axis label */
	char *label_str;
} chart_axis_type;

/* data type for legend box */
typedef struct chart_legend_type {
	unsigned int
	 disabled           : 1,
	 x_space_eating     : 1,
	 y_space_eating     : 1,
	 text_before_sample : 1,
	 arranged_by_row    : 1,
	 auto_margins       : 1,
	 h_align            : 2, /* left, center, right */
	 v_align            : 2, /* top, middle, bottom */
	 title_align        : 2; /* left, center, right */
	int h_off, v_off;
	int max_per_row_or_column;
	int t_margin, b_margin, l_margin, r_margin, row_margin, column_margin;
	struct thing_attributes box_attr;
	struct thing_attributes label_attr;
	struct thing_attributes title_attr;
	char *title_str;
} chart_legend_type;

typedef struct chart_info_type {
	/* info on strata */
	int stacked_strata;

	/* info on bars */
	double bar_margin;  /* 0 to 500% */
	double bar_overlap; /* -100% to 100% */

	/* info on pies */
	int start_angle;

	/* default data type (for new groups) */
	int def_type;
} chart_info_type;

typedef struct chart_datasource_type {
	int type;
	int ix;
} chart_datasource_type;


typedef struct sheet_part_element {
	int type;
	int n;
	char *val;
	struct sheet_part_element **item;
} sheet_part_element;

typedef struct sheet_part {
	char *name;
	sheet_part_element item;
} sheet_part;

typedef struct sheet_data {
	int isa_val;		/* 0->null, 1->char 2->double */
	union {
		double val;
		char   *str;
	}u;
}sheet_data;

typedef struct chart_sheet_info {
    int        n_cols;
    int        n_rows;
	int        n_parts;
	int        n_col_sz;
	int        *col_sz;
	char       *range;
    char       **legends;
	char       **categories;
	sheet_data ***data;
	sheet_part **part;
} chart_sheet_info;

typedef struct chart_sheet_range  {
	int tl_row;
	int tl_col;
	int br_row;
	int br_col;
} chart_sheet_range;

/* generic data type for chart */
typedef struct chart_type {
	/* type info */
	unsigned int
	 type     : 4, /* PIE, CAT-VAL, VAL-VAL, HIST, SPIDER, STR-VAL, SURFACE */
	 old_type : 4,

	/* chart orientation */
	 orientation : 1, /* HORIZONTAL, VERTICAL */
	 old_orient  : 1, /* HORIZONTAL, VERTICAL */

	/* default treatment for nulls */
	 null_format : 2, /* zero, span, gap */

	/* title alignments (left, center, right) */
	 title_align    : 2,
	 subtitle_align : 2,
	 footer_align   : 2,

	/* title expansion */
	 title_expand    : 1,
	 subtitle_expand : 1,
	 footer_expand   : 1,

	/* chart decoration info */
	 close_top   : 1,
	 close_bot   : 1,
	 close_lft   : 1,
	 close_rt    : 1,
	 close_bg    : 1,
	 close_fg    : 1,
	 maj_x_grids : 1,
	 min_x_grids : 1,
	 maj_y_grids : 1,
	 min_y_grids : 1,
	 maj_z_grids : 1,
	 min_z_grids : 1,

	/* range supression flag */
	 no_ranges   : 1;

	unsigned int
	/* redraw flag */
	 needs_redraw: 1,
	 needs_object: 1;

	struct thing_attributes maj_x_attr;
	struct thing_attributes min_x_attr;
	struct thing_attributes maj_y_attr;
	struct thing_attributes min_y_attr;
	struct thing_attributes maj_z_attr;
	struct thing_attributes min_z_attr;

	/* axis info */
	int def_x_offset;
	int def_y_offset;
	int n_x_axes, old_n_x_axes;
	int n_y_axes, old_n_y_axes;
	int n_z_axes, old_n_z_axes;
	struct chart_axis_type *x_axis;
	struct chart_axis_type *y_axis;
	struct chart_axis_type *z_axis;

	/* overall extent, margins and adjusted extent */
	struct vector extent;
	struct vector margin;
	struct vector adj_ex;
	struct vector body;
	int depth;
	/* type specific info */
	struct chart_info_type info;

	/* titles */
	struct thing_attributes title_attr;
	struct thing_attributes subtitle_attr;
	struct thing_attributes footer_attr;
	char *title;
	char *subtitle;
	char *footer;
	int title_x_off;
	int title_y_off;
	int subtitle_x_off;
	int subtitle_y_off;
	int footer_x_off;
	int footer_y_off;
	
	/* data group info */
	int def_grp_type;
	int n_groups;
	struct chart_data_group_type *grp;
	int old_n_groups;
	struct chart_old_data_group_type *old_grp;

	/* legend format */
	struct chart_legend_type legend;

	/* default attribute */
	struct thing_attributes def_attr;

	/* 3D info */
	struct chart_3D_effect_type three_d;
	struct chart_3D_effect_type old_three_d;
	struct projection_volume volume;
	
	/* datasource */
	int    col_sliced;   /* flag to tell whether data is column sliced */
	struct chart_datasource_type data_src;
	struct chart_sheet_info *sheet;
	
	/* actual drawing of chart (links things to chart)
	 * order of things in group (an entry may be NULL)
	 * -frame background
	 * -minor horizontal grids
	 * -major horizontal grids
	 * -minor vertical grids
	 * -major vertical grids
	 * -minor z grids
	 * -major z grids
	 * -frame bg borders
	 * -bg axes
	 * -3D special effects layer
	 * -group data
	 * -group data labels
	 * -frame fg borders
	 * -axes
	 * -title
	 * -subtitle
	 * -footer
	 * -legend box */
	struct thing *t;
	struct thing *old_t;
	int old_shadow;
	/* name of chart */
	char name[STRPATH];
} chart_type;

/*****************************************************************************
 SS <===> GE interface data structures 
 to bypass C <==> ELF <==> C communication.
 ****************************************************************************/
typedef struct chart_datum {
    unsigned int use_label_x_offset  : 1,
                 use_label_y_offset  : 1,
                 use_label_string    : 1,
                 use_display_type    : 1,
                 use_x               : 1,
                 use_y               : 1,
                 use_z               : 1;
	int label_x_offset;
	int label_y_offset;
	char *label_string;
	int display_type; /* BG_PLOT, BG_SUBS, BG_SPAN, BG_GAP, BG_NULL */
	int nx; /* number of x's */
	int ny; /* number of y's */
	int nz; /* number of z's */
	double x[5];
	double y[5];
	double z[5];
} chart_datum;

typedef struct chart_number_format {
	unsigned int use_style                 : 1,
	             use_aux_style_info        : 1, 
                 use_units                 : 1,
                 use_radix_places          : 1,
                 use_prefix                : 1, 
                 use_suffix                : 1,
                 use_am_str                : 1,
                 use_pm_str                : 1,       
                 use_true_str              : 1,
                 use_false_str             : 1,
                 use_thousands_seperator   : 1,
                 use_decimal_seperator     : 1,
                 use_display_format_errors : 1,
                 use_trim_string           : 1;
	number_format num_fmt;
} chart_number_format;


typedef struct chart_trend {
	unsigned int use_type         : 1,
                 use_t_order      : 1,
                 use_r_squared    : 1,
                 use_correlations : 1;
	char   type[STRPATH];
	int    t_order;
	double r_squared;
	double correlations[13];
} chart_trend;

typedef struct chart_group {
	unsigned int use_type             : 1,
	             use_stack_id         : 1,
                 use_custom_part_name : 1,
                 use_x_axis           : 1,
                 use_y_axis           : 1,
                 use_z_axis           : 1,
                 use_label_x_offset   : 1,
                 use_label_y_offset   : 1,
                 use_label_type       : 1,
                 use_label_format     : 1,
                 use_chart_datum      : 1,
                 use_legend_string    : 1,
                 use_label_alignment  : 1,
                 use_data_range       : 1,
                 use_hidden           : 1,
                 use_trend            : 1;
	char type[STRPATH];
	int  stack_id;
	char custom_part_name[STRPATH];
	int  x_axis;
	int  y_axis;
	int  z_axis;
	int  label_x_offset;
	int  label_y_offset;
	char label_type[STRPATH];
	chart_number_format label_format;
	int  n;                               /* number of data points */
	chart_datum *data;
	char legend_string[STRPATH];
	int  label_alignment;
	char *data_range;
	int  hidden;
	chart_trend trend;		
} chart_group;

typedef struct chart_type_info {
	unsigned int use_bar_margin   : 1,
                 use_bar_overlap  : 1,
                 use_starting_pos : 1;
	int bar_margin;
	int bar_overlap;
	int starting_pos;
}chart_type_info;

typedef struct chart_axis_labels {
	unsigned int use_type            : 1,
                 use_tight_label     : 1,
                 use_label           : 1,	
                 use_label_x_margin  : 1,
                 use_label_y_margin  : 1,	
                 use_label_alignment : 1,
                 use_tik_format      : 1,
                 use_tight_tiks      : 1,
                 use_tik_range       : 1,
                 use_tik_labels      : 1,
                 use_tik_margin      : 1,
                 use_tik_alignment   : 1,
                 use_tik_filter      : 1,
                 use_tik_flags       : 1;
	char type[STRPATH];
	int tight_label;
	char label[STRPATH];
	int label_x_margin;
	int label_y_margin;
	int label_alignment;
	chart_number_format tik_format;
	int tight_tiks;
	char tik_range[STRPATH];
	char **tik_labels;
	int n_tik_labels;
	int tik_margin;
	int tik_alignment;
	int tik_filter;
	int *tik_flags;
	int n_tik_flags;
	
} chart_axis_labels;

typedef struct chart_axis_info {
	unsigned int use_value_type              : 1,
                 use_log10                   : 1,
                 use_auto_max                : 1,
                 use_max_value               : 1,
                 use_auto_min                : 1,
                 use_min_value               : 1,
                 use_auto_bas                : 1,
                 use_bas_value               : 1,
                 use_position                : 1,
                 use_value                   : 1,
                 use_minor_tik_type          : 1,
                 use_minor_tik_size          : 1,
                 use_auto_n_minor_tiks       : 1,
                 use_n_minor_tiks            : 1,
                 use_major_tik_type          : 1,
                 use_major_tik_size          : 1,
                 use_auto_n_major_tiks       : 1,
                 use_n_major_tiks            : 1,
                 use_tik_format              : 1,
                 use_label_x_margin          : 1,
                 use_label_y_margin          : 1,
                 use_tik_range               : 1,
                 use_tik_filter              : 1,
                 use_tik_labels              : 1,
                 use_tik_flags               : 1,
                 use_label                   : 1,
                 use_bar_margin              : 1,
                 use_bar_overlap             : 1,	
                 use_label_alignment         : 1,
                 use_tik_alignment           : 1,	
                 use_tik_margin              : 1,	
                 use_tight_label             : 1,
                 use_tight_tiks              : 1,	
                 use_use_tik_attributes      : 1,	
                 use_behind_data             : 1,
                 use_hidden                  : 1,	
                 use_inverted                : 1;	
	
	int    value_type;
	int    log10;
	int	   auto_max;
	double max_value;
	int	   auto_min;
	double min_value;
	int	   auto_bas;
	double bas_value;
	double position;
	double value;
	int    minor_tik_type;
	double minor_tik_size;
	int    auto_n_minor_tiks;
	int    n_minor_tiks;
	int    major_tik_type;
	double major_tik_size;
	int    auto_n_major_tiks;
	int    n_major_tiks;
	chart_number_format tik_format;
	int    label_x_margin;
	int    label_y_margin;
	char   tik_range[STRPATH];
	int    tik_filter;
	char   **tik_labels;
	int    *tik_flags;
	char   label[STRPATH];
	int    bar_margin;
	int    bar_overlap;
	int    label_alignment;
	int    tik_alignment;
	int    tik_margin;
	int    tight_label;
	int    tight_tiks;
	int    use_tik_attributes;
	int    behind_data;
	int    hidden;
	int    inverted;
}chart_axis_info;

typedef struct chart_decorations {
	unsigned int use_close_top        : 1,
                 use_close_bottom     : 1,
                 use_close_left       : 1,
                 use_close_right      : 1,
                 use_major_h_grids    : 1,
                 use_minor_h_grids    : 1,
                 use_major_v_grids    : 1,
                 use_minor_v_grids    : 1,
                 use_close_foreground : 1,
                 use_close_background : 1,
                 use_major_z_grids    : 1,
                 use_minor_z_grids    : 1;

	int close_top;
	int close_bottom;
	int close_left;
	int close_right;
	int major_h_grids;
	int minor_h_grids;
	int major_v_grids;
	int minor_v_grids;
	int close_foreground;
	int close_background;
	int major_z_grids;
	int minor_z_grids;
	
}chart_decorations;

typedef struct chart_legend {
	unsigned int use_disabled              : 1,
                 use_uses_x_space          : 1,
                 use_uses_y_space          : 1,
                 use_text_before_sample    : 1,	
                 use_arrange_by_row        : 1,
                 use_max_per_row_or_column : 1,
                 use_title                 : 1,	
                 use_horizontal_alignment  : 1,
                 use_horizontal_offset     : 1,
                 use_vertical_alignment    : 1,	
                 use_vertical_offset       : 1,
                 use_top_margin            : 1,
                 use_bottom_margin         : 1,	
                 use_row_margin            : 1,
                 use_left_margin           : 1,
                 use_right_margin          : 1,	
                 use_column_margin         : 1,
                 use_proportional_margins  : 1,
                 use_title_alignment       : 1,
                 use_label_range           : 1;

	int uses_x_space;
	int uses_y_space;
	int text_before_sample;
	int arrange_by_row;
	int max_per_row_or_column;
	char title[STRPATH];
	int horizontal_alignment;
	int horizontal_offset;
	int vertical_alignment;
	int vertical_offset;
	int top_margin;
	int bottom_margin;
	int row_margin;
	int left_margin;
	int right_margin;
	int column_margin;
	int proportional_margins;
	int title_alignment;
	int label_range;
} chart_legend;


typedef struct chart_titles {
	unsigned int use_title                      : 1,
                 use_title_alignment            : 1,
                 use_expand_margin_for_title    : 1,
                 use_title_x_offset             : 1,
                 use_title_y_offset             : 1,
                 use_subtitle                   : 1,
                 use_subtitle_alignment         : 1,
                 use_expand_margin_for_subtitle : 1,
                 use_subtitle_x_offset          : 1,
                 use_subtitle_y_offset          : 1,
                 use_footer                     : 1,
                 use_footer_alignment           : 1,
                 use_expand_margin_for_footer   : 1,
                 use_footer_x_offset            : 1,
                 use_footer_y_offset            : 1;

	char *title;
	int title_alignment;
	int expand_margin_for_title;
	int title_x_offset;
	int title_y_offset;
	char *subtitle;
	int subtitle_alignment;
	int expand_margin_for_subtitle;
	int subtitle_x_offset;
	int subtitle_y_offset;
	char *footer;
	int footer_alignment;
	int expand_margin_for_footer;
	int footer_x_offset;
	int footer_y_offset;
} chart_titles;

typedef struct chart_3D_effect {
	unsigned int use_enabled      : 1,
                 use_yaw          : 1,	
                 use_pitch        : 1,	
                 use_projection   : 1,	
                 use_depth        : 1,	
                 use_right_angles : 1;

	int enabled;
	int yaw;
	int pitch;
	int projection;	
	int depth;
	int right_angles;
}chart_3D_effect;

typedef struct chart_draw {
	char   name[STRPATH];
	int    suppress;
	int    got_size;
	vector extent;
} chart_draw;
