/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: cvtmain.c
   Applix ASC version: 82
   Date: Fri Apr 16 21:11:25 1999
*/
#include <stdio.h>
#include <signal.h>
#include <aos.h>
#include <atmattr.h>
#include <scac.h>
#include <getstr23.h>
#include <getstr10.h>
#include <impexp.h>
#include <atm.h>
#include <atmwin.h>
#include <axtext.h>
#include <atxint.h>
#include <doctypes.h>
#include <afile.h>
#include <ge.h>
#include <afileimg.h>
#include <cvt.h>
#include <savemode.h>
#include <fsrts.h>
#include <ax_des.h>
#include <cvtmain.p>


/****************************************************************************
 *
 * module: geBackupFile - copy contents of file 1 into file 2
 */
static int
geBackupFile(ge, read_file, write_file)
graphics_main_type *ge;
char *read_file;
char *write_file;
{
	int err, buf_len, read_only;
	FILE *fp1, *fp2;
	char *buf;

	/* set the locks */
	read_only = TRUE;
	err = AFTaskLockFile(read_file, ge->pid, &read_only, FALSE);
	if ( err != AOK )
		return(err);
	read_only = FALSE;
	err = AFTaskLockFile(write_file, ge->pid, &read_only, FALSE);
	if ( err != AOK )
	{
		AFTaskUnlockFile(read_file, ge->pid, TRUE);
		return(err);
	}

	/* open the existing file */
	fp1 = AFTaskfopen(ge->pid, read_file, READ_BINARY_STR);
	if ( !fp1 )
		err = FEErrno+errno;
	else
	{
		/* get the size of file */
		fseek(fp1, 0L, 2);
		buf_len = ftell(fp1);

		/* empty file ?? */
		if ( buf_len <= 0 )
		{
			AFfclose(read_file, FALSE, fp1);
			fp2 = AFTaskfopen(ge->pid, write_file, READ_TRUNC_STR);
			if ( !fp2 )
				err = FEErrno+errno;
			else if ( AFfclose(write_file, FALSE, fp2) != 0 )
				err = FEErrno+errno;
		}
		else /* a valid document */
		{
			buf = (char *)gloc(ge, buf_len+1);
			fseek(fp1, 0L, 0);
			if ( fread(buf, buf_len, 1, fp1) != 1 )
				err = FEErrno+errno;
			if ( AFfclose(read_file, FALSE, fp1) != 0 )
				err = FEErrno+errno;
	
			/* if read is succesfull */
			if ( err == AOK )
			{
				fp2 = AFTaskfopen(ge->pid, write_file, READ_TRUNC_STR);
				if ( !fp2 )
					err = FEErrno+errno;
				else
				{
					/* write the buffer to disk */
					if ( fwrite(buf, buf_len, 1, fp2) != 1 )
						err = FEErrno+errno;
	
					if ( err == AOK )
					{
						/* done */
						if ( AFfclose(write_file, FALSE, fp2) != 0 )
							err = FEErrno+errno;
					}
					else
						AFfclose(write_file, FALSE, fp2);
				}
			}
			gfree(ge, buf);
		}
	}

	/* unlock the files and return */
	AFTaskUnlockFile(read_file,  ge->pid, TRUE);
	AFTaskUnlockFile(write_file, ge->pid, FALSE);
	return(err);
}

/****************************************************************************
 *
 * module:
 */
geSetFilePermission(ge, gdoc_info)
graphics_main_type *ge;
ge_file_info_type *gdoc_info;
{
	int pmode;

	pmode = AxChmodValueFromCodes(gdoc_info->pathname,
								  gdoc_info->grp_access,
								  gdoc_info->all_access);
	chmod(gdoc_info->pathname, pmode);
}

/****************************************************************************
 *
 * module: geDoFileSave - write the buffer to a file
 */
geDoFileSave(ge, gdoc_info)
graphics_main_type *ge;
ge_file_info_type *gdoc_info;
{
	char *p, dir[STRPATH+1], backup[STRPATH+1];
	int err, b_err, recoverable;
	FILE *fp;

	/* copy the contents of the original file to a backup file */
	if ( gdoc_info->on_disk == FALSE ) /* no backup needed */
	{
		recoverable = FALSE;
		err = AOK;
	}
	else /* file exists - make a backup copy */
	{
		AxParsePath(gdoc_info->pathname, dir, backup);
		if ( gdoc_info->save_mode == GE_SAVE_AS_AX_BITMAP )
			p = NXLT("/px%%%%%%%%.im");
		else
			p = NXLT("/gr%%%%%%%%.ag");
		if ( AxTasklessTempFile(strcat(dir, p), backup) == 0 )
			err = FEErrno+errno;
		else
		{
			err = geBackupFile(ge, gdoc_info->pathname, backup);
			if ( err != AOK )
				AxDeleteTasklessTempfile(backup);
		}
		recoverable = (err == AOK) ? TRUE : FALSE;
	}

	/* write the file to disk */
	if ( err == AOK )
	{
		fp = AFTaskfopen(ge->pid, gdoc_info->pathname, READ_TRUNC_STR); /* r+ */
		if ( !fp )
			err = FEErrno+errno;
		else
		{
			/* write the buffer to disk */
			if ( fwrite(gdoc_info->buf, gdoc_info->buf_len,1,fp) != 1 )
				err = FEErrno+errno;

			if ( err == AOK )
			{
				/* close the file (set permissions, too) */
				if ( AFfclose(gdoc_info->pathname, FALSE, fp) != 0 )
					err = FEErrno+errno;
				else /* set permissions */
				{
					geSetFilePermission(ge, gdoc_info);
					err = AOK;
				}
			}
			else
				AFfclose(gdoc_info->pathname, FALSE, fp);
		}

		/* done */
		if ( recoverable )
		{
			if ( err == AOK )
				AxDeleteTasklessTempfile(backup);
			else
			{
				b_err = geBackupFile(ge, backup, gdoc_info->pathname);
				if ( b_err == AOK )
					AxDeleteTasklessTempfile(backup);
			}
		}
	}
	return(err);
}

/****************************************************************************
 *
 * module:	G C t h i m b l e - GECVT main routine
 */
GCthimble(ge, gdoc_info, write_mode, fpage_only, do_encryption)
graphics_main_type *ge;
ge_file_info_type *gdoc_info;
int write_mode;
int fpage_only;
int do_encryption;
{
	int err, buf_len, hdr_len, tmp_keys, valid_buf;
	cvt_global_type cvt;
	key_tables *ktab;
	key_ring *kr;
	char *buf, hdr[STRPATH+1], *tmp_buf;
	AFileInfo f_info;

	/* set the inital state */
	mlInit(ge, &cvt);
	cvt.fileName = (char *)0;
	cvt.unixName = (char *)0;

	if ( write_mode == FALSE )
	{
		err = AxDecryptBuffer(gdoc_info->buf, gdoc_info->buf_len, ge->pid);
		if ( err == AOK )
		{
			/* convert the ascii data in the buffer to binary */
			cvt.inputBufferSize = gdoc_info->buf_len;
			err = grRead(ge, &cvt, gdoc_info->buf, fpage_only);
			if ( cvt.delete_file )
				unlink(cvt.fileName);
		}
	}
	else  /* saving an existing document */
	{
		if ( gdoc_info->save_mode == GE_UNCOMPRESSED_HEX ||
			 gdoc_info->save_mode == GE_UNCOMPRESSED_BIT6 )
			cvt.compression = COMP_RAW;
		if ( gdoc_info->save_mode == GE_UNCOMPRESSED_HEX ||
			 gdoc_info->save_mode == GE_COMPRESSED_HEX )
			cvt.encoding = HEX;
		/* is a save allowed? */
		if ( gdoc_info->read_only )
			return(FEEACCES);

		/* get password for encryption */
		if ( do_encryption )
		{
			tmp_keys = AxGetWKeys(&ktab, &kr);
			cvt.info.encrypted = (kr != (key_ring *)0) ? TRUE : FALSE;
		}
		else
			cvt.info.encrypted = FALSE;

		/* copy the binary data onto a buffer */
		if ( gdoc_info->save_mode == GE_SAVE_AS_AX_BITMAP )
			err = imWrite(ge, &cvt, &gdoc_info->buf, &gdoc_info->buf_len);
		else
			err = grWrite(ge, &cvt, &gdoc_info->buf, &gdoc_info->buf_len);

		/* encrypt the data in place (skipping the header and footer) */
		if ( cvt.info.encrypted )
		{
			/* get the data ptr and count */
			AFileGetBuffer(gdoc_info->buf, gdoc_info->buf_len,
			               &buf, &buf_len);

			/* encrypt the data in place */
			AxEncrypt(ktab, kr, (unsigned char *)buf, buf_len);
			if ( tmp_keys )
				AxDestroyKeys(&kr);
		}

		/* write the buffer to disk */
		if ( err == AOK && gdoc_info->save_to_disk )
		{
			/* overwrite the original file */
			err = geDoFileSave(ge, gdoc_info);
			gfree(ge, gdoc_info->buf);
			gdoc_info->buf_len = 0;
			gdoc_info->buf = (char *)0;
		}
	}
	/* done */
	mlCleanupCvt(ge, &cvt);
	return(err);
}

/****************************************************************************
 *
 * module: G E O p e n F i l e 
 */
GEOpenFile(ge, gdoc_info, fp, fpage_only)
graphics_main_type *ge;
ge_file_info_type *gdoc_info;
FILE *fp;
int fpage_only;
{
	int err;

	/* get the size of the data */
	fseek(fp, 0L, 2);
	gdoc_info->buf_len = ftell(fp);
	gdoc_info->buf = (char *)gloc(ge, gdoc_info->buf_len+1);

	/* read the data from the file */
	fseek(fp, 0L, 0);
	if ( fread(gdoc_info->buf, gdoc_info->buf_len, 1, fp) != 1 )
		err = FEErrno+errno;
	else
	{
		gdoc_info->buf[gdoc_info->buf_len] = 0;

		/* convert the ascii info to a binary ge */
		err = GCthimble(ge, gdoc_info, FALSE, fpage_only, FALSE);
		if ( err == AOK )
			gResetPage(ge);
	}
	/* done */
	gfree(ge, gdoc_info->buf);
	gdoc_info->buf_len = 0;
	gdoc_info->buf = (char *)0;

	return(err);
}

/****************************************************************************
 *
 * module:	c v t F o r X f e r I n
 */
cvtForXferIn(ge, t_ge, buf, byte_count, outline_mode)
graphics_main_type *ge;
graphics_main_type *t_ge;
char *buf;
int byte_count;
int outline_mode;
{
	ge_file_info_type gdoc_info;
	char outfile[STRPATH+1], path[STRPATH+1];
	cvt_global_type cvt;
	int err, filtered;
	AFileInfo info;
	FILE *fp;

	/* check header of buffer. */
	err = AFileInfoFromHeader(buf, &info);
	if (  err == AOK )
	{
		if ( info.docType != AX_DOC_TYPE_ASTERX_GRAPHICS )
			err = !AOK;
		else
		{
			/* copy the binary data from a buffer to a GE structure */
			memset((char *)&gdoc_info, 0, sizeof(gdoc_info));
			gdoc_info.buf = buf;
			gdoc_info.buf_len = byte_count;
			err = GCthimble(t_ge, &gdoc_info, FALSE, FALSE, FALSE);
		}
	}
	else if ( err == FETooNew )
	{
		/* create a temp file */
		if ( AxTempFile(NXLT("zz%%%%%%"), path) == 0 )
			err = FEErrno+errno;
		else
		{
			/* write the buffer to file */
			fp = fopen(path, WRITE_STR);
			if ( !fp )
				err = FEErrno+errno;
			else
			{
				if ( fwrite(buf, byte_count, 1, fp) != 1 )
				{
					err = FEErrno+errno;
					fclose(fp);
				}
				else /* successful write */
				{
					if ( fclose(fp) != 0 )
						err = FEErrno+errno;
					else
						err = AOK;
				}
			}

			/* buff now in file */
			if ( err == AOK )
			{
				/* filter the file to current revision */
				filtered = geFilterDocToCurRev(ge, path, outfile);
				if ( filtered == FALSE )
					err = FETooNew;
				else 
				{
					/* get the size of the data */
					fp = fopen(outfile, READ_BINARY_STR);
					if (!fp )
						err = FEErrno+errno;
					else
					{
						memset((char *)&gdoc_info, 0, sizeof(gdoc_info));
						err = GEOpenFile(t_ge, &gdoc_info, fp, FALSE);
						fclose(fp);
					}
					AxDeleteTempfile(outfile);
				}

				/* done - delete */
				AxDeleteTempfile(path);
			}
		}
	}
	return(err);
}
