/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: fb.h
   Applix ASC version: 23
   Date: Mon Nov 16 13:56:00 1998
*/
#ifndef FB_H_DEFINED
#define FB_H_DEFINED

#if GE_THIMBLE
#define UNDO_NONE   	0
#define UNDO_BUFFER 	1
#define UNDO_FILE   	2
#define UNDO_COLORMAP   3
#define UNDO_IMAGE   	4
#endif

#ifdef DEBUG 
#if DEBUG 
#undef NDEBUG 
#else 
#undef NDEBUG
#define NDEBUG
#endif 
#endif

#define CROP      (-1)
#define CELL        0
#define PREV        1
#define MAX_REGIONS 2

#define CROP_LEFT	0
#define CROP_RIGHT	1
#define CROP_TOP	2
#define CROP_BOT	3

#define X_UNIT	500
#define Y_UNIT	500

#define TOOL_XCELL	6
#define TOOL_YCELL	6

#define MAX_XCELL	14
#define MAX_YCELL	14
#define MIN_XCELL	2
#define MIN_YCELL	2

#define SMALL_XCELL	10
#define SMALL_YCELL	10
#define SMALLER_XCELL	6
#define SMALLER_YCELL	6

#define XSEP		50
#define YSEP		50
#define XCROP		300
#define YCROP		300

#define MIN_WINX	6000
#define MIN_WINY	(MIN_WINX / 2 + 1000)
#define NUM_TOOLS	18
#define DEF_TOOL	0
#define TOOL_HEIGHT	5

/* pen styles */
#define PAINT_PEN   0
#define WASH_PEN    1
#define TOGGLE_PEN  2
#define CERASE_PEN  3
#define ERASE_PEN   4
#define BLEND_PEN   5

/* pen modes */
#define PX_STROKE   0
#define PX_LINE     1
#define PX_FLOOD    2
#define PX_RECT     3
#define PX_ELL      4
#define PX_GRAB     6
#define PX_SPRAY    7
#define PX_DISP     24
#define PX_DRAG     25
/* macros */

#ifdef NDEBUG
#define SHOWRECT(r)
#else
#define SHOWRECT(r)	showRect(r)
#endif

#define BYTE_AT(x,y,m)	((long) (y) * (m) + (x >> 3))
#define IN_RECT(x,y,r)	(r->x1 <= x && x < r->x2 && r->y1 <= y && y < r->y2)

/***********************************************************************/
#ifdef MAIN 	/* || GE_THIMBLE */
unsigned char mask[8] = {
	0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01
};
unsigned char tool[NUM_TOOLS][TOOL_HEIGHT] = {
	{ 0x00, 0x00, 0x20, 0x00, 0x00 },
	{ 0x00, 0x20, 0x70, 0x20, 0x00 },
	{ 0x00, 0x70, 0x70, 0x70, 0x00 },
	{ 0x20, 0x70, 0xF8, 0x70, 0x20 },
	{ 0x70, 0xF8, 0xF8, 0xF8, 0x70 },
	{ 0xF8, 0xF8, 0xF8, 0xF8, 0xF8 },
	{ 0x00, 0x20, 0x20, 0x20, 0x00 },
	{ 0x00, 0x00, 0x70, 0x00, 0x00 },
	{ 0x00, 0x10, 0x20, 0x40, 0x00 },
	{ 0x00, 0x40, 0x20, 0x10, 0x00 },
	{ 0x00, 0x50, 0x20, 0x50, 0x00 },
	{ 0x20, 0x20, 0x20, 0x20, 0x20 },
	{ 0x00, 0x00, 0xF8, 0x00, 0x00 },
	{ 0x20, 0x20, 0xF8, 0x20, 0x20 },
	{ 0x08, 0x10, 0x20, 0x40, 0x80 },
	{ 0x80, 0x40, 0x20, 0x10, 0x08 },
	{ 0x88, 0x50, 0x20, 0x50, 0x88 },
	{ 0x50, 0x00, 0xa8, 0x00, 0x50 },
};
#else	/* #ifdef MAIN */
extern unsigned char mask[8], tool[NUM_TOOLS][TOOL_HEIGHT];
#endif	/* #ifdef MAIN #else */

/****************************************************************************\
*                     pixel editor typedefs                                  *
\****************************************************************************/
typedef struct region_image_type {
    char *background;
    char *image;
	char *mask;
    int width, height;
    int clipX, clipY, clipWid, clipHyt;
    int oldX, oldY;
    int clippedWid, clippedHyt;
} region_image_type;

typedef struct selected_region_type {
	struct thing *t;
	struct region_image_type cell;
	struct region_image_type prev;
	int x_off, y_off;
	int x, y, wid, hyt;
	unsigned char *data;
	unsigned char *mask;
} selected_region_type;

typedef struct px_main_type {
	unsigned int undisplayed      : 1,
	             color_from_image : 2,
	             four_connected   : 1,
	             zoomed           : 1,
	             normal           : 1,
	             squeezed         : 1,
	             changed          : 1,
	             edited           : 1,
	             color            : 1,
	             flooding         : 1,
	             erasing          : 1,
	             filling          : 1,
	             soft_filling     : 1,
	             grid_lines       : 1,
	             stroke_aborted   : 1,
	             rubber_banded    : 1,
	             erase_trans      : 1,
	             wash_trans       : 1,
	             tmp_fg_trans     : 1,
	             tmp_bg_trans     : 1,
	             fg_trans         : 1,
	             bg_trans         : 1,
	             geColorpalOn     : 1;

	/* brush info */
	int brush_wid;
	int brush_hyt;
	unsigned char *brush_dat;
	int mode, temp_mode;
	int pen_style;
	int toggle_style;
	int toggle_state;
	int spray_seed;
	unsigned char spray_density;
	struct TmRect32 bounds;
	struct TmPoint32 point[65];
	int nPoints;

	/* stroke info */
	struct TmPoint32 first;
	struct TmPoint32 last;
	struct TmPoint32 first_cell;
	struct TmPoint32 last_cell;
	int stroke_mode;

	/* color info */
	int erase_color;
	int wash_color;
	int temp_fg;
	int fg;    
	int temp_bg;
	int bg;
	attr_type attr;

	/* image info */
	struct thing *t;
	poly_type  org_trans;
	int img_ix;
	unsigned char *img[2];
	unsigned char *msk[2];
	int img_bytes, limg_bytes, img_lag, limg_lag;
	int msk_bytes, lmsk_bytes, msk_lag, lmsk_lag;
	int xSize, ySize;
	int modulo;
 
	/* selction info */
	struct selected_region_type sel;

	/* window and cell info */
	int xCrop,  yCrop;
	int xSep,   ySep;
	int xCell,  yCell;
	int xCellO, yCellO;
	int xCellS, yCellS;
	int xBorg,  yBorg;
	int xOff,   yOff;
	int xPos,   yPos;
	int xWin,   yWin;
	int xNum,   yNum;
	int rez, x_rez, y_rez;
	int event_region;
	int h_scroll, v_scroll;
	struct vector region[2];
	struct vector sqr;

	/* cropping info */
	int cropHit;
	int cropTop, cropLeft, cropBot, cropRight;
	struct vector cropBox[4];
	int cropped;

	/* undo info */
	int undo_mode;
	unsigned char *undo_img;
	unsigned char *undo_msk;
	unsigned char *undo_data;
	unsigned char *undo_mask;
 
	cmap_type undo_cmap;
	cmap_type orig_cmap;
	struct thing *orig_t;
	int image_convert;
	int orig_color;
	int pixel_table[256];

	thing_attributes def_attr;
} px_main_type;

/****************************************************************************/

#endif /* FB_H_DEFINED */

