/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: ge.h
   Applix ASC version: 343
   Date: Fri Jan 29 21:13:38 1999
*/
#ifndef GE_H_DEFINED
#define GE_H_DEFINED

#ifndef TURBO
#define TURBO 0
#endif
#ifndef DOWORK
#define DOWORK 0
#endif
#define ALLOW_JAVA_GE 1

#ifdef DEBUG
#if DEBUG
#undef NDEBUG
#else
#ifndef NDEBUG
#define NDEBUG
#endif
#endif
#endif

/* redefine template as it is a C++ keyword */
#define template template_t

/****************************************************************************\
*                  graphics editor macros                                    *
\****************************************************************************/
#undef MIN
#undef MAX
#define MIN(a, b) ((a) < (b) ? (a) : (b))
#define MAX(a, b) ((a) > (b) ? (a) : (b))

#define LE(a, b)  ((a) <= (b))
#define LT(a, b)  ((a) <  (b))

#define STRNEQ(a, b, n) (strncmp(a, b, n) == 0)
#define STREQ(a, b)     (strcmp(a, b) == 0)

#ifdef ABS
#undef ABS
#endif
#define ABS(a)       ((a) < 0 ? -(a) : (a))
#define ABSVAL(a, b) if ((a = b) < 0) { a = -a; }
#define SWAP(a, b)   { register int tmp; tmp = a; a = b; b = tmp; }

#define COUNT(s)    (sizeof(s) / sizeof(*s))

#define MID(a, b)   (((long) (a) + (long) (b)) / 2)

#define VECTFIX(v)  { if ((v)->x2 == SAME) (v)->x2 = (v)->x1; \
if ((v)->y2 == SAME) (v)->y2 = (v)->y1; }

#define FULL_ELL(t) ((t)->data.ell->start_angle == 0 && (t)->data.ell->end_angle == MAXANG)

#ifndef GECVT
#define XLENTOVP(s)   (double)( (double)(s) * ((double)ge->xform.x_res/1000.0) )
#define YLENTOVP(s)   (double)( (double)(s) * ((double)ge->xform.y_res/1000.0) )
#define XTOVP(x)      (XLENTOVP(x - ge->state.win_orig.x1) + ge->state.vp.x1)
#define YTOVP(y)      (YLENTOVP(y - ge->state.win_orig.y1) + ge->state.vp.y1)
#define XLENTOWIN(s)  (double)( (double)(s) * (1000.0/(double)ge->xform.x_res) )
#define YLENTOWIN(s)  (double)( (double)(s) * (1000.0/(double)ge->xform.y_res) )
#define XTOWIN(x)     (XLENTOWIN(x - ge->state.vp.x1) + ge->state.win_orig.x1)
#define YTOWIN(y)     (YLENTOWIN(y - ge->state.vp.y1) + ge->state.win_orig.y1)
#else /* for the converters (don't use scale factors or display offsets) */
#define XLENTOVP(s)  (double)(s)
#define YLENTOVP(s)  (double)(s)
#define XTOVP(x)     (XLENTOVP((x) - ge->state.win_orig.x1) + ge->state.vp.x1)
#define YTOVP(y)     (YLENTOVP((y) - ge->state.win_orig.y1) + ge->state.vp.y1)
#define XLENTOWIN(s) (double)(s)
#define YLENTOWIN(s) (double)(s)
#define XTOWIN(x)    (XLENTOWIN((x) - ge->state.vp.x1) + ge->state.win_orig.x1)
#define YTOWIN(y)    (YLENTOWIN((y) - ge->state.vp.y1) + ge->state.win_orig.y1)
#endif

#define COORD(a)    (* ((coord *) &(a)))

#define ISA(type, t) (t->type == type)

#define ASSERT(ex)  { if (!(ex)) badAssert(__FILE__, __LINE__); }
#define ASSURE(act, op, val)    ASSERT((int) (act) op (int) val)
#define LOGGER(body)

/* ge filing stuff */
#define FILE_ID1 "V1.02 Alis Graphics Editor"
#define FILE_ID2 "V1.02 Asterix Graphics Editor"
#define FILE_ID3 "V1.10 Asterix Graphics Editor"

/* grid sizes */
#define POINTS      14
#define PICAS       83

#define SIXTEENTH   63
#define TENTH      100
#define EIGHTH     125
#define QUARTER    250
#define HALF       500

#define MILLIMETER  39
#define CENTIMETER 394

#define PT_BLOCK_SIZE 256

/* modes for GE */
#define GE_MODE 0
#define PX_MODE 1
#define GP_MODE 2

/* presentation and page modes */
#define START_MODE  0
#define SLIDE_ED    1
#define SLIDE_MSTR  2
#define SLIDE_SORT  3
#define HOUT_MSTR   4
#define OLIN_ED     5
#define OLIN_MSTR   6
#define SNOTES_ED   7
#define SNOTES_MSTR 8
#define BLANK_MODE  9
#define TITLE_MSTR  10

/* Arrow types */
#define ARROW_1   0
#define ARROW_2   1
#define ARROW_3   2
#define ARROW_4   3
#define ARROW_5   4
#define ARROW_6   5
#define ARROW_7   6
#define ARROW_8   7
#define ARROW_9   8
#define ARROW_10  9
#define ARROW_11 10
#define ARROW_12 11
#define ARROW_13 12
#define ARROW_14 13

/* marker types */
#define MRKR_NONE 0
#define MRKR_01   1
#define MRKR_02   2
#define MRKR_03   3
#define MRKR_04   4
#define MRKR_05   5
#define MRKR_06   6
#define MRKR_07   7
#define MRKR_08   8
#define MRKR_09   9
#define MRKR_10  10
#define MRKR_11  11
#define MRKR_12  12
#define MRKR_13  13
#define MRKR_14  14

/* types of template objects */
#define TEMPLATE_OTHER  0
#define TEMPLATE_TITLE  1 
#define TEMPLATE_BLIST  2
#define TEMPLATE_GRAFX  3
#define TEMPLATE_CHART  4
#define TEMPLATE_SLIDE  5
#define TEMPLATE_NOTES  6
#define TEMPLATE_HOUTS  7
#define TEMPLATE_S1     8
#define TEMPLATE_S2     9
#define TEMPLATE_N1    10
#define TEMPLATE_N2    11
#define TEMPLATE_NONE  12
#define TEMPLATE_TABLE 13
#define TEMPLATE_NUM_LIST 14
 
/* template layout types */ 
#define TITLE_PAGE           0
#define BASIC_PAGE           1
#define TWO_COLUMN_PAGE      2
#define TWO_ROW_PAGE         3
#define SPLT_COL_AND_COL_PG  4
#define COL_AND_SPLT_COL_PG  5
#define SPLT_ROW_OVR_ROW_PG  6
#define ROW_OVR_SPLT_ROW_PG  7
#define FOUR_OBJECTS_PAGE    8
#define BLANK_PAGE           9
#define BIG_OBJECT_PAGE      10
 
/* template size types */ 
#define TMP_TITLE_1   0
#define TMP_TITLE_2   1
#define TMP_SUBTITLE  2
#define TMP_FULLSIZE  3
#define TMP_COLUMN_1  4
#define TMP_COLUMN_2  5
#define TMP_ROW_1     6
#define TMP_ROW_2     7
#define TMP_SMALL_1   8
#define TMP_SMALL_2   9
#define TMP_SMALL_3  10
#define TMP_SMALL_4  11
#define TMP_S1       12
#define TMP_S2       13
#define TMP_N1       14
#define TMP_N2       15
#define TMP_H2_1     16
#define TMP_H2_2     17
#define TMP_H3_1     18
#define TMP_H3_2     19
#define TMP_H3_3     20
#define TMP_H4_1     21
#define TMP_H4_2     22
#define TMP_H4_3     23
#define TMP_H4_4     24
#define TMP_H5_1     25
#define TMP_H5_2     26
#define TMP_H5_3     27
#define TMP_H5_4     28
#define TMP_H5_5     29
#define TMP_H6_1     30
#define TMP_H6_2     31
#define TMP_H6_3     32
#define TMP_H6_4     33
#define TMP_H6_5     34
#define TMP_H6_6     35
#define TMP_FULLPAGE 36

/* copied from atmtxtin.h to here */
#define STK_NONE        0
#define STK_PEN         1
#define STK_ONE_LINE    2
#define STK_TWO_LINE    3
#define STK_RECT        4
#define STK_ELLIPSE     5
#define STK_HVLINE      6
#define STK_MLINE       7
#define STK_MRECT       8
#define STK_SQUARE      9
#define STK_CIRCLE     10
#define STK_MAP        11
#define STK_ONE_CURVE  12
#define STK_TWO_CURVE  13
#define STK_WRECT      14
#define STK_R_RECT     15
#define STK_R_SQR      16
#define STK_R_POLY     17
#define STK_R_SQR_POLY 18
#define STK_CURVE_END  19

/* Chart object types */
#define ISA_NOTHING  0
#define ISA_CHART    1
#define ISA_TITLE    2
#define ISA_SUBTITLE 3
#define ISA_FOOTER   4
#define ISA_LEGEND   5
#define ISA_AXIS     6
#define ISA_AXLAB    7
#define ISA_DATGRP   8
#define ISA_DATLAB   9
#define ISA_BODY    10
#define ISA_GRID    11

#define ISA_X_AXIS  50
#define ISA_Y_AXIS  51
#define ISA_Z_AXIS  52
#define SMALL_DOC_PAGE_COUNT 4

/****************************************************************************\
*                  graphics editor typedefs                                  *
\****************************************************************************/

typedef unsigned char  ubyte;
#ifndef BOOL_DEFINED
typedef int bool;
#define BOOL_DEFINED 1
#endif

typedef struct coord {
    int x1, y1;
} coord;
typedef struct vector {
    int x1, y1, x2, y2;
} vector;

typedef struct ge_source_info {
	unsigned int use_page_ix : 1,
	             use_x       : 1,
	             use_y       : 1,
	             use_wid     : 1,
	             use_hyt     : 1;
	int page_ix;
	int x;
	int y;
	int wid;
	int hyt;
} ge_source_info;

typedef struct ge_mouse_info {
	unsigned int
	 relative  : 1, /* relative to window  */
	 b1        : 1, /* button 1 pressed    */
	 b2        : 1, /* button 2 pressed    */
	 b3        : 1, /* button 3 pressed    */
	 shifted   : 1, /* shift key pressed   */
	 controled : 1, /* control key pressed */
	 alted     : 1; /* alt key pressed     */
	int x;
	int y;
} ge_mouse_info;

typedef struct ge_audio_type {
    unsigned int play:1,
                 type:2;
    int id;
    int volume;
} ge_audio_type;

typedef struct ge_hdrftr_str_type {
    char *string;
    int face_ix;
    int pointsize;
    int color_ix;
    unsigned int bold                 : 1, /* TRUE or FALSE */
                 italic               : 1, /* TRUE or FALSE */
                 strike_thru          : 1, /* TRUE or FALSE */
                 underline            : 2, /* none(0), single(1), double(2) */
                 underline_words_only : 1; /* TRUE or FALSE */
} ge_hdrftr_str_type;

typedef struct ge_hdrftr_line_type {
    ge_hdrftr_str_type inner;
    ge_hdrftr_str_type center;
    ge_hdrftr_str_type outer;
} ge_hdrftr_line_type;

typedef struct ge_hdrftr_page_type {
    int hdr_offset;
    int ftr_offset;
    int hdr_n_lines;
    struct ge_hdrftr_line_type *hdr;
    int ftr_n_lines;
    struct ge_hdrftr_line_type *ftr;
} ge_hdrftr_page_type;

typedef struct ge_hdrftr_main_type {
    unsigned int use_even_odd : 1, /* TRUE or FALSE */
                 use_first    : 1, /* TRUE or FALSE */
                 use_final    : 1; /* TRUE or FALSE */
    ge_hdrftr_page_type base;
    ge_hdrftr_page_type even;
    ge_hdrftr_page_type odd;
    ge_hdrftr_page_type first;
    ge_hdrftr_page_type final;
} ge_hdrftr_main_type;

typedef struct ge_arrow_info_type {
    coord p;
    int dx;
    int dy;
    long len;
} ge_arrow_info_type;

typedef struct attr_flags { 
    /* fill attributes */
    unsigned int fill_pat       : 1,
                 fpat_type      : 1,
                 fill_fg        : 1,
                 fill_bg        : 1,

    /* line attributes */
                 line_pat       : 1,
                 lpat_type      : 1,
                 line_fg        : 1,
                 line_bg        : 1,
                 mode           : 1,
                 line_style     : 1,
                 line_thickness : 1,
                 left_arrow     : 1,    
                 right_arrow    : 1,
                 join_symbols   : 1, 

    /* drop shadow attributes */
                 shadow         : 1,
                 shadow_color   : 1,
                 shadow_x       : 1,
                 shadow_y       : 1,

    /* text attributes */
                 face           : 1,
                 size           : 1,
                 ruling         : 1,
                 strike         : 1,
                 weight         : 1,
                 slant          : 1,
                 x_scale        : 1,
                 shear          : 1,
                 angle          : 1,
	             x_off          : 1,
                 y_sub_sup      : 1;

	/* paragraph attributes */
	unsigned int l_indent1      : 1,
	             l_indent2      : 1,
	             r_indent       : 1,
                 bullet         : 1,
                 bullet_scale   : 1,
                 bullet_color   : 1,
                 bullet_font    : 1,
                 bullet_glyph   : 1,
                 space_above    : 1,
                 space_below    : 1,
                 space_inside   : 1,

    /* text field attributes */
                 h_just         : 1,
                 v_just         : 1,
                 field_space    : 1,
                 field_x_scale  : 1,
                 field_y_scale  : 1,
                 field_shear    : 1,
                 field_angle    : 1,
                 l_margin       : 1,
                 r_margin       : 1,
                 t_margin       : 1,
                 b_margin       : 1,

    /* style id */
                 style          : 1,

    /* elf callback id */
                 callback       : 1,

    /* user defined tag field */
                 tag            : 1;
} attr_flags;

typedef struct thing_attributes {
    /* flag list of active attributes */
    attr_flags use; 

    /* fill attributes */
    int fill_pat;
    int fpat_type;
    int fill_fg;
    int fill_bg;

    /* line attributes */
    int line_pat;
    int lpat_type;
    int line_fg;
    int line_bg;
    int mode;
    int line_style;
    int line_thickness;
    int left_arrow;
    int right_arrow;
    int join_symbols;

    /* drop shadow attributes */
    int shadow;
    int shadow_color;
    int shadow_x;
    int shadow_y;

    /* text attributes */
    int face;
    int size;
    int ruling;
    int strike;
    int weight;
    int slant;
    int x_scale;
    int shear;
    int angle;
    int x_off;
    int y_sub_sup;

	/* paragraph attributes */
	int l_indent1;
	int l_indent2;
	int r_indent;
    int bullet;
	int bullet_scale;
	int bullet_color;
	int bullet_font;
	int bullet_glyph;
	int space_above;
	int space_below;
    int space_inside;

    /* text field attributes */
    int h_just;
    int v_just;
    int field_space;
    int field_x_scale;
    int field_y_scale;
    int field_shear;
    int field_angle;
    int l_margin;
    int r_margin;
    int t_margin;
    int b_margin;

    /* style id */
    int style;

    /* elf callback id */
    int callback;

    /* user definable tag field */
    int tag;
} thing_attributes;

typedef struct gr_group_type {
    int num;
    int sub_ix;
    struct thing **list;
} gr_group_type;

typedef struct gr_textbox_type {
    struct poly_type trans;
    struct thing **strings;
    int n_strings;
    int x_off;
    int y_off;
    int field_wid;
    coord fixed_pos;
    int v_align; /* 0 - top, 1- center, 2 - bottom */
	int def_tabsize;
	int n_tabs;
	int *tabstops;
    unsigned int wrap_txt : 1,
                 sub_txt  : 1;
} gr_textbox_type;

typedef struct gr_image_type {
    struct poly_type trans;
    struct ge_image_type info;
    int last_x_rez;
    int last_y_rez;
    int cmapID;
    int id;
} gr_image_type;

typedef struct gr_string_type {
    tx_char *chars;
    int n_chars;
} gr_string_type;

typedef struct gr_line_type {
    int sshape_type;
	int end1_cpnt;
	int end2_cpnt;
	int end1_cpnt_dir; /* direction of connector */
	int end2_cpnt_dir; /* direction */
    TmPoint32 sshape_pnt[3]; /* from 0 to 21600 (relative to object zero) */
} gr_line_type;

typedef struct gr_rect_type {
    int x_corner_radius;
    int y_corner_radius;
    int scale_corners;
    int x_corner_numerator;
    int y_corner_numerator;
    int sshape_type;
    TmPoint32 sshape_pnt[6]; /* from 0 to 21600 (relative to object zero) */
    struct thing *txt_t;
} gr_rect_type;

typedef struct gr_ellipse_type {
    int n_sides;
    int start_angle;
    int end_angle;
    int sshape_type;
    TmPoint32 sshape_pnt[6]; /* from 0 to 21600 (relative to object zero) */
    struct thing *txt_t;
} gr_ellipse_type;

typedef gr_ellipse_type gr_reg_poly_type;

typedef struct ge_build_info_type {
    unsigned int use_reverse_build  : 1,
                 use_inactive_color : 2;

    int build_effect;
    int build_rule; /* 0 - all at once , 1- I`st .. */ 
    int build_size; /* 0 - by paragraph, 1 - by word, 2 - by letter */
    int build_delay;
    int inactive_color;
    int sequence;
    int build_throttle;
    struct ge_audio_type baudio;
} ge_build_info_type;

typedef struct gr_template_t_type {
    unsigned int inactive     : 1,
                 grayed       : 1,
                 enable_build : 1;
    int type;
    int size;
    int ix;
    struct vector area;
    struct thing *child_t;
    struct ge_build_info_type *build;
    struct thing *crect_t;
} gr_template_t_type;

typedef struct gr_inset_type {
	struct vector area;
	int x_zoom;
	int y_zoom;
	unsigned int scale_mode   : 2,
	             proportional : 1,
	             centered     : 1,
	             unsaved      : 1,
	             is_open      : 1,
	             type         : 3;
	int ix;
} gr_inset_type;

typedef union thing_type_data {
    char                      *gen;
    struct gr_group_type      *group;
    struct gr_textbox_type    *txtbox;
    struct gr_image_type      *img;
    struct gr_string_type     *str;
    struct gr_line_type       *line;
    struct gr_rect_type       *rect;
    struct gr_ellipse_type    *ell;
    gr_reg_poly_type          *rpoly;
    struct gr_template_t_type *temp;
    struct gr_inset_type      *inset;
} thing_type_data;

typedef struct thing_mark_type {
    /* state flags */
    unsigned int fixed          : 1,
                 hidden         : 1,
                 locked         : 1,
                 shaped         : 1,
                 marked         : 1,
                 was_marked     : 1,
                 ghosted        : 1,
                 recolored      : 1,
                 templated      : 1,
                 mastered       : 1,
                 top_attr       : 1,
                 used           : 1,
                 no_shadow      : 1,
                 for_shadow     : 1,
                 emboss_shadow  : 1,

    /* draw cache flags */
                 in_draw_cache  : 1,
                 in_erase_cache : 1,

    /* chart flags */
                 isa_chart      : 1,
                 isa_title      : 1,
                 isa_subtitle   : 1,
                 isa_body       : 1,
                 isa_footer     : 1,
                 isa_legend     : 1,
	             isa_axis       : 1, 
	             isa_axlab      : 1,
	             isa_datgrp     : 1, 
	             isa_datlab     : 1,
				 isa_grid		: 1,
	             txt_measured   : 1,
	             child_text     : 1;
} thing_mark_type;

typedef struct thing {
    /* object attributes */
    thing_attributes attr;

    /* object state */
    thing_mark_type flags;

    /* object identity (LINE, IMAGE, TEXT...) */
    int type;

    /* object family (all associated objects have the same non-zero id) */
    int t_class;

    /* object type specific data */
    thing_type_data data;

    /* object path(s) */
    poly_type *paths;
    int n_paths;

	/* object anchor point(s) */
	int n_cpnts;
	int *cpnt;

    /* object extent */
    vector extent;

    /* object ink bounder */
    vector ink;

    /* object handle (unique) */
    unsigned int id;

    /* chart handle (unique) */
    unsigned int chart_ix;

	/* layer containing thing */
	unsigned int layer;

	/* clip info */
	clip_path *clip;
} thing;

typedef struct ge_rotate {
    float cos_num;
    float sin_num; 
    float denom;
} ge_rotate;

typedef struct ge_shear {
    float x;
    float y;
    int top;
    int left;
} ge_shear;

typedef struct ge_transform_type {
    unsigned int shearing     : 1,
                 rotate       : 1,
                 rotate_by_90 : 1,
                 scale_x      : 1,
                 scale_y      : 1,
                 zoom_to_fit  : 1;
    struct vector translate;
    struct vector scale;
    struct ge_rotate rot;
    struct ge_shear shear;
    int x_res;
    int y_res;
    int tmp_x_res;
    int tmp_y_res;
} ge_transform_type;

typedef struct ge_slide_show {
	unsigned int full_screen :1,
	             all_pages   :1;
	int page1;
	int pageN;
	int at_end;
} ge_slide_show;

typedef struct ge_display_buffer {
	int wid;
	int hyt;
	unsigned char *dat;
} ge_display_buffer;

typedef struct session {
    unsigned int terminated     : 1,
                 deleted        : 1,
                 translating    : 1,
                 rotating       : 1,
                 shearing       : 1,
                 editing        : 1,
                 editing_pix    : 1,
                 shaping        : 1,
                 hiding         : 1,
                 scrolling      : 1,
                 importing      : 1,
                 caching        : 1,
                 licensed       : 1,
                 full_license   : 1,

                 autoGrid       : 1,
                 stickyDots     : 1,
                 rulerOn        : 1,
                 gridOn         : 1,
                 readOut        : 1,
                 pageOn         : 1,
                 paletteOn      : 1,
                 colorpalOn     : 1,
                 bordersOn      : 1,
                 solidTextOn    : 1,
                 keylineOn      : 1,
                 guidesOn       : 1,
                 fullGrids      : 1,
                 autoStick      : 1,
                 equalAspect    : 1,
                 slowMotion     : 1,
                 metricUser     : 1,
                 colorXWD       : 1;

    unsigned int invertXWD      : 1,
                 colorTiff      : 1,
                 digitizeDeep   : 1,
                 grabAtEdge     : 1,
                 no_borders     : 1,
                 disable_undo   : 1,

                 tool_lock      : 1,
                 locked         : 1,
                 silent         : 1,
                 no_map_win     : 1,
                 msgup          : 1,
                 reset_marks    : 1,
                 reset_page     : 1,
                 just_marked    : 1,
                 just_unmarked  : 1,
                 att_changed    : 1,
                 remaster_pages : 1,
                 remaster_notes : 1,
                 force_color    : 1,
                 im_hex         : 1,
                 im_raw         : 1,
                 line_first     : 1,

                 no_handles     : 1,
                 no_drawing     : 1,
                 no_templates   : 1,
                 no_hourglass   : 1,
                 no_tools       : 1,
                 no_colors      : 1,
                 no_rulers      : 1,
                 no_hilite      : 1,
                 autosaveOn     : 1,
                 no_outliner    : 1;

    unsigned int displayed      : 1,
                 no_display     : 1,
                 need_display   : 1,
                 force_register : 1,
                 force_deselect : 1,
                 printing       : 1,
                 disable_poke   : 1,
                 busy           : 1,
                 use_java_attr  : 1,
                 backup_exists  : 1,
                 needs_backup   : 1,
                 java_display   : 1,
                 java_cached    : 1,
                 top_level_task : 1,
                 building       : 1;

    /* the most recently created object */
    struct thing *new_t;

    /* the last object the mouse stuck to */
    struct thing *sticky_t;

    /* distance from object that counts as "on" object */
    int pick_distance;

    /* maximum number of objects displayed prior to outlining */
    int display_max;

    /* number of times "smoother" does the smooth thing */
    int smooth_loops;

    /* mouse motion constraints */
    int constraint;

    /* customizable line width settings */
    int line_1;
    int line_2;
    int line_3;
    int line_4;

    /* current grid size */
    long dpi;
    long gridSize;
    long unit_precision;
    long unit_size;
    char *unit_name;

    /* number of images in file */
    unsigned long imCount;
    unsigned long tCount;
    unsigned long assocCount;

	/* current display buffer (an image of what is on the client screen) */
	ge_display_buffer img_src;

    /* offset to corner of grabbed object */
    coord corner_offset;

    /* currently marked and selected tools */
    int marked_tool;
    int curr_tool;

    /* window origin */
    coord win_orig;
    vector win_bounds;

    /* viewport to display window into (should match window by default) */
    vector vp;

    /* text cursor location within window */
    coord text_pos;
    int text_ix1;
    int text_ix2;
    int text_ix3;
    int last_text_ix1;
    int last_text_ix2;

    /* text typing attribute (at current insertion point) */
    thing_attributes *type_attr;

	/* current layer */
	int layer;

	/* default tab size */
	int def_tabsize;

    /* default slide info */
    int wrap_slide_show;
    int default_slide_effect;
    int default_slide_delay;
    int default_slide_throttle;
    int default_use_slide_build;
    struct ge_build_info_type default_build;
    struct ge_slide_show show_option;

	/* current presentation mode */
	int presentation_mode;
	int old_presentation_mode;

    /* size used by GE file */
    vector total_area;
    vector drawn_area;
    vector orig_total_area;
    vector orig_drawn_area;

    /* status message */
    char *status_msg;

    /* auto save info */
    int autoSaveInterval;
    long lastAutoSaveTime;

    /* macro to run in run_graphic */
    tx_char *run_macro;
    int n_mchars;
} session;

typedef struct ge_print_info_type {
    unsigned int use_content_border : 1;
    unsigned int use_crop_marks     : 1;
    unsigned int use_regis_marks    : 1;
    unsigned int separate           : 1;
    unsigned int landscape          : 1;

    /* printable area size and location */
    int page_x, page_wid;
    int page_y, page_hyt;

    /* printer id */
    int id;

    /* in mills */
    int wid;
    int hyt;
    int bleed;
    vector margin;

	/* location of guides */
	TmPoint32 guide_pt; /* offset from page center */

    /* in percent */
    int h_scale;
    int v_scale;

    /* headers and footers */
    struct ge_hdrftr_main_type hdrftr;
} ge_print_info_type;

/* must match palette.c struct */
typedef struct GEPalette_t {
    long timeChanged;
    int toolType;   
    int fillPattern;
    int linefillPattern;
    int lineThickness;
    int lineStyle;  
    int arrow1;  
    int arrow2;
    int join_symbols;
    char **parts;   
 
    int ffg;
    int fbg;
    int lfg;
    int lbg;
    int shadow_color;

    int ribbon;
 
    int page_ix;
    int n_pages;
    int n_colors;
    int *cinfo;

    unsigned int
     toolLocked    : 1,    
     editingPix    : 1,  
     shadowBoxUp   : 1,
     repaintColors : 1,
     repaintFlag   : 1;
} GEPalette_t;

typedef struct ge_win_info_type {
    struct TmCrePacket params;

    /* max window size */
    coord max_win;

    /* graphic unit sizes */
    long grMWid;
    long grMHyt;
    long grPWid;
    long grPHyt;

    /* text cell mil sizes */
    long txWid;
    long txHyt;

    /* smallest fill expressible in mils */
    int patWid;
    int patHyt;

    /* some often used sizes based on resolution */
    int xPel, yPel;
    int xDot, yDot;
} ge_win_info_type;

typedef struct ge_draw_cache_type {
    int n_hidden;
    int n_drawn;
    vector draw_area;
    vector hide_area;
} ge_draw_cache_type;

typedef struct udfill_type {
    int is_pict;
    int wid;
    int hyt;
    int n_chars;
    int cmapID;
    tx_char *name;
    unsigned char *data;
} udfill_type;

typedef struct udgrad_type {
    int type;
    int xoff, yoff;
    int angle;
    int n_chars;
    tx_char *name;
} udgrad_type;

typedef struct layer_type {
    unsigned int
     hidden : 1,
     locked : 1,
     prints : 1,
     backgd : 1;
    int n_chars;
    tx_char *name;
} layer_type;

typedef struct udo_type {
    struct thing *t;
    int n_chars;
    tx_char *name;
} udo_type;

typedef struct ge_callback_type {
    unsigned int
     on_drag     : 1,
     on_double   : 1,
     on_single   : 1;

    char *name;

    int n_args;
    char **args;
} ge_callback_type;

typedef struct ge_hook_type {
    char *type;
    char *name;
} ge_hook_type;

typedef struct ge_tag_type {
    int n_strs;
    char **str;
} ge_tag_type;

typedef struct ax_message_type {
    long type;         /* message type */
    long size;         /* message size */
    char *data;        /* message data */
    long sender_type;  /* type of task that sent mesage */
    long sender_task;  /* task id if sent by a task rather than system */
} ax_message_type;

typedef struct ge_event_filter_type {
    int num_timeouts;
    int max_timeouts;
    int timeout_time;
	int next_timeout_time;
} ge_event_filter_type;

typedef struct ge_stroke_info_type {
	unsigned int double_click    :1,
                 triple_click    :1,
                 quadruple_click :1,
                 constrained     :1,
                 panningOn       :1;
    TmPoint32 constraining_pt;
    TmPoint32 down;
    TmPoint32 last;
    TmPoint32 at;

    /* panning info */
    int panInterval;
    long lastPanTime;
} ge_stroke_info_type;

typedef struct ge_readout_type {
    int displayed;
    int x1, y1;
    int x2, y2;
    char label1[32];
    char label2[32];
} ge_readout_type;

typedef struct ge_stroke_echo_type {
    unsigned int pointer_rep;
    unsigned int report;
    unsigned int at_button;
    unsigned int displayed;
    int pointer_act;
    int action;
    int last_action;
    TmPoint32 pt[8];
} ge_stroke_echo_type;

typedef struct ge_event_type {
    unsigned int type : 4;
    unsigned int last : 4; 
    unsigned int echoed       : 1;
    unsigned int was_filtered : 1;
    unsigned int mouse_moved  : 1;
    unsigned int is_a_paste   : 1;
    unsigned int shift_pick   : 1;
    unsigned int boundary_hit : 4;
    unsigned int last_bound_hit : 4;
    ge_event_filter_type filter;
    ge_stroke_info_type mouse;
    ge_readout_type readout;
    ge_stroke_echo_type stroke_echo;
    ax_message_type msg;
} ge_event_type;

typedef struct ge_file_info_type {
    /* document state */
    unsigned int
     on_disk        : 1,
     for_wp         : 1, /* wp child */
     for_ss         : 1, /* ss - chart child */
     for_mail       : 1, /* mail - child */
     for_other      : 1, /* Other child */
     read_only      : 1,
     read_only_set  : 1,
     new_doc        : 1,
     doc_changed    : 1,
     was_saved      : 1,
     icon_title     : 1,
     save_to_disk   : 1,
     no_update      : 1,
     foreign_type   : 1,
     registered     : 1;

    /* version */
    int version;

    /* save flags */
    int save_mode;
    int quiet_mode;
    int paste_mode;
    vector paste_win;

    /* file permissions */
    int grp_access;
    int all_access;

    /* identity */
    char name[STRPATH];
    char pathname[STRPATH];
	int doc_type;

    /* file buffer */
    int buf_len;
    char *buf;
} ge_file_info_type;

typedef struct ge_edit_pt {
    TmPoint32 p;
    TmPoint32 c1;
    TmPoint32 c2;
    unsigned int selected : 1,
                 c1_on    : 1,
                 c2_on    : 1,
                 smooth   : 1,
                 first    : 1,
                 final    : 1;
} ge_edit_pt;

typedef struct ge_edit_pts {
    int closed;
    int n_pts;
    ge_edit_pt *pt;
} ge_edit_pts;

typedef struct ge_edit_path {
    int n_paths;
    ge_edit_pts *path;
} ge_edit_path;

typedef struct ge_path_edit_save_info {
    int n_paths;
    struct poly_type *path;
    thing t;
} ge_path_edit_save_info;

typedef struct ge_edit_info_type { 
    unsigned int new_obj         : 1,
                 changed         : 1,
                 marked          : 1,
                 was_marked      : 1,
                 plus            : 1,
                 swap            : 1,
                 was_closed      : 1,
                 is_endpoint     : 1,
                 first_pt        : 1,
                 final_pt        : 1,
                 read_out_was_on : 1,
                 area_selecting  : 1;
    struct ge_path_edit_save_info save;
    int point_type;
    int point_ix;
    int path_ix;
    int edit_mode;
    int shear_type;
    int top, left;
    int hyt, wid;
    int lastr, theta;
    int x_size, y_size;
    int delta_x, delta_y;
    int plen, clen;
    TmPoint32 *pnts, *ctls;
    TmPoint32 pts_before[4], pts_at[4], pts_after[4];
    axtext txt;
    atext  otxt;
    struct thing *t, *box;
    int (*func)();
} ge_edit_info_type;

typedef struct ge_inset_info {
    /* active field flags */
    unsigned int
     use_clear_mode    : 1,
     use_scale_mode    : 1,
     use_display_mode  : 1,
     use_proportional  : 1,
     use_prescale_x    : 1,
     use_prescale_y    : 1,
     use_source        : 1,
     use_src_x         : 1,
     use_src_y         : 1,
     use_src_wid       : 1,
     use_src_hyt       : 1,
     use_dst_x         : 1,
     use_dst_y         : 1,
     use_dst_wid       : 1,
     use_dst_hyt       : 1,
     use_clip_x        : 1,
     use_clip_y        : 1,
     use_clip_wid      : 1,
     use_clip_hyt      : 1,
     use_change_cursor : 1,
     use_page_ix       : 1,

    /* format flags */
     clear_mode        : 4,
     scale_mode        : 2,
     proportional      : 1,
     change_cursor     : 1;

	/* more active field flags */
	unsigned int
	 use_audio_play    : 1,
	 use_audio_type    : 1,
	 use_audio_path    : 1,
	 use_volume        : 1;

    /* prescale factors */
    int prescale_x;
    int prescale_y;

    /* destination size and offset */
    int dst_x;
    int dst_y;
    int dst_wid;
    int dst_hyt;

    /* clipping info */
    int clip_x;
    int clip_y;
    int clip_wid;
    int clip_hyt;

    /* source size, offset, and optional alternate image */
    int src_x;
    int src_y;
    int src_wid;
    int src_hyt;
    long source;

    /* widget handle */
    long widID;

    /* special effect for display */
    int display_mode;

    /* which page to inset */
    int page_ix;

	/* audio info */
	int audio_play;
	int audio_type;
	char *audio_path;
	int volume;
} ge_inset_info;

typedef struct ge_menubar_type {
    int time;
    int curr_plot_type;
    int no_bold_or_ital;
    int bold, ital, ulin, size, face;
    int h_just;
	int (*func)();
} ge_menubar_type;

typedef struct ge_marked_type {
    int kind;
    int total;
    int last_total;
    int b_texts;
    int s_texts;
    int ells;
    int rects;
    int strokes;
    int polys;
    int images;
    int lines;
    int udos;
    int rpolys;
    int c_images;
    int rrects;
    int curves;
    int insets;
    int links;
    int embeds;
    int gfxs;
    int wpxs;
    int ssxs;
    int templates;
    int regions;
    int charts;
    int locked;
    int combines;
    int combinables;
    int assocs;
    struct thing *first;
} ge_marked_type;

typedef struct ge_path_shadow_draw_info {
    unsigned int active;
    int x, y;
} ge_path_shadow_draw_info;

typedef struct ge_vars_type {
    char *var, *val;
} ge_vars_type;

typedef struct ge_rt_inset_type {
	char *exec; /* macro to execute when opened */
	char *filt; /* macro to execute for filtering */
	char *data;
	int type;
	unsigned int
	 icon_only  : 1,
	 bad_handle : 1, /* runtime - toss handle on next reformat */
	 bad_format : 1; /* runtime - reformat before next display */
} ge_rt_inset_type;

typedef struct ge_link_type {
    char *path;
	ge_rt_inset_type rt_inset;
} ge_link_type;

typedef struct ge_embed_type {
	char *edit_path;
	int unsaved;
    int size;
    char *data;
	ge_rt_inset_type rt_inset;
} ge_embed_type;

typedef struct ge_clipart_type {
	char *edit_path;
	int unsaved;
    int size;
    char *data;
	ge_rt_inset_type rt_inset;
} ge_clipart_type;

typedef struct outline_text_info {
	int type;
	int tx_type;
	int size;
	int n_chars;
	char *olin_txt;
	thing *t;
}outline_text_info;

typedef struct ge_undo_element_type {
    int nobjs;
    int pos, old_pos;
	int page_mode;
	int scale_mode;
	int templated;
    vector extent;
    vector old_vec;
    vector new_vec;
    int l_ix;
    int obj1;
    int obj2;
    int obj3;
    int obj4;
    int n_chars;
    int n_charts;
    int n_cmaps;
    int n_links;
    int n_embeds;
    int n_pages;
	int no_delete;
    thing *t;
    tx_char *name;
    char *chart;
    cmap_type *cmap;
    ge_link_type *link;
    ge_embed_type *embed;
    layer_type *layer;
    udfill_type *fill;
    thing_mark_type flags;
    ge_print_info_type *print;
    struct ge_page_type *page;
    struct ge_template_type *template;
} ge_undo_element_type;

typedef struct ge_undo_type {
    int n;
    int code;
    struct ge_undo_element_type *info;
} ge_undo_type;

typedef struct ge_main_path_type {
    int max_pnts;
    int max_ctls;
    poly_type sub_path;
} ge_main_path_type;

typedef struct ge_color_scheme_type {
    /* object info */
    int title_fg, title_bg;
    int text_fg, text_bg;
    int line_fg, line_bg;
    int fill_fg, fill_bg;
    int accent_color[15];
    int shadow_color;

	int n_other_colors;
	int *other_color;

    /* page info */
    int page_pat;
    int page_type;
    int page_fg;
    int page_bg;
} ge_color_scheme_type;

typedef struct ge_page_type {
    unsigned int
     hide_objs          : 1, /* don't display objects in template layer     */
     hide_pg_backgd     : 1, /* don't display page fill pattern             */
     use_backgd         : 1, /* use the page background, not the template's */
     use_color          : 1, /* use the page color, not the template's      */
     compress_outline   : 1,
     use_slide_build    : 1,
	 hide_slide         : 1, /* hide slide slide show */
	 digitize_image     : 1; /* for filters */

    /* paint list for this page */
	int n_cpoints;
	TmPoint32 *cpnt;
    int n_layers;
    int layer_ix;
    struct layer_type *layer;
    struct thing t;

    /* temp storage */
    struct thing *old_t;

    /* speaker notes for this page */
    struct thing *notes_t;

    /* outline for this page */
    tx_char *outline_text;
    int n_ochars;

    /* slide info */
    int slide_effect;
    int slide_delay;
    int slide_throttle;

    /* page layout index */
    int layout_ix;
    int object1;
    int object2;
    int object3;
    int object4;

    /* color scheme info */
    struct ge_color_scheme_type c_scheme;

	/* audio info */
	struct ge_audio_type audio;

	/* build info */
    int n_builds; /* num of objs having active builds */
    struct thing **build_list; /* list of objs having active builds */
    struct ge_build_info_type build;

	/* optional hdrftrs per page - will overide default hdrftrs */
    struct ge_hdrftr_page_type *hf_page;

	/* currently used only in filter */
	int n_outlines;
	int n_id;
	struct outline_text_info *otext;
} ge_page_type;

/* for slide sorted - never saved */
typedef struct ge_pagelist_type {
	thing *t;
	int new_ix;
	int n_rows;
	int n_cols;
	int m_wid;
	int m_hyt;
	int p_wid;
	int p_hyt;
} ge_pagelist_type;

typedef struct ge_oseg_type {
	unsigned int strip_newline : 1,
				compressed     : 1,
				selected       : 1,
				empty          : 1,
				deleted        : 1;

	int type;
	int type_ix;
	int ix1;
	int ix2;
	int n_chars;
	int n_lines;
	thing *src_t;
	thing *t;
} ge_oseg_type;

typedef struct ge_oseg_list {
	unsigned int compress_all   : 1,
	             txt_active     : 1,
	             reset          : 1,
	             reset_mrkrs    : 1,
	             reset_mrkpos   : 1,
	             show_colors    : 1,
	             hide_formating : 1,
				 hide_txt       : 1;

	struct ge_oseg_type *seg;
	int n;
	int cursor_ix;
	int cur_off;
	thing *t;
} ge_oseg_list;

/* write as part of SESSION in file */
typedef struct ge_template_type {
	/* slide master */
	struct ge_page_type s_master_page;
	struct vector slide_heading;
	struct vector slide_content;
	struct thing_attributes t0_attr;
	struct thing_attributes t1_attr;
	struct thing_attributes b0_attr;
	struct thing_attributes b1_attr;
	struct thing_attributes b2_attr;
	struct thing_attributes b3_attr;
	struct thing_attributes b4_attr;
	struct thing_attributes b5_attr;

	/* title master */
	struct ge_page_type *t_master_page;
	struct vector title_heading;
	struct vector title_content;
	struct thing_attributes ct0_attr;
	struct thing_attributes ct1_attr;

	/* handout master */
	struct ge_page_type h_master_page;

	/* outline master */
	struct ge_page_type o_master_page;
	struct ge_oseg_list o_list;

	/* notes master */
	struct ge_page_type n_master_page;
	struct vector notes_preview;
	struct vector notes_content;
	struct thing_attributes n0_attr;
	struct thing_attributes n1_attr;
	struct thing_attributes n2_attr;
	struct thing_attributes n3_attr;
	struct thing_attributes n4_attr;
	struct thing_attributes n5_attr;

	/* current page mode (slides, notes, handouts, outline, sorter) */
	int page_mode;

	/* old page info (run-time only) */
	int old_n_pages;
	int old_page_ix;
	int old_next_page_ix;
	struct ge_page_type *old_page;
	struct ge_pagelist_type *old_pagelist;

	int n_schemes;
    struct ge_color_scheme_type *c_schemeList;

	/* attr transition tracking */
	unsigned int
	 rem_suppress   : 1,
	 s1_suppress    : 1,
	 s1_weight      : 1,
	 s1_slant       : 1,
	 s1_strike      : 1,
	 s1_ruling      : 1,
	 s2_suppress    : 1,
	 s2_weight      : 1,
	 s2_slant       : 1,
	 s2_strike      : 1,
	 s2_ruling      : 1,
	 n2_suppress    : 1,
	 n2_weight      : 1,
	 n2_slant       : 1,
	 n2_strike      : 1,
	 n2_ruling      : 1;
} ge_template_type;

typedef struct ge_sequence_list {
	int s_ix;
	int e_ix;
	struct thing *t;
} ge_sequence_list;

typedef struct ge_auto_shape_info {
	int n_paths;
	poly_type *paths;
	TmPoint32 txt_pt[2];
}ge_auto_shape_info;

#include <fb.h>

/****************************************************************************\
*                  THE graphics editor typedef                               *
\****************************************************************************/
typedef struct graphics_main_type {
	/* stuff common to all apps */
	genericAppStruct appInfo; /* generic app info - */
	int pid;            /* task id                               */

    /* identity */
    struct ge_file_info_type file; /* name, permissions, save mode, ...     */
    long id;                       /* widget id of drawing surface          */
	int w_xoff;                    /* widget x offset                       */
	int w_yoff;                    /* widget y offset                       */

	/* image hot spot  */
	struct TmPoint32 hotspot;
	
    /* device and ATM stuff */
    struct ge_win_info_type win;

    /* event info (and how to handle them) */
    struct ge_event_type ev;

    /* printer and page format info */
	struct ge_print_info_type *handout_print;
	struct ge_print_info_type *outline_print;
	struct ge_print_info_type *notes_print;
    struct ge_print_info_type print;
    int x_zero_pt;
    int y_zero_pt;

    /* description of current graphic state - for graphics editor */
    struct thing_attributes *tmp_attr;
    struct thing_attributes def_attr;
    struct thing_attributes drp_attr;
    struct thing_attributes sys_attr;
    struct ge_marked_type marked;
    struct ge_undo_type undo;
    struct session state;
    vector scroll_area;
    int mode;

    /* current state - for palette */
    struct GEPalette_t pal;

    /* current state - for menubar */
    struct ge_menubar_type menub;

    /* the paint list(s) */
    int n_pages;
    int page_ix;
    int next_page_ix;
    struct ge_page_type *page;
	struct ge_pagelist_type *pagelist;
    struct thing *t;

    /* current drawing info */
    int n_paths;
    int max_n_paths;
    struct ge_main_path_type *path;
    struct attr_type path_attr;
    struct attr_type java_attr;
    struct thing_attributes cur_attr;
    struct ge_transform_type xform;
    struct ge_draw_cache_type cache;
    struct ge_path_shadow_draw_info shadow;
    struct vector clip;
    struct vector *inset_clip;
    struct clip_path *clip_path;

    /* state of the current edit */
    struct ge_edit_info_type ed;
    struct thing *prim_t;
    struct thing *text_t;
    int px_is_color;
    int px_height;
    int px_width;
#if 1
	px_main_type *px;
#else
    char *px;
#endif

    /* user defined property descriptions (not associated with a page) */
    int n_faces;
    char **face;
    int n_fills;
    struct udfill_type *fill;
    int n_grads;
    struct udgrad_type *grad;
    int n_cmaps;
    struct cmap_type *cmap;
    int n_styles;
    struct thing_attributes *style;
    int n_callbacks;
    struct ge_callback_type *callback;
    int n_tags;
    struct ge_tag_type *tag;
    int n_cmts;
    struct ge_vars_type *cmt;
    int n_links;
    struct ge_link_type *link;
    int n_embeds;
    struct ge_embed_type *embed;
    int n_carts;
    struct ge_clipart_type *cart;
    int n_hooks;
    struct ge_hook_type *hook;
    int n_udos;
    struct udo_type *udo;
    int n_charts;
    char *chart;
	char *pInfo;
    struct ge_template_type *template;

    /* file reading in progress */
    int locked;

    /* just in case */
    int invalid;
} graphics_main_type;

/****************************************************************************\
*                  graphics editor constants                                 *
\****************************************************************************/
/* Events */
#define NON_EVENT   15
#define MOUSE_UP    0
#define MOUSE_DN    1
#define MOUSE_MV    2
#define KEYBOARD    3
#define FINISH_UP   4
#define LOG_KEY     5
#define CHOICE      6
#define PICK        7
#define MENUBAR     8

/* char field align codes */
#define LEFT_ALIGN      0
#define CENTER_ALIGN    1
#define RIGHT_ALIGN     2
#define JUSTIFIED_ALIGN 3

/* char field vertical align codes */
#define TOP_ALIGN       0
#define BASELINE_ALIGN  1
#define BOTTOM_ALIGN    2

/* Align menu codes */
#define LEFT_JUST   0
#define CTR_JUST    1
#define RIGHT_JUST  2
#define TOP_JUST    3
#define MID_JUST    4
#define BOT_JUST    5
#define EQUAL_X     6
#define EQUAL_Y     7
#define X_EQUAL_Y   8
#define Y_EQUAL_X   9
#define X_MIRROR    10
#define Y_MIRROR    11
#define XY_SMALLER  12
#define XY_LARGER   13
#define X_EQ_CDELTA 14
#define Y_EQ_CDELTA 15
#define X_EQ_SDELTA 16
#define Y_EQ_SDELTA 17
#define X_ADJACENT  18
#define Y_ADJACENT  19
#define XY_MOVE     20
#define XY_SCALE    21
#define XY_ROTATE   22

#ifdef MAXSHORT
#undef MAXSHORT
#endif

/* some maximum values */
#undef  MAXINT
#define MAXINT     2147483647
#define MAXUINT    4294967295
#define MAXSHORT   32767
#define MAXUSHORT  65535
#define MAXUWORD   MAXUINT

#define GE_PEND_TIME MAXSHORT /* ~9 hours */
#define GE_MAXTEXT (MAXSHORT - 100)

/* a position value to indicate second value equality to first */
#define SAME       MAXINT

/* define boundaries of ge page */
#define WIN_MIN      0
#define MAX_WID 128000
#define WIN_MAX (WIN_MIN + MAX_WID)

/* direction markers for tranformation control */
#define SOUTH 0x0001
#define EAST  0x0002
#define NORTH 0x0004
#define WEST  0x0008
#define XLATE (NORTH | SOUTH | EAST | WEST)

/* shape markers for edit control */
#define SHAPEHGD 0x0200
#define SHAPEVGD 0x0400
#define SHAPEELL 0x0800
#define SHAPEVTX 0x1000
#define SHAPESEG 0x2000
#define SHAPECTL 0x4000
#define SHAPETXT 0x8000

#ifdef TEXT
#undef TEXT
#endif

/* object types */
#define TEXTBOX    1
#define IMAGE      2
#define TEXT       3
#define POLY       4
#define STROK      5
#define GROUP      6
#define LINE       7
#define RECT       8
#define ELL        9
#define RPOL      10
#define INSET     11
#define TEMPLATE  12
#define SUBTXTBOX 13
#define REGION    14

/* object sub-types */
#define INSET_NONE     0
#define INSET_LINKED   1
#define INSET_EMBEDED  2
#define INSET_SELF     3
#define INSET_SELF2    4
#define INSET_SELF3    5
#define INSET_OLIN_CTL 6
#define INSET_CLIPART  7

/* built in ge tools */
#define P_POINT      0
#define P_TEXT       1
#define P_LINE       2
#define P_POLY       3
#define P_RECT       4
#define P_ELL        5
#define P_STROK      6
#define P_DIGITIZE   7
#define P_SQUARE     8
#define P_CIRCLE     9
#define P_CURVE     10
#define P_C_DIGITZE 11
#define P_ZOOM      12
#define P_TXTBOX    13
#define P_UDO       14

/* constant for extraction for query@ */
#define P_PLACEHOLDR 15

/* Polyline states */
#define G_POINT 0
#define G_LAST  1
#define G_FINI  2

/* ink types */
#define PROCESS 0
#define SPOT    1

/* shadow types */
#define NO_SHADOW     0
#define BACK_DSHADOW  1
#define LOCAL_DSHADOW 2
#define BACK_EMBOSS   3
#define LOCAL_EMBOSS  4

/* draw cache flags */
#define DRAW 0
#define HIDE 1

/* Truly miscellaneous */

#define INITIAL (-1)
#define APPEND  0

#define MAX_EDIT    5
#define LIST_SIZE   10

#undef  FALSE
#undef  TRUE
#undef  UNKNOWN
#define FALSE   0
#define TRUE    1
#define UNKNOWN 2

#ifndef NULL
#define NULL    0
#endif

#undef NULL_OBJ
#define NULL_OBJ    0

#define FGEPREFIX   "GE"
#define FBGPREFIX   "BG"
#define FSUFFIX ".qm"

#define MAX_PATN    12
#define MAXANG  3600
#define EPSILON 90  

#define X_LINE_MOVE 750
#define Y_LINE_MOVE 750

#define RULER_WIDTH 375
#define MILS_P_INCH 1000

/* pi/1800 */
#define C_FACT  0.0017453292519943295

#define X_PEL 14
#define Y_PEL 14

#define XINC    13
#define YINC    15
#define SMALL_ARROW 180

/* chart macro */
#define IS_FLAG_SET(t, val)                                \
{                                                          \
	if ( t &&                                              \
	     !t->flags.isa_axis   && !t->flags.isa_axlab    && \
	     !t->flags.isa_datgrp && !t->flags.isa_legend   && \
	     !t->flags.isa_title  && !t->flags.isa_subtitle && \
	     !t->flags.isa_footer && !t->flags.isa_grid)       \
		val = TRUE;                                        \
	else                                                   \
		val = FALSE;                                       \
}

/* maximum value contained in a sshape_pnt */
#define ANCHOR_MAX    21600.0

/****************************************************************************/
#endif /* GE_H_DEFINED */
