/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: ag31.c
   Applix ASC version: 29
   Date: Mon Nov 16 13:58:52 1998
*/
#include <aos.h>
#include <atm.h>
#include <atmwin.h>
#include <axtext.h>
#include <atxint.h>
#include <img.h>
#include <ge.h>
#include <n_style.h>
#include <bg.h>
#include <afileimg.h>
#include <cvttoks.h>
#include <cvt.h>

#if FOR_FILTER
#	include <stiorts.h>
#else
#	include <fsrts.h>
#	define FOR_FILTER 0
#endif

#define NOHAMMER 	0
#define HAMMER 		1

#define RUN 0
#define DMP 1
#define MINRUN  3
#define MAXRUN  128
#define MAXDMP  128


int mailOutBit6(), mailOutHex();
int ag31writeRAW(), ag31writeRUN(), ReadWriteLZW();



/****************************************************************************
 *
 * module:	ag31OutSpecialChars - handle the specials
 */
ag31OutSpecialChars(inBuffer,outBuffer,len,mode)
tx_char *inBuffer, *outBuffer;
int len, mode; 
{
	int inlen, count=0;
	tx_char *inP, *outP;

	if ((mode == HAMMER) && (outBuffer == NULL))
		return;

	inP = inBuffer;
	outP = outBuffer;
	inlen = len;
	while(len--)
	{
		if ((*inP == '\\') || (*inP == '"'))
		{
			count++;
			if (mode == HAMMER)
			{
				*outP++ = '\\';
				*outP++ = *inP;
			}
		}
		else if (*inP == '\n')
		{
			count++;
			if (mode == HAMMER)
			{
				*outP++ = '\\';
				*outP++ = 'n';
			}
		}
		else if (mode == HAMMER)
			*outP++ = *inP;

		inP++;
	}
	return(inlen+count);
}

/****************************************************************************
 *
 * module:	a g 3 1 O u t S t r i n g
 */
ag31OutString(ge, len, str, cvt, suppress_quote)
graphics_main_type *ge;
int len;
tx_char *str;
cvt_global_type *cvt;
int suppress_quote;
{
	int i,llen,size;
	mbchar *strP, *outBuf;
	tx_char *inbufP;

	/* count the characters first */
	llen = ag31OutSpecialChars(str,NULL,len,NOHAMMER);

	/* str must be null terminated for NormalizeOut16, so  */
	/* we MUST make a local copy since we cannot count on  */
	/* the necessary tx_char being available in str.       */
	size = ((llen+1)*sizeof(tx_char));
	inbufP = (tx_char*)malloc(size);

	/* now hammer the corrected chars in */
	ag31OutSpecialChars(str,inbufP,len,HAMMER);
	inbufP[llen] = 0;


	/* Pump it thru the encoder */
    /* worst case: a wide char turns into  4 bytes */
    size = (llen+1)*4;  
 
	outBuf = (mbchar*)malloc(size);  /* worst case */
	len = NormalizeOut16(inbufP, outBuf, size, &cvt->info);

	/* open quote */
	if ( suppress_quote )
		llen = 0;
	else
	{
		llen = 1;
		cputc(ge,cvt, '\"');
	}
	strP = outBuf;

	while(len--)
	{
		cputc(ge,cvt,*strP++);

		llen++;
		if ( (MAX_LINE-1) < llen )
		{
			if (strP[-1] == '\\')
				cputc(ge,cvt,*strP++);
			cputc(ge,cvt,'\n');
			cputc(ge,cvt,' ');
			llen=0;
		}
	}

	/* close quote */
	cputc(ge,cvt, '\"');
	free(inbufP);
	free(outBuf);
}



/****************************************************************************
 *
 * module:	a g 3 1 O u t A t t r s - Write an object's attrs to file
 */
ag31OutAttrs(ge, attr, cmap_ix, cvt)
graphics_main_type *ge;
thing_attributes *attr;
int cmap_ix;
cvt_global_type *cvt;
{
	int wroteSomething=FALSE;
	int attr_undval, attr_thruval, cur_undval, cur_thruval;
	int ln_size;

	if ( attr->fill_pat != cvt->curr_atts.fill_pat ||
	     attr->fpat_type != cvt->curr_atts.fpat_type )
	{
		/* write attr->fpat_off twice until x,y offs happens */
		cvt->linepos += cprintf(ge, cvt," %s %d %d %d %d %d",
			tok[TFILL_PAT],attr->fill_pat ,attr->fpat_type,
			0, 0, 0);

		/* update the current default state */
		cvt->curr_atts.fill_pat = attr->fill_pat;
		cvt->curr_atts.fpat_type = attr->fpat_type;
	}

	if ( MAX_LINE < cvt->linepos )
		{ cprintf(ge, cvt, "\n"); cvt->linepos=0; wroteSomething=TRUE; }

	if ( attr->fill_bg != cvt->curr_atts.fill_bg )
		cvt->linepos += cprintf(ge, cvt," %s %d", tok[TBACK_COLOR],
			cvt->curr_atts.fill_bg = attr->fill_bg);

	if ( MAX_LINE < cvt->linepos )
		{ cprintf(ge, cvt, "\n"); cvt->linepos=0; wroteSomething=TRUE; }

	if (attr->fill_fg != cvt->curr_atts.fill_fg)
		cvt->linepos += cprintf(ge, cvt," %s %d", tok[TFORE_COLOR],
			cvt->curr_atts.fill_fg = attr->fill_fg);

	if ( MAX_LINE < cvt->linepos )
		{ cprintf(ge, cvt, "\n"); cvt->linepos=0; wroteSomething=TRUE; }

	if ( attr->line_pat != cvt->curr_atts.line_pat ||
	     attr->lpat_type != cvt->curr_atts.lpat_type )
	{
		/* write attr->lpat_off twice until x,y offs happens */
		cvt->linepos += cprintf(ge, cvt," %s %d %d %d %d %d",
			tok[TLINE_PAT],attr->line_pat ,attr->lpat_type,
			0, 0, 0);

		/* update the current default state */
		cvt->curr_atts.line_pat  = attr->line_pat;
		cvt->curr_atts.lpat_type = attr->lpat_type;
	}
	if ( MAX_LINE < cvt->linepos )
		{ cprintf(ge, cvt, "\n"); cvt->linepos=0; wroteSomething=TRUE; }

	if ( attr->line_bg != cvt->curr_atts.line_bg )
		cvt->linepos += cprintf(ge, cvt," %s %d", tok[TLINE_BACK_COLOR],
			cvt->curr_atts.line_bg = attr->line_bg);

	if ( MAX_LINE < cvt->linepos )
		{ cprintf(ge, cvt, "\n"); cvt->linepos=0; wroteSomething=TRUE; }

	if ( attr->line_fg != cvt->curr_atts.line_fg )
		cvt->linepos += cprintf(ge, cvt," %s %d", tok[TLINE_FORE_COLOR],
			cvt->curr_atts.line_fg = attr->line_fg);

	if ( MAX_LINE < cvt->linepos )
		{ cprintf(ge, cvt, "\n"); cvt->linepos=0; wroteSomething=TRUE; }

	if ( cvt->curr_atts.line_thickness != attr->line_thickness )
	{
		cvt->curr_atts.line_thickness = attr->line_thickness;
		if ( cvt->curr_atts.line_thickness < 0 )
			cvt->curr_atts.line_thickness = 0;
		cvt->linepos += cprintf(ge, cvt," %s %d", tok[TTHICKNESS],
			cvt->curr_atts.line_thickness);
	}
	if ( MAX_LINE < cvt->linepos )
		{ cprintf(ge, cvt, "\n"); cvt->linepos=0; wroteSomething=TRUE; }

	if ( cvt->curr_atts.line_style != attr->line_style )
		cvt->linepos += cprintf(ge, cvt, " %s %d", tok[TSTYLE],
			cvt->curr_atts.line_style = attr->line_style);
	
	if ( MAX_LINE < cvt->linepos )
		{ cprintf(ge, cvt, "\n"); cvt->linepos=0; wroteSomething=TRUE; }

	if ( cvt->curr_atts.left_arrow != attr->left_arrow )
		cvt->linepos += cprintf(ge, cvt, " %s %d", tok[TL_ARROW],
			cvt->curr_atts.left_arrow = attr->left_arrow);

	if ( MAX_LINE < cvt->linepos )
		{ cprintf(ge, cvt, "\n"); cvt->linepos=0; wroteSomething=TRUE; }

	if ( cvt->curr_atts.right_arrow != attr->right_arrow )
		cvt->linepos += cprintf(ge, cvt, " %s %d", tok[TR_ARROW],
			cvt->curr_atts.right_arrow = attr->right_arrow);

	if ( MAX_LINE < cvt->linepos )
		{ cprintf(ge, cvt, "\n"); cvt->linepos=0; wroteSomething=TRUE; }

	if ( cvt->curr_atts.join_symbols != attr->join_symbols )
		cvt->linepos += cprintf(ge, cvt, " %s %d", tok[TMARKER],
			cvt->curr_atts.join_symbols = attr->join_symbols);

	if ( MAX_LINE < cvt->linepos )
		{ cprintf(ge, cvt, "\n"); cvt->linepos=0; wroteSomething=TRUE; }

	/* be sure the shadow color is valid */
	if (attr->shadow_color >= ge->cmap[0].n_colors)
		attr->shadow_color = 0;

	if (attr->shadow != cvt->curr_atts.shadow)
		cvt->linepos += cprintf(ge, cvt," %s %d", tok[TSHADOW],
			cvt->curr_atts.shadow = attr->shadow);

	if ( MAX_LINE < cvt->linepos )
		{ cprintf(ge, cvt, "\n"); cvt->linepos=0; wroteSomething=TRUE; }

	if (attr->shadow_color != cvt->curr_atts.shadow_color)
		cvt->linepos += cprintf(ge, cvt," %s %d", tok[TSHADOWCOLOR],
			cvt->curr_atts.shadow_color = attr->shadow_color);

	if ( MAX_LINE < cvt->linepos )
		{ cprintf(ge, cvt, "\n"); cvt->linepos=0; wroteSomething=TRUE; }

	if ((attr->shadow_x != cvt->curr_atts.shadow_x) ||
	    (attr->shadow_y != cvt->curr_atts.shadow_y))
	{
			cvt->linepos += cprintf(ge, cvt," %s %d %d",
				tok[TSHADOWSIZE],
				cvt->curr_atts.shadow_x = attr->shadow_x,
				cvt->curr_atts.shadow_y = attr->shadow_y);
	}
	if ( MAX_LINE < cvt->linepos )
		{ cprintf(ge, cvt, "\n"); cvt->linepos=0; wroteSomething=TRUE; }

	if ( cmap_ix != cvt->cmapID )
		cvt->linepos += cprintf(ge, cvt," %s %d\n", tok[TCOLORMAP],
			cvt->cmapID = cmap_ix);

	if ( MAX_LINE < cvt->linepos )
		{ cprintf(ge, cvt, "\n"); cvt->linepos=0; wroteSomething=TRUE; }

	if ( attr->style != cvt->curr_atts.style )
		cvt->linepos += cprintf(ge, cvt," %s %d", tok[TSTYS],
			cvt->curr_atts.style = attr->style);

	if ( MAX_LINE < cvt->linepos )
		{ cprintf(ge, cvt, "\n"); cvt->linepos=0; wroteSomething=TRUE; }

	if ( attr->face != cvt->curr_atts.face )
		cvt->linepos += cprintf(ge, cvt," %s %d",
			tok[TFONT], cvt->curr_atts.face = attr->face);

	if ( MAX_LINE < cvt->linepos )
		{ cprintf(ge, cvt, "\n"); cvt->linepos=0; wroteSomething=TRUE; }

	if ( attr->size != cvt->curr_atts.size )
		cvt->linepos += cprintf(ge, cvt," %s %d",
			tok[TSIZE], cvt->curr_atts.size = attr->size);

	if ( MAX_LINE < cvt->linepos )
		{ cprintf(ge, cvt, "\n"); cvt->linepos=0; wroteSomething=TRUE; }

	if (attr->ruling != cvt->curr_atts.ruling)
	{
		cvt->linepos += cprintf(ge, cvt," %s %s",tok[TUNDERLINE],und[attr->ruling]);
		cvt->curr_atts.ruling = attr->ruling;
	}

	if (attr->strike != cvt->curr_atts.strike)
	{
		cvt->linepos += cprintf(ge, cvt," %s %s", tok[TTHRU], tst[attr->strike]);
		cvt->curr_atts.strike = attr->strike;
	}

	if ( MAX_LINE < cvt->linepos )
		{ cprintf(ge, cvt, "\n"); cvt->linepos=0; wroteSomething=TRUE; }

	if (attr->weight != cvt->curr_atts.weight )
		cvt->linepos += cprintf(ge, cvt," %s %s",
		    tok[TBOLD], tst[cvt->curr_atts.weight = attr->weight]);

	if ( MAX_LINE < cvt->linepos )
		{ cprintf(ge, cvt, "\n"); cvt->linepos=0; wroteSomething=TRUE; }

	if (attr->slant != cvt->curr_atts.slant )
		cvt->linepos += cprintf(ge, cvt," %s %s",
		    tok[TITALICS], tst[cvt->curr_atts.slant = attr->slant]);

	if ( MAX_LINE < cvt->linepos )
		{ cprintf(ge, cvt, "\n"); cvt->linepos=0; wroteSomething=TRUE; }

	if (attr->x_scale != cvt->curr_atts.x_scale )
		cvt->linepos += cprintf(ge, cvt," %s %d", tok[TSCALE],
			(cvt->curr_atts.x_scale = attr->x_scale));
	
	if ( MAX_LINE < cvt->linepos )
		{ cprintf(ge, cvt, "\n"); cvt->linepos=0; wroteSomething=TRUE; }

	if (attr->shear != cvt->curr_atts.shear )
		cvt->linepos += cprintf(ge, cvt," %s %d", tok[TSHEAR],
			(cvt->curr_atts.shear = attr->shear));
	
	if ( MAX_LINE < cvt->linepos )
		{ cprintf(ge, cvt, "\n"); cvt->linepos=0; wroteSomething=TRUE; }

	if (attr->angle != cvt->curr_atts.angle )
		cvt->linepos += cprintf(ge, cvt," %s %d", tok[TANGLE],
			(cvt->curr_atts.angle = attr->angle));
	
	if ( MAX_LINE < cvt->linepos )
		{ cprintf(ge, cvt, "\n"); cvt->linepos=0; wroteSomething=TRUE; }

	if ( attr->x_off != cvt->curr_atts.x_off )
		cvt->linepos += cprintf(ge, cvt," %s %d 0", tok[TXYOFFS],
			(cvt->curr_atts.x_off = attr->x_off));
	
	if ( MAX_LINE < cvt->linepos )
		{ cprintf(ge, cvt, "\n"); cvt->linepos=0; wroteSomething=TRUE; }

	if ( attr->y_sub_sup != cvt->curr_atts.y_sub_sup )
		cvt->linepos += cprintf(ge, cvt," %s 0 %d",tok[TSUBSUP],
			(cvt->curr_atts.y_sub_sup = attr->y_sub_sup));

	if ( MAX_LINE < cvt->linepos )
		{ cprintf(ge, cvt, "\n"); cvt->linepos=0; wroteSomething=TRUE; }

	if (attr->h_just != cvt->curr_atts.h_just )
		cvt->linepos += cprintf(ge, cvt," %s %s",
		    tok[THOR_JUST], jst[cvt->curr_atts.h_just = attr->h_just]);

	if ( MAX_LINE < cvt->linepos )
		{ cprintf(ge, cvt, "\n"); cvt->linepos=0; wroteSomething=TRUE; }

	if (attr->v_just != cvt->curr_atts.v_just )
		cvt->linepos += cprintf(ge, cvt," %s %s",
		    tok[TVER_JUST], vjst[(cvt->curr_atts.v_just = attr->v_just)]);

	if ( MAX_LINE < cvt->linepos )
		{ cprintf(ge, cvt, "\n"); cvt->linepos=0; wroteSomething=TRUE; }

	if (attr->field_space != cvt->curr_atts.field_space )
	{	
		ln_size = attr->field_space * 72.0/1000.0 + 0.5;
		cvt->linepos += cprintf(ge, cvt," %s %d",
		    tok[TLINE_SPACE], cvt->curr_atts.field_space = ln_size);
	}
	if ( MAX_LINE < cvt->linepos )
		{ cprintf(ge, cvt, "\n"); cvt->linepos=0; wroteSomething=TRUE; }

	if ( attr->l_margin != cvt->curr_atts.l_margin ||
	     attr->r_margin != cvt->curr_atts.r_margin ||
	     attr->t_margin != cvt->curr_atts.t_margin ||
	     attr->b_margin != cvt->curr_atts.b_margin )
	  cvt->linepos += cprintf(ge, cvt," %s %d %d %d %d", tok[TMARGINS],
				  cvt->curr_atts.l_margin = attr->l_margin,
				  cvt->curr_atts.r_margin = attr->r_margin,
				  cvt->curr_atts.t_margin = attr->t_margin,
				  cvt->curr_atts.b_margin = attr->b_margin);
	
	if ( MAX_LINE < cvt->linepos )
	  { cprintf(ge, cvt, "\n"); cvt->linepos=0; wroteSomething=TRUE; }
	
	if ((attr->field_x_scale != cvt->curr_atts.field_x_scale) ||
	    (attr->field_y_scale != cvt->curr_atts.field_y_scale))
	  {
	    cvt->linepos += cprintf(ge, cvt," %s %d,%d",tok[TFLD_XYSCALE],
				    (cvt->curr_atts.field_x_scale=attr->field_x_scale),
				    (cvt->curr_atts.field_y_scale=attr->field_y_scale));
	  }
	
	if ( MAX_LINE < cvt->linepos )
		{ cprintf(ge, cvt, "\n"); cvt->linepos=0; wroteSomething=TRUE; }

	if ((attr->field_shear != cvt->curr_atts.field_shear) ||
		(attr->field_angle != cvt->curr_atts.field_angle))
	{
		cvt->linepos += cprintf(ge, cvt," %s %d %d",tok[TSHR_ANG],
			(cvt->curr_atts.field_shear=attr->field_shear),
			(cvt->curr_atts.field_angle=attr->field_angle));
	}

	if (cvt->linepos)
	{
		cputc(ge,cvt,'\n');
		cvt->linepos = 0;
		wroteSomething=TRUE;
	}

	return(wroteSomething);
}
ag31OutLayer(ge, layer, cvt)
graphics_main_type *ge;
int layer;
cvt_global_type *cvt;
{

	if ( MAX_LINE < cvt->linepos )
		{ cprintf(ge, cvt, "\n"); cvt->linepos=0; }

	if ( layer != cvt->curr_layer )
	{
		cvt->linepos += cprintf(ge, cvt," %s %d", tok[TLAYERS], layer);
		cvt->curr_layer = layer;
	}
}

/****************************************************************************
 *
 * module:	a g 3 1 O u t O b j H d r - Write the object's header to file
 */
ag31OutObjHdr(ge, t, cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	int x1,y1,x2,y2, ii, idTag;
	ge_tag_type *tag;

	/* Put any tags (comments) associated with an object */
	idTag = FALSE;
	if (t->attr.use.tag)
	{
		tag = &ge->tag[t->attr.tag];

		if (tag)
		{
			/* wrap tags at 80 characters */
			for (ii=0; ii<tag->n_strs; ii++)
			{
				cprintf(ge, cvt,"\n#%s",tag->str[ii]);
				if (!strncmp(tag->str[ii],"#id",3))
					idTag = TRUE;
			}
		}
	}

	
#if vamsi	
	/* type of object (note charts are zero based) */
	cprintf(ge, cvt,"\n%s ", pnm[t->type]);
#endif
	
	/* type of object (note charts are zero based) */
     if ((t->type == GROUP) && (t->chart_ix))
          cprintf(ge, cvt,"\n%s %d ", tok[TCHART], t->chart_ix-1);
     else
          cprintf(ge, cvt,"\n%s ", pnm[t->type]);

	if (t->attr.callback > 0)
		cprintf(ge, cvt,"\n%s %d ", tok[TCBACK], t->attr.callback - 1);
	
	/* HIDDEN mark */
	if (t->flags.hidden)
		cprintf(ge, cvt,"%s ", tok[THIDDEN]);

	/* FIXED mark */
	if (t->flags.fixed)
		cprintf(ge, cvt,"%s", tok[TFIXED]);

	/* need a newline */
	cprintf(ge, cvt, "\n");
	cvt->linepos = 0;

#if FOR_FILTER
	if ((t->type == TEXTBOX) &&
	    (t->extent.x1 == t->extent.x2) &&
	    (t->extent.y1 == t->extent.y2))
	{
		/* a filter file. Use the position keyword. */
		cprintf(ge, cvt," %s (%d,%d)\n",
			tok[TPOSITION], t->extent.x1-PAGE_ZEROX, t->extent.y1-PAGE_ZEROY);
		return;
	}

#endif

	/* object's bounding box  - no more 'same' */
		if (t->extent.x1 <= t->extent.x2)
			{ x1 = t->extent.x1; x2 = t->extent.x2; }
		else
			{ x2 = t->extent.x1; x1 = t->extent.x2; }
		
		if (t->extent.y1 <= t->extent.y2)
			{ y1 = t->extent.y1; y2 = t->extent.y2; }
		else
			{ y2 = t->extent.y1; y1 = t->extent.y2; }

	cprintf(ge, cvt," %s (%d,%d) %s (%d,%d)\n",
		tok[TORIGIN], x1-PAGE_ZEROX, y1-PAGE_ZEROY,
		tok[TCORNER], x2-PAGE_ZEROX, y2-PAGE_ZEROY);
	
}

/****************************************************************************
 *
 * module:	a g 3 1 O u t N a m e - write out a palette part name (obsolete)
 */
ag31OutName(obj)
udo_type *obj; /* ??? */
{
}

/****************************************************************************
 *
 * module: a g 3 1 O u t T r a n s
 */
ag31OutTrans(ge, trans, cvt)
graphics_main_type *ge;
poly_type *trans;
cvt_global_type *cvt;
{
	if ((trans->nPoints > 0) || (trans->nCtrls > 0))
		cprintf(ge, cvt," %s\n",tok[TTRANS]);
	ag31OutPoints(ge, trans, cvt);
	ag31OutCtrls(ge, trans, cvt);
}

/****************************************************************************
 *
 * module: a g 3 1 O u t P a t h
 */
ag31OutPath(ge, PathP, nPaths, cvt)
graphics_main_type *ge;
poly_type *PathP;
int nPaths;
cvt_global_type *cvt;
{
	int ii;

	if (nPaths > 0 )
		cprintf(ge, cvt," %s\n",tok[TOUTLINE]);
	for (ii=0; ii<nPaths; ii++)
	{
		ag31OutPoints(ge, &PathP[ii], cvt);
		ag31OutCtrls(ge, &PathP[ii], cvt);
	}
}

/****************************************************************************
 *
 * module: a g 3 1 O u t P o i n t s
 */
ag31OutPoints(ge, PathP, cvt)
graphics_main_type *ge;
poly_type *PathP;
cvt_global_type *cvt;
{
	int ii, x, y, llen;

	/* write any Point infomation */

	/* get count of coord pairs and write them to file */
	if (PathP->nPoints > 0)
	{
			llen = cprintf(ge, cvt," %s ", tok[TPOINTS]);

		for (ii=0; ii<PathP->nPoints; ii++)
		{
			x = PathP->points[ii].x - PAGE_ZEROX;
			y = PathP->points[ii].y - PAGE_ZEROY;
			if ( MAX_LINE < llen )
			{
				cprintf(ge, cvt, "\n        ");
				llen=0;
			}
			llen += cprintf(ge, cvt,"(%d,%d)", x, y);
		}

		if ( llen )
			cputc(ge,cvt,'\n');
	}
}

/****************************************************************************
 *
 * module: a g 3 1 O u t C t r l s
 */
ag31OutCtrls(ge, PathP, cvt)
graphics_main_type *ge;
poly_type *PathP;
cvt_global_type *cvt;
{
	int ii, x, y, llen;

	/* write any control points */


	if (PathP->nCtrls > 0)
	{
		llen = cprintf(ge, cvt," %s ", tok[TCONTROL_PTS]);

		for (ii=0; ii<PathP->nCtrls; ii++)
		{
			x = PathP->controls[ii].x - PAGE_ZEROX;
			y = PathP->controls[ii].y - PAGE_ZEROY;
			llen += cprintf(ge, cvt,"(%d,%d)", x, y);
			if ( MAX_LINE < llen )
			{
				cprintf(ge, cvt, "\n                ");
				llen=0;
			}
		}

		if ( llen )
			cputc(ge,cvt,'\n');
	}
}

/****************************************************************************
 *
 * module: a g 3 1 O u t E l l
 */
ag31OutEll(ge,t,cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	poly_type path;
	TmPoint32 pnts[16], ctls[16], *points, *controls;
	int sz, ii;
	thing newt;

	cvt->linepos = 0;
#if 0
	if (cvt->as_1_1 && (t->data.ell->n_sides > 0))
	{
		/* convert the regular poly to a path */
		getEllPath(t,&path);
		ag31OutPath(ge,&path,1,cvt);
		free(path.points);
		return;
	}

	if (cvt->as_1_1 && (!isRectForm(t->paths[0].points)))
	{
		path.points   = pnts;
		path.controls = ctls;

		ellToPath(t->paths->points,
			t->data.ell->start_angle,
			t->data.ell->end_angle,
			FALSE, &path);

		/* change the ell path to a poly path */
		newt.paths = (poly_type*)malloc(sizeof(poly_type));
		sz = (path.nPoints/2 + 1) * sizeof(TmPoint32);
		newt.paths->points = (TmPoint32*)malloc(sz);
		sz = path.nCtrls * sizeof(TmPoint32);
		newt.paths->controls = (TmPoint32*)malloc(sz);;
		newt.paths->nPoints = path.nPoints/2;
		newt.paths->nCtrls = path.nCtrls;
		for (ii=0; ii<newt.paths->nCtrls; ii++ )
			newt.paths->controls[ii] = ctls[ii];
		for (ii=0; ii<newt.paths->nPoints; ii++ )
			newt.paths->points[ii] = pnts[ii*2];

		/* close the path */
		newt.paths->points[ii] = pnts[0];
		newt.paths->nPoints++;

		ag31OutPoly(ge,&newt,cvt);
		free(newt.paths->points);
		free(newt.paths->controls);
		free(newt.paths);
		return;
	}
#endif

		ag31OutPath(ge, t->paths, t->n_paths,cvt);

	if (t->data.ell->start_angle != cvt->startAngle)
		cvt->linepos += cprintf(ge, cvt," %s %d", tok[TSTART_ANGLE],
			cvt->startAngle = t->data.ell->start_angle);
	if (t->data.ell->end_angle != cvt->endAngle)
		cvt->linepos += cprintf(ge, cvt," %s %d", tok[TEND_ANGLE],
			cvt->endAngle = t->data.ell->end_angle);
	if ((t->data.ell->n_sides != cvt->nsides) )
		cvt->linepos += cprintf(ge, cvt," %s %d", tok[TNSIDES],
			cvt->nsides = t->data.ell->n_sides);

	if ( cvt->linepos )
		cputc(ge,cvt,'\n');
}

/****************************************************************************
 *
 * module: a g 3 1 O u t R e c t
 */
ag31OutRect(ge,t,cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	int llen=0;


	if ((t->data.rect->x_corner_radius != cvt->x_corner_radius) ||
		(t->data.rect->y_corner_radius != cvt->y_corner_radius))
	{
		llen += cprintf(ge, cvt," %s %d %d\n", tok[TXYRAD],
			cvt->x_corner_radius = t->data.rect->x_corner_radius,
			cvt->y_corner_radius = t->data.rect->y_corner_radius);
	}
	if (t->data.rect->scale_corners != cvt->scale_corners)
	{
		llen += cprintf(ge, cvt," %s %d\n", tok[TSCALE_CORNERS],
			cvt->scale_corners = t->data.rect->scale_corners);
	}

	if ((t->data.rect->x_corner_numerator != cvt->x_corner_numerator) ||
		(t->data.rect->y_corner_numerator != cvt->y_corner_numerator))
	{
			llen += cprintf(ge, cvt," %s %d %d\n", tok[TPOSITION],
				cvt->x_corner_numerator = t->data.rect->x_corner_numerator,
				cvt->y_corner_numerator = t->data.rect->y_corner_numerator);
	}

	ag31OutPath(ge, t->paths, t->n_paths,cvt);
}

/****************************************************************************
 *
 * module: a g 3 1 O u t L i n e
 */
ag31OutLine(ge,t,cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{

  ag31OutPath(ge, t->paths, t->n_paths,cvt);
}

/****************************************************************************
 *
 * module: a g 3 1 O u t S t r o k e
 */
ag31OutStroke(ge,t,cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{

	ag31OutPoly(ge,t,cvt);
}

extern short THIMpid;
/****************************************************************************
 *
 * module: a g 3 1 O u t P o l y
 */
ag31OutPoly(ge,t,cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	poly_type path;
	int npts;			/* number of points converted poly */
	TmPoint32 *newPointsP=NULL;	/* vertices of converted poly */

	/* no combined object in 3.x */
	ag31OutPath(ge, t->paths, 1, cvt);
}

/****************************************************************************
 *
 * module:  a g 3 1 O u t T e x t  - write the string(s) Rev 2.1 style
 */
ag31OutTextString(ge,cvt,t)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
{
	if (t->data.str->n_chars > 0)
	{
		cputc(ge,cvt,' ');
		ag31OutString(ge,t->data.str->n_chars,
		   t->data.str->chars,cvt, FALSE);
		cputc(ge,cvt,'\n');
	}
}


/****************************************************************************
 *
 * module:  a g 3 1 O u t T e x t  - write the string(s)
 */
ag31OutText(ge,t,cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	int ii;
	thing *txt_t;

	/* Write out the actual strings */
	for (ii=0; ii<t->data.txtbox->n_strings; ii++)
	{
		txt_t = t->data.txtbox->strings[ii];

		cprintf(ge, cvt," %s\n", pnm[TEXT]);
		ag31OutAttrs(ge, &txt_t->attr, 0, cvt);
		ag31OutPath(ge, txt_t->paths, txt_t->n_paths, cvt);
		ag31OutTextString(ge,cvt,txt_t);
	}
}

/****************************************************************************
 *
 * module:	a g 3 1 O u t B o x T e x t -  TEXTBOX
 */
ag31OutTextBox(ge,t,cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{


	if ((cvt->x_off != t->data.txtbox->x_off) ||
		(cvt->y_off != t->data.txtbox->y_off))
	{
			cprintf(ge, cvt," %s %d %d\n", tok[TTXTXYOFF],
				cvt->x_off = t->data.txtbox->x_off,
				cvt->y_off = t->data.txtbox->y_off);
	}

	/* draw the bounding polygon(s) */
	ag31OutPath(ge, t->paths, t->n_paths, cvt);
	ag31OutTrans(ge, &t->data.txtbox->trans, cvt);

	/* the actual text */
	ag31OutText(ge,t,cvt);
}

/****************************************************************************
 *
 * module:	
 */
ag31OutTextBoxStrings(ge,t,cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	int i;

	t = (thing *)cvtInsertNewLines(ge, t);
	if ( t->type == GROUP )
	{
		for ( i=0; i<t->data.group->num; i++)
			ag31OutPart(ge, t->data.group->list[i], cvt);
	}
	else
	{
		ag31OutObjHdr(ge, t, cvt);
		ag31OutAttrs(ge, &t->attr, 0, cvt);
		ag31OutLayer(ge, t->layer, cvt);
		ag31OutTextBox(ge, t, cvt);
	}
}

/****************************************************************************
 *
 * module:	a g 3 1 O u t G r o u p - export a Grouped object
 */
ag31OutGroup(ge, t, cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	int ii;

	/* Most confusing mistake that ever was.  Not really
	   a user defined object, simply a group of objects. 
	 */

	for ( ii=0; ii<t->data.group->num; ii++ )
		ag31OutPart(ge, t->data.group->list[ii], cvt);

	if ( t->chart_ix )
          cprintf(ge, cvt,"%s %s\n", tok[TEND], tok[TCHART]);
     else
          cprintf(ge, cvt,"%s %s\n", tok[TEND], pnm[GROUP]);
}

/****************************************************************************
 *
 * module:	a g 3 1 O u t P a r t - top-level part 
 */
ag31OutPart(ge, t, cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	int cmap_ix;
	layer_type *layer;

	if ( t == NULL )
	{
		    cprintf(ge, cvt,"\n%s\n",pnm[0]);
	}
	else if( t->type >= TEXTBOX && t->type <= ELL )
	{
		cmap_ix = t->type == IMAGE ? t->data.img->cmapID : 0;

		if ( t->type == GROUP && t->data.group->num == 0)
			return;
		/* if layer is locked /hidden set the hidden flag */
		layer = &ge->page[0].layer[t->layer];
		if ( layer && (layer->locked||layer->hidden) )
			t->flags.hidden = TRUE;

		if ( t->type != TEXTBOX )
		{
			ag31OutObjHdr(ge, t, cvt);
			ag31OutAttrs(ge, &t->attr, cmap_ix, cvt);
			ag31OutLayer(ge, t->layer, cvt);
		}
		switch ( t->type )
		{
			case ELL:     ag31OutEll    (ge,t,cvt); break;
			case RECT:    ag31OutRect   (ge,t,cvt); break;
			case LINE:    ag31OutLine   (ge,t,cvt); break;
			case STROK:   ag31OutStroke (ge,t,cvt); break;
			case POLY:    ag31OutPoly   (ge,t,cvt); break;
			case TEXTBOX: ag31OutTextBoxStrings(ge,t,cvt); break;
			case IMAGE:   ag31OutImage  (ge,t,cvt); break;
			case GROUP:   ag31OutGroup  (ge,t,cvt); break;
		}
	}
}

/****************************************************************************
 *
 * module:	a g 3 1 O u t P i c t
 */
ag31OutPict(ge, t, cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	int i;

	/* note the start of the picture segment */
	cprintf(ge, cvt,"%s", tok[TPICTURE]);

	if (!t || (t->type == 0) || (t->type == GROUP && t->data.group->num == 0))
	{
		cprintf(ge, cvt,"\n%s %s\n\n", tok[TEND], tok[TPICTURE]);
		return;
	}


	/* write out all the parts in the document */
	if ( t->type != GROUP )
		ag31OutPart(ge, t, cvt);
	else
	{
		for ( i=0; i<t->data.group->num; i++ )
			ag31OutPart(ge, t->data.group->list[i], cvt);
	}

	/* note the end of the picture segment */
	cprintf(ge, cvt,"%s %s\n\n", tok[TEND], tok[TPICTURE]);

}


/****************************************************************************
 *
 * module:	a g 3 1 O u t P a l e t
 */
ag31OutPalet(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int ii;
	udo_type udoP;


	/* for each UDO */
	for (ii=0; ii<ge->n_udos; ii++)
	{
	  cprintf(ge, cvt,"%s ", tok[TPART]);
	  ag31OutString(ge,ge->udo[ii].n_chars, ge->udo[ii].name, cvt, FALSE);
	  ag31OutPart(ge, ge->udo[ii].t, cvt);
	}
	if ( ge->n_udos > 0 )
		cprintf(ge, cvt, "\n");
}

/****************************************************************************
 *
 * module:	a g 3 1 O u t N a m e d F i l l s - write out user defined fill defs
 */
ag31OutNamedFills(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int i, j, k, byte_size, byte_width, data_wid, llen, datum;
	udfill_type fill_pack, *fillp;

	if ( ge->n_fills <= 0 )
		return;

	/* note the start of the fill list */
	cprintf(ge, cvt,"%s\n", tok[TFILLNAMES]);

	/* write each fill */
	for ( i=0; i<ge->n_fills; i++ )
	{
		/* convert the fill data to mono, lsb format */
		fill_pack = ge->fill[i];
		fillp = &ge->fill[i];
		data_wid = (fill_pack.wid+7)/8 + ((fill_pack.wid+7)/8)%2;
		byte_size = fill_pack.hyt * data_wid;
		fill_pack.data = (unsigned char *)gloc(ge, byte_size);
		if ( fill_pack.cmapID > 0 )
			memset(fill_pack.data, 0xFF, byte_size);
		else for ( j=byte_size-1; j>=0; j-- )
			fill_pack.data[j] = BitSpin(fillp->data[j]);

		/* print the "Header" */
		mlOutString(ge, fill_pack.n_chars, fill_pack.name, cvt, FALSE);
		cprintf(ge, cvt," %d %d\n ", fill_pack.hyt, fill_pack.wid);;

		/* output the data 8 bits at a time (not byte aligned) */
		llen = 0;
		byte_width = (fill_pack.wid+7)/8;
		for ( j=0; j<fill_pack.hyt; j++ )
		{
			for ( k=0; k<byte_width; k++ )
			{
				datum = fill_pack.data[j*data_wid + k];
				llen += mailOutBit6(ge, datum&0xff, llen, cvt);
				if ( MAX_LINE < llen )
				{
					cputc(ge,cvt,'\n');
					cputc(ge,cvt,' ');
					llen=0;
				}
			}
			llen += mailOutBit6(ge, -1, llen, cvt);
		}

		/* done */
		if ( llen > 0 )
			cputc(ge,cvt,'\n');
		gfree(ge, fill_pack.data);
	}

	/* note the end of the fill list */
	cprintf(ge, cvt,"%s %s\n\n", tok[TEND], tok[TFILLNAMES]);
}

/****************************************************************************
 *
 * module:	a g 3 1 O u t F a c e s
 */
ag31OutFaces(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
  	int i;
	int len;
	tx_char tbuff[500];


	/* note the start of the face list */
	if (ge->n_faces == 0)
		return;

	cprintf(ge, cvt,"%s\n", tok[TFONT]);

	for ( i=0; i<ge->n_faces; i++ )
	  {
	    len = MultiToWideStr(ge,ge->face[i], tbuff);
	    ag31OutString(ge, len, tbuff, cvt, FALSE );
	    cputc(ge,cvt,'\n');
	  }
	if (ge->n_faces == 0)
	  {
	    len = MultiToWideStr(ge,fnt[0], tbuff);
	    ag31OutString(ge, len, tbuff, cvt, FALSE );
	    cputc(ge,cvt,'\n');
	  }

	/* note the end of the fill list */
	cprintf(ge, cvt,"%s %s\n\n", tok[TEND], tok[TFONT]);
}




/****************************************************************************
 *
 * module:	a g 3 1 O u t C o l o r m a p - write the colormap to the file
 */
ag31OutColormap(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int ii, jj;
	color_type *colorP;
	int len, same;
	tx_char tbuff[200];
	graphics_main_type localGE;

	
	if (ge->n_cmaps == 0) /* gotta have 1 */
	  {
	    colormapInit(ge);
	    return; /* since the next condition is satisfied */
	  }

	for (ii=0; ii<ge->n_cmaps; ii++ )
	{
		/* note the start of the colormap */
		cprintf(ge, cvt,"%s\n", tok[TCOLORMAP]);

		/* ordinal list of color names, definitions, and ink types */
		colorP = ge->cmap[ii].color;
		for ( jj=0; jj<ge->cmap[ii].n_colors; jj++ )
		{
			/* first the name of the color */
			len = MultiToWideStr(ge,colorP[jj].name, tbuff );
			ag31OutString(ge, len, tbuff, cvt, FALSE );

			/* then the ink type */
			cprintf(ge, cvt," %d ", colorP[jj].ink_type);

			/* finally the cmyk definition */
			cprintf(ge, cvt,"%d %d %d %d %d\n", colorP[jj].c,
			                                colorP[jj].m,
			                                colorP[jj].y,
			                                colorP[jj].k,
			                                colorP[jj].see_thru);
		}

		/* note the end of the colormap */
		cprintf(ge, cvt,"%s %s\n\n", tok[TEND], tok[TCOLORMAP]);
	}
}

/****************************************************************************
 *
 * module:	a g 3 1 O u t L a y e r s
 */
ag31OutLayers(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int i, j, len;
	layer_type *l_pack;
	tx_char tbuff[400];

	if (ge->page[0].n_layers == 0)
		return;

	/* note the start of the layers info section */
	cprintf(ge, cvt,"%s\n", tok[TLAYERS]);

	/* output info for each layer */
	for ( i=0; i<ge->page[0].n_layers; i++ )
	{
		l_pack = &ge->page[0].layer[i];

		/* print the layer name */
		ag31OutString(ge,l_pack->n_chars, l_pack->name, cvt, FALSE);

		/* locked and hidden tags */
		cprintf(ge, cvt," %d %d\n", l_pack->locked, l_pack->hidden);

	}

	/* note the end of the layers info section */
	cprintf(ge, cvt,"%s %s\n\n", tok[TEND], tok[TLAYERS]);
}

/****************************************************************************
 *
 * module:  a g 3 1 O u t S e s s i o n I n f o
 */
ag31OutSessionInfo(ge,cvt,t)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
{
	struct ge_print_info_type *prnt;
	struct session *sess;


	cprintf(ge, cvt,"%s\n", tok[TSESSION]);

	/* window info */
	ag31OutWindowSize(ge, cvt, t);

	/* document info */
	sess = &ge->state;
	if ( sess->stickyDots )
		cprintf(ge, cvt," %s\n",tok[TSTICKY]);
	if ( sess->autoGrid )
		cprintf(ge, cvt," %s\n",tok[TAUTOGRID]);
	if ( sess->rulerOn )
		cprintf(ge, cvt," %s\n",tok[TRULERON]);
	if ( sess->pageOn )
		cprintf(ge, cvt," %s 0 0\n",tok[TPAGEON]);
	if ( sess->readOut )
		cprintf(ge, cvt," %s\n",tok[TREADOUT]);
	if ( sess->gridOn )
		cprintf(ge, cvt," %s\n",tok[TGRIDS]);
	if ( sess->im_hex || sess->im_raw )
	{
		if ( sess->im_hex && sess->im_raw )
			cprintf(ge, cvt," %s 3\n",tok[TIMAGE_FMT]);
		else if ( sess->im_hex )
			cprintf(ge, cvt," %s 2\n",tok[TIMAGE_FMT]);
		else if ( sess->im_raw )
			cprintf(ge, cvt," %s 1\n",tok[TIMAGE_FMT]);
		else
			cprintf(ge, cvt," %s 0\n",tok[TIMAGE_FMT]);
	}

	if (sess->dpi != 1000 || sess->gridSize != 250 || sess->unit_size != 1000)
	{
		cprintf(ge, cvt," %s %d %s %d %s %d %d \"%s\"\n",
			tok[TGRID_DPI], sess->dpi, tok[TGRID_FACTOR], sess->gridSize,
			tok[TUNITS], sess->unit_size,sess->unit_precision,sess->unit_name);
	}

	if ((sess->line_1 != 27) || /* defaults */
	    (sess->line_2 != 40) ||
	    (sess->line_3 != 54))
	{
		cprintf(ge, cvt," %s %d %d %d\n",tok[TWIDTH],
				sess->line_1, sess->line_2, sess->line_3);
	}

	/* print setup info */
	prnt = &ge->print;

	prnt->margin.y1 += 500;
	prnt->margin.y2 += 500;
	/* printable area - position, size, and zero point */
	if ( prnt->page_wid == 0 && prnt->page_hyt == 0 ) /* HACK */
	{
		cprintf(ge, cvt," %s 59750 58000\n", tok[TPAGEXY]);
		cprintf(ge, cvt," %s 8500 ", tok[TPAGEWID]);
		cprintf(ge, cvt," %s 11000\n", tok[TPAGEHYT]);
		cprintf(ge, cvt," %s", tok[TPAGEZEROS]);
		cprintf(ge, cvt," %d %d\n", 59700 + prnt->margin.x1,
		                        58000 + prnt->margin.y1);
	}
	else
	{
		cprintf(ge, cvt," %s %d %d\n", tok[TPAGEXY], prnt->page_x, prnt->page_y);
		cprintf(ge, cvt," %s %d ",  tok[TPAGEWID], prnt->page_wid);
		cprintf(ge, cvt," %s %d\n", tok[TPAGEHYT], prnt->page_hyt);
		cprintf(ge, cvt," %s", tok[TPAGEZEROS]);
		cprintf(ge, cvt," %d %d\n", ge->x_zero_pt, ge->y_zero_pt+500);
	}

	/* paper type */
	cprintf(ge, cvt," %s %d", tok[TPRINTWID], prnt->wid);
	cprintf(ge, cvt," %s %d", tok[TPRINTHYT], prnt->hyt);
	if ( prnt->landscape )
		cprintf(ge, cvt," %s", tok[TPRINTLAND]);
	cprintf(ge, cvt, "\n");
	cprintf(ge, cvt," %s %d %d %d %d\n", tok[TPRINTMARG],
	                                 prnt->margin.x1,
	                                 prnt->margin.y1,
	                                 prnt->margin.x2,
	                                 prnt->margin.y2);
	cprintf(ge, cvt, " %s %s %s %s\n", tok[TPRINTMARKS],
	                               tst[prnt->use_content_border],
	                               tst[prnt->use_crop_marks],
	                               tst[prnt->use_regis_marks]);
	if ( prnt->bleed )
		cprintf(ge, cvt," %s %d \n", tok[TPRINTBLEED], prnt->bleed);

	/* be sure the shadow color is valid */
	if (ge->def_attr.shadow_color >= ge->cmap[0].n_colors)
		ge->def_attr.shadow_color = 0;

	/* default attributes */
	cvt->linepos=0;
	ag31OutAttrs(ge, &ge->def_attr, 0, cvt);
	ag31OutLayer(ge, ge->state.layer, cvt);

	cprintf(ge, cvt,"%s %s\n\n", tok[TEND], tok[TSESSION]);
}

/****************************************************************************
 *
 * module:  a g 3 1 O u t H d r s F t r s
 */
ag31OutHdrsFtrs(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	if ( ag31NeedHdrsFtrs(ge,cvt) == FALSE )
		return;

	/* set defaults */
	memset(&cvt->hfstr,0,sizeof(ge_hdrftr_str_type));
	cvt->hfstr.pointsize = 120000;
	cvt->hfstr.color_ix = 1;

	/* write out the header */
	cprintf(ge, cvt,"%s", tok[THDRFTR]);
	if (ge->print.hdrftr.use_first)
		cprintf(ge, cvt,"  %s %s", tok[TUSE],tok[TFIRST]);
	if (ge->print.hdrftr.use_even_odd)
		cprintf(ge, cvt,"  %s %s", tok[TUSE],tok[TEVENODD]);
	if (ge->print.hdrftr.use_final)
		cprintf(ge, cvt,"  %s %s", tok[TUSE],tok[TFINAL]);
	cprintf(ge, cvt,"\n");

	/* write out the special first page hdr/ftr info */
	if (ge->print.hdrftr.first.hdr_offset ||
		ge->print.hdrftr.first.hdr_n_lines ||
		ge->print.hdrftr.first.ftr_offset ||
		ge->print.hdrftr.first.ftr_n_lines)
	{
		cprintf(ge, cvt," %s\n", tok[TFIRST]);
		if (ge->print.hdrftr.first.hdr_offset ||
			ge->print.hdrftr.first.hdr_n_lines)
		{
			cvt->hdr_ftr = &ge->print.hdrftr.first;
			ag31OutHdrPage(ge,cvt);
		}
		if (ge->print.hdrftr.first.ftr_offset ||
			ge->print.hdrftr.first.ftr_n_lines)
		{
			cvt->hdr_ftr = &ge->print.hdrftr.first;
			ag31OutFtrPage(ge,cvt);
		}
		cprintf(ge, cvt," %s %s\n", tok[TEND], tok[TFIRST]);
	}

	/* write out the basic page hdr/ftr info (also used for odd pages) */
	if (ge->print.hdrftr.base.hdr_offset ||
		ge->print.hdrftr.base.hdr_n_lines ||
		ge->print.hdrftr.base.ftr_offset ||
		ge->print.hdrftr.base.ftr_n_lines)
	{
		cprintf(ge, cvt," %s\n", tok[TBASE]);
		if (ge->print.hdrftr.base.hdr_offset ||
			ge->print.hdrftr.base.hdr_n_lines)
		{
			cvt->hdr_ftr = &ge->print.hdrftr.base;
			ag31OutHdrPage(ge,cvt);
		}
		if (ge->print.hdrftr.base.ftr_offset ||
			ge->print.hdrftr.base.ftr_n_lines)
		{
			cvt->hdr_ftr = &ge->print.hdrftr.base;
			ag31OutFtrPage(ge,cvt);
		}
		cprintf(ge, cvt," %s %s\n", tok[TEND], tok[TBASE]);
	}

	/* write out the special even page hdr/ftr info */
	if (ge->print.hdrftr.even.hdr_offset ||
		ge->print.hdrftr.even.hdr_n_lines ||
		ge->print.hdrftr.even.ftr_offset ||
		ge->print.hdrftr.even.ftr_n_lines)
	{
		cprintf(ge, cvt," %s\n", tok[TEVEN]);
		if (ge->print.hdrftr.even.hdr_offset ||
			ge->print.hdrftr.even.hdr_n_lines)
		{
			cvt->hdr_ftr = &ge->print.hdrftr.even;
			ag31OutHdrPage(ge,cvt);
		}
		if (ge->print.hdrftr.even.ftr_offset ||
			ge->print.hdrftr.even.ftr_n_lines)
		{
			cvt->hdr_ftr = &ge->print.hdrftr.even;
			ag31OutFtrPage(ge,cvt);
		}
		cprintf(ge, cvt," %s %s\n", tok[TEND], tok[TEVEN]);
	}

	/* write out the special final page hdr/ftr info */
	if (ge->print.hdrftr.final.hdr_offset ||
		ge->print.hdrftr.final.hdr_n_lines ||
		ge->print.hdrftr.final.ftr_offset ||
		ge->print.hdrftr.final.ftr_n_lines)
	{
		cprintf(ge, cvt," %s\n", tok[TFINAL]);
		if (ge->print.hdrftr.final.hdr_offset ||
			ge->print.hdrftr.final.hdr_n_lines)
		{
			cvt->hdr_ftr = &ge->print.hdrftr.final;
			ag31OutHdrPage(ge,cvt);
		}
		if (ge->print.hdrftr.final.ftr_offset ||
			ge->print.hdrftr.final.ftr_n_lines)
		{
			cvt->hdr_ftr = &ge->print.hdrftr.final;
			ag31OutFtrPage(ge,cvt);
		}
		cprintf(ge, cvt," %s %s\n", tok[TEND], tok[TFINAL]);
	}

	cprintf(ge, cvt,"%s %s\n\n", tok[TEND], tok[THDRFTR]);

	/* make sure hfstr attrs are in sync with curr_atts */
	if ( cvt->hfstr.face_ix != cvt->curr_atts.face )
		cvt->curr_atts.face = cvt->hfstr.face_ix;
	if ( cvt->hfstr.pointsize != cvt->curr_atts.size )
		cvt->curr_atts.size = cvt->hfstr.pointsize;
	if ( cvt->hfstr.color_ix != cvt->curr_atts.fill_fg )
		cvt->curr_atts.fill_fg = cvt->hfstr.color_ix;
	if ( cvt->hfstr.bold != cvt->curr_atts.weight )
		cvt->curr_atts.weight = cvt->hfstr.bold;
	if ( cvt->hfstr.italic != cvt->curr_atts.slant )
		cvt->curr_atts.slant = cvt->hfstr.italic;
	if ( cvt->hfstr.strike_thru != cvt->curr_atts.strike )
		cvt->curr_atts.strike = cvt->hfstr.strike_thru;
	if ( cvt->hfstr.underline != cvt->curr_atts.ruling )
		cvt->curr_atts.ruling = cvt->hfstr.underline;
}

/****************************************************************************
 *
 * module:  a g 3 1 N e e d H d r F t r s
 */
ag31NeedHdrsFtrs(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	struct ge_hdrftr_main_type *hf;
	int needed;

	hf = &ge->print.hdrftr;
	if ( hf->base.hdr_n_lines > 0 || hf->base.ftr_n_lines > 0 )
		needed = TRUE;
	else
		needed = FALSE;
	if ( needed == FALSE  && hf->use_even_odd )
	{
		if ( hf->even.hdr_n_lines > 0 || hf->even.ftr_n_lines > 0 )
			needed = TRUE;
	}
	if ( needed == FALSE && hf->use_first )
	{
		if ( hf->first.hdr_n_lines > 0 || hf->first.ftr_n_lines > 0 )
			needed = TRUE;
	}
	if ( needed == FALSE && hf->use_final )
	{
		if ( hf->final.hdr_n_lines > 0 || hf->final.ftr_n_lines > 0 )
			needed = TRUE;
	}
	return(needed);
}

/****************************************************************************
 *
 * module:  a g 3 1 O u t H d r P a g e
 */
ag31OutHdrPage(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int ii;

	cprintf(ge, cvt,"  %s ", tok[THEADER]);
	cprintf(ge, cvt,"%s %d ", tok[TOFFSET], cvt->hdr_ftr->hdr_offset);
	cprintf(ge, cvt,"%s %d\n", tok[TNLINES], cvt->hdr_ftr->hdr_n_lines);
	for (ii=0; ii<cvt->hdr_ftr->hdr_n_lines; ii++)
	{
		cvt->hf = &cvt->hdr_ftr->hdr[ii];
		ag31OutHF(ge,cvt,ii);
	}
	cprintf(ge, cvt,"  %s %s\n", tok[TEND], tok[THEADER]);
}

/****************************************************************************
 *
 * module:  a g 3 1 O u t F t r P a g e
 */
ag31OutFtrPage(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int ii;

	cprintf(ge, cvt,"  %s ", tok[TFOOTER]);
	cprintf(ge, cvt,"%s %d ", tok[TOFFSET], cvt->hdr_ftr->ftr_offset);
	cprintf(ge, cvt,"%s %d\n", tok[TNLINES], cvt->hdr_ftr->ftr_n_lines);
	for (ii=0; ii<cvt->hdr_ftr->ftr_n_lines; ii++)
	{
		cvt->hf = &cvt->hdr_ftr->ftr[ii];
		ag31OutHF(ge,cvt,ii);
	}
	cprintf(ge, cvt,"  %s %s\n", tok[TEND], tok[TFOOTER]);
}

/****************************************************************************
 *
 * module:	ag31OutHF
 */
ag31OutHF(ge,cvt,lineNum)
graphics_main_type *ge;
cvt_global_type *cvt;
int lineNum;
{
	cvt->ico = &cvt->hf->inner;
	cprintf(ge, cvt,"  %s %d\n", tok[TINNER], lineNum);
	ag31OutHFstring(ge,cvt);
	cprintf(ge, cvt,"  %s %s\n", tok[TEND], tok[TINNER]);

	cvt->ico = &cvt->hf->center;
	cprintf(ge, cvt,"  %s %d\n", tok[TCENTER], lineNum);
	ag31OutHFstring(ge,cvt);
	cprintf(ge, cvt,"  %s %s\n", tok[TEND], tok[TCENTER]);

	cvt->ico = &cvt->hf->outer;
	cprintf(ge, cvt,"  %s %d\n", tok[TOUTER], lineNum);
	ag31OutHFstring(ge,cvt);
	cprintf(ge, cvt,"  %s %s\n", tok[TEND], tok[TOUTER]);
}

/****************************************************************************
 *
 * module:	ag31OutHFstring
 */
ag31OutHFstring(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	tx_char *str;
	int len, i;

	if ( !cvt->ico->string || !cvt->ico->string[0] )
		cprintf(ge, cvt,"   %s \"\"\n", pnm[TEXT]);
	else
	{   
		len = strlen(cvt->ico->string);
		str = (tx_char *)malloc((len+1) * sizeof(tx_char));
		for ( i=0; i<len; i++ )
			str[i] = cvt->ico->string[i];
		str[i] = (tx_char)0;

		cprintf(ge, cvt,"   %s \"", pnm[TEXT]);
		ag31OutString(ge,len, str, cvt, TRUE);
		cprintf(ge, cvt,"\n");

		free(str);
	}

	if (cvt->hfstr.face_ix != cvt->ico->face_ix)
		cprintf(ge, cvt,"   %s %d\n", tok[TFONT],
			cvt->hfstr.face_ix = cvt->ico->face_ix);

	if (cvt->hfstr.pointsize != cvt->ico->pointsize)
		cprintf(ge, cvt,"   %s %d\n", tok[TSIZE],
			cvt->hfstr.pointsize = cvt->ico->pointsize);

	if (cvt->hfstr.color_ix != cvt->ico->color_ix)
		cprintf(ge, cvt,"   %s %d\n", tok[TFORE_COLOR],
			cvt->hfstr.color_ix = cvt->ico->color_ix);

	if (cvt->hfstr.bold != cvt->ico->bold)
		cprintf(ge, cvt,"   %s %s\n", tok[TBOLD],
			tst[cvt->hfstr.bold = cvt->ico->bold]);

	if (cvt->hfstr.italic != cvt->ico->italic)
		cprintf(ge, cvt,"   %s %s\n", tok[TITALICS],
			tst[cvt->hfstr.italic = cvt->ico->italic]);

	if (cvt->hfstr.strike_thru != cvt->ico->strike_thru)
		cprintf(ge, cvt,"   %s %s\n", tok[TTHRU],
			tst[cvt->hfstr.strike_thru = cvt->ico->strike_thru]);

	if (cvt->hfstr.underline != cvt->ico->underline)
		cprintf(ge, cvt,"   %s %d\n", tok[TUNDERLINE],
			cvt->hfstr.underline = cvt->ico->underline);

	if (cvt->hfstr.underline_words_only != cvt->ico->underline_words_only)
		cprintf(ge, cvt,"   %s %s\n", tok[TWORDSONLY],
			tst[cvt->hfstr.underline_words_only = cvt->ico->underline_words_only]);
}

/****************************************************************************
 *
 * module:  a g 3 1 O u t W i n d o w S i z e
 */
ag31OutWindowSize(ge,cvt,t)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
{
	int wid, hyt;

	/* set the window width */
	wid = ge->state.drawn_area.x2 - ge->state.drawn_area.x1;
	if ( wid < (5.5 * MILS_P_INCH) )
		wid = 5.5 * MILS_P_INCH;

	/* set the window height */
	hyt = ge->state.drawn_area.y2 - ge->state.drawn_area.y1;
	if ( hyt == 0 )
		hyt = 7.5 * MILS_P_INCH;
	else if ( hyt < (5.9 * MILS_P_INCH) )
		hyt = 5.9 * MILS_P_INCH;

	cprintf(ge, cvt," %s %d %d\n", tok[TDOCDIMENSIONS], wid, hyt);
}

/****************************************************************************
 *
 * module:	a g 3 1 O u t V e r s i o n
 */
ag31OutVersion(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int ii;

	cprintf(ge, cvt,"*START GRAPHICS VERSION=311 ENCODING=7BIT\n\n");
	cvt->info.encoding = 1;
	return;	     	
}



/****************************************************************************
 *
 * module:	a g 3 1 O u t H o o k s
 */
ag31OutHooks(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int i;

	for ( i=0; i<ge->n_hooks; i++ )
	  cprintf(ge, cvt,"%s %s %s\n", ATOK_HOOK,ge->hook[i].type,ge->hook[i].name);
}
/****************************************************************************
 *
 * module:	a g 3 1 O u t E n d e r
 */
ag31OutEnder(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{

		AGetDefaultFileInfo(&cvt->info);
		cvt->info.docType = AX_DOC_TYPE_ASTERX_GRAPHICS;
		AFileEnderFromInfo(&cvt->info,cvt->buf);
		cprintf(ge, cvt,"%s\n",cvt->buf);

}

/****************************************************************************
 *
 * module:	a g 3 1 O u t C h a r t I n f o
 */
ag31OutChart(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int ii;
	chart_type *chart;

	if (ge->n_charts)
	{
		cprintf(ge, cvt,"%s ", tok[TCHART]);
		for (ii=0; ii<ge->n_charts; ii++)
		{
			chart = (chart_type*)&ge->chart[ii];
			if (!chart)
				continue;
			ag31OutChartData(ge,cvt,chart);
		}
		cprintf(ge, cvt,"%s %s\n\n", tok[TEND], tok[TCHART]);
	}
}

/***************************************************************************
 *
 * module:	a g 3 1 O u t C a l l b a c k s
 */
ag31OutCallbacks(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt; 
{
	int ii,jj;
	ge_callback_type *cback;
	char *arg;
	int len;
	tx_char tbuff[400];

	if (ge->n_callbacks)
	{
		cprintf(ge, cvt,"%s", tok[TCBACK]);
		for (ii=0; ii<ge->n_callbacks; ii++)
		{
			cback = &ge->callback[ii];
			cprintf(ge, cvt,"\n %s ",tok[TMACRO]);
			len = MultiToWideStr(ge, cback->name, tbuff );
			ag31OutString(ge, len, tbuff, cvt, FALSE );
			cprintf(ge, cvt," %d %d", TRUE, TRUE);

			for (jj=0; jj<cback->n_args; jj++)
			{
				/* check for wrapping */
				cputc(ge,cvt,' ');
				len = MultiToWideStr(ge,cback->args[jj], tbuff );
				ag31OutString(ge, len, tbuff, cvt, FALSE );
			}
		}
		cprintf(ge, cvt,"\n%s %s\n\n", tok[TEND], tok[TCBACK]);
	}
}


/************************************************************************
 *
 * module:	a g 3 1 M a p T o 1 2 5
 */
ag31MapTo125(ge,cvt, src_t, dst_t)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *src_t, *dst_t;
{
	unsigned char *srcP, *dstP, *outCmap;
	int size, ii, cmap_ix;
	unsigned int r,g,b;
	cmap_type *srcCmap;

	if ( src_t->type == IMAGE && dst_t->type == IMAGE )
	{
		srcP = src_t->data.img->info.data;
		dstP = dst_t->data.img->info.data;
		cmap_ix = src_t->data.img->cmapID;
		srcCmap = &ge->cmap[cmap_ix];

		/* start by making a map from the input colormap */
		outCmap = (unsigned char*)malloc(srcCmap->n_colors);
		AxSet125PixVal(NULL, FALSE, NULL, NULL, NULL, 16, TRUE);
		for (ii=srcCmap->n_colors; ii>=0; --ii)
		{
			if (srcCmap->color[ii].see_thru)
				outCmap[ii] = 124; /* set to white */
			else
			{
				ExtractRGBCmapEntry(ge, cvt, cmap_ix, ii, &r, &g, &b);
				AxSet125PixVal(outCmap+ii,FALSE,r,g,b,0,FALSE);
			}
		}
		size = src_t->data.img->info.hyt * src_t->data.img->info.data_byte_wid;
		for ( ii=size; ii>=0; --ii)
			dstP[ii] = outCmap[srcP[ii]];
		free(outCmap);
	}
}


/************************************************************************
 *
 * module:	a g 3 1 O u t I m a g e 
 */
ag31OutImage(ge, t, cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	char *epsP;
	int llen, ii, (*encodeFunc)(), (*compressFunc)();


	ag31OutPath(ge, t->paths, t->n_paths, cvt);
	ag31OutTrans(ge, &t->data.img->trans, cvt);

	/* if there is an EPS image, write it out. */
	if (t->data.img->info.eps_seg != NULL)
	{
		/* the width, height, size and a leading space */
		cprintf(ge, cvt," %s %d %d %d\n ", tok[TEPS],
			t->data.img->info.wid,
			t->data.img->info.hyt,
			t->data.img->info.eps_seg_size);

		epsP = t->data.img->info.eps_seg;
		while ( *epsP != '\0')
		{
			cputc(ge,cvt, *epsP);
			if (*epsP == '\n')
				cputc(ge,cvt, ' ');
			epsP++;
		}

		/* the end of the EPS segment */
		cprintf(ge, cvt,"%s %s\n",tok[TEND], tok[TEPS]);
	}

	cprintf(ge, cvt," %s %d", tok[TWIDTH], t->data.img->info.wid);
	cprintf(ge, cvt," %s %d", tok[THEIGHT], t->data.img->info.hyt);
	cprintf(ge, cvt," %s %d", tok[TDEPTH], t->data.img->info.dep);


	if ( cvt->compression != COMP_LZW)
		cprintf(ge, cvt," %s %s", tok[TCOMPRESSION], cmp[cvt->compression-1]);

	if ((cvt->encoding != BIT6) )
		cprintf(ge, cvt," %s %s", tok[TENCODING], enc[cvt->encoding]);
	
	encodeFunc = (cvt->encoding == BIT6) ? mailOutBit6 : mailOutHex;

	if (cvt->compression == COMP_RAW)
		compressFunc = ag31writeRAW;
	else if (cvt->compression == COMP_RUN)
		compressFunc = ag31writeRUN;
	else
		compressFunc = ReadWriteLZW;

	llen = cprintf(ge, cvt,"\n %s ", tok[TDATA]);

	/* DO IT! */
	(*compressFunc)(ge, cvt,llen,t->data.img->info.hyt,
		t->data.img->info.data_byte_wid,
		t->data.img->info.data,encodeFunc, FALSE);
	if ((t->data.img->info.dep == 8) && (t->data.img->info.mask))
	{
		llen = cprintf(ge, cvt,"\n%s ", tok[TMASK]);
		(*compressFunc)(ge, cvt,llen,t->data.img->info.hyt,
			t->data.img->info.mask_byte_wid,
			t->data.img->info.mask,encodeFunc,FALSE);
	}

	cputc(ge,cvt,'\n');
	return;
}

/****************************************************************************
 *
 * module:	W r i t e R A W - Write data in Raw format
 */
ag31writeRAW(ge, cvt,llen,hyt,wid,imgP,func,mode)
graphics_main_type *ge;
cvt_global_type *cvt;
int llen,hyt, wid;
unsigned char *imgP;
int (*func)(), mode;
{
	int cwid;

	while (0 <= --hyt)
	{
		cwid = wid;
		while (0 <= --cwid)
		{
			llen += (*func)(*imgP++ & 0xFF, llen, cvt);
			if (MAX_LINE < llen)
			{
				cputc(ge,cvt,'\n');
				cputc(ge,cvt,' ');
				llen = 0;
			}
		}
		llen += (*func)(-1, llen, cvt);
	}
}

/****************************************************************************
 *
 * module:	p a c k R U N - Pack a scanline of `snum' bytes into a buffer.
 */
ag31packRUN(ge, inP,outP,inWid,outWid,buffer)
graphics_main_type *ge;
char *inP, *outP;
int inWid,*outWid;
char *buffer; /* temp space */
{
	register char *buf;
	register int nbuf, rstart;
	int i, mode;
	char c, lastc, *startP;
	
	startP = outP;
	buf = buffer + 1;
	buf[0] = lastc = c = *inP++;
	nbuf = 1;
	rstart = 0;
	mode = DMP;

	for (i=1; i<inWid; i++)
	{
		buf[nbuf++] = c = *inP++;

		switch (mode) {

		case DMP:
			if (MAXDMP < nbuf)
			{
				buffer[0] = nbuf - 2;
				memcpy(outP,buffer,nbuf);
				outP += nbuf;
				/* check for outP overflow */
				buf[0] = c;
				nbuf = 1;
				rstart = 0;
				break;
			}

			if (c != lastc)
			{
				rstart = nbuf - 1;	/* 1st of poss run */
				break;
			}

			if (rstart && nbuf - rstart < MINRUN)
				break;

			if (0 < rstart)
			{
				buffer[0] = rstart - 1;
				memcpy(outP,buffer,rstart + 1);
				outP += (rstart + 1);
				/* check for outP overflow */
			}

			mode = RUN;
			break;

		case RUN:
			if (c == lastc && nbuf - rstart <= MAXRUN)
				break;

			*outP++ = rstart - nbuf + 2;
			*outP++ = lastc;
			/* check for outP overflow */
			buf[0] = c;
			nbuf = 1;
			rstart = 0;
			mode = DMP;
			break;
		}

		lastc = c;
	}

	if (mode == DMP)
	{
		buffer[0] = nbuf - 1;
		memcpy(outP,buffer,nbuf + 1);
		outP += (nbuf + 1);
		/* check for outP overflow */
	}
	else
	{
		*outP++ = rstart - nbuf + 1;
		*outP++ = lastc;
		/* check for outP overflow */
	}

	*outWid = (int)(outP - startP);
	return inWid;
}





/****************************************************************************
 *
 * module:	W r i t e R U N - Write data in RUN LENGTH ENCODED format
 */
ag31writeRUN(ge,cvt,llen,hyt,wid,imgP,func,mode)
graphics_main_type *ge;
cvt_global_type *cvt;
int llen,hyt, wid;
unsigned char *imgP;
int (*func)(), mode;
{
	int cmpWid;
	unsigned char *outP, *pP, *tmpBuf;

	outP = (unsigned char*)calloc(2,wid); /* worst case */
	tmpBuf = (unsigned char*)calloc(1,260); /* max */

	while (0 <= --hyt)
	{
		if (ag31packRUN(imgP,outP,wid,&cmpWid,tmpBuf) != wid)
			return(!AOK);
		imgP += wid;
		pP = outP;
		while (0 <= --cmpWid)
		{
			llen += (*func)(*pP++, llen, cvt);
			if (MAX_LINE < llen)
			{
				cputc(ge,cvt,'\n');
				cputc(ge,cvt,' ');
				llen = 0;
			}
		}
		llen += (*func)(-1, llen, cvt);
	}

	free(outP);
	free(tmpBuf);
	return(AOK);
}


/****************************************************************************
 *
 * module:  ag31OutChartData
 */
ag31OutChartData(ge,cvt,chart)
graphics_main_type *ge;
cvt_global_type *cvt;
register chart_type *chart;
{
	tx_char tbuff[500];
	register int ii;
	int len;

	/* sanity check */
	if ( !chart )
		return;

	/* name of chart */
	len = MultiToWideStr(ge,chart->name, tbuff );
	ag31OutString(ge, len, tbuff, cvt, FALSE );
	cprintf(ge, cvt,"\n");

	/* chart type  */
	cprintf(ge, cvt," %s ", tok[TTYPE]);
	len = MultiToWideStr(ge,chrtypes[chart->type],  tbuff );
	ag31OutString(ge, len, tbuff, cvt, FALSE );

	/* chart orientation */
	if ( chart->orientation ) 
		cprintf(ge, cvt," %s\n", tok[TVERT]);
	else 
		cprintf(ge, cvt," %s\n", tok[THORZ]);

	/* title, subtitle, and footer strings */
	if ( chart->title )
	{
		cprintf(ge, cvt," %s ", tok[TTITLE]);
		len = MultiToWideStr(ge,chart->title, tbuff );
		ag31OutString(ge, len, tbuff, cvt, FALSE );
		cprintf(ge, cvt,"\n");
	}
	if ( chart->subtitle )
	{
		cprintf(ge, cvt," %s ", tok[TSUBTITLE]);
		len = MultiToWideStr( ge,chart->subtitle, tbuff );
		ag31OutString(ge, len, tbuff, cvt, FALSE );
		cprintf(ge, cvt,"\n");
	}
	if ( chart->footer )
	{
	  	cprintf(ge, cvt," %s ", tok[TFOOT]);
		len = MultiToWideStr(ge, chart->footer, tbuff );
		ag31OutString(ge, len, tbuff, cvt, FALSE );
		cprintf(ge, cvt,"\n");
	}
	cprintf(ge, cvt," %s %d %d %d\n", tok[TTITLE_ALIGN],
	                              chart->title_align,
	                              chart->subtitle_align,
	                              chart->footer_align);

	/* chart decorations (closing lines and grid lines) */
	cprintf(ge, cvt," %s %d %d %d %d\n", tok[TCLOSE], chart->close_top,
	                                              chart->close_bot,
	                                              chart->close_lft,
	                                              chart->close_rt);
	cprintf(ge, cvt," %s %d %d %d %d\n", tok[TGRIDS], chart->maj_x_grids,
	                                              chart->min_x_grids,
	                                              chart->maj_y_grids,
	                                              chart->min_y_grids);

	cprintf(ge, cvt," %s %d %d\n", tok[TOFFSET], chart->def_x_offset,
	                                         chart->def_y_offset);

	/* chart size and margins */
	cprintf(ge, cvt," %s %d %d %d %d\n", tok[TEXTENT], chart->extent.x1,
	                                               chart->extent.y1,
	                                               chart->extent.x2,
	                                               chart->extent.y2);
	cprintf(ge, cvt," %s %d %d %d %d\n", tok[TMARGIN], chart->margin.x1,
	                                               chart->margin.y1,
	                                               chart->margin.x2,
	                                               chart->margin.y2);
 
	/* count of original axes and groups (as of last rendering) */
	cprintf(ge, cvt," %s %d %d %d %d %d\n", tok[TCHART_COUNTS],
	                                 chart->old_n_x_axes,
	                                 chart->old_n_y_axes,
	                                 chart->old_n_z_axes,
	                                 chart->old_n_groups,
	                                 chart->old_type);

	/* default null treatment */
	switch(chart->null_format)
	{
		default:
		case 0: cprintf(ge, cvt," %s %s\n", tok[TNULL_FMT], tok[TSUBS]); break;
		case 1: cprintf(ge, cvt," %s %s\n", tok[TNULL_FMT], tok[TSPAN]); break;
		case 2: cprintf(ge, cvt," %s %s\n", tok[TNULL_FMT], tok[TGAP]); break;
	}
	
	/* type specific info */
	cprintf(ge, cvt," %s %d\n",    tok[TPIE_INFO], chart->info.start_angle);
	cprintf(ge, cvt," %s %f %f\n", tok[TBAR_INFO], chart->info.bar_margin,
	                                           chart->info.bar_overlap);
	cvt->linepos=0;

	/* the axes */
	if (chart->n_x_axes)
	{
		for (ii=0; ii<chart->n_x_axes; ii++)
		{
			cprintf(ge, cvt," %s ", tok[TX_AXIS]);
			if ( !chart->x_axis[ii].label_str )
				cprintf(ge, cvt,"\"\"");
			else
			{
				len = MultiToWideStr(ge,chart->x_axis[ii].label_str, tbuff );
				ag31OutString(ge, len, tbuff, cvt, FALSE );
			}
			cprintf(ge, cvt, "\n");
			ag31OutChartAxisType(ge,cvt,&chart->x_axis[ii]);
			cprintf(ge, cvt," %s %s\n", tok[TEND], tok[TX_AXIS]);
		}
	}
	if (chart->n_y_axes)
	{
		for (ii=0; ii<chart->n_y_axes; ii++)
		{
			cprintf(ge, cvt," %s ", tok[TY_AXIS]);
			if ( !chart->y_axis[ii].label_str )
				cprintf(ge, cvt,"\"\"");
			else
			{
			  len = MultiToWideStr(ge,chart->y_axis[ii].label_str, tbuff );
			  ag31OutString(ge, len, tbuff, cvt, FALSE );
			}
			cprintf(ge, cvt, "\n");
			ag31OutChartAxisType(ge,cvt,&chart->y_axis[ii]);
			cprintf(ge, cvt," %s %s\n", tok[TEND], tok[TY_AXIS]);
		}
	}

	cvt->linepos=0;

	/* group data */
	for (ii=0; ii<chart->n_groups; ii++)
		ag31OutChartDataGroupType(ge,cvt,&chart->grp[ii]);
 
	/* legend box */
	ag31OutChartLegendType(ge,cvt,&chart->legend);
 
	/* write the default attributes */
	cvt->linepos = cprintf(ge, cvt,"%s ", tok[TDEFATR]);
	if (AttrCmp(ge, &chart->def_attr, &cvt->curr_atts))
		ag31OutAttrs(ge, &chart->def_attr, 0, cvt);
	else
		cputc(ge,cvt,'\n');

	cvt->linepos = cprintf(ge, cvt,"%s ", tok[TMAJHRZ]);
	if (AttrCmp(ge, &chart->maj_x_attr, &cvt->curr_atts))
		ag31OutAttrs(ge, &chart->maj_x_attr, 0, cvt);
	else
		cputc(ge,cvt,'\n');

	cvt->linepos = cprintf(ge, cvt,"%s ", tok[TMINHRZ]);
	if (AttrCmp(ge, &chart->min_x_attr, &cvt->curr_atts))
		ag31OutAttrs(ge, &chart->min_x_attr, 0, cvt);
	else
		cputc(ge,cvt,'\n');

	cvt->linepos = cprintf(ge, cvt,"%s ", tok[TMAJVRT]);
	if (AttrCmp(ge, &chart->maj_y_attr, &cvt->curr_atts))
		ag31OutAttrs(ge, &chart->maj_y_attr, 0, cvt);
	else
		cputc(ge,cvt,'\n');

	cvt->linepos = cprintf(ge, cvt,"%s ", tok[TMINVRT]);
	if (AttrCmp(ge, &chart->min_y_attr, &cvt->curr_atts))
		ag31OutAttrs(ge, &chart->min_y_attr, 0, cvt);
	else
		cputc(ge,cvt,'\n');

	cvt->linepos = cprintf(ge, cvt,"%s ", tok[TTITLEATR]);
	if (AttrCmp(ge, &chart->title_attr, &cvt->curr_atts))
		ag31OutAttrs(ge, &chart->title_attr, 0, cvt);
	else
		cputc(ge,cvt,'\n');

	cvt->linepos = cprintf(ge, cvt,"%s ", tok[TSUBTITLEATR]);
	if (AttrCmp(ge, &chart->subtitle_attr, &cvt->curr_atts))
		ag31OutAttrs(ge, &chart->subtitle_attr, 0, cvt);
	else
		cputc(ge,cvt,'\n');

	cvt->linepos = cprintf(ge, cvt,"%s ", tok[TFOOTATR]);
	if (AttrCmp(ge, &chart->footer_attr, &cvt->curr_atts))
		ag31OutAttrs(ge, &chart->footer_attr, 0, cvt);
	else
		cputc(ge,cvt,'\n');
}

ag31OutNumberFormat(ge,cvt,fmt)
graphics_main_type *ge;
cvt_global_type *cvt;
number_format *fmt;
{
    tx_char tbuff[500];
	int ii, len;

	/* the number format token */
	cprintf(ge, cvt,"  %s ", tok[TNUM_FMT]);
	cprintf(ge, cvt,"%d ",  fmt->id);
	cprintf(ge, cvt,"%d ",  fmt->aux_id);
	cprintf(ge, cvt,"%d ",  fmt->display_format_errors);
	cprintf(ge, cvt,"%d ",  fmt->trim_string);
	cprintf(ge, cvt,"%d ",  fmt->units);
	cprintf(ge, cvt,"%d", fmt->n_radix_places);
	cprintf(ge, cvt,"\n   ");

	/* the prefix and suffix strings */
	if ( !fmt->prefix )
		cprintf(ge, cvt,"\"\" ");
	else
	{
		len = MultiToWideStr(ge,fmt->prefix, tbuff );
		mlOutString(ge,len, tbuff, cvt, FALSE);
		cprintf(ge, cvt," ");
	}
	if ( !fmt->suffix )
		cprintf(ge, cvt,"\"\"");
	else
	{
		len = MultiToWideStr(ge,fmt->suffix, tbuff );
		mlOutString(ge,len, tbuff, cvt, FALSE);
	}
	cprintf(ge, cvt,"\n   ");

	/* am and pm strings */
	if ( !fmt->am )
		cprintf(ge, cvt,"\"\" ");
	else
	{
		len = MultiToWideStr(ge,fmt->am, tbuff );
		mlOutString(ge,len, tbuff, cvt, FALSE);
		cprintf(ge, cvt," ");
	}
	if ( !fmt->pm )
		cprintf(ge, cvt,"\"\"");
	else
	{
		len = MultiToWideStr(ge,fmt->pm, tbuff );
		mlOutString(ge,len, tbuff, cvt, FALSE);
	}
	cprintf(ge, cvt,"\n   ");

	/* the boolean strings */
	if ( !fmt->true )
		cprintf(ge, cvt,"\"\" ");
	else
	{
		len = MultiToWideStr(ge,fmt->true, tbuff );
		mlOutString(ge,len, tbuff, cvt, FALSE);
		cprintf(ge, cvt," ");
	}
	if ( !fmt->false )
		cprintf(ge, cvt,"\"\"");
	else
	{
		len = MultiToWideStr(ge,fmt->false, tbuff );
		mlOutString(ge,len, tbuff, cvt, FALSE);
	}
	cprintf(ge, cvt,"\n   ");

	/* the separator characters */
	len = 1;
	tbuff[1] = (tx_char)0;
	tbuff[0] = (tx_char)fmt->t_sep;
	mlOutString(ge, len, tbuff, cvt, FALSE);
	cprintf(ge, cvt," ");
	tbuff[0] = (tx_char)fmt->d_sep;
	mlOutString(ge, len, tbuff, cvt, FALSE);
	cprintf(ge, cvt,"\n");
}

/****************************************************************************
 *
 * module:  ag31OutChartAxisType
 */
ag31OutChartAxisType(ge,cvt,axis)
graphics_main_type *ge;
cvt_global_type *cvt;
chart_axis_type *axis;
{
    tx_char tbuff[500];
	int ii, len;

	/* type and space consumed (for additional axes) */
	cprintf(ge, cvt,"  ");
	cprintf(ge, cvt,"%d ",  axis->type);
	cprintf(ge, cvt,"%d ",  axis->x_offset);
	cprintf(ge, cvt,"%d ",  axis->y_offset);
	cprintf(ge, cvt,"%d\n", axis->tik_margin);

	/* tik and label alignment */
	cprintf(ge, cvt,"  ");
	cprintf(ge, cvt,"%d ",  axis->label_align);
	cprintf(ge, cvt,"%d\n", axis->tik_align);

	/* axis format info */
	cprintf(ge, cvt,"  ");
	cprintf(ge, cvt,"%d ",  axis->hidden);
	cprintf(ge, cvt,"%d ",  axis->auto_max);
	cprintf(ge, cvt,"%d ",  axis->auto_min);
	cprintf(ge, cvt,"%d ",  axis->auto_bas);
	cprintf(ge, cvt,"%d ",  axis->auto_n_major_tiks);
	cprintf(ge, cvt,"%d ",  axis->auto_n_minor_tiks);
	cprintf(ge, cvt,"%d ",  axis->major_tik_format);
	cprintf(ge, cvt,"%d ",  axis->minor_tik_format);
	cprintf(ge, cvt,"%d ",  axis->pos);
	cprintf(ge, cvt,"%f ",  axis->pos_val);
	cprintf(ge, cvt,"%f ",  axis->min_val);
	cprintf(ge, cvt,"%f ",  axis->max_val);
	cprintf(ge, cvt,"%f ",  axis->bas_val);
	cprintf(ge, cvt,"%f\n", axis->first_delta);

	cprintf(ge, cvt,"  ");
	cprintf(ge, cvt,"%d %d %f %f\n", axis->major_sz,
	                             axis->n_maj_steps,
	                             axis->maj_delta,
	                             axis->maj_step);
	cprintf(ge, cvt,"  ");
	cprintf(ge, cvt,"%d %d %f %f\n", axis->minor_sz,
	                             axis->n_min_steps,
	                             axis->min_delta,
	                             axis->min_step);

	/* numerical format for tiks */
	ag31OutNumberFormat(ge,cvt,&axis->tik_format);

	/* bar info */
	cprintf(ge, cvt,"  %s %f %f\n", tok[TBAR_INFO],
	        axis->bar_margin,
	        axis->bar_overlap);

	/* tik labels (for catagory axes) */
	for ( ii=0; ii<axis->n; ii++ )
	{
		cprintf(ge, cvt, "  %s ", tok[TLABEL]);
		if ( !axis->tik_str[ii] )
			cprintf(ge, cvt,"\"\"");
		else
		{
			len = MultiToWideStr(ge,axis->tik_str[ii], tbuff);
			ag31OutString(ge,len, tbuff, cvt, FALSE);
		}
		cprintf(ge, cvt,"\n");
	}
 
	/* axis attribute info */
	cvt->linepos = cprintf(ge, cvt,"  %s ", tok[TAXATTR]);
	if (AttrCmp(ge, &axis->attr, &cvt->curr_atts))
		ag31OutAttrs(ge, &axis->attr, 0, cvt);
 
	cvt->linepos = cprintf(ge, cvt,"  %s ", tok[TAXLABATTR]);
	if (AttrCmp(ge, &axis->label_attr, &cvt->curr_atts))
		ag31OutAttrs(ge, &axis->label_attr, 0, cvt);
 
	cvt->linepos = cprintf(ge, cvt,"  %s ", tok[TAXTIKLABATTR]);
	if (AttrCmp(ge, &axis->tik_label_attr, &cvt->curr_atts))
		ag31OutAttrs(ge, &axis->tik_label_attr, 0, cvt);
}

/****************************************************************************
 *
 * module:  ag31OutChartDataGroupType
 */
ag31OutChartDataGroupType(ge,cvt,grp)
graphics_main_type *ge;
cvt_global_type *cvt;
chart_data_group_type *grp;
{
	tx_char tbuff[500];
	int len, ii;

	/* the beginning of the data group and the group name */
	cprintf(ge, cvt," %s ", tok[TDATGROUP]);
	len = MultiToWideStr(ge,grp->name, tbuff);
	ag31OutString(ge,len, tbuff, cvt, FALSE);
	cprintf(ge, cvt, " %s", grp->disabled ? tst[0] : tst[1]);
	cprintf(ge, cvt,"\n");

	/* the legend string */
	cprintf(ge, cvt,"  %s ", tok[TLABEL]);
	if ( !grp->legend_str )
		cprintf(ge, cvt,"\"\"");
	else
	{
		len = MultiToWideStr(ge,grp->legend_str, tbuff);
		ag31OutString(ge,len, tbuff, cvt, FALSE);
	}
	cprintf(ge, cvt,"\n");

	/* data type (line, bar, etc...), stack id, and custom string */
	if( grp->type == BG_STACKED_STRATA )
		grp->type = BG_STRATA;
	cprintf(ge, cvt,"  %s %d %d ", tok[TDATATYPE], grp->type, grp->stack_id);
	if ( !grp->part )
		cprintf(ge, cvt,"\"\"");
	else
	{
		len = MultiToWideStr(ge,grp->part, tbuff);
		ag31OutString(ge,len, tbuff, cvt, FALSE);
	}
	cprintf(ge, cvt,"\n");

	/* the axes this data is drawn against */
	cprintf(ge, cvt,"  %s %d %d %d\n", tok[TAXES], grp->x_axis,
	                                           grp->y_axis,
	                                           grp->z_axis);

	/* data label info */
	ag31OutNumberFormat(ge,cvt,&grp->label_frmt);
	cprintf(ge, cvt, "   %d %d %d %d\n", grp->data_label_type,
	                                 grp->label_x_off,
	                                 grp->label_y_off,
	                                 grp->label_align);

	/* write each data point, calculating grp->n at read time */
	for (ii=0; ii<grp->n; ii++)
		ag31OutChartDataType(ge,cvt,&grp->dat[ii]);

	/* write the data label and data attributes */
	cvt->linepos = cprintf(ge, cvt,"  %s ", tok[TGRPLABATTR]);
	if (AttrCmp(ge, &grp->label_attr, &cvt->curr_atts))
		ag31OutAttrs(ge, &grp->label_attr, 0, cvt);
	cvt->linepos = cprintf(ge, cvt,"  %s ", tok[TGRPATTR]);
	if (AttrCmp(ge, &grp->attr, &cvt->curr_atts))
		ag31OutAttrs(ge, &grp->attr, 0, cvt);

	/* the end of the data group */
	cprintf(ge, cvt," %s %s\n", tok[TEND], tok[TDATGROUP]);
}

/****************************************************************************
 *
 * module:  ag31OutChartDataType
 */
ag31OutChartDataType(ge,cvt,dat)
graphics_main_type *ge;
cvt_global_type *cvt;
chart_data_point_type *dat;
{
	tx_char tbuff[500];
	int len, ii;

	/* the data point token and the display type */
	cprintf(ge, cvt,"  %s ",tok[TDAT]);
	switch(dat->display_flag)
	{
		default:
		case BG_PLOT: cprintf(ge, cvt, "%s ", tok[TPLOT]);      break;
		case BG_SUBS: cprintf(ge, cvt, "%s ", tok[TSUBS]);      break;
		case BG_SPAN: cprintf(ge, cvt, "%s ", tok[TSPAN]);      break;
		case BG_GAP:  cprintf(ge, cvt, "%s ", tok[TGAP]);       break;
		case BG_NULL: cprintf(ge, cvt, "%s ", tok[TNULL_VAL]);  break;
	}

	/* the data point label */
	if (!dat->label_str)
		cprintf(ge, cvt,"\"\"");
	else
	{
		len = MultiToWideStr(ge, dat->label_str, tbuff );
		ag31OutString(ge, len, tbuff, cvt, FALSE );
	}
	if ( dat->local_x )
		cprintf(ge, cvt," %d", dat->label_x_off);
	else
		cprintf(ge, cvt," %s", tok[TX]);
	if ( dat->local_y )
		cprintf(ge, cvt," %d", dat->label_y_off);
	else
		cprintf(ge, cvt," %s", tok[TY]);
	cprintf(ge, cvt, "\n");

	/* data point x position and auxiliary data */
	cprintf(ge, cvt, "  ");
	for (ii=0; ii<dat->nx; ii++)
		cprintf(ge, cvt," %f",dat->x[ii]);
	cprintf(ge, cvt, " %s\n", tok[TX]);

	/* data point y position and auxiliary data */
	cprintf(ge, cvt, "  ");
	for (ii=0; ii<dat->ny; ii++)
		cprintf(ge, cvt," %f",dat->y[ii]);
	cprintf(ge, cvt, " %s\n", tok[TY]);

	/* data point z position and auxiliary data */
	cprintf(ge, cvt, "  ");
	for (ii=0; ii<dat->nz; ii++)
		cprintf(ge, cvt," %f",dat->z[ii]);
	cprintf(ge, cvt, " %s\n", tok[TZ]);
}

/****************************************************************************
 *
 * module:  ag31OutChartLegendType
 */
ag31OutChartLegendType (ge,cvt,legend)
graphics_main_type *ge;
cvt_global_type *cvt;
chart_legend_type *legend;
{
    tx_char tbuff[500];
    int len;

	if (!legend)
		return;

	/* the legend token and the optional legend title */
	cprintf(ge, cvt," %s ",tok[TLEGEND]);
	if (!legend->title_str)
		cprintf(ge, cvt,"\"\"");
	else
	{
	  len = MultiToWideStr(ge, legend->title_str, tbuff );
	  ag31OutString(ge, len, tbuff, cvt, FALSE );
	}
	cprintf(ge, cvt,"\n");

	/* the legend position */
	cprintf(ge, cvt,"  %d %d\n", legend->h_align, legend->h_off);
	cprintf(ge, cvt,"  %d %d\n", legend->v_align, legend->v_off);
	cprintf(ge, cvt,"  %d\n",    legend->title_align);

	/* the legend format */
	cprintf(ge, cvt,"  %d",  legend->disabled);
	cprintf(ge, cvt, " %d",  legend->x_space_eating);
	cprintf(ge, cvt, " %d",  legend->y_space_eating);
	cprintf(ge, cvt, " %d",  legend->text_before_sample);
	cprintf(ge, cvt, " %d",  legend->arranged_by_row);
	cprintf(ge, cvt, " %d\n",legend->max_per_row_or_column);

	/* legend margins */
	cprintf(ge, cvt,"  %d",   legend->t_margin);
	cprintf(ge, cvt, " %d",   legend->b_margin);
	cprintf(ge, cvt, " %d",   legend->l_margin);
	cprintf(ge, cvt, " %d",   legend->r_margin);
	cprintf(ge, cvt, " %d",   legend->row_margin);
	cprintf(ge, cvt, " %d ",  legend->column_margin);
	cprintf(ge, cvt, " %d\n", legend->auto_margins);

	/* the attributes of the various parts of the legend */
	cvt->linepos = cprintf(ge, cvt,"  %s ",tok[TLEGBOXATTR]);
	if (AttrCmp(ge, &legend->box_attr, &cvt->curr_atts))
		ag31OutAttrs(ge, &legend->box_attr, 0, cvt);
	else
		cputc(ge,cvt,'\n');

	cvt->linepos = cprintf(ge, cvt,"  %s ",tok[TLEGLABATTR]);
	if (AttrCmp(ge, &legend->label_attr, &cvt->curr_atts))
		ag31OutAttrs(ge, &legend->label_attr, 0, cvt);
	else
		cputc(ge,cvt,'\n');

	cvt->linepos = cprintf(ge, cvt,"  %s ",tok[TLEGTITATTR]);
	if (AttrCmp(ge, &legend->title_attr, &cvt->curr_atts))
		ag31OutAttrs(ge, &legend->title_attr, 0, cvt);
	else
		cputc(ge,cvt,'\n');

	/* done with legends */
	cprintf(ge, cvt," %s %s\n",tok[TEND],tok[TLEGEND]);
	cvt->linepos = 0;
}

/****************************************************************************
 *
 * module:	G r a p h i c s W r i t e
 */
ag31Write(ge, cvt, t, outBuffer, outCount)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
char **outBuffer;
int *outCount;
{

	mlDumpTxtObjs(ge, t);

	if (outBuffer)
	{
		cvt->buffit = TRUE;
		cvt->outBuff = cvt->outBufPtr = (char*)malloc(CVTBLKSIZE);
		cvt->outBufEndPtr = cvt->outBufPtr + CVTBLKSIZE;
	}
	else
	{
		/* check for succesful open of the file */
		if (!(cvt->fp = (FILE *)cvtGetFP(ge,cvt->fileName, "w+")) )
			return(!AOK);
	}

	InitTokens(ge);

	/* *START GRAPHICS and version string */
	ag31OutVersion(ge, cvt);

	/* user definable fields */
	ag31OutHooks(ge,cvt);
	ag31OutColormap(ge, cvt);
	ag31OutFaces(ge, cvt);
	ag31OutLayers(ge, cvt);
	ag31OutNamedFills(ge, cvt);
	ag31OutCallbacks(ge, cvt);
	ag31OutHdrsFtrs(ge, cvt);

	/* default state */
	ag31OutSessionInfo(ge, cvt, t);

	/* charts */
	ag31OutChart(ge, cvt);

	/* user defined parts */
	ag31OutPalet(ge, cvt);

	/* the paint list */
	ag31OutPict(ge, t, cvt);

	/* finish up with *END GRAPHICS */
	ag31OutEnder(ge, cvt);

	if (outBuffer)
	{
		*outBuffer = cvt->outBuff;
		*outCount = cvt->outBufPtr - cvt->outBuff;
	}
	else
		cvtCloseFile(ge,cvt);

	return(AOK);
}

