/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: ag400.c
   Applix ASC version: 11
   Date: Mon Nov 16 13:58:52 1998
*/
#include <aos.h>
#include <atm.h>
#include <atmwin.h>
#include <axtext.h>
#include <atxint.h>
#include <img.h>
#include <ge.h>
#include <n_style.h>
#include <bg.h>
#include <afileimg.h>
#include <cvttoks.h>
#include <cvt.h>

#if FOR_FILTER
#	include <stiorts.h>
#else
#	include <fsrts.h>
#	define FOR_FILTER 0
#endif

#define NOHAMMER 	0
#define HAMMER 		1

#define RUN 0
#define DMP 1
#define MINRUN  3
#define MAXRUN  128
#define MAXDMP  128


int mailOutBit6();
int mailOutHex();
static void mlOutPart();
static void mlOutImage();
static int WriteRAW();
static int WriteRUN();
int ReadWriteLZW();

static int
mlOutIndent(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	if ( cvt->linepos != 0 )
	{
		cprintf(ge, cvt, "\n");
		cvt->linepos = 0;
	}
	switch ( cvt->indent )
	{
		default:
		case 9: cprintf(ge, cvt, "         "); return(9);
		case 8: cprintf(ge, cvt, "        ");  return(8);
		case 7: cprintf(ge, cvt, "       ");   return(7);
		case 6: cprintf(ge, cvt, "      ");    return(6);
		case 5: cprintf(ge, cvt, "     ");     return(5);
		case 4: cprintf(ge, cvt, "    ");      return(4);
		case 3: cprintf(ge, cvt, "   ");       return(3);
		case 2: cprintf(ge, cvt, "  ");        return(2);
		case 1: cprintf(ge, cvt, " ");         return(1);
		case 0:                            return(0);
	}
}

static void
mlOutNewline(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	/* add a newline if we need one */
	if ( MAX_LINE < cvt->linepos )
	{
		cprintf(ge, cvt, "\n");
		cvt->linepos = 0;
	}
}

/****************************************************************************
 *
 * module:	CvtOutSpecialChars - handle the specials
 */
static int
CvtOutSpecialChars(ge, inBuffer,outBuffer,len,mode)
graphics_main_type *ge;
tx_char *inBuffer, *outBuffer;
int len, mode; 
{
	int inlen, count=0;
	tx_char *inP, *outP;

	if ((mode == HAMMER) && (outBuffer == NULL))
		return;

	inP = inBuffer;
	outP = outBuffer;
	inlen = len;
	while(len--)
	{
		if ((*inP == '\\') || (*inP == '"'))
		{
			count++;
			if (mode == HAMMER)
			{
				*outP++ = '\\';
				*outP++ = *inP;
			}
		}
		else if (*inP == '\n')
		{
			count++;
			if (mode == HAMMER)
			{
				*outP++ = '\\';
				*outP++ = 'n';
			}
		}
		else if (mode == HAMMER)
			*outP++ = *inP;

		inP++;
	}
	return(inlen+count);
}

/****************************************************************************
 *
 * module:	m l O u t S t r i n g
 */
static int
mlOutString(ge, len, str, cvt, suppress_quote)
graphics_main_type *ge;
int len;
tx_char *str;
cvt_global_type *cvt;
int suppress_quote;
{
	int i,llen,size;
	mbchar *strP, *outBuf;
	tx_char *inbufP;

	/* count the characters first */
	llen = CvtOutSpecialChars(ge, str, NULL, len, NOHAMMER);

	/* str must be null terminated for NormalizeOut16, so  */
	/* we MUST make a local copy since we cannot count on  */
	/* the necessary tx_char being available in str.       */
	size = ((llen+1)*sizeof(tx_char));
	inbufP = (tx_char*)gloc(ge, size);

	/* now hammer the corrected chars in */
	CvtOutSpecialChars(ge, str, inbufP, len, HAMMER);
	inbufP[llen] = 0;

	/* Pump it thru the encoder (worst case: a char turns into 4 bytes) */
	size = (llen+1)*4;  
	outBuf = (mbchar *)gloc(ge, size);  /* worst case */
	len = NormalizeOut16(inbufP, outBuf, size, &cvt->info);

	/* open quote */
	if ( suppress_quote )
		llen = 0;
	else
	{
		llen = 1;
		cputc(ge, cvt, '\"');
	}

	/* write the string to the file (or buffer) */
	strP = outBuf;
	while(len--)
	{
		/* append the character to the output */
		cputc(ge, cvt,*strP++);
		llen++;

		/* check for wrap */
		if ( (MAX_LINE-1) < llen )
		{
			/* don't allow escapes to seperate */
			if ( strP[-1] == '\\' )
			{
				cputc(ge, cvt, *strP++);
				len--;
			}

			/* advance to the next line */
			cputc(ge, cvt,'\n');
			cputc(ge, cvt,' ');
			llen = 0;
		}
	}

	/* close quote */
	cputc(ge, cvt, '\"');
	gfree(ge, inbufP);
	gfree(ge, outBuf);
	return(llen);
}

static int
cvt_need_fill_attrs(ge,attr1, attr2)
graphics_main_type *ge;
thing_attributes *attr1;
thing_attributes *attr2;
{
	if ( attr1->fill_fg != attr2->fill_fg )
		return(TRUE);
	if ( attr1->fill_bg != attr2->fill_bg )
		return(TRUE);
	if ( attr1->fill_pat != attr2->fill_pat )
		return(TRUE);
	if ( attr1->fpat_type != attr2->fpat_type )
		return(TRUE);
	return(FALSE);
}

static int
cvt_need_line_attrs(ge, attr1, attr2)
graphics_main_type *ge;
thing_attributes *attr1;
thing_attributes *attr2;
{
	if ( attr1->line_fg != attr2->line_fg )
		return(TRUE);
	if ( attr1->line_bg != attr2->line_bg )
		return(TRUE);
	if ( attr1->line_pat != attr2->line_pat )
		return(TRUE);
	if ( attr1->lpat_type != attr2->lpat_type )
		return(TRUE);
	return(FALSE);
}

static int
cvt_need_shad_attrs(ge, attr1, attr2)
graphics_main_type *ge;
thing_attributes *attr1;
thing_attributes *attr2;
{
	if ( attr1->shadow_color != attr2->shadow_color )
		return(TRUE);
	if ( attr1->shadow != attr2->shadow )
		return(TRUE);
	if ( attr1->shadow_x != attr2->shadow_x )
		return(TRUE);
	if ( attr1->shadow_y != attr2->shadow_y )
		return(TRUE);
	return(FALSE);
}

static int
cvt_attr_indent(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	if ( cvt->linepos == 0 )
		cvt->linepos += mlOutIndent(ge, cvt);
	else
		cvt->linepos += cprintf(ge, cvt, " ");
}

/****************************************************************************
 *
 * module:	m l O u t A t t r s - Write an object's attrs to file
 */
static int
mlOutAttrs(ge, attr, cmap_ix, cvt)
graphics_main_type *ge;
thing_attributes *attr;
int cmap_ix;
cvt_global_type *cvt;
{
	/* write the style */
	if ( attr->style != cvt->curr_atts.style )
	{
		/* write the current indentation */
		mlOutIndent(ge, cvt);

		cprintf(ge, cvt,"%s %d\n", tok[TSTYS], attr->style);
		cvt->curr_atts.style = attr->style;
		cvt->linepos = 0;
	}

	/* write the colormap */
	if ( cmap_ix )
	{
		/* write the current indentation */
		mlOutIndent(ge, cvt);

		/* write the colormap id (images only) */
		cprintf(ge, cvt,"%s %d\n", tok[TCOLORMAP], cmap_ix);
		cvt->cmapID = cmap_ix;
		cvt->linepos = 0;
	}

	/* write the fill pattern and color info */
	if ( cvt_need_fill_attrs(ge, attr, &cvt->curr_atts) )
	{
		/* write the current indentation */
		mlOutIndent(ge, cvt);

		/* write the attribute */
		if ( attr->fpat_type == 6 )
		{
			switch ( ge->grad[attr->fill_pat].type )
			{
				case 0: attr->fpat_type = 2; break;
				case 1: attr->fpat_type = 3; break;
				case 2: attr->fpat_type = 4; break;
				case 3: attr->fpat_type = 2; break;
			}
		}
		cprintf(ge, cvt, "%s <%d %d %d %d %d %d %d>\n", tok[TBACKFILL],
		              attr->fill_fg,
		              attr->fill_bg,
		              attr->fpat_type,
		              attr->fill_pat,
		              0,
		              0,
		              0);

		/* update the current default state */
		cvt->curr_atts.fill_fg   = attr->fill_fg;
		cvt->curr_atts.fill_bg   = attr->fill_bg;
		cvt->curr_atts.fpat_type = attr->fpat_type;
		cvt->curr_atts.fill_pat  = attr->fill_pat;
		cvt->linepos = 0;
	}

	/* write the line fill pattern and color info */
	if ( cvt_need_line_attrs(ge, attr, &cvt->curr_atts) )
	{
		/* write the current indentation */
		mlOutIndent(ge, cvt);

		if ( attr->lpat_type == 6 )
		{
			switch ( ge->grad[attr->line_pat].type )
			{
				case 0: attr->lpat_type = 2; break;
				case 1: attr->lpat_type = 3; break;
				case 2: attr->lpat_type = 4; break;
				case 3: attr->lpat_type = 2; break;
			}
		}
		/* write the attribute */
		cprintf(ge, cvt, "%s <%d %d %d %d %d %d %d>\n", tok[TLINEFILL],
		              attr->line_fg,
		              attr->line_bg,
		              attr->lpat_type,
		              attr->line_pat,
		              0,
		              0,
		              0);

		/* update the current default state */
		cvt->curr_atts.line_fg   = attr->line_fg;
		cvt->curr_atts.line_bg   = attr->line_bg;
		cvt->curr_atts.lpat_type = attr->lpat_type;
		cvt->curr_atts.line_pat  = attr->line_pat;
		cvt->linepos = 0;
	}

	/* write the shadow info */
	if ( cvt_need_shad_attrs(ge, attr, &cvt->curr_atts) )
	{
		/* write the current indentation */
		mlOutIndent(ge, cvt);

		/* write the attribute */
		cprintf(ge, cvt, "%s <%d %d %d %d>\n", tok[TSHADOW],
		              attr->shadow_color,
		              attr->shadow,
		              attr->shadow_x,
		              attr->shadow_y);

		/* update the current default state */
		cvt->curr_atts.shadow_color = attr->shadow_color;
		cvt->curr_atts.shadow       = attr->shadow;
		cvt->curr_atts.shadow_x     = attr->shadow_x;
		cvt->curr_atts.shadow_y     = attr->shadow_y;
		cvt->linepos = 0;
	}

	/* line attribute info */
	if ( cvt->curr_atts.line_thickness != attr->line_thickness )
	{
		cvt_attr_indent(ge, cvt);
		cvt->curr_atts.line_thickness = attr->line_thickness;
		if ( cvt->curr_atts.line_thickness < 0 )
			cvt->curr_atts.line_thickness = 0;
		cvt->linepos += cprintf(ge, cvt,"%s %d", tok[TTHICKNESS],
			cvt->curr_atts.line_thickness);
		mlOutNewline(ge, cvt);
	}
	if ( cvt->curr_atts.line_style != attr->line_style )
	{
		cvt_attr_indent(ge, cvt);
		cvt->curr_atts.line_style = attr->line_style;
		cvt->linepos += cprintf(ge, cvt, "%s %d", tok[TSTYLE], attr->line_style);
		mlOutNewline(ge, cvt);
	}
	if ( cvt->curr_atts.left_arrow != attr->left_arrow )
	{
		cvt_attr_indent(ge, cvt);
		cvt->curr_atts.left_arrow = attr->left_arrow;
		cvt->linepos += cprintf(ge, cvt, "%s %d", tok[TL_ARROW], attr->left_arrow);
		mlOutNewline(ge, cvt);
	}
	if ( cvt->curr_atts.right_arrow != attr->right_arrow )
	{
		cvt_attr_indent(ge, cvt);
		cvt->curr_atts.right_arrow = attr->right_arrow;
		cvt->linepos += cprintf(ge, cvt, "%s %d", tok[TR_ARROW], attr->right_arrow);
		mlOutNewline(ge, cvt);
	}
	if ( cvt->curr_atts.join_symbols != attr->join_symbols )
	{
		cvt_attr_indent(ge, cvt);
		cvt->curr_atts.join_symbols = attr->join_symbols;
		cvt->linepos += cprintf(ge, cvt, "%s %d", tok[TMARKER], attr->join_symbols);
		mlOutNewline(ge, cvt);
	}

	/* text attribute info */
	if ( attr->face != cvt->curr_atts.face )
	{
		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge, cvt, "%s %d", tok[TFONT], attr->face);
		cvt->curr_atts.face = attr->face;
		mlOutNewline(ge, cvt);
	}
	if ( attr->size != cvt->curr_atts.size )
	{
		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge, cvt, "%s %d", tok[TSIZE], attr->size);
		cvt->curr_atts.size = attr->size;
		mlOutNewline(ge, cvt);
	}
	if ( attr->ruling != cvt->curr_atts.ruling )
	{
		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge, cvt, "%s %s",tok[TUNDERLINE],und[attr->ruling]);
		cvt->curr_atts.ruling = attr->ruling;
		mlOutNewline(ge, cvt);
	}
	if ( attr->strike != cvt->curr_atts.strike )
	{
		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge, cvt, "%s %s", tok[TTHRU], tst[attr->strike]);
		cvt->curr_atts.strike = attr->strike;
		mlOutNewline(ge, cvt);
	}
	if ( attr->weight != cvt->curr_atts.weight )
	{
		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge, cvt, "%s %s", tok[TBOLD], tst[attr->weight]);
		cvt->curr_atts.weight = attr->weight;
		mlOutNewline(ge, cvt);
	}
	if ( attr->slant != cvt->curr_atts.slant )
	{
		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge, cvt, "%s %s", tok[TITALICS], tst[attr->slant]);
		cvt->curr_atts.slant = attr->slant;
		mlOutNewline(ge, cvt);
	}
	if ( attr->h_just != cvt->curr_atts.h_just )
	{
		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge, cvt,"%s %s", tok[THOR_JUST], jst[attr->h_just]);
		cvt->curr_atts.h_just = attr->h_just;
		mlOutNewline(ge, cvt);
	}
	if ( attr->v_just != cvt->curr_atts.v_just )
	{
		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge, cvt,"%s %s", tok[TVER_JUST],vjst[attr->v_just]);
		cvt->curr_atts.v_just = attr->v_just;
		mlOutNewline(ge, cvt);
	}
	if ( attr->field_space != cvt->curr_atts.field_space )
	{
		cvt_attr_indent(ge, cvt);
		attr->field_space = attr->field_space *72/1000 + 0.5;
		cvt->linepos += cprintf(ge, cvt,"%s %d",tok[TLINE_SPACE],attr->field_space);
		cvt->curr_atts.field_space = attr->field_space;
		mlOutNewline(ge, cvt);
	}
	if ( attr->l_margin != cvt->curr_atts.l_margin ||
	     attr->r_margin != cvt->curr_atts.r_margin ||
	     attr->t_margin != cvt->curr_atts.t_margin ||
	     attr->b_margin != cvt->curr_atts.b_margin )
	{
		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge, cvt,"%s <%d %d %d %d>", tok[TMARGINS],
		                        attr->l_margin,
		                        attr->r_margin,
		                        attr->t_margin,
		                        attr->b_margin);
		cvt->curr_atts.l_margin = attr->l_margin;
		cvt->curr_atts.r_margin = attr->r_margin;
		cvt->curr_atts.t_margin = attr->t_margin;
		cvt->curr_atts.b_margin = attr->b_margin;
		mlOutNewline(ge, cvt);
	}

	/* some transform info */
	if ( attr->x_scale != cvt->curr_atts.x_scale )
	{
		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge, cvt,"%s %d", tok[TSCALE], attr->x_scale);
		cvt->curr_atts.x_scale = attr->x_scale;
		mlOutNewline(ge, cvt);
	}
	if ( attr->shear != cvt->curr_atts.shear )
	{
		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge, cvt,"%s %d", tok[TSHEAR], attr->shear);
		cvt->curr_atts.shear = attr->shear;
		mlOutNewline(ge, cvt);
	}
	if ( attr->angle != cvt->curr_atts.angle )
	{
		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge, cvt,"%s %d", tok[TANGLE], attr->angle);
		cvt->curr_atts.angle = attr->angle;
		mlOutNewline(ge, cvt);
	}
	if ((attr->field_x_scale != cvt->curr_atts.field_x_scale) ||
		(attr->field_y_scale != cvt->curr_atts.field_y_scale))
	{
		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge, cvt,"%s <%d %d>",tok[TFLD_XYSCALE],
				                attr->field_x_scale,
				                attr->field_y_scale);
		cvt->curr_atts.field_x_scale = attr->field_x_scale;
		cvt->curr_atts.field_y_scale = attr->field_y_scale;
		mlOutNewline(ge, cvt);
	}
	if ((attr->field_shear != cvt->curr_atts.field_shear) ||
		(attr->field_angle != cvt->curr_atts.field_angle))
	{
		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge, cvt,"%s <%d %d>",tok[TSHR_ANG],
			                    attr->field_shear,
			                    attr->field_angle);
		cvt->curr_atts.field_shear = attr->field_shear;
		cvt->curr_atts.field_angle = attr->field_angle;
		mlOutNewline(ge, cvt);
	}
	if ( cvt->linepos != 0 )
	{
		cprintf(ge, cvt, "\n");
		cvt->linepos = 0;
	}
}
static int
mlOutLayer(ge, cvt, layer)
graphics_main_type *ge;
cvt_global_type *cvt;
int layer;
{
	/* write the layer */
	if ( layer != cvt->curr_layer )
	{
		/* write the current indentation */
		mlOutIndent(ge, cvt);

		cprintf(ge, cvt,"%s %d\n", tok[TLAYERS], layer);
		cvt->curr_layer = layer;
		cvt->linepos = 0;
	}
}

/****************************************************************************
 *
 * module:	m l O u t O b j C o m m e n t
 */
static void
mlOutObjComment(ge, cvt, t)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
{
	int ii, needs_tag;
	ge_tag_type *tag;

	/* quick exit */
	if ( !t )
		return;

	/* Put any tags (comments) associated with an object */
	needs_tag = TRUE;
	if ( t->attr.use.tag )
	{
		tag = &ge->tag[t->attr.tag];
		if ( tag ) for ( ii=0; ii<tag->n_strs; ii++ )
		{
			/* indent */
			mlOutIndent(ge, cvt);

			/* comment */
			cvt->linepos = cprintf(ge, cvt, "#%s", tag->str[ii]);
#if (FOR_FILTER && (!TURBO)) /* rather helpful */
			if ( !strncmp(tag->str[ii], "#id", 3) )
				needs_tag = FALSE;
#endif
		}
	}

#if 0 /* rather helpful */
	if ( needs_tag )
	{
		mlOutIndent(ge, cvt);
		cvt->linepos = cprintf(ge, cvt, "#id %d", t->id);
	}
#endif
}

/****************************************************************************
 *
 * module:	m l O u t O b j H d r - Write the object's header to file
 */
static void
mlOutObjHdr(ge, cvt, t, no_indent)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
int no_indent;
{
	int x1, y1, x2, y2;

	/* write the object type */
	if ( no_indent == FALSE )
		mlOutIndent(ge, cvt);
	if ( t->type == ELL && t->data.ell->n_sides > 0 )
		cprintf(ge, cvt,"%s", pnm2[RPOL]);
	else
		cprintf(ge, cvt,"%s", pnm2[t->type]);

	/* write the object's position */
	if ( t->extent.x1 <= t->extent.x2 )
	{
		x1 = t->extent.x1 - PAGE_ZEROX;
		x2 = t->extent.x2 - PAGE_ZEROX;
	}
	else
	{
		x2 = t->extent.x1 - PAGE_ZEROX;
		x1 = t->extent.x2 - PAGE_ZEROX;
	}
	if ( t->extent.y1 <= t->extent.y2 )
	{
		y1 = t->extent.y1 - PAGE_ZEROY;
		y2 = t->extent.y2 - PAGE_ZEROY;
	}
	else
	{
		y2 = t->extent.y1 - PAGE_ZEROY;
		y1 = t->extent.y2 - PAGE_ZEROY;
	}
	/* all non groups need a position */
	if ( t->type != GROUP )
		cprintf(ge, cvt, " %s (%d,%d)\n", tok[TAT], x1, y1);

	/* groups write any special info (chart, blend, extrusion, ...) */
	else if ( t->chart_ix )
		cprintf(ge, cvt, " %s <1 %d>\n", tok[TGRP_INFO], t->chart_ix-1);

	/* a group with no special info - just a return */
	else
		cprintf(ge, cvt, "\n");

	/* update the current pos */
	cvt->dx = x1;
	cvt->dy = y1;
	
	/* write any persistant marks */
	cvt->linepos = 0;
	cvt->indent++;
	if ( t->attr.callback > 0 )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s %d\n", tok[TCBACK], t->attr.callback-1);
	}
	if ( t->flags.hidden )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s %s\n", tok[THIDDEN], tst[1]);
	}
	if ( t->flags.locked )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s %s\n", tok[TLOCKED], tst[1]);
	}
	if ( t->flags.fixed )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s %s\n", tok[TFIXED], tst[1]);
	}
	if ( t->flags.top_attr )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s %s\n", tok[TTOP_ATTR], tst[1]);
	}
	if ( t->t_class != 0 )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s %d\n", tok[TCLASS], t->t_class);		
	}	
}


/****************************************************************************
 *
 * module: m l O u t P n t s - write any Point infomation
 */
static void
mlOutPnts(ge, cvt, pathP, type_trans)
graphics_main_type *ge;
cvt_global_type *cvt;
poly_type *pathP;
int type_trans;
{
	int ii, x, y;

	/* get count of coord pairs and write them to file */
	if ( pathP->nPoints > 0 )
	{
		/* header */
		cvt->linepos = mlOutIndent(ge, cvt);
		if ( type_trans )
			cvt->linepos += cprintf(ge, cvt,"%s ", tok[TT_PNTS]);
		else
			cvt->linepos += cprintf(ge, cvt,"%s ", tok[TPNTS]);

		/* point list */
		for ( ii=0; ii<pathP->nPoints; ii++ )
		{
			x = pathP->points[ii].x - PAGE_ZEROX - cvt->dx;
			y = pathP->points[ii].y - PAGE_ZEROY - cvt->dy;
			if ( MAX_LINE < cvt->linepos )
			{
				cvt->linepos = 0;
				if ( type_trans )
				{
					cprintf(ge, cvt, "\n       ");
					cvt->linepos = 7 + mlOutIndent(ge, cvt);
				}
				else
				{
					cprintf(ge, cvt, "\n     ");
					cvt->linepos = 5 + mlOutIndent(ge, cvt);
				}
			}
			cvt->linepos += cprintf(ge, cvt,"(%d,%d)", x, y);
		}
		if ( cvt->linepos )
		{
			cputc(ge, cvt,'\n');
			cvt->linepos = 0;
		}
	}
}

/****************************************************************************
 *
 * module: m l O u t C t l s - write any control points
 */
static void
mlOutCtls(ge, cvt, pathP, type_trans)
graphics_main_type *ge;
cvt_global_type *cvt;
poly_type *pathP;
int type_trans;
{
	int ii, x, y;

	if ( pathP->nCtrls > 0 )
	{
		/* header */
		cvt->linepos = mlOutIndent(ge, cvt);
		if ( type_trans )
			cvt->linepos += cprintf(ge, cvt,"%s ", tok[TT_CTLS]);
		else
			cvt->linepos += cprintf(ge, cvt,"%s ", tok[TCTLS]);

		/* point list */
		for ( ii=0; ii<pathP->nCtrls; ii++ )
		{
			x = pathP->controls[ii].x - PAGE_ZEROX - cvt->dx;
			y = pathP->controls[ii].y - PAGE_ZEROY - cvt->dy;
			if ( MAX_LINE < cvt->linepos )
			{
				cvt->linepos = 0;
				if ( type_trans )
				{
					cprintf(ge, cvt, "\n       ");
					cvt->linepos = 7 + mlOutIndent(ge, cvt);
				}
				else
				{
					cprintf(ge, cvt, "\n     ");
					cvt->linepos = 5 + mlOutIndent(ge, cvt);
				}
			}
			cvt->linepos += cprintf(ge, cvt,"(%d,%d)", x, y);
		}
		if ( cvt->linepos )
		{
			cputc(ge, cvt,'\n');
			cvt->linepos = 0;
		}
	}
}

/****************************************************************************
 *
 * module: m l O u t T r a n s
 */
static void
mlOutTrans(ge, trans, cvt)
graphics_main_type *ge;
poly_type *trans;
cvt_global_type *cvt;
{
	mlOutPnts(ge, cvt, trans, TRUE);
	mlOutCtls(ge, cvt, trans, TRUE);
}

/****************************************************************************
 *
 * module: m l O u t P a t h
 */
static void
mlOutPath(ge, pathP, nPaths, cvt)
graphics_main_type *ge;
poly_type *pathP;
int nPaths;
cvt_global_type *cvt;
{
	int ii;
	for ( ii=0; ii<nPaths; ii++ )
	{
		mlOutPnts(ge, cvt, &pathP[ii], FALSE);
		mlOutCtls(ge, cvt, &pathP[ii], FALSE);
	}
}

/****************************************************************************
 *
 * module: m l O u t P o l y
 */
static void
mlOutPoly(ge,t,cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	poly_type path;
	int npts;				/* number of points converted poly */
	TmPoint32 *newPointsP=NULL;	/* vertices of converted poly */

	mlOutPath(ge, t->paths, t->n_paths,cvt);
}

/****************************************************************************
 *
 * module: m l O u t E l l
 */
static void
mlOutEll(ge,t,cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	poly_type path;
	TmPoint32 pnts[16], ctls[16], *points, *controls;
	int sz, ii;
	thing newt;

	cvt->linepos = 0;


	/* some specific attributes */
	if ( t->data.ell->start_angle != cvt->startAngle )
	{
		cvt_attr_indent(ge, cvt);
		cvt->startAngle = t->data.ell->start_angle;
		cvt->linepos += cprintf(ge, cvt,"%s %d",tok[TSTART_ANGLE],cvt->startAngle);
		mlOutNewline(ge, cvt);
	}
	if ( t->data.ell->n_sides == 0 && t->data.ell->end_angle != cvt->endAngle )
	{
		cvt_attr_indent(ge, cvt);
		cvt->endAngle = t->data.ell->end_angle;
		cvt->linepos += cprintf(ge, cvt,"%s %d", tok[TEND_ANGLE], cvt->endAngle);
		mlOutNewline(ge, cvt);
	}
	if ( t->data.ell->n_sides > 0 && !cvt->as_1_1 )
	{
		cvt_attr_indent(ge, cvt);
		cvt->nsides = t->data.ell->n_sides;
		cvt->linepos += cprintf(ge, cvt,"%s %d", tok[TNSIDES], cvt->nsides);
		mlOutNewline(ge, cvt);
	}

	if ( !cvt->as_1_1 )
		mlOutPath(ge, t->paths, t->n_paths,cvt);
}
static void
mlOutRPoly(ge,t,cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	mlOutEll(ge,t,cvt);
}
static void
mlOutInset(ge,t,cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
}
static void
mlOutTemp(ge,t,cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
}

/****************************************************************************
 *
 * module: m l O u t R e c t
 */
static void
mlOutRect(ge,t,cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{

	if ((t->data.rect->x_corner_radius != cvt->x_corner_radius) ||
		(t->data.rect->y_corner_radius != cvt->y_corner_radius))
	{
		cvt_attr_indent(ge, cvt);
		cvt->x_corner_radius = t->data.rect->x_corner_radius;
		cvt->y_corner_radius = t->data.rect->y_corner_radius;
		cprintf(ge, cvt,"%s <%d %d>", tok[TXYRAD],
		        cvt->x_corner_radius,
		        cvt->y_corner_radius);
		mlOutNewline(ge, cvt);
	}
	if (t->data.rect->scale_corners != cvt->scale_corners)
	{
		cvt_attr_indent(ge, cvt);
		cvt->scale_corners = t->data.rect->scale_corners;
		cvt->linepos += cprintf(ge, cvt,"%s %d", tok[TSCALE_CORNERS],
		                        cvt->scale_corners);
		mlOutNewline(ge, cvt);
	}

	if ((t->data.rect->x_corner_numerator != cvt->x_corner_numerator) ||
		(t->data.rect->y_corner_numerator != cvt->y_corner_numerator))
	{
		cvt_attr_indent(ge, cvt);
		cvt->x_corner_numerator = t->data.rect->x_corner_numerator;
		cvt->y_corner_numerator = t->data.rect->y_corner_numerator;
		cvt->linepos += cprintf(ge, cvt,"%s <%d %d>", tok[TPOSITION],
		                        cvt->x_corner_numerator,
		                        cvt->y_corner_numerator);
		mlOutNewline(ge, cvt);
	}

	mlOutPath(ge, t->paths, t->n_paths,cvt);
}

/****************************************************************************
 *
 * module: m l O u t L i n e
 */
static void
mlOutLine(ge,t,cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	mlOutPath(ge, t->paths, t->n_paths,cvt);
}

/****************************************************************************
 *
 * module: m l O u t S t r o k e
 */
static void
mlOutStroke(ge,t,cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	mlOutPoly(ge,t,cvt);
}

/****************************************************************************
 *
 * module: m l O u t P a r a I n f o
 */
static void
mlOutParaInfo(ge,cvt,t)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
{
	if ( cvt->field_wid  != t->data.txtbox->field_wid )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s <%d %d %d %d>\n", tok[TH_SPACE],
		        t->data.txtbox->field_wid,
		        0, 0, 0);

		cvt->field_wid  = t->data.txtbox->field_wid;
	}
}

/****************************************************************************
 *
 * module:  m l O u t T e x t S t r i n g
 */
static void
mlOutTextString(ge,cvt,t)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
{
	if ( t->data.str->n_chars > 0 )
	{
		mlOutIndent(ge, cvt);
		mlOutString(ge, t->data.str->n_chars, t->data.str->chars, cvt, FALSE);
		cputc(ge, cvt,'\n');
		cvt->linepos = 0;
	}
}

/****************************************************************************
 *
 * module:  m l O u t T e x t  - write the string(s)
 */
static void
mlOutText(ge,t,cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	int ii;
	thing *txt_t;

	/* Write out the actual strings */
	for ( ii=0; ii<t->data.txtbox->n_strings; ii++ )
	{
		/* write the TEXT token */
		mlOutIndent(ge, cvt);
		txt_t = t->data.txtbox->strings[ii];
		cprintf(ge, cvt,"%s\n", pnm2[TEXT]);
		cvt->linepos = 0;

		/* write the string */
		cvt->indent++;
		mlOutAttrs(ge, &txt_t->attr, 0, cvt);
		mlOutPath(ge, txt_t->paths, txt_t->n_paths, cvt);
		mlOutTextString(ge, cvt,txt_t);
		cvt->indent--;
	}
}

/****************************************************************************
 *
 * module:	m l O u t B o x T e x t -  TEXTBOX
 */
static void
mlOutTextBox(ge,t,cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{

	/* don't write empty text strings */
	if ( t->n_paths <= 0 && t->data.txtbox->n_strings <= 0 )
		return;
	/* text attrs */
	mlOutAttrs(ge, &t->attr, 0, cvt);
	mlOutLayer(ge, cvt, t->layer);
	if ((cvt->x_off != t->data.txtbox->x_off) ||
		(cvt->y_off != t->data.txtbox->y_off))
	{
		cvt_attr_indent(ge, cvt);
		cprintf(ge, cvt,"%s <%d %d>", tok[TTXTXYOFF],
			cvt->x_off = t->data.txtbox->x_off,
			cvt->y_off = t->data.txtbox->y_off);
		mlOutNewline(ge, cvt);
	}

	/* draw the bounding polygon(s) */
	mlOutPath(ge, t->paths, t->n_paths, cvt);
	mlOutTrans(ge, &t->data.txtbox->trans, cvt);

	/* the paragraph info for text */
	mlOutParaInfo(ge, cvt, t);

	/* the actual text */
	mlOutText(ge,t,cvt);
}

/****************************************************************************
 *
 * module:	m l O u t G r o u p - export a Grouped object
 */
static void
mlOutGroup(ge, t, cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	int ii;

	/* write the children */
	for ( ii=0; ii<t->data.group->num; ii++ )
	{
		mlOutObjComment(ge, cvt, t->data.group->list[ii]);
		mlOutPart(ge, cvt, t->data.group->list[ii], FALSE);
	}

	/* terminate the group */
	cvt->indent--;
	mlOutIndent(ge, cvt);
	cvt->indent++;
	cprintf(ge, cvt, "%s %s\n", tok[TEND], pnm2[GROUP]);
}

/****************************************************************************
 *
 * module:	m l O u t P a r t - top-level part 
 */
static void
mlOutPart(ge, cvt, t, no_indent)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
int no_indent;
{
	/* null objects became valid in rev3.0 */
	if ( !t )
	{
		if ( no_indent == FALSE )
			mlOutIndent(ge, cvt);
		cprintf(ge, cvt, "%s\n", pnm2[0]);
	}
	else if ( t->type >= TEXTBOX && t->type <= ELL )
	{
		/* write the type, position, and any persistant marks */
		mlOutObjHdr(ge, cvt, t, no_indent);

		/* write out the attributes */
		if ( t->type == IMAGE )
			mlOutAttrs(ge, &t->attr, t->data.img->cmapID, cvt);
		else
			mlOutAttrs(ge, &t->attr, 0, cvt);
		mlOutLayer(ge, cvt, t->layer);

		/* write the type specific info */
		switch ( t->type )
		{
			case ELL:      mlOutEll    (ge,t,cvt); break;
			case RECT:     mlOutRect   (ge,t,cvt); break;
			case LINE:     mlOutLine   (ge,t,cvt); break;
			case STROK:    mlOutStroke (ge,t,cvt); break;
			case POLY:     mlOutPoly   (ge,t,cvt); break;
			case TEXTBOX:  mlOutTextBox(ge,t,cvt); break;
			case IMAGE:    mlOutImage  (ge,t,cvt); break;
			case GROUP:    mlOutGroup  (ge,t,cvt); break;
			case RPOL:     mlOutRPoly  (ge,t,cvt); break;
			case INSET:    mlOutInset  (ge,t,cvt); break;
			case TEMPLATE: mlOutTemp   (ge,t,cvt); break;
		}

		/* restore the indentation level (bumped in mlOutObjHdr) */
		cvt->indent--;
		cvt->dx = 0;
		cvt->dy = 0;
	}
}

/****************************************************************************
 *
 * module:	m l O u t S l i d e I n f o
 */
static void
mlOutSlideInfo(ge, cvt, page)
graphics_main_type *ge;
cvt_global_type *cvt;
ge_page_type *page;
{
}

/****************************************************************************
 *
 * module:	m l O u t L a y e r s
 */
static void
mlOutLayers(ge, cvt, page)
graphics_main_type *ge;
cvt_global_type *cvt;
ge_page_type *page;
{
	layer_type *l_pack;
	tx_char tbuff[400];
	char *def_name, str2[STRPATH+1];
	int i, j, len;

	if ( page->n_layers == 0 )
		return;

	/* don't need to write the default layer */
	if ( page->n_layers == 1 )
	{
		def_name = XLT("Default");
		len = strlen(def_name);
		for ( i=0; i<len; i++ )
			tbuff[i] = (tx_char)def_name[i];
		tbuff[i] = (tx_char)0;
		l_pack = &page->layer[0];
		if ( l_pack->n_chars == len &&
		     tx_strncmp(tbuff, l_pack->name, len) == 0 &&
		     l_pack->locked == FALSE &&
		     l_pack->hidden == FALSE &&
		     l_pack->prints == TRUE )
			return;
	}

	/* note the start of the layers info section */
	cprintf(ge, cvt,"%s\n", tok[TLAYERS]);

	/* output info for each layer */
	for ( i=0; i<page->n_layers; i++ )
	{
		l_pack = &page->layer[i];

		/* print the layer name */
		cprintf(ge, cvt, "<");
		mlOutString(ge, l_pack->n_chars, l_pack->name, cvt, FALSE);

		/* locked and hidden tags */
		cprintf(ge, cvt," %d %d %d %d>\n",
		        l_pack->locked,
		        l_pack->hidden,
		        l_pack->prints,
		        l_pack->backgd);
	}

	/* note the end of the layers info section */
	cprintf(ge, cvt,"%s %s\n", tok[TEND], tok[TLAYERS]);
}

/****************************************************************************
 *
 * module:	m l O u t P i c t
 */
static void
mlOutPict(ge, cvt, page)
graphics_main_type *ge;
cvt_global_type *cvt;
ge_page_type *page;
{
	thing *t, *temp_t;
	int i;

	/* note the start of the picture segment */
	cprintf(ge, cvt, "%s\n", tok[TPICTURE]);
	t = &page->t;
	if ( t && t->type != 0 && (t->type != GROUP || t->data.group->num > 0) )
	{
		/* write out all the parts in the document */
		if ( t->type != GROUP )
		{
			mlOutObjComment(ge, cvt, t);
			mlOutPart(ge, cvt, t, FALSE);
		}

		else for ( i=0; i<t->data.group->num; i++ )
		{
			temp_t = t->data.group->list[i];
			if ( temp_t )
			{
				if ( temp_t->type == TEMPLATE )
					temp_t = temp_t->data.temp->child_t;
				mlOutObjComment(ge, cvt, temp_t);
				mlOutPart(ge, cvt, temp_t, FALSE);
			}
		}
	}

	/* note the end of the picture segment */
	cprintf(ge, cvt,"%s %s\n", tok[TEND], tok[TPICTURE]);
}

/****************************************************************************
 *
 * module:	m l O u t N o t e s
 */
static void
mlOutNotes(ge, cvt, page)
graphics_main_type *ge;
cvt_global_type *cvt;
ge_page_type *page;
{
}

/****************************************************************************
 *
 * module:	m l O u t P a g e s
 */
static void
mlOutPages(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int i;
	for ( i=0; i<ge->n_pages; i++ )
	{
		/* write the slide info for this page */
		mlOutSlideInfo(ge, cvt, &ge->page[i]);

		/* write the page's layers */
		mlOutLayers(ge, cvt, &ge->page[i]);

		/* write the page's paint list */
		mlOutPict(ge, cvt, &ge->page[i]);

		/* write the speaker notes for this page */
		mlOutNotes(ge, cvt, &ge->page[i]);

		/* done */
		cprintf(ge, cvt, "\n");
	}
}

/****************************************************************************
 *
 * module:	m l O u t P a l e t
 */
static void
mlOutPalet(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	udo_type udoP;
	int ii;

	/* for each UDO */
	for ( ii=0; ii<ge->n_udos; ii++ )
	{
		/* write any comments */
		mlOutObjComment(ge, cvt, ge->udo[ii].t);

		/* write the part token and the part name */
		if ( cvt->linepos != 0 )
			mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s ", tok[TPART]);
		mlOutString(ge, ge->udo[ii].n_chars, ge->udo[ii].name, cvt, FALSE);

		/* write the object associated with the part */
		cprintf(ge, cvt, " ");
		mlOutPart(ge, cvt, ge->udo[ii].t, TRUE);
	}
	if ( ge->n_udos > 0 )
		cprintf(ge, cvt, "\n");
}

/****************************************************************************
 *
 * module:	m l O u t N a m e d F i l l s - write out user defined fill defs
 */
static void
mlOutNamedFills(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int i, n, t, bit6, (*func)();
	int byte_width, llen, datum;
	udfill_type *fill_pack;
	char *data, *string;

	if ( ge->n_fills <= 0 )
		return;

	if ( cvt->encoding == BIT6 )
	{
		bit6 = TRUE;
		func = mailOutBit6;
	}
	else
	{
		bit6 = FALSE;
		func = mailOutHex;
	}

	/* note the start of the fill list */
	cprintf(ge, cvt,"%s\n", tok[TFILLNAMES]);
	for ( i=0; i<ge->n_fills; i++ )
	{
		fill_pack = &ge->fill[i];

		/* print the "Header" */
		cprintf(ge, cvt, " <");
		mlOutString(ge, fill_pack->n_chars, fill_pack->name, cvt, FALSE);
		cprintf(ge, cvt," %d %d %d %d>\n ", fill_pack->hyt, fill_pack->wid,
										fill_pack->cmapID, bit6);

		/* output the data 8 bits at a time */
		llen = 0;
		if (fill_pack->cmapID > 0 )
			byte_width = fill_pack->wid+fill_pack->wid%2;
		else
			byte_width = (fill_pack->wid+7)/ 8 +((fill_pack->wid+7)/ 8)%2;
		for ( n=0; n<fill_pack->hyt; n++)
		{
			for ( t=0; t<byte_width; t++ )
			{
				datum = fill_pack->data[n*byte_width + t];
				llen += (*func)(ge, datum & 0xff, llen, cvt);
				if ( MAX_LINE < llen )
				{
					cprintf(ge, cvt, "\n ");
					llen=0;
				}
			}
			llen += (*func)(ge, -1, llen, cvt);
		}
		if ( llen > 0 )
			cputc(ge, cvt,'\n');
	}

	/* note the end of the fill list */
	cprintf(ge, cvt,"%s %s\n\n", tok[TEND], tok[TFILLNAMES]);
}

/****************************************************************************
 *
 * module:	m l O u t F a c e s
 */
static void
mlOutFaces(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int i;
    int len;
    tx_char tbuff[500];

	/* note the start of the face list */
	if ( ge->n_faces == 0 )
		return;

	/* don't write the default */
	if ( ge->n_faces == 1 && strcmp(NXLT("Times"), ge->face[0]) == 0 )
		return;

	/* write the face list */
	cprintf(ge, cvt,"%s\n", tok[TFONT]);
	for ( i=0; i<ge->n_faces; i++ )
	{
		len = MultiToWideStr(ge, ge->face[i], tbuff);
        mlOutString(ge,  len, tbuff, cvt, FALSE );
		cputc(ge, cvt,'\n');
	}
	cprintf(ge, cvt,"%s %s\n\n", tok[TEND], tok[TFONT]);
}

/****************************************************************************
 *
 * module:	m l O u t C o l o r m a p - write the colormap to the file
 */
static void
mlOutColormap(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int ii, jj;
	color_type *colorP;
    int len, same;
    tx_char tbuff[200];
	graphics_main_type localGE;

	if (ge->n_cmaps == 0) /* gotta have 1 */
	{
		colormapInit(ge);
		return; /* since the next condition is satisfied */
	}

	/* If there's only 1 colormap and it's the default */
	/* colormap , don't bother writing it. */ 
	if (ge->n_cmaps == 1) 
	{
		/* get a default colormap for comparison */
		localGE.pid = 0;
		localGE.n_cmaps = 0;
		localGE.cmap = NULL;
		colormapInit(&localGE);

		/* do the comparison */
		same = foundCmap(ge, &ge->cmap[0],localGE.cmap);

		/* free the comparison map */
		freeCmap(&localGE,localGE.cmap);

		if (same == TRUE)
			return;
	}

	for (ii=0; ii<ge->n_cmaps; ii++ )
	{
		/* note the start of the colormap */
		cprintf(ge, cvt,"%s\n", tok[TCOLORMAP]);

		/* ordinal list of color names, definitions, and ink types */
		colorP = ge->cmap[ii].color;
		for ( jj=0; jj<ge->cmap[ii].n_colors; jj++ )
		{
			/* first the name of the color */
			len = MultiToWideStr(ge, colorP[jj].name, tbuff );
			cprintf(ge, cvt, "<");
            mlOutString(ge,  len, tbuff, cvt, FALSE );

			/* then the ink type */
			cprintf(ge, cvt," %d ", colorP[jj].ink_type);

			/* finally the cmyk definition */
			cprintf(ge, cvt,"%d %d %d %d %d>\n", colorP[jj].c,
			                                colorP[jj].m,
			                                colorP[jj].y,
			                                colorP[jj].k,
			                                colorP[jj].see_thru);
		}

		/* note the end of the colormap */
		cprintf(ge, cvt,"%s %s\n\n", tok[TEND], tok[TCOLORMAP]);
	}
}

/****************************************************************************
 *
 * module:  m l O u t W i n d o w S i z e
 */
static void
mlOutWindowSize(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int wid, hyt;

	/* set the window width */
	wid = ge->state.drawn_area.x2 - ge->state.drawn_area.x1;
	if ( wid <= 0 )
		wid = 7.5 * MILS_P_INCH;

	/* set the window height */
	hyt = ge->state.drawn_area.y2 - ge->state.drawn_area.y1;
	if ( hyt <= 0 )
		hyt = 5.5 * MILS_P_INCH;

	cprintf(ge, cvt," %s <%d %d>\n", tok[TWIN_SIZE], wid, hyt);
}

/****************************************************************************
 *
 * module:  m l O u t S e s s i o n I n f o
 */
static void
mlOutSessionInfo(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	struct ge_print_info_type *prnt;
	struct session *sess;

	/* begin */
	cprintf(ge, cvt,"%s\n", tok[TSESSION]);

	/* window info */
	mlOutWindowSize(ge, cvt);

	/* document info */
	sess = &ge->state;
	if ( sess->stickyDots )
		cprintf(ge, cvt," %s %s\n",tok[TSTICKY], tst[1]);
	if ( sess->autoGrid )
		cprintf(ge, cvt," %s %s\n",tok[TAUTOGRID], tst[1]);
	if ( sess->rulerOn || sess->pageOn || sess->readOut ||
	     sess->gridOn  || !sess->bordersOn )
	{
		cprintf(ge, cvt, " %s <%d %d %d %d %d>\n", tok[TVIEW],
		        sess->rulerOn,
		        sess->pageOn,
		        sess->readOut,
		        sess->gridOn,
				sess->bordersOn);
	}
	if ( sess->im_hex || sess->im_raw )
	{
		if ( sess->im_hex && sess->im_raw )
			cprintf(ge, cvt," %s 3\n",tok[TIMAGE_FMT]);
		else if ( sess->im_hex )
			cprintf(ge, cvt," %s 2\n",tok[TIMAGE_FMT]);
		else if ( sess->im_raw )
			cprintf(ge, cvt," %s 1\n",tok[TIMAGE_FMT]);
		else
			cprintf(ge, cvt," %s 0\n",tok[TIMAGE_FMT]);
	}

	if (sess->dpi != 1000 || sess->gridSize != 250 || sess->unit_size != 1000)
	{
		cprintf(ge, cvt," %s %d %s %d %s <%d %d \"%s\">\n",
			tok[TGRID_DPI], sess->dpi,
			tok[TGRID_FACTOR], sess->gridSize,
			tok[TUNITS], sess->unit_size,sess->unit_precision,sess->unit_name);
	}

	if ((sess->line_1 != 28) || /* defaults */
	    (sess->line_2 != 42) ||
	    (sess->line_3 != 56))
	{
		cprintf(ge, cvt," %s <%d %d %d>\n",tok[TL_WIDTHS],
				sess->line_1, sess->line_2, sess->line_3);
	}

	/* print setup info */
	prnt = &ge->print;

	/* printable area - position, size, and zero point */
	cvt->linepos = 0;
	if ( prnt->page_wid != 0 || prnt->page_hyt != 0 )
	{
		if ( prnt->page_x != 59750 || prnt->page_y != 58000 )
		{
			cvt->linepos += cprintf(ge, cvt," %s <%d %d>", tok[TPAGEXY],
			                             prnt->page_x - 59750,
			                             prnt->page_y - 58000);
		}
		if ( ge->x_zero_pt != 60250 || ge->y_zero_pt != 59000 )
		{
			cvt->linepos += cprintf(ge, cvt," %s <%d %d>", tok[TPAGEZEROS],
			                             ge->x_zero_pt - 59750,
			                             ge->y_zero_pt - 58000);
		}
		if ( cvt->linepos > 0 )
		{
			cprintf(ge, cvt, "\n");
			cvt->linepos = 0;
		}
		if ( prnt->page_wid != 8500 )
		{
			cvt->linepos += cprintf(ge, cvt," %s %d",tok[TPAGEWID],
			                        prnt->page_wid);
		}
		if ( prnt->page_hyt != 11000 )
		{
			cvt->linepos += cprintf(ge, cvt," %s %d",tok[TPAGEHYT],
			                        prnt->page_hyt);
		}
		if ( cvt->linepos > 0 )
		{
			cprintf(ge, cvt, "\n");
			cvt->linepos = 0;
		}
	}

	/* paper type */
	if ( prnt->wid != 8500 )
		cvt->linepos += cprintf(ge, cvt," %s %d", tok[TPRINTWID], prnt->wid);
	if ( prnt->hyt != 11000 )
		cvt->linepos += cprintf(ge, cvt," %s %d", tok[TPRINTHYT], prnt->hyt);
	if ( prnt->landscape )
		cvt->linepos += cprintf(ge, cvt," %s %s", tok[TPRINTLAND], tst[1]);
	if ( cvt->linepos > 0 )
		cprintf(ge, cvt, "\n");
	cvt->linepos = 0;
	if ( prnt->margin.x1 != 500 || prnt->margin.y1 != 1000 ||
	     prnt->margin.x2 != 500 || prnt->margin.y2 != 1000 )
	{
		cprintf(ge, cvt," %s <%d %d %d %d>\n", tok[TPRINTMARG],
		                                   prnt->margin.x1,
		                                   prnt->margin.y1,
		                                   prnt->margin.x2,
		                                   prnt->margin.y2);
	}
	if ( prnt->use_content_border ||
	     prnt->use_crop_marks ||
	     prnt->use_regis_marks )
	{
		cprintf(ge, cvt, " %s <%d %d %d>\n", tok[TPRINTMARKS],
		                                 prnt->use_content_border,
		                                 prnt->use_crop_marks,
		                                 prnt->use_regis_marks);
	}
	if ( prnt->bleed )
		cprintf(ge, cvt," %s %d\n", tok[TPRINTBLEED], prnt->bleed);

	/* what page is active */
	if ( ge->page_ix != 0 )
		cprintf(ge, cvt, " %s %d\n", tok[TCURRENT_PAGE], ge->page_ix+1);

	/* default attributes */
	cvt->linepos = 0;
	cvt->indent = 1;
	mlOutAttrs(ge, &ge->def_attr, 0, cvt);
	mlOutLayer(ge, cvt, ge->state.layer);
	cvt->indent = 0;

	/* done */
	cprintf(ge, cvt,"%s %s\n\n", tok[TEND], tok[TSESSION]);
}

/****************************************************************************
 *
 * module:  m l N e e d H d r F t r s
 */
static int
mlNeedHdrsFtrs(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	struct ge_hdrftr_main_type *hf;
	int needed;

	hf = &ge->print.hdrftr;
	if ( hf->base.hdr_n_lines > 0 || hf->base.ftr_n_lines > 0 )
		needed = TRUE;
	else
		needed = FALSE;
	if ( needed == FALSE  && hf->use_even_odd )
	{
		if ( hf->even.hdr_n_lines > 0 || hf->even.ftr_n_lines > 0 )
			needed = TRUE;
	}
	if ( needed == FALSE && hf->use_first )
	{
		if ( hf->first.hdr_n_lines > 0 || hf->first.ftr_n_lines > 0 )
			needed = TRUE;
	}
	if ( needed == FALSE && hf->use_final )
	{
		if ( hf->final.hdr_n_lines > 0 || hf->final.ftr_n_lines > 0 )
			needed = TRUE;
	}
	return(needed);
}

/****************************************************************************
 *
 * module:	mlOutHFstring
 */
static void
mlOutHFstring(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	tx_char *str;
	int len, i;

	if ( !cvt->ico->string || !cvt->ico->string[0] )
		cprintf(ge, cvt,"   %s \"\"\n", pnm2[TEXT]);
	else
	{   
		len = strlen(cvt->ico->string);
		str = (tx_char *)gloc(ge, (len+1) * sizeof(tx_char));
		for ( i=0; i<len; i++ )
			str[i] = cvt->ico->string[i];
		str[i] = (tx_char)0;

		cprintf(ge, cvt,"   %s \"", pnm2[TEXT]);
		mlOutString(ge, len, str, cvt, TRUE);
		cprintf(ge, cvt,"\n");

		gfree(ge, str);
	}
	if (cvt->hfstr.face_ix != cvt->ico->face_ix)
		cprintf(ge, cvt,"   %s %d\n", tok[TFONT],
			cvt->hfstr.face_ix = cvt->ico->face_ix);

	if (cvt->hfstr.pointsize != cvt->ico->pointsize)
		cprintf(ge, cvt,"   %s %d\n", tok[TSIZE],
			cvt->hfstr.pointsize = cvt->ico->pointsize);

	if (cvt->hfstr.color_ix != cvt->ico->color_ix)
		cprintf(ge, cvt,"   %s %d\n", tok[TFORE_COLOR],
			cvt->hfstr.color_ix = cvt->ico->color_ix);

	if (cvt->hfstr.bold != cvt->ico->bold)
		cprintf(ge, cvt,"   %s %s\n", tok[TBOLD],
			tst[cvt->hfstr.bold = cvt->ico->bold]);

	if (cvt->hfstr.italic != cvt->ico->italic)
		cprintf(ge, cvt,"   %s %s\n", tok[TITALICS],
			tst[cvt->hfstr.italic = cvt->ico->italic]);

	if (cvt->hfstr.strike_thru != cvt->ico->strike_thru)
		cprintf(ge, cvt,"   %s %s\n", tok[TTHRU],
			tst[cvt->hfstr.strike_thru = cvt->ico->strike_thru]);

	if (cvt->hfstr.underline != cvt->ico->underline)
		cprintf(ge, cvt,"   %s %d\n", tok[TUNDERLINE],
			cvt->hfstr.underline = cvt->ico->underline);

	if (cvt->hfstr.underline_words_only != cvt->ico->underline_words_only)
		cprintf(ge, cvt,"   %s %s\n", tok[TWORDSONLY],
			tst[cvt->hfstr.underline_words_only = cvt->ico->underline_words_only]);
}

/****************************************************************************
 *
 * module:	mlOutHF
 */
static void
mlOutHF(ge,cvt,lineNum)
graphics_main_type *ge;
cvt_global_type *cvt;
int lineNum;
{
	cvt->ico = &cvt->hf->inner;
	cprintf(ge, cvt,"  %s %d\n", tok[TINNER], lineNum);
	mlOutHFstring(ge,cvt);
	cprintf(ge, cvt,"  %s %s\n", tok[TEND], tok[TINNER]);

	cvt->ico = &cvt->hf->center;
	cprintf(ge, cvt,"  %s %d\n", tok[TCENTER], lineNum);
	mlOutHFstring(ge,cvt);
	cprintf(ge, cvt,"  %s %s\n", tok[TEND], tok[TCENTER]);

	cvt->ico = &cvt->hf->outer;
	cprintf(ge, cvt,"  %s %d\n", tok[TOUTER], lineNum);
	mlOutHFstring(ge,cvt);
	cprintf(ge, cvt,"  %s %s\n", tok[TEND], tok[TOUTER]);
}

/****************************************************************************
 *
 * module:  m l O u t F t r P a g e
 */
static void
mlOutFtrPage(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int ii;

	cprintf(ge, cvt,"  %s ", tok[TFOOTER]);
	cprintf(ge, cvt,"%s %d ", tok[TOFFSET], cvt->hdr_ftr->ftr_offset);
	cprintf(ge, cvt,"%s %d\n", tok[TNLINES], cvt->hdr_ftr->ftr_n_lines);
	for (ii=0; ii<cvt->hdr_ftr->ftr_n_lines; ii++)
	{
		cvt->hf = &cvt->hdr_ftr->ftr[ii];
		mlOutHF(ge,cvt,ii);
	}
	cprintf(ge, cvt,"  %s %s\n", tok[TEND], tok[TFOOTER]);
}

/****************************************************************************
 *
 * module:  m l O u t H d r P a g e
 */
static void
mlOutHdrPage(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int ii;

	cprintf(ge, cvt,"  %s ", tok[THEADER]);
	cprintf(ge, cvt,"%s %d ", tok[TOFFSET], cvt->hdr_ftr->hdr_offset);
	cprintf(ge, cvt,"%s %d\n", tok[TNLINES], cvt->hdr_ftr->hdr_n_lines);
	for (ii=0; ii<cvt->hdr_ftr->hdr_n_lines; ii++)
	{
		cvt->hf = &cvt->hdr_ftr->hdr[ii];
		mlOutHF(ge,cvt,ii);
	}
	cprintf(ge, cvt,"  %s %s\n", tok[TEND], tok[THEADER]);
}

/****************************************************************************
 *
 * module:  m l O u t H d r s F t r s
 */
static void
mlOutHdrsFtrs(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	if ( mlNeedHdrsFtrs(ge,cvt) == FALSE )
		return;

	/* set defaults */
	memset(&cvt->hfstr,0,sizeof(ge_hdrftr_str_type));
	cvt->hfstr.pointsize = 120000;
	cvt->hfstr.color_ix = 1;

	/* write out the header */
	cprintf(ge, cvt,"%s", tok[THDRFTR]);
	if (ge->print.hdrftr.use_first)
		cprintf(ge, cvt,"  %s %s", tok[TUSE],tok[TFIRST]);
	if (ge->print.hdrftr.use_even_odd)
		cprintf(ge, cvt,"  %s %s", tok[TUSE],tok[TEVENODD]);
	if (ge->print.hdrftr.use_final)
		cprintf(ge, cvt,"  %s %s", tok[TUSE],tok[TFINAL]);
	cprintf(ge, cvt,"\n");

	/* write out the special first page hdr/ftr info */
	if (ge->print.hdrftr.first.hdr_offset ||
		ge->print.hdrftr.first.hdr_n_lines ||
		ge->print.hdrftr.first.ftr_offset ||
		ge->print.hdrftr.first.ftr_n_lines)
	{
		cprintf(ge, cvt," %s\n", tok[TFIRST]);
		if (ge->print.hdrftr.first.hdr_offset ||
			ge->print.hdrftr.first.hdr_n_lines)
		{
			cvt->hdr_ftr = &ge->print.hdrftr.first;
			mlOutHdrPage(ge,cvt);
		}
		if (ge->print.hdrftr.first.ftr_offset ||
			ge->print.hdrftr.first.ftr_n_lines)
		{
			cvt->hdr_ftr = &ge->print.hdrftr.first;
			mlOutFtrPage(ge,cvt);
		}
		cprintf(ge, cvt," %s %s\n", tok[TEND], tok[TFIRST]);
	}

	/* write out the basic page hdr/ftr info (also used for odd pages) */
	if (ge->print.hdrftr.base.hdr_offset ||
		ge->print.hdrftr.base.hdr_n_lines ||
		ge->print.hdrftr.base.ftr_offset ||
		ge->print.hdrftr.base.ftr_n_lines)
	{
		cprintf(ge, cvt," %s\n", tok[TBASE]);
		if (ge->print.hdrftr.base.hdr_offset ||
			ge->print.hdrftr.base.hdr_n_lines)
		{
			cvt->hdr_ftr = &ge->print.hdrftr.base;
			mlOutHdrPage(ge,cvt);
		}
		if (ge->print.hdrftr.base.ftr_offset ||
			ge->print.hdrftr.base.ftr_n_lines)
		{
			cvt->hdr_ftr = &ge->print.hdrftr.base;
			mlOutFtrPage(ge,cvt);
		}
		cprintf(ge, cvt," %s %s\n", tok[TEND], tok[TBASE]);
	}

	/* write out the special even page hdr/ftr info */
	if (ge->print.hdrftr.even.hdr_offset ||
		ge->print.hdrftr.even.hdr_n_lines ||
		ge->print.hdrftr.even.ftr_offset ||
		ge->print.hdrftr.even.ftr_n_lines)
	{
		cprintf(ge, cvt," %s\n", tok[TEVEN]);
		if (ge->print.hdrftr.even.hdr_offset ||
			ge->print.hdrftr.even.hdr_n_lines)
		{
			cvt->hdr_ftr = &ge->print.hdrftr.even;
			mlOutHdrPage(ge,cvt);
		}
		if (ge->print.hdrftr.even.ftr_offset ||
			ge->print.hdrftr.even.ftr_n_lines)
		{
			cvt->hdr_ftr = &ge->print.hdrftr.even;
			mlOutFtrPage(ge,cvt);
		}
		cprintf(ge, cvt," %s %s\n", tok[TEND], tok[TEVEN]);
	}

	/* write out the special final page hdr/ftr info */
	if (ge->print.hdrftr.final.hdr_offset ||
		ge->print.hdrftr.final.hdr_n_lines ||
		ge->print.hdrftr.final.ftr_offset ||
		ge->print.hdrftr.final.ftr_n_lines)
	{
		cprintf(ge, cvt," %s\n", tok[TFINAL]);
		if (ge->print.hdrftr.final.hdr_offset ||
			ge->print.hdrftr.final.hdr_n_lines)
		{
			cvt->hdr_ftr = &ge->print.hdrftr.final;
			mlOutHdrPage(ge,cvt);
		}
		if (ge->print.hdrftr.final.ftr_offset ||
			ge->print.hdrftr.final.ftr_n_lines)
		{
			cvt->hdr_ftr = &ge->print.hdrftr.final;
			mlOutFtrPage(ge,cvt);
		}
		cprintf(ge, cvt," %s %s\n", tok[TEND], tok[TFINAL]);
	}

	cprintf(ge, cvt,"%s %s\n\n", tok[TEND], tok[THDRFTR]);

	/* make sure hfstr attrs are in sync with curr_atts */
	if ( cvt->hfstr.face_ix != cvt->curr_atts.face )
		cvt->curr_atts.face = cvt->hfstr.face_ix;
	if ( cvt->hfstr.pointsize != cvt->curr_atts.size )
		cvt->curr_atts.size = cvt->hfstr.pointsize;
	if ( cvt->hfstr.color_ix != cvt->curr_atts.fill_fg )
		cvt->curr_atts.fill_fg = cvt->hfstr.color_ix;
	if ( cvt->hfstr.bold != cvt->curr_atts.weight )
		cvt->curr_atts.weight = cvt->hfstr.bold;
	if ( cvt->hfstr.italic != cvt->curr_atts.slant )
		cvt->curr_atts.slant = cvt->hfstr.italic;
	if ( cvt->hfstr.strike_thru != cvt->curr_atts.strike )
		cvt->curr_atts.strike = cvt->hfstr.strike_thru;
	if ( cvt->hfstr.underline != cvt->curr_atts.ruling )
		cvt->curr_atts.ruling = cvt->hfstr.underline;
}

/****************************************************************************
 *
 * module:	m l O u t H o o k s
 */
static void
mlOutHooks(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int i;

	for ( i=0; i<ge->n_hooks; i++ )
		cprintf(ge, cvt,"%s %s %s\n", ATOK_HOOK,ge->hook[i].type,ge->hook[i].name);
}

/****************************************************************************
 *
 * module:	m l O u t L i n k s
 */
static void
mlOutLinks(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int i;

	for ( i=0; i<ge->n_links; i++ )
		cprintf(ge, cvt,"%s %s\n", ATOK_LINK, ge->link[i].path);
}

/****************************************************************************
 *
 * module:	m l O u t E m b e d
 */
static void
mlOutEmbed(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int i, n, size;
	char *data;

	for ( i=0; i<ge->n_embeds; i++ )
	{
		size = ge->embed[i].size - 1; /* don't count NULL at end */
		data = ge->embed[i].data;
		n = 0;

		while ( n < size )
		{
			cputc(ge, cvt, *data);
			data++;
			n++;
		}
	}
}

/****************************************************************************
 *
 * module:	m l O u t H e a d e r
 */
static void
mlOutHeader(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int i;

	/* write the header line */
	AFileHeaderFromInfo(&cvt->info, cvt->buf);

	cprintf(ge, cvt, "%s\n", cvt->buf);

	/* dump any meta comments now */
	for ( i=0; i<ge->n_cmts; i++ )
		cprintf(ge, cvt, "** \"%s\" %s\n", ge->cmt[i].var, ge->cmt[i].val);

	/* write the hook list */
	mlOutHooks(ge, cvt);

	/* write the link list */
	mlOutLinks(ge, cvt);

	/* write the embedded object list */
	mlOutEmbed(ge, cvt);

	/* add a newline after the header */
	cprintf(ge, cvt,"\n");
}

/****************************************************************************
 *
 * module:	m l O u t E n d e r
 */
static void
mlOutEnder(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	AFileEnderFromInfo(&cvt->info,cvt->buf);
	cprintf(ge, cvt,"%s\n",cvt->buf);
}

/***************************************************************************
 *
 * module:	m l O u t C a l l b a c k s
 */
static void
mlOutCallbacks(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt; 
{
	int ii,jj;
	ge_callback_type *cback;
	char *arg;
    int len;
    tx_char tbuff[400];

	if (ge->n_callbacks)
	{
		cprintf(ge, cvt,"%s\n", tok[TCBACK]);
		for (ii=0; ii<ge->n_callbacks; ii++)
		{
			/* write the macro */
			cback = &ge->callback[ii];
			cprintf(ge, cvt," %s <", tok[TMACRO]);
			len = MultiToWideStr(ge, cback->name, tbuff );
			mlOutString(ge,  len, tbuff, cvt, FALSE );
			cprintf(ge, cvt," %d %d %d>\n",
			        cback->on_drag,
			        cback->on_single,
			        cback->on_double);

			/* write the args */
			for ( jj=0; jj<cback->n_args; jj++ )
			{
				/* check for wrapping */
				cprintf(ge, cvt, " ");
				len = MultiToWideStr(ge, cback->args[jj], tbuff );
				mlOutString(ge, len, tbuff, cvt, FALSE);
				cprintf(ge, cvt, "\n");
			}
		}
		cprintf(ge, cvt,"%s %s\n\n", tok[TEND], tok[TCBACK]);
	}
}

/************************************************************************
 *
 * module:	m l O u t S t y l e s
 */
static void
mlOutStyles(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt; 
{
}

/************************************************************************
 *
 * module:	m l M a p T o 1 2 5
 */
static void
mlMapTo125(ge,cvt, src_t, dst_t)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *src_t, *dst_t;
{
	unsigned char *srcP, *dstP, *outCmap;
	int size, ii, cmap_ix;
	unsigned int r,g,b;
	cmap_type *srcCmap;

	if ( src_t->type == IMAGE && dst_t->type == IMAGE )
	{
		srcP = src_t->data.img->info.data;
		dstP = dst_t->data.img->info.data;
		cmap_ix = src_t->data.img->cmapID;
		srcCmap = &ge->cmap[cmap_ix];

		/* start by making a map from the input colormap */
		outCmap = (unsigned char*)gloc(ge, srcCmap->n_colors);
		AxSet125PixVal(NULL, FALSE, NULL, NULL, NULL, 16, TRUE);
		for (ii=srcCmap->n_colors; ii>=0; --ii)
		{
			if (srcCmap->color[ii].see_thru)
				outCmap[ii] = 124; /* set to white */
			else
			{
				ExtractRGBCmapEntry(ge, cvt, cmap_ix, ii, &r, &g, &b);
				AxSet125PixVal(outCmap+ii,FALSE,r,g,b,0,FALSE);
			}
		}
		size = src_t->data.img->info.hyt * src_t->data.img->info.data_byte_wid;
		for ( ii=size; ii>=0; --ii)
			dstP[ii] = outCmap[srcP[ii]];
		gfree(ge, outCmap);
	}
}

/****************************************************************************
 *
 * module:	W r i t e R A W - Write data in Raw format
 */
static int
WriteRAW(ge, cvt,llen,hyt,wid,imgP,func,mode)
graphics_main_type *ge;
cvt_global_type *cvt;
int llen,hyt, wid;
unsigned char *imgP;
int (*func)(), mode;
{
	int cwid;

	while (0 <= --hyt)
	{
		cwid = wid;
		while (0 <= --cwid)
		{
			llen += (*func)(ge, *imgP++ & 0xFF, llen, cvt);
			if (MAX_LINE < llen)
			{
				cputc(ge, cvt,'\n');
				cputc(ge, cvt,' ');
				llen = 0;
			}
		}
		llen += (*func)(ge, -1, llen, cvt);
	}
}

/****************************************************************************
 *
 * module:	p a c k R U N - Pack a scanline of `snum' bytes into a buffer.
 */
static int
packRUN(ge, inP,outP,inWid,outWid,buffer)
graphics_main_type *ge;
char *inP, *outP;
int inWid,*outWid;
char *buffer; /* temp space */
{
	register char *buf;
	register int nbuf, rstart;
	int i, mode;
	char c, lastc, *startP;
	
	startP = outP;
	buf = buffer + 1;
	buf[0] = lastc = c = *inP++;
	nbuf = 1;
	rstart = 0;
	mode = DMP;

	for (i=1; i<inWid; i++)
	{
		buf[nbuf++] = c = *inP++;

		switch (mode) {

		case DMP:
			if (MAXDMP < nbuf)
			{
				buffer[0] = nbuf - 2;
				memcpy(outP,buffer,nbuf);
				outP += nbuf;
				/* check for outP overflow */
				buf[0] = c;
				nbuf = 1;
				rstart = 0;
				break;
			}

			if (c != lastc)
			{
				rstart = nbuf - 1;	/* 1st of poss run */
				break;
			}

			if (rstart && nbuf - rstart < MINRUN)
				break;

			if (0 < rstart)
			{
				buffer[0] = rstart - 1;
				memcpy(outP,buffer,rstart + 1);
				outP += (rstart + 1);
				/* check for outP overflow */
			}

			mode = RUN;
			break;

		case RUN:
			if (c == lastc && nbuf - rstart <= MAXRUN)
				break;

			*outP++ = rstart - nbuf + 2;
			*outP++ = lastc;
			/* check for outP overflow */
			buf[0] = c;
			nbuf = 1;
			rstart = 0;
			mode = DMP;
			break;
		}

		lastc = c;
	}

	if (mode == DMP)
	{
		buffer[0] = nbuf - 1;
		memcpy(outP,buffer,nbuf + 1);
		outP += (nbuf + 1);
		/* check for outP overflow */
	}
	else
	{
		*outP++ = rstart - nbuf + 1;
		*outP++ = lastc;
		/* check for outP overflow */
	}

	*outWid = (int)(outP - startP);
	return inWid;
}

/****************************************************************************
 *
 * module:	W r i t e R U N - Write data in RUN LENGTH ENCODED format
 */
static int
WriteRUN(ge, cvt,llen,hyt,wid,imgP,func,mode)
graphics_main_type *ge;
cvt_global_type *cvt;
int llen,hyt, wid;
unsigned char *imgP;
int (*func)(), mode;
{
	int cmpWid;
	unsigned char *outP, *pP, *tmpBuf;

	outP = (unsigned char*)gloc(ge, 2*wid); /* worst case */
	memset(outP, 0x0, 2*wid);
	tmpBuf = (unsigned char*)gloc(ge,260); /* max */
	memset(tmpBuf, 0x0, 260);

	while (0 <= --hyt)
	{
		if (packRUN(ge, imgP,outP,wid,&cmpWid,tmpBuf) != wid)
			return(!AOK);
		imgP += wid;
		pP = outP;
		while (0 <= --cmpWid)
		{
			llen += (*func)(ge, *pP++, llen, cvt);
			if (MAX_LINE < llen)
			{
				cputc(ge, cvt,'\n');
				cputc(ge, cvt,' ');
				llen = 0;
			}
		}
		llen += (*func)(ge, -1, llen, cvt);
	}

	gfree(ge, outP);
	gfree(ge, tmpBuf);
	return(AOK);
}

/************************************************************************
 *
 * module:	m l O u t I m a g e - ascii-ize an image
 */
static void
mlOutImage(ge, t, cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	char *epsP;
	int llen, ii, (*encodeFunc)(), (*compressFunc)();

	mlOutPath(ge, t->paths, t->n_paths, cvt);
	mlOutTrans(ge, &t->data.img->trans, cvt);

	/* if there is an EPS image, write it out. */
	if (t->data.img->info.eps_seg != NULL)
	{
		/* the width, height, size and a leading space */
		cprintf(ge, cvt," %s %d %d %d\n ", tok[TEPS],
			t->data.img->info.wid,
			t->data.img->info.hyt,
			t->data.img->info.eps_seg_size);

		llen = t->data.img->info.eps_seg_size;
		epsP = t->data.img->info.eps_seg;
		while ( *epsP != '\0')
		{
			cputc(ge, cvt, *epsP);
			if (*epsP == '\n')
				cputc(ge, cvt, ' ');
			epsP++;
		}

		/* the end of the EPS segment */
		cprintf(ge, cvt,"%s %s\n",tok[TEND], tok[TEPS]);
	}

	cprintf(ge, cvt," %s %d", tok[TWIDTH], t->data.img->info.wid);
	cprintf(ge, cvt," %s %d", tok[THEIGHT], t->data.img->info.hyt);
	cprintf(ge, cvt," %s %d", tok[TDEPTH], t->data.img->info.dep);


	if ( cvt->compression != COMP_LZW)
		cprintf(ge, cvt," %s %s", tok[TCOMPRESSION], cmp[cvt->compression-1]);

	if ((cvt->encoding != BIT6) || cvt->as_1_1)
		cprintf(ge, cvt," %s %s", tok[TENCODING], enc[cvt->encoding]);
	
	encodeFunc = (cvt->encoding == BIT6) ? mailOutBit6 : mailOutHex;

	if (cvt->compression == COMP_RAW)
		compressFunc = WriteRAW;
	else if (cvt->compression == COMP_RUN)
		compressFunc = WriteRUN;
	else
		compressFunc = ReadWriteLZW;

	llen = cprintf(ge, cvt,"\n %s ", tok[TDATA]);

	/* DO IT! */
	(*compressFunc)(ge, cvt,llen,t->data.img->info.hyt,
		t->data.img->info.data_byte_wid,
		t->data.img->info.data,encodeFunc, FALSE);
	if ((t->data.img->info.dep == 8) && (t->data.img->info.mask))
	{
		llen = cprintf(ge, cvt,"\n %s ", tok[TMASK]);
		(*compressFunc)(ge, cvt,llen,t->data.img->info.hyt,
			t->data.img->info.mask_byte_wid,
			t->data.img->info.mask,encodeFunc,FALSE);
	}

	cputc(ge, cvt,'\n');
	return;
}

/****************************************************************************
 *
 * module:  mlOutNumberFormat
 */
static void
mlOutNumberFormat(ge,cvt,fmt)
graphics_main_type *ge;
cvt_global_type *cvt;
number_format *fmt;
{
    tx_char tbuff[500];
	int ii, len;

	/* format info */
	mlOutIndent(ge, cvt);
	cvt->linepos += cprintf(ge, cvt, "%s <", tok[TNUM_FMT]);
	cvt->linepos += cprintf(ge, cvt,"%d ", fmt->id);
	cvt->linepos += cprintf(ge, cvt,"%d ", fmt->aux_id);
	cvt->linepos += cprintf(ge, cvt,"%d ", fmt->display_format_errors);
	cvt->linepos += cprintf(ge, cvt,"%d ", fmt->trim_string);
	cvt->linepos += cprintf(ge, cvt,"%d ", fmt->units);
	cvt->linepos += cprintf(ge, cvt,"%d ", fmt->n_radix_places);

	/* the prefix and suffix strings */
	if ( !fmt->prefix )
		cvt->linepos += cprintf(ge, cvt,"\"\" ");
	else
	{
		len = MultiToWideStr(ge, fmt->prefix, tbuff );
		cvt->linepos += mlOutString(ge, len, tbuff, cvt, FALSE);
		cvt->linepos += cprintf(ge, cvt," ");
	}
	if ( !fmt->suffix )
		cvt->linepos += cprintf(ge, cvt,"\"\" ");
	else
	{
		len = MultiToWideStr(ge, fmt->suffix, tbuff );
		cvt->linepos += mlOutString(ge, len, tbuff, cvt, FALSE);
		cvt->linepos += cprintf(ge, cvt," ");
	}

	/* am and pm strings */
	if ( !fmt->am )
		cvt->linepos += cprintf(ge, cvt,"\"\" ");
	else
	{
		len = MultiToWideStr(ge, fmt->am, tbuff );
		cvt->linepos += mlOutString(ge, len, tbuff, cvt, FALSE);
		cvt->linepos += cprintf(ge, cvt," ");
	}
	if ( !fmt->pm )
		cvt->linepos += cprintf(ge, cvt,"\"\" ");
	else
	{
		len = MultiToWideStr(ge, fmt->pm, tbuff );
		cvt->linepos += mlOutString(ge, len, tbuff, cvt, FALSE);
		cvt->linepos += cprintf(ge, cvt," ");
	}

	/* the boolean strings */
	if ( !fmt->true )
		cvt->linepos += cprintf(ge, cvt,"\"\" ");
	else
	{
		len = MultiToWideStr(ge, fmt->true, tbuff );
		cvt->linepos += mlOutString(ge, len, tbuff, cvt, FALSE);
		cvt->linepos += cprintf(ge, cvt," ");
	}
	if ( !fmt->false )
		cvt->linepos += cprintf(ge, cvt,"\"\" ");
	else
	{
		len = MultiToWideStr(ge, fmt->false, tbuff );
		cvt->linepos += mlOutString(ge, len, tbuff, cvt, FALSE);
		cvt->linepos += cprintf(ge, cvt," ");
	}

	/* the separator characters */
	len = 1;
	tbuff[1] = (tx_char)0;
	tbuff[0] = (tx_char)fmt->t_sep;
	cvt->linepos += mlOutString(ge, len, tbuff, cvt, FALSE);
	cvt->linepos += cprintf(ge, cvt," ");
	tbuff[0] = (tx_char)fmt->d_sep;
	cvt->linepos += mlOutString(ge, len, tbuff, cvt, FALSE);
	cvt->linepos += cprintf(ge, cvt,">\n");

	/* done */
	cvt->linepos = 0;
}

/****************************************************************************
 *
 * module:  mlOutChartLegendType
 */
static void
mlOutChartLegendType (ge,cvt,legend)
graphics_main_type *ge;
cvt_global_type *cvt;
chart_legend_type *legend;
{
    tx_char tbuff[500];
    int len;

	if (!legend)
		return;

	/* the legend token and the optional legend title */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s ",tok[TLEGEND]);
	if (!legend->title_str)
		cprintf(ge, cvt,"\"\"");
	else
	{
        len = MultiToWideStr(ge,  legend->title_str, tbuff );
        mlOutString(ge,  len, tbuff, cvt, FALSE );
	}
	cprintf(ge, cvt,"\n");
	cvt->linepos = 0;

	/* the legend position */
	cvt->indent++;
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt, "%s <",   tok[TLEGALIGN]);
	cprintf(ge, cvt, "%d %d ", legend->h_align, legend->h_off);
	cprintf(ge, cvt, "%d %d ", legend->v_align, legend->v_off);
	cprintf(ge, cvt, "%d>\n",  legend->title_align);
	cvt->linepos = 0;

	/* the legend format */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt, "%s ",   tok[TLEGFMT]);
	cprintf(ge, cvt, "<%d",   legend->disabled);
	cprintf(ge, cvt, " %d",   legend->x_space_eating);
	cprintf(ge, cvt, " %d",   legend->y_space_eating);
	cprintf(ge, cvt, " %d",   legend->text_before_sample);
	cprintf(ge, cvt, " %d",   legend->arranged_by_row);
	cprintf(ge, cvt, " %d>\n",legend->max_per_row_or_column);
	cvt->linepos = 0;

	/* legend margins */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt, "%s ",   tok[TLEGMARG]);
	cprintf(ge, cvt, "<%d",   legend->t_margin);
	cprintf(ge, cvt, " %d",   legend->b_margin);
	cprintf(ge, cvt, " %d",   legend->l_margin);
	cprintf(ge, cvt, " %d",   legend->r_margin);
	cprintf(ge, cvt, " %d",   legend->row_margin);
	cprintf(ge, cvt, " %d ",  legend->column_margin);
	cprintf(ge, cvt, " %d>\n",legend->auto_margins);

	/* the attributes of the various parts of the legend */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n",tok[TLEGBOXATTR]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &legend->box_attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &legend->box_attr, 0, cvt);
		cvt->indent--;
	}

	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n",tok[TLEGLABATTR]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &legend->label_attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &legend->label_attr, 0, cvt);
		cvt->indent--;
	}

	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n",tok[TLEGTITATTR]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &legend->title_attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &legend->title_attr, 0, cvt);
		cvt->indent--;
	}

	/* done with legends */
	cvt->indent--;
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s %s\n",tok[TEND],tok[TLEGEND]);
	cvt->linepos = 0;
}

/****************************************************************************
 *
 * module:  mlOutChartDataType
 */
static void
mlOutChartDataType(ge,cvt,dat)
graphics_main_type *ge;
cvt_global_type *cvt;
chart_data_point_type *dat;
{
	tx_char tbuff[500];
	int len, ii;

	/* the data point token and the display type */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <",tok[TDAT]);
	switch(dat->display_flag)
	{
		default:
		case BG_PLOT: cprintf(ge, cvt, "%d ", BG_PLOT); break;
		case BG_SUBS: cprintf(ge, cvt, "%d ", BG_SUBS); break;
		case BG_SPAN: cprintf(ge, cvt, "%d ", BG_SPAN); break;
		case BG_GAP:  cprintf(ge, cvt, "%d ", BG_GAP);  break;
		case BG_NULL: cprintf(ge, cvt, "%d ", BG_NULL); break;
	}

	/* the data point label */
	if (!dat->label_str)
		cprintf(ge, cvt,"\"\"");
	else
	{
		len = MultiToWideStr(ge,  dat->label_str, tbuff );
		mlOutString(ge,  len, tbuff, cvt, FALSE );
	}

	/* the data label offsets */
	if ( dat->local_x )
		cprintf(ge, cvt," %d", dat->label_x_off);
	else
		cprintf(ge, cvt," \"\"");
	if ( dat->local_y )
		cprintf(ge, cvt," %d", dat->label_y_off);
	else
		cprintf(ge, cvt," \"\"");
	cprintf(ge, cvt, ">\n");
	cvt->linepos = 0;

	/* data point x position and auxiliary data */
	cvt->indent++;
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt, "%s <", tok[TX]);
	for ( ii=0; ii<dat->nx; ii++ )
	{
		if ( ii != 0 )
			cprintf(ge, cvt, " ");
		cprintf(ge, cvt,"%f",dat->x[ii]);
	}
	cprintf(ge, cvt, ">\n");
	cvt->linepos = 0;

	/* data point y position and auxiliary data */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt, "%s <", tok[TY]);
	for ( ii=0; ii<dat->ny; ii++ )
	{
		if ( ii != 0 )
			cprintf(ge, cvt, " ");
		cprintf(ge, cvt,"%f",dat->y[ii]);
	}
	cprintf(ge, cvt, ">\n");
	cvt->linepos = 0;

	/* data point z position and auxiliary data */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt, "%s <", tok[TZ]);
	for ( ii=0; ii<dat->nz; ii++ )
	{
		if ( ii != 0 )
			cprintf(ge, cvt, " ");
		cprintf(ge, cvt,"%f",dat->z[ii]);
	}
	cprintf(ge, cvt, ">\n");
	cvt->linepos = 0;
	cvt->indent--;
}

/****************************************************************************
 *
 * module:  mlOutChartDataGroupType
 */
static void
mlOutChartDataGroupType(ge,cvt,grp)
graphics_main_type *ge;
cvt_global_type *cvt;
chart_data_group_type *grp;
{
	tx_char tbuff[500];
	int len, ii;

	/* the beginning of the data group and the group name */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <", tok[TDATGROUP]);
    len = MultiToWideStr(ge, grp->name, tbuff);
    mlOutString(ge, len, tbuff, cvt, FALSE);
	cprintf(ge, cvt, " %d>\n", grp->disabled);
	cvt->linepos = 0;

	/* the legend string */
	cvt->indent++;
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s ", tok[TLABEL]);
	if ( !grp->legend_str )
		cprintf(ge, cvt,"\"\"");
	else
	{
		len = MultiToWideStr(ge, grp->legend_str, tbuff);
		mlOutString(ge, len, tbuff, cvt, FALSE);
	}
	cprintf(ge, cvt,"\n");
	cvt->linepos = 0;

	/* data type (line, bar, etc...), stack id, and custom string */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <%d %d ", tok[TDATATYPE], grp->type, grp->stack_id);
	if ( !grp->part )
		cprintf(ge, cvt,"\"\"");
	else
	{
		len = MultiToWideStr(ge, grp->part, tbuff);
		mlOutString(ge, len, tbuff, cvt, FALSE);
	}
	cprintf(ge, cvt," %d %d", grp->new, grp->old_ix);
	cprintf(ge, cvt," %d", grp->disabled);
	cprintf(ge, cvt,">\n");
	cvt->linepos = 0;

	/* the axes this data is drawn against */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <%d %d %d>\n", tok[TAXES], grp->x_axis,
	                                   grp->y_axis,
	                                   grp->z_axis);
	cvt->linepos = 0;

	/* data label info */
	mlOutNumberFormat(ge,cvt,&grp->label_frmt);
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt, "%s <%d %d %d %d>\n", tok[TDATALABEL],
	                                   grp->data_label_type,
	                                   grp->label_x_off,
	                                   grp->label_y_off,
	                                   grp->label_align);
	cvt->linepos = 0;

	/* write each data point, calculating grp->n at read time */
	for (ii=0; ii<grp->n; ii++)
		mlOutChartDataType(ge,cvt,&grp->dat[ii]);

	/* write the data label and data attributes */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n", tok[TGRPLABATTR]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &grp->label_attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &grp->label_attr, 0, cvt);
		cvt->indent--;
	}

	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n", tok[TGRPATTR]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &grp->attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &grp->attr, 0, cvt);
		cvt->indent--;
	}

	/* the end of the data group */
	cvt->indent--;
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s %s\n", tok[TEND], tok[TDATGROUP]);
	cvt->linepos = 0;
}

/****************************************************************************
 *
 * module:  mlOutChartAxisType
 */
static void
mlOutChartAxisType(ge,cvt,axis)
graphics_main_type *ge;
cvt_global_type *cvt;
chart_axis_type *axis;
{
    tx_char tbuff[500];
	int ii, len;

	/* type and space consumed (for additional axes) */
	cvt->indent++;
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <",   tok[TTIK_TYPE]);
	cprintf(ge, cvt,"%d ",    axis->type);
	cprintf(ge, cvt,"%d ",    axis->x_offset);
	cprintf(ge, cvt,"%d ",    axis->y_offset);
	cprintf(ge, cvt,"%d>\n",  axis->tik_margin);
	cvt->linepos = 0;

	/* tik and label alignment */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <",  tok[TTIK_ALIGN]);
	cprintf(ge, cvt,"%d ",   axis->label_align);
	cprintf(ge, cvt,"%d ",   axis->tik_align);
	cprintf(ge, cvt,"%d ",   axis->tight_label);
	cprintf(ge, cvt,"%d>\n", axis->tight_tiks);
	cvt->linepos = 0;

	/* axis format info */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <",  tok[TAXIS_FMT]);
	cprintf(ge, cvt,"%d ",   axis->hidden);
	cprintf(ge, cvt,"%d ",   axis->auto_max);
	cprintf(ge, cvt,"%d ",   axis->auto_min);
	cprintf(ge, cvt,"%d ",   axis->auto_bas);
	cprintf(ge, cvt,"%d ",   axis->auto_n_major_tiks);
	cprintf(ge, cvt,"%d ",   axis->auto_n_minor_tiks);
	cprintf(ge, cvt,"%d ",   axis->major_tik_format);
	cprintf(ge, cvt,"%d ",   axis->minor_tik_format);
	cprintf(ge, cvt,"%d ",   axis->pos);
	cprintf(ge, cvt,"%f ",   axis->pos_val);
	cprintf(ge, cvt,"%f ",   axis->min_val);
	cprintf(ge, cvt,"%f ",   axis->max_val);
	cprintf(ge, cvt,"%f ",   axis->bas_val);
	cprintf(ge, cvt,"%f ",   axis->first_delta);
	cprintf(ge, cvt,"%d ",   axis->invalid_tiks);
	cprintf(ge, cvt,"%d ",   axis->draw_behind);
	cprintf(ge, cvt,"%d>\n", axis->use_tik_attr);	
	cvt->linepos = 0;

	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s ",  tok[TAXIS_MAJOR]);
	cprintf(ge, cvt,"<%d %d %f %f>\n", axis->major_sz,
	                                   axis->n_maj_steps,
	                                   axis->maj_delta,
	                                   axis->maj_step);
	cvt->linepos = 0;
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s ",  tok[TAXIS_MINOR]);
	cprintf(ge, cvt,"<%d %d %f %f>\n", axis->minor_sz,
	                                   axis->n_min_steps,
	                                   axis->min_delta,
	                                   axis->min_step);
	cvt->linepos = 0;

	/* numerical format for tiks */
	mlOutNumberFormat(ge,cvt,&axis->tik_format);

	/* bar info */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <%f %f>\n", tok[TBAR_INFO],
	                                axis->bar_margin,
	                                axis->bar_overlap);
	cvt->linepos = 0;

	/* tik label filter (for catagory axes) */
	cprintf(ge, cvt,"  %s %d\n", tok[TLABEL_FILTER], axis->filter_labels);

	/* tik labels (for catagory axes) */
	for ( ii=0; ii<axis->n; ii++ )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt, "%s <", tok[TLABEL]);
		if ( !axis->tik_str[ii] )
			cprintf(ge, cvt,"\"\"");
		else
		{
			len = MultiToWideStr(ge, axis->tik_str[ii], tbuff);
			mlOutString(ge, len, tbuff, cvt, FALSE);
		}
		if ( axis->display )
			cprintf(ge, cvt, " %d", axis->display[ii]);
		else
			cprintf(ge, cvt, " 1");
		cprintf(ge, cvt,">\n");
		cvt->linepos = 0;
	}
 
	/* axis attribute info */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n", tok[TAXATTR]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &axis->attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &axis->attr, 0, cvt);
		cvt->indent--;
	}
 
 
	/* axis tik attribute info */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n", tok[TAXTIKATTR]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &axis->tik_attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &axis->tik_attr, 0, cvt);
		cvt->indent--;
	}

	
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n", tok[TAXLABATTR]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &axis->label_attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &axis->label_attr, 0, cvt);
		cvt->indent--;
	}
 
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n", tok[TAXTIKLABATTR]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &axis->tik_label_attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &axis->tik_label_attr, 0, cvt);
		cvt->indent--;
	}

	/* done */
	cvt->indent--;
}

/****************************************************************************
 *
 * module:  mlOutChartData
 */
static void
mlOutChartData(ge,cvt,chart)
graphics_main_type *ge;
cvt_global_type *cvt;
register chart_type *chart;
{
	register int ii, len;
	tx_char tbuff[500];

	/* chart token and name of chart */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s ", tok[TCHART]);
	len = MultiToWideStr(ge, chart->name, tbuff);
	mlOutString(ge, len, tbuff, cvt, FALSE);
	cprintf(ge, cvt,"\n");
	cvt->linepos = 0;
	cvt->indent++;

	/* chart type and orientation */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <%d %d %d>\n", tok[TTYPE], 
								chart->type, 
								chart->orientation,
								chart->old_orient);
	cvt->linepos = 0;

	/* title, subtitle, and footer strings */
	if ( chart->title )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s ", tok[TTITLE]);
		len = MultiToWideStr(ge, chart->title, tbuff );
		mlOutString(ge,  len, tbuff, cvt, FALSE );
		cprintf(ge, cvt,"\n");
		cvt->linepos=0;
	}
	if ( chart->subtitle )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt," %s ", tok[TSUBTITLE]);
		len = MultiToWideStr(ge, chart->subtitle, tbuff );
		mlOutString(ge,  len, tbuff, cvt, FALSE );
		cprintf(ge, cvt,"\n");
		cvt->linepos=0;
	}
	if ( chart->footer )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt," %s ", tok[TFOOT]);
		len = MultiToWideStr(ge,  chart->footer, tbuff );
		mlOutString(ge,  len, tbuff, cvt, FALSE );
		cprintf(ge, cvt,"\n");
		cvt->linepos=0;
	}
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <%d %d %d %d %d %d>\n", tok[TTITLE_ALIGN],
									    chart->title_align,
	                               	         chart->subtitle_align,
	                               		    chart->footer_align,
								   	    chart->title_expand,
	                               	         chart->subtitle_expand,
	                               	         chart->footer_expand);
	cvt->linepos=0;

	/* chart decorations (closing lines and grid lines) */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt, "%s <%d %d %d %d %d %d>\n", tok[TCLOSE], 
					   				     chart->close_top,
	                               	          chart->close_bot,
	                               	          chart->close_lft,
	                                             chart->close_rt,
									     chart->close_bg,
									     chart->close_fg);
	cvt->linepos=0;
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <%d %d %d %d %d %d>\n", tok[TGRIDS], 
								   chart->maj_x_grids,
 								   chart->min_x_grids,
	                        	             chart->maj_y_grids,
	                        	             chart->min_y_grids,
								   chart->maj_z_grids,
								   chart->min_z_grids);
	cvt->linepos=0;
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <%d %d %d %d %d %d %d %d>\n", tok[TOFFSET], 
										     chart->def_x_offset,
	                                                  chart->def_y_offset,
											chart->title_x_off,
											chart->title_y_off,		
											chart->subtitle_x_off,
											chart->subtitle_y_off,
											chart->footer_x_off,
											chart->footer_y_off);
	cvt->linepos=0;

	/* chart size and margins */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <%d %d %d %d>\n", tok[TEXTENT], 
								   chart->extent.x1,
	                                      chart->extent.y1,
	                                      chart->extent.x2,
	                                      chart->extent.y2);
	cvt->linepos=0;
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <%d %d %d %d>\n", tok[TMARGIN], chart->margin.x1,
	                                      chart->margin.y1,
	                                      chart->margin.x2,
	                                      chart->margin.y2);
	cvt->linepos=0;
 
	/* count of original axes and groups (as of last rendering) */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <%d %d %d %d %d>\n", tok[TCHART_COUNTS],
	                                         chart->old_n_x_axes,
	                                         chart->old_n_y_axes,
	                                         chart->old_n_z_axes,
	                                         chart->old_n_groups,
	                                         chart->old_type);
	cvt->linepos=0;

	/* default null treatment */
	mlOutIndent(ge, cvt);
	switch ( chart->null_format )
	{
		default:
		case 0:cprintf(ge, cvt,"%s %s\n",  tok[TNULL_FMT], tok[TPLOT]); break;
		case 1: cprintf(ge, cvt,"%s %s\n", tok[TNULL_FMT], tok[TSUBS]); break;
		case 2: cprintf(ge, cvt,"%s %s\n", tok[TNULL_FMT], tok[TSPAN]); break;
		case 3: cprintf(ge, cvt,"%s %s\n", tok[TNULL_FMT], tok[TGAP]); break;
	}
	cvt->linepos=0;

	/* 3D */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt, "%s <%d %d %d %d %d %d %d %d %d %d %d %d>\n",
	        tok[TTHREE_D],
	        chart->three_d.enabled, chart->old_three_d.enabled,
	        chart->three_d.pitch, chart->old_three_d.pitch,
	        chart->three_d.yaw, chart->old_three_d.yaw,
	        chart->three_d.projection, chart->old_three_d.projection,
	        chart->three_d.depth, chart->old_three_d.depth,
			chart->three_d.right_angles, chart->old_three_d.right_angles);
	cvt->linepos=0;

	/* chart type specific info */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <%d %f %f %d %d %d %d>\n",tok[TCHART_INFO],
										chart->info.stacked_strata, 	
										chart->info.bar_margin,
										chart->info.bar_overlap,	
										chart->info.start_angle,
										chart->info.def_type,
										chart->data_src.type,
										chart->data_src.ix);
	cvt->linepos=0;		

	
	/* chart type misc info */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <%d>\n",tok[TMISC_CHART_INFO],
						   chart->needs_redraw);
								
	cvt->linepos=0;		

#if 0	
	/* dont bother to create a new token */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <%d %d>\n", tok[TDATA_INFO],
							  chart->data_src.type,
							  chart->data_src.ix);
	cvt->linepos=0;		
#endif	
	
	
	/* the axes */
	if ( chart->n_x_axes )
	{
		for ( ii=0; ii<chart->n_x_axes; ii++ )
		{
			mlOutIndent(ge, cvt);
			cprintf(ge, cvt,"%s ", tok[TX_AXIS]);
			if ( !chart->x_axis[ii].label_str )
				cprintf(ge, cvt,"\"\"");
			else
			{
				len = MultiToWideStr(ge, chart->x_axis[ii].label_str, tbuff );
				mlOutString(ge,  len, tbuff, cvt, FALSE );
			}
			cprintf(ge, cvt, "\n");
			cvt->linepos=0;
			mlOutChartAxisType(ge,cvt,&chart->x_axis[ii]);
			mlOutIndent(ge, cvt);
			cprintf(ge, cvt,"%s %s\n", tok[TEND], tok[TX_AXIS]);
			cvt->linepos=0;
		}
	}
	if ( chart->n_y_axes )
	{
		for ( ii=0; ii<chart->n_y_axes; ii++ )
		{
			mlOutIndent(ge, cvt);
			cprintf(ge, cvt,"%s ", tok[TY_AXIS]);
			if ( !chart->y_axis[ii].label_str )
				cprintf(ge, cvt,"\"\"");
			else
			{
				len = MultiToWideStr(ge, chart->y_axis[ii].label_str, tbuff );
				mlOutString(ge,  len, tbuff, cvt, FALSE );
			}
			cprintf(ge, cvt, "\n");
			cvt->linepos=0;
			mlOutChartAxisType(ge,cvt,&chart->y_axis[ii]);
			mlOutIndent(ge, cvt);
			cprintf(ge, cvt,"%s %s\n", tok[TEND], tok[TY_AXIS]);
			cvt->linepos=0;
		}
	}
	if ( chart->n_z_axes )
	{
		for ( ii=0; ii<chart->n_z_axes; ii++ )
		{
			mlOutIndent(ge, cvt);
			cprintf(ge, cvt,"%s ", tok[TZ_AXIS]);
			if ( !chart->z_axis[ii].label_str )
				cprintf(ge, cvt,"\"\"");
			else
			{
				len = MultiToWideStr(ge, chart->z_axis[ii].label_str, tbuff );
				mlOutString(ge,  len, tbuff, cvt, FALSE );
			}
			cprintf(ge, cvt, "\n");
			cvt->linepos=0;
			mlOutChartAxisType(ge,cvt,&chart->z_axis[ii]);
			mlOutIndent(ge, cvt);
			cprintf(ge, cvt,"%s %s\n", tok[TEND], tok[TZ_AXIS]);
			cvt->linepos=0;
		}
	}

	/* old_data type and related info such as old_ix*/
	for ( ii=0; ii<chart->old_n_groups; ii++ )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt, "%s <%d %d %d %d>\n", tok[TOLD_GRP], 
		        chart->old_grp[ii].n,
		        chart->old_grp[ii].type,
		        chart->old_grp[ii].stack_id,
		        chart->old_grp[ii].stack_ax);
	}
	cvt->linepos = 0;

	/* group data */
	for (ii=0; ii<chart->n_groups; ii++)
		mlOutChartDataGroupType(ge,cvt,&chart->grp[ii]);

	/* legend box */
	mlOutChartLegendType(ge,cvt,&chart->legend);

	/* write the default attributes */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n", tok[TDEFATR]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &chart->def_attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &chart->def_attr, 0, cvt);
		cvt->indent--;
	}

	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n", tok[TMAJHRZ]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &chart->maj_x_attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &chart->maj_x_attr, 0, cvt);
		cvt->indent--;
	}

	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n", tok[TMINHRZ]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &chart->min_x_attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &chart->min_x_attr, 0, cvt);
		cvt->indent--;
	}

	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n", tok[TMAJVRT]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &chart->maj_y_attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &chart->maj_y_attr, 0, cvt);
		cvt->indent--;
	}

	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n", tok[TMINVRT]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &chart->min_y_attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &chart->min_y_attr, 0, cvt);
		cvt->indent--;
	}
	
	if ( chart->three_d.enabled && chart->n_z_axes > 0 )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s\n", tok[TMAJDPT]);
		cvt->linepos = 0;
		if ( AttrCmp(ge, &chart->maj_z_attr, &cvt->curr_atts) )
		{
			cvt->indent++;
			mlOutAttrs(ge, &chart->maj_z_attr, 0, cvt);
			cvt->indent--;
		}

		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s\n", tok[TMINDPT]);
		cvt->linepos = 0;
		if ( AttrCmp(ge, &chart->min_z_attr, &cvt->curr_atts) )
		{
			cvt->indent++;
			mlOutAttrs(ge, &chart->min_z_attr, 0, cvt);
			cvt->indent--;
		}
	}
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n", tok[TTITLEATR]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &chart->title_attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &chart->title_attr, 0, cvt);
		cvt->indent--;
	}

	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n", tok[TSUBTITLEATR]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &chart->subtitle_attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &chart->subtitle_attr, 0, cvt);
		cvt->indent--;
	}

	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n", tok[TFOOTATR]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &chart->footer_attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &chart->footer_attr, 0, cvt);
		cvt->indent--;
	}

	/* done */
	cprintf(ge, cvt,"%s %s\n\n", tok[TEND], tok[TCHART]);
	cvt->linepos = 0;
}

/****************************************************************************
 *
 * module:	m l O u t C h a r t I n f o
 */
static void
mlOutChart(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	chart_type *chart;
	int i;

	for ( i=0; i<ge->n_charts; i++ )
	{
		chart = (chart_type *)&ge->chart[i];
		if ( chart )
			mlOutChartData(ge, cvt, chart);
	}
}

/****************************************************************************
 *
 * module:	G r a p h i c s W r i t e
 */
ag400Write(ge, cvt, outBuffer, outCount)
graphics_main_type *ge;
cvt_global_type *cvt;
char **outBuffer;
int *outCount;
{
	/* init */
	mlDumpTxtObjs(ge, ge->t);
	if ( outBuffer )
	{
		cvt->buffit = TRUE;
		cvt->outBuff = cvt->outBufPtr = (char*)gloc(ge, CVTBLKSIZE);
		cvt->outBufEndPtr = cvt->outBufPtr + CVTBLKSIZE;
	}
	else if (!(cvt->fp = (FILE *)cvtGetFP(ge, cvt->fileName, "w+")) )
		return(!AOK);
	InitTokens(ge);

	/* generic file header (*START GRAPHICS...) */
    cvt->info.encoding = 1; /* assume file IS encoded */
    cvt->info.version = 400;
    cvt->info.old_version = 420;
    cvt->info.min_version = 320;
    cvt->info.docType = AX_DOC_TYPE_ASTERX_GRAPHICS;
    cvt->info.content = 0; /* none */
	mlOutHeader(ge, cvt);

	/* self contained segments */
	mlOutColormap(ge, cvt);
	mlOutFaces(ge, cvt);
	mlOutCallbacks(ge, cvt);

	/* fills may reference colors */
	mlOutNamedFills(ge, cvt);

	/* styles may reference colors, faces or callbacks */
	mlOutStyles(ge, cvt);

	/* headers/footers may reference colors and faces */
	mlOutHdrsFtrs(ge, cvt);

	/* default state may reference styles, colors, faces or callbacks */
	mlOutSessionInfo(ge, cvt);

	/* charts may reference styles, colors, faces or callbacks */
	mlOutChart(ge, cvt);

	/* palet parts may reference styles, colors, faces or callbacks */
	mlOutPalet(ge, cvt);

	/* paint list may reference styles, colors, faces or callbacks */
	mlOutPages(ge, cvt);

	/* finish up with *END GRAPHICS */
	mlOutEnder(ge, cvt);

	/* done */
	if ( !outBuffer )
		cvtCloseFile(ge, cvt);
	else
	{
		*outBuffer = cvt->outBuff;
		*outCount = cvt->outBufPtr - cvt->outBuff;
	}
	return(AOK);
}
