/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: fsdith.c
   Applix ASC version: 7
   Date: Mon Nov 16 13:58:51 1998
*/
#if AXMS
#include <windows.h>
#endif

#if !FOR_FILTER
#include <thim_gen.h>
#endif


#define INIT_ERR_ROW 1
#define ERROR_PROP 1
#define SERPENTINE 1

#define FS_SCALE 1	/* 1024 */
#define HALF_FS_SCALE 0	/* 512 */

#define SET_XY_BIT(ptr,wid,col,row) \
    ptr[((wid)*(row))+((col)/8)] |= ((unsigned char)(((unsigned char)0x80 >> ((col)%8)) & 0xFF)) 

#define CLR_XY_BIT(ptr,wid,col,row) \
    ptr[((wid)*(row))+((col)/8)] &= ~((unsigned char)((0x80 >> ((col)%8)) & 0xFF)) 

/****************************************************************************
 *
 * module:  FloydSteinbergDither - dither an 8 bit buffer of
 *                                 intensities to 1 bit.
 *
 *  5 May 93  SJP  Rewrote to do serpentine scanning and random error
 *                 row initialization.  Based on Poskanzer.
 */
void
FloydSteinbergDither (inBuffer, inWidth, inHeight, alignSize, outBuffer)
short *inBuffer;
int inWidth, inHeight;
int alignSize;
unsigned char *outBuffer;
{
	int outByteWidth, inByteWidth;
    register int row, col;
    int endstop;
    int in_index;
    int fs_direction = 1;	/* 1=> left to right, 0=> right to left */
    register int sum;
	short *inbuf;
    short *this_err;
    short *next_err;
    short *tmp_err;	/* used when switching direction */
    register int threshold = 128 * FS_SCALE;
	unsigned char *outbuf;


	/* How is the output buffer aligned? (Add other values as necessary.) */
	if (alignSize == 16)
		outByteWidth = (((inWidth + 15) / 8) & ~1);
	else /* 8 */
		outByteWidth = (inWidth + 7) / 8;

    inByteWidth = inWidth;

    /* time/space tradeoff; this way we dont have to check at the end
       of each col */
    this_err = (short *)malloc((inByteWidth +2) * sizeof(short));
    next_err = (short *)malloc((inByteWidth +2) * sizeof(short));
    memset(this_err, '\0', (inByteWidth +2) * sizeof(short));
    memset(next_err, '\0', (inByteWidth +2) * sizeof(short));
#if INIT_ERR_ROW
#if AXMS
    srand((unsigned int)(time(0) ^ GetCurrentProcessId()));	/* seed */
#else
    srand((unsigned int)(time(0) ^ getpid()));	/* seed */
#endif
    for (col=0; col<inByteWidth+2; col++)
        this_err[col] = 63 - (rand() & 127);	/* fill array with
                                                   +/-0..FS_SCALE/8 */
#endif
    memset(outBuffer, '\0', inHeight * outByteWidth);

	inbuf = inBuffer;
	outbuf = outBuffer;
	for (row=0; row<inHeight; row++)
	{
        memset(next_err, 0, sizeof(short) * (inWidth +2));

        if ( fs_direction )
		{
            col = 0;		/* start at left */
            endstop = inWidth; /* end at right */
        }
        else
        {
            col = inWidth -1;	/* start at right */
            endstop = -1;	/* end at left */
        }
        do
		{
            /* note the col+1; col means different things when refering
               to inbuf than to this_err[]/next_err[].
               */
            in_index = (row * inWidth) + col;
#if ERROR_PROP
            sum = (inbuf[in_index] * FS_SCALE) + this_err[col +1];
#else
            sum = inbuf[in_index] * FS_SCALE;
#endif

			if ( sum > threshold) /* becomes white */
			{
#if 0	/* do nothing, already white (paper) */
                CLR_XY_BIT(outbuf, outByteWidth, col, row);
#endif
#if ERROR_PROP
                sum = sum - threshold - HALF_FS_SCALE - 127;
#endif
			}
            else
                SET_XY_BIT(outbuf, outByteWidth, col, row);

            if ( fs_direction )
            {
#if ERROR_PROP
                this_err[col + 2] += ( sum * 7 ) / 16;
                next_err[col    ] += ( sum * 3 ) / 16;
                next_err[col + 1] += ( sum * 5 ) / 16;
                next_err[col + 2] += ( sum     ) / 16;
#endif
                col++;
            }
            else
            {
#if ERROR_PROP
                this_err[col    ] += ( sum * 7 ) / 16;
                next_err[col + 2] += ( sum * 3 ) / 16;
                next_err[col + 1] += ( sum * 5 ) / 16;
                next_err[col    ] += ( sum     ) / 16;
#endif
                col--;
            }
		}
        while ( col != endstop);
        tmp_err = this_err;	/* advance to the 'next' error row */
        this_err = next_err;
        next_err = tmp_err;
#if SERPENTINE
        fs_direction = !fs_direction;	/* change direction */
#endif
	}
    free(this_err);
    free(next_err);
}


