/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: imo.c
   Applix ASC version: 25
   Date: Tue Jan  5 21:14:49 1999
*/
#include <aos.h>
#include <attr.h>
#include <atm.h>
#include <atmwin.h>
#include <axtext.h>
#include <atxint.h>
#include <img.h>
#include <ge.h>
#include <afileimg.h>
#include <cvt.h>
#include <stiorts.h>


typedef struct {
	unsigned char r,g,b;
}CMAP_ENTRY;

typedef struct {
	char magic[2];				/* "AX" */
	unsigned char wid[2];		/* MSB format only */
	unsigned char hyt[2];		/* hyt and wid byte aligned */
	unsigned char dep[2];		/* currently 1 or 8 */
	unsigned char hasMask;		/* TRUE if there is a mask section */
	unsigned char nColors[2];	/* should be 0 for dep 1 */
	CMAP_ENTRY *CMAP_entries;	/* nColors of these */
	unsigned char *image;		/* ((wid+7)/8)*hyt or wid*hyt bytes */
	unsigned char *mask;		/* ((wid+7)/8)*hyt bytes if present */
} IM;

/***************************************************************************/
static double percent_levels[] = {0.0, 50.0, 75.0, 87.0, 95.0, 100.0};
/***********************************************************************/
static void
create_Cmap(ge, cmap)
graphics_main_type *ge;
cmap_type *cmap;
{
	int r, g, b, ix, pix_ix, c, m, y, k;

	/* create the colormap */
	cmap->n_colors = 256;
	cmap->color = (color_type *)gloc(ge, 256 * sizeof(color_type));

	/* load the transparent color */
	cmap->color[0].see_thru = TRUE;
	cmap->color[0].ink_type = 0;
	cmap->color[0].c = 0;
	cmap->color[0].m = 0;
	cmap->color[0].y = 0;
	cmap->color[0].k = 0;
	cmap->color[0].name = (char *)gloc(ge, 12);
	strcpy(cmap->color[0].name, NXLT("Transparent"));

	/* load black */
	cmap->color[1].see_thru = FALSE;
	cmap->color[1].ink_type = 0;
	cmap->color[1].c = 0;
	cmap->color[1].m = 0;
	cmap->color[1].y = 0;
	cmap->color[1].k = 255;
	cmap->color[1].name = (char *)gloc(ge, 6);
	strcpy(cmap->color[1].name, NXLT("Black"));

	/* white in slot 2 */
	cmap->color[2].see_thru = FALSE;
	cmap->color[2].ink_type = 0;
	cmap->color[2].c = 0;
	cmap->color[2].m = 0;
	cmap->color[2].y = 0;
	cmap->color[2].k = 0;
	cmap->color[2].name = (char *) gloc(ge, 6);
	strcpy(cmap->color[2].name, NXLT("White"));

	/* load the rest of the grey values */
	for ( ix=3, r=1; r<5; r++ )
	{
		cmap->color[ix].see_thru = FALSE;
		cmap->color[ix].ink_type = 0;
		cmap->color[ix].c = 0;
		cmap->color[ix].m = 0;
		cmap->color[ix].y = 0;
		cmap->color[ix].k = 255 - (int)percent_levels[r] *255/100.0;
		cmap->color[ix].name = (char *)gloc(ge, 9);
		sprintf(cmap->color[ix].name, XLT("color%03d"), ix);
		ix++;
	}

	/* load the remaining active colormap entries */
	for ( r=0; r<6; r++ )
	{
		for ( g=0; g<6; g++ )
		{
			for ( b=0; b<6; b++ )
			{
				/* don't redo the grey values */
				if ( r == g && g == b )
					continue;

				/* a non-tranparent process color */
				cmap->color[ix].see_thru = FALSE;
				cmap->color[ix].ink_type = 0;

				/* calc the cmy values */
				c = 255.0 - percent_levels[r] * 255.0 / 100.0;
				m = 255.0 - percent_levels[g] * 255.0 / 100.0;
				y = 255.0 - percent_levels[b] * 255.0 / 100.0;

				/* calc the black value */
				if ( c == 0 || m == 0 || y == 0 )
					k = 0;
				else
				{
					if ( c < m )
						k = (c < y) ? c : y;
					else
						k = (m < y) ? m : y;
					k = k/2;
				}

				/* set the cmyk values */
				cmap->color[ix].c = MAX(0, c-k);
				cmap->color[ix].m = MAX(0, m-k);
				cmap->color[ix].y = MAX(0, y-k);
				cmap->color[ix].k = k;

				/* name the color */
				cmap->color[ix].name = (char *)gloc(ge, 9);
				sprintf(cmap->color[ix].name, XLT("color%03d"), ix);
				ix++;
			}
		}
	}

	/* load the rest of the table with "white" */
	for (; ix < 256; ix++)
	{
		cmap->color[ix].see_thru = FALSE;
		cmap->color[ix].ink_type = 0;
		cmap->color[ix].c = 0;
		cmap->color[ix].m = 0;
		cmap->color[ix].y = 0;
		cmap->color[ix].k = 0;
		cmap->color[ix].name = (char *)gloc(ge, 9);
		sprintf(cmap->color[ix].name, XLT("color%03d"), ix);
	}
}

/***************************************************************************/
outVersionInfo(ge, cvt )
graphics_main_type *ge;
cvt_global_type *cvt;
{
	char *buf;
	int len;

	buf = (char *)gloc(ge, 80);
	AGetDefaultFileInfo(&cvt->info);
	cvt->info.docType = AX_DOC_TYPE_ASTERX_BITMAP;
	AFileHeaderFromInfo(&cvt->info, buf);
	cprintf(ge, cvt,"%s\n",buf);
	gfree(ge, buf);
}
/***************************************************************************/
outEnderInfo(ge, cvt )
graphics_main_type *ge;
cvt_global_type *cvt;
{
	char *buf;
	buf = (char *)gloc(ge, 80);
	AGetDefaultFileInfo(&cvt->info);
	cvt->info.docType = AX_DOC_TYPE_ASTERX_BITMAP;
	AFileEnderFromInfo(&cvt->info,buf);
	cprintf(ge, cvt,"\n%s\n",buf); 
	gfree(ge, buf);
}

/****************************************************************************
 *
 * module:	i m W r i t e
 */
imWrite(ge, cvt, outBuffer, outCount)
graphics_main_type *ge;
cvt_global_type *cvt;
char **outBuffer;
int *outCount;
{
	int status;

	int ii, colormapID, nColors, has_mask, byteWid, hyt,maskWid, wid, c,m,y,k;
	int dep,see_thru,ink_type,size;
	int x_spot,y_spot;
	unsigned char *imgPtr, magic[2];
	cmap_type temp_cmap,*cmap;
	tx_char tbuff[200];
	int len,same;
	thing *t;

	if ( outBuffer )
	{
		cvt->buffit = TRUE;
		cvt->outBuff = cvt->outBufPtr = (char*)gloc(ge, CVTBLKSIZE);
		cvt->outBufEndPtr = cvt->outBufPtr + CVTBLKSIZE;
	}
	else if ( !(cvt->fp = (FILE *)cvtGetFP(ge, cvt->fileName,"w+")) )
		return (!AOK);

	/* check that last object is an image */
	t = cvt->lastobj = ge->t->data.group->list[ge->t->data.group->num-1];
	if (t->type != IMAGE)
		return (!AOK);

	wid = t->data.img->info.wid;
	hyt = t->data.img->info.hyt;
	dep = t->data.img->info.dep;
	byteWid = t->data.img->info.data_byte_wid;
	x_spot = y_spot = 0;

	/* New Magic Line */
	outVersionInfo(ge, cvt);

	/* width  */
	cprintf(ge, cvt,"WIDTH  %d\n",wid);

	/* height */
	cprintf(ge, cvt,"HEIGHT  %d\n",hyt);

	/* depth  */
	cprintf(ge, cvt,"DEPTH  %d\n",dep);

	/* Mask Flag */
	if ((dep > 1) && (t->data.img->info.mask))
			has_mask = TRUE;
	else
			has_mask = FALSE;

	if ( ge->hotspot.x != 0 && ge->hotspot.y != 0 )
		cprintf(ge, cvt,"HOTSPOT (%d,%d)\n",ge->hotspot.x, ge->hotspot.y);

	/* Number of Colors */
	if (dep == 8)
	{
		colormapID = t->data.img->cmapID;
		create_Cmap(ge, &temp_cmap);
		same = foundCmap(ge, &ge->cmap[colormapID],&temp_cmap);
		if ( same == FALSE )
		{
			cmap = &ge->cmap[colormapID];
			nColors = cmap->n_colors;
			if ( nColors < 256 )
			{
				ii = nColors;
				cmap->color = (color_type *)regloc(ge, cmap->color,
											 256 *sizeof(color_type));
				for ( ; ii<256; ii++ )
				{
					cmap->color[ii].see_thru = FALSE;
					cmap->color[ii].ink_type = 0;
					cmap->color[ii].c = 0;
					cmap->color[ii].m = 0;
					cmap->color[ii].y = 0;
					cmap->color[ii].k = 0;
					cmap->color[ii].name = (char *)gloc(ge, 9);
					sprintf(cmap->color[ii].name, "color%03d", ii);
				}
			}
			nColors = 256;
			cprintf(ge, cvt,"COLORMAP ");

			for (ii=0; ii<nColors; ii++)
			{
				see_thru = cmap->color[ii].see_thru;
				ink_type = cmap->color[ii].ink_type;
				k = cmap->color[ii].k;
				c = cmap->color[ii].c;
				m = cmap->color[ii].m;
				y = cmap->color[ii].y;
				len = MultiToWideStr(ge, cmap->color[ii].name,tbuff );
	
				cprintf(ge, cvt,"\n");
				mlOutString(ge, len, tbuff, cvt, FALSE);
				cprintf(ge, cvt,"%02X",c);
				cprintf(ge, cvt,"%02X",m);
				cprintf(ge, cvt,"%02X",y);
				cprintf(ge, cvt,"%02X",k);
				cprintf(ge, cvt,"%01X",ink_type);
				cprintf(ge, cvt,"%01X",see_thru);
			}
			cprintf(ge, cvt,"\n");
			cprintf(ge, cvt,"END COLORMAP\n");
		}
		freeCmap(ge, &temp_cmap);
	}

	/* put the image data */
	imgPtr = t->data.img->info.data;
	size = hyt*byteWid;
	cprintf(ge, cvt,"DATA");
	for ( ii=0; ii<size; ii++, imgPtr++ ) 
	{
		if ( !(ii % 35) )
			cprintf(ge, cvt,"\n");
		c = *imgPtr;
		cprintf(ge, cvt,"%02X",c);
	}

	/* do we need this ?? */
	if (has_mask)
	{
		cprintf(ge, cvt,"\n");
		imgPtr = t->data.img->info.mask;
		maskWid = t->data.img->info.mask_byte_wid;
		size = hyt*maskWid;
		cprintf(ge, cvt,"MASK");
		for ( ii=0; ii<size; ii++, imgPtr++ ) 
		{
			if ( !(ii % 35) )
				cprintf(ge, cvt,"\n");
			c = *imgPtr;
			cprintf(ge, cvt,"%02X",c);
		}
	}

	/* End info */
	outEnderInfo(ge, cvt);

	if ( !outBuffer )
		cvtCloseFile(ge, cvt);
	else
	{
		*outBuffer = cvt->outBuff;
		*outCount = cvt->outBufPtr - cvt->outBuff;
	}
	return(AOK);
}
/***************************************************************************/
