/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: AxEditMask.h
   Applix ASC version: 1
   Date: Mon Nov 16 13:53:48 1998
*/
#ifndef AXEDITMASK_H
#define AXEDITMASK_H

// AxEditMask.h / AxEditMask.C represent a parallel to AxEditMask.java

class AxEditMask {
	char *hard_placeholders;
    char *hard_literals;
	char *edit_mask;
	int edit_mask_len;
	int *placeholders;
	int *literals;

	int in_array(char *inarray, char ccode);
	
	// Return whether the mchars[] character at index 'dex' is an
	// escape character (that is, a backslash which isn't itself escaped).
	// This function must be called in the context of a forward iteration
	// thru the mchars[]; param last_escape_dex is < dex during scan.
	//
	int is_escape_char(char *mchars, int dex, int last_escape_dex);

	// convert the passed plain index (user-visible entry field index)
	// to its corresponding index in the mchars.  The two indices can
	// differ because of the presence of escape chars '\' in the mchars.
	//
	int to_mask_dex(char *mchars, int mcharslen, int plainDex);
						 
	// return whether the mchars character at index 'dex' is a
	// placeholder character.  This is defined as (char's code is
	// in hardcoded placeholder charcodes) && (char is not escaped)
	//
  	int is_placeholder_index(char *mchars, int mcharslen, int dex);

	// return whether the mchars character at index 'dex' is a
	// literal character.  This is defined as (char is escaped) ||
	// (char is in hardcoded literal charcodes)
	//
  	int is_literal_index(char *mchars, int mcharslen, int dex);

	int placeholders_exist();

	// Return whether the 'inchar' is acceptable to the placeholder
	// at mask index 'dex'. This is the quick-n-cheesy way to do this.
	// Zero is returned if the input is unacceptable.
	//
	int accepts_char (char inchar, int dex);

public:

	AxEditMask (char *inmask);

	~AxEditMask();

	// Return whether passed inmask differs from current edit_mask
	//
	int mask_differs(char *inmask);

	// Return the length of the edit mask.  The mask-constrained
	// text field will always have exactly this text length.
	//
	int mask_length();

	// Return the string representing the "empty" value of the field.
	// Caller must delete[] the returned string.
	//
	char *initial_value ();

	// Return the next higher legal cursor index.
	//
	// param curdex:  the current cursor index in the field.
	//
	int increment_cursor_position (int curdex);

	// Return the next lower legal cursor index.
	//
	// param curdex:  the current cursor index in the field.
	//
	int decrement_cursor_position (int curdex);

	// return the smallest legal cursor position
	//
	int home_cursor();

	// return the largest legal cursor position
	//
	int end_cursor();

	// Perform "deletion" of characters over the indicated range.
	// Passed curstring is modified in place -- its length is assumed to
	// be >= edit_mask_len
	//
	// param curstring:  the current value of the field, in its entirety
	// param dex:        the starting index of the deletion
	// param len:        the #chars to delete from curstring
	//
	// Note that "deleting" characters from the string means to change
	// them to spaces if they occupy placeholder positions, do-nothing if
	// they occupy literal positions.
	//
	void deleteChars(char *curstring, int dex, int len);

	// Perform overwrite of the passed overstring onto the passed curstring.
	// Param curstring is modified in place -- its length is assumed to
	// be >= edit_mask_len
	//
	// param curstring:  the current value of the field, in its entirety
	// param overstring: the incoming string segment, to apply to field at dex
	// param overlen:    the length of overstring
	// param dex:        the starting index of the overwrite
	//
	// Overwriting behaves according to a "keystroke entry" analog,
	// that is, how would the field behave if each character of 'overstring'
	// were typed in individually.
	//
	void overwriteChars(char *curstring, char *overstring,
						int overlen, int dex);

	// test whether the passed overstring will be fully accepted by
	// the editmask.  Returns the number of characters of overstring that
	// would be accepted
	//
	// param curstring:  the current field value, in its entirety
	// param overstring: the incoming string, to test
	// param overlen:    the length of overstring
	// param dex:        the starting index of the "test overwrite"
	//
	int testChars (char *curstring, char *overstring,
				   int overlen, int dex);

};

#endif AXEDITMASK_H
