/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: arcell.c
   Applix ASC version: 6
   Date: Mon Nov 16 13:53:32 1998
*/
#include <aos.h>
#include <paratm.h>

/* normally 15 coords per quadrant */
static short coeffs[16] = {
	    0,  1713,  3406,  5063,  6664,  8192,  9630, 10963,
	12176, 13255, 14189, 14968, 15582, 16026, 16294, 16384
};
#define SHIFT   14

/* reduce coordinate output for ellipses with minimum radius smaller than... */
#define THRESH	30

#if 0	/* how does this look?: */

	radius		#pts/ellipse
	======		============
	16383 to 31		60
	30 to 16		30
	15 to 11		20
	10 to 8			15
	7 				12
	6				10
	5	 			9
	4				7
	3				5
	2				3
	1				2

#endif

/***********************************************************/
/* exact endpoint interpolator */
int AxEndArc(ang, sine, cosine, idx)
int ang;
register short *cosine, *sine;
short *idx;			/* index after last */
{
	register int tmp, lidx, ndx, interp;

	/* angle as 1/6000th circle */
	tmp = (ang * 10L) / 6;

	/* interpolation index and error */
	lidx = tmp / 100;
	interp = tmp - lidx * 100;
	if (60 < (ndx = lidx))
		ndx -= 60;

	/* interpolate sine and cosine */
	tmp = ndx <= 30 ? ndx - 15 : 45 - ndx;
	*sine = 0 < tmp ? coeffs[tmp] : -coeffs[-tmp];
	tmp = ndx <= 15 ? ndx : ndx <= 45 ? 30 - ndx : ndx - 60;
	*cosine = 0 < tmp ? coeffs[tmp] : -coeffs[-tmp];

	/* result index */
	*idx = lidx;

	/* any error? */
	if (interp) {
		if (60 < (ndx = ++lidx))
			ndx -= 60;

		/* interpolate sine */
		tmp = ndx <= 30 ? ndx - 15 : 45 - ndx;
		tmp = 0 < tmp ? coeffs[tmp] : -coeffs[-tmp];
		tmp -= *sine;
		*sine += tmp * interp / 100;

		/* interpolate cosine */
		tmp = ndx <= 15 ? ndx : ndx <= 45 ? 30 - ndx : ndx - 60;
		tmp = 0 < tmp ? coeffs[tmp] : -coeffs[-tmp];
		tmp -= *cosine;
		*cosine += tmp * interp / 100;
	}

	/* return TRUE if index is exact */
	return interp == 0;
}

/***********************************************************/

int ellipse(rect, stang, arcang, p)
TmRect32 *rect;
short stang;				/* start angle */
short arcang;				/* arc angle */
TmPoint32 p[];
{
	register TmPoint32 *pp;
	register int tmp, ndx, incr;
	int rx, ry, cx, cy;
	short s_idx, s_sine, s_cosine;
	short e_idx, e_sine, e_cosine;

	rx = rect->x_size >> 1;
	ry = rect->y_size >> 1;
	cx = rect->x_position + rx;
	cy = rect->y_position + ry;
	if (rx < 0)
		rx = -rx;
	if (ry < 0)
		ry = -ry;

	/* output density is a function of radius */
	incr = rx < ry ? rx : ry;
	incr = (incr ? THRESH / incr : THRESH) + 1;

	pp = p;

	if (arcang < 3600) {
		/* interpolate for start angle & beginning index */
		AxEndArc(stang, &s_sine, &s_cosine, &s_idx);

		/* interpolate for ending angle & ending index */
		if (AxEndArc(stang + arcang, &e_sine, &e_cosine, &e_idx))
			e_idx--;	/* if no interp, back up last index */

		/* wedge-ify */
		pp->x = cx;
		pp->y = cy;
		++pp;

		/* first interpolated point */
		pp->x = cx + ((long) rx * s_cosine >> SHIFT);
		pp->y = cy + ((long) ry * s_sine   >> SHIFT);
		++pp;
	}
	else {
		/* full circle; no need for endpoint interps */
		s_idx = 0;
		e_idx = 59;

		/* first interpolated point at 12:00 high */
		pp->x = cx;
		pp->y = cy - ry;
		++pp;
	}

	/* all the middle points */
	s_idx += incr;
	while (s_idx <= e_idx) {

		/* handle passing thru 3600 */
		if (60 < (ndx = s_idx))
			ndx -= 60;

		/* quadrant 1 points */
		while (s_idx <= e_idx && ndx <= 15) {
			pp->x = cx + ((long) rx *  coeffs[ndx     ] >> SHIFT);
			pp->y = cy + ((long) ry * -coeffs[15 - ndx] >> SHIFT);
			++pp;
			ndx = s_idx += incr;
		}

		/* quadrant 2 points */
		while (s_idx <= e_idx && ndx <= 30) {
			pp->x = cx + ((long) rx *  coeffs[30 - ndx] >> SHIFT);
			pp->y = cy + ((long) ry *  coeffs[ndx - 15] >> SHIFT);
			++pp;
			ndx = s_idx += incr;
		}

		/* quadrant 3 points */
		while (s_idx <= e_idx && ndx <= 45) {
			pp->x = cx + ((long) rx * -coeffs[ndx - 30] >> SHIFT);
			pp->y = cy + ((long) ry *  coeffs[45 - ndx] >> SHIFT);
			++pp;
			ndx = s_idx += incr;
		}

		/* quadrant 4 points */
		while (s_idx <= e_idx && ndx <= 60) {
			pp->x = cx + ((long) rx * -coeffs[60 - ndx] >> SHIFT);
			pp->y = cy + ((long) ry * -coeffs[ndx - 45] >> SHIFT);
			++pp;
			ndx = s_idx += incr;
		}
	}

	/* last interpolated point */
	if (arcang < 3600) {
		pp->x = cx + ((long) rx * e_cosine >> SHIFT);
		pp->y = cy + ((long) ry * e_sine   >> SHIFT);
		++pp;
	}

	/* very last is always the same as the very first */
	*pp++ = *p;

	/* return the coord count */
	return pp - p;
}

