/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: awfuncs.c
   Applix ASC version: 72.2
   Date: Tue Jan 19 21:10:46 1999
*/
#include <stdio.h>
#include <stdlib.h>

#include <basicwin.h>
#include <aos.h>
#include <xim.h>
#include <wchars.h>
#include <widefont.h>
#include <wide_extern.h>

#if NOLIB_THIMBLE && FOR_FILTER				/* For MSW filters */
#if BUILD_STANDALONE
extern void OleGetUserDir(char**);
extern void OleGetSystemDir(char**);
#else
extern void (*OFEXPORT_OleGetUserDir)(char**);
extern void (*OFEXPORT_OleGetSystemDir)(char**);
#endif
#elif !NOLIB_THIMBLE						/* For axmain */
extern char *ThimUserDir;
extern char *ThimApplixDir;
#endif

#ifdef getenv
#undef getenv
#endif

#define DEFAULT_UNK_CHAR '?' /* a replacement char when we don't know what how
							  * to replace a unicode char
							  */

/* Used in int vs. mbchar comparisons; mbchar is currently a signed char */
#define UC     unsigned char

#if BUILD_KANJI
unsigned int MBunicode2sjis_char();
#endif

/************************************************************/
/* The followiing are some useful wide strings   */
/************************************************************/
wchar WideNullString[] = {0};

#if NOLIB_THIMBLE
#define get_static_buffer nolib_get_static_buffer

mbchar *get_static_buffer(size)
int size;
{
	static int lastSize;
	static mbchar *lastPtr;

	if ( !lastPtr )
		lastPtr = malloc(lastSize = 1000);
	if ( size > lastSize )
		lastPtr = realloc(lastPtr,lastSize = size);
	return(lastPtr);
}
#endif

mbchar *get_static_buffer();




/** this is a special function used as a convenience to convert a string
**  from wide to multi-byte, without having to allocate a buffer.  It
**  converts into a static buffer, and returns a pointer to the buffer.
**  THEREFORE, do not call this routine twice and expect the result of the
**  first call to be valid!    **/

mbchar *WToMbOneStr( in_str )
wchar *in_str;
{
    int len;
    mbchar *one_static_str;

    len = (Wstrlen(in_str)+1) * sizeof(wchar);
    one_static_str = get_static_buffer( len );
    WToMbStr( in_str, one_static_str );
    return( one_static_str );
}

mbchar *WnToMbOneStr( in_str, len )
wchar *in_str;
int len;        /* length in characters, not bytes */
{
    mbchar *one_static_str;

    one_static_str = (mbchar *)get_static_buffer( (len+1)*sizeof(wchar) );
    WnToMbStr( in_str, one_static_str, len );
    return( one_static_str );
}




/** this is a special function used as a convenience to convert a string
**  from multi-byte to wide, without having to allocate a buffer.  It
**  converts into a static buffer, and returns a pointer to the buffer.
**  THEREFORE, do not call this routine twice and expect the result of the
**  first call to be valid!    **/

wchar *MbToWOneStr( in_str )
mbchar *in_str;
{
    int len;
    wchar *one_static_str;

    len = (MBstrlen(in_str)+1) * sizeof(wchar);
    one_static_str = (wchar *)get_static_buffer( len );
    MbToWStr( in_str, one_static_str );
    return( one_static_str );
}


/** Return the number of real wide characters in a wide string
**  The input length is the number of characters, not bytes    */

int Wnum_wide_chars( in_str )
wchar *in_str;
{
    int num_wide = 0;

    while( *in_str != (wchar)0 )
       {
       if ( IS_WIDE_CHAR( *in_str ) || IS_NARROW_KANA(*in_str) )
          num_wide++;
       in_str++;
       }
    return( num_wide );
}


HasWideChars_n( in_str, len )
wchar *in_str;
int len;
{
   register int i;
#if BUILD_WIDE
   /* return whether a string contains any wide characters */
   for ( i=0; i<len; i++ )
       {
       if ( IS_WIDE_CHAR( *in_str ) || IS_NARROW_KANA(*in_str) )
          return( TRUE );
       in_str++;
       }
#endif
   return( FALSE );
}
   



int32 WMbstrcmp( str1, str2 )
wchar *str1;
mbchar *str2;
{
#if BUILD_WIDE
    return(Wstrcmp(str1, MbToWOneStr(str2)));
#else
    return( strcmp( str1, str2 ) );
#endif
}



int32 WMbstrncmp( str1, str2, n )
wchar *str1;
mbchar *str2;
int n;
{
#if BUILD_WIDE
    return(Wstrncmp(str1, MbToWOneStr(str2), n));
#else
    return( strncmp( str1, str2, n ) );
#endif
}




/* add a mixed byte string to the end of a wide string */
wchar *WMbstrcat( base, add )
wchar *base;
mbchar *add;
{
    wchar *p;

    if ( add != (mbchar *)NULL )
       {
       p = base + Wstrlen( base );
       MbToWStr( add, p );
       }

    return( base );
}






/* add a number of bytes from a mixed byte string to the end of a wide string */
wchar *WMbstrncat( base, add, len )
wchar *base;
mbchar *add;
int len;
{
    wchar *p;

    if ( add != (mbchar *)NULL )
       {
       p = base + Wstrlen( base );
       MbnToWStr( add, p, len );
       }

    return( base );
}


/* add a wide string to the end of a mixed byte string */
mbchar *MbWstrcat( base, add )
mbchar *base;
wchar *add;
{
    mbchar *p;

    if ( add != (wchar *)NULL )
       {
       p = base + MBstrlen( base );
       WToMbStr( add, p );
       }

    return( base );
}



/* add a number of characters from a wide string to the end of a */
/*  mixed byte string */
mbchar *MbWstrncat( base, add, len )
mbchar *base;
wchar *add;
int len;
{
    mbchar *p;

    if ( add != (wchar *)NULL )
       {
       p = base + MBstrlen( base );
       WnToMbStr( add, p, len );
       }

    return( base );
}




/* Ashisuto added for check the input chars validation.
 * First time used in axatmrt/xmenter.c, EntInsertText().
 * Return value:
 *  if wch is "in" mbstr, returns the pointer in mbstr of wchr's 1st occurrence
 *  else, returns NULL
 * 95.09.13
 */
mbchar *
MbWstrchr(mbstr, wch)
mbchar  *mbstr;
wchar   wch;
{
    wchar   *wchP;
    mbchar  *mbP;
 
    if (mbstr == NULL)
        return(NULL);
 
    wchP = MbToWOneStr(mbstr);
    while (*wchP) {
        if (*wchP == wch)
            return(mbstr);
        if (IS_MULTI_BYTE(mbstr))
            mbstr += 2;
        else
            mbstr++;
        wchP++;
    }
    return(NULL);
}


/* FUNCTION to convert a unicode character to an "equivalent" ascii character */
void
MBunicode2ax(wc, mbstr, flag, n_chars)
unsigned short wc;
mbchar **mbstr;
int *flag;
int *n_chars; /* len of the return string */
{
	*mbstr = (mbchar*)malloc(4);
   	(*mbstr)[0] = DEFAULT_UNK_CHAR;
   	(*mbstr)[1] = 0;

#if BUILD_KANJI
	MBunicode2euc(wc, *mbstr);
	if ( (*mbstr)[0] != DEFAULT_UNK_CHAR )
	{
		*flag = 4; /* EUC */
		*n_chars = 2 /* ? */;
		return;
	}
#endif

	*flag = 1;  /* use "regular" ISO latin font */
		    /* set flag to 2 for Symbol font */
	*n_chars = 1;
	switch (wc)
	{
		case 0x20ac:		/* EURO */
			(*mbstr)[0]=0x80;
			(*mbstr)[1] = 0;
		break;
		case 0x02c6:		/* ^ */
			(*mbstr)[0] = '^';
			(*mbstr)[1] = 0;
		break;
		
		case 0x02dc:		/* ~ */
			(*mbstr)[0] = '~';
			(*mbstr)[1] = 0;
		break;
	
		case 0x2013:		/* en-dash (-) */
			(*mbstr)[0] = '-';
			(*mbstr)[1] = 0;
		break;

		case 0x2014:		/* em-dash (--) */
			(*mbstr)[0] = '-';
			(*mbstr)[1] = '-';
			(*mbstr)[2] = 0;
			*n_chars = 2;
		break;
		
		case 0x2015:		/* ~ */
			(*mbstr)[0] = '~';
			(*mbstr)[1] = 0;
		break;
		
		case 0x2018:		/* ` */
			(*mbstr)[0] = '`';
			(*mbstr)[1] = 0;
		break;

		case 0x2019:		/* ' */
			(*mbstr)[0] = '\'';
			(*mbstr)[1] = 0;
		break;

		case 0x201a:		/* , */
			(*mbstr)[0] = ',';
			(*mbstr)[1] = 0;
		break;
		case 0x2022:		/* . */
			(*mbstr)[0] = '.';
			(*mbstr)[1] = 0;
			*flag = 2;
		break;

		case 0x2026:		/* ... */
			(*mbstr)[0] = '.';
			(*mbstr)[1] = '.';
			(*mbstr)[2] = '.';
			(*mbstr)[3] = 0;
			*n_chars = 3;
		break;

		case 0x201C:		/* `` */
			(*mbstr)[0] = '`';
			(*mbstr)[1] = '`';
			(*mbstr)[2] = 0;
			*n_chars = 2;
		break;

		case 0x201D:		/* '' */
			(*mbstr)[0] = '\'';
			(*mbstr)[1] = '\'';
			(*mbstr)[2] = 0;
			*n_chars = 2;
		break;

		case 0x201E:		/* ,, */
			(*mbstr)[0] = ',';
			(*mbstr)[1] = ',';
			(*mbstr)[2] = 0;
			*n_chars = 2;
		break;

		case 0x2039:		/* < */
			(*mbstr)[0] = '<';
			(*mbstr)[1] = 0;
		break;

		case 0x203A:		/* > */
			(*mbstr)[0] = '>';
			(*mbstr)[1] = 0;
		break;
		case 0x2122:		/* 00D4 TM*/
			(*mbstr)[0] = 0x00D4;
			(*mbstr)[1] = 0;
			*flag = 2;
		break;


		default:
	    	/* here we ask for the user preference for unknown chars */
	    	(*mbstr)[0] = DEFAULT_UNK_CHAR;
	    	(*mbstr)[1] = 0;
			*flag = 0;
			*n_chars = 1;
		break;
	}
}



#if AX_X11      /*    {    */

int WToMbStr( in_str, out_str )
wchar *in_str;
mbchar *out_str;
{
    /** convert a wide to a mixed-byte string.  This returns the number of */
    /**  bytes in the resultant mixed-byte string                          */
#if !BUILD_WIDE
    MBstrcpy( out_str, in_str );
    return( strlen( out_str ));
#else

    wchar *inP;
    mbchar *outP;

    inP = in_str;
    outP = out_str;

    while ( *inP != (wchar)NULL )
      {
      if ( IS_WIDE_CHAR( *inP ) || IS_NARROW_KANA( *inP ) )
         {
         /* a multi-byte character */
         MAKE_MULTI_FROM_WIDE( inP, outP );
         }
      else
         {
         /* a single byte character */
         *outP++ = (*inP & 0xFF);
         }
      inP++;
      }
   *outP = 0;
   return(outP - out_str);
#endif
}


int WnToMbStr( in_str, out_str, len )
wchar *in_str;
mbchar *out_str;
int len;            /* the length of the wide string */
{
    /** convert a wide to a mixed-byte string.  This returns the number of */
    /**  bytes in the resultant mixed-byte string                          */

#if !BUILD_WIDE
    MBstrncpy( out_str, in_str, len );
    *(out_str+len) = 0;
    return( strlen(out_str) );
#else
    wchar *inP, *endP;
    mbchar *outP;
    int char_len;

    if ( len < 0 )
        /* saftey feature */
        len = Wstrlen( in_str );

    inP = in_str;
    endP = inP + len;
    outP = out_str;

    while ( (inP < endP) && (*inP != (wchar)NULL) )
      {
      if ( IS_WIDE_CHAR( *inP ) || IS_NARROW_KANA(*inP) )
         {
         /* a multi-byte character */
         MAKE_MULTI_FROM_WIDE( inP, outP );
         }
      else
         {
         /* a single byte character */
         *outP++ = (*inP & 0xFF);
         }
      inP++;
      }
   *outP = 0;
   return(outP - out_str);
#endif
}


/** return the lengtth of a mixed byte string in characters, not bytes */
int MbCharLen( in_str )
mbchar *in_str;
{
#if !BUILD_WIDE
    return( strlen( in_str ) );
#else
    register mbchar *inP;
    register int length;

    inP = in_str;
    length = 0;
    while ( *inP != (mbchar)NULL )
      {
      if ( IS_MULTI_BYTE(inP) )
         {
         /* a multi-byte character */
         inP++;                 /* skip an extra character */
         }
      inP++;
      length++;
      }

   return( length );
        
#endif
}



/**********************************************************************/
/* return a byte pointer to the i'th character (not byte) in the
** input multi-byte string.  The input character postion is zero relative,
** so character "1" is the second character in the input string.
** (This function means that we do not have to
** fully expand the string to wide characters to do certain operations */
/**********************************************************************/

mbchar *MbCharPos( in_str, char_pos )
mbchar *in_str;
int char_pos;
{
    mbchar *inP;

#if !BUILD_WIDE
    return( in_str+char_pos );
#else
    inP = in_str;
    while ( (*inP != (mbchar)NULL) && (char_pos > 0) )
      {
      if ( IS_MULTI_BYTE(inP) )
         {
         /* a multi-byte character */
         inP++;                 /* skip an extra character */
         }
      inP++;
      char_pos--;
      }

   return( inP );
        
#endif
}



/**************************************************************************/
/* Convert a mixed-byte string to a wide character string, and return the */
/* number of characters (not bytes) in the string                         */
/**************************************************************************/

int MbToWStr( in_str, out_str )
mbchar *in_str;
wchar *out_str;
{
#if !BUILD_WIDE
    MBstrcpy( out_str, in_str );
    return( MBstrlen(out_str) );
#else
    mbchar *inP;
    wchar *outP;
    int len;

    len = 0;
    inP = in_str;
    outP = out_str;

    if ( inP == (mbchar *)NULL )
       {
       if ( out_str != (wchar *)NULL )
           *out_str = (wchar)0;
       return(0);
       }

    while ( *inP != (mbchar)0 )
      {
      if ( IS_MULTI_BYTE(inP) )
         {
         /* a multi-byte character */
         *outP = MAKE_WIDE_FROM_MULTI( inP );
         inP++;                 /* skip an extra character */
         }
      else
         {
         /* a single byte character */
         *outP = (*inP & 0xFF);
         }
      outP++;
      inP++;
      len++;
      }
   *outP = 0;
   return(len);
#endif
}




/**************************************************************************/
/** Convert a certain number of bytes of a mixed-byte string to a wide */
/** string.  Returns the number of characters (not bytes) in the result  */
/*  wide string.              */

int MbnToWStr( in_str, out_str, len )
mbchar *in_str;
wchar *out_str;
int len;
{
#if !BUILD_WIDE
    MBstrncpy( out_str, in_str, len );
    *(out_str+len) = 0; 

    return( len );
#else

    mbchar *inP, *endP;
    wchar *outP;
    int out_len = 0;

    inP = in_str;
    endP = in_str + len - 1;
    outP = out_str;

    if (len == 0) {
        if (out_str != NULL)
            *out_str = 0;
        return(0);
    }
 
    while ( inP <= endP )
      {
      if ( IS_MULTI_BYTE(inP) && (inP < endP) )  /* careful,don't look off end string */
         {
         /* a multi-byte character */
         *outP = MAKE_WIDE_FROM_MULTI( inP );
         inP++;                 /* skip an extra character */
         out_len++;
         }
      else
         {
         /* a single byte character */
         *outP = (*inP & 0xFF);
         out_len++;
         }
      outP++;
      inP++;
      }

    *outP = 0;
    return( out_len );
#endif
}

   


/** return the number of wide chars, (as opposed to chars from the Latin-1
**  set) which are in the passed string.  The passed length if the number of
**  bytes in the input which shuld be looked at, not the number of characters */

int MBnum_wide_chars_n( in_str, len )
mbchar *in_str;
int len; 
{
    mbchar *inP, *endP;
    int num_wide = 0;

#if !BUILD_WIDE
    return( 0 );
#else 

    inP = in_str;
    endP = in_str + len - 1;

    while ( inP <= endP )
      {
      if ( IS_MULTI_BYTE(inP) && (inP < endP) )  /* careful,don't look off end string */
         {
         /* a multi-byte character */
         inP++;                 /* skip an extra character */
         num_wide++;
         }
      inP++;
      }

    return( num_wide );
#endif
}









#if BUILD_ORIENT    /*  {  */

/**************************************************************************/
/** Convert a mixed-byte string to a certain number if characters in a wide */
/** string.  Returns the number of characters (not bytes) in the result  */
/*  wide string.              */
/**************************************************************************/

int MbToWnStr( in_str, out_str, max_len )
mbchar *in_str;
wchar *out_str;
int max_len;
{
    register mbchar *inP;
    register wchar *outP;
    register int out_len = 0;

    inP = in_str;
    outP = out_str;

    while ( (inP != (mbchar *)NULL) && (inP != (mbchar)0) && (max_len != 0) )
      {
      if ( IS_MULTI_BYTE(inP) ) 
         {
         /* a multi-byte character */
         *outP = MAKE_WIDE_FROM_MULTI( inP );
         inP++;                 /* skip an extra character */
         out_len++;
         }
      else
         {
         /* a single byte character */
         *outP = (*inP & 0xFF);
         out_len++;
         }

      outP++;
      inP++;

      if ( out_len == max_len )
         break;
      }

    *outP = 0;
    return( out_len );
}
#endif     /*  }  BUILD_ORIENT   */


#else     /*   } AX_X11 vs AXMS {   */


/*************************************************************************/
/* On NT we use the native multi-byte to wide conversion functions       */
/* so that any multi-byte strings we may produce are compativle with     */
/* any system functions we may pass them to.  In japanese, the multi-byte */
/* format us Shift-JIS                                                   */
/*************************************************************************/

int WToMbStr( in_str, out_str )
wchar *in_str;
mbchar *out_str;
{
    /** convert a wide to a mixed-byte string.  This returns the number of */
    /**  bytes in the resultant mixed-byte string                          */
#if !BUILD_WIDE
    MBstrcpy( out_str, in_str );
    return( strlen( out_str ));
#else
    return( wcstombs( out_str, in_str, 9999 ) ); /* 9999 means dest is big */
#endif
}


int WnToMbStr( in_str, out_str, len )
wchar *in_str;
mbchar *out_str;
int len;            /* the length of the wide string */
{
    /** convert a wide to a mixed-byte string.  This returns the number of */
    /**  bytes in the resultant mixed-byte string                          */

#if !BUILD_WIDE
    MBstrncpy( out_str, in_str, len );
    *(out_str+len) = 0;
    return( strlen(out_str) );
#else
    int mb_len;
    wchar save_char;

    if ( len < 0 )
        /* saftey feature */
        len = Wstrlen( in_str );

    save_char = *(in_str + len);
    *(in_str + len) = 0;
    mb_len = wcstombs( out_str, in_str, 9999 ); /* 9999 means dest is big */
    *(in_str + len) = save_char;
    return( mb_len );
#endif
}


/** return the lengtth of a mixed byte string in characters, not bytes */
int MbCharLen( in_str )
mbchar *in_str;
{
#if !BUILD_WIDE
    return( strlen( in_str ) );
#else
   int length;
   length = _mbslen(in_str);
   return( length );
#endif
}



/**********************************************************************/
/* return a byte pointer to the i'th character (not byte) in the
** input multi-byte string.  The input character postion is zero relative,
** so character "1" is the second character in the input string.
** (This function means that we do not have to
** fully expand the string to wide characters to do certain operations */
/**********************************************************************/
#ifndef MS_CharNext
#define MS_CharNext CharNext
#endif

mbchar *MbCharPos( in_str, char_pos )
mbchar *in_str;
int char_pos;
{
    mbchar *inP;

#if !BUILD_WIDE
    return( in_str+char_pos );
#else
    inP = in_str;
    while ( (*inP != (mbchar)NULL) && (char_pos > 0) )
      {
      inP = MS_CharNext( inP );
      char_pos--;
      }
   return( inP );
#endif
}



/**************************************************************************/
/* Convert a mixed-byte string to a wide character string, and return the */
/* number of characters (not bytes) in the string                         */
/**************************************************************************/

int MbToWStr( in_str, out_str )
mbchar *in_str;
wchar *out_str;
{
#if !BUILD_WIDE
    MBstrcpy( out_str, in_str );
    return( MBstrlen(out_str) );
#else
   int len;
   len = mbstowcs( out_str, in_str, strlen(in_str) );
   if ( len < 0 )
       {
       /* invalid multi-byte char encountered */
       *out_str = 0;
       len = 0;
       }
   else if ( len == strlen(in_str) )
   {
      /* when calling "len = mbstowcs(wcstr, mbstr, count)", So sprach
       * Microsoft: "When return value equals count, the terminal NULL will
       * not be set in target string." 
       * Do it by ourself. Ashisuto, 1996.12.02
       */
      out_str[len] = 0;
   }
   return(len);
#endif
}




/**************************************************************************/
/** Convert a certain number of bytes of a mixed-byte string to a wide */
/** string.  Returns the number of characters (not bytes) in the result  */
/*  wide string.              */

int MbnToWStr( in_str, out_str, len )
mbchar *in_str;
wchar *out_str;
int len;
{
#if !BUILD_WIDE
    MBstrncpy( out_str, in_str, len );
    *(out_str+len) = 0; 
    return( len );
#else
    int out_len;
    out_len = mbstowcs( out_str, in_str, len );
   if ( out_len < 0 )
       /* invalid multi-byte char encountered */
       out_len = 0;
   else if ( out_len == len )
      out_str[out_len] = 0;

    *(out_str+out_len) = 0;
    return( out_len );
#endif
}



/** return the number of wide chars, (as opposed to chars from the Latin-1
**  set) which are in the passed string.  The passed length if the number of
**  bytes in the input which shuld be looked at, not the number of characters */

int MBnum_wide_chars_n( in_str, len )
mbchar *in_str;
int len; 
{
#if !BUILD_WIDE
    return( 0 );
#else 
    int num_wide, char_len;
    mbchar save_char;
    save_char = *(in_str+len);
    *(in_str+len) = 0;
    char_len = _mbslen( in_str );
    num_wide = strlen(in_str) - char_len;
    *(in_str+len) = save_char;
    return( num_wide );
#endif
}



#if BUILD_ORIENT    /*  {  */

/**************************************************************************/
/** Convert a mixed-byte string to a certain number if characters in a wide */
/** string.  Returns the number of characters (not bytes) in the result  */
/*  wide string.              */
/**************************************************************************/

int MbToWnStr( in_str, out_str, max_len )
mbchar *in_str;
wchar *out_str;
int max_len;
{
    register mbchar *inP;
    register int out_len, num_chars;

    inP = in_str;

    /* find the end of the chars we want */
    while ( (*inP != 0) && (num_chars < max_len) )
       {
       inP = MS_CharNext( inP );
       num_chars++;
       }

    out_len = mbstowcs( out_str, in_str, (inP-in_str) );
    if ( out_len < 0 )
       /* invalid multi-byte char encountered */
       out_len = 0;
    *(out_str+out_len) = 0;

    return( out_len );
}
#endif              /*  }  BUILD_ORIENT  */

#endif    /*   }  AXMs vs X11 */







































/** Return whether the input char pointer points to the fist byte of a wide */
/**   character in an external ascii file                                */

int is_external_ascii_multi_byte(inP)
mbchar  *inP;
{
    char    *code_type;
 
#if !BUILD_KANJI
    return( IS_MULTI_BYTE( inP ) );
#else
    if ( ascii_is_shift_jis() )
        /* as jis */
        return( IS_SHIFT_JIS(inP) );
    else
        /* as euc */
        return( IS_MULTI_BYTE(inP) );
#endif
 
}






external_ascii_to_wide_str( in_str, out_str )
mbchar *in_str;
wchar *out_str;
{
    int len;

#if !BUILD_KANJI    /*  {  */
    len = MbToWStr( in_str, out_str );
    return( len );
#else
    /* Ashisuto changed for JAPANESE build. 95.09.13 */
	/* Ashisuto added first time to enable reading Applix OpenMail attached
	 * text files, which are encoded in JIS. Enabling automatic JIS code 
	 * recognition. 
	 * 1998.02.25
	 */
    if (is_jis_str(in_str)) {
        mbchar  *tmp_str;
 
        tmp_str = (mbchar *)malloc(2*(MBstrlen(in_str)+1));
        MBjis2euc_str(in_str, tmp_str);
        len = MbToWStr(tmp_str, out_str);
        free(tmp_str);
    } else if (ascii_is_shift_jis()) {
        mbchar  *tmp_str;
 
        tmp_str = (mbchar *)malloc(2*(MBstrlen(in_str)+1));
        MBstrcpy(tmp_str, in_str);
        MBsjis2euc_str(tmp_str);
        len = MbToWStr(tmp_str, out_str);
        free(tmp_str);
    } else {
        len = MbToWStr(in_str, out_str);
    }
    return( len );
#endif    /*  }  BUILD_KANJI */
}




#if 0        /*   {  */

#define MSB(s)  (unsigned char)(((s) >> 8) & 0xFF)
#define LSB(s)  (unsigned char)((s) & 0xFF)
#define ODD(n)  (unsigned char)((n) & 1)

wide_to_shift_jis(ins, outs)
wchar   *ins;
mbchar  *outs;
{
    int i;
 
    if (MSB(*ins) <= 0x5e)
        outs[0] = (mbchar)((wchar)(MSB(*ins) + 1) / 2 + 0x70);
    else
        outs[0] = (mbchar)((wchar)(MSB(*ins) + 1) / 2 + 0xb0);
    if (ODD(MSB(*ins))) {
        if ((outs[1] = LSB(*ins) + 0x1f) >= 0x7f)
            outs[1]++;
    } else
        outs[1] = LSB(*ins) + 0x7e;
}
 

wchar
shift_jis_to_wide(inP)
mbchar  *inP;
{
    wchar   outw;
 
    if ((*inP >= 0x81) && (*inP <= 0x9f)) {
        outw = (*inP << 1) - 0x0e1;
    } else if ((*inP >= 0xe0) && (*inP <= 0xef)) {
        outw = (*inP << 1) - 0x161;
    }
    if ((*(inP+1)) >= 0x9f)
        outw++;
    outw <<= 8;
 
    if (((*(inP+1)) >= 0x40) && ((*(inP+1)) <= 0x7e)) {
        outw |= (*(inP+1)) - 0x1f;
    } else if (((*(inP+1)) >= 0x80) && ((*(inP+1)) <= 0x9e)) {
        outw |= (*(inP+1)) - 0x1f - 1;
    } else if (((*(inP+1)) >= 0x9f) && ((*(inP+1)) <= 0xfe)) {
        outw |= (*(inP+1)) - 0x7e;
    }
    return(outw);
}
#endif  /*  if zero'd out }  */



/** convert a wide to a mixed-byte string for the external environemnt  */
/** This returns the number of bytes in the resultant mixed-byte string */

wide_to_external_ascii( in_str, out_str )
wchar *in_str;
mbchar *out_str;
{
    wchar *inP;
    mbchar *outP;
    int len;

   len = WToMbStr(in_str, out_str);
#if BUILD_KANJI    /*  {  */
   /* Ashisuto changed for JAPANESE build. 95.09.13 */
    if (ascii_is_shift_jis())
       len = MBeuc2sjis_str(out_str);
#else     /*  }  {  BUILD_KANJI  */
   return(len);
#endif      /*  }  */
}


 
    
mbchar *wide_to_external_ascii_one( in_str )
wchar *in_str;
{
    int len;
    mbchar *one_static_str;
 
    len = (Wstrlen(in_str)+1) * sizeof(wchar);
    one_static_str = get_static_buffer( len );
    wide_to_external_ascii( in_str, one_static_str );
    return( one_static_str );
}


#if 0
WCharType(wch)
wchar   wch;
{
    int char_type;
 
    if (IS_WIDE_CHAR(wch))
        char_type = 1;  /* WIDE_FONT */
    else if (IS_NARROW_KANA(wch))
        char_type = 2;  /* NARROW_KANA_FONT */
    else
        char_type = 0;  /* NARROW_FONT */
 
    return(char_type);
}
#endif
 




#if BUILD_KANJI   /*  {  */

/* Ashisuto added for convinience.	1998.02.25 */
int IS_SJIS_NARROW_KANA(mbP)
mbchar *mbP;
{
	if ((UC)*mbP >= 0xa1 && (UC)*mbP <= 0xdf)
		return(TRUE);
	else
		return(FALSE);
}


/* return whether  we are to treat external ascii files as being encoded in */
/* Shift-JIS or EUC                                     */
int ascii_is_shift_jis()
{
    static int env_says_shift_jis = -1;
    char *code_type;

    if ( env_says_shift_jis < 0 )
       {
       /* initialize from environment one time only */
       code_type = (char *)getenv("AXASCII");
       if (code_type)
          if (strcasecmp(code_type, NXLT("sjis")) == 0)
             env_says_shift_jis = 1;
          else
             env_says_shift_jis = 0;
       else
          env_says_shift_jis = env_is_sjis();
       }
    return( env_says_shift_jis );
}



/* return if we should treat the keyboard input as Shift-JIS or as EUC code. */
int lang_is_sjis()
{
   static int  lang_is_sjis = -1;
    char *lang_type;
 
   if (lang_is_sjis == -1) {
       lang_type = (char *)getenv("LANG");
       if ( lang_type == (char *)NULL )
           {
           lang_is_sjis = 0;               /* default */
           return(lang_is_sjis);
           }

#if HP800V3
       if (!strcmp(lang_type, "japanese.euc") ||
               !strcmp(lang_type, "ja_JP.eucJP") || 
               !strcmp(lang_type, "ja_JP.UJIS") || 
               !strcmp(lang_type, "ja_JP.EUC"))
           lang_is_sjis = 0;
       else
           lang_is_sjis = 1;
#elif RIOS
       if (!strcmp(lang_type, "ja_JP"))
           lang_is_sjis = 0;
       else
           lang_is_sjis = 1;
#elif SUN4S || SUNIS
       if (strcmp(lang_type, "ja_JP.PCK"))
           lang_is_sjis = 0;
       else
           lang_is_sjis = 1;
#elif NEC_EWS
       /* For NEC_EWS: japan, ja_JP.EUC, ja_JP.eucJP - are EUC */
       if (strcmp(lang_type, "ja_JP.SJIS"))
           lang_is_sjis = 0;
       else
           lang_is_sjis = 1;
#else
       lang_is_sjis = 0;    /* default as euc, such as SunOS4.1.x. */
#endif
   }
   return(lang_is_sjis);
}
 


/* return if we should treat the title appearing over application windows
   as Shift-JIS or as EUC code. */
int env_is_sjis()
{
   static int  wm_is_sjis = -1;
    char *wm_type;
 
   if (wm_is_sjis == -1) {
       wm_type = (char *)getenv("AXKANJI");
       if (wm_type) {
            if (!strcasecmp(wm_type, NXLT("sjis")))
                wm_is_sjis = 1;
            else
                wm_is_sjis = 0;
        } else {
            wm_is_sjis = lang_is_sjis();
        }
   }
   return(wm_is_sjis);
}
 
 
 


static void MBeuc2sjis(inP)
mbchar  *inP;
{
    if (*inP & 0x01) {
        *(inP+1) = (mbchar)((UC)*(inP+1) - 0x61);
        if ((UC)*(inP+1) >= 0x7f)
            (*(inP+1))++;
    } else {
        *(inP+1) = (mbchar)((UC)*(inP+1) - 0x02);
    }
 
    if ((UC)*inP <= 0xde)
        *inP = (mbchar)((((UC)*inP - 0x7f) / (unsigned)2) + 0x70);
    else
        *inP = (mbchar)((((UC)*inP - 0x7f) / (unsigned)2) + 0xb0);
}



static void MBsjis2euc(inP)
mbchar        *inP;
{
      if (((UC)*inP >= 0x81) && ((UC)*inP <= 0x9f)) {
              *inP = (mbchar)(((UC)*inP << 1) - 0x061);
      } else if (((UC)*inP >= 0xe0) && ((UC)*inP <= 0xef)) {
              *inP = (mbchar)(((UC)*inP << 1) - 0xe1);
      }
      if ((UC)*(inP+1) >= 0x9f)
              (*inP)++; 

      if (((UC)*(inP+1) >= 0x40) && ((UC)*(inP+1) <= 0x7e)) {
              *(inP+1) = (mbchar)((UC)*(inP+1) + 0x61);
      } else if (((UC)*(inP+1) >= 0x80) && ((UC)*(inP+1) <= 0x9e)) {
              *(inP+1) = (mbchar)((UC)*(inP+1) + 0x60);
      } else if (((UC)*(inP+1) >= 0x9f) && ((UC)*(inP+1) <= 0xfe)) {
              *(inP+1) = (mbchar)((UC)*(inP+1) + 0x02);
      }
}


MBeuc2sjis_str(str)
mbchar  *str;
{
    mbchar  *strP;
 
    if (str == NULL)
        return;
    while (*str) {
        if (IS_MULTI_BYTE(str)) {
            if ((UC)*str == 0x8e) {
               /* Since the Shift-JIS code of Narrow Kana is just the LSB of EUC
                 * code, we must throw away the MSB to being Shift-JIS code */
                for (strP = str; *strP = *(strP + 1); strP++);
            } else {
                MBeuc2sjis(str);
                str++;
            }
        }
        str++;
    }
}


MBsjis2euc_str(str)
mbchar  *str;
{
    mbchar *strP;
 
    if (str == NULL)
        return;
    while (*str) {
        if (IS_SHIFT_JIS(str)) {
            if (IS_SJIS_NARROW_KANA(str)) {
               /* Since the Shift-JIS code of Narrow Kana is just the LSB of EUC
                 * code, we must add the MSB of '0x8e' to being EUC code */
                for (strP = str + MBstrlen(str); strP >= str; strP--)
                    *(strP + 1) = *strP;
                *str++ = 0x8e;
            } else {
                MBsjis2euc(str);
                str++;
            }
        }
        str++;
    }
}

/* FUNCTION to convert a wide Unicode char to a multi-byte EUC str */
MBunicode2euc(w, mbstr)
wchar w;
mbchar *mbstr;			/* Must be 3 bytes long */
{
	unsigned int sjis;
	int one_byte=FALSE;

	if ( w == 0 || mbstr == NULL)
		return;
	sjis = MBunicode2sjis_char((unsigned int)w);
	if (!sjis) {
		/* No SJIS for that unicode, so return a question mark */
		mbstr[0] = DEFAULT_UNK_CHAR;
    	mbstr[1] = 0;
    	mbstr[2] = 0;
	}
	else {
		if (sjis > 0xff) {
			mbstr[0] = (sjis & 0xff00) >> 8;
			mbstr[1] = sjis & 0x00ff;
		}
		else {
			mbstr[0] = sjis;
			mbstr[1] = 0;
			one_byte = TRUE;
		}
    	mbstr[2] = 0;

		/* Narrow Kana is a range of 8-bit chars (0xa1 - 0xdf).
			EUC is the same value preceded by 0x8e */
		if (one_byte && IS_SJIS_NARROW_KANA(mbstr)) {
			mbstr[1] = mbstr[0];
			mbstr[0] = 0x8e;
		}
		else
			MBsjis2euc(mbstr);
	}
}


/* FUNCTION to convert a wide Unicode char to an SJIS value.
	Returns an unsigned int representing a 1 or 2-byte SJIS value */

#define UNI2SJIS_TABLE_UNREAD   0
#define UNI2SJIS_TABLE_READ     1
#define UNI2SJIS_TABLE_FAILED   2
#define BAD_SJIS                0

unsigned int 
MBunicode2sjis_char(uni)
wchar uni;
{
	static int status = UNI2SJIS_TABLE_UNREAD;
	char	filename[STRPATH+1];
	static unsigned int *table=NULL;
	char	*ax_system_dir, *thim_user_dir;
	FILE *fp;
	char	buf[1001];
	unsigned int		uni_table, sjis_table;
	unsigned int		uni_val;
	int		lines=0;
	int		ix;
	char	str[100];

	if (status == UNI2SJIS_TABLE_UNREAD) {
		/* Look for override xlation table in user's home dir */
#if NOLIB_THIMBLE && FOR_FILTER
#if BUILD_STANDALONE
		OleGetUserDir(&thim_user_dir);
#else
		(*OFEXPORT_OleGetUserDir)(&thim_user_dir);
#endif
#elif !NOLIB_THIMBLE
		thim_user_dir = ThimUserDir;
#else
		return(BAD_SJIS);			/* Can't use this fcn in aagtwy */
#endif
		if (!thim_user_dir)
			return(BAD_SJIS);
		sprintf( filename, "%s/%s", thim_user_dir, NXLT("uni2sjis.dat"));
		fp = fopen(filename, "r");
		if (fp == (FILE*)NULL) {
			/* Use standard xlation table in axdata/lang */
#if NOLIB_THIMBLE && FOR_FILTER
#if BUILD_STANDALONE
		OleGetSystemDir(&ax_system_dir);
#else
		(*OFEXPORT_OleGetSystemDir)(&ax_system_dir);
#endif
#elif !NOLIB_THIMBLE
		ax_system_dir = ThimApplixDir;
#endif
			if (!ax_system_dir)
				return(BAD_SJIS);
			sprintf( filename, "%s/jpn/%s", 
							ax_system_dir, NXLT("uni2sjis.dat"));
			fp = fopen(filename, "r");
			if (fp == (FILE*)NULL) {
				status = UNI2SJIS_TABLE_FAILED;
				goto failed;
			}
		}

		/* Allocate enough for table from 0x0 to 0xFFFF */
		table = (unsigned int*)calloc( 65536 * sizeof (unsigned int), 1);
		if (!table) {
			status = UNI2SJIS_TABLE_FAILED;
			goto failed;
		}
		while( fgets( buf, 1000, fp)) {
			++lines;
			if (*buf && (*buf != '#')) {
				if ((sscanf(buf, "%x %x", &uni_table, &sjis_table)) == 2) {
					if (uni_table > 0xffff) {
						continue;			/* Too big */
					}
					if (sjis_table > 0xffff) {
						continue;			/* Too big */
					}
					if (table[uni_table]) {
						continue;			/* Duplicate values */
					}
					table[uni_table] = sjis_table;
				}
			}
		}
		fclose (fp);
		status = UNI2SJIS_TABLE_READ;
	}

	if (status == UNI2SJIS_TABLE_READ) {
		return( table[uni & 0xffff]);
	}

failed:
	return(BAD_SJIS);

}


#define SHIFT_IN	0x0e
#define SHIFT_OUT	0x0f
#define ESCAPE		0x1b
#define NEW_LINE	0x0a
#define IN_ASCII	0
#define IN_KANJI	1
#define IN_KANA		2
#define ESCAPE_NEW	"\033$B"
#define ESCAPE_OLD 	"\033$@"
#define ESCAPE_ROM	"\033(J"
#define ESCAPE_ASC	"\033(B"
#define ESCAPE_HAN	"\033(I"

MBeuc2jis_str(inP, outP)
mbchar  *inP;
mbchar  *outP;
{
	int	state = IN_ASCII;
	mbchar	*out = outP;
	int	new_state;

	while (*inP) {
		if (IS_MULTI_BYTE(inP)) {
			if ((UC)*inP == 0x8e)
				new_state = IN_KANA;
			else
				new_state = IN_KANJI;
		} else {
			new_state = IN_ASCII;
		}

		if (state != new_state) {
			/* close old state */
			if (state == IN_KANJI) {
				MBstrcpy(outP, ESCAPE_ROM);  /* 21Apr98 rjd Was ESCAPE_OUT */
				outP += 3;
			} else if (state == IN_KANA) {
				*outP++ = SHIFT_OUT;
			}
			/* change state */
			state = new_state;
			/* open new state */
			if (state == IN_KANJI) {
				MBstrcpy(outP, ESCAPE_NEW);  /* 21Apr98 rjd Was ESCAPE_IN */
				outP += 3;
			} else if (state == IN_KANA) {
				*outP++ = SHIFT_IN;
			}
		}

		if (state == IN_KANJI) {
			*outP++ = *inP++ & 0x7f;
			*outP++ = *inP++ & 0x7f;
		} else if (state == IN_KANA) {
			/* skip the leading byte 0x8e */
			inP++;
			*outP++ = *inP++ & 0x7f;
		} else {
			*outP++ = *inP++;
		}
	}

	if (state == IN_KANJI) {
		/* close KANJI */
		MBstrcpy(outP, ESCAPE_ROM);  /* 21Apr98 rjd Was ESCAPE_OUT */
		outP += 3;
	} else if (state == IN_KANA) {
		/* close KANA */
		*outP++ = SHIFT_OUT;
	}
	*outP = '\0';
	return(outP - out);
}

MBjis2euc_str(inP, outP)
mbchar  *inP;
mbchar  *outP;
{
	mbchar	*out = outP;
	static int	state = IN_ASCII;

	while (*inP) {
		if (*inP == ESCAPE) {
			if (strncmp(inP, ESCAPE_NEW, strlen(ESCAPE_NEW)) == 0) {
				inP += strlen(ESCAPE_NEW);
				state = IN_KANJI;
			} else if (strncmp(inP, ESCAPE_OLD, strlen(ESCAPE_OLD)) == 0) {
				inP += strlen(ESCAPE_OLD);
				state = IN_KANJI;
			} else if (strncmp(inP, ESCAPE_ROM, strlen(ESCAPE_ROM)) == 0) {
				inP += strlen(ESCAPE_ROM);
				state = IN_ASCII;
			} else if (strncmp(inP, ESCAPE_ASC, strlen(ESCAPE_ASC)) == 0) {
				inP += strlen(ESCAPE_ASC);
				state = IN_ASCII;
			} else if (strncmp(inP, ESCAPE_HAN, strlen(ESCAPE_HAN)) == 0) {
				inP += strlen(ESCAPE_HAN);
				state = IN_KANA;
			} else {
				/* corrupted codes, ignore the ESCAPE code */
				inP++;
			}
		} else if (*inP == SHIFT_IN) {
			inP++;
			state = IN_KANA;
		} else if (*inP == SHIFT_OUT) {
			inP++;
			state = IN_ASCII;
		} else if (*inP == NEW_LINE) {
			*outP++ = *inP++;	/* ignore the state */
		} else {	/* text */
			if (state == IN_KANJI) {
				*outP++ = *inP++ | 0x80;
				*outP++ = *inP++ | 0x80;
			} else if (state == IN_KANA) {
				*outP++ = 0x8e;
				*outP++ = *inP++ | 0x80;
			} else {
				*outP++ = *inP++;
			}
		}
	}
	*outP = '\0';
	return(outP - out);
}

MBsjis2jis_str(inP, outP)
mbchar  *inP;
mbchar  *outP;
{
	int	state = IN_ASCII;
	mbchar	*out = outP;
	int	new_state;

	while (*inP) {
		if (IS_SHIFT_JIS(inP)) {
			if (IS_SJIS_NARROW_KANA(inP))
				new_state = IN_KANA;
			else
				new_state = IN_KANJI;
		} else {
			new_state = IN_ASCII;
		}

		if (state != new_state) {
			/* close old state */
			if (state == IN_KANJI) {
				MBstrcpy(outP, ESCAPE_ROM);  /* 21Apr98 rjd Was ESCAPE_OUT */
				outP += 3;
			} else if (state == IN_KANA) {
				*outP++ = SHIFT_OUT;
			}
			/* change to new state */
			state = new_state;
			/* open new state */
			if (state == IN_KANJI) {
				MBstrcpy(outP, ESCAPE_NEW);  /* 21Apr98 rjd Was ESCAPE_IN */
				outP += 3;
			} else if (state == IN_KANA) {
				*outP++ = SHIFT_IN;
			}
		}

		if (state == IN_KANJI) {
			if ((UC)*inP >= 0x81 && (UC)*inP <= 0x9f)
				*outP = (mbchar)(((UC)*inP << 1) - 0xe1);
			else
				*outP = (mbchar)(((UC)*inP << 1) - 0x161);
			if ((UC)*(inP+1) >= 0x9f)
				(*outP)++;
			inP++;
			outP++;
			if ((UC)*inP >= 0x40 && (UC)*inP <= 0x7e)
				*outP = (mbchar)((UC)*inP - 0x1f);
			else if ((UC)*inP >= 0x80 && (UC)*inP <= 0x9e)
				*outP = (mbchar)((UC)*inP - 0x20);
			else
				*outP = (mbchar)((UC)*inP - 0x7e);
			inP++;
			outP++;
		} else if (state == IN_KANA) {
			*outP++ = *inP++ & 0x7f;
		} else {
			*outP++ = *inP++;
		}
	}

	if (state == IN_KANJI) {
		/* close KANJI */
		MBstrcpy(outP, ESCAPE_ROM);  /* 21Apr98 rjd Was ESCAPE_OUT */
		outP += 3;
	} else if (state == IN_KANA) {
		/* close KANA */
		*outP++ = SHIFT_OUT;
	}
	*outP = '\0';
	return(outP - out);
}

MBjis2sjis_str(inP, outP)
mbchar  *inP;
mbchar  *outP;
{
	mbchar	*out = outP;
	static int	state;

	while (*inP) {
		if (*inP == ESCAPE) {
			if (strncmp(inP, ESCAPE_NEW, strlen(ESCAPE_NEW)) == 0) {
				inP += strlen(ESCAPE_NEW);
				state = IN_KANJI;
			} else if (strncmp(inP, ESCAPE_OLD, strlen(ESCAPE_OLD)) == 0) {
				inP += strlen(ESCAPE_OLD);
				state = IN_KANJI;
			} else if (strncmp(inP, ESCAPE_ROM, strlen(ESCAPE_ROM)) == 0) {
				inP += strlen(ESCAPE_ROM);
				state = IN_ASCII;
			} else if (strncmp(inP, ESCAPE_ASC, strlen(ESCAPE_ASC)) == 0) {
				inP += strlen(ESCAPE_ASC);
				state = IN_ASCII;
			} else if (strncmp(inP, ESCAPE_HAN, strlen(ESCAPE_HAN)) == 0) {
				inP += strlen(ESCAPE_HAN);
				state = IN_ASCII;
			} else {
				/* corrupted codes, ignore the ESCAPE code */
				inP++;
			}
		} else if (*inP == SHIFT_IN) {
			inP++;
			state = IN_KANA;
		} else if (*inP == SHIFT_OUT) {
			inP++;
			state = IN_ASCII;
		} else if (*inP == NEW_LINE) {
			*outP++ = *inP++;	/* ignore the state */
		} else {	/* text */
			if (state == IN_KANJI) {
				if ((UC)*inP <= 0x5e)
					*outP = (mbchar)(((UC)(*inP) + 1) / 2 + 0x70);
				else
					*outP = (mbchar)(((UC)(*inP) + 1) / 2 + 0xb0);
				inP++;
				outP++;
				if (*(inP-1) & 1) {
					*outP = (mbchar)((UC)*inP + 0x1f);
					if ((UC)*outP >= 0x7f)
						(*outP)++;
				} else
					*outP = (mbchar)((UC)*inP + 0x7e);
				inP++;
				outP++;
			} else if (state == IN_KANA) {
				*outP++ = *inP++ | 0x80;
			} else {
				*outP++ = *inP++;
			}
		}
	}
	*outP = '\0';
	return(outP - out);
}

is_pure_ascii_str(inP)
mbchar	*inP;
{
	if (!inP)
		return(TRUE);
	while (*inP) {
		if (*inP & 0x80)
			return(FALSE);
		else
			inP++;;
	}
	return(TRUE);
}

is_euc_str(inP)
mbchar	*inP;
{
	if (!inP)
		return(FALSE);
	while (*inP) {
		if ((*inP & 128) == 0)
			inP++;
		else if (IS_MULTI_BYTE(inP))
			inP +=2;
		else
			return(FALSE);
	}
	return(TRUE);
}

is_sjis_str(inP)
mbchar	*inP;
{
	if (!inP || !*inP)
		return(FALSE);
	while (*inP) {
		if ((*inP & 128) == 0)
			inP++;
		else if (IS_SJIS_NARROW_KANA(inP))
			inP++;
		else if (IS_SHIFT_JIS(inP))
			inP += 2;
		else
			return(FALSE);
	}
	return(TRUE);
}

is_jis_str(inP)
mbchar  *inP;
{
	if ( strstr(inP, ESCAPE_NEW) || strstr(inP, ESCAPE_OLD) ||
			strstr(inP, ESCAPE_ROM) || strstr(inP, ESCAPE_ASC) ||
			strstr(inP, ESCAPE_HAN) || strchr(inP, SHIFT_IN) )
		return(TRUE);
	else
		return(FALSE);
}

#define BASE64		"B"
#define QP			"Q"
#define ISO_2022_JP	"ISO-2022-JP"
#define US_ASCII	"US-ASCII"
#define B_US_ASCII	1
#define B_JP_ASCII	2
#define B_JP_KANA	3
#define B_JP_KANJI1	4
#define B_JP_KANJI2	5
static char base16[] = "0123456789ABCDEF";
static char base64[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
static char	*CODE_SWITCH[6] = {"", "\033(B", "\033(J", "\033(I", "\033$@", "\033$B"};

index64(c)
char	c;
{
	char	*p;

	p = strchr(base64, c);
	if (p)
		return(p - base64);
	else
		return(-1);
}

hexchar(ch)
char ch;
{
	char *p;

	p = strchr(base16, toupper(ch));
	if (p == NULL)
		return(-1);
	else
		return(p-base16);
}

is_wraped_mime(string)
char	*string;
{
	char	*checkP = string;

	while (isspace(*checkP))
		checkP++;
	if (strncmp(checkP, "=?", 2))
		return(FALSE);
	checkP += 2;
	if (strncmp(checkP, ISO_2022_JP, strlen(ISO_2022_JP)))
		return(FALSE);
	checkP += strlen(ISO_2022_JP);
	if (*checkP++ != '?')
		return(FALSE);
	if (strncmp(checkP, BASE64, strlen(BASE64)))
		return(FALSE);
	checkP += strlen(BASE64);
	if (*checkP++ != '?')
		return(FALSE);
	if (strstr(checkP, "?=") == NULL)
		return(FALSE);

	return(TRUE);

}

fold_mime(src, dst, stop, encoding)
char	*src;
char	*dst;
char	*stop;
char	*encoding;
{
	char	c1, c2, c3, c4;
	int		a1, a2, a3, a4;

	if (strcasecmp(encoding, BASE64) == 0) {
		while (src < stop) {
			while(isspace(c1 = *src++));
			while(isspace(c2 = *src++));
			while(isspace(c3 = *src++));
			while(isspace(c4 = *src++));

			if ((c1 == '=') || (c2 == '='))
				break;

			a1 = index64(c1);
			a2 = index64(c2);
			*dst++ = (a1 << 2) | (a2 >> 4);

			if (c3 == '=')
				break;

			a3 = index64(c3);
			*dst++ = ((a2 & 0x0f) << 4) | (a3 >> 2);

			if (c4 == '=')
				break;

			a4 = index64(c4);
			*dst++ = ((a3 & 0x03) << 6) | a4;
		}
	} else if (strcasecmp(encoding, QP) == 0) {
		int sawnewline = 1, neednewline = 0;
		while (src < stop) {
            if (neednewline) {
                *dst++ = '\n';
                neednewline = 0;
            }
			c1 = *src++;
            if (c1 == '=') {
                sawnewline = 0;
                c1 = *src++;
                if (c1 == '\n') {
                    /* ignore it */
                    sawnewline = 1;
                } else {
                    c2 = *src++;
                    a1 = hexchar(c1);
                    a2 = hexchar(c2);
                    *dst++ = (char)(a1 << 4 | a2);
                    if (c2 == '\n')
						sawnewline = 1;
                }
            } else {
                if (c1 == '\n') {
                    sawnewline = 1;
                    neednewline = 1;
                } else {
                    sawnewline = 0;
                    *dst++ = c1;
                }
            }
		}
    	if (neednewline) {
        	*dst++ = '\n';
        	neednewline = 0;
    	}    
	} else {
		*dst = '\0';
		return(FALSE);
	}
	*dst = '\0';
	return(TRUE);
}

/* Assume the "in" string here is always has the form of :
 *     [<ESCAPE>XY]<ESCAPE>$@......<ESCAPE>(J
 *	The '.'s representing encoded characters.
 * Resolve the internal eating space chars problems.
 */
folded_to_jis(src, dst, eat_prev, eat_post)
char	*src;
char	*dst;
int		*eat_prev, *eat_post;
{
	char	c;
	int		bset;

	*eat_prev = FALSE;
	*eat_post = FALSE;
	while ((c = *src++)) {
		if (c != ESCAPE)
			goto end_escape;
start_escape:
		switch (*src++) {	/* ISO-2022-JP */
			case '(':
				switch (*src++) {
					case 'B':
						bset = B_US_ASCII;
						break;
					case 'J':
						bset = B_JP_ASCII;
						break;
					case 'I':
						bset = B_JP_KANA;
						break;
					default:
						goto failed;
				}
				break;
			case '$':
				switch (*src++) {
					case '@':
						bset = B_JP_KANJI1;
						break;
					case 'B':
						bset = B_JP_KANJI2;
						break;
					default:
						goto failed;
				}
				break;
			default:
				goto failed;
		}
		if ((c = *src++) == ESCAPE) {
			switch (bset) {
				case B_US_ASCII:
					*eat_prev = FALSE;
					*eat_post = FALSE;
					break;
				case B_JP_ASCII:
					*eat_prev = TRUE;
					*eat_post = TRUE;
					break;
				case B_JP_KANA:
					*eat_prev = TRUE;
					*eat_post = TRUE;
					break;
				case B_JP_KANJI1:
					*eat_prev = TRUE;
					*eat_post = FALSE;
					break;
				case B_JP_KANJI2:
					*eat_prev = FALSE;
					*eat_post = TRUE;
					break;
				default:
					break;
			}
			goto start_escape;
		} else {
			MBstrcpy(dst, CODE_SWITCH[bset]);
			dst += strlen(dst);
		}
end_escape:
		if (c)
			*dst++ = c;
		else
			break;
	}
	*dst = 0;
	return(TRUE);
failed:
	MBstrcpy(dst, src);
	return(FALSE);	
}

int
fold_mime_to_jis(src, dst, stopP, eat_prev, eat_post)
char	*src, *dst, **stopP;
int		*eat_prev, *eat_post;
{
	char	*checkP;
	char	charset[80];
	char	encoding[16];
	char	*tmpbuf = NULL;
	char	*stop = *stopP;

	/* do check if it is a real Japanese MIME */
	checkP = src + 2;
	if (strncasecmp(checkP, ISO_2022_JP, strlen(ISO_2022_JP)) == 0)
		MBstrcpy(charset, ISO_2022_JP);
	else if (strncasecmp(checkP, US_ASCII, strlen(US_ASCII)) == 0)
		MBstrcpy(charset, US_ASCII);
	else
		goto failed;
	checkP += strlen(charset);
	if (*checkP++ != '?')
		goto failed;
	if (strncasecmp(checkP, BASE64, strlen(BASE64)) == 0)
		MBstrcpy(encoding, BASE64);
	else if (strncasecmp(checkP, QP, strlen(QP)) == 0)
		MBstrcpy(encoding, QP);
	else
		goto failed;
	checkP += strlen(encoding);
	if (*checkP++ != '?')
		goto failed;

	/* passed all checks. */
	/* copy Japanese mime encoding start */
	src = checkP;
	stop = strstr(src, "?=");

	/* unfold MIME formatted string */
	tmpbuf = (char *)malloc(3*(stop-src));
	if (fold_mime(src, tmpbuf, stop, encoding) == FALSE)
		goto failed;
	if (folded_to_jis(tmpbuf, dst, eat_prev, eat_post) == FALSE)
		goto failed;

	*stopP = stop + 2;
	free(tmpbuf);
	return(TRUE);

failed:
	if (tmpbuf)
		free(tmpbuf);
	return(FALSE);
}

MBmime2jis(src, dst)
char	*src, *dst;
{
	char	*mime;
	int		eat_prev, eat_post;
	char	*tmpbuf;

	tmpbuf = (char *)malloc(strlen(src));

	while (mime = strstr(src, "=?")) {

		/* (1) copy ascii part */
		while (src < mime)
			*dst++ = *src++;

		/* (2) if not really MIME, copy the remainder and return */
		if (strstr(src, "?=") == NULL) {
			MBstrcpy(dst, src);
			return(FALSE);
		}

		/* (3) convert pure MIME part to JIS encoding and copy it */
		src = strstr(mime, "?=") + 2;
		if (fold_mime_to_jis(mime, tmpbuf, &src, &eat_prev, &eat_post)) {
		/* (4) eat left/right space characters conditionally */
			if (eat_prev && isspace(*(dst-1)))
				dst--;
			if (eat_post && isspace(*src))
				src++;
			MBstrcpy(dst, tmpbuf);
			dst += strlen(dst);
		} else {
			strncpy(dst, mime, src-mime);
			dst += src-mime;
		}

	}
	MBstrcpy(dst, src);
	free(tmpbuf);
	return(TRUE);
}

char *
unfold_jis_to_mime(sP, dP, stopP)
char	*sP, *dP, *stopP;
{
	int		a1, a2, a3;

	/* copy Japanese mime encoding start */
	sprintf(dP, "=?%s?%s?", ISO_2022_JP, BASE64);
	dP += strlen(dP);

	/* convert pure encoded (JIS) string to MIME_JP format */
	while (sP < stopP) {
		a1 = *sP++;
		*dP++ = base64[a1>>2];
		if (sP == stopP) {
			*dP++ = base64[(a1&0x03)<<4];
			*dP++ = '=';
			*dP++ = '=';
		} else {
			a2 = *sP++;
			*dP++ = base64[((a1&0x03)<<4)|(a2>>4)];
			if (sP == stopP) {
				*dP++ = base64[(a2&0x0f)<<2];
				*dP++ = '=';
			} else {
				a3 = *sP++;
				*dP++ = base64[((a2&0x0f)<<2)|(a3>>6)];
				*dP++ = base64[a3&0x3f];
			}
		}
	}

	/* copy Japanese mime encoding end */
	MBstrcpy(dP, "?= ");
	dP += strlen(dP);
	return(dP);
}

MBjis2mime(oldstr, newstr)
char	*oldstr, *newstr;
{
	char	*jisP;
	char	*srcP = oldstr;
	char	*dstP = newstr;

	while (jisP = strpbrk(srcP, NXLT("\033\016"))) {  /* 21Apr98 rjd Was ESCAPE, and w/o NXLT */

		/* (1) copy ascii part */
		while (srcP < jisP)
			*dstP++ = *srcP++;

		/* (2) find out the pure JIS piece */
		while (*srcP) {
			if (*srcP == ESCAPE) {
				if (strncmp(srcP, ESCAPE_NEW, strlen(ESCAPE_NEW)) == 0) {
					srcP += strlen(ESCAPE_NEW);
				} else if (strncmp(srcP, ESCAPE_OLD, strlen(ESCAPE_OLD)) == 0) {
					srcP += strlen(ESCAPE_OLD);
				} else if (strncmp(srcP, ESCAPE_HAN, strlen(ESCAPE_HAN)) == 0) {
					srcP += strlen(ESCAPE_HAN);
				} else {
					/* not real jis, bad control code mixed */
					MBstrcpy(dstP, jisP);  /* 21Apr98 rjd Was strcpy */
					return(FALSE);
				}
				srcP = strchr(srcP, ESCAPE);
				if (srcP == NULL) {
					MBstrcpy(dstP, jisP);  /* 21Apr98 rjd Was strcpy */
					return(FALSE);
				}
				if (strncmp(srcP, ESCAPE_ROM, strlen(ESCAPE_ROM)) == 0) {
					srcP += strlen(ESCAPE_ROM);
				} else if (strncmp(srcP, ESCAPE_ASC, strlen(ESCAPE_ASC)) == 0) {
					srcP += strlen(ESCAPE_ASC);
				} else {
					MBstrcpy(dstP, jisP);  /* 21Apr98 rjd Was strcpy */
					return(FALSE);
				}
			} else {
				if (strchr(srcP+1, SHIFT_OUT) == NULL) {
					MBstrcpy(dstP, jisP);  /* 21Apr98 rjd Was strcpy */
					return(FALSE);
				} else {
					srcP = strchr(srcP+1, SHIFT_OUT) + 1;
				}
			}
			if ((*srcP != ESCAPE) && (*srcP != SHIFT_IN))
				break;
		}

		/* (3) convert pure JIS part to mime format and copy it */
		dstP = unfold_jis_to_mime(jisP, dstP, srcP);
	}
	MBstrcpy(dstP, srcP);
	return(TRUE);
}



/* return HTML encoding for Japanese:
 *  0:Unknown, 1:ASCII, 2:EUC, 3:Shift-JIS, 4:JIS.
 */
int html_kanji_encoding()
{
    static int ax_html_kanji_encoding = -1;
    char *code_type;

    if ( ax_html_kanji_encoding == -1 ) {
       /* initialize from environment one time only */
       code_type = (char *)getenv("AXHTMLKANJI");
       if (code_type) {
          if (strcasecmp(code_type, "ascii") == 0)
              ax_html_kanji_encoding = 1;
          else if (strcasecmp(code_type, "euc") == 0)
              ax_html_kanji_encoding = 2;
          else if (strcasecmp(code_type, "sjis") == 0)
              ax_html_kanji_encoding = 3;
          else if (strcasecmp(code_type, "jis") == 0)
              ax_html_kanji_encoding = 4;
          else
              ax_html_kanji_encoding = 0;
       } else {
          if (ascii_is_shift_jis())
             ax_html_kanji_encoding = 3;
          else
             ax_html_kanji_encoding = 2;
       }
    }
    return( ax_html_kanji_encoding );
}

/*
 * Ashisuto created for automatic checking multi-byte string Japanese encoding.
 * (EUC, Shift-JIS, JIS)    96.03.08.
 * Starting at first time for importing Japanese HTML file.
 * This method will be used until we implement "CodePage" for Japanese version.
 *
 * Check the mbstr encoding and return the original mbstr encoding type.
 * Change the mbstr to Applixware internal encoding if change_it flag is set.
 * The type_hint parameter is used for overlapping encoding judgement.
 * Return value:
 *  0:Unknown, 1:ASCII, 2:EUC, 3:Shift-JIS, 4:JIS.
 * Caution: If change_it flag is set, the buffer size of mbstr must be large
 *  enough to contain the converted new mbstr. (The buffer may be not enough
 *  only on JIS encoding, and long Narrow Katakana char strings. These are
 *  rare cases. Most likely happenning in spreadsheet's or table's cell
 *  strings where are in lack of display spaces.)
 */
int
CheckEncodingMbStr(mbstr, change_it)
mbchar *mbstr;
int change_it;
{
    int     type_hint;
    int     type_ret;
    mbchar  *mbP = mbstr;
    char    *kanjiCodeStr;
    mbchar  *mb_buf;
 
    type_hint = html_kanji_encoding();
    type_ret = 0;
    /* Find first char from where we can do judge */
    while (*mbP) {
        if (*mbP == ESCAPE) /* may be JIS */
            break;
        if ((UC)*mbP > 0x80) {
            if (IS_MULTI_BYTE(mbP))
                if (IS_SHIFT_JIS(mbP) && !IS_SJIS_NARROW_KANA(mbP)) {
                    if (type_ret == 0)
                        type_ret = 5;   /* EUC or SJIS, half determined. */
                    mbP++;          /* double byte char */
                } else if (IS_SJIS_NARROW_KANA(mbP) && type_hint == 3) {
                    type_ret = 5;   /* EUC or SJIS, half determined. */
                } else {
                    type_ret = 2;
                    break;
                }
            else if (IS_SHIFT_JIS(mbP)) {
                type_ret = 3;
                break;
            }
        }
        mbP++;
    }
 
    /* Judge str encoding type */
    if (type_ret == 0) {        /* ASCII or JIS or Unknown */
        if (*mbP == 0)          /* End of mbstr, single byte string */
            type_ret = 1;       /* ASCII */
        else                    /* Now *mbP == ESCAPE, JIS ? */
            if (strchr((char*)(mbP+1), ESCAPE))
                type_ret = 4;   /* JIS, still a rough judge */
            else
                type_ret = 0;   /* Unknown */
    } else if (type_ret == 5) {
        if ((type_hint == 2) || (type_hint == 3))   /* all code overlapped */
            type_ret = type_hint;   /* EUC or Shift-JIS, using user setting */
        else
            type_ret = 2;       /* Applixware jpn default EUC encoding */
    }
 
    /* Do change */
    if (change_it) {
        switch (type_ret) {
            case 0:     /* Unknown */
            case 1:     /* ASCII */
            case 2:     /* EUC */
                break;
            case 3:     /* Shift-JIS */
                MBsjis2euc_str(mbstr);
                break;
            case 4:     /* JIS */
                mb_buf = (mbchar *)malloc(2*MBstrlen(mbstr)+1);
                MBjis2euc_str(mbstr, mb_buf);
                MBstrcpy(mbstr, mb_buf);
                free(mb_buf);
                break;
            default:
                break;
        }
    }
    return(type_ret);
}



 

#else     /*  } BUILD_KANJI  {   */



MBeuc2sjis_str()
{
return( FALSE );
}
MBsjis2euc_str()
{
return( FALSE );
}

mbchar *MBstrWchr()
{ return( (mbchar *)1); }

 


#endif    /*  }  else BUILD_KANJI */





#if BUILD_WIDE   /*  {  */
   /* these exist only in a wide build, ootherwise they are defined */
   /* back to the native equivalents                                */
int Wstrlen( wchar *p )
{
    int count = 0;
    
    if ( p == (wchar *)NULL )
       return( 0 );
    while ( *p++ != (wchar)0 )
       {
       count++;
       }
    return( count );
}

/*************************************************************************
function: WMbstreq - Compare a wide string, and a multi-byte string,
return TRUE if equal */

WMbstreq (s1,s2)
register wchar *s1;
register mbchar *s2;
{
    return(Wstreq(s1, MbToWOneStr(s2)));
}

/*************************************************************************
function: Wstreq - Compare two Wide strings, return TRUE if equal
*/
Wstreq (s1,s2)
register wchar *s1;
register wchar *s2;
{
    for ( ; *s1 == *s2; s1++,s2++ )
        {
        if ( *s1 == (wchar)0 )
            return(TRUE);
        }
    return(FALSE);
}



int32 Wstrncmp( str1, str2, n )
wchar *str1, *str2;
int n;
{
    if (str1 == str2)
        return(0);

    while ( n > 0 )
        {
        if ( *str1 != *str2 )
           return( ((unsigned int)*str1) - ((unsigned int)*str2) );
        str1++;
        str2++;
        n--;
        }

    /* strings are the same */
    return( 0 );
}



wchar *WMbstrcpy( base, add )
wchar *base;
mbchar *add;
{
    if ( add != (mbchar *)NULL )
      {
      MbToWStr( add, base );
      }

    return( base );
}

wchar *Wstrcat( base, add )
wchar *base, *add;
{
    wchar *p;

    p = base + Wstrlen( base );
    Wstrcpy( p, add );
    return( base );
}

wchar *Wstrncat( base, add, n )
wchar *base, *add;
int		n;
{
    wchar *p;
    p = base + Wstrlen( base );
    Wstrncpy( p, add, n );
    return( base );
}


int32 Wstrcmp( str1, str2 )
wchar *str1, *str2;
{
    if (str1 == str2)
        return(0);

    while ( (*str1 != 0) && (*str2 != 0) ) 
        {
        if ( *str1 != *str2 )
           break;
        str1++;
        str2++;
        }

    return( *str1 - *str2 );
}



wchar * Wstrcpy( destination, source )
wchar *destination, *source;
{
    register wchar *dest;

    if (destination == source)
        return(destination);

    dest = destination;
    *dest = (wchar)0;
    if ( source == (wchar *)NULL )
       return;
    while ( *source != (wchar)0 )
       {
       *dest++ = *source++;
       }
    *dest = (wchar)0;
    return( destination );
}



wchar *Wstrncpy( destination, source, n )
wchar *destination, *source;
int		n;
{
    register wchar *dest;

    dest = destination;
    *dest = (wchar)0;

    if ( source == (wchar *)NULL )
       return;

    while ( *source != (wchar)0 && n != 0 )
       {
       *dest++ = *source++;
	   n--;
       }
    *dest = (wchar)0;
    return( destination );
}







wchar *Wstrpbrk( wstr, search_chars )
wchar *wstr;
wchar *search_chars;
{
    register wchar *searchP;
    while ( *wstr != 0 )
      {
      searchP = search_chars;
      while ( *searchP != 0 )
         {
         if ( *searchP == *wstr )
             return( wstr );
         searchP++;
         }
      wstr++;
      }
    /* fell through, did not find any char */
    return( (wchar *)NULL );
}



wchar *Wstrtok( str, seperators )
wchar *str;
mbchar *seperators;
{
    static wchar *lastP;
    wchar *p;
    wchar *ret;

    if ( str != (wchar *)NULL )
       {
       /* a new string to be processed */
       lastP = str;
       }

    if ( *lastP == (wchar)0 )
       {
       /* nothing more in the string */
       return( (wchar *)NULL );
       }

    p = lastP;

    /* skip any leading seperators */
    while ( (*p != (wchar)NULL) && 
            (MBstrchr( (char *)seperators, (int)*p ) != NULL) )
       {
       p++;
       }
    if ( *p == (wchar)0 )
       {
       /* no more tokens in the string */
       return( (wchar *)NULL );
       }

    ret = p;     /* return value */

    /* now find the end of this token */
    while ( (*p != (wchar)NULL) &&
            (MBstrchr( (char *)seperators, (int)*p ) == NULL) )
       {
       p++;
       }
    if ( *p == (wchar)NULL )
       /* hit the end of the string */
       lastP = p;
    else
       /* still more in the string, start with the next character next time */
       lastP = p+1;

    *p = (wchar)0;       /* terminate the token */

    return( ret );
}



wchar *Windex( wstr, target )
wchar *wstr;
unsigned int target;
{
    while ( *wstr && (*wstr != target) )
       { wstr++; }

    if ( *wstr == (wchar)0 )
       return( (wchar *)NULL );
    else
       return( wstr );
}



wchar *Wrindex( wstr, target )
wchar *wstr;
unsigned int target;
{
    register wchar *endP;

    endP = wstr + Wstrlen(wstr)-1;

    while( endP >= wstr )
       {
       if ( *endP == target )
          break;
       endP--;
       }

    if ( endP < wstr )
       return( (wchar *)NULL );
    else
       return( endP );
}




#endif          /*  } BUILD_WIDE  */



/************************************************************************
  These functions are needed on both wide and non-wide builds
***/
/* this is really only converting from UTF to generic multi-byte;
 * it just happens to be be regular ascii or sjis because 
 */

int
MButf2sjis(char *inP, char *outP)
{
	char ch;
    char *istr, *ostr;
    int x, y, z, val;
 
    if ( inP && outP )
    {
        istr = inP;
        ostr = outP;
        *istr++;    /* skip past length bytes that writeUTF() writes out */
        while ( (*istr & 0x000000FF) != 0xC0 &&
				(*(istr+1) & 0x000000FF) != 0x80 ) /* NULL is represented by
													* these 2 bytes.
													*/
        {
			ch = *istr;
            if ( !(ch & 0x80) )
            {
                /* 7-bit ascii */
                *ostr++ = *istr;
            }
            else if ( ((ch & 0x000000FF) << 5) == 0x6 )
            {
                /* convert from 2 utf bytes to 2 sjis bytes */
                x = ((*istr++ & 0x000000FF) - 192) << 6;
                y = ((*istr & 0x000000FF) - 128);
                val = x + y;
 
                *ostr++ = (val & 0xFF00) >> 8;
                *ostr++ = (val & 0xFF);
            }
            else if ( ((ch & 0x000000FF) >> 5) == 0x7 )
            {
                /* convert from 3 utf bytes to 2 sjis bytes */
                x = ((*istr++ & 0x000000FF) - 224) << 12;
                y = ((*istr++ & 0x000000FF) - 128) << 6;
                z = ((*istr & 0x000000FF) - 128);
                val = x + y + z;
 
                *ostr++ = (val & 0xFF00) >> 8;
                *ostr++ = (val & 0xFF);
            }
            *istr++;
        }
        *ostr = 0;
    }
 
    return 0;
}
/***********************************************************************/
