/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axconn.h
   Applix ASC version: 10
   Date: Mon Nov 16 13:53:45 1998
*/
#ifndef AXCONN_H
#define AXCONN_H

#include <limits.h>
#include "axhtab.h"

#define CENTRAL_EV 1      /* central (uniprocess) event management */
#define MAXEVENT INT_MAX  /* maximum allowable event number */

/* a conn instance
 */
typedef struct axconn_t {
	/* link arrays
	 */
	int numLinkTypes;
	axhtab typeTable;       /* link-type-name -> linkTables index */
	axhtab *linkTables;     /* link-name -> link instance */
	/* event info
	 */
	int evnum;            /* # of last handled event */
	axhtab evTable;       /* linkname -> last event# from that link */
} axconn;

/* an axconn "scope"
 */
typedef struct axconn_c {
	void *name;                          /* scope name */
	int namelen;
	axconn *            (*conn)();       /* construct */
	void                (*dconn)();      /* destruct */
	int                 (*link)();       /* construct a link */
	void                (*dlink)();      /* destruct a link */
	int                 (*locate)();     /* object-name -> handle */
	void *              (*linkCall)();   /* link-specific function */
#if CENTRAL_EV
	int                 (*newEvent)();   /* produce a new event# */
#endif
	axhtab                locators;      /* object-name -> its locator */
	axhtab                linkFuncs;     /* link-type-name -> link func */
#if CENTRAL_EV
	int                    evnum;        /* # of next event */
#endif
} axconnC;

/* a link instance
 */
typedef struct axlink_t {
	unsigned int direction : 1;  /* point-er sees true, pointee sees false */
	axconnC *scope;              /* name scope et al */
	int linkType;                /* user-defined */
	void *data;                  /* ptr to link-specific data */
} axlink;

/* a reference to a remote object
 */
typedef struct remoteRef_t {
	void *rHandle;
	axconn **rConn;
} remoteRef;

/* link_funcs_t associates a linktype's link-function,
 * link-data-copier, and link-data-destructor symbols
 */
typedef struct link_funcs_t {
	void *(*func)();              /* for doing link work */
	void *(*linkDataCopier)();    /* for copying link data */
	void (*linkDataDestructor)(); /* for destroying link data */
} link_funcs;

#endif    /* AXCONN_H */

