/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axdnd.c
   Applix ASC version: 21
   Date: Mon Jan 11 16:28:31 1999
*/
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
/* Although in ``/ax/pgms/axatmrt/'', this header file is treated as a
   part of system header files. */
#include <DrgDrpLib.h>

#include <sys/stat.h>
#include <sys/param.h>

#if SUPPORT_HP_VUE
/* *Must* be before aos.h */
#define _axt_h_
#define _NO_PROTO
#include <X11/Intrinsic.h>
#if HP800V3
#include <Motif1.2/Xv/Drag.h>
#include <Motif1.2/Xv/Init.h>
#else
#include <Motif1.1/Xv/Drag.h>
#include <Motif1.1/Xv/Init.h>
#endif /* not HP800V3 */
#endif

#include <aos.h>
#include <axdnd.h>
#include <usrevent.h>
#include <scac.h>
#include <xmlib.h>


extern short DpyWid, DpyHt;
extern Display *Dpy;
extern unsigned long FGpixel;
extern unsigned long BGpixel;


/* AxNoMessaging is TRUE if we want to skip hp-vue drag-n-drop. If TRUE
   does not connect with the Broadcast Message Server (BMS));
   FALSE (default: drag-n-drop enabled).
   This flag is set to TRUE in axmain.c by looking at the -noMessaging
   flag. */
int AxDndNoMessaging = FALSE;

#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN 50
#endif

#if 0
/* temporary */
#define TMPLOG(m) logit m
#else
#define TMPLOG(m)
#endif

/* Drop Item description (local to axdnd.c) */
typedef struct drop_item_t {
	char name[STRPATH + 1]; /* file name */
	int  type;              /* file type */
	char host[MAXHOSTNAMELEN + 1]; /* host name */
	struct drop_item_t *next;
} drop_item_t;

/* Registered window description (local to axdnd.c): Used for tracking
   purposes. */
typedef struct regwin_t {
	Display *dpy; /* X Display connection */
	Window win;   /* X window id */
	int pid;      /* ax task id */
	int wid;      /* ATM window id */
	struct regwin_t *next;
	struct regwin_t *prev;
} regwin_t;

#if SUPPORT_HP_VUE
typedef struct hp_vue_droplist_t {
	int  type;
	char name[STRPATH + 1];
	char machine[MAXHOSTNAMELEN + 1];
	struct hp_vue_droplist_t *next;
} hp_vue_droplist_t;
#endif

#if (SUPPORT_IXI || SUPPORT_OLWM)
static Atom AxXA_FILE_NAME = None;
#endif

#if SUPPORT_IXI
/* Per display context */
static DrgDrpContext ixi_dd_context;
#endif

#if SUPPORT_OLWM
/* Selection owned by the Drop Site Database Manager (DSDM) */
static Atom AxXA_SUN_DRAGDROP_DSDM     = None;
static Atom AxXA_SUN_DRAGDROP_INTEREST = None;
static Atom AxXA_SUN_DRAGDROP_TRIGGER = None;
static Atom AxXA_SUN_DRAGDROP_DONE  = None;
static Atom AxXA_SUN_DRAGDROP_ACK   = None;
static Atom AxXA_DRAGDROP_PROP = None;
#endif

/* Type of desktop manager we are running under. */
static int AxDeskFlag = 0;

/* Head of registered window list */
static regwin_t *RegWinHead = (regwin_t *) NULL;

#if SUPPORT_IXI
static void ixi_initialize(/* Display *dpy */);
static void ixi_register_window(/* Display *dpy, Window win, Window
								   target */);
static void ixi_drop_callback(/*void *closure, Display *dpy,   Window win,
								Time time,  unsigned int x, unsigned int y,
								DrgDrpStyle styles[], unsigned int n_styles,
								DrgDrpInfo data[], unsigned int n_formats */);
static void ixi_unregister_window(/* Display *dpy, Window win */);
static int  ixi_filter_event(/* XEvent *ev */);
#endif

#if SUPPORT_OLWM
static void olwm_initialize(/* Display *dpy */);
static void olwm_register_window(/* Display *dpy, Window win,
									Window target */);
static void olwm_unregister_window(/* Display *dpy, Window win */);
static int  olwm_filter_event(/* XEvent *ev */);
#endif

#if SUPPORT_HP_VUE
static void hp_vue_initialize(/* Display *dpy */);
static hp_vue_droplist_t *hp_vue_parse_droplist(/* char *types_str,
												   char *objects_str*/);
static void hp_vue_drop_callback(/* Display *dpy, Window win, int x, int y,
									Window root, int root_x, int root_y,
									unsigned int mods, String types,
									String objects */);
static void hp_vue_destroy_droplist(/* hp_vue_droplist_t *head */);
static void hp_vue_register_window(/* Display *dpy, Window win,
									  Window target */);
static void hp_vue_unregister_window(/* Display *dpy, Window win */);
#endif

static void add_to_regwin_list(/* Display *dpy, Window win */);
static void delete_from_regwin_list(/* Display *dpy, Window win */);
static unsigned int determine_desktop_manager(/* */);
static int  is_registered_window(/* Display *dpy, Window win */);
static int determine_file_type(/* char *name */);
static void poke_list_box(/* widget_t *w, char *hostname, char *filename*/);
static int send_drop_message_to_object(/* Window win, drop_item_t *dp */);

void
AxDndInitialize(dpy)
Display *dpy;
{
	static int dnd_init_done = FALSE;
	
	if (dnd_init_done)
		return;
	else
		dnd_init_done = TRUE;

	AxDeskFlag = determine_desktop_manager(dpy);
#if SUPPORT_IXI
	if (HAVE_IXI(AxDeskFlag))
		ixi_initialize(dpy);
#endif

#if SUPPORT_OLWM
	if (HAVE_OLWM(AxDeskFlag))
		olwm_initialize(dpy);
#endif

#if SUPPORT_HP_VUE
	if (HAVE_HP_VUE(AxDeskFlag))
		hp_vue_initialize(dpy);
#endif

	TMPLOG(("AxDndInitialize done.:\n"));
	return;
}

void
AxDndRegisterWindow(dpy, win, target)
Display *dpy;
Window  win;
Window  target;
{
	extern Window ShellWindowFromXWindow(); /* tmx.c */

#if SUPPORT_IXI
	if (HAVE_IXI(AxDeskFlag))
		ixi_register_window(dpy, win, target);
#endif

#if SUPPORT_OLWM
	if (HAVE_OLWM(AxDeskFlag)) {
		Window shell_win;

		olwm_register_window(dpy, win, target);

		/* Also register the shell window if there is one */
		if ((shell_win = ShellWindowFromXWindow(win)) != (Window) None)
			olwm_register_window(dpy, shell_win, target);
	}
#endif

#if SUPPORT_HP_VUE
	if (HAVE_HP_VUE(AxDeskFlag))
		hp_vue_register_window(dpy, win, target);
#endif
	TMPLOG(("AxDndRegisterWindow:\n"));
	TMPLOG(("\tdpy = <0x%lx>\twin = <0x%lx>\n", dpy, win));
	return;
}

void
AxDndUnregisterWindow(dpy, win)
Display *dpy;
Window  win;
{
	/* The unregister call in hp-vue 3.0 *needs* an "already
	   registered" dpy-win.  dnd will stop working if passed a dpy-win
	   pair that was not previously registered. */
	if (!is_registered_window(dpy, win))
		return;

#if SUPPORT_IXI
	if (HAVE_IXI(AxDeskFlag))
		ixi_unregister_window(dpy, win);
#endif

#if SUPPORT_OLWM
	if (HAVE_OLWM(AxDeskFlag))
		olwm_unregister_window(dpy, win);
#endif

#if SUPPORT_HP_VUE
	if (HAVE_HP_VUE(AxDeskFlag))
		hp_vue_unregister_window(dpy, win);
#endif

	delete_from_regwin_list(dpy, win);
	return;
}

/* Return `False' if event is consumed; `True' otherwise. */
int
AxDndHandleEvent(ev)
XEvent *ev;
{
	if (ev->xany.type == SelectionNotify) {
		/* Only olwm dragdrop uses selection mechanism */
		if (!HAVE_OLWM(AxDeskFlag))
			return(True);
	}
#if SUPPORT_IXI
	if (HAVE_IXI(AxDeskFlag))
		if (!ixi_filter_event(ev))
			return(False);
#endif

#if SUPPORT_OLWM
	if (HAVE_OLWM(AxDeskFlag))
		if (!olwm_filter_event(ev))
			return(False);
#endif

	/* Note: For HP_VUE, the Xt/Xv library directly calls-back the
	   drop handler.  Therefore, don't need a hp_vue_filter_event()
	   function. */
	return(True);
}

/***********************************************************************
function: add_to_regwin_list - Add dpy-win pair to list private list
of registered windows.
*/
static void
add_to_regwin_list(dpy, win)
register Display *dpy;
register Window  win;
{
	regwin_t *rp;

	if (!(rp = (regwin_t *) calloc(1, sizeof(regwin_t))))
		return;

	rp->dpy = dpy;
	rp->win = win;

	if (RegWinHead) {
		rp->next = RegWinHead;
		RegWinHead->prev = rp;
	}
	RegWinHead = rp;
	return;
}

/***********************************************************************
function: delete_from_regwin_list - Remove dpy-win pair from list of
registered windows.
*/
static void
delete_from_regwin_list(dpy, win)
register Display *dpy;
register Window  win;
{
	register regwin_t *rp;

	TMPLOG(("Deleting window <0x%lx> ***\n", win));
	for (rp = RegWinHead; rp; rp = rp->next) {
		if ((rp->win == win) && (rp->dpy == dpy)) {
			if (rp->prev)
				rp->prev->next = rp->next;
			if (rp->next)
				rp->next->prev = rp->prev;
			if (rp == RegWinHead)
				RegWinHead = rp->next;
			free(rp);
			rp = (regwin_t *) NULL;
			return;
		}
	}
	/* dpy-win not found in registered list. */
	return;
}

/***********************************************************************
function: is_registered_window - Return TRUE if dpy-win pair exists in
the list of registered windows; FALSE otherwise.
*/
static int
is_registered_window(dpy, win)
register Display *dpy;
register Window  win;
{
	register regwin_t *rp;

	for (rp = RegWinHead; rp; rp = rp->next)
		if ((rp->win == win) && (rp->dpy == dpy))
			return(TRUE);

	return(FALSE);
}

/***********************************************************************
function: determine_desktop_manager - Return a flag with appropriate
bits set by determining which desktop/window managers can be supported.
*/
static unsigned int
determine_desktop_manager(dpy)
Display *dpy;
{
	int have_olwm = FALSE;
	int have_ixi  = FALSE;
	unsigned int dtflags = 0;

#if SUPPORT_OLWM
	/* Test for the presence of the Drop Site Database Manager (DSDM). */
	AxXA_SUN_DRAGDROP_DSDM = XInternAtom(dpy, NXLT("_SUN_DRAGDROP_DSDM"), True);
	if (AxXA_SUN_DRAGDROP_DSDM != None)
		dtflags |= AXDTM_FLAG_OLWM; /* will have olwm */
#endif

#if SUPPORT_IXI
	dtflags |= AXDTM_FLAG_IXI; /* will have ixi X.desktop */
#endif

#if SUPPORT_HP_VUE
	if (!AxDndNoMessaging)
		dtflags |= AXDTM_FLAG_HP_VUE; /* will have hp-vue */
#endif

	return(dtflags);
}

/***********************************************************************
function: determine_file_type - stat(2) the file NAME and return one
of the DRAG_FILETYPE_XXX file types.
*/
static int
determine_file_type(name)
char *name;
{
	struct stat sbuf;
	int filetype;

	if (!name)
		return(DRAG_FILETYPE_UNKNOWN);

	if (stat(name, &sbuf) == -1)
		return(DRAG_FILETYPE_UNKNOWN);

	switch (sbuf.st_mode & S_IFMT) {
	case S_IFREG:
		filetype = DRAG_FILETYPE_REG;
		break;
	case S_IFDIR:
		filetype = DRAG_FILETYPE_DIR;
		break;
	case S_IFCHR:
		filetype = DRAG_FILETYPE_CHR;
		break;
	case S_IFBLK:
		filetype = DRAG_FILETYPE_BLK;
		break;
#if !X386_NEW
	case S_IFLNK:
		filetype = DRAG_FILETYPE_LNK;
		break;
	case S_IFSOCK:
		filetype = DRAG_FILETYPE_SOCK;
		break;
#endif
	case S_IFIFO:
		filetype = DRAG_FILETYPE_FIFO;
		break;
	default: /* huh? */
		filetype = DRAG_FILETYPE_UNKNOWN;
		break;
	}
	return(filetype);
}

static void
poke_list_box(w, hostname, filename)
widget_t *w;
char *hostname;
char *filename;
{
	data_ptr aP;
#define POKE_LIST_DROP 10030
	
	aP = AxMakeArray(2);
	aP = AxAddStrToArray(aP, 0, filename);
	aP = AxAddStrToArray(aP, 1, hostname);
	XmPoker(POKE_LIST_DROP, aP, w, FALSE);
	AxFreeData(aP);
	return;
}

static int
send_drop_message_to_object(win, dp)
Window win;
drop_item_t *dp;
{
	sys_event_t sys_event;
	drop_item_t drop_item;
	int wid, pid;
	AXWidget w = (AXWidget) NULL;

	if ((win == (Window) None) || !dp)
		return(-1);

	if (!(wid = ATMWindowFromXWindow(win))) {
		TMPLOG(("ATMWindowFromXWindow FAILS for win <0x%lx>\n", win));
		if ((wid = ATMWindowFromShellWindow(win)) == -1) {
			TMPLOG(("ATMWindowFromShellWindow also FAILED!!\n"));
			/* must be list box */
			w = (AXWidget) WidgetOfXWindow(win);
			if (w != (AXWidget) NULL) {
				if (w->type == W_LIST) {
					TMPLOG(("Sending poke to list box\n"));
					poke_list_box(w, drop_item.host, drop_item.name);
					return(0);
				} else {
					TMPLOG(("object not a list box\n"));
					return(-1); /* object not a list box */
				}
			} else {
				TMPLOG(("object not an app or a list box\n"));
				return(-1); /* object not an app or a list box */
			}
		}
	}
/* object is an app window */
	if ((pid = PidFromATMWindow(wid)) == -1) {
		TMPLOG(("PidFromATMWindow FAILS for wid %d\n", wid));
		return(-1);
	}
	memset((char *) &sys_event, 0, sizeof(sys_event_t));
	sys_event.type = DROP_EVENT;
	sys_event.ev.drop.filetype = dp->type;
	strcpy(sys_event.ev.drop.name, dp->name);
	strcpy(sys_event.ev.drop.machine, dp->host);
	AxSendMes(pid, (char *) &sys_event, sizeof(sys_event_t), APSYS);
	TMPLOG(("Sent message to pid %d!!\n", pid));
	return(0);
}

/************************************************************************
  IXI specific code. */
#if SUPPORT_IXI

static void
ixi_initialize(dpy)
Display *dpy;
{
	if (AxXA_FILE_NAME == None)
		AxXA_FILE_NAME = XInternAtom(dpy, NXLT("FILE_NAME"), False);
	/* FIXME, check return */
	DrgDrpSetupFilter(dpy, &ixi_dd_context, ixi_drop_callback);
	return;
}

static void
ixi_register_window(dpy, win, target)
Display *dpy;
Window  win;
Window  target;
{
	DrgDrpRegisterWindow(dpy, win, target, &AxXA_FILE_NAME, 1);
	add_to_regwin_list(dpy, win);
	return;
}

static void
ixi_unregister_window(dpy, win)
Display *dpy;
Window  win;
{
	DrgDrpUnregisterWindow(dpy, win);
	return;
}

/* Return `False' if event was consumed; `True' otherwise. */
static int
ixi_filter_event(ev)
XEvent *ev;
{
	return(DrgDrpFilterEvent(ev, &ixi_dd_context));
}

static void
ixi_drop_callback(closure, dpy, win, time, x, y, styles, n_styles,
				  data, n_formats)
void *closure;
Display *dpy;
Window  win;
Time time;
unsigned int x, y;
DrgDrpStyle styles[];
unsigned int n_styles;
DrgDrpInfo data[];
unsigned int n_formats;
{
	register unsigned int ix;
	register char *fname = (char *) NULL;
	int wid, pid;
	sys_event_t sys_event;
	drop_item_t drop_item;
	AXWidget w;

	memset((char *) &drop_item, 0, sizeof(drop_item_t));
	for (ix = 0; ix < n_formats; ix++) {
		if (data[ix].format == AxXA_FILE_NAME) {
			fname = (char *) data[ix].data;
			strcpy(drop_item.name, fname);
			if (access(fname, 0) == -1)
				return;
			drop_item.type = determine_file_type(fname);
		}
	}

	if (!fname)
		return; /* did not get a filename */
/* Send drop message to appropriate task. */
	if ((w = (AXWidget)WidgetOfXWindow(win)) != (AXWidget)NULL) {
		/* Note: When list box implementation code is changed to
		   follow the ``elf event model'', make callback right here. */
		if ((w->type == W_LIST) && (w->callback)) {
			SendListDropMessage(w, fname); /* in xmlist.c */
			return;
		}
		/* else, drop through */
	}

/* Not a list box, must be an app window. */
	if (!(wid = ATMWindowFromXWindow(win)))
		return;
	if ((pid = PidFromATMWindow(wid)) == -1)
		return;
	memset((char *) &sys_event, 0, sizeof(sys_event_t));
	sys_event.type = DROP_EVENT;
	sys_event.ev.drop.filetype = drop_item.type;
	strcpy(sys_event.ev.drop.name, drop_item.name);
	if (gethostname(drop_item.host, MAXHOSTNAMELEN) == -1)
		return;
	strcpy(sys_event.ev.drop.machine, drop_item.host);
	AxSendMes(pid, (char *) &sys_event, sizeof(sys_event_t), APSYS);
	return;
}

#endif /* SUPPORT_IXI */

#if SUPPORT_OLWM
/************************************************************************
  OLWM specific code.
  See Also:
				 Drag-and-Drop Protocol Specification
						Open Windows Version 3
						   Stuart W. Marks
					  Sun Microsystems, Inc. */
static void
olwm_initialize(dpy)
Display *dpy;
{
	AxXA_SUN_DRAGDROP_INTEREST = XInternAtom(dpy,
											 NXLT("_SUN_DRAGDROP_INTEREST"), True);
	if (AxXA_SUN_DRAGDROP_INTEREST == None)
		goto errout;

	AxXA_SUN_DRAGDROP_TRIGGER = XInternAtom(dpy,
											NXLT("_SUN_DRAGDROP_TRIGGER"), True);
	if (AxXA_SUN_DRAGDROP_TRIGGER == None)
		goto errout;

	AxXA_DRAGDROP_PROP = XInternAtom(dpy, NXLT("_AX_DRAGDROP_PROP"), False);
	if (AxXA_DRAGDROP_PROP == None)
		goto errout;

	AxXA_SUN_DRAGDROP_DONE = XInternAtom(dpy, NXLT("_SUN_DRAGDROP_DONE"), True);
	if (AxXA_SUN_DRAGDROP_DONE == None)
		goto errout;

	AxXA_SUN_DRAGDROP_ACK = XInternAtom(Dpy, NXLT("_SUN_DRAGDROP_ACK"), False);
	if (AxXA_SUN_DRAGDROP_ACK == None)
		goto errout;

	if (AxXA_FILE_NAME == None)
		AxXA_FILE_NAME = XInternAtom(dpy, NXLT("FILE_NAME"), False);
	if (AxXA_FILE_NAME == None)
		goto errout;
	TMPLOG(("olwm_initialize: DONE\n"));
	return;

errout:
	/* An error has occured creating one of the atoms, turn off the
	   OLWM bit in `AxDeskFlag'. */
	AxDeskFlag &= (~(AXDTM_FLAG_OLWM));
	return;
}

static void
olwm_register_window(dpy, win, target)
register Display *dpy;
register Window  win;
register Window  target;
{
	register int num_elem = 0;
	int interest_contents[20];

	/* interest contents */
	interest_contents[num_elem++] = 0; /* version number */
	interest_contents[num_elem++] = 1; /* site count */

	/* drop site description */
	interest_contents[num_elem++] = win; /* XID of window to receive events */
	interest_contents[num_elem++] = target; /* client-specified site ID
												(uninterpreted) */
	interest_contents[num_elem++] = AXOL_SITE_FLAG_DEFAULT;
	interest_contents[num_elem++] = 0; /* passing a rect-list */
	interest_contents[num_elem++] = 1; /* number of rects */
	interest_contents[num_elem++] = 0; /* x */
	interest_contents[num_elem++] = 0; /* y */
	interest_contents[num_elem++] = (int) DpyWid; /* width */
	interest_contents[num_elem++] = (int) DpyHt;  /* height */
	XChangeProperty(dpy, win, AxXA_SUN_DRAGDROP_INTEREST,
					AxXA_SUN_DRAGDROP_INTEREST, 32, PropModeReplace,
					(unsigned char *) interest_contents, num_elem);
	add_to_regwin_list(dpy, win);
	TMPLOG((NXLT("\tolwm-register-window: registering window <0x%lx>\n"),
			win));
	return;
}

/* Return `False' if event was consumed; `True' otherwise. */
static int
olwm_filter_event(ev)
XEvent *ev;
{
	static Atom selection = None;
	static Window win;
	static Window target;
	static unsigned int x;
	static unsigned int y;
	static int flags;
	static Time time;
	int err, actual_format;
	unsigned long nitems, bytes_after;
	unsigned long *data = 0;
	long long_size = 10000;
	Atom actual_type;
	sys_event_t sys_event;
	drop_item_t drop_item;
	int wid, pid;
	widget_t *w = (widget_t *) NULL;
	
	if (!ev)
		return(True);

	if (ev->xany.type == ClientMessage) {
		TMPLOG((NXLT("Received ClientMessage message_type = <%s>\n"),
				XGetAtomName(Dpy, ev->xclient.message_type)));
		if (ev->xclient.message_type != AxXA_SUN_DRAGDROP_TRIGGER)
			return(True); /* not a dragdrop message */
	} else if (ev->xany.type == SelectionNotify) {
		if (ev->xselection.property != AxXA_DRAGDROP_PROP)
			return(True); /* not a dragdrop related message */
		TMPLOG((NXLT("Received SelectionNotify property = <%s>\n"),
				XGetAtomName(Dpy, ev->xselection.property)));
	} else
		return(True); /* unexpected event type */

	if (ev->xany.type == ClientMessage) {
		win = ev->xclient.window;
		selection = (Atom) ev->xclient.data.l[0]; /* transient seln */
		time      = (Time) ev->xclient.data.l[1];
		x = (unsigned int) (ev->xclient.data.l[2] & 0xFFFF0000);
		y = (unsigned int) ((ev->xclient.data.l[2] & 0x0000FFFF) << 16);
		target = (Window) ev->xclient.data.l[3];
		flags  = (int) ev->xclient.data.l[4];
		
		TMPLOG((NXLT("ClientMessage receives WINDOW = <0x%lx>, TARGET = <0x%lx>\n"),
				win, target));
		/* don't handle (flags & AXOL_TRG_DELETE)
		   or (flags & AXOL_TRG_FORWARD) */

		if (flags & AXOL_TRG_TRANSIENT)
			XConvertSelection(Dpy, selection, AxXA_FILE_NAME,
							  AxXA_DRAGDROP_PROP, win, time);

		if (flags & AXOL_TRG_ACKNOWLEDGE)
			XConvertSelection(Dpy, selection, AxXA_SUN_DRAGDROP_ACK,
							  AxXA_DRAGDROP_PROP, win, time);

	} else if (ev->xany.type == SelectionNotify) {
		if (selection == None) {
			/* not expecting any dragdrop-related selection notifies */
			return(True);
		}
		if (ev->xselection.property == None) {
			/* botch: ConvertSelection() failed.  Doesn't happen;
			   calling code makes sure property is valid. */
			goto reset_and_out;
		}
		if (ev->xselection.property == AxXA_SUN_DRAGDROP_ACK) {
			/* Just an ack-ack.  Nothing more to do. */
			return(False); /* event consumed */
		}
		err = XGetWindowProperty(Dpy, win, ev->xselection.property, 0,
								 long_size, True, AnyPropertyType,
								 &actual_type, &actual_format,
								 &nitems, &bytes_after, (unsigned char **)&data);
		if (err != Success)
			goto reset_and_out; /* botch */

		if (actual_type == AxXA_SUN_DRAGDROP_ACK) {
			if (data)
				XFree(data);
			return(False); /* event consumed */
		}
		if ((actual_type != XA_STRING) || !data)
			goto reset_and_out; /* botch */

		memset((char *) &drop_item, 0, sizeof(drop_item_t));
		memset((char *) &sys_event, 0, sizeof(sys_event_t));

		strcpy(drop_item.name, (char *)data);
		if (access(data, 0) == -1) /* data holds a file name */
			goto reset_and_out;

		drop_item.type = determine_file_type(data);
		if (gethostname(drop_item.host, MAXHOSTNAMELEN) == -1)
			goto reset_and_out;

		if (!is_registered_window(Dpy, win)) {
			TMPLOG(("%d NOT REGISTERED WINDOW\n", win));
			goto reset_and_out;
		}
		send_drop_message_to_object(win, (drop_item_t *) &drop_item);
		goto reset_and_out;
	}
	if (data)
		XFree(data);
	return(False);

reset_and_out:
	selection = None; /* done with this transient seln */
	if (data)
		XFree(data);
	return(False); /* event consumed */
}

static void
olwm_unregister_window(dpy, win)
Display *dpy;
Window  win;
{
	return;
}
#endif /* SUPPORT_OLWM */

#if SUPPORT_HP_VUE

/************************************************************************
  HP-VUE specific code. */

static void
hp_vue_initialize(dpy)
Display *dpy;
{
	extern void AxRealXtToolkitInitialize();

	AxRealXtToolkitInitialize(); /* in axt.c */
	XSync(dpy, False);
#if HP800V3
	if (XvInit(dpy, NXLT("AxHPVue")) == False)
		AxDeskFlag &= (~(AXDTM_FLAG_HP_VUE)); /* failed */
#else
	XvInit(dpy, NXLT("AxHPVue"));
#endif
	return;
}

static hp_vue_droplist_t *
hp_vue_parse_droplist(types_str, objects_str)
char *types_str;
char *objects_str;
{
	hp_vue_droplist_t *head = (hp_vue_droplist_t *) NULL;
	hp_vue_droplist_t *dp   = (hp_vue_droplist_t *) NULL;
	register char *tp = types_str;
	register char *op = objects_str;
	register char *p;
	int nobj;
	char machine[MAXHOSTNAMELEN + 1];
	char typestr[STRLONG + 1];
	char name[STRPATH + 1];

	if (!tp || !op)
		return((hp_vue_droplist_t *) NULL); /* invalid args */

	/* count the number of ',' in `types_str' to get number of
	   objects. */
	for (tp = types_str, nobj = 1; *tp; tp++)
		if (*tp == ',')
			nobj++;

	/* get the machine name */
	for (p = machine, op = objects_str; op; op++) {
		if ((*op == ' ') && (*(op + sizeof(char)) == '-'))
			break;
		*p++ = *op;
	}
	*p = '\0';

    for (; *op; op++)
        if (*op == '/') /* leading '/' guranteed */
            break;

	for (tp = types_str; nobj; nobj--) {
		if (!(dp = (hp_vue_droplist_t *) calloc(1,
										 sizeof(hp_vue_droplist_t))))
			return((hp_vue_droplist_t *) NULL);
		if (!head) {
			head = dp;
		} else {
			dp->next = head;
			head = dp;
		}

		strcpy(dp->machine, machine);

		/* gather type string */
		for (p = typestr; *tp; tp++) {
			if (*tp == ',') {
				tp++; /* for next pass */
				break;
			}
			*p++ = *tp;
		}
		*p = '\0';
		/* note: ignore `typestr'. Use `determine_file_type()' to find
		   out file type. */

		/* gather file name string */
		for (p = name; *op; op++) {
			if (*op == ' ') { /* no spaces in file names!! */
				op++; /* for next pass */
				break;
			}
			*p++ = *op;
		}
		*p = '\0';

		strcpy(dp->name, name);
		dp->type = determine_file_type(dp->name);
	}
	
	return(head);
}

static void
hp_vue_drop_callback(dpy, win, x, y, root, root_x, root_y, mods, types,
					objects)
Display *dpy;
Window  win;
int x;
int y;
Window root;
int root_x;
int root_y;
unsigned int mods;
String types;
String objects;
{
	int wid;
	int pid;
	sys_event_t sys_event;
	register hp_vue_droplist_t *droplist;
	register hp_vue_droplist_t *dp;
	widget_t *w = (widget_t *) NULL;
	if (!(droplist = hp_vue_parse_droplist(types, objects)))
		return;

	for (dp = droplist; dp; dp = dp->next) {
		if ((w = WidgetOfXWindow(win)) != (widget_t *) NULL) {
			/* Note: When list box implementation code is changed to
			   follow the ``elf event model'', make callback right here. */
			if ((w->type == W_LIST) && (w->callback)) {
				SendListDropMessage(w, dp->name); /* in xmlist.c */
				continue;
			} /* else, drop through */
		} /* else, drop through */
		
		/* Not a list box, must be an app window. */
		if (!(wid = ATMWindowFromXWindow(win)))
			return;

		if ((pid = PidFromATMWindow(wid)) == -1)
			return;

		memset(&sys_event, 0, sizeof(sys_event_t));
		sys_event.type = DROP_EVENT;
		sys_event.ev.drop.filetype = dp->type;
		strcpy(sys_event.ev.drop.name, dp->name);
		strcpy(sys_event.ev.drop.machine, dp->machine);
		AxSendMes(pid, (char *) &sys_event, sizeof(sys_event_t), APSYS);
	}
	hp_vue_destroy_droplist(droplist);
	return;
}

static void
hp_vue_destroy_droplist(head)
register hp_vue_droplist_t *head;
{
	register hp_vue_droplist_t *p;
	register hp_vue_droplist_t *next;

	if (!head)
		return;

	p = head;
	while (p) {
		next = p->next;
		free(p);
		p = next;
	}
	return;
}

/*ARGSUSED*/
static void
hp_vue_register_window(dpy, win, target)
Display *dpy;
Window win;
Window target; /* unused */
{
	char secondary_hotspot[STRLONG + 1];
	XSync(dpy, False);
	sprintf(secondary_hotspot, NXLT("=%dx%d+0+0-Device "), DpyWid, DpyHt);
	XvRegisterDragWindow(dpy, win, 2, BGpixel, FGpixel, secondary_hotspot,
						 NULL, NULL);
	XvRegisterDropCallback(hp_vue_drop_callback);
	add_to_regwin_list(dpy, win);
	return;
}

static void
hp_vue_unregister_window(dpy, win)
Display *dpy;
Window win;
{
	XvRegisterInvalidDropWindow(dpy, win, DpyWid, DpyHt);
	return;
}

#endif /* SUPPORT_HP_VUE */

#if 0
/* temporary */
static int
logit(va_alist) va_dcl
{
	va_list argp;
	char *fmt;
	static FILE *logfp = (FILE *) NULL;

	if (logfp == (FILE *) NULL)
		logfp = fopen(NXLT("/tmp/log-dnd"), "w");

	if (logfp == (FILE *) NULL)
		return(0);
	va_start(argp);
	fmt = va_arg(argp, char *); /* LOK pa */
	vfprintf(logfp, fmt, argp);
	va_end(argp);
	fflush(logfp);
	return(0);
}
#endif


/* GNU Emacs settings - This (including the "\n^L") MUST be the last thing
 *                      in this file!
 * Local variables:
 * mode: c
 * tab-width: 4
 * end:
 */
