/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axfont.c
   Applix ASC version: 31
   Date: Mon Nov 16 13:53:40 1998
*/
#include <axtext.h>
#include <axfont.h>
#include <wchars.h>

extern char *constructForDisplay(), *axGetScreenMetrics();
extern char *axGetFontRecordName(), *axGetFontRecord();
extern char *axGetTargetMetrics();
extern int isBaseFontSet();
extern XCharStruct *getCharStruct();
extern mbchar *OrientFace[];

int getScreenFontMetrics(),
	copyScreenFontMetrics(),
	freeScreenFontMetrics(),
	getTargetFontMetrics(),
	copyTargetFontMetrics(),
	freeTargetFontMetrics(),
	initTargetCharMetrics();

XCharStruct *targetCharStruct();

#ifndef IS_ORIENT_CHAR
#define IS_ORIENT_CHAR( p ) ((unsigned)p > (unsigned)256)
#endif

#define SPACECHAR (' ')
#define HYPHEN ('-')
#define NONBREAKSPACE   (SPACECHAR + 0x80)  /* font convention */   


int
getScreenFontMetrics(name, dp, sp, fontAtts)
char *name;
char **dp;       /* out: to be returned as screenFontMetrics type */
int *sp;         /* out: item size */
char *fontAtts;  /* in: ptr to textAttrs */
{
#if 0	 /* a stub */
	screenFontMetrics *sm;
	char *errString;
	tx_char junk = (tx_char) 'a';
	
	sm = (screenFontMetrics *) malloc (sizeof(screenFontMetrics));
	memset((char *) sm, 0, sizeof(screenFontMetrics));
	sm->scaled = (axtext *) malloc (sizeof(axtext));
	memset((char *) sm->scaled, 0, sizeof(axtext));

	setDefaultTextAttrs(sm->scaled, (textAttrs *) fontAtts,
						TXAllAtts);        /* set atts */
	insChars(sm->scaled, &junk, 0, 1);     /* add dummy character */
	     /* get display resource data */
	if (errString = constructForDisplay(sm->scaled)) {
		free((char *) errString);
		daxtext(sm->scaled);
		free((char *) sm->scaled);
		free((char *) sm);
		return FALSE;
	}
	if (sm->scaled->atts[0].qDraw) {
		    /* we can use x letterform metrics directly
			 * note that it would be best to have the target
			 * fontTabEnt reference here.. has to do with
			 * AXMS alternate-screenfont-bashing in the constructor
			 * getScreenMetrics
			 */
		if (errString =
			axGetScreenMetrics(sm->scaled->atts[0].screenFontName,
							   (fontTabEnt *) NULL,
							   &sm->direct)) {
			freeErrString(errString);
			return FALSE;
		}
		daxtext(sm->scaled);       /* we can pitch the axtext stuff */
		free((char *) sm->scaled);
		sm->scaled = (axtext *) NULL;
		sm->kickoutCount = kickOutCounterValue(SCREEN_METRICS);
	}
	*sp = 0;  /* a lie!  but for now, no kickouts on
				 SCREENFONT_METRICS */
	*dp = (char *) sm;
	return TRUE;
#else	/* to prevent intelligent compilers from complaining */
	return(FALSE);
#endif   /* if 0 (stub) */	
}

int
copyScreenFontMetrics(name, onto, sizep, from)
char *name;
char **onto;
int *sizep;
char *from;
{
	return FALSE; /* this copy op not defined yet */
}

int
freeScreenFontMetrics(dp)
char **dp;
{
#if 0	 /* a stub */
	screenFontMetrics *sm;

	if (!dp)
		return FALSE;
	sm = (screenFontMetrics *) *dp;
	if (sm->scaled) {
		daxtext(sm->scaled);
		free((char *) sm->scaled);
	}
	    /* member "direct" is a SCREEN_METRICS ref; don't free */
	
	free((char *) sm);
	return TRUE;
#else	/* to prevent intelligent compilers from complaining */
	return(TRUE);
#endif	
}

int
getTargetFontMetrics(name, dp, sp, fontAtts)
char *name;
char **dp;       /* out: to be returned as targetFontMetrics type */
int *sp;         /* out: item size */
char *fontAtts;  /* in: ptr to textAttrs */
{
	int tmp;
	textAttrs *atts;
	targetFontAttrs *tfattrsP;
	targetFontMetrics *target_fm;
	fontMetrics *metrics;
	fontTabEnt *ftab;
	char *errString, *recName;

	tfattrsP = (targetFontAttrs *)fontAtts;
	atts = (textAttrs *) &(tfattrsP->attrs);

	if (errString = axGetFontRecordName(atts->fontFamily,
										atts->weight,
										atts->slant,
										&recName)) {
		freeErrString(errString);
		return FALSE;
	}
	if (errString = axGetFontRecord(recName, &ftab)) {
		freeErrString(errString);
		return FALSE;
	}
	if (errString = axGetTargetMetrics(ftab->targetMetricsNames[0],
									   ftab, &metrics)) {
		freeErrString(errString);
		return FALSE;
	}

	target_fm = (targetFontMetrics *) malloc (sizeof(targetFontMetrics));
	memset((char *) target_fm, 0, sizeof(targetFontMetrics));

	target_fm->resolution = tfattrsP->resolution;

	target_fm->metrics = metrics;
	target_fm->kickoutCount = kickOutCounterValue(TARGET_METRICS);

	    /* fill out whole-font data on target_fm->local now, at res dpi
		 */
	target_fm->local = (fontMetrics *) malloc(sizeof(fontMetrics));
	memset((char *) target_fm->local, 0, sizeof(fontMetrics));
	target_fm->local->xfont = (XFontStruct *) malloc (sizeof(XFontStruct));
	*target_fm->local->xfont = *metrics->xfont;
	target_fm->local->xfont->ext_data = (XExtData *) NULL;
	target_fm->local->xfont->properties = (XFontProp *) NULL;
	        /* alloc blank per_char array, same size as on metrics */
	tmp = target_fm->local->xfont->max_byte1 -
		target_fm->local->xfont->min_byte1 + 1;
	tmp *= target_fm->local->xfont->max_char_or_byte2 -
		target_fm->local->xfont->min_char_or_byte2 + 1;
	tmp *= sizeof(XCharStruct);
	if (tmp <= 0) {
		free((char *) target_fm);
		return FALSE;
	}
	target_fm->local->xfont->per_char = (XCharStruct *) malloc(tmp);
	memset((char *) target_fm->local->xfont->per_char, 0, tmp);
	
	        /* convert whole-font numbers to resdpi
			 */
	target_fm->pointsize = (float) atts->pointsize / 10000.0;
	target_fm->local->xfont->ascent = mToPix(target_fm->local->xfont->ascent,
											 target_fm->pointsize,
											 target_fm->resolution);
	target_fm->local->xfont->descent = mToPix(target_fm->local->xfont->descent,
											  target_fm->pointsize,
											  target_fm->resolution);

	target_fm->local->xfont->min_bounds.lbearing =
		mToPix(target_fm->local->xfont->min_bounds.lbearing,
			   target_fm->pointsize, target_fm->resolution);
	target_fm->local->xfont->min_bounds.rbearing =
		mToPix(target_fm->local->xfont->min_bounds.rbearing,
			   target_fm->pointsize, target_fm->resolution);
	target_fm->local->xfont->min_bounds.width =
		mToPix(target_fm->local->xfont->min_bounds.width,
			   target_fm->pointsize, target_fm->resolution);
	target_fm->local->xfont->min_bounds.ascent =
		mToPix(target_fm->local->xfont->min_bounds.ascent,
			   target_fm->pointsize, target_fm->resolution);
	target_fm->local->xfont->min_bounds.descent =
		mToPix(target_fm->local->xfont->min_bounds.descent,
			   target_fm->pointsize, target_fm->resolution);

	target_fm->local->xfont->max_bounds.lbearing =
		mToPix(target_fm->local->xfont->max_bounds.lbearing,
			   target_fm->pointsize, target_fm->resolution);
	target_fm->local->xfont->max_bounds.rbearing =
		mToPix(target_fm->local->xfont->max_bounds.rbearing,
			   target_fm->pointsize, target_fm->resolution);
	target_fm->local->xfont->max_bounds.width =
		mToPix(target_fm->local->xfont->max_bounds.width,
			   target_fm->pointsize, target_fm->resolution);
	target_fm->local->xfont->max_bounds.ascent =
		mToPix(target_fm->local->xfont->max_bounds.ascent,
			   target_fm->pointsize, target_fm->resolution);
	target_fm->local->xfont->max_bounds.descent =
		mToPix(target_fm->local->xfont->max_bounds.descent,
			   target_fm->pointsize, target_fm->resolution);

	target_fm->local->underlinePosition =
		mToPix(target_fm->metrics->underlinePosition,
			   target_fm->pointsize, target_fm->resolution);
	target_fm->local->underlineThickness =
		mToPix(target_fm->metrics->underlineThickness,
			   target_fm->pointsize, target_fm->resolution);
	target_fm->local->xHeight = 
		mToPix(target_fm->metrics->xHeight,
			   target_fm->pointsize, target_fm->resolution);

	target_fm->getCharMetrics = initTargetCharMetrics;

	*sp = 0;  /* a lie!  but for now, no kickouts on
				 TARGETFONT_METRICS */
	*dp = (char *) target_fm;
	return TRUE;
}


/* : initTargetCharMetrics - read the scalable character metrics
   onto the targetFontMetrics' local copy, at resolution res dpi */
int
initTargetCharMetrics(ch, fm)
tx_char ch;    /* character code */
targetFontMetrics *fm;
{
	register XCharStruct *sc;
	register XCharStruct *dest;

	sc = getCharStruct((int)ch, fm->metrics->xfont);
	dest = getCharStruct((int)ch, fm->local->xfont);
	dest->lbearing = mToPix(sc->lbearing, fm->pointsize, fm->resolution);
	dest->rbearing = mToPix(sc->rbearing, fm->pointsize, fm->resolution);
	dest->width = mToPix(sc->width, fm->pointsize, fm->resolution);
	dest->ascent = mToPix(sc->ascent, fm->pointsize, fm->resolution);
	dest->descent = mToPix(sc->descent, fm->pointsize, fm->resolution);
	return(TRUE);
}

int
copyTargetFontMetrics(name, onto, sizep, from)
char *name;
char **onto;
int *sizep;
char *from;
{
	return FALSE; /* this copy op not defined yet */
}

int
freeTargetFontMetrics(dp)
char **dp;
{
	targetFontMetrics *tm;

	if (!dp)
		return FALSE;
	tm = (targetFontMetrics *) *dp;

	if (tm->local) {
		freeTargetMetrics((char **) &tm->local);
	}
	    /* tm member 'metrics' is TARGET_METRICS ref; don't free */

	free((char *) tm);
	return TRUE;
}





#if BUILD_ORIENT        /*  {  */
mbchar *sansserif_orient_face();
mbchar *serif_orient_face();
mbchar *serif_narrow_kana_face();
mbchar *sansserif_narrow_kana_face();

WCharType(wch)			/* 22 Apr 98 rjd Did Jon omit this on purpose? */
wchar   wch;
{
    int char_type;

    if (IS_WIDE_CHAR(wch))
        char_type = 1;  /* WIDE_FONT */
    else if (IS_NARROW_KANA(wch))
        char_type = 2;  /* NARROW_KANA_FONT */
    else
        char_type = 0;  /* NARROW_FONT */

    return(char_type);
}

targetFontMetrics *VerifyFontMetrics(ch, fm)
int ch;
targetFontMetrics *fm;
{
	static targetFontMetrics *last_fm = NULL;
	static targetFontMetrics *last_ret_fm = NULL;
	static int last_char_type = -1;
	static int last_font_type = -1;
	static int sans_type, size, bold, italic;
	static char	family[30];
	targetFontMetrics	*new_fm;
	mbchar	*face;
	int 	char_type, font_type;

	char_type = WCharType(ch);

	if (last_fm == fm && last_char_type == char_type)
		return(last_ret_fm);

	last_char_type = char_type;

	if (last_fm == fm) {
		font_type = last_font_type;
	} else {
		getTargetFontAttr(fm, family, &size, &bold, &italic);
		font_type = IsWideFontFace(family);
#if BUILD_KANJI
		if (IsNarrowKanaFontFace(family))
			font_type = NARROW_KANA_FONT;
#endif
		sans_type = IsSansFontFace(family);
		last_font_type = font_type;
		last_fm = fm;
	}

	if (char_type == font_type) {
		last_ret_fm = fm;
		return(fm);
	}

	switch (char_type) {
		case NARROW_FONT:
			if (sans_type)
				face = (mbchar *)NXLT(axDefaultSansSerifFont);
			else
				face = (mbchar *)NXLT(axDefaultSerifFont);
			break;
		case WIDE_FONT:
			if (sans_type)
				face = sansserif_orient_face();
			else
				face = serif_orient_face();
			break;
#if BUILD_KANJI
		case NARROW_KANA_FONT:
			if (sans_type)
				face = sansserif_narrow_kana_face();
			else
				face = serif_narrow_kana_face();
			break;
#endif
	}

	/* all shortcut failed. prepare new target font metrics */
	if (axGetTargetFontMetrics(face, bold, italic, size*10000, &new_fm))
		last_ret_fm = new_fm;
	else
		last_ret_fm = fm;

	return(last_ret_fm);
}
#endif      /*  }  BUILD_ORIENT */





XCharStruct *
targetCharStruct(ch, fm)
int ch;
targetFontMetrics *fm;
{
	int row, col, rowWid, idx;
	XCharStruct *xc;
	XFontStruct *xf;
#if BUILD_KANJI
    static XCharStruct auxc;
#endif

#if BUILD_ORIENT
    fm = VerifyFontMetrics(ch, fm);
#endif

    xf = fm->local->xfont;

	if (xf->max_byte1 == 0) {
		col = (unsigned int)(ch & 0xff);
		if (col >= xf->min_char_or_byte2 &&
			col <= xf->max_char_or_byte2) {
			if (xf->per_char == (XCharStruct *) NULL) {
				xc = &xf->min_bounds;
			}
			else {
				xc = &xf->per_char[col - xf->min_char_or_byte2];
			}
		}
		else {
			xc = &xf->min_bounds;
		}
	}
	else {
		row = (unsigned int)((ch >> 8) & 0xff);
		col = (unsigned int) (ch & 0xff);
		if (row >= xf->min_byte1 && 
            row <= xf->max_byte1 &&
			col >= xf->min_char_or_byte2 &&
			col <= xf->max_char_or_byte2) {
			
			rowWid = xf->max_char_or_byte2 -
				xf->min_char_or_byte2 + 1;
			idx = ((row - xf->min_byte1) * rowWid) +
				col - xf->min_char_or_byte2;
			if (xf->per_char == NULL) {
				xc = &xf->min_bounds;
			}
			else {
				xc = &xf->per_char[idx];
			}
		}

#if BUILD_KANJI
        else if ((xf->fid == 0) && (row == 0x8e)) {
            auxc.lbearing = xf->min_bounds.lbearing;
            auxc.rbearing = xf->min_bounds.rbearing / 2;    /* Half width */
            auxc.width = xf->min_bounds.width / 2;  /* Half width */
            auxc.ascent = xf->min_bounds.ascent;
            auxc.descent = xf->min_bounds.descent;
            auxc.attributes = xf->min_bounds.attributes;
            xc = &auxc;
        } else if ((xf->fid && xf->max_byte1 > 0) && (row == 0x8e)) {
            auxc.lbearing = xf->min_bounds.lbearing;
            auxc.rbearing = xf->min_bounds.rbearing / 2;    /* Half width */
            auxc.width = xf->min_bounds.width / 2;  /* Half width */
            auxc.ascent = xf->max_bounds.ascent;
            auxc.descent = xf->max_bounds.descent;
            auxc.attributes = xf->max_bounds.attributes;
            xc = &auxc;
        }
#endif
		else {
			xc = &xf->min_bounds;
		}
	}
	if (!xc->width) {
		(*fm->getCharMetrics)(ch, fm);
	}
	return (xc);
}


#if !FOR_FILTER   /*  {  */

/***************************************************************************
function: ImgTextWidth - Return width of text in mils
(Used to be WpImgTextWidth in the WP - moved here so SS can use this too).
*/
int ImgTextWidth (face, size, bold, italic, wstr, len)
mbchar *face;   /* font family */
int size;       /* in points */
int bold;       /* boolean */
int italic;     /* boolean */
wchar *wstr;    /* the string (must be null terminated if len is -1) */
int len;        /* num chars to measure (-1 if entire string) */
{
    int ix, width;
    mbchar *wide_face = (mbchar *)NULL;
    mbchar narrow_kana_face[60];
    targetFontMetrics *fs, *wide_fs;
    targetFontMetrics *narrow_kana_fs, *use_fs;
    XCharStruct *xc;
        
    if (len == -1)
        len = Wstrlen( wstr );

    if (!axGetTargetFontMetrics(face, bold, italic,
                                size * 10000, &fs)) {
        return 0;
    }

#if BUILD_ORIENT
    wide_fs = (targetFontMetrics *)NULL;
    narrow_kana_fs = (targetFontMetrics *)NULL;
    if ( fs->metrics->xfont->max_byte1 > 0)
        {
        /* we selected a wide font */
        wide_fs = fs;
        wide_face = (mbchar *)face;
        fs = (targetFontMetrics *)NULL;
        }

    for (ix = 0, width = 0; ix < len; ix++) {
        if ( IS_WIDE_CHAR( (*wstr) ) )
            {
            if ( wide_fs == (targetFontMetrics *)NULL )
                {
                /* need to find a new wide font */
                wide_face = OrientFace[ get_default_face() ];
                if (!axGetTargetFontMetrics(wide_face, bold, italic,
                                            size * 10000, &wide_fs))
                    return 0;
                }
            use_fs = wide_fs;
            }
#if BUILD_KANJI
        else if ( IS_NARROW_KANA( (*wstr) ) )
            {
            if ( narrow_kana_fs == (targetFontMetrics *)NULL )
                {
                if ( wide_face == (mbchar *)NULL )
                   wide_face = OrientFace[ get_default_face() ];
                MBstrcpy( narrow_kana_face, NXLT("Narrow") );
                MBstrcat( narrow_kana_face, wide_face );
                if (!axGetTargetFontMetrics(narrow_kana_face, bold, italic,
                                            size * 10000, &narrow_kana_fs))
                    return 0;
                }
            use_fs = narrow_kana_fs;
            }
#endif
        else
            {
            if ( fs == (targetFontMetrics *)NULL )
                {
                if (!axGetTargetFontMetrics(NXLT(axDefaultSerifFont), bold, italic,
                                            size * 10000, &fs))
                    return 0;
                }
            use_fs = fs;
            }

        xc = targetCharStruct((int) *wstr, use_fs);
        width += xc->width;
        wstr++;
    }
#else
    for (ix = 0, width = 0; ix < len; ix++) {
        xc = targetCharStruct((int) *wstr, fs);
        width += xc->width;
        wstr++;
    }
#endif
    return(width);
}


/*************************************************************************
function: SSImgTxtFitWidth - A wordwrap modified version.
*   Note that a line break is forced when a non-breaking space is detected.
*/
int SSImgTxtFitWidth (wtext, len, width, face, size, bold,
                      italic, lastWord, wordLen)
register wchar *wtext;  /* not null terminated */
register int len;           /* number of chars in text */
register int width; /* available width in mils */
mbchar *face;   /* font family */
int size;       /* in points */
int bold;       /* boolean */
int italic;     /* boolean */
int *lastWord;      /* out: number of chars to end of last word that fits */
int *wordLen;       /* out: length of string to end of last word that fits */
{
    int ix;
    int total;
    int lwx, lwlen, spaced;
    targetFontMetrics *fs, *wide_fs, *use_fs, *narrow_kana_fs;
    XCharStruct *xc;
    register wchar wch, next_wch;
    register int is_wide_ch;
    mbchar *wide_face, narrow_kana_face[60];

    *lastWord = *wordLen = 0;
    wide_face = (mbchar *)NULL;
    wide_fs = (targetFontMetrics *)NULL;
    narrow_kana_fs = (targetFontMetrics *)NULL;

    if ( len <= 0 ) {
        return(0);
    }

    if (!axGetTargetFontMetrics(face, bold, italic, size * 10000, &fs)) {
        return (0);
    }
    if ( fs->metrics->xfont->max_byte1 > 0 )
        {
        /* we selected a wide font */
        wide_fs = fs;
        wide_face = (mbchar *)face;
        fs = (targetFontMetrics *)NULL;
        }
    else
        wide_fs = (targetFontMetrics *)NULL;

    /* trashy loop below!, needs comments !!1 */
    lwlen = 0;
    lwx = 0;
    spaced = FALSE;
    total = 0;
    for( ix = 0; ix < len; ix++) {
        wch = *wtext++;
#if BUILD_ORIENT
        is_wide_ch = IS_WIDE_CHAR(wch);
#if BUILD_KANJI
       /* Ashisuto changed for JAPANESE build, considering that the
        * wide chars are everywhere breakable.(Since at cells in SpreadSheets,
        * generally there is no sufficient room to arrange wraped strings as
        * pretty as in Words. The "no_line_start/no_line_end" rule is not
        * important here)  95.09.14 */
       if (wch == ' ' || is_wide_ch || IS_NARROW_KANA(wch))
#else
       next_wch = *wtext;
       if ( (wch == ' ') ||
           (is_wide_ch && IsOKtoBreakOriental( wch, next_wch )) )
#endif
#else
        if ( wch == ' ')
#endif
            {
            if (!spaced) 
                {
                lwlen = total; /* length up to end of last word */
                lwx = ix;   /* number of chars to end of last word */
                spaced = TRUE;
                }
            }
        else
            spaced = FALSE;

#if BUILD_ORIENT
        if ( is_wide_ch )
            {
            if ( wide_fs == (targetFontMetrics *)NULL )
                {
                if (!axGetTargetFontMetrics(OrientFace[get_default_face()],
                                            bold,
                                            italic, 
                                            size * 10000,
                                            &wide_fs))
                    return (0);
                }
            use_fs = wide_fs;
            }
#if BUILD_KANJI
        else if ( IS_NARROW_KANA( wch ) )
            {
            if ( narrow_kana_fs == (targetFontMetrics *)NULL )
                {
                if ( wide_face == (mbchar *)NULL )
                   wide_face = OrientFace[ get_default_face() ];
                MBstrcpy( narrow_kana_face, NXLT("Narrow") );
                MBstrcat( narrow_kana_face, wide_face );
                if (!axGetTargetFontMetrics(narrow_kana_face, bold, italic,
                                            size * 10000, &narrow_kana_fs))
                    return 0;
                }
            use_fs = narrow_kana_fs;
            }
#endif
        else
            {
            if ( fs == (targetFontMetrics *)NULL )
                {
                if (!axGetTargetFontMetrics(NXLT(axDefaultSerifFont), bold,
                                            italic,
                                            size * 10000, &fs))
                    return(0);
                }
            use_fs = fs;
            }
#else
        use_fs = fs;
#endif
        xc = targetCharStruct((int) wch, use_fs);
        total += xc->width;

        if ((width <= total) || (wch == NONBREAKSPACE)) {
            total -= xc->width;  /* report total that is within limit */
            if(wch == NONBREAKSPACE)
                lwx = 0;    /* force no word boundary */
            break;
        }

        /* hyphen ends a word, and is included in the word it ends.
        End of string ends a word. */

        if ( (wch == HYPHEN) || (ix == len-1) ) { 
            lwlen = total;
            lwx = ix+1;
        }
    }

    if (lwx == 0) {
        /* no word boundary found */
        *lastWord = ix;
        *wordLen = total;
    }
    else {
        *lastWord = lwx;
        *wordLen = lwlen;
    }
    return(ix);
}

#endif   /* !FOR_FILTER  } */

/***************************************************************************
function: ImgTextHeight - Return height and ascent in mils
(Used to be WpImgTextHeight in the WP - moved here so SS can use this too).
*/
int
ImgTextHeight (face, size, bold, italic, ascentP)
mbchar *face;   /* font family */
int size;       /* in points */
int bold;       /* boolean */
int italic;     /* boolean */
int *ascentP;   /* RETURN */
{
    axtext txt;
    textAttrs atts;
    unsigned long flags;
    int descent, tmp;

    memset((char *)&txt, 0, sizeof(txt));
    flags = TXfontFamily|TXweight|TXslant|TXpointsize;
    atts.fontFamily = (char *)face;
    atts.weight     = bold;
    atts.slant      = italic;
    atts.pointsize  = size * 10000;
    setDefaultTextAttrs(&txt, &atts, flags);
    tmp = getTextHeight(&atts, ascentP, &descent);
    daxtext(&txt);
    return(*ascentP + descent);
}

/***************************************************************************
function: scaledCharWid - Get text character's horizontal dimensions from
XFontStruct-format printer metrics data.  Return setwidth in mils.
*/
int scaledCharWid (textchar, xf, pointsize, lbear, rbear)
tx_char textchar;
XFontStruct *xf;
int pointsize;
int *lbear, *rbear;
{
    int row, col, rowWid, idx;
    XCharStruct *xc;


    row = (textchar >> 8) & 0xff;
    col = textchar & 0xff;
    if (row >= xf->min_byte1 && row <= xf->max_byte1 &&
        col >= xf->min_char_or_byte2 && col <= xf->max_char_or_byte2) {
        rowWid = xf->max_char_or_byte2 - xf->min_char_or_byte2 + 1;
        idx = ((row - xf->min_byte1) * rowWid) + col - xf->min_char_or_byte2;
        if (xf->per_char == NULL) {
            xc = &xf->min_bounds;
        }
        else {
            xc = &xf->per_char[idx];
        }
    }

    else {
        xc = &xf->min_bounds;
    }
    *lbear = mToMils(xc->lbearing, (float) pointsize);
    *rbear = mToMils(xc->rbearing, (float) pointsize);
    return mToMils(xc->width, (float) pointsize);
}




#if BUILD_ORIENT && 0
WCharType(wch)
wchar   wch;
{
    int char_type;

    if (IS_WIDE_CHAR(wch))
        char_type = 1;  /* WIDE_FONT */
    else if (IS_NARROW_KANA(wch))
        char_type = 2;  /* NARROW_KANA_FONT */
    else
        char_type = 0;  /* NARROW_FONT */

    return(char_type);
}
#endif
 

