/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axgfs.c
   Applix ASC version: 26
   Date: Wed May 12 21:10:36 1999
*/
#include <stdio.h>
#include <stdlib.h>
#include <aos.h>
#include <axtext.h>
#include <axhtab.h>
#include <axgfs.h>

#define MAX_XLFD 257

/* external function declarations
 */
extern int axhTestItem();
extern int Agethostname();
#if !FOR_FILTER
extern int AxRunProgram();
#endif
extern char *axFontmetricsPathname();
extern char *GetProfileByStr();
extern char *axgfsReadFonts();
extern void axtFormScalableName();
extern void axhGetItem(), axhGetItemI();
extern void axhInitTable(), axhDeleteTable();
extern void axGetEncodingMap();
extern XCharStruct *getCharStruct();
char *axgfsNewPLetter();
void axgfsFreePLetter();

/* external globals read in this module
 */
extern int ScreenRes, TargetRes;
extern char *DirSlash;
extern Display *Dpy;
extern GC PixGC, PixClearGC;

/* non-static functions defined here
 */
int axRealDisplayResolution();
int isGFSxlfdName();
int haveGFSServer();
void axgfsInit();
void axgfsInitCharStruct();
void axgfsDrawString();
void axgfsWDrawString();
void axgfsFreeFontStruct();
void axgfsFreeAllPLetters();
#if !FOR_FILTER
char *axgfsStartFontManager();
#endif
XFontStruct *axgfsOpenFont();
int axgfsFreeExtension();
axletter *axgfsIntGetLetter();
char *axgfsIntGetFid();
AXFSServer *AxGFSServer = NULL;

/* static functions
 */
static int axgfsErrorFunc();
static int axgfsIOErrorFunc();
static int axgfsX11R5();
static int serverHasScalables();
static char *axNewEncodingMap();
static void axFreeEncodingMap();
static void xlfdTemplate();
static void noCharBox();
static float axgfsGetMilliemFactor();

/* static globals
 */
static axfsBitmapFormat AxGFSBitmapFormat;
static axfsBitmapFormat AxGFSPrinterBitmapFormat;
static axhtab AxGFSxlfdNames;
static axhtab AxEncodingMaps;
static float AxGFSMilliemFactor;

void
axgfsInit()
{
	char **names, *err;
	char *configPath, *sName, *serverName;
	char *galliumDir, *envVal, *gName;
	int numNames, ix, itmp;

	/* init pathname to GFS config file:  <galliumdir>/fs/config
	 */
#if !FOR_FILTER	
	galliumDir = GetProfileByStr(NXLT("axGFSdir"));
#else	
	galliumDir = (char *) NULL;
#endif	
	if (galliumDir && *galliumDir) {
		gName = (char *) NULL;    /* no need to free base dir name */
	}
	else {
		/* by default, the gallium directory is called "gallium",
		 * and is located under axdata/fontmetrics
		 */
		galliumDir = axFontmetricsPathname(NXLT("gallium"));
		gName = galliumDir;
	}
	itmp = strlen(galliumDir) + strlen(NXLT("fs")) +
		strlen(NXLT("config")) + 1;
	itmp += 2 * strlen(DirSlash);
	configPath = (char *) malloc(itmp);
	strcpy(configPath, galliumDir);
	strcat(configPath, DirSlash);
	strcat(configPath, NXLT("fs"));
	strcat(configPath, DirSlash);
	strcat(configPath, NXLT("config"));

	/* place the gallium directory path in the environment as $AXGFSDIR
	 */
	envVal = (char *) malloc(strlen(galliumDir) + 10); /* plus "AXGFSDIR=" */
	strcpy(envVal, NXLT("AXGFSDIR="));
	strcat(envVal, galliumDir);
	if (putenv(envVal)) {
		free(envVal);     /* didn't work; error message will result */
	}
	if (gName) {
		free(gName);
	}
	
	/* get the font server connection address; try profile
	 * value (e.g., "tcp/nova89:7001") first before using "imbedded" name.
	 */
#if !FOR_FILTER
	serverName = GetProfileByStr(NXLT("axGFSName"));
#else	
	serverName = (char *) NULL;
#endif	
	if (serverName && *serverName) {
		/* use profile value as connection target; this means we'll
		 * try to connect to a separate font server process.  As of
		 * Sep '95 this method prevents access to charcodes > 255..
		 */
		sName = (char *) NULL;   /* no imbedded name to free */
	}
	else {
		/* connect to imbedded font server code
		 */
		itmp = strlen(NXLT("imbedded:")) + strlen(configPath) + 1;
		sName = (char *) malloc(itmp);
		strcpy(sName, NXLT("imbedded:"));
		strcat(sName, configPath);
		serverName = sName;
	}
	
#if BUILD_ORIENT
	AxGFSServer = (AXFSServer *) NULL;
#else
	/* load font records representing the fonts available
	 * via the gallium server.
	 */
	err = axgfsReadFonts(configPath, &AxGFSxlfdNames);
	if (err) {
		AxGFSServer = (AXFSServer *) NULL;
		if (strcmp(err, AXGFS_NO_CONFIG)) {
#if !FOR_FILTER			
			/* the message is other than cant-find-config-file; report it
			 */
			ElfbNewTask(NXLT("cant_open_gallium_server$"), err);
#endif			
		}
		free(err);
	}
	else {
		/* open connection to gallium server
		 */
		if (AxGFSServer == NULL)
			AxGFSServer = FSOpenServer(serverName);
		else { /* InitFontInfo() called to re-read fonts */
			FSReconfigureServer(AxGFSServer);
			FSFlush(AxGFSServer);
		}
	}
	if (sName) {
		free(sName);
	}
	
	if (AxGFSServer) {
		
		/* install non-fatal and I/0 error handlers
		 */
		FSSetErrorHandler(axgfsErrorFunc);
		FSSetIOErrorHandler(axgfsIOErrorFunc);
		
		/* Determine the bitmap format for font server requests.
		 */
		AxGFSBitmapFormat = axfsBitmapFormatImageRectMin;
		
		/* Determine the bit order of the server. Beware the possibility
		 * of bit order conversion from Font Server to Client, then again
		 * from Client to Display Server.
		 */
		if (BitmapBitOrder( Dpy ) == MSBFirst)
			AxGFSBitmapFormat |= axfsBitmapFormatBitOrderMSB;
		else
			AxGFSBitmapFormat |= axfsBitmapFormatBitOrderLSB;
		
		if (ImageByteOrder( Dpy ) == MSBFirst)
			AxGFSBitmapFormat |= axfsBitmapFormatByteOrderMSB;
		else
			AxGFSBitmapFormat |= axfsBitmapFormatByteOrderLSB;
		
		if (BitmapPad( Dpy ) == 64)
			AxGFSBitmapFormat |= axfsBitmapFormatScanlinePad64;
		else if (BitmapPad( Dpy ) == 32)
			AxGFSBitmapFormat |= axfsBitmapFormatScanlinePad32;
		else if (BitmapPad( Dpy ) == 16)
			AxGFSBitmapFormat |= axfsBitmapFormatScanlinePad16;
		else if (BitmapPad( Dpy ) == 8)
			AxGFSBitmapFormat |= axfsBitmapFormatScanlinePad8;
		
		if (BitmapUnit( Dpy ) == 64)
			AxGFSBitmapFormat |= axfsBitmapFormatScanlineUnit64;
		else if (BitmapUnit( Dpy ) == 32)
			AxGFSBitmapFormat |= axfsBitmapFormatScanlineUnit32;
		else if (BitmapUnit( Dpy ) == 16)
			AxGFSBitmapFormat |= axfsBitmapFormatScanlineUnit16;
		else if (BitmapUnit( Dpy ) == 8)
			AxGFSBitmapFormat |= axfsBitmapFormatScanlineUnit8;

		/* supply a fixed value for the printer bit order
		 */
		AxGFSPrinterBitmapFormat = axfsBitmapFormatByteOrderMSB |
									axfsBitmapFormatBitOrderMSB |
									axfsBitmapFormatImageRectMin |
									axfsBitmapFormatScanlinePad8 |
									axfsBitmapFormatScanlineUnit8 ;
	}
	
	AxGFSMilliemFactor = axgfsGetMilliemFactor(Dpy);
	
	axhInitTable(&AxEncodingMaps, 31, axNewEncodingMap, axFreeEncodingMap);
#endif
}

XFontStruct *
axgfsOpenFont(xlfd, resolution, pointsize, ftab)
char *xlfd;               /* in */
int resolution;           /* in, ==ScreenRes or ==TargetRes */
int pointsize;            /* in, meaningful only when resolution==ScreenRes */
fontTabEnt *ftab;         /* in */
{
	int numChars, itmp;
	char fullname[MAX_XLFD], *xnamep;
    unsigned char  *prop_data;
	Font tfont, other_id;
	XFontStruct *fs;
    axfsFontHeader header;
    axfsPropInfo      properties;
    axfsPropOffset   *offsets;

	if (resolution == TargetRes) {
		/* this is a target metrics fontstruct lookup; the xlfd name
		 * is fully filled out at real screen res, and AXGFS_TARGET_POINTSIZE.
		 */
		xnamep = xlfd;
	}
	else {
		/* this is a screen metrics lookup; the xlfd name is a template
		 * which must be finished using pointsize and resolution.
		 */
		axtFormScalableName(xlfd, pointsize, resolution, (axctm *)0, fullname);
		xnamep = fullname;
	}
	other_id = (Font) 0;
    tfont = FSOpenBitmapFont(AxGFSServer, 0, 0, xnamep, &other_id);
	if (other_id) {
		tfont = other_id;
	}
	if (!tfont) {
		return ((XFontStruct *) NULL);
	}
	
	if (FSQueryXInfo(AxGFSServer, tfont, &header,
					 &properties, &offsets, &prop_data) != AXFSSuccess) {
		return ((XFontStruct *) NULL);
	}
    FSFree((char *) offsets);
    FSFree((char *) prop_data);

	/* calloc font struct and init all non-zero members.
	 */
	fs = (XFontStruct *) malloc(sizeof(XFontStruct));
	memset((void *) fs, 0, sizeof(XFontStruct));
	fs->fid = tfont;
    fs->direction = header.draw_direction;
	fs->min_char_or_byte2 = header.char_range.min_char.low;
	fs->max_char_or_byte2 = header.char_range.max_char.low;
	fs->min_byte1 = header.char_range.min_char.high;
	fs->max_byte1 = header.char_range.max_char.high;
	fs->default_char = header.default_char.low +
		(header.default_char.high << 8);
	fs->min_bounds.lbearing = (short) header.min_bounds.left;
	fs->min_bounds.rbearing = (short) header.min_bounds.right;
	fs->min_bounds.width = (short) header.min_bounds.width;
	fs->min_bounds.ascent = (short) header.min_bounds.ascent;
	fs->min_bounds.descent = (short) header.min_bounds.descent;
	fs->max_bounds.lbearing = (short) header.max_bounds.left;
	fs->max_bounds.rbearing = (short) header.max_bounds.right;
	fs->max_bounds.width = (short) header.max_bounds.width;
	fs->max_bounds.ascent = (short) header.max_bounds.ascent;
	fs->max_bounds.descent = (short) header.max_bounds.descent;
	fs->ascent = (int) header.font_ascent;
	fs->descent = (int) header.font_descent;
	
	/* include the per_char array (which is initialized to zeros).
	 */
	numChars = fs->max_char_or_byte2 - fs->min_char_or_byte2 + 1;
	numChars *= fs->max_byte1 - fs->min_byte1 + 1;
	itmp = numChars * sizeof(XCharStruct);
	fs->per_char = (XCharStruct *) malloc(itmp);
	memset((void *) fs->per_char, 0, itmp);  /* to be filled on demand */

	if (resolution == TargetRes) {
		/* We're filling out target metrics, in milliems.  Convert,
		 * assuming that the font has been opened at AXGFS_TARGET_POINTSIZE.
		 */
		fs->min_bounds.lbearing = 
			(short) ((float) header.min_bounds.left * AxGFSMilliemFactor);
		fs->min_bounds.rbearing = 
			(short) ((float) header.min_bounds.right * AxGFSMilliemFactor);
		fs->min_bounds.width = 
			(short) ((float) header.min_bounds.width * AxGFSMilliemFactor);
		fs->min_bounds.ascent =
			(short) ((float) header.min_bounds.ascent * AxGFSMilliemFactor);
		fs->min_bounds.descent =
			(short) ((float) header.min_bounds.descent * AxGFSMilliemFactor);
		fs->max_bounds.lbearing =
			(short) ((float) header.max_bounds.left * AxGFSMilliemFactor);
		fs->max_bounds.rbearing =
			(short) ((float) header.max_bounds.right * AxGFSMilliemFactor);
		fs->max_bounds.width =
			(short) ((float) header.max_bounds.width * AxGFSMilliemFactor);
		fs->max_bounds.ascent =
			(short) ((float) header.max_bounds.ascent * AxGFSMilliemFactor);
		fs->max_bounds.descent =
			(short) ((float) header.max_bounds.descent * AxGFSMilliemFactor);
		fs->ascent = 
			(int) ((float) header.font_ascent * AxGFSMilliemFactor);
		fs->descent =
			(int) ((float) header.font_descent * AxGFSMilliemFactor);
	}
	
	/* add our font extension data to this struct.
	 */
	axgfsAddFontExtension(fs, resolution, pointsize, xlfd, ftab);
	
	return (fs);
}




int
axRealDisplayResolution()
{
	return (100);   /* how real can you get? */
}

void
axgfsFreeFontStruct(fs)
XFontStruct *fs;
{
	if (fs) {
		if (fs->fid) {
			FSCloseFont(AxGFSServer, fs->fid);
		}
		if (fs->ext_data) {
			axgfsFreeExtension(fs->ext_data);
			free(fs->ext_data);
		}
		if (fs->per_char) {
			free(fs->per_char);
		}
		free(fs);
	}
}

/* : haveGFSServer - report whether we've got a GFS server connection
 */
int
haveGFSServer()
{
	return (AxGFSServer != (AXFSServer *) NULL);
}

/* : isGFSxlfdName - return whether the passed xlfd name is available from
 * the Gallium server.  If true, return a reference to the name's xlfd template
 */
int
isGFSxlfdName(xlfd, retname)
char *xlfd;              /* in: xlfd name in some specific pointsize and res */
fontfile_xlfd **retname; /* out: ref from AxGFSxlfdNames (don't free) */
{
	int dashcount;
	char templ[MAX_XLFD], *inp, *outp;
	fontfile_xlfd *xf;
	
	if (!AxGFSServer || !AxGFSxlfdNames.width) {
		*retname = (fontfile_xlfd *) NULL;
		return FALSE;
	}

	/* copy the xlfd onto the template, zeroing out the metric
	 * fields while doing so  (fields: 7, 8, 9, 10, 12)
	 */
	for (inp = xlfd, outp = templ, dashcount = 0; *inp; ) {
		if (*inp == '-') {
			dashcount++;
		}
		if ((dashcount == 7) || (dashcount == 8) || (dashcount == 9) ||
			(dashcount == 10) || (dashcount == 12)) {

			*outp++ = '-';   /* copy out zeroed metric */
			*outp++ = '0';
			inp++;
			while (*inp && *inp != '-') {   /* skip rest of input field */
				inp++;
			}
		}
		else {
			*outp++ = *inp++;
		}
	}
	*outp = '\0';

	if (axhTestItem(templ, &AxGFSxlfdNames)) {
		axhGetItem(templ, &AxGFSxlfdNames, (char *) 0, (char **) &xf);
		*retname = xf;
		return (TRUE);
	}
	else {
		*retname = (fontfile_xlfd *) NULL;
		return (FALSE);
	}
}

#if !FOR_FILTER
/* : axgfsStartFontManager - start up the gallium font manager
 * program.  Return of (char *) NULL means OK, otherwise, caller
 * must free returned error message.
 */
char *
axgfsStartFontManager(galliumDir)
char *galliumDir;
{
	int start_server, cmdlen, ix, n_fontdirs;
	AXFSServer *serverConn;
	char hostnam[MAX_XLFD], *serverName, *managerProg;
	char **new_fontdirs, *serverProg;
	char **fontdirs = (char **) NULL;
	int set_path = TRUE;
	int retcode = 0;
	
	if (!galliumDir || !getenv(NXLT("AXGFSDIR"))) {
		return ((char *) NULL);
	}
	
	/* axmain currently links the internal version of the
	 * gfs server (libgfsi.a), and so we do our outline font work
	 * via this internal connection.  By contrast, the font manager
	 * program needs the external server.  For the font-install
	 * session, start the server on the current host, port 7001.
	 * example server name: "tcp/bigkitty:7001"
	 */
	start_server = axgfsX11R5();  /* start gfs server only if X server is R5 */
	
	if (start_server) {
		if (Agethostname(hostnam, MAX_XLFD - 1)) {
			start_server = FALSE;
		}
	}
	if (start_server) {
		
		serverName = (char *)
			malloc(strlen(hostnam) + 11);  /* room for "tcp/:7001" */
		strcpy(serverName, NXLT("tcp/"));
		strcat(serverName, hostnam);
		strcat(serverName, NXLT(":7001"));
		
		serverConn = FSOpenServer(serverName);
		
		if (!serverConn) {
			/* can't connect to the gfs server; treat this as the need
			 * to actually start it up.
			 */
			
			/* <serverName> is also the font path element for the
			 * gfs server.  Before starting gfs, look thru the current
			 * font paths, and clip out <serverName> if it is found there.
			 */
			fontdirs = XGetFontPath(Dpy, &n_fontdirs);
			if (fontdirs) {
				int numgfs, numnew, newix;
				
				/* count how many times <serverName> appears
				 */
				for (ix = 0, numgfs = 0; ix < n_fontdirs; ix++) {
					if (!strcmp(fontdirs[ix], serverName)) {
						numgfs++;
						break;
					}
				}
				if (numgfs) {
					/* need to clip out <serverName>
					 */
					numnew = n_fontdirs - numgfs;
					if (numnew <= 0) {
						free(serverName);
						XFreeFontPath(fontdirs);
						return((char *) NULL);
					}
					new_fontdirs = (char **) malloc
						(numnew * sizeof(char *));
					for (ix = 0, newix = 0; ix < n_fontdirs; ix++) {
						if (strcmp(fontdirs[ix], serverName)) {
							new_fontdirs[newix] = (char *)
								malloc(strlen(fontdirs[ix]) + 1);
							strcpy(new_fontdirs[newix], fontdirs[ix]);
							newix++;
						}
					}
					XSetFontPath(Dpy, new_fontdirs, numnew);
					XSync(Dpy, 0);  /* update xserver before starting gfs */
					for (ix = 0; ix < numnew; ix++) {
						free(new_fontdirs[ix]);
					}
					free(new_fontdirs);
				}
				XFreeFontPath(fontdirs);
			}
		
			/* The gfs servers's program file and startup script live in
			 * the "fs" subdirectory of galliumDir.  Use this command:
			 * "<galliumDir>/fs/startgfs <galliumDir>"
			 */
			cmdlen = (strlen(galliumDir) * 2) + 20;  /* plus "/fs/startgfs " */
			serverProg = (char *) malloc(cmdlen);
			strcpy(serverProg, galliumDir);
			strcat(serverProg, NXLT("/fs/startgfs "));
			strcat(serverProg, galliumDir);
			
			retcode = AxRunProgram (serverProg, "", "", "");
			free(serverProg);

			/* now add the server to the font path if not already there.
			 */
			fontdirs = XGetFontPath(Dpy, &n_fontdirs);
			if (fontdirs) {
				for (ix = 0; ix < n_fontdirs; ix++) {
					if (!strcmp(fontdirs[ix], serverName)) {
						set_path = FALSE;
						break;
					}
				}
				if (set_path) {
					/* Note the copying in this block -- can't mix
					 * our and Xlib's allocators, which differ on
					 * some ports.  Also, serverName freed below, not here.
					 */
					new_fontdirs = (char **) malloc
						((n_fontdirs + 1) * sizeof(char *));
					for (ix = 0; ix < n_fontdirs; ix++) {
						new_fontdirs[ix] = (char *)
							malloc(strlen(fontdirs[ix]) + 1);
						strcpy(new_fontdirs[ix], fontdirs[ix]);
					}
					new_fontdirs[n_fontdirs] = serverName;
					
					XSetFontPath(Dpy, new_fontdirs, n_fontdirs + 1);
					XSync(Dpy, 0);  /* update Xserver before fontfm started */
					
					for (ix = 0; ix < n_fontdirs; ix++) {
						free(new_fontdirs[ix]);
					}
					free(new_fontdirs);
				}
			}
			XFreeFontPath(fontdirs);
		}
		else {
			FSCloseServer(serverConn);
		}
		
		free(serverName);
	}

	/* start up the font manager program.  
	 * This program and its startups script live in galliumDir directly,
	 * invoked as "<galliumDir>/startfontfm <galliumDir>"
	 */
	cmdlen = (strlen(galliumDir) * 2) + 20;  /* plus "/startfontfm " */
	managerProg = (char *) malloc(cmdlen);
	strcpy(managerProg, galliumDir);
	strcat(managerProg, NXLT("/startfontfm "));
	strcat(managerProg, galliumDir);
	
	retcode = AxRunProgram (managerProg, "", "", "");
	free(managerProg);
	
	return ((char *) NULL);
}
#endif

/* : axgfsFreeAllPLetters - free all the letterforms on the
 * passed printer font extension instance.
 */
void
axgfsFreeAllPLetters(ext)
axpfs_ext *ext;
{
	if (!ext || (ext->com.resolution != TargetRes)) {
		return;
	}
	axhDeleteTable(&ext->com.letters, (void *) NULL);
	axhInitTable(&ext->com.letters, 31,
				 axgfsNewPLetter, axgfsFreePLetter);
}




void
axgfsInitCharStruct(fs, charcode, resolution, xc)
XFontStruct *fs;            /* in */
int charcode;               /* in */
int resolution;             /* in */
XCharStruct *xc;            /* in/out: init caller's memory */
{
	int ret, tcode, ix;
	int haveChar = TRUE;
	axfsChar2b tchar;
	axfsCharInfo *fc;
	axfs_ext *ext;

	ext = (axfs_ext *) fs->ext_data->private_data;
	
	if (ext->com.nativeCodes) {
		ix = charcode - ext->com.nativeCodes->mincode;
		tcode = ext->com.nativeCodes->codes[ix];
	}
	else {
		tcode = charcode;
	}
	
	tchar.high = (tcode >> 8) & 0xff;
	tchar.low = tcode & 0xff;
	
	if (ext->com.resolution == ScreenRes) {
		/* first fetch the letterform for the passed font, charcode.
		 * this is reasonable for screenfonts because of the high
		 * likelihood that the letterform will be needed.
		 */
		int idx = axgfsGetQLetterIndex(charcode, fs);
		ext->qletters[idx] = axgfsIntGetLetter(charcode, fs, NULL, (Font) fs->fid,
											   TRUE, FALSE);
	}

	if (tcode >= 0) {
		ret = FSQueryXExtents16(AxGFSServer, (Font) fs->fid, False,
								&tchar, 1, &fc);
		if (ret == AXFSSuccess) {
			xc->lbearing = (short) fc[0].left;
			xc->rbearing = (short) fc[0].right;
			xc->width = (short) fc[0].width;
			xc->ascent = (short) fc[0].ascent;
			xc->descent = (short) fc[0].descent;
			FSFree((char *) fc);
 			if (!(xc->rbearing - xc->lbearing) && !xc->width) {
				haveChar = FALSE;
			}
		}
		else {
			haveChar = FALSE;
		}
	}
	else {
		haveChar = FALSE;
	}
	if (!haveChar) {
		/* assume that no letterform is present at this codepoint.
		 * provide "error box" metrics instead.
		 */
		r32 bbox;
		noCharBox(fs, &bbox);
		xc->lbearing = 0;
		xc->rbearing = bbox.width;
		xc->width = bbox.width;
		xc->ascent = -bbox.origin.y;
		xc->descent = 0;
	}
	if (haveChar && (resolution == TargetRes)) {
		/* We're filling out target metrics, in milliems.  Convert.
		 */
		xc->lbearing = (short) ((float) xc->lbearing * AxGFSMilliemFactor);
		xc->rbearing = (short) ((float) xc->rbearing * AxGFSMilliemFactor);
		xc->width = (short) ((float) xc->width * AxGFSMilliemFactor);
		xc->ascent = (short) ((float) xc->ascent * AxGFSMilliemFactor);
		xc->descent = (short) ((float) xc->descent * AxGFSMilliemFactor);
	}
}

static float
axgfsGetMilliemFactor()
{
	float ffactor;
	int displayRes;

	displayRes = axRealDisplayResolution();
	
	/* this factor will produce milliem numbers when multiplied
	 * against the dpi figures of a font opened at AXGFS_TARGET_POINTSIZE
	 */
	ffactor = ((float) 72000) /
		((float) AXGFS_TARGET_POINTSIZE * (float) displayRes);
	
	return (ffactor);
}

static int
axgfsX11R5()
{
	if (strstr(ServerVendor(Dpy), "X11/NeWS")) {
		/* sunOS4 has scalable fonts available thru the Xlib API,
		 * but is not X11R5;  don't know of any other such cases.
		 */
		return FALSE;
	}
	else {
		if (serverHasScalables()) {
			return TRUE;
		}
		else {
			return FALSE;
		}
	}
}

static int
serverHasScalables()
{
	char **fonts;
	int n_fonts;
	
	fonts = XListFonts(Dpy, "-*-*-*-*-*-*-0-0-*-*-*-0-*-*", 2, &n_fonts);
	if (fonts) {
		XFreeFontNames(fonts);
	}
	else {
		return (FALSE);
	}
	return (n_fonts > 0);
}

/*==============================*/








char *
axgfsIntGetFid(xlfd_template, pointsize, resolution, ctm)
char *xlfd_template;    /* in */
int pointsize;          /* in; in points */
int resolution;         /* in; one of ScreenRes, TargetRes */
axctm *ctm;             /* in */
{
	char fullname[MAX_XLFD + 64];
	Font tfont, other_id;
	
	axtFormScalableName(xlfd_template, pointsize, resolution, ctm, fullname);
	other_id = (Font) 0;
    tfont = FSOpenBitmapFont(AxGFSServer, 0, 0, fullname, &other_id);
	if (other_id) {
		tfont = other_id;
	}
	return ((char *) tfont);
}


axletter *
axgfsIntGetLetter(charcode, fs, ctm, fid, pix, bbox)
int charcode;     /* in */
XFontStruct *fs;  /* in */
axctm *ctm ;	  /* in, but used only in MS */
Font fid;         /* in */
int pix;          /* in: bool; whether to return pixmap or just bits */
int bbox;         /* in: bool; whether to include a bbox in the return */
{
	int ret, tcode, haveLetter, haveCodepoint, idx;
	unsigned int width, height;
	XImage *charImage;
	Pixmap letterPix;
	axfsChar2b tchar;
	axfsBitmapFormat bitformat;
	r32 errorCharBox;
	axfs_com *extcom;
	axletter *thisLetter = (axletter *) 0;
	axfsCharInfo *fc = (axfsCharInfo *) 0;
	axfsOffset *offsets = (axfsOffset *) 0;
	unsigned char *glyph_data = (unsigned char *) 0;
	
	if (!fs || !fid) {
		return ((axletter *) NULL);
	}

	/* use this fid,charcode to get a letterform
	 */
	extcom = (axfs_com *) fs->ext_data->private_data;
	if (extcom->nativeCodes) {
		idx = charcode - extcom->nativeCodes->mincode;
		charcode = extcom->nativeCodes->codes[idx];
	}
	tcode = charcode;
	haveCodepoint = haveLetter = TRUE;
	if (tcode < 0) {
		haveLetter = haveCodepoint = FALSE;
	}
	else {
		tchar.high = (tcode >> 8) & 0xff;
		tchar.low = tcode & 0xff;
		bitformat = pix ? AxGFSBitmapFormat : AxGFSPrinterBitmapFormat;
		ret = FSQueryXBitmaps16(AxGFSServer, fid, bitformat,
								False, &tchar, 1, &offsets, &glyph_data);
		if (ret != AXFSSuccess) {
			haveLetter = haveCodepoint = FALSE;
		}
		else {
			ret = FSQueryXExtents16(AxGFSServer, fid, False, &tchar, 1, &fc);
			if (ret != AXFSSuccess) {
				goto all_done;
			}
			width = (unsigned int) ((int) fc[0].right - (int) fc[0].left);
			height = (unsigned int) ((int) fc[0].ascent + (int) fc[0].descent);
			if (!(width && height)) {
				haveLetter = FALSE;  /* no letterform raster */
			}
			if (!(haveLetter || fc[0].width)) {
				/* assume font doesn't supply anything for this codepoint
				 */
				haveCodepoint = FALSE;
			}
		}
	}
	
	if (!(haveLetter && haveCodepoint)) {
		noCharBox(fs, &errorCharBox);
	}
	
	if (pix) {
		/* draw the letterform onto a pixmap, return the pixmap handle.
		 */
		if (haveLetter && haveCodepoint) {
			charImage =
				XCreateImage(Dpy,
							 DefaultVisual(Dpy, DefaultScreen(Dpy)),
							 1, XYPixmap, 0,
							 (char *) &glyph_data[offsets[0].position],
							 width, height, BitmapPad(Dpy), 0);
			if (!charImage) {
				goto all_done;
			}
		}
		else {
			width = errorCharBox.width;
			height = errorCharBox.height;
		}
		letterPix = XCreatePixmap(Dpy, DefaultRootWindow(Dpy),
								  width, height, 1);
		XFillRectangle(Dpy, letterPix, PixClearGC, 0, 0, width, height);
		if (haveLetter && haveCodepoint) {
			/* shoot the letterform onto the pixmap
			 */
			XPutImage(Dpy, letterPix, PixGC, charImage, 0, 0, 0, 0,
					  width, height);
			charImage->data = (char *) NULL; /* bits belong to gfslib */
			XDestroyImage(charImage);
		}
		else if ((fid == fs->fid) && !haveCodepoint) {
			/* no letterform at this codepoint; draw box onto the pixmap..
			 * (for screen letterforms, condition fid==fs->fid implies
			 * that the letterform is untransformed; we do error-box
			 * for this condition).
			 */
			if (width > 0 && height > 0) {
				XDrawRectangle(Dpy, letterPix, PixGC, 1, 1,
							   width-2 > 0 ? width - 2 : 0,
							   height-2 > 0 ? height - 2 : 0);
			}
		}
		thisLetter = (axletter *) malloc (sizeof(axletter));
		memset((void *) thisLetter, 0, sizeof(axletter));
		thisLetter->bits = (void *) letterPix;
	}
	else {
		/* return the letterform bits directly
		 */
		thisLetter = (axletter *) malloc (sizeof(axletter));
		memset((void *) thisLetter, 0, sizeof(axletter));
		if (haveLetter) {
			thisLetter->bits = (void *) malloc(offsets[0].length);
			memcpy(thisLetter->bits, (void *) &glyph_data[offsets[0].position],
				   offsets[0].length);
		}
	}
	if (bbox) {
		thisLetter->bbox = (r32 *) malloc(sizeof(r32));
		if (haveLetter) {
			thisLetter->bbox->origin.x = (int) fc[0].left;
			thisLetter->bbox->origin.y = (int) -(fc[0].ascent);
			thisLetter->bbox->width = width;
			thisLetter->bbox->height = height;
		}
		else {
			*(thisLetter->bbox) = errorCharBox;
		}
	}
 all_done:
	if (glyph_data) {
		FSFree((char *) glyph_data);
	}
	if (offsets) {
		FSFree((char *) offsets);
	}
	if (fc) {
		FSFree((char *) fc);
	}
	return (thisLetter);
}

/* Return the dimensions of a box to represent a missing codepoint.
 */
static void
noCharBox(fs, bbox)
XFontStruct *fs;       /* in */
r32 *bbox;             /* out: init caller's memory */
{
	int width;

	width = fs->min_bounds.width;
	if (width < 4) {
		width = 4;
	}
	bbox->width = width;
	bbox->height = fs->ascent;
	bbox->origin.x = 0;
	bbox->origin.y = -fs->ascent;
}

static int
axgfsErrorFunc(server, fserror)
AXFSServer *server;
AXFSErrorEvent *fserror;
{
	return (TRUE);
}

/* %%%%%% This func needs a bool param telling whether
 * the passed server handle is imbedded.  We currently
 * use only the imbedded flavor, and so we aren't worried
 * about the IO error condition.  We SHOULD be worried if
 * this condition occurs on a connection to an external
 * server, however....
 */
static int
axgfsIOErrorFunc(server)
AXFSServer *server;
{
	return (TRUE);
}

static char *
axNewEncodingMap(mapname, notUsed)
char *mapname;
char *notUsed;
{
	char *fromname, *toname, *ip, *op;
	ax_encoding_map *emap;
	int len;
	
	/* the mapname is expected to be a comma-separated pairing
	 * of the from/to encoding filenames, with no extra space,
	 * for instance, "latin1.enc,bics.enc".  extract fromname and toname
	 */
	len = strlen(mapname) + 1;
	fromname = (char *) malloc(len);
	toname = (char *) malloc(len);
	memset((void *) fromname, 0, len);
	memset((void *) toname, 0, len);
	for (ip = mapname, op = fromname; *ip && *ip != ','; ) {
		*op++ = *ip++;
	}
	if (*ip != ',') {
		free(fromname);
		free(toname);
		return ((char *) NULL);
	}
	for (op = toname, ip++; *ip; ) {
		*op++ = *ip++;
	}
	axGetEncodingMap(fromname, toname, &emap);
	free(fromname);
	free(toname);
	return ((char *) emap);
}

static void
axFreeEncodingMap(mapp, notUsed)
char *mapp;
char *notUsed;
{
	ax_encoding_map *emap = (ax_encoding_map *) mapp;
	if (emap) {
		if (emap->codes) {
			free(emap->codes);
		}
		free(emap);
	}
}
isGFSOutlineSupported()
{
	int maj_opcode, frst_ev, frst_err_code;

	return(FSQueryExtension(AxGFSServer, "GLYPH_OUTLINES", &maj_opcode,
							&frst_ev, &frst_err_code));
}
#if 0
axgfxGetGlyphOutline(fs, charcode, fp)
XFontStruct *fs;
int charcode;     /* in */
FILE *fp; /* temp */
{
	int i, j, p_ix, xm, xd, ym, yd, sz;
	int n_operands, n_points, n_contours;
	axfsCharInfo metrics;
	GFSOperand *operands;
	GFSPoint *points;
	Font font;
	int n_pnts;
	TmPoint32 *out_pnts;

	font = (Font)fs->fid;
#if (LINUX && !LIBC5) /* newer fontfs code calls this FTGetGlyphOutline */
	FTGetGlyphOutline(AxGFSServer, font, (unsigned short)charcode,
						True, &metrics, &n_contours,
						&n_operands, &operands,
						&n_points, &points);
#else
	GFSGetGlyphOutline(AxGFSServer, font, (unsigned short)charcode,
						True, &metrics, &n_contours,
						&n_operands, &operands,
						&n_points, &points);
#endif
	fprintf(fp, "Num Of Contours For Char  %c is %d \n", charcode, n_contours);
	for ( p_ix = 0, i=0; i< n_operands; i++ )
	{
		switch ( operands[i] )
		{
			case StartContour:
				fprintf(fp, "StartContour\n");
				xm = (points[p_ix].x & 0xFFFF0000)>>16;
				xd = (points[p_ix].x & 0x0000FFFF);
				ym = (points[p_ix].y & 0xFFFF0000)>>16;
				yd = (points[p_ix].y & 0x0000FFFF);
				fprintf(fp, "\t %d - (%d.%d, %d.%d)\n", p_ix, xm, xd, ym, yd);
				p_ix++;
				n_pnts++;
				sz = n_pnts*sizeof(TmPoint32);
				out_pnts = (TmPoint32 *)realloc(out_pnts, sz);
				out_pnts[n_pnts-1].x = 0;
				out_pnts[n_pnts-1].y = 0;
			break;
			case Move:
				fprintf(fp, "Move\n");
				xm = (points[p_ix].x & 0xFFFF0000)>>16;
				xd = (points[p_ix].x & 0x0000FFFF);
				ym = (points[p_ix].y & 0xFFFF0000)>>16;
				yd = (points[p_ix].y & 0x0000FFFF);
				fprintf(fp, "\t %d - (%d.%d, %d.%d)\n", p_ix, xm, xd, ym, yd);
				p_ix++;
			break;
			case Curve:
				fprintf(fp, "Curve\n");
				for (j=0; j< 3; j++ )
				{
					xm = (points[p_ix].x & 0xFFFF0000)>>16;
					xd = (points[p_ix].x & 0x0000FFFF);
					ym = (points[p_ix].y & 0xFFFF0000)>>16;
					yd = (points[p_ix].y & 0x0000FFFF);
					fprintf(fp, "\t %d - (%d.%d, %d.%d)\n", p_ix, xm, xd, ym, yd);
					p_ix++;
				}
			break;
			case EndContour:
				fprintf(fp, "EndContour\n");
			break;
		}
	}
	if ( operands != NULL )
		FSFree((char *)operands);
	if ( points != NULL )
		FSFree((char *)points);
}

axgfxGetOutline(name)
char *name;
{
	int n_operands, n_points, n_contours;
	axfsCharInfo metrics;
	GFSOperand *operands;
	GFSPoint *points;
	XFontStruct *fs;

	if ( isGFSOutlineSupported())
	{
		fontfile_xlfd *xf;

		if (isGFSxlfdName(name, &xf))
		{
			FILE *fp;
			int c;

			fs = axgfsOpenFont(xf->xlfd_template, ScreenRes,
								axtDecipointSize(name) / 10, xf->ftab);
			fp = fopen("/tmp/foo", "w");
			if ( fp )
			{
				for ( c=65; c < 128; c++ )
					axgfxGetGlyphOutline(fs, c, fp);

				fclose(fp);
			}
		}
	}
}
#endif
