/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axgfs.h
   Applix ASC version: 6
   Date: Fri May  7 21:11:05 1999
*/
#ifndef _AXGFS_H
#define _AXGFS_H

#if AX_X11
#include <X11/Xlib.h>
#include <X11/Xmd.h>
#endif

/* The pointsize at which fonts are opened to derive milliem metrics.
 * We sure hope that at this size, hinting effects fall out....
 */
#define AXGFS_TARGET_POINTSIZE 200

/* Supplying a hard-coded extension number as we do here is quite a hack;
 * really should use XRegisterExtension (methinks), but can't find this call.
 * This introduces the slim possibility that a real XFontStruct extension
 * will appear on the return of XLoadQueryFont, and just happen to have
 * exactly this extension number.
 */
#define AXGFS_FONT_EXTENSION 8388605

/* If this macro returns true then we think that the fs->fid was
 * returned to us by the Gallium font server; if false, then we
 * assume that fs->fid was returned by the X server.  Fids returned
 * by a particular server can be used only with that server.
 */
#define IS_AXGFS_FONT(fs) (fs->ext_data &&\
						   (fs->ext_data->number == AXGFS_FONT_EXTENSION))

#define AXGFS_FONT_RES(fs) \
(((axfs_com *) (fs->ext_data->private_data))->resolution)

/* This error message signifies that no Gallium config file could be found
 * It doesn't get printed; instead it means to fail silently, disabling gfs.
 */
#define AXGFS_NO_CONFIG  (NXLT("NoConfig"))

/* return type for letterform bitmaps
 */
typedef struct axletter_t {
	void *bits;               /* letterform bitmap, or pixmap handle */
	r32 *bbox;                /* optional bounding box */
} axletter;

/* type used for mapping codepoints in one encoding to another
 */
typedef struct ax_encoding_map_t {
	int mincode;              /* offset term; minimum codepoint in 'from' */
	int maxcode;              /* maximum codepoint in 'from' */
	int *codes;               /* 'from' codepoint -> 'to' codepoint */
} ax_encoding_map;

/* common portion of screen/printer font extension types
 */
typedef struct axfs_com_t {
	int resolution;            /* typically TargetRes or ScreenRes */
	char *xlfd;                /* ref into AxGFSxlfdNames -- do not free */
	axhtab fids;               /* axctm -> fid of r6-style xformed xlfd */
	axhtab letters;            /* axformed_lettername -> axletter pointer */ 
	ax_encoding_map *nativeCodes;  /* optional codepoint map array */
} axfs_com;

/* screenfont extension type
 */
typedef struct axfs_ext_t {    /* we use this type on XFontStruct->ext_data */
	axfs_com com;              /* screen/printer common stuff; must be 1st */
	int pointsize;
	int numChars;        /* #elements in qletters array */
	axletter **qletters; /* unxformed, indexed identically to fs->per_char */
} axfs_ext;

/* printerfont extension type
 */
typedef struct axpfs_ext_t {   /* we use this type on XFontStruct->ext_data */
	axfs_com com;              /* screen/printer common stuff; must be 1st */
} axpfs_ext;

/* name type for transformed screen letters
 */
typedef struct axformed_lettername_t {
	unsigned short charcode;
	axctm ctm;
} axformed_lettername;	

/* name type for printer font fids
 */
typedef struct axpfid_t {
	int pointsize;
	axctm ctm;
} axpfid;

/* name type for printer letters
 */
typedef struct plettername_t {
	unsigned short charcode;
	axpfid pfid;
} plettername;

/* type used during gfs config file reading
 */
typedef struct fontfile_xlfd_t {
	char *xlfd_template;
	char *dir;                  /* gfs directory where file was found */
	char *filename;             /* font file name, e.g., "timesb.ttf" */
	fontTabEnt *ftab;           /* font record associated with this xlfd */
} fontfile_xlfd;


#if !AXMS    /* { */
typedef unsigned long AXFSID;

typedef struct AXFSalternate {
    Bool        subset;
    char       *name;
}           AXFSAlternateServer;

/* extension stuff */
typedef struct AXFSExtData {
    int         number;     /* number returned by FSRegisterExtension */
    struct AXFSExtData *next;    /* next item on list of data for structure */
    int         (*free_private) (); /* called to free private storage */
    char       *private_data;   /* data private to this extension. */
}           AXFSExtData;
 
typedef struct {        /* public to extension, cannot be changed */
    int         extension;  /* extension number */
    int         major_opcode;   /* major op-code assigned by server */
    int         first_event;    /* first event number for the extension */
    int         first_error;    /* first error number for the extension */
}           AXFSExtCodes;
 
typedef struct AXFSExtent {
    struct AXFSExtent *next; /* next in list */
    AXFSExtCodes  codes;      /* public information, all extension told */
    int         (*close_server) (); /* routine to call when connection
                     * closed */
    int         (*error) ();    /* who to call when an error occurs */
    int         (*error_string) (); /* routine to supply error string */
    char       *name;
}           AXFSExtension;

/* server data structure */
typedef struct AXFSServer {
    struct AXFSServer *next;
    int         fd;
    int         proto_version;
    char       *vendor;
    int         byte_order;
    int         vnumber;
    int         release;
    int         resource_id;
    struct AXFSQEvent *head,
               *tail;
    int         qlen;
    unsigned long last_request_read;
    unsigned long request;
    char       *last_req;
    char       *buffer;
    char       *bufptr;
    char       *bufmax;
    unsigned    max_request_size;
    char       *server_name;
    char       *auth_data;
    AXFSAlternateServer *alternate_servers;
    int         num_alternates;
    AXFSExtData  *ext_data;
    AXFSExtension *ext_procs;
    int         ext_number;
    Bool        (*event_vec[132]) ();
                Status(*wire_vec[132]) ();
    char       *scratch_buffer;
    unsigned long scratch_length;
    int         (*synchandler) ();
    unsigned long flags;
}           AXFSServer;

typedef struct {
    int         type;
    unsigned long serial;
    Bool        send_event;
    AXFSServer   *server;
}           AXFSAnyEvent;

typedef union AXFSEvent {
    int         type;
    AXFSAnyEvent  fsany;
}           AXFSEvent;

typedef struct AXFSQEvent {
    struct AXFSQEvent *next;
    AXFSEvent     event;
}           AXFSQEvent;

typedef struct {
    int         type;
    AXFSServer   *server;
    AXFSID        resourceid;
    unsigned long serial;
    unsigned char error_code;
    unsigned char request_code;
    unsigned char minor_code;
}           AXFSErrorEvent;

typedef struct {
    INT16   left B16,
                right B16;
    INT16   width B16;
    INT16   ascent B16,
                descent B16;
    CARD16  attributes B16;
}           axfsCharInfo;
 
typedef struct {
    CARD8       high;
    CARD8       low;
}           axfsChar2b;
 
typedef struct {
    axfsChar2b    min_char,
                max_char;
}           axfsRange;

typedef struct {
    CARD32  flags B32;
    axfsRange     char_range;
    CARD8   draw_direction;
    CARD8   pad;
    axfsChar2b    default_char;
    axfsCharInfo  min_bounds,
                max_bounds;
    INT16   font_ascent B16,
                font_descent B16;
    /* propinfo */
}           axfsFontHeader;

typedef struct  {
    CARD32  position B32;
    CARD32  length B32;
}       axfsOffset;
 
typedef struct {
    axfsOffset    name;
    axfsOffset    value;
    CARD8   type;
    BYTE        pad0;
    CARD16  pad1 B16;
}           axfsPropOffset;
 
typedef struct {
    CARD32  num_offsets B32;
    CARD32  data_len B32;
    /* offsets */
    /* data */
}       axfsPropInfo;




#if ALPHA || (LINUX && __alpha__)
typedef unsigned int  axfsBitmapFormat;
typedef unsigned int  axfsBitmapFormatMask;
#else
typedef unsigned long  axfsBitmapFormat;
typedef unsigned long  axfsBitmapFormatMask;
#endif


#define axfsBitmapFormatByteOrderLSB        (0)
#define axfsBitmapFormatByteOrderMSB        (1L << 0)
#define axfsBitmapFormatBitOrderLSB         (0)
#define axfsBitmapFormatBitOrderMSB         (1L << 1)

#define axfsBitmapFormatImageRectMin        (0L << 2)

#define axfsBitmapFormatScanlinePad8        (0L << 8)
#define axfsBitmapFormatScanlinePad16       (1L << 8)
#define axfsBitmapFormatScanlinePad32       (2L << 8)
#define axfsBitmapFormatScanlinePad64       (3L << 8)
 
#define axfsBitmapFormatScanlineUnit8       (0L << 12)
#define axfsBitmapFormatScanlineUnit16      (1L << 12)
#define axfsBitmapFormatScanlineUnit32      (2L << 12)
#define axfsBitmapFormatScanlineUnit64      (3L << 12)


#define    AXFSSuccess               -1

AXFSServer   *FSOpenServer();
#endif    /* } !AXMS */


#endif  /* _AXGFS_H */
