/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axtfm.c
   Applix ASC version: 16
   Date: Tue May  4 21:11:02 1999
*/
#include <stdio.h>
#include <axtext.h>

/**************************************************************************\
 *                      tag segment keys                                  * 
\**************************************************************************/
#define  tagSUBFILE          400 
#define  tagCPYRIGHT        401
#define  tagCOMMENT          402
#define  tagSYMBOLMAP        403
#define  tagSYMBOLSETDIR     404
#define  tagUNIQUEASSOCID    405
#define  tagPOINT            406 
#define  tagNOMINALPOINT     407
#define  tagDESIGNUNITS      408
#define  tagTYPESTRUCT       410
#define  tagSTROKEWT         411
#define  tagSPACING          412
#define  tagSLANT            413
#define  tagAPPEARWIDTH      414
#define  tagSERIFSTYLE       415
#define  tagTYPESTYLE        416
#define  tagTYPEFACE         417
#define  tagTFSOURCE         418
#define  tagAVERAGEWD        419
#define  tagMAXWIDTH         420
#define  tagINTERWORDSP      421
#define  tagRECLINESP        422
#define  tagCAPHEIGHT        423
#define  tagXHEIGHT          424
#define  tagASCENT           425
#define  tagDESCENT          426
#define  tagLOWERASCENT      427
#define  tagLOWERDESCENT     428
#define  tagUNDERDEPTH       429
#define  tagUNDERTHICK       430
#define  tagUPPERACCENT      431
#define  tagLOWERACCENT      432
#define  tagHORIZONTALESC    433
#define  tagVERTICALESC      434
#define  tagLEFTEXTENT       435
#define  tagRIGHTEXTENT      436
#define  tagCHARASCENT       437
#define  tagCHARDESCENT      438
#define  tagKERNPAIRS        439
#define  tagSECTORKERN       440
#define  tagTRACKKERN        441
#define  tagTFSELECTSTR      442

/**************************************************************************\
 *                    Data type definitions                               * 
\**************************************************************************/
#define  tagBYTE           1
#define  tagASCII          2
#define  tagSHORT          3
#define  tagLONG           4
#define  tagRATIONAL       5
#define  tagSIGNEDBYTE    16
#define  tagSIGNEDSHORT   17
#define  tagSIGNEDLONG    18

/**************************************************************************\
 *                    Data structures                                     *
\**************************************************************************/
struct general_type {
	long          n_chars;
	short         n_symbol_sets;
};

struct metrics_type {
	double        point;
	double        nominal_pointsize;
	double        design_units;
	unsigned char weight;
	unsigned char appearance_width;
	unsigned char serif_style;
	unsigned char type_style;
	unsigned char type_struct;
	short         spacing;
	short         slant;
	double        average_width;
	short         maximum_width;
	short         word_spacing;
	short         line_spacing;
	short         capheight;
	short         x_height;
	short         ascent;
	short         descent;
	short         lowercase_ascent;
	short         lowercase_descent;
	short         underscore_descent;
	short         underscore_thickness;
	short         uppercase_accent_height;
	short         lowercase_accent_height;
};

struct symbolset_directory_type {
	unsigned char name[30];
	unsigned char selection_name[5];
	short        *index;
	short         index_len;  /* size of symbolIndex */
};

struct symbol_type {
	short                           *symbol_map;
	struct symbolset_directory_type *symbolset_directory;
};

struct character_type {
	short         horizontal_escapement;
	short         vertical_escapement;
	short         left_extent;
	short         right_extent;
	short         character_ascent;
	short         character_descent;
};

struct pairs_type {
	short         ix1;
	short         ix2;
	short         val;
};

struct sector_type {
	short         ix;
	short        *left;
	short        *right;
};

struct track_type {
	short         val;
	short         max_pointsize;
	short         min_pointsize;
	short         max_kern;
	short         min_kern;
};

struct kerning_type {
	short               n_pairs;
	struct pairs_type  *pairs;

	short               n_chars;
	short               n_sectors;
	struct sector_type *sector;

	short               n_tracks;
	struct track_type  *track;
};

struct face_type {
	struct general_type    general;
	struct metrics_type    metrics;
	struct symbol_type     symbol;
	struct character_type *character;
	struct kerning_type    kerning;
};

struct tfm_type {
	struct face_type *face;
	short n_faces;
};

extern FILE *getFontmetricsFile();
extern int freeTargetMetrics();
extern int fontMetricsSize();
extern int qRound();

/*****************************************************************************/
static void
get_string(fp, string)
FILE *fp;
char *string;
{
	int i = 0;
	while ((string[i++] = fgetc(fp)) != 0);
}

/*****************************************************************************/
static short
get_short(fp, intel_order)
FILE *fp;
int intel_order;
{
	short b1,b2;

	if ( intel_order )
	{
		b1 = (short)fgetc(fp);
		b2 = (short)fgetc(fp);
	}
	else
	{
		b2 = (short)fgetc(fp);
		b1 = (short)fgetc(fp);
	}
	return ((b2<<8) + b1);
}

/*****************************************************************************/
static long
get_long(fp, intel_order)
FILE *fp;
int intel_order;
{
	long b1,b2,b3,b4;

	if ( intel_order )
	{
		b1 = (long)fgetc(fp);
		b2 = (long)fgetc(fp);
		b3 = (long)fgetc(fp);
		b4 = (long)fgetc(fp);
	}
	else
	{
		b4 = (long)fgetc(fp);
		b3 = (long)fgetc(fp);
		b2 = (long)fgetc(fp);
		b1 = (long)fgetc(fp);
	}
	return((b4<<24)+(b3<<16)+(b2<<8)+b1);
}

/*****************************************************************************/
static double
get_double(fp, intel_order)
FILE *fp;
int intel_order;
{
	double num, den;

	num = (double)get_long(fp, intel_order); /* numerator   */
	den = (double)get_long(fp, intel_order); /* denominator */

	return (num/den);
}

/*****************************************************************************/
static void
get_n_faces(tfm, fp, intel_order)
struct tfm_type *tfm;
FILE *fp;
int intel_order;
{
	long curr_offset; /* current offset into the file */
	long base_offset; /* offset to first directory */
	long face_offset; /* offset to additional directories */
	long next_face;   /* value at the offset reigon */
	short n_tags;     /* number of tag entries in a face */
	int done;

	/* get the offset to the first directory, and seek to it */
	curr_offset = ftell(fp);
	base_offset = get_long(fp, intel_order);
	fseek(fp, base_offset, 0);
	done = FALSE;

	/* keep adding faces as long as there are more directories */
	tfm->n_faces = 1;
	while ( !done )
	{
		/* get the offset to the actual face (skipping tags) */
		n_tags = get_short(fp, intel_order);
		face_offset = ((n_tags * 12) + base_offset + 2); 
		fseek(fp, face_offset, 0);

		/* get next face offset */
		next_face = get_long(fp, intel_order);
		if ( next_face == 0L )
			done = TRUE;
		else
		{
			fseek(fp, next_face, 0);
			tfm->n_faces += 1;
			/* shouldn't this update the base offset each time? */
		}
	}

	/* reset the position */
	fseek(fp, curr_offset, 0);
}

/*****************************************************************************/
static void
free_tfm(tfm)
struct tfm_type *tfm;
{
	int tx, ix;
	struct face_type *tf;
	struct symbolset_directory_type *sd;
	struct kerning_type *kn;

	if (!tfm || !tfm->face)
		return;

	for (tx = 0; tx < tfm->n_faces; tx++) {
		tf = &tfm->face[tx];
		if (tf->symbol.symbol_map) {
			free((char *) tf->symbol.symbol_map);
		}
		if (tf->symbol.symbolset_directory) {
			for (ix = 0; ix < tf->general.n_symbol_sets; ix++) {
				sd = &tf->symbol.symbolset_directory[ix];
				if (sd->index) {
					free((char *) sd->index);
				}
			}
			free((char *) tf->symbol.symbolset_directory);
		}
		if (tf->character) {
			free((char *) tf->character);
		}
		kn = &tf->kerning;
		if (kn->pairs) {
			free((char *) kn->pairs);
		}
		if (kn->sector) {
			for (ix = 0; ix < kn->n_chars; ix++) {
				if (kn->sector[ix].left) {
					free((char *) kn->sector[ix].left);
				}
				if (kn->sector[ix].right) {
					free((char *) kn->sector[ix].right);
				}
			}
			free((char *) kn->sector);
		}
		if (kn->track) {
			free((char *) kn->track);
		}
	}
	free((char *) tfm->face);
	free((char *) tfm);
}

    /* : readTfmFile - initialize a tfm_type instance from
	   a .tfm font metrics file */

static struct tfm_type *
readTfmFile(name, encoding)
char *name;        /* e.g., "92500238.tfm" */
char *encoding;    /* e.g., "ECMA-94 Latin 1" */
{
	unsigned char jnk_str[256]; /* for strings we read but don't use */
	FILE *fp;
	struct tfm_type *tfm;
	short type_size[19]; /* stores size in bytes of each type */
	short tshort;
	int ok = TRUE;
	int ix, iy, iz, kx, tmp, numSymbolSets, thisSet, found;
	int intel_order;
	unsigned char tmpStr[30];     /* >= size of name member of
							symbolset_directory_type struct */
	int16 ptype,          /* processor type (intel vs motorola) */
	     n_tags,    /* number of tags per typeface */
	     tag_value,      /* numerical tag value */
	     data_type,      /* char, short, long, ... */
	     n_pairs,   /* # kern pairs per symbol set */
	     n_chars,
	     n_sectors, /* # kern sectors per symbol set */
	     n_tracks;  /* # kern tracks per symbol set */
	long dir_offset,     /* offset to first directory */
	     block_size,     /* size of data for a tag (in its own
							datatype) */
	     length,         /* size in bytes of data for a tag */
	     offset,         /* offset to data, or data area for, a tag */
	     nextOffset,
	     tmpOffset,
	     tagDataOffset,
	     index_offset,   /* offset to character index array */
	     index_length;   /* length of character index array */

	fp = getFontmetricsFile(name);
	if (!fp) {
		return (struct tfm_type *) NULL;
	}
	fseek(fp, 0L, 0);  /* set ptr to file start */

	type_size[tagBYTE]       = 1;
	type_size[tagASCII]      = 1;
	type_size[tagSHORT]      = 2;
	type_size[tagLONG]       = 4;   /* byte sizes */
	type_size[tagRATIONAL]   = 8;
	type_size[tagSIGNEDBYTE] = 1;
	type_size[tagSIGNEDSHORT]= 2;
	type_size[tagSIGNEDLONG] = 4;

	tfm = (struct tfm_type *) malloc (sizeof(struct tfm_type));
	memset((char *) tfm, 0, sizeof(struct tfm_type));

	ptype = get_short(fp, TRUE);          /* get byte order flag */
	if (ptype == 0X4949) {
		intel_order = TRUE;	  /* set byte order to Intel */
	}
	else if (ptype == 0X4D4D) {
		intel_order = FALSE;
	}
	else {
		ok = FALSE;
		goto alldone;
	}

	/* read and discrad tfm version number */
	fgetc(fp);
	fgetc(fp);
	
	    /* get number  of typefaces in the tfm file, alloc mem */
	get_n_faces(tfm, fp, intel_order);
	tfm->face = (struct face_type *) malloc
		(tmp = tfm->n_faces * sizeof(struct face_type));
	memset((char *) tfm->face, 0, tmp);

	for(ix = 0; ix < tfm->n_faces; ix++) {
		dir_offset = get_long(fp, intel_order);
		fseek(fp, dir_offset, 0);   /* get to offset position */
		n_tags = get_short(fp, intel_order);
		for (iz = 0; iz < n_tags; iz++) {
			    /* read in info about each type tag */
			tag_value = get_short(fp, intel_order);
			data_type = get_short(fp, intel_order);
			block_size = get_long(fp, intel_order);
			length = (type_size[data_type] * block_size); /* #bytes */

			if (length > 4) {
				    /* data is offset from here */
				offset = get_long(fp, intel_order);
				tagDataOffset = ftell(fp);
				fseek(fp, offset, 0);
			}
			else
				offset = 0L;
			
			switch(tag_value) {
				
			case tagSUBFILE: /* must be present to ensure a valid tfm file */
				get_long(fp, intel_order); /* not used here */
				break;

			case tagCPYRIGHT:
				get_string(fp, jnk_str); /* not used here */
				if (length <= 4) {
				        /* data not offset; skip any unused bytes */
					for(tmp = block_size; tmp < 4; tmp++)	  
						fgetc(fp);
				}
				break;
				
			case tagCOMMENT:
				get_string(fp, jnk_str); /* not used here */
				if (length <= 4) {
					for (tmp = block_size; tmp < 4; tmp++)
						fgetc(fp);
				}
				break;

			case tagSYMBOLMAP:
				    /* this tag tells how many letterforms
					   are in the typeface */
				tfm->face[ix].general.n_chars = block_size;

				    /* allocate memory for the symbol set map and characters.
					   the length is the total number of bytes
					   (typesize[data_type] * n_chars) */
				tfm->face[ix].symbol.symbol_map = (int16 *) malloc
					(tmp = length * sizeof(int16));
				memset((char *) tfm->face[ix].symbol.symbol_map, 0, tmp);

				tfm->face[ix].character =
					(struct character_type *) malloc
						(tmp = tfm->face[ix].general.n_chars *
						 sizeof(struct character_type));
				memset((char *) tfm->face[ix].character,
					   0, tmp);
			
				for (tmp = 0; tmp < block_size; tmp++) {
					    /* read in MSL number for every character */
					tfm->face[ix].symbol.symbol_map[tmp] =
						get_short(fp, intel_order);
				}
				break;

			case tagSYMBOLSETDIR:
				numSymbolSets = block_size / 14;
				    /* we'll read thru all the symbol sets, but will
					   store only the one set (if found) whose name
					   matches the passed-in encoding name */
				found = FALSE;
				tfm->face[ix].general.n_symbol_sets = 1;
				    /* alloc space for the directory */
				tfm->face[ix].symbol.symbolset_directory = 
					(struct symbolset_directory_type *) malloc
						(sizeof (struct symbolset_directory_type));
				memset((char *) tfm->face[ix].symbol.symbolset_directory,
					   0, sizeof(struct symbolset_directory_type));

				for (iy = 0; iy < numSymbolSets; iy++) {
					    /* read offset to symbol set name */
					nextOffset = get_long(fp, intel_order);
					tmpOffset = ftell(fp);  /* save current position */
					fseek(fp, nextOffset, 0);
					get_string(fp, tmpStr);  /* read symset name */
					if (encoding && tmpStr && !strcmp(tmpStr, encoding)) {
						thisSet = TRUE;
						found = TRUE;
						strcpy(tfm->face[ix].symbol.symbolset_directory[0].name, tmpStr);
					}
					else {
						thisSet = FALSE;
					}
					fseek(fp, tmpOffset, 0);    /* restore position */
					    /* read offset to selection name */
					nextOffset = get_long(fp, intel_order);
					tmpOffset = ftell(fp);
					fseek(fp, nextOffset, 0);
					get_string(fp, tmpStr);     /* read selection name */
					if (thisSet) {
						strcpy(tfm->face[ix].symbol.symbolset_directory[0].selection_name, tmpStr);
					}
					fseek(fp, tmpOffset, 0);    /* restore position */
					    /* read offset to index array */
					index_offset = get_long(fp, intel_order);
					    /* read length of index array */
					index_length = get_short(fp, intel_order);
					tmpOffset = ftell(fp);       /* save position */
					if (thisSet) {
						    /* allocate for the symbolset index array */
						tfm->face[ix].symbol.
							symbolset_directory[0].index_len =
								index_length;
						tfm->face[ix].symbol.
							symbolset_directory[0].index =
								(short *) malloc (index_length *
												  type_size[tagSHORT]);
						memset((char *) tfm->face[ix].symbol.
							   symbolset_directory[0].index, 0,
							   index_length * type_size[tagSHORT]);
					}
					fseek(fp, index_offset, 0);    /* to array data */
					for (kx = 0; kx < index_length; kx++) {
						tshort = get_short(fp, intel_order);
						if (thisSet) {
							tfm->face[ix].symbol.
								symbolset_directory[0].index[kx] =
									tshort;
						}
					}
					fseek(fp, tmpOffset, 0);    /* restore position */
				}
				if (!found) {
					ok = FALSE;
					goto alldone;
				}
				break;

			case tagUNIQUEASSOCID:
				get_string(fp, jnk_str); /* not used here */
				if (length <= 4) {
					for (tmp = block_size; tmp < 4; tmp++)
						fgetc(fp);    /* skip unused bytes */
				}
				break;
			
			case tagPOINT:    /* point size */
				tfm->face[ix].metrics.point =
					get_double(fp, intel_order);
				break;

			case tagNOMINALPOINT:    /* nominal point size */
				tfm->face[ix].metrics.nominal_pointsize =
					get_double(fp, intel_order);
				break;

			case tagDESIGNUNITS:
				tfm->face[ix].metrics.design_units =
					get_double(fp, intel_order);
				break;
			
			case tagTYPESTRUCT:
				tfm->face[ix].metrics.type_struct = fgetc(fp);
				fgetc(fp);
				get_short(fp, intel_order);  /* skip unused */
				break;

			case tagSTROKEWT:
				tfm->face[ix].metrics.weight =
					fgetc(fp);
				fgetc(fp);
				get_short(fp, intel_order);  /* skip unused */
				break;

			case tagSPACING:
				tfm->face[ix].metrics.spacing =
					get_short(fp, intel_order);
				get_short(fp, intel_order);  /* skip unused */
				break;

			case tagSLANT:
				tfm->face[ix].metrics.slant =
					get_short(fp, intel_order);
				get_short(fp, intel_order);    /* skip unused */
				break;

			case tagAPPEARWIDTH:
				tfm->face[ix].metrics.appearance_width =
					fgetc(fp);
				fgetc(fp);                /* skip unused */
				get_short(fp, intel_order);
				break;

			case tagSERIFSTYLE:
				tfm->face[ix].metrics.serif_style = fgetc(fp);
				fgetc(fp);               /* skip unused */
				get_short(fp, intel_order);
				break;

			case tagTYPESTYLE:
				tfm->face[ix].metrics.type_style = fgetc(fp);
				fgetc(fp);
				get_short(fp, intel_order);  /* skip unused */
				break;

			case tagTYPEFACE:
				get_string(fp, jnk_str); /* not used here */
				if (length <= 4) {
					for (tmp = block_size; tmp < 4; tmp++)
						fgetc(fp);  /* skip unused */
				}
				break;

			case tagTFSOURCE:
				get_string(fp, jnk_str); /* not used here */
				if (length <= 4) {
					for (tmp = block_size; tmp < 4; tmp++)
						fgetc(fp);  /* skip unused */
				}
				break;

			case tagAVERAGEWD:
				tfm->face[ix].metrics.average_width =
					get_double(fp, intel_order);
				break;

			case tagMAXWIDTH:
				tfm->face[ix].metrics.maximum_width =
					get_short(fp, intel_order);
				get_short(fp, intel_order);   /* skip unused */
				break;

			case tagINTERWORDSP:
				tfm->face[ix].metrics.word_spacing = get_short(fp, intel_order);
				get_short(fp, intel_order);   /* skip unused */
				break;

			case tagRECLINESP:
				tfm->face[ix].metrics.line_spacing = get_short(fp, intel_order);
				get_short(fp, intel_order);  /* skip unused */
				break;

			case tagCAPHEIGHT:
				tfm->face[ix].metrics.capheight =
					get_short(fp, intel_order);
				get_short(fp, intel_order);  /* skip unused */
				break;

			case tagXHEIGHT:
				tfm->face[ix].metrics.x_height = get_short(fp, intel_order);
				get_short(fp, intel_order);  /* skip unused */
				break;

			case tagASCENT:
				tfm->face[ix].metrics.ascent =
					get_short(fp, intel_order);
				get_short(fp, intel_order);  /* skip unused */
				break;

			case tagDESCENT:
				tfm->face[ix].metrics.descent =
					get_short(fp, intel_order);
				get_short(fp, intel_order);  /* skip unused */
				break;

			case tagLOWERASCENT:
				tfm->face[ix].metrics.lowercase_ascent =
					get_short(fp, intel_order);
				get_short(fp, intel_order);  /* skip unused */
				break;

			case tagLOWERDESCENT:
				tfm->face[ix].metrics.lowercase_descent =
					get_short(fp, intel_order);
				get_short(fp, intel_order);  /* skip unused */
				break;

			case tagUNDERDEPTH:
				tfm->face[ix].metrics.underscore_descent =
					get_short(fp, intel_order);
				get_short(fp, intel_order);  /* skip unused */
				break;

			case tagUNDERTHICK:
				tfm->face[ix].metrics.underscore_thickness =
					get_short(fp, intel_order);
				get_short(fp, intel_order);  /* skip unused */
				break;

			case tagUPPERACCENT:
				tfm->face[ix].metrics.uppercase_accent_height =
					get_short(fp, intel_order);
				get_short(fp, intel_order);  /* skip unused */
				break;
				
			case tagLOWERACCENT:
				tfm->face[ix].metrics.lowercase_accent_height =
					get_short(fp, intel_order);
				get_short(fp, intel_order);  /* skip unused */
				break;

			case tagHORIZONTALESC:
				for (kx = 0; kx < block_size; kx++) {
					tfm->face[ix].character[kx].
						horizontal_escapement = get_short(fp, intel_order);
				}
				break;

			case tagVERTICALESC:
				for (kx = 0; kx < block_size; kx++) {
					tfm->face[ix].character[kx].
						vertical_escapement = get_short(fp, intel_order);
				}
				break;

			case tagLEFTEXTENT:
				for (kx = 0; kx < block_size; kx++) {
					tfm->face[ix].character[kx].
						left_extent = get_short(fp, intel_order);
				}
				break;

			case tagRIGHTEXTENT:
				for (kx = 0; kx < block_size; kx++) {
					tfm->face[ix].character[kx].
						right_extent = get_short(fp, intel_order);
				}
				break;

			case tagCHARASCENT:
				for (kx = 0; kx < block_size; kx++) {
					tfm->face[ix].character[kx].
						character_ascent = get_short(fp, intel_order);
				}
				break;

			case tagCHARDESCENT:
				for (kx = 0; kx < block_size; kx++) {
					tfm->face[ix].character[kx].
						character_descent = get_short(fp, intel_order);
				}
				break;

			case tagKERNPAIRS:
				tfm->face[ix].kerning.pairs = (struct pairs_type *)
					malloc(length);
				memset((char *) tfm->face[ix].kerning.pairs,
					   0, length);
				n_pairs = get_short(fp, intel_order);
				tfm->face[ix].kerning.n_pairs = n_pairs;
				for (kx = 0; kx < n_pairs; kx++) {
					tfm->face[ix].kerning.pairs[kx].ix1 =
						get_short(fp, intel_order);
					tfm->face[ix].kerning.pairs[kx].ix2 =
						get_short(fp, intel_order);
					tfm->face[ix].kerning.pairs[kx].val =
						get_short(fp, intel_order);
				}
				break;

			case tagSECTORKERN:
				n_chars = get_short(fp, intel_order);
				tfm->face[ix].kerning.n_chars = n_chars;
				tfm->face[ix].kerning.sector =
					(struct sector_type *) malloc
						(tmp = n_chars *
						 sizeof(struct sector_type));
				memset((char *)
					   tfm->face[ix].kerning.sector,
					   0, tmp);
				n_sectors = get_short(fp, intel_order);
				tfm->face[ix].kerning.n_sectors = n_sectors;

				for (kx = 0; kx < n_chars; kx++) {

					tfm->face[ix].kerning.sector[kx].ix =
						get_short(fp, intel_order);

					    /* alloc and fill left and right sector data
						   for this kernchar */
					tfm->face[ix].kerning.sector[kx].left =
						(short *) malloc
							(tmp = n_sectors * type_size[tagSHORT]);
					memset((char *)
						   tfm->face[ix].kerning.
						   sector[kx].left,
						   0, tmp);
					tfm->face[ix].kerning.sector[kx].right =
						(short *) malloc
							(tmp = n_sectors * type_size[tagSHORT]);
					memset((char *)
						   tfm->face[ix].kerning.
						   sector[kx].right,
						   0, tmp);
					for (tmp = 0; tmp < n_sectors; tmp++) {
						tfm->face[ix].kerning.sector[kx].
							left[tmp] = get_short(fp, intel_order);
					}
					for (tmp = 0; tmp < n_sectors; tmp++) {
						tfm->face[ix].kerning.sector[kx].
							right[tmp] = get_short(fp, intel_order);
					}
				}
				break;

			case tagTRACKKERN:
				n_tracks = get_short(fp, intel_order);
				tfm->face[ix].kerning.n_tracks = n_tracks;
				tfm->face[ix].kerning.track = (struct track_type *)
					malloc (length);
				memset((char *) tfm->face[ix].kerning.track,
					   0, length);
				for (kx = 0; kx < n_tracks; kx++) {
					tfm->face[ix].kerning.track[kx].val =
						get_short(fp, intel_order);
					tfm->face[ix].kerning.track[kx].max_pointsize =
						get_short(fp, intel_order);
					tfm->face[ix].kerning.track[kx].min_pointsize =
						get_short(fp, intel_order);
					tfm->face[ix].kerning.track[kx].max_kern =
						get_short(fp, intel_order);
					tfm->face[ix].kerning.track[kx].min_kern =
						get_short(fp, intel_order);
				}
				break;
			
			case tagTFSELECTSTR:
				get_string(fp, jnk_str); /* not used here */
				if (length <= 4) {
					for (tmp = block_size; tmp < 4; tmp++)
						fgetc(fp);    /* skip unused */
				}
				break;

			default:
				get_long(fp, intel_order);  /* skip past unknown tag */
				break;

			}  /* end switch */

			if (offset !=0L) {
				    /* if the data was offset, reset the position */
				fseek(fp, tagDataOffset, 0);
			}
		}
	}
 alldone:
	fclose(fp);
	if (ok) {
		return tfm;
	}
	else {
		free_tfm(tfm);
		return (struct tfm_type *) NULL;
	}
}

    /* : tfmToFontMetrics - initialize a fontMetrics instance
       from a tfm_type instance */

static fontMetrics *
tfmToFontMetrics(tfm)
struct tfm_type *tfm;  /* in: font metrics in tfm form */
{
	fontMetrics *fm;
	struct symbolset_directory_type *sd;
	struct face_type *tf;
	struct character_type *tfmChar;
	XCharStruct *xChar;
	float dtmp, dfactor;
	int ok = TRUE;
	int minLbearing, maxLbearing, minRbearing, maxRbearing;
	int minWidth, maxWidth, minAscent, maxAscent;
	int minDescent, maxDescent;
	int lbear, rbear, width, ascent, descent;
	int ix, tmp, row, col, rowWid, symIndex;

	if (!tfm)
		return ((fontMetrics *) NULL);

	fm = (fontMetrics *) malloc (sizeof(fontMetrics));
	memset((char *) fm, 0, sizeof(fontMetrics));

	/* we're looking only at the 0th typeface in the tfm */
	tf = &tfm->face[0];

	/* we're looking only at the 0th symbol set directory */
	sd = &tf->symbol.symbolset_directory[0];

	/* set dfactor so that a tfm number times the factor yields milliems */
	if (tf->metrics.design_units == 0.0) {
		ok = FALSE;
		goto alldone;
	}
	dfactor = 1000.0 / tf->metrics.design_units;

	fm->encodingName = malloc(strlen(sd->name) + 1);
	strcpy(fm->encodingName, sd->name);

	fm->PCLSymbolSetID = malloc(strlen(sd->selection_name) + 1);
	strcpy(fm->PCLSymbolSetID, sd->selection_name);

	fm->underlinePosition = qRound
		((float) tf->metrics.underscore_descent * dfactor);
	fm->underlineThickness = qRound
		((float) tf->metrics.underscore_thickness * dfactor);
	fm->xHeight = qRound
		((float) tf->metrics.x_height * dfactor);
	fm->capHeight = qRound
		((float) tf->metrics.capheight * dfactor);
	fm->averageWidth = qRound
		((float) tf->metrics.average_width * dfactor);

	    /* fill out the XFontStruct portion of the fontMetrics.
		   (would be nice to fill out members like direction,
		   default_char, all_chars_exist, but just leave zeroed) */
			   
	fm->xfont = (XFontStruct *) malloc (sizeof(XFontStruct));
	memset((char *) fm->xfont, 0, sizeof(XFontStruct));

	    /* take the character range from the symbol set directory's
		   index_len member; not much else to go on */
	fm->xfont->min_byte1 = 0;
	tmp = sd->index_len - 1;
	fm->xfont->max_byte1 = (unsigned) ((tmp >> 8) & 0xff);
	fm->xfont->min_char_or_byte2 = 0;
	fm->xfont->max_char_or_byte2 = (unsigned) (tmp & 0xff);
	
	fm->xfont->ascent = qRound
		((float) tf->metrics.ascent * dfactor);
	fm->xfont->descent = qRound
		((float) -tf->metrics.descent * dfactor);

	    /* alloc per-char metrics array */
	fm->xfont->per_char = (XCharStruct *) malloc
		(tmp = sd->index_len * sizeof(XCharStruct));
	memset((char *) fm->xfont->per_char, 0, tmp);
	    /* as the per-char array is filled, remember min/max
		   character metrics; these go on min_bounds, max_bounds */
	minLbearing = minRbearing = minWidth =
		minAscent = minDescent = MAXLRU;     /* bigger than any */
	maxLbearing = maxRbearing = maxWidth =
		maxAscent = maxDescent = -MAXLRU;

	for (ix = 0; ix < sd->index_len; ix++) {
		symIndex = (int) sd->index[ix];
		if (symIndex >= 0) {
			    /* ix is a valid codepoint in the encoding;
				 * get the address of its charmetrics struct on tfm
				 */
			tfmChar = &tf->character[symIndex];
			    /* get the corresponding XCharStruct address on
				 * fm->xfont->per_char
				 */
			row = (ix >> 8) & 0xff;
			col = ix & 0xff;
			rowWid = fm->xfont->max_char_or_byte2 -
				fm->xfont->min_char_or_byte2 + 1;
			tmp = ((row - fm->xfont->min_byte1) * rowWid) +
				col - fm->xfont->min_char_or_byte2;
			xChar = &fm->xfont->per_char[tmp];
			    /* copy metrics; test for max/min */
			        /* get tfm metrics in milliem units */
			lbear = qRound ((float) tfmChar->left_extent * dfactor);
			rbear = qRound ((float) tfmChar->right_extent * dfactor);
			width = qRound
				((float) tfmChar->horizontal_escapement * dfactor);
			ascent = qRound
				((float) tfmChar->character_ascent * dfactor);
			descent = qRound
				((float) tfmChar->character_descent * dfactor);
			       /* test max/min */
			if (lbear < minLbearing)
				minLbearing = lbear;
			if (lbear > maxLbearing)
				maxLbearing = lbear;
			if (rbear < minRbearing)
				minRbearing = rbear;
			if (rbear > maxRbearing)
				maxRbearing = rbear;
			if (width < minWidth)
				minWidth = width;
			if (width > maxWidth)
				maxWidth = width;
			if (ascent < minAscent)
				minAscent = ascent;
			if (ascent > maxAscent)
				maxAscent = ascent;
			if (-descent < minDescent)
				minDescent = -descent;
			if (-descent > maxDescent)
				maxDescent = -descent;

			    /* copy onto fm->xfont->per_char element */
			xChar->lbearing = lbear;
			xChar->rbearing = rbear;
			xChar->width = width;
			xChar->ascent = ascent;
			xChar->descent = -descent;
		}
	}
	
	    /* fill out max_bounds, min_bounds */
	fm->xfont->min_bounds.lbearing = minLbearing;
	fm->xfont->min_bounds.rbearing = minRbearing;
	fm->xfont->min_bounds.width = minWidth;
	fm->xfont->min_bounds.ascent = minAscent;
	fm->xfont->min_bounds.descent = minDescent;
	
	fm->xfont->max_bounds.lbearing = maxLbearing;
	fm->xfont->max_bounds.rbearing = maxRbearing;
	fm->xfont->max_bounds.width = maxWidth;
	fm->xfont->max_bounds.ascent = maxAscent;
	fm->xfont->max_bounds.descent = maxDescent;

	if (fm->xfont->min_byte1 == 0) {
		    /* hack alert: what we're doing in this block
			 * is invalid unless the encoding is compatible
			 * with iso latin or unicode.  There are two
			 * hacks:  hack-1) test whether codepoint 32 (in
			 * compatible encodings this is the space char)
			 * has width, and if not supply width as:
			 * tfm->typeface[0].metrics.word_spacing
			 * and hack-2) test whether codepoint 160 (in
			 * compatible encodings this is the nonbreakable space)
			 * has width, and if not supply width from codepoint 32
			 */
		if (fm->xfont->max_char_or_byte2 -
			fm->xfont->min_char_or_byte2 > 32) {
			tmp = 32 - fm->xfont->min_char_or_byte2;
			xChar = &fm->xfont->per_char[tmp];
			if (!xChar->width) {
				tmp = tfm->face[0].metrics.word_spacing;
				xChar->width = qRound((float) tmp * dfactor);
			}
		}
		
		if (fm->xfont->max_char_or_byte2 -
			fm->xfont->min_char_or_byte2 > 160) {
			tmp = 160 - fm->xfont->min_char_or_byte2;
			xChar = &fm->xfont->per_char[tmp];
			if (!xChar->width) {
				tmp = 32 - fm->xfont->min_char_or_byte2;
				*xChar = fm->xfont->per_char[tmp];
			}
		}
	}
	
alldone:
	if (ok) {
		return fm;
	}
	else {
		freeTargetMetrics(fm);
		return ((fontMetrics *) NULL);
	}
}

/**************************************************************************\
 * getTfmMetrics: read TaggedFontMetrics file, fill out fontMetrics       *
\**************************************************************************/
int
getTfmMetrics(name, mp, sizep, encoding)
char  *name;
char **mp;       /* data to return as a fontMetrics struct */
int   *sizep;
char  *encoding; /* name of encoding, e.g., "ECMA-94 Latin 1" */
{
	struct tfm_type *tfm;
	fontMetrics *fm;
	int success;

    /* get contents of named tfm file onto tfm struct */
	if ( !(tfm = readTfmFile(name, encoding)) )
	{
		*mp = (char *) NULL;
		*sizep = 0;
		success = FALSE;
	}
	else /* now copy from tfm struct onto fontMetrics struct */
	{
		if ( !(fm = tfmToFontMetrics(tfm, sizep)) )
		{
			*mp = (char *) NULL;
			*sizep = 0;
			success = FALSE;
		}
		else
		{
			*mp = (char *) fm;
			*sizep = fontMetricsSize(fm);
			success = TRUE;
		}
		free_tfm(tfm);
	}
	return(success);
}
