/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axtrutyp.c
   Applix ASC version: 2
   Date: Mon Nov 16 13:53:47 1998
*/
#include <stdio.h>
#include <aos.h>
#include <axtrutyp.h>

#ifndef SEEK_SET
#define	SEEK_SET	0
#endif
#ifndef SEEK_END
#define	SEEK_END	2
#endif

int axttFontEncoding();

static int absSeek();
static int byteSizeOfFile();
static int readCmapTable();
static int readNamingTable();
static int readNameRecord();
static int readNameString();
static int readOffsetTable();
static int readTableDirectory();
static int readTTencodingNames();
static unsigned short read16();
static unsigned long read32();

/* : axttFontEncoding - return integer code representing the
 * character encoding of the passed trutype font file
 */
int                                   /* return 0, or <0 error code */
axttFontEncoding(ttname, encoding_int)
char *ttname;                         /* in: full pathname of tt font file */
int *encoding_int;                    /* out: e.g., AXENC_LATIN1 */
{
#if AX_X11
	int ix, err;
	FILE *ttf;
	offset_table offTab;
	table_directory tabDir;
	axfont_name_data nd;
	
	if (ttname) {
		ttf = fopen(ttname, READ_BINARY_STR);
		if (ttf) {
			if ((err = readOffsetTable(ttf, &offTab)) < 0) {
				fclose (ttf);
				return (err);
			}
			memset((void *) &nd, 0, sizeof(axfont_name_data));
			for (ix = 0; ix < (int) offTab.numTables; ix++) {
				if ((err = readTableDirectory(ttf, &tabDir)) < 0) {
					break;
				}
				if (tabDir.tag == (unsigned long) TAG_CMAP) {
					if ((err = readTTencodingNames(ttf, &tabDir, &nd)) < 0) {
						break;
					}
					*encoding_int = nd.internalEncoding;
					break;
				}
			}
			fclose(ttf);
		}
		else {
			err = AXTT_OPEN_IN;
		}
	}
	else {
		err = AXTT_OPEN_IN;
	}
	return (err);
#else
	return (0);
#endif
}

#if AX_X11           /* static funcs in here are X11-only { */

static int           /* return 0, or int error code */
absSeek(fp, offset)
FILE *fp;		     /* in */
long offset;		 /* in: do an absolute seek to this offset */
{
	if (fseek(fp, offset, SEEK_SET)) {
		return (AXTT_ABS_SEEK);
	}
	return (0);
}

static int           /* return byte size of file, or <0 int error code */
byteSizeOfFile(fp)
FILE *fp;
{
	long fsize;
	
	if (fseek(fp, 0, SEEK_END)) {
		return (AXTT_END_SEEK);
	}
	fsize = ftell(fp);
	absSeek(fp, 0);
	return ((int) fsize);
}

/* read16, read32 assume the file data arrives in moto (big end) order.
 */
static unsigned short      /* return 16bit data word */
read16(fp, errint)
FILE *fp;                  /* in */
int *errint;               /* out: error if < 0 */
{
	unsigned char clow, chigh;
	
	if (!fread((char *) &chigh, 1, sizeof(unsigned char), fp)) {
		*errint = AXTT_READ_IN;
		return (0);
	}
	if (!fread((char *) &clow, 1, sizeof(unsigned char), fp)) {
		*errint = AXTT_READ_IN;
		return (0);
	}
	*errint = 0;
	return ((unsigned short) (clow | ((chigh << 8) & 0xff00)));
}

static unsigned long       /* return 32bits of data in an unsigned long */
read32(fp, errint)
FILE *fp;                  /* in */
int *errint;               /* out: error if < 0 */
{
	unsigned short slow, shigh;
	
	shigh = read16(fp, errint);
	if (*errint < 0) {
		return (0);
	}
	slow = read16(fp, errint);
	if (*errint < 0) {
		return (0);
	}
	return ((unsigned long) (slow | ((shigh << 16) & 0xffff0000)));
}

static int                  /* return 0, or <0 error code */
readCmapDirTable(fp, table)
FILE *fp;
cmap_dir_table *table;	    /* out: init caller's memory */
{
	int err = 0;
	
	table->table_version = read16(fp, &err);
	if (err >= 0) {
		table->numCmaps = read16(fp, &err);
	}
	return (err);
}

static int               /* return 0, or <0 error code */
readCmapTable(fp, table)
FILE *fp;
cmap_table *table;	     /* out: init caller's memory */
{
	int err = 0;
	
	table->platform_id = read16(fp, &err);
	if (err >= 0) {
		table->encoding_id = read16(fp, &err);
		if (err >= 0) {
			table->cmap_offset = read32(fp, &err);
		}
	}
	return (err);
}

static int                 /* return 0, or <0 error code */
readNamingTable(fp, table)
FILE *fp;
naming_table *table;	   /* out: init caller's memory */
{
	int err = 0;
	
	table->format_selector = read16(fp, &err);
	if (err >= 0) {
		table->numNameRecords = read16(fp, &err);
		if (err >= 0) {
			table->stringOffset = read16(fp, &err);
		}
	}
	return (err);
}

static int                 /* return 0, or <0 error code */
readNameRecord(fp, nr)
FILE *fp;
name_record *nr;	       /* out: init caller's memory */
{
	int err = 0;
	
	nr->platform_id = read16(fp, &err);
	if (err < 0) {
		return (err);
	}
	nr->encoding_id = read16(fp, &err);
	if (err < 0) {
		return (err);
	}
	nr->language_id = read16(fp, &err);
	if (err < 0) {
		return (err);
	}
	nr->name_id = read16(fp, &err);
	if (err < 0) {
		return (err);
	}
	nr->string_length = read16(fp, &err);
	if (err < 0) {
		return (err);
	}
	nr->string_offset = read16(fp, &err);
	
	return (err);
}

static int                       /* return 0, or <0 error return */
readNameString(fp, nr, str)
FILE *fp;		                 /* in */
name_record *nr;	             /* in */
char *str;		                 /* out: init caller's memory */
{
	int nbytes = (int) nr->string_length;
	int err = 0;
	int wideEncoding = FALSE;
	
	if (nbytes <= 0) {
		return (err);
	}
	
	if (!fread(str, nbytes, 1, fp)) {
		return (AXTT_READ_IN);
	}
	if ((nr->platform_id == PLATFORM_MICROSOFT) &&
		((enum microsoft_encoding_id) nr->encoding_id) == UGL) {
		
		wideEncoding = TRUE;	/* UCS-16 */
	}
	else if (!nr->platform_id) {
		wideEncoding = TRUE;	/* "Apple Unicode" */
	}

	if (wideEncoding) {
		int ix, n_dbytes, all_hibytes_zero;
		
		n_dbytes = ((nbytes >> 1) << 1);
		for (ix = 0, all_hibytes_zero = TRUE; ix < n_dbytes; ix++) {
			if (!(ix % 2) && str[ix]) {
				all_hibytes_zero = FALSE;
			}
		}
		if (all_hibytes_zero) {
			int numUCS;
			char *ocp;
			unsigned short *longcopy, *icp;
			
			/* copy the string from 16bit to 8bit.
			 */
			numUCS = nbytes >> 1;
			longcopy = (unsigned short *) malloc
				((numUCS + 1) * sizeof(unsigned short));
			memcpy((void *) longcopy, (void *) str, nbytes);
			longcopy[numUCS] = 0;
			for (ix = 0; ix < numUCS; ix++) {
				str[ix] = (char) (longcopy[ix] & 0xff);
			}
			str[numUCS] = '\0';
			free(longcopy);
		}
	}
}

static int                      /* return 0, or <0 error code */
readOffsetTable(fp, table)
FILE *fp;
offset_table *table;	        /* out: init caller's memory */
{
	int err = 0;
	
	table->sfnt_version = (int) read32(fp, &err);
	if (err < 0) {
		return (err);
	}
	table->numTables = read16(fp, &err);
	if (err < 0) {
		return (err);
	}
	table->searchRange = read16(fp, &err);
	if (err < 0) {
		return (err);
	}
	table->entrySelector = read16(fp, &err);
	if (err < 0) {
		return (err);
	}
	table->rangeShift = read16(fp, &err);
	
	return (err);
}

static int                      /* return 0, or <0 error code */
readTableDirectory(fp, table)
FILE *fp;
table_directory *table;	        /* out: init caller's memory */
{
	int err = 0;
	
	table->tag = read32(fp, &err);
	if (err < 0) {
		return (err);
	}
	table->checkSum = read32(fp, &err);
	if (err < 0) {
		return (err);
	}
	table->offset = read32(fp, &err);
	if (err < 0) {
		return (err);
	}
	table->length = read32(fp, &err);
	
	return (err);
}

static int                                 /* return 0, or <0 error code */
readTTencodingNames(ttf, tabDir, nameData)
FILE *ttf;			                       /* in */
table_directory *tabDir;	               /* in */
axfont_name_data *nameData;	               /* out: write caller's memory */
{
	int ix, tmpEnc, latin1, err;
	int accumEncoding;
	long current_offset;
	cmap_dir_table cdir;
	cmap_table cmap;
	enum platform_id platID;
	enum mac_encoding_id tmac, macEncoding;
	enum iso_encoding_id tiso, isoEncoding;
	enum microsoft_encoding_id tmsoft, msoftEncoding;

	current_offset = ftell(ttf);
	err = 0;
	
	if ((err = absSeek(ttf, tabDir->offset)) < 0) {
		return (err);
	}
	if ((err = readCmapDirTable(ttf, &cdir)) < 0) {
		return (err);
	}
	
	macEncoding = (enum mac_encoding_id) -1;	 /* out-of-band values */
	isoEncoding = (enum iso_encoding_id) -1;
	msoftEncoding = (enum microsoft_encoding_id) -1;

	/* loop thru the cmaps, and save first valid endcoding numbers found.
	 */
	for (ix = 0; ix < (int) cdir.numCmaps; ix++) {
		if ((err = readCmapTable(ttf, &cmap)) < 0) {
			return (err);
		}
		platID = cmap.platform_id;
		tmpEnc = cmap.encoding_id;

		tmac = (enum mac_encoding_id) -1;	  /* out-of-band values */
		tiso = (enum iso_encoding_id) -1;
		tmsoft = (enum microsoft_encoding_id) -1;
		
		switch (platID) {
			
		case PLATFORM_MACINTOSH:
			tmac = (enum mac_encoding_id) tmpEnc;
			break;
			
		case PLATFORM_ISO:
			tiso = (enum iso_encoding_id) tmpEnc;
			break;
			
		case PLATFORM_MICROSOFT:
			tmsoft = (enum microsoft_encoding_id) tmpEnc;
			break;
			
		}
		
		if (((int) macEncoding == -1) && ((int) tmac >= 0)) {
			macEncoding = tmac;
		}
		if (((int) isoEncoding == -1) && ((int) tiso >= 0)) {
			isoEncoding = tiso;
		}
		if (((int) msoftEncoding == -1) && ((int) tmsoft >= 0)) {
			msoftEncoding = tmsoft;
		}
	}
	
	/* Assign to nameData our extremely coarse guess as to encoding;
	 * we're basically looking for an excuse to say latin1.
	 */
	latin1 = FALSE;
	if ((int) msoftEncoding >= 0) {
		if (msoftEncoding == UGL) {
			latin1 = TRUE;	    /* 'page 0' of UCS16 is latin1 */
		}
	}
	if ((int) isoEncoding >= 0) {
		latin1 = TRUE;		/* all defined vals consistent with latin1 */
	}
	if (latin1) {
		nameData->internalEncoding = AXENC_LATIN1;
	}
	else {
		nameData->internalEncoding = AXENC_UNSPECIFIED;
	}
	
	err = absSeek(ttf, current_offset);
	return (err);
}

#endif          /* if AX_X11 } */
