/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axtrutyp.h
   Applix ASC version: 1
   Date: Mon Nov 16 13:53:47 1998
*/
#ifndef AXTRUTYP_H
#define AXTRUTYP_H


/* platform id's
 */
enum platform_id {
	PLATFORM_MACINTOSH = 1,
	PLATFORM_ISO = 2,
	PLATFORM_MICROSOFT = 3,
};

/* platform-specific encoding id's -- MAC
 */
enum mac_encoding_id {
	ROMAN = 0,
	JAPANESE = 1,
	CHINESE = 2,
	KOREAN = 3,
	ARABIC = 4,
	HEBREW = 5,
	GREEK = 6,
	RUSSIAN = 7,
	RSYMBOL = 8,
	DEVANAGARI = 9,
	GURMUKHI = 10,
	GUJARATI = 11,
	ORIYA = 12,
	BENGALI = 13,
	TAMIL = 14,
	TELUGU = 15,
	KANNADA = 16,
	MALAYALAM = 17,
	SINHALESE = 18,
	BURMESE = 19,
	KHMER = 20,
	THAI = 21,
	LAOTIAN = 22,
	GEORGIAN = 23,
	ARMENIAN = 24,
	MALDIVIAN = 25,
	TIBETAN = 26,
	MONGOLIAN = 27,
	GEEZ = 28,
	SLAVIC = 29,
	VIETNAMESE = 30,
	SINDHI = 31,
	UNINTERPRETED = 32,
};

/* platform-specific encoding id's -- ISO
 */
enum iso_encoding_id {
	ASCII_7BIT = 0,
	ISO_646 = 1,
	ISO_8859_1 = 2,
};

/* platform-specific encoding id's -- MICROSOFT
 */
enum microsoft_encoding_id {
	UNDEFINED,
	UGL,
};

/* font name string id's
 */
enum font_namestring_id {
	COPYRIGHT = 0,
	FONT_FAMILY = 1,
	FONT_SUBFAMILY = 2,
	UNIQUE_ID = 3,
	FULL_FONT_NAME = 4,
	VERSION = 5,
	POSTSCRIPT_NAME = 6,
	TRADEMARK = 7,
};

/* integer error codes; OK always returns 0; error codes always < 0
 */
#define AXTT_ABS_SEEK (-2)
#define AXTT_END_SEEK (-3)
#define AXTT_READ_IN  (-4)
#define AXTT_OPEN_IN  (-5)

#define NUM_TT_REQ_TABLES 10
#define TAG_CMAP 0x636d6170    /* "cmap" */
#define TAG_NAME 0x6e616d65    /* "name" */

/* not real slick regarding all the encodings listed above;
 * we tend to see a truetype file either as latin1 or not
 */
#define AXENC_LATIN1      1
#define AXENC_UNSPECIFIED 2

/* this type is for holding accumulating the name info we want.
 */
typedef struct axfont_name_data_t {
	char *unique_id;
	char *copyright;
	char *familyName;
	char *subName;
	char *postscriptName;
	char *xlfd;
	char *ttfileName;
	int ttfileSize;
	int internalEncoding;
} axfont_name_data;

/* typedefs for those portions of the truetype file
 * that we look at.
 */
typedef struct offset_table_t {          /* offset table starts at byte 0 */
	unsigned int    sfnt_version;
	unsigned short  numTables;           /* #table_directory_t's to follow */
	unsigned short  searchRange;
	unsigned short  entrySelector;
	unsigned short  rangeShift;
} offset_table;

typedef struct table_directory_t {       /* these follow the offset table */
	unsigned long tag;                   /* 4-byte identifier of this table */
	unsigned long checkSum;
	unsigned long offset;                /* file offset of this table */
	unsigned long length;                /* length of this table */
} table_directory;

typedef struct cmap_dir_table_t {
	unsigned short table_version;
	unsigned short numCmaps;
} cmap_dir_table;

typedef struct cmap_table_t {
	unsigned short platform_id;
	unsigned short encoding_id;
	/* cmap_offset stores the offset (measured from the start of this table)
	 * to the cmap subtable for this encoding_id.
	 */
	unsigned long  cmap_offset;
} cmap_table;

typedef struct namimg_table_t {
	unsigned short format_selector;      /* ==0 */
	unsigned short numNameRecords;       /* #name_record_t's to follow */
	/* stringOffset stores file offset (measured from start of this table)
	 * to this table's string data storage
	 */
	unsigned short stringOffset;
} naming_table;

typedef struct name_record_t {
	unsigned short platform_id;
	unsigned short encoding_id;
	unsigned short language_id;
	unsigned short name_id;
	/* string_length stores the length in bytes of the string data for
	 * this name record
	 */
	unsigned short string_length;
	/* string_offset stores the offset (mesured from the start of this table's
	 * string data area) of the string data for this name record.
	 */
	unsigned short string_offset;
} name_record;


#endif /* AXTRUTYP_H */
