/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axtxint.h
   Applix ASC version: 70
   Date: Wed Jan 20 21:12:40 1999
*/
#ifndef _AXTXINT_H
#define _AXTXINT_H


/* stub defines */
#ifndef TMA_NO
#define TMA_NO 0
#endif

#ifndef BLACK
#define BLACK 1
#endif

#ifndef WHITE
#define WHITE 0
#endif

#ifndef NULL
#define NULL 0
#endif

#include <basicwin.h>    /* need some Xtypedefs */

#ifndef _ATM_H
#include <atm.h> /* need poly_type */
#endif

#define MAXLFD  257                /* max length of XLFD plus null */
#define DEFAULT_RASTER_TOLERANCE 5 /* screen pixels of "fit" error */

    /* reasons why the 'fullDisp' member of eState might be nonzero; bitwise
     */
#define FORMAT_OVERLAP   0x1       /* lines format w/ vertical overlap */
#define COMPLEX_XFORM    0x2       /* dissimilar transforms in text */

    /* this type has one member for each member of fontTabEnt
     */
typedef struct font_tabent_use {
    unsigned int xNamesKnown           : 1;
    unsigned int scalable              : 1;
    unsigned int fixedWidth            : 1;
    unsigned int isWindowsFont         : 1;
    unsigned int isPrinterFont         : 1;
	unsigned int gallium               : 1;
    unsigned int globalName            : 1;
    unsigned int familyName            : 1;
    unsigned int aliasFor              : 1;
    unsigned int encodingWidth         : 1;
    unsigned int rasterTolerance       : 1;
    unsigned int toleranceLowCutoff    : 1;
    unsigned int toleranceHighCutoff   : 1;
    unsigned int weight                : 1;
    unsigned int slant                 : 1;
    unsigned int numPointsizes         : 1;
    unsigned int pointsizes            : 1;
    unsigned int numTargetMetricsNames : 1;
    unsigned int targetMetricsNames    : 1;
    unsigned int xNameCount            : 1;
    unsigned int xNames                : 1;
    unsigned int scalableName          : 1;
    unsigned int PostScriptPrintName   : 1;
    unsigned int PCLPrintName          : 1;
    unsigned int PostScriptAlternate   : 1;
    unsigned int PCLAlternate          : 1;
    unsigned int WindowsAlternate      : 1;
    unsigned int XAlternate            : 1;
    unsigned int largeXAlternate       : 1;
    unsigned int Type1FontFileName     : 1;
    unsigned int encodingFileName      : 1;
    unsigned int encodingName          : 1;
    unsigned int internalEncodingName  : 1;
    unsigned int sizeEM                : 1;
    unsigned int sans                  : 1;
    unsigned int avgWidth              : 1;
    unsigned int charSet               : 1;
    unsigned int pitchAndFamily        : 1;
#if BUILD_KANJI
    unsigned int LIPSPrintName         : 1;
#endif
} fontTabEntUse;

    /* the fontTabEnt holds the data from a FontRecord in the
     * fontMap.dir file.  Some additional info is kept here also,
     * as more is learned about the screenfont resources available
     * during a session
     */
typedef struct font_tabent {
    unsigned int xNamesKnown : 1; /* whether font info lookup done */
    unsigned int scalable    : 1; /* whether scalable available */
    unsigned int fixedWidth  : 1;
    unsigned int isWindowsFont : 1;  /* a windows truetype font */
    unsigned int isPrinterFont : 1;
	unsigned int gallium : 1;     /* accessible via Gallium font server */
    char *globalName;
    char *familyName;             /* this is the menu name */
    char *aliasFor;               /* familyName is an alias for this name */
    int  encodingWidth;           /* 8, 16, .. */
    int  rasterTolerance;
    int  toleranceLowCutoff;      /* tolerate fit error above this ptsize */
    int  toleranceHighCutoff;
    int  weight;                  /* bold */
    int  slant;                   /* ital */
    int  numPointsizes;
    int  *pointsizes;             /* for restriction on pointsize */
    int  numTargetMetricsNames;
    char **targetMetricsNames;    /* .afm or .tfm file name(s) */
    int  xNameCount;
    char **xNames;                /* xlfd's of X fonts for this ent */
    char *scalableName;           /* points to xName of scalable if present */
    char *PostScriptPrintName;    /* e.g., "Times-BoldItalic" */
    char *PCLPrintName;           /* tail of pcl selection sequence */
    char *PostScriptAlternate;    /* substitute-display font for PS */
    char *PCLAlternate;           /* substitute-display font for PCL */
    char *WindowsAlternate;       /* substitute-display font for Windows */
    char *XAlternate;             /* substitute-display font for X */
    char *largeXAlternate;        /* like XAlternate, but for big sizes only */
    char *Type1FontFileName;      /* file name, for downloading */
    char *encodingFileName;       /* .enc filename for app-visible encoding */
    char *encodingName;           /* e.g., "ECMA-94 Latin 1", for .tfm reads */
    char *internalEncodingName;   /* .enc filename for encoding inside font */
    int sizeEM;               /* font design unit for scalable numbers */
    int avgWidth;
    unsigned char charSet;    /* character set, used on NT only for now. */
    unsigned char pitchAndFamily;  /* NT only for now; see doc for LOGFONT */
    fontTabEntUse use;
   int  sans;                    /* sans serif */
#if BUILD_KANJI
   char *LIPSPrintName;  /* e.g., "Mincho-Medium" */
#endif
} fontTabEnt;

typedef struct family_member {
    char *propName;
    char *fontName;
} familyMember;

typedef struct font_family {
    int familyNumber;             /* records order in fontmap.dir */
    int numMembers;
    char *aliasFor;
    familyMember **members;
#if AXMS
    unsigned int isWindowsFont : 1;
    unsigned int isPrinterFont : 1;
#endif
} fontFamilyRecord;

typedef struct font_metrics {
    XFontStruct *xfont;
    char *encodingName; /* for example, "ISOLatin1Encoding" */
    char *PCLSymbolSetID;
    int underlinePosition;
    int underlineThickness;
    int xHeight;
    int capHeight;
    int averageWidth;
    int resolution;               /* dpi, (or 0 for milliems) */
    unsigned int noBearings : 1;  /* no char-level raster bboxes */
} fontMetrics;

typedef struct char_metrics {
    unsigned int charCode;
    int yAdvance;
    XCharStruct xchar;
} charMetrics;

typedef struct font_context {
        /* these pointers hold references to cache objects.
         * don't assume they're up to date; don't free them, either
         */
    fontTabEnt *font;
    fontMetrics *screen;
    fontMetrics *target;
} fontContext;

typedef struct {
    unsigned int unity : 1;
    int angle;
    int xscale;
    int yscale;
    int xshear;
} axformTerms;

typedef struct {
        /* postcript-like coord transformation matrix elements
         */
    float ca;
    float cb;
    float cc;
    float cd;
    int ctx;
    int cty; 
} axctm;

typedef struct {
    unsigned int known :1;
    unsigned int unity :1;
    unsigned int rotated : 1;
        /* ctm's for both directions
         */
    axctm forward;
    axctm reverse;
} axform;

typedef struct poly_t {
    unsigned int rectangular   :1;
    int numPaths;
    poly_type *paths;
} poly;

typedef struct axtTabInfo {
    int defTabSize;
    int *tabStops;
    int n_tabs;
} axtTabInfo;

typedef struct whole_field {
    unsigned int widthBounder :1;    /* simple reFit formatting */
    int nChars;
    int formattedWidth;              /* used with widthBounder formatting */
    p32 pos;        /* setpos of upstream-most character */
    p32 fixedPos;   /* point to remain stationary during edit */
    r32 bbox;       /* ink bounder */
    poly bounder;   /* format text to stay inside this */
    int width;      /* optional width for use with widthBounder formatting */
    int bounderMiny;
    int bounderMaxy;
    int lineHeight; /* vertical distance between baselines */
    int leftOffset; /* these are tolerances for 'staying inside' bounder */
    int rightOffset;
    int topOffset;
    int botOffset;
    int alignment;         /* left, right, center, justified, .. */
    int verticalAlignment; /* top, bottom, center */
    int xZoom;      /* in percent units (unzoomed is 100%) */
    int yZoom;
    axtTabInfo tabInfo;
    axformTerms transform;
    int (*getIntercepts)(); /* return intercepts for a given y pos */
    int (*lineFormatter)(); /* lay out individual lines */
#if FOR_FILTER
    void *cvt; /* pointer to raster info - should be null unless exporting */
#endif
} wholeField;

typedef struct font_props {
    char *familyName; /* for example, "Times" */
    char *fontName; /* for example, "Times-Roman" */
    int weight;
    int slant;
    float ptsize; /* in points */
    int ptsizeInt; /* integer * 10000: e.g. 14.4 pts -> 144000 */
    int underlinePosition; /* in milliems */
    int underlineThickness; /* in milliems */
    r32 bbox; /* font bounding box, mils */
} fontProps;

typedef struct tx_attr {
    unsigned int qDraw   : 1,   /* if TRUE just use XDrawString16 */
             useFgPixel  : 1,   /* if TRUE, member fgPixel meaningful */
             hilite      : 1,
             strike_thru : 1;
    int supx;                    /* non-accumulating x offset */
    int supy;                    /* non-accumulating y offset */
    int kernx;                   /* accumulating x offset */
    int kerny;                   /* accumulating y offset */
    int ruling;                  /* AXRULING_SINGLE, e.g., */
    fontContext fconx;           /* font metrics and table reference */
    fontProps fontAttr;          /* user-specified attr info */
    p32 offset;                  /* measured from current set position */
    int strokecolor;             /* fg color of letterform outline stroke */
    int strokebcolor;
    attr_type *strokepattern;    /* fill information for outline stroke */
    attr_type *fillpattern;      /* fill information for letterform fill */
    char *screenFontName;        /* resolved xlfd name */
    int fillcolor;               /* packed cmyk letterform fill fg color */
    int fillbcolor;              /* packed cmyk letterform fill bg color */
    unsigned long fgPixel;       /* direct pass-thru of fill fg color */
    tx_char noRuleChars[2];      /* tab and or space */
    axformTerms transform;       /* angle, x scale, e.g., */
    axform tctm;                 /* target (reference) xform matrix coeffs */
    axform fctm;                 /* same as tctm but without zoom */
    int nChars;                  /* number of chars under this attr segment */
    axtIndent *indent;           /* apply to bounder intercepts */
    axtDropShadow *dropShadow;
} txAttr;

typedef struct per_char {
    tx_char charcode;
    int width; /* horizontal projection of bbox in mils */
    int height; /* vertical proj of bbox in mils */
    p32 orig;  /* hv projection of bbox 'origin', mils */
    p32 offset; /* formatting offset; affects downstream chars, mils */
    p32 kern;  /* kerning, curvefitting; affects downstream chars, mils */
    p32 fkern; /* same as kern but without zoom */
    p32 sup;  /* sub,superscript; affects this character only, mils */
    r32 sbox; /* untransformed source metrics, screenpix */
    b32 bbox; /* transformed footprint in mils */
    int padvance; /* untransformed target advance in mils */
    int sadvance; /* untransformed screen advance in screenpix */
    p32 advance; /* transformed advance in mils */
    p32 fadvance; /* same as advance but without zoom (for formatting) */
    axform *ctm;    /* transform data for this char */
    char *letterName; /* used if cacheing letterforms */
    unsigned int letterNameKnown : 1;
    unsigned int screenMetrics : 1; /* whether known */
    unsigned int targetMetrics : 1; /* whether known */
    unsigned int changed : 1;       /* since last redisplay */
    unsigned int deleted : 1;       /* since last redisplay */
    unsigned int damaged : 1;       /* since last redisplay */
    unsigned int inserted : 1;      /* since last redisplay */
    unsigned int moved : 1;         /* since last redisplay */
    unsigned int hidden : 1;        /* since last reFit */
    unsigned int temp : 1;          /* various temporary uses */
    unsigned int hyphen : 1;
    unsigned int nonbreakspace : 1;
    unsigned int linebreak : 1;     /* formatter started line with this */
    unsigned int hanging : 1;       /* formatter 'hung' this (space) char */
    unsigned int hilite : 1;        /* this char should be hilighted */
} perChar;

    /* directly associate a perChar and its segment for convenience */
typedef struct ch_seg {
    txAttr *txAp;
    perChar *pc;
} chSeg;

typedef struct e_state {
    int disp;  /* if true, text believed to be on-screen */
        /* editPos: smallest index affected by edit -- support for
         * an optimization where the min-redisp code assumes
         * that no attention is needed for that portion of
         * the text upstream of editPos.  This opt introduces
         * the restriction that no edits can be performed
         * upstream of the editPos during the edited state;
         * if you want to do such an edit, display first
         * to reset the old-copy info
         */
    int editPos;
	
	/* bg pixmap for lifting text
	 */
    Pixmap background;
    unsigned long backgroundPixel;   /* for simple color rather than pixmap */
    int javaPixel;   /* for simple color rather than pixmap */
    r32 backRect;   /* window-rel position, size of background pix */
	
	/* cursor stuff
	 */
    int cursorPos;
    int cursorType;
    XPoint lastCursorPoints[4];
    int lastCursorPointcount;
    r32 lastCursorClip;
	
    r32 fieldBounder;
    int fullDisp;                /* unknown damage, so draw in entirety */
	
    unsigned int
	 cursorOn         : 1, /* cursor belived to be on-screen */
	 cursorActive     : 1,
	 lastCursorActive : 1,
	 inEdit           : 1, /* display implies drawCursor */
	 forceBGDraw      : 1; /* draw the background, no matter what */
} eState;

typedef struct ax_text {
    int pid;
	void *gContext;        /* handle of ax graphics context */
    tx_char *textstr;
    wholeField whole;
    txAttr  *atts;
    perChar *items;
    txAttr *userDefaults;
    unsigned int userDefaultFlags;
        /* some buffer lengths
         */
    int wCharBuflen;
    int txAttrBuflen;
    int perCharBuflen;
        /* state bits expressing how fully filled out instance is
         */
    unsigned int
     constructed   : 1, /* att info and text string */
     measured      : 1, /* target (printer) metrics */
     displayable   : 1, /* screen metrics and resources */
     draw_shadow   : 1, /* draw shadows */
     draw_emboss   : 1, /* draw emboss */
     shadow_only   : 1; /* draw only shadows */
    
    eState editState;
    r32 clipRect;
    struct ax_text *old;
} axtext;

typedef struct point_pair {
    p32 p0;
    p32 p1;
} pointPair;

typedef struct intercept_t {
    int paircount;
    pointPair *pairs;
} intercept;

typedef struct tab_stop {
    int type;
    int leadering;
    int decimalChar;
    p32 pos;
} tabStop;

typedef struct tab_ruler {
    int numTabs;
    tabStop *tabs;
} tabRuler;

typedef struct line_format {
    intercept edges;
    int justificationType;
    tabRuler *ruler;
} lineFormat;

#define NO_CHAR ((tx_char) 0) /* disallows code 0 as 'no rule' char.. */

/* support for insert/delete/overstrike
 */
#define DELTA 5
#define TYPE_W_CHAR 0
#define TYPE_TXATTR 1
#define TYPE_PERCHAR 2

/* edit types
 */
#define RESET 0
#define INS_CHARS 1
#define OVERSTRIKE_CHARS 2
#define DEL_CHARS 3
#define SET_OFFSETS 4
#define APPLY_ATTR 5

/* which edge of edit range
 */
#define RIGHT_EDGE 1
#define LEFT_EDGE 2

/* an impossibly? large coordinate value
 */
#if !AXMS       /* already defined under NT */
#define INFINITE (0x7fffffff)
#endif  /* DAMNT */

#include <fontcach.h>
#endif /* _AXTXINT_H */


/* The axtext datatype is used for interactive text display.  It
 * is managed using code in axtext.c.  The axtext.c code uses the
 * font cache code in fontcache.c for learning about available
 * screen resources, and screen & 'target' metrics.
 *
 * A major goal of the axtext.c code is to provide support for
 * arbitrary pointsize text, even when no scalable screenfonts are
 * present during a session.  Other graphical effects like rotation,
 * nonsquare scaling, and shear are also supported.  Polytext is
 * supported.
 *
 * The text attribute description used with the axtext.c code is
 * intended to be 'sparse', meaning that it is always possible to
 * use an incomplete attribute description at the api.  Attrs expressed
 * there override any 'user default' attrs in effect, which inturn
 * override the internal hardcoded default attrs.  Therefore one can
 * perform edits to a persistent instance whose history involved
 * unrelated edits by someone else -- it isn't necessary to understand
 * how to do those past edits in order to work with the instance.
 * 
 * The display and editing calls defined in axtext.c cause the
 * 'construction' of data onto an instance of axtext.  Much of this
 * data comes from the font cache.
 * 
 * The construction is implemented in routines called text1, axtMeasuredText,
 * and axtDisplayableText.  displayable text derives from measured text,
 * which derives from text1.
 * Text1 does no more than to establish a thorough internal
 * attribute description.  No font cache data is used at that point.
 * axtMeasuredText establishes the measurements (in target coordinates) of the
 * text instance.  It uses the fontcache to resolve the attribute info
 * into the actual target font metric names needed for the measurement.
 * measured text is the 'most derived' form of axtext common to both screen
 * display and printing. axtDisplayableText is geared toward screen display;
 * it determines whether straight-to-screen or image-based display will
 * be used, by using further fontcache information.  The rules for
 * deciding how to display are, roughly:
 *           Is the text to be rotated, sheared, mirrored, etc. ?
 *                        -> use image-based display
 *           else, is a scalable screenfont available?
 *                        -> go straight-to-screen
 *           else, are the target metrics close enough to the
 *                 available raster screenfont metrics?
 *                        -> go straight-to-screen
 *           else,
 *                        -> use image-based display
 * 
 * If axtDisplayable has to resort to image-display, it will cache the
 * letterforms in the fontcache.
 *
 * The construction involves dynamic memory usage.  So, before the first
 * use of an axtext structure in some scope, always make sure it's zeroed,
 * and before leaving that scope, always deconstruct the structure using
 * daxtext().
 *
 * The screen display follows a minimal-lift, minimal-set pattern, which
 * relies on comparing the state of the text since last display with the
 * current state of the text.
 */
