/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: bitblt.c
   Applix ASC version: 19
   Date: Mon Nov 16 13:53:33 1998
*/
#include <basicwin.h>
#include <errno.h>
#include <stdio.h>
#include <aos.h>
#include <tmi.h>
#include <atm.h>
#include <atmattr.h>
#include <oldwin.h>
#include <tmassert.h>
#include <laytext.h>
#include <devsetup.h>



extern Display *Dpy;
extern GC Gc1;

extern Drawable HostWin;
extern unsigned long FGpixel, BGpixel, PlaneMask;
extern AxRectangle VisRect;
extern int ToggleFG;

#if IBM_G && IBM_PHM == 0  /* { */
extern char *DpyBase, *OffBase;
#else
#define DpyBase DPY_BASE
#define OffBase OFF_BASE
#endif  /*  }  */

extern short ATMtrace;
extern FILE *ATfp;
extern short TermReleased;
extern short DpyDepth;
extern short ColorDisplay;
extern short FGcolor, BGcolor;
extern A_Bitmap ToneBM;
extern unsigned long AlisPixels[];

int resolution = 1;
unsigned char fieldMask[] = { 0, 0x80, 0xC0, 0xE0, 0xF0, 0xF8, 0xFC, 0xFE, 0xFF };
unsigned char msbfBitMask[] = {0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01};
unsigned char lsbfBitMask[] = {0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80};


/************************************************************/
#define TMZEROS     0
#define TMS_A_D     1
#define TMS_A_ID    2
#define TMS     3
#define TMIS_A_D    4
#define TMD     5
#define TMS_X_D     6
#define TMS_O_D     7
#define TMIS_A_ID   8
#define TMIS_X_D    9
#define TMID        10
#define TMS_O_ID    11
#define TMIS        12
#define TMIS_O_D    13
#define TMIS_O_ID   14
#define TMONES      15
/************************************************************/

/*
 * This blitter does not accept a clipping rectangle.
 * Routines such as ATMlayerop do clipping before calling ATMbitblt.
 */


Pixmap
mapSave(rp, sx, sy, wid, ht)
    AxRectangle *rp;            /* rectangle as a structure */
    int sx, sy, wid, ht;    /* rectangle as origin and size */
{
#if 0  /* XCopyArea except screen-to-screen is much too slow */ 
        /* also, doesn't work on PMAX! */
    short discard;
    Pixmap pm;

    /* X window system does not always give exposure event when needed */
    WMwinsize(&discard, &discard);      /* another X Windows workaround */

    if (rectInRect(rp, &VisRect)) {
        if ( (pm = XCreatePixmap(Dpy, HostWin, wid, ht, DpyDepth)) != 0 ) {
            /* very slow operation */
            XSetPlaneMask(Dpy, Gc1, PlaneMask);
            XCopyArea(Dpy, HostWin, pm, Gc1, sx, sy, wid, ht, 0, 0);
        }
        return pm;
    } else
#endif  /* #if 0 */ 

        return 0;
}

/*
 * X11 version
 *
 *  side effect: destination Bitmap and its Layer can be ejected
 */

Pixmap GetStippleMap();

ATMbitblt(sb, rp, db, pp, trmode, htone)
A_Bitmap *sb, *db;  /* -> source, dest bitmaps */
AxRectangle *rp;        /* -> source rectangle */
Point *pp;      /* -> origin of dest rectangle */
int trmode;     /* transfer mode (display function code) */
int htone;      /* halftone indicator */
{
    int sx, sy, wid, ht, dx, dy;
    char *bmap;
    Pixmap pm;
    int tmp, cmode;
    unsigned long pel, pmask;
    XImage ximg;
    Window rootWinID;
    int Xpos, Ypos;
    int useStencil;
    int maskMode;
    unsigned int bgPel;
    unsigned int winWidth, winHeight;
    unsigned int bordWidth, winDepth;
    extern short NoColor;           /* forced monochrome */

    if (TermReleased)
        return;

    /* calc source and destination info */
    if (sb != NULL)
    {
        sx = rp->origin.x - sb->maprect.origin.x;
        sy = rp->origin.y - sb->maprect.origin.y;
    }
    wid = rp->corner.x - rp->origin.x;
    ht = rp->corner.y - rp->origin.y;
    dx = pp->x - db->maprect.origin.x;
    dy = pp->y - db->maprect.origin.y;

    pmask = PlaneMask;
    pel = FGpixel;
    if ( !ColorDisplay )
        cmode = trmode;
    else
    {
        if ( trmode == TMS_X_D )
        {
            pel = AlisPixels[0] ^ AlisPixels[1];
            pmask = AllPlanes;
            cmode = TMS_X_D;
        }
        else if ( trmode == 16 )
        {
            pel = AllPlanes;
            pmask = AllPlanes;
            cmode = TMS_X_D;
        }
        else
            cmode = TMS;
    }

    XSetPlaneMask(Dpy, Gc1, pmask);
    XSetForeground(Dpy, Gc1, pel);

    /* set up sourceless blits */
    if(trmode==TMZEROS||trmode==TMID||trmode==TMD||trmode==TMONES||trmode==17)
    {
        switch(trmode)
        {
            case TMZEROS:
                pel = BGpixel;
                cmode = GXcopy;
            break;
            case TMONES:
                pel = FGpixel;
                cmode = GXcopy;
            break;
            case TMID:
                pel = AlisPixels[0] ^ AlisPixels[1];
                cmode = GXxor;
            break;
            case 17:
                pel = AllPlanes;
                cmode = GXxor;
            break;
        }

        XSetClipMask(Dpy, Gc1, None);
        XSetState(Dpy, Gc1, pel, BGpixel, cmode, AllPlanes);
        if ( htone == 0 )
            XSetFillStyle(Dpy, Gc1, FillSolid);
        else
        {
            XSetStipple(Dpy, Gc1, GetStippleMap(htone-1));
            XSetFillStyle(Dpy, Gc1, FillOpaqueStippled);
        }
        XFillRectangle(Dpy, HostWin, Gc1, dx, dy, wid, ht);
        XSetFillStyle(Dpy, Gc1, FillSolid);
        htone = 0;
    }
    else if (sb == A_Display && db == A_Display)
    {
        XSetGraphicsExposures(Dpy, Gc1, TRUE);
        XSetClipMask(Dpy, Gc1, None);
        XSetFunction(Dpy, Gc1, trmode);
        XCopyArea(Dpy, HostWin, HostWin, Gc1, sx, sy, wid, ht, dx, dy);
        XSetPlaneMask(Dpy, Gc1, PlaneMask);
        XSetGraphicsExposures(Dpy, Gc1, FALSE);
    }
    else if (db == A_Display)   /* memory to display */
    {
        if (sb->wmF)    /* map is a window manager resource */
        {
            if (sb->depthF)     /* map is a full Pixmap */
            {
                assert(sb->mapdata != NULL);
                /* very slow operation */
                XCopyArea(Dpy, (Pixmap) sb->mapdata, HostWin, Gc1,
                        sx, sy, wid, ht, dx, dy);
            }
            else            /* it's a bitmap: for now, assume stencil  */
            {
                if (ColorDisplay) {
                    XSetState(Dpy, Gc1, 0x7f, 0, GXandInverted, PlaneMask);
                    XCopyPlane(Dpy, (Pixmap) sb->mapdata, HostWin, Gc1,
                        sx, sy, wid, ht, dx, dy, 1);
                    XSetState(Dpy, Gc1, FGpixel, 0, GXor, PlaneMask);
                    XCopyPlane(Dpy, (Pixmap) sb->mapdata, HostWin, Gc1,
                        sx, sy, wid, ht, dx, dy, 1);
                } else {
                    cmode = ToggleFG? GXandInverted : GXor;
                    XSetState(Dpy, Gc1, FGpixel, 0, cmode, PlaneMask);
                    XCopyPlane(Dpy, (Pixmap) sb->mapdata, HostWin, Gc1,
                        sx, sy, wid, ht, dx, dy, 1);
                }
            }
        }
        else            /* map is in ATM memory */
        {
            if (trmode==TMS_O_D || trmode==TMIS_O_D)
                useStencil = TRUE;
            else
                useStencil = FALSE;

            bmap = sb->mapdata;
            ximg.height = sb->maprect.corner.y - sb->maprect.origin.y;
            ximg.width = sb->maprect.corner.x - sb->maprect.origin.x;
            if (sb->depthF == 1 && DpyDepth == 8)
            {
                /* COLOR IMAGE support */
                ximg.depth = 8;
                ximg.format = ZPixmap;
                ximg.bits_per_pixel = 8;
                ximg.bytes_per_line = sb->mapwid + sb->mapwid%2;
            }
            else
            {
                ximg.depth = 1;
                ximg.format = XYBitmap;
                ximg.bits_per_pixel = 1;
                ximg.bytes_per_line = sb->mapwid;
            }
            ximg.xoffset = 0;
            ximg.data = bmap;
#if PMAX || X386_NEW || VAX || CLIX
            ximg.byte_order = LSBFirst;
#else
            ximg.byte_order = MSBFirst;
#endif
            ximg.bitmap_unit = 16;
            ximg.bitmap_bit_order = LSBFirst;
            ximg.bitmap_pad = 16;

            if(ximg.bits_per_pixel == 8)
            {
                /* color image; source transfers only */
                trmode = GXcopy;
                bgPel = BGpixel;
            }
            else if (useStencil)
            {
                /* the tranfer is or-type. Use stencil pass to zero the
                 * foreground while leaving the background alone, then use
                 * the 2nd pass to or the fg pixval onto the foreground, while
                 * again leaving the background alone.
                 */
                XSetState(Dpy,Gc1,~AllPlanes,AllPlanes,GXand,AllPlanes);
                XPutImage(Dpy, HostWin, Gc1, &ximg, sx, sy, dx, dy, wid, ht);
                bgPel = ~AllPlanes;
            }
            else
                bgPel = BGpixel;

            XSetState(Dpy,Gc1,FGpixel,bgPel,trmode,AllPlanes);

            if (ximg.depth==1 && FGpixel==BGpixel && trmode==TMS)
            {
                XFillRectangle(Dpy,HostWin,Gc1,dx,dy,ximg.width,ximg.height);
            }
            else if (!useStencil || FGpixel != 0) {
                /* if !useStencil, this is the only output call
                 * and is mandatory.  if useStencil, then we
                 * need to do this only if the FGpixel is some-
                 * thing other than the zero val just cut out
                 * during the useStencil section above */
                XPutImage(Dpy, HostWin, Gc1, &ximg, sx, sy, dx, dy, wid, ht);
            }
        }
    }
    else if (sb == A_Display)   /* display to memory */
    {
        /* use whole map as destination; assume source transfer mode */

        if (db->wmF)            /* window manager resource */
        {
            /* deallocate old Pixmap if any */
            /* (should this ever happen?) */

            if ( (pm = (Pixmap) db->mapdata) != 0 )
            {
                TLOGGER((ATfp, "!!!bitblt XFreePixmap(0x%x)\n", pm));
                XFreePixmap(Dpy, pm);
            }

            /* allocate new Pixmap and save screen in it */

            if ( (pm = mapSave(rp, sx, sy, wid, ht)) == 0 )
            {
                TLOGGER((ATfp, "bitblt mapSave failed\n"));
                db->inMemF = FALSE;     /* indicate bitmap ejected */
            }
            db->mapdata = (char *) pm;
        }
        else
        {
            db->mapdata = (char *) XGetImage(Dpy, HostWin, sx, sy, wid, ht,
                    (DpyDepth!=1 && NoColor)||ColorDisplay ? AllPlanes : 1,
                    (DpyDepth!=1 && NoColor)||ColorDisplay ? ZPixmap:XYPixmap);
#if 0 /* ifndef NDEBUG */  /* { */
            {
                XImage *xip;


                xip = (XImage *) db->mapdata;
                TLOGGER((ATfp, NXLT("XImage wid=%d ht=%d\n"),
                        xip->width, xip->height));
                TLOGGER((ATfp, NXLT("bpl=%d xoff=%d bu=%d bp=%d\n"),
                        xip->bytes_per_line, xip->xoffset, xip->bitmap_unit,
                        xip->bitmap_pad));
                TLOGGER((ATfp, NXLT("byord=%d bitord=%d\n"), xip->byte_order,
                        xip->bitmap_bit_order));
printf("bit order %d, byte order %d\n", xip->bitmap_bit_order, xip->byte_order);
#if 1           /* 1 to see image blasted to the upper left of the window */
                XPutImage(Dpy, HostWin, Gc1, xip, 0, 0, 0, 0, wid, ht);
#endif
            }
#endif  /*  }  */
        }
    }
}


polyf(bp,polyPoints,nPts,trmode,clipRp)
A_Bitmap *bp;
TmPoint32  *polyPoints;
int nPts;
int trmode;
AxRectangle *clipRp;
{
    extern int halftoneMode;
    int i, cmode;
    unsigned long pel, pmask;
    XPoint x_polyPts[10];
    XRectangle fillClipRect;

    if( nPts > 10 ) nPts = 10;
    for(i=0; i<nPts; i++)
    {
        x_polyPts[i].x = polyPoints[i].x;
        x_polyPts[i].y = polyPoints[i].y;
    }
    pmask = PlaneMask;
    pel = FGpixel;
    switch( trmode )
    {
        case TMID:
            pmask = AllPlanes;
            cmode = GXxor;
            pel = (WhitePixel(Dpy, DefaultScreen(Dpy))) ^
                  (BlackPixel(Dpy, DefaultScreen(Dpy)));
            break;

        case TMONES:
            pel = FGpixel;
            cmode = GXcopy;
            break;

        default:
            case TMZEROS:
            pel = BGpixel;
            cmode = GXcopy;
            break;
    }
    fillClipRect.x = clipRp->origin.x;
    fillClipRect.y = clipRp->origin.y;
    fillClipRect.width = clipRp->corner.x - clipRp->origin.x;
    fillClipRect.height = clipRp->corner.y - clipRp->origin.y;
    XSetClipRectangles(Dpy, Gc1,
                       0, 0,
                       &fillClipRect,
                       1, Unsorted);
    XSetState(Dpy, Gc1, pel, BGpixel, cmode, pmask);
    if( halftoneMode )
    {
        XSetStipple(Dpy, Gc1, GetStippleMap( halftoneMode-1 ) );
        XSetFillStyle(Dpy, Gc1, FillOpaqueStippled);
    }
    else
    {
        XSetFillStyle(Dpy, Gc1, FillSolid);
    }
    XFillPolygon(Dpy, HostWin, Gc1, x_polyPts, nPts, Convex, CoordModeOrigin);
    XSetClipMask(Dpy,Gc1,None);
}

/*******************************************/
/* Some utils to support "image text mode", which
 * assumes default screen treatment of 75dpi, and
 * a printer res. of 300dpi.  In 300dpi image text
 * mode, objects are represented on the screen so
 * that their pixel dimensions are the same as on
 * the printer.
 */

/***************************************************/

/* the following macro comes from fileimage.h */
#define BYTE_WIDTH(w)       ((((w) + 15) >> 3) & ~1)

/****************** window system dependent screenswap ***********************/

/*
 * WMscreenswap -- perform screenswap in window system dependent way
 */
WMscreenswap(bp)
    A_Bitmap *bp;
{
    Pixmap pm;
    int xx, yy, wid, ht;        /* screen rectangle */

    xx = bp->maprect.origin.x;
    yy = bp->maprect.origin.y;
    wid = bp->maprect.corner.x - bp->maprect.origin.x;
    ht = bp->maprect.corner.y - bp->maprect.origin.y;
    
    XSetFunction(Dpy, Gc1, GXcopy);
    XSetPlaneMask(Dpy, Gc1, PlaneMask);
    pm = mapSave(&bp->maprect, xx, yy, wid, ht);
    assert(bp->mapdata);
    XCopyArea(Dpy, (Pixmap) bp->mapdata, HostWin, Gc1, 0, 0, wid, ht, xx, yy);
    TLOGGER((ATfp, "WMscreenswap XFreePixmap(0x%x)\n", bp->mapdata));
    XFreePixmap(Dpy, (Pixmap) bp->mapdata);
    bp->mapdata = (char *) pm;      /* new Pixmap or 0 */
    if (pm == 0)
    {
        TLOGGER((ATfp, "WMscreenswap XPixmapSave failed\n"));
        bp->inMemF = FALSE;     /* indicate bitmap ejected */
    }
}





#if 0  /* { */
int
scaleData(imData, imWid, imHyt, imBytes, destWid, destHyt, destBytes)
char *imData;
int  *imWid, *imHyt, *imBytes, destWid, destHyt, *destBytes;
{
    unsigned char *newImBuf, *scanBuf;
    int maxBytes, (*xScale)(), (*yScale)();
    int xShrinkIm(), xStretchIm(), xCopyIm();
    int yShrinkIm(), yStretchIm(), yCopyIm();

    /* allocate any needed buffers */
    *destBytes = BYTE_WIDTH(destWid);
    maxBytes = (*imBytes > *destBytes) ? *imBytes : *destBytes;
    newImBuf = (unsigned char *)malloc(*destBytes * destHyt);
    if(newImBuf == NULL)
    {
        return(!AOK);
    }
    scanBuf  = (unsigned char *)malloc(maxBytes);
    if(scanBuf == NULL)
    {
        free(newImBuf);
        return(!AOK);
    }

    /* determine what x scaling needs to be done... */
    if(*imWid > destWid)
        xScale = xShrinkIm;
    else if(*imWid < destWid)
        xScale = xStretchIm;
    else
        xScale = xCopyIm;

    /* and what y scaling needs to be done */
    if(*imHyt > destHyt)
        yScale = yShrinkIm;
    else if(*imHyt < destHyt)
        yScale = yStretchIm;
    else
        yScale = yCopyIm;

    /* Go to it */
    (*yScale)(xScale,
              imData, *imWid, *imHyt, *imBytes,
              newImBuf, destWid, destHyt, destBytes,
              scanBuf); 
}

int
yStretch(xScale,
         imData, imWid, imHyt, imBytes,
         newImBuf, destWid, destHyt, destBytes,
         scanBuf);
int (*xScale)();
char *imData;
unsigned char *newImBuf, *scanBuf;
int imWid, imHyt, imBytes, destWid, destHyt, destBytes;
{
    int linesLeft, lineIncrement, modulusMarker, lineLength;
    int yScaleFactor, yScaleModulus;
    unsigned char *imPtr, *scanPtr;

    yScaleFactor  = destHyt / imHyt;
    yScaleModulus = destHyt % imHyt;
    modulusMarker = imHyt >> 1;
    imPtr = (unsigned char *)imData;

    linesLeft = imHyt;
    while(linesLeft > 0)
    {
        /* set the line increment, adding in the remainder at some point */
        lineIncrement = yScaleFactor;
        if((modulusMarker -= yScaleModulus) < 0)
        {
            /* add an extra line now */
            modulusMarker += linesLeft;
            lineIncrement++;
        }

        /* the line increment should never exceed the number of lines left */
        if(lineIncrement > linesLeft)
            lineIncrement = linesLeft;

        /* get the current line from the image */
        lineLength = imBytes;
        scanPtr = scanBuf;
        while(lineLength--)
            *scanPtr++ = *imPtr++;

        /* now do any required x scaling on this line */
        memset(newImBuf, 0, destBytes);
        (*xScale)(scanBuf, newImBuf, destWid, destHyt, imWid, imHyt);

        /* replicate this line over the next few lines */
        while(

    }
}
#endif  /*  }  */


