/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: buttncom.c
   Applix ASC version: 38
   Date: Tue Jan 12 21:12:20 1999
*/
#include <basicwin.h>
#include <ax_xm.h>
#include <global.h>
#include <xmlib.h>
#include <xmpixel.h>
#include <axtext.h>

#include <buttncom.p>


extern GC SunGC, ShadowGC;
extern GC ForeGC, BackGC, RecessGC;
extern Display *Dpy;
extern int UseWidgetColors;
extern short ColorDisplay;
extern int Thim2dMode;

extern int AxAlternateDialogFont;
extern int AxCharAbove, SmallCharAbove;
extern AXWidget_pushbutton ButtonArmed;
extern XFontStruct *AxDlgMainFont, *AxDlgMiniFont;

#if AXMS
#define AX_SS_TABBUTTON_EXT		6

extern PIXEL_VALUE SunPixel, RecessPixel;
extern PIXEL_VALUE ForePixel, InactiveTextPixel;
#endif


centerButtonText(width, height, name, font, small_font, xp, yp)
int width;           /* button width */
int height;          /* button height */
mbchar *name;        /* button label */
XFontStruct *font;   /* button label font */
int small_font;      /* ancient "small-font" boolean, probably always false */
int *xp, *yp;        /* in/out: in as button rect origin, out as setpoint */
{
    int wid, hyt, ascent;
    XFontStruct *fs;

    if (font || AxAlternateDialogFont) {
        if (font) {
            fs = font;
        }
        else {
            fs = small_font ? AxDlgMiniFont: AxDlgMainFont;
        }
        wid = AXTextWidth(fs, name, MBstrlen(name));
        hyt = fs->ascent + fs->descent;
        ascent = fs->ascent + 1;
    } else {
        /* font is axsab-16 or its small variant;
           use measurement compatible with pre-3.2 ax */
        if (small_font) {
            wid = SmallBasicStrlen(name);
            ascent = hyt = SmallCharAbove;
            ascent--;
        } else {
            wid = BasicStrlen(name);
            ascent = hyt = AxCharAbove;
            ascent--;
        }
    }
    
    *xp += (width / 2) - (wid / 2);
    *yp += ((height - hyt) / 2) + ascent;
}   


SSDrawWorkAreaLine(win, x1, y1, x2, y2)
Window win;
int x1, y1;             /* positon to start */
int x2, y2;             /* positon to end */
{

    XDrawLine(Dpy, win, ShadowGC, x1, y1, x2, y2);
}




/*
 * S S D r a w W o r k A r e a B u t t o n ()
 *
 * Draws a fake button, in the ss work area
 * This has nothing to do with the widget kit.
 */
SSDrawWorkAreaButton(tssP, win, ptsize, color, x, y, width, height, 
                     clip, on, align, str, draw2D, editbutt, objname)
	void	*tssP;
	Window win;
	int 	ptsize, 
			x, y,           /* positon to start */
		width, height,  /* dimensions of the button */
		on,  /* whether to draw the button in off or on(depressed) state */
		align,
		draw2D, /* Just Draw a Rectangle */
		editbutt ; /* this is to identify where this call was made, 
					like is it from editline X,Y buttons if yes it would 
					be set to TRUE else false */
	unsigned long color;
	r32 *clip;
	char *str;
	char *objname;
{
	XGCValues values;
	GC outfgc;
	GC outbgc;
	int xPos, yPos;
#if AXMS
	int bThick = 1;
#else
	int bThick = 2;
#endif
	int textWidth=0, fontHeight=0;
	XRectangle rects[1];
	XFontStruct *font;
	int printfactor = 100;
	int slant, weight ;
	mbchar  fontFamily[100];
	int ascent, img_width, img_height, ruling;

    if(clip != NULL) {
		int x1, y1, x2, y2;
		x1 = MAX(x, clip->origin.x);
		y1 = MAX(y, clip->origin.y);
		x2 = MIN(x + width-1, clip->origin.x + clip->width-1);
		y2 = MIN(y + height-1, clip->origin.y + clip->height-1);
		if((x1 > x2) || (y1 > y2))
			return;	/* rectangles don't intersect at all */
        rects[0].x = x1;
        rects[0].y = y1;
        rects[0].width = x2 - x1 + 1;
        rects[0].height = y2 - y1 + 1;
	}else	{
		/* set button width/height for clip */
        rects[0].x = x;
        rects[0].y = y;
        rects[0].width = width;
        rects[0].height = height;
	}


	ssGetNonFontAttrValues(tssP, str, &font, fontFamily, &weight, &slant, 
						   &ruling, &ascent, ptsize, &img_width,
						   &img_height, editbutt);

    XGetGCValues(Dpy, ForeGC, (GCForeground|GCFont), &values);

    if (values.font == 0)
        values.font = font->fid;

    if(color != 0L)
        XSetForeground(Dpy,ForeGC, color);


    XSetClipRectangles(Dpy, ForeGC, 0,0,rects, 1, Unsorted);
    XSetClipRectangles(Dpy, BackGC, 0,0,rects, 1, Unsorted);
    XSetClipRectangles(Dpy, SunGC, 0,0,rects, 1, Unsorted);
    XSetClipRectangles(Dpy, ShadowGC, 0,0,rects, 1, Unsorted);
    XSetClipRectangles(Dpy, RecessGC, 0,0,rects, 1, Unsorted);
    
    if(str != (char *) NULL) {
        if(font != (XFontStruct *)NULL) {
            axSetFont(Dpy, ForeGC, font);
            axSetFont(Dpy, BackGC, font);
            textWidth = AXTextWidth(font, str, strlen(str));
            fontHeight = font->ascent+font->descent;
        }
		else {
            textWidth = img_width;
            fontHeight = img_height;
        }
    }

    switch(align) {  /* alignment of button text */
		
	case 0:
		xPos = x;                              /* left aligned */
		break;
		
	case 1:
		xPos = x + ((width - textWidth)/2);    /* center aligned */
		break;
		
	case 2:
		xPos = x + (width - textWidth);        /* right aligned */
		break;
		
	default:
		xPos = x + (width - textWidth)/2;      /* default to center */
		break;
		
    }
	
    if (!font) {
        yPos = (y + height/2) + ascent/2 - (img_height - ascent)/2;
    }
	else {
        yPos = (y + height/2) + font->ascent/2 - font->descent/2;
    }
    

    if(on) {
        outfgc = BackGC;
        outbgc = ForeGC;
    }
	else {
        outfgc = ForeGC;
        outbgc = BackGC;
    }

    if(draw2D) {
        XFillRectangle(Dpy, win, outbgc, x, y, width, height);
        if(on) {
            XDrawRectangle(Dpy,win, outfgc,x+bThick,y+bThick, 
						   width-2*bThick, height-2*bThick);
        }
		else {
            XDrawRectangle(Dpy,win, outfgc,x-1,y-1, width, height);
        }
    }
	else {
        /* fill it with the control panel background color */
        if(on && ColorDisplay) {
			XFillRectangle(Dpy,win,RecessGC,x,y,width-2,height-2);
		}
        else
            XFillRectangle(Dpy, win, outbgc, x, y, width, height);

		/* draw the button's bevel */
		if (on) {
			AxDrawBevel((AXWidget) NULL, win, x, y, width, height,
						bThick, BEVEL_IN);
            bThick++;
            if(draw2D || !ColorDisplay)
                XDrawRectangle(Dpy, win, outfgc, x+bThick, y+bThick,
							   width-2*bThick, height-2*bThick);
		}
		else {
			AxDrawBevel((AXWidget) NULL, win, x, y, width, height,
						bThick, BEVEL_OUT);
		}
    }

	if (!(objname && *objname != '\0')){/* to draw the string title rather then 
										bitmap **/
		if(str != (char *) NULL) {
			if (!font) {
				PIXEL_VALUE fore, back;
				wchar   wstr[300];

				MbToWStr(str, wstr);
				if (on & ColorDisplay) {
					fore = AXGetForegroundPixel(Dpy, outbgc);
					back = AXGetBackgroundPixel(Dpy, outbgc);
				}
				else {
					fore = AXGetForegroundPixel(Dpy, outfgc);
					back = AXGetBackgroundPixel(Dpy, outfgc);
				}
				ssDrawButtonImgText(tssP, win, wstr, printfactor, xPos, yPos, 
									fontFamily, ptsize, weight, slant,
									fore, back, ruling, FALSE, clip);
			}
			else {
				if(on && ColorDisplay)
					axDrawString(Dpy, win, outbgc, font, xPos, yPos,
								 str, strlen(str));
				else
					axDrawString(Dpy, win, outfgc, font, xPos, yPos,
								 str, strlen(str));
			}
		}
	} else {
		ssDrawButtonAsBitmap(tssP, objname, x, y, width, height,(SSRECT *) clip);
	}

    if (font != (XFontStruct *)NULL) {
		/* restore gc fonts to initial vals
		 */
        axSetFont(Dpy, ForeGC, AxDlgMainFont);
        axSetFont(Dpy, BackGC, AxDlgMainFont);
	}
    if(color != 0L)
        XSetForeground(Dpy,ForeGC, values.foreground);

	XSetClipMask(Dpy,ForeGC,None);
	XSetClipMask(Dpy,BackGC,None);
	XSetClipMask(Dpy,SunGC,None);
	XSetClipMask(Dpy,ShadowGC,None);
	XSetClipMask(Dpy,RecessGC,None);

}

/*
 * SSButtonDimensions
 */
SSButtonDimensions(tssP, str, ptsize, width, height)
	void	*tssP;
	char	*str;	
	int 	ptsize, 
			*width, *height;  /* RET'd dimensions of the button */
{
	XGCValues values;
	int bThick = 2;
	int textWidth=0, fontHeight=0;
	XFontStruct *font;
	int printfactor = 100;
	int slant, weight ;
	mbchar  fontFamily[100];
	int ascent, img_width, img_height, ruling;


	ssGetNonFontAttrValues(tssP, str, &font, fontFamily, &weight, &slant, 
						   &ruling, &ascent, ptsize, &img_width,
						   &img_height, FALSE);

    XGetGCValues(Dpy, ForeGC, (GCForeground|GCFont), &values);

    if (values.font == 0)
        values.font = font->fid;

    if(str != (char *) NULL) {
        if(font != (XFontStruct *)NULL) {
            textWidth = AXTextWidth(font, str, strlen(str));
            fontHeight = font->ascent+font->descent;
        } else {
            textWidth = img_width;
            fontHeight = img_height;
        }
    }
	*width = textWidth;
	*height = fontHeight;
}

/*
 * SS3dMode()
 * Returns TRUE if SS Work area buttons ( row  & col descriptors) are to be
 * drawn in 3d mode, FALSE if not
 */
SS3dMode()
{
extern short NoColor;
    if(Thim2dMode)
        {/* 2d mode in profile */
        return(FALSE);
        }
    return(TRUE);
}



/*
*   ssDrawPlatesPushButton - draw one of the buttons on the SS plates widget
*   return TRUE if local paint code was used, FALSE otherwise
*/
int ssDrawPlatesPushButton(w)   
register AXWidget_pushbutton w;
{
	int x,y, width,height, bthick;
	GC gc;
	int x1, x2;
	int offset;
#if AXMS
	int npts;
	XPoint pts[7];
	BOOL reset_gc = FALSE;

	offset = AX_SS_TABBUTTON_EXT;
#else
	pixcache_t *pxc;
    /* get the pixmap just for sizes */
    pxc = Pixmapper(NXLT("ssfrltab"));
	offset = pxc->width;
#endif

    if(w->font && !w->pixmapLabel) {

        w->mapped = TRUE;
        if(UseWidgetColors || w->value != 0)
            ssDrawColorTab(w);
        else
            ssDrawNoColorTab(w);

        x = w->xpos - offset;
        y = w->ypos;

#if AXMS
		/* set up a poly the shape of the tab */
		npts = 7;
		pts[0].x = w->xpos - offset + 1;
		pts[0].y = w->ypos + w->height - 2;
		pts[1].x = pts[0].x;
		pts[1].y = w->ypos + offset;
		pts[2].x = w->xpos;
		pts[2].y = w->ypos + 1;
		pts[3].x = w->xpos + w->width;
		pts[3].y = pts[2].y;
		pts[4].x = w->xpos + w->width + offset;// - 1;
		pts[4].y = pts[1].y;
		pts[5].x = pts[4].x;
		pts[5].y = pts[0].y;
		pts[6].x = pts[0].x;
		pts[6].y = pts[0].y;
#endif

		/* draw the top line */
        XDrawLine(Dpy, WIN_OF(w), SunGC, w->xpos, y, w->xpos+w->width-1, y);
#if !AXMS
	/* don't draw a second line on Windows because we use a thickness of 1 */
        if ( w->value != 0 )
		{
            XDrawLine(Dpy, WIN_OF(w), SunGC, w->xpos, y+1, 
					  w->xpos+w->width-1, y+1);
        }
#endif

        /* draw the bottom line */
        if(0 == w->value )  {
            /* button off */
            y = w->ypos + w->height-2;
            XDrawLine(Dpy, WIN_OF(w), SunGC, w->xpos - offset, 
                y, w->xpos + w->width + offset - 1, y);
#if !AXMS
	/* don't draw a second line on Windows because we use a thickness of 1 */
            XDrawLine(Dpy, WIN_OF(w), SunGC, w->xpos - offset, 
                y+1, w->xpos + w->width + offset - 1, y+1);
#endif
        }
		else {
            /* button on */
            y = w->ypos + w->height - 2;
            x1 = w->xpos - offset + w->thickness;
#if AXMS
            x2 = w->xpos + w->width + offset - 1;
            XDrawLine(Dpy, WIN_OF(w), BackGC, x1, y, x2, y);
#else
            x2 = w->xpos + w->width + offset - 5;
            XDrawLine(Dpy, WIN_OF(w), BackGC, x1, y, x2, y);
            XDrawLine(Dpy, WIN_OF(w), BackGC, x1, y+1, x2, y+1);
#endif
        }
        
        /* display the text */
        x = w->xpos;
        y = w->ypos;
        width = w->width;
        height  = w->height;
        if ( w->thickness )
            bthick = w->thickness;
        else
            bthick = 2;

        /* fill the button with foreground or background value */
        if ( w->value ) {
            /* button down position */
            if (UseWidgetColors) {
#if AXMS
                XFillPolygon(Dpy, WIN_OF(w), BackGC, pts, npts,
								Nonconvex, CoordModeOrigin);
#else
                XFillRectangle(Dpy, WIN_OF(w), BackGC,
                               w->xpos, w->ypos+2, w->width,
                               w->height-4);
#endif
            }
            else {
                XFillRectangle(Dpy, WIN_OF(w), BackGC, w->xpos,
							   w->ypos+2, w->width,
                               w->height-4);
            }
        
        }
		else {
            /* button up */
            if(UseWidgetColors) {
#if AXMS
                XFillPolygon(Dpy, WIN_OF(w), RecessGC, pts, npts,
								Nonconvex, CoordModeOrigin);
#else
                XFillRectangle(Dpy,WIN_OF(w), RecessGC ,w->xpos, w->ypos+1, 
							   w->width, w->height-3);
#endif
            }
			else {
#if 0
                XFillRectangle(Dpy,WIN_OF(w), SunGC ,w->xpos, w->ypos+1, 
							   w->width, w->height-3);
#endif
            }
        }

        x += bthick;
        y += bthick;

        if ( !w->noBorder )
		{   /* apply 1pix border around image */
            x++;
            y++;
        }

        if (w->value) {
            gc = ForeGC;
		}
        else {
            if ( UseWidgetColors )
			{
#if AXMS
				/* recolor the gc */
				if ( InactiveTextPixel == RecessPixel )
					XSetForeground(Dpy, SunGC, ForePixel);
				else
					XSetForeground(Dpy, SunGC, InactiveTextPixel);
				reset_gc = TRUE;
#endif
                gc = SunGC; /* text for in-active tab */
			}
            else
                gc = BackGC;
        }
        x = w->xpos;
        y = w->ypos;
        centerButtonText(w->width, w->height, w->name, w->font,
						 w->small_font, &x, &y);
        axSetFont(Dpy, gc, w->font);
        axDrawString(Dpy, WIN_OF(w), gc, w->font, x, y,
					 w->name, MBstrlen(w->name));
        axSetFont(Dpy, gc, AxDlgMainFont);
#if AXMS
		if ( reset_gc )
			XSetForeground(Dpy, SunGC, SunPixel);
#endif
        return (TRUE);                         /* did paint work here */
    }

    if ( w->pixmapLabel || w->font || w->color) {
        return (FALSE);                        /* didn't do paint work */
    }

	return (TRUE);   /* safer to claim we're finished painting */
}

static void ssDrawColorTab(w)
register AXWidget_pushbutton w;
{
	GC gc;
	pixcache_t *pxc;
	int x,y;
	int npts;
	XPoint pts[3];

    /* draw the left folder tab */
#if AXMS
    gc = SunGC;
	npts = 3;
	pts[0].x = w->xpos;
	pts[0].y = w->ypos;
	pts[1].x = w->xpos - AX_SS_TABBUTTON_EXT;
	pts[1].y = w->ypos + AX_SS_TABBUTTON_EXT;
	pts[2].x = w->xpos - AX_SS_TABBUTTON_EXT;
	pts[2].y = w->ypos + w->height - 2;
	XDrawLines(Dpy, WIN_OF(w), gc, pts, npts, CoordModeOrigin);
#else
    gc  = ForeGC;
    if(w->value == 0)   {
        if(UseWidgetColors)
            pxc = Pixmapper(NXLT("ssbkltab"));
        else
            pxc = Pixmapper(NXLT("ssbwltab"));
    } else  {
        if(UseWidgetColors)
            pxc = Pixmapper(NXLT("ssfrltab"));
        else
            pxc = Pixmapper(NXLT("ssbwltab"));
            
    }
    if(!pxc)
        return;

    x = w->xpos  - pxc->width;
    y = w->ypos;
#endif


    /* draw the right tab */
#if AXMS
    gc = ShadowGC;
	pts[0].x = w->xpos + w->width;
	pts[0].y = w->ypos;
	pts[1].x = w->xpos + w->width + AX_SS_TABBUTTON_EXT;
	pts[1].y = w->ypos + AX_SS_TABBUTTON_EXT;
	pts[2].x = w->xpos + w->width + AX_SS_TABBUTTON_EXT;
	if ( w->value )
		pts[2].y = w->ypos + w->height - 1;
	else
		pts[2].y = w->ypos + w->height - 2;
	XDrawLines(Dpy, WIN_OF(w), gc, pts, npts, CoordModeOrigin);
#else
    XCopyArea(Dpy,pxc->pix,WIN_OF(w),gc,0,0,pxc->width,pxc->height,x,y);
    if(w->value == 0)   {
        if(UseWidgetColors)
            pxc = Pixmapper(NXLT("ssbkrtab"));
        else
            pxc = Pixmapper(NXLT("ssbwrtab"));
    } else  {
        if(UseWidgetColors)
            pxc = Pixmapper(NXLT("ssfrrtab"));
        else
            pxc = Pixmapper(NXLT("ssbwrtab"));
    }
    if(!pxc)
        return;

    x = w->xpos + w->width;
    XCopyArea(Dpy,pxc->pix,WIN_OF(w),gc,0,0,pxc->width,pxc->height,x,y);
#endif
}


static void ssDrawNoColorTab(w)
register AXWidget_pushbutton w;
{
int x,y;
pixcache_t *pxc;
XPoint points[7];

    /* get pixmap just for sizes */
    pxc = Pixmapper(NXLT("ssfrltab"));

    x = w->xpos  - pxc->width;
    y = w->ypos;
    if(w->value == 0)   {
        points[0].x = x+1;
        points[0].y = y + pxc->height-1;
        points[1].x = x+1;
        points[1].y = y + 6;
        points[2].x = w->xpos;
        points[2].y = y;
        points[3].x = w->xpos + w->width+1;
        points[3].y = y;
        points[4].x = w->xpos + w->width + pxc->width-1;
        points[4].y = y + 6;
        points[5].x = w->xpos + w->width + pxc->width-1;
        points[5].y = y + pxc->height-1;
        points[6].x = x+1;
        points[6].y = y + pxc->height-1;

        /* first clear the button area */
        XFillRectangle(Dpy, WIN_OF(w), BackGC, x, y,
                            points[5].x - points[0].x + 2, pxc->height);        

        XFillPolygon(Dpy, WIN_OF(w), SunGC, points, 7, Convex,CoordModeOrigin);
        XDrawLines(Dpy, WIN_OF(w), ShadowGC, &points[3], 3, CoordModeOrigin);
        return;
    }

    points[0].x = x;
    points[0].y = y + pxc->height-1;
    points[1].x = x;
    points[1].y = y + 6;
    points[2].x = w->xpos;
    points[2].y = y;
    points[3].x = w->xpos + w->width+1;
    points[3].y = y;
    XDrawLines(Dpy, WIN_OF(w), SunGC, points, 4, CoordModeOrigin);
    points[0].x = x+1;
    points[0].y = y + pxc->height-1;
    points[1].x = x+1;
    points[1].y = y + 7;
    points[2].x = w->xpos;
    points[2].y = y+1;
    points[3].x = w->xpos + w->width+1;
    points[3].y = y+1;
    XDrawLines(Dpy, WIN_OF(w), SunGC, points, 4, CoordModeOrigin);

    points[0].x = w->xpos+w->width+1;
    points[0].y = y;
    points[1].x = w->xpos + w->width + pxc->width;
    points[1].y = y + 6;
    points[2].x = w->xpos + w->width + pxc->width;
    points[2].y = y + pxc->height-1;
    XDrawLines(Dpy, WIN_OF(w), ShadowGC, points, 3, CoordModeOrigin);

    points[0].x = w->xpos+w->width-1;
    points[0].y = y;
    points[1].x = w->xpos + w->width + pxc->width-1;
    points[1].y = y + 6;
    points[2].x = w->xpos + w->width + pxc->width-1;
    points[2].y = y + pxc->height-1;
    XDrawLines(Dpy, WIN_OF(w), ShadowGC, points, 3, CoordModeOrigin);
}


