/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: defterm.c
   Applix ASC version: 8
   Date: Mon Nov 16 13:53:33 1998
*/
#include <stdio.h>
#include <aos.h>
#include <tmi.h>

#if !AXMS
#if (SYS&VMSSYS)
#else	/* if (SYS&VMSSYS) */
#if SYS & SYS5 || LINUX
#include <termio.h>
#else
#include <sgtty.h>
#endif
#endif	/* (SYS&VMSSYS) */
#else
#define B9600   13      /* defined for NT build */
#endif  /* !AXMS, DAMNT */

#include <defterm.h>
#include <getstr13.h>

extern int TermFD;
extern FILE *ATfp;
extern short ATMdebug;
extern short ATMtrace;
extern int MapMode;
extern short TermReleased;

/* following globals remember the current cursor position, needed to optimize
	next cursor positioning */
short CursX;
short CursY;

/* foloowing variable is set if highlight and underscore are
	embedded attributes, so that the cursor has to be positioned
	before setting those attrs.
*/
short ATMemb = 0;
short ATMinvEmb = 0;	/* invisible embedded attributes: this is
				the case where attributes are embedded
				but do not occupy a space eg. HP2622.
			*/

	/* this is the subroutine that impliments termcap in ATM */
	/* terminal commands, that are supported now, are defined 
	   in defterm.c */  
char *tgetstr();
char *tgoto();
char *getenv();
#if TEK
char bp[2048];
#else
char bp[1024];
#endif
char *area;
char TermCtlbuf[250];
char *comm[15];
short CommFlag[15];	/* 0 : corresponding term entry not founds in
								termcap.
						*/ 
short ospeed;
char PC;
short cur;
short CrashFlag;
short tempFlag;		/* NB - 11 Feb 88 */

/* reverse video sequence for ibm316X */
static char specSeq[] = {033, 064, 0101};

init_term(termName)
	char termName[];
{
	char *strcpy(), *strcat();
	char id[3];
	short choice;
	short i,j;
	char mesBuf[150];

#if GMODE == 0			/* not in workstation graphics mode */

#if !(SYS&VMSSYS)
	ospeed = B9600;
#endif	/* if !(SYS&VMSSYS) */
	area = TermCtlbuf;
	cur = 0;
	if(tent(termName)!= 1)
		ATMabort(ERRTERMTC); /* "terminal name not found in termcap" */
/*
	Get the character sequence of terminal command, id's are defined in
	termcat(5). Any change here should be reflected in defterm.h .
*/
	strcpy(mesBuf,"\0");
	CommFlag[CM] = tstr(NXLT("cm"),1);
	if(CommFlag[CM] == 0)
		strcat(mesBuf,NXLT("cm "));

	CommFlag[SO] = tstr(NXLT("so"),1);
	if ( strcmp(termName, NXLT("ibm3161")) == 0 || strcmp(termName, NXLT("ibm3163")) == 0 )
		comm[SO] = specSeq;						/* override */
	if(CommFlag[SO] == 0)
		strcat(mesBuf,NXLT("so "));

	CommFlag[SE] = tstr(NXLT("se"),1);
	if(CommFlag[SE] == 0)
		strcat(mesBuf,NXLT("se "));

	CommFlag[CL] = tstr(NXLT("cl"),1);
	if(CommFlag[CL] == 0)
		strcat(mesBuf,NXLT("cl "));

	CommFlag[CE] = tstr(NXLT("ce"),1);
	if(CommFlag[CE] == 0)
		strcat(mesBuf,NXLT("ce "));

	CommFlag[CD] = tstr(NXLT("cd"),1);
	if(CommFlag[CD] == 0)
		strcat(mesBuf,NXLT("cd "));

	if(CommFlag[CD] = tstr(NXLT("pc"),0)) 
		PC=comm[PCC][0];
	else
		PC = '\0';
	CommFlag[UP] = tstr(NXLT("up"),0);
	CommFlag[DO] = tstr(NXLT("do"),0);
	CommFlag[ND] = tstr(NXLT("nd"),0);
	CommFlag[BC] = tstr(NXLT("bc"),0);
	CommFlag[CR] = tstr(NXLT("cr"),0);
	CommFlag[NN] = tstr(NXLT("nn"),0);

	/* NB - 11 Feb 88 - on terminals that do not have us and ue entries
	 * (i.e. no underline mode) use so and se entries (standout mode)
	 * instead.
	 */
	tempFlag = CrashFlag;

	CommFlag[US] = tstr(NXLT("us"),1);
	if(CommFlag[US] == 0)
	{
		CommFlag[US] = CommFlag[SO];
		comm[cur-1] = comm[SO];
	}

	CommFlag[UE] = tstr(NXLT("ue"),1);
	if(CommFlag[UE] == 0)
	{
		CommFlag[UE] = CommFlag[SE];
		comm[cur-1] = comm[SE];
	}

	CrashFlag = tempFlag;

/* get the no of spaces left by stand-out mode  */
	ATMemb = tnum(NXLT("sg"));
	if (ATMemb == -1) ATMemb = 0;	/* default 0 */
#if 0
		strcat(mesBuf,NXLT("sg "));
#endif
	
#if 0
	if(ATMtrace)
	{
		for(i=0;i<15;++i)
		{
			if(CommFlag[i])
			{
				fprintf(ATfp,NXLT("\n%d: len:%d:") ,i,strlen(comm[i]));
				for(j=0;j<strlen(comm[i]);++j)
					fprintf(ATfp,NXLT(" %o"),comm[i][j]);
			}
			fflush(ATfp);
		}
		fprintf(ATfp,NXLT("\nATMemb:%d"),ATMemb);
	}
#endif				/* of #if 0 */

	if(CrashFlag)
	{
		ATMab3(mesBuf, ERRENTTC);
	}

/* for a non-graphics terminal, initialize the screen and  cursor position
*/
	if (!MapMode)
	{
		CursX = 0;
		CursY = 0;
	}

#endif  /* #if GMODE == 0 */

}  /* init_term() */

tent(name)
char *name;
{
#if GMODE == 0
	return(tgetent(bp,name));
#endif
}

tnum(id)
char *id;
{
#if GMODE == 0
	short i;

	if((i=tgetnum(id)) == -1)
	{
#if 0
		CrashFlag = 1;
#endif
	}
	return(i);
#endif
}

tflag(id)
char *id;
{
#if GMODE == 0
	return(tgetflag(id));
#endif
}

tstr(id, flag)
char *id;
int flag;
{
#if GMODE == 0
#if 0  /* obsolete */
	char buf[80];

	comm[cur] = area;
	++cur;
	if(tgetstr(id,&area)==0)
#endif
	if ( (comm[cur++] = tgetstr(id, &area)) == NULL )
	{
		if(flag)
		{
			CrashFlag = 1;
		}
		return(0);
	}
	else
	{
		return(1);
	}
#endif
}

term_ce(x,y)
short x;
short y;
{
#if GMODE == 0
	term_cm(x,y);
	tpt(comm[CE]);
#endif
}

term_cd(x,y)
short x;
short y;
{
#if GMODE == 0
	term_cm(x,y);
	tpt(comm[CD]);
#endif
}

term_so(x,y)
short x;
short y;
{
#if GMODE == 0
	if((ATMemb) || (ATMinvEmb))
	{
		term_cm(x,y);
		CursX = -1;
	}
	tpt(comm[SO]);
#endif
}

term_se(x,y)
short x;
short y;
{
#if GMODE == 0
	if((ATMemb) || (ATMinvEmb))
	{
		term_cm(x,y);
		CursX = -1;
	}
	tpt(comm[SE]);
#endif
}

term_us(x,y)
short x;
short y;
{
#if GMODE == 0
	if((ATMemb) || (ATMinvEmb))
	{
		term_cm(x,y);
		CursX = -1;
	}
	tpt(comm[US]);
#endif
}

term_ue(x,y)
short x;
short y;
{
#if GMODE == 0
	if((ATMemb) || (ATMinvEmb))
	{
		term_cm(x,y);
		CursX = -1;
	}
	tpt(comm[UE]);
#endif
}

/*	This low level fuction position the cursor at pos(x,y).
	Cursor positioning is optimized to use least possible character sequence.
*/

char *tgoto();
term_cm(destcol,destline)
short destcol;
short destline;
{
#if GMODE == 0
	char *adr;
	short cursX;
	short cursY;
	extern short CursOptimise;
	extern short RG_wid;
	extern short RG_ht;

#if 0
	if(ATMtrace)
	{
		fprintf(ATfp,NXLT("\nTERMCAP:tgoto line:%d col:%d"),destline,destcol);
		fflush(ATfp);
	}
#endif
#if 1
	cursX = CursX;
	cursY = CursY;
	CursX = destcol;
	CursY = destline;

#if 0
	if(ATMtrace)
	{
		fprintf(ATfp,NXLT("\ncursX:%d destcol:%d  cursY:%d destline:%d"),
			cursX,destcol,cursY,destline);
		fflush(ATfp);
	}
#endif
	if(cursX != -1)
	{
		
		switch(destline-cursY)
		{
			case(0):
				switch(destcol-cursX)
				{
					case(0):
						/* same cursor position */
						if(ATMtrace)
							fprintf(ATfp,NXLT("same position\n"));
						return(0);
						break;
					case(1):
						if (CursOptimise)
						{
						/* shift with right arrow */
							if(ATMtrace)
								fprintf(ATfp,NXLT("right arrow\n"));
							if(CommFlag[ND])
							{
								tpt(comm[ND]);
								return(0);
							}
						}
						break;
					case(-1):
						if (CursOptimise)
						{
						/* shift with left arrow */
							if(ATMtrace)
								fprintf(ATfp,NXLT("left arrow\n"));
							if(CommFlag[BC])
							{
								tpt(comm[BC]);
								return(0);
							}
						}
						break;
					default:
						if (CursOptimise)
						{
							if(destcol == 0)
							{
						/* use carriage return */
								if(ATMtrace)
									fprintf(ATfp,NXLT("carriage ret\n"));
								if(CommFlag[CR])
								{
									tpt(comm[CR]);
									return(0);
								}
							}
						}
#if 0
						if(ATMtrace)
							fprintf(ATfp,NXLT("use tgoto\n"));
#endif
						break;
				}
				break;
			case(1):
				if (CursOptimise)
				{
					switch(destcol-cursX)
					{
					case(0):
						/* down arrow key */
						if(ATMtrace)
							fprintf(ATfp,NXLT("down arrow\n"));
						if(CommFlag[DO])
						{
							tpt(comm[DO]);
							return(0);
						}
						break;
					default:
						if(destcol == 0)
						{
						/* use newline */
							if(ATMtrace)
								fprintf(ATfp,NXLT("new line\n"));
							if(CommFlag[NN])
							{
								tpt(comm[NN]);
								return(0);
							}
						}
#if 0
						if(ATMtrace)
							fprintf(ATfp,NXLT("use tgoto\n"));
#endif
						break;
					}
				}
				break;
			case(-1):
				if (CursOptimise)
				{
					switch(destcol-cursX)
					{
					case(0):
						/* up arrow key */
						if(ATMtrace)
							fprintf(ATfp,NXLT("up arrow\n"));
						if(CommFlag[UP])
						{
							tpt(comm[UP]);
							return(0);
						}
						break;
					default:
						/* use tgoto() */
#if 0
						if(ATMtrace)
							fprintf(ATfp,NXLT("use tgoto\n"));
#endif
						break;
					}
				}
				break;
			default:
				/* use tgoto() */
				break;
		}
	}
#endif

	if(ATMtrace)
		fprintf(ATfp,NXLT("calling tgoto\n"));
	adr=tgoto(comm[CM],destcol,destline);
	tpt(adr);
	if (!CursOptimise) GText ( (destcol * RG_wid), (destline * RG_ht) );
#endif
}

char OutStr[128];
short OutStrPtr;
int ATMputchar(c)
	char c;
{
	OutStr[OutStrPtr] = c;
	++OutStrPtr;
#if 0
	if(ATMtrace)
	{
		fprintf(ATfp,NXLT("\nATMputchar:%o"),c);
		fflush(ATfp);
	}
#endif
}

tpt(cp)
char *cp;
{
#if GMODE == 0
	if (TermReleased) return;
#if (SYS&VMSSYS)
	if (!cp) return;
#endif
	OutStrPtr = 0;
	tputs(cp,24,ATMputchar);
	write(TermFD, OutStr, OutStrPtr);
#endif
}
