/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: devsetup.h
   Applix ASC version: 1
   Date: Mon Nov 16 13:53:33 1998
*/
#define POSTSCRIPTDEV 70
#define SCREEN75	106

/*  This file is in two parts, old format stuff first, then new format. */

#define OLDMXGAMS		384	/* Number of generic attr. maps */

#define OLDMXCHARINDEX	256	/* Number of characters in a font */

#define OLDDSF_LEN		32	/* max length of "axfonts/..." file specifier */

/*
 *  The following numbers are multiplied by the renditions to get
 *  a generic attribute mapping index.  For example, if font 2, size 1,
 *  underlined was desired, its gam index would be 218 (2*96+24+2).
 *  The maximum possible index is 383 (one less than MXGAMS above).
 */
#define OLDFFONTOFFSET	96	/* Font attribute (0..3) */
#define OLDFSIZEOFFSET	24	/* Size attribute (0..3) */
#define OLDFSSOFFSET	8	/* Sub/superscript (0..2) */
#define OLDFIIOFFSET	4	/* Italics (0..1) */
#define OLDFUUOFFSET	2	/* Underline (0..1) */
#define OLDFBBOFFSET	1	/* Bold (0..1) */


/* globals */

struct devchar {
	short DispMWid;
	short DispMHt;
	short DispPWid;
	short DispPHt;
	short DefChMWid;
	short DefChMHt;
	short DefChPWid;
	short DefChPHt;
	short GraphCap;
	short DevProtocol;
	short PrivDLen;
};
/*
 *  The generic attribute map table consists of a header and a physical
 *  realization code for each of the 384 combinations of generic attributes.
 *  Each physical realization code has a font part and a private part.
 *  The header consists of a one word bit mask that can be used to
 *  get the font part.
 */
struct gamtab {
	short FontMask;
	short gams[OLDMXGAMS];
};

struct fontbuf {
	short NomChHt;			/* Nominal character height */
	short NomBase;			/* Baseline height */
	short NomChWid;			/* Nominal width */
	short LastChar;			/* Last valid character */
	short FontType;			/* 0 - mono; 2 - proportional */
	char  Widths[OLDMXCHARINDEX];	/* Bytes containing char widths */
};

/************************************************/
/* Following information is new in version 1.20 */
/************************************************/

#define DSIVER		2	/* Device setup information version 2 */

#define MXAFMS		1152	/* Number of attribute-font maps */

#define MXCHARINDEX	256	/* Number of characters in a font */

#define DSF_LEN		28	/* max length of file specifier */

/*
 *  The following numbers are multiplied by the renditions to get
 *  a font map index.  For example, if font 2, size 1, italic was
 *  desired, its index would be 206 (2*96+1*12+1*2).
 *  Possible indices range from 0 to 671 (one less than MXGAMS above).
 *
 *  Color, underline, capitalization, etc. don't appear here because
 *  they are always handled algorithmically and never effect the
 *  character widths.
 */
#define FFONTOFFSET	96	/* Font attribute (0..6) */
#define FSIZEOFFSET	12	/* Size attribute (0..7) */
#define FSSOFFSET	4	/* Sub/superscript (0..2) */
#define FIIOFFSET	2	/* Italics (0..1) */
#define FBBOFFSET	1	/* Bold (0..1) */

/*
 *  Use the following to determine the default local font number:
 *     (curSetup.FontMap[FMAPDEFAULT] & ALGFONTBITS)
 */
#define FMAPDEFAULT (FFONTOFFSET * 5 + FSIZEOFFSET)

/*
 *   The following values appear in the FontMap table when algorithmic
 *   methods should be used to get the corresponding attribute.  For
 *   example, if a bold font exists, it will be used as is, and the
 *   font map table entry would have the ALGBOLDBIT clear.  But if
 *   no bold font existed, the ALGBOLDBIT would be set.  This is done
 *   on an entry by entry basis because some devices will have bold
 *   fonts only in certain sizes, etc.
 *
 *   The top four bits are used here.  This leaves the bottom 12 bits
 *   for actual fonts.  This limits the number of fonts per device to
 *   4096, which is no problem since the FontMap table has only 672 entries.
 *   To get the actual font number, the local font number is used as
 *   an index in the local-global font array to get the global number.
 *   The global number is then used as an index in the font address array,
 *   which contains the file address of the actual font descriptor.
 *
 *   The algorithms for bolding and italicizing must not change the
 *   character advance.
 *
 *   The desired algorithm for subscripting and superscripting is to switch
 *   to the next smaller size (if possible) and to change the baseline
 *   so that subscripted text is centered on the old baseline and
 *   superscripted text is centered on the old cap height.  For this
 *   purpose, the center is halfway between the baseline and the cap
 *   height.
 *
 *   Unfortunately, we do not know the cap height, and neither do we
 *   want to switch fonts in midstream.  So the algorithm we really use
 *   is to switch fonts in the gam table itself, and to supply values
 *   to displace the baseline for subscripting and for superscripting.
 *   These values are determined from the small font only at present
 *   (January 87).
 */
#define ALGBITS         0xF000   /* Top four bits are used. */
#define ALGFONTBITS     0x0FFF   /* Bottom 12 bits for fonts */
#define ALGBOLDBIT      0x8000   /* Algorithmic bold */
#define ALGITALICBIT    0x4000   /* Algorithmic italicizing */
#define ALGSUPERBIT     0x2000   /* Algorithmic superscripting */
#define ALGSUBBIT       0x1000   /* Algorithmic subscripting */

/*
 *   Character set value used in the dsiSetup structure.
 */
#define CSET_ALIS       0   /* Alis-Adobe character set */
#define CSET_DEC        1   /* Dec Multinational */
#define CSET_NCR        2
#define CSET_HP         3   /* HP Roman-8 */
#define CSET_IBM        4

/*
 *  The following structures describe the devinfo file.
 *  The file has a header, then a set of pointers to setups, and then
 *  a set of pointers to font descriptors.  The pointers are absolute
 *  addresses in the file - they are longs, and can be used directly
 *  as an argument to fseek.  Following the two sets of pointers are
 *  the dsiSetup and dsiFontDesc structures, in no particular order.
 *
 *  Following each dsiSetup structure is a array of the global font
 *  numbers used in that setup.  They are indexed by the device's
 *  local font number (the entry in the FontMap array).  The entries
 *  are shorts.
 *
 *  The text and graphic unit size values should be used for conversion.
 *  They will often be more accurate than the display or cell sizes.
 *
 *  The graphic unit size will differ when the graphics resolution
 *  is different than the text resolution.  This can be the case on the
 *  HP LaserJet, where graphics resolution can be 75, 100, 150, or
 *  300 dpi.  Graphics resolution is mainly of interest to programs
 *  that use images since neither the atm nor the pf provide scaling.
 *
 *  The font map table consists of a font number for each
 *  of the 672 possible combinations of generic attributes.
 *
 *  The font description contains the pixel widths of the characters
 *  and the file name of the font.  It contains the height of the font,
 *  the baseline distance, and the cap height.  The height of a font is
 *  the distance from the top of the tallest character to the bottom
 *  of the lowest descender, plus any leading.  The baseline distance
 *  is the distance from the top of the font to the baseline.  The cap
 *  height is the distance from the baseline to the top of a capital
 *  letter.  The baseline distance could be the same as the cap height
 *  if no letter extends above the top of a capital letter and if no
 *  leading is built in above the font.  Accented capital letters
 *  often extend above the cap height.
 *
 *  The cap height is not currently used in algorithmic sub and super-
 *  scripting, although that is why it is included here.  Instead,
 *  two additional values are provided which are used to offset the
 *  baseline of the font when doing sub or superscripting. (Jan 87)
 *
 *  Partial path names are treated as starting in $AOSDIR/axfonts.
 *  If it's a non turbo version, it will check in curdir/axfonts first.
 *  Full path names (those starting with "/") are used as is.
 *  
 */

struct dsiHeader {
	short dsiVersion;
	short numSetups;
	short numFonts;
};

/*
 *  Following the dsiHeader are two arrays.  The first array contains
 *  the file addresses (longs) of the dsiSetup structures in the file,
 *  indexed by setup number.  The second array contains the file addresses
 *  of the dsiFontDesc structures in the file.
 */

struct dsiSetup {
	short DispMWid;        /* Display size in mils */
	short DispMHt;
	short TextUMWid;       /* Text unit size in mils and pixels */
	short TextUMHt;
	short TextUPWid;
	short TextUPHt;
	short GrafUMWid;       /* Graphics unit size in mils and pixels */
	short GrafUMHt;
	short GrafUPWid;
	short GrafUPHt;
	short GrafCap;         /* Graphics capability (RGMODE, etc) */
	short GrafExt;         /* Extra bits.  TM_SLOWGRFOPS lives here. */
	short CharSet;         /* Base character set of device */
	short MaxFont;         /* Maximum local font number. */
	short CellLeading;     /* Leading added to default font for cell */
	short FontMap[MXAFMS]; /* Fonts for each attribute combination */
	short MonoCellGpix;    /* Monomode cell wid in gpix */
	short MonoFace;        /* Monomode fontface attval (a la atmattr.h) */
	char apcap[DSF_LEN];   /* null terminated apcap file spec */
};

/*
 *  Following each dsiSetup is an array of global font numbers (shorts).
 *  The array has an entry for each local font number.
 */

struct dsiFontDesc {
	short NomChHt;             /* Nominal character height (MUST BE FIRST) */
	short NomBase;             /* Baseline distance, or ascent */
	short NomCapHeight;        /* Height of capital letters */
	short SubOffset;           /* Subscripting offset (usually positive) */
	short SupOffset;           /* Superscripting offset (usually negative) */
	short NomChWid;            /* Nominal width */
	unsigned char Widths[MXCHARINDEX];
	                           /* Bytes containing char widths */
#if MIPS || RIOS || SGI
	signed char LBearing[MXCHARINDEX];
                                   /* Bytes containing left bearings */
	signed char RBearing[MXCHARINDEX];
                                   /* Bytes containing right bearings */
	signed char FontPath[DSF_LEN];
                                   /* null terminated font file spec, or name */
#else
	char LBearing[MXCHARINDEX];
	                           /* Bytes containing left bearings */
	char RBearing[MXCHARINDEX];
	                           /* Bytes containing right bearings */
	char FontPath[DSF_LEN];    /* null terminated font file spec, or name */
#endif
};
/*
 * createflag values for TmGetXFontName
 */
#define	CREATEFONT			1
#define	SELECTFONT			0
