/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: egcd.h
   Applix ASC version: 3
   Date: Mon Nov 16 13:53:46 1998
*/
#ifndef EGCD_H_INCLUDED

#define EGCD_H_INCLUDED 1


typedef struct wi {
    unsigned char    wisegno;           /* segment number */
    unsigned char    wiseqno;           /* kanji number */
    unsigned short   wirecno;           /* yomi number */
}   WORDID;

typedef struct tagCB {
    unsigned char    acbcmd;     /* HDML command code                    */
    unsigned char    acbdir;     /* HDML sub command code                */
    unsigned short   acbrsp;     /* HDML errror code                     */

    unsigned char    acbsmd;     /* system mode                          */
    unsigned char    acbimd;     /* input mode                           */

    unsigned short   acbtpn;     /* total phrase number                  */
    unsigned short   acbcpn;     /* current phrase number                */
    unsigned short   acbcrs;     /* cursor start position                */
    unsigned short   acbcre;     /* cursor end position                  */
    unsigned short   acbrmj;     /* romaji start position                */
    unsigned short   acbswd;     /* screen width                         */
    unsigned short   acbtpg;     /* total page number                    */
    unsigned short   acbcpg;     /* current page number                  */
    unsigned short   acbspg;     /* homonym page size                    */
    unsigned short   acbchm;     /* current homonym number               */
    unsigned short   acbhyl;     /* homonym yomi length                  */
    unsigned short   acbmik;     /* max  input key                       */

    unsigned short  *acbebp;     /* pointer to element buffer            */
    unsigned char   *acbsbp;     /* pointer to screen buffer             */
    unsigned char   *acbobp;     /* pointer to origenal buffer           */
    unsigned char   *acbhpp;     /* pointer to homonym page buffer       */
    unsigned char   *acbhep;     /* pointer to homonym element buffer    */
    unsigned char   *acbhbp;     /* pointer to homonym buffer            */
    WORDID          *acbwid;     /* pointer to homonym word id. buffer   */
    unsigned char   *acbhym;     /* pointer to homonym yomi buffer       */
}   CB,*ACB;


/*   acbsmd  definition  */
#define     CSMCVT      0x80    /* 1:convert 0:hommony  */
#define     CSMCEL      0x40    /* 1:convert 0:no-conv. */
#define     CSMAGE      0x20    /* age flag             */

/*   acbcmd  definition  */
#define     CMOVE       1       /* cursor move          */
#define     RVCVT       2       /* recverse convert     */
#define     TXCVT       3       /* kanatokanji convert  */
#define     CHCVT       4       /* change phrase        */
#define     PHFIX       5       /* phrase fix           */
#define     AGEHOM      6       /* age homonym          */
#define     CLEAR       7       /* clear control block  */
#define     CDCVT       8       /* code convert         */
#define     CANHOM      9       /* cancel homonym       */
#define     NEXTHOM     10      /* next homonym         */ 
#define     PREVHOM     11      /* previous homonym     */
#define     CLLEARN     12      /* cl learn             */

/*   acbimd  definition  */
#define     IMDJCD      0x80    /* JIS code     */
#define     IMDRMJ      0x40    /* romaji in    */
#define     IMDZEN      0x20    /* zen kaku     */
#define     IMDHRC      0x10    /* hira kana    */

/*   acbdir definition  */
#define     DRLEFT      0       /* left         */
#define     DRRIGHT     1       /* right        */
#define     DRDOWN      2       /* down         */
#define     DRUP        3       /* up           */
#define     DRHOME      4       /* home         */
#define     DRLAST      5       /* last         */

typedef struct _EGC_JTABLE {
    int sel_no;
    int sel_jis[10];
    int sel_char[10];
    unsigned char start_char[10];
    unsigned char start_jis;
    CB cb;
    } EGC_JTABLE;


typedef struct _EGC_JBUF {
    int hlwidth;
    int hlstart;
    int kanji_buf_len;
    unsigned short kanji_buf[1024];
    } EGC_JBUF;


#define EGC_CONVNEXT_KEY	XK_space
#define EGC_SPACE_KEY		XK_space
#define EGC_LEFT_KEY		XK_Left
#define EGC_RIGHT_KEY		XK_Right
#define EGC_UP_KEY    		XK_Up
#define EGC_DOWN_KEY		XK_Down
#define EGC_CONVERT_KEY		XK_Kanji
#define EGC_BACKSPACE_KEY	XK_BackSpace
#define EGC_DELETE_KEY		XK_Delete
#define EGC_ESCAPE_KEY		XK_Escape
#define EGC_COMMIT_KEY		XK_Execute
#define EGC_RETURN_KEY		XK_Return
#define EGC_ENTER_KEY		XK_KP_Enter


#define EGC_LEFT            1
#define EGC_RIGHT           2
#define EGC_UP              3
#define EGC_DOWN            4


#define EGC_CONTINUE        1
#define EGC_COMMIT          2
#define EGC_KANAMODE_CHANGE 3
#define EGC_NOTHING         4


#endif
