/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: fontcach.h
   Applix ASC version: 8
   Date: Mon Nov 16 13:53:33 1998
*/
#ifndef _FONTCACHE_H
#define _FONTCACHE_H

/* operation types FIND, LOAD, LOOKUP; their rules:
 *
 *          item is found              not found
 *          -------------              ---------
 * FIND     return data                call fetch, add the item
 * LOAD     overwrite data             add item
 * LOOKUP   return data                return false
 */
#define FIND 0
#define LOAD 1
#define LOOKUP 2  /* 'fetchless' find */

	/* struct for declaring functions for maintaining a datatype */
typedef struct typeFunc_t {
	int (*fetch)();
	int (*copy)();
	int (*toss)();
} typeFuncs;

#define NUM_LISTS 14		/* # types so far */
#define FONT_TABLE_WIDTH 20 /* hash table width */
#define ENCODING_TABLE_WIDTH 256
#define MAXLRU (0x7fffffff)
#define MAXSHORT16 (0x7fff)

	/* types */
#define CACHELISTS 0	/* cache list names paired with index numbers */
#define	TARGET_METRICS 1 /* printer metrics disguised as XFontStruct */
#define	SCREEN_METRICS 2 /* XFontStruct */
#define SCREENFONT_METRICS 3   /* for app screen text measurements */
#define TARGETFONT_METRICS 4   /* for app printer text measurements */
#define	LETTERFORM 5	/* chacter XImage */
#define FONTFAMILY 6	/* family name and the font names of its members */
#define FONTRECORD 7	/* data for a particular font */
#define ENCODING 8	/* encoding file names paired with encoding name strs */
#define LATIN1	9	/* the rest are character names paired with codes */
#define FONTSPECIFIC 10
#define AXBATS 11
#define ADOBESTANDARD 12
#define REFERENCE_CELL 13  /* monospace XFontStruct; metrics only */

/*
 * the limits should be liberal, to
 * reduce the likelihood that the limit
 * is hit before the hash table fills.
 * In a simple example (one where the item
 * sizes are fixed and known beforehand),
 * use a minimum limit something like:
 * (LIST_SIZE * <item size> * 2)
 *
 * Some examples of practical estimates:
 * latin screen/target metrics table: ~3k bytes
 * ksc5601 screen/target metrics table:
 * ~9000 per-char structs, or ~100k bytes.
 * letterform: ~1000 bytes is big....
 */
#define TARGETLIMITFACTOR (9000) /* this could be smaller (re: BBox) */
#define SCREENLIMITFACTOR (9000)
#define LETTERLIMITFACTOR (2000)
#define FONTFAMILYFACTOR (500)
#define FONTRECORDFACTOR (2000)

/* use a limit corresponding to 256*(20 chars),
 * a good test for the evenness of distribution
 * of the shifted-xor hash applied to adobe
 * glyphname strings (where 20 chars mean is
 * assumed to be liberal) */
#define LATINLIMITFACTOR (20)

/* a setting for the XCharStruct attributes field,
 * when using the struct for target metrics */
#define UNUSED ((unsigned short)0x8000);

/* the hardcoded xlfd strings for Times-Roman screenfonts */
#define AXDEFTIMES8 \
  NXLT("-adobe-times-medium-r-normal--8-80-75-75-p-44-iso8859-1")
#define AXDEFTIMES10 \
  NXLT("-adobe-times-medium-r-normal--10-100-75-75-p-54-iso8859-1")
#define AXDEFTIMES12 \
  NXLT("-adobe-times-medium-r-normal--12-120-75-75-p-64-iso8859-1")
#define AXDEFTIMES14 \
  NXLT("-adobe-times-medium-r-normal--14-140-75-75-p-74-iso8859-1")
#define AXDEFTIMES18 \
  NXLT("-adobe-times-medium-r-normal--18-180-75-75-p-94-iso8859-1")
#define AXDEFTIMES24 \
  NXLT("-adobe-times-medium-r-normal--24-240-75-75-p-124-iso8859-1")

#endif /*  _FONTCACHE_H  */

