/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: fontset.c
   Applix ASC version: 4
   Date: Mon Nov 16 13:53:47 1998
*/
#if BUILD_ORIENT    /*  {  */

#include <basicwin.h>

#include <aos.h>
#include <xim.h>
#include <wchars.h>
#include <ximcomm.h>

extern Display *Dpy;


typedef struct font_name_struct {
  int font_height;
  char *font_name;
  }  font_name_struct;

FontSet FontInfoFromIdList;

mbchar *sansserif_orient_face();
mbchar *serif_orient_face();
mbchar *sansserif_narrow_kana_face();
mbchar *getXFontFace();

#if !FOR_FILTER
extern XFontStruct *AxDlgMainFont;      /* basic dialog font */
#endif

/**************************************************************************/
/* for a given font-set, look through our cache and return the font struct
* for an element in the font set.  Some of these elements may be null
* since they have not yet been referenced.
* If a request is made for a font in the set which has not yet been loaded,
*   that font will be loaded.  */
/**************************************************************************/

XFontStruct *FontSetFontStruct( font_set, font_request )
FontSet font_set;
int font_request;
{
    XFontStruct *ret_font;
    if ( font_request == WIDE_FONT )
       {
       if ( font_set->wide_font_id == (Font)NULL )
          {
          /* wide font has not yet been loaded, load the default */
          GetWideFontFromSize( font_set );
          }
       if ( font_set->wide_font == (XFontStruct *)NULL )
          font_set->wide_font = XQueryFont( Dpy, font_set->wide_font_id );
       ret_font = font_set->wide_font;
       }
    if ( font_request == NARROW_FONT )
       {
       if ( font_set->narrow_font_id == (Font)NULL )
          {
          /* narrow font has not yet been loaded, load the default */
          GetNarrowFontFromSize( font_set );
          }
       if ( font_set->narrow_font  == (XFontStruct *)NULL )
          font_set->narrow_font = XQueryFont( Dpy, font_set->narrow_font_id );
       ret_font = font_set->narrow_font;
       }
#if BUILD_KANJI || HYUNDAI
    if ( font_request == NARROW_KANA_FONT )
       {
       if ( font_set->narrow_kana_id == (Font)NULL )
          {
          /* narrow kana has not yet been loaded, load the default */
          GetNarrowKanaFontFromSize( font_set );
          }
       if ( font_set->narrow_kana  == (XFontStruct *)NULL )
          font_set->narrow_kana = XQueryFont( Dpy, font_set->narrow_kana_id );
       ret_font = font_set->narrow_kana;
       }
#endif

    if ( ret_font == (XFontStruct *)NULL )
       {
       if ( font_set->narrow_font_id != (Font)NULL )
          /* first level fall-back is to return the western font */
          ret_font = font_set->narrow_font;
#if !FOR_FILTER
       else 
          /* as a really bogus fallback, return the dialog font */
          ret_font = AxDlgMainFont;
#endif
       }

    return ( ret_font );
}

/**************************************************************************/
/* for a given font-set, look through our cache and return the font-id 
* for an element in the font set.  Some of these elements may be null
* since they have not yet been referenced.
* If a request is made for a font in the set which has not yet been loaded,
*   that font will be loaded.  */
/**************************************************************************/

Font FontSetFontId( font_set, font_request )
FontSet font_set;
int font_request;
{
    Font default_font;
    Font return_font;

    default_font = (Font)0;
    if ( font_set->narrow_font_id != (Font)0 )
       /* use the standard western font as the default */
       default_font = font_set->narrow_font_id;

    if ( font_request == WIDE_FONT )
       {
       if ( font_set->wide_font_id == (Font)NULL )
          {
          /* wide font has not yet been loaded, load the default */
          GetWideFontFromSize( font_set );
          }
       return_font = font_set->wide_font_id;
       }
    if ( font_request == NARROW_FONT )
       {
       if ( font_set->narrow_font_id == (Font)NULL )
          {
          /* narrow font has not yet been loaded, load the default */
          GetNarrowFontFromSize(  font_set );
          }
       return_font = font_set->narrow_font_id;
       }
#if BUILD_KANJI || HYUNDAI
    if ( font_request == NARROW_KANA_FONT )
       {
       if ( font_set->narrow_kana_id == (Font)NULL )
          {
          /* narrow kana has not yet been loaded, load the default */
          GetNarrowKanaFontFromSize( font_set );
          }
       return_font = font_set->narrow_kana_id;
       }
#endif

    if ( return_font == (Font)0 )
       return( default_font );
    else
       return( return_font );
}

/**************************************************************************/
/* Given an font-id, find the corresponding font set and return the font
*    set pointer, the type of the font-id which was input.  If there is
*    no font set which contains the requested font-id, then a new font set
*    will be created with thes element.  If a font set is created, the
*    third argument "font" may point to an XFontStruct which will be used
*    to determine the type of the font.  If this is NULL, then an 
*    XQueryFont call will be made to get the info (this may be expensive)   */
/**************************************************************************/

FontSet FindFontSet( fid, type, font )
Font fid;
int *type;         /* OUT: one of NARROW_FONT, WIDE_FONT, NARROW_KANA_FONT */
XFontStruct *font;     /* supplied by caller or null if we don't have it */
{
    FontSet info;
    static FontSet last_info;
    static int last_type = -1;
    static Font last_fid = -1;
    int     point, sans;

    if ( fid == last_fid )
       {
       *type = last_type;
       return( last_info );    /* the quick route */
       }

    info = FontInfoFromIdList;
    
    while ( info != (FontSet )NULL )
        {
        if ( info->narrow_font_id == fid )
            {
            *type = NARROW_FONT;
            break;
            }
        if ( info->wide_font_id == fid )
            {
            *type = WIDE_FONT;
            break;
            }
        if ( info->narrow_kana_id == fid )
            {
            *type = NARROW_KANA_FONT;
            break;
            }
        info = info->next;
        }


#if BUILD_KANJI         /*  {  */
	/* Ashisuto added to find the UNIUQE fontset. Since wide fonts and narrow
	 * fonts are not unique, two fontsets may have same narrow font, but with
	 * different wide fonts, and vice versa. This plurality has confused us
	 * in Spreadsheets' cells: There is only one font choice for each cell, 
	 * if I change font attribute among Narrow Fonts and Wide Fonts at the
	 * same cell with both English and Japanese chars included, the result 
	 * fontset will be un-predictable. It is even worse that the print out 
	 * result will different from screen displaying. This has confused us for
	 * a long while. To setup the unique wide-narrow font correspondence for
	 * look up font set, we must decide the corresponding mapping from wide
	 * to narrow or from narrow to wide. Following we try our best to make out
	 * the unique correspondence. 
	 * 1995.11.07
	 */
	if (info) {
		FontSet run_info = info;
		FontSet better_info = NULL;
		mbchar	*face, *default_face;

		while (run_info != NULL) {
			face = NULL;
			switch (*type) {
				case NARROW_FONT:
					if (run_info->narrow_font_id == fid) {
						if (run_info->wide_font) {
							if (run_info->sans)
								default_face = sansserif_orient_face();
							else
								default_face = serif_orient_face();
							face = getXFontFace(run_info->wide_font);
						} else if (run_info->narrow_kana) {
							if (run_info->sans)
								default_face = sansserif_narrow_kana_face();
							else
								default_face = serif_narrow_kana_face();
							face = getXFontFace(run_info->narrow_kana);
						} else {	/* only one font in fontset */
							better_info = run_info;
						}
					}		
					break;
				case WIDE_FONT:
					if (run_info->wide_font_id == fid) {
						if (run_info->narrow_font) {
							if (run_info->sans)
								default_face = axDefaultSansSerifFont;
							else
								default_face = axDefaultSerifFont;
							face = getXFontFace(run_info->narrow_font);
						} else {	/* no narrow font in fontset */
							better_info = run_info;
						}
					}		
					break;
				case NARROW_KANA_FONT:
					if (run_info->narrow_kana_id == fid) {
						if (run_info->narrow_font) {
							if (run_info->sans)
								default_face = axDefaultSansSerifFont;
							else
								default_face = axDefaultSerifFont;
							face = getXFontFace(run_info->narrow_font);
						} else {	/* no narrow font in fontset */
							better_info = run_info;
						}
					}
					break;
				default:
					break;
			}

			if (face && (strcmp(face, default_face) == 0) )	/* Best fontset ! */
				break;

			run_info = run_info->next;
		}
		if (run_info)
			info = run_info;	/* The best. (Unique correspondence) */
		else if (better_info)
			info = better_info;	/* The un-filled. (Also unique) */
		else
			info = NULL;		/* No unique corresponding. Shall create one */
	}					
#endif       /*  }  BUILD_KANJI  */


    if ( info != (FontSet )NULL )
        {
        /* we found it */
        last_fid = fid;
        last_info = info;
        last_type = *type;
        return( info );
        }

    /*fell through, did not find it, make a new entry */
    if ( font == (XFontStruct *)NULL )
        {
        /* take the long route by quering the server */
        /* and then add a new element at the head of the list */
        /* to remember this for next time                     */
        font = XQueryFont( Dpy, fid );

        if ( font == (XFontStruct *)NULL )
           {
           last_info = (FontSet )NULL;
           last_fid = -1;
           last_type = -1;
           return( last_info );          /* non-graceful bail out */
           }
        }

    info = (FontSet )malloc( sizeof(font_info_struct) );
    memset( info, 0, sizeof(font_info_struct) );

    info->next = FontInfoFromIdList;
    FontInfoFromIdList = info;

    /* the size is put in here only as a fail safem it should normally */
    /*  be overwritten at an upper level which has the full XLFD */
    info->height = font->ascent + font->descent;
    if (getXFontAttr(font, NULL, &point, NULL, NULL, &sans)) {
        info->pointsize = point;
        info->sans = sans;
    }


    if ( font->max_byte1 > 0 )
       {
       info->wide_font_id = font->fid;
       info->wide_font = font;
       *type = WIDE_FONT;
       }
    else
       {
       info->narrow_font_id = font->fid;
       info->narrow_font = font;
       *type = NARROW_FONT;
       }
    last_fid = fid;
    last_info = info;
    last_type = *type;
    return( info );
}


#endif   /*  } BUILD_ORIENT  */

