/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: inoutstr.c
   Applix ASC version: 63
   Date: Wed Apr 14 21:14:18 1999
*/
#include <aos.h>
#include <afile.h>
#include <doctypes.h>
#include <wchars.h>
#include <getstr10.h>
#include <revstamp.h>
#include <inoutstr.p>

void AGetDefaultFileInfo();
mbchar *ADocTypeTokFromCode();



#if !BUILD_WIDE
/** these functions always use wide characters **/
#ifdef wchar
#undef wchar
#endif

#define wchar unsigned short
#endif




/* Needs to be defined here for SS filters */
#ifndef streq
#define streq(x,y) !strcmp(x,y)
#endif

/* Processing of Asterx Standard File records
    When reading, use:
        AFileInfoFromHeader(...&info) - decode header into file info
        NormalizeIn(...&info) - to decode lines 
    When writing, use:
        AGetDefaultFileInfo(&info) - get default values for file header
        info.docType = whatever.
        AFileHeaderFromInfo(...&info) - to construct a header line
        NormalizeOut(...&info) - to encode lines 
        AFileEnderFromInfo(...&info) - to construct a file ender line

*/

#ifndef MBNXLT
#define MBNXLT(a) (mbchar *)a
#endif


/* Standard Tokens in Asterx Files */
#define ATOK_NONE MBNXLT("NONE")
#define ATOK_7BIT MBNXLT("7BIT")
#define ATOK_VERSION MBNXLT("VERSION")
#define ATOK_ENCODING MBNXLT("ENCODING")
#define ATOK_ENCRYPTED MBNXLT("ENCRYPTED")
#define ATOK_ON MBNXLT("ON")
#define ATOK_OFF MBNXLT("OFF")
#define ATOK_WORDS MBNXLT("WORDS")
#define ATOK_GRAPHICS MBNXLT("GRAPHICS")
#define ATOK_BITMAP MBNXLT("RASTER")
#define ATOK_SPREADSHEETS MBNXLT("SPREADSHEETS")
#define ATOK_EQUATIONS MBNXLT("EQUATIONS")
#define ATOK_MACROS MBNXLT("MACROS")
#define ATOK_ASCII MBNXLT("ASCII")
#define ATOK_ELFARRAY MBNXLT("ELFARRAY")
#define ATOK_ELFTABLE MBNXLT("ELFTABLE")
#define ATOK_QUERYDATA MBNXLT("QUERYDATA")
#define ATOK_BINARY MBNXLT("BINARY")
#define ATOK_HEX MBNXLT("HEX")
#define ATOK_BUILDER MBNXLT("BUILDER")
#define ATOK_OBJ_CLASS MBNXLT("OBJCLASS")
#define ATOK_BUILDER_TURBO MBNXLT("BUILDER-TURBO")
#define ATOK_BUILDER_ABD MBNXLT("BUILDER-DISTRIBUTION")
#define ATOK_MENUBAR MBNXLT("MENUBAR")
#define ATOK_CONTENT MBNXLT("C")
#define ATOK_VVAR1 MBNXLT(" %s=%d/%d/%d")
#define ATOK_VVAR2 MBNXLT(" %s=%d/%d")
#define ATOK_CVARIABLE MBNXLT(" %s=%s")
#define ATOK_SLASH MBNXLT("/")
#define SPACE MBNXLT(" ")
#define EQUALS MBNXLT("=")

extern int AFTaskLockFile();
extern FILE * AFTaskfopen();
extern int AFfclose();
extern char *TaskAlloc();

/***************************************************************************
function: AGetCurrentVersion - get current version number
*/
int AGetCurrentVersion()
{
	return(BUILD_REVISION_100_FMT);
}

/***************************************************************************
function: ADocIsCurRev - given a file, return TRUE if the document was
	saved in current rev, else return FALSE
*/
int ADocIsCurRev(pathname)
	char *pathname;
{
	char headerLine[AREADBUFSIZE+1];
	FILE *fp;
	AFileInfo revInfo;
	int retval = FALSE;

	fp = fopen(pathname, READ_STR);
	if(NULL == fp)	
		return(FALSE);

	if(NULL == fgets(headerLine, AREADBUFSIZE, fp))	
		retval = FALSE;
	else if (AOK != AFileInfoFromHeader(headerLine, &revInfo))	
		retval = FALSE;	
	else if(revInfo.version != AGetCurrentVersion())	
		retval = FALSE;
	else
		retval = TRUE;
	fclose(fp);
	return(retval);
}

/***************************************************************************
function: AGetDefaultFileInfo - get default values for file header, including
current version number.
*/
void AGetDefaultFileInfo(info_ptr)
AFileInfo *info_ptr;    /* out: default info */
{
    /* initialize info to default values */
    info_ptr->encoding = 1; /* assume file IS encoded */
    info_ptr->version = info_ptr->old_version = AGetCurrentVersion();
    info_ptr->min_version = 320;
    info_ptr->docType = AX_DOC_TYPE_DATA; /* simply a data file */
    info_ptr->content = 0; /* none */
    info_ptr->encrypted = 0; /* not encrypted */
    return;
}

/***************************************************************************
function: AFileInfoFromHeader - recognize and process asterx file header.
Given first line of a file, determine if it is an Asterx standard file
header, and if so, get the info in the header and return TRUE, If it is
not, set header to default values and return FALSE. This determines the
character encoding used in the file.
Returns:
    AOK
    FENoHeader  - line is not a Applix header line
    FETooNew    - current software rev is less than min_version in header
    FEInvalidHeader - some mistake in header.
*/
typeErr AFileInfoFromHeader(headerLine, info_ptr)
mbchar *headerLine; /* in: first line of file, null terminated,
                        with or without newline; or NULL, to get defaults. */
AFileInfo *info_ptr;    /* out: info about this file */
{
    register mbchar *tok, *name, *value, *subval;
    int version, current_software_version, i;
    static mbchar line[ALINELENGTH+1];
    mbchar *exp[4];
    typeErr status;

    /* init (assume no encoding for reading) */
    status = AOK;
    AGetDefaultFileInfo(info_ptr);
    current_software_version = info_ptr->version; /* save this from defaults */
    info_ptr->encoding = 0;

    /* parse the string */
    if ( !headerLine )
        status = FENoHeader;
    else
    {
        /* get the first "token" from the string (*START or *BEGIN) */
        MBstrncpy(line, headerLine, ALINELENGTH);
        line[ALINELENGTH]=0;
        tok = MBstrtok(line, ATOK_SEPARATORS);
        if ( !tok )
            status = FENoHeader; /* not a header at all. */
        else
        {
            /* rev 1.x/2.x header */
            value = NXLT("2278791755874641370");
            if ( MBstrncmp(tok, value, strlen(value)) == 0 )
                version = 2;

            /* rev 3.x header? */
            else if ( MBstreq(tok, ATOK_OLDSTART) )
                version = 3;

            /* rev 4.x header */
            else if ( MBstreq(tok, ATOK_START) )
                version = 4;

            else /* none of the above - return an error */
            {
                if ( MBstrncmp(tok, ATOK_ASTERIX, strlen(ATOK_ASTERIX)) )
                    status = FENoHeader;      /* no header, no asterisk */
                else
                    status = FEInvalidHeader; /* starts with an asterisk */
            }
        }

        if ( status == AOK && version == 2 ) /* very old file */
        {
            info_ptr->version     = 210;
            info_ptr->old_version = 210;
            info_ptr->min_version = 210;

            /* get the second "token" from the string (machine type) */
            tok = MBstrtok((mbchar *)0, ATOK_SEPARATORS);

            /* get the third "token" from the string (document type) */
            tok = MBstrtok((mbchar *)0, ATOK_SEPARATORS);
            if ( tok )
            {
                if (MBstreq(tok, NXLT("2")) )
                    info_ptr->docType = AX_DOC_TYPE_ASTERX_SS;
                else if (MBstreq(tok, NXLT("5")) )
                    info_ptr->docType = AX_DOC_TYPE_ASTERX_GRAPHICS;
                else if (MBstreq(tok, NXLT("6")) )
                    info_ptr->docType = AX_DOC_TYPE_ASTERX_WORDS;
                else if (MBstreq(tok, NXLT("8")) )
                    info_ptr->docType = AX_DOC_TYPE_ASTERX_MACROS;
            }
        }

        else if ( status == AOK ) /* rev 3.x or greater */
        {
            /* get the second "token" from the string (document type) */
            tok = MBstrtok((mbchar *)0, ATOK_SEPARATORS);
            if ( tok )
            {
                if (MBstreq(tok, ATOK_WORDS))
                    info_ptr->docType = AX_DOC_TYPE_ASTERX_WORDS;
                else if (MBstreq(tok, ATOK_GRAPHICS))
                    info_ptr->docType = AX_DOC_TYPE_ASTERX_GRAPHICS;
                else if (MBstreq(tok, ATOK_SPREADSHEETS))
                    info_ptr->docType = AX_DOC_TYPE_ASTERX_SS;
                else if (MBstreq(tok, ATOK_MACROS))
                    info_ptr->docType = AX_DOC_TYPE_ASTERX_MACROS;
                else if (MBstreq(tok, ATOK_EQUATIONS))
                    info_ptr->docType = AX_DOC_TYPE_ASTERX_EQUATIONS;
                else if (MBstreq(tok, ATOK_ASCII))
                    info_ptr->docType = AX_DOC_TYPE_ASCII;
                else if (MBstreq(tok, ATOK_ELFARRAY))
                    info_ptr->docType = AX_DOC_TYPE_ASTERX_ELFARRAY;
                else if (MBstreq(tok, ATOK_ELFTABLE))
                    info_ptr->docType = AX_DOC_TYPE_ASTERX_ELF_TABLE;
                else if (MBstreq(tok, ATOK_QUERYDATA))
                    info_ptr->docType = AX_DOC_TYPE_ASTERX_QUERYDATA;
                else if (MBstreq(tok, ATOK_BUILDER))
                    info_ptr->docType = AX_DOC_TYPE_ASTERX_BUILDER;
                else if (MBstreq(tok, ATOK_BUILDER_TURBO))
                    info_ptr->docType = AX_DOC_TYPE_ASTERX_BUILDER_TURBO;
                else if (MBstreq(tok, ATOK_BUILDER_ABD))
                    info_ptr->docType = AX_DOC_TYPE_ASTERX_BUILDER_ABD;
                else if (MBstreq(tok, ATOK_OBJ_CLASS))
                    info_ptr->docType = AX_DOC_TYPE_ASTERX_OBJ_CLASS;
                else if (MBstreq(tok, ATOK_BITMAP))
                    info_ptr->docType = AX_DOC_TYPE_ASTERX_BITMAP;
                else if (MBstreq(tok, ATOK_HEX))
                    info_ptr->docType = AX_DOC_TYPE_HEX;
                else if (MBstreq(tok, ATOK_MENUBAR))
                    info_ptr->docType = AX_DOC_TYPE_ASTERX_MENUBAR;
            }

            /* there may be 4 other arg pairs (name=value) */
            exp[0] = MBstrtok((mbchar *)0, ATOK_SEPARATORS);
            exp[1] = MBstrtok((mbchar *)0, ATOK_SEPARATORS);
            exp[2] = MBstrtok((mbchar *)0, ATOK_SEPARATORS);
            exp[3] = MBstrtok((mbchar *)0, ATOK_SEPARATORS);
            for ( i=0; i<4; i++ )
            {
                if ( !exp[i] )
                    continue;

                name  = MBstrtok(exp[i], EQUALS);
                value = MBstrtok((mbchar *)0, ATOK_SEPARATORS);
                if ( value )
                {
                    if ( MBstreq(name, ATOK_CONTENT) )
                        info_ptr->content = (char *)value;
                    else if ( MBstreq(name, ATOK_ENCODING) )
                    {
                        if ( MBstreq(value, ATOK_7BIT) )
                            info_ptr->encoding = 1;
                    }
                    else if ( MBstreq(name, ATOK_ENCRYPTED) )
                    {
                    	if ( MBstreq(value, ATOK_ON) )
                    		info_ptr->encrypted = 1;
                    }
                    else if ( MBstreq(name, ATOK_VERSION) )
                    {
						/* 'Old' V3 files will only have a single number
							in the VERSION= token, but filtered V3 files
							may have a triple; e.g. VERSION=311/300/400
							for a file run thru a 4.0 filter */
						subval = MBstrtok(value, ATOK_SLASH); 
						if ( subval ) 
							info_ptr->version = atoi(subval);

						/* min version */ 
						subval = MBstrtok((mbchar *)0, ATOK_SLASH); 
						if ( subval ) 
							info_ptr->min_version = atoi(subval); 
						else 
							info_ptr->min_version = info_ptr->version;

						/* original version (before filtering) */ 
						subval = MBstrtok((mbchar *)0, ATOK_SLASH); 
						if ( subval ) 
							info_ptr->old_version = atoi(subval); 
						else 
							info_ptr->old_version = info_ptr->version;
                    } /* VERSION */
                }
            } /* for */
        
            /* check minimum version */
            if ( info_ptr->min_version > current_software_version )
                status = FETooNew;
        } /* v3 or later */
    }
    return(status);
}

/***************************************************************************
function: AFileHeaderFromInfo - to construct a line for use as a file header.
*/
void AFileHeaderFromInfo(info_ptr, header)
AFileInfo *info_ptr;/* in: info for header */
mbchar *header;     /* out: buffer for header line */
{
    mbchar *docTypeTok;
    int len;

    /* makes something like: "*BEGIN WORDS VERSION=400/320 ENCODING=7BIT" */

    /* *BEGIN WORDS... */
    strcpy(header, ATOK_START);
    strcat(header, SPACE);
    docTypeTok = ADocTypeTokFromCode(info_ptr->docType);
    strcat(header, docTypeTok);

    /* ...VERSION=400/320... */
    len = MBstrlen(header);
    if ( info_ptr->old_version && info_ptr->old_version != info_ptr->version )
    {
        sprintf(&header[len], ATOK_VVAR1, ATOK_VERSION,
                info_ptr->version,
                info_ptr->min_version,
                info_ptr->old_version);
    }
    else
    {
        sprintf(&header[len], ATOK_VVAR2, ATOK_VERSION,
                info_ptr->version,
                info_ptr->min_version);
    }

    /* ...ENCODING=7BIT... */
    if ( info_ptr->encoding )
    {
        len = MBstrlen(header);
        sprintf(&header[len], ATOK_CVARIABLE, ATOK_ENCODING, ATOK_7BIT);
    }
    else
    {
        len = MBstrlen(header);
        sprintf(&header[len], ATOK_CVARIABLE, ATOK_ENCODING, ATOK_NONE);
    }

    /* ...C=???... */
    if ( info_ptr->content )
    {
        len = MBstrlen(header);
        if (len + 2 + MBstrlen(ATOK_CONTENT) + MBstrlen(info_ptr->content)
                < ALINELENGTH )
        sprintf(&header[len], ATOK_CVARIABLE, ATOK_CONTENT, info_ptr->content);
    }

    /* ...ENCRYPTED=ON... */
	if ( info_ptr->encrypted )
	{
		len = MBstrlen(header);
		sprintf(&header[len], ATOK_CVARIABLE, ATOK_ENCRYPTED, ATOK_ON);
	}

    return;
}


/***************************************************************************
function: AFileHeaderFromInfoV3 - to construct a line for use as a file 
header for a 3.11 doc. Note keywords are hardwired to their 311 values.

In the passed info_ptr, the version member should reflect the original
version of the file - e.g. 400 if being called by wp400311
*/
void AFileHeaderFromInfoV3(info_ptr, header)
AFileInfo *info_ptr;/* in: info for header */
mbchar *header;     /* out: buffer for header line */
{
    mbchar *docTypeTok;
    int len;

    /* makes something like: "*START WORDS VERSION=311 ENCODING=7BIT" */

    /* *START WORDS... */
    strcpy(header, MBNXLT("*START "));
    docTypeTok = ADocTypeTokFromCode(info_ptr->docType);
    strcat(header, docTypeTok);

    /* ...VERSION=311/300/400... */
	/* That is, this is a 311 file which can be opened by any 3.x
		software, and it comes from (e.g.) 4.0 */
    len = MBstrlen(header);
    sprintf(&header[len], MBNXLT(" VERSION=311/300/%d"), info_ptr->version);

    /* ...ENCODING=7BIT... */
    if ( info_ptr->encoding )
    {
        len = MBstrlen(header);
        sprintf(&header[len], MBNXLT(" ENCODING=7BIT"));
    }
    else
    {
        len = MBstrlen(header);
        sprintf(&header[len], MBNXLT(" ENCODING=NONE"));
    }
    return;
}

/***************************************************************************
function: ADocTypeTokFromCode - return pointer to string for type of doc
*/
mbchar *ADocTypeTokFromCode(docType)
int docType; /* as defined in doctypes.h */
{
    switch ( docType )
    {
        case AX_DOC_TYPE_ASTERX_WORDS:         return(ATOK_WORDS        );
        case AX_DOC_TYPE_ASTERX_GRAPHICS:      return(ATOK_GRAPHICS     );
        case AX_DOC_TYPE_ASTERX_SS:            return(ATOK_SPREADSHEETS );
        case AX_DOC_TYPE_ASTERX_MACROS:        return(ATOK_MACROS       );
        case AX_DOC_TYPE_ASTERX_EQUATIONS:     return(ATOK_EQUATIONS    );
        case AX_DOC_TYPE_ASCII:                return(ATOK_ASCII        );
        case AX_DOC_TYPE_ASTERX_ELFARRAY:      return(ATOK_ELFARRAY     );
        case AX_DOC_TYPE_ASTERX_ELF_TABLE:     return(ATOK_ELFTABLE     );
        case AX_DOC_TYPE_ASTERX_QUERYDATA:     return(ATOK_QUERYDATA    );
        case AX_DOC_TYPE_ASTERX_BUILDER:       return(ATOK_BUILDER      );
        case AX_DOC_TYPE_ASTERX_BUILDER_TURBO: return(ATOK_BUILDER_TURBO);
        case AX_DOC_TYPE_ASTERX_BUILDER_ABD:   return(ATOK_BUILDER_ABD  );
        case AX_DOC_TYPE_ASTERX_OBJ_CLASS:     return(ATOK_OBJ_CLASS    );
        case AX_DOC_TYPE_ASTERX_BITMAP:        return(ATOK_BITMAP       );
        case AX_DOC_TYPE_HEX:                  return(ATOK_HEX          );
        case AX_DOC_TYPE_ASTERX_MENUBAR:       return(ATOK_MENUBAR      );
        default:                               return(ATOK_BINARY       );
    }
}

/***************************************************************************
function: AFileEnderFromInfo - return string containing file ender line. 
*/
mbchar *AFileEnderFromInfo(info_ptr, header)
AFileInfo *info_ptr;/* in: info for header */
mbchar *header;     /* out: buffer for header line */
{
    mbchar *docTypeTok;

    /* makes something like: "*END WP" */
    strcpy(header, ATOK_END);
    strcat(header, SPACE);
    docTypeTok = ADocTypeTokFromCode(info_ptr->docType);
    strcat(header, docTypeTok);
    return;
}

/***************************************************************************
function: AFileGetBuffer - gets pointer to start of data, and data len
*/
void AFileGetBuffer(in_buf, in_len, data_ptr, data_len)
mbchar * in_buf;
int in_len;
mbchar ** data_ptr;
int *data_len;
{
	int ix, len;
	mbchar *buf;

	/* init */
	len = in_len;
	buf = in_buf;

	/* find the ichar AFTER the first newline (end of the header) */
	for ( ix=0; ix<in_len; ix++ )
	{
		if ( in_buf[ix] == '\n' )
		{
			buf = &in_buf[++ix];
			len -= (ix+1);
			break;
		}
	}

	/* done */
	*data_ptr = buf;
	*data_len = len;
}

/*************************************************************************/
/* Herin follow a set of routines which should be called immediatly before
writing anything to disk, and immediatly after reading something from disk. 
These routines will take care of encoding 8 and 16 bit characters in a 7
bit manner.  The encoding used is as follows:
   let 'c' be a character, and "\" be the ESCAPER
   if ( c < 128 )
      output c
   if ( 128 <= c < 256 )
      output c as "\[a-p][a-p]"   where a==>0 ... p==>15 
   else if ( c >= 256 )
      output c as "\[ -_][ -_][ -_]"  (that's [blank - underscore])
              6 bits -^    ^    ^
                5 bits ----|    |
                     5 bits ----|
Since the encodeing chars for 8 bit chars [a-p] do not overlap with the
encoding chars for 16 bit chars [blank to underscore] we can tell the two
apart just by looking at the character after the ESCAPER.

ESCAPERs in the original text are doubled.

Note: The routines below take an argument specifying whether the output
should be encoded or not.  Thevalues for this flag on output are:
    0 - do not encode
    1 - always encode     
    2 - encode if the users profile says to encode, else do not encode

On the reading side the choice is simply:
   0 - do not read as encoded
   1 - read as encoded

If text is being exported from Aster*x the you should never encode since you
want to leave characters in their native format, so other software (and the user)
can read them directly.

When reading foreign text (ie. ascii import) you should not read it as encoded
since ESCAPERs may cause character conversions.  Even if incoming text is
read as unencode, it will still be scanned for wide characters, so that the
read functions can be used to read multi-byte text from the outside world.
*/

/* character used to escape codes in encoded text */
#define ESCAPER '^' 

int NormalizeOut( inP, outP, max_size, info_ptr , encodeNewlines)
register mbchar *inP;
register mbchar *outP;
register int max_size;
AFileInfo *info_ptr;    /* encoding info or NULL */
int encodeNewlines;	/* TRUE to encode all embedded newlines, 
						not the last new newline, FALSE to leave all newlines
						alone */
{
	return(AxNormalizeOut( inP, outP, max_size, info_ptr , FALSE));
}

int FileNormalizeOut( inP, outP, max_size, info_ptr , encodeNewlines)
register mbchar *inP;
register mbchar *outP;
register int max_size;
AFileInfo *info_ptr;    /* encoding info or NULL */
int encodeNewlines;	/* TRUE to encode all embedded newlines, 
						not the last new newline, FALSE to leave all newlines
						alone */
{
	int hasNewline=FALSE;
	int len=strlen(inP);

	if( inP[len-1] == '\n')	{
		hasNewline = TRUE;
		inP[len-1] = '\0';	/* remove new line */
	}
	len = AxNormalizeOut( inP, outP, max_size, info_ptr, encodeNewlines );
	if(hasNewline)	{
		outP[len++] = '\n';	/* put back the new line */
		outP[len] = '\0';
	}
	return(len);
}
                          
int AxNormalizeOut( inP, outP, max_size, info_ptr, encodeNewlines )
register mbchar *inP;
register mbchar *outP;
register int max_size;
AFileInfo *info_ptr;    /* encoding info or NULL */
int encodeNewlines;	/* TRUE to encode newlines, FALSE not to encode them */
{
    register int encode;
    mbchar errbuff[80];
    int bytes_out;
    register int len, ix, str_len;
    register mbchar msb, lsb;
	int encodeChar; 

    encode = (info_ptr) ? info_ptr->encoding : 1;
    len = 0;
    if ( encode == 0 )
       {
       /* do not encode, just output as raw bytes */
       while ( *inP != (mbchar)0 )
          {
          if ( len+1  > max_size )
                goto all_done;
          len++;
          *outP++ = *inP;
          inP++;
          }
       *outP = '\0';
       return( len );
       }

	str_len = MBstrlen(inP);	

    /** we are to output the string in encoded form */
    /* 8-bit characters are output as "\[a-p][a-p]"  */
    /* encode all embedded newlines but the terminating newline */
    /* any ESCAPERs in the text are doubled  */
	for(ix=0; ix < str_len; ix++, inP++) {

		if( ((unsigned)(*inP) >= 128)	||
			( ((unsigned)(*inP) < 32) && 
			(encodeNewlines || ( (unsigned)(*inP) != '\n'))) )	{
             /* an 8-bit character */
             if ( len+3 >= max_size )
                goto all_done;
             len = len + 3;
             *outP++ = ESCAPER;
             *outP++ = ((*inP & 0xF0) >> 4) + 'a';
             *outP++ = (*inP & 0xF) + 'a';
        } else {
             /* a standard character */
             if ( *inP > 0 )
                len++;
             if ( *inP == ESCAPER )
                /* double the ESCAPERs */
                len++;
             if ( len >= max_size )
                /* not enough rootm */
                goto all_done;

             *outP++ = (*inP & 0xFF);
             if ( *inP == ESCAPER )
                *outP++ = ESCAPER;
		}
	}

all_done:
    *outP = '\0';
    return( len );
}

int NormalizeOut16( inP, outP, max_size, info_ptr)
register wchar *inP;
register mbchar *outP;
register int max_size;
AFileInfo *info_ptr;    /* encoding info or NULL */
{

    register int encode;
    register int len;
    register mbchar msb, lsb;
    int out_char;

    encode = (info_ptr) ? info_ptr->encoding : 1;

    len = 0;
    if ( encode == 0 )
       {
       /* do not encode, just output in most-sig-byte first order */
       /* without any null bytes                           */
       while ( *inP != (wchar)0 )
          {
          if ( IS_ORIENT_CHAR( ((wchar)*inP) ) )
             {
             if ( len+2 > max_size )
                goto all_done;
             MAKE_MULTI_FROM_WIDE( inP, outP );
             len = len + SIZE_MULTI_FROM_WIDE;
             }
          else
             {
             if ( len+1  > max_size )
                goto all_done;
             len++;
             *outP++ = *inP;
             }
          inP++;
          }
       *outP = '\0';
       return( len );
       }

    /** we are to output the string in encoded form */
    /* 8-bit characters are output as "\[a-p][a-p]"  */
    /* 16 bit characters are output as ESCAPERs followed by three characters */
    /* the first char represents 6 bits, and the next two represet 5 bits each*/
    /* The range of the first character is <blank> ==> <_>, for 64 unique */
    /*  values, none of which overlap the characters used for the 8 bit chars */
    /* any ESCAPERs in the text are doubled  */
    while ( 1 )
       {
       if ( IS_ORIENT_CHAR( *inP ) )
          {
          /* a wide character */
          if ( len+5 >= max_size )
              goto all_done;
          len = len + 4;
          *outP++ = ESCAPER;

          /* As a bit of a hack, we avoid putting double quotes */
          /* into the output, but instead substitute a back quote */
          out_char = ((*inP & 0xFC00) >> 10) + ' ';
          if ( out_char == '"' )
              out_char = '`';
          *outP++ = out_char;
          out_char = ((*inP & 0x3E0) >> 5) + ' ';
          if ( out_char == '"' )
              out_char = '`';
          *outP++ = out_char;
          out_char = (*inP & 0x1F) + ' ';
          if ( out_char == '"' )
              out_char = '`';
          *outP++ = out_char;
          }
       else
          {
          if ( ((unsigned)(*inP) >= 128) || ( *inP && ((unsigned)(*inP) < 32) && ((unsigned)(*inP) != '\n') ))
             {
             /* an 8-bit character */
             if ( len+3 >= max_size )
                goto all_done;
             len = len + 3;
             *outP++ = ESCAPER;
             *outP++ = ((*inP & 0xF0) >> 4) + 'a';
             *outP++ = (*inP & 0xF) + 'a';
             }
          else
             {
             /* a standard character */
             if ( *inP > 0 )
                len++;
             if ( *inP == ESCAPER )
                /* double the ESCAPERs */
                len++;
             if ( len >= max_size )
                /* not enough rootm */
                goto all_done;

             *outP++ = (*inP & 0xFF);
             if ( *inP == ESCAPER )
                *outP++ = ESCAPER;
             }
          }
        
       if ( *inP == 0 )
          break;
       inP++;
       }

all_done:
    *outP = '\0';
    return( len );
}


          



/** convert an external string into internal format **/
/** returns the number of bytes (not characters) in the result string **/
/** arguments: inP - pointer to string to convert                     **/
/**            start_outP - pointer to where result should go         **/
/**            max_len - maximum number of bytes which can be put in  **/
/**                      the result area                              **/
/**            info_ptr - inforamtion about how conversion should be done **/
 
int NormalizeIn( inP, start_outP, max_len, info_ptr )
register mbchar *inP;
register mbchar *start_outP;
int max_len;
AFileInfo *info_ptr;    /* encoding info or NULL */
{
    register mbchar *outP;
    int num_bytes;
    register mbchar *p;
    register int len;
    int allow_decode;
    wchar in_char;
    

    len = 0;
    outP = start_outP;
    allow_decode = (info_ptr) ? info_ptr->encoding : TRUE;

    while ( (*inP != (mbchar)NULL) && (len < max_len-1) )
       {
       if ( allow_decode )
          {
          /* we are to scan for ESCAPER encoded characters */
          if ( (*inP== ESCAPER) && (*(inP+1) == ESCAPER) )
             {
             /* an escaped ESCAPER, should only be one ESCAPER */
             *outP++ = ESCAPER;
             inP = inP + 2;
             len++;
             continue;
             }
          if ( *inP== ESCAPER )
             {
             /* possible start of encoded 8 or 16 bit character */
             p = inP + 1;
             if ( (*p >= 'a') && (*p <= 'p') )
                {
                /* possible encoded 8 bit character */        
                if ( (*(p+1) >= 'a') && (*(p+1) <= 'p') )
                    {
                    /* ok, a valid 8 bit character */
                    *outP++ = ((*p - 'a') << 4) | (*(p+1) - 'a');
                    inP = inP + 3;
                    len++;
                    continue;
                    }
                }
             else if ( (*p >= ' ') && (*p <= '`') )
                {
                /* possible start of encoded 16 bit character */
                if ( (*(p+1) >= ' ') && (*(p+1) <= '`') &&
                     (*(p+2) >= ' ') && (*(p+2) <= '`') )
                   {
                   /* a 16 bit value */
                   /* As a bit of a hack, we avoid putting double quotes */
                   /* into the output, but instead substitute a back quote */
                   if ( *p == '`' )
                       *p = '"';
                   if ( *(p+1)== '`' )
                       *(p+1) = '"';
                   if ( *(p+2)== '`' )
                       *(p+2) = '"';

                   in_char = ((*p - ' ') << 10) | ((*(p+1) - ' ') << 5) |
                                                 (*(p+2) - ' ');
                   MAKE_MULTI_FROM_WIDE( (&in_char), outP );
                   len = len + SIZE_MULTI_FROM_WIDE;
                   inP = inP + 4;
                   continue;
                   }
                }
             }      /* end of "it is a ESCAPER */
   
          }    /* end of allow_decode */
       /* a normal character */
       *outP++ = *inP++;
       len++;
       }
    *outP = 0;

    return( len );
}


    


/** convert an external string into internal format **/
/** returns the number of characters in the result string **/
/** If the "allow_decode" is TRUE we will look for, and convert ESCAPER */
/**    encoded charactes.  If this is FALSE we do not look for ESCAPER  */
/**    encoded characters, and do only mixed-byte->wide_char conversions  */

int NormalizeIn16( inP, start_outP, max_len, info_ptr )
register mbchar *inP;
register wchar *start_outP;
int max_len;
AFileInfo *info_ptr;    /* encoding info or NULL */
{
    register wchar *outP;
    register mbchar *p;
    register int len;
    int allow_decode;

    if ( info_ptr == (AFileInfo *)0 )
        allow_decode = FALSE;
    else if ( info_ptr == (AFileInfo *)1  )
        allow_decode = TRUE;
    else
        allow_decode = (info_ptr) ? info_ptr->encoding : TRUE;

    len = 0;
    outP = start_outP;

    while ( (*inP != (mbchar)NULL) && (len < max_len-1) )
       {
       if ( IS_MULTI_BYTE( inP ) )
          {
          /* we have a direct multi-byte character */
          *outP = MAKE_WIDE_FROM_MULTI( inP );
          inP = inP + 2;
          outP++;
          len++;
          continue;
          }
       if ( allow_decode )
          {
          /* we are to scan for ESCAPER encoded characters */
          if ( (*inP== ESCAPER) && (*(inP+1) == ESCAPER) )
             {
             /* an escaped ESCAPER, should only be one ESCAPER */
             *outP++ = ESCAPER;
             inP = inP + 2;
             len++;
             continue;
             }
          if ( *inP== ESCAPER )
             {
             /* possible start of encoded 8 or 16 bit character */
             p = inP + 1;
             if ( (*p >= 'a') && (*p <= 'p') )
                {
                /* possible encoded 8 bit character */        
                if ( (*(p+1) >= 'a') && (*(p+1) <= 'p') )
                    {
                    /* ok, a valid 8 bit character */
                    *outP++ = ((*p - 'a') << 4) | (*(p+1) - 'a');
                    inP = inP + 3;
                    len++;
                    continue;
                    }
                }
             else if ( (*p >= ' ') && (*p <= '`') )
                {
                /* possible start of encoded 16 bit character */
                if ( (*(p+1) >= ' ') && (*(p+1) <= '`') &&
                     (*(p+2) >= ' ') && (*(p+2) <= '`') )
                   {
                   /* As a bit of a hack, we avoid putting double quotes */
                   /* into the output, but instead substitute a back quote */
                   if ( *p == '`' )
                       *p = '"';
                   if ( *(p+1)== '`' )
                       *(p+1) = '"';
                   if ( *(p+2)== '`' )
                       *(p+2) = '"';

                   *outP = ((*p - ' ') << 10) | ((*(p+1) - ' ') << 5) |
                                                 (*(p+2) - ' ');


                   /* a 16 bit value */
#if HYUNDAI   /*  {  */
                   /* On Solaris, characters in the font are kept without */
                   /*  the high-bit of each byte set, but under X11R5/SunOS */+
                   /*  the characters in the font are kept with the high bit */
                   /*  set, at least as long as the system uses a type "zero" */
                   /*  font.  This is indicated by the font name ending in   */
                   /*  "-0" instead of "-1".                                 */
                   /* Here we make sure that the wide character we */
                   /*  have is correct for this system                     */
#if SUN4S
                   *outP = *outP & 0x7F7F;  /* clear high bits of both bytes */
#else
                   /* X11R5 under SunOS, using type "zero" font */
                   *outP = *outP | 0x8080;  /* set both high bits  */
#endif
#endif    /* }  HYUNDAI  */
                   outP++;
                   len++;
                   inP = inP + 4;
                   continue;
                   }
                }
             }      /* end of "it is a ESCAPER */
   
          }    /* end of allow_decode */
       /* a normal character */
       *outP++ = (unsigned)*inP++;
       len++;
       }
    *outP = 0;

    return( len );
}

/***************************************************************************
function: encodeMBText - to encode text strings. Escapes newlines and double
quotes using backslash. Potentially doubles size of input string.
*/
void encodeMBText(outP, inP)
register mbchar *outP; /* should point to buffer twice as big as input string */
register mbchar *inP;
{
    while (*inP) {
        if (*inP == '\\' || *inP == '"') {
            *outP++ = '\\';
            *outP++ = *inP;
        }
        else if (*inP == '\n') {
            *outP++ = '\\';
            *outP++ = 'n';
        }
        else 
            *outP++ = *inP;
        inP++;
    }
    *outP = 0;
}

/***************************************************************************
function: decodeMBText - to decode text strings with escaped newlines and quotes
works in place.
*/
void decodeMBText(textP)
mbchar *textP;
{
    register mbchar *outP;
    register mbchar *inP;

    inP = outP = textP;
    while (*inP) {
        if (*inP == '\\') {
            inP++;
            if (*inP == 'n')  {
                *outP++ = '\n';
                inP++;
            }
            else {
                *outP++ = *inP++;
            }
        }
        else
            *outP++ = *inP++;
    }
    *outP = 0;
}

/***************************************************************************
function: encodeWText - to encode text strings. Escapes newlines and double
quotes using backslash. Potentially doubles size of input string.
*/
void encodeWText(outP, inP)
register wchar *outP; /* should point to buffer twice as big as input string */
register wchar *inP;
{
    while (*inP) {
        if (*inP == '\\' || *inP == '"') {
            *outP++ = '\\';
            *outP++ = *inP;
        }
        else if (*inP == '\n') {
            *outP++ = '\\';
            *outP++ = 'n';
        }
        else 
            *outP++ = *inP;
        inP++;
    }
    *outP = 0;
}

/***************************************************************************
function: decodeWText - to decode text strings with escaped newlines and quotes
works in place.
*/
void decodeWText(textP)
wchar *textP;
{
    register wchar *outP;
    register wchar *inP;

    inP = outP = textP;
    while (*inP) {
        if (*inP == '\\') {
            inP++;
            if (*inP == 'n') {
                *outP++ = '\n';
                inP++;
            }
            else {
                *outP++ = *inP++;
            }
        }
        else
            *outP++ = *inP++;
    }
    *outP = 0;
}

/***************************************************************************
function: AFileValidateHeaders - checks the header and ender for "parity"
*/
int AFileValidateHeaders (mbchar *buf, int buflen)
{
	mbchar hdr[AREADBUFSIZE+1], edr[AREADBUFSIZE+1];
	int ix, hdrlen, edrlen, is_valid, found;
	AFileInfo finfo;

	/* find the header */
	for ( ix=hdrlen=0; ix<buflen; ix++ )
	{
		if ( buf[ix] == '\n' )
		{
			hdrlen = ix+1;
			break;
		}
	}
	if ( hdrlen <= 0 )
		is_valid = FALSE;
	else
	{
		/* build a matching ender */
		strncpy(hdr, buf, hdrlen);
		if ( AFileInfoFromHeader(hdr, &finfo) != AOK )
			is_valid = FALSE;
		else
		{
			AFileEnderFromInfo(&finfo, edr);
			edrlen = strlen(edr);
			found = FALSE;
			for ( ix=buflen-1; ix>hdrlen; ix-- )
			{
				if ( buf[ix] == '\n' )
				{
					if ( (ix+edrlen) > buflen )
						continue;
					else if ( strncmp(&buf[ix+1], edr, edrlen) != 0 )
						continue;
					else
					{
						is_valid = TRUE;
						break;
					}
				}
			}
			if ( ix <= hdrlen )
				is_valid = FALSE;
		}
	}
	return(is_valid);
}

void 
AxOverrideEncryptionHeader(mbchar *buf, int buf_len)
{
    mbchar encrypt_buf[ALINELENGTH+1];
	int ix, j, hdr_len, encrypt_len;

	encrypt_len = sprintf(encrypt_buf, ATOK_CVARIABLE, ATOK_ENCRYPTED, ATOK_ON);

	if ( buf_len < ALINELENGTH )
		hdr_len = buf_len;
	else
		hdr_len = ALINELENGTH;
	for ( ix=0; ix<hdr_len; ix++ )
	{
		if ( strncmp(&buf[ix], encrypt_buf, encrypt_len ) == 0 )
		{
			for (j=0; j < encrypt_len; j++ )
			{
				if ( ( j+ix) >= hdr_len )
					break;
				buf[j+ix] = ' ';
			}
			break;
		}
	}
}
