/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: laytext.h
   Applix ASC version: 2
   Date: Mon Nov 16 13:53:34 1998
*/
#ifndef _LAYTEXT_H_
#define _LAYTEXT_H_

/*
 * Need some xtypedefs
 */
#include <basicwin.h>
#if 0
#if AX_X11
#ifndef _XLIB_H_
#include <X11/Xlib.h>
#endif
#endif  /* DAMNT */

#if AXMS      /* BOGUS X definitions to allow NT build */
#define Drawable unsigned long
#define XImage   char
#endif          /* DAMNT */
#endif


/* Usage of FieldInfo for field clearing and clipping
 * purposes:
 *--------------------------------------------------
 * 
 * Used for field clearing, and for ensuring that we
 * don't display any text
 * info that our applications don't know the extent
 * of.  Typically, this would be glyph bits which
 * jut outside of the "simple cell" for the glyph
 * (on the left.)  The simple cell is as tall as the
 * glyph, and extends from the current set point to
 * that point plus the advance width (setwidth.)
 * We intend to include in the bounding box any
 * glyph bits which jut to the right for the last
 * glyph bounded.
 * We could dispense with this if the applications
 * would accept and use info which expressed any
 * leftward jut.  This would break implicit assumptions
 * that when text is positioned at x, it is impossible
 * for it to extend to the left of x.  As for rightward
 * jut, the applications are already accustomed to the
 * atm telling them how long a piece of text really is
 * (within the app-specified field limits.)
 *
 * The "set" boolean is used because of the structure
 * of text output calls in the code.  Text output info
 * often follows: wrTx->laytext->WMtext (field read/write)
 * but there are many direct calls to laytext->WMtext, as
 * well as direct calls to WMtext.  A higher-level call
 * such as wrTx sets the flag and the textBox before doing
 * output (which may result in multiple laytext calls,)
 * then resets it when done.  If a lower level call finds
 * the info unset, it must set it; when done with text
 * output, it will reset the info only if it set it itself.
 *
 * resetClip is used to control whether WMtext should
 * reset the GC's clipmask when done. E.g., wrTx suppresses
 * clip reset until the last of the n WMtext calls it uses
 * for one field output.  (This prevents glyph damage at
 * text attr boundaries.)
 *
 * 
 * Use of FieldInfo for image text mode:
 *----------------------------------------
 *
 * "Image text mode" is used for setting text onto a pixmaps so
 * that scaling can be applied to the text prior to display.
 *
 * The textBox is a bounding box in regular display coords,
 * as used for non-image-mode field clearing and clipping.
 * 
 * The printerBox has the same info as textBox, but expressed
 * in target device pixels and with the origin zeroed by convention.
 * (since it's possible for the target device to be the screen device,
 * "printerBox" can be a misnomer..)
 *
 * The scaledTextBox is in screen pixels; it describes the rectangle
 * used for XPutImage
 *
 * The dispBox is formed by intersecting scaledTextBox with
 * any higher-level clip info (such as app clip, or expose event
 * clip.)  ClipBox is provided for convenience.
 *
 * xdelta is sort of like the delta member of the XTextItem;
 * it expresses the x offset of the latest WMtext call from
 * the starting x coord of the overall field. used for offsetting
 * onto textPixmap.
 *
 * textItemIndex is used for imageText field reads, in helping to
 * control nonmatching-string-lifting work.
 *
 * a textItemList is maintained for each atm_write_tx or dispf
 * pass; there's one item per character.  A field info struct is
 * on each item (called charInfo), and stores per-character info;
 * so a fieldInfo instance is used globally as FieldInfo for
 * whole-field stuff, and then again on the textItem list for
 * individual chars.
 */
struct fieldInfo 
{
	unsigned int			set: 1;
	unsigned int			resetClip: 1;
	unsigned int			imageTextMode: 1;
	unsigned int			rotated: 1;
	unsigned int			underlined: 1;
	unsigned int			transformed: 1; /* angle!=0, scale!=1 */

	AxRectangle				textBox;
	AxRectangle				printerBox;
	AxRectangle				scaledTextBox;
	AxRectangle				ClipBox;
	AxRectangle				dispBox;
	int						maxAboveBaseline;
	int						targMaxAboveBaseline;
	Drawable				textPixmap;
	XImage					*textImage;
	int						xdelta;
	int						textItemIndex;
	int						underlineType;
	struct _textItem		*textItems;
};

#define TXPIXMAP_HEIGHT 75	/* max font height + sub/sup offsets; device 106 */

/*
 * Modifiers for computing text bounding box (AxGetTxRect)
 *
 * SIMPLECLIP: Clip left edge of text field right at the
 * text position.  Don't worry about any glyph black bits
 * which might extend outside the 1st cell to the left.
 *
 * LEFTBEARINGCLIP: Adjust the left edge of the text field
 * cliprect to include any glyph blackbits which might extend
 * to the left of the text position.  That is, account for
 * the left bearing of the 1st glyph in the text field.
 *
 * LEFTHILITECLIP: Adjust the left edge of the text field
 * cliprect so that it starts just after the advance width
 * of the 1st glyph in the text field.  This is done during
 * field read text setting (setting the nonmatching substring
 * of the field read) where the nonmatching string starts just
 * after a highlighted character. To help explain:
 * In all cases where the
 * field read is not in its first-display state and is resetting
 * a substring which doesn't start with the first visible char-
 * acter in the field, the character just to its left (the last
 * matching glyph) is included in order to repair any damage to
 * it that might have occurred during the previous lifting of
 * text.  If that character is hilit, then all that we try to
 * repair is that portion of the glyph which extends out of the
 * charcell to the right.  To do this, the left edge of the clip
 * is placed on the right edge of the cell.
 */
#define SIMPLECLIP	0
#define LEFTBEARINGCLIP 1
#define LEFTHILITECLIP 2

#define VERTICAL 0
#define HORIZONTAL 1 /* AxTextScale asks: if(horizontal) */

/* used for image text */

typedef struct _textItem {
	struct _textItem *next;
	struct _textItem *prev;
	struct fieldInfo charInfo;	 /* used in per-character context */
	TmPoint32 charBox[4];  /* xformed corners of per-char bounding box */
	unsigned int mapped: 1;/* xformed glyph image on charInfo->textImage */
	unsigned int drawn: 1; /* unxformed glyph image on charInfo->textPixmap */
} textItem;

#define LISTAPPEND 0x7fffffff /* an impossibly large list index */

#endif /* _LAYTEXT_H_ */

