/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: maininit.c
   Applix ASC version: 50
   Date: Fri Apr 30 21:12:59 1999
*/
#include <aos.h>
#include <basicwin.h>
#include <scac.h>
#include <style.h>
#include <paratm.h>
#include <xim.h>
#include <getstr5.h>
#include <getstr8.h>
#include <ax_des.h>


int AxBetaTest;     /* TRUE if this is a beta test site */
int AxMetric;       /* TRUE if running metric mode */
int AxShortNames;   /* TRUE if profile imposes short (SYS5) filenames */

extern short NoColor;   /* THIMBLE is X11 base let's try color */
extern foreignChars UTsymbols;
extern foreignChars UTprofsym;
extern Display *Dpy;
extern char *axDefaultSerifFont;
extern char *axDefaultSansSerifFont;
extern char *axDefaultMonospaceFont;
extern Window AxTopWinNeverMapped;  /* created by AxGetDpyAndGCs */

extern void AxSetEncryptionState();
extern int AxGetEncryptionState();
extern void AxLoadKeytables();

#if BUILD_ORIENT
extern XIM input_method;
#endif


#include <maininit.p>

/****************************************************************************/
static void read_encryption_state()
{
	char *val = (char *)UPbyName("axUseFileEncryption");
	if ( !val || strcmp(val, "0") == 0 )
		AxSetEncryptionState(FALSE);
	else
		AxSetEncryptionState(TRUE);
	AxLoadKeytables(AxGetEncryptionState());
}

/****************************************************************************/
void massive_init(argc,argv)
int argc;
char **argv;
{
    char *sisP;
    short i;
    int c;
    short protocol;
    extern int ACoptind;
    extern char *ACoptarg;
        static short traceLogging = 0;  /* short-term; easy way to initialize */
    static int extraLogging = 0;
    static int demoLoop = 0;        /* looping demo playback */
    char *jfile;
    int fd;
    int lineNumber;
    extern int AxIOblocks;
    static struct TmCrePacket ATMpacket;
    short xwid;
    long widget_id;
    int status;

#if !BUILD_ORIENT && !ALPHA && !AXMS && !(LINUX && __alpha__)
    /* here is a bit of code to keep us from allocating 1024 byte widgets */
    /* since the Asian builds have more things in the widgets, and pointers */
    /* on the ALPHA are 8 bytes, both of these environments are already over */
    /* the limit.                                                           */
    if ( sizeof(axw_widget) > 496 )
        tempAbort("The size of widget_t has exceeded 496, make it smaller");
#endif

    sisP = (char *) NULL;

    SetThimbleDirs(argv[0]);        /*  Turns XLT func on -- lhunt */
    InitThimble();
    HelpInit();

#if AXMS
    /* install the optional true-type fonts which match the ones we ship with */
    /* Gallium on the Unix side.                                              */
    InstallOptionalTruetypeFonts();
#endif


/* Turn on signal catchers to avoid cores */
#if AX_X11
    THIMset_sigs();
#endif

#if QATOOL
    /* should move to a qa init function  -lance */
    if ( getenv("QACAPTURE") )
        QaStartEventCapture("/tmp/qakeys");
    else if ( getenv("QAPLAY") )
        QaStartEventPlayback("/tmp/qakeys");
#endif


    /*
     * initialize AC and environment; get options from command line
     *
     * force some switches on by setting flags
     * (should be done under overrides, below; clearing at this point is
     * redundant initialization, not an override)
     */

/* Process once-only defaults */

    NoColor = UPnumVal(PROF_NO_COLOR_);
    AxBetaTest = UPnumVal(PROF_BETA_TEST_);
    AxMetric = UPnumVal(PROF_MEASURE_);
    AxShortNames = UPnumVal(PROF_SHORT_NAMES_);
    AxIOblocks = 8160;
	AxReadLockState();
	AxReadACLState();
	read_encryption_state();

    /* Now get the profile symbols - currency, decimal etc.. */
    UTgetsym();
    UTsymbols = UTprofsym;
    UTsymbols.use_me = TRUE;

    /* find out if this device supports audio */
    GetSupportsAudio();

    ACinitXferBin();        /* create xfer bin doc */

    /* Initialize ATM */
    init_atm();
    
#if AXMS
    /*  the Windows/NT case */
    AxGetDpyAndGCs();
#endif

    axSetDefaultFontNames();

#if AX_X11
	AxCreateMemInfoAtoms();
#endif

#if BUILD_ORIENT
    /** initialize the oriental font and input mechanism */
    /* initialize the connection to the conversion server */
    input_method = AXOpenIM(Dpy, NULL, (char *)NULL, 
                                                (char *)NULL, &status);
    if ( input_method == (XIM)NULL && (status != NOT_LANG_LOCALE) )
        {
        printf( XLT("WARNING: No Asian entry possible:") );
        if ( status == NO_KANJI_FONT )
            printf( XLT("    Unable to access Asian fonts \n") );
        else if ( status == NO_KANJI_CONVERTER )
            printf( XLT("    Could not connect to the conversion server\n") );
        else
            printf( XLT("    Unknown return status from AXOpenIM() = %d\n"),
                                                            status );
        }
#endif

    /* create window 0 - the status line HISTORICAL! */
    TmCreate(&xwid, &widget_id, &ATMpacket);

}   /* massive_init() */

void
axSetDefaultFontNames()
{
    char *prefval = GetProfileByStr(NXLT("axPrintDefaultType"));

#if AXMS
    axDefaultSerifFont = NXLT("Times New Roman");
    axDefaultSansSerifFont = NXLT("Arial");
    axDefaultMonospaceFont = NXLT("Courier New");
#else
    if (!strcmp(prefval, NXLT("PCL5")))
        {
        axDefaultSerifFont = NXLT("CG Times");      /* pcl5 defaults */
        axDefaultSansSerifFont = NXLT("Univers");
        axDefaultMonospaceFont = NXLT("Courier");
        }
    else
        {
        axDefaultSerifFont = NXLT("Times");         /* postscript defaults */
        axDefaultSansSerifFont = NXLT("Helvetica");
        axDefaultMonospaceFont = NXLT("Courier");
        }
#endif
}
