/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: oldwin.h
   Applix ASC version: 2
   Date: Mon Nov 16 13:53:44 1998
*/
#if AX_X11
#define DEFDEV      DS_X11

#define CELL_UNDER	16	/* "cell underline" position */
#define CELL_HT		17
#define CELL_WID	9

#define RAST_HT		864	/* raster dimensions, pixels */
#define RAST_WID	1024

#endif	/* AX_X11 */



/*
 * min window size in cells
 */

#define MINW_COLS	15
#define MINW_ROWS	0

/*
 * max window sizes in mils
 */




#define MC_XMIN		0
#define MC_YMIN		0


/*
 * larger than any screen dimension but safely under MAXSHORTINT
 * (not convertible to mils)
 */
#define MAX_PELS	10000

/********************************************************************/

/*
 * function codes for 2-operand bitblt()
 *
 * the commonly used (e.g. in Smalltalk and DEC-10/20)
 * numbers that identify as well as define
 * the Boolean functions
 *
 *	-> OBSOLETE (phase out): replaced by symbols in atmattr.h
 */
#define B_CLEAR		4		/* bit clear (dest &= ~src) */
#define B_OR		7		/* dest |= src */
#define B_XOR		6		/* dest ^= src */
#define B_STORE		3		/* dest = src */



/* function codes for 1-operand bitblt() used with rectf() */

#define R_CLEAR		0		/* dest = 0 (OK to use) */
#define R_OR		15		/* OBSOLETE (phase out) */
#define R_SET		15		/* dest = 1 (OK to use) */
#define R_XOR		10		/* OBSOLETE (phase out) */
#define R_COMP		10		/* complement bits of dest (OK to use) */
#define R_STORE		R_OR	/* OBSOLETE (phase out) */

/***** point *****/

typedef struct {short x, y;} Point;	/* device coordinates */

/***** graphics vector (2 pairs of coordinates) *****/

typedef struct {
	short x1;		/* start point, x */
	short y1;		/* start point, y */
	short x2;		/* end point, x */
	short y2;		/* end point, y */
} Vector;

typedef struct {
	int x1;		/* start point, x */
	int y1;		/* start point, y */
	int x2;		/* end point, x */
	int y2;		/* end point, y */
} Vector32;

/***** AxRectangle *****/

typedef struct {
	Point origin;
	Point corner;
} AxRectangle;
/*
 * Note: The cell associated with the origin is the
 * lowest-numbered cell inside the rectangle.
 * The cell associated with the corner is diagonally
 * opposite and immediately outside the rectangle.
 *
 * Caution: AxRectangles are always passed by reference.
 */

/*
 * A_Bitmap
 *
 * Bitmaps are defined in screen coordinates.
 * If a bitmap is not associated with a particular part of the
 * screen, the origin of its rectangle should be (0, 0).
 *
 * The origin of the bitmap's coordinate system is assumed to
 * be on a 16-bit word boundary.
 * The rectangle of the bitmap, which
 * defines its exact boundaries, is not necessarily aligned on
 * a word boundary.   Thus, the locations of word boundaries
 * are determined by the (x, y) coordinate values, not by the
 * positions of bits relative to the bitmap.
 *
 * Note: The declaration (char *) is used for system-dependent
 * main memory addresses.   Such addresses are assumed to be
 * in units of bytes.
 */
typedef struct mapTag {
	char *mapdata;		/* -> system-dependent image definition */
	short mapwid;			/* image width in bytes */
	AxRectangle maprect;		/* image rectangle */

	/* flags */

	unsigned newF: 1;	/* new or discarded image: no content */

	unsigned inMemF: 1;	/* image available (in main memory pool) */
	unsigned modF: 1;	/* modified while in memory (not used) */

	/* esF and interF are meaningful only on IBM PC family */
	unsigned esF: 1;	/* in 8088 extra segment */
	unsigned interF: 1;	/* interleaved organization */
	/*
	 * depthF and wmF apply to X Window System;
	 * senses are chosen so 1-deep maps in ATM have zero values
	 */
	unsigned depthF: 1;		/* map has same depth as Display ("pixmap") */
	unsigned wmF: 1;		/* map is a window manager resource */

	/* image memory management extension */

	unsigned long imsiz;	/* image size, bytes */
	char *imaddr;			/* image address in memory */
	struct mapTag *older;		/* LRU list link */
	struct mapTag *younger;		/* LRU list link */
	struct mapTag *nextinpool;	/* pool list link */
	struct mapTag *previnpool;	/* pool list link */
	struct pdTag *bpool;		/* -> pool descriptor */
	struct layTag *mapLayer;	/* -> owner Layer or NULL */
	/*
	 * note: imsiz and width must be even numbers
	 */
} A_Bitmap;


/***** list of clipping rectangles (list element) *****/

typedef struct CLtag
{
	struct CLtag *link;
	AxRectangle rect;
	A_Bitmap *bmap;
	char *clipWin;			/* window manager resource or 0 */
} ClipList;

/***** Layer *****/

typedef struct layTag {
	AxRectangle rect;		/* image rectangle */
	struct obsTag *obs;	/* linked list of obscured rectangles */
	short window_id;	/* ID of layer's window */

	unsigned mixed : 1; /* Flag to indicate mixed-mode Layer:
							character-mapped with graphics added */
	unsigned regenF : 1;	/* regenerated image flag */
	unsigned needRegen : 1;	/* whole image needs regeneration */
	unsigned needGRegen : 1; /* graphics in image need regeneration */
	unsigned needBRegen : 1;	/* borders need regeneration */
	unsigned tabGlyph : 1;	/* draw glyphs to display tab characters */

	AxRectangle curClip;	/* current clipping rectangle */
	struct layTag *front;	/* adjacent layer in front */
	struct layTag *back;	/* adjacent layer behind */
	ClipList *clp;
} Layer;

/***** obscured list element *****/

typedef struct obsTag {
	Layer *lobs;		/* frontmost obscuring Layer */
	A_Bitmap *bmap;		/* where the obscured data resides */
	struct obsTag *next;	/* chaining */
	struct obsTag *prev;
} Obscured;

/* externals that are widely used */

extern A_Bitmap *A_Display;
extern int MapMode;
