/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: palette.c
   Applix ASC version: 150
   Date: Tue Apr  6 21:57:11 1999
*/
#define APPLIX_WIDGETS 1

#include <stdio.h>
#include <basicwin.h>
#include <ax_xm.h>
#include <aos.h>
#include <akeys.h>
#include <aiconids.h>
#include <scac.h>
#include <ge_syms.h>
#include <elhelper.h>
#include <adlgpoke.h>
#include <getstr5.h>
#include <xmlib.h>
#include <elxwins.h>
#include <xmres.h>
#include <atm.h>
#include <atmwin.h>
#include <axtext.h>
#include <atxint.h>
#include <ge.h>


#define PALT_GE 0
#define PALT_PX 1

#define BUT_UP  0   /* inactive, up */
#define BUT_DOWN 1  /* button widget is down */
#define BUT_NA 2    /* dimmed */
#define BUT_HIDE 3  /* nodisplay */

/* must match xmlib.h */
#define W_LABEL  2
#define W_BUTTON 5
#define W_ENTER  9
#define W_ICON  10
#define W_POPUP 12
#define W_SCALE 25
#define W_ROWCOL 33

/* must match elf_xwins.c */
#define LOCK_FOR_PALETTE    999

/* PALETTE STRUCTURE */

#define PIX_START (20000-PDEX_tool1)

typedef struct {
    AXWidget id;
    int scroll_pos;
} PartsType;

static GEPalette_t CurPalette;  /* current palette state */
static GEPalette_t **PaletteList;   /* one per possible window */
static AXWidget **PaletteWidgets;       /* widget List per window */
static AXWidget **PaletteButtons;       /* widget buttons per window */
static PartsType **PaletteParts;    /* parts lib info per window */

typedef struct {
    int x;
    int y;
    char *name; /* icon file name */
    char *altname;  /* icon file name when depressed */
} pltdata;

/* tool buttons */
#define PD_NUM_TOOLS 14 /* Number of items in tools area, same as P_UDO */
#define PDEX_point    0
#define PDEX_alpha    1
#define PDEX_line     2
#define PDEX_poly     3
#define PDEX_box      4
#define PDEX_ellip    5
#define PDEX_stk      6
#define PDEX_image    7
#define PDEX_square   8
#define PDEX_circle   9
#define PDEX_curve    10
#define PDEX_c_image  11
#define PDEX_zoom     12
#define PDEX_alphabox 13

/* drop shadow buttons */
#define PDEX_shadowButton 14
#define PDEX_shadowColor  15

/* fill pattern and color buttons */
#define PDEX_fillpat 16
#define PDEX_fill_fg 17
#define PDEX_fill_bg 18

/* line pattern and color buttons */
#define PDEX_linepat 19
#define PDEX_line_fg 20
#define PDEX_line_bg 21

/* line thickness */
#define PDEX_line0 22
#define PDEX_line1 23
#define PDEX_line2 24
#define PDEX_line4 25
#define PDEX_line6 26
#define PDEX_linex 27

/* line style, ends and ends/arrows */
#define PDEX_lstyle 28
#define PDEX_aleft  29
#define PDEX_aright 30
#define PDEX_mrkr   31

/* bitmaps for line style and markers (OpenLook) */
#define PDEX_stylemap 32
#define PDEX_lendmap  33
#define PDEX_rendmap  34
#define PDEX_markmap  35


/* pixel editing tools */
#define OLD_PDEX_tool1   41
#define PDEX_tool1       36
#define PDEX_tool2       37
#define PDEX_tool3       38
#define PDEX_tool4       39
#define PDEX_tool5       40
#define PDEX_tool6       41
#define PDEX_tool7       42
#define PDEX_tool8       43
#define PDEX_tool9       44
#define PDEX_tool10      45
#define PDEX_tool11      46
#define PDEX_tool12      47
#define PDEX_pxfill_fg   48
#define PDEX_pxfill_bg   49


#define PALT_NUM_BUTTONS 50
#define PRES_NUM_BUTTONS 5

#define TOOL_BASE PDEX_point
#define FILL_BASE PDEX_fillpat
#define LINE_BASE PDEX_linepat
#define THICK_BASE PDEX_line0
#define LINESTYLE_BASE PDEX_lstyle
#define PIXED_TOOL_BASE PDEX_tool1

static int repaintPalette = FALSE;
extern int UseWidgetColors;
extern int WidgetNeedsUpdated;
extern int ElfLanguage;
extern int TmAmSE;
extern int axolAdjPopupPos();
extern AXWidget WidgetOfWindow();
extern AXWidget_popup AXmCreatePopupMenu();
extern AXWidget AXmCreateScrolledList();
extern AXWidget ZXmCreateIconImage();
extern AXWidget_rowcol AXmCreateRowColumn();
extern AXWidget_pushbutton AXmCreatePushButtonGadget();
#if AXMS
extern int WidthOfDialogZero;
#endif

static pltdata PaletteBasics[PALT_NUM_BUTTONS] = {
    /* tools (40x20) */
    { 7,  4, NXLT("gp_arrow"), NXLT("gp_arrow")},
    { 7, 30, NXLT("gp_alpha"), NXLT("gp_alpha")},
    {53,108, NXLT("gp_line"),  NXLT("gp_line")},
    { 7,134, NXLT("gp_poly"),  NXLT("gp_poly")},
    { 7, 56, NXLT("gp_rect"),  NXLT("gp_rect")},
    { 7, 82, NXLT("gpellips"), NXLT("gpellips")},
    { 7,108, NXLT("gp_strok"), NXLT("gp_strok")},
    { 7,160, NXLT("gp_image"), NXLT("gp_image")},
    {53, 56, NXLT("gp_squre"), NXLT("gp_squre")},
    {53, 82, NXLT("gp_circ"),  NXLT("gp_circ")},
    {53,134, NXLT("gp_curve"), NXLT("gp_curve")},
    {53,160, NXLT("gpcimage"), NXLT("gpcimage")},
    {53,  4, NXLT("gp_zoom"),  NXLT("gp_zoom")},
    {53, 30, NXLT("gptxtbox"), NXLT("gptxtbox")},

    /* drop shadow buttons */
    {24,379, NXLT("gp_shad"),  NXLT("gp_shad")},  /* 28x14 */
    {61,379, NXLT("gp_color"), NXLT("gp_color")}, /* 14x14 */

    /* fill pattern and color */
    {59,195, NXLT("gp_fill"),  NXLT("gp_fill")},  /* 34x34 */
    {59,238, NXLT("gp_color"), NXLT("gp_color")}, /* 14x14 */
    {79,238, NXLT("gp_color"), NXLT("gp_color")}, /* 14x14 */

    /* line pattern and color */
    { 7,195, NXLT("gp_lfill"), NXLT("gp_lfill")}, /* 34x34 */
    { 7,238, NXLT("gp_color"), NXLT("gp_color")}, /* 14x14 */
    {27,238, NXLT("gp_color"), NXLT("gp_color")}, /* 14x14 */

    /* line thickness (should be 8x20) */
    {82,271, NXLT("gpline0"), NXLT("gpline0")}, /*  8x20 */
    {69,271, NXLT("gpline1"), NXLT("gpline1")}, /*  8x20 */
    {55,271, NXLT("gpline2"), NXLT("gpline2")}, /*  8x20 */
    {41,271, NXLT("gpline4"), NXLT("gpline4")}, /*  8x20 */
    {27,271, NXLT("gpline6"), NXLT("gpline6")}, /*  8x20 */
    { 9,271, NXLT("gplinex"), NXLT("gplinex")}, /* 12x20 */

    /* line style */
    {43,351, NXLT("gplinsol"), NXLT("gplinsol")},
    {15,317, NXLT("gp_aleft"), NXLT("gp_aleft")},
    {72,317, NXLT("gp_arigt"), NXLT("gp_arigt")},
    {44,317, NXLT("gp_mkr00"), NXLT("gp_mkr00")},

    /* line style bitmaps */
    {21,336, NXLT("gplinsol"), ""},
    {16,302, NXLT("gp_aleft"), ""},
    {57,302, NXLT("gp_arigt"), ""},
    {45,302, NXLT("gp_mkr00"), ""},

    /* pixed tools */
    { 7,  4, NXLT("pxstroke"), NXLT("pxstroke")},
    {53,  4, NXLT("px_wash"),  NXLT("px_wash")},
    { 7, 50, NXLT("pxsprayc"), NXLT("pxsprayc")},
    {53, 50, NXLT("px_blend"), NXLT("px_blend")},
    { 7, 96, NXLT("px_erase"), NXLT("px_erase")},
    {53, 96, NXLT("pxcerase"), NXLT("pxcerase")},
    { 7,142, NXLT("px_line"),  NXLT("px_line")},
    {53,142, NXLT("px_rect"),  NXLT("px_rect")},
    { 7,188, NXLT("px_ell"),   NXLT("px_ell")},
    {53,188, NXLT("px_flood"), NXLT("px_flood")},
    { 7,234, NXLT("pxfilter"), NXLT("pxfilter")},
    {53,234, NXLT("px_reg"),   NXLT("px_reg")},
    { 7,284, NXLT("gp_color"), NXLT("gp_color")},
    {53,284, NXLT("gp_color"), NXLT("gp_color")},
};

/*
 * Same data as PaletteBasics, but for mono openlook mode.
 */
static pltdata axolPaletteBasics[PALT_NUM_BUTTONS] = {
    /* tools (40x20) */
    { 7,  4, NXLT("gp_arrow"), NXLT("gp_arrow")},
    { 7, 29, NXLT("gp_alpha"), NXLT("gp_alpha")},
    {52,104, NXLT("gp_line"),  NXLT("gp_line")},
    { 7,129, NXLT("gp_poly"),  NXLT("gp_poly")},
    { 7, 54, NXLT("gp_rect"),  NXLT("gp_rect")},
    { 7, 79, NXLT("gpellips"), NXLT("gpellips")},
    { 7,104, NXLT("gp_strok"), NXLT("gp_strok")},
    { 7,154, NXLT("gp_image"), NXLT("gp_image")},
    {52, 54, NXLT("gp_squre"), NXLT("gp_squre")},
    {52, 79, NXLT("gp_circ"),  NXLT("gp_circ")},
    {52,129, NXLT("gp_curve"), NXLT("gp_curve")},
    {52,154, NXLT("gpcimage"), NXLT("gpcimage")},
    {52,  4, NXLT("gp_zoom"),  NXLT("gp_zoom")},
    {52, 29, NXLT("gptxtbox"), NXLT("gptxtbox")},

    /* drop shadow buttons */
    {24,379, NXLT("gp_shad"),  NXLT("gp_shad")},  /* 28x14 */
    {61,379, NXLT("gp_color"), NXLT("gp_color")}, /* 14x14 */

    /* fill pattern and color */
    {59,195, NXLT("gp_fill"),  NXLT("gp_fill")},  /* 34x34 */
    {59,238, NXLT("gp_color"), NXLT("gp_color")}, /* 14x14 */
    {79,238, NXLT("gp_color"), NXLT("gp_color")}, /* 14x14 */

    /* line pattern and color */
    { 7,195, NXLT("gp_lfill"), NXLT("gp_lfill")}, /* 34x34 */
    { 7,238, NXLT("gp_color"), NXLT("gp_color")}, /* 14x14 */
    {27,238, NXLT("gp_color"), NXLT("gp_color")}, /* 14x14 */

    /* line thickness (should be 8x20) */
    {83,271, NXLT("gpline0"), NXLT("gpline0")}, /*  8x20 */
    {69,271, NXLT("gpline1"), NXLT("gpline1")}, /*  8x20 */
    {55,271, NXLT("gpline2"), NXLT("gpline2")}, /*  8x20 */
    {41,271, NXLT("gpline4"), NXLT("gpline4")}, /*  8x20 */
    {27,271, NXLT("gpline6"), NXLT("gpline6")}, /*  8x20 */
    { 9,271, NXLT("gplinex"), NXLT("gplinex")}, /* 12x20 */

    /* line style */
    {43,351, NXLT("gplinsol"), NXLT("gplinsol")},
    {15,317, NXLT("gp_aleft"), NXLT("gp_aleft")},
    {71,317, NXLT("gp_arigt"), NXLT("gp_arigt")},
    {43,317, NXLT("gp_mkr00"), NXLT("gp_mkr00")},

    /* line style bitmaps */
    {21,336, NXLT("gplinsol"), ""},
    {16,302, NXLT("gp_aleft"), ""},
    {57,302, NXLT("gp_arigt"), ""},
    {45,302, NXLT("gp_mkr00"), ""},

    /* pixed tools */
    { 7,  4, NXLT("pxstroke"), NXLT("pxstroke")},
    {52,  4, NXLT("px_wash"),  NXLT("px_wash")},
    { 7, 49, NXLT("pxsprayc"), NXLT("pxsprayc")},
    {52, 49, NXLT("px_blend"), NXLT("px_blend")},
    { 7, 94, NXLT("px_erase"), NXLT("px_erase")},
    {52, 94, NXLT("pxcerase"), NXLT("pxcerase")},
    { 7,139, NXLT("px_line"),  NXLT("px_line")},
    {52,139, NXLT("px_rect"),  NXLT("px_rect")},
    { 7,184, NXLT("px_ell"),   NXLT("px_ell")},
    {52,184, NXLT("px_flood"), NXLT("px_flood")},
    { 7,229, NXLT("pxfilter"), NXLT("pxfilter")},
    {52,229, NXLT("px_reg"),   NXLT("px_reg")},
    { 7,279, NXLT("gp_color"), NXLT("gp_color")},
    {52,279, NXLT("gp_color"), NXLT("gp_color")},
};

static pltdata mwmPaletteBasics[PALT_NUM_BUTTONS] = {
    /* tools (40x20) && 39x20 */
    { 7,  4, NXLT("gp_arrow"), NXLT("gp_arrow")},
    { 7, 30, NXLT("gp_alpha"), NXLT("gp_alpha")},
    {51,108, NXLT("gp_line"),  NXLT("gp_line")},
    { 7,134, NXLT("gp_poly"),  NXLT("gp_poly")},
    { 7, 56, NXLT("gp_rect"),  NXLT("gp_rect")},
    { 7, 82, NXLT("gpellips"), NXLT("gpellips")},
    { 7,108, NXLT("gp_strok"), NXLT("gp_strok")},
    { 7,160, NXLT("gp_image"), NXLT("gp_image")},
    {51, 56, NXLT("gp_squre"), NXLT("gp_squre")},
    {51, 82, NXLT("gp_circ"),  NXLT("gp_circ")},
    {51,134, NXLT("gp_curve"), NXLT("gp_curve")},
    {51,160, NXLT("gpcimage"), NXLT("gpcimage")},
    {51,  4, NXLT("gp_zoom"),  NXLT("gp_zoom")},
    {51, 30, NXLT("gptxtbox"), NXLT("gptxtbox")},

    /* drop shadow buttons */
    {24,346, NXLT("gp_shad"),  NXLT("gp_shad")},  /* 28x14 */
    {61,346, NXLT("gp_color"), NXLT("gp_color")}, /* 14x14 */

    /* fill pattern and color */
    {54,195, NXLT("gp_fill"),  NXLT("gp_fill")},  /* 34x34 */
    {54,238, NXLT("gp_color"), NXLT("gp_color")}, /* 14x14 */
    {74,238, NXLT("gp_color"), NXLT("gp_color")}, /* 14x14 */

    /* line pattern and color */
    { 7,195, NXLT("gp_lfill"), NXLT("gp_lfill")}, /* 34x34 */
    { 7,238, NXLT("gp_color"), NXLT("gp_color")}, /* 14x14 */
    {27,238, NXLT("gp_color"), NXLT("gp_color")}, /* 14x14 */

    /* line thickness (should be 8x20) */
    {80,267, NXLT("gpline0"), NXLT("gpline0")}, /*  8x20 */
    {67,267, NXLT("gpline1"), NXLT("gpline1")}, /*  8x20 */
    {53,267, NXLT("gpline2"), NXLT("gpline2")}, /*  8x20 */
    {39,267, NXLT("gpline4"), NXLT("gpline4")}, /*  8x20 */
    {25,267, NXLT("gpline6"), NXLT("gpline6")}, /*  8x20 */
    { 7,267, NXLT("gplinex"), NXLT("gplinex")}, /* 12x20 */

    /* line style */
    { 7,319, NXLT("gp_ls_01"), NXLT("gp_ls_01")}, /* 82x12 */
    { 7,297, NXLT("gp_aleft"), NXLT("gp_aleft")},
    {59,297, NXLT("gp_arigt"), NXLT("gp_arigt")},
    {41,297, NXLT("gp_mkr00"), NXLT("gp_mkr00")},

    /* line style bitmaps (inactive for MOTIF) */
    {21,329, NXLT("gplinsol"), ""},
    {16,305, NXLT("gp_aleft"), ""},
    {57,305, NXLT("gp_arigt"), ""},
    {45,305, NXLT("gp_mkr00"), ""},

    /* pixed tools */
    { 7,  4, NXLT("pxstroke"), NXLT("pxstroke")},
    {51,  4, NXLT("px_wash"),  NXLT("px_wash")},
    { 7, 50, NXLT("pxsprayc"), NXLT("pxsprayc")},
    {51, 50, NXLT("px_blend"), NXLT("px_blend")},
    { 7, 96, NXLT("px_erase"), NXLT("px_erase")},
    {51, 96, NXLT("pxcerase"), NXLT("pxcerase")},
    { 7,142, NXLT("px_line"),  NXLT("px_line")},
    {51,142, NXLT("px_rect"),  NXLT("px_rect")},
    { 7,188, NXLT("px_ell"),   NXLT("px_ell")},
    {51,188, NXLT("px_flood"), NXLT("px_flood")},
    { 7,234, NXLT("pxfilter"), NXLT("pxfilter")},
    {51,234, NXLT("px_reg"),   NXLT("px_reg")},
    { 9,284, NXLT("gp_color"), NXLT("gp_color")},
    {53,284, NXLT("gp_color"), NXLT("gp_color")},
};


static char *leftstrs[] = {
    NXLT("gp_l_a01"),
    NXLT("gp_l_a02"),
    NXLT("gp_l_a03"),
    NXLT("gp_l_a04"),
    NXLT("gp_l_a05"),
    NXLT("gp_l_a06"),
    NXLT("gp_l_a07"),
    NXLT("gp_l_a08"),
    NXLT("gp_l_a09"),
    NXLT("gp_l_a10"),
    NXLT("gp_l_a11"),
    NXLT("gp_l_a12"),
    NXLT("gp_l_a13"),
    NULL};
static char *ol_midstrs[] = {
    NXLT("gplinsol"),
    NXLT("gpoldash"),
    NXLT("gp_oldot"),
    NXLT("gpolddot"),
    NXLT("gpoltdot"),
    NXLT("gpoldddt"),
    NXLT("gpolldsh"),
    NULL};
static char *midstrs[] = {
    NXLT("gp_ls_01"),
    NXLT("gp_ls_02"),
    NXLT("gp_ls_03"),
    NXLT("gp_ls_04"),
    NXLT("gp_ls_05"),
    NXLT("gp_ls_06"),
    NXLT("gp_ls_07"),
    NULL};
static char *rightstrs[] = {
    NXLT("gp_r_a01"),
    NXLT("gp_r_a02"),
    NXLT("gp_r_a03"),
    NXLT("gp_r_a04"),
    NXLT("gp_r_a05"),
    NXLT("gp_r_a06"),
    NXLT("gp_r_a07"),
    NXLT("gp_r_a08"),
    NXLT("gp_r_a09"),
    NXLT("gp_r_a10"),
    NXLT("gp_r_a11"),
    NXLT("gp_r_a12"),
    NXLT("gp_r_a13"),
    NULL};
static char *mrkrstrs[] = {
    NXLT("gp_mkr00"),
    NXLT("gp_mkr01"),
    NXLT("gp_mkr02"),
    NXLT("gp_mkr03"),
    NXLT("gp_mkr04"),
    NXLT("gp_mkr05"),
    NXLT("gp_mkr06"),
    NXLT("gp_mkr07"),
    NXLT("gp_mkr08"),
    NXLT("gp_mkr09"),
    NXLT("gp_mkr10"),
    NXLT("gp_mkr11"),
    NXLT("gp_mkr12"),
    NXLT("gp_mkr13"),
    NULL};

static int fillvals[] = { /* from pat id to widget val */
     0,  1,  3,  5,  7,  9, 10, 11, 12, 13,
    14,  4, 16,  2,  8, 21, 22, 24, 23, 17,
    18, 15,  6, 19, 20, 25, 26, 27, 28, 29,
};
static int fill_ids[] = { /* from widget val to pat id */
     0,  1, 13,  2, 11,  3, 22,  4, 14,  5,
     6,  7,  8,  9, 10, 21, 12, 19, 20, 23,
    24, 15, 16, 18, 17, 25, 26, 27, 28, 29,
};
static int percents[] = { /* from widget val to percentage */
    0, 0, 125, 250, 375, 500, 625, 750, 875, 1000,
};
static char *fillmaps[] = {
    NXLT("gp_ff00"),
    NXLT("gp_ff01"),
    NXLT("gp_ff13"),
    NXLT("gp_ff02"),
    NXLT("gp_ff11"),
    NXLT("gp_ff03"),
    NXLT("gp_ff19"),
    NXLT("gp_ff04"),
    NXLT("gp_ff14"),
    NXLT("gp_ff05"),

    NXLT("gp_ff06"),
    NXLT("gp_ff07"),
    NXLT("gp_ff08"),
    NXLT("gp_ff09"),
    NXLT("gp_ff10"),
    NXLT("gp_ff20"),
    NXLT("gp_ff12"),
    NXLT("gp_ff21"),
    NXLT("gp_ff22"),
    NXLT("gp_ff23"),

    NXLT("gp_ff24"),
    NXLT("gp_ff15"),
    NXLT("gp_ff16"),
    NXLT("gp_ff18"),
    NXLT("gp_ff17"),
    NXLT("gp_ff25"),
    NXLT("gp_ff26"),
    NXLT("gp_ff27"),
    NXLT("gp_ff28"),
    NXLT("gp_ff29"),
    NULL};
static char *fillstrs[31];

#include <palette.p>

void
init_fillstrs()
{
    static int  inited = FALSE;
    if (!inited)
    {
        fillstrs[0] = XLT("Fill: None");
        fillstrs[1] = XLT("Fill: 0% foreground");
        fillstrs[2] = XLT("Fill: 12.5% foreground");
        fillstrs[3] = XLT("Fill: 25% foreground");
        fillstrs[4] = XLT("Fill: 37.5% foreground");
        fillstrs[5] = XLT("Fill: 50% foreground");
        fillstrs[6] = XLT("Fill: 62.5% foreground");
        fillstrs[7] = XLT("Fill: 75% foreground");
        fillstrs[8] = XLT("Fill: 87.5% foreground");
        fillstrs[9] = XLT("Fill: 100% foreground");

        fillstrs[10] = XLT("Fill: Vertical Hatch");
        fillstrs[11] = XLT("Fill: Horizontal Hatch");
        fillstrs[12] = XLT("Fill: Horizontal-Vertical Crosshatch");
        fillstrs[13] = XLT("Fill: Diagonal Hatch");
        fillstrs[14] = XLT("Fill: Back Diagonal Hatch");
        fillstrs[15] = XLT("Fill: Diagonal Crosshatch");
        fillstrs[16] = XLT("Fill: Mixed Crosshatch");
        fillstrs[17] = XLT("Fill: Diamonds");
        fillstrs[18] = XLT("Fill: Squares");
        fillstrs[19] = XLT("Fill: Waves");

        fillstrs[20] = XLT("Fill: ZigZag");
        fillstrs[21] = XLT("Fill: Scales");
        fillstrs[22] = XLT("Fill: Weave");
        fillstrs[23] = XLT("Fill: Bricks");
        fillstrs[24] = XLT("Fill: Dots");
        fillstrs[25] = XLT("Fill: Diagonal Bricks");
        fillstrs[26] = XLT("Fill: Linear Gradient");
        fillstrs[27] = XLT("Fill: Radial Gradient");
        fillstrs[28] = XLT("Fill: Rectangular Gradient");
        fillstrs[29] = XLT("Fill: Custom Bitmap");

        fillstrs[30] = NULL;

        inited = TRUE;
    }
}
static char *linestrs[31];

void
init_linestrs()
{
    static int  inited = FALSE;
    if (!inited)
    {
        linestrs[0] = XLT("Line Fill: None");
        linestrs[1] = XLT("Line Fill: 0% foreground");
        linestrs[2] = XLT("Line Fill: 12.5% foreground");
        linestrs[3] = XLT("Line Fill: 25% foreground");
        linestrs[4] = XLT("Line Fill: 37.5% foreground");
        linestrs[5] = XLT("Line Fill: 50% foreground");
        linestrs[6] = XLT("Line Fill: 62.5% foreground");
        linestrs[7] = XLT("Line Fill: 75% foreground");
        linestrs[8] = XLT("Line Fill: 87.5% foreground");
        linestrs[9] = XLT("Line Fill: 100% foreground");

        linestrs[10] = XLT("Line Fill: Vertical Hatch");
        linestrs[11] = XLT("Line Fill: Horizontal Hatch");
        linestrs[12] = XLT("Line Fill: Horizontal-Vertical Crosshatch");
        linestrs[13] = XLT("Line Fill: Diagonal Hatch");
        linestrs[14] = XLT("Line Fill: Back Diagonal Hatch");
        linestrs[15] = XLT("Line Fill: Diagonal Crosshatch");
        linestrs[16] = XLT("Line Fill: Mixed Crosshatch");
        linestrs[17] = XLT("Line Fill: Diamonds");
        linestrs[18] = XLT("Line Fill: Squares");
        linestrs[19] = XLT("Line Fill: Waves");

        linestrs[20] = XLT("Line Fill: ZigZag");
        linestrs[21] = XLT("Line Fill: Scales");
        linestrs[22] = XLT("Line Fill: Weave");
        linestrs[23] = XLT("Line Fill: Bricks");
        linestrs[24] = XLT("Line Fill: Dots");
        linestrs[25] = XLT("Line Fill: Diagonal Bricks");
        linestrs[26] = XLT("Line Fill: Linear Gradient");
        linestrs[27] = XLT("Line Fill: Radial Gradient");
        linestrs[28] = XLT("Line Fill: Rectangular Gradient");
        linestrs[29] = XLT("Line Fill: Custom Bitmap");

        linestrs[30] = NULL;
        
        inited = TRUE;
    }
}
/****************************************************************************/

#define CALLBACKER(cbuf,func,more) {cbuf[0].callback = (XtCallbackProc)(func); cbuf[0].closure = (caddr_t)(more);}

extern AXWidget_popup CurrentPopupButton;


/****************************************************************************/
/* a global that should only be set during TmCreate (and immediately unset!) */
static int ge_as_pixed = FALSE;
void AxSetPixedPalette()
{
    ge_as_pixed = TRUE;
}
void AxUnsetPixedPalette()
{
    ge_as_pixed = FALSE;
}

/****************************************************************************/
void AxLoadPaletteColormap(w, n_colors, c_names, c_maps, fg_vals, bg_vals, wide)
AXWidget_main w;
int n_colors;
char **c_names;
char **c_maps;
int *fg_vals;
int *bg_vals;
int wide;
{
	int an, wID;
	Arg args[7];

	if ( !w || !PaletteList )
		return;
	wID = AxWindowIdFromShellWidget(w);
	if ( wID == -1 || !PaletteList[wID] )
		return;

	/* load up the args */
	an = 0;
	XtSetArg(args[an], XmNitemCount,      n_colors); an++;
	XtSetArg(args[an], AXmNbtnPixmapList, c_maps);   an++;
	XtSetArg(args[an], XmNitems,          c_names);  an++;
	XtSetArg(args[an], XmNfg_color,       fg_vals);  an++;
	XtSetArg(args[an], XmNbg_color,       bg_vals);  an++;
	XtSetArg(args[an], XmNorientation,    wide?0:5); an++;

	/* set the args onto the approriate widgets */
    if ( PaletteWidgets[wID][PDEX_shadowColor] )
        AxXtSetValues(PaletteWidgets[wID][PDEX_shadowColor], args, an);
    if ( PaletteWidgets[wID][PDEX_fill_fg] )
        AxXtSetValues(PaletteWidgets[wID][PDEX_fill_fg],     args, an);
    if ( PaletteWidgets[wID][PDEX_fill_bg] )
        AxXtSetValues(PaletteWidgets[wID][PDEX_fill_bg],     args, an);
    if ( PaletteWidgets[wID][PDEX_pxfill_fg] )
        AxXtSetValues(PaletteWidgets[wID][PDEX_pxfill_fg],   args, an);
    if ( PaletteWidgets[wID][PDEX_pxfill_bg] )
        AxXtSetValues(PaletteWidgets[wID][PDEX_pxfill_bg],   args, an);
    if ( PaletteWidgets[wID][PDEX_line_fg] )
        AxXtSetValues(PaletteWidgets[wID][PDEX_line_fg],     args, an);
    if ( PaletteWidgets[wID][PDEX_line_bg] )
        AxXtSetValues(PaletteWidgets[wID][PDEX_line_bg],     args, an);
}

/****************************************************************************/
int line_thickness(palP)
GEPalette_t *palP;
{
    switch ( palP->lineThickness )
    {
        case  0: return(PDEX_line0);
        case  1: return(PDEX_line1);
        case  2: return(PDEX_line2);
        case  3: return(PDEX_line4);
        case  4: return(PDEX_line6);
        default: return(PDEX_linex);
    }
}

/****************************************************************************/
pltdata *
getPaletteEntry(index)
int index;
{
    return(&PaletteBasics[index]);
}

/****************************************************************************/
static void
GEscalerChanged(widget, wid, data)
AXWidget widget;
int wid;
XmRowColumnCallbackStruct *data;
{
}

/****************************************************************************/
static void
GEPaletteTask(args)
char *args;
{
	int pid;
	VfyContext(0);
	ACprocIt(APELFM, args, &pid, FALSE, FALSE, "GR_PALETTE_LAUNCHER$", FALSE);
}

/****************************************************************************/
int
AxSetPalette(w, newP, parts)
AXWidget_main w;
GEPalette_t *newP;
int parts;
{
    int new_ffg, new_fbg, new_lfg, new_lbg, new_scolor,new_ribbon;
    int newFill, newLinefill, newArrows, newLineStyle, newMarkers;
    int scheme_ix,mode,repaintFlag;
    int tool, new_toolchoice = FALSE;
    char str[STRPATH];
    graphics_main_type *ge;
    GEPalette_t *oldP;
    int winID, pid;

    /* sanity test */
    if ( TmAmSE == TRUE || !w || w->type != W_MAIN )
        return(!AOK);

    /* need a window to avoid a crash - return if we don't have one */
    winID = AxWindowIdFromShellWidget(w);
    if ( winID == -1 )
        return(!AOK);

    /* get the graphic info from the task (must be an APGE task) */
    pid = PidFromATMWindow(winID);
    if ( AxIsGraphicsTask(pid) == FALSE )
        return(!AOK);
    ge = (graphics_main_type *)AxGetGFXfromPID(pid);

    /* first time for this window - initialize the palette */
    if ( !PaletteList || !PaletteList[winID] )
    {
        InitPalette(w);
        PaintGEPalette(w->palette, (XExposeEvent *) NULL, TRUE, TRUE);
    }

    /* get a snapshot of the palette before any changes are made */
    oldP = PaletteList[winID];

    /* take care of the UDO list box */
    if ( parts )
    {
        PartsListMan ( w,
                       newP->parts,
                       newP->toolType-PD_NUM_TOOLS,
                       newP->toolType < PIX_START ? BUT_UP : BUT_HIDE);
    }

    /* has the current tool changed? */
    if ( oldP->toolType != newP->toolType || oldP->repaintFlag )
    {
        /* lift the old tool */
        if ( oldP->toolType >= 0 )
        {
            tool = oldP->toolType;
            if ( ((tool < PD_NUM_TOOLS) && (newP->toolType < PIX_START)) ||
                 ((tool >= PIX_START) && (newP->toolType >= PIX_START)) )
            {
                tool = (tool >= PIX_START) ? tool-PIX_START : tool;
                if(!newP->editingPix && tool >= PDEX_tool1 && tool <= PDEX_tool7)
                    WidgetMan(tool,w,BUT_HIDE);
                else
                    WidgetMan(tool,w,BUT_UP);
            }

            /* if the list box hasn't been dealt with, do so now */
            if ( !parts )
                PartsListMan(w,
                             (char**)NULL,
                             oldP->toolType-PD_NUM_TOOLS,
                             oldP->toolType < PIX_START ? BUT_UP : BUT_HIDE);
        }

        /* depress the new tool */
        if ( newP->toolType >= 0 )
        {
            tool = newP->toolType;
            if ( (tool < PD_NUM_TOOLS) || (tool >= PIX_START) )
            {
                tool = (tool >= PIX_START) ? tool-PIX_START : tool;
                if(!newP->editingPix && tool >= PDEX_tool1 && tool <= PDEX_tool6)
                    WidgetMan(tool,w,BUT_HIDE);
                tool = (tool >= PIX_START) ? tool-PIX_START : tool;
                    WidgetMan(tool,w,BUT_DOWN);
            }

            /* if the list box hasn't been dealt with, do so now */
            if ( !parts )
                PartsListMan(w,
                             (char**)NULL,
                             newP->toolType-PD_NUM_TOOLS,
                             newP->toolType < PIX_START ? BUT_UP : BUT_HIDE);
        }

        /* note that the cursor may need to change */
        new_toolchoice = TRUE;
    }

    /* has the line thickness changed? */
    if ( oldP->lineThickness != newP->lineThickness || oldP->repaintFlag )
    {
        if ( oldP->lineThickness >= 0 )
            WidgetMan(line_thickness(oldP), w, BUT_UP);

        if ( newP->lineThickness >= 0 )
            WidgetMan(line_thickness(newP), w, BUT_DOWN);
    }

    /* if the drop shadow button was pressed, lift it */
    if ( oldP->shadowBoxUp || oldP->repaintFlag )
    {
        WidgetMan(PDEX_shadowButton,w,newP->editingPix?BUT_HIDE:BUT_UP);
        newP->shadowBoxUp = FALSE;
    }

    /* get the ribbon fill type (fg, bg, etc) option value */
    AxGetColorpalMode(w, &mode);
    newP->ribbon = mode;

    /* update the state for parts who need current state to draw correctly */
    newLineStyle = oldP->lineStyle != newP->lineStyle;
    newArrows = oldP->arrow1 != newP->arrow1 || oldP->arrow2 != newP->arrow2;
    newMarkers = oldP->join_symbols != newP->join_symbols;
    new_ffg = (oldP->ffg != newP->ffg) || oldP->repaintColors;
    new_fbg = (oldP->fbg != newP->fbg) || oldP->repaintColors;
    new_lfg = (oldP->lfg != newP->lfg) || oldP->repaintColors;
    new_lbg = (oldP->lbg != newP->lbg) || oldP->repaintColors;
    new_scolor = (oldP->shadow_color != newP->shadow_color) ||
                       oldP->repaintColors;
    new_ribbon = oldP->ribbon != newP->ribbon;
    newFill = oldP->fillPattern != newP->fillPattern;
    newLinefill = oldP->linefillPattern != newP->linefillPattern;
    repaintFlag = oldP->repaintFlag != newP->repaintFlag;
    *PaletteList[winID] = *newP;

    /* has the line style changed? */
    if ( newLineStyle || repaintFlag )
    {
        WidgetMan(PDEX_stylemap, w, BUT_UP);
        WidgetMan(PDEX_lstyle,   w, BUT_UP);
    }

    /* have end-styles changed? */
    if ( newArrows || repaintFlag )
    {
        WidgetMan(PDEX_aright,  w, BUT_UP);
        WidgetMan(PDEX_aleft,   w, BUT_UP);
        WidgetMan(PDEX_rendmap, w, BUT_UP);
        WidgetMan(PDEX_lendmap, w, BUT_UP);
    }

    /* have line join markers changed? */
    if ( newMarkers || repaintFlag )
    {
        WidgetMan(PDEX_markmap, w, BUT_UP);
        WidgetMan(PDEX_mrkr,    w, BUT_UP);
    }

    /* have the colors changed */
    if ( new_ffg || repaintFlag )
    {
        WidgetMan(PDEX_pxfill_fg,w,newP->editingPix?BUT_UP:BUT_HIDE);
        WidgetMan(PDEX_fill_fg,  w,newP->editingPix?BUT_HIDE:BUT_UP);
        scheme_ix = newP->ffg;
    }
    if ( new_fbg || repaintFlag )
    {
        WidgetMan(PDEX_pxfill_bg,w,newP->editingPix?BUT_UP:BUT_HIDE);
        WidgetMan(PDEX_fill_bg,w,newP->editingPix?BUT_HIDE:BUT_UP);
        scheme_ix = newP->fbg;
    }
    if ( new_lfg || repaintFlag )
    {
        WidgetMan(PDEX_line_fg,w,newP->editingPix?BUT_HIDE:BUT_UP);
        scheme_ix = newP->lfg;
    }
    if ( new_lbg || repaintFlag )
    {
        WidgetMan(PDEX_line_bg,w,newP->editingPix?BUT_HIDE:BUT_UP);
        scheme_ix = newP->lbg;
    }
    if ( new_scolor || repaintFlag )
    {
        WidgetMan(PDEX_shadowColor,w,newP->editingPix?BUT_HIDE:BUT_UP);
        scheme_ix = newP->shadow_color;
    }

    /* has the fill pattern changed? */
    if ( newP->fillPattern == 21 ||
         newFill || new_ffg || new_fbg || repaintFlag )
    {
        WidgetMan(PDEX_fillpat,w,newP->editingPix?BUT_HIDE:BUT_UP);
    }

    /* has the line pattern changed? */
    if ( newP->linefillPattern == 21 ||
         newLinefill || new_lfg || new_lbg || repaintFlag )
    {
        WidgetMan(PDEX_linepat,w,newP->editingPix?BUT_HIDE:BUT_UP);
    }

    /* update the cursor if tool has changed */
    if ( new_toolchoice )
        GESetPointerFromPalette(w);

	/* update the color ribbon */
    if ( new_lfg || new_lbg || new_ffg  || new_fbg ||
        new_scolor || new_ribbon || repaintFlag )
    {
        switch(newP->ribbon)
        {
            case 0: scheme_ix = newP->ffg;break;
            case 1: scheme_ix = newP->fbg;break;
            case 2: scheme_ix = newP->lfg;break;
            case 3: scheme_ix = newP->lbg;break;
            case 4: scheme_ix = newP->shadow_color;break;
        }
        name_from_scheme(ge, scheme_ix, str);
        AxSetRibbon(w, scheme_ix, str, newP->editingPix);
    }

    /* reset and return */
    oldP->repaintColors = FALSE;
    oldP->repaintFlag = FALSE;
    return(AOK);
}
/****************************************************************************/
int
TmSetPalette(wd, newP, parts)
long wd;
GEPalette_t *newP;
int parts;
{
	return(AxSetPalette((AXWidget_main)wd, newP, parts));
}

/****************************************************************************/
static void
FinishLeftChoice(w, wid,data)
AXWidget_popup w;
int wid;        /* winID as set in widgetman call */
XmRowColumnCallbackStruct *data;
{
    Arg args[20];
    GEPalette_t *palP, newPalette;
    int value;
    char elfArgString[STRPATH];

    /* release the button */
    XtSetArg(args[0],XmNvalue,FALSE);
    AxXtSetValues((AXWidget)CurrentPopupButton,args,1);

    value = (int)data->data;
    palP = PaletteList[wid];
    newPalette = *palP;
    if ( newPalette.arrow1 != value )
    {
        sprintf(elfArgString, NXLT("%d GE_PAL_SET_LINE_END1$ %d "), wid, value);
		GEPaletteTask(elfArgString);

        /* update the palette */
        newPalette.arrow1 = value;
        AxSetPalette(w->parent->parent, &newPalette, FALSE);
        *palP = newPalette;
    }
}

/****************************************************************************/
static void
FinishMidChoice(w, wid, data)
AXWidget_popup w;
int wid;        /* winID as set in widgetman call */
XmRowColumnCallbackStruct *data;
{
    Arg args[20];
    GEPalette_t *palP, newPalette;
    int value;
    char elfArgString[STRPATH];

    /* release the button */
    XtSetArg(args[0],XmNvalue,FALSE);
    AxXtSetValues((AXWidget)CurrentPopupButton,args,1);

    value = (int)data->data;
    palP = PaletteList[wid];
    newPalette = *palP;
    if ( newPalette.lineStyle != value )
    {
        /* update the ge */
        sprintf(elfArgString, NXLT("%d GE_PAL_SET_LINE_STYLE$ %d "), wid,value);
		GEPaletteTask(elfArgString);

        /* update the palette */
        newPalette.lineStyle = value;
        AxSetPalette(w->parent->parent, &newPalette, FALSE);
        *palP = newPalette;
    }
}

/****************************************************************************/
static void
FinishRightChoice(w, wid, data)
AXWidget_popup w;
int wid;        /* winID as set in widgetman call */
XmRowColumnCallbackStruct *data;
{
    Arg args[20];
    GEPalette_t *palP, newPalette;
    int value;
    char elfArgString[STRPATH];

    /* release the button */
    XtSetArg(args[0],XmNvalue,FALSE);
    AxXtSetValues((AXWidget)CurrentPopupButton,args,1);

    value = (int)data->data;
    palP = PaletteList[wid];
    newPalette = *palP;
    if ( newPalette.arrow2 != value )
    {
        /* update the ge */
        sprintf(elfArgString, NXLT("%d GE_PAL_SET_LINE_END2$ %d "), wid, value);
		GEPaletteTask(elfArgString);

        /* update the palette */
        newPalette.arrow2 = value;
        AxSetPalette(w->parent->parent, &newPalette, FALSE);
        *palP = newPalette;
    }
}

/****************************************************************************/
static void
FinishMrkrChoice(w, wid, data)
AXWidget_popup w;
int wid;        /* winID as set in widgetman call */
XmRowColumnCallbackStruct *data;
{
    Arg args[20];
    GEPalette_t *palP, newPalette;
    int value;
    char elfArgString[STRPATH];

    /* release the button */
    XtSetArg(args[0],XmNvalue,FALSE);
    AxXtSetValues((AXWidget)CurrentPopupButton,args,1);

    value = (int)data->data;
    palP = PaletteList[wid];
    newPalette = *palP;
    if ( newPalette.join_symbols != value )
    {
        /* update the ge */
        sprintf(elfArgString, NXLT("%d GE_PAL_SET_LINE_JOIN$ %d "), wid, value);
		GEPaletteTask(elfArgString);

        /* update the palette */
        newPalette.join_symbols = value;
        AxSetPalette(w->parent->parent, &newPalette, FALSE);
        *palP = newPalette;
    }
}

/****************************************************************************/
static void FinishFFillPatChoice(widget,wid,data)
AXWidget_rowcol widget;
int wid;        /* winID as set in widgetman call */
XmRowColumnCallbackStruct *data;
{
    char elfArgString[STRPATH], *ElfSymbolNameFromId();
    GEPalette_t *palP;
    int val;

    /* get the new value */
    val = (int)data->data;
    palP = PaletteList[wid];
    if ( fill_ids[val] == palP->fillPattern && val != 29 )
        return; 
    if ( val < 0 || val > 29 )
        return; 

    if ( val == 0 || (val > 9 && val <= 25 ) ) /* indexed pattern */
    {
        sprintf(elfArgString, NXLT("%d GR_SET_FILL@ %d "), wid, fill_ids[val]);
		GEPaletteTask(elfArgString);
    }
    else if ( val >= 1 && val <= 9 ) /* percentage pattern */
    {
        sprintf(elfArgString, NXLT("%d GR_SET_PERCENT_FILL@ %d "),
		        wid,
		        percents[val]);
		GEPaletteTask(elfArgString);
    }
    else if ( val == 26 ) /* linear gradient */
    {
        sprintf(elfArgString, NXLT("%d GE_PROCESS_FILL_GRAD_FOR_PAL 0"), wid); 
		GEPaletteTask(elfArgString);
    }
    else if ( val == 27 ) /* radial gradient */
    {
        sprintf(elfArgString, NXLT("%d GE_PROCESS_FILL_GRAD_FOR_PAL 1"), wid); 
		GEPaletteTask(elfArgString);
    }
    else if ( val == 28 ) /* rectangular gradient */
    {
        sprintf(elfArgString, NXLT("%d GE_PROCESS_FILL_GRAD_FOR_PAL 2"), wid); 
		GEPaletteTask(elfArgString);
    }
    else if ( val == 29 ) /* user defined (don't update the palette) */
    {
        sprintf(elfArgString, NXLT("%d GR_BITFILL_DLG@"), wid);
		GEPaletteTask(elfArgString);
    }
}

/****************************************************************************/
static void FinishLFillPatChoice(widget,wid,data)
AXWidget_rowcol widget;
int wid;        /* winID as set in widgetman call */
XmRowColumnCallbackStruct *data;
{
    char elfArgString[STRPATH], *ElfSymbolNameFromId();
    GEPalette_t *palP;
    int val;

    /* get the new value */
    val = (int)data->data;
    palP = PaletteList[wid];
    if ( fill_ids[val] == palP->linefillPattern && val != 29 )
        return;
    if ( val < 0 || val > 29 )
        return;

    if ( val == 0 || (val > 9 && val <= 25 ) ) /* indexed pattern */
    {
        sprintf(elfArgString, NXLT("%d GR_SET_LINE_FILL@ %d "),
		        wid,
                fill_ids[val]);
		GEPaletteTask(elfArgString);
    }
    else if ( val >= 1 && val <= 9 ) /* percentage pattern */
    {
        sprintf(elfArgString, NXLT("%d GR_SET_PERCENT_LINE_FILL@ %d "),
                wid,
		        percents[val]);
		GEPaletteTask(elfArgString);
    }
    else if ( val == 26 ) /* linear gradient */
    {
        sprintf(elfArgString, NXLT("%d GE_PROCESS_LINE_GRAD_FOR_PAL 0"), wid);
		GEPaletteTask(elfArgString);
    }
    else if ( val == 27 ) /* radial gradient */
    {
        sprintf(elfArgString, NXLT("%d GE_PROCESS_LINE_GRAD_FOR_PAL 1"), wid);
		GEPaletteTask(elfArgString);
    }
    else if ( val == 28 ) /* rectangular gradient */
    {
        sprintf(elfArgString, NXLT("%d GE_PROCESS_LINE_GRAD_FOR_PAL 2"), wid);
		GEPaletteTask(elfArgString);
    }
    else if ( val == 29 ) /* user defined */
    {
        sprintf(elfArgString, NXLT("%d GR_BITLINE_DLG@"), wid);
		GEPaletteTask(elfArgString);
    }
}

/****************************************************************************/
static void
FinishFFillFgChoice(widget,wid,data)
AXWidget_rowcol widget;
int wid;        /* winID as set in widgetman call */
XmRowColumnCallbackStruct *data;
{
	char elfArgString[STRPATH];
	int val;

	val = (int)data->data;
	if ( val != PaletteList[wid]->ffg )
	{
		sprintf(elfArgString, NXLT("%d GR_SET_COLOR_BY_CMAP_INDEX@ %d 0"),
		        wid,
		        val);
		GEPaletteTask(elfArgString);
	}
}

/****************************************************************************/
static void
FinishFFillBgChoice(widget,wid,data)
AXWidget_rowcol widget;
int wid;        /* winID as set in widgetman call */
XmRowColumnCallbackStruct *data;
{
	char elfArgString[STRPATH];
	int val;

	val = (int)data->data;
	if ( val != PaletteList[wid]->fbg )
	{
		sprintf(elfArgString, NXLT("%d GR_SET_COLOR_BY_CMAP_INDEX@ %d 1"),
		        wid,
		        val);
		GEPaletteTask(elfArgString);
	}
}

/****************************************************************************/
static void
FinishLFillFgChoice(widget,wid,data)
AXWidget_rowcol widget;
int wid;        /* winID as set in widgetman call */
XmRowColumnCallbackStruct *data;
{
	char elfArgString[STRPATH];
	int val;

	val = (int)data->data;
	if ( val != PaletteList[wid]->lfg )
	{
		sprintf(elfArgString, NXLT("%d GR_SET_COLOR_BY_CMAP_INDEX@ %d 2"),
		        wid,
		        val);
		GEPaletteTask(elfArgString);
	}
}

/****************************************************************************/
static void
FinishLFillBgChoice(widget,wid,data)
AXWidget_rowcol widget;
int wid;        /* winID as set in widgetman call */
XmRowColumnCallbackStruct *data;
{
	char elfArgString[STRPATH];
	int val;

	val = (int)data->data;
	if ( val != PaletteList[wid]->lbg )
	{
		sprintf(elfArgString, NXLT("%d GR_SET_COLOR_BY_CMAP_INDEX@ %d 3"),
		        wid,
		        val);
		GEPaletteTask(elfArgString);
	}
}

/****************************************************************************/
static void
FinishShadowColorChoice(widget,wid,data)
AXWidget_rowcol widget;
int wid;        /* winID as set in widgetman call */
XmRowColumnCallbackStruct *data;
{
	char elfArgString[STRPATH];
	int val;

	val = (int)data->data;
	if ( val != PaletteList[wid]->shadow_color )
	{
		sprintf(elfArgString, NXLT("%d GR_SET_COLOR_BY_CMAP_INDEX@ %d 4"),
		        wid,
		        val);
		GEPaletteTask(elfArgString);
	}
}

/****************************************************************************/
int getButtonType(pos)
int pos;
{
    switch(pos)
    {
        /* color row-col buttons */
        case PDEX_fill_fg:
        case PDEX_fill_bg:
        case PDEX_line_fg:
        case PDEX_line_bg:
        case PDEX_shadowColor:
        case PDEX_pxfill_fg:
        case PDEX_pxfill_bg:
            return(0);

        /* main ge tool array */
        case PDEX_point:
        case PDEX_zoom:
        case PDEX_box:
        case PDEX_square:
        case PDEX_ellip:
        case PDEX_circle:
        case PDEX_poly:
        case PDEX_stk:
        case PDEX_curve:
        case PDEX_line:
        case PDEX_alpha:
        case PDEX_alphabox:
        case PDEX_image:
        case PDEX_c_image:
            return(1);
    
        /* pixel editor tool array */
        case PDEX_tool1:
        case PDEX_tool2:
        case PDEX_tool3:
        case PDEX_tool4:
        case PDEX_tool5:
        case PDEX_tool6:
        case PDEX_tool7:
        case PDEX_tool8:
        case PDEX_tool9:
        case PDEX_tool10:
        case PDEX_tool11:
        case PDEX_tool12:
            return(2);
    
        /* drop shadow dialog button */
        case PDEX_shadowButton:
            return(3);

        /* fill pattern row-column widget */
        case PDEX_fillpat:
            return(4);

        /* line fill pattern row-column widget */
        case PDEX_linepat:
            return(5);

        /* color row-column widgets */
            /*  return(6);  */

        /* openlookable pulldown widgets */
        case PDEX_aright:
        case PDEX_aleft:
        case PDEX_lstyle:
        case PDEX_mrkr:
            return(7);

        default:
            return(8);
    
        /* line thickness button array */
        case PDEX_linex:
        case PDEX_line6:
        case PDEX_line4:
        case PDEX_line2:
        case PDEX_line1:
        case PDEX_line0:
            return(9);

    }
}

/*************************************************************************
function: DisplayPopup - Produce the pulldown
*/
static void
DisplayPopup(w, wid, pos)
AXWidget_palette w;
int wid;
int pos;
{
    Arg args[20];
    int an = 0;
    int bx,by,bw,bh;    /* parent button info */
    int xadj, yadj;
    static XtCallbackRec Callback[2];   /* standard template */
    char **strs;
    int arrow1, arrow2, lineStyle, mrkr, void_val;

    CurrentPopupButton = (AXWidget_popup)PaletteWidgets[wid][pos];
    arrow1 = PaletteList[wid]->arrow1;
    arrow2 = PaletteList[wid]->arrow2;
    lineStyle = PaletteList[wid]->lineStyle;
    mrkr = PaletteList[wid]->join_symbols;
    void_val = 0;
    XmWidgetSizeAndPos((AXWidget)CurrentPopupButton,&bx,&by,&bw,&bh);
    yadj = 0;
    xadj = -4;
    XtSetArg(args[an],XmNoldWidget,0); an++;
    XtSetArg(args[an],XmNflags,0); an++;
    switch ( pos )
    {
        case PDEX_aleft:
            strs = leftstrs;
            CALLBACKER(Callback,FinishLeftChoice,wid);
            XtSetArg(args[an],XmNitemCount,13); an++;
            XtSetArg(args[an],XmNvalue,arrow1); an++;
            yadj -= arrow1*18 + 4;
        break;
        case PDEX_lstyle:
            strs = midstrs;
            CALLBACKER(Callback,FinishMidChoice,wid);
            XtSetArg(args[an],XmNitemCount,7); an++;
            XtSetArg(args[an],XmNvalue,lineStyle); an++;
            yadj -= lineStyle*18 + 4;
        break;
        case PDEX_aright:
            strs = rightstrs;
            CALLBACKER(Callback,FinishRightChoice,wid);
            XtSetArg(args[an],XmNitemCount,13); an++;
            XtSetArg(args[an],XmNvalue,arrow2); an++;
            yadj -= arrow2*18 + 4;
        break;
        case PDEX_mrkr:
            strs = mrkrstrs;
            CALLBACKER(Callback,FinishMrkrChoice,wid);
            XtSetArg(args[an],XmNitemCount,14); an++;
            XtSetArg(args[an],XmNvalue,mrkr); an++;
            yadj -= mrkr*18 + 4;
        break;
    }
    XtSetArg(args[an],XmNx,bx+xadj); an++;
    XtSetArg(args[an],XmNy,by+yadj); an++;
    XtSetArg(args[an],XmNitems,strs); an++;
    XtSetArg(args[an],XmNactivateCallback,Callback); an++;
    XtSetArg(args[an],XmNgePaletteButtonType,getButtonType(pos)); an++;
    AXmCreatePopupMenu(w, NXLT("arrows"), args, an);
}

/*************************************************************************
function: GEButtonByPos - Respond to widget activity in GE palette
*/
void GEButtonByPos(w, wid, ix,dbl_clck)
AXWidget_pushbutton w;     /* button */
int wid;        /* ATM window id           */
int ix;         /* widget index in palette */
int dbl_clck;  /* TRUE if dbl click event */
{
    register int val;
    GEPalette_t *palP, currPal;
    char elfArgString[STRPATH], *ElfSymbolNameFromId(), *str, *mstr;
    static AXWidget_palette curr;

    /* get the palette info */
    palP = PaletteList[wid];
    currPal = *palP;

    /* set the correct fields on the palette structure */
    curr = (AXWidget_palette)w->parent; /* palette */
    switch ( ix )
    {
        case PDEX_shadowButton:
            sprintf(elfArgString, "%d GR_SHADOW_DLG@", wid);
			GEPaletteTask(elfArgString);
            palP->shadowBoxUp = TRUE;
            AxSetPalette(curr->parent, &currPal, FALSE);
            *palP = currPal;
        break;

        case PDEX_point:
        case PDEX_zoom:
        case PDEX_alpha:
        case PDEX_alphabox:
        case PDEX_line:
        case PDEX_poly:
        case PDEX_box:
        case PDEX_ellip:
        case PDEX_stk:
        case PDEX_image:
        case PDEX_square:
        case PDEX_circle:
        case PDEX_curve:
        case PDEX_c_image:
            if ( atoi(GetProfileByCode(PROF_TOOL_LOCK_)) != -1 )
                mstr = dbl_clck ? NXLT("GR_TOOL_LOCK@") : NXLT("GR_TOOL_PICK@");
            else
                mstr = dbl_clck ? NXLT("GR_TOOL_PICK@") : NXLT("GR_TOOL_LOCK@");
            str = (char *)createToolString(ix);
            if ( str )
                sprintf(elfArgString, "%d %s %s", wid, mstr, str);
            else
                sprintf(elfArgString, "%d %s %d", wid, mstr, ix);
			GEPaletteTask(elfArgString);
            currPal.toolType = ix;
            currPal.toolLocked = dbl_clck ? TRUE : FALSE;
            AxSetPalette(curr->parent, &currPal, FALSE);
            *palP = currPal;
        break;

        case PDEX_tool1:
        case PDEX_tool2:
        case PDEX_tool3:
        case PDEX_tool4:
        case PDEX_tool5:
        case PDEX_tool6:
        case PDEX_tool7:
        case PDEX_tool8:
        case PDEX_tool9:
        case PDEX_tool10:
        case PDEX_tool11:
        case PDEX_tool12:
            val = PIX_START + ix;
            if ( atoi(GetProfileByCode(PROF_TOOL_LOCK_)) != -1 )
                mstr = dbl_clck ? NXLT("GR_TOOL_LOCK@") : NXLT("GR_TOOL_PICK@");
            else
                mstr = dbl_clck ? NXLT("GR_TOOL_PICK@") : NXLT("GR_TOOL_LOCK@");
            str = (char *)createToolString(val);
            if ( str )
                sprintf(elfArgString, "%d %s %s", wid, mstr, str);
            else
                sprintf(elfArgString, "%d %s %d", wid, mstr, val);
			GEPaletteTask(elfArgString);
            currPal.toolType = val;
            AxSetPalette(curr->parent, &currPal, FALSE);
            *palP = currPal;
        break;

        case PDEX_lstyle:
        case PDEX_aright:
        case PDEX_aleft:
        case PDEX_mrkr:
            DisplayPopup(curr, wid,ix);
        break;
    
        case PDEX_linex:
            palP->lineThickness = 5;
            AxSetPalette(curr->parent, &currPal, FALSE);
            *palP = currPal;
            sprintf(elfArgString, "%d GR_LINE_WIDTH_DLG@", wid);
			GEPaletteTask(elfArgString);
        break;

        case PDEX_line6:
            currPal.lineThickness = 4;
            AxSetPalette(curr->parent, &currPal, FALSE);
            *palP = currPal;
            sprintf(elfArgString, "%d GE_PAL_SET_LINE_WIDTH$ 4 ", wid);
			GEPaletteTask(elfArgString);
        break;
        case PDEX_line4:
            currPal.lineThickness = 3;
            AxSetPalette(curr->parent, &currPal, FALSE);
            *palP = currPal;
            sprintf(elfArgString, "%d GE_PAL_SET_LINE_WIDTH$ 3 ", wid);
			GEPaletteTask(elfArgString);
        break;
        case PDEX_line2:
            currPal.lineThickness = 2;
            AxSetPalette(curr->parent, &currPal, FALSE);
            *palP = currPal;
            sprintf(elfArgString, "%d GE_PAL_SET_LINE_WIDTH$ 2 ", wid);
			GEPaletteTask(elfArgString);
        break;
        case PDEX_line1:
            currPal.lineThickness = 1;
            AxSetPalette(curr->parent, &currPal, FALSE);
            *palP = currPal;
            sprintf(elfArgString, "%d GE_PAL_SET_LINE_WIDTH$ 1 ", wid);
			GEPaletteTask(elfArgString);
        break;
        case PDEX_line0:
            currPal.lineThickness = 0;
            AxSetPalette(curr->parent, &currPal, FALSE);
            *palP = currPal;
            sprintf(elfArgString, "%d GE_PAL_SET_LINE_WIDTH$ 0 ", wid);
			GEPaletteTask(elfArgString);
        break;

        default:
            AxSetPalette(curr->parent, &currPal, FALSE);
            *palP = currPal;
        break;
    }
}

/*************************************************************************
function: GEPaletteButtonPushed - Recieve a callback
*/
void GEPaletteButtonPushed (w,wid,data)
AXWidget_pushbutton w;
int wid;
XmAnyCallbackStruct *data;
{
    extern unsigned long TimeOfButtonPress; /* event time of button down */
    extern int DoubleMillisec;  /* dbl click interval */
    unsigned long interval;
    static unsigned long LastTime;
    static AXWidget_pushbutton LastWidget;
    register int ix;
    int dbl_click;

    /* find the widget in the widow's palette list */
    for ( ix=0; ix<PALT_NUM_BUTTONS; ix++ )
        if ( w == (AXWidget_pushbutton)PaletteWidgets[wid][ix] )
            break;

    if ( ix < PALT_NUM_BUTTONS )
    {
        /* determine if this is a double click */
        interval = TimeOfButtonPress - DoubleMillisec;
        dbl_click = (w == LastWidget && interval <= LastTime) ? TRUE : FALSE;

        /* note the last button activity */
        LastTime = TimeOfButtonPress;
        LastWidget = w;

        /* now dispatch the event */
        GEButtonByPos(w, wid, ix, dbl_click);
    }
}

/*************************************************************************
function: GEPalettePartSelect - select a palette part as the current tool
*/
void GEPalettePartSelect(w, winID, data)
AXWidget_palette w;
int winID;
XmListCallbackStruct *data;
{
    char elfArgString[STRPATH], *mstr;
    GEPalette_t *palP, currPal;
    extern int ElfLanguage;
    int pos;

    /* ignore events not sent to a window */
    if ( !winID )
        return;

    /* ignore selections that are not in the valid range */
    pos = data->item_position - 1;
    if ( pos < 0 )
        return;

    /* now select the part as the current tool */
    if ( atoi(GetProfileByCode(PROF_TOOL_LOCK_)) != -1 )
    {
        if ( data->reason == XmCR_DEFAULT_ACTION )
            mstr = NXLT("GR_TOOL_LOCK@");
        else
            mstr = NXLT("GR_TOOL_PICK@");
    }
    else
    {
        if ( data->reason == XmCR_DEFAULT_ACTION )
            mstr = NXLT("GR_TOOL_PICK@");
        else
            mstr = NXLT("GR_TOOL_LOCK@");
    }
    sprintf(elfArgString, "%d %s %d", winID, mstr, pos+PD_NUM_TOOLS);
	GEPaletteTask(elfArgString);

    /* make sure the palette is correctly set */
    palP = PaletteList[winID];
    currPal = *palP;
    currPal.toolType = pos + PD_NUM_TOOLS;
    currPal.toolLocked = (data->reason == XmCR_DEFAULT_ACTION) ? TRUE : FALSE;
    AxSetPalette(w->parent->parent, &currPal, FALSE);
    *palP = currPal;
}

/****************************************************************************/
static int
undisplayedWidget(w, palID, state)
AXWidget_main w;
int palID;
int state;
{
    int pal, colorpal, rulers;
    int no_display_flag;
    int winID;

    winID = AxWindowIdFromShellWidget(w);
    if ( PaletteList[winID]->editingPix && palID < PDEX_tool1
                                        && state != BUT_HIDE )
        return TRUE;   /* a button which shouldn't be drawn now */

    AxGetDecorations(winID, &pal, &colorpal, &rulers);
    no_display_flag = !pal;
    if ( pal && palID >= 0 )
    {
        switch ( palID )
        {
            case PDEX_lendmap:
            case PDEX_rendmap:
            case PDEX_stylemap:
            case PDEX_markmap:
                no_display_flag = TRUE;
            break;
        }
    }
    return ( no_display_flag );
}

/*************************************************************************
function: PartsListMan - Create/Change parts list
*/
static PartsListMan (w,lstrs,selected,state)
AXWidget_main  w;
char **lstrs;   /* NULL termed array of str ptrs */
int selected;   /* which one active or -1 */
int state;
{
    AXWidget box;
    Arg args[30];
    int n = 0;
    int size;
    int winID;      /* ATM window id */
    extern char *StrFromArray();
    static XtCallbackRec Callback[2];   /* standard template */

    if ( undisplayedWidget(w, -1, state) )
        return;

    /* get the palette */

    winID = AxWindowIdFromShellWidget(w);
    /* is there a selected item in the list box? */
    selected = MAX(selected,-1);

    /* count the number of entries in the list box */
    if ( lstrs ) for ( size = 0; lstrs[size]; size++ );

    /* set the args for list box creation/recreation */
    XtSetArg(args[n],XmNoldWidget,PaletteParts[winID]->id); n++;
    if ( state == BUT_HIDE || ge_as_pixed )
    {
        XtSetArg(args[n],XmNflags,WFLG_NODISPLAY); n++;
    }
    else
    {
        XtSetArg(args[n],XmNflags,0); n++;
    }
    CALLBACKER(Callback,GEPalettePartSelect,winID);
    XtSetArg(args[n],XmNx,7); n++;
    XtSetArg(args[n],XmNlistSizePolicy,XmCONSTANT); n++;
    XtSetArg(args[n],XmNscrollBarDisplayPolicy,XmAS_NEEDED); n++;
    XtSetArg(args[n],XmNautomaticSelection,TRUE); n++;
    XtSetArg(args[n],XmNsingleSelectionCallback,Callback); n++;
    XtSetArg(args[n],XmNdoubleClickInterval,1); n++;
    if ( lstrs )
    {
        XtSetArg(args[n],XmNitemCount,size); n++;
        XtSetArg(args[n],XmNitems,lstrs); n++;
    }
#if AXMS
	/* NT/95 will change the width of a listbox depending on font mode */
	if ( WidthOfDialogZero < 7 )
	{
    	XtSetArg(args[n],XmNwidth,82); n++;
	}
	else
	{
    	XtSetArg(args[n],XmNwidth,80); n++;
	}
#else
    XtSetArg(args[n],XmNwidth,69); n++;
#endif
    XtSetArg(args[n],XmNselectedItemCount,1); n++;
    XtSetArg(args[n],XmNselectedItems,selected); n++;
    XtSetArg(args[n],AXmNlistHScroll, TRUE); n++;
    XtSetArg(args[n],XmNmenuHelpWidget,3575); n++;

    /* create/recreate the widget */
    box = AXmCreateScrolledList((AXWidget)w->palette, NXLT("parts"), args, n);

    PaletteParts[winID]->id = box;
}

/****************************************************************************/
static int widgetType(index)
int index;
{
    switch ( index )
    {
        case PDEX_fillpat:
        case PDEX_fill_fg:
        case PDEX_fill_bg:
        case PDEX_linepat:
        case PDEX_line_fg:
        case PDEX_line_bg:
        case PDEX_shadowColor:
        case PDEX_pxfill_fg:
        case PDEX_pxfill_bg:
            return ( W_ROWCOL );

        case PDEX_lstyle:
        case PDEX_aleft:
        case PDEX_aright:
        case PDEX_mrkr:
            return ( W_POPUP );

        case PDEX_lendmap:
        case PDEX_rendmap:
        case PDEX_stylemap:
        case PDEX_markmap:
            return ( W_ICON );

        default:
            return ( W_BUTTON );
    }
}

/*************************************************************************
function: WidgetMan - Create/Change simple pushbutton from struct
*/
void
WidgetMan (pos,w,state)
int pos;                /* which button in PaletteWidgets */
AXWidget_main w;
int state;              /* new setting for the button */
{
    Arg args[20];
    int an = 0;
    static XtCallbackRec Callback[2];   /* standard template */
    AXWidget newWidg;
    AXWidget_palette parent;
    register pltdata *p;
    int type, val, fg, bg, jnk;
    extern int Thim2dMode;
    char *name;
    register int winID;     /* ATM win id */

    parent = w->palette;
    winID = AxWindowIdFromShellWidget(w);
    /* not a real button - don't draw */
    if ( pos >= PALT_NUM_BUTTONS )
        return;

    if ( undisplayedWidget(w, pos, state) )
        return;

    /* get the physical layout of the palette */
    p = &mwmPaletteBasics[pos];

    /* get the widget associated with the entry in the palette array */
    XtSetArg(args[an],XmNoldWidget,PaletteWidgets[winID][pos]); an++;

    /* set the widget flags */
    switch ( state )
    {
        case BUT_NA:   XtSetArg(args[an],XmNflags,WFLG_INACTIVE);  an++; break;
        case BUT_HIDE: XtSetArg(args[an],XmNflags,WFLG_NODISPLAY); an++; break;
        default:       XtSetArg(args[an],XmNflags,0);              an++; break;
    }

    /* set the position of the widget */
    XtSetArg(args[an],XmNx,p->x); an++;
    XtSetArg(args[an],XmNy,p->y); an++;

    /* manage the widget */
    type = widgetType(pos);
    if ( type == W_POPUP )
    {
        CALLBACKER(Callback,GEPaletteButtonPushed,winID);
        XtSetArg(args[an],XmNactivateCallback,Callback); an++;
        XtSetArg(args[an],XmNmenuHelpWidget,(3500+pos)); an++;
        XtSetArg(args[an],XmNvalue,(state==BUT_DOWN)); an++;
        XtSetArg(args[an],XmNgePaletteButtonType,getButtonType(pos)); an++;
        XtSetArg(args[an],XmNlabelPixmap,TRUE); an++;
        XtSetArg(args[an],XmNalternatePixmap,p->altname); an++;

        /* otherwise use normal button for pulldowns */
        if ( pos == PDEX_aright )
        {
            XtSetArg(args[an],XmCArmCallback,TRUE); an++;
            XtSetArg(args[an],XmNwhichButton,Button1); an++;
            val = PaletteList[winID]->arrow2;
            if ( val < 0 || val > 12 )
                val = 0;
            newWidg = (AXWidget)AXmCreatePushButtonGadget((AXWidget)parent,
												rightstrs[val], args,an);
        }
        else if ( pos == PDEX_aleft )
        {
            XtSetArg(args[an],XmCArmCallback,TRUE); an++;
            XtSetArg(args[an],XmNwhichButton,Button1); an++;
            val = PaletteList[winID]->arrow1;
            if ( val < 0 || val > 12 )
                val = 0;
            newWidg = (AXWidget)AXmCreatePushButtonGadget((AXWidget)parent,
											leftstrs[val], args,an);
        }
        else if ( pos == PDEX_lstyle )
        {
            XtSetArg(args[an],XmCArmCallback,TRUE); an++;
            XtSetArg(args[an],XmNwhichButton,Button1); an++;
            val = PaletteList[winID]->lineStyle;
            if ( val < 0 )
                val = 0;
            newWidg = (AXWidget)AXmCreatePushButtonGadget((AXWidget)parent,
												midstrs[val], args,an);
        }
        else if ( pos == PDEX_mrkr )
        {
            XtSetArg(args[an],XmCArmCallback,TRUE); an++;
            XtSetArg(args[an],XmNwhichButton,Button1); an++;
            val = PaletteList[winID]->join_symbols;
            if ( val < 0 )
                val = 0;
            newWidg = (AXWidget)AXmCreatePushButtonGadget((AXWidget)parent,
											mrkrstrs[val], args,an);
        }
        else
        {
            XtSetArg(args[an],XmCArmCallback,TRUE); an++;
            XtSetArg(args[an],XmNwhichButton,Button1); an++;
            newWidg = (AXWidget)AXmCreatePushButtonGadget((AXWidget)parent,
												p->name, args,an);
        }

#if AXMS
        /* make sure NT knows this is a W_POPUP and not an actual button */
        newWidg->style = type;
#endif
    }

    else if ( type == W_ROWCOL )
    {
        XtSetArg(args[an],XmNmenuHelpWidget,(3500+pos)); an++;
        XtSetArg(args[an],XmNgePaletteButtonType,getButtonType(pos)); an++;
        XtSetArg(args[an],XmNwhichButton,Button1); an++;
        XtSetArg(args[an],ZXmNstyle,0); an++;
        if ( pos == PDEX_fillpat )
        {
            /* set the appearance of the static button */
            val = PaletteList[winID]->fillPattern;
            if ( val < 0 )
                name = NXLT("gp_nfll");
            else
                name = fillmaps[fillvals[val]];
            XtSetArg(args[an],XmNlabelPixmap,TRUE); an++;
            XtSetArg(args[an],XmNalternatePixmap,name); an++;
            AxRowColColors((AXWidget_rowcol)PaletteWidgets[winID][PDEX_fill_fg],
                           PaletteList[winID]->ffg, &fg, &jnk);
            AxRowColColors((AXWidget_rowcol)PaletteWidgets[winID][PDEX_fill_fg],
                           PaletteList[winID]->fbg, &jnk, &bg);
            XtSetArg(args[an],AXmNpfg_color,fg); an++;
            XtSetArg(args[an],AXmNpbg_color,bg); an++;

            /* set up the child buttons */
            if ( val < 0 )
                XtSetArg(args[an],XmNvalue,-1);
            else
                XtSetArg(args[an],XmNvalue,fillvals[val]);
            an++;
            if ( !PaletteWidgets[winID][pos] )
            {
                init_fillstrs();
                XtSetArg(args[an],XmNitemCount,30); an++;
                XtSetArg(args[an],AXmNbtnPixmapList,fillmaps); an++;
                XtSetArg(args[an],XmNitems,fillstrs); an++;
                XtSetArg(args[an],XmNorientation,10); an++;
            }

            /* set the callback */
            CALLBACKER(Callback,FinishFFillPatChoice,winID);
            XtSetArg(args[an],XmNactivateCallback,Callback); an++;
        }
        else if ( pos == PDEX_linepat )
        {
            /* set the appearance of the static button */
            val = PaletteList[winID]->linefillPattern;
            if ( val < 0 ) /* mixed fill */
                name = NXLT("gp_nlfll");
            else if ( val == 0 ) /* no fill */
                name = fillmaps[fillvals[5]];
            else if ( val == 29 ) /* user defined fill */
                name = NXLT("gp_l_udf");
            else
                name = fillmaps[fillvals[val]];
            XtSetArg(args[an],XmNlabelPixmap,TRUE); an++;
            XtSetArg(args[an],XmNalternatePixmap,name); an++;
            AxRowColColors((AXWidget_rowcol)PaletteWidgets[winID][PDEX_fill_fg],
                           PaletteList[winID]->lfg, &fg, &jnk);
            AxRowColColors((AXWidget_rowcol)PaletteWidgets[winID][PDEX_fill_fg],
                           PaletteList[winID]->lbg, &jnk, &bg);
            XtSetArg(args[an],AXmNpfg_color,fg); an++;
            XtSetArg(args[an],AXmNpbg_color,bg); an++;

            /* set up the child buttons */
            if ( val < 0 )
                XtSetArg(args[an],XmNvalue,-1);
            else
                XtSetArg(args[an],XmNvalue,fillvals[val]);
            an++;
            if ( !PaletteWidgets[winID][pos] )
            {
                init_linestrs();
                XtSetArg(args[an],XmNitemCount,30); an++;
                XtSetArg(args[an],AXmNbtnPixmapList,fillmaps); an++;
                XtSetArg(args[an],XmNitems,linestrs);an++;
                XtSetArg(args[an],XmNorientation,10); an++;
            }

            /* set the callback */
            CALLBACKER(Callback,FinishLFillPatChoice,winID);
            XtSetArg(args[an],XmNactivateCallback,Callback); an++;
        }
        else
        {
            if ( pos == PDEX_shadowColor )
            {
                /* set the appearance of the static button */
                val = PaletteList[winID]->shadow_color;
                XtSetArg(args[an],XmNlabelPixmap,TRUE); an++;
                if ( val < 0 )
                {
                    XtSetArg(args[an],XmNalternatePixmap,p->altname); an++;
                    XtSetArg(args[an],AXmNpfg_color,0); an++;
                    XtSetArg(args[an],AXmNpbg_color,0); an++;
                    XtSetArg(args[an],XmNvalue,-1); an++;
                }
                else
                {
                    AxRowColColors(
						(AXWidget_rowcol)PaletteWidgets[winID][PDEX_fill_fg],
                        val, &fg, &bg);
                    if ( fg )
                        XtSetArg(args[an],XmNalternatePixmap,p->altname);
                    else
                        XtSetArg(args[an],XmNalternatePixmap,NXLT("gp_ncolr"));
                    an++;
                    XtSetArg(args[an],AXmNpfg_color,fg); an++;
                    XtSetArg(args[an],AXmNpbg_color,fg); an++;
                    XtSetArg(args[an],XmNvalue,val); an++;
                }

                /* set the child buttons */
                if ( !PaletteWidgets[winID][pos] )
                {
                    XtSetArg(args[an],XmNitemCount,0);
                    an++;
                }

                /* set the callback */
                CALLBACKER(Callback,FinishShadowColorChoice,winID);
                XtSetArg(args[an],XmNactivateCallback,Callback); an++;
            }
            else if ( pos == PDEX_fill_fg )
            {
                /* set the appearance of the static button */
                val = PaletteList[winID]->ffg;
                XtSetArg(args[an],XmNlabelPixmap,TRUE); an++;
                if ( val < 0 )
                {
                    XtSetArg(args[an],XmNalternatePixmap,p->altname); an++;
                    XtSetArg(args[an],AXmNpfg_color,0); an++;
                    XtSetArg(args[an],AXmNpbg_color,0); an++;
                    XtSetArg(args[an],XmNvalue,-1); an++;
                }
                else
                {
                    AxRowColColors(
						(AXWidget_rowcol)PaletteWidgets[winID][PDEX_fill_fg],
                        val, &fg, &bg);
                    if ( fg )
                        XtSetArg(args[an],XmNalternatePixmap,p->altname);
                    else
                        XtSetArg(args[an],XmNalternatePixmap,NXLT("gp_ncolr"));
                    an++;
                    XtSetArg(args[an],AXmNpfg_color,fg); an++;
                    XtSetArg(args[an],AXmNpbg_color,fg); an++;
                    XtSetArg(args[an],XmNvalue,val); an++;
                }

                /* set the child buttons */
                if ( !PaletteWidgets[winID][pos] )
                {
                    XtSetArg(args[an],XmNitemCount,0);
                    an++;
                }

                /* set the callback */
                CALLBACKER(Callback,FinishFFillFgChoice,winID);
                XtSetArg(args[an],XmNactivateCallback,Callback); an++;
            }
            else if ( pos == PDEX_fill_bg )
            {
                /* set the appearance of the static button */
                val = PaletteList[winID]->fbg;
                XtSetArg(args[an],XmNlabelPixmap,TRUE); an++;
                if ( val < 0 )
                {
                    XtSetArg(args[an],XmNalternatePixmap,p->altname); an++;
                    XtSetArg(args[an],AXmNpfg_color,0); an++;
                    XtSetArg(args[an],AXmNpbg_color,0); an++;
                    XtSetArg(args[an],XmNvalue,-1); an++;
                }
                else
                {
                    AxRowColColors(
						(AXWidget_rowcol)PaletteWidgets[winID][PDEX_fill_fg],
                        val, &fg, &bg);
                    if ( fg )
                        XtSetArg(args[an],XmNalternatePixmap,p->altname);
                    else
                        XtSetArg(args[an],XmNalternatePixmap,NXLT("gp_ncolr"));
                    an++;
                    XtSetArg(args[an],AXmNpfg_color,fg); an++;
                    XtSetArg(args[an],AXmNpbg_color,fg); an++;
                    XtSetArg(args[an],XmNvalue,val); an++;
                }

                /* set the child buttons */
                if ( !PaletteWidgets[winID][pos] )
                {
                    XtSetArg(args[an],XmNitemCount,0);
                    an++;
                }

                /* set the callback */
                CALLBACKER(Callback,FinishFFillBgChoice,winID);
                XtSetArg(args[an],XmNactivateCallback,Callback); an++;
            }
            else if ( pos == PDEX_line_fg )
            {
                /* set the appearance of the static button */
                val = PaletteList[winID]->lfg;
                XtSetArg(args[an],XmNlabelPixmap,TRUE); an++;
                if ( val < 0 )
                {
                    XtSetArg(args[an],XmNalternatePixmap,p->altname); an++;
                    XtSetArg(args[an],AXmNpfg_color,0); an++;
                    XtSetArg(args[an],AXmNpbg_color,0); an++;
                    XtSetArg(args[an],XmNvalue,-1); an++;
                }
                else
                {
                    AxRowColColors(
						(AXWidget_rowcol)PaletteWidgets[winID][PDEX_fill_fg],
                        val, &fg, &bg);
                    if ( fg )
                        XtSetArg(args[an],XmNalternatePixmap,p->altname);
                    else
                        XtSetArg(args[an],XmNalternatePixmap,NXLT("gp_ncolr"));
                    an++;
                    XtSetArg(args[an],AXmNpfg_color,fg); an++;
                    XtSetArg(args[an],AXmNpbg_color,fg); an++;
                    XtSetArg(args[an],XmNvalue,val); an++;
                }

                /* set the child buttons */
                if ( !PaletteWidgets[winID][pos] )
                {
                    XtSetArg(args[an],XmNitemCount,0);
                    an++;
                }

                /* set the callback */
                CALLBACKER(Callback,FinishLFillFgChoice,winID);
                XtSetArg(args[an],XmNactivateCallback,Callback); an++;
            }
            else if ( pos == PDEX_line_bg )
            {
                /* set the appearance of the static button */
                val = PaletteList[winID]->lbg;
                XtSetArg(args[an],XmNlabelPixmap,TRUE); an++;
                if ( val < 0 )
                {
                    XtSetArg(args[an],XmNalternatePixmap,p->altname); an++;
                    XtSetArg(args[an],AXmNpfg_color,0); an++;
                    XtSetArg(args[an],AXmNpbg_color,0); an++;
                    XtSetArg(args[an],XmNvalue,-1); an++;
                }
                else
                {
                    AxRowColColors(
						(AXWidget_rowcol)PaletteWidgets[winID][PDEX_fill_fg],
                        val, &fg, &bg);
                    if ( fg )
                        XtSetArg(args[an],XmNalternatePixmap,p->altname);
                    else
                        XtSetArg(args[an],XmNalternatePixmap,NXLT("gp_ncolr"));
                    an++;
                    XtSetArg(args[an],AXmNpfg_color,fg); an++;
                    XtSetArg(args[an],AXmNpbg_color,fg); an++;
                    XtSetArg(args[an],XmNvalue,val); an++;
                }

                /* set the child buttons */
                if ( !PaletteWidgets[winID][pos] )
                {
                    XtSetArg(args[an],XmNitemCount,0);
                    an++;
                }

                /* set the callback */
                CALLBACKER(Callback,FinishLFillBgChoice,winID);
                XtSetArg(args[an],XmNactivateCallback,Callback); an++;
            }
            else if ( pos == PDEX_pxfill_fg )
            {
                /* set the appearance of the static button */
                val = PaletteList[winID]->ffg;
                XtSetArg(args[an],XmNlabelPixmap,TRUE); an++;
                if ( val < 0 )
                {
                    XtSetArg(args[an],XmNalternatePixmap,p->altname); an++;
                    XtSetArg(args[an],AXmNpfg_color,0); an++;
                    XtSetArg(args[an],AXmNpbg_color,0); an++;
                    XtSetArg(args[an],XmNvalue,-1); an++;
                }
                else
                {
                    AxRowColColors(
						(AXWidget_rowcol)PaletteWidgets[winID][PDEX_pxfill_fg],
                        val, &fg, &bg);
                    if ( fg )
                        XtSetArg(args[an],XmNalternatePixmap,p->altname);
                    else
                        XtSetArg(args[an],XmNalternatePixmap,NXLT("gp_ncolr"));
                    an++;
                    XtSetArg(args[an],AXmNpfg_color,fg); an++;
                    XtSetArg(args[an],AXmNpbg_color,fg); an++;
                    XtSetArg(args[an],XmNvalue,val); an++;
                }

                /* set the child buttons */
                if ( !PaletteWidgets[winID][pos] )
                {
                    XtSetArg(args[an],XmNitemCount,0);
                    an++;
                }

                /* set the callback */
                CALLBACKER(Callback,FinishFFillFgChoice,winID);
                XtSetArg(args[an],XmNactivateCallback,Callback); an++;
            }
            else if ( pos == PDEX_pxfill_bg )
            {
                /* set the appearance of the static button */
                val = PaletteList[winID]->fbg;
                XtSetArg(args[an],XmNlabelPixmap,TRUE); an++;
                if ( val < 0 )
                {
                    XtSetArg(args[an],XmNalternatePixmap,p->altname); an++;
                    XtSetArg(args[an],AXmNpfg_color,0); an++;
                    XtSetArg(args[an],AXmNpbg_color,0); an++;
                    XtSetArg(args[an],XmNvalue,-1); an++;
                }
                else
                {
                    AxRowColColors(
						(AXWidget_rowcol)PaletteWidgets[winID][PDEX_pxfill_bg],
                        val, &fg, &bg);
                    if ( fg )
                        XtSetArg(args[an],XmNalternatePixmap,p->altname);
                    else
                        XtSetArg(args[an],XmNalternatePixmap,NXLT("gp_ncolr"));
                    an++;
                    XtSetArg(args[an],AXmNpfg_color,fg); an++;
                    XtSetArg(args[an],AXmNpbg_color,fg); an++;
                    XtSetArg(args[an],XmNvalue,val); an++;
                }

                /* set the child buttons */
                if ( !PaletteWidgets[winID][pos] )
                {
                    XtSetArg(args[an],XmNitemCount,0);
                    an++;
                }

                /* set the callback */
                CALLBACKER(Callback,FinishFFillBgChoice,winID);
                XtSetArg(args[an],XmNactivateCallback,Callback); an++;
            }
        }

        /* reuse the old widget if possible */
        /* AXmCreateRowColumn() reuses old widget */
        newWidg=(AXWidget)AXmCreateRowColumn((AXWidget)parent,p->name,args,an);
        PaletteWidgets[winID][pos] = newWidg;
    }

    else if ( type == W_ICON )
    {
        if ( pos == PDEX_rendmap )
        {
            XtSetArg(args[an],XmNmaxWidth,29); an++;
            XtSetArg(args[an],XmNmaxHeight,12); an++;
            val = PaletteList[winID]->arrow2;
            if ( val < 0 || val > 12 )
                val = 0;
            XtSetArg(args[an],XmNvalue,rightstrs[val]); an++;
            newWidg = ZXmCreateIconImage((AXWidget)parent,p->name,args,an);
        }
        else if ( pos == PDEX_lendmap )
        {
            XtSetArg(args[an],XmNmaxWidth,29); an++;
            XtSetArg(args[an],XmNmaxHeight,12); an++;
            val = PaletteList[winID]->arrow1;
            if ( val < 0 || val > 12 )
                val = 0;
            XtSetArg(args[an],XmNvalue,leftstrs[val]); an++;
            newWidg = ZXmCreateIconImage((AXWidget)parent,p->name,args,an);
        }
        else if ( pos == PDEX_stylemap )
        {
            XtSetArg(args[an],XmNmaxWidth,60); an++;
            XtSetArg(args[an],XmNmaxHeight,12); an++;
            val = PaletteList[winID]->lineStyle;
            if ( val < 0 )
                val = 0;
            XtSetArg(args[an],XmNvalue,midstrs[val]); an++;
            newWidg = ZXmCreateIconImage((AXWidget)parent,p->name,args,an);
        }
        else if ( pos == PDEX_markmap )
        {
            XtSetArg(args[an],XmNmaxWidth,12); an++;
            XtSetArg(args[an],XmNmaxHeight,12); an++;
            val = PaletteList[winID]->join_symbols;
            if ( val < 0 )
                val = 0;
            XtSetArg(args[an],XmNvalue,mrkrstrs[val]); an++;
            newWidg = ZXmCreateIconImage((AXWidget)parent, p->name, args,an);
        }
    }

    else /* W_BUTTON */
    {
        CALLBACKER(Callback,GEPaletteButtonPushed,winID);
        XtSetArg(args[an],XmNactivateCallback,Callback); an++;
        XtSetArg(args[an],XmNmenuHelpWidget,(3500+pos)); an++;
        XtSetArg(args[an],XmNvalue,(state==BUT_DOWN)); an++;
        XtSetArg(args[an],XmNgePaletteButtonType,getButtonType(pos)); an++;
        XtSetArg(args[an],XmNalternatePixmap,p->altname); an++;
        XtSetArg(args[an],XmNwhichButton,Button1); an++;
        XtSetArg(args[an],XmNmultipleSelectionCallback,TRUE); an++;
        XtSetArg(args[an],XmNlabelPixmap,TRUE); an++;
        newWidg = (AXWidget)AXmCreatePushButtonGadget((AXWidget)parent,
													  p->name, args,an);
    }

    PaletteWidgets[winID][pos] = newWidg;
    AxXtManageChild(PaletteWidgets[winID][pos]);
}

/****************************************************************************/
static void 
DrawWidgets(mode,w)
int mode;
AXWidget_main w;
{
    int palette_tool, state;
    int winID;

    winID = AxWindowIdFromShellWidget(w);
    /* draw GE palette */
    if ( mode == PALT_GE && !ge_as_pixed )
    {
        for(palette_tool=0; palette_tool<PALT_NUM_BUTTONS; palette_tool++)
        {
            state = (palette_tool < PDEX_tool1 ? BUT_UP : BUT_HIDE);
            WidgetMan(palette_tool,w,state);
        }
        PartsListMan(w,(char**)NULL,-1,BUT_UP);
    }

    /* or draw PIXED palette */
    else
    {
        for(palette_tool=0; palette_tool<PALT_NUM_BUTTONS; palette_tool++)
        {
            state = (palette_tool >= PDEX_tool1 ? BUT_UP : BUT_HIDE);
            WidgetMan(palette_tool,w,state);
        }
        PartsListMan(w,(char**)NULL,-1,BUT_HIDE);
    }
}

/*************************************************************************
function: InitPalette - Create the palette widgets
*/
void InitPalette (w)
AXWidget_main  w;
{
    int winID;      /* ATM window id */
    static char *strs[] = {NULL};

    winID = AxWindowIdFromShellWidget(w);
    /* initial allocation of palette structure arrays */
    if ( !PaletteList )
    {
        /* 100 is safe, but should use TM_MAXWIN */
        PaletteList    = (GEPalette_t**)calloc(100,sizeof(GEPalette_t*));
        PaletteWidgets = (AXWidget**)calloc(100,sizeof(AXWidget*));
        PaletteParts   = (PartsType**)calloc(100,sizeof(PartsType*));
    }

    /* clear the palette struct of a previous window */
    if ( PaletteList[winID] )
    {
        /* this is a free on reenter, not on exit (sloppy) */
        free(PaletteList[winID]);
        free(PaletteWidgets[winID]);
        free(PaletteParts[winID]);
    }

    /* allocate and set the new palette structure for this window */
    PaletteWidgets[winID] = (AXWidget*)calloc(PALT_NUM_BUTTONS,sizeof(AXWidget));
    PaletteParts[winID]   = (PartsType*)calloc(1,sizeof(PartsType));
    PaletteList[winID]    = (GEPalette_t*)calloc(1,sizeof(GEPalette_t));
    memset((char *)PaletteList[winID],-1,sizeof(GEPalette_t));

    PaletteList[winID]->shadowBoxUp = FALSE;
    PaletteList[winID]->editingPix = FALSE;
    PaletteList[winID]->repaintFlag = TRUE;
    PaletteList[winID]->ribbon = 0;

    /* display the window's palette */
    DrawWidgets(PALT_GE,w);
}

/****************************************************************************/
void ReinitPalette(w)
AXWidget_main  w;
{
    int winID;

    winID = AxWindowIdFromShellWidget(w);
     PaletteList[winID]->repaintFlag = TRUE;
    if ( PaletteList[winID]->editingPix )
        DrawWidgets(PALT_PX, w);
    else
        DrawWidgets(PALT_GE, w);
}

/****************************************************************************/
void TmPixedPalette(wd)
long wd;
{
	AXWidget_main w;
    int winID;

	w = (AXWidget_main)wd;
    winID = AxWindowIdFromShellWidget(w);
    if ( !TmAmSE )
    {
        PaletteList[winID]->editingPix = TRUE;
        DrawWidgets(PALT_PX, w);
    }
}

/****************************************************************************/
clearPalette(wd)
long wd;
{
    int winID;
    GEPalette_t *oldP;
	AXWidget_main  w;

    /* ignore palettes when printing */
    if ( TmAmSE == TRUE )
        return(AOK);

	w = (AXWidget_main)wd;
    winID = AxWindowIdFromShellWidget(w);

    /* first time for this window - initialize the palette */
    if ( !PaletteList || !PaletteList[winID] )
    {
        InitPalette(w);
        PaintGEPalette(w->palette, (XExposeEvent *) NULL, TRUE, TRUE);
    }

    oldP = PaletteList[winID];
    oldP->repaintColors = TRUE;
}

/****************************************************************************/
void TmGEPalette(wd)
long wd;
{
    int winID;
	AXWidget_main w;

	w = (AXWidget_main)wd;
    winID = AxWindowIdFromShellWidget(w);
    if ( !TmAmSE )
    {
        PaletteList[winID]->editingPix = FALSE;
        DrawWidgets(PALT_GE, w);
    }
}

/****************************************************************************/
int TmGetPalette(wd, pal)
long wd;
GEPalette_t *pal;
{
    int winID;
	AXWidget_main w;

    if (TmAmSE == TRUE) {
        /* ignore palettes when printing */
        return(AOK);
    }
	w = (AXWidget_main)wd;
    if ( !w || w->type != W_MAIN )
        return(!AOK);
    winID = AxWindowIdFromShellWidget(w);
    if ( !PaletteList || !PaletteList[winID] )
        memset((char *)pal,0,sizeof(GEPalette_t));
    else
        *pal = *PaletteList[winID];
    return(AOK);
}

/****************************************************************************/
TmRedrawListbox(w)
AXWidget_main w;
{
    if (TmAmSE == TRUE) {
        /* ignore palettes when printing */
        return(AOK);
    }
    PaintGEListBox((AXWidget)w->palette);
}

/*************************************************************************
function: GESetPointerFromPalette - Update mouse pointer in work area
*/
int
GESetPointerFromPalette(w)
AXWidget_main w;
{
    GEPalette_t *pP;
    int winID;                /* ATM window of GE */

    winID = AxWindowIdFromShellWidget(w);
    if ( winID <= 0 ) /* sanity */
        return(FALSE);

    pP = PaletteList[winID];
    if ( !pP )
        return(FALSE);

    AxSetCursorShape(w, 0, pP->toolType);
    return(TRUE);
}

/****************************************************************************/
#define PAL_SLIDE_EDITOR   0
#define PAL_HANDOUT_EDITOR 1
#define PAL_OUTLINE_EDITOR 2
#define PAL_NOTES_EDITOR   3
#define PAL_SLIDE_SHOW     4

static void grPresSlideButton();
static void grPresHandoutButton();
static void grPresOutlineButton();
static void grPresNotesButton();
static void grPresSlideShowButton();
void (*pres_func[])() = {
    grPresSlideButton,
    grPresHandoutButton,
    grPresOutlineButton,
    grPresNotesButton,
    grPresSlideShowButton,
};
static pltdata pixbuttons[PRES_NUM_BUTTONS] = {
    { 0,  0, NXLT("gp_pres1"), NXLT("gp_pres1")},
    { 0,  0, NXLT("gp_pres2"), NXLT("gp_pres2")},
    { 0,  0, NXLT("gp_pres3"), NXLT("gp_pres3")},
    { 0,  0, NXLT("gp_pres4"), NXLT("gp_pres4")},
    { 0,  0, NXLT("gp_pres5"), NXLT("gp_pres5")},
};

PaintGEButton(w, exP, regen, layered)
register AXWidget_buttonbar w;
XExposeEvent *exP;
int regen;
int layered;
{
    register AXWidget childP;
    register void (*func)();

    if ( !w || !w->display )
        return;
    
    for ( childP=w->widgets; childP; childP=childP->nextChild )
    {
        func = childP->painter;
        if ( func && childP->display )
            (*func)(childP, exP, TRUE, layered);
    }
}

AxSetButtonBarButton(w, ix)
AXWidget_main w;
int ix;
{
    Arg args[1];
    int i, n;

    /* need a shell widget with a presentation button bar and buttons */
    if ( !w || w->type != W_MAIN || !w->butonBar || !w->butonBar->buttons )
        return;

    /* depress the active button, raise the others */
    n = PRES_NUM_BUTTONS - 1;
    if ( ix < n ) for ( i=0; i<n; i++ )
    {
        XtSetArg(args[0], XmNvalue, (i == ix) ? TRUE : FALSE);
        AxXtSetValues((AXWidget)w->butonBar->buttons[i], args, 1);
    }
    XtSetArg(args[0], XmNvalue, FALSE);
    AxXtSetValues((AXWidget)w->butonBar->buttons[n], args, 1);
}

InitButtonPalette(w)
AXWidget_main w;
{
    static XtCallbackRec Callback[2];
    int i, j, yoff, an, winID, help_id;
    register pltdata *p;
    AXWidget_buttonbar wp;
    AXWidget_pushbutton newWidg;
    Arg args[10];

    winID = AxWindowIdFromShellWidget(w);
    wp = w->butonBar;

    yoff = w->butonBar->height=0;
    if ( wp->value == 4 )
    {
        /* no outliner */
        wp->num_buttons = 5;
        help_id = 3560;
        j = 0;
        for ( i=0; i<PRES_NUM_BUTTONS; i++ )
        {
            /* get the button info */
            CALLBACKER(Callback, pres_func[i], winID);
            p = &pixbuttons[i];
            an = 0;

            /* create the child button */
            XtSetArg(args[an], XmNoldWidget, 0);                an++;
            if ( i == 2 ) /* PAL_OUTLINE_EDITOR */
            {
                /* NODISPLAY */
                XtSetArg(args[an], XmNflags, 0x01);             an++;
                XtSetArg(args[an], XmNlabelPixmap, FALSE);      an++;
            }
            else
            {
                XtSetArg(args[an], XmNflags,     0);            an++;
                XtSetArg(args[an], XmNlabelPixmap, TRUE);       an++;
            }
            XtSetArg(args[an], XmNx, 0);                        an++;
            XtSetArg(args[an], XmNy, yoff);                     an++;
            XtSetArg(args[an], XmNactivateCallback, Callback);  an++;
            XtSetArg(args[an], XmNmenuHelpWidget, help_id);     an++;
            XtSetArg(args[an], XmNvalue, (i==0));               an++;
            XtSetArg(args[an], XmNalternatePixmap, p->altname); an++;
            XtSetArg(args[an], XmNwhichButton, Button1);        an++;
            newWidg = AXmCreatePushButtonGadget((AXWidget)w->butonBar,
												p->name, args, an);
            AxXtManageChild((AXWidget)newWidg);

            /* update the button bar size and the next button offset */
            if ( i != 2 ) /* PAL_OUTLINE_EDITOR */
            {
                w->butonBar->height += newWidg->height;
                yoff += newWidg->height;
            }
            w->butonBar->buttons[i] = newWidg;
            help_id++;
        }
    }
    else
    {
        help_id = 3560;
        for ( i=0; i<PRES_NUM_BUTTONS; i++ )
        {
            /* get the button info */
            CALLBACKER(Callback, pres_func[i], winID);
            p = &pixbuttons[i];
            an = 0;

            /* create the child button */
            XtSetArg(args[an], XmNoldWidget, 0);                an++;
            XtSetArg(args[an], XmNflags,     0);                an++;
            XtSetArg(args[an], XmNx, 0);                        an++;
            XtSetArg(args[an], XmNy, yoff);                     an++;
            XtSetArg(args[an], XmNactivateCallback, Callback);  an++;
            XtSetArg(args[an], XmNmenuHelpWidget, help_id);     an++;
            XtSetArg(args[an], XmNvalue, (i==0));               an++;
            XtSetArg(args[an], XmNalternatePixmap, p->altname); an++;
            XtSetArg(args[an], XmNwhichButton, Button1);        an++;
            XtSetArg(args[an], XmNlabelPixmap, TRUE);           an++;
            newWidg = AXmCreatePushButtonGadget((AXWidget)w->butonBar,
												p->name, args, an);
            AxXtManageChild((AXWidget)newWidg);

            /* update the button bar size and the next button offset */
            w->butonBar->height += newWidg->height;
            yoff += newWidg->height;
            w->butonBar->buttons[i] = newWidg;
            help_id++;
        }
    }
}

static void
grPresButton(w, winID, data, ix)
AXWidget w;
int winID;
XmRowColumnCallbackStruct *data;
int ix;
{
    char arg_str[STRPATH];
    if ( !w || !w->parent || !w->parent->parent )
        return;

    /* raise the other buttons (unless it's the slideshow button) */
    AxSetButtonBarButton(w->parent->parent, ix);

    /* do any associated action */
    switch(ix)
    {
        case PAL_SLIDE_SHOW: /* slide show button */
            if ( ThimIsShifted() )
                sprintf(arg_str, "%d gr_slideshow_main@", winID);
            else
                sprintf(arg_str, "%d gr_slideshow_dlg@", winID);
			GEPaletteTask(arg_str);
        break;

        case PAL_NOTES_EDITOR:
            if ( ThimIsShifted() )
                sprintf(arg_str, "%d gr_edit_speaker_notes_master@", winID);
            else
                sprintf(arg_str, "%d gr_edit_speaker_notes@", winID);
			GEPaletteTask(arg_str);
        break;

        case PAL_OUTLINE_EDITOR:
            if ( ThimIsShifted() )
                sprintf(arg_str, "%d gr_edit_outline_master@", winID);
            else
                sprintf(arg_str, "%d gr_edit_outline@", winID);
			GEPaletteTask(arg_str);
        break;

        case PAL_HANDOUT_EDITOR:
            if ( ThimIsShifted() )
                sprintf(arg_str, "%d gr_edit_handout_master@", winID);
            else
                sprintf(arg_str, "%d gr_sort_slides@", winID);
			GEPaletteTask(arg_str);
        break;

        default:
        case PAL_SLIDE_EDITOR:
            if ( ThimIsShifted() )
                sprintf(arg_str, "%d gr_edit_slide_master@", winID);
            else
                sprintf(arg_str, "%d gr_edit_slide@", winID);
			GEPaletteTask(arg_str);
        break;
    }
}

static void
grPresSlideButton(w, winID, data)
AXWidget w;
int winID;
XmRowColumnCallbackStruct *data;
{
    grPresButton(w, winID, data, PAL_SLIDE_EDITOR);
}

static void
grPresHandoutButton(w, winID, data)
AXWidget w;
int winID;
XmRowColumnCallbackStruct *data;
{
    grPresButton(w, winID, data, PAL_HANDOUT_EDITOR);
}

static void
grPresOutlineButton(w, winID, data)
AXWidget w;
int winID;
XmRowColumnCallbackStruct *data;
{
    grPresButton(w, winID, data, PAL_OUTLINE_EDITOR);
}

static void
grPresNotesButton(w, winID, data)
AXWidget w;
int winID;
XmRowColumnCallbackStruct *data;
{
    grPresButton(w, winID, data, PAL_NOTES_EDITOR);
}

static void 
grPresSlideShowButton(w, winID, data)
AXWidget w;
int winID;
XmRowColumnCallbackStruct *data;
{
    grPresButton(w, winID, data, PAL_SLIDE_SHOW);
}

AxInactiveButtonBar(wd)
long wd;
{
    register AXWidget childP;
	AXWidget_main w;

	w = (AXWidget_main)wd;
    if ( !w || !w->butonBar )
        return;
    AxXtSetSensitive((AXWidget)w->butonBar, FALSE);
    for ( childP=w->butonBar->widgets; childP; childP=childP->nextChild )
        AxXtSetSensitive(childP, FALSE);
    w->butonBar->update_widget = TRUE;
    WidgetNeedsUpdated = TRUE;
}

AxactiveButtonBar(wd)
long wd;
{
    register AXWidget childP;
	AXWidget_main w;

	w = (AXWidget_main)wd;
    if ( !w || !w->butonBar )
        return;
    AxXtSetSensitive((AXWidget)w->butonBar, TRUE);
    for ( childP=w->butonBar->widgets; childP; childP=childP->nextChild )
        AxXtSetSensitive(childP, TRUE);
    w->butonBar->update_widget = TRUE;
    WidgetNeedsUpdated = TRUE;
}
