/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: parint.h
   Applix ASC version: 24
   Date: Mon Nov 16 13:53:37 1998
*/
#ifndef PARINT_INCLUDED
#define PARINT_INCLUDED 1

/*
	Parameters for Client Services Layer of Main ATM
*/

/*
 * Note: includes oldwin.h.
 * Lower level modules use oldwin.h in place of paratmint.h.
 * Also note: paratmint.h depends on stdio.h, atmwin.h, atmreg.h,
 * atmattr.h and atmtextin.h.
 */

#include <oldwin.h>
#include <qmenu.h>
#include <amenudef.h>
#include <atmtxtin.h>

#ifndef MAX_SELECTS
#define MAX_SELECTS 50  /* selction/home line options */
#endif

#ifndef _LAYTEXT_H_
#include <laytext.h> /* for fieldInfo typedef */
#endif

#if AXMS
#include <windows.h>
#endif


/***********
 * Defines *
 ***********/
 
/*
 * attribute buffer size in 'TmTextDescriptor"s:
 * = max number of 'TmTextDescriptor's in an attribute string
 * (based on worst case of an attribute change with every character)
 */
#define MXATTRBUF	( 1 + 2 * \
			((TM_TAASIZE - sizeof(TmTextDescriptor)) \
			/ (2 * sizeof(TmTextDescriptor) + sizeof(char))) )

#define MAX_CHAN	20

#define HEADER_LENGTH	8

/* text attributes */

#define DFONT_DES	((TMA_MONO1 << TMFPFONT) | (TMA_12PT << TMFPSIZ))

#define CHECK_BIT(w, b)			((w & (1L << (b))) != 0)
#define EXT_ATTR(a, m, c)		(int) (((a) >> (c)) & (TmTextDescriptor) (m))

/*
 * desktop definitions used in setting up runtime values
 */

#define NUL				'\0'		/* use NULCHAR instead */
#define ZERO			0
#define RESET			0
#define SET				1
#define SPACE			' '
#define BLANK(ch)		((ch) == SPACE || (ch) == TM_NBSPACE \
								|| (ch) == '\t' || (ch) == '\n')
#define NONBLANK(ch)	(!BLANK(ch))

#define AND			&&
#define OR			||

#define END 			0
#define MILLS_TO_PIXEL		1
#define PIXEL_TO_MILLS		2

#define MAX_RULER_UNIT_NAME    32

		/* mouse defines */


/********************
 * Type Definitions *
 ********************/

		/* window database */

struct  wind_struct 
{
	/* general window data */

	/* flags */

	/* border present flags (TMODE, RGMODE) */

	unsigned leftBord : 1;	/* left border present */
	unsigned rightBord : 1;	/* right border present */
	unsigned botBord : 1;	/* bottom border present */
	unsigned topBord : 1;	/* top border present */

	unsigned clipSupplied : 1;	/* appl has supplied clip rect */
	unsigned regenWait : 1;		/* atm waiting for requested regen
									to begin */
	unsigned gRegenWait : 1;	/* waiting for requested gr. regen to begin */
	unsigned g_mod_flag : 1;	/* Flag to indicate if the window has */
								/* switched to graphics window	*/
	unsigned frWait : 1;	/* deferring char mode output until field read */
	unsigned lineUsed: 1;	/* clear menuless mode window before reuse */
	unsigned recording: 1;	/* recording requests from application */
	unsigned noLKparse: 1;	/* suppress LK parsing when window active */
	unsigned cloverEnable: 1;	/* enable clear on overstriking 1st char */
	unsigned resizeSent: 1;		/* resize event not yet acknowledged */
	unsigned promptByElf: 1;	/* prompt status due to ELF prompt mode */
	unsigned motionF: 1;	/* user took action to move cursor */
	unsigned regenBlank: 1;	/* (in regenWait): key was not sent */

	unsigned hasBeenMapped: 1;	/* TRUE after first exposure event */
	unsigned imageTextMode: 1;

	short smallestFontSize;	/* nominally 6 */
	short largestFontSize;	/* nominally 36 */
	short displaySizeShift;	/* display larger chars than print */

	short window_id;

	/* always make sure NEED_X_INFO section is the same
	 * size as !NEED_X_INFO section.. some modules define
	 * NEED_X_INFO true, others false.. */
#if NEED_X_INFO
	Window xwin;		/* work area */
	Window topwindow;	/* main window */
	Window editwin;		/* editline window for SS */
	Window inCellWin;	/* incell edit window for SS **/
	Window rowwin;		/* row window for SS */
	Window colwin;		/* col window for SS */
	AXWidget_main shell;		/* widget containing main window */
	AXWidget_main topwidget;	/* main window widget */
	AXWidget_work workWidget;	/* widget of xwin */
	AXWidget_scroll vbar;		/* vert scroll bar */
	AXWidget_scroll hbar;		/* horiz scroll bar */
	AXWidget_menu menu_bar;	/* menu bar */
	long haveTempTitle;	/* TRUE if title is replace by a temporary msg */
	AXWidget_palette palette;		/* left widget */
	AXWidget_ruler ruler;		/* widget below menu bar */
	AXWidget_plates plates;		/* widget below menu bar in SS */
	AXWidget zoomer;		/* sister window */
	AXWidget_debug debugger;	/* for macro editor */
	Pixmap xMap;        /* for window raster dump */
	AXWidget_corner corner;		/* bottom right corner */
	Colormap colorLUT;
	AXWidget_colorpal colorpal;	/* bottom widget */
	AXWidget_ge_ruler GExruler;	/* bottom widget */
	AXWidget_ge_ruler GEyruler;	/* right widget */
	long reserved[3];
#else
        unsigned long xwin;                                                
        long topwindow;                                                    
        unsigned long editwin;  /* Window is a typedef for unsigned long */
		Window inCellWin;		/* incell edit window for SS **/
        unsigned long rowwin;                                              
        unsigned long colwin;                                              
        long reserved[21];                                                 
#endif

	short p_id;
	short ap_id;
	short sub_window;
	struct TmCrePacket wind_attr;
	short option_len;
	short title_len;
	struct TmRegPacket *regions;
	short cur_reg_num ;
	struct hlist *hlhead;	/* highlighted regions list */
	short help_num;

	short creOrd;		/* creation order number */

	short busy_flag;
	short mouse_type;
	short mouse_state;

	short x_cur_pos;
	short y_cur_pos;
	TmLineDescriptor lin_des;
	TmTextDescriptor write_tx_des;
	TmTextDescriptor read_tx_des;
    short bgcolor, fgcolor;
	short trans_mode;
	short txAlign;			/* top-left or base-left alignment of text */
	short spWid;			/* default space width in pixel */
	short spCnt1;			/* space counts set by alter text extent */
	short spCnt2;			/* spCnt1 is no of spaces with spChange+1 */
							/* spCnt2 is no of spaces with spChange */
	short spChng1;			/* added to spWid to effect total change 
								in text extent (for spCnt1) */
	short spChng2;			/* added to spWid to effect total change 
								in text extent (for spCnt2) */
	Layer *layer;			/* layered bitmap for window */
	Layer *iconlayer;		/* layered bitmap for set-aside icon */
	struct TmBounds *bounds; /* border icons data base */
	AxRectangle inner;		/* inner clipping rectangle */
	AxRectangle outer;		/* outer clipping rectangle */
	AxRectangle appClip;		/* application clipping rectangle */
	short RegenClippingMode;	/* TRUE while doing partial regen */
	short DamagedRectPending;	/* will go into RegenClippingMode */
	AxRectangle damagedRect;	/* area that needs regened on LK_REGEN */
	AxRectangle fixingDamage;	/* current clip from regen by app */
	TmTabStop *tabs;
	short numTabs;
	short decimalChar;		/* character used in decimal tabs */

/* following variables are useful in TmModify */
	short newPS;			/* new position and size */
	short newTit;			/* new title */
	short newBord;			/* new bordercontrols */

	/* field read status */

	short tc_x;			/* old text cursor x */
	XPoint tcPoints[4]; /* as of last time cursor was turned on */
	XPoint tcInactivePoints[4]; /* used with openlook style cursor */
	short tcIsOn;		/* text cursor is on boolean */
	short tcType;		/* current text cursor type */
	short tcActive;		/* tracks window focus */

	short term_flag;
	short text_overtype;
	AxRectangle frClip;			/* clipping rectangle for field read */
	short frTrMode;				/* transfer mode for field read */
	short fldAbove;				/* ascent of field */
	short fldBelow;				/* descent of field */

	/* pick and stroke input */

	union TmTermData td;		/* struct still used by pick/stroke tmouse.c */
	short td_type;
	short td_len;
	short init_str_flag;
	short init_stk_flag;
	struct TmPickTrigger p_trig;
	struct TmStkTrigger s_trig;
								/* local echoing of strokes */
	short stkCode;				/* action code for stroke echoing */
	short stkReportF;			/* report individual stroke points flag */
	short useButF;				/* use button-down as first anchor point */
	TmPoint32 stkAnc[2];		/* anchor points for stroke echoing */
	short stkStyle;				/* line style for stroke echoing */
	AxRectangle stkClip;			/* clipping rectangle for stroke echoing */

	char *keyBindings;	/* window's keybindings table (see tmkeys.c) */
	short winWasBorrowed;/* 0 or winID of stolen X-window */

	long docId;		/* 0 or ID of document owned by window */
	int privateMenuBar;	/* override from using icon_id */

	/* misc */

	int widget_uid;		/* incrementing uid when window gets created */
    short near_mouse;   /* create/activate near mouse activity */
    short can_index;    /* TRUE if INDEX key has any effect */
    short regen_locked; /* am a window waiting to do regen */
	menu_item_def *popups;	/* array of popup window info */

	/* imagetext mode stuff */
	int imageTextZoom; /* x1, x2, x4 */
	int imageTextTarget;   /* target setup reference for default metrics */
	TmPoint32 imageBox[4]; /* footprint of text object as displayed */
	TmPoint32 imageVBox[4]; /* footprint of last pass text object */
	int Tx, Ty;			   /* translation */
	double Hx, Hy;		   /* shear */
	double Sx, Sy;         /* scale */
	double sinAng, cosAng; /* rotate */
	int devTx, devTy;		/* same transform stuff for target metrics */
	double devHx, devHy;
	double devSx, devSy;
	double devSinAng, devCosAng;
	TmPoint32 baseAdjust; /* adjustment to baseline position on screen */
	struct fieldInfo textFieldInfo;
	struct fieldInfo fieldVInfo; /* previous dispf info */
	short useOldScaler;

	/* grid / ruler info */
	int gridSize;
	int x_zero_pt;
	int y_zero_pt;
	int x_scroll_origin;
	int y_scroll_origin;
	int x_zoom;
	int y_zoom;
	int dpi;
	int unitSize;
	int unitPrec;
	char unitName[MAX_RULER_UNIT_NAME];

	/* polyline/bezier curve info */
	int num_points;
	TmPoint32 *point_list;
	int num_control_points;
	TmPoint32 *control_list;

	/* line and fill info */
	int line_width, line_style, end1, end2, join_style, fill_style;
	int line_fg, line_bg, line_pat, fill_fg, fill_bg, fill_pat;
	int ffg_cyan, ffg_magn, ffg_yell, ffg_blck;
	int fbg_cyan, fbg_magn, fbg_yell, fbg_blck;
	int lfg_cyan, lfg_magn, lfg_yell, lfg_blck;
	int lbg_cyan, lbg_magn, lbg_yell, lbg_blck;
	int line_mode, lfill_type, lfill_ang, lfill_off;
	int ffill_type, ffill_ang, ffill_off;
	char *line_pat_name, *fill_pat_name;

};  /* end wind_struct definition */


#define wind_pnt Wp

/* name change: character height above baseline */
#define CharAbove	CharHome

		/* device characteristics table */

typedef struct 
{
	short h_desktop;
	short w_desktop;
	short d_height;
	short d_width;
} DcharTab;

		/* algorithmic text attributes (text descriptor field values) */

typedef struct
{
	short cc, hi, uu;
} AlgoAttr;

/***********
 * Globals *
 ***********/


		/* array of window control blocks */

extern DLL_IMPORT struct wind_struct window[], *wind_pnt;

		/* request dispatching globals */

extern DLL_IMPORT short cur_ap_id;
extern DLL_IMPORT short cur_p_id;
extern DLL_IMPORT short mes_len;

		/* general globals */

extern DLL_IMPORT int TermFD;	/* file descriptor for terminal output */
extern DLL_IMPORT FILE *fd2;		/* log file pointer */
extern DLL_IMPORT FILE *ATfp;		/* ATM trace file pointer */
extern DLL_IMPORT short ATMdebug;		/* debug switch */
extern DLL_IMPORT short ATMtrace;		/* trace switch */

extern DLL_IMPORT short cur_win;		/* active window ID */
extern DLL_IMPORT short cur_flag;		/* cursor on/off indicator */

extern DLL_IMPORT short ACwin;		/* ID of window created by APAC */
extern DLL_IMPORT short IntWin;		/* ID of window created by APINT */
extern DLL_IMPORT long WindowCreationCounter;    /* to supply uid's for each win */

		/* string globals */

extern DLL_IMPORT char IntString[];
extern DLL_IMPORT char HelpTx0[];
extern DLL_IMPORT char HelpTx1[];
extern DLL_IMPORT short IntLen;


		/* mouse variables */

extern DLL_IMPORT short mous_x;
extern DLL_IMPORT short mous_y;
extern DLL_IMPORT short mx;
extern DLL_IMPORT short my;

		/* built-in globals */


		/* external functions that must be declared */

char *ATMmalloc();
char *ATMrealloc();


#endif    /*  PARINT_INCLUDED  */
