/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: pol.c
   Applix ASC version: 128
   Date: Thu Jan 28 21:10:48 1999
*/
#include <stdio.h>
#include <math.h>
#include <aos.h>
#include <tmi.h>
#include <oldwin.h>
#include <atm.h>
#include <atmattr.h>
#include <colors.h>
#include <basicwin.h>

#if AXMS
#define hypot _hypot
#define GRADIENT_FILL(fill) (fill == RECT_FILL_TYPE || \
                            fill == LINEAR_FILL_TYPE || \
                            fill == RADIAL_FILL_TYPE || \
                            fill == BILINEAR_FILL_TYPE)

extern int Running_Windows95;
#endif
#define ABS(x) (((x) > 0) ? (x) : -(x))
#define N_GRAD_PATS 8

#include <pol.p>
extern DLL_IMPORT Window   RootWin;
extern Display *Dpy;
extern GC Gc1, Gc2;
extern Drawable HostWin;
extern int PlaneMask;
extern unsigned long FGpixel, BGpixel;
extern Pixmap ScratchMap;
extern unsigned long AlisPixels[];
extern unsigned long WidgetPixels[];

static char basicdash[2] = {  7, 5 };
static char fat__dott[2] = {  3, 5 };
static char dot__dash[4] = {  7, 5, 3, 5 };
static char thin_dott[2] = {  1, 1 };
static char dtdt_dash[6] = {  7, 5, 3, 5, 3, 5 };
static char long_dash[2] = { 14, 5 };

static int dash_offset[] = { 0, 0, 0, 0, 0, 0 };
static int dash_length[] = { 2, 2, 4, 2, 6, 2 };       
static char *dash_list[] = {
    basicdash, fat__dott, dot__dash, thin_dott, dtdt_dash, long_dash
};


/*******************************************************************/
getRoundPenOutline(lineWidth, numSegments, pattern)
int lineWidth;
int *numSegments;
short **pattern;
{
    int x,y, i, n;
    double d;
    short *p;

    /* special cases - hand tune the smallest lines */
    if ( lineWidth == 1 )
    {
        *numSegments = 1;
        *pattern = (short *)TaskAlloc(0, 2*(*numSegments)*sizeof(short));
        (*pattern)[0] = (*pattern)[1] = 0;
        return;
    }
    else if ( lineWidth == 2 )
    {
        *numSegments = 4;
        *pattern = (short *)TaskAlloc(0, 2*(*numSegments)*sizeof(short));
        (*pattern)[0] = 0;
        (*pattern)[1] = 0;
        (*pattern)[2] = 1;
        (*pattern)[3] = 0;
        (*pattern)[4] = 0;
        (*pattern)[5] = 1;
        (*pattern)[6] = -1;
        (*pattern)[7] = 0;
        return;
    }
    else if ( lineWidth == 3 )
    {
        *numSegments = 5;
        *pattern = (short *)TaskAlloc(0, 2*(*numSegments)*sizeof(short));
        (*pattern)[0] = 0;
        (*pattern)[1] = -1;
        (*pattern)[2] = 1;
        (*pattern)[3] = 1;
        (*pattern)[4] = -1;
        (*pattern)[5] = 1;
        (*pattern)[6] = -1;
        (*pattern)[7] = -1;
        (*pattern)[8] = 1;
        (*pattern)[9] = 0;
        return;
    }
    else if ( lineWidth == 4 )
    {
        *numSegments = 12;
        *pattern = (short *)TaskAlloc(0, 2*(*numSegments)*sizeof(short));
        (*pattern)[0] = 0;
        (*pattern)[1] = -1;
        (*pattern)[2] = 1;
        (*pattern)[3] = 0;
        (*pattern)[4] = 0;
        (*pattern)[5] = 1;
        (*pattern)[6] = 1;
        (*pattern)[7] = 0;
        (*pattern)[8] = 0;
        (*pattern)[9] = 1;
        (*pattern)[10] = -1;
        (*pattern)[11] = 0;
        (*pattern)[12] = 0;
        (*pattern)[13] = 1;
        (*pattern)[14] = -1;
        (*pattern)[15] = 0;
        (*pattern)[16] = 0;
        (*pattern)[17] = -1;
        (*pattern)[18] = -1;
        (*pattern)[19] = 0;
        (*pattern)[20] = 0;
        (*pattern)[21] = -1;
        (*pattern)[22] = 1;
        (*pattern)[23] = 0;
        return;
    }
    else if ( lineWidth == 5 )
    {
        *numSegments = 16;
        *pattern = (short *)TaskAlloc(0, 2*(*numSegments)*sizeof(short));
        (*pattern)[0] = -1;
        (*pattern)[1] = -1;
        (*pattern)[2] = 0;
        (*pattern)[3] = -1;
        (*pattern)[4] = 1;
        (*pattern)[5] = 0;
        (*pattern)[6] = 1;
        (*pattern)[7] = 0;
        (*pattern)[8] = 0;
        (*pattern)[9] = 1;
        (*pattern)[10] = 1;
        (*pattern)[11] = 0;
        (*pattern)[12] = 0;
        (*pattern)[13] = 1;
        (*pattern)[14] = 0;
        (*pattern)[15] = 1;
        (*pattern)[16] = -1;
        (*pattern)[17] = 0;
        (*pattern)[18] = 0;
        (*pattern)[19] = 1;
        (*pattern)[20] = -1;
        (*pattern)[21] = 0;
        (*pattern)[22] = -1;
        (*pattern)[23] = 0;
        (*pattern)[24] = 0;
        (*pattern)[25] = -1;
        (*pattern)[26] = -1;
        (*pattern)[27] = 0;
        (*pattern)[28] = 0;
        (*pattern)[29] = -1;
        (*pattern)[30] = 0;
        (*pattern)[31] = -1;
        return;
    }

    /* greater than 5 pix wide, no need for compatability - calc circle */
    p = *pattern = (short *)TaskAlloc(0, 2*lineWidth*sizeof(short));
    x = 0;
    y = lineWidth / 2;
    d = 5.0 / 4.0 - (double)lineWidth / 2.0;
    n = p[0] = p[1] = 0;

    /* get the circle from 90 to 45 degrees */
    while ( y > x )
    {
        if ( d < 0 )
        {
            d += 2 * x++ + 3;
            n++;
            p[2*n]     = 1;
            p[2*n + 1] = 0;
        }
        else
        {
            d += 2 * (x++ - y--) + 5;
            n++;
            p[2*n]     = 0;
            p[2*n + 1] = 1;
            n++;
            p[2*n]     = 1;
            p[2*n + 1] = 0;
        }
    }

    /* copy the 2nd octant points to the other 7 octants */
    p = *pattern = (short *)realloc(*pattern, 16 * n * sizeof(short));
    for ( i=0; i<n; i++ )
    {
        /* second octant value at this index */
        x = p[2*i];
        y = p[2*i+1];

        /* first to eighth (backwards), (x,y) = (y,x) */
        p[4*n - (2*i+2)] = y;
        p[4*n - (2*i+1)] = x;

        /* eighth to seventh, (x,y) = (-y,x) */
        p[4*n + 2*i]   = (-1) * y;
        p[4*n + 2*i+1] = x;

        /* seventh to sixth (backwards), (x,y) = (-x,y) */
        p[8*n - (2*i+2)] = (-1) * x;
        p[8*n - (2*i+1)] = y;

        /* sixth to fifth, (x,y) = (-x,-y) */
        p[8*n + 2*i]   = (-1) * x;
        p[8*n + 2*i+1] = (-1) * y;
 
        /* fifth to fourth (backwards), (x,y) = (-y,-x) */
        p[12*n - (2*i+2)] = (-1) * y;
        p[12*n - (2*i+1)] = (-1) * x;
 
        /* fourth to third, (x,y) = (y,-x) */
        p[12*n + 2*i]   = y;
        p[12*n + 2*i+1] = (-1) * x;
 
        /* third return to second (backwards), (x,y) = (x,-y) */
        p[16*n - (2*i+2)] = x;
        p[16*n - (2*i+1)] = (-1) * y;
    }
 
    /* set the first point and set the number of point pairs */
    p[1] = 1 + (-1) * lineWidth / 2;
    *numSegments = 8 * n;
}

/****************************************************************************/
rotate_around_center(pt, center, sinAng, cosAng)
XPoint *pt, *center;
double sinAng, cosAng;
{
    int xt, yt;

    xt = pt->x - center->x;
    yt = pt->y - center->y;
    pt->x = (double)xt*cosAng + (double)yt*sinAng + (double)center->x;
    pt->y = (double)yt*cosAng - (double)xt*sinAng + (double)center->y;
}

/****************************************************************************/
get_percent_ix(percent)
int percent;
{
    int ix;
    if ( percent < 62.5 )
        ix = 0;
    else if ( percent < 187.5 )
        ix = 12;
    else if ( percent < 312.5 )
        ix = 1;
    else if ( percent < 437.5 )
        ix = 10;
    else if ( percent < 562.5 )
        ix = 2;
    else if ( percent < 687.5 )
        ix = 21;
    else if ( percent < 812.5 )
        ix = 3;
    else if ( percent < 937.5 )
        ix = 13;
    else
        ix = 4;
    return(ix);
}

/****************************************************************************/
get_gradient_map(pmap, ix, inv)
Pixmap *pmap;
int ix, inv;
{
    switch ( ix )
    {
        case 0:  *pmap = inv ? GetStippleImap(0)  : GetStippleMap(0);   break;
        case 1:  *pmap = inv ? GetStippleImap(12) : GetStippleMap(12);  break;
        case 2:  *pmap = inv ? GetStippleImap(1)  : GetStippleMap(1);   break;
        case 3:  *pmap = inv ? GetStippleImap(10) : GetStippleMap(10);  break;
        case 4:  *pmap = inv ? GetStippleImap(2)  : GetStippleMap(2);   break;
        case 5:  *pmap = inv ? GetStippleMap(10)  : GetStippleImap(10); break;
        case 6:  *pmap = inv ? GetStippleImap(3)  : GetStippleMap(3);   break;
        case 7:  *pmap = inv ? GetStippleImap(13) : GetStippleMap(13);  break;
        default: *pmap = inv ? GetStippleImap(4)  : GetStippleMap(4);   break;
    }
}

/****************************************************************************/
Pixmap
get_color_pixmap(wid,hyt,data,cmap,inv)
int wid,hyt;
char *data;
cmap_type *cmap;
int inv;
{
    Pixmap fill_map;
    color_type *tmp_color;
    unsigned char *imgData;
    char *loc_buf;
    int i, r,g,b,k, byte_wid, sz;
    col_table col_tab[256];

    byte_wid = wid + wid%2;
    sz = byte_wid*hyt;
    loc_buf = (char *)TaskAlloc(0, sz*sizeof(char));
    memcpy(loc_buf,data,sz);

    fill_map = (Pixmap)AxCreateColorPixmap(loc_buf,wid,hyt,cmap);

    TaskFree(0, loc_buf);
    return(fill_map);
}

/****************************************************************************/
Pixmap
get_mono_pixmap(data, wid,hyt,inv)
char *data;
int wid,hyt;
int inv;
{
    int byte_wid, sz;
    Pixmap fill_map;
    char *loc_buf;

    byte_wid = (wid+7)/8 + ((wid+7)/8)%2;
    sz = byte_wid*hyt;
    loc_buf = (char *)TaskAlloc(0, sz*sizeof(char));
    memcpy(loc_buf,data,sz);

    fill_map = (Pixmap)AxCreatePMFromBM(loc_buf,wid,hyt,1, inv);

    TaskFree(0, loc_buf);
    return(fill_map);
}

/****************************************************************************/
do_color_picture_pattern(xwin, x1, y1, wid, hyt, attr, for_line)
Drawable xwin;
int x1, y1, wid, hyt;
attr_type *attr;
int for_line;
{
	ge_image_type imP;
	int x_off, y_off;
	TmPoint32 pnt;
	poly_type xform;

	convPixToMils(x1, y1, &x1, &y1);
	convPixToMils(wid, hyt, &wid, &hyt);
	pnt.x = x1;
	pnt.y = y1;
	if ( for_line == FALSE )
	{
		memset(&imP, 0, sizeof(ge_image_type));
		memset(&xform, 0, sizeof(poly_type));
		imP.hyt = attr->f_dhyt;
		imP.wid = attr->f_dwid;
		imP.dep = attr->f_ddep;
		imP.data = attr->f_data;
		if ( attr->f_ddep == 8 )
			imP.data_byte_wid = imP.wid + imP.wid%2;
		xform.nPoints = 5;
		xform.points =  (TmPoint32 *)TaskAlloc(0, 5*sizeof(TmPoint32));
		xform.points[0].x = x1;
		xform.points[0].y = y1;
		xform.points[1].x = x1 + wid;
		xform.points[1].y = y1;
		xform.points[2].x = x1 + wid;
		xform.points[2].y = y1 + hyt;
		xform.points[3].x = x1;
		xform.points[3].y = y1 + hyt;
		xform.points[4].x = x1;
		xform.points[4].y = y1;
		do_ax_pimage(xwin, 0, &pnt, &imP, attr, &xform,
	                1.0, 1.0, attr->f_cmap);
		TaskFree(0, xform.points);
	}
}

/****************************************************************************/
do_mask_pattern(xwin, x1, y1, wid, hyt, attr, for_line)
Drawable xwin;
int x1, y1, wid, hyt;
attr_type *attr;
int for_line;
{
    int ix, c, m, y, k, inv;
    Pixmap fill_map;
    unsigned long pel;

    /* get the pattern for filling */
    if ( for_line )
    {
        inv = attr->lfg.see_thru ? TRUE : FALSE;
        if ( attr->l_type != 1 ) /* assume built in fill, get map by index */
            fill_map = inv ? GetStippleImap(attr->l_pat) : 
                              GetStippleMap(attr->l_pat);
        else
        {
            /* user defined fill */
            fill_map = get_mono_pixmap(attr->l_data, attr->l_dwid,
                                        attr->l_dhyt, inv);
        }

        if ( inv )
        {
            if ( attr->use_lbg )
                pel = AlisPixels[TMA_WHITE];
            else
            {
                c = attr->lbg.c;
                m = attr->lbg.m;
                y = attr->lbg.y;
                k = attr->lbg.k;
                AxGetRealPixel(2, c,m,y,k, 255, &pel);
            }
        }
        else
        {
            if ( attr->use_lfg )
                pel = AlisPixels[TMA_BLACK];
            else
            {
                c = attr->lfg.c;
                m = attr->lfg.m;
                y = attr->lfg.y;
                k = attr->lfg.k;
                AxGetRealPixel(2, c,m,y,k, 255, &pel);
            }
        }
    }
    else
    {
        inv = attr->ffg.see_thru ? TRUE : FALSE;
        if ( attr->f_type != 1 ) /* assume built in fill, get map by index */
            fill_map = inv ? GetStippleImap(attr->f_pat) : 
                             GetStippleMap(attr->f_pat);
        else
        {
            /* user defined fill */
            fill_map = get_mono_pixmap(attr->f_data, attr->f_dwid,
                                        attr->f_dhyt, inv);
        }

        if ( inv )
        {
            if ( attr->use_fbg )
                pel = AlisPixels[TMA_WHITE];
            else
            {
                c = attr->fbg.c;
                m = attr->fbg.m;
                y = attr->fbg.y;
                k = attr->fbg.k;
                AxGetRealPixel(2, c,m,y,k, 255, &pel);
            }
        }
        else
        {
            if ( attr->use_fbg )
                pel = AlisPixels[TMA_BLACK];
            else
            {
                c = attr->ffg.c;
                m = attr->ffg.m;
                y = attr->ffg.y;
                k = attr->ffg.k;
                AxGetRealPixel(2, c,m,y,k, 255, &pel);
            }
        }
    }

    /* draw */
    XSetForeground(Dpy, Gc1, pel);
    XSetFunction(Dpy, Gc1, GXcopy);
    XSetStipple(Dpy, Gc1, fill_map);
    XFillRectangle(Dpy, xwin, Gc1, x1, y1, wid, hyt);

    /* done */
    if ( for_line && attr->l_type == 1 )
        AxFreePixmap(fill_map);
    else if ( !for_line && attr->f_type == 1 )
        AxFreePixmap(fill_map);
}

/****************************************************************************/
do_tile_pattern(xwin, x1, y1, wid, hyt, attr, for_line)
Drawable xwin;
int x1, y1, wid, hyt;
attr_type *attr;
int for_line;
{
    unsigned long f_pel, b_pel;
    int ix, c,m,y,k;
    Pixmap fill_map;

    /* get the pattern for filling */
    if ( for_line )
    {
        if ( attr->l_type != 1 ) /* assume built in fill, get map by index */
            fill_map = GetStippleMap(attr->l_pat);
        else
        {
            /* user defined fill */
            fill_map = get_mono_pixmap(attr->l_data, attr->l_dwid,
                                        attr->l_dhyt, FALSE);
        }
        if ( attr->use_lfg )
            f_pel = AlisPixels[TMA_BLACK];
        else
        {
            c = attr->lfg.c;
            m = attr->lfg.m;
            y = attr->lfg.y;
            k = attr->lfg.k;
            AxGetRealPixel(2, c,m,y,k, 255, &f_pel);
        }
        if ( attr->use_lbg )
            b_pel = AlisPixels[TMA_WHITE];
        else
        {
            c = attr->lbg.c;
            m = attr->lbg.m;
            y = attr->lbg.y;
            k = attr->lbg.k;
            AxGetRealPixel(2, c,m,y,k, 255, &b_pel);
        }
    }
    else
    {
        if ( attr->f_type != 1 ) /* assume built in fill, get map by index */
            fill_map = GetStippleMap(attr->f_pat);
        else
        {
            /* user defined fill */
            fill_map = get_mono_pixmap(attr->f_data, attr->f_dwid,
                                        attr->f_dhyt, FALSE);
        }
        if ( attr->use_ffg )
            f_pel = AlisPixels[TMA_BLACK];
        else
        {
            c = attr->ffg.c;
            m = attr->ffg.m;
            y = attr->ffg.y;
            k = attr->ffg.k;
            AxGetRealPixel(2, c,m,y,k, 255, &f_pel);
        }
        if ( attr->use_fbg )
            b_pel = AlisPixels[TMA_WHITE];
        else
        {
            c = attr->fbg.c;
            m = attr->fbg.m;
            y = attr->fbg.y;
            k = attr->fbg.k;
            AxGetRealPixel(2, c,m,y,k, 255, &b_pel);
        }
    }

    /* draw */
    XSetForeground(Dpy, Gc1, f_pel);
    XSetBackground(Dpy, Gc1, b_pel);
    XSetFunction(Dpy, Gc1, GXcopy);
    XSetStipple(Dpy, Gc1, fill_map);
    XFillRectangle(Dpy, xwin, Gc1, x1, y1, wid, hyt);

    /* done */
    if ( for_line && attr->l_type == 1 )
        AxFreePixmap(fill_map);
    else if ( !for_line && attr->f_type == 1 )
        AxFreePixmap(fill_map);
}

/****************************************************************************/
do_color_tile_pattern(xwin, x1, y1, wid, hyt, attr, for_line)
Drawable xwin;
int x1, y1, wid, hyt;
attr_type *attr;
int for_line;
{
    unsigned long f_pel, b_pel, cmapper[256];
    int i, c,m,y,k;
    Pixmap fill_map;

    /* get the pattern for filling */
    if ( for_line )
    {
        fill_map = get_color_pixmap(attr->l_dwid, attr->l_dhyt,
                                    attr->l_data, attr->l_cmap,0);

        if ( attr->use_lfg )
            f_pel = AlisPixels[TMA_BLACK];
        else
        {
            c = attr->lfg.c;
            m = attr->lfg.m;
            y = attr->lfg.y;
            k = attr->lfg.k;
            AxGetRealPixel(2, c,m,y,k, 255, &f_pel);
        }
        if ( attr->use_lbg )
            b_pel = AlisPixels[TMA_WHITE];
        else
        {
            c = attr->lbg.c;
            m = attr->lbg.m;
            y = attr->lbg.y;
            k = attr->lbg.k;
            AxGetRealPixel(2, c,m,y,k, 255, &b_pel);
        }
    }
    else
    {
        fill_map = get_color_pixmap(attr->f_dwid, attr->f_dhyt,
                                    attr->f_data, attr->f_cmap,0);
        if ( attr->use_ffg )
            f_pel = AlisPixels[TMA_BLACK];
        else
        {
            c = attr->ffg.c;
            m = attr->ffg.m;
            y = attr->ffg.y;
            k = attr->ffg.k;
            AxGetRealPixel(2, c,m,y,k, 255, &f_pel);
        }
        if ( attr->use_fbg )
            b_pel = AlisPixels[TMA_WHITE];
        else
        {
            c = attr->fbg.c;
            m = attr->fbg.m;
            y = attr->fbg.y;
            k = attr->fbg.k;
            AxGetRealPixel(2, c,m,y,k, 255, &b_pel);
        }
    }

    /* draw */
    XSetForeground(Dpy, Gc1, f_pel);
    XSetBackground(Dpy, Gc1, b_pel);
    XSetFunction(Dpy, Gc1, GXcopy);
    XSetTile(Dpy, Gc1, fill_map);
    XFillRectangle(Dpy, xwin, Gc1, x1, y1, wid, hyt);

    /* done */
    AxFreePixmap(fill_map);
}

/****************************************************************************/
do_mask_bilinear_gradient(xwin, x1, y1, wid, hyt, attr, for_line)
Drawable xwin;
int x1, y1, wid, hyt;
attr_type *attr;
int for_line;
{
    double angle, sinAng, cosAng;
    int left, right, top, bottom;
    int i, j, patTopHyt, patBotHyt, inv, thyt, bhyt;
    XPoint pts[5], center;
    int c, m, y, k, ix, x_off, y_off;
    Pixmap pattern;
    unsigned long pel;
#if AXMS
    int orig_x = x1, orig_y = y1;
    int orig_wid = wid, orig_hyt = hyt;
#endif

    center.x = x1 + wid/2;
    center.y = y1 + hyt/2;
    y_off = attr->f_yoff/200.0 * hyt;
    thyt = center.y - y1 + y_off;
    bhyt = y1 + hyt - center.y - y_off - 1;

    /* rotate the coordinates */
    angle  = (double)attr->f_ang / 180.0 * M_PI;
    sinAng = (double)sin(angle);
    cosAng = (double)cos(angle);
    pts[0].x = pts[2].x = x1;
    pts[0].y = pts[1].y = y1;
    pts[1].x = pts[3].x = x1 + wid;
    pts[2].y = pts[3].y = y1 + hyt;
    rotate_around_center(&pts[0], &center, sinAng, cosAng);
    rotate_around_center(&pts[1], &center, sinAng, cosAng);
    rotate_around_center(&pts[2], &center, sinAng, cosAng);
    rotate_around_center(&pts[3], &center, sinAng, cosAng);

    /* recalc the size and position */
    x1  = MIN( MIN( MIN(pts[0].x, pts[1].x), pts[2].x), pts[3].x);
    y1  = MIN( MIN( MIN(pts[0].y, pts[1].y), pts[2].y), pts[3].y);
    wid = MAX( MAX( MAX(pts[0].x, pts[1].x), pts[2].x ), pts[3].x ) - x1;
    hyt = MAX( MAX( MAX(pts[0].y, pts[1].y), pts[2].y ), pts[3].y ) - y1;

    /* set the color */
    if ( for_line )
    {
        inv = attr->lfg.see_thru ? TRUE : FALSE;
        if ( inv )
        {
            c = attr->lbg.c;
            m = attr->lbg.m;
            y = attr->lbg.y;
            k = attr->lbg.k;
        }
        else
        {
            c = attr->lfg.c;
            m = attr->lfg.m;
            y = attr->lfg.y;
            k = attr->lfg.k;
        }
    }
    else
    {
        inv = attr->ffg.see_thru ? TRUE : FALSE;
        if ( inv )
        {
            c = attr->fbg.c;
            m = attr->fbg.m;
            y = attr->fbg.y;
            k = attr->fbg.k;
        }
        else
        {
            c = attr->ffg.c;
            m = attr->ffg.m;
            y = attr->ffg.y;
            k = attr->ffg.k;
        }
    }
    y_off = attr->f_yoff/200.0 * hyt;
    thyt = hyt/2  + y_off;
    bhyt = hyt/2 - y_off;
    patTopHyt = thyt/(N_GRAD_PATS + 1);
    patBotHyt = bhyt/(N_GRAD_PATS + 1);
    AxGetRealPixel(2, c,m,y,k, 255, &pel);
#if AXMS    /* { */
    /* NT only!  Windows 95 functionality is done at the AxFill level */
    win_mask_bilinear_gradient(xwin, Gc1, x1, y1, wid, hyt, orig_x, orig_y,
                    orig_wid, orig_hyt, patTopHyt, pel, inv, sinAng, cosAng);
#else    /* } { */
    XSetForeground(Dpy, Gc1, pel);
    XSetFunction(Dpy, Gc1, GXcopy);

    /* draw the gradient thru the mask */
    left   = x1;
    right  = x1 + wid;
    top    = inv ? y1 + hyt - patBotHyt : y1 + thyt - patTopHyt;
    bottom = inv ? y1 + hyt : y1 + thyt;
    pts[0].x = pts[4].x = pts[3].x = left;
    pts[0].y = pts[4].y = pts[1].y = top;
    pts[1].x = pts[2].x = right;
    pts[2].y = pts[3].y = bottom;
    rotate_around_center(&pts[0], &center, sinAng,cosAng);
    rotate_around_center(&pts[1], &center, sinAng,cosAng);
    rotate_around_center(&pts[2], &center, sinAng,cosAng);
    rotate_around_center(&pts[3], &center, sinAng,cosAng);
    rotate_around_center(&pts[4], &center, sinAng,cosAng);
    XSetFillStyle (Dpy, Gc1, FillSolid);
    XFillPolygon(Dpy, xwin, Gc1, pts, 5, Complex, CoordModeOrigin);
    XSetFillStyle (Dpy, Gc1, FillStippled);

    for( k=0; k<N_GRAD_PATS; k++ )
    {
        j = N_GRAD_PATS - k;
        top    = inv ? y1 + (k+2)*patTopHyt : y1 + thyt + (k+1)*patBotHyt;
        bottom = inv ? y1 + (k+1)*patTopHyt : y1 + thyt + (k+2)*patBotHyt;
        pts[0].x = pts[4].x = pts[3].x = left;
        pts[0].y = pts[4].y = pts[1].y = top;
        pts[1].x = pts[2].x = right;
        pts[2].y = pts[3].y = bottom;
        rotate_around_center(&pts[0], &center, sinAng,cosAng);
        rotate_around_center(&pts[1], &center, sinAng,cosAng);
        rotate_around_center(&pts[2], &center, sinAng,cosAng);
        rotate_around_center(&pts[3], &center, sinAng,cosAng);
        rotate_around_center(&pts[4], &center, sinAng,cosAng);
        get_gradient_map(&pattern, j, FALSE);
        XSetStipple(Dpy, Gc1, pattern);
        XFillPolygon(Dpy, xwin, Gc1, pts, 5, Complex, CoordModeOrigin);

        top    = inv ? y1+thyt + (k+2)*patBotHyt : y1 + (k+1)*patTopHyt;
        bottom = inv ? y1+thyt + (k+1)*patBotHyt : y1 + (k+2)*patTopHyt;
        pts[0].x = pts[4].x = pts[3].x = left;
        pts[0].y = pts[4].y = pts[1].y = top;
        pts[1].x = pts[2].x = right;
        pts[2].y = pts[3].y = bottom;
        rotate_around_center(&pts[0], &center, sinAng,cosAng);
        rotate_around_center(&pts[1], &center, sinAng,cosAng);
        rotate_around_center(&pts[2], &center, sinAng,cosAng);
        rotate_around_center(&pts[3], &center, sinAng,cosAng);
        rotate_around_center(&pts[4], &center, sinAng,cosAng);
        get_gradient_map(&pattern, k, FALSE);
        XSetStipple(Dpy, Gc1, pattern);
        XFillPolygon(Dpy, xwin, Gc1, pts, 5, Complex, CoordModeOrigin);
    }

    left   = x1;
    right  = x1 + wid;
    top    = inv ? y1 : y1 + thyt;
    bottom = inv ? y1 + patTopHyt : y1 + thyt + patBotHyt;
    pts[0].x = pts[4].x = pts[3].x = left;
    pts[0].y = pts[4].y = pts[1].y = top;
    pts[1].x = pts[2].x = right;
    pts[2].y = pts[3].y = bottom;
    rotate_around_center(&pts[0], &center, sinAng,cosAng);
    rotate_around_center(&pts[1], &center, sinAng,cosAng);
    rotate_around_center(&pts[2], &center, sinAng,cosAng);
    rotate_around_center(&pts[3], &center, sinAng,cosAng);
    rotate_around_center(&pts[4], &center, sinAng,cosAng);
    XSetFillStyle (Dpy, Gc1, FillSolid);
    XFillPolygon(Dpy, xwin, Gc1, pts, 5, Complex, CoordModeOrigin);

#endif    /* } AXMS */
}

/****************************************************************************/
do_tile_bilinear_gradient(xwin, x1, y1, wid, hyt, attr, for_line)
Drawable xwin;
int x1, y1, wid, hyt;
attr_type *attr;
int for_line;
{
#if AXMS    /* { */

    /* NT only!  Windows 95 functionality is done at the AxFill level */
    win_bilinear_gradient(xwin, Gc1, x1, y1, wid, hyt, attr, for_line);

#else    /* AXMS } { AX_X11 */
    double angle, sinAng, cosAng, c,m,y, c_inc, m_inc, y_inc;
    int c_steps, t_steps, i, j, k, n, patTopHyt, patBotHyt;
    int f_c,f_y,f_m,f_k, b_c,b_m,b_y,b_k, avail;
    int left, right, btop, tbot, bbot, y_off, thyt, bhyt, ttop;
    unsigned long back_pixel, fore_pixel;
    XPoint pts[5], center;
    Pixmap pattern;

    y_off = attr->f_yoff/200.0 * hyt;
    center.x = x1 + wid/2;
    center.y = y1 + hyt/2;
    thyt = center.y - y1 + y_off;
    bhyt = y1 + hyt - center.y - y_off - 1;

    /* rotate the coordinates */
    angle  = (double)attr->f_ang / 180.0 * M_PI;
    sinAng = (double)sin(angle);
    cosAng = (double)cos(angle);
    pts[0].x = pts[2].x = x1;
    pts[0].y = pts[1].y = y1;
    pts[1].x = pts[3].x = x1 + wid;
    pts[2].y = pts[3].y = y1 + hyt;
    rotate_around_center(&pts[0], &center, sinAng, cosAng);
    rotate_around_center(&pts[1], &center, sinAng, cosAng);
    rotate_around_center(&pts[2], &center, sinAng, cosAng);
    rotate_around_center(&pts[3], &center, sinAng, cosAng);

    /* recalc the size and position */
    x1  = MIN( MIN( MIN(pts[0].x, pts[1].x), pts[2].x), pts[3].x);
    y1  = MIN( MIN( MIN(pts[0].y, pts[1].y), pts[2].y), pts[3].y);
    wid = MAX( MAX( MAX(pts[0].x, pts[1].x), pts[2].x ), pts[3].x ) - x1;
    hyt = MAX( MAX( MAX(pts[0].y, pts[1].y), pts[2].y ), pts[3].y ) - y1;

    /* remove black component from color step calcs */
    if ( for_line )
    {
        f_c = MIN(attr->lfg.c + attr->lfg.k, 255);
        f_m = MIN(attr->lfg.m + attr->lfg.k, 255);
        f_y = MIN(attr->lfg.y + attr->lfg.k, 255);
        f_k = 0;
        b_c = MIN(attr->lbg.c + attr->lbg.k, 255);
        b_m = MIN(attr->lbg.m + attr->lbg.k, 255);
        b_y = MIN(attr->lbg.y + attr->lbg.k, 255);
        b_k = 0;
    }
    else
    {
        f_c = MIN(attr->ffg.c + attr->ffg.k, 255);
        f_m = MIN(attr->ffg.m + attr->ffg.k, 255);
        f_y = MIN(attr->ffg.y + attr->ffg.k, 255);
        f_k = 0;
        b_c = MIN(attr->fbg.c + attr->fbg.k, 255);
        b_m = MIN(attr->fbg.m + attr->fbg.k, 255);
        b_y = MIN(attr->fbg.y + attr->fbg.k, 255);
        b_k = 0;
    }

    /* draw the top half */
    y_off = attr->f_yoff/200.0 * hyt;
    if ( thyt > 0 )
    {
        thyt = hyt/2  + y_off;
        /* get the number of divisible sections */
        avail = MAX(1, (int)(thyt/(N_GRAD_PATS+1)));

        /* get the maximum number of color steps */
        c_steps = ABS(f_c - b_c);
        c_steps = MAX(ABS(f_m - b_m), c_steps);
        c_steps = MAX(ABS(f_y - b_y), c_steps);
        c_steps = MIN(avail, c_steps);

        /* get the number of color steps FOR THIS DISPLAY */
        t_steps = 1;
        left  = x1;
        right = x1 + wid;
        ttop = y1;
        tbot = y1;
        btop = y1+hyt;
        bbot = y1+hyt;
        if ( c_steps > 0 )
        {
            c_inc = (double)(f_c - b_c) / (double)(c_steps);
            m_inc = (double)(f_m - b_m) / (double)(c_steps);
            y_inc = (double)(f_y - b_y) / (double)(c_steps);
            c = b_c;
            m = b_m;
            y = b_y;
            AxGetRealPixel(2, b_c, b_m, b_y, 0, 255, &back_pixel);

            for ( i=0; i<c_steps; i++ )
            {
                /* increment the foreground color */
                c += c_inc;
                m += m_inc;
                y += y_inc;
                AxGetRealPixel(2, (int)c, (int)m, (int)y, 0, 255, &fore_pixel);

                /* if there is no color change, widen patterns and try again */
                if ( fore_pixel != back_pixel )
                {
                    t_steps++;
                    back_pixel = fore_pixel;
                }
            }
            AxGetRealPixel(2, f_c, f_m, f_y, 0, 255, &fore_pixel);
            if ( fore_pixel != back_pixel )
                t_steps++;

            /* tile in the gradient patterns */
            patTopHyt = thyt / (t_steps*N_GRAD_PATS+1);
            while ( patTopHyt == 0 && t_steps > 1 )
            {
                t_steps--;
                patTopHyt = thyt / (t_steps*N_GRAD_PATS+1);
            }


            c_inc = (double)(f_c - b_c) / (double)(c_steps);
            m_inc = (double)(f_m - b_m) / (double)(c_steps);
            y_inc = (double)(f_y - b_y) / (double)(c_steps);
            c = b_c;
            m = b_m;
            y = b_y;
            AxGetRealPixel(2, b_c, b_m, b_y, 0, 255, &back_pixel);
            if ( patTopHyt > 0 ) for ( i=n=0; i<c_steps && n < t_steps; i++ )
            {
                /* increment the foreground color */
                c += c_inc;
                m += m_inc;
                y += y_inc;
                AxGetRealPixel(2, (int)c, (int)m, (int)y, 0, 255, &fore_pixel);

                /* if there is no color change, skip it */
                if ( fore_pixel == back_pixel )
                    continue;

                /* do the bg/fg mix - 100%,87%,75%,62%,50%,37%,25%,12% */
                XSetState(Dpy, Gc1, fore_pixel, back_pixel, GXcopy, AllPlanes);

                for( j=0; j<N_GRAD_PATS; j++ )
                {
                    ttop    = y1+n*N_GRAD_PATS*patTopHyt + j*patTopHyt;
                    tbot    = y1+n*N_GRAD_PATS*patTopHyt + (j+1)*patTopHyt +2;
                    pts[0].x = pts[4].x = pts[3].x = left;
                    pts[0].y = pts[4].y = pts[1].y = ttop;
                    pts[1].x = pts[2].x = right;
                    pts[2].y = pts[3].y = tbot;
                    rotate_around_center(&pts[0], &center, sinAng,cosAng);
                    rotate_around_center(&pts[1], &center, sinAng,cosAng);
                    rotate_around_center(&pts[2], &center, sinAng,cosAng);
                    rotate_around_center(&pts[3], &center, sinAng,cosAng);
                    rotate_around_center(&pts[4], &center, sinAng,cosAng);
                    get_gradient_map(&pattern, j, FALSE);
                    XSetStipple(Dpy, Gc1, pattern);
                    XFillPolygon(Dpy, xwin, Gc1, pts, 5, Complex, CoordModeOrigin);
                }

                /* set the background color to the previous foreground */
                back_pixel = fore_pixel;
                n++;
            }
            else
            {
                /* fill with 100% background */
                XSetState(Dpy, Gc1, fore_pixel, back_pixel, GXcopy, AllPlanes);
                XSetFillStyle (Dpy, Gc1, FillSolid);
                pts[0].x = pts[4].x = pts[3].x = left;
                pts[0].y = pts[4].y = pts[1].y = y1;
                pts[1].x = pts[2].x = right;
                pts[2].y = pts[3].y = y1+thyt;
                rotate_around_center(&pts[0], &center, sinAng,cosAng);
                rotate_around_center(&pts[1], &center, sinAng,cosAng);
                rotate_around_center(&pts[2], &center, sinAng,cosAng);
                rotate_around_center(&pts[3], &center, sinAng,cosAng);
                rotate_around_center(&pts[4], &center, sinAng,cosAng);
                XFillPolygon(Dpy, xwin, Gc1, pts, 5, Complex, CoordModeOrigin);

                ttop = y1;
                tbot = y1 + thyt;
            }
        }
    }

    /* draw the bottom half */
    if ( bhyt > 0 )
    {
        bhyt = hyt/2 - y_off;
        /* get the number of divisible sections */
        avail = MAX(1, (int)(bhyt/(N_GRAD_PATS+1)));

        /* get the maximum number of color steps */
        c_steps = ABS(f_c - b_c);
        c_steps = MAX(ABS(f_m - b_m), c_steps);
        c_steps = MAX(ABS(f_y - b_y), c_steps);
        c_steps = MIN(avail, c_steps);

        /* get the number of color steps FOR THIS DISPLAY */
        if ( c_steps > 0 )
        {
            c_inc = (double)(f_c - b_c) / (double)(c_steps);
            m_inc = (double)(f_m - b_m) / (double)(c_steps);
            y_inc = (double)(f_y - b_y) / (double)(c_steps);
            c = b_c;
            m = b_m;
            y = b_y;
            AxGetRealPixel(2, b_c, b_m, b_y, 0, 255, &back_pixel);
            t_steps = 1;
            for ( i=0; i<c_steps; i++ )
            {
                /* increment the foreground color */
                c += c_inc;
                m += m_inc;
                y += y_inc;
                AxGetRealPixel(2, (int)c, (int)m, (int)y, 0, 255, &fore_pixel);

                /* if there is no color change, widen patterns and try again */
                if ( fore_pixel != back_pixel )
                {
                    t_steps++;
                    back_pixel = fore_pixel;
                }
            }

            AxGetRealPixel(2, f_c, f_m, f_y, 0, 255, &fore_pixel);
            if ( fore_pixel != back_pixel )
                t_steps++;

            /* tile in the gradient patterns */
            patBotHyt = bhyt / (t_steps*N_GRAD_PATS+1);

            while ( patBotHyt == 0 && t_steps > 1 )
            {
                t_steps--;
                patBotHyt = bhyt / (t_steps*N_GRAD_PATS+1);
            }

            c_inc = (double)(f_c - b_c) / (double)(c_steps);
            m_inc = (double)(f_m - b_m) / (double)(c_steps);
            y_inc = (double)(f_y - b_y) / (double)(c_steps);
            c = b_c;
            m = b_m;
            y = b_y;
            left  = x1;
            right = x1 + wid;
            AxGetRealPixel(2, b_c, b_m, b_y, 0, 255, &back_pixel);
            if ( patBotHyt > 0 ) for ( i=n=0; i<c_steps && n < t_steps; i++ )
            {
                /* increment the foreground color */
                c += c_inc;
                m += m_inc;
                y += y_inc;
                AxGetRealPixel(2, (int)c, (int)m, (int)y, 0, 255, &fore_pixel);

                /* if there is no color change, skip it */
                if ( fore_pixel == back_pixel )
                    continue;

                /* do the bg/fg mix - 100%,87%,75%,62%,50%,37%,25%,12% */
                XSetState(Dpy, Gc1, fore_pixel, back_pixel, GXcopy, AllPlanes);

                for( j=0; j<N_GRAD_PATS; j++ )
                {
                    btop = y1+hyt - n*N_GRAD_PATS*patBotHyt - (j+1)*patBotHyt;
                    bbot = y1+hyt - n*N_GRAD_PATS*patBotHyt - (j)*patBotHyt +2;
                    pts[0].x = pts[4].x = pts[3].x = left;
                    pts[0].y = pts[4].y = pts[1].y = btop;
                    pts[1].x = pts[2].x = right;
                    pts[2].y = pts[3].y = bbot;
                    rotate_around_center(&pts[0], &center, sinAng,cosAng);
                    rotate_around_center(&pts[1], &center, sinAng,cosAng);
                    rotate_around_center(&pts[2], &center, sinAng,cosAng);
                    rotate_around_center(&pts[3], &center, sinAng,cosAng);
                    rotate_around_center(&pts[4], &center, sinAng,cosAng);
                    get_gradient_map(&pattern, j, FALSE);
                    XSetStipple(Dpy, Gc1, pattern);
                    XFillPolygon(Dpy, xwin, Gc1, pts, 5, Complex, CoordModeOrigin);
                }

                /* set the background color to the previous foreground */
                back_pixel = fore_pixel;
                n++;
            }
            else
            {
                /* fill with 50% background */
                XSetState(Dpy, Gc1, back_pixel, fore_pixel, GXcopy, AllPlanes);
                XSetFillStyle (Dpy, Gc1, FillSolid);
                pts[0].x = pts[4].x = pts[3].x = left;
                pts[0].y = pts[4].y = pts[1].y = y1+thyt;
                pts[1].x = pts[2].x = right;
                pts[2].y = pts[3].y = y1+hyt;
                rotate_around_center(&pts[0], &center, sinAng,cosAng);
                rotate_around_center(&pts[1], &center, sinAng,cosAng);
                rotate_around_center(&pts[2], &center, sinAng,cosAng);
                rotate_around_center(&pts[3], &center, sinAng,cosAng);
                rotate_around_center(&pts[4], &center, sinAng,cosAng);
                XFillPolygon(Dpy, xwin, Gc1, pts, 5, Complex, CoordModeOrigin);
                btop = y1 + thyt;
            }
        }
    }

    /* fill remainder with pure forground */
    AxGetRealPixel(2, f_c, f_m, f_y, 0, 255, &fore_pixel);
    XSetState(Dpy, Gc1, fore_pixel, back_pixel, GXcopy, AllPlanes);
    XSetFillStyle (Dpy, Gc1, FillSolid);
    pts[0].x = pts[4].x = pts[3].x = left;
    pts[0].y = pts[4].y = pts[1].y = ttop;
    pts[1].x = pts[2].x = right;
    pts[2].y = pts[3].y = btop;
    rotate_around_center(&pts[0], &center, sinAng,cosAng);
    rotate_around_center(&pts[1], &center, sinAng,cosAng);
    rotate_around_center(&pts[2], &center, sinAng,cosAng);
    rotate_around_center(&pts[3], &center, sinAng,cosAng);
    rotate_around_center(&pts[4], &center, sinAng,cosAng);
    XFillPolygon(Dpy, xwin, Gc1, pts, 5, Complex, CoordModeOrigin);
#endif    /* } AX_X11 */
}

/****************************************************************************/
do_mask_linear_gradient(xwin, x1, y1, wid, hyt, attr, for_line)
Drawable xwin;
int x1, y1, wid, hyt;
attr_type *attr;
int for_line;
{
    double angle, sinAng, cosAng;
    int left, right, top, bottom;
    int i, j, patHyt, inv, y_off;
    XPoint pts[5], center;
    int c, m, y, k, ix;
    Pixmap pattern;
    unsigned long pel;
    int orig_x, orig_y, orig_wid, orig_hyt;

    orig_x = x1;
    orig_y = y1;
    orig_wid = wid;
    orig_hyt = hyt;

    /* rotate the coordinates */
    y_off = attr->f_yoff/100.0 * hyt;
    angle  = (double)attr->f_ang / 180.0 * M_PI;
    sinAng = (double)sin(angle);
    cosAng = (double)cos(angle);
    center.x = x1 + wid/2;
    center.y = y1 + hyt/2;
    pts[0].x = pts[2].x = x1;
    pts[0].y = pts[1].y = y1;
    pts[1].x = pts[3].x = x1 + wid;
    pts[2].y = pts[3].y = y1 + hyt;
    rotate_around_center(&pts[0], &center, sinAng, cosAng);
    rotate_around_center(&pts[1], &center, sinAng, cosAng);
    rotate_around_center(&pts[2], &center, sinAng, cosAng);
    rotate_around_center(&pts[3], &center, sinAng, cosAng);

    /* recalc the size and position */
    x1  = MIN( MIN( MIN( MIN(pts[0].x, pts[1].x), pts[2].x), pts[3].x), orig_x);
    y1  = MIN( MIN( MIN( MIN(pts[0].y, pts[1].y), pts[2].y), pts[3].y), orig_y);
    wid = MAX( MAX( MAX(pts[0].x, pts[1].x), pts[2].x ), pts[3].x ) - x1;
    hyt = MAX( MAX( MAX(pts[0].y, pts[1].y), pts[2].y ), pts[3].y ) - y1;
    center.x = x1 + wid/2;
    center.y = y1 + hyt/2;

    /* set the color */
    if ( for_line )
    {
        inv = attr->lfg.see_thru ? TRUE : FALSE;
        if ( inv )
        {
            c = attr->lbg.c;
            m = attr->lbg.m;
            y = attr->lbg.y;
            k = attr->lbg.k;
        }
        else
        {
            c = attr->lfg.c;
            m = attr->lfg.m;
            y = attr->lfg.y;
            k = attr->lfg.k;
        }
    }
    else
    {
        inv = attr->ffg.see_thru ? TRUE : FALSE;
        if ( inv )
        {
            c = attr->fbg.c;
            m = attr->fbg.m;
            y = attr->fbg.y;
            k = attr->fbg.k;
        }
        else
        {
            c = attr->ffg.c;
            m = attr->ffg.m;
            y = attr->ffg.y;
            k = attr->ffg.k;
        }
    }
    AxGetRealPixel(2, c,m,y,k, 255, &pel);
#if AXMS    /* { */
    /* NT only!  Windows 95 functionality is done at the AxFill level */
    win_mask_linear_gradient(xwin, Gc1, x1, y1, wid, hyt, orig_x, orig_y,
                            orig_wid, orig_hyt, pel, inv, sinAng, cosAng,
                            y_off);
#else    /* } { */
    XSetForeground(Dpy, Gc1, pel);
    XSetFunction(Dpy, Gc1, GXcopy);

    /* draw the gradient thru the mask */
    patHyt = hyt/(N_GRAD_PATS + 1);
    if ( y_off >= 0 )
        y1 = y1 + y_off;
    else
        hyt =  hyt - (-1)*y_off;
    left  = x1;
    right = x1 + wid;
    for( j=0; j<N_GRAD_PATS; j++ )
    {
        top    = inv ? y1+hyt - (j+2)*patHyt : y1 + j*patHyt;
        bottom = inv ? y1+hyt - (j+1)*patHyt : y1 + (j+1)*patHyt;
        pts[0].x = pts[4].x = pts[3].x = left;
        pts[0].y = pts[4].y = pts[1].y = top;
        pts[1].x = pts[2].x = right;
        pts[2].y = pts[3].y = bottom;
        rotate_around_center(&pts[0], &center, sinAng,cosAng);
        rotate_around_center(&pts[1], &center, sinAng,cosAng);
        rotate_around_center(&pts[2], &center, sinAng,cosAng);
        rotate_around_center(&pts[3], &center, sinAng,cosAng);
        rotate_around_center(&pts[4], &center, sinAng,cosAng);
        get_gradient_map(&pattern, j, FALSE);
        XSetStipple(Dpy, Gc1, pattern);
        XFillPolygon(Dpy, xwin, Gc1, pts, 5, Complex, CoordModeOrigin);
    }

    /* fill remainder with pure forground */
    if ( y_off >= 0 )
    {
        XSetFillStyle (Dpy, Gc1, FillSolid);
        top = inv ? y1 : y1 + N_GRAD_PATS*patHyt;
        bottom = inv ? y1 + hyt - N_GRAD_PATS*patHyt : y1 + hyt + y_off;
        pts[0].x = pts[4].x = pts[3].x = left;
        pts[0].y = pts[4].y = pts[1].y = top;
        pts[1].x = pts[2].x = right;
        pts[2].y = pts[3].y = bottom;
        rotate_around_center(&pts[0], &center, sinAng,cosAng);
        rotate_around_center(&pts[1], &center, sinAng,cosAng);
        rotate_around_center(&pts[2], &center, sinAng,cosAng);
        rotate_around_center(&pts[3], &center, sinAng,cosAng);
        rotate_around_center(&pts[4], &center, sinAng,cosAng);
        XFillPolygon(Dpy, xwin, Gc1, pts, 5, Complex, CoordModeOrigin);
    }
    else if ( inv == FALSE )
    {
        top    = y1 + y_off + hyt;
        bottom = y1 + hyt + (-1)*y_off;
        pts[0].x = pts[4].x = pts[3].x = left;
        pts[0].y = pts[4].y = pts[1].y = top;
        pts[1].x = pts[2].x = right;
        pts[2].y = pts[3].y = bottom;
        rotate_around_center(&pts[0], &center, sinAng,cosAng);
        rotate_around_center(&pts[1], &center, sinAng,cosAng);
        rotate_around_center(&pts[2], &center, sinAng,cosAng);
        rotate_around_center(&pts[3], &center, sinAng,cosAng);
        rotate_around_center(&pts[4], &center, sinAng,cosAng);
        XSetFillStyle (Dpy, Gc1, FillSolid);
        XFillPolygon(Dpy, xwin, Gc1, pts, 5, Complex, CoordModeOrigin);
    }
#endif    /* } AXMS */
}

/****************************************************************************/
do_tile_linear_gradient(xwin, x1, y1, wid, hyt, attr, for_line)
Drawable xwin;
int x1, y1, wid, hyt;
attr_type *attr;
int for_line;
{
    double angle, sinAng, cosAng, c,m,y, c_inc, m_inc, y_inc, t_steps;
    int c_steps, i, j, n, patHyt, f_c,f_y,f_m,f_k, b_c,b_m,b_y,b_k, avail;
    int left, right, top, bottom, y_off, skip_count, pr_cnt, s_off;
    unsigned long back_pixel, fore_pixel;
    XPoint pts[5], center;
    Pixmap pattern;
    int orig_x, orig_y, orig_wid, orig_hyt;

    orig_x = x1;
    orig_y = y1;
    orig_wid = wid;
    orig_hyt = hyt;

    /* rotate the coordinates */
    y_off = attr->f_yoff/100.0 * hyt;
    angle  = (double)attr->f_ang / 180.0 * M_PI;
    sinAng = (double)sin(angle);
    cosAng = (double)cos(angle);
    center.x = x1 + wid/2;
    center.y = y1 + hyt/2;
    pts[0].x = pts[2].x = x1;
    pts[0].y = pts[1].y = y1;
    pts[1].x = pts[3].x = x1 + wid;
    pts[2].y = pts[3].y = y1 + hyt;
    rotate_around_center(&pts[0], &center, sinAng, cosAng);
    rotate_around_center(&pts[1], &center, sinAng, cosAng);
    rotate_around_center(&pts[2], &center, sinAng, cosAng);
    rotate_around_center(&pts[3], &center, sinAng, cosAng);

    /* recalc the size and position */
    x1  = MIN( MIN( MIN(pts[0].x, pts[1].x), pts[2].x), pts[3].x);
    y1  = MIN( MIN( MIN(pts[0].y, pts[1].y), pts[2].y), pts[3].y);
    wid = MAX(MAX(MAX(pts[0].x,pts[1].x),pts[2].x),pts[3].x);
    hyt = MAX(MAX(MAX(pts[0].y, pts[1].y),pts[2].y),pts[3].y);
    wid = wid - x1;
    hyt = hyt - y1;

    /* remove black component from color step calcs */
    if ( for_line )
    {
        f_c = MIN(attr->lfg.c + attr->lfg.k, 255);
        f_m = MIN(attr->lfg.m + attr->lfg.k, 255);
        f_y = MIN(attr->lfg.y + attr->lfg.k, 255);
        f_k = 0;
        b_c = MIN(attr->lbg.c + attr->lbg.k, 255);
        b_m = MIN(attr->lbg.m + attr->lbg.k, 255);
        b_y = MIN(attr->lbg.y + attr->lbg.k, 255);
        b_k = 0;
    }
    else
    {
        f_c = MIN(attr->ffg.c + attr->ffg.k, 255);
        f_m = MIN(attr->ffg.m + attr->ffg.k, 255);
        f_y = MIN(attr->ffg.y + attr->ffg.k, 255);
        f_k = 0;
        b_c = MIN(attr->fbg.c + attr->fbg.k, 255);
        b_m = MIN(attr->fbg.m + attr->fbg.k, 255);
        b_y = MIN(attr->fbg.y + attr->fbg.k, 255);
        b_k = 0;
    }

    /* get the number of divisible sections */
    avail = MAX(1, (int)(hyt/(N_GRAD_PATS+1)));

    /* get the maximum number of color steps */
    c_steps = ABS(f_c - b_c);
    c_steps = MAX(ABS(f_m - b_m), c_steps);
    c_steps = MAX(ABS(f_y - b_y), c_steps);
    c_steps = MIN(avail, c_steps);
    if ( c_steps == 0 )
    {
        patHyt = 0;
        left  = x1;
        right = x1 + wid;
    }
    else
    {
        t_steps = c_steps;
        /* tile in the gradient patterns */
        patHyt = hyt / (c_steps*N_GRAD_PATS+1);
        if ( patHyt <= 0 )
            patHyt = 1;
        left  = x1;
        right = x1 + wid;
        skip_count = 0;
        pr_cnt = 0;
        if ( y_off >= 0 )
            y1  = y1 + y_off;
        else
            hyt =  hyt - (-1)*y_off;
        t_steps = N_GRAD_PATS*c_steps;
        c_inc = (double)(f_c - b_c) / (double)(c_steps);
        m_inc = (double)(f_m - b_m) / (double)(c_steps);
        y_inc = (double)(f_y - b_y) / (double)(c_steps);
        c = b_c;
        m = b_m;
        y = b_y;
    }   
    AxGetRealPixel(2, b_c, b_m, b_y, 0, 255, &back_pixel);
    AxGetRealPixel(2, f_c, f_m, f_y, 0, 255, &fore_pixel);
#if AXMS    /* { */
    /* NT only!  Windows 95 functionality is done at the AxFill level */    
    win_linear_gradient(xwin, Gc1, x1, y1, wid, hyt, c_steps,
                        t_steps, c_inc, m_inc, y_inc, c, m, y,
                        back_pixel, orig_x, orig_y, orig_wid, orig_hyt,
                        sinAng, cosAng, y_off);
#else    /* } { */

    for ( i=n=0; i<c_steps; i++ )
    {
        /* increment the foreground color */
        c += c_inc;
        m += m_inc;
        y += y_inc;
        AxGetRealPixel(2, (int)c, (int)m, (int)y, 0, 255, &fore_pixel);

        /* if there is no color change, skip it */
        if ( fore_pixel == back_pixel )
        {
            skip_count += N_GRAD_PATS;
            continue;
        }

        /* do the bg/fg mix - 100%,87%,75%,62%,50%,37%,25%,12% */
        XSetState(Dpy, Gc1, fore_pixel, back_pixel, GXcopy, AllPlanes);
        skip_count += N_GRAD_PATS;
        s_off = (skip_count - pr_cnt)/N_GRAD_PATS;
        for ( j=0; j<N_GRAD_PATS; j++ )
        {
            top    = y1 + (s_off*j + pr_cnt)*hyt/t_steps;
            bottom = y1 + (s_off*(j+1) + pr_cnt)*(double)hyt/t_steps; 
            pts[0].x = pts[4].x = pts[3].x = left;
            pts[0].y = pts[4].y = pts[1].y = top;
            pts[1].x = pts[2].x = right;
            pts[2].y = pts[3].y = bottom;
            rotate_around_center(&pts[0], &center, sinAng, cosAng);
            rotate_around_center(&pts[1], &center, sinAng, cosAng);
            rotate_around_center(&pts[2], &center, sinAng, cosAng);
            rotate_around_center(&pts[3], &center, sinAng, cosAng);
            rotate_around_center(&pts[4], &center, sinAng, cosAng);
            get_gradient_map(&pattern, j, FALSE);
            XSetStipple(Dpy, Gc1, pattern);
            XFillPolygon(Dpy,xwin,Gc1, pts,5,Complex,CoordModeOrigin);
        }

        /* set the background color to the previous foreground */
        back_pixel = fore_pixel;
        pr_cnt = skip_count;
    }

    /* fill remainder with pure forground */
    if ( y_off >= 0 )
    {
        AxGetRealPixel(2, b_c, b_m, b_y, 0, 255, &back_pixel);
        XSetState(Dpy, Gc1, fore_pixel, back_pixel, GXcopy, AllPlanes);
        XSetFillStyle (Dpy, Gc1, FillSolid);
        if ( patHyt <= 0 )
            top = y1;
        else
            top = bottom;
        bottom = y1 + hyt;
        pts[0].x = pts[4].x = pts[3].x = left;
        pts[0].y = pts[4].y = pts[1].y = top;
        pts[1].x = pts[2].x = right;
        pts[2].y = pts[3].y = bottom;
        rotate_around_center(&pts[0], &center, sinAng,cosAng);
        rotate_around_center(&pts[1], &center, sinAng,cosAng);
        rotate_around_center(&pts[2], &center, sinAng,cosAng);
        rotate_around_center(&pts[3], &center, sinAng,cosAng);
        rotate_around_center(&pts[4], &center, sinAng,cosAng);
        XFillPolygon(Dpy, xwin, Gc1, pts, 5, Complex, CoordModeOrigin);


        if ( patHyt > 0 )
        {
            y1 = y1 - y_off;
            XSetState(Dpy, Gc1, back_pixel, fore_pixel, GXcopy, AllPlanes);
            XSetFillStyle (Dpy, Gc1, FillSolid);
            top = y1;
            bottom = y1 + y_off;
            pts[0].x = pts[4].x = pts[3].x = left;
            pts[0].y = pts[4].y = pts[1].y = top;
            pts[1].x = pts[2].x = right;
            pts[2].y = pts[3].y = bottom;
            rotate_around_center(&pts[0], &center, sinAng,cosAng);
            rotate_around_center(&pts[1], &center, sinAng,cosAng);
            rotate_around_center(&pts[2], &center, sinAng,cosAng);
            rotate_around_center(&pts[3], &center, sinAng,cosAng);
            rotate_around_center(&pts[4], &center, sinAng,cosAng);
            XFillPolygon(Dpy, xwin, Gc1, pts, 5, Complex, CoordModeOrigin);
        }
    }
    else
    {
        AxGetRealPixel(2, b_c, b_m, b_y, 0, 255, &back_pixel);
        XSetState(Dpy, Gc1, fore_pixel, back_pixel, GXcopy, AllPlanes);
        XSetFillStyle (Dpy, Gc1, FillSolid);
        if ( patHyt <= 0 )
            top = y1;
        else
            top = bottom;
        bottom = y1 + hyt;
        pts[0].x = pts[4].x = pts[3].x = left;
        pts[0].y = pts[4].y = pts[1].y = top;
        pts[1].x = pts[2].x = right;
        pts[2].y = pts[3].y = bottom;
        rotate_around_center(&pts[0], &center, sinAng,cosAng);
        rotate_around_center(&pts[1], &center, sinAng,cosAng);
        rotate_around_center(&pts[2], &center, sinAng,cosAng);
        rotate_around_center(&pts[3], &center, sinAng,cosAng);
        rotate_around_center(&pts[4], &center, sinAng,cosAng);
        XFillPolygon(Dpy, xwin, Gc1, pts, 5, Complex, CoordModeOrigin);
        AxGetRealPixel(2, b_c, b_m, b_y, 0, 255, &back_pixel);
        XSetState(Dpy, Gc1, fore_pixel, back_pixel, GXcopy, AllPlanes);
        XSetFillStyle (Dpy, Gc1, FillSolid);

        if ( patHyt > 0 )
        {
            top = y1 + hyt;
            bottom = y1 + hyt + (-1)*y_off;
            pts[0].x = pts[4].x = pts[3].x = left;
            pts[0].y = pts[4].y = pts[1].y = top;
            pts[1].x = pts[2].x = right;
            pts[2].y = pts[3].y = bottom;
            rotate_around_center(&pts[0], &center, sinAng,cosAng);
            rotate_around_center(&pts[1], &center, sinAng,cosAng);
            rotate_around_center(&pts[2], &center, sinAng,cosAng);
            rotate_around_center(&pts[3], &center, sinAng,cosAng);
            rotate_around_center(&pts[4], &center, sinAng,cosAng);
            XFillPolygon(Dpy, xwin, Gc1, pts, 5, Complex, CoordModeOrigin);
        }
    }
#endif    /* } !AXMS */
}

/****************************************************************************/
do_tile_rect_gradient(xwin, x1, y1, wid, hyt, attr, for_line)
Drawable xwin;
int x1, y1, wid, hyt;
attr_type *attr;
int for_line;
{
    double angle, sinAng, cosAng, c,m,y, c_inc, m_inc, y_inc, tmp_wid, patWid;
    int c_steps, step, i, j, patHyt, tmpPatWid, tmpPatHyt;
    int f_c,f_y,f_m,f_k, b_c,b_m,b_y,b_k;
    double x_off, y_off, t_steps;
    XPoint pts[5], src_pts[2], cur_pts[5], center;
    unsigned long back_pixel, fore_pixel;
    int wid1, wid2, hyt1, hyt2, skip_count, pr_cnt, s_off;
    int orig_x, orig_y, orig_wid, orig_hyt;
    Pixmap pattern;

    orig_x = x1;
    orig_y = y1;
    orig_wid = wid;
    orig_hyt = hyt;
    /* rotate the coordinates */
    x_off  = attr->f_xoff/200.0 * wid;
    y_off  = attr->f_yoff/200.0 * hyt;
    angle  = (double)attr->f_ang / 180.0 * M_PI;
    sinAng = (double)sin(angle);
    cosAng = (double)cos(angle);
    center.x = x1 + x_off + wid/2;
    center.y = y1 + y_off + hyt/2;
    pts[0].x = pts[2].x = src_pts[0].x = x1;
    pts[0].y = pts[1].y = src_pts[0].y = y1;
    pts[1].x = pts[3].x = src_pts[1].x = x1 + wid;
    pts[2].y = pts[3].y = src_pts[1].y = y1 + hyt;
    rotate_around_center(&pts[0], &center, sinAng, cosAng);
    rotate_around_center(&pts[1], &center, sinAng, cosAng);
    rotate_around_center(&pts[2], &center, sinAng, cosAng);
    rotate_around_center(&pts[3], &center, sinAng, cosAng);

    /* recalc the size and position */
    x1  = MIN( MIN( MIN( MIN(pts[0].x, pts[1].x), pts[2].x), pts[3].x), orig_x);
    y1  = MIN( MIN( MIN( MIN(pts[0].y, pts[1].y), pts[2].y), pts[3].y), orig_y);
    wid = MAX(MAX(MAX(MAX(pts[0].x,pts[1].x),pts[2].x),pts[3].x),src_pts[1].x);
    hyt = MAX(MAX(MAX(MAX(pts[0].y, pts[1].y),pts[2].y),pts[3].y),src_pts[1].y);
    wid = wid - x1;
    hyt = hyt - y1;
    wid1 = center.x - x1;
    wid2 = x1 + wid - center.x;
    hyt1 = center.y - y1;
    hyt2 = y1 + hyt - center.y;

    /* remove black component from color step calcs */
    if ( for_line )
    {
        f_c = MIN(attr->lfg.c + attr->lfg.k, 255);
        f_m = MIN(attr->lfg.m + attr->lfg.k, 255);
        f_y = MIN(attr->lfg.y + attr->lfg.k, 255);
        f_k = 0;
        b_c = MIN(attr->lbg.c + attr->lbg.k, 255);
        b_m = MIN(attr->lbg.m + attr->lbg.k, 255);
        b_y = MIN(attr->lbg.y + attr->lbg.k, 255);
        b_k = 0;
    }
    else
    {
        f_c = MIN(attr->ffg.c + attr->ffg.k, 255);
        f_m = MIN(attr->ffg.m + attr->ffg.k, 255);
        f_y = MIN(attr->ffg.y + attr->ffg.k, 255);
        f_k = 0;
        b_c = MIN(attr->fbg.c + attr->fbg.k, 255);
        b_m = MIN(attr->fbg.m + attr->fbg.k, 255);
        b_y = MIN(attr->fbg.y + attr->fbg.k, 255);
        b_k = 0;
    }

    /* set the color steps */
    c_steps = ABS(f_c - b_c);
    c_steps = MAX(ABS(f_m - b_m), c_steps);
    c_steps = MAX(ABS(f_y - b_y), c_steps);
    while ( (patHyt = orig_hyt/(2*c_steps*N_GRAD_PATS+1)) <= 0 && c_steps )
        c_steps -= 1;
    if ( c_steps == 0 )
    {
        patHyt = orig_hyt / (N_GRAD_PATS+1);
        c_steps = 1;
    }

    /* get the number of color steps FOR THIS DISPLAY */
    if ( c_steps > 0 )
    {
        c_inc = (double)(f_c - b_c) / (double)(c_steps);
        m_inc = (double)(f_m - b_m) / (double)(c_steps);
        y_inc = (double)(f_y - b_y) / (double)(c_steps);
        c = b_c;
        m = b_m;
        y = b_y;
        AxGetRealPixel(2, b_c, b_m, b_y, 0, 255, &back_pixel);

    }
    c_inc = (double)(f_c - b_c) / (double)(c_steps);
    m_inc = (double)(f_m - b_m) / (double)(c_steps);
    y_inc = (double)(f_y - b_y) / (double)(c_steps);
    patWid = (double)((double)patHyt * (double)wid / (double)hyt);

    /* initialize the starting color */
    c = f_c;
    m = f_m;
    y = f_y;

#if AXMS    /* { */
    /* NT only!  Windows 95 functionality is done at the AxFill level */
    win_rect_gradient(xwin, Gc1, x1,y1, wid, hyt, x_off, y_off,
                            c_steps, c_inc, m_inc, y_inc, c, m, y, orig_x,
                            orig_y, orig_wid, orig_hyt, sinAng, cosAng);
#else    /* } { */

    AxGetRealPixel(2, (int)c, (int)m, (int)y, 0, 255, &fore_pixel);
    AxGetRealPixel(2, (int)b_c, (int)b_m, (int)b_y, 0, 255, &back_pixel);
    XSetState(Dpy, Gc1, fore_pixel, back_pixel, GXcopy, AllPlanes);
    XSetFillStyle (Dpy, Gc1, FillSolid);
    cur_pts[0].x = cur_pts[4].x = cur_pts[3].x = x1;
    cur_pts[0].y = cur_pts[4].y = cur_pts[1].y = y1;
    cur_pts[1].x = cur_pts[2].x = x1 + wid;
    cur_pts[2].y = cur_pts[3].y = y1 + hyt;
    XFillPolygon(Dpy, xwin, Gc1, cur_pts, 5, Complex, CoordModeOrigin);

    c = f_c;
    m = f_m;
    y = f_y;
    pts[0].x = x1;
    pts[0].y = y1;
    pts[1].x = x1 + wid;
    pts[1].y = y1 + hyt;
    tmpPatHyt = patHyt;
    skip_count = 0;
    pr_cnt = 0;
    t_steps = N_GRAD_PATS*c_steps;

    AxGetRealPixel(2, (int)c, (int)m, (int)y, 0, 255, &fore_pixel);

    /* tile in the gradient patterns */
    XSetFillStyle (Dpy, Gc1, FillOpaqueStippled);
    for ( i=0, step=0; i<c_steps; i++ )
    {
        /* increment the foreground color */
        c -= c_inc;
        m -= m_inc;
        y -= y_inc;
        AxGetRealPixel(2, (int)c, (int)m, (int)y, 0, 255, &back_pixel);

        /* if there is no color change, widen patterns and try again */
        if ( fore_pixel == back_pixel )
        {
            skip_count += N_GRAD_PATS;
            continue;
        }
        /* do the bg/fg mix - 100%,87%,75%,62%,50%,37%,25%,12% */
        XSetState(Dpy, Gc1, fore_pixel, back_pixel, GXcopy, AllPlanes);
        skip_count += N_GRAD_PATS;
        s_off = (skip_count - pr_cnt)/N_GRAD_PATS;
        for ( j=0; j<N_GRAD_PATS; j++ )
        {
            cur_pts[0].x = pts[0].x + (s_off*j + pr_cnt)*(double)wid1/t_steps;
            cur_pts[0].y = pts[0].y + (s_off*j + pr_cnt)*(double)hyt1/t_steps;
            cur_pts[1].x = pts[1].x - (s_off*j + pr_cnt)*(double)wid2/t_steps;
            cur_pts[1].y = pts[0].y + (s_off*j + pr_cnt)*(double)hyt1/t_steps;
            cur_pts[2].x = pts[1].x - (s_off*j + pr_cnt)*(double)wid2/t_steps;
            cur_pts[2].y = pts[1].y - (s_off*j + pr_cnt)*(double)hyt2/t_steps;
            cur_pts[3].x = pts[0].x + (s_off*j + pr_cnt)*(double)wid1/t_steps;
            cur_pts[3].y = pts[1].y - (s_off*j + pr_cnt)*(double)hyt2/t_steps;
            cur_pts[4].x = pts[0].x + (s_off*j + pr_cnt)*(double)wid1/t_steps;
            cur_pts[4].y = pts[0].y + (s_off*j + pr_cnt)*(double)hyt1/t_steps;

            rotate_around_center(&cur_pts[0], &center, sinAng, cosAng);
            rotate_around_center(&cur_pts[1], &center, sinAng, cosAng);
            rotate_around_center(&cur_pts[2], &center, sinAng, cosAng);
            rotate_around_center(&cur_pts[3], &center, sinAng, cosAng);
            rotate_around_center(&cur_pts[4], &center, sinAng, cosAng);

            get_gradient_map(&pattern, j, TRUE);
            XSetStipple(Dpy, Gc1, pattern);
            XFillPolygon(Dpy, xwin, Gc1, cur_pts, 5, Complex, CoordModeOrigin); 
        }

        /* set the background color to the previous foreground */
        fore_pixel = back_pixel;
        pr_cnt = skip_count;
    }
#endif    /* } AXMS */
}

/****************************************************************************/
do_tile_radial_gradient(xwin, x1, y1, wid, hyt, attr, for_line)
Drawable xwin;
int x1, y1, wid, hyt;
attr_type *attr;
int for_line;
{
    double c,m,y, c_inc,m_inc,y_inc, x_pos,y_pos, x_size,y_size, cx,cy;
    int center, c_steps, i, j, n, tint_steps;
    int f_c,f_y,f_m,f_k, b_c,b_m,b_y,b_k, avail;
    double x_off, y_off, dx, dy, t_steps;
    unsigned long back_pixel, fore_pixel;
    int skip_count, pr_cnt, s_off;
    Pixmap pattern;
#if AXMS
    int orig_x = x1, orig_y = y1;
#endif

    /* no angle info */
    /* find the center of the object */
    x_off = attr->f_xoff/100.0 * wid;
    y_off = attr->f_yoff/100.0 * hyt;

    /* set the position and size of the field */
    cx = x_off + x1 + wid/2;
    cy = y_off + y1 + hyt/2;
    wid = wid + ABS(x_off);
    hyt = hyt + ABS(y_off);
    x_size = y_size = hypot((double)wid, (double)hyt);
    x_pos = cx - x_size / 2;
    y_pos = cy - y_size / 2;

    /* remove black component from color step calcs */
    if ( for_line )
    {
        f_c = MIN(attr->lfg.c + attr->lfg.k, 255);
        f_m = MIN(attr->lfg.m + attr->lfg.k, 255);
        f_y = MIN(attr->lfg.y + attr->lfg.k, 255);
        f_k = 0;
        b_c = MIN(attr->lbg.c + attr->lbg.k, 255);
        b_m = MIN(attr->lbg.m + attr->lbg.k, 255);
        b_y = MIN(attr->lbg.y + attr->lbg.k, 255);
        b_k = 0;
    }
    else
    {
        f_c = MIN(attr->ffg.c + attr->ffg.k, 255);
        f_m = MIN(attr->ffg.m + attr->ffg.k, 255);
        f_y = MIN(attr->ffg.y + attr->ffg.k, 255);
        f_k = 0;
        b_c = MIN(attr->fbg.c + attr->fbg.k, 255);
        b_m = MIN(attr->fbg.m + attr->fbg.k, 255);
        b_y = MIN(attr->fbg.y + attr->fbg.k, 255);
        b_k = 0;
    }

    /* get the number of divisible sections */
    avail = MAX(1, (int)(y_size/(2*N_GRAD_PATS+1)));

    /* get the maximum number of color steps */
    c_steps = ABS(f_c - b_c);
    c_steps = MAX(ABS(f_m - b_m), c_steps);
    c_steps = MAX(ABS(f_y - b_y), c_steps);
    c_steps = MIN(avail, c_steps);

    /* tile in the gradient patterns */
#if AXMS    /* { */
    /* NT only!  Windows 95 functionality is done at the AxFill level */
    win_radial_gradient(xwin, Gc1, orig_x, orig_y, wid, hyt, x_off, y_off,
                                c_steps, f_c, f_m, f_y, b_c, b_m, b_y); 
#else    /* } { */

    AxGetRealPixel(2, f_c, f_m, f_y, 0, 255, &fore_pixel);
    XSetState(Dpy, Gc1, fore_pixel, fore_pixel, GXcopy, AllPlanes);
    XSetFillStyle (Dpy, Gc1, FillSolid);
    XFillRectangle(Dpy, xwin, Gc1, x1, y1, wid, hyt);

    /* tile in the gradient patterns */
    XSetFillStyle (Dpy, Gc1, FillOpaqueStippled);
    x1  = x_pos;
    y1  = y_pos;
    wid = x_size;
    hyt = y_size;
    c_inc = (double)(f_c - b_c) / (double)(c_steps);
    m_inc = (double)(f_m - b_m) / (double)(c_steps);
    y_inc = (double)(f_y - b_y) / (double)(c_steps);
    c = f_c;
    m = f_m;
    y = f_y;
    AxGetRealPixel(2, f_c, f_m, f_y, 0, 255, &fore_pixel);
    skip_count = 0;
    pr_cnt = 0;
    t_steps = 2*N_GRAD_PATS*c_steps;
    for ( i=0; i<c_steps; i++ )
    {
        /* increment the foreground color */
        c -= c_inc;
        m -= m_inc;
        y -= y_inc;
        AxGetRealPixel(2, (int)c, (int)m, (int)y, 0, 255, &back_pixel);

        /* if there is no color change, skip it */
        if ( fore_pixel == back_pixel )
        {
            skip_count += N_GRAD_PATS;
            continue;
        }

        /* do the bg/fg mix - 100%,87%,75%,62%,50%,37%,25%,12% */
        XSetState(Dpy, Gc1, fore_pixel, back_pixel, GXcopy, AllPlanes);
        skip_count += N_GRAD_PATS;
        s_off = (skip_count - pr_cnt)/N_GRAD_PATS;
        for ( j=0; j<N_GRAD_PATS; j++ )
        {
            get_gradient_map(&pattern, j, TRUE);
            XSetStipple(Dpy, Gc1, pattern);
            x1  = x_pos + (s_off*j + pr_cnt)*y_size/t_steps;
            y1  = y_pos + (s_off*j + pr_cnt)*y_size/t_steps;;
            wid = x_size - (s_off*j+pr_cnt)*2.0*y_size/t_steps;
            hyt = y_size - (s_off*j+pr_cnt)*2.0*y_size/t_steps;
            XFillArc(Dpy, xwin, Gc1, x1, y1, wid, hyt, 0*64, 360*64);
        }

        /* set the background color to the previous foreground */
        fore_pixel = back_pixel;
        pr_cnt = skip_count;
    }
#endif    /* } AXMS */
}

/********************************************************************
* doAttrFill - 
*/
doAttrFill(xwin, x, y, wid, hyt, attr, for_line)
Drawable xwin;
int x, y, wid, hyt;
attr_type *attr;
int for_line;
{
    int type, ddep, is_pic;
    color_type *fg, *bg;

    if ( for_line )
    {
        type = attr->l_type;
        ddep = attr->l_ddep;
        is_pic = attr->l_pic;

        fg = &attr->lfg;
        bg = &attr->lbg;
    }
    else
    {
        type = attr->f_type;
        ddep = attr->f_ddep;
        is_pic = attr->f_pic;

        fg = &attr->ffg;
        bg = &attr->fbg;
        if ( wid <= 0 || hyt <= 0 )
        	is_pic = FALSE;
    }

    switch ( type )
    {
		/* 0/1 bitmap fill */
    	case BUILTIN_FILL_TYPE: /* 0 */
    	case USERDEF_FILL_TYPE: /* 1 */
			if ( is_pic ) /* display as picture */
			{
				do_color_picture_pattern(xwin, x, y, wid, hyt, attr, for_line);
			}
			else if ( ddep == 8 ) /* color pattern */
			{
				XSetFillStyle (Dpy, Gc1, FillTiled);
				do_color_tile_pattern(xwin, x, y, wid, hyt, attr, for_line);
			}
			/* draw foreground and background color */
			else if ( bg->see_thru == FALSE && fg->see_thru == FALSE )
			{
				XSetFillStyle (Dpy, Gc1, FillOpaqueStippled);
				do_tile_pattern(xwin, x, y, wid, hyt, attr, for_line);
			}

			/* bg or fg color must be transparent - draw one color only */
			else if ( bg->see_thru == FALSE || fg->see_thru == FALSE )
			{
				XSetFillStyle (Dpy, Gc1, FillStippled);
				do_mask_pattern(xwin, x, y, wid, hyt, attr, for_line);
			}
		break;
    	case LINEAR_FILL_TYPE: /* 2 - linear gradient */
			/* draw foreground and background color */
			if ( bg->see_thru == FALSE && fg->see_thru == FALSE )
			{
				XSetFillStyle (Dpy, Gc1, FillOpaqueStippled);
				do_tile_linear_gradient(xwin, x, y, wid, hyt, attr, for_line);
			}

			/* bg or fg color must be transparent - draw one color only */
			else if ( bg->see_thru == FALSE || fg->see_thru == FALSE )
			{
				XSetFillStyle (Dpy, Gc1, FillStippled);
				do_mask_linear_gradient(xwin, x, y, wid, hyt, attr, for_line);
			}
		break;
		case RADIAL_FILL_TYPE: /* 3 - radial gradient */
			/* draw foreground and background color */
			if ( bg->see_thru == FALSE && fg->see_thru == FALSE )
			{
				XSetFillStyle (Dpy, Gc1, FillOpaqueStippled);
				do_tile_radial_gradient(xwin, x, y, wid, hyt, attr, for_line);
			}

			/* bg or fg color must be transparent - draw one color only */
			else if ( bg->see_thru == FALSE || fg->see_thru == FALSE )
			{
#if 0  
				/* was AXMS. save for later?   -donato */
				do_mask_radial_gradient(xwin, x,y, wid, hyt,attr,for_line);
#else
				XSetFillStyle (Dpy, Gc1, FillStippled);
				do_tile_radial_gradient(xwin, x, y, wid, hyt, attr, for_line);
#endif
			}
		break;
		case RECT_FILL_TYPE: /* 4 - rect gradient */
			/* draw foreground and background color */
			if ( bg->see_thru == FALSE && fg->see_thru == FALSE )
			{
				XSetFillStyle (Dpy, Gc1, FillOpaqueStippled);
				do_tile_rect_gradient(xwin, x, y, wid, hyt, attr, for_line);
			}

			/* bg or fg color must be transparent - draw one color only */
			else if ( bg->see_thru == FALSE || fg->see_thru == FALSE )
			{
#if 0
				/* was AXMS. save for later?   -donato */
				do_mask_rect_gradient(xwin, x, y, wid, hyt, attr, for_line);
#else
				XSetFillStyle (Dpy, Gc1, FillStippled);
				do_tile_rect_gradient(xwin, x, y, wid, hyt, attr, for_line);
#endif
			}
		break;
		case BILINEAR_FILL_TYPE: /* 6 - bilinear gradient */
			/* draw foreground and background color */
			if ( bg->see_thru == FALSE && fg->see_thru == FALSE )
			{
				XSetFillStyle (Dpy, Gc1, FillOpaqueStippled);
				do_tile_bilinear_gradient(xwin, x, y, wid, hyt, attr, for_line);
			}

			/* bg or fg color must be transparent - draw one color only */
			else if ( bg->see_thru == FALSE || fg->see_thru == FALSE )
			{
				XSetFillStyle (Dpy, Gc1, FillStippled);
				do_mask_bilinear_gradient(xwin, x, y, wid, hyt, attr, for_line);
			}
		break;
    }
}

int
check_for_fill(attr)
attr_type *attr;
{
    /* the following cases result in nothing being drawn */
    if ( attr->mode == TMS && attr->ffg.see_thru && attr->fbg.see_thru )
        return(FALSE);
    if ( attr->fill_flag == FALSE )
        return(FALSE);
    return(TRUE);
}

/********************************************************************
* create_clip_mask - create clip mask according to the clip path
*/
static Pixmap
create_clip_mask(xwin, clip, clip_info, x_off, y_off)
Drawable xwin;
TmRect32 *clip;
clip_path *clip_info;
int *x_off;
int *y_off;
{
	int sz, i, j, found, x, y, cminX, cminY, cmaxX, cmaxY;
	int x1, y1, x2, y2, x_size, y_size, wid, hyt, x_p, y_p;
    Pixmap temp_mask, clip_mask;
    unsigned long gc_mask;
	XRectangle rect[1];
#if AXMS
	POINT **cpoints;
	HRGN hrgn, final_hrgn;
#else
    XPoint **cpoints;
#endif
	poly_type *path;
    XGCValues gcv;
	GC gc;

#if AXMS
	sz = clip_info->n_paths *sizeof(POINT *);
	cpoints = (POINT **)TaskAlloc(0, sz);
#else
	sz = clip_info->n_paths *sizeof(XPoint *);
	cpoints = (XPoint **)TaskAlloc(0, sz);
#endif

	for ( found=FALSE, i=0; i< clip_info->n_paths; i++ )
	{
		path = &clip_info->paths[i];
#if AXMS
		cpoints[i] = (POINT *)TaskAlloc(0, path->nPoints*sizeof(POINT));
#else
		cpoints[i] = (XPoint *)TaskAlloc(0, path->nPoints*sizeof(XPoint));
#endif
	
		for ( j=0; j< path->nPoints; j++ )
		{
			convMilsToPix(path->points[j].x, path->points[j].y, &x, &y);
			cpoints[i][j].x = x;
			cpoints[i][j].y = y;
			if ( !found )
			{
				found = TRUE;
				cmaxX = cminX = x;
				cmaxY = cminY = y;
			}

			if ( x < cminX )
				cminX = x;
			else if ( x > cmaxX )
				cmaxX = x;

			if ( y < cminY )
				cminY = y;
			else if ( y > cmaxY )
				cmaxY = y;
		}
	}

	if ( clip )
	{
		x1 = clip->x_position;
		y1 = clip->y_position;
		x2 = clip->x_size + x1 - 1;
		y2 = clip->y_size + y1 - 1;
		convMilsToPix(x1, y1, &x_p, &y_p);
		convMilsToPix(x2, y2, &x_size, &y_size);

		if ( x_p > cminX )
			cminX = x_p;
		if ( y_p > cminX )
			cminY = y_p;

		if ( cmaxX > x_size )
			cmaxX = x_size;
		if ( cmaxY > y_size )
			cmaxY = y_size;
	}
	wid = cmaxX - cminX + 1;
	hyt = cmaxY - cminY + 1;

#if AX_X11
	/* create a cleared pixmap to use as a mask */
	clip_mask = XCreatePixmap(Dpy, xwin, wid, hyt, 1);
	temp_mask = XCreatePixmap(Dpy, xwin, wid, hyt, 1);

	gc_mask = GCForeground | GCBackground;
	memset((char *)&gcv, 0, sizeof(XGCValues));
	gcv.foreground = 0;
	gcv.background = 1;
	gc = axCreateGC(Dpy, clip_mask, DefaultScreen(Dpy), 1,
					(XFontStruct *) 0, gc_mask, &gcv);
	if ( clip )
	{
		rect[0].x = x_p - cminX;
		rect[0].y = y_p - cminY;
		rect[0].width  = x_size - x_p;
		rect[0].height = y_size - y_p;
		XSetClipRectangles(Dpy, gc, 0, 0, rect, 1, Unsorted);
	}

	XFillRectangle(Dpy, clip_mask, gc, 0, 0, wid, hyt);

	XSetForeground(Dpy, gc, 1);
	XSetBackground(Dpy, gc, 0);
	XSetFunction(Dpy, gc, GXcopy);
#endif

	/* fill the first path */
	path = &clip_info->paths[0];

#if AXMS
	final_hrgn = MS_CreatePolygonRgn(cpoints[0], path->nPoints, WINDING);
#else
	for ( j=0; j< path->nPoints; j++ )
	{
		cpoints[0][j].x -= cminX;
		cpoints[0][j].y -= cminY;
	}
	XFillPolygon(Dpy, clip_mask, gc, cpoints[0], path->nPoints,
					  Complex, CoordModeOrigin);
#endif

	for ( i=1; i< clip_info->n_paths; i++ )
	{
		path = &clip_info->paths[i];

#if AXMS
		hrgn = MS_CreatePolygonRgn(cpoints[i], path->nPoints, WINDING);
		MS_CombineRgn(final_hrgn, hrgn, final_hrgn, RGN_OR);
		MS_DeleteObject(hrgn);
#else
		for ( j=0; j< path->nPoints; j++ )
		{
			cpoints[i][j].x -= cminX;
			cpoints[i][j].y -= cminY;
		}
		XFillPolygon(Dpy, temp_mask, gc, cpoints[i], path->nPoints,
						  Complex, CoordModeOrigin);
		XSetFunction(Dpy, gc, GXand);
		XCopyArea(Dpy, temp_mask, clip_mask, gc, 0, 0, wid, hyt, 0, 0);
		XSetGraphicsExposures(Dpy, gc, FALSE);

		/* clear the surface */
		XSetFunction(Dpy, gc, GXcopy);
		XFillRectangle(Dpy, temp_mask, gc, 0, 0, wid, hyt);
#endif
	}

	for ( i=0; i<clip_info->n_paths; i++ )
		TaskFree(0, cpoints[i]);
	TaskFree(0, cpoints);
#if AX_X11
	XFreePixmap(Dpy, temp_mask);
	XFreeGC(Dpy, gc);
#endif

	/* done */
	*x_off = cminX;
	*y_off = cminY;
#if AXMS
	return final_hrgn;
#else
	return(clip_mask);
#endif
}

/********************************************************************
* AxFillCompoundPoly - multi path filling routine
*/
AxFillCompoundPoly(w, n_polys, n_open_pts, pt, attr, clip, clip_info)
AXWidget w;
int n_polys;
int *n_open_pts;
TmPoint32 **pt;
attr_type *attr;
TmRect32 *clip;
clip_path *clip_info;
{
    int minX, minY, maxX, maxY, wid, hyt, i, ix, *n_pts, cminX, cminY;
    int x_p, y_p, x_size, y_size, x1, y1, x2, y2, x_off, y_off, j, sz;
    Pixmap mask, clip_mask;
    XRectangle rect[1];
    XPoint **points;
    unsigned long gc_mask;
    Drawable xwin;
	poly_type *path;
    XGCValues gcv;
    GC gc;

    /* if there's no work to do, quick return */
    if ( attr->ffg.see_thru && attr->fbg.see_thru )
        return;
    if ( n_polys <= 0 )
        return;

    /* fills must be closed, so close polys as needed */
    n_pts = (int *)TaskAlloc(0, n_polys * sizeof(int *));
    for ( ix=0; ix<n_polys; ix++ )
    {
        n_pts[ix] = n_open_pts[ix];
        if ( n_pts[ix] > 2 &&
             (pt[ix][n_pts[ix]-1].x != pt[ix][0].x ||
              pt[ix][n_pts[ix]-1].y != pt[ix][0].y) )
        {
            pt[ix][n_pts[ix]] = pt[ix][0];
            n_pts[ix] += 1;
        }
    }

    /* copy the ATM points into XPoints */
    maxX = minX = pt[0][0].x;
    maxY = minY = pt[0][0].y;
    points = (XPoint **)TaskAlloc(0, n_polys * sizeof(XPoint *));
    for ( ix=0; ix<n_polys; ix++ )
    {
        /* convert the list to a list of points */
        points[ix] = (XPoint *)TaskAlloc(0, n_pts[ix] * sizeof(XPoint));
        for ( i=0; i<n_pts[ix]; i++ )
        {
            points[ix][i].x = pt[ix][i].x;
            points[ix][i].y = pt[ix][i].y;

            if ( points[ix][i].x < minX )
                minX = points[ix][i].x;
            else if ( points[ix][i].x > maxX )
                maxX = points[ix][i].x;

            if ( points[ix][i].y < minY )
                minY = points[ix][i].y;
            else if ( points[ix][i].y > maxY )
                maxY = points[ix][i].y;
        }
    }
    wid = maxX - minX + 1;
    hyt = maxY - minY + 1;
    if ( wid == 0 || hyt == 0 )
    {
        TaskFree(0, n_pts);
        for ( i=0; i<n_polys; i++ )
            TaskFree(0, points[i]);
        TaskFree(0, points);
        return;
    }

    /* get the drawable */
    xwin = AxGetDrawableFromWidget(w, &x_off, &y_off);

    /* create a cleared pixmap to use as a mask */
    mask = XCreatePixmap(Dpy, xwin, wid, hyt, 1);
    gc_mask = GCForeground | GCBackground;
    memset((char *)&gcv, 0, sizeof(XGCValues));
    gcv.foreground = 0;
    gcv.background = 1;
    gc = axCreateGC(Dpy, mask, DefaultScreen(Dpy), 1,
                    (XFontStruct *) 0, gc_mask, &gcv);
    XFillRectangle(Dpy, mask, gc, 0, 0, wid, hyt);

    /* set the clip */
	/* handle clip path */
	if ( clip_info )
	{
		clip_mask = create_clip_mask(xwin, clip, clip_info, &cminX, &cminY);

		/* set the pixmap as a clip mask */
#if AXMS
		axmsXSetClipMaskAsRegion(gc, (HRGN)clip_mask);
#else
		XSetClipOrigin(Dpy, gc, cminX, cminY);
		XSetClipMask(Dpy, gc, clip_mask);
#endif
	}
    else
	{
		clip_mask = (Pixmap)NULL;
		if ( clip )
		{
			x1 = clip->x_position;
			y1 = clip->y_position;
			x2 = clip->x_size + x1 - 1;
			y2 = clip->y_size + y1 - 1;
			conv_i_p32(&x1, &y1, &x_p, &y_p, 1);
			conv_i_p32(&x2, &y2, &x_size, &y_size,1);
			rect[0].x = x_p - minX;
			rect[0].y = y_p - minY;
			rect[0].width  = x_size -x_p + 1;
			rect[0].height = y_size -y_p + 1;
			XSetClipRectangles(Dpy, gc, 0, 0, rect, 1, Unsorted);
		}
	}

    /* copy the shape onto the pixmap */
    XSetForeground(Dpy, gc, 1);
    XSetBackground(Dpy, gc, 0);
    XSetFunction(Dpy, gc, GXxor);
    for ( ix=0; ix<n_polys; ix++ )
    {
        if ( n_pts[ix] <= 2 )
            continue;

        /* set the points to the zero point */
        for ( i=0; i<n_pts[ix]; i++ )
        {
            points[ix][i].x -= minX;
            points[ix][i].y -= minY;
        }

        XFillPolygon(Dpy, mask, gc, points[ix], n_pts[ix],
                     Complex, CoordModeOrigin);
    }

#if AXMS
    if ( Running_Windows95 && GRADIENT_FILL(attr->f_type) )
        {
        AxWinDrawGradientImage(xwin, minX, minY, wid, hyt, attr, mask, FALSE);
        attr->f_type = -1;  /* just to flag that we are done. */
        }
#endif

    /* set the pixmap as a clip mask */
    XSetClipOrigin(Dpy, Gc1, minX, minY);
    XSetClipMask(Dpy, Gc1, mask);

    /* do the fill */
    doAttrFill(xwin, minX, minY, wid, hyt, attr, FALSE);

    /* done */
    XFreePixmap(Dpy, mask);
    XFreeGC(Dpy, gc);
    TaskFree(0, n_pts);
    for ( ix=0; ix<n_polys; ix++ )
        TaskFree(0, points[ix]);
    TaskFree(0, points);

#if AX_X11
	if ( clip_mask )
		XFreePixmap(Dpy, clip_mask);
#endif
}

/********************************************************************
* do_ax_picture_poly - display picture in path
*/
do_ax_picture_poly(xwin, n_polys, n_pts, pt, attr, clip, clip_info)
Drawable xwin;
int n_polys;
int *n_pts;
TmPoint32 **pt;
attr_type *attr;
TmRect32 *clip;
clip_path *clip_info;
{
    unsigned long gc_mask;
    XGCValues gcv;
	XPoint **cpts, **pts;
    Pixmap mask, clip_mask;
    XRectangle rect[1];
    int wid, hyt, x_size, y_size, i, ix, found;
    int x1, y1, x2, y2, x_p, y_p, cminX, cminY;
	int minX, minY, maxX, maxY;
    GC gc;

    wid = maxX - minX + 1;
    hyt = maxY - minY + 1;


    /* copy the ATM points into XPoints */
    pts  = (XPoint **)TaskAlloc(0, n_polys * sizeof(XPoint *));
    cpts = (XPoint **)TaskAlloc(0, n_polys * sizeof(XPoint *));
    for ( ix=found=0; ix<n_polys; ix++ )
    {
        /* get the min values */
        if ( !found )
        {
            if ( pt[ix] )
            {
                minX = maxX = pt[ix][0].x;
                minY = maxY = pt[ix][0].y;
                found = TRUE;
            }
        }

        /* convert the list to a list of points */
        pts[ix] = (XPoint *)TaskAlloc(0,  n_pts[ix] * sizeof(XPoint));
        cpts[ix] = (XPoint *)TaskAlloc(0, n_pts[ix] * sizeof(XPoint));
        for ( i=0; i<n_pts[ix]; i++ )
        {
            pts[ix][i].x = pt[ix][i].x;
            pts[ix][i].y = pt[ix][i].y;

            cpts[ix][i].x = pts[ix][i].x;
            cpts[ix][i].y = pts[ix][i].y;

            if ( pts[ix][i].x < minX )
                minX = pts[ix][i].x;
            else if ( pts[ix][i].x > maxX )
                maxX = pts[ix][i].x;

            if ( pts[ix][i].y < minY )
                minY = pts[ix][i].y;
            else if ( pts[ix][i].y > maxY )
                maxY = pts[ix][i].y;
        }
    }

    /* create a cleared pixmap to use as a mask */
    mask = XCreatePixmap(Dpy, xwin, wid, hyt, 1);
    gc_mask = GCForeground | GCBackground;
    memset((char *)&gcv, 0, sizeof(XGCValues));
    gcv.foreground = 0;
    gcv.background = 1;
    gc = axCreateGC(Dpy, mask, DefaultScreen(Dpy), 1,
                    (XFontStruct *) 0, gc_mask, &gcv);
    XFillRectangle(Dpy, mask, gc, 0, 0, wid, hyt);

    /* copy the shape onto the pixmap */
    if ( clip_info )
	{
		clip_mask = create_clip_mask(xwin, clip, clip_info, &cminX, &cminY);

		/* set the pixmap as a clip mask */
#if AXMS
		axmsXSetClipMaskAsRegion(gc, (HRGN)clip_mask);
#else
		XSetClipOrigin(Dpy, gc, cminX, cminY);
		XSetClipMask(Dpy, gc, clip_mask);
#endif
	}
    else
	{
		clip_mask = (Pixmap)NULL;
		if ( clip )
		{
			x1 = clip->x_position;
			y1 = clip->y_position;
			x2 = clip->x_size + x1 - 1;
			y2 = clip->y_size + y1 - 1;
			conv_i_p32(&x1, &y1, &x_p, &y_p, 1);
			conv_i_p32(&x2, &y2, &x_size, &y_size,1);
			rect[0].x = x_p;
			rect[0].y = y_p;
			rect[0].width  = x_size -x_p + 1;
			rect[0].height = y_size -y_p + 1;
			XSetClipRectangles(Dpy, gc, 0, 0, rect, 1, Unsorted);
		}
	}

    XSetForeground(Dpy, gc, 1);
    XSetBackground(Dpy, gc, 0);
    XSetFunction(Dpy, gc, GXxor);

    /* do the actual fill */
    if ( (minX != maxX) && (minY != maxY) ) for ( ix=0; ix<n_polys; ix++ )
    {
        if ( n_pts[ix] <= 2 || !cpts[ix] )
            continue;
		for ( i=0; i< n_pts[ix]; i++ )
		{
			cpts[ix][i].x -= minX;
			cpts[ix][i].y -= minY;
		}
        /* render the contained area of the fill */
        XFillPolygon(Dpy, mask, gc, cpts[ix], n_pts[ix], Complex,CoordModeOrigin);
    }

    /* set the pixmap as a clip mask */
	XSetClipOrigin(Dpy, Gc1, minX, minY);
    XSetClipMask(Dpy, Gc1, mask);

    /* do the fill */
	do_color_picture_pattern(xwin, minX, minY, wid, hyt, attr, FALSE);

    /* reset the clip */
    XSetClipMask(Dpy, Gc1, None);

    /* do the line fill */
    if ( (minX != maxX) && (minY != maxY) ) for ( i=0; i<n_polys; i++ )
    {
        if ( n_pts[i] <= 2 || !pts[i] )
            continue;

        /* make sure all the pixels on the stroke are rendered, as well */
        XSetLineAttributes(Dpy, Gc1, 0, LineSolid, CapButt, JoinBevel);
        XDrawLines(Dpy, xwin, Gc1, pts[i],n_pts[i], CoordModeOrigin);
    }
    /* done */
    XFreePixmap(Dpy, mask);
    XFreeGC(Dpy, gc);
    for ( i=0; i<n_polys; i++ )
    {
        TaskFree(0, pts[i]);
        TaskFree(0, cpts[i]);
    }
    TaskFree(0, pts);
    TaskFree(0, cpts);
#if AX_X11
	if ( clip_mask )
		XFreePixmap(Dpy, clip_mask);
#endif
}

/********************************************************************
* AxFillSimplePoly - mono path patterned filling routine
*/
AxFillSimplePoly(w, n_polys, n_open_pts, pt, attr, clip, clip_info)
AXWidget w;
int n_polys;
int *n_open_pts;
TmPoint32 **pt;
attr_type *attr;
TmRect32 *clip;
clip_path *clip_info;
{
    int minX,minY, maxX,maxY, ix, i, *n_pts, inv, found;
    int c,m,y,k, cminX, cminY;
    int x_off, y_off, wid, hyt;
    unsigned long fg_pel, bg_pel;
    XPoint **pts;
    Pixmap fill_map;
    Pixmap clip_mask;
    Drawable xwin;

    /* if there's no work to do, quick return */
    if ( attr->ffg.see_thru && attr->fbg.see_thru )
    {
        if ( !attr->use_ffg && !attr->use_fbg )
            return;
    }
    if ( n_polys <= 0 )
        return;

    /* fills must be closed, so close polys as needed */
    n_pts = (int *)TaskAlloc(0, n_polys * sizeof(int *));
    for ( ix=0; ix<n_polys; ix++ )
    {
        n_pts[ix] = n_open_pts[ix];
        if ( n_pts[ix] > 2 &&
             (pt[ix][n_pts[ix]-1].x != pt[ix][0].x ||
              pt[ix][n_pts[ix]-1].y != pt[ix][0].y) )
        {
            pt[ix][n_pts[ix]] = pt[ix][0];
            n_pts[ix] += 1;
        }
    }

    /* copy the ATM points into XPoints */
    pts = (XPoint **)TaskAlloc(0, n_polys * sizeof(XPoint *));
    for ( ix=found=0; ix<n_polys; ix++ )
    {
        /* get the min values */
        if ( !found )
        {
            if ( pt[ix] )
            {
                minX = maxX = pt[ix][0].x;
                minY = maxY = pt[ix][0].y;
                found = TRUE;
            }
        }

        /* convert the list to a list of points */
        pts[ix] = (XPoint *)TaskAlloc(0, n_pts[ix] * sizeof(XPoint));
        for ( i=0; i<n_pts[ix]; i++ )
        {
            pts[ix][i].x = pt[ix][i].x;
            pts[ix][i].y = pt[ix][i].y;

            if ( pts[ix][i].x < minX )
                minX = pts[ix][i].x;
            else if ( pts[ix][i].x > maxX )
                maxX = pts[ix][i].x;

            if ( pts[ix][i].y < minY )
                minY = pts[ix][i].y;
            else if ( pts[ix][i].y > maxY )
                maxY = pts[ix][i].y;
        }
    }

    xwin = AxGetDrawableFromWidget(w, &x_off, &y_off);

    /* set the map to fill the poly */
    inv = attr->ffg.see_thru ? TRUE : FALSE;
    wid = maxX - minX + 1;
    hyt = maxY - minY + 1;
	if ( wid <= 0 || hyt <= 0 )
		attr->f_pic = FALSE;
    if ( attr->f_type == USERDEF_FILL_TYPE && attr->f_pic )
    {
		do_ax_picture_poly(xwin, n_polys, n_pts, pt, attr, clip, clip_info);
		TaskFree(0, n_pts);
		for ( ix=0; ix<n_polys; ix++ )
			TaskFree(0, pts[ix]);
		TaskFree(0, pts);
		return;
	}
	/* handle clip path */
	if ( clip_info )
	{
		clip_mask = create_clip_mask(xwin, clip, clip_info, &cminX, &cminY);

		/* set the pixmap as a clip mask */
#if AXMS
		axmsXSetClipMaskAsRegion(Gc1, (HRGN)clip_mask);
#else
		XSetClipOrigin(Dpy, Gc1, cminX, cminY);
		XSetClipMask(Dpy, Gc1, clip_mask);
#endif
	}
    else
	{
		clip_mask = (Pixmap)NULL;
		if ( clip )
			AxSetClip(Gc1, clip);
	}

    if ( attr->f_type == USERDEF_FILL_TYPE ) /* user defined fill */
    {
        if ( attr->f_ddep == 8 )
            fill_map = get_color_pixmap(attr->f_dwid, attr->f_dhyt,
                                        attr->f_data, attr->f_cmap,inv);
        else
            fill_map = get_mono_pixmap(attr->f_data, attr->f_dwid,
                                        attr->f_dhyt, inv);
    }
    else
        fill_map = inv ? GetStippleImap(attr->f_pat):GetStippleMap(attr->f_pat);
    if ( attr->fill_abs )
        XSetTSOrigin(Dpy, Gc1, attr->f_xoff, attr->f_yoff);
    else
        XSetTSOrigin(Dpy, Gc1, minX + attr->f_xoff, minY + attr->f_yoff);

	if ( attr->f_ddep == 8 )
		XSetTile(Dpy, Gc1, fill_map);
	else
		XSetStipple(Dpy, Gc1, fill_map);
    /* set the color of the fill */
    if ( attr->use_ffg )
        fg_pel = AlisPixels[TMA_BLACK];
    else
    {
        AxGetRealPixel(2, attr->ffg.c, attr->ffg.m, attr->ffg.y,
                          attr->ffg.k, 255, &fg_pel);
    }
    if ( attr->use_fbg )
        bg_pel = AlisPixels[TMA_WHITE];
    else
    {
        AxGetRealPixel(2, attr->fbg.c,
                              attr->fbg.m,
                              attr->fbg.y,
                              attr->fbg.k, 255, &bg_pel);
    }

    switch(attr->mode)
    {
        case TMS_X_D:
            XSetState(Dpy, Gc1,
                      AlisPixels[1]^AlisPixels[0], bg_pel, GXxor, PlaneMask);
            XSetFillStyle (Dpy, Gc1, FillSolid);
        break;
        case TMZEROS:
            XSetState(Dpy, Gc1, AlisPixels[0], bg_pel, GXcopy, PlaneMask);
            XSetFillStyle (Dpy, Gc1, FillSolid);
        break;
        case TMONES:
            XSetState(Dpy, Gc1, AlisPixels[1], bg_pel, GXcopy, PlaneMask);
            XSetFillStyle (Dpy, Gc1, FillSolid);
        break;
        case 16: /* widget background, like TMZEROS */
            XSetState(Dpy, Gc1, WidgetPixels[0], bg_pel, GXcopy, PlaneMask);
            XSetFillStyle (Dpy, Gc1, FillSolid);
        break;
        case 17: /* widget foreground, like TMONES */
            XSetState(Dpy, Gc1, WidgetPixels[1], bg_pel, GXcopy, PlaneMask);
            XSetFillStyle (Dpy, Gc1, FillSolid);
        break;
        default:
            if ( attr->ffg.see_thru )
                XSetState(Dpy, Gc1, bg_pel, fg_pel, GXcopy, PlaneMask);
            else
                XSetState(Dpy, Gc1, fg_pel, bg_pel, GXcopy, PlaneMask);

            if ( attr->f_ddep == 8 )
                XSetFillStyle (Dpy, Gc1, FillTiled);
            else
            {
                if ( attr->ffg.see_thru || attr->fbg.see_thru )
                    XSetFillStyle (Dpy, Gc1, FillStippled);
                else
                    XSetFillStyle (Dpy, Gc1, FillOpaqueStippled);
            }
        break;

    }

    /* do the actual fill */
    if ( (minX != maxX) && (minY != maxY) ) for ( i=0; i<n_polys; i++ )
    {
        if ( n_pts[i] <= 2 || !pts[i] )
            continue;

        /* render the contained area of the fill */
        XFillPolygon(Dpy, xwin, Gc1, pts[i],n_pts[i], Complex,CoordModeOrigin);

        /* make sure all the pixels on the stroke are rendered, as well */
        XSetLineAttributes(Dpy, Gc1, 0, LineSolid, CapButt, JoinBevel);
        XDrawLines(Dpy, xwin, Gc1, pts[i],n_pts[i], CoordModeOrigin);
    }

    /* done */
	if ( attr->f_type == 1 ) /* user defined fill */
		AxFreePixmap(fill_map);
    TaskFree(0, n_pts);
    for ( ix=0; ix<n_polys; ix++ )
        TaskFree(0, pts[ix]);
    TaskFree(0, pts);
#if AX_X11
	if ( clip_mask )
		XFreePixmap(Dpy, clip_mask);
#endif
}

/********************************************************************
* AxFillPoly - top level generic path filling routine
*/
AxFillPoly(w, n_polys, n_open_pts, pt, attr, clip, clip_info)
AXWidget w;
int n_polys;
int *n_open_pts;
TmPoint32 **pt;
attr_type *attr;
TmRect32 *clip;
clip_path *clip_info;
{
    if ( check_for_fill(attr) == FALSE )
        return;

    /* special modes draw solid lines only */
    if ( attr->mode != TMS )
    {
        attr->f_type = 0;
        attr->f_pat  = 4;
    }

    /* convert percent fill to pattern fill */
    if ( attr->f_type == 5 )
    {
        attr->f_type = 0;
        attr->f_pat = get_percent_ix(attr->f_pat);
    }

    /* intersect multiple paths using even-odd fill rule */
    if ( n_polys > 1 )
        AxFillCompoundPoly(w, n_polys, n_open_pts, pt, attr, clip, clip_info);

    /* simple path, need at least 2 points */
    else if ( n_open_pts[0] > 2 )
    {
		switch ( attr->f_type )
		{
			case LINEAR_FILL_TYPE:
			case RADIAL_FILL_TYPE:
			case RECT_FILL_TYPE:
			case BILINEAR_FILL_TYPE:
            	AxFillCompoundPoly(w, n_polys, n_open_pts, pt, attr,
									  clip, clip_info);
			break;
			default:
				AxFillSimplePoly(w, n_polys, n_open_pts, pt, attr,
									clip, clip_info);
			break;
		}
    }
}

/********************************************************************
* AxDrawFilledPoly - patterned path stroking routine
*/
AxDrawFilledPoly(w, n_pts, pt, attr, clip, clip_info)
AXWidget w;
int n_pts;
TmPoint32 *pt;
attr_type *attr;
TmRect32 *clip;
clip_path *clip_info;
{
    int minX, minY, maxX, maxY, line_wid, wid, hyt, i, j, n_segs;
    int x_p, y_p, x_size, y_size;
    int x_off, y_off, cminX, cminY;
    int l_width, l_style, l_cap, l_join;
    unsigned long gc_mask;
    XPoint *points;
    short *seg_off;
    XGCValues gcv;
    Pixmap mask, clip_mask;
    GC gc;
    Drawable xwin;

    /* convert the list to a list of points */
    points = (XPoint *)TaskAlloc(0, n_pts * sizeof(XPoint));
    maxX = minX = pt[0].x;
    maxY = minY = pt[0].y;
    for ( i=0; i<n_pts; i++ )
    {
        points[i].x = pt[i].x;
        points[i].y = pt[i].y;

        if ( points[i].x < minX )
            minX = points[i].x;
        else if ( points[i].x > maxX )
            maxX = points[i].x;

        if ( points[i].y < minY )
            minY = points[i].y;
        else if ( points[i].y > maxY )
            maxY = points[i].y;
    }

    /* set the line to the zero point (offset by it's width) */
    line_wid = ceil((double)attr->l_wid / 2.0);
    minX -= line_wid;
    minY -= line_wid;
    maxX += line_wid;
    maxY += line_wid;
    for ( i=0; i<n_pts; i++ )
    {
        points[i].x -= minX;
        points[i].y -= minY;
    }
    wid = maxX - minX +1;
    hyt = maxY - minY +1;

    xwin = AxGetDrawableFromWidget(w, &x_off, &y_off);
    /* create a cleared pixmap to use as a mask */
    mask = XCreatePixmap(Dpy, xwin, wid, hyt, 1);
    gc_mask = GCForeground | GCBackground;
    memset((char *)&gcv, 0, sizeof(XGCValues));
    gcv.foreground = 0;
    gcv.background = 1;
    gc = axCreateGC(Dpy, mask, DefaultScreen(Dpy), 1,
                    (XFontStruct *) 0, gc_mask, &gcv);
    XFillRectangle(Dpy, mask, gc, 0, 0, wid, hyt);

    /* set the clip */
	/* handle clip path */
	if ( clip_info )
	{
		clip_mask = create_clip_mask(xwin, clip, clip_info, &cminX, &cminY);

		/* set the pixmap as a clip mask */
#if AXMS
		axmsXSetClipMaskAsRegion(gc, (HRGN)clip_mask);
#else
		XSetClipOrigin(Dpy, gc, cminX, cminY);
		XSetClipMask(Dpy, gc, clip_mask);
#endif
	}
    else
	{
		clip_mask = (Pixmap)NULL;
		if ( clip )
			AxSetClip(gc, clip);
	}

    XSetForeground(Dpy, gc, 1);
    XSetBackground(Dpy, gc, 0);
    if ( attr->l_style != TM_SOLID )
    {
        l_width = (attr->l_wid == 1) ? 0 : attr->l_wid;
        l_style = LineOnOffDash;
        l_cap   = CapButt;
        l_join  = JoinBevel;
        XSetLineAttributes(Dpy, gc, l_width, l_style, l_cap, l_join);

        XSetDashes(Dpy, gc,
                   dash_offset[attr->l_style-1],
                   dash_list[attr->l_style-1],
                   dash_length[attr->l_style-1]);
        XDrawLines(Dpy, mask, gc, points, n_pts, CoordModeOrigin);
    }
    else
    {
        l_width = 0;
        l_style = LineSolid;
        l_cap   = CapRound;
        l_join  = JoinMiter;
        XSetLineAttributes(Dpy, gc, l_width, l_style, l_cap, l_join);

        getRoundPenOutline(attr->l_wid, &n_segs, &seg_off);
        for ( i=0; i<n_segs; i++ )
        {
            if ( seg_off[2*i] || seg_off[2*i+1] )
            {
                for ( j=0; j<n_pts; j++ )
                {
                    points[j].x += seg_off[2*i];
                    points[j].y += seg_off[2*i+1];
                }
            }
            XDrawLines(Dpy,mask, gc, points, n_pts, CoordModeOrigin);
        }
        TaskFree(0, seg_off);
    }

#if AXMS
    if ( Running_Windows95 && GRADIENT_FILL(attr->l_type) )
        {
        AxWinDrawGradientImage(xwin, minX, minY, wid, hyt, attr, mask, TRUE);
        attr->l_type = -1;  /* just to flag that we are done. */
        }
#endif

    /* set the pixmap as a clip mask */
    XSetClipOrigin(Dpy, Gc1, minX, minY);
    XSetClipMask(Dpy, Gc1, mask);

    /* do the line fill */
    doAttrFill(xwin, minX, minY, wid, hyt, attr, TRUE);

    /* done */
    XFreePixmap(Dpy, mask);
    XFreeGC(Dpy, gc);
    TaskFree(0, points);
#if AX_X11
	if ( clip_mask != (Pixmap)NULL )
		XFreePixmap(Dpy, clip_mask);
#endif
}

/********************************************************************
* AxDrawSolidPoly - solid fg path stroking routine
*/
AxDrawSolidPoly(w, n_pts, pt, attr, clip, clip_info)
AXWidget w;
int n_pts;
TmPoint32 *pt;
attr_type *attr;
TmRect32 *clip;
clip_path *clip_info;
{
    int l_wid, l_style, l_cap, l_join, mode, planes;
    int minX, minY, maxX, maxY, sz, x;
    int cminX, cminY, cmaxX, cmaxY, found;
    int c, m, y, k, wid, hyt, i, j, n_segs, ix;
    int x_off, y_off;
    int x_p, y_p, x_size, y_size, x1, y1, x2, y2;
    unsigned long fg_pel, bg_pel;
    XPoint *points;
    short *seg_off;
    Drawable xwin;
    Pixmap mask, temp_mask, clip_mask;
    XGCValues gcv;
    unsigned long gc_mask;
	GC gc;

    /* drawing a fg color line with transparent color is a noop */
    if ( attr->mode == TMS && attr->lfg.see_thru )
        return;

    /* allocate XPoints */
    points = (XPoint *)TaskAlloc(0, n_pts * sizeof(XPoint));
    maxX = minX = pt[0].x;
    maxY = minY = pt[0].y;
    for(i=0; i<n_pts; i++)
    {
        points[i].x = pt[i].x;
        points[i].y = pt[i].y;
    }

    xwin = AxGetDrawableFromWidget(w, &x_off, &y_off);

	/* handle clip path */
	if ( clip_info != (clip_path *)0 )
	{
		clip_mask = create_clip_mask(xwin, clip, clip_info, &cminX, &cminY);

		/* set the pixmap as a clip mask */
#if AXMS
		axmsXSetClipMaskAsRegion(Gc1, (HRGN)clip_mask);
#else
		XSetClipOrigin(Dpy, Gc1, cminX, cminY);
		XSetClipMask(Dpy, Gc1, clip_mask);
#endif
	}
	else
	{
		clip_mask = (Pixmap)NULL;
		if ( clip )
			AxSetClip(Gc1, clip);
	}

    /* set color, mode, and mask */
    switch ( attr->mode )
    {
        case TMS_X_D:
            mode   = GXxor;
            planes = AllPlanes;
            fg_pel = AlisPixels[TMA_WHITE] ^ AlisPixels[TMA_BLACK];
            bg_pel = AlisPixels[TMA_WHITE];
        break;

        case TMZEROS:
            mode   = GXcopy;
            planes = PlaneMask;
            fg_pel = AlisPixels[TMA_WHITE];
            bg_pel = BGpixel;
        break;

        case 16: /* widget background, like TMZEROS */
            mode   = GXcopy;
            planes = PlaneMask;
            fg_pel = WidgetPixels[0];
            bg_pel = BGpixel;
        break;

        case TMONES:
            mode   = GXcopy;
            planes = PlaneMask;
            fg_pel = AlisPixels[TMA_BLACK];
            bg_pel = BGpixel;
        break;

        case 17: /* widget foreground, like TMONES */
            mode   = GXcopy;
            planes = PlaneMask;
            fg_pel = WidgetPixels[1];
            bg_pel = BGpixel;
        break;

        default:
            mode   = GXcopy;
            planes = PlaneMask;
            c = attr->lfg.c;
            m = attr->lfg.m;
            y = attr->lfg.y;
            k = attr->lfg.k;
            AxGetRealPixel(2, c,m,y,k, 255, &fg_pel);
            bg_pel = BGpixel;
        break;
    }
    if ( attr->use_lfg )
        fg_pel = AlisPixels[TMA_BLACK];
    if ( attr->use_lbg )
        bg_pel = AlisPixels[TMA_WHITE];
    XSetState(Dpy, Gc1, fg_pel, bg_pel, mode, planes);


    /* draw the solid line */
    XSetFillStyle (Dpy, Gc1, FillSolid);
    if ( attr->l_style != TM_SOLID )
    {
        l_wid = (attr->l_wid == 1) ? 0 : attr->l_wid;
        l_style = LineOnOffDash;
        l_cap   = CapButt;
        l_join  = JoinBevel;
        XSetLineAttributes(Dpy, Gc1, l_wid, l_style, l_cap, l_join);

        XSetDashes(Dpy, Gc1,
                   dash_offset[attr->l_style-1],
                   dash_list[attr->l_style-1],
                   dash_length[attr->l_style-1]);
        XDrawLines(Dpy, xwin, Gc1, points, n_pts, CoordModeOrigin);
    }
    else
    {
        l_wid = 0;
        if ( mode != GXxor )
            l_style = LineSolid;
        else
        {
            XSetDashes(Dpy, Gc1, dash_offset[3], dash_list[3], dash_length[3]);
            l_style = LineDoubleDash;
        }
        l_cap   = CapRound;
        l_join  = JoinMiter;
        XSetLineAttributes(Dpy, Gc1, l_wid, l_style, l_cap, l_join);

        getRoundPenOutline(attr->l_wid, &n_segs, &seg_off);
        for ( i=0; i<n_segs; i++ )
        {
            if ( seg_off[2*i] || seg_off[2*i+1] )
            {
                for ( j=0; j<n_pts; j++ )
                {
                    points[j].x += seg_off[2*i];
                    points[j].y += seg_off[2*i+1];
                }
            }
            XDrawLines(Dpy, xwin, Gc1, points, n_pts, CoordModeOrigin);
        }
        TaskFree(0, seg_off);
    }

    /* done */
    TaskFree(0, points);
#if AX_X11
	if ( clip_mask != (Pixmap)NULL )
		XFreePixmap(Dpy, clip_mask);
#endif
}

/********************************************************************
* AxDrawPoly - top level generic path stroking routine
*/
AxDrawPoly(w, n_pts, pt, attr, clip, clip_info)
AXWidget w;
int n_pts;
TmPoint32 *pt;
attr_type *attr;
TmRect32 *clip;
clip_path *clip_info;
{
    /* the following cases result in no line being drawn */
    if ( attr->mode == TMS && attr->lfg.see_thru && attr->lbg.see_thru )
        return;
    if ( attr->line_flag == FALSE )
        return;
    if ( attr->l_wid <= 0 )
        return;
    if ( n_pts < 2 )
        return;

    /* special modes draw solid lines only */
    if ( attr->mode != TMS )
    {
        attr->l_type = 0;
        attr->l_pat  = 4;
    }

    /* convert percent fill to pattern fill (display only) */
    if ( attr->l_type == 5 )
    {
        attr->l_type = 0;
        attr->l_pat = get_percent_ix(attr->l_pat);
    }

    if ( attr->l_type == 2 ) /* linear gradient */
        AxDrawFilledPoly(w, n_pts, pt, attr, clip, clip_info);
    else if ( attr->l_type == 3 ) /* radial gradient */
        AxDrawFilledPoly(w, n_pts, pt, attr, clip, clip_info);
    else if ( attr->l_type == 4 ) /* racangular gradient */
        AxDrawFilledPoly(w, n_pts, pt, attr, clip, clip_info);
    else if ( attr->l_type == 6 ) /* bilinear gradient */
        AxDrawFilledPoly(w, n_pts, pt, attr, clip, clip_info);
    else if ( attr->l_type == 1 ) /* user defined fill */
        AxDrawFilledPoly(w, n_pts, pt, attr, clip, clip_info);
    else if ( attr->l_type == 0 ) /* built in fill */
    {
        if ( attr->l_pat == 4 ) /* solid foreground color */
            AxDrawSolidPoly(w, n_pts, pt, attr, clip, clip_info);
        else
            AxDrawFilledPoly(w, n_pts, pt, attr, clip, clip_info);
    }
}

/****************************************************************************/
AxDrawPoints(w, nPts, pts, x1, x2, y1, y2)
AXWidget w;
int nPts;
TmPoint *pts;
int x1, x2, y1, y2;
{
    XRectangle rects[1];
    unsigned long oldFG, oldBG;
    int i;
    int x_off, y_off;
    Drawable xwin;

    /* convert from mils (1000 dpi) to screen res */
    for ( i=0; i<nPts; i++ )
        milPix ( &(pts[i].x), &(pts[i].y) );

    /* set the graphic context */
    milPix32 ( &x1, &y1 );
    milPix32 ( &x2, &y2 );
    rects[0].x = x1;
    rects[0].y = y1;
    rects[0].width  = x2 - x1;
    rects[0].height = y2 - y1;
    XSetClipRectangles(Dpy, Gc1, 0, 0, rects, 1, Unsorted);
    oldFG = FGpixel;
    oldBG = BGpixel;
    WMsetPixels(TMA_BLACK, TMA_WHITE);
    XSetState(Dpy, Gc1, FGpixel, BGpixel, GXcopy, PlaneMask);
    XSetLineAttributes(Dpy, Gc1, 0, LineSolid, CapButt, JoinBevel);

    /* draw the dots */
    xwin = AxGetDrawableFromWidget(w, &x_off, &y_off);
    XDrawPoints(Dpy, xwin, Gc1, (XPoint *)pts, nPts, CoordModeOrigin);

    /* restore state */
    FGpixel = oldFG;
    BGpixel = oldBG;
}

/****************************************************************************/
AxDrawLines(w, nPts, pts, x1, x2, y1, y2)
AXWidget w;
int nPts;
TmPoint *pts;
int x1, x2, y1, y2;    /* clip dimensions */
{
    XRectangle rects[1];
    unsigned long oldFG, oldBG;
    int i;
    int x_off, y_off;
    Drawable xwin;

    /* convert from mils (1000 dpi) to screen res */
    for ( i=0; i<nPts; i++ )
        milPix ( &(pts[i].x), &(pts[i].y) );

    /* set the graphic context */
    milPix32 ( &x1, &y1 );
    milPix32 ( &x2, &y2 );
    rects[0].x = x1;
    rects[0].y = y1;
    rects[0].width  = x2 - x1;
    rects[0].height = y2 - y1;
    XSetClipRectangles(Dpy, Gc1, 0, 0, rects, 1, Unsorted);
    oldFG = FGpixel;
    oldBG = BGpixel;
    WMsetPixels(TMA_BLACK, TMA_WHITE);
    XSetState(Dpy, Gc1, FGpixel, BGpixel, GXcopy, PlaneMask);
    XSetLineAttributes(Dpy, Gc1, 0, LineSolid, CapButt, JoinBevel);

    /* draw the lines */
    xwin = AxGetDrawableFromWidget(w, &x_off, &y_off);
    XDrawLines(Dpy, xwin, Gc1, (XPoint *)pts, nPts, CoordModeOrigin);

    /* restore state */
    FGpixel = oldFG;
    BGpixel = oldBG;
}

/****************************************************************************/
AxFillrect(w, x,y,wid,hyt, mode, dither_id, cmap,c1,c2, x0,y0, clip)
AXWidget w;
int x, y, wid, hyt, mode, dither_id;
cmap_type *cmap; 
int c1, c2;
TmRect32 *clip;
{
    unsigned long fg_pel, bg_pel;
    AxRectangle rDst;
    color_type *col;
    int ts_x, ts_y;
    int x_off, y_off;
    Drawable xwin;

    /* set the context */
    switch(mode)
    {
        default:
            /* get the fg color from the colormap */
            col = &cmap->color[c1];
            AxGetRealPixel(2, col->c,col->m,col->y,col->k, 255, &fg_pel);

            /* get the bg color from the colormap */
            col = &cmap->color[c2];
            AxGetRealPixel(2, col->c,col->m,col->y,col->k, 255, &bg_pel);

            mode = GXcopy;
        break;
        case TMZEROS:
            fg_pel = AlisPixels[0];
            bg_pel = AlisPixels[1];
            mode = GXcopy;
        break;
        case TMONES:
            fg_pel = AlisPixels[1];
            bg_pel = AlisPixels[0];
            mode = GXcopy;
        break;
        case 16:
        case 17:
        case TMID:
        case TMS_X_D:
            fg_pel = AlisPixels[0] ^ AlisPixels[1];
            bg_pel = AlisPixels[0] ^ AlisPixels[1];
            mode = GXxor;
        break;
    }

    /* set the state */
    if ( clip )
        AxSetClip(Gc1, clip);
    XSetState(Dpy, Gc1, fg_pel, bg_pel, mode, AllPlanes);
    if ( dither_id == 0 )
        XSetFillStyle(Dpy, Gc1, FillSolid);
    else
    {
        XSetTSOrigin(Dpy, Gc1, x0, y0);
        XSetStipple(Dpy, Gc1, GetStippleMap(dither_id-1));
        XSetFillStyle(Dpy, Gc1, FillOpaqueStippled);
    }

    /* fill the area with the current color */
    xwin = AxGetDrawableFromWidget(w, &x_off, &y_off);
    XFillRectangle(Dpy, xwin, Gc1, x, y, wid, hyt);
}

/****************************************************************************/
ppText(w, pos, str, len)
AXWidget w;
TmPoint32 *pos;
char *str;
int len;
{
    Drawable xwin;
    int x_off, y_off;

    /* set the clip */
    XSetClipMask(Dpy, Gc2, None);

    /* draw */
    if ( len > 0 )
    {
        /* draw the string */
        xwin = AxGetDrawableFromWidget(w, &x_off, &y_off);
        axDrawString(Dpy, xwin, Gc2, (XFontStruct *) 0,
                     pos->x, pos->y, str, len);
    }
}

/****************************************************************************/
void TmPPLine(w, nPoints, points)
AXWidget w;
int nPoints;
TmPoint32 *points;
{
    Drawable xwin;
    XPoint *pts;
    int i;
    int x_off, y_off;

    /* set the clip */
    XSetClipMask(Dpy, Gc2, None);

    /* draw */
    if ( nPoints > 0 )
    {
        /* copy the ATM points into X Points */
        pts = (XPoint *)TaskAlloc(0, sizeof(XPoint) * nPoints);
        for ( i=0; i<nPoints; i++ ) 
        {    
            /* copies an int into a short!!! */ 
            pts[i].x = points[i].x; 
            pts[i].y = points[i].y;
        }

        /* draw the line onto the drawable */
        xwin = AxGetDrawableFromWidget(w, &x_off, &y_off);
        XDrawLines(Dpy, xwin, Gc2, pts, nPoints, CoordModeOrigin);

        /* done */
        TaskFree(0, pts);
    }
}

/******************************************************************/
AxDrawHandle(w, x,y, ix, h_pos,v_pos, clip)
AXWidget w;
int x, y;
int ix;
TmRect32 *clip;
{
    unsigned long fg_pel, bg_pel;
    Drawable xwin;
    int x_off, y_off;

    /* set the clip */
    if ( clip )
        AxSetClip(Gc2, clip);

    /* set the handle position */
    conv_i_p32(&x,&y, &x,&y, 1); /* mills to pixels */
    switch(h_pos)
    {
        default: x -= 4; break;
        case 1:  x -= 9; break;
        case 2:  x += 1; break;
    }
    switch(v_pos)
    {
        default: y -= 4; break;
        case 1:  y -= 9; break;
        case 2:  y += 1; break;
    }

    /* do the draw (should be 9x9 pixmaps) */
    xwin = AxGetDrawableFromWidget(w, &x_off, &y_off);
    XCopyPlane(Dpy, GetStippleMap(29+ix), xwin, Gc2, 0,0, 9,9, x,y, 1);
}

AxGetLineOutline(closed, in_n_pts, in_pt, out_n_pts, out_pt, x_wid, y_wid)
int closed;
int in_n_pts, *out_n_pts;
TmPoint32 *in_pt;
TmPoint32 **out_pt;
int x_wid, y_wid;
{
    double sinAng, cosAng, delta_x, delta_y, h;
    int i, ix, dx, dy, n_pts;
    TmPoint32 *pt;

    /* for now! (should depend on endcaps!) */
    if ( in_n_pts < 2 )
    {
        *out_n_pts = 0;
        *out_pt = (TmPoint32 *)0;
        return(AOK);
    }

    /* create a container to hold all the points */
    n_pts = (in_n_pts-1) * 4 + 1;
    if ( closed )
        n_pts += 2;
    pt = (TmPoint32 *)TaskAlloc(0, n_pts * sizeof(TmPoint32));

    /* start it */
    delta_x = in_pt[1].x - in_pt[0].x;
    delta_y = in_pt[1].y - in_pt[0].y;
    h = hypot(delta_x, delta_y);
    sinAng = delta_y/h;
    cosAng = delta_x/h;
    dx = x_wid * sinAng;
    dy = y_wid * cosAng;
    pt[0].x = in_pt[i-1].x - dx;
    pt[0].y = in_pt[i-1].y + dy;
    ix = 1;

    /* add an endcap */
    if ( !closed )
    {
    }

    /* create the points "above" */
    for ( i=1; i<in_n_pts; i++ )
    {
        delta_x = in_pt[i].x - in_pt[i-1].x;
        delta_y = in_pt[i].y - in_pt[i-1].y;
        h = hypot(delta_x, delta_y);
        sinAng = delta_y/h;
        cosAng = delta_x/h;
        dx = x_wid * sinAng;
        dy = y_wid * cosAng;

        pt[ix].x = in_pt[i-1].x + dx;
        pt[ix].y = in_pt[i-1].y - dy;
        ix++;

        pt[ix].x = in_pt[i].x + dx;
        pt[ix].y = in_pt[i].y - dy;
        ix++;
    }

    /* is it closed? */
    if ( closed )
    {
        pt[ix++] = pt[1];
        pt[ix++] = pt[0];
    }

    /* create the points below */
    for ( i=in_n_pts-1; i>1; i-- )
    {
        delta_x = in_pt[i].x - in_pt[i-1].x;
        delta_y = in_pt[i].y - in_pt[i-1].y;
        h = hypot(delta_x, delta_y);
        sinAng = delta_y/h;
        cosAng = delta_x/h;
        dx = x_wid * sinAng;
        dy = y_wid * cosAng;

        pt[ix].x = in_pt[i-1].x - dx;
        pt[ix].y = in_pt[i-1].y + dy;
        ix++;

        pt[ix].x = in_pt[i-1].x - dx;
        pt[ix].y = in_pt[i-1].y + dy;
        ix++;
    }

    /* finish it (endcap) */
    if ( !closed )
    {
    }

    /* done */
    *out_n_pts = n_pts;
    *out_pt = pt;
    return(AOK);
}

AxGetForegroundPixel()
{
    unsigned long bg_pel;
    AxGetRealPixel(2, 0,0,0, 255, 255, &bg_pel);
    return( bg_pel);
}
