/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: simplwin.c
   Applix ASC version: 20
   Date: Mon Nov 16 13:53:43 1998
*/
#include <basicwin.h>
#include <aos.h>
#include <xmlib.h>
#include <paratm.h>    /* for Wp global */
#include <parint.h>
#include <scac.h>    /* for APSS */

#include <simplwin.p>

extern win_list_t *TotalWindowsList;   /* for resource tracking */ 
extern int NumWindows; 



LRESULT CALLBACK
SimpleClassProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    PAINTSTRUCT ps;
    XEvent xevent;
    LRESULT retval;
    Window win;
    BOOL bEvent;


    /* Find the 'simple' window associated with the event,
     * which was created by AxCrtSimpleWindow.
     */
    win = AxGetPackagedSimpleHWND(hWnd);
    if ( !win || !SetWpFromX11Window(win) )
        {
        /* have no idea about this window yet */
        return MS_DefWindowProc(hWnd, message, wParam, lParam);
        }

    memset(&xevent,0,sizeof(XEvent));
	bEvent = WinToXEvent(hWnd,message,wParam,lParam, &xevent, (AXWidget)NULL);

    /* HACK ALERT! The SS child windows (editwin, rowwin, colwin) were      */
    /* implemented in X as real child windows, not widgets. The event mask  */
    /* dictated that all mouse events in rowwin and colwin were to be       */
    /* ignored, X passing them along to the SS main window automatically.   */
    /* The editwin, however, processes its own mouse events. To complicate  */
    /* matters further, we must pass mouse movement events to the SS in order*/
    /* to set the edit window cursor, another free ride in X. So we get:    */
    /*              mouse move           mouse button   */
    /*  editwin         SS                  editwin     */
    /*  rowwin          SS                  SS          */
    /*  colwin          SS                  SS          */
    /* Where the entry corresponds to the window that needs to receive      */
    /* and process the event.  The general case is that all events except   */
    /* WM_PAINT are passed along to the parent window by returning          */
    /* HTTRANSPARENT but note that we CANNOT simply pass the move events to */
    /* the SS window since the editwin will never receive the subsequent    */
    /* buttonpress event. To keep the SS code as generic as possible, the   */
    /* hack is done here. My apologies to all future generations. Long live */
    /* the widget.  jimr                                                    */

    retval = 0;
    switch (message)
        {
        case WM_PAINT:
            if ( IgnorePaintMessages() )
               /* defer painting until later to handle "full drag" mode */
               break;

            MS_BeginPaint(hWnd, &ps);
            if ( bEvent )
                {
                /* we should be using the clipping DC, but we can't because *
                 * the SS checks against a previously created window handle *
                 * to see if it's dealing with a row/col window.(BRAIN-DEAD)*/
                xevent.xany.window = win;
                DoAppEvent(&xevent);
                }
            MS_EndPaint(hWnd, &ps);
            break;


        case WM_NCHITTEST:
            /* See Above Comment */
            if ( (Wp->ap_id == APSS) && (((Window)Wp->editwin == win) ||
                                        ((Window)Wp->inCellWin == win)) )
                retval = MS_DefWindowProc(hWnd, message, wParam, lParam);
            else
                retval = HTTRANSPARENT;
            break;


        case WM_MOUSEMOVE:
            if ( bEvent && (Wp->ap_id == APSS) && (((Window)Wp->editwin == win)
                             || ((Window)Wp->inCellWin == win)) )
                {
                /* find and add the x offset to the x coord */
#if 0
                w = TotalWindowsList;
                while ( w && (w->win != win) )
                    w = w->next;
                if ( !w )
                    return(retval); /* hopefully never get here */
#endif
                //if ( w )
                //    {
                xevent.xany.window = win; //hWnd;
                DoAppEvent(&xevent);
                //    }
                }
            else
                retval = HTTRANSPARENT; /* should never get here */
            break;


        case WM_MOUSEACTIVATE:
            /* if user clicks in editwin, make sure SS is brought *
             * to forefront if it isn't already.                  */
			/* Also send the FocusIn/FocusOut events to the edit windows. */
            if ( (Wp->ap_id == APSS) && (((Window)Wp->editwin == win) ||
                            ((Window)Wp->inCellWin == win)) )
				{
				/* First, send a FocusOut to the other editwin (regardless
                 * of whether it actually had focus beforehand).
				 */
				WinToXEvent((HWND)NULL, WM_KILLFOCUS,
								(WPARAM)0, (LPARAM)0,
								&xevent, (AXWidget)NULL);
				if ( (Window)Wp->editwin == win )
					{
                	xevent.xany.window = Wp->inCellWin;
					}
				else
					{
                	xevent.xany.window = (Window)Wp->editwin;
					}
				DoAppEvent(&xevent);

				/* Now send a FocusIn to the window we are activating. */
				WinToXEvent(hWnd, WM_SETFOCUS,
								(WPARAM)0, (LPARAM)0,
								&xevent, (AXWidget)NULL);
                xevent.xany.window = win;
				DoAppEvent(&xevent);
                retval =  MA_ACTIVATE;
				}
            break;


        case WM_LBUTTONDOWN:
        case WM_MBUTTONDOWN:
        case WM_RBUTTONDOWN:
        case WM_LBUTTONUP:
        case WM_MBUTTONUP:
        case WM_RBUTTONUP:
            if ( bEvent && (Wp->ap_id == APSS) && (((Window)Wp->editwin == win)
                                     || ((Window)Wp->inCellWin == win)))
                {
                /* allow the ss edit window to get keypress events */
                xevent.xany.window = win;
                DoAppEvent(&xevent);
                }
            else
                retval = HTTRANSPARENT;
            break;


        default:
            /* let NT send to window just beneath, in all */
            /* our uses this IS the parent. */
            retval = HTTRANSPARENT;
        }

    return(retval);
}



/********************************************************************/
HWND
AxCrtSimpleWindow(Display *dpy, Window parent, int x, int y,
        unsigned int wid, unsigned int ht, unsigned int bord,
        unsigned long border, unsigned long back)
{
    short obj_type;
    win_list_t *w;
    HWND hwnd, parent_hwnd;


    obj_type = GET_DRAWABLE_TYPE(parent);
    if ( obj_type == AX_HWND_TYPE )
        {
        parent_hwnd = (HWND)GET_DRAWABLE_HANDLE(parent);
        }
    else
        {
        AXWidget parent_widget;
        HDC hdc = (HDC)GET_DRAWABLE_HANDLE(parent);
        parent_widget = GetWidgetFromDC(hdc);
        parent_hwnd = parent_widget->winid;
        }

    //if ( !parent_widget )
    //    tempAbort("could not get parent widget in AxCrtWindow");

    if ( !parent_hwnd )
        return (HWND)NULL;/* parent wasn't created (probably windowless app) */

    hwnd = MS_CreateWindow(NXLT("SimpleClass"), "", WS_CHILD | WS_CLIPSIBLINGS,
                                x, y, wid, ht, parent_hwnd, (HMENU)NULL,
                                (HANDLE)NULL, (LPVOID)NULL);
    MS_SetWindowPos(hwnd, HWND_BOTTOM, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);

    /* Add simple_hwnd and HDC to TotalWindowsList for later */
    /* searching since ret_win is not unique but hwnd is. */
    w = (win_list_t*)malloc(sizeof(win_list_t));
    memset( w, 0, sizeof(win_list_t) );
    w->win = AxPackageHandle(hwnd, AX_HWND_TYPE);
    //w->xOffset = x;
    w->hwnd = hwnd;
    w->parentHwnd = parent_hwnd;
    w->next = TotalWindowsList;
    if ( w->next )
        {
        w->next->prev = w;
        }
    w->prev = NULL;
    TotalWindowsList = w;
    NumWindows++;

    return w->win;
}

static Window
AxGetPackagedSimpleHWND(HWND hwnd)
{
    win_list_t *wp;
    for ( wp = TotalWindowsList; wp; wp = wp->next )
        {
        if ( wp->hwnd == hwnd )
            return wp->win;
        }
    return (Window)NULL;
}

