/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: tmouse.c
   Applix ASC version: 31
   Date: Tue Apr  6 21:57:46 1999
*/
#include <stdio.h>
#include <math.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <aos.h>
#include <paripc.h>
#include <tmi.h>
#include <paratm.h>
#include <tmassert.h>
#include <paramin.h>
#include <parint.h>
#include <parout.h>
#include <bounds.h>
#include <getstr13.h>
#include <mainhelp.h>
#include <in.h>
#include <getstr4.h>
#include <scac.h>
#include <usrevent.h>
#include <iconic.h>
#include <elhelper.h>
#include <tmouse.p>

#define ABS(x) (((x) > 0) ? (x) : -(x))

#define X_THRESH    35
#define Y_THRESH    35

#define P_STRK_MAX 400     /* square of max distance for pick-stroke */

#define UNCLIPPED_STK 0
#define CLIP__LEFT 1
#define CLIP_RIGHT 2
#define CLIP_BELOW 4
#define CLIP_ABOVE 8

/* symbols worth removing */




short MousWid, MousHt, Mc_xMax, Mc_yMax;

/* globals */
short MevButton;
extern int AxPointerMotion;
extern int DialogBoxActive; /* TRUE while a dialog box is active */
extern unsigned long TimeOfButtonPress;
extern short PopupIgnoreButtonUp;
extern int THIMhelpMode;
extern struct FullShrk FSdata[];

static void stroke_end();
static void pick();
struct wind_struct *getActiveWin();
AxRectangle *getMrect(), *absRect();

short ButtonIsUp;       /* Hint for popups logic */
short IgnoreButtonUp;   /* set by residual popup activity */
static int stk_flag;             /* stroke state */
static TmPoint32 StkPel;    /* current stroke point, screen pixels */
static TmPoint32 StkPix;    /* current stroke point, pixels - adjustable */
static TmPoint32 StkPrev;   /* previous stroke point */
static TmPoint32 StkPt0;    /* penultimate to previous point */
static TmPoint32 StkCtrl0;  /* penultimate control to previous point */
static TmPoint32 StkSource; /* source to use for constraint */
TmPoint ButDown;            /* button-down point of stroke or pick, pixels */
static struct wind_struct *StkWp;  /* window of current stroke */

static short is_constrained;
static short PenCount;      /* point counter for pen strokes */
static short stk_clip_code; /* clip for previously reported stroke point */

static void stroke_action();
void atm_mous(){}

/***************************************************************************
 *
 * ptr_ev -- send a pointer motion event to the owning app
 */
void ptr_ev(wID, x, y, shifted, controlled)
int wID, x, y, shifted, controlled;
{
    TmPoint32 mouse_pos;
    sys_event_t uEvent;

    uEvent.type = PTR_EVENT; 
    uEvent.ev.ptr.shifted = shifted;
    uEvent.ev.ptr.controlled = controlled;
    conv_i_p32(&uEvent.ev.ptr.x, &uEvent.ev.ptr.y, &x, &y, PIXEL_TO_MILLS);
    AxSendMes(window[wID].p_id, &uEvent, sizeof(uEvent), APPOINTER);
}

/***************************************************************************
 *
 * constrainStk -- force the stroke point globals to a fixed angle
 */
static void constrainStroke(windowP)
struct wind_struct *windowP;
{
    int dx, dy, code;

    /* is this stroke constrainable? */
    code = windowP->stkCode;
    if ( code != TMK_PEN &&
         code != TMK_ONE_LINE &&
         code != TMK_TWO_LINE &&
         code != TMK_HVLINE &&
         code != TMK_ONE_CURVE &&
         code != TMK_TWO_CURVE )
        return;

    if ( is_constrained == FALSE )
    {
        if ( code == TMK_PEN ) /* doesn't scroll */
        {
            StkSource.x = mous_x;
            StkSource.y = mous_y;
        }
        else
        {
            StkSource.x = windowP->stkAnc[0].x;
            StkSource.y = windowP->stkAnc[0].y;
        }
        is_constrained = TRUE;
    }
    else
    {
        /* get the stroke so far */
        dx = StkSource.x - mous_x;
        dy = StkSource.y - mous_y;

        /* constrain it */
        if ( ABS(dx - dy) < ABS((int)(dx/2.0)) ) /* 45 deg y increasing line */
            mous_y = StkSource.y - dx;
        else if ( ABS(dx + dy) < ABS((int)(dx/2.0)) ) /* other 45 deg line */
            mous_y = StkSource.y + dx;
        else if ( ABS(dx) > ABS(dy) ) /* vertical line */
            mous_y = StkSource.y;
        else /* horizontal line */
            mous_x = StkSource.x;
    }
}

void resetConstraint()
{
    is_constrained = FALSE;
}

/***************************************************************************
 *
 * tmSetStroke -- set current stroke point globals to mouse position
 */
static void tmSetStroke( windowP )
struct wind_struct *windowP;
{
    /* constrain mouse motion */
    if ( ThimIsControled() )
        constrainStroke( windowP );
    else
        is_constrained = FALSE;

    /* set globals */
    StkPel.x = StkPix.x = (int)mous_x;
    StkPel.y = StkPix.y = (int)mous_y;
}

/***************************************************************************
 *
 * button_down -- button-down event (could be pick or stroke)
 * Note: A pick should always be preceded by a button-down.
 */
static button_down( winP )
struct wind_struct *winP;
{
    int winPicked;
    int pos;
    AxRectangle optRect;

    LImodExit();
    ButDown.x = mous_x;
    ButDown.y = mous_y;

    /* initialize stroke handling logic in case a stroke follows */
    is_constrained = FALSE;
    ButtonIsUp = FALSE;
    tmSetStroke( winP );                   /* no longer needed here ? */
    stk_flag = FIRST_STK;       /* first-point (uncommitted) state */
    stk_clip_code = UNCLIPPED_STK; /* OK to report first clipped point */
}

/***************************************************************************
 *
 * stroke_motion -- handle stroke motion event
 */
static void stroke_motion( winP )
struct wind_struct *winP;
{
    int winID;
    AxRectangle rr;

    if ( IgnoreButtonUp || (winP == (struct wind_struct *)NULL) )
        return;

    /* the first point in the stroke handler */
    if (stk_flag == FIRST_STK)
    {
        /* if we're in the inner rectangle of active window */
        if ( inRect(ButDown.x, ButDown.y, &winP->inner ) )
        {
            stroke_action( winP );
            stk_flag = CONT_STK;
        }

        /* set window context of stroke (if needed) */
        if (stk_flag != FIRST_STK)
        {
            is_constrained = FALSE;
            StkWp = winP;
        }
    }
    
    /* handle other points */
    if (stk_flag != FIRST_STK)
    {
        /* update current point globals */
        tmSetStroke( winP ) ;

        /* do local stroke actions */
        stroke_action( winP );
    }
}

/***************************************************************************
 *
 * stroke_end -- process end (button-up) point of stroke
 */
static void stroke_end( winP )
struct wind_struct *winP;
{
    int dx, dy, pos;
    AxRectangle rr;

    ButtonIsUp = TRUE;

    /* in case was residual from popup activity */
    if ( IgnoreButtonUp )
    {
        stk_flag = RESET;
        return;
    }

    /* set current stroke point globals */
    tmSetStroke( winP );

    /* do the button up thing */
    switch(stk_flag)
    {
        case(CONT_STK):
            stk_flag = RESET;
            stroke_action( winP );
        break;

        case(FIRST_STK):
            stk_flag = RESET;
            dx = mous_x - ButDown.x;
            dy = mous_y - ButDown.y;

            /* if it hasn't moved enough, it's a pick */
            if ( dx*dx + dy*dy < P_STRK_MAX )
                pick( winP );
        break;
    }
}


/***************************************************************************
 *
 * mev -- process a mouse event, assumes (mous_x, mous_y) up-to-date
 */
void Mouse_Event(mousEvent, winP )
int mousEvent;
struct wind_struct *winP;
{
    /* Don't allow events in PIXED work area during Elf recording */
    if ( IsElfInLearnMode() && (AppIdOfCurrWindow() == APPIXED) )
        return;

    /* set the window context for the stroke */
    if ( stk_flag == RESET )
        StkWp = winP;

    /* dispatch the event */
    switch(mousEvent)
    {
        default:
        case(MOUSE_MOTION) :         break;
        case(BUT_DOWN)     : button_down( winP );   break;
        case(SINGLE_PICK)  : pick( winP ); break;
        case(STROKE_MOTION): stroke_motion( winP );   break;
        case(STROKE_END)   : stroke_end( winP );   break;
    }
}

/***************************************************************************
 *
 * send_mice -- generate an APSYS event
 */
static void
send_mice(wp, eventType)
struct wind_struct *wp;
int eventType;
{
    sys_event_t uEvent;

    uEvent.type = eventType;
    switch ( eventType )
    {
        case STROKE_EVENT:
            uEvent.ev.stroke.x = wp->td.td_strok.x_strok;
            uEvent.ev.stroke.y =  wp->td.td_strok.y_strok;
            uEvent.ev.stroke.type = wp->td.td_strok.tst.type;
            uEvent.ev.stroke.button = wp->td.td_strok.tst.button;
            uEvent.ev.stroke.hvStatus =wp->td.td_strok.tst.hv_status;
            uEvent.ev.stroke.shifted = wp->td.td_strok.tst.shifted;
            uEvent.ev.stroke.controlled = wp->td.td_strok.tst.controled;
            uEvent.ev.stroke.firstPoint = wp->td.td_strok.tst.first_point;
            uEvent.ev.stroke.lastPoint = wp->td.td_strok.tst.last_point;
            uEvent.ev.stroke.boundary = wp->td.td_strok.tst.boundary;
            uEvent.ev.stroke.echoed = wp->td.td_strok.tst.echoed;
            uEvent.ev.stroke.clipped = wp->td.td_strok.tst.clipped;
            AxSendMes(wp->p_id, &uEvent, sizeof(uEvent), APSYS);
        break;
        case PICK_EVENT:
            uEvent.ev.pick.x = wp->td.td_pick.x_pick;
            uEvent.ev.pick.y = wp->td.td_pick.y_pick;
            uEvent.ev.pick.type = wp->td.td_pick.tpt.type;
            uEvent.ev.pick.button = wp->td.td_pick.tpt.button;
            uEvent.ev.pick.quickClick = wp->td.td_pick.tpt.quick_click;
            uEvent.ev.pick.shifted = wp->td.td_pick.tpt.shifted;
            uEvent.ev.pick.controlled = wp->td.td_pick.tpt.controled;
            uEvent.ev.pick.helpBit = wp->td.td_pick.tpt.helpBit;
            AxSendMes(wp->p_id, &uEvent, sizeof(uEvent), APSYS);
        break;
    }
}

/***************************************************************************
 *
 * chk_bound -- return TRUE if point is on boundary
 */
static int chk_bound(winP,pt)
struct wind_struct *winP;
TmPoint *pt; /* stroke point in window mils */
{
    return ( pt->x <= 0 || pt->x >= winP->wind_attr.x_size ||
             pt->y <= 0 || pt->y >= winP->wind_attr.y_size );
}




/***************************************************************************
 *
 * stroke_action -- process an application stroke point
 * Note: stk_flag indicates first/last point cases
 * Note: End points are always reported, while intermediate points may not
 *       be depending on stroke mode and clipping considerations.
 */
static void
stroke_action( winP )
struct wind_struct *winP;
{
    TmPoint pt;
    int clipped;            /* point outside clipping rectangle flag */
    AxRectangle *rp;
    int clip_code;
    int interReport;        /* report intermediate stroke points? */
    short tempx, tempy;
    short x,y;
    int button, shift, ctrl;


    /* determine if the stroke was clipped */
    if ( winP->stkClip.origin.x >= 0 )
        rp = &winP->stkClip;
    else
        rp = &winP->appClip;
    clip_code = UNCLIPPED_STK;
    clipped = !inRect((short)StkPel.x, (short)StkPel.y, rp);
    if ( clipped ) /* point outside clip rect? */
    {
        /* left or right? */
        if ( StkPel.x < rp->origin.x )
            clip_code = CLIP__LEFT;
        else if ( StkPel.x >= rp->corner.x )
            clip_code = CLIP_RIGHT;

        /* top or bottom? */
        if ( StkPel.y < rp->origin.y )
            clip_code |= CLIP_BELOW;
        else if ( StkPel.y >= rp->corner.y )
            clip_code |= CLIP_ABOVE;
    }

    /* changes in clip code result in forced reporting of intermediate point */
    if ( stk_clip_code != clip_code )
        stk_clip_code = clip_code;


    /* report the mouse point */
    button = MevButton;
    shift = ThimIsShifted();
    ctrl = ThimIsControled();
    winP->td.td_strok.tst.b1strok = TRUE; /* only kind supported */
    winP->td.td_strok.tst.first_point = stk_flag == FIRST_STK;
    winP->td.td_strok.tst.last_point = stk_flag == RESET;
    winP->td.td_strok.tst.echoed = winP->stkCode != TMK_NONE;
    tempx = (short)StkPel.x - winP->inner.origin.x;
    tempy = (short)StkPel.y - winP->inner.origin.y;
    conv_i_p(&pt.x, &pt.y, &tempx, &tempy, PIXEL_TO_MILLS);
    winP->td.td_strok.tst.boundary = chk_bound(winP,&pt);
    winP->td.td_strok.tst.clipped = clipped;
    winP->td.td_strok.tst.button = button;
    winP->td.td_strok.tst.shifted = shift;
    winP->td.td_strok.tst.controled = ctrl;
    winP->td.td_strok.x_strok = pt.x;
    winP->td.td_strok.y_strok = pt.y;
    winP->td_len = 1;
    winP->td_type = TM_STROKE;
    winP->term_flag = TM_STK_EVENT;
    winP->td.td_strok.tst.type = ButtonEventType();
    send_mice(winP, STROKE_EVENT);

    if ( stk_flag == RESET )
        winP->stkCode = TMK_NONE;

    if ( IsElfInLearnMode() )
    {
        x = pt.x;
        y = pt.y;
        milPix(&x,&y);
        if ( stk_flag == FIRST_STK )
        {
            if ( winP->td.td_strok.tst.type == TM_NORMAL_SELECT )
                ElfRecord(ELFREC_STK_START_SELECT,x,y);
            if ( winP->td.td_strok.tst.type == TM_EXTENDED_SELECT )
                ElfRecord(ELFREC_STK_START_EXTENDED,x,y);
            if ( winP->td.td_strok.tst.type == TM_MULTIPLE_SELECT )
                ElfRecord(ELFREC_STK_START_MULTIPLE,x,y);
        }
        else if ( stk_flag == RESET )
            ElfRecord(ELFREC_STK_END,x,y,winP->td.td_strok.tst.hv_status);
        else if (winP->td.td_strok.tst.boundary)
            ElfRecord(ELFREC_STK_BOUND,x,y);
        else
            ElfRecord(ELFREC_STK_POINT,x,y);
    }
}

/*************************************************************************
 *
 * ElfbStroke -- Send a stroke event to active window (from Elf)
 */
void ElfbStroke (type,xpos,ypos,button,shift,ctrl,hv_stat,stk_type)
int type;   /* 1-start, 0-mid, 2-end, 3-boundary */
int xpos, ypos; /* in pixels, window relative */
int hv_stat, button, shift, ctrl;
int stk_type;
{
    short x,y, tempx, tempy;
    struct wind_struct *winP;
 
    if ( cur_win == TM_NULWIN )
        return;
    winP = &window[ cur_win ];

    StkPel.x = (int)(xpos + winP->inner.origin.x);
    StkPel.y = (int)(ypos + winP->inner.origin.y);
    winP->td.td_strok.tst.b1strok = TRUE; /* only kind supported */
    winP->td.td_strok.tst.first_point = type == 1;
    winP->td.td_strok.tst.last_point = type == 2;
    winP->td.td_strok.tst.echoed = 1;
    tempx = (short)StkPel.x - winP->inner.origin.x;
    tempy = (short)StkPel.y - winP->inner.origin.y;
    conv_i_p(&x, &y, &tempx, &tempy, PIXEL_TO_MILLS);
    winP->td.td_strok.tst.boundary = FALSE; /* OBSOLETE */
    winP->td.td_strok.tst.clipped = FALSE;  /* ??c */
    winP->td.td_strok.tst.button = button;
    winP->td.td_strok.tst.shifted = shift;
    winP->td.td_strok.tst.controled = ctrl;
    winP->td.td_strok.tst.hv_status = hv_stat;
    winP->td.td_strok.tst.type = stk_type;
    winP->td.td_strok.x_strok = x;
    winP->td.td_strok.y_strok = y;
    winP->td_len = 1;
    winP->td_type = TM_STROKE;
    winP->term_flag = TM_STK_EVENT;
    send_mice(winP, STROKE_EVENT);
}

/************************************************************************\
*                     Pick handling and reporting                        *
\************************************************************************/

/*************************************************************************
 *
 * pick_term
 */
static void pick_term(winP,x,y,button,quick,shift,ctrl,pick_type)
struct wind_struct *winP;
int x,y,button,quick,shift,ctrl,pick_type;
{
    short tempx;
    short tempy;

    winP->term_flag = TM_PIK_EVENT;
    winP->td.td_pick.tpt.b1clk1 = SET;
    winP->td.td_pick.tpt.b1clk2 = RESET;
    winP->td.td_pick.tpt.helpBit = 0;
    winP->td.td_pick.tpt.button = button;
    winP->td.td_pick.tpt.shifted = shift;
    winP->td.td_pick.tpt.controled = ctrl;
    winP->td.td_pick.tpt.type = pick_type;

    if ( (quick < 0) || (quick > 4) )
        quick = 0;   /* just to be safe */
    winP->td.td_pick.tpt.quick_click = quick; /* # quick clicks AFTER first */

    tempx = x - winP->inner.origin.x;
    tempy = y - winP->inner.origin.y;
    conv_i_p(&tempx, &tempy, &tempx, &tempy, PIXEL_TO_MILLS);

    winP->td.td_pick.x_pick = tempx;
    winP->td.td_pick.y_pick = tempy;

    winP->td_type = TM_PICK_TD;
    winP->td.td_pick.reg_id = TM_NULREG;
    if ( IsElfInLearnMode() &&
        (AppIdOfCurrWindow() == APGE) || (AppIdOfCurrWindow() == APBG) ) {
        if ( pick_type == TM_NORMAL_SELECT )
            ElfRecord(ELFREC_PICK_SELECT,mous_x,mous_y,quick);
        else if ( pick_type == TM_EXTENDED_SELECT )
            ElfRecord(ELFREC_PICK_EXTENDED,mous_x,mous_y,quick);
        else if ( pick_type == TM_MULTIPLE_SELECT )
            ElfRecord(ELFREC_PICK_MULTIPLE,mous_x,mous_y,quick);
        else if ( pick_type == TM_PASTE_SELECT )
            ElfRecord(ELFREC_PICK_PASTE,mous_x,mous_y,quick);
    }
}

/*************************************************************************
 *
 * pickInner -- check for and handle pick in inner rectangle of window
 */
static void pickInner( winP )
struct wind_struct *winP;
{
    extern int DoubleMillisec, DoublePixSpan;
    static short last_x, last_y;
    static int quick = 0; /* # of quick clicks AFTER first */
    static unsigned long last_time;
    short i,j;
    short x1i,y1i;
    short x2i,y2i;
    short done_flag;
    AxRectangle rr;
    int x, y;

    if ( winP == (struct wind_struct *)NULL )
        return;

    if ( inRect(mous_x, mous_y, &winP->inner) )
    {
        x = mous_x;
        y = mous_y;
        if ( (TimeOfButtonPress - last_time < DoubleMillisec) &&
             (ABS(last_x - x) < DoublePixSpan) &&
             (ABS(last_y - y) < DoublePixSpan) &&
             (quick < 4) )
            quick++;
        else
            quick = 0;
        last_time = TimeOfButtonPress;
        last_x = x;
        last_y = y;
        pick_term ( winP, x, y, MevButton, 
                    quick,
                    ThimIsShifted(),
                    ThimIsControled(),
                    ButtonEventType());

        if ( winP->term_flag != RESET )
            send_mice(winP, PICK_EVENT);
    }
}

/*************************************************************************
 *
 * pick -- handle a single or double pick
 */
static void pick( winP )
struct wind_struct *winP;
{
    is_constrained = FALSE;
    ButtonIsUp = TRUE;
    stk_flag = RESET;

    /* in case was residual from popup/dialog activity */
    if ( IgnoreButtonUp )
        return;

    if ( THIMhelpMode )
        ElfbSetHelpMode(FALSE);

    stk_flag = RESET;

    pickInner( winP );
}

/*************************************************************************
 *
 * ElfbPick -- Send a pick (called by Elf)
 */
void ElfbPick (type,xpos,ypos,button,quick,shift,ctrl,pick_type)
int xpos, ypos; /* in pixels, window relative */
int quick, button, shift, ctrl;
int pick_type;
{
    struct wind_struct *winP;

    if ( cur_win == TM_NULWIN )
        return;
    winP = &window[ cur_win ];

    mous_x = xpos + winP->inner.origin.x;
    mous_y = ypos + winP->inner.origin.y;
    pick_term(winP,mous_x,mous_y,button,quick,shift,ctrl,pick_type);
    send_mice(winP, PICK_EVENT);
}
