/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: tmreg.c
   Applix ASC version: 5
   Date: Mon Nov 16 13:53:35 1998
*/
#include <stdio.h>
#include <fcntl.h>
#include <aos.h>
#include <paripc.h>
#include <tmi.h>
#include <tmassert.h>
#include <paratm.h>
#include <paratmgr.h>
#include <parreg.h>
#include <parint.h>

#define UNHIGH	0
#define HIGH	1

/*
 * element of highlighted region list
 */
struct hlist
{
	struct hlist *hl_link;
	short hl_rid;				/* region ID */
	AxRectangle hl_rect;
};

		/* TmDreg function */

atm_dreg(mes, mesLen, flag)
char *mes;
short mesLen;
short flag;
{
	union 
	{
	    struct SDregData dreg_data;
	    char rec_buf[sizeof(struct SDregData)];
	} rec_pack;

	short rx, wrx;
	short ev_fd;
	short res_fd;
	struct TmRegPacket *regp;

	ATMmesMov(mes, rec_pack.rec_buf, mesLen, sizeof(struct SDregData), flag);

#if CHECK
	PCkDreg(rec_pack.rec_buf);	/* check packet */
#endif
	SetWp(&window[rec_pack.dreg_data.window_id ]);
	TLOGGER((ATfp,"\ndefine regions from %d\n",Wp->window_id));

	if ( (Wp->regenWait) || ( (Wp->gRegenWait) && (Wp->g_mod_flag) ) ) 
		return;	/* regen requested, regions closed */

	if(Wp->cur_reg_num == 0)
	{
	    Wp->regions = (struct TmRegPacket *)
			ATMmalloc(sizeof(struct TmRegPacket)
				* rec_pack.dreg_data.r_def_num);
	}
	else
	{
	    Wp->regions = (struct TmRegPacket *)
			ATMrealloc(Wp->regions,
				sizeof(struct TmRegPacket) *
				(rec_pack.dreg_data.r_def_num + 
				Wp->cur_reg_num));
	}

	/* for regions being defined loop */

	for (rx = 0; rx < rec_pack.dreg_data.r_def_num; rx++)
	{
		regp = &rec_pack.dreg_data.r_def_table[rx];

		/* convert measurements to pixels */

	    mils_dc(&regp->x_position,
	    		&regp->y_position,
	    		&regp->x_position,
	    		&regp->y_position);
	    mils_dc(&regp->x_size,
	    		&regp->y_size,
	    		&regp->x_size,
	    		&regp->y_size);
		/*
		* convert position from window-relative
		* to screen-relative
		*/
	    regp->x_position += Wp->inner.origin.x;
	    regp->y_position += Wp->inner.origin.y;

		/* do highlighting */

	    if ( regp->action.pick == TM_DISABLE
	    		|| regp->action.pick == TM_ENABLE)
	    {
	    	regp->action.pickstate = TM_INACTIVE;
	    }
		wrx = Wp->cur_reg_num + rx;
	    Wp->regions[wrx] = *regp;

	    if (regp->action.pickstate == TM_ACTIVE)
	    {
			hiReg(wrx, TRUE);
			addHreg(wrx);
	    }
	}  /* for regions being defined loop */

	Wp->cur_reg_num += rec_pack.dreg_data.r_def_num;
}


		/* TmMreg function */

atm_mreg(mes, mesLen, flag)
char *mes;
short mesLen;
short flag;
{
	union 
	{
	    struct SMregData mreg_data;
	    char rec_buf[sizeof(struct SMregData)];
	} rec_pack;

	short rx, wrx;
	short j;
	short ev_fd;
	short res_fd;
	struct TmRegMPacket *mregp;
	struct TmRegPacket *wregp;

	ATMmesMov(mes, rec_pack.rec_buf, mesLen, sizeof(struct SMregData), flag);

#if CHECK
	PCkMreg(rec_pack.rec_buf);	/* check packet */
#endif
	SetWp(&window[rec_pack.mreg_data.window_id ]);
	TLOGGER((ATfp,"modify regions from %d\n",Wp->window_id));

	if ( (Wp->regenWait) || ( (Wp->gRegenWait) && (Wp->g_mod_flag) ) ) 
		return;	/* regen requested, regions closed */

	/* for regions being modified loop */

	for (rx = 0; rx <rec_pack.mreg_data.r_mod_num; rx++)
	{
		mregp = &rec_pack.mreg_data.r_mod_table[rx];

		/* find region index within window (wrx) */

	    wrx = 0;
		while (wrx < Wp->cur_reg_num && Wp->regions[wrx].r_id != mregp->r_id)
		{
			wrx++;
		}
	    if (wrx < Wp->cur_reg_num)				/* found match */
	    {
			wregp = &Wp->regions[wrx];
			wregp->action.pick = mregp->action.pick;
			wregp->action.open = mregp->action.open;
			if (((wregp->action.pick == TM_DISABLE)
						OR (wregp->action.pick == TM_ENABLE))
					AND (wregp->action.pickstate == TM_ACTIVE))
			{
				hlUhlReg(wrx);		/* active to inactive */
			}
			if ((wregp->action.pick == TM_HIGHLIGHT)
					OR (wregp->action.pick == TM_HIGH_UNHIGH))
			{
				/* region has highlighting action */

				if ((mregp->action.pickstate == TM_ACTIVE)
		    			AND (wregp->action.pickstate == TM_INACTIVE))
				{
					hlUhlReg(wrx);	/* inactive to active */
		    	}
		    	else if ((mregp->action.pickstate == TM_INACTIVE)
		        		AND (wregp->action.pickstate == TM_ACTIVE))
		        {
					hlUhlReg(wrx);		/* active to inactive */
		        }
			}  /* region has highlighting action */

			if (mregp->action.openstate != TM_NO_CHANGE)
				wregp->action.openstate = mregp->action.openstate;

	    }  /* found match */
	}  /* for regions being modified loop */
}  /* atm_mreg() */

/*
 * TmRclose function -- close region if it exists, else do nothing
 */
void atm_rclose(mes, mesLen, flag)
char *mes;
short mesLen;
short flag;
{
	union
	{
		struct SRcloseData rclose_data;
		char rec_buf[sizeof(struct SRcloseData)];
	} rec_pack;

	short i;
	short j;
	short ev_fd;
	short res_fd;

	ATMmesMov(mes, rec_pack.rec_buf, mesLen, sizeof(struct SRcloseData), flag);

#if CHECK
	PCkRclose(rec_pack.rec_buf);	/* check packet */
#endif
	TLOGGER((ATfp, "atm_rclose() rid=%d\n", rec_pack.rclose_data.r_id));
	SetWp(&window[rec_pack.rclose_data.window_id ]);
	j=0;
	while (j < Wp->cur_reg_num
				&& rec_pack.rclose_data.r_id != Wp->regions[j].r_id)
	{
	    j++;
	}

	if (j < Wp->cur_reg_num) closeReg(Wp,j);
}
/*
 * closeReg -- delete a region
 */
closeReg(w, rx)
struct wind_struct *w;
short rx;
{
	short i;

	assert(w == Wp);
	if (Wp->regions[rx].action.pickstate == TM_ACTIVE)
	{
		hlUhlReg(rx);
	}
	Wp->cur_reg_num--;
	for (i=rx; i < Wp->cur_reg_num; ++i)
	{
		Wp->regions[i] = Wp->regions[i+1];
	}
	Wp->regions = (struct TmRegPacket *)
					ATMrealloc( Wp->regions,
					sizeof(struct TmRegPacket) * Wp->cur_reg_num );
}
/*
 * caReg -- internal close all regions function
 */
caReg(outF)
	int outF;		/* perform output flag */
{
	short i;

	TLOGGER((ATfp, "caReg() cur_reg_num=%d\n", Wp->cur_reg_num));
	for (i=0; i<Wp->cur_reg_num; ++i)
	{
		if (Wp->regions[i].action.pickstate == TM_ACTIVE)
		{
			if (outF)
				hlUhlReg(i);
			else
				rmHreg(i);
		}
	}
	assert(Wp->hlhead == NULL);
	Wp->cur_reg_num = 0;
	ATMfree(Wp->regions);
	Wp->regions = NULL;
}

		/* TmARclose function */

void atm_arclose(mes, mesLen, flag)
char *mes;
short mesLen;
short flag;
{
	union
	{
	    struct SARcloseData arclose_data;
	    char rec_buf[sizeof(struct SARcloseData)];
	} rec_pack;
	short i;

	ATMmesMov(mes, rec_pack.rec_buf, mesLen, sizeof(struct SARcloseData), flag);

#if CHECK
	PCkARclose(rec_pack.rec_buf);	/* check packet */
#endif
	SetWp(&window[rec_pack.arclose_data.window_id ]);
	caReg(TRUE);		/* close all regions; perform output */
}

/*
 * regArect -- convert a region to a AxRectangle in absolute screen pixels
 */
AxRectangle *
regArect(regp, rectp)
	struct TmRegPacket *regp;
	AxRectangle *rectp;
{
	rectp->origin.x = regp->x_position;
	rectp->origin.y = regp->y_position;
	rectp->corner.x = rectp->origin.x + regp->x_size;
	rectp->corner.y = rectp->origin.y + regp->y_size;
	return rectp;
}
/*
 * hiReg -- highlight or unhighlight a region
 */
hiReg(ri, onOff)
	short ri;			/* region index */
	short onOff;
{
	AxRectangle rr;

	TLOGGER((ATfp, "hiReg(%d %d)\n", ri, onOff));
	regArect(&Wp->regions[ri], &rr);
	lrectf(Wp->layer, &rr, R_XOR);
}
/*
 * hlUhlReg -- invert pick state of a region
 *
 *			refers to window (Wp); inverts pixels,
 *			inverts pickstate variable; updates list
 *			of highlighted regions
 */
hlUhlReg(rx)
short rx;
{
	if (Wp->regions[rx].action.pickstate == TM_ACTIVE)
	{
		Wp->regions[rx].action.pickstate = TM_INACTIVE;
		hiReg(rx, FALSE);
		rmHreg(rx);
	}
	else
	{
		Wp->regions[rx].action.pickstate = TM_ACTIVE;
		hiReg(rx, TRUE);
		addHreg(rx);
	}
}
/*
 * regWrect -- compute window-relative AxRectangle for a region
 */
AxRectangle *
regWrect(regp, rectp)
	struct TmRegPacket *regp;
	AxRectangle *rectp;
{
	rectp->origin.x = regp->x_position - Wp->inner.origin.x;
	rectp->origin.y = regp->y_position - Wp->inner.origin.y;
	rectp->corner.x = rectp->origin.x + regp->x_size;
	rectp->corner.y = rectp->origin.y + regp->y_size;
	return rectp;
}
/*
 * addHreg -- add a region to the highlighed regions list
 */
addHreg(rx)
	short rx;
{
	struct hlist *hlp;
	struct TmRegPacket *regp;
	AxRectangle rr;

	regp = &Wp->regions[rx];
	TLOGGER((ATfp, "addHreg() rid=%d\n", regp->r_id));
	hlp = (struct hlist *) ATMmalloc(sizeof(struct hlist));
	hlp->hl_rid = regp->r_id;
	hlp->hl_rect = *regWrect(regp, &rr);
	hlp->hl_link = Wp->hlhead;
	Wp->hlhead = hlp;
}
/*
 * rmHreg -- remove a region from the highlighted regions list
 *
 *			window Wp
 */
rmHreg(rx)
	short rx;
{
	struct hlist *hlp, *prev;
	short rid;

	rid = Wp->regions[rx].r_id;
	TLOGGER((ATfp, "rmHreg() rid=%d\n", rid));
	hlp = Wp->hlhead;
	assert(hlp != NULL);
	while (hlp->hl_rid != rid)
	{
		prev = hlp;
		hlp = hlp->hl_link;
		assert(hlp != NULL);
	}
	if (hlp == Wp->hlhead)
		Wp->hlhead = hlp->hl_link;
	else
		prev->hl_link = hlp->hl_link;
	ATMfree(hlp);
}
/*
 * hiAllReg -- unhighlight or rehighlight all highlighted regions
 *
 *			inverts state of display if in grahics mode
 */
hiAllReg(rectp, onOff)
	AxRectangle *rectp;
	short onOff;
{
	struct hlist *hlp;
	AxRectangle rr;

	for (hlp = Wp->hlhead; hlp != NULL; hlp = hlp->hl_link)
	{
		absRect(&hlp->hl_rect, &rr);	/* to absolute (screen) pixels */
		if (rectXrect(rectp, &rr))
				lrectf(Wp->layer, &rr, R_XOR);
	}
}
