/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: tmwin.c
   Applix ASC version: 20
   Date: Wed Feb 24 21:13:18 1999
*/
#include <stdio.h>
#include <fcntl.h>
#include <aos.h>
#include <paripc.h>
#include <tmi.h>
#include <tmassert.h>

#include <atm.h>
#include <akeys.h>
#include <atmwin.h>
#include <atmreg.h>
#include <atmattr.h>
#include <atmtxtin.h>
#include <elhelper.h>

#include <paratmgr.h>
#include <parwin.h>
#include <parint.h>
#include <bounds.h>
#include <scac.h>
#include <getstr13.h>
#include <tmwin.p>

#define NULLAID     0


struct wind_struct *make_curwin();
struct wind_struct *getActiveWin();
char *strcpy();
int minWinHt();
extern int MapMode;
extern short GrayBack;
extern short LocTmR;        /* local call flag */
extern short LocWid;        /* window_id of a locally created window */
extern char DemoPlayback;
extern short FSWid;
extern short FSHt;
extern short TestMode;
extern short MenulessMode;
extern short Xmpc;
extern short MinWWid, MinWHt;   /* min window size, mils */
extern short Opening;
extern short RastWid, RastHt;
extern short DpyWid;
extern AxRectangle GrayRect;
extern short LastInputWasMouse; /* used to override xpos,ypos on window create */

/* globals */

struct wind_struct *ActWp = NULL;   /* active window pointer */
extern struct wind_struct *StatWp;  /* status line window pointer */

/* global to this module */

STATIC short CurOptLen = 0;
extern short CreCount;      /* creation order counter */
 
extern short ACwin;
void TestForNoWindowsLeft ();

extern int ThimNoWindowsLeft;   /* goes TRUE after terming last window */


/*************************************************************************
function: TermByApp - Terminate all windows owned by a task
*/
void TermByApp (pid)
int pid;
{
    register int ix;

    for ( ix = 1; ix < TM_MAXWIN; ix++ ) {
        if ( (window[ix].busy_flag) && (window[ix].p_id == pid) )
            TmTerm(ix);
    }
}

/* TmTerm function */
void atm_term(mes, mesLen, flag) 
char *mes;
short mesLen;
short flag;
{
    union
    {
        struct STermData term_data;
        char rec_buf[sizeof(struct STermData)];
    } rec_pack;

    union
    {
        struct RTermData term_data;
        char sen_buf[MAX_FM_ATM];
    } sen_pack;

    extern Layer *FrontLayer, *BackLayer;
    short i,j;
    short sp_flag;
    short winID;
    int actWin;     /* window being terminated is the active window */
    struct wind_struct *windowP, *tempWp;
    short ev_fd;
    short res_fd;
    typeErr status;
    int aWinWasActive;

    ATMmesMov(mes, rec_pack.rec_buf, mesLen, sizeof(struct STermData), flag);

    winID = rec_pack.term_data.window_id;
    TLOGGER((ATfp, "win %d\n", winID));
    if ( winExists(winID) )
    {
        status = TMR_OK;

#if BUILD_KANJI
    /** TEMP  */
        windowP = &window[winID];
        UpdateHostWin( windowP );
        Wp = NULL;
#else
        SetWp(&window[winID]);
        windowP = Wp;
#endif
        if (windowP->sub_window != TM_NULWIN)   /* window has subordinate */
        {
            /* cut off subordinate */
            tempWp = &window[Wp->sub_window];
            tempWp->wind_attr.parent_id = TM_NULWIN;
            tempWp->wind_attr.modifiers.sub_window = FALSE;
        }
    }
    else
    {
        status = TMR_NOWIN;
    }
    if ( status != TMR_OK ) tempAbort("attempt to term a non-window");
    windowP->busy_flag = FALSE;
    DestroyAxWindow( windowP );
    getActiveWin( cur_win );
    TestForNoWindowsLeft();
    return;
}  /* atm_term() */

/*************************************************************************
function: TestForNoWindowsLeft - Set global if no windows alive
*/
void TestForNoWindowsLeft ()
{
    register int i;
    for ( i = 1; i < TM_MAXWIN; i++ ) {
        if ( window[i].busy_flag ) {
            ThimNoWindowsLeft = FALSE;
            return;
        }
    }
    ThimNoWindowsLeft = TRUE;
}

/* Reset various flags and release the malloc'ed memory in wind_struct 
    after a window is terminated */

void resWindStruct( windowP )
struct wind_struct *windowP;
{
    ATMfree(windowP->wind_attr.title_line);
    ATMfree((char *)windowP->tabs);
    memset(windowP,0,sizeof(struct wind_struct));
}

/*****************************************************************/
/*
 * TmModify function
 *
 * Following flags, if FALSE, disable bar calculations in tmbounds.c.
 * Due to rounding off, the atm could get wrong pixel position
 * values and leave trash on screen.
 */
short BarVChanged = 1;
short BarHChanged = 1;

/*
 * strEq -- check string equality for length-defined strings
 *
 *              null strings are considered equal
 */
int
strEq(s1, len1, s2, len2)
    char s1[], s2[];
    int len1, len2;
{
    int i;

    if (len1 != len2) return FALSE;
    for (i = 0; i < len1; i++) if (s1[i] != s2[i]) return FALSE;
    return TRUE;
}

/*************************************************************/
/* TmSasideAll function */

void atm_saside(mes, mesLen, flag)
char *mes;
short mesLen;
short flag;

{
    union
    {
        struct SSasideData sa_data;
        char rec_buf[sizeof(struct SSasideData)];
    } rec_pack;

    ATMmesMov(mes, rec_pack.rec_buf, mesLen, sizeof(struct SSasideData), flag);

    saAll();
}

void saAll(){}

/*************************************************************/
/* TmCurWin function */

void atm_curwin(mes, mesLen, flag, mesback)
char *mes;
short mesLen;
short flag;
char *mesback;  /* return msg buffer for local (flag=2) calls THIMBLE-only */

{
    union
    {
        struct RCurWinData curwin_data;
        char sen_buf[sizeof(struct RCurWinData)];
    } sen_pack;

    short ev_fd;
    short res_fd;

    sen_pack.curwin_data.window_id = cur_win;
    sen_pack.curwin_data.status = TMR_OK;
    if (flag==2) {    /* local call */
        memcpy(mesback,sen_pack.sen_buf,sizeof(struct RCurWinData));
        return;
    }
}

/*************************************************************/
/* TmMkCurWin function */

void atm_mk_curwin(mes, mesLen, flag)
char *mes;
short mesLen;
short flag;
{
    union
    {
        struct SMkCurWinData mk_curwin;
        char rec_buf[sizeof(struct SMkCurWinData)];
    } rec_pack;
    short cw;
    short i;

    ATMmesMov(mes, rec_pack.rec_buf, mesLen, sizeof(struct SMkCurWinData),flag);

#if 0
    PCkMkCurWin(rec_pack.rec_buf);
#endif
    cw = rec_pack.mk_curwin.window_id ;
#if DIO_CLIENT  /* temporarily limit check to this version */
    if (cw == ACwin) return;
#endif
    if (cw == TM_NULWIN || winExists(cw))
        make_curwin(cw);
}


/****************************************************************/
/*
 * make_curwin -- activate window
 *
 *      Deactivate the currently active window and then
 *      activate the specified window.
 *
 *      Activating the null window puts the appropriate entries
 *      on the option line.
 *
 *      If the specified window is already active, do nothing, unless
 *      argument is TM_NULWIN, in which case, assume ATM is being
 *      initialized.
 *
 *      Side effects:
 *          - turns mouse cursor on
 *          - sets clipping rectangle to application clipping rectangle
 */
struct wind_struct *make_curwin(winID)
    int winID;
{
    int old = cur_win;
    struct wind_struct *windowP;

    cur_win = winID;
    windowP = getActiveWin( cur_win );
    if ( (winID > 0) && (old != winID) && IsElfInLearnMode() )
        ElfRecord(ELFREC_WIN, windowP->wind_attr.title_line);
    return( windowP );
}

/*****************************************************************/
/*
 * winExists -- does window ID refer to an existing window?
 *
 *  TRUE: yes
 *  FALSE: bad ID, nonexistent window, or TM_NULWIN
 */
int
winExists(winID)
    short winID;
{
    return (winID >= 0 && winID < TM_MAXWIN && window[winID].busy_flag);
}
/*****************************************************************/
/*
 * getOwnWin - get window context if application can access window
 *
 *      sets global Wp
 */
int                 /* boolean: success (access granted) */
getOwnWin(winID)
short winID;
{
    Wp = NULL;
    if (winID <= 0 || winID >= TM_MAXWIN)
        return FALSE;
    SetWp(&window[winID]);
    if (Wp->busy_flag)      /* any existing window other than status line */
    {
        return TRUE;
    }
    else
    {
        Wp = NULL;
        return FALSE;
    }
}
/***************************************************************/
/*
 * get active window and set global pointer Wp
 *
 *      return TRUE - success;  FALSE - no active window (Wp NULL)
 */
struct wind_struct *getActiveWin( current_win )
int current_win;
{
    if (current_win == TM_NULWIN)
    {
        Wp = (struct wind_struct *)NULL;
    }
    else
    {
        Wp = &window[ current_win ];
        UpdateHostWin( Wp );
    }
    return( Wp );
}



#if 0
struct wind_struct *getWp()
{
    if (cur_win == TM_NULWIN) {
        ClearWp();
        return(NULL);
    }
    else {
        SetWp(&window[cur_win]);
        ThimSetHostWinFromWp();
        return(Wp);
    }
}
#endif



