/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: tree.h
   Applix ASC version: 13
   Date: Mon Nov 16 13:53:41 1998
*/
#ifndef _xm_tree_h_
#define _xm_tree_h_

#define NODE_ID(p) ((long) (p))

/* Node Display Information */
typedef struct {
	int xpos;
	int ypos;
	unsigned int width;
	unsigned int height;
	char *icon_pixmap; /* display icon pixmap - User settable */
	char *title; /* display title - User settable */
	unsigned int title_width;
	unsigned int title_height;
	unsigned int pixmap_width;
	unsigned int pixmap_height;
} nodeDispInfo;

/* Tree Node Attributes */
typedef struct {
	unsigned int dd: 1;       /* is a directory displayer node */
	unsigned int selected: 1; /* node has been selected */
	unsigned int spare2: 1;
	unsigned int spare3: 1;
	unsigned int spare4: 1;
	unsigned int spare5: 1;
	unsigned int spare6: 1;
	unsigned int spare7: 1;
} tree_node_attr_t;
	
/* Definition of a Tree Node */
typedef struct treeNode {
	nodeDispInfo info;
	int total_width;  /* accumulated width of this node + children */
	int total_height; /* accumulated height of this node + children */
	int value; /* node value - User settable */
	int level;
	union {
		struct treeNode *parent;/* (level > 0) */
		long *cache; /* (level == 0) IntsArray */
	} u;
	long *sub_nodes; /* IntsArray */
	tree_node_attr_t attr;
} treeNode;

/* tree.c */
extern void SetNodeIconPixmap AXP_((treeNode *nodeP, char *icon_pixmap));
extern void SetNodeValue AXP_((treeNode *nodeP, int value));
extern void SetNodeTitle AXP_((treeNode *nodeP, char *title));
extern treeNode *FindNodeByTitle AXP_((treeNode *anyP, char *title));
extern treeNode *FindNodeByValue AXP_((treeNode *anyP, int value));
extern treeNode *CrtTreeNode AXP_((treeNode *parent));
extern treeNode *FindTopNode AXP_((treeNode *nodeP));
extern treeNode *NodePtrFromBinaryData AXP_((data_binary *dP));
extern int FindNode AXP_((treeNode *topP, treeNode *nodeP));
extern void DelTreeNode AXP_((treeNode *nodeP));
extern void DelTreeSubnodes AXP_((treeNode *nodeP));
extern void ResetTreeDispInfo AXP_((treeNode *topP));
extern void DestroyTree AXP_((treeNode *topP));
extern void FastDestroyTree AXP_((treeNode *topP));
extern void DumpTreeCache AXP_((treeNode *topP));
extern char *ExpandNodeTitle AXP_((treeNode *nodeP, char *sepStr));
extern char *ExpandNodeValue AXP_((treeNode *nodeP, char *sepStr));
extern data_ptr ElfArrayOfTreeNodes AXP_((long *nodep_list));
extern data_ptr ElfArrayOfSubnodes AXP_((treeNode *nodeP));
extern data_ptr ElfFormatOfTreeNode AXP_((treeNode *nodeP));
extern data_ptr TreeDataToElfArray AXP_((treeNode *nodeP));

#endif /* not _xm_tree_h_ */



/* GNU Emacs settings - This (including the "\n^L") MUST be the last thing
 *                      in this file!
 * Local variables:
 * mode: c
 * tab-width: 4
 * end:
 */
