/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: usrevent.h
   Applix ASC version: 8
   Date: Mon Nov 16 13:53:35 1998
*/
#include <apheader.h>

/* event types */
#define PICK_EVENT    0
#define STROKE_EVENT  1
#define REGEN_EVENT   2
#define CONFIG_EVENT  3
#define PTR_EVENT     4
#define DROP_EVENT    5

/* mouse flags for picks and strokes */
#define L_BUTTON 0
#define M_BUTTON 0
#define R_BUTTON 0

/* horizontal/vertical line flags for picks and strokes */
#define N_LINE 0
#define V_LINE 1
#define H_LINE 2

/* the event structures passed to the apps */
typedef struct pick_event_t {
	int x;
	int y;
	int type;
	int button;
	int quickClick;
	unsigned int shifted : 1;
	unsigned int controlled : 1;
	unsigned int helpBit : 1;
} pick_event_t;

typedef struct stroke_event_t {
	int x;
	int y;
	int type;
	int button;
	int hvStatus;
	unsigned int shifted : 1;
	unsigned int controlled : 1;
	unsigned int firstPoint : 1;
	unsigned int lastPoint : 1;
	unsigned int boundary : 1;
	unsigned int echoed : 1;
	unsigned int clipped : 1;
} stroke_event_t;

typedef struct pointer_event_t {
	int x, y;
	unsigned int shifted : 1;
	unsigned int controlled : 1;
} pointer_event_t;

typedef struct regen_event_t {
	int x1, y1;
	int x2, y2;
} regen_event_t;

typedef struct config_event_t {
	int wid, hyt;
} config_event_t;

typedef struct drop_event_t {
	unsigned int mods;
	int  filetype;
	char name[STRPATH+1];
	char machine[STRLONG+1];
} drop_event_t;

typedef struct sys_event_t {
	int type;
	int wid;	/* ATM window id */
	union {
		stroke_event_t  stroke;
		pick_event_t    pick;
		regen_event_t   regen;
		config_event_t  config;
		pointer_event_t ptr;
		drop_event_t drop;
	} ev;
} sys_event_t;

#define UNCHANGED 0
#define UNIT_INC  1
#define UNIT_DEC  2
#define PAGE_INC  3
#define PAGE_DEC  4
#define NEW_VAL   5

typedef struct scroll_event_t {
	int h_type, v_type;
	int h_value;
	int v_value;
	int wid;	/* ATM window id */
} scroll_event_t;
