/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: wchars.h
   Applix ASC version: 26
   Date: Fri Jan 22 21:12:20 1999
*/
#if !BUILD_WIDE    /* **********************     {  **********************   */

  /* define a set of dummy macros for the non-wide builds */
#define IS_WIDE_CHAR(p) (  0 )
#define IS_MULTI_BYTE(p) ( 0 )
#define IS_NARROW_KANA(p) ( 0 )
#define IS_ORIENT_CHAR( p )  ( 0 )
#define IS_SHIFT_JIS( p ) ( 0 )

#define MAKE_WIDE_FROM_MULTI(inP) (*inP)

/* we set the high bit on the byte to make sure that we do not generate any */
/*   'sensitive' characters like quote or backslash by just clipping off    */
/*   a wide character                                                       */
#define MAKE_MULTI_FROM_WIDE( inP, outP ) *outP++ = \
                       ((((unsigned)*inP) & (unsigned)0xFF) | (unsigned)0x80);

/* how many bytes result from converting a wide char to a multi-byte string */
#define SIZE_MULTI_FROM_WIDE 1




#else     /*  } ****************** BUILD_WIDE  ****************** {  */
/* how many bytes result from converting a wide char to a multi-byte string */
#define SIZE_MULTI_FROM_WIDE sizeof(wchar)

#if BUILD_KANJI    /*   {  ---------------------------  */
#define IS_WIDE_CHAR(p) ( ((unsigned)p > (unsigned)256) && \
										((unsigned)p < (unsigned)0x8000) )
#define IS_NARROW_KANA(p) ( ((unsigned)(p) > (unsigned)0x8E00) && \
										((unsigned)(p) <= (unsigned)0x8EFF) )
#define IS_ORIENT_CHAR( p ) ((unsigned)p > (unsigned)256)

/*JFP encodes the characters on disk (basically as EUC), with the high bit of */
/* both bytes set.  The high bit is not actually used to reference the */
/* glyph in the font however since the font is a JIS font           */
#define IS_MULTI_BYTE(p) \
  ((((unsigned)(*(p) & (unsigned)0xFF) >= (unsigned)0xa1) && \
    ((unsigned)(*(p) & (unsigned)0xFF) <= (unsigned)0xf4) && \
    ((unsigned)(*((p)+1) & (unsigned)0xFF) >= (unsigned)0xa1) && \
    ((unsigned)(*((p)+1) & (unsigned)0xFF) <= (unsigned)0xfe) ) || \
    ((unsigned)(*(p) & (unsigned)0xFF) == (unsigned)0x8E) )
       /* the last clause is because narrow kana chara are "0x8Exx" */  

#if BUILD_KANJI
#define IS_SHIFT_JIS( p ) \
   ((((((unsigned)(*(p) & (unsigned)0xFF) >= (unsigned)0x81) && \
         ((unsigned)(*(p) & (unsigned)0xFF) <= (unsigned)0x9f)) ||  \
     (((unsigned)(*(p) & (unsigned)0xFF) >= (unsigned)0xe0) && \
        ((unsigned)(*(p) & (unsigned)0xFF) <= (unsigned)0xef))) &&  \
     (((unsigned)(*((p)+1) & (unsigned)0xFF) >= (unsigned)0x40) && \
        ((unsigned)(*((p)+1) & (unsigned)0xFF) <= (unsigned)0xfe))) || \
     (((unsigned)(*(p) & (unsigned)0xFF) >= 0xa1) && \
        ((unsigned)(*(p) & (unsigned)0xFF) <= 0xdf)) )
   /* the last two lines are for narrow kana characters */

#else
#define IS_SHIFT_JIS( p ) ( 0 )
#endif


#define MAKE_WIDE_FROM_MULTI(inP) \
   ( ((unsigned)(*(inP) & (unsigned)0xFF) == (unsigned)0x8E) ? \
     ((unsigned)0x8E00 | (unsigned)(*(inP+1) & (unsigned)0xFF)) : \
     ((unsigned)(*inP & (unsigned)0x7F) << 8) | (unsigned)(*(inP+1) & (unsigned)0x7F) ) 


#define MAKE_MULTI_FROM_WIDE(inP,outP ) \
     { *outP++ = ((unsigned)(*inP >> 8) & (unsigned)0xFF) | (unsigned)0x80; \
       *outP++ = (unsigned)(*inP & 0xFF) | (unsigned)0x80; }


#elif BUILD_KOREAN   /*   }  ------ KANJI {  -------------------------------  */

    /* The Korean KSC5601 character set starts at 0x2121 and goes to 0x7d7e */
    /* when stored as multi-byte we assume that the high bit is always set */
    /*   on both bytes                                            */
#define IS_WIDE_CHAR(p) ((unsigned)p > (unsigned)256)
#define IS_NARROW_KANA(p) ( 0 )
#define IS_ORIENT_CHAR( p ) ((unsigned)p > (unsigned)256)

#define IS_MULTI_BYTE(p) \
  (((unsigned)(*(p) & (unsigned)0xFF) >= (unsigned)0xa1) && \
    ((unsigned)(*(p) & (unsigned)0xFF) <= (unsigned)0xfd) && \
    ((unsigned)(*((p)+1) & (unsigned)0xFF) >= (unsigned)0xa1) && \
    ((unsigned)(*((p)+1) & (unsigned)0xFF) <= (unsigned)0xfe) ) 


#if HYUNDAI && SUN4 && !SUN4S
    /** The SunOS X11R5 fonts leave the high bits set on the characters */
#define MAKE_WIDE_FROM_MULTI(inP) \
     (  ((unsigned)(*inP << 8) | \
        ((unsigned)(*(inP+1)) )))
#else
    /** The Solaris fonts clip the high bits from both bytes */
#define MAKE_WIDE_FROM_MULTI(inP) \
     (  ((unsigned)(*inP & (unsigned)0x7F) << 8) | \
        ((unsigned)(*(inP+1) & (unsigned) 0x7F) )  )
#endif



#define MAKE_MULTI_FROM_WIDE(inP,outP ) \
     { *outP++ = ((unsigned)(*inP >> 8) & (unsigned)0xFF) | (unsigned)0x80; \
       *outP++ = (unsigned)(*inP & 0xFF) | (unsigned)0x80; }

#elif BUILD_CHINA   /*  } ----- KOREAN {  -------------------------------  */

    /* The Chinese GB-2312 character set starts at 0x2121 and goes to 0x777e */
    /* when stored as multi-byte we assume that the high bit is always set */
    /*   on both bytes                                            */
#define IS_WIDE_CHAR(p) ((unsigned)p > (unsigned)256)
#define IS_NARROW_KANA(p) ( 0 )
#define IS_ORIENT_CHAR( p ) ((unsigned)p > (unsigned)256)

#define IS_MULTI_BYTE(p) \
  (((unsigned)(*(p) & (unsigned)0xFF) >= (unsigned)0xa1) && \
    ((unsigned)(*((p)+1) & (unsigned)0xFF) >= (unsigned)0xa1) && \
    ((unsigned)(*((p)+1) & (unsigned)0xFF) <= (unsigned)0xfe) ) 


#define MAKE_WIDE_FROM_MULTI(inP) \
     (  ((unsigned)(*inP & (unsigned)0x7F) << 8) | \
        ((unsigned)(*(inP+1) & (unsigned) 0x7F) )  )


#define MAKE_MULTI_FROM_WIDE(inP,outP ) \
     { *outP++ = ((unsigned)(*inP >> 8) & (unsigned)0xFF) | (unsigned)0x80; \
       *outP++ = (unsigned)(*inP & 0xFF) | (unsigned)0x80; }

#else    /*  }  ----- CHINA { --------------------  */
		/* non-orient wide build */

#define IS_WIDE_CHAR(p) ( ((unsigned)p > (unsigned)256) )
#define IS_NARROW_KANA(p) (0)
#define IS_ORIENT_CHAR(p) (0)
#define IS_MULTI_BYTE(p) (0)

#define MAKE_MULTI_FROM_WIDE(inP,outP ) \
     { *outP++ = ((unsigned)(*inP >> 8) & (unsigned)0xFF) | (unsigned)0x80; \
       *outP++ = (unsigned)(*inP & 0xFF); }
	   
#define MAKE_WIDE_FROM_MULTI(inP) \
     (  ((unsigned)(*inP & (unsigned)0xFF) << 8) | \
        ((unsigned)(*(inP+1) & (unsigned) 0xFF) )  )

#endif     /*  non-orient build }  */

#endif     /*  BUILD_WIDE  }  */




/** define the return values from HandleOffTheSpot() */
#define NO_SPOT_INTEREST 0
#define SPOT_HANDLED     1
#define SPOT_RETURNS_STRING -1

#define KANJI_STATUS_PAD 7    /* for display of ststus at bottom of dialogs */

