/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: widgfmt.h
   Applix ASC version: 16
   Date: Mon Nov 16 13:53:46 1998
*/
#ifndef _WIDGFMT_H
#define _WIDGFMT_H

/* gadget_set_t stores gadget_size_t instances for a particular
 * dialog box
 */
typedef struct gadget_set_t {
	int newEdgeId;
	axhtab gadgets;
	axconnC *scope;
} gadget_set;

/* gadget_size_t describes the 'realized' size of an elf-visible widget
 */
typedef struct gadget_size_t {
	unsigned int updateElf : 1;     /* will need to update the elf data */
	unsigned int fixedWidth : 1;
	unsigned int fixedHeight : 1;
	int gadgId;                     /* id of this gadget_size instance */
	struct gadget_size_t *edges[4];    /* for containers, refs to edges */
	struct gadget_size_t *containerTop; /* for container edges, ref to whole */
	int gadgetType;                 /* elf-level gadget type */
	int gadgetMinorType;
	axconn *conn;                   /* stores alignment relationships */
	r32 minsize;                    /* minimum allowable realized size */
	r32 proposedsize;               /* proposed new size */
	r32 currentsize;                /* realized size as of last adjustment */
} gadget_size;

typedef struct new_gsize_t {
	gadget_size sizeData;
	axconnC *scope;
} new_gsize;

/* align_link_data_t carries details for an
 * "alignment link" between two widgets
 */
    /* these alignTypes are one-to-one with the align link types 
	 */
enum alignType {
	EDGE_TO_EDGE =    1,
	ROW_COL =         2,
	BUMPER =          3,
	SAME_WIDTH =      4,
	SAME_HEIGHT =     5,
	FIXED_WIDTH =     6,
	FIXED_HEIGHT =    7,
};  

enum edgeType {
	OUTER_RIGHT =      1,
	OUTER_LEFT =       2,
	OUTER_TOP =        3,
	OUTER_BOTTOM =     4,
	INNER_RIGHT =      5,
	INNER_LEFT =       6,
	INNER_TOP =        7,
	INNER_BOTTOM =     8,
	CENTER =           9,
};

typedef struct align_link_data_t {
	enum alignType type;
	int minDistance;
	int maxDistance;
	enum edgeType toEdge;
	enum edgeType fromEdge;
} align_link_data;

/* size_change_t expresses a proposed change in size or
 * position of a gadget_size instance
 */
typedef struct size_change_t {
	int callerType;                  /* e.g., DLG_ALIGN */
	int callerMinorType;             /* e.g., HORIZONTAL */
	r32 currentsize;
	int leftChange;
	int rightChange;
	int topChange;
	int bottomChange;
	align_link_data *linkData;       /* can be null */
} size_change;

/* linkname_t is used for the lookup names of individual links
 */
typedef struct linkname_t {
	enum alignType type;
	int toid;
	enum edgeType toedge; 
} linkname;

#endif /* _WIDGFMT_H */

