/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: widgtcom.c
   Applix ASC version: 137
   Date: Thu May  6 21:12:07 1999
*/
#if AX_X11
#include <X11/keysym.h>
#endif
#include <basicwin.h>
#include <dwt2xm.h>
#include <aos.h>
#include <xmlib.h>
#include <elxwins.h>
#include <xmpixel.h>
#include <xmdlged.h>
#include <xmtable.h>
#include <editmask.h>
#include <intsarr.h>

/* for option menu on GE ribbon */
#define AXSAB_16 \
    (NXLT("-applix-axsab-bold-r-normal--16-160-75-75-p-65-iso8859-1"))

#define THREE_STATE_TOGGLE 3 /* Note must be > XmHORIZONTAL !! */

int AxWindowCreationCounter = AX_WINDOW_UID_BASE;  /* inc uid for each win */
                                                   /*  w->window_uid */
extern Display *Dpy;
extern AXWidget WindowList;
extern int NumWidgets;
extern int WidgetNeedsUpdated;  /* Set when a widget needs repainted */
extern AXWidget TotalWidgetsList;       /* for resource tracking */
extern XFontStruct *AxDlgMainFont;
extern int SmallCharHeight;

extern int AlwaysWide;      /* always make room for toggles/checks */
extern int AxNoFocusManage;  /* if TRUE, don't set focus unless asked to */
extern int NoTransient;/* if FALSE, use window mgr's Transient for dialogs */
extern int ShowAccelerators;
extern int ThimBellVolume;
extern int UseChecksForToggles; /* use checkmarks rather than toggle boxes */
extern int Thim2dMode;
extern int DoubleMillisec;
extern int AxMarkMenuMnemonics; /* if TRUE: underline menu key mnemonics */
extern long WinMask;  
extern XSetWindowAttributes WinAtt;
extern Cursor ThimPointerCursor;
extern Cursor ThimBusyCursor; 
extern char *IconSizeStr; 

#if AX_X11
extern AXWidget_menu AxPulldownActive; /* non-zero while menubar pulldown
                                            is active */
#endif

struct XMlibStr {
    PIXEL_VALUE FGpixel;    /* foreground color for app work areas */
    PIXEL_VALUE BGpixel;    /* backround color for app work areas */
    int FullRegenForUpdate;  /* Set by UnMapWidget,
                                            hint for WidgetNeedsUpdated */
};
extern struct XMlibStr xmlibStr; 


#include <widgtcom.p>



int 
SetNWidgetName(w, name, len)
register AXWidget w;
register char *name;
register int len;
{
    if ((!w) || (!name) || (len <= 0))
        return (-1);

    FreeWidgetName(w);
    w->name = (mbchar *) malloc(len + 1);
    MBstrncpy(w->name, name, len);
    w->name[len] = 0;
    return (0);
}


int SetWidgetName(w, name)
register AXWidget w;
register char *name;
{
    if (!w || !name)
        return (-1);
    if (w->name == (mbchar *)name) /* same pointer == same name */
        return (0);
    if (w->name && !strcmp(w->name,name)) /* same name */
        return(0);
    FreeWidgetName(w);
    w->name = (mbchar *) malloc(strlen(name) + 1);
    MBstrcpy(w->name, name);

    return (0);
}

static int
read_accelerator(arr, ix)
data_ptr arr;
int ix;
{
	mbchar *sAcc = NULL;
	if (ix >= 0) {
		sAcc = AxStrFromArray(arr, ix);
	}
	else {
		sAcc = AxStrFromDataPtr(arr);
	}
	if (0 == sAcc || 0 == *sAcc) {
		return 0;
	}
	else {
		/* take the 1st character as the accelerator
		 */
		return (Atoupper((unsigned int) *sAcc));
	}
}

int
SetWidgetAccelerator(ww, info)
AXWidget ww;
data_ptr info;               /* typically single str; can be an array */
{
	if (!info) {
		return FALSE;
	}
	switch (ww->type) {
	case W_RADIO:
		if (ww->auxData && AxIsArray(info)) {
			wComposite *cp;
			int ax = 0;
			/* first count how many toggles this radio array has;
			 * accept new accelerator keys only if passed array has
			 * the same #elements as this radio's #toggles
			 */
			for (cp = ww->auxData->compositeList; cp; cp = cp->next) {
				if (cp->type == WLINKDOWN && W_TOGGLE == cp->w->type) {
					ax++;
				}
			}
			if (ax == AxArraySize(info)) {
				ax = 0;
				for (cp = ww->auxData->compositeList; cp; cp = cp->next) {
					if (cp->type == WLINKDOWN && W_TOGGLE == cp->w->type) {
						cp->w->accelerator_key =
							read_accelerator(info, ax++);
					}
				}
			}
		}
		break;
	case W_RECT:
		if (AxIsArray(info)) {
			AXWidget_rect rw = (AXWidget_rect) ww;
			if (AxArraySize(info) == rw->tabctl.num_layers) {
				int rx;
				for (rx = 0; rx < rw->tabctl.num_layers; rx++) {
					rw->tabctl.layerlist[rx].accelerator_key =
						read_accelerator(info, rx);
				}
			}
		}
		else {
			ww->accelerator_key = read_accelerator(info, -1);
		}
		break;
	default:
		ww->accelerator_key = read_accelerator(info, -1);
		break;
	}
	return TRUE;
}

int 
GetWidgetSize(type)
int type;
{
    int w_size;

    switch (type)
    {
        case W_DIALOG:
            w_size = sizeof(axw_dialog);
        break;
        case W_LABEL:
            w_size = sizeof(axw_label);
        break;
        case W_TOGGLE:
            w_size = sizeof(axw_toggle_button);
        break;
        case W_RADIO:
            w_size = sizeof(axw_radio_box);
        break;
        case W_BUTTON:
            w_size = sizeof(axw_push_button);
        break;
        case W_OPTION:
            w_size = sizeof(axw_option_menu);
        break;
        case W_LIST:
            w_size = sizeof(axw_list_box);
        break;
        case W_SCROLL:
            w_size = sizeof(axw_scrollbar);
        break;
        case W_ENTER:
            w_size = sizeof(axw_entry_field);
        break;
        case W_ICON:
            w_size = sizeof(axw_icon);
        break;
        case W_HGROOVE:
            w_size = sizeof(axw_any);
        break;
        case W_POPUP:
            w_size = sizeof(axw_popup);
        break;
        case W_RECT:
            w_size = sizeof(axw_rect);
        break;
        case W_MAIN:
            w_size = sizeof(axw_main);
        break;
        case W_WORK:
            w_size = sizeof(axw_work);
        break;
        case W_MENUBAR:
            w_size = sizeof(axw_menu);
        break;
        case W_PALETTE:
            w_size = sizeof(axw_palette);
        break;
        case W_RULER:
            w_size = sizeof(axw_ruler);
        break;
        case W_STATUS:
            w_size = sizeof(axw_status);
        break;
        case W_DEBUG:
            w_size = sizeof(axw_debug);
        break;
        case W_CORNER:
            w_size = sizeof(axw_corner);
        break;
        case W_TOOLBAR:
            w_size = sizeof(axw_toolbar);
        break;
        case W_EDITBOX:
            w_size = sizeof(axw_editbox);
        break;
        case W_SPLITTER:
            w_size = sizeof(axw_splitter);
        break;
        case W_SCALE:
            w_size = sizeof(axw_scale);
        break;
        case W_COLORPAL:
            w_size = sizeof(axw_colorpal);
        break;
        case W_RIBBON:
            w_size = sizeof(axw_ribbon);
        break;
        case W_GE_XRULER:
            w_size = sizeof(axw_ge_ruler);
        break;
        case W_GE_YRULER:
            w_size = sizeof(axw_ge_ruler);
        break;
        case W_ALIGN:
            w_size = sizeof(axw_any);
        break;
        case W_TABLE:
            w_size = sizeof(axw_table);
        break;
#if AX_X11
        case W_KANJI_KANA:
            w_size = sizeof(axw_kanji_kana);
        break;
        case W_KANJI_STATUS:
            w_size = sizeof(axw_kanji);
        break;
#endif
        case W_ROWCOL:
            w_size = sizeof(axw_rowcol);
        break;
        case W_CANVAS:
            w_size = sizeof(axw_canvas);
        break;
        case W_TRAY:
            w_size = sizeof(axw_tray);
        break;
        case W_PLATES:
            w_size = sizeof(axw_plates);
        break;
        case W_CHOOSE:
            w_size = sizeof(axw_color_chooser);
        break;
        case W_PAGEBAR:
            w_size = sizeof(axw_page);
        break;
        case W_BUTBAR:
            w_size = sizeof(axw_buttonbar);
        break;
        case W_COMBO_BOX:
            w_size = sizeof(axw_combo_box);
        break;
        default:
            w_size = sizeof(axw_any);
        break;
    }
    return(w_size);
}

/***********************************************************************
function: InitWidget - Alloc widget
*/
AXWidget
InitWidget(type)
register int type;
{
    register AXWidget w;
    register int size;

    size = (int)GetWidgetSize(type);

    w = (AXWidget) calloc(1, size);
    w->type = type;
    return ((AXWidget) w);
}

/***********************************************************************
function: CrtWidget - Alloc widget, setup header
*/
AXWidget CrtWidget(type, parent, name)
register int type;
register AXWidget parent;    /* parent not necessarily of type W_MAIN */
register char *name;
{
    register AXWidget w;

    w = (AXWidget)InitWidget(type);
#if AXMS
    if ( w->type == W_PALETTE )
        w->initial_paint = TRUE;
#endif
    w->prev = NULL;
    w->parent = (AXWidget_main)parent;
#if BUILD_ORIENT && !AXMS
    if (w->type == W_MAIN)
    {
        ((AXWidget_main)w)->conversion_start = -1;
        ((AXWidget_main)w)->conversion_end = -1;
        ((AXWidget_main)w)->conv_phrase_start = -1;
        ((AXWidget_main)w)->conv_phrase_end = -1;
        ((AXWidget_main)w)->conv_change_start = -1;
        ((AXWidget_main)w)->conv_change_end = -1;
    }
#endif
    if (parent)
        w->display = w->parent->display;
    else
        w->display = TRUE;

    if (parent) {
        if (type == W_ICON) {
            /* thread at head of list */
            w->nextChild = w->parent->widgets;
            w->parent->widgets = w;
        } else {
            AXWidget tw;
            /* thread downstream of any icon widgets */
            /* Note: if this is a button on a toolbar we will be threading */
            /*       this to the toolbar widget.  There are portions of the */
            /*       code which still reference the toolbar buttons via the */
            /*       "widgets" chain, instead of via the toolBarInfo array. */
            /*       This incorrect use does not occur on Windows.- jon     */
#if AXMS
            if ( w->parent->type == W_TOOLBAR )
                goto do_name;
#endif

            tw = w->parent->widgets;
            while (tw && tw->nextChild && tw->type == W_ICON) {
                tw = tw->nextChild;
            }
            if (tw) {
                w->nextChild = tw->nextChild;
                tw->nextChild = w;
            } else {
                w->nextChild = w->parent->widgets;
                w->parent->widgets = w;
            }
        }
    }
do_name:
    SetNWidgetName(w, name, MAX_NAME);
    w->taskOwner = THIMpid;

    w->update_widget = TRUE;
    WidgetNeedsUpdated = TRUE;

#if !AXMS               /* MS Windows list boxes have automatic scroll bars */
    if (w->type == W_LIST)
    {
        AXWidget_list wl;

        wl = (AXWidget_list)w;
        wl->scroll_hist = (scroll_history *) calloc(1, sizeof(scroll_history));
        wl->curPos = -1;
        wl->maxWidth = -1;
    }
#endif

    if ( w->type == W_DIALOG )
        ((AXWidget_dialog) w)->window_uid = ++AxWindowCreationCounter;


    /* Thread to total widgets list for resource tracker */
    w->next = TotalWidgetsList;
    if (w->next) {
        w->next->prev = w;
    }
    TotalWidgetsList = w;
    NumWidgets++;
    return ((AXWidget) w);
}




/** taken from xmlist.c **/
int ListStrsMaxWidth(w)
AXWidget_list w;
{
    int ix, line_width, max_width;
    char *p;


    /* find longest item in list */
    max_width = 0;
    for ( ix = 0; ix < w->num_strs; ix++ ) {
        p = w->strs[ix];
#if AXMS
        /* for now? */
        line_width = strlen( p );
#else        
        line_width = IconedStringWidth(w, p);
#endif
        if (line_width > max_width)
            max_width = line_width;
    }
    return(max_width);
}
/**                      **/

int ComboStrsMaxWidth(w)
AXWidget_combo_box w;
{
    int ix, line_width, max_width;
    char *p;


    /* find longest item in list */
    max_width = 0;
    for ( ix = 0; ix < w->num_buttons; ix++ ) {
        p = w->strs[ix];
#if AXMS
        /* for now? */
        line_width = strlen( p );
#else        
        line_width = ComboStringWidth(w, p);
#endif
        if (line_width > max_width)
            max_width = line_width;
    }
    return(max_width);
}

void
xmSetWidgetDisplay(ww, display)
AXWidget ww;          /* in/out */
int display;          /* bool: TRUE - display widget, FALSE - hide widget */
{
    wComposite *cp;
    
    if (!ww) {
        return;
    }
    
    if (display) {
        if (!ww->mapped) {
            ww->update_widget = WidgetNeedsUpdated = TRUE;
        }
        if (ww->tabctl_linktab) {
            if (xmWidgetIsHidden(ww)) {
                /* widget belongs to a tab-control rect layer that's hidden
                 */
                ww->saved_display = TRUE;
                ww->display = FALSE;
            }
            else {
                ww->display = ww->saved_display = TRUE;
            }
        }
        else {
            ww->display = ww->saved_display = TRUE;
        }
    }
    else {
        if (ww->mapped) {
            ww->parent->regen_widget = TRUE;
            ww->update_widget = WidgetNeedsUpdated = TRUE;
        }
        ww->display = ww->saved_display = FALSE;
    }
    if (ww->auxData) {
        /* keep display state of composite kids in sync; this is
         * done here mostly for widgets with scrollbars.
         */
        for (cp = ww->auxData->compositeList; cp; cp = cp->next) {
            if ((WLINKDOWN == cp->type) && cp->w) {
                cp->w->display = ww->display;
                cp->w->saved_display = ww->display;
            }
        }
    }
}

void SetWidgetFlags( w_any, value )
AXWidget_any w_any;
int value;
{

    xmSetWidgetDisplay(w_any, !(value & WFLG_NODISPLAY));
    
    if (value & WFLG_INACTIVE)
        {
        if (!w_any->disabled && w_any->type == W_LIST)
            {
            /* state toggled */
            w_any->regen_widget = TRUE;
            }
        w_any->disabled = TRUE;
        }
    else
        {
        if (w_any->disabled && w_any->type == W_LIST)
            {
            /* state toggled */
            w_any->regen_widget = TRUE;
            }
        w_any->disabled = FALSE;
        }
    if (value & WFLG_NO_ECHO)
        w_any->no_echo = TRUE;
    else
        w_any->no_echo = FALSE;
    w_any->flags = value;

    return;
}




/* function: SetWidgetColors - set widget or button colors */
void SetWidgetColors( in_w, color_array, color_val, is_foreground )
AXWidget_any in_w;
int color_array[];    /* array of colors for W_ROWCOL */
XtArgVal color_val;   /* single color value for everyting else */
int is_foreground;    /* if TRUE, set foreground, else set background */
{
    int j;
    char abort_msg[200];
    AXWidget_rowcol     w_rowcol;
    AXWidget_pushbutton w_button;
    AXWidget_icon       w_icon;
    AXWidget_list       w_list;
    AXWidget_editbox    w_editbox;
    AXWidget_canvas     w_canvas;


    if ( in_w->type == W_ROWCOL)
        {
        w_rowcol = (AXWidget_rowcol)in_w;
        /*  color buttons for rowcol for ge. */
        for (j = 0; j < w_rowcol->num_buttons; j++)
            {
            if ( is_foreground )
               w_rowcol->buttons[j]->color = color_array[j];
            else
               w_rowcol->buttons[j]->bg_color = color_array[j];
            }
        }

    else if ( in_w->type == W_BUTTON )
        {
        w_button = (AXWidget_pushbutton)in_w;
        w_button->color = color_val;
        }

    else if ( in_w->type == W_ICON )
        {
        w_icon = (AXWidget_icon)in_w;
        if ( is_foreground )
            w_icon->color = color_val;
        else
            w_icon->bg_color = color_val;
        }

    else if ( in_w->type == W_LIST )
        {
        w_list = (AXWidget_list)in_w;
        w_list->color = color_val;
        }

    else if ( in_w->type == W_EDITBOX )
        {
        w_editbox = (AXWidget_editbox)in_w;
        w_editbox->color = color_val;
        }
    else if ( in_w->type == W_CANVAS )
        {
        w_canvas = (AXWidget_canvas)in_w;
        w_canvas->color = color_val;
        }
    else {
        sprintf(abort_msg, NXLT("Attempt to modify color on widget type %d"),
                in_w->type);
        tempAbort(abort_msg);
    }
    return;
}


void SetWidgetItems(in_w, value)
AXWidget_any in_w;
char **value;
{
    int j, size;
    char *p;
    AXWidget_list w_list;
    AXWidget_popup  w_popup;
    AXWidget_rowcol w_rowcol;
    AXWidget_combo_box w_combo;

    size = sizeof( axw_push_button );

   if ( in_w->type == W_POPUP )
       {
       w_popup = (AXWidget_popup)in_w;
       for ( j = 0; j < w_popup->num_buttons; j++ )
           {
           w_popup->buttons[j] = (AXWidget_pushbutton )(calloc(1, size));
           SetWidgetName((AXWidget)w_popup->buttons[j], value[j] );
           }
       return;
       }
   else if ( in_w->type == W_LIST )
        {
        w_list = (AXWidget_list)in_w;
        w_list->regen_widget = TRUE;
        for ( j = 0; j < w_list->num_strs; j++ )
            {
            p = value[j];
            w_list->strs[j] = (char *) strcpy(malloc(strlen(p)+1), p);
            }
        if ( w_list->list_hscroll )
            {
            /* get max text width */
            w_list->maxLen = (short) ListStrsMaxWidth(w_list);
            }
        else
            w_list->maxLen = 0;
        }
   else if ( in_w->type == W_ROWCOL )
        {
        w_rowcol = (AXWidget_rowcol)in_w;
        w_rowcol->regen_widget = TRUE;
        if ( w_rowcol->num_buttons )
            w_rowcol->strs = (char **) calloc 
                (1, w_rowcol->num_buttons * sizeof(char *));
        for ( j = 0; j < w_rowcol->num_buttons; j++ )
            {
            p = value[j];
            w_rowcol->strs[j] = (char *) strcpy(malloc(strlen(p)+1), p);    
            }
        }
    else if (in_w->type == W_COMBO_BOX)
        {
        w_combo = (AXWidget_combo_box) in_w;
        w_combo->regen_widget = TRUE;
        if (w_combo->num_buttons)
            {
            for (j = 0; j < w_combo->num_buttons; j++)
                {
                p = value[j];
                w_combo->strs[j] = (char *) strcpy(malloc(strlen(p) + 1), p);
                }
            }
        w_combo->maxLen = (short) ComboStrsMaxWidth(w_combo);
        }
    else if ( in_w->type == W_BUTBAR )
        {
        AXWidget_buttonbar w_butbar;
        w_butbar = (AXWidget_buttonbar)in_w;
        w_butbar->regen_widget = TRUE;
        for ( j = 0; j < w_butbar->num_buttons; j++ )
            {
            p = value[j];
            w_butbar->buttons[j] = (AXWidget_pushbutton) 
                strcpy(malloc(strlen(p)+1),p);
            }
        }
    else if ( in_w->type == W_RADIO )
        {
        AXWidget_radio w_radio;
        w_radio = (AXWidget_radio)in_w;
        w_radio->regen_widget = TRUE;
        for ( j=0; j<w_radio->num_buttons; j++ )
            {
            p = value[j];
            w_radio->buttons[j] = (AXWidget_togglebutton)
                strcpy(malloc(strlen(p)+1),p);
            }
        }
    else if ( in_w->type == W_ROWCOL )
        {
        AXWidget_rowcol w_rowcol;
        w_rowcol = (AXWidget_rowcol)in_w;
        w_rowcol->regen_widget = TRUE;
        for ( j=0; j<w_rowcol->num_buttons; j++ )
            {
            p = value[j];
            w_rowcol->buttons[j] = (AXWidget_pushbutton)
                strcpy(malloc(strlen(p)+1),p);
            }
        }
    else if ( in_w->type == W_OPTION )
        {
        AXWidget_option w_option;
        w_option = (AXWidget_option)in_w;
        w_option->regen_widget = TRUE;
        for ( j=0; j<w_option->num_buttons; j++ )
            {
            p = value[j];
#if AXMS
            w_option->strs[j] = strcpy(malloc(strlen(p)+1),p);
#else
            w_option->buttons[j] = (AXWidget_togglebutton)
                strcpy(malloc(strlen(p)+1),p);
#endif
            }
        }
}    

void SetWidgetItemCount(in_w, value)
AXWidget in_w;
int value;
{
    int j, sz;
    AXWidget_list w_list;
    AXWidget_popup w_popup;
    AXWidget_rowcol w_rowcol;
    AXWidget_buttonbar w_buton;
    AXWidget_combo_box w_cbox;
    AXWidget_option w_omenu;

    if ( in_w->type == W_LIST )
    {
        w_list = (AXWidget_list)in_w;
        if ( w_list->strs )
        {
            for ( j = 0; j < w_list->num_strs ; j++ )
                free( w_list->strs[j] );
            free( w_list->strs );
        }

        w_list->num_strs = value;
        if ( value == 0 )
           value = 1;    /* don't allocate nothing */
        w_list->strs = (char **)calloc(1, (value * sizeof(char *)) );
    }
    else
    {
        if ( in_w->type == W_ROWCOL )
        {
            w_rowcol = (AXWidget_rowcol)in_w;
            if ( w_rowcol->strs )
            {
                /* Needed here because if item-count changes, we
                   will not know how many strs to free in the
                   XmNitems case for W_ROWCOL. */
                for ( j = 0; j < w_rowcol->num_buttons; j++ )
                    {
                    if ( w_rowcol->strs[j] )
                        free( w_rowcol->strs[j] );
                    w_rowcol->strs[j] = (char *) NULL;
                   }
                free( w_rowcol->strs );
                w_rowcol->strs = (char **) NULL;
            }

            if ( w_rowcol->buttons )
            {
                for ( j = 0; j < w_rowcol->num_buttons ; j++ )
                {
                    FreeWidgetName((AXWidget)w_rowcol->buttons[j] );
                    free( w_rowcol->buttons[j] );
                    w_rowcol->buttons[j] = (AXWidget_pushbutton ) NULL;
                }
                free( w_rowcol->buttons );
                w_rowcol->buttons = (AXWidget_pushbutton *) NULL;
            }

            w_rowcol->num_buttons = value;
            w_rowcol->buttons = (AXWidget_pushbutton*)calloc(1,
                                          w_rowcol->num_buttons*sizeof(char*));
        }
        else if ( in_w->type == W_POPUP )
        {
            w_popup = (AXWidget_popup)in_w;
            if ( w_popup->buttons )
            {
                for ( j = 0; j < w_popup->num_buttons; j++ )
                    {
                    FreeWidgetName( (AXWidget)w_popup->buttons[j] );
                    free( w_popup->buttons[j] );
                    w_popup->buttons[j] = (AXWidget_pushbutton ) NULL;
                    }
                free( w_popup->buttons );
                w_popup->buttons = (AXWidget_pushbutton *) NULL;
            }

            w_popup->num_buttons = value;
            w_popup->buttons = (AXWidget_pushbutton*)calloc(1,
                                   w_popup->num_buttons*sizeof(char*));
        }

        else if ( in_w->type == W_BUTBAR )
        {
            w_buton = (AXWidget_buttonbar)in_w;
            if ( w_buton->buttons )
            {
                for ( j = 0; j < w_buton->num_buttons; j++ )
                    {
                    FreeWidgetName((AXWidget)w_buton->buttons[j]);
                    free(w_buton->buttons[j]);
                    w_buton->buttons[j] = (AXWidget_pushbutton)NULL;
                    }
                free(w_buton->buttons);
                w_buton->buttons = (AXWidget_pushbutton*)NULL;
            }
            w_buton->num_buttons = value;
            w_buton->buttons = (AXWidget_pushbutton*)calloc(1,
                                    w_buton->num_buttons*sizeof(char*));
        }
        else if ( in_w->type == W_COMBO_BOX )
        {
            w_cbox = (AXWidget_combo_box)in_w;
            if (w_cbox->strs)
            {
                for (j = 0; j < w_cbox->num_buttons; j++)
                {
                    if (w_cbox->strs[j])
                        free(w_cbox->strs[j]);
                }
                free(w_cbox->strs);
            }
            w_cbox->num_buttons = value;
            sz = w_cbox->num_buttons * sizeof(char *);
            w_cbox->strs = (char **) calloc(1, sz);
        }
#if AXMS
        else if ( in_w->type == W_OPTION )
        {
            w_omenu = (AXWidget_option)in_w;
            if (w_omenu->strs)
            {
                for (j = 0; j < w_omenu->num_buttons; j++)
                {
                    if (w_omenu->strs[j])
                        free(w_omenu->strs[j]);
                }
                free(w_omenu->strs);
            }
            w_omenu->num_buttons = w_omenu->num_strs = value;
            sz = w_omenu->num_buttons * sizeof(char *);
            w_omenu->strs = (char **) calloc(1, sz);
        }
#endif
    }
}

/************************************************
function:  SetButtonPixmaps  - for W_ROWCOL only 
*/
void SetButtonPixmaps( w, pixmap_names )
AXWidget_rowcol w;
char **pixmap_names;
{
    int j;
    int size;

#if AXMS
    size = sizeof( axw_push_button );
#else
    size = sizeof( widget_t );
#endif
    for ( j=0; j < w->num_buttons; j++ )
        {
        w->buttons[j] = (AXWidget_pushbutton)calloc(1, size);
        SetWidgetName( (AXWidget)w->buttons[j], pixmap_names[j] );
        }
}


void SetWidgetValues(in_w, value)
AXWidget in_w;
AxArgVal value;
{
    char *p;
    pixcache_t *pix;
    AXWidget_icon  w_icon;

    in_w->update_widget = WidgetNeedsUpdated = TRUE;
    if ( in_w->type == W_ENTER )
        {
        p = (char*)(value);
        SetWidgetString((AXWidget_entry)in_w, p);
        }

    else if ( in_w->type == W_ICON )
        {
        w_icon = (AXWidget_icon)in_w;
        SetWidgetName((AXWidget)w_icon, (char *)value);
        pix = Pixmapper(w_icon->name);
        if ( pix )
            {   /* was 'if ( pix && !w->color )' why?? */
            if ( w_icon->maxWidth > 0 && w_icon->maxHeight > 0 )
                {
                w_icon->width = MIN((unsigned int)pix->width,
                                    (unsigned int)w_icon->maxWidth);
                w_icon->height = MIN((unsigned int)pix->height,
                                    (unsigned int)w_icon->maxHeight);
                }
            else
                {
                w_icon->width = pix->width;
                w_icon->height = pix->height;
                }
            }
        }

    else if ( in_w->type == W_LIST )
        {
        if ( in_w->scroll_pos != value )
            {
            in_w->scroll_pos = value;
            in_w->regen_widget = TRUE;
            }
        }

    else if ( in_w->type == W_SCROLL || in_w->type == W_PAGEBAR)
        {
        in_w->scroll_pos = value;
        }
    else if ( in_w->type == W_COMBO_BOX )
        {
        int ival = 0;
        AXWidget_combo_box cw = (AXWidget_combo_box) in_w;
        
        if (AxIsInt((data_ptr) value))
            {
            ival = AxIntFromDataPtr((data_ptr) value);
            if ((ival < -1) || (ival >= cw->num_buttons))
                {
                ival = 0;
                }
            if (cw->flags & WFLG_EDITABLE && cw->edit_win.text)
                {
                /* clear any "none of the above" (typed-in) text when
                 * accepting an explicit integer value
                 */
                free(cw->edit_win.text);
                cw->edit_win.text = 0;
                cw->edit_win.text_len = 0;
                cw->edit_win.edited = FALSE;
                }
            }
        else if (AxIsString((data_ptr) value) && (cw->flags & WFLG_EDITABLE))
            {
            ival = -1;   /* denotes "none of the above" to the combo code */
            }
        cw->value = ival;
        }
    else
        {
        in_w->value = value;
        }
}





void SetVerticalScrollBar(in_w, newP)
AXWidget in_w;
AXWidget_scroll newP;
{
    AXWidget_scroll *scrollP;
    char msg[80];

    if ( in_w->type == W_MAIN )
        scrollP = &(((AXWidget_main) in_w)->vScrollBar);
    else if ( in_w->type == W_TABLE )
        scrollP = &(((AXWidget_table) in_w)->vScrollBar);
    else
       {
       sprintf(msg, NXLT("SetVerticalScrollBar called with widget type=%d\n"),
                                                               in_w->type);
       taskAbort(msg);
       }
    if ( newP != *scrollP ) {
        if ( *scrollP ) {
            newP->window = (*scrollP)->window;
            /* destroy old widget but leave window intact */
            ClipFromWindowList((AXWidget)*scrollP);
            /* null out window handle so that XtDestroy won't destroy it */
            (*scrollP)->window = 0;
            AxDestroyWidget((AXWidget)*scrollP);
        }
        *scrollP = newP;
        AddToWindowList((AXWidget)newP);
    }
    return;
}


void SetHorizontalScrollBar(in_w, newP)
AXWidget in_w;
AXWidget_scroll newP;
{
    AXWidget_scroll *scrollP;
    char msg[80];

    if ( in_w->type == W_MAIN )
        scrollP = &(((AXWidget_main) in_w)->hScrollBar);
    else if ( in_w->type == W_TABLE )
        scrollP = &(((AXWidget_table) in_w)->hScrollBar);
    else
       {
       sprintf(msg,
               NXLT("SetHorizontalScrollBar called with widget type=%d\n"),
                                                               in_w->type);
       taskAbort(msg);
       }
    if ( newP != *scrollP ) {
        if ( *scrollP ) {
            newP->window = (*scrollP)->window;
            /* destroy old widget but leave window intact */
            ClipFromWindowList((AXWidget)*scrollP);
            /* null out window handle so that XtDestroy won't destroy it */
            (*scrollP)->window = 0;
            AxDestroyWidget((AXWidget)*scrollP);
        }
        *scrollP = newP;
        AddToWindowList((AXWidget)newP);
    }
    return;
}


/**************************************************************
function: SetRibbonWString - Set string on ribbon.
*/
int SetRibbonWString(w, str)
AXWidget_ribbon w;
mbchar *str;
{
    register int real_len;
    int len;

    if ( !w )
        return(0);
    if ( !str )
        str = (mbchar *) "";

    if ( w->wstr != (wchar *) NULL )
        free( w->wstr );
    len = MBstrlen( str );
    w->wstr = (wchar *) malloc( (len + 1) * sizeof(wchar) );
    real_len = MbnToWStr( str, w->wstr, len );

    return(0);
}


/*************************************************************************
function: SetActiveWidget - Set the text cursor to a specific entry field
*/
void SetActiveWidget(in_widget, pending_delete)
AXWidget in_widget;
int pending_delete;
{
    int active_changed;
    AXWidget_entry w_enter;
    AXWidget_editbox w_editbox;
    AXWidget_combo_box w_combo_box;

    if (!in_widget || ((in_widget->type != W_ENTER) &&
               (in_widget->type != W_EDITBOX) &&
               (in_widget->type != W_COMBO_BOX))) {
        
        /* currently this func sets focus only on ENTER, EDITBOX....
       (and now COMBO_BOX too! - EJD =b)
         */
        return;
    }
    
    
    if (!WIN_OF(in_widget))
        return; /* W_ENTER and W_EDITBOX require a valid container. */

    active_changed = ResetActiveTextsExceptFor(in_widget);
    if (in_widget->type == W_ENTER) {
        w_enter = (AXWidget_entry)in_widget;
        if ( pending_delete ) {
            w_enter->pending_delete = TRUE;
            w_enter->curPos = -1;
            active_changed = TRUE;
        }
        /* we need to cast the return of strlen to short because on some
         * systems (e.g. NT), strlen returns unsigned int which causes
         * curPos (a short) to be promoted to unsigned which then produces
         * incorrect results when curPos is -1.          -donato
         */
        else if (  w_enter->curPos > (short)Wstrlen(w_enter->wstr) ) {
            /* someone shortened the string */
            w_enter->curPos = MIN((int)w_enter->curPos,Wstrlen(w_enter->wstr));
            w_enter->pending_delete = FALSE; /* sanity */
            active_changed = TRUE;
        }
        else if ( -1 == w_enter->curPos ) {
            if ( w_enter->wstr ) {
                w_enter->curPos = Wstrlen(w_enter->wstr);  /* default to end */
            }
        }

        if ( active_changed ) {
            /* we do this now for the sake of PaintEnter */
            w_enter->hasFocus = TRUE;
            PaintEnter(w_enter, NULL, TRUE, TRUE);
        }
    }
    else if (in_widget->type == W_EDITBOX) {
        w_editbox = (AXWidget_editbox)in_widget;
        EditBoxDrawCursor(w_editbox, TRUE);
    }
    else if (in_widget->type == W_COMBO_BOX) {
      w_combo_box = (AXWidget_combo_box)in_widget;
      if (w_combo_box->flags & WFLG_EDITABLE) {
#if AX_X11        
          CbActivateEditWin(w_combo_box);
#endif        
      }
    }
    in_widget->hasFocus = TRUE;
}


/*************************************************************************
function: ElfbGetActiveWidget - Return widget of text cursor
*/
AXWidget_entry ElfbGetActiveWidget(in_w)
AXWidget in_w;
{
    AXWidget w;
    AXWidget_entry w_enter;
#if AXMS
    HWND focus_win;
#endif

    w = in_w;
    w_enter = (AXWidget_entry) NULL;

    if ((w != NULL) && (w->type == W_MAIN))
        w = (AXWidget)((AXWidget_main)w)->dialog;
    if (w == NULL) {
        return ((AXWidget_entry) NULL);
    }

#if AXMS    /* { */
    w_enter = NULL;
    focus_win = MS_GetFocus();
    for ( w = w->widgets; w; w = w->nextChild )
        {
        if ( w->type == W_ENTER || w->type == W_EDITBOX )
            {
            w_enter = (AXWidget_entry)w;
            if ( w_enter->winid == focus_win )
                {
                w_enter->hasFocus = TRUE;
                return( w_enter );
                }
            }
        }
    
    /*** The focus has shifted somewhere else so try                      **/
    /*** the old, unreliable way since we do not handle all of the focus  **/
    /*** change messages from Windows to set our own widget state         **/
    /*** but we need this to support the toolbar "cut/copy/paste" buttons **/
    /*** since the very act of picking on them removes the focus from the **/
    /*** windows widget, so wwe will not find it in the loop above - jon  **/
	w = in_w;
#endif

    for (w = w->widgets; w; w = w->nextChild)
    {
        w_enter = (AXWidget_entry) w;
        if ((w_enter->type == W_ENTER) && !(w_enter->disabled) &&
			(w_enter->hasFocus))
            break;
        if ((w_enter->type == W_EDITBOX) && (w_enter->hasFocus))
            break;
        w_enter = (AXWidget_entry) NULL;
    }
    return ((AXWidget_entry) w_enter);
}


/*************************************************************************
function: PaintWidget - Put widget on screen
*/
void PaintWidget(w)
register AXWidget w;
{
    register void (*func) ();

    func = w->painter;
    if (func) {
        (*func) (w, NULL, TRUE, TRUE);
        w->update_widget = FALSE;
    }
}


/*************************************************************************
function: XmDlgInWidget - Return the dialog sub-widget id
*/
AXWidget_dialog XmDlgInWidget(w)
AXWidget_main w;
{
    if (!w || !w->dialog)
        return ((AXWidget_dialog)w);
    else
        return (w->dialog);
}

/*************************************************************************
function: XmButtonInWidget - Return widget of button in a widget
This is only used by the DC.  It is used to get the zoom button
from the corner widget.
*/
AXWidget XmButtonInWidget(w)
AXWidget w;
{
    AXWidget childP;

    if (!w)
        return (NULL);
    for (childP = w->widgets; childP; childP = childP->nextChild) {
        if (childP->type == W_BUTTON)
            return (childP);
    }
    return (NULL);
}

/*************************************************************************
function: XmWidgetName - Express fetch of widget name
*/
mbchar *XmWidgetName(w)
register AXWidget w;
{
    return (w->name);
}

/*************************************************************************
function: XmWidgetWidth - Express call for width of a widget
*/
XmWidgetWidth(w)
register AXWidget w;
{
    return (w->width);
}

/*************************************************************************
function: XmWidgetSizeAndPos - Express call for widget info
*/
void XmWidgetSizeAndPos(w, x, y, width, height)
register AXWidget w;
register int *x, *y, *width, *height;
{
    if (!w)
        return;

    *x = w->xpos;
    *y = w->ypos;
    *width = w->width;
    *height = w->height;
}


Pixmap
GetWidgetPixmap(w)
AXWidget_pushbutton w;
{
    if (!w)
        return((Pixmap) NULL);
    WIDGET_PARANOID( w, W_BUTTON );
    return((Pixmap) w->pixmap);
}

/* Get pixmap of first child widget */
Pixmap
GetWidgetChildPixmap(w)
AXWidget w;
{
    if (!w || !w->widgets)
        return((Pixmap) NULL);

    if ( w->type == W_COLORPAL )
        return((Pixmap) ((AXWidget_ribbon) w->widgets)->pixmap);
    else
        return((Pixmap) ((AXWidget_pushbutton) w->widgets)->pixmap);
}


/****************************************************************
function: WidgetExposed - Return whether widget needs redrawn
*/
int WidgetExposed(w, exP)
register AXWidget w;
XExposeEvent *exP;
{
    register int wx, wy, ex, ey, wwidth, wheight, ewidth, eheight;

#if AXMS
    if ( w->type == W_BUTTON && w->style != W_POPUP )
       {
       if ( ((AXWidget_pushbutton)w)->owner_draw && (w->window != NULL) )
          /* Windows will generate a subsequent DRAW_ITEM for exposed buttons */
          /* se do not draw them now.  If we do we get a double-draw    */
          return( FALSE );
       }
#endif

    if (!exP)           /* sometimes expose-all is encoded as NULL exP */
        return (TRUE);

    wx = w->xpos;
    if (w->type == W_SPLITTER)  /* special animal */
        wy = w->ypos + ((AXWidget_main) w)->value - w->height / 2;
    else
        wy = w->ypos;
    wwidth = w->width;
    wheight = w->height;
    ex = exP->x;
    ey = exP->y;
    ewidth = exP->width;
    eheight = exP->height;

    if ((ey + eheight <= wy) || /* above */
        (ex + ewidth <= wx) ||  /* left */
        (wy + wheight <= ey) || /* below */
        (wx + wwidth <= ex))    /* right */
        return (FALSE);
    else
        return (TRUE);
}

/****************************************************************
 * function: AxGadgetExposed - Return whether elf-visible widget needs
 * to be redrwan.  This is analogous to WidgetExposed, but is called on
 * the elements of (AXWidget_dialog)->layerList; for instance, this would
 * be called on a list box, but not on its scrollbars (which aren't visible
 * from the elf layer.  layerList holds handles to elf-visible widgets)
 */
int                               /* return whether needs redraw */
AxGadgetExposed(w, updateRect)
register AXWidget w;
r32 *updateRect;                  /* area that needs to be updated */
{
#if AXMS
    if ( w->type == W_BUTTON && w->style != W_POPUP )
       {
       if ( ((AXWidget_pushbutton)w)->owner_draw )
          /* Windows will generate a subsequent DRAW_ITEM for exposed buttons */
          /* se do not draw them now.  If we do we get a double-draw    */
          return FALSE;
       }
#endif
    
    return (widgetIntersectsUpdateRect(w, updateRect));
}

int FreeWidgetName(w)
register AXWidget w;
{
    if (!w)
        return (-1);
    if (w->name)
        free(w->name);
    w->name = (mbchar *) 0;
    return (0);
}

/*************************************************************************
function: RemoveWidgetFromParent - Remove from parent's children list
*/
void RemoveWidgetFromParent(w)
register AXWidget w;
{
    register AXWidget p = (AXWidget)w->parent;

    if (!p || (w->type == W_MAIN))      /* no parent */
        return;
    
    if ((w->type == W_BUTTON) && (p->type == W_DIALOG)) {
        AXWidget_dialog dw = (AXWidget_dialog)p;

            /* these are also child references
             */
        if (dw->defaultButton == (AXWidget_pushbutton)w) {
            dw->defaultButton = (AXWidget_pushbutton) NULL;
        }
        if (dw->currentDefaultButton == (AXWidget_pushbutton)w) {
            dw->currentDefaultButton = (AXWidget_pushbutton) NULL;
        }
        if (dw->cancelButton == (AXWidget_pushbutton)w) {
            dw->cancelButton = (AXWidget_pushbutton) NULL;
        }
    }
    
    if (p->widgets == w) {
        p->widgets = w->nextChild;
        return;
    }
    
    p = p->widgets;
    while (TRUE) {
        if (!p) {
#if ZEROED_MEM
            printf(NXLT("RemoveWidgetFromParent: ORPHAN widget @ 0x%lx\n"), w);
#endif
            return;
        }
        if (p->nextChild == w) {
            p->nextChild = w->nextChild;
            return;
        }
        p = p->nextChild;
    }
}


/*************************************************************************
function: SetRadioValue - Select a button
Return boolean as to whether anyone got set.
*/
int SetRadioValue(tP, display)
register AXWidget_togglebutton tP;
int display;
{
    register AXWidget_radio w;
    register int ix;


    WIDGET_PARANOID(tP, W_TOGGLE);
    w = tP->in_radio;
    if ( !w )
        return(FALSE);
    for ( ix = 0; ix < w->num_buttons; ix++ ) {
        if ( w->buttons[ix]->value ) {
            /* if this button is the one that's on */
            if ( w->buttons[ix] == (AXWidget_togglebutton)tP ) {
                /* if the on button is the one that was picked */
                return(FALSE);
            }
            /* some other button was picked; paint this one off */
            w->buttons[ix]->value = FALSE;
            if ( display )
                PaintToggle(w->buttons[ix], NULL, TRUE, TRUE);
        }
        else if ( w->buttons[ix] == (AXWidget_togglebutton)tP ) {
            /* the picked button is one that is off,
             * so turn it on */
            w->buttons[ix]->value = TRUE;
                        w->value = ix;
            if ( display )
                PaintToggle(w->buttons[ix], NULL, TRUE, TRUE);
        }
    }
    return(TRUE);
}



/***********************************************************************
function: AttachOptionMenuToRibbon - 
*/
void AttachOptionMenuToRibbon(w, names, num_names, winID)
AXWidget_ribbon w;
char **names;
int  num_names;
int  winID;
{
    static XtCallbackRec Callback[2];   /* standard template */
    int num_buttons, an, ix, x_offset, option_ix;
    AXWidget buttons[20], pulldown;
    void GeRibbonOptionChanged();
    AXWidget_option childP, option_menu;
    Arg args[20];
    char *name;
    
    if (!w || w->type != W_RIBBON || !names)
        return;

    /* create option (or reuse) the option widget, noting the old value */
    an = 0;
    XtSetArg(args[an], XmNoldWidget, 0); an++;
    option_ix = 0;
    childP = (AXWidget_option)w->widgets;
    while ( childP )
    {
        if ( childP->type != W_OPTION )
            childP = (AXWidget_option)childP->nextChild;
        else
        {
            XtSetArg(args[0], XmNoldWidget, childP);   /* rewrite 0th arg */
            if ( childP->num_buttons == num_names )
                option_ix = childP->value;
            break;
        }
    }
    XtSetArg(args[an], XmNflags, 0); an++;
    CALLBACKER(Callback, GeRibbonOptionChanged, winID);
    XtSetArg(args[an], XmNactivateCallback, Callback); an++;
    
#if AXMS    /* { */
    /* Windows option menu is not made up of a separate pulldown with *
     * pushbuttons. Windows will manage and draw the strings for us.  */
    option_menu = AXmCreatePulldownMenu((AXWidget)w, "", args, an);
    AxWinDoOptionMenu(option_menu, num_names, names, 0, 1, names[option_ix]);
    option_menu->xpos = w->width - option_menu->width;

#else       /* } { */
    pulldown = (AXWidget)AXmCreatePulldownMenu((AXWidget)w, "", args, an);
    x_offset = w->width - pulldown->width;

    /* create button gadgets (old ones deleted by AXmCreatePulldownMenu) */
    for ( ix=num_buttons=0; ix<num_names; ix++ )
    {
        name = names[ix];
        if (!name)
            continue;

        an = 0;
        XtSetArg(args[an],XmNoldWidget, 0); an++;
        XtSetArg(args[an],XmNflags, 0); an++;
        XtSetArg(args[an],XmNwhichButton, Button1); an++;
        XtSetArg(args[an],XmNaxFont, AxMakeStrData(-1,(mbchar *)AXSAB_16));an++;
        buttons[ix] = AXmCreateToggleButtonGadget(pulldown, name, args, an);
        num_buttons++;
    }
    
    /* associate buttons and pulldown with option menu and ribbon */
    an = 0;
    AxXtManageChildren();
    XtSetArg(args[an], XmNx, x_offset); an++;
    XtSetArg(args[an], XmNy, 1); an++;
    XtSetArg(args[an], XmNsubMenuId, pulldown); an++;
    XtSetArg(args[an], XmNlabelString, ""); an++;
    XtSetArg(args[an], XmNvalue, option_ix); an++;
    XtSetArg(args[an], XmNmenuHistory, buttons[option_ix]); an++;
    XtSetArg(args[an], XmNaxFont, AxMakeStrData(-1, (mbchar *) AXSAB_16)); an++;
    option_menu = (AXWidget_option)AXmCreateOptionMenu((AXWidget)w,"",args,an);

#endif    /* } AXMS */
    AxXtManageChild((AXWidget)option_menu);
    PaintWidget((AXWidget)option_menu);
}


/*************************************************************************
function: AxSetWidgetVal - Change an attribute of a widget
*/
void AxSetWidgetVal (w,code,val)
AXWidget w;
int code;
int val;
{
    Arg arglist[10];

    XtSetArg(arglist[0],code,val);
    AxXtSetValues(w,arglist,1);
}


/*** wrapper for AXtRealizeWidget ***/
void AxRealizeWidget(w)
AXWidget_main w;
{
#if AXMS
    WXShowWindow(w);
#else
    AXtRealizeWidget(w);
#endif
}    

/*** wrapper for AxXtDestroyWidget ***/
void AxDestroyWidget(w)
AXWidget w;
{
#if AXMS
    WXDestroyWidget(w);
#else
    AxXtDestroyWidget(w);
#endif
}    



/* Return the name of the button corresponding to the option value VALUE.
   Return NULL on failure. */
char *
AXmGetOptionButtonName(w, value)
register AXWidget_option w;
register int value;
{
    if ((!w) || w->type != W_OPTION)
        return((char *) NULL);
#if AXMS
    if ((value < 0) || (value >= w->num_strs) || (!w->strs))
        return((char *) NULL);        
    return((char *) w->strs[value]);

#else

    if ((value < 0) || (value >= w->num_buttons) || (!w->buttons))
        return((char *) NULL);
    return((char *) w->buttons[value]->name);
#endif
}

/*************************************************************************
function: AxGetSizeOfWidget - returns content width and height
*/
void AxGetSizeOfWidget(wd, wid, hyt)
AXWidget wd;
int *wid, *hyt;
{
    int thickness;

    *wid = 0;
    *hyt = 0;
    if ( wd ) switch ( wd->type )
    {
        /* the widgets with no edges */
        case W_MAIN:
        case W_DIALOG:
        case W_WORK:
        case W_EDITBOX:
        case W_TABLE:
        case W_CANVAS:
        case W_TRAY:
        case W_PLATES:
        default:
            *wid = wd->width;
            *hyt = wd->height;
        break;

        /* the widgets with edges */
        case W_BUTTON:
        case W_ICON:
        case W_RECT:
        case W_PALETTE:
        case W_RULER:
        case W_MENUBAR:
        case W_STATUS:
        case W_DEBUG:
        case W_CORNER:
        case W_TOOLBAR:
        case W_COLORPAL:
        case W_GE_XRULER:
        case W_GE_YRULER:
        case W_ROWCOL:
            thickness = wd->thickness ? wd->thickness : 2;
            if ( thickness < 0 )
            {
                *wid = wd->width  + 2*thickness;
                *hyt = wd->height + 2*thickness;
            }
            else
            {
                *wid = wd->width  - 2*thickness;
                *hyt = wd->height - 2*thickness;
            }
        break;

        /* the irregular widgets */
        case W_KANJI_KANA:
        case W_KANJI_STATUS:
        case W_HGROOVE:
        case W_POPUP:
        case W_OPTION:
        case W_LIST:
        case W_SCROLL:
        case W_ENTER:
        case W_LABEL:
        case W_TOGGLE:
        case W_RADIO:
        case W_ALIGN:
        case W_SPLITTER:
        case W_SCALE:
        case W_RIBBON:
        break;
    }
}

/*************************************************************************
function: AxGetPosOfWidget - returns content x and y from widget
*/
void AxGetPosOfWidget(wd, xpos, ypos)
AXWidget wd;
int *xpos, *ypos;
{
    int thickness;
    if ( !wd )
    {
        *xpos = 0;
        *ypos = 0;
    }
    else switch ( wd->type )
    {
        /* the widgets with no window and no edges */
        case W_DIALOG:
        case W_WORK:
        case W_EDITBOX:
        case W_TABLE:
        case W_TRAY:
        case W_PLATES:
        default:
            *xpos = wd->xpos;
            *ypos = wd->ypos;
        break;

        /* the widgets with no window and with edges */
        case W_BUTTON:
        case W_ICON:
        case W_RECT:
        case W_PALETTE:
        case W_RULER:
        case W_MENUBAR:
        case W_STATUS:
        case W_DEBUG:
        case W_CORNER:
        case W_TOOLBAR:
        case W_COLORPAL:
        case W_GE_XRULER:
        case W_GE_YRULER:
        case W_ROWCOL:
            thickness = wd->thickness ? wd->thickness : 2;
            if ( thickness < 0 )
            {
                *xpos = wd->xpos - thickness;
                *ypos = wd->ypos - thickness;
            }
            else
            {
                *xpos = wd->xpos + thickness;
                *ypos = wd->ypos + thickness;
            }
        break;

        /* the widgets with windows or ??? */
        case W_MAIN:
        case W_CANVAS:
        case W_KANJI_KANA:
        case W_KANJI_STATUS:
        case W_HGROOVE:
        case W_POPUP:
        case W_OPTION:
        case W_LIST:
        case W_SCROLL:
        case W_ENTER:
        case W_LABEL:
        case W_TOGGLE:
        case W_RADIO:
        case W_ALIGN:
        case W_SPLITTER:
        case W_SCALE:
        case W_RIBBON:
            *xpos = 0;
            *ypos = 0;
        break;
    }
}


/*************************************************************************
function: PaintCorner - Paint the corner area of the window
*/
void PaintCorner (w, exP, regen, layered)
register AXWidget w;
register XExposeEvent *exP;
int regen;
int layered;
{
    AXWidget childP;
    register void (*func)();

    for ( childP = w->widgets; childP; childP = childP->nextChild ) {
        func = childP->painter;
        if ( func && childP->display )
            (*func)(childP, exP, TRUE, TRUE);
    }
}

/* : doWidgetUpdates - this is called from AXtAppNextEvent when
 * WidgetNeedsUpdated is true
 */
void
doWidgetUpdates()
{
    register AXWidget w, bp;
    int ix, intersectsDamage;
    
    /* first unmap those widgets that need it */
    
    for ( bp = WindowList; bp; bp = bp->nextWin ) {
        if ( (bp->type != W_DIALOG) && (bp->type != W_PALETTE) ) {
            if ( bp->regen_widget ) {
                for ( w = bp->widgets; w; w = w->nextChild ) {
                    if ( !w->display && w->mapped ) {
                        UnMapWidget(w);
                    }
                    else {
                        w->update_widget = TRUE;
                    }
                }
            }
        }
        if (bp->window) {
            for ( w = bp->widgets; w; w = w->nextChild ) {
                if ( w->update_widget ) {
                    if ( !w->display && w->mapped ) {
                        UnMapWidget(w);
                    }
                }
            }
        }
#if AXMS
        /* on Windows, the toolbar keeps its widgets in the tool_bar_info *
         * struct. so let's make sure we go thru that.      -donato       */
        if ( bp->type == W_TOOLBAR ) {
            for ( ix=0; ix < MAX_TOOLBAR_ITEMS; ix++ ) {
                w = ((AXWidget_tool)bp)->toolBarInfo->widgets[ix];
                if ( w && w->update_widget && !w->display && w->mapped )
                    UnMapWidget(w);
            }
        }
#endif
    }
    
    /* now do the updates on containers and their containees */
    
    for ( bp = WindowList; bp; bp = bp->nextWin ) {
        if ( bp->widgets ) {
            if (!(bp->flags & WFLG_IN_EDIT) && bp->layerList) {
                    /* the layerList reflects back-to-front layering
                     * for display purposes.  Regarding WFLG_IN_EDIT:
                     * dialog editor does its own layered painting.
                     */
                r32 updateRect, thisRect;
                int insidePanel;
                wLayer *ly;

                if (bp->auxData && bp->auxData->damageRect.width) {
                    updateRect = bp->auxData->damageRect;
                    memset((void *) &bp->auxData->damageRect, 0, sizeof(r32));
                }
                else {
                    memset((void *) &updateRect, 0, sizeof(r32));
                }
                
                for ( ly = bp->layerList; ly; ly = ly->next ) {
                    w = ly->w;
                    if (!w->display) {
                        w->update_widget = FALSE;
                        continue;
                    }
                    intersectsDamage =
                        widgetIntersectsUpdateRect(w, &updateRect);
                    if (w->update_widget || intersectsDamage) {
                        if (w->painter) {
#if AXMS
                            w->main_loop_paint = TRUE; /* no focus changes */
#endif
                            if (!w->update_widget && (W_RECT == w->type)) {
                                /* for panels, try to avoid full repaint
                                 */
                                if (updateRect.width && updateRect.height) {
                                    insidePanel = 
                                        xmRectInsidePanel(&updateRect, 
														(AXWidget_rect)w);
                                }
                                else {
                                    widgetFootprint(w, &thisRect);
                                    insidePanel =
                                        xmRectInsidePanel(&thisRect, 
														(AXWidget_rect)w);
                                }
                            }
                            else {
                                insidePanel = FALSE;
                            }
                            
                            if (insidePanel) {
                                /* paint only the affected part of (W_RECT) w
                                 */
                                xmPaintPanelFace((AXWidget_rect)w, 
												 updateRect.origin.x,
                                                 updateRect.origin.y,
                                                 (int) updateRect.width,
                                                 (int) updateRect.height);
                            }
                            else {
                                w->painter(w, NULL, (intersectsDamage ||
                                           xmlibStr.FullRegenForUpdate), TRUE);
                            }
#if AXMS
                            w->main_loop_paint = FALSE;
#endif
                            if (!insidePanel) {
                                /* include this widget's redisplay
                                 * damage in updateRect
                                 */
                                widgetFootprint(w, &thisRect);
                                if (!updateRect.width) {
                                    updateRect = thisRect;
                                }
                                else {
                                    unionRectangle(&thisRect, &updateRect,
                                                   &updateRect);
                                }
                            }
                        }
                        w->update_widget = FALSE;
                    }
                }
            }
            else {
                /* no layerList -- just do the updates without
                   considering update-damage to other containees */
                
                for ( w = bp->widgets; w; w = w->nextChild ) {
                    if ( w->update_widget && w->display ) {
                        if ( w->painter ) {
#if AXMS
                            w->main_loop_paint = TRUE; /* no focus changes */
#endif
                            w->painter(w, NULL,
                                       xmlibStr.FullRegenForUpdate, TRUE);
#if AXMS
                            w->main_loop_paint = FALSE;
#endif
                        }
                        w->update_widget = FALSE;
                    }
                }
            }
        }
    }
    
    /* update those that weren't updated during the container pass */
    
    for ( bp = WindowList; bp; bp = bp->nextWin ) {
#if AXMS
        /* on Windows, the toolbar keeps its widgets in the tool_bar_info *
         * struct. so let's make sure we go thru that.      -donato       */
        if ( bp->type == W_TOOLBAR ) {
            UpdateToolBarWidgets((AXWidget_tool)bp);
        }
#endif
        if ( bp->update_widget ) {
            if ( bp->painter ) {
                bp->painter(bp, NULL, xmlibStr.FullRegenForUpdate, TRUE);
            }
            bp->update_widget = FALSE;
        }
    }
}


/* : widgetIntersectsUpdateRect - this is called by doWidgetUpdates et. al.
 * return whether w intersects the passed update rectangle
 */
int
widgetIntersectsUpdateRect(w, updateRect)
AXWidget w;
r32 *updateRect;
{
    r32 thisRect;
    int margin, doublemargin;
#if AXMS
	int running_windows = TRUE;
#else
	int running_windows = FALSE;
#endif
    
    if (!updateRect->width || !updateRect->height)
        return (FALSE);
    
    if (w->flags & WFLG_IN_EDIT) {
        margin = deMarginSize(w);
    }
    else {
        margin = 1;
    }
    doublemargin = margin + margin;

    /* take a conservative size estimate using xmFocusSelectorRect
     */
    if (running_windows || (W_RECT == w->type) || (W_OPTION == w->type)) {
        /* widget's selector rect is smaller than whole widget for these types.
         */
        thisRect.origin.x = w->xpos - margin;
        thisRect.origin.y = w->ypos - margin;
        thisRect.width = w->width + doublemargin;
        thisRect.height = w->height + doublemargin;
    }
    else {
        xmFocusSelectorRect(w, &thisRect);   /* >= size of widget */
        thisRect.origin.x -= margin;
        thisRect.origin.y -= margin;
        thisRect.width += doublemargin;
        thisRect.height += doublemargin;
    }
    
    if ((w->type == W_RECT) &&
        (!((AXWidget_rect) w)->graphic &&
         !((AXWidget_rect) w)->background_image &&
         !((AXWidget_rect) w)->widgetColor)) {
        
        int thickness;
    
        /* this widget is a transparent panel; test whether the
         * updateRect is totally inside it
         */
        thickness = abs(w->thickness) + margin;
        thisRect.origin.x += thickness;
        thisRect.origin.y += thickness;
        thisRect.width -= (2 * thickness);
        thisRect.height -= (2 * thickness);
        if ((RECT_LABELED == w->style) || (RECT_LAYERED == w->style)) {
            /* account for smaller inner rect for these types
             */
            AXWidget_rect rw = (AXWidget_rect) w;
            thisRect.origin.y += rw->tabctl.top_inset;
            thisRect.height -= rw->tabctl.top_inset;
        }
        if (rectInsideRect(updateRect, &thisRect)) {
            return (FALSE);
        }
        else {
            /* go back to outer dimension and check for intersection
             */
            thisRect.origin.x = w->xpos - margin;
            thisRect.origin.y = w->ypos - margin;
            thisRect.width = w->width + doublemargin;
            thisRect.height = w->height + doublemargin;
            if ((RECT_LABELED == w->style) || (RECT_LAYERED == w->style)) {
                /* account for extra northern extent for these types.
                 * Use the labeled measurement (the larger) for both
                 */
                XFontStruct *labelfont;
                labelfont = w->font ? w->font : AxDlgMainFont;
                thisRect.origin.y -= labelfont->ascent / 2;
                thisRect.height += labelfont->ascent / 2;
            }
            return (rectanglesIntersect(&thisRect, updateRect));
        }
    }
    else {
        if (rectanglesIntersect(&thisRect, updateRect)) {
            return (TRUE);     /* "top" widget already intersects */
        }
        else if (w->auxData) {
            wComposite *cp;

            /* there may be composite kids; see whether any intersect
             */
            for (cp = w->auxData->compositeList; cp; cp = cp->next) {
                if (cp->type == WLINKDOWN) {
                    thisRect.origin.x = cp->w->xpos - margin;
                    thisRect.origin.y = cp->w->ypos - margin;
                    thisRect.width = cp->w->width + doublemargin;
                    thisRect.height = cp->w->height + doublemargin;
                    if (rectanglesIntersect(&thisRect, updateRect)) {
                        return (TRUE);
                    }
                }
            }
        }
        return (FALSE);
    }
}

/*******************************
 * : widgetFootprint - return size of passed widget,
 * including any 'composite kids', for example, for
 * an entry field, include the size of its label if present
 */
void
widgetFootprint(w, wsize)
AXWidget w;
r32 *wsize;           /* out: return total footprint */
{
    r32 trect;
    wComposite *cp;
    
    if (!w || !wsize)
        return;
    
    wsize->origin.x = w->xpos;
    wsize->origin.y = w->ypos;
    wsize->width = w->width;
    wsize->height = w->height;
    

        /* test for types whose extents differ
         * from their xpos, ypos, width, height
         */
    if (w->type == W_SPLITTER) {
        wsize->origin.y = w->ypos + w->value - (w->height / 2);
    }
    else if (w->type == W_ENTER) {
        wsize->origin.y -= 4;
        wsize->origin.x += 2;
    }
    
    if ((w->type != W_OPTION) && w->auxData) {
        
            /* W_OPTION is excluded here since its composite kids are
             * the pulldown buttons -- these buttons affect geometry
             * only during the time that the pulldown is open (a state
             * during which widget footprints currently aren't measured.)
             * To include them will actually throw off the measurement.
             */
        for (cp = w->auxData->compositeList; cp; cp = cp->next) {
            if (cp->type == WLINKDOWN) {
                /*
                 * include size of composite kid
                 */
                trect.origin.x = cp->w->xpos;
                trect.origin.y = cp->w->ypos;
                trect.width = cp->w->width;
                trect.height = cp->w->height;
                unionRectangle(&trect, wsize, wsize);
            }
        }
    }
}


/*************************************************************************
function: FindGadget - Find pointer sensitive widget by x/y
*/
AXWidget
FindGadget(win, x, y)
Window win;
register int x,y;
{
    register AXWidget wp;

    for ( wp = WindowList; wp; wp = wp->nextWin )
#if AXMS
        if ( wp->winid == win )
#else
        if ( wp->window == win )
#endif
            break;

    if ( !wp )
        return(NULL);

    for ( wp = wp->widgets; wp; wp = wp->nextChild )
    {
        if ( wp->disabled || !wp->mapped || !wp->display )
            continue;
        if ( (wp->type == W_BUTTON) ||
            ((wp->type == W_TOGGLE) && 
                (!((AXWidget_togglebutton) wp)->in_option)) ||
            ((wp->type == W_LABEL) && (wp->helpstr)) ||
            (wp->type == W_OPTION) ||
            (wp->type == W_SCALE) ||
            (wp->type == W_ROWCOL) ||
            (wp->type == W_RECT && RECT_LAYERED == wp->style) ||
            (wp->type == W_COMBO_BOX) ) {
            
            if (wp->parent->type == W_PLATES) {
                if(ssFindPlatesGadget((AXWidget_pushbutton)wp, x, y))
                    return(wp);
            }
            else if (wp->type == W_RECT) {
                AXWidget_rect rw = (AXWidget_rect) wp;
                if ( (x >= rw->xpos) && (x < rw->xpos + rw->width) &&
                     (y >= rw->ypos) &&
                     (y < rw->ypos + rw->tabctl.top_inset) ) {
                    
                    return (wp);    /* falls in tabs area somewhere */
                }
                else {
                    continue;
                }
            }
            else if ( (x >= wp->xpos) && (x < wp->xpos+wp->width) &&
                       (y >= wp->ypos) && (y < wp->ypos+wp->height) ) {
                return(wp);
            }
        }
        else if ( wp->type == W_SPLITTER ) {
            if ( (x >= wp->xpos) && (x < wp->xpos+wp->width) &&
                    (y >= wp->ypos+wp->value-wp->height/2) &&
                    (y < wp->ypos+wp->value+wp->height/2) )
                return(wp);
        }
    }
    return(NULL);
}



/*************************************************************************
function: SetTopPendingDelete - Set the initial keyboard-focus widget
*/
AXWidget SetTopPendingDelete (w)
	register AXWidget w;        /* the dialog box */
{
	register AXWidget best_w;
	AXWidget_dialog d_w;
	AXWidget wtmp;
    
	for ( wtmp = w->widgets; wtmp; wtmp = wtmp->nextChild ) {
		if (wtmp->hasFocus) {
			return (wtmp);  /* take this as an elf-specified 1st-focus widg */
		}
	}

	best_w = NULL;
	d_w = (AXWidget_dialog) w;
	{
		int anyType = TRUE;  /* whether to consider non-W_ENTER,W_EDITBOX */
		for ( wtmp = w->widgets; wtmp; wtmp = wtmp->nextChild ) {
			if ((W_ENTER == wtmp->type) || (W_EDITBOX == wtmp->type) ||
				(W_COMBO_BOX == wtmp->type)) {
				anyType = FALSE;  /* one or both preferred types are present */
			}
		}
		for ( w = w->widgets; w; w = w->nextChild ) {
			if (anyType || (W_ENTER == w->type) || (W_EDITBOX == w->type)
				|| (W_COMBO_BOX == w->type)) {
				if (xmCanTakeKeyFocus(d_w, w, FALSE)) {
					if ( !best_w )
						best_w = w;
					if ( w->ypos < best_w->ypos )
						best_w = w;
					else if ( (w->ypos == best_w->ypos) &&
							  (w->xpos <= best_w->xpos) )
						best_w = w;
				}
			}
		}
	}
    if (best_w) {
		AXWidget tw;
		if ((best_w->type == W_ENTER) || (best_w->type == W_COMBO_BOX)) {
			best_w->pending_delete = TRUE;
		}
		if (W_RADIO == best_w->type) {
			tw = xmRadioActiveToggle(best_w);
		}
		else {
			tw = best_w;
		}
		tw->hasFocus = TRUE;
    }
    return(best_w);
}


/*************************************************************************
function: AxXtGetValues - Write back to args array
*/
void AxXtGetValues (w,args,cnt)
     register AXWidget w;
     register Arg args[];
     register int cnt;
{
  register int ix;
  AXWidget_list wl;

  for ( ix = 0; ix < cnt; ix++ ) {
    switch ( args[ix].name ) {
    case XXmNrevstamp:
      *((int*)(args[ix].value)) = ((AXWidget_dialog)w)->revstamp;
      break;
    case XXmNvalue:
      if ( w->type == W_ENTER)
    {
#if BUILD_WIDE
#if AXMS
      if ( w->mb_string != (mbchar *)NULL )
        free( w->mb_string );
      w->mb_string = (mbchar *)malloc( (Wstrlen(w->wstr)+1)*sizeof(wchar) );
      WToMbStr( w->wstr, w->mb_string );
      *((char**)(args[ix].value)) = (char *)w->mb_string;
#else
      *((char**)(args[ix].value)) = (char *)GetWidgetString(w);
#endif
#else
      *((char**)(args[ix].value)) = (char *)((AXWidget_entry)w)->wstr;
#endif
    }
    else if ( w->type == W_SCROLL )
    {
        *((int*)(args[ix].value)) = ((AXWidget_scroll)w)->scroll_pos;
    }
      else
    *((int*)(args[ix].value)) = w->value;
      break;
    case XXmNwidth:
      *((int*)(args[ix].value)) = w->width;
      break;
    case XXmNheight:
      *((int*)(args[ix].value)) = w->height;
      break;
    case XXmNx:
      *((int*)(args[ix].value)) = w->xpos;
      break;
    case XXmNy:
      *((int*)(args[ix].value)) = w->ypos;
      break;
    case XXmNbackground:
      *((int*)(args[ix].value)) = xmlibStr.BGpixel;
      break;
    case XXmNforeground:
      *((int*)(args[ix].value)) = xmlibStr.FGpixel;
      break;
    case XXmNitemCount:
      if ( w->type == W_LIST )
    {
      wl = (AXWidget_list)w;
      *((int*)(args[ix].value)) = wl->num_strs;
    }
      break;
    case XXmNvisibleItemCount:
      if ( w->type == W_LIST ) 
    {
      wl = (AXWidget_list)w;
      *((int*)(args[ix].value)) =
        wListVisibleItemCount(wl->height, wl->font,
                  wl->textShadow, wl->list_hscroll);
    }
      else if ( w->type == W_SCROLL )
    {
      *((int*)(args[ix].value)) = ((AXWidget_scroll)w)->scroll_max;
    }
      break;
    case XXmNitems:
      if ( w->type == W_LIST )
    {
      wl = (AXWidget_list)w;
      *((char ***)(args[ix].value)) = wl->strs;
    }
      else if (w->type == W_ROWCOL )
    {
      *((AXWidget_pushbutton  **)(args[ix].value)) = ((AXWidget_rowcol)w)->buttons;
    }
      break;
    case XXmNflags:
      *((int*)(args[ix].value)) = (int)(w->flags & 0xffff);
      break;
    case XXmNnoPrimarySeln:
      if (  w->type == W_EDITBOX )
    {
      *((int*)(args[ix].value)) = ((AXWidget_editbox)w)->no_primary_seln;
    }
      break;
    case XXmNsliderSize:
    if ( w->type == W_SCROLL )
    {
      *((int*)(args[ix].value)) = ((AXWidget_scroll)w)->scroll_per;
    }
      break;
    default:
      abort();
    }
  }
}

/*************************************************************************
function: AxXtSetValues - Write standard args
*/
void AxXtSetValues (w,args,cnt)
register AXWidget w;
register Arg args[];
register int cnt;
{
    register int ix, j, tempint, sz;
    Cursor lcursor = None;
    char *cp;
    int att_code;
    AXWidget newP;
    AXWidget_menu menuP;
    AXWidget_list wl;
    AXWidget_rowcol wr;
    AXWidget_pushbutton wp;
    AXWidget_option w_option;
    AXWidget_editbox w_editbox;
    AXWidget_main w_main;
    AXWidget_ribbon w_rbn;
    AXWidget_radio w_radio;
    AXWidget_entry w_enter;
    AXWidget_scroll w_scroll;
    AXWidget_scale w_scale;
    AXWidget_dialog wd;
    AXWidget_popup w_popup;
    AXWidget_togglebutton w_toggle;
    AXWidget_splitter w_splitter;
    AXWidget_color_chooser w_choose;
    AXWidget_tray w_tray;
    AXWidget_table w_t;
    AXWidget_canvas w_c;
    AXWidget_rect w_rect;
    AXWidget_page w_page;
    AXWidget_icon w_icon;
    AXWidget_combo_box w_cbox;
    XFontStruct *tfont;


    for ( ix = 0; ix < cnt; ix++ ) {
        att_code = args[ix].name;
        switch ( att_code ) {
        case XXmNdefaultPosition:
        case XXmNdialogStyle:
        case XXmNoldWidget:
            break;
        case XXmNflags:
            SetWidgetFlags( w, (int)args[ix].value );
            break;
        case XXmNuid:
            if ((int) args[ix].value && w) {
                int tuid;
                if (w->parent && (W_DIALOG == w->parent->type)) {
                    char *notUsed;
                    wd = (AXWidget_dialog) w->parent;
                    if (!wd->uidWidgets) {
                        /* init table associating uid's with AXWidget handles
                         */
                        axInitUidWidgets(wd);
                    }
                    tuid = (int) (args[ix].value);
                    axhGetItemI((void *) &tuid, sizeof(int), wd->uidWidgets,
                                (char *) w, &notUsed);
                    w->db_uid = tuid;
                }
                else if (W_DIALOG == w->type) {
                    w->db_uid = (int) (args[ix].value);
                }
            }
            break;
        case XXmNvalue:
            SetWidgetValues(w, args[ix].value); 
            break;
        case XXmNelfObject:
            if (!w->elfObject)
                w->elfObject = AxTaskCopyElfData(0,(data_ptr)args[ix].value);
            break;
        case XXmNfg_color:
            SetWidgetColors(w, (int *)(args[ix].value), args[ix].value, TRUE);
            break;
        case XXmNbg_color:
            SetWidgetColors(w, (int *)(args[ix].value), args[ix].value, FALSE);
            break;
        case XXmNitems:
            SetWidgetItems(w, (char**)args[ix].value);
            break;
        case XXmNitemCount:
            SetWidgetItemCount(w, args[ix].value);
            break;
        case XXmNactivateCallback:
        case XXmNcancelCallback:
        case XXmNvalueChangedCallback:
        case XXmNentryCallback:
        case XXmNsingleSelectionCallback:
        case XXmNdefaultActionCallback:
        case XXmNextendedSelectionCallback:
        case XXmNincrementCallback:
        case XXmNdecrementCallback:
        case XXmNpageIncrementCallback:
        case XXmNpageDecrementCallback:
        case XXmNbrowseSelectionCallback:
        case XXmNokCallback:
            w->callback = ((XtCallbackRec*)args[ix].value)->callback;
            w->callbacktag = ((XtCallbackRec*)args[ix].value)->closure;
            break;

        case XXmNaxFont:
            tfont = lookupNamedFont((data_ptr)args[ix].value);
            if (w->font != tfont) {
                w->update_widget = WidgetNeedsUpdated = TRUE;
            }
            w->font = tfont;
            if (w->type == W_TABLE ) {
                TblSetFont((AXWidget_table)w, (data_ptr)args[ix].value);
            }
            break;
        case XXmNshadow:
            if (w->shadow) {
                AxTaskFreeElfData(0, w->shadow);
                w->shadow = (data_ptr) NULL;
            }
            if (args[ix].value) {
                w->shadow = AxTaskCopyElfData(0, (data_ptr) args[ix].value);
            }
            break;
        case XXmNtextShadow:
            if (w->textShadow) {
                AxTaskFreeElfData(0, w->textShadow);
                w->textShadow = (data_ptr) NULL;
            }
            if (args[ix].value) {
                w->textShadow =
                    AxTaskCopyElfData(0, (data_ptr) args[ix].value);
                if (w->type == W_TABLE && ((AXWidget_table)w)->info.table) {
                    int xoff, yoff;
                    table_info *tp;
                    tp = (table_info *)((AXWidget_table)w)->info.table;
                    if (tp && tp->font) {
                        xmTextShadowOffsets(tp->font, &xoff, &yoff);
                        tp->char_below = tp->font->descent + yoff + 2;
                    }
                }
            }
            break;
        case XXmNtextColor:
            if (w->textColor)
                AxFreeWColor(w->textColor);
            w->textColor = AxCreateWColor(w, (data_ptr) args[ix].value);
            break;
        case XXmNwidgetColor:
            if (w->widgetColor)
                AxFreeWColor(w->widgetColor);
            w->widgetColor = AxCreateWColor(w, (data_ptr) args[ix].value);
            break;
        case XXmNreferenceFont:
            w->referenceFont = lookupNamedFont((data_ptr)args[ix].value);
            break;
        case XXmNcompositeLinkUp: /* dlg editor */
            addCompositeLink(w, WLINKUP, (AXWidget ) args[ix].value);
            break;
        case XXmNcompositeLinkDown: /* dlg editor */
            addCompositeLink(w, WLINKDOWN, (AXWidget ) args[ix].value);
            break;
        case XXmNinEdit:        /* dlg editor */
            if (args[ix].value)
                w->flags |= WFLG_IN_EDIT;
            else
                w->flags &= ~WFLG_IN_EDIT;
            break;
        case XXmNx:
            w->xpos = args[ix].value;
            break;
        case XXmNy:
            w->ypos = args[ix].value;
            break;
        case XXmNwidth:
            w->width = args[ix].value;
            break;
        case XXmNheight:
            w->height = args[ix].value;
            break;
        case XXmNborderWidth:
        case XXmNthickness:
            w->thickness = args[ix].value;
            break;
        case XXmNIconId:
        case XXmNstyle:
            w->style = args[ix].value;
            break;
        case XXmNhelpStr:
            free(w->helpstr);
            w->helpstr =
                strcpy((char *)malloc(strlen((char *)args[ix].value)+1),
                       (char *)args[ix].value);
            break;
        case XXmNrminfo:
            SetRmInfoFrmDatPtr((AXWidget_popup)w, (data_ptr)args[ix].value);
            break;
        case XXmNmultipleSelectionCallback:
            w->multi_pick = args[ix].value;   
            break;
        case XXmNmenuHelpWidget:
            w->helpid = args[ix].value;
            break;
        case XXmNuseWorkColors:
            if ((w->useWorkColors != args[ix].value) && w->window) {
                if (w->flags & WFLG_IN_EDIT)
                    XFreePixmap(Dpy, w->window);
                else
                    AxDestroyWindow(Dpy, w, w->window);
                w->window = 0;
            }
            w->useWorkColors = args[ix].value;
            break;
        case XXmNeventHandlers:
            setEventHandlers(w, (eventHandlerList *) args[ix].value);
            break;
        case XXmNminorEventNotify:
            w->minorEventNotify = (args[ix].value != 0);
            break;
        case XXmNelfAccess:
            w->elf_access = TRUE;
            break;
        case XXmNnotTransient:
            w->notTransient = TRUE;
            break;
        case XXmNDefaultCursor: /* default cursor on the dlg box */
            ZXmBindCursor((char*)args[ix].value, &lcursor); 
            AXDefineDefaultCursor(w, lcursor);
            break;

        case XXmNbackgroundImage: {
            char *im_name = NULL;
            im_name = (char *) args[ix].value;
            if (!w->background_image) { /* Can only be set once */
                if (im_name) {
                    w->background_image = (char *) malloc(strlen(im_name) + 1);
                    strcpy(w->background_image, im_name);
                }
            }
            break;
        }
        case XXmNgraphic:
            if (w->graphic)
                AxDestroyAssociatedGraphic(w->graphic);
            w->graphic = AxTaskCopyElfData(0,(data_ptr)args[ix].value);
            w->update_widget = WidgetNeedsUpdated = TRUE;
            break;
        case XXmNuse2d:
            w->use2d = TRUE;
            break;

        case XXmNsmallFont:
            w->small_font = TRUE;
            if ( w->type == W_LABEL) {
                w->width = SmallBasicStrlen(w->name);
                w->height = SmallCharHeight;
            }
            break;

        case XXmNmeasureWtext:  /* dlg editor */
            if (!w->auxData) {
                w->auxData = newAuxData();
            }
            w->auxData->wEdit.measureText = TRUE;
            break;
		case XXmNaccelerator_key:
			SetWidgetAccelerator(w, (data_ptr)args[ix].value);
			break;
        default:
            switch (w->type) {
            case W_MAIN:
                w_main = (AXWidget_main)w;
                switch (att_code) {
                case XXmNmenuBar:
#if AX_X11                      /* { */
                    menuP = (AXWidget_menu)(args[ix].value);
                    if (!menuP)
                        break;
                    if ( menuP != w_main->menuBar ) {
                        if ( w_main->menuBar ) {
                            menuP->window = w_main->menuBar->window;
                            /* destroy old wdgt but leave window */
                            if ( AxPulldownActive == w_main->menuBar )
                                ResetActive();
                            ClipFromWindowList((AXWidget)w_main->menuBar);
                            free(w_main->menuBar->menu_strs);
                            free(w_main->menuBar->menu_starts);
                            /* null out window handle so that XtDestroy
                             * won't destroy it */
                            w_main->menuBar->window = 0;
                            AxXtDestroyWidget((AXWidget)w_main->menuBar);
                            menuP->update_widget = TRUE;
                            WidgetNeedsUpdated = TRUE;
                        }
                        w_main->menuBar = menuP;
                        AddToWindowList((AXWidget)menuP);
                    }
#endif /* } */
                    break;
                case XXmNresize:
                    w_main->variable_size = args[ix].value;
                    break;
                case XXmNmaxWidth:
                    w_main->maxWidth = args[ix].value;
                    break;
                case XXmNmaxHeight:
                    w_main->maxHeight = args[ix].value;
                    break;
                case XXmNworkWindow:
                    if ( !w_main->workWindow )
                        w_main->workWindow=(AXWidget_work)(args[ix].value);
                    break;
                case XXmNhorizontalScrollBar:
                    newP = (AXWidget)(args[ix].value);
                    SetHorizontalScrollBar((AXWidget)w_main,(AXWidget_scroll)newP);
                    break;
                case XXmNverticalScrollBar:
                    newP = (AXWidget)(args[ix].value);
                    SetVerticalScrollBar((AXWidget)w_main,(AXWidget_scroll)newP);
                    break;
                case XXmNcolorpal:
                    w_main->colorpal =(AXWidget_colorpal)args[ix].value;
                    break;
                case XXmNpalette:
                    w_main->palette =(AXWidget_palette)(args[ix].value);
                    break;
                case XXmNplates:
                    w_main->plates = (AXWidget_plates)(args[ix].value);
                    break;
                case XXmNruler:
                    w_main->ruler = (AXWidget_ruler)(args[ix].value);
                    break;
                case XXmNminWidth:
                    w_main->minWidth = args[ix].value;
                    break;
                case XXmNminHeight:
                    w_main->minHeight = args[ix].value;
                    break;
                case XXmNtopmostwindow: /* override-redirect */
                    if (w_main->dialog) {
                        w_main->dialog->topmostwindow =
                            args[ix].value ? TRUE : FALSE;
                    }
                    break;
                case XXmNmessageString:
                    SetNWidgetName(w,(char *)args[ix].value, MAX_NAME);
                    break;
                }
                break;
            case W_DIALOG:
                wd = (AXWidget_dialog)w;
                switch (att_code)
                    {
                    case XXmNdisableElfRecord:
                        /* w is of type W_DIALOG but disable_record
                         * property sits on parent W_MAIN */
                        w->parent->disable_record = args[ix].value;
                        break;
                    case XXmNnoExpand:
                        wd->noExpand = args[ix].value;
                        break;
                    case XXmNautoUnmanage:
                        wd->autoUnmanage = args[ix].value;
                        break;
                    case XXmNresize:
                        wd->variable_size = args[ix].value;
                        break;
                    case XXmNrevstamp:
                        wd->revstamp = args[ix].value;
                        break;
                    case XXmNminWidth:
                        wd->minWidth = args[ix].value;
                        break;
                    case XXmNminHeight:
                        wd->minHeight = args[ix].value;
                        break;
                    case XXmNmaxWidth:
                        wd->maxWidth = args[ix].value;
                        break;
                    case XXmNmaxHeight:
                        wd->maxHeight = args[ix].value;
                        break;
                    case XXmNdefaultButton:
                        wd->defaultButton =
                            (AXWidget_pushbutton)(args[ix].value);
                        wd->currentDefaultButton = wd->defaultButton;
                        wd->update_widget = WidgetNeedsUpdated = TRUE;
                        break;
                    case XXmNcancelButton:
                        wd->cancelButton =
                            (AXWidget_pushbutton)(args[ix].value);
                        break;
                    case XXmNpokeCodes:
                        if ( ElfbIsArray((data_ptr)args[ix].value) ) {
                            j = ElfbArraySize((data_ptr)args[ix].value);
                            if (wd->pokeCodes) {
                                free(wd->pokeCodes);
                            }
                            wd->pokeCodes = CreateIntsArray(j);
                            for ( j=0;j<IntsArraySize(wd->pokeCodes);j++) {
                                tempint = IntFromArray((data_ptr)args[ix].value,j);
                                WriteIntsArray(wd->pokeCodes,j,tempint);
                            }
                        }
                        else if (wd->pokeCodes) {
                            free(wd->pokeCodes);
                            wd->pokeCodes = (long *) NULL;
                        }
                        break;
                    case XXmNtopmostwindow: /* override-redirect */
                        wd->topmostwindow = args[ix].value ? TRUE : FALSE;
                        break;
                    case XXmNmessageString:
                        SetNWidgetName(w,(char *)args[ix].value, MAX_NAME);
                        break;
                    }
                break;
            case W_RADIO:
                w_radio = (AXWidget_radio)w;
                switch (att_code) {
                case XXmNupdate:
                    /* x, y, or num_buttons changed: unmap and update */
                    UnMapWidget((AXWidget)w_radio);
                    sz = args[ix].value;
                    /* free extra buttons if num_buttons decreased */
                    if ( sz > 0 && sz < w_radio->num_buttons ) {
                        for ( j=sz;j<w_radio->num_buttons; j++ ) {
                            AxXtDestroyWidget((AXWidget)w_radio->buttons[j]);
                        }
                        w_radio->num_buttons = sz;
                        sz *= sizeof(axw_toggle_button *);
                        w_radio->buttons = (AXWidget_togglebutton *)
                            realloc(w_radio->buttons, sz);
                    }
                    /* reset widths and heights  */
                    w_radio->width = 0;
                    w_radio->height = 0;
                    for ( j = 0; j < w_radio->num_buttons; j++ ) {
                        w_radio->buttons[j]->width = 0;
                        w_radio->buttons[j]->height = 0;
                    }
                    break;
                case XXmNmessageString:
                    SetNWidgetName(w,(char *)args[ix].value, MAX_NAME);
                    break;
                }
                break;
            case W_ENTER:
                w_enter = (AXWidget_entry)w;
                switch (att_code) {
                case XXmNcolumns:
                    w_enter->width = wEnterPixWidth(args[ix].value,
                                                    w_enter->font, w_enter->textShadow);
                    break;
                case XXmNmaxLength:
                    w_enter->maxLen = args[ix].value;
                    break;
                case XXmNtextOptions:
                    free(w_enter->validChars);
                    w_enter->validChars = 0;
                    cp = (char *)args[ix].value;
                    if (cp) {
                        sz = strlen((char *)cp)+1;
                        w_enter->validChars = strcpy((char *)malloc(sz), cp);
                    }
                    break;
                case XXmNmonospaceFont:
                    if (args[ix].value) {
                        if (w_enter->monospace_font == FALSE) {
                            w_enter->update_widget = TRUE;
                            WidgetNeedsUpdated = TRUE;
                            /* force a full-regen */
                            xmlibStr.FullRegenForUpdate = TRUE;
                        }
                        w_enter->monospace_font = TRUE;
                    }
                    else {
                        if (w_enter->monospace_font == TRUE) {
                            w_enter->update_widget = TRUE;
                            WidgetNeedsUpdated = TRUE;
                            /* force a full-regen */
                            xmlibStr.FullRegenForUpdate = TRUE;
                        }
                        w_enter->monospace_font = FALSE;
                    }
                    break;
                case XXmNmessageString:
                    SetWidgetString((AXWidget_entry)w, (char*)args[ix].value);
                    break;
                case XXmNedit_mask:
					/* the edit mask constrains input to the entry field.
					 * In particular, the field text string always has the
					 * same #chars as the edit mask
					 */
					if (AxIsArray((data_ptr) args[ix].value)) {
						/* array form: <<mask-string><masktype-int>>
						 */
						cp = AxStrFromArray((data_ptr) args[ix].value, 0);
					}
					else {
						/* string form: <mask-string>
						 */
						cp = AxStrFromDataPtr((data_ptr) args[ix].value);
					}
					if (cp && *cp) {
						void *tmask = w_enter->edit_win.editmask;
						if (!tmask || EditMaskDiffers(tmask, cp)) {
							/* nonnull edit mask arriving for 1st time, or
							 * new mask value.
							 */
							w_enter->edit_win.editmask = NewEditMask(cp);
							tmask = w_enter->edit_win.editmask;
							if (w_enter->wstr) {
								free(w_enter->wstr);
							}
							w_enter->wstr = EditMaskInitialValue(tmask);
						}
					}
					break;
                }
                break;
            case W_EDITBOX:
                w_editbox = (AXWidget_editbox)w;
                switch (att_code) {
                case XXmNno_picks:
                    w_editbox->no_picks = args[ix].value;
                    break;
                case XXmNnoPrimarySeln:
                    w_editbox->no_primary_seln = args[ix].value;
                    break;
                case XXmNmonospaceFont:
                    if (args[ix].value) {
                        if (w_editbox->monospace_font == FALSE) {
                            w_editbox->update_widget = TRUE;
                            WidgetNeedsUpdated = TRUE;
                            /* force a full-regen */
                            xmlibStr.FullRegenForUpdate = TRUE;
                        }
                        w_editbox->monospace_font = TRUE;
                    }
                    else {
                        if (w_editbox->monospace_font == TRUE) {
                            w_editbox->update_widget = TRUE;
                            WidgetNeedsUpdated = TRUE;
                            /* force a full-regen */
                            xmlibStr.FullRegenForUpdate = TRUE;
                        }
                        w_editbox->monospace_font = FALSE;
                    }
                    break;
                case XXmNmaxLength:
                    w_editbox->maxLen = args[ix].value;
                    break;
                case XXmNmessageString:
                    SetNWidgetName(w,(char *)args[ix].value, MAX_NAME);
                    break;
                }
                break;
            case W_SPLITTER:
                w_splitter = (AXWidget_splitter)w;
                switch (att_code) {
                case XXmNmaxHeight:
                    w_splitter->maxHeight = args[ix].value;
                    break;
                case XXmNmessageString:
                    SetNWidgetName(w,(char *)args[ix].value, MAX_NAME);
                    break;
                }
                break;
            case W_TOGGLE:
                w_toggle = (AXWidget_togglebutton)w;
                switch (att_code) {
                case XXmNset:
                    if ( args[ix].value )
                        SetRadioValue(w_toggle,FALSE);
                    w_toggle->value = args[ix].value;
                    break;
                case XXmNsmallFont:
                    w->small_font = TRUE;
                    break;
                case XXmNwhichButton:
                    w_toggle->whichButton = args[ix].value;
                    break;
                case XXmNorientation:
                    /* note: , this field holds THREE_STATE_TOGGLE */
                    w_toggle->scroll_type = args[ix].value;
                    break;
                case XXmNmessageString:
                    SetNWidgetName(w,(char *)args[ix].value, MAX_NAME);
                    break;
                }
                break;
            case W_PAGEBAR:
                w_page = (AXWidget_page)w;
                switch (att_code) {
                case XXmNmaximum:
                    w_page->scroll_max = args[ix].value;
                    w_page->update_widget = WidgetNeedsUpdated = TRUE;
                    break;
                case XXmNmessageString:
                    SetNWidgetName(w,(char *)args[ix].value, MAX_NAME);
                    break;
                }
                break;
            case W_SCROLL:
                w_scroll = (AXWidget_scroll)w;
                switch (att_code) {
                case XXmNbuttonUpScroll:
                    w_scroll->buttonUpScroll = args[ix].value;
                    break;
                case XXmNmaximum:
                    w_scroll->scroll_max = args[ix].value;
                    w_scroll->update_widget = WidgetNeedsUpdated = TRUE;
                    break;
                case XXmNorientation:
                    w_scroll->scroll_type = (short) args[ix].value;
                    break;
                case XXmNsliderSize:
                    w_scroll->scroll_per = args[ix].value;
                    w_scroll->update_widget = WidgetNeedsUpdated = TRUE;
                    break;
                case XXmNmessageString:
                    SetNWidgetName(w,(char *)args[ix].value, MAX_NAME);
                    break;
                }
                break;
            case W_RIBBON:
                w_rbn = (AXWidget_ribbon)w;
                switch (att_code) {
                case XXmNpixmap:
                    if (w_rbn->pixmap != (Pixmap) args[ix].value) {
                        if (w_rbn->pixmap)
                            AxFreePixmap(w_rbn->pixmap);
                        w_rbn->pixmap = (Pixmap) args[ix].value;
                    }
                    break;
                case XXmNsliderSize:
                    w_rbn->scroll_per = args[ix].value;
                    w_rbn->update_widget = WidgetNeedsUpdated = TRUE;
                    break;
                case XXmNmaximum:
                    w_rbn->scroll_max = args[ix].value;
                    w_rbn->update_widget = WidgetNeedsUpdated = TRUE;
                    break;
                case XXmNselectedItemCount:
                    free(w_rbn->selections);
                    w_rbn->selections = CreateIntsArray(args[ix].value);
                    break;
                case XXmNselectedItems:
                    if ( w_rbn->elf_access ) {
                            if ( ElfbIsArray((data_ptr)args[ix].value) ) {
                                sz = IntsArraySize(w_rbn->selections);
                                for (j=0;j<sz; j++ ) {
                                    tempint = IntFromArray((data_ptr)args[ix].value,j);
                                    WriteIntsArray(w_rbn->selections,j,tempint);
                                }
                            }
                            else {
                                tempint = IntFromDataPtr((data_ptr)args[ix].value);
                                WriteIntsArray(w_rbn->selections,0,tempint);
                            }
                        }
                    else {
                        WriteIntsArray(w_rbn->selections,0,args[ix].value);
                    }
                    w_rbn->update_widget = TRUE;
                    WidgetNeedsUpdated = TRUE;
                    break;
                case XXmNmessageString:
                    sz = strlen((char *)args[ix].value)+1;
                    w_rbn->wstr = (wchar *) malloc(sz * sizeof(wchar));
                    MbToWStr((mbchar *)args[ix].value, w_rbn->wstr);
                    break;
                }
                break;
                case W_POPUP:
                w_popup = (AXWidget_popup)w;
                switch (att_code) {
                case XXmNwhichButton:
                    w_popup->whichButton = args[ix].value;
                    break;
                case XXmNlabelPixmap:
                    w_popup->pixmapLabel = args[ix].value;
                    break;
                case XXmNgePaletteButtonType:
                    w_popup->gePaletteButtonType = args[ix].value;
                    break;
                case XXmNmessageString:
                    SetNWidgetName(w,(char *)args[ix].value, MAX_NAME);
                    break;
                }
                break;
            case W_BUTTON:
                wp = (AXWidget_pushbutton)w;
                switch (att_code) {
                case XXmNpixmap:
                    wp->pixmap = (Pixmap)args[ix].value;
                    break;
                case XXmNlabelPixmap:
                    wp->pixmapLabel = args[ix].value;
                    break;
                case XXmNgePaletteButtonType:
                    wp->gePaletteButtonType = args[ix].value;
                    break;
                case XXmNwhichButton:
                    wp->whichButton = args[ix].value;
                    break;
                case XXmNalternatePixmap:
                    cp = (char *)args[ix].value;
                    free(wp->alternatePixmap);
                            if ( cp ) {
                                sz = strlen(cp)+1;
                                wp->alternatePixmap = strcpy(malloc(sz),cp);
                            }
                            else
                                wp->alternatePixmap = (char *)NULL;
                    break;
                case XXmNbuttonType:
                    wp->buttonType = args[ix].value;
                    break;
                case XXmNnoArmHighlight:
                    wp->no_arm_reverse = TRUE;
                    break;
                case XXmCArmCallback:
                    wp->arm_callback = TRUE;
                    break;
                case XXmNmessageString:
                    SetNWidgetName(w,(char *)args[ix].value, MAX_NAME);
                    break;
                }
                break;
            case W_LIST:
                wl = (AXWidget_list)w;
                switch (att_code) {
                case XXmNcurPos:
                    /* save old curPos in maxWidth */
                    wl->maxWidth = wl->curPos;
                    wl->curPos = (short) args[ix].value;
                    break;
                case XXmNno_picks:
                    wl->no_picks = args[ix].value;
                    break;
                case XXmNbuttonUpScroll:
                    /* This attribute is applicable to child scrollbars
                       of the table and list widgets.  The list might not
                       have scrollbars at this point.  We therefore set
                       `buttonUpScroll' here.  The list widgets will
                       propogate this value to their scrollbars at
                       Realize() time. */
                    wl->buttonUpScroll = args[ix].value;
                    
                    if (wl->hScrollBar)
                        wl->hScrollBar->buttonUpScroll = args[ix].value;
                    if (wl->vScrollBar)
                        wl->vScrollBar->buttonUpScroll = args[ix].value;
                    break;
                case XXmNdoubleClickInterval:
                    wl->double_click = args[ix].value;
                    break;
                case XXmNresize:
                    /* HACK for listbox resize; get the list to
                     * recompute scrollbar geometry */
                    if ( args[ix].value ) {
                        UnMapWidget((AXWidget)wl); /* destroy window */
                        if ( wl->vScrollBar ) {
                            AxXtDestroyWidget((AXWidget)wl->vScrollBar);
                            wl->vScrollBar = (AXWidget_scroll) NULL;
                        }
                        if ( wl->hScrollBar ) {
                            AxXtDestroyWidget((AXWidget)wl->hScrollBar);
                            wl->hScrollBar = (AXWidget_scroll ) NULL;
                        }
                    }
                    break;
                case XXmNhideListVScroll:
                    wl->hide_vscroll = args[ix].value;
                    break;
                    
                case XXmNlistHScroll:
                    if ( wl->list_hscroll != args[ix].value ) {
                        wl->list_hscroll = args[ix].value;
                        if ( wl->list_hscroll ) {
                            /* get max text width */
                            if ( wl->maxLen == 0 ) /* others XmNitems*/
                                wl->maxLen =(short)ListStrsMaxWidth(wl);
                        }
                        else {
                            wl->maxLen = 0;
                        }
                    }
                    break;
                case XXmNselectedItemCount:
                    /* must precede selectedItems */
                    free(wl->selections);
                    wl->selections = CreateIntsArray(args[ix].value);
                    break;
                case XXmNselectedItems:
                    if ( wl->elf_access ) {
                        if ( ElfbIsArray((data_ptr)args[ix].value) ) {
                            for (j=0;j<IntsArraySize(wl->selections);j++ ) {
                                tempint = IntFromArray((data_ptr)args[ix].value,j);
                                WriteIntsArray(wl->selections,j,tempint);
                            }
                        }
                        else {
                            tempint = IntFromDataPtr((data_ptr)args[ix].value);
                            WriteIntsArray(wl->selections,0,tempint);
                        }
                    }
                    else {
                        WriteIntsArray(wl->selections,0,args[ix].value);
                    }
                    wl->update_widget = TRUE;
                    WidgetNeedsUpdated = TRUE;
                    break;
                case XXmNmaximum:
                    wl->scroll_max = args[ix].value;
                    wl->update_widget = WidgetNeedsUpdated = TRUE;
                    break;
                case XXmNvisibleItemCount:
                    wl->scroll_max = args[ix].value;
                    wl->height = wListPixHeight(args[ix].value,
                                                wl->font,
                                                wl->textShadow,
                                                wl->list_hscroll);
                    break;
                case XXmNmonospaceFont:
                    if (args[ix].value ) {
                        if (wl->monospace_font == FALSE) {
                            wl->update_widget = TRUE;
                            WidgetNeedsUpdated = TRUE;
                            /* force a full-regen */
                            xmlibStr.FullRegenForUpdate = TRUE;
                        }
                        wl->monospace_font = TRUE;
                    }
                    else {
                        if (wl->monospace_font == TRUE) {
                            wl->update_widget = TRUE;
                            WidgetNeedsUpdated = TRUE;
                            /* force a full-regen */
                            xmlibStr.FullRegenForUpdate = TRUE;
                        }
                        wl->monospace_font = FALSE;
                    }
                    break;
                    
                case XXmNforce_list_item_display:
                    wl->force_list_item_display = TRUE;
                    wl->update_widget = TRUE;
                    WidgetNeedsUpdated = TRUE;
                    break;
                case XXmNmessageString:
                    SetNWidgetName(w,(char *)args[ix].value, MAX_NAME);
                    break;
                }
                break;
            case W_RECT:
                w_rect = (AXWidget_rect) w;
                switch (att_code) {
                case XXmNmessageString:
                    SetNWidgetName(w,(char *)args[ix].value, MAX_NAME);
                    if (RECT_LABELED == w_rect->style) {
                        w_rect->update_widget =
                            WidgetNeedsUpdated = TRUE;
                    }
                    break;
                            
                case XXmNtablayers:
                    SetTabLayers(w_rect, (data_ptr) args[ix].value, FALSE);
                    w_rect->update_widget = WidgetNeedsUpdated = TRUE;
                    break;
                            
                case XmNactiveTablayer:
                    if (w_rect->tabctl.active_layer !=
                        (int) args[ix].value) {
                        w_rect->tabctl.active_layer =
                            (int) args[ix].value;
                        w_rect->update_widget =
                            WidgetNeedsUpdated = TRUE;
                    }
                    break;
                }
                break;
            case W_TRAY:
                w_tray = (AXWidget_tray)w;
                switch (att_code) {
                case XXmNhideListVScroll:
                    w_tray->hide_vscroll = args[ix].value;
                    break;
                case XXmNlistHScroll:
                    if ( w_tray->list_hscroll != args[ix].value ) {
                        w_tray->list_hscroll = args[ix].value;
                        w_tray->maxLen = 0;
                    }
                    break;
                case XXmNresize:
                    /* HACK for listbox resize; get the list to
                     * recompute scrollbar geometry */
                    if ( args[ix].value ) {
                        UnMapWidget((AXWidget)w_tray); /* destroy window */
                        if ( w_tray->vScrollBar ) {
                            AxXtDestroyWidget((AXWidget)w_tray->vScrollBar);
                            w_tray->vScrollBar = (AXWidget_scroll)NULL;
                        }
                        if ( w_tray->hScrollBar ) {
                            AxXtDestroyWidget((AXWidget)w_tray->hScrollBar);
                            w_tray->hScrollBar = (AXWidget_scroll)NULL;
                        }
                    }
                    break;
                case XXmNmessageString:
                    SetNWidgetName(w,(char *)args[ix].value, MAX_NAME);
                    break;
                }
                break;
            case W_CANVAS:
                w_c = (AXWidget_canvas)w;
                switch (att_code) {
                case XXmNhideListVScroll:
                    w_c->hide_vscroll = args[ix].value;
                    break;
                case XXmNlistHScroll:
                    if ( w_c->list_hscroll != args[ix].value ) {
                        w_c->list_hscroll = args[ix].value;
                        w_c->maxLen = 0;
                    }
                    break;
                case XXmNresize:
                    /* HACK for listbox resize; get the list to
                     * recompute scrollbar geometry */
                    if ( args[ix].value ) {
                        UnMapWidget((AXWidget)w_c); /* destroy window */
                        if ( w_c->vScrollBar ) {
                            AxXtDestroyWidget((AXWidget)w_c->vScrollBar);
                            w_c->vScrollBar = (AXWidget_scroll)NULL;
                        }
                        if ( w_c->hScrollBar ) {
                            AxXtDestroyWidget((AXWidget)w_c->hScrollBar);
                            w_c->hScrollBar = (AXWidget_scroll)NULL;
                        }
                    }
                    break;
                case XXmNmessageString:
                    SetNWidgetName(w,(char *)args[ix].value, MAX_NAME);
                    break;
                }
                break;
            case W_ICON:
                w_icon = (AXWidget_icon)w;
                switch (att_code) {
                case XXmNmaxWidth:
                    w_icon->maxWidth = args[ix].value;
                    break;
                case XXmNmaxHeight:
                    w_icon->maxHeight = args[ix].value;
                    break;
                case XXmNpixmap:
                    w_icon->pixmap = (Pixmap)args[ix].value;
                    break;
                case XXmNmessageString:
                    SetNWidgetName(w,(char *)args[ix].value, MAX_NAME);
                    break;
                }
                break;
            case W_SCALE:
                w_scale = (AXWidget_scale)w;
                switch (att_code) {
                case XXmNminValue:
                    w_scale->minValue = args[ix].value;
                    break;
                case XXmNmaxValue:
                    w_scale->maxValue = args[ix].value;
                    break;
                case XXmNno_label:
                    w_scale->no_label = (args[ix].value != 0);
                    break;
                case XXmNmessageString:
                    SetNWidgetName(w,(char *)args[ix].value, MAX_NAME);
                    break;
                }
                break;
            case W_ROWCOL:
                wr = (AXWidget_rowcol)w;
                switch (att_code) {
                case XXmNpfg_color:
                    wr->color = (int) args[ix].value;
                    wr->update_widget = WidgetNeedsUpdated = TRUE;
                    break;
                case XXmNpbg_color:
                    wr->bg_color = (int) args[ix].value;
                    w->update_widget = WidgetNeedsUpdated = TRUE;
                    break;
                case XXmNwhichButton:
                    wr->whichButton = args[ix].value;
                    break;
                case XXmNminValue:
                    /* holds button text relative x position. */
                    wr->minValue = args[ix].value;
                    break;
                case XXmNmaxValue:
                    /* holds button text relative y position. */
                    wr->maxValue = args[ix].value;
                    break;
                case XXmNalternatePixmap:
                    cp = (char*)args[ix].value;
                    free(wr->alternatePixmap);
                    if ( cp )
                        wr->alternatePixmap = strcpy(malloc(strlen(cp)+1),cp);
                    else
                        wr->alternatePixmap = (char *)NULL;
                    break;
                case XXmNgePaletteButtonType:
                    wr->gePaletteButtonType = args[ix].value;
                    break;
                case XXmNorientation:
                    /* note: this field holds the number of columns . */
                    wr->scroll_type = args[ix].value;
                    break;
                case XXmNsliderSize:
                    /* 0 means normal pixmap buttons.
                       != 0 means text buttons (for eqn editor). */
                    wr->scroll_per = args[ix].value;
                    break;
                case XXmNlabelPixmap:
                    wr->pixmapLabel = args[ix].value;
                    break;
                case XXmNbtnPixmapList:
                    {
                        char **pixmap_names = (char **) args[ix].value;
                        for (j=0; j<wr->num_buttons; j++) {
                            wr->buttons[j] = (AXWidget_pushbutton)
                                InitWidget(W_BUTTON);
                            SetWidgetName((AXWidget)wr->buttons[j], pixmap_names[j]);
                        }
                    }
                    break;
                case XXmNpixmask:
                    break;
                case XXmNmessageString:
                    SetNWidgetName(w,(char *)args[ix].value, MAX_NAME);
                    break;
                }
                break;
            case W_CHOOSE:
                w_choose = (AXWidget_color_chooser)w;
                switch (att_code) {
                case XXmNhideListVScroll:
                    w_choose->hide_vscroll = args[ix].value;
                    break;
                case XXmNmessageString:
                    SetNWidgetName(w,(char *)args[ix].value, MAX_NAME);
                    break;
                }
                break;
            case W_TABLE:
                w_t = (AXWidget_table)w;
                switch (att_code) {
                    /* the following group of things is designed to 
                       fall through to make one generic function call */
                case XXmNmarkerWidth:
                case XXmNallowHiddenColumns:
                case XXmNallowColumnResizing:
                case XXmNsuppressRowMarkers:
                case XXmNallowColumnReordering:
                case XXmNhideVerticalGrid:
                case XXmNhideHorizontalGrid:
                case XXmNuseDataRequestModel:
                case XXmNalisStyleScroll:
                case XXmNvscrollLength:
                case XXmNvscrollOrigin:
                case XXmNhscrollLength:
                case XXmNhscrollOrigin:
                case XXmNshowRowNumbers:
                case XXmNRowColCreate:
                    SetTableAttributes((AXWidget)w_t,att_code,args[ix].value);
                    break;
                case XXmNhideListVScroll:
                    w_t->hide_vscroll = args[ix].value;
                    break;
                    
                case XXmNlistHScroll:
                    w_t->list_hscroll = args[ix].value;
                    break;
                case XXmNbuttonUpScroll:
                    /* This attribute is applicable to child scrollbars
                       of the table .  The table might not have scrollbars
                       at this point.  We therefore generically set
                       `buttonUpScroll' here.  The table widgets will
                       propogate this value to their scrollbars at 
                       Realize() time. */
                    w_t->buttonUpScroll = args[ix].value;
                    
                    /* If table, set attribute on scrollbars as well */
                    /* Set attr if there is a v/h scrollbar */
                    if (w_t->hScrollBar)
                        w_t->hScrollBar->buttonUpScroll = args[ix].value;
                    if (w_t->vScrollBar)
                        w_t->vScrollBar->buttonUpScroll = args[ix].value;
                    break;
                case XXmNmaximum:
                    w_t->scroll_max = args[ix].value;
                    w_t->update_widget = WidgetNeedsUpdated = TRUE;
                    break;
                case XXmNresize:
                    /* HACK for listbox resize; get the list to
                     * recompute scrollbar geometry */
                    if ( args[ix].value ) {
                        UnMapWidget((AXWidget)w_t); /* destroy window */
                        if ( w_t->vScrollBar ) {
                            AxXtDestroyWidget((AXWidget)w_t->vScrollBar);
                            w_t->vScrollBar = (AXWidget_scroll) NULL;
                        }
                        if ( w_t->hScrollBar ) {
                            AxXtDestroyWidget((AXWidget)w_t->hScrollBar);
                            w_t->hScrollBar = (AXWidget_scroll) NULL;
                        }
                    }
                    break;
                case XXmNhorizontalScrollBar:
                    newP = (AXWidget)(args[ix].value);
                    SetHorizontalScrollBar((AXWidget)w_t,(AXWidget_scroll)newP);
                    break;
                case XXmNverticalScrollBar:
                    newP = (AXWidget)(args[ix].value);
                    SetVerticalScrollBar((AXWidget)w_t,(AXWidget_scroll)newP);
                    break;
                case XXmNvisibleItemCount:
                    w_t->height = wTablePixHeight(args[ix].value,
                                                  w_t->thickness);
                    break;
                case XXmNlabelColor:
                    {
                        table_info *tp;
                        tp = (table_info *)w_t->info.table;
                        if (tp && tp->headingColor) {
                            AxFreeWColor(tp->headingColor);
                        }
                        if ( tp )
                            tp->headingColor =
                                AxCreateWColor(w, (data_ptr) args[ix].value);
                    }
                    break;
                case XXmNlabelFont:
                    TblSetHeadingFont(w_t, (data_ptr)args[ix].value);
                    break;
                case XXmNlabelShadow:
                    {
                        table_info *tp;
                        tp = (table_info *)w_t->info.table;
                        if (tp && tp->headingShadow) {
                            AxTaskFreeElfData(0, tp->headingShadow);
                            tp->headingShadow = (data_ptr) NULL;
                        }
                        if (tp && args[ix].value) {
                            tp->headingShadow =
                                AxTaskCopyElfData(0, (data_ptr)args[ix].value);
                            if (tp->headingFont) {
                                int xoff, yoff;
                                xmTextShadowOffsets(tp->headingFont,
                                                    &xoff, &yoff);
                                tp->heading_char_below =
                                    tp->headingFont->descent + yoff + 2;
                            }
                        }
                    }
                    break;
                case XXmNmessageString:
                    SetNWidgetName(w,(char *)args[ix].value, MAX_NAME);
                    break;
                }
                break;
            case W_OPTION:
                w_option = (AXWidget_option)w;
                switch (att_code) {
                case XXmNmenuHistory:
                    w_option->optionValue = (AXWidget)(args[ix].value);
                    break;
                case XXmNmessageString:
                    SetNWidgetName(w,(char *)args[ix].value, MAX_NAME);
                    break;
#if AXMS
                case XXmNlabelColor:
                    if (w_option->labelColor)
                        AxFreeWColor(w_option->labelColor);
                    w_option->labelColor =
                        AxCreateWColor(w, (data_ptr)args[ix].value);
                    break;
#endif
                }
                break;
            case W_COMBO_BOX:
                w_cbox = (AXWidget_combo_box)w;
                switch (att_code) {
                    
                case XXmNhorizontalScrollBar:
                    newP = (AXWidget)(args[ix].value);
                    SetHorizontalScrollBar((AXWidget)w_cbox,(AXWidget_scroll)newP);
                    break;
                case XXmNverticalScrollBar:
                    newP = (AXWidget)(args[ix].value);
                    SetVerticalScrollBar((AXWidget)w_cbox,(AXWidget_scroll)newP);
                    break;
                case XXmNbuttonUpScroll:
                    w_cbox->buttonUpScroll = args[ix].value;
                    
                    /* If table, set attribute on scrollbars as well */
                    /* Set attr if there is a v/h scrollbar */
                    if (w_cbox->hScrollBar) {
                        w_cbox->hScrollBar->buttonUpScroll = args[ix].value;
                    }
                    if (w_cbox->vScrollBar) {
                        w_cbox->vScrollBar->buttonUpScroll = args[ix].value;
                    }
                    break;
                case XXmNmessageString:
                    SetNWidgetName(w,(char *)args[ix].value, MAX_NAME);
                    break;
                }
                break;
            }
        }
    }
}


data_ptr ElfGetSysVar();

static data_ptr preloadJavaImgsFromTreeData (nameBuf, aP)
char *nameBuf;
data_ptr aP;
{
    register int ix = 0;
    register char *name = AxStrFromArray(aP, 1);
    int asize = AxArraySize(aP);
    if (0 == asize)
        return NULL;
    memcpy(nameBuf+6,name,strlen(name));
    nameBuf[6+strlen(name)] = 0;
    if (!ElfGetSysVar(nameBuf)) {
        data_ptr strP = ELF_CREATE_STRING(nameBuf+6);
        ElfCallByName("POST_JAVA_IMAGE$", NULL, 1, strP);
        ELF_DATA_FREE(strP);
    }
    aP = AxArrayElement(aP,4);
    asize = AxArraySize(aP);
    for (ix = 0; ix < asize; ++ix)
        preloadJavaImgsFromTreeData(nameBuf, AxArrayFromArray(aP,ix));
    return NULL;
}

/*************************************************************************
function: AxfPreloadJavaImgsFromStrs - Scan listbox strs for bitmaps
*/
data_ptr AxfPreloadJavaImgsFromStrs (data_ptr argDatum)
{
    register data_ptr strs = AxArrayElement(argDatum,0);
    register int isTray = (0 != AxIntFromArray(argDatum,1));
    register data_ptr strP;
    register int ix;
    register int size = ElfbArraySize(strs); /* also validates is actually an array */
    char nameBuf[6+STRNORMAL+1];
    char *p, *name, *line;
    data_ptr retval = NULL;

    strcpy(nameBuf,"jv$im$");
    if (isTray) {
        return preloadJavaImgsFromTreeData(nameBuf, strs);
    }
    for ( ix = 0; ix < size; ix++ ) {
        strP = AX(strs)->vals[ix];
        if ( strP && !IS_INTEGER(strP) && (SX(strP)->type == TYPE_STRING) ) {
            line = SX(strP)->text;
            while ( p = strstr(line,"\\i(") ) { /* isolate name... */
                name = p + 3;
                line = name;
                p = strchr(name,')');
                if ( p && (p-name <= STRNORMAL) ) {
                    memcpy(6+nameBuf,name,p-name);
                    nameBuf[6+p-name] = 0;
                    if ( !ElfGetSysVar(nameBuf) ) {
                        strP = ELF_CREATE_STRING(nameBuf+6);
                        ElfCallByName("POST_JAVA_IMAGE$", NULL, 1, strP);
                        ELF_DATA_FREE(strP);
                    }
                }
            }
        }
    }
    return(retval);
}



/*************************************************************************
function: ElfbWidgetProfile - Set volume, speed, color stuff
*/
void
ElfbWidgetProfile (volume,click_speed,use_toggles,narrow_menu,accels,
           pointer_curs,busy_curs,noTrans,
           d2mode,icon_size,mark_menu_mnemonics, wmfocus)
     int volume,click_speed;
     int use_toggles,narrow_menu,accels;
     char *pointer_curs, *busy_curs;
     int noTrans;
     int d2mode;
     char *icon_size;
     int mark_menu_mnemonics;
     int wmfocus;
{
  ThimBellVolume = 2 * (MAX(0,MIN(100,volume))) - 100;
  DoubleMillisec = click_speed;
  UseChecksForToggles = !use_toggles;
  AlwaysWide = !narrow_menu;
  ShowAccelerators = accels;
  ZXmBindCursor(pointer_curs,&ThimPointerCursor);
  ZXmBindCursor(busy_curs,&ThimBusyCursor);
  WinMask |= CWCursor;
  WinAtt.cursor = ThimPointerCursor;
  AXmRestoreNormalCursorPointer(0);
#if AX_X11
  if ( AxMarkMenuMnemonics != mark_menu_mnemonics ) {
    AxMarkMenuMnemonics = mark_menu_mnemonics;
    AxRePaintAllMenuBars();
  }
#endif
  if ( (Thim2dMode && !d2mode) || (!Thim2dMode && d2mode) ) {
    Thim2dMode = d2mode;
    AxRePaintAll(0);
  }
  NoTransient = noTrans;
  IconSizeStr = strcpy(malloc(strlen(icon_size)+1),icon_size);
  AxNoFocusManage = wmfocus;
}


int
xmWidgetKeyPressHandler(wp, eventP, ksym, kbuf, num_chars, in_menu_bar)
AXWidget wp;       /* in: widget getting the keypress */
XEvent *eventP;    /* in: keypress event */
KeySym ksym;
wchar *kbuf;       /* in: key string associated with ksym, from LookupString */
int num_chars;     /* num_chars in kbuf */
int in_menu_bar;   /* bool: passed ksym has a menubar binding */
{
	int consumed = FALSE;
    
	if (!wp) {
		return (FALSE);
	}
  
	if (wp->flags & WFLG_IN_EDIT) {
		return (FALSE); /* widget-specific key input off during dialog edit */
	}
    
	switch (wp->type) {
	case W_SCALE:
		if (in_menu_bar) {
			break;
		}
		consumed = ScaleKeyPressEvent((AXWidget_scale)wp, eventP, ksym, kbuf, 
																	num_chars);
		break;

	case W_RECT:
		if (in_menu_bar) {
			break;
		}
		if (RECT_LAYERED == wp->style) {
			consumed = TabctlKeyPressHandler((AXWidget_rect)wp, eventP, ksym, kbuf, num_chars);
		}
		break;

	case W_RADIO:
		if (in_menu_bar) {
			break;
		}
		consumed = RadioKeyPressEvent((AXWidget_togglebutton)wp, eventP, ksym, kbuf, num_chars);
		break;

#if AX_X11    /* { */
	case W_TABLE:
#if 0				/* Allow TblKeyPressCB to do Euro check first */
		if (in_menu_bar && (ksym != XK_Tab)) {
			break;
		}
#endif
		consumed = TblKeyPressCB((AXWidget_table)wp, eventP, in_menu_bar);
		break;
        
	case W_TRAY:
		if (in_menu_bar) {
			break;
		}
		consumed = TrayEvent((AXWidget_tray)wp, eventP);
		break;
        
	case W_CHOOSE:
		if (in_menu_bar) {
			break;
		}
		consumed = ColorChooserEvent((AXWidget_color_chooser)wp, eventP);
		break;
        
	case W_CANVAS:
		if (in_menu_bar) {
			break;
		}
		consumed = CanvasEvent((AXWidget_canvas)wp, eventP);
		break;
        
	case W_LIST:
		if (in_menu_bar) {
			break;
		}
		consumed = ListEvent((AXWidget_list)wp, eventP);
		break;

	case W_BUTTON:
		if (in_menu_bar) {
			break;
		}
		consumed = VerbButtonKeyPressEvent((AXWidget_pushbutton)wp, eventP, ksym, kbuf, num_chars);
		break;
        
	case W_COMBO_BOX:
		consumed = ComboKeyPressEvent((AXWidget_combo_box)wp, eventP, ksym, kbuf,num_chars, in_menu_bar);
		break;
        
	case W_TOGGLE:
		if (in_menu_bar) {
			break;
		}
		consumed = ToggleKeyPressEvent((AXWidget_togglebutton)wp, eventP, ksym, kbuf, num_chars);
		break;
        
	case W_OPTION:
		if (in_menu_bar) {
			break;
		}
		consumed = OptionKeyPressEvent((AXWidget_option)wp, eventP, ksym, kbuf, num_chars);
		break;
        
	case W_ENTER:
#if 0				/* Allow EntKeyPressCB to do Euro check first */
		if (in_menu_bar) {
			break;
		}
#endif
		ArmButton(NULL, FALSE, eventP); /* if needed at all, why only here? */
		consumed = EntKeyPressCB((AXWidget_entry)wp, eventP, in_menu_bar);
		break;
        
	case W_EDITBOX:
#if 0				/* Allow editboxKeyEvent to do Euro check first */
		if (in_menu_bar && !(XK_Return == ksym || XK_Execute == ksym)) {
			break;
		}
#endif
		consumed = editboxKeyEvent((AXWidget_editbox)wp, eventP, in_menu_bar);
		break;
#endif    /* } AX_X11 */
        
	}
	return (consumed);
}

static AXWidget
findInKeyFocusLinks(d_w, oldP, next)
AXWidget_dialog d_w;
AXWidget oldP;
int next; /* TRUE for next, FALSE for prev */
{
	int dex, iasize, iax;
	AXWidget w, focusWidget;

	if (!d_w || !oldP)
		return(NULL);

	if (!d_w->keyFocusLinks)
		return(NULL);
	if (!InIntsArray(d_w->keyFocusLinks, (long)oldP))
		return(NULL);
	focusWidget = oldP;
	iasize = IntsArraySize(d_w->keyFocusLinks);

findagain:
	for (iax = 0; iax < iasize; iax++) {
		w = (AXWidget) ReadIntsArray(d_w->keyFocusLinks, iax);
		if (w != focusWidget)
			continue;
		if (next == TRUE) {
			iax++;
			if (iax >= iasize)
				return(NULL);
		} else {
			iax--;
			if (iax < 0)
				return(NULL);
		}
		w = (AXWidget) ReadIntsArray(d_w->keyFocusLinks, iax);
		if (w != NULL) {
			if (xmCanTakeKeyFocus(d_w, w, FALSE)) {
				return(w);
			} else {
				/* Widget is probably inactive -- find the one after it */
				focusWidget = w;
				goto findagain;
			}
		}
		/* w is NULL here.  It is possible that we have traversed the
		   keyFocusLinks list from oldP onward, and hit inactive widgets
		   all along.  We will return focusWidget in that case.
		   The caller will then verify that focusWidget passes
		   xmCanTakeKeyFocus().  If it does not, it will find the next
		   geometrically placed widget after focusWidget */
		if (oldP != focusWidget) { /* focusWidget can never be null */
			return(focusWidget);
		} /* else, break -- focusWidget is the first (or last) in the list
			 of keyFocusLinks. */
		break;
	}
	return(NULL);
}

/*************************************************************************
function: TabToNextWidget - Default response to tab key by focusing "next"
W_ENTER or W_EDITBOX widget.
*/
AXWidget
TabToNextWidget(oldP, parentP)
     AXWidget oldP;
     AXWidget parentP;   /* oldP->parent (dialog if oldP is null) */
{
  char *prof = UPbyName(NXLT("axDlgTabVertically"));
    
  /* no double/triple/etc. counting for radio buttons, instead, just refer
	 to the label (of type W_RADIO) */
  if (oldP && oldP->type == W_TOGGLE &&
	  ((AXWidget_togglebutton)oldP)->in_radio != NULL) {
	  oldP = (AXWidget)((AXWidget_togglebutton)oldP)->in_radio;
  }

  if (prof && atoi(prof)) {
    return(TabToNextWidgetVert(oldP, parentP));
  }
  else {
    return(TabToNextWidgetHoriz(oldP, parentP));
  }
}

/*************************************************************************
function: TabToPrevWidget - Default response to tab key by focusing "prev"
W_ENTER or W_EDITBOX widget.
*/
AXWidget 
TabToPrevWidget (oldP, parentP)
     AXWidget oldP;
     AXWidget_any parentP;   /* oldP->parent (dialog if oldP is null) */
{
  char *prof = UPbyName(NXLT("axDlgTabVertically"));

  /* no double/triple/etc. counting for radio buttons, instead, just refer
	 to the label (of type W_RADIO) */
  if (oldP && oldP->type == W_TOGGLE &&
	  ((AXWidget_togglebutton)oldP)->in_radio != NULL) {
	  oldP = (AXWidget)((AXWidget_togglebutton)oldP)->in_radio;
  }

  if (prof && atoi(prof)) {
    return(TabToPrevWidgetVert(oldP, parentP));
  }
  else {
    return(TabToPrevWidgetHoriz(oldP, parentP));
  }
}

AXWidget 
TabToNextWidgetVert (oldP,parentP)
     AXWidget oldP;
     AXWidget_any parentP;   /* oldP->parent (dialog if oldP is null) */
{
  register AXWidget best_w, w;
  register AXWidget_dialog d_w;

  if ( !parentP )
    return(NULL);
    
  if ( parentP->type == W_MAIN )
    d_w = ((AXWidget_main)parentP)->dialog;
  else
    d_w = (AXWidget_dialog)parentP;

  best_w = NULL;
    
  if (oldP) {
    for ( w = d_w->widgets; w; w = w->nextChild ) {
      if (xmCanTakeKeyFocus(d_w, w, FALSE)) {

    /* make radio toggles look like the radio label's geometry */
	if (w->type == W_TOGGLE &&
		((AXWidget_togglebutton)w)->in_radio != NULL) {
		w = (AXWidget)((AXWidget_togglebutton)w)->in_radio;
	}
                
    if ( (w == oldP) || (w->xpos < oldP->xpos) )
      continue;  /* reject old itself, or any left of old */
                
    if ( w->xpos == oldP->xpos ) {
      if ( w->ypos <= oldP->ypos )
        continue;        /* reject those above old */
      else if ( best_w ) {
        if ( w->ypos < best_w->ypos ) {
          /* prefer topmost of those below
           */
          best_w = w;
        }
      }
    }
    if ( !best_w ) {
      best_w = w; /* first found at or below old */
    }
    if ( w->xpos < best_w->xpos ) {
      /* prefer leftmost of those at or right of old
       */
      best_w = w;
    }
    else if ( (w->xpos == best_w->xpos) &&
          (w->ypos <= best_w->ypos) ) {

      /* prefer topmost of those at same x as best
       */
      best_w = w;
    }
      }
    }
  }
  if (best_w || (best_w = axTopLeftmostWidget(d_w))) {
    xmTakeKeyFocus(best_w, NULL, (W_ENTER == best_w->type));
#if AXMS
	/* if we have a radio array, make sure we set the focus on the
	 * currently selected toggle button.
	 */
	if ( best_w->type == W_RADIO )
		best_w = xmRadioActiveToggle(best_w);
#endif
    return(best_w);
  }
  else {
    return (NULL);
  }
}

AXWidget 
TabToPrevWidgetVert (oldP, parentP)
     AXWidget oldP;
     AXWidget_any parentP;   /* oldP->parent (dialog if oldP is null) */
{
  register AXWidget best_w, w;
  register AXWidget_dialog d_w;

  if ( !parentP )
    return(NULL);
  if ( parentP->type == W_MAIN )
    d_w = ((AXWidget_main)parentP)->dialog;
  else
    d_w = (AXWidget_dialog)parentP;

  best_w = NULL;
    
  if (oldP) {
    for ( w = d_w->widgets; w; w = w->nextChild ) {
      if (xmCanTakeKeyFocus(d_w, w, FALSE)) {
                
    /* make radio toggles look like the radio label's geometry */
	if (w->type == W_TOGGLE &&
		((AXWidget_togglebutton)w)->in_radio != NULL) {
		w = (AXWidget)((AXWidget_togglebutton)w)->in_radio;
	}
                
    if ( (w == oldP) || (w->xpos > oldP->xpos) )
      continue;  /* reject old or those right of old */
                
    if ( w->xpos == oldP->xpos ) {
      if ( w->ypos >= oldP->ypos ) {
        continue;  /* reject those below old */
      }
      else if ( best_w ) {
        if ( w->ypos > best_w->ypos ) {
          /* prefer bottommost of those above
           */
          best_w = w;
        }
      }
    }
                
    if ( !best_w ) {
      best_w = w;    /* first of those to left of old */
    }
    if ( w->xpos > best_w->xpos ) {
      best_w = w;    /* prefer rightmost of those to left */
    }
    else if ( (w->xpos == best_w->xpos) &&
          (w->ypos >= best_w->ypos) ) {
      /* prefer bottommost of those at same x
       */
      best_w = w;
    }
      }
    }
  }
  if ( best_w || (best_w = axBottomRightmostWidget(d_w))) {
    xmTakeKeyFocus(best_w, NULL, (W_ENTER == best_w->type));
#if AXMS
	/* if we have a radio array, make sure we set the focus on the
	 * currently selected toggle button.
	 */
	if ( best_w->type == W_RADIO )
		best_w = xmRadioActiveToggle(best_w);
#endif
    return(best_w);
  }
  else {
    return (NULL);
  }
}

AXWidget 
TabToNextWidgetHoriz (oldP,parentP)
     AXWidget oldP;
     AXWidget_any parentP;   /* oldP->parent (dialog if oldP is null) */
{
  register AXWidget best_w, w;
  register AXWidget_dialog d_w;

  if ( !parentP )
    return(NULL);
  if ( parentP->type == W_MAIN )
    d_w = ((AXWidget_main)parentP)->dialog;
  else
    d_w = (AXWidget_dialog)parentP;

  if ( !oldP )
    goto do_top;

  /* look ahead */
  best_w = NULL;

  best_w = findInKeyFocusLinks(d_w, oldP, TRUE);
  if (best_w) {
	  if (xmCanTakeKeyFocus(d_w, best_w, FALSE))
		  goto ok;
	  else {
		  /* Returned widget is inactive -- make it the oldP and
			 continue on to find the next best widget */
		  oldP = best_w;
		  best_w = NULL;
	  }
  }

  for ( w = d_w->widgets; w; w = w->nextChild ) {
    if (xmCanTakeKeyFocus(d_w, w, FALSE)) {

	/* make radio toggles look like the radio label's geometry */
	if (w->type == W_TOGGLE &&
		((AXWidget_togglebutton)w)->in_radio != NULL) {
		w = (AXWidget)((AXWidget_togglebutton)w)->in_radio;
	}
                
      if ( (w == oldP) || (w->ypos < oldP->ypos) )
    continue;
      if ( w->ypos == oldP->ypos ) {
    if ( w->xpos <= oldP->xpos )
      continue;
    else if ( best_w ) {
      if ( w->xpos < best_w->xpos )
        best_w = w;
    }
      }
      if ( !best_w )
    best_w = w;
      if ( w->ypos < best_w->ypos )
    best_w = w;
      else if ( (w->ypos == best_w->ypos) && (w->xpos <= best_w->xpos) )
    best_w = w;
    }
  }
  if ( best_w ) {
  ok:
    xmTakeKeyFocus(best_w, NULL, (W_ENTER == best_w->type));
#if AXMS
	/* if we have a radio array, make sure we set the focus on the
	 * currently selected toggle button.
	 */
	if ( best_w->type == W_RADIO )
		best_w = xmRadioActiveToggle(best_w);
#endif
    return(best_w);
  }

  /* select top widget */

do_top:
  best_w = NULL;
  for ( w = d_w->widgets; w; w = w->nextChild ) {
    if (xmCanTakeKeyFocus(d_w, w, FALSE)) {
      if ( !best_w )
    best_w = w;
      if ( w->ypos < best_w->ypos )
    best_w = w;
      else if ( (w->ypos == best_w->ypos) && (w->xpos <= best_w->xpos) )
    best_w = w;
    }
  }
  if ( best_w )
    goto ok;
  return(NULL);
}

AXWidget 
TabToPrevWidgetHoriz (oldP, parentP)
     AXWidget oldP;
     AXWidget parentP;   /* oldP->parent (dialog if oldP is null) */
{
  register AXWidget best_w, w;
  register AXWidget_dialog d_w;

  if ( !parentP )
    return(NULL);
  if ( parentP->type == W_MAIN )
    d_w = ((AXWidget_main)parentP)->dialog;
  else
    d_w = (AXWidget_dialog)parentP;

  if ( !oldP )
    goto do_bottom;

  /* look behind */

  best_w = NULL;

  best_w = findInKeyFocusLinks(d_w, oldP, FALSE);
  if (best_w) {
	  if (xmCanTakeKeyFocus(d_w, best_w, FALSE))
		  goto ok;
	  else {
		  /* Returned widget is inactive -- make it the oldP and
			 continue on to find the next best widget */
		  oldP = best_w;
		  best_w = NULL;
	  }
  }

  for ( w = d_w->widgets; w; w = w->nextChild ) {
    if (xmCanTakeKeyFocus(d_w, w, FALSE)) {
                
    /* make radio toggles look like the radio label's geometry */
	if (w->type == W_TOGGLE &&
		((AXWidget_togglebutton)w)->in_radio != NULL) {
		w = (AXWidget)((AXWidget_togglebutton)w)->in_radio;
	}
		
      if ( (w == oldP) || (w->ypos > oldP->ypos) )
    continue;
      if ( w->ypos == oldP->ypos ) {
    if ( w->xpos >= oldP->xpos )
      continue;
    else if ( best_w ) {
      if ( w->xpos > best_w->xpos )
        best_w = w;
    }
      }
      if ( !best_w )
    best_w = w;
      if ( w->ypos > best_w->ypos )
    best_w = w;
      else if ( (w->ypos == best_w->ypos) && (w->xpos >= best_w->xpos) )
    best_w = w;
    }
  }
  if ( best_w ) {
  ok:
    xmTakeKeyFocus(best_w, NULL, (W_ENTER == best_w->type));
#if AXMS
	/* if we have a radio array, make sure we set the focus on the
	 * currently selected toggle button.
	 */
	if ( best_w->type == W_RADIO )
		best_w = xmRadioActiveToggle(best_w);
#endif
    return(best_w);
  }

  /* select bottom widget */

do_bottom:
  best_w = NULL;
  for ( w = d_w->widgets; w; w = w->nextChild ) {
    if (xmCanTakeKeyFocus(d_w, w, FALSE)) {
      if ( !best_w )
    best_w = w;
      if ( w->ypos > best_w->ypos )
    best_w = w;
      else if ( (w->ypos == best_w->ypos) && (w->xpos >= best_w->xpos) )
    best_w = w;
    }
  }
  if ( best_w )
    goto ok;
  return(NULL);
}


static AXWidget
axTopLeftmostWidget(parentP)
     AXWidget_dialog parentP;
{
  AXWidget best_w, w;
  register AXWidget_dialog d_w;

  if ( parentP->type == W_MAIN )
    d_w = ((AXWidget_main)parentP)->dialog;
  else
    d_w = (AXWidget_dialog)parentP;

  best_w = NULL;
  for ( w = parentP->widgets; w; w = w->nextChild ) {
    if (xmCanTakeKeyFocus(d_w, w, FALSE)) {
      if ( !best_w )
    best_w = w;
      if ( w->xpos < best_w->xpos )
    best_w = w;
      else if ( (w->xpos == best_w->xpos) && (w->ypos <= best_w->ypos) )
    best_w = w;
    }
  }
    
  return (best_w);
}

static AXWidget
axBottomRightmostWidget(parentP)
     AXWidget_dialog parentP;
{
  AXWidget best_w, w;
  register AXWidget_dialog d_w;
    
  if ( parentP->type == W_MAIN )
    d_w = ((AXWidget_main)parentP)->dialog;
  else
    d_w = (AXWidget_dialog)parentP;

  best_w = NULL;
  for ( w = parentP->widgets; w; w = w->nextChild ) {
    if (xmCanTakeKeyFocus(d_w, w, FALSE)) {
      if ( !best_w )
    best_w = w;
      if ( w->xpos > best_w->xpos )
    best_w = w;
      else if ( (w->xpos == best_w->xpos) && (w->ypos >= best_w->ypos) )
    best_w = w;
    }
  }
  return (best_w);
}

static int
AxWidgetHasAccelerator(ww, ksym, tab_ix)
AXWidget ww;
KeySym ksym;
int *tab_ix;  /* out, meaningful if ww is a tabcontrol */
{
	AXWidget_rect rw;
	
	if (0 == ksym) {
		return FALSE;
	}
	
	*tab_ix = 0;
	
	switch (ww->type) {
	case W_RECT:
		rw = (AXWidget_rect) ww;
		if (RECT_LAYERED == rw->style) {
			if (rw->tabctl.num_layers > 0) {
				int rx;
				for (rx = 0; rx < rw->tabctl.num_layers; rx++) {
					if (Atoupper((unsigned int) ksym) ==
						rw->tabctl.layerlist[rx].accelerator_key) {
						
						*tab_ix = rx;
						return TRUE;
					}
				}
			}
		}
		break;
	default:
		return Atoupper((unsigned int) ksym) == ww->accelerator_key;
	}
	return FALSE;
}


/* Find the first widget after the currently-selected widget that has
	the desired accelerator
*/
int
AxDoDialogAccelerator(dw, ksym)
AXWidget_dialog dw;
KeySym ksym;
{
	AXWidget curr, start, end;
	int ret;

	/* Do we know what widget is currently selected? */
	curr = xmKeyFocusWidget(dw);

	/* If so... */
	if (curr) {

		/* First look after current, thru the end of the dbox */
		start = curr->nextChild;
		end = NULL;
		ret = AxDoDialogAcceleratorLoop(dw, ksym, start, end);
		if (ret)
			return( ret);

		/* Then look from the start of the box to the current widget */
		start = dw->widgets;
		end = curr;
		ret = AxDoDialogAcceleratorLoop(dw, ksym, start, end);
		return( ret);
	}

	/* Else just search entire dialog box from top to bottom */
	else {
		start = dw->widgets;
		end = NULL;
		ret = AxDoDialogAcceleratorLoop(dw, ksym, start, end);
		return( ret);
	}
}

/* Look from 'start' to 'end' for the widget with the desired accelerator.
	Select it and return TRUE if found, else return FALSE.
	If 'end' is NULL, look until we run out of widgets.
*/
int
AxDoDialogAcceleratorLoop(dw, ksym, start, end)
AXWidget_dialog dw;
KeySym ksym;
AXWidget start, end;
{
	AXWidget ww, tab_ix;
	
	for (ww = start; ww; ww = ww->nextChild) {
		if (xmCanTakeKeyFocus(dw, ww, TRUE) &&
			AxWidgetHasAccelerator(ww, ksym, (int *)&tab_ix)) {

			if (!(ww->flags & WFLG_IN_EDIT)) {
				/* several widget types should be "clicked" besides
				 * just getting the keyfocus
				 */
				switch (ww->type) {
				case W_RECT:
					if (RECT_LAYERED == ww->style) {  /* should be so */
						xmtcSwitchLayer((AXWidget_rect) ww, (long)tab_ix);
					}
					break;
				case W_TOGGLE:
					xmToggleAction((AXWidget_togglebutton)ww);
					break;
				case W_BUTTON:
					if (ww->callback) {
						XmAnyCallbackStruct data;
						data.reason = EV_CHANGED_;
						AxCallback(ww->callback,ww, ww->callbacktag,
								   (char *) &data);
					}
					break;
				}
			}
			
			xmTakeKeyFocus(ww, (XButtonEvent *) NULL, FALSE);
			return (TRUE);
		}

		/* If caller asked us to stop looking at this widget, then stop */
		if (ww == end)
			return( FALSE);
	}
	return (FALSE);
}

int
xmToggleThreeStatePress(wt)
     AXWidget_togglebutton wt;
{
  if (wt->value == 2)
    wt->value = FALSE;
  else if (wt->value == FALSE)
    wt->value = TRUE;
  else if (wt->value = TRUE)
    wt->value = 2;
  return (wt->value);
}

xmToggleAction(wt)
AXWidget_togglebutton wt;
{
	XmToggleButtonCallbackStruct tog_data;
	XmRowColumnCallbackStruct rad_data;
	
	if (wt->in_radio) {
		AXWidget ww;
		if (!SetRadioValue(wt, TRUE)) {
			return FALSE;
		}
		ww = (AXWidget) wt->in_radio;
		rad_data.data = (char*) ww->value;
		rad_data.widget = (Widget) ww;
		if (ww->callback)
			AxCallback(ww->callback, ww, ww->callbacktag, (caddr_t)&rad_data);
	}
	else if (wt->like_radio) {
		if (wt->value != TRUE) {  /* if already set, no-op */
			tog_data.set = wt->value = TRUE;
			PaintToggle(wt, NULL, TRUE, TRUE);
			if (wt->callback) {
				AxCallback(wt->callback,(AXWidget)wt, wt->callbacktag, (caddr_t)&tog_data);
			}
		}
	}
	else {
		if (wt->scroll_type == THREE_STATE_TOGGLE) {
			wt->value = xmToggleThreeStatePress(wt);
			tog_data.set = wt->value;
		}
		else {
			tog_data.set = wt->value = (wt->value != TRUE);
		}
		PaintToggle(wt, NULL, TRUE, TRUE);
		if (wt->callback)
			AxCallback(wt->callback, (AXWidget)wt, wt->callbacktag,(caddr_t)&tog_data);
	}
	return TRUE;
}

