/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: wincolor.c
   Applix ASC version: 45
   Date: Mon Nov 16 13:53:43 1998
*/
#include <stdio.h>
#include <math.h>
#include <basicwin.h>
#include <errno.h>
#include <aos.h>
#include <tmi.h>    /* just for defines */
#include <atm.h>    /* just for defines */
#include <atmattr.h>    /* just for defines */
#include <getstr5.h>
#include <axwinclr.h>

#define N_ALIS_PIX 16
#define N_AX_PIX 242


HPALETTE   AxWinPaletteHandle;
LOGPALETTE *AxWinPalette;


/* profile color control flags */
extern int AxLoadApplicationColors; /* load app colors into default colormap */
extern int AxLoadWidgetColors;      /* load widget colors into def colormap */
extern short NoColor;      /* indicates mono apps - forced if set by profile */
int AxNoWidgetColor;       /* indicates mono widgets - forced if profiled */
extern int AxNColorLevels; /* the height of the color cube */
extern int AxNColors;      /* total number of colors used by apps */
extern short ColorDisplay; /* color available for apps and being used */
extern int UseWidgetColors; /* color available for widgets and being used */

extern HDC RootWin;
extern Running_Windows95;
extern PALETTEENTRY Ax64Defs[];
extern short DpyWid, DpyHt;
short ToggleFG;

HDC Dpy;        /* handle to DEVICE CONTEXT */
unsigned long FGpixel, BGpixel;
PIXEL_VALUE ButtonFacePixel;
PIXEL_VALUE BackPixelRGB, RecessPixelRGB;
PIXEL_VALUE ForeWorkPixelRGB, BackWorkPixelRGB;
PIXEL_VALUE InactiveTextPixel;

/* pixel indices allocated by applixware */
extern unsigned long ForeWorkPixel, BackWorkPixel;
extern unsigned long SunWorkPixel, ShadowWorkPixel, RecessWorkPixel;
extern unsigned long ForePixel, BackPixel;
extern unsigned long SunPixel, ShadowPixel, RecessPixel;

unsigned long AlisPixels[N_ALIS_PIX];
unsigned long AxPixels[PALETTESIZE];
unsigned long AxColors[256];
unsigned long WidgetPixels[10];


/* table of old ALIS colors */
int PelR[] = {
        0xffff, /* white */         0x0000, /* black */
        0xffff, /* ~ desktop */     0xa000, /* desktop color (Tina) */
        0xffff, /* red */           0x0000, /* cyan */
        0x0000, /* green */         0x8000, /* purple (Tina) */
        0x0000, /* blue (Tina) */   0xffff, /* yellow */
        0x6000, /* dark gray */     0x3000, /* olive */
        0xa000, /* light gray */    0xa000, /* dark red */
        0xffff, /* orange */        0x0000 /* dark blue (Tina) */
};
int PelG[] = {
        0xffff, /* white */         0x0000, /* black */
        0xffff, /* ~ desktop */     0xc000, /* desktop color */
        0x0000, /* red */           0xffff, /* cyan */
        0xc000, /* green */         0x0000, /* purple (Tina) */
        0x6000, /* blue (Tina) */   0xffff, /* yellow */
        0x6000, /* dark gray */     0x8000, /* olive */
        0xa000, /* light gray */    0x0000, /* dark red */
        0x8000, /* orange */        0x0000 /* dark blue (Tina) */
};
int PelB[] = {
        0xffff, /* white */         0x0000, /* black */
        0xffff, /* ~ desktop */     0xffff, /* desktop color */
        0x3000, /* red */           0xffff, /* cyan */
        0x3000, /* green */         0xffff, /* purple (Tina)  */
        0xffff, /* blue (Tina) */   0x0000, /* yellow */
        0x6000, /* dark gray */     0x2000, /* olive */
        0xa000, /* light gray */    0x0000, /* dark red */
        0x0000, /* orange */        0xa000 /* dark blue (Tina) */
};

#include <wincolor.p>



void
allocatePixels(void)
{
    int i;

    /* init the pixel arrays */
    for ( i=0; i < PALETTESIZE; i++ )
        AxPixels[i] = COLOR_BLACK;
    if ( DpyDepth == 8 ) 
        AxPixels[215] = 255;
    else
        AxPixels[215] = COLOR_WHITE;

    WidgetPixels[0] = COLOR_BLACK;
    for ( i=1; i<10; i++ )
        WidgetPixels[i] = COLOR_WHITE;

    AlisPixels[0] = COLOR_WHITE;    /* WhitePixel() */
    for ( i=1; i<N_ALIS_PIX; i++ )
        AlisPixels[i] = COLOR_BLACK;

    /* set the global application color indicators */
    UseWidgetColors = (DpyDepth != 1 && AxNoWidgetColor == FALSE);
    ColorDisplay = ( DpyDepth != 1 && NoColor == FALSE );
}


void
initAlisPixels(void)
{
    return;
}



/*************************************************************************
function: WMsetPixels - set system pixel values by id
*/
WMsetPixels(int fg, int bg)
{
    int black, white;

    if (fg == TMA_DEF)
        fg = TMA_BLACK;

    if (bg == TMA_DEF)
        bg = TMA_WHITE;

    if (!ColorDisplay)
    {
        /* account for extended color requests */
        bg = bg>15 ? (bg==17 ? TMA_WHITE : TMA_BLACK) : bg;
        fg = fg>15 ? (fg==17 ? TMA_WHITE : TMA_BLACK) : fg;

        /* set the pixel index */
        white = COLOR_WHITE;
        black = COLOR_BLACK;
        if ( ToggleFG )
        {
            BGpixel = ( bg == TMA_WHITE ) ? black : white;
            FGpixel = ( fg == TMA_WHITE ) ? black : white;
        }
        else
        {
            BGpixel = ( bg == TMA_WHITE ) ? white : black;
            FGpixel = ( fg == TMA_WHITE ) ? white : black;
        }
    }
    else
    {
        /* the ALIS colors end with color 15 */
        FGpixel = (fg>15) ? AxPixels[fg-N_ALIS_PIX] : AlisPixels[fg];
        BGpixel = (bg>15) ? AxPixels[bg-N_ALIS_PIX] : AlisPixels[bg];
    }

    return 0; /* this func should be void, but an obsolete prototype
               * file is being included declaring it as int, so this is
               * a temp compiler-gag.
               */
}

void
WMsetColors(int fg, int bg)
{
    WMsetPixels(fg, bg);
}



/*** move these to colorcom.c  ***/
get_2_level_color(val, max) /* 0, 100 */
int *val, max;
{
    int temp_color;

    temp_color = 100 * *val / max;
    if      ( temp_color < 51 )
        *val = 0;
    else
        *val = 1;
}
get_3_level_color(val, max) /* 0, 50, 100 */
int *val, max;
{
    int temp_color;

    temp_color = 100 * *val / max;
    if      ( temp_color < 40 )
        *val = 0;
    else if ( temp_color < 82 )
        *val = 1;
    else
        *val = 2;
}
get_4_level_color(val, max) /* 0, 50, 75, 100 */
int *val, max;
{
    int temp_color;

    temp_color = 100 * *val / max;
    if      ( temp_color < 40 )
        *val = 0;
    else if ( temp_color < 61 )
        *val = 1;
    else if ( temp_color < 82 )
        *val = 2;
    else
        *val = 3;
}
get_old_5_level_color(val, max) /* 0, 25, 50, 75, 100 */
int *val, max;
{
    int temp_color;

    temp_color = 100 * *val / max;
    if      ( temp_color < 13 )
        *val = 0;
    else if ( temp_color < 40 )
        *val = 1;
    else if ( temp_color < 61 )
        *val = 2;
    else if ( temp_color < 82 )
        *val = 3;
    else
        *val = 4;
}
get_5_level_color(val, max) /* 0, 50, 75, 87, 100 */
int *val, max;
{
    int temp_color;

    temp_color = 100 * *val / max;
    if      ( temp_color < 40 )
        *val = 0;
    else if ( temp_color < 61 )
        *val = 1;
    else if ( temp_color < 82 )
        *val = 2;
    else if ( temp_color < 92 )
        *val = 3;
    else
        *val = 4;
}
get_6_level_color(val, max) /* 0, 50, 75, 87, 95, 100 */
int *val, max;
{
    int temp_color;

    temp_color = 100 * *val / max;
    if      ( temp_color < 40 )
        *val = 0;
    else if ( temp_color < 61 )
        *val = 1;
    else if ( temp_color < 82 )
        *val = 2;
    else if ( temp_color < 92 )
        *val = 3;
    else if ( temp_color < 99 )
        *val = 4;
    else
        *val = 5;
}
/*******************************************/




/*** move this to common file?? */
/***************************************************************************
function: create_sys_cmap - create a 256 color colormap (for bitmap editor)
*/
void create_sys_cmap(pid, cmap)
int pid;
cmap_type *cmap;
{
    extern int Ax64Values[][6];
    int r, g, b, ix, pix_ix, c, m, y, k;


    /* create the colormap */
    cmap->n_colors = 256;
    cmap->color = (color_type *)TaskAlloc(pid, 256 * sizeof(color_type));

    /* load the transparent color */
    cmap->color[0].see_thru = TRUE;
    cmap->color[0].ink_type = 0;
    cmap->color[0].c = 0;
    cmap->color[0].m = 0;
    cmap->color[0].y = 0;
    cmap->color[0].k = 0;
    cmap->color[0].name = (char *)TaskAlloc(pid, 12);
    strcpy(cmap->color[0].name, NXLT("Transparent"));

    /* now load black */
    cmap->color[1].see_thru = FALSE;
    cmap->color[1].ink_type = 0;
    cmap->color[1].c = 0;
    cmap->color[1].m = 0;
    cmap->color[1].y = 0;
    cmap->color[1].k = 255;
    cmap->color[1].name = (char *)TaskAlloc(pid, 6);
    strcpy(cmap->color[1].name, NXLT("Black"));
    axGetColormapIndex(2, cmap->color[1].c,
                          cmap->color[1].m,
                          cmap->color[1].y,
                          cmap->color[1].k,
                          255,
                          &pix_ix);

    /* now load the rest of the grey values */
    for ( ix=2, r=0; r<AxNColorLevels; r++ )
    {
        cmap->color[ix].see_thru = FALSE;
        cmap->color[ix].ink_type = 0;
        cmap->color[ix].c = 0;
        cmap->color[ix].m = 0;
        cmap->color[ix].y = 0;
        cmap->color[ix].k = 255 - (double)Ax64Values[0][r]*255.0/31.0;
        cmap->color[ix].name = (char *)TaskAlloc(pid, 9);
        sprintf(cmap->color[ix].name, "color%03d", ix);
        axGetColormapIndex(2, cmap->color[ix].c,
                              cmap->color[ix].m,
                              cmap->color[ix].y,
                              cmap->color[ix].k,
                              255,
                              &pix_ix);
        ix++;
    }
    
    /* load the colormap entries */
    for ( r=0; r<AxNColorLevels; r++ )
    {
        for ( g=0; g<AxNColorLevels; g++ )
        {
            for ( b=0; b<AxNColorLevels; b++ )
            {
                /* don't redo the grey values */
                if ( r == g && g == b )
                    continue;

                /* set the color as a non-tranparent process color */
                cmap->color[ix].see_thru = FALSE;
                cmap->color[ix].ink_type = 0;

                /* calc the cmy values */
                c = 255 - (double)Ax64Values[0][r]*255.0/31.0;
                m = 255 - (double)Ax64Values[1][g]*255.0/31.0;
                y = 255 - (double)Ax64Values[2][b]*255.0/31.0;

                /* calc the black value */
                if ( c == 0 || m == 0 || y == 0 )
                    k = 0;
                else
                {
                    if ( c < m )
                        k = ( c < y ) ? c : y;
                    else
                        k = ( m < y ) ? m : y;
                }

                /* set the cmyk values */
                cmap->color[ix].c = c - k;
                cmap->color[ix].m = m - k;
                cmap->color[ix].y = y - k;
                cmap->color[ix].k = k;

                /* name the color */
                cmap->color[ix].name = (char *)TaskAlloc(pid, 9);
                sprintf(cmap->color[ix].name, "color%03d", ix);

                /* create map from sys pix val to sys independent val */
                axGetColormapIndex(2, cmap->color[ix].c,
                                      cmap->color[ix].m,
                                      cmap->color[ix].y,
                                      cmap->color[ix].k,
                                      255,
                                      &pix_ix);

                /* advance to the next color */
                ix++;
            }
        }
    }

    /* load the rest of the table with "white" */
    for ( ; ix<256; ix++ )
    {
        cmap->color[ix].see_thru = FALSE;
        cmap->color[ix].ink_type = 0;
        cmap->color[ix].c = 0;
        cmap->color[ix].m = 0;
        cmap->color[ix].y = 0;
        cmap->color[ix].k = 0;
        cmap->color[ix].name = (char *)TaskAlloc(pid, 9);
        sprintf(cmap->color[ix].name, "color%03d", ix);
        axGetColormapIndex(2, cmap->color[ix].c,
                              cmap->color[ix].m,
                              cmap->color[ix].y,
                              cmap->color[ix].k,
                              255,
                              &pix_ix);
    }
}



/****************************************/
void 
GetColorIndex(int r, int g, int b, UINT *cindex)
{
    unsigned int v1,v2,v3,
                 max_v = 255;
    
    v1 = r;
    v2 = g;
    v3 = b;

    switch (AxNColorLevels)
        {
        case 1:  /* black or white */
            if (v1 == max_v && v2 == max_v && v3 == max_v)
                *cindex = 1;
            else
                *cindex = 0;
            break;
        case 2:  /* 8 colors */
            get_2_level_color(&v1, max_v);
            get_2_level_color(&v2, max_v);
            get_2_level_color(&v3, max_v);
            *cindex = v1 * 4 + v2 * 2 + v3;
            break;
        case 3:  /* 27 colors */
            get_3_level_color(&v1, max_v);
            get_3_level_color(&v2, max_v);
            get_3_level_color(&v3, max_v);
            *cindex = v1 * 9 + v2 * 3 + v3;
            break;
        case 4:  /* 64 colors */
            get_4_level_color(&v1, max_v);
            get_4_level_color(&v2, max_v);
            get_4_level_color(&v3, max_v);
            *cindex = v1 * 16 + v2 * 4 + v3;
            break;
        case 5:  /* 125 colors */
            get_5_level_color(&v1, max_v);
            get_5_level_color(&v2, max_v);
            get_5_level_color(&v3, max_v);
            *cindex = v1 * 25 + v2 * 5 + v3;
            break;
        case 6:  /* 216 colors */
        default:
            get_6_level_color(&v1, max_v);
            get_6_level_color(&v2, max_v);
            get_6_level_color(&v3, max_v);
            *cindex = v1 * 36 + v2 * 6 + v3;
            break;
        }
}



static unsigned long
GetAxWinPaletteIndex(COLORREF color)
{
    unsigned long pixel;

    pixel = MS_GetNearestPaletteIndex(AxWinPaletteHandle, color);

#if 0
    /* if index is out of range, set to white. this is very possible, *
     * since the empty slots in the global palette (which will only   *
     * have up to 216 colors) are white and the MS function seems to  *
     * return the last slot it finds.              -donato            */
    if ( pixel >= (unsigned)(AxNColors + 10) )
    if ( pixel == 255 )
        pixel = (AxNColors + 10) - 1;    /* white pixel */
#endif
    return pixel;
}


COLORREF
Get16BitColor(COLORREF in_rgb)
{
    int r, g, b;
    COLORREF out_rgb;

    normalize_rgb_to_555(MS_GetRValue(in_rgb),
                        MS_GetGValue(in_rgb),
                        MS_GetBValue(in_rgb),
                        &r, &g, &b);
    out_rgb = (r & RED_MASK16) + (g & GREEN_MASK16) + (b & BLUE_MASK16);
    return out_rgb;
}


void
SetupUserColors(void)
{
  if ( DpyDepth > 8 )
    {
    ForeWorkPixel   = MS_GetSysColor(COLOR_WINDOWTEXT);
    BackWorkPixel   = MS_GetSysColor(COLOR_WINDOW);
    SunWorkPixel    = MS_GetSysColor(COLOR_BTNHIGHLIGHT);
    ShadowWorkPixel = MS_GetSysColor(COLOR_BTNSHADOW);
    RecessWorkPixel = MS_GetSysColor(COLOR_SCROLLBAR);

    BackPixel   = MS_GetSysColor(COLOR_BTNFACE);
    ForePixel   = MS_GetSysColor(COLOR_BTNTEXT);
    SunPixel    = MS_GetSysColor(COLOR_BTNHIGHLIGHT);
    ShadowPixel = MS_GetSysColor(COLOR_BTNSHADOW);
    RecessPixel = MS_GetSysColor(COLOR_SCROLLBAR);
    ButtonFacePixel = MS_GetSysColor(COLOR_BTNFACE);

	InactiveTextPixel = MS_GetSysColor(COLOR_INACTIVECAPTIONTEXT);
    }
  else
    {
    ForeWorkPixel   = GetAxWinPaletteIndex(MS_GetSysColor(COLOR_WINDOWTEXT));
    BackWorkPixel   = GetAxWinPaletteIndex(MS_GetSysColor(COLOR_WINDOW));
    SunWorkPixel    = GetAxWinPaletteIndex(MS_GetSysColor(COLOR_BTNHIGHLIGHT));
    ShadowWorkPixel = GetAxWinPaletteIndex(MS_GetSysColor(COLOR_BTNSHADOW));
    RecessWorkPixel = GetAxWinPaletteIndex(MS_GetSysColor(COLOR_SCROLLBAR));

    BackPixel   = GetAxWinPaletteIndex(MS_GetSysColor(COLOR_BTNFACE));
    ForePixel   = GetAxWinPaletteIndex(MS_GetSysColor(COLOR_BTNTEXT));
    SunPixel    = GetAxWinPaletteIndex(MS_GetSysColor(COLOR_BTNHIGHLIGHT));
    ShadowPixel = GetAxWinPaletteIndex(MS_GetSysColor(COLOR_BTNSHADOW));
    RecessPixel = GetAxWinPaletteIndex(MS_GetSysColor(COLOR_SCROLLBAR));
    ButtonFacePixel = GetAxWinPaletteIndex(MS_GetSysColor(COLOR_BTNFACE));

	InactiveTextPixel = GetAxWinPaletteIndex(
									MS_GetSysColor(COLOR_INACTIVECAPTIONTEXT));
    }

    /* HACK: if user background color is white then we set AlisPixels[0]   *
     * to 255, which is the Windows user color index for white. Otherwise  *
     * we set it to the AxWinPalette index. The reason for this is because *
     * MS_GetDIBits always seems to return 0xFF for white bits.            */
#if 0
    if ( Running_Windows95 && (BackWorkPixel == (AxNColors + 10) - 1) )
        {
        /* white */
        AlisPixels[0] = 0xFF;
        BackWorkPixel = 0xFF;
        }
    else
#endif
    AlisPixels[0] = BackWorkPixel;
    //AlisPixels[0] = RetrieveDIBWorkPixel();
    AlisPixels[1] = ForeWorkPixel;

    /* set up some global RGB values for cases where *
     * we'll use the Windows static colors.          */
    BackPixelRGB = MS_GetSysColor(COLOR_BTNFACE);
    RecessPixelRGB = MS_GetSysColor(COLOR_SCROLLBAR);
    ForeWorkPixelRGB = MS_GetSysColor(COLOR_WINDOWTEXT);
    BackWorkPixelRGB = MS_GetSysColor(COLOR_WINDOW);
}

void
AxWinSetPixels(void)
{
    int ix;
    UINT pixel_index;

    /* don't include the Windows static colors */
    for ( ix=10; ix < 246; ix++ )
    /*for ( ix=0; ix < PALETTESIZE; ix++ ) */
        {
        
        GetColorIndex(AxWinPalette->palPalEntry[ix].peRed,
                       AxWinPalette->palPalEntry[ix].peGreen,
                       AxWinPalette->palPalEntry[ix].peBlue,
                       &pixel_index);
        if ( DpyDepth > 8 )
            {
            AxPixels[pixel_index] =
                            RGB(AxWinPalette->palPalEntry[ix].peRed,
                                AxWinPalette->palPalEntry[ix].peGreen,
                                AxWinPalette->palPalEntry[ix].peBlue);
            }
        else
            AxPixels[pixel_index] = ix;
        }

    if ( DpyDepth > 8 )
        {
        /* we don't need the palette structure anymore, so delete it. */
        free(AxWinPalette);
        AxWinPalette = NULL;
        }
}

void
AxWinSetPalette(void)
{
    int i, j;

    AxWinPalette = (LOGPALETTE*)calloc(1, sizeof(LOGPALETTE) +
                            (sizeof (PALETTEENTRY) * (PALETTESIZE)));
    if ( !AxWinPalette )
        taskAbort(NXLT("AxSetWinPalette: Not enough memory for palette."));

    AxWinPalette->palVersion    = 0x300;
    AxWinPalette->palNumEntries = PALETTESIZE;

    /* fill in intensities for all palette entry colors */
    /* start the Windows palette at index 10, because any colors we *
     * put in 0-9 won't get returned properly in MS_GetDIBits.      */
    for ( i=0, j=10; i < (PALETTESIZE - 10); i++, j++)
        {
        AxWinPalette->palPalEntry[j].peRed   = Ax64Defs[i].peRed;
        AxWinPalette->palPalEntry[j].peBlue  = Ax64Defs[i].peBlue;
        AxWinPalette->palPalEntry[j].peGreen = Ax64Defs[i].peGreen;
        AxWinPalette->palPalEntry[j].peFlags = 0;
        }

    /* set up white in unused sections of the palette. */
    for ( i = AxNColors+10; i < PALETTESIZE; i++ )
        {
        AxWinPalette->palPalEntry[i].peRed   = 255;
        AxWinPalette->palPalEntry[i].peBlue  = 255;
        AxWinPalette->palPalEntry[i].peGreen = 255;
        AxWinPalette->palPalEntry[i].peFlags = 0;
        }

    SetupIdentityPalette();

#if 0
    /* set up background pixel in the last non-static palette entry. */
    back_rgb = MS_GetSysColor(COLOR_WINDOW);
    AxWinPalette->palPalEntry[245].peRed   = MS_GetRValue(back_rgb);
    AxWinPalette->palPalEntry[245].peBlue  = MS_GetBValue(back_rgb);
    AxWinPalette->palPalEntry[245].peGreen = MS_GetGValue(back_rgb);
    AxWinPalette->palPalEntry[245].peFlags = PC_NOCOLLAPSE;

    AlisPixels[0] = 245;
#endif

    /*  create a logical color palette according the information
     *  in the LOGPALETTE structure.
     */
    if ( DpyDepth == 8 )
        AxWinPaletteHandle = MS_CreatePalette(AxWinPalette);

    /* set up global palette indices. */
    SetupUserColors();
}


/****************************************************************
 * this function places the 20 Windows static colors into our
 * logical palette so that user-defined Windows colors can be
 * properly accessed and used.
 ****************************************************************/
void
SetupIdentityPalette(void)
{
    int i, j;
    HDC hdc;
    PALETTEENTRY temp_pal[10];


    hdc = MS_GetDC(NULL);
    /* get the first 10 static colors. */
    MS_GetSystemPaletteEntries(hdc, 0, 10, temp_pal);
    for ( i=0; i < 10; i++ )
        {
        AxWinPalette->palPalEntry[i].peRed   = temp_pal[i].peRed;
        AxWinPalette->palPalEntry[i].peBlue  = temp_pal[i].peBlue;
        AxWinPalette->palPalEntry[i].peGreen = temp_pal[i].peGreen;
        AxWinPalette->palPalEntry[i].peFlags = temp_pal[i].peFlags;
        }
    /* get the last 10 static colors. */
    MS_GetSystemPaletteEntries(hdc, 246, 10, temp_pal);
    for ( i=246, j=0; i < 255; i++, j++ )
        {
        AxWinPalette->palPalEntry[i].peRed   = temp_pal[j].peRed;
        AxWinPalette->palPalEntry[i].peBlue  = temp_pal[j].peBlue;
        AxWinPalette->palPalEntry[i].peGreen = temp_pal[j].peGreen;
        AxWinPalette->palPalEntry[i].peFlags = temp_pal[j].peFlags;
        }
    MS_ReleaseDC(NULL, hdc);
}


/**********************************************************************
 * Retrieves the color index of the specified color from the palette
 * of the specified image.
 ***/
void
axWinGetImageColormapIndex(int c,int m,int y,int k, XImage *ximg, int *cindex)
{
    BYTE red, green, blue;
    COLORREF color;

    red   = 255 - MIN(255, c + k);
    green = 255 - MIN(255, m + k);
    blue  = 255 - MIN(255, y + k);
    color = RGB(red, green, blue);
    *cindex = MS_GetNearestPaletteIndex(ximg->palette_handle, color);
}


PIXEL_VALUE
axmsGetPixelFromDIBImageData(unsigned char *img_data, int width, int depth,
                                                            int x, int y)
{
    int bytes_per_line, image_index;
    PIXEL_VALUE pixel;
    unsigned char *pixel_byteP;

    bytes_per_line = ((width * depth)+7)/8;
    bytes_per_line = ((bytes_per_line + 3)/4) * 4;
    pixel_byteP = img_data + (y * bytes_per_line) + ((x * depth) / 8);

    if ( depth == 8 )
        {
        image_index = ( (width * y) + x );
        pixel = img_data[image_index];
        }
    /* byte order ASSUMPTION for greater than 8 bit images: LSBFirst */
    else if ( depth == 16 )
        {
        pixel = *((unsigned short *)pixel_byteP);
        }
    else if ( depth == 24 )
        {
        pixel = *pixel_byteP;
        pixel_byteP++;
        pixel |= *pixel_byteP << 8;
        pixel_byteP++;
        pixel |= *pixel_byteP << 16;
        }
    else if ( depth == 32 )
        {
        pixel = *((unsigned int *)pixel_byteP);
        }
    return pixel;
}

PIXEL_VALUE
axmsRetrieveDIBValue(int color_value)
{
    int nColors;
    HDC hdc, mem_dc;
    HBRUSH hbrush, old_brush;
    HPALETTE old_palette;
    PIXEL_VALUE pixel;
    HBITMAP hbm, old_hbm;
    unsigned char *img_data;
    int scan_lines, img_size;
    BITMAPINFO  *bm_infoP;
    ULONG       sizeBMI;


    mem_dc = MS_CreateCompatibleDC(RootWin);
    hbm = MS_CreateCompatibleBitmap(RootWin, 1, 1);
    old_hbm = MS_SelectObject(mem_dc, hbm);
    old_palette = SelectAndRealizeAxWinPalette(mem_dc);

    hbrush = MS_CreateSolidBrush(GET_COLOR_VALUE(color_value));
    old_brush = MS_SelectObject(mem_dc, hbrush);
    MS_PatBlt(mem_dc, 0, 0, 1, 1, PATCOPY);

    MS_SelectObject(mem_dc, old_hbm);
    MS_SelectObject(mem_dc, old_brush);
    UnSelectPalette(mem_dc, old_palette);
    MS_DeleteDC(mem_dc);

    hdc = MS_GetDC(NULL);
    old_palette = SelectAndRealizeAxWinPalette(hdc);

    /* First call GetDIBits with a zero'ed structure (particularly     *
     * the bm_infoP->bmiHeader.biBitCount field) to get the image size */
    bm_infoP = (BITMAPINFO *)calloc( 1, sizeof(BITMAPINFO) );
    bm_infoP->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);

    MS_GetDIBits(hdc, hbm, (UINT)0, (UINT)0, (LPBYTE)NULL,
               (LPBITMAPINFO)bm_infoP, (DWORD)DIB_RGB_COLORS);

    /* Now that we know the size, allocate space for the image data */
    scan_lines = bm_infoP->bmiHeader.biHeight;
    img_size   = bm_infoP->bmiHeader.biSizeImage;
    img_data = (char*)malloc(img_size);

    /** Allocate space for the color table (if necessary) */
    /** from \mstools\samples\mandel\dibmp.c **/
    /*
     * Note: 24 bits per pixel has no color table.  So, we don't have to
     * allocate memory for retrieving that.  Otherwise, we do.
     */
    nColors = GetNumDIBColors(&bm_infoP->bmiHeader);
    if ( nColors )
        {
        /* has color table */
        sizeBMI = sizeof(BITMAPINFO) + sizeof(RGBQUAD) * nColors;
        /* I need more memory for the color table */
        bm_infoP = (BITMAPINFO*)realloc(bm_infoP, sizeBMI);
        }
    if ( bm_infoP->bmiHeader.biCompression == BI_BITFIELDS )
        {
        /* need more space for color mask values (3 double words). */
        sizeBMI = sizeof(BITMAPINFO) + (sizeof(DWORD) * 3);
        bm_infoP = (BITMAPINFO*)realloc( bm_infoP, sizeBMI );
        }

    /* get the bits and the color table */
    MS_GetDIBits(hdc, hbm, 0, scan_lines, (LPSTR)img_data,
                                bm_infoP, DIB_RGB_COLORS);
    pixel = axmsGetPixelFromDIBImageData(img_data, 1,
                            bm_infoP->bmiHeader.biBitCount, 0, 0);

    /* cleanup */
    UnSelectPalette(hdc, old_palette);
    MS_ReleaseDC(NULL, hdc);
    MS_DeleteObject(hbm);
    MS_DeleteObject(hbrush);
    free(bm_infoP);
    free(img_data);

    return pixel;
}





#if 0    /* { */
void GetPaletteColorFromIndex(pixel, pal_color)
unsigned long pixel;
COLORREF *pal_color;
{
    *pal_color = PALETTEINDEX(pixel);
}


void GetPaletteColorFromRGB(rgb_val, pal_color)
unsigned long rgb_val;
COLORREF *pal_color;
{
    BYTE red, green, blue;
    
    red   = MS_GetRValue(rgb_val);
    green = MS_GetGValue(rgb_val);
    blue  = MS_GetBValue(rgb_val);
    *pal_color = PALETTERGB(red, green, blue);
}

static PIXEL_VALUE
RetrieveDIBWorkPixel(void)
{
    int nColors;
    HDC hdc, mem_dc;
    HBRUSH hbrush, old_brush;
    HPALETTE old_palette;
    PIXEL_VALUE back_work_pixel;
    HBITMAP hbm, old_hbm;
    unsigned char *img_data;
    int scan_lines, img_size;
    BITMAPINFO  *bm_infoP;
    ULONG       sizeBMI;


    mem_dc = MS_CreateCompatibleDC(RootWin);
    hbm = MS_CreateCompatibleBitmap(RootWin, 2, 2);
    old_hbm = MS_SelectObject(mem_dc, hbm);
    old_palette = SelectAndRealizeAxWinPalette(mem_dc);

    hbrush = MS_CreateSolidBrush(GET_COLOR_VALUE(BackWorkPixel));
    old_brush = MS_SelectObject(mem_dc, hbrush);
    MS_PatBlt(mem_dc, 0, 0, 2, 2, PATCOPY);

    MS_SelectObject(mem_dc, old_hbm);
    MS_SelectObject(mem_dc, old_brush);
    UnSelectPalette(mem_dc, old_palette);
    MS_DeleteDC(mem_dc);


    hdc = MS_GetDC(NULL);
    old_palette = SelectAndRealizeAxWinPalette(hdc);

    /* First call GetDIBits with a zero'ed structure (particularly     *
     * the bm_infoP->bmiHeader.biBitCount field) to get the image size */
    bm_infoP = (BITMAPINFO *)calloc( 1, sizeof(BITMAPINFO) );
    bm_infoP->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);

    MS_GetDIBits(hdc, hbm, (UINT)0, (UINT)0, (LPBYTE)NULL,
               (LPBITMAPINFO)bm_infoP, (DWORD)DIB_RGB_COLORS);

    /* Now that we know the size, allocate space for the image data */
    scan_lines = bm_infoP->bmiHeader.biHeight;
    img_size   = bm_infoP->bmiHeader.biSizeImage;
    img_data = (char*)malloc(img_size);

    /** Allocate space for the color table (if necessary) */
    /** from \mstools\samples\mandel\dibmp.c **/
    /*
     * Note: 24 bits per pixel has no color table.  So, we don't have to
     * allocate memory for retrieving that.  Otherwise, we do.
     */
    nColors = GetNumDIBColors(&bm_infoP->bmiHeader);
    if ( nColors )
        {
        /* has color table */
        sizeBMI = sizeof(BITMAPINFO) + sizeof(RGBQUAD) * nColors;
        /* I need more memory for the color table */
        bm_infoP = realloc(bm_infoP, sizeBMI);
        }
    if ( bm_infoP->bmiHeader.biCompression == BI_BITFIELDS )
        {
        /* need more space for color mask values (3 double words). */
        sizeBMI = sizeof(BITMAPINFO) + (sizeof(DWORD) * 3);
        bm_infoP = realloc( bm_infoP, sizeBMI );
        }

    /* get the bits and the color table */
    MS_GetDIBits(hdc, hbm, 0, scan_lines, (LPSTR)img_data,
                                bm_infoP, DIB_RGB_COLORS);
    back_work_pixel = img_data[0];

    /* cleanup */
    UnSelectPalette(hdc, old_palette);
    MS_ReleaseDC(NULL, hdc);
    MS_DeleteObject(hbm);
    MS_DeleteObject(hbrush);
    free(bm_infoP);
    free(img_data);

    return back_work_pixel;
}
#endif    /* } #if 0 */
