/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: winprint.c
   Applix ASC version: 11
   Date: Mon Nov 16 13:53:46 1998
*/
#include <stdio.h>
#include <string.h>
#include <basicwin.h>
#include <winspool.h>
#include <commdlg.h>
#include <aos.h>
#include <eldata.h>
#include <axtext.h>
#include <winprint.p>


extern HANDLE hInst;
extern int Running_Windows95;



data_ptr
GetWinPrinters(void)
{
    data_ptr retP;
    int printer_index = 0;
#define MAX_PRINTERS 100
#define PRINTER_NAME_SIZE 100
    char printer_list[MAX_PRINTERS][PRINTER_NAME_SIZE];
    union {
        PRINTER_INFO_4 *printP;
        PRINTER_INFO_5 *win95_printP;
    } printer_info;
    unsigned char *buffP;
    DWORD size_needed, num_entries, level, initial_size;
    int i, j, ret, duplicate;
    char *nameP;
    int request;

    initial_size = 1024; /* just some "big enough" size for the printer info.
                          * we'll re-enum if necessary.
                          */
    request = (PRINTER_ENUM_FAVORITE | PRINTER_ENUM_LOCAL);

    if ( Running_Windows95 )
        level = 5;
    else
        level = 4;

    buffP = (unsigned char *)malloc(initial_size);
enum_printers:
    ret = MS32_EnumPrinters(request, (char*)NULL, level, buffP, initial_size,
                                                &size_needed, &num_entries);
    /* do we need more space? */
    if ( size_needed > initial_size )
        {
        initial_size = size_needed;
        buffP = (unsigned char *)realloc(buffP, initial_size);
        goto enum_printers;
        }
    else if ( !ret )
        /* some error, skip it */
        return( (data_ptr)NULL );

    if ( Running_Windows95 )
        printer_info.win95_printP = (PRINTER_INFO_5 *)buffP;
    else
        printer_info.printP = (PRINTER_INFO_4 *)buffP;

    for ( i=0; i < (int) num_entries; i++ )
       {
       duplicate = FALSE;
       nameP = printer_info.printP->pPrinterName;
       for ( j=0; j<printer_index; j++ )
          {
          /* make sure we do not duplicate an existing entry */
          duplicate = FALSE;
          if ( stricmp( printer_list[j], nameP ) == 0 )
              {
              /* a duplicate */
              duplicate = TRUE;
              break;
              }
          }
       if ( !duplicate )
           {
           strcpy( printer_list[printer_index], nameP );
           printer_index = printer_index + 1;
           if ( printer_index == MAX_PRINTERS )
              break;
           }
        if ( Running_Windows95 )
            printer_info.win95_printP++;
        else
            printer_info.printP++;
       }
    free( buffP );

    /* now build the return array */
    retP = MkArray(printer_index);
    for ( i=0; i < printer_index; i++ )
       {
       retP = AxAddStrToArray( retP, i, printer_list[ i ] );
       }

    return( retP );
}

void
GetWinPrinterAverageWidth(fontTabEnt *fentP)
{
    LOGFONT logfont;
    HFONT win_font;
    int delete_dc;
    HDC printer_hdc;
    TEXTMETRIC tm;

    
    /* I would like to use the average char width number which comes out */
    /* of the NEWTEXTMETRIC structure passed as part of the enumeration, but */
    /* experience has shown that this number is bogus.  With the          */
    /* LaserWriter-NTR driver, the average-char-width is exactly the      */
    /* same for all fonts (including Helvetica-Narrow).  THe waverage     */
    /* width number retrieved by actually crreating a font seems accurate, */
    /* so that is what we do here.                                         */
    /* create a logical font, in the size of the em-square.          */
    /*   This will give us the un-hinted sizes                       */
    memset( &logfont, 0, sizeof(logfont) );
    logfont.lfHeight = fentP->sizeEM;

    if ( fentP->weight )
       logfont.lfWeight = FW_BOLD;   /* bold */
    if ( fentP->slant )
       logfont.lfItalic = 1;     /* italic */

    logfont.lfCharSet = fentP->charSet;
    logfont.lfPitchAndFamily = fentP->pitchAndFamily;
    strcpy( logfont.lfFaceName, fentP->familyName );

#if 0
    /* printer fonts will have a " (p)" appended to the name */
    p = logfont.lfFaceName + strlen(logfont.lfFaceName) - 4;
    if ( strcmp( p, " (P)") == 0 )
       *p = '\0';     /* take off printer font designation */
#endif
 
    win_font = MS_CreateFontIndirect( &logfont );
    if ( win_font == (HFONT)NULL )
        {
        /* the best we can do */
        fentP->avgWidth = -fentP->avgWidth;  /* value from enumeration is */
        }                                    /* best we can do            */
    else
       {
       printer_hdc = AxGetPrinterHDC(NULL, &delete_dc);
       if ( !printer_hdc )
           {
           WXErrorMessage(XLT("Invalid printer requested."));
           return;
           }

       /* now select the font into the hDC so that we can get information */
       MS_SelectObject( printer_hdc, win_font );
       memset( &tm, 0, sizeof(tm) );
       MS_GetTextMetrics( printer_hdc, &tm );
       fentP->avgWidth = tm.tmAveCharWidth;
       if ( delete_dc )
          MS_DeleteDC(printer_hdc);
       }
}


HDC
AxGetPrinterHDC(char *printer_name, int *delete_dc)
{
    HDC printer_hdc;
    char *profile_printer;

    if ( printer_name && printer_name[0] )
        {
        printer_hdc = GetPrinterHDCfromName(printer_name);
        *delete_dc = TRUE;
        }
    else
        {
        /* if printer name is null, we look for the profile value first,
         * then the Windows system default as a last resort.
         */
        profile_printer = UPbyName(NXLT("axPrintPrinter"));
        if ( profile_printer )
            {
            printer_hdc = GetPrinterHDCfromName(profile_printer);
            if ( !printer_hdc )
                /* don't know what else to do if above failed! */
                printer_hdc = GetDefaultPrinterHDC(FALSE, NULL);
            *delete_dc = TRUE;
            }
        else
            {
            printer_hdc = GetDefaultPrinterHDC(FALSE, NULL);
            *delete_dc = TRUE;
            }
        }

    return printer_hdc;
}


HDC
GetDefaultPrinterHDC(int print_to_file, char **printer_name)
{
    HDC printer_hdc;
    PRINTDLG pdlg;
    DEVMODE *dmP;

    /* To get the HDC of the current printer, fill out a PRINTDLG
       structure, and set it's flags to (PD_RETURNDC | PD_RETURNDEFAULT).
       This will tell PrintDlg() to create you an HDC but not show
       the Print dialog box. */
    pdlg.lStructSize = sizeof(PRINTDLG);
    pdlg.hwndOwner = NULL;
    pdlg.hDevMode = (HANDLE)0;
    pdlg.hDevNames = (HANDLE)0;
    if ( print_to_file )
        /* just get default printer and toggle 'print to file' on */
        pdlg.Flags = PD_RETURNDC | PD_RETURNDEFAULT | PD_PRINTTOFILE;
    else
        /* just get default printer */
        pdlg.Flags = PD_RETURNDC | PD_RETURNDEFAULT;
    pdlg.nFromPage = 0;
    pdlg.nToPage = 0;
    pdlg.nMinPage = 0;
    pdlg.nMaxPage = 0;
    pdlg.nCopies = 0;
    pdlg.hInstance = (HANDLE)hInst;

    if ( MS_PrintDlg(&pdlg) )
        {
        printer_hdc = pdlg.hDC;
#if DEBUG_RESOURCES
        remember_HDC(printer_hdc);
#endif
        }
    else
        printer_hdc = NULL;

    /* if we are printing to file, return the default printer name,
     * found in the DEVMODE struct.
     */
    if ( pdlg.hDevMode && print_to_file )
        {
        dmP = (LPDEVMODE) MS_GlobalLock(pdlg.hDevMode);
        *printer_name = malloc(strlen(dmP->dmDeviceName)+1);
        strcpy(*printer_name, dmP->dmDeviceName);
        MS_GlobalUnlock(pdlg.hDevMode);
        }

    return printer_hdc;
}


HDC
GetPrinterHDCfromName(char *name)
{
    HDC hdc;

    /* can we assume "winspool" for the printer?  -donato */
    /* Ideally, we should be getting the HDC from the Windows Print Dialog */
    hdc = MS_CreateDC("WINSPOOL", name, NULL, NULL);
    return hdc;
}


