/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: wintoolb.c
   Applix ASC version: 33
   Date: Mon Nov 16 13:53:42 1998
*/
#include <windows.h>
#include <stdlib.h>
#include <string.h>
#include <aos.h>
#include <xmlib.h>
#include <xmpixel.h>

extern AXWidget_popup PopupArmed;
extern AXWidget_rowcol RowColumnArmed;
extern int THIMhelpMode;
extern Cursor ThimPointerCursor;


struct XMlibStr {
    PIXEL_VALUE FGpixel;    /* foreground color for app work areas */
    PIXEL_VALUE BGpixel;    /* backround color for app work areas */
    int FullRegenForUpdate;  /* Set by UnMapWidget,
                                            hint for WidgetNeedsUpdated */
};
extern struct XMlibStr xmlibStr; 


#include <wintoolb.p>




/*********************************************************
function: RealizeToolBar - create toolbar window
*/
void
RealizeToolBar(AXWidget_tool toolbar)
{
    int release_dc = FALSE;
    int tool_width, tool_height;
    RECT rect;

    WIDGET_PARANOID(toolbar, W_TOOLBAR);
	if ( !toolbar->display )
		return;

    if ( toolbar->width == 0 )
        {
        MS_GetClientRect(toolbar->parent->winid, &rect);
        tool_width  = rect.right;
        tool_height = AxToolBarHeight(toolbar->parent, tool_width);
        }

    if ( toolbar->height == 0 )
        {
        /* toolbar not visible, so don't create it */
        return;
        }

    if ( toolbar->winid )
        {
        MS_DestroyWindow(toolbar->winid);
        toolbar->winid  = (HWND)NULL;
        /* toolbar has been destroyed, so remove the button uids *
         * from the global list, so that they may be used again. */
        ClearToolBarButtonUids(toolbar);
        }

    toolbar->winid = MS_CreateWindow(NXLT("ToolBarClass"), toolbar->name,
                                      WS_VISIBLE | WS_CHILD | WS_CLIPSIBLINGS,
                                      toolbar->xpos, toolbar->ypos,
                                      toolbar->width, toolbar->height,
                                      toolbar->parent->winid, (HMENU)NULL,
                                      (HANDLE)NULL, NULL);

    RealizeToolBarWidgets(toolbar);
    if ( (int)toolbar->window <= 1 )
        {
        toolbar->window = MS_GetDC(toolbar->winid);
        release_dc = TRUE;
        MS_SetClassLong(toolbar->winid, GCL_HCURSOR, (LONG)ThimPointerCursor);
        }
    PaintToolBar(toolbar, 0, 0, FALSE);

    if ( release_dc )
        {
        MS_ReleaseDC(toolbar->winid, toolbar->window);
        toolbar->window = NULL;
        }
	toolbar->mapped = TRUE;
}


/************************************************************************
 FUNCTION : ToolBarClassProc( HWND, UINT, WPARAM, LPARAM )
 PURPOSE  : Processes messages for the ToolBarClass
*/
LRESULT CALLBACK
ToolBarClassProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    int           widgetId;
    HDC           hDC;
    PAINTSTRUCT   ps;
    AXWidget      w;
    AXWidget_tool tb;
    XEvent xevent;
    LRESULT retval;


    /* Find toolbar associated with the event */
    tb = (AXWidget_tool) GetWidgetFromHWND(hWnd, W_TOOLBAR);
    if ( !tb )
        return MS_DefWindowProc(hWnd, message, wParam, lParam);

    retval = 0;
    switch( message )
        {
        case WM_SETCURSOR:
            {
            /* make sure help cursor is set when necessary. */
            if ( THIMhelpMode )
                {
                MS_SetCursor(tb->parent->help_cursor);
                retval = TRUE;
                }
            else
                retval = MS_DefWindowProc(hWnd, message, wParam, lParam);
            break;
            }


        case WM_MOUSEMOVE:
            {
            UINT state_mask = GetButtonMask(wParam);
            /* X automatically grabs the pointer on a button down,
             * so we do the same here on Windows if the button is down
             * and the mouse is moving (we release it on button up).
             */
            if ( state_mask == Button1Mask || state_mask == Button3Mask )
                MS_SetCapture(hWnd);
            break;
            }

        case WM_MOUSEACTIVATE:
            if ( TaskOfWidgetPended((AXWidget_main)tb) )
                {
                /* user tried to activate a pended dialog (i.e. a modal *
                 * dialog is taking precedence).        -donato         */
                RingTheBell();
                retval = MA_NOACTIVATEANDEAT;
                }
            else
                {
                /* if a rowcol or popup widget is active while the *
                 * user has activated another window, we must make *
                 * sure the widget is destroyed.                   */
                if ( RowColumnArmed )
                    {
                    PaintRowColumnPulldown(RowColumnArmed,OPTION_DONE,&xevent);
                    RowColumnArmed = NULL;
                    }
                if ( PopupArmed )
                    {
                    PaintPopupPulldown(PopupArmed, OPTION_DONE, &xevent);
                    PopupArmed = NULL;
                    }

                /* we don't want the toolbar to be "activated". *
                 * Focus should stay on the work space.         */
                if ( tb->parent->winid == MS_GetFocus() )
                    retval = MA_NOACTIVATE;
                else
                    retval = MA_ACTIVATE;
                }
            break;


        case WM_LBUTTONUP:
            if ( MS_GetCapture() == hWnd )
                MS_ReleaseCapture();
            break;


        case WM_COMMAND:
            widgetId = GetWidgetIdFromMenuId(tb, wParam);
            if ( widgetId >= 0 )
                {
                w = WIDGET_FROM_UID(widgetId);
                if ( !w->disabled )
                    CallBackToMenu(tb->parent, wParam);

                /* set focus back to the app space. otherwise *
                 * focus would stay on the toolbar widget.    */
                if ( tb->parent->workWindow )
                    MS_SetFocus(tb->parent->workWindow->winid);
                else if ( tb->parent->dialog )
                    MS_SetFocus(tb->parent->winid);
                }
            break;


        case WM_PAINT:
            hDC = MS_BeginPaint(hWnd, &ps);
            tb->window = hDC;
            PaintToolBar(tb, 0, 0, FALSE);
            MS_EndPaint(hWnd, &ps);
            tb->window = NULL;
            break;


        case WM_DRAWITEM:
            /* toolbar buttons have menu ids instead of widget ids,    *
             * but here we need a widget id so we can draw the button. */
            widgetId = GetWidgetIdFromMenuId(tb, wParam);
            if ( widgetId < 0 )
                /* can't find widget id, just continue. */
                retval = MS_DefWindowProc(hWnd, message, wParam, lParam);
            else
                {
                DrawButtonEvent(widgetId, lParam);
                retval = TRUE;
                }
            break;


        default:
            retval = MS_DefWindowProc(hWnd, message, wParam, lParam);
        }

    return retval;
}



static int
GetWidgetIdFromMenuId(AXWidget_tool toolbar, WPARAM menu_id)
{
    int ix;
    AXWidget pb;

    WIDGET_PARANOID(toolbar, W_TOOLBAR);
    for ( ix = 0; ix < MAX_TOOLBAR_ITEMS; ix++ )
        {
        pb = toolbar->toolBarInfo->widgets[ix];
        if ( !pb )    /* menu separators are null widgets in the array */
            continue;

        if ( ((AXWidget_pushbutton) pb)->menu_uid == (int)menu_id )
            return pb->uid;
        }
    return -1;
}



static void
ClearToolBarButtonUids(AXWidget_tool toolbar)
{
    int ix;
    AXWidget pb;


    WIDGET_PARANOID(toolbar, W_TOOLBAR);
    for ( ix = 0; ix < MAX_TOOLBAR_ITEMS; ix++ )
        {
        pb = toolbar->toolBarInfo->widgets[ix];
        if ( !pb )    /* menu separators are null widgets in the array */
            continue;

        if ( pb->uid )
            {
            CLEAR_WIDGET_FROM_UID_LIST(pb->uid);
            pb->uid = 0;
            }
        pb->winid = NULL;
        pb->window = NULL;
        pb->mapped = FALSE;
        }
}


void
UpdateToolBarWidgets(AXWidget_tool w_tool)
{
    int ix;
    AXWidget_pushbutton pb;

    for ( ix=0; ix < MAX_TOOLBAR_ITEMS; ix++ )
        {
        pb = (AXWidget_pushbutton) w_tool->toolBarInfo->widgets[ix];
        if ( pb && pb->update_widget && pb->display && pb->painter )
            {
            if ( !pb->mapped )
                {
                pb->menu_uid = GetUIDfromMenu(w_tool->parent->menuBar,
                                    w_tool->toolBarInfo->menuP[ix]->word,
                                    w_tool->toolBarInfo->menuP[ix]->word2);
                RealizePushButton(pb, TRUE);
                }
            pb->painter(pb, NULL, xmlibStr.FullRegenForUpdate, FALSE);
            pb->update_widget = FALSE;
            }
        }
}

