/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xim.h
   Applix ASC version: 23
   Date: Mon Nov 16 13:53:43 1998
*/
#if !XIM_INCLUDED

#define XIM_INCLUDED 1


/**********************************************************************/

#define MAX_OFF_SPOT 300  /* max number of characters in off-the-spot widget */
#define MAX_BYTE_OFF_SPOT (MAX_OFF_SPOT*2)

#if BUILD_ORIENT


#ifdef XIMPreeditArea   /* this is defined by X11 R5 if it supports the*/
                        /*   X Input Method */
#define USE_XIM 1
#endif

#if !USE_XIM              /* {  */

/**********************************************************************/

#include <egcd.h>

typedef char *XPointer;
typedef char *XVaNestedList;



typedef unsigned long XIMStyle;
 
typedef struct {
    unsigned short count_styles;
    XIMStyle *supported_styles;
} XIMStyles;


#define XNArea "area"
#define XNInputStyle "inputStyle"
#define XNClientWindow "clientWindow"
#define XNPreeditAttributes "preeditAttributes"
#define XNStatusAttributes "statusAttributes"
#define XNPreeditStartCallback "preeditStartCallback"
#define XNPreeditDoneCallback "preeditDoneCallback"
#define XNPreeditDrawCallback "preeditDrawCallback"
#define XNPreeditCaretCallback "preeditCaretCallback"
#define XNStatusStartCallback "statusStartCallback"
#define XNStatusDoneCallback "statusDoneCallback"
#define XNStatusDrawCallback "statusDrawCallback"
#define XNForeground "foreGround"
#define XNBackground "backGround"
#define XNFocusWindow "focusWindow"




/* types for XCreateIC()   */
#define XIMPreeditArea      0x0001
#define XIMPreeditCallbacks 0x0002
#define XIMStatusCallbacks  0x0200
#define XIMStatusArea       0x0100

/** status values from XmbLookupString()  */
#define XLookupNone     1
#define XLookupChars        2
#define XLookupKeySym       3
#define XLookupBoth     4



typedef char *XFontSet;


typedef struct XIM_struct {
   int type;
   } XIM_struct;

typedef XIM_struct *XIM;

typedef struct XIMDB_struct {
   int type;
   } XIMDB_struct;

typedef XIMDB_struct *XmDataBase;



#ifndef XPointer
#define XPointer unsigned char *
#define XIMStatusDataType int
#endif

typedef void (*XIMProc)();

typedef struct {
    XPointer client_data;
    XIMProc callback;
    } XIMCallback;


typedef unsigned char XIMFeedback;

/* define the possible feedback styles */
#define XIMReverse 1
#define XIMUnderline 2
#define XIMHighlight 4

typedef struct _XIMText {
    unsigned short length;
    XIMFeedback    *feedback;
    int            encoding_is_wchar;
    union  {
           unsigned char *multi_byte;
           unsigned short *wide_char;
           } string;
    } XIMText;


typedef struct _XIMPreeditDrawCallbackStruct {
    int caret;
    int chg_first;
    int chg_length;
    XIMText  *text;
    }  XIMPreeditDrawCallbackStruct;

/* the XIMStatusDataType field below can have the following values */
#define XIMTextType 1
#define XIMBitmapType 2

typedef struct _XIMStatusDrawCallbackStruct {
    XIMStatusDataType type;
    union {
        XIMText *text;
        Pixmap  bitmap;
          } data;
    } XIMStatusDrawCallbackStruct;



typedef struct IC_struct {
	EGC_JBUF       jbuf;	     /* current data for this input context */
	EGC_JTABLE     *egc_table;   /* current server communication buffer */
	unsigned char  *mb_buffP;    /* the mb version of "jbuf.kanji_buf"  */
	int            len_mb_buff;  /* length of mb_buffP, in "wchar" units */
	unsigned short	*output_buf;    /* for SetXLookupStaticString() */
	int            total_hmnyms; /* total number of homonyms */

	/* 
	 * various input status, we hope these valid for most conversion server 
	 */
	unsigned
	/* Four preedit status */
	                in_edition,	/* 0:not inputed, 1:inputed */
                    in_replace,	/* 0:not replaced, 1:replaced */
	                in_conversion,	/* 0:not converted, 1:converted */
	                in_homonyms,	/* 0:no homonyms, 1:some homonyms */
	/* Two conversion system modes */
	                in_automatic,	/* auto conversion or manual */
	                in_learning,	/* learning into dictionary or not */
	/* Six input modes */
	                in_code,	/* jis code input or not */
	                in_symbol,	/* symbol input or not */
	                in_english,	/* English chars  or Japanese chars */
	                in_roma,	/* phonetic input or direct input */
	                in_hiragana,	/* hiragana or katakana */
	                in_zenkaku;	/* wide chars or narrow chars */
	/* 								*/
	/* The combinations of above six input modes which have reality	*/
	/* are listed below (total 10): 				*/
	/*								*/
	/* (1) in_code=1 						*/
	/* (2) in_symbol=1						*/
	/* in_code=0 and in_symbol=0 					*/
	/*	in_english=1						*/
	/* (3) 		in_zenkaku=1 					*/
	/* (4) 		in_zenkaku=0					*/
	/* in_code=0 and in_symbol=0 and in_egnlish=0			*/
	/*	in_roma=1						*/
	/* (5)		in_hiragana=1					*/
	/*		in_hiragana=0 					*/
	/* (6)			in_zenkaku=1				*/
	/* (7)			in_zenkaku=0				*/
	/*	in_roma=0						*/
	/* (8)		in_hiragana=1					*/
	/*		in_hiragana=0 					*/
	/* (9)			in_zenkaku=1 				*/
	/* (10)			in_zenkaku=0 				*/
	/*								*/


	/* We allow first phrase with extra conversion varieties.	*/
	/* Wheather in edition, in conversion or in homonyms, the	*/
	/* exchange of char set from kanji-hiragana to katakana or	*/
	/* hankaku-katakana is permitted. And when conversion is	*/
	/* started again, we let it go to the normal kanji-hiragana	*/
	/* form and do conversion. 					*/
	/* The 10 forms are:						*/
	/* 	DEFAULT_FORM(kanji-kana, or english as inputed),	*/
	/*	HIRAGANA_FORM(hiragana),				*/
	/*	KATAKANA_FORM(katakana),				*/
	/*	HANKAKU_KATAKANA_FORM(hankaku-katakana).		*/
	/* (if in english input)					*/
	/*	WIDE_roma_FORM(wide-roma),				*/
	/*	WIDE_Roma_FORM(Wide-Roma),				*/
	/*	WIDE_ROMA_FORM(WIDE-ROMA),				*/
	/*	NORMAL_roma_FORM(roma),					*/
	/*	NORMAL_Roma_FORM(Roma),					*/
	/*	NORMAL_ROMA_FORM(ROMA).					*/
	int		first_phrase_form;
												 

	struct axw_main          main_w;
	struct axw_entry_field          entry_widget;
#if 0
	Widget          homonym_widget;
#endif
	/* the following are the callback functions for this input context */
	XIMCallback     preedit_start;
	XIMCallback     preedit_done;
	XIMCallback     preedit_draw;
	XIMCallback     preedit_cursor_move;
	XIMCallback     status_start;
	XIMCallback     status_done;
	XIMCallback     status_draw;
}               IC_struct;

typedef IC_struct *XIC;


XVaNestedList XVaCreateNestedList();
XIC XCreateIC();
XIM XOpenIM();

#endif    /*  }   end of X11 R5 exclusion */


/* The definition of 10 forms of first phrase */
#define DEFAULT_FORM		0
#define HIRAGANA_FORM		1
#define KATAKANA_FORM		2
#define HANKAKU_KATAKANA_FORM	3
/*
#define WIDE_roma_FORM		4
#define WIDE_Roma_FORM		5
#define WIDE_ROMA_FORM		6
#define NORMAL_roma_FORM	7
#define NORMAL_Roma_FORM	8
#define NORMAL_ROMA_FORM	9
*/


#define CONVERSION_TOGGLE    XK_Henkan_Mode
#define LEARN_TOGGLE         XK_F9
#define CODE_INPUT_TOGGLE    XK_F10
#if SUN4
/** on a sun keyboard F11 and F12 actually generate XK_SunF36 and XK_SunF37 */
#include <X11/Sunkeysym.h>
#define MANUAL_AOTO_TOGGLE   XK_SunF36
#define INPUT_MODE_CHANGE    XK_SunF37
#else
#define MANUAL_AOTO_TOGGLE   XK_F11
#define INPUT_MODE_CHANGE    XK_F12
#endif


/* status return values from AXOpenIM()  */
#define NO_KANJI_FONT -1
#define NO_KANJI_CONVERTER -2
#define INVALID_LOCALE -3
#define NOT_LANG_LOCALE -4


#define PREEDIT_START 0
#define PREEDIT_DONE  1
#define PREEDIT_DRAW  2
#define PREDIT_CURSOR_MOVE 3
#define STATUS_START 4
#define STATUS_DONE 5
#define STATUS_DRAW 6


XIM AXOpenIM();




/***************************************************************/
/* define some elements for a pseudo-font-set                  */
/***************************************************************/
#define NARROW_FONT 0
#define WIDE_FONT 1
#define NARROW_KANA_FONT 2


typedef struct font_info_struct {
    struct font_info_struct *next;
    int height;
    int pointsize;
    int sans;
    Font narrow_font_id;
    Font wide_font_id;
    Font narrow_kana_id;
    XFontStruct *narrow_font;
    XFontStruct *narrow_kana;
    XFontStruct *wide_font;
    } font_info_struct;

typedef font_info_struct *FontSet;


#include <wide_extern.h>




#endif   /* BUILD_ORIENT */

#endif   /* XIM_INCLUDED */
