/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xmbutton.c
   Applix ASC version: 89
   Date: Mon Nov 16 13:53:35 1998
*/
#include <ax_xm.h>
#include <aos.h>
#include <xmlib.h>
#include <elxwins.h>
#include <xmpixel.h>
#include <axtext.h>

typedef struct {
    Pixmap map, imap;
    int hyt, wid;
} StippleMap;

extern AXWidget_pushbutton ButtonArmed;

extern int AxCharAbove, SmallCharAbove;
extern short ColorDisplay;

extern int AxOpenLook;
extern int UseWidgetColors;
extern int AxBigFont;
extern int NoTransient;
extern int Thim2dMode;
extern int AxNewMailMenuIcon;
extern int AxAlternateDialogFont;
extern int SmallCharHeight, AxCharHeight;
extern AXWidget ToolArmedForUpdate;

extern GC ForeGC, BackGC, RecessGC;
extern GC DisabledGC, SunGC, ShadowGC;
extern GC SmallFontGC, SmallFontBackGC;

extern GC olTextGC, olDisabledTextGC, olHilitTextGC;
extern GC olMiniTextGC, olMiniDisabledTextGC, olMiniHilitTextGC;
extern GC olButtonGC, olDisabledButtonGC, olHilitButtonGC;
extern GC olMiniButtonGC, olMiniHilitButtonGC;
extern GC olAsterixGC, olHilitAsterixGC, olBusyButtonGC;

extern GC olClrForeButtonGC, olClrBackButtonGC;
extern GC olClrSunButtonGC, olClrShadowButtonGC, olClrRecessButtonGC;
extern GC olClrSunMiniButtonGC, olClrShadowMiniButtonGC;
extern GC olClrBackMiniButtonGC, olClrRecessMiniButtonGC;

extern XFontStruct *AxDlgMainFont;
extern XFontStruct *Txfont, *MiniTxfont;
extern XFontStruct *Grafont, *MiniButtonFont;
extern Display *Dpy;

extern unsigned long AxPlaneMask;
extern GC AxColorGC;


extern pixcache_t *Pixmapper();
extern int wShadowIsOn();
extern void *AxRenderInsetGraphic();
extern void paintCompositeWidget();
extern AXWidget topWidget();

void ArmButton(), PaintButton();
void xmDrawRectShadow(), xmDrawRectAntiShadow();
void xmColorFillRectangle(), xmLiftRectShadow();


/*************************************************************************
function: ArmButton - Set/Reset button & make callback
*/
void
ArmButton (w,up,eventP)
register AXWidget_pushbutton w;
register int up;    /* TRUE if button up, FALSE on down */
register XEvent *eventP;
{
    XmAnyCallbackStruct data;
    int x, y;
    AXWidget_pushbutton tw;
    AXWidget child;

    /* w->changed is used to hilight/unhilight button
     * depending on pointer motion inside/outside button.
     * When a button is pressed, w->changed is FALSE.
     * w->changed becomes TRUE if pointer moves outside. */

    if ( ButtonArmed) {
        if (eventP && eventP->type==MotionNotify) { 
            /* motion while button selected: repaint button to reflect 
             * new state (if needed), but don't reset ButtonArmed */
            x = eventP->xmotion.x;
            y = eventP->xmotion.y;
            if ( (x >= w->xpos) && (x < w->xpos+w->width)
                        && (y >= w->ypos) && (y < w->ypos+w->height) ) {
                /* motion inside button */
                if ( w->changed ) {
                    w->changed = FALSE;
                    w->value = !w->value;
                    PaintButton(w, NULL, TRUE, TRUE);
                }
            }
            else if ( !w->minorEventNotify ) {
                /* motion outside button */
                if ( !w->changed ) {
                    w->changed = TRUE;
                    w->value = !w->value;
                    PaintButton(w, NULL, TRUE, TRUE);
                }
            }
            return;
        }
        /* if here, either ButtonRelease or ResetActive() */
        if ( ButtonArmed->parent->type == W_TOOLBAR ) {
            ToolArmedForUpdate = (AXWidget)ButtonArmed;
        }
        else if (w && w->gePaletteButtonType && up) 
            ; /* unfortunate hack needed to eliminate awful flicker */
        else if ( w && w->minorEventNotify )
            ; /* let application disarm... */
        else if ( w && eventP && (eventP->type == ButtonRelease) ) {
			x = eventP->xbutton.x;
			y = eventP->xbutton.y;
			if ( (x >= ButtonArmed->xpos) &&
				(x < ButtonArmed->xpos+ButtonArmed->width) &&
				(y >= ButtonArmed->ypos) &&
				(y < ButtonArmed->ypos+ButtonArmed->height) ) {
				
				ButtonArmed->value = !ButtonArmed->value;
				PaintButton(ButtonArmed, NULL, TRUE, TRUE);
			}
        }
        else if ( w && eventP && (eventP->type == KeyRelease) ) {
			ButtonArmed->value = !ButtonArmed->value;
			PaintButton(ButtonArmed, NULL, TRUE, TRUE);
        }
    }
    if ( !w ) {
        if ( ButtonArmed ) {
            if ( up && ButtonArmed->minorEventNotify ) {
                w = ButtonArmed;
                data.reason = EV_DISARMED_;
                if ( w->callback )
                    AxCallback(w->callback,w,w->callbacktag,&data);
            }
            ButtonArmed->changed = FALSE;
            ButtonArmed = 0;
        }
        return;
    }

    if ( eventP ) {
        if (eventP->type==ButtonPress || eventP->type==ButtonRelease) {
            if ((eventP->xbutton.button & 0xff) != w->whichButton) {
                return;
            }
        }
        else if (eventP->type == MotionNotify) {
            if (w->whichButton == Button1) {
                if ( (eventP->xmotion.state & Button1Mask) == 0 )
                    return;
            }
            else if (w->whichButton == Button3) {
                if ( (eventP->xmotion.state & Button3Mask) == 0 )
                    return;
            }
        }
    }

    data.event = eventP;
    if ( up ) {
        if ( !ButtonArmed || w != ButtonArmed ) {
            if ( ButtonArmed ) {
                ButtonArmed->changed = FALSE;
                ButtonArmed = 0;
            }
            return;
        }
        ButtonArmed->changed = FALSE;
        ButtonArmed = 0;

		if (w->parent->type == W_DIALOG)
		{
			AXWidget_dialog d_w;
			d_w = (AXWidget_dialog)w->parent;
			if ( d_w->autoUnmanage )
                AxXtDestroyWidget(w->parent->parent); /* destroy dlg shell */
        }

        if ( w->arm_callback ) /* type is downstroke reporter */
            return;
        if ( w->minorEventNotify )
            data.reason = EV_DISARMED_;
        else
            data.reason = EV_CHANGED_;
        if ( w->callback )
            AxCallback(w->callback,w,w->callbacktag,&data);
        else if ( w->parent->callback )
            AxCallback(w->parent->callback,w,w->parent->callbacktag,&data);
    }
    else {
        if ( w->minorEventNotify ) {
            data.reason = EV_ARMED_;
            w->value = TRUE;
            PaintButton(ButtonArmed = w, NULL, TRUE, TRUE);
            if ( w->callback && w->minorEventNotify) {
                AxCallback(w->callback,w,w->callbacktag,&data);
            }
        }
        else {
            data.reason = EV_CHANGED_;
            if ( w->gePaletteButtonType )
                w->value = TRUE;
            else
                w->value = !w->value;
            PaintButton(ButtonArmed = w, NULL, TRUE, TRUE);
            if ( w->callback && w->arm_callback ) { /* type is downstroke reporter */
                ButtonArmed->changed = FALSE;
                ButtonArmed = 0;    /* does not stay actively sensitive */
                AxCallback(w->callback,w,w->callbacktag,&data);
            }
        }
    }
}

/***************************************************************************/
/* THIS FUNCTION IS FOR UNIX/X ONLY!  the WINDOWS version is in WINBUTTN.C */
/* ALL COMMON BUTTON FUNCTIONS SHOULD GO IN BUTTNCOM.C   - donato          */
/*************************************************************************
function: PaintButton - Paint a W_BUTTON instance
*/
void
PaintButton (wp, xev, regen, layered)
register AXWidget wp;
XEvent *xev;
int regen;
int layered;
{
    register GC gc;
    register int width, height;
    register int bthick, bvalue;
    int x, y;
    int xoff, yoff;
    AXWidget w;
	AXWidget_pushbutton wb;
    pixcache_t *pxc;
    Pixmap pix;

    if (wp->parent->type == W_PLATES) {
        if(ssDrawPlatesPushButton(wp))
			return;
    }

    if ( !wp->display )
        return;
	
	if ( !WIN_OF(wp) ) {
		return;
	}

	 /* allow option menu itself to manage its buttons */
    if ( (wp->type == W_TOGGLE) && ((AXWidget_togglebutton)wp)->in_option )
        return;

	if (wp->parent && (wp->parent->type == W_DIALOG) &&
				   !((AXWidget_dialog)wp->parent)->exposed)
		return;

	if (layered) {
		if (wp->flags & WFLG_IN_EDIT) {
			if (dePaintWidget(wp)) {
				return;
			}
		}
		else {
			if (topWidget(wp) == wp) {
				paintCompositeWidget(wp, xev, regen);
				return;
			}
		}
	}
	wb = (AXWidget_pushbutton)wp;
    if (!wb->width || !wb->height || wb->pixmapLabel) {
        wVerbButtonSize(wb->name, wb->pixmapLabel, wb->font,
                        wb->thickness, wb->shadow, wb->textShadow,
                        &wb->width, &wb->height);
    }

    wb->mapped = TRUE;
	
    if ( wb->thickness )
        bthick = wb->thickness;
    else
        bthick = 2;
	
	if ( wb->disabled )
		bvalue = FALSE;
	else
		bvalue = wb->value;
    
    x = wb->xpos;
    y = wb->ypos;
    width = wb->width;
    height = wb->height;
    
    if ( wShadowIsOn(wb->shadow) ) {
        if ((bvalue && !wb->disabled) ||
			(wb->pixmapLabel && wb->disabled)) {
			
			/* lift shadow while pressed, or if disabled pixbutton
			 */
			xmLiftRectShadow(wb, BackGC, x, y, width, height);
        }
        else {
            xmDrawRectShadow(wb, x, y, width, height);
        }
    }
    
	/* draw the face
	 */
    if ( bvalue ) {
		/* draw the button face in its pressed state
		 */
		if (UseWidgetColors) {
			/* note that we're using the recessed color
			 * despite the face color of the button..
			 */
			XFillRectangle(Dpy, WIN_OF(wb), RecessGC, x+2, y+2,
						   width-4, height-4);
		}
		else {
			XFillRectangle(Dpy, WIN_OF(wb), ForeGC, x+3, y+3,
						   width-6, height-6);
		}
    }
    else {
		/* draw the button face in its unpressed state
		 */
		if (wb->widgetColor) {
			xmColorFillRectangle(wb, WIN_OF(wb), BackGC, x, y, width, height);
		}
		else {
			XFillRectangle(Dpy,WIN_OF(wb),BackGC,x,y,width,height);
		}
    }
    if ( wb->disabled ) {
        gc = DisabledGC;
    }
	else if ( bvalue ) {
		gc = UseWidgetColors ? ForeGC : BackGC;
	}
	else {
		gc = ForeGC;
	}
	if (wb->pixmapLabel) {
		
		/* draw the pixmap onto the button
		 */
		int px = x;
		int py = y;
		
		px += bthick;
		py += bthick;
		
		if ( !wb->noBorder ) {  /* apply 1pix border around image */
			px++;
			py++;
		}
		
		if ( bvalue && wb->alternatePixmap ) {
			pxc = Pixmapper(wb->alternatePixmap);
			if ( pxc )
				pix = pxc->pix;
		}
		else {
			pxc = Pixmapper(wb->name);
			if ( pxc ) {
				if ( bvalue && !wb->disabled && pxc->invert_pix )
					pix = pxc->invert_pix;
				else
					pix = pxc->pix;
			}
		}
		if ( !pxc )
			return;
		
		if (pxc->clip_mask) {
			XSetClipOrigin(Dpy, gc, px, py);
			XSetClipMask(Dpy, gc, pxc->clip_mask);
		}
		XCopyArea(Dpy,pix,WIN_OF(wb),gc,0,0,pxc->width,pxc->height,px,py);
		if (pxc->clip_mask) {
			XSetClipMask(Dpy, gc, None);
		}
#if INVERTABLE
		if ( bvalue && !wb->no_arm_reverse ) {
			XFillRectangle(Dpy, WIN_OF(wb), InvertGC, px, py,
						   pxc->width, pxc->height);
		}
#endif
        if (ButtonArmed && wb->parent->type==W_PALETTE)
		{
			/* simulate radio effect for palette buttons.
			 * do this hack only during the ButtonArmed state,
			 * and only when the armed button is painted.
			 */
			AXWidget_pushbutton  saveArmed;
			    /* shut off ButtonArmed temporarily to
				 * prevent recursion into this block
				 */
			saveArmed = ButtonArmed;
			ButtonArmed = (AXWidget_pushbutton) NULL;
            for (w=wb->parent->widgets;w;w=w->nextChild) {
                if (wb->gePaletteButtonType ==
					((AXWidget_pushbutton)w)->gePaletteButtonType)
				{
                    if (w->value && wb->name!=w->name && w->display) {
                        w->value = bvalue ? FALSE : TRUE;
                        PaintButton(w, xev, regen, layered);
                        w->value = TRUE;
                        break;
                    }
                }
            }
			ButtonArmed = saveArmed;    /* restore */
        }
	}
	else {
		/* draw the button's label
		 */
		int tx = x;
		int ty = y;
		centerButtonText(wb->width, wb->height, wb->name, wb->font,
						 wb->small_font, &tx, &ty);
		xmDrawStringAndAccelerator(wb, WIN_OF(wb), gc, tx, ty, wb->name,
								   MBstrlen(wb->name), wb->textShadow,
								   wb->textColor, wb->widgetColor,
								   wb->accelerator_key);
	}
	
	/* draw the bevel
	 */
	if (wb->hasFocus && wb->parent->hasFocus) {
		/* reduce button bevel 1pix all around;
		 * focus indicator will pick up the difference
		 */
		x++;
		y++;
		width -= 2;
		height -= 2;
	}
	if ( wb->pixmapLabel && wb->disabled ) {
		AxDrawBevel(wb, WIN_OF(wb), x, y, width, height, bthick, BEVEL_NONE);
	}
	else if ( wb->use2d || Thim2dMode ) {
		AxDrawBevel(wb, WIN_OF(wb), x, y, width, height,
					bthick, BEVEL_SIMPLE);
	}
	else if ( bvalue ) {
		AxDrawBevel(wb, WIN_OF(wb), x, y, width, height,
					bthick, BEVEL_IN);
	}
	else {
		AxDrawBevel(wb, WIN_OF(wb), x, y, width, height,
					bthick, BEVEL_OUT);
		if (wShadowIsOn(wb->shadow)) {
			xmDrawRectAntiShadow(wb, x, y, width, height);
		}
	}
	if (wb->graphic) {
		AxRenderInsetGraphic(wb, wb->graphic);
	}
	
    if ( (wb->parent->type == W_DIALOG ) &&
		 ((AXWidget_dialog)wb->parent)->currentDefaultButton == wb ) {
		
		/* draw the default indicator
		 */
#if 0        /* motif-style */
        if ( wb->use2d || Thim2dMode ) {
			AxDrawBevel(wb, WIN_OF(wb), x - 3, y - 3, width + 6,
						height + 6, 1, BEVEL_SIMPLE);
		}
        else {
			AxDrawBevel(wb, WIN_OF(wb), x - 3, y - 3, width + 6,
						height + 6, 1, BEVEL_IN);
		}
#else        /* windows-style */
		char dashes[4];
		XSetLineAttributes(Dpy, ForeGC, 0, LineOnOffDash, CapButt,
						   JoinMiter);
		dashes[0] = 1;
		dashes[1] = dashes[2] = dashes[3] = 0;
		XSetDashes(Dpy, ForeGC, 0, dashes, 1);
		XDrawRectangle(Dpy, WIN_OF(wb), ForeGC, x + 3, y + 3,
					   width - 7, height - 7);
		XSetLineAttributes(Dpy, ForeGC, 0, LineSolid, CapButt, JoinMiter);
#endif
    }
}
