/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xmchoose.h
   Applix ASC version: 11
   Date: Mon Nov 16 13:53:46 1998
*/
#ifndef _xm_choose_h_
#define _xm_choose_h_

#include <xmlib.h>

#define MAX_H_LEV 36  /* max number of Hue levels. x-axis. */
#define MAX_C_LEV 10  /* max number of Chroma levels. y-axis. */

typedef struct CcColor {
	int r, g, b;
	float h, v, c;
	unsigned long pixel;
} CcColor;

/* slider styles. */
#define SLIDER_HORIZ 1
#define SLIDER_VERT  2

#define SLIDER_THICKNESS 16
#define SLIDER_BARLEN    24
#define SLIDER_MINLEN    (96 + SLIDER_BARLEN)

/* mini-slider */
typedef struct {
	AXWidget parent;
	char *name;
	int x, y, width, height;
	int style; /* SLIDER_VERT/SLIDER_HORIZ */
	int min_value, max_value, step_value;
	int value;
	int cur_pos; /* pixel position */
	int length;
	int armed;
	void (*value_changed_cb)();
	caddr_t closure;
} CcSlider;

/* Color Chooser Data (Private to xmchoose.c and xmlib.c::XtSetValues()) */
typedef struct chooser_data {
	int win_width;
	int win_height;
	int pad_x, pad_y, pad_width, pad_height;
	int sample_x, sample_y, sample_width, sample_height;
	int pad_armed;
	int selected_cell_idx; /* idx into the `colors' array */
	CcColor colors[MAX_H_LEV * MAX_C_LEV];
	CcSlider *h_slider, *s_slider, *v_slider;
} chooser_data;

/* Event types */
typedef struct {
	int type; /* one of the types defined in evreason.h */
} CcAnyEvent;

typedef struct {
	int type;
	float hue, saturation, brightness;
	int red, green, blue;
} CcColorChangeEvent;

typedef union {
	CcAnyEvent any;
	CcColorChangeEvent value;
} ChooserEventData;

/* Public fcns */
Widget AXmCreateColorChooser AXP_((widget_t *parentWidget, char *name,
								   Arg args[], int cnt));
extern int  ColorChooserEvent AXP_((widget_t *w, XEvent *evp));
extern void DrawColorChooserBorder AXP_((widget_t *w));
extern void PaintColorChooser AXP_((AXWidget_color_chooser w,
									XExposeEvent *exP, int regen,
									int layered));

#endif /* !_xm_choose_h_ */


/* GNU Emacs settings - This (including the "\n^L") MUST be the last thing
 *                      in this file!
 * Local variables:
 * mode: c
 * tab-width: 4
 * end:
 */
