/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xmcursor.c
   Applix ASC version: 48
   Date: Mon Nov 16 13:53:45 1998
*/
#include <stdio.h>
#include <dwt2xm.h>
#include <basicwin.h>
#if AX_X11
#include <X11/cursorfont.h>
#else
#include <x2wincur.h>
#endif
#include <aos.h>
#include <thimble.h>
#include <paripc.h>
#include <xmlib.h>
#include <tmi.h>
#include <paratm.h>
#include <parint.h>
#include <paratmgr.h>
#include <usrevent.h>
#include <parwin.h>
#include <aiconids.h>
#include <cursors.h>
#include <axbitmap.h>
#include <java_syms.h>
#include <elxwins.h>


#define MIT_CURSOR_FONT_ID 0
#define OL_CURSOR_FONT_ID 1
#define MAX_GUN 65535L 
#define MAX_APP_CURSORS 20

typedef struct {
    Cursor curs;
    int iconid;
} app_curs_info;

typedef struct {
    char *name;  /* Asterix cursor_name, from cursors.h */
    int font_id; /* font containing cursor glyph */
    int id;      /* for MIT_CURSOR_FONT_ID, from <X11/cursorfont.h>,
                  * for other fonts, index into font */
} cursormap;

static app_curs_info AppCursors[MAX_APP_CURSORS];

static cursormap CursorMap[] = {    /* maps our names to curfont.h */
    {XMITC_X_cursor,MIT_CURSOR_FONT_ID,XC_X_cursor},
    {XMITC_arrow,MIT_CURSOR_FONT_ID,XC_arrow},
    {XMITC_based_arrow_down,MIT_CURSOR_FONT_ID,XC_based_arrow_down},
    {XMITC_based_arrow_up,MIT_CURSOR_FONT_ID,XC_based_arrow_up},
    {XMITC_boat,MIT_CURSOR_FONT_ID,XC_boat},
    {XMITC_bogosity,MIT_CURSOR_FONT_ID,XC_bogosity},
    {XMITC_bottom_left_corner,MIT_CURSOR_FONT_ID,XC_bottom_left_corner},
    {XMITC_bottom_right_corner,MIT_CURSOR_FONT_ID,XC_bottom_right_corner},
    {XMITC_bottom_side,MIT_CURSOR_FONT_ID,XC_bottom_side},
    {XMITC_bottom_tee,MIT_CURSOR_FONT_ID,XC_bottom_tee},
    {XMITC_box_spiral,MIT_CURSOR_FONT_ID,XC_box_spiral},
    {XMITC_center_ptr,MIT_CURSOR_FONT_ID,XC_center_ptr},
    {XMITC_circle,MIT_CURSOR_FONT_ID,XC_circle},
    {XMITC_clock,MIT_CURSOR_FONT_ID,XC_clock},
    {XMITC_coffee_mug,MIT_CURSOR_FONT_ID,XC_coffee_mug},
    {XMITC_cross,MIT_CURSOR_FONT_ID,XC_cross},
    {XMITC_cross_reverse,MIT_CURSOR_FONT_ID,XC_cross_reverse},
    {XMITC_crosshair,MIT_CURSOR_FONT_ID,XC_crosshair},
    {XMITC_diamond_cross,MIT_CURSOR_FONT_ID,XC_diamond_cross},
    {XMITC_dot,MIT_CURSOR_FONT_ID,XC_dot},
    {XMITC_dotbox,MIT_CURSOR_FONT_ID,XC_dotbox},
    {XMITC_double_arrow,MIT_CURSOR_FONT_ID,XC_double_arrow},
    {XMITC_draft_large,MIT_CURSOR_FONT_ID,XC_draft_large},
    {XMITC_draft_small,MIT_CURSOR_FONT_ID,XC_draft_small},
    {XMITC_draped_box,MIT_CURSOR_FONT_ID,XC_draped_box},
    {XMITC_exchange,MIT_CURSOR_FONT_ID,XC_exchange},
    {XMITC_fleur,MIT_CURSOR_FONT_ID,XC_fleur},
    {XMITC_gobbler,MIT_CURSOR_FONT_ID,XC_gobbler},
    {XMITC_gumby,MIT_CURSOR_FONT_ID,XC_gumby},
    {XMITC_hand1,MIT_CURSOR_FONT_ID,XC_hand1},
    {XMITC_hand2,MIT_CURSOR_FONT_ID,XC_hand2},
    {XMITC_heart,MIT_CURSOR_FONT_ID,XC_heart},
    {OLC_hourglass,OL_CURSOR_FONT_ID,26},
    {XMITC_icon,MIT_CURSOR_FONT_ID,XC_icon},
    {XMITC_iron_cross,MIT_CURSOR_FONT_ID,XC_iron_cross},
    {XMITC_left_ptr,MIT_CURSOR_FONT_ID,XC_left_ptr},
    {XMITC_left_side,MIT_CURSOR_FONT_ID,XC_left_side},
    {XMITC_left_tee,MIT_CURSOR_FONT_ID,XC_left_tee},
    {XMITC_leftbutton,MIT_CURSOR_FONT_ID,XC_leftbutton},
    {XMITC_ll_angle,MIT_CURSOR_FONT_ID,XC_ll_angle},
    {XMITC_lr_angle,MIT_CURSOR_FONT_ID,XC_lr_angle},
    {XMITC_man,MIT_CURSOR_FONT_ID,XC_man},
    {XMITC_middlebutton,MIT_CURSOR_FONT_ID,XC_middlebutton},
    {XMITC_mouse,MIT_CURSOR_FONT_ID,XC_mouse},
    {XMITC_pencil,MIT_CURSOR_FONT_ID,XC_pencil},
    {XMITC_pirate,MIT_CURSOR_FONT_ID,XC_pirate},
    {XMITC_plus,MIT_CURSOR_FONT_ID,XC_plus},
    {XMITC_question_arrow,MIT_CURSOR_FONT_ID,XC_question_arrow},
    {XMITC_right_ptr,MIT_CURSOR_FONT_ID,XC_right_ptr},
    {XMITC_right_side,MIT_CURSOR_FONT_ID,XC_right_side},
    {XMITC_right_tee,MIT_CURSOR_FONT_ID,XC_right_tee},
    {XMITC_rightbutton,MIT_CURSOR_FONT_ID,XC_rightbutton},
    {XMITC_rtl_logo,MIT_CURSOR_FONT_ID,XC_rtl_logo},
    {XMITC_sailboat,MIT_CURSOR_FONT_ID,XC_sailboat},
    {XMITC_sb_down_arrow,MIT_CURSOR_FONT_ID,XC_sb_down_arrow},
    {XMITC_sb_h_double_arrow,MIT_CURSOR_FONT_ID,XC_sb_h_double_arrow},
    {XMITC_sb_left_arrow,MIT_CURSOR_FONT_ID,XC_sb_left_arrow},
    {XMITC_sb_right_arrow,MIT_CURSOR_FONT_ID,XC_sb_right_arrow},
    {XMITC_sb_up_arrow,MIT_CURSOR_FONT_ID,XC_sb_up_arrow},
    {XMITC_sb_v_double_arrow,MIT_CURSOR_FONT_ID,XC_sb_v_double_arrow},
    {XMITC_shuttle,MIT_CURSOR_FONT_ID,XC_shuttle},
    {XMITC_sizing,MIT_CURSOR_FONT_ID,XC_sizing},
    {XMITC_spider,MIT_CURSOR_FONT_ID,XC_spider},
    {XMITC_spraycan,MIT_CURSOR_FONT_ID,XC_spraycan},
    {XMITC_star,MIT_CURSOR_FONT_ID,XC_star},
    {OLC_stop_watch,OL_CURSOR_FONT_ID,6},
    {XMITC_target,MIT_CURSOR_FONT_ID,XC_target},
    {XMITC_tcross,MIT_CURSOR_FONT_ID,XC_tcross},
    {XMITC_top_left_arrow,MIT_CURSOR_FONT_ID,XC_top_left_arrow},
    {XMITC_top_left_corner,MIT_CURSOR_FONT_ID,XC_top_left_corner},
    {XMITC_top_right_corner,MIT_CURSOR_FONT_ID,XC_top_right_corner},
    {XMITC_top_side,MIT_CURSOR_FONT_ID,XC_top_side},
    {XMITC_top_tee,MIT_CURSOR_FONT_ID,XC_top_tee},
    {XMITC_trek,MIT_CURSOR_FONT_ID,XC_trek},
    {XMITC_ul_angle,MIT_CURSOR_FONT_ID,XC_ul_angle},
    {XMITC_umbrella,MIT_CURSOR_FONT_ID,XC_umbrella},
    {XMITC_ur_angle,MIT_CURSOR_FONT_ID,XC_ur_angle},
    {XMITC_watch,MIT_CURSOR_FONT_ID,XC_watch},
    {XMITC_xterm,MIT_CURSOR_FONT_ID,XC_xterm},
    {NULL,0},
};

extern int ATMinitCOMPLETE;    /* TRUE if initialization completed */
extern int THIMhelpMode;       /* TRUE while in help mode */
extern int AxElfrtMode;        /* TRUE if elfrt */
extern int AxOpenLook;         /* TRUE while running under OpenLook */
extern int InSysHourglassMode; /* TRUE while all windows glassed */
extern int ElfInLearnMode;     /* TRUE while in learn mode */
extern int JavaMode;

extern AXWidget WindowList;

extern Cursor ThimPointerCursor, ThimRotateCursor;
extern Cursor ThimUpDownCursor, ThimLeftRightCursor;
extern Cursor ThimTopRightCursor, ThimTopLeftCursor;
extern Cursor ThimBottomRightCursor, ThimBottomLeftCursor;
extern Cursor ThimFleurCursor, ThimTextCursor;
extern Cursor ThimBusyCursor;
extern Cursor ThimRecordCursor;
extern Cursor ThimNoRecordCursor;
extern Cursor ThimHelpCursor;
extern Cursor ThimZoomInCursor, ThimZoomOutCursor;
extern Cursor ThimWPDefaultCursor;
extern Cursor ThimMEDefaultCursor;
extern Cursor ThimEQNDefaultCursor;
extern Cursor ThimSelectCursor;
extern Cursor ThimAddCursor;
extern Cursor ThimDeleteCursor;
extern Cursor ThimNoopCursor;
extern Cursor ThimBoltCursor;
extern Cursor ThimRightMouseCursor;

extern XSetWindowAttributes WinAtt;

extern Pixmap ScratchMap; /* screen size pixmap available as app resource */
extern Display *Dpy;      /* Resolution of DisplayWidget */
extern int PlaneMask;     /* depth of display */
extern short THIMpid;     /* pid of task */
extern GC Gc1;            /* grahics context shared by all windows */
extern XFontStruct *olCursorfont;
extern Window RootWin;
extern context_block *ActiveTasks[MAX_PIDS+1]; /* array of tasks .in_use */
extern GC axCreateGC();

static void RestoreNormalCursorPointer();

#if AXMS
extern HANDLE hInst;
#endif

/*************************************************************************
function: ZXmBindCursor - Set a variable to a standard MIT/OL cursor
*/
ZXmBindCursor (name,cursorP)
char *name;     /* from cursors.h */
Cursor *cursorP;
{
    unsigned int font_id, glyph_id, mask_id, ix, success = FALSE;
    XColor fore, back;
    mbchar temp_name[80], *p;

    /* clear the cursor */
    AxFreeCursor(*cursorP);
    *cursorP = None;

    if ( !name || !name[0] )
        return(FALSE);

    *cursorP = (Cursor)get_custom_cursor(name);
    if ( *cursorP != None )
        success = TRUE;
    else for ( ix=0; CursorMap[ix].name; ix++ )
    {
        if ( streq(name, CursorMap[ix].name) )
        {
            /* get the font and the glyph in the font */
            font_id = (unsigned int) CursorMap[ix].font_id;
            glyph_id = (unsigned int) CursorMap[ix].id;

            /* easy, it's already done */
            if ( font_id == MIT_CURSOR_FONT_ID )
            {
                success = TRUE;
#if AXMS
                if ( glyph_id == -1 )
                   { 
                   /** get this cursor by name from our resources */
                   /** first we must replace blanks in the X name */
                   /** with underscores since that is how we have encoded */
                   /** the resource names (which can not contain blanks) */
                   MBstrcpy( temp_name, name );
                   p = temp_name;
                   while ( *p != 0 )
                      {
                      if ( *p == ' ' )
                        *p = '_';
                      p++;
                      }
                   *cursorP = MS_LoadCursor( hInst, temp_name );
                   if ( cursorP == (Cursor *)NULL )
                      success = FALSE;
                   }
                else
#endif
                *cursorP = XCreateFontCursor(Dpy, glyph_id);
            }

#if !AXMS
            /* do we need to create it? */
            else if ( (!AxElfrtMode || AxOpenLook)
                    && (font_id == OL_CURSOR_FONT_ID) )
            {
                success = TRUE;
                mask_id = glyph_id + 1;
                fore.red = fore.green = fore.blue = (unsigned short) 0;
                back.red = back.green = back.blue = (unsigned short) MAX_GUN;
                fore.flags = back.flags = DoRed | DoGreen | DoBlue;
                *cursorP = XCreateGlyphCursor(Dpy, olCursorfont->fid,
                            olCursorfont->fid, glyph_id, mask_id,
                            &fore, &back);
            }
#endif
        }
    }
    return(success);
}

/*************************************************************************
function: ZXmCreateCursor - Set a variable to a image-file based cursor
*/
void ZXmCreateCursor (dat_name,msk_name, x,y, cursorP)
char *dat_name, *msk_name;
unsigned int x,y;
Cursor *cursorP;
{
    unsigned int wid, hyt, dep;
    XColor fore, back;
    XImage  msk_xP;
    int jnk, stat, x_hot, y_hot, wid_ret, hyt_ret;
    unsigned char map[768];
    Pixmap dat_map, msk_map;
    GC datGC, mskGC;
    unsigned long vmask;
    XGCValues vals;
    bitmap_struct bmap;
    bitmap_struct bmask;

    /* clear the old cursor */
    if ( *cursorP != None )
        XFreeCursor(Dpy, *cursorP);

    *cursorP = None;
    memset(&bmap, 0, sizeof(bitmap_struct));
    memset(&bmask, 0, sizeof(bitmap_struct));
    /* read the new one */
    if ( dat_name && dat_name[0] )
    {
        /* get the bits for the cursor */
        bmap.pid = THIMpid;
        bmap.display = FALSE;
        bmap.needcolor = FALSE;
        bmap.image.data = (char *)NULL;
        bmap.mask.data = (char *)NULL;
        stat = AxIMtoXImage(dat_name,dat_name, &bmap);
        if (stat != AOK )
            return;
        if ( bmap.image.depth != 1 )
        {
            if ( bmap.image.data && stat == AOK )
                free(bmap.image.data);
            if ( bmap.mask.data && stat == AOK )
                free(bmap.mask.data);
            return;
        }

        if ( x == 0 && y == 0 )
        {
            x_hot = bmap.x_spot;
            y_hot = bmap.y_spot;
        }
        else
        {
            x_hot = x;
            y_hot = y;
        }

#if AX_X11    /* don't need pixmaps on NT */
        /* create the cursor pixmap */
        fore.flags = back.flags = DoRed | DoGreen | DoBlue;
        fore.red = fore.green = fore.blue = (unsigned short) 0;
        back.red = back.green = back.blue = (unsigned short) MAX_GUN;
        vmask = GCForeground | GCBackground;
        vals.foreground = 1;
        vals.background = 0;
        dat_map = XCreatePixmap(Dpy, RootWin, bmap.image.width,
                                bmap.image.height,1);
        datGC = axCreateGC(Dpy, dat_map, DefaultScreen(Dpy), 1,
						   (XFontStruct *) 0, vmask, &vals);
        XPutImage(Dpy, dat_map, datGC, &bmap.image,
                  0, 0, 0, 0, bmap.image.width, bmap.image.height);
#endif

        /* read the bits for the mask */
        msk_map = None;
        if ( msk_name && msk_name[0] )
        {
            msk_xP.data =  NULL;
            bmask.pid =THIMpid;
            bmask.display =FALSE;
            bmask.needcolor =FALSE;
            bmask.image.data =(char *)NULL;
            bmask.mask.data =(char *)NULL;
            stat = AxIMtoXImage(msk_name,msk_name,&bmask);
            if ( bmask.image.depth != 1 )
            {
                if ( bmask.image.data && stat == AOK )
                    TaskFree( bmask.pid,bmask.image.data );
                if ( bmask.mask.data && stat == AOK )
                    TaskFree( bmask.pid,bmask.mask.data );
            }
            else if ( bmask.image.width != bmap.image.width ||
                    bmask.image.height != bmap.image.height )
            {
                if ( bmask.image.data && stat == AOK )
                    TaskFree( bmask.pid,bmask.image.data );
                if ( bmask.mask.data && stat == AOK )
                    TaskFree( bmask.pid,bmask.mask.data );
            }
            else
            {
#if AX_X11            
                /* create the cursor mask pixmap */
                msk_map = XCreatePixmap(Dpy, RootWin,
                                        bmask.image.width,
                                        bmask.image.height,
                                        1);
                mskGC = axCreateGC(Dpy, msk_map, DefaultScreen(Dpy), 1,
								   (XFontStruct *) 0, vmask, &vals);
                XPutImage(Dpy, msk_map, mskGC, &bmask.image,
                          0, 0, 0, 0, bmask.image.width, bmask.image.height);
#endif
            }
        }
        else
        {
#if AX_X11            
                /* create the cursor mask pixmap */
                msk_map = XCreatePixmap(Dpy, RootWin,
                                        bmap.image.width,
                                        bmap.image.height,
                                        1);
                mskGC = axCreateGC(Dpy, msk_map, DefaultScreen(Dpy), 1,
								   (XFontStruct *) 0, vmask, &vals);
                XPutImage(Dpy, msk_map, mskGC, &bmap.image,
                          0, 0, 0, 0, bmap.image.width, bmap.image.height);
#endif
        }

        /* create the cursor */
#if AXMS    /* see comment at XCreatePixmapCursor in x2win.c */       
        *cursorP = CreateWinCursor(&bmap.image, &bmask.image, &fore, &back,
                                                            x_hot, y_hot);
#else
        *cursorP = XCreatePixmapCursor(Dpy, dat_map,msk_map,
                                            &fore, &back, x_hot,y_hot);

        /* free resources */
        XFreePixmap(Dpy, dat_map);
        XFreeGC(Dpy, datGC);
#endif
        if ( bmap.image.data )
            TaskFree(bmap.pid, bmap.image.data);
        if ( bmap.mask.data )
            TaskFree(bmap.pid, bmap.mask.data);
        if ( msk_map != None )
        {
#if AX_X11   /* don't free pixmaps on NT, because they weren't created */
            XFreePixmap(Dpy, msk_map);
            XFreeGC(Dpy, mskGC);
#endif
            if ( bmask.image.data )
                TaskFree(bmask.pid, bmask.image.data);
            if ( bmask.mask.data )
                TaskFree(bmask.pid, bmask.mask.data);
        }
    }
}

/*************************************************************************
function: ZXmGetCursor - Return cursor for a window type
*/
Cursor ZXmGetCursor (iconid)
int iconid;
{
    int ix;

    /* sanity */
    if ( !iconid )
        return(None);

    /* find the matching cursor */
    for ( ix=0; ix<MAX_APP_CURSORS ; ix++ )
    {
        if ( AppCursors[ix].iconid == iconid )
            break;
    }
    if ( ix == MAX_APP_CURSORS )
        return(None);
    return(AppCursors[ix].curs);
}

/*************************************************************************
function: AXSetActiveCursor - Set the cursor for the widget in question
*/
int AXSetActiveCursor(w)
AXWidget w;
{
    Cursor new_cursor;

    /* sanity */
#if AXMS
    if ( !w->winid )
        return FALSE;
#else
    if ( !w->window )
        return FALSE;
#endif


    /* is the wait cursor active? */
    new_cursor = 0;
    if ( w->wait_cursor )
    {
        if ( w->wait_cursor != w->curr_cursor )
            new_cursor = w->wait_cursor;
    }

    /* is the help cursor active? */
    else if ( w->help_cursor )
    {
        if ( w->help_cursor != w->curr_cursor )
            new_cursor = w->help_cursor;
    }

    /* is the application temp cursor active? */
    else if ( w->app_cursor )
    {
        if ( w->app_cursor != w->curr_cursor )
            new_cursor = w->app_cursor;
    }

    /* is the external temp cursor active? */
    else if ( w->tmp_cursor )
    {
        if ( w->tmp_cursor != w->curr_cursor )
            new_cursor = w->tmp_cursor;
    }

    /* is the default cursor active? */
    else if ( w->def_cursor )
    {
        if ( w->def_cursor != w->curr_cursor )
            new_cursor = w->def_cursor;
    }

    /* no cursor defined on this widget - use the default */
    else if ( ThimPointerCursor )
    {
        new_cursor = ThimPointerCursor;
        w->def_cursor = new_cursor;
    }
 
    /* no cursor defined on this widget - use the default */
    else
        w->def_cursor = new_cursor = ThimPointerCursor;

    /* now set the cursor (if there is one) */
    if ( new_cursor )
    {
        w->curr_cursor = new_cursor;
		
#if AX_X11
	if ((w->flags & WFLG_IN_EDIT) && (W_DIALOG != w->type))
	{
	    return FALSE;   /* such "windows" are pixmaps during edit */
	}
#endif
		
#if AXMS
        /* we have a separate call because Windows wants a window handle
         * (w->winid), rather than a handle to a device context (w->window).
         */
		if (TaskIsJavaBased(w->taskOwner)) {
			if (w->type == W_MAIN) {
				int jwId = JavaWinIdFromWidget(w);
				JavaSetTaskCursor(jwId,
								  (w->curr_cursor==w->wait_cursor) ?
								  AX_HOURGLASS_CURSOR : AX_DEFAULT_CURSOR);
			}
		}
		else
        	AxWinDefineCursor(w->winid, w->curr_cursor);
#else
	if (w->type == W_COMBO_BOX) {
	    XDefineCursor(Dpy, ((AXWidget_combo_box)w)->edit_win.win,
			  w->curr_cursor);
	    /* we only want a text cursor in the editable window itself */
	    if (w->curr_cursor == ThimTextCursor)
		XDefineCursor(Dpy, w->window, ThimPointerCursor);
	    else
		XDefineCursor(Dpy, w->window, w->curr_cursor);
	} else if (w->type == W_ENTER) {
	    XDefineCursor(Dpy, ((AXWidget_entry)w)->edit_win.win,
			  w->curr_cursor);
	} else {
		if (TaskIsJavaBased(w->taskOwner)) {
			if (w->type == W_MAIN) {
				int jwId = JavaWinIdFromWidget(w);
				JavaSetTaskCursor(jwId,
								  (w->curr_cursor==w->wait_cursor) ?
								  AX_HOURGLASS_CURSOR : AX_DEFAULT_CURSOR);
			}
		}
		else
			XDefineCursor(Dpy, w->window, w->curr_cursor);
	}
#endif
        return TRUE;
    }
    else
        return FALSE;
}

/*************************************************************************
function: AXDefineWaitCursor - Set the wait cursor (window & widget)
*/
void AXDefineWaitCursor(w,ptr)
AXWidget w;
Cursor ptr;
{
    /* sanity */
    if ( !w->window )
        return;

 
    /* if the showing hourglass flag is set, leave this alone!!! */
    if (!ptr && (TaskInfo[w->taskOwner].showing_hourglass
        || TaskInfo[w->taskOwner].slaved_hourglass))
        return;

    if ( w->wait_cursor != ptr )
    {
        w->wait_cursor = ptr;
        AXSetActiveCursor(w);
    }
}

/*************************************************************************
function: AXDefineHelpCursor - Set the help cursor (window & widget)
*/
void AXDefineHelpCursor(w,ptr)
AXWidget w;
Cursor ptr;
{
    /* sanity */
    if ( !w->window )
        return;

    if ( w->help_cursor != ptr )
    {
        w->help_cursor = ptr;
        AXSetActiveCursor(w);
    }
}

/*************************************************************************
function: AXDefineTmpCursor - Set the tmp cursor (window & widget)
*/
void AXDefineTmpCursor(w,ptr)
AXWidget w;
Cursor ptr;
{
    /* sanity */
    if ( !w->window )
        return;

    if ( w->tmp_cursor != ptr )
    {
        w->tmp_cursor = ptr;
        AXSetActiveCursor(w);
    }
}

/*************************************************************************
function: AXDefineAppCursor - Set the application cursor (window & widget)
*/
void AXDefineAppCursor(w,ptr)
AXWidget w;
Cursor ptr;
{
    /* sanity */
    if ( !w->window )
        return;

    if ( w->app_cursor != ptr )
    {
        w->app_cursor = ptr;
        AXSetActiveCursor(w);
    }
}

/*************************************************************************
function: AXDefineDefaultCursor - Set the default cursor (window & widget)
*/
void AXDefineDefaultCursor(w,ptr)
AXWidget w;
Cursor ptr;
{
    /* sanity */
    if ( !w->window )
        return;

    if ( w->def_cursor != ptr )
    {
        w->def_cursor = ptr;
        AXSetActiveCursor(w);
    }
}

/*************************************************************************
function: CursorAllWindows - Set the default cursor on all windows
*/
static void CursorAllWindows (curs, code, widgets_too)
Cursor curs;
int code, widgets_too;
{
    register AXWidget wp;

    /* sanity */
    if ( !ATMinitCOMPLETE )
        return;

    for ( wp=WindowList; wp; wp=wp->nextWin )
    {
        if ( !wp->window )
            continue;       /* possible? */

        if ( !widgets_too ) {
            if ( (wp->type != W_DIALOG) && (wp->type != W_WORK) &&
                 (wp->type != W_ENTER) && (wp->type != W_LIST) &&
                 (wp->type != W_EDITBOX) )
                continue;
        }

        switch(code)
        {
            default:
            case 0: AXDefineDefaultCursor(wp,curs); break;
            case 1: AXDefineTmpCursor(wp,curs);     break;
            case 2: AXDefineAppCursor(wp,curs);     break;
            case 3: AXDefineHelpCursor(wp,curs);    break;
            case 4: AXDefineWaitCursor(wp,curs);    break;
        }
    }
    XFlush(Dpy);
}

static void HgTaskWindows (pid,slaved_mode)
register int pid; /* task's id */
register int slaved_mode;   /* Used when task wating on blocked task */
{
    register AXWidget wp;
    win_list_t *winP;
    register int ix;
    Cursor cursor;
	int found = FALSE;
	
    /* set the busy cursor */
	for ( wp=WindowList; wp; wp=wp->nextWin ) {
		if ( !wp->window || (wp->taskOwner != pid) )
			continue;
		found = TRUE;
		AXDefineWaitCursor(wp,ThimBusyCursor);
	}
	if (!found && TaskInfo[pid].TaskUsesJavaDisplay) {
		JavaSetTaskCursor(TaskInfo[pid].uid,AX_HOURGLASS_CURSOR);
	}
	XFlush(Dpy);

    /* set the task wait flag */
    if ( slaved_mode )
        TaskInfo[pid].slaved_hourglass = TRUE;
    else
        TaskInfo[pid].showing_hourglass = TRUE;
}

/*************************************************************************
function: AxmHourglassTaskWindows - Set the hour on task's windows
*/
void AxmHourglassTaskWindows (taskid)
register int taskid; /* task's uid */
{
    register int pid, ix;

    /* sanity */
    if ( !ThimBusyCursor )      /* init complete? */
        return;
    pid = ThimPidFromUID(taskid);
    if ( !pid )     /* bogus call */
        return;

    HgTaskWindows(pid,FALSE);               /* make wins HG */
    for ( ix = 0; ActiveTasks[ix]; ix++ ) { /* make pended tasks wins HG */
        if ( ActiveTasks[ix]->ipc_response_waiting == pid )
            HgTaskWindows(ThimPidFromUID(ActiveTasks[ix]->uid),TRUE);
    }
}

AxmReleaseSlavedCursors(pid)
int pid;
{
    int ix;

    for ( ix = 0; ActiveTasks[ix]; ix++ ) { /* make pended tasks wins HG */
        if ( ActiveTasks[ix]->ipc_response_waiting == pid )
            RestoreNormalCursorPointer(ThimPidFromUID(ActiveTasks[ix]->uid),TRUE);
    }
}

/*************************************************************************
function: AXmRestoreNormalCursorPointer - Reload current cursors
*/
void AXmRestoreNormalCursorPointer (taskid)
int taskid; /* task's uid or 0 for all windows */
{
    register int pid;

    if ( taskid == 0 )
        pid = taskid;
    else {
        pid = ThimPidFromUID(taskid);
        if ( !pid )
            return;
    }
    AxmReleaseSlavedCursors(pid);
    RestoreNormalCursorPointer(pid,FALSE);
}

static void RestoreNormalCursorPointer (pid,from_slaved)
int pid; /* task's dex id or 0 for all windows */
int from_slaved;
{
    register int ix, found = FALSE;
    register AXWidget wp;
    win_list_t *winP;
    Cursor cursor;

    /* get the pid */
    if ( pid != 0 ) {
        if ( from_slaved ) {
            if ( !TaskInfo[pid].slaved_hourglass )
                return;
            TaskInfo[pid].slaved_hourglass = FALSE;
            if ( TaskInfo[pid].showing_hourglass )
                return;     /* just toss "slaved" nature */
        }
        else if ( TaskInfo[pid].showing_hourglass ) {
            TaskInfo[pid].showing_hourglass = FALSE;
            TaskInfo[pid].slaved_hourglass = FALSE;
        }
        else
            return;
    }
/**	
	if (pid && TaskInfo[pid].TaskUsesJavaDisplay) {
		JavaSetTaskCursor(TaskInfo[pid].uid,AX_DEFAULT_CURSOR);
		return;
	}
**/	
    if ( !ATMinitCOMPLETE )
        return;

    /* restore default cursor (obsolete) */
    WinAtt.cursor = ThimPointerCursor;

    /* set default cursors for each window */
    for ( wp=WindowList; wp; wp=wp->nextWin )
    {
        /* sanity check */
        if ( !wp->window )
            continue;
        if ( pid && (wp->taskOwner != pid) )
            continue;
        if ( !pid && (TaskInfo[wp->taskOwner].showing_hourglass
                    || TaskInfo[wp->taskOwner].slaved_hourglass) )
            continue;

        /* main windows (not accessable via the ELF) */
        if ( (wp->type == W_WORK) && 
              wp->parent &&
              (wp->parent) && (wp->parent->style) &&
              (wp->parent->type == W_MAIN) )
        {
            /* check if  the task is a apge  */
            /* the GE sets the defaults according to current tool */
            if ( AxIsGraphicsTask(pid) == FALSE )
                found = FALSE;
            else
                found = !GESetPointerFromPalette(wp->parent);

            /* set the standard defaults by icon id */
            if ( !found )
            {
                for ( ix=0; ix<MAX_APP_CURSORS; ix++ )
                {
                    if ( AppCursors[ix].iconid == wp->parent->style )
                        break;
                }
                if ( ix == MAX_APP_CURSORS )
                    AXDefineDefaultCursor(wp,ThimPointerCursor);
                else
                    AXDefineDefaultCursor(wp,AppCursors[ix].curs);
            }
        }

        /* entry fields and edit boxes get a text cursor */
        else if ( wp->type == W_ENTER || wp->type == W_EDITBOX )
            AXDefineDefaultCursor(wp, ThimTextCursor);

        /* everything else get's the def pointer cursor */
        else
            AXDefineDefaultCursor(wp,ThimPointerCursor);

        /* disable temp and wait cursors */
        AXDefineTmpCursor(wp,  0);
        AXDefineWaitCursor(wp, 0);
		found = TRUE;
    }

	if (!found && pid && TaskInfo[pid].TaskUsesJavaDisplay) {
		JavaSetTaskCursor(TaskInfo[pid].uid,AX_DEFAULT_CURSOR);
		return;
	}

    /* ...and do it NOW!!!! */
    XFlush(Dpy);
}

/*************************************************************************
function: ElfbMapCursor - Bind a default cursor to a window type (by icon)
*/
void ElfbMapCursor (iconid, curs)
int iconid;
char *curs;
{
    int ix;

    /* sanity */
    if ( !iconid )
        return;

    /* find the current icon id of the window in question */
    for ( ix=0; (ix<MAX_APP_CURSORS) && AppCursors[ix].iconid; ix++ )
    {
        if ( AppCursors[ix].iconid == iconid )
            break;
    }

    /* if there are too many cursors, quietly fail */
    if ( ix < MAX_APP_CURSORS )
    {
        AppCursors[ix].iconid = iconid;
        ZXmBindCursor(curs, &AppCursors[ix].curs);
    }

    /* set defaults and clear temp and wait flags */
    AXmRestoreNormalCursorPointer(0);
}

/*************************************************************************
function: ElfbMapDefaultCursor - Bind a default cursor to a window type (by icon)
NOTE: ElfbMapCursor maps the _TEXT_ pointer for the WP/ME/EQN, this maps
the _DEFAULT_ pointer.
*/
void ElfbMapDefaultCursor (iconid, curs)
int iconid;
char *curs;
{
    /* NOTE: The calling ELF macro will use the defines in
     * elf_apps.elf/app_ids_.am for the iconid.  This file
     * uses aiconids.h.  Make sure that the defines in the
     * two files match. */

    switch ( iconid ) {
        case IC_DC:
            ZXmBindCursor(curs, &ThimWPDefaultCursor);
            break;
        case IC_CE:
            ZXmBindCursor(curs, &ThimMEDefaultCursor);
            break;
        case IC_EQN:
            ZXmBindCursor(curs, &ThimEQNDefaultCursor);
            break;
        case IC_HTML:
			/* !!0 use WP cursor for now */
            ZXmBindCursor(curs, &ThimWPDefaultCursor);
            break;
        default:
            return;
    }

    /* set defaults and clear temp and wait flags */
    AXmRestoreNormalCursorPointer(0);
}

/*************************************************************************
function: XmPutBusyCursor - Post the hourglass to ALL windows
Use profile busy cursor if cursor_name is NULL or invalid
*/
void XmPutBusyCursor (cursor_name)
char *cursor_name;  /* from cursors.h */
{
    Cursor busyCursor = None;
    int did_cursor;

	if( JavaMode) { 
		/* even if InSysHourglassMode is set, still
		   set the hourglass, as it may be for a different client */
		JavaSetTaskCursor(0,AX_HOURGLASS_CURSOR);
	}
    else if ( !InSysHourglassMode && ATMinitCOMPLETE )
    {
        /* get the busy cursor */
        did_cursor = ZXmBindCursor(cursor_name, &busyCursor);
        if ( did_cursor == FALSE )
            busyCursor = ThimBusyCursor;

        /* set the cursor on all windows */
        CursorAllWindows(busyCursor, 4, TRUE);
        /* clear the cursor */
        if ( did_cursor )
            AxFreeCursor(busyCursor);
        XSync(Dpy,FALSE);
    }
	InSysHourglassMode = TRUE;
}

/*************************************************************************
function: XmPutHourglass - Post the hourglass to ALL windows
*/
void XmPutHourglass()
{
    XmPutBusyCursor(NULL);
}

/*************************************************************************
function: XmPutSomeHourglass - Put hourglass around locus window
Currently hourglass's the entire world ??1
*/
void XmPutSomeHourglass ()
{
    XmPutBusyCursor(NULL);
}

/*************************************************************************
function: XmClrHourglass - Remove hourglass
*/
void XmClrHourglass ()
{
    register AXWidget wp;
	if( InSysHourglassMode && JavaMode) {
		/** To optimize no. of messages, clear hourglass for all tasks
		  * and then set hourglass for the tasks that are still
		  * showing hourglass
		  **/
		context_block *tInfo;
		int ix;
		JavaSetTaskCursor(0,AX_DEFAULT_CURSOR);
		/* clear wait_cursor value for all windows, as we removed it 
		   for all */
		for ( wp=WindowList; wp; wp=wp->nextWin ) {
			wp->wait_cursor = 0;
		}
		for (ix=0; ActiveTasks[ix]; ix++) {
			tInfo = ActiveTasks[ix];
			if ((tInfo->showing_hourglass || tInfo->slaved_hourglass) &&
				tInfo->TaskUsesJavaDisplay)
				HgTaskWindows(ThimPidFromUID(tInfo->uid),
							  tInfo->slaved_hourglass);				
		}
	}	
    else if ( InSysHourglassMode && ATMinitCOMPLETE )
    {
        CursorAllWindows(0, 4, TRUE);
    }
	InSysHourglassMode = FALSE;
}


/*************************************************************************
function: ElfbStartLearn - Turn on learn mode (set temp cursor)
*/
void ElfbStartLearn ()
{
    register AXWidget wp;
    register AXWidget_main parent;
    char *UPbyCode();

    /* already in learn mode */
    if ( ElfInLearnMode )
        return;

    XmClrHourglass();   /* prevents lockup during recording */
    for ( wp=WindowList; wp; wp=wp->nextWin )
    {
        if ( !wp->window )
            continue;       /* possible? */

        parent = wp->parent;
        while ( parent )
        {
            if ( parent->type == W_MAIN )
                break;
            parent = parent->parent;
        }
        if ( (parent && parent->disable_record) || (wp->style == IC_DC) )
            AXDefineTmpCursor(wp,ThimNoRecordCursor);
        else
            AXDefineTmpCursor(wp,ThimRecordCursor);
    }
    WinAtt.cursor = ThimRecordCursor;
    ElfInLearnMode = BTRUE;
}

/*************************************************************************
function: ElfbEndLearn - Turn off learn mode (clear temp cursor)
*/
void ElfbEndLearn ()
{
    /* not in learn mode */
    if ( !ElfInLearnMode )
        return;

    XmClrHourglass();   /* prevents lockup during recording */
    WinAtt.cursor = ThimPointerCursor;
    ElfInLearnMode = FALSE;
    CursorAllWindows(0, 1, TRUE);
}


/*************************************************************************
function: ElfbSetHelpMode - Set the THIMhelpMode variable
*/
void ElfbSetHelpMode (val)
{
    if ( val )
    {
        SystemMessage(XLT("?tIn HELP mode, select a command or function key"));
        XmClrHourglass();
        CursorAllWindows(ThimHelpCursor, 3, TRUE);
        THIMhelpMode = TRUE;
    }
    else
    {
        SystemMessage("");
        THIMhelpMode = FALSE;
        CursorAllWindows(0, 3, TRUE);
    }
    UpdateAllToolBars();
}

/*************************************************************************
function: AxSetCursorShape - Set the default GE cursor according to index
*/
AxSetCursorShape(w, cursor_ix, tool_ix)
AXWidget_main w;
int cursor_ix;
int tool_ix;
{
	int use_app_cursor;
    Cursor ptr_cursor;

	if (w)
	{
	    use_app_cursor = TRUE;
	    switch ( cursor_ix )
	    {
	    case 0x0010: ptr_cursor = ThimSelectCursor;      break;
	    case 0x0020: ptr_cursor = ThimAddCursor;         break;
	    case 0x0030: ptr_cursor = ThimDeleteCursor;      break;
	    case 0x0040: ptr_cursor = ThimNoopCursor;        break;
	    case 4:      ptr_cursor = ThimUpDownCursor;      break;
	    case 1:      ptr_cursor = ThimUpDownCursor;      break;
            case 2:      ptr_cursor = ThimLeftRightCursor;   break;
            case 8:      ptr_cursor = ThimLeftRightCursor;   break;
            case 6:      ptr_cursor = ThimTopRightCursor;    break;
            case 12:     ptr_cursor = ThimTopLeftCursor;     break;
            case 3:      ptr_cursor = ThimBottomRightCursor; break;
            case 9:      ptr_cursor = ThimBottomLeftCursor;  break;
            case 15:     ptr_cursor = ThimFleurCursor;       break;
            case 0x8000: ptr_cursor = ThimTextCursor;        break;
            default:
                use_app_cursor = FALSE;
                if ( tool_ix == 0 || tool_ix == 13 ) 
                    ptr_cursor = ThimPointerCursor;
                else if ( tool_ix == 1 )
                    ptr_cursor = ThimTextCursor;
                else if ( tool_ix == 12 )
                    ptr_cursor = ThimZoomInCursor;
                else 
                    ptr_cursor = ZXmGetCursor(IC_GE);
            break;
        }

	    if ( use_app_cursor )
		AXDefineAppCursor(w->workWindow, ptr_cursor);
	    else
	    {
		AXDefineDefaultCursor(w->workWindow, ptr_cursor);
		AXDefineAppCursor(w->workWindow, 0);
	    }
	}
}

/*************************************************************************
function: AxSetRotateCursor - Set the default GE cursor for rotation
*/
AxSetRotateCursor(w)
AXWidget_main w;
{

	if (w)
		AXDefineAppCursor(w->workWindow, ThimRotateCursor);
}

/*************************************************************************
function: 
*/
void VerifyCursor(old_curs, new_curs)
Cursor old_curs;
Cursor new_curs;
{
    register AXWidget w;

    for ( w=WindowList; w; w=w->nextWin )
    {
        if ( !w->window )
            continue;       /* possible? */

        if ( w->def_cursor == old_curs )
        {
            w->def_cursor = new_curs;
            AXSetActiveCursor(w);
        }
        if ( w->tmp_cursor == old_curs )
        {
            w->tmp_cursor = new_curs;
            AXSetActiveCursor(w);
        }
        if ( w->app_cursor == old_curs )
        {
            w->app_cursor = new_curs;
            AXSetActiveCursor(w);
        }
        if ( w->wait_cursor == old_curs )
        {
            w->wait_cursor = new_curs;
            AXSetActiveCursor(w);
        }
        if ( w->curr_cursor == old_curs )
        {
            w->curr_cursor = new_curs;
            AXSetActiveCursor(w);
        }
    }
    XFlush(Dpy);
}
