/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xmevent.c
   Applix ASC version: 11
   Date: Mon Nov 16 13:53:37 1998
*/
#include <aos.h>
#include <xmlib.h>

static void tossEventRec(), copyEventHandler(), setEventMask();
static int matchesEvent();
static eventRec *newEventRec();

#define NonMaskableMask ((unsigned long)0x80000000L)

    /* elf will refer to the event handlers by name, not symbol */
#define NUMPROCS 2
static char *eventProcNames[] = {
    NXLT("dlgEventProc"),
    NXLT("dlgChildEventProc"),
};
void dlgEventProc(), dlgChildEventProc();
static void (*eventProcs[])() = {
    dlgEventProc,
    dlgChildEventProc,
};



/* : InsertEventHandler - insert event handler onto the
 * passed widget's local event_table */
int
InsertEventHandler(w, eventMask, nonmaskable, procName,
                        clientData, listPos)
AXWidget w;
unsigned long eventMask;
int nonmaskable; /* unused here, as if always False */
char *procName;
char *clientData; /* unused here, as if always null */
int listPos; /* unused here, as if always XtListHead */
{
    int ix;
    void (*tproc)() = (void(*)()) NULL;
    for(ix = 0; ix < NUMPROCS; ix++) {
        if (strcmp(procName, eventProcNames[ix]) == 0) {
            tproc = eventProcs[ix];
        }
    }
    if (tproc) {
        ZXtInsertEventHandler(w, eventMask, nonmaskable,
            tproc, clientData, listPos);
    }
    return TRUE;
}

int
ElfbRemoveEventHandlers(w)
AXWidget w;
{
    if (w) {
        xmRemoveAllEventHandlers(w);
    }
    return TRUE;
}

void
ZXtInsertEventHandler(w, evmask, nonmaskable, proc, clientData, pos)
AXWidget w;
unsigned long evmask;
int nonmaskable; /* unused here, as if always False */
void (*proc)();
char *clientData;
int pos; /* unused here, as if always XtListHead */
{
    eventRec *ep;

    if (!w || !proc)
        return;

        /* check whether matching handler is already here */
    ep = w->event_table;
    while(ep) {
        if (ep->proc == proc && ep->nonmaskable == nonmaskable) {
            setEventMask(evmask, &ep->eventMask);
            return;
        }
        ep = ep->next;
    }
        /* need to insert a new list item */
    ep = newEventRec();
    ep->proc = proc;
    ep->eventMask = evmask;
#if 0
    ep->clientData = ?????
#endif
    ep->nonmaskable = nonmaskable;
    ep->next = w->event_table;
    w->event_table = ep;
}

void
ZXtRemoveEventHandler(w, evmask, nonmaskable, proc, clientData, pos)
AXWidget w;
unsigned long evmask;
int nonmaskable; /* unused here, as if always False */
void (*proc)();
char *clientData;
int pos; /* unused here, as if always XtListHead */
{
    eventRec *ep, *lastp;

    if (!w || !proc)
        return;
    
    ep = w->event_table;
    lastp = (eventRec *) NULL;
    while(ep) {
        if (ep->proc == proc && ep->nonmaskable == nonmaskable) {
                /* toss this item */
            if (lastp == (eventRec *) NULL) {
                    /* matching item was 1st on list */
                w->event_table = ep->next;
            }
            else {
                lastp->next = ep->next;
            }
            tossEventRec(ep);
            return;
        }
        lastp = ep;
        ep = ep->next;
    }
}

void
xmRemoveAllEventHandlers(w)
AXWidget w;
{
    eventRec *ep, *nep;

    if (!w || !w->event_table)
        return;
    ep = w->event_table;
    while(ep) {
        nep = ep->next;
        tossEventRec(ep);
        ep = nep;
    }
    w->event_table = (eventRec *) NULL;
}

/* : localEventDispatch - check local event dispatch table,
 * call any procs that match the passed event type, then
 * return whether the event was 'consumed' by the proc */
int
localEventDispatch(w, event)
AXWidget w;
XEvent *event;
{
    eventRec *ep;
    int consumed;

    if (!w || !w->event_table || !event)
        return FALSE;
    ep = w->event_table;
    while(ep) {
        if (matchesEvent(ep, event->type)) {
            (*ep->proc)(w, ep->clientData, event, &consumed);
            return(consumed);
        }
        ep = ep->next;
    }
    return FALSE;
}

    /* this is used from xmlib.c:AxXtSetValues, when event handlers
     * are specified on a dialog elf array, as opposed to calling
     * InsertEventHandler directly for an existing widget */
int
setEventHandlers(w, list)
AXWidget w;
eventHandlerList *list;
{
    int ix;
    if (!list || !list->eventProcNames || !list->eventMasks)
        return FALSE;
    if (list->numProcs <= 0)
        return FALSE;
    for (ix = 0; ix < list->numProcs; ix++) {
        InsertEventHandler(w, list->eventMasks[ix], FALSE,
                list->eventProcNames[ix], (char *) NULL, 0);
    }
}

static void
tossEventRec(ep)
eventRec *ep;
{
    /* clientData ??? */
    free((char *) ep);
}

void
copyEventHandlerList(from, to)
eventRec *from, **to;
{
    eventRec *fh, *tmp, **th, *newList;

    if (!from || !to)
        return;
            /* toss any list currently on *to */
    tmp = *to;
    while(tmp) {
        fh = tmp->next;
        tossEventRec(tmp);
        tmp = fh;
    }
    newList = (eventRec *) NULL;
    for (fh = from, th = &newList; fh; fh = fh->next, th = &((*th)->next)) {
        copyEventHandler(fh, th);
    }
    *to = newList;
}

static void
copyEventHandler(from, to)
eventRec *from, **to;
{
    eventRec *ep;
    if (!from || !to)
        return;
            /* toss anything currently on *to */
    if (*to) {
        tossEventRec(*to);
    }
    *to = newEventRec();
    (*to)->proc = from->proc;
    (*to)->eventMask = from->eventMask;
    (*to)->clientData = from->clientData;
    (*to)->nonmaskable = from->nonmaskable;
}

    /* make sure that the bits on from are also on *to */
static void
setEventMask(from, to)
unsigned long from, *to;
{
    int i;
    unsigned long m;
    for(i = 0, m = 1; i < 32; i++, m <<= 1 ) {
        if (from & m) {
            *to |= m;
        }
    }
}

    /* adapted from Xt's Event.c */
static unsigned long masks[] = {
    0,              /* Error, should never see  */
    0,              /* Reply, should never see  */
    KeyPressMask,           /* KeyPress         */
    KeyReleaseMask,         /* KeyRelease       */
    ButtonPressMask,        /* ButtonPress      */
    ButtonReleaseMask,      /* ButtonRelease        */
    PointerMotionMask       /* MotionNotify     */
        | Button1MotionMask
        | Button2MotionMask
        | Button3MotionMask
        | Button4MotionMask
        | Button5MotionMask
        | ButtonMotionMask,
    EnterWindowMask,        /* EnterNotify      */
    LeaveWindowMask,        /* LeaveNotify      */
    FocusChangeMask,        /* FocusIn          */
    FocusChangeMask,        /* FocusOut         */
    KeymapStateMask,        /* KeymapNotify     */
    ExposureMask,           /* Expose           */
    NonMaskableMask,        /* GraphicsExpose, in GC    */
    NonMaskableMask,        /* NoExpose, in GC      */
    VisibilityChangeMask,       /* VisibilityNotify     */
    SubstructureNotifyMask,     /* CreateNotify     */
    StructureNotifyMask     /* DestroyNotify        */
        | SubstructureNotifyMask,
    StructureNotifyMask     /* UnmapNotify      */
        | SubstructureNotifyMask,
    StructureNotifyMask     /* MapNotify        */
        | SubstructureNotifyMask,
    SubstructureRedirectMask,   /* MapRequest       */
    StructureNotifyMask     /* ReparentNotify       */
        | SubstructureNotifyMask,
    StructureNotifyMask     /* ConfigureNotify      */
        | SubstructureNotifyMask,
    SubstructureRedirectMask,   /* ConfigureRequest     */
    StructureNotifyMask     /* GravityNotify        */
        | SubstructureNotifyMask,
    ResizeRedirectMask,     /* ResizeRequest        */
    StructureNotifyMask     /* CirculateNotify      */
        | SubstructureNotifyMask,
    SubstructureRedirectMask,   /* CirculateRequest     */
    PropertyChangeMask,     /* PropertyNotify       */
    NonMaskableMask,        /* SelectionClear       */
    NonMaskableMask,        /* SelectionRequest     */
    NonMaskableMask,        /* SelectionNotify      */
    ColormapChangeMask,     /* ColormapNotify       */
    NonMaskableMask,        /* ClientMessage        */
    NonMaskableMask         /* MappingNotify        */
};

    /* return whether evMask contains evType */
static int
matchesEvent(ep, evType)
eventRec *ep;
int evType;
{
    evType &= 0x7f; /* SendEvent events have 8th bit set */
    if (masks[evType] == NonMaskableMask) {
        return (ep->nonmaskable);
    }
    return (masks[evType] & ep->eventMask);
}

static eventRec *
newEventRec()
{
    eventRec *ep;
    ep = (eventRec *) malloc (sizeof(eventRec));
    memset((char *) ep, 0, sizeof(eventRec));
    return (ep);
}

