/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xmgerulr.c
   Applix ASC version: 30
   Date: Tue Apr 20 21:12:45 1999
*/
#include <basicwin.h>
#include <ax_xm.h>
#include <aos.h>
#include <axtext.h>
#include <xmlib.h>
#include <elxwins.h>

#include <atmwin.h> 
#include <atmattr.h>
#include <parint.h>

#if AX_X11
#include <X11/keysym.h>
#endif

#include <wchars.h>
#include <xmpixel.h>
#include <ctype.h>
#include <eldata.h>
#include <elf_env.h>
#include <dwt2xm.h>

#define GE_YRULER_WIDTH 32    /* cribbed from xmlib.c */
#define GE_XRULER_HEIGHT 32

#define LABELFONT  NXLT(axDefaultSerifFont)
#define LABELSIZE  120000
#define LABELWEIGHT 1

typedef struct tickmark_t {
    unsigned int labelled : 1;
    unsigned int display : 1;
    int ticklen;
    r32 bbox;
    int value;   /* in mils */
       /* fValue is the floating-point parallel of value;
          it should really be the one value used, with all
          the grid/ruler painting and user interface replumbed
          for floating point; in the current integer method,
          fValue serves only to generate reasonable label vals */
    float fValue;
    char labelstr[MAX_RULER_UNIT_NAME];
} tickMark;

extern int AxOpenLook, ScreenRes;
extern Display *Dpy;
extern Window HostWin;
extern GC ForeGC, BackGC, SunGC, ShadowGC;
#if AX_X11
extern long WinMask;
extern XSetWindowAttributes WinAtt;
#endif

#include <xmgerulr.p>


/************************************************************************ 
 * : measureGrids - convert the unitSize (dots per unit) and
 *                  gridSize (dots per grid) to:
 *                  units per grid and grids per unit */
static void
measureGrids(int unitSize, int gridSize,
             int *units_per_grid,
             int *grids_per_unit,
             int *n_gaps, int **gap)
{
	int *tgap, n, upg, gpu;

	/* determine grid counts */
	if ( unitSize < gridSize ) /* a grid is larger than a unit */
	{
		upg = gridSize / unitSize;
		gpu = 1;
	}
	else /* round up grids per unit */
	{
		upg = 1;
		gpu = (unitSize%gridSize) ? unitSize/gridSize+1 : unitSize/gridSize;
	}

	/* determine grid positions */
	if ( gpu >= 16 )
	{
		gpu = 16;
		n = 5;
		tgap = (int *)malloc(n * sizeof(int));
		tgap[0] = unitSize/16;
		tgap[1] = unitSize/8;
		tgap[2] = unitSize/4;
		tgap[3] = unitSize/2;
		tgap[4] = unitSize/1;
	}
	else if ( gpu >= 12 )
	{
		gpu = 12;
		n = 3;
		tgap = (int *)malloc(n * sizeof(int));
		tgap[0] = unitSize/12;
		tgap[1] = unitSize/2;
		tgap[2] = unitSize/1;
	}
	else if ( gpu >= 10 )
	{
		gpu = 10;
		n = 3;
		tgap = (int *)malloc(n * sizeof(int));
		tgap[0] = unitSize/10;
		tgap[1] = unitSize/2;
		tgap[2] = unitSize/1;
	}
	else if ( gpu >= 8 )
	{
		gpu = 8;
		n = 4;
		tgap = (int *)malloc(n * sizeof(int));
		tgap[0] = unitSize/8;
		tgap[1] = unitSize/4;
		tgap[2] = unitSize/2;
		tgap[3] = unitSize/1;
	}
	else if ( gpu >= 4 )
	{
		gpu = 4;
		n = 3;
		tgap = (int *)malloc(n * sizeof(int));
		tgap[0] = unitSize/4;
		tgap[1] = unitSize/2;
		tgap[2] = unitSize/1;
	}
	else if ( gpu >= 2 )
	{
		gpu = 2;
		n = 2;
		tgap = (int *)malloc(n * sizeof(int));
		tgap[0] = unitSize/2;
		tgap[1] = unitSize/1;
	}
	else if ( gpu >= 1 )
	{
		gpu = 1;
		n = 1;
		tgap = (int *)malloc(n * sizeof(int));
		tgap[0] = gridSize;
	}

	/* done */
	*units_per_grid = upg;
	*grids_per_unit = gpu;
	*n_gaps = n;
	*gap = tgap;
}

static void
initRulerLabel(float value,  /* in: tickmark value in mils */
               float unit,   /* in: unit size in mils */
               float gridf,  /* in: unit size in mils */
               int unitPrec, /* in: unit precision */
               char *lstr,   /* in/out: MAX_RULER_UNIT_NAME size */
               int units_per_grid)
{
    char *cp;
    int lastix, ix, tmp;
    char scratch[1024]; /* paranoid about sprintf */
    float dval;
	
	if ( units_per_grid > 1 ) /* unitsize is less than gridsize */
	{
		/* sanity check */
		if ( unitPrec <= 0 || unitPrec > 3 )
			unitPrec = 1;
		if ( unitPrec == 1)
			value *= 10;
		else if ( unitPrec == 2)
			value *= 100;
		else
			value *= 1000;
		value *= gridf;
		tmp = (int)value;
		value = (float) tmp;
		dval = value / gridf;
		dval = dval / unit;
		if ( unitPrec == 1)
		{
			dval = dval / 10.0;
			sprintf(scratch, "%.1f", dval);  /* convert to ascii */
		}
		else if ( unitPrec == 2)
		{
			dval = dval / 100.0;
			sprintf(scratch, "%.2f", dval);  /* convert to ascii */
		}
		else if (unitPrec == 3)
		{
			dval = dval / 1000.0;
			sprintf(scratch, "%.3f", dval);  /* convert to ascii */
		}
		lastix = strlen(scratch);    /* trim trailing 0's */
		cp = &scratch[lastix];
		while(*cp == '0') {
			cp--;
		}
		*cp = '\0';
	}
	else /* unitsize is greater than gridsize */
	{
		dval = value / unit;
		if ( dval < 0 )
			dval -= 0.5;
		else
			dval += 0.5;
		if (dval == -0.0)
			dval = 0.0;
		tmp = (int) dval;
		dval = (float) tmp;

		if (dval == -0.0)
			dval = 0.0;

        /* trunc the value in order to filter
           stuff like "16.062", when "16" is desired.
           This chop defeats any attempt to display labels
           for fractional-unit tickmarks.  A solution might
           involve knowing the fraction (1/n) in here, then:
           a) magnify float by n, b) trunc, c) copy to float,
           d) divide by n.  For now, just go whole-units */

		sprintf(scratch, "%.4f", dval);  /* convert to ascii */
		lastix = strlen(scratch) - 1;    /* trim trailing 0's */
		cp = &scratch[lastix];
		ix = lastix;
		while(*cp == '0' && *cp != '.' && ix >= 0) {
			ix--;
			cp--;
		}
		*cp = '\0';

		if (ix > MAX_RULER_UNIT_NAME - 1) {
				/* need to trunc this string however cruelly */
			ix = MAX_RULER_UNIT_NAME - 1;
			scratch[ix] = '\0';
			cp = &scratch[ix + 1];
		}

		--cp;
		if (*cp == '.')
			*cp = '\0';  /* trim decimal point if endmost */

	}
    strcpy(lstr, scratch);  /* copy to caller mem */
}

/************************************************************************/
static void
rulerLabelSizes (tickMark *ticks,  /* in: array describing ruler tickmarks */
                 int numTicks,
                 int *maxWidth,    /* out: report widest width found */
                 int *maxHeight)   /* out: report widest height found */
{
    int ix, jx, len, maxwid, maxht;
    unsigned int flags;
    axtext txt;
    textAttrs atts;
    tx_char *txchars;
    char *err;

    maxwid = maxht = 0;
    atts.fontFamily = LABELFONT;
    atts.pointsize = LABELSIZE;
    atts.weight = LABELWEIGHT;
    flags = (unsigned int) (TXpointsize | TXfontFamily | TXweight);

    for (ix = 0; ix < numTicks; ix++) {
        if (ticks[ix].labelled && ticks[ix].labelstr) {

                /* create the labelstr text */
            memset((char *) &txt, 0, sizeof(axtext));
            len = strlen(ticks[ix].labelstr);
            txchars = (tx_char *) malloc
                ((len + 1) * sizeof(tx_char));
            for (jx = 0; jx < len; jx++) {
                txchars[jx] = (tx_char) ticks[ix].labelstr[jx];
            }
            setDefaultTextAttrs(&txt, &atts, flags);
            insChars(&txt, txchars, 0, len);

                /* measure the labelstr text */
            if (err = getOverallBounder(&txt, 0, -1,
                                        &ticks[ix].bbox)) {
                freeErrString(err);
            } else {
                if (ticks[ix].bbox.width > maxwid)
                    maxwid = ticks[ix].bbox.width;
                if (ticks[ix].bbox.height > maxht)
                    maxht = ticks[ix].bbox.height;
            }

            daxtext(&txt);
            free((char *) txchars);
                
        }
    }
    *maxWidth = maxwid;
    *maxHeight = maxht;
}

/************************************************************************/
/* : axComputeTickMarks - produce an array of tickMark
 * structs for use in painting the ruler widget.  Caller
 * must free the returned array
 */
static tickMark *
axComputeTickMarks (int horiz,    /* in: bool; whether for horizontal ruler */
                    int scrollPos,/* in: scroll position in mils */
                    int scrollPer,/* in: scrollrange percent on display, mils */
                    int zeroPt,   /* in: grid origin coord in mils */
                    int zoom,     /* in: zoom factor, reference val 1000 */
                    int unitSize, /* in: size of ruler unit in mils */
                    int gridSize, /* in: size of grid cell in mils */
                    int unitPrec, /* in: precision of grid  */
                    int *numMarks)/* out: report size of returned tickMark array */
{
    int mil_gap, ix, tMils, n_ticks;
    int t0Mils;
    int tmp, *gap, n_gaps, tx, ty, jx, sz;
    int ticklen, maxWidth, maxSize, droppedLabels;
    int majorWidth, foundWidth, u0Mils, grids_per_unit, units_per_grid;
	int val;
    float fMils, f0Mils, fGridSize, fsize;
    tickMark *ticks;
	double min_gap, big_gap;

	/* sanity check */
	if ( zoom == 0 || gridSize == 0 )
		return((tickMark *)0);

	/* determine the number of tick marks that can fit in viewport */
	if ( scrollPer % gridSize ) /* viewport size (scrollPer) contains zoom */
		n_ticks = scrollPer / gridSize + 1;
	else
		n_ticks = scrollPer / gridSize;
	if ( n_ticks <= 0 )
		return((tickMark *)0);

	/* determine the grid counts per unit (or vice versa) */
	measureGrids(unitSize, gridSize, &units_per_grid, &grids_per_unit,
	             &n_gaps, &gap);
	if ( grids_per_unit == 0 )
		return((tickMark *)0);

	/* determine interval between tick marks (5 pix minimum) */
	min_gap = (double)axtPixToMils(5, ScreenRes) * 1000.0 / (double)zoom;
	for ( ix=0; ix<n_gaps; ix++ ) /* get first gap that is large enough */
	{
		big_gap = (double)gap[ix];
		if ( big_gap >= min_gap )
			break;
	}
	while ( big_gap < min_gap )
		big_gap *= 2;
	mil_gap = big_gap;
	if ( mil_gap == 0 )
	{
		free(gap);
		return((tickMark *)0);
	}
	fGridSize = (float)unitSize / (float)grids_per_unit;

	/* create tick marks array */
	sz = n_ticks * sizeof(tickMark);
	ticks = (tickMark *)malloc(sz);
	memset((char *) ticks, 0, sz);

	/* determine starting point(s) */
	if ( horiz )
	{
		/* get tick mark pos of first tick at unit boundary */
		tx = scrollPos;
		ty = 0;
		if ( units_per_grid > 1 )
			axGridToRuler(tx, ty, zeroPt, 0, gridSize, gridSize, &tx, &ty);
		else
			axGridToRuler(tx, ty, zeroPt, 0, unitSize, unitSize, &tx, &ty);
		u0Mils = tx;
		ticks[0].value = tx;
		fMils = tx;
		ticks[0].fValue = fMils;

		/* get first tick mark pos */
		tx = scrollPos;
		ty = 0;
		axGridToRuler(tx, ty, zeroPt, 0, mil_gap, mil_gap, &tx, &ty);
		t0Mils = tx;
	}
	else /* vertical */
	{
		/* get tick mark pos of first tick at unit boundary */
		tx = 0;
		ty = scrollPos;
		if ( units_per_grid > 1 )
			axGridToRuler(tx, ty, 0, zeroPt, gridSize, gridSize, &tx, &ty);
		else
			axGridToRuler(tx, ty, 0, zeroPt, unitSize, unitSize, &tx, &ty);
		u0Mils = ty;
		ticks[0].value = ty;
		fMils = ty;
		ticks[0].fValue = fMils;

		/* get first tick mark pos */
		tx = 0;
		ty = scrollPos;
		axGridToRuler(tx, ty, 0, zeroPt, mil_gap, mil_gap, &tx, &ty);
		t0Mils = ty;
	}

        /* initialize each tickmark array element */
    tMils = ticks[0].value;
    f0Mils = ticks[0].fValue;
	fGridSize = (float)unitSize/(float)gridSize;
    for (ix = 0, tmp = -1, foundWidth = FALSE; ix < n_ticks; ix++) {

		/*	tMils += gridSize;	*/
		if ( units_per_grid > 1 )
			fMils = (float)(ix+1)*(float) gridSize;
		else
			fMils = (float)(ix+1)*(float) unitSize / (float) grids_per_unit;
        ticks[ix].value = tMils + (int)fMils;
        ticks[ix].fValue = f0Mils + fMils;
        if (((ticks[ix].value - t0Mils) % mil_gap) == 0) {
            ticks[ix].display = TRUE;
        }

             /* determine the tickmark size; the 'most major'
                increments get the largest size of 16pix */
        for (jx = n_gaps - 1, ticklen = 16; jx >= 0;
             jx--, ticklen /= 2) {

            if (!gap[jx])
                continue;
            if (((ticks[ix].value - u0Mils) % gap[jx]) == 0) {
                ticks[ix].ticklen = ticklen;
                break;
            }
        }

        if (jx == n_gaps - 1) {
                /* this is one of the 'most major' tickmarks */
			val = (int)ticks[ix].fValue;

			/* make sure that labels displays at unitsize */
			ticks[ix].value = val;

            ticks[ix].labelled = TRUE;
            ticks[ix].ticklen += 4;  /* extra emphasis if labelled */
			/* assign a label */
			fsize = ticks[ix].fValue - (float)zeroPt;
            initRulerLabel(fsize, (float) unitSize, fGridSize, unitPrec,
								  &ticks[ix].labelstr[0], units_per_grid);
			/* remember the distance between 'most major' tickmarks */
            if (tmp == -1) {
                tmp = ticks[ix].value;
            }
            else if (!foundWidth) {
                majorWidth = ticks[ix].value - tmp;
                majorWidth = (int)
                    (((float) majorWidth * (float) zoom) / 1000.0);
                foundWidth = TRUE;
            }
        }
    }

        /* record the sizes of the labels */
    rulerLabelSizes(ticks, n_ticks, &maxWidth, &maxSize);
    maxSize = maxWidth;  /* whether vertical or horizontal */

    if (foundWidth) {
            /* test whether to drop some labels */
        jx = 2;   /* initial 'drop factor' (save even labels only) */
        droppedLabels = FALSE;
        while ((2 * maxSize) > majorWidth) {
            for (ix = 0; ix < n_ticks; ix++) {
                if (!ticks[ix].display || !ticks[ix].labelled)
                    continue;
                tmp = ((int) ticks[ix].value - zeroPt);
                tmp /= unitSize;
                if (tmp % jx) {
                    ticks[ix].labelled = FALSE;  /* drop */
                    ticks[ix].display = FALSE;
                    droppedLabels = TRUE;
                }
            }
            jx *= 2;
            majorWidth *= 2;
        }
    }

	/* done */
    free((char *)gap);
    *numMarks = n_ticks;
    return (ticks);
}

static void
paintRulerUnitString(long wdg,                      /* in: widget id */
                     int x, int y, int wid, int ht, /* in: panel in pix */
                     char *unitStr,                 /* in */
                     r32 *clip)                     /* in */
{
    unsigned int flags;
    int len, jx, mx, my, mht, mwid;
    p32 pos;
    r32 bbox;
    axtext txt;
    textAttrs atts;
    tx_char *txchars;
    char *err;

    atts.fontFamily = LABELFONT;
    atts.pointsize = LABELSIZE;
    atts.weight = LABELWEIGHT;
    flags = (unsigned int) (TXpointsize | TXfontFamily | TXweight);

        /* create the unitstring text */
    memset((char *) &txt, 0, sizeof(axtext));
    len = strlen(unitStr);
	/*make sure length of string is less than 3 */
	if ( len > 3 )
		len = 3;
    txchars = (tx_char *) malloc((len + 1) * sizeof(tx_char));
    for (jx = 0; jx < len; jx++) {
        txchars[jx] = (tx_char) unitStr[jx];
    }
    setDefaultTextAttrs(&txt, &atts, flags);
    insChars(&txt, txchars, 0, len);

        /* determine the position, paint, cleanup */
    mx = axtPixToMils(x, ScreenRes);
    my = axtPixToMils(y, ScreenRes);
    mwid = axtPixToMils(wid, ScreenRes);
    mht = axtPixToMils(ht, ScreenRes);
    if (err = getOverallBounder(&txt, 0, -1, &bbox)) {
        freeErrString(err);
        return;
    }
        /* center-align for now; will need to tune */
    pos.x = -bbox.origin.x;
    pos.y = -bbox.origin.y;
    pos.x += (mwid - bbox.width) / 2;
    pos.y += (mht - bbox.height) / 2;
    pos.x += mx;
    pos.y += my;
    setPos(&txt, &pos);

    if (err = dispTx(wdg, &txt, clip)) {
        freeErrString(err);
    }

    free((char *) txchars);
    daxtext(&txt);
}

static void
paintRulerLabel( long wdg,         /* in: widget of ruler */
                 int horiz,       /* in: bool; whether horiz ruler */
                 tickMark *mark,  /* in: data for this tick mark */
                 int rulerStart,  /* in: ruler origin coord in mils */
                 int zoom,        /* in: zoom factor, reference val 100 */
                 r32 *clip)       /* in: clip in mils */
{
    unsigned int flags;
    int len, jx, tmp;
    p32 pos;
    axtext txt;
    textAttrs atts;
    tx_char *txchars;
    char *err;

    atts.fontFamily = LABELFONT;
    atts.pointsize = LABELSIZE;
    atts.weight = LABELWEIGHT;
    flags = (unsigned int) (TXpointsize | TXfontFamily | TXweight);
    if (!horiz) {
        flags |= TXangle;
        atts.angle = 900000;
    }

        /* create the labelstr text */
    memset((char *) &txt, 0, sizeof(axtext));
    len = strlen(mark->labelstr);
    txchars = (tx_char *) malloc
        ((len + 1) * sizeof(tx_char));
    for (jx = 0; jx < len; jx++) {
        txchars[jx] = (tx_char) mark->labelstr[jx];
    }
    setDefaultTextAttrs(&txt, &atts, flags);
    insChars(&txt, txchars, 0, len);

        /* determine the position for the label */
    tmp = mark->value - rulerStart;
    tmp = (int) ((float) (tmp * zoom) / 1000.0);
    if (horiz) {
        pos.x = tmp - mark->bbox.origin.x + 40;
        pos.y = axtPixToMils(GE_XRULER_HEIGHT - 8, ScreenRes);
    }
    else {
        pos.x = axtPixToMils(GE_YRULER_WIDTH - 8, ScreenRes);
            /* we use bbox x here because it was measured unrotated */
        pos.y = tmp - mark->bbox.origin.x - 40;
    }

        /* set position, display, cleanup */
    setPos(&txt, &pos);
    if (err = dispTx(wdg, &txt, clip)) {
        freeErrString(err);
    }

    free((char *) txchars);
    daxtext(&txt);
}


/*************************************************************************
function: PaintGEyruler - Regen widget
*/
void
PaintGEyruler( AXWidget_ge_ruler w,
               XExposeEvent *exP,
               int full_regen,
               int layered)
{
    char *unitName;
    tickMark *ticks;
    int winID, etch, xpos, ypos, ix, hyt, wid;
    int xScrollPos, yScrollPos, xScrollPer, yScrollPer;
    int xZeroPt, yZeroPt, xZoom, yZoom, unitSize, gridSize;
    int pixpos, numMarks, rulerStart, tmp, unitPrec;
    XRectangle clip;
    r32 tclip;

    if ( !w || !w->display )
        return;

    if ((w->flags & WFLG_IN_EDIT) && layered) {
        if(dePaintWidget((AXWidget)w))
            return;
    }

    if ( !w->window ) {
#if AXMS
        RealizeGEyruler(w);
#else
        w->window = AxCrtWindow(Dpy,WIN_OF(w),
                                w->xpos, w->ypos,
                                w->width, w->height,
                                0,CopyFromParent,InputOutput,
                                CopyFromParent,WinMask,&WinAtt);

        XSelectInput(Dpy, w->window, TM_ALL_INPUTS_MASK);
        AxMapWindow((AXWidget)w, Dpy, w->window, w->name);
#endif
        w->mapped = TRUE;
    }

    if ( !w->parent->workWindow )
        return;
    winID = ATMWindowFromXWindow(w->parent->workWindow->window);
    if ( !winID )
        return;

    if ( full_regen ) {
        XFillRectangle(Dpy,w->window,BackGC,0,0,w->width,w->height);
        w->value = -1;   /* since this action lifts tracker */
    }

    /* draw ruler's shadow/outline */
#if AXMS
	etch = 1;
#else
    etch = 2;
#endif
    xpos = 0;
    ypos = 0;
    wid = w->width-1;

	/* colorpal and ge_rulers hang off W_MAIN (parent) widget */
    if ( (w->parent->colorpal && w->parent->colorpal->display) ||
         (w->parent->ge_xruler && w->parent->ge_xruler->display) )
        hyt = w->height;
    else
        hyt = w->height-1;
    AxDrawShadow( w->window, SunGC, ShadowGC, etch,
                 xpos, ypos, wid, hyt,
                 OLSHADOW_NODRAW );

        /* prepare clip, both for ForeGC, and for ruler label paints */
    clip.x = (short) (xpos + etch);
    clip.y = (short) (ypos + etch);
    clip.width = (unsigned short) (wid - (2 * etch) - 1);
    clip.height = (unsigned short) (hyt - (2 * etch) - 1);
    tclip.origin.x = axtPixToMils((int) clip.x, ScreenRes);
    tclip.origin.y = axtPixToMils((int) clip.y, ScreenRes);
    tclip.width = (unsigned int) axtPixToMils((int) clip.width,
                                              ScreenRes);
    tclip.height = (unsigned int) axtPixToMils((int) clip.height,
                                               ScreenRes);
    XSetClipRectangles(Dpy, ForeGC, 0, 0, &clip, 1, Unsorted);

    /* draw the ruling line */
    XDrawLine ( Dpy, w->window, ForeGC,
                4, etch+1,
                4, w->height - (2 * etch) - 1 );

    /* get the ruler info from the atm window array (in mils) */
    AxGetGridInfo(winID, &xScrollPos, &yScrollPos, &xScrollPer,
                  &yScrollPer, &xZeroPt, &yZeroPt, &xZoom, &yZoom,
                  &unitSize, &gridSize, &unitPrec, &unitName);

    ticks = axComputeTickMarks(FALSE, yScrollPos, yScrollPer, yZeroPt,
                               yZoom, unitSize, gridSize, unitPrec, &numMarks);
    if (ticks) {
        /* draw the tickmarks and labels */
        rulerStart = yScrollPos;
        for (ix = 0; ix < numMarks; ix++) {
            if (!ticks[ix].display)
                continue;
            tmp = ticks[ix].value - rulerStart;
            tmp = (int) ((float) (tmp * yZoom) / 1000.0);
            pixpos = milsToPix(tmp, ScreenRes);
            XDrawLine(Dpy, w->window, ForeGC, 4, pixpos,
                      4 + ticks[ix].ticklen, pixpos);
            if (ticks[ix].labelled && ticks[ix].labelstr) {
                paintRulerLabel((long)w, FALSE, &ticks[ix],
                                rulerStart, yZoom, &tclip);
            }
        }
        free((char *) ticks);
    }

    XSetClipMask(Dpy, ForeGC, None);  /* remove clip */
}

/*************************************************************************
function: PaintGExruler - Regen widget
*/
void
PaintGExruler( AXWidget_ge_ruler w,
               XExposeEvent *exP,
               int full_regen,
               int layered)
{
    char *unitName;
    tickMark *ticks;
    int winID, etch, xpos, ypos, wid, hyt, ix, tmp;
    int xScrollPos, yScrollPos, xScrollPer, yScrollPer;
    int xZeroPt, yZeroPt, xZoom, yZoom, unitSize, gridSize;
    int pixpos, numMarks, rulerStart, unitPrec;
    XRectangle clip;
    r32 tclip;

    if ( !w || !w->display )
        return;

    if ((w->flags & WFLG_IN_EDIT) && layered) {
        if (dePaintWidget((AXWidget)w))
            return;
    }

    if ( !w->window ) {
#if AXMS
        RealizeGExruler(w);
#else
        w->window = AxCrtWindow(Dpy,WIN_OF(w),
                                w->xpos, w->ypos,
                                w->width, w->height,
                                0,CopyFromParent,InputOutput,
                                CopyFromParent,WinMask,&WinAtt);
        XSelectInput(Dpy, w->window, TM_ALL_INPUTS_MASK);
        AxMapWindow((AXWidget)w, Dpy, w->window, w->name);
#endif
        w->mapped = TRUE;
    }

    if ( !w->parent->workWindow )
        return;
    winID = ATMWindowFromXWindow(w->parent->workWindow->window);
    if ( !winID )
        return;

    if ( full_regen ) {
        XFillRectangle(Dpy,w->window,BackGC,0,0,w->width,w->height);
        w->value = -1;   /* since this action lifts tracker */
    }

    /* draw shadow/outline for ruler */
#if AXMS
	etch = 1;
#else
    etch = 2;
#endif
    xpos = 0;
    wid = w->width - GE_YRULER_WIDTH;
    ypos = 0;

	/* colorpal hangs off W_MAIN (parent) widget */
    hyt = (w->parent->colorpal && w->parent->colorpal->display)
                                                    ? w->height : w->height-1;
    AxDrawShadow( w->window, SunGC, ShadowGC, etch,
                 xpos, ypos, wid, hyt,
                 OLSHADOW_NODRAW );

        /* prepare clip; both for ForeGC, and for label paints */
    clip.x = (short) (xpos + etch);
    clip.y = (short) (ypos + etch);
    clip.width = (unsigned short) (wid - (2 * etch) - 1);
    clip.height = (unsigned short) (hyt - (2 * etch) - 1);
    tclip.origin.x = axtPixToMils((int) clip.x, ScreenRes);
    tclip.origin.y = axtPixToMils((int) clip.y, ScreenRes);
    tclip.width = (unsigned int) axtPixToMils((int) clip.width,
                                              ScreenRes);
    tclip.height = (unsigned int) axtPixToMils((int) clip.height,
                                               ScreenRes);
    XSetClipRectangles(Dpy, ForeGC, 0, 0, &clip, 1, Unsorted);

    /* draw the ruling line */
    XDrawLine ( Dpy, w->window, ForeGC,
                etch + 1, 4,
                wid - (2 * etch) - 1, 4 );

    /* get the ruler info from the atm window array (in mils) */
    AxGetGridInfo(winID, &xScrollPos, &yScrollPos, &xScrollPer,
                  &yScrollPer, &xZeroPt, &yZeroPt, &xZoom, &yZoom,
                  &unitSize, &gridSize, &unitPrec, &unitName);

    ticks = axComputeTickMarks(TRUE, xScrollPos, xScrollPer, xZeroPt,
                               xZoom, unitSize, gridSize, unitPrec, &numMarks);

    if (ticks) {
        rulerStart = xScrollPos;
        /* draw the tickmarks and labels */
        for (ix = 0; ix < numMarks; ix++) {
            if (!ticks[ix].display)
                continue;
            tmp = ticks[ix].value - rulerStart;
            tmp = (int) ((float) (tmp * xZoom) / 1000.0);
            pixpos = milsToPix(tmp, ScreenRes);
            XDrawLine(Dpy, w->window, ForeGC, pixpos, 4,
                      pixpos, 4 + ticks[ix].ticklen); 
            if (ticks[ix].labelled && ticks[ix].labelstr) {
                paintRulerLabel((long)w, TRUE, &ticks[ix],
                                rulerStart, xZoom, &tclip);
            }
        }
        free((char *) ticks);
    }

    /* draw shadow/outline for units box */
    xpos = wid;
    wid = GE_YRULER_WIDTH - 1;
    XSetClipMask(Dpy, ForeGC, None);  /* remove ForeGC's clip now */
    AxDrawShadow( w->window, SunGC, ShadowGC, etch,
                 xpos, ypos, wid, hyt,
                 OLSHADOW_NODRAW );

    /* draw an etch just inside the unit name panel */
    XDrawRectangle(Dpy,w->window,ForeGC,
                   xpos + etch,
                   ypos + etch,
                   wid - (2 * etch) - 1,
                   hyt - (2 * etch) - 1);

    /* rewrite label clip for the units panel */
    tclip.origin.x = axtPixToMils(xpos + etch + 1, ScreenRes);
    tclip.origin.y = axtPixToMils(ypos + etch + 1, ScreenRes);
    tclip.width = (unsigned int) 
        axtPixToMils(wid - 2 * (etch + 1), ScreenRes);
    tclip.height = (unsigned int)
        axtPixToMils(hyt - 2 * (etch + 1), ScreenRes);

    /* draw the unit name */
    paintRulerUnitString((long)w, xpos + etch + 1,
                         ypos + etch + 1,
                         wid - 2 * (etch + 1),
                         hyt - 2 * (etch + 1),
                         unitName, &tclip);
}
