/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xmpage.c
   Applix ASC version: 28
   Date: Sun Jan 17 21:13:00 1999
*/
#include <math.h>
#include <basicwin.h>
#if AX_X11
#include <X11/keysym.h>
#endif
#include <ax_xm.h>
#include <aos.h>
#include <dwt2xm.h>
#include <xmlib.h>
#include <elxwins.h>
#include <wchars.h>
#include <thimble.h>
#include <ctype.h>
#include <elf_env.h>
#include <eldata.h>
#include <evreason.h>
#include <xmpixel.h>
#include <xmwcolor.h>
#include <intsarr.h>
#include <ge_syms.h>
#include <xmpage.p>

/* Standard Externs for Widget Modules */

extern int WidgetNeedsUpdated; /* Set when a widget needs repainted */
extern int AxCharAbove, AxCharHeight, AxCharBelow;
extern int AxOpenLook;
extern int AxBigFont;
extern int UseWidgetColors;
extern Display *Dpy;
extern short DpyDepth;
extern int AxButtonDown;
extern short THIMpid;

extern unsigned long FGpixel;
extern long WinMask;
extern XSetWindowAttributes WinAtt;
extern Pixmap UnknownPixmap;
extern AXWidget ClickWidget;
extern unsigned long BackWorkPixel;
extern Colormap WorkColormap;
extern GC ForeGC, BackGC, Gc1;
extern GC SunGC, ShadowGC, DisabledGC, GrayLineGC;
extern GC RecessGC;
extern GC SmallFontGC;     /* smaller sans-serif system font, forground mode */
extern GC olClrSunButtonGC, olClrShadowButtonGC, olClrRecessButtonGC;
extern GC MonospaceGC;       /* monspace system font, forground mode */
extern GC MonospaceBackGC;   /* monspace system font, background mode */
extern GC MonospaceBoldGC, MonospaceBoldBackGC;
extern GC MonospaceItalGC, MonospaceItalBackGC;
extern GC MonospaceItalBoldGC, MonospaceItalBoldBackGC;
extern XFontStruct *AxDlgMainFont;  /* Bolded dialog font */
extern XFontStruct *MonospaceFont, *MonospaceBoldFont;
extern XFontStruct *MonospaceItalFont, *MonospaceItalBoldFont;
extern XFontStruct *AxStatusLineFont;
XFontStruct *AxDlgMiniFont, *MonospaceFont, *MonospaceBoldFont;
extern int DoubleMillisec;
extern unsigned long ForePixel;
extern unsigned long BackPixel;
extern short ColorDisplay;
extern short NoColor;
extern Window RootWin;
extern Cursor ThimCrosshairCursor;
extern Cursor ThimPointerCursor;
extern Cursor ThimLeftRightCursor;
extern unsigned long AlisPixels[];
extern unsigned long WidgetPixels[];
/* End of Standard Externs */

#if !BUILD_WIDE
#define axDrawString16 axDrawString
#endif

/****************************************************************************
 *
 * module: UpdatePageInfo
 */
void
UpdatePageInfo(w)
AXWidget_page w;
{
    XRectangle rect[1];
    int len, x, y, width, height, x_off, y_off;
    char pageinfo[80];
    register GC gc;
	XFontStruct *restore_fs;

	if (w->disabled) {
		gc = DisabledGC;
		restore_fs = AxDlgMainFont;
	}
	else {
		gc = SmallFontGC;
		restore_fs = AxDlgMiniFont;
	}

    WIDGET_PARANOID( w, W_PAGEBAR );

#if AXMS
    x = 23;
    y = 0;
    x_off = 83;
    y_off = 0;
    width = w->width - 83;
    height = w->height;
#else
    x = 21;
    y = 2;
    x_off = 54;
    y_off = 6;
    width = w->width - 78;
    height = w->height - 4;
#endif

    if ( UseWidgetColors == FALSE )
	{
        XFillRectangle(Dpy, w->window, BackGC, x, y, width, height);
	}
    else
	{
        XFillRectangle(Dpy, w->window, RecessGC, x, y, width, height);
	}
    rect[0].x = x;
    rect[0].y = y + 1;
    rect[0].width = w->width - x_off;
    rect[0].height = w->height - y_off;
    XSetClipRectangles(Dpy, gc, 0, 0, rect, 1, Unsorted);

    sprintf(pageinfo, XLT("%d of %d"), w->scroll_pos, w->scroll_max);
    len = MBstrlen(pageinfo);

    /* get the font info of the gc */
    axSetFont(Dpy, gc, AxStatusLineFont);
    x = w->xpos + (w->width-36)/2 - len*6/2 + 1;
    y = w->ypos + AxStatusLineFont->max_bounds.ascent + 2;
    axDrawString(Dpy, w->window, gc, (XFontStruct *) 0, x, y, pageinfo, len);

    /* restore the gc */
    axSetFont(Dpy, gc, restore_fs);
    XSetClipMask(Dpy, gc, None);
}

/****************************************************************************
 *
 * module: axmPaintJmpButton
 */
void
axmPaintJmpButton(w, drw, x, y, wid, hyt, button_up)
AXWidget w;
Drawable drw;
int x, y;
int wid, hyt;
int button_up;
{
	int thickness;
    XPoint pt[5];

    if ( w->disabled )
        return;

#if AXMS
	thickness = 1;
#else
	thickness = 2;
#endif

    if ( UseWidgetColors == FALSE )
    {
        pt[0].x = x;
        pt[0].y = y;
        pt[1].x = x + wid - 1;
        pt[1].y = y;
        pt[2].x = x + wid - 1;
        pt[2].y = y + hyt - 1;
        pt[3].x = x;
        pt[3].y = y + hyt - 1;
        pt[4].x = x;
        pt[4].y = y;
        XFillPolygon(Dpy, drw, ForeGC, pt, 5, Convex, CoordModeOrigin);
    }
    else if ( button_up )
    {
        XFillRectangle(Dpy, drw, BackGC, x,y, wid,hyt);
        AxDrawShadow(drw,SunGC,ShadowGC,thickness,x,y,wid,hyt,OLSHADOW_NODRAW);
    }
    else
    {
        XFillRectangle(Dpy,drw,RecessGC, x,y, wid,hyt);
        AxDrawShadow(drw,ShadowGC,SunGC,thickness,x,y,wid,hyt,OLSHADOW_NODRAW);
    }
}

/****************************************************************************
 *
 * module: axmPaintGoToButton
 */
void
axmPaintGoToButton(w, drw, x, y, button_up)
AXWidget w;
Drawable drw;
int x, y;
int button_up;
{
	pixcache_t *pxc;
    int width, height, x_off, y_off;

	/* draw the button outline */
#if AXMS
    x += 2;
    width  = w->height - 4;
    height = w->height;
    x_off = 2;
    y_off = 4;
#else
    width = height = w->height - 4;
    x_off = y_off = 2;
#endif
	XSetForeground(Dpy, Gc1, FGpixel);
	XSetFunction(Dpy, Gc1, GXcopy);
	XSetFillStyle(Dpy, Gc1, FillSolid);
	XSetClipMask(Dpy, Gc1, 0);
	if ( UseWidgetColors == TRUE )
		axmPaintJmpButton(w, drw, x, y, width, height, button_up);
	else
		XDrawRectangle(Dpy, drw, Gc1, x, y, width, height);

	/* draw the button content */
	pxc = (pixcache_t *)Pixmapper(NXLT("quest"));
	if ( pxc->clip_mask )
	{
		XSetClipOrigin(Dpy, Gc1, x + x_off, y + y_off);
		XSetClipMask(Dpy, Gc1, pxc->clip_mask);
	}
    XCopyArea(Dpy, pxc->pix, drw, Gc1, 0, 0, pxc->width, pxc->height,
                                                x + x_off, y + y_off);
	if ( pxc->clip_mask )
		XSetClipMask(Dpy, Gc1, None);
}

void
axmPaintDcrButton(w, drw, x,y, button_up)
AXWidget w;
Drawable drw;
int x, y;
int button_up;
{
    XPoint pt[4];
    int width, height, y_off;

    /* draw the button outline */
#if AXMS
    width = w->height - 3;
    height = w->height;
    y_off = 7;
#else
    width = height = w->height - 4;
    y_off = 5;
#endif
    XSetForeground(Dpy, Gc1, FGpixel);
    XSetFunction(Dpy, Gc1, GXcopy);
    XSetFillStyle(Dpy, Gc1, FillSolid);
    XSetClipMask(Dpy, Gc1, 0);
    if ( UseWidgetColors == TRUE )
		axmPaintJmpButton(w, drw, x, y, width, height, button_up);
    else
		XDrawRectangle(Dpy, drw, Gc1, x, y, width, height);

	/* draw the minus sign */
    pt[0].x = x + 3;
    pt[0].y = y + y_off;
    pt[1].x = pt[0].x + 6;
    pt[1].y = pt[0].y;
    pt[2].x = pt[1].x;
    pt[2].y = pt[1].y + 2;
    pt[3].x = pt[2].x - 6;
    pt[3].y = pt[2].y;
    XFillPolygon(Dpy, drw, Gc1, pt, 4, Convex, CoordModeOrigin);
}

void
axmPaintIncrButton(w, drw, x,y, button_up)
AXWidget w;
Drawable drw;
int x, y;
int button_up;
{
    XPoint pt[12];
    int width, height, y_off;

    /* draw the button outline */
#if AXMS
    x += 1;
    width = w->height - 3;
    height = w->height;
    y_off = 5;
#else
    width = height = w->height - 4;
    y_off = 3;
#endif
    XSetForeground(Dpy, Gc1, FGpixel);
    XSetFunction(Dpy, Gc1, GXcopy);
    XSetFillStyle(Dpy, Gc1, FillSolid);
    XSetClipMask(Dpy, Gc1, 0);
    if ( UseWidgetColors == TRUE )
		axmPaintJmpButton(w, drw, x, y, width, height, button_up);
    else
		XDrawRectangle(Dpy, drw, Gc1, x, y, width, height);

    /* draw the button content */
    pt[0].x = x + 5;
    pt[0].y = y + y_off;
    pt[1].x = pt[0].x + 2;
    pt[1].y = pt[0].y;
    pt[2].x = pt[1].x;
    pt[2].y = pt[1].y + 2;
    pt[3].x = pt[2].x + 2;
    pt[3].y = pt[2].y;
    pt[4].x = pt[3].x;
    pt[4].y = pt[3].y + 2;
    pt[5].x = pt[4].x - 2;
    pt[5].y = pt[4].y;
    pt[6].x = pt[5].x;
    pt[6].y = pt[5].y + 2;
    pt[7].x = pt[6].x - 2;
    pt[7].y = pt[6].y;
    pt[8].x = pt[7].x;
    pt[8].y = pt[7].y - 2;
    pt[9].x = pt[8].x - 2;
    pt[9].y = pt[8].y;
    pt[10].x = pt[9].x;
    pt[10].y = pt[9].y - 2;
    pt[11].x = pt[10].x + 2;
    pt[11].y = pt[10].y;
    XFillPolygon(Dpy, drw, Gc1, pt, 12, Convex, CoordModeOrigin);
}

/****************************************************************************
 *
 *     module: PaintPageBar
 */
void
PaintPageBar(w, exP, regen, layered)
AXWidget w;
XExposeEvent *exP;
int regen;
int layered;
{
    int wid, hyt, x, y;
    Drawable drw;

    if ( !w->window )
        return;

    /* calc position and size info */
    drw = w->window;
    x = w->xpos;
    y = w->ypos;
    wid = w->width;
    hyt = w->height;

    /* clear the background */
    if ( UseWidgetColors == FALSE )
        XFillRectangle(Dpy, drw, BackGC,   x,y, wid,hyt);
    else
        XFillRectangle(Dpy, drw, RecessGC, x,y, wid,hyt);

#if AX_X11
    /* draw the shadow around the button bar */
    AxDrawShadow(drw, ShadowGC, SunGC, 2, x,y, wid,hyt, OLSHADOW_NODRAW);
#endif

    /* draw the buttons */
    if ( !w->disabled )
    {
        /* draw the first page button */
#if AXMS
        x = w->xpos;
        y = w->ypos;
        wid = 7;
#else
        x = w->xpos+2;
        y = w->ypos+2;
        wid = 7;
        hyt = hyt - 4;
#endif
        axmPaintJmpButton(w, drw, x,y, wid,hyt, TRUE);

        /* draw the page left arrow */
        axmPaintLeftArrow(w, drw, 7, 0, TRUE);

        /* draw the page right arrow */
        axmPaintRightArrow(w, drw, w->width - 42, 0, TRUE);

        /* draw the final page button */
        x = w->width-45;
        axmPaintJmpButton(w, drw, x,y, wid,hyt, TRUE);

        /* draw the decr page button */
        x = w->width-38;
        axmPaintDcrButton(w, drw, x,y, TRUE);

        /* draw the incr page button */
        x = w->width-26;
        axmPaintIncrButton(w, drw, x,y, TRUE);

        /* jump to new page button */
        x = w->width-14;
		axmPaintGoToButton(w, drw, x, y, TRUE);
    }

    /* draw the page x of n label */
    UpdatePageInfo((AXWidget_page)w);
}

/****************************************************************************
 *
 *     module: ArmPageButton
 */
void
ArmPageButton(w, eventP)
AXWidget w;
XEvent *eventP;
{
    int win_id, new_x, new_y, y1, y2, hyt, y_inside;
    char str[STRPATH];
    static int x, is_empty;
    Drawable drw;

    if ( !w || w->disabled )
        return;

    /* depress the correct button */
    if ( eventP->type == ButtonPress )
    {
        x = eventP->xbutton.x;
        y1 = w->ypos + 2;
        hyt = w->height - 4;
        drw = w->window;

        /* goto first page button */
        if ( x >= 2 && x < 9 )
            axmPaintJmpButton(w, drw, w->xpos+2, y1, 7, hyt, FALSE);

        /* goto previous page button */
        else if ( x < 20 )
            axmPaintLeftArrow(w, drw, 7, 0, FALSE);

        /* goto (create) next page button */
        else if ( x >= w->width-56 && x < w->width-45 )
            axmPaintRightArrow(w, drw, w->width - 42, 0, FALSE);

        /* goto final page button */
        else if ( x >= w->width-45 && x < w->width-38 )
            axmPaintJmpButton(w, drw, w->width-45, y1, 7, hyt, FALSE);

        /* delete current page button */
        else if ( x >= w->width-38 && x < w->width-26 )
            axmPaintDcrButton(w, drw, w->width-38, y1, FALSE);

        /* insert (and goto) new page button */
        else if ( x >= w->width-26 && x < w->width-14 )
            axmPaintIncrButton(w, drw, w->width-26, y1, FALSE);

        /* jump to new page button */
        else if ( x >= w->width-14 && x < w->width-2 )
            axmPaintGoToButton(w, drw, w->width-14, y1, FALSE);
    }

    /* raise any buttons that were depressed */
    else if ( eventP->type == ButtonRelease )
    {
        win_id = w->parent ? AxWindowIdFromShellWidget(w->parent) : 0;
        new_x = eventP->xbutton.x;
        y1 = w->ypos + 2;
        hyt = w->height - 4;
        drw = w->window;

        /* is new y position within widget bounds */
        new_y = eventP->xbutton.y;
        y2 = y1 + hyt - 1;
        y_inside = (new_y >= y1 && new_y <= y2) ? TRUE : FALSE;

        /* jump to first page button */
        if ( x >= 2 && x < 9 )
        {
            /* raise the button */
            axmPaintJmpButton(w,drw, w->xpos+2, y1, 7, hyt, TRUE);

            /* jump to the first page */
            if ( win_id && new_x >= 2 && new_x < 9 && y_inside )
            {
                sprintf(str, NXLT("%d GR_FIRST_PAGE@"), win_id);
                ElfStartNewTaskWithArgs(GEElfMacroId(ELFM_GE_PALETTE_LAUNCHER_), str);
            }
        }

        /* goto previous page button */
        else if ( x < 20 ) /* size of arrow */
        {
            /* raise the button */
            axmPaintLeftArrow(w, drw, 7, 0, TRUE);

            /* jump to the previous page */
            if ( win_id && x < 20 && y_inside )
            {
                sprintf(str, NXLT("%d GR_PREVIOUS_PAGE@"), win_id);
                ElfStartNewTaskWithArgs(GEElfMacroId(ELFM_GE_PALETTE_LAUNCHER_), str);
            }
        }

        /* goto (create) next page button */
        else if ( x >= w->width-56 && x < w->width-45 )
        {
            /* raise the button */
            axmPaintRightArrow(w, drw, w->width - 42, 0, TRUE);

            /* (create and) jump to the next page */
            if ( win_id && x >= w->width-56 && x < w->width-45 && y_inside )
            {
                sprintf(str, NXLT("%d GR_NEXT_PAGE_DLG@"), win_id);
                ElfStartNewTaskWithArgs(GEElfMacroId(ELFM_GE_PALETTE_LAUNCHER_), str);
            }
        }

        /* goto final page button */
        else if ( x >= w->width-45 && x < w->width-38 )
        {
            /* raise the button */
            axmPaintJmpButton(w,drw, w->width-45, y1,7, hyt,TRUE);

            /* jump to the final page */
            if ( win_id && x >= w->width-45 && x < w->width-38 && y_inside )
            {
                sprintf(str, NXLT("%d GR_LAST_PAGE@"), win_id);
                ElfStartNewTaskWithArgs(GEElfMacroId(ELFM_GE_PALETTE_LAUNCHER_), str);
            }
        }

        /* delete current page button */
        else if ( x >= w->width-38 && x < w->width-26 )
        {
            /* raise the button */
            axmPaintDcrButton(w,drw, w->width-38, y1, TRUE);

            /* delete the current page */
            if ( win_id && x >= w->width-38 && x < w->width-26 && y_inside )
            {
                if ( ((AXWidget_page)w)->scroll_max == 1 && w->scroll_pos == 1 )
                {
                    sprintf(str, NXLT("%d GR_DELETE_PAGE1_DLG@"), win_id);
                    ElfStartNewTaskWithArgs(GEElfMacroId(ELFM_GE_PALETTE_LAUNCHER_), str);
                }
                else
                {
                    sprintf(str, NXLT("%d GR_DELETE_PAGE_DLG@"), win_id);
                    ElfStartNewTaskWithArgs(GEElfMacroId(ELFM_GE_PALETTE_LAUNCHER_), str);
                }
            }
        }

        /* insert (and goto) new page button */
        else if ( x >= w->width-26 && x < w->width-14 )
        {
            /* raise the button */
            axmPaintIncrButton(w,drw, w->width-26, y1, TRUE);

            /* insert a new page */
            if ( win_id && x >= w->width-26 && x < w->width-14 && y_inside )
            {
				if ( ThimIsShifted() )
				{
					is_empty = get_page_info_from_ge(PidFromATMWindow(win_id));
					if ( is_empty )
						sprintf(str, NXLT("%d GR_INSERT_PAGE_DLG@"), win_id);
					else
						sprintf(str, NXLT("%d GR_DUPLICATE_PAGE_DLG@"), win_id);
				}
				else
					sprintf(str, NXLT("%d GR_INSERT_PAGE_DLG@"), win_id);
                ElfStartNewTaskWithArgs(GEElfMacroId(ELFM_GE_PALETTE_LAUNCHER_), str);
            }
        }
    
        /* jump to new page button */
        else if ( x >= w->width-14 && x < w->width-2 )
        {
            /* raise the button */
            axmPaintGoToButton(w,drw, w->width-14, y1, TRUE);

            /* insert a new page */
            if ( win_id && x >= w->width-14 && x < w->width-2 && y_inside )
            {
				sprintf(str, NXLT("%d GR_JUMP_TO_PAGE@"), win_id);
                ElfStartNewTaskWithArgs(GEElfMacroId(ELFM_GE_PALETTE_LAUNCHER_), str);
            }
        }
    }
}

void
AxModifyPageBar(wd, pos, max)
long wd;
int pos, max;
{
	AXWidget_main w;
    Arg arglist[2];
    int an;

	w = (AXWidget_main)wd;
    if ( !w || !w->pageBar || w->pageBar->disabled )
        return;
    an = 0;
    if ( w->pageBar->scroll_max != max )
    {
        XtSetArg(arglist[an], AXmNmaximum, max);
        an++;
    }
    if ( w->pageBar->scroll_pos != pos )
    {
        XtSetArg(arglist[an], AXmNvalue, pos);
        an++;
    }
    if ( an )
        AxXtSetValues((AXWidget)w->pageBar, arglist, an);
}

void
AxXmRepaintPageBar(wd)
long wd;
{
	AXWidget_main w;

	w = (AXWidget_main)wd;
    if ( !w || !w->pageBar || !w->pageBar->display )
        return;

    PaintPageBar((AXWidget)w->pageBar, (XExposeEvent *)0, TRUE, FALSE);
}

void
AxInactivePageBar(wd)
long wd;
{
	AXWidget_main w;

	w = (AXWidget_main)wd;
    if ( !w || !w->pageBar )
        return;
    AxXtSetSensitive((AXWidget)w->pageBar, FALSE);
    w->pageBar->update_widget = TRUE;
    WidgetNeedsUpdated = TRUE;
}

void
AxactivePageBar(wd)
long wd;
{
	AXWidget_main w;

	w = (AXWidget_main)wd;
    if ( !w || !w->pageBar )
        return;
    AxXtSetSensitive((AXWidget)w->pageBar, TRUE);
    w->pageBar->update_widget = TRUE;
    WidgetNeedsUpdated = TRUE;
}
