/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xmscroll.c
   Applix ASC version: 33
   Date: Mon Nov 16 13:53:36 1998
*/
#include <ax_xm.h>
#include <aos.h>
#include <xmlib.h>
#include <elxwins.h>

#define MIN_ELEV_PIX 12

extern AXWidget_scroll ScrollBarArmed;    /* currently armed W_SCROLL */
extern AXWidget_scroll StrokingBar;

extern int AxOpenLook;
extern int UseWidgetColors;
extern int AxBigFont;
extern Display *Dpy;
extern Pixmap ScratchMap; /* screen size pixmap available as app resource */
extern short DpyDepth;

extern GC ForeGC, BackGC, RecessGC;
extern GC SunGC, ShadowGC;
extern GC olButtonGC, olDisabledButtonGC;
extern GC olClrBackButtonGC;
extern GC olClrSunButtonGC, olClrShadowButtonGC, olClrRecessButtonGC;

extern XFontStruct *Grafont;

extern long WinMask;
extern XSetWindowAttributes WinAtt;

extern int DoubleMillisec;
extern unsigned long TimeOfButtonPress; /* event time of button down */


extern Window AxCrtWindow();
extern GC axCreateGC();

void PaintScrollBar(), ScrollBarEvent();
void xmDestroyScrollPixmap();
static void PaintVScrollBar(), PaintHScrollBar(), olScrollBarEvent();
static void axolDrawScrollBar(), axolDrawClrScrlElev();
static int axolScrollReason(), axolScrollDisplayStyle();
static void CrtScrollPixmap();
static void ClearScrollPixmap();


/*
 * For motif, dont paint stroking bar.  If QA people insist on stroking
 * bars for motif, flip STROKING_BAR to 1
 */
#define STROKING_BAR 0


/*************************************************************************
function: ScrollBarEvent - Handle mouse event on scroll bar
??1 Scroll tracking stinks and needs to be re-done!
*/
void
ScrollBarEvent (w,eventP)
register AXWidget_scroll w;
XEvent *eventP;
{
    register int x,y,width,height,delta,total_delta;
    static int last_x, last_y, first_point, first_pos, first_pix;
    XmScrollBarCallbackStruct data;
    int tempReason;
    int doCallBack = TRUE;
    int old_scroll_pos;

    /* if the bar is full size, ignore all events */

    if ( w->scroll_max < w->scroll_per )
        return;

    x = eventP->xbutton.x;
    y = eventP->xbutton.y;
    width = w->width;
    height = w->height;

    if ( eventP->type == ButtonPress ) {
        last_x = x;
        last_y = y;
        first_pos = w->scroll_pos;
        first_pix = w->pospix;
        if ( w->scroll_type == XmVERTICAL ) {
            if ( y < width ) {
                data.reason = XmCR_DECREMENT;
                w->armed = XmCR_DECREMENT;
                ScrollBarArmed = w;
            }
            else if ( y > height-width ) {
                data.reason = XmCR_INCREMENT;
                w->armed = XmCR_INCREMENT;
                ScrollBarArmed = w;
            }
            else if ( y < width+w->pospix )
                data.reason = XmCR_PAGE_DECREMENT;
            else if ( y > width+w->pospix+w->forepix )
                data.reason = XmCR_PAGE_INCREMENT;
            else {
                if ( IsElfInLearnMode() && (w->parent->type == W_MAIN) )
                    return; /* ignore bar drag in learn mode */
                first_point = y;
                StrokingBar = w;
#if STROKING_BAR
                PaintScrollBar(w,NULL,FALSE,TRUE);
#endif
                return;
            }
        }
        else {
            if ( x < height ) {
                data.reason = XmCR_DECREMENT;
                w->armed = XmCR_DECREMENT;
                ScrollBarArmed = w;
            }
            else if ( x > width-height ) {
                data.reason = XmCR_INCREMENT;
                w->armed = XmCR_INCREMENT;
                ScrollBarArmed = w;
            }
            else if ( x < height+w->pospix )
                data.reason = XmCR_PAGE_DECREMENT;
            else if ( x > height+w->pospix+w->forepix )
                data.reason = XmCR_PAGE_INCREMENT;
            else {
                if ( IsElfInLearnMode() && (w->parent->type == W_MAIN) )
                    return; /* ignore bar drag in learn mode */
                first_point = x;
                StrokingBar = w;
                PaintScrollBar(w,NULL,FALSE,TRUE);
                return;
            }
        }
        if ( doCallBack && w->callback )
            AxCallback(w->callback,w,w->callbacktag,&data);
    }
    else if ((eventP->type == MotionNotify) ||
             (eventP->type == ButtonRelease)){
        if ( w->scroll_type == XmVERTICAL ) {
            delta = y - last_y;
            total_delta = y - first_point;
        }
        else {
            delta = x - last_x;
            total_delta = x - first_point;
        }
        if ( !delta && (eventP->type == MotionNotify) )
            return;
        last_x = x;
        last_y = y;
        old_scroll_pos = w->scroll_pos;
        if (w->availpix) {
            w->scroll_pos =
                ((first_pix+total_delta)*w->scroll_max)/w->availpix;
        }
        else {
            w->scroll_pos = 0;
        }
        if ( w->scroll_pos + w->scroll_per > w->scroll_max )
            w->scroll_pos = w->scroll_max - w->scroll_per;
        if ( w->scroll_pos < 0 )
            w->scroll_pos = 0;
        if ( eventP->type == ButtonRelease )
            StrokingBar = NULL;
        PaintScrollBar(w,NULL,FALSE,TRUE);
        if ((eventP->type == ButtonRelease) && (w->scroll_pos != first_pos)) {
            data.reason = XmCR_VALUE_CHANGED;
            data.value = w->scroll_pos;
            if ( doCallBack && w->callback )
                AxCallback(w->callback,w,w->callbacktag,&data);
        }
        else if ( eventP->type == MotionNotify && !w->buttonUpScroll) {
/*          if ( (w->scroll_pos/10) != (old_scroll_pos/10) ) */
            if (w->scroll_pos != old_scroll_pos) {
                data.reason = XmCR_VALUE_CHANGED;
                data.value = w->scroll_pos;
                if ( doCallBack && w->callback )
                    AxCallback(w->callback,w,w->callbacktag,&data);
            }
        }
    }
}

/*************************************************************************
function: PaintScrollBar - Paint a vertical scroll bar
*/
void PaintScrollBar (w, exP, full_regen, layered)
register AXWidget_scroll w;
register XExposeEvent *exP;     /* NULL or exposure info */
int full_regen; /* TRUE if regen all */
int layered;
{
    int elevHeight, height;

    if ( w->scroll_type == XmVERTICAL )
        PaintVScrollBar(w, exP, full_regen, layered);
    else
        PaintHScrollBar(w, exP, full_regen, layered);

}

static void
PaintVScrollBar (w, exP, full_regen, layered)
register AXWidget_scroll w;
register XExposeEvent *exP;     /* NULL or exposure info */
int full_regen; /* TRUE if regen all */
int layered;
{
    int width, height;
    int new_window = FALSE;
    int use_scratchmap = FALSE;
    int clearHt, clearOffset;
    int elevHeight;
    Drawable drw;
    int three_d = (UseWidgetColors && !AxUseBigFont());
    GC tgc;
	int paint_elevator = TRUE;

    if ( !w->display )
        return;
	
    if (w->parent && (w->parent->type == W_DIALOG) &&
		!w->parent->window) {
        return;
	}

    if ((w->flags & WFLG_IN_EDIT) && layered) {
        if (dePaintWidget(w))
            return;
    }

    w->mapped = TRUE;

    if ( full_regen && exP && w->window && (exP->window != w->window) )
        return;

    if ( !w->window ) { /* only for scroll bars within dialogs */
        if (w->flags & WFLG_IN_EDIT) { /* under dialog editor */
            w->window = AxCreateQueryPixmap(Dpy, w->parent->window,
                            w->width, w->height, DpyDepth);
            XFillRectangle(Dpy, w->window, BackGC, 0, 0, w->width, w->height);
        }
        else {
			Window host_window;
			if (w->parent->type == W_COMBO_BOX)
				host_window = ((AXWidget_combo_box)w->parent)->pulldown;
			else
				host_window = WIN_OF(w);
            w->window = AxCrtWindow(Dpy,host_window,
                    w->xpos, w->ypos, w->width, w->height,
                    0,CopyFromParent,InputOutput,
                    CopyFromParent,WinMask,&WinAtt);
            XSelectInput(Dpy, w->window, TM_ALL_INPUTS_MASK);
            XMapWindow(Dpy,w->window);
            AddToWindowList(w);
            if (w->wapixmap)
                xmDestroyScrollPixmap(w);
            CrtScrollPixmap(w, w->width, w->height);
            ClearScrollPixmap(w, w->width, w->height, BackGC);
        }
        new_window = TRUE;
    } else {
        if (w->wapixmap) {
            if ((w->pixmapWidth  != w->width) ||
                (w->pixmapHeight != w->height)) {
                xmDestroyScrollPixmap(w);
                CrtScrollPixmap(w, w->width, w->height);
                ClearScrollPixmap(w, w->width, w->height, BackGC);
            }
        }
    }
    width = w->width;
    height = w->height;

    /* select drawable */
    if (w->wapixmap) {
        drw = w->wapixmap;
    }
    else if (!w->wapixmap && ScratchMap &&
             !(w->flags & WFLG_IN_EDIT)) {
        drw = ScratchMap;
        use_scratchmap = TRUE;
        full_regen = TRUE;  /* draw whole scrollbar */
    }
    else {
        drw = w->window;
    }

    if ( full_regen ) {
        tgc = three_d ? RecessGC : BackGC;
        XFillRectangle(Dpy,drw,tgc,0,0,width,height);
        DrawShadow(drw,ShadowGC,SunGC,2,0,0,width,height,
                   OLSHADOW_NODRAW);
        axmPaintTopArrow(w, drw, 0, 0, TRUE);
        axmPaintBottomArrow(w, drw, 0, w->height, TRUE);
    }

	if ( w->height <= (MIN_ELEV_PIX + 32) ) {
		paint_elevator = FALSE;   /* not enough room. 32 is 2x arrow height */
	}

    /* Apply scroll bar */
    
    if ( paint_elevator && !full_regen ) {    /* lift old elevator */
        tgc = three_d ? RecessGC : BackGC;
        XFillRectangle(Dpy,drw,tgc,2,width+w->pospix,
                       width-4,w->forepix);

    }
    
    w->availpix = height - 2*width;
    if ( (w->scroll_max == 0) || (w->scroll_max <= w->scroll_per) ) {
        w->forepix = w->availpix;
        w->pospix = 0;
    }
    else {
        /* 
          Pospix is massaged to make sure the scroll bar is not below
          the elevator area.
          */
        if (w->scroll_max) {
            w->forepix = (w->scroll_per * w->availpix)/w->scroll_max;
        }
        else {
            w->forepix = 0;
        }
        if ( w->forepix < MIN_ELEV_PIX )
            w->forepix = MIN_ELEV_PIX;
        if (w->scroll_max) {
            w->pospix = (w->scroll_pos * w->availpix)/w->scroll_max;
        }
        else {
            w->pospix = 0;
        }
        if ( w->pospix + w->forepix > w->availpix )
            w->pospix = w->availpix - w->forepix;
    }
#if STROKING_BAR
    if ( w == StrokingBar ) {
		if ( paint_elevator ) {
			DrawShadow(drw,ShadowGC,SunGC,2,
					   2,width+w->pospix,width-4,w->forepix,OLSHADOW_NODRAW);
		}
	}
    else {
#endif
		if ( paint_elevator ) {
			DrawShadow(drw,SunGC,ShadowGC,2,
					   2,width+w->pospix,width-4,w->forepix,OLSHADOW_NODRAW);
		}
#if STROKING_BAR
    }
#endif
    if ( paint_elevator ) {
        tgc = three_d ? BackGC : ForeGC;
        XFillRectangle(Dpy,drw,tgc,4,width+w->pospix+2,
                       width-8,w->forepix-4);
    }
    if (w->wapixmap) {
        XCopyArea(Dpy, w->wapixmap, w->window, BackGC, 0, 0,
                  w->pixmapWidth, w->pixmapHeight, 0, 0);
    }
    else if (use_scratchmap) {
        XCopyArea(Dpy, ScratchMap, w->window, BackGC, 0, 0,
                  w->width, w->height, 0, 0);
    }

    if (w->flags & WFLG_IN_EDIT) {
        XCopyArea(Dpy, w->window, w->parent->window, ForeGC, 0, 0,
                    w->width, w->height, w->xpos, w->ypos);
    }
}

static void
PaintHScrollBar (w, exP, full_regen, layered)
register AXWidget_scroll w;
register XExposeEvent *exP;
int full_regen; /* TRUE if regen all */
int layered;
{
    int width, height;
    int clearWid, clearOffset;
    int elevHeight;
    Drawable drw;
    GC tgc;
    int three_d = (UseWidgetColors && !AxUseBigFont());
    int use_scratchmap = FALSE;
	int paint_elevator = TRUE;

    if ( !w->display )
        return;

    if (w->parent && (w->parent->type == W_DIALOG) &&
		!w->parent->window) {
        return;
	}

    if ((w->flags & WFLG_IN_EDIT) && layered) {
        if (dePaintWidget(w))
            return;
    }

    w->mapped = TRUE;

    if ( !w->window ) { /* only for scroll bars within dialogs */
        if (w->flags & WFLG_IN_EDIT) { /* under dialog editor */
            w->window = AxCreateQueryPixmap(Dpy, w->parent->window,
                            w->width, w->height, DpyDepth);
            XFillRectangle(Dpy, w->window, BackGC, 0, 0, w->width, w->height);
        }
        else {
			Window host_window;
			if (w->parent->type == W_COMBO_BOX)
				host_window = ((AXWidget_combo_box)w->parent)->pulldown;
			else
				host_window = WIN_OF(w);
            w->window = AxCrtWindow(Dpy,host_window,
                    w->xpos, w->ypos, w->width, w->height,
                    0,CopyFromParent,InputOutput,
                    CopyFromParent,WinMask,&WinAtt);
            XSelectInput(Dpy, w->window, TM_ALL_INPUTS_MASK);
            XMapWindow(Dpy,w->window);
            AddToWindowList(w);
            if (w->wapixmap)
                xmDestroyScrollPixmap(w);
            CrtScrollPixmap(w, w->width, w->height);
            ClearScrollPixmap(w, w->width, w->height, BackGC);
        }
    } else {
        if (w->wapixmap) {
            if ((w->pixmapWidth  != w->width) ||
                (w->pixmapHeight != w->height)) {
                xmDestroyScrollPixmap(w);
                CrtScrollPixmap(w, w->width, w->height);
                ClearScrollPixmap(w, w->width, w->height, BackGC);
            }
        }
    }
    width = w->width;
    height = w->height;

    /* select drawable */
    if (w->wapixmap) {
        drw = w->wapixmap;
    }
    else if (!w->wapixmap && ScratchMap &&
             !(w->flags & WFLG_IN_EDIT)) {
        drw = ScratchMap;
        use_scratchmap = TRUE;
        full_regen = TRUE;  /* draw whole scrollbar */
    }
    else {
        drw = w->window;
    }

    if ( full_regen ) {
        tgc = three_d ? RecessGC : BackGC;
        XFillRectangle(Dpy,drw,tgc,0,0,width,height);
        DrawShadow(drw,ShadowGC,SunGC,2,0,0,width,height,
                   OLSHADOW_NODRAW);
        
        axmPaintLeftArrow(w, drw, 0, 0, TRUE);
        axmPaintRightArrow(w, drw, w->width, 0, TRUE);
    }
	
	if ( w->width <= (MIN_ELEV_PIX + 32) ) {
		paint_elevator = FALSE;  /* not enough room. (32 is 2x arrow wid) */
	}

    /* Apply scroll bar */
    
    if ( paint_elevator && !full_regen ) {    /* lift old elevator */
        tgc = three_d ? RecessGC : BackGC;
        XFillRectangle(Dpy,drw,tgc,
                       height+w->pospix,2,
                       w->forepix,height-4);
    }
    
    w->availpix = width - 2*height;
    if ( (w->scroll_max == 0) || (w->scroll_max <= w->scroll_per) ) {
        w->forepix = w->availpix;
        w->pospix = 0;
    }
    else {
        if (w->scroll_max) {
            w->forepix = (w->scroll_per * w->availpix)/w->scroll_max;
            w->pospix = (w->scroll_pos * w->availpix)/w->scroll_max;
        }
        else {
            w->forepix = w->pospix = 0;
        }
    }
	
    if ( w->forepix < MIN_ELEV_PIX )
        w->forepix = MIN_ELEV_PIX;    /* make sure always visible */

	if ( w->pospix + w->forepix > w->availpix )
		w->pospix = w->availpix - w->forepix;
#if STROKING_BAR
	if ( w == StrokingBar ) {
		if ( paint_elevator ) {
			DrawShadow(drw,ShadowGC,SunGC,2,
					   height+w->pospix,2,w->forepix,height-4,OLSHADOW_NODRAW);
		}
	}
	else {
#endif
		if ( paint_elevator ) {
			DrawShadow(drw,SunGC,ShadowGC,2,
					   height+w->pospix,2,w->forepix,height-4,OLSHADOW_NODRAW);
		}
#if STROKING_BAR
	}
#endif
		
	if ( paint_elevator ) {
		tgc = three_d ? BackGC : ForeGC;
		XFillRectangle(Dpy,drw,tgc,
					   height+w->pospix+2,4,
					   w->forepix-4,height-8);
	}

    if (w->wapixmap) {
        XCopyArea(Dpy, w->wapixmap, w->window, BackGC, 0, 0,
                  w->pixmapWidth, w->pixmapHeight, 0, 0);
    }
    else if (use_scratchmap) {
        XCopyArea(Dpy, ScratchMap, w->window, BackGC, 0, 0,
                  w->width, w->height, 0, 0);
    }
    if (w->flags & WFLG_IN_EDIT) {
        XCopyArea(Dpy, w->window, w->parent->window, ForeGC, 0, 0,
                    w->width, w->height, w->xpos, w->ypos);
    }
}

/*
 * Larger scrollbars will be OLSCRDISP_FULL.  As the size of the
 * bar decreases, the style changes to _ABBREV, or even to _MIN.
 */
static int
axolScrollDisplayStyle( width, height, orientation )
int width, height;
short orientation;
{
    int scrollDisp;

    if (orientation == XmVERTICAL) {
        if (height > OLSCRV_FULL_LIMIT) {
            scrollDisp = OLSCRDISP_FULL; /* full size bar, cable and anchors */
        } else if ( height > OLSCRV_ABBREV_LIMIT) {
            scrollDisp = OLSCRDISP_ABBREV; /* abbrev bar, anchors */
        } else {
            scrollDisp = OLSCRDISP_MIN; /* abbrev bar only */
        }
    } else {
        if (width > OLSCRH_FULL_LIMIT) {
            scrollDisp = OLSCRDISP_FULL; /* full size bar, cable and anchors */
        } else if ( width > OLSCRH_ABBREV_LIMIT) {
            scrollDisp = OLSCRDISP_ABBREV; /* abbrev bar, anchors */
        } else {
            scrollDisp = OLSCRDISP_MIN; /* abbrev bar only */
        }
    }
    return (scrollDisp);
}

/*************************************************************************
function: CrtScrollPixmap - Create a pixmap for scroll area.
*/
static void
CrtScrollPixmap(w, width, height)
register AXWidget_scroll w;
int width;
int height;
{
    unsigned int depth;
    int screen;
#if 1
    if ((width <= 0) || (height <= 0))
        goto badret;

    screen = DefaultScreen(Dpy);
    depth  = DefaultDepth(Dpy, screen); /* GCs also need to be of same depth */
    if (!(w->wapixmap = AxCreateQueryPixmap(Dpy, w->window,
                                            width, height, depth)))
        goto badret;
    w->pixmapWidth  = width;
    w->pixmapHeight = height;
    return;
badret:
    if (w->wapixmap)
        XFreePixmap(Dpy, w->wapixmap);
    w->wapixmap = (Pixmap) NULL;
    w->pixmapWidth = w->pixmapHeight = 0;
#else
    w->wapixmap = (Pixmap) NULL;
    w->pixmapWidth = w->pixmapHeight = 0;
#endif
    return;
}

/*************************************************************************
function: xmDestroyScrollPixmap - Free pixmap for scroll.
*/
void
xmDestroyScrollPixmap(w)
AXWidget_scroll w;
{
    if (!w->wapixmap) {
        w->pixmapWidth = w->pixmapHeight = 0;   /* sanity */
        return;
    }
    XFreePixmap(Dpy, w->wapixmap);
    w->wapixmap = (Pixmap) NULL;
    w->pixmapWidth = w->pixmapHeight = 0;
    return;
}

/*************************************************************************
function: CleanupScrollPixmap - cleanup a pixmap
*/
static void
ClearScrollPixmap(w, width, height, gc)
AXWidget_scroll w;
int width;
int height;
GC  gc;
{
    int screen;

    if (!w->wapixmap)
        return;
    if ((width <= 0) || (height <= 0))
        return;
    XFillRectangle(Dpy, w->wapixmap, gc, 0, 0, width, height);
    return;
}
