/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xmstring.c
   Applix ASC version: 63
   Date: Mon Dec 21 21:11:35 1998
*/
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <basicwin.h>

#if SUN4
#include <X11/Sunkeysym.h>
#endif
#if AX_X11
#include <X11/keysym.h>
#endif

#include <aos.h>
#include <xmlib.h>      /* contains widget_t def, OpenLook defs */
#include <wchars.h>
#include <xim.h>
#include <ximcomm.h>

#if HYUNDAI && HP800
#include <aiconids.h>
#endif

#include <getstr5.h>


#if BUILD_ORIENT && !AXMS    /*  {  */
#if USE_XIM
#include <locale.h>
#endif

extern XIC ICWhichHasFocus;

#if BUILD_KOREAN
#if OPENWIN
#define KOREAN_WIDE_BLANK 0x2121
#else
#define KOREAN_WIDE_BLANK 0xA1A1
#endif
#endif

#endif    /*  } BUILD_ORIENT */

int AXLookupString();
void TrueFree();
void *TrueMalloc();
int XLookupString();


extern char *ThimApplixDir;        /* Where the THIMBLE executable lies */
extern char *ThimCustomDir;        /* axlocal */
extern char *ThimUserDir;          /* $HOME/axhome */
extern win_list_t *TotalWindowsList;
extern Display *Dpy;
extern Cursor ThimPointerCursor;
extern Cursor ThimTextCursor;
extern char *DirSlash;

#if SUN4 && !USE_XIM && BUILD_KANJI
extern unsigned long   KanaMask;   /* defined in egcdintr.c */
#endif


int SizeMapTable = -1;
void MapKey();


typedef struct compose_table_struct_ {
    unsigned char seq[2];
    unsigned char value;
    } compose_table_struct;
static int ReadComposeTableFile();
#define COMPOSE_TABLE_SIZE 129   /* number of entries in the compose table */

/* Support 'dead' keys. On some keyboards, to enter a European accented
	character, user hits a 'dead' key (e.g. acute accent) and then
	the plain character (e.g. 'U') to get the accented character
	(thus U-acute). */
typedef struct dead_key_table_struct_ {
	KeySym deadKey;
	KeySym nextKey;
	unsigned char value;
    } dead_key_table_struct;
static int ReadDeadKeyTableFile();
static unsigned int read_int_or_hex();

/* In Latin1, only the high-bit chars can be 'European accented chars',
	so 129 entries is plenty.

	!!0 This is NOT TRUE for Unicode; there are more accented chars
	to be supported and the table will be longer !!0 */
#define DEAD_KEY_TABLE_SIZE 129   /* # of entries in the dead key table */


/*******************************************************************/
/* The wide-character version of the ever popular XLookupString        */
/* This RETURNS(!) a pointer to a buffer containing the character      */
/*   string, rather than placing the characters in a caller's supplied */
/*   buffer.  This is done so that off-the-spot mechanisms can return  */
/*   arbitrarily long strings of characters                            */
/* This returns the number of characters (not bytes) in the returned   */
/*   character buffer                                                  */
/***********************************************************************/

#define TEMP_LEN 2000       /* plenty big, and on the stack */

int AXLookupString16(eventP, buf, ksymP)
    XEvent *eventP;     /* KeyPress or KeyRelease */
    wchar **buf;        /* RETURN, pointer to a static buffer */
    KeySym *ksymP;      /* RETURN */
{

    mbchar temp_buff[TEMP_LEN];
    static wchar *static_buff = (wchar *) NULL; /* where the data gets
                             * put */
    static int static_buf_len = -1;
    int byte_len, char_len;

    byte_len = AXLookupString(eventP, temp_buff, TEMP_LEN, ksymP);
    temp_buff[byte_len] = '\0';

    char_len = (byte_len + 1) * sizeof(wchar);  /* worst case, each byte
                             * is char */
    if (static_buf_len < char_len)
        {
        /** need a bigger static buffer */
        if (static_buff != (wchar *) NULL)
            TrueFree((char *) static_buff);
        static_buff = (wchar *) TrueMalloc(char_len);
        static_buf_len = char_len;
        }

    char_len = MbToWStr(temp_buff, static_buff);

    *buf = static_buff; /* return pointer to the characters */
    return (char_len);  /* return the number of characters */
}

static Time lastTime;   /* timestamp from last event processed */
static int lastType;    /* event type of last event processed */
static unsigned int lastCode;   /* keycode from last event */
static mbchar lastBuf[XLS_BUF_LEN + 1]; /* string buffer from last event */
static unsigned long lastSerial;

static KeySym lastKsym; /* keysym from last event processed */
static int lastNbytes;  /* length of string buffer from last event */
static XComposeStatus compose;


#if SUN4 && !USE_XIM && BUILD_ORIENT
SetXLookupStaticString( str, keysym )
wchar *str;
KeySym keysym;
{
    lastNbytes = WToMbStr( str, lastBuf );
    lastKsym = keysym;
}
#endif




/*************************************************************************
function: AXLookupString - Wrapper for XLookupString to avoid too many
Xlib calls for same event.  (This is not really an attempt at optimization,
but a work-around for a bug where multiple calls to XLookupString for the
same event loses compose key sequence information.)
*/
int AXLookupString(eventP, buf, len, ksymP)
    XEvent *eventP;     /* KeyPress or KeyRelease */
    mbchar *buf;        /* RETURN caller-supplied buffer */
    int len;        /* length of caller-suplied buffer in bytes, not chars */
    KeySym *ksymP;      /* RETURN */
{
    int ret_val;
    XKeyEvent *kep;
    int status;
    mbchar *p, *p2;
    static int compose_state = -1;		/* needs initialization */
    static KeySym compose_key = 0;
    static unsigned char compose_seq[2];
    static compose_table_struct compose_table[COMPOSE_TABLE_SIZE];
    static int dead_key_state = -1;		/* needs initialization */
    static dead_key_table_struct dead_key_table[DEAD_KEY_TABLE_SIZE];
	static KeySym first_dead_key, last_dead_key;
	static KeySym dead_key=0;
	static dead_key_inx;
    unsigned char real_char;
    KeySym real_keysym;
    int ix, map_index;

#if HYUNDAI && HP800
   AXWidget main_w;
   int syw_i;
   unsigned char tmp_buf;
#endif

#if RIOS
    static int use_xlookup = -1;
    int i, state_index;
    char keyname[40], state_str[80];
    mbchar *map_str;
    unsigned int save_keystate;
    char *ksname;
    static int debug_display = -1;
 
    if ( debug_display == -1 )
        {
        /* toggle a runtime flage to dump keystrokes */
        if ( getenv(NXLT("RS6000_DEBUG")) != (char *)NULL )
            debug_display = 1;
        else
            debug_display = 0;
        }
#endif


    if (!eventP || ((eventP->type != KeyPress) &&
            (eventP->type != KeyRelease)))
        {
        if (ksymP)
            *ksymP = 0;
        return (0);
        }

    kep = (XKeyEvent *)eventP;

    /* We set the serial field to a uniquly incrementec value in the */
    /*   function AXtAppNextEvent() so we can tell when we are processing */
    /*   a new event, as opposed to being handed a packet by an application */
    /*   which was already looked at by the widget code.  This is all done
    /*   because we do not want to call XLookupString again with the same */
    /*   event as that can screw up compose key processing, and really  */
    /*   screws up the orient input mechanism (XIM) processing - jon    */
    if ( eventP->xany.serial != lastSerial )
        {
        /* new event */
        lastSerial = eventP->xany.serial;
        lastBuf[0] = 0;

#if HP800  && !BUILD_ORIENT  /* { */
        /** On HP machines, in order to get access to the compose key **/
        /** processing, one must call the special HP routine, instead of **/
        /** the normal XLookupString.  We pass a zero as the last argument **/
        /** which specifies that we want the results in the LATIN1 char **/
        /** set, instead of Roman8.  This shouldbe done on the HP-60000 **/
        /** line as well, but it requires HP/UX 8.0 or later, and we must **/
        /** still build the HP 60000 line on 7.x   - jon                **/
        lastNbytes = XLookupString((XKeyEvent *) eventP, lastBuf,
                     XLS_BUF_LEN, &lastKsym, &compose);
        
#endif      /*  } **/

#if RIOS    /*  {  */
      if ( use_xlookup == -1 )
        {
        if ( getenv("AX_DO_RS6000_COMPOSE") != (char *)NULL )
            use_xlookup = FALSE;
        else
            use_xlookup = TRUE;
        }
     if ( use_xlookup )
        {
        /* The generic case, use the standard XLookupString() call   */
        /* This will not handle compose keys from the IBM european */
        /*   keyboards.  To use those you must set "AX_DO_RS6000_COMPOSE" */
        lastNbytes = XLookupString((XKeyEvent *) eventP, lastBuf,
                       XLS_BUF_LEN, &lastKsym, &compose);
        }
    else
        {
        /** on the RS-6000, I must use XLookupMapping to get a string **/
        /** since that is the way composed keys get processed         **/
        /** Since XLookupMapping is not a 1-for- replacement of       **/
        /** XLookupString, we must get the keysym seperately - jon     **/
        kep = (XKeyEvent *) eventP;
        save_keystate = kep->state;

        /** we need to clear any modifier flags, since this wonderful **/
        /** IBM routine does not return a string for an alpha character **/
        /** if "Mod1" is on, which hoses the Motif accellerators - jon **/
        kep->state = kep->state &
                       ~(Mod1Mask | Mod3Mask |  Mod4Mask | ControlMask );
        map_str = XLookupMapping((XKeyEvent *) eventP, &lastNbytes);
        kep->state = save_keystate;
        if ( map_str != (char *)NULL )
            *(map_str+lastNbytes) = '\0';

        if ( (debug_display == 1) && (eventP->type == KeyPress) )
            {
            printf(NXLT("-------------------------------------------------------\n"));
            printf(NXLT("XLookupMapping returns %d bytes\n"), lastNbytes );
            if ( lastNbytes > 0 )
               {
               printf(NXLT("string returned = <%s>   <"), map_str);
 
               for ( i=0; i<lastNbytes; i++ )
                  {
                  printf(NXLT(" 0x%x "), (int)(*(map_str+i)) );
                  }
               printf(">\n");
               }
            }


        if (lastNbytes == 0)
            {
            /** We set map_str to null because the RS6000 will still return a */
            /**  character pointer from XLookupMapping (the character from    */
            /**  a previous call) if there is not string for this character */
            map_str = (mbchar *) NULL;
            lastBuf[0] = '\0';
            }
        else
            {
            /* a real string */
            strcpy(lastBuf, map_str);
            }

        state_index = 0;/** assume not shifted **/
        if ((kep->state & ShiftMask) && (lastNbytes > 0) &&
            (lastBuf[0] != '\t') && (lastBuf[0] != '\n') &&
            (lastBuf[0] != ' ') )
            {
            /** we only want to pass the shift indicator for key events */
            /** which result in printable characters.  This is because, */
            /** once again, the RS6000sees fit to respond to a shifted  */
            /** function key event with a keysym of "unknown key"       */
            /** You can only get XK_F3 for the unshifted event, BUT ... */
            /** if you call this with a tab or return ( both which have */
            /** the buff length = 1) with the shift bit on you will get */
            /** back "unknown key", which breaks <shift-tab> and        */
            /** <shift-return>  (what a piece of crap)-jon              */
            /** More crap - also broken for <shift><space> - jon        */
            state_index = LockMapIndex; /* shift is in effect */
            }

        lastKsym = XLookupKeysym((XKeyEvent *) eventP, state_index);

        if ( (debug_display == 1) && (eventP->type == KeyPress) )
            {
            ksname = XKeysymToString (lastKsym);
            state_str[0] = '\0';
            if ( save_keystate & ControlMask )
                strcat( state_str, NXLT("Control") );
            if ( save_keystate & ShiftMask )
                strcat( state_str, NXLT(" Shift") );
            if ( save_keystate & LockMask )
                strcat( state_str, NXLT(" Lock") );
            if ( save_keystate & Mod1Mask )
                strcat( state_str, NXLT(" Mod1") );
            if ( save_keystate & Mod2Mask )
                strcat( state_str, NXLT(" Mod2") );
            if ( save_keystate & Mod3Mask )
                strcat( state_str, NXLT(" Mod3") );
            if ( save_keystate & Mod4Mask )
                strcat( state_str, NXLT(" Mod4") );
            if ( save_keystate & Mod5Mask )
                strcat( state_str, NXLT(" Mod5") );
 
            if ( ksname != (char *)0 )
                strcpy( keyname, ksname );
            else
                strcpy( keyname, XLT("Unknown key name") );
 
            printf (NXLT("Key is XK_%s, keycode %d (0x%x), keysym=0x%lx,\n\t\tstate=(%s)\n"),keyname, kep->keycode, kep->keycode, lastKsym, state_str);
           }

        if ( (lastKsym >= 0x18000000) && (lastNbytes > 0) )
           {
           /* the keysym reflects a "dead" key like 'XK_dead_grave' but */
           /* there is a real string.  This must be a case of pressing */
           /* <alt><XK_dead_grave> to get a real string.  Fix things up */
           /* so this is transparent to the upper levels by making the    */
           /* keysym the value of the first char in the buffer, since for */
           /* printable chars, their XK values are the char values */
           /* What a hack to deal with IBM crap!                   */
           lastKsym = lastBuf[0];
           }
 
        /* finally, strip the MOD2 and MOD5 bits used by IBM to drive */
        /* conversion so the upper levels do not see these            */
        kep->state = kep->state & ~(Mod2Mask |  Mod5Mask);
        }

#endif   /** }  RIOS   */

#if (!RIOS && !HP800) || BUILD_ORIENT    /*  {  */
        /* most platforms */
#if BUILD_ORIENT && !AXMS  /*  {  */
        if ( ICWhichHasFocus != (XIC)NULL )
           {
#if HP800 && !HYUNDAI /* If this module is add, when we type ENG key,
                  the last hangul data at the buffer is lost - 1994.8.4. SYW */
           if ( eventP->type == KeyRelease )
              {
              /* we must make sure that we do not pass the key-release event */
              /* for the conversion on/off key to XmbLookupString since */
              /* this function is too stupid to realize that this is a key */
              /* release event, and turns conversion off !! (crap from HP */
              lastKsym = XLookupKeysym( eventP, 0 );
              if ( (lastKsym == XK_Meta_L) || (lastKsym == XK_Meta_R)  ||
                   (lastKsym == XK_Henkan_Mode) )
                 {
                 /* this is key-release of the on-off key, bail out */
                 lastNbytes = 0;
                 lastBuf[0] = (mbchar)0;
                 lastKsym = XK_Shift_L;   /* an inocuous value */
                 goto end_of_lookup;
                 }
              }
#endif

#if HP800 && BUILD_KANJI
        /* Ashisuto added for the function XmbLookupString() only accepts
         * the KeyPress events as processible events while the XLookupString()
         * accepts both the KeyPress and KeyRelease events. The needs has
         * arised from the operation of "Press the key continously in some
         * dialog box's entry field, on HP platform", the unpredicable chars
         * will be returned after an abusing of function XmbLookupString().
         * 96.01.30
         */
        if (eventP->type == KeyRelease) {
            lastNbytes = XLookupString(eventP, lastBuf, XLS_BUF_LEN, 
                                                              &lastKsym, NULL);
            status = XLookupBoth;
        } else {
           lastNbytes = XmbLookupString( ICWhichHasFocus, eventP, lastBuf,
                       XLS_BUF_LEN, &lastKsym, &status );
           if (lastNbytes == XLS_BUF_LEN) 
              {
              /* HP-UX HACK !!! 9.x or 10.x returns garbages when user do
               * very fast key hitting! Checked all parameters of
               * XmbLookupString()'s and they were all OK!
               * Not problem of Applixware. - Ashisuto, 1997.02.25.
               */
              lastNbytes = 0;
              lastBuf[0] = (mbchar)0;
              lastKsym = XK_Shift_L;   /* an inocuous value */
              goto end_of_lookup;
              }
           }
#else
           lastNbytes = XmbLookupString( ICWhichHasFocus, eventP, lastBuf,
                       XLS_BUF_LEN, &lastKsym, &status );
#endif
 

#if HP800 && !HYUNDAI /* If this module is add, when we type ENG key,
                  the last hangul data at the buffer is lost - 1994.8.4. SYW */
           if ( (lastKsym == XK_Meta_L) || (lastKsym == XK_Meta_R)  ||
                   (lastKsym == XK_Henkan_Mode) || (lastNbytes < 0) )
              {
              /* this is  the on-off key, bail out */
              lastNbytes = 0;
              lastBuf[0] = (mbchar)0;
              lastKsym = XK_Shift_L;   /* an inocuous value */
              goto end_of_lookup;
              }
#endif
           if ( (status != XLookupChars) && (status != XLookupBoth) )
              {
              /* no string returned */
              lastNbytes = 0;
              lastBuf[0] = (mbchar)0;
              }
           if ( lastNbytes > 1 )
              {
              /* more than one byte is a sure sign of oriental string input */
              /* make sure that the keysym is a "printable" quantity since */
              /* if the conversion server is set up to use Control-K as the */
              /* 'fix' key, we would return keysym=Control and a multi-byte */
              /* string which really confuses the upper levels          */
#if BUILD_KANJI
           /* Ashisuto added to JAPANESE build.   95.09.14 */
           lastBuf[lastNbytes] = 0;
           if (lang_is_sjis()) {
               /* Here shift-jis returned, convert to Applixware inner code */
               unsigned char temp_buff[XLS_BUF_LEN+1];
               strcpy(temp_buff, lastBuf);
               MBsjis2euc_str(temp_buff);  /* shift-jis to euc func */
               strcpy(lastBuf, temp_buff);
               lastNbytes = strlen(lastBuf);
           }
#endif
#if HP800 && !HYUNDAI
              {
              p = lastBuf + lastNbytes - 1;
              p2 = lastBuf;
              if ( IS_MULTI_BYTE( p2 ) && (*p == 13) )
                 {
                 /* the stupid HP input mechanism returns a "carriage return" */
                 /* used to commit the input, as part of the string.  So */
                 /* we take it off                                       */
                 *p = 0;
                 lastNbytes--;
                 }
              /* also we reset the spot location */
              ResetSpotLocation( eventP->xany.window );
              }
#endif
#if SUN4 && !USE_XIM && BUILD_KANJI
              /* we'll use the key state to decide if the "kana mode" is on,
               * it may not be set here, especially in the case of SUN4,
               * BUILD_KANJI. At least we wish the "kana mode" on/off state
               * be preserved.  Ashisuto edited. Jun 25, 94. */
              eventP->xkey.state = eventP->xkey.state & ~KanaMask;
#else
              eventP->xkey.state = 0;
#endif
              lastKsym = XK_A;
              }
#if BUILD_KANJI
              else if ( lang_is_sjis() && lastNbytes==1 && lastBuf[0]&0x80 )
              {
                   /* Ashisuto added for JAPANESE build.
                    * This is the Shift-JIS Katakana single byte char,
                    * should be converted to Applixware internal EUC code.
                    * 1997.12.15
                    */
                   unsigned char temp_buff[3];
                   lastBuf[lastNbytes] = 0;
                   strcpy(temp_buff, lastBuf);
                   MBsjis2euc_str(temp_buff);  /* shift-jis to euc func */
                   strcpy(lastBuf, temp_buff);
                   lastNbytes = strlen(lastBuf);
              }
#endif
           }
        else
#endif   /* }  BUILD_ORIENT */

        lastNbytes = XLookupString((XKeyEvent *) eventP, lastBuf,
                       XLS_BUF_LEN, &lastKsym, &compose);
        lastBuf[lastNbytes] = 0;
#if 0
printf("ksym = %d (0x%x)\n", lastKsym, lastKsym);
#endif

#if SUN4 || SUNIS  /** {  SUN HACK */
        /* first of all, the "F11" and "F12" keys do not return the keysym */
        /* one would expect.  Instead they return some wierd sun propietary */
        /* keysyms!   Map these wierd things to something normal            */
#ifdef SunXK_F36
        if ( (lastKsym == SunXK_F36) || (lastKsym == 0x1000FF10) )
            /* 0x1000FF10 is "XK_SunF36" on SunOS 4.1.2 */
            lastKsym = XK_F11;
        else if ( (lastKsym == SunXK_F37) || (lastKsym == 0x1000FF11) )
            /* 0x1000FF11 is "XK_SunF37" on SunOS 4.1.2 */
            lastKsym = XK_F12;
#endif
#ifdef XK_SunF36
        /* this is being built on SunOS 4.1.2 or earlier */
        if ( (lastKsym == XK_SunF36) || (lastKsym == 0x1005FF10) )
           lastKsym = XK_F11;
        else if ( (lastKsym == XK_SunF37) || (lastKsym == 0x1005FF11) )
           lastKsym = XK_F12;
#endif

        kep = (XKeyEvent *)eventP;
#if SUN4 && !USE_XIM && BUILD_KANJI
        /*
         * sometimes Mod3Mask used as status mask of XK_Mode_switch key,
         * which is important in countries with varies input modes. So
         * here we separate it out from num-lock status mask, here it is
         * the KanaMask, which we had already set it in
         * XmbLookupString().   Ashisuto added. 94/06/21
         */
        if ((kep->state & ((Mod3Mask | Mod4Mask) & ~KanaMask)) &&
              (lastBuf[0] == 0) &&
              (((lastKsym >= XK_F24) && (lastKsym <= XK_F35)) ||
               ((lastKsym >= XK_Home) && (lastKsym <= XK_KP_Divide))) )
#else
        if ((kep->state & (Mod3Mask | Mod4Mask)) && (lastBuf[0] == 0) &&
             (((lastKsym >= XK_F24) && (lastKsym <= XK_F35)) ||
              ((lastKsym >= XK_Home) && (lastKsym <= XK_KP_Divide))) )
#endif
            {
            /** here comes a real hack for Open Windows 3. They broke */
            /** the num-lock key so we no longer get back digits.  */
            /** Instead you get a function key code and a mod3 or mod4*/
            /**  state bit and no keystring.  So here we fudge it to */
            /** make it look like OpenWindow 2 which worked fine    */

            /* the keymapping which exists on the sun systems I have seen */
            /* (run 'xmodmap -pke' to see it) has the actual keypad keysym */
            /* in the "control" position (3'd key across in the keymap table */
            /* for example, the standard mapping for the "4" key on the  */
            /* numeric keypad looks like:                                */
            /*     "Left F30 KP_4"                                       */
            /* On this basis, even though the num-lock key sets MOD3     */
            /*   or MOD4 (depending on the keyboard type and whether     */
            /*   you are runing against a remote machine running a different */
            /*   version of OpenWindows.  Under the normal OpenWindows-3 */
            /*   situation MOD3 is set.) we will look at the "control"   */
            /*   postion to get the 'real' keypad keysym                 */
            real_char = '\0';
            map_index = ControlMapIndex;
            real_keysym = XLookupKeysym( kep, map_index );
            if ( real_keysym != 0 )
               /* found it */
               lastKsym = real_keysym;
            }
#if SUN4 && !USE_XIM && BUILD_KANJI
        /* with the same reason above, Ashisuto edited. */
        else if ( (lastKsym == XK_Delete) &&
                   (kep->state & ((Mod3Mask | Mod4Mask) & ~KanaMask)) )
#else
        else if ( (lastKsym == XK_Delete) &&
                  (kep->state & (Mod3Mask | Mod4Mask)) )
#endif
        {
        /* Ashisuto added the more delicate judgement for SunOS platforms:
         * This bug is fixed from Solaris-2.5(both SPARC and x86), there
         * the keysym returned is the right symbol KP_Decimal when Num_Lock
         * is on. The another reason we must change it here is that if the
         * Num_Lock is on, another key labeled "Delete" which is not
         * on the Keypad when pressed will become the period char, this
         * is the bug belonged to Applixware. We correct it here.
         * 1996.07.19
         */
        for (ix = 0; XKeycodeToKeysym(Dpy, kep->keycode, ix) != NoSymbol; ix++);
        if (ix > 1) 
            {	
            /* The keysym of Del key on keypad is not unique! */
            /* special case for the "Del" key on the keypad because the */
            /* keystring returned when the 'Del'key is pressed, even with */
            /* the num-lock key down, is not null */
            int use_comma;
            lastKsym = XK_KP_Decimal;
            use_comma = UPnumVal(PROF_NUMBERS_);
            if ( use_comma )
               lastBuf[0] = ',';
            else
               lastBuf[0] = '.';
            lastBuf[1] = '\0';
            lastNbytes = 1;
            }
        }


#endif    /** }  SUN HACK */

#if LINUX
#ifdef XK_ISO_Left_Tab /* XFree86 makes S-Tab look like S-ISO_Left_Tab */
		if (lastKsym == XK_ISO_Left_Tab)
			lastKsym = XK_Tab;
#endif
#endif /* LINUX */
		
#if BUILD_KANJI
       /* Ashisuto added because if conversion server is "canna", the
        * <RETURN> keysym would not be set correctly.  95.10.16 */
       if (lastNbytes == 1 && lastBuf[0] == '\n')
           lastKsym = XK_Return;
#endif

                      
#endif /*  }   (!RIOS && !HP800) || BUILD_ORIENT ie. most platforms */


       if ( (lastKsym >= XK_KP_Multiply) && (lastKsym <= XK_KP_Equal) &&
               (lastBuf[0] == 0) )
          {
          /* we have a keypad keysym but no string.  There seem to be */
          /* at least a couple of servers which do this when the num-lock */
          /* key is down.  All we need to do is fabricate the correct string */
          /* (based on the assumption that the KP keysyms are contiguous */
          /*      and in order!)                                         */
          lastBuf[0] = lastKsym & 0x3F;
          lastBuf[1] = 0;
          lastNbytes = 1;
          }
       else if ( (lastKsym >= XK_KP_Space) && (lastKsym < XK_KP_Equal ) &&
              (lastBuf[0] == 0) )
          {
          /* make sure we have a printable string for the printable */
          /* keys in this range                                     */
          lastNbytes = 1;
          if ( lastKsym == XK_KP_Space )
             lastBuf[0] = ' ';
          else if ( lastKsym == XK_KP_Tab )
             lastBuf[0] = '\t';
          else
             {
             /* not one of the printable chars */
             lastNbytes = 0;
             lastBuf[0] = 0;
             }
          lastBuf[1] = 0;
          }
       else if ( lastKsym == XK_KP_Separator )
          {
          /* dot or comma as appropriate for the language */
          int use_comma;
          use_comma = UPnumVal(PROF_NUMBERS_);
          if ( use_comma )
             lastBuf[0] = ',';
          else
             lastBuf[0] = '.';
          lastBuf[1] = 0;
          }
#if 0
       /* we do allow the user to treat the "enter" key on the numeric */
       /* keypad differently from the "return" key.  Taking this out   */
       /* may cause some problem down the road, but leaving it in causes */
       /* bug #18296 - jon                                                 */
       else if ( lastKsym == XK_KP_Enter )
          {
          lastKsym = XK_Return;
          lastBuf[0] = 0x0D;
          }
#endif

       


            
#if AX_X11     /*   {   */
        /** The following section is in place to implement a compose key */
        /** processing functionality, for those platforms which do not */
        /** doirectly supply compose key access to European chars      */
        if ( compose_state != 0  )
            {
            if ( compose_state == -1 )
               {
               /* determine if we should do our own compose key processing, */
               /* and if so, attempt to load the compose table     */
               compose_state = 0;   /* assume no compose */
               if ( getenv("AX_LOCAL_COMPOSE") != (char *)NULL )
                   {
                   compose_key = atoi( getenv("AX_LOCAL_COMPOSE") );
                   if ( compose_key <= 256 )
                      /* funny value, use the default */
                      compose_key = XK_Multi_key;
       
                   /* we are to do our own compose handling */
                   if ( ReadComposeTableFile( compose_table ) )
                      compose_state = 1;   /* enable base leve compose state */
                   else
                      /* error - bail out */
                      goto end_compose;
                   }
               }

            if ( eventP->type == KeyPress )
               {
               /* Do our own local compose key processing. This is driven */
               /* by a key event (which the user can define via the       */
               /* AX_LOCAL_COMPOSE environemnt variable).   We keep track of */
               /* the state and two subsequent characters, and return a null */
               /* event until we get the composed char, which is then returned*/
               if ( lastKsym == compose_key )
                  {
                  /* start of compose sequence */
                  compose_state = 2;
                  lastBuf[0] = 0;
                  lastNbytes = 0;
                  }
               else if ( lastBuf[0] != 0  )
                  {
                  /* we want only "printable" keystroke events */
                  if ( compose_state == 2 )
                     {
                     /* this should be the first char of a compose sequence */
                     /* suck up this keystroke */
                     compose_seq[0] = lastBuf[0];
   ignore_key:
                     lastBuf[0] = 0;
                     lastNbytes = 0;
                     lastKsym = XK_Shift_L;   /* an inocuous value */
                     compose_state = 3;
                     }
                  else if ( compose_state == 3 )
                     {
                     /* this should complete the compose key processing */
                     compose_seq[1] = lastBuf[0];
                     lastBuf[0] = 0;
                     lastNbytes = 0;
   
                     /* look it up */
                     for ( ix=0; ix < COMPOSE_TABLE_SIZE; ix++ )
                        {
                        if ( compose_table[ix].value == 0 )
                            {
                            /* end of table, did not find what we want, reset */
                            /* and return an invalid key to get a beep */
                            lastBuf[0] = '\0';
                            lastNbytes = 0;
                            lastKsym = 0;
                            break;
                            }
                        /* check the table entry.  We allow the user to enter */
                        /*  the 2 compose chars in either order          */
                        if ( ((compose_table[ix].seq[0] == compose_seq[0]) &&
                              (compose_table[ix].seq[1] == compose_seq[1])) ||
                             ((compose_table[ix].seq[0] == compose_seq[1]) &&
                              (compose_table[ix].seq[1] == compose_seq[0])) )
                            {
                            /* found it, return the composed char */
                            lastBuf[0] = compose_table[ix].value;
                            lastNbytes = 1;
                            lastKsym = lastBuf[0];  /* 8859 chars this is OK */
                            break;
                            }
                        }
                     compose_state = 1;
                     }     /* end of compose_state == 2 */
   
                  }    /* end of "printable character */
               }       /* end of KeyPress event   */
   
            }  /* end of compose_state != 0 */
end_compose:

		if ( dead_key_state != 0  ) {			/* zero -> no dead keys */
			if ( dead_key_state == -1 ) {
				/* Determine if we should do our own dead-key processing,
					and if so, attempt to load the dead-key table.
					Dead-key processing can be inhibited by removing
					the file, making it empty, or setting an
					environment variable. */
				dead_key_state = 0;   			/* assume no dead keys */ 
				if ( getenv("AX_NO_DEAD_KEYS") != (char *)NULL )
					goto end_dead_keys;

				if ( ReadDeadKeyTableFile( dead_key_table, 
									&first_dead_key, &last_dead_key) )
					dead_key_state = 1;			/* Ready to process */
				else
					goto end_dead_keys;
			}

			/* Ready to process dead keys. See if current keypress is
				a dead key. If so, remember what it is, and replace
				it with a no-op key */
            if ( eventP->type == KeyPress ) {
				if (dead_key_state == 1) {
					if (lastKsym && (lastKsym >= first_dead_key) &&
								(lastKsym <= last_dead_key)) {
						for (ix=0; ix < DEAD_KEY_TABLE_SIZE; ix++) {
	
							/* If this dead key is in the table, we remember
								it, but "hide" it from the caller */
							if (dead_key_table[ix].deadKey == lastKsym) {
								dead_key = lastKsym;
								dead_key_inx = ix;		/* for performance */
								lastBuf[0] = 0;
								lastNbytes = 0;
								lastKsym = 0;
								dead_key_state = 2;
							}
	
							/* If we run out of entries, this is not a dead key
								and needs to be processed normally */
							else if ((dead_key_table[ix].deadKey > lastKsym) ||
										(dead_key_table[ix].value == 0)) {
								break;
							}
						}
					}
				}
	
				/* Prior key was a dead key. Given a new printable key, see 
					if the dead-key/new-key combo is in the table. If so, 
					substitute the table value for the new key. If not, 
					leave the new key alone and drop out of the dead_key 
					state. If this is NOT a printable key (e.g. a Shift)
					leave it alone but do not drop out of the dead-key
					state. */
				else if ((dead_key_state == 2) && (lastBuf[0] != 0)) {
					for (ix=dead_key_inx; ix < DEAD_KEY_TABLE_SIZE; ix++) {
						if (dead_key_table[ix].deadKey == dead_key) {
							if (dead_key_table[ix].nextKey == lastKsym) {
								/* Found a match!! */
								lastBuf[0] = dead_key_table[ix].value;
								lastBuf[1] = '\0';
								/* For Latin1 chars, XK_ value == ascii value */
								lastKsym = (KeySym)lastBuf[0];
								break;
							}
							else if (dead_key_table[ix].nextKey > lastKsym)
								break;
						}
						else {
							/* The current keystroke does not pair with the
								dead key to produce anything. */
							break;
						}
					}
	
					/* Reset the dead-key state */
					dead_key_state = 1;
					dead_key = 0;
				}
			} /* keypress */
		} /* doing dead keys */

end_dead_keys:
        /* the following section is in place to allow a user to map one */
        /* key to another key.  Only the keystring is changed           */
        if ( SizeMapTable && (eventP->type == KeyPress) )
            {
            MapKey( lastBuf, kep );
            }

#endif    /* } AX_X11  */

        }           /* end of "if this is the same event as last call */

end_of_lookup:

	if (XK_KP_Decimal == lastKsym) {
		if (UPnumVal(PROF_NUMBERS_) && (MIN(len, lastNbytes) >= 1)) {
			lastBuf[0] = ',';  /* use comma as decimal sep, per preference */
		}
	}

    if (buf)
        memcpy(buf, lastBuf, MIN(len, lastNbytes));
    if (ksymP)
        *ksymP = lastKsym;
	
    return (MIN(len, lastNbytes));
}

static int ReadComposeTableFile( table )
compose_table_struct table[];
{
    FILE *ff;
    unsigned int char1, char2, value;
    int ix;
    char buff[201];
    char filename[ STRPATH+1 ];

    sprintf( filename, "%s/%s", ThimApplixDir, "compose.sp" );

    ff = fopen( filename, "r" );
    if ( ff == (FILE *)NULL )
       /* no compose processing */
       return( FALSE );

    ix = 0;
    while ( fgets( buff, 200, ff) != (char *)NULL )
       {
       if ( buff[0] == '#' ) 
           /* comment line */
           continue;
       char1 = char2 = value = 0;
       char1 = buff[0];
       char2 = buff[2];
       if ( buff[4] > '9' )
           {
           /* an alpha hex digit */
           buff[4] = toupper( buff[4] );
           value = 10 + (buff[4] - 'A');
           }
       else
           value = buff[4] - '0';
       value = value * 16;

       if ( buff[5] > '9' )
           {
           /* an alpha hex digit */
           buff[5] = toupper( buff[5] );
           value = value + (10 + (buff[5] - 'A'));
           }
       else
           value = value + (buff[5] - '0');

       if ( (char1 < ' ') || (char2 < ' ') || (value == 0) )
           /* quit on the first error */
           break;

       table[ix].seq[0] = char1;
       table[ix].seq[1] = char2;
       table[ix].value = value;
       ix++;
       }
    table[ix].seq[0] = 0;
    table[ix].seq[1] = 0;
    table[ix].value = 0;

    if ( ix > 0 )
       /* OK */
       return( TRUE );
    else
       return( FALSE );
}

/* 'Dead' accent key + letter -> European accented letter */
static int ReadDeadKeyTableFile( table, first_dead_key, last_dead_key )
dead_key_table_struct table[];
KeySym *first_dead_key, *last_dead_key;
{
    FILE *ff;
    KeySym char1, char2;
    unsigned int value;
    int ix;
    char buff[201];
    char filename[ STRPATH+1 ];
	char *p;
	int nEntries=0;

	/* Look in $HOME/axhome */
    sprintf( filename, "%s%s%s", ThimUserDir, DirSlash, NXLT("deadkeys.dat"));
    ff = fopen( filename, "r" );
    if ( ff != (FILE *)NULL )
		goto read_file;

	/* Look in axlocal */
    sprintf( filename, "%s%s%s", ThimCustomDir, DirSlash, NXLT("deadkeys.dat"));
    ff = fopen( filename, "r" );
    if ( ff != (FILE *)NULL )
		goto read_file;

	/* Look in axdata */
    sprintf( filename, "%s%s%s", ThimApplixDir, DirSlash, NXLT("deadkeys.dat"));
    ff = fopen( filename, "r" );
    if ( ff == (FILE *)NULL )
		/* no dead-key processing */
		return( FALSE );

read_file:
	*first_dead_key = 0xFFFF;
	*last_dead_key = 0;
    ix = 0;
    while ( fgets( buff, 200, ff) != (char *)NULL ) {
		if (nEntries >= DEAD_KEY_TABLE_SIZE)
			break;
		if ( buff[0] == '#' ) 
			/* comment line */ 
			continue;

			/* File format:
				accent_code  letter_code  result_value
			   where:
				accent_code = XK_ value of dead key, in integer or hex;
					thus 65104 or 0xFE50 for XK_dead_grave
				letter_code = XK_ value of letter key, in integer or hex;
					thus 65 or 0xA1 for 'A'
				result_value = Latin1 value of resulting accented letter,
					in interger or hex; thus 192 or 0xC0 for A-grave.
					Note for values < 256, XK_value = Latin1 value!
			*/
		char1 = char2 = value = 0;
		p = &(buff[0]);
		if (*p) {
			while (*p && ((*p == ' ') || (*p == '\t')))
				++p;
			char1 = read_int_or_hex( &p);
			while (*p && ((*p == ' ') || (*p == '\t')))
				++p;
			char2 = read_int_or_hex( &p);
			while (*p && ((*p == ' ') || (*p == '\t')))
				++p;
			value = read_int_or_hex( &p);

			if (char1 && char2 && value) {
				table[nEntries].deadKey = char1;
				table[nEntries].nextKey = char2;
				table[nEntries].value = value;
				++nEntries;

				if (char1 < *first_dead_key)
					*first_dead_key = char1;
				if (char1 > *last_dead_key)
					*last_dead_key = char1;
			}
		}
	} /* while */
	/* Signal end of table */
	table[nEntries].deadKey = 0;
	table[nEntries].nextKey = 0;
	table[nEntries].value = 0;

	if (nEntries > 0)
		return( TRUE );
	else
		return( FALSE );
}

/* Read an integer or hex value; return integer (or zero if problem) */
static unsigned int
read_int_or_hex( pp)
char **pp;
{
	unsigned int result=0;
	int is_hex=FALSE;
	char c;
	int digit;

	/* Test the first 2 chars for '0x', meaning hex */
	if ((**pp == '0') && (*(*pp+1) == 'x')) {
		is_hex = TRUE;
		*pp = *pp + 2;
	}
	while ((c = **pp) && ((**pp != ' ') && (**pp != '\t'))) {
		++(*pp);
		if ((c >= '0') && (c <= '9'))
			digit = (int) (c - '0');
		else if (is_hex) {
			c = toupper(c);
			if ((c >= 'A') && (c <= 'F'))
				digit = (int) (c - 'A') + 10;
			else
				return(0);
		}
		else
			return(0);

		/* Add the good digit to the value */
		result = result * (is_hex ? 16 : 10);	/* shift up one place */
		result += digit;						/* and add last digit */
	} /* while */
	return( result);
}

#if HP800 && BUILD_ORIENT   /*  {  */
ResetSpotLocation( win )
Window win;
{
    AXWidget_main main_w;
    AXWidget_kanji status_w;
    XVaNestedList preedit_attr;
    XPoint spot;
 
    main_w = FindMainWidgetFromWindow( win );
    if ( main_w == NULL )
        return;
    if ( (main_w->ic == NULL) || (main_w->kanji_status == NULL))
        return;

    status_w = main_w->kanji_status;

    spot.x = status_w->xpos;
    spot.y = status_w->ypos + status_w->height - 3;
                 
    preedit_attr = XVaCreateNestedList(NULL, XNSpotLocation, &spot, NULL );
    XSetICValues( main_w->ic, XNPreeditAttributes, preedit_attr, NULL);
                 
    return;
}

#endif  /*  }  */



/** The following ssection is some code to allow the user to map a single */
/** keystroke to another single keystroke.  We change only the key-string */
/* not the keycode.  The user's table may also set/reset the "shift" and */
/* "ModMask" flags on the key event                                         */

typedef struct _map_table {
    unsigned char OldChar;
    unsigned char state;
    unsigned char NewChar;
    unsigned char new_state;
    } map_table;

map_table *MapTable;

int TwoHexDigits( p )
char *p;
{
    int digit;

    if ( (*p >= 'A') && (*p <= 'F') )
        digit = *p - 'A' + 10;
    else 
        digit = *p - '0';
    digit = digit << 4;
    p++;
    if ( (*p >= 'A') && (*p <= 'F') )
        digit = digit + (*p - 'A' + 10);
    else
        digit = digit + (*p - '0');

    return( digit );
}

   /* Read the key-map file.  It is structured EXACTLY as:    */
   /* key-as-2-hex-digits state-as-2-hex-digit new-key new-state */
   /* For example, to convert "a6" shifted to "B8 unshifted :    */
   /* A6 00 B9 01                                        */

   /* defined in "X.h" the current state are:            */
   /* #define ShiftMask       (1<<0)      */
   /* #define LockMask        (1<<1)      */
   /* #define ControlMask     (1<<2)      */
   /* #define Mod1Mask        (1<<3)      */
   /* #define Mod2Mask        (1<<4)      */
   /* #define Mod3Mask        (1<<5)      */
   /* #define Mod4Mask        (1<<6)      */
   /* #define Mod5Mask        (1<<7)      */
 

   
void ReadMapTableFile()
{
    int i;
    FILE *ff;
    unsigned int char1, char2, value;
    int ix;
    char buff[201], *p;
    char filename[ 1025 ];
    map_table *entryP;
 
    sprintf( filename, "%s/%s", ThimApplixDir, "char_map.sp" );
 
    ff = fopen( filename, "r" );
    if ( ff == (FILE *)NULL )
       /* no compose processing */
       return;
 
    ix = SizeMapTable = 0;
    while ( fgets( buff, 200, ff) != (char *)NULL )
       {
       if ( (buff[0] == '#') || (strlen(buff) < 11) )
           /* comment line */
           continue;
       SizeMapTable++;
       }

   entryP = MapTable = (map_table *)TrueCalloc( 1, 
                                         (SizeMapTable * sizeof(map_table)) );

   fseek( ff, 0, SEEK_SET );
   while ( fgets( buff, 200, ff) != (char *)NULL )
       {
       if ( (buff[0] == '#') || (strlen(buff) < 11) )
           /* comment line */
           continue;
       p = buff;
       entryP->OldChar = TwoHexDigits( p );
       p = p + 3;
       entryP->state = TwoHexDigits( p );
       p = p + 3;
       entryP->NewChar = TwoHexDigits( p );
       p = p + 3;
       entryP->new_state = TwoHexDigits( p );
       entryP++;
       }

    fclose( ff );
    return;
}
void MapKey( charbuff, kep )
unsigned char *charbuff;
XKeyEvent *kep;
{
    map_table *entryP;
    int state_bits;
    int i;

    if ( SizeMapTable == -1 )
        {
        /* read the mapping file and set things up */
        SizeMapTable = 0;
        ReadMapTableFile( );
        }

    if ( *charbuff == '\0' )
        /* no string */
        return;

    entryP = MapTable;

    /* we do not consider the Control or Lock flags.  We do this with */
    /* the Control flags, because most of the time it will not matter */
    /* since Controled keystrokes do not generate printable string.   */
    /* We ignore the lock mask, just because it is there              */
    state_bits = kep->state & ~(ControlMask | LockMask);
    for ( i=0;  i < SizeMapTable; i++ )
        {
        if ( (entryP->OldChar == *charbuff) && (entryP->state == kep->state) )
           {
           /* found a match */
           *charbuff = entryP->NewChar;
           kep->state = entryP->new_state | 
                             (kep->state & (ControlMask | LockMask));
           return;
           }
        entryP++;
        }

    return;
}

/********* end of key-map section          */

