/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xmwcolor.c
   Applix ASC version: 10
   Date: Mon May 10 21:13:08 1999
*/
#include <math.h>
#include <aos.h>
#include <xmlib.h>
#include <elxwins.h>
/*#include <thimble.h> */
#include <elf_env.h>
#include <eldata.h>
#include <evreason.h>

extern unsigned long ForePixel;
extern unsigned long AlisPixels[];

/***********************************************************************
function: AxCreateWColor - Create a color definition.
Return pointer to allocated color definition.  Caller must free
pointer with call to `AxFreeWColor ()'.
Return NULL on error.  Caller will treat that as 'no color' (default).
*/
wcolor_t *
AxCreateWColor(w, dP)
AXWidget w;
data_ptr dP;
{
	wcolor_t *color = NULL;
	data_ptr cvP; /* color value */
	int icolor;
	int color_type = 0;

	if (!dP)
		return(NULL);

	if (AxIsInt(dP)) {
		/* Old convention of color encoding scheme:
		   icolor is TRUE            -- use work colors.
		   icolor is FALSE           -- default (do not use work colors)
		   icolor of form 0x01rrggbb -- use color rrggbb. */
		icolor = AxIntFromDataPtr(dP);
		switch (w->type) {
		case W_LIST:
		case W_EDITBOX:
		case W_TABLE:
			/* icolor is a boolean for work-area colors */
			break;
		default:
			/* icolor is an rgb */
			break;
		}
	} else {
		if (!AxIsArray(dP))
			return(NULL); /* botch -- nocolor */
			
		color_type = AxIntFromArray(dP, 0);
		cvP = AxArrayElement(dP, 1);
		switch (color_type) {
		case WCSPACE_RGB:
			if (!AxIsArray(cvP))
				return(NULL); /* botch -- bad args */
			color = (wcolor_t *) calloc(1, sizeof(wcolor_t));
			color->any.type = color_type;
			color->rgb.r = AxIntFromArray(cvP, 0);
			color->rgb.g = AxIntFromArray(cvP, 1);
			color->rgb.b = AxIntFromArray(cvP, 2);
			AxGetRealPixel(0, color->rgb.r, color->rgb.g, color->rgb.b,
						   0, 255, &color->rgb.pixel);
			break;
		case WCSPACE_CMYK:
			if (!AxIsArray(cvP))
				return(NULL); /* botch -- bad args */
			color = (wcolor_t *) calloc(1, sizeof(wcolor_t));
			color->any.type = color_type;
			color->cmyk.c = AxIntFromArray(cvP, 0);
			color->cmyk.m = AxIntFromArray(cvP, 1);
			color->cmyk.y = AxIntFromArray(cvP, 2);
			color->cmyk.k = AxIntFromArray(cvP, 3);
			AxGetRealPixel(2, color->cmyk.c, color->cmyk.m, color->cmyk.y, color->cmyk.k,
						   255, &color->cmyk.pixel);
			break;
		case WCSPACE_HSB: {
			int r, g, b;
			if (!AxIsArray(cvP))
				return(NULL); /* botch -- bad args */
			color = (wcolor_t *) calloc(1, sizeof(wcolor_t));
			color->any.type = color_type;
			color->hsb.h = AxFloatFromArray(cvP, 0);
			color->hsb.s = AxFloatFromArray(cvP, 1);
			color->hsb.b = AxFloatFromArray(cvP, 2);
			r = g = b = 0;
			HSBToRGB(&r, &g, &b, color->hsb.h, color->hsb.s,
					 color->hsb.b);
			AxGetRealPixel(0, r, g, b, 0, 255, &color->rgb.pixel);
			break;
		}
		case WCSPACE_NAMED:
			color = (wcolor_t *) calloc(1, sizeof(wcolor_t));
			color->any.type = color_type;
			color->named.color_name = calloc(1, strlen(AxStrFromDataPtr(cvP)));
			strcpy(color->named.color_name, AxStrFromDataPtr(cvP));
			AxPixelFromColorname(color->named.color_name, &color->named.pixel);
			break;
		case WCSPACE_WKIT:
			color = (wcolor_t *) calloc(1, sizeof(wcolor_t));
			color->any.type = color_type;
			color->wkit.pixel = ForePixel;
			break;
		case WCSPACE_WORK:
			color = (wcolor_t *) calloc(1, sizeof(wcolor_t));
			color->any.type = color_type;
			color->work.pixel = AlisPixels[1];
			break;
		default:
			return(NULL); /* botch -- unknown type -- no color. */
			break;
		}
	}
	return(color);
}

/***********************************************************************
function: AxFreeWColor - Free storage allocated by a call to
`AxCreateWColor ()'.
Free all allocated memory inside CP and then CP itself.  Caller must
set CP to NULL for sanity.
*/
void
AxFreeWColor(cp)
wcolor_t *cp;
{
	if (!cp)
		return;

	switch (cp->any.type) {
	case WCSPACE_NAMED:
		if (cp->named.color_name)
			free(cp->named.color_name);
		cp->named.color_name = NULL;
		break;
	default:
		break;
	}
	free(cp);
	return;
}

/***********************************************************************
function: HSBToRGB - Convert from HSB to RGB color space.
H --> [0.0 .. 360.0]
S --> [0.0 .. 100.0]
B --> [0.0 .. 100.0]
*/
void
HSBToRGB(rp, gp, bp, h, s, v)
int *rp, *gp, *bp; /* OUT [0 .. 255] */
float h, s, v; /* IN */
{
	int r, g, b;
	float p, q, t;
	int i;
	float f;

	r = g = b = 0;

	s = s * 0.01; /* make it [0-1] */
	v = v * 0.01; /* make it [0-1] */

	/* Straight out of Foley, van Dam ... */
	if (s == 0.0) {
		/* ignore `h' */
		r = v * 255;
		g = v * 255;
		b = v * 255;
		goto finish;
	}
	if (h == 360.0)
		h = 0.0;
	h = h / 60.0; /* h is now in [0,6] */
	i = (int) floor((double) h);
	f = h - i;
	p = v * (1 - s);
	q = v * (1 - (s * f));
	t = v * (1 - (s * (1 - f)));

	/* max for applix get color is 255 */
	p = p * 255.0;
	q = q * 255.0;
	t = t * 255.0;
	v = v * 255.0;
	switch ((int) i) {
	case 0:
		r = (int) v;
		g = t;
		b = p;
		break;
	case 1:
		r = q;
		g = (int) v;
		b = p;
		break;
	case 2:
		r = p;
		g = (int) v;
		b = t;
		break;
	case 3:
		r = p;
		g = q;
		b = (int) v;
		break;
	case 4:
		r = t;
		g = p;
		b = (int) v;
		break;
	case 5:
		r = (int) v;
		g = p;
		b = q;
		break;
	default:
		r = g = b = 0;
		break;
	}
finish:
	*rp = r;
	*gp = g;
	*bp = b;
	return;
}

/***********************************************************************
function: RGBToHSB - Convert RGB to HSB
*/
void
RGBToHSB(oh, os, ob, ir, ig, ib)
float *oh; /* OUT: [0.0 .. 360.0] */
float *os; /* OUT: [0.0 .. 100.0] */
float *ob; /* OUT: [0.0 .. 100.0] */
int ir, ig, ib; /* IN: [0 .. 255] */
{
	float h, s, b; /* s, b are [0.0 .. 1.0], h is [0.0 .. 360.0] */
	float red, green, blue; /* r, g, b are [0.0 .. 1.0] */
	float max_value, min_value, delta_value;

	h = s = b = 0.0;
	red = green = blue = 0.0;

	red   = (ir / 255.0);
	green = (ig / 255.0);
	blue  = (ib / 255.0);

	/* Straight out of Foley, van Dam ... */
	max_value = MAX(red, (MAX(green, blue)));
	min_value = MIN(red, (MIN(green, blue)));

	/* calculate brightness. */
	b = max_value;

	/* calculate saturation */
	if (max_value != 0.0)
		s = (max_value - min_value) /  max_value;
	else
		s = 0.0;

	/* calculate hue */
	if (s == 0.0) {
		h = 0.0; /* undefined */
		goto finish;
	}
	delta_value = max_value - min_value;
	if (red == max_value)
		h = (green - blue) / delta_value;
	else if (green == max_value)
		h = 2.0 + ((blue - red) / delta_value);
	else if (blue == max_value)
		h = 4.0 + ((red - green) / delta_value);

	h = h * 60; /* convert to degrees */
	if (h < 0.0)
		h += 360.0;
finish:
	*oh = h;
	*os = s * 100.0;
	*ob = b * 100.0;
	return;
}

/***********************************************************************
function: AxGetHSBFromWColor - Extract h, s, b from COLOR.  Return 0
on success; -1 otherwise.
*/
int
AxGetHSBFromWColor(color, hp, sp, bp)
wcolor_t *color;
float *hp, *sp, *bp;
{
	if (!hp || !sp || !bp)
		return(-1);
	if (!color) {
		*hp = 0.0;
		*sp = 0.0;
		*bp = 0.0;
		return(-1);
	}
	switch (color->any.type) {
	case WCSPACE_RGB: {
		int r, g, b;
		r = (int) color->rgb.r;
		g = (int) color->rgb.g;
		b = (int) color->rgb.b;

		if (r < 0)
			r = 0;
		if (r > 255)
			r = 255;
		if (g < 0)
			g = 0;
		if (g > 255)
			g = 255;
		if (b < 0)
			b = 0;
		if (b > 255)
			b = 255;
		RGBToHSB(hp, sp, bp, r, g, b);
		break;
	}
	case WCSPACE_HSB:
		*hp = color->hsb.h;
		*sp = color->hsb.s;
		*bp = color->hsb.b;
		if (*hp < 0.0)
			*hp = 0.0;
		if (*hp > 360.0)
			*hp = 360.0;
		if (*sp < 0.0)
			*sp = 0.0;
		if (*sp > 100.0)
			*sp = 100.0;
		if (*bp < 0.0)
			*bp = 0.0;
		if (*bp > 100.0)
			*bp = 100.0;
		break;
	default:
		*hp = 0.0;
		*sp = 0.0;
		*bp = 0.0;
		return(-1);
		/*NOTREACHED*/
		break;
	}
	return(0);
}

/*
 * : wColorsEqual - compare two wcolor_t instances
 * the passed colors are assumed to be initialized
 * fully, that is, so that their pixel members are
 * initialized.
 */
int                     /* return c1 == c2 */
wColorsEqual(c1, c2)
wcolor_t *c1;           /* in */
wcolor_t *c2;           /* in */
{
	if (!c1 && !c2) {
		return (TRUE);
	}
	else if (!c1) {
		return (FALSE);
	}
	else if (!c2) {
		return (FALSE);
	}

	return (c1->any.pixel == c2->any.pixel);
}

/***********************************************************************
						 Elf Layer Interface
 ***********************************************************************/

/***********************************************************************
function: AxCreateWColorData - Create an elf-datum from the color spec CP.
Creates an elf-array/elf-integer suitable for call to `AxCreateWColor ()'
or to write out into a dialog box file.
*/
data_ptr
AxCreateWColorData(cp)
wcolor_t *cp;
{
	data_ptr dP = NULL; /* return datum */
	data_ptr cvP = NULL; /* color values */

	if (!cp)
		return(NULL);
	switch (cp->any.type) {
	case WCSPACE_RGB:
		/* dP is [ type [ r g b ] ] */
		dP  = AxMakeArray(2);
		cvP = AxMakeArray(3);

		cvP = AxAddIntToArray(cvP, 0, (int) cp->rgb.r);
		cvP = AxAddIntToArray(cvP, 1, (int) cp->rgb.g);
		cvP = AxAddIntToArray(cvP, 2, (int) cp->rgb.b);

		dP = AxAddIntToArray(dP, 0, cp->any.type);
		dP = AxAddArrayToArray(dP, 1, cvP);
		break;
	case WCSPACE_HSB:
		/* dP is [ type [ h s b ] ] */
		dP = AxMakeArray(2);
		cvP = AxMakeArray(3);

		cvP = AxAddFloatToArray(cvP, 0, (double) cp->hsb.h);
		cvP = AxAddFloatToArray(cvP, 1, (double) cp->hsb.s);
		cvP = AxAddFloatToArray(cvP, 2, (double) cp->hsb.b);

		dP = AxAddIntToArray(dP, 0, cp->any.type);
		dP = AxAddArrayToArray(dP, 1, cvP);
		break;
	case WCSPACE_NAMED:
		/* dP is [ type color_name ] */
		dP = AxMakeArray(2);
		dP = AxAddIntToArray(dP, 0, cp->any.type);
		dP = AxAddStrToArray(dP, 1, cp->named.color_name);
		break;
	case WCSPACE_WKIT:
	case WCSPACE_WORK:
		/* dP is [ type ]   -- for now */
		dP = AxMakeArray(1);
		dP = AxAddIntToArray(dP, 0, cp->any.type);
		break;
	default:
		break;
	}
	return(dP);
}


/* GNU Emacs settings - This (including the "\n^L") MUST be the last thing
 *                      in this file!
 * Local variables:
 * mode: c
 * tab-width: 4
 * end:
 */
