/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: aspeller.h
   Applix ASC version: 12
   Date: Mon Feb  1 21:13:46 1999
*/
#ifndef ASPELLER_INCL
#define ASPELLER_INCL

/*** Parts of filenames of Proximity files ***/
#define LEX_SUFFIX		".uh"
#define CORELEX_SUFFIX	".clx"
#define HY_SUFFIX		".hyp"
#define CLAM_PREFIX		".ax_"
#define CLAM_SUFFIX		".dct"
#define DETECT_SUFFIX	".env"
#define PHON_SUFFIX1	"phon"
#define PHON_SUFFIX2	".env"
#define PHON_SUFFIX		"phon.env"
#define SYNONYM_SUFFIX	".syn"


/*** Essential Proximity defines ***/
/* See Proximity file scdef.h for many useful defines, including:
	- Word type "doflags" flags (e.g. IW_COMMON, IW_PROPER, IW_ACRONYM...)
	- Location codes for isword() (e.g. IW_NOTFOUND, IW_CORELEX...)
	- Common string sizes (e.g. MAXWORD, LONGWORD)
	- Hyphenation rankings (HY_NORM, HY_PREF, HY_UNPREF)
*/

#ifndef UCHAR
typedef unsigned char   UCHAR;		/* keep in synch with proxport.h */
#endif
#ifndef VOID
typedef void VOID;
#endif
#ifndef USHORT
typedef unsigned short USHORT;
#endif

/* Hide the Proximity structures that we pass around but don't modify */
#ifndef CLAM
typedef void CLAM;
#endif
#ifndef CACHE
typedef void CACHE;
#endif
#ifndef CORELEX
typedef void CORELEX;
#endif
#ifndef LEX
typedef void LEX;
#endif
#ifndef COREHEAD
typedef void COREHEAD;
#endif
#ifndef TOKEN
typedef void TOKEN;
#endif

/*We must define our own value for the maximum word length, since the */
/*token "MAXWORD" is defined by Microsoft include files to be 0xFFF.  */
/*The use of our own define will avoid possible compile time confusion.  */
/*Note: I have NOT changed all the entries on the proximity sources since */
/*   none of these include any Windows include files. - jon           */
#define MAX_PROX_WORD 32    
#define LONGWORD	(2*MAX_PROX_WORD)

/*** Foreign language support ***/
/* Applix language codes match Proximity codes (see language.h) */
#define NOLANG			0	/* Invalid language code */
#define ENGLISH			1     /* American English */
#define GERMAN			2
#define FRENCH			3
#define SPANISH			4
#define ITALIAN			5
#define BRITISH			6     /* British English */
#define SWEDISH			7
#define DANISH			8
#define NORWEGIAN		9
#define DUTCH			10
#define PORTUGUESE		11
#define BRAZILIAN		12 	/* Brazilian Portuguese */
#define CANADIAN_FRENCH	13
#define SWISS_GERMAN	14 
#define NYNORSK			15	/* Norwegian dialect */
#define FINNISH			16
#define CATALAN			17
#define MAXLANG			17

/*** Codes for "locations" of found words. ***/
/* It is assumed that the dictionary codes, specified by the caller 
	(i.e. for site, user, and document dictionaries) will not conflict 
	with these numbers. */
#define IN_STOPCACHE	1001
#define IN_CACHE		1002
#define IN_CORELEX		1003
#define IN_LINGUIBASE	1004

/*** Thesaurus support ***/
/* Keep in synch with thes.h */
#define MAXMEAN 		50
#define MAXLISTS        5
#define TH_ALLPOS       0x0000FFFFL
#define TH_ALLLIST      0x00FF0000L
#define TH_INFALL       0x02000000L



/* See Proximity file thes.h for the "word list type" code defines (e.g.
	TH_SYN, TH_CMP, etc.) as well as other useful defines. */

/*** Current language and spelling/hyphenation options status of Words ***/
#define HYPH_NONE		0x0			/* No hyphenation for words */
#define HYPH_DICT		0x1			/* Look in dictionary for hyph points */
#define HYPH_CALC		0x2			/* Calculate hyph points */
#define HYPH_DICT_CALC	(HYPH_DICT + HYPH_CALC)	/* Look in dictionary for hyph
											points; if not there, calculate */

/* Proximity hyphenation ranking values (see cschar.h and sahyphen.h) */
#define HY_NORM		0x04
#define HY_PREF		0x05
#define HY_UNPREF	0x06
#define HY_TEMP		0x07
#define HY_GRMKK	0x10
#define HY_DBLCNS	0x20
#define HY_HARD		0x70

/* Various Proximity defines (see detect.h, scdef.h, clam.h) */
#define DET_NEW		0		/* New token */
#define DET_CLR		1		/* New position, not sentence start */
#define DET_RST		2		/* Ignore current input */
#define DET_SEN		3		/* Set sentence start */

#define TOK_MORE	0		/* add more characters */
#define TOK_MADE	1       /* token made; call tokdesc */
#define TOK_DONE	2       /* tokdesc finished */
#define TOK_NULL	3		/* No text characters in the token */
#define TOK_ERROR	4		/* tokdesc failure */
#define TOK_REPEAT	5       /* token is repeat of previous */
#define TOK_EOF		6       /* token is end of input */
#define TOK_NUMBER	7       /* token is number */
#define TOK_DATE	8		/* token is a date */
#define TOK_TIME	9		/* token is a time */
#define TOK_NONALPH	10      /* nonalphabnumeric token */
#define TOK_MISP	11      /* A misspelled token */
#define TOK_SENT	12      /* An improper start of sentence. */
#define TOK_ROMAN	13      /* A Roman numeral */
#define TOK_PPREF	14      /* An invalid punctuation prefix */
#define TOK_PSUFF	15      /* An invalid punctuation suffix */
#define TOK_OUTLN	16		/* An outline marker */

#define SP_CORRECT	0x1     /* Word list can be used for correction */
#define SP_DETECT	0x2     /* Word list can be used for detection */
#define SP_HYPHEN	0x4     /* Use this word list for hyphen verify */
#define SP_STOP		0x8     /* Stop if word in list but flags mismatch */
#define SP_MASK		0xF     /* Mask for the Searchpath flags */

#define CL_EOF		(-2)
#define CL_INIT		0x10
#define CL_PAIR     0x20
#define CL_READONLY	0x40

#define IW_NOTFOUND     0x0000
#define PRX_OKAY    0
#define PRX_ERROR	(-1)



typedef struct langnode {
    short   num;
    char    name[32];
    struct {
         unsigned use_compound_spelling : 1;  /* use compund word spelling */
         unsigned use_dictionary_hyph   : 1;  /* use dict based hyphenation */
         unsigned try_algorithmic_hyph  : 1;  /* use algoritmic hyph if no  */
                                       /*   hyphenation points from the dict */
         } flags;
} LangNode;

/* Here are all the detect codes Proximity can issue (via TOK_XXX values).
	Not all of them are necessarily "spelling errors" */
typedef struct {
	unsigned int unused0 : 1;		/* 0: Unused */
	unsigned int unused1 : 1;		/* 1: Unused */
	unsigned int unused2 : 1;		/* 2: Unused */
	unsigned int unused3 : 1;		/* 3: Unused */
	unsigned int unused4 : 1;		/* 4: Unused */
	unsigned int repeat : 1;		/* 5: Repeated word */
	unsigned int unused6 : 1;		/* 6: Unused */
	unsigned int number : 1;		/* 7: A number */
	unsigned int date : 1;			/* 8: A date */
	unsigned int time : 1;			/* 9: A time */
	unsigned int nonAlph : 1;		/* 10: Not alphabetic (contains symbols) */
	unsigned int misspell : 1;		/* 11: Misspelled */
	unsigned int startSent : 1;		/* 12: Bad start of sentence (needs cap) */
	unsigned int roman : 1;			/* 13: A Roman numeral */
	unsigned int punctPre : 1;		/* 14: Invalid punctuation prefix */
	unsigned int punctSuff : 1;		/* 15: Invalid punctuation suffix */
	unsigned int outline : 1;		/* 16: An outline marker */
	unsigned int wordPS : 1;		/* 17: Word contains invalid prefix/suffix*/
	unsigned int prefix : 1;		/* 18: Invalid prefix */
	unsigned int wordWithPS : 1;	/* 19: Word attached to invalid pre/suffix*/
	unsigned int suffix : 1;		/* 20: Invalid suffix */
	unsigned int parens : 1;		/* 21: Text contains parentheses */
	unsigned int badParens : 1;		/* 22: Word with paren text is invalid */
	unsigned int needParens : 1;	/* 23: Word without paren text is invalid */
	unsigned int goodCmpdPart : 1;	/* 24: Valid part of a compound word */
	/* #25 is German only */
	unsigned int cmpd : 1;			/* 25: A compound word */
	unsigned int badCmpd : 1;		/* 26: An invalid compound word */
	unsigned int goodCmpd : 1;		/* 27: A valid compound word */
	unsigned int slash : 1;			/* 28: A slash that separates words */
	unsigned int unneededHyphen : 1;/* 29: Word does not need a hyphen */
	/* #30 - #32 are French and Italian only */
	unsigned int contract : 1;		/* 30: Could contract word w/prev word */
	unsigned int apos : 1;			/* 31: Word contains apostrophe prefix */
	unsigned int misspellApos : 1;	/* 32: Invalid word contains apos prefix */
	unsigned int hyphen : 1;		/* 33: A hyphen */
	unsigned int extraHyphen : 1;	/* 34: Invalid hyphen separator */
	unsigned int hasHyphen : 1;		/* 35: Word contains hyphens */
	unsigned int misspellHyphen : 1;/* 36: Invalid word contains hyphens */
	unsigned int hasPosses : 1;		/* 37: Word is in possessive form */
	unsigned int misspellPosses : 1;/* 38: Invalid word is in possessive form */
	/* #39 - #40 are French and Italian only */
	unsigned int aposPrefix : 1;	/* 39: Apostrophe prefix of a prefixed wd */
	unsigned int badAposPrefix : 1;	/* 40: Invalid apostrophe prefix */
	unsigned int posses : 1;		/* 41: Possesive part of word */
	unsigned int badPosses : 1;		/* 42: Invalid possesive part of word */
	/* # 43 is German and Dutch only */
	unsigned int linkage : 1;		/* 43: Linkage symbol */
	unsigned int hyNumber : 1;		/* 44: Hyphenated number (e.g phone #) */
} sp_error_mask;

/*** Translation tables: Aster*x-to-Proximity and Proximity-to-Aster*x ***/
/* This was taken from rev 6.3 of Aster*x 2.0 pgms/includes/prox_trans.h */
/* !!1 Use latest version of this file; it gets corrected for weird
	chearacters every so often. */

#define AX2PROX1 { \
	'\000',	'\001',	'\002',	'\003',	'\004',	'\005',	'\006',	'\007',	\
	'\010',	'\011',	'\012',	'\013',	'\014',	'\015',	'\016',	'\017',	\
	'\020',	'\021',	'\022',	'\023',	'\024',	'\025',	'\026',	'\027',	\
	'\030',	'\031',	'\032',	'\033',	'\034',	'\035',	'\036',	'\037',	\
	' ',	'!',	'"',	'#',	'$',	'%',	'&',	'\'',	\
	'(',	')',	'*',	'+',	',',	'-',	'.',	'/',	\
	'0',	'1',	'2',	'3',	'4',	'5',	'6',	'7',	\
	'8',	'9',	':',	';',	'<',	'=',	'>',	'?',
#define AX2PROX2 \
	'@',	'A',	'B',	'C',	'D',	'E',	'F',	'G',	\
	'H',	'I',	'J',	'K',	'L',	'M',	'N',	'O',	\
	'P',	'Q',	'R',	'S',	'T',	'U',	'V',	'W',	\
	'X',	'Y',	'Z',	'[',	'\\',	']',	'^',	'_',	\
	'`',	'a',	'b',	'c',	'd',	'e',	'f',	'g',	\
	'h',	'i',	'j',	'k',	'l',	'm',	'n',	'o',	\
	'p',	'q',	'r',	's',	't',	'u',	'v',	'w',	\
	'x',	'y',	'z',	'{',	'|',	'}',	'~',	'\177',
#define AX2PROX3 \
	'?',	'?',	'?',	'?',	'?',	'?',	'?',	'?',	\
	'?',	'?',	'?',	'?',	'?',	'?',	'?',	'?',	\
	'?',	'?',	'?',	'?',	'?',	'?',	'?',	'?',	\
	'?',	'?',	'?',	'?',	'?',	'?',	'?',	'?',	\
	'?',	'\255',	'?',	'?',	'?',	'?',	'?',	'?',	\
	'?',	'?',	'\246',	'\256',	'?',	'?',	'?',	'\304',	\
	'\370',	'?',	'?',	'?',	'?',	'?',	'?',	'?',	\
	'?',	'?',	'\247',	'\257',	'?',	'?',	'?',	'\250',
#define AX2PROX4 \
	'\260',	'\265',	'\273',	'\303',	'\216',	'\217',	'\222',	'\200',	\
	'\261',	'\220',	'\274',	'\271',	'\262',	'\266',	'\275',	'\272',	\
	'?',	'\245',	'\263',	'\267',	'\276',	'\307',	'\231',	'?',	\
	'\351',	'\264',	'\270',	'\277',	'\232',	'?',	'?',	'\341',	\
	'\205',	'\240',	'\203',	'\302',	'\204',	'\206',	'\221',	'\207',	\
	'\212',	'\202',	'\210',	'\211',	'\215',	'\241',	'\214',	'\213',	\
	'?',	'\244',	'\225',	'\242',	'\223',	'\306',	'\224',	'?',	\
	'\355',	'\227',	'\243',	'\226',	'\201',	'?',	'?',	'\310',	\
}

#define PROX2AX1 { \
	'\000',	'\001',	'\002',	'\003',	'\004',	'\005',	'\006',	'\007',	\
	'\010',	'\011',	'\012',	'\013',	'\014',	'\015',	'\016',	'\017',	\
	'\020',	'\021',	'\022',	'\023',	'\024',	'\025',	'\026',	'\027',	\
	'\030',	'\031',	'\032',	'\033',	'\034',	'\035',	'\036',	'\037',	\
	' ',	'!',	'"',	'#',	'$',	'%',	'&',	'\'',	\
	'(',	')',	'*',	'+',	',',	'-',	'.',	'/',	\
	'0',	'1',	'2',	'3',	'4',	'5',	'6',	'7',	\
	'8',	'9',	':',	';',	'<',	'=',	'>',	'?',
#define PROX2AX2 \
	'@',	'A',	'B',	'C',	'D',	'E',	'F',	'G',	\
	'H',	'I',	'J',	'K',	'L',	'M',	'N',	'O',	\
	'P',	'Q',	'R',	'S',	'T',	'U',	'V',	'W',	\
	'X',	'Y',	'Z',	'[',	'\\',	']',	'^',	'_',	\
	'`',	'a',	'b',	'c',	'd',	'e',	'f',	'g',	\
	'h',	'i',	'j',	'k',	'l',	'm',	'n',	'o',	\
	'p',	'q',	'r',	's',	't',	'u',	'v',	'w',	\
	'x',	'y',	'z',	'{',	'|',	'}',	'~',	'\177',
#define PROX2AX3 \
	'\307',	'\374',	'\351',	'\342',	'\344',	'\340',	'\345',	'\347',	\
	'\352',	'\353',	'\350',	'\357',	'\356',	'\354',	'\304',	'\305',	\
	'\311',	'\346',	'\306',	'\364',	'\366',	'\362',	'\373',	'\371',	\
	'?',	'\326',	'\334',	'?',	'?',	'?',	'?',	'?',	\
	'\341',	'\355',	'\363',	'\372',	'\361',	'\321',	'\252',	'\272',	\
	'\277',	'?',	'?',	'?',	'?',	'\241',	'\253',	'\273',	\
	'\300',	'\310',	'\314',	'\322',	'\331',	'\301',	'\315',	'\323',	\
	'\332',	'\313',	'\317',	'\302',	'\312',	'\316',	'\324',	'\333',
#define PROX2AX4 \
	'?',	'?',	'\343',	'\303',	'\257',	'n',	'\365',	'\325',	\
	'\377',	'\377',	'l',	'L',	'?',	'?',	'?',	'?',	\
	'?',	'?',	'?',	'?',	'?',	'?',	'?',	'?',	\
	'?',	'?',	'?',	'?',	'?',	'?',	'?',	'?',	\
	'?',	'\337',	'?',	'?',	'?',	'?',	'?',	'?',	\
	'?',	'\330',	'?',	'?',	'?',	'\370',	'?',	'?',	\
	'?',	'?',	'?',	'?',	'?',	'?',	'?',	'?',	\
	'\260',	'?',	'?',	'?',	'?',	'?',	'?',	'?',	\
}


#ifndef ASPELLER_PROTO_INCL
#define ASPELLER_PROTO_INCL

/*** Prototypes for public Speller functions ***/
extern AX_BOOLEAN SpIsLanguageOK AXP_((int language));
extern int SpGetCurrentLanguage AXP_((void));
extern int SpGetLanguageCode AXP_((char *langStr));
extern void SpGetLanguageName AXP_((int langCode,char *langName));
extern typeErr SpCloseSpeller AXP_((void));

extern typeErr SpListValidSpellings AXP_((char *word,int language,char      **dictionaries,int maxGet,int *gottenP,char words[][MAX_PROX_WORD+1],AX_BOOLEAN *inList));
extern typeErr SpWordSpelledCorrectly AXP_((char *word,int language, char   **dictionaries,AX_BOOLEAN *spellOK));
extern typeErr SpWordProximityData AXP_((char *word,int language, char      **dictionaries,char query[MAX_PROX_WORD+1],int *qflags,int *qlen,char qlookup[     MAX_PROX_WORD+1],char qinfo[MAX_PROX_WORD+1],int *wlist,int *wflags,int *iswordRet));
extern typeErr SpCheckString AXP_((char *string,int startOffset,int language,  char **dictionaries,sp_error_mask *mask,int *errP,int *errStartP,int *errEndP ,int *scanEndP));

extern AX_BOOLEAN SpIsHyphenationOK AXP_((int language));
extern typeErr SpHyphenateWord AXP_((char *word,int language, char          **dictionaries,AX_BOOLEAN rank,int how,int *howDid,char hyPoints[MAX_PROX_WORD+1]));

extern AX_BOOLEAN SpIsThesaurusOK AXP_((int language));
extern typeErr SpWhatThesaurusData AXP_((int language,int *thesDataP));
extern typeErr SpGetThesaurus AXP_((char *word,int language,int *numMeaningsP, char *meanings[MAXMEAN+1],char **words[][MAXLISTS]));
extern typeErr SpCloseThesaurus AXP_((void));

extern AX_BOOLEAN SpCanAddAnotherDictionary AXP_((void));
extern typeErr SpCreateDictionary AXP_((int language,char *dict));
extern typeErr SpRemoveDictionary AXP_((char *dfile));
extern typeErr SpAddWord AXP_((char *word,char *correct,int language, char *dfile));
extern typeErr SpRemoveWord AXP_((char *word,char *correct,int language,     char *dfile));
extern typeErr SpGetDictWords AXP_((int language,char *dfile,AX_BOOLEAN firstTime,   int maxGet,int *gottenP,char words[][LONGWORD+2]));
extern void SpInitErrMask AXP_((sp_error_mask *mask));
extern void SpSetErrMaskValue AXP_((sp_error_mask *mask,int code,AX_BOOLEAN value));
extern AX_BOOLEAN SpGetErrMaskValue AXP_((sp_error_mask *mask,int code));

#endif /* ASPELLER_PROTO_INCL */

#endif /* ASPELLER_INCL */
