/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: atm.h
   Applix ASC version: 27
   Date: Wed Feb 24 21:13:16 1999
*/
#ifndef _ATM_H
#define _ATM_H

#include <aos.h>
#define TMR_OK      AOK
#include <getstr13.h>   /* returned status values (TMR_) */

#define TMR_OWN     TMR_NOCRE   /* old equivalent symbol */
#define TMR_OPE     TMR_SELE    /* old equivalent symbol */
#define TMR_DDEV    TMR_DEVUN   /* old equivalent symbol */
#define TMR_NOACT   TMR_NOFUN   /* old equivalent symbol */

#define BUILTIN_FILL_TYPE  0
#define USERDEF_FILL_TYPE  1
#define LINEAR_FILL_TYPE   2
#define RADIAL_FILL_TYPE   3
#define RECT_FILL_TYPE     4
#define GRAY_FILL_TYPE     5
#define BILINEAR_FILL_TYPE 6

/* null identifier values */

#define TM_NULHELP	(-1)
#define TM_NULWIN	(-1)
#define TM_NULREG	(-1)

/* absolute/relative position option */

#define TM_ABSOLUTE	0
#define TM_RELATIVE	1

/* color types */
#define PROCESS 0
#define SPOT    1

/* device setup number for ASCII file.  Other setups in paratmgr.h */

#define TM_DS_LPR1	0	/* standard 66x80 line printer */

/* short sized points and rects */
typedef struct TmPoint {
	short x;
	short y;
} TmPoint;
typedef struct TmRect {
	short x_position;
	short y_position;
	short x_size;
	short y_size;
} TmRect;

/* int sized points and rects */
typedef struct TmPoint32 {
	int x;
	int y;
} TmPoint32;
typedef struct TmRect32 {
	int x_position;
	int y_position;
	int x_size;
	int y_size;
} TmRect32;

/* a color */
typedef struct color_type {
    unsigned int ink_type : 31;
	unsigned int see_thru :  1;
    unsigned int c        :  8;
    unsigned int m        :  8;
    unsigned int y        :  8;
    unsigned int k        :  8;
    char *name;
} color_type;

/* a colormap */
typedef struct cmap_type {
    int n_colors;
    struct color_type *color;
} cmap_type;

/* a polygon (includes bezier control points) */
typedef struct poly_type {
	int nPoints, nCtrls;
	TmPoint32 *points;
	TmPoint32 *controls;
} poly_type;

/* some attribute info for polygons */
typedef struct attr_type {
	/* back fill attrs */
	struct color_type ffg, fbg;
	int f_type, f_pat, f_ang;
	int f_dhyt, f_dwid, f_ddep;
	int f_xoff, f_yoff;
	char *f_data;
	char *f_sdata;
	cmap_type *f_cmap;

	/* line fill attrs */
	struct color_type lfg, lbg;
	int l_type, l_pat, l_ang;
	int l_dhyt, l_dwid, l_ddep;
	int l_xoff, l_yoff;
	char *l_data;
	char *l_sdata;
	cmap_type *l_cmap;

	/* shadow attrs */
	int s_type;
	struct color_type sfg;
	int s_xoff, s_yoff;

	/* line style attrs */
	int l_style;
	int l_wid;
	int l_cap;  /* cap style: TM_CAP_ROUND, TM_CAP_BUTT, TM_CAP_PROJECTING */

	/* offset in mills for absolute (doc relative zero point) filling */
	int x_off;
	int y_off;

	/* transfer mode: TMS_X_D, TMZEROS, TMONES, or TMS (default) */
	int mode;

	/* flags */
	unsigned int line_first : 1,
	             fill_flag  : 1,
	             line_flag  : 1,
	             fill_abs   : 1,
	             line_abs   : 1,
	             use_ffg    : 1,
	             use_fbg    : 1,
	             use_lfg    : 1,
	             use_lbg    : 1,
				 f_remaped  : 1,
				 l_remaped  : 1,
				 f_pic      : 1,
				 l_pic      : 1;
} attr_type;

typedef struct ge_image_type {
    int hyt, wid, dep;
    int data_byte_wid;
    int mask_byte_wid;
    unsigned char *data;
    unsigned char *mask;
    unsigned char *sys_data;
    unsigned char *sys_mask;
    unsigned char *java_data;
    unsigned char *java_mask;
    char *eps_seg;
    int eps_seg_size;
} ge_image_type;

typedef struct ge_image_transform_type {
      int hyt, wid;
      int x_t, y_t;
      double x_x, x_y;
      double y_x, y_y;
      double x_scale;
      double y_scale;
} ge_image_transform_type;

typedef struct clip_path {
	struct poly_type *paths;
	int n_paths;
} clip_path;

#include <basicwin.h>
#include <xmlib.h>
#if 0
#include <atmproto.h>
#endif

#endif /* _ATM_H */
