/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: atmattr.h
   Applix ASC version: 4
   Date: Mon Nov 16 13:52:52 1998
*/
#ifndef ATMATTR_INCLUDED
#define ATMATTR_INCLUDED 1

/*
 * parameters for callers of ATM attributes
 * and characteristics functions
 *
 *
 * line attributes
 *
 * note: values of line styles and line thicknesses cannot
 * be changed without rewriting parts of layline.c module
 *
 * Revision history:
 *   12/05/86  GAB  Added new font names, sizes, color, a few other attributes.
 *                  Changed some old values to match new ones.
 *                  Updated TmTextDescriptor bit masks and offsets.
 *                  Fields added to Features packet.
 *                  Color fields added to TmSetWinAttr.
 *   04/15/87  GAB  Added TM_ALISFONT.
 *   01/04/90 BOBZ  Added five extra font face names.
 */

typedef struct {
	short line_style;
	short line_thickness;
} TmLineDescriptor;

/* line style */

#define TM_SOLID	0
#define TM_DASHED	1
#define TM_DOTTED	2
#define TM_DADOT	3
#define TM_THIN_DOTTED	4

/* line thickness */

#define TM_VTHIN	0
#define TM_THIN 	1
#define TM_MEDIUM	2
#define TM_THICK	3
#define TM_VTHICK	4

/* line cap style */

#define TM_CAP_ROUND		0  /* default */
#define TM_CAP_BUTT			1  /* not supported on screen 7/9/92 */
#define TM_CAP_PROJECTING	2  /* not yet supported 7/9/92 */


/* text attributes */

typedef unsigned long TmTextDescriptor;

/* Old TmTextDescriptor bit masks (before 1.20) */
#define TMOLDFMBB		3
#define TMOLDFMUU		3
#define TMOLDFMII		3
#define TMOLDFMSS		7
#define TMOLDFMCC		7
#define TMOLDFMHI		3
#define TMOLDFMSIZ		7
#define TMOLDFMFONT	7

/* Old TmTextDescriptor bit positions for attribute fields */
#define TMOLDFPBB		0
#define TMOLDFPUU		2
#define TMOLDFPII		4
#define TMOLDFPSS		6
#define TMOLDFPCC		9
#define TMOLDFPHI		12
#define TMOLDFPSIZ		14
#define TMOLDFPFONT	17

/* TmTextDescriptor bit masks */
#define TMFMBB		3	/* Bold */
#define TMFMUU		15	/* Underline, strike through */
#define TMFMII		3	/* Italic */
#define TMFMSS		7	/* Sub- and superscript */
#define TMFMCC		7	/* Capitalization */
#define TMFMHI		3	/* Highlight */
#define TMFMHY		7	/* Hyphenation */
#define TMFMSIZ		31	/* Size */
#define TMFMFONT	15	/* Font face */
#define TMFMCOLOR	15	/* Foreground color */

/* TmTextDescriptor bit positions for attribute fields */
#define TMFPBB		0
#define TMFPUU		2
#define TMFPII		6
#define TMFPSS		8
#define TMFPCC		11
#define TMFPHI		14
#define TMFPHY		16
#define TMFPSIZ		19
#define TMFPFONT	24
#define TMFPCOLOR	28

/* generic attributes */

/* generally applicable */
#define TMA_NO		0
#define TMA_SET		1
#define TMA_DEF		2
#define TMA_PRIVATE     3  /* Reserved for application use. Undefined result. */

/* superscript/subscript */
#define TMA_SUP		1
#define TMA_SUB		4

/* capitalization */
#define TMA_LC		1	/* Lower case */
#define TMA_UC		4	/* Upper case */
#define TMA_IC		5	/* Initial caps */
#define TMA_SC		6	/* Small caps */

/* hyphenation (DC internal use) */
#define TMA_HYPHOK		1
#define TMA_SOFTHYPH	4

/* old font sizes */
#define TMA_SMALL	1
#define TMA_REGSIZE	4
#define TMA_LARGE	5
#define TMA_XTRALG	6

/* new point sizes (a point is approximately 1/72 of an inch) */
#define TMA_6PT        7
#define TMA_8PT        8
#define TMA_10PT       1
#define TMA_12PT       4
#define TMA_14PT       5
#define TMA_18PT       6
#define TMA_24PT       9
#define TMA_36PT       10

/* old font names */
#define TMA_F1		1
#define TMA_F2		4
#define TMA_F3		5
#define TMA_F4		6

/* new font names */
#define TMA_SERIF1      1
#define TMA_SANS1       4
#define TMA_MONO1       5
#define TMA_SYMBOL1     6
#define TMA_SERIF2      7
#define TMA_SANS2       8
#define TMA_SYMBOL2     9
#define TMA_SYSFONT     10
/* some extra font names */
#define TMA_FACE8		11
#define TMA_FACE9		12
#define TMA_FACE10		13
#define TMA_FACE11		14
#define TMA_FACE12		15

/*
 * underline options
 * - note: codes have a subfield structure used internally by the ATM
 */
#define TMA_UNDER       1   /* Normal underline */
#define TMA_DOUBLEUNDER	8	/* Double underline */
#define TMA_STRIKETHRU  4   /* Strike through */
#define TMA_SU			5	/* strike through & underline */
#define TMA_SDU			6	/* strike through & double underline */
#define TMA_BU			9	/* broken underline */
#define TMA_BDU			10	/* broken double underline */
#define TMA_BS			12	/* broken strike through */
#define TMA_BSU			13	/* broken strike through & underline */
#define TMA_BSDU		14	/* broken strike through & double underline */

/* colors */
#define TMA_WHITE    0   /* White (Media color - usually white) */
#define TMA_BLACK    1   /* Black (Opposite of media color) */
#define TMA_RED      4   /* Red */
#define TMA_LBLUE    5   /* Light blue (Cyan) */
#define TMA_GREEN    6   /* Green */
#define TMA_PURPLE   7   /* Purple (Magenta) */
#define TMA_BLUE     8   /* Blue */
#define TMA_YELLOW   9   /* Yellow */
#define TMA_DGRAY    10  /* Dark gray */
#define TMA_OLIVE    11  /* Olive (yellow green) */
#define TMA_LGRAY    12  /* Light gray */
#define TMA_DRED     13  /* Dark red */
#define TMA_ORANGE   14  /* Orange */
#define TMA_DBLUE    15  /* Dark blue */


/* special character code */

#define TM_NBSPACE		0240	/* no-break space */

/* transfer mode */
			/*
			   S - source 	D - destination
			   I - invert 	A - and	  O - or   X - xor
			*/

#define TMZEROS		0
#define TMS_A_D		1
#define TMS_A_ID	2
#define TMS		3
#define TMIS_A_D	4
#define TMD		5
#define TMS_X_D		6
#define TMS_O_D		7
#define TMIS_A_ID	8
#define TMIS_X_D	9
#define TMID		10
#define TMS_O_ID	11
#define TMIS		12
#define TMIS_O_D	13
#define TMIS_O_ID	14
#define TMONES 		15

/* text alignment parameters */
#define TM_TOPLEFT	0
#define TM_BASELEFT	1

/* justification parameter for field read */

#define TM_LEFT_JUST	0
#define TM_CENTERED		1
#define TM_RIGHT_JUST	2
#define TM_FULL_JUST	3

/* tab stop list handling options supplied to TmTInquireTxExtent */

#define TM_IGNORE	0	/* ignore usage flags in tab stop list */
#define TM_ACCEPT	1	/* accept existing usage flags as input */
#define TM_COMPUTE	2	/* compute and store usage flags */

/* string status values returned by inquire text extent */

#define TM_INLIMIT	0	/* string length within stated limit */
#define TM_NOWORD	1	/* limit exceeded, no word bdry found */
#define TM_WORD		2	/* limit exceeded, word bdry found */

/* flag definitions */

/* mouse types */
#define TM_MCDEF	0	/* defualt */
#define TM_MCHG		1	/* hour-glass */
#define TM_MCARR	2	/* arrow */


/* terminal features */
struct TmFeatures
	{
		short txGrMode;	/* text/graphics mode */
		short mcellHt;	/* height of char-cell in mills */
		short pcellHt;	/* height of char-cell in pixels */
		short mcellWid;	/* width of char-cell in mills */
		short pcellWid;	/* width of char-cell in pixels */
		short defdev;	/* default device setup number */
		short txMHt;	/* text unit mil and pixel sizes */
		short txPHt;
		short txMWid;
		short txPWid;
		short grMHt;	/* graphic unit mil and pixel sizes */
		short grPHt;
		short grMWid;
		short grPWid;
		long modified;	/* when dev desc was last modified */
		short extBits;	/* other information such as speed */
		short res1;	/* reserved */
	};

#define TM_TXMODE	0	/* text mode ATM */
#define TM_GRMODE	1	/* full graphics mode ATM */
#define TM_RGMODE	2	/* remote graphics mode ATM */

/* individual bit values in extBits */

#define TM_SLOWGRFOPS	1	/* Graphics ops are leisurely */
				/* if (...->extBits & TM_SLOWGRFOPS) */
#define TM_MONOPOS	2	/* Device supports monospace positions only */
#define TM_ALISFONT	4	/* Device uses Alis standard font widths */
#define TM_VSTRETCH	8	/* ATM rescaling vertically so pixels measure square */


/* Page categories for TmPage() call */

#define TM_FIRSTPAGE  1
#define TM_EVENPAGE   2
#define TM_ODDPAGE    3

/* calls */

#if 1
typeErr TmInq();
typeErr TmDchar();
typeErr TmDcursor();
typeErr TmInqAttr();
typeErr TmSetWinAttr();
typeErr TmSetDevAttr();
typeErr TmSetSpWid();
typeErr TmAlterTxExtent();
typeErr TmInqTxExtent();
typeErr TmRInqTxExtent();
#endif

/* SYNOPSIS */

#if 0				/* documentation only */
*************************************************

typeErr TmInq (window_id, x_position, y_position, x_size, y_size, h_sc_pos, 
		v_sc_pos, curtype, x_cur_pos, y_cur_pos)
short window_id;
short *x_position;
short *y_position;
short *x_size;
short *y_size;
short *h_sc_pos;
short *v_sc_pos;
short *curtype;
short *x_cur_pos;
short *y_cur_pos;

 return status

 TMR_OK
 TMR_OWN

*************************************************

typeErr TmDchar(h_desktop,w_desktop,d_height,d_width, features, pSetup)
short *h_desktop;
short *w_desktop;
short *d_height;
short *d_width;
struct TmFeatures *features
short pSetup;

 return status

 TMR_OK

*************************************************

typeErr TmDcursor(window_id,text_overtype,mouse_type,mouse_state,x_cur_pos,
	y_cur_pos)
short window_id;
short text_overtype;
short mouse_type;
short mouse_state;
short x_cur_pos;
short y_cur_pos;

 return status

 TMR_OK
 TMR_NCUR

*************************************************

typeErr TmInqAttr(window_id, x_cur_pos, y_cur_pos, transfer_mode,
	lin_des, tx_des, fg_color, bg_color, cliprect);
short window_id;
short *x_cur_pos;
short *y_cur_pos;
TmLineDescriptor *lin_des;
TmTextDescriptor *tx_des;
short *fg_color;
short *bg_color;
TmRect *cliprect;

 return status

 TMR_OK
 TMR_OWN

*************************************************

typeErr TmSetWinAttr(window_id, x_cur_pos, y_cur_pos, lin_des,
	read_tx_des, write_tx_des, trans_mode, cliprect, txAlign,
	fg_color, bg_color);
short window_id;
short x_cur_pos,y_cur_pos;	/* currently ignored */
TmLineDescriptor *lin_des;
TmTextDescriptor *write_tx_des;
TmTextDescriptor *read_tx_des;
short trans_mode;
TmRect *cliprect;	/* NULL = don't change */
short txAlign;		/* text alignment */
short fg_color;
short bg_color;

 return status

 TMR_OK
 TMR_OWN
 TMR_PARM

*************************************************

typeErr TmSetDevAttr(fg_color,bg_color,entry_color,color_index);
short fg_color;
short bg_color;
short entry_color;
short color_index;

 return status

 TMR_OK
 TMR_PARM

*************************************************

typeErr TmSetSpWid(sp_width);
short sp_width;

 return status

 TMR_OK

*************************************************

typeErr TmAlterTxExtent(window_id,space_count,total_change);
short window_id;
short space_count;
short total_change;

 return status

 TMR_OK
 TMR_OWN

*************************************************

typeErr TmInqTxExtent(string, attr_string, byte_count, pSetup, limit,
		defAttr, sp_width, str_status, cut_pt1, cut_pt2, extent1, extent2,
		extent3, space_count, maxHeight, aboveBline)
char *string;
TmTextDescriptor *attr_string;
short byte_count;
short pSetup;
short limit;
TmTextDescriptor defAttr;
short sp_width;
short *str_status;
short *cut_pt1;
short *cut_pt2;
short *extent1;
short *extent2;
short *extent3;
short *space_count;
short *maxHeight;
short *aboveBline;

 return status

 TMR_OK
 TMR_NGC

*************************************************
#endif


#endif     /*   ATMATTR_INCLUDED    */
