/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: atmtxtin.h
   Applix ASC version: 5
   Date: Mon Nov 16 13:52:52 1998
*/
#ifndef ATMTXTIN_INCLUDED
#define ATMTXTIN_INCLUDED 1

/*
	Parameters for Callers of ATM Text and Input Functions

	Revision History:

	01-20-89  CRJ  Update TM_TAASIZE for change in IPC buffer size.
	02-08-89  SPE  Added TMK_SQUARE and TMK_CIRCLE local stroke symbols
	08-02-89 BOBZ  Added tab leadering types.
	12-01-89 BOBZ  Added TmTxExtentData, TmTxExtentFlags struct defs.
	02-03-91 BOBZ  Change TmEInq flag FJendCheck into reportPix. [4.3]
	31-05-91 cch   decimal character for each tab
	10-08-91 BOBZ  Add TmFRPacket variant w/ 32bit field size/pos [6.3]
*/

#include <qmenu.h>

/*
 * max combined size of text and accompanying attribute string (bytes)
 *
 * - ATM build checks this value against internal parameters.
 * - Expression is (MAX_TO_ATM - sizeof(struct SInitStrData) - MAXTCMAP).
 * - Size of structure is max anticipated on any platform.
 */
#define TM_TAASIZE		(3996 - 116 - 28)

struct TmISflags
{
	unsigned absrel : 1;			/* not used */
	unsigned align_option : 2;
	unsigned insert_over : 1;
	unsigned clr_typ_ahead : 1;
	unsigned allow_long : 1;
	unsigned con_ignore : 1;		/* not used */
	unsigned sup_cursor : 1;
	unsigned sup_init : 1; /* suppress output of initial string (see NOTE) */
	unsigned sup_blank : 1;	/* suppress initial blanking */
	unsigned sup_write : 1; /* suppress final writing of string */
	unsigned pass_all : 1;	/* suppress translation to logical keys */
	unsigned monospace : 1;	/* monospace centering and full justification */
	unsigned clear_over : 1; /* clear default if first character overtyped */
	unsigned highlighted : 1; /* field read is highlighted (see NOTE) */
	unsigned tabinput : 1; /* accept input of embedded tabs */

	unsigned reserv2 : 16; /* should be zero */
};
/*	NOTE: In V2 (for COLOR), while a field read is underway:
 *	if `sup_init' is
 *	    FALSE, the ATM removes chars with TMS_X_D and draws with TMS_X_D
 *	    TRUE, if `highlighted' is
 *	        FALSE, the ATM removes chars with TMZEROS and draws with TMS
 *	        TRUE, the ATM removes chars with TMONES and draws with TMIS
 */

typedef struct
{
	unsigned tabType : 3;		/* 1 spare bit */
	unsigned leadering : 2;
	unsigned usage : 2;			/* tab character correspondence option */
	unsigned reserved : 1;		/* should be zero */
	char decimalChar;		/* decimal character for this tab, default is 0 */
	short tabX;					/* local device x position, window pixels */
	short devTabX;		/* position on target device (possibly expanded) */
} TmTabStop;

/* tab types */

#define TM_LEFT_TAB				0
#define TM_CENTER_TAB			1
#define TM_RIGHT_TAB			2
#define TM_DECIMAL_TAB			3
#define TM_DECTAB1	(TM_DECIMAL_TAB + 1)	/* atm internal use only */
#define TM_DECTAB2	(TM_DECIMAL_TAB + 2)	/* atm internal use only */

/* tab leadering values */

#define TM_BLANK_LDR			0
#define TM_DOTTED_LDR			1
#define TM_SOLID_LDR			2

#define TM_MAX_TABSTOPS			100

/* flag definitions */

/* align_option */

#ifndef TM_LEFT_JUST	/* also defined in atmattr.h */
#define TM_LEFT_JUST	0
#endif
#ifndef TM_CENTERED		/* also defined in atmattr.h */
#define TM_CENTERED		1
#endif
#ifndef TM_RIGHT_JUST	/* also defined in atmattr.h */
#define TM_RIGHT_JUST	2
#endif
#ifndef TM_FULL_JUST	/* also defined in atmattr.h */
#define TM_FULL_JUST	3
#endif

/* insert_over */

#define TM_INSERT		0
#define TM_OVERSTRIKE	1

/* clr_typ_ahead */

#define TM_NO_CLEAR_TYP	0
#define TM_CLEAR_TYP	1

/* allow_long */

#define TM_NO_ALLOW_LONG	0
#define TM_ALLOW_LONG		1

/* con_ignore */

#define TM_NO_CON_IGNORE	0
#define TM_CON_IGNORE		1

/* sup_cursor */

#define TM_NO_SUP_CURSOR	0
#define TM_SUP_CURSOR		1

union TmPreData
{
	char char_seq[8];
	struct {
		char ch;
		TmTextDescriptor attr;
	} over_char;
	short log_key;
};
/*
 * Predata types are the following subset of the termination data types:
 * TM_CHAR_SEQ, TM_LOG_KEY, and TM_ATTCHAR_SEQ.
 */

/* masks for optional termination conditions */

#define TMF_LT_BOUNDARY_FAULT		0x00000001L
#define TMF_RT_BOUNDARY_FAULT		0x00000002L
#define TMF_STR_BOUNDARY_FAULT		0x00000004L	/* not used */
#define TMF_SPACE_THRESHOLD			0x00000008L
#define TMF_DISP_OVERFLOW			0x00000010L
#define TMF_BUF_FULL				0x00000020L
#define TMF_F_CLIP					0x00000100L
#define TMF_FIRST_THRESHOLD			0x00004000L
#define TMF_ADD_AT_END				0x00080000L
#define TMF_LAST_OVERSTRUCK			0x00100000L
#define TMF_VDISP_OVERFLOW			0x00200000L
#define TMF_TAB_OVERFLOW			0x00800000L

#define TMF_OPT_TERM				0x00B8413FL	/* all optional conditions */

/* disable local action logical key defines */

#define TMF_INSERT_MODE		00001
#define TMF_OVERSTRIKE_MODE	00002
#define TMF_DC_FORWARD		00004
#define TMF_DC_BACKWARD		00010
#define TMF_DW_FORWARD		00020
#define TMF_DW_BACKWARD		00040
#define TMF_D_EOL			00100
#define TMF_D_BEG_OF_LINE	00200
#define TMF_MC_FORWARD		00400
#define TMF_MC_BACKWARD		01000
#define TMF_MW_FORWARD		02000
#define TMF_MW_BACKWARD		04000
#define TMF_M_EOL			010000
#define TMF_M_BEG_OF_LINE	020000
#define TMF_INS_BLANK		040000
#define TMF_MODE_TOGGLE		0200000

struct TmGEflags
{
	unsigned fr_term		: 1;
	unsigned insert_over	: 1;
	unsigned change    		: 1;
	unsigned motion			: 1;
	unsigned qa_event		: 1;	/* TRUE if running a QA script */
}; 

/* flag definition */

/* fr_term */

#define TM_NO_FIELD_READ	0
#define TM_FIELD_READ_TERM	1

/* insert_over: same as in TmISflags */

/* change */
#ifndef TM_NO_CHANGE	/* also defined in atmreg.h, atmwin.h */
#define TM_NO_CHANGE	0
#endif
#define TM_CHANGE	1

/* td_type */

#define TM_CHAR_SEQ		0
#define TM_LOG_KEY		1	/* caution: not the same as TM_LOGICAL_KEY */
#define TM_PICK_TD		2
#define TM_STROKE		3
#define TM_ATTCHAR_SEQ	4	/* char & attrs */
#define TM_CUCLIP		5	/* cursor clip */
#define TM_LOGWID		6	/* logical key + window id */
#define TM_IB_ATTCHAR	7	/* char (blank) & attrs for insert space key */
#define TM_KEY_SEQ		8	/* logical key, code page, char code */
#define	TM_MOTION		9	/* pointer motion without any buttons pressed */

/* event types defines */

#define TM_LT_BOUNDARY_FAULT	1
#define TM_RT_BOUNDARY_FAULT	2
#define TM_STR_BOUNDARY_FAULT	3	/* not used */
#define TM_SPACE_THRESHOLD		4
#define TM_DISP_OVERFLOW		5
#define TM_BUF_FULL				6
#define TM_BUF_OVERFLOW			7
#define TM_CONTCHAR_TERM		8	/* not used */
#define TM_F_CLIP				9	/* cursor clip by field boundary */
#define TM_LOGICAL_KEY			10  /* caution: not the same as TM_LOG_KEY */
#define TM_LK_EVENT				10  /* new, less confusing symbol */
#define TM_PIK_EVENT			11
#define TM_STK_EVENT			12
#define TM_RESET_EVENT			13
#define TM_CH_TERM				14
#define TM_FIRST_THRESHOLD		15
#define TM_CURSOR_CLIP			16	/* clip by window clipping rectangle */
#define TM_UC_EVENT				17	/* user command event */
#define TM_MENU_REQ				18	/* menu request */
#define TM_ATT_BUF_OVERFLOW		19	/* attribute buffer overflow */
#define TM_ADD_AT_END			20	/* appending in overstrike mode */
#define TM_LAST_OVERSTRUCK		21	/* last unprotected char overstruck */
#define TM_VDISP_OVERFLOW		22	/* vertical overflow of initial field */
#define TM_XFER_CONTROL			23  /* re: mouse xchg w/ "peer" app */
#define TM_TAB_OVERFLOW			24	/* wysiwyg tab causes negative advance */
#define TM_MOTION_EVENT			25  /* pointer motion without any buttons */
									/* keys or buttons pressed */

/* base logical key value for option line picks */

#define TM_OPT_LOG_KEY		128
#define TM_INTERRUPT		TM_OPT_LOG_KEY

struct TmPickTrigger
{
	unsigned b1clk1	: 1;		/* enable regular picks */
	unsigned b1clk2 : 1;		/* enable special ("open") picks */
};

struct TmStkTrigger
{
	unsigned b1strok : 1;		/* enable stroking */
};

/* button values for pick and stroke types */

#define TM_L_BUTTON		0		/* left (or only) button */
#define TM_M_BUTTON		1		/* middle button (3-button device) */
#define TM_R_BUTTON		2		/* right button (device with 2 or more) */

/* button event type (based on combination of button and modifier keys) */
#define TM_UNKNOWN_SELECT	0
#define TM_IGNORE_SELECT	1
#define TM_NORMAL_SELECT	2
#define TM_EXTENDED_SELECT	3
#define TM_MULTIPLE_SELECT	4
#define TM_PASTE_SELECT		5

struct TmPickType
{
	unsigned b1clk1	: 1;		/* regular pick */
	unsigned b1clk2 : 1;		/* special ("open") pick */
	unsigned helpBit : 1;
	unsigned button : 3;
#if THIMBLE
	unsigned shifted : 1;		/* shift key was down */
	unsigned controled : 1;		/* shift key was down */
	unsigned quick_click : 3;	/* number of quick clicks AFTER first */
	unsigned type : 3;  /* logical event based on button,keys,Motif/OL */
#endif
};

struct TmStkType
{
	unsigned b1strok 	: 1;	/* always '1' */
	unsigned button   	: 2;
	unsigned first_point	: 1;
	unsigned last_point	: 1;
	unsigned boundary	: 1;	/* stroke point at window boundary */
	unsigned echoed : 1;	/* stroke point echoed by ATM */
	unsigned clipped : 1;	/* stroke point outside clipping rect */
	unsigned hv_status : 2;	/* h/v line status */
#if THIMBLE
	unsigned shifted : 1;		/* shift key was down */
	unsigned controled : 1;		/* control key was down */
	unsigned type : 3;  /* logical event based on button,keys,Motif/OL */
#endif
};

/* hv_status values */

#define TM_N_LINE	0
#define TM_V_LINE	1
#define TM_H_LINE	2

typedef struct
{
	struct TmPickType tpt;
	short reg_id;
	short x_pick;
	short y_pick;
} TmPickDescriptor;

typedef struct
{
	struct TmStkType tst;
	short x_strok;
	short y_strok;
} TmStkDescriptor;

typedef struct {
	unsigned str_dev	: 1;
	unsigned pick_dev	: 1;
	unsigned strok_dev	: 1;
} DeviceIds;

#if THIMBLE
typedef struct {
	short log_key;		/* set to LK_REGEN */
	short x_position;	/* top left corner of damage */
	short y_position;
	short x_end;		/* bottom right corner */
	short y_end;
} TmRegenData;
#endif

union TmTermData
{
	char char_seq[8];
	struct {
		char ch;
		TmTextDescriptor attr;
	} over_char;
	short log_key;
#if THIMBLE
	TmRegenData regenData;	/* valid only for LK_REGEN */
#endif
	TmPickDescriptor td_pick;
	TmStkDescriptor td_strok;
	struct {
		unsigned topSide	: 1;
		unsigned botSide	: 1;
		unsigned leftSide	: 1;
		unsigned rightSide	: 1;
	} clipCuSide;
	struct {
		short log_key;
		short window_id;
	} logwid;
	struct {
		short log_key;
		short code_page;
		unsigned char ch;
	} key_seq;
};

/* local action codes for strokes */

#define	TMK_NONE       0	/* no local echoing (default) */
#define TMK_PEN        1	/* connect the dots */
#define TMK_ONE_LINE   2	/* rubber-band line */
#define TMK_TWO_LINE   3	/* rubber-band vertex */
#define TMK_RECT       4	/* rubber-band rectangle */
#define TMK_ELLIPSE    5	/* rubber-band ellipse */
#define TMK_HVLINE     6	/* horizontal or vertical line */
#define TMK_MLINE      7	/* movable line */
#define TMK_MRECT      8	/* movable rectangle */
#define TMK_SQUARE     9	/* rubber-band square */
#define TMK_CIRCLE    10	/* rubber-band circle */
#define TMK_MAP       11	/* movable image */
#define TMK_ONE_CURVE 12	/* rubber-band 1 section curve */
#define TMK_TWO_CURVE 13	/* rubber-band 2 section curve */
#define TMK_WRECT     14	/* rubber-band window proportioned rectangle */

struct TmFRpacket		/* main field read parameter packet */
	{
		struct TmISflags flags;
		short field_x;			/* field x position (first character) */
		short field_y;			/* field y position (first character) */
		short field_width;		/* display field width in mils */
		short buf_len;			/* text buffer length */
		short str_len;			/* length of initial string in buffer */
		short cursor_pos;		/* character position of cursor or -1 */
		short milCurPos;		/* pos in window mils if cursor_pos is -1 */
		short scroll_state;		/* horizontal scrolling state */
		long term_map;			/* termination condition enable bitmap */
		long dis_table;			/* local action disable bitmap */
		short pred_type;		/* pre-data type */
		short pred_len;			/* pre-data length */
		union TmPreData pred;	/* pre-data */
		char *buf;				/* text buffer */
		char *charMap;			/* termination character bitmap */
		short mapLen;			/* length of charMap */
		struct QMselect *qmSelect;	/* atm menu selection or NULL */
	};

struct TmFRpacket32		/* main field read parameter packet */
	{
		struct TmISflags flags;
		int field_x;			/* field x position (first character) */
		int field_y;			/* field y position (first character) */
		int field_width;		/* display field width in mils */
		short buf_len;			/* text buffer length */
		short str_len;			/* length of initial string in buffer */
		short cursor_pos;		/* character position of cursor or -1 */
		short milCurPos;		/* pos in window mils if cursor_pos is -1 */
		short scroll_state;		/* horizontal scrolling state */
		long term_map;			/* termination condition enable bitmap */
		long dis_table;			/* local action disable bitmap */
		short pred_type;		/* pre-data type */
		short pred_len;			/* pre-data length */
		union TmPreData pred;	/* pre-data */
		char *buf;				/* text buffer */
		char *charMap;			/* termination character bitmap */
		short mapLen;			/* length of charMap */
		struct QMselect *qmSelect;	/* atm menu selection or NULL */
	};

/*
 * extended parameter packet for field read
 *
 * character level attributes, floating field support, word wrap support,
 * wysiwyg support
 */
struct TmFREpacket		/* field read extension packet */
	{
		short attrBufLen;	/* attribute buffer length (text descriptors) */
		TmTextDescriptor *attrBuf;		/* attribute string buffer or NULL */
		TmTextDescriptor overrideAttr;	/* (not used if attrBuf is NULL) */
		short leftBound;		/* left boundary of editable string or -1 */
		short rightBound;		/* right boundary of editable string or -1 */
		short space_threshold;	/* available space threshold, mils */
		short first_threshold;	/* first word size threshold or -1 */
		short devSetup;			/* device setup for wysiwyg target or -1 */
		short devWidth;		/* field width on wysiwyg target device or -1 */
						/* following are used only if attrBuf is non-NULL: */
		TmTextDescriptor leftAttr;	/* attrs left of field */
		TmTextDescriptor rightAttr;	/* attrs right of field */
		char leftChar;			/* character left of field */
		char rightChar;			/* character right of field */
		short devFieldX;		/* field x position on target device */
	};

struct TmOFWpacket		/* old field write parameter packet (internal use) */
	{
		struct TmISflags flags;
		short field_x;
		short field_y;
		short field_width;
		short str_len;
		short scroll_state;
		char *string;
		TmTextDescriptor *attr_string;
		short field_top;	/* top (lowest value) y coordinate to clear */
		short field_height;	/* vertical range to clear;
								-1: use bounding box (ignore field_top) */
	};

struct TmFWpacket		/* field write parameter packet */
	{
		struct TmISflags flags;
		short field_x;
		short field_y;
		short field_width;
		short str_len;
		short scroll_state;
		char *string;
		TmTextDescriptor *attr_string;
		short field_top;	/* top (lowest value) y coordinate to clear */
		short field_height;	/* vertical range to clear;
								-1: use bounding box (ignore field_top) */
		short devSetup;		/* device setup for wysiwyg target or -1 */
		short devFieldX;	/* field x position on target device */
	};

struct TmHLpacket			/* home line parameter packet */
	{
		short x_position;
		short y_position;
		char *longString;	/* long form string */
		short lenLong;		/* string length or -1 */
		char *shortString;	/* short form string */
		short lenShort;		/* string length or -1 */
		short sep1;			/* first separator or -1 */
		short sep2;			/* second separator or -1 */
		short hHidNum;		/* no of entries in help id array */
		long *helpID;		/* array of help IDs */
	};

/*
 * support for sharing display/mouse/keyboard between ATM and
 * another ("peer") process.
 */
struct  xfer_data
{
    short       xfer_type;
    short       xfer_flags;
    short       win_x;          /* upper left corner in pixels */
    short       win_y;
    short       win_x_siz;      /* in pixels */
    short       win_y_siz;
    short       mous_x;         /* in pixels */
    short       mous_y;
};

/*
 * Xfer types
 */
#define RELINQUISH          1   /* sender is "giving up" indicated devices */
#define RELINQUISH_RQST     2   /* demand a relinquish */

/*
 * Xfer flags -- single bit
 */
#define XFER_MOUSE          1   /* sender is not reading/displaying the mouse */
#define XFER_KEYBOARD       2   /* sender is not reading the keyboard */
#define XFER_DISPLAY        4   /* sender is not writing shared fb area */
#define WINDOW_DEF          8   /* packet's window information is valid */
#define SHUTDOWN            16  /* sender will be calling TmTerm */
 
/*
 * Argument defines for TmXfer()
 */
#define SEND            1       /* peer application -> ATM */
#define RECV            2       /* peer application <- ATM */
 
#define BLOCK           0       /* block until send() returns */
#define NONBLOCK        1       /* return immed. if can't send whole packet */


/*
 * Info for calling TmEInqTxExtent
 */

struct TmTxExtentFlags
{
	unsigned doTabs		: 1;
	unsigned fldRel		: 1;
	unsigned measureAll	: 1;
	unsigned reportPix	: 1; /* output numbers in pixels, not mils */
};

struct TmTxExtentData
{
	/*
	 * Inputs
	 */
	struct TmTxExtentFlags	flags;
	short		pSetup;	/* devnum, -1, or -2 (screen in mono mode) */
	unsigned char *string;
	short		byte_count;
	TmTextDescriptor *attr_string;
	TmTextDescriptor defAttr;
	short		byteindex;		/* measure up to this byte index */
		/* for TmRInq behavior use index = byte_count or strlen */
	short		limit;		/* measure up to this mil limit */
	short		sp_width;
	short		field_just;	/* (if fldRel==1) align_option value */
	short		field_width;	/* (if fldRel==1), in mils */
	short		x_pos;		/* (if doTabs==1) pos of field in pix */
	TmTabStop	*tabs;		/* (if doTabs==1) tab positions in pix, too */
	int			numTabs;	/* (if doTabs==1) */
	int			tabUsage;	/* (if doTabs==1) */
	short		decimalChar;	/* (if doTabs==1) */
	/*
	 * Outputs
	 */
	short		str_status;
	short		word_cut;		/* old cut_pt1 */
	short		char_cut;		/* old cut_pt2 */
	short		word_extent;	/* old extent1 */
	short		char_extent;	/* old extent2 */
	short		string_extent;	/* old extent3 (set if measureAll==1) */
	short		space_count;
	short		string_maxHeight;	/* set if measureAll==1 */
	short		string_maxAboveLine;/* set if measureAll==1 */
	short		word_maxHeight;		/* as of word_cut */
	short		word_maxAboveLine;	/* as of word_cut */
	short		char_maxHeight;		/* as of char_cut */
	short		char_maxAboveLine;	/* as of char_cut */
	short		minHeight;
	short		minAboveLine;
	/*
	 * Tab overflow numbers.  These are set whenever doTabs==1.
	 *
	 * The overflow1 numbers are meaningful when using EInq against
	 * the screen with tabUsage==TM_ACCEPT.  They express the extent
	 * of the 1st tab overflow encountered in measuring the string.
	 * (When TM_ACCEPT, the tabstops will be chosen not according to
	 * to text measurements, but by using an ordered association of
	 * tab characters in the text string to tabstop usage bits as set
	 * prior to calling EInq(ACCEPT).  The usage bits were set according
	 * to the way the same text would set on the target.)  In this
	 * screen-ACCEPT case, it's possible for a piece of text to overshoot
	 * the tabstop that will actually be used on the target.  The amount
	 * of overshoot is reported in the overflow1_stopx and overflow1_ov
	 * numbers.  They say in effect, "if the tabstop had been only
	 * <overflow1_ov> pixels further to the right, overflow wouldn't have
	 * occurred on this stop."  Overflow1_stopx is the pixel position of
	 * the tabstop, and overflow1_ov is the # pixels of overshoot.  The
	 * intent is to use these numbers to determine the minimum increase
	 * in expansion factor needed to eliminate this instance of overflow.
	 * When overflow1_ov is nonzero, then overflow1_ov and overflow1_stopx
	 * are the ONLY outputs of EInq that are meaningful; EInq immediatley
	 * returns TMR_LONG; the expansion factor must be increased, the
	 * string re-measured, and only when overflow1_ov==0 can the string
	 * be displayed reasonably.
	 *
	 * overflow1_ov!=0 should only be possible when comparing measurements
	 * between 'non-matched' screen-printer pairs (like raster screen vs.
	 * lineprinter, or dumb terminal vs. postscript.)
	 *
	 * Overflow2 is always meaningful when doTabs==1.  This number refers
	 * to cases where text which is to be: a) centered on a TM_CENTER_TAB
	 * tabstop, b) right-justified on a TM_RIGHT_TAB or DECIMAL_TAB stop,
	 * and can't strictly be placed in this manner because it won't fit.
	 * (You can see this effect by typing into a right justified tab until
	 * the tabwidth shrinks to its minimum; subsequent text will then set
	 * 'normally' rather than being right-justified on the stop.)  If we
	 * say that characters which couldn't fit in this way were 'pushed',
	 * then overflow2 expresses the highest number of 'pushed' characters
	 * (on a single tabstop) that were encountered across the string.  Ov2
	 * is routine, so EInq runs to completion regardless of this value;
	 * all outputs are meaningful so long as overflow1_ov==0.  The intent
	 * is to allow the application to detect instances where characters
	 * were pushed on the target but not on the screen, or vice versa.
	 * Such cases will be real wysiwyg errors.  But beware -- it is not
	 * possible in general to always achieve an expansion factor which
	 * eliminates overflow1_ov (more important) and produces screen/target
	 * ov2's which are both zero or both nonzero.  If you want to try, tho,
	 * remember that increasing the expansion factor tends to decrease ov2
	 * on the screen.
	 *
	 * Ov2 can be nonzero for either matched or nonmatched comparisons.  But
	 * only for nonmatched should it be possible for the number to be zero
	 * on the screen and nonzero on the target, or vice versa.  In fact,
	 * when matched, the screen-target ov2's should always be equal.
	 */
	short		tab_overflow1_stopx;	/* pix (set if doTabs==1) */
	short		tab_overflow1_ov;		/* pix (set if doTabs==1) */
	short		tab_overflow2;		/* characters (set if doTabs==1) */
};


/* calls */

#if 1
typeErr TmFieldWrite();
typeErr TmWriteTx();
typeErr TmInitString();
typeErr TmGetEvent();
typeErr TmReset();
typeErr TmInitPick();
typeErr TmIStroke();
typeErr TmBeep();
typeErr TmStkAction();
typeErr TmFRead();
typeErr TmHomeLine();
typeErr TmPrompt();
#endif


#endif     /*   ATMTXTIN_INCLUDED   */
