/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: fsrts.h
   Applix ASC version: 18
   Date: Mon Nov 16 13:52:52 1998
*/
#ifndef AX_FSRTS_H
#define AX_FSRTS_H

    /* The type of material which a document contains */
    /* DO NOT CHANGE THESE VALUES - THEY MUST MATCH VALUES IN matinfo */ 
#define MNone   0
#define MTable  1
#define MSpreadsheet 2
#define MCalculator 3
#define MGraph  4
#define MDrawing  5
#define MText  6
#define MExternal  7
#define MUserCommand  8
#define MCalendar  9
#define MQuery  10
#define MUDF    11
#define MFSTable  12
#define MAudio 13
#define MPostit 14
#define MMacro 15
#define MDialog 16
#define MEquation 17
#define MApplixBuilder 18
#define MBitmap 19
#define MApplixBuilderTurbo 20
#define MApplixSlideShow 21
#define MApplixBuilderDistribution 22

    /* A null filing system id */
#define FSNULL  (long) -1

typedef struct tofcEntry {      /* one for each table of contents items */
    char *name;             /* name of entry */
    int task;               /* copied from docFileData's task number */
    long size;              /* number of bytes */
    long filepos;   /* used for sequential SegRead & SegWrite */
    int have_written;       /* set on first write */
    int orphaned;   /* unlinked, buit not closed */
    int num_blocks; /* how many blocks allocated */
    char **blocks;  /* ptr to table of blocks */
} tofcEntry;

typedef struct docFileData {
    long id;                /* file system id for this file */
    char *name;             /* Unix system name */
    int mode;               /* DFM_ values */
    int task;               /* task having this file open */
    char revstamp[20];      /* application defined */
    int fd;                 /* 0 unless opened in DFM_PARTIAL mode */
    int delete_me;  /* TRUE if this is a zombie */
    tofcEntry **segP;       /* ptr to null termed table of file ptrs */
} docFileData;

typedef struct {
    char *name;
    long map, imap; /* HACK - works as long as a pixmap handle is long */
    int wid, hyt;
} user_pixmap;

typedef struct doc_info {       /* all info about a document */
    struct doc_info *next;  /* next in chain or NULL */
    char pathname[STRPATH+1];
    long id;                            /* The document's id */
    int type;   /* The document's material type */
    unsigned long dtmod;
    int num_openers;        /* how many tasks have this doc open */
    short *openers;         /* id's (THIMpid) of those tasks */
    int num_lookers;        /* how many tasks have declared interest */
    short *lookers;         /* id's (THIMpid) of those tasks */
    int num_refers;         /* how many tasks have done a FSGetRefDoc */
    short *refers;          /* id's (THIMpid) of those tasks */
    short owner;            /* task that opened for writing */
    int childDoc;           /* TRUE if this is a member of another file */
    int nPixmaps;
    user_pixmap *pm;
} doc_info;

/* Format of entries read from a 'KS' formatted file.  This includes
        FS KS files, setups, printers and personal profile.
*/

typedef struct ks_entry {       /* line from kitchen sink or profile files */
        struct ks_entry *next;  /* next in chain or NULL */
        char *name;             /* field name */
        char *value;            /* field value */
        int dirty;              /* true to save to local profile */
        int locked;             /* true of value locked by administrator */
} ks_entry;

    /* Info on a document */
typedef struct {
    char pathname[STRPATH+1];
    int is_locked;
    long DocId;     /* address of this struct */
    char DocName[AX_MAX_FILENAME+1];
    int DocMatType;     /* The document's material type */
    docFileData *docFileId; /* id of virtual document file */
    doc_info *sysFileId;    /* id in system pool of doc */
    ks_entry *ksData;   /* id of kitchen sink data */
    int saveMode;   /* 0-binary, 1-ASCII... */
    int page_ix;    /* src page info */
    int src_type;   /* src type */
    unsigned onDisk   : 1,  /* TRUE if file opened/saved under this name */
            childDoc  : 1,  /* TRUE if local inset */
            canIwrite : 1,  /* access check for this user */
            grpAccess : 2,  /* 0-none, 1-read, 2-write */
            allAccess : 2,  /* 0-none, 1-read, 2-write */
            stuff     : 25;
} DocInfoType;

    /* A filing system location */
typedef struct {
    char CabName [STRPATH+1];   /* The cabinet */
    long DrawerId;      /* The drawer */
    long FolderId;      /* The folder */
    long SectionId;     /* The section */
    long SubsectionId;      /* The subsection */
} FSLocType;
    

    /* A null filing system location */
extern DLL_IMPORT FSLocType FSNullLoc;

    /* The current cabinet */
extern DLL_IMPORT struct CabDesc FSCurrCab;

    /* The current filing system location */
extern DLL_IMPORT FSLocType FSCurrLoc;

    /* A null document */
extern DLL_IMPORT DocInfoType FSNullDoc; 

    /* The current document */
extern DLL_IMPORT DocInfoType FSCurrDoc; 

    /* The transfer bin */
extern DLL_IMPORT DocInfoType XferDoc;

    /* A null cabinet */
extern DLL_IMPORT char FSNullCab []; 

    /* Is the current cabinet shared? */
extern DLL_IMPORT short FSCabIsShared;

    /* If the current cabinet is shared, does it 
    reside on this host? */
extern DLL_IMPORT short FSIsLocal;

    /* Declaration of runtimes */

typeErr FSInit ();      /* Initialize FS */
typeErr FSKInit ();     /* Initialize FS - 
                   specify size of kitchen sink cache */
typeErr FSTerm ();      /* Terminate FS */

typeErr FSCrtDoc ();        /* Create document (short form) */
typeErr FSDelDoc ();        /* Delete document */
typeErr FSOKToDelete ();    /* Indicates whether an application may delete 
                    the doc from its Clerk menu */
typeErr FSDelCDoc ();       /* Delete document and its closure  */
typeErr FSPrgDoc ();        /* Purge document */
typeErr FSGtDoc ();     /* Get all info on a document */
typeErr FSChgDoc ();        /* Change a document */
typeErr FSPreOpen ();       /* Prepare a doc for opening */
typeErr FSOpnDoc ();        /* Open document */
typeErr FSClsDoc ();        /* Close document */
typeErr FSChkPtDoc ();      /* Checkpoint an open doc */
typeErr FSRevert ();        /* Revert composition table to last checkpoint*/
typeErr FSLkDoc ();     /* Lock document */
typeErr FSUnlkDoc ();       /* Unlock document */
short FSIsLocked ();        /* Return indication whether a doc is locked */
typeErr FSGetDirs ();       /* Get UNIX dirs associated with a document */
typeErr FSGetMetaDoc ();    /* Get metadocument info  (e.g., fmt tbl) */
typeErr FSSetMetaDoc ();    /* Set metadocument info */
typeErr FSGetRefDoc ();     /* Get interdocument reference */
typeErr FSGetVirtId ();     /* Get virtual id of a doc referenced 
                    from some other doc */
typeErr FSGtRefDocGivenName ();
                /* Get information on a refence document 
                    given its external name */
typeErr FSSetRefDoc ();     /* Set interdocument reference */
typeErr FSDelRefDoc ();     /* Delete an interdocument reference */
typeErr FSStRfFlg ();       /* Set flags for an interdoc reference */
typeErr FSGtRfFlg ();       /* Get flags for an interdoc reference */
typeErr FSBlessRefs ();     /* Synchronize composition table with 
                    updated boom files */
typeErr FSRefCnt ();        /* Determine the number of documents 
                    which reference a document */
typeErr FSDtLastMod ();     /* Get a doc's date last modified */
typeErr FSGtDcName ();      /* Get external name of a document */
typeErr FSGtDcInfo ();      /* Get info on a document given its external 
                   name */
typeErr FSGDInfoGivenID (); /* Get info on a document given its unique ID */
typeErr FSGDIGFInfo ();     /* Get info on a doc in the absence of 
                    an environment  (e.g., Janitor) */
typeErr FSGtCkSrcInfo ();   /* Get checkout info on a document in a 
                    shared cabinet given its DocInfo */
typeErr FSGtCkDocInfoGivenName ();  /* Get checkout info on a doc in a shared cab
                    given its name  */
typeErr FSGCkSrcInfoGivenId (); /* Get chekout info on a doc in a shared cab 
                    given its id */
typeErr FSGtCkCpyInfo ();   /* Get chekout info on a doc in a private cab */
typeErr FSPurgeDoc ();      /* Delete the contents of a document without 
                   removing the document from the filing 
                   system */
typeErr FSSlipDoc ();       /* Slip a document back into the filing system*/
typeErr FSCpyDoc ();        /* Copy a document to an already existing doc */
typeErr FSCtAndCpyDoc ();   /* Copy a document to a new doc */
typeErr FSCCtAndCpyDoc ();  /* Copy documents */
typeErr FSSvCnt ();     /* Save and continue under new name */
typeErr FSCCkIn ();     /* Check a document into a shared cabinet */
typeErr FSUpdate ();        /* Check a document into a shared cabinet */
typeErr FSCkOut ();     /* Check a document out of a shared cabinet */
typeErr FSISCRefDoc ();     /* Start iteration through all referenced docs */
typeErr FSDocFiles ();      /* Determine the UNIX files which 
                    constitute a document's contents. */
typeErr FSSetArchFlds ();   /* Set archive fields in a doc header */
typeErr FSGetArchFlds ();   /* Get archive field values for a doc */
typeErr FSGetDocField ();   /* Get the value of a field in a 
                    document's header */
typeErr FSSetDocField ();   /* Set the value of a field in the doc header*/
int FSNameCount ();     /* Get the # of docs having a given name in the 
                   current cab */



typeErr FSGtLoc ();     /* Get filing system location info given
                   dividers' external names */
typeErr FSGtLcAttrs ();     /* Get attributes of a filing system location
                    given ids */
typeErr FSGLAttrsGivenNames (); /* Get attributes of a filing system location
                    given names */
typeErr FSGtCrLcAttrs ();   /* Get attributes of the current filing system
                   location */
typeErr FSGtCurrLoc ();     /* Get current filing system location */
typeErr FSStCrLoc ();       /* Set current filing system location */
typeErr FSChUser();     /* Change current user */
typeErr FSGtCrDoc ();       /* Get current document */
typeErr FSStCrDoc ();       /* Set current document */
typeErr FSChildChgdLoc ();  /* Inform parents that a child process changed
                    the current filing system location */


#define FLAGLEN 2

#define FSO_READ    0   /* Open just for reading */
#define FSO_OWNER   1   /* Open for read/write */
#define FSO_READ_WRITE FSO_OWNER
#define FSO_EXISTS  2   /* In conjunction with OWNER, specifies file must
                            already exist.  Used in -v mode app opens. */
#define FSO_FILEOK  4   /* Ok if file lacks a header */

typeErr FSOpenDocument();   /* Open and load passed doc info struct */
typeErr FSRenameDoc();  /* Change name & id of doc */
DocInfoType *ACxferDoc();   /* Return ptr to XferDoc */



/* Prototypes for the file locking and related functions */
FILE *GetFilePtrFromCache AXP_((char *pathname, char *mode_str));
int CloseFilePtrFromCache AXP_((char *pathname, FILE *fileP ));
int GetFileDescrFromCache AXP_((char *pathname, int mode));
int CloseFileDescrFromCache AXP_((char *pathname, int fd));
#endif  /* whole file */
