/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: userp.h
   Applix ASC version: 2
   Date: Mon Nov 16 13:52:51 1998
*/
#define UPBUFSIZE 512

/*
 * ENTRY POINTS:
 */

/*
 * module:	UPget - Get profile for a given user and application, possibly
						without looking up the pathname of the user's cabinet.
						(New call, old one being phased out 1/28/85)
 * input:	appID - Application ID
 *			cabinet_path - pathname string to the user's personal cabinet or
				NULL if you want the pathname looked up.
 *			uName - user's name or NULL if just want to use this process's
 *					user name
 * output:	probuffer - Buffer in DIA format of user profile info
 * 			buflen - Profile buffer length
 * returns:	typeErr
 * algorithm:
 * side effects:
 */

extern typeErr UPget( /* IN appID, IN cabinet_path, IN uName, OUT probuffer, INOUT buflen */);
/* int		appID;			/* application ID */
/* char	*cabinet_path;		/* pathname of user's personal cabinet */
/* char	*uName;				/* user's name */
/* char	*probuffer;			/* profile info (DIA format) */
/* int		*buflen;		/* length of profile buffer */


/*
 * module:	UPgetinfo - Get profile for a given user and application
 * input:	appID - Application ID
 *			uName - user's name or NULL if just want to use this process's
 *					user name
 * output:	probuffer - Buffer in DIA format of user profile info
 * 			buflen - Profile buffer length
 * returns:	typeErr
 * algorithm:
 * side effects:
 */

extern	typeErr UPgetinfo( /* IN appID, IN uName, OUT probuffer, INOUT buflen */ );
/* int		appID;			/* application ID */
/* char		*uName;			/* user's name */
/* char		*probuffer;		/* profile info (DIA format) */
/* int		*buflen;		/* length of profile buffer */


/*
 * module:	UPput - Put a DIA format record back into the user's profile, possibly
						without looking up the pathname of the user's cabinet.
						(New call, old one being phased out 1/28/85)
 * input:	appID - Application ID
 *			cabinet_path - pathname string to the user's personal cabinet or
				NULL if you want the pathname looked up.
 *			uName - user's name or NULL if just want to use this process's
 *					user name
 * 			probuffer - Buffer in DIA format of user profile info
 * 			buflen - Profile buffer length
 * returns:	typeErr
 * algorithm:
 * side effects:
 */

extern typeErr UPput( /* IN appID, IN cabinet_path, IN uName, IN probuffer, IN buflen */);
/* int		appID;		/* application ID */
/* char	*cabinet_path;	/* pathname of user's personal cabinet */
/* char	*uName;			/* user's name */
/* char	*probuffer;		/* profile info (DIA format) */
/* int		buflen;		/* length of profile buffer */


/*
 * module:	UPputinfo - Put a DIA format record back into the user's profile
 * input:	appID - Application ID
 *			uName - user's name or NULL if just want to use this process's
 *					user name
 * 			probuffer - Buffer in DIA format of user profile info
 * 			buflen - Profile buffer length
 * returns:	typeErr
 * algorithm:
 * side effects:
 */

extern	typeErr UPputinfo( /* IN appID, IN uName, IN probuffer, IN buflen */ );
/* int		appID;		/* application ID */
/* char		*uName;		/* user's name */
/* char		*probuffer;	/* profile info (DIA format) */
/* int		buflen;		/* length of profile buffer */


/*
 *  Error values that may be returned:
 */

#define	UPOPENERR	(typeErr)(ERRPROF+1)	/* Can't open user profile file */
#define	UPIOERR		(typeErr)(ERRPROF+2)	/* Read, write or seek I/O error */
#define	UPNFDERR	(typeErr)(ERRPROF+3)	/* Can't find info for this application */

#define DtPAPPID	(typeDt)(DtUP_SHORTPRIV+1)  /* application ID */

/*
 * Types for User Profile information.
 *
 * Do not assume any order for the DIA tuples!
 *
 */

#define DtPGAMESS	(typeDt)(DtUP_SHORTPRIV+2)  /* general audible messages */
#define DtPGAALARM	(typeDt)(DtUP_SHORTPRIV+3)  /* general audible alarms */
#define DtPGDATEFOR	(typeDt)(DtUP_SHORTPRIV+4)  /* general date format */
#define DtPGTIMEFOR	(typeDt)(DtUP_SHORTPRIV+5)  /* general time format */
#define DtPGDOWFOR	(typeDt)(DtUP_SHORTPRIV+6)  /* general day of week format */
#define DtPGNUMFOR	(typeDt)(DtUP_SHORTPRIV+7)  /* general number format */
#define DtPGSELABR	(typeDt)(DtUP_SHORTPRIV+8)  /* general abbreviation number */
#define DtPGDELAY	(typeDt)(DtUP_SHORTPRIV+9)  /* general mouse delay */
#define DtPGCHKPNT	(typeDt)(DtUP_SHORTPRIV+10)  /* general checkpoint interval */
												/* other general preferences are
												   defined at the end of this file */

#define DtPDASSIGN	(typeDt)(DtUP_SHORTPRIV+11)  /* delegation assignments */

#define DtPMURGENT	(typeDt)(DtUP_SHORTPRIV+12)  /* mail urgent */
#define DtPMPURGENT	(typeDt)(DtUP_SHORTPRIV+13)  /* mail prompt urgent */
#define DtPMCERTIFIED	(typeDt)(DtUP_SHORTPRIV+14)  /* mail certified */
#define DtPMPCERTIFIED	(typeDt)(DtUP_SHORTPRIV+15)  /* mail prompt certified */
#define DtPMPRIVATE	(typeDt)(DtUP_SHORTPRIV+16)  /* mail private */
#define DtPMPPRIVATE	(typeDt)(DtUP_SHORTPRIV+17)  /* mail prompt private */
#define DtPMREPLY	(typeDt)(DtUP_SHORTPRIV+18)  /* mail reply requested */
#define DtPMPREPLY	(typeDt)(DtUP_SHORTPRIV+19)  /* mail prompt reply requested */
#define DtPMFINAL	(typeDt)(DtUP_SHORTPRIV+20)  /* mail revisable/final form */
#define DtPMPFINAL	(typeDt)(DtUP_SHORTPRIV+21)  /* mail prompt revisable/final form */
#define DtPMDELAYED	(typeDt)(DtUP_SHORTPRIV+22)  /* mail delayed delivery */
#define DtPMTAGS	(typeDt)(DtUP_SHORTPRIV+23)  /* mail specific tags */
#define DtPMEXPIRE	(typeDt)(DtUP_SHORTPRIV+24)  /* mail expiration dates  */
#define DtPMBEHALF	(typeDt)(DtUP_SHORTPRIV+25)  /* mail behalf another */
#define DtPMCDEFAULTS	(typeDt)(DtUP_SHORTPRIV+26)  /* mail confirm defaults */
#define DtPMCNAME	(typeDt)(DtUP_SHORTPRIV+27)  /* mail confirm name */
#define DtPMUNIXMAIL	(typeDt) (DtUP_SHORTPRIV+56) /* receive Unix mail */
#define DtPMFWDDEL	(typeDt) (DtUP_SHORTPRIV+57) /* delete messages after forwarding */
#define DtPMSENDFOR	(typeDt) (DtUP_SHORTPRIV+58) /* Use 'send for' */
#define DtPMPHFILE	(typeDt) (DtUP_SHORTPRIV+59) /* file outgoing phone messages */
#define DtPMMVCP	(typeDt) (DtUP_SHORTPRIV+60) /* Move or Copy documents from inbox. 0 => copy, 1 => move */
#define DtPMCOD	(typeDt)(DtUP_SHORTPRIV+61)	/* mail c-o-d */
#define DtPMPCOD	(typeDt)(DtUP_SHORTPRIV+62)	/* mail prompt c-o-d */
#define DtPMAFORWARD    (typeDt)(DtUP_SHORTPRIV+64) /* Auto forward flag */

#define DtPTVERSIONS	(typeDt)(DtUP_SHORTPRIV+28)  /* text no. of versions */
#define DtPTSAVEOLD	(typeDt)(DtUP_SHORTPRIV+29)  /* text save text of old versions */
#define DtPTCURSOR	(typeDt)(DtUP_SHORTPRIV+30)  /* text initial cursor position */
#define DtPTINSERT	(typeDt)(DtUP_SHORTPRIV+31)  /* text initial insert/overtype */
#define DtPTCLEAR	(typeDt)(DtUP_SHORTPRIV+32)  /* text clear search string */

#define DtPCMEETINGS	(typeDt)(DtUP_SHORTPRIV+33)  /* calendar attend meetings */
#define DtPCDISPLAYED	(typeDt)(DtUP_SHORTPRIV+34)  /* calendar displayed */
#define DtPCACTIVITY	(typeDt)(DtUP_SHORTPRIV+35)  /* calendar use activity id's */
#define DtPCDELEGATIONS	(typeDt)(DtUP_SHORTPRIV+36)  /* calendar leave delegations */
#define DtPCPDELEGATIONS	(typeDt)(DtUP_SHORTPRIV+37)  /* calendar prompt leave delegations */
#define DtPCPREEMPTIVE	(typeDt)(DtUP_SHORTPRIV+38)  /* calendar pre-emptive scheduling */
#define DtPCPPREEMPTIVE	(typeDt)(DtUP_SHORTPRIV+39)  /* calendar prompt pre-emptive scheduling */
#define DtPCEVENTS	(typeDt)(DtUP_SHORTPRIV+40)  /* calendar events private */
#define DtPCPEVENTS	(typeDt)(DtUP_SHORTPRIV+41)  /* calendar prompt events private */
#define DtPCRESOURCES	(typeDt)(DtUP_SHORTPRIV+42)  /* calendar schedule resources */
#define DtPCPRESOURCES	(typeDt)(DtUP_SHORTPRIV+43)  /* calendar prompt schedule resources */
#define DtPCARCHIVE	(typeDt)(DtUP_SHORTPRIV+44)  /* calendar archive activities */
#define DtPCPARCHIVE	(typeDt)(DtUP_SHORTPRIV+45)  /* calendar prompt archive activities */
#define DtPCSCHEDULE	(typeDt)(DtUP_SHORTPRIV+46)  /* calendar scheduling time slot */
#define DtPCPSCHEDULE	(typeDt)(DtUP_SHORTPRIV+47)  /* calendar prompt scheduling time slot */

#define DtPPPNAME	(typeDt)(DtUP_SHORTPRIV+48)  /* print prompt for printer name */
#define DtPPPSETUP	(typeDt)(DtUP_SHORTPRIV+49)  /* print prompt for setup */
#define DtPPPASETUP	(typeDt)(DtUP_SHORTPRIV+50)  /* print prompt for alternate setup */
#define DtPPCOMPLETED	(typeDt)(DtUP_SHORTPRIV+51)  /* print notify when completed */
#define DtPPPCOMPLETED	(typeDt)(DtUP_SHORTPRIV+52)  /* print prompt completed */
#define DtPPBANNER	(typeDt)(DtUP_SHORTPRIV+53)  /* print use banner page */
#define DtPPPBANNER	(typeDt)(DtUP_SHORTPRIV+54)  /* print prompt for banner page */
#define DtPGMEASURE	(typeDt)(DtUP_SHORTPRIV+55)  /* general measurement system */
#define DtPGSMALL	(typeDt) (DtUP_SHORTPRIV+61)  /* Use small menus */
#define DtPGCRL		(typeDt) (DtUP_SHORTPRIV+63)  /* Currency location */
#define DtPGCRS		(typeDt) (DtUP_CHARPRIV+23)   /* currency symbol */



#define DtPPHID         (typeDt)(DtUP_LONGPRIV+1)    /* printer host id */
#define DtPGDURALARM    (typeDt)(DtUP_LONGPRIV+2)    /* duration of alarms */
#define DtPGFRQALARM    (typeDt)(DtUP_LONGPRIV+3)    /* alarm frequency */



#define DtPMDRAWER	(typeDt)(DtUP_CHARPRIV+3)  /* mail default drawer */
#define DtPMFOLDER	(typeDt)(DtUP_CHARPRIV+4)  /* mail default folder */
#define DtPMSECTION	(typeDt)(DtUP_CHARPRIV+5)  /* mail default section */

#define DtPTDRAWER	(typeDt)(DtUP_CHARPRIV+6)  /* mail default drawer */
#define DtPTFOLDER	(typeDt)(DtUP_CHARPRIV+7)  /* mail default folder */
#define DtPTSECTION	(typeDt)(DtUP_CHARPRIV+8)  /* mail default section */

#define DtPPNAME	(typeDt)(DtUP_CHARPRIV+9)  /* print printer name */
#define DtPPSETUP	(typeDt)(DtUP_CHARPRIV+10)  /* print setup */
#define DtPPANAME	(typeDt)(DtUP_CHARPRIV+11)  /* print alt printer name */
#define DtPPASETUP	(typeDt)(DtUP_CHARPRIV+12)  /* print alternate setup */
#define DtPPCLASS       (typeDt)(DtUP_CHARPRIV+13)  /* printer class */

#define DtPMSUBSECT	(typeDt)(DtUP_CHARPRIV+14)  /* mail default subsection */
#define DtPMSTYLEGUIDE	(typeDt) (DtUP_CHARPRIV+15) /* style guide for messages from mail menu */
#define DtPMSNAME	(typeDt) (DtUP_CHARPRIV+16) /* name to use in 'send for' */
#define DtPMPDRAWER	(typeDt) (DtUP_CHARPRIV+17) /* drawer for filing outgoing phone messages */
#define DtPMPFOLDER	(typeDt) (DtUP_CHARPRIV+18) /* folder */
#define DtPMPSECTION	(typeDt) (DtUP_CHARPRIV+19) /* section */
#define DtPMPSSECTION	(typeDt) (DtUP_CHARPRIV+20) /* subsection */
#define DtPMCCNAME		(typeDt) (DtUP_CHARPRIV+21) /* ccname */
#define DtPMAFNAME  (typeDt) (DtUP_CHARPRIV+24) /* Auto forwarding name */


#define DtPDSEND	(typeDt)(DtUP_CONSPRIV+1)  /* delegation users access and send mail */
#if BLD_FEDACL
#define UPSENDMAX	20	/* Maximum number of names for DtPDSEND */
#else
#define UPSENDMAX       6       /* Maximum number of names for DtPDSEND */
#endif

#define DtPDACCEPT	(typeDt)(DtUP_CONSPRIV+2)  /* delegation users accept assignments */
#define UPACCEPTMAX	6	/* Maximum number of names for DtPDACCEPT */

#define DtPDVIEW	(typeDt)(DtUP_CONSPRIV+5)  /* delegation users view and send mail */
#define UPVIEWMAX	20	/* Maximum number of names for DtPDSEND */

/* Startup preferences menu */

#define DtPSAPPLIC	(typeDt) (DtUP_CONSPRIV+3) /* applications to start at
												startup time */
#define DtPSCHOICES	(typeDt) (DtUP_CONSPRIV+4) /* binary choices as to startup */
#define DtPSLOGON	(typeDt) (DtUP_CHARPRIV+22) /* Name of document containing logon command */

#define DtPSBORDERS	(typeDt) (DtUP_SHORTPRIV+62)
		/* Flag indicating old (1) or new(2) style borders */

#define DtPMENVELOPE	(typeDt) (DtUP_CHARPRIV+30) /* Name of document containing envelopes */
#define DtPMFILEPATH	(typeDt) (DtUP_CHARPRIV+31) /* Name of dir for attached files */



	/* Moshe 11/30/87 - filing preferences */
#define DtPFPASSWORD	(typeDt) (DtUP_CHARPRIV+40)  /* Signature password */
#define DtPFTITLE	(typeDt) (DtUP_CHARPRIV+41)  /* Signature title */
#define DtPFACLDISTLIST	(typeDt) (DtUP_CHARPRIV+42)  /* Access control distribution list */
#define DtPFNAME 	(typeDt) (DtUP_CHARPRIV+43)  /* Signature name */

