/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xferfmt.h
   Applix ASC version: 2
   Date: Mon Nov 16 13:52:51 1998
*/
#define MDRevNo	2

/*========================= header structure =======================*/
struct impMD {

	/* Header and version stuff comes first.  The first three fields must
	   stay the same size and type forever so that programs can check
	   for old revisions. */
	short 	impMDsize;	  /* size of the header structure in bytes*/
	char 	machineType;  /* machine type on which file was
							 constructed. Contents to be defined. */
	char 	MDrevision;	  /* Revision number of metadata format */
	long	creID;		  /* File identification information for use by
							 the creating program.  For DC created files,
							 is 1 if the udf file contains text from a
							 single text area in the DC doc, otherwise
							 its a zero. */
	short	creAppid;	  /* Application id of creating program  See
							 scac.h */
	short	creRevision;  /* Revision number of creating program */

	/* Sizes and offsets for the rest of the file start here. */
	short	colCount;	  /* Count of column definitions and length 
							 of Order Vector, including mark column. */
	short	locOrderVector;	/* Byte position in the file of the global or
							   defaultColumn Order Vector. The order Vector 
							   is a sequence of unsigned chars whose values 
							   represent the display order of the columns in
							   records. The 0-th character contains the 
							   column number of the leftmost displayed column.
							   The 1st char, next leftmost, and so on. There
							   are colCount chars. The order vector must 
							   include the Mark column.  */
	short	locColDef;	  /* Byte position in the file of the Column 
							 Definition structures. The structures are 
							 defined below as struct colDefinition. 
							 There are colCount definitions. The column
							 definition structrues must include the Mark 
							 column as the 0th structure.  */
	short	lenCNameStr;  /* Length of Column Heading string, including 
							 a trailing null. */
	short	locCNameStr;  /* Byte position in the file of the Column 
							 Headings string. The string is lenCNameStr 
							 chars long, including a trailing null. The
							 offset and length of each column heading is
							 contained in the column definition structures.*/
	short	recSize;	  /* byte length of table records */
	short	rngCount;	  /* Count of valid line range structures */
	short	locRangeTbl;  /* Byte position in the file of the line Range 
							 structures. There are rngCount structures. 
							 The structures specify the beginning and 
							 ending line numbers of a range which is 
							 validly in the file, as well as the byte 
							 offset into the file where the first line in 
							 the range will be found. Each line will be
							 recSize characters long. If a range contains
							 no records, its end record number must be 
							 exactly one less than its start record number,
							 i.e. start 1, end 0.*/
	char 	maxHeadLines; /* Maximum number of heading lines.
							 Domain: 0-9 */
	char	usedHeadLines;/* Not used in Rev1.20. Will contain the number
							 of heading lines actually required by the
							 current heading texts and column widths. */

	/* Flags start here. */
	char	allowColMarking;/* Allow columns to be marked.
							   Values are: 0 (no), 1 (yes). */
	char 	allowRowMarking;/* Allow rows to be marked. Values
							   are: 0 (no), 1 (yes). */
	char 	fmtMark;	  /* Indicates if mark column (0) should be
							 formatted by the IMP or is already formatted
							 in the file by the creating application */
	char 	truncCols;	  /* Indicates if text in columns is to be 
							 truncated or line wrapped.
							 0 => line wrap. 1 => truncate */
	char	editFormat;	  /* Indicates if the Personal Database user
							 can change the table format def
							 0 => allowed, 1 => not allowed */
	char	opnAny;		  /* Indicates if any fields are openable.
							 0 => look at each column's setting
							 1 => no fields are openable (ignore individual
							      column settings) */

	/* Unused stuff and extra space starts here. */
	long	extraSpace[4];/* Extra space for extensions. */
	short	idOrderVector;/* UNDEFINED - may go away next rev unless someone
							 shows an interest. */
	short	idColDef;	  /* UNDEFINED - may go away next rev unless someone
							  shows an interest. */
	short	idCNameStr;	  /* UNDEFINED - may go away next rev unless someone
							  shows an interest. */
	short	idRangeTbl;	  /* UNDEFINED - may go away next rev unless someone
							  shows an interest. */
};

/*========================= colDefinition structure =======================*/
struct colDefinition {

	/* First come the column descriptions. */
	short	ofsColumn;	  /* offset into each record where field
							 data will be found. 0-relative. */
	unsigned short	disColumn;	  /* display width of field in mils */
	short	lenColumn;	  /* character length of field in record */
	short	colType;	  /* DIA type for data validation */
	short	colValid;	  /* short for additional validation info */
	short	colFormat;	  /* DIA type for file format of data */

	/* Here are the basic column header descriptions. */
	short	lenHeader;	  /* Length of column header text */
	short	ofsHeader;	  /* Offset into header text for column header.*/

	/* Here is the support for proportional fonts. */
	char	separation;	  /* The number of space characters (in the
							 font and size, etc for this column) which are
							 to follow this column as a separator between
							 this column and the next column.  Currently,
							 (4/85) the same value MUST be used for all 
							 columns; a reader of the UDF file may look 
							 at any column's separation and use that 
							 value consistently. */
	char 	justColumn;	  /* Justification of data in display field.
							 Values are TM_LEFT_JUST, TM_CENTERED,
							 TM_RIGHT_JUST and TM_FULL_JUST as defined
							 in atmtextin.h */
	char 	fntColumn;	  /* Code for font to be used. Values are 
							 TMA_NO (a don't care value), TMA_F1, TMA_F2,
							 TMA_F3, and TMA_F4.  TMA_F3 will be a monospace
							 font if a monospace font is available.*/
	char 	fszColumn;	  /* code for font size to be used or DONTCARE */
	short	xpos;		  /* Not used in Rev1.20. Will contain the x
							 coordinate of the leading edge of the column
							 in mils. */
	short	leading;	  /* Not used in Rev1.20. Will contain the between
						     line spacing of the column in a manner TBD. */

	/* Here is the fancy column header formatting stuff. */
	short	hdStartLine;  /* Not used in Rev1.20. Will contain the 0-origin
							 number of the first heading line to contain
							 text for this column.  */
	short	hdMaxWidth;	  /* Not used in Rev1.20. Will contain the maximum
						     width in mils of any line of heading text for
							 this column. */
	short	cutPoints[10];/* Not used in Rev1.20. Will contain the offset
						     into header text of each line of the header. */

	/* Help support is here. */
	long	hlpColumn;	  /* Help id for help on column header. */
	long	hlpField;	  /* Help id for help on fields in column. */

	/* Flags start here. */
	char 	mrkColumn;	  /* 0 => column is not markable.
							 1 => column is markable */
	char 	opnColumn;	  /* 0 => instances of fields in column are 
							      not openable.
							 1 => instances are openable. */
	char	hidColumn;	  /* 0 => column is not hidden.  
							 1 => column is hidden. */
	char	extraChar;	  /* UNDEFINED extra space for future expansion*/

	/* Application specific stuff and extra space brings up the rear. */
	long	appSpecific;  /* Extra long for application specific
							 information.  DC created UDF files record 
							 a complete character attribute word here as
							 default attribute for the field.  The format
							 is the same as used by the ATM.  Some PDB UDF
							 files record the unique id of the format
							 definition column here for use when reloading
							 data after a format change. */
	union {
		struct {
			long	extraSpace [7];	/* UNDEFINED extra space */
			short	extraShort;		/* UNDEFINED extra space */
		} fillr;
		struct { /* For DC to DC transfers only (which use the above extra
					space as follows) */
			unsigned short	leftCent[3];
			unsigned short	right[3];
			char	lines;
			char	alnCodes;
		} dcdc;
		struct { /* For PDB import/export */
			short	sortOrder; 		/* 0=none, +=ascending -=descending */
			short	type; 			/* Internal PDB data type (DT-level) */
			char	size[6]; 		/* Size string for floats */
		} pdb;
	} spce;
};

/* ---Some redefinitions that apply to DC to DC transfers ----*/
#define xLoc   colValid	  /* Temporarily recorded here so that we can
							 later recompute the display width of the 
							 field based on xpos of the next field. Its
							 always cleared to 0 before file is written*/
#define tdefAtt	appSpecific; /* The default attribute for this column */

/*========================= tblRange structure ==========================*/
struct tblRange {

	/* The first three items are the basic range table. */
	long	rngOffset;	  /* byte offset into file for first line
							 in this range of lines.
							 if positive points to first line
							 if negative, abs value points to next 
							 list of range tables */
	short	rngStart;	  /* Inclusive starting line number of first 
							 line in this range. */
	short	rngEnd;		  /* Inclusive ending line number of last line 
							 in this range. Must be 1 less than rngStart 
							 if the range is empty. */

	short	colCount;	  /* Count of column definitions and length of 
							 Order Vector, including mark column.
							 0 -- implies take standard colCount, recSize
							 and ColumnDef from the header impMD */

	/* the next three items will be ignored if colCount == 0 */
	short	recSize;	  /* byte length of table records */

	long	ColumnDef;	  /* Byte position in the file of the
							 Column Definition structures. The
							 structures are defined below as
							 struct colDefinition. There are
							 colCount definitions. The column
							 definition structrues must include
							 the Mark column as the 0th structure.*/

	long	orderVector;  /* Byte position in the file of the
							 Column Order Vector. The order
							 Vector is a sequence of unsigned
							 chars whose values represent the
							 display order of the columns in
							 records. The 0-th character contains
							 the column number of the leftmost
							 displayed column. the 1st char,
							 next leftmost, and so on. There
							 are colCount chars. The order vector
							 must include the Mark column.
							 0 -- implies take standard orderVector
							 from the header impMD */

	/* ------------ Extra items for use in DC to DC transfers -------*/
	short	fractW;		  /* interWPI spacing fraction. */
	short	fractL;		  /* inter line spacing fraction */
	char	atBot;		  /* allow at bottom flag. */
	char	decimalChar;
	char	hyphenate;	  /* Not used in Rev 1. */
	char	fillChar;
	long	fill[4];

	/* Obsolete stuff and extra space. */
	short	idColDef;	  /* UNDEFINED - may go away next rev unless someone
							 shows an interest. */
	short	idOrderVector;/* UNDEFINED - may go away next rev unless someone
							 shows an interest. */
	long	fillLong;		
};

/* GNU Emacs settings - This MUST be the last thing in this file!
 * Local variables:
 * mode: c
 * tab-width: 4
 * comment-column: 26
 * end:
 */
