/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: pfdef.h
   Applix ASC version: 8
   Date: Mon Nov 16 13:56:50 1998
*/
#define CNULL			((char *) 0)
#define SNULL			((short *) 0)
#define LNULL			((long *) 0)
#define FNULL			((FILE *) 0)
#define REGISTER		register

#define NUL			'\0'
#define ZERO		0
#define RESET		0
#define SET			1
#define SPACE		' '
#define NONBREAK_SPACE ((unsigned char) 0xa0)


#define END 			0
#define MILLS_TO_PIXEL		1
#define PIXEL_TO_MILLS		2

#define MONOSPACED 	1
#define PROPORTIONAL	2
#define NUM_CHARS	256

#define UNDERLINE    1
#define DOUBLEUNDER  2
#define STRIKETHRU   4
#define BROKEN       8

#define SUPERSCRIPT  1
#define SUBSCRIPT    2

#define SIZE6PT   4
#define SIZE8PT   5
#define SIZE10PT  0
#define SIZE12PT  1
#define SIZE14PT  2
#define SIZE18PT  3
#define SIZE24PT  6
#define SIZE36PT  7

#define FONTSERIF1  0
#define FONTSANS1   1
#define FONTMONO1   2
#define FONTSYM1    3
#define FONTSERIF2  4
#define FONTSYS1    5
#define FONTSYM2    6
#define FONTFACE8	7
#define FONTFACE9	8
#define FONTFACE10	9
#define FONTFACE11	10
#define FONTFACE12	11

/* fseek commands */
#define FROM_BEG	0
#define FROM_HERE	1
#define FROM_END	2

/*
 *  This inline function returns width of given ASCII char.
 *  It is used by field write for justification, etc.
 */
#define charwid(cc) (PfCurFontDesc->Widths[(unsigned char)cc])


/*
 *  Define globals to be used in determining the technique for pattern
 *  filling in PostScript.  The imagemask method performs transparent fills
 *  but does not produce crisp patterns.  The setscreen method cannot
 *  produce transparent fills, but produces crisp patterns.  Default will be
 *  imagemask method.  Acapps switch 'n' can be used to turn on the setscreen
 *  method, but will be invalid with 'PostScript color' printer setups.
 */
#define SETSCREEN 0
#define IMAGEMASK 1


/***** point *****/

typedef struct {short x, y;} Point;	/* device coordinates */
typedef struct {int x, y;} Point32;	/* device coordinates */

/***** graphics vector (2 pairs of coordinates) *****/

typedef struct {
	short x1;		/* start point, x */
	short y1;		/* start point, y */
	short x2;		/* end point, x */
	short y2;		/* end point, y */
} Vector;

/***** AxRectangle *****/

typedef struct {
	Point origin;
	Point corner;
} AxRectangle;

/*
 * Note: The cell associated with the origin is the
 * lowest-numbered cell inside the rectangle.
 * The cell associated with the corner is diagonally
 * opposite and immediately outside the rectangle.
 */
typedef struct {
	AxRectangle map;		/* location on page */
	int byte_width;		/* width in bytes */
	char *raster;		/* address of raster */
} PBitMap;


/********************
 * Type Definitions *
 ********************/

		/* window control block definition */

struct  wind_struct 
{
	/* general window data */

	struct TmCrePacket wind_attr;

	TmLineDescriptor lin_des;
	TmTextDescriptor write_tx_des;
	int trans_mode;
	int txAlign;		/* top-left or base-left alignment of text */
	int spWid;		/* default space width in pixel */
	AxRectangle inner;	/* inner clipping rectangle */
	AxRectangle appClip;	/* application clipping rectangle */
	PBitMap *bm;		/* page bitmap */
	unsigned int scaledText : 1; /* text may be scaled/rotated */
	Point32 imageBox[4];		/* text footprint when scaledText==1 */
};  /* end wind_struct definition */

struct GLYPH {
	int width, height, b_width, x_offset, y_offset, advance;
	char *bits;
};

struct FONT {
	int height, ascent, width, b_width;
	struct GLYPH *glyph[NUM_CHARS];
};

typedef struct {
	int len;
	char *cp;
} permBuf;

typedef struct {
	short	tabType;
	short ldr;
	short	devTabX;
	short	decimalChar;
} TabStop;

typedef struct {
	TabStop	*stops;
	short	numTabs;
	short	decimalChar;
} TabInfo;

/* built-in globals */


#if !AXMS && !LINUX
/* NT declares these is stdlib.h, Linux declares these in stdio.h */
extern int sys_nerr;
extern char *sys_errlist[];
#endif

/* external functions that must be declared */

char *PFmalloc();
long InLong();


/* Polygon shape constants for use with plotters */

#define	FREEFORM	0
#define	WEDGE   	1
#define RECTANGLE	2

