/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: pfdispat.c
   Applix ASC version: 67
   Date: Sun Jan 17 21:13:04 1999
*/
#include <stdio.h>
#include <signal.h>
#include <setjmp.h>
#include <ctype.h>
#include <aos.h>
#include <paripc.h>
#include <scac.h>
#include <dia.h>
#include <fsrts.h>
#include <print.h>
#include <rsval.h>
#include <pf.h>
#include <atm.h>
#include <atmwin.h>
#include <atmreg.h>
#include <atmattr.h>
#include <atmtxtin.h>
#include <devsetup.h>
#include <pfint.h>
#include <pfextra.h>
#include <sff.h>
#include <getstr5.h>
#include <axprint.h>

#if AXMS
#include <winspool.h>
#endif


#include <pfdispat.p>

/*---------------------------------------------------------------------------------------
	external data
---------------------------------------------------------------------------------------*/
extern  short	IPCdebug;
extern  char	*ACoptarg;
extern  char	*optarg;
extern  char	ThimTempDir[];
extern  int		PfemitLx, PfemitLy;
extern  char	**PfEpsTitles;
extern  char	*DirSlash;
extern	PSinserts PfPsInsertData;


/*---------------------------------------------------------------------------------------
	external procedures
---------------------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------------------
	global procedures
---------------------------------------------------------------------------------------*/
void	pf_markpage() ;

/*---------------------------------------------------------------------------------------
	global data
---------------------------------------------------------------------------------------*/
int		PfJobToPSPrinter;		/* if TRUE: job to printer, not file */

/* global which controls whether image transformation is done
 * in the printer or in the GE
 * TRUE:	transform in GE, print as raster
 *		 - HUGE FILES, VERY LARGE DATA TRANSFER TIME, but may? have
 *			 slightly smaller interpreter time
 * FALSE: use untransformed image, do all transformations in printer
 *		 - MUCH SMALLER FILES, esp. for large color images (atleast 16,
 *			 recorded upto 100 times smaller), but may be a little more
 *			 resource-intensive (sometimes doesn't print on some printers)
 */
int			PfPrintImageAsRaster;
char		*IFName, *OFName, *JFName;
jmp_buf		PfAbortEnv;


/*---------------------------------------------------------------------------------------
	local data
---------------------------------------------------------------------------------------*/
static  jmp_buf		env;
static  AX_BOOLEAN  envset;
static  AX_BOOLEAN  abend;
static  AX_BOOLEAN  PfAbandonCalledAlready;
static  int			parray_n ;
static  long		*parray ;
static  int			start_page ;

/*---------------------------------------------------------------------------------------
	main program of driver process
---------------------------------------------------------------------------------------*/

int PFthimble(pi)
	PF_info *pi ;

{
	int i;
	int retval = 0;
	char *prof_se_file;
	char *new_OFName = (char *) 0;
	int filemask;

	IPCdebug = 0;

	clear_globals();

	/* assign arguments to their proper places */
	OFName = pi->outfile;
	IFName = pi->infile;
	PF.PfDoc_name = pi->docname;
	PF.PfUser_name = pi->user;
	PF.PfHost_name = pi->host;
	PF.PfReverse = pi->reverse ;	
	if (setjmp(PfAbortEnv) != 0)
	{
		retval = -2;
		goto leave_PFthimble;
	}

	PF.PfPrinterName = pi->printer;
	PF.PfBanner = pi->banner;
	PF.PfShell_command = pi->command;
	PF.PfCopies = pi->copies;
	PF.PfDev = pi->device;

	filemask = umask(0);
	/*
		*  Override default protection of read-write for alis only,
		*  since lp daemon must be able to read, and possibly delete,
		*  our printer ready files.
		*  Allowing people to read the log files is nice, too.
		*/

	if (!PF.PfUser_name)
		PF.PfUser_name = NXLT("John Doe");
	if (!PF.PfHost_name)
		PF.PfHost_name = NXLT("unknown host");
	if (!PF.PfDoc_name)
		PF.PfDoc_name = NXLT("unknown document");

	if (!IFName)
		tempAbort( NXLT("PFthimble() - no input serializer file specified") );
	
	if (!OFName)
		tempAbort( NXLT("PFthimble() - no output file specified") );

	if (!PF.PfDoc_date)
	{
		PF.PfDoc_date = (char *)ctime(&pi->date);
		PF.PfDoc_date[24] = '\0';
	}
	if (PF.PfDoc_date[0] == ' ')
		PF.PfDoc_date++;

	/* check user profile to see if image transformation
		* is to be done in the GE or printer */
	if ( UPnumVal(PROF_PRINT_XFORMEDIMG_) )
		PfPrintImageAsRaster = TRUE;
	else
		PfPrintImageAsRaster = FALSE;
	
	init_pf(PF.PfDev);

	memcpy((char *)&PF.pi, (char *)pi, sizeof(PF_info)) ;

	if (PF.pi.source_all && *PF.pi.source_all == 0)
		PF.pi.source_all = NULL ;
	if (PF.pi.source_1st && *PF.pi.source_1st == 0)
		PF.pi.source_1st = NULL ;
	if (PF.pi.source_last && *PF.pi.source_last == 0)
		PF.pi.source_last = NULL ;
	if (PF.pi.dest_all && *PF.pi.dest_all == 0)
		PF.pi.dest_all = NULL ;
	if (PF.pi.dest_1st && *PF.pi.dest_1st == 0)
		PF.pi.dest_1st = NULL ;
	if (PF.pi.dest_last && *PF.pi.dest_last == 0)
		PF.pi.dest_last = NULL ;

#if !AXMS
	/*
		*  if either the shell command or the output file name start with
		*  "/dev/" then we are going directly to the device.
		*/
	if ( (PF.PfShell_command != NULL) && 
			(strncmp ("/dev/", PF.PfShell_command, 5) == 0) )
		OFName = PF.PfShell_command;
#endif

	if (OFName != NULL)
	{
#if AX_X11
		if (OFName[0] == *DirSlash)
		{
			if (strncmp ("/dev/", OFName, 5) == 0)
			{
				PF.DevDirect = TRUE;
				/* remove extra trailing spaces */
				for (i = strlen(OFName)-1; OFName[i] == SPACE; i--);
				OFName[++i] = NUL;
			}
		}
		else
		{
			/* change name so file ends up in ../PRTMP */
			new_OFName = malloc((unsigned) (strlen(OFName)
											+ strlen("/tmp") + 2));
			sprintf(new_OFName, "/tmp/%s", OFName);
			OFName = new_OFName;
		}
#else
		/* a Windows name may start with a drive specifier, ie. "d:\...\..." */
		/* or if it is direct to the device will look like "COM2:"		*/
		if ( pi->print_to_file )
			{
			if ( (OFName[0] != *DirSlash) && (OFName[1] != ':') &&
					(OFName[4] != ':') )
				{
				new_OFName = malloc(strlen(OFName) + strlen(ThimTempDir) + 2);
				sprintf(new_OFName, "%s\\%s", ThimTempDir, OFName );
				}
			/* on Windows writing to a file takes the "direct" route */
			PF.DevDirect = TRUE;	
			}
#endif
	}

	if ( PF.PfCopies <= 0)
	{
		Debugf("Requested %d copies\n", PF.PfCopies);
		Warnf(NXLT("Print formatter warning.\nRequested %d copies\n"), PF.PfCopies);
	}

#if !AXMS
	if (PF.PfShell_command == CNULL || *PF.PfShell_command == 0)
	{
		Debugs("Shell command is null.\n");
		Warns(NXLT("Print formatter warning.\nShell command is null.\n"));
	}
#endif

	PF.PfPrintMultiCopiesInLang = 1; /* How many copies in the language */
	PfJobToPSPrinter = FALSE;
	if ( !pi->print_to_file )
	{
		PfJobToPSPrinter = TRUE;
		if ( PF.PfCopies > 1 )
		{
			if ( UPnumVal(PROF_PRINT_COLLATE_) )
				PF.PfPrintMultiCopiesInLang = 1;
			else
			{
				/* transfer the onus to the printer language */
				PF.PfPrintMultiCopiesInLang = PF.PfCopies;
				PF.PfCopies = 1;
			}
		}
	}

	do {
		if (PF.DevDirect)
		{
			if (!OutOpen (OFName))
				PFabort1("Can't open output file: '%s'", OFName);
		}
		else
		{
			if (!OutCreate(OFName))
#if AXMS
				PFabort1("Can't open printer: '%s'", PF.PfPrinterName );
#else
				PFabort1("Can't open output file: '%s'", OFName);
#endif
		}

		Debugf("OFName = %s\n", OFName);
				
		InOpen (IFName);
		PF.PfPageNum = 0;
		PF.PfPageCat = TM_FIRSTPAGE;
		StartJob ();
		PF.PfJob_done = FALSE;
		start_page = TRUE ;
		while (TRUE)
		{
			if (next_se_record() != 0)
			{		
				InClose ();
				OutJustClose();

				/* remove input serializer file */
				if (PF.in_remove == TRUE)
					unlink(IFName);

				pfFreeData();
				if (new_OFName)
				{
					free(new_OFName);
					new_OFName = (char *)NULL;
				}
				retval = -1;
				goto leave_PFthimble;
			}
			if (PF.PfAbortInProgress)
			{
				PF.LastPage = TRUE;
				PF.PfPageNum++;
				NewPage(FALSE, FALSE, FALSE, FALSE, (TmRect32 *)0, (TmRect32 *)0) ;
				abend = TRUE;
				break;
			}
			if (PF.PfJob_done)
				break;
		}
			/*  Next we finish up the document, but we must handle
				*  cases where interrupts occurs during EndJob.
				*  If we're already aborting, we just do the second
				*  EndJob below and we're done.	If we're not yet aborting,
				*  we execute the first and then check to see whether
				*  an abort occurred, in which case we do EndJob again.
				*/
		PfWriteIt();
		if (!PF.PfAbortInProgress)
			EndJob ();
		if (PF.PfAbortInProgress)
			EndJob();
		
		InClose ();
		OutClose ();

	} while (PF.DevDirect && --PF.PfCopies > 0 && !PF.PfAbortInProgress);

	if (PF.in_remove == TRUE)
	{
		/* should we make a copy of the input serializer file ? */
		/* the profile option overrides environment var */
		prof_se_file = GetProfileByStr(NXLT("axPrintSerialFile"));
		if ( *prof_se_file != '\0' )
			/** leave se file around for printer development */
			Acopy( IFName, prof_se_file );
		else if ( (prof_se_file = getenv(NXLT("SERIAL"))) != (char *)NULL )
			Acopy( IFName, prof_se_file );
		unlink(IFName); /* delete the original serializer file */
	}
	
	pfFreeData();
	if (new_OFName)
		free(new_OFName);

	retval = 0;

leave_PFthimble:
    umask(filemask);
	return(retval);
}

/*---------------------------------------------------------------------------------------
	pfFreeData
---------------------------------------------------------------------------------------*/
void pfFreeData()
{
	freeEPSData();
	freePSinserts(&PfPsInsertData);
	pfFreePatTab();
	pfFreePBitMap(PF.PfPage.bm);
	if (PF.PfTabInfo.stops)
		free( PF.PfTabInfo.stops);
	if (parray)
		free(parray) ;
	if (PF.slist)
		free(PF.slist) ;
}

/*---------------------------------------------------------------------------------------
	next_se_record
---------------------------------------------------------------------------------------*/
static int next_se_record()
{
	int type;
	int error_check;
	int use_crops, use_regis, use_bounds;
	long loc;
	long mark;
	TmRect32 bounds, clip;

	PF.PfSEByteAlignment = 0;	/* to keep track of se file byte alignment */
	envset = TRUE;
	if (setjmp(env))
	{
		envset = FALSE;
		PF.PfJob_done = TRUE;
		PF.PfCopies = 1;
		/* set number of copies so we don't loop in PF.DevDirect case */
		abend = TRUE;
		return(-1);
	}

	if (PF.PfReverse && start_page)
	{
		if (parray_n == 1)
		{
			PF.LastPage = TRUE;
			PF.PfPageNum++;
			PF.PfJob_done = TRUE;
			pf_term();
			return (0) ;
		}
		mark = parray[parray_n - 2] ;
		InGotoMark(mark) ;
		parray_n -- ;
	}

	type = InWord ();

	switch (type)
	{
		case SfCreate:
			if (PF.PfReverse)
			{
				if (start_page)
					start_page = FALSE ;
				else
				{
					start_page = TRUE ;
					return (0) ;
				}
			}
			PF.PfLandscape = InWord();
			PF.PfPageMilWidth = InInt();
			PF.PfPageMilHeight = InInt();
			PF.PfLeftMarginMil = InInt();
			use_crops = InInt();
			use_regis = InInt();
			use_bounds = InInt();
			InBlock((char *)&bounds, sizeof(TmRect32)) ;
			InBlock((char *)&clip, sizeof(TmRect32)) ;
			PF.inputTray = InInt() ;
			PF.outputTray = InInt() ;
			PF.duplex = InInt() ;
						
			loc = InMark();
			error_check = InWord ();
			if (error_check!=SfEndCall)
			{
				Debugf ("Expected end of call, found %d\n", error_check);
				PFabort (NXLT("File out of synch - didn't read end of call"));
			}
			type = InWord();
			PF.LastPage = (type == SfTerm) ? TRUE : FALSE;
			InGotoMark(loc);
			PF.PfPageNum++;
			if (PF.PfPageNum == 1)
				PF.PfPageCat = TM_FIRSTPAGE;
			else if (PF.PfPageNum & 1)
				PF.PfPageCat = TM_ODDPAGE;
			else
				PF.PfPageCat = TM_EVENPAGE;
			pf_create(use_crops, use_regis, use_bounds, &bounds, &clip);
			break;

		case SfTerm:
			if (PF.PfReverse)
			{
				start_page = TRUE ;
				return (0) ;
			}
			PF.LastPage = TRUE;
			PF.PfPageNum++;
			PF.PfJob_done = TRUE;
			pf_term();
			break;

		case SfPolyline:
			pf_pline();
			break;

		case SfArc:
			pf_Arc();
			break;

		case SfSetWinAttr:
			pf_set_win();
			break;

		case SfGMode:
			break;

		case SfImage:
			pf_image();
			break;

		case SfTabs:
			pf_tabs();
			break;

		case SfScaledTxt:
			pf_imgtxt();
			break;

		case SfInclude:
			break;				 /* EPS include OBSOLETE! */

		case SfText:
			pfText();
			break;

		case SfSave:
			pf_save();
			break;

		case SfRestore:
			pf_restore();
			break;

		case SfClip:
			pf_clip();
			break;

		case SfPrControl:
			pf_prcontrol();

		default:
			break;
	}

	error_check = InWord ();
	if (error_check != SfEndCall)
	{
		Debugf ("Expected end of call, found %d\n", error_check);
		PFabort(NXLT("File out of sync - didn't read end of call.")) ;
	}
	envset = FALSE;
	return(0);
}


/*---------------------------------------------------------------------------------------
	Abandon

	Abandon() attempts to flush the document.	It should be called
	when no more input can be accepted, or when some other condition
	makes it impossible to continue formatting the document.
---------------------------------------------------------------------------------------*/

Abandon(reason)
	char *reason;
{	
	Debugf("Abandon(%s)\n", reason);

#if AXMS
	if ( PF.PfPrinterHandle != (HANDLE)NULL );
		{
		MS_AbortPrinter( PF.PfPrinterHandle );
		MS32_ClosePrinter(  PF.PfPrinterHandle );
		PF.PfPrinterHandle = (HANDLE)NULL;
		}
#endif


	if (PfAbandonCalledAlready)
	{
		PF.PfAbortInProgress = TRUE;
		if (envset)
			longjmp(env, 2);
	}
	PfAbandonCalledAlready = TRUE;

	if (envset)
		longjmp(env, 1);
	else
		PFabort(reason);
}

/*---------------------------------------------------------------------------------------
	pf_unlink() unlinks files created by the pf.

	If the stderr and stdout files are non-null, the info
	is reported to ACerrlog before they are unlinked.

	The debug log file (zzpfdebug) is not removed.

	The output file is not removed if the -r switch is on,
	or if the #<r># switch is present in the command line.

	The output file is not removed if the -r switch is on.

---------------------------------------------------------------------------------------*/
void pf_unlink(NormalTerm)
	AX_BOOLEAN NormalTerm;
{

	if (PF.in_remove)
		unlink (IFName);
	if (PF.out_remove && !PF.DevDirect)
		unlink (OFName);
	if (JFName != CNULL)
		unlink (JFName);


#if DEBUGCOMP

	if (PFdebug)
	{
		if (NormalTerm)
			Debugs("\nnormal termination\n");
		else
			Debugs("\nabort\n");
		fclose(PFdfp);
	}
#endif

}


/*---------------------------------------------------------------------------------------
	freeEPSData
---------------------------------------------------------------------------------------*/
static void freeEPSData()
{
	int i;

	if ( PfEpsTitles ) {
		for ( i = 0; i < PF.PfNumEPSIncluded; i++ ) {
			if ( PfEpsTitles[i] )
				free(PfEpsTitles[i]);
		}
		free(PfEpsTitles);
	}
}

/*---------------------------------------------------------------------------------------
	clear_globals

	This module will simply clear the global variables in order to 
	ensure a clean slate 
---------------------------------------------------------------------------------------*/
static void clear_globals() 
{

	/* Set up defaults before checking arguments. */
	PF.PfUser_name = CNULL;
	PF.PfHost_name = CNULL;
	PF.PfDoc_name = CNULL;
	IFName = CNULL;
	OFName = CNULL;
	PF.PfShell_command = CNULL;
	PF.PfDoc_date = CNULL;

	parray_n = 0 ;
	parray = NULL ;

	PF.PfCopies = 1;
	PF.PfBanner = FALSE;
	PF.DevDirect = FALSE;
	PF.PfReverse = FALSE;

	memset(&PF.PfPage,0,sizeof(struct wind_struct));
	memset(&PF.PfTabInfo,0,sizeof(TabInfo));
	PF.PfGXPixFactor = PF.PfGYPixFactor = PF.PfGXMilFactor = PF.PfGYMilFactor = 0;
	PF.PfTXPixFactor = PF.PfTYPixFactor = PF.PfTXMilFactor = PF.PfTYMilFactor = 
	PF.PfDotsPerInch = 0;
	PF.PfLandscape = FALSE;
	PF.LastPage = FALSE;
	PF.PfPrintMultiCopiesInLang = FALSE;
	PfJobToPSPrinter = FALSE;
	PfPrintImageAsRaster = FALSE;
	PF.PfPageMilWidth = PF.PfPageMilHeight = PF.PfPagePixWidth = PF.PfPagePixHeight = 0;

#ifdef DEBUGCOMP
	PF.PFdfp = (FILE *)NULL;
	PF.PFdebug = FALSE;
#endif

	PF.PfUnderAC = TRUE;		/* yes, always in Aster*x */
	PF.PfDev = PF.PfCopies = 0;
	PF.PfJob_done = PF.PfAbortInProgress = FALSE;
	PF.PfHardX = PF.PfHardY = 0;
	PF.PfPatXMul = PF.PfPatYMul = 1;
	PF.PfDev_align = TM_BASELEFT;
	PF.PfPrinterName = (char *)NULL;
	PF.PfBanner = FALSE;
	PF.PfSe_version = 4;

	PF.PfPageCat = PF.PfPageNum = 0;
	PF.PfWin_fg_color = TMA_BLACK;
	PF.PfWin_bg_color = TMA_WHITE;
	PF.PfIn_text = PF.PfText_ok = TRUE;
	PF.PfPoly_shape = PF.PfPoly_radius = PF.PfStart_angle = PF.PfArc_angle = 0;
	memset(&PF.PfWedge_center,0,sizeof(TmPoint));
	PF.ClipGSaveLevel = PF.D70FontNumber = 0;
	PF.InsertSpec = (char *)NULL;
	PF.small = PF.pagetop = PF.PfNumDocPages = 
	PF.PfNumTotalPages = PF.left = PF.right = PF.top = PF.bottom = 0;
	PF.eightbit = PF.eightbitDingbats = FALSE;
	PF.cur_thick = PF.cur_style = -1;
	PF.monospace = FALSE;
	PF.fontNeeded = FALSE;
	PF.PfNumEPSIncluded = PF.PfNumEPSAddedFonts = PF.PfNumEPSProcessed = 0;
	PF.CurPat = -1;
	PF.PFifdes = (FILE *)NULL;
	PF.envset = PF.abend = FALSE;
	PF.in_remove = PF.out_remove = TRUE;
	
	PF.FirstPage = TRUE;
	PF.FirstCreate = TRUE;
	PfAbandonCalledAlready = FALSE;
	PF.PfDolinePatStart = PF.PfDolinePatPtr = CNULL;
	PF.PfDolineBlack = PF.PfDolineCount = 0;
	PfemitLx = PfemitLy = 0;
	PF.scount = 0 ;
	PF.slist = NULL ;
}

/*---------------------------------------------------------------------------------------
	pf_markpage

	Save the se file position of the last word read in parray
---------------------------------------------------------------------------------------*/
void pf_markpage()
{
	long mark ;

	mark = InMark() - sizeof(short) ;
	if (parray_n == 0)
	{
		parray = (long *)malloc(sizeof(long)) ;
		*parray = mark ;
		parray_n = 1 ;
	}
	else
	{
		parray_n++ ;
		parray = (long *)realloc(parray, parray_n * sizeof(long)) ;
		parray[parray_n - 1] = mark ;
	}
}

/*---------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------*/
int pf_tabs()
{
#if 0
	int nt, n;
	TabStop *tsp;

	Debugs ("pf_tabs()\n");

	if (PF.PfTabInfo.stops) PFfree((char *)PF.PfTabInfo.stops);

	nt = PF.PfTabInfo.numTabs = InWord();
	PF.PfTabInfo.decimalChar = InWord();
	PF.PfTabInfo.stops = (TabStop *) PFmalloc(nt * sizeof(TabStop));
	for(tsp=PF.PfTabInfo.stops, n=0; n<nt; n++, tsp++)
	{
		tsp->tabType = InWord();
		tsp->ldr = InWord();
		tsp->devTabX = InWord() + xgpixtotpix(PF.PfXPixOffset);
		tsp->decimalChar = InWord();
	} 
#endif
	return(0);
}
