/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: pfextra.c
   Applix ASC version: 31
   Date: Mon May 10 21:13:16 1999
*/
#include    <stdio.h>
#include    <aos.h>
#include    <pf.h>
#include    <pfextra.h>
#include    <elf.h>
#include    <elf_env.h>
#include    <afile.h>
#include    <paratm.h>
#include    <pfint.h>
#include    <axtext.h>
#include    <axhtab.h>
#include    <axgfs.h>
#include    <pfsoft.h>

#include    <pfextra.p>

extern  char    *ThimApplixDir;
extern  char    *ThimCustomDir;
extern  char    *AGetAxHomeDir();


#define AIR(ch) ((ch)==' '|| (ch)=='\t' || (ch)=='\n' || (ch)=='\0')

/*
 * names of (version3.0) adobe printer description file
 * keys whose values can be stuffed into postscript output.
 */
#define NUM_PPD_KEYS 28
static
ppdKeyTabEnt PPDKeys[NUM_PPD_KEYS] = {
        /* paper controls */
    {0, NXLT("*PageSize") },
    {0, NXLT("*PaperTray") },
    {0, NXLT("*PageRegion") },
    {0, NXLT("*InputSlot") },
    {0, NXLT("*ManualFeed") },
    {0, NXLT("*OutputBin") },
    {0, NXLT("*OutputOrder") },
    {0, NXLT("*Collator") },
    {0, NXLT("*Duplex") },
        /* the rest are queries, or system
            * management keys; good luck.. */
    {0, NXLT("*?PageSize") }, 
    {0, NXLT("*?PaperTray") },
    {0, NXLT("*?ImageableArea") },
    {0, NXLT("*?InputSlot") },
    {0, NXLT("*?ManualFeed") },
    {0, NXLT("*?OutputBin") },
    {0, NXLT("*?OutputOrder") },
    {0, NXLT("*?Collator") },
    {0, NXLT("*?Duplex") },
    {0, NXLT("*?FontList") },
    {0, NXLT("*?FontQuery") },
    {0, NXLT("*?FileSystem") },
    {0, NXLT("*?Resolution") },
    {0, NXLT("*SetResolution") },
    {0, NXLT("*PatchFile") },
    {0, NXLT("*Reset") },
    {0, NXLT("*Password") },
    {0, NXLT("*ExitServer") },
    {0, NXLT("*MainKeyword") },
};

    /* for handling ppd include files. */
#define MAX_NEST 11 /* which works out to 10 levels of nesting max */
FILE * PPDstack[MAX_NEST] = {
    NULL, NULL, NULL, NULL, NULL, NULL, NULL, 
    NULL, NULL, NULL, NULL
};

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
int initPPDkeys()
{
    int i, prHash();
    ppdKeyTabEnt *keys = PPDKeys;

    for (i = 0; i < NUM_PPD_KEYS; i++) {
        keys[i].hash = prHash(keys[i].key, strlen(keys[i].key));
    }
    return TRUE;
}

/*-----------------------------------------------------------------------------
    is the alleged key one of those whose value is postscript code?
-----------------------------------------------------------------------------*/
int isPPDInvocation(keyname)
    char *keyname; /* ppd file line believed to contain key */
{
    char *tp;
    int i, len, hash, prHash();
    ppdKeyTabEnt *keys = PPDKeys;

    if (!keyname)
        return FALSE;

        /* get leftjust, contiguous, non-whitespace portion of key */
    tp = keyname;
    len = 0;
    while(!AIR(*tp) && *tp != '\0') {
        tp++;
        len++;
    }
    /* 1st 'len' chars of keyname to be compared with table vals */
    hash = prHash(keyname, len);
    for (i = 0; i < NUM_PPD_KEYS; i++) {
        if (keys[i].hash == hash) {
            if (strncmp(keyname, keys[i].key, len) == 0) {
                return TRUE;
            }
        }
    }
    return FALSE;
}

/*-----------------------------------------------------------------------------
    'key' could be an alternative expression of the 'primary'
    keyname, e.g., "Papier Ledger" as an alt. for "PageSize Ledger".
    Return whether linebuf represents 'key'.
-----------------------------------------------------------------------------*/
int containsPPDkey(linebuf, key, len)
    char *linebuf;  /* line from ppdfile */
    char *key;      /* keyname we're looking for */
    int len;        /* of keyname */
{
    /* ppd key format:  *<key name>[/<alt. key name>]:
        * The alt. name is typically a foreign-language
        * expression of the keyname; there may be 0 or more
        * of them. */
    char *lp, *kp;

    /* skip the leading '*' chars which are present if this is going to match */
    len--;
    kp = key + 1;
    lp = linebuf + 1;
    while(TRUE) {
        if (strncmp(lp, kp, len) == 0) {
            return TRUE;
        }
        while(*lp != '/' && *lp != ':' && *lp != '\0') {
            lp++;
        }
        if (*lp == ':' || *lp == '\0') {
            return FALSE;
        }
        else {
            lp++; /* skip the '/' */
        }
    }
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
static char *Finder(name)
    char *name;
{
    data_ptr searchList;
    int searchListCount;
    int ix;
    char path[STRPATH];
	char *retP;

again:
    ElfbGetSysVar(NXLT("elf_search_list@"));
    searchList = ElfRetData;
    searchListCount = ElfbArraySize(searchList);
    if ( 0 == searchListCount ) {
        ElfSetStandardSearchList();
        goto again;
    }
    ElfRetData = NULL;
    for ( ix = 0; ix < searchListCount; ix++ )
    {
        AxMakePath(path, StrFromArray(searchList,ix), NXLT("ax_prctl"));
        if ( 0 == access(path,R_OK) )
        {
            ElfDataFree(searchList);
			retP = (char *)malloc(strlen(path)+1);
			strcpy(retP, path);
            return(retP);
        }
    }
    return(NULL);
}

/*-----------------------------------------------------------------------------
    Find and open for reading the file name which should be in axlocal
    Start by looking in user's home, etc. etc.
-----------------------------------------------------------------------------*/
FILE *OpenAxLocal(name)
    char *name ;
{
    char path[2048] ;
    char *p ;
    char *homedir = AGetAxHomeDir();
    FILE *fp;

    if (name == NULL || *name == 0)
        return NULL ;

#if AXMS
    if (*name == '\\' || name[1] == ':')
#else
    if (*name == '/')
#endif
    {
        fp = fopen(name, READ_BINARY_STR);
        return fp ;
    }

    AxMakePath(path, homedir, name);
    if ((access(path, R_OK) == 0))
    {
        fp = fopen(path, READ_BINARY_STR);
        return fp ;
    }

    AxMakePath(path, ThimCustomDir, name);
    if ((access(path, R_OK) == 0))
    {
        fp = fopen(path, READ_BINARY_STR);
        return fp ;
    }

    p = Finder(name);
    if (p == NULL)
        return (NULL);
    fp = fopen(p, READ_BINARY_STR);
	free(p);
    return fp ;
}

/*-----------------------------------------------------------------------------
    Looking for "ax_prctl" which would contain the
    extra postscript text fragments.
-----------------------------------------------------------------------------*/
char *openPSinsertsFile(len)
    int *len;        /* return this */
{
    FILE *fp;
    char  *spec ;
    long lsiz;

    fp = OpenAxLocal(NXLT("ax_prctl"));
    if (fp == NULL)
    {
        *len = 0;
        return (NULL);
    }

    fseek(fp, 0, 2);
    lsiz = ftell(fp);
    spec = (char *)malloc((int)lsiz + 1);
    fseek(fp, 0, 0);
    fread(spec, sizeof(char), (int)lsiz, fp);
    spec[lsiz] = '\0';
    fclose(fp);
    *len = (int)lsiz;
    return spec;
}

/*-----------------------------------------------------------------------------
    Set up a PSinserts instance
    initPSinserts first determines which section of the
    spec, if any, corresponds to printerName.  It then
    calls scanInsertSpec with that subsection
-----------------------------------------------------------------------------*/
int initPSinserts(spec, speclen, ins, printerName)
    char *spec;
    int speclen;
    PSinserts *ins;
    char *printerName;
{
    char *subspec;
    register char *cp, lastChar='\n';
    char *valStart, *valEnd;
    int sublen, i, plen;
    int inStatement = FALSE; /* in <name> = "value" */
    int inComment = FALSE;  /* inside a comment */
    int inText = FALSE;      /* in "value" */
    int inName = FALSE;      /* in <name> */
    int inPrinterStatement = FALSE; /* in PrinterName = "value" */
    int inPrinterSection = FALSE;   /* between printer statements */
        /* set subspec, sublen to correspond to that section
            * of spec which extends from just after the statement:
            *        PrinterName = "<printerName>"
            * to the next PrinterName keyword, or to end of file
            * if none found */
                /* printer names can 'share' a section by using the
                    * syntax: PrinterName = "<p1>", "<p2>", .. , "<pn>" */
    subspec = (char *) NULL;
    sublen = 0;
    valStart = valEnd = (char *) NULL;
    plen = strlen(printerName);
    for(cp=spec, i=0; i<speclen; i++, cp++) {
        if (*cp == '\n')
            inComment = FALSE;
        if (inStatement) {
            if (inText) {
                if (*cp == '"' && lastChar != '\\') {
                    inText = FALSE; /* found trailing quote of value str */
                    if (inPrinterStatement) {
                        valEnd = cp; /* valuestr is that of printer name */
                        if (valStart && (valEnd - valStart > 2)) {
                                /* (they're aimed at the double quotes;
                                    * hence check against 2 for real len) */
                            valStart++;
                            if (strncmp(valStart, printerName, valEnd-valStart) == 0) {
                                    /* valuestr matches printerName */
                                if (cp - spec < speclen - 1) {
                                        /* i.e., there's more stream left */
                                    subspec = cp + 1;
                                    inPrinterSection = TRUE;
                                }
                            }
                        }
                    }
                }
            }
            else {
                if (*cp == '"' && lastChar != '\\') {
                    /* found start of value str */
                    inText = TRUE;
                    inName = FALSE;
                    if (inPrinterStatement) {
                        valStart = cp;
                    }
                }
                else if (*cp == '\n' && lastChar != '\\') {
                    /* statement terminated: nonescaped newline */
                    inStatement = inText = inName = FALSE;
                }
                else if (*cp == '\'' && lastChar != '\\') {
                    /* statement terminated: start of comment */
                    inComment = TRUE;
                    inStatement = inText = inName = FALSE;
                }
                else if (!inName && !AIR(*cp) && *cp != ',') {
                    /* statement terminated: start of another statement */
                    inName = TRUE;
                    if (strncmp(NXLT("PrinterName"), cp, 11) == 0) {
                        inPrinterStatement = TRUE;
                        if (inPrinterSection)
                            break;
                    }
                }
            }
        }
        else if (!inComment) {
            if (*cp == '\'' && lastChar != '\\')
                inComment = TRUE;
            else if (!AIR(*cp) && *cp != '"' && *cp != ',') {
                /* start of a statement */
                inStatement = TRUE;
                if (strncmp(NXLT("PrinterName"), cp, 11) == 0) {
                    inPrinterStatement = TRUE;
                    if (inPrinterSection)
                        break;
                }
            }
        }
        if (inPrinterSection)
            sublen++;
        lastChar = *cp;
    }
    scanInsertSpec(subspec, sublen, ins);
}

/*-----------------------------------------------------------------------------
    get insert data from a stream containing insert statements
-----------------------------------------------------------------------------*/
int scanInsertSpec(spec, speclen, ins)
    char *spec;
    int speclen;
    PSinserts *ins;
{
    int i, numStatements, search;
    int inStatement = FALSE; /* in <name> = "value" */
    int inText = FALSE;      /* in "value" */
    int inName = FALSE;      /* in <name> */
    int inComment = FALSE;  /* inside a comment */
    char *cp;
    char lastChar = '\0';
    patData *curInsert;

    if (!spec || !ins || speclen == 0) {
        return FALSE;
    }
    /* the insert spec statements are of the form:
        * <action name> = "value"+ There may be whitespace
        * and elf-style comments in between.  dbl quote chars
        * and ' chars can be escaped via backslash.
        */
        /* count #statements */
    for(cp=spec, i=0, numStatements=0; i<speclen; i++, cp++) {
        if (*cp == '\n')
            inComment = FALSE;
        if (inStatement) {
            if (inText) {
                if (*cp == '"' && lastChar != '\\') {
                    inText = FALSE; /* found trailing quote of value str */
                }
            }
            else {
                if (*cp == '"' && lastChar != '\\') {
                        /* found start of value str */
                    inText = TRUE;
                    inName = FALSE;
                }
                else if (*cp == '\n' && lastChar != '\\') {
                        /* statement termintated: nonescaped newline */
                    inStatement = inText = inName = FALSE;
                }
                else if (*cp == '\'' && lastChar != '\\') {
                        /* statement termintated: start of comment */
                    inComment = TRUE;
                    inStatement = inText = inName = FALSE;
                }
                else if (!inName && !AIR(*cp) && *cp != ',') {
                        /* statement terminated: start of another statement */
                    inName = TRUE;
                    numStatements++;
                }
            }
        }
        else if (!inComment) {
            if (*cp == '\'' && lastChar != '\\')
                inComment = TRUE;
            else if (!AIR(*cp) && *cp != '"' && *cp != ',') {
                inStatement = inName = TRUE;
                numStatements++; /* start of a statement */
            }
        }
        lastChar = *cp;
    }
    ins->numInserts = numStatements;
        /* alloc that many patData instances onto inserts */
    i = numStatements * sizeof(patData);
    ins->inserts = (patData *)malloc(i);
    memset((char *)ins->inserts, 0, i);
    ins->numInserts = numStatements;
        /* rescan, only this time determine type of each statement,
            * and fully init the patData structs.  While at it, note on
            * numStatements how many are of pattern-search type */
    curInsert = ins->inserts;
    for(cp=spec, i=0, numStatements=0; i<speclen; i++, cp++) {
        if (*cp == '\n')
            inComment = FALSE;
        if (inStatement) {
            if (inText) {
                if (*cp == '"' && lastChar != '\\') {
                    inText = FALSE;
                }
            }
            else {
                if (*cp == '"' && lastChar != '\\') {
                    inText = TRUE;
                    inName = FALSE;
                }
                else if (*cp == '\n' && lastChar != '\\') {
                    inStatement = inText = inName = FALSE;
                }
                else if (*cp == '\'' && lastChar != '\\') {
                    inComment = TRUE;
                    inStatement = inText = inName = FALSE;
                }
                else if (!inName && !AIR(*cp) && *cp != ',') {
                        /* start of a new statement */
                    inName = TRUE;
                    initPatData(curInsert, cp, &search);
                    if (search)
                        numStatements++;
                    curInsert++;
                }
            }
        }
        else if (!inComment) {
            if (*cp == '\'' && lastChar != '\\')
                inComment = TRUE;
            else if (!AIR(*cp) && *cp != '"' && *cp != ',') {
                    /* start of a new statement */
                inStatement = inName = TRUE;
                initPatData(curInsert, cp, &search);
                if (search)
                    numStatements++;
                curInsert++;
            }
        }
        lastChar = *cp;
    }
        /* now to initialize the search patterns member */
    initSearchPats(ins, numStatements);
        /* init table of printer description file keys */
    initPPDkeys();
    return TRUE;
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
void freePSinserts(ins)
    PSinserts *ins ;
{
    int i;
    patData *pd;

    if (!ins->inserts)
        return;
    for(i=0; i<ins->numInserts; i++) {
        pd = &ins->inserts[i];
        if (pd->pattern)
            free(pd->pattern);
        if (pd->text)
            free(pd->text);
    }
    free(ins->inserts);
    ins->inserts = (patData *) NULL;
    if (ins->searchPatterns.pats) {
        free(ins->searchPatterns.pats);
        ins->searchPatterns.pats = (patData **) NULL;
    }
    memset((char *) ins, 0, sizeof(PSinserts));
}

/*-----------------------------------------------------------------------------
    Initialize patData instance from the passed in text,
    which at this point is expected to be:
         <action name> = "value str" ["value str"]
    Syntax isn't checked -- it's expected to be ok....
-----------------------------------------------------------------------------*/
static void initPatData(pd, text, search)
    patData *pd;
    char *text;
    int *search; /* return whether action is of search type */
{
    register int i, len;
    register char *tp;
    char *cp, *copyp, *from, *to;
    char lastChar = '\0';

    /* the action name has no spaces in it */
    cp = tp = text;
    while(*tp && !AIR(*tp) && *tp != '=')
        tp++;
    pd->action = getActionType(cp, tp - text, search);
        /* For non-search-type actions, we're looking for
            * one double-quoted string representing the insert
            * text.  For search-type actions, we're looking for
            * two; the first is the search pattern, the second,
            * the insert text */
    while (*tp && *tp != '"') {
        tp++;
    }
    cp = tp = tp + 1; /* point to start of text */
    while (*tp && !(*tp == '"' && lastChar != '\\')) {
        lastChar = *tp;
        tp++;
    }
    len = tp - cp;
        /* have to copy the string in order to pitch any
            * nonescaped newlines that might be present..
            * owner of the patData instance must free this copy */
    copyp = (char *)malloc(len + 1);
    to = copyp;
    from = cp;
    lastChar = '\0';
    for (i=0; i<len; i++) {
        if ((*from != '\n' && *from != '\\') || lastChar == '\\')
            *to++ = *from;
        lastChar = *from;
        from++;
    }
    *to = '\0';
    pd->text = copyp;

    if (*search) {
            /* what we just copied is actually the search pattern */
        pd->pattern = copyp;
            /* now we must look for the insert text */
        tp++; /* skip dbl quote at end of last stringval */
        while (*tp && *tp != '"') {
            tp++;
        }
        cp = tp = tp + 1; /* point to start of text */
        while (*tp && !(*tp == '"' && lastChar != '\\')) {
            lastChar = *tp;
            tp++;
        }
        len = tp - cp;
            /* have to copy the string in order to pitch any
                * nonescaped newlines that might be present..
                * owner of the patData instance must free this copy */
        copyp = (char *)malloc(tp - cp + 1);
        to = copyp;
        from = cp;
        lastChar = '\0';
        for (i=0; i<len; i++) {
            if ((*from != '\n' && *from != '\\') || lastChar == '\\')
                *to++ = *from;
            lastChar = *from;
            from++;
        }
        *to = '\0';
        pd->text = copyp;
        pd->patlen = strlen(pd->pattern);
        pd->hash = prHash(pd->pattern, pd->patlen);
    }
    else {
        pd->patlen = pd->hash = 0;
    }
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
static void initSearchPats(ins, numStatements)
    PSinserts *ins;
    int numStatements; /* of search type */
{
    int i, j, inOrder;
    patData *tp;
    searchPat *gp = &ins->searchPatterns;
    if (numStatements == 0) {
        gp->numPatterns = 0;
        gp->pats = (patData **)0;
        gp->hash = 0;
        return;
    }
    gp->numPatterns = numStatements;
    gp->pats = (patData **) malloc(numStatements * sizeof(patData *));
    memset((char *)(gp->pats), 0, numStatements * sizeof(patData *));
    gp->hash = 0;
        /* init pats array with addrs of those ins->inserts which
            * are of the search variety */
    for (i=0, j=0; i<ins->numInserts; i++) {
        if (searchType(ins->inserts[i].action))
            gp->pats[j++] = &ins->inserts[i];
    }
        /* the gp->pats array should be sorted from shortest
            * search pattern to longest for prSearch's sake. */
    inOrder = FALSE;
    while(!inOrder) {
        inOrder = TRUE;
        for(i=0; i<numStatements; i++) {
            if (i+1 < numStatements &&
                ((gp->pats[i])->patlen > (gp->pats[i+1])->patlen)) {
                tp = gp->pats[i];
                gp->pats[i] = gp->pats[i+1];
                gp->pats[i+1] = tp;
                inOrder = FALSE;
            }
        }
    }
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
static int searchType(type)
int type;
{
    return(type == InsertJustBeforePattern ||
            type == InsertJustAfterPattern ||
            type == ReplacePatternWith);
}

/*-----------------------------------------------------------------------------
    convert action name string to type int
-----------------------------------------------------------------------------*/
static int getActionType(name, len, searchType)
    char *name;
    int len;
    int *searchType; /* return whether true */
{
    *searchType = FALSE;
    if (strncmp(name, NXLT("InsertAtPrologBottom"), len) == 0)
        return InsertAtPrologBottom;
    if (strncmp(name, NXLT("InsertAtPrologTop"), len) == 0)
        return InsertAtPrologTop;
    if (strncmp(name, NXLT("InsertAtSetupBottom"), len) == 0)
        return InsertAtSetupBottom;
    if (strncmp(name, NXLT("InsertAtSetupTop"), len) == 0)
        return InsertAtSetupTop;
    if (strncmp(name, NXLT("InsertAtPageSetupBottom"), len) == 0)
        return InsertAtPageSetupBottom;
    if (strncmp(name, NXLT("InsertAtPageSetupTop"), len) == 0)
        return InsertAtPageSetupTop;
    if (strncmp(name, NXLT("InsertAtFirstPageSetupBottom"), len) == 0)
        return InsertAtFirstPageSetupBottom;
    if (strncmp(name, NXLT("InsertAtFirstPageSetupTop"), len) == 0)
        return InsertAtFirstPageSetupTop;
    if (strncmp(name, NXLT("InsertAtTrailerTop"), len) == 0)
        return InsertAtTrailerTop;
    if (strncmp(name, NXLT("InsertAtTrailerBottom"), len) == 0)
        return InsertAtTrailerBottom;
    if (strncmp(name, NXLT("InsertAtPageTrailerTop"), len) == 0)
        return InsertAtPageTrailerTop;
    if (strncmp(name, NXLT("InsertAtPageTrailerBottom"), len) == 0)
        return InsertAtPageTrailerBottom;
    if (strncmp(name, NXLT("InsertAtFirstPageTrailerTop"), len) == 0)
        return InsertAtFirstPageTrailerTop;
    if (strncmp(name, NXLT("InsertAtFirstPageTrailerBottom"), len) == 0)
        return InsertAtFirstPageTrailerBottom;
    if (strncmp(name, NXLT("InsertJustBeforePattern"), len) == 0) {
        *searchType = TRUE;
        return InsertJustBeforePattern;
    }
    if (strncmp(name, NXLT("InsertJustAfterPattern"), len) == 0) {
        *searchType = TRUE;
        return InsertJustAfterPattern;
    }
    if (strncmp(name, NXLT("ReplacePatternWith"), len) == 0) {
        *searchType = TRUE;
        return ReplacePatternWith;
    }
    if (strncmp(name, NXLT("PPDFileName"), len) == 0)
        return PPDFileName;
    if (strncmp(name, NXLT("PPDInvocation"), len) == 0)
        return PPDInvocation;
    return -1;
}

/*-----------------------------------------------------------------------------
    this hash function must match code used in prSearch below
-----------------------------------------------------------------------------*/
static int prHash(str, len)
    char *str;
    int len;
{
    int i;
    int hash = 0;
    char *np = str;
    for (i=0; i<len; i++) {
        hash = (hash << 1) ^ (*np++);
    }
    return (hash);
}

/*-----------------------------------------------------------------------------
    The patterns on *gp are arranged from shortest to longest.
    Starting from 'reset', hash the input up to the length of
    the shortest pattern & compare with that pattern's hash.
    If no explicit match, go on to the next pattern, etc..
-----------------------------------------------------------------------------*/
int prSearch(gp, ip, fp, fplen, cp, cplen)
    searchPat *gp; /* in/out: info regarding what to search for etc.. */
    char *ip;        /* in: input stream pos at last reset */
    char **fp;      /* out: stuff to place on output first if nonnull */
    int *fplen; /* out */
    char **cp;      /* out: stuff to place on output after checking fp */
    int *cplen; /* out */
{
    int i, j, match;
    patData *pd;
    char *tp;

    match = FALSE;
    for (i = 0; i < gp->numPatterns; i++) {
        pd = gp->pats[i];
        if (gp->offset > pd->patlen)
            continue;
        for (j = gp->offset; j < pd->patlen; j++) {
            if (ip[gp->offset] == '\0') {
                    /* ran out of input */
                *fplen = *cplen = 0;
                *cp = *fp = (char *)NULL;
                return 0;
            }
            else {
                gp->hash = (gp->hash << 1) ^ (ip[gp->offset]);
                gp->offset++;
            }
        }
        if (gp->offset == pd->patlen && gp->hash == pd->hash) {
            if (strncmp(pd->pattern, ip, gp->offset) == 0) {
                    /* do the action */
                switch (pd->action) {
                    case InsertJustBeforePattern:
                        *fp = pd->text;
                        *fplen = strlen(pd->text);
                        *cp = ip;
                        *cplen = pd->patlen;
                        break;
                    case InsertJustAfterPattern:
                        *fp = ip;
                        *fplen = pd->patlen;
                        *cp = pd->text;
                        *cplen = strlen(pd->text);
                        break;
                    case ReplacePatternWith:
                        *fp = (char *)NULL;
                        *fplen = 0;
                        *cp = pd->text;
                        *cplen = strlen(pd->text);
                        break;
                }
                gp->offset = gp->hash = 0; /* reset */
                return(pd->patlen);
            }
        }
    }
    if (!match) {
        *fp = (char *)NULL; /* nothing to insert */
        *fplen = 0;
        *cp = ip; /* copy input straight thru */
#if 1
        *cplen = 1;
        gp->offset = gp->hash = 0; /* reset */
        return(1);
#else
            /* feeble attempt to improve speed of simplistic algo */
        match = FALSE; /* i.e., no pat char matches stream yet */
        for (tp = ip, i = 0; !match; i++, tp++) {
            for (j = 0; j < gp->numPatterns; j++) {
                pd = gp->pats[j];
                if (i < pd->patlen && *tp == pd->pattern[i]) {
                    match = TRUE;
                    break;
                }
            }
            if (i >= pd->patlen)
                break;
        }
        *cplen = i;
        gp->offset = gp->hash = 0; /* reset */
        return(i);
#endif
    }
    return(0);
}

/*-----------------------------------------------------------------------------
    get references to all extra postscript fragments which are
    of the specified type.
-----------------------------------------------------------------------------*/
patData **getInserts(ins, type, num)
    PSinserts *ins; /* in */
    int type; /* in */
    int *num; /* out: return # inserts */
{
    int i, tnum;
    patData *pd = (patData *)NULL;
    patData **retData = (patData **)NULL;
    if (!ins || !ins->inserts) {
        *num = 0;
        return retData;
    }
            /* first count how many fit the type */
    for(i = 0, tnum = 0; i < ins->numInserts; i++) {
        pd = &ins->inserts[i];
        if (pd->action == type) {
            tnum++;
        }
    }
    if (tnum == 0) {
        *num = 0;
        return retData;
    }
        /* alloc space for return array */
    retData = (patData **) malloc(tnum * sizeof(patData *));
    memset((char *)retData, 0, tnum * sizeof(patData *)); /* ??? */
    *num = tnum;
        /* fill out array */
    for(i = 0, tnum = 0; i < ins->numInserts; i++) {
        pd = &ins->inserts[i];
        if (pd->action == type) {
            retData[tnum++] = pd;
        }
    }
    return retData;
}

/*-----------------------------------------------------------------------------
    fetch the ps string corresponding to the key from ppdFile;
    return either null, or the ps string (which the caller must free).
-----------------------------------------------------------------------------*/
char *doPPDInvocation(ppdFile, key)
    FILE *ppdFile;
    char *key;
{
    long fileMark = 0;
    int tFileMark = 0;
    char linebuf[257], *cp, *tp, *textstr;
    int len, lineCount, textLen, isPPDInvocation(), containsPPDkey();
    FILE *popPPDfile(), *pushPPDfile(); /* for nested includes */
    void closeNestedPPDfiles();
    int inText = FALSE; /* whether pointer inside ps text */
    int found = FALSE;
    if (!ppdFile || !key || key[0] != '*')
        return((char *)NULL);
            /* copy passed in file stream pointer
                * onto PPDstack as 0th entry (top level of nesting) */
    PPDstack[0] = ppdFile;
            /* search for the key's value */
    len = strlen(key);
    lineCount = textLen = 0;
    fseek(ppdFile, 0, 0);
    while(TRUE) {
        tFileMark = ftell(ppdFile);
        if (fgets(linebuf, 256, ppdFile) == NULL)
            if ((ppdFile = popPPDfile()) == NULL)
                break; /* hit end of topmost file */
        if (inText) {
                /* looking for end of text data, signified
                    * by double quote. */
            if (strncmp(linebuf, NXLT("*End"), 4) == 0) {
                closeNestedPPDfiles();
                return ((char *)NULL); /* should've found quote */
            }
            else {
                lineCount++;
                cp = linebuf;
                while(*cp != '"' && *cp != '\0') {
                    textLen++;
                    cp++;
                }
                if (*cp != '"') {
                    continue; /* more line(s) of text to follow */
                }
                else {
                    inText = FALSE;
                    break;
                }
            }
        }
        else {
            if (strncmp(linebuf, NXLT("*Include"), 8) == 0) {
                    /* could be an include file */
                ppdFile = pushPPDfile(ppdFile, linebuf);
            }
                /* look for the key, which should be left-
                    * justified on the line */
            if (linebuf[0] == '*' && /* all ppd keys start with '*' */
                containsPPDkey(linebuf, key, len) &&
                isPPDInvocation(linebuf)) {
                    /* look for and measure the ps text;
                        * it should be wrapped in quotes */
                cp = &linebuf[len];
                while(*cp != '"' && *cp != '\0')
                    cp++;
                if (*cp != '"') {
                    /* any useful text would've started with a quote */
                    closeNestedPPDfiles();
                    return((char *)NULL);
                }
                inText = found = TRUE;
                fileMark = tFileMark;
                lineCount = 1;
                textLen = 0;
                tp = cp + 1;
                while(*tp != '"' && *tp != '\0') {
                    textLen++;
                    tp++;
                }
                if (*tp != '"') {
                    continue; /* possible multiline entry */
                }
                    /* found start and end of text on same line */
                cp++;
                textstr = (char *)malloc(textLen + 1);
                memcpy(textstr, cp, textLen);
                textstr[textLen] = '\0';
                closeNestedPPDfiles();
                return textstr;
            }
        }
    }
    if (found) {
            /* multiline case; return to file position where
                * start of text was found, and reread onto return mem. */
        fseek(ppdFile, fileMark, 0);
        textstr = (char *)malloc(textLen + 1);
        textstr[0] = '\0';
        inText = FALSE;
        while(lineCount--) {
            if (fgets(linebuf, 256, ppdFile) == NULL) {
                freeInsertText(textstr);
                closeNestedPPDfiles();
                return ((char *)NULL);
            }
            if (inText) {
                tp = cp = linebuf;
                while(*tp != '"' && *tp != '\0')
                    tp++;
                *tp = '\0';
                strcat(textstr, cp);
            }
            else {
                inText = TRUE;
                cp = linebuf;
                while(*cp != '"' && *cp != '\0')
                    cp++;
                if (*cp == '"') {
                    cp++;
                    if (*cp != '\0')
                        strcat(textstr, cp);
                }
            }
        }
        closeNestedPPDfiles();
        return textstr;
    }
    else {
        return((char *)NULL);
    }
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
void closeNestedPPDfiles()
{
    /* The caller of doPPDInvocation opens and closes
        * the ppd file pointer.  If present on the PPDstack,
        * it is on the 0th element.  Any other stream ptrs
        * on PPDstack would be those of nested ppd's, and
        * are unknown to the caller.  Close all streams but
        * the 0th */
    int i;
    for (i = 1; i < MAX_NEST; i++) {
        if (PPDstack[i] != NULL) {
            fclose(PPDstack[i]);
            PPDstack[i] = NULL;
        }
    }
}

/*-----------------------------------------------------------------------------
    Open a nested ppd file.  includeLine is expected to
    be of form:  *Include <filename>
-----------------------------------------------------------------------------*/
FILE *pushPPDfile(stream, includeLine)
    FILE *stream;
    char *includeLine;
{
    char *cp, *tp;
    int i;
    if (!includeLine) {
        return(stream);
    }
        /* this should address the 1st whitespace
            * char between the *Include key and the
            * file name string */
    cp = &includeLine[8];
        /* advance to 1st nonwhite */
    while(AIR(*cp) && *cp != '\0') {
        cp++;
    }
    if (*cp == '\0') {
        return(stream);
    }
        /* cp now at start of alleged filename */
    tp = cp;
    while(!AIR(*tp) && *tp != '\0') {
        tp++;
    }
    *tp = '\0'; /* use 1st contig., nonwhite group as filename */
        /* try to open the file */
            /* find first open slot */
    for(i = 0; i < MAX_NEST; i++) {
        if (PPDstack[i] == NULL) {
            break;
        }
    }
    if (i == MAX_NEST) {
        /* we've hit our maximum level of nesting; act as if
            * we couldn't open the file.. */
        return(stream);
    }
    if ((PPDstack[i] = fopen(cp, READ_BINARY_STR)) == NULL) { 
        return(stream);
    }
    return (PPDstack[i]);
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
FILE *popPPDfile()
{
    int i;
    for (i = 1; i < MAX_NEST; i++) {
        if (PPDstack[i] == NULL) {
            break;
        }
    }
    if (i == 1) {
        /* there's only one entry on the stack;
            * that of the caller-visible stream pointer
            * for the top-level file; caller will close it */
        return (NULL);
    }
        /* close deepest; return next one up */
    i--;
    fclose(PPDstack[i]);
    PPDstack[i] = NULL;
    return(PPDstack[i-1]);
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
void freeInsertsFile(fileImage)
    char *fileImage;
{
    free(fileImage);
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
void freeInsertList(list)
    char *list;
{
    free(list);
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
void freeInsertText(text)
    char *text;
{
    free(text);
}





/*-----------------------------------------------------------------------------
    pfSoftFontRegister

    Register a font for downloading
-----------------------------------------------------------------------------*/
PF_soft *pfSoftFontRegister(ftab, pointsize, ctm)
    fontTabEnt *ftab;
    int pointsize ;
    axctm *ctm ;
{
    int ix ;
    PF_soft s ;
    PF_soft *sl = PF.slist ;

    s.ftab = ftab ;
    s.pointsize = pointsize ;
    s.ctm = *ctm ;

    for (ix = 0 ; ix < PF.scount ; ix++)
    {
        if (s.ftab == sl[ix].ftab &&
            s.pointsize == sl[ix].pointsize &&
            s.ctm.ca == sl[ix].ctm.ca &&
            s.ctm.cb == sl[ix].ctm.cb &&
            s.ctm.cc == sl[ix].ctm.cc &&
            s.ctm.cd == sl[ix].ctm.cd &&
            s.ctm.ctx == sl[ix].ctm.ctx &&
            s.ctm.cty == sl[ix].ctm.cty)
            return &sl[ix] ;
    }

    s.id = 20000 + PF.scount ;
    memset(s.needed, 0, 256) ;
    PF.scount++ ;
    sl = (PF_soft *)realloc(sl, PF.scount * sizeof(PF_soft)) ;
    sl[PF.scount-1] = s ;
    PF.slist = sl ;
    return &sl[PF.scount-1] ;
}

/*-----------------------------------------------------------------------------
    pfSoftCharRegister

    Register character x in softfont s as needed
-----------------------------------------------------------------------------*/
void pfSoftCharRegister(s, x)
    PF_soft *s ;
    tx_char x ;
{
    s->needed[(unsigned char)x] = TRUE ;
}


