/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: pfinf.c
   Applix ASC version: 30
   Date: Mon Nov 16 13:56:50 1998
*/
#include <stdio.h>
#include <errno.h>
#include <seprelud.h>
#include <pf.h>
#include <aos.h>
#include <paratm.h>
#include <paripc.h>
#include <devsetup.h>
#include <pfint.h>
#include <axtext.h>

#include <pfinf.p>

#define TBUFSIZ 80

FILE *PFifdes;

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
InOpen (fname)
char *fname;
{
#if AXMS
    PFifdes = fopen (fname, "rb");
#else
    PFifdes = fopen (fname, "r");
#endif
	/* open a file to read */
    if (PFifdes == FNULL)
	{
		Debug3f(NXLT("Error opening serial file '%s': errno = %d: %s\n"),
				fname, errno,
				((unsigned) errno < sys_nerr) ? sys_errlist[errno] : "?");
		PFabort ("Can't open input (serial) file.");
	}
    InPrelude ();
	/* skip prelude stuff */
}

/*-----------------------------------------------------------------------------
 *  InClose closes the input file.
-----------------------------------------------------------------------------*/
InClose ()
{
    fclose(PFifdes);
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
long InMark ()
{
	long mark;

	mark = 0;
	if (PFifdes == FNULL)
		PFabort ("InMark when input file not open.");
	else
		mark = ftell (PFifdes);

	Debugf ("InMark.  mark = %ld\n", mark);
	return mark;
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
InGotoMark (mark)
long mark;
{
	Debugf ("InGotoMark.  mark = %ld\n", mark);
	if (PFifdes != FNULL && fseek (PFifdes, mark, 0) == -1)
		PFabort ("InGotoMark() seek error.");
}

/*-----------------------------------------------------------------------------
 *  InByte reads a byte from the input file.
 *  It returns the byte.
-----------------------------------------------------------------------------*/
char InByte ()
{
	int c;

	PF.PfSEByteAlignment = !PF.PfSEByteAlignment;
	c = getc (PFifdes);
	if (c == EOF)
	{
		Abandon(NXLT("Unexpected end of file."));
		return 0;
	}
    return c;
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
InAlign()
{
	if ( PF.PfSEByteAlignment )
		InByte();
}

/*-----------------------------------------------------------------------------
 *  InWord reads a word from the input file.
 *  It returns the word.
-----------------------------------------------------------------------------*/
InWord ()
{
	/* version 2 or greater */
	short w;
		
	w = (InByte() << 8) & 0xff00;
	w |= InByte() & 0xff;
		
	return (int) w;
}

/*-----------------------------------------------------------------------------
 *  InLong reads a long from the input file.
 *  It returns the long.
-----------------------------------------------------------------------------*/
long InLong ()
{
	/* version 2 or greater */
	long l;

	l = ((long)InWord() << 16) & 0xffff0000;
	l |= InWord() & 0xffff;
	return l;
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
int InInt()
{
	int ii;

	ii = ((int)InWord() << 16) & 0xffff0000;
	ii |= InWord() & 0xffff;
	return ii;
}

/*-----------------------------------------------------------------------------
 *  InCString reads a counted string.  It reads n+1 characters into *s.
 *  The n+1'th character is a null.
 *  If there an odd number of characters, an extra byte is read so that
 *  each read starts on a word boundary.
-----------------------------------------------------------------------------*/
InCString (s, n)
register char *s;
register int n;
{
    register int i;

    for (i=0;i<n;i++)
        *s++ = InByte ();
    *s='\0';
    if (i & 1)
    	InByte ();
}

/*-----------------------------------------------------------------------------
 *  InBlock reads n characters into s
-----------------------------------------------------------------------------*/
InBlock (s, n)
register char *s;
register int n;
{
    register int i;

    for (i=0;i<n;i++)
        *s++ = InByte ();
}

/*-----------------------------------------------------------------------------
 *  InString - Reads null terminated string.
 *  Reads out pad char if present.
-----------------------------------------------------------------------------*/
InString (s)
register char *s;
{
	register int	i=1;	/* string plus null (and pad) will be even # */

    while (*s = InByte ()) { s++; i++; }
    if (i & 1)
		InByte ();
}

/*-----------------------------------------------------------------------------
 *  InString - Reads null terminated string.
-----------------------------------------------------------------------------*/
InNString(s)
register char *s;
{
    while ( *s = InByte () ) {
		s++;
	}
}

/*-----------------------------------------------------------------------------
 * : InTxCharString - reads in a string of tx_char's.
 * No null.  Returned string must be freed by caller
-----------------------------------------------------------------------------*/
int
InTxCharString(textString)
tx_char **textString; /* out */
{
	tx_char *tbuf;
	int ix;
	int len = InInt();
	if (!textString) {
		return (0);
	}
	if (!len) {
		*textString = (tx_char *) NULL;
		return (0);
	}
	tbuf = (tx_char *) malloc (len * sizeof (tx_char ));
	if (sizeof(tx_char) == 1) {
		InNString((char *) tbuf);
	}
	else {
		for (ix = 0; ix < len; ix++) {
			tbuf[ix] = (tx_char) InWord();
		}
	}
	*textString = tbuf;
	return len;
}


/*-----------------------------------------------------------------------------
 *  InPrelude skips the serial file prelude.
 *  It reads the three prelude lines.  Each prelude line is terminated
 *  by a '\n'.
-----------------------------------------------------------------------------*/
static InPrelude ()
{
	int len;
    char tbuf[TBUFSIZ];
    long file_pos;

file_pos = ftell(PFifdes);
    fgets (tbuf, TBUFSIZ, PFifdes);
file_pos = ftell(PFifdes);
	len = strlen(tbuf);
	if (len <= 0)
		PFabort("Unknown se file prelude.");
	if (tbuf[len - 1] == '\n')
		tbuf[len - 1] = 0;

    Debugf ("Prelude 1: '%s'\n", tbuf);
	if (strcmp(SE_PRL_V1, tbuf) == 0)
	{
		Debugs("SE file format version 1.\n");
		PF.PfSe_version = 1;
	}
	else if (strcmp(SE_PRL_V2, tbuf) == 0)
	{
		Debugs("SE file format version 2.\n");
		PF.PfSe_version = 2;
	}
	else if (strcmp(SE_PRL_V3, tbuf) == 0)
	{
		Debugs("SE file format version 3.\n");
		PF.PfSe_version = 3;
	}
	else if (strcmp(SE_PRL_V4, tbuf) == 0)
	{
		Debugs("SE file format version 4.\n");
		PF.PfSe_version = 4;
	}
	else
		PFabort("Unknown se file prelude.");
    fgets (tbuf, TBUFSIZ, PFifdes);
file_pos = ftell(PFifdes);
    Debugf ("Prelude 2: '%s'\n", tbuf);
    fgets (tbuf, TBUFSIZ, PFifdes);
file_pos = ftell(PFifdes);
    Debugf ("Prelude 3: '%s'\n", tbuf);

    file_pos = ftell(PFifdes);

    if (file_pos & 1L)
	    InByte ();

}

/*-----------------------------------------------------------------------------
	 * read in axtext.h:textAttrs instance.  the passed
	 * arguments are treated as pointers, and returned
	 * info is allocated onto them -- caller must free
-----------------------------------------------------------------------------*/
int InTextAttrs(atts, fontName, underPos, underThick, strikePos)
textAttrs **atts; /* out */
char **fontName;  /* out */
int *underPos;    /* out - underline position relative to baseline */
int *underThick;  /* out - underline thickness */
int *strikePos;  /* out - strikethru baserel position */
{
	int len;
	textAttrs *tatts;

	if (!atts || !fontName) {
		return FALSE;
	}
	tatts = (textAttrs *) malloc (sizeof(textAttrs));
	memset((char *)tatts, 0, sizeof(textAttrs));
	len = InInt(); /* length of font family string */
	if (len) {
		tatts->fontFamily = (char *) malloc (len + 1);
		InCString(tatts->fontFamily, len);
	}
	tatts->weight         = InInt();
	tatts->slant          = InInt();
	tatts->pointsize      = InInt();
	tatts->ruling         = InInt();
	tatts->strike_thru    = InInt();
	tatts->supx           = InInt();
	tatts->supy           = InInt();
	tatts->kernx          = InInt();
	tatts->kerny          = InInt();
	tatts->noRuleChars[0] = InWord();
	tatts->noRuleChars[1] = InWord();
	tatts->angle          = InInt();
	tatts->xscale         = InInt();
	tatts->yscale         = InInt();
	tatts->xshear         = InInt();
	tatts->fillcolor      = InInt();
	tatts->fillbcolor     = InInt();
	tatts->strokecolor    = InInt();
	tatts->strokebcolor   = InInt();
	len                   = InInt(); /* bool: whether there's a fillpattern */
	if (len) {

	/* The (address of) the memory returned by malloc() will already have
	   char * alignment.  attr_type * alignment is probably more
	   restrictive and, as we will eventually access tatts->fillpattern as
	   an attr_type *, that is what we should initialize it to.
	*/

#if CLIX && 0 /* must be a char* for clix to build  */
		tatts->fillpattern = (char *) malloc (sizeof(attr_type));
#else
		tatts->fillpattern = (attr_type *) malloc (sizeof(attr_type));
#endif  /* CLIX */
		pf_read_poly_attr(tatts->fillpattern);
	}

	*atts = tatts; /* caller should use pfFreeTextAttrs to free */

	len = InInt(); /* length of font name string */
	if (len) {
		*fontName = (char *) malloc (len + 1);
		InCString(*fontName, len);
	}
	else {
		*fontName = (char *) NULL;
	}
	*underPos = InInt();
	*underThick = InInt();
	*strikePos = InInt();
	return TRUE;
}

/*-----------------------------------------------------------------------------
	 * Read whole-field text properties; currently,
	 * this is just transform info
-----------------------------------------------------------------------------*/
int
InTxFieldProps(atts, xZoom, yZoom, draw_shadow, shadow_only, draw_emboss)
textAttrs *atts;
int *xZoom, *yZoom;
int *draw_shadow;
int *shadow_only;
int *draw_emboss;
{
	if (!atts)
		return FALSE;
	atts->angle = InInt();
	atts->xscale = InInt();
	atts->yscale = InInt();
	atts->xshear = InInt();
	*xZoom = InInt();
	*yZoom = InInt();
	*draw_shadow = InInt();
	*shadow_only = InInt();
	*draw_emboss = InInt();
	return TRUE;
}
