/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: pfpol.c
   Applix ASC version: 16
   Date: Mon Nov 16 13:56:50 1998
*/
#include <stdio.h>
#include <aos.h>
#include <afile.h>
#include <pf.h>
#include <paratm.h>
#include <paratmgr.h>
#include <devsetup.h>
#include <pfint.h>

#include <pfpol.p>

#define MAX_CUT	50
#define PAT_WID		8
#define PAT_HT		8

#define MAX_N_PATS	12

/* array of pattern definitions */

PBitMap *pattab[MAX_N_PATS];

static PBitMap *patbitmap;

int CurPat;

int PfemitLx, PfemitLy;

static int cut[MAX_CUT];


/***********************************************************************\
*			    POLY-CLIP STUFF				*
\***********************************************************************/

#define MAX_LEVEL	4
#define MAX_POLY	TM_POINT_MAX

#define BELOW(ty)	((clipArray[2].y <= ty) << 3)
#define RIGHT(tx)	((clipArray[2].x <= tx) << 2)
#define ABOVE(ty)	((ty < clipArray[0].y) << 1)
#define LEFT(tx)	(tx < clipArray[0].x)

static TmPoint clipArray[5], *clipdPtr, clipdPoly[MAX_POLY * 2];
static *edgePtr, edgePoly[MAX_POLY * 2];
static short edgeCnt;

/***********************************************************************/

PfLoadClip()
{
	int left, right, top, bottom;

	left = PF.PfPage.appClip.origin.x;
	top = PF.PfPage.appClip.origin.y;
	right = PF.PfPage.appClip.corner.x;
	bottom = PF.PfPage.appClip.corner.y;

	Debug4f(NXLT("PfLoadClip: left %d, top %d, right %d, bottom %d\n"),
			left, top, right, bottom);

	clipArray[0].x = left;	clipArray[0].y = top;
	clipArray[1].x = right;	clipArray[1].y = top;
	clipArray[2].x = right;	clipArray[2].y = bottom;
	clipArray[3].x = left;	clipArray[3].y = bottom;
	clipArray[4].x = left;	clipArray[4].y = top;
}

/***********************************************************************/

static interc(i, x1, y1, x2, y2, rx, ry)
short i, x1, y1, x2, y2, *rx, *ry;
{
	long dx, dy;

	dx = x2 - x1;
	dy = y2 - y1;
	if (i & 1) {
		if (!dy)
			return FALSE;
		*ry = clipArray[i].y;
		if ((y1 < *ry) != (y2 < *ry)) {
			*rx = x1 + dx * (*ry - y1) / dy;
			return TRUE;
		}
		return FALSE;
	}
	if (!dx)
		return FALSE;
	*rx = clipArray[i].x;
	if ((x1 < *rx) != (x2 < *rx)) {
		*ry = y1 + dy * (*rx - x1) / dx;
		return TRUE;
	}
	return FALSE;
}

/***********************************************************************/

static emit(code, x, y)
int code;
short x, y;
{
	static short lx, ly;

	if (clipdPtr != clipdPoly && x == PfemitLx && y == PfemitLy)
		return;

	clipdPtr->x = PfemitLx = x;
	clipdPtr->y = PfemitLy = y;
	clipdPtr++;
	*edgePtr++ = code;
}

/***********************************************************************/

static clipSeg(lvl, x1, y1, x2, y2)
short lvl, x1, y1, x2, y2;
{
	short v, h, tmp, i, tx, ty;

	for (i = lvl; i < 4; i++) {
		if (interc(i, x1, y1, x2, y2, &tx, &ty)) {
			clipSeg(i + 1, x1, y1, tx, ty);
			clipSeg(i + 1, tx, ty, x2, y2);
			return;
		}
	}

	tx = (x1 >> 1) + (x2 >> 1);
	ty = (y1 >> 1) + (y2 >> 1);
	h = RIGHT(tx);
	if (h == 0) h = LEFT(tx);
	v = BELOW(ty);
	if (v == 0) v = ABOVE(ty);

	tmp = v | h;
	if (!tmp) {
		emit(FALSE, x1, y1);
		emit(TRUE, x2, y2);
	}
	else if (v && h) {
		tmp >>= 2;
		if (1 < tmp)
			tmp = 5 - tmp;
		emit(FALSE, clipArray[tmp].x, clipArray[tmp].y);
	}
}

/***********************************************************************/

PfClipPoly(pts, cnt, fill, closed)
TmPoint **pts;
int *cnt;
int closed, fill;
{
	REGISTER int i;
	REGISTER TmPoint *p1, *p2;

	p2 = *pts;
	p1 = p2++;
	clipdPtr = clipdPoly;
	*(edgePtr = edgePoly) = 0;
	for (i = 1; i < *cnt; p1++, p2++, i++)
		clipSeg(0, p1->x, p1->y, p2->x, p2->y);

	if (!closed)
		edgeCnt = edgePtr - edgePoly;

	if (fill || closed) {
		clipSeg(0, p1->x, p1->y, (*pts)->x, (*pts)->y);
		if (clipdPtr != clipdPoly)
			emit(*edgePoly, clipdPoly[0].x, clipdPoly[0].y);
	}

	if (closed)
		edgeCnt = edgePtr - edgePoly;

	*pts = clipdPoly;
	*cnt = clipdPtr - clipdPoly;
}

/***********************************************************************/

int PfBorderLine(indx)
short indx;
{
	return indx < edgeCnt && edgePoly[indx];
}

void
pfFreePatTab()
{
	int ix;
	void pfFreePBitMap();
	for (ix = 0; ix < MAX_N_PATS; ix++) {
		pfFreePBitMap(pattab[ix]);
		pattab[ix] = (PBitMap *) 0;
	}
}

void
pfFreePBitMap(patbitmap)
PBitMap *patbitmap;
{
	if(patbitmap) {
		if (patbitmap->raster) {
			free(patbitmap->raster);
		}
		free(patbitmap);
	}
}

/******************************** END **********************************/
