/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: pfwin.c
   Applix ASC version: 21
   Date: Mon Nov 16 13:56:50 1998
*/
#include <stdio.h>
#include <paripc.h>
#include <pf.h>

#include <atm.h>
#include <atmwin.h>
#include <atmreg.h>
#include <atmattr.h>
#include <atmtxtin.h>

#include <devsetup.h>
#include <sff.h>

#include <pfint.h>

#include <pfwin.p>

static void set_clip();
static void pd_save();
static void pd_restore();

static struct wind_struct save;

/*
 *	A place to save the page descriptor
 */

/*
 *   pd_init() initializes the page descriptor, including
 *   clipping rectangles and writing modes, etc.
 */
pd_init(x, y, w, h)
int x, y, w, h;
{
	Debug4f ("pd_init(%d, %d, %d, %d)\n", x, y, w, h);

	PF.PfPage.trans_mode = TMS;
	PF.PfPage.write_tx_des = 0;
	PF.PfPage.spWid = -1;	/* means use current font space width */
	PF.PfPage.txAlign = TM_TOPLEFT;
	PF.PfPage.lin_des.line_style = TM_SOLID;
	PF.PfPage.lin_des.line_thickness = TM_VTHIN;
	PF.PfPage.scaledText = FALSE;

	/* set up clipping rectangles */

	set_clip (x, y, w, h);

	pd_save();
	/* save it now so that restore will work even if we forget to save. */

} /* end of pd_init */

/*
 *   Set clipping rectangles - device and application
 */
static void
set_clip(x, y, w, h)
int x, y, w, h;
{
	Debug4f("set_clip(%d, %d, %d, %d)\n", x, y, w, h);

	PF.PfPage.inner.origin.x = x;
	PF.PfPage.inner.origin.y = y;
	PF.PfPage.inner.corner.x = x + w;
	PF.PfPage.inner.corner.y = y + h;
	PF.PfPage.appClip = PF.PfPage.inner;
}

/*
 *	The next two routines "pd_save" and "pd_restore" save and restore
 *	the state of the page descriptor.
 */
static void
pd_save()
{
	Debugs("pd_save()\n");

	save.lin_des = PF.PfPage.lin_des;
	save.write_tx_des = PF.PfPage.write_tx_des;
	save.trans_mode = PF.PfPage.trans_mode;
	save.txAlign = PF.PfPage.txAlign;
	save.spWid = PF.PfPage.spWid;
	save.inner = PF.PfPage.inner;
	save.appClip = PF.PfPage.appClip;
	save.scaledText = PF.PfPage.scaledText;
	save.imageBox[0].x = PF.PfPage.imageBox[0].x;
	save.imageBox[0].y = PF.PfPage.imageBox[0].y;
	save.imageBox[1].x = PF.PfPage.imageBox[1].x;
	save.imageBox[1].y = PF.PfPage.imageBox[1].y;
	save.imageBox[2].x = PF.PfPage.imageBox[2].x;
	save.imageBox[2].y = PF.PfPage.imageBox[2].y;
	save.imageBox[3].x = PF.PfPage.imageBox[3].x;
	save.imageBox[3].y = PF.PfPage.imageBox[3].y;

	Debug4f(NXLT("inner: %d, %d, %d, %d\n"),
			save.inner.origin.x, save.inner.origin.y,
			save.inner.corner.x, save.inner.corner.y);
	Debug4f(NXLT("appClip: %d, %d, %d, %d\n"),
			save.appClip.origin.x, save.appClip.origin.y,
			save.appClip.corner.x, save.appClip.corner.y);
}

static void
pd_restore()
{
	Debugs("pd_restore()\n");

	PF.PfPage.lin_des = save.lin_des;
	PF.PfPage.write_tx_des = save.write_tx_des;
	PF.PfPage.trans_mode = save.trans_mode;
	PF.PfPage.txAlign = save.txAlign;
	PF.PfPage.spWid = save.spWid;
	PF.PfPage.inner = save.inner;
	PF.PfPage.appClip = save.appClip;
	PF.PfPage.imageBox[0].x = save.imageBox[0].x;
	PF.PfPage.imageBox[0].y = save.imageBox[0].y;
	PF.PfPage.imageBox[1].x = save.imageBox[1].x;
	PF.PfPage.imageBox[1].y = save.imageBox[1].y;
	PF.PfPage.imageBox[2].x = save.imageBox[2].x;
	PF.PfPage.imageBox[2].y = save.imageBox[2].y;
	PF.PfPage.imageBox[3].x = save.imageBox[3].x;
	PF.PfPage.imageBox[3].y = save.imageBox[3].y;

	Debug4f(NXLT("inner: %d, %d, %d, %d\n"),
			save.inner.origin.x, save.inner.origin.y,
			save.inner.corner.x, save.inner.corner.y);
	Debug4f(NXLT("appClip: %d, %d, %d, %d\n"),
			save.appClip.origin.x, save.appClip.origin.y,
			save.appClip.corner.x, save.appClip.corner.y);
}


/*****************************************************************/
/* TmCreate function */
pf_create(use_crops, use_regis, use_bounds, bounds, clip)
int use_crops, use_regis, use_bounds;
TmRect32 *bounds, *clip;
{
	extern long InMark();

	Debugs ("pf_create()\n");

	PF.PfPagePixWidth = xmiltopix(PF.PfPageMilWidth);
	PF.PfPagePixHeight = ymiltopix(PF.PfPageMilHeight);
	PF.pagetop = PF.PfPagePixHeight;
	set_clip(PF.PfXPixOffset, PF.PfYPixOffset, PF.PfPagePixWidth , PF.PfPagePixHeight);
	if ( PF.FirstCreate == FALSE )
		NewPage(FALSE, use_crops, use_regis, use_bounds, bounds, clip);
	else 
	{
		/* the first page */
		PF.PfFileMark = InMark();
		NewPage(!PF.PfBanner, use_crops, use_regis, use_bounds, bounds, clip);
		PF.FirstCreate = FALSE;
	}

}  /* pf_create() */

/*****************************************************************/
/*
 *  PfPage_init -- initialize page descriptor
 *
 *  Fill in default values, and set up clipping rectangles.
 *
 *  Assumes correct device characteristics have been loaded.
 */
PfPage_init()
{

	pd_init (PF.PfXPixOffset, PF.PfYPixOffset, PF.PfPagePixWidth, PF.PfPagePixHeight);

	PF.PfHardX = 0;
	PF.PfHardY = 0;

}  /* PfPage_init() */

/*****************************************************************/
/* TmTerm function */

pf_term()
{
	Debugs ("pf_term()\n");
	pf_create(FALSE, FALSE, FALSE, (TmRect32 *)0, (TmRect32 *)0);
}
