/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: ssedit.h
   Applix ASC version: 20
   Date: Mon Nov 16 13:57:03 1998
*/
#ifndef SSEDIT_INCL	/* { */
#define SSEDIT_INCL
#define ADDRSTR_LEN	12
#define EDIT_LINE_PAD 0
#define EDIT_TEXT_PAD 5

typedef struct edit_text_attr 
	{
	unsigned long foreground, background, plane_mask;
	int function;
	Font font;
	} EDIT_TEXT_ATTR;

#define SS_BUTTONS                 2
#define SS_EDIT_BUTTON_OFF         0
#define SS_EDIT_BUTTON1_ON         1
#define SS_EDIT_BUTTON2_ON         2

#define SS_EDIT_AREA_BUTTON1       1
#define SS_EDIT_AREA_BUTTON2       2

#define SS_BUTTON_OFF              0
#define SS_BUTTON_ON               1

#define MAXADDRSTR "ZZZ:ZZZ99999"
#define MAX_EXPOSE_BUF	(MAX_CELLSTR+1)
#define MAX_EDIT_LINES	200

#define DEF_EDIT_PTSIZE	     14
#if AXMS
#define DEF_EDIT_FONTNAME "Arial"
#define DEF_ADDR_FONTNAME "MS Console"
#define DEF_BUTT_FONTNAME "Wingdings"
#else
#define DEF_EDIT_FONTNAME "Helvetica"
#define DEF_ADDR_FONTNAME "Avant Garde"
#define DEF_BUTT_FONTNAME "Zapf Dingbats"
#endif

typedef struct ss_editline {	
	wchar wstr[MAX_CELLSTR+1]; /* string being entered in to the edit line */
	int winx, winy;	/* position of edit window relative to SS windows origin */
	int startx;              /* xpos where text starts in the edit window */
	int endx;				 /* where text ends in the editwindow */
	int	starty;              /* ypos where editline starts in the SS window */
	int addrstrlen;          /* TextWidth of MAXADDRSTR, set at InitEditLine */
	int buttx[SS_BUTTONS];   /* xpos where button1 of editline starts */
	int buttstate[SS_BUTTONS];
	int buttwidth[SS_BUTTONS];
	int buttmode;            /* Control Button Mode for editline */
	unsigned int pix_height; /* height of editline in pixels at any instant*/
	unsigned int pix_width;  /* width of editline in pixels */
	int pointsize;           /* Point Size of Edit Line */
	int	ElfEditSize;		/** do we need this !!0 ***/
	GC	EditGC;				/** editline GC **/
	XFontStruct *edit_font;  /* font for text in the edit line */
	XFontStruct *addr_font;  /* font for cell addr,style in the edit line */
	XFontStruct *butt_font;  /* font for Edit Mode Symbols in the editline*/
	unsigned int line_height;/* height of text char in the edit line */
	unsigned int max_char_width;	/* bounding box width */
	unsigned int addr_char_height;/* height of cell addr char in the editline*/
	int addr_char_width;        /* width of cell addr char in the editline */
	int butt_char_width;        /* width of Edit Mode Symbols in the editline*/
	int butt_char_height;       /* height of Edit Mode Symbols in the editline*/
	int num_lines;	            /* number of lines in the edit line */
	short num_chars[MAX_EDIT_LINES+1];/* number of chars in each line */
	boolean editmode;	            /* TRUE if in edit mode */
	short cursor_pos;           /* If in edit mode, the cursor pos */
	unsigned long foreground, 	/* foreground color in the editline */
		background;				/* the foreground and background */
	boolean cursor_up;          /* TRUE if cursor is displayed */
	short cursor_state;			/* active or inactive */
	int x,y;                    /* coordinates of the cursor   */
	Window xwin;    			/* The editline X-window handle */
	short openedit;             /* length of cell addr str */
	short len;                  /* length of the text string */
	int linex; /* where the separator line between cell addr and text falls*/
	short maxchars;             /* maxchars editline can hold */

	/* transient elements used during a stroke */
	int firstpoint, curpoint;
	unsigned long buttontime;
	boolean buttondown;
	boolean stroking;
	boolean clickIsDouble;

	/* marking parameters */
	boolean marked;	/* TRUE if an area in the editline is marked */
	int mark_start, mark_end;	/* marked region boundaries */

	Cursor pointer;	/* pointer shape in the editline */

	boolean focus;	/* TRUE if editwindow has focus */
	boolean keyboard_active;	/* TRUE if keyboard is active */
	boolean awaiting_expose;/* TRUE if waiting for an expose event */
	int expose_buf_count;		/* num chars in the the expose buf */
	wchar wexpose_buf[MAX_EXPOSE_BUF];	/* use for keys typed when waiting */
									/* for an expose to happen */
#if BUILD_ORIENT
    int start_conv;
    int end_conv;
    int start_phrase;
    int end_phrase;
#endif
	unsigned int	isInCellEdit:1,	/* TRUE if inline cell editing */
					unMapNow:1,		/* TRUE to unmap the editline */
					isUnMapped:1,	/* TRUE if window is currently unmapped */	
					uline:3,	/** for the underline stuff **/
					mapped:1,	/** is this editline mapped **/
					leftJustify:1, /** more for the incell editing. Is this **/ 
					rightJustify:1, /** cell left, right, center justified, */
					centerJustify:1, /** wrap text or default **/
					wrapText:1,		
					scrollTextOn:1, /* of space is limited and so we have
										to used some kind of scrolling 
										then this flag is set to true 
										else false **/
					startDispLines:8, /** start of the display line **/
					endDispLines:8; /** end of the display line **/
}	SS_EDITLINE;


/* For OpenLook Only */
#define CURSOR_ACTIVE	1	 /* cursor when editwindow has keyboard focus */
#define CURSOR_INACTIVE	0	 /* cursor when editwindow doesn't have the */
							 /* keyboard focus */
#endif	/* } SSEDIT_INCL */
