/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: allcpool.c
   Applix ASC version: 17
   Date: Tue Jan 12 21:12:19 1999
*/
#include <stdio.h>
#include <aos.h>
#include <thimble.h>
#include <amemtrak.h>

typedef struct {
	int blockSize;		/* an exact power of two */
	int itemSize;		/* all items must be of same size */
	int task;			/* same task owns all the data! */
	char *freeItem;		/* head of free list */
	char *current;		/* addr of current (unfilled) pool */
	int poolDex;		/* how much of current pool has been filled (in chars) */
	char *poolList;		/* chain of pools */
} pool_info_t;

/* Copied from tmzmallo.c - THIS MUST MATCH THE DEF'N IN THAT FILE EXACTLY! */
/* ALWAYS keep this structure size a multiple of 8 */
typedef struct mem_header
{
    short pid;      /* task that owns this data */
    short free_dex;     /* free list index number + dex */
    struct mem_header *next;/* next of this app type or NULL */
    struct mem_header *prev;/* prev of this app type or NULL */
#if ALPHA || (LINUX && __alpha__) || SGI
    int dummy;
#if MEM_CALL_TRACE
    int *ret_addr;
#endif
#else
    int *ret_addr;      /* return address to caller */
#endif
}   mem_header;

#if 0
#define BLK 16		/* memory header size */
#else
#define BLK (sizeof(mem_header))
#endif

static short PoolDex[] = {
	1,                        /* don't waste a 1K pool for 0-byte requests */
	1, 1, 1, 1, 1, 1, 1, 1,   /*  1- 8 bytes use  8-byte pool */
	2, 2, 2, 2, 2, 2, 2, 2,   /*  9-16 bytes use 16-byte pool */
	3, 3, 3, 3, 3, 3, 3, 3,   /* 17-24 bytes use 24 byte pool */
	4, 4, 4, 4, 4, 4, 4, 4,   /* 25-32 bytes use 32 byte pool */
	5, 5, 5, 5, 5, 5, 5, 5,   /* 33-40 bytes use 40 byte pool */
	6, 6, 6, 6, 6, 6, 6, 6,   /* 41-48 bytes use 48 byte pool */
	7, 7, 7, 7, 7, 7, 7, 7,   /* 49-56 bytes use 56 byte pool */
	8, 8, 8, 8, 8, 8, 8, 8    /* 57-64 bytes use 64 byte pool */
};
static int UnPoolDex[] = {	0, 8, 16, 24, 32, 40, 48, 56, 64 };

#define TASK_POOL_LIMIT 64		/* bytes for auto-pooling */
#define MAX_POOL_DEX 8          /* 9 pools, (0-8)*8 */

extern char *TaskAlloc(), *TaskCalloc(), *TaskRealloc();
void AxTaskPoolCheckPointer();
void AxPoolFindPointer();

/* Turn me ON for automatic debugging when memory is freed (SLOOOOW!) */
#define POOL_DEBUG	0

/*************************************************************************
function: AxPoolCreate - Create a pool resource
*/
char *AxPoolCreate (task,blockSize,itemSize)
int task;		/* task that owns the data */
int blockSize;	/* should be EXACT power of two e.g. 2048 */
int itemSize;	/* size of each item in pool (e.g. cell, gob...) */
{
	pool_info_t *p;

	p = (pool_info_t *)TaskCalloc(task,1,sizeof(pool_info_t));
	p->blockSize = blockSize - BLK;
	p->itemSize = itemSize;
	p->task = task;
	return((char*)p);
}

/*************************************************************************
function: AxPoolDestroy - Destroy pool struct and all the data
*/
void AxPoolDestroy (infoP)
char *infoP;
{
	pool_info_t *p;
	char *lp, *next;

	p = (pool_info_t*)infoP;
	for ( lp = p->poolList; lp; lp = next ) {
		next = *((char**)lp);
		TaskFree(p->task,lp);
	}
	TaskFree(p->task,p);
}

/*************************************************************************
function: AxPoolAllocate - Allocate 1 item from passed pool handle, rtn ptr
*/
char *AxPoolAllocate (pPtr)
char *pPtr;
{
	register char *px;
	register pool_info_t *infoP;

	infoP = (pool_info_t *)pPtr;

	RETADDR(infoP);

	/* If something on free list, use it! */
	if ( px = infoP->freeItem ) {
		infoP->freeItem = *((char**)px);
		return(px);
	}

	/* If there is no current pool, allocate one, thread to infoP->poolList */
	if ( !infoP->current ) {
		infoP->current = (char*)TaskAlloc(infoP->task,infoP->blockSize);
		*((char**)infoP->current) = infoP->poolList;	/* 1st long is thread */
		infoP->poolList = infoP->current;
		infoP->poolDex = 8;	/* sizeof(char**); not enough for HP dbl-align */
	}

	/* Allocate from current pool. If it becomes full, close it out */
	px = infoP->current + infoP->poolDex;
	infoP->poolDex += infoP->itemSize;
	if ( infoP->poolDex + infoP->itemSize > infoP->blockSize )
		infoP->current = NULL;
	return(px);
}

/*************************************************************************
function: AxPoolFree - Allocate item from pool
*/
char *AxPoolFree (pPtr,p)
register char *pPtr;
char *p;
{
register pool_info_t *infoP;
	
	infoP = (pool_info_t *) pPtr;

	*((char**)p) = infoP->freeItem;
	infoP->freeItem = p;
}

/***************** TASK BASED AUTO-POOLING *****************************/


/*************************************************************************
function: AxTaskPoolAllocate - Allocate a fixed size thing from tasks auto-pools
*/
char *AxTaskPoolAllocate (task,size)
int task;		/* THIMpid style number */
int size;		/* size in bytes */
{
	register context_block *cP;
	register char *p;
	register short dex;

	RETADDR(task);
#if 0
	size = ((size + 7) >> 3) << 3;		/* 8byte-align */
	if ( size > TASK_POOL_LIMIT )
		return(TaskAlloc(task,size));
#else
	if ( size > TASK_POOL_LIMIT )
		return(TaskAlloc(task,size));
	dex = PoolDex[size];
	size = UnPoolDex[dex];
#endif
	
	cP = &TaskInfo[task];
	if ( !cP->allocPools )
#if 0
		cP->allocPools = (char**)TaskCalloc(task,TASK_POOL_LIMIT+1,sizeof(char*));
#else
	    cP->allocPools = (char**)TaskCalloc(task,MAX_POOL_DEX+1,sizeof(char*));
#endif
	if ( NULL == (p = cP->allocPools[dex]) )
		p = cP->allocPools[dex] = AxPoolCreate(task,1024,size);
	return(AxPoolAllocate(p));
}

/*************************************************************************
function: AxTaskPoolCallocate - Allocate a fixed size thing from the tasks pools
                                and zero the memory before returning
*/
char *AxTaskPoolCallocate (task,n,size)
int task;		/* THIMpid style number */
int size;		/* size in bytes */
{
	register char *p;
	register int s;
	
	s = n * size;
	if ( s > TASK_POOL_LIMIT )
		return(TaskCalloc(task, n, size ));
			   
	p = AxTaskPoolAllocate( task, s );
	memset( p, 0, s );
	return( p );
}

/*************************************************************************
function: AxTaskPoolFree - Free data gotten by AxTaskPoolAllocate
*/
void AxTaskPoolFree (task,p,size)
int task;
char *p;
int size;
{
	register short dex;
	
	if ( !p )
		return;
#if 0
	size = ((size + 7) >> 3) << 3;		/* 8byte-align */
	if ( size > TASK_POOL_LIMIT ) {
		TaskFree(task,p);
		return;
	}
#else
	if ( size > TASK_POOL_LIMIT ) {
		TaskFree(task,p);
		return;
	}
	dex = PoolDex[size];
	size = UnPoolDex[dex];
#endif

#if POOL_DEBUG
	AxTaskPoolCheckPointer(task,p,size);
#endif
	AxPoolFree(TaskInfo[task].allocPools[dex],p);
}

/*************************************************************************
function: AxTaskPoolRealloc - Slow resize logic
*/
char *AxTaskPoolRealloc (task,p,newsize,oldsize)
int task;
char *p;
int newsize, oldsize;
{
	register char *newP;

	RETADDR(task);
#if 0
	newsize = ((newsize + 7) >> 3) << 3;		/* 8byte-align */
#else
	if ( newsize <= TASK_POOL_LIMIT )
		newsize = UnPoolDex[PoolDex[newsize]];
#endif
	if ( !p ) {
		newP = AxPoolAllocate(task,newsize);
		return(newP);
	}
#if 0
	oldsize = ((oldsize + 7) >> 3) << 3;		/* 8byte-align */
#else
	if (oldsize <= TASK_POOL_LIMIT )
		oldsize = UnPoolDex[PoolDex[oldsize]];
#endif
	if ( newsize == oldsize )
		return(p);
	if ( (newsize > TASK_POOL_LIMIT) && (oldsize > TASK_POOL_LIMIT) )
		return(TaskRealloc(task,p,newsize));
	newP = AxTaskPoolAllocate(task,newsize);
	memcpy(newP,p,MIN(newsize,oldsize));
	AxTaskPoolFree(task,p,oldsize);
	return(newP);
}


/*************************************************************************
function: AxTaskPoolCheckPointer - Sanity check on (alleged) pool memory

Given a pointer to what is supposed to be pooled memory, this function
verifies that the pointer:
	A) Falls within the blocks allocated for the task's pool of the 
		passed size
	B) Does NOT fall within the free list for this pool!

Note if test A fails, a number of possibilities exist:
	- The wrong size may have been passed
	- A different task may have actually allocated this memory
	- This memory was allocated by TaskAlloc, not AxTaskPoolAllocate

The function taskAborts if the test fails, else simply returns.
*/
void AxTaskPoolCheckPointer (task,ptr,size)
int task;
char *ptr;
int size;
{
	register short dex;
	register context_block *cP;
	register pool_info_t *infoP;
	register char *p;
	register char *pool;
	char	buf[256];

	
	if ( !p )
		return;

	/* Large allocations are passed to TaskAlloc; we don't test these
		in any way */
	if ( size > TASK_POOL_LIMIT )
		return;
	dex = PoolDex[size];
	size = UnPoolDex[dex];

	cP = &TaskInfo[task];
	if ( !cP->allocPools ) {
		sprintf( buf, "pooled memory error, task %d has no pooled memory",
						task);
		taskAbort( buf);
	}

	infoP = (pool_info_t*)(cP->allocPools[dex]);
	if (!infoP) {
		sprintf( buf, 
			"pooled memory error, task %d has no pooled memory for size %d",
						task, size);
		taskAbort( buf);
	}

	/* Is this pointer on the free list? (Error if so) */
	for (p=infoP->freeItem; p; p=*((char**)p) ) {
		if (ptr == p) {
			sprintf( buf, "pooled memory error, task %d has already freed ptr",
						task);
			taskAbort( buf);
		}
	}

	/* Does pointer fall into one of the blocks in this pool? (Error if not) */
	for (pool=infoP->poolList; pool; pool=*((char**)pool) ) {
		if ((ptr >= pool) && (ptr <= (pool+infoP->blockSize)))
			return;
	}

	sprintf( buf, 
			"pooled memory error, ptr out of range for task %d pooled memory",
						task);

	/* Try to find who really does own this pointer. Print to stdout. */
	AxPoolFindPointer(ptr);

	/* Abort with the error message */
	taskAbort( buf);
}


/*************************************************************************
function: AxPoolFindPointer - Get info on (alleged) pool memory

Given a pointer to what is supposed to be pooled memory, this function
	determines the task that owns the memory, the size pool it falls
	within, and whether it is on the free list. The info is printed 
	to stdout.
*/
void AxPoolFindPointer (ptr)
void *ptr;
{
	int ix;
	context_block *cP;
	short dex;
	int size;
	pool_info_t *infoP;
	char *pool;
	char *p;
	int free=FALSE;
	int in_use=TRUE;

	for ( ix = 0; ix < MAX_PIDS; ix++ ) {
		cP = &TaskInfo[ix];
		if ( cP->in_use && cP->allocPools ) {
			for ( dex = 1; dex <= MAX_POOL_DEX; dex++ ) {
				if (infoP = (pool_info_t*)(cP->allocPools[dex])) {

					/* Does pointer fall into one of the blocks in this pool? */
					for (pool=infoP->poolList; pool; pool=*((char**)pool) ) {
						if (((char *)ptr >= pool) && 
								((char *)ptr <= (pool+infoP->blockSize))) {

							/* Is pointer in the range of already-used memory? 
								(If location is near the end of the current
								memory block, it may not have been used yet) */
							if ((char *)ptr >= (pool+infoP->poolDex)) 
								in_use = FALSE;

							/* Is pointer on free list? */
							else for (p=infoP->freeItem; p; p=*((char**)p) ) {
								if (ptr == p) {
									free = TRUE;
									break;
								}
							}

							size = UnPoolDex[dex];
							fprintf( stdout, 
										"%lx is in task %d size %d pool %s\n",
										ptr, ix, size,
										(free ? "(ON FREE LIST)" :
											(in_use ? "(in use)" :
											"(not yet used)")));
							return;
						}
					}
				}
			}
		}
	}

	fprintf( stdout, "%lx is NOT in any task's memory pool\n", ptr);
}
