/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: asciipf.c
   Applix ASC version: 6
   Date: Mon Nov 16 13:53:24 1998
*/
#include <stdio.h>
#include <aos.h>
#include <pageform.h>
#include <fsrts.h>
#include <errno.h>
#include <getstr4.h>

static int PFasciiFD;

/*************************************************************************
function: PFtoAscii - Convert PF data to ascii form
*/
typeErr PFtoAscii (docP)
DocInfoType *docP;
{
	char pathname[STRPATH+1];
	typeErr err;
	extern int errno;
	struct pageInfo data;
	char strs[MAXTEXT];
	int fd;		/* file desc */
	int cnt;
	int first_page_special;
	int even_pages_special;
	struct pageVersion hdr;	/* file header */
	int line;

	sprintf(pathname,"%s/1.pf",docP->pathname);
	fd = OpenFileObjFmode(docP,pathname,"r");
	if ( (fd <= 0) && (errno == ENOENT) )
		return(AOK);		/* nothin to do */
	if ( fd <= 0 )
		return(ErrnoErr());

/*
	Read file groups into memory.
	The following relies on the assumption that the odd group always comes first
	int the file and may be followed by a first and/or an even group in
	that order.
*/

	cnt = DocHardRead(fd,&hdr,sizeof(hdr));
	if ( (cnt <= 0) || (hdr.version != PFREV) || (hdr.flag != -1) )
		return(ERR_ILLVAL);	/* ?? better code */

	sprintf(pathname,"%s/1.pfa",docP->pathname);
	PFasciiFD = OpenFileObjFmode(docP,pathname,"w");

	
	Putter(NXLT("printFormatRev %d"),1);
	while ( DocHardRead(fd,&data,sizeof(data)) > 0 ) {
		cnt = DocHardRead(fd,strs,sizeof(strs));
		if ( cnt <= 0 )
			goto wrap;
		PutAscii(&data,strs);
	}
wrap:
	CloseFileObject(fd);
	CloseFileObject(PFasciiFD);
	return(AOK);
}

/*************************************************************************
function: PutAscii - Send header/footer data to ascii
*/
static PutAscii (dataP,strbuf)
struct pageInfo *dataP;
char *strbuf;
{
	int line;

	Putter(NXLT("pageType %d"),dataP->pageType);
	Putter(NXLT("lMargin %d"),dataP->lMargin);
	Putter(NXLT("rMargin %d"),dataP->rMargin);
	Putter(NXLT("tMargin %d"),dataP->tMargin);
	Putter(NXLT("bMargin %d"),dataP->bMargin);
	Putter(NXLT("hDistance %d"),dataP->hDistance);
	Putter(NXLT("fDistance %d"),dataP->fDistance);
	Putter(NXLT("revBars %d"),dataP->revBars);
	Putter(NXLT("dAtt %d"),dataP->dAtt);
	Putter(NXLT("mControl %d"),dataP->mControl);
	Putter(NXLT("HlineCount %d"),dataP->pHeader.lineCount);
	Putter(NXLT("FlineCount %d"),dataP->pFooter.lineCount);

	for ( line = 0; line < dataP->pHeader.lineCount; line++ ) {
		if ( dataP->pHeader.hfLines[line].leftText.pText != -1 ) {
			Putter(NXLT("HleftText %d %X"),line,
					dataP->pHeader.hfLines[line].leftText.pAtt);
			Putter(NXLT("%s"),strbuf+dataP->pHeader.hfLines[line].leftText.pText);
		}
		if ( dataP->pHeader.hfLines[line].centText.pText != -1 ) {
			Putter(NXLT("HcentText %d %X"),line,
						dataP->pHeader.hfLines[line].centText.pAtt);
			Putter(NXLT("%s"),strbuf+dataP->pHeader.hfLines[line].centText.pText);
		}
		if ( dataP->pHeader.hfLines[line].rightText.pText != -1 ) {
			Putter(NXLT("HrightText %d %X"),line,
						dataP->pHeader.hfLines[line].rightText.pAtt);
			Putter(NXLT("%s"),strbuf+dataP->pHeader.hfLines[line].rightText.pText);
		}
	}
	for ( line = 0; line < dataP->pFooter.lineCount; line++ ) {
		if ( dataP->pFooter.hfLines[line].leftText.pText != -1 ) {
			Putter(NXLT("FleftText %d %X"),line,
						dataP->pFooter.hfLines[line].leftText.pAtt);
			Putter(NXLT("%s"),strbuf+dataP->pFooter.hfLines[line].leftText.pText);
		}
		if ( dataP->pFooter.hfLines[line].centText.pText != -1 ) {
			Putter(NXLT("FcentText %d %X"),line,
						dataP->pFooter.hfLines[line].centText.pAtt);
			Putter(NXLT("%s"),strbuf+dataP->pFooter.hfLines[line].centText.pText);
		}
		if ( dataP->pFooter.hfLines[line].rightText.pText != -1 ) {
			Putter(NXLT("FrightText %d %X"),line,
						dataP->pFooter.hfLines[line].rightText.pAtt);
			Putter(NXLT("%s"),strbuf+dataP->pFooter.hfLines[line].rightText.pText);
		}
	}
}

/*************************************************************************
function: Putter - Actually write line to file
*/
static Putter (str,a,b,c,d,e,f)
char *str;
int a,b,c,d,e,f;
{
	char buf[2000];

	sprintf(buf,str,a,b,c,d,e,f);
	strcat(buf,"\n");
	FSfputs(buf,PFasciiFD);
}

/*************************************************************************
function: AsciiToPF - Convert ascii pf to binary pf data
*/
AsciiToPF (docP)
DocInfoType *docP;
{
	struct pageInfo data;
	char strbuf[MAXTEXT];
	int fd;		/* file desc */
	char asciiname[STRPATH+1];
	char binaryname[STRPATH+1];
	typeErr err;
	int line, att;
	char linebuf[2001];
	struct pageVersion hdr;	/* file header */
	int tempint;
	int bufpos;
	int gotStandard = FALSE;
	int gotEven = FALSE;
	int gotOdd = FALSE;
	int gotFirst = FALSE;

/* open ascii */

	sprintf(asciiname,"%s/1.pfa",docP->pathname);
	PFasciiFD = OpenFileObjFmode(docP,asciiname,"r");
	if ( PFasciiFD <= 0 )
		return(ErrnoErr());

/* create binary */

	sprintf(binaryname,"%s/1.pf",docP->pathname);
	fd = OpenFileObjFmode(docP,binaryname,"w+");
	memset(&hdr,0,sizeof(hdr));
	hdr.version = PFREV;
	hdr.flag = -1;
	DocHardWrite(fd,&hdr,sizeof(hdr));
	memset(strbuf,0,sizeof(strbuf));

	if ( !FSfgets(linebuf,2000,PFasciiFD) )
		goto corrupted;
	if ( 1 != sscanf(linebuf,NXLT("printFormatRev %d\n"),&tempint) )
		goto corrupted;
	if ( !FSfgets(linebuf,2000,PFasciiFD) )
		goto corrupted;

	while ( TRUE )	{	/* actually till  EOF */

/* init the group for textless */

		for ( line = 0; line < XTRA_HF_LINES; line++ ) {
			data.pHeader.hfLines[line].leftText.pText = -1;
			data.pHeader.hfLines[line].leftText.pAtt = 0;
			data.pHeader.hfLines[line].centText.pText = -1;
			data.pHeader.hfLines[line].centText.pAtt = 0;
			data.pHeader.hfLines[line].rightText.pText = -1;
			data.pHeader.hfLines[line].rightText.pAtt = 0;
			data.pFooter.hfLines[line].leftText.pText = -1;
			data.pFooter.hfLines[line].leftText.pAtt = 0;
			data.pFooter.hfLines[line].centText.pText = -1;
			data.pFooter.hfLines[line].centText.pAtt = 0;
			data.pFooter.hfLines[line].rightText.pText = -1;
			data.pFooter.hfLines[line].rightText.pAtt = 0;
		}

/*	read the required data */

		if ( 1 != sscanf(linebuf,NXLT("pageType %d\n"),&tempint) ) goto corrupted;
		data.pageType = tempint;
		if ( data.pageType == stdInitial )
			gotStandard = TRUE;
		if ( data.pageType == oddInitial )
			gotOdd = TRUE;
		if ( data.pageType == evInitial )
			gotEven = TRUE;
		if ( data.pageType == frstInitial )
			gotFirst = TRUE;

		if ( !FSfgets(linebuf,2000,PFasciiFD) ) goto corrupted;
		if ( 1 != sscanf(linebuf,NXLT("lMargin %d\n"),&tempint) ) goto corrupted;
		data.lMargin = tempint;

		if ( !FSfgets(linebuf,2000,PFasciiFD) ) goto corrupted;
		if ( 1 != sscanf(linebuf,NXLT("rMargin %d\n"),&tempint) ) goto corrupted;
		data.rMargin = tempint;

		if ( !FSfgets(linebuf,2000,PFasciiFD) ) goto corrupted;
		if ( 1 != sscanf(linebuf,NXLT("tMargin %d\n"),&tempint) ) goto corrupted;
		data.tMargin = tempint;

		if ( !FSfgets(linebuf,2000,PFasciiFD) ) goto corrupted;
		if ( 1 != sscanf(linebuf,NXLT("bMargin %d\n"),&tempint) ) goto corrupted;
		data.bMargin = tempint;

		if ( !FSfgets(linebuf,2000,PFasciiFD) ) goto corrupted;
		if ( 1 != sscanf(linebuf,NXLT("hDistance %d\n"),&tempint) ) goto corrupted;
		data.hDistance = tempint;

		if ( !FSfgets(linebuf,2000,PFasciiFD) ) goto corrupted;
		if ( 1 != sscanf(linebuf,NXLT("fDistance %d\n"),&tempint) ) goto corrupted;
		data.fDistance = tempint;

		if ( !FSfgets(linebuf,2000,PFasciiFD) ) goto corrupted;
		if ( 1 != sscanf(linebuf,NXLT("revBars %d\n"),&tempint) ) goto corrupted;
		data.revBars = tempint;

		if ( !FSfgets(linebuf,2000,PFasciiFD) ) goto corrupted;
		if ( 1 != sscanf(linebuf,NXLT("dAtt %d\n"),&tempint) ) goto corrupted;
		data.dAtt = tempint;

		if ( !FSfgets(linebuf,2000,PFasciiFD) ) goto corrupted;
		if ( 1 != sscanf(linebuf,NXLT("mControl %d\n"),&tempint) ) goto corrupted;
		data.mControl = tempint;

		if ( !FSfgets(linebuf,2000,PFasciiFD) ) goto corrupted;
		if ( 1 != sscanf(linebuf,NXLT("HlineCount %d\n"),&tempint) ) goto corrupted;
		data.pHeader.lineCount = tempint;

		if ( !FSfgets(linebuf,2000,PFasciiFD) ) goto corrupted;
		if ( 1 != sscanf(linebuf,NXLT("FlineCount %d\n"),&tempint) ) goto corrupted;
		data.pFooter.lineCount = tempint;

		if ( !FSfgets(linebuf,2000,PFasciiFD) )	{ /* no more lines, all done */
			DocHardWrite(fd,&data,sizeof(data));
			DocHardWrite(fd,strbuf,MAXTEXT);
			goto all_done;
		}

		bufpos = 0;
		while ( TRUE ) {	/* breaks when no line spec */
			if ( 2 == sscanf(linebuf,NXLT("HleftText %d %X\n"),&line,&att) ) {
				data.pHeader.hfLines[line].leftText.pText = bufpos;
				data.pHeader.hfLines[line].leftText.pAtt = att;
			}
			else if ( 2 == sscanf(linebuf,NXLT("HcentText %d %X\n"),&line,&att) ) {
				data.pHeader.hfLines[line].centText.pText = bufpos;
				data.pHeader.hfLines[line].centText.pAtt = att;
			}
			else if ( 2 == sscanf(linebuf,NXLT("HrightText %d %X\n"),&line,&att) ) {
				data.pHeader.hfLines[line].rightText.pAtt = att;
				data.pHeader.hfLines[line].rightText.pText = bufpos;
			}
			else if ( 2 == sscanf(linebuf,NXLT("FleftText %d %X\n"),&line,&att) ) {
				data.pFooter.hfLines[line].leftText.pText = bufpos;
				data.pFooter.hfLines[line].leftText.pAtt = att;
			}
			else if ( 2 == sscanf(linebuf,NXLT("FcentText %d %X\n"),&line,&att) ) {
				data.pFooter.hfLines[line].centText.pText = bufpos;
				data.pFooter.hfLines[line].centText.pAtt = att;
			}
			else if ( 2 == sscanf(linebuf,NXLT("FrightText %d %X\n"),&line,&att) ) {
				data.pFooter.hfLines[line].rightText.pAtt = att;
				data.pFooter.hfLines[line].rightText.pText = bufpos;
			}
			else
				break;

			if ( !FSfgets(linebuf,2000,PFasciiFD) ) goto corrupted;
			linebuf[strlen(linebuf)-1] = 0;	/* clip the newline */
			strcpy(strbuf+bufpos,linebuf);
			bufpos += strlen(linebuf)+1;
			/* fetch next line group */
			if ( !FSfgets(linebuf,2000,PFasciiFD) )	{ /* no more lines, all done */
				DocHardWrite(fd,&data,sizeof(data));
				DocHardWrite(fd,strbuf,MAXTEXT);
				goto all_done;
			}
		}

/* add group to file */

		DocHardWrite(fd,&data,sizeof(data));
		DocHardWrite(fd,strbuf,MAXTEXT);
	}

/* DC seems to require a standard page Ref! */

all_done:
	data.pHeader.lineCount = 0;
	data.pFooter.lineCount = 0;
	for ( line = 0; line < XTRA_HF_LINES; line++ ) {
		data.pHeader.hfLines[line].leftText.pText = -1;
		data.pHeader.hfLines[line].leftText.pAtt = 0;
		data.pHeader.hfLines[line].centText.pText = -1;
		data.pHeader.hfLines[line].centText.pAtt = 0;
		data.pHeader.hfLines[line].rightText.pText = -1;
		data.pHeader.hfLines[line].rightText.pAtt = 0;
		data.pFooter.hfLines[line].leftText.pText = -1;
		data.pFooter.hfLines[line].leftText.pAtt = 0;
		data.pFooter.hfLines[line].centText.pText = -1;
		data.pFooter.hfLines[line].centText.pAtt = 0;
		data.pFooter.hfLines[line].rightText.pText = -1;
		data.pFooter.hfLines[line].rightText.pAtt = 0;
	}

	if ( gotOdd && !gotEven ) {
		data.pageType = evInitial;
		DocHardWrite(fd,&data,sizeof(data));
		DocHardWrite(fd,strbuf,MAXTEXT);
	}
	if ( gotEven && !gotOdd ) {
		data.pageType = oddInitial;
		DocHardWrite(fd,&data,sizeof(data));
		DocHardWrite(fd,strbuf,MAXTEXT);
	}
	if ( !gotEven && !gotOdd && gotFirst && !gotStandard ) {
		data.pageType = stdInitial;
		DocHardWrite(fd,&data,sizeof(data));
		DocHardWrite(fd,strbuf,MAXTEXT);
	}
	CloseFileObject(fd);
	CloseFileObject(PFasciiFD);
	return;

corrupted:
	CloseFileObject(fd);
	CloseFileObject(PFasciiFD);
	DeleteFileObject(docP,asciiname);
	DeleteFileObject(docP,binaryname);
}
