/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: ax_des.c
   Applix ASC version: 18
   Date: Wed Apr 14 21:16:31 1999
*/
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#if AXMS
#include <process.h>
#else
#include <unistd.h>
#endif
#include <time.h>
#include <aos.h>
#include <ax_des.h>
#include <eldata.h>
#include <afile.h>
#include <getstr10.h>

#define TRUE 1
#define FALSE 0

#define MAXPATH 512

#define MODE_ENCRYPT 1
#define MODE_DECRYPT 2

extern "C" int AxUseEncryption;
extern "C" key_tables *AxKtab, *AxSKtab;
extern "C" key_ring *AxKeys, *AxSKeys;
extern "C" data_ptr ElfRetData;

extern "C" {
	int AxCreateKeys(key_tables *, key_ring **, char *);
	void AxDestroyKeys(key_ring **);
	int AxAddKeys(char *);
	int AxAddSKeys(char *);
	int AxGetRKeys2(key_tables **, key_ring **, int);
	int AxGetRKeys(key_tables **, key_ring **);
	int AxGetWKeys(key_tables **, key_ring **);
	int AxGetRSKeys(key_tables **, key_ring **);
	int AxGetWSKeys(key_tables **, key_ring **);
	void AxInitCrypt (key_tables *);
	void AxInitKeys (key_tables *, key_ring *, char *);
	void AxInitializeEncryption (key_tables *, key_ring *, char *);
	void AxLoadKeytables (int);
	void AxEncrypt(key_tables *, key_ring *, unsigned char*, int);
	void AxDecrypt(key_tables *, key_ring *, unsigned char*, int);
	void AxSetEncryptionState(int);
	int AxGetEncryptionState();

	char * AxStrPtrFromDataPtr(data_ptr);
	data_ptr ElfbNewTaskWithData(char *, char *, int *, int *);
	int ElfCallByName (...);
	void AxFreeData (data_ptr);
	void AFileGetBuffer(mbchar *, int in_len, mbchar **, int *);
	int AFileValidateHeaders (mbchar *, int );
	typeErr AFileInfoFromHeader(mbchar *, AFileInfo *);
	int AxDecryptBuffer(char *, int, int);
	void AxOverrideEncryptionHeader(mbchar *, int );
}

key_tables *AxKtab, *AxSKtab;
key_ring *AxKeys, *AxSKeys;

/****************************************************************************/
static unsigned char bytemask[8] = {
	0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01
};

static unsigned int intmask[32] = {
	0x80000000, 0x40000000, 0x20000000, 0x10000000,
	0x08000000, 0x04000000, 0x02000000, 0x01000000,
	0x00800000, 0x00400000, 0x00200000, 0x00100000,
	0x00080000, 0x00040000, 0x00020000, 0x00010000,
	0x00008000, 0x00004000, 0x00002000, 0x00001000,
	0x00000800, 0x00000400, 0x00000200, 0x00000100,
	0x00000080, 0x00000040, 0x00000020, 0x00000010,
	0x00000008, 0x00000004, 0x00000002, 0x00000001
};

static int pc1[56] = {
	57, 49, 41, 33, 25, 17,  9,  1, 58, 50, 42, 34, 26, 18,
	10,  2, 59, 51, 43, 35, 27, 19, 11,  3, 60, 52, 44, 36,
	63, 55, 47, 39, 31, 23, 15,  7, 62, 54, 46, 38, 30, 22,
	14,  6, 61, 53, 45, 37, 29, 21, 13,  5, 28, 20, 12,  4
};
static int pc2[48] = {
	14, 17, 11, 24,  1,  5,  3, 28, 15,  6, 21, 10,
	23, 19, 12,  4, 26,  8, 16,  7, 27, 20, 13,  2,
	41, 52, 31, 37, 47, 55, 30, 40, 51, 45, 33, 48,
	44, 49, 39, 56, 34, 53, 46, 42, 50, 36, 29, 32
};

static int rots[16] = {
	1, 1, 2, 2, 2, 2, 2, 2, 1, 2, 2, 2, 2, 2, 2, 1
};

static int esel[48] = {
	32,  1,  2,  3,  4,  5,  4,  5,  6,  7,  8,  9,
	 8,  9, 10, 11, 12, 13, 12, 13, 14, 15, 16, 17,
	16, 17, 18, 19, 20, 21, 20, 21, 22, 23, 24, 25,
	24, 25, 26, 27, 28, 29, 28, 29, 30, 31, 32,  1
};

static int perm32[32] = {
	16,  7, 20, 21, 29, 12, 28, 17,  1, 15, 23, 26,  5, 18, 31, 10,
	 2,  8, 24, 14, 32, 27,  3,  9, 19, 13, 30,  6, 22, 11,  4, 25
};

static int initial_perm[64] = {
	58, 50, 42, 34, 26, 18, 10,  2, 60, 52, 44, 36, 28, 20, 12, 4,
	62, 54, 46, 38, 30, 22, 14,  6, 64, 56, 48, 40, 32, 24, 16, 8,
	57, 49, 41, 33, 25, 17,  9,  1, 59, 51, 43, 35, 27, 19, 11, 3,
	61, 53, 45, 37, 29, 21, 13,  5, 63, 55, 47, 39, 31, 23, 15, 7
};
static int final_perm[64] = {
	40,  8, 48, 16, 56, 24, 64, 32, 39,  7, 47, 15, 55, 23, 63, 31,
	38,  6, 46, 14, 54, 22, 62, 30, 37,  5, 45, 13, 53, 21, 61, 29,
	36,  4, 44, 12, 52, 20, 60, 28, 35,  3, 43, 11, 51, 19, 59, 27,
	34,  2, 42, 10, 50, 18, 58, 26, 33,  1, 41,  9, 49, 17, 57, 25
};

static int sbox_table[8][4][16]= {
        { { 14,  4, 13,  1,  2, 15, 11,  8,  3, 10,  6, 12,  5,  9,  0,  7 },
          {  0, 15,  7,  4, 14,  2, 13,  1, 10,  6, 12, 11,  9,  5,  3,  8 },
          {  4,  1, 14,  8, 13,  6,  2, 11, 15, 12,  9,  7,  3, 10,  5,  0 },
          { 15, 12,  8,  2,  4,  9,  1,  7,  5, 11,  3, 14, 10,  0,  6, 13 }
        },

        { { 15,  1,  8, 14,  6, 11,  3,  4,  9,  7,  2, 13, 12,  0,  5, 10 },
          {  3, 13,  4,  7, 15,  2,  8, 14, 12,  0,  1, 10,  6,  9, 11,  5 },
          {  0, 14,  7, 11, 10,  4, 13,  1,  5,  8, 12,  6,  9,  3,  2, 15 },
          { 13,  8, 10,  1,  3, 15,  4,  2, 11,  6,  7, 12,  0,  5, 14,  9 }
        },

        { { 10,  0,  9, 14,  6,  3, 15,  5,  1, 13, 12,  7, 11,  4,  2,  8 },
          { 13,  7,  0,  9,  3,  4,  6, 10,  2,  8,  5, 14, 12, 11, 15,  1 },
          { 13,  6,  4,  9,  8, 15,  3,  0, 11,  1,  2, 12,  5, 10, 14,  7 },
          {  1, 10, 13,  0,  6,  9,  8,  7,  4, 15, 14,  3, 11,  5,  2, 12 }
        },

        { {  7, 13, 14,  3,  0,  6,  9, 10,  1,  2,  8,  5, 11, 12,  4, 15 },
          { 13,  8, 11,  5,  6, 15,  0,  3,  4,  7,  2, 12,  1, 10, 14,  9 },
          { 10,  6,  9,  0, 12, 11,  7, 13, 15,  1,  3, 14,  5,  2,  8,  4 },
          {  3, 15,  0,  6, 10,  1, 13,  8,  9,  4,  5, 11, 12,  7,  2, 14 }
        },

        { {  2, 12,  4,  1,  7, 10, 11,  6,  8,  5,  3, 15, 13,  0, 14,  9 },
          { 14, 11,  2, 12,  4,  7, 13,  1,  5,  0, 15, 10,  3,  9,  8,  6 },
          {  4,  2,  1, 11, 10, 13,  7,  8, 15,  9, 12,  5,  6,  3,  0, 14 },
          { 11,  8, 12,  7,  1, 14,  2, 13,  6, 15,  0,  9, 10,  4,  5,  3 }
        },

        { { 12,  1, 10, 15,  9,  2,  6,  8,  0, 13,  3,  4, 14,  7,  5, 11 },
          { 10, 15,  4,  2,  7, 12,  9,  5,  6,  1, 13, 14,  0, 11,  3,  8 },
          {  9, 14, 15,  5,  2,  8, 12,  3,  7,  0,  4, 10,  1, 13, 11,  6 },
          {  4,  3,  2, 12,  9,  5, 15, 10, 11, 14,  1,  7,  6,  0,  8, 13 }
        },

        { {  4, 11,  2, 14, 15,  0,  8, 13,  3, 12,  9,  7,  5, 10,  6,  1 },
          { 13,  0, 11,  7,  4,  9,  1, 10, 14,  3,  5, 12,  2, 15,  8,  6 },
          {  1,  4, 11, 13, 12,  3,  7, 14, 10, 15,  6,  8,  0,  5,  9,  2 },
          {  6, 11, 13,  8,  1,  4, 10,  7,  9,  5,  0, 15, 14,  2,  3, 12 }
        },

        { { 13,  2,  8,  4,  6, 15, 11,  1, 10,  9,  3, 14,  5,  0, 12,  7 },
          {  1, 15, 13,  8, 10,  3,  7,  4, 12,  5,  6, 11,  0, 14,  9,  2 },
          {  7, 11,  4,  1,  9, 12, 14,  2,  0,  6, 10, 13, 15,  3,  5,  8 },
          {  2,  1, 14,  7,  4, 10,  8, 13, 15, 12,  9,  0,  3,  5,  6, 11 }
        }
};
char test_str[20] = "Default test string";

/****************************************************************************/
#define ASCII2BIN(c) ((c)>='a'?(c-59):(c)>='A'?((c)-53):(c)-'.')
#define BIN2ASCII(c) ((c)>=38?((c)-38+'a'):(c)>=12?((c)-12+'A'):(c)+'.')
#define BITMASK(i) ( (1<<(11-(i)%12+3)) << ((i)<12?16:0) )
#define SBA(sb, v) (*(unsigned int *)((char *)(sb)+(v)))
#define S_LOOKUP(i,s) sbox_table[(i)][(((s)>>4) & 0x2)|((s) & 0x1)][((s)>>1) & 0xf];

/****************************************************************************/

/****************************************************************************\
\****************************************************************************/

/****************************************************************************/
static void shuffle_sbox(unsigned int *sbox, unsigned int saltbits)
{
	unsigned int ix, val;
	for ( ix=4096; ix--; )
	{
		val = (sbox[0] ^ sbox[1]) & saltbits;
		*sbox++ ^= val;
		*sbox++ ^= val;
	}
}

/****************************************************************************/
void AxInitCrypt (key_tables *ktab)
{
	unsigned int mask_1, mask_2, ix, jx, val, saltbits, *sb;
	int comes_from_f_bit, comes_from_e_bit;
	int comes_from_word, bit_within_word;
	int bit_ix, bitval, box_ix;
	int j1, j2, s1, s2;
	int o_bit, o_int;
	long c;

	/* set the default salt */
	strcpy(ktab->salt, "..");

	/* Create table used for pc1 permutation when generating keys */
	for ( bit_ix=0; bit_ix<56; bit_ix++ )
	{
		bitval = pc1[bit_ix] - 1;
		mask_1 = bytemask[bitval % 8 + 1];
		mask_2 = intmask[bit_ix % 28 + 4];
		for ( ix=0; ix<128; ix++ )
		{
			if ( ix&mask_1 )
		      ktab->pc1_table[bitval/8][bit_ix/28][ix] |= mask_2;
		}
	}

	/* Create table used for pc2 permutation when generating keys */
	for ( bit_ix=0; bit_ix<48; bit_ix++ )
	{
		bitval = pc2[bit_ix] - 1;
		mask_1 = bytemask[bitval % 7 + 1];
		mask_2 = BITMASK(bit_ix % 24);
		for ( ix=0; ix<128; ix++ )
		{
			if ( ix&mask_1 )
				ktab->pc2_table[bitval/7][ix] |= mask_2;
		}
	}

	/* Create table used to do combined 32 bit permutation and e expansion */
	for ( bit_ix=0; bit_ix<48; bit_ix++ )
	{
		bitval = perm32[esel[bit_ix]-1]-1;
		mask_1 = bytemask[bitval % 8];
		for ( ix=256; ix--; )
		{
			if ( ix&mask_1 )
				ktab->eperm32tab[bitval/8][ix][bit_ix/24] |= BITMASK(bit_ix%24);
		}
	}

	/* Create the s boxes */
	for ( box_ix=0; box_ix<4; box_ix++ )
	{
		switch(box_ix)
		{
			default:
			case 0: sb = ktab->sb0;
			case 1: sb = ktab->sb1;
			case 2: sb = ktab->sb2;
			case 3: sb = ktab->sb3;
		}
		for ( j1=0; j1<64; j1++ )
		{
			s1 = S_LOOKUP(2*box_ix, j1);
			for ( j2=0; j2<64; j2++ )
			{
				s2 = S_LOOKUP(2*box_ix+1, j2);
				val = ((s1<<4)|s2) << (24-8*box_ix);

				ix = ((j1<<6)|j2)<<1;
				sb[ix  ]  = ktab->eperm32tab[0][(val>>24)&0xff][0];
				sb[ix+1]  = ktab->eperm32tab[0][(val>>24)&0xff][1];
				sb[ix  ] |= ktab->eperm32tab[1][(val>>16)&0xff][0];
				sb[ix+1] |= ktab->eperm32tab[1][(val>>16)&0xff][1];
				sb[ix  ] |= ktab->eperm32tab[2][(val>> 8)&0xff][0];
				sb[ix+1] |= ktab->eperm32tab[2][(val>> 8)&0xff][1];
				sb[ix  ] |= ktab->eperm32tab[3][(val)    &0xff][0];
				sb[ix+1] |= ktab->eperm32tab[3][(val)    &0xff][1];
			}
		}
    }

	/* Create inverse matrix for esel */
	for ( bit_ix=48; bit_ix--; )
	{
		ktab->e_inverse[esel[bit_ix]-1   ] = bit_ix;
		ktab->e_inverse[esel[bit_ix]-1+32] = bit_ix+48;
	}

	/* Create matrix used to undo the E expansion and for final permutation */
	for ( bit_ix=0; bit_ix<64; bit_ix++ )
	{
		/* See where bit belongs in the two 32 bit longs... */
		o_int = bit_ix / 32; /* 0..1  */
		o_bit = bit_ix % 32; /* 0..31 */

		/* ...and find a bit in the e permutated value setting this bit. */
		comes_from_f_bit = final_perm[bit_ix] - 1;      /* 0..63 */
		comes_from_e_bit = ktab->e_inverse[comes_from_f_bit]; /* 0..95 */
		comes_from_word  = comes_from_e_bit / 6;        /* 0..15 */
		bit_within_word  = comes_from_e_bit % 6;        /* 0..5  */

		mask_1 = intmask[bit_within_word + 26];
		mask_2 = intmask[o_bit];
		for ( ix=64; ix--; )
		{
			if ( ix&mask_1 )
				ktab->efp[comes_from_word][ix][o_int] |= mask_2;
		}
	}

	/* Permute the sboxes to reflect the changed e selection table */
	saltbits = 0;
	for ( ix=0; ix<2; ix++ )
	{
		c = ASCII2BIN(ktab->salt[ix]);
		if ( c < 0 || c > 63 )
				c = 0;
		for ( jx=0; jx<6; jx++ )
		{
			if ( (c>>jx)&0x1 )
				saltbits |= BITMASK(6*ix+jx);
		}
	}
	shuffle_sbox(ktab->sb0, ktab->current_saltbits^saltbits);
	shuffle_sbox(ktab->sb1, ktab->current_saltbits^saltbits);
	shuffle_sbox(ktab->sb2, ktab->current_saltbits^saltbits);
	shuffle_sbox(ktab->sb3, ktab->current_saltbits^saltbits);
	ktab->current_saltbits = saltbits;
}

/****************************************************************************\
\****************************************************************************/

/****************************************************************************/
static void create_key_table (key_tables *ktab, char *key_str, unsigned int *key_table)
{
	unsigned int v, v1, v2, *permuted_key;
	int ix;

	v1 = v2 = 0;
	permuted_key = &ktab->pc1_table[0][0][0];
	for ( ix=8; ix--; )
	{
		v1 |= permuted_key[*key_str   & 0x7f];
		permuted_key += 128;

		v2 |= permuted_key[*key_str++ & 0x7f];
		permuted_key += 128;
	}

	for ( ix=0; ix<16; ix++ )
	{
		permuted_key = &ktab->pc2_table[0][0];
		v1 = (v1 << rots[ix]) | (v1 >> (28 - rots[ix]));
		v  = permuted_key[(v1 >> 21) & 0x7f]; permuted_key += 128;
		v |= permuted_key[(v1 >> 14) & 0x7f]; permuted_key += 128;
		v |= permuted_key[(v1 >>  7) & 0x7f]; permuted_key += 128;
		v |= permuted_key[(v1      ) & 0x7f]; permuted_key += 128;
		*key_table++ = v;

		v = 0;
		v2 = (v2 << rots[ix]) | (v2 >> (28 - rots[ix]));
		v |= permuted_key[(v2 >> 21) & 0x7f]; permuted_key += 128;
		v |= permuted_key[(v2 >> 14) & 0x7f]; permuted_key += 128;
		v |= permuted_key[(v2 >>  7) & 0x7f]; permuted_key += 128;
		v |= permuted_key[(v2      ) & 0x7f];
		*key_table++ = v;
	}
}

/****************************************************************************/
void AxInitKeys (key_tables *ktab, key_ring *kr, char *key_str)
{
	unsigned int key_table[16][2], val;
	int ix, jx;

	/* copy the key string onto the key ring */
	if ( key_str )
	{
		if ( strlen(key_str) < KEY_LEN )
			strcpy(kr->key_str, key_str);
		else
			strncpy(kr->key_str, key_str, KEY_LEN);
	}

	/* create the encryption key */
	create_key_table(ktab, kr->key_str, &key_table[0][0]);
	for ( ix=0; ix<16; ix++ )
	{
		for ( jx=0; jx<2; jx++ )
			kr->encrypt_key[ix][jx] = key_table[ix][jx];
	}

	/* create the decryption key */
	create_key_table(ktab, kr->key_str, &key_table[0][0]);
	for ( ix=0; ix<8; ix++ )
	{
		val = key_table[15-ix][0];
		kr->decrypt_key[15-ix][0] = key_table[ix][0];
		kr->decrypt_key[ix][0] = val;

		val = key_table[15-ix][1];
		kr->decrypt_key[15-ix][1] = key_table[ix][1];
		kr->decrypt_key[ix][1] = val;
	}
}

/****************************************************************************/
void AxInitializeEncryption (key_tables *ktab, key_ring *kr, char *key_str)
{
	AxInitCrypt(ktab);
	AxInitKeys(ktab, kr, key_str);
}

/****************************************************************************/
void AxLoadKeytables(int do_load)
{
	if ( do_load == FALSE )
		AxKtab = AxSKtab = (key_tables *)0;
	else
	{
        AxKtab = (key_tables *)TaskCalloc(0, 1, sizeof(key_tables));
        AxInitCrypt(AxKtab);

        AxSKtab = (key_tables *)TaskCalloc(0,1, sizeof(key_tables));
        AxInitCrypt(AxSKtab);
    }
    AxKeys = AxSKeys = (key_ring *)0;
}

/****************************************************************************/
void AxSetEncryptionState(int state)
{
	AxUseEncryption = state ? TRUE : FALSE;
}

/****************************************************************************/
int AxGetEncryptionState()
{
	return(AxUseEncryption);
}

/****************************************************************************\
\****************************************************************************/

/****************************************************************************/
static void final_permute(key_tables *ktab,
                          unsigned int l1, unsigned int l2,
                          unsigned int r1, unsigned int r2,
                          unsigned int *val1,
                          unsigned int *val2)
{
	unsigned int v1, v2, x;

	x = (l1 ^ l2) & ktab->current_saltbits;
	l1 ^= x;
	l2 ^= x;

	x = (r1 ^ r2) & ktab->current_saltbits;
	r1 ^= x;
	r2 ^= x;

	v1 = v2 = 0;
	l1 >>= 3;
	l2 >>= 3;
	r1 >>= 3;
	r2 >>= 3;

	v1 |= ktab->efp[15][ r2         & 0x3f][0];
	v2 |= ktab->efp[15][ r2 & 0x3f][1];
	v1 |= ktab->efp[14][(r2 >>= 6)  & 0x3f][0];
	v2 |= ktab->efp[14][ r2 & 0x3f][1];
	v1 |= ktab->efp[13][(r2 >>= 10) & 0x3f][0];
	v2 |= ktab->efp[13][ r2 & 0x3f][1];
	v1 |= ktab->efp[12][(r2 >>= 6)  & 0x3f][0];
	v2 |= ktab->efp[12][ r2 & 0x3f][1];

	v1 |= ktab->efp[11][ r1         & 0x3f][0];
	v2 |= ktab->efp[11][ r1 & 0x3f][1];
	v1 |= ktab->efp[10][(r1 >>= 6)  & 0x3f][0];
	v2 |= ktab->efp[10][ r1 & 0x3f][1];
	v1 |= ktab->efp[ 9][(r1 >>= 10) & 0x3f][0];
	v2 |= ktab->efp[ 9][ r1 & 0x3f][1];
	v1 |= ktab->efp[ 8][(r1 >>= 6)  & 0x3f][0];
	v2 |= ktab->efp[ 8][ r1 & 0x3f][1];

	v1 |= ktab->efp[ 7][ l2         & 0x3f][0];
	v2 |= ktab->efp[ 7][ l2 & 0x3f][1];
	v1 |= ktab->efp[ 6][(l2 >>= 6)  & 0x3f][0];
	v2 |= ktab->efp[ 6][ l2 & 0x3f][1];
	v1 |= ktab->efp[ 5][(l2 >>= 10) & 0x3f][0];
	v2 |= ktab->efp[ 5][ l2 & 0x3f][1];
	v1 |= ktab->efp[ 4][(l2 >>= 6)  & 0x3f][0];
	v2 |= ktab->efp[ 4][ l2 & 0x3f][1];

	v1 |= ktab->efp[ 3][ l1         & 0x3f][0];
	v2 |= ktab->efp[ 3][ l1 & 0x3f][1];
	v1 |= ktab->efp[ 2][(l1 >>= 6)  & 0x3f][0];
	v2 |= ktab->efp[ 2][ l1 & 0x3f][1];
	v1 |= ktab->efp[ 1][(l1 >>= 10) & 0x3f][0];
	v2 |= ktab->efp[ 1][ l1 & 0x3f][1];
	v1 |= ktab->efp[ 0][(l1 >>= 6)  & 0x3f][0];
	v2 |= ktab->efp[ 0][ l1 & 0x3f][1];

	*val1 = v1;
	*val2 = v2;
}

/****************************************************************************/
void crypt_block (key_tables *ktab, unsigned int *key, char *block)
{
	unsigned int l1, l2, r1, r2, v1, v2, s, x, ix;
	int i;

	/* Do initial permutation + E expansion */
	for ( i=0,l1=0,l2=0; i<24; i++ )
	{
		if ( block[initial_perm[esel[i]-1]-1] )
			l1 |= BITMASK(i);
	}
	for ( ; i<48; i++ )
	{
		if ( block[initial_perm[esel[i]-1]-1] )
			l2 |= BITMASK(i-24);
	}

	for ( i=0,r1=0,r2=0; i<24; i++ )
	{
		if ( block[initial_perm[esel[i]-1+32]-1] )
			r1 |= BITMASK(i);
	}
	for ( ; i<48; i++ )
	{
		if ( block[initial_perm[esel[i]-1+32]-1] )
			r2 |= BITMASK(i-24);
	}

	/* Do DES inner loops */
	ix = 1;
	while (ix--)
	{
		for ( i=8; i--; )
		{
			s = *key++ ^ r1;
			l1 ^= SBA(ktab->sb1, s & 0xffff);
			l2 ^= SBA(ktab->sb1, (s & 0xffff) + 4);
			l1 ^= SBA(ktab->sb0, s >>= 16);  
			l2 ^= SBA(ktab->sb0, (s)          + 4);

			s = *key++ ^ r2;
			l1 ^= SBA(ktab->sb3, s & 0xffff);
			l2 ^= SBA(ktab->sb3, (s & 0xffff) + 4);
			l1 ^= SBA(ktab->sb2, s >>= 16);  
			l2 ^= SBA(ktab->sb2, (s)          + 4);

			s = *key++ ^ l1;
			r1 ^= SBA(ktab->sb1, s & 0xffff);
			r2 ^= SBA(ktab->sb1, (s & 0xffff) + 4);
			r1 ^= SBA(ktab->sb0, s >>= 16);  
			r2 ^= SBA(ktab->sb0, (s)          + 4);

			s = *key++ ^ l2;
			r1 ^= SBA(ktab->sb3, s & 0xffff);
			r2 ^= SBA(ktab->sb3, (s & 0xffff) + 4);
			r1 ^= SBA(ktab->sb2, s >>= 16);  
			r2 ^= SBA(ktab->sb2, (s)          + 4);
		}

		s  = l1;
		l1 = r1;
		r1 = s;

		s  = l2;
		l2 = r2;
		r2 = s;
	}

	/* final conversion, and convert to bit array */
	final_permute(ktab, l1, l2, r1, r2, &v1, &v2);
	for ( i=0; i<32; i++ )
		*block++ = (v1 & intmask[i]) != 0;
	for ( i=0; i<32; i++ )
		*block++ = (v2 & intmask[i]) != 0;
}

/****************************************************************************/
void AxCrypt(key_tables *ktab, unsigned int *key, unsigned char* data, int len)
{
	unsigned char c, block[64];
	int ix, jx, kx, ind;

	for ( ix=0; ix<len; ix+=8 )
	{
		for( jx=0, ind=0; jx<8; jx++ )
		{
			c = data[ix+jx];
			for ( kx=0; kx<8; kx++ )
				block[ind++] = (c & bytemask[kx]) != 0;
		}

		crypt_block(ktab, key, (char *)block);

		for ( jx=0, ind=0; jx<8; jx++ )
		{
			c = 0;
			for ( kx=0; kx<8; kx++ )
			{
				if ( block[ind++] )
					c |= bytemask[kx];
			}
			data[ix+jx] = c;
		}
	}
}

/****************************************************************************/
void AxEncrypt(key_tables *ktab, key_ring *kr, unsigned char* data, int len)
{
	AxCrypt(ktab, &kr->encrypt_key[0][0], data, len);
}

/****************************************************************************/
void AxDecrypt(key_tables *ktab, key_ring *kr, unsigned char* data, int len)
{
	AxCrypt(ktab, &kr->decrypt_key[0][0], data, len);
}

/****************************************************************************\
\****************************************************************************/

/****************************************************************************/
int AxCreateKeys(key_tables *ktab, key_ring **kr, char *passwd)
{
	if ( ktab == (key_tables *)0 )
		return(FALSE);
	else
	{
		*kr = (key_ring *)TaskCalloc(0, 1, sizeof(key_ring));
		AxInitKeys(ktab, *kr, passwd);
		return(TRUE);
	}
}

/****************************************************************************/
void AxDestroyKeys(key_ring **kr)
{
	TaskFree(0, *kr);
	*kr = (key_ring *)0;
}

/****************************************************************************/
int AxAddKeys(char *passwd)
{
	return(AxCreateKeys(AxKtab, &AxKeys, passwd));
}

/****************************************************************************/
int AxGetWKeys(key_tables **ktab, key_ring **kr)
{
	data_ptr retP;
	char *pwd;
	int temp;

	temp = FALSE;
	*ktab = AxKtab;
	if ( *ktab == (key_tables *)0 || AxGetEncryptionState() == FALSE )
		*kr = (key_ring *)0;
	else    
	{       
		*kr = AxKeys;
		if ( *kr == (key_ring *)0 )
		{ 
			if ( ElfCallByName("do_w_passwd$", &retP, 0) != AOK )
			{
				*ktab = (key_tables *)0;
				*kr = (key_ring *)0;
			}       
			else
			{
				if ( AxKeys != (key_ring *)0 )
					*kr = AxKeys;
				else if ( !retP )
				{
					*ktab = (key_tables *)0;
					*kr = (key_ring *)0;
				}       
				else
				{
					pwd = (char *)AxStrPtrFromDataPtr(retP);
					AxCreateKeys(*ktab, kr, pwd);
					temp = TRUE;
	            }
				if ( retP )
					AxFreeData(retP);
				if ( ElfRetData )
					ElfRetData = (data_ptr)0;
			}
		}
	}
	return(temp);
}

/****************************************************************************/
static int do_get_r_keys(key_tables **ktab, key_ring **kr, int force_temp)
{
	data_ptr retP;
	char *pwd;
	int temp;

	temp = force_temp;
	*ktab = AxKtab;
	if ( *ktab == (key_tables *)0 )
	{
		AxLoadKeytables(TRUE);
		*ktab = AxKtab;
	}
	if ( *ktab == (key_tables *)0 )
		*kr = (key_ring *)0;
	else    
	{
		*kr = temp ? (key_ring *)0 : AxKeys;
		if ( *kr == (key_ring *)0 )
		{ 
			if ( ElfCallByName("do_r_passwd$", &retP, 0) != AOK )
			{
				*ktab = (key_tables *)0;
				*kr = (key_ring *)0;
			}       
			else
			{
				if ( temp == FALSE && AxKeys != (key_ring *)0 )
					*kr = AxKeys;
				else if ( !retP )
				{
					*ktab = (key_tables *)0;
					*kr = (key_ring *)0;
				}       
				else
				{
					pwd = (char *)AxStrPtrFromDataPtr(retP);
					AxCreateKeys(*ktab, kr, pwd);
					temp = TRUE;
				}
				if ( retP )
					AxFreeData(retP);
				if ( ElfRetData )
					ElfRetData = (data_ptr)0;
			}
		}
	}
	return(temp);
}

/****************************************************************************/
int AxGetRKeys(key_tables **ktab, key_ring **kr)
{
	return(do_get_r_keys(ktab, kr, TRUE));
}

/****************************************************************************/
int AxGetRKeys2(key_tables **ktab, key_ring **kr, int force_temp)
{
	return(do_get_r_keys(ktab, kr, force_temp));
}

/****************************************************************************/
int AxAddSKeys(char *passwd)
{
	return(AxCreateKeys(AxSKtab, &AxSKeys, passwd));
}

/****************************************************************************/
int AxGetWSKeys(key_tables **ktab, key_ring **kr)
{
        data_ptr retP;
        char *pwd;
        int temp;

	temp = FALSE;
        *ktab = AxSKtab;
        if ( *ktab == (key_tables *)0 )
                *kr = (key_ring *)0;
        else    
        {       
                *kr = AxSKeys;
                if ( *kr == (key_ring *)0 )
                { 
			if ( ElfCallByName("do_w_passwd$", &retP, 0) != AOK )
			{
				*ktab = (key_tables *)0;
				*kr = (key_ring *)0;
			}       
			else
			{
				if ( AxSKeys != (key_ring *)0 )
					*kr = AxSKeys;
				else if ( !retP )
				{
					*ktab = (key_tables *)0;
					*kr = (key_ring *)0;
				}       
				else
				{
					pwd = (char *)AxStrPtrFromDataPtr(retP);
					AxCreateKeys(*ktab, kr, pwd);
					temp = TRUE;
	                        }
				if ( retP )
					AxFreeData(retP);
				if ( ElfRetData )
					ElfRetData = (data_ptr)0;
			}
		}
	}
	return(temp);
}

/****************************************************************************/
int AxGetRSKeys(key_tables **ktab, key_ring **kr)
{
        data_ptr retP;
        char *pwd;
        int temp;

	temp = TRUE;
        *ktab = AxSKtab;
        if ( *ktab == (key_tables *)0 )
                *kr = (key_ring *)0;
        else    
        {       
		if ( ElfCallByName("do_r_passwd$", &retP, 0) != AOK )
		{
			*ktab = (key_tables *)0;
			*kr = (key_ring *)0;
		}       
		else
		{
			if ( !retP )
			{
				*ktab = (key_tables *)0;
				*kr = (key_ring *)0;
			}       
			else
			{
				pwd = (char *)AxStrPtrFromDataPtr(retP);
				AxCreateKeys(*ktab, kr, pwd);
				AxFreeData(retP);
			}
			if ( ElfRetData )
				ElfRetData = (data_ptr)0;
		}
	}
	return(temp);
}

int AxDecryptBuffer(char *in_buf, int in_buf_len, int pid)
{
	int err, buf_len, hdr_len, tmp_keys, valid_buf, i;
	key_tables *ktab;
	key_ring *kr;
	char *buf, hdr[STRPATH+1], *tmp_buf;
	AFileInfo f_info;

	/* check the header for encryption */
	AFileGetBuffer(in_buf, in_buf_len, &buf, &buf_len);
	hdr_len = in_buf_len - buf_len;
	if ( hdr_len < STRPATH && hdr_len > 0 )
	{
		strncpy(hdr, in_buf, hdr_len);
		hdr[hdr_len] = (char)0;
		err = AFileInfoFromHeader(hdr, &f_info);
		if ( err == AOK && f_info.encrypted )
		{
			/* get the decryption key */
			tmp_keys = AxGetRKeys2(&ktab, &kr, FALSE);
			if ( kr == (key_ring *)0)
				return FEbadDecrypt;

			/* try the default key key first */
			valid_buf = FALSE;
			if ( tmp_keys == FALSE )
			{
				/* backup the buffer, just in case */
				tmp_buf = (char *)TaskAlloc(pid, buf_len);
				memcpy(tmp_buf, buf, buf_len);

				/* decrypt that buffer */
				AxDecrypt(ktab, kr, (unsigned char *)buf, buf_len);

				/* validate it */
				valid_buf = AFileValidateHeaders(in_buf, in_buf_len);

				/* restore buf, get tmp key from user */
				if ( valid_buf == FALSE )
				{
					memcpy(buf, tmp_buf, buf_len);
					tmp_keys = AxGetRKeys2(&ktab, &kr, TRUE);
				}

				/* free the buffer */
				TaskFree(pid, tmp_buf);

				/* return if attempt failed, and no new key */
				if ( kr == (key_ring *)0)
					return FEbadDecrypt;
			}

			/* if not decrypted yet, try getting another key */
			if ( valid_buf == FALSE )
			{
				AxDecrypt(ktab, kr, (unsigned char *)buf, buf_len);
				if ( tmp_keys )
					AxDestroyKeys(&kr);
				valid_buf = AFileValidateHeaders(in_buf, in_buf_len);
				if ( valid_buf == FALSE )
					return FEbadDecrypt;
			}
			if ( valid_buf )
				AxOverrideEncryptionHeader(in_buf,  in_buf_len);
		}
		else
			return AOK;
	}
	else
		return AOK;

	return err;
}
