/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axcontext.c
   Applix ASC version: 9
   Date: Sat Apr 17 21:17:36 1999
*/
#include <stdio.h>
#include <aos.h>
#include <scac.h>
#include <thimble.h>

#define AXSHLIB_DEFINES_ONLY
#include <axshlib.h>			/* Need spelling shared library name */

#include <axcontext.p>



/*************************************************************************
function:  AxReturnSSUidForGE
*/
int
AxReturnSSUidForGE(pathname, quiet, read_only)
char *pathname;
int quiet;
int read_only;
{
	char *sspathname;
	int old_pid, uid, top_pid, pid;
	int i;
	
	/* save current pid */
	old_pid = THIMpid;
	uid = -1;

	/* is this task the SS? */
	top_pid = ThimTopPidFromPid(THIMpid);
	if ( TaskInfo[top_pid].appid == APSS )
		uid = TaskInfo[top_pid].uid;
	else
   	{
		/* is this task a GE, running as a SS parasite? */
		if ( TaskInfo[top_pid].appid == APGE && TaskInfo[top_pid].parasite_of )
		{
			pid = TaskInfo[top_pid].parasite_of;
			if ( TaskInfo[pid].appid == APSS )
				uid = TaskInfo[pid].uid;
		}

		/* try to find an SS with the datafile open */
		if ( uid < 0 && pathname )
   		{
   			for ( i=0; i<MAX_PIDS; i++ )
		   	{	
				if ( TaskInfo[i].appid == APSS )
			   	{
			   		VfyContext(i);
					sspathname = (char *)GetSSPathname();
			   		if ( strcmp(sspathname, pathname) == 0 )
		   			{
						uid = TaskInfo[i].uid;
	   					break;
					}
				}
			}
			VfyContext(old_pid);

			/* no success - open SS and make the task a parasite of the GE */
			if ( uid < 0 )
			{
				data_ptr argsP, uidP;
				argsP = AxMakeArray(5);
				argsP = AxAddStrToArray(argsP, 0, pathname);
				argsP = AxAddIntToArray(argsP, 2, quiet);
				argsP = AxAddIntToArray(argsP, 3, read_only);
				uidP = AxfOpenDoc(argsP);
				uid = AxIntFromDataPtr(uidP);
				AxFreeData(uidP);
				AxFreeData(argsP);
			}
		}
	}
	return(uid);
	
}

/*************************************************************************
function: AxSSSetFuncCallCounter - note the builtin function calling depth
*		for a spreadsheet task
*/
AxSSSetFuncCallCounter(ssPid, count)
	int ssPid;
	int count;
{
	TaskInfo[ssPid].ss_func_call_counter =  count;
   return (TRUE);
}

/*************************************************************************
function:  AxSSFuncCallCounter - return the builtin function calling depth
*/
AxSSFuncCallCounter(ssPid)
	int ssPid;
{
	return(TaskInfo[ssPid].ss_func_call_counter);
}

/*************************************************************************
function: AxSSSetTaskCallingBltinFunc - make a note of the AxInTask at the
*	time a builtin function was called in the SS.
*/
AxSSSetTaskCallingBltinFunc(ssPid)
	int ssPid;
{
	TaskInfo[ssPid].ss_task_calling_bltin_func = AxInTask;
   return (TRUE);
}

/*************************************************************************
function: *	AxSSTaskCallingBltinFunc - return the AxInTask at the time of
*		calling the built in function in a Spreadsheet task
*/
AxSSTaskCallingBltinFunc(ssPid)
	int ssPid;
{
	return(TaskInfo[ssPid].ss_task_calling_bltin_func);
}



