/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axdir.c
   Applix ASC version: 36
   Date: Mon Dec 21 21:11:27 1998
*/
#include <aos.h>
#if AX_NT
#include <windows.h>
#endif
#if AX_UNIX     /*  {  */

#if X386_NEW || M_88K || SYSV4      /*   {   */

  /** These are here because the system call returns a variable sized */
  /** block and we want to malloc an array of them so they all need to */
  /**  be the same size    - jon         */
#define MAXNAMLEN   512     /* maximum filename length */
struct dirent               /* data from readdir() */
    {
    long        d_ino;      /* inode number of entry */
    off_t       d_off;      /* offset of disk direntory entry */
    unsigned short  d_reclen;   /* length of this record */
    char        d_name[255];    /* JJM was 1 - name of file */
    };
#define DIRBUF      1048        /* buffer size for fs-indep. dirs */
 
typedef struct
    {
    int dd_fd;          /* file descriptor */
    int dd_loc;         /* offset in block */
    int dd_size;        /* amount of valid data */
    char    *dd_buf;        /* directory block */
    }   DIR;            /* stream data from opendir() */
 
extern DIR      *opendir();
extern struct dirent    *readdir();
extern long     telldir();
extern void     seekdir();
extern int      closedir();
#define direct dirent       /* JJM */
 
#else       /*  }  {  */
#if RIOS || CLIX || ALPHA || SUN4 || LINUX || SGI /* { */
#include <dirent.h>
#if (!SUN4) /* { */
#define direct dirent
#endif		/* } */
#else  /* } { */
#include <sys/dir.h>
#endif  /* } */
#endif     /*  }  */

#endif     /*  }  */
#define ALCSTR(p) strcpy((char *)malloc(strlen(p)+1),p)

#if (AX_UNIX && (SUN4 || SGI || LINUX))
/***************************************************************************
function: axListFilesInDirectory - list all the files in the passed
directory name. List is null terminated.
AxListFilesFreer should be called to release the memory.
*/
static char **
axListFilesInDirectory (dirName)
char *dirName;
{
	register int count;
	register struct dirent *direntP;
	DIR *dirP;
	char **retP;
	
	retP = NULL;
	if ((dirP = opendir(dirName)) == NULL) {
		retP = (char **)malloc(sizeof(char *));
		retP[0] = NULL;
	}
	else {
		retP = (char **)malloc(sizeof(char *));
		for (count = 0, direntP = readdir(dirP); direntP;
			 			count++, direntP = readdir(dirP)) {
			retP = (char **)realloc(retP, (count+2)*sizeof(char*));
			retP[count] = ALCSTR(direntP->d_name);
		}
		retP[count] = NULL;
		closedir(dirP);
	}
	return(retP);
}
#endif /* AX_UNIX && SUN4 */

#if (AX_UNIX && CLIX)
/***********************************************************************
function: clix_list_files_in_dir - List all files in directory DIRNAME.
List is null terminated.
Call AxListFilesFreer() to release memory.
*/
static char **
clix_list_files_in_dir(dirname)
char *dirname;
{
	register int cnt;
	register struct dirent *dp;
	DIR *dirp;
	char **retP;
	char file[15];

	retP = (char **) malloc(sizeof(char *));
	retP[0] = NULL;

	if ((dirp = opendir(dirname)) == NULL)
		return(retP);

	for (cnt = 0, dp = readdir(dirp); dp; cnt++, dp = readdir(dirp)) {
		retP = (char **) realloc(retP, (cnt + 2) * sizeof(char *));
		/* See <sys/dirent.h> on clix.  Defn of `d_name' member is:
		   char d_name[14]; */
		strncpy(file, dp->d_name, 14);
		file[14] = '\0';
		retP[cnt] = ALCSTR(file);
	}

	retP[cnt] = NULL;
	closedir(dirp);

	return(retP);
}
#endif /* (AX_UNIX && CLIX) */

/*************************************************************************
function: AxListFilesInDir - Return malloced list of filenames
The array is null pointer termed.
At leaast on Unix machines, if the return value is NULL, errno is set.
AxListFilesFreer should be called to release the memory.
*/
char **AxListFilesInDir (dirname)
char *dirname;
{
#if AX_NT
	WIN32_FIND_DATA data;
	HANDLE h;
	int cnt = 0;
    int len;
	char **retP;
	char name[STRPATH+11];
    int more_files = TRUE;

    len = strlen( dirname );
    if ( *(dirname+len-1) == '\\' )
       /* already a trailing slash */
       sprintf(name,"%s*.*",dirname);
    else
	   sprintf(name,"%s\\*.*",dirname);

	retP = (char**)calloc(1,sizeof(char*));
	h = FindFirstFile(name,&data);
	if ( h == (HANDLE)-1 )
		return(retP);
    
	while ( more_files ) {
        if ( data.dwFileAttributes & 
                  (FILE_ATTRIBUTE_HIDDEN | FILE_ATTRIBUTE_SYSTEM) )
           {
           /* skip hidden or system files */
           more_files = FindNextFile(h,&data);
           continue;
           }
        if ( (strcmp( data.cFileName, "." ) == 0 ) ||
             (strcmp( data.cFileName, ".." ) == 0 ) )
           {
           /* strip these which are always reported */
           more_files = FindNextFile(h,&data);
           continue;
           }

		retP = (char**)realloc(retP,(cnt+2)*sizeof(char*));
		retP[cnt++] = ALCSTR(data.cFileName);
		AXLOG(("AxListFilesInDir %s\n",data.cFileName));
        more_files = FindNextFile(h,&data);
	}
	FindClose(h);
	retP[cnt] = 0;
	return(retP);
#endif		/* AX_NT */

#if AX_UNIX
	struct direct **namelist;
	int cnt, ix;
	char **retP;

	if (!AxAccessAllowed(dirname)) {
		char **retP;

		retP = (char**)malloc(sizeof(char*));
		retP[0] = 0;
		return(retP);
	}
#if (SUN4 || SGI || LINUX)
	/* I am scared to delete any of this interleaved web of ifdefs and
	  elifs, so I am including another flavour of my own listFiles() - Babu */
	return(axListFilesInDirectory(dirname));	/* scandir() fails for /pcfs
												   mounted filesystems -Babu */
#else
#if CLIX
	return(clix_list_files_in_dir(dirname));
#else

#if SYSV4 || M_88K || RIOS || SCOUNIX
    cnt = our_scandir(dirname, &namelist, NULL, NULL);
#else
	cnt = scandir(dirname, &namelist, NULL, NULL);
#endif
	if (cnt < 1) {			/* 'scandir' returns -1 if no such directory */
		retP = (char**)malloc(sizeof(char*));
		retP[0] = 0;
		return(retP);
	}
	retP = (char**)malloc((cnt+1)*sizeof(char*));
	for ( ix = 0; ix < cnt; ix++ ) {
#if SUN4 || M_88K || HP800 || HP300 || RIOS || ALPHA || USL54 || SCOUNIX || SGI || SUNIS || LINUX || NEC_EWS
		retP[ix] = ALCSTR(namelist[ix]->d_name);
#else
		retP[ix] = ALCSTR(namelist[ix]->gd_name);
#endif
		free(namelist[ix]);
	}
	free(namelist);
	retP[ix] = 0;
	return(retP);
#endif		/* SUN4 */
#endif      /* CLIX */
#endif		/* AX_UNIX */
}

/*************************************************************************
function: AxListFilesFreer - Release memory from AxListFilesInDir
*/
void AxListFilesFreer (p)
char **p;
{
	int ix;

	if ( !p )
		return;
	for ( ix = 0; p[ix]; ix++ ) {
		AXLOG(("AxListFilesFreer %s\n",p[ix]));
		free(p[ix]);
	}
	free(p);
}

/*    #if (SYSV4 && ELFRT_ONLY)   */
#if SYSV4 || M_88K || RIOS || SCOUNIX

/* Code below lifted from the lovely agetnfn.c... */

#define direct dirent
#include <sys/types.h>  /* for ino_t and off_t */
#include <sys/param.h>  /* for MAXNAMELEN */

/*
 * NOTE:
 * We need our own "struct dirent" because we do a structure assignment in
 * scandir().  On SYSV ports, the d_name[] element of "struct dirent" is a
 * char array of size 2, so a structure assignment copies only the first two
 * chars of the filename.
 */

 
/*
 *  Name: our_scandir() - scan a directory.
 *
 *  Description: Scandir reads the directory dirname and builds an array of
 *      pointers to directory entries using malloc(3).  The second parameter is
 *      a pointer to an array of structure pointers. The third parameter is a
 *      pointer to a routine which is called with a pointer to a directory
 *      entry and should return a none zero value if the directory should be
 *      included in the array.  If this pointer is null, than all the directory
 *      entries will be included.  The last argument is a pointer to a routine
 *      which is passed to qsort(3) to sort the completed array.  If this
 *      pointer is null, the array is not sorted.
 *
 *      Scandir returns the number of entries in the array and a pointer to the
 *      array through the parameter namelist.
 */

our_scandir(dirname,namelist,select,compar)
char *dirname;
struct direct ***namelist;
int (*select)();
int (*compar)();
{
    DIR *fp;
    struct direct *dir,*dirptr,**nameptr;
    int array_cnt = 0;
    int dir_cnt = 0;
 
    if(!(fp = opendir(dirname)))
        return(-1);
 
    while(dir = readdir(fp)) {
        if(!dir->d_ino || (select && !(*select)(dir)))
            continue;
        if(dir_cnt == array_cnt) {
            if(array_cnt) {
                array_cnt *= 2;
                if(!(nameptr = (struct direct **)

                    malloc((unsigned int)
                        (sizeof(struct direct *) * array_cnt))))
                    goto free_mem;
                bcopy((char *)*namelist,(char *)nameptr,(int)
                    (sizeof(struct direct *) * dir_cnt));
                free((char *)*namelist);
                *namelist = nameptr;
            }
            else {
                array_cnt = 64;
                if(!(*namelist = nameptr = (struct direct **)
                    malloc((unsigned int)(sizeof(struct direct *) * array_cnt))))
                    return(-1);
            }
        }
        if(!(dirptr = (struct direct *)
            malloc((unsigned int)sizeof(struct direct))))
            goto free_mem;
        *dirptr = *dir;
        nameptr[dir_cnt++] = dirptr;
    }
    closedir(fp);
 
    if(compar && dir_cnt > 1)
        qsort(nameptr,dir_cnt,(int)sizeof(struct direct *),*compar);

 
    return(dir_cnt);
 
free_mem:
    closedir(fp);
    nameptr = *namelist;
    while(dir_cnt--)
        free((char *)nameptr[dir_cnt]);
    free((char *)nameptr);
    return(-1);
}
#endif  /* SYS&SYS5 */
 

