/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axfs2.c
   Applix ASC version: 25
   Date: Wed Mar 10 21:09:56 1999
*/
#define NO_APX_INTERNATIONAL 1


#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <aos.h>
#include <fsrts.h>
#include <afile.h>
#include <getstr10.h>
#include <doctypes.h>
#include <thimble.h>
#include <scac.h>



extern doc_info *DocList;       /* Head of documents list */


extern char *ThimApplixDir; /* executables -> /usr/bin//thimble.data */
extern char ThimTempDir[];  /* dir to put allemp files in */
extern char *DirSlash;
extern char *AxSocketName;



char *ThimTextSfx = NXLT("aw");
char *ThimSprdSfx = NXLT("as");
static char ThimPictSfx[] = NXLT("ag");
static char ThimElfcSfx[] = NXLT("am");
static char ThimQuerySfx[] = NXLT("aq");
static char ThimEqnSfx[] = NXLT("ae");
static char ThimAudiSfx[] = NXLT("aa");
static char ThimPostSfx[] = NXLT("apt");
static char ThimMacrSfx[] = NXLT("aM");
static char ThimDlgSfx[] = NXLT("d");
static char ThimBldrSfx[] = NXLT("ab");         /* Applix Builder */
static char ThimBldrTurboSfx[] = NXLT("abo");   /* Applix Builder Turbo*/
static char ThimBldrDistSfx[] = NXLT("abd");   /* Applix Builder Turbo*/
static char ThimBmapSfx[] = NXLT("im");         /* Applix Bitmap */
static char ThimSlideSfx[] = NXLT("ap");        /* Applix SlideShow */

char *SuffixFromDocType();
char *Amktemp();
elfData AxCallAxMain();
elfData AxElfCallProgram();

/*************************************************************************
function: MyError - Wrapper so error can be thrown from axmain/standalone
*/
static MyError(c,s,o)
int c;
char *s, *o;
{
#if FOR_VERSION_FILTERS
    AxError(c,s,o);
#else
    ElfStrAbort(c,s,o);
#endif
}

AxAppendTemplateStr(str, doc_type)
char *str;
int doc_type;
{
    strcat(str, NXLT("a%%%%%%%."));
    strcat(str, SuffixFromDocType(doc_type));
}

/*************************************************************************
function: FSFilterDocToCurRev - passed the path to an Applixware doc,
this function applies whatever filter is necessary to produce a current
revision document.  Called from the application file open'ers, as well
as the various filter readers which expect a document of the current rev
as input.
infile = name of the file to filter
outfile (in) pointer to spaces for at least STRPATH+1 characters
outfile (out) contains name of doc in current rev. this may be the
same as infile or it may be the name of a temporary file. The caller must
delete this temp file after use, BUT ONLY IF THIS FUNCTION RETURNS TRUE!

Return value = TRUE iff a filter has been run and a temp file has been created
which contains the filtered version of the document.

Errors are thrown using AxError().

*/
#ifdef AX_USE_PROTOS
int
FSFilterDocToCurRev(
    char *infile,
    char *outfile)
#else
int
FSFilterDocToCurRev(infile,outfile)
char *infile;
char *outfile;
#endif
{
    AFileInfo axSWInfo;
    long sw_rev;

    AGetDefaultFileInfo(&axSWInfo);

    sw_rev = axSWInfo.version;

    sw_rev = (sw_rev == 320) ? 400 : sw_rev;

    return FSFilterDocToAnyRev(infile,outfile,sw_rev);
}

/*************************************************************************
function: FSFilterDocToAnyRev - Like FSFilterDocToCurRev, but accepts
the desired rev number (e.g. 311) as an argument. Intended to be used
by the cross-rev filters for insets, etc; in this case the filter
determines the target rev.
returns TRUE iff a filter has been run and a temp file created.
*/
#ifdef AX_USE_PROTOS
int
FSFilterDocToAnyRev(
    char *infile,
    char *outfile,
    long rev)
#else
int
FSFilterDocToAnyRev(infile,outfile,rev)
char *infile;
char *outfile;
int rev;
#endif
{
    int filterRun = FALSE;
    FILE *fp;
    int read_only=TRUE;
    struct stat status;
    AFileInfo axDocInfo;
    mbchar *line, lineBuf[301];
    char filename[STRPATH+1];
    char pathname[STRPATH+1];
    char cmd[STRPATH+128];
    char buf[301];
    char filtname[STRNORMAL+1];
    char template[STRPATH+1];
    char *tempfile;
    int foundFilter,elfFilter;
    long doc_rev, sw_rev=rev;
    typeErr ret;
    char *str, *obj;
    int code, pid;
    elfData args, data, pathdata, result;
    char errstr[1000];
    char errobj[1000];


    /*
     * retrieve revision info from the doc
     */

    line = lineBuf;

#if FOR_VERSION_FILTERS
    fp = fopen( infile, READ_STR );
#else
    AFTaskLockFile(infile, THIMpid, &read_only, TRUE);
    fp = (FILE *)AFTaskfopen( THIMpid, infile, READ_STR);
#endif
    if (fp)
    {
        line = (mbchar *)fgets( (char *)line, 300, fp );

        if (ferror(fp))
        {
#if FOR_VERSION_FILTERS
            fclose( fp );
#else
            AFfclose( infile, TRUE, fp);
            AFTaskUnlockFile(infile, THIMpid, TRUE);
#endif
            MyError(ErrnoErr(), NULL, infile);
        }

        ret = AFileInfoFromHeader(lineBuf, &axDocInfo);
        if ( (ret == FENoHeader) || (ret == FEInvalidHeader) )
        {
            /*
             * failed to retrieve rev info from doc header -
             * not an Applixware doc?  Let the default behavior
             * handle this condition
             */

            doc_rev = 0;
            sw_rev = 0;

#if FOR_VERSION_FILTERS      
            fclose( fp );
#else
            AFfclose( infile, TRUE, fp);
            AFTaskUnlockFile(infile, THIMpid, TRUE);
#endif
            strcpy(outfile,infile);

        }
        else
        {
#if FOR_VERSION_FILTERS      
            fclose( fp );
#else
            AFfclose( infile, TRUE, fp);
            AFTaskUnlockFile(infile, THIMpid, TRUE);
#endif
            doc_rev = axDocInfo.version;

            /*
             * treat rev 320 as 400 to handle docs created
             * with internal rev software never seen by
             * customers...
             */

            doc_rev = (doc_rev == 320) ? 400 : doc_rev;

            sw_rev = (sw_rev == 320) ? 400 : sw_rev;

            /*
             * compare the software revision to the rev of
             * the document we're trying to read
             */

            if (sw_rev == doc_rev)
            {
                /* same rev - no filtering necessary */

                strcpy(outfile,infile);

            }
            else
            {
                switch(axDocInfo.docType)
                {
                    case AX_DOC_TYPE_ASTERX_WORDS:
                    case AX_DOC_TYPE_ASTERX_MACROS:
                    case AX_DOC_TYPE_ASTERX_EQUATIONS:
                        sprintf(filtname,"wp%d%d",doc_rev,sw_rev);
                        break;

                    case AX_DOC_TYPE_ASTERX_GRAPHICS:
                        sprintf(filtname,"gr%d%d",doc_rev,sw_rev);
                        break;

                    case AX_DOC_TYPE_ASTERX_SS:
                        sprintf(filtname,"ss%d%d",doc_rev,sw_rev);
                        break;

                    case AX_DOC_TYPE_ASTERX_QUERYDATA:
                        sprintf(filtname,"aq%d%d",doc_rev,sw_rev);
                        break;

                    case AX_DOC_TYPE_ASTERX_BUILDER:
                        sprintf(filtname,"ab%d%d",doc_rev,sw_rev);
                        break;

                    case AX_DOC_TYPE_ASTERX_BITMAP:
                        sprintf(filtname,"im%d%d",doc_rev,sw_rev);
                        break;

                    default:
                        /* Ship this filter as needed */
                        sprintf(filtname,"xx%d%d",doc_rev,sw_rev);
                        break;
                }

                sprintf(pathname, NXLT("%s%sfilters%s%s"),
                  ThimApplixDir,DirSlash,DirSlash,filtname);
#if AXMS
                strcat(pathname, NXLT(".exe"));
#endif
                errno = 0;

                /*
                 * see if the filter exists - if not, give the
                 * original file to the app and hope for the best
                 */ 

                foundFilter = (Astat(pathname, &status) == 0);
                elfFilter = FALSE;
                if (!foundFilter) {
                    strcat(pathname, ".am");
                    foundFilter = (Astat(pathname, &status) == 0);
                    elfFilter = TRUE;
                }
                if (foundFilter)
                {
                    if (status.st_size != 0)
                    {
                        /*
                         * zero-length filters are used to indicate
                         * that the app can open this cross-rev doc
                         * without any modifications
                         */


                        /* Specify a temp file in which to put the
                            filtered file. When built into axmain, we
                            can use the high-level temp file fcn (that
                            will cause the temp file to be deleted when
                            the axmain exits). When linked into the
                            standalone filters, however, we use a lower
                            level function that merely generates the
                            name of an available temp file. The caller
                            should delete this file when no longer needed. */

#if !FOR_VERSION_FILTERS                    /* If linked into axmain... */
                        template[0] = (char)0;
                        AxAppendTemplateStr(template, axDocInfo.docType);
                        if ( !AxTempFile(template, filename) )
                        {
                            MyError(ErrnoErr(), NULL, filename);
                        }
#else                                       /* Else a standalone filter... */
                        sprintf( template, NXLT("%s%s"), 
                                                ThimTempDir, DirSlash);
                        AxAppendTemplateStr(template, axDocInfo.docType);
                        tempfile = Amktemp( template);
                        if (!tempfile)
                        {
                            MyError(ErrnoErr(), NULL, XLT("temp file"));
                        }
                        strcpy( filename, tempfile);
#endif

                        sprintf(cmd, NXLT("%s %s %s"),
                          pathname, infile, filename);

                        /* Support ELF api */
                        data = (elfData) AxMakeArray(6);
                        data = (elfData) AxAddStrToArray(data,0,infile);
                        data = (elfData) AxAddStrToArray(data,1,filename);
                        data = (elfData) AxAddStrToArray(data,2,ThimTempDir);
                        data = (elfData) AxAddStrToArray(data,3,ThimApplixDir);
                        data = (elfData) AxAddStrToArray(data,4,DirSlash);
                        data = (elfData) AxAddStrToArray(data,5,AxSocketName);
                        code = 0;

#if FOR_VERSION_FILTERS
                        /* Filters use this type of call */
                        if (elfFilter) {
                            /* get axmain to install and run a macro */
                            args = (elfData) AxMakeArray(2);
                            args = (elfData) AxAddStrToArray(args, 0, pathname);
                            args = (elfData) AxAddArrayToArray(args, 1, data);
                            result = AxCallAxMain(AxSocketName,
                                                NXLT("run_filter_for_filter$"),
                                                args,
                                                &code, &str, &obj);
                        }
                        else {
                            /* run program directly from here */
                            result = AxElfCallProgram( pathname, data,
                                        &code,&str,&obj);
                        }
#else
                        /* Axmain calls via ELF */
                        pathdata = (elfData) AxMakeStrData(-1, pathname);
                        pid = THIMpid;
                        AxTaskLock();
                        code = ElfCallByName(NXLT("run_filter_with_args$"), 
                                &result, 2, pathdata, data);
                        AxTaskUnlock();
                        VfyContext(pid);
                        if (code != AOK) {
                            str = AxStrPtrFromArray(result,0);
                            strncpy(errstr,str,1000);
                            errstr[999] = 0;
                            str = errstr;
                            obj = AxStrPtrFromArray(result,1);
                            strncpy(errobj,obj,1000);
                            errobj[999] = 0;
                            obj = errobj;
                        }
                        AxFreeData(result);
                        AxFreeData(pathdata);
                        AxFreeData(data);
#endif
                        if (code != 0)
                            MyError(code,str,obj);

                        /* point at the new (filtered) file */
                        strcpy(outfile, filename);
                        filterRun = TRUE;
                    }
                    else
                    {
                        strcpy(outfile, infile);
                    }
                }
                else
                {
                    /* filter doesn't exist - return input file */

                    strcpy(outfile, infile);
                }
            }
        }
    }
    else
    {
        MyError(ErrnoErr(), NULL, infile);
    }

    return filterRun;
}


 
#if !FOR_VERSION_FILTERS

/*************************************************************************
function: FSPathOfDocId - Given a doc id, return pathname
This is derived from FSGDInfoGivenID, to be ued by thim_io.c
*/
typeErr FSXPathOfDocId (pid,pathname)
int pid;
char *pathname;
{
    register doc_info *p;
    register int id;
 
    id = TaskInfo[pid].fscurrdoc;
    for ( p = DocList; p; p = p->next ) {
        if ( p->id == id ) {
            strcpy(pathname,p->pathname);
            return(AOK);
        }
    }
    return(FEENOENT);
}

#endif
 

/*************************************************************************
function: GetAppidFromDocType - return applix appid given doctype
*/
int GetAppidFromDocType(doctype)
int doctype; /* as defined in doctypes.h */
{
    int appid; /* return the appid */
  
    switch(doctype)
    {
        case AX_DOC_TYPE_ASTERX_WORDS:
            appid = APDC;
            break;
        case AX_DOC_TYPE_ASTERX_MACROS:
            appid = APELFDEBUG;
            break;
        case AX_DOC_TYPE_ASTERX_EQUATIONS:
            appid = APEQN;
            break;
        case AX_DOC_TYPE_ASTERX_GRAPHICS:
            appid = APGE;
            break;
        case AX_DOC_TYPE_ASTERX_SS:
            appid = APSS;
            break;
        case AX_DOC_TYPE_ASTERX_QUERYDATA:
            appid = -1;
            break;
        case AX_DOC_TYPE_ASTERX_BUILDER:
            appid = -1;
            break;
        case AX_DOC_TYPE_ASTERX_BUILDER_TURBO:
            appid = -1;
            break;
        case AX_DOC_TYPE_ASTERX_SLIDESHOW:
            appid = APGE;
            break;
        default:
            appid = -1;
    }
    return(appid);
}

/*************************************************************************
function: SuffixFromDocType - return applix suffix given doctype
*/
char *SuffixFromDocType(doctype)
int doctype; /* as defined in doctypes.h */
{
    char *suffix;
    static char empty[] = "";
 
    switch(doctype)
    {
        case AX_DOC_TYPE_ASTERX_WORDS:
            suffix = ThimTextSfx;
            break;
        case AX_DOC_TYPE_ASTERX_MACROS:
            suffix = ThimElfcSfx;
            break;
        case AX_DOC_TYPE_ASTERX_EQUATIONS:
            suffix = ThimEqnSfx;
            break;
        case AX_DOC_TYPE_ASTERX_GRAPHICS:
            suffix = ThimPictSfx;
            break;
        case AX_DOC_TYPE_ASTERX_SS:
            suffix = ThimSprdSfx;
            break;
        case AX_DOC_TYPE_ASTERX_QUERYDATA:
            suffix = ThimQuerySfx;
            break;
        case AX_DOC_TYPE_ASTERX_BUILDER:
            suffix = ThimBldrSfx;
            break;
        case AX_DOC_TYPE_ASTERX_BUILDER_TURBO:
            suffix = ThimBldrTurboSfx;
            break;
        case AX_DOC_TYPE_ASTERX_BUILDER_ABD:
            suffix = ThimBldrDistSfx;
            break;
        case AX_DOC_TYPE_ASTERX_SLIDESHOW:
            suffix = ThimSlideSfx;
            break;
        default:
            suffix = empty;
            break;
    }
    return suffix;
}

/*************************************************************************
function: ThimSfxFromType - Given a doc type, return suffix
*/
char *ThimSfxFromType (type)
int type;
{
    switch ( type ) {
    case MSpreadsheet:
        return(ThimSprdSfx);
    case MQuery:
        return(ThimQuerySfx);
    case MEquation:
        return(ThimEqnSfx);
    case MDrawing:
        return(ThimPictSfx);
    case MText:
        return(ThimTextSfx);
    case MUserCommand:
        return(ThimElfcSfx);
    case MAudio:
        return(ThimAudiSfx);
    case MPostit:
        return(ThimPostSfx);
    case MMacro:
        return(ThimMacrSfx);
    case MDialog:
        return(ThimDlgSfx);
    case MApplixBuilder:
        return(ThimBldrSfx);
    case MApplixBuilderTurbo:
        return(ThimBldrTurboSfx);
    case MApplixBuilderDistribution:
        return(ThimBldrDistSfx);
    case MBitmap:
        return(ThimBmapSfx);
    case MApplixSlideShow:
        return(ThimSlideSfx);
    }
    return("");
}
 
 

/*************************************************************************
function: DocTypeFromName - Return type of dir based on dirname suffix
*/
DocTypeFromName (name)
char *name;
{
    char *p;
 
    for ( p = name + strlen(name) - 1; p > name; p-- ) {
        if ( *p == '.' ) {
            p++;
#if AXMS
            /* need to do a case-insensitive search on NT. */
            if ( !stricmp(p,ThimTextSfx) )  return(MText);
            if ( !stricmp(p,ThimPictSfx) )  return(MDrawing);
            if ( !stricmp(p,ThimSprdSfx) )  return(MSpreadsheet);
            if ( !stricmp(p,ThimQuerySfx) ) return(MQuery);
            if ( !stricmp(p,ThimEqnSfx) )   return(MEquation);
            if ( !stricmp(p,ThimElfcSfx) )  return(MUserCommand);
            if ( !stricmp(p,ThimAudiSfx) )  return(MAudio);
            if ( !stricmp(p,ThimPostSfx) )  return(MPostit);
            if ( !stricmp(p,ThimMacrSfx) )  return(MMacro);
            if ( !stricmp(p,ThimDlgSfx) )   return(MDialog);
            if ( !stricmp(p,ThimBldrSfx) )  return(MApplixBuilder);
            if ( !stricmp(p,ThimBldrTurboSfx) )  return(MApplixBuilderTurbo);
            if ( !stricmp(p,ThimBldrDistSfx) )  return(MApplixBuilderDistribution);
            if ( !stricmp(p,ThimBmapSfx) )  return(MBitmap);
            if ( !stricmp(p,ThimSlideSfx) )  return(MApplixSlideShow);
#else
            if ( streq(p,ThimTextSfx) )  return(MText);
            if ( streq(p,ThimPictSfx) )  return(MDrawing);
            if ( streq(p,ThimSprdSfx) )  return(MSpreadsheet);
            if ( streq(p,ThimQuerySfx) ) return(MQuery);
            if ( streq(p,ThimEqnSfx) )   return(MEquation);
            if ( streq(p,ThimElfcSfx) )  return(MUserCommand);
            if ( streq(p,ThimAudiSfx) )  return(MAudio);
            if ( streq(p,ThimPostSfx) )  return(MPostit);
            if ( streq(p,ThimMacrSfx) )  return(MMacro);
            if ( streq(p,ThimDlgSfx) )   return(MDialog);
            if ( streq(p,ThimBldrSfx) )  return(MApplixBuilder);
            if ( streq(p,ThimBldrTurboSfx) )  return(MApplixBuilderTurbo);
            if ( streq(p,ThimBldrDistSfx) )  return(MApplixBuilderDistribution);
            if ( streq(p,ThimBmapSfx) )  return(MBitmap);
            if ( streq(p,ThimSlideSfx) )  return(MApplixSlideShow);
#endif
            return(0);
        }
    }
    return(0);
}
 

