/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axiobase.c
   Applix ASC version: 10
   Date: Mon Nov 16 13:53:26 1998
*/
#include <stdio.h>
#include <aos.h>
#include <thimble.h>
#if AXMS
#include <fcntl.h>
#endif  /* DAMNT */

char *AxMBufGets AXP_((
					char *inBuf,		/* buffer from AxReadFileIntoTaskBuffer */
					int *bufPosP,		/* IN/OUT buffer position */
					int bufSize			/* total size of buf */
					));
char *AxReadFileIntoTaskBuf AXP_((
					int task,
					char *filename,
					int *sizeP
					));

char *AxMFgets AXP_((FILE *fp));
void AxMFreset();

extern char *TaskAlloc(), *TaskRealloc();
extern void AxMfreset();

static char *axMIOBuf;		/* malloc based IO buffer buffer */
static int axMIOBufSize;	/* current size */

/*************************************************************************
function: AxReadFileIntoBuf - Allocate a buffer in current task, read Unix file into it
Returns NULL on I/O error, errno set
*/
char *AxReadFileIntoBuf (filename,sizeP)
char *filename;
int *sizeP;
{
	return(AxReadFileIntoTaskBuf(THIMpid,filename,sizeP));
}

/*************************************************************************
function: AxReadFileIntoTaskBuf - Allocate a buffer in passed task, read Unix file into it
Returns NULL on I/O error, errno set
*/
char *AxReadFileIntoTaskBuf (task,pfilename,sizeP)
int task;
char *pfilename;
int *sizeP;
{
	char filename[STRPATH+1], *buf;
	int read_only, err, size;
	struct stat status;
	FILE *fp;

	/* stat the file */
	strcpy(filename, pfilename);
	AxSysDosUnixPath(filename);
	if ( -1 == Astat(filename, &status) )
		return(NULL);
	*sizeP = size = status.st_size;
	if ( size <= 0 )
		return(NULL);
	buf = (char *)TaskAlloc(task,size);
	if ( !buf )
		return(NULL);

	/* lock and open the file for reading */
	read_only = TRUE;
	err = AFTaskLockFile(filename, task, &read_only, FALSE);
	if ( err != AOK )
		return(NULL);
	fp = (FILE *)AFTaskfopen(task, filename, READ_BINARY_STR); /* r */
	if ( !fp )
	{
		AFTaskUnlockFile(filename, task, read_only);
		return(NULL);
	}

	/* read the file into the buffer */
	if ( fread(buf, size, 1, fp) != 1 )
	{
		TaskFree(task, buf);
		buf = (char *)0;
	}

	/* close and unlock the file */
	AFfclose(filename, read_only, fp);
	AFTaskUnlockFile(filename, task, read_only);
	return(buf);
}

/*************************************************************************
function: AxMBufGets - Sort of like AxMFgets, but buffer based
Returns line ptr, tosses \n character at end.
The returned ptr is malloc'd... The caller should NOT free.  This
function re-uses the buffer.
*/
char *AxMBufGets (inBuf,bufPosP,bufSize)
register char *inBuf;		/* buffer from AxReadFileIntoTaskBuffer */
register int *bufPosP;		/* IN/OUT buffer position */
register int bufSize;		/* total size of buf */
{
	register int len;
	register char *inP = inBuf + *bufPosP;
	register char *endP = inBuf + bufSize;

	inBuf += *bufPosP;

	if ( !axMIOBuf )
		axMIOBuf = TaskAlloc(0,axMIOBufSize = 1000);

	while ( inP < endP ) {
		if ( *inP == '\n' ) {
			len = inP - inBuf + 1;		/* including newline but not NULL */
			if ( len > axMIOBufSize ) {
				axMIOBufSize = len;			/* make buffer bigger */
				axMIOBuf = TaskRealloc(0,axMIOBuf,axMIOBufSize);
			}
			memcpy(axMIOBuf,inBuf,len-1);
			axMIOBuf[len-1] = 0;
			*bufPosP += len;
#if AXMS
            if ( axMIOBuf[len-2] == 0x0D )
               {
               /* this is a DOS created file, with a CR-NL pair on the end */
               /* step on the CR */
               axMIOBuf[len-2] = 0;
               }
#endif
			return(axMIOBuf);
		}
		inP++;
	}
	if ( inP == inBuf ) {
		errno = 0;
		AxMfreset();
		return(NULL);
	}
	else {
		len = inP - inBuf + 1;		/* including trailing NULL */
		if ( len > axMIOBufSize ) {
			axMIOBufSize = len;			/* make buffer bigger */
			axMIOBuf = TaskRealloc(0,axMIOBuf,axMIOBufSize);
		}
		memcpy(axMIOBuf,inBuf,len-1);
		axMIOBuf[len-1] = 0;
		*bufPosP += len-1;
		return(axMIOBuf);
	}
}

/*************************************************************************
function: AxMFgets - Read file line into buffer, return buffer ptr, tosses \n
The returned ptr is malloc'd... The caller should NOT free.  This
function re-uses the buffer.
*/
char *AxMFgets (fp)
FILE *fp;
{
	register int pos = 0;
	register int len;

	if ( !axMIOBuf )
		axMIOBuf = TaskAlloc(0,axMIOBufSize = 1000);

	while ( fgets(axMIOBuf+pos,axMIOBufSize-pos,fp) ) {
		len = strlen(axMIOBuf+pos);
		if ( len == axMIOBufSize-pos-1 ) {	/* filled buf */
			if ( axMIOBuf[pos+len-1] == '\n' ) {	/* exact hit on line length * */
				axMIOBuf[pos+len-1] = 0;			/* toss newline */
				return(axMIOBuf);					/* done */
			}
			axMIOBufSize = 2*axMIOBufSize;			/* make buffer bigger */
			axMIOBuf = TaskRealloc(0,axMIOBuf,axMIOBufSize);
			pos += len;								/* loop for more */
		}
		else {
			if ( axMIOBuf[pos+len-1] == '\n' )
				axMIOBuf[pos+len-1] = 0;	/* toss newline */
			return(axMIOBuf);				/* done */
		}
	}

	/* If got here, first read failed or last line was not newline termed */

	if ( pos != 0 )
		return(axMIOBuf);
	else {
		AxMfreset();
		return(NULL);
	}
}

/*************************************************************************
function: AxMfreset - frees memory buffer
Not critical that this be called, but a nice idea
*/
void AxMfreset ()
{
	free(axMIOBuf);
	axMIOBuf = NULL;
}

