/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axole.c
   Applix ASC version: 5
   Date: Thu Mar 11 21:14:18 1999
*/
#if AXMS
#include <windows.h>
#endif
#include <aos.h>
#include <adld.h>  /* shared lib stuff (Adlopen, etc) */


typedef int (*AXCLSID_FUNC)(char *, char *);

/****************************************************************************/
/* this code stolen from MS OLE implementation for Solaris */
#if !AXMS    /* { */
typedef struct tagGUID
{
    unsigned long     Data1;
    unsigned short    Data2;
    unsigned short    Data3;
    unsigned char     Data4[8];
} GUID;

typedef GUID CLSID;
typedef GUID IID;
#endif    /* } !AXMS */

/* The byte combination that identifies a file as a storage of some kind. */
const unsigned char SIGSTG[] = {0xd0,0xcf,0x11,0xe0,0xa1,0xb1,0x1a,0xe1};
#define CBSIGSTG sizeof(SIGSTG)

/* Identifier for first bytes of Beta 2 Docfiles. */
const unsigned char SIGSTG_B2[] = {0x0e,0x11,0xfc,0x0d,0xd0,0xcf,0x11,0xe0};
#define CBSIGSTG_B2 sizeof(SIGSTG_B2)

/* The first portion of a storage file. */
struct SStorageFile {
    unsigned char    abSig[CBSIGSTG];    /*  Signature */
    CLSID            _clid;              /*  Class Id */
};

#include <axole.p>

int
AxIsOleStorageFile(char *filename)
{
	int retval, nbytes;
	FILE *fp;
    struct SStorageFile stgfile;

	retval = 0;
	if ( filename != (char*)NULL )
	{
	    fp = fopen(filename, READ_BINARY_STR);
		if ( fp != (FILE*)NULL )
		{
			if ( !SkipMacHeader(fp) )
    	    	fseek(fp, 0, SEEK_SET);
        	nbytes = fread((void*)&stgfile, 1, sizeof(stgfile), fp);
            if ( nbytes == sizeof(stgfile) )
            {        
                if ( memcmp((void *)stgfile.abSig,SIGSTG,CBSIGSTG) == 0 ||
                     memcmp((void *)stgfile.abSig,SIGSTG_B2,CBSIGSTG_B2) == 0 )
	            {
        	        retval = 1;		/* file is an OLE storage */
	            }
            }
			fclose(fp);
	    }
	}
	return retval;
}
/****************************************************************************/


int
AxGetClassIdFromOleFile(char *filename, char *clsid_str)
{
	int retval;
	void *handle;
	AXCLSID_FUNC func;
#if AXMS
	char *lib_name = "axole.dll";
#elif HP800V3
	char *lib_name = "libaxole.sl";
#else
	char *lib_name = "libaxole.so";
#endif
	char *func_name = "axGetCLSID";

	retval = !AOK;

	handle = Adlopen(lib_name, ADL_BIND_IMMEDIATE | ADL_BIND_FIRST);
	if ( handle )
	{
#if RIOS
		func = (AXCLSID_FUNC)handle;
#else
		func = (AXCLSID_FUNC)Adlsym(handle, func_name, ADL_TYPE_PROCEDURE);
#endif
		if ( func )
		{
			retval = func(filename, clsid_str);
		}

		Adlclose(handle);
	}

	return retval;
}

