/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axshcmd.c
   Applix ASC version: 28
   Date: Tue Jan 19 21:10:46 1999
*/
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <setjmp.h>
#include <axsocket.h>
#include <errno.h>
#include <signal.h>
#if AXMS
#include <process.h>
#include <winmap.h>
#endif

#define THIM_STATIC 1    /** do not redefine such things as "exit()"  */
#include <aos.h>
#include <elfapi.h>
#include <axnet.h>

int received_sig_usr1 = 0;

#if !AXMS && !LINUX
extern char *sys_errlist[]; /* Unix errro strings */
#endif


#if PMAX  /* { */
char *strdup(str)
char *str;
{
  return(strcpy(malloc(strlen(str)+1),str));
}
#endif  /* } PMAX */


#include <axshcmd.p>



/***********************************************************************
function: GotAlarm - Signal catcher for alarm going off
*/
void GotAlarm ()
{
}

/*************************************************************************
function: AxRunProcess - Perform a shell command, return status
*/
static int AxRunProcess(user_id,argc,argv,cmdline,pended,for_axmain,console_process)
int user_id;        /* for setuid */
int argc;
char **argv;
char *cmdline;	/* if non-NULL and argc=0, use this rather than argc/argv */
int pended;
int for_axmain;
int console_process;/* NT/95: so we know whether to hide the console window. */
{
    int status;
    int shpid;
    int jpid;
    int alarm_count, alarm_set;

	if ( cmdline && !argc ) {
		status = axRunProcess(user_id,cmdline,pended,for_axmain,console_process);
		return(status);
	}
#if AX_X11          /*  {  */
    shpid = fork();
    if ( shpid == -1 )
        return(errno);
    if ( !shpid ) {
        if ( user_id != getuid() ) {
            if ( 0 != setuid(user_id) )
                _exit(errno);
        }
        execvp(argv[0],argv);
        _exit(errno);
    }
    if ( !pended )
        return(shpid);

    /* wait for exec'ed process to finish */
    if ( for_axmain && received_sig_usr1 )
        return(0);
    alarm_set = FALSE;
    alarm_count = 0;
    while ( TRUE ) {
        if ( for_axmain && received_sig_usr1 == FALSE ) {
            signal(SIGALRM, (void (*)())GotAlarm);
            alarm_set = TRUE;
            alarm_count++;
            alarm(1);
        }
        jpid = wait(&status);
        if ( alarm_set )
            alarm(0);
        if ( jpid == shpid ) /* child exited */
            break;
        if ( for_axmain == FALSE || received_sig_usr1 )
        {
            if ( jpid == -1 && errno == EINTR )
                return(0);
        }
        if ( alarm_count > 30 )
            return(ETIMEDOUT);
    }
    status = status & 0xffff; /*only 16 bits are defined*/

    /* If it is 0x7f then either the child stopped (whatever that means) or 
    the shcmd() routine failed and explicitly returned this value.  Don't do 
    any further interpretation. 
    If not 0x7f then, the child terminated due to a signal; treat this as a
    serious error. */

    if ( status&0xff ) {
        return(127);
    }

    /* The child exited and exit code is in upper byte. */

    return(status >> 8);

#else    /* }  {  */
    /* Windows/NT  */
    if ( pended )
        status = _spawnvp( _P_WAIT, argv[0], argv );
    else {
        /* status is actually the pid */
		if ( !cmdline )
#if 1
			abort();
#else
			tempAbort("AxRunProcess does not support argc/argv model yet!");
#endif
        status = axmsStartProcessNopend(cmdline, !console_process, TRUE);
	}
    return( status );
#endif   /*  }  */
}

/*************************************************************************
function: axRunProcess - Perform shell command, return status
Issue a shell command and return its status
Double quoted args are passed as a single string after
the quotes are removed to make sh -c "string" work
The caller is presumed to have set the SIGCHLD signal to SIG_IGN!
If this is not a pended call, return child's pid.
*/
static int axRunProcess(user_id,passed_cmd,pended,for_axmain,console_process)
int user_id;        /* for setuid */
char *passed_cmd;
int pended;
int for_axmain;
int console_process;/* NT/95: so we know whether to hide the console window. */
{
    register int cnt = 0;
    register char *p, *next;
    char *argv[1001];
    char cmd[10001];        /* plenty! */

    /* parse the command string */
    while ( *passed_cmd == ' ' ) passed_cmd++;  /* ignore leading blanks */

    memset(cmd, 0, 10001);
    if ( strlen(passed_cmd) > 10001 )
        strncpy(cmd,passed_cmd,10001);
    else
        strcpy(cmd, passed_cmd);

    for ( p = cmd, argv[cnt] = p; *p; p++ ) {
        if ( cnt == 1000 )                      /* nonsense case */
            break;
        if ( *p == ' ' ) {
            *p++ = 0;
            cnt++;
            while ( *p == ' ' ) p++;
            if ( !p )
                break;
            argv[cnt] = p;
        }
        else if ( *p == '\"' ) {
            next = strchr(p+1,'\"');
            if ( next )
                p = next;
        }
        else if ( *p == '\'' ) {
            next = strchr(p+1,'\'');
            if ( next )
                p = next;
        }
    }
    argv[++cnt] = NULL;
	return(AxRunProcess(user_id,cnt,argv,passed_cmd,pended,for_axmain,console_process));
}

/*************************************************************************
function: AxCallProcess - Pend for a process, return its exit code
*/
int AxCallProcess (cmd)
char *cmd;
{
#if AX_X11
    return(axRunProcess(getuid(),cmd,TRUE,FALSE,FALSE));
#else
    /* Windows/NT, no numreic uxer-id, hope we don't ever need it */
    return(axRunProcess(0,cmd,TRUE,FALSE,FALSE));
#endif
}

/*************************************************************************
function: AxCallUidProcess - Pend for a process, return its exit code
*/
int AxCallUidProcess (user_id,cmd)
int user_id;        /* as in setuid() */
char *cmd;
{
    return(axRunProcess(user_id,cmd,TRUE,FALSE,FALSE));
}

/*************************************************************************
function: AxStartProcess - Start a process, return its process id
*/
int AxStartProcess (cmd, console_process)
char *cmd;
int console_process;
{
#if AX_X11
    return(axRunProcess(getuid(),cmd,FALSE,FALSE,FALSE));
#else
    /* Windows/NT, no numreic uxer-id, hope we don't ever need it */
    return(axRunProcess(0,cmd,FALSE,FALSE,console_process));
#endif

}

/*************************************************************************
function: AxCallAxmain - Pend for an axmain process, return its exit code
*/
int AxCallAxmain (cmd)
char *cmd;
{
#if AX_X11
    return(axRunProcess(getuid(),cmd,TRUE,TRUE,FALSE));
#else
    /* Windows/NT, no numreic uxer-id, hope we don't ever need it */
    return(axRunProcess( 0,cmd,TRUE,TRUE,FALSE));
#endif
}

char *ErrnoTextFromUnix();

/*************************************************************************
function: AxElfCallProgram - Start program with Elf data, get Elf data result
If there is an error, return code/str/obj set non-zero;
*/
elfData AxElfCallProgram(progName,funcData,code,str,obj)
char *progName;
elfData funcData;
int *code;
char **str;
char **obj;
{
    int channel;
    elfData response;
    char sockname[50];
    char cmd[2000];
    int errval;

    *code = 0;

/* Create a named-pipe command channel & start child process.
   The AxCallProcess will return programs exit code only if the program exits
   before calling AxOpenServer().  If the called program succeeds in
   calling AxOpenServer, the AxCallProcess returns zero. */

    strcpy(sockname,"/tmp/cpXXXXXX");
    mktemp(sockname);
    sprintf(cmd,"%s %s -slave -fork",progName,sockname);
    errval = AxCallProcess(cmd);
    if ( errval != 0 ) {
        *code = errval;
        *str = ErrnoTextFromUnix(errval);
        *obj = strdup(progName);
        return(NULL);
    }

/* Connect to child process */

    channel = AxClientConnect(0,sockname,NULL,code,str,obj);
    if ( channel == -1 )
        return(NULL);


/* Send command line to child process */

    AxSendMessageToServer(channel,1,funcData,code,str,obj);
    if ( *code != 0 ) { /* child crashed while waiting for cmdline */
        AxClientDisConnect(channel);
        return(NULL);
    }

/* get return data - child calling AxError yields error code here */

    response = AxGetResponseFromServer(channel,code,str,obj);
    if ( *code != 0 ) { /* child called AxError */
        AxClientDisConnect(channel);
        return(NULL);
    }

/* Close connection */
    
    AxClientDisConnect(channel);

    return(response);
}

/*************************************************************************
function: AxCallAxMain - Call Elf code in axmain process
Not for publication to customers!  Uses undocument axnet$(16).
*/
elfData AxCallAxMain (sockname, funcname, funcData, code,str,obj)
char *sockname;     /* as per axnet$(16) ... AxSocketName */
char *funcname;
elfData funcData;
int *code;
char **str;
char **obj;
{
    int channel;
    elfData callArray, response;

    *code = 0;
    channel = AxClientConnect(0,sockname,NULL,code,str,obj);
    if ( channel == -1 )
        return(NULL);
    callArray = AxMakeArray(2);
    callArray = AxAddStrToArray(callArray,0,funcname);
    callArray = AxAddDataToArray(callArray,1,AxCopyData(funcData));
    AxSendMessageToServer(channel, AXMAIN_CALL_MACRO_GET_RESULT, callArray, code,str,obj);
    if ( *code != 0 ) {
        AxClientDisConnect(channel);
        return(NULL);
    }
    response = AxGetResponseFromServer(channel, code, str, obj);
    AxClientDisConnect(channel);
    return(response);
}

