/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axstack.c
   Applix ASC version: 21
   Date: Fri May  7 21:11:02 1999
*/
#include <stdio.h>
#if SUN4 || (LINUX && __sparc__)
#include <alloca.h>	
#endif
#include <aos.h>
#include <setjmp.h>
#include <paripc.h>
#include <scac.h>
#include <thimble.h>
#include <axstack.p>

/*
Stack swapper for THIMBLE tasks.
The stack itself is swapped to the depth of the current suspending action.
*/

#if HP800
#define GrowsDown 0             /* ?? should be dynamic */
#else
#define GrowsDown 1		/* ?? should be dynamic */
#endif
#define MAX_STACK_DIFF (75000 * sizeof(char *))
#define INC_STACK_DIFF (75000 * sizeof(char *))
#define EXCESS_STACK (125 * sizeof(char *)) /* 100 stack (regs backup ) beyond vars to keep */
#define MARGIN_STACK 1024

#define SLOG 0
#if SLOG
#define SLOGGER(body) printf body ; printf("\n"); fflush(stdout)
#else
#define SLOGGER(body)
#endif

#ifdef sparc
#define FLUSH_REGISTER_WINDOWS asm(NXLT("ta 3"))
#else
#define FLUSH_REGISTER_WINDOWS 
#endif

int AxInTask;	/* 0 or task the owns the current stack */

/*
MAX_STACK_DIFF is the max difference between the stack size of the
largest suspended task and the amount of stack the kernel uses at activation
time.  It is essentially the max stack depth of a task when it calls
IPCawait (e.i. SuspendTask() ).  The suspend and activate code have an unused
char buffer of this size to make sure that async events like signals
cannot damage the stack while it is being re-written - before the longjump.
MAX_STACK_DIFF has been replaced with INC_STACK_DIFF and the stacker
is resursive as needed.
*/


/*************************************************************************
function: StartTask - Save initial task state and call task's entry func
This is called by the kernel after the task's BSS and DATA have been set up.
This actually starts the task.
Note that when tasks are started, THIMpid & AxInTask are the same.
*/
#if !ALPHA && !(LINUX && __alpha__)
StartTask (id,func,argc,argv,usesElfArgs,elfArgv)
int id;
int (*func)();
int argc;
char **argv;
int usesElfArgs;
data_ptr elfArgv;
{
	int junk;
	register int size;
	data_ptr elfaP=NULL;

	TaskInfo[THIMpid].start_me = FALSE;
	if ( GrowsDown )
		size = StackBase - (char*)&junk + EXCESS_STACK;
	else
		size = (char*)&junk - StackBase + EXCESS_STACK;
	TaskInfo[THIMpid].stack_size = size;
	if ( NULL == (TaskInfo[THIMpid].stack = (char *)malloc(size)) ) {
		SystemMessage(XLT("!Not enough memory to start task"));
		THIMexit(127);
	}
	if ( GrowsDown )
		memcpy(TaskInfo[THIMpid].stack,&junk-EXCESS_STACK,size);
	else
		memcpy(TaskInfo[THIMpid].stack,&junk+EXCESS_STACK,size);
	SLOGGER((NXLT("Start %s (%d) sz %d stk %lx"),argv[0],THIMpid,size,TaskInfo[THIMpid].stack));
	AxInTask = THIMpid;
	TaskInfo[AxInTask].activationData = THIMpid;
	if(usesElfArgs && elfArgv != NULL) { 
		/* copy Elf args to new task's memory pool */
		elfaP = elfArgv;
		elfArgv = AxTaskCopyElfData(THIMpid,elfaP);
		AxTaskFreeElfData(0,elfaP);
	}
	AxRunTask();
	(*func)(argc,argv,usesElfArgs,elfArgv);
}
#else
StartTask (id,func,argc,argv,usesElfArgs,elfArgv)
int id;
int (*func)();
int argc;
char **argv;
int usesElfArgs;
data_ptr elfArgv;
{
	int junk;
	register int size;
	data_ptr elfaP=NULL;
	register double *src, *dst;

	TaskInfo[THIMpid].start_me = FALSE;
	size = StackBase - (char*)&junk + EXCESS_STACK;
	TaskInfo[THIMpid].stack_size = size;
	if ( NULL == (TaskInfo[THIMpid].stack = (char *)malloc(size)) ) {
		SystemMessage(XLT("!Not enough memory to start task"));
		THIMexit(127);
	}
	src = (double*)&junk-EXCESS_STACK;
        dst = (double*)TaskInfo[THIMpid].stack;
	size = size >> 3;
	while ( --size >= 0 ) *dst++ = *src--;
	SLOGGER((NXLT("Start %s (%d) sz %d stk %lx"),argv[0],THIMpid,size,TaskInfo[THIMpid].stack));
	AxInTask = THIMpid;
	TaskInfo[AxInTask].activationData = THIMpid;
	if(usesElfArgs && elfArgv != NULL) { 
		/* copy Elf args to new task's memory pool */
		elfaP = elfArgv;
		elfArgv = AxTaskCopyElfData(THIMpid,elfaP);
		AxTaskFreeElfData(0,elfaP);
	}
	AxRunTask();
	(*func)(argc,argv,usesElfArgs,elfArgv);
}
#endif

/*************************************************************************
function: SuspendTask - Save task's stack and branch back to kernel
This is called by a task to become suspended.  Control returns to the
setjmp in the kernel.  When the task becomes activated again via ActivateTask,
the control is passed to the setjmp in this routine.
*/
SuspendTask (xxx)
int xxx;
{
	int junk[1];
	register int size;
	int *p;

	SLOGGER((NXLT("Suspend %s task %d"),TaskInfo[AxInTask].argv[0],AxInTask));
#if sparc
	/* We must call a function which will recursively call itself a number */
    /*   of times so that the register window gets rotated enough to put a */
    /*  real stack frame on the stack. (See a SPARC processor architecture */
    /*  book for a description of register windows) - jon                  */
	FLUSH_REGISTER_WINDOWS;
	rotate_register_wins(1,0,0);
#endif
	TaskInfo[AxInTask].activationData = THIMpid;	/* usually same as AxInTask */
	MemParanoid();
#if ALPHA || (LINUX && __alpha__)
	size = StackBase - (char*)junk + EXCESS_STACK;
	if ( _setjmp(TaskInfo[AxInTask].jump) ) {
#else
	if ( axsetjmp(TaskInfo[AxInTask].jump) ) {
#endif
		SLOGGER((NXLT("Restart %s task %d"),TaskInfo[AxInTask].argv[0],AxInTask));
		AxRunTask();
		return;
	}

#if !ALPHA && !(LINUX && __alpha__)
	if ( GrowsDown )
		size = StackBase - (char*)junk + EXCESS_STACK;
	else
		size = (char*)junk - StackBase + EXCESS_STACK;
#endif
	if ( size > TaskInfo[AxInTask].stack_size ) {
		free(TaskInfo[AxInTask].stack);
		TaskInfo[AxInTask].stack = (char *)malloc(size);
	}
	TaskInfo[AxInTask].stack_size = size;
	oldstack((long*)TaskInfo[AxInTask].stack,TaskInfo[AxInTask].stack_size,0);/* saves & branches */
	nop_func(junk);	/* to make sure it gets allocated by compiler */
	nop_func(junk);	/* to make sure it gets allocated by compiler */
}
/* The save & branch is split out becuase we dont' know how the
compiler will order junk versus lots_of_stack */
static oldstack(dst,cnt, so_far)
#if !ALPHA && !(LINUX && __alpha__)
register long *dst;
#else
register double *dst;
#endif 
register int cnt;
register int so_far;
{
	char lots_of_stack[INC_STACK_DIFF];

	DoOldStack(dst,cnt);
	nop_func(lots_of_stack);/* to make sure it gets allocated by compiler */
	nop_func(lots_of_stack);/* to make sure it gets allocated by compiler */
}

#if !ALPHA && !(LINUX && __alpha__)
DoOldStack(dst,cnt)
register long *dst;
register int cnt;
{
	register long *src;

	AxInTask = 0;
	src = (long*)StackBase;
	cnt = cnt >> 2;
	if ( GrowsDown )
		while ( --cnt >= 0 ) *dst++ = *src--;
	else
		while ( --cnt >= 0 ) *dst++ = *src++;
	axlongjmp(KernelJump,1);
}
#else
DoOldStack(dst,cnt)
register double *dst;
register int cnt;
{
	register double *src;

	AxInTask = 0;
	src = (double*)StackBase;
	cnt = cnt >> 3;
	while ( --cnt >= 0 ) *dst++ = *src--;
	_longjmp(KernelJump,1);
}
#endif

/*************************************************************************
function: ActivateTask - Load task's stack and branch to suspension point
This is called by the kernel to re-activate a suspended task.  The control
is passed to the SuspendTask setjmp.
*/
#if VAX
static ax_jmp_buf todo;       /* cannot be passed cause of stack mash */
#else
static jmp_buf todo;	/* cannot be passed cause of stack mash */
#endif
ActivateTask (id)
int id;
{
	DoActivateTask(id, 0, TaskInfo[id].stack_size);
}

#if SUN4 || (LINUX && (__alpha__ || __sparc__))	/* { sun and solaris */
DoActivateTask(id, so_far, cnt)
	int id, so_far;
{
	char *lots_of_stack;
	/* push the stack pointer by atlease cnt bytes */
	lots_of_stack = (char *)alloca(cnt + MARGIN_STACK);
	lots_of_stack[0] = '\0';
	VfyContext(TaskInfo[id].activationData);
	DoReallyActivateTask(id);
}
#else	/* }{ */
DoActivateTask(id, so_far, cnt)
	int id, so_far;
{
	char lots_of_stack[INC_STACK_DIFF];
	if ( so_far+INC_STACK_DIFF < cnt )
		DoActivateTask(id, so_far+INC_STACK_DIFF, cnt);
	else	{
		VfyContext(TaskInfo[id].activationData);
		DoReallyActivateTask(id);
		nop_func(lots_of_stack);/* to make sure it gets allocated by compiler */
		nop_func(lots_of_stack);/* to make sure it gets allocated by compiler */
	}
}
#endif	/* } */



#if !ALPHA && !(LINUX && __alpha__)
DoReallyActivateTask(id)
int id;
{
	register long *src;
	register long *dst;
	register int cnt;

	AxInTask = id;
	SLOGGER((NXLT("Activate %s task %d"),TaskInfo[id].argv[0],id));
#if VAX
	memcpy(todo,TaskInfo[id].jump,sizeof(ax_jmp_buf));
#else
	memcpy(todo,TaskInfo[id].jump,sizeof(jmp_buf));
#endif
	src = (long*)TaskInfo[id].stack;
	cnt = TaskInfo[id].stack_size >> 2;
	dst = (long*)StackBase;
	if ( GrowsDown )
		while ( --cnt >= 0 ) *dst-- = *src++;
	else
		while ( --cnt >= 0 ) *dst++ = *src++;
#if VAX
	axlongjmp(todo,1);
#else
	longjmp(todo,1);
#endif
}
#else
DoReallyActivateTask(id)
int id;
{
	register double *src;
	register double *dst;
	register int cnt;

	AxInTask = id;
	SLOGGER((NXLT("Activate task %d"),id));
	memcpy(todo,TaskInfo[id].jump,sizeof(jmp_buf));
	src = (double*)TaskInfo[id].stack;
	cnt = TaskInfo[id].stack_size >> 3;
	dst = (double*)StackBase;
	while ( --cnt >= 0 ) *dst-- = *src++;
	_longjmp(todo,1);
}
#endif

static nop_func(void *ignored){}

rotate_register_wins(a, b, c )
int a, b, c;
{
  a = a + 1;
  if ( a == 10 )
     return;
  else
     rotate_register_wins( a, b, c );

return;
}


