/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: lockrsrc.c
   Applix ASC version: 10
   Date: Mon Nov 16 13:53:26 1998
*/
#include <stdio.h>
#include <aos.h>
#include <thimble.h>
#include <paripc.h>
#include <scac.h>

typedef struct lock_info {
	struct lock_info *next;
	char *name;			/* unique name of this resource */
	int locked_by;		/* locking task id */
	int num_waiters;	/* how many tasks waiting on this resource */
	int waiters[100];	/* plenty!!! */
} lock_info;

#include <lockrsrc.p>

static lock_info *Locked;	/* Chain */

/*************************************************************************
function: AxLockResource - Keep other well behaved tasks away
If the resource is already locked, block waiting for a lock.
*/
void AxLockResource (name,taskid)
register mbchar *name;	/* name of resource... better be unique! */
register int taskid;	/* this task's uid */
{
	register lock_info *p;
	register int ix;
	long senderType, senderTask, size, msgType;
	int code;
	char *msgP;
	int pid;

	for ( p = Locked; p; p = p->next ) {
		if ( streq(p->name, name) )
			break;
	}

	if ( !p ) {
		/* create the lock */
		p = (lock_info*)malloc(sizeof(lock_info));
		p->name = strcpy(malloc(strlen(name)+1),name);
		p->next = Locked;
		p->locked_by = taskid;
		p->num_waiters = 0;
		Locked = p;
		return;
	}
	if ( p->locked_by == taskid )	/* already have lock, just return */
		return;

    /* Make sure (for fail-safe reasons) that the task which owns     */
    /* the resource still exists                                              */
    pid = ThimPidFromUID( p->locked_by );
    if ( pid == 0 )
       {
       /* task which locked the resource if gone!  Have seen this happen  */
       /* running scripts, but nevery figured out how. - jon              */
       /* pretend we get the resource                                     */
       p->locked_by = taskid;
       return;
       }

    /* If got here, resource locked.  Append self as waiter, wait for notify. */
	p->waiters[p->num_waiters] = taskid;
	if ( p->num_waiters == 100 )	/* will never happen */
		tempAbort(NXLT("Too many tasks waiting on resource"));
	p->num_waiters++;

msg_getter:

	TaskInfo[AxInTask].waiting_on_socket = TRUE;/* prevents RPC interleave */
	TaskInfo[THIMpid].waiting_on_socket = TRUE;/* prevents RPC interleave */
	code = AxGetMessage(IPC_INF,&msgP,&size,&senderType,&senderTask,&msgType);
	TaskInfo[AxInTask].waiting_on_socket = FALSE;
	TaskInfo[THIMpid].waiting_on_socket = FALSE;

	if ( senderType != APREGIS ) {
		AxDeferMessage();
		goto msg_getter;
	}
	AxMessageAbsorbed();
	AxLoadDeferedMessages();
}

/*************************************************************************
function: AxUnlockResource - Let another task use this resource
Also called at task termination time unlock all resources used by this task.
*/
void AxUnlockResource (name,taskid)
register mbchar *name;	/* name of resource of NULL for all */
register int taskid;	/* my uid */
{
	register lock_info *p, *prev;
	register int ix;
    int pid;

/* If task term case, toss this task from all waiter lists */

	if ( !name ) {
		for ( p = Locked; p; p = p->next ) {
			/* clip this task from waiters list */
			for ( ix = 0; ix < p->num_waiters; ix++ ) {
				if ( taskid == p->waiters[ix] ) {
					p->num_waiters--;		/* I am no longer waiting */
					for ( ; ix < p->num_waiters; ix++ )	/* shift others foward */
						p->waiters[ix] = p->waiters[ix+1];
					break;
				}
			}
		}
	}

/* Unlock the resource (On task term, unlock all resources) */

reloop:
	prev = NULL;
	for ( p = Locked; p; prev = p, p = p->next ) {
		if ( (p->locked_by == taskid) && (!name || streq(p->name, name)) ) {
			if ( p->num_waiters == 0 ) {
				free(p->name);
				if ( prev )
					prev->next = p->next;
				else
					Locked = p->next;
				free(p);
				goto reloop;		/* prev may now be corrupt */
			}
			else {
				/* clip waiter from list, wake it up */
				p->num_waiters--;
				p->locked_by = p->waiters[0];
				for ( ix = 0; ix < p->num_waiters; ix++ )
					p->waiters[ix] = p->waiters[ix+1];
                pid = ThimPidFromUID( p->locked_by );
                if ( pid == 0 )
                    {
                    /* locking task has, for some reason, managed to go away */
                    /* and not free this resource, look for other waiters    */
                    goto reloop;
                    }
				IPCXsendMes(IPC_SRT,ThimPidFromUID(p->locked_by),"",1,APREGIS,0);
			}
		}
	}
}

/*************************************************************************
function: AxWaitOnResource - Allows taskX to wait on a semaphore anybody can set
*/
void AxWaitOnResource (char *name, int keepLocked, int taskid)
{
	AxLockResource(name,taskid);	/* make sure we can get this resource */
	AxUnlockResource(name,taskid);
	AxLockResource(name,0);
	AxLockResource(name,taskid);	/* wait on this call till AxNotifyResource() */
	if ( !keepLocked )
		AxUnlockResource(name,taskid);
}

/*************************************************************************
function: AxNotifyResource - Allows taskX to wait on a semaphore anybody can set
*/
void AxNotifyResource (char *name)
{
	while ( !AxIsResourceLocked(name) )
		AxSleepMilliseconds(100);
	AxUnlockResource(name,0);
}

/* !!0 BARRY:  This is for the temp fix for the Elf call -> WP during pended
 * procs.  Delete once we've resolved that issue */

/***************************************************************************
function: AxIsResourceLocked 
*/
int AxIsResourceLocked (name)
mbchar *name;
{
	register lock_info *p;

	for ( p = Locked; p; p = p->next ) {
		if ( streq(p->name, name) )
			return(TRUE);
	}

	return(FALSE);
}

/*************************************************************************
function: ElfResourceStatus - Return array for resources$ function
*/
data_ptr ElfResourceStatus ()
{
	data_ptr aP, sP, wP;
	int numap = 0, ix;
	register lock_info *p;

	aP = AxMakeArray(0);
	for ( p = Locked; p; p = p->next ) {
		sP = AxMakeArray(0);
		sP = AxAddStrToArray(sP,0,p->name);
		sP = AxAddIntToArray(sP,1,p->locked_by);
		sP = AxAddIntToArray(sP,2,p->num_waiters);
		if ( p->num_waiters ) {
			wP = AxMakeArray(p->num_waiters);
			for ( ix = 0; ix < p->num_waiters; ix++ )
				wP = AxAddIntToArray(wP,ix,p->waiters[ix]);
			sP = AxWriteArray(sP,3,wP);
		}
		aP = AxWriteArray(aP,numap++,sP);
	}
	return(aP);
}

