/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: profile.c
   Applix ASC version: 67
   Date: Mon Nov 16 13:53:22 1998
*/
#include <aos.h>
#include <afile.h>
#include <getstr5.h>
#include <getstr7.h>
#include <getstr10.h>
#include <thimble.h>
#include <elf.h>
#include <elapi.h>
#include <aprofpok.h>
#include <scac.h>
#include <fsrts.h>
#include <java_syms.h>

#define MAX_KS_LINE 256	/* max char in a KS file line */
#define KS_SEP_CHAR ':'	/* entry separator in KS files */
#define KS_COMMENT_CHAR ':'	/* same as sep -> no name means no value */
#define KS_LOCK_CHAR '*'    /* used to lock profile values */

/* NOTE: 'STREQ' is case insensitive, while 'streq' is case sensitive. */
#define STREQ(a,b) (Astrcmp(a,b) == 0)

typedef struct ks_file {
	struct ks_file *next;	/* next in thread */
	ks_entry *entries;	/* head of ks entry pairs */
	int changed;		/* TRUE if stuff added/removed */
	char name[STRPATH+1];	/* filename of the KS file */
} ks_file;

#include <profile.p>	/* must be after typedefs */

extern char *GetFromKS(), *NeedFromKS(), *FSKSLookup();
extern char *ThimSfxFromType(), *DocPath();
extern char *xGetFromKS(), *ErrnoTextFromUnix();
extern typeErr IsValidPathFilename(), KSAdd(), ErrnoErr();
extern typeErr ThimDcInfoFromPath() , FSThimGtDcInfo(), FSRegisterDoc();
extern typeErr getStr();
extern char *AxPoolCreate();
extern char *AxGetStr();
extern int AxDontSaveProfiles;

char *Vmalloc(), *Vcalloc(), *Vfree(), *Vrealloc(), *strcpy();
char *UPneedByCode();
char *UPbyName();
typeErr UPIsLocked();
char *AGetHomeDir();
char *AGetAxHomeDir();
char *GetProfileByCode();

extern char *AxLangSuffix[];
extern char *AxStrflName;
extern char *ThimUserBackupDir;
extern char *DirSlash;
extern char *ThimSessionElfDir;

static char *ProfileName = NXLT("ax_prof4");
static char *OldProfileName = NXLT("ax_prof3");
static char *VeryOldProfileName = NXLT(".ax_profile");

char *ThimProfileFileName;	/* global prof - visible to outside world */

typeErr UPerrcode;	/* generic error code (like errno)... NO CONSUMERS! */

static ks_file ProfileData;		/* global Personal profile file info */
static ks_file LProfileData;	/* local Personal profile file info */
static ks_file OrigProfileData;	/* original global profile file info */

char *KsThreadPool;

static char *JavaClientProfiles[] = {
	"wpShowModifiedStatus",
	NULL
};

/*************************************************************************
function: AllcStr - Malloc a str based on arg  TASK 0!
*/
char *AllcStr (p)
register char *p;
{
	return(strcpy((char *)TaskAlloc(0,strlen(p)+1),p));
}

/*************************************************************************
function: AxProfileInterest - Declare interest in profile variables
*/
AxProfileInterest (codes)
typeErr *codes;
{
	int cnt, ix, size;
	char *p;

	for ( cnt=0; codes[cnt]; cnt++ );

	size = (cnt+1) * sizeof(char *);
	TaskInfo[THIMpid].profile_interest = (char**)TaskAlloc(THIMpid, size);
	for ( ix=0; ix<cnt; ix++ )
	{
		p = AxGetStr(codes[ix]);
		TaskInfo[THIMpid].profile_interest[ix] =
		 strcpy((char *)TaskAlloc(THIMpid, strlen(p)+1), p);
	}
	TaskInfo[THIMpid].profile_interest[cnt] = NULL;
}

/*************************************************************************
function: AxNotifyProfilers - Notify listeners of a profile change
*/
AxNotifyProfilers (str)
char *str;
{
	int ix, jx;
	char *p;
	profile_poke_msg_t msg;
	data_ptr infoP;

	if ( !str || !str[0] )
		return;
	for ( ix = 0; ix < MAX_PIDS; ix++ ) {
		if ( ix == THIMpid )	/* never notify self! */
			continue;
		if ( !TaskInfo[ix].in_use )
			continue;
		if ( !TaskInfo[ix].profile_interest )
			continue;
		for ( jx = 0; p = TaskInfo[ix].profile_interest[jx]; jx++ ) {
			if ( streq(p,str) ) {
				if ( ix == 0 )
					AxSyncAgainstProfile(str);
				else {
					msg.msgtype = 0;
					strncpy(msg.name,p,STRLONG);
					msg.name[STRLONG] = 0;
					IPCXsendMes(IPC_SRT,ix,&msg,sizeof(msg),APPROFM,0);
				}
			}
		}
	}
	for ( ix = 0; p = JavaClientProfiles[ix]; ix++ ) {
		if ( streq(p,str) ) {
			infoP = AxMakeArray(2);
			AxAddStrToArray(infoP,0,str);
			AxAddStrToArray(infoP,1,UPbyName(str));
			AxJavaCoreNotifier(JV_PROFILE_CHANGE,infoP);
		}
	}
}

/*************************************************************************
function: AxSyncAgainstProfile - Sync global vars against profile settings
*/
static void AxSyncAgainstProfile (char *str)
{
	extern int AxMemoryWarning;	/* in bytes before warning box */
	extern int AxMemoryLimit;		/* in bytes before tearing task down */

	free(ThimUserElfDir);
	free(ThimUserModelsDir);
	free(ThimUserBackupDir);

	ThimUserElfDir = AllcStr(UPneedByCode(PROF_CMDS_DIR_));
	ThimUserModelsDir = AllcStr(UPneedByCode(PROF_WP_TEMPLATES_DIR));
	ThimUserBackupDir = AllcStr(UPneedByCode(PROF_BACKUP_DIR_));

#if JavaMode
	AxMemoryWarning = 900000000;
	AxMemoryLimit = 900000000;
#else
	AxMemoryWarning = atoi(UPneedByCode(PROF_AX_MEMWARN));
	AxMemoryLimit = atoi(UPneedByCode(PROF_AX_MEMLIMIT));
	if ( !AxMemoryWarning )
		AxMemoryWarning = 15000000;
	else
		AxMemoryWarning *= 1000000;
	if ( !AxMemoryLimit )
		AxMemoryLimit = 100000000;
	else
		AxMemoryLimit *= 1000000;
#endif
}


/*************************************************************************
function: ResetUPdata - Hacky way to re-read profile
This should only be used by the out0of-rev logic because it doesn't
clear the old profile.
*/
ResetUPdata ()
{
	ProfileData.entries = NULL;
}

/*************************************************************************
function: FindThimbleFile - Search some dirs for a file
Returns TRUE if found the file.
*/
FindThimbleFile (mode,name,pathname,dirP)
IN int mode;	/* where to start/end search */
IN char *name;	/* entryname of file */
OUT char *pathname;	/* returned full pathname */
OUT int *dirP;		/* tells what kind of dir found in */
{
	struct stat status;
	extern char ElfDir[],
				*AxLocalLangDir,
				*AxLangDemosDir;

	switch ( mode ) {
	case THIM_FIND_ELF:			/* ELF docs */
		*dirP = THIM_FIND_ELF;
		if (ThimSessionElfDir) {
			sprintf(pathname, "%s%s%s", ThimSessionElfDir, DirSlash, name);
			AxSysDosUnixPath(pathname);
			if (0 == stat(pathname,&status)) return(TRUE);
		}
		sprintf(pathname,"%s%s%s",ThimUserElfDir,DirSlash,name);
		AxSysDosUnixPath(pathname);
		if ( 0 == stat(pathname,&status) ) return(TRUE);
		*dirP = THIM_FIND_CUSTOM;
		sprintf(pathname,"%s%self%s%s",ThimCustomDir,DirSlash,DirSlash,name);
		AxSysDosUnixPath(pathname);
		if ( 0 == stat(pathname,&status) ) return(TRUE);
		*dirP = THIM_FIND_AXLOCALLANG;
		sprintf(pathname,"%s%s%s",AxLocalLangDir,DirSlash,name);
		AxSysDosUnixPath(pathname);
		if ( 0 == stat(pathname,&status) ) return(TRUE);
		sprintf(pathname,"%s%s%s",AxLangDemosDir,DirSlash,name);
		AxSysDosUnixPath(pathname);
		if ( 0 == stat(pathname,&status) ) return(TRUE);
		*dirP = 0;
		strcpy(pathname,name);
		AxSysDosUnixPath(pathname);
		if ( 0 == stat(pathname,&status) ) return(TRUE);
		return(FALSE);

	case THIM_FIND_MODEL:	/* Find style guide in user's models+ */
		*dirP = mode;
		sprintf(pathname,"%s%s%s",ThimUserModelsDir,DirSlash,name);
		AxSysDosUnixPath(pathname);
		if ( 0 == stat(pathname,&status) ) return(TRUE);
		*dirP = THIM_FIND_CUSTOM;
		sprintf(pathname,"%s%smodels%s%s",ThimCustomDir,DirSlash,DirSlash,name);
		/* sprintf(pathname,"%s%Cmodels%C%s",ThimCustomDir,DirSlash,DirSlash,name);
		 * even %c looks wrong... SJP
		 */
		AxSysDosUnixPath(pathname);
		if ( 0 == stat(pathname,&status) ) return(TRUE);
		*dirP = 0;
		strcpy(pathname,name);
		AxSysDosUnixPath(pathname);
		if ( 0 == stat(pathname,&status) ) return(TRUE);
		return(FALSE);

	case THIM_FIND_USER:
		return(FALSE);

	case THIM_FIND_CUSTOM:
		*dirP = THIM_FIND_CUSTOM;
		sprintf(pathname,"%s%s%s",ThimCustomDir,DirSlash,name);
		AxSysDosUnixPath(pathname);
		if ( 0 == stat(pathname,&status) ) return(TRUE);
		*dirP = THIM_FIND_AXLOCALLANG;
		sprintf(pathname,"%s%s%s",AxLocalLangDir,DirSlash,name);
		AxSysDosUnixPath(pathname);
		if ( 0 == stat(pathname,&status) ) return(TRUE);
		return(FALSE);

	case THIM_FIND_USERPLUS:
	case THIM_FIND_CUSTOMPLUS:
		*dirP = THIM_FIND_CUSTOM;
        sprintf(pathname,"%s%s%s",ThimCustomDir,DirSlash,name);
		AxSysDosUnixPath(pathname);
		if ( 0 == stat(pathname,&status) ) return(TRUE);
		*dirP = THIM_FIND_AXLOCALLANG;
		sprintf(pathname,"%s%s%s",AxLocalLangDir,DirSlash,name);
		AxSysDosUnixPath(pathname);
		if ( 0 == stat(pathname,&status) ) return(TRUE);
	case THIM_FIND_APPLIX:
		*dirP = THIM_FIND_APPLIX;
        sprintf(pathname,"%s%s%s",ThimApplixDir,DirSlash,name);
		AxSysDosUnixPath(pathname);
		if ( 0 == stat(pathname,&status) ) return(TRUE);
	case THIM_FIND_AXLANG:
		*dirP = THIM_FIND_AXLANG;
        sprintf(pathname,"%s%s%s",AxLangDir,DirSlash,name);
		AxSysDosUnixPath(pathname);
		if ( 0 == stat(pathname,&status) ) return(TRUE);
	case THIM_FIND_AXLOCALLANG:
		*dirP = THIM_FIND_AXLOCALLANG;
        sprintf(pathname,"%s%s%s",AxLocalLangDir,DirSlash,name);
		AxSysDosUnixPath(pathname);
		if ( 0 == stat(pathname,&status) ) return(TRUE);
	default:
		return(FALSE);
	}
}

/*************************************************************************
function: AxReadProfile
read the user's profile file.  Always reads the 'global' version; then
reads the 'local' version, first looking for one from the current rev,
but will check backwards for a file from one of the previous two revs.
Currently the earlier rev profile filenames are hard coded;  when we
start to follow a consistant naming convention, this function should
use a template to determine the names of earlier rev profile files.
*/
char *
AxReadProfile(new_prof, new_home)
int *new_home, *new_prof;
{
	char newfile[STRPATH+1], buf[STRPATH+200], *home, *str;
	char multiLangProfile[STRPATH+1], *lang_str;
	register ks_entry *p, *q, *next;
	register char *valp;
	int	lang = 0;
	int found_file;
	int local_override;
	typeErr err;

	*new_home = FALSE;
	*new_prof = FALSE;

	if ( !ProfileData.entries )
	{
		/*	Are we in a multi-lang environment? */

		lang_str = getenv(NXLT("AXLANG"));
		if ( lang_str )
			lang = atoi(lang_str);
		else
			lang = 0;

		/*
		 * First, find the global version and read it;  check
		 * the environment variable AXPROFILE for a hint as to
		 * it's location...
		 */

		str = getenv(NXLT("AXPROFILE"));
		if ( str && str[0] )
		{
			home = (char *)0;
			strcpy(ProfileData.name, str);
			if ( -1 == Aaccess(ProfileData.name,R_OK) )
			{
				sprintf(buf, NXLT("Could not access profile: %s: %s"),
		          ErrnoTextFromUnix(errno), ProfileData.name);
				tempAbort(buf);
			}
		}
		else if ( AxDontSaveProfiles )
		{
			/* Don't save profiles on exit */

			home = (char *)0;

			sprintf(ProfileData.name, "%s%s%s", ThimApplixDir,
			  DirSlash, ProfileName);

			AxSysDosUnixPath(ProfileData.name);

			if ( -1 == Aaccess(ProfileData.name,R_OK) )
				tempAbort(NXLT("Could not find Applix profiles!"));
		}
		else
		{
			/* check custom dir, and exec dir */

			sprintf(ProfileData.name, "%s%s%s", ThimCustomDir,
			  DirSlash,ProfileName);

			AxSysDosUnixPath(ProfileData.name);

			if ( -1 == Aaccess(ProfileData.name, R_OK) )
			{
				if (lang)
				{
					/*	check multi-lang before default ax_prof4 in
					 *	ThimApplixDir environment */

					sprintf(multiLangProfile, "%s%s%s%s%s.%s",
								ThimCustomDir, DirSlash,
								AxLangSuffix[lang], DirSlash, ProfileName,
								AxLangSuffix[lang]);

					AxSysDosUnixPath(multiLangProfile);

					if ( -1 == Aaccess(multiLangProfile, R_OK) )
					{
						/* check system lang dir */

						sprintf(multiLangProfile, "%s%s%s%s%s.%s",
									ThimApplixDir, DirSlash,
									AxLangSuffix[lang], DirSlash, ProfileName,
									AxLangSuffix[lang]);

						AxSysDosUnixPath(multiLangProfile);

						if ( -1 == Aaccess(multiLangProfile, R_OK) )
							lang = 0;
						else
							sprintf(ProfileData.name, multiLangProfile);
					}
					else
						sprintf(ProfileData.name, multiLangProfile);
				}

				if ( ! lang)
				{
					/*	2 possibilities - multilang env and they didn't
					 *	bother to set AXLANG so it defaults to 1 - so check
					 *	the CustomDir for English anyway, using hard-coded 1
					 */

					sprintf(multiLangProfile, "%s%s%s%s%s.%s",
								ThimCustomDir, DirSlash,
								AxLangSuffix[1], DirSlash, ProfileName,
								AxLangSuffix[1]);

					AxSysDosUnixPath(multiLangProfile);

					if ( -1 == Aaccess(multiLangProfile, R_OK) )
					{
						/* no multi-lang env ; check the Applix dir */

						sprintf(ProfileData.name, "%s%s%s", ThimApplixDir,
						  DirSlash, ProfileName);

						AxSysDosUnixPath(ProfileData.name);

						if ( -1 == Aaccess(ProfileData.name, R_OK) )
							tempAbort(NXLT("Could not find any profiles!"));
					}
					else
						sprintf(ProfileData.name, multiLangProfile);
				}
			}
		}

		/*
		 * if we found a global profile, read it;  if not, punt
		 */

		ProfileData.entries = ReadProfileFile(ProfileData.name);

		/*
		 * create the axhome directory if it doesn't already exist
		 */

		home = AGetAxHomeDir();

		if ( -1 == Aaccess(home, F_OK) )
		{
			/* no axhome - create it */

			if ( -1 == mkdir(home, 0777) )
			{
				sprintf(buf, NXLT("Could not create home directory: %s: %s"),
		          ErrnoTextFromUnix(errno), home);
				tempAbort(buf);
			}
			*new_home = TRUE;
		}

		/* locate the local profile and read that */

		found_file = FALSE;
		local_override = FALSE;

		str = getenv(NXLT("AXPROFILE_USER"));
		if ( str && str[0] )
		{
			strcpy(LProfileData.name, str);

			if ( -1 == Aaccess(LProfileData.name,R_OK) )
			{
				sprintf(buf, NXLT("Could not access profile: %s: %s"),
		          ErrnoTextFromUnix(errno), LProfileData.name);
				tempAbort(buf);
			}
			else
			{
				local_override = TRUE;
				found_file = TRUE;
			}
		}
		else
		{
			/*
			 * check for a local profile from the current rev first
			 */

			sprintf(LProfileData.name,"%s%s%s",home,DirSlash,ProfileName);

			AxSysDosUnixPath(LProfileData.name);

			if ( Aaccess(LProfileData.name, R_OK) != 0 )
			{
				*new_prof = TRUE;

				/*
				 * if no profile from the current rev was found
				 * look for one from the previous rev...
				 */

				sprintf(LProfileData.name,"%s%s%s",home,DirSlash,
				  OldProfileName);

				AxSysDosUnixPath(LProfileData.name);

				if ( Aaccess(LProfileData.name, R_OK) != 0 )
				{
					/*
					 * last try - look back one more rev for the
					 * 2.1 profile file, which lived in the user's
					 * home directory...
					 */

					sprintf(LProfileData.name,"%s%s%s",AGetHomeDir(),
					  DirSlash,VeryOldProfileName);

					AxSysDosUnixPath(LProfileData.name);

					if ( Aaccess(LProfileData.name, R_OK) == 0 )
						found_file = TRUE;
				}
				else
					found_file = TRUE;
			}
			else
				found_file = TRUE;
		}

		if (found_file)
			LProfileData.entries =
			  ReadProfileFile(LProfileData.name);

		/*
		 * save the original profile info for comparison to the
		 * current data at save time - this prevents things that
		 * are changed, then changed back during the same session
		 * from being written to the profile
		 */

		OrigProfileData.entries = ReadProfileFile(ProfileData.name);

		/*
		 * combine the personal and global profile values by
		 * overwriting global things with local things...
		 */

		for ( q = ProfileData.entries; q; q = q->next )
		{
			for ( p = LProfileData.entries; p; p = p->next )
			{
				if ( streq(q->name,p->name) )
				{
					if (!q->locked)
					{
						TaskFree(0,q->value);

						if ( (valp = (char *)TaskAlloc(0,
						  strlen(p->value)+1)) == NULL )
							ElfStrAbort(FENoMemory,NULL,NULL);

						strcpy(valp,p->value);

						q->value = valp;
					}

					q->dirty = TRUE;

					/*
					 * use dirty flag in local version to
					 * indicate that this entry has been
					 * represented in the global version
				 	 */

					p->dirty = TRUE;

					break;
				}
			}
		}

		/*
		 * add things that exist only in the local version
		 * to the combined list
		 */

		for ( p = LProfileData.entries; p; p = p->next )
		{
			if (!p->dirty)
			{
				err = KSAdd(&ProfileData,p->name,p->value,TRUE);

				if ( err != AOK )
					ElfAbort(err);
			}
		}

		/* set the profile name */

		if (!local_override)
			sprintf(LProfileData.name,"%s%s%s",home,DirSlash,ProfileName);

		strcpy(ProfileData.name,LProfileData.name);

		ThimProfileFileName = ProfileData.name;
	}

	return(home);
}

/*************************************************************************
function: AxSyncProfile
*/
AxSyncProfile(new_prof, new_home, new_home_name)
int new_prof, new_home;
char *new_home_name;
{
	char buf[STRPATH+200];
	char *pval;
	int checkpoint;

	checkpoint = FALSE;

	if ( new_prof )
	{
		/*
		 * in consideration of those users who are updating from a
		 * previous rev, if these values are already defined in an
		 * existing profile file, don't mash them here...
		 */

		pval = GetProfileByCode(PROF_CMDS_DIR_);
		if (!pval || !pval[0])
		{
			sprintf(buf,"%s%s%s", new_home_name, DirSlash, NXLT("macros"));
			AxSysDosUnixPath(buf);
			AddProfileByCode(PROF_CMDS_DIR_, buf);

			checkpoint = TRUE;
		}

		pval = GetProfileByCode(PROF_WP_TEMPLATES_DIR);
		if (!pval || !pval[0])
		{
			sprintf(buf,"%s%s%s", new_home_name, DirSlash, NXLT("template"));
			AxSysDosUnixPath(buf);
			AddProfileByCode(PROF_WP_TEMPLATES_DIR, buf);

			checkpoint = TRUE;
		}

		pval = GetProfileByCode(PROF_BACKUP_DIR_);
		if (!pval || !pval[0])
		{
			sprintf(buf,"%s%s%s", new_home_name, DirSlash, NXLT("backup"));
			AxSysDosUnixPath(buf);
			AddProfileByCode(PROF_BACKUP_DIR_, buf);

			checkpoint = TRUE;
		}
	}

	if ( new_home )
		checkpoint = TRUE;

	if ( checkpoint )
		ProfileSave();
}

/*************************************************************************
function: UPbyName - Return profile value or NULL
*/
char *UPbyName (name)
char *name;
{
	/* parse the profile and return the value */
	return(xGetFromKS(name, ProfileData.entries));
}

/*************************************************************************
function: UPbyCode - Return profile value or NULL
*/
char *UPbyCode (code)
int code;
{
	char name[STRPATH+1];

	if ( getStr(code,name) != AOK )
		return(NULL);
	return(UPbyName(name));
}

/*************************************************************************
function: GetProfileByCode - Same as UPneedByCode
*/
char *GetProfileByCode (code)
int code;
{
	register char *p;

	if ( NULL == (p = UPbyCode(code)) )
		return("");
	return(p);
}

/*************************************************************************
function: GetProfileByStr - Fetch profile str (or empty string)
*/
char *GetProfileByStr (str)
char *str;
{
	register char *p;

	if ( NULL == (p = UPbyName(str)) )
		return("");
	return(p);
}

/*************************************************************************
function: UPneedByCode - Return profile value or ""
*/
char *UPneedByCode (code)
int code;
{
	register char *p;

	if ( NULL == (p = UPbyCode(code)) )
		return("");
	return(p);
}

/*************************************************************************
function: UPdblVal - Read a float
*/
double UPdblVal (code)
int code;
{
	register char *p;
	double atof();

	if ( NULL == (p = UPbyCode(code)) )
		return(0);
	return(atof(p));
}

/*************************************************************************
function: UPnumVal - Return a numeric Profile value
Returns 0 if no entry in profile.
*/
UPnumVal (code)
int code;
{
	register char *p;

	if ( NULL == (p = UPbyCode(code)) )
		return(0);
	return(atoi(p));
}

/*************************************************************************
function: ProfileSave - Save current profile info out to profile file
*/
ProfileSave ()
{
	if ( !AxDontSaveProfiles )
		KSSave(&ProfileData);
}

/*************************************************************************
function: ElfbProfiler - Read/Write specified profile file

For read:
Read another profile in, over-writing current values
If the new file has vars that the current state has, the state is updated
to reflect the new file.  Other vars are no affected.

For Write:
*/
void ElfbProfiler (filename,readit)
char *filename;
int readit;
{
	ks_entry *ksP, *p;
	ks_file fileInfo;
	typeErr err;

	if ( readit ) {
		if ( !filename[0] )
			filename = ProfileData.name;
		ksP = ReadProfileFile(filename);
		if ( !ksP )
			ElfStrAbort(UPerrcode,NULL,filename);
		for ( p = ksP; p; p = p->next ) {
			err = KSAdd(&ProfileData,p->name,p->value,p->dirty);
			AxNotifyProfilers(p->name);
			if ( err != AOK )
				ElfAbort(err);
		}
	}
	else if ( !filename[0] )	/* write to default profile, ignore errors */
		ProfileSave();
	else {						/* write to spec'd file, error on err */
		memset(&fileInfo,0,sizeof(fileInfo));
		fileInfo.entries = ProfileData.entries;
		strcpy(fileInfo.name,filename);
		err = KSSave(&fileInfo);
		if ( err != AOK )
			ElfStrAbort(err,NULL,filename);
	}
}

/*************************************************************************
function: ReadProfileFile - Read the asterix profile file into memory list
This returns the head of the list or NULL if there were any read errors.
If the dtmodP is not NULL, this also returns the datestamp of the file.
*/
static ks_entry *ReadProfileFile (filename)
char *filename;
{
	register char *p, *pdst;
	register ks_entry *kshead = NULL, *ksP;
	register int len;
	register FILE *fp;
	char *bad_profile_name;
 	char buf[MAX_KS_LINE+2];
	char name[MAX_KS_LINE+1];
	char value[MAX_KS_LINE+1];
	int linecnt = 0;
	int locked;
    ks_entry *ks_ptr, *next, *newhead = NULL;

	if ( !KsThreadPool )
		KsThreadPool = AxPoolCreate(0,128,sizeof(ks_entry));

	fp = fopen(filename,"r");
	if ( fp == NULL ) {
		UPerrcode = ErrnoErr();
		return(NULL);
	}
	while ( fgets(buf,MAX_KS_LINE,fp) ) {
		locked = FALSE;
		linecnt++;

		len = strlen(buf);

		if (len == 0)
			continue;

		if ( buf[0] == KS_COMMENT_CHAR )
		{
			if (len >= 3)
			{
				if ( (buf[1] == KS_LOCK_CHAR) && (buf[2] == KS_COMMENT_CHAR) )
					locked = TRUE;
				else
					continue;
			}
			else
				continue;
		}

		if ( buf[len-1] == '\n' )
			buf[--len] = '\0';	/* clip newline */

		p = buf;

		if (locked)
		{
			*p++;
			*p++;
			*p++;
		}

        pdst = name;

		if ( buf[0] == '\0' )
			continue;		/* blank line */

		while ( (*p != '\0') && (*p != KS_SEP_CHAR) )
		{
			/* copy name of this entry up to sep char*/
			*pdst++ = *p++;
		}

		if ( *p != KS_SEP_CHAR ) {			/* no sep char!, abandon */
			register ks_entry *next;
			UPerrcode = FEBadKSline;	/* ??1 misleading, use new code */
abandon:	for ( ksP = kshead; ksP; ksP = next ) {
				next = ksP->next;
				AxPoolFree(KsThreadPool, (char*)ksP);
			}
			fclose(fp);

			/*	
			 * move the bad profile aside so that it can be
			 * recovered by hand...
			 */

			bad_profile_name =
			  (char *)malloc( (strlen(filename)+4) * sizeof(char) );

			sprintf(bad_profile_name, "%s.bad", filename);

			 if ( Arename(filename, bad_profile_name) != 0 )
				printf(XLT("Unable to rename bad profile: %s\n"), filename);
			else
			{
				printf(XLT("Moved bad profile to: %s\n"), bad_profile_name);
				printf(XLT(" - See line: %d\n"), linecnt);
			}

			return(NULL);
		}
		*pdst = '\0';				/* null term */
		FullTrim(name,name);		/* trim spaces */

		if ( name[0] == '\0' )
			continue;

		p++;		/* get beyond marker */
		pdst = value;			/* copy value string */
		while ( *p != '\0')
			*pdst++ = *p++;		

		*pdst = '\0';				/* null term */

        if ( strcmp( NXLT("axPrintShellCmd"), name ) != 0 )
        {
            /** we trim blanks on everything but the print shell command    **/
            /**  because some machine, notably the integraph machine, require**/
            /**  a blank between the trailing "-q" and the print queue name **/
            /**  which will be concatenated on the end when we use this     **/
            FullTrim(value,value);      /* trim spaces */
        }

		ksP = (ks_entry*)AxPoolAllocate(KsThreadPool);
		ksP->next = kshead;

		if ( NULL == (ksP->name = (char *)TaskAlloc(0,strlen(name)+1)) )
			goto nomem;
		if ( NULL == (ksP->value = (char *)TaskAlloc(0,strlen(value)+1)) )
			goto nomem;

		strcpy(ksP->name,name);
		strcpy(ksP->value,value);

		ksP->dirty = FALSE;

		ksP->locked = ( locked ) ? TRUE : FALSE;

		kshead = ksP;			/* thread to head of list */
	}
	fclose(fp);

    /* flip the ks entries so the file's first will be the first in the list */
	for ( ks_ptr = kshead; ks_ptr; ks_ptr = next ) 
        {
		next = ks_ptr->next;
		ks_ptr->next = newhead;
		newhead = ks_ptr;
	    }
	return(newhead);

nomem:
	UPerrcode = ErrnoErr();
	goto abandon;
}

/*************************************************************************
function: UPAddNumVal - Add a numeric item to profile
*/
UPAddNumVal (code,value)
int code;
int value;
{
	char buf[100];

	sprintf(buf,"%d",value);
	AddProfileByCode(code,buf);
}

/*************************************************************************
function: AddProfileByCode - Add/Change a profile entry by strfl code
*/
AddProfileByCode (code,value)
int code;
char *value;
{
	char codestr[300];
	typeErr err;
	int dirty = TRUE;

	if ( err = getStr(code,codestr) )
		return(err);
	err = KSAdd(&ProfileData,codestr,value,dirty);
	AxNotifyProfilers(codestr);
	return(err);
}

/*************************************************************************
function: ElfbAddProfile - Add/Change a personal profile option
*/
ElfbAddProfile (code,value)
char *code;		/* actual index string (e.g. "tempDir") in profile */
char *value;
{
	typeErr err;
	int dirty = TRUE;

	err = KSAdd(&ProfileData,code,value,dirty);
	AxNotifyProfilers(code);
	if ( err != AOK ) ElfAbort(err);
}

/*************************************************************************
function: KSAdd - Add a line to Profile or Filing KS files
*/
typeErr KSAdd (ksP,name,value,dirty)
register ks_file *ksP;
register char *name, *value;
register int dirty;
{
	register ks_entry *p;
	register char *valp, *namep;

	if ( NULL == (valp = (char *)TaskAlloc(0,strlen(value)+1)) )
		return(FENoMemory);	/* cannot happen, axmain blocks for more */
	strcpy(valp,value);

/* Try to re-use an existing field */

	for ( p = ksP->entries; p; p = p->next )
	{
		if ( streq(p->name,name) )
		{
			if (p->locked) {
				TaskFree(0,valp);
				if (streq( value, p->value))
					return(AOK);		/* No-op; no error (122069) */
				ElfStrAbort(FEEACCES,XLT("Profile Locked"),p->name);
			}
			else
			{
				free(p->value);
				p->value = valp;
				p->dirty = dirty;
				p->locked = FALSE;
				ksP->changed = TRUE;
				return(AOK);
			}
		}
	}

/* allocate a new line */

	if ( NULL == (namep = (char *)TaskAlloc(0,strlen(name)+1)) )
		return(FENoMemory);
	strcpy(namep,name);
	p = (ks_entry*)AxPoolAllocate(KsThreadPool);
	p->next = ksP->entries;
	p->value = valp;
	p->name = namep;
	p->dirty = dirty;
	p->locked = FALSE;
	ksP->entries = p;
	ksP->changed = TRUE;
	return(AOK);
}

/*************************************************************************
function: KSDelete - Delete a line from a KS image
*/
KSDelete (ksP,name)
register ks_file *ksP;
register char *name;
{
	register ks_entry *p, *next, **prevP;

	prevP = &ksP->entries;
	for ( p = ksP->entries; p; p = next ) {
		next = p->next;
		if ( streq(name,p->name) ) {
			*prevP = p->next;
			free(p->name);
			free(p->value);
			free(p);
		}
		else
			prevP = &p->next;
	}
}

/*************************************************************************
function: NeedFromKS - Get an entry from KS or empty string
*/
char *NeedFromKS (name,ks)
register char *name;
register ks_entry *ks;
{
	register char *p;
	static char *empty = "";

	p = GetFromKS(name,ks);
	if ( p == NULL )
		return(empty); 
	else
	return (p);        /* in either case, return a value */
}


/*************************************************************************
function: GetFromKS - Get an entry from ks data
If named entry not there, returns NULL
*/
char *GetFromKS (name,ks)
register char *name;
register ks_entry *ks;
{
	while ( ks ) {
		if ( streq(name,ks->name) )
			return(ks->value);
		ks = ks->next;
	}
	return(NULL);
}
/*************************************************************************
function: xGetFromKS - Get an entry from ks data - char senstive version
If named entry not there, returns NULL
*/
char *xGetFromKS (name,ks)
register char *name;
register ks_entry *ks;
{
	while ( ks ) {
		if ( 0 == strcmp(name,ks->name) )
			return(ks->value);
		ks = ks->next;
	}
	return(NULL);
}

/*************************************************************************
function: UPIsLocked - returns TRUE if the passed profile entry is
locked, else FALSE.  If not found, entry is not locked.
*/
typeErr UPIsLocked(name)
register char *name;
{
	register ks_entry *ksP;

	for ( ksP = ProfileData.entries; ksP; ksP = ksP->next )
	{
		if ( 0 == strcmp(name,ksP->name) )
		{
			if (ksP->locked)
				return(TRUE);
			else
				return(FALSE);
		}
	}

	return(FALSE);
}

/*************************************************************************
function: KSSave - Save a profile or FS kitchen sink file
modified to only write profile entries to the local file that differ
from the global version - rob 7/1/94
*/
static typeErr KSSave (ksP)
ks_file *ksP;
{
	register ks_entry *ksp, *jsp;
	register FILE *fp;
	char name[STRPATH+1];
	register int code;
	register int found = FALSE;


#if AXMS
	sprintf(name,"%s.x",ksP->name);		/* UNIX profile is already 8 chars */
#else
	sprintf(name,"%sx",ksP->name);		/* first to a tempfile */
#endif
	if ( NULL == (fp = fopen(name,WRITE_STR)) )
		return(ErrnoErr());

	for (ksp = ksP->entries; ksp; ksp = ksp->next)
	{
		/* see if it belongs in the local profile */

		if (ksp->value && ksp->value[0]) 		/* && (ksp->dirty) ) */
		{
			/* see if it differs from the global version */

			for (jsp = OrigProfileData.entries; jsp; jsp = jsp->next)
			{
				if ( streq(ksp->name,jsp->name) )
				{
					if (jsp->locked)
					{
						found = TRUE;
						continue;
					}

					found = TRUE;

					if ( !streq(ksp->value,jsp->value) )
					{
						fprintf(fp,"%s:%s\n",ksp->name,ksp->value);
					}

					break;
				}
			}

			/* not found in global profile - put in local file */

			if (!found)
				fprintf(fp,"%s:%s\n",ksp->name,ksp->value);

			found = FALSE;

		}
	}

	code = ferror(fp);

	if ( 0 != fclose(fp) )
		code = errno;
	if ( code )
	{
		Adelete(name);
		return(ErrnoErr());
	}
	if ( Arename(name,ksP->name) != 0 ) {	/* rename to final file */
		Adelete(name);
		return(ErrnoErr());
	}
	ksP->changed = FALSE;
	return(AOK);
}
