/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: rpc_cli.c
   Applix ASC version: 34
   Date: Mon Nov 16 13:53:24 1998
*/
#include <sys/types.h>
#include <sys/stat.h>
#include <axsocket.h>

#if AXMS
#include <basicwin.h>
void SetSocketAsyncCallback();
extern HWND WXDummyWnd;
#endif

#include <errno.h>
#include <signal.h>

#define THIM_STATIC 1    /** do not redefine such things as "exit()"  */
#include <aos.h>
#include <elfapi.h>
#include <axnet.h>

/* NT declares this is stdlib.h */
#if !AXMS && !LINUX
extern char *sys_errlist[]; /* Unix errro strings */
#endif
extern int AxRPCNonBlockingIO;  /* If TRUE, use AxFdTimedReadBytes() */

extern void AxErrnoError();
extern char *ErrnoTextFromUnix();

int NoteWinIPCUse();
void NoteWinIPCDone();
int AxSockMakeChannelNonBlocking();
int AxSocketClose();
int AxSockUidFromFd();
void ShutDownSocket();

/************************************************************************
function: AxClientConnect - Connect to a server
Returns channel
returns -1 on failure, errno set
*/
int AxClientConnect (socknum,sockname,passed_hostname,code,str,obj)
char *sockname; /* NULL or local process connection file */
char *passed_hostname;  /* only if !sockname */
int socknum;    /* 0 or port number */
int *code;
char **str, **obj;
{
    SOCKET sockfd;
    struct sockaddr_in in_serv_addr;
    struct hostent *info;
    extern int errno;
    int isUnix = FALSE;
    char hostname[101];
    short short_socknum;
#if !AXMS 
    static struct sockaddr_un un_serv_addr;
#endif


    *str = (char *)NULL;
    *obj = (char *)NULL;

    if ( sockname && (*sockname != '\0') )
        isUnix = TRUE;

    if ( isUnix ) {
#if AXMS 
        *code = -1;
        *str = NXLT("UNIX sockets not supported on this platform");
        *obj = NXLT("socket");
        return(-1);
#else
        if ( (sockfd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0 ) {
            *code = ErrnoErr();
            *str = ErrnoTextFromUnix(errno);
            *obj = NXLT("socket");
            return(-1);
        }
#endif
    }
    else {
            NoteWinIPCUse();
            sockfd = socket(AF_INET, SOCK_STREAM, 0);
            if ( sockfd == INVALID_SOCKET )
                {
#if AX_X11
                *code = ErrnoErr();
#else
                *code = WSAGetLastError();
#endif
                *str = ErrnoTextFromUnix(errno);
                *obj = NXLT("socket");
                NoteWinIPCDone();
                return(-1);
                }
    }

    if ( isUnix ) {
#if AXMS 
        *code = -1;
        *str = NXLT("UNIX sockets not supported on this platform");
        *obj = NXLT("socket");
        return(-1);
#else
        bzero(&un_serv_addr, sizeof(un_serv_addr));
        un_serv_addr.sun_family = AF_UNIX;
        strcpy(un_serv_addr.sun_path,sockname);
        if ( connect( sockfd, (struct sockaddr *)(&un_serv_addr), 
                                                sizeof(un_serv_addr)) < 0 ) {
            *code = ErrnoErr();
            *str = ErrnoTextFromUnix(errno);
            *obj = NXLT("connect");
            close(sockfd);
            return(-1);
        }
#endif
    }
    else {
        /* Use TCP sockets */
        /* If no hostname passed in, assume current host */
        if ( passed_hostname ) {
            strncpy(hostname,passed_hostname,100);
            hostname[100] = 0;
        }
        else
            hostname[0] = 0;
        if ( !hostname[0] )
            gethostname(hostname,100);
        info = gethostbyname(hostname);
        if ( !info ) { /* chris 19 oct 93. */
            *code = -1;
            if ( passed_hostname )
                *str = passed_hostname;
            else
                *str = NXLT("current host");
            *obj = NXLT("gethostbyname");
#if AX_X11
            close(sockfd);
#else
            closesocket(sockfd);
#endif
            NoteWinIPCDone();
            return(-1);
        }

        bzero(&in_serv_addr, sizeof(in_serv_addr));
        in_serv_addr.sin_family = AF_INET;
        in_serv_addr.sin_addr.s_addr = *((long*)(info->h_addr));
        short_socknum = socknum;
        in_serv_addr.sin_port = htons( short_socknum );
        if ( connect( sockfd, (struct sockaddr *)(&in_serv_addr),
                                       sizeof(in_serv_addr)) == SOCKET_ERROR ) 
            {
#if AX_X11
            *code = ErrnoErr();
#else
            *code = WSAGetLastError();
#endif
            *str = ErrnoTextFromUnix( *code );
            *obj = NXLT("connect");
#if AX_X11
            close(sockfd);
#else
            closesocket(sockfd);
#endif
            NoteWinIPCDone();
            return(-1);
            }
    }
    if ( AxRPCNonBlockingIO ) { /* If TRUE, use AxFdTimedReadBytes() */
        if ( !AxSockMakeChannelNonBlocking(sockfd) ) {
            *code = ErrnoErr();
            *str = ErrnoTextFromUnix(errno);
            *obj = NXLT("nonblocking");
#if AX_X11
            close(sockfd);
#else
            closesocket(sockfd);
#endif
            NoteWinIPCDone();
            return(-1);
        }
    }

#if AXMS && !ELF_COMM_STAND_ALONE
    /* built into axmain, but not elfapi.a */
    SetSocketAsyncCallback( sockfd );
#endif

    *code = 0;
    return(sockfd);
}

/************************************************************************
function: AxClientDisConnect - Shut connection to server
Accepts both real & non-blocking channels.  Non-blocking channels
have values greater than 100000 per axsocket.c.  What a hack!
This can be called with a non-blocking small-number channel from elsock.c
in "case 3", the AxNetCoonect returns a small numbered channel!
*/
void AxClientDisConnect (channel)
int channel;
{
    if ( AxRPCNonBlockingIO ) {
        if ( channel > 100000 )
            AxSocketClose(channel);
        else if ( AxSockUidFromFd(channel) ) {
            channel = AxSockUidFromFd(channel);
            AxSocketClose(channel);
        }
        else {
            ShutDownSocket( channel );
        }
    }
    else {
        ShutDownSocket( channel );
    }
}





/* This is a centralized routine to do the right thing when a socket is *
/*   to be closed                                       */
void ShutDownSocket( sockfd )
int sockfd;
{
    shutdown( sockfd, 2 );

#if AX_X11
    close( sockfd );
#else
    /* Windows NT */
    closesocket( sockfd );
    NoteWinIPCDone();
#endif

}

#if AX_X11
int NoteWinIPCUse()
{ return(0); }
void NoteWinIPCDone()
{ return; }
#endif



/** Some specific hooks and routines for Windows-NT */
#if AXMS    /*  {  */

/*static int socket_count = -1;*/

int NoteWinIPCUse()
{
#if 0
/* I believe this is obsolete. no'op-ing it for now.
 * should eventually just remove all calls to it.   -donato
 */
    int ret;
    WSADATA wsaData;
    
    if ( socket_count == -1 )
        {
        ret = WSAStartup(0x0101, &wsaData);
        if ( ret != 0 )
            {
            ret = WSAGetLastError();
            return( ret );
            }
        if ( (LOBYTE( wsaData.wVersion ) != 1 ) &&
             (HIBYTE( wsaData.wVersion ) != 1 ) )
            {
            WSACleanup();
            return( WSAVERNOTSUPPORTED );
            }
        }
    socket_count++;
#endif
    return(0);
}


void NoteWinIPCDone()
{
#if 0
/* I believe this is obsolete. no'op-ing it for now.
 * should eventually just remove all calls to it.   -donato
 */
    socket_count--;
   
    if ( socket_count == -1 )
       WSACleanup();
#endif
    return;
}




#if !ELF_COMM_STAND_ALONE   /*  { don't async do socket stuff for elfapi.lib */

extern Window AxTopWinNeverMapped;     /* created by AxGetDpyAndGCs */


void SetSocketAsyncCallback( sockfd )
{
    int ret;
    long flags;

    /* Set up a "callback" message, so that when data arrives on this socket */
    /* the system will generate a message which we get in the main */
    /* MS_GetMessage() loop.  This way we can 'wait' in only one spot.  */
    flags = (FD_READ | FD_WRITE);
    ret = WSAAsyncSelect( sockfd, WXDummyWnd, SOCKET_PSEUDO_EVENT,
                                       flags  );
    if ( ret != 0 )
        tempAbort( NXLT("bad return from WSAAsyncSelect()") );
}

#endif   /*  }  end of not ELF_COMM_STAND_ALONE   */



#endif    /*  } AMS  */

/*************************************************************************
function: AxNetElfChannelCall - Call Elf code from C client
On error, code is set to non-zero
*/
elfData AxNetElfChannelCall(channel,funcName,Args,code,str,obj)
int channel;    /* created by AxNetElfServiceConnect() */
char *funcName; /* must be in a registered axnet service! */
elfData Args;       /* array of args... each element disnt arg (up to 10) */
int *code;          /* returned error code */
char **str;         /* error message */
char **obj;         /* object of error */
{
    elfData response, ElfFuncNamePtr;

    ElfFuncNamePtr = AxMakeStrData(-1,funcName);
    axSendIt(channel,ElfFuncNamePtr,Args,code,str,obj);
    if ( *code != 0 )
        return(NULL);

    response = AxGetResponseFromServer(channel,code,str,obj);
    if ( *code != 0 )
        return(NULL);
    return(response);
}

/*************************************************************************
function: AxNetElfServiceConnect - Used by C clients to talk to Elf Services
Returns a comms channel for use by AxNetElfChannelCall().
On error, returns -1 nad sets code...
*/
int AxNetElfServiceConnect(hostName,fullServiceName,code,str,obj)
char *hostName;             /* or NULL for local machine */
char *fullServiceName;      /* username:servicename */
int *code;          /* returned error code */
char **str;         /* error message */
char **obj;         /* object of error */
{
    int isRemote, channel;
    elfData response, request;
    char sockName[1025];

    if ( !hostName )
        hostName = "";
    if ( hostName[0] )
        isRemote = TRUE;
    else
        isRemote = FALSE;

/* Connect to axnert on the specified host */

    channel = AxNetConnect(hostName,code,str,obj);
    if ( *code != 0 )
        return(-1);

/* Connect to user's axmain via axnet */

    request = AxMakeArray(2);
    request = AxAddStrToArray(request,0,fullServiceName);
    request = AxAddIntToArray(request,1,isRemote);
	/* request will be free'd by AxSendMessageToServer */
    AxSendMessageToServer(channel,AXNET_FIND_SERVICE,request,code,str,obj);
    if ( *code != 0 )
        return(-1);
    response = AxGetResponseFromServer(channel,code,str,obj);
    if ( *code != 0 ) {
        AxClientDisConnect(channel);
        return(-1);
    }

/* If "find" response is NULL, that means that the channel used to
    connect to axnet is now connection to the axmain.
    If the response is number 0, the service does not exist in the
    remote axmain.
    Otherwise, use the returned name as the name of the pipe to
    communicate thru... open connection to the axmain.
    This gets a channel to the service in the axmain (creates a new listener$)
*/

    if ( !response )
        channel = channel;
    else {
        if ( AxIsNumber(response) && (AxIntFromDataPtr(response) == 0) ) {
            AxClientDisConnect(channel);
            *code = 1;
            *str = NXLT("No such ELF service");
            *obj = fullServiceName;
            return(-1);
        }
        strcpy(sockName,AxStrPtrFromDataPtr(response));
        AxFreeData(response);
        AxClientDisConnect(channel);
        channel = AxClientConnect(0,sockName,hostName,code,str,obj);
        if ( *code != 0 )
            return(-1);
    }
    return(channel);
}

