/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: setshlib.c
   Applix ASC version: 15
   Date: Tue Apr 27 15:00:41 1999
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#if !AXMS
#include <sys/utsname.h>
#endif
#include <aos.h>


#if HP800V3
#define LIB_PATH "SHLIB_PATH"
#else
#if RIOS
#define LIB_PATH "LIBPATH"
#else
#if AXMS
#define LIB_PATH "PATH"
#else
#define LIB_PATH "LD_LIBRARY_PATH"
#endif
#endif
#endif

#if AXDIRaxdata
#define THIMBLE_DATA NXLT("axdata")
#define ENT_DIR NXLT("ent")
#else
#if AXDIRdaxdata
#define THIMBLE_DATA NXLT("daxdata")
#define ENT_DIR NXLT("dent")
#else
#if AXDIRpaxdata
#define THIMBLE_DATA NXLT("paxdata")
#define ENT_DIR NXLT("pent")
#else
I will not compile ; because no axdir is defined;
#endif
#endif
#endif

/* If we get a lot of these, the following should be replace with
a #include <sys_protos_def1> and #include <sys_protos_def2> containg
the #if/#endif - rather than clutter each source with #if/#endif pairs. */
#if (SUN4 && !SUN4S)
#define malloc use_sys_prototype
#endif
#include <setshlib.p>
#if (SUN4 && !SUN4S)
#undef malloc
#endif

static void deal_with_sun(char *str, char*WhereIAm);

/***************************************************************************
function: set_shared_lib - set the places where we ship shared libs
*/
void
set_shared_lib(char *WhereIAm)
{
	char mylibs[10*STRPATH+100];
	char *newlibpath, *p, *str;
#if AXMS
	char sep = ';';
#else
	char sep = ':';
#endif

	/* Everybody gets /applix/axexec/axdata/axshlib */

	sprintf(mylibs,"%s=%s/%s/axshlib",LIB_PATH,WhereIAm,THIMBLE_DATA);
	str = mylibs+strlen(mylibs);

	/* Each platform gets its own axajre (and for now jre115/116/113) */

#if SUN4S || SUNIS
	deal_with_sun(str,WhereIAm);
#elif AXMS
	sprintf(str,";%s\\axajre\\bin", WhereIAm);
	str += strlen(str);
	sprintf(str,";%s\\jre115\\bin", WhereIAm);
#elif LINUX && LIBC5 /* only using Java version in libc5 version */
	sprintf(str,":%s/axajre/lib/i386/green_threads", WhereIAm);
	str += strlen(str);
	sprintf(str,":%s/jre115/lib/i386/green_threads", WhereIAm);
#elif RIOS
	strcat(str,":/usr/jdk_base/lib/aix/native_threads");
#elif ALPHA
	sprintf(str,":%s/jre116/lib/alpha/native_threads", WhereIAm);
#elif HP800V3
	sprintf(str,":%s/axajre/lib/PA_RISC/green_threads", WhereIAm);
	str += strlen(str);
	sprintf(str,":%s/jre116/lib/PA_RISC/green_threads", WhereIAm);
#elif SGI
	sprintf(str,":%s/jre116/lib/sgi/green_threads", WhereIAm);
#endif

	/* append the current library path to our library path */
	p = getenv(LIB_PATH);
	if ( p ) {
		newlibpath = malloc(strlen(mylibs)+1 + strlen(p)+1);
		sprintf(newlibpath,"%s%c%s",mylibs,sep,p);
	}
	else {
		newlibpath = malloc(strlen(mylibs)+1);
		strcpy(newlibpath,mylibs);
	}
	putenv(newlibpath);
}


static void deal_with_sun(char *str, char *WhereIAm) {
	char *p;
#if SUN4S || SUNIS
	int e, maj_v, min_v, use_green, ix;
	struct utsname name;
#if SUNIS
	char *machine = "i386";
#else
	char *machine = "sparc";
#endif
	e = uname(&name);
	if ( e < 0 || name.release == NULL || strlen(name.release) < 3 )
		use_green = TRUE;
	else
	{
		/* get the major version number */
		p = name.release;
		ix = 0;
		while ( p && isdigit((int)*p) )
		{
			str[ix++] = *p++;
		}
		if ( ix <= 0 )
			maj_v = 0;
		else
		{
			str[ix] = 0;
			maj_v = atoi(str);
		}

		/* skip past the seperator */
		while ( p && isdigit((int)*p) == FALSE )
			p++;

		/* get the minor version number */
		ix = 0;
		while ( p && isdigit((int)*p) )
		{
			str[ix++] = *p++;
		}
		str[ix] = 0;
		if ( ix <= 0 )
			min_v = 0;
		else
		{
			str[ix] = 0;
			min_v = atoi(str);
		}

		/* 5.6 or better uses native threads */
		use_green = ( maj_v < 5 || min_v < 6 ) ? TRUE : FALSE;
	}

	if ( use_green ) {
		sprintf(str,":%s/axajre/lib/%s/green_threads", WhereIAm, machine);
		str += strlen(str);
		sprintf(str,":%s/jre115/lib/%s/green_threads", WhereIAm, machine);
	}
	else {
		sprintf(str,":%s/axajre/lib/%s/native_threads", WhereIAm, machine);
		str += strlen(str);
		sprintf(str,":%s/jre115/lib/%s/native_threads", WhereIAm, machine);
	}
#endif
}
