/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: shelf.c
   Applix ASC version: 2
   Date: Wed May 19 21:17:01 1999
*/
#if AXDIRaxdata
#define THIMBLE_DATA NXLT("axdata")
#else
#if AXDIRdaxdata
#define THIMBLE_DATA NXLT("daxdata")
#else
I will not compile ; because no axdir is defined;
#endif
#endif

#if AXMS   /* { */
#include <stdio.h>
#include <windows.h>
#include <aos.h>
#include <winmap.h>
#include <shelf.p>

int APIENTRY
WinMain(HINSTANCE hInstance,
        HINSTANCE hPrevInstance,
        LPSTR lpCmdLine,
        int nCmdShow)
{
	int ix, status, ret, len;
    char command_line[4000], elfmain_str[1025];
	char *here;

    PROCESS_INFORMATION process_info;
    STARTUPINFO startup;
    PVOID err_str;
    DWORD buf_len;

	status = EXIT_FAILURE;
	here = WhereAmI(__argv[0]);
	set_shared_lib(here);

	sprintf(elfmain_str, "%s/%s/elfmain.exe", here, THIMBLE_DATA);

    strcpy(command_line, "\"");
    strcat(command_line, elfmain_str);
    strcat(command_line, "\" ");

    for ( ix=1; ix < __argc; ix++ )
    {
       /* add on the rest of the arguments, quoting everything in case */
       /* the field contains a blank                                   */
       strcat(command_line, "\"");
       strcat(command_line, __argv[ix]);
       len = strlen(__argv[ix]);
       if ( *(__argv[ix]+len-1) == '\\' )
            /* so we don't end up escaping the next quote... */
           strcat(command_line, " \" ");
       else
           strcat(command_line, "\" ");
    }

    /* set the wierd Windows global so things get passed along */
    _fileinfo = TRUE;

    memset(&startup, 0, sizeof( STARTUPINFO));
    memset(&process_info, 0, sizeof(PROCESS_INFORMATION));
    ret = MS_CreateProcess(NULL, command_line, 
                           NULL,     /* process security attributes */
                           NULL,     /* thread security attributes */
                           TRUE,     /* process inherits handles   */
                           (DWORD)0, /* creation flags */
                           (LPVOID)0, /* address of new environment block */
                           (LPCSTR)NULL, /* address of current directory */
                           &startup,
                           &process_info);

    if ( ret )
	{
		status = EXIT_SUCCESS;
	}
	else
    {
        /* some error in starting the process */
        ret = MS_GetLastError();    /* for debugging */
        /* make sure we print the error message to stderr. */
        buf_len = MS_FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
                                   FORMAT_MESSAGE_FROM_SYSTEM,
                                   NULL, ret,
                                   0, /* let system decide language! */
                                   (LPTSTR)&err_str, 0, NULL);

        if ( err_str )
        {
			printf((char*)err_str);
            /* need to free the string allocated by FormatMessage */
            MS_LocalFree(err_str);
        }
    }
    
    MS_CloseHandle(process_info.hThread);
    MS_CloseHandle(process_info.hProcess);

	return status;
}

#else    /* } { */

#include <stdio.h>
#include <sys/types.h>
#include <locale.h>
#include <errno.h>
#include <stdlib.h>
#include <aos.h>
#include <afile.h>
#include <ashell.h>
#include <elapi.h>
#include <axnet.h>
#include <shelf.p>

/*************************************************************************
function: main - Main entry point
*/
main (argc,argv)
int argc;
char *argv[];
{
	char *WhereIAm;		/* result from WhereAmI */
	char pathname[STRPATH+1];

	set_locale_c();
	WhereIAm = WhereAmI(argv[0]);
	set_shared_lib(WhereIAm);
	sprintf(pathname,"%s/%s/elfmain",WhereIAm,THIMBLE_DATA);
	argv[0] = pathname;
	execv(argv[0],argv);
	fprintf(stderr,"%s: %s\n",argv[0],ErrnoTextFromUnix(errno));
	_exit(127);
}

/***********************************************************************
function: set_locale_c - Set all locale variables to NXLT("C").  Our
product assumes default (american) locale.  We do all translations ourselves.
Eg. In sprintf(), we want the %f format for number 3.14 to yield 3.14
and not 3,14 (german).
*/
static void set_locale_c ()
{
	char *p;

#if !BUILD_ORIENT
    /* Asian languages use these to drive the input mechanism */
	p = strdup(NXLT("LANG=C"));         putenv(p);
	p = strdup(NXLT("LC_ALL=C"));       putenv(p);
#endif
	p = strdup(NXLT("LC_MONETARY=C"));  putenv(p);
	p = strdup(NXLT("LC_NUMERIC=C"));   putenv(p);
	p = strdup(NXLT("LC_TIME=C"));      putenv(p);
	return;
}

/*************************************************************************
function: ErrnoTextFromUnix - Read Unix errno text
*/
char *ErrnoTextFromUnix (err)
int err;
{
    static char *buf;
#if !LINUX
    extern char *sys_errlist[];
    extern int sys_nerr;
#endif

    if ( err < sys_nerr )
        return((char*)sys_errlist[err]); /* cast gets rid of 'const' warning */
    if ( !buf )
        buf = malloc(300);
    sprintf(buf,"err %d",err);
    return(buf);
}
#endif   /* } */
