/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: thim_se.c
   Applix ASC version: 98
   Date: Thu Feb 11 21:11:24 1999
*/
#include <stdio.h>
#include <aos.h>
#include <afile.h>
#include <signal.h>
#include <errno.h>
#include <paripc.h>
#include <dia.h>
#include <fsrts.h>
#include <print.h>
#include <rsval.h>
#include <atm.h>
#include <atmwin.h>
#include <atmreg.h>
#include <atmattr.h>
#include <atmtxtin.h>
#include <atmgraph.h>
#include <akeys.h>
#include <paratmgr.h>
#include <parattr.h>
#include <parout.h>
#include <parwin.h>
#include <atmimage.h>       /* Image call data structure */
#include <parimage.h>   /* Image packet structure */
#include <scac.h>
#include <devsetup.h>
#include <pargmod.h>        /* Graphics mode packet structure */
#include <sff.h>
#include <getstr10.h>
#include <axtext.h>


/*-----------------------------------------------------------------------------
	define
-----------------------------------------------------------------------------*/

#define CNULL       ((char *) NULL)
#define SNULL       ((short *) NULL)
#define LNULL       ((long *) NULL)
#define FNULL       ((FILE *) NULL)


#define BUFFER_SIZE 31000   /* Big local buffer is a win! */

#define MAX_BUF_LEN 1024

#define NEWLINE     ('\n')

#define SE_PRL_V4	NXLT("SE serial file V4")
#define SE_PRL2		NXLT("string 2")
#define SE_PRL3		NXLT("string 3")


        /* window control block definition */

typedef struct {short x, y;} Point; /* device coordinates */

typedef struct {
    short x1;       /* start point, x */
    short y1;       /* start point, y */
    short x2;       /* end point, x */
    short y2;       /* end point, y */
} Vector;

typedef struct {
    Point origin;
    Point corner;
} AxRectangle;

typedef struct {
    int x;
    int y;
} Point32;

static struct {
    TmLineDescriptor lin_des;
    TmTextDescriptor write_tx_des;
    TmTextDescriptor read_tx_des;
    short trans_mode;
    short txAlign;          /* top-left or base-left alignment of text */
    short spWid;            /* default space width in pixel */
    short spCnt1;           /* space counts set by alter text extent */
    short spCnt2;           /* spCnt1 is no of spaces with spChange+1 */
                            /* spCnt2 is no of spaces with spChange */
    short spChng1;          /* added to spWid to effect total change 
                                in text extent (for spCnt1) */
    short spChng2;          /* added to spWid to effect total change 
                                in text extent (for spCnt2) */
    AxRectangle inner;      /* inner clipping rectangle */
    AxRectangle outer;      /* outer clipping rectangle */
    AxRectangle appClip;        /* application clipping rectangle */
} page;

/* this typedef is used in seText to manage a bullet-display list  */
typedef struct bullet_call_t {
    r32 bRect; /* bullet position */
    txAttr *txAp;             /* text attr of text assoc. w/ bullet */
} bulletCall;


#include <thim_se.p>

/*-----------------------------------------------------------------------------
	global external procedures
-----------------------------------------------------------------------------*/
extern	char	*TrueMalloc();
extern	void	axtGetZoom(), axtGetFieldTransform();
extern	void	axtWriteTargetCtm(), pfGetTransformTerms();
extern	void	axtAdjustRulingEndpoints();
extern	int		mToMils(), axtHasOffset(), axtHasBullet();
extern	void	freeTextAttrs();

/*-----------------------------------------------------------------------------
	global internal procedures
-----------------------------------------------------------------------------*/
extern	typeErr	SEinit() ;
extern	typeErr	SEterm() ;
extern	void	se_newpage() ;

/*-----------------------------------------------------------------------------
	local data
-----------------------------------------------------------------------------*/
static	int		pageMilWidth;
static	int		pageMilHeight;
static	int		landscape;
static	char	*SEfname ;
static	int		SEfdes ;
static	int		SETermStatus ;
static	int		bufidx ;
static	char	*buf = (char *)0;

/*-----------------------------------------------------------------------------
	local procedures
-----------------------------------------------------------------------------*/
static	void	seRecordBullet();
static	void	OutTextAttrs() ;
static	void	OutTxFieldProps() ;
static	void	OutPrelude() ;
static	void	OutByte() ;
static	void	OutWord() ;
static	void	OutLong() ;
static	void	OutInt() ;
static	void	OutNString() ;
static	void	OutString() ;
static	void	OutCString() ;
static	void	OutTxCharString() ;
static	void	OutAlign() ;
static	void	WriteIt () ;
static	void	page_init() ;

static	void	se_gmode() ;
static	void	se_write_poly_attr() ;


/*-----------------------------------------------------------------------------
	SEinit
-----------------------------------------------------------------------------*/
typeErr SEinit(filename)
IN char *filename;  /* file to create */
{
    short height = 11000,
            width = 8500,
            orient = 0;

    typeErr e;

	/* get setup info */

    pageMilHeight = height;
    pageMilWidth = width;
    landscape = (orient == 1);  /* name this thing?? */

	/* open output file */

    SEfname = filename;
#if AXMS
    /* we must make sure to open this file in binary mode, so carriage */
    /* returns do not get turned into CR-LF pairs which screw up the byte */
    /* count.  You can not do a 'creat' and set binary mode.            */
    SEfdes = open( filename, (_O_CREAT | _O_BINARY | _O_RDWR), (_S_IWRITE | _S_IREAD ) );
#else
    SEfdes = creat(filename,0666);
#endif
    if ( -1 == SEfdes )
        {
        e = ErrnoErr();
        Adelete(filename);
        return(e);
        }
    if (buf == NULL)
        if (NULL == (buf = TrueMalloc(BUFFER_SIZE)) )
            return(ErrnoErr());

	SETermStatus = AOK ;
    bufidx = 0;
    OutPrelude ();
    page_init();    /* initialize base window management */
    return(AOK);
}

/*-----------------------------------------------------------------------------
	SEevDispat - dispatches on the event evInd with data in mes, length mesLen.

-----------------------------------------------------------------------------*/
SEevDispat(evInd, mes, mesLen, rtnBuf)
int evInd;
char *mes;
short mesLen;
char *rtnBuf;
{
    TmRect32 bnds;
    TmRect32 clip;

    clip.x_position = clip.y_position = 0 ;
    clip.x_size = 8500 ;
    clip.y_size = 11000 ;

    switch(evInd)
    {
        case QATM_CREATE:
            se_newpage(FALSE, 8500, 11000, 0, FALSE, FALSE, FALSE, &bnds, &clip, 0, 0, 0);
            break;
        case QATM_TERM:
            SEterm();
            break;
        case QATM_MODIFY:
            err_return(rtnBuf);
            break;
        case QATM_CURWIN:
            err_return(rtnBuf);
            break;
        case QATM_INQ:
            se_inq(rtnBuf);
            break;
        case QATM_DCURSOR:
            err_return(rtnBuf);
            break;
        case QATM_TEXT_EXT:
             /*se_text_ext(mes, mesLen, rtnBuf);*/
             break;
        case QATM_SCROLL:
            err_return(rtnBuf);
            break;
#if 0
        case QATM_PLINE:
            se_pline(mes, mesLen);
            break;
#endif
        case QATM_Arc:
            se_Arc(mes, mesLen);
            break;
        case QATM_SET_WIN:
            se_set_win(mes, mesLen);
            break;
        case QATM_GMODE:
            se_gmode(mes, mesLen);
            break;
        case QATM_PIMAGE:
            se_pimage(mes, mesLen);
            break;
        case QATM_TABS:
            se_tabs(mes, mesLen, mesLen);
            break;
        case QATM_IMGTXT:
            se_imgtxt(mes, mesLen);
            break;
        case QATM_INCLUDE:
            err_return(rtnBuf);
            break;
        default:
            err_return(rtnBuf);
            break;
    };

}

/*-----------------------------------------------------------------------------
 * err_return - this routine returns the error TMR_PRINTER to the runtime
 * that called the SE.  It is used for all routines that are defined to
 * return data but have no meaning in printing.
-----------------------------------------------------------------------------*/
err_return (rtnBuf)
char *rtnBuf;
{

    struct {
        char filler1[IPC_HEAD_LEN];
        short status;
    } err_packet;

    err_packet.status = TMR_PRINTER;
    memcpy(rtnBuf,(char *) &err_packet, sizeof(err_packet));
}


/*-----------------------------------------------------------------------------
	se_inq
-----------------------------------------------------------------------------*/
se_inq (rtnBuf)
char *rtnBuf;
{
    memset(rtnBuf,0,sizeof(struct RInqData));
}


/*-----------------------------------------------------------------------------
	TmSetWinAttr function
-----------------------------------------------------------------------------*/
se_set_win(mes, mesLen)
char *mes;
short mesLen;
{
    short x, y, x_size, y_size;
    union 
    {
        struct SSetWinData set_win_data;
        char rec_buf[sizeof(struct SSetWinData)];
    } rec_pack;

    memcpy(rec_pack.rec_buf,mes,mesLen);

    OutWord (SfSetWinAttr);
    page.lin_des = rec_pack.set_win_data.lin_des;
    OutWord (page.lin_des.line_style);
    OutWord (page.lin_des.line_thickness);
    OutLong ((long) (page.write_tx_des = rec_pack.set_win_data.write_tx_des));
    OutWord (page.trans_mode = rec_pack.set_win_data.trans_mode);
    OutWord (page.txAlign = rec_pack.set_win_data.txAlign);

    OutWord (rec_pack.set_win_data.fg_color);   /* foreground color */
    OutWord (rec_pack.set_win_data.bg_color);   /* background color */


    x = rec_pack.set_win_data.cliprect.x_position;
    y = rec_pack.set_win_data.cliprect.y_position;
    x_size = rec_pack.set_win_data.cliprect.x_size;
    y_size = rec_pack.set_win_data.cliprect.y_size;

    if (x_size >= 0)
    {
		/* application supplied clipping rectangle */
        page.appClip.origin.x = x;
        page.appClip.origin.y = y;
        page.appClip.corner.x = x + x_size;
        page.appClip.corner.y = y + y_size;
    }
    OutWord (page.appClip.origin.x);
    OutWord (page.appClip.origin.y);
    OutWord (page.appClip.corner.x);
    OutWord (page.appClip.corner.y);

    OutWord (SfEndCall);

}


/*-----------------------------------------------------------------------------
	se_imgtxt
-----------------------------------------------------------------------------*/
se_imgtxt(mes, mesLen)
char *mes;
short mesLen;
{
    union
    {   
        struct SSetImgTxtMode set_imgtxt_data;
        char rec_buf[sizeof(struct SSetImgTxtMode)];
    } rec_pack;

    memcpy(rec_pack.rec_buf,mes,mesLen);
    OutWord (SfScaledTxt);
    OutWord ((short)rec_pack.set_imgtxt_data.on);
    OutWord ((short)rec_pack.set_imgtxt_data.imageBox[0].x);
    OutWord ((short)rec_pack.set_imgtxt_data.imageBox[0].y);
    OutWord ((short)rec_pack.set_imgtxt_data.imageBox[1].x);
    OutWord ((short)rec_pack.set_imgtxt_data.imageBox[1].y);
    OutWord ((short)rec_pack.set_imgtxt_data.imageBox[2].x);
    OutWord ((short)rec_pack.set_imgtxt_data.imageBox[2].y);
    OutWord ((short)rec_pack.set_imgtxt_data.imageBox[3].x);
    OutWord ((short)rec_pack.set_imgtxt_data.imageBox[3].y);
    OutWord (SfEndCall);
}

/*-----------------------------------------------------------------------------
	se_write_poly_attr
-----------------------------------------------------------------------------*/
static void se_write_poly_attr(attr)
attr_type *attr;
{
    register int i;
    register int j;
    register scan_bytes;

    /* write the fill attributes */
    OutWord (attr->fill_flag);
    OutWord (attr->f_type);
    OutWord (attr->f_pat);
    OutWord (attr->f_ang);
    OutWord (attr->f_xoff);
    OutWord (attr->f_yoff);

    OutWord (attr->s_type);
    OutWord (attr->s_xoff);
    OutWord (attr->s_yoff);

    /* write any passed in fill fill pattern */
    OutWord (attr->f_dhyt);
    OutWord (attr->f_dwid);
    OutWord (attr->f_ddep);
    scan_bytes = 2*((attr->f_dwid * attr->f_ddep + 15) / 16);
#if 0
    for ( i=0; i<attr->f_dhyt; i++ )
        for ( j=0; j<scan_bytes; j++ )
            OutByte ( attr->f_data[i*scan_bytes + j] );
#endif
    OutCString(attr->f_data, scan_bytes * attr->f_dhyt) ;
    OutAlign ();
    if (attr->f_cmap)
    {
        OutWord(attr->f_cmap->n_colors) ;
        OutCString((char *)attr->f_cmap->color,
                attr->f_cmap->n_colors * sizeof(color_type)) ;
    }
    else
        OutWord(0) ;
    OutAlign ();

    /* write the line style */
    OutWord (attr->l_style);
    OutWord (attr->l_wid);
    OutWord (attr->l_cap);

    /* write the line attributes */
    OutWord (attr->line_flag);
    OutWord (attr->l_type);
    OutWord (attr->l_pat);
    OutWord (attr->l_ang);
    OutWord (attr->l_xoff);
    OutWord (attr->l_yoff);

    /* write any passed in line fill pattern */
    OutWord (attr->l_dhyt);
    OutWord (attr->l_dwid);
    OutWord (attr->l_ddep);
    scan_bytes = 2*((attr->l_dwid * attr->l_ddep + 15) / 16);
#if 0
    for ( i=0; i<attr->l_dhyt; i++ )
        for ( j=0; j<scan_bytes; j++ )
            OutByte ( attr->l_data[i*scan_bytes + j] );
#endif
    OutCString(attr->l_data, scan_bytes * attr->l_dhyt) ;
    OutAlign ();
    if (attr->l_cmap)
    {
        OutWord(attr->l_cmap->n_colors) ;
        OutCString((char *)attr->l_cmap->color,
                attr->l_cmap->n_colors * sizeof(color_type)) ;
    }
    else
        OutWord(0) ;
    OutAlign ();

    /* write the color attributes */
    OutWord (attr->ffg.c);
    OutWord (attr->ffg.m);
    OutWord (attr->ffg.y);
    OutWord (attr->ffg.k);
    OutWord (attr->ffg.see_thru);
    OutWord (attr->fbg.c);
    OutWord (attr->fbg.m);
    OutWord (attr->fbg.y);
    OutWord (attr->fbg.k);
    OutWord (attr->fbg.see_thru);
    OutWord (attr->lfg.c);
    OutWord (attr->lfg.m);
    OutWord (attr->lfg.y);
    OutWord (attr->lfg.k);
    OutWord (attr->lfg.see_thru);
    OutWord (attr->lbg.c);
    OutWord (attr->lbg.m);
    OutWord (attr->lbg.y);
    OutWord (attr->lbg.k);
    OutWord (attr->lbg.see_thru);
    OutWord (attr->sfg.c);
    OutWord (attr->sfg.m);
    OutWord (attr->sfg.y);
    OutWord (attr->sfg.k);
    OutWord (attr->sfg.see_thru);

    /* override transfer mode */
    OutWord (attr->mode);

    /* absolute fill info */
    OutWord (attr->fill_abs);
    OutWord (attr->line_abs);
    OutWord (attr->x_off);
    OutWord (attr->y_off);

    /* write the other flags */
    OutWord (attr->line_first);
    OutWord (attr->use_ffg);
    OutWord (attr->use_fbg);
    OutWord (attr->use_lfg);
    OutWord (attr->use_lbg);
}

/*-----------------------------------------------------------------------------
	se_clip
-----------------------------------------------------------------------------*/
se_clip(clip)
TmRect32 *clip;
{

    OutWord (SfClip);
    OutCString((char *)clip, sizeof(TmRect32)) ;
    OutWord (SfEndCall);
}

/*-----------------------------------------------------------------------------
	se_prcontrol
-----------------------------------------------------------------------------*/
se_prcontrol(n, buf)
int n ;
char *buf ;
{
    int k = strlen(buf) ;

    OutWord (SfPrControl);
    OutInt(n) ;
    OutInt(k) ;
    OutCString(buf, k) ;
    OutWord (SfEndCall);
}

/*-----------------------------------------------------------------------------
	ax_se_pline

    TmPolyline function for printing
-----------------------------------------------------------------------------*/
ax_se_pline(wID, nPolys, attr, poly_listP, clip, clip_info)
int wID;
int nPolys;
attr_type *attr;
poly_type *poly_listP;
TmRect32 *clip;
clip_path *clip_info;
{
    int i, j, scan_bytes;
    register poly_type *polyP;
    register TmPoint32 *pointP;

    OutWord (SfPolyline);

    /* write the polygon data */
    OutWord (nPolys);
    polyP = poly_listP;
    for ( i=0; i<nPolys; i++ )
    {
        OutWord (polyP->nPoints);
        pointP = polyP->points;

        for ( j=0; j<polyP->nPoints; j++ )
        {
            OutInt (pointP->x);
            OutInt (pointP->y);
            pointP++;
        }
        OutWord (polyP->nCtrls);
        pointP = polyP->controls;
        for ( j=0; j<polyP->nCtrls; j++ )
        {
            OutInt (pointP->x);
            OutInt (pointP->y);
            pointP++;
        }
        polyP++;
    }

    /* write the poly attributes to the se file */
    se_write_poly_attr(attr);

    /* clipping rectangle */
    OutCString((char *)clip, sizeof(TmRect32));

	if ( clip_info )
	{
		/* write the polygon data */
		OutWord (clip_info->n_paths);
		for ( i=0; i<clip_info->n_paths; i++ )
		{
			polyP = &clip_info->paths[i];
			OutWord (polyP->nPoints);

			for ( j=0; j<polyP->nPoints; j++ )
			{
				pointP = &polyP->points[j];
				OutInt (pointP->x);
				OutInt (pointP->y);
			}
			OutWord (polyP->nCtrls);

			for ( j=0; j<polyP->nCtrls; j++ )
			{
				pointP = &polyP->controls[j];
				OutInt (pointP->x);
				OutInt (pointP->y);
			}
		}
	}
	else
		OutWord (0);
    OutWord (SfEndCall);
}


/*-----------------------------------------------------------------------------
	se_pline

    old TmPolyline function for printing
-----------------------------------------------------------------------------*/
se_pline(mes, mesLen)
char *mes;
short mesLen;
{
}

/*-----------------------------------------------------------------------------
	ax_se_arc

    TmArc function for printing
-----------------------------------------------------------------------------*/
ax_se_arc(wID, bound_box, start_angle, arc_angle, attr)
int wID;
TmRect32 *bound_box;
int start_angle;
int arc_angle;
attr_type *attr;
{

    OutWord (SfArc);

    /* write the ellipse data */
    OutWord (bound_box->x_position);
    OutWord (bound_box->y_position);
    OutWord (bound_box->x_size);
    OutWord (bound_box->y_size);
    OutWord (start_angle);
    OutWord (arc_angle);

    /* write the poly attributes to the se file */
    se_write_poly_attr(attr);

    OutWord (SfEndCall);
}

/*-----------------------------------------------------------------------------
	se_Arc

    old TmArc function for printing
-----------------------------------------------------------------------------*/
se_Arc(mes, mesLen)
char *mes;
short mesLen;
{
}

/*-----------------------------------------------------------------------------
	se_tabs

    TmSetTabStops function
-----------------------------------------------------------------------------*/
se_tabs(mes, mesLen, buffered)
char *mes;
short mesLen;
short buffered;
{
    union
    {
        struct StabsData data;
        char rec_buf[MAX_BUF_LEN];
    }rec_pack;
    TmTabStop stops[200];/* plenty, needed to get good PMAX alignment */
    short  ntabs, i, typ, ldr;

    memcpy(rec_pack.rec_buf,mes,mesLen);

    ntabs = rec_pack.data.numTabs;

    OutWord (SfTabs);
    OutWord (rec_pack.data.numTabs);
    OutWord (rec_pack.data.decimalChar);
    memcpy(stops,&rec_pack.rec_buf[sizeof(rec_pack.data)],
                                ntabs*sizeof(TmTabStop));
    for(i=0; i<ntabs; i++)
    {
        typ = (short) stops[i].tabType;
        ldr = (short) stops[i].leadering;
        OutWord (typ);
        OutWord (ldr);
        OutWord (stops[i].devTabX);
        OutWord ((short)((stops[i].decimalChar) &  0xff));
    }
    OutWord (SfEndCall);
}

/*-----------------------------------------------------------------------------
    se_pimage - Taken from TmPutImage function
-----------------------------------------------------------------------------*/
se_pimage(mes, mesLen)
char *mes;
short mesLen;
{
}

/*-----------------------------------------------------------------------------
    ax_se_pimage
	
	'TmPutImage2' function. Need to do the masks!...
-----------------------------------------------------------------------------*/
ax_se_pimage(wID, pnt, imP, attr, xform, x_scale, y_scale, cmap, clip)
int wID;
TmPoint32 *pnt;
ge_image_type *imP;
attr_type *attr;
poly_type *xform;
double x_scale, y_scale;
cmap_type *cmap;
TmRect32 *clip;
{
    register int ix;
    int horiz;
    register int vert_scan;
    unsigned char *tmp_data_ptr;

    if (!imP)
        return;

    if (xform)
    {
            /* scale the transform data */
        if (xform->points)
        {
            for (ix = 0; ix < xform->nPoints; ix++)
            {
                xform->points[ix].x = (int)(xform->points[ix].x * x_scale);
                xform->points[ix].y = (int)(xform->points[ix].y * y_scale);
            }
        }
        if (xform->controls)
        {
            for (ix = 0; ix < xform->nCtrls; ix++)
            {
                xform->controls[ix].x = (int)(xform->controls[ix].x * x_scale);
                xform->controls[ix].y = (int)(xform->controls[ix].y * y_scale);
            }
        }
    }

    OutWord (SfImage);
    OutWord (0);    /* pixel flag */
        
    if ( attr->mode != TMS )
        OutWord (attr->mode);
    else if ( attr->ffg.see_thru && attr->fbg.see_thru )
        OutWord (TMD);
    else if ( attr->ffg.see_thru )
        OutWord (TMIS_O_D);
    else if ( attr->fbg.see_thru )
        OutWord (TMS_O_D);
    else
        OutWord (TMS);
    if ( attr->use_ffg )
    {
        OutByte(0x00);
        OutByte(0x00);
        OutByte(0x00);
        OutByte(0xff);
    }
    else
    {
        OutByte(attr->ffg.c);
        OutByte(attr->ffg.m);
        OutByte(attr->ffg.y);
        OutByte(attr->ffg.k);
    }
    if ( attr->use_fbg )
    {
        OutByte(0x00);
        OutByte(0x00);
        OutByte(0x00);
        OutByte(0x00);
    }
    else
    {
        OutByte(attr->fbg.c);
        OutByte(attr->fbg.m);
        OutByte(attr->fbg.y);
        OutByte(attr->fbg.k);
    }
    OutInt (pnt->x);    /* x pos */
    OutInt (pnt->y);    /* y pos */
    OutInt (imP->wid);  /* x size */
    OutInt (imP->hyt);  /* y size */
    OutWord (imP->dep); /* image depth */
    OutLong (xform->points[0].x);
    OutLong (xform->points[0].y);
    OutLong (xform->points[1].x);
    OutLong (xform->points[1].y);
    OutLong (xform->points[2].x);
    OutLong (xform->points[2].y);
    OutLong (xform->points[3].x);
    OutLong (xform->points[3].y);
    OutLong (xform->points[4].x);
    OutLong (xform->points[4].y);
#if BYTE_WID_IN_SE
    OutWord (imP->data_byte_wid);   /* x width of data in bytes */
#endif
    tmp_data_ptr = imP->data;
    for (vert_scan=0; vert_scan<imP->hyt; vert_scan++)
    {
        for (horiz=0; horiz < imP->data_byte_wid; horiz++)
        {
            /* put bytes of image data into se stream */
            OutByte (*tmp_data_ptr++); /* is PutWord() better? */
        }
    }

    OutAlign();
    OutWord (imP->mask_byte_wid);   /* NB. BYTE wid; not bits */
    if ( (imP->mask_byte_wid != 0) && (imP->mask != NULL) )
    {
        tmp_data_ptr = imP->mask;
        for (vert_scan=0; vert_scan<imP->hyt; vert_scan++)
        {
            for (horiz=0; horiz < imP->mask_byte_wid; horiz++)
            {
                OutByte (*tmp_data_ptr++);
            }
        }
    }

    /* highjack vert_scan for the colormap */
    OutLong(cmap->n_colors);
    for (vert_scan=0; vert_scan<cmap->n_colors; vert_scan++)
    {
        if (cmap->color[vert_scan].see_thru)
            OutInt(1);
        else
            OutInt(0);
        OutByte(cmap->color[vert_scan].c);
        OutByte(cmap->color[vert_scan].m);
        OutByte(cmap->color[vert_scan].y);
        OutByte(cmap->color[vert_scan].k);
    }

    /* clipping rectangle */
    OutCString((char *)clip, sizeof(TmRect32)) ;

    /* Now for the EPS data */
    OutLong (imP->eps_seg_size);    /* number of bytes of eps file */
    OutCString(imP->eps_seg, imP->eps_seg_size);
    if (imP->eps_seg_size & 1)
        OutByte('\0');  /* OUCH! InCString() only reads even numbers
                            of bytes... why? */

    OutWord(SfEndCall);
}

/*-----------------------------------------------------------------------------
	seText
-----------------------------------------------------------------------------*/
int seText(txt, clip, shadow_type)
axtext *txt;
r32 *clip ;
int shadow_type;
{
    int charcount, charix, tcharix, width, ix;
    int pieceCount, charLen, pieceIx, listsize;
    txAttr *txAp;
    perChar *pc;
    p32 piecePos, curpos, prevSup, tPrevSup, bpt;
    int bulletListSize = 0;
    bulletCall *bulletList = (bulletCall *) NULL;

    if (!txt) {
        return FALSE;
	}

    if (txt->whole.nChars) {
		
		OutWord (SfText);
		OutInt (shadow_type);

		/* out field props
		 */
		OutTxFieldProps(txt);
		
		/* clipping rectangle
		 */
		OutCString((char *)clip, sizeof(r32)) ;

		/* out total # attr segs; -1 denotes whole
		 */
		OutInt (segCount(txt, -1, &txAp));
		charix = 0;
		curpos = txt->whole.pos;
		prevSup.x = prevSup.y = 0;
	}
	
	txAp = &txt->atts[0];
	charcount = txt->whole.nChars;
	
	if (charcount) {
		while(charcount) {
			if (!txAp->nChars) {
				OutInt (0); /* a pieceCount of zero */
				txAp++;
				continue;
			}
			/* for each attr seg, output the text in
			 * the largest possible 'pieces' that can
			 * be handled by a PS show operator, e.g..
			 * first, count how many pieces there will be
			 */
			tcharix = charix;
			charLen = 0;
			tPrevSup.x = tPrevSup.y = 0;
			for (ix = 0, pieceCount = 1; ix < txAp->nChars; ix++) {
				if (axtHasOffset(&txt->items[tcharix], &tPrevSup) && charLen) {
					pieceCount++;
					charLen = 1;
				}
				else {
					charLen++;
				}
				tPrevSup = txt->items[tcharix].sup;
				tcharix++;
			}
			OutInt (pieceCount);
			/* now write attr info for this seg
			 */
			OutTextAttrs(txt, charix, txAp->nChars, txAp->fontAttr.fontName,
						 mToMils(txAp->fontAttr.underlinePosition,
								 txAp->fontAttr.ptsize),
						 mToMils(txAp->fontAttr.underlineThickness,
								 txAp->fontAttr.ptsize),
						 strikeThruOffset(txAp));
			
			/* now write the pieces
			 */
			charLen = 0;
			width = 0;
			piecePos = curpos;
			pieceIx = tcharix = charix;
			
			/* adjust piecePos for sub/sup, kern, offset
			 */
			piecePos.x += txt->items[pieceIx].sup.x;
			piecePos.y += txt->items[pieceIx].sup.y;
			piecePos.x += txt->items[pieceIx].kern.x;
			piecePos.y += txt->items[pieceIx].kern.y;
			piecePos.x += txt->items[pieceIx].offset.x;
			piecePos.y += txt->items[pieceIx].offset.y;
			
			for (ix = 0; ix < txAp->nChars; ix++) {
				pc = &txt->items[tcharix];
				if (txAp->indent) {
					p32 tpos;
					tpos = curpos;
					tpos.x += pc->offset.x + pc->kern.x;
					tpos.y += pc->offset.y + pc->kern.y;
					seRecordBullet(txt, txAp, tcharix, &tpos, &bulletListSize,
								   &bulletList, shadow_type);
				}
				if (axtHasOffset(pc, &prevSup) && charLen) {
					/* output pos, width, string
					 */
					OutInt (piecePos.x);
					OutInt (piecePos.y);
					OutInt (width);
					OutTxCharString(&txt->items[pieceIx], charLen);
					if (txAp->ruling || txAp->strike_thru) {
						/* output the right endpoint of the
						 * ruling for the piece just completed;
						 * derive from curpos
						 */
						p32 leftpos, rightpos;
						leftpos = piecePos;
						rightpos = curpos;
						if (ix && tcharix) {
							/* curpos tracks changes to the set position
							 * and so doesn't contain sup; look one char back
							 * to add in the piece-end's superscript val
							 */
							rightpos.x += txt->items[tcharix - 1].sup.x;
							rightpos.y += txt->items[tcharix - 1].sup.y;
						}
						axtAdjustRulingEndpoints(txt, pieceIx,
												 tcharix - pieceIx,
												 &leftpos, &rightpos,
												 &leftpos, &rightpos);
						OutInt(leftpos.x);
						OutInt(leftpos.y);
						OutInt(rightpos.x);
						OutInt(rightpos.y);
					}
					charLen = 1;
					if (!pc->hidden)
						width = pc->padvance;
					piecePos = curpos;
					pieceIx = tcharix;
					if (ix > 0) {
						piecePos.x += txt->items[pieceIx].kern.x;
						piecePos.y += txt->items[pieceIx].kern.y;
						piecePos.x += txt->items[pieceIx].offset.x;
						piecePos.y += txt->items[pieceIx].offset.y;
						
						piecePos.x += txt->items[pieceIx].sup.x;
						piecePos.y += txt->items[pieceIx].sup.y;
						if ( txt->items[pieceIx].charcode == '\t' )
							piecePos.x += txt->items[pieceIx].advance.x;
					}
				}
				else {
					charLen++;
					/* the width is being accumed for use with a
					 * show operator; this will be done under a
					 * coordinate transform to effect rotation,
					 * scale, etc., so use the untransformed width
					 */
					if (!pc->hidden)
						width += pc->padvance; /* padvance is untransformed */
				}
				prevSup = pc->sup;
				tcharix++;
				
				curpos.x += pc->advance.x + pc->kern.x + pc->offset.x;
				curpos.y += pc->advance.y + pc->kern.y + pc->offset.y;
			}
			if (charLen > 0) {
				/* caboose of this attr seg
				 */
				OutInt (piecePos.x);
				OutInt (piecePos.y);
				OutInt (width);
				OutTxCharString(&txt->items[pieceIx], charLen);
				if (txAp->ruling || txAp->strike_thru) {
					p32 leftpos, rightpos;
					leftpos = piecePos;
					rightpos = curpos;
					if (ix && tcharix) {
						rightpos.x += txt->items[tcharix - 1].sup.x;
						rightpos.y += txt->items[tcharix - 1].sup.y;
					}
					axtAdjustRulingEndpoints(txt, pieceIx, tcharix - pieceIx,
											 &leftpos, &rightpos,
											 &leftpos, &rightpos);
					OutInt(leftpos.x);
					OutInt(leftpos.y);
					OutInt(rightpos.x);
					OutInt(rightpos.y);
				}
			}
			charix += txAp->nChars;
			charcount -= txAp->nChars;
			txAp++;
		}

		OutWord (SfEndCall);
		
	}
	else if (txAp->indent) {
		/* no chars, but a defined indent/bullet attr; record the bullet
		 */
		seRecordBullet(txt, txAp, 0, &txt->whole.pos,
					   &bulletListSize, &bulletList, shadow_type);
	}
	
    if (bulletListSize) {
        txAttr *ba;
        r32 *bRect;

        /* draw the bullets
         */
        for (ix = 0; ix < bulletListSize; ix++) {
            ba = bulletList[ix].txAp;
			bRect =  &bulletList[ix].bRect;
            ba->indent->bullet->bulletDraw(ba->indent,
                                           ba->fontAttr.ptsizeInt,
                                           ba->fillpattern,
										   bRect, shadow_type, FALSE);
			free(ba);
        }
        free(bulletList);
    }

    return(TRUE);
}

/*-----------------------------------------------------------------------------
	seText
-----------------------------------------------------------------------------*/
static void 
seRecordBullet(txt, txAp, charix, charpos, listSize, bulletList, shadow_type)
axtext *txt;                      /* in */
txAttr *txAp;                     /* in */
int charix;                       /* in */
p32 *charpos;                     /* in */
int *listSize;                    /* in/out */
bulletCall **bulletList;          /* in/out */
int shadow_type;
{
    r32 bRect;
    bulletCall *bList;
    int tsize = *listSize;
	int sfg, sdx, sdy, no_prnt;
    color_type new_color;
	txAttr *newAp;
        
    if ( !txAp->indent )
        return;
    if ( !axtHasBullet(txt, txAp, &txt->items[charix], charix) )
        return;
    if ( !axtNeedBullet(txt, txAp->indent, charix) )
        return;
    axtBulletRect(txt, txAp, charpos, &bRect);
    if (!bRect.width || !bRect.height)
        return;

	/* add the bullet to the passed display list
	 */
	newAp = (txAttr *)malloc(sizeof(txAttr));
	memcpy(newAp, txAp, sizeof(txAttr));
	if ( shadow_type == 2 ) {
		no_prnt = axtGetShadowDisplayInfo(txt, txAp, shadow_type);
		if ( no_prnt ) {
			sdx = 0;
			sdy = 0;
		}
		else {
			sdx = txAp->fillpattern->s_xoff;
			sdy = txAp->fillpattern->s_yoff;
			newAp->fillpattern->ffg = txAp->fillpattern->sfg;
			newAp->fillpattern->fbg = txAp->fillpattern->sfg;
			newAp->fillpattern->lfg = txAp->fillpattern->sfg;
			newAp->fillpattern->lbg = txAp->fillpattern->sfg;
			sfg = 0;
			sfg |= newAp->fillpattern->ffg.c << 24;
			sfg |= newAp->fillpattern->ffg.m << 16;
			sfg |= newAp->fillpattern->ffg.y << 8;
			sfg |= newAp->fillpattern->ffg.k;
			newAp->fillcolor = sfg;
		}
	}
	else if ( shadow_type == 1 ) {
		no_prnt = axtGetShadowDisplayInfo(txt, txAp, shadow_type);
		if ( no_prnt ) {
			sdx = 0;
			sdy = 0;
		}
		else if ( txAp->fillpattern->s_type < 3 ) {
			sdx = 0;
			sdy = 0;
		}
		else {
			sdx = (-1)*txAp->fillpattern->s_xoff;
			sdy = (-1)*txAp->fillpattern->s_yoff;
			GetWashoutColor(&txAp->fillpattern->sfg, &new_color);
			newAp->fillpattern->sfg.c = new_color.c;
			newAp->fillpattern->sfg.m = new_color.m;
			newAp->fillpattern->sfg.y = new_color.y;
			newAp->fillpattern->sfg.k = new_color.k;
			
			newAp->fillpattern->ffg = newAp->fillpattern->sfg;
			newAp->fillpattern->fbg = newAp->fillpattern->sfg;
			newAp->fillpattern->lfg = newAp->fillpattern->sfg;
			newAp->fillpattern->lbg = newAp->fillpattern->sfg;
			
			sfg = 0;
			sfg |= newAp->fillpattern->ffg.c << 24;
			sfg |= newAp->fillpattern->ffg.m << 16;
			sfg |= newAp->fillpattern->ffg.y << 8;
			sfg |= newAp->fillpattern->ffg.k;
			newAp->fillcolor = sfg;
			
		}
	}
	else {
		sdx = 0;
		sdy = 0;
	}
	*bulletList = (bulletCall *)
		realloc(*bulletList, (tsize + 1) * sizeof(bulletCall));
	bList = *bulletList;
	
	bList[tsize].bRect.origin.x = bRect.origin.x + sdx;
	bList[tsize].bRect.origin.y = bRect.origin.y + sdx;
	bList[tsize].txAp = newAp;
	
	(*listSize)++;
}

/*-----------------------------------------------------------------------------
    Write out the attr info for one mono-attribute segment
    of an axtext instance
-----------------------------------------------------------------------------*/
static void OutTextAttrs(txt, pos, len, fontname, underpos, underthick, strikepos)
axtext *txt;
int pos, len;
char *fontname;
int underpos;
int underthick;
int strikepos;
{
    textAttrs *textProps;
    int attlen, tlen; 
    if(!getTextAttrs(txt, pos, len, &textProps, &attlen)) {
            /* use the internal hardcoded defaults */
        textProps = (textAttrs *) malloc (sizeof(textAttrs));
        memset((char *)textProps, 0, sizeof(textAttrs));
        getHardcodedDefaults(textProps, 0);
    }
    if (textProps->fontFamily) {
        tlen = strlen(textProps->fontFamily);
        OutInt(tlen);
        OutCString(textProps->fontFamily, tlen);
    }
    else {
        OutInt(0);
    }
    OutInt(textProps->weight);
    OutInt(textProps->slant);
    OutInt(textProps->pointsize);
    OutInt(textProps->ruling);
    OutInt(textProps->strike_thru);
    OutInt(textProps->supx);
    OutInt(textProps->supy);
    OutInt(textProps->kernx);
    OutInt(textProps->kerny);
    OutWord(textProps->noRuleChars[0]);
    OutWord(textProps->noRuleChars[1]);
    OutInt(textProps->angle);
    OutInt(textProps->xscale);
    OutInt(textProps->yscale);
    OutInt(textProps->xshear);
    OutInt(textProps->fillcolor);
    OutInt(textProps->fillbcolor);
    OutInt(textProps->strokecolor);
    OutInt(textProps->strokebcolor);
    if (textProps->fillpattern) {
        OutInt(1);  /* bool: "there is a fillpattern" */
        se_write_poly_attr(textProps->fillpattern);
    }
    else {
        OutInt(0);
    }
        /* include the (derived) font name attribute */
    tlen = strlen(fontname);
    OutInt(tlen);
    OutCString(fontname, tlen);
        /* as well as underline dimensions */
    OutInt(underpos);
    OutInt(underthick);
    OutInt(strikepos);
    freeTextAttrs(textProps, attlen);
}

/*-----------------------------------------------------------------------------
    Write out the whole-field properties.
    Currently, this is just the field transform, plus zoom.
-----------------------------------------------------------------------------*/
static void OutTxFieldProps(txt)
axtext *txt;
{
    textAttrs fieldProps;
    int xZoom, yZoom;
    void axtGetFieldTransform();

    axtGetFieldTransform(txt, &fieldProps);
    axtGetZoom(txt, &xZoom, &yZoom);
    OutInt(fieldProps.angle);
    OutInt(fieldProps.xscale);
    OutInt(fieldProps.yscale);
    OutInt(fieldProps.xshear);
    OutInt(xZoom);
    OutInt(yZoom);
    OutInt(txt->draw_shadow);
    OutInt(txt->shadow_only);
    OutInt(txt->draw_emboss);
}

/*-----------------------------------------------------------------------------
    SEunlink() removes the serial file.
-----------------------------------------------------------------------------*/
void SEunlink()
{
    if (SEfdes != -1)
    {
        close(SEfdes);
        SEfdes = -1;
    }
    if (SEfname)
    {
        Adelete(SEfname);
        SEfname = CNULL;
    }
}

/*-----------------------------------------------------------------------------
    se_save  - saves current graphics state 
-----------------------------------------------------------------------------*/
void se_save()
{

    OutWord(SfSave);
    OutWord (SfEndCall);

}

/*-----------------------------------------------------------------------------
    se_restore - restores last saved graphics state 
-----------------------------------------------------------------------------*/
void se_restore()
{
    OutWord(SfRestore);
    OutWord (SfEndCall);
}


/*-----------------------------------------------------------------------------
    OutPrelude writes the standard prelude into the output buffer.
-----------------------------------------------------------------------------*/
static void OutPrelude ()
{
    OutNString (SE_PRL_V4);
    OutByte ('\n');
    OutNString (SE_PRL2);
    OutByte ('\n');
    OutNString (SE_PRL3);
    OutByte ('\n');
    OutAlign ();
}

/*-----------------------------------------------------------------------------
    Outbyte writes a byte into the output buffer.
-----------------------------------------------------------------------------*/
static void OutByte(b)
char b;
{
    if (bufidx >= BUFFER_SIZE)
    {
        WriteIt ();
        bufidx = 0;
    }
    if(buf)
        buf[bufidx++]=b;
}

/*-----------------------------------------------------------------------------
    OutWord write a word.
-----------------------------------------------------------------------------*/
static void OutWord (ww)
int ww;
{
    OutAlign ();
    OutByte (ww >> 8);
    OutByte (ww);
}

/*-----------------------------------------------------------------------------
    OutLong writes the bottom 32 bits of a long
-----------------------------------------------------------------------------*/
static void OutLong(lx)
long lx;
{
    OutWord ((int) (lx >> 16));
    OutWord ((int) lx);
}

/*-----------------------------------------------------------------------------
    same thing, but explicitly for ints
-----------------------------------------------------------------------------*/
static void OutInt (ix)
int ix;
{
    OutWord ((int) (ix >> 16));
    OutWord ((int) ix);
}

/*-----------------------------------------------------------------------------
    OutNString writes the non-null characters of a null terminated string.
-----------------------------------------------------------------------------*/
static void OutNString (s)
char *s;
{
    while (*s) OutByte (*s++);
}

/*-----------------------------------------------------------------------------
    This strange function always writes an odd number of
    bytes; if the passed string has an even strlen, then
    it will write a null to the output stream, otherwise
    not.  This was not the original intent at all, but it
    matches its correspondent, pfinf.c:InString()
-----------------------------------------------------------------------------*/
static void OutString (s)
char *s;
{
    register int    i=1;

    while (*s) { OutByte (*s++); i++; }
    if (i & 1) OutByte(0);
}

/*-----------------------------------------------------------------------------
    OutCString writes a string of len characters.
-----------------------------------------------------------------------------*/
static void OutCString (s, len)
char *s;
register int len;
{
    if (s == NULL)
    {
        while (len-- > 0)
            OutByte(0) ;
    }
    else {
        while (len-- > 0)
            OutByte (*s++);
    }
}

/*-----------------------------------------------------------------------------
    OutTxCharString - writes out a counted
    string of tx_char's. No null.
-----------------------------------------------------------------------------*/
static void OutTxCharString(pc, len)
register perChar *pc;
register int len;
{
    register int ix;
    register int tlen;
    
    for (ix = 0, tlen = len; ix < len; ix++)
    {
        if ((int) pc[ix].charcode == NEWLINE || pc[ix].hidden)
        {
                /* filter newlines and hidden chars.  note that
                    the receiving code expects each TxCharString
                    to be ready for direct setting, so it's assumed
                    that newlines and hidden characters occur in a
                    fashion that won't break the assumption, e.g.,
                    always at the beginning or ends of lines, or
                    making up whole lines */
            tlen--;
        }
    }
    OutInt (tlen);
    if (sizeof(tx_char) == 1)
    {
        for (ix = 0; ix < len; ix++)
        {
            if ((int) pc[ix].charcode != NEWLINE && !pc[ix].hidden)
                OutByte((char) pc[ix].charcode);
        }
    }
    else
    {
        for (ix = 0; ix < len; ix++)
        {
            if ((int) pc[ix].charcode != NEWLINE && !pc[ix].hidden)
                OutWord((int) pc[ix].charcode);
        }
    }
}

/*-----------------------------------------------------------------------------
    OutAlign ensures that the next byte will appear on a word boundary.
-----------------------------------------------------------------------------*/
static void OutAlign ()
{
    if (bufidx&1)
		OutByte(0);
}

/*-----------------------------------------------------------------------------
    WriteIt actually writes the buffer; everything else just
    loads the buffer.
-----------------------------------------------------------------------------*/
static void WriteIt()
{
    int cnt;

    if (SEfdes != -1 && buf) {
        cnt = write(SEfdes, buf, bufidx);
        if ( cnt != bufidx )
		{
			SETermStatus = ErrnoErr() ;
			SEunlink() ;
#if 0
            exit(ErrnoErr());
#endif
	    }
    }
}

/*-----------------------------------------------------------------------------
    TmGMode function
-----------------------------------------------------------------------------*/
static void se_gmode(mes, mesLen)
char *mes;
short mesLen;
{
    union 
    {
        struct SGModeData g_mode_data;
        char rec_buf[sizeof(struct SGModeData)];
    } rec_pack;

    memcpy(rec_pack.rec_buf,mes,mesLen);

    OutWord (SfGMode);
    OutWord (rec_pack.g_mode_data.graphics_on);
    OutWord (rec_pack.g_mode_data.grect.x_position);
    OutWord (rec_pack.g_mode_data.grect.y_position);
    OutWord (rec_pack.g_mode_data.grect.x_size);
    OutWord (rec_pack.g_mode_data.grect.y_size);
    OutWord (SfEndCall);
}

/*-----------------------------------------------------------------------------
    TmCreate function
-----------------------------------------------------------------------------*/
void se_newpage(landscape, pageWid, pageHt, leftMargin, use_crops, use_regis, use_bounds,
			bounds, clip,
            in_tray, out_tray, duplex)
int landscape;
int pageHt, pageWid;
int leftMargin;
int use_crops, use_regis, use_bounds;
TmRect32 *bounds, *clip;
int in_tray, out_tray, duplex ;
{
    OutWord (SfCreate);
    OutWord (landscape);
    OutInt (pageWid);
    OutInt (pageHt);
    OutInt (leftMargin);
    OutInt (use_crops);
    OutInt (use_regis);
    OutInt (use_bounds);
    OutCString((char *)bounds, sizeof(TmRect32)) ;
    OutCString((char *)clip, sizeof(TmRect32)) ;
    OutInt (in_tray);
    OutInt (out_tray);
    OutInt (duplex);
#if 0
    if ( bounds )
    {
        OutInt (bounds->x_position);
        OutInt (bounds->y_position);
        OutInt (bounds->x_size);
        OutInt (bounds->y_size);
    }
    else
    {
        OutInt (0);
        OutInt (0);
        OutInt (0);
        OutInt (0);
    }
#endif
    OutWord (SfEndCall);

    page.appClip.origin.x = 0;
    page.appClip.origin.y = 0;
    page.appClip.corner.x = pageWid;
    page.appClip.corner.y = pageHt;
}


/*-----------------------------------------------------------------------------
    page_init -- initialize page descriptor
    Fill in default values, and set up clipping rectangles.
    Assumes correct device characteristics have been loaded.
-----------------------------------------------------------------------------*/
static void page_init()
{
    page.trans_mode = TMS;
    page.spWid = -1;
    page.lin_des.line_style = TM_SOLID;
    page.lin_des.line_thickness = TM_VTHIN;

	/* set up clipping rectangles */

    page.inner.origin.x = page.inner.origin.y = 0;
    page.inner.corner.x = 8500; /* default */
    page.inner.corner.y = 11000;    /* default */
    page.appClip = page.inner;
}

/*-----------------------------------------------------------------------------
    TmTerm function - normal termination
    If the DC should crash during formatting, the SE never gets terminated 
    this way.
-----------------------------------------------------------------------------*/
typeErr SEterm()
{
    if (SEfdes != -1)
	{
        OutWord(SfTerm);
        OutWord(SfEndCall);
        WriteIt();
        if (close(SEfdes) == -1)
            SETermStatus = ErrnoErr() ;
		SEfdes = -1;
    }
	return SETermStatus ;
}



