/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: whereami.c
   Applix ASC version: 17
   Date: Thu Apr 22 14:36:47 1999
*/
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <axsocket.h>
#include <errno.h>
#include <signal.h>
#include <setjmp.h>
#include <aos.h>
#include <afile.h>

#if AXMS
#include <winmap.h>

char *WhereAmI (char *pname)
{
    char path_buff[STRPATH+1];
    int ret;
    char *retP;
    char *p;

    ret = MS_GetModuleFileName( (HMODULE)NULL, path_buff, sizeof(path_buff) );
    if ( ret == 0 )
        {
        ret = MS_GetLastError();
        printf("WhereAmI - GetModuleFileName fails, error = %d\n", ret );
        return( (char *)NULL );
        }

    /* strip off the executable name to leave just the containing dir */
    p = path_buff + ret - 1;
    while ( *p != '\\' )
       { p--; }
    *p = '\0';

    retP = malloc(strlen(path_buff) + 1);
    strcpy( retP, path_buff );
    return( retP );
}
  


#else                       /* }   {  */
    
void ParsePath();

static char CurrentDir[STRPATH+1];

int AxNoSymbolicFileLinks;	/* For HP CD/ROM */

/*************************************************************************
function: WhereAmI - Returns the directory containing the thimble program
*/
char *WhereAmI (pname)
register char *pname;	/* prog name as appears in main's *argv[0] */
{
	char path[STRPATH+1], one_name[STRPATH+1], *getenv();
	register char *p;
	register int ix;
	struct stat status;
	char dir[STRPATH+1];
	char entryname[STRPATH+1];

	CurrentDir[0] = 0;
#if X386_NEW || HP300 || HP800 || M_88K || CLIX || SYSV4 || AXMS || SGI || NEC_EWS || LINUX
	if ( !getcwd(CurrentDir, STRPATH+1) ) {
		perror(NXLT("Current directory"));
		exit(1);
	}
#else
	if ( !getwd(CurrentDir) ) {
		printf(NXLT("Current directory: %s\n"),CurrentDir);
		exit(1);
	}
#endif
	if ( strlen(CurrentDir) > STRPATH-50 ) {
		printf(NXLT("Current directory's name is too long, cannot start.\n"));
		exit(1);
	}

/* Produces a full pathname of where the Thimble executables lie */

	if ( *pname == '/' || *pname == '.') {	/* already absolute path */
		strcpy(path,pname);
		goto wrap_up;
	}
	p = getenv(NXLT("PATH"));
	if ( !p || !(*p) ) {
		printf(NXLT("Could not access environment variable PATH\n"));
		exit(1);
	}
	while ( TRUE ) {
		ix = 0;
		while ( *p && (*p != ':') )
			one_name[ix++] = *p++;
		one_name[ix] = 0;		/* null term it */
		if ( strlen(one_name) ) {	/* non-empty */
			sprintf(path,"%s/%s",one_name,pname);
			if ((0 == access(path,X_OK)) &&
				(0 == stat(path,&status)) &&
				((status.st_mode & S_IFDIR) != S_IFDIR) )
				goto wrap_up;
		}
		else {			/* empty path, use current dir */
			sprintf(path,"%s/%s",CurrentDir,pname);
			if ((0 == access(path,X_OK)) &&
				(0 == stat(path,&status)) &&
				((status.st_mode & S_IFDIR) != S_IFDIR) ) {
				goto wrap_up;
			}
		}
		if ( *p == 0 ) {
			printf(NXLT("Could not find %s in $PATH (%s)\n"),pname,getenv(NXLT("PATH")));
			exit(1);
		}
		p++;	/* get beyond the current ':' */
	}
wrap_up:
	AbsolutifyPathname(dir,path,TRUE);	/* resolve relative refs, links */
	ParsePath(dir,dir,entryname);
	return(strcpy(malloc(strlen(dir)+1),dir));
}

/*************************************************************************
function: AbsolutifyPathname - Fix relative and symbolic paths
Src/Dst allowed to be same address.
*/
AbsolutifyPathname (dst,src,expand_last)
char *dst;	/* returned fully absolute path */
char *src;
int expand_last;	/* TRUE if should expand entryname */
{
	char *dstbegins = dst;
	int got_slash = FALSE;
	int dot_count = 0;
	char buf[STRPATH+1];

	if ( src == dst ) {
		strcpy(buf,src);
		src = buf;
	}
	TruePath(dst,src,expand_last);
	for ( src = dst; *src; src++ ) {
		if ( *src == '/' ) {
			if ( got_slash ) {
				if ( dot_count == 0 )	/* remove / */
					dst--;
				else if ( dot_count == 1 )	/* remove /. */
					dst -= 2;
				else if ( dot_count == 2 ) {	/* remove /word/.. */
					dst -= 4;
					while ( (dst > dstbegins) && (*dst != '/') )
						dst--;
					if ( dst < dstbegins )	/* sanity */
						dst = dstbegins;
				}
			}
			got_slash = TRUE;
			dot_count = 0;
		}
		else if ( got_slash && *src == '.' )
			dot_count++;
		else {
			got_slash = FALSE;
			dot_count = 0;
		}
		*dst++ = *src;
	}
	*dst = 0;
}

/*************************************************************************
function: TruePath - Convert soft link to full pathnames
*/
TruePath (dst,src,convert_last)
char *dst, *src;
int convert_last;	/* TRUE/FALSE */
{
	char *dstbuf = dst;
	char inbuf[STRPATH+1];
	int len;
	int expanded = FALSE;	/* TRUE if a link expansion occurred */

again:
	strcpy(inbuf,src);
	src = inbuf;

	if ( *src != '/' )
		sprintf(dstbuf,"%s/",CurrentDir);
	else {
		strcpy(dstbuf,"/");
		src++;
	}

/* trim off trailing / */

	while ( inbuf[(len = strlen(inbuf))-1] == '/' ) {
		if ( len == 1 )		/* case of src == '/' */
			return;
		inbuf[len-1] = 0;
	}

/* build up path */

	len = strlen(dstbuf);
	while ( *src ) {
		if ( *src == '/' ) {
			dst[len] = 0;		/* null term the partial path */
			if ( ExpandLink(dstbuf) ) {
				len = strlen(dstbuf);
				expanded = TRUE;
			}
			dst[len++] = '/';
			src++;
		}
		else
			dst[len++] = *src++;
	}
	dst[len] = 0;
	
	if ( convert_last ) {
		if ( ExpandLink(dstbuf) )
			expanded = TRUE;
	}
/* Retry expansions until there are not expansions.
	This is needed to handle the case of /tmp being a link to /var/tmp
	which is a link to /usr/var/tmp.  Without the subsequent expansions
	the /var is never expanded to /usr/var! */

	if ( expanded ) {
		expanded = FALSE;
		src = dstbuf;
		goto again;
	}
	return;
}

/*************************************************************************
function: ExpandLink - Expand a symblic link
*/
ExpandLink (name)
char *name;
{
	char newbuf[STRPATH+1];
	int len;
	int retval = FALSE;
	extern int AxNoSymbolicFileLinks;

	if ( AxNoSymbolicFileLinks )	/* For HP CD/ROM */
		return(FALSE);

again:
#if !X386_NEW && !AXMS
	if ( -1 == (len = readlink(name,newbuf,STRPATH)) )
		return(retval);
#else
    return(retval);
#endif
	retval = TRUE;
	newbuf[len] = 0;

/* clip trailing /'s from link */

	while ( newbuf[(len = strlen(newbuf))-1] == '/' )
		newbuf[len-1] = 0;

/* replace if leading /, else append */

	if ( newbuf[0] == '/' )
		strcpy(name,newbuf);
	else {
		char *p, *strrchr();
		p = strrchr(name,'/');	/* get rid of old entryname */
		if ( !p )
			strcpy(name,newbuf);	/* best guess */
		else
			strcpy(p+1,newbuf);
	}
	goto again;
}
#endif                    /*  }  AX_X11  */


/*************************************************************************
function: ParsePath - Split a pathname into dir & entry names
The pathname and the dirname portion are allowed to be the same address.
*/
void ParsePath (path,dir,entryname)
char *path;
char *dir;
char *entryname;
{
	register char *p;

	if ( path != dir )
		strcpy(dir,path);
	for ( p = dir + strlen(dir) -1; p >= dir; p-- ) {
#if AXMS
		if ( *p == '/' || *p == '\\' ) {
#else
		if ( *p == '/' ) {
#endif
			strcpy(entryname,p+1);
			*p = 0;
			if ( p == dir )
				strcpy(dir,"/");
			return;
		}
	}
	strcpy(entryname,dir);
	dir[0] = 0;
}


