/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: tm1files.c
   Applix ASC version: 16
   Date: Tue Mar  9 21:12:21 1999
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#define AX_USE_PROTOS 1
#include <aos.h>
#include <elfapi.h>
#include <tm1files.h>

/***************/
/* Definitions */
/***************/

#define FALSE 0
#define TRUE  1

#if (SUN4S && !SUNIS) || LINUX
/* sanjay -- keep in sync will ALLOW_VIO in r6_sys_sf.c */
#define ALLOW_VIO 1
#else
#define ALLOW_VIO 0
#endif

#if ALLOW_VIO
extern char *vio_gets(char *buf, int bufsiz, int fd);
extern int vio_puts(char *buf, int fd);
extern int vio_fopen(char *name, char *modestr);
extern int vio_fclose(int fd);
extern int vio_tmpfile();
extern int vio_rewind(int fd);
extern int vio_access(char *name, int mode);
extern char **vio_listfiles(char *dirname);
extern void vio_listfilesfreer(char **files);
extern int  vio_do_rename (char *oldname, char *newname);
#define FILE_OPEN_ERR (-1)
#define FILE_PTR int
#define fopen  vio_fopen
#define fclose vio_fclose
#define fgets  vio_gets
#define fputs  vio_puts
#define tmpfile vio_tmpfile
#define rewind  vio_rewind
#define access  vio_access
#define rename  vio_do_rename
#else
#define FILE_OPEN_ERR (NULL)
typedef FILE *FILE_PTR;
#endif /* not ALLOW_VIO */

/**********************************/
/* Prototypes of Local Procedures */
/**********************************/

data_ptr
cda_file_write_substructure(char *file_with_path, char *startcharstr,
			    char *specific_instance, data_ptr twoDarray,
			    int file_exists, int replace);
data_ptr
make_file_error_return(char *errstr);
data_ptr
make_file_okay_return(data_ptr rData);
void
write_current_line(char *readbuffer,
		   FILE_PTR newtm1file, 
		   int copying_file, 
		   char *sptr);
void
get_next_line(char **out_cptr, 
	      char *readbuffer, 
	      FILE_PTR tm1file, 
	      char **out_sptr);
void
write_2D_array(data_ptr twoDarray,
	       FILE_PTR newtm1file);

/**************************/
/* File Reading Functions */
/**************************/

/**************************************************************************
 * do_parse_cda_for_charlines - opens a file looking for all lines that
 *     begin with the specified character; arguments are:
 *       [0] = filename, with complete pathname
 *       [1] = string of a single character, representing the header char
 */
data_ptr
do_parse_cda_for_charlines(data_ptr funcData)
{
  char file_with_path[1025];
  FILE_PTR tm1file;
  char buffer[512];
  char errstr[512];
  char *cptr;
  char *sptr;
  data_ptr rData, retData;
  int asize;
  char currchar, startchar;

  if (AxArraySize(funcData) < 2) {
    return(make_file_error_return(
				  "Insufficient number of arguments to file parsing procedure"));
  }
  if (! AxIsString(AxArrayElement(funcData,0))) {
    return(make_file_error_return("First argument is not a string"));
  }
  cptr = AxStrFromArray(funcData,0);
  strcpy(file_with_path,cptr);

  if (! AxIsString(AxArrayElement(funcData,1))) {
    return(make_file_error_return("Second argument is not a string"));
  }
  cptr = AxStrFromArray(funcData,1);
  startchar = *cptr;

  tm1file = fopen(file_with_path, "r");
  if (FILE_OPEN_ERR == tm1file) {
    sprintf(errstr, "Cannot open file %s to read", file_with_path);
    return(make_file_error_return(errstr));
  }
  asize = 0;
  get_next_line(&cptr, buffer, tm1file, &sptr);
  rData = AxMakeArray(0);
  while (NULL != cptr) {
    if (buffer[0] == startchar) {
      /* skip over the initial character and comma to look at the
         balance of the line */
      if ('\0' == buffer[1]) {
	cptr = NULL;
      } else {
	cptr = &(buffer[2]);

	/* get rid of any commas, carriage returns or double quotes */
	cptr = strtok(cptr, ",\n\"");

	/* add the value to the return data array */
	rData = AxAddStrToArray(rData, asize++, cptr);
      }
    }

    /* get the next text line from the file */
    get_next_line(&cptr, buffer, tm1file, &sptr);
  }
  fclose(tm1file);

  retData = make_file_okay_return(rData);
  return(retData);
} /* do_parse_cda_for_charlines */

/**************************************************************************
 * do_parse_cda_for_substructure - opens a file looking for a line that
 *    begins with the specified character, and whose balance is the
 *    specified string (note that this comparison is *NOT*
			*    case-sensitive).  If such an entry is found, the contents of 
			*    the file between this entry and the next is parsed into a 2D
			*    (line, element) array.  The header line may also be parsed into
			*    this array and returned.  The arguments are:
			*        [0] = filename, with complete pathname
			*        [1] = string containing start character
			*        [2] = string of the specific instance (subset, view, query)
			*              we're looking for 
 *        [3] = optional, flag for whether to return the header line
 *              along with the rest of the substructure; default to
 *              FALSE
 */
data_ptr
do_parse_cda_for_substructure(data_ptr funcData)
{
  char file_with_path[1025];
  char specific_instance[512];
  int return_header_line;
  FILE_PTR tm1file;
  char buffer[512];
  char errstr[512];
  char *cptr;
  char *sptr;
  char *sptr2;
  data_ptr rData, tmpData, retData;
  int asize, bsize, ix, isnum, slen, tok_count, previous_was_num;
  char currchar, startchar;

  /* check basic argument correctness */
  if (AxArraySize(funcData) < 3) {
    return(make_file_error_return(
      "Insufficient number of arguments to file entry parsing procedure"));
  }

  /* get the filename */
  if (! AxIsString(AxArrayElement(funcData,0))) {
    return(make_file_error_return("First argument is not a string"));
  }
  cptr = AxStrFromArray(funcData,0);
  strcpy(file_with_path, cptr);

  /* get the string containing the header character */
  if (! AxIsString(AxArrayElement(funcData,1))) {
    return(make_file_error_return("Second argument is not a string"));
  }
  cptr = AxStrFromArray(funcData,1);
  startchar = *cptr;

  /* get the string containing the specific instance we're looking for */
  if (! AxIsString(AxArrayElement(funcData,2))) {
    return(make_file_error_return("Third argument is not a string"));
  }
  cptr = AxStrFromArray(funcData,2);
  strcpy(specific_instance,cptr);
  
  /* get the flag indicating whether or not to return the header line
     as well */
  if (AxArraySize(funcData) > 3) {
    if (! AxIsInt(AxArrayElement(funcData,3))) {
      return(make_file_error_return("Fourth argument is not an integer"));
    }
    return_header_line = TRUE && AxIntFromArray(funcData,3);
  } else {
    return_header_line = FALSE;
  }
  
  /* open the file for reading */
  tm1file = fopen(file_with_path, "r");
  if (FILE_OPEN_ERR == tm1file) {
    sprintf(errstr, "Cannot open file %s to read", file_with_path);
    return(make_file_error_return(errstr));
  }
  
  /* setup the initial part of the return value */
  asize = 0;
  rData = AxMakeArray(0);
  
  /* start reading in the lines */
  get_next_line(&cptr, buffer, tm1file, &sptr);
  
  /* as long as we have a valid line */
  slen = strlen(specific_instance);
  while (NULL != cptr) {
    /* check that it's the correct type of line */
    if (buffer[0] == startchar) {
      if (',' != buffer[1]) {
	goto do_get_next_line;
      }

      /* get the string following the "%c," */
      cptr = &(buffer[2]);
      if ((0 == strncasecmp(cptr, specific_instance, sizeof(char) * slen)) &&
	  ((',' == buffer[slen+2]) || ('\0' == buffer[slen+2]) ||
	   ('\n' == buffer[slen+2]))) {
	/* this is the instance we're looking for */

	/* check to see whether this line should be returned */
	if (! return_header_line) {
	  get_next_line(&cptr, buffer, tm1file, &sptr);
	}

	/* iterate over the comma-delimited elements of this line */
	while (NULL != cptr) {
	  /* setup the initial part of the return value for this line */
	  bsize = 0;
	  tmpData = AxMakeArray(0);

	  /* parse out the commas */
	  sptr = buffer;
	  cptr = strtok(sptr, ",\n");
	  tok_count = 0;

	  /* add each token (without comma) to the array */
	  while (NULL != cptr) {
	    /* check if the current token can be a numeric value */
	    if ((1 == tok_count) && (! previous_was_num)) {
	      /* if we're looking at the second element of a line, and
		 the previous element wasn't a number, do not try to
		 interpret this one as a number either */
	      isnum = FALSE;
	    } else {
	      /* check to see if the current token can be a number */
	      isnum = TRUE;
	      sptr2 = cptr;
	      for (ix=0; (ix < strlen(sptr)); ix++) {
		if (((*sptr2 > '9') || (*sptr2 < '0')) && 
		    (*sptr2 != '-') &&
		    (*sptr2 != '.')) {
		  isnum = FALSE;
		  break;
		}
		sptr2++;
	      }
	    }

	    /* add the appropriate type of element to the return value
	       for this line */
	    if (isnum) {
	      ix = atoi(cptr);
	      tmpData = AxAddIntToArray(tmpData, bsize++, ix);
	      previous_was_num = TRUE;
	    } else {
	      if (0 == strcmp(cptr,"\"\"")) {
		bsize ++;
	      } else {
		tmpData = AxAddStrToArray(tmpData, bsize++, cptr);
	      }
	      previous_was_num = FALSE;
	    }

	    /* get the next element */
	    cptr = strtok(NULL, ",\n");
	    tok_count++;
	  } /* while (NULL != cptr) for elements */

	  rData = AxAddArrayToArray(rData, asize++, tmpData);
	  get_next_line(&cptr, buffer, tm1file, &sptr);
	  if (buffer[0] == startchar) {
	    goto after_loop;
	  }
	} /* while (NULL != cptr) for lines */
	goto after_loop;
      } /* if (0 == strcasecmp(cptr, specific_instance)) */
    } /* if (buffer[0] == startchar) */
    /* if we haven't found the specified instance yet */
  do_get_next_line:
    get_next_line(&cptr, buffer, tm1file, &sptr);
  } /* while (NULL != cptr) for file lines */
  
 after_loop:
  fclose(tm1file);
  retData = make_file_okay_return(rData);
  return(retData);
} /* do_parse_cda_for_substructure */

/***********************************/
/* Specific File Writing Functions */
/***********************************/

/**************************************************************************
 * do_remove_substructure_in_cda - removes the given entry from the given
 *     file; arguments are:
 *          [0] = filename, with complete pathname
 *          [1] = header line's starting character
 *          [2] = specific instance string
 */
data_ptr
do_remove_substructure_in_cda(data_ptr funcData)
{
  char file_with_path[1025];
  char startcharstr[2];
  char specific_instance[512];
  char *cptr;
  data_ptr retData, tData;

  /* check basic argument correctness */
  if (AxArraySize(funcData) < 3) {
    return(make_file_error_return(
      "Insufficent number of arguments to file entry removing procedure"));
  }

  /* get the filename */
  if (! AxIsString(AxArrayElement(funcData,0))) {
    return(make_file_error_return("First argument is not a string"));
  }
  cptr = AxStrFromArray(funcData,0);
  strcpy(file_with_path, cptr);

  /* get the header line starting character */
  if (! AxIsString(AxArrayElement(funcData,1))) {
    return(make_file_error_return("Second argument is not a string"));
  }
  cptr = AxStrFromArray(funcData,1);
  startcharstr[0] = *cptr;
  startcharstr[1] = '\0';
  
  /* get the specific instance */
  if (! AxIsString(AxArrayElement(funcData,2))) {
    return(make_file_error_return("Third argument is not a string"));
  }
  cptr = AxStrFromArray(funcData,2);
  strcpy(specific_instance,cptr);
  
  tData = AxMakeArray(0); /* place holder in following proc call */
  retData = cda_file_write_substructure(file_with_path, startcharstr,
					specific_instance, tData, TRUE, FALSE);
  AxFreeData(tData);
  return(retData);
} /* do_remove_substructure_in_cda */

/**************************************************************************
 * do_add_or_replace_substructure_in_cda - takes the specified file, and
 *     writes the entry given; if the entry already exists, it is
 *     replaced; otherwise, a new entry is added.  The arguments are:
 *        [0] = filename, with complete pathname
 *        [1] = start character for this file's header lines
 *        [2] = string of the specific instance (subset, view, query)
 *              we're looking for 
 *        [3] = a two-D (line, element) data_ptr array of stuff to
 *              write for this entry
 *        [4] = flag indicating whether the file already exists or not
 */
data_ptr
do_add_or_replace_substructure_in_cda(data_ptr funcData)
{
  char file_with_path[1025];
  char startcharstr[2];
  char specific_instance[512];
  data_ptr twoDarray;
  int file_exists;
  char *cptr;
  data_ptr retData;

  /* check basic argument correctness */
  if (AxArraySize(funcData) < 5) {
    return(make_file_error_return(
				  "Insufficient number of arguments to file entry writing procedure"));
  }

  /* get the filename */
  if (! AxIsString(AxArrayElement(funcData,0))) {
    return(make_file_error_return("First argument is not a string"));
  }
  cptr = AxStrFromArray(funcData,0);
  strcpy(file_with_path,cptr);

  /* get the start character */
  if (! AxIsString(AxArrayElement(funcData,1))) {
    return(make_file_error_return("Second argument is not a string"));
  }
  cptr = AxStrFromArray(funcData,1);
  startcharstr[0] = *cptr;
  startcharstr[1] = '\0';

  /* get the specific instance */
  if (! AxIsString(AxArrayElement(funcData,2))) {
    return(make_file_error_return("Third argument is not a string"));
  }
  cptr = AxStrFromArray(funcData,2);
  strcpy(specific_instance,cptr);

  /* get the two-D (line, element) array to write */
  if (! AxIsArray(AxArrayElement(funcData,3))) {
    return(make_file_error_return("Fourth argument is not an array"));
  }
  twoDarray = AxArrayFromArray(funcData,3);

  /* get the flag indicating whether the file already exists or not */
  if (! AxIsInt(AxArrayElement(funcData,4))) {
    return(make_file_error_return("Fifth argument is not an integer"));
  }
  file_exists = TRUE && AxIntFromArray(funcData,4);

  /* write the entry to the file */
  retData = cda_file_write_substructure(file_with_path, startcharstr,
					specific_instance, twoDarray, file_exists, TRUE);

  /* return the answer */
  return(retData);
} /* do_add_or_replace_substructure_in_cda */

/*********************************/
/* General File Writing Function */
/*********************************/

/**************************************************************************
 * cda_file_write_substructure - modifies the given entry in the given
 *     file; if "replace" is true, writes the given twoD (line, element)
 *     array to the given file, using the specified starting character
 *     as the header line indicator; if "replace" is false, it deletes
 *     the given entry from the file
 */
data_ptr
cda_file_write_substructure(char *file_with_path, char *startcharstr,
			    char *specific_instance, data_ptr twoDarray,
			    int file_exists, int replace)
{
  FILE_PTR tm1file;
  FILE_PTR newtm1file;
  char errstr[512];
  char readbuffer[512];
  char fmtbuffer[512];
  char *cptr;
  char *sptr;
  char startchar;
  data_ptr rData, retData;
  int slen;
  int foundit = FALSE;
  int reading_done = FALSE;
  int copying_file = TRUE;
  static int mypid = 0;
  char tmpfilename[1024 + 1];

  /* setup initial values used later */
  startchar = startcharstr[0];
  slen = strlen(specific_instance);

  /* open the temporary file to write to */
  sprintf(tmpfilename, "/tmp/%ld.%s", getpid(), startcharstr);
  newtm1file = fopen(tmpfilename, "w");
  if (FILE_OPEN_ERR == newtm1file) {
    return(make_file_error_return("Cannot open temporary file"));
  }

  /* if we need to modify an existing file */
  if (file_exists) {
    /* open the current comma-delimited ascii file to read from */
    tm1file = fopen(file_with_path, "r");
    if (FILE_OPEN_ERR == tm1file) {
      fclose(newtm1file);
      sprintf(errstr, "Cannot open file %s for reading", file_with_path);
      return(make_file_error_return(errstr));
    }

    /* start reading through the file */
    get_next_line(&cptr, readbuffer, tm1file, &sptr);

    /* for each line in the file */
    while (NULL != cptr) {
      /* check if the current line is of the correct type */
      if (readbuffer[0] == startchar) {
	/* since this starts a new entry, if we had previously stopped
	   copying a file, we should resume it here */
	copying_file = TRUE;

	if (',' != readbuffer[1]) {
	  goto do_write_current_line;
	}

	/* get the string following the "%c," */
	cptr = &(readbuffer[2]);

	/* check and see if it's the instance we're looking for */
	if ((! foundit) &&
	    (0 == strncasecmp(cptr, specific_instance, sizeof(char) * slen)) &&
	    ((',' == readbuffer[slen+2]) || ('\0' == readbuffer[slen+2]) ||
	     ('\n' == readbuffer[slen+2]))) {
	  /* if this is the instance we're looking for, remember that
	     we've found it */
	  foundit = TRUE;
	  if (replace) {
	    /* write out the contents of the 2D array here */
	    write_2D_array(twoDarray, newtm1file);
	  }
	  /* this is for the case where we're deleting an entry */
	  copying_file = FALSE;
	}
      }

    do_write_current_line:
      /* write out the current line, if necessary */
      write_current_line(readbuffer, newtm1file, copying_file, sptr);

      /* get the next input line */
      get_next_line(&cptr, readbuffer, tm1file, &sptr);
    } /* while (NULL != cptr) */
    fclose(tm1file);
  } /* if (file_exists) */

  /* if we haven't found the definition to replace, then we need to
     add it to the end of the new file */
  if (! foundit) {
    write_2D_array(twoDarray, newtm1file);
  }

  fclose(newtm1file);
  newtm1file = fopen(tmpfilename, "r");
  if (FILE_OPEN_ERR == newtm1file) {
	  return(make_file_error_return("Cannot reopen temporary file"));
  }

  /* reopen the file, reading new into old, so that it will have the 
     correct filename */
  tm1file = fopen(file_with_path, "w");
  if (FILE_OPEN_ERR == tm1file) {
    fclose(newtm1file);
    sprintf(errstr, "Cannot open file %s for writing", file_with_path);
    return(make_file_error_return(errstr));
  }
  get_next_line(&cptr, readbuffer, newtm1file, &sptr);
  while (NULL != cptr) {
    write_current_line(readbuffer, tm1file, TRUE, sptr);
    get_next_line(&cptr, readbuffer, newtm1file, &sptr);
  }

  /* close out both files; the temporary file will automatically be
     deleted */
  fclose(tm1file);
  fclose(newtm1file);
  unlink(tmpfilename); /* this is a realMode file anyway... */
  /* return a value indicating that all's well */
  rData = NULL;
  retData = make_file_okay_return(rData);
  return(retData);
} /* cda_file_write_substructure */

/***********************************************************************
function: tm1_file_exists - Return TRUE if file exists;  FALSE otherwise
NOTE: Caller makes sure that filename is a file entry inside the data
directory.
*/
int
tm1_file_exists(char *filename)
{
	int fd;
	if (!filename || !strlen(filename))
		return(FALSE);
	if (0 == access(filename, 0))
		return(TRUE);
	return(FALSE);
}

/***********************************************************************
function: tm1_file_system_files -
*/
#if ALLOW_VIO
data_ptr
tm1_file_system_files(char *dirname, char *wildcard)
{
	char **files;
	int ix, rx;
	data_ptr retP = NULL;

	files = vio_listfiles(dirname);
	if (!files)
		return(NULL);
	retP = AxMakeArray(0);
	rx = 0;
	for (ix = 0; files[ix]; ix++) {
		if (!WildMatches(files[ix], wildcard))
			continue;
		retP = AxAddStrToArray(retP, rx++, files[ix]);
	}
	vio_listfilesfreer(files);
	return(retP);
}
#else
data_ptr
tm1_file_system_files(char *dirname, char *wildcard)
{
	int a;
	a=10;
}
#endif

/***********************************************************************
function: tm1_file_rename - 
*/
int
tm1_file_rename(char *oldname, char *newname)
{
	if (!oldname || !newname)
		return;
	if (!strlen(oldname) || !strlen(newname))
		return;
	return(rename(oldname, newname));
}

int
tm1_write_ascii_lines(data_ptr lines, char *newname)
{
	FILE_PTR newfp;
	int numlines, ix;

	newfp = fopen(newname, "w");
	if (newfp == FILE_OPEN_ERR)
		return(-1);
	numlines = AxArraySize(lines);
	for (ix = 0; ix < numlines; ix++) {
		fputs(AxStrFromArray(lines, ix), newfp);
		fputs("\n", newfp);
	}
	fclose(newfp);
	return(0);
}

data_ptr
tm1_read_ascii_lines(char *file)
{
	FILE_PTR fp;
	data_ptr lines;
	int numlines;
	char buf[1024 * 2];
	char *p;

	numlines = 0;
	lines = AxMakeArray(0);

	fp = fopen(file, "r");
	if (fp == FILE_OPEN_ERR)
		return(lines);
	while (TRUE) {
		p = fgets(buf, sizeof(buf), fp);
		if (NULL == p)
			break;
		buf[strlen(buf) - 1] = '\0';
		lines = AxAddStrToArray(lines, numlines++, p);
	}
	fclose(fp);
	return(lines);
}

/**************************************************************************
 * write_2D_array - writes each (first index) element of the 2D
 *     data_ptr array to a line in the given file; each (second index)
 *     element of the 2D array is separated from the others by a comma
 */
void
write_2D_array(data_ptr twoDarray,
	       FILE_PTR newtm1file)
{
  int aval, bval, ix, jx, itmp;
  data_ptr tData1, tData2;
  char *sptr;
  char fmtbuffer[512];
  double fval;

  aval = AxArraySize(twoDarray);
  for (ix=0; (ix < aval); ix++) {
    /* the first index indicates the line number within the
       entry */
    tData1 = AxArrayFromArray(twoDarray, ix);
    bval = AxArraySize(tData1);
    for (jx=0; (jx < bval); jx++) {
      /* the second index indicates the placement of the element
	 within the comma-delimited string */
      tData2 = AxArrayElement(tData1, jx);
      if (AxIsString(tData2)) {
	sptr = AxStrFromDataPtr(tData2);
	if (0 == jx) {
	  strcpy(fmtbuffer, sptr);
	} else {
	  sprintf(fmtbuffer, "%s,%s", fmtbuffer, sptr);
	}
      } else if (AxIsInt(tData2)) {
	itmp = AxIntFromDataPtr(tData2);
	if (0 == jx) {
	  sprintf(fmtbuffer, "%d", itmp);
	} else {
	  sprintf(fmtbuffer, "%s,%d", fmtbuffer, itmp);
	}
      } else if (AxIsFloat(tData2)) {
	fval = AxFloatFromDataPtr(tData2);
	if (0 == jx) {
	  sprintf(fmtbuffer, "%f", fval);
	} else {
	  sprintf(fmtbuffer, "%s,%f", fmtbuffer, fval);
	}
      } else {
	if (0 == jx) {
	  strcpy(fmtbuffer, "Unknown Object");
	} else {
	  sprintf(fmtbuffer, "%s,%s", fmtbuffer, "Unknown Object");
	}
      }
    }
    sprintf(fmtbuffer, "%s\n", fmtbuffer);
    fputs(fmtbuffer, newtm1file);
  }
} /* write_2D_array */

/******************************/
/* Line Read/Write Procedures */
/******************************/

/**************************************************************************
 * write_current_line - write the line in "readbuffer" to the file
 *     specified by "newtm1file", but only if "copying_file" is TRUE;
 *     if "sptr" is non-NULL, that means that the line, when read in,
 *     was terminated by a '\n', and should be returned to its
 *     original, '\n'-terminated state before being written out again
 */
void
write_current_line(char *readbuffer,
		   FILE_PTR newtm1file, 
		   int copying_file, 
		   char *sptr)
{
  if (copying_file) {
    if (sptr) {
      *sptr = '\n';
    }
    fputs(readbuffer, newtm1file);
  }
} /* write_current_line */

/**************************************************************************
 * get_next_line - returns the next line of text from "tm1file" in
 *     "readbuffer"; "out_cptr" is the value returned from the fgets
 *     call (which is NULL if the end-of-file is reached); "out_sptr"
 *     is NULL if the text line was *not* terminated by a ^M or a \n,
 *     and points to the \0 if it was
 */
void
get_next_line(char **out_cptr, 
	      char *readbuffer, 
	      FILE_PTR tm1file, 
	      char **out_sptr)
{
  char *cptr;
  char *sptr;
  char *tptr;

  /* read in the next line of text from the file */
  cptr = fgets(readbuffer, 512, tm1file);

  /* if the last character is a carriage-return, null it out */
  if (strlen(readbuffer) > 0) {
    sptr = &(readbuffer[strlen(readbuffer)-1]);
    if ('\n' == *sptr) {
      *sptr = '\0';
    }
  } else {
    sptr = NULL;
  }

  /* save away the current sptr */
  tptr = sptr;

  /* see if the current last character is a ^M, and null it out, if
     necessary */
  if (strlen(readbuffer) > 0) {
    sptr = &(readbuffer[strlen(readbuffer)-1]);
    if ('\015' == *sptr) {	/* ^M */
      *sptr = '\0';
    } else {
      sptr = tptr;
    }
  }
  *out_cptr = cptr;
  *out_sptr = sptr;
} /* get_next_line */

/*****************************************/
/* Procedures for Creating Return Values */
/*****************************************/

/**************************************************************************
 * make_file_error_return - makes a return value signifying that the
 *    procedure call failed
 */
data_ptr
make_file_error_return(char *errstr)
{
  data_ptr rData;

  rData = AxMakeArray(2);
  rData = AxAddIntToArray(rData, 0, -1);
  rData = AxAddStrToArray(rData, 1, errstr);
  return(rData);
} /* make_file_error_return */

/**************************************************************************
 * make_file_okay_return - makes a return value signifying that the
 *    procedure call succeeded
 */
data_ptr
make_file_okay_return(data_ptr rData)
{
  data_ptr retData;

  retData = AxMakeArray(1);
  retData = AxAddIntToArray(retData, 0, 0);
  if (NULL != rData) {
    retData = AxAddDataToArray(retData, 1, rData);
  }
  return(retData);
} /* make_file_okay_return */
