/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: elfdump.c
   Applix ASC version: 2
   Date: Fri Jun  4 21:14:35 1999
*/
/****************************************************************************\
                                 includes
\****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

/****************************************************************************\
                                  defines
\****************************************************************************/
#define OPCODES "EXECUTBALE OPCODES"

#define GF_HDR "#Global Functions\n"
#define GF_STR1 "GLOBAL FUNCTIONS"
#define GF_STR2 "END FUNCTIONS INFO"
			
#define EF_HDR "#External Functions\n"
#define EF_STR1 "EXTERNAL LINKAGE"
#define EF_STR2 "EXTERNS TABLE"
			
#define GV_HDR "#Global Variables\n"
#define GV_STR1 "GLOBAL VARIABLES"
#define GV_STR2 "GLOBALS TABLE"
			
#define OV_HDR "#Object Variables\n"
#define OV_STR1 "OBJECTVAR VARIABLES"
#define OV_STR2 "OBJECTVARS TABLE"
			
#define SV_HDR "#System Variables\n"
#define SV_STR1 "SYSTEMVAR VARIABLES"
#define SV_STR2 "SYSTEMVARS TABLE"

#define COMMENT_OPEN "/*"
#define COMMENT_CLOSE "*/"
#define COMMENT_OPEN_LEN 2
#define COMMENT_CLOSE_LEN 2
#define COMMENT_MIN_LEN (COMMENT_OPEN_LEN+COMMENT_CLOSE_LEN)

#define TRUE 1
#define FALSE 0

#define LINE_LEN 1024

/****************************************************************************\
                             global variables
\****************************************************************************/
char line[LINE_LEN+1];

/****************************************************************************/
static char *
extract_first_comment (char *line)
{
	char *ptr, *ptr1, *ptr2;

	if ( line == NULL )
		ptr = NULL;
	else
	{
		ptr1 = strstr(line, COMMENT_OPEN);
		if ( ptr1 == NULL || strlen(ptr1) < COMMENT_MIN_LEN )
			ptr = NULL;
		else
		{
			ptr1 += COMMENT_OPEN_LEN;
			ptr2 = strstr(ptr1, COMMENT_CLOSE);
			if ( ptr2 == NULL )
				ptr = NULL;
			else
			{
				ptr = ptr1;
				*ptr2 = (char)0;
			}
		}
	}
	return(ptr);
}

/****************************************************************************/
static char *
strip_spaces (char *line)
{
	char *ptr, *ptr1, *ptr2;
	int len;

	if ( line == NULL )
		ptr = NULL;
	else
	{
		/* remove leading spaces */
		ptr1 = line;
		len = strlen(ptr1);
		ptr2 = &line[len];
		while ( ptr1 < ptr2 && isspace(*ptr1) )
			ptr1++;

		/* remove trailing spaces */
		while ( ptr1 < ptr2 && isspace(*(ptr2-1)) )
			ptr2--;

		/* finish up */
		*ptr2 = (char)0;
		ptr = strlen(ptr1) <= 0 ? NULL : ptr1;
	}
	return(ptr);
}

/****************************************************************************/
static void
parse_file (FILE *fp, char *header, char *str1, char *str2)
{
	char *comment, *content;
	int n_funcs, reading;

	/* search for the opcode comment in the file... */
	reading = TRUE;
	while ( reading && fgets(line, LINE_LEN, fp) )
	{
		comment = extract_first_comment(line);
		if ( comment != NULL )
		{
			content = strip_spaces(comment);
			if ( content && strcmp(content, OPCODES) == 0 )
			{
				printf("#\n%s#\n", header);
				reading = FALSE;
			}
		}
	}

	/* ...then find the start comment... */
	reading = TRUE;
	while ( reading && fgets(line, LINE_LEN, fp) )
	{
		comment = extract_first_comment(line);
		if ( comment != NULL )
		{
			content = strip_spaces(comment);
			if ( content && strcmp(content, str1) == 0 )
				reading = FALSE;
		}
	}

	/* ...read the "function strings" until you reach the end comment */
	reading = TRUE;
	while ( reading && fgets(line, LINE_LEN, fp) )
	{
		comment = extract_first_comment(line);
		if ( comment != NULL )
		{
			content = strip_spaces(comment);
			if ( content )
			{
				if ( strcmp(content, str2) == 0 )
					reading = FALSE;
				else
				{
					printf("%s\n", content);
					n_funcs++;	
				}
			}
		}
	}
}

/****************************************************************************/
static void
parse_file_by_code (FILE *fp, int code)
{

	switch ( code )
	{
		default:
		case 0: parse_file(fp, GF_HDR, GF_STR1, GF_STR2); break;
		case 1: parse_file(fp, EF_HDR, EF_STR1, EF_STR2); break;
		case 2: parse_file(fp, GV_HDR, GV_STR1, GV_STR2); break;
		case 3: parse_file(fp, OV_HDR, OV_STR1, OV_STR2); break;
		case 4: parse_file(fp, SV_HDR, SV_STR1, SV_STR2); break;
	}
}

/****************************************************************************/
int
main (int argc, char **argv)
{
	int r_val, n_files, n_codes, *code, ix, jx, sz;
	char **file;
	FILE *fp;
#if 0
	{
	int wait = TRUE;
	while ( wait )
		sleep(2);
	}
#endif

	/* init */
	n_files = n_codes = 0;
	file = (char **)0;
	code = (int *)0;
	r_val = 0;

	/* parse the command line options */
	for ( ix=1; ix<argc; ix++ )
	{
		if ( argv[ix][0] != '-' )
		{
			sz = (n_files+1) * sizeof(char *);
			file = (char **)realloc(file, sz);
			if ( file == NULL )
			{
				n_files = n_codes = 0;
				r_val = -1;
				break;
			}
			else
			{
				file[n_files] = argv[ix];
				n_files++;
			}
		}
		else
		{
			sz = (n_codes+1) * sizeof(int);
			code = (int *)realloc(code, sz);
			if ( code == NULL )
			{
				n_files = n_codes = 0;
				r_val = -1;
				break;
			}
			else
			{
				if ( strcmp(argv[ix], "-sv") == 0 )
					code[n_codes] = 4;
				else if ( strcmp(argv[ix], "-ov") == 0 )
					code[n_codes] = 3;
				else if ( strcmp(argv[ix], "-gv") == 0 )
					code[n_codes] = 2;
				else if ( strcmp(argv[ix], "-ef") == 0 )
					code[n_codes] = 1;
				else /* if ( strcmp(argv[ix], "-gf") == 0 ) */
					code[n_codes] = 0;
				n_codes++;
			}
		}
	}

	/* verify command line options */
	if ( n_codes == 0 && r_val == 0 )
	{
		n_codes = 5;
		code = (int *)malloc(n_codes*sizeof(int));
		for ( jx=0; jx<n_codes; jx++ )
			code[jx] = jx;
	}

	/* parse the file(s) */
	if ( n_files <= 0 )
	{
		if ( r_val == 0 )
			r_val = -1;
	}
	else
	{
		for ( ix=0; ix<n_files; ix++ )
		{
			/* open the file */
			if ( file[ix] == NULL )
				continue;
			fp = fopen(file[ix], "r");
			if ( !fp )
				r_val |= (-1);
			else
			{
				/* perform requested ops on file */
				for ( jx=0; jx<n_codes; jx++ )
				{
					parse_file_by_code(fp, code[jx]);
					rewind(fp);
				}

				/* close the file */
				fclose(fp);
			}
		}
	}
	return (r_val);
}
