/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: hex2bin.c
   Applix ASC version: 4
   Date: Tue Jan 12 21:12:34 1999
*/
#include <stdio.h>

#define TRUE 1
#define FALSE 0

static int fetchbyte(char *p, FILE *fp);
char *strchr();

char *infile, *outfile;

main(argc,argv)
int argc;
char *argv[];
{
	FILE *rfp, *wfp;
	char buf[300];
	int len;
	char byte, byte2;
	unsigned short checksum, sum = 0;

	infile = argv[2];
	outfile = argv[1];

/* open input file */

#if AXMS
	rfp = fopen(infile,"rb");    /* open in 'binary' mode */
#else
	rfp = fopen(infile,"r");
#endif
	if ( !rfp ) {
		fprintf(stderr,"Could not open input file %s\n",infile);
		exit(1);
	}

/* open output file */

#if AXMS
	wfp = fopen(outfile,"wb");
#else
	wfp = fopen(outfile,"w");
#endif
	if ( !wfp ) {
		fprintf(stderr,"Could not open output file %s\n",outfile);
		exit(1);
	}

/* fetch checksum */

	if ( !fetchbyte(&byte,rfp) || !fetchbyte(&byte2,rfp) ) {
		fprintf(stderr,"Could not read checksum %s\n",outfile);
		unlink(outfile);
		exit(1);
	}
	checksum = byte << 8;
	checksum |= byte2 & 255;

/* fetch and write data */

	while ( fetchbyte(&byte,rfp) ) {
		sum += byte & 255;
		if ( 1 != fwrite(&byte,1,1,wfp) ) {
			fprintf(stderr,"Could not write %s\n",outfile);
			unlink(outfile);
			exit(1);
		}
	}
	if ( sum != checksum ) {
		fprintf(stderr,"File corrupted - bad checksum\n",outfile);
		unlink(outfile);
		exit(1);
	}
	fclose(rfp);
	fclose(wfp);
	exit(0);
}

/*************************************************************************
function: fetchbyte - Fetch a hex byte from file, ignore non-hex lines
Real paranoid coding style!
*/
static
fetchbyte(p,fp)
char *p;
FILE *fp;
{
	static char *chars = "0123456789ABCDEF";
	int ch;
	char *posP;
	int result;

retry:
	ch = fgetc(fp);
	if ( ch == EOF )
		return(FALSE);
	if ( ch == '\n' )
		goto retry;
	posP = strchr(chars,ch);
	if ( !posP ) {		/* ignore chars till end of line */
		while ( TRUE ) {
			ch = fgetc(fp);
			if ( ch == EOF )
				return(FALSE);
			if ( ch == '\n' )
				goto retry;
		}
	}
	result = ((posP-chars) << 4);
	ch = fgetc(fp);
	posP = strchr(chars,ch);
	if ( !posP ) {
		fprintf(stderr,"File corrupted\n",outfile);
		unlink(outfile);
		exit(1);
	}
	result = result | ((posP-chars) & 15);
	*p = result;
	return(TRUE);
}
