/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: make_externs.c
   Applix ASC version: 2
   Date: Mon Jun  7 21:16:31 1999
*/
/****************************************************************************\
                                 includes
\****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

/****************************************************************************\
                                  defines
\****************************************************************************/
#define OPCODES "EXECUTBALE OPCODES"

#define EF_HDR "#External Functions\n"
#define EF_STR1 "EXTERNAL LINKAGE"
#define EF_STR2 "EXTERNS TABLE"

#define COMMENT_OPEN "/*"
#define COMMENT_CLOSE "*/"
#define COMMENT_OPEN_LEN 2
#define COMMENT_CLOSE_LEN 2
#define COMMENT_MIN_LEN (COMMENT_OPEN_LEN+COMMENT_CLOSE_LEN)

#define TRUE 1
#define FALSE 0

#define LINE_LEN 1024

/****************************************************************************\
                             global variables
\****************************************************************************/
char line[LINE_LEN+1];

/****************************************************************************/
static char *
extract_first_comment (char *line)
{
	char *ptr, *ptr1, *ptr2;

	if ( line == NULL )
		ptr = NULL;
	else
	{
		ptr1 = strstr(line, COMMENT_OPEN);
		if ( ptr1 == NULL || strlen(ptr1) < COMMENT_MIN_LEN )
			ptr = NULL;
		else
		{
			ptr1 += COMMENT_OPEN_LEN;
			ptr2 = strstr(ptr1, COMMENT_CLOSE);
			if ( ptr2 == NULL )
				ptr = NULL;
			else
			{
				ptr = ptr1;
				*ptr2 = (char)0;
			}
		}
	}
	return(ptr);
}

/****************************************************************************/
static char *
strip_spaces (char *line)
{
	char *ptr, *ptr1, *ptr2;
	int len;

	if ( line == NULL )
		ptr = NULL;
	else
	{
		/* remove leading spaces */
		ptr1 = line;
		len = strlen(ptr1);
		ptr2 = &line[len];
		while ( ptr1 < ptr2 && isspace(*ptr1) )
			ptr1++;

		/* remove trailing spaces */
		while ( ptr1 < ptr2 && isspace(*(ptr2-1)) )
			ptr2--;

		/* finish up */
		*ptr2 = (char)0;
		ptr = strlen(ptr1) <= 0 ? NULL : ptr1;
	}
	return(ptr);
}

/****************************************************************************/
static void
parse_file (FILE *fp, FILE *outfp, char *header, char *str1, char *str2)
{
	char *comment, *content;
	int n_funcs, reading;

	/* search for the opcode comment in the file... */
	reading = TRUE;
	while ( reading && fgets(line, LINE_LEN, fp) )
	{
		comment = extract_first_comment(line);
		if ( comment != NULL )
		{
			content = strip_spaces(comment);
			if ( content && strcmp(content, OPCODES) == 0 )
			{
				fprintf(outfp, "#\n%s#\n", header);
				reading = FALSE;
			}
		}
	}

	/* ...then find the start comment... */
	reading = TRUE;
	while ( reading && fgets(line, LINE_LEN, fp) )
	{
		comment = extract_first_comment(line);
		if ( comment != NULL )
		{
			content = strip_spaces(comment);
			if ( content && strcmp(content, str1) == 0 )
				reading = FALSE;
		}
	}

	/* ...read the "function strings" until you reach the end comment */
	reading = TRUE;
	while ( reading && fgets(line, LINE_LEN, fp) )
	{
		comment = extract_first_comment(line);
		if ( comment != NULL )
		{
			content = strip_spaces(comment);
			if ( content )
			{
				if ( strcmp(content, str2) == 0 )
					reading = FALSE;
				else
				{
					fprintf(outfp, "%s\n", content);
					n_funcs++;	
				}
			}
		}
	}
}

/****************************************************************************/
static void
parse_globals (FILE *fp, FILE *outfp)
{
	parse_file(fp, outfp, EF_HDR,EF_STR1,EF_STR2);
}

/****************************************************************************/
int
main (int argc, char **argv)
{
	int r_val, n_files, ix, jx, sz;
	char **file, *outfile;
	FILE *fp, *outfp;
#if 0
	{
	int wait = TRUE;
	while ( wait )
		sleep(2);
	}
#endif

	/* parse the command line options */
	if ( argc < 3 )
	{
		printf("usage: make_externs <outfile> <file list>\n");
		r_val = -1;
	}
	else
	{
		r_val = 0;
		outfile = argv[1];
		n_files = 0;
		file = (char **)0;
		for ( ix=2; ix<argc; ix++ )
		{
			sz = (n_files+1) * sizeof(char *);
			file = (char **)realloc(file, sz);
			if ( file == NULL )
			{
				printf("Memory allocation error\n");
				n_files = 0;
				r_val = -1;
				break;
			}
			else
			{
				file[n_files] = argv[ix];
				n_files++;
			}
		}

		/* open the output file */
		outfp = fopen(outfile, "w");
		if ( !outfp )
		{
			printf("Failed to open %s\n", outfile);
			r_val |= (-1);
		}
		else
		{
			/* parse the file(s) */
			for ( ix=0; ix<n_files; ix++ )
			{
				/* open the file for reading */
				if ( file[ix] == NULL )
					continue;
				fp = fopen(file[ix], "r");
				if ( !fp )
				{
					printf("Failed to open %s\n", file[ix]);
					r_val |= (-1);
					break;
				}

				parse_globals(fp, outfp);

				/* close the file */
				fclose(fp);
			}
		}
	}
	return (r_val);
}
