/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: make_imports.c
   Applix ASC version: 2
   Date: Mon Jun  7 21:16:32 1999
*/
/****************************************************************************\
                                 includes
\****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

/****************************************************************************\
                                  defines
\****************************************************************************/
#define TRUE 1
#define FALSE 0

#define LINE_LEN 1024

/****************************************************************************\
                             global variables
\****************************************************************************/
char line[LINE_LEN+1];

/****************************************************************************/
int
cmp_func (const void *input1, const void *input2)
{
	char *str1, *str2;
	str1 = *(char **)input1;
	str2 = *(char **)input2;
	return(strcmp(str1, str2));
}

/****************************************************************************/
static int
read_file (char *file, int *out_n, char ***out_array)
{
	char **array;
	int sz, n, r_val;
	FILE *fp;

	array = *out_array;
	n = *out_n;

	fp = fopen(file, "r");
	if ( fp == NULL )
		r_val = -1;
	else
	{
	        while ( fgets(line, LINE_LEN, fp) )
		{
			if ( line[0] == '#' )
				continue;

			if ( (n%100) == 0 )
			{
				sz = (n+100) * sizeof(char *);
				array = (char **)realloc(array, sz);
			}
			array[n] = (char *)malloc(strlen(line)+1);
			strcpy(array[n], line);
			n++;
		}
		fclose(fp);
		r_val = 0;
	}

	*out_n = n;
	*out_array = array;
	return (r_val);
}

/****************************************************************************/
static void
sdiff (int n_a1s,char **a1, int n_a2s,char **a2, int *out_n,char ***out_d)
{
	int r_val, n, ix, ix1, ix2, sz, res, tmp;
	char **diff;

	/* init */
	diff = NULL;
	n = 0;

	/* sort the arrays */
	qsort(a1, n_a1s, sizeof(char *), cmp_func);
	qsort(a2, n_a2s, sizeof(char *), cmp_func);

	/* compare the lists, creating a new list */
	for ( ix1=ix2=0; ix1<n_a1s && ix2<n_a2s; )
	{
		res = strcmp(a1[ix1], a2[ix2]);
		if ( res < 0 )
		{
			/* copy the string to the diff list */
			if ( (n%100) == 0 )
			{
				sz = (n+100) * sizeof(char *);
				diff = (char **)realloc(diff, sz);
			}
			diff[n] = a1[ix1++];

			/* advance list1 past the matching string(s) */
			while ( ix1<n_a1s && strcmp(diff[n], a1[ix1]) == 0 )
				ix1++;
			n++;
		}
		else if ( res == 0 )
		{
			/* advance list1 past the matching string(s) */
			while ( ix1<n_a1s && strcmp(a2[ix2], a1[ix1]) == 0 )
				ix1++;

			/* advance list2 past the matching string(s) */
			tmp = ix2;
			while ( ix2<n_a2s && strcmp(a2[tmp], a2[ix2]) == 0 )
				ix2++;
		}
		else
			ix2++;
	}
	*out_d = diff;
	*out_n = n;
}

/****************************************************************************/
static int
mod_read (char *file1, char *file2, int *out_n, char ***out_array)
{
	int r_val, n_a1s, n_a2s, n;
	char **a1, **a2, **array;

	array = a1 = a2 = NULL;
	n = n_a1s = n_a2s = 0;
	r_val = read_file(file1, &n_a1s, &a1);
	if ( r_val == 0 )
	{
		r_val = read_file(file2, &n_a2s, &a2);
		if ( r_val == 0 )
			sdiff(n_a1s,a1, n_a2s,a2, &n,&array);
	}
	*out_array = array;
	*out_n = n;
	return(r_val);
}

/****************************************************************************/
static int
elf_read (char *file1, char *file2, int *out_n, char ***out_array)
{
	int r_val, n_a1s, n_a2s, n;
	char **a1, **a2, **array;

	array = NULL;
	n = 0;
	r_val = read_file(file1, &n, &array);
	if ( r_val == 0 )
	{
		r_val = read_file(file2, &n, &array);
		if ( r_val == 0 )
			qsort(array, n, sizeof(char *), cmp_func);
	}
	*out_array = array;
	*out_n = n;
	return(r_val);
}

/****************************************************************************/
int
main (int argc, char **argv)
{
	char *outfile, *mod_globals, *mod_externs, *elf_globals, *elf_externs;
	int n, n_a1s, n_a2s, ix, r_val;
	char **array, **a1, **a2;
	FILE *outfp;
#if 0
	{
	int wait = TRUE;
	while ( wait )
		sleep(2);
	}
#endif
	/* init */
	n_a1s = 0;
	a1 = NULL;
	n_a2s = 0;
	a2 = NULL;
	array = NULL;
	n = 0;
	r_val = 0;

	/* parse the command line options */
	if ( argc < 6 )
	{
		printf("usage: make_imports <outfile> <module global file> ");
		printf("<module extern file> <elf global file> ");
		printf("<elf extern file>\n");
		return(-1);
	}
	outfile = argv[1];
	mod_globals = argv[2];
	mod_externs = argv[3];
	elf_globals = argv[4];
	elf_externs = argv[5];

	/* open the output file */
	outfp = fopen(outfile, "w");
	if ( !outfp )
	{
		printf("Failed to open %s\n", outfile);
		return(-1);
	}
	else
	{
		/* copy the entries in file1 that are NOT in file2 into array */
		r_val = mod_read(mod_externs, mod_globals, &n_a1s,&a1);
		if ( r_val == 0 )
		{
			/* read the elf files into a sorted array */
			r_val = elf_read (elf_globals, elf_externs, &n_a2s,&a2);
			if ( r_val == 0 )
			{
				/* copy the entries in a1 that are NOT in a2 */
				sdiff(n_a1s,a1, n_a2s,a2, &n,&array);
				for ( ix=0; ix<n; ix++ )
					fprintf(outfp, "%s", array[ix]);
			}
		}
		fclose(outfp);
	}
	return (r_val);
}
