/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: mlist_id.c
   Applix ASC version: 21.0.1
   Date: Mon Feb  1 21:14:00 1999
*/
#include <stdio.h>
#include <string.h>
#include <axtime.h>
#if !AXMS
#include <sys/utsname.h>
#endif

/**************************************************************************/
/* Here it is - these are used to set the version number in the defines   *
 * in revstamp.h and axdata/revstamp, which in turn are used as the       *
 * definitive version numbers in the product itself!                      */
/**************************************************************************/
#define VERSION_DOT_FMT		"4.42"
#define VERSION_100_FMT		"442"

main(argc,argv)
int argc;
char *argv[];
{
	FILE *ifp, *ofp;
	char *p, *mlistname = argv[2], *revname = argv[1];
	char buf[101];
#if !AXMS
	struct utsname name;
#endif
	int len;
	char *strtloc;
	time_t tloc;
	struct tm	*t;
	char month_eng[31], month_frn[32], month_grm[32];

/* Open files */

	ifp = fopen(mlistname,"r");
	if ( !ifp ) {
		printf("Could not open %s for reading \n",mlistname);
		exit(1);
	}
	ofp = fopen(revname,"w");
	if ( !ofp ) {
		printf("Could not open %s for writing\n",revname);
		exit(1);
	}

    fprintf(ofp, "#define BUILD_REVISION %s\n", VERSION_DOT_FMT);
    fprintf(ofp, "#define BUILD_REVISION_100_FMT %s\n", VERSION_100_FMT);

/* fetch line , get to id pos (after second space)
	prototype: "thimble.ml ctrl 4.190 REV2.1B 12/13/90" */

	if ( !fgets(buf,100,ifp) ) {
		printf("Could not read %s\n",mlistname);
		exit(1);
	}
	p = buf;
	p = strchr(buf,' ');
	if ( !p ) {
err:	printf("Badly formed %s\n",mlistname);
		exit(1);
	}
	while ( *p == ' ' ) p++;		/* p now points to "ctrl */
	p = strchr(p,' ');
	if ( !p )
		goto err;
	while ( *p == ' ' ) p++;		/* p now points to "user */
	p = strchr(p,' ');
	if ( !p )
		goto err;
	while ( *p == ' ' ) p++;		/* p now points to "4.190 */

/* isolate id and append newline, write to file */

	strcpy(buf,p);					/* buf now begins with 4.190 */
	len = strlen(buf)-1;
	if (buf[len] == '\n')
		buf[len] = ' ';
#if 1
	p = strchr(buf,' ');			/* p now points after 4.190 */
	if ( !p ) 
		goto err;
	
	*p = 0;
#endif

#if RELEASE_VERSION
	/*fprintf(ofp,"#define BUILD_VERSION  %s\n",buf);*/
	fprintf(ofp,"#define BUILD_VERSION_STR  \"%s\"\n",buf);
#else
	/*fprintf(ofp,"#define BUILD_VERSION  0\n");*/
	fprintf(ofp,"#define BUILD_VERSION_STR  \"0\"\n");
#endif

	tloc = time((time_t*)0);
	strtloc = ctime(&tloc);
	len = strlen(strtloc)-1;
	if (strtloc[len] == '\n')
		strtloc[len] = 0;

	/* Construct a more friendly build time in the format Mmmmm dd, yyyy */
	t = localtime( &tloc);
	if (t) {
		switch(t->tm_mon) {
			case 0: 
			default:
				strcpy(month_eng, "January");
				strcpy(month_frn, "Janvier");
				strcpy(month_grm, "Januar");
				break;
			case 1:
				strcpy(month_eng, "February");
				strcpy(month_frn, "Fvrier");
				strcpy(month_grm, "Februar");
				break;
			case 2:
				strcpy(month_eng, "March");
				strcpy(month_frn, "Mars");
				strcpy(month_grm, "Mrz");
				break;
			case 3:
				strcpy(month_eng, "April");
				strcpy(month_frn, "Avril");
				strcpy(month_grm, "April");
				break;
			case 4:
				strcpy(month_eng, "May");
				strcpy(month_frn, "Mai");
				strcpy(month_grm, "Mai");
				break;
			case 5:
				strcpy(month_eng, "June");
				strcpy(month_frn, "Juin");
				strcpy(month_grm, "Juni");
				break;
			case 6:
				strcpy(month_eng, "July");
				strcpy(month_frn, "Juillet");
				strcpy(month_grm, "Juli");
				break;
			case 7:
				strcpy(month_eng, "August");
				strcpy(month_frn, "Aot");
				strcpy(month_grm, "August");
				break;
			case 8:
				strcpy(month_eng, "September");
				strcpy(month_frn, "Septembre");
				strcpy(month_grm, "September");
				break;
			case 9:
				strcpy(month_eng, "October");
				strcpy(month_frn, "Octobre");
				strcpy(month_grm, "Oktober");
				break;
			case 10:
				strcpy(month_eng, "November");
				strcpy(month_frn, "Novembre");
				strcpy(month_grm, "November");
				break;
			case 11:
				strcpy(month_eng, "December");
				strcpy(month_frn, "Dcembre");
				strcpy(month_grm, "Dezember");
				break;
		}
		fprintf(ofp,"#define BUILD_DATE_ENG       \"%s %d, %d\"\n",
								month_eng, t->tm_mday, 1900+t->tm_year);
		fprintf(ofp,"#define BUILD_DATE_FRN       \"%s %d, %d\"\n",
								month_frn, t->tm_mday, 1900+t->tm_year);
		fprintf(ofp,"#define BUILD_DATE_GRM       \"%s %d, %d\"\n",
								month_grm, t->tm_mday, 1900+t->tm_year);
	}

#if !AXMS
	uname(&name);
#endif

	fprintf(ofp,"#define BUILD_DATE_AND_TIME  \"%s\"\n",strtloc);
	fprintf(ofp,"#define ASC_BUILDDATIME      \"@(#) Applix: %s\"\n",strtloc);
	fprintf(ofp,"#define ASC_WHAT_STR         \"@(#)Version: %s(%s) %s\"\n",
					VERSION_DOT_FMT, buf,strtloc);
#if !AXMS
	fprintf(ofp,"#define ASC_WHAT_MACHINE      \"%s %s %s %s\"\n",name.sysname,name.release,name.version,name.machine);
	fprintf(ofp,"#define ASC_WHAT_MACHINE2     \"@(#)Machine: %s %s %s %s\"\n",name.sysname,name.release,name.version,name.machine);
#else
	fprintf(ofp,"#define ASC_WHAT_MACHINE      \"NT or Windows 95\"\n");
	fprintf(ofp,"#define ASC_WHAT_MACHINE2     \"@(#)Machine: NT or Windows 95\"\n");
#endif

	if ( 0 != fclose(ofp) ) {
		printf("Could not close %s\n",argv[1]);
		exit(1);
	}
	
#if 0
	if ( strlen(buf) != write(fileno(ofp),buf,strlen(buf)) ) {
		printf("Could not write %s\n",revname);
		exit(1);
	}
#endif
	exit(0);
}
