/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: sortdiff.c
   Applix ASC version: 1
   Date: Fri Jun  4 21:14:36 1999
*/
/****************************************************************************\
                                 includes
\****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/****************************************************************************\
                                  defines
\****************************************************************************/
#define TRUE 1
#define FALSE 0

#define LINE_LEN 1024

/****************************************************************************\
                             global variables
\****************************************************************************/
char line[LINE_LEN];

/****************************************************************************/
int
cmp_func (const void *input1, const void *input2)
{
	char *str1, *str2;
	str1 = *(char **)input1;
	str2 = *(char **)input2;
	return(strcmp(str1, str2));
}

/****************************************************************************/
int
main (int argc, char **argv)
{
	int r_val, n_a1s, n_a2s, n_diffs, ix, ix1, ix2, sz, res, tmp;
	char **a1, **a2, **diff;
	FILE *fp1, *fp2;
#if 0
	{
	int wait = FALSE;
	while ( wait )
		sleep(2);
	}
#endif

	/* open the files */
	if ( argc < 3 )
		return (-1);
	fp1 = fopen(argv[1], "r");
	if ( fp1 == NULL )
		return (-1);
	fp2 = fopen(argv[2], "r");
	if ( fp2 == NULL )
	{
		fclose(fp1);
		return (-1);
	}

	/* read the lines of file 1 into a1 (skip lines starting with #) */
	a1 = NULL;
	n_a1s = 0;
        while ( fgets(line, LINE_LEN, fp1) )
	{
		if ( line[0] == '#' )
			continue;

		if ( (n_a1s%100) == 0 )
		{
			sz = (n_a1s+100) * sizeof(char *);
			a1 = (char **)realloc(a1, sz);
		}
		a1[n_a1s] = (char *)malloc(strlen(line)+1);
		strcpy(a1[n_a1s], line);
		n_a1s++;
	}


	/* read the lines of file 2 into a2 (skip lines starting with #) */
	a2 = NULL;
	n_a2s = 0;
        while ( fgets(line, LINE_LEN, fp2) )
	{
		if ( line[0] == '#' )
			continue;

		if ( (n_a2s%100) == 0 )
		{
			sz = (n_a2s+100) * sizeof(char *);
			a2 = (char **)realloc(a2, sz);
		}
		a2[n_a2s] = (char *)malloc(strlen(line)+1);
		strcpy(a2[n_a2s], line);
		n_a2s++;
	}

	/* close the files */
	fclose(fp1);
	fclose(fp2);

	/* sort the arrays */
	qsort(a1, n_a1s, sizeof(char *), cmp_func);
	qsort(a2, n_a2s, sizeof(char *), cmp_func);

	/* compare the lists, creating a new list */
	diff = NULL;
	n_diffs = 0;
	for ( ix1=ix2=0; ix1<n_a1s && ix2<n_a2s; )
	{
		res = strcmp(a1[ix1], a2[ix2]);
		if ( res < 0 )
		{
			/* copy the string to the diff list */
			if ( (n_diffs%100) == 0 )
			{
				sz = (n_diffs+100) * sizeof(char *);
				diff = (char **)realloc(diff, sz);
			}
			diff[n_diffs++] = a1[ix1++];

			/* advance list1 past the matching string(s) */
			while ( ix1<n_a1s && strcmp(diff[n_diffs], a1[ix1]) == 0 )
				ix1++;
		}
		else if ( res == 0 )
		{
			/* advance list1 past the matching string(s) */
			while ( ix1<n_a1s && strcmp(a2[ix2], a1[ix1]) == 0 )
				ix1++;

			/* advance list2 past the matching string(s) */
			tmp = ix2;
			while ( ix2<n_a2s && strcmp(a2[tmp], a2[ix2]) == 0 )
				ix2++;
		}
		else
			ix2++;
	}

	/* print the sorted, diffed list */
	for ( ix=0; ix<n_diffs; ix++ )
		printf("%s", diff[ix]);

	return (r_val);
}
