/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: strip_tr.c
   Applix ASC version: 3
   Date: Mon Nov 16 13:57:13 1998
*/
#include <stdio.h>
#include <errno.h>

#define TRUE 1
#define FALSE 0


static char *Complex_Strip( char *str );


main(argc,argv)
int argc;
char *argv[];
{
	FILE *rfp, *wfp;
	char buf[300];
    char *infile, *outfile;
    int ret;

	infile = argv[2];
	outfile = argv[1];

    /* open input file */
#if AXMS
	rfp = fopen(infile,"rb");    /* open in 'binary' mode */
#else
	rfp = fopen(infile,"r");
#endif
	if ( !rfp ) {
		fprintf(stderr,"Could not open input file %s\n",infile);
		exit(1);
	}

    /* open output file */
#if AXMS
	wfp = fopen(outfile,"wb");
#else
	wfp = fopen(outfile,"w");
#endif
	if ( !wfp ) {
		fprintf(stderr,"Could not open output file %s\n",outfile);
		exit(1);
	}

    while ( fgets( buf, 300, rfp ) != (char *)NULL )
       {
       if ( buf[0] != '"' )
           {
           /* not an elf string, just output */
           ret = fprintf( wfp, "%s", buf );
           if ( ret < 0 )
              {
              fprintf(stderr,"Error reading <%s>, errno = %d\n", infile, errno );
              exit(1);
              }
           continue;
           }

       /* this is an elf string, strip off any possible translation tag */
       ret = fprintf( wfp, "\"%s", Complex_Strip( &buf[1] ) );
       if ( ret < 0 )
          {
          fprintf(stderr,"Error writing <%s>, errno = %d\n", outfile, errno  );
          exit(1);
          }
       }

	fclose(rfp);
	ret = fclose(wfp);
    if ( ret != 0 )
       {
       fprintf(stderr,"Error on close of <%s>, errno = %d\n", outfile, errno  );
       exit( 1 );
       }
	exit(0);
}


/** take a "complex" string, which is composed of a tag **
/**   is followed by 3 pipe symbols, followed by   **/
/**   the English equivalent, and return the English string.     **/
/** If the string does not contain the 3 pipe symbols in a row delimiter */
/**   then the input string is returned in its entirety        */

/*	Search from front of string, not back, because expression might have
 *	'|' in it -- lhunt 12 aug 94 */
 
static char *Complex_Strip( char *str )
{
    register char *p, *end;
 
    if ( str == (char *)NULL )
        return( str );
    if ( *str == '\0' )
        return( str );
 
    end = str + strlen(str) - 3;     /* end of string */
    for ( p = str; p < end ; p++)
        {
			if ( (*p == '|') && (*(p+1) == '|') && (*(p+2) == '|') )
			{
			   /** triple pipe found, return english */
			   return( (p+3) );
		   }
        }
 
    /* fell through, no triple pipe, just return original */
    return( str );
}
 

