#!/bin/csh 
#
# Copyright (C) 1983-1999 Applix, Inc.
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# 02111-1307, USA.
# 
# Applix Inc. is the original author of this material.  Please send
# generally useful enhancements to Applix in the form of all modified 
# source files including an explanation of the changes.  For more
# information see our website at www.applixware.org.
# 
# Do not remove or modify the license information above or the following
# Applix source control information from this file.
# 
# Applix filename: win_elf_partial_link.csh
# Applix ASC version: 4
# Date: Thu Jun  3 21:16:31 1999
# 
#
##################################################################################
# Usage: elf_partial_link elfimports.xx edump.pl sdiff.pl elfsyms1.dat usrmacros.dat <elf compiler generated c files>
##################################################################################
#
# Implementation Notes:
# 
# elf_partial_link links a modules's .els files and generates import list file
#   elfimports.<module>, list of macros defined elfglobals.<module>
# and list of all macro calls made elfexterns.<module>
# Inputs:
#   edump.pl      - Perl script that dumps elf macros info from elfc
#                   generated C files
#   sdiff.pl      - Perl script that diffs two sorted files and outputs
#                   the lines which are in file1 and not in file2
#   elfsyms1.dat  - Elf built in macros list
#   usrmacros.dat - User macros or macros supported by external utilities
#   
#   Elf generated C files
# Outputs:
#   elfimports.<module> - Passed on command line
#   elfglobals.<module> - assumes same extension as elfimports file
#   elfexterns.<module> - assumes same extension as elfimports file
# where <module> is one of db,ge,wp,ss etc.,
#   
#
# elf_link takes multiple import files as input and resolves them by using
# corresponding  elfglobals files and uses
# elfexterns files to add Error line number and file name to the
# unresolved macros
#
# 
if ($#argv < 7) then
    echo "Usage Error: elf_partial_link <import/export file> <export/import file> edump.pl sdiff.pl elfsyms1.dat usrmacros.dat <elfc generated c files>"
    exit 1
end
if ($1 == "-axms") then
    set AXMS=1
    shift
end
set file1=$1:t
set base1=$file1:r
if ($base1 == "elfimports") then
	set IMP_FILE=$1
	set EXP_FILE=$2
else
	set EXP_FILE=$1
	set IMP_FILE=$2
end
set PERL=perl
set ELFDUMP=$PERL $3
set SDIFF=$PERL $4
set ELF_BUILTINS=$5
set ELF_EXTERNALS=$6
#shift used args, remaining args are input file names
shift;shift;shift;shift;shift;shift
set TOOLS_DIR=$ELFDUMP:h
set OBJDIR=$IMP_FILE:h
set ext=$IMP_FILE:e
set GLOB_FILE=elfglobals.$ext
set EXT_FILE=elfexterns.$ext
set ERR_FILE=inv_names.$ext
if ($AXMS == 1) then
    set RM=hrm -f
else
    set RM="/bin/rm -f"
end
cd $OBJDIR
#Remove all the files that will be generated (for noclobber mode support)
$RM $GLOB_FILE $EXT_FILE $IMP_FILE zzelflnk.t1 zzelflnk.t2
$ELFDUMP -gf -gv $* >$GLOB_FILE
$ELFDUMP -ef $* >$EXT_FILE
$SDIFF  $EXT_FILE $GLOB_FILE >zzelflnk.t1
cat $ELF_BUILTINS $ELF_EXTERNALS > zzelflnk.t2
$SDIFF zzelflnk.t1 zzelflnk.t2 >$IMP_FILE
$RM zzelflnk.t1 zzelflnk.t2
$RM errors.t1 $ERR_FILE
#Check for invalid macro/variable names
$PERL -ne 'if(/^##/) { $file=$_; } else { if(\!/^[]\[#\!\?]/ && \!/^(.*[_\$\@])$/) { print $file,$_; $file=""; } }' $GLOB_FILE $EXT_FILE > errors.t1
if ( ! -z errors.t1 ) then
	echo "Error: Invalid Internal macro or global variable names" >$ERR_FILE
	cat errors.t1 >>$ERR_FILE
	echo "Invalid name Error(s) written to $OBJDIR/$ERR_FILE"
end
$RM errors.t1
setenv LIB_NAME "libax$ext"
grep -v "^#" $GLOB_FILE | tee elfexports.tmp |	wc -l | $PERL -lane 'print "\@1\n\@",$F[0]+1, "\n\"$ENV{'LIB_NAME'}";' >$EXP_FILE
# sort and prefix with '"'
$PERL -e '{ @f=sort(<STDIN>); foreach $i (@f){print "\"",$i;}}' < elfexports.tmp  >>$EXP_FILE
$RM elfexports.tmp
exit 0
