/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: achmod.c
   Applix ASC version: 4
   Date: Mon Nov 16 13:57:58 1998
*/
#if APOLLO
#include <stdio.h>
#endif
#include <utloc.h>
#include <getstr9.h>

#define MAXLINE 200

#if SYS&VMSSYS
static int err,mode,do_chmod();
#endif

/*
.BEGIN MANPAGE
.NAME Achmod - Change the access permissions of files.
.DESCRIPTION
	Achmod() sets the access permissions of specified files to a specified mode.
.ARGUMENT objects
	The address of a string which identifies atleast one file.
	Objects may identify multiple files by using wildcards and
	multiple pathnames separated by spaces or tabs.  "*" and "?" are
	legal wildcards.
.ARGUMENT ac_mode
	The access permissions to set.
.RETURN VALUE
	Achmod() returns 0 if it succeeds.
.SEE ALSO
	chmod(2)
.NOTES
	The system call is more efficient and should be used if the pathnames
	are known.
.END MANPAGE
*/

/*	BEGIN SYNOPSIS */
typeErr Achmod(objects,ac_mode) 
char *objects;
int ac_mode;
/*	END SYNOPSIS */
{
	RNAME("Achmod");

#if SYS&AEGISO

	FILE *fp;
    char line[MAXLINE];
    int err = AOK;
    char *tempName = NXLT("zzchmod.%%%%%%");
	char *p;
    extern char *Amktemp();

	TRACE((UT,rname,TRLEV_1,"pathnames %s ACL %o",objects,ac_mode));
    str_aegis_wild(line,objects);
    pgm_save_output("ld",mktemp(tempName),5,"-c","-sn","-nhd","-un",line);
    
	if(!(fp = fopen(tempName,"r"))) {
		TRACE((UT,rname,TRLEV_2,"opening of %s  failed",tempName));
		return(err);
	}

	TRACE((UT,rname,TRLEV_4,"value of fp is 0x%x\ncalling fgets,fp"));
    while(fgets(line,MAXLINE,fp)) {
		TRACE((UT,rname,TRLEV_4,"file name whose Acl to be changed %s",line));
		if(p = rindex(line,'\n'))
			*p = '\0';
        /* Need to unescape the pathname here! */
        if(chmod(line,ac_mode)) {
			TRACE((UT,rname,TRLEV_2,
				NXLT("chmod sys call failed for file %s mode %o"),line,ac_mode));
            err = S9CHMOD;
            break;
        }
	TRACE((UT,rname,TRLEV_4,"acl changed for %s",line));
    }
    fclose(fp);

	unlink(tempName);

	TRACE((UT,rname,TRLEV_2,"exiting Achmod"));
    return(err);

#else

#if SYS&VMSSYS

	err = AOK;
	mode = ac_mode;

	shell$to_vms(objects,do_chmod,1);

	return(err);

#else

	char cmdBuf[MAXLINE];

	if(!strpbrk(objects," \t?*")) {
		if(chmod(objects,ac_mode))
			return(S9CHMOD);
		return(AOK);
	}
	sprintf(cmdBuf,"chmod %o %s",ac_mode,objects);
	if(system(cmdBuf))
		return(S9CHMOD);
	return(AOK);

#endif
#endif
}  

#if SYS&VMSSYS

/* For wildcarded file names, this is called */
/* once for each matching file name. */

static do_chmod(fname,type)
char *fname;
int type;
{
    globalvalue int SHELL$K_DIRECTORY;
	int len;
    char *p;
    char dir_name[STRPATH + 1],stemp[STRNORMAL + 1];
 
	/*
	 * The chmod shell command applies to both directories and ordinary files.
	 * However, we must do some conversions for directory names.
	 * Examples:
	 *
	 *      Unix syntax     shell$to_vms result     chmod needs
	 *      -----------     -------------------     -----------
	 *      /d/x/y/z        d:[x.y.z]               d:[x.y]z.dir
	 *      /d/x/y          d:[x.y]                 d:[x]y.dir
	 *      /d/x            d:[x]                   d:[000000]x.dir
	 */
    if(type == SHELL$K_DIRECTORY) {
		len = strlen(fname);
        memcpy(dir_name,fname,len + 1);
		fname = dir_name;
        p = dir_name + len - 1;
        if(*p == ']') {
            memcpy(p,".dir",5);
			while(--p > dir_name && *p != '.');
            if(p > dir_name)
                *p = ']';
            else {
				p = dir_name;
				while(*p++ != '[');
				len += 4 - (dir_name - p);
				memcpy(stemp,p,len);
				memcpy(p,"000000]",7);
				memcpy(p + 7,stemp,len);
            }
        }
    }
    if(chmod(fname,mode))
		err = S9CHMOD;
 
    return(1);
}

#endif
