/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: acopy.c
   Applix ASC version: 3
   Date: Mon Nov 16 13:57:58 1998
*/
#include <utloc.h>
#include <getstr9.h>

#if SYS&VMSSYS
#include <stat.h>
struct copy_file_tag {
	struct copy_file_tag *next;
	char name[1];
};
static int add_to_list();
static struct copy_file_tag *copy_list;
#endif

/*
.BEGIN MANPAGE
.NAME Acopy - Copy files.
.DESCRIPTION
	Acopy() copies specified files to a specified destination.
.ARGUMENT acpy_source
	The address of a string which may identifies at least one file.
	Acpy_source may identify multiple files by using wildcards and
	multiple pathnames separated by spaces or tabs.  "*" and "?" are
	legal wildcards.
.ARGUMENT acpy_target
	The address of a string which identifies a destination directory.
	If acpy_source identifies a single file, acpy_target may identify a
	file or directory.
.RETURN VALUE
	Acopy() returns 0 if it succeeds.
.SEE ALSO
	AcopyFile(3A) AcopyTree(3A)
.END MANPAGE
*/

/* BEGIN SYNOPSIS */
typeErr Acopy(acpy_source,acpy_target)
char *acpy_source;
char *acpy_target;
/* END SYNOPSIS */
{
#if SYS&VMSSYS
	struct copy_file_tag *next;
	char sources[STRPATH + 1];
	char *p;
	struct stat lstat;
	int isdir;

	if(strpbrk(acpy_source," \t?*")) {
		
		strcpy(sources,acpy_source);
		copy_list = NULL;
		p = sources;

		while(p = strtok(p," \t")) {
			shell$to_vms(p,add_to_list,1);
			p = NULL;
		}

		if(!copy_list)
			return(AOK);	/* Nothing to copy */

		if(!Astat(acpy_target,&lstat) && lstat.st_mode&S_IFDIR)
			isdir = TRUE;
		else
			isdir = FALSE;

		if(copy_list->next && !isdir)
			return(S9COPY);

		while(copy_list) {
			acpy_source = copy_list->name;
			if(isdir) {
				if(p = rindex(acpy_source,'/'))
					p++;
				else
					p = acpy_source;
				sprintf(sources,"%s/%s",acpy_target,p);
				p = sources;
			}
			else
				p = acpy_target;
			VMS$copyfile(acpy_source,p);
			next = copy_list->next;
			free((char *)copy_list);
			copy_list = next;
		}

		return(AOK);
	}
#else
#if SYS&AEGISO
    char sources[2 * STRPATH];
    char target[2 * STRPATH];         
    char *sp;

    if(strpbrk(acpy_source," \t?*")) {
    	sp = sources;
    	*sp++ = '(';
    	*sp++ = ' ';
    	str_aegis_wild(sp,acpy_source);
    	sp += strlen(sp);
    	*sp++ = ' ';
    	*sp++ = ')';
    	*sp++ = '\0';
        str_aegis_wild(target,acpy_target);
        if(pgm_system("cpf",4,"-r","-nq",sources,target))
	    	return(AOK);
	    return(S9COPY);
	}

   	str_aegis_wild(target,acpy_target);
   	str_aegis_wild(sources,acpy_source);
#else    
    char  acpy_cmd[STRPATH + STRPATH + 5];

	if(strpbrk(acpy_source," \t?*")) {
    	sprintf(acpy_cmd,"cp %s %s 2>/dev/null",acpy_source,acpy_target);
    	if(system(acpy_cmd))
        	return(S9COPY);
		return(AOK);
	}
#endif
#endif

   	if(AcopyFile(acpy_source,acpy_target))
	    return(S9COPY);
	return(AOK);

}

#if SYS&VMSSYS
/* For a wildcarded file specification, this is */
/* called once for each matching file name. */
/* Stick them all in a linked list. */

globalvalue int SHELL$K_FILE;

static int add_to_list(fname,type)
char *fname;
int type;
{
    struct copy_file_tag *new;

    /* The rm shell command does not delete directories. */
    if(type == SHELL$K_FILE) {
        new = (struct copy_file_tag *)
            malloc(sizeof(struct copy_file_tag) + strlen(fname));
        new->next = copy_list;
        strcpy(new->name,fname);
        copy_list = new;
    }
    return(1);
}
#endif
