/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: adld.c
   Applix ASC version: 30
   Date: Mon Jun 28 21:13:54 1999
*/
#if (SUN4 || SCOUNIX || UNIXWARE || NCR || SUNIS || AVIION || SGI || LINUX || ALPHA || NEC_EWS) 
#include <dlfcn.h>
#define ADLD_SUPPORTED 1
#endif

#if HP800V3 && !HYUNDAI
#include <dl.h>
#define ADLD_SUPPORTED 1
#endif

#if AXMS
#include <windows.h>
#include <winmap.h>
#include <getstr10.h>
#define ADLD_SUPPORTED 1
#endif

#if RIOS
#include <stdio.h>
#include <sys/ldr.h>
#define ADLD_SUPPORTED 1
#endif

#include <adld.h>

#ifndef NULL
#define NULL 0
#endif

/* PORTING: Follow the description for Adlopen(), Adlclose(), Adlsym(),
   Adlerror() in the SUN4 case. */

static char *last_adl_error = NULL;


#if (SUN4 || SUN4S || SCOUNIX || UNIXWARE || NCR || SUNIS || AVIION || SGI || ALPHA || NEC_EWS || LINUX)   /*   {  */

/***********************************************************************
function: Adlopen - Attach the shared library named by `path' to the
process.  `mode' must be one or more flags (or'ed) of:
    ADL_BIND_IMMEDIATE -- resolve symbol references when the library is
                          loaded;
    ADL_BIND_DEFERRED  -- delay resolution until actual reference.
	ADL_BIND_FIRST     -- On HP800, to place the library at the head of
						  the symbol search order.
						  On all other systems this flag is ignored.
						  This should not be used for user shared
						  libraries, as they will be able
						  to override applixware internal functions.
Return a handle suitable for use with subsequent calls to Adlsym(),
Adlclose() on success;  NULL on failure.
*/
void *
Adlopen(path, mode)
char *path;
int mode;
{
    int dl_mode;
#if SUN4 && !SUN4S  /* Sun OS */
    dl_mode = RTLD_LAZY;    /* the only supported on SUNOS */
#else
    if (mode & ADL_BIND_IMMEDIATE)
        dl_mode = RTLD_NOW;
    else /* ADL_BIND_DEFERRED */
        dl_mode = RTLD_LAZY;
#endif

    /* check global bit */
    if ((mode & ADL_BIND_GLOBAL) == ADL_BIND_GLOBAL) 
        dl_mode |= RTLD_GLOBAL;

    return(dlopen(path, dl_mode));
}

/***********************************************************************
function: Adlsym - Return the address of an exported symbol `symbol' from
the shared lirary.  `handle' is the handle gotten from a call to Adlopen().
  TYPE specifies the type of the symbol:
    ADL_TYPE_PROCEDURE -- requested symbol is a procedure.
    ADL_TYPE_DATA      -- requested symbol is data.
    ADL_TYPE_UNDEFINED -- requested symbol is undefined.
Return address of `symbol' on success;  NULL on failure.
*/
/*ARGSUSED*/
void *
Adlsym(handle, symbol, type)
void *handle;
char *symbol;
int type; /* ignored */
{
    return(dlsym(handle, symbol));
}


/***********************************************************************
function: Adlerror - Return a string describing the last error that occured
during a Adlopen(), Adlsym() or Adlclose().  At each call to Adlerror(), the
"last error" string is reset.
*/
char *
Adlerror()
{
    static char *no_error = "";
    last_adl_error = dlerror();
    if (!last_adl_error)
        last_adl_error = no_error;
    return(last_adl_error);
}

/***********************************************************************
function: Adlclose - Delete refrence to the shared object referenced by
`handle'.  Return 0 on sucess;  -1 on failure.
*/
int
Adlclose(handle)
void *handle;
{
    if (dlclose(handle) != 0)
        return(-1);
    return(0);
}
#endif /* }  SUN4 et al. */



#if HP800V3 && !HYUNDAI   /*  {  */

/***********************************************************************
function: Adlopen - Attach the shared library named by `path' to the
process.  `mode' must be one or more flags (or'ed) of:
    ADL_BIND_IMMEDIATE -- resolve symbol references when the library is
                          loaded;
    ADL_BIND_DEFERRED  -- delay resolution until actual reference.
	ADL_BIND_FIRST     -- On HP800, to place the library at the head of
						  the symbol search order.
						  On all other systems this flag is ignored.
						  This should not be used for user shared
						  libraries, as they will be able
						  to override applixware internal functions.
Return a handle suitable for use with subsequent calls to Adlsym(),
Adlclose() on success;  NULL on failure.
*/
void *
Adlopen(path, mode)
char *path;
int mode;
{
    shl_t h;
    int shl_mode;
	shl_mode = DYNAMIC_PATH;
    if (mode & ADL_BIND_IMMEDIATE)
        shl_mode |= BIND_IMMEDIATE;
    else
        shl_mode |= BIND_DEFERRED;
	if (mode & ADL_BIND_FIRST)
		shl_mode |= BIND_FIRST;
    h = shl_load(path, shl_mode, 0L);
    return((void *) h);
}

void *
Adlsym(handle, symbol, type)
void *handle;
char *symbol;
int type;
{
    shl_t h = (shl_t) handle;
    int shl_type;
    int e;
    void *addr = NULL;
    int n, ix;
    struct shl_symbol *symlist = NULL;
    struct shl_symbol *sp;

    if (type == ADL_TYPE_PROCEDURE)
        shl_type = TYPE_PROCEDURE;
    else if (type == ADL_TYPE_DATA)
        shl_type = TYPE_DATA;
    else
        shl_type = TYPE_UNDEFINED;

#if 0 /* FIXME, this does not work!  Returns -1 and sets errno to EINVAL.
         What is wrong here? */
    e = shl_findsym(h, symbol, shl_type, &addr);
    if (e < 0)
        return(NULL);
#endif
    /* This works. */
    n = shl_getsymbols(h, shl_type, EXPORT_SYMBOLS, malloc, &symlist);
    if (n > 0) {
        for (ix = 0, sp = symlist; symlist && (ix < n); ix++) {
            if (!strcmp(symbol, sp->name)) {
                addr = sp->value;
                break;
            }
            sp++;
        }
    }
    if (symlist)
        free(symlist);

    return(addr);
}

char *
Adlerror()
{
    static char *err = "";
    last_adl_error = err;
    return(last_adl_error);
}

int
Adlclose(handle)
void *handle;
{
    shl_t h = (shl_t) handle;
    if (shl_unload(h) != 0)
        return(-1);
    return(0);
}
#endif /* }    HP800V3 */





#if AXMS    /*  {  */
/***********************************************************************
function: Adlopen - Attach the shared library named by `path' to the
process.  `mode' is not used on Windows/NT
Return a handle suitable for use with subsequent calls to Adlsym(),
Adlclose() on success;  NULL on failure.
*/
#define MS_LoadLibrary LoadLibrary
#define MS_FreeLibrary FreeLibrary
#define MS_GetProcAddress GetProcAddress

static int last_lib_error = -1;

void *Adlopen(path, mode)
char *path;
int mode;
{
    HINSTANCE lib_handle;

    last_lib_error = -1;
    lib_handle = MS_LoadLibrary( path );
    if ( lib_handle < (HINSTANCE)HINSTANCE_ERROR )
       {
       if ( lib_handle == (HINSTANCE)0 )
           last_lib_error = MS_GetLastError();
       else
           last_lib_error = (int)lib_handle;
       lib_handle = (HINSTANCE)NULL;
       }
    return( (void *)lib_handle );
}


/***********************************************************************
function: Adlsym - Return the address of an exported symbol `symbol' from
the shared lirary.  `handle' is the handle gotten from a call to Adlopen().
  "type" is not used on Windows/NT
Return address of `symbol' on success;  NULL on failure.
*/
void *Adlsym(handle, symbol, type)
void *handle;
char *symbol;
int type; /* ignored */
{
    FARPROC ret_ptr;

    last_lib_error = -1;
    ret_ptr = MS_GetProcAddress( (HINSTANCE)handle, symbol );
    if ( ret_ptr == (FARPROC)NULL )
       last_lib_error = MS_GetLastError();
    return( (void *)ret_ptr );
}
 

/***********************************************************************
function: Adlerror - Return a string describing the last error that occured
during a Adlopen(), Adlsym() or Adlclose().  At each call to Adlerror(), the
"last error" string is reset.
*/
char *Adlerror(void)
{
    static char err_str[261];    /* is this big enough? */
    static char local_error[10];

    if ( last_lib_error == -1 )
        {
        local_error[0] = '\0';
        return( local_error );
        }
    else
        {
        if ( last_lib_error < HINSTANCE_ERROR )
            {
            sprintf( local_error, "%d", last_lib_error );
            last_lib_error = -1;
            return( local_error );
            }
        else
            {
            /* probably FEEADDRNOTAVAIL  */
            MS_FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NULL, last_lib_error,
                             MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                             (LPTSTR)err_str, 260, NULL);
            last_lib_error = -1;
            return err_str;
            }
        }
}
          
          

/***********************************************************************
function: Adlclose - Delete refrence to the shared object referenced by
`handle'.  Return 0 on sucess;  -1 on failure.
*/
int Adlclose(handle)
void *handle;
{
    MS_FreeLibrary( (HINSTANCE)handle );
    return(0);
}

#endif   /*  }  AXMS  */

 






#if RIOS /* { */

/***********************************************************************
function: Adlopen - Attach the shared library named by `path' to the
process.  `mode' must be one or more flags (or'ed) of:
    ADL_BIND_IMMEDIATE -- resolve symbol references when the library is
                          loaded;
    ADL_BIND_DEFERRED  -- delay resolution until actual reference.
	ADL_BIND_FIRST     -- On HP800, to place the library at the head of
						  the symbol search order.
						  On all other systems this flag is ignored.
						  This should not be used for user shared
						  libraries, as they will be able
						  to override applixware internal functions.
Return a handle suitable for use with subsequent calls to Adlsym(),
Adlclose() on success;  NULL on failure.
*/
void *
Adlopen(path, mode)
char *path;
int mode;
{
	void* handle;
    int dl_mode;
	typeErr err;
	
	if (mode & ADL_BIND_IMMEDIATE)
		dl_mode = L_NOAUTODEFER;
	else
		dl_mode = 0;
	dl_mode |= L_LIBPATH_EXEC;
	handle = load(path, dl_mode, NULL);
	if(NULL != handle && mode & ADL_BIND_IMMEDIATE)  {
		/* resolve import symbols in shared lib from main exports */
		err = loadbind(0, Adlopen,  handle);
		if (err != 0)
			return(NULL);
	}
	return(handle);
}

/***********************************************************************
function: Adlsym - Return the address of an exported symbol `symbol' from
the shared lirary.  `handle' is the handle gotten from a call to Adlopen().
  TYPE specifies the type of the symbol:
    ADL_TYPE_PROCEDURE -- requested symbol is a procedure.
    ADL_TYPE_DATA      -- requested symbol is data.
    ADL_TYPE_UNDEFINED -- requested symbol is undefined.
Return address of `symbol' on success;  NULL on failure.
*/
/*ARGSUSED*/
void *
Adlsym(handle, symbol, type)
void *handle;
char *symbol;
int type; /* ignored */
{
	/* Adlsym not supported on RIOS */
	return(NULL);
}

/***********************************************************************
function: Adlerror - Return a string describing the last error that occured
during a Adlopen(), Adlsym() or Adlclose().  At each call to Adlerror(), the
"last error" string is reset.
*/
char *
Adlerror()
{
    static char *no_error = "";
	void* errmsgs[100];
	loadquery(L_GETMESSAGES, errmsgs, 100);
	last_adl_error = (char *)errmsgs[0];
    if (!last_adl_error)
        last_adl_error = no_error;
    return(last_adl_error);
}

/***********************************************************************
function: Adlclose - Delete refrence to the shared object referenced by
`handle'.  Return 0 on sucess;  -1 on failure.
*/
int
Adlclose(handle)
void *handle;
{
	if (unload(handle) != 0)
		return(-1);
    return(0);
}

#endif /* } */



#if !ADLD_SUPPORTED    /*  {  */
   /** Stub routines for platforms without shared library support */
void *
Adlopen(path, mode)
char *path;
int mode;
{
    return NULL;
}

void *
Adlsym(handle, symbol, type)
void *handle;
char *symbol;
int type;
{
    return NULL;
}

char *
Adlerror()
{
    return NULL;
}

int
Adlclose(handle)
void *handle;
{
    return -1;
}
#endif /* }  ADLD_UNSUPP */


/* GNU Emacs settings - This (including the "\n^L") MUST be the last thing
 *                      in this file!
 * Local variables:
 * mode: c
 * tab-width: 4
 * end:
 */
