/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: adu.c
   Applix ASC version: 4
   Date: Mon Nov 16 13:57:58 1998
*/
#include <sys/types.h>
#include <sys/stat.h>
#if !(SYS&VMSSYS) && !AXMS
#include <sys/param.h>
#if !SYSV4 && !AXMS
#include <sys/dir.h>
#endif
#endif
#include <utloc.h>


/* Assume standard definitions for all ports */

#define BUFSIZE 1024
#ifdef NSHIFT
#undef NSHIFT
#endif
#define NSHIFT 10
#ifdef NINDIR
#undef NINDIR
#endif 
#if !(SYS&VMSSYS)
#define NINDIR (BUFSIZE/sizeof(daddr_t))
#endif
#define DIRECT	10
#define MAXNODES 500

static long BlockSize();

/*
.BEGIN MANPAGE
.NAME Adu - Summarize disk usage.
.DESCRIPTION
	Given a file or directory name, Adu() recursively examines all the files
	and sub-directories and totals the number of kilobytes of disk space
	used.
.ARGUMENT filename
	The address of a string which identifies a directory or file.
.RETURN VALUE
	Adu() returns the number of kilobytes of disk space used.
.SEE ALSO
	Agetcwd() Astat(3A)
.END MANPAGE
*/

/* BEGIN SYNOPSIS */
long Adu(filename)
char *filename;
/* END SYNOPSIS */
{
	int level = 0,start_dir;
	char *place = NULL;
	struct stat statbuf;
	char path[STRPATH + 1],name_for_stat[STRPATH + 1];
	char fname[STRNORMAL + 1];
	long size = 0;
#if !(SYS&VMSSYS)
	int i;
	struct inodeinfo {
		dev_t	dev;
		ino_t	ino;
	} inodeInfo[MAXNODES];
	int	linkcount = 0;

	bzero((char *)inodeInfo,sizeof(struct inodeinfo) * MAXNODES);
#endif

	strcpy(name_for_stat,filename);

get_info:
	if(!Astat(name_for_stat,&statbuf)) {
#if !(SYS&VMSSYS)
		if(statbuf.st_nlink > 1) {
			for(i = 0; i <= linkcount; i++) {
				if(inodeInfo[i].ino == statbuf.st_ino &&
					inodeInfo[i].dev == statbuf.st_dev)
					goto get_next;
			}
			if(linkcount < MAXNODES) {
				inodeInfo[linkcount].dev = statbuf.st_dev;
				inodeInfo[linkcount++].ino = statbuf.st_ino;
			}
		}
#endif
		if(statbuf.st_mode&S_IFDIR) {
			start_dir = TRUE;
			level++;
			strcpy(path,name_for_stat);
		}
		size += BlockSize((long)statbuf.st_size);
	}

get_next:
	while(level) {
		if(AgetNfilename(start_dir,path,fname,&place)) {
			start_dir = FALSE;
			sprintf(name_for_stat,"%s/%s",path,fname);
			goto get_info;
		}
		if(--level) {
			start_dir = FALSE;
			*rindex(path,'/') = '\0';
		}
	}

	return(size);
}

static long BlockSize(size)
long size;
{
	long tot = (size >> NSHIFT) + ((size & (BUFSIZE - 1)) ? 1 : 0);
#if !(SYS&VMSSYS) && !AXMS
	long blocks;

	if((blocks = tot) > DIRECT)
		tot += ((blocks - DIRECT - 1) >> NSHIFT) + 1;
	if(blocks > DIRECT + NINDIR)
		tot += ((blocks - DIRECT - NINDIR - 1) >> (NSHIFT << 1)) + 1;
	if(blocks > DIRECT + NINDIR + (NINDIR * NINDIR))
		tot++;
#endif
	return(tot);
}


