/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: agethom.c
   Applix ASC version: 14
   Date: Mon Nov 16 13:57:58 1998
*/
#include <stdlib.h>
#include <utloc.h>
#include <afile.h>

#if !AXMS
#	include <apwd.h>
	extern struct passwd *getpwnam();
#endif

char *AxNTgetDrive();
static char *ThimUserDirEntName = NXLT("axhome"),
			ThimUserDir[STRPATH];

/**************************************************************************
	AGetHomeDir - Obtain the user's home directory.

	AGetHomeDir() provides a generic interface for obtaining a user's
	home directory.

	On successful completion, AGetHomeDir() returns The address
	of a string which identifies the user.

	See also getpwnam(3) Agetuser(3A)

    When succesful, Agethome() returns a pointer to a static buffer
	that holds the pathname of the home directory of user.  On error,
	0 is returned.

*/

char *AGetHomeDir()
{
    char *homedir, *homedrive, *username;

#if AXMS       /* {  */
	static char home_dir[128];

	/* the DOS command prompt on NT uses these */
    if ((homedir = getenv(NXLT("HOMEPATH"))) != NULL)
	{
    	if ((homedrive = getenv(NXLT("HOMEDRIVE"))) != NULL)
		{
			sprintf( home_dir, "%s%s", homedrive, homedir);
    		return((char *)home_dir);
		}
	}

	/* Next, try the 'dos' way */
    if ((homedir = getenv(NXLT("INIT"))) != NULL)
        return(homedir);

	/* okay, try the NT way (hack for now) */
    if ((username = getenv(NXLT("USERNAME"))) != NULL)
	{
		sprintf( home_dir, "%s\\%s", AxNTgetDrive(), username);
        if ( -1 != Aaccess(home_dir, F_OK) )
           /* directory seems to exist, take it */
    	   return((char *)home_dir);
	}

    /* as a last restort, make it the "C" drive */
    return( "C:" );

#else    /*   }  AXMS  / UNIX  {  */

    /* the UNIX way **/
    if ((homedir = getenv(NXLT("HOME"))) != NULL)
        return(homedir);
    if ((homedir = getenv(NXLT("home"))) != NULL)
        return(homedir);

#if !SYSV
	{
		struct passwd *p;
		if(p = getpwnam(Agetuser()))
			return(p->pw_dir);
	}
#endif

	return((char *)NULL);
#endif  /*  } AXMS */

}


char	*AGetAxHomeDir()
{
	char	*userdir;

	/*	Allow total flexibility -- if they have an environment variable set
	 *	to determine where home is, accept it
	 */
    if ((userdir = getenv(NXLT("AXHOME"))) != NULL)
        return(userdir);

    if ((userdir = getenv(NXLT("axhome"))) != NULL)
        return(userdir);

	if (ThimUserDir[0] == '\0')
	{
		userdir = AGetHomeDir();
#if AXMS
		sprintf (ThimUserDir, "%s\\%s", userdir, ThimUserDirEntName);
#else
		sprintf (ThimUserDir, "%s/%s", userdir, ThimUserDirEntName);
#endif
	}

	return (ThimUserDir);
}
