/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: amkdir.c
   Applix ASC version: 3
   Date: Mon Nov 16 13:57:58 1998
*/
#include <errno.h>
#include <utloc.h>
#if SYS&VMSSYS
#include <stat.h>
#endif
#if SYS&AEGISO
#include </sys/ins/base.ins.c>
#include </sys/ins/pgm.ins.c>
#include </sys/ins/error.ins.c>
#include </sys/ins/streams.ins.c>
#define MAX_ARG 2
#endif


/*
 * This routine is the Applix mkdir subroutine. For Berkeley systems, it
 * executes the mkdir system call. For all other Unixes, it executes
 * the mknod system call. If the system call fails, then it falls back to
 *	the old standard of shcmd ("mkdir <path>").
 */

/*
.BEGIN MANPAGE
.NAME Amkdir - Create a directory.
.DESCRIPTION
	Amkdir() provides a generic interface to create a directory.
.ARGUMENT path
	The address of a string which identifies the directory to be created.
.SEE ALSO
	mkdir(2)
.RETURN VALUES
	Returns 0 if the directory is successfully created.
.END MANPAGE
*/

/* BEGIN SYNOPSIS */
Amkdir (path)
char *path;
/* END SYNOPSIS */
{
	char cmdline[STRPATH + 7];


#if SYS&B4_2 || HP || RT /*  {  */
	if(!mkdir(path,0770))
		return(0);
#else                    /*  }  {  */
#if AXMS                 /*  {  */
    if ( _mkdir(path) == -1 )
       /*  error */
       return( errno );
    else
       return( 0 );
#else                   /*  }  {  */
	int len;
	char d_path[STRPATH + 1],dd_path[STRPATH + 1];

	if(!mknod(path,040770,0)) {
		/*
		 * Construct "path"/. (d_path - for path with a single dot)
		 * and link "path" to it.
		 */
		if(len = strlen(path))
			memcpy(d_path,path,len);
		d_path[len++] = '/';
		d_path[len++] = '.';
		d_path[len] = '\0';
		link(path,d_path);

		/*
		 * We now need a second copy to find out the parent directory
		 * (.. or dd_path), but first concatenate another "." to d_path
		 * to construct the name "path"/..
		 */
		d_path[len++] = '.';
		d_path[len] = '\0';
		len -= 3;
		while(--len > 0 && d_path[len] != '/');
		if(len > 0) {
			/*
			 * trim the trailing directory name and preceeding
			 * "/" from the created path and use that name
			 * as the parent directory
			 */
			memcpy(dd_path,d_path,len);
			dd_path[len] = '\0';
		}
		else {
			/*
			 * We simply made the new directory in the current directory,
			 * so set dd_path to be "."
			 */
			dd_path[0] = '.';
			dd_path[1] = '\0';
		}
		link(dd_path,d_path);
		return(0);
	}
#endif    /*  }   */
#endif    /*  }   */

#if !AXMS
	/* If all else fails, use the fallback position of shcmd ("mkdir ...") */
	if(errno == EEXIST || (errno == EPERM && !Aaccess(path,0)))
		return(0);
	sprintf(cmdline,"mkdir %s",path);
	return(shcmd(cmdline));
#endif
}

