/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: aregex.c
   Applix ASC version: 3
   Date: Fri Jan  8 21:11:43 1999
*/
#include <utloc.h>
#include <aregexp.h>
#include <varargs.h>

char *ut_loc1;

static char *ut_braslist[NBRA];
static char *ut_braelist[NBRA];
static int	ut_bravar[NBRA];
static char	*ut_st[SSIZE + 1];
static char	**ut_eptr;
static char **ut_lptr;
static int	ut_cflg;

static char *ut_execute(),*ut_advance(),*ut_xpop();

static ut_cclass(/* register char *set, register int c, int af */);
static ut_xpush(/* int i, char *p */);
static ut_getrnge(/* int *i,*j, char *k */);

/*
.BEGIN MANPAGE
.NAME Aregex - Execute a compiled regular expression
.SYNOPSIS
extern char *ut_loc1;

char *Aregex(re, subject [,ret0, ...]);
char *re, *subject, *ret0 ...;
.DESCRIPTION
	Aregex() provides a generic interface to execute a compiled regular
	expression pattern against a specified string.
.ARGUMENT re
	The address of the compiled regular expression pattern which was
	returned by Aregcmp().
.ARGUMENT subject
	The address of the string which is to be compared.
.ARGUMENT ret0, ...
	The addresses of buffers to receive the values matching the regular
	expression "tagged" pieces, i.e., portions of the regular expression
	enclosed by \( and \).
.NOTE
	Ut_loc1 points to where the match began.
.SEE ALSO
	Aregcmp(3A), ed(1) free(3) malloc(3)
.RETURN VALUE
	Aregex() returns 0 on failure or a pointer to the next unmatched character
	on success.
.END MANPAGE
*/
/* BEGIN SYNOPSIS */
/* END SYNOPSIS */

char *Aregex(va_alist)
va_dcl
{
	va_list ap;
	char *addrc,*addrl;
	int in;
	char *cur;
	register char **adx;
	register char *p1,*p2;
	char *args[NBRA];
 
	va_start(ap);
	addrc = va_arg(ap,char *);
	addrl = va_arg(ap,char *);
	for(in = 0; in < NBRA; in++) {
		ut_braslist[in] = NULL;
		ut_bravar[in] = -1;
		args[in] = va_arg(ap,char *);
	}
	va_end(ap);
	ut_cflg = 0;
	cur = ut_execute(addrc,addrl);
	adx = args;
	for(in = 0; in < NBRA; in++) {
		if((p1 = ut_braslist[in]) && ut_bravar[in] >= 0) {
			p2 = adx[ut_bravar[in]];
			while(p1 < ut_braelist[in])
				*p2++ = *p1++;
			*p2 = '\0';
		}
	}
	if(!ut_cflg)
		return((addrl == cur) ? (char *)NULL : cur);
	else
		return(cur);
}
 
static char *ut_execute(p2,addrl)
register char *p2;
char *addrl;
{
	register char *p1,*p,c;
 
	p1 = addrl;
	ut_eptr = &ut_st[SSIZE];
	ut_lptr = &ut_st[0];
	if(*p2 == CIRCFL) {
		ut_loc1 = p1;
		return((p = ut_advance(p1,++p2)) ? p : addrl);
	}
	/* fast check for first character */
	if(*p2 == CCHR) {
		c = p2[1];
		do {
			if(*p1 != c)
				continue;
			ut_eptr = &ut_st[SSIZE];
			ut_lptr = &ut_st[0];
			if(p = ut_advance(p1,p2)) {
				ut_loc1 = p1;
				return(p);
			}
		} while(*p1++);
		return(addrl);
	}
	/* regular algorithm */
	do {
		ut_eptr = &ut_st[SSIZE];
		ut_lptr = &ut_st[0];
		if(p = ut_advance(p1, p2)) {
			ut_loc1 = p1;
			return(p);
		}
	} while (*p1++);
	return(addrl);
}
 
static char *ut_advance(lp,ep)
register char *lp;
register char *ep;
{
	register char *curlp;
	char *sep,*dp;
	int i,lcnt,dcnt,gflg;
 
	gflg = 0;
	while(1) {
		switch(*ep++) {
		  case CCHR:
			if(*ep++ == *lp++)
				continue;
			return((char *)NULL);
 
		  case EGRP|RNGE:
			return(lp);
		  case EGRP:
		  case GRP:
			ep++;
			continue;
 
		  case EGRP|STAR:
			ut_xpop(0);
		  case EGRP|PLUS:
			ut_xpush(0,++ep);
			return(lp);
 
		  case CDOT:
			if(*lp++)
				continue;
			return((char *)NULL);
 
		  case CDOL:
			if(!*lp)
				continue;
			lp++;
			return((char *)NULL);
 
		  case CEOF:
			ut_cflg = 1;
			return(lp);
 
		  case TGRP:
		  case TGRP|A768:
		  case TGRP|A512:
		  case TGRP|A256:
			i = (((ep[-1]&0x3)<<8) + (*ep)&0xff);
			ep++;
			ut_xpush(0,ep + i + 2);
			gflg = 1;
			ut_getrnge(&lcnt,&dcnt,&ep[i]);
			while(lcnt--)
				if(!(lp = ut_advance(lp,ep)))
					return((char *)NULL);
			ut_xpush(1,curlp=lp);
			while(dcnt-- && (dp = ut_advance(lp,ep)))
				ut_xpush(1,lp = dp);
			ep = ut_xpop(0);
			goto star;
		  case CCHR|RNGE:
			sep = ep++;
			ut_getrnge(&lcnt,&dcnt,ep);
			while(lcnt--) {
				if(*lp++ != *sep)
					return((char *)NULL);
			}
			curlp = lp;
			while(dcnt-- && *lp++ == *sep);
			if(dcnt < 0)
				lp++;
			ep += 2;
			goto star;
		  case CDOT|RNGE:
			ut_getrnge(&lcnt,&dcnt,ep);
			while(lcnt--)
				if(!(*lp++))
					return((char *)NULL);
			curlp = lp;
			while(dcnt-- && *lp++);
			if(dcnt < 0)
				lp++;
			ep += 2;
			goto star;
		  case CCL|RNGE:
		  case NCCL|RNGE:
			ut_getrnge(&lcnt,&dcnt,(ep + (*ep&0xff)));
			while(lcnt--)
				if(!ut_cclass(ep,*lp++,ep[-1]==(CCL|RNGE)))
					return((char *)NULL);
			curlp = lp;
			while(dcnt-- && ut_cclass(ep,*lp++,ep[-1]==(CCL|RNGE)));
			if(dcnt < 0)
				lp++;
			ep += (*ep + 2);
			goto star;
		  case CCL:
			if(ut_cclass(ep,*lp++,1)) {
				ep += *ep;
				continue;
			}
			return((char *)NULL);
 
		  case NCCL:
			if(ut_cclass(ep,*lp++,0)) {
				ep += *ep;
				continue;
			}
			return((char *)NULL);
 
		  case CBRA:
			ut_braslist[*ep++] = lp;
			continue;
 
		  case CKET:
			ut_braelist[*ep] = lp;
			ut_bravar[*ep] = ep[1];
			ep += 2;
			continue;
 
		  case CDOT|PLUS:
			if(!(*lp++))
				return((char *)NULL);
		  case CDOT|STAR:
			curlp = lp;
			while(*lp++);
			goto star;
 
		  case CCHR|PLUS:
			if(*lp++ != *ep)
				return((char *)NULL);
		  case CCHR|STAR:
			curlp = lp;
			while(*lp++ == *ep);
			ep++;
			goto star;
 
		  case PGRP:
		  case PGRP|A256:
		  case PGRP|A512:
		  case PGRP|A768:
			if(!(lp = ut_advance(lp,ep+1)))
				return((char *)NULL);
		  case SGRP|A768:
		  case SGRP|A512:
		  case SGRP|A256:
		  case SGRP:
			i = (((ep[-1]&0x3) << 8) + (*ep&0xff));
			ep++;
			ut_xpush(0,ep + i);
			ut_xpush(1,curlp=lp);
			while(dp = ut_advance(lp,ep))
				ut_xpush(1,lp = dp);
			ep = ut_xpop(0);
			gflg = 1;
			goto star;
 
		  case CCL|PLUS:
		  case NCCL|PLUS:
			if(!ut_cclass(ep,*lp++,ep[-1] == (CCL|PLUS)))
				return((char *)NULL);
		  case CCL|STAR:
		  case NCCL|STAR:
			curlp = lp;
			while(ut_cclass(ep,*lp++, ep[-1] == (CCL|STAR) || ep[-1] == (CCL|PLUS)));
			ep += *ep;
			goto star;
 
		  star:
			do {
				if(!gflg)
					lp--;
				else if(!(lp = ut_xpop(1)))
					break;
				if(dp = ut_advance(lp, ep)) 
					return(dp);
			} while(lp > curlp);
			return((char *)NULL);
 
		  default:
			return((char *)NULL);
		}
	}
}

static ut_cclass(set,c,af)
register char *set;
register int c;
int af;
{
	register int n;
 
	if(!c)
		return(0);
	n = *set++;
	while(--n) {
		if(*set == MINUS) {
			if((set[2] - set[1]) < 0)
				return(0);
			if(*++set <= c) {
				if(c <= *++set)
					return(af);
			}
			else
				++set;
			++set;
			n -= 2;
			continue;
		}
		if(*set++ == c)
			return(af);
	}
	return(!af);
}
static ut_xpush(i,p)
int i;
char *p;
{
	if(i)
		*ut_lptr++ = p;
	else
		*ut_eptr-- = p;
}
static char *ut_xpop(i)
int i;
{
	if(i)
		return((ut_lptr < &ut_st[0]) ? (char *)NULL : *--ut_lptr);
	else
		return((ut_eptr > &ut_st[SSIZE]) ? (char *)NULL : *++ut_eptr);
}
static ut_getrnge(i,j,k)
int *i,*j;
char *k;
{
	*i = *k++&0xff;
	if((*k&0xff) == 0xff)
		*j = 20000;
	else
		*j = (*k&0xff) - *i;
}
