/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: armdir.c
   Applix ASC version: 7
   Date: Mon Nov 16 13:57:58 1998
*/
#include <direct.h>
#include <utloc.h>
#include <exitcode.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>


/* BEGIN SYNOPSIS */
/*
.BEGIN MANPAGE
.NAME Armdir - Clean or delete a directory.
.SYNOPSIS
Armdir(path,dirToo)
char *path;
int dirToo;
.ARGUMENT path
    The address of a string which identifies a directory or file.
.ARGUMENT dirToo
    0 remove only files.

    1 remove all files and directories including path.

    2 remove all files and directories excluding path.
.DESCRIPTION
    Armdir() provides a generic interface for cleaning or deleting
    a directory.
.SEE ALSO
    rmdir(2)
.RETURN VALUE
    Armdir() returns 0 on success and -1 on error.
.END MANPAGE
*/
/* END SYNOPSIS */




Armdir(filename, dirToo)
char *filename;
int dirToo;
{
    int init = TRUE;
    int err = 0;
    char temp[STRPATH + 1];
    char name[AX_MAX_FILENAME + 1];
    struct stat statbuf;
    char *place = NULL, *p;

#define SLASH_STRING "\\"
#define SLASH_CHAR '\\'
#define WHOLE_ROOT "\\*"

    if (Astat(filename, &statbuf))
        {
        if (errno == ENOENT)
            return (0);
        return (-1);
        }
    if (!strcmp(filename, SLASH_STRING) || !strcmp(filename, WHOLE_ROOT))
        {
        /* forget about deleting the entire disk! */
        errno = EACCES;
        return (-1);
        }

    if ((statbuf.st_mode & S_IFMT) == S_IFDIR)
        {
        /* the file is a directory */
        while (AgetNfilename(init, filename, name, &place))
            {
            init = FALSE;
            if (*name == '.' && dirToo != TRUE)
                continue;
            sprintf(temp, "%s%s%s", filename, SLASH_STRING, name);
            err = Armdir(temp, (dirToo == 2) ? TRUE : dirToo);
            }
        if (!err && dirToo == TRUE)
            err = rmdir(filename);
        }
    else
        {
        /* If we are deleting the directory too, you can always delete
         * the file. Look for a '/'. If there is none, then if the 1st
         * char is not a '.', then we can delete the file. If there is
         * a '/', then if the 1st char after the '/' is not a '.', then
         * we can delete the file. Otherwise, leave the file there */
        if (dirToo == TRUE ||
         ((p = strrchr(filename, SLASH_CHAR)) == NULL && *filename != '.') ||
            (p && *(p + 1) != '.'))
            err = unlink(filename);
        }
    return (err);
}

