/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: asprint.c
   Applix ASC version: 3
   Date: Mon Nov 16 13:57:58 1998
*/
#include <ctype.h>
#include <varargs.h>
#define NO_APX_INTERNATIONAL 1
#include <aos.h>

#ifdef Asprintf
#undef Asprintf
#endif

LINTCOMMENT(ARGSUSED2)
Asprintf(va_alist)
va_dcl
{
	va_list args;
	char *str;
	char *fmt;
#if SYS&VMSSYS
	static char nil[] = NXLT("(null)");
	int c;
	int left,sign,alternate;
	int width,precision,pad;
	int size,base;
	char buffer[32 + 1];
	char *p;
	long value;
	unsigned long uvalue;
	int len;
	char *(*func)();
	int decpt;
	int fsign;
	char *exp;
	extern char *ecvt(),*fcvt();
#endif

	va_start(args);
	str = va_arg(args,char *);
	fmt = va_arg(args,char *);

#if SYS&VMSSYS
#define pushc_m(c) *str++ = c;
	while(c = *fmt++) {
		if(c != '%') {
			pushc_m(c)
			continue;
		}

		left = 0;
		sign = 0;
		alternate = 0;
		width = 0;
		precision = -1;
		pad = ' ';
		size = 0;
		base = 0;

		while(c = *fmt++) {
			switch(c) {
			  case '-':
				left++;
				continue;
			  case '+':
				sign = c;
				continue;
			  case ' ':
				if(!sign)
					sign = c;
				continue;
			  case '#':
				alternate++;
				continue;
			}
			break;
		}

		if(c == '0') {
			pad = c;
			c = *fmt++;
		}
		if(c == '*') {
			width = va_arg(args,int);
			c = *fmt++;
		}
		else while(isdigit(c)) {
			width = (10 * width) + c - '0';
			c = *fmt++;
		}
 
		if(c == '.') {
			precision = 0;
			if((c = *fmt++) == '*') {
				precision = va_arg(args,int);
				c = *fmt++;
			}
			else while(isdigit(c)) {
				precision = (10 * precision) + c - '0';
				c = *fmt++;
			}
		}

		if(c == 'l') {
			size++;
			c = *fmt++;
		}
 
		p = buffer;
		func = 0;

		switch(c) {
		  case 'u':
			sign = 0;
		  case 'd':
			base = 10;
			alternate = 0;
			break;
		  case 'o':
			base = 8;
			sign = 0;
			break;
		  case 'X':
		  case 'x':
			base = 16;
			sign = 0;
			break;
		  case 'g':
		  case 'G':
			if(!precision)
				precision++;
			base++;
		  case 'e':
		  case 'E':
			if(precision == -1)
				precision = 6;
			if(!base) {
				precision++;
				base++;
			}
			func = ecvt;
		  case 'f':
			if(!func)
				func = fcvt;
			if(precision == -1)
				precision = 6;
			p = (*func)(va_arg(args,double),precision,&decpt,&fsign);
			precision -= base;
			break;
		  case 's':
			p = va_arg(args,char *);
			if(!p)
				p = nil;
			break;
		  case 'c':
			*p = va_arg(args,int);
			*(p + 1) = '\0';
			break;
		  case '\0':
			goto ret;
		  default:				/* also covers %% */
			*p = c;
			*(p + 1) = '\0';
		}

		if(func) {
			if(fsign)
				sign = '-';
			base = strlen(p);
			if(c == 'g' || c == 'G') {
				if(decpt < -4 || decpt > base)
					c -= 2;
				if(!alternate && base > 1)
					while(*(p + base - 1) == '0' && --base > 1);
			}
			exp = buffer;
			value = (decpt <= 0) ? (1 - decpt) : (decpt - 1);
			if(c == 'e' || c == 'E') {
				pad = ' ';
				do
					*exp++ = '0' + value % 10;
				while(value /= 10);
				if(exp == buffer + 1)
					*exp++ = '0';
				*exp++ = (decpt <= 0) ? '-' : '+';
				len = exp - buffer + 1;
				if(alternate || base > 1)
					len++;
			}
			else {
				if(decpt <= 0)
					len = value;
				else if(decpt > base)
					len = decpt - base;
				else
					len = 0;
				if(alternate || precision > 0)
					len++;
			}
			width -= base + len;
			if(sign)
				width--;

			if(pad == '0' && sign) {
				pushc_m(sign)
				sign = 0;
			}
			if(!left || pad == '0')
				while(width-- > 0)
					pushc_m(pad)
			if(sign)
				pushc_m(sign)
			if(exp == buffer && decpt <= 0) {
				while(value--) 
					pushc_m('0')
				if(base || alternate)
					pushc_m('.')
			}
			while(base--) {
				pushc_m(*p++)
				if(value--)
					continue;
				if(base || alternate)
					pushc_m('.')
			}
			while(value-- > 0) {
				pushc_m('0')
				if(!value && alternate)
					pushc_m('.')
			}
			if(exp != buffer) {
				pushc_m(c)
				while(exp-- > buffer)
					pushc_m(*exp)
			}
			if(left && pad != '0')
				while(width-- > 0)
					pushc_m(' ')
			continue;
		}

		if(base) {
			if(c == 'u') {
				if(size)
					uvalue = va_arg(args,unsigned long);
				else
					uvalue = va_arg(args,unsigned int);
			}
			else {
				if(size)
					value = va_arg(args,long);
				else
					value = va_arg(args,int);
				if(value < 0) {
					sign = '-';
					uvalue = -value;
				}
				else
					uvalue = value;
			}

			do {
				if((value = uvalue % base) < 10)
					*p++ = '0' + value;
				else
					*p++ = ((c == 'X') ? 'A' : 'a') + value - 10;
			}
			while(uvalue /= base);
			len = p - buffer;

			if((precision -= len) > 0)
				width -= precision;
			width -= len;
			if(sign) {
				width--;
				if(pad == '0') {
					pushc_m(sign)
					sign = 0;
				}
			}
			else if(alternate) {
				width--;
				if(pad == '0')
					pushc_m('0')
				if(base == 16) {
					width--;
					if(pad == '0')
						pushc_m(c)
				}
			}
			if(!left || pad == '0') {
				while(width-- > 0)
					pushc_m(pad)
			}
			if(sign)
				pushc_m(sign)
			else if(alternate && pad == ' ') {
				pushc_m('0')
				if(base == 16)
					pushc_m(c)
			}
			while(precision-- > 0)
				pushc_m('0')
			while(len--)
				pushc_m(*(--p))
			if(left && pad != '0')
				while(width-- > 0)
					pushc_m(' ')
			continue;
		}

		pad = ' ';
		len = strlen(p);
		if(precision != -1 && len > precision)
			len = precision;
		width -= len;
		if(!left)
			while(width-- > 0)
				pushc_m(pad)
		while(len--)
			pushc_m(*p++)
		if(left)
			while(width-- > 0)
				pushc_m(pad)
	}
ret:
	pushc_m('\0')

#else
	vsprintf(str,fmt,args);
#endif

	va_end(args);
	return(0);
}
