/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: aunique.c
   Applix ASC version: 4
   Date: Mon Nov 16 13:57:59 1998
*/
#include <utloc.h>


/* Generate a unique name given a buffer containing a filename prefix
   (e.g., "/util/fs/tmp/zzfs").  Aunique will add unique stuff to the end
   of the string (e.g., "/util/fs/tmp/zzfs1234567_123"), and return a
   pointer to the string.

   BE SURE you have enough room in the string for the unique stuff
   which Aunique will tack on to the end. Leave MAXSIMPLE+1 
   characters.

   The unique filename is built from the time (to the second) plus an
   additional alpha character which is incremented each time the routine
   is called (modulo 'z').  The uniqueness of the filenames produced by
   this routine therefore  assumes that this routine will not be called
   27 times within one second.
*/

#if SUN
#define MAXSIMPLE   21   /* This WORKS on Suns */
#else
#define MAXSIMPLE   12   /* short enough even for System-V! */
#endif
#define MAXFILE     256


char *Aunique(s)
char *s;
{

	static char filename [MAXFILE];
	char temp[MAXSIMPLE];
	int i, num_time_digits;
	char *p;
	static char Unique_str[] = "a";
	long dumbtime;

    /* first get any profix the user passed us */
    if( s != (char *)NULL) 
        {
        /* we were passed a non-null filename prefix */
        if(p = strrchr(s,'/'))
           {
           /* there is a slash, pick up the rest of the string */
           strcpy( filename, (p+1) );
           *(p+1) = 0;                  /* take prefix off of stem */
           } 
        else
           {
           /* use the string we were passed */
           strcpy( filename, s );
           *s = '\0';                /* we return the whole thing later */
           }
        }
    else
        /* no filename at all */
        filename[0] = '\0';


    /* start with the unique character */
    strcat( filename, Unique_str );

    /* increment the unique character  */
    Unique_str[0] = Unique_str[0] + 1;
    if ( Unique_str[0] > 'z' )
       Unique_str[0] = 'a';

    /* add on the host-id, this is important for distributed login where
    *   several hosts may be generating temp file names into the same
    *   fstmp directory                                            */
    Asprintf( temp, "%lx", 1);
    strcat( filename, temp );

    /* add on as many of the low order digits of the time as will
     * fit within the MAXSIMPLE size filename.  By using the low order
     * digits we pick up the seconds, getting maximum granularity  
     */
    
    /*
     * time() IS A SYSTEM CALL you should ALWAYS provide a VALID address !!
     */

    Asprintf( temp, "%ld", time(&dumbtime) );
    num_time_digits = MAXSIMPLE - strlen( filename );

    /* we start part way through the time string to get the last digits */
    i = strlen(temp) - num_time_digits;
    if ( i< 0 )
        i = 0;
    strcat( filename, &temp[i] );

    if ( s != (char *)NULL )
       {
       strcat( s, filename );
       return(s);
       }
    else
       {
       return( filename );
       }
}

