/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: bits.c
   Applix ASC version: 3
   Date: Mon Nov 16 13:57:59 1998
*/
#include <utloc.h>

#ifdef setbit
#undef setbit
#endif

#ifdef clrbit
#undef clrbit
#endif

/* Module: bits.c
 *
 * Routines to SET, CLEAR, and TEST a bit. Each function requires a
 * character pointer to the 'bit buffer' and an unsigned bit number,
 * which represents the offset into the buffer of the specified
 * bit to set, clear or test.
 */

#define CHARSIZE 8

/*
.BEGIN MANPAGE
.NAME setbit - Set a bit.
.ARGUMENT ptr
	The address of a bit buffer.
.ARGUMENT num
	The bit number to set.
.DESCRIPTION
	Setbit() sets a specified bit in a character array.
.NOTES
	The highest bit in the first character is bit 0.
.SEE ALSO
	clrbit(3A) tstbit(3A)
.RETURN VALUE
	None.
.END MANPAGE
*/

/* BEGIN SYNOPSIS */
setbit(ptr,num)
char *ptr;
REGISTER unsigned int num;
/* END SYNOPSIS */
{
	REGISTER unsigned char *theptr;

	theptr = (unsigned char *)(ptr + (num/CHARSIZE));
	*theptr |= 1 << (int)(CHARSIZE - 1 - (num % CHARSIZE));

}

/*
.BEGIN MANPAGE
.NAME clrbit - Clear a bit.
.ARGUMENT ptr
	The address of a bit buffer.
.ARGUMENT num
	The bit number to clear.
.DESCRIPTION
	Clrbit() clears a specified bit in a character array.
.NOTES
	The highest bit in the first character is bit 0.
.SEE ALSO
	setbit(3A) tstbit(3A)
.RETURN VALUE
	None.
.END MANPAGE
*/

/* BEGIN SYNOPSIS */
clrbit(ptr,num)
char *ptr;
REGISTER unsigned int num;
/* END SYNOPSIS */
{
	REGISTER unsigned char *theptr;

	theptr = (unsigned char *)(ptr + (num/CHARSIZE));
	*theptr &= ~(1 << (int)(CHARSIZE - 1 - (num % CHARSIZE)));
}

/* 'tstbit': TEST a specified bit.
 * Inputs: 1) Pointer to a character buffer,
 *	   2) Bit number to test within the buffer.
 * Returns: '0', if bit is off else '1' if bit is on.
*/

/*
.BEGIN MANPAGE
.NAME tstbit - Test a bit.
.ARGUMENT ptr
	The address of a bit buffer.
.ARGUMENT num
	The bit number to test.
.DESCRIPTION
	Tstbit() tests a specified bit in a character array.
.NOTES
	The highest bit in the first character is bit 0.
.SEE ALSO
	clrbit(3A) tstbit(3A)
.RETURN VALUE
	Tstbit() returns 1 if the bit is set and 0 if it is not set.
.END MANPAGE
*/

/* BEGIN SYNOPSIS */
tstbit(ptr,num)
char *ptr;
REGISTER unsigned int num;
/* END SYNOPSIS */
{
	if((*(ptr + (num/CHARSIZE))) & (1 << (int)(CHARSIZE - 1 - (num%CHARSIZE))))
		return(1);
	return(0);
}
