/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: datecmp.c
   Applix ASC version: 4
   Date: Mon Dec 21 21:12:33 1998
*/
#include <aos.h>
#include <datetime.h>

/*
.BEGIN MANPAGE
.NAME datecmp - Compare Applix format date-time values.
.SYNOPSIS
#include <datetime.h>
.ARGUMENT arg1
	An Applix format date-time as obtained from ParseDate(3A).
.ARGUMENT arg2
	An Applix format date-time as obtained from ParseDate(3A).
.DESCRIPTION
    Datecmp() compares two dates in Applix format.
    If either date is a date-only value (no time specified),
	only the date portion of the two arguments are compared,
	otherwise both date and time values are compared.
	Seconds value never enter into the comparison.
    DTNULL is considered less than any date value.
.BUGS
    A date-time value with a time of 12:00AM (midnight) is
    considered to be a date-only value.
.SEE ALSO
	ParseDate(3A) ut_daycmp(3A) ut_timecmp(3a)
.RETURN VALUE
	Datecmp() -1, 0 or 1 depending on whether arg1 is less than, equal to,
	or greater than arg2.
.END MANPAGE
*/

/* BEGIN SYNOPSIS */
datecmp(arg1,arg2)
unsigned long arg1;
unsigned long arg2;
/* END SYNOPSIS */
{
	unsigned long lc;
	unsigned long tmp1, tmp2;

	if( arg1 == arg2 )
		return( 0 );		/* handles case of DTNULL==DTNULL among others */
	if( arg1 == DTNULL )
		return( -1 );
	if( arg2 == DTNULL )
		return( 1 );

	tmp1 = arg1;
	tmp2 = arg2;
	
	/* base comparison on everything except the seconds field. */

	lc = 60/DTTIMEGRANULARITY;
	tmp1 /= lc;
	tmp2 /= lc;		

	/* determine if comparison is to be based on date or time values only. */

	if( (tmp1 / 1440) == 0 || (tmp2 / 1440) == 0 ) {
		tmp1 %= 1440;
		tmp2 %= 1440;
	}
	else if( (tmp1 % 1440) == 0 || (tmp2 % 1440) == 0) {
		tmp1 /= 1440;	/*  60*24  */
		tmp2 /= 1440;
	}

	return( (tmp1 < tmp2) ? -1 : ( (tmp1 == tmp2) ? 0 : 1 ) );
}
