/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: err_map.c
   Applix ASC version: 13
   Date: Mon Dec 21 21:12:34 1998
*/
#include <sys/types.h>
#include <sys/stat.h>
#include <axsocket.h>
#include <errno.h>
#include <signal.h>
#include <getstr10.h>

#define THIM_STATIC 1    /** do not redefine such things as "exit()"  */
#include <aos.h>
#include <eldata.h>

#if AXMS
#include <errnohack.h>
#endif



/*************************************************************************
function: ErrnoErr - Return an Alis errcode based on errno value
*/
typeErr ErrnoErr ()
{
#if USE_GLOBAL_ALLOC           /*  {  */
    /* we are runnin in NT, as a standalone library, so do NOT map the */
    /* error code to some Alis value.                                  */
    return( errno );
#else                          /*  }  {  */
    switch (errno) { 
    case 0:
        return( FEErrno );      /* errno not set */
    case EPERM:
        return( FEEPERM  );     /* Not owner */
    case ENOENT:
        return( FEENOENT  );    /* No such file or directory */
    case ESRCH:
        return( FEESRCH  );     /* No such process */
    case EINTR:
        return( FEEINTR  );     /* Interrupted system call */
    case EIO:
        return( FEEIO  );       /* I/O error */
    case ENXIO:
        return( FEENXIO  );     /* No such device or address */
    case E2BIG:
        return( FEE2BIG  );     /* Arg list too long */
    case ENOEXEC:
        return( FEENOEXEC  );   /* Exec format error */
    case EBADF:
        return( FEEBADF  );     /* Bad file number */
    case ECHILD:
        return( FEECHILD  );        /* No children */
    case EAGAIN:
        return( FEEAGAIN  );        /* No more processes */
    case ENOMEM:
        return( FEENOMEM  );        /* Not enough core */
    case EACCES:
        return( FEEACCES  );        /* Permission denied */
    case EFAULT:
        return( FEEFAULT  );        /* Bad address */
#if !AXMS
    case ENOTBLK:
        return( FEENOTBLK  );       /* Block device required */
#endif
    case EBUSY:
        return( FEEBUSY  );     /* Mount device busy */
    case EEXIST:
        return( FEEEXIST  );        /* File exists */
    case EXDEV:
        return( FEEXDEV  );     /* Cross-device link */
    case ENODEV:
        return( FEENODEV  );        /* No such device */
    case ENOTDIR:
        return( FEENOTDIR  );       /* Not a directory*/
    case EISDIR:
        return( FEEISDIR  );        /* Is a directory */
    case EINVAL:
        return( FEEINVAL  );        /* Invalid argument */
    case ENFILE:
        return( FEENFILE  );        /* File table overflow */
    case EMFILE:
        return( FEEMFILE  );        /* Too many open files */
    case ENOTTY:
        return( FEENOTTY  );        /* Not a typewriter */
#if !AXMS
    case ETXTBSY:
        return( FEETXTBSY  );       /* Text file busy */
#endif
    case EFBIG:
        return( FEEFBIG  );     /* File too large */
    case ENOSPC:
        return( FEENOSPC  );        /* No space left on device */
    case ESPIPE:
        return( FEESPIPE  );        /* Illegal seek */
    case EROFS:
        return( FEEROFS  );     /* Read-only file system */
    case EMLINK:
        return( FEEMLINK  );        /* Too many links */
    case EPIPE:
        return( FEEPIPE  );     /* Broken pipe */

    /* math software */
    case EDOM:
        return( FEEDOM  );      /* Argument too large */
    case ERANGE:
        return( FEERANGE  );        /* Result too large */

    /* non-blocking and interrupt i/o */
/* solaris/ibm defines EWOULDBLOCK to EAGAIN */
/* Dec defines EDEADLK to EWOULDBLOCK  */
#if !(SUN4S || NCR || UNIXWARE || SUNIS || PMAX || RIOS || AVIION || ALPHA || SGI || LINUX || NEC_EWS)
    case EWOULDBLOCK:
        return( FEEWOULDBLOCK  );       /* Operation would block */
#endif
    case EINPROGRESS:
        return( FEEINPROGRESS  );       /* Operation now in progress */
    case EALREADY:
        return( FEEALREADY  );      /* Operation already in progress */
    /* ipc/network software */

    /* argument errors */
    case ENOTSOCK:
        return( FEENOTSOCK  );      /* Socket operation on non-socket */
    case EDESTADDRREQ:
        return( FEEDESTADDRREQ  );      /* Destination address required */
    case EMSGSIZE:
        return( FEEMSGSIZE  );      /* Message too long */
    case EPROTOTYPE:
        return( FEEPROTOTYPE  );        /* Protocol wrong type for socket */
    case ENOPROTOOPT:
        return( FEENOPROTOOPT  );       /* Protocol not available */
    case EPROTONOSUPPORT:
        return( FEEPROTONOSUPPORT  );   /* Protocol not supported */
    case ESOCKTNOSUPPORT:
        return( FEESOCKTNOSUPPORT  );   /* Socket type not supported */
    case EOPNOTSUPP:
        return( FEEOPNOTSUPP  );    /* Operation not supported on socket */
    case EPFNOSUPPORT:
        return( FEEPFNOSUPPORT  );  /* Protocol family not supported */
    case EAFNOSUPPORT:
        return( FEEAFNOSUPPORT  );  /* Address family not supported by protocol family */
    case EADDRINUSE:
        return( FEEADDRINUSE  );        /* Address already in use */
    case EADDRNOTAVAIL:
        return( FEEADDRNOTAVAIL  ); /* Can't assign requested address */

    /* operational errors */
    case ENETDOWN:
        return( FEENETDOWN  );      /* Network is down */
    case ENETUNREACH:
        return( FEENETUNREACH  );       /* Network is unreachable */
    case ENETRESET:
        return( FEENETRESET  );     /* Network dropped connection on reset */
    case ECONNABORTED:
        return( FEECONNABORTED  );      /* Software caused connection abort */
    case ECONNRESET:
        return( FEECONNRESET  );        /* Connection reset by peer */
    case ENOBUFS:
        return( FEENOBUFS  );       /* No buffer space available */
    case EISCONN:
        return( FEEISCONN  );       /* Socket is already connected */
    case ENOTCONN:
        return( FEENOTCONN  );      /* Socket is not connected */
    case ESHUTDOWN:
        return( FEESHUTDOWN  );     /* Can't send after socket shutdown */
    case ETOOMANYREFS:
        return( FEETOOMANYREFS  );      /* Too many references: can't splice */
    case ETIMEDOUT:
        return( FEETIMEDOUT  );     /* Connection timed out */
    case ECONNREFUSED:
        return( FEECONNREFUSED  );      /* Connection refused */

    /* */
    case ELOOP:
        return( FEELOOP  );     /* Too many levels of symbolic links */
    case ENAMETOOLONG:
        return( FEENAMETOOLONG  );  /* File name too long */

    /* should be rearranged */
    case EHOSTDOWN:
        return( FEEHOSTDOWN  );     /* Host is down */
    case EHOSTUNREACH:
        return( FEEHOSTUNREACH  );  /* No route to host */
/* rs600 defines ENOTEMPTY to EEXIST */
#if !RIOS
    case ENOTEMPTY:
        return( FEENOTEMPTY  );     /* Directory not empty */
#endif
    /* quotas & mush */
#if !SUN4S && !HP800V3 && !SUNIS && !SGI && !SCOUNIX && !UNIXWARE && !NCR && !NEC_EWS && !LINUX
    case EPROCLIM:
        return( FEEPROCLIM  );      /* Too many processes */
#endif
#if !SCOUNIX
    case EUSERS:
        return( FEEUSERS  );        /* Too many users */
#endif
#if !SUN4S && !SUNIS && !SCOUNIX && !UNIXWARE && !NCR && !NEC_EWS
    case EDQUOT:
        return( FEEDQUOT  );        /* Disc quota exceeded */
#endif

    /* Network File System */
    case ESTALE:
        return( FEESTALE  );        /* Stale NFS file handle */
    case EREMOTE:
        return( FEEREMOTE  );       /* Too many levels of remote in path */

    /* streams */
#if !PMAX && !AXMS
    case ENOSTR:
        return( FEENOSTR  );        /* Device is not a stream */
    case ETIME:
        return( FEETIME  );     /* Timer expired */
#if !SCOUNIX
    case ENOSR:
        return( FEENOSR  );     /* Out of streams resources */
#endif
#endif
#if !AXMS
    case ENOMSG:
        return( FEENOMSG  );        /* No message of desired type */
#endif
#if !PMAX && !AXMS
    case EBADMSG:
        return( FEEBADMSG  );       /* Trying to read unreadable message */
#endif

    /* SystemV IPC */
#if !AXMS
    case EIDRM:
        return(     FEEIDRM  );     /* Identifier removed */
#endif

    /* SystemV Record Locking */
    case EDEADLK:
        return(     FEEDEADLK  );       /* Deadlock condition. */
    case ENOLCK:
        return(     FEENOLCK  );        /* No record locks available. */
#if !AXMS && !PMAX
    case EPROTO:
        return(     FEEPROTO  );        /* Protocol error */
#endif
    default:
        return(FENOMATCH);          /* no match found */

    }
#endif                             /*  } not NT standalone */
}




