/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: formatd.c
   Applix ASC version: 54
   Date: Sat Jan 23 21:14:35 1999
*/

/*
.BEGIN MANPAGE
.NAME Formatdt - Format an Applix date-time value.
.SYNOPSIS
#include <datetime.h>
.ARGUMENT dtvalue
	An Applix format date as obtained from ParseDate(3A).
.ARGUMENT formatopt
	Specifies the format to be generated.
	The legal values are defined in "datetime.h"
.ARGUMENT retstrP
	The address of a buffer that will receive the formatted date-time string.
	The buffer should be at least 30 bytes long.
.DESCRIPTION
	Formatdt() generates a character string rendition of the given date-time
	value.
.SEE ALSO
	ParseDate(3A)
.RETURN VALUE
	None.
.END MANPAGE
*/

#include <stdio.h>
#include <axtime.h>
#include <string.h>
#include <aos.h>
#include <stdlib.h>
#include <thimble.h>		/*	for the THIM_FIND_X #def */
#include <dt.h>
#include <datetime.h>
#include <getstr9.h>
#include <formatd.p>


int	MaxDtFmt = 0,
	MaxTmFmt = 0;

dttmfmt	dtfmt[100],		/*	holds date formats read in from datetime.sp */
	tmfmt[50];		/*	holds time formats read in from datetime.sp */

extern char		*AxLangDir;			/* for location of datetime.sp */
extern char		*DirSlash;			/* for location of datetime.sp */

mbchar 	 DysPermo[12] = {31,29,31,30,31,30,31,31,30,31,30,31};

#define	MAX_LANGS	21		/*	must match context.c's MAX_LANGS define */

int
id_lang()
{
    mbchar	*AxL;
    int		AxLanguage;

    AxL = (mbchar *)getenv(NXLT("AXLANG"));
    if ( AxL )
        AxLanguage = atoi(AxL);
    else{   
        AxL = UPbyName(NXLT("axDisplayLanguage"));
        if ( !AxL || AxL[0] == (char)0 )
            AxLanguage = 1;
        else
            AxLanguage = atoi(AxL);
    }
    if ( AxLanguage < 1 || AxLanguage >= MAX_LANGS )
        AxLanguage = 1;

	return (AxLanguage);
}

mbchar	*
get_date_ordinal(day, st, nd, rd, th)
int	day;
mbchar	*st, *nd, *rd, *th;
{
	switch (day % 10){
		case 1:
			if (day == 11)
				return (th);
			else
				return ( st );
			break;
		case 2:
			if (day == 12)
				return (th);
			else
				return ( nd );
			break;
		case 3:
			if (day == 13)
				return (th);
			else
				return ( rd );
			break;
		default:
			return ( th );
			break;
	}
}


/*	Has been turned into a wrapper for Formadt_lang; see
 *	edit #25 for reason */

typeErr
Formatdt (dtvalue, datefmt, retstrP)
unsigned long	dtvalue;	   		/*	incoming Applix-fmt datetime */
int				datefmt;
mbchar			*retstrP;			 
{
	typeErr retval = AOK,
	lang = 1;			/*	default to English */

	lang = id_lang();				/*	identify running language */

	retval = Formatdt_lang_val(dtvalue, datefmt, retstrP, lang, FALSE, FALSE);
	return (retval);
}

typeErr
Formatdt_unix (dtvalue, datefmt, retstrP)
unsigned long	dtvalue;	   		/*	incoming UNIX datetime */
int				datefmt;
mbchar			*retstrP;			 
{
	return(Formatdt_lang_val(dtvalue, datefmt, retstrP, id_lang(), FALSE,TRUE));
}

typeErr
axFormatdt (dtvalue, datefmt, retstrP, fullYr)
unsigned long	dtvalue;	   		/*	incoming Applix-fmt datetime */
int				datefmt;
mbchar			*retstrP;			 
int				fullYr;	/* TRUE to always generate full year for year */
{
	return(Formatdt_lang_val(dtvalue, datefmt, retstrP, id_lang(), fullYr,FALSE));
}

typeErr
Formatdt_lang (dtvalue, datefmt, retstrP, lang, fullYr)
unsigned long	dtvalue;	   		/*	incoming Applix-fmt datetime */
int				datefmt;
mbchar			*retstrP;			 
int				lang;
int				fullYr;	/* TRUE to always generate full year for year */
{
	return(Formatdt_lang_val(dtvalue, datefmt, retstrP, lang, fullYr,FALSE));
}

typeErr
Formatdt_lang_components (day, month, year, hour, min, sec, datefmt, retstrP, lang, fullYr)
int				day;
int				month;
int				year;
int				hour;
int				min;
int				sec;
int				datefmt;
mbchar			*retstrP;			 
int				lang;
int				fullYr;	/* TRUE if always generate full year for year */
{
	AxTimeStruct axt;
	
	axt.day = day;
	axt.month = month;
	axt.year = year;
	axt.hour = hour;
	axt.minute = min;
	axt.second = sec;
	axt.fraction = 0;
	
	return (Formatdt_AxTime_lang(axt,datefmt,retstrP,lang,fullYr));

}

typeErr
Formatdt_lang_val (dtvalue, datefmt, retstrP, lang, fullYr, unixval)
unsigned long	dtvalue;	   	/*	incoming Applix-fmt, or UNIX datetime */
int				datefmt;
mbchar			*retstrP;			 
int				lang;
int				fullYr;	/* TRUE if always generate full year for year */
int				unixval; /* TRUE if dtvalue is a UNIX datetime value */
{
	int	day 	= 0,
		month 	= 0,
		year 	= 0,
		hour 	= 0,
		min 	= 0,
		sec 	= 0;

	/* what to do with no date value to parse */
	if( dtvalue == DTNULL){
		MBstrcpy (retstrP, "");
		return (!AOK);
	}
	/* convert dtvalue to AxTimeStruct */
	if (unixval)
		decodeUnixDates (dtvalue, &day, &month, &year, &hour, &min, &sec);
	else{
		decodeApplixDates (dtvalue, &day, &month, &year, &hour, &min, &sec);
	}
		return(Formatdt_lang_components(day, month, year, hour, min, sec, datefmt, retstrP, lang, fullYr));
}

typeErr
Formatdt_AxTime (axt, datefmt, retstrP)
AxTimeStruct	axt; /* the time, decomposed. */
int				datefmt;
mbchar			*retstrP;			 
{
	return(Formatdt_AxTime_lang(axt, datefmt, retstrP, id_lang(),FALSE));
}

typeErr
Formatdt_AxTime_lang (axt, datefmt, retstrP, lang, fullYr)
AxTimeStruct	axt; /* the time, decomposed. */
int				datefmt;
mbchar				*retstrP;			 
int				lang;
int				fullYr;	/* TRUE if always generate full year for year */
{
	mbchar	wkdy[STRPATH],		/*	the abb'd weekday name */
		weekday[STRPATH],	/*	the full weekday name */
		mnth[STRPATH],		/*	the abb'd month name */
		monthN[STRPATH],	/*	the full month name */
		th[STRPATH],		/*	whatever serves for 'n-th' */
		st[STRPATH],		/*	whatever serves for '1st' */
		nd[STRPATH],		/*	whatever serves for '2nd' */
		rd[STRPATH],		/*	whatever serves for '3rd' */
		am[STRPATH],		/*	whatever serves for 'am' */
		pm[STRPATH],		/*	whatever serves for 'pm' */
		emp_date[STRPATH];	/*	support Japanese format */
	int	wfmt = 0,
		tfmt = 0,
		dfmt = 0,
		day3 = 0, day = 0, month = 0, year = 0,
		hour = 0, min = 0, sec = 0,
		i = 0,
		preceding = 0,
		use_ampm = FALSE,
		len = 0, chr = '\0',
		list = 0;				/*	list number to get from getStr */
	static 	int	loaded = FALSE;			/*	are the datetime fmts loaded? */
    	int 	did_emperor_date = FALSE;
    	
	if ( ! loaded && ! fmts_loaded()){
		if (load_datetime_fmts() != AOK)
			return (!AOK);
		else
			loaded = TRUE;
	}

	/* start with a clean slate */
    	*retstrP = (mbchar)0;

	/* what time is it? */
	day = axt.day;
	month = axt.month;
	year = axt.year;
	hour = axt.hour;
	min = axt.minute;
	sec = axt.second + ((int)(axt.fraction + .5)); /* round to second */
	day3 = ax_get_3digit_day(day, month, year);

	/* set what the formats are */

#if BUILD_KANJI
/* Make ugly changes here because of some unknown reasons.
 * when the date format in SS=>Style=>DocumentSettings=>Date=>dd.mm.yyyy
 * was set, the date string would not displayed. we trace it to ssdspstr.c,
 * NewDspStr() and to display.c DisplayLine(), still not found the answer.
 */
   if (datefmt == 0)
       datefmt = 13;
#endif

	wfmt	  = datefmt/DTWFBASE;
	datefmt  -= wfmt*DTWFBASE;
	tfmt	  = datefmt/DTTFBASE;
	dfmt	  = datefmt - tfmt*DTTFBASE;

	/*	check for limits; if documents were exchanged without a datetime.sp
	 *	being exchanged to support differently-defined user-defined formats,
	 *	we won't know it, and the cell or area may use user-defined formats
	 *	that aren't defined in current one.  If the dfmt or tfmt are above
	 *	the current max's, reset them to 1.
	 */

	if (dfmt > MaxDtFmt)
		dfmt = 1;

	if (tfmt > MaxTmFmt)
		tfmt = 1;

	if ((set_names(lang, wkdy, weekday, mnth, monthN, month, day, year,
			am, pm, th, st, nd, rd)) != AOK)
		return (!AOK);

	if (wfmt % 2 == 1){
		switch (wfmt * DTWFBASE){
			case DTWFFBEG:
				MBstrcat (retstrP, weekday);
				break;
			case DTWFBEG:
				MBstrcat (retstrP, wkdy);
				break;
		}
		preceding = TRUE;
	}

	if (dfmt){
		if (preceding)
			MBstrcat (retstrP, " ");

		preceding = TRUE;

		use_ampm = check_for_ampm (&dtfmt[dfmt], &hour);

		for (i = 0; i < dtfmt[dfmt].toklen; i++){
			switch (dtfmt[dfmt].tok[i]){
				case DU_FMON:
					MBstrcat (retstrP, monthN);
					break;
				case DU_AMON:
					MBstrcat (retstrP, mnth);
					break;
				case DU_1MON:
					sprintf (&retstrP[strlen(retstrP)], "%d", month);
					break;
				case DU_2MON:		/*	zero-padded */
					sprintf (&retstrP[strlen(retstrP)], "%02d", month);
					break;
				case DU_1DAY:
					sprintf (&retstrP[strlen(retstrP)], "%d", day);
					break;
				case DU_3DAY:
					sprintf(&retstrP[strlen(retstrP)], "%03d", day3);
					break;
				case DU_2DAY:
					sprintf (&retstrP[strlen(retstrP)], "%02d", day);
					break;
				case DU_4YR:
					sprintf (&retstrP[strlen(retstrP)], "%d", year);
					break;
				case DU_2YR:
					if(fullYr)
						sprintf (&retstrP[strlen(retstrP)], "%d", year);
					else
						sprintf (&retstrP[strlen(retstrP)], "%02d", year % 100);
					break;
				case DU_1DATE2YR:
					if(fullYr){
						sprintf (&retstrP[strlen(retstrP)], 
								"%d%02d%02d", year, month, day);
					}else{
						sprintf (&retstrP[strlen(retstrP)], 
								"%02d%02d%02d", year % 100, month, day);
					}
					break;
				case DU_1DATE4YR:
					sprintf (&retstrP[strlen(retstrP)], 
								"%d%02d%02d", year, month, day);
					break;
				case DU_TH:
					MBstrcat (retstrP, get_date_ordinal(day, st, nd, rd, th));
					break;
                case DU_WST:
                    get_western_date (year, month, day, emp_date);
                    MBstrcat (retstrP, emp_date);
                    did_emperor_date = TRUE;
                    break;
				case DU_EMP:
					get_emperor_date (year, month, day, emp_date);
					MBstrcat (retstrP, emp_date);
                    			did_emperor_date = TRUE;
					break;
				case DU_SLASH:
					MBstrcat (retstrP, "/");
					break;
				case DU_DASH:
					MBstrcat (retstrP, "-");
					break;
				case DU_PERIOD:
					MBstrcat (retstrP, ".");
					break;
				case DU_COLON:
					MBstrcat (retstrP, ":");
					break;
				case DU_SPACE:
					MBstrcat (retstrP, " ");
					break;
				case DU_COMMA:
					MBstrcat (retstrP, ",");
					break;
				case DU_FWKDY:			/*	Full weekday name */
					MBstrcat (retstrP, weekday);
					break;
				case DU_AWKDY:			/*	Abbrev. weekday name */
					MBstrcat (retstrP, wkdy);
					break;
				/* Times are present for those datefmts that include times,
				 *	such as sybase or oracle
				 */
				case DU_1HR:
					sprintf (&retstrP[strlen(retstrP)], "%d", hour);
					break;
				case DU_2HR:
					sprintf (&retstrP[strlen(retstrP)], "%02d", hour);
					break;
				case DU_24HR:
					sprintf (&retstrP[strlen(retstrP)], "%02d", hour);
					break;
				case DU_MIN:
					sprintf (&retstrP[strlen(retstrP)], "%02d", min);
					break;
				case DU_SEC:
					sprintf (&retstrP[strlen(retstrP)], "%02d", sec);
					break;
				case DU_1TIME:
					sprintf (&retstrP[strlen(retstrP)], 
								"%02d%02d", hour, min);
					break;
				case DU_AM:
					if (use_ampm == DU_PM)
						MBstrcat (retstrP, pm);
					else
						MBstrcat (retstrP, am);
					break;
				case DU_PM:
					if (use_ampm == DU_AM)
						MBstrcat (retstrP, am);
					else
						MBstrcat (retstrP, pm);
					break;
				default:
					len = strlen (retstrP);
					chr = dtfmt[dfmt].tok[i] - DU_UNIDENTIF;
					if (chr > 0)
					{
						retstrP[len++] = chr;
						retstrP[len] = '\0';
					}
					break;
			}
            if (did_emperor_date)
               break;
		}
	}

	if (tfmt){
		if (preceding)
			MBstrcat (retstrP, " ");

		preceding = TRUE;

		/*	
		 *	If token list contains DU_AM or DU_PM,
		 *	set use_ampm to DU_AM or DU_PM depending on hour value and
		 *	re-set the hour if necessary for pm
		 */
		use_ampm = check_for_ampm (&tmfmt[tfmt], &hour);

		for (i = 0; i < tmfmt[tfmt].toklen; i++){
			switch (tmfmt[tfmt].tok[i]){
				case DU_1HR:
					sprintf (&retstrP[strlen(retstrP)], "%d", hour);
					break;
				case DU_2HR:
					sprintf (&retstrP[strlen(retstrP)], "%02d", hour);
					break;
				case DU_24HR:
					sprintf (&retstrP[strlen(retstrP)], "%02d", hour);
					break;
				case DU_MIN:
					sprintf (&retstrP[strlen(retstrP)], "%02d", min);
					break;
				case DU_SEC:
					sprintf (&retstrP[strlen(retstrP)], "%02d", sec);
					break;
				case DU_1TIME:
					sprintf (&retstrP[strlen(retstrP)], "%02d%02d", hour, min);
					break;
				case DU_AM:
					if (use_ampm == DU_PM)
						MBstrcat (retstrP, pm);
					else
						MBstrcat (retstrP, am);
					break;
				case DU_PM:
					if (use_ampm == DU_AM)
						MBstrcat (retstrP, am);
					else
						MBstrcat (retstrP, pm);
					break;
				case DU_TH:
					MBstrcat (retstrP, get_date_ordinal(day, st, nd, rd, th));
					break;
				case DU_SLASH:
					MBstrcat (retstrP, "/");
					break;
				case DU_DASH:
					MBstrcat (retstrP, "-");
					break;
				case DU_PERIOD:
					MBstrcat (retstrP, ".");
					break;
				case DU_COLON:
					MBstrcat (retstrP, ":");
					break;
				case DU_SPACE:
					MBstrcat (retstrP, " ");
					break;
				case DU_COMMA:
					MBstrcat (retstrP, ",");
					break;
				default:
					len = strlen (retstrP);
					chr = tmfmt[tfmt].tok[i] - DU_UNIDENTIF;
					if (chr > 0){
						retstrP[len++] = chr;
						retstrP[len] = '\0';
					}
					break;
			}
		}
	}

	if (wfmt && (wfmt % 2 == 0)){
		if (preceding)
			MBstrcat (retstrP, " ");

		switch (wfmt * DTWFBASE){
			case DTWFFEND:
				MBstrcat (retstrP, weekday);
				break;
			case DTWFEND:
				MBstrcat (retstrP, wkdy);
				break;
		}
	}
	MBstrcat (retstrP, "");		/*	finish with a null */
	
#if BUILD_KANJI
    if  (did_emperor_date )
        {
        /* need to convert all the digits to wide digits */
        wchar wtemp[100];
        MbToWStr( retstrP, wtemp );
        MakeConstantCharSet( wtemp );
        WToMbStr( wtemp, retstrP );
        }
#endif
	return (AOK);
}


/*	Has been turned into a wrapper for Formadt_str_lang; see
 *	edit #25 for reason */

typeErr
Formatdt_str (dtvalue, datefmt_str, retstrP)
unsigned long	dtvalue;	   		/*	incoming Applix-fmt datetime */
mbchar		*datefmt_str,
		*retstrP;			 
{
	return( Formatdt_str_lang (dtvalue, datefmt_str, retstrP, id_lang()));
}

typeErr
Formatdt_str_unix (dtvalue, datefmt_str, retstrP)
unsigned long	dtvalue;	   		/*	incoming unix datetime value */
mbchar			*datefmt_str,
				*retstrP;			 
{
	return(Formatdt_str_lang_val(dtvalue,datefmt_str,retstrP,id_lang(),TRUE));
}

/* yet another layer of wrapper added may 1 1996 to enable formatting
of unix datetime values, which are precise to the second, unlike
applix datetime values */

typeErr
Formatdt_str_lang (dtvalue, datefmt_str, retstrP, lang)
unsigned long	dtvalue;	   		/*	incoming Applix-fmt datetime */
mbchar		*datefmt_str,
		*retstrP;			 
int		lang;
{
	return( Formatdt_str_lang_val(dtvalue, datefmt_str, retstrP, lang, FALSE));
}

typeErr
Formatdt_str_lang_val (dtvalue, datefmt_str, retstrP, lang, unixval)
unsigned long		dtvalue;	/* incoming datetime value */
mbchar			*datefmt_str,
			*retstrP;			 
int			lang;
int			unixval; 	/* TRUE if dtvalue is a UNIX datetime value */
{
	AxTimeStruct axt;
	int	day = 0, month = 0, year = 0,
		hour = 0, min = 0, sec = 0;

	/*	what to do with no date value to parse */
	if( dtvalue == DTNULL){
		MBstrcpy (retstrP, "");
		return (!AOK);
	}

	/* convert dtvalue to AxTimeStruct */
	if (unixval)
		decodeUnixDates (dtvalue, &day, &month, &year, &hour, &min, &sec);
	else
		decodeApplixDates (dtvalue, &day, &month, &year, &hour, &min, &sec);

	axt.day = day;
	axt.month = month;
	axt.year = year;
	axt.hour = hour;
	axt.minute = min;
	axt.second = sec;
	axt.fraction = 0;
	
	return(Formatdt_str_AxTime_lang(axt, datefmt_str, retstrP, lang));
}

typeErr
Formatdt_str_AxTime (axt, datefmt_str, retstrP)
AxTimeStruct	axt;	   		/*	incoming datetime value */
mbchar			*datefmt_str,
				*retstrP;			 
{
	return(Formatdt_str_AxTime_lang(axt, datefmt_str, retstrP,id_lang()));
}

typeErr
Formatdt_str_AxTime_lang (axt, datefmt_str, retstrP, lang)
AxTimeStruct	axt;	   		/*	incoming datetime value */
mbchar		*datefmt_str,
		*retstrP;			 
int		lang;
{
	mbchar		wkdy[STRPATH],		/*	the abb'd weekday name */
			weekday[STRPATH],	/*	the full weekday name */
			mnth[STRPATH],		/*	the abb'd month name */
			monthN[STRPATH],	/*	the full month name */
			th[STRPATH],		/*	whatever serves for 'th' */
			st[STRPATH],		/*	whatever serves for '1st' */
			nd[STRPATH],		/*	whatever serves for '2nd' */
			rd[STRPATH],		/*	whatever serves for '3rd' */
			am[STRPATH],		/*	whatever serves for 'am' */
			pm[STRPATH],		/*	whatever serves for 'pm' */
			emp_date[STRPATH],	/*	support Japanese format */
			*ptr,				/*	skim thru the fmt string */
			rstr[240];
	int		day3 = 0, day = 0, month = 0, year = 0,
			hour = 0, min = 0, sec = 0,
			use_ampm = 0;
	static int	loaded = FALSE;			/*	are the datetime fmts loaded? */
	int 		did_emperor_date = FALSE;


	if ( ! loaded && ! fmts_loaded()){
		if (load_datetime_fmts() != AOK)
			return (!AOK);
		else
			loaded = TRUE;
	}

	/*	start with a clean slate */
    	*retstrP = (mbchar)0;
	memset (rstr, 0, sizeof(rstr));

	/* what time is it? */
	day = axt.day;
	month = axt.month;
	year = axt.year;
	hour = axt.hour;
	min = axt.minute;
	sec = axt.second + ((int)(axt.fraction + .5)); /* round to second */
	day3 = ax_get_3digit_day(day, month, year);

	if ((set_names(lang, wkdy, weekday, mnth, monthN, month, day, year,
			am, pm, th, st, nd, rd)) != AOK)
		return (!AOK);

	use_ampm = check_string_for_ampm (datefmt_str, &hour);

	ptr = datefmt_str;

	while (*ptr != '\0' && *ptr != '\n'){
		switch (*ptr){
			case 'M':	/*	Mmmm  or  Mm */
						if (MBstrncmp (ptr, FMON_NAME, FMON_LEN) == 0
							|| (MBstrncmp (ptr, AMON_NAME, AMON_LEN) == 0))
						{
							if (MBstrncmp (ptr, FMON_NAME, FMON_LEN) == 0){
								ptr += FMON_LEN;
								MBstrcat (rstr, monthN);
							}
							else if (MBstrncmp (ptr, AMON_NAME, AMON_LEN) == 0){
								ptr += AMON_LEN;
								MBstrcat (rstr, mnth);
							}
						}
						else{
							rstr[MBstrlen(rstr)] = *ptr;
							ptr++;
						}

						break;

			case 'm':	/*	mm (months) or mi (minutes) or m (1-2digit month)*/
						if (MBstrncmp (ptr, MON2_NAME, MON2_LEN) == 0){
							ptr += MON2_LEN;
							sprintf (&rstr[strlen(rstr)], "%02d", month);
						}
						else if (MBstrncmp (ptr, MIN_NAME, MIN_LEN) == 0){
							ptr += MIN_LEN;
							sprintf (&rstr[strlen(rstr)], "%02d", min);
						}
						else if (MBstrncmp (ptr, MON1_NAME, MON1_LEN) == 0){
							ptr += MON1_LEN;
							sprintf (&rstr[strlen(rstr)], "%d", month);
						}
						else{
							rstr[MBstrlen(rstr)] = *ptr;
							ptr++;
						}
						break;
			case 'd':	
						/* ddd */
						if (MBstrncmp(ptr, DAY3_NAME, DAY3_LEN) == 0){
							/* ddd */
							ptr += DAY3_LEN;
							sprintf (&rstr[strlen(rstr)], "%03d", day3);
						}
						/* dd */
						if (MBstrncmp (ptr, DAY2_NAME, DAY2_LEN) == 0){
							ptr += DAY2_LEN;
							sprintf (&rstr[strlen(rstr)], "%02d", day);
						}
						else if (MBstrncmp (ptr, DAY1_NAME, DAY1_LEN) == 0){
							/*	d	*/
							ptr += DAY1_LEN;
							sprintf (&rstr[strlen(rstr)], "%d", day);
						}
						else{
							rstr[MBstrlen(rstr)] = *ptr;
							ptr++;
						}
						break;

			case 'y':	/*	no good way of #def-ing this */
						if (MBstrncmp (ptr, NXLT("yyyymmdd"), strlen (NXLT("yyyymmdd"))) == 0){
							ptr += strlen (NXLT("yyyymmdd"));
							sprintf (&rstr[strlen(rstr)],
										"%d%02d%02d", year, month, day);
						}
						else if (MBstrncmp (ptr, NXLT("yymmdd"), strlen (NXLT("yymmdd"))) == 0){
							ptr += strlen (NXLT("yymmdd"));
							sprintf (&rstr[strlen(rstr)], 
								"%02d%02d%02d", year % 100, month, day);
						}
						/*	yyyy   or  yy */
						else if (MBstrncmp (ptr, YR4_NAME, YR4_LEN) == 0){
							ptr += YR4_LEN;
							sprintf (&rstr[strlen(rstr)], "%d", year);
						}
						else if (MBstrncmp (ptr, YR2_NAME, YR2_LEN) == 0){
							ptr += YR2_LEN;
							sprintf (&rstr[strlen(rstr)], "%02d", year % 100);
						}
						else{
							rstr[MBstrlen(rstr)] = *ptr;
							ptr++;
						}
						break;

			case 'W':	/*	Wwww  or  Ww */
						if (MBstrncmp (ptr, FWKDY_NAME, FWKDY_LEN) == 0){
							ptr += FWKDY_LEN;
							MBstrcat (rstr, weekday);
						}
						else if (MBstrncmp (ptr, AWKDY_NAME, AWKDY_LEN) == 0){
							ptr += AWKDY_LEN;
							MBstrcat (rstr, wkdy);
						}
						else{
							rstr[MBstrlen(rstr)] = *ptr;
							ptr++;
						}
						break;

			case 'h':	/*	no good way of #def-ing this */
						if (MBstrncmp (ptr, NXLT("hhmi"), strlen (NXLT("hhmi"))) == 0){
							ptr += strlen (NXLT("hhmi"));
							sprintf (&rstr[strlen(rstr)],
										"%02d%02d", hour, min);
						}
						/*	hh */
						else if (MBstrncmp (ptr, HR2_NAME, HR2_LEN) == 0){
							ptr += HR2_LEN;
							sprintf (&rstr[strlen(rstr)], "%02d", hour);
						}
						else{
							rstr[MBstrlen(rstr)] = *ptr;
							ptr++;
						}
						break;

			case 'H':	/*	HH */
						if (MBstrncmp (ptr, HR24_NAME, HR24_LEN) == 0){
							ptr += HR24_LEN;
							sprintf (&rstr[strlen(rstr)], "%02d", hour);
						}
						else{
							rstr[MBstrlen(rstr)] = *ptr;
							ptr++;
						}
						break;

			case 's':	/*	ss */
						if (MBstrncmp (ptr, SEC_NAME, SEC_LEN) == 0){
							ptr += SEC_LEN;
							sprintf (&rstr[strlen(rstr)], "%02d", sec);
						}
						else{
							if (*ptr == '\\')
								ptr++;
							rstr[MBstrlen(rstr)] = *ptr;
							ptr++;
						}
						break;

			case 'p':	/*	 pm */
						if (MBstrncmp (ptr, PM_NAME, PM_LEN) == 0){
							ptr += PM_LEN;
							if (use_ampm == DU_AM)
								MBstrcat (rstr, am);
							else if (use_ampm == DU_PM)
								MBstrcat (rstr, pm);
						}
						else{
							rstr[MBstrlen(rstr)] = *ptr;
							ptr++;
						}
						break;

			case 't':	/*	th */
						if (MBstrncmp (ptr, TH_NAME, TH_LEN) == 0){
							ptr += TH_LEN;
							MBstrcat (rstr, get_date_ordinal(day, st, nd, rd,
							th));
						}
						else{
							rstr[MBstrlen(rstr)] = *ptr;
							ptr++;
						}
						break;

            case 'C':   /*  CCCC - Western date; Japanese date support */
                        if (MBstrncmp (ptr, WST_NAME, WST_LEN) == 0){
                            ptr += WST_LEN;
                            get_western_date (year, month, day, emp_date);
                            MBstrcat (rstr, emp_date);
                            did_emperor_date = TRUE;
                        }
                        else{
                            rstr[MBstrlen(rstr)] = *ptr;
                            ptr++;
                        }
                        break;
 

			case 'E':	/*	EE - Emperor date; Japanese date support */
						if (MBstrncmp (ptr, EMP_NAME, EMP_LEN) == 0){
							ptr += EMP_LEN;
							get_emperor_date (year, month, day, emp_date);
							MBstrcat (rstr, emp_date);
							did_emperor_date = TRUE;
						}
						else{
							rstr[MBstrlen(rstr)] = *ptr;
							ptr++;
						}
						break;

			default:	/*	Stuff anything else onto the fmt string verbatim, */
						/*	one unrecognized char at a time.  A char preceded */
						/*	by a backslash is also stuffed onto the string.  */
						if (*ptr == '\\')
							ptr++;
						rstr[MBstrlen(rstr)] = *ptr;
						ptr++;
						break;
		}
	}

	if (*ptr == '\n')		/*	annihilate an ending return */
		*ptr = '\0';


	MBstrcat (rstr, "");		/*	finish with a null */

#if BUILD_KANJI
    if  (did_emperor_date ){
        /* need to convert all the digits to wide digits */
        wchar wtemp[100];
        MbToWStr( rstr, wtemp );
        MakeConstantCharSet( wtemp );
        WToMbStr( wtemp, rstr );
        }
#endif
	strcpy (retstrP, rstr);
	return (AOK);
}

void
decodeUnixDates (dtval, day, month, year, hour, min, sec)
unsigned long	dtval;
int				*day, *month, *year, *hour, *min, *sec;
{
	struct tm *t, *localtime();

	t = localtime(&dtval);
	*day = t->tm_mday;
	*month = t->tm_mon+1;
	*year = t->tm_year+1900;
	*hour = t->tm_hour;
	*min = t->tm_min;
	*sec = t->tm_sec;
	return;
}

void
decodeApplixDates (dtval, day, month, year, hour, min, sec)
unsigned long	dtval;
int				*day, *month, *year, *hour, *min, *sec;
{
	unsigned long	lyr, lmo, ldy, lhr, lmin, lsec, lc, lcyr, lcmo, lcdy, lchr;

	if (dtval == DTNULL)
	{
		return;
	}

	lc   = 60/DTTIMEGRANULARITY;
	lcyr = lc*(unsigned long)599040;			/*	   60*24*32*13	 	 */
	lcmo = lc*46080;					/*	   60*24*32		 */
	lcdy = lc*1440;						/*	   60*24		 */
	lchr = lc*60;

	lyr		 = dtval/lcyr;
	dtval -= lyr*lcyr;

	lmo		 = dtval/lcmo;
	dtval -= lmo*lcmo;

	ldy		 = dtval/lcdy;
	dtval -= ldy*lcdy;

	lhr		 = dtval/lchr;
	dtval -= lhr*lchr;

	lmin             = dtval/lc;
	dtval -= lmin*lc;

	lsec   = dtval*DTTIMEGRANULARITY;

	*year = lyr + DTTIMEBASEYEAR;
	*month = lmo;
	*day = ldy;
	*hour = lhr;
	*min= lmin;
	*sec = lsec;
}


typeErr
set_names (lang, wkdy, weekday, mnth, month, month_nb, day, year, am, pm, th,
			st, nd, rd)
int		lang;					/*	Proximity language number */
mbchar	*wkdy, *weekday, *mnth, *month, *am, *pm, *th, *st, *nd, *rd;
int		day, month_nb;
unsigned long	year;
{
	static mbchar
			wkdys[STRPATH],		/*	abb'd weekday names */
			weekdays[STRPATH],	/*	full weekday names */
			mnths[STRPATH],		/*	abb'd month names */
			months[STRPATH];		/*	full month names */
	int		daycnt = 0,
			daylen = 0,
			ix = 0,
			day_abbrevs = 0,
			day_names = 0,
			month_abbrevs = 0,
			month_names = 0;
	unsigned long	lyear = year - DTTIMEBASEYEAR;
	static int		init = 1;

	if (init != lang)
		init = lang;

	/*	if lang outside 1-17 range (Eng-Jpn), defaults to 1 (Eng) */
	set_string_nums (lang, &day_abbrevs, &day_names,
							&month_abbrevs, &month_names);

	if (getStr (day_abbrevs, wkdys) != AOK)
		return (!AOK);

	if (getStr (day_names, weekdays) != AOK)
		return (!AOK);

	if (getStr (month_abbrevs, months) != AOK)
		return (!AOK);

	if (getStr (month_names, mnths) != AOK)		/*	also has am,pm,th */
		return (!AOK);

	/*	find weekday names */
	daycnt = lyear + (lyear + 1)/4 + 1;		/*	+1 since we base this on 1848
											 *	which was a leap year.*/
	if (year > 1900) 
		daycnt--;
	if (year % 4 == 0 && year % 100 != 0 || year % 400 == 0)
		DysPermo[1] = 29;
	else
		DysPermo[1] = 28;
	daylen = 0;
	for (ix = 1; ix < month_nb ; ix++)  {
		daylen += DysPermo[ix-1];
	}
	daylen += day;
	daycnt = daycnt + ((daylen - 1) % 7);

	ListExtract ((daycnt % 7) + 'a', wkdys, 0, wkdy);	/*	0 == case as is */
	ListExtract ((daycnt % 7) + 'a', weekdays, 0, weekday);

	/*	find month names, am, pm & th abbreviations */
	ListExtract( 'a', mnths, -1, am);	/*	-1 == lower case */
	ListExtract( 'b', mnths, -1, pm);
	ListExtract( 'c', mnths, -1, th);
	ListExtract( 'p', mnths, -1, st);
	ListExtract( 'q', mnths, -1, nd);
	ListExtract( 'r', mnths, -1, rd);
	ListExtract( month_nb +'c', mnths, 0, mnth);	/*	0 == case as is */
	ListExtract( month_nb +'c', months, 0, month);

	return (AOK);
}


void
set_string_nums (lang, day_abbrevs, day_names, month_abbrevs, month_names)
int	lang, *day_abbrevs, *day_names, *month_abbrevs, *month_names;
{
	switch (lang) {
		default:
		case 1:
			*day_abbrevs = SDTDYS;
			*day_names = SDTDAYS;
			*month_abbrevs = SDTMOS;
			*month_names = SDTVALS;
			break;
		case 2:
			*day_abbrevs = GRMDYS;
			*day_names = GRMDAYS;
			*month_abbrevs = GRMMOS;
			*month_names = GRMVALS;
			break;
		case 3:
			*day_abbrevs = FRNDYS;
			*day_names = FRNDAYS;
			*month_abbrevs = FRNMOS;
			*month_names = FRNVALS;
			break;
		case 4:
			*day_abbrevs = SPNDYS;
			*day_names = SPNDAYS;
			*month_abbrevs = SPNMOS;
			*month_names = SPNVALS;
			break;
		case 5:
			*day_abbrevs = ITLDYS;
			*day_names = ITLDAYS;
			*month_abbrevs = ITLMOS;
			*month_names = ITLVALS;
			break;
		case 6:
			*day_abbrevs = BRTDYS;
			*day_names = BRTDAYS;
			*month_abbrevs = BRTMOS;
			*month_names = BRTVALS;
			break;
		case 7:
			*day_abbrevs = SWDDYS;
			*day_names = SWDDAYS;
			*month_abbrevs = SWDMOS;
			*month_names = SWDVALS;
			break;
		case 8:
			*day_abbrevs = DANDYS;
			*day_names = DANDAYS;
			*month_abbrevs = DANMOS;
			*month_names = DANVALS;
			break;
		case 9:
			*day_abbrevs = NRWDYS;
			*day_names = NRWDAYS;
			*month_abbrevs = NRWMOS;
			*month_names = NRWVALS;
			break;
		case 10:
			*day_abbrevs = DUTDYS;
			*day_names = DUTDAYS;
			*month_abbrevs = DUTMOS;
			*month_names = DUTVALS;
			break;
		case 11:
			*day_abbrevs = PRTDYS;
			*day_names = PRTDAYS;
			*month_abbrevs = PRTMOS;
			*month_names = PRTVALS;
			break;
		case 12:
			*day_abbrevs = BRZDYS;
			*day_names = BRZDAYS;
			*month_abbrevs = BRZMOS;
			*month_names = BRZVALS;
			break;
		case 13:
			*day_abbrevs = CFRDYS;
			*day_names = CFRDAYS;
			*month_abbrevs = CFRMOS;
			*month_names = CFRVALS;
			break;
		case 14:
			*day_abbrevs = SGRDYS;
			*day_names = SGRDAYS;
			*month_abbrevs = SGRMOS;
			*month_names = SGRVALS;
			break;
		case 15:
			*day_abbrevs = NYNDYS;
			*day_names = NYNDAYS;
			*month_abbrevs = NYNMOS;
			*month_names = NYNVALS;
			break;
		case 16:
			*day_abbrevs = FINDYS;
			*day_names = FINDAYS;
			*month_abbrevs = FINMOS;
			*month_names = FINVALS;
			break;
		case 17:
			*day_abbrevs = JPNDYS;
			*day_names = JPNDAYS;
			*month_abbrevs = JPNMOS;
			*month_names = JPNVALS;
			break;
	}
}


/*	see if the format requires am/pm notation and adjust the hour accordingly */

int
check_for_ampm (tfmt, hour)
dttmfmt	*tfmt;
int		*hour;
{
	int		i;

	for (i = 0; i < tfmt->toklen; i++)
	{
		if (tfmt->tok[i] == DU_AM || tfmt->tok[i] == DU_PM)
		{
			if (*hour >= 12)			/*	12:30 is past noon, pm */
			{
				if (*hour > 12)			/*	but only decrement hour if > 12 */
					*hour = *hour - 12;
				return (DU_PM);
			}
			else
			{
				if (*hour < 1)				/*	e.g.:  00:045			 */
					*hour = *hour + 12;		/*	in 12-hr fmt is 12:45 am */
				return (DU_AM);
			}
		}
	}

	return (FALSE);
}

/*	see if the format requires am/pm notation and adjust the hour accordingly */

int
check_string_for_ampm (datefmt_str, hour)
mbchar	*datefmt_str;
int		*hour;
{
	char	*loc = NULL;

	loc = strpbrk (datefmt_str, PM_NAME);

	if (loc == NULL)
		return (FALSE);
	else{
		while ( (*(loc - 1) == '\\') || (strncmp (loc, PM_NAME, PM_LEN) != 0) ){
			loc++;
			loc = strpbrk (loc, PM_NAME);
			if (loc == NULL)
				return (FALSE);
			else
				continue;
		}

	}

	/*	If it got through that last loop, it found an 'am/pm' token;
	 *	so reset the hour if it's over 12
	 */

	if (*hour >= 12){			/*	12:30 is past noon, pm */
		if (*hour > 12)			/*	but only decrement hour if > 12 */
			*hour = *hour - 12;
		return (DU_PM);
	}
	else{
		if (*hour < 1)				/*	e.g.:  00:045			 */
			*hour = *hour + 12;		/*	in 12-hr fmt is 12:45 am */
		return (DU_AM);
	}
}


/* If we use format string which includes some non-ascii chars in sprintf(),
 * the results sometimes would be unpredicable(st least we found one case
 * in HP800V3). So we assume separated multi-byte strings here. */
mbchar showa[] = {(mbchar)0xbe, (mbchar)0xbc, (mbchar)0xcf, (mbchar)0xc2, (mbchar)0x00};
mbchar heisei[] = {(mbchar)0xca, (mbchar)0xbf, (mbchar)0xc0, (mbchar)0xae, (mbchar)0x00};
mbchar yrJi[] = {(mbchar)0xc7, (mbchar)0xaf, (mbchar)0x00};
mbchar monJi[] = {(mbchar)0xb7, (mbchar)0xee, (mbchar)0x00};
mbchar dyJi[] = {(mbchar)0xc6, (mbchar)0xfc, (mbchar)0x00};

typeErr
get_western_date (yr, mon, dy, wst_dt)            /*  !!  STUB */
int       yr, mon, dy;
mbchar    *wst_dt;
{
  sprintf((char *)wst_dt, "%d%s%d%s%d%s", yr, yrJi, mon, monJi, dy, dyJi);
  return (AOK);
}


typeErr
get_emperor_date (yr, mon, dy, emp_dt)			/* !!  STUB */
int		yr, mon, dy;
mbchar	*emp_dt;
{
    int num;

    if ( (yr >= 1926) && (yr < 1989) )
        {
        /* a showa date */
        num = yr - 1925;
        sprintf( (char *)emp_dt, "%s%d%s%d%s%d%s", showa, num, yrJi, mon, monJi, dy, dyJi );
        }
    else if ( yr >= 1989 )
        {
        /* a heisei date */
        num = yr - 1988;
        sprintf( (char *)emp_dt, "%s%d%s%d%s%d%s", heisei, num, yrJi, mon, monJi, dy, dyJi );
        }
    else
        {
        /* before the emperors we know */
        MBstrcpy( (char *)emp_dt, "???? " );
        }

	return (AOK);
}



/*---------------- Routines to load datetime fmts from file -----------------*/



static char	loaded = FALSE;	/*	so that fmts_loaded() can be used to check it */



typeErr
fmts_loaded()
{
	return (loaded);
}



/* 
 * GE will load datetime formats and store the values in its data structure
 * but load_datetime_fmts() stores the values in a static structs. Thus, a
 * variation of load_datetime_fmts() is written to access the values.
 */
typeErr
load_datetime_formats(dfmt, tfmt, n_d, n_t )
dttmfmt dfmt[];
dttmfmt tfmt[];
int *n_d;
int *n_t;
{
        int i;
        if (load_datetime_fmts() != AOK)
          return (!AOK);
	*n_d = MaxDtFmt;
	*n_t = MaxTmFmt;

        
        for ( i=0; i<100; i++ ) 
        {
             MBstrcpy(dfmt[i].fmt, dtfmt[i].fmt);
             dfmt[i].toklen = dtfmt[i].toklen;
             memcpy(dfmt[i].tok, dtfmt[i].tok, 120); 
        }
        for ( i=0; i<50; i++ ) 
        {
             MBstrcpy(tfmt[i].fmt, tmfmt[i].fmt);
             tfmt[i].toklen = tmfmt[i].toklen;
             memcpy(tfmt[i].tok, tmfmt[i].tok, 120); 
        }
        return(AOK);
}


typeErr
load_datetime_fmts()
{
	/*	open the datetime.sp file and read it in, parsing the formats
	 *	into the global dtfmt and tmfmt arrays as we go */

	mbchar	execdir[STRPATH],
			buf[1024];
	int		i = 0;
	FILE	*dtfp = NULL,
			*ofp = NULL;


	if (loaded == TRUE)
		return (AOK);

	/*	use DirSlash to coddle DOS slash requirements
		sprintf (execdir, "%s%s%s",AxLangDir,DirSlash,NXLT("datetime.sp"));
	*/

	/*	Look in all the standard places:
		axlocal, axlocal/elf/eng, axdata, axdata/eng
	 */
	if (! FindThimbleFile (THIM_FIND_CUSTOMPLUS, XLT2("Name of datetime file, ending needs to be 3-letter lang suffix", "datetime.sp"), execdir, &i))
		return (!AOK);

	/*	read-only, so no DOS footwork necessary */
	if ( (dtfp = fopen (execdir, "r")) == NULL)
		return (!AOK);

#if 0
	if ( (ofp = fopen ("/ax/lpgms/datetime.sp.out", "w")) == NULL)
		return (!AOK);
#endif

	memset (dtfmt, 0, sizeof(dtfmt));
	memset (tmfmt, 0, sizeof(tmfmt));

	/*	date and time formats start at 1, so that the 0'th element
	 *	can be used to signify there is no time or date format, as
	 *	per datetime.h */
	MaxDtFmt = 1;
	MaxTmFmt = 1;

	while (fgets (buf, sizeof (buf), dtfp) != NULL)
	{
		if (buf[0] != '#')
			break;
	}

#if 0
	fprintf (ofp, "%s\n\n", "dates");
#endif

	do
	{
		if (buf[0] == '#')
			break;

		if (MaxDtFmt < 100)		/*	Max possible fmt */
		{
			/*	read in and parse the date fmts */
			if (get_datetime_fmt (buf, &dtfmt[MaxDtFmt]) == AOK)
			{
#if 0
				fprintf (ofp, "%s\t = \t", buf);
				for (i = 0; i < dtfmt[MaxDtFmt].toklen; i++)
					fprintf (ofp, "%d ", dtfmt[MaxDtFmt].tok[i]);
				fprintf (ofp, "\n");
#endif

				MBstrcpy (dtfmt[MaxDtFmt].fmt, buf);
				MaxDtFmt++;
			}
			else
				return (!AOK);
		}
	} while ( (fgets (buf, sizeof (buf), dtfp) != NULL) &&  (buf[0] != '#'));

#if 0
	fprintf (ofp, "\n\n%s\n\n", "times");
#endif

	while (fgets (buf, sizeof (buf), dtfp) != NULL  && buf[0] != '#')
	{
		/*	read in and parse the time fmts */
		if (get_datetime_fmt (buf, &tmfmt[MaxTmFmt]) == AOK)
		{
#if 0
			fprintf (ofp, "%s\t = \t", buf);
			for (i = 0; i < tmfmt[MaxTmFmt].toklen; i++)
				fprintf (ofp, "%d ", tmfmt[MaxTmFmt].tok[i]);
			fprintf (ofp, "\n");
#endif

			MBstrcpy (tmfmt[MaxTmFmt].fmt, buf);
			MaxTmFmt++;
		}
		else
			return (!AOK);
	}

	loaded = TRUE;
	return (AOK);
}


typeErr
get_datetime_fmt (buf, fmt)
mbchar	*buf;
dttmfmt	*fmt;
{
	/*	Getting the format consists of parsing the 'Mmmm dd, yyyy' constructs
	 *	to build up a regular expression string.
	 */

	mbchar	*ptr = buf;					/*	read thru buf */
	int		indx = 0,					/*	counter for tok array */
			i = 0;						/*	 counter */


	fmt->sep = DU_SPACE;		/*	default separator for fmt */


	/*	See if there's a comment at the end.  If so, back up to the end of
	 *	the format and step on it there. */

	while (*ptr != '\0' && *ptr !=  '#' && *ptr != '\n')
		ptr ++;

	if (*ptr == '#')
	{
		--ptr;
		while (*ptr == '\t' || *ptr == ' ')
			--ptr;
		++ptr;
		*ptr = '\0';
	}

	ptr = buf;


	/*	Now we've isolated the format, comment or no comment */

	while (*ptr != '\0' && *ptr != '\n')
	{
		switch (*ptr)
		{
			case 'M':	/*	Mmmm  or  Mm */
						if (MBstrncmp (ptr, FMON_NAME, FMON_LEN) == 0
							|| (MBstrncmp (ptr, AMON_NAME, AMON_LEN) == 0))
						{
							if (MBstrncmp (ptr, FMON_NAME, FMON_LEN) == 0)
							{
								ptr += FMON_LEN;
								fmt->tok[indx++] = DU_FMON;
							}
							else if (MBstrncmp (ptr, AMON_NAME, AMON_LEN) == 0)
							{
								ptr += AMON_LEN;
								fmt->tok[indx++] = DU_AMON;
							}
						}
						else
							fmt->tok[indx++] = DU_UNIDENTIF + (*ptr++);
						break;

			case 'm':	/*	mm (months)  or  mi (minutes) or m (1-2digit month*/
						if (MBstrncmp (ptr, MON2_NAME, MON2_LEN) == 0)
						{
							ptr += MON2_LEN;
							fmt->tok[indx++] = DU_2MON;
						}
						else if (MBstrncmp (ptr, MIN_NAME, MIN_LEN) == 0)
						{
							ptr += MIN_LEN;
							fmt->tok[indx++] = DU_MIN;
						}
						else if (MBstrncmp (ptr, MON1_NAME, MON1_LEN) == 0)
						{
							ptr += MON1_LEN;
							fmt->tok[indx++] = DU_1MON;
						}
						else
							fmt->tok[indx++] = DU_UNIDENTIF + (*ptr++);
						break;

			case 'd':
						if (MBstrncmp(ptr, DAY3_NAME, DAY3_LEN) == 0)
						{
							/* ddd */
							ptr += DAY3_LEN;
							fmt->tok[indx++] = DU_3DAY;
						}
						else if (MBstrncmp (ptr, DAY2_NAME, DAY2_LEN) == 0)
						{	/*	dd */
							ptr += DAY2_LEN;
							fmt->tok[indx++] = DU_2DAY;
						}
						else if (MBstrncmp (ptr, DAY1_NAME, DAY1_LEN) == 0)
						{
							/*	d	*/
							ptr += DAY1_LEN;
							fmt->tok[indx++] = DU_1DAY;
						}
						else
							fmt->tok[indx++] = DU_UNIDENTIF + (*ptr++);
						break;

			case 'y':	/*	no good way of #def-ing this */
						if (MBstrncmp (ptr, NXLT("yyyymmdd"), strlen (NXLT("yyyymmdd"))) == 0)
						{
							ptr += strlen (NXLT("yyyymmdd"));
							fmt->tok[indx++] = DU_1DATE4YR;
						}
						else if (MBstrncmp (ptr, NXLT("yymmdd"), strlen (NXLT("yymmdd"))) == 0)
						{
							ptr += strlen (NXLT("yymmdd"));
							fmt->tok[indx++] = DU_1DATE2YR;
						}
						/*	yyyy   or  yy */
						else if (MBstrncmp (ptr, YR4_NAME, YR4_LEN) == 0)
						{
							ptr += YR4_LEN;
							fmt->tok[indx++] = DU_4YR;
						}
						else if (MBstrncmp (ptr, YR2_NAME, YR2_LEN) == 0)
						{
							ptr += YR2_LEN;
							fmt->tok[indx++] = DU_2YR;
						}
						else
							fmt->tok[indx++] = DU_UNIDENTIF + (*ptr++);
						break;

			case 'W':	/*	Wwww  or  Ww */
						if (MBstrncmp (ptr, FWKDY_NAME, FWKDY_LEN) == 0)
						{
							ptr += FWKDY_LEN;
							fmt->tok[indx++] = DU_FWKDY;
						}
						else if (MBstrncmp (ptr, AWKDY_NAME, AWKDY_LEN) == 0)
						{
							ptr += AWKDY_LEN;
							fmt->tok[indx++] = DU_AWKDY;
						}
						else
							fmt->tok[indx++] = DU_UNIDENTIF + (*ptr++);
						break;

			case 'H':	/*	HH */
						if (MBstrncmp (ptr, NXLT("HHmi"), strlen (NXLT("HHmi"))) == 0)
						{
							ptr += strlen (NXLT("HHmi"));
							fmt->tok[indx++] = DU_1TIME; /* treat like "hhmi" */
						}
						else if (MBstrncmp (ptr, HR24_NAME, HR24_LEN) == 0)
						{
							ptr += HR24_LEN;
							fmt->tok[indx++] = DU_24HR;
						}
						else
							fmt->tok[indx++] = DU_UNIDENTIF + (*ptr++);
						break;

			case 'h':	/*	no good way of #def-ing this */
						if (MBstrncmp (ptr, NXLT("hhmi"), strlen (NXLT("hhmi"))) == 0)
						{
							ptr += strlen (NXLT("hhmi"));
							fmt->tok[indx++] = DU_1TIME;
						}
						/*	hh */
						else if (MBstrncmp (ptr, HR2_NAME, HR2_LEN) == 0)
						{
							ptr += HR2_LEN;
							fmt->tok[indx++] = DU_2HR;
						}
						else
							fmt->tok[indx++] = DU_UNIDENTIF + (*ptr++);
						break;

			case 's':	/*	ss */
						if (MBstrncmp (ptr, SEC_NAME, SEC_LEN) == 0)
						{
							ptr += SEC_LEN;
							fmt->tok[indx++] = DU_SEC;
						}
						else
							fmt->tok[indx++] = DU_UNIDENTIF + (*ptr++);
						break;

			case 'p':	/*	 pm */
						if (MBstrncmp (ptr, PM_NAME, PM_LEN) == 0)
						{
							ptr += PM_LEN;
							fmt->tok[indx++] = DU_PM;
						}
						else
							fmt->tok[indx++] = DU_UNIDENTIF + (*ptr++);
						break;

			case 't':	/*	th */
						if (MBstrncmp (ptr, TH_NAME, TH_LEN) == 0)
						{
							ptr += TH_LEN;
							fmt->tok[indx++] = DU_TH;
						}
						else
							fmt->tok[indx++] = DU_UNIDENTIF + (*ptr++);
						break;

            case 'C':   /*  CCCC - Western date; Japanese date support */
                        if (MBstrncmp (ptr, WST_NAME, WST_LEN) == 0)
                        {
                            ptr += WST_LEN;
                            fmt->tok[indx++] = DU_WST;
                        }
                        else
                            fmt->tok[indx++] = DU_UNIDENTIF + (*ptr++);
                        break;

			case 'E':	/*	EE - Emperor date; Japanese date support */
						if (MBstrncmp (ptr, EMP_NAME, EMP_LEN) == 0)
						{
							ptr += EMP_LEN;
							fmt->tok[indx++] = DU_EMP;
						}
						else
							fmt->tok[indx++] = DU_UNIDENTIF + (*ptr++);
						break;

			case '/':	ptr++;
						fmt->sep = fmt->tok[indx++] = DU_SLASH;
						break;

			case '-':	ptr++;
						fmt->sep = fmt->tok[indx++] = DU_DASH;
						break;

			case '.':	ptr++;
						fmt->sep = fmt->tok[indx++] = DU_PERIOD;
						break;

			case ',':	ptr++;
						fmt->tok[indx++] = DU_COMMA;
						break;

			case ' ':	ptr++;
						fmt->tok[indx++] = DU_SPACE;
						break;

			case ':':	ptr++;
						fmt->sep = fmt->tok[indx++] = DU_COLON;
						break;

			default:	/*	Stuff anything else onto the fmt string verbatim, */
						/*	one unrecognized char at a time.  A char preceded */
						/*	by a backslash is also stuffed onto the string.  */
						if (*ptr == '\\')
							ptr++;
						fmt->tok[indx++] = DU_UNIDENTIF + (*ptr++);
						break;
		}
	}

	if (*ptr == '\n')		/*	annihilate an ending return */
		*ptr = '\0';

	fmt->toklen = indx;

	return (AOK);
}

short  axDays[12] = {
	  31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };


int ax_get_3digit_day(day, month, year)
int day, month, year;
{
int ix, leap_year;
int foo = 0;

	/* get the number of days since start of the year and return that value */
	--month; /* since we go from 0 - 1 less in month to get the number of days*/
	if (month < 0)
		return(0);

	for (ix = 0; ix < month; ++ix) {
		foo += axDays[ix];
	}

	leap_year = (((year % 4) == 0) && ((year% 100)!= 0)) || 
		(0 == (year % 400)) || (year == 1900);

	if (foo > 59 )
		foo += (leap_year ? 1 : 0);
	foo += day;

	return(foo);
}

void ax_extract_date(year, ddd_day, day, month)
int year;
int ddd_day;
int *day, *month;
{
	int ix;
	int tday;
	int leap_year;

	leap_year = (((year % 4) == 0) && ((year% 100)!= 0)) || 
		(0 == (year % 400)) || (year == 1900);

	tday = ddd_day;

	for (ix = 0 ; ix <= 11; ++ix) {
		tday -= (axDays[ix] + ((leap_year && ix == 1) ? 1 : 0)) ;
		if (tday < 0) {
			tday += axDays[ix];
			break;
		}
	}

	*month =  ix +1;
	*day = tday;
}
