/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: gtdeflt.c
   Applix ASC version: 4
   Date: Mon Nov 16 13:58:00 1998
*/
#include <aos.h>
#include <getstr18.h>
#include <getstr5.h>

extern typeErr getStr();


/************************************************************/
/* Function:  Get_Default_Margins                           */
/*                                                          */
/*            This function will read the default margins   */
/*  and header and footer distances from Applixstfl and     */
/*  return the values in mills.                             */
/*                                                          */
/************************************************************/

Get_Default_Margins(top, bottom, left, right, hdist, fdist)
int *top, *bottom, *left, *right, *hdist, *fdist;
{
	double val, UPdblVal(), hdr;
	extern int AxMetric;

/*

Strange machinations here!  The profile top/bottom include the
distance for the hdr/ftr, but the reported values are subtracted
for the caller.
*/

/* left margin */

	val = UPdblVal(PROF_LEF_MARGIN);
	if ( AxMetric ) {
		if ( val <= 0.0 )
			val = 1.0;
		val = val / 2.54;
	}
	else if ( val <= 0.0 )
		val = 0.5;
	*left = val * 1000;

/* right margin */

	val = UPdblVal(PROF_RIT_MARGIN);
	if ( AxMetric ) {
		if ( val <= 0.0 )
			val = 1.0;
		val = val / 2.54;
	}
	else if ( val <= 0.0 )
		val = 0.5;
	*right = val * 1000;

/* top & hdr */

	val = UPdblVal(PROF_TOP_MARGIN);
	if ( AxMetric ) {
		if ( val <= 0.0 )
			val = 2.0;
		val = val / 2.54;
	}
	else if ( val <= 0.0 )
		val = 1.0;

	hdr = UPdblVal(PROF_HDR_MARGIN);
	if ( AxMetric ) {
		if ( hdr <= 0.0 )
			hdr = 1.0;
		hdr = hdr / 2.54;
	}
	else if ( hdr <= 0.0 )
		hdr = 0.5;

	if ( val < hdr )
		val = hdr;
	*top = hdr * 1000;
	*hdist = (val - hdr) * 1000;

/* bottom & ftr */

	val = UPdblVal(PROF_BOT_MARGIN);
	if ( AxMetric ) {
		if ( val <= 0.0 )
			val = 2.0;
		val = val / 2.54;
	}
	else if ( val <= 0.0 )
		val = 1.0;

	hdr = UPdblVal(PROF_FTR_MARGIN);
	if ( AxMetric ) {
		if ( hdr <= 0.0 )
			hdr = 1.0;
		hdr = hdr / 2.54;
	}
	else if ( hdr <= 0.0 )
		hdr = 0.5;

	if ( val < hdr )
		val = hdr;
	*bottom = hdr * 1000;
	*fdist = (val - hdr) * 1000;
}


/***********************************************************/
/* Function:  Get_Default_Language                         */
/*                                                         */
/*            This function will read a string from        */
/*  from Applixstrfl and return the value as a int.        */
/*                                                         */
/***********************************************************/

Get_Default_Language(language)
int *language;
{
	char def_language[STRLONG];

	getStr(DEFLANG, def_language);
	*language = atoi(def_language);

}
