/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: helpman.c
   Applix ASC version: 11
   Date: Mon Dec 21 21:12:33 1998
*/
#include <stdio.h>
#include <aos.h>

typedef struct help_data_t {
	struct help_data_t *next;
	int filepos;
#if ALPHA || (LINUX && __alpha__)
	int foo;			/* Need for proper mem alignment (see cm.c) */
#endif
	char name[1];		/* varying extent */
} help_data_t;

#define HASH_PRIME 1011 /* Hash basis (must be a prime number) */
static FILE *HelpDataFile;
static help_data_t *HelpLists[HASH_PRIME];
static help_data_t* HelpHashTable;		/* The actual data */

extern char	*AxLangDir;		/*	Where language-dependent system files are */
extern char *AxLangSuffix[];
extern int AxLanguage;

/*************************************************************************
function: HelpManFd - Return file desc of database
*/
HelpManFd ()
{
	if ( !HelpDataFile )
		return(-1);
	return(fileno(HelpDataFile));
}

/*************************************************************************
function: HelpScreenPos - Search for a topic
*/
HelpScreenPos (name)
char *name;
{
	register help_data_t *p;

	if ( !HelpDataFile )
		return(0);
	for ( p = HelpLists[hash(name)]; p; p = p->next ) {
		if ( streq(name,p->name) )
			return(p->filepos);
	}
	return(0);
}

#if !ELFRT_ONLY
/*************************************************************************
function: HelpInit - Load lookup info, leave file open
*/
HelpInit ()
{
	int hashLens[HASH_PRIME];
	char filename[STRPATH+1];
	help_data_t *p;
	char *cp;
	int ix, jx, size;
	register int len,pad;
	FILE *fp;

    if ( AxLanguage <= 1 )
        /* english */
        sprintf(filename,"%s/help_pro.dat",AxLangDir );
    else
        /* a language other than english */
        sprintf( filename, "%s/help_pro.%s",
                                    AxLangDir, AxLangSuffix[AxLanguage]);

	HelpDataFile = fp = fopen(filename,"r");
	if ( !fp )
		return;

/* read hash lens */

	if ( 1 != fread(hashLens,sizeof(hashLens),1,fp) )
		goto abandon;

/* read alloc word */

	if ( 1 != fread(&size,sizeof(size),1,fp) )
		goto abandon;

/* read table */

	HelpHashTable = (help_data_t*)malloc(size+sizeof(help_data_t));
	if ( !HelpHashTable )
		goto abandon;
	if ( 1 != fread(HelpHashTable,size,1,fp) )
		goto abandon;

/* thread structs together */

	p = HelpHashTable;
	cp = (char*)p;
	for ( ix = 0; ix < HASH_PRIME; ix++ ) {
		for ( jx = 0; jx < hashLens[ix]; jx++ ) {
			if ( jx == 0 )
				HelpLists[ix] = p;
			cp += 2*sizeof(char *);	/* header on help_data_t */
			len = strlen(cp)+1;
			pad = len%sizeof(char *);
			cp += len;
			if ( pad )
				cp += sizeof(char *)-pad;
			if ( jx == hashLens[ix]-1 )
				p->next = 0;
			else
				p->next = (help_data_t*)cp;
			p = (help_data_t*)cp;
		}
		p->next = 0;
	}
	return;

abandon:
	HelpDataFile = NULL;
	if ( fp )
		fclose(fp);
	return;
}
#endif

