/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: istrcmp.c
   Applix ASC version: 13
   Date: Mon Nov 16 13:57:59 1998
*/
#include <stdio.h>
#include <aos.h>
#include <wchars.h>

/* 
.BEGIN MANPAGE 
.NAME Istrcmp - Perform Applix international string comparison.
.ARGUMENTS s1
	The address of a string.
.ARGUMENTS s2 
	The address of a string. 
.ARGUMENTS langnum 
	A Proximity language number; determines which sorting order gets used.
.ARGUMENTS case_sensitive 
	TRUE/FALSE
.DESCRIPTION 
	Istrcmp() is similar to strcmp() except characters get mapped to
	their international equivalents prior to comparison.
.RETURN VALUE 
	Istrcmp() returns an integer greater than, equal to, or less than 0, 
	accordingly as s1 is lexicographically greater than, equal to or less 
	than s2.
.END MANPAGE 
*/

#undef strcmp

#if 1	/*{*/

	/*	This is #if 0'd while being fixed in lpgms.  When changing it
	 *	back to #if 1, remember to change the decl in strinit.c to extern
	 */

unsigned int Atoupper();

unsigned char Imap[(16+1)*2][256];	/*	# of proximity langs + 1 because
							 *	they're 1-based; * 2 for non-prox langs */

/* BEGIN SYNOPSIS */ 
Istrcmp(s1,s2,langnum,case_sensitive)
register unsigned char *s1;
register unsigned char *s2;
register int langnum;
register int case_sensitive;
/* END SYNOPSIS */
{
	register int c1,c2;
#if BUILD_DEBUG
    int ret2;
    ret2 = strcmp( s1, s2 );
#endif
	
	if(s1 == s2)
		return(0);

    if ( strcmp(s1, s2) == 0 )
        return(0);

	if( Imap[langnum][1] == 0 )		/*	no map for this language */
		langnum = 1;				/*	 revert to English */


#if BUILD_ORIENT
	if (IS_MULTI_BYTE(s1)) {
		c1 = MAKE_WIDE_FROM_MULTI(s1);
		s1 += 2;
	} else if (case_sensitive) {
		c1 = *(Imap[langnum] + *s1++);
	} else {
 		c1 = *s1++;
		c1 = *(Imap[langnum] + Atoupper(c1));
	}
	if (IS_MULTI_BYTE(s2)) {
		c2 = MAKE_WIDE_FROM_MULTI(s2);
		s2 += 2;
	} else if (case_sensitive) {
		c2 = *(Imap[langnum] + *s2++);
	} else {
 		c2 = *s2++;
		c2 = *(Imap[langnum] + Atoupper(c2));
	}

	while( (c1 == c2) && (c1 != 0) ) {
		if (IS_MULTI_BYTE(s1)) {
			c1 = MAKE_WIDE_FROM_MULTI(s1);
			s1 += 2;
		} else if (case_sensitive) {
			c1 = *(Imap[langnum] + *s1++);
		} else {
			 c1 = *s1++;
			c1 = *(Imap[langnum] + Atoupper(c1));
		}
		if (IS_MULTI_BYTE(s2)) {
			c2 = MAKE_WIDE_FROM_MULTI(s2);
			s2 += 2;
		} else if (case_sensitive) {
			c2 = *(Imap[langnum] + *s2++);
		} else {
			 c2 = *s2++;
			c2 = *(Imap[langnum] + Atoupper(c2));
		}
	}
#else
	if (case_sensitive)
	{
        c1 = *(Imap[langnum] + *s1);
        c2 = *(Imap[langnum] + *s2);
		while( (c1 == c2) && *s1 && *s2) 
			{
            s1++;    s2++;
            c1 = *(Imap[langnum] + *s1);
            c2 = *(Imap[langnum] + *s2);
            }
	}
	else
	{
		while( *s1 && *s2)
		{
			/*	Atoupper correctly toupper's accented chars, but is
			 *	very dependent on specific char tests for its subroutine,
			 *	Aislower() which give incorrect results if we set c1 and c2
			 *	to the imap value in the loop then Atoupper() that value.
			 *	So we'll Atoupper() our char here instead of in the while loop
			 *	because the correct result is still returned if the while
			 *	loop fails for lack of c1 or c2.
			 */
		 	c1 = *s1++;
		 	c2 = *s2++;

			c1 = *(Imap[langnum] + Atoupper(c1));
			c2 = *(Imap[langnum] + Atoupper(c2));
			if (c1 != c2)
				return(c1 - c2);	/* strings are not equal */
		}
		c1 = *s1;
		c2 = *s2;
		if(c1)
			c1 = *(Imap[langnum] + Atoupper(c1));
		if(c2)
			c2 = *(Imap[langnum] + Atoupper(c2));
		return(c1 - c2);
	}
#endif


#if BUILD_DEBUG
    if ( case_sensitive )
        {
        int ret1;
        ret1 = c1 - c2;
        if ( ((ret1 < 0) && (ret2 >=0)) ||
             ((ret1 == 0) && (ret2 != 0)) ||
             ((ret1 > 0) && (ret2 <=0)) )
           /* different resutls, which should be the same in English ! */
           ret2 = 0;    /* a place to put a breakpoint */
        }
#endif
	return(c1 - c2);
}



/*** do a wide string comparison */
int IWstrcmp(str1, str2, langnum, case_sensitive)
register wchar *str1;
register wchar *str2;
register int langnum;
register int case_sensitive;
{
    register unsigned int char1,char2;
 
    if (str1 == str2)
        return(0);
 
	if (case_sensitive)
	{
    while ( ((char1 = *str1++) != 0) && ((char2 = *str2++) != 0)) 
        {
        if ( !IS_ORIENT_CHAR(char1) )
            char1 = *(Imap[langnum] + char1);
        if ( !IS_ORIENT_CHAR(char2) )
            char2 = *(Imap[langnum] + char2);
        if (char1 != char2)
            break;
        }
	}
	else
	{
    while ( ((char1 = *str1++) != 0) && ((char2 = *str2++) != 0)) 
        {
        if ( !(IS_ORIENT_CHAR(char1)) )
            char1 = Atoupper(*(Imap[langnum] + char1));
        if ( !(IS_ORIENT_CHAR(char2)) )
            char2 = Atoupper(*(Imap[langnum] + char2));
        if (char1 != char2)
            break;
        }
	}
    return(char1 - char2);
}
 

#endif    /*  } BUILD_WIDE */

