/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: parsnum.c
   Applix ASC version: 25
   Date: Thu Jan 28 21:10:59 1999
*/
#if !HP300
/* Causes a redeclaration error on HP */
#include <string.h>
#endif
#include <math.h>
#if X386_NEW
#include <stdlib.h>
#endif

#include <setjmp.h>
#include <stdlib.h>

#include <ctype.h>
#include <utloc.h>
#include <style.h>

#define LOGGER(x)

extern foreignChars UTsymbols;
extern foreignChars USAchars;
foreignChars UTsymbols =
{ 0, 0, '.', ',', 1, "$", 1, "$", 1, 0, 0, 0, 1, "","",0, ""};

foreignChars USAchars =
{ 0, 0, '.', ',', 1, "$", 1, "$", 1, 0, 0, 0, 1, "","",0, ""};

foreignChars    *special_symbols_ptr = NULL;

extern foreignChars UTprofsym;

#define MAX_MANTISSA    1.0


int FPExceptionType = 0;

#include <parsnum.p>

/*
.BEGIN MANPAGE
.NAME ParseNum - Parse a numeric string.
.SYNOPSIS
#include <style.h>
.ARGUMENT type
    PNmINTEGER implies that the numeric string must contain only
    numeric characters.  Leading an trailing space characters are also accepted.

    PNmREGULAR implies that the numeric string may contain a currency
    symbol, percent sign, decimal point, thousand separators, or
    an exponential symbol (for scientific style).
.ARGUMENT str
    The address of the numeric string to be parsed.
.ARGUMENT ret_style
    The address to receive the style of the numeric string.
.ARGUMENT ret_value
    The address to receive the numeric value of the string.
.DESCRIPTION
    Given an ASCII string, ParseNum() attempts to parse the string into
    a double number and return the style (e.g., currency, percent) and
    precision as specified in style.h.  ParseNum() uses the external
    foreignChars UTsymbols to identify the currency symbol, decimal point,
    and thousands separator and to return whether a decimal point was found
    and whether the currency symbol was leading.
.NOTES
    Precision is the number of digits after the decimal point. (A precision
    of 0 does not imply whether there was a decimal point or not.
    The variable UTsymbols.decimal_found gives this info.)
.RETURN VALUE
    ParseNum() returns 1 if it successfully parsed the string and
    0 if it fails.
.END MANPAGE
*/

/* 
 ***********************************************************************
 Revision History: Prem July 31, 1990.
 added correct parsing of scientic numbers.

 Cleanup:
    8/28/93 - Prem
    In future, we don;'t need to compute the value using axpow() as we
    go on. Also for scientific numbers we don't need to parse the mantissa
    and exponent separately.

    All that needs to be done is to collect all the numeric characters
    i.e; [1-9][eE][+-][\.] into a string, while making sure that these are
    found in right sequence and call strtod to convert it to a number.
 ************************************************************************
 */
/* BEGIN SYNOPSIS */
ParseNum(type,str,ret_style, ret_prec, ret_value)
short type;
#if AVIION
unsigned char *str;
#else
unsigned char *str;
#endif
short *ret_style;
short  *ret_prec;
double *ret_value;
/* END SYNOPSIS */
{
register int c;
register int totalchars = 0;
unsigned char numberstr[100];
char  mantissa_str[100];
register int style;
int precision,digit_count;
int is_negative,got_decimal,got_left_paren,got_digit,got_comma,c_foo;
double value,dec_value,base_science,gotexponentsym;

    /* additional variables added for parsing a scientific number */
    double mantissa=0.0;
    int exp=0;
#if AVIION
    unsigned char *expstr;
    unsigned char *numstr;
#else
    char *expstr;
    char *numstr;
#endif
    short isnegmantissa = 0, isnegexp = 0,len = 0;
    foreignChars    *local_sym_ptr;

    numberstr[0] = '\0';

    /* set the symbols ptr */
    if(special_symbols_ptr != NULL)
        local_sym_ptr = special_symbols_ptr;
    else    /* shouldn't reach here, but if you did .. */
        local_sym_ptr = &USAchars;

    /* return value 0.0 if error */
    *ret_style = SS_STL_GENERAL;
    *ret_value = 0.0;
    *ret_prec = 0;

    /* set up auxiliary  return status stuff.  For example, the ss
       uses local_sym_ptr->decimal_found to decide the precision of
       currency. */
    local_sym_ptr->decimal_found = 0;
    local_sym_ptr->curr_sym_leading = 0;

    while((c = *str++) == ' '); /* pass by leading spaces */
    if(!c)      /* return if string was just spaces */
        return(FALSE);

    style = SS_STL_GENERAL;
    is_negative = FALSE;
    got_left_paren = FALSE;
    got_comma = FALSE;

    if(type == PNmINTEGER) 
        {
        /* check for minus */
        if(c == '-') 
            {
            is_negative = TRUE;
            numberstr[totalchars++] = c;
            while((c = *str++) == ' ');
            }
        }
    else 
        {
        /* check for $, $-, (, ($, or - */
        if(local_sym_ptr->moneyLen > 0 &&
           !Astrncmp(str - 1,local_sym_ptr->moneyChar, local_sym_ptr->moneyLen))
            {
            style = SS_STL_CURRENCY;
            local_sym_ptr->curr_sym_leading = 1;
            str += local_sym_ptr->moneyLen - 1;
            while((c = *str++) == ' ');
            if(c == '-') 
                {               /* is number negative? */
                is_negative = TRUE;
                numberstr[totalchars++] = c;
                while((c = *str++) == ' ');
                }
            }
        else if(c == '(') 
            {
            is_negative = TRUE;
            numberstr[totalchars++] = '-';
            while((c = *str++) == ' ');
            got_left_paren = TRUE;
            if(local_sym_ptr->moneyLen > 0 &&
               !Astrncmp(str - 1,local_sym_ptr->moneyChar,
                                                    local_sym_ptr->moneyLen))
                {
                style = SS_STL_CURRENCY;
                local_sym_ptr->curr_sym_leading = 1;
                str += local_sym_ptr->moneyLen - 1;
                while((c = *str++) == ' ');
                }
            else
                style = SS_STL_COMMA;
            }
        else if(c == '-') 
            {
            is_negative = TRUE;
            numberstr[totalchars++] = '-';
            isnegmantissa = TRUE;
            while((c = *str++) == ' ');
            }

        /* You are pointing to begining of the string */
        /* after the sign symbol, if any , */
        /* Also, note that str is pointing to second digit, and not the first*/
           numstr = str-1;
        }

    value = 0.0;
    dec_value = 0.0;
    got_decimal = FALSE;
    precision = 0;
    got_digit = FALSE;
    digit_count = 0;
    gotexponentsym = FALSE;

    for(; c; c = *str++) 
        {
        /* check for digit */
#if BUILD_KANJI
        /* Here c may be half of an EUC code(Applixware internal encoding),
         * while the environment may be Shift-JIS, it had happend that even
         * the 8th bit of c was set, isdigit(c) returned TRUE. This maight
         * happend because of the in-coincidence of setlocale() and isdigit().
         */
        if(!(c&0x80) && isdigit(c))
#else
        if(isdigit(c)) 
#endif
            {
            got_digit = TRUE;

            /* ignore digits after the 15'th significant digit.  This
               is important to the spreadsheet. */
            if(digit_count>MAXDIGITS)
                continue;
            numberstr[totalchars++] = c;
            if(digit_count || (c != '0'))
                digit_count++;

            if(got_decimal)
                {
                if(!gotexponentsym)
                    {/* meaning that these digits are before the exponent */
                    /* symbol */
                    precision++;
                    }
                /* append the characters to our mantissa string */
                mantissa_str[len++] = c;
                }
            value = value * 10.0 + (c - '0');
            continue;
            }

        if(c == ' ')        /* if space, must be spaces to end */
            break;

        if(type == PNmINTEGER)
            return(FALSE);

        /* check for decimal point character */
        if(c == local_sym_ptr->decimalChar && !got_decimal ) 
            {
            if(style == SS_STL_SCIENCE)
                return(FALSE);
            got_decimal = TRUE;
            /* put a literal '.' since numberstr and also mantissa_str */
            /* are used for scientific format only and we call atof */
            /* for conversions */
            numberstr[totalchars++] = '.';
            local_sym_ptr->decimal_found = 1;
            dec_value = value;
            value = 0.0;


            /* Found a period here     */
            /* Have to normalize the mantissa, so that */
            /* the period is always after the first digit */
            /* and we have to adjust the exponent if we moved */
            /* the decimal point */

            /* Construct the mantissa string */

            if( *numstr == local_sym_ptr->decimalChar)
                mantissa_str[0] = '0';
            else
                mantissa_str[0] = *numstr; /* first digit in the number string*/
            mantissa_str[1] = '.'; /* Place the decimal point here */

            /* move the remaining digits in the mantissa to, after the */
            /* decimal point in the mantissa_str */
            if( *numstr == local_sym_ptr->decimalChar)
                len = 0;
            else
                len = (int) (((long)(str-1))  - ((long) (numstr + 1)));
            /* copy to the third place in the mantissa_str, since */
            /* zeroth position for first significant digit, and */
            /* second place for decimal point */
            if(len > 0)
                strncpy(&(mantissa_str[2]) , (numstr+1), len);

            /* adjust the exponent */
            exp = len;

            /* Now make len the current index in the mantissa string */
            /* len was equal to number of characters between the */
            /* first digit and the decimal point. To make it the index */
            /* such that it points to end of mantissa_str we have to */
            /* add 2 i.e; we have to into account the first character */
            /* and the decimal point */
            len = len + 2;

            continue;
            }

        /* comma only legal if general/comma/money */
        if(c == local_sym_ptr->commaChar) 
            {
            if(style != SS_STL_GENERAL && style != SS_STL_COMMA && style != SS_STL_CURRENCY)
                return(FALSE);
            if(style == SS_STL_GENERAL)
                style = SS_STL_COMMA;
            continue;
            }

        if((c == 'e' || c == 'E') && style == SS_STL_GENERAL) 
            {
            if(!got_digit)
                return(FALSE);
            style = SS_STL_SCIENCE;
            gotexponentsym = TRUE;
            numberstr[totalchars++] = c;
            if(got_decimal)
                {
                /* Now terminate our mantissa string here, since we */
                /* found the exponent character */
                mantissa_str[len] = '\0';
                }
            else
                { /* there is no decimal point in this scientific */
                /* number. So plant a decimal point and adjust */
                /* exponent */
                len = ((long) (str - 1))- ((long) numstr); /* number of mantissa digits */
                mantissa_str[0] = *numstr; /* plant the first digit */
                mantissa_str[1] = '.'; /* decimal point here */
                strncpy(&mantissa_str[2],(numstr+1), (len-1));
                mantissa_str[len+1] = '\0'; 
                exp += (len - 1);
                }
 

            /* The exponent part of the number starts here */
            /* Keep track of the exponent , So create a pointer */
            /* to point to this part of the string */
            /* expstr includes the sign character if any */
            /* Here c is either 'E' or 'e' and str is pointing to */
            /* the next character after the exponent symbol */
            expstr = str;
            

            base_science = dec_value + (value * axpow(10.0,(double)-precision));
            if(is_negative)
                base_science = -base_science;
            value = 0.0;
            dec_value = 0.0;
            is_negative = FALSE;
            got_decimal = FALSE;
            if((c = *str++) == '-')
                {
                /* found a negative sign before the exponent */
                /* set "is negative exponent" flag to TRUE */
                isnegexp = TRUE;
                is_negative = TRUE;
                numberstr[totalchars++] = '-';
                }
            else if(c != '+')
                str--;
            continue;
            }

        /* check for percent, that must be trailing */
        if(c == '%' && style == SS_STL_GENERAL) 
            {
            style = SS_STL_PERCENT;
            c = *str++;
            break;
            }

        /* check for right paren, that must be trailing */
        if(c == ')' && got_left_paren) 
            {
            c = *str++;
            break;
            }

        /* check for trailing currency */
        if(local_sym_ptr->moneyLen > 0 &&
            !Astrncmp(str - 1,local_sym_ptr->moneyChar,
                                                    local_sym_ptr->moneyLen) &&
            (style == SS_STL_COMMA || style == SS_STL_GENERAL)) 
            {
            style = SS_STL_CURRENCY;
            str += local_sym_ptr->moneyLen - 1;
            if((c = *str++) == ')' && got_left_paren)
                c = *str++;
            break;
            }

        return(FALSE);              /* unrecognized character */
        }

    numberstr[totalchars] = '\0';

    while(c == ' ')
        c = *str++;

    if(c || !got_digit)
        return(FALSE);

    if(style != SS_STL_SCIENCE)
        {/* get the most accurate value, don't use the value generated */
        /* by using axpow as we were parsing the string. There is an error */
        /* associated with using axpow */

#if ALPHA /*Alpha atof()/strtod() seg faults when passwd non-ASCII. Hack for*/
          /*3.11 FCS to avoid crash when using Special Chrs in SS -sean */
        c_foo = (int)numberstr[0];
        if (isascii(c_foo))
          value = atof(numberstr);
        else {
          value = 0.0;
          return(FALSE);
        }
#else
        value = atof(numberstr);
#endif
        
        }

    if(style == SS_STL_PERCENT)
        value /= 100.0;

    *ret_value = value;

    if(style == SS_STL_SCIENCE) 
        {
        value = dec_value + (value * axpow(10.0,(double)-precision));

    /*
     ***********************************************************************
     Now evaluate the mantissa and exponent separately, and check for the
     max limit defined. The following table will give an example of how to
     check for mantissa and exponent.

     Mantissa       exponent        status
     --------       --------        ------

     1.0            25              valid.

     1.1            25              invalid.

     9.0            24              valid.


     From the above table , we can find out that if the exponent
     is greater than the MAX_EXPONENT , the number is clearly
     out of range. If the exponent is equal to MAX_EXPONENT then the
     mantissa has to be less than the MAX_MANTISSA.


     Number is VALID    If    { exponent < MAX_EXPONENT }

                                 OR

                              {
                                (exponent == MAX_EXPONENT)

                                  AND

                                 mantissa <= MAX_MANTISSA for +ve exponent
                                 mantissa >= MAX_MANTISSA for -ve exponent
                               }

     ***********************************************************************
     */
       /*
        * If a number is like this 0.005E10, it has to be made 
        * like 5.0e8
        */
        if(mantissa_str[0] == '0')
            {/* number of starts with a zero */
            len = 0;
            /* point to first digit afer decimal point */
            str = &(mantissa_str[2]); 
            /* get to the first significant digit after decimal point */
            while( (*str == '0') && (*str != '\0'))
                {
                str++;
                len++;
                }

            if(*str != '\0')
                {/* skip the first significant digit also */
                mantissa_str[0] = *str;
                str++;
                len++;
                }
            exp -= len; 
            strcpy(&mantissa_str[2], str);
            }

        /* Now calculate the mantissa */
        mantissa =  fabs( atof(mantissa_str));

        /* Now calculate the exponent */
        exp = exp + atoi(expstr);


        /* Now actually check for inrange */

        /* If exponent is greater than MAX_EXPONENT then */
        /* out of range */
        if( abs(exp) > MAX_EXPONENT )
            {
            *ret_style = style + ((precision > 9) ? 9 : precision);
            if(mantissa == 0.0)
                { 
                *ret_value = 0.0;
                return(TRUE);
                }
            *ret_style = SS_STL_GENERAL_NP;
            return(FALSE);
            }

        /* If exponent is equal to MAX_EXPONENT check mantissa */
        if( abs(exp) == MAX_EXPONENT)
            {
            /* exponent is equal to MAX_EXPONENT */
            if( isnegexp)
                { /* negative exp */
                if( mantissa < MAX_MANTISSA )
                    {
                    *ret_style = SS_STL_GENERAL_NP;
                    return(FALSE); /* can't handle numbers smaller than */
                                    /* this */
                    }
                 }
            else
                { /* positive exponent */
                if(mantissa > MAX_MANTISSA )
                    {
                    *ret_style = SS_STL_GENERAL_NP;
                    *ret_prec = SS_MAX_PREC;
                    return(FALSE); /* can't handle numbers larger than */
                                    /* this */
                    }

                 }
            }
        /* this is how scientic numbers are check for validity */
        /* before, clearly this is very insufficient to invalidate*/
        /* a scientific number, so this is being ifzeroed out */

        if(is_negative) 
            {
            is_negative = FALSE;
            value = -value;
            }
        value = atof(numberstr);
        *ret_value = is_negative ? -value : value;
        }

    /* put precision in style. */ 
    *ret_prec =  ((precision > 9) ? 9 : precision);
    *ret_style = style;
    return(TRUE);
}

jmp_buf pow_env;

double axpow(x,y)
double x,y;
{
double retval;
int tmp = FPExceptionType;
    FPExceptionType = 2;
    LOGGER(("axpow() "));
    if ( setjmp(pow_env) != 0 )
        {
        FPExceptionType = tmp;
        LOGGER((" <<<<< axpow from LONGJMP"));
        return((double)1.0e+38); /* maybe this should be SPL_ERROR */
        }
    if(x== 0.0 && y < 0.0)
        retval = 0.0;
    else 
        retval = pow(x,y);
    LOGGER(("RETVAL = %lf\n",retval));
    FPExceptionType = tmp;
    return(retval);
}


/* The following function allows an external caller to use the ParseNum
    function, without having to worry about include files, etc.

    This wrapper allows the caller to easily distinguish between an
    integer and non-integer result (e.g. so the string "14" will not
    be interpreted as 13.999999). Numbers in scientific notation are
    considered not to be integer, since it is non-trivial to check if
    they are or not.

    The primary motivation for this function was to allow the ELF-based
    WP Sorting and Calculation functions to "properly" convert strings
    (representing numbers in the WP) into numeric values, which could
    be used by these ELF functions. - D Miller 10/19/93
    
    fix for Bug#103353, some floating point libraries like HP800,SCO UNIX
    cannot handle int cast for numbers > MAXINT, so set is_int false for
    int underflow/overflow cases - karun 07/15/96
*/
int ParseNumSimple ( str, dval, ival, is_int)
unsigned char   *str;   /* String to parse - should represent number, possibly
                        surrounded by parenthesis and/or whitespace */
int     *is_int;    /* OUT: 1: passed number is an interger value
                            0: passed number is a float or exponent */
double  *dval;      /* OUT: Float value of passed number */
int     *ival;      /* OUT: Integer value of passed number - ignore unless
                        is_int is set to 1. */
{
    short ret_style, ret_prec;
    int retval;

    /* Grab desired format rules from prefs (e.g. what symbol is
        decimal point) */
    UTgetsym(); 
    special_symbols_ptr = &UTprofsym;       /* set by above call */

    /* Parse the string */
    retval = ParseNum( PNmREGULAR, str, &ret_style, &ret_prec, dval);

    /* If no decimal point, treat as an int */
    /* Exponentials are treated as floats because they can force a float 
        without a decimal point: e.g. 32e-1 = 3.2 */
    if (!special_symbols_ptr->decimal_found && !(ret_style & SS_STL_SCIENCE)
        && *dval > MININT+2 && *dval < MAXINT-2 ) { 
        /* 14 could be represented as 13.99999 or 14.00001 */
        if ( *dval >= 0 ) {
            *ival = (int)(*dval + 0.5);
            *is_int = TRUE;
        }
        else {
            *ival = (int)(*dval - 0.5);
            *is_int = TRUE;
        }
    }
    else {
        *ival = 0;
        *is_int = FALSE;
    }
    return( retval);
}
