/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: tempdir.c
   Applix ASC version: 4
   Date: Mon Nov 16 13:58:01 1998
*/
#if AXMS
#include <windows.h>
#endif

#include <stdlib.h>

#include <aos.h>
#include <afile.h>

#ifndef MS_GetTempPath
#define MS_GetTempPath GetTempPath
#endif

#ifdef getenv
#undef getenv
#endif

char *AtempDir()
{
    char path_buff[STRPATH+1];
    char win_dir_path[STRPATH+1];
    static char *pathP = (char *)NULL;
    int ret;
    char *p;

    if ( pathP != (char *)NULL )
       /* already set */
       return( pathP );

#if AXMS                        /*  {  */
    /* Windos/NT */
    /* this does not work well for aagtwy, since when run as a system     */
    /* service, PATH is not set, so that we end up with the current      */
    /* working directory, which is the system directory, ie. "c:\WINNT40\" */
    /* Better to try for the environment ourselves (since that is all      */
    /* "GetTempPath" really does) in case we are running as a user, but     */
    /* if these fail (meaning we are running as the system), then go for     */
    /* a default of "C:\temp"                                               */
    ret = MS_GetTempPath( sizeof(path_buff), path_buff );
    if ( ret != 0 )
       {
       /* return value is length of path, not including null terminator */
       /* create a permenent version of the string for current and future */
       p = path_buff + strlen(path_buff) - 1;
       if ( (*p == '\\') || (*p == '/') )
           /* we keep path without trailing slash */
           *p = '\0';

       GetWindowsDirectory( win_dir_path, sizeof(win_dir_path) );
       if ( stricmp(win_dir_path, path_buff) == 0 )
           {
           /* we got the system directory, we must be running as a */
           /*  LocalSystem account.   Default to "C:\temp"         */
           goto default_setting;
           }

       pathP = (char *)malloc( strlen(path_buff) + 1 );
       strcpy( pathP, path_buff );
       return( pathP );
       }
#endif                    /*  }  AXMS  */


default_setting:
    /* try the environment variables */
    p = getenv(NXLT("AX_TEMP_DIR"));
    if ( p != (char *)NULL )
       /* found one */
       goto return_path;

    p = getenv(NXLT("tmp"));
    if ( p != (char *)NULL )
       goto return_path;
    p = getenv(NXLT("TMP"));
    if ( p != (char *)NULL )
       goto return_path;

    p = getenv(NXLT("temp"));
    if ( p != (char *)NULL )
       goto return_path;
    p = getenv(NXLT("TEMP"));
    if ( p != (char *)NULL )
       goto return_path;

    /* very last choice, hard coded */
#if AXMS
    p = "\\temp";     /* temp dir on current drive */
    if ( -1 == access( p, W_OK ) )
         {
         /* very last chance */
         p = "c:\\temp";
         }
#else
    p = "/tmp";
#endif

return_path:
    pathP = (char *)malloc( strlen(p) + 1 );
    strcpy( pathP, p );              /* save static copy */
    return( pathP );
}


