/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: utconvt.c
   Applix ASC version: 15
   Date: Fri Jan  8 21:11:44 1999
*/
#include <stdio.h>
#include <stdlib.h>

#if SYSV4 || AXMS || CLIX || M_88K || SCOUNIX
#include <fcntl.h>
#else
#include <sys/file.h>
#endif

#include <utloc.h>
#include <atmattr.h>

#define SIZE 32768

#define XTAB NXLT("native.tbl")

/* characters which are undefined or don't map */
#define FLAGCHAR '^'

static char *fgetl();

extern char *ThimApplixDir;

static char *Native = NULL;
static char *Sym1 = NULL;
static char *Sym2 = NULL;

static int InitTab();
static int GetValue(/* char *valP */);

/*
.BEGIN MANPAGE
.NAME ut_file_convert - Converts 8-bit text characters in a file
	between machine native set and the Applix internal set.
.ARGUMENT source
	The address of a string which identifies the source file.
.ARGUMENT destination
	The address of a string which identifies the destination file.
.ARGUMENT char_set_id
	A bit array set as follows:

	0 bit set for internal to native, otherwise native to internal.

	1 bit set for symbol set 1.

	2 bit set for symbol set 2.
.DESCRIPTION
	Ut_file_convert() is used to convert ascii files between
	the machine native character set and the Applix internal
	character set.
.RETURN VALUE
	Ut_file_convert() returns 0, if successful.  Upon error,
	ut_file_convert() returns -1.
.SEE ALSO
	ut_convert(3A)
.END MANPAGE
*/
/* BEGIN SYNOPSIS */
ut_file_convert(source,destination,char_set_id)
char *source;
char *destination;
int char_set_id;
/* END SYNOPSIS */
{
	int k = -1,file = -1,fd = -1;
	char *buf = NULL;

#if !HP800 && !HP300 && !HP800V3
    if ( strcmp( source, destination ) == 0 )
       {
       /* source and destination are the same file.  On all platforms */
       /* other than HP, "ut_convert" is a no-op, so do not bother    */
       /* copying the file to itself.  (Also hangs on NT because it   */
       /* never gets tot he end of the file                           */
       return( 1 );
       }
#endif

	if((fd = open(source,O_RDONLY)) == -1 ||
		(file = open(destination,O_WRONLY|O_CREAT,0660)) == -1 ||
		!(buf = malloc((unsigned int)SIZE)))
		goto ret;

	while(k = read(fd,buf,SIZE)) {
		if(k == -1 || ut_convert(buf,k,char_set_id,buf) ||
			write(file,buf,k) != k)
			break;
	}

ret:
	if(fd != -1)
		close(fd);
	if(file != -1)
		close(file);
	if(buf)
		free(buf);
	return(k);
}





/*
.BEGIN MANPAGE
.NAME ut_convert - Converts 8-bit text characters between native machine
	set and the Applix internal set.
.ARGUMENT in_text
	The address of the text array to translate.
.ARGUMENT text_len
	The length of text array.
.ARGUMENT char_set_id
	A bit array set as follows:

	0 bit set for internal to native, otherwise native to internal.

	1 bit set for symbol set 1.

	2 bit set for symbol set 2.
.ARGUMENT out_text
	The address the buffer to receive the translated text.
.DESCRIPTION
	Ut_convert() initializes character translation arrays from the
	human-readable file: 'native.tbl'.
	If the file doesn't exist, an identity mapping will be done.
.RETURN VALUE
	None.
.NOTES
	Any characters which are undefined are translated to '^'.
	The '^' character was chosen because it is rarely used and moreover
	is a legal character for document names, Alis names, etc. (unlike '~').
.END MANPAGE
*/

/* BEGIN SYNOPSIS */
ut_convert(in_text,text_len,char_set_id,out_text)
REGISTER char *in_text;
REGISTER int text_len;
REGISTER int char_set_id;
REGISTER char *out_text;
/* END SYNOPSIS */
{

	static int init_flag = 0;
	REGISTER char *map;

	if(!init_flag)
		init_flag = InitTab();

	switch(char_set_id & ~0x1) {
	  case 0:
		map = Native;
		break;
	  case 2:
		map = Sym1;
		break;
	  case 4:
		map = Sym2;
		break;
	  default:
		map = (char *)0;
	}

	if(!map || init_flag == -1) {
		if(in_text != out_text)
			bcopy(in_text,out_text,text_len);
		return(0);
	}
	if(char_set_id & 0x1)
		map += 256;
		
	while(text_len--)
		*out_text++ = *(map + (*in_text++ & 0xff));
	return(0);
}


static int InitTab()
{
	static char sep[] = " : ";
	static char file_name[] = XTAB;
	char *memory;
	char line[128 + 1],*ptr;
	int native,applix,font;
	char *map;
	FILE *fp;

	/*
	 * Look for the translation table file first in current directory,
	 * then in 'axdata' 
	 */
	if(!(fp = fopen(file_name,"r"))) {
	    sprintf(line,"%s/%s",ThimApplixDir,file_name);
		if(!(fp = fopen(line,"r")))
			return(-1);
	}

	if(!(memory = malloc((unsigned int)(256 * 6)))) {
		fclose(fp);
		return(-1);
	}
	memset(memory,FLAGCHAR,(256 * 6));
		
	while(fgetl(line,128,fp)) {
		if(*line == '\0')
			continue;
		if(!(ptr = strtok(line,sep)))
			continue;
		native = GetValue(ptr);
		if(!(ptr = strtok((char *)0,sep)))
			continue;
		applix = GetValue(ptr);
		if(!(ptr = strtok((char *)0,sep)))
			continue;
		font = GetValue(ptr);
		map = memory + (256 * 2 * font);
		switch(font) {
		  case 0:
			Native = map;
			break;
		  case 1:
			Sym1 = map;
			break;
		  case 2:
			Sym2 = map;
			break;
		  default:
			continue;
		}
		*(map + native) = applix;
		*(map + 256 + applix) = native;
	}

	fclose(fp);
	return(1);
}


static int GetValue(valP)
char *valP;	/* Ptr to value in ascii */
{
	int value;

	/*
	 * Check mode of expression (hex, octal, decimal, or character)
	 */
	if(*valP == '0') {
		if(*(valP + 1) == 'x' || *(valP + 1) == 'X') /* Hex mode */
			value = strtol(valP + 2,(char **)0,16);
		else	/* Octal */
			value = strtol(valP + 1,(char **)0,8);
	}
	else if(*valP == '\'') /* Character */
		value = *(valP + 1);
	else
		value = strtol(valP,(char **)0,10);

	return((int)(value & 0xff));
}


/*
 * Like fgets but ignores escaped characters and comments delimited by
 * # and (non-escaped) newlines. (An escaped character is preceeded by
 * a '\'.)
 */
static char *fgetl(line,len,fp)
char *line;
register int len;
register FILE *fp;
{
	register int c,escape = 0,comment = 0;
	register char *cs;

	cs = line;
	while((comment || --len > 0) && (c = getc(fp)) != -1) {
		switch(c) {
		  case '\\':
			if(escape)
				break;
			escape = 1;
			continue;

		  case '#':
			if(escape)
				break;
			if(!comment) {
				len++;
				comment = 1;
			}
			continue;

		  case '\n':
			if(escape) {
				if(!comment)
					len++;
				escape = 0;
				continue;
			}
			if(comment) {
				comment = 0;
				if(cs == line)	/* Nothing on the line */
					continue;
			}
			goto ret;
		}
		if(!comment)
			*cs++ = c;
		escape = 0;
	}
ret:
	*cs = '\0';
	return((c == -1 && cs == line) ? (char *)0 : line);
}




