/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: uterror.c
   Applix ASC version: 4
   Date: Mon Nov 16 13:58:00 1998
*/
#include <utloc.h>
#include <errno.h>

int D_errno = 0;

void ut_error(rname,code)
char *rname;
int code;
{
#if !DEBUG
	D_errno = code;
	return;
#else
	static struct {
		int num;
		char *sym;
		char *msg;
	} errtbl[] = {
		ED_BADARG,NXLT("ED_BADARG"),NXLT("Invalid argument specified"),
		ED_NOREC,NXLT("ED_NOREC"),NXLT("A record has not been read"), 
		ED_BADCMP,NXLT("ED_BADCMP"),NXLT("Illegal field-date comparison"),
		ED_NOPRED,NXLT("ED_NOPRED"),NXLT("No predicate list"),
		ED_BADPRED,NXLT("ED_BADPRED"),NXLT("ALL is not allowed"),
		ED_BADFLD,NXLT("ED_BADFLD"),NXLT("Invalid field identifier"),
		ED_BADOPER,NXLT("ED_BADOPER"),NXLT("Illegal comparison operator"),
		ED_BADUID,NXLT("ED_BADUID"),NXLT("No current record uid"),
		ED_MAXOPN,NXLT("ED_MAXOPN"),NXLT("Maximum open"),
		ED_RDMODE,NXLT("ED_RDMODE"),NXLT("Writes not allowed in read mode"),
		ED_NOTUNIQ,NXLT("ED_NOTUNIQ"),NXLT("Not unique key"),
		ED_DELREC,NXLT("ED_DELREC"),NXLT("Record deleted or already used"),
		ED_BADMODE,NXLT("ED_BADMODE"),NXLT("Bad open mode"),
		ED_BADATTR,NXLT("ED_BADATTR"),NXLT("Invalid field attribute"),
		ED_NOWHAT,NXLT("ED_NOWHAT"),NXLT("Whatpred is null"),
		ED_BADSRCH,NXLT("ED_BADSRCH"),NXLT("Algorithmic search error"),
		ED_TBLOPN,NXLT("ED_TBLOPN"),NXLT("Table already open"),
		ED_TOOBIG,NXLT("ED_TOOBIG"),NXLT("Record too large"),
		ED_BADOFF,NXLT("ED_BADOFF"),NXLT("Bad file offset"),
		ED_BADBTPG,NXLT("ED_BADBTPG"),NXLT("Bad Btree page"),
		ED_INTERNAL,NXLT("ED_INTERNAL"),NXLT("Internal failure"),
		ED_BADREC,NXLT("ED_BADREC"),NXLT("Read bad datafile record"),
		ED_ACOPY,NXLT("ED_ACOPY"),NXLT("Acopy failed"),
		ED_MALLOC,NXLT("ED_MALLOC"),NXLT("malloc failed"),
		ED_OPEN,NXLT("ED_OPEN"),NXLT("open failed"),
		ED_READ,NXLT("ED_READ"),NXLT("read failed"),
		ED_UNLINK,NXLT("ED_UNLINK"),NXLT("unlink failed"),
		ED_LSEEK,NXLT("ED_LSEEK"),NXLT("lseek failed"),
		ED_WRITE,NXLT("ED_WRITE"),NXLT("write failed"),
		ED_REALLOC,NXLT("ED_REALLOC"),NXLT("realloc failed"),
		ED_FSTAT,NXLT("ED_FSTAT"),NXLT("fstat failed"),
		ED_STAT,NXLT("ED_STAT"),NXLT("stat failed"),
		ED_LOCK,NXLT("ED_LOCK"),NXLT("flock, lockf, or locking failed"),
		ED_RE_COMP,NXLT("ED_RE_COMP"),NXLT("re_comp failed"),
		ED_RE_EXEC,NXLT("ED_RE_EXEC"),NXLT("re_exec failed"),
		ED_CLOSE,NXLT("ED_CLOSE"),NXLT("close failed"),
		0,"",""
	};
	REGISTER int i;

	if(code == ED_NORMAL) {
		D_errno = ED_NORMAL;
		return;
	}

	for(i = 0; errtbl[i].num && errtbl[i].num != code; i++);

	if(errtbl[i].num == 0)
		trace(0,rname,TRLEV_2,"Unknown error code %d",code);
	else if(code > 0)
		trace(0,rname,TRLEV_2,NXLT("%s: %s (errno = %d)"),errtbl[i].sym,
				 errtbl[i].msg,errno);
	else
		trace(0,rname,TRLEV_2,"%s: %s",errtbl[i].sym,errtbl[i].msg);

	D_errno = code;
	return;
#endif
}
