/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: uttrace.c
   Applix ASC version: 3
   Date: Mon Nov 16 13:57:57 1998
*/
#include <utloc.h>

#if SUN
#include <sys/types.h>
#include <sys/timeb.h>
#include <sys/time.h>
#include <sys/resource.h>
#endif

#if SYS&B4_2
#include <varargs.h>
#else
#include <varargs.h>
#endif

#if DEBUG
FILE *D_trcfp = NULL;
int D_trlevel = TRLEV_1|TRLEV_2|TRLEV_3|TRLEV_4;
int D_trtime = FALSE;
int D_trperf = FALSE;
int D_traceon = FALSE;
int D_trsubsys = DB|A_|D_|RX;
char *D_rname = NULL;
#endif

LINTCOMMENT(VARARGS0)

void is_this_used_trace(va_alist)
va_dcl
{
#if DEBUG
	static char trfmt[] = "%s: %s:[%d] ";
	va_list arg;
	int trlevel,subnum,r,s;
	char *subsys,*rname,*fmt,*sp;
	static FILE *trcfp = NULL;
	char str[16],fbuf[26];
#if SUN
	static char timefmt[] = "%d.%d: ";
	static char perffmt[] = "faults = %d (%d), swaps = %d, real I/O = %d/%d,\
stime = %ld:%ld\n";
	struct rusage usage;
	struct timeb time;
#endif

	va_start(arg);

	subnum = va_arg(arg,int);
	if(subnum && !(subnum&D_trsubsys))
		return;
	switch(subnum) {
	  case DB:
		subsys = "DB";
		break;
	  case A_:
		subsys = "A_";
		break;
	  case D_:
		subsys = "D_";
		break;
	  case RX:
		subsys = "RX";
		break;
	  case R_:
		subsys = "R_";
		break;
	  case 0:
		subsys = "ER";
		break;
	  default:
		return;
	}

	rname = va_arg(arg,char *);

	trlevel = va_arg(arg,int);
	if(!(trlevel&D_trlevel))
		return;

	if(D_trlevel&(TRINCL|TREXCL)) {
		r = ListCk(rname,D_rname,1);
		if((D_trlevel&TRINCL && r != 'i') || (D_trlevel&TREXCL && r == 'e'))
			return;
	}

#if SUN
	if(D_trtime) {
		ftime(&time);
		fprintf(D_trcfp,timefmt,time.time,time.millitm);
	}
#endif

	fprintf(D_trcfp,trfmt,subsys,rname,trlevel);

	fmt = va_arg(arg,char *);

	while(TRUE) {
		while(*fmt != '%') {
			if(*fmt == '\0') {
				putc('\n',D_trcfp);
#if SUN
				if(D_trperf) {
					getrusage(RUSAGE_SELF,&usage);
					fprintf(D_trcfp,perffmt,usage.ru_minflt,usage.ru_majflt,
							usage.ru_nswap,usage.ru_inblock,usage.ru_oublock,
							usage.ru_stime.tv_sec,usage.ru_stime.tv_usec);
				}
#endif
				fflush(D_trcfp);
				return;
			}
			putc(*fmt,D_trcfp);
			fmt++;		/* past the char */
		}
		fmt++;			/* past the % */
		s = FALSE;
		if(*fmt == 'l') {
			s = TRUE;
			fmt++;		/* past the l */
		}
		r = 0;
		sp = str;
		switch(*fmt) {
		  case 'd':
			r = 10;
			break;
		  case 'o':
			r = 8;
			break;
		  case 'x':
#if AT
			s = TRUE;	/* pointers should be longs on AT */
#endif
			r = 16;
			break;
		  case 's':
			sp = va_arg(arg,char *);
			break;
		  case 'c':
			*str = (char)va_arg(arg,int);
			*(str + 1) = '\0';
			break;
		  case 'f':
			if(s)
				sprintf(fbuf,"%.16e",va_arg(arg,double));
			else
				sprintf(fbuf,"%.16e",va_arg(arg,float));
			sp = fbuf;
			break;
		  case 'b':
			sp = va_arg(arg,char *);
			if(!s)
				s = *(short *)sp;
			while(s--) {
				fputc((char)('0' | (*(unsigned char *)sp >> 4)),D_trcfp);
				fputc((char)('0' | (*sp++ & 0x0f)),D_trcfp);
			}
			sp = NULL;
			break;
		  default:
			*str = *fmt; /* also covers %% */
			*(str + 1) = '\0';
		}
		if(r) {
			if(s)
				ax_ltoa(str,r,va_arg(arg,long));
			else
				ax_itoa(str,r,va_arg(arg,int));
		}
		if(sp)
			fputs(sp,D_trcfp);
		fmt++;			/* past the [doxscfb] */
	} /* while */

	va_end(arg);

#else
	return;
#endif
} /* rx_trace */
