/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: valaddr.c
   Applix ASC version: 4
   Date: Mon Nov 16 13:58:00 1998
*/
#include <stdio.h>
#include <ctype.h>


/*
* Definition of conditional symbols (within aos.h, which is within utloc.h).
*/
#include <utloc.h>	/* This includes aos.h !!! so keep it here */


/*
* Conditional system includes.
*/
#if B_SOCKETS || UNISYS
#if MIPS
#include <sys/types.h>
#include <bsd/netdb.h>
#include <bsd/netinet/in.h>
#include <bsd/sys/socket.h>
#else
#if PCS
#include <bsd/sys/types.h>
#include <bsd/netdb.h>
#include <bsd/netinet/in.h>
#include <bsd/sys/uio.h>
#include <bsd/sys/socket.h>
#else
#include <sys/types.h>
#include <netdb.h>
#include <netinet/in.h>
#endif 		/* if PCS */
#endif 		/* if MIPS */
#if SYS&B4_2 || UNISYS
#include <sys/uio.h>
#include <sys/socket.h>
#endif 		/* if SYS&B4_2 || UNISYS */
#endif 		/* if B_SOCKETS || UNISYS */

#if EXCELAN && (!UNISYS)
#include <sys/types.h>
#if ALTOS
#include <exos/misc.h>
#include <sys/socket.h>
#include <netinet/in.h>
#endif
#if X386
#include <exos/sys/socket.h>
#endif
#if ATT3B2
#include <sys/socket.h>
#include <sys/in.h>
#include <netdb.h>
#endif
#if PCS
#include <bsd/sys/types.h>
#include <bsd/sys/socket.h>
#endif		/* if PCS */
#if !ATT3B2
struct hostent
{	/* stuff from netdb.h */
	char *h_name;
	char **h_aliases;
	int h_addrtype;
	int h_length;
	char *h_addr;
};
struct hostent *gethostbyaddr();
struct hostent *gethostbyname();
#endif
#endif

#if HP
#include <sys/types.h>
#include <misc.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <netdb.h>
#endif

#if NCR
#include <sys/socket.h>
#endif

#if NECEWS
#include <net/socket.h>
#endif

#if SGM
#if EXCELAN
#include <exos/exos/misc.h>
#include <exos/sys/socket.h>
#include <exos/netinet/in.h>
#else
#include <sys/socket.h>
#endif		/* if EXCELAN */
#endif		/* if SGM */

/*
* ALIS global includes.
*/
#include <scac.h>
#include <getstr16.h>
#if EXCELAN && (!UNISYS)
#include <paripc.h>
#endif


/*
* Local defines.
*/

#if SYS&AEGISO
#define ADDRLN 8
#endif

#if (SYS&B4_2)==0	/* may not exist if using yellow pages */
#define ETCHOSTS	"/etc/hosts"
#endif

/*
* External definitions.
*/
extern	typeErr	getStr();
#if B_SOCKETS|HP|UNISYS|ATT3B2
extern	char	* inet_ntoa ();
#endif




int validNetAddr (addr, hn)
char * * addr;
char	hn[];
{
	*hn = '\0';

#if SYS&AEGISO
/*
 * Validate that the address contains only kexadecimal digits only, and
 * that it is at most 8 characters long.
 * (NOTE, a validation that the node exists may also be warranted later!)
 */
  {
	register int	i, c;
	char		* s;


	s = * addr;
	for (i = 0; (c = (int) s[i]) != 0 && i < ADDRLN; i ++)
	{
        	if (!isxdigit(c))
		{
			goto errtn;
		}
		if (islower(c))
			s[i] = _toupper(c);
	}

	if (s[i] != '\0')
	{
		goto errtn;
	}

	goto goodret;
  }
#endif

#if B_SOCKETS|EXCELAN|HP|UNISYS
/*
 * Scan through the /etc/hosts file or its equivalent
 * and return true if the supplied address string is
 * in the file. Returns false if the file was not found.
 * Puts appropriate error message on status line.
 */
  {
#if (!X386)
	struct in_addr laddr;
#endif
	unsigned long iaddr, inet_addr();
	struct hostent *hp;
	register char	* s;
	register int	i;


/*
* Note, due to a bug in inet_addr on SUN's, we will trap out the bad
* format that consists of too many tupples before calling the system
* runtime.
*/
	for (i = 0, s = *addr; *s != '\0'; s++)
		if (*s == '.')
			i++;
	if (i > 3)
	{
		goto errtn;
	}
	if ((iaddr = inet_addr (* addr)) == -1)
	{
		goto errtn;
	}
#if SYS&B4_2==0		/* may not exist if using yellow pages */
	if (Aaccess(ETCHOSTS, 04) == -1)
	{
		goto errtn;	/* assumed to be a non-networked release */
	}
#endif
	if ((hp = gethostbyaddr(&iaddr, sizeof iaddr, AF_INET)) == NULL)
	{
		goto errtn;
	}

	strcpy (hn, hp->h_name);
#if B_SOCKETS|HP|UNISYS
	memcpy ((char *)&laddr, (char *)&iaddr, sizeof iaddr);
	strcpy (*addr, inet_ntoa(laddr));
#endif
#if EXCELAN && (!UNISYS)
#if X386
/*
* This must get reordered, for display purposes, because
* inet_addr has put it in network order (on a 386).
*/
	sprintf(*addr, "%d.%d.%d.%d",
		    (short)(iaddr & 0xffL),
		    (short)((iaddr >> 8) & 0xffL),
		    (short)((iaddr >> 16) & 0xffL),
		    (short)((iaddr >> 24) & 0xffL));
#else
	sprintf(*addr, "%d.%d.%d.%d",
		    (short)((iaddr >> 24) & 0xffL),
		    (short)((iaddr >> 16) & 0xffL),
		    (short)((iaddr >> 8) & 0xffL),
		    (short)(iaddr & 0xffL));
#endif
#endif
	goto goodret;
  }
#endif



errtn:			/* default return path */
	return(0);

goodret:
	return (1);
}




#if SYS&VMSSYS
getNetAddr(hname,addr)
char *hname;
char addr[];
{
	/*
	 * Host address is same as host name for DECnet.
	 */
	strcpy(addr,hname);
	return(0);
}
#else

int getNetAddr (hname, addr)
char	*hname;
char	addr[];
/*
* The address buffer indicated by the caller must be large
* enough to hold a maximum size network address.  This size will
* vary on a port by port basis.  Currently STRNORMAL should be
* large enough for existing ports.
*/
{
	*addr = '\0';

#if (SYS&AEGISO)
  {
	goto goodret;
  }
#endif

#if B_SOCKETS|EXCELAN|HP|UNISYS
/*
 * Scan through the /etc/hosts file or its equivalent
 * and return true if the supplied host name is
 * in the file. Returns false if the file was not found.
 */
  {
#if X386
	long laddr;
#else
	struct in_addr laddr;
#endif
	struct hostent *hp;

#if SYS&B4_2==0		/* may not exist if using yellow pages */
	if (Aaccess(ETCHOSTS, 04) == -1)
	{
		goto er2rtn;	/* assumed to be a non-networked release */
	}
#endif
	if ((hp = gethostbyname(hname)) == (struct hostent *)0)
	{
		goto er1rtn;
	}
#if ATT3B2
        laddr = *((struct in_addr *)hp->h_addr);
#else
	memcpy((char *)&laddr, hp->h_addr, hp->h_length);
#endif
#if B_SOCKETS|HP|UNISYS|ATT3B2
	strcpy (addr, inet_ntoa(laddr));
#endif
#if EXCELAN && (!UNISYS) && (!ATT3B2)
#if X386
/*
* This must get reordered, for display purposes, because
* inet_addr has put it in network order (on a 386).
*/
	sprintf(addr, "%d.%d.%d.%d",
		    (short)(laddr & 0xffL),
		    (short)((laddr >> 8) & 0xffL),
		    (short)((laddr >> 16) & 0xffL),
		    (short)((laddr >> 24) & 0xffL));
#else
	sprintf(addr, "%d.%d.%d.%d",
		    (short)((laddr >> 24) & 0xffL),
		    (short)((laddr >> 16) & 0xffL),
		    (short)((laddr >> 8) & 0xffL),
		    (short)(laddr & 0xffL));
#endif
#endif
	goto goodret;
  }
#endif



goodret:			/* default return path */
	return (0);

er1rtn:
	return(-1);

er2rtn:
	return(-2);
}
#endif



#if EXCELAN & !NECEWS & !UNISYS & !ATT3B2

unsigned long inet_addr (cp)
register char *cp;
/*
 * Returns an internet address given a character string in "." notation.
 * See inet_addr(3N) for details
 */
{
	register int nparts;
	unsigned long addr, val;
	unsigned long part[4];
	int		rtn;

	for (nparts=0; (rtn = gettodot(cp, &val)) == 0 && nparts < 4; )
	{
		part[nparts++] = val;
		cp = NULL;
	}
	if (rtn == -2 || rtn == 0)
		return (-1L);

	if (nparts == 1)
		addr = part[0];
	else
	{
		addr = (part[0] & 0xff) << 24;
		if (nparts == 2)
			addr |= (part[1] & 0xffffff);
		else if (nparts == 3)
			addr |= (part[1] & 0xff) << 16 | (part[2] & 0xffff);
		else if (nparts == 4)
			addr |= (part[1] & 0xff) << 16 | (part[2] &0xff) << 8
							| (part[3] & 0xff);
	}
	return(htonl(addr));
}


int	gettodot (ocp, valP)
char		*ocp;
unsigned long	*valP;
{
	register unsigned long val;
	register int base;
	register char c;
	static char *cp;

	*valP = 0L;
	if (ocp != NULL)
		cp = ocp;
	if (*cp == '\0')
		return(-1);

	base = 10;
	if (*cp == '0')
	{
		base = 8;
		cp++;
	}
	if (*cp == 'x' || *cp == 'X')
	{
		base = 16;
		cp++;
	}
	val = 0;
	while (c = *cp)
	{
		if (isdigit(c))
		{
			val = (val * base) + (c - '0');
			cp++;
		} else if (base == 16 && isxdigit(c))
		{
			val = (val << 4) + (c + 10 - (islower(c) ? 'a' : 'A'));
			cp++;
		} else if (c == '.')
		{
			cp++;
			break;
		} else
		{
			cp++;
			return (-2);
		}
	}
	*valP = val;
	return(0);
}



#define DELIM " \t"

struct hostent *gethostbyaddr(addr, len, type)
char *addr;
int len, type;
/*
 * A partial implementation -- returns only h_name
 */
{
	register char *hstr;
	unsigned long along, laddr;
	FILE *fp;
	unsigned long inet_addr();
	char hbuf[BUFSIZ];
	static struct hostent hp;
	static char hname[32];

	hp.h_name = hname;
	*hname = '\0';
	if ((fp = fopen(ETCHOSTS, "r")) == NULL)
		return((struct hostent *) NULL);
	memcpy((char *)&laddr, addr, len);
	while (fgets(hbuf, sizeof hbuf, fp))
	{
		*strchr(hbuf, '\n') = '\0';
		if (*hbuf == '#')
			continue;
		if ((hstr = strtok(hbuf, DELIM)) == (char*)0)
			continue;
		if ((along = inet_addr(hstr)) == -1)
			continue;
		if (along == laddr)
		{
			strncpy(hname, strtok((char *) NULL, DELIM),
								sizeof hname);
			break;
		}
	}
	fclose(fp);
	if (*hname != '\0')
		return(&hp);
	else
		return ((struct hostent *) NULL);
}



#define DELIM " \t"

struct hostent *gethostbyname(hname)
char *hname;
/*
 * A partial implementation -- 
 *	- returns only h_addr with h_length set to 4
 */
{
	register char *astr;
	register char *hstr;
	FILE *fp;
	unsigned long inet_addr();
	char hbuf[BUFSIZ];
	static struct hostent hp;
	static unsigned long laddr;

	hp.h_addr = (char *)&laddr;
	hp.h_length = 4;
	laddr = 0L;

	if ((fp = fopen(ETCHOSTS, "r")) == NULL)
		return((struct hostent *)0);
	while (laddr == 0 && fgets(hbuf, sizeof hbuf, fp))
	{
		/*
		 * Wipe out comments & nl
		 */
		if (*hbuf == '#')
			continue;
		if (astr = strrchr(hbuf, '#'))
			*astr = '\0';
		else
			*strchr(hbuf, '\n') = '\0';
		/*
		 * get address
		 */
		if ((astr = strtok(hbuf, DELIM)) == (char*)0)
			continue;
		/*
		 * get official names and aliases & compare
		 */
		while (laddr == 0 && (hstr = strtok((char *)NULL, DELIM))) {
			if (strcmp(hname, hstr) == 0) {
				if ((laddr = inet_addr(astr)) == -1L)
					laddr = 0L;
			}
		}
	}
	fclose(fp);

	if (laddr != 0L)
		return(&hp);
	else
		return ((struct hostent *)0);
}

#endif
