/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xllib.c
   Applix ASC version: 11
   Date: Mon Nov 16 13:58:01 1998
*/
#include <stdio.h>
#include <aos.h>
#include <afile.h>

/*	These variables are private to this file */

static char *AxLangDir;
static char *DirSlash;
static char	*axlangs[] = {    		/*      More than we need right now */
	"",					/*	Lang abbrevs and order based on proximity defines */
	NXLT("eng"),
	NXLT("grm"),
	NXLT("frn"),
	NXLT("spn"),
	NXLT("itl"),
	NXLT("brt"),
	NXLT("swd"),
	NXLT("dan"),
	NXLT("nrw"),
	NXLT("dut"),
	NXLT("prt"),
	NXLT("brz"),
	NXLT("cfr"),
	NXLT("sgr"),
	NXLT("nyn"),
	NXLT("fin"),
	NXLT("jpn"),
	NXLT("kor"),     /* Korea */
	NXLT("prc"),     /* China - mainland  (Peoples Republic of China (PRC) */
	NXLT("twn")      /* Chinese - Taiwan  */
};

#ifdef getenv
#undef getenv
#endif

/**************************************************************************/
#define TAB 9
#define MAX_KS_LINE 256       /* max char in a KS file line */
#define KS_SEP_CHAR ':' /* entry separator in KS files */
#define KS_COMMENT_CHAR ':' /* same as sep -> no name means no value */

typedef struct profile_entry {
	struct profile_entry *next;
	char *name;
	char *value;
} profile_entry;

static struct profile_entry *standalone_profile;

static int
standaloneFullTrim(dst, src)
register char *dst, *src;
{
	register char *p = dst;
 
	/* trim leading spaces */
	while ( (*src == ' ') || (*src == TAB) ) src++;
	if ( src != dst )
		strcpy(dst,src);
 
	/* trim trailing spaces */
	for ( src = dst + strlen(dst) - 1; src > dst; src-- )
		if ( (*src != ' ') && (*src != TAB) )
			break;
	*(src+1) = 0;
}

int
standaloneReadProfile(filename)
char *filename;
{
	profile_entry *profP, *profHead, *next, *newhead;
	char name[MAX_KS_LINE+2], value[MAX_KS_LINE+2], *pdst;
	char buf[MAX_KS_LINE+2], *p;
	register int len;
	FILE *fp;

	standalone_profile = profHead = (profile_entry *)0;
	fp = fopen(filename,"r");
	if ( fp == NULL )
		return(1); /* NO DATA IN PROFILE - could not open file */

	while ( fgets(buf,2000,fp) )
	{
		/* comment */
		if ( buf[0] == KS_COMMENT_CHAR )
			continue;

		/* blank line */
		if ( buf[0] == '\0' )
			continue;

		/* zero length line */
		len = strlen(buf);
		if ( len == 0 )
			continue;

		/* clip newline */
		if ( buf[len-1] == '\n' )
			buf[--len] = '\0';
             
		/* copy name of this entry up to sep char*/
		p = buf;
		pdst = name;
		while ( (*p != '\0') && (*p != KS_SEP_CHAR) )
			*pdst++ = *p++;

		/* found null in file stream */
		if ( *p != KS_SEP_CHAR )
		{
			continue;
#if 0
            fclose(fp);
			return(profHead ? 2 : 1); /* found a null in the stream */
#endif
        }    

		/* null terminate the profile entry name */
		*pdst = '\0';

		/* remove spaces between name and argument in profile */
		standaloneFullTrim(name, name);
		if ( name[0] == '\0' )
			continue;

		/* read the argument */
		p++;        /* get beyond marker */
		pdst = value;           /* copy value string */
		while ( *p != '\0')
			*pdst++ = *p++;
		*pdst = '\0';
		if ( strcmp( NXLT("axPrintShellCmd"), name ) != 0 )
		{
			/** we trim blanks on everything but the print shell command    **/
			/**  because some machines (integraph) require                  **/
			/**  a blank between the trailing "-q" and the print queue name **/
			/**  which will be concatenated on the end when we use this     **/
			standaloneFullTrim(value,value);      /* trim spaces */
		}

		/* create a new profile node */
		profP = (profile_entry *)malloc(sizeof(profile_entry));
		if ( profP == (profile_entry *)0 )
		{
			fclose(fp);
			return(profHead ? 2 : 1); /* no more memory */
		}

		/* add the name and the value */
		profP->name  = (char *)malloc(strlen(name)+1);
		if ( !profP->name )
		{
			free(profP);
			fclose(fp);
			standalone_profile = profHead;
			return(profHead ? 2 : 1); /* no more memory */
		}
		profP->value = (char *)malloc(strlen(value)+1);
		if ( !profP->value )
		{
			free(profP->name);
			free(profP);
			fclose(fp);
			standalone_profile = profHead;
			return(profHead ? 2 : 1); /* no more memory */
		}
        strcpy(profP->name,  name);
        strcpy(profP->value, value);

		/* thread to head of list */
        profP->next = profHead;
        profHead = profP;
    }
 
	/* flip the ks entries so the file's first will be the first in the list */
	newhead = (profile_entry *)0;
	for ( profP=profHead; profP; profP=next )
	{
		next = profP->next;
		profP->next = newhead;
		newhead = profP;
		if ( next )
			standalone_profile = next;
	}

	fclose(fp);
	return(AOK);
}


char *
standaloneUPbyName(name)
char *name;
{
	profile_entry *profP;

	profP = standalone_profile;

	while ( profP )
	{
		if ( 0 == strcmp(name, profP->name) )
			return(profP->value);
		profP = profP->next;
	}
	return(NULL);
}

char *
standalonegetlangdir(axdata_dir, suffix)
char *axdata_dir, *suffix;
{

	char *p, *prof_lang;
	char buf[STRPATH];
	char *AxL;					/* set by getenv ("AXLANG") */
	int AxLanguage;
#if HYUNDAI && SUN4    /* because of redeclaration of strchr */
	char *getenv();
#else
	 char *getenv(), *strchr();
#endif

 
	if (axdata_dir == NULL)
		return (NULL);
 
 
	if ( (p = strchr (axdata_dir, '\\')) != NULL)
		DirSlash = "\\";
	else
		DirSlash = "/";
 
 
	/*  Now find out what language the user wants to run in */
	AxL = getenv (NXLT("AXLANG"));
	if ( AxL )
		AxLanguage = atoi (AxL);
	else
	{
		prof_lang = (char *)standaloneUPbyName(NXLT("axDisplayLanguage"));
		if ( !prof_lang || prof_lang[0] == (char)0 )
			AxLanguage = 1;
		else
			AxLanguage = atoi(prof_lang);
	}

	if ( AxLanguage < 1 || AxLanguage > 20 )
		AxLanguage = 1;                     /*  default to English */

	if (suffix != NULL)
		sprintf(suffix,"%s",axlangs[AxLanguage]);

	sprintf(buf,"%s%s%s",axdata_dir,DirSlash,axlangs[AxLanguage]);
	AxLangDir = strcpy(malloc(strlen(buf)+1),buf);

	/* done */
	return (AxLangDir);
}

char	*
getlangdir (axdata_dir, suffix)
char	*axdata_dir, *suffix;
{
	char	*AxL,					/* set by getenv ("AXLANG") */
			*prof_lang,
			*p, 
#if HYUNDAI && SUN4    /* because of redeclaration of strchr */
         *getenv(),
#else
         *getenv(), *strchr(),
#endif
			buf[STRPATH];
	int		AxLanguage;					/* set by getenv ("AXLANG") */


	if (axdata_dir == NULL)
		return (NULL);


	if ( (p = strchr (axdata_dir, '\\')) != NULL)
		DirSlash = "\\";
	else
		DirSlash = "/";


	/*	Now find out what language the user wants to run in */

	AxL = getenv (NXLT("AXLANG"));
	if ( AxL )
		AxLanguage = atoi (AxL);
	else
	{
		prof_lang = (char *)UPbyName(NXLT("axDisplayLanguage"));
		if ( !prof_lang || prof_lang[0] == (char)0 )
			AxLanguage = 1;
		else
			AxLanguage = atoi(prof_lang);
	}
	if ( AxLanguage < 1 || AxLanguage > 20 )
		AxLanguage = 1;						/*	default to English */

	if (suffix != NULL)
		sprintf(suffix,"%s",axlangs[AxLanguage]);

	sprintf(buf,"%s%s%s",axdata_dir,DirSlash,axlangs[AxLanguage]);	
	AxLangDir = strcpy(malloc(strlen(buf)+1),buf);

	/* Verify AxLangDir directory exists */

	/* done */
	return (AxLangDir);
}


char	*
getslashchar (axdata_dir)
char	*axdata_dir;
{
	char	*p, 
#if HYUNDAI && SUN4    /* because of redeclaration of strchr */
         *getenv();
#else
         *getenv(), *strchr();
#endif

	if (axdata_dir == NULL)
		return (NULL);


	if ( (p = strchr (axdata_dir, '\\')) != NULL)
		DirSlash = "\\";
	else
		DirSlash = "/";

	return (DirSlash);
}

/*************************************************************************
 * function getLangSuffixFromPath - get suffix info from langdir
 *
 */
getLangSuffixFromPath(langdir, suffix)
char *langdir;
char *suffix;
{
	char *p, *lsuf;
	int i, found;

	found = FALSE;
	if ( langdir )
	{
#if AXMS
		lsuf = strrchr(langdir,'\\');
#else
		lsuf = strrchr(langdir,'/');
#endif
		if ( lsuf )
		{
			p = lsuf+1;
			if ( p ) for ( i=0; i<20; i++ )
			{
				if (strcmp(p,axlangs[i]) == 0 ) 
				{
					found = TRUE;
					break;
				}
			}
		}
	}
	if ( found == FALSE )
		p = NXLT("eng");
	sprintf(suffix,"%s",p);
}
