/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: cvti.c
   Applix ASC version: 539
   Date: Wed Mar 24 21:14:35 1999
*/
#include <math.h>
#include <aos.h>
#include <atmattr.h>
#include <atmgraph.h>
#include <attr.h>
#include <atm.h>
#include <getstr10.h>
#include <getstr23.h>
#include <atmwin.h>
#include <axtext.h>
#include <atxint.h>
#include <img.h>
#include <ge.h>
#include <n_style.h>
#include <bg.h>
#include <afileimg.h>
#include <cvttoks.h>
#include <cvt.h>
#include <fontmap.h>

#if FOR_FILTER
#	undef exit
#	include <stiorts.h>
#else
#	include <fsrts.h>
#endif

#define MAGIC_STRING NXLT("2278791755874641370") /* keep for old files */
#define ALLOWDECODE 1
#if 1
#include <cvti.p>
#endif

extern int (*TokFuncs[])();
extern int mailInBit6();
extern int mailInHex();
extern int cungetc();
extern int bungetc();
extern int cgetc();
extern int bgetc();
extern char *bfgets();
extern char *cfgets();
thing *mlInPart();

char *MapFontFamily();
extern int TokPointers[26];

/****************************************************************************
 *
 * module:	m l I n E r r o r - for debugging only
 */
void
mlInError(ge, cvt, err)
graphics_main_type *ge;
cvt_global_type *cvt;
int err;
{
#if FOR_FILTER
	fprintf(stderr, NXLT("Err #%d; line %d: `%s'\n"),err,cvt->lineno,cvt->buf);
	fflush(stderr);
#endif
}

/****************************************************************************
 *
 * module:	c v t S e g m e n t T e r m i n a t o r
 */
cvtSegmentTerminator(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	if ( cvt->tok.id == TCOLORMAP )
		return(TRUE);
	if ( cvt->tok.id == TFILLNAMES )
		return(TRUE);
	if ( cvt->tok.id == THDRFTR )
		return(TRUE);
	if ( cvt->info.version >= 300 && cvt->tok.id == TSESSION )
		return(TRUE);
	if ( cvt->tok.id == TCHART )
		return(TRUE);
	if ( cvt->tok.id == TPART )
		return(TRUE);
	if ( cvt->tok.id == TPICTURE )
		return(TRUE);
	return(FALSE);
}

/****************************************************************************
 *
 * module:	cvtUpdateStatus
 */
void
cvtUpdateStatus(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
#if !FOR_FILTER
	char str[256];
	int size;

	size = 100.0 * (cvt->inBufPtr - cvt->inputBuff) / cvt->inputBufferSize;
	size = size / 5 * 5;
	sprintf(str, XLT("?tReading Data: %d%%"), size);
	textMsg(ge, str);
#endif
}

/****************************************************************************
 *
 * module:	cvtTokToInt
 */
cvtTokToInt(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int val;
	if ( cvt->tok.type != TOK_NUM || cvt->tok.n_vals <= 0 )
		val = 0;
	else
	{
		if ( cvt->tok.val[0].type == 0 )
			val = cvt->tok.val[0].u.ival;
		else if ( cvt->tok.val[0].type == 1 )
			val = cvt->tok.val[0].u.fval;
		else
			val = 0;
	}
	return(val);
}

/****************************************************************************
 *
 * module:	cvtTokToFlt
 */
double
cvtTokToFlt(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	double val;
	if ( cvt->tok.type != TOK_NUM || cvt->tok.n_vals <= 0 )
		val = 0.0;
	else
	{
		if ( cvt->tok.val[0].type == 0 )
			val = cvt->tok.val[0].u.ival;
		else if ( cvt->tok.val[0].type == 1 )
			val = cvt->tok.val[0].u.fval;
		else
			val = 0.0;
	}
	return(val);
}

/****************************************************************************
 *
 * module:	cvtGetStrFromTokVal
 */
char *
cvtGetStrFromTokVal(ge, cvt, ix)
graphics_main_type *ge;
cvt_global_type *cvt;
int ix;
{
	char *str;
	int size;

	if ( cvt->tok.n_vals <= 0 || cvt->tok.val[ix].type != 2 )
		str = (char *)0;
	else
	{
#if !BUILD_ORIENT
		size = (cvt->tok.val[ix].u.cval.n_chars+1) * sizeof(char);
#else
        /* Ashisuto edited for wide chars.  95.10.16 */
        size = (cvt->tok.val[ix].u.cval.n_chars+1) *
                sizeof(*(cvt->tok.val[ix].u.cval.str));
#endif
		if ( size == 0 )
			str = (char *)0;
		else
		{
			str = (char *)gloc(ge, size);
			WideToMultiStr(ge, cvt->tok.val[ix].u.cval.str, str);
		}
	}
	return(str);
}

/****************************************************************************
 *
 * module:	cvtTokToStr
 */
char *
cvtTokToStr(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	return(cvtGetStrFromTokVal(ge, cvt, 0));
}

/****************************************************************************
 *
 * module:	cvtClearTokVals
 */
void
cvtClearTokVals(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int i;

	/* clear the token argument list */
	if ( cvt->tok.n_vals > 0 && cvt->tok.val )
	{
		if ( cvt->tok.val )
		{
			for ( i=0; i<cvt->tok.n_vals; i++ )
			{
				if ( cvt->tok.val[i].type == 2 && cvt->tok.val[i].u.cval.str )
					gfree(ge, cvt->tok.val[i].u.cval.str);
			}
			gfree(ge, cvt->tok.val);
		}
		cvt->tok.n_vals = 0;
		cvt->tok.val = (cvt_token_val_type *)0;
	}
}

void
cvtTokToBuf(ge, cvt, offset)
graphics_main_type *ge;
cvt_global_type *cvt;
int offset;
{
	char *buf;
	int ch;

	buf = &cvt->buf[offset];
	ch = (*cvt->getcFunc)(ge, cvt);
	while ( isalnum(ch) || ch == '.' || ch == '-' || ch == '*' || ch == '_' )
	{
		if (islower(ch))
			ch = toupper(ch);
		else if (ch == '-')
			ch = '_';
		*buf++ = ch;
		ch = (*cvt->getcFunc)(ge, cvt);
	}
	(*cvt->ungetcFunc)(ge, cvt, ch);
	*buf = '\0';
}

void
cvtReadNum(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int ch, force_neg, is_flt, ival, eval, size, n_i, n_f, i, is_exp;
	char *buf, *ibuf, *fbuf, ebuf[4];
	double fval;
	int is_neg;
	
	/* collect a (possibly signed) number (assume tok.type == TOK_NUM) */
	buf    = cvt->buf;
	ibuf   = (char *)0;
	fbuf   = (char *)0;
	n_i    = n_f = 0;
	is_exp = FALSE;
	is_neg = FALSE;
	/* get the first character (it could be a sign) */
	ch = (*cvt->getcFunc)(ge, cvt);
	force_neg = (ch == '-') ? TRUE : FALSE;
	if ( ch != '+' )
	{
		*buf++ = ch;
		n_i++;
	}

	/* get the rest of the characters */
	ch = (*cvt->getcFunc)(ge, cvt);
	while (isdigit(ch))
	{
		*buf++ = ch;
		n_i++;

		ch = (*cvt->getcFunc)(ge, cvt);
	}
	*buf = '\0';
	ibuf = (char *)gloc(ge, n_i+1);
	strcpy(ibuf, cvt->buf);
	ival = atoi(ibuf);

	/* if the string ending in a decimal, read the post-decimal chars */
	if ( ch != '.' )
	{
		(*cvt->ungetcFunc)(ge, cvt, ch);
		is_flt = FALSE;
	}
	else
	{
		buf = cvt->buf;
		ch = (*cvt->getcFunc)(ge, cvt);
		while (isdigit(ch))
		{
			*buf++ = ch;
			n_f++;

			ch = (*cvt->getcFunc)(ge, cvt);
		}
		if ( ch == 101 )
		{
			/* Must be an exponent */
			is_exp = TRUE;
			ch = (*cvt->getcFunc)(ge, cvt);
			if ( !isdigit(ch) )
				is_neg = (ch == '-') ? TRUE : FALSE;
			
			ch = (*cvt->getcFunc)(ge, cvt);
			if ( isdigit(ch) ) 
			{
				ebuf[0] = ch;
				ch      = (*cvt->getcFunc)(ge, cvt);
				ebuf[1] = ch;
				if ( (ch = (*cvt->getcFunc)(ge, cvt)) != 0x20 && isdigit(ch) )
					ebuf[2] = ch;
				else
					ebuf[2] = '\0';
				ebuf[3] = '\0';
			}
			else 
			{	
				(*cvt->ungetcFunc)(ge, cvt, ch);
				is_exp = FALSE;
			}
			if ( is_exp )
			{
				eval = atoi(ebuf);
				eval = is_neg ? (-1)*eval: eval;
			}
			
		}
		*buf = '\0';
		(*cvt->ungetcFunc)(ge, cvt, ch);
		if ( n_f > 0 )
		{
			fbuf = (char *)gloc(ge, n_f+1);
			strcpy(fbuf, cvt->buf);
		}
		else
		{
			n_f = 1;
			fbuf = (char *)gloc(ge, n_f+1);
			strcpy(fbuf, "0");
		}

		/* format a string */
		if ( !force_neg || ival != 0 )
			sprintf(cvt->buf,"%s.%s00000", ibuf, fbuf);
		else
		{
			for ( i=0; i<n_i; i++ )
			{
				if ( ibuf[i] != '0' )
					break;
			}
			if ( i < n_i )
				sprintf(cvt->buf,"%s.%s00000", &ibuf[i], fbuf);
			else
				sprintf(cvt->buf,"-0.%s00000", fbuf);
		}

		/* scanf the string into a float */
		if ( sscanf(cvt->buf, "%lf", &fval) != 1 )
			fval = 0.0;
		is_flt = TRUE;
	}
	gfree(ge, ibuf);
	if ( fbuf )
		gfree(ge, fbuf);

	/* update the token arguments */
	cvt->tok.n_vals++;
	size = cvt->tok.n_vals * sizeof(cvt_token_val_type);
	cvt->tok.val = (cvt_token_val_type *)regloc(ge, cvt->tok.val, size);
	if ( is_flt )
	{
		if ( is_exp )
		{
			fval = fval * pow(10, eval); 
		}
		cvt->tok.val[cvt->tok.n_vals-1].u.fval = fval;
		cvt->tok.val[cvt->tok.n_vals-1].type = 1;
	}
	else
	{
		cvt->tok.val[cvt->tok.n_vals-1].u.ival = ival;
		cvt->tok.val[cvt->tok.n_vals-1].type = 0;
	}
}

/**************************************************************************
 * Function: c v t R e a d N a N
 */
void
cvtReadNaN(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int ch, ch1, ch2, sz, add_tok=FALSE;
	
	ch = (*cvt->getcFunc)(ge, cvt);
	if ( ch == 'N' )
	{
		ch1 = (*cvt->getcFunc)(ge, cvt);
		ch2 = (*cvt->getcFunc)(ge, cvt);
		if ( ch1 == 'a' && ch2 == 'N' )
			add_tok = TRUE;
	}
	
	/* update the token arguments, by adding a 0 to handle NaN */
	if ( add_tok == TRUE )
	{	
		cvt->tok.n_vals++;
		sz = cvt->tok.n_vals * sizeof(cvt_token_val_type);
		cvt->tok.val = (cvt_token_val_type *)regloc(ge, cvt->tok.val, sz);
		cvt->tok.val[cvt->tok.n_vals-1].u.ival = 0;
		cvt->tok.val[cvt->tok.n_vals-1].type = 0;
	}
}

/****************************************************************************
 *
 * module:	NormalizeIn2_1
 */
NormalizeIn2_1(ge,cvt,inStr,len,outStr)
graphics_main_type *ge;
cvt_global_type *cvt;
mbchar *inStr;
int len;
mbchar *outStr;
{
	int  d1, d2, size, l;
	mbchar *inP, *outP, ch;

	inP = inStr;
	outP = outStr;
	for (l=0; l<len; l++)
	{
		ch = *inP++;
		if (ch == '^')
		{
			d1 = *inP++;
			d1 -= d1 <= '9' ? '0' : '7';
			d2 = *inP++;
			d2 -= d2 <= '9' ? '0' : '7';
			ch = (d1 << 4) + d2;
		}
		else if (ch == '\\')
			ch = *inP++;

		*outP++ = ch;
	}
	*outP = '\0';
	return(MBstrlen(outStr));
}

/****************************************************************************
 *
 * module:	CvtInSpecialChars
 */
CvtInSpecialChars(ge, wstr,numcharsIn)
graphics_main_type *ge;
tx_char *wstr;
int numcharsIn;
{
	tx_char *inP, *outP;
	int ii, numcharsOut;

	numcharsOut = numcharsIn;
	inP = outP = wstr;
	for (ii=0; ii<numcharsIn; ii++)
	{
		if (*inP == '\\')
		{
			*outP = *++inP;
			if (*outP == 'n')
				*outP = '\n';
			else if (*outP == 't')
				*outP = '\t';
			numcharsOut--;
		}
		else
			*outP = *inP;

		inP++;outP++;
	}
	wstr[numcharsOut] = 0;
	return(numcharsOut);
}

void
cvtReadStr(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int ok2get, ignore, len, size, ch, n_chars;
	mbchar *newstr;
	tx_char *wstr;
	char *buf;

	/* toss the opening quotes, and begin parsing string */
	ch = (*cvt->getcFunc)(ge, cvt);
	ok2get = cvt->n_bufblks * CVTBUFLEN - 1;
	ignore = FALSE;
	buf = cvt->buf;
	while ( TRUE )
	{
		/* increase buffer size as needed */
		if ( (buf - cvt->buf) >= ok2get )
		{
			len = buf - cvt->buf;
			CvtBiggerBuffer(ge, cvt, 1);
			buf = cvt->buf + len;
			ok2get += CVTBUFLEN;
		}

		/* get the next character */
		ch = (*cvt->getcFunc)(ge, cvt);
		if ( ignore )
		{
			*buf++ = ch;
			ignore = FALSE;
			continue;
		}

		/* special character treatment */
		if ( ch == '\n' )
		{
			/* toss a space after a newline */
			cvt->lineno++;
			ch = (*cvt->getcFunc)(ge, cvt);
			if ( ch != ' ' )
				*buf++ = ch;
		}
		else if ( ch == '"' )
			break;
		else if ( ch == '\\' )
		{
			*buf++ = ch;
			ignore = TRUE;
		}
		else
			*buf++ = ch;
	}
	*buf = '\0';

	/* convert the buffer from 8 bits to 16 bits (per character) */
	len = buf - cvt->buf;
	size = (len+1) * sizeof(tx_char);
	wstr = (tx_char *)gloc(ge, size);
	if ( cvt->info.version < 300 )
	{
		newstr = (mbchar *)gloc(ge, size);
		NormalizeIn2_1(ge, cvt, (mbchar *)cvt->buf, len, newstr);
		n_chars = CvtMbToWStr(ge, (unsigned char *)newstr, (unsigned char *)wstr);
		gfree(ge, newstr);
	}
	else
	{
		n_chars = NormalizeIn16((mbchar *)cvt->buf, wstr, len+1, &cvt->info);
		n_chars = CvtInSpecialChars(ge, wstr, n_chars);
	}

	/* update the token arguments */
	cvt->tok.n_vals++;
	size = cvt->tok.n_vals * sizeof(cvt_token_val_type);
	cvt->tok.val = (cvt_token_val_type *)regloc(ge, cvt->tok.val, size);
	cvt->tok.val[cvt->tok.n_vals-1].type = 2;
	cvt->tok.val[cvt->tok.n_vals-1].u.cval.n_chars = n_chars;
	cvt->tok.val[cvt->tok.n_vals-1].u.cval.str = wstr;
}

/****************************************************************************
 *
 * module:	cvtObjTokSearch
 */
void
cvtObjTokSearch(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	register int i;

	/* new format - if it begins with a dot, must be an object */
	if ( cvt->buf[0] == '.' )
	{
		cvt->tok.type = TOK_OTY;
		for ( i=FIRST_OBJ; i<=LAST_OBJ; i++ )
		{
			if ( STREQ(cvt->buf, pnm2[i]) )
			{
				cvt->tok.id = i;
				return;
			}
		}
		cvt->tok.id = -1; /* unknown object */
	}

	/* old format */
	else for ( i=FIRST_OBJ; i<=LAST_OBJ; i++ )
	{
		if ( STREQ(cvt->buf, pnm[i]) )
		{
			cvt->tok.type = TOK_OTY;
			cvt->tok.id = i;
			return;
		}
	}
}
void
cvtReadObj(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	/* load the buffer */
	cvtTokToBuf(ge, cvt, 0);

	/* parse the object lists */
	cvt->tok.id = 0;
	cvtObjTokSearch(ge, cvt);
}

/****************************************************************************
 *
 * module:	c v t T o k S e a r c h - search the token array for key
 */
cvtTokTokSearch(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int curr;
	char *key;

	/* Search tokens with the same first letter as the key */
	key = cvt->buf;
	curr = TokPointers[cvt->buf[0] - 'A'];

	if (curr == (-1)) /* no token starts with this letter */
		return(INITIAL);

	/* At least 1 token begins with the same letter. */
	/* 2nd letter match and linear search from here. */

	while ((*key == *(tok[curr])) && (curr <= TLAST_TOK))
	{
		if ((tok[curr][1] == key[1]) && (!strcmp(key,tok[curr]))) /* found it */
			return(curr);
		curr++;
	}

	return(INITIAL);
}

/****************************************************************************
 *
 * module:	cvtAuxTokSearch
 */
void
cvtAuxTokSearch(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	register int i;
	for ( i=0; i<LSTS; i++ )
	{
		if ( STREQ(cvt->buf, lst[i]) )
		{
			cvt->tok.type = TOK_AUX;
			cvt->tok.id = i;
			return;
		}
	}
	for ( i=0; i<AROWS; i++ )
	{
		if ( STREQ(cvt->buf, aro[i]) )
		{
			cvt->tok.type = TOK_AUX;
			cvt->tok.id = i;
			return;
		}
	}
	for ( i=0; i<AROTS; i++ )
	{
		if ( STREQ(cvt->buf, art[i]) )
		{
			cvt->tok.type = TOK_AUX;
			cvt->tok.id = i;
			return;
		}
	}
	for ( i=0; i<LTHS; i++ )
	{
		if ( STREQ(cvt->buf, lth[i]) )
		{
			cvt->tok.type = TOK_AUX;
			cvt->tok.id = i;
			return;
		}
	}
	for ( i=0; i<FNTS1; i++ )
	{
		if ( STREQ(cvt->buf, fnt1[i]) )
		{
			cvt->tok.type = TOK_AUX;
			cvt->tok.id = i;
			return;
		}
	}
	for ( i=0; i<FNTS2; i++ )
	{
		if ( STREQ(cvt->buf, fnt2[i]) )
		{
			cvt->tok.type = TOK_AUX;
			cvt->tok.id = i;
			return;
		}
	}
	for ( i=0; i<SIZS; i++ )
	{
		if ( STREQ(cvt->buf, siz[i]) )
		{
			cvt->tok.type = TOK_AUX;
			cvt->tok.id = i;
			return;
		}
	}
	for ( i=0; i<SIZS1; i++ )
	{
		if ( STREQ(cvt->buf, siz1[i]) )
		{
			cvt->tok.type = TOK_AUX;
			cvt->tok.id = i;
			break;
		}
	}
	if ( cvt->tok.type == TOK_AUX )
	{
		switch(cvt->tok.id)
		{
			default:
			case 0: cvt->tok.id = 5; return;
			case 1: cvt->tok.id = 1; return;
			case 2: cvt->tok.id = 3; return;
			case 3: cvt->tok.id = 7; return;
		}
	}
	for ( i=0; i<JSTS; i++ )
	{
		if ( STREQ(cvt->buf, jst[i]) )
		{
			cvt->tok.type = TOK_AUX;
			cvt->tok.id = i;
			return;
		}
	}
	for ( i=0; i<JSTS; i++ )
	{
		if ( STREQ(cvt->buf, vjst[i]) )
		{
			cvt->tok.type = TOK_AUX;
			cvt->tok.id = i;
			return;
		}
	}
	for ( i=0; i<UNDS; i++ )
	{
		if ( STREQ(cvt->buf, und[i]) )
		{
			cvt->tok.type = TOK_AUX;
			cvt->tok.id = i;
			return;
		}
	}
	for ( i=0; i<CMPS; i++ )
	{
		if ( STREQ(cvt->buf, cmp[i]) )
		{
			cvt->tok.type = TOK_AUX;
			cvt->tok.id = i;
			return;
		}
	}
	for ( i=0; i<ENCS; i++ )
	{
		if ( STREQ(cvt->buf, enc[i]) )
		{
			cvt->tok.type = TOK_AUX;
			cvt->tok.id = i;
			return;
		}
	}
}

void
cvtReadTok(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int i, tok_id;

	/* read the token */
	cvtTokToBuf(ge, cvt, 0);

	/* is it OFF or ON ? */
	if ( STREQ(cvt->buf, tst[0]) )
	{
		cvt->tok.type = TOK_TST;
		cvt->tok.n_vals = 0;
		cvt->tok.id = 0;
	}
	else if ( STREQ(cvt->buf, tst[1]) )
	{
		cvt->tok.type = TOK_TST;
		cvt->tok.n_vals = 0;
		cvt->tok.id = 1;
	}

	/* is it FALSE or TRUE */
	else if ( STREQ(cvt->buf, sts[0]) )
	{
		cvt->tok.type = TOK_STS;
		cvt->tok.n_vals = 0;
		cvt->tok.id = 0;
	}
	else if ( STREQ(cvt->buf, sts[1]) )
	{
		cvt->tok.type = TOK_STS;
		cvt->tok.n_vals = 0;
		cvt->tok.id = 1;
	}

	/* look for other types of tokens */
	else
	{
		/* reinit */
		cvt->tok.type = TOK_UNK;
		cvt->tok.n_vals = 0;
		cvt->tok.id = 0;

		/* simple linear lookup search for object type */
		cvtObjTokSearch(ge, cvt);
		if ( cvt->tok.type == TOK_UNK )
		{
			/* Not object type, so try keyword/argument pair */
			tok_id = cvtTokTokSearch(ge, cvt);
			if ( tok_id == INITIAL )
				cvtAuxTokSearch(ge, cvt); /* try the aux lists */
			else
			{
				/* found the keyword... */
				cvt->tok.type = TOK_TOK;
				cvt->tok.id = tok_id;
			}
		}
	}
}

void
cvtReadGrp(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int ch;

	/* groups can contain strings and numbers ONLY */
	ch = (*cvt->getcFunc)(ge, cvt);
	while ( ch != '>' )
	{
		/* Is it a newline? If so, advance the counter and continue */
		if ( ch == '\n' )
			cvt->lineno++;

		/* Is it a number or a sign? If so, must be a number */
		else if (ch == '-' || ch == '+' || isdigit(ch))
		{
			(*cvt->ungetcFunc)(ge, cvt, ch);
			cvtReadNum(ge, cvt);
		}

		/* Is it a quote? If so, must be a string */
		else if ( ch == '"' )
		{
			(*cvt->ungetcFunc)(ge, cvt, ch);
			cvtReadStr(ge, cvt);
		}
		/* what if it's not a number or a string - CRASH? */
		else if ( ch == 'N' )
		{
			(*cvt->ungetcFunc)(ge, cvt, ch);
			cvtReadNaN(ge, cvt);
		}


		/* next character */
		ch = (*cvt->getcFunc)(ge, cvt);
	}
}

void
cvtReadToken(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	/* clear the token argument list */
	cvtClearTokVals(ge, cvt);

	/* parse the token */
	switch ( cvt->tok.type )
	{
		case TOK_OTY: cvtReadObj(ge, cvt); break;
		case TOK_NUM: cvtReadNum(ge, cvt); break;
		case TOK_STR: cvtReadStr(ge, cvt); break;
		case TOK_TOK: cvtReadTok(ge, cvt); break;
		case TOK_GRP: cvtReadGrp(ge, cvt); break;
	}
}


void
cvtReadComment(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int slen, len, size;

	(*cvt->fgetsFunc)(ge, cvt, cvt->buf, 256);
	slen = strlen(cvt->buf);
	if ( cvt->buf[slen-1] == '\n' )
	{
		slen--;
		cvt->buf[slen] = 0;
	}
	cvt->lineno++;
	len = cvt->n_comments++;
	size = cvt->n_comments * sizeof(char*);
	cvt->comments = (char **)regloc(ge, cvt->comments, size);
	cvt->comments[len] = strcpy((char*)gloc(ge, slen+1), cvt->buf);
}

void
cvtReadEmbedded(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int done, level, n, size, sz1, sz2, sz3;
	ge_embed_type embed;
	char buf[AREADBUFSIZE];

	/* init */
	sz1 = strlen(ATOK_START);
	sz2 = strlen(ATOK_OLDSTART);
	sz3 = strlen(ATOK_END);
	memset((char *)&embed, 0, sizeof(ge_embed_type));
	embed.size = strlen(cvt->buf) + 1;
	embed.data = (char *)gloc(ge, embed.size);
	strcpy(embed.data, cvt->buf);

	/* read the embedded material */
	done = FALSE;
	level = 0;
	while ( !done )
	{
		/* read a line */
		(*cvt->fgetsFunc)(ge, cvt, buf, AREADBUFSIZE);

		/* add the line to the buffer */
		n = strlen(buf);
		embed.data = (char *)regloc(ge, embed.data, (embed.size+n));
		/*	strcat(embed.data, buf);	*/
		sprintf(&embed.data[embed.size-1], "%s", buf);
		embed.size += n;

		/* if it's a *START, increment the level */
		if ( buf[0] == '*' && strncmp(buf, ATOK_START, sz1) == 0 )
			level++;
		else if ( buf[0] == '*' && strncmp(buf, ATOK_OLDSTART, sz2) == 0 )
			level++;

		/* else if it's an *END */
		else if ( buf[0] == '*' && strncmp(buf, ATOK_END, sz3) == 0 )
		{
			if ( level == 0 )
				done = TRUE;
			else
				level--;
		}
	}
	embed.data[embed.size-1]= '\0';

	/* expand the embed list */
	size = sizeof(ge_embed_type) * (ge->n_embeds+1);
	ge->embed = (ge_embed_type *)regloc(ge, ge->embed, size);
	ge->embed[ge->n_embeds] = embed;
	ge->n_embeds++;
}

void
cvtReadStandard(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int size, len, n;
	mbchar *p, *q;
	AFileInfo info;

	/* reinsert the leading '*' */
	cvt->buf[0] = '*';
	cvtTokToBuf(ge, cvt, 1);

	/* ** */
	if ( cvt->buf[1] == '*' )
	{
		len = ge->n_cmts++;
		size = ge->n_cmts * sizeof(ge_vars_type);
		ge->cmt = (ge_vars_type *)regloc(ge, ge->cmt, size);

		/* get the variable string */
		mlInTok(ge, cvt, FALSE);
		ge->cmt[len].var = (char *)cvtTokToStr(ge, cvt);

		/* the rest is the value */
		(*cvt->fgetsFunc)(ge, cvt, cvt->buf, 256);
		n = strlen(cvt->buf);
		ge->cmt[len].val = (char *)gloc(ge, n+1);
		strcpy(ge->cmt[len].val, cvt->buf+1);
		ge->cmt[len].val[n-2] = (char)0; /* step on the newline */
	}

	/* *START or *BEGIN */
	else if ( (strcmp(cvt->buf, ATOK_START) == 0) ||
	          (strcmp(cvt->buf, ATOK_OLDSTART) == 0) )
	{
		/* get the rest of the line */
		len = strlen(cvt->buf);
		(*cvt->fgetsFunc)(ge, cvt, cvt->buf+len, 256);

		/* *START GRAPHICS (or *START ? ? ?) */
		AFileInfoFromHeader(cvt->buf, &info);
		if ( cvt->vers_found || info.docType != AX_DOC_TYPE_ASTERX_GRAPHICS )
			cvtReadEmbedded(ge, cvt);
		else
		{
			cvt->info     = info;
			cvt->majorVer = cvt->info.version / 100;
			cvt->minorVer = cvt->info.version % 100;
			cvt->vers_found = TRUE;
		}
	}

	/* *END */
	else if ( strcmp(cvt->buf, ATOK_END) == 0 )
	{
		/* MUST be end GRAPHICS */
		if ( mlInTok(ge, cvt, FALSE) == TOK_TOK && cvt->tok.id == TGRAPHICS )
			cvt->done = TRUE;

		/* An end with no start! Bail out. */
		else
			longjmp(cvt->jmpPoint, SGE_NOFILE);
	}

	/* *LINK */
	else if ( strcmp(cvt->buf, ATOK_LINK) == 0 )
	{
		/* increase the size of the link list */
		size = sizeof(ge_link_type);
		ge->link = (ge_link_type *)regloc(ge, ge->link, size*(ge->n_links+1));
		memset((char *)&ge->link[ge->n_links], 0, size);

		/* copy the path into the cvt->buf */
		(*cvt->fgetsFunc)(ge, cvt, cvt->buf, 256);
		n = strlen(cvt->buf);

		/* copy the path from the cvt->buf to the link structure */
		ge->link[ge->n_links].path = (char *)gloc(ge, n+1);
		strcpy(ge->link[ge->n_links].path, cvt->buf+1);
		ge->link[ge->n_links].path[n-2] = (char)0; /* step on the newline */
		ge->n_links++;
	}

	/* *HOOK */
	else if ( strcmp(cvt->buf, ATOK_HOOK) == 0 )
	{
		/* increase the size of the hook list */
		size = sizeof(ge_hook_type) * (ge->n_hooks+1);
		ge->hook = (ge_hook_type *)regloc(ge, ge->hook, size);

		/* copy the line into the cvt->buf (space hook_type space hook_name) */
		(*cvt->fgetsFunc)(ge, cvt, cvt->buf, 256);
		p = cvt->buf;
		while (*p && (*p == ' '))
			p++;
		if (p[0] == 0)
			return; /* garbage */
		q = p;
		while (*p && (*p != ' '))
			p++;
		if (p[0] == 0)
			return; /* garbage */
		len = p - q;
		ge->hook[ge->n_hooks].type = (char*)gloc(ge, len+1);
		strncpy(ge->hook[ge->n_hooks].type,q,len);
		ge->hook[ge->n_hooks].type[len] = 0;
		while (*p && (*p == ' '))
			p++;
		if (p[0] == 0)
			return; /* garbage */
		q = p;
		while (*p && (*p != ' '))
			p++;
		len = p-q;
		if (q[len-1] == '\n')
		{
			q[len-1] = 0;
			len--;
		}
		ge->hook[ge->n_hooks].name = (char*)gloc(ge, len+1);
		strcpy(ge->hook[ge->n_hooks].name,q);
		ge->n_hooks++;
	}
	else /* just dump the rest of the line */
		(*cvt->fgetsFunc)(ge, cvt, cvt->buf, 256);


	/* set the token type */
	if ( cvt->done )
		cvt->tok.type = TOK_STD;
	else
		cvt->tok.type = TOK_UNK;
}

/****************************************************************************
 *
 * module: c v t P a r s e T o k
 */
void
cvtParseTok(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int ch;

	/* read a token from the input file */
	cvt->tok.type = TOK_UNK;
	while ( cvt->tok.type == TOK_UNK )
	{
		/* read a single 'character' */
		ch = (*cvt->getcFunc)(ge, cvt);

		/* Is it a comment? If so, read the comment and continue */
		if ( ch == COMMENT )
			cvtReadComment(ge, cvt);

		/* Is it an asterix? If so, read the meta-token and continue */
		else if ( ch == '*' )
			cvtReadStandard(ge, cvt);

		/* Is it a newline? If so, advance the counter and continue */
		else if ( ch == '\n' )
			cvt->lineno++;

		/* Is it an opening angle bracket? If so, must be a group of args */
		else if ( ch == '<' )
			cvt->tok.type = TOK_GRP;

		/* Is it a dot? If so, must be an object type token */
		else if ( ch == '.' )
			cvt->tok.type = TOK_OTY;

		/* Is it a letter? If so, must be a token */
		else if ( isalpha(ch) )
			cvt->tok.type = TOK_TOK;

		/* Is it a number or a sign? If so, must be a number */
		else if (ch == '-' || ch == '+' || isdigit(ch))
			cvt->tok.type = TOK_NUM;

		/* Is it a quote? If so, must be a string */
		else if ( ch == '"' )
			cvt->tok.type = TOK_STR;
	}

	/* return the identifying character to the buffer */
	(*cvt->ungetcFunc)(ge, cvt, ch);

	/* parse the token (and it's args, if any) */
	cvtReadToken(ge, cvt);
}

/****************************************************************************
 *
 * module:	m l I n T o k
 */
mlInTok(ge, cvt, handle_errs)
graphics_main_type *ge;
cvt_global_type *cvt;
int handle_errs;
{
	/* if untok is true, reuse the last token */
	if ( cvt->untok )
		cvt->untok = FALSE;

	else /* read a new token from the file (or buffer) */
		cvtParseTok(ge, cvt);

	/* if the caller doesn't want unknown tokens, keep looking */
	if ( handle_errs && cvt->tok.type == TOK_UNK )
	{
		/* if the token is not recognized, toss it's argument... */
		cvtParseTok(ge, cvt);

		/* ...and read the next token */
		mlInTok(ge, cvt, TRUE);
	}
	return(cvt->tok.type);
}

/****************************************************************************
 *
 * module:	ReadTABSTOPS
 */
void
ReadTABSTOPS(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	int i, sz;
	thing *txt_t;

	mlInTok(ge, cvt, FALSE);
	if ( t->type == TEXTBOX )
		txt_t = t;
	else if ( cvt->active_t )
		txt_t = cvt->active_t;
	else
		txt_t = (thing *)0;
	if ( txt_t )
	{
		txt_t->data.txtbox->def_tabsize  = cvt->tok.val[0].u.ival;
		txt_t->data.txtbox->n_tabs       = cvt->tok.val[1].u.ival;
		if ( txt_t->data.txtbox->n_tabs > 0 )
		{
			sz = txt_t->data.txtbox->n_tabs*sizeof(int);
			txt_t->data.txtbox->tabstops = (int *)gloc(ge, sz);
			for ( i=0; i< cvt->tok.n_vals-2; i++ )
				txt_t->data.txtbox->tabstops[i] = cvt->tok.val[i+2].u.ival;
		}
	}
}

/****************************************************************************
 *
 * module:	ReadCAT
 */
void
ReadCAT(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	int x, y;

	mlInCoord(ge, cvt, &x, &y, TRUE);
	cvt->dx = x - PAGE_ZEROX;
	cvt->dy = y - PAGE_ZEROY;
	if ( cvt->active_t )
	{
		cvt->active_t->extent.x1 = cvt->active_t->extent.x2 = x;
		cvt->active_t->extent.y1 = cvt->active_t->extent.y2 = y;
	}
}

/****************************************************************************
 *
 * module:	ReadCANGLE
 */
void
ReadCANGLE(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->active_t->attr.angle = cvtTokToInt(ge, cvt);
	cvt->curr_atts.angle = cvt->active_t->attr.angle;
}

/****************************************************************************
 *
 * module:	ReadCSCALE
 */
void
ReadCSCALE(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->active_t->attr.x_scale = cvtTokToInt(ge, cvt);
	if ( t->type == RECT ) {
		if ( t->data.rect->txt_t->attr.x_scale == 0 )
			t->data.rect->txt_t->attr.x_scale = 1;
	}
	else if ( t->type == ELL ) {
		if ( t->data.ell->txt_t->attr.x_scale == 0 )
			t->data.ell->txt_t->attr.x_scale = 1;		
	}
	cvt->curr_atts.x_scale = cvt->active_t->attr.x_scale;
}

/****************************************************************************
 *
 * module:	ReadCXYOFFS
 */
void
ReadCXYOFFS(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->active_t->attr.x_off = cvt->tok.val[0].u.ival;
	cvt->curr_atts.x_off = cvt->active_t->attr.x_off;
}

/****************************************************************************
 *
 * module:	ReadCTXTXYOFF
 */
void
ReadCTXTXYOFF(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->active_t->data.txtbox->x_off = cvt->tok.val[0].u.ival;
	cvt->active_t->data.txtbox->y_off = cvt->tok.val[1].u.ival;
}

/****************************************************************************
 *
 * module:	ReadCL_SPACE
 */
void
ReadCL_SPACE(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->active_t->attr.field_space = cvtTokToInt(ge, cvt);
	cvt->curr_atts.field_space = cvt->active_t->attr.field_space;
}

/****************************************************************************
 *
 * module:	ReadCMARGINS
 */
void
ReadCMARGINS(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->active_t->attr.l_margin = cvt->tok.val[0].u.ival;
	cvt->active_t->attr.r_margin = cvt->tok.val[1].u.ival;
	cvt->active_t->attr.t_margin = cvt->tok.val[2].u.ival;
	cvt->active_t->attr.b_margin = cvt->tok.val[3].u.ival;
	cvt->curr_atts.l_margin = cvt->active_t->attr.l_margin;
	cvt->curr_atts.r_margin = cvt->active_t->attr.r_margin;
	cvt->curr_atts.t_margin = cvt->active_t->attr.t_margin;
	cvt->curr_atts.b_margin = cvt->active_t->attr.b_margin;
}

/****************************************************************************
 *
 * module:	ReadCSUBSUP
 */
void
ReadCSUBSUP(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->active_t->attr.y_sub_sup = cvt->tok.val[1].u.ival;
	cvt->curr_atts.y_sub_sup = cvt->active_t->attr.y_sub_sup;
}

/****************************************************************************
 *
 * module:	ReadCSHADOW
 */
void
ReadCSHADOW(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->active_t->attr.shadow_color = cvt->tok.val[0].u.ival;
	cvt->active_t->attr.shadow       = cvt->tok.val[1].u.ival;
	cvt->active_t->attr.shadow_x     = cvt->tok.val[2].u.ival;
	cvt->active_t->attr.shadow_y     = cvt->tok.val[3].u.ival;

	cvt->curr_atts.shadow_color      = cvt->active_t->attr.shadow_color;
	cvt->curr_atts.shadow            = cvt->active_t->attr.shadow;
	cvt->curr_atts.shadow_x          = cvt->active_t->attr.shadow_x;
	cvt->curr_atts.shadow_y          = cvt->active_t->attr.shadow_y;
}

/****************************************************************************
 *
 * module:	ReadCFONT
 */
void
ReadCFONT(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	int face_ix;

	mlInTok(ge, cvt, FALSE);
	if ( cvt->tok.type == TOK_NUM )
	{
		if ( cvt->tok.val[0].type == 0 )
			face_ix = cvt->tok.val[0].u.ival;
		else if ( cvt->tok.val[0].type == 1 )
			face_ix = cvt->tok.val[0].u.fval;
		else
			face_ix = CvtAddFont(ge,cvt,fnt[0]);
	}
	else
	{
		if ( cvt->tok.type == TOK_AUX )
			face_ix = CvtAddFont(ge,cvt,fnt[cvt->tok.id]);

		else /* fake it */
		{
			face_ix = CvtAddFont(ge,cvt,fnt[0]);
			mlInError(ge, cvt, ML_UNKTAV);
		}
	}
	if ( face_ix >= ge->n_faces )
		cvt->active_t->attr.face = 0;
	else
		cvt->active_t->attr.face = face_ix;
	cvt->curr_atts.face = cvt->active_t->attr.face;
}

/****************************************************************************
 *
 * module:	ReadCSHEAR
 */
void
ReadCSHEAR(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->active_t->attr.shear = cvtTokToInt(ge, cvt);
	cvt->curr_atts.shear = cvt->active_t->attr.shear;
}

/****************************************************************************
 *
 * module:	ReadCSIZE
 */
void
ReadCSIZE(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	int size;

	mlInTok(ge, cvt, FALSE);
	if ( cvt->tok.type == TOK_NUM )
	{
		if ( cvt->tok.val[0].type == 0 )
			size = cvt->tok.val[0].u.ival;
		else if ( cvt->tok.val[0].type == 1 )
			size = cvt->tok.val[0].u.fval;
		else
			size = 180000;
	}
	else
	{
		if ( cvt->tok.type == TOK_AUX )
			size = siz2val1[cvt->tok.id];
		else
		{
			size = 180000;
			mlInError(ge, cvt, ML_UNKTAV);
		}
	}

	if ( size <= 0 )
		size = 180000;
	cvt->active_t->attr.size = size;
	cvt->curr_atts.size = cvt->active_t->attr.size;
}

/****************************************************************************
 *
 * module:	ReadCTHRU
 */
void
ReadCTHRU(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( cvt->tok.type == TOK_TST )
		cvt->active_t->attr.strike = cvt->tok.id;
	else if ( cvt->tok.type == TOK_NUM )
		cvt->active_t->attr.strike = cvtTokToInt(ge, cvt);
	else
		mlInError(ge, cvt, ML_TAVEXP);
	cvt->curr_atts.strike = cvt->active_t->attr.strike;
}

/****************************************************************************
 *
 * module:	ReadCSHR-ANG
 */
void
ReadCSHR_ANG(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->active_t->attr.field_shear = cvt->tok.val[0].u.ival;
	cvt->active_t->attr.field_angle = cvt->tok.val[1].u.ival;
	cvt->curr_atts.field_shear = cvt->active_t->attr.field_shear;
	cvt->curr_atts.field_angle = cvt->active_t->attr.field_angle;
}

/****************************************************************************
 *
 * module:	ReadCTHICKNESS
 */
void
ReadCTHICKNESS(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( cvt->tok.type == TOK_NUM )
		cvt->active_t->attr.line_thickness = cvtTokToInt(ge, cvt);
	else  if ( cvt->tok.type != TOK_AUX )
		mlInError(ge, cvt, ML_PAVEXP);
	else
	{
		if ( cvt->tok.id == LTHS )
			cvt->active_t->attr.line_thickness = 1;
		else
			cvt->active_t->attr.line_thickness = lth2val[cvt->tok.id];
	}
	cvt->curr_atts.line_thickness = cvt->active_t->attr.line_thickness;
}

/****************************************************************************
 *
 * module:	ReadCFLD_XYSCALE
 */
void
ReadCFLD_XYSCALE(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
    int f_x_scale, f_y_scale;

	mlInTok(ge, cvt, FALSE);
	f_x_scale = cvt->tok.val[0].u.ival;
	f_y_scale = cvt->tok.val[1].u.ival;

	if ( f_x_scale <= -MAXINT/2 )
		f_x_scale = 10000;
	else if ( f_x_scale >= MAXINT/2 )
		f_x_scale = 10000;
	if ( f_y_scale <= -MAXINT/2 )
		f_y_scale = 10000;
	else if ( f_y_scale >= MAXINT/2 )
		f_y_scale = 10000;
	if ( f_x_scale == 0 )
		f_x_scale = 10000;
	if ( f_y_scale == 0 )
		f_y_scale = 10000;
	cvt->active_t->attr.field_x_scale = f_x_scale;
	cvt->active_t->attr.field_y_scale = f_y_scale;
	cvt->curr_atts.field_x_scale = cvt->active_t->attr.field_x_scale;
	cvt->curr_atts.field_y_scale = cvt->active_t->attr.field_y_scale;
}

/****************************************************************************
 *
 * module:	ReadCH_SPACE
 */
void
ReadCH_SPACE(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->active_t->data.txtbox->field_wid  = cvt->tok.val[0].u.ival;
}

/****************************************************************************
 *
 * module:	ReadCSTYLE
 */
void
ReadCSTYLE(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( cvt->tok.type == TOK_NUM )
		cvt->active_t->attr.line_style = cvtTokToInt(ge, cvt);
	else if ( cvt->tok.type == TOK_AUX )
		cvt->active_t->attr.line_style = cvt->tok.id;
	else
		mlInError(ge, cvt, ML_PAVEXP);
	cvt->curr_atts.line_style = cvt->active_t->attr.line_style;
}

/****************************************************************************
 *
 * module:	ReadCV_SPACE
 */
void
ReadCV_SPACE(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
cvt_part_creation_type *obj;
thing *t;
{
	mlInTok(ge, cvt, FALSE);
	cvt->active_t->attr.space_above  = cvt->tok.val[0].u.ival;
	cvt->active_t->attr.space_inside = cvt->tok.val[1].u.ival;
	cvt->active_t->attr.space_below  = cvt->tok.val[2].u.ival;
	cvt->curr_atts.space_above = cvt->active_t->attr.space_above;
	cvt->curr_atts.space_inside = cvt->active_t->attr.space_inside;
	cvt->curr_atts.space_below = cvt->active_t->attr.space_below;
}

/****************************************************************************
 *
 * module:	R e a d CP A R A
 */
void
ReadCPARA(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
cvt_part_creation_type *obj;
thing *t;
{
	mlInTok(ge, cvt, FALSE);
	cvt->active_t->attr.l_indent1    = cvt->tok.val[0].u.ival;
	cvt->active_t->attr.l_indent2    = cvt->tok.val[1].u.ival;
	cvt->active_t->attr.r_indent     = cvt->tok.val[2].u.ival;
	cvt->active_t->attr.bullet       = cvt->tok.val[3].u.ival;
	cvt->active_t->attr.bullet_scale = cvt->tok.val[4].u.ival;
	cvt->active_t->attr.bullet_color = cvt->tok.val[5].u.ival;
	cvt->active_t->attr.bullet_font  = cvt->tok.val[6].u.ival;
	cvt->active_t->attr.bullet_glyph = cvt->tok.val[7].u.ival;
	cvt->curr_atts.l_indent1 = cvt->active_t->attr.l_indent1;
	cvt->curr_atts.l_indent2 = cvt->active_t->attr.l_indent2;
	cvt->curr_atts.r_indent = cvt->active_t->attr.r_indent;
	cvt->curr_atts.bullet = cvt->active_t->attr.bullet;
	cvt->curr_atts.bullet_scale = cvt->active_t->attr.bullet_scale;
	cvt->curr_atts.bullet_color = cvt->active_t->attr.bullet_color;
	cvt->curr_atts.bullet_font = cvt->active_t->attr.bullet_font;
	cvt->curr_atts.bullet_glyph = cvt->active_t->attr.bullet_glyph;
}

/****************************************************************************
 *
 * module:	ReadCHOR_JUST
 */
void
ReadCHOR_JUST(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	
	if ( t->type == ELL ) 
	{
		ReadHORIZONTAL_JUSTIFICATION(ge, cvt, t->data.ell->txt_t, obj);
		t->data.ell->txt_t->attr.h_just = cvt->curr_atts.h_just;
	}
	else 
	{
		ReadHORIZONTAL_JUSTIFICATION(ge, cvt, t->data.rect->txt_t, obj);
		t->data.rect->txt_t->attr.h_just = cvt->curr_atts.h_just;
	}
}

/****************************************************************************
 *
 * module:	ReadCVER_JUST
 */
void
ReadCVER_JUST(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	if ( t->type == ELL ) 
	{
		ReadVERTICAL_JUSTIFICATION(ge, cvt, t->data.ell->txt_t, obj);
		t->data.ell->txt_t->attr.h_just = cvt->curr_atts.v_just;
	}
	else 
	{
		ReadVERTICAL_JUSTIFICATION(ge, cvt, t->data.rect->txt_t, obj);
		t->data.rect->txt_t->attr.h_just = cvt->curr_atts.v_just;
	}
}

/****************************************************************************
 *
 * module:	ReadCUNDERLINE
 */
void
ReadCUNDERLINE(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	int ruling;

	mlInTok(ge, cvt, FALSE);
	if ( cvt->tok.type == TOK_TST )
		ruling = cvt->tok.id;
	else if ( cvt->tok.type == TOK_NUM )
		ruling = cvtTokToInt(ge, cvt);
	else if ( cvt->tok.type == TOK_AUX )
		ruling = cvt->tok.id;
	else
	{
		mlInError(ge, cvt, ML_UNKTAV);
		ruling = 0;
	}
	if ( ruling & 0x4 )
		cvt->active_t->attr.strike = TRUE;
	cvt->active_t->attr.ruling = ruling;
	cvt->curr_atts.ruling = cvt->active_t->attr.ruling;
}

/****************************************************************************
 *
 * module:	ReadCBOLD
 */
void
ReadCBOLD(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( cvt->tok.type == TOK_TST )
		cvt->active_t->attr.weight = cvt->tok.id;
	else if ( cvt->tok.type == TOK_NUM )
		cvt->active_t->attr.weight = cvtTokToInt(ge, cvt);
	else
		mlInError(ge, cvt, ML_TAVEXP);
	cvt->curr_atts.weight = cvt->active_t->attr.weight;
}

/****************************************************************************
 *
 * module:	ReadCITALICS
 */
void
ReadCITALICS(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( cvt->tok.type == TOK_TST )
		cvt->active_t->attr.slant = cvt->tok.id;
	else if ( cvt->tok.type == TOK_NUM )
		cvt->active_t->attr.slant = cvtTokToInt(ge, cvt);
	else
		mlInError(ge, cvt, ML_TAVEXP);
	cvt->curr_atts.slant = cvt->active_t->attr.slant;
}

void
ReadCBACKFILL(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	/* fg, bg, type, id, angle, xoffset, yoffset */
	mlInTok(ge, cvt, FALSE);
	cvt->active_t->attr.fill_fg   = cvt->tok.val[0].u.ival;
	cvt->active_t->attr.fill_bg   = cvt->tok.val[1].u.ival;
	cvt->active_t->attr.fpat_type = cvt->tok.val[2].u.ival;
	cvt->active_t->attr.fill_pat  = cvt->tok.val[3].u.ival;
	cvt->curr_atts.fill_fg = cvt->active_t->attr.fill_fg;
	cvt->curr_atts.fill_bg = cvt->active_t->attr.fill_bg;
	cvt->curr_atts.fpat_type = cvt->active_t->attr.fpat_type;
	cvt->curr_atts.fill_pat = cvt->active_t->attr.fill_pat;
}

void
ReadCLINEFILL(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	/* fg, bg, type, id, angle, xoffset, yoffset */
	mlInTok(ge, cvt, FALSE);
	cvt->active_t->attr.line_fg   = cvt->tok.val[0].u.ival;
	cvt->active_t->attr.line_bg   = cvt->tok.val[1].u.ival;
	cvt->active_t->attr.lpat_type = cvt->tok.val[2].u.ival;
	cvt->active_t->attr.line_pat  = cvt->tok.val[3].u.ival;
	cvt->curr_atts.line_fg = cvt->active_t->attr.line_fg;
	cvt->curr_atts.line_bg = cvt->active_t->attr.line_bg;
	cvt->curr_atts.lpat_type = cvt->active_t->attr.lpat_type;
	cvt->curr_atts.line_pat = cvt->active_t->attr.line_pat;
}

/****************************************************************************
 *
 * module:	m l I n C h e c k P a g e 
 */
void
mlInCheckPage(ge, cvt, page)
graphics_main_type *ge;
cvt_global_type *cvt;
ge_page_type *page;
{
	char *def_name, str2[STRPATH+1];
	int len, sz, i, n_colors;

	/* check audio info */
	if ( cvt->audio_found )
		cvt->audio_found = FALSE;
	else
	{
		cvt->audio.play = FALSE;
		cvt->audio.type = INSET_NONE;
		cvt->audio.id = 0;
		cvt->audio.volume = 50;
	}
	page->audio = cvt->audio;

	if ( cvt->hdrftr_found )
	{
		cvt->hdrftr_found = FALSE;
		sz = sizeof(ge_hdrftr_page_type);
		page->hf_page = (ge_hdrftr_page_type *)gloc(ge, sz);
		memcpy(page->hf_page, &cvt->hf_main.base, sz);
		memset(&cvt->hf_main, 0, sizeof(ge_hdrftr_main_type));
	}

	/* check slide info */
	if ( cvt->slide_found == FALSE )
	{
		page->n_ochars = 0;
		page->outline_text = (tx_char *)0;

		page->n_ochars = 0;

		cvt->slide_effect = ge->state.default_slide_effect;
		cvt->slide_delay = ge->state.default_slide_delay;
		cvt->slide_throttle = ge->state.default_slide_throttle;
	}
	else
	{
		page->n_ochars = cvt->n_outline_chars;
		page->outline_text = cvt->outline_text;

		page->layout_ix = cvt->layout_ix;
		page->object1 = cvt->object1;
		page->object2 = cvt->object2;
		page->object3 = cvt->object3;
		page->object4 = cvt->object4;
		page->hide_objs = cvt->hide_objs;
		page->hide_pg_backgd = cvt->hide_pg_backgd;
		page->use_backgd = cvt->use_backgd;
		page->use_color = cvt->use_color;
		page->hide_slide = cvt->hide_slide;
		page->c_scheme = cvt->c_scheme;


		n_colors = ge->cmap[0].n_colors;
		if ( page->c_scheme.title_fg >= n_colors )
			page->c_scheme.title_fg = 0;
		if ( page->c_scheme.title_bg >= n_colors )
			page->c_scheme.title_bg = 0;
		if ( page->c_scheme.text_fg >= n_colors )
			page->c_scheme.text_fg = 0;
		if ( page->c_scheme.line_fg >= n_colors )
			page->c_scheme.line_fg = 0;
		if ( page->c_scheme.line_bg >= n_colors )
			page->c_scheme.line_bg = 0;
		if ( page->c_scheme.fill_fg >= n_colors )
			page->c_scheme.fill_fg = 0;
		if ( page->c_scheme.fill_bg >= n_colors )
			page->c_scheme.fill_bg = 0;
		if ( page->c_scheme.page_fg >= n_colors )
			page->c_scheme.page_fg = 0;
		if ( page->c_scheme.page_bg >= n_colors )
			page->c_scheme.page_bg = 0;
		if ( page->c_scheme.shadow_color >= n_colors )
			page->c_scheme.shadow_color = 0;
		for ( i=0; i<16; i++ )
		{
			if ( page->c_scheme.accent_color[i] >= n_colors )
				page->c_scheme.shadow_color = 0;
		}
		cvt->outline_text = (tx_char *)0;
		cvt->n_outline_chars = 0;
		cvt->notes_text = (tx_char *)0;
		cvt->n_notes_chars = 0;
		cvt->c_scheme.n_other_colors = 0;
		cvt->c_scheme.other_color = (int *)0;

		cvt->hide_objs = FALSE;
		cvt->hide_pg_backgd = FALSE;
		cvt->use_color = FALSE;
		cvt->use_backgd = FALSE;
		cvt->slide_found = FALSE;
		cvt->hide_slide = FALSE;
	}

	/* check slide build info */
	if ( cvt->slide_build_found == FALSE )
	{
		page->use_slide_build = FALSE;
		memcpy(&page->build, &ge->state.default_build,
							 sizeof(ge_build_info_type));
	}
	else
	{
		page->use_slide_build = cvt->use_slide_build;
		memcpy(&page->build, &cvt->build, sizeof(ge_build_info_type));
		cvt->slide_build_found = FALSE;
	}

	page->slide_effect = cvt->slide_effect;
	page->slide_delay  = cvt->slide_delay;
	page->slide_throttle  = cvt->slide_throttle;

	/* check layers */
	if ( cvt->layers_found == FALSE )
	{
		def_name = XLT("Default");
		len = strlen(def_name);
		page->n_layers = 1;
		sz = sizeof(layer_type);
		page->layer = (layer_type *)gloc(ge, sz);
		memset(page->layer, 0x0, sz);
		sz = sizeof(tx_char);
		page->layer[0].name = (tx_char *)gloc(ge, (len+1)* sz);
		memset(page->layer[0].name, 0x0, (len+1)* sz);
#if BUILD_ORIENT
        page->layer[0].n_chars = CvtMbToWStr(ge, def_name, page->layer[0].name);
#else
		for ( i=0; i<len; i++ )
			page->layer[0].name[i] = (tx_char)def_name[i]; 
		page->layer[0].name[i] = (tx_char)0;
		page->layer[0].n_chars = len;
#endif
		page->layer[0].locked = FALSE;
		page->layer[0].hidden = FALSE;
		page->layer[0].prints = TRUE;
		page->layer[0].backgd = FALSE;
		page->layer_ix = 0;
	}
	else
	{
		page->n_layers = cvt->n_layers;
		page->layer = cvt->layer;
		cvt->n_layers = 0;
		cvt->layer = (layer_type *)0;
		cvt->layers_found = FALSE;
		page->layer_ix = 0;
	}
}

/****************************************************************************
 *
 * module:	m l I n C o o r d - collect an (x,y) position
 */
mlInCoord(ge, cvt, x, y, same_disabled)
graphics_main_type *ge;
cvt_global_type *cvt;
int *x, *y;
int same_disabled; /* FALSE: allow SAME; TRUE: no SAME, error; UNKNOWN: untok */
{
	int bad_x, bad_y;

	/* read the x value */
	mlInTok(ge, cvt, FALSE);
	if ( cvt->tok.type == TOK_TOK && cvt->tok.id == TSAME )
	{
		if ( same_disabled == FALSE )
		{
			same_disabled = TRUE;
			*x = SAME;
			bad_x = FALSE;
		}
		else
		{
			if ( same_disabled == UNKNOWN )
				cvt->untok = TRUE;
			bad_x = TRUE;
		}
	}
	else /* must be a number */
	{
		*x = PAGE_ZEROX + cvt->dx;
		if ( cvt->tok.type == TOK_NUM )
		{
			*x += cvtTokToInt(ge, cvt);
			bad_x = FALSE;
		}
		else
		{
			cvt->untok = TRUE;
			bad_x = TRUE;
		}
	}

	/* read the y value */
	if ( bad_x )
		bad_y = TRUE;
	else
	{
		mlInTok(ge, cvt, FALSE);
		if ( cvt->tok.type == TOK_TOK && cvt->tok.id == TSAME )
		{
			if ( same_disabled == FALSE )
			{
				*y = SAME;
				bad_y = FALSE;
			}
			else
			{
				if ( same_disabled == UNKNOWN )
					cvt->untok = TRUE;
				bad_y = TRUE;
			}
		}
		else /* must be a number */
		{
			*y = PAGE_ZEROY + cvt->dy;
			if ( cvt->tok.type == TOK_NUM )
			{
				*y += cvtTokToInt(ge, cvt);
				bad_y = FALSE;
			}
			else
			{
				cvt->untok = TRUE;
				bad_y = TRUE;
			}
		}
	}

	return((bad_x||bad_y) ? FALSE : TRUE);
}

/****************************************************************************
 *
 * module:	m l I n C t r l s
 */
mlInCtls(ge,cvt,pathP)
graphics_main_type *ge;
cvt_global_type *cvt;
poly_type *pathP;
{
	int x, y, n_pts, size;

	n_pts = 0;
	while ( mlInCoord(ge, cvt, &x, &y, UNKNOWN) )
	{
		size = (n_pts + 1) * sizeof(TmPoint32);
		pathP->controls = (TmPoint32 *)regloc(ge, pathP->controls, size);
		pathP->controls[n_pts].x = x;
		pathP->controls[n_pts].y = y;
		n_pts++;
	}
	pathP->nCtrls = n_pts;
	if ( pathP->nCtrls > (2*pathP->nPoints-2) )
		cvt->point_damage = TRUE;
	cvt->untok = TRUE;
}

/****************************************************************************
 *
 * module:	m l I n P o i n t s
 */
mlInPnts(ge,cvt,pathP,allowZeroLength)
graphics_main_type *ge;
cvt_global_type *cvt;
poly_type *pathP;
int allowZeroLength;
{
	int x, y, last_x, last_y, n_pts, size;

	if ( mlInCoord(ge, cvt, &x, &y, UNKNOWN) )
	{
		pathP->points = (TmPoint32*)gloc(ge, sizeof(TmPoint32));
		pathP->points[0].x = x;
		pathP->points[0].y = y;
		n_pts = 1;

		last_x = x;
		last_y = y;
		while ( mlInCoord(ge, cvt, &x, &y, UNKNOWN) )
		{
			if ( !allowZeroLength && last_x == x && last_y == y )
				cvt->point_damage = TRUE;

			size = (n_pts + 1) * sizeof(TmPoint32);
			pathP->points = (TmPoint32 *)regloc(ge, pathP->points, size);
			pathP->points[n_pts].x = x;
			pathP->points[n_pts].y = y;
			last_x = x;
			last_y = y;
			n_pts++;
		}
		pathP->nPoints = n_pts;
	}
	cvt->untok = TRUE;
}

/****************************************************************************
 *
 * module:	mlFillPolyType - fills a transform or path
 */
mlFillPolyType(ge, cvt, pathP, allowZeroLength)
graphics_main_type *ge;
cvt_global_type *cvt;
poly_type *pathP;
int allowZeroLength;
{
	int got_pnts, got_ctls;

	/* we've got points and/or controls coming, at most 1 of each */
	got_pnts = got_ctls = FALSE;
	mlInTok(ge, cvt, FALSE);
	switch ( cvt->tok.id ) /* assume TOK_TOK */
	{
		default:
			cvt->untok = TRUE;
		break;
		case TPNTS:
		case TT_PNTS:
		case TPOINTS:
			mlInPnts(ge, cvt, pathP, allowZeroLength);
			got_pnts = TRUE;
		break;
		case TCTLS:
		case TT_CTLS:
		case TCONTROL_PTS:
			mlInCtls(ge, cvt, pathP);
			got_ctls = TRUE;
		break;
	}

	/* get the other one, if it's there */
	mlInTok(ge, cvt, FALSE);
	switch ( cvt->tok.id ) /* assume TOK_TOK */
	{
		default:
			cvt->untok = TRUE;
		break;
		case TPNTS:
		case TT_PNTS:
		case TPOINTS:
			if ( got_pnts )
				cvt->untok = TRUE;
			else
				mlInPnts(ge, cvt, pathP, allowZeroLength);
		break;
		case TCTLS:
		case TT_CTLS:
		case TCONTROL_PTS:
			if ( got_ctls )
				cvt->untok = TRUE;
			else
				mlInCtls(ge, cvt, pathP);
		break;
	}
}

/****************************************************************************
 *
 * module:	mlInPath - creates and fills a path
 */
mlInPath(ge, cvt, pathP, nPaths, allowZeroLength)
graphics_main_type *ge;
cvt_global_type *cvt;
poly_type **pathP;
int *nPaths;
int allowZeroLength;
{
	poly_type *newP;
	int n_paths;

	/* add another path onto the list */
	n_paths = *nPaths;
	*nPaths += 1;
	*pathP = (poly_type*)regloc(ge, *pathP, (*nPaths)*sizeof(poly_type));

	/* initialize the new path */
	newP = *pathP + n_paths;
	newP->nPoints = 0;
	newP->nCtrls = 0;
	newP->points = (TmPoint32 *)0;
	newP->controls = (TmPoint32 *)0;

	/* read the path dataonto the new path */
	mlFillPolyType(ge, cvt, newP, allowZeroLength);

	if ( newP->nPoints <= 1 )
	{
		*nPaths -= 1;
		*pathP = (poly_type*)regloc(ge, *pathP, (*nPaths)*sizeof(poly_type));
	}
}

/****************************************************************************
 *
 * module:	mlInSetAttrs
 */
mlInSetAttrs(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	int fgTrans, bgTrans, i, n_colors;
	layer_type *layer;

	if (cvt->majorVer == GVER1_MAJOR)
	{
		/* set the line attrs to the fill attrs */
		cvt->curr_atts.line_bg = cvt->curr_atts.fill_bg;
		cvt->curr_atts.line_fg = cvt->curr_atts.fill_fg;
	}

	/* set the generic object attributes */
	t->attr = cvt->curr_atts;
	t->layer = cvt->curr_layer;

	if (cvt->curr_atts.use.tag == TRUE)
		cvt->curr_atts.use.tag = FALSE; /* off by default */

	if (cvt->trans )
	{
		/* "TRANSPARENT" tok found, set the
		 * background colors to transparent
		 */
		SetToTransparent(ge,cvt,t,TBACK_COLOR);
		SetToTransparent(ge,cvt,t,TLINE_BACK_COLOR);
	}

	/* set any object specific fields now */
	if (t->type == ELL )
	{
		t->data.ell->start_angle = cvt->startAngle;
		t->data.ell->end_angle   = cvt->endAngle;
		t->data.ell->n_sides     = cvt->nsides;
		for ( i=0; i<6; i++ )
		{
			t->data.ell->sshape_pnt[i].x = cvt->sshape_pnt[i].x;
			t->data.ell->sshape_pnt[i].y = cvt->sshape_pnt[i].y;
		}
	}
	else if (t->type == TEXTBOX )
	{
		t->data.txtbox->x_off = cvt->x_off;
		t->data.txtbox->y_off = cvt->y_off;

		t->data.txtbox->field_wid = cvt->field_wid;
		t->data.txtbox->v_align = cvt->v_align;
		t->data.txtbox->sub_txt = cvt->sub_txt;
								  
		if (cvt->majorVer == GVER1_MAJOR)
		{
			/* make text as solid fill */
			t->attr.fpat_type = GRAY_FILL_TYPE;
			t->attr.fill_pat = 1000;
			for (i=0; i<t->data.txtbox->n_strings; i++)
			{
				t->data.txtbox->strings[i]->attr.fpat_type = GRAY_FILL_TYPE;
				t->data.txtbox->strings[i]->attr.fill_pat = 1000;
			}
		}

	}
	else if (t->type == RECT )
	{
		t->data.rect->x_corner_radius    = cvt->x_corner_radius;
		t->data.rect->y_corner_radius    = cvt->y_corner_radius;
		t->data.rect->scale_corners      = cvt->scale_corners;
		t->data.rect->x_corner_numerator = cvt->x_corner_numerator;
		t->data.rect->y_corner_numerator = cvt->y_corner_numerator;
		t->data.rect->sshape_type        = cvt->sshape_type;
		for ( i=0; i<6; i++ )
		{
			t->data.rect->sshape_pnt[i].x = cvt->sshape_pnt[i].x;
			t->data.rect->sshape_pnt[i].y = cvt->sshape_pnt[i].y;
		}
	}
	else if (t->type == LINE )
	{
		t->data.line->sshape_type   = cvt->lshape_type;
		t->data.line->end1_cpnt     = cvt->end1_cpnt;
		t->data.line->end2_cpnt     = cvt->end2_cpnt;
		t->data.line->end1_cpnt_dir = cvt->end1_cpnt_dir;
		t->data.line->end2_cpnt_dir = cvt->end2_cpnt_dir;
		for ( i=0; i<3; i++ )
		{
			t->data.line->sshape_pnt[i].x = cvt->lshape_pnt[i].x;
			t->data.line->sshape_pnt[i].y = cvt->lshape_pnt[i].y;
		}
	}
	else if ( t->type == IMAGE )
	{
		t->attr.fill_pat    = 0;
		t->attr.fpat_type   = BUILTIN_FILL_TYPE;
		t->data.img->cmapID = cvt->cmapID;
	}
	else if ( t->type == TEMPLATE )
	{
		t->data.temp->type     = cvt->temp_type;
		t->data.temp->size     = cvt->temp_size;
		t->data.temp->inactive = cvt->temp_inactive;
		t->data.temp->ix       = cvt->temp_chart_ix;
		t->data.temp->area     = cvt->temp_area;
		if ( t->data.temp->area.x1 >= t->data.temp->area.x2 )
			t->data.temp->area.x2 = t->data.temp->area.x1 + 1;
		if ( t->data.temp->area.y1 >= t->data.temp->area.y2 )
			t->data.temp->area.y2 = t->data.temp->area.y1 + 1;
	}

	if (cvt->curr_atts.callback)
	{
		t->attr.use.callback = TRUE;
		cvt->curr_atts.callback = 0;
	}

	/* put other color checks as well */
	n_colors = ge->cmap[0].n_colors;
	if ( t->attr.fill_fg >= n_colors )
		t->attr.fill_fg = 0;
	if ( t->attr.fill_bg >= n_colors )
		t->attr.fill_bg = 0;
	if ( t->attr.line_fg >= n_colors )
		t->attr.line_fg = 0;
	if ( t->attr.line_bg >= n_colors )
		t->attr.line_bg = 0;
	if ( t->attr.shadow_color >= n_colors )
		t->attr.shadow_color = 0;
	if ( t->attr.bullet_color >= n_colors )
		t->attr.bullet_color = 0;

	fgTrans = ge->cmap[0].color[t->attr.line_fg].see_thru;
	bgTrans = ge->cmap[0].color[t->attr.line_bg].see_thru;
	if (fgTrans && bgTrans)
		t->attr.mode = TMD;
	else if (fgTrans)
		t->attr.mode = TMS_O_D;
	else if (bgTrans)
		t->attr.mode = TMIS_O_D;
	else
		t->attr.mode = TMS;

	if ( cvt->info.version < 320 )
	{
		layer = &ge->page[ge->page_ix].layer[t->layer];
		if ( layer && (layer->locked||layer->hidden) && t->flags.hidden )
			t->flags.hidden = FALSE;
	}
}

/****************************************************************************
 *
 * module:	m l I n P a t t e r n
 */
mlInPattern(ge, atr, cvt)
graphics_main_type *ge;
int atr;
cvt_global_type *cvt;
{
	int *patt, *type, new_patt;
	
	if ( atr == TPATTERN )
	{
		patt = &cvt->curr_atts.fill_pat;
		type = &cvt->curr_atts.fpat_type;
	}
	else /* TLINE_PATTERN */
	{
		patt = &cvt->curr_atts.line_pat;
		type = &cvt->curr_atts.lpat_type;
	}
		
	mlInTok(ge, cvt, FALSE);
	switch ( cvt->tok.type )
	{
		case TOK_NUM:
			if ( cvt->info.version < 300 )
			{
				*patt = cvtTokToInt(ge, cvt) + 1;
				*type = BUILTIN_FILL_TYPE;
			}
			else
			{
				*patt = cvtTokToInt(ge, cvt);
				mlInTok(ge, cvt, FALSE);
				*type = cvtTokToInt(ge, cvt);
				mlInTok(ge, cvt, FALSE);
				mlInTok(ge, cvt, FALSE);
			}
		break;
		case TOK_TST: /* pre rev 3 - PATTERN OFF */
			*patt = 0;
			*type = BUILTIN_FILL_TYPE ;
		break;
		default:
			mlInError(ge, cvt, ML_NUMNEXP);
		break;
	}

	/* old builtin's that emulate greyscales should use greyscales */
	if ( *type == BUILTIN_FILL_TYPE )
	{
		new_patt = CvtGetPercentFill(ge, *patt);
		if ( new_patt != (*patt) )
		{
			*type = GRAY_FILL_TYPE;
			*patt = new_patt;
		}
	}
}

/****************************************************************************
 *
 * module:	m l I n N a m e - read in a palet part name (obsolete)
 */
mlInName(ge,cvt,t)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
{
	mlInTok(ge, cvt, FALSE);
}

/****************************************************************************
 *
 * module:	mlCopyTags
 */
mlCopyTags(ge,cvt,t)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
{
	int ii,ntags;
	ge_tag_type *tag;

	/* Comments must occur before the */
	/* object they are associated with. */

	if (cvt->n_comments)
	{
		if (cvt->tags_found == FALSE)
			cvt->tags_found = TRUE;
		ntags = ge->n_tags++;
		ge->tag = (ge_tag_type*)regloc(ge, ge->tag,ge->n_tags*sizeof(ge_tag_type));
		tag = &ge->tag[ntags];
		tag->n_strs = cvt->n_comments;
		tag->str = (char**)gloc(ge, tag->n_strs*sizeof(char*));
		for (ii=0; ii<tag->n_strs; ii++)
		{
			tag->str[ii] = strcpy((char*)gloc(ge, strlen(cvt->comments[ii])+1),
							cvt->comments[ii]);
		}
		for (ii=0; ii<cvt->n_comments; ii++)
			gfree(ge, cvt->comments[ii]);
		gfree(ge, cvt->comments);
		cvt->comments = NULL;
		cvt->n_comments = 0;
		cvt->curr_atts.tag = ntags;
		cvt->curr_atts.use.tag = TRUE;
	}
}

/****************************************************************************
 *
 * module:	mlTrans5ToTrans7
 */
mlTrans5ToTrans7(ge, cvt, t)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
{
#if !FOR_FILTER

	axtext txt;
    thing *txt_t;
    tx_char *txP;
	b32 bbox;
	p32 baseLineOffset;
	textAttrs tp;
	unsigned int tflags, ii;
	int xLen, yLen;
	TmPoint32 minn, maxx;

	memset(&txt,0,sizeof(axtext));
	bbox.p0.x = t->data.txtbox->trans.points[0].x - PAGE_ZEROX;
	bbox.p0.y = t->data.txtbox->trans.points[0].y - PAGE_ZEROY;
	bbox.p1.x = t->data.txtbox->trans.points[1].x - PAGE_ZEROX;
	bbox.p1.y = t->data.txtbox->trans.points[1].y - PAGE_ZEROY;
	bbox.p2.x = t->data.txtbox->trans.points[2].x - PAGE_ZEROX;
	bbox.p2.y = t->data.txtbox->trans.points[2].y - PAGE_ZEROY;
	bbox.p3.x = t->data.txtbox->trans.points[3].x - PAGE_ZEROX;
	bbox.p3.y = t->data.txtbox->trans.points[3].y - PAGE_ZEROY;

	/* set some attrs */
	memset(&tp,0,sizeof(textAttrs));
	txt_t = t->data.txtbox->strings[0];
	tp.numChars = txt_t->data.str->n_chars;
	tp.fontFamily = ge->face[txt_t->attr.face];
	tp.weight = txt_t->attr.weight;
	tp.slant =  txt_t->attr.slant;
	tp.pointsize = txt_t->attr.size / 10000;
	tp.ruling = txt_t->attr.ruling;
	if ( txt_t->attr.strike )
		tp.ruling |= 0x4;
	tflags = TXfontFamily|TXweight|TXslant|TXpointsize|TXruling;

	getXformFromRev2BBox(&txt,txt_t->data.str->chars,txt_t->data.str->n_chars,
		tp.pointsize, &bbox, &baseLineOffset, &tp, &tflags,0,0);

	t->data.txtbox->trans.points =
		(TmPoint32*)regloc(ge, t->data.txtbox->trans.points,7*sizeof(TmPoint32));
	t->data.txtbox->trans.nPoints = 7;

	/* the baseLineOffset is wrt to the minn coord, */
	/* so make the bbox also wrt to the minn coord. */

	/* first, calculate the minn and maxx coord */

	minn.x = maxx.x = bbox.p0.x;
	minn.y = maxx.y = bbox.p0.y;
	if (bbox.p1.x < minn.x) minn.x = bbox.p1.x;
	if (bbox.p1.y < minn.y) minn.y = bbox.p1.y;
	if (bbox.p2.x < minn.x) minn.x = bbox.p2.x;
	if (bbox.p2.y < minn.y) minn.y = bbox.p2.y;
	if (bbox.p3.x < minn.x) minn.x = bbox.p3.x;
	if (bbox.p3.y < minn.y) minn.y = bbox.p3.y;
	if (bbox.p1.x > maxx.x) maxx.x = bbox.p1.x;
	if (bbox.p1.y > maxx.y) maxx.y = bbox.p1.y;
	if (bbox.p2.x > maxx.x) maxx.x = bbox.p2.x;
	if (bbox.p2.y > maxx.y) maxx.y = bbox.p2.y;
	if (bbox.p3.x > maxx.x) maxx.x = bbox.p3.x;
	if (bbox.p3.y > maxx.y) maxx.y = bbox.p3.y;

	/* reset the box */
	bbox.p0.x -= minn.x;
	bbox.p0.y -= minn.y;
	bbox.p1.x -= minn.x;
	bbox.p1.y -= minn.y;
	bbox.p2.x -= minn.x;
	bbox.p2.y -= minn.y;
	bbox.p3.x -= minn.x;
	bbox.p3.y -= minn.y;
	maxx.x -= minn.x;
	maxx.y -= minn.y;

	t->data.txtbox->trans.points[1].x = bbox.p0.x + t->extent.x1;
	t->data.txtbox->trans.points[1].y = bbox.p0.y + t->extent.y1;
	t->data.txtbox->trans.points[2].x = bbox.p1.x + t->extent.x1;
	t->data.txtbox->trans.points[2].y = bbox.p1.y + t->extent.y1;
	t->data.txtbox->trans.points[3].x = bbox.p2.x + t->extent.x1;
	t->data.txtbox->trans.points[3].y = bbox.p2.y + t->extent.y1;
	t->data.txtbox->trans.points[4].x = bbox.p3.x + t->extent.x1;
	t->data.txtbox->trans.points[4].y = bbox.p3.y + t->extent.y1;

	t->data.txtbox->trans.points[0].x = t->extent.x1 + baseLineOffset.x;
	t->data.txtbox->trans.points[0].y = t->extent.y1 + baseLineOffset.y;
	t->data.txtbox->trans.points[5].x = t->extent.x1 + baseLineOffset.x;
	t->data.txtbox->trans.points[5].y = t->extent.y1 + baseLineOffset.y;

	/* grab lengths to calculate point 6 (0:1 as 6:2) */
	xLen = t->data.txtbox->trans.points[0].x - t->data.txtbox->trans.points[1].x;
	yLen = t->data.txtbox->trans.points[0].y - t->data.txtbox->trans.points[1].y;
	t->data.txtbox->trans.points[6].x = t->data.txtbox->trans.points[2].x + xLen;
	t->data.txtbox->trans.points[6].y = t->data.txtbox->trans.points[2].y + yLen;

	t->extent.x2 = t->extent.x1 + maxx.x;
	t->extent.y2 = t->extent.y1 + maxx.y;

	cvt->x_off = baseLineOffset.x;
	cvt->y_off = baseLineOffset.y;

	cvt->curr_atts.field_angle = tp.angle;
	cvt->curr_atts.field_shear = tp.xshear;

	/* HACK HACK - allow for error on the side of UNSCALED text */
	cvt->curr_atts.field_x_scale =
		((tp.xscale>9500)&&(tp.xscale<10500)) ? 10000 : tp.xscale;
	cvt->curr_atts.field_y_scale =
		((tp.yscale>9500)&&(tp.yscale<10500)) ? 10000 : tp.yscale;
	if (cvt->curr_atts.field_x_scale == 0)
		cvt->curr_atts.field_x_scale = 10000;
	if (cvt->curr_atts.field_y_scale == 0)
		cvt->curr_atts.field_y_scale = 10000;

#endif
}

/****************************************************************************
 *
 * module:	mlMakeTrans7
 */
mlMakeTrans7(ge, cvt, t)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
{
	/* take the minimal info and complete it.  */
	int ii, sz, hyt, wid, xt, yt, width, height;
	int valign; /* 0 == top , 1 == baseline, 2 == bottom */
	int halign; /* 0 == left, 1 == center, 2 == right */
	coord lowerRightCorner, baseline;
	double sinAng, cosAng, xShear;
	poly_type *trans;

	t->extent.x2 = 0;
	t->extent.y2 = 0;
	valign = t->attr.v_just;
	halign = t->attr.h_just;
	getTextInfo(ge, t, &lowerRightCorner, &baseline);

	if (halign == CTR_JUST)
	{
		wid = (lowerRightCorner.x1 - t->extent.x1)/2;
		t->extent.x1 -= wid;
		lowerRightCorner.x1 -= wid;
	}
	else if (halign == RIGHT_JUST)
	{
		wid = lowerRightCorner.x1 - t->extent.x1;
		t->extent.x1 -= wid;
		lowerRightCorner.x1 -= wid;
	}

	if (valign == 1) /* Baseline */
	{
		t->extent.y1 -= baseline.y1;
		lowerRightCorner.y1 -= baseline.y1;
	}

	else if (valign == 2) /* Bottom */
	{
		hyt = lowerRightCorner.y1 - t->extent.y1;
		t->extent.y1 -= hyt;
		lowerRightCorner.y1 -= hyt;
	}

	hyt = lowerRightCorner.y1 - t->extent.y1;
	wid = lowerRightCorner.x1 - t->extent.x1;

	t->extent.x2 = t->extent.x1 + wid;
	t->extent.y2 = t->extent.y1 + hyt;
	t->data.txtbox->x_off = cvt->x_off = baseline.x1;
	t->data.txtbox->y_off = cvt->y_off = baseline.y1;

	trans = &t->data.txtbox->trans;
	trans->nPoints = 7;
	trans->nCtrls = 0;
	sz = 7 * sizeof(TmPoint32);
	trans->points = (TmPoint32 *)regloc(ge, trans->points, sz);
	trans->controls = (TmPoint32 *)0;
	trans->points[0].x = 0;
	trans->points[0].y = baseline.y1;
	trans->points[1].x = 0;
	trans->points[1].y = 0;
	trans->points[2].x = wid;
	trans->points[2].y = 0;
	trans->points[3].x = wid;
	trans->points[3].y = hyt;
	trans->points[4].x = 0;
	trans->points[4].y = hyt;
	trans->points[5].x = 0;
	trans->points[5].y = baseline.y1;
	trans->points[6].x = wid;
	trans->points[6].y = baseline.y1;

	if (cvt->curr_atts.field_angle != 0)
	{
		cvt->curr_atts.field_angle = cvt->curr_atts.field_angle / 1000;
		sinAng = sin((double)cvt->curr_atts.field_angle *  C_FACT);
		cosAng = cos((double)cvt->curr_atts.field_angle *  C_FACT);

		for (ii=0; ii<7; ii++)
		{
			xt = trans->points[ii].x;
			yt = trans->points[ii].y;

			/* rotate */
			trans->points[ii].x = ((xt*cosAng + yt*sinAng) + 0.5);
			trans->points[ii].y = ((yt*cosAng - xt*sinAng) + 0.5);
		}
	}

	/* calculate the angle and shear values */
	extractSizeAndAngle(ge, &trans->points[1],
						&trans->points[2],
						&trans->points[3],
						&trans->points[4],
						&width,&height,&sinAng, &cosAng, &xShear);
	if ( asin(sinAng) < 0 )
		cvt->curr_atts.field_angle =
			3600000 - 180.0 * acos(cosAng)*10000.0/M_PI;
	else
		cvt->curr_atts.field_angle = 180.0 * acos(cosAng)*10000.0/M_PI;

	cvt->curr_atts.field_shear = xShear * 10000.0;

	for (ii=0; ii<7; ii++)
	{
		trans->points[ii].x += t->extent.x1;
		trans->points[ii].y += t->extent.y1;
	}

	cvt->curr_atts.field_x_scale = cvt->curr_atts.field_y_scale = 10000;
}

/****************************************************************************
 *
 * module:	mlInGroup
 */
mlInGroup(ge,t, cvt, obj)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
cvt_part_creation_type *obj;
{
	int size, saw_child;
	thing *grpt;

	saw_child = FALSE;
	while ( !obj->done && !cvt->done ) switch ( mlInTok(ge, cvt, TRUE) )
	{
		case TOK_TOK:
			(TokFuncs[cvt->tok.id])(ge, cvt, t, obj);
		break;

		case TOK_STR:
			mlInError(ge, cvt, ML_STRNEXP);
		break;

		case TOK_NUM:
			mlInError(ge, cvt, ML_NUMNEXP);
		break;

		case TOK_OTY:
			/* set the attribute on the textbox if not already done */
			if ( saw_child == FALSE )
			{
				mlInSetAttrs(ge, cvt, t, obj);
				saw_child = TRUE;
			}

			/* read the child (it may be a null) */
			grpt = cvt->tok.id ? mlInPart(ge, cvt, cvt->tok.id) : (thing *)0;
			size = (t->data.group->num+1) * sizeof(thing*);
			t->data.group->list = (thing**)regloc(ge, t->data.group->list, size);
			t->data.group->list[t->data.group->num] = grpt;
			t->data.group->num++;
			cvtUpdateStatus(ge, cvt);
		break;
	}

	/* set the attribute on the group if not already done */
	if ( saw_child == FALSE )
		mlInSetAttrs(ge, cvt, t, obj);
	calcExtents(ge, t);
	primFix(ge, t, t->type);

	/* done */
	return (AOK );
}

/****************************************************************************
 *
 * module:	mlInEll
 */
int
mlInEll(ge,t,cvt, obj)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
cvt_part_creation_type *obj;
{
	thing *s_t;
	thing_attributes cattr;
	int saw_text = FALSE;
	
	cvt->active_t = (thing *)0;
	while ( !obj->done && !cvt->done ) switch ( mlInTok(ge, cvt, TRUE) )
	{
		case TOK_TOK:
			if ( cvt->tok.id == TOUTLINE )
				mlInPath(ge, cvt, &t->paths, &t->n_paths, TRUE);
			else if ( cvt->tok.id == TPNTS )
			{
				cvt->untok = TRUE;
				mlInPath(ge, cvt, &t->paths, &t->n_paths, TRUE);
			}
			else if ( cvt->tok.id == TCHART )
			{
				cvt->untok = TRUE;
				obj->done = TRUE;
			}
			else if ( cvt->tok.id == TCTXT )
			{
				cvt->untok = FALSE;
				saw_text = TRUE;
				mlInSetAttrs(ge, cvt, t, obj);
				memcpy(&cattr, &cvt->curr_atts, sizeof(thing_attributes));
				t->data.ell->txt_t = (thing *)CvtPartGenesis2(ge, cvt,
													TEXTBOX, &t->attr);
				cvt->active_t = t->data.ell->txt_t;
				cvt->dx = 0;
				cvt->dy = 0;
			}
			else if ( cvt->tok.id == TCSTR )
			{
				s_t = (thing *)CvtPartGenesis2(ge, cvt, TEXT, &cvt->curr_atts);
				addStrToTextBox(ge, t->data.ell->txt_t, s_t, -1);
				cvt->active_t = s_t;
			}

			else
				(TokFuncs[cvt->tok.id])(ge, cvt, t, obj);
		break;

		case TOK_STR:
			if ( t->data.ell->txt_t )
            {
				
                memcpy(&cvt->active_t->attr, &cvt->curr_atts,
                       sizeof(thing_attributes));
                cvt->active_t->data.str->n_chars = 
					cvt->tok.val[0].u.cval.n_chars;
                cvt->active_t->data.str->chars = cvt->tok.val[0].u.cval.str;
                cvt->tok.val[0].u.cval.n_chars = 0;
                cvt->tok.val[0].u.cval.str = (tx_char *)0;
            }
            else
                mlInError(ge, cvt, ML_STRNEXP);			
		break;
		case TOK_NUM:
			mlInError(ge, cvt, ML_NUMNEXP);
		break;

		case TOK_OTY:
			cvt->untok = TRUE; /* unTok the OTY */
			obj->done = TRUE;
		break;
	}

	/* repair (a pre rev-300 doc) */
	if ( t->n_paths == 0 )
	{
		t->n_paths = 1;
		t->paths = (poly_type*)gloc(ge, sizeof(poly_type));
		memset(t->paths, 0x0, sizeof(poly_type));
		t->paths->nPoints = 5;
		t->paths->points = (TmPoint32*)gloc(ge, 5*sizeof(TmPoint32));
		t->paths->points[0].x = t->paths->points[3].x =
		t->paths->points[4].x = t->extent.x1;
		t->paths->points[1].x = t->paths->points[2].x = t->extent.x2;
		t->paths->points[0].y = t->paths->points[1].y =
		t->paths->points[4].y = t->extent.y1;
		t->paths->points[2].y = t->paths->points[3].y = t->extent.y2;
	}
	
	
	if ( saw_text == FALSE )
		mlInSetAttrs(ge, cvt, t, obj);
	else
		memcpy(&cvt->curr_atts, &cattr, sizeof(thing_attributes));
	if ( cvt->info.version > 311 )
		t->data.ell->n_sides = 0;
	calcExtents(ge, t);
	primFix(ge, t, t->type);
	cvt->active_t = (thing *)0;

	/* done */
	return (AOK );
}

/****************************************************************************
 *
 * module:	mlInRPoly
 */
mlInRPoly(ge, t, cvt, obj)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
cvt_part_creation_type *obj;
{
	t->type = ELL;
	while ( !obj->done && !cvt->done ) switch ( mlInTok(ge, cvt, TRUE) )
	{
		case TOK_TOK:
			if ( cvt->tok.id == TOUTLINE )
				mlInPath(ge, cvt, &t->paths, &t->n_paths, TRUE);
			else if ( cvt->tok.id == TPNTS )
			{
				cvt->untok = TRUE;
				mlInPath(ge, cvt, &t->paths, &t->n_paths, TRUE);
			}
			else if ( cvt->tok.id == TCHART )
			{
				cvt->untok = TRUE;
				obj->done = TRUE;
			}
			else
				(TokFuncs[cvt->tok.id])(ge, cvt, t, obj);
		break;

		case TOK_STR:
			mlInError(ge, cvt, ML_STRNEXP);
		break;

		case TOK_NUM:
			mlInError(ge, cvt, ML_NUMNEXP);
		break;

		case TOK_OTY:
			cvt->untok = TRUE; /* unTok the OTY */
			obj->done = TRUE;
		break;
	}
	mlInSetAttrs(ge, cvt, t, obj);
	calcExtents(ge, t);		
	primFix(ge, t, t->type);

	/* done */
	return (AOK );
}

/****************************************************************************
 *
 * module:	mlInInset
 */
mlInInset(ge, t, cvt, obj)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
cvt_part_creation_type *obj;
{
	int type = (-1);
	int ix = (-1);
	while ( !obj->done && !cvt->done ) switch ( mlInTok(ge, cvt, TRUE) )
	{
		case TOK_TOK:
			if ( cvt->tok.id == TOUTLINE )
				mlInPath(ge, cvt, &t->paths, &t->n_paths, TRUE);
			else if ( cvt->tok.id == TPNTS )
			{
				cvt->untok = TRUE;
				mlInPath(ge, cvt, &t->paths, &t->n_paths, TRUE);
			}
			else if ( cvt->tok.id == TINS_AREA )
			{
				mlInTok(ge, cvt, FALSE);
				t->data.inset->area.x1 = cvt->tok.val[0].u.ival + PAGE_ZEROX;
				t->data.inset->area.y1 = cvt->tok.val[1].u.ival + PAGE_ZEROY;
				t->data.inset->area.x2 = cvt->tok.val[2].u.ival + PAGE_ZEROX;
				t->data.inset->area.y2 = cvt->tok.val[3].u.ival + PAGE_ZEROY;
			}
			else if ( cvt->tok.id == TINS_REF )
			{
				mlInTok(ge, cvt, FALSE);
				type = t->data.inset->type = cvt->tok.val[0].u.ival;
				ix = cvt->tok.val[1].u.ival;
				t->data.inset->ix = ix;
				if ( ix < 0 )
				{
					type = t->data.inset->type = INSET_NONE;
					t->data.inset->ix = 0;
				}
				if ( cvt->tok.n_vals > 2 )
				{
					if ( type == INSET_LINKED )
					{
						t->data.inset->unsaved = cvt->tok.val[2].u.ival;
						if ( ix < ge->n_links )
						{
							ge->link[ix].rt_inset.type = cvt->tok.val[3].u.ival;
							ge->link[ix].rt_inset.icon_only =
							 cvt->tok.val[4].u.ival;
							ge->link[ix].rt_inset.exec =
							 (char *)cvtGetStrFromTokVal(ge, cvt, 5);
							ge->link[ix].rt_inset.filt =
							 (char *)cvtGetStrFromTokVal(ge, cvt, 6);
							if ( ge->link[ix].rt_inset.exec &&
							     ge->link[ix].rt_inset.exec[0] == (char)0 )
							{
								gfree(ge, ge->link[ix].rt_inset.exec);
								ge->link[ix].rt_inset.exec = (char *)0;
							}
							if ( ge->link[ix].rt_inset.filt &&
							     ge->link[ix].rt_inset.filt[0] == (char)0 )
							{
								gfree(ge, ge->link[ix].rt_inset.filt);
								ge->link[ix].rt_inset.filt = (char *)0;
							}
						}
						else
						{
							t->data.inset->type = INSET_NONE;
							t->data.inset->ix = 0;
						}
					}
					else if ( type == INSET_EMBEDED )
					{
						t->data.inset->unsaved = cvt->tok.val[2].u.ival;
						if ( ix < ge->n_embeds )
						{
							ge->embed[ix].unsaved = t->data.inset->unsaved;
							ge->embed[ix].rt_inset.type = cvt->tok.val[3].u.ival;
							ge->embed[ix].rt_inset.icon_only =
							 cvt->tok.val[4].u.ival;
							ge->embed[ix].rt_inset.exec =
							 (char *)cvtGetStrFromTokVal(ge, cvt, 5);
							ge->embed[ix].rt_inset.filt =
							 (char *)cvtGetStrFromTokVal(ge, cvt, 6);
							if ( ge->embed[ix].rt_inset.exec &&
							     ge->embed[ix].rt_inset.exec[0] == (char)0 )
							{
								gfree(ge, ge->embed[ix].rt_inset.exec);
								ge->embed[ix].rt_inset.exec = (char *)0;
							}
							if ( ge->embed[ix].rt_inset.filt &&
							     ge->embed[ix].rt_inset.filt[0] == (char)0 )
							{
								gfree(ge, ge->embed[ix].rt_inset.filt);
								ge->embed[ix].rt_inset.filt = (char *)0;
							}
						}
						else
						{
							t->data.inset->type = INSET_NONE;
							t->data.inset->ix = 0;
						}
					}
				}
				if ( cvt->tok.n_vals > 7 )
				{
					t->data.inset->scale_mode = cvt->tok.val[7].u.ival;
					t->data.inset->proportional = cvt->tok.val[8].u.ival;
					t->data.inset->centered = cvt->tok.val[9].u.ival;
					t->data.inset->x_zoom = cvt->tok.val[10].u.ival;
					t->data.inset->y_zoom = cvt->tok.val[11].u.ival;
				}
#if !FOR_FILTER
				if ( cvt->tok.n_vals > 12 )
				{
					if ( type == INSET_LINKED )
					{
						ge->link[ix].rt_inset.data =
						 (char *)cvtGetStrFromTokVal(ge, cvt, 12);
						ge->link[ix].rt_inset.data =
						 (char *)AxTaskInsetStrToElfData(ge->pid,
						  ge->link[ix].rt_inset.data);
						ge->link[ix].rt_inset.bad_handle = TRUE;
					}
					else if ( type == INSET_EMBEDED )
					{
						ge->embed[ix].rt_inset.data =
						 (char *)cvtGetStrFromTokVal(ge, cvt, 12);
						ge->embed[ix].rt_inset.data =
						 (char *)AxTaskInsetStrToElfData(ge->pid,
						  ge->embed[ix].rt_inset.data);
						ge->embed[ix].rt_inset.bad_handle = TRUE;
					}
				}
#endif
			}
			else if ( cvt->tok.id == TCHART )
			{
				cvt->untok = TRUE;/* unTok the chart */
				obj->done = TRUE;
			}
			else
				(TokFuncs[cvt->tok.id])(ge, cvt, t, obj);
		break;

		case TOK_STR:
			mlInError(ge, cvt, ML_STRNEXP);
		break;

		case TOK_NUM:
			mlInError(ge, cvt, ML_NUMNEXP);
		break;

		case TOK_OTY:
			cvt->untok = TRUE; /* unTok the OTY */
			obj->done = TRUE;
		break;
	}
	mlInSetAttrs(ge, cvt, t, obj);
	calcExtents(ge, t);
	primFix(ge, t, t->type);

	/* done */
	return(AOK);
}

/****************************************************************************
 *
 * module:	mlInTemp
 */
mlInTemp(ge, t, cvt, obj)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
cvt_part_creation_type *obj;
{
	thing_attributes cattr;
	int sz, enable_build, saw_rect;
	thing *child_t;

	memcpy((char *)&t->attr, (char *)&ge->def_attr, sizeof(thing_attributes));
	memcpy(&cattr, &cvt->curr_atts, sizeof(thing_attributes));
	cvt->active_t = (thing *)0;
	saw_rect = FALSE;
	while ( !obj->done && !cvt->done ) switch ( mlInTok(ge, cvt, TRUE) )
	{
		case TOK_TOK:
			if ( cvt->tok.id == TOUTLINE )
				mlInPath(ge, cvt, &t->paths, &t->n_paths, TRUE);
			else if ( cvt->tok.id == TPNTS )
			{
				cvt->untok = TRUE;
				if ( cvt->active_t )
					mlInPath(ge, cvt, &cvt->active_t->paths,
					                  &cvt->active_t->n_paths, TRUE);
				else
					mlInPath(ge, cvt, &t->paths, &t->n_paths, TRUE);
			}
			else if ( cvt->tok.id == TT_PNTS )
			{
				cvt->untok = TRUE;
				mlFillPolyType(ge, cvt, &t->data.txtbox->trans,TRUE);
			}
			else if ( cvt->tok.id == TCHART )
			{
				cvt->untok = TRUE; /* unTok the chart */
				obj->done = TRUE;
			}
			else if ( cvt->tok.id == TBUILD_INFO )
			{
				sz = sizeof(ge_build_info_type);
				t->data.temp->build = (ge_build_info_type *)gloc(ge, sz);
				memset(t->data.temp->build, 0, sz);
				mlInBuildInfo(ge, cvt, t->data.temp->build, &enable_build);
				t->data.temp->enable_build = enable_build;
			}
			else if ( cvt->tok.id == TCRECT )
			{
				saw_rect = TRUE;
				mlInSetAttrs(ge, cvt, t, obj);
				memcpy(&cattr, &cvt->curr_atts, sizeof(thing_attributes));
				t->data.temp->crect_t = (thing *)CvtPartGenesis2(ge, cvt,
													RECT, &t->attr);
				cvt->active_t = t->data.temp->crect_t;
				cvt->dx = 0;
				cvt->dy = 0;
			}
			else
				(TokFuncs[cvt->tok.id])(ge, cvt, t, obj);
		break;

		case TOK_STR:
			mlInError(ge, cvt, ML_STRNEXP);
		break;

		case TOK_NUM:
			mlInError(ge, cvt, ML_STRNEXP);
		break;

		/* templates have child objects (may be null) */
		case TOK_OTY:
			if ( t->data.temp->crect_t )
				calcExtents(ge, t->data.temp->crect_t);
			memcpy(&cvt->curr_atts, &cattr, sizeof(thing_attributes));
			child_t = cvt->tok.id ? mlInPart(ge, cvt, cvt->tok.id) : (thing *)0;
			t->data.temp->child_t = child_t;
			cvtUpdateStatus(ge, cvt);
			obj->done = TRUE;
		break;
	}
	if ( !saw_rect )
		mlInSetAttrs(ge, cvt, t, obj);
	calcExtents(ge, t);
	primFix(ge, t, t->type);
	cvt->active_t = (thing *)0;

	/* done */
	return(AOK);
}

/****************************************************************************
 *
 * module:	mlInLine
 */
mlInLine(ge,t, cvt, obj)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
cvt_part_creation_type *obj;
{
	while ( !obj->done && !cvt->done ) switch ( mlInTok(ge, cvt, TRUE) )
	{
		case TOK_TOK:
			if (cvt->tok.id == TOUTLINE)
				mlInPath(ge, cvt, &t->paths, &t->n_paths, FALSE);
			else if ( cvt->tok.id == TPNTS )
			{
				cvt->untok = TRUE;
				mlInPath(ge, cvt, &t->paths, &t->n_paths, FALSE);
			}
			else if ( cvt->tok.id == TCHART)
			{
				cvt->untok = TRUE; /* unTok the chart */
				obj->done = TRUE;
			}
			else
				(TokFuncs[cvt->tok.id])(ge, cvt, t, obj);
		break;

		case TOK_STR:
			mlInError(ge, cvt, ML_STRNEXP);
		break;

		case TOK_NUM:
			mlInError(ge, cvt, ML_NUMNEXP);
		break;

		case TOK_OTY:
			cvt->untok = TRUE; /* unTok the OTY */
			obj->done = TRUE;
		break;
	}

	/* repair (a pre rev-300 doc) */
	if ( t->n_paths == 0 )
	{
		t->n_paths = 1;
		t->paths = (poly_type*)gloc(ge, sizeof(poly_type));
		memset(t->paths, 0x0, sizeof(poly_type));
		t->paths->nPoints = 2;
		t->paths->points = (TmPoint32*)gloc(ge, 2*sizeof(TmPoint32));
		t->paths->points[0].x = t->extent.x1;
		t->paths->points[0].y = t->extent.y1;
		t->paths->points[1].x = t->extent.x2;
		t->paths->points[1].y = t->extent.y2;
	}
	if ( (t->paths->points[0].x == t->paths->points[1].x) &&
          (t->paths->points[0].y == t->paths->points[1].y) )
		t->paths->points[1].x = t->paths->points[0].x+1;

	mlInSetAttrs(ge, cvt, t, obj);
	calcExtents(ge, t);
	primFix(ge, t, t->type);

	/* done */
	return (AOK );
}

/****************************************************************************
 *
 * module:	mlInRect
 */
mlInRect(ge,t, cvt, obj)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
cvt_part_creation_type *obj;
{
	thing *s_t;
	int saw_text;
	thing_attributes cattr;

	saw_text = FALSE;
	cvt->active_t = (thing *)0;
	while ( !obj->done && !cvt->done ) switch ( mlInTok(ge, cvt, TRUE) )
	{
		case TOK_TOK:
			if ( cvt->tok.id == TOUTLINE )
				mlInPath(ge, cvt, &t->paths, &t->n_paths, TRUE);
			else if ( cvt->tok.id == TPNTS )
			{
				cvt->untok = TRUE;
				mlInPath(ge, cvt, &t->paths, &t->n_paths, TRUE);
			}
			else if ( cvt->tok.id == TCHART )
			{
				cvt->untok = TRUE;/* unTok the chart */
				obj->done = TRUE;
			}
			else if ( cvt->tok.id == TCTXT )
			{
				cvt->untok = FALSE;
				saw_text = TRUE;
				mlInSetAttrs(ge, cvt, t, obj);
				memcpy(&cattr, &cvt->curr_atts, sizeof(thing_attributes));
				t->data.rect->txt_t = (thing *)CvtPartGenesis2(ge, cvt,
													TEXTBOX, &t->attr);
				cvt->active_t = t->data.rect->txt_t;
				cvt->dx = 0;
				cvt->dy = 0;
				cvt->active_t->data.txtbox->def_tabsize = ge->state.def_tabsize;
			}
			
			else if ( cvt->tok.id == TCSTR )
			{
				s_t = (thing *)CvtPartGenesis2(ge, cvt, TEXT, &cvt->curr_atts);
				addStrToTextBox(ge, t->data.rect->txt_t, s_t, -1);
				cvt->active_t = s_t;
			}
			else
				(TokFuncs[cvt->tok.id])(ge, cvt, t, obj);
		break;

		case TOK_STR:
			if ( t->data.rect->txt_t )
			{
				memcpy(&cvt->active_t->attr, &cvt->curr_atts,
				       sizeof(thing_attributes));
				cvt->active_t->data.str->n_chars = cvt->tok.val[0].u.cval.n_chars;
				cvt->active_t->data.str->chars = cvt->tok.val[0].u.cval.str;
				cvt->tok.val[0].u.cval.n_chars = 0;
				cvt->tok.val[0].u.cval.str = (tx_char *)0;
			}
			else
				mlInError(ge, cvt, ML_STRNEXP);
		break;

		case TOK_NUM:
			mlInError(ge, cvt, ML_NUMNEXP);
		break;

		case TOK_OTY:
			cvt->untok = TRUE; /* unTok the OTY */
			obj->done = TRUE;
		break;
	}

	/* repair (a pre rev-300 doc) */
	if ( t->n_paths == 0 )
	{
		t->n_paths = 1;
		t->paths = (poly_type*)gloc(ge,sizeof(poly_type));
		memset(t->paths, 0x0, sizeof(poly_type));
		t->paths->nPoints = 5;
		t->paths->points = (TmPoint32*)gloc(ge, 5*sizeof(TmPoint32));
		t->paths->points[0].x = t->paths->points[3].x =
		t->paths->points[4].x = t->extent.x1;
		t->paths->points[1].x = t->paths->points[2].x = t->extent.x2;
		t->paths->points[0].y = t->paths->points[1].y =
		t->paths->points[4].y = t->extent.y1;
		t->paths->points[2].y = t->paths->points[3].y = t->extent.y2;
	}
	if ( saw_text == FALSE )
		mlInSetAttrs(ge, cvt, t, obj);
	else
	{
		memcpy(&cvt->curr_atts, &cattr, sizeof(thing_attributes));
	}
	calcExtents(ge, t);
	primFix(ge, t, t->type);
	cvt->active_t = (thing *)0;
	if ( t->data.rect->txt_t )
		calcExtents(ge, t->data.rect->txt_t);

	/* done */
	return (AOK );
}

/****************************************************************************
 *
 * module:	mlInPoly
 */
mlInPoly(ge,t, cvt, obj)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
cvt_part_creation_type *obj;
{
	while ( !obj->done && !cvt->done ) switch ( mlInTok(ge, cvt, TRUE) )
	{
		case TOK_TOK:
			if ( cvt->tok.id == TOUTLINE)
				mlInPath(ge, cvt, &t->paths, &t->n_paths, FALSE);
			else if ( cvt->tok.id == TPNTS )
			{
				cvt->untok = TRUE;
				mlInPath(ge, cvt, &t->paths, &t->n_paths, FALSE);
			}
			else if ( cvt->tok.id == TCHART)
			{
				cvt->untok = TRUE; /* unTok the chart */
				obj->done = TRUE;
			}
			else
				(TokFuncs[cvt->tok.id])(ge, cvt, t, obj);
		break;

		case TOK_STR:
			mlInError(ge, cvt, ML_STRNEXP);
		break;

		/* repair (pre rev-300 doc) */
		case TOK_NUM:
			cvt->untok = TRUE;/* unTok the NUM */
			mlInPath(ge, cvt, &t->paths, &t->n_paths, FALSE);
			mlInPnts(ge,cvt,t->paths,FALSE);
		break;

		case TOK_OTY:
			cvt->untok = TRUE;/* unTok the OTY */
			obj->done = TRUE;
		break;
	}

	/* repair (a pre rev-300 doc), see mlInLine */
	if ( t->n_paths == 0 )
	{
		t->n_paths = 1;
		t->paths = (poly_type*)gloc(ge, sizeof(poly_type));
		memset(t->paths, 0x0, sizeof(poly_type));
		t->paths->nPoints = 2;
		t->paths->points = (TmPoint32*)gloc(ge, 2*sizeof(TmPoint32));
		t->paths->points[0].x = t->extent.x1;
		t->paths->points[0].y = t->extent.y1;
		t->paths->points[1].x = t->extent.x2;
		t->paths->points[1].y = t->extent.y2;
	}
	if ( (t->paths->points[0].x == t->paths->points[1].x) &&
          (t->paths->points[0].y == t->paths->points[1].y) )
		t->paths->points[1].x = t->paths->points[0].x+1;
	
	mlInSetAttrs(ge, cvt, t, obj);
	calcExtents(ge, t);
	primFix(ge, t, t->type);

	/* done */
	return (AOK );
}

/****************************************************************************
 *
 * Module: cvtAddTextRect - t not yet added to paint list
 */
cvtAddTextRect(ge, cvt, t) 
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
{
#if 0
	int x1, x2, y1, y2;
	thing *rect_t;

	rect_t = (thing *)CvtPartGenesis2(ge,cvt,RECT,&cvt->curr_atts);
	if ( rect_t )
	{
		/* get the rect size */
		x1 = t->extent.x1;
		y1 = t->extent.y1;
		x2 = t->extent.x2;
		y2 = t->extent.y2;
  
		/* set the type */
		rect_t->type = RECT;
		rect_t->attr = t->attr;
		rect_t->attr.fill_bg = t->attr.fill_bg;
		rect_t->attr.fill_fg = t->attr.fill_fg;
		rect_t->attr.line_fg = t->attr.line_fg;
		rect_t->attr.line_bg = t->attr.line_bg;
		rect_t->attr.fill_pat= 0;

		/* set the rect */
		rect_t->n_paths = 1;
		rect_t->paths = (poly_type*)gloc(ge, sizeof(poly_type));
		memset(rect_t->paths, 0x0, sizeof(poly_type));
		rect_t->paths->nPoints = 5;
		rect_t->paths->points = (TmPoint32*)gloc(ge, 5*sizeof(TmPoint32));
		rect_t->paths->points[0].x = x1;
		rect_t->paths->points[0].y = y1;
		rect_t->paths->points[1].x = x2;
		rect_t->paths->points[1].y = y1;
		rect_t->paths->points[2].x = x2;
		rect_t->paths->points[2].y = y2;
		rect_t->paths->points[3].x = x1;
		rect_t->paths->points[3].y = y2;
		rect_t->paths->points[4].x = x1;
		rect_t->paths->points[4].y = y1;
 
		/* finish up */
		calcExtents(ge, rect_t);
		primFix(ge, rect_t, rect_t->type);
		cvtAddToPaintList(ge, rect_t, -1);
		cvt->lastobj = rect_t;
	}
#endif
}

/****************************************************************************
 *
 * Module: cvtRepairTextbox
 */
cvtRepairTextbox(ge, cvt, t)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
{
	thing *str_t;
	axtext txt;
	int i;

	if ( cvt->info.version <= 420 )
	{
		/* make sure the entire textbox shares the container line spacing */
		if ( cvt->info.version <= 400 )
		{
			for ( i=0; i<t->data.txtbox->n_strings; i++ )
			{
				str_t = t->data.txtbox->strings[i];
				if ( str_t == (thing *)0 )
					continue;
				str_t->attr.field_space  = t->attr.field_space;
				str_t->attr.space_inside = t->attr.space_inside;
				str_t->attr.space_above  = t->attr.space_above;
				str_t->attr.space_below  = t->attr.space_below;
			}
		}

		/* make sure the entire textbox shares the container shadows */
		if ( cvt->info.version < 420 )
		{
			for ( i=0; i<t->data.txtbox->n_strings; i++ )
			{
				str_t = t->data.txtbox->strings[i];
				if ( str_t == (thing *)0 )
					continue;
				str_t->attr.shadow       = t->attr.shadow;
				str_t->attr.shadow_color = t->attr.shadow_color;
				str_t->attr.shadow_x     = t->attr.shadow_x;
				str_t->attr.shadow_y     = t->attr.shadow_y;
			}
		}

		if ( t->data.txtbox->field_wid )
		{
			gInstantiateTxt(ge, t, &txt, TRUE, FALSE, TRUE);
			daxtext(&txt);
		}
		else if ( cvt->ch_data.wid )			
		{
			/* if field_wid is not set use the field_wid from ch_data. */
			/* Note: ch_data is valid only if cvt->isa_chart is true   */
			if ( t->n_paths )
			{	
				disposePaths(ge, t->n_paths, t->paths);
				t->paths = (poly_type *) 0;
				t->n_paths = 0;
			}
			t->data.txtbox->field_wid = cvt->ch_data.wid;
			t->flags.txt_measured     = FALSE;
			calcExtents(ge, t);
		}

	}
}

/****************************************************************************
 *
 * Module: cvtRepairText
 */
cvtRepairText(ge, cvt, t, s_t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
thing *s_t;
cvt_part_creation_type *obj;
{
	/* repair (most probably either 2.1 or filtered 2.1) */
	if ( t->data.txtbox->trans.nPoints == 5 )
	{
		mlInSetAttrs(ge, cvt, t, obj);
		mlTrans5ToTrans7(ge, cvt,t);
		mlInSetAttrs(ge, cvt, t, obj);
		calcExtents(ge, t);
		primFix(ge, t, t->type);
	}

	/* repair */
	else if  ( t->data.txtbox->trans.nPoints != 7 && t->n_paths == 0 )
	{
		/* get rid of bad transform bounders */
		if ( t->data.txtbox->trans.nPoints && t->data.txtbox->trans.points )
		{
			gfree(ge, t->data.txtbox->trans.points);
			t->data.txtbox->trans.points = (TmPoint32 *)0;
			t->data.txtbox->trans.nPoints = 0;
		}

		/* create a new (un)transform(ed) bounder */
		mlMakeTrans7(ge,cvt,t);
		mlInSetAttrs(ge, cvt, t, obj);
		t->attr.v_just = s_t->attr.v_just = 0; /* Top just */
		t->attr.h_just = s_t->attr.h_just = LEFT_JUST;
		calcExtents(ge, t);
		primFix(ge, t, t->type);
	}
}

/****************************************************************************
 *
 * module:	mlInText
 */
mlInText(ge,t, cvt, obj)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
cvt_part_creation_type *obj;
{
	int ii, str_ix, sz, field_space, old_field_space;
	thing *s_t;
	tx_char *txP;
	char *chr;

	/* create a text container */
	s_t = (thing *)makeThing(ge, TEXT, (thing_attributes *)0);
	if ( !s_t )
		longjmp(cvt->jmpPoint, SGE_NOFILE);

	/* read the text info from the file */
	field_space = cvt->curr_atts.field_space;
	while ( !obj->done && !cvt->done ) switch ( mlInTok(ge, cvt, TRUE) )
	{
		case TOK_TOK:
			if ( cvt->tok.id == TOUTLINE )
				mlInPath(ge, cvt, &s_t->paths, &s_t->n_paths, TRUE);
			else if ( cvt->tok.id == TPNTS )
			{
				cvt->untok = TRUE;
				mlInPath(ge, cvt, &s_t->paths, &s_t->n_paths, TRUE);
			}
			else if ( cvt->tok.id == TCHART )
			{
				cvt->untok = TRUE; /* unTok the chart */
				obj->done = TRUE;
			}
			else
				(TokFuncs[cvt->tok.id])(ge, cvt, t, obj);
		break;

		case TOK_STR:
			s_t->data.str->n_chars = cvt->tok.val[0].u.cval.n_chars;
			s_t->data.str->chars = cvt->tok.val[0].u.cval.str;
			cvt->tok.val[0].u.cval.n_chars = 0;
			cvt->tok.val[0].u.cval.str = (tx_char *)0;
			obj->sawChild = TRUE;
		break;

		/* repair (pre rev-300 doc) */
		case TOK_NUM:
			obj->sawTxtCorner = obj->sawCorner;
			cvt->untok = TRUE; /* unTok the NUM */
			mlInPnts(ge,cvt,&t->data.txtbox->trans,TRUE);
		break;

		case TOK_OTY:
			cvt->untok = TRUE; /* unTok the OTY */
			obj->done = TRUE;
		break;
	}
	if ( !s_t->data.str->chars ) /* need to free */
	{
		s_t->data.str->chars = (tx_char *)gloc(ge, sizeof(tx_char));
		s_t->data.str->chars[0] = (tx_char)0;
	}

	/* apply the current attributes to the newly created text object */
	if ( cvt->info.version >= 420 )
		mlInSetAttrs(ge, cvt, s_t, obj);
	else
	{
		old_field_space = cvt->curr_atts.field_space;
		cvt->curr_atts.field_space  = field_space;
		cvt->curr_atts.space_inside = field_space;
		cvt->curr_atts.space_above  = 0;
		cvt->curr_atts.space_below  = 0;
		mlInSetAttrs(ge, cvt, s_t, obj);
		cvt->curr_atts.field_space  = old_field_space;
		cvt->curr_atts.space_inside = old_field_space;
	}

	/* make sure it's a textbox */
	if ( t->type == TEXT )
	{
		t->type = TEXTBOX;
		gfree(ge, t->data.gen);
		t->data.gen = (char *)0;
		makeData(ge, t, t->type);
	}

	/* add the text to the textbox */
	addStrToTextBox(ge, t, s_t, -1);

	/* repair (most probably either 2.1 or filtered 2.1) */
	cvtRepairText(ge, cvt, t, s_t, obj);

	/* done */
	return (AOK );
}

/****************************************************************************
 *
 * module:	mlInBoxText
 */
mlInBoxText(ge,t, cvt, obj)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
cvt_part_creation_type *obj;
{
	int stat;

	stat = !AOK;
	t->data.txtbox->def_tabsize = ge->state.def_tabsize;
	while ( !obj->done && !cvt->done ) switch ( mlInTok(ge, cvt, TRUE) )
	{
		case TOK_TOK:
			if ( cvt->tok.id == TOUTLINE )
				mlInPath(ge, cvt, &t->paths, &t->n_paths, TRUE);
			else if ( cvt->tok.id == TPNTS )
			{
				cvt->untok = TRUE;
				mlInPath(ge, cvt, &t->paths, &t->n_paths, TRUE);
			}
			else if ( cvt->tok.id == TTRANS )
				mlFillPolyType(ge, cvt, &t->data.txtbox->trans,TRUE);
			else if ( cvt->tok.id == TT_PNTS )
			{
				cvt->untok = TRUE;
				mlFillPolyType(ge, cvt, &t->data.txtbox->trans,TRUE);
			}
			else if ( cvt->tok.id == TCHART )
			{
				cvt->untok = TRUE; /* unTok the chart */
				obj->done = TRUE;
			}
			else
				(TokFuncs[cvt->tok.id])(ge, cvt, t, obj);
		break;

		case TOK_STR:
			mlInError(ge, cvt, ML_STRNEXP);
		break;

		case TOK_NUM:
			mlInError(ge, cvt, ML_STRNEXP);
		break;

		/* text boxes have child text objects */
		case TOK_OTY:
			if ( cvt->tok.id == TEXT )
			{
				/* set the attribute on the textbox if not already done */
				if (obj->sawChild == FALSE)
					mlInSetAttrs(ge, cvt, t, obj);

				/* read the child object */
				stat = mlInText(ge,t, cvt, obj);
				obj->done = FALSE; /* force another loop */
				obj->sawChild = TRUE;
			}
			else
			{
				cvt->untok = TRUE; /* unTok the OTY */
				obj->done = TRUE;
			}
		break;
	}

	/* object has both a bounder and a transform - use transform */
	if ( t->n_paths > 0 && t->data.txtbox->trans.nPoints > 0 )
	{
		disposePaths(ge, t->n_paths, t->paths);
		t->paths = (poly_type *) 0;
		t->n_paths = 0;
	}

	if ( t->n_paths <= 0 && t->data.txtbox->n_strings <= 0 )
	{
		disposeThing(ge, t, TRUE);
		stat = !AOK;
	}
	else /* don't set the attribute on the textbox if it's already been done */
	{
		if ( obj->sawChild == FALSE )
			mlInSetAttrs(ge, cvt, t, obj);
		calcExtents(ge, t);
		primFix(ge, t, t->type);
		stat = AOK;
	}

	/* done */
	cvtRepairTextbox(ge, cvt, t);
	return(stat);
}

/****************************************************************************
 *
 * module:	mlInImg
 */
mlInImg(ge,t,cvt, obj)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
cvt_part_creation_type *obj;
{
	while ( !obj->done && !cvt->done ) switch ( mlInTok(ge, cvt, TRUE) )
	{
		case TOK_TOK:
			if ( cvt->tok.id == TOUTLINE )
			{
				if ( cvt->majorVer == GVER1_MAJOR )
				{
					mlFillPolyType(ge, cvt, &t->data.img->trans,TRUE);
					mlInPnts(ge,cvt,&t->data.img->trans,TRUE);
				}
				else
					mlInPath(ge, cvt, &t->paths, &t->n_paths, TRUE);
			}
			else if ( cvt->tok.id == TPNTS )
			{
				cvt->untok = TRUE;
				mlInPath(ge, cvt, &t->paths, &t->n_paths, TRUE);
			}
			else if ( cvt->tok.id == TTRANS )
				mlFillPolyType(ge, cvt, &t->data.img->trans,TRUE);
			else if ( cvt->tok.id == TT_PNTS )
			{
				cvt->untok = TRUE;
				mlFillPolyType(ge, cvt, &t->data.img->trans,TRUE);
			}
			else if ( cvt->tok.id == TCHART )
			{
				cvt->untok = TRUE; /* unTok the chart */
				obj->done = TRUE;
			}
			else
				(TokFuncs[cvt->tok.id])(ge, cvt, t, obj);
		break;

		case TOK_STR:
			mlInError(ge, cvt, ML_STRNEXP);
		break;

		case TOK_NUM:
			mlInError(ge, cvt, ML_NUMNEXP);
		break;

		case TOK_OTY:
			cvt->untok = TRUE; /* unTok the OTY */
			obj->done = TRUE;
		break;
	}

	/* repair (pre rev-300) */
	if ( (t->data.img->info.dep == 8) && (t->data.img->info.mask == NULL) )
		mlCreateImgMask(ge, t, cvt,0xFF);

	/* repair (create a unique colormap for pre rev-300 images) */
	if ( (cvt->majorVer == GVER1_MAJOR) && (t->data.img->info.dep == 8) )
	{
		cvt->cmapID = ge->n_cmaps;
		mlOldStyleImgCmap(ge, cvt);
	}

	mlInSetAttrs(ge, cvt, t, obj);
	mlCheckColorRefs(ge, cvt, t);
	calcExtents(ge, t);
	primFix(ge, t, t->type);

	/* done */
	return (AOK );
}

/****************************************************************************
 *
 * module:	c v t R e p a i r P a t h
 */
cvtRepairPath(ge, cvt, t)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
{
	int i, j, last_j, n_pnts, n_ctls, n_old_ctls, n_old_pnts, sz;
	TmPoint32 *pnt, *ctl, *old_pnt, *old_ctl;

	for ( i=0; i<t->n_paths; i++ )
	{
		n_old_pnts = t->paths[i].nPoints;
		n_old_ctls = t->paths[i].nCtrls;
		if ( n_old_ctls > 0 )
		{
			sz = n_old_pnts * sizeof(TmPoint32);
			pnt = (TmPoint32 *)gloc(ge, sz);
			old_pnt = t->paths[i].points;
			pnt[0] = old_pnt[0];
			n_pnts = 1;

			sz = (2*n_old_pnts - 2) * sizeof(TmPoint32);
			ctl = (TmPoint32 *)gloc(ge, sz);
			old_ctl = t->paths[i].controls;
			n_ctls = 0;

			last_j = 0;
			for ( j=1; j<n_old_pnts; j++ )
			{
				if ( old_pnt[j-1].x == old_pnt[j].x &&
				     old_pnt[j-1].y == old_pnt[j].y )
					continue;
				else
				{
					pnt[n_pnts] = old_pnt[j];
					n_pnts++;

					ctl[n_ctls] = old_ctl[2*last_j];
					n_ctls++;
					ctl[n_ctls] = old_ctl[2*j - 1];
					n_ctls++;

					last_j = j;
				}
			}

			gfree(ge, old_pnt);
			gfree(ge, old_ctl);

			sz = n_pnts * sizeof(TmPoint32);
			t->paths[i].points = (TmPoint32 *)regloc(ge, pnt, sz);
			t->paths[i].nPoints = n_pnts;

			sz = n_ctls * sizeof(TmPoint32);
			t->paths[i].controls = (TmPoint32 *)regloc(ge, ctl, sz);
			t->paths[i].nCtrls = n_ctls;
		}
		else if ( n_old_pnts > 0 )
		{
			sz = n_old_pnts * sizeof(TmPoint32);
			pnt = (TmPoint32 *)gloc(ge, sz);
			old_pnt = t->paths[i].points;
			pnt[0] = old_pnt[0];
			n_pnts = 1;

			for ( j=1; j<n_old_pnts; j++ )
			{
				if ( old_pnt[j-1].x == old_pnt[j].x &&
				     old_pnt[j-1].y == old_pnt[j].y )
					continue;
				else
				{
					pnt[n_pnts] = old_pnt[j];
					n_pnts++;
				}
			}

			gfree(ge, old_pnt);

			sz = n_pnts * sizeof(TmPoint32);
			t->paths[i].points = (TmPoint32 *)regloc(ge, pnt, sz);
			t->paths[i].nPoints = n_pnts;

			t->paths[i].controls = (TmPoint32 *)0;
			t->paths[i].nCtrls = 0;
		}
	}
}

/****************************************************************************
 *
 * module:	m l I n P a r t - read a graphic thing
 */
thing *
mlInPart(ge, cvt, t_type)
graphics_main_type *ge;
cvt_global_type *cvt;
int t_type;
{
	cvt_part_creation_type obj;
	thing *t;
	int stat;

	/* set-up pages as needed */
	if ( !ge->t && cvt->info.version < 320 )
	{
		cvt->inserted_page = TRUE;
		cvtAppendPage(ge, cvt, &cvt->curr_atts);
		mlInCheckPage(ge, cvt, &ge->page[ge->page_ix]);
	}

	/* create only if it's a valid type */
	if ( t_type == -1 )
		t = (thing *)0;
	else
	{
		/* create the object */
		t = (thing *)CvtPartGenesis2(ge, cvt, t_type, (thing_attributes *)0);
		if ( !t )
			longjmp(cvt->jmpPoint, SGE_NOFILE);

		/* initialize the object */
		cvt->dx = 0;
		cvt->dy = 0;
		mlCopyTags(ge,cvt,t);
		memset(&obj, 0, sizeof(obj));
		switch ( t_type )
		{
			case ELL:        stat = mlInEll    (ge, t, cvt, &obj); break;
			case LINE:       stat = mlInLine   (ge, t, cvt, &obj); break;
			case RECT:       stat = mlInRect   (ge, t, cvt, &obj); break;
			case STROK:      stat = mlInPoly   (ge, t, cvt, &obj); break;
			case POLY:       stat = mlInPoly   (ge, t, cvt, &obj); break;
			case TEXT:       stat = mlInText   (ge, t, cvt, &obj); break;
			case TEXTBOX:    stat = mlInBoxText(ge, t, cvt, &obj); break;
			case IMAGE:      stat = mlInImg    (ge, t, cvt, &obj); break;
			case GROUP:      stat = mlInGroup  (ge, t, cvt, &obj); break;
			case RPOL:       stat = mlInRPoly  (ge, t, cvt, &obj); break;
			case INSET:      stat = mlInInset  (ge, t, cvt, &obj); break;
			case TEMPLATE:   stat = mlInTemp   (ge, t, cvt, &obj); break;
		}
		if ( stat != AOK )
			t = (thing *)0;
		else
		{
			t->flags.isa_chart = cvt->isa_chart;
			if ( cvt->point_damage && cvt->isa_chart)
			{
				cvt->point_damage = FALSE;
				cvtRepairPath(ge, cvt, t);
			}
		}
	}
	return(t);
}

/****************************************************************************
 *
 * module:	ReadNULL - an empty function for tokens not needing a reader
 */
void
ReadNULL(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
}

/****************************************************************************
 *
 * module:	ReadCLIPPATH  
 */
void
ReadCLIPPATH(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	clip_path *cinfo;
	int i, j, sz;

	mlInTok(ge, cvt, FALSE);
	cinfo = (clip_path *)gloc(ge, sizeof(clip_path));
	memset(cinfo, 0, sizeof(clip_path));
	cinfo->n_paths = 1;
	sz = cinfo->n_paths * sizeof(poly_type);
	cinfo->paths = (poly_type *)gloc(ge, sz);
	memset(cinfo->paths, 0, sz);
	cinfo->paths[0].nPoints = 5; 
	cinfo->paths[0].nCtrls = 0;
	cinfo->paths[0].controls = (TmPoint32 *)0;
	sz = cinfo->paths[0].nPoints*sizeof(TmPoint32);
	cinfo->paths[0].points = (TmPoint32 *)gloc(ge, sz);
	for ( i=0, j=0; j< cvt->tok.n_vals; i++ )
	{
		cinfo->paths[0].points[i].x = cvt->tok.val[j].u.ival + PAGE_ZEROX;
		cinfo->paths[0].points[i].y = cvt->tok.val[j+1].u.ival + PAGE_ZEROY;
		j+= 2;
	}
	t->clip = cinfo;
	cvt->untok = FALSE; /* read a new token */
}

/****************************************************************************
 *
 * module:	ReadCLIPMASK  - OLD TOKEN 
 */
void
ReadCLIPMASK(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	clip_path *clip;

	clip = (clip_path *)gloc(ge, sizeof(clip_path));
	memset(clip, 0, sizeof(clip_path));
	mlInPath(ge, cvt, &clip->paths, &clip->n_paths, TRUE);
	t->clip = clip;
	cvt->untok = FALSE; /* read a new token */
	mlInTok(ge, cvt, FALSE); /* get rid of token after END */
}

/****************************************************************************
 *
 * module:	ReadH_SPACE
 */
void
ReadH_SPACE(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->field_wid  = cvt->tok.val[0].u.ival;
}

/****************************************************************************
 *
 * module:	ReadHLINK
 */
void
ReadHLINK(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	int val;
	mlInTok(ge, cvt, FALSE);
	val = cvt->tok.val[0].u.ival;
}

/****************************************************************************
 *
 * module:	ReadAT
 */
void
ReadAT(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	int x, y;

	mlInCoord(ge, cvt, &x, &y, TRUE);
	cvt->dx = x - PAGE_ZEROX;
	cvt->dy = y - PAGE_ZEROY;
	if ( t )
	{
		t->extent.x1 = t->extent.x2 = x;
		t->extent.y1 = t->extent.y2 = y;
	}
}

/****************************************************************************
 *
 * module:	ReadORIGIN
 */
void
ReadORIGIN(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	ReadAT(ge, cvt, t, obj);
	cvt->dx = cvt->dy = 0;
}

/****************************************************************************
 *
 * module:	ReadCORNER
 */
void
ReadCORNER(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	int val1, val2;

	if (obj)
	{
		obj->sawCorner = TRUE;
		obj->sawTxtCorner = TRUE;
	}
	mlInCoord(ge, cvt, &val1, &val2, FALSE);
	if (t)
	{
		if (val1 == SAME)
			t->extent.x2 = t->extent.x1;
		else
			t->extent.x2 = val1;

		if (val2 == SAME)
			t->extent.y2 = t->extent.y1;
		else
			t->extent.y2 = val2;
	}
}

/****************************************************************************
 *
 * module:	ReadPOSITION
 */
void
ReadPOSITION(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	int x, y;

	if ( !t )
		mlInError(ge, cvt, ML_POSNEXP);

	else
	{
		if ( t->type == RECT )
		{
			if ( cvt->info.version < 320 )
			{
				mlInTok(ge, cvt, FALSE);
				cvt->x_corner_numerator = cvtTokToInt(ge, cvt);
				mlInTok(ge, cvt, FALSE);
				cvt->y_corner_numerator = cvtTokToInt(ge, cvt);
			}
			else
			{
				mlInTok(ge, cvt, FALSE);
				cvt->x_corner_numerator = cvt->tok.val[0].u.ival;
				cvt->y_corner_numerator = cvt->tok.val[1].u.ival;
			}
		}
		else if ( (t->type == TEXT) || (t->type == TEXTBOX) )
		{
			mlInCoord(ge, cvt, &x, &y, TRUE);
			t->extent.x1 = t->extent.x2 = x;
			t->extent.y1 = t->extent.y2 = y;
		}
		else
			mlInError(ge, cvt, ML_POSNEXP);
	}
}

/****************************************************************************
 *
 * module:	ReadV_ALIGN
 */
void
ReadV_ALIGN(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( t->type == RECT && t->data.rect->txt_t )
		t->data.rect->txt_t->data.txtbox->v_align = cvt->tok.val[0].u.ival;
	else if ( t->type == ELL && t->data.ell->txt_t )
		t->data.ell->txt_t->data.txtbox->v_align = cvt->tok.val[0].u.ival;
	else
		cvt->v_align = cvt->tok.val[0].u.ival;
}

/****************************************************************************
 *
 * module:	ReadSUBTXT
 */
void
ReadSUBTXT(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( t->type == RECT && t->data.rect->txt_t )
		t->data.rect->txt_t->data.txtbox->sub_txt = cvt->tok.val[0].u.ival;
	else if ( t->type == ELL && t->data.ell->txt_t )
		t->data.ell->txt_t->data.txtbox->sub_txt = cvt->tok.val[0].u.ival;
	else
		cvt->sub_txt = cvt->tok.val[0].u.ival;
}

/****************************************************************************
 *
 * module:	ReadLSHAPE
 */
void
ReadLSHAPE(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->lshape_type     = cvt->tok.val[0].u.ival;
	cvt->end1_cpnt       = cvt->tok.val[1].u.ival;
	cvt->end2_cpnt       = cvt->tok.val[2].u.ival;
	cvt->end1_cpnt_dir   = cvt->tok.val[3].u.ival;
	cvt->end2_cpnt_dir   = cvt->tok.val[4].u.ival;
	cvt->lshape_pnt[0].x = cvt->tok.val[5].u.ival;
	cvt->lshape_pnt[0].y = cvt->tok.val[6].u.ival;
	cvt->lshape_pnt[1].x = cvt->tok.val[7].u.ival;
	cvt->lshape_pnt[1].y = cvt->tok.val[8].u.ival;
	cvt->lshape_pnt[2].x = cvt->tok.val[9].u.ival;
	cvt->lshape_pnt[2].y = cvt->tok.val[10].u.ival;
}

/****************************************************************************
 *
 * module:	ReadSSHAPE
 */
void
ReadSSHAPE(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->sshape_type     = cvt->tok.val[0].u.ival;
	cvt->sshape_pnt[0].x = cvt->tok.val[1].u.ival;
	cvt->sshape_pnt[0].y = cvt->tok.val[2].u.ival;
	cvt->sshape_pnt[1].x = cvt->tok.val[3].u.ival;
	cvt->sshape_pnt[1].y = cvt->tok.val[4].u.ival;
	cvt->sshape_pnt[2].x = cvt->tok.val[5].u.ival;
	cvt->sshape_pnt[2].y = cvt->tok.val[6].u.ival;
	cvt->sshape_pnt[3].x = cvt->tok.val[7].u.ival;
	cvt->sshape_pnt[3].y = cvt->tok.val[8].u.ival;
	cvt->sshape_pnt[4].x = cvt->tok.val[9].u.ival;
	cvt->sshape_pnt[4].y = cvt->tok.val[10].u.ival;
	if ( cvt->tok.n_vals > 11 )
	{
	    cvt->sshape_pnt[5].x = cvt->tok.val[11].u.ival;
   	    cvt->sshape_pnt[5].y = cvt->tok.val[12].u.ival;
	}
	else
	{
	    cvt->sshape_pnt[5].x = 0;
   		cvt->sshape_pnt[5].y = 0;
	}
}

/****************************************************************************
 *
 * module:	ReadISA
 */
void
ReadISA(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	/* Parent part */
	mlInName(ge, cvt,t);
}

/****************************************************************************
 *
 * module:	ReadFIXED
 */
void
ReadFIXED(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	int fixed;

	if ( cvt->info.version < 320 )
		fixed = TRUE;
	else
	{
		mlInTok(ge, cvt, FALSE);
		if ( cvt->tok.type != TOK_TST )
			fixed = TRUE;
		else
			fixed = cvt->tok.id;
	}

	if ( t )
		t->flags.fixed = fixed;
}

/****************************************************************************
 *
 * module:	ReadRECOLOR
 */
void
ReadRECOLOR(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( t )
	{
		if ( cvt->tok.type != TOK_TST )
			t->flags.recolored = TRUE;
		else
			t->flags.recolored = cvt->tok.id;
	}
}

/****************************************************************************
 *
 * module:	ReadTEMPLATED
 */
void
ReadTEMPLATED(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( t )
	{
		if ( cvt->tok.type != TOK_TST )
			t->flags.templated = TRUE;
		else
			t->flags.templated = cvt->tok.id;
	}
}

/****************************************************************************
 *
 * module:	ReadCLASS
 */
void
ReadCLASS(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{

	mlInTok(ge, cvt, FALSE);
	t->t_class = cvt->tok.val[0].u.ival;

}

/****************************************************************************
 *
 * module:	ReadGHOSTED
 */
void
ReadGHOSTED(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
}

setChartData(ge, cvt, t, obj, val)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
int val;
{
	chart_type *ge_chart, *chart;
	if ( t )
	{	
		cvt->ch_data.chart_ix = t->chart_ix;
		ge_chart = (chart_type *)ge->chart;
		chart    = (chart_type*)&ge_chart[cvt->ch_data.chart_ix-1];
		cvt->ch_data.wid = ABS(chart->extent.x2-chart->extent.x1);
		cvt->ch_data.hyt = ABS(chart->extent.y2-chart->extent.y1);	
	}
	else if ( val )
	{
		cvt->ch_data.chart_ix = val;
		ge_chart = (chart_type *)ge->chart;
		chart    = (chart_type*)&ge_chart[cvt->ch_data.chart_ix-1];
		cvt->ch_data.wid = ABS(chart->extent.x2-chart->extent.x1);
		cvt->ch_data.hyt = ABS(chart->extent.y2-chart->extent.y1);	
	}
}
/****************************************************************************
 *
 * module:	ReadGRP_INFO
 */
void
ReadGRP_INFO(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	switch(cvt->tok.val[0].u.ival)
	{
		case 1:
			cvt->isa_chart = TRUE;
			t->chart_ix = cvt->tok.val[1].u.ival+1;
			if ( t->chart_ix <= ge->n_charts )
				setChartData(ge, cvt, t, obj, 0);
		break;
	}
}

/****************************************************************************
 *
 * module:	ReadSHR-ANG
 */
void
ReadSHR_ANG(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	if ( cvt->info.version < 320 )
	{
		mlInTok(ge, cvt, FALSE);
		cvt->curr_atts.field_shear = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		cvt->curr_atts.field_angle = cvtTokToInt(ge, cvt);
	}
	else
	{
		mlInTok(ge, cvt, FALSE);
		cvt->curr_atts.field_shear = cvt->tok.val[0].u.ival;
		cvt->curr_atts.field_angle = cvt->tok.val[1].u.ival;
	}
}

/****************************************************************************
 *
 * module:	ReadFLD-XYSCALE
 */
void
ReadFLD_XYSCALE(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
    int f_x_scale, f_y_scale;

	if ( cvt->info.version < 320 )
	{
		mlInTok(ge, cvt, FALSE);
		f_x_scale = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		f_y_scale = cvtTokToInt(ge, cvt);
	}
	else
	{
		mlInTok(ge, cvt, FALSE);
		f_x_scale = cvt->tok.val[0].u.ival;
		f_y_scale = cvt->tok.val[1].u.ival;
	}

	if ( f_x_scale <= -MAXINT/2 )
		f_x_scale = 10000;
	else if ( f_x_scale >= MAXINT/2 )
		f_x_scale = 10000;
	if ( f_y_scale <= -MAXINT/2 )
		f_y_scale = 10000;
	else if ( f_y_scale >= MAXINT/2 )
		f_y_scale = 10000;
	if ( f_x_scale == 0 )
		f_x_scale = 10000;
	if ( f_y_scale == 0 )
		f_y_scale = 10000;
	cvt->curr_atts.field_x_scale = f_x_scale;
	cvt->curr_atts.field_y_scale = f_y_scale;
}

/****************************************************************************
 *
 * module:	ReadHIDDEN
 */
void
ReadHIDDEN(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	int hidden;

	if ( cvt->info.version < 320 )
		hidden = TRUE;
	else
	{
		mlInTok(ge, cvt, FALSE);
		if ( cvt->tok.type != TOK_TST )
			hidden = TRUE;
		else
			hidden = cvt->tok.id;
	}

	if ( t )
		t->flags.hidden = hidden;
}

/****************************************************************************
 *
 * module:	ReadLOCKED
 */
void
ReadLOCKED(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	int locked;

	if ( cvt->info.version < 320 )
		locked = TRUE;
	else
	{
		mlInTok(ge, cvt, FALSE);
		if ( cvt->tok.type != TOK_TST )
			locked = TRUE;
		else
			locked = cvt->tok.id;
	}

	if ( t )
		t->flags.locked = locked;
}

/****************************************************************************
 *
 * module:	ReadTOP_ATTR
 */
void
ReadTOP_ATTR(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( t )
	{
		if ( cvt->tok.type != TOK_TST )
			t->flags.top_attr = FALSE;
		else
			t->flags.top_attr = cvt->tok.id;
	}
}

/****************************************************************************
 *
 * module:	ReadISA_BODY
 */
void
ReadISA_BODY(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( t )
	{
		if ( cvt->tok.type != TOK_TST )
			t->flags.isa_body = FALSE;
		else
			t->flags.isa_body = cvt->tok.id;
	}
}

/****************************************************************************
 *
 * module:	ReadISA_TITLE
 */
void
ReadISA_TITLE(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( t )
	{
		if ( cvt->tok.type != TOK_TST )
			t->flags.isa_title = FALSE;
		else
			t->flags.isa_title = cvt->tok.id;
	}
}

/****************************************************************************
 *
 * module:	ReadISA_SUBTITLE
 */
void
ReadISA_SUBTITLE(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( t )
	{
		if ( cvt->tok.type != TOK_TST )
			t->flags.isa_subtitle = FALSE;
		else
			t->flags.isa_subtitle = cvt->tok.id;
	}
}

/****************************************************************************
 *
 * module:	ReadISA_FOOTER
 */
void
ReadISA_FOOTER(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( t )
	{
		if ( cvt->tok.type != TOK_TST )
			t->flags.isa_footer = FALSE;
		else
			t->flags.isa_footer = cvt->tok.id;
	}
}

/****************************************************************************
 *
 * module:	ReadISA_LEGEND
 */
void
ReadISA_LEGEND(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( t )
	{
		if ( cvt->tok.type != TOK_TST )
			t->flags.isa_legend = FALSE;
		else
			t->flags.isa_legend = cvt->tok.id;
	}
}


/****************************************************************************
 *
 * module:	ReadISA_AXIS
 */
void
ReadISA_AXIS(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( t )
	{
		if ( cvt->tok.type != TOK_TST )
			t->flags.isa_axis = FALSE;
		else
			t->flags.isa_axis = cvt->tok.id;
	}
}


/****************************************************************************
 *
 * module:	ReadISA_AXLAB
 */
void
ReadISA_AXLAB(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( t )
	{
		if ( cvt->tok.type != TOK_TST )
			t->flags.isa_axlab = FALSE;
		else
			t->flags.isa_axlab = cvt->tok.id;
	}
}

/****************************************************************************
 *
 * module:	ReadISA_DATGRP
 */
void
ReadISA_DATGRP(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( t )
	{
		if ( cvt->tok.type != TOK_TST )
			t->flags.isa_datgrp = FALSE;
		else
			t->flags.isa_datgrp = cvt->tok.id;
	}
}
/****************************************************************************
 *
 * module:	ReadISA_DATLAB
 */
void
ReadISA_DATLAB(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( t )
	{
		if ( cvt->tok.type != TOK_TST )
			t->flags.isa_datlab = FALSE;
		else
			t->flags.isa_datlab = cvt->tok.id;
	}
}

/****************************************************************************
 *
 * module:	ReadISA_GRID
 */
void
ReadISA_GRID(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( t )
	{
		if ( cvt->tok.type != TOK_TST )
			t->flags.isa_grid = FALSE;
		else
			t->flags.isa_grid = cvt->tok.id;
	}
}


/****************************************************************************
 *
 * module:	ReadSTART_ANGLE
 */
void
ReadSTART_ANGLE(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->startAngle = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:	ReadEND_ANGLE
 */
void
ReadEND_ANGLE(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->endAngle = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:	ReadNSIDES
 */
void
ReadNSIDES(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->nsides = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:	ReadTEMPLATE_FLAGS
 */
void
ReadTEMPLATE_FLAGS(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->temp_type     = cvt->tok.val[0].u.ival;
	cvt->temp_size     = cvt->tok.val[1].u.ival;
	cvt->temp_inactive = cvt->tok.val[2].u.ival;
	cvt->temp_chart_ix = cvt->tok.val[3].u.ival;
}

/****************************************************************************
 *
 * module:	ReadAREA
 */
void
ReadAREA(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->temp_area.x1 = cvt->tok.val[0].u.ival;
	cvt->temp_area.y1 = cvt->tok.val[1].u.ival;
	cvt->temp_area.x2 = cvt->tok.val[2].u.ival;
	cvt->temp_area.y2 = cvt->tok.val[3].u.ival;
}

void
ReadT_POS(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->temp_area.x1 = cvt->tok.val[0].u.ival + PAGE_ZEROX;
	cvt->temp_area.y1 = cvt->tok.val[1].u.ival + PAGE_ZEROY;
	cvt->temp_area.x2 = cvt->tok.val[2].u.ival + PAGE_ZEROX;
	cvt->temp_area.y2 = cvt->tok.val[3].u.ival + PAGE_ZEROY;
}

/****************************************************************************
 *
 * module:	ReadPART
 */
void
ReadPART(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	obj->done = TRUE;
	cvt->untok = TRUE;
}

/****************************************************************************
 *
 * module:	ReadHYPERLINKS
 */
void
ReadHYPERLINKS(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	obj->done = TRUE;
	cvt->untok = TRUE;
}

/****************************************************************************
 *
 * module:	ReadPICTURE - end of Palette? assume so...
 */
void
ReadPICTURE(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	obj->done = TRUE;
	cvt->untok = TRUE;
}

/****************************************************************************
 *
 * module:	ReadBEGIN
 */
void
ReadBEGIN(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	/* no-op */
}

/****************************************************************************
 *
 * module:	ReadEND
 */
void
ReadEND(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	/* END GROUP? */
	obj->done = TRUE;

	/* ENDing something other than a GROUP? Then untok the END */
	if ( t->type != GROUP )
		cvt->untok = TRUE;

	/* ENDing a GROUP? (any boolean algebra gurus?) */
	else
	{
		mlInTok(ge, cvt, FALSE);
		if ( cvt->tok.type == TOK_OTY && cvt->tok.id != GROUP )
			mlInError(ge, cvt, ML_GENDEXP);

		else if ( cvt->tok.type == TOK_TOK && cvt->tok.id != TCHART )
			mlInError(ge, cvt, ML_GENDEXP);

		else if ( cvt->tok.type != TOK_OTY && cvt->tok.type != TOK_TOK )
			mlInError(ge, cvt, ML_GENDEXP);
	}
}

/****************************************************************************
 *
 * module:	ReadSTYLE
 */
void
ReadSTYLE(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( cvt->tok.type == TOK_NUM )
		cvt->curr_atts.line_style = cvtTokToInt(ge, cvt);
	else if ( cvt->tok.type == TOK_AUX )
		cvt->curr_atts.line_style = cvt->tok.id;
	else
		mlInError(ge, cvt, ML_PAVEXP);
}

/****************************************************************************
 *
 * module:	ReadSTYS
 */
void
ReadSTYS(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->curr_atts.style = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:	ReadARROW
 */
void
ReadARROW(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	int curr_arrow;

	mlInTok(ge, cvt, FALSE);
	if ( cvt->tok.type != TOK_AUX )
		mlInError(ge, cvt, ML_PAVEXP);
	else switch ( cvt->tok.id )
	{
		default:
			cvt->curr_atts.left_arrow = cvt->curr_atts.right_arrow = 0;
		break;

		case 0:
			if ( cvt->curr_atts.left_arrow )
				curr_arrow = cvt->curr_atts.left_arrow;
			if ( cvt->curr_atts.right_arrow )
				curr_arrow = cvt->curr_atts.right_arrow;
			cvt->curr_atts.left_arrow = 0;
			cvt->curr_atts.right_arrow = (curr_arrow) ? curr_arrow : 1;
		break;

		case 1:
			if ( cvt->curr_atts.left_arrow )
				curr_arrow = cvt->curr_atts.left_arrow;
			if ( cvt->curr_atts.right_arrow )
				curr_arrow = cvt->curr_atts.right_arrow;
			cvt->curr_atts.left_arrow = (curr_arrow) ? curr_arrow : 1;
			cvt->curr_atts.right_arrow = 0;
		break;

		case 3:
			if ( cvt->curr_atts.left_arrow )
				curr_arrow = cvt->curr_atts.left_arrow;
			if ( cvt->curr_atts.right_arrow )
				curr_arrow = cvt->curr_atts.right_arrow;
			cvt->curr_atts.left_arrow =  (curr_arrow) ? curr_arrow : 1;
			cvt->curr_atts.right_arrow = (curr_arrow) ? curr_arrow : 1;
		break;
	}
}

/****************************************************************************
 *
 * module:	ReadARROW_STYLE
 */
void
ReadARROW_STYLE(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( cvt->tok.type != TOK_AUX )
		mlInError(ge, cvt, ML_PAVEXP);
	else
	{
		if ( cvt->curr_atts.left_arrow )
			cvt->curr_atts.left_arrow = cvt->tok.id+1;
		if ( cvt->curr_atts.right_arrow )
			cvt->curr_atts.right_arrow = cvt->tok.id+1;
	}
}

/****************************************************************************
 *
 * module:	ReadARTYP
 */
void
ReadARTYP(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	ReadARROW_STYLE(ge, cvt, t, obj);
}

/****************************************************************************
 *
 * module:	ReadMARKER
 */
void
ReadMARKER(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( cvt->tok.type == TOK_NUM )
		cvt->curr_atts.join_symbols = cvtTokToInt(ge, cvt);
	if ( cvt->tok.type == TOK_TST )
	{
		if ( cvt->info.version < 300 && cvt->tok.id == 0 )
			cvt->curr_atts.right_arrow = cvt->curr_atts.left_arrow = 0;
		else
			cvt->curr_atts.join_symbols = cvt->tok.id;
	}
	else
		mlInError(ge, cvt, ML_PAVEXP);
}

/****************************************************************************
 *
 * module:	ReadL_ARROW
 */
void
ReadL_ARROW(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( cvt->tok.type == TOK_TST )
		cvt->curr_atts.left_arrow = cvt->tok.id;
	else if ( cvt->tok.type == TOK_NUM )
		cvt->curr_atts.left_arrow = cvtTokToInt(ge, cvt);
	else
		mlInError(ge, cvt, ML_PAVEXP);
}

/****************************************************************************
 *
 * module:	ReadLARROW
 */
void
ReadLARROW(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	ReadL_ARROW(ge, cvt, t, obj);
}

/****************************************************************************
 *
 * module:	ReadR_ARROW
 */
void
ReadR_ARROW(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( cvt->tok.type == TOK_TST )
		cvt->curr_atts.right_arrow = cvt->tok.id;
	else if ( cvt->tok.type == TOK_NUM )
		cvt->curr_atts.right_arrow = cvtTokToInt(ge, cvt);
	else
		mlInError(ge, cvt, ML_PAVEXP);
}

/****************************************************************************
 *
 * module:	ReadRARROW
 */
void
ReadRARROW(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	ReadR_ARROW(ge, cvt, t, obj);
}

/****************************************************************************
 *
 * module:	ReadTHICKNESS
 */
void
ReadTHICKNESS(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( cvt->tok.type == TOK_NUM )
		cvt->curr_atts.line_thickness = cvtTokToInt(ge, cvt);
	else  if ( cvt->tok.type != TOK_AUX )
		mlInError(ge, cvt, ML_PAVEXP);
	else
	{
		if ( cvt->tok.id == LTHS )
			cvt->curr_atts.line_thickness = 1;
		else
			cvt->curr_atts.line_thickness = lth2val[cvt->tok.id];
	}
}

/****************************************************************************
 *
 * module:	ReadTHRU
 */
void
ReadTHRU(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( cvt->tok.type == TOK_TST )
		cvt->curr_atts.strike = cvt->tok.id;
	else if ( cvt->tok.type == TOK_NUM )
		cvt->curr_atts.strike = cvtTokToInt(ge, cvt);
	else
		mlInError(ge, cvt, ML_TAVEXP);
}

/****************************************************************************
 *
 * module:	ReadPATTERN
 */
void
ReadPATTERN(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInPattern(ge, TPATTERN, cvt);
}


/****************************************************************************
 *
 * module:	ReadLINE_PATTERN
 */
void
ReadLINE_PATTERN(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInPattern(ge, TLINE_PATTERN, cvt);
}

void
ReadBACKFILL(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	/* fg, bg, type, id, angle, xoffset, yoffset */
	mlInTok(ge, cvt, FALSE);
	cvt->curr_atts.fill_fg   = cvt->tok.val[0].u.ival;
	cvt->curr_atts.fill_bg   = cvt->tok.val[1].u.ival;
	cvt->curr_atts.fpat_type = cvt->tok.val[2].u.ival;
	cvt->curr_atts.fill_pat  = cvt->tok.val[3].u.ival;
}

void
ReadLINEFILL(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	/* fg, bg, type, id, angle, xoffset, yoffset */
	mlInTok(ge, cvt, FALSE);
	cvt->curr_atts.line_fg   = cvt->tok.val[0].u.ival;
	cvt->curr_atts.line_bg   = cvt->tok.val[1].u.ival;
	cvt->curr_atts.lpat_type = cvt->tok.val[2].u.ival;
	cvt->curr_atts.line_pat  = cvt->tok.val[3].u.ival;
}

/****************************************************************************
 *
 * module:	ReadFILL_PAT
 */
void
ReadFILL_PAT(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->curr_atts.fill_pat = cvtTokToInt(ge, cvt);
	mlInTok(ge, cvt, FALSE);
	cvt->curr_atts.fpat_type = cvtTokToInt(ge, cvt);
	mlInTok(ge, cvt, FALSE);
	mlInTok(ge, cvt, FALSE);
	mlInTok(ge, cvt, FALSE);
}

/****************************************************************************
 *
 * module:	ReadLINE_PAT
 */
void
ReadLINE_PAT(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->curr_atts.line_pat = cvtTokToInt(ge, cvt);
	mlInTok(ge, cvt, FALSE);
	cvt->curr_atts.lpat_type = cvtTokToInt(ge, cvt);
	mlInTok(ge, cvt, FALSE);
	mlInTok(ge, cvt, FALSE);
	mlInTok(ge, cvt, FALSE);
}

/****************************************************************************
 *
 * module:	ReadLINE_COLOR
 */
void
ReadLINE_COLOR(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	/* read the next token and toss it */
	mlInTok(ge, cvt, FALSE);
}

/****************************************************************************
 *
 * module:	ReadREVERSE
 */
void
ReadREVERSE(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	/* read the next token and toss it */
	mlInTok(ge, cvt, FALSE);
}

/****************************************************************************
 *
 * module:	ReadTRANSPARENT
 */
void
ReadTRANSPARENT(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( STREQ(cvt->buf, tst[1]) )
		cvt->trans = TRUE;
	else
		cvt->trans = FALSE;
}

/****************************************************************************
 *
 * module:	ReadBACKGROUND_COLOR
 */
void
ReadBACKGROUND_COLOR(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	int val;

	mlInTok(ge, cvt, FALSE);
	if ( cvt->tok.type != TOK_NUM )
		cvt->curr_atts.fill_bg = 2;
	else
	{
		if ( cvt->tok.n_vals <= 0 )
			val = 0;
		else
		{
			if ( cvt->tok.val[0].type == 0 )
				val = cvt->tok.val[0].u.ival;
			else if ( cvt->tok.val[0].type == 1 )
				val = cvt->tok.val[0].u.fval;
			else
				val = 0;
		}

		if ( cvt->info.version < 300 )
			cvt->curr_atts.fill_bg = clr2val[val];
		else
			cvt->curr_atts.fill_bg = val;
	}
}

/****************************************************************************
 *
 * module:	ReadBACK_COLOR
 */
void
ReadBACK_COLOR(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	ReadBACKGROUND_COLOR(ge, cvt, t, obj);
}

/****************************************************************************
 *
 * module:	ReadFOREGROUND_COLOR
 */
void
ReadFOREGROUND_COLOR(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	int val;

	mlInTok(ge, cvt, FALSE);
	if ( cvt->tok.type != TOK_NUM )
		cvt->curr_atts.fill_fg = 1;
	else
	{
		if ( cvt->tok.n_vals <= 0 )
			val = 1;
		else
		{
			if ( cvt->tok.val[0].type == 0 )
				val = cvt->tok.val[0].u.ival;
			else if ( cvt->tok.val[0].type == 1 )
				val = cvt->tok.val[0].u.fval;
			else
				val = 1;
		}

		if ( cvt->info.version < 300 )
			cvt->curr_atts.fill_fg = clr2val[val];
		else
			cvt->curr_atts.fill_fg = val;
	}
}

/****************************************************************************
 *
 * module:	ReadFORE_COLOR
 */
void
ReadFORE_COLOR(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	ReadFOREGROUND_COLOR(ge, cvt, t, obj);
}

/****************************************************************************
 *
 * module:	ReadLINE_BACKGROUND_COLOR
 */
void
ReadLINE_BACKGROUND_COLOR(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	int val;

	mlInTok(ge, cvt, FALSE);
	if ( cvt->tok.type != TOK_NUM )
		cvt->curr_atts.line_bg = 2;
	else
	{
		if ( cvt->tok.n_vals <= 0 )
			val = 0;
		else
		{
			if ( cvt->tok.val[0].type == 0 )
				val = cvt->tok.val[0].u.ival;
			else if ( cvt->tok.val[0].type == 1 )
				val = cvt->tok.val[0].u.fval;
			else
				val = 0;
		}

		if ( cvt->info.version < 300 )
			cvt->curr_atts.line_bg = clr2val[val];
		else
			cvt->curr_atts.line_bg = val;
	}
}

/****************************************************************************
 *
 * module:	ReadLINE_BACK_COLOR
 */
void
ReadLINE_BACK_COLOR(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	ReadLINE_BACKGROUND_COLOR(ge, cvt, t, obj);
}

/****************************************************************************
 *
 * module:	ReadLINE_FOREGROUND_COLOR
 */
void
ReadLINE_FOREGROUND_COLOR(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	int val;

	mlInTok(ge, cvt, FALSE);
	if ( cvt->tok.type != TOK_NUM )
		cvt->curr_atts.line_fg = 1;
	else
	{
		if ( cvt->tok.n_vals <= 0 )
			val = 1;
		else
		{
			if ( cvt->tok.val[0].type == 0 )
				val = cvt->tok.val[0].u.ival;
			else if ( cvt->tok.val[0].type == 1 )
				val = cvt->tok.val[0].u.fval;
			else
				val = 1;
		}

		if ( cvt->info.version < 300 )
			cvt->curr_atts.line_fg = clr2val[val];
		else
			cvt->curr_atts.line_fg = val;
	}
}

/****************************************************************************
 *
 * module:	ReadLINE_FORE_COLOR
 */
void
ReadLINE_FORE_COLOR(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	ReadLINE_FOREGROUND_COLOR(ge, cvt, t, obj);
}

/****************************************************************************
 *
 * module:	ReadCOLORMAP
 */
void
ReadCOLORMAP(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( cvt->tok.type == TOK_NUM )
	{
		if ( cvt->tok.val[0].type == 0 )
			cvt->cmapID = cvt->tok.val[0].u.ival;
		else if ( cvt->tok.val[0].type == 1 )
			cvt->cmapID = cvt->tok.val[0].u.fval;
	}
}

/****************************************************************************
 *
 * module:	ReadSHADOW
 */
void
ReadSHADOW(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	if ( cvt->info.version < 320 )
	{
		mlInTok(ge, cvt, FALSE);
		cvt->curr_atts.shadow = cvtTokToInt(ge, cvt);
	}
	else /* color, type xoff, yoff */
	{
		mlInTok(ge, cvt, FALSE);
		cvt->curr_atts.shadow_color = cvt->tok.val[0].u.ival;
		cvt->curr_atts.shadow       = cvt->tok.val[1].u.ival;
		cvt->curr_atts.shadow_x     = cvt->tok.val[2].u.ival;
		cvt->curr_atts.shadow_y     = cvt->tok.val[3].u.ival;
	}
}

/****************************************************************************
 *
 * module:	ReadSHADOWCOLOR
 */
void
ReadSHADOWCOLOR(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->curr_atts.shadow_color = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:	ReadSHADOWSIZE
 */
void
ReadSHADOWSIZE(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->curr_atts.shadow_x = cvtTokToInt(ge, cvt);
	mlInTok(ge, cvt, FALSE);
	cvt->curr_atts.shadow_y = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:	R e a d P A R A
 */
void
ReadPARA(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
cvt_part_creation_type *obj;
thing *t;
{
	mlInTok(ge, cvt, FALSE);
	cvt->curr_atts.l_indent1    = cvt->tok.val[0].u.ival;
	cvt->curr_atts.l_indent2    = cvt->tok.val[1].u.ival;
	cvt->curr_atts.r_indent     = cvt->tok.val[2].u.ival;
	cvt->curr_atts.bullet       = cvt->tok.val[3].u.ival;
	cvt->curr_atts.bullet_scale = cvt->tok.val[4].u.ival;
	cvt->curr_atts.bullet_color = cvt->tok.val[5].u.ival;
	cvt->curr_atts.bullet_font  = cvt->tok.val[6].u.ival;
	cvt->curr_atts.bullet_glyph = cvt->tok.val[7].u.ival;
}

/****************************************************************************
 *
 * module:	ReadV_SPACE
 */
ReadV_SPACE(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
cvt_part_creation_type *obj;
thing *t;
{
	mlInTok(ge, cvt, FALSE);
	cvt->curr_atts.space_above  = cvt->tok.val[0].u.ival;
	cvt->curr_atts.space_inside = cvt->tok.val[1].u.ival;
	cvt->curr_atts.space_below  = cvt->tok.val[2].u.ival;
}

/****************************************************************************
 *
 * module:	ReadLAYERS
 */
ReadLAYERS(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->curr_layer = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:	ReadFONT
 */
ReadFONT(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	int face_ix;

	mlInTok(ge, cvt, FALSE);
	if ( cvt->tok.type == TOK_NUM )
	{
		if ( cvt->tok.val[0].type == 0 )
			face_ix = cvt->tok.val[0].u.ival;
		else if ( cvt->tok.val[0].type == 1 )
			face_ix = cvt->tok.val[0].u.fval;
		else
			face_ix = CvtAddFont(ge,cvt,fnt[0]);
	}
	else
	{
		if ( cvt->tok.type == TOK_AUX )
			face_ix = CvtAddFont(ge,cvt,fnt[cvt->tok.id]);

		else /* fake it */
		{
			face_ix = CvtAddFont(ge,cvt,fnt[0]);
			mlInError(ge, cvt, ML_UNKTAV);
		}
	}
	if ( face_ix >= ge->n_faces )
		cvt->curr_atts.face = 0;
	else
		cvt->curr_atts.face = face_ix;
}

/****************************************************************************
 *
 * module:	ReadSIZE
 */
ReadSIZE(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	int size;

	mlInTok(ge, cvt, FALSE);
	if ( cvt->tok.type == TOK_NUM )
	{
		if ( cvt->tok.val[0].type == 0 )
			size = cvt->tok.val[0].u.ival;
		else if ( cvt->tok.val[0].type == 1 )
			size = cvt->tok.val[0].u.fval;
		else
			size = 180000;
	}
	else
	{
		if ( cvt->tok.type == TOK_AUX )
			size = siz2val1[cvt->tok.id];
		else
		{
			size = 180000;
			mlInError(ge, cvt, ML_UNKTAV);
		}
	}

	if ( size <= 0 )
		size = 180000;
	cvt->curr_atts.size = size;
}

/****************************************************************************
 *
 * module:	ReadHORIZONTAL_JUSTIFICATION
 */
ReadHORIZONTAL_JUSTIFICATION(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	int h_just;

	mlInTok(ge, cvt, FALSE);
	if ( cvt->tok.type == TOK_TST )
		h_just = cvt->tok.id;
	else if ( cvt->tok.type == TOK_NUM )
		h_just = cvtTokToInt(ge, cvt);
	else if ( cvt->tok.type == TOK_AUX )
		h_just = cvt->tok.id;
	else if ( cvt->tok.type == TOK_TOK && cvt->tok.id == TCENTER )
		h_just = 1;
	else
	{
		mlInError(ge, cvt, ML_UNKTAV);
		h_just = 0;
	}
	cvt->curr_atts.h_just = h_just;
}

/****************************************************************************
 *
 * module:	ReadHOR_JUST
 */
ReadHOR_JUST(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	ReadHORIZONTAL_JUSTIFICATION(ge, cvt, t, obj);
}

/****************************************************************************
 *
 * module:	ReadUNDERLINE
 */
ReadUNDERLINE(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	int ruling;

	mlInTok(ge, cvt, FALSE);
	if ( cvt->tok.type == TOK_TST )
		ruling = cvt->tok.id;
	else if ( cvt->tok.type == TOK_NUM )
		ruling = cvtTokToInt(ge, cvt);
	else if ( cvt->tok.type == TOK_AUX )
		ruling = cvt->tok.id;
	else
	{
		mlInError(ge, cvt, ML_UNKTAV);
		ruling = 0;
	}
	if ( ruling & 0x4 )
		cvt->curr_atts.strike = TRUE;
	cvt->curr_atts.ruling = ruling;
}

/****************************************************************************
 *
 * module:	ReadBOLD
 */
ReadBOLD(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( cvt->tok.type == TOK_TST )
		cvt->curr_atts.weight = cvt->tok.id;
	else if ( cvt->tok.type == TOK_NUM )
		cvt->curr_atts.weight = cvtTokToInt(ge, cvt);
	else
		mlInError(ge, cvt, ML_TAVEXP);
}

/****************************************************************************
 *
 * module:	ReadITALICS
 */
ReadITALICS(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( cvt->tok.type == TOK_TST )
		cvt->curr_atts.slant = cvt->tok.id;
	else if ( cvt->tok.type == TOK_NUM )
		cvt->curr_atts.slant = cvtTokToInt(ge, cvt);
	else
		mlInError(ge, cvt, ML_TAVEXP);
}

/****************************************************************************
 *
 * module:	ReadWIDECHARS
 */
ReadWIDECHARS(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	cvt->widechars = TRUE;
}

/****************************************************************************
 *
 * module:	ReadANGLE
 */
ReadANGLE(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->curr_atts.angle = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:	ReadCBACK
 */
ReadCBACK(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->curr_atts.callback = cvtTokToInt(ge, cvt)+1;
}

/****************************************************************************
 *
 * module:	ReadSCALE
 */
ReadSCALE(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->curr_atts.x_scale = cvtTokToInt(ge, cvt);
	if ( cvt->curr_atts.x_scale == 0 )
		cvt->curr_atts.x_scale = 1;
}

/****************************************************************************
 *
 * module:	ReadSCALE_CORNERS
 */
ReadSCALE_CORNERS(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if (t && (t->type == RECT))
		cvt->scale_corners = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:	ReadSHEAR
 */
ReadSHEAR(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->curr_atts.shear = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:	ReadMASK
 */
ReadMASK(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInImgMask(ge, t, cvt);
}


/****************************************************************************
 *
 * module:	ReadDATA
 */
ReadDATA(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	/* can't see DATA twice */
	if ( obj->sawData )
		mlInError(ge, cvt, ML_DATANEXP);
	else
	{
		obj->sawData = TRUE;
		mlInImgData(ge,cvt,t,obj);
	}
}

/****************************************************************************
 *
 * module:	ReadFORMAT
 */
ReadFORMAT(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
}

/****************************************************************************
 *
 * module:	ReadWIDTH
 */
ReadWIDTH(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->width = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:	ReadHEIGHT
 */
ReadHEIGHT(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->height = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:	ReadDEPTH
 */
ReadDEPTH(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->depth = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:	ReadL_WIDTHS
 */
ReadL_WIDTHS(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	ge->state.line_1 = cvt->tok.val[0].u.ival;
	ge->state.line_2 = cvt->tok.val[1].u.ival;
	ge->state.line_3 = cvt->tok.val[2].u.ival;
}

/****************************************************************************
 *
 * module:	ReadLINE_SPACE
 */
ReadLINE_SPACE(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	int space;
	mlInTok(ge, cvt, FALSE);
	space = cvtTokToInt(ge, cvt);
	if ( space <= 0 )
		space = 18;
	cvt->curr_atts.field_space  = space * 1000.0 / 72.0 + 0.5;
	cvt->curr_atts.space_inside = cvt->curr_atts.field_space;
	cvt->curr_atts.space_above  = 0;
	cvt->curr_atts.space_below  = 0;
}

/****************************************************************************
 *
 * module:	ReadL_SPACE
 */
ReadL_SPACE(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->curr_atts.field_space = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:	ReadMARGINS
 */
ReadMARGINS(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	if ( cvt->info.version < 320 )
	{
		mlInTok(ge, cvt, FALSE);
		cvt->curr_atts.l_margin = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		cvt->curr_atts.r_margin = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		cvt->curr_atts.t_margin = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		cvt->curr_atts.b_margin = cvtTokToInt(ge, cvt);
	}
	else
	{
		mlInTok(ge, cvt, FALSE);
		cvt->curr_atts.l_margin = cvt->tok.val[0].u.ival;
		cvt->curr_atts.r_margin = cvt->tok.val[1].u.ival;
		cvt->curr_atts.t_margin = cvt->tok.val[2].u.ival;
		cvt->curr_atts.b_margin = cvt->tok.val[3].u.ival;
	}
}

/****************************************************************************
 *
 * module:	ReadSUBSUP
 */
ReadSUBSUP(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	if ( cvt->info.version < 320 )
	{
		mlInTok(ge, cvt, FALSE);
		mlInTok(ge, cvt, FALSE);
		cvt->curr_atts.y_sub_sup = cvtTokToInt(ge, cvt);
	}
	else
	{
		mlInTok(ge, cvt, FALSE);
		cvt->curr_atts.y_sub_sup = cvt->tok.val[1].u.ival;
	}
}

/****************************************************************************
 *
 * module:	ReadVERTICAL_JUSTIFICATION
 */
ReadVERTICAL_JUSTIFICATION(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	int v_just;

	mlInTok(ge, cvt, FALSE);
	if ( cvt->tok.type == TOK_TST )
		v_just = cvt->tok.id;
	else if ( cvt->tok.type == TOK_NUM )
		v_just = cvtTokToInt(ge, cvt);
	else if ( cvt->tok.type == TOK_AUX )
		v_just = cvt->tok.id;
	else
	{
		mlInError(ge, cvt, ML_UNKTAV);
		v_just = 0;
	}
	cvt->curr_atts.v_just = v_just;
}

/****************************************************************************
 *
 * module:	ReadVER_JUST
 */
ReadVER_JUST(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	ReadVERTICAL_JUSTIFICATION(ge, cvt, t, obj);
}

/****************************************************************************
 *
 * module:	ReadXYOFFS
 */
ReadXYOFFS(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	if ( cvt->info.version < 320 )
	{
		mlInTok(ge, cvt, FALSE);
		cvt->curr_atts.x_off = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
	}
	else
	{
		mlInTok(ge, cvt, FALSE);
		cvt->curr_atts.x_off = cvt->tok.val[0].u.ival;
	}
}

/****************************************************************************
 *
 * module:	ReadTXTXYOFF
 */
ReadTXTXYOFF(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	if ( cvt->info.version < 320 )
	{
		mlInTok(ge, cvt, FALSE);
		cvt->x_off = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		cvt->y_off = cvtTokToInt(ge, cvt);
	}
	else
	{
		mlInTok(ge, cvt, FALSE);
		cvt->x_off = cvt->tok.val[0].u.ival;
		cvt->y_off = cvt->tok.val[1].u.ival;
	}
}

/****************************************************************************
 *
 * module:	ReadXYRAD
 */
ReadXYRAD(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	int ok;

	ok = (t && t->type == RECT) ? TRUE : FALSE;
	if ( cvt->info.version < 320 )
	{
		mlInTok(ge, cvt, FALSE);
		if ( ok )
			cvt->x_corner_radius = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		if ( ok )
			cvt->y_corner_radius = cvtTokToInt(ge, cvt);
	}
	else
	{
		mlInTok(ge, cvt, FALSE);
		if ( ok )
		{
			cvt->x_corner_radius = cvt->tok.val[0].u.ival;
			cvt->y_corner_radius = cvt->tok.val[1].u.ival;
		}
	}
}

/****************************************************************************
 *
 * module:	ReadXYRES
 */
ReadXYRES(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	cvt->xres = cvtTokToInt(ge, cvt);
	mlInTok(ge, cvt, FALSE);
	cvt->yres = cvtTokToInt(ge, cvt);
	obj->sawXYres = TRUE;
}

/****************************************************************************
 *
 * module:	ReadDAT
 */
ReadDAT(ge,cvt,dat)
graphics_main_type *ge;
cvt_global_type *cvt;
chart_data_point_type *dat;
{
	int ii, done;

	if ( cvt->info.version < 320 )
	{
		mlInTok(ge, cvt, FALSE);
		switch ( cvt->tok.id ) /* assume it returned TOK_TOK */
		{
			default:
			case TPLOT:     dat->display_flag = BG_PLOT; break;
			case TSUBS:     dat->display_flag = BG_SUBS; break;
			case TSPAN:     dat->display_flag = BG_SPAN; break;
			case TGAP:      dat->display_flag = BG_GAP;  break;
			case TNULL_VAL: dat->display_flag = BG_NULL; break;
		}

		mlInTok(ge, cvt, FALSE);
		dat->label_str = cvtTokToStr(ge, cvt);

		if ( mlInTok(ge, cvt, FALSE) == TOK_TOK )
		{
			dat->local_x = FALSE;
			dat->label_x_off = 0;
		}
		else
		{
			dat->local_x = TRUE;
			dat->label_x_off = cvtTokToInt(ge, cvt);
		}
		if ( mlInTok(ge, cvt, FALSE) == TOK_TOK )
		{
			dat->local_y = FALSE;
			dat->label_y_off = 0;
		}
		else
		{
			dat->local_y = TRUE;
			dat->label_y_off = cvtTokToInt(ge, cvt);
		}
		if ( cvt->info.version < 311 )
		{
			if ( dat->local_x && dat->label_x_off == 0 )
				dat->local_x = FALSE;
			if ( dat->local_y && dat->label_y_off == 0 )
				dat->local_y = FALSE;
		}

		/* read the x's */
		ii = 0;
		while ( mlInTok(ge, cvt, FALSE) != TOK_TOK )
		{
			if ( ii < 5 )
				dat->x[ii] = cvtTokToFlt(ge, cvt);
			ii++;
		}
		dat->nx = ii;

		/* read the y's */
		ii = 0;
		while ( mlInTok(ge, cvt, FALSE) != TOK_TOK )
		{
			if ( ii < 5 )
				dat->y[ii] = cvtTokToFlt(ge, cvt);
			ii++;
		}
		dat->ny = ii;
	
		/* read the z's */
		ii = 0;
		while ( mlInTok(ge, cvt, FALSE) != TOK_TOK )
		{
			if ( ii < 5 )
				dat->z[ii] = cvtTokToFlt(ge, cvt);
			ii++;
		}
		dat->nz = ii;
	}
	else
	{
		/* read the plot and label data */
		mlInTok(ge, cvt, FALSE);
		dat->display_flag = cvt->tok.val[0].u.ival;
		dat->label_str = cvtGetStrFromTokVal(ge, cvt, 1);
		if ( cvt->tok.val[2].type == 2 )
		{
			dat->local_x = FALSE;
			dat->label_x_off = 0;
		}
		else
		{
			dat->local_x = TRUE;
			dat->label_x_off = cvt->tok.val[2].u.ival;
		}
		if ( cvt->tok.val[3].type == 2 )
		{
			dat->local_y = FALSE;
			dat->label_y_off = 0;
		}
		else
		{
			dat->local_y = TRUE;
			dat->label_y_off = cvt->tok.val[3].u.ival;
		}

		done = FALSE;
		while ( !done && mlInTok(ge, cvt, FALSE) == TOK_TOK )
		{
			switch ( cvt->tok.id )
			{
				case TX:
					mlInTok(ge, cvt, FALSE);
					for ( ii=0; ii<cvt->tok.n_vals && ii < 5; ii++ )
						dat->x[ii] = cvt->tok.val[ii].u.fval;
					dat->nx = MIN(5, cvt->tok.n_vals);
				break;

				case TY:
					mlInTok(ge, cvt, FALSE);
					for ( ii=0; ii<cvt->tok.n_vals && ii < 5; ii++ )
						dat->y[ii] = cvt->tok.val[ii].u.fval;
					dat->ny = MIN(5, cvt->tok.n_vals);
				break;

				case TZ:
					mlInTok(ge, cvt, FALSE);
					for ( ii=0; ii<cvt->tok.n_vals && ii < 5; ii++ )
						dat->z[ii] = cvt->tok.val[ii].u.fval;
					dat->nz = MIN(5, cvt->tok.n_vals);
				break;

				default:
					done = TRUE;
					cvt->untok = TRUE;
				break;
			}
		}
	}
}

/****************************************************************************
 *
 * module:ReadDataGroup
 */
ReadDataGroup(ge, cvt, grp)
graphics_main_type *ge;
cvt_global_type *cvt;
chart_data_group_type *grp;
{
	chart_data_point_type *dat;
	int ndatgps, size, done, ix;
	char *str;

	/* read the name and the active flag */
	if ( cvt->info.version < 320 )
	{
		mlInTok(ge, cvt, FALSE);
		str = (char *)cvtTokToStr(ge, cvt);
		if ( !str )
			grp->name[0] = (char)0;
		else
		{
			strcpy(grp->name, str);
			gfree(ge, str);
		}

		mlInTok(ge, cvt, FALSE);
		grp->disabled = cvt->tok.id ? FALSE : TRUE;
	}
	else
	{
		mlInTok(ge, cvt, FALSE);
		str = cvtGetStrFromTokVal(ge, cvt, 0);
		if ( !str )
			grp->name[0] = (char)0;
		else
		{
			strcpy(grp->name, str);
			gfree(ge, str);
		}
		grp->disabled = cvt->tok.val[1].u.ival;
	}

	/* read the rest */
	done = FALSE;
	while ( !done && mlInTok(ge, cvt, TRUE) )
	{
		if ( cvt->tok.type != TOK_TOK )
			continue;

		switch ( cvt->tok.id )
		{
			default:
				(TokFuncs[cvt->tok.id])(ge, cvt, NULL, NULL);
			break;

			case TLABEL:
				mlInTok(ge, cvt, FALSE);
				grp->legend_str = cvtTokToStr(ge, cvt);
			break;

			case TDATATYPE:
				if ( cvt->info.version < 320 )
				{
					mlInTok(ge, cvt, FALSE);
					grp->type = cvtTokToInt(ge, cvt);
					mlInTok(ge, cvt, FALSE);
					grp->stack_id = cvtTokToInt(ge, cvt);
					mlInTok(ge, cvt, FALSE);
					grp->part = cvtTokToStr(ge, cvt);
				}
				else
				{
					mlInTok(ge, cvt, FALSE);
					grp->type = cvt->tok.val[0].u.ival;
					grp->stack_id = cvt->tok.val[1].u.ival;
					grp->part = cvtGetStrFromTokVal(ge, cvt, 2);
					if ( cvt->tok.n_vals >= 4 )
					{	
						grp->new    = cvt->tok.val[3].u.ival;
					    grp->old_ix = cvt->tok.val[4].u.ival;
					}
					if ( cvt->tok.n_vals >= 5 )
						grp->disabled    = cvt->tok.val[5].u.ival;
				}
			break;

			case TOLD_DATATYPE:
				mlInTok(ge, cvt, FALSE);
				/* grp->old_type =     */ cvtTokToInt(ge, cvt);
				/* grp->old_stack_id = */ cvtTokToInt(ge, cvt);
				/* grp->old_part =     */ cvtTokToStr(ge, cvt);
			break;

			case TAXES:
				if ( cvt->info.version < 320 )
				{
					mlInTok(ge, cvt, FALSE);
					grp->x_axis = cvtTokToInt(ge, cvt);
					mlInTok(ge, cvt, FALSE);
					grp->y_axis = cvtTokToInt(ge, cvt);
					mlInTok(ge, cvt, FALSE);
					grp->z_axis = cvtTokToInt(ge, cvt);
				}
				else
				{
					mlInTok(ge, cvt, FALSE);
					grp->x_axis = cvt->tok.val[0].u.ival;
					grp->y_axis = cvt->tok.val[1].u.ival;
					grp->z_axis = cvt->tok.val[2].u.ival;
				}
			break;

			case TDATALABEL:
				mlInTok(ge, cvt, FALSE);
				grp->data_label_type = cvt->tok.val[0].u.ival;
				grp->label_x_off = cvt->tok.val[1].u.ival;
				grp->label_y_off = cvt->tok.val[2].u.ival;
				grp->label_align = cvt->tok.val[3].u.ival;
			break;

			case TTREND:
                    mlInTok(ge, cvt, FALSE);
                    grp->trend_info.type  = cvt->tok.val[0].u.ival;
                    grp->trend_info.order = cvt->tok.val[1].u.ival;
                    grp->trend_info.r_squared = cvt->tok.val[2].u.fval;
                    grp->trend_info.n_correlations  = cvt->tok.val[3].u.ival;
				    for( ix = 0; ix < grp->trend_info.n_correlations; ix ++)
						grp->trend_info.correlations[ix] = 
							cvt->tok.val[4+ix].u.fval;
               break;

			case TDAT:
				ndatgps = grp->n++;
				size = grp->n * sizeof(chart_data_point_type);
				grp->dat = (chart_data_point_type*)regloc(ge,grp->dat,size);
				dat = &grp->dat[ndatgps];
				memset(dat,0,sizeof(chart_data_point_type));
				ReadDAT(ge,cvt,dat);
			break;

			case TNUM_FMT:
				ReadNUM_FMT(ge, cvt, &grp->label_frmt);
				if ( cvt->info.version < 320 )
				{
					mlInTok(ge, cvt, FALSE);
					grp->data_label_type = cvtTokToInt(ge, cvt);
					mlInTok(ge, cvt, FALSE);
					grp->label_x_off = cvtTokToInt(ge, cvt);
					mlInTok(ge, cvt, FALSE);
					grp->label_y_off = cvtTokToInt(ge, cvt);
					if ( cvt->info.version >= 311 )
					{
						mlInTok(ge, cvt, FALSE);
						grp->label_align = cvtTokToInt(ge, cvt);
					}
				}
			break;

			case TGRPLABATTR:
			break;

			case TGRPBUF:
				mlInTok(ge, cvt, FALSE);
				grp->enable_bufs = cvt->tok.val[0].u.ival;
				grp->n_bufs      = cvt->tok.val[1].u.ival;
			break;
				
			case TGRPATTR:
				grp->label_attr = cvt->curr_atts;
			break;

			case TEND:
				done = TRUE;
			break;
		}
	}

	/* must have gotten a TEND */
	grp->attr = cvt->curr_atts;
	mlInTok(ge, cvt, FALSE); /* DATGROUP */

	/* some quick repair */
	if ( cvt->info.version < 311 )
	{
		grp->label_align = grp->attr.h_just;
		grp->attr.h_just = 1;
	}
}

/****************************************************************************
 *
 * module:
 */
ReadDATGROUP(ge, cvt, t, chart)
graphics_main_type *ge;
cvt_global_type *cvt;
char *t;
chart_type *chart;
{
	chart_data_group_type *grp;
	int ngrps, size;

	ngrps = chart->n_groups++;
	size = chart->n_groups * sizeof(chart_data_group_type);
	chart->grp = (chart_data_group_type *)regloc(ge, chart->grp, size);
	grp = &chart->grp[ngrps];
	memset(grp, 0, sizeof(chart_data_group_type));
	ReadDataGroup(ge, cvt, grp);
}

/****************************************************************************
 *
 * module:
 */
ReadLEGEND(ge, cvt, t, chart)
graphics_main_type *ge;
cvt_global_type *cvt;
char *t;
chart_type *chart;
{
	chart_legend_type *legend;
	thing_attributes *att;
	int done;

	/* read the legend title */
	legend = &chart->legend;
	mlInTok(ge, cvt, FALSE);
	legend->title_str = cvtTokToStr(ge, cvt);

	if ( cvt->info.version < 320 )
	{
		/* read the legend position */
		mlInTok(ge, cvt, FALSE);
		legend->h_align = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		legend->h_off = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		legend->v_align = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		legend->v_off = cvtTokToInt(ge, cvt);
		if ( cvt->info.version >= 311 )
		{
			mlInTok(ge, cvt, FALSE);
			legend->title_align = cvtTokToInt(ge, cvt);
		}

		/* read the legend format */
		mlInTok(ge, cvt, FALSE);
		legend->disabled = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		legend->x_space_eating = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		legend->y_space_eating = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		legend->text_before_sample = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		legend->arranged_by_row = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		legend->max_per_row_or_column = cvtTokToInt(ge, cvt);

		/* read the legend margins */
		mlInTok(ge, cvt, FALSE);
		legend->t_margin = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		legend->b_margin = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		legend->l_margin = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		legend->r_margin = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		legend->row_margin = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		legend->column_margin = cvtTokToInt(ge, cvt);
		if ( cvt->info.version >= 311 )
		{
			mlInTok(ge, cvt, FALSE);
			legend->auto_margins = cvtTokToInt(ge, cvt);
		}
	}

	/* read the attributes of the various parts of the legend */
	att = (thing_attributes *)0;
	done = FALSE;
	while ( !done && mlInTok(ge, cvt, TRUE) )
	{
		/* assume tokens (fragile!) */
		if ( cvt->tok.type != TOK_TOK )
			continue;

		switch ( cvt->tok.id )
		{
			default:
				(TokFuncs[cvt->tok.id])(ge, cvt, t, NULL);
			break;
			case TLEGALIGN:
				mlInTok(ge, cvt, FALSE);
				legend->h_align = cvt->tok.val[0].u.ival;
				legend->h_off = cvt->tok.val[1].u.ival;
				legend->v_align = cvt->tok.val[2].u.ival;
				legend->v_off = cvt->tok.val[3].u.ival;
				legend->title_align = cvt->tok.val[4].u.ival;
			break;

			case TLEGFMT:
				mlInTok(ge, cvt, FALSE);
				legend->disabled = cvt->tok.val[0].u.ival;
				legend->x_space_eating = cvt->tok.val[1].u.ival;
				legend->y_space_eating = cvt->tok.val[2].u.ival;
				legend->text_before_sample = cvt->tok.val[3].u.ival;
				legend->arranged_by_row = cvt->tok.val[4].u.ival;
				legend->max_per_row_or_column = cvt->tok.val[5].u.ival;
			break;

			case TLEGMARG:
				mlInTok(ge, cvt, FALSE);
				legend->t_margin = cvt->tok.val[0].u.ival;
				legend->b_margin = cvt->tok.val[1].u.ival;
				legend->l_margin = cvt->tok.val[2].u.ival;
				legend->r_margin = cvt->tok.val[3].u.ival;
				legend->row_margin = cvt->tok.val[4].u.ival;
				legend->column_margin = cvt->tok.val[5].u.ival;
				legend->auto_margins = cvt->tok.val[6].u.ival;
			break;

			case TLEGBOXATTR:
				if ( att )
					memcpy(att,&cvt->curr_atts,sizeof(thing_attributes));
				att = &legend->box_attr;
			break;

			case TLEGLABATTR:
				if ( att )
					memcpy(att,&cvt->curr_atts,sizeof(thing_attributes));
				att = &legend->label_attr;
			break;

			case TLEGTITATTR:
				if ( att )
					memcpy(att,&cvt->curr_atts,sizeof(thing_attributes));
				att = &legend->title_attr;
			break;

			case TEND:
				done = TRUE;
			break;
		}
	}
	if ( att )
		memcpy(att,&cvt->curr_atts,sizeof(thing_attributes));
	mlInTok(ge, cvt, FALSE); /* LEGEND */

	/* some quick repair */
	if ( cvt->info.version < 311 )
	{
		legend->title_align = legend->title_attr.h_just;
		legend->title_attr.h_just = 1;
	}
}

/****************************************************************************
 *
 * module:	ReadNUM_FMT
 */
ReadNUM_FMT(ge,cvt, fmt)
graphics_main_type *ge;
cvt_global_type *cvt;
number_format *fmt;
{
	char *str;

	if ( cvt->info.version < 320 )
	{
		/* format info */
		mlInTok(ge, cvt, FALSE);
		fmt->id = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		fmt->aux_id = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		fmt->display_format_errors = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		fmt->trim_string = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		fmt->units = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		fmt->n_radix_places = cvtTokToInt(ge, cvt);

		/* numerical format strings */
		mlInTok(ge, cvt, FALSE);
		fmt->prefix = cvtTokToStr(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		fmt->suffix = cvtTokToStr(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		fmt->am = cvtTokToStr(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		fmt->pm = cvtTokToStr(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		fmt->true = cvtTokToStr(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		fmt->false = cvtTokToStr(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		str = cvtTokToStr(ge, cvt);
		if ( !str )
			fmt->t_sep = ',';
		else
		{
			fmt->t_sep = str[0];
			gfree(ge, str);
		}
		mlInTok(ge, cvt, FALSE);
		str = cvtTokToStr(ge, cvt);
		if ( !str )
			fmt->d_sep = ',';
		else
		{
			fmt->d_sep = str[0];
			gfree(ge, str);
		}
	}
	else
	{
		/* format info */
		mlInTok(ge, cvt, FALSE);
		fmt->id = cvt->tok.val[0].u.ival;
		fmt->aux_id = cvt->tok.val[1].u.ival;
		fmt->display_format_errors = cvt->tok.val[2].u.ival;
		fmt->trim_string = cvt->tok.val[3].u.ival;
		fmt->units = cvt->tok.val[4].u.ival;
		fmt->n_radix_places = cvt->tok.val[5].u.ival;

		/* format strings */
		fmt->prefix = cvtGetStrFromTokVal(ge, cvt, 6);
		fmt->suffix = cvtGetStrFromTokVal(ge, cvt, 7);
		fmt->am = cvtGetStrFromTokVal(ge, cvt, 8);
		fmt->pm = cvtGetStrFromTokVal(ge, cvt, 9);
		fmt->true = cvtGetStrFromTokVal(ge, cvt, 10);
		fmt->false = cvtGetStrFromTokVal(ge, cvt, 11);
		str = cvtGetStrFromTokVal(ge, cvt, 12);
		if ( !str )
			fmt->t_sep = ',';
		else
		{
			fmt->t_sep = str[0];
			gfree(ge, str);
		}
		str = cvtGetStrFromTokVal(ge, cvt, 13);
		if ( !str )
			fmt->d_sep = ',';
		else
		{
			fmt->d_sep = str[0];
			gfree(ge, str);
		}
	}
}


#ifdef ISA_VALUE
#undef ISA_VALUE
#define ISA_VALUE 0
#else
#define ISA_VALUE 0
#endif

#ifdef ISA_ARRAY
#undef ISA_ARRAY
#define ISA_ARRAY 1
#else
#define ISA_ARRAY 1
#endif
#define NUM_ELEMS_IN_ROW 10

ReadELEM(ge, cvt, item, indx)
graphics_main_type *ge;
cvt_global_type *cvt;
sheet_part_element *item;
int indx;
{
	item->type = ISA_VALUE;
	item->n    = 1;
	item->item = (sheet_part_element **) NULL;
	item->val  = cvtGetStrFromTokVal(ge, cvt, indx);
}


ReadARR(ge, cvt, item)
graphics_main_type *ge;
cvt_global_type *cvt;
sheet_part_element *item;
{
	sheet_part_element *n_item;
	int i, len, sz;
	
	if ( strcmp(cvt->buf, tok[TELEM]) == 0)
	{
		mlInTok(ge, cvt, FALSE);
		ReadELEM(ge, cvt, item, 0);
	}
	else if ( strcmp(cvt->buf, tok[TARR]) == 0)
	{	
		mlInTok(ge, cvt, FALSE);
		len = cvt->tok.n_vals;
		item->type = ISA_ARRAY;
		item->n    = len;
		item->val  = (char *)0;
		sz         = item->n * sizeof(sheet_part_element *);		
		item->item = (sheet_part_element **)gloc(ge, sz);
		for(i=0; i<len; i++)
		{
			sz  = sizeof(sheet_part_element);
			item->item[i] = (sheet_part_element *)gloc(ge, sz);
			n_item        = item->item[i];
			n_item->type  = ISA_VALUE;
			n_item->n     = 1;
			n_item->item  = (sheet_part_element **) NULL;			
			n_item->val   = cvtGetStrFromTokVal(ge, cvt, i);			
		}
	}
}

ReadSH_ARR(ge, cvt, t, chart)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
chart_type *chart;
{
	int i, j, m, n, sz, len, indx, tokenized;
	sheet_part_element *item, *n_item;
	chart_sheet_info *sheet;
	sheet_part *part;
	char *str;

	sheet = chart->sheet;
	if ( !chart->sheet->part )
	{
		sheet->n_parts = 1;
		sz             = sheet->n_parts * sizeof(sheet_part *);
		sheet->part    = (sheet_part **) gloc(ge, sz);
	}
	else
	{
		sheet->n_parts++;
		sz             = sheet->n_parts * sizeof(sheet_part *);
		sheet->part    = (sheet_part **) regloc(ge, sheet->part, sz);		
	}
	indx               = sheet->n_parts-1;
	sheet->part[indx]  = (sheet_part *) gloc(ge, sizeof(sheet_part));
	part = sheet->part[indx];
	mlInTok(ge, cvt, FALSE);
	part->name = cvtGetStrFromTokVal(ge, cvt, 0);

	mlInTok(ge, cvt, FALSE); /* read ELEM/ARR */
	item       = &part->item;
	item->type = ISA_ARRAY;
	item->val  = (char *)0;
	item->n    = 0;
	item->item = (sheet_part_element **) 0;
	if ( strcmp(cvt->buf, tok[TELEM]) == 0 || strcmp(cvt->buf, tok[TARR]) == 0 )
		tokenized = TRUE;
	else
		tokenized = FALSE;
	while ( tokenized || (cvt->tok.n_vals && cvt->tok.val[0].type == TOK_STR) )
	{
		item->n++;
		sz         = item->n * sizeof(sheet_part_element *);
		item->item = (sheet_part_element **)regloc(ge, item->item, sz);
		indx       = item->n - 1;
		item->item[indx] = (sheet_part_element *) 
								gloc(ge, sizeof(sheet_part_element));
		n_item = item->item[indx];
		if ( tokenized )
			ReadARR(ge, cvt, n_item);
		else
		{
			n_item->type = ISA_VALUE;
			n_item->n    = 1;
			n_item->item = (sheet_part_element **) NULL;
			n_item->val  = cvtGetStrFromTokVal(ge, cvt, 0);
		}

		/* read the next token */
		mlInTok(ge, cvt, FALSE);
		if ( strcmp(cvt->buf, tok[TELEM]) == 0 ||
		     strcmp(cvt->buf, tok[TARR]) == 0 )
			tokenized = TRUE;
		else
			tokenized = FALSE;
	}
	cvt->untok = TRUE;
}

ReadSH_ELEM(ge, cvt, t, chart)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
chart_type *chart;
{
	chart_sheet_info *sheet;
	sheet_part *part;
	sheet_part_element *item, *n_item;
	int i, j, m, n, sz, len, indx;
	char *str;

	sheet = chart->sheet;
	if ( !chart->sheet->part )
	{
		sheet->n_parts = 1;
		sz             = sheet->n_parts * sizeof(sheet_part *);
		sheet->part    = (sheet_part **) gloc(ge, sz);
	}
	else
	{
		sheet->n_parts++;
		sz             = sheet->n_parts * sizeof(sheet_part *);
		sheet->part    = (sheet_part **) regloc(ge, sheet->part, sz);		
	}
	indx           = sheet->n_parts-1;
	sheet->part[indx] = (sheet_part *) gloc(ge, sizeof(sheet_part));
	part = sheet->part[indx];
	mlInTok(ge, cvt, FALSE);
	part->name = cvtGetStrFromTokVal(ge, cvt, 0);
	item = &part->item;
	mlInTok(ge, cvt, FALSE); /* read elem */
	mlInTok(ge, cvt, FALSE); /* read element str */
	ReadELEM(ge, cvt, item, 0);
}


ReadROWDAT(ge, cvt, t, chart)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
chart_type *chart;
{
	chart_sheet_info *sheet;
	sheet_part *part;
	sheet_part_element *item, *n_item;
	int i, j, m, n, sz, len;
	char *str;
	/* first read the row data */
	mlInTok(ge, cvt, FALSE);
	len = cvt->tok.n_vals;
	if ( !chart->sheet )
	{	
		chart->sheet = (chart_sheet_info *)gloc(ge,sizeof(chart_sheet_info));
		memset((void *)chart->sheet, 0, sizeof(chart_sheet_info));
		sheet = chart->sheet;
		sheet->n_cols = len - 1;
		if ( chart->col_sliced == TRUE ) {
			sz = sheet->n_cols * sizeof(char *);
			sheet->legends = (char **) gloc(ge, sz);
			for(i=1; i<len; i++)
				sheet->legends[i-1]= cvtGetStrFromTokVal(ge, cvt, i);
		}
		else {
			sz = sheet->n_cols * sizeof(char *);
			sheet->categories = (char **) gloc(ge, sz);
			for(i=1; i<len; i++)
				sheet->categories[i-1]= cvtGetStrFromTokVal(ge, cvt, i);
		}
	}
	else
	{	
		sheet = chart->sheet;
		sheet->n_rows++;
		if ( chart->col_sliced == TRUE ) {
			/* realloc categories strs */		
			sz             = sheet->n_rows * sizeof(char *);
			sheet->categories = (char **) regloc(ge, sheet->categories, sz);
			sheet->categories[sheet->n_rows-1]=cvtGetStrFromTokVal(ge, cvt, 0);
		}
		else {
			/* realloc legend strs */		
			sz             = sheet->n_rows * sizeof(char *);
			sheet->legends = (char **) regloc(ge, sheet->legends, sz);
			sheet->legends[sheet->n_rows-1] = cvtGetStrFromTokVal(ge, cvt, 0);
		}
		
		/* realloc data stuff */
		sz = sheet->n_rows * sizeof(sheet_data **);
		sheet->data = (sheet_data ***) regloc(ge, sheet->data, sz);
		i = sheet->n_rows - 1;

		sz = (len-1) * sizeof(sheet_data *);
		sheet->data[i]  = (sheet_data **) gloc(ge, sz);
		for(j=0; j<len-1; j++)
		{
			sheet->data[i][j] = (sheet_data *) gloc(ge,sizeof(sheet_data));
			switch(cvt->tok.val[j+1].type)
			{
				case 0: /* is a number */
				case 1:
					sheet->data[i][j]->isa_val = 2;
					sheet->data[i][j]->u.val   = cvt->tok.val[j+1].u.fval;
				break;
				case 2:
					sheet->data[i][j]->isa_val = 1;
					sheet->data[i][j]->u.str = cvtGetStrFromTokVal(ge,cvt,j+1);
				break;
			}
		}
	}

}

ReadCOLUMNS(ge, cvt, t, chart)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
chart_type *chart;
{
	chart_sheet_info *sheet;
	sheet_part *part;
	sheet_part_element *item, *n_item;
	int i, j, len, *col_sz;
	char *str;
	
	mlInTok(ge, cvt, FALSE);
	len = cvt->tok.n_vals;
	sheet = chart->sheet;
	sheet->n_col_sz = len;
	col_sz = (int *) gloc(ge, len * sizeof(int));
	for ( i=0; i<len; i++)
		col_sz[i] = cvt->tok.val[i].u.ival;
	sheet->col_sz = col_sz;
}



/****************************************************************************
 *
 * module:	ReadAXIS
 */
ReadAXIS(ge,cvt,chart,axis)
graphics_main_type *ge;
cvt_global_type *cvt;
chart_type *chart;
chart_axis_type *axis;
{
	chart_data_point_type *dat;
	thing_attributes *att;
	int done, n, size;

	/* the axis label */
	mlInTok(ge, cvt, FALSE);
	axis->label_str = cvtTokToStr(ge, cvt);

	if ( cvt->info.version < 320 )
	{
		/* type and space consumed (for additional axes) */
		mlInTok(ge, cvt, FALSE);
		axis->type = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		axis->x_offset = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		axis->y_offset = cvtTokToInt(ge, cvt);
		if ( cvt->info.version >= 311 )
		{
			mlInTok(ge, cvt, FALSE);
			axis->tik_margin = cvtTokToInt(ge, cvt);
			mlInTok(ge, cvt, FALSE);
			axis->label_align = cvtTokToInt(ge, cvt);
			mlInTok(ge, cvt, FALSE);
			axis->tik_align = cvtTokToInt(ge, cvt);
		}
		else
			axis->tik_margin = 50;

		/* axis format info */
		mlInTok(ge, cvt, FALSE);
		axis->hidden = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		axis->auto_max = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		axis->auto_min = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		axis->auto_bas = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		axis->auto_n_major_tiks = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		axis->auto_n_minor_tiks = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		axis->major_tik_format = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		axis->minor_tik_format = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		axis->pos = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		axis->pos_val = cvtTokToFlt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		axis->min_val = cvtTokToFlt(ge, cvt);
		mlInTok(ge, cvt, FALSE);	
		axis->max_val = cvtTokToFlt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		axis->bas_val = cvtTokToFlt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		axis->first_delta = cvtTokToFlt(ge, cvt);
		if ( cvt->info.version > 311 )
		{
			mlInTok(ge, cvt, FALSE);
			axis->invalid_tiks = cvtTokToInt(ge, cvt);
		}

		mlInTok(ge, cvt, FALSE);
		axis->major_sz = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		axis->n_maj_steps = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		axis->maj_delta = cvtTokToFlt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		axis->maj_step = cvtTokToFlt(ge, cvt);

		mlInTok(ge, cvt, FALSE);
		axis->minor_sz = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		axis->n_min_steps = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		axis->min_delta = cvtTokToFlt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		axis->min_step = cvtTokToFlt(ge, cvt);

		/* numerical format for tiks */
		if ( mlInTok(ge, cvt, FALSE) == TOK_TOK)
			ReadNUM_FMT(ge, cvt, &axis->tik_format);

		/* initialize the bar stuff */
		axis->bar_overlap = chart->info.bar_overlap;
		axis->bar_margin  = chart->info.bar_margin;
		if ( axis->bas_val < axis->min_val )
			axis->bas_val = axis->min_val;
	}

	att = (thing_attributes *)0;
	done = FALSE;
	while ( !done && mlInTok(ge, cvt, TRUE) )
	{
		if ( cvt->tok.type != TOK_TOK )
			continue;

		switch ( cvt->tok.id )
		{
			default:
				(TokFuncs[cvt->tok.id])(ge, cvt, NULL, NULL);
			break;
			case TTIK_TYPE:
				mlInTok(ge, cvt, FALSE);
				axis->type       = cvt->tok.val[0].u.ival;
				axis->x_offset   = cvt->tok.val[1].u.ival;
				axis->y_offset   = cvt->tok.val[2].u.ival;
				axis->tik_margin = cvt->tok.val[3].u.ival;
			break;

			case TTIK_ALIGN:
				mlInTok(ge, cvt, FALSE);
				axis->label_align = cvt->tok.val[0].u.ival;
				axis->tik_align   = cvt->tok.val[1].u.ival;
				axis->tight_label = cvt->tok.val[2].u.ival;				
				axis->tight_tiks  = cvt->tok.val[3].u.ival;
			break;

			case TAXIS_FMT:
				mlInTok(ge, cvt, FALSE);
				axis->hidden            = cvt->tok.val[0].u.ival;
				axis->auto_max          = cvt->tok.val[1].u.ival;
				axis->auto_min          = cvt->tok.val[2].u.ival;
				axis->auto_bas          = cvt->tok.val[3].u.ival;
				axis->auto_n_major_tiks = cvt->tok.val[4].u.ival;
				axis->auto_n_minor_tiks = cvt->tok.val[5].u.ival;
				axis->major_tik_format  = cvt->tok.val[6].u.ival;
				axis->minor_tik_format  = cvt->tok.val[7].u.ival;
				axis->pos               = cvt->tok.val[8].u.ival;
				axis->pos_val           = cvt->tok.val[9].u.fval;
				axis->min_val           = cvt->tok.val[10].u.fval;
				axis->max_val           = cvt->tok.val[11].u.fval;
				axis->bas_val           = cvt->tok.val[12].u.fval;
				axis->first_delta       = cvt->tok.val[13].u.fval;
				axis->invalid_tiks      = cvt->tok.val[14].u.ival;
				axis->draw_behind       = cvt->tok.val[15].u.ival;
				axis->use_tik_attr      = cvt->tok.val[16].u.ival;
				if ( cvt->tok.n_vals > 17 )
					axis->inverted = cvt->tok.val[17].u.ival;
			break;

			case TAXIS_MAJOR:
				mlInTok(ge, cvt, FALSE);
				axis->major_sz    = cvt->tok.val[0].u.ival;
				axis->n_maj_steps = cvt->tok.val[1].u.ival;
				axis->maj_delta   = cvt->tok.val[2].u.fval;
				axis->maj_step    = cvt->tok.val[3].u.fval;
			break;

			case TAXIS_MINOR:
				mlInTok(ge, cvt, FALSE);
				axis->minor_sz    = cvt->tok.val[0].u.ival;
				axis->n_min_steps = cvt->tok.val[1].u.ival;
				axis->min_delta   = cvt->tok.val[2].u.fval;
				axis->min_step    = cvt->tok.val[3].u.fval;
			break;

			case TNUM_FMT:
				ReadNUM_FMT(ge, cvt, &axis->tik_format);
			break;
				
			case TBAR_INFO:
				if ( cvt->info.version < 320 )
				{
					mlInTok(ge, cvt, FALSE);
					axis->bar_margin = cvtTokToFlt(ge, cvt);
					mlInTok(ge, cvt, FALSE);
					axis->bar_overlap = cvtTokToFlt(ge, cvt);
				}
				else
				{
					mlInTok(ge, cvt, FALSE);
					axis->bar_margin  = cvt->tok.val[0].u.fval;
					axis->bar_overlap = cvt->tok.val[1].u.fval;
				}
			break;

			case TLABEL_FILTER:
				mlInTok(ge, cvt, FALSE);
				axis->filter_labels = cvt->tok.val[0].u.ival;
				if ( cvt->tok.n_vals == 1 )
				{	
					axis->c_tiks[0] = 0;
					axis->c_tiks[1] = 0;
				}
				else
				{	
					axis->c_tiks[0] = cvt->tok.val[1].u.ival;
					axis->c_tiks[1] = cvt->tok.val[2].u.ival;
				}
			break;

			case TLABEL:
				mlInTok(ge, cvt, FALSE);
				if ( cvt->info.version < 320 )
				{
					/* read the string */
					if ( cvt->tok.type == TOK_STR )
					{
						n = axis->n++;
						size = axis->n * sizeof(char *);
						axis->tik_str = 
								(char **)regloc(ge, axis->tik_str,size);
						axis->tik_str[n] = cvtTokToStr(ge, cvt);
					}

					/* read the display flag */
					if ( cvt->info.version > 311 )
					{
						size = axis->n * sizeof(int);
						axis->display = 
								(int *)regloc(ge, axis->display, size);
						mlInTok(ge, cvt, FALSE);
						axis->display[n] = cvtTokToInt(ge, cvt);
					}
					else
					{
						size = axis->n * sizeof(int);
						axis->display = 
								(int *)regloc(ge, axis->display, size);
						axis->display[n] = TRUE;
					}
				}
				else
				{
					n = axis->n++;

					size = axis->n * sizeof(char *);
					axis->tik_str=(char **)regloc(ge, axis->tik_str, size);
					axis->tik_str[n] = cvtGetStrFromTokVal(ge, cvt, 0);
					size = axis->n * sizeof(int);
					axis->display = (int *)regloc(ge, axis->display, size);
					axis->display[n] = cvt->tok.val[1].u.ival;
				}
			break;

			case TAXATTR:
				if ( att )
					memcpy(att,&cvt->curr_atts,sizeof(thing_attributes));
				att = &axis->attr;
			break;

			case TAXTIKATTR:
				if ( att )
					memcpy(att,&cvt->curr_atts,sizeof(thing_attributes));
				att = &axis->tik_attr;
			break;

			case TAXLABATTR:
				if ( att )
					memcpy(att,&cvt->curr_atts,sizeof(thing_attributes));
				att = &axis->label_attr;
			break;

			case TAXTIKLABATTR:
				if ( att )
					memcpy(att,&cvt->curr_atts,sizeof(thing_attributes));
				att = &axis->tik_label_attr;
			break;

			case TEND:
				done = TRUE;
			break;
		}
	}
	if ( att )
		memcpy(att,&cvt->curr_atts,sizeof(thing_attributes));
	mlInTok(ge, cvt, FALSE); /* AXIS */

	/* some quick repair */
	if ( cvt->info.version < 311 )
	{
		axis->label_align = axis->label_attr.h_just;
		axis->tik_align = axis->tik_label_attr.h_just;
		axis->tik_label_attr.h_just = axis->label_attr.h_just = 1;
	}
}

/****************************************************************************
 *
 * module:	ReadX_AXIS
 */
ReadX_AXIS(ge, cvt, t, chart)
graphics_main_type *ge;
cvt_global_type *cvt;
char *t;
chart_type *chart;
{
	chart_axis_type *axis;
	int naxis, size;

	naxis = chart->n_x_axes++;
	size = chart->n_x_axes * sizeof(chart_axis_type);
	chart->x_axis = (chart_axis_type *)regloc(ge, chart->x_axis, size);
	axis = &chart->x_axis[naxis];
	memset(axis, 0, sizeof(chart_axis_type));
	ReadAXIS(ge,cvt,chart,axis);
}

/****************************************************************************
 *
 * module:	ReadY_AXIS
 */
ReadY_AXIS(ge, cvt, t, chart)
graphics_main_type *ge;
cvt_global_type *cvt;
char *t;
chart_type *chart;
{
	chart_axis_type *axis;
	int naxis, size;

	naxis = chart->n_y_axes++;
	size = chart->n_y_axes * sizeof(chart_axis_type);
	chart->y_axis = (chart_axis_type *)regloc(ge, chart->y_axis, size);
	axis = &chart->y_axis[naxis];
	memset(axis, 0, sizeof(chart_axis_type));
	ReadAXIS(ge, cvt, chart, axis);
}

/****************************************************************************
 *
 * module:	ReadZ_AXIS
 */
ReadZ_AXIS(ge, cvt, t, chart)
graphics_main_type *ge;
cvt_global_type *cvt;
char *t;
chart_type *chart;
{
	chart_axis_type *axis;
	int naxis, size;

	naxis = chart->n_z_axes++;
	size = chart->n_z_axes * sizeof(chart_axis_type);
	chart->z_axis = (chart_axis_type *)regloc(ge, chart->z_axis, size);
	axis = &chart->z_axis[naxis];
	memset(axis, 0, sizeof(chart_axis_type));
	ReadAXIS(ge, cvt, chart, axis);
}

/****************************************************************************
 *
 * module:	ReadCHART_COUNTS
 */
ReadCHART_COUNTS(ge, cvt, t, chart)
graphics_main_type *ge;
cvt_global_type *cvt;
char *t;
chart_type *chart;
{
	if ( cvt->info.version < 320 )
	{
		mlInTok(ge, cvt, FALSE);
		chart->old_n_x_axes = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		chart->old_n_y_axes = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		chart->old_n_z_axes = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		chart->old_n_groups = cvtTokToInt(ge, cvt);
		if ( cvt->info.version >= 311 )
		{
			mlInTok(ge, cvt, FALSE);
			chart->old_type = cvtTokToInt(ge, cvt);
		}
		else
			chart->old_type = chart->type;
	}
	else
	{
		mlInTok(ge, cvt, FALSE);
		chart->old_n_x_axes = cvt->tok.val[0].u.ival;
		chart->old_n_y_axes = cvt->tok.val[1].u.ival;
		chart->old_n_z_axes = cvt->tok.val[2].u.ival;
		chart->old_n_groups = cvt->tok.val[3].u.ival;
		chart->old_type     = cvt->tok.val[4].u.ival;
	}
}

/****************************************************************************
 *
 * module:	ReadPIE_INFO
 */
ReadPIE_INFO(ge, cvt, t, chart)
graphics_main_type *ge;
cvt_global_type *cvt;
char *t;
chart_type *chart;
{
	mlInTok(ge, cvt, FALSE);
	chart->info.start_angle = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:	ReadMARGIN
 */
ReadMARGIN(ge, cvt, t, chart)
graphics_main_type *ge;
cvt_global_type *cvt;
char *t;
chart_type *chart;
{
	if ( cvt->info.version < 320 )
	{
		mlInTok(ge, cvt, FALSE);
		chart->margin.x1 = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		chart->margin.y1 = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		chart->margin.x2 = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		chart->margin.y2 = cvtTokToInt(ge, cvt);
	}
	else
	{
		mlInTok(ge, cvt, FALSE);
		chart->margin.x1 = cvt->tok.val[0].u.ival;
		chart->margin.y1 = cvt->tok.val[1].u.ival;
		chart->margin.x2 = cvt->tok.val[2].u.ival;
		chart->margin.y2 = cvt->tok.val[3].u.ival;
	}
}

/****************************************************************************
 *
 * module:	ReadEXTENT
 */
ReadEXTENT(ge, cvt, t, chart)
graphics_main_type *ge;
cvt_global_type *cvt;
char *t;
chart_type *chart;
{
	if ( cvt->info.version < 320 )
	{
		mlInTok(ge, cvt, FALSE);
		chart->extent.x1 = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		chart->extent.y1 = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		chart->extent.x2 = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		chart->extent.y2 = cvtTokToInt(ge, cvt);
	}
	else
	{
		mlInTok(ge, cvt, FALSE);
		chart->extent.x1 = cvt->tok.val[0].u.ival;
		chart->extent.y1 = cvt->tok.val[1].u.ival;
		chart->extent.x2 = cvt->tok.val[2].u.ival;
		chart->extent.y2 = cvt->tok.val[3].u.ival;
	}
}

/****************************************************************************
 *
 * module:	ReadOFFSET
 */
ReadOFFSET(ge, cvt, t, chart)
graphics_main_type *ge;
cvt_global_type *cvt;
char *t;
chart_type *chart;
{
	if ( cvt->info.version < 320 )
	{
		mlInTok(ge, cvt, FALSE);
		chart->def_x_offset = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		chart->def_y_offset = cvtTokToInt(ge, cvt);
	}
	else
	{
		mlInTok(ge, cvt, FALSE);
		chart->def_x_offset   = cvt->tok.val[0].u.ival;
		chart->def_y_offset   = cvt->tok.val[1].u.ival;
		chart->title_x_off    = cvt->tok.val[2].u.ival;
		chart->title_y_off    = cvt->tok.val[3].u.ival;
		chart->subtitle_x_off = cvt->tok.val[4].u.ival;
		chart->subtitle_y_off = cvt->tok.val[5].u.ival;
		chart->footer_x_off   = cvt->tok.val[6].u.ival;
		chart->footer_y_off   = cvt->tok.val[7].u.ival;
	}
}

/****************************************************************************
 *
 * module:	ReadGRIDS
 */
ReadGRIDS(ge, cvt, t, chart)
graphics_main_type *ge;
cvt_global_type *cvt;
char *t;
chart_type *chart;
{
	if ( cvt->info.version < 320 )
	{
		mlInTok(ge, cvt, FALSE);
		chart->maj_x_grids = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		chart->min_x_grids = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		chart->maj_y_grids = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		chart->min_y_grids = cvtTokToInt(ge, cvt);
	}
	else
	{
		mlInTok(ge, cvt, FALSE);
		chart->maj_x_grids = cvt->tok.val[0].u.ival;
		chart->min_x_grids = cvt->tok.val[1].u.ival;
		chart->maj_y_grids = cvt->tok.val[2].u.ival;
		chart->min_y_grids = cvt->tok.val[3].u.ival;
		chart->maj_z_grids = cvt->tok.val[4].u.ival;
		chart->min_z_grids = cvt->tok.val[5].u.ival;
	}
}

/****************************************************************************
 *
 * module:	ReadCLOSE
 */
ReadCLOSE(ge, cvt, t, chart)
graphics_main_type *ge;
cvt_global_type *cvt;
char *t;
chart_type *chart;
{
	if ( cvt->info.version < 320 )
	{
		mlInTok(ge, cvt, FALSE);
		chart->close_top = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		chart->close_bot = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		chart->close_lft = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		chart->close_rt = cvtTokToInt(ge, cvt);
	}
	else
	{
		mlInTok(ge, cvt, FALSE);
		chart->close_top = cvt->tok.val[0].u.ival;
		chart->close_bot = cvt->tok.val[1].u.ival;
		chart->close_lft = cvt->tok.val[2].u.ival;
		chart->close_rt  = cvt->tok.val[3].u.ival;
		chart->close_bg  = cvt->tok.val[4].u.ival;
		chart->close_fg  = cvt->tok.val[5].u.ival;
	}
}

/****************************************************************************
 *
 * module:	ReadHORZ
 */
ReadHORZ(ge, cvt, t, chart)
graphics_main_type *ge;
cvt_global_type *cvt;
char *t;
chart_type *chart;
{
	chart->orientation = chart->old_orient = 0;
}

/****************************************************************************
 *
 * module:	ReadVERT
 */
ReadVERT(ge, cvt, t, chart)
graphics_main_type *ge;
cvt_global_type *cvt;
char *t;
chart_type *chart;
{
	chart->orientation = chart->old_orient = 1;
}

/****************************************************************************
 *
 * module:	ReadINDEX
 */
ReadINDEX(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
}

/****************************************************************************
 *
 * module:	ReadTYPE - originally for images, now unused
 */
ReadTYPE(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
}

/****************************************************************************
 *
 * module:	ReadCOMPRESSION
 */
ReadCOMPRESSION(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	if ( mlInTok(ge, cvt, FALSE) != TOK_AUX )
		mlInError(ge, cvt, ML_UNKIMG);
	else
		cvt->compression = cvt->tok.id+1;
}

/****************************************************************************
 *
 * module:	ReadENCODING
 */
ReadENCODING(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	if ( mlInTok(ge, cvt, FALSE) != TOK_AUX )
		mlInError(ge, cvt, ML_UNKIMG);
	else
		cvt->encoding = cvt->tok.id;
}

/****************************************************************************
 *
 * module:	ReadEPS
 */
ReadEPS(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	int readingLines, nChars;
	int x, y, num, temp, top;
	char **tokP, *chP, epsLine[1024];
	int found;

	/* prepare the eps segement */
	memset(epsLine, 0, 1024);

	/* read the width */
	mlInTok(ge, cvt, FALSE);
	cvt->epsWid = cvtTokToInt(ge, cvt);

	/* read the height */
	mlInTok(ge, cvt, FALSE);
	cvt->epsHyt = cvtTokToInt(ge, cvt);

	if ( cvt->majorVer >= GVER_MAJOR )
	{
		/* then the size is stored */
		mlInTok(ge, cvt, FALSE);
		num = cvtTokToInt(ge, cvt);
		t->data.img->info.eps_seg_size = num;
		t->data.img->info.eps_seg = chP = (char*)gloc(ge, num+1);

		/* get to next line */
		while ((*cvt->getcFunc)(ge, cvt) != '\n')
			;

		/* read the eps data */
		(*cvt->fgetsFunc)(ge, cvt, epsLine, 1024);
		if (strncmp(epsLine, NXLT(" END EPS"), 7))
		{
			do {
				if (*epsLine == ' ')
				{
					num = strlen(epsLine) - 1;
					memcpy(chP,epsLine+1,num);
				}
				else
				{
					num = strlen(epsLine);
					memcpy(chP,epsLine,num);
				}
				chP += num;
				(*cvt->fgetsFunc)(ge, cvt, epsLine, 1024);
			} while(strncmp(epsLine, NXLT(" END EPS"), 7));
			*chP = 0;
		}
	}
	else /* rev 2, so do it the hard way */
	{
		/* get to next line */
		while ((*cvt->getcFunc)(ge, cvt) != '\n')
			;

		(*cvt->fgetsFunc)(ge, cvt, epsLine, 1024); /* toss x and y offset */
		(*cvt->fgetsFunc)(ge, cvt, epsLine, 1024);

		t->data.img->info.eps_seg_size = 0;
		(*cvt->fgetsFunc)(ge, cvt, epsLine, 1024);
		if (strncmp(epsLine, NXLT("END EPS"), 7))
		{
			num = strlen(epsLine); /* malloc for filters */
			t->data.img->info.eps_seg = chP = (char*)gloc(ge, num+1);
			do {
				num = strlen(epsLine);
				t->data.img->info.eps_seg_size += num;
				temp = t->data.img->info.eps_seg_size;
				chP = (char*)regloc(ge, t->data.img->info.eps_seg,temp);
				t->data.img->info.eps_seg = chP;
				chP = t->data.img->info.eps_seg + 
					 (t->data.img->info.eps_seg_size - num);
				memcpy(chP,epsLine,num);
				(*cvt->fgetsFunc)(ge, cvt, epsLine, 1024);
			} while(strncmp(epsLine, NXLT("END EPS"), 7));
			*chP = 0;
		}
	}
}

/****************************************************************************
 *
 * module:	ReadNULLOBJ
 */
ReadNULLOBJ(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
}

/****************************************************************************
 *
 * module:
 */
ReadOPAGEXY(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( !cvt->templ_print )
		mlInCheckPrint(ge, cvt);
	ge->outline_print->page_x = cvt->tok.val[0].u.ival + 59750;
	ge->outline_print->page_y = cvt->tok.val[1].u.ival + 58000;
}

/****************************************************************************
 *
 * module:
 */
ReadOPAGEHYT(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{

	mlInTok(ge, cvt, FALSE);
	if ( !cvt->templ_print )
		mlInCheckPrint(ge, cvt);
	ge->outline_print->page_hyt = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:
 */
ReadOPAGEWID(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( !cvt->templ_print )
		mlInCheckPrint(ge, cvt);
	ge->outline_print->page_wid = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:
 */
ReadOPRINTMARKS(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( !cvt->templ_print )
		mlInCheckPrint(ge, cvt);
	ge->outline_print->use_content_border = cvt->tok.val[0].u.ival;
	ge->outline_print->use_crop_marks     = cvt->tok.val[1].u.ival;
	ge->outline_print->use_regis_marks    = cvt->tok.val[2].u.ival;
}

/****************************************************************************
 *
 * module:
 */
ReadOPRINTBLEED(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( !cvt->templ_print )
		mlInCheckPrint(ge, cvt);
	ge->outline_print->bleed = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:
 */
ReadOPRINTWID(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( !cvt->templ_print )
		mlInCheckPrint(ge, cvt);
	ge->outline_print->wid = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:
 */
ReadOPRINTHYT(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( !cvt->templ_print )
		mlInCheckPrint(ge, cvt);
	ge->outline_print->hyt = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:
 */
ReadOPRINTLAND(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( !cvt->templ_print )
		mlInCheckPrint(ge, cvt);
	ge->outline_print->landscape = cvt->tok.id;
}

/****************************************************************************
 *
 * module:
 */
ReadOPRINTGUIDES(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( !cvt->templ_print )
		mlInCheckPrint(ge, cvt);
	ge->outline_print->guide_pt.x = cvt->tok.val[0].u.ival;
	ge->outline_print->guide_pt.y = cvt->tok.val[1].u.ival;
}

/****************************************************************************
 *
 * module:
 */
ReadOPRINTMARG(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( !cvt->templ_print )
		mlInCheckPrint(ge, cvt);
	ge->outline_print->margin.x1 = cvt->tok.val[0].u.ival;
	ge->outline_print->margin.y1 = cvt->tok.val[1].u.ival;
	ge->outline_print->margin.x2 = cvt->tok.val[2].u.ival;
	ge->outline_print->margin.y2 = cvt->tok.val[3].u.ival;
}

/****************************************************************************
 *
 * module:
 */
ReadHPAGEXY(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( !cvt->templ_print )
		mlInCheckPrint(ge, cvt);
	ge->handout_print->page_x = cvt->tok.val[0].u.ival + 59750;
	ge->handout_print->page_y = cvt->tok.val[1].u.ival + 58000;
}

/****************************************************************************
 *
 * module:
 */
ReadHPAGEHYT(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{

	mlInTok(ge, cvt, FALSE);
	if ( !cvt->templ_print )
		mlInCheckPrint(ge, cvt);
	ge->handout_print->page_hyt = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:
 */
ReadHPAGEWID(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( !cvt->templ_print )
		mlInCheckPrint(ge, cvt);
	ge->handout_print->page_wid = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:
 */
ReadHPRINTMARKS(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( !cvt->templ_print )
		mlInCheckPrint(ge, cvt);
	ge->handout_print->use_content_border = cvt->tok.val[0].u.ival;
	ge->handout_print->use_crop_marks     = cvt->tok.val[1].u.ival;
	ge->handout_print->use_regis_marks    = cvt->tok.val[2].u.ival;
}

/****************************************************************************
 *
 * module:
 */
ReadHPRINTBLEED(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( !cvt->templ_print )
		mlInCheckPrint(ge, cvt);
	ge->handout_print->bleed = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:
 */
ReadHPRINTWID(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( !cvt->templ_print )
		mlInCheckPrint(ge, cvt);
	ge->handout_print->wid = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:
 */
ReadHPRINTHYT(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( !cvt->templ_print )
		mlInCheckPrint(ge, cvt);
	ge->handout_print->hyt = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:
 */
ReadHPRINTLAND(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( !cvt->templ_print )
		mlInCheckPrint(ge, cvt);
	ge->handout_print->landscape = cvt->tok.id;
}

/****************************************************************************
 *
 * module:
 */
ReadHPRINTGUIDES(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( !cvt->templ_print )
		mlInCheckPrint(ge, cvt);
	ge->handout_print->guide_pt.x = cvt->tok.val[0].u.ival;
	ge->handout_print->guide_pt.y = cvt->tok.val[1].u.ival;
}

/****************************************************************************
 *
 * module:
 */
ReadHPRINTMARG(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( !cvt->templ_print )
		mlInCheckPrint(ge, cvt);
	ge->handout_print->margin.x1 = cvt->tok.val[0].u.ival;
	ge->handout_print->margin.y1 = cvt->tok.val[1].u.ival;
	ge->handout_print->margin.x2 = cvt->tok.val[2].u.ival;
	ge->handout_print->margin.y2 = cvt->tok.val[3].u.ival;
}

/****************************************************************************
 *
 * module:
 */
ReadNPAGEXY(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( !cvt->templ_print )
		mlInCheckPrint(ge, cvt);
	ge->notes_print->page_x = cvt->tok.val[0].u.ival + 59750;
	ge->notes_print->page_y = cvt->tok.val[1].u.ival + 58000;
}

/****************************************************************************
 *
 * module:
 */
ReadNPAGEHYT(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{

	mlInTok(ge, cvt, FALSE);
	if ( !cvt->templ_print )
		mlInCheckPrint(ge, cvt);
	ge->notes_print->page_hyt = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:
 */
ReadNPAGEWID(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( !cvt->templ_print )
		mlInCheckPrint(ge, cvt);
	ge->notes_print->page_wid = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:
 */
ReadNPRINTMARKS(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( !cvt->templ_print )
		mlInCheckPrint(ge, cvt);
	ge->notes_print->use_content_border = cvt->tok.val[0].u.ival;
	ge->notes_print->use_crop_marks     = cvt->tok.val[1].u.ival;
	ge->notes_print->use_regis_marks    = cvt->tok.val[2].u.ival;
}

/****************************************************************************
 *
 * module:
 */
ReadNPRINTBLEED(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( !cvt->templ_print )
		mlInCheckPrint(ge, cvt);
	ge->notes_print->bleed = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:
 */
ReadNPRINTWID(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( !cvt->templ_print )
		mlInCheckPrint(ge, cvt);
	ge->notes_print->wid = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:
 */
ReadNPRINTHYT(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( !cvt->templ_print )
		mlInCheckPrint(ge, cvt);
	ge->notes_print->hyt = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:
 */
ReadNPRINTLAND(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( !cvt->templ_print )
		mlInCheckPrint(ge, cvt);
	ge->notes_print->landscape = cvt->tok.id;
}

/****************************************************************************
 *
 * module:
 */
ReadNPRINTGUIDES(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( !cvt->templ_print )
		mlInCheckPrint(ge, cvt);
	ge->notes_print->guide_pt.x = cvt->tok.val[0].u.ival;
	ge->notes_print->guide_pt.y = cvt->tok.val[1].u.ival;
}

/****************************************************************************
 *
 * module:
 */
ReadNPRINTMARG(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( !cvt->templ_print )
		mlInCheckPrint(ge, cvt);
	ge->notes_print->margin.x1 = cvt->tok.val[0].u.ival;
	ge->notes_print->margin.y1 = cvt->tok.val[1].u.ival;
	ge->notes_print->margin.x2 = cvt->tok.val[2].u.ival;
	ge->notes_print->margin.y2 = cvt->tok.val[3].u.ival;
}

/****************************************************************************
 *
 * module:	ReadPAGEHYT
 */
ReadPAGEHYT(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	ge->print.page_hyt = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:	ReadPRINTSCALE
 */
ReadPRINTSCALE(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	ge->print.h_scale = cvtTokToInt(ge, cvt);
	mlInTok(ge, cvt, FALSE);
	ge->print.v_scale = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:	ReadPRINTBLEED
 */
ReadPRINTBLEED(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	ge->print.bleed = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:	ReadPRINTWID
 */
ReadPRINTWID(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	ge->print.wid = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:	ReadPRINTHYT
 */
void
ReadPRINTHYT(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	ge->print.hyt = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:	ReadPRINTMARKS
 */
void
ReadPRINTMARKS(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	if ( cvt->info.version < 320 )
	{
		mlInTok(ge, cvt, FALSE);
		ge->print.use_content_border = cvt->tok.id ? TRUE : FALSE;
		mlInTok(ge, cvt, FALSE);
		ge->print.use_crop_marks = cvt->tok.id ? TRUE : FALSE;
		mlInTok(ge, cvt, FALSE);
		ge->print.use_regis_marks = cvt->tok.id ? TRUE : FALSE;
	}
	else
	{
		mlInTok(ge, cvt, FALSE);
		ge->print.use_content_border = cvt->tok.val[0].u.ival;
		ge->print.use_crop_marks     = cvt->tok.val[1].u.ival;
		ge->print.use_regis_marks    = cvt->tok.val[2].u.ival;
	}
}

/****************************************************************************
 *
 * module:	ReadPRINTLAND
 */
void
ReadPRINTLAND(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	if ( cvt->info.version < 320 )
		ge->print.landscape = TRUE;
	else
	{
		mlInTok(ge, cvt, FALSE);
		ge->print.landscape = cvt->tok.id;
	}
}

/****************************************************************************
 *
 * module:
 */
void
ReadPRINTGUIDES(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( !cvt->templ_print )
		mlInCheckPrint(ge, cvt);
	ge->print.guide_pt.x = cvt->tok.val[0].u.ival;
	ge->print.guide_pt.y = cvt->tok.val[1].u.ival;
}

/****************************************************************************
 *
 * module:	ReadPRINTMARG
 */
void
ReadPRINTMARG(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	if ( cvt->info.version < 320 )
	{
		mlInTok(ge, cvt, FALSE);
		ge->print.margin.x1 = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		ge->print.margin.y1 = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		ge->print.margin.x2 = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		ge->print.margin.y2 = cvtTokToInt(ge, cvt);
	}
	else
	{
		mlInTok(ge, cvt, FALSE);
		ge->print.margin.x1 = cvt->tok.val[0].u.ival;
		ge->print.margin.y1 = cvt->tok.val[1].u.ival;
		ge->print.margin.x2 = cvt->tok.val[2].u.ival;
		ge->print.margin.y2 = cvt->tok.val[3].u.ival;
	}
}

/****************************************************************************
 *
 * module:	ReadPAGEWID
 */
void
ReadPAGEWID(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	ge->print.page_wid = cvtTokToInt(ge, cvt);
}

/****************************************************************************
 *
 * module:	ReadPAGEXY
 */
void
ReadPAGEXY(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	if ( cvt->info.version < 320 )
	{
		mlInTok(ge, cvt, FALSE);
		ge->print.page_x = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		ge->print.page_y = cvtTokToInt(ge, cvt);
	}
	else
	{
		mlInTok(ge, cvt, FALSE);
		ge->print.page_x = cvt->tok.val[0].u.ival + 59750;
		ge->print.page_y = cvt->tok.val[1].u.ival + 58000;
	}
}

/****************************************************************************
 *
 * module:	ReadPAGEZEROS
 */
void
ReadPAGEZEROS(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	if ( cvt->info.version < 320 )
	{
		mlInTok(ge, cvt, FALSE);
		ge->x_zero_pt = cvtTokToInt(ge, cvt);
		mlInTok(ge, cvt, FALSE);
		ge->y_zero_pt = cvtTokToInt(ge, cvt);
	}
	else
	{
		mlInTok(ge, cvt, FALSE);
		ge->x_zero_pt = cvt->tok.val[0].u.ival + 59750;
		if ( cvt->info.version < 420 )
			ge->y_zero_pt = cvt->tok.val[1].u.ival + 58000;
		else
			ge->y_zero_pt = cvt->tok.val[1].u.ival + 58500;
	}
	ge->state.win_orig.x1 = ge->x_zero_pt;
	ge->state.win_orig.y1 = ge->y_zero_pt;
}

/****************************************************************************
 *
 * module:	ReadDOCDIMENSIONS
 */
void
ReadDOCDIMENSIONS(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	/* they should both return TOK_NUM */
	mlInTok(ge, cvt, FALSE);
	cvt->geWidth = cvtTokToInt(ge, cvt);
	mlInTok(ge, cvt, FALSE);
	cvt->geHeight = cvtTokToInt(ge, cvt);
	ge->state.drawn_area.x1 = 0;
	ge->state.drawn_area.y1 = 0;
	ge->state.drawn_area.x2 = cvt->geWidth;
	ge->state.drawn_area.y2 = cvt->geHeight;
}

/****************************************************************************
 *
 * module:	ReadWIN_SIZE
 */
void
ReadWIN_SIZE(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	ge->state.drawn_area.x1 = 0;
	ge->state.drawn_area.y1 = 0;
	ge->state.drawn_area.x2 = cvt->geWidth  = cvt->tok.val[0].u.ival;
	ge->state.drawn_area.y2 = cvt->geHeight = cvt->tok.val[1].u.ival;
}

/****************************************************************************
 *
 * module:	mlInAttrBlock - for template attributes
 */
void
mlInAttrBlock(ge, cvt, attr)
graphics_main_type *ge;
cvt_global_type *cvt;
thing_attributes *attr;
{
	int done;

	done = FALSE;
	while ( !done && mlInTok(ge, cvt, TRUE) )
	{
		/* assume tokens (fragile!) */
		if ( cvt->tok.type != TOK_TOK )
			continue;

		switch ( cvt->tok.id )
		{
			default:
				(TokFuncs[cvt->tok.id])(ge, cvt, (thing *)0, NULL);
			break;
			case TEND:
				done = TRUE;
			break;
		}
	}
	mlInTok(ge, cvt, FALSE); /* END OF BLOCK TOKEN */
	memcpy(attr, &cvt->curr_atts, sizeof(thing_attributes));
}

/****************************************************************************
 *
 * module:	mlInSlideMaster
 */
void
mlInSlideMaster(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int found_t, done, sz, enable_build;
	thing *bg_t;
	vector ext;

	if ( cvt->info.version >= 320 )
	{
		mlInCheckCMaps(ge, cvt);
		mlInCheckFaces(ge, cvt);
		mlInCheckCBacks(ge, cvt);
		mlInCheckFills(ge, cvt);
		mlInCheckStyles(ge, cvt);
	}

	mlInCheckTemplate(ge, cvt);

	/* read the info */
	done = found_t = FALSE;
	while ( !done && mlInTok(ge, cvt, TRUE) )
	{
		if ( cvt->tok.type == TOK_OTY )
		{
			mlInCheckPage(ge, cvt, &ge->template->s_master_page);
			bg_t = cvt->tok.id ? mlInPart(ge, cvt, cvt->tok.id) : (thing *)0;
			ge->template->s_master_page.t = *bg_t;
			gfree(ge, bg_t);
			found_t = TRUE;
		}
		else if ( cvt->tok.type == TOK_TOK ) switch ( cvt->tok.id )
		{
			case TS_EXT1:
				mlInTok(ge, cvt, FALSE);
				ext.x1 = cvt->tok.val[0].u.ival + PAGE_ZEROX;
				ext.y1 = cvt->tok.val[1].u.ival + PAGE_ZEROY;
				ext.x2 = cvt->tok.val[2].u.ival + PAGE_ZEROX;
				ext.y2 = cvt->tok.val[3].u.ival + PAGE_ZEROY;
				ge->template->slide_heading = ext;
			break;
			case TS_EXT2:
				mlInTok(ge, cvt, FALSE);
				ext.x1 = cvt->tok.val[0].u.ival + PAGE_ZEROX;
				ext.y1 = cvt->tok.val[1].u.ival + PAGE_ZEROY;
				ext.x2 = cvt->tok.val[2].u.ival + PAGE_ZEROX;
				ext.y2 = cvt->tok.val[3].u.ival + PAGE_ZEROY;
				ge->template->slide_content = ext;
			break;
			case TT_ATTR0:
				mlInAttrBlock(ge, cvt, &ge->template->t0_attr);
			break;
			case TT_ATTR1:
				mlInAttrBlock(ge, cvt, &ge->template->t1_attr);
			break;
			case TS_ATTR0:
				mlInAttrBlock(ge, cvt, &ge->template->b0_attr);
			break;
			case TS_ATTR1:
				mlInAttrBlock(ge, cvt, &ge->template->b1_attr);
			break;
			case TS_ATTR2:
				mlInAttrBlock(ge, cvt, &ge->template->b2_attr);
			break;
			case TS_ATTR3:
				mlInAttrBlock(ge, cvt, &ge->template->b3_attr);
			break;
			case TS_ATTR4:
				mlInAttrBlock(ge, cvt, &ge->template->b4_attr);
			break;
			case TS_ATTR5:
				mlInAttrBlock(ge, cvt, &ge->template->b5_attr);
			break;
			case TSLIDE_INFO:
				mlInSlideInfo(ge, cvt);
			break;
			case TBUILD_INFO:
				mlInBuildInfo(ge, cvt, &cvt->build, &enable_build);
				/* note that we found the slide build info */
				cvt->use_slide_build = enable_build;
				cvt->slide_build_found = TRUE;
			break;
			case TLAYERS:
				mlInLayers(ge, cvt);
			break;
			case TEND:
			{
				done = TRUE; /* get rid of SLIDE_MASTER after end */
				mlInTok(ge, cvt, FALSE); 
			}
		}
	}
	if ( !found_t )
		mlInCheckPage(ge, cvt, &ge->template->s_master_page);
}

/****************************************************************************
 *
 * module:	mlInTitleMaster
 */
void
mlInTitleMaster(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int found_t, done, sz;
	thing *bg_t;
	vector ext;

	if ( cvt->info.version >= 320 )
	{
		mlInCheckCMaps(ge, cvt);
		mlInCheckFaces(ge, cvt);
		mlInCheckCBacks(ge, cvt);
		mlInCheckFills(ge, cvt);
		mlInCheckStyles(ge, cvt);
	}

	/* create an empty template */
	mlInCheckTemplate(ge, cvt);

	/* read the info */
	sz = sizeof(ge_page_type);
	ge->template->t_master_page = (ge_page_type *)gloc(ge, sz);
	memset((char *)ge->template->t_master_page, 0, sz);
	sz = sizeof(thing_attributes);
	memcpy((char*)&ge->template->ct0_attr, (char*)&ge->def_attr, sz);
	memcpy((char*)&ge->template->ct1_attr, (char*)&ge->def_attr, sz);

	done = found_t = FALSE;
	while ( !done && mlInTok(ge, cvt, TRUE) )
	{
		if ( cvt->tok.type == TOK_OTY )
		{
			mlInCheckPage(ge, cvt, ge->template->t_master_page);
			bg_t = cvt->tok.id ? mlInPart(ge, cvt, cvt->tok.id) : (thing *)0;
			ge->template->t_master_page->t = *bg_t;
			gfree(ge, bg_t);
			found_t = TRUE;
		}
		else if ( cvt->tok.type == TOK_TOK ) switch ( cvt->tok.id )
		{
			case TS_EXT1:
				mlInTok(ge, cvt, FALSE);
				ext.x1 = cvt->tok.val[0].u.ival + PAGE_ZEROX;
				ext.y1 = cvt->tok.val[1].u.ival + PAGE_ZEROY;
				ext.x2 = cvt->tok.val[2].u.ival + PAGE_ZEROX;
				ext.y2 = cvt->tok.val[3].u.ival + PAGE_ZEROY;
				ge->template->title_heading = ext;
			break;
			case TS_EXT2:
				mlInTok(ge, cvt, FALSE);
				ext.x1 = cvt->tok.val[0].u.ival + PAGE_ZEROX;
				ext.y1 = cvt->tok.val[1].u.ival + PAGE_ZEROY;
				ext.x2 = cvt->tok.val[2].u.ival + PAGE_ZEROX;
				ext.y2 = cvt->tok.val[3].u.ival + PAGE_ZEROY;
				ge->template->title_content = ext;
			break;
			case TT_ATTR0:
				mlInAttrBlock(ge, cvt, &ge->template->ct0_attr);
			break;
			case TT_ATTR1:
				mlInAttrBlock(ge, cvt, &ge->template->ct1_attr);
			break;
			case TSLIDE_INFO:
				mlInSlideInfo(ge, cvt);
			break;
			case TLAYERS:
				mlInLayers(ge, cvt);
			break;
			case TEND:
			{
				done = TRUE; /* get rid of SLIDE_MASTER after end */
				mlInTok(ge, cvt, FALSE); 
			}
		}
	}
	if ( !found_t )
		mlInCheckPage(ge, cvt, ge->template->t_master_page);
}

/****************************************************************************
 *
 * module:	mlInOutlineMaster
 */
void
mlInOutlineMaster(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int found_t, done, sz;
	thing *bg_t;

	if ( cvt->info.version >= 320 )
	{
		mlInCheckCMaps(ge, cvt);
		mlInCheckFaces(ge, cvt);
		mlInCheckCBacks(ge, cvt);
		mlInCheckFills(ge, cvt);
		mlInCheckStyles(ge, cvt);
	}

	/* create an empty template */
	mlInCheckTemplate(ge, cvt);

	/* read the info */
	done = found_t = FALSE;
	while ( !done && mlInTok(ge, cvt, TRUE) )
	{
		if ( cvt->tok.type == TOK_OTY )
		{
			mlInCheckPage(ge, cvt, &ge->template->o_master_page);
			bg_t = cvt->tok.id ? mlInPart(ge, cvt, cvt->tok.id) : (thing *)0;
			ge->template->o_master_page.t = *bg_t;
			gfree(ge, bg_t);
			found_t = TRUE;
		}
		else if ( cvt->tok.type == TOK_TOK ) switch ( cvt->tok.id )
		{
			case TSLIDE_INFO:
				mlInSlideInfo(ge, cvt);
			break;
			case TLAYERS:
				mlInLayers(ge, cvt);
			break;
			case TEND:
			{
				done = TRUE; /* get rid of SLIDE_MASTER after end */
				mlInTok(ge, cvt, FALSE); 
			}
		}
	}
	if ( !found_t )
		mlInCheckPage(ge, cvt, &ge->template->o_master_page);
}

/****************************************************************************
 *
 * module:	mlInHandoutMaster
 */
void
mlInHandoutMaster(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int found_t, done, sz;
	thing *bg_t;

	if ( cvt->info.version >= 320 )
	{
		mlInCheckCMaps(ge, cvt);
		mlInCheckFaces(ge, cvt);
		mlInCheckCBacks(ge, cvt);
		mlInCheckFills(ge, cvt);
		mlInCheckStyles(ge, cvt);
	}

	/* create an empty template */
	mlInCheckTemplate(ge, cvt);

	/* read the info */
	done = found_t = FALSE;
	while ( !done && mlInTok(ge, cvt, TRUE) )
	{
		if ( cvt->tok.type == TOK_OTY )
		{
			mlInCheckPage(ge, cvt, &ge->template->h_master_page);
			bg_t = cvt->tok.id ? mlInPart(ge, cvt, cvt->tok.id) : (thing *)0;
			ge->template->h_master_page.t = *bg_t;
			gfree(ge, bg_t);
			found_t = TRUE;
		}
		else if ( cvt->tok.type == TOK_TOK ) switch ( cvt->tok.id )
		{
			case TSLIDE_INFO:
				mlInSlideInfo(ge, cvt);
			break;
			case TLAYERS:
				mlInLayers(ge, cvt);
			break;
			case TEND:
			{
				done = TRUE; /* get rid of SLIDE_MASTER after end */
				mlInTok(ge, cvt, FALSE); 
			}
		}
	}
	if ( !found_t )
		mlInCheckPage(ge, cvt, &ge->template->h_master_page);
}

/****************************************************************************
 *
 * module:	mlInNotesMaster
 */
void
mlInNotesMaster(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int found_t, done, sz;
	thing *bg_t;
	vector ext;

	if ( cvt->info.version >= 320 )
	{
		mlInCheckCMaps(ge, cvt);
		mlInCheckFaces(ge, cvt);
		mlInCheckCBacks(ge, cvt);
		mlInCheckFills(ge, cvt);
		mlInCheckStyles(ge, cvt);
	}

	/* create an empty template */
	mlInCheckTemplate(ge, cvt);

	/* read the info */
	done = found_t = FALSE;
	while ( !done && mlInTok(ge, cvt, TRUE) )
	{
		if ( cvt->tok.type == TOK_OTY )
		{
			mlInCheckPage(ge, cvt, &ge->template->n_master_page);
			bg_t = cvt->tok.id ? mlInPart(ge, cvt, cvt->tok.id) : (thing *)0;
			ge->template->n_master_page.t = *bg_t;
			gfree(ge, bg_t);
			found_t = TRUE;
		}
		else if ( cvt->tok.type == TOK_TOK ) switch ( cvt->tok.id )
		{
			case TN_EXT1:
				mlInTok(ge, cvt, FALSE);
				ext.x1 = cvt->tok.val[0].u.ival + PAGE_ZEROX;
				ext.y1 = cvt->tok.val[1].u.ival + PAGE_ZEROY;
				ext.x2 = cvt->tok.val[2].u.ival + PAGE_ZEROX;
				ext.y2 = cvt->tok.val[3].u.ival + PAGE_ZEROY;
				ge->template->notes_preview = ext;
			break;
			case TN_EXT2:
				mlInTok(ge, cvt, FALSE);
				ext.x1 = cvt->tok.val[0].u.ival + PAGE_ZEROX;
				ext.y1 = cvt->tok.val[1].u.ival + PAGE_ZEROY;
				ext.x2 = cvt->tok.val[2].u.ival + PAGE_ZEROX;
				ext.y2 = cvt->tok.val[3].u.ival + PAGE_ZEROY;
				ge->template->notes_content = ext;
			break;
			case TN_ATTR0:
				mlInAttrBlock(ge, cvt, &ge->template->n0_attr);
			break;
			case TN_ATTR1:
				mlInAttrBlock(ge, cvt, &ge->template->n1_attr);
			break;
			case TN_ATTR2:
				mlInAttrBlock(ge, cvt, &ge->template->n2_attr);
			break;
			case TN_ATTR3:
				mlInAttrBlock(ge, cvt, &ge->template->n3_attr);
			break;
			case TN_ATTR4:
				mlInAttrBlock(ge, cvt, &ge->template->n4_attr);
			break;
			case TN_ATTR5:
				mlInAttrBlock(ge, cvt, &ge->template->n5_attr);
			break;
			case TSLIDE_INFO:
				mlInSlideInfo(ge, cvt);
			break;
			case TLAYERS:
				mlInLayers(ge, cvt);
			break;
			case TEND:
			{
				done = TRUE; /* get rid of SLIDE_MASTER after end */
				mlInTok(ge, cvt, FALSE); 
			}
		}
	}
	if ( !found_t )
		mlInCheckPage(ge, cvt, &ge->template->n_master_page);
}

/****************************************************************************
 *
 * module:	ReadZOOM
 */
void
ReadZOOM(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	mlInTok(ge, cvt, FALSE);
	ge->xform.zoom_to_fit = cvt->tok.val[0].u.ival;
	ge->xform.x_res = cvt->tok.val[1].u.ival * 10;
	ge->xform.y_res = cvt->tok.val[2].u.ival * 10;
	if ( cvt->tok.n_vals > 3 )
	{
		ge->state.win_orig.x1 = cvt->tok.val[3].u.ival + ge->x_zero_pt;
		ge->state.win_orig.y1 = cvt->tok.val[4].u.ival + ge->y_zero_pt;
	}
}

/****************************************************************************
 *
 * module:	ReadREZ
 */
void
ReadREZ(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	mlInTok(ge, cvt, FALSE);
	ge->xform.zoom_to_fit = cvt->tok.val[0].u.ival;
	ge->xform.x_res = cvt->tok.val[1].u.ival;
	ge->xform.y_res = cvt->tok.val[2].u.ival;
	if ( cvt->tok.n_vals > 3 )
	{
		ge->state.win_orig.x1 = cvt->tok.val[3].u.ival + ge->x_zero_pt;
		ge->state.win_orig.y1 = cvt->tok.val[4].u.ival + ge->y_zero_pt;
	}
}

/****************************************************************************
 *
 * module:	ReadRUN_MACRO
 */
void
ReadRUN_MACRO(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	mlInTok(ge, cvt, FALSE);
	ge->state.run_macro = cvt->tok.val[0].u.cval.str;
	ge->state.n_mchars  = cvt->tok.val[0].u.cval.n_chars;
	cvt->tok.val[0].u.cval.str = (tx_char *)0;
	cvt->tok.val[0].u.cval.n_chars = 0;
}

/****************************************************************************
 *
 * module:	ReadSLIDE_SHOW
 */
void
ReadSLIDE_SHOW(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	mlInTok(ge, cvt, FALSE);
	ge->state.show_option.full_screen = cvt->tok.val[0].u.ival;
	ge->state.show_option.all_pages   = cvt->tok.val[1].u.ival;
	ge->state.show_option.page1       = cvt->tok.val[2].u.ival;
	ge->state.show_option.pageN       = cvt->tok.val[3].u.ival;
	ge->state.show_option.at_end      = cvt->tok.val[4].u.ival;
}

/****************************************************************************
 *
 * module:	ReadSESSION
 */
void
ReadSESSION(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
 	int val, done, size, jnk, n_colors;

	if ( cvt->info.version >= 320 )
	{
		mlInCheckCMaps(ge, cvt);
		mlInCheckFaces(ge, cvt);
		mlInCheckCBacks(ge, cvt);
		mlInCheckFills(ge, cvt);
		mlInCheckStyles(ge, cvt);
	}
	else
		mlInCheckCMaps(ge, cvt);

	mlInCheckSession(ge, cvt);

	/* read the info */
	done = FALSE;
	while ( !done && mlInTok(ge, cvt, TRUE) )
	{
		if ( cvt->tok.type != TOK_TOK )
			continue;

		else switch ( cvt->tok.id )
		{
			/* document info */
			case TDOCDIMENSIONS:
				ReadDOCDIMENSIONS(ge, cvt, t, obj);
				cvt->tok.id = TSESSION;
			break;

			case TSTICKY:
				if ( cvt->info.version < 320 )
					ge->state.stickyDots = TRUE;
				else
				{
					mlInTok(ge, cvt, FALSE);
					ge->state.stickyDots = cvt->tok.id;
				}
			break;

			case TAUTOGRID:
				if ( cvt->info.version < 320 )
					ge->state.autoGrid = TRUE;
				else
				{
					mlInTok(ge, cvt, FALSE);
					ge->state.autoGrid = cvt->tok.id;
				}
			break;

			case TVIEW:
				mlInTok(ge, cvt, FALSE);
				ge->state.rulerOn    = cvt->tok.val[0].u.ival;
				ge->state.pageOn     = cvt->tok.val[1].u.ival;
				ge->state.readOut    = cvt->tok.val[2].u.ival;
				ge->state.gridOn     = cvt->tok.val[3].u.ival;
				ge->state.bordersOn  = cvt->tok.val[4].u.ival;
				if ( cvt->tok.n_vals > 5 )
					ge->state.guidesOn = cvt->tok.val[5].u.ival;
			break;

			case TRULERON:
				ge->state.rulerOn = TRUE;
			break;
	
			case TPAGEON:
				ge->state.pageOn = TRUE;
				if ( cvt->info.version < 320 )
				{
					mlInTok(ge, cvt, FALSE);
					mlInTok(ge, cvt, FALSE);
				}
			break;

			case TREADOUT:
				ge->state.readOut = TRUE;
			break;

			case TGRIDS:	
				ge->state.gridOn = TRUE;
			break;

			/* the default image file format */
			case TIMAGE_FMT:
				mlInTok(ge, cvt, FALSE);
				val = cvtTokToInt(ge, cvt);
				switch ( val )
				{
					case 3: ge->state.im_hex = ge->state.im_raw = TRUE; break;
					case 2: ge->state.im_hex = TRUE; break;
					case 1: ge->state.im_raw = TRUE; break;
				}
			break;

			case TGRIDSIZE:
				mlInTok(ge, cvt, FALSE);
				ge->state.gridSize = cvtTokToInt(ge, cvt);
				if ( ge->state.gridSize == 394 )
				{
					/* old cm unit */
					ge->state.unit_size = 394;
					ge->state.unit_precision = 2;
					strcpy(ge->state.unit_name, XLT("cm"));
				}
				else
				{
					ge->state.unit_size = 1000;
					ge->state.unit_precision = 3;
					strcpy(ge->state.unit_name, XLT("in."));
				}
				ge->state.dpi = 1000;
			break;

			case TGRID_DPI:
				mlInTok(ge, cvt, FALSE);
				ge->state.dpi = cvtTokToInt(ge, cvt);
				if ( ge->state.dpi <= 0 )
					ge->state.dpi = 1000;
			break;
	
			case TGRID_FACTOR:
				mlInTok(ge, cvt, FALSE);
				ge->state.gridSize = cvtTokToInt(ge, cvt);
				if ( ge->state.gridSize <= 0 )
					ge->state.gridSize = 250;
			break;
	
			case TGRID_UNITS:
				ge->state.unit_precision = 3;
				mlInTok(ge, cvt, FALSE);
				ge->state.unit_size = cvtTokToInt(ge, cvt);
				if ( ge->state.unit_size <= 0 )
					ge->state.unit_size = 1000;
				mlInTok(ge, cvt, FALSE);
				gfree(ge, ge->state.unit_name);
				ge->state.unit_name = cvtTokToStr(ge, cvt);
			break;

			case TNOTE_SIZE:
				mlInTok(ge, cvt, FALSE);
			break;

			case TCURRENT_PAGE:
				mlInTok(ge, cvt, FALSE);
				cvt->page_ix = cvtTokToInt(ge, cvt) - 1;
			break;
	
			case TSLIDE_CYCLE:
				mlInTok(ge, cvt, FALSE);
				ge->state.wrap_slide_show = cvtTokToInt(ge, cvt);
			break;

			case TSLIDE_STYLE:
				mlInTok(ge, cvt, FALSE);
				ge->state.default_slide_effect = cvt->tok.val[0].u.ival;
				ge->state.default_slide_delay = cvt->tok.val[1].u.ival;
				if ( cvt->tok.n_vals > 2 )
					ge->state.default_slide_throttle = cvt->tok.val[2].u.ival;
			break;

			case TUNITS:
				if ( ge->state.unit_name )
					gfree(ge, ge->state.unit_name);
				if ( cvt->info.version < 320 )
				{
					mlInTok(ge, cvt, FALSE);
					ge->state.unit_size = cvtTokToInt(ge, cvt);
					mlInTok(ge, cvt, FALSE);
					ge->state.unit_precision = cvtTokToInt(ge, cvt);
					mlInTok(ge, cvt, FALSE);
					ge->state.unit_name = cvtTokToStr(ge, cvt);
				}
				else
				{
					mlInTok(ge, cvt, FALSE);
					ge->state.unit_size = cvt->tok.val[0].u.ival;
					ge->state.unit_precision = cvt->tok.val[1].u.ival;
#if !BUILD_ORIENT
					size = (cvt->tok.val[2].u.cval.n_chars+1) * sizeof(char);
#else
                    /* Ashisuto edited for wide chars.  95.10.16 */
                    size = (cvt->tok.val[2].u.cval.n_chars+1) *
                            sizeof(*(cvt->tok.val[2].u.cval.str));
#endif
					ge->state.unit_name = (char *)gloc(ge, size);
					WideToMultiStr(ge, cvt->tok.val[2].u.cval.str,
					               ge->state.unit_name);
				}
			break;
	
			case TWIDTH:
				mlInTok(ge, cvt, FALSE);
				ge->state.line_1 = cvtTokToInt(ge, cvt);
				mlInTok(ge, cvt, FALSE);
				ge->state.line_2 = cvtTokToInt(ge, cvt);
				mlInTok(ge, cvt, FALSE);
				ge->state.line_3 = cvtTokToInt(ge, cvt);
			break;
	
			/* pre-ver 3 attributes */
			case TCHRATTR:
				mlInTok(ge, cvt, FALSE);
				val = cvtTokToInt(ge, cvt);
				cvt->curr_atts.face = CvtAddFont(ge,cvt,fnt[GET_ATTR(val,FONT)]);
				cvt->curr_atts.size = siz2val1[GET_ATTR(val,SIZE)];
				cvt->curr_atts.ruling = GET_ATTR(val,ULIN);
				if ( cvt->curr_atts.ruling & 0x4 )
				{
					cvt->curr_atts.ruling &= ~0x4;
					cvt->curr_atts.strike = TRUE;
				}
				cvt->curr_atts.weight = GET_ATTR(val,BOLD);
				cvt->curr_atts.slant = GET_ATTR(val,ITAL);
			break;
			case TTXTATTR:
				mlInTok(ge, cvt, FALSE);
				val = cvtTokToInt(ge, cvt);
				cvt->curr_atts.h_just = GET_ATTR(val,HJUST);
			break;
			case TLINATTR:
				mlInTok(ge, cvt, FALSE);
				val = cvtTokToInt(ge, cvt);
				cvt->curr_atts.line_thickness = GET_ATTR(val,LSIZ);
				cvt->curr_atts.line_style = GET_ATTR(val,PATN);
			break;
			case TPOLATTR:
				mlInTok(ge, cvt, FALSE);
				val = cvtTokToInt(ge, cvt);
				cvt->curr_atts.fill_pat = GET_ATTR(val,PATN);
				jnk = GET_ATTR(val,WMODE);
			break;
			case TCOLATTR:
				mlInTok(ge, cvt, FALSE);
				val = cvtTokToInt(ge, cvt);
				cvt->curr_atts.fill_fg = GET_ATTR(val,FCOLR) + 1;
				cvt->curr_atts.fill_bg = GET_ATTR(val,BCOLR) + 1;
			break;
			case TDEF_TABSIZE:
				mlInTok(ge, cvt, FALSE);
				ge->state.def_tabsize = cvt->tok.val[0].u.ival;
			break;
	
			default:
				if ( cvtSegmentTerminator(ge, cvt) && cvt->tok.id != TSESSION )
				{
					done = TRUE;
					cvt->untok = TRUE;
				}
				else if ( cvt->tok.id == TEND )
				{
					done = TRUE;	/* get rid of SESSION after end */
					mlInTok(ge, cvt, FALSE); 
				}
				else
					(TokFuncs[cvt->tok.id])(ge, cvt, t, obj);
			break;
		}
	}

	/* done */
	ge->def_attr = cvt->curr_atts;
	n_colors = ge->cmap[0].n_colors;
	if ( ge->def_attr.fill_fg >= n_colors )
		ge->def_attr.fill_fg = 0;
	if ( ge->def_attr.fill_bg >= n_colors )
		ge->def_attr.fill_bg = 0;
	if ( ge->def_attr.line_fg >= n_colors )
		ge->def_attr.line_fg = 0;
	if ( ge->def_attr.line_bg >= n_colors )
		ge->def_attr.line_bg = 0;
	if ( ge->def_attr.shadow_color >= n_colors )
		ge->def_attr.shadow_color = 0;
	if ( ge->def_attr.bullet_color >= n_colors )
		ge->def_attr.bullet_color = 0;
}

/****************************************************************************
 *
 * module:	SetHfAttrs
 */
void
SetHfAttrs(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	cvt->hfstr.face_ix = cvt->curr_atts.face;
	cvt->hfstr.pointsize = cvt->curr_atts.size;
	cvt->hfstr.color_ix = cvt->curr_atts.fill_fg;
	cvt->hfstr.bold = cvt->curr_atts.weight;
	cvt->hfstr.italic = cvt->curr_atts.slant;
	cvt->hfstr.strike_thru = cvt->curr_atts.strike;
	cvt->hfstr.underline = cvt->curr_atts.ruling;
	cvt->hfstr.underline_words_only = cvt->words_only;
}

/****************************************************************************
 *
 * module:	CvtICO - get an inner, center, or outer header,footer line
 */
void
CvtICO(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	cvt_part_creation_type dummy_obj;
	thing dummy_t;
	int done;

	done = FALSE;
	while ( !done && mlInTok(ge, cvt, TRUE) ) switch ( cvt->tok.type )
	{
		case TOK_TOK:
			if ( cvt->tok.id == TEND )
				done = TRUE;
			else
				(TokFuncs[cvt->tok.id])(ge, cvt, &dummy_t, &dummy_obj);
		break;

		case TOK_OTY:
			mlInTok(ge, cvt, FALSE);
			cvt->hfstr.string = cvtTokToStr(ge, cvt);
		break;

		default:
			mlInError(ge, cvt, 0);
		break;
	}
	mlInTok(ge, cvt, FALSE); /* get rid of the matching token */

	SetHfAttrs(ge, cvt);
	memcpy(cvt->ico , &cvt->hfstr, sizeof(ge_hdrftr_str_type));
}

/****************************************************************************
 *
 * module:	CvtReadHdrFtr
 */
void
CvtReadHdrFtr(ge, cvt, nlines, offs)
graphics_main_type *ge;
cvt_global_type *cvt;
int *nlines, *offs;
{
	int done, size, val;

	done = FALSE;
	while ( !done && mlInTok(ge, cvt, TRUE) )
	{
		if ( cvt->tok.type != TOK_TOK )
			continue;

		switch ( cvt->tok.id )
		{
			case TEND:
				done = TRUE;
			break;

			case TINNER:
				mlInTok(ge, cvt, FALSE);
				val = cvtTokToInt(ge, cvt);
				cvt->ico = &cvt->hf[val].inner;
				CvtICO(ge, cvt);
			break;

			case TCENTER:
				mlInTok(ge, cvt, FALSE);
				val = cvtTokToInt(ge, cvt);
				cvt->ico = &cvt->hf[val].center;
				CvtICO(ge, cvt);
			break;

			case TOUTER:
				mlInTok(ge, cvt, FALSE);
				val = cvtTokToInt(ge, cvt);
				cvt->ico = &cvt->hf[val].outer;
				CvtICO(ge, cvt);
			break;

			case TNLINES:
				mlInTok(ge, cvt, FALSE);
				*nlines = cvtTokToInt(ge, cvt);
				size = (*nlines) *sizeof(ge_hdrftr_line_type);
				if( size > 0)
					cvt->hf = (ge_hdrftr_line_type*)gloc(ge,size);
				memset(cvt->hf, 0x0, size);
			break;

			case TOFFSET:
				mlInTok(ge, cvt, FALSE);
				*offs = cvtTokToInt(ge, cvt);
			break;
		}
	}
	mlInTok(ge, cvt, FALSE); /* get rid of the HEADER token */
}

/****************************************************************************
 *
 * module:	CvtHdrFtr
 */
void
CvtHdrFtr(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int done, nlines, offs;

	done = FALSE;
	while ( !done && mlInTok(ge, cvt, TRUE) )
	{
		if ( cvt->tok.type != TOK_TOK )
			continue;

		switch ( cvt->tok.id )
		{
			case TEND:
				done = TRUE;
			break;

			case THEADER:
				CvtReadHdrFtr(ge, cvt, &nlines, &offs);
				cvt->hfpage.hdr = cvt->hf;
				cvt->hfpage.hdr_n_lines = nlines;
				cvt->hfpage.hdr_offset = offs;
				cvt->hf = (ge_hdrftr_line_type *)0;
			break;

			case TFOOTER:
				CvtReadHdrFtr(ge, cvt, &nlines, &offs);
				cvt->hfpage.ftr = cvt->hf;
				cvt->hfpage.ftr_n_lines = nlines;
				cvt->hfpage.ftr_offset = offs;
				cvt->hf = (ge_hdrftr_line_type *)0;
			break;
		}
	}
	mlInTok(ge, cvt, FALSE); /* get rid of the FINAL token */

	/* copy the data onto the header/footer and reset */
	memcpy(cvt->hdr_ftr, &cvt->hfpage, sizeof(ge_hdrftr_page_type));
	memset((char *)&cvt->hfpage, 0, sizeof(ge_hdrftr_page_type));
}

/****************************************************************************
 *
 * module:	ReadWORDSONLY
 */
void
ReadWORDSONLY(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	mlInTok(ge, cvt, FALSE);
	if ( cvt->tok.type == TOK_TST )
		cvt->words_only = cvt->tok.id;
	else
		cvt->words_only = FALSE;
}

/****************************************************************************
 *
 * module:  mlInFaces
 */
void
mlInFaces(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int done;
    char *face_name, *new_name;

	done = FALSE;
	while( !done && mlInTok(ge, cvt, TRUE) )
	{ 
		if ( cvt->tok.type == TOK_TOK && cvt->tok.id == TEND )
			done = TRUE;
		else
		{
			ge->n_faces++;
			ge->face = (char **)regloc(ge, ge->face, ge->n_faces*sizeof(char*));
            face_name = cvtTokToStr(ge, cvt);
#if !FOR_FILTER
            /* in axmain we need to make sure that the names for Adobe 35n */
            /* fonts as used on UNIX get converted to/from the names for   */
            /* these same fonts as enumerated by the postscript drivers on NT */
            new_name = MapFontFamily( face_name, FOREIGN_TO_NATIVE );
            if ( face_name != new_name )
               {
               /* the name was mapped to a local equivalent */
               gfree( ge, face_name );
               face_name = (char *)gloc( ge, (strlen(new_name)+1) );
               MBstrcpy( face_name, new_name );
               }
#endif
			ge->face[ge->n_faces-1] = face_name;
		}
	}
	mlInTok(ge, cvt, FALSE); /* get rid of FONT token after END */
	cvt->faces_found = TRUE;
}

/****************************************************************************
 *
 * module:  mlInCallBacks
 */
void
mlInCallBacks(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int done, len, size, nargs;
	ge_callback_type *cback;

	cvt->cbacks_found = TRUE;
	done = FALSE;
	while ( !done && mlInTok(ge, cvt, TRUE) )
	{
		/* must be arguments */
		if ( cvt->tok.type == TOK_STR )
		{
			nargs = cback->n_args++;
			size = cback->n_args * sizeof(char *);
			cback->args = (char **)regloc(ge, cback->args, size);
			cback->args[nargs] = cvtTokToStr(ge, cvt);
		}

		/* has to be MACRO or END */
		if ( cvt->tok.type != TOK_TOK )
			continue;

		if ( cvt->tok.id == TEND )
			done = TRUE;

		else
		{
			/* init */
			len = ge->n_callbacks++;
			size = ge->n_callbacks * sizeof(ge_callback_type);
			ge->callback = (ge_callback_type *)regloc(ge, ge->callback, size);
			cback = &ge->callback[len];
			memset((char*)cback,0, sizeof(ge_callback_type));
			if ( cvt->info.version < 320 )
			{
				mlInTok(ge, cvt, FALSE);
				cback->name = cvtTokToStr(ge, cvt);

				mlInTok(ge, cvt, FALSE);
				mlInTok(ge, cvt, FALSE);
			}
			else
			{
				mlInTok(ge, cvt, FALSE);
				cback->name = cvtGetStrFromTokVal(ge, cvt, 0);
				cback->on_drag     = cvt->tok.val[1].u.ival;
				cback->on_single   = cvt->tok.val[2].u.ival;
				cback->on_double   = cvt->tok.val[3].u.ival;
			}
		}
	}
	mlInTok(ge, cvt, FALSE); /* get rid of token after END */
}

/****************************************************************************
 *
 * module:  mlInHyperlinks
 */
void
mlInHyperlinks(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	char *path, *target;
	int done;

	cvt->hlinks_found = TRUE;
	done = FALSE;
	while ( !done && mlInTok(ge, cvt, TRUE) )
	{
		if ( cvt->tok.id == TEND )
			done = TRUE;
		else if ( cvt->tok.id == THLINK )
		{
			path = cvtGetStrFromTokVal(ge, cvt, 0);
			target = cvtGetStrFromTokVal(ge, cvt, 1);
		}
	}
	mlInTok(ge, cvt, FALSE); /* get rid of token after END */
}

/****************************************************************************
 *
 * module:
 */
void
mlInNamedGrads(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int ix, sz;
	int byte_wid,data_wid;

	/* better clear here */
	ge->n_grads = 0;
	ge->grad = (udgrad_type *)0;

	/* read the fills into the cleared struct */
	while( mlInTok(ge, cvt, FALSE) != TOK_TOK )
	{
		ix = ge->n_grads++;
		sz = ge->n_grads * sizeof(udgrad_type);
		ge->grad = (udgrad_type *)regloc(ge, ge->grad, sz);

		/* read the header segment */
		ge->grad[ix].name = cvt->tok.val[0].u.cval.str;
		ge->grad[ix].n_chars = cvt->tok.val[0].u.cval.n_chars;
		cvt->tok.val[0].u.cval.str = (tx_char *)0;
		cvt->tok.val[0].u.cval.n_chars = 0;

		ge->grad[ix].type = cvt->tok.val[1].u.ival;
		ge->grad[ix].angle = cvt->tok.val[2].u.ival;
		ge->grad[ix].xoff = cvt->tok.val[3].u.ival;
		ge->grad[ix].yoff = cvt->tok.val[4].u.ival;

	}
	mlInTok(ge, cvt, FALSE); /* get rid of token after END */
}

/****************************************************************************
 *
 * module:	m l I n N a m e d F i l l s
 */
void
mlInNamedFills(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int i, ix, val, done, byte_index, datum, n_chars, (*func)();
	int byte_wid,data_wid, sz;
	udfill_type *f_pack;
	char *buf;
    unsigned char *temp_data, *indataP,*outdataP;

	/* better clear here */
	ge->n_fills = 0;
	ge->fill = (udfill_type *)0;
	if ( cvt->info.version >= 320 )
		mlInCheckCMaps(ge, cvt);

	/* read the fills into the cleared struct */
	while( mlInTok(ge, cvt, FALSE) != TOK_TOK )
	{
		ix = ge->n_fills++;
		sz = sizeof(udfill_type);
		ge->fill = (udfill_type *)regloc(ge, ge->fill, ge->n_fills*sz);
		f_pack = &ge->fill[ix];
		memset(f_pack, 0, sz);

		/* read the header segment */
		if ( cvt->info.version < 320 )
		{
			buf = cvtTokToStr(ge, cvt);
			n_chars = strlen(buf);
			f_pack->name = (tx_char*)gloc(ge, (n_chars+1)*sizeof(tx_char));
			for ( i=0; i<n_chars; i++ )
				f_pack->name[i] = (tx_char)buf[i];
			f_pack->name[i] = (tx_char)0;
			f_pack->n_chars = n_chars;

			mlInTok(ge, cvt, FALSE);
			f_pack->hyt = cvtTokToInt(ge, cvt);

			mlInTok(ge, cvt, FALSE);
			f_pack->wid = cvtTokToInt(ge, cvt);
			f_pack->cmapID = 0;
			byte_wid = (f_pack->wid+7)/ 8;
			func = mailInBit6;
		}
		else
		{
			f_pack->name = cvt->tok.val[0].u.cval.str;
			f_pack->n_chars = cvt->tok.val[0].u.cval.n_chars;
			cvt->tok.val[0].u.cval.str = (tx_char *)0;
			cvt->tok.val[0].u.cval.n_chars = 0;

			f_pack->hyt = cvt->tok.val[1].u.ival;
			f_pack->wid = cvt->tok.val[2].u.ival;
			f_pack->cmapID = cvt->tok.val[3].u.ival;
			if ( f_pack->cmapID > 0 )
				byte_wid = f_pack->wid+f_pack->wid%2;
			else
				byte_wid = (f_pack->wid+7)/ 8+ ((f_pack->wid+7)/ 8)%2;
			func = cvt->tok.val[4].u.ival ? mailInBit6 : mailInHex;
			/* use filll as picture */
			if ( cvt->tok.n_vals > 4 )
				f_pack->is_pict = cvt->tok.val[5].u.ival;
		}

		/* read the data segment */
		byte_index = 0;
		sz = f_pack->hyt * byte_wid;
		f_pack->data = (unsigned char *)gloc(ge, sz);
		memset(f_pack->data, 0x0, sz);
		for ( i=0; i<f_pack->hyt; i++ )
			while ( (datum = (*func)(ge, cvt)) != (-1) )
				f_pack->data[byte_index++] = datum & 0xff;
		if ( cvt->info.version < 320 )
		{
			data_wid = (f_pack->wid+7)/ 8+ ((f_pack->wid+7)/ 8)%2;
			sz = f_pack->hyt * data_wid;
			if ( data_wid != byte_wid )
			{
				temp_data = f_pack->data;
				f_pack->data = (unsigned char *)gloc(ge, sz);

				indataP = temp_data;
				outdataP = f_pack->data;
				for ( i=0;i<f_pack->hyt;i++ )
				{
					memcpy(outdataP,indataP,byte_wid);
					indataP += byte_wid;
					outdataP += data_wid;
				}
				gfree(ge, temp_data);
			}

			for (i=f_pack->hyt*data_wid; i>=0; --i)
				f_pack->data[i] = BitSpin(f_pack->data[i]);
		}
	}
	mlInTok(ge, cvt, FALSE); /* get rid of token after END */
	cvt->fills_found = TRUE;
}

/****************************************************************************
 *
 * module:	mlInHdrFtr
 */
void
mlInHdrFtr(ge, cvt, hf)
graphics_main_type *ge;
cvt_global_type *cvt;
ge_hdrftr_main_type *hf;
{
	int done;

	if ( cvt->info.version >= 320 )
	{
		mlInCheckCMaps(ge, cvt);
		mlInCheckFaces(ge, cvt);
	}

	done = FALSE;
	while ( !done && mlInTok(ge, cvt, TRUE) )
	{
		if ( cvt->tok.type != TOK_TOK )
			continue;

		else switch ( cvt->tok.id )
		{
			case TEND:
				done = TRUE;
			break;
	
			case TUSE:
				mlInTok(ge, cvt, FALSE);
				if ( cvt->tok.id == TFIRST )
					hf->use_first = TRUE;
				else if ( cvt->tok.id == TEVENODD )
					hf->use_even_odd = TRUE;
				else if ( cvt->tok.id == TFINAL )
					hf->use_final = TRUE;
			break;
	
			case TFIRST:
				cvt->hdr_ftr = &hf->first;
				CvtHdrFtr(ge, cvt);
			break;
	
			case TEVEN:
				cvt->hdr_ftr = &hf->even;
				CvtHdrFtr(ge, cvt);
			break;
	
			case TFINAL:
				cvt->hdr_ftr = &hf->final;
				CvtHdrFtr(ge, cvt);
			break;
	
			default:
				cvt->hdr_ftr = &hf->base;
				CvtHdrFtr(ge, cvt);
			break;
		}
	}
	mlInTok(ge, cvt, FALSE); /* get rid of the HDRFTR token */
	cvt->hdrftr_found = TRUE;
}

/****************************************************************************
 *
 * module:	CalcNumBars
 */
void
CalcNumBars(ge,cvt, chart_ix)
graphics_main_type *ge;
cvt_global_type *cvt;
int chart_ix;
{
	int grp_ix, stack_id, stack_ax, i, type, size;
	chart_data_group_type *grp;
	chart_type *ge_chart, *chart;

	ge_chart = (chart_type *)ge->chart;
	chart    = (chart_type*)&ge_chart[chart_ix];
	for ( grp_ix=0; grp_ix<chart->n_groups; grp_ix++ )
	{
		grp = &chart->grp[grp_ix];
		type = grp->type;
		stack_id = grp->stack_id;
		stack_ax = grp->x_axis;
		if	( type == BG_BAR  ||
		      type == BG_LBAR ||
		      type == BG_CBAR ||
		      type == BG_STACKED_CBAR )
		{
			if ( stack_ax < chart->n_x_axes )
				chart->x_axis[stack_ax].n_simple_bars++;
		}

		else if ( type == BG_STACKED_BAR || type == BG_CENT_BAR )
		{
			if ( stack_ax < chart->n_x_axes )
			{
				/* look for the stack in the stack list */
				for ( i=0; i<chart->x_axis[stack_ax].n_stacked_bars; i++ )
				{
					if ( chart->x_axis[stack_ax].stacks[i] == stack_id )
						break;
				}

				/* if it's not there, add it */
				if ( i >= chart->x_axis[stack_ax].n_stacked_bars )
				{
					chart->x_axis[stack_ax].n_stacked_bars++;
					size = chart->x_axis[stack_ax].n_stacked_bars * sizeof(int);
					chart->x_axis[stack_ax].stacks =
					 (int *)regloc(ge, chart->x_axis[stack_ax].stacks, size);
					chart->x_axis[stack_ax].stacks[i] = stack_id;
				}
			}
		}
	}
}

/****************************************************************************
 *
 * module:	mlNewChart
 */
void
mlNewChart(ge, cvt, newchart)
graphics_main_type *ge;
cvt_global_type *cvt;
chart_type **newchart;
{
	chart_type *chart, *ge_chart;
	int chartnum, size;

	chartnum = ge->n_charts++;
	size = ge->n_charts * sizeof(chart_type);
	ge_chart = (chart_type *)regloc(ge, ge->chart, size);
	ge->chart = (char *)ge_chart;
	*newchart = chart = &ge_chart[chartnum];
	memset(chart,0,sizeof(chart_type));
}

/****************************************************************************
 *
 * module:	ReadCHART
 */
void
ReadCHART(ge, cvt, t, obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	int done, chartnum, i, sz, old_count;
	chart_type *chart, *ge_chart;
	thing_attributes *att;
	char *type;

	/* got a chart token - create a chart and read it's name */
	mlNewChart(ge, cvt, &chart);
	mlInTok(ge, cvt, FALSE);
	strcpy(chart->name, cvtTokToStr(ge, cvt));

	/* process the chart data */
	att = (thing_attributes *)0;
	old_count = 0;
	done = FALSE;
	while ( !done && mlInTok(ge, cvt, TRUE) ) switch ( cvt->tok.type )
	{
		case TOK_TOK: switch(cvt->tok.id)
		{
			case TEND:
				done = TRUE;
			break;

			case TTYPE:
				mlInTok(ge, cvt, FALSE);
				if ( cvt->info.version < 320 )
				{
					/* chart type */
					type = cvtTokToStr(ge, cvt);
					if ( !type )
						chart->type = 0;
					else
					{
						for ( i=0; i<6; i++ )
							if ( !strcasecmp(chrtypes[i], type) )
								break;
						chart->type = i % 6;
						gfree(ge, type);
					}
				}
				else
				{
					chart->type = cvt->tok.val[0].u.ival;
					chart->orientation = cvt->tok.val[1].u.ival;
					chart->old_orient  = cvt->tok.val[2].u.ival;
				}
			break;

			case TTITLE:
				mlInTok(ge, cvt, FALSE);
				chart->title = cvtTokToStr(ge, cvt);
			break;

			case TSUBTITLE:
				mlInTok(ge, cvt, FALSE);
				chart->subtitle = cvtTokToStr(ge, cvt);
			break;

			case TFOOT:
				mlInTok(ge, cvt, FALSE);
				chart->footer = cvtTokToStr(ge, cvt);
			break;

			case TTITLE_ALIGN:
				if ( cvt->info.version < 320 )
				{
					mlInTok(ge, cvt, FALSE);
					chart->title_align = cvtTokToInt(ge, cvt);
					mlInTok(ge, cvt, FALSE);
					chart->subtitle_align = cvtTokToInt(ge, cvt);
					mlInTok(ge, cvt, FALSE);
					chart->footer_align = cvtTokToInt(ge, cvt);
					/* These were never written/read in <320 */
					chart->title_expand   = 1;
					chart->subtitle_expand= 1;
					chart->footer_expand  = 1;

				}
				else
				{
					mlInTok(ge, cvt, FALSE);
					chart->title_align    = cvt->tok.val[0].u.ival;
					chart->subtitle_align = cvt->tok.val[1].u.ival;
					chart->footer_align   = cvt->tok.val[2].u.ival;
					chart->title_expand   = cvt->tok.val[3].u.ival;
					chart->subtitle_expand= cvt->tok.val[4].u.ival;
					chart->footer_expand  = cvt->tok.val[5].u.ival;
				}
			break;

			case TNULL_FMT:
				mlInTok(ge, cvt, FALSE);
				switch(cvt->tok.id) /* assume it returned TOK_TOK */
				{
					default:
					case TPLOT: chart->null_format = 0; break;
					case TSUBS: chart->null_format = 1; break;
					case TSPAN: chart->null_format = 2; break;
					case TGAP:  chart->null_format = 3; break;
				}
			break;

			case TTHREE_D:
				mlInTok(ge, cvt, FALSE);
				chart->three_d.enabled    = cvt->tok.val[0].u.ival;
				chart->three_d.pitch      = cvt->tok.val[2].u.ival;
				chart->three_d.yaw        = cvt->tok.val[4].u.ival;
				chart->three_d.projection = cvt->tok.val[6].u.ival;
				chart->three_d.depth      = cvt->tok.val[8].u.ival;

				chart->old_three_d.enabled    = cvt->tok.val[1].u.ival;
				chart->old_three_d.pitch      = cvt->tok.val[3].u.ival;
				chart->old_three_d.yaw        = cvt->tok.val[5].u.ival;
				chart->old_three_d.projection = cvt->tok.val[7].u.ival;
				chart->old_three_d.depth      = cvt->tok.val[9].u.ival;
				if ( cvt->tok.n_vals > 10 )
				{
					chart->three_d.right_angles     = cvt->tok.val[10].u.ival;
					chart->old_three_d.right_angles = cvt->tok.val[11].u.ival;
				}
				else
				{
					chart->three_d.right_angles       = FALSE;
					chart->old_three_d.right_angles   = FALSE;
				}
			break;
				
			case TOLD_GRP:
				mlInTok(ge, cvt, FALSE);
				i = old_count;
				old_count++;
				sz = old_count * sizeof(chart_old_data_group_type);
				chart->old_grp = (chart_old_data_group_type *)
				                  regloc(ge, chart->old_grp, sz);
				chart->old_grp[i].n        = cvt->tok.val[0].u.ival;
				chart->old_grp[i].type     = cvt->tok.val[1].u.ival;
				chart->old_grp[i].stack_id = cvt->tok.val[2].u.ival;
				chart->old_grp[i].stack_ax = cvt->tok.val[3].u.ival;
			break;

			case TCHART_INFO: 
				mlInTok(ge, cvt, FALSE);
				chart->info.stacked_strata = cvt->tok.val[0].u.ival;
				chart->info.bar_margin  = cvt->tok.val[1].u.fval;
				chart->info.bar_overlap = cvt->tok.val[2].u.fval;
				chart->info.start_angle = cvt->tok.val[3].u.ival;
				chart->info.def_type  = cvt->tok.val[4].u.ival;
				chart->data_src.type  = cvt->tok.val[5].u.ival;
				chart->data_src.ix    = cvt->tok.val[6].u.ival;       
		    break;
			case TMISC_CHART_INFO: 
				mlInTok(ge, cvt, FALSE);
				chart->needs_redraw = cvt->tok.val[0].u.ival;
				if ( cvt->tok.n_vals > 1 )
						chart->col_sliced = cvt->tok.val[1].u.ival;
		    break;
			case TBAR_INFO:
				if ( cvt->info.version < 320 )
				{
					mlInTok(ge, cvt, FALSE);
					chart->info.bar_margin = cvtTokToFlt(ge, cvt);
					mlInTok(ge, cvt, FALSE);
					chart->info.bar_overlap = cvtTokToFlt(ge, cvt);
				}
				else
				{
					mlInTok(ge, cvt, FALSE);
					chart->info.bar_margin  = cvt->tok.val[0].u.fval;
					chart->info.bar_overlap = cvt->tok.val[1].u.fval;
				}
			break;

			case TDEFATR:
				if ( att )
					memcpy(att,&cvt->curr_atts,sizeof(thing_attributes));
				att = &chart->def_attr;
			break;

			case TMAJHRZ:
				if ( att )
					memcpy(att,&cvt->curr_atts,sizeof(thing_attributes));
				att = &chart->maj_x_attr;
			break;

			case TMINHRZ:
				if ( att )
					memcpy(att,&cvt->curr_atts,sizeof(thing_attributes));
				att = &chart->min_x_attr;
			break;

			case TMAJVRT:
				if ( att )
					memcpy(att,&cvt->curr_atts,sizeof(thing_attributes));
				att = &chart->maj_y_attr;
			break;

			case TMINVRT:
				if ( att )
					memcpy(att,&cvt->curr_atts,sizeof(thing_attributes));
				att = &chart->min_y_attr;
			break;
				
			case TMAJDPT:
				if ( att )
					memcpy(att,&cvt->curr_atts,sizeof(thing_attributes));
				att = &chart->maj_z_attr;
			break;

				
			case TMINDPT:
				if ( att )
					memcpy(att,&cvt->curr_atts,sizeof(thing_attributes));
				att = &chart->min_z_attr;
			break;
			

			case TTITLEATR:
				if ( att )
					memcpy(att,&cvt->curr_atts,sizeof(thing_attributes));
				att = &chart->title_attr;
			break;

			case TSUBTITLEATR:
				if ( att )
					memcpy(att,&cvt->curr_atts,sizeof(thing_attributes));
				att = &chart->subtitle_attr;
			break;

			case TFOOTATR:
				if ( att )
					memcpy(att,&cvt->curr_atts,sizeof(thing_attributes));
				att = &chart->footer_attr;
			break;
			default:
				(TokFuncs[cvt->tok.id])(ge, cvt, t, chart);
			break;
		}
		break;

		case TOK_STR: /* got a new chart */
			mlNewChart(ge, cvt, &chart);
			strcpy(chart->name, cvtTokToStr(ge, cvt));
		break;

		case TOK_STS: /* Just a tag indicating that this file has charts */
		return;

		default: /* bad chart! */
			mlFreeCharts(ge);
		break;
	}

	if ( att )
		memcpy(att, &cvt->curr_atts, sizeof(thing_attributes));
	mlInTok(ge, cvt, FALSE); /* get rid of the CHART token */
}
/****************************************************************************
 *
 * module:	mlInCHART
 */
void
mlInCHART(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	cvt_part_creation_type obj;
	thing t;

	if ( cvt->info.version >= 320 )
	{
		mlInCheckCMaps(ge, cvt);
		mlInCheckFaces(ge, cvt);
		mlInCheckCBacks(ge, cvt);
		mlInCheckFills(ge, cvt);
		mlInCheckStyles(ge, cvt);
		mlInCheckSession(ge, cvt);
	}
	ReadCHART(ge, cvt, &t, &obj);
}

char *dump_chart0(ge)
graphics_main_type *ge;
{
	chart_type *ge_chart, *chart;
	if ( ge->n_charts > 0 )
	{
		ge_chart = (chart_type *)ge->chart;
		chart = &ge_chart[0];
		return(chart->name);
	}
	else
		return((char *)0);
}

/****************************************************************************
 *
 * module:	c v t R e p a i r C h a r t
 */
void
cvtRepairAttr(ge, attr)
graphics_main_type *ge;
thing_attributes *attr;
{
	attr->field_space  = attr->size * 1000.0 / 720000.0;
	attr->space_inside = attr->field_space;
	attr->space_above  = 0;
	attr->space_below  = 0;
	attr->l_margin = attr->field_space / 2.0;
	attr->r_margin = attr->field_space / 2.0;
	attr->t_margin = attr->field_space;
	attr->b_margin = attr->field_space / 4.0;
}

/****************************************************************************
 *
 * module:	cvtRepairChart
 */
cvtRepairChart(ge, cvt, chart_t)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *chart_t;
{
	chart_type *ge_chart, *chart;
	chart_axis_type *axis;
	int stat, i, sz;
	
	/* get the chart structure */
	ge_chart = (chart_type *)ge->chart;
	chart = &ge_chart[chart_t->chart_ix-1];
	stat = AOK;

	/* if there is no version info, get rid of the thing */
	if ( cvt->vers_found == FALSE )
		stat = !AOK;   

	/* else if we're less than 3.20, fix the chart and get rid of the thing */
	else if ( cvt->info.version < 400 )
	{
		/* fix textbox line spacing and margins */
		cvtRepairAttr(ge, &chart->def_attr);
		cvtRepairAttr(ge, &chart->title_attr);
		cvtRepairAttr(ge, &chart->subtitle_attr);
		cvtRepairAttr(ge, &chart->footer_attr);
		cvtRepairAttr(ge, &chart->maj_x_attr);
		cvtRepairAttr(ge, &chart->min_x_attr);
		cvtRepairAttr(ge, &chart->maj_y_attr);
		cvtRepairAttr(ge, &chart->min_y_attr);
		cvtRepairAttr(ge, &chart->legend.box_attr);
		cvtRepairAttr(ge, &chart->legend.label_attr);
		cvtRepairAttr(ge, &chart->legend.title_attr);
		for ( i=0; i<chart->n_x_axes; i++ )
		{			
			cvtRepairAttr(ge, &chart->x_axis[i].attr);
			cvtRepairAttr(ge, &chart->x_axis[i].label_attr);
			cvtRepairAttr(ge, &chart->x_axis[i].tik_label_attr);
		}
		for ( i=0; i<chart->n_y_axes; i++ )
		{
			cvtRepairAttr(ge, &chart->y_axis[i].attr);
			cvtRepairAttr(ge, &chart->y_axis[i].label_attr);
			cvtRepairAttr(ge, &chart->y_axis[i].tik_label_attr);
		}
		for ( i=0; i<chart->n_z_axes; i++ )
		{
			cvtRepairAttr(ge, &chart->z_axis[i].attr);
			cvtRepairAttr(ge, &chart->z_axis[i].label_attr);
			cvtRepairAttr(ge, &chart->z_axis[i].tik_label_attr);
		}
		for ( i=0; i<chart->n_groups; i++ )
		{
			cvtRepairAttr(ge, &chart->grp[i].label_attr);
			cvtRepairAttr(ge, &chart->grp[i].attr);
		}

		if ( cvt->info.version < 320 )
		{
			chart->three_d.enabled    = 0;
			chart->three_d.pitch      = -45;
			chart->three_d.yaw        = -10;
			chart->three_d.projection = 0;
			chart->three_d.depth      = 50;
		
			chart->old_three_d.enabled    = 0;
			chart->old_three_d.pitch      = -45;
			chart->old_three_d.yaw        = -10;
			chart->old_three_d.projection = 0;
			chart->old_three_d.depth      = 50;
		
			for ( i=0; i<chart->n_x_axes; i++ )
			{			
				axis = &chart->x_axis[i];
				if ( axis->tik_str && axis->minor_tik_format == 0 )
					axis->minor_tik_format = 1;
				axis->tight_label = axis->tight_tiks = FALSE;
			}

			for ( i=0; i<chart->n_y_axes; i++ )
			{			
				axis = &chart->y_axis[i];
				if ( axis->tik_str && axis->minor_tik_format == 0 )
					axis->minor_tik_format = 1;
			}
		}

		/* get rid of the old rendering */
		chart->needs_redraw = TRUE;      
		stat = !AOK;
	}
	else if ( cvt->info.version <= 420 )
		chart->needs_redraw = TRUE;   
	
	/* done */
	if ( stat != AOK )
	{
#if !FOR_FILTER
		disposeThing(ge, chart_t, TRUE);
#else
		/* turn the chart into a group (filters can't rebuild the chart!) */
		if ( cvt->as_1_1 ) /* final destination will be a *.ag file */
			disposeThing(ge, chart_t, TRUE);
		else
		{
			chart_t->flags.isa_chart = FALSE;
			chart_t->chart_ix = 0;
			stat = AOK;
		}
#endif
		chart->old_n_x_axes = 0;
		chart->old_n_y_axes = 0;
		chart->old_n_z_axes = 0;
		chart->old_n_groups = 0;
		chart->old_type = 0;
		chart->t = (thing *)0;
	}
	return(stat);
}

/****************************************************************************
 *
 * module:	m l I n C o l o r m a p
 */
void
mlInColormap(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int ink, c, y, m, k, see_thru, size;
	char *name;

	/* allocate space for the new colormap */
	size = (ge->n_cmaps+1) * sizeof(cmap_type);
	ge->cmap = (cmap_type *)regloc(ge, ge->cmap, size);
	memset((char *)&ge->cmap[ge->n_cmaps], 0, sizeof(cmap_type));
	ge->n_cmaps++;

	/* get the colors from the file, a line at a time */
	while ( mlInTok(ge, cvt, TRUE) != TOK_TOK )
	{ 
		if ( cvt->info.version < 320 )
		{
			/* copy the name already read in by mlInTok above */
			name = cvtTokToStr(ge, cvt);

			/* get ink type */
			mlInTok(ge, cvt, FALSE);
			ink = cvtTokToInt(ge, cvt);

			/* get color definition */
			mlInTok(ge, cvt, FALSE);
			c = cvtTokToInt(ge, cvt);
			mlInTok(ge, cvt, FALSE);
			y = cvtTokToInt(ge, cvt);
			mlInTok(ge, cvt, FALSE);
			m = cvtTokToInt(ge, cvt);
			mlInTok(ge, cvt, FALSE);
			k = cvtTokToInt(ge, cvt);
			mlInTok(ge, cvt, FALSE);
			see_thru = cvtTokToInt(ge, cvt);
		}
		else
		{
			/* get the name first */
#if !BUILD_ORIENT
			size = (cvt->tok.val[0].u.cval.n_chars+1) * sizeof(char);
#else
            /* Ashisuto edited for wide chars.  95.10.16 */
            size = (cvt->tok.val[0].u.cval.n_chars+1) *
                    sizeof(*(cvt->tok.val[0].u.cval.str));
#endif
			name = (char *)gloc(ge, size);
			WideToMultiStr(ge, cvt->tok.val[0].u.cval.str, name);

			/* get ink type */
			ink = cvt->tok.val[1].u.ival;

			/* get color definition */
			c = cvt->tok.val[2].u.ival;
			y = cvt->tok.val[3].u.ival;
			m = cvt->tok.val[4].u.ival;
			k = cvt->tok.val[5].u.ival;
			see_thru = cvt->tok.val[6].u.ival;
		}

		/* add color to the map */
		mlLoadCmapEntry(ge, ge->n_cmaps-1, name, ink, c, y, m, k, see_thru);
		if ( name )
			gfree(ge, name);
	}
	mlInTok(ge, cvt, FALSE); /* get rid of COLORMAP token after END */
	cvt->cmaps_found = TRUE;
}

/****************************************************************************
 *
 * module:	m l C m a p 2 5 6
 */
void
mlCmap256(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int size;

	size = (ge->n_cmaps+1) * sizeof(cmap_type);
	ge->cmap = (cmap_type *)regloc(ge, ge->cmap, size);
	memset((char *)&ge->cmap[ge->n_cmaps], 0, sizeof(cmap_type));
	ge->n_cmaps++;
	cvt->cmaps_found = TRUE;
	mlLoadAxImDefCmap(ge, cvt, ge->n_cmaps-1);
}

/****************************************************************************
 *
 * module:	m l O l d S t y l e I m g C m a p
 */
void
mlOldStyleImgCmap(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int nm, size;

	size = (ge->n_cmaps+1) * sizeof(cmap_type);
	ge->cmap = (cmap_type*)regloc(ge, ge->cmap, size);
	memset((char *)&ge->cmap[ge->n_cmaps], 0, sizeof(cmap_type));
	nm = ge->n_cmaps++;
	cvt->cmaps_found = TRUE;

	/* Create the (hopefully) equivalent for pre 3.0 files */
	/* which contain color images */
	mlLoadCmapEntry(ge, nm, XLT("color000"), 0, 0, 0, 0, 255, 0);
	mlLoadCmapEntry(ge, nm, XLT("color001"), 0, 65, 65, 0, 190, 0);
	mlLoadCmapEntry(ge, nm, XLT("color002"), 0, 123, 123, 0, 132, 0);
	mlLoadCmapEntry(ge, nm, XLT("color003"), 0, 197, 197, 0, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color004"), 0, 255, 255, 0, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color005"), 0, 65, 0, 65, 190, 0);
	mlLoadCmapEntry(ge, nm, XLT("color006"), 0, 65, 0, 0, 190, 0);
	mlLoadCmapEntry(ge, nm, XLT("color007"), 0, 123, 58, 0, 132, 0);
	mlLoadCmapEntry(ge, nm, XLT("color008"), 0, 197, 132, 0, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color009"), 0, 255, 190, 0, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color010"), 0, 123, 0, 123, 132, 0);
	mlLoadCmapEntry(ge, nm, XLT("color011"), 0, 123, 0, 58, 132, 0);
	mlLoadCmapEntry(ge, nm, XLT("color012"), 0, 123, 0, 0, 132, 0);
	mlLoadCmapEntry(ge, nm, XLT("color013"), 0, 197, 74, 0, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color014"), 0, 255, 132, 0, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color015"), 0, 197, 0, 197, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color016"), 0, 197, 0, 132, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color017"), 0, 197, 0, 74, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color018"), 0, 197, 0, 0, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color019"), 0, 255, 58, 0, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color020"), 0, 255, 0, 255, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color021"), 0, 255, 0, 190, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color022"), 0, 255, 0, 132, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color023"), 0, 255, 0, 58, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color024"), 0, 255, 0, 0, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color025"), 0, 0, 65, 65, 190, 0);
	mlLoadCmapEntry(ge, nm, XLT("color026"), 0, 0, 65, 0, 190, 0);
	mlLoadCmapEntry(ge, nm, XLT("color027"), 0, 58, 123, 0, 132, 0);
	mlLoadCmapEntry(ge, nm, XLT("color028"), 0, 132, 197, 0, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color029"), 0, 190, 255, 0, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color030"), 0, 0, 0, 65, 190, 0);
	mlLoadCmapEntry(ge, nm, XLT("color031"), 0, 0, 0, 0, 190, 0);
	mlLoadCmapEntry(ge, nm, XLT("color032"), 0, 58, 58, 0, 132, 0);
	mlLoadCmapEntry(ge, nm, XLT("color033"), 0, 132, 132, 0, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color034"), 0, 190, 190, 0, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color035"), 0, 58, 0, 123, 132, 0);
	mlLoadCmapEntry(ge, nm, XLT("color036"), 0, 58, 0, 58, 132, 0);
	mlLoadCmapEntry(ge, nm, XLT("color037"), 0, 58, 0, 0, 132, 0);
	mlLoadCmapEntry(ge, nm, XLT("color038"), 0, 132, 74, 0, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color039"), 0, 190, 132, 0, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color040"), 0, 132, 0, 197, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color041"), 0, 132, 0, 132, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color042"), 0, 132, 0, 74, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color043"), 0, 132, 0, 0, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color044"), 0, 190, 58, 0, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color045"), 0, 190, 0, 255, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color046"), 0, 190, 0, 190, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color047"), 0, 190, 0, 132, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color048"), 0, 190, 0, 58, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color049"), 0, 190, 0, 0, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color050"), 0, 0, 123, 123, 132, 0);
	mlLoadCmapEntry(ge, nm, XLT("color051"), 0, 0, 123, 58, 132, 0);
	mlLoadCmapEntry(ge, nm, XLT("color052"), 0, 0, 123, 0, 132, 0);
	mlLoadCmapEntry(ge, nm, XLT("color053"), 0, 74, 197, 0, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color054"), 0, 132, 255, 0, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color055"), 0, 0, 58, 123, 132, 0);
	mlLoadCmapEntry(ge, nm, XLT("color056"), 0, 0, 58, 58, 132, 0);
	mlLoadCmapEntry(ge, nm, XLT("color057"), 0, 0, 58, 0, 132, 0);
	mlLoadCmapEntry(ge, nm, XLT("color058"), 0, 74, 132, 0, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color059"), 0, 132, 190, 0, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color060"), 0, 0, 0, 123, 132, 0);
	mlLoadCmapEntry(ge, nm, XLT("color061"), 0, 0, 0, 58, 132, 0);
	mlLoadCmapEntry(ge, nm, XLT("color062"), 0, 0, 0, 0, 132, 0);
	mlLoadCmapEntry(ge, nm, XLT("color063"), 0, 74, 74, 0, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color064"), 0, 132, 132, 0, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color065"), 0, 74, 0, 197, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color066"), 0, 74, 0, 132, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color067"), 0, 74, 0, 74, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color068"), 0, 74, 0, 0, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color069"), 0, 132, 58, 0, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color070"), 0, 132, 0, 255, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color071"), 0, 132, 0, 190, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color072"), 0, 132, 0, 132, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color073"), 0, 132, 0, 58, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color074"), 0, 132, 0, 0, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color075"), 0, 0, 197, 197, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color076"), 0, 0, 197, 132, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color077"), 0, 0, 197, 74, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color078"), 0, 0, 197, 0, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color079"), 0, 58, 255, 0, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color080"), 0, 0, 132, 197, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color081"), 0, 0, 132, 132, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color082"), 0, 0, 132, 74, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color083"), 0, 0, 132, 0, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color084"), 0, 58, 190, 0, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color085"), 0, 0, 74, 197, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color086"), 0, 0, 74, 132, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color087"), 0, 0, 74, 74, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color088"), 0, 0, 74, 0, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color089"), 0, 58, 132, 0, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color090"), 0, 0, 0, 197, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color091"), 0, 0, 0, 132, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color092"), 0, 0, 0, 74, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color093"), 0, 0, 0, 0, 58, 0);
	mlLoadCmapEntry(ge, nm, XLT("color094"), 0, 58, 58, 0, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color095"), 0, 58, 0, 255, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color096"), 0, 58, 0, 190, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color097"), 0, 58, 0, 132, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color098"), 0, 58, 0, 58, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color099"), 0, 58, 0, 0, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color100"), 0, 0, 255, 255, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color101"), 0, 0, 255, 190, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color102"), 0, 0, 255, 132, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color103"), 0, 0, 255, 58, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color104"), 0, 0, 255, 0, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color105"), 0, 0, 190, 255, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color106"), 0, 0, 190, 190, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color107"), 0, 0, 190, 132, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color108"), 0, 0, 190, 58, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color109"), 0, 0, 190, 0, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color110"), 0, 0, 132, 255, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color111"), 0, 0, 132, 190, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color112"), 0, 0, 132, 132, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color113"), 0, 0, 132, 58, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color114"), 0, 0, 132, 0, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color115"), 0, 0, 58, 255, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color116"), 0, 0, 58, 190, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color117"), 0, 0, 58, 132, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color118"), 0, 0, 58, 58, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color119"), 0, 0, 58, 0, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color120"), 0, 0, 0, 255, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color121"), 0, 0, 0, 190, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color122"), 0, 0, 0, 132, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color123"), 0, 0, 0, 58, 0, 0);
	mlLoadCmapEntry(ge, nm, XLT("color124"), 0, 0, 0, 0, 0, 0);
}

void
mlInDOCDIMENSIONS(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	cvt_part_creation_type obj;
	thing t;

	ReadDOCDIMENSIONS(ge, cvt, &t, &obj);
}

/****************************************************************************
 *
 * module:	mlInBuildInfo
 */
void
mlInBuildInfo(ge, cvt, build, enable_build)
graphics_main_type *ge;
cvt_global_type *cvt;
ge_build_info_type *build;
int *enable_build;
{
	mlInTok(ge, cvt, TRUE);
	*enable_build             = cvt->tok.val[0].u.ival;
	build->use_reverse_build  = cvt->tok.val[1].u.ival;
	build->build_effect       = cvt->tok.val[2].u.ival;
	build->build_rule         = cvt->tok.val[3].u.ival;
	build->build_size         = cvt->tok.val[4].u.ival;
	build->build_delay        = cvt->tok.val[5].u.ival;
	build->use_inactive_color = cvt->tok.val[6].u.ival;
	build->inactive_color     = cvt->tok.val[7].u.ival;
	build->sequence           = cvt->tok.val[8].u.ival;
	build->build_throttle     = cvt->tok.val[9].u.ival;
	build->baudio.play        = cvt->tok.val[10].u.ival;
	build->baudio.type        = cvt->tok.val[11].u.ival;
	build->baudio.id          = cvt->tok.val[12].u.ival;
	build->baudio.volume      = cvt->tok.val[13].u.ival;
}

/****************************************************************************
 *
 * module:	m l I n S l i d e I n f o - read the SLIDE_INFO...END SLIDE_INFO
 */
void
mlInSlideInfo(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int done, type, sz, i, val;

	done = FALSE;
	cvtUpdateStatus(ge, cvt);
	cvt->slide_effect = 0;
	cvt->slide_delay = 5;
	cvt->slide_throttle = 1;
	while ( !done && !cvt->done )
	{ 
		type = mlInTok(ge, cvt, TRUE);
		if ( type == TOK_TOK )
		{
			/* if it's an END, assume following SLIDE_INFO and toss it */
			if ( cvt->tok.id == TEND )
			{
				mlInTok(ge, cvt, FALSE);
				done = TRUE;
			}

			/* otherwise evaluate the token */
			else switch ( cvt->tok.id )
			{
				case TSLIDE_OUTLINE:
					mlInTok(ge, cvt, FALSE);
					cvt->outline_text = cvt->tok.val[0].u.cval.str;
					cvt->n_outline_chars = cvt->tok.val[0].u.cval.n_chars;
					cvt->tok.val[0].u.cval.str = (tx_char *)0;
					cvt->tok.val[0].u.cval.n_chars = 0;
				break;

				case TSLIDE_STYLE:
					mlInTok(ge, cvt, FALSE);
					cvt->slide_effect = cvt->tok.val[0].u.ival;
					cvt->slide_delay = cvt->tok.val[1].u.ival;
					if ( cvt->tok.n_vals > 2 )
						cvt->slide_throttle = cvt->tok.val[2].u.ival;
				break;

				case TLAYOUT_ID:
					mlInTok(ge, cvt, FALSE);
					cvt->layout_ix = cvt->tok.val[0].u.ival;
					cvt->object1 = cvt->tok.val[1].u.ival;
					cvt->object2 = cvt->tok.val[2].u.ival;
					cvt->object3 = cvt->tok.val[3].u.ival;
					cvt->object4 = cvt->tok.val[4].u.ival;
				break;

				case THIDE_OBJECTS:
					mlInTok(ge, cvt, FALSE);
					cvt->hide_objs = cvt->tok.val[0].u.ival;
				break;

				case THIDE_BACKGD:
					mlInTok(ge, cvt, FALSE);
					cvt->hide_pg_backgd = cvt->tok.val[0].u.ival;
				break;

				case TPAGE_BACKGD:
					mlInTok(ge, cvt, FALSE);
					cvt->c_scheme.page_pat = cvt->tok.val[0].u.ival;
					cvt->c_scheme.page_type = cvt->tok.val[1].u.ival;
					cvt->c_scheme.page_fg  = cvt->tok.val[2].u.ival;
					cvt->c_scheme.page_bg  = cvt->tok.val[3].u.ival;
					cvt->use_backgd = TRUE;
				break;

				case TC_SCHEME:
					mlInTok(ge, cvt, FALSE);
					cvt->c_scheme.title_fg = cvt->tok.val[0].u.ival;
					cvt->c_scheme.title_bg = cvt->tok.val[1].u.ival;
					cvt->c_scheme.text_fg  = cvt->tok.val[2].u.ival;
					cvt->c_scheme.text_bg  = cvt->tok.val[3].u.ival;
					cvt->c_scheme.line_fg  = cvt->tok.val[4].u.ival;
					cvt->c_scheme.line_bg  = cvt->tok.val[5].u.ival;
					cvt->c_scheme.fill_fg  = cvt->tok.val[6].u.ival;
					cvt->c_scheme.fill_bg  = cvt->tok.val[7].u.ival;
					cvt->c_scheme.accent_color[0] = cvt->tok.val[8].u.ival;
					cvt->c_scheme.accent_color[1] = cvt->tok.val[9].u.ival;
					cvt->c_scheme.accent_color[2] = cvt->tok.val[10].u.ival;
					cvt->c_scheme.accent_color[3] = cvt->tok.val[11].u.ival;
					cvt->c_scheme.accent_color[4] = cvt->tok.val[12].u.ival;
					cvt->c_scheme.accent_color[5] = cvt->tok.val[13].u.ival;
					cvt->c_scheme.accent_color[6] = cvt->tok.val[14].u.ival;
					cvt->c_scheme.accent_color[7] = cvt->tok.val[15].u.ival;
					cvt->c_scheme.accent_color[8] = cvt->tok.val[16].u.ival;
					cvt->c_scheme.accent_color[9] = cvt->tok.val[17].u.ival;
					cvt->c_scheme.accent_color[10] = cvt->tok.val[18].u.ival;
					cvt->c_scheme.accent_color[11] = cvt->tok.val[19].u.ival;
					cvt->c_scheme.accent_color[12] = cvt->tok.val[20].u.ival;
					cvt->c_scheme.accent_color[13] = cvt->tok.val[21].u.ival;
					cvt->c_scheme.accent_color[14] = cvt->tok.val[22].u.ival;
					cvt->c_scheme.shadow_color = cvt->tok.val[23].u.ival;
					if ( cvt->tok.n_vals > 24 )
					{
						cvt->c_scheme.n_other_colors = cvt->tok.n_vals - 24;
						sz = cvt->c_scheme.n_other_colors*sizeof(int);
						cvt->c_scheme.other_color = (int *)
									  regloc(ge, cvt->c_scheme.other_color, sz);
						for ( i=24; i<cvt->tok.n_vals; i++ )
						{
							val = cvt->tok.val[i].u.ival;
							if ( val >= ge->cmap[0].n_colors )
								cvt->c_scheme.other_color[i-24] = 0;
							else
								cvt->c_scheme.other_color[i-24] = val;
						 }
					 }
					cvt->use_color = TRUE;
				break;
				case THIDE_SLIDE:
					mlInTok(ge, cvt, FALSE);
					cvt->hide_slide = cvt->tok.val[0].u.ival;
				break;
			}
		}
	}

	/* note that we found the slide info */
	cvt->slide_found = TRUE;
}

/****************************************************************************
 *
 * module:	m l I n L a y e r s - read the LAYERS...END LAYERS data
 */
void
mlInLayers(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int ix, size;

	/* get the layers from the file, a line at a time */
	cvtUpdateStatus(ge, cvt);
	while( mlInTok(ge, cvt, TRUE) != TOK_TOK )
	{ 
		/* allocate space for the new layer */
		ix = cvt->n_layers;
		cvt->n_layers++;
		size = cvt->n_layers * sizeof(layer_type);
		cvt->layer = (layer_type *)regloc(ge, cvt->layer, size);
		
		/* move the name already read in by mlInTok above */
		if ( cvt->info.version < 320 )
		{
			if ( cvt->tok.type != TOK_STR )
			{
				cvt->layer[ix].name = (tx_char *)0;
				cvt->layer[ix].n_chars = 0;
			}
			else
			{
				/* do we need to null terminate? */
				cvt->layer[ix].name = cvt->tok.val[0].u.cval.str;
				cvt->layer[ix].n_chars = cvt->tok.val[0].u.cval.n_chars;
				cvt->tok.val[0].u.cval.str = (tx_char *)0;
				cvt->tok.val[0].u.cval.n_chars = 0;
			}

			/* get locked flag */
			mlInTok(ge, cvt, FALSE);
			cvt->layer[ix].locked = cvtTokToInt(ge, cvt);

			/* get hidden flag */
			mlInTok(ge, cvt, FALSE);
			cvt->layer[ix].hidden = cvtTokToInt(ge, cvt);

			/* get the print flag */
			if ( cvt->info.version < 312 )
				cvt->layer[ix].prints = TRUE;
			else
			{
				mlInTok(ge, cvt, FALSE);
				cvt->layer[ix].prints = cvtTokToInt(ge, cvt);
			}
			cvt->layer[ix].backgd = FALSE;
		}
		else
		{
			/* do we need to null terminate? */
			cvt->layer[ix].name = cvt->tok.val[0].u.cval.str;
			cvt->layer[ix].n_chars = cvt->tok.val[0].u.cval.n_chars;
			cvt->tok.val[0].u.cval.str = (tx_char *)0;
			cvt->tok.val[0].u.cval.n_chars = 0;

			cvt->layer[ix].locked = cvt->tok.val[1].u.ival;
			cvt->layer[ix].hidden = cvt->tok.val[2].u.ival;
			cvt->layer[ix].prints = cvt->tok.val[3].u.ival;
			if ( cvt->tok.n_vals > 4 )
				cvt->layer[ix].backgd = cvt->tok.val[4].u.ival;
			else
				cvt->layer[ix].backgd = FALSE;
		}
	}
	mlInTok(ge, cvt, FALSE); /* get rid of LAYERS token after END */

	/* note that we found layers */
	cvt->layers_found = TRUE;
}

/****************************************************************************
 *
 * module:	
 */
void
mlInPageAudioInfo(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	while( mlInTok(ge, cvt, TRUE) != TOK_TOK )
	{ 
		cvt->audio.play   = cvt->tok.val[0].u.ival;
		cvt->audio.type   = cvt->tok.val[1].u.ival;
		cvt->audio.id     = cvt->tok.val[2].u.ival;
		cvt->audio.volume = cvt->tok.val[3].u.ival;
	}
	mlInTok(ge, cvt, FALSE); /* get rid of AUDIO1 token after END */

	if ( cvt->info.version < 430 )
	{
		if ( cvt->audio.type == 0 )
			cvt->audio.type = INSET_LINKED;
	}

	/* note that we found the audio info */
	cvt->audio_found = TRUE;
}

/****************************************************************************
 *
 * module:	m l I n P i c t - read the PICTURE...END PICTURE data
 */
void
mlInPict(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	cvt_part_creation_type dummy_obj;
	thing *t, dummy_t;
	chart_type *chart;
	int done, val, size;

	if ( cvt->info.version < 320 )
		mlInCheckCMaps(ge, cvt);
	else
	{
		mlInCheckCMaps(ge, cvt);
		mlInCheckFaces(ge, cvt);
		mlInCheckCBacks(ge, cvt);
		mlInCheckFills(ge, cvt);
		mlInCheckStyles(ge, cvt);
		mlInCheckSession(ge, cvt);
	}

	/* create a new page (if a temp page wasn`t needed) */
	if ( cvt->inserted_page )
		cvt->inserted_page = FALSE;
	else
	{
		cvtAppendPage(ge, cvt, &cvt->curr_atts);
		mlInCheckPage(ge, cvt, &ge->page[ge->page_ix]);
	}

	cvt->lastobj = NULL;
	done = FALSE;
	cvtUpdateStatus(ge, cvt);
	while ( !done && !cvt->done ) switch ( mlInTok(ge, cvt, TRUE) )
	{
		/* if it's a non-null object, read it */
		case TOK_OTY:
			t = cvt->tok.id ? mlInPart(ge, cvt, cvt->tok.id) : (thing *)0;
			if ( t )
			{
				if ( cvt->isa_chart )
				{
					if ( t->type == GROUP && cvtRepairChart(ge,cvt,t) != AOK )
						t = (thing *)0;
					cvt->isa_chart = FALSE;
					memset(&cvt->ch_data, 0, sizeof(chart_info));
				}
				cvtAddToPaintList(ge, t, -1);
			}
			cvtUpdateStatus(ge, cvt);
		break;

		case TOK_TOK:
			if ( cvt->tok.id == TEND )
			{
				mlInTok(ge, cvt, FALSE);
				done = TRUE;
			}
			else if ( cvt->tok.id == TCHART )
			{
				mlInTok(ge, cvt, FALSE); /* get the chart number */
				val = cvtTokToInt(ge, cvt);
				cvt->isa_chart = TRUE;
				if ( val <= ge->n_charts )
					setChartData(ge, cvt, (thing *) 0, 
								 (cvt_part_creation_type *) 0, val+1);
				t = mlInPart(ge, cvt, GROUP);
				if ( t )
				{
					/* set the chart id */
					t->chart_ix = val+1;

					/* fix the chart (if needed) */
					if ( cvtRepairChart(ge, cvt, t) == AOK )
						cvtAddToPaintList(ge, t, -1);
				}
				cvt->isa_chart = FALSE;
				memset(&cvt->ch_data, 0, sizeof(chart_info));
			}
			else
				(TokFuncs[cvt->tok.id])(ge, cvt, &dummy_t, &dummy_obj);
			cvtUpdateStatus(ge, cvt);
		break;

		default:
			cvtUpdateStatus(ge, cvt);
		break;
	}

	/* set the group size of the paint list */
	ge->t->type = GROUP;
	memset((char *)&ge->t->flags, 0, sizeof(thing_mark_type));

	/* if we need to calculate a picture size, we can do it here */
	calcExtents(ge, ge->t);

	/* set the area used by graphics */
	if ( cvt->pict_found )
	{
		unionExtents(ge, ge->t, &ge->t->extent, &ge->t->ink, TRUE);
	}
	else
	{
		ge->state.total_area.x1 = ge->t->extent.x1;
		ge->state.total_area.y1 = ge->t->extent.y1;
		ge->state.total_area.x2 = ge->t->extent.x2;
		ge->state.total_area.y2 = ge->t->extent.y2;
		cvt->pict_found = TRUE;
	}
}

/****************************************************************************
 *
 * module:	mlInNotes
 */
void
mlInNotes(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	ge_page_type *page;
	int done, type;

	mlInCheckCMaps(ge, cvt);
	mlInCheckFaces(ge, cvt);
	mlInCheckCBacks(ge, cvt);
	mlInCheckFills(ge, cvt);
	mlInCheckStyles(ge, cvt);
	mlInCheckSession(ge, cvt);

	done = FALSE;
	page = &ge->page[ge->page_ix];
	cvtUpdateStatus(ge, cvt);
	while ( !done && !cvt->done )
	{ 
		type = mlInTok(ge, cvt, TRUE);
		if ( type == TOK_OTY )
		{
			if ( cvt->tok.id )
				page->notes_t = mlInPart(ge, cvt, cvt->tok.id);
			else
				page->notes_t = (thing *)0;
		}
		else if ( type == TOK_TOK )
		{
			/* if it's an END, assume following NOTES and toss it */
			if ( cvt->tok.id == TEND )
			{
				mlInTok(ge, cvt, FALSE);
				done = TRUE;
			}
		}
	}
}

/****************************************************************************
 *
 * module:	m l I n U D O
 */
void
mlInUDO(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int i, ix, n_chars, sz;
	tx_char *name;
	thing *t;

	if ( cvt->info.version < 320 )
		mlInCheckCMaps(ge, cvt);
	else
	{
		mlInCheckCMaps(ge, cvt);
		mlInCheckFaces(ge, cvt);
		mlInCheckCBacks(ge, cvt);
		mlInCheckFills(ge, cvt);
		mlInCheckStyles(ge, cvt);
		mlInCheckSession(ge, cvt);
	}

	/* get a UDO name */
	if ( mlInTok(ge, cvt,FALSE) != TOK_STR )
	{
		mlInError(ge, cvt, ML_NAMEXP);
		name = (tx_char *)0;
		n_chars = 0;
	}
	else
	{
		/* do we need to null terminate? */
		name = cvt->tok.val[0].u.cval.str;
		n_chars = cvt->tok.val[0].u.cval.n_chars;
		cvt->tok.val[0].u.cval.str = (tx_char *)0;
		cvt->tok.val[0].u.cval.n_chars = 0;
	}

	/* read the GROUP token */
	if ( mlInTok(ge, cvt,FALSE) == TOK_TOK )
		gfree(ge, name);
	else
	{
		/* create the object */
		t = cvt->tok.id ? mlInPart(ge, cvt, cvt->tok.id) : (thing *)0;

		/* expand the UDO list */
		ix = ge->n_udos;
		sz = (ge->n_udos+1)*sizeof(udo_type);
		ge->udo = (udo_type *)regloc(ge, ge->udo, sz);
		ge->udo[ix].n_chars = n_chars;
		ge->udo[ix].name = name;
		ge->udo[ix].t = t;
		geVerifyPNameCollision(ge, &ge->udo[ix]);
		ge->n_udos++;
	}

	cvt->udos_found = TRUE;
}

/****************************************************************************
 *
 * module:	ReadDEF_PAGE
 */
void
ReadDEF_PAGE(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	mlInTok(ge, cvt, FALSE);
	switch(cvt->tok.val[0].u.ival)
	{
		default:
		case 1:
			grInitBasePrint(ge, &ge->print, FALSE, FALSE, FALSE);
		break;

		case 2:
			grInitBasePrint(ge, &ge->print, TRUE, FALSE, FALSE);
		break;
	}
	ge->x_zero_pt = ge->print.page_x + ge->print.margin.x1;
	ge->y_zero_pt = ge->print.page_y + ge->print.margin.y1;
	ge->state.win_orig.x1 = ge->x_zero_pt;
	ge->state.win_orig.y1 = ge->y_zero_pt;
}

/****************************************************************************
 *
 * module:	mlInSESSION
 */
void
mlInSESSION(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	cvt_part_creation_type obj;
	thing t;

	ReadSESSION(ge, cvt, &t, &obj);
}

/****************************************************************************
 *
 * module:	m l I n V e r s i o n - obsolete
 */
mlInVersion(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int alisVersion = FALSE;
	int val, ch, ii, tok;
	double fval;

	if ( mlInTok(ge, cvt, FALSE) != TOK_NUM )
		mlInError(ge, cvt, ML_VEREXP);
	else
	{
		fval = cvtTokToFlt(ge, cvt);
		cvt->majorVer = fval;
		cvt->minorVer = (fval - (double)cvt->majorVer) * 10.0;
	}

	/* get the product string */
	mlInTok(ge, cvt, FALSE);
	if ( !strncmp(cvt->buf, NXLT("ALIS"), 4) )
		alisVersion = TRUE;
	else if ( strncmp(cvt->buf, NXLT("ASTER"), 5) )
		mlInError(ge, cvt, ML_APPEXP);

	/* get the application version string */
	switch ( cvt->appID )
	{
		/* get the Graphics Editor application string */
		default:
		case G_APP:
			mlInTok(ge, cvt, FALSE); /* assume TOK_TOK */
			if ( cvt->tok.id != TGRAPHICS )
				mlInError(ge, cvt, ML_APPEXP);
			else
			{
				mlInTok(ge, cvt, FALSE);
				if (cvt->tok.id != TEDITOR)
					mlInError(ge, cvt, ML_APPEXP);
			}
		break;

		/* BG string - BUSINESS GRAPHICS (ALIS) or GRAPHICS (ASTERX/ASTERIX) */
		case B_APP:
			/* get the string */
			mlInTok(ge, cvt, FALSE); /* BUSINESS or GRAPHICS */
			if ( alisVersion )
				mlInTok(ge, cvt, FALSE); /* GRAPHICS */

			/* consume all trailing characters until a number is found */
			while ( !isdigit(ch = (*cvt->getcFunc)(ge, cvt)) );
			(*cvt->ungetcFunc)(ge, cvt, ch);
		break;

		case F_APP:
		break;
	}

	/* session info was stored on version line in pre-rev3 software */
	if ( cvt->majorVer == GVER1_MAJOR && cvt->appID != B_APP )
	{
		/* read the session info from the version line */
		mlInSESSION(ge, cvt);

		/* no colormap in pre-rev3 software, init colormap */
		mlInCheckCMaps(ge, cvt);
	}

	if ( cvt->majorVer < GVER_MAJOR )
	{
		/* set the rev 2 state */
		ge->print.page_x = 59750;
		ge->print.page_y = 58000;
		ge->print.page_wid = 33000;
		ge->print.page_hyt = 35000;
		ge->x_zero_pt = ge->print.page_x + ge->print.margin.x1;
		ge->y_zero_pt = ge->print.page_y + ge->print.margin.y1;
		ge->state.win_orig.x1 = ge->x_zero_pt;
		ge->state.win_orig.y1 = ge->y_zero_pt;
		cvt->curr_atts.face = CvtAddFont(ge,cvt,NXLT("Times"));
		cvt->faces_found = TRUE;
	}
	return(AOK);
}

/****************************************************************************
 *
 * module:	c h e c k I m g
 */
checkImg(ge, cvt,t,obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{
	/* check for missing mandatory information */
	/* Don't check for depth. If not present assume 1 */
	if ((cvt->width == INITIAL)       ||
		(cvt->height == INITIAL))
				return ML_MISIMG;

	checkOlin(ge, cvt,t,obj);
	return (0);
}

/****************************************************************************
 *
 * module:	m l I n I m g D a t a
 */
void
mlInImgData(ge,cvt,t,obj)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
cvt_part_creation_type *obj;
{	
	int err, num_colors, size;
	unsigned char *buff8Bit;
	unsigned char *r,*g,*b;

	/* before collecting data, make sure we got all necessary info already */
	err = checkImg(ge, cvt, t, obj);
	if ( err )
		mlInError(ge, cvt, err);
	else
	{
		/* construct image thing */
		t->data.img->info.hyt = cvt->height;
		t->data.img->info.wid = cvt->width;
		t->data.img->info.dep = cvt->depth;
		if ( cvt->depth == 1 )
			t->data.img->info.data_byte_wid = BYTE_WIDTH(cvt->width);
		else if (cvt->depth == 8)
			t->data.img->info.data_byte_wid = (cvt->width + 1) & (~1);
		else /* 24 bit. Currently only converters. */
			t->data.img->info.data_byte_wid = cvt->width;

		size = cvt->height*t->data.img->info.data_byte_wid;
		t->data.img->info.data = (unsigned char *)gloc(ge,size);
		memset(t->data.img->info.data, 0x0, size);

		mlInGetData(ge, cvt,t,t->data.img->info.data);
	}
}


/****************************************************************************
 *
 * module:	m l I n I m g M a s k
 */
void
mlInImgMask(ge, t, cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	mlCreateImgMask(ge, t, cvt, 0x0);
	mlInGetData(ge, cvt,t,t->data.img->info.mask);
}


/****************************************************************************
 *
 * module:	m l I n G e t D a t a
 */
mlInGetData(ge, cvt,t,destP)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
unsigned char *destP;
{
	int hyt, (*func)(), datum, byte_wid, err;
	unsigned char *sdP, *tmpP, *imgP, *bufP;

	hyt = t->data.img->info.hyt;
	imgP = destP;
	func = (cvt->encoding == BIT6) ? mailInBit6 : mailInHex;
	byte_wid = t->data.img->info.data_byte_wid;

	if (cvt->compression == COMP_RAW)
	{
		while (0 <= --hyt)
		{
			while ((datum = (*func)(ge, cvt)) != (-1))
				*imgP++ = datum & 0xFF;
		}
	}
	else if (cvt->compression == COMP_RUN)
	{
		sdP = (unsigned char*)gloc(ge, 2*byte_wid);
		memset(sdP, 0x0, 2*byte_wid);
		bufP = (unsigned char*)gloc(ge, 2*byte_wid); /* allow for overrun */
		memset(bufP, 0x0, 2*byte_wid);
		while (0 <= --hyt)
		{
			tmpP = sdP;
			while ((datum = (*func)(ge, cvt)) != (-1))
				*tmpP++ = datum & 0xFF;
			if (unpackRUN(ge, sdP,bufP,byte_wid) != byte_wid)
				return(!AOK);
			memcpy(imgP,bufP,byte_wid);
			imgP += byte_wid;
		}
		gfree(ge, sdP);
		gfree(ge, bufP);
	}
	else /* COMP_LZW */
	{
		if (ReadWriteLZW(ge, cvt,0,0,0,destP,func,TRUE))
			return(!AOK);
	}
	return(AOK);
}

/***********************************************************************
 *
 * module: u n p a c k R U N - Unpack a scanline of `dnum' bytes from a buffer.
 */
unpackRUN(ge, src, dst, dnum)
graphics_main_type *ge;
register char *src;		/* pointer to read bytes from */
register char *dst;	/* destination pointer */
int dnum;		/* number of decompressed bytes to read */
{
	register int n, c;
	int dn;

	dn = dnum;
	while (0 < dnum)
	{
		n = (*src++) & 0xFF;

		if (n & 0x80)
		{
			n = 257 - n;
			dnum -= n; /* check that dnum is nonnegative */
			c = (*src++) & 0xFF; /* buffer overflow check */
			while (n--)
				*dst++ = c;
		}
		else
		{
			n = n + 1;
			dnum -= n; /* check that dnum is nonnegative */
			memcpy(dst,src,n); /* buffer overflow check */
			src += n;
			dst += n;
		}
	}
	return dn;
}

void
mlInPageHdrFtrInfo(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	memset((char *)&cvt->hfstr, 0, sizeof(ge_hdrftr_str_type));
	cvt->curr_atts.size = cvt->hfstr.pointsize = 120000;
	cvt->hfstr.color_ix = 1;
	if ( !cvt->templ_print )
		mlInCheckPrint(ge, cvt);
	mlInHdrFtr(ge, cvt, &cvt->hf_main);
}

/****************************************************************************
 *
 * module:	mlCheckColorRefs
 */
mlCheckColorRefs(ge,cvt,t)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
{
	int ii, jj, cmap_ix;
	cmap_type *cmap;
	char filler[16];

	if ( !t )
		return;

	if ( t->type == IMAGE )
	{
		if ( t->data.img->info.dep != 8 )
			t->data.img->cmapID = 0;
		else
		{
			/* reference to a cmap that doesn't exist. */
			cmap_ix = t->data.img->cmapID;
			if ( cmap_ix >= ge->n_cmaps )
			{
				cmap_ix = t->data.img->cmapID = ge->n_cmaps;
				MakeNewCmap(ge,cvt);
				mlLoadAxImDefCmap(ge, cvt, ge->n_cmaps-1);
			}
			else if ( ge->cmap[cmap_ix].n_colors == 0 )
				mlLoadAxImDefCmap(ge, cvt, cmap_ix);

			/* make sure the colormap has 256 entries */
			cmap = &ge->cmap[cmap_ix];
			if ( cmap->n_colors < 256 && cmap_ix != 0 )
			{
				strcpy(filler,XLT("User"));
				for (jj=cmap->n_colors; jj<256; jj++)
				{
					sprintf(filler+4,"%d",jj);
					mlLoadCmapEntry(ge, cmap_ix ,filler, 0,0,0,0,0,0);
				}
			}
		}
	}
}
void
mlInCheckPrint(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	if ( cvt->templ_print == FALSE )
	{
		grInitPresentationPrint(ge, FALSE, FALSE, FALSE);
		cvt->templ_print = TRUE;
	}
}
void
mlInCheckCMaps(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	if ( cvt->cmaps_found == FALSE || !ge->cmap )
	{
		mlAxColormap(ge, cvt);
		cvt->cmaps_found = TRUE;
	}
}
void
mlInCheckFaces(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	if ( cvt->faces_found == FALSE || ge->n_faces == 0 )
	{
		ge->face = (char**)gloc(ge, sizeof(char*));
		ge->face[0] = (char*)gloc(ge, strlen(fnt[0])+1);
		strcpy(ge->face[0],fnt[0]);
		ge->n_faces = 1;
		cvt->faces_found = TRUE;
	}
}
void
mlInCheckCBacks(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	if ( cvt->cbacks_found == FALSE )
	{
		ge->n_callbacks = 0;
		ge->callback = (ge_callback_type *)0;
		cvt->cbacks_found = TRUE;
	}
}
void
mlInCheckFills(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	if ( cvt->fills_found == FALSE )
	{
		ge->n_fills = 0;
		ge->fill = (udfill_type *)0;
		cvt->fills_found = TRUE;
	}
}
void
mlInCheckStyles(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	if ( cvt->styles_found == FALSE )
	{
		ge->n_styles = 0;
		ge->style = (thing_attributes *)0;
		cvt->styles_found = TRUE;
	}
}
void
mlInCheckSession(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	if ( cvt->sess_found == FALSE )
	{
		ge->state.drawn_area.x1 = 0;
		ge->state.drawn_area.y1 = 0;
		ge->state.drawn_area.x2 = cvt->geWidth  = 5500;
		ge->state.drawn_area.y2 = cvt->geHeight = 5900;
		if ( cvt->info.version < 420 )
			grInitBasePrint(ge, &ge->print, FALSE, TRUE, FALSE);
		else
			grInitBasePrint(ge, &ge->print, FALSE, FALSE, FALSE);
		ge->x_zero_pt = ge->print.page_x + ge->print.margin.x1;
		ge->y_zero_pt = ge->print.page_y + ge->print.margin.y1;
		ge->state.win_orig.x1 = ge->x_zero_pt;
		ge->state.win_orig.y1 = ge->y_zero_pt;
		ge->state.default_slide_effect = 0;
		ge->state.default_slide_delay = 5;
		ge->state.default_slide_throttle = 1;
		if ( cvt->info.version < 420 )
		{
			cvt->curr_atts.shadow_x = 150;
			cvt->curr_atts.shadow_y = 150;
			ge->def_attr.shadow_x = 150;
			ge->def_attr.shadow_y = 150;
		}
		
		cvt->sess_found = TRUE;
	}
}
void
mlInCheckParts(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	if ( cvt->udos_found == FALSE )
	{
		ge->n_udos = 0;
		ge->udo = (udo_type *)0;
		cvt->udos_found = TRUE;
	}
}
void
mlInCheckTags(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	if ( cvt->tags_found == FALSE )
	{
		ge->n_tags = 0;
		ge->tag = (ge_tag_type *)0;
		cvt->tags_found = TRUE;
	}
}
void
mlInCheckTemplate(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int sz;

	/* create an empty template */
	if ( cvt->tmp_found == FALSE )
	{
		sz = sizeof(ge_template_type);
		ge->template = (ge_template_type *)gloc(ge, sz);
		memset((char *)ge->template, 0, sz);
		sz = sizeof(thing_attributes);
		memcpy((char*)&ge->template->t0_attr, (char*)&ge->def_attr, sz);
		memcpy((char*)&ge->template->t1_attr, (char*)&ge->def_attr, sz);
		memcpy((char*)&ge->template->b0_attr, (char*)&ge->def_attr, sz);
		memcpy((char*)&ge->template->b1_attr, (char*)&ge->def_attr, sz);
		memcpy((char*)&ge->template->b2_attr, (char*)&ge->def_attr, sz);
		memcpy((char*)&ge->template->b3_attr, (char*)&ge->def_attr, sz);
		memcpy((char*)&ge->template->b4_attr, (char*)&ge->def_attr, sz);
		memcpy((char*)&ge->template->b5_attr, (char*)&ge->def_attr, sz);
		memcpy((char*)&ge->template->n0_attr, (char*)&ge->def_attr, sz);
		memcpy((char*)&ge->template->n1_attr, (char*)&ge->def_attr, sz);
		memcpy((char*)&ge->template->n2_attr, (char*)&ge->def_attr, sz);
		memcpy((char*)&ge->template->n3_attr, (char*)&ge->def_attr, sz);
		memcpy((char*)&ge->template->n4_attr, (char*)&ge->def_attr, sz);
		memcpy((char*)&ge->template->n5_attr, (char*)&ge->def_attr, sz);
		ge->template->page_mode = SLIDE_ED;
		cvt->tmp_found = TRUE;
	}
	mlInCheckPrint(ge, cvt);
}

/****************************************************************************
 *
 * module:	mlInFinishUp
 */
void
mlInFinishUp(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int ii;

	mlInCheckCMaps(ge, cvt);
	mlInCheckFaces(ge, cvt);
	mlInCheckCBacks(ge, cvt);
	mlInCheckFills(ge, cvt);
	mlInCheckStyles(ge, cvt);
	mlInCheckSession(ge, cvt);
	mlInCheckParts(ge, cvt);
	mlInCheckTags(ge, cvt);
	if ( ge->n_pages <= 0 )
	{
		cvtAppendPage(ge, cvt, &cvt->curr_atts);
		mlInCheckPage(ge, cvt, &ge->page[ge->page_ix]);
	}

	/* calculate number of bars for any charts */
	for (ii=0; ii<ge->n_charts; ii++)
		CalcNumBars(ge,cvt, ii);

	/* set the object count (for creating id's) */
	ge->state.tCount = cvt->count;

	/* make sure we're pointing at the correct page */
	if ( cvt->page_ix >= ge->n_pages )
		cvt->page_ix = ge->n_pages-1;
	ge->page[cvt->page_ix].layer_ix = ge->state.layer;
	ge->page_ix = cvt->page_ix;
	if ( ge->state.show_option.all_pages )
	{
		ge->state.show_option.page1 = 0;
		ge->state.show_option.pageN = ge->n_pages-1;
	}
#if 0
	for (ii=0; ii<ge->n_pages; ii++)
		grGenerateBuildList(ge, &ge->page[ii]);
#endif
}

/****************************************************************************
 *
 * module:	mlInTossSegment
 */
void
mlInTossSegment(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int tok_id, tok_type;

	/* get the segment id and type */
	tok_type = cvt->tok.type;
	tok_id = cvt->tok.id;

	/* remove everything until the next END is found */
	while ( TRUE )
	{
		/* get the next token */
		mlInTok(ge, cvt, FALSE);

		/* if an END token, check it's arg */
		if ( cvt->tok.type == TOK_TOK && cvt->tok.id == TEND )
		{
			/* if it match segment id and type, we're done */
			mlInTok(ge, cvt, FALSE);
			if ( cvt->tok.type == tok_type && cvt->tok.id == tok_id )
				break;
		}
	}
}

/****************************************************************************
 *
 * module:	c v t I n i t I n p u t S t r e a m
 */
cvtInitInputStream(ge, cvt, inputBuffer)
graphics_main_type *ge;
cvt_global_type *cvt;
char *inputBuffer;
{
	/* if there is a file, read from it */
	if ( cvt->fileName )
	{
		/* open */
		errno = 0;
		cvt->fp = (FILE *)cvtGetFP(ge, cvt->fileName, "r");
		if ( !cvt->fp )
			return(FEErrno+errno);

		/* init reading tools */
		cvt->fgetsFunc = cfgets;
		cvt->getcFunc = cgetc;
		cvt->ungetcFunc = cungetc;
	}
	else /* read from the buffer */
	{
		/* init buffers */
		cvt->inputBuff = cvt->inBufPtr = inputBuffer;

		/* init reading tools */
		cvt->fgetsFunc = bfgets;
		cvt->getcFunc = bgetc;
		cvt->ungetcFunc = bungetc;
	}
	return(AOK);
}

/****************************************************************************
 *
 * module:	grRead - Converts a file to GE thing structure
 */
grRead(ge, cvt, inputBuffer, fpage_only)
graphics_main_type *ge;
cvt_global_type *cvt;
char *inputBuffer;
int fpage_only;
{
	int err;

	/* a quick error handler (mostly for hitting an EOF while reading) */
	if ( err = setjmp(cvt->jmpPoint) )
	{
		/* clean up */
		if ( cvt->fileName )
			cvtCloseFile(ge, cvt);
		return(err);
	}

	/* delete all existing data and start anew */
	cvtInitAttrs(ge, &cvt->curr_atts);
	ResetGEstate(ge);

	/* init the token reader */
	InitTokens(ge);
	err = cvtInitInputStream(ge, cvt, inputBuffer);
	if ( err != AOK )
		return(err);

	/* now read the GE ascii file (a segment at a time) */
	while ( cvt->done == FALSE ) switch ( mlInTok(ge, cvt, FALSE) )
	{
		case TOK_TOK:
			if ( cvt->done )
				break;

			switch ( cvt->tok.id )
			{
				/* generic file header (*START GRAPHICS...) */
				case TVERSION:
					mlInVersion(ge, cvt);
				break;

				/* self contained segments */
				case TCOLORMAP:
					mlInColormap(ge, cvt);
				break;
				case TFONT:
					mlInFaces(ge, cvt);
				break;
				case TCBACK:
					mlInCallBacks(ge, cvt);
				break;
				case THYPERLINKS:
					mlInHyperlinks(ge, cvt);
				break;

				/* fills may reference colors */
				case TFILLNAMES:
					mlInNamedFills(ge, cvt);
				break;

				case TGRAD:
					mlInNamedGrads(ge, cvt);
				break;

				/* styles may reference colors, faces or callbacks */

				/* headers/footers may reference colors and faces */
				case THDRFTR:
					memset((char *)&cvt->hfstr, 0, sizeof(ge_hdrftr_str_type));
					cvt->curr_atts.size = cvt->hfstr.pointsize = 120000;
					cvt->hfstr.color_ix = 1;
					mlInHdrFtr(ge, cvt, &(ge->print.hdrftr));
					cvt->hdrftr_found = FALSE;
				break;

				case THHDRFTR: /* handouts */
					memset((char *)&cvt->hfstr, 0, sizeof(ge_hdrftr_str_type));
					cvt->curr_atts.size = cvt->hfstr.pointsize = 120000;
					cvt->hfstr.color_ix = 1;
					if ( !cvt->templ_print )
						mlInCheckPrint(ge, cvt);
					mlInHdrFtr(ge, cvt, &(ge->handout_print->hdrftr));
					cvt->hdrftr_found = FALSE;
				break;

				case TNHDRFTR: /* notes */
					memset((char *)&cvt->hfstr, 0, sizeof(ge_hdrftr_str_type));
					cvt->curr_atts.size = cvt->hfstr.pointsize = 120000;
					cvt->hfstr.color_ix = 1;
					if ( !cvt->templ_print )
						mlInCheckPrint(ge, cvt);
					mlInHdrFtr(ge, cvt, &(ge->notes_print->hdrftr));
					cvt->hdrftr_found = FALSE;
				break;

				case TOHDRFTR: /* outline */
					memset((char *)&cvt->hfstr, 0, sizeof(ge_hdrftr_str_type));
					cvt->curr_atts.size = cvt->hfstr.pointsize = 120000;
					cvt->hfstr.color_ix = 1;
					if ( !cvt->templ_print )
						mlInCheckPrint(ge, cvt);
					mlInHdrFtr(ge, cvt, &(ge->outline_print->hdrftr));
					cvt->hdrftr_found = FALSE;
				break;

				/* state may reference styles, colors, faces or callbacks */
				case TSESSION:
					mlInSESSION(ge, cvt);
				break;

				/* masters may reference styles, colors, faces or callbacks */
				case TSLIDE_MASTER:
					mlInSlideMaster(ge, cvt);
				break;
				case TTITLE_MASTER:
					mlInTitleMaster(ge, cvt);
				break;
				case THANDOUT_MASTER:
					mlInHandoutMaster(ge, cvt);
				break;
				case TOUTLINE_MASTER:
					mlInOutlineMaster(ge, cvt);
				break;
				case TNOTES_MASTER:
					mlInNotesMaster(ge, cvt);
				break;

				/* charts may reference styles, colors, faces or callbacks */
				case TCHART:
					mlInCHART(ge, cvt);
				break;

				/* parts may reference styles, colors, faces or callbacks */
				case TPART:
					mlInUDO(ge, cvt);
				break;

				/* SLIDE_INFO, LAYERS, AUDIO, PICTURE, NOTES make up a page */
				case TPHDRFTR:
					mlInPageHdrFtrInfo(ge, cvt);
				break;
				case TSLIDE_INFO:
					mlInSlideInfo(ge, cvt);
				break;
				case TLAYERS:
					mlInLayers(ge, cvt);
				break;
				case TAUDIO1:
					mlInPageAudioInfo(ge, cvt);
				break;
				case TPICTURE:
					mlInPict(ge, cvt);
					if (cvt->majorVer < GVER_MAJOR || fpage_only)
						cvt->done = TRUE;
				break;
				case TNOTES:
					mlInNotes(ge, cvt);
				break;

				/* no longer in use as a discreet segment */
				case TDOCDIMENSIONS:
					mlInDOCDIMENSIONS(ge, cvt);
				break;

				/* scan to find the END, toss all in between */
				default:
					mlInTossSegment(ge, cvt);
					mlInError(ge, cvt, ML_ILLTOK);
				break;
			}
		break;

		case TOK_OTY:
			if ( cvt->info.version < 320 )
				mlInCheckCMaps(ge, cvt);
			else
			{
				mlInCheckCMaps(ge, cvt);
				mlInCheckFaces(ge, cvt);
				mlInCheckCBacks(ge, cvt);
				mlInCheckFills(ge, cvt);
				mlInCheckStyles(ge, cvt);
				mlInCheckSession(ge, cvt);
			}
			mlInPart(ge, cvt, cvt->tok.id);
		break;

		case TOK_STD:
		break;

		default:
			mlInTossSegment(ge, cvt);
			mlInError(ge, cvt, ML_ILLTOK);
		break;
	}

	/* we're done */
	mlInFinishUp(ge, cvt);
	if ( cvt->fileName )
		cvtCloseFile(ge, cvt);

#if !TURBO
	mlDumpTxtObjs(ge, ge->t);
#endif

	return(AOK);
}
