/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: cvto.c
   Applix ASC version: 360
   Date: Thu Apr 15 21:13:50 1999
*/
#include <aos.h>
#include <atm.h>
#include <getstr10.h>
#include <getstr23.h>
#include <getstr25.h>
#include <atmwin.h>
#include <axtext.h>
#include <atxint.h>
#include <img.h>
#include <ge.h>
#include <n_style.h>
#include <bg.h>
#include <afileimg.h>
#include <cvttoks.h>
#include <cvt.h>
#include <math.h>
#if FOR_FILTER
#include <stiorts.h>
#else
#include <fsrts.h>
#define FOR_FILTER 0
#endif
#include <cvto.p>

#define NOHAMMER 0
#define HAMMER 1

#ifdef ISA_VALUE
#undef ISA_VALUE
#define ISA_VALUE 0
#else
#define ISA_VALUE 0
#endif

#ifdef ISA_ARRAY
#undef ISA_ARRAY
#define ISA_ARRAY 1
#else
#define ISA_ARRAY 1
#endif
#define NUM_ELEMS_IN_ROW 10

#ifdef MAGNITUDE
#undef MAGNITUDE
#define LOG10E          0.4342944819032518276
#define LOG10(a)        (LOG10E * log(a))
#define POW10(a)        pow(10.0, a)
#define MAGNITUDE(a)    POW10(floor(LOG10(a)))
#else
#define LOG10E          0.4342944819032518276
#define LOG10(a)        (LOG10E * log(a))
#define POW10(a)        pow(10.0, a)
#define MAGNITUDE(a)    POW10(floor(LOG10(a)))
#endif

#define RUN 0
#define DMP 1
#define MINRUN  3
#define MAXRUN  128
#define MAXDMP  128


extern int mailOutBit6();
extern int mailOutHex();
extern int ReadWriteLZW();
static void mlOutPart();

int Rev2Colors[16][4] = {
	{0, 0, 0, 0}, /* white */
	{0, 0, 0, 255}, /* black */
	{0xFFFF,0xFFFF,0xFFFF,0xFFFF}, /*desktop */
	{0xFFFF,0xFFFF,0xFFFF,0xFFFF}, /*desktop */
	{0, 255, 207, 0}, /* red */
	{255, 0, 0, 0}, /* cyan */
	{192, 0, 144, 63}, /* green */
	{127, 255, 0, 0}, /* purple */
	{255, 159, 0, 0}, /* blue */
	{0, 0, 255, 0}, /* yellow */
	{0, 0, 0, 159}, /* drk grey */
	{80, 0, 96, 127}, /* olive */
	{0, 0, 0, 95}, /* light gray */
	{0, 160, 160, 95}, /* drk red */
	{0, 127, 255, 0}, /* orange */
	{160, 160, 0, 95} /* dark blue */
};
int Rev2Arrows[9] = {0,1,2,3,4,3,1,4,2};
int Rev2Fills[26] = {0,0,1,1,2,2,3,3,4,4,5,6,7,8,9,10,2,3,4,5,6,7,8,9,10,11};

/****************************************************************************
 *
 * module:	m l O u t I n d e n t
 */
static int
mlOutIndent(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	if ( cvt->linepos != 0 )
	{
		cprintf(ge, cvt, "\n");
		cvt->linepos = 0;
	}
	switch ( cvt->indent )
	{
		default:
		case 9: cprintf(ge, cvt, "         "); return(9);
		case 8: cprintf(ge, cvt, "        ");  return(8);
		case 7: cprintf(ge, cvt, "       ");   return(7);
		case 6: cprintf(ge, cvt, "      ");    return(6);
		case 5: cprintf(ge, cvt, "     ");     return(5);
		case 4: cprintf(ge, cvt, "    ");      return(4);
		case 3: cprintf(ge, cvt, "   ");       return(3);
		case 2: cprintf(ge, cvt, "  ");        return(2);
		case 1: cprintf(ge, cvt, " ");         return(1);
		case 0:                                return(0);
	}
}

/****************************************************************************
 *
 * module:	m l O u t N e w l i n e
 */
static void
mlOutNewline(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	/* add a newline if we need one */
	if ( MAX_LINE < cvt->linepos )
	{
		cprintf(ge, cvt, "\n");
		cvt->linepos = 0;
	}
}

/****************************************************************************
 *
 * module:	CvtOutSpecialChars - handle the specials
 */
static int
CvtOutSpecialChars(ge, inBuffer,outBuffer,len,mode)
graphics_main_type *ge;
tx_char *inBuffer, *outBuffer;
int len, mode; 
{
	int inlen, count=0;
	tx_char *inP, *outP;

	if ((mode == HAMMER) && (outBuffer == NULL))
		return(0);

	inP = inBuffer;
	outP = outBuffer;
	inlen = len;
	while(len--)
	{
		if ((*inP == '\\') || (*inP == '"'))
		{
			count++;
			if (mode == HAMMER)
			{
				*outP++ = '\\';
				*outP++ = *inP;
			}
		}
		else if (*inP == '\n')
		{
			count++;
			if (mode == HAMMER)
			{
				*outP++ = '\\';
				*outP++ = 'n';
			}
		}
		else if (*inP == '\t')
		{
			count++;
			if (mode == HAMMER)
			{
				*outP++ = '\\';
				*outP++ = 't';
			}
		}
		else if (mode == HAMMER)
			*outP++ = *inP;

		inP++;
	}
	return(inlen+count);
}

/****************************************************************************
 *
 * module:	m l O u t S t r i n g
 */
int
mlOutString(ge, len, str, cvt, suppress_quote)
graphics_main_type *ge;
int len;
tx_char *str;
cvt_global_type *cvt;
int suppress_quote;
{
	int i,llen,size;
	mbchar *strP, *outBuf;
	tx_char *inbufP;

	/* count the characters first */
	llen = CvtOutSpecialChars(ge, str, NULL, len, NOHAMMER);

	/* str must be null terminated for NormalizeOut16, so  */
	/* we MUST make a local copy since we cannot count on  */
	/* the necessary tx_char being available in str.       */
	size = ((llen+1)*sizeof(tx_char));
	inbufP = (tx_char*)gloc(ge, size);

	/* now hammer the corrected chars in */
	CvtOutSpecialChars(ge, str, inbufP, len, HAMMER);
	inbufP[llen] = 0;

	/* Pump it thru the encoder (worst case: a char turns into 4 bytes) */
	size = (llen+1)*4;  
	outBuf = (mbchar *)gloc(ge, size);  /* worst case */
	len = NormalizeOut16(inbufP, outBuf, size, &cvt->info);

	/* open quote */
	if ( suppress_quote )
		llen = 0;
	else
	{
		llen = 1;
		cputc(ge, cvt, '\"');
	}

	/* write the string to the file (or buffer) */
	strP = outBuf;
	while(len--)
	{
		/* append the character to the output */
		cputc(ge, cvt,*strP++);
		llen++;

		/* check for wrap */
		if ( (MAX_LINE-1) < llen )
		{
			/* don't allow escapes to seperate */
			if ( strP[-1] == '\\' )
			{
				cputc(ge, cvt, *strP++);
				len--;
			}

			/* advance to the next line */
			cputc(ge, cvt,'\n');
			cputc(ge, cvt,' ');
			llen = 0;
		}
	}

	/* close quote */
	cputc(ge, cvt, '\"');
	gfree(ge, inbufP);
	gfree(ge, outBuf);
	return(llen);
}

/****************************************************************************
 *
 * module:	p a c k R U N - Pack a scanline of `snum' bytes into a buffer.
 */
static int
packRUN(ge, inP,outP,inWid,outWid,buffer)
graphics_main_type *ge;
char *inP, *outP;
int inWid,*outWid;
char *buffer; /* temp space */
{
	register char *buf;
	register int nbuf, rstart;
	int i, mode;
	char c, lastc, *startP;
	
	startP = outP;
	buf = buffer + 1;
	buf[0] = lastc = c = *inP++;
	nbuf = 1;
	rstart = 0;
	mode = DMP;

	for (i=1; i<inWid; i++)
	{
		buf[nbuf++] = c = *inP++;

		switch (mode) {

		case DMP:
			if (MAXDMP < nbuf)
			{
				buffer[0] = nbuf - 2;
				memcpy(outP,buffer,nbuf);
				outP += nbuf;
				/* check for outP overflow */
				buf[0] = c;
				nbuf = 1;
				rstart = 0;
				break;
			}

			if (c != lastc)
			{
				rstart = nbuf - 1;	/* 1st of poss run */
				break;
			}

			if (rstart && nbuf - rstart < MINRUN)
				break;

			if (0 < rstart)
			{
				buffer[0] = rstart - 1;
				memcpy(outP,buffer,rstart + 1);
				outP += (rstart + 1);
				/* check for outP overflow */
			}

			mode = RUN;
			break;

		case RUN:
			if (c == lastc && nbuf - rstart <= MAXRUN)
				break;

			*outP++ = rstart - nbuf + 2;
			*outP++ = lastc;
			/* check for outP overflow */
			buf[0] = c;
			nbuf = 1;
			rstart = 0;
			mode = DMP;
			break;
		}

		lastc = c;
	}

	if (mode == DMP)
	{
		buffer[0] = nbuf - 1;
		memcpy(outP,buffer,nbuf + 1);
		outP += (nbuf + 1);
		/* check for outP overflow */
	}
	else
	{
		*outP++ = rstart - nbuf + 1;
		*outP++ = lastc;
		/* check for outP overflow */
	}

	*outWid = (int)(outP - startP);
	return inWid;
}

/****************************************************************************
 *
 * module:	W r i t e R U N - Write data in RUN LENGTH ENCODED format
 */
static int
WriteRUN(ge, cvt,llen,hyt,wid,imgP,func,mode)
graphics_main_type *ge;
cvt_global_type *cvt;
int llen,hyt, wid;
unsigned char *imgP;
int (*func)(), mode;
{
	int cmpWid;
	unsigned char *outP, *pP, *tmpBuf;

	outP = (unsigned char*)gloc(ge, 2*wid); /* worst case */
	memset(outP, 0x0, 2*wid);
	tmpBuf = (unsigned char*)gloc(ge,260); /* max */
	memset(tmpBuf, 0x0, 260);

	while (0 <= --hyt)
	{
		if (packRUN(ge, imgP,outP,wid,&cmpWid,tmpBuf) != wid)
			return(SGE_COMPRESSION);
		imgP += wid;
		pP = outP;
		while (0 <= --cmpWid)
		{
			llen += (*func)(ge, *pP++, llen, cvt);
			if (MAX_LINE < llen)
			{
				cputc(ge, cvt,'\n');
				cputc(ge, cvt,' ');
				llen = 0;
			}
		}
		llen += (*func)(ge, -1, llen, cvt);
	}

	gfree(ge, outP);
	gfree(ge, tmpBuf);
	return(AOK);
}

/****************************************************************************
 *
 * module:	W r i t e R A W - Write data in Raw format
 */
static int
WriteRAW(ge, cvt,llen,hyt,wid,imgP,func,mode)
graphics_main_type *ge;
cvt_global_type *cvt;
int llen,hyt, wid;
unsigned char *imgP;
int (*func)(), mode;
{
	int cwid;

	while (0 <= --hyt)
	{
		cwid = wid;
		while (0 <= --cwid)
		{
			llen += (*func)(ge, *imgP++ & 0xFF, llen, cvt);
			if (MAX_LINE < llen)
			{
				cputc(ge, cvt,'\n');
				cputc(ge, cvt,' ');
				llen = 0;
			}
		}
		llen += (*func)(ge, -1, llen, cvt);
	}
}

/****************************************************************************
 *
 * module:	mlGetClosest21Color
 */
static int
mlGetClosest21Color(ge, c,m,y,k)
graphics_main_type *ge;
int c,m,y,k;
{
	int ii, d, mdist, close;

	mdist = 100000;   close = -1;
	for (ii=0; ii<16; ii++)
	{
		if ((ii == 2) || (ii == 3))
			continue;
		d = ABS(c - Rev2Colors[ii][0]) +
		    ABS(m - Rev2Colors[ii][1]) +
		    ABS(y - Rev2Colors[ii][2]) +
		    ABS(k - Rev2Colors[ii][3]);
		if (d<mdist)
		{
			mdist=d;
			close=ii;
		}
	}
	if (close < 0)
		close = 1; /* black default */
	return (close);
}

/****************************************************************************
 *
 * module:	m l O u t A t t r s 2 1
 */
static void
mlOutAttrs21(ge, attr, cvt)
graphics_main_type *ge;
thing_attributes *attr;
cvt_global_type *cvt;
{
	cmap_type *cmap;
	int att, new_l_arr,new_r_arr, cur_l_arr, cur_r_arr;
	int new_l_arr_state,new_r_arr_state, cur_l_arr_state, cur_r_arr_state;

	if ( attr->face != cvt->curr_atts.face )
	{
		cprintf(ge, cvt," %s %s", tok[TFONT],"SERIF1");
		cvt->curr_atts.face = attr->face;
	}
	if ( attr->size != cvt->curr_atts.size )
	{
		cvt->curr_atts.size = attr->size;
		cprintf(ge, cvt," %s PT%d\n", tok[TSIZE], GetAxFontSize(ge, attr->size/10000));
	}

	/* use only the fill colors */
	if ( attr->fill_fg != cvt->curr_atts.fill_fg )
	{
		cmap = &ge->cmap[0];
		att = mlGetClosest21Color(ge, cmap->color[attr->fill_fg].c, cmap->color[attr->fill_fg].m,
			cmap->color[attr->fill_fg].y, cmap->color[attr->fill_fg].k);
		cvt->curr_atts.fill_fg = attr->fill_fg;
		cprintf(ge, cvt," FOREGROUND-COLOR %d\n", att);
	}

	if ( attr->fill_bg != cvt->curr_atts.fill_bg )
	{
		cmap = &ge->cmap[0];
		att = mlGetClosest21Color(ge, cmap->color[attr->fill_bg].c, cmap->color[attr->fill_bg].m,
			cmap->color[attr->fill_bg].y, cmap->color[attr->fill_bg].k);
		cvt->curr_atts.fill_bg = attr->fill_bg;
		cprintf(ge, cvt," BACKGROUND-COLOR %d\n", att);
	}

	if ( attr->line_thickness != cvt->curr_atts.line_thickness )
	{
		if (attr->line_thickness == 0)
			att = 7;
		else
		{
			att = attr->line_thickness  / 14;
			if (att > 4)
				att = 4;
		}
		cprintf(ge, cvt," THICKNESS %s\n", lth[att]);
		cvt->curr_atts.line_thickness = attr->line_thickness;
	}

	if ( cvt->curr_atts.line_style != attr->line_style )
	{
		switch (attr->line_style)
		{
			case 1:
			case 6: /* dash */
				att = 1;
				break;
			case 2:
			case 4:
				att = 2; /* dotted */
				break;
			case 3:
			case 5:
				att = 3; /* dash-dotted */
				break;
			default:
				att = 0; /* solid */
				break;
		}
		cprintf(ge, cvt," STYLE %s\n", lst[att]);
		cvt->curr_atts.line_style = attr->line_style;
	}
	if (( cvt->curr_atts.left_arrow != attr->left_arrow ) ||
	    (  cvt->curr_atts.right_arrow != attr->right_arrow ))
	{
		new_l_arr =
			((attr->left_arrow >= 1) && (attr->left_arrow <= 8)) ? attr->left_arrow : 0;
		new_r_arr =
			((attr->right_arrow >= 1) && (attr->right_arrow <= 8)) ? attr->right_arrow : 0;
		new_l_arr_state = ((new_l_arr >= 1) && (new_l_arr <= 8)) ? TRUE : FALSE;
		new_r_arr_state = ((new_r_arr >= 1) && (new_r_arr <= 8)) ? TRUE : FALSE;

		cur_l_arr = ((cvt->curr_atts.left_arrow >= 1) &&
			(cvt->curr_atts.left_arrow <= 8)) ? cvt->curr_atts.left_arrow : 0;
		cur_r_arr = ((cvt->curr_atts.left_arrow >= 1) &&
			(cvt->curr_atts.left_arrow <= 8)) ? cvt->curr_atts.left_arrow : 0;
		cur_l_arr_state = ((cur_l_arr >= 1) && (cur_l_arr <= 8)) ? TRUE : FALSE;
		cur_r_arr_state = ((cur_r_arr >= 1) && (cur_r_arr <= 8)) ? TRUE : FALSE;

		if ((cur_l_arr_state != new_l_arr_state) ||
		    (cur_r_arr_state != new_r_arr_state))
		{
			if ((new_l_arr_state == TRUE) && (new_r_arr_state == TRUE))
				cprintf(ge, cvt," ARROW BOTH-ENDS\n");
			else if ((new_l_arr_state == TRUE) && (new_r_arr_state == FALSE))
				cprintf(ge, cvt," ARROW FIRST-END\n");
			else if ((new_l_arr_state == FALSE) && (new_r_arr_state == TRUE))
				cprintf(ge, cvt," ARROW LAST-END\n");
			else
				cprintf(ge, cvt," MARKER OFF\n");
		}
		if ((cur_l_arr != new_l_arr) || (cur_r_arr != new_r_arr))
		{
			if ((new_l_arr != 0) && (new_r_arr != 0))
				att = Rev2Arrows[new_r_arr]; /* arbitrarily choose right */
			else if (new_l_arr != 0)
				att = Rev2Arrows[new_l_arr];
			else if (new_r_arr != 0)
				att = Rev2Arrows[new_r_arr];
			else
				att = 0;
			if (att)
				 cprintf(ge, cvt," ARROW-STYLE ARROW%d\n",att);
				
		}
	    cvt->curr_atts.right_arrow = attr->right_arrow;
	    cvt->curr_atts.left_arrow = attr->left_arrow;
	}

	if ((cvt->curr_atts.fill_pat != attr->fill_pat) ||
		(cvt->curr_atts.fpat_type != attr->fpat_type))
	{
		if (attr->fill_pat == 0)
			cprintf(ge, cvt," PATTERN OFF\n");
		else
			cprintf(ge, cvt," PATTERN %d\n",Rev2Fills[attr->fill_pat%26]);

		cvt->curr_atts.fill_pat = attr->fill_pat;
		cvt->curr_atts.fpat_type = attr->fpat_type;
	}

	cputc(ge, cvt,'\n');
}

/****************************************************************************
 *
 * module:
 */
static int
cvt_need_fill_attrs(ge,attr1, attr2)
graphics_main_type *ge;
thing_attributes *attr1;
thing_attributes *attr2;
{
	if ( attr1->fill_fg != attr2->fill_fg )
		return(TRUE);
	if ( attr1->fill_bg != attr2->fill_bg )
		return(TRUE);
	if ( attr1->fill_pat != attr2->fill_pat )
		return(TRUE);
	if ( attr1->fpat_type != attr2->fpat_type )
		return(TRUE);
	return(FALSE);
}

/****************************************************************************
 *
 * module:
 */
static int
cvt_need_line_attrs(ge, attr1, attr2)
graphics_main_type *ge;
thing_attributes *attr1;
thing_attributes *attr2;
{
	if ( attr1->line_fg != attr2->line_fg )
		return(TRUE);
	if ( attr1->line_bg != attr2->line_bg )
		return(TRUE);
	if ( attr1->line_pat != attr2->line_pat )
		return(TRUE);
	if ( attr1->lpat_type != attr2->lpat_type )
		return(TRUE);
	return(FALSE);
}

/****************************************************************************
 *
 * module:
 */
static int
cvt_need_shad_attrs(ge, attr1, attr2)
graphics_main_type *ge;
thing_attributes *attr1;
thing_attributes *attr2;
{
	if ( attr1->shadow_color != attr2->shadow_color )
		return(TRUE);
	if ( attr1->shadow != attr2->shadow )
		return(TRUE);
	if ( attr1->shadow_x != attr2->shadow_x )
		return(TRUE);
	if ( attr1->shadow_y != attr2->shadow_y )
		return(TRUE);
	return(FALSE);
}

/****************************************************************************
 *
 * module:
 */
static int
cvt_need_para_attrs(ge, attr1, attr2)
graphics_main_type *ge;
thing_attributes *attr1;
thing_attributes *attr2;
{
	if ( attr1->l_indent1 != attr2->l_indent1 )
		return(TRUE);
	if ( attr1->l_indent2 != attr2->l_indent2 )
		return(TRUE);
	if ( attr1->r_indent != attr2->r_indent )
		return(TRUE);
	if ( attr1->bullet != attr2->bullet )
		return(TRUE);
	if ( attr1->bullet_scale != attr2->bullet_scale )
		return(TRUE);
	if ( attr1->bullet_color != attr2->bullet_color )
		return(TRUE);
	if ( attr1->bullet_font != attr2->bullet_font )
		return(TRUE);
	if ( attr1->bullet_glyph != attr2->bullet_glyph )
		return(TRUE);
	return(FALSE);
}

/****************************************************************************
 *
 * module:
 */
static int
cvt_need_spacing_attrs(ge, attr1, attr2)
graphics_main_type *ge;
thing_attributes *attr1;
thing_attributes *attr2;
{
	if ( attr1->space_above != attr2->space_above )
		return(TRUE);
	if ( attr1->space_inside != attr2->space_inside )
		return(TRUE);
	if ( attr1->space_below != attr2->space_below )
		return(TRUE);
	return(FALSE);
}

/****************************************************************************
 *
 * module:
 */
static void
cvt_attr_indent(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	if ( cvt->linepos == 0 )
		cvt->linepos += mlOutIndent(ge, cvt);
	else
		cvt->linepos += cprintf(ge, cvt, " ");
}

/****************************************************************************
 *
 * module:	m l O u t A t t r s - Write an object's attrs to file
 */
static void
mlOutCAttrs(ge, attr, cvt, needs_text_attrs)
graphics_main_type *ge;
thing_attributes *attr;
cvt_global_type *cvt;
int needs_text_attrs;
{
	/* write the fill pattern and color info */
	if ( cvt_need_fill_attrs(ge, attr, &cvt->curr_atts) )
	{
		/* write the current indentation */
		mlOutIndent(ge, cvt);

		/* write the attribute */
		cprintf(ge, cvt, "%s <%d %d %d %d 0 0 0>\n", tok[TCBACKFILL],
		              attr->fill_fg,
		              attr->fill_bg,
		              attr->fpat_type,
		              attr->fill_pat);

		/* update the current default state */
		cvt->curr_atts.fill_fg   = attr->fill_fg;
		cvt->curr_atts.fill_bg   = attr->fill_bg;
		cvt->curr_atts.fpat_type = attr->fpat_type;
		cvt->curr_atts.fill_pat  = attr->fill_pat;
		cvt->linepos = 0;
	}

	/* write the line fill pattern and color info */
	if ( cvt_need_line_attrs(ge, attr, &cvt->curr_atts) )
	{
		/* write the current indentation */
		mlOutIndent(ge, cvt);

		/* write the attribute */
		cprintf(ge, cvt, "%s <%d %d %d %d 0 0 0>\n", tok[TCLINEFILL],
		              attr->line_fg,
		              attr->line_bg,
		              attr->lpat_type,
		              attr->line_pat);

		/* update the current default state */
		cvt->curr_atts.line_fg   = attr->line_fg;
		cvt->curr_atts.line_bg   = attr->line_bg;
		cvt->curr_atts.lpat_type = attr->lpat_type;
		cvt->curr_atts.line_pat  = attr->line_pat;
		cvt->linepos = 0;
	}

	/* write the shadow info */
	if ( cvt_need_shad_attrs(ge, attr, &cvt->curr_atts) )
	{
		/* write the current indentation */
		mlOutIndent(ge, cvt);

		/* write the attribute */
		cprintf(ge, cvt, "%s <%d %d %d %d>\n", tok[TCSHADOW],
		              attr->shadow_color,
		              attr->shadow,
		              attr->shadow_x,
		              attr->shadow_y);

		/* update the current default state */
		cvt->curr_atts.shadow_color = attr->shadow_color;
		cvt->curr_atts.shadow       = attr->shadow;
		cvt->curr_atts.shadow_x     = attr->shadow_x;
		cvt->curr_atts.shadow_y     = attr->shadow_y;
		cvt->linepos = 0;
	}

	/* line attribute info */
	if ( cvt->curr_atts.line_thickness != attr->line_thickness )
	{
		cvt_attr_indent(ge, cvt);
		cvt->curr_atts.line_thickness = attr->line_thickness;
		if ( cvt->curr_atts.line_thickness < 0 )
			cvt->curr_atts.line_thickness = 0;
		cvt->linepos += cprintf(ge, cvt,"%s %d", tok[TCTHICKNESS],
			cvt->curr_atts.line_thickness);
		mlOutNewline(ge, cvt);
	}
	if ( cvt->curr_atts.line_style != attr->line_style )
	{
		cvt_attr_indent(ge, cvt);
		cvt->curr_atts.line_style = attr->line_style;
		cvt->linepos += cprintf(ge, cvt, "%s %d", tok[TCSTYLE], attr->line_style);
		mlOutNewline(ge, cvt);
	}
	if ( needs_text_attrs )
	{
		/* text attribute info */
		if ( cvt_need_para_attrs(ge, attr, &cvt->curr_atts) )
		{
			/* write the current indentation */
			mlOutIndent(ge, cvt);

			/* write the attribute */
			cprintf(ge, cvt, "%s <%d %d %d %d %d %d %d %d>\n", tok[TCPARA],
						  attr->l_indent1,
						  attr->l_indent2,
						  attr->r_indent,
						  attr->bullet,
						  attr->bullet_scale,
						  attr->bullet_color,
						  attr->bullet_font,
						  attr->bullet_glyph);

			/* update the current default state */
			cvt->curr_atts.l_indent1    = attr->l_indent1;
			cvt->curr_atts.l_indent2    = attr->l_indent2;
			cvt->curr_atts.r_indent     = attr->r_indent;
			cvt->curr_atts.bullet       = attr->bullet;
			cvt->curr_atts.bullet_scale = attr->bullet_scale;
			cvt->curr_atts.bullet_color = attr->bullet_color;
			cvt->curr_atts.bullet_font  = attr->bullet_font;
			cvt->curr_atts.bullet_glyph = attr->bullet_glyph;
			cvt->linepos = 0;
		}
		if ( cvt_need_spacing_attrs(ge, attr, &cvt->curr_atts) )
		{
			/* write the current indentation */
			mlOutIndent(ge, cvt);

			/* write the attribute */
			cprintf(ge, cvt, "%s <%d %d %d>\n", tok[TCV_SPACE],
					attr->space_above,
					attr->space_inside,
					attr->space_below);

			/* update the current default state */
			cvt->curr_atts.space_above  = attr->space_above;
			cvt->curr_atts.space_inside = attr->space_inside;
			cvt->curr_atts.space_below  = attr->space_below;
			cvt->linepos = 0;
		}

		if ( attr->face != cvt->curr_atts.face )
		{
			cvt_attr_indent(ge, cvt);
			cvt->linepos += cprintf(ge, cvt, "%s %d", tok[TCFONT], attr->face);
			cvt->curr_atts.face = attr->face;
			mlOutNewline(ge, cvt);
		}
		if ( attr->size != cvt->curr_atts.size )
		{
			cvt_attr_indent(ge, cvt);
			cvt->linepos += cprintf(ge, cvt, "%s %d", tok[TCSIZE], attr->size);
			cvt->curr_atts.size = attr->size;
			mlOutNewline(ge, cvt);
		}
		if ( attr->ruling != cvt->curr_atts.ruling )
		{
			cvt_attr_indent(ge, cvt);
			cvt->linepos += cprintf(ge, cvt, "%s %s",tok[TCUNDERLINE],und[attr->ruling]);
			cvt->curr_atts.ruling = attr->ruling;
			mlOutNewline(ge, cvt);
		}
		if ( attr->strike != cvt->curr_atts.strike )
		{
			cvt_attr_indent(ge, cvt);
			cvt->linepos += cprintf(ge, cvt, "%s %s", tok[TCTHRU], tst[attr->strike]);
			cvt->curr_atts.strike = attr->strike;
			mlOutNewline(ge, cvt);
		}
		if ( attr->weight != cvt->curr_atts.weight )
		{
			cvt_attr_indent(ge, cvt);
			cvt->linepos += cprintf(ge, cvt, "%s %s", tok[TCBOLD], tst[attr->weight]);
			cvt->curr_atts.weight = attr->weight;
			mlOutNewline(ge, cvt);
		}
		if ( attr->slant != cvt->curr_atts.slant )
		{
			cvt_attr_indent(ge, cvt);
			cvt->linepos += cprintf(ge, cvt, "%s %s", tok[TCITALICS], tst[attr->slant]);
			cvt->curr_atts.slant = attr->slant;
			mlOutNewline(ge, cvt);
		}
		if ( attr->x_off != cvt->curr_atts.x_off )
		{
			cvt_attr_indent(ge, cvt);
			cvt->linepos += cprintf(ge, cvt,"%s <%d 0>", tok[TCXYOFFS],
									attr->x_off);
			cvt->curr_atts.x_off = attr->x_off;
			mlOutNewline(ge, cvt);
		}
		if ( attr->y_sub_sup != cvt->curr_atts.y_sub_sup )
		{
			cvt_attr_indent(ge, cvt);
			cvt->linepos += cprintf(ge, cvt,"%s <0 %d>",tok[TCSUBSUP],
									attr->y_sub_sup);
			cvt->curr_atts.y_sub_sup = attr->y_sub_sup;
			mlOutNewline(ge, cvt);
		}
		if ( attr->h_just != cvt->curr_atts.h_just )
		{
			cvt_attr_indent(ge, cvt);
			cvt->linepos += cprintf(ge, cvt,"%s %s", tok[TCHOR_JUST], jst[attr->h_just]);
			cvt->curr_atts.h_just = attr->h_just;
			mlOutNewline(ge, cvt);
		}
		if ( attr->v_just != cvt->curr_atts.v_just )
		{
			cvt_attr_indent(ge, cvt);
			cvt->linepos += cprintf(ge, cvt,"%s %s", tok[TCVER_JUST],vjst[attr->v_just]);
			cvt->curr_atts.v_just = attr->v_just;
			mlOutNewline(ge, cvt);
		}
		if ( attr->field_space != cvt->curr_atts.field_space )
		{
			cvt_attr_indent(ge, cvt);
			cvt->linepos += cprintf(ge,cvt,"%s %d",tok[TCL_SPACE],attr->field_space);
			cvt->curr_atts.field_space = attr->field_space;
			mlOutNewline(ge, cvt);
		}
		if ( attr->l_margin != cvt->curr_atts.l_margin ||
			 attr->r_margin != cvt->curr_atts.r_margin ||
			 attr->t_margin != cvt->curr_atts.t_margin ||
			 attr->b_margin != cvt->curr_atts.b_margin )
		{
			cvt_attr_indent(ge, cvt);
			cvt->linepos += cprintf(ge, cvt,"%s <%d %d %d %d>", tok[TCMARGINS],
									attr->l_margin,
									attr->r_margin,
									attr->t_margin,
									attr->b_margin);
			cvt->curr_atts.l_margin = attr->l_margin;
			cvt->curr_atts.r_margin = attr->r_margin;
			cvt->curr_atts.t_margin = attr->t_margin;
			cvt->curr_atts.b_margin = attr->b_margin;
			mlOutNewline(ge, cvt);
		}

		/* some transform info */
		if ( attr->x_scale != cvt->curr_atts.x_scale )
		{
			cvt_attr_indent(ge, cvt);
			cvt->linepos += cprintf(ge, cvt,"%s %d", tok[TCSCALE], attr->x_scale);
			cvt->curr_atts.x_scale = attr->x_scale;
			mlOutNewline(ge, cvt);
		}
		if ( attr->shear != cvt->curr_atts.shear )
		{
			cvt_attr_indent(ge, cvt);
			cvt->linepos += cprintf(ge, cvt,"%s %d", tok[TCSHEAR], attr->shear);
			cvt->curr_atts.shear = attr->shear;
			mlOutNewline(ge, cvt);
		}
		if ( attr->angle != cvt->curr_atts.angle )
		{
			cvt_attr_indent(ge, cvt);
			cvt->linepos += cprintf(ge, cvt,"%s %d", tok[TCANGLE], attr->angle);
			cvt->curr_atts.angle = attr->angle;
			mlOutNewline(ge, cvt);
		}
		if ((attr->field_x_scale != cvt->curr_atts.field_x_scale) ||
			(attr->field_y_scale != cvt->curr_atts.field_y_scale))
		{
			cvt_attr_indent(ge, cvt);
			cvt->linepos += cprintf(ge, cvt,"%s <%d %d>",tok[TCFLD_XYSCALE],
									attr->field_x_scale,
									attr->field_y_scale);
			cvt->curr_atts.field_x_scale = attr->field_x_scale;
			cvt->curr_atts.field_y_scale = attr->field_y_scale;
			mlOutNewline(ge, cvt);
		}
		if ((attr->field_shear != cvt->curr_atts.field_shear) ||
			(attr->field_angle != cvt->curr_atts.field_angle))
		{
			cvt_attr_indent(ge, cvt);
			cvt->linepos += cprintf(ge, cvt,"%s <%d %d>",tok[TCSHR_ANG],
									attr->field_shear,
									attr->field_angle);
			cvt->curr_atts.field_shear = attr->field_shear;
			cvt->curr_atts.field_angle = attr->field_angle;
			mlOutNewline(ge, cvt);
		}
	}
	if ( cvt->linepos != 0 )
	{
		cprintf(ge, cvt, "\n");
		cvt->linepos = 0;
	}
}

/****************************************************************************
 *
 * module:	m l O u t A t t r s - Write an object's attrs to file
 */
static void
mlOutAttrs(ge, attr, cmap_ix, cvt)
graphics_main_type *ge;
thing_attributes *attr;
int cmap_ix;
cvt_global_type *cvt;
{
	if ( cvt->as_1_1 )
	{
		mlOutAttrs21(ge, attr, cvt);
		return;
	}

	/* write the style */
	if ( attr->style != cvt->curr_atts.style )
	{
		/* write the current indentation */
		mlOutIndent(ge, cvt);

		cprintf(ge, cvt,"%s %d\n", tok[TSTYS], attr->style);
		cvt->curr_atts.style = attr->style;
		cvt->linepos = 0;
	}

	/* write the colormap */
	if ( cmap_ix )
	{
		/* write the current indentation */
		mlOutIndent(ge, cvt);

		/* write the colormap id (images only) */
		cprintf(ge, cvt,"%s %d\n", tok[TCOLORMAP], cmap_ix);
		cvt->cmapID = cmap_ix;
		cvt->linepos = 0;
	}

	/* write the fill pattern and color info */
	if ( cvt_need_fill_attrs(ge, attr, &cvt->curr_atts) )
	{
		/* write the current indentation */
		mlOutIndent(ge, cvt);

		/* write the attribute */
		cprintf(ge, cvt, "%s <%d %d %d %d 0 0 0>\n", tok[TBACKFILL],
		              attr->fill_fg,
		              attr->fill_bg,
		              attr->fpat_type,
		              attr->fill_pat);

		/* update the current default state */
		cvt->curr_atts.fill_fg   = attr->fill_fg;
		cvt->curr_atts.fill_bg   = attr->fill_bg;
		cvt->curr_atts.fpat_type = attr->fpat_type;
		cvt->curr_atts.fill_pat  = attr->fill_pat;
		cvt->linepos = 0;
	}

	/* write the line fill pattern and color info */
	if ( cvt_need_line_attrs(ge, attr, &cvt->curr_atts) )
	{
		/* write the current indentation */
		mlOutIndent(ge, cvt);

		/* write the attribute */
		cprintf(ge, cvt, "%s <%d %d %d %d 0 0 0>\n", tok[TLINEFILL],
		              attr->line_fg,
		              attr->line_bg,
		              attr->lpat_type,
		              attr->line_pat);

		/* update the current default state */
		cvt->curr_atts.line_fg   = attr->line_fg;
		cvt->curr_atts.line_bg   = attr->line_bg;
		cvt->curr_atts.lpat_type = attr->lpat_type;
		cvt->curr_atts.line_pat  = attr->line_pat;
		cvt->linepos = 0;
	}

	/* write the shadow info */
	if ( cvt_need_shad_attrs(ge, attr, &cvt->curr_atts) )
	{
		/* write the current indentation */
		mlOutIndent(ge, cvt);

		/* write the attribute */
		cprintf(ge, cvt, "%s <%d %d %d %d>\n", tok[TSHADOW],
		              attr->shadow_color,
		              attr->shadow,
		              attr->shadow_x,
		              attr->shadow_y);

		/* update the current default state */
		cvt->curr_atts.shadow_color = attr->shadow_color;
		cvt->curr_atts.shadow       = attr->shadow;
		cvt->curr_atts.shadow_x     = attr->shadow_x;
		cvt->curr_atts.shadow_y     = attr->shadow_y;
		cvt->linepos = 0;
	}

	/* line attribute info */
	if ( cvt->curr_atts.line_thickness != attr->line_thickness )
	{
		cvt_attr_indent(ge, cvt);
		cvt->curr_atts.line_thickness = attr->line_thickness;
		if ( cvt->curr_atts.line_thickness < 0 )
			cvt->curr_atts.line_thickness = 0;
		cvt->linepos += cprintf(ge, cvt,"%s %d", tok[TTHICKNESS],
			cvt->curr_atts.line_thickness);
		mlOutNewline(ge, cvt);
	}
	if ( cvt->curr_atts.line_style != attr->line_style )
	{
		cvt_attr_indent(ge, cvt);
		cvt->curr_atts.line_style = attr->line_style;
		cvt->linepos += cprintf(ge, cvt, "%s %d", tok[TSTYLE], attr->line_style);
		mlOutNewline(ge, cvt);
	}
	if ( cvt->curr_atts.left_arrow != attr->left_arrow )
	{
		cvt_attr_indent(ge, cvt);
		cvt->curr_atts.left_arrow = attr->left_arrow;
		cvt->linepos += cprintf(ge, cvt, "%s %d", tok[TL_ARROW], attr->left_arrow);
		mlOutNewline(ge, cvt);
	}
	if ( cvt->curr_atts.right_arrow != attr->right_arrow )
	{
		cvt_attr_indent(ge, cvt);
		cvt->curr_atts.right_arrow = attr->right_arrow;
		cvt->linepos += cprintf(ge, cvt, "%s %d", tok[TR_ARROW], attr->right_arrow);
		mlOutNewline(ge, cvt);
	}
	if ( cvt->curr_atts.join_symbols != attr->join_symbols )
	{
		cvt_attr_indent(ge, cvt);
		cvt->curr_atts.join_symbols = attr->join_symbols;
		cvt->linepos += cprintf(ge, cvt, "%s %d", tok[TMARKER], attr->join_symbols);
		mlOutNewline(ge, cvt);
	}

	/* text attribute info */
	if ( cvt_need_para_attrs(ge, attr, &cvt->curr_atts) )
	{
		/* write the current indentation */
		mlOutIndent(ge, cvt);

		/* write the attribute */
		cprintf(ge, cvt, "%s <%d %d %d %d %d %d %d %d>\n", tok[TPARA],
		              attr->l_indent1,
		              attr->l_indent2,
		              attr->r_indent,
		              attr->bullet,
		              attr->bullet_scale,
		              attr->bullet_color,
		              attr->bullet_font,
		              attr->bullet_glyph);

		/* update the current default state */
		cvt->curr_atts.l_indent1    = attr->l_indent1;
		cvt->curr_atts.l_indent2    = attr->l_indent2;
		cvt->curr_atts.r_indent     = attr->r_indent;
		cvt->curr_atts.bullet       = attr->bullet;
		cvt->curr_atts.bullet_scale = attr->bullet_scale;
		cvt->curr_atts.bullet_color = attr->bullet_color;
		cvt->curr_atts.bullet_font  = attr->bullet_font;
		cvt->curr_atts.bullet_glyph = attr->bullet_glyph;
		cvt->linepos = 0;
	}
	if ( cvt_need_spacing_attrs(ge, attr, &cvt->curr_atts) )
	{
		/* write the current indentation */
		mlOutIndent(ge, cvt);

		/* write the attribute */
		cprintf(ge, cvt, "%s <%d %d %d>\n", tok[TV_SPACE],
		        attr->space_above,
		        attr->space_inside,
		        attr->space_below);

		/* update the current default state */
		cvt->curr_atts.space_above  = attr->space_above;
		cvt->curr_atts.space_inside = attr->space_inside;
		cvt->curr_atts.space_below  = attr->space_below;
		cvt->linepos = 0;
	}

	if ( attr->face != cvt->curr_atts.face )
	{
		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge, cvt, "%s %d", tok[TFONT], attr->face);
		cvt->curr_atts.face = attr->face;
		mlOutNewline(ge, cvt);
	}
	if ( attr->size != cvt->curr_atts.size )
	{
		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge, cvt, "%s %d", tok[TSIZE], attr->size);
		cvt->curr_atts.size = attr->size;
		mlOutNewline(ge, cvt);
	}
	if ( attr->ruling != cvt->curr_atts.ruling )
	{
		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge, cvt, "%s %s",tok[TUNDERLINE],und[attr->ruling]);
		cvt->curr_atts.ruling = attr->ruling;
		mlOutNewline(ge, cvt);
	}
	if ( attr->strike != cvt->curr_atts.strike )
	{
		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge, cvt, "%s %s", tok[TTHRU], tst[attr->strike]);
		cvt->curr_atts.strike = attr->strike;
		mlOutNewline(ge, cvt);
	}
	if ( attr->weight != cvt->curr_atts.weight )
	{
		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge, cvt, "%s %s", tok[TBOLD], tst[attr->weight]);
		cvt->curr_atts.weight = attr->weight;
		mlOutNewline(ge, cvt);
	}
	if ( attr->slant != cvt->curr_atts.slant )
	{
		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge, cvt, "%s %s", tok[TITALICS], tst[attr->slant]);
		cvt->curr_atts.slant = attr->slant;
		mlOutNewline(ge, cvt);
	}
	if ( attr->x_off != cvt->curr_atts.x_off )
	{
		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge, cvt,"%s <%d 0>", tok[TXYOFFS],
		                        attr->x_off);
		cvt->curr_atts.x_off = attr->x_off;
		mlOutNewline(ge, cvt);
	}
	if ( attr->y_sub_sup != cvt->curr_atts.y_sub_sup )
	{
		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge, cvt,"%s <0 %d>",tok[TSUBSUP],
		                        attr->y_sub_sup);
		cvt->curr_atts.y_sub_sup = attr->y_sub_sup;
		mlOutNewline(ge, cvt);
	}
	if ( attr->h_just != cvt->curr_atts.h_just )
	{
		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge, cvt,"%s %s", tok[THOR_JUST], jst[attr->h_just]);
		cvt->curr_atts.h_just = attr->h_just;
		mlOutNewline(ge, cvt);
	}
	if ( attr->v_just != cvt->curr_atts.v_just )
	{
		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge, cvt,"%s %s", tok[TVER_JUST],vjst[attr->v_just]);
		cvt->curr_atts.v_just = attr->v_just;
		mlOutNewline(ge, cvt);
	}
	if ( attr->field_space != cvt->curr_atts.field_space )
	{
		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge,cvt,"%s %d",tok[TL_SPACE],attr->field_space);
		cvt->curr_atts.field_space = attr->field_space;
		mlOutNewline(ge, cvt);
	}
	if ( attr->l_margin != cvt->curr_atts.l_margin ||
	     attr->r_margin != cvt->curr_atts.r_margin ||
	     attr->t_margin != cvt->curr_atts.t_margin ||
	     attr->b_margin != cvt->curr_atts.b_margin )
	{
		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge, cvt,"%s <%d %d %d %d>", tok[TMARGINS],
		                        attr->l_margin,
		                        attr->r_margin,
		                        attr->t_margin,
		                        attr->b_margin);
		cvt->curr_atts.l_margin = attr->l_margin;
		cvt->curr_atts.r_margin = attr->r_margin;
		cvt->curr_atts.t_margin = attr->t_margin;
		cvt->curr_atts.b_margin = attr->b_margin;
		mlOutNewline(ge, cvt);
	}

	/* some transform info */
	if ( attr->x_scale != cvt->curr_atts.x_scale )
	{
		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge, cvt,"%s %d", tok[TSCALE], attr->x_scale);
		cvt->curr_atts.x_scale = attr->x_scale;
		mlOutNewline(ge, cvt);
	}
	if ( attr->shear != cvt->curr_atts.shear )
	{
		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge, cvt,"%s %d", tok[TSHEAR], attr->shear);
		cvt->curr_atts.shear = attr->shear;
		mlOutNewline(ge, cvt);
	}
	if ( attr->angle != cvt->curr_atts.angle )
	{
		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge, cvt,"%s %d", tok[TANGLE], attr->angle);
		cvt->curr_atts.angle = attr->angle;
		mlOutNewline(ge, cvt);
	}
	if ((attr->field_x_scale != cvt->curr_atts.field_x_scale) ||
		(attr->field_y_scale != cvt->curr_atts.field_y_scale))
	{
		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge, cvt,"%s <%d %d>",tok[TFLD_XYSCALE],
				                attr->field_x_scale,
				                attr->field_y_scale);
		cvt->curr_atts.field_x_scale = attr->field_x_scale;
		cvt->curr_atts.field_y_scale = attr->field_y_scale;
		mlOutNewline(ge, cvt);
	}
	if ((attr->field_shear != cvt->curr_atts.field_shear) ||
		(attr->field_angle != cvt->curr_atts.field_angle))
	{
		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge, cvt,"%s <%d %d>",tok[TSHR_ANG],
			                    attr->field_shear,
			                    attr->field_angle);
		cvt->curr_atts.field_shear = attr->field_shear;
		cvt->curr_atts.field_angle = attr->field_angle;
		mlOutNewline(ge, cvt);
	}
	if ( cvt->linepos != 0 )
	{
		cprintf(ge, cvt, "\n");
		cvt->linepos = 0;
	}
}

/****************************************************************************
 *
 * module:	m l O u t H o o k s
 */
static void
mlOutHooks(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int i;

	for ( i=0; i<ge->n_hooks; i++ )
	{
		cprintf(ge, cvt, "%s %s %s\n",
		        ATOK_HOOK,
		        ge->hook[i].type,
		        ge->hook[i].name);
	}
}

/****************************************************************************
 *
 * module:	m l O u t L i n k s
 */
static void
mlOutLinks(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int i;

	for ( i=0; i<ge->n_links; i++ )
		cprintf(ge, cvt,"%s %s\n", ATOK_LINK, ge->link[i].path);
}

/****************************************************************************
 *
 * module:	m l O u t E m b e d
 */
static void
mlOutEmbed(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int i, n, size;
	char *data;

	for ( i=0; i<ge->n_embeds; i++ )
	{
		size = ge->embed[i].size - 1; /* don't count NULL at end */
		data = ge->embed[i].data;
		n = 0;

		while ( n < size )
		{
			cputc(ge, cvt, *data);
			data++;
			n++;
		}
		/* add a newline after the end */
		cprintf(ge, cvt,"\n");
	}
}

/****************************************************************************
 *
 * module:	m l O u t H e a d e r
 */
static int
mlOutHeader(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int i, offset, encrypted;

	if ( cvt->as_1_1 )
	{
		cprintf(ge, cvt,"VERSION 1.1 ASTERX GRAPHICS EDITOR\n\n");
		offset = 35;
	}
	else
	{
		/* write the header line */
		encrypted = cvt->info.encrypted;
		AGetDefaultFileInfo(&cvt->info);
		cvt->info.docType = AX_DOC_TYPE_ASTERX_GRAPHICS;
		cvt->info.encrypted = encrypted;

		/*hammer the minimum version to force prev rev to go through filter */
		cvt->info.min_version = 420;
		AFileHeaderFromInfo(&cvt->info, cvt->buf);
		offset = cprintf(ge, cvt, "%s\n", cvt->buf);

		/* dump any meta comments now */
		for ( i=0; i<ge->n_cmts; i++ )
		{
			if ( ge->cmt[i].var == (char *)0 )
				continue;
			cprintf(ge, cvt, "** \"%s\" %s\n", ge->cmt[i].var, ge->cmt[i].val);
		}

		/* write the hook list */
		mlOutHooks(ge, cvt);

		/* write the link list */
		mlOutLinks(ge, cvt);

		/* write the embedded object list */
		mlOutEmbed(ge, cvt);

		/* add a newline after the header */
		cprintf(ge, cvt,"\n");
	}
	return(offset);
}

/****************************************************************************
 *
 * module:	m l O u t C o l o r m a p - write the colormap to the file
 */
static void
mlOutColormap(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int ii, jj;
	color_type *colorP;
    int len, old_len, same;
    tx_char *tbuff;
	graphics_main_type localGE;

	if (ge->n_cmaps == 0) /* gotta have 1 */
	{
		colormapInit(ge);
		return; /* since the next condition is satisfied */
	}

	if (cvt->as_1_1)
		return;

	/* If there's only 1 colormap and it's the default */
	/* colormap , don't bother writing it. */ 
	if (ge->n_cmaps == 1) 
	{
		/* get a default colormap for comparison */
		localGE.pid = 0;
		localGE.n_cmaps = 0;
		localGE.cmap = NULL;
		colormapInit(&localGE);

		/* do the comparison */
		same = foundCmap(ge, &ge->cmap[0],localGE.cmap);

		/* free the comparison map */
		freeCmap(&localGE,localGE.cmap);

		if (same == TRUE)
			return;
	}

	for (ii=0; ii<ge->n_cmaps; ii++ )
	{
		/* note the start of the colormap */
		cprintf(ge, cvt,"%s\n", tok[TCOLORMAP]);

		/* ordinal list of color names, definitions, and ink types */
		colorP = ge->cmap[ii].color;
		tbuff = (tx_char *)0;
		old_len = len = 0;
		for ( jj=0; jj<ge->cmap[ii].n_colors; jj++ )
		{
			/* first the name of the color */
			old_len = len;
			len = strlen(colorP[jj].name)*2 + 1;
			if ( len > old_len )
				tbuff = (tx_char *)regloc(ge, tbuff, len*sizeof(tx_char));
			len = MultiToWideStr(ge, colorP[jj].name, tbuff);
			cprintf(ge, cvt, "<");
            mlOutString(ge,  len, tbuff, cvt, FALSE );

			/* then the ink type */
			cprintf(ge, cvt," %d ", colorP[jj].ink_type);

			/* finally the cmyk definition */
			cprintf(ge, cvt,"%d %d %d %d %d>\n", colorP[jj].c,
			                                colorP[jj].m,
			                                colorP[jj].y,
			                                colorP[jj].k,
			                                colorP[jj].see_thru);
		}
		if ( tbuff )
			gfree(ge, tbuff);

		/* note the end of the colormap */
		cprintf(ge, cvt,"%s %s\n\n", tok[TEND], tok[TCOLORMAP]);
	}
}

/****************************************************************************
 *
 * module:	m l O u t F a c e s
 */
static void
mlOutFaces(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	tx_char tbuff[500];
	int i, len;

	if (cvt->as_1_1)
		return; /* set up mapping ? */

	/* note the start of the face list */
	if ( ge->n_faces == 0 )
		return;

	/* don't write the default */
	if ( ge->n_faces == 1 && strcmp(NXLT("Times"), ge->face[0]) == 0 )
		return;

	/* write the face list */
	cprintf(ge, cvt,"%s\n", tok[TFONT]);
	for ( i=0; i<ge->n_faces; i++ )
	{
		len = MultiToWideStr(ge, ge->face[i], tbuff);
        mlOutString(ge,  len, tbuff, cvt, FALSE );
		cputc(ge, cvt,'\n');
	}
	cprintf(ge, cvt,"%s %s\n\n", tok[TEND], tok[TFONT]);
}

/***************************************************************************
 *
 * module:	m l O u t C a l l b a c k s
 */
static void
mlOutCallbacks(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt; 
{
	int ii,jj;
	ge_callback_type *cback;
	char *arg;
    int len;
    tx_char tbuff[400];

	if (ge->n_callbacks)
	{
		cprintf(ge, cvt,"%s\n", tok[TCBACK]);
		for (ii=0; ii<ge->n_callbacks; ii++)
		{
			/* write the macro */
			cback = &ge->callback[ii];
			cprintf(ge, cvt," %s <", tok[TMACRO]);
			len = MultiToWideStr(ge, cback->name, tbuff );
			mlOutString(ge,  len, tbuff, cvt, FALSE );
			cprintf(ge, cvt," %d %d %d>\n",
			        cback->on_drag,
			        cback->on_single,
			        cback->on_double);

			/* write the args */
			for ( jj=0; jj<cback->n_args; jj++ )
			{
				/* check for wrapping */
				cprintf(ge, cvt, " ");
				len = MultiToWideStr(ge, cback->args[jj], tbuff );
				mlOutString(ge, len, tbuff, cvt, FALSE);
				cprintf(ge, cvt, "\n");
			}
		}
		cprintf(ge, cvt,"%s %s\n\n", tok[TEND], tok[TCBACK]);
	}
}

/****************************************************************************
 *
 * module:	m l O u t N a m e d F i l l s - write out user defined fill defs
 */
static void
mlOutNamedFills(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int i, n, t, bit6, (*func)();
	int byte_width, llen, datum;
	udfill_type *f_pack;
	char *data, *string;

	if ( ge->n_fills <= 0 )
		return;

	if ( cvt->encoding == BIT6 )
	{
		bit6 = TRUE;
		func = mailOutBit6;
	}
	else
	{
		bit6 = FALSE;
		func = mailOutHex;
	}

	/* note the start of the fill list */
	cprintf(ge, cvt,"%s\n", tok[TFILLNAMES]);
	for ( i=0; i<ge->n_fills; i++ )
	{
		f_pack = &ge->fill[i];

		/* print the "Header" */
		cprintf(ge, cvt, " <");
		mlOutString(ge, f_pack->n_chars, f_pack->name, cvt, FALSE);
		cprintf(ge, cvt," %d %d %d %d %d>\n ", f_pack->hyt, f_pack->wid,
										f_pack->cmapID, bit6, f_pack->is_pict);

		/* output the data 8 bits at a time */
		llen = 0;
		if (f_pack->cmapID > 0 )
			byte_width = f_pack->wid+f_pack->wid%2;
		else
			byte_width = (f_pack->wid+7)/ 8 +((f_pack->wid+7)/ 8)%2;
		for ( n=0; n<f_pack->hyt; n++)
		{
			for ( t=0; t<byte_width; t++ )
			{
				datum = f_pack->data[n*byte_width + t];
				llen += (*func)(ge, datum & 0xff, llen, cvt);
				if ( MAX_LINE < llen )
				{
					cprintf(ge, cvt, "\n ");
					llen=0;
				}
			}
			llen += (*func)(ge, -1, llen, cvt);
		}
		if ( llen > 0 )
			cputc(ge, cvt,'\n');
	}

	/* note the end of the fill list */
	cprintf(ge, cvt,"%s %s\n\n", tok[TEND], tok[TFILLNAMES]);
}

/****************************************************************************
 *
 * module:	
 */
static void
mlOutNamedGrads(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	udgrad_type *grad_pack;
	int i;

	if ( ge->n_grads <= 0 )
		return;

	/* note the start of the fill list */
	cprintf(ge, cvt,"%s\n", tok[TGRAD]);
	for ( i=0; i<ge->n_grads; i++ )
	{
		grad_pack = &ge->grad[i];
		cprintf(ge, cvt, " <");
		mlOutString(ge, grad_pack->n_chars, grad_pack->name, cvt, FALSE);
		cprintf(ge, cvt," %d %d %d %d>\n",
		        grad_pack->type,
		        grad_pack->angle,
		        grad_pack->xoff,
		        grad_pack->yoff);
	}

	/* note the end of the grad list */
	cprintf(ge, cvt,"%s %s\n\n", tok[TEND], tok[TGRAD]);
}

/************************************************************************
 *
 * module:	m l O u t S t y l e s
 */
static void
mlOutStyles(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt; 
{
}

/****************************************************************************
 *
 * module:  m l N e e d H d r F t r s
 */
static int
mlNeedHdrsFtrs(ge,cvt, hf)
graphics_main_type *ge;
cvt_global_type *cvt;
ge_hdrftr_main_type *hf;
{
	int needed;

	if ( hf->base.hdr_n_lines > 0 || hf->base.ftr_n_lines > 0 )
		needed = TRUE;
	else
		needed = FALSE;
	if ( needed == FALSE  && hf->use_even_odd )
	{
		if ( hf->even.hdr_n_lines > 0 || hf->even.ftr_n_lines > 0 )
			needed = TRUE;
	}
	if ( needed == FALSE && hf->use_first )
	{
		if ( hf->first.hdr_n_lines > 0 || hf->first.ftr_n_lines > 0 )
			needed = TRUE;
	}
	if ( needed == FALSE && hf->use_final )
	{
		if ( hf->final.hdr_n_lines > 0 || hf->final.ftr_n_lines > 0 )
			needed = TRUE;
	}
	return(needed);
}

/****************************************************************************
 *
 * module:	mlOutHFstring
 */
static void
mlOutHFstring(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	tx_char *str;
	int len, i;

	if ( !cvt->ico->string || !cvt->ico->string[0] )
		cprintf(ge, cvt,"   %s \"\"\n", pnm2[TEXT]);
	else
	{   
		len = strlen(cvt->ico->string);
		str = (tx_char *)gloc(ge, (len+1) * sizeof(tx_char));
		for ( i=0; i<len; i++ )
			str[i] = cvt->ico->string[i];
		str[i] = (tx_char)0;

		cprintf(ge, cvt,"   %s \"", pnm2[TEXT]);
		mlOutString(ge, len, str, cvt, TRUE);
		cprintf(ge, cvt,"\n");

		gfree(ge, str);
	}
	if (cvt->hfstr.face_ix != cvt->ico->face_ix)
		cprintf(ge, cvt,"   %s %d\n", tok[TFONT],
			cvt->hfstr.face_ix = cvt->ico->face_ix);

	if (cvt->hfstr.pointsize != cvt->ico->pointsize)
		cprintf(ge, cvt,"   %s %d\n", tok[TSIZE],
			cvt->hfstr.pointsize = cvt->ico->pointsize);

	if (cvt->hfstr.color_ix != cvt->ico->color_ix)
		cprintf(ge, cvt,"   %s %d\n", tok[TFORE_COLOR],
			cvt->hfstr.color_ix = cvt->ico->color_ix);

	if (cvt->hfstr.bold != cvt->ico->bold)
		cprintf(ge, cvt,"   %s %s\n", tok[TBOLD],
			tst[cvt->hfstr.bold = cvt->ico->bold]);

	if (cvt->hfstr.italic != cvt->ico->italic)
		cprintf(ge, cvt,"   %s %s\n", tok[TITALICS],
			tst[cvt->hfstr.italic = cvt->ico->italic]);

	if (cvt->hfstr.strike_thru != cvt->ico->strike_thru)
		cprintf(ge, cvt,"   %s %s\n", tok[TTHRU],
			tst[cvt->hfstr.strike_thru = cvt->ico->strike_thru]);

	if (cvt->hfstr.underline != cvt->ico->underline)
		cprintf(ge, cvt,"   %s %d\n", tok[TUNDERLINE],
			cvt->hfstr.underline = cvt->ico->underline);

	if (cvt->hfstr.underline_words_only != cvt->ico->underline_words_only)
		cprintf(ge, cvt,"   %s %s\n", tok[TWORDSONLY],
			tst[cvt->hfstr.underline_words_only = cvt->ico->underline_words_only]);
}

/****************************************************************************
 *
 * module:	mlOutHF
 */
static void
mlOutHF(ge,cvt,lineNum)
graphics_main_type *ge;
cvt_global_type *cvt;
int lineNum;
{
	cvt->ico = &cvt->hf->inner;
	cprintf(ge, cvt,"  %s %d\n", tok[TINNER], lineNum);
	mlOutHFstring(ge,cvt);
	cprintf(ge, cvt,"  %s %s\n", tok[TEND], tok[TINNER]);

	cvt->ico = &cvt->hf->center;
	cprintf(ge, cvt,"  %s %d\n", tok[TCENTER], lineNum);
	mlOutHFstring(ge,cvt);
	cprintf(ge, cvt,"  %s %s\n", tok[TEND], tok[TCENTER]);

	cvt->ico = &cvt->hf->outer;
	cprintf(ge, cvt,"  %s %d\n", tok[TOUTER], lineNum);
	mlOutHFstring(ge,cvt);
	cprintf(ge, cvt,"  %s %s\n", tok[TEND], tok[TOUTER]);
}

/****************************************************************************
 *
 * module:  m l O u t H d r P a g e
 */
static void
mlOutHdrPage(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int ii;
	cprintf(ge, cvt,"  %s ", tok[THEADER]);
	cprintf(ge, cvt,"%s %d ", tok[TOFFSET], cvt->hdr_ftr->hdr_offset);
	cprintf(ge, cvt,"%s %d\n", tok[TNLINES], cvt->hdr_ftr->hdr_n_lines);
	for (ii=0; ii<cvt->hdr_ftr->hdr_n_lines; ii++)
	{
		cvt->hf = &cvt->hdr_ftr->hdr[ii];
		mlOutHF(ge,cvt,ii);
	}
	cprintf(ge, cvt,"  %s %s\n", tok[TEND], tok[THEADER]);
}

/****************************************************************************
 *
 * module:  m l O u t F t r P a g e
 */
static void
mlOutFtrPage(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int ii;
	cprintf(ge, cvt,"  %s ", tok[TFOOTER]);
	cprintf(ge, cvt,"%s %d ", tok[TOFFSET], cvt->hdr_ftr->ftr_offset);
	cprintf(ge, cvt,"%s %d\n", tok[TNLINES], cvt->hdr_ftr->ftr_n_lines);
	for (ii=0; ii<cvt->hdr_ftr->ftr_n_lines; ii++)
	{
		cvt->hf = &cvt->hdr_ftr->ftr[ii];
		mlOutHF(ge,cvt,ii);
	}
	cprintf(ge, cvt,"  %s %s\n", tok[TEND], tok[TFOOTER]);
}

/****************************************************************************
 *
 * module:
 */
static void
mlOutOutlineHdrsFtrs(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	ge_hdrftr_main_type *hf;
	ge_print_info_type *prnt;

	prnt = ge->outline_print;
	if ( prnt == (ge_print_info_type *)0 )
		return;
	hf = &(ge->outline_print->hdrftr);
	if ( mlNeedHdrsFtrs(ge,cvt, hf) == FALSE )
		return;

	/* set defaults */
	memset(&cvt->hfstr,0,sizeof(ge_hdrftr_str_type));
	cvt->hfstr.pointsize = 120000;
	cvt->hfstr.color_ix = 1;

	/* write out the header */
	cprintf(ge, cvt,"%s", tok[TOHDRFTR]);
	if (hf->use_first)
		cprintf(ge, cvt,"  %s %s", tok[TUSE],tok[TFIRST]);
	if (hf->use_even_odd)
		cprintf(ge, cvt,"  %s %s", tok[TUSE],tok[TEVENODD]);
	if (hf->use_final)
		cprintf(ge, cvt,"  %s %s", tok[TUSE],tok[TFINAL]);
	cprintf(ge, cvt,"\n");

	/* write out the special first page hdr/ftr info */
	if (hf->first.hdr_offset ||
		hf->first.hdr_n_lines ||
		hf->first.ftr_offset ||
		hf->first.ftr_n_lines)
	{
		cprintf(ge, cvt," %s\n", tok[TFIRST]);
		if (hf->first.hdr_offset ||
			hf->first.hdr_n_lines)
		{
			cvt->hdr_ftr = &hf->first;
			mlOutHdrPage(ge,cvt);
		}
		if (hf->first.ftr_offset ||
			hf->first.ftr_n_lines)
		{
			cvt->hdr_ftr = &hf->first;
			mlOutFtrPage(ge,cvt);
		}
		cprintf(ge, cvt," %s %s\n", tok[TEND], tok[TFIRST]);
	}

	/* write out the basic page hdr/ftr info (also used for odd pages) */
	if (hf->base.hdr_offset ||
		hf->base.hdr_n_lines ||
		hf->base.ftr_offset ||
		hf->base.ftr_n_lines)
	{
		cprintf(ge, cvt," %s\n", tok[TBASE]);
		if (hf->base.hdr_offset ||
			hf->base.hdr_n_lines)
		{
			cvt->hdr_ftr = &hf->base;
			mlOutHdrPage(ge,cvt);
		}
		if (hf->base.ftr_offset ||
			hf->base.ftr_n_lines)
		{
			cvt->hdr_ftr = &hf->base;
			mlOutFtrPage(ge,cvt);
		}
		cprintf(ge, cvt," %s %s\n", tok[TEND], tok[TBASE]);
	}

	/* write out the special even page hdr/ftr info */
	if (hf->even.hdr_offset ||
		hf->even.hdr_n_lines ||
		hf->even.ftr_offset ||
		hf->even.ftr_n_lines)
	{
		cprintf(ge, cvt," %s\n", tok[TEVEN]);
		if (hf->even.hdr_offset ||
			hf->even.hdr_n_lines)
		{
			cvt->hdr_ftr = &hf->even;
			mlOutHdrPage(ge,cvt);
		}
		if (hf->even.hdr_offset ||
			hf->even.hdr_n_lines)
		{
			cvt->hdr_ftr = &hf->even;
			mlOutFtrPage(ge,cvt);
		}
		cprintf(ge, cvt," %s %s\n", tok[TEND], tok[TEVEN]);
	}

	/* write out the special final page hdr/ftr info */
	if (hf->final.hdr_offset ||
		hf->final.hdr_n_lines ||
		hf->final.ftr_offset ||
		hf->final.ftr_n_lines)
	{
		cprintf(ge, cvt," %s\n", tok[TFINAL]);
		if (hf->final.hdr_offset ||
			hf->final.hdr_n_lines)
		{
			cvt->hdr_ftr = &hf->final;
			mlOutHdrPage(ge,cvt);
		}
		if (hf->final.ftr_offset ||
			hf->final.ftr_n_lines)
		{
			cvt->hdr_ftr = &hf->final;
			mlOutFtrPage(ge,cvt);
		}
		cprintf(ge, cvt," %s %s\n", tok[TEND], tok[TFINAL]);
	}

	cprintf(ge, cvt,"%s %s\n\n", tok[TEND], tok[TOHDRFTR]);

	/* make sure hfstr attrs are in sync with curr_atts */
	if ( cvt->hfstr.face_ix != cvt->curr_atts.face )
		cvt->curr_atts.face = cvt->hfstr.face_ix;
	if ( cvt->hfstr.pointsize != cvt->curr_atts.size )
		cvt->curr_atts.size = cvt->hfstr.pointsize;
	if ( cvt->hfstr.color_ix != cvt->curr_atts.fill_fg )
		cvt->curr_atts.fill_fg = cvt->hfstr.color_ix;
	if ( cvt->hfstr.bold != cvt->curr_atts.weight )
		cvt->curr_atts.weight = cvt->hfstr.bold;
	if ( cvt->hfstr.italic != cvt->curr_atts.slant )
		cvt->curr_atts.slant = cvt->hfstr.italic;
	if ( cvt->hfstr.strike_thru != cvt->curr_atts.strike )
		cvt->curr_atts.strike = cvt->hfstr.strike_thru;
	if ( cvt->hfstr.underline != cvt->curr_atts.ruling )
		cvt->curr_atts.ruling = cvt->hfstr.underline;
}

/****************************************************************************
 *
 * module:
 */
static void
mlOutNotesHdrsFtrs(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	ge_hdrftr_main_type *hf;
	ge_print_info_type *prnt;

	prnt = ge->notes_print;
	if ( prnt == (ge_print_info_type *)0 )
		return;
	hf = &(ge->notes_print->hdrftr);
	if ( mlNeedHdrsFtrs(ge,cvt, hf) == FALSE )
		return;

	/* set defaults */
	memset(&cvt->hfstr,0,sizeof(ge_hdrftr_str_type));
	cvt->hfstr.pointsize = 120000;
	cvt->hfstr.color_ix = 1;

	/* write out the header */
	cprintf(ge, cvt,"%s", tok[TNHDRFTR]);
	if (hf->use_first)
		cprintf(ge, cvt,"  %s %s", tok[TUSE],tok[TFIRST]);
	if (hf->use_even_odd)
		cprintf(ge, cvt,"  %s %s", tok[TUSE],tok[TEVENODD]);
	if (hf->use_final)
		cprintf(ge, cvt,"  %s %s", tok[TUSE],tok[TFINAL]);
	cprintf(ge, cvt,"\n");

	/* write out the special first page hdr/ftr info */
	if (hf->first.hdr_offset || hf->first.hdr_n_lines ||
		hf->first.ftr_offset || hf->first.ftr_n_lines)
	{
		cprintf(ge, cvt," %s\n", tok[TFIRST]);
		if (hf->first.hdr_offset || hf->first.hdr_n_lines)
		{
			cvt->hdr_ftr = &hf->first;
			mlOutHdrPage(ge,cvt);
		}
		if (hf->first.ftr_offset || hf->first.ftr_n_lines)
		{
			cvt->hdr_ftr = &hf->first;
			mlOutFtrPage(ge,cvt);
		}
		cprintf(ge, cvt," %s %s\n", tok[TEND], tok[TFIRST]);
	}

	/* write out the basic page hdr/ftr info (also used for odd pages) */
	if (hf->base.hdr_offset || hf->base.hdr_n_lines ||
		hf->base.ftr_offset || hf->base.ftr_n_lines)
	{
		cprintf(ge, cvt," %s\n", tok[TBASE]);
		if (hf->base.hdr_offset ||
			hf->base.hdr_n_lines)
		{
			cvt->hdr_ftr = &hf->base;
			mlOutHdrPage(ge,cvt);
		}
		if (hf->base.ftr_offset || hf->base.ftr_n_lines)
		{
			cvt->hdr_ftr = &hf->base;
			mlOutFtrPage(ge,cvt);
		}
		cprintf(ge, cvt," %s %s\n", tok[TEND], tok[TBASE]);
	}

	/* write out the special even page hdr/ftr info */
	if (hf->even.hdr_offset || hf->even.hdr_n_lines ||
		hf->even.ftr_offset || hf->even.ftr_n_lines)
	{
		cprintf(ge, cvt," %s\n", tok[TEVEN]);
		if (hf->even.hdr_offset || hf->even.hdr_n_lines)
		{
			cvt->hdr_ftr = &hf->even;
			mlOutHdrPage(ge,cvt);
		}
		if (hf->even.ftr_offset || hf->even.ftr_n_lines)
		{
			cvt->hdr_ftr = &hf->even;
			mlOutFtrPage(ge,cvt);
		}
		cprintf(ge, cvt," %s %s\n", tok[TEND], tok[TEVEN]);
	}

	/* write out the special final page hdr/ftr info */
	if (hf->final.hdr_offset || hf->final.hdr_n_lines ||
		hf->final.ftr_offset || hf->final.ftr_n_lines)
	{
		cprintf(ge, cvt," %s\n", tok[TFINAL]);
		if (hf->final.hdr_offset || hf->final.hdr_n_lines)
		{
			cvt->hdr_ftr = &hf->final;
			mlOutHdrPage(ge,cvt);
		}
		if (hf->final.ftr_offset || hf->final.ftr_n_lines)
		{
			cvt->hdr_ftr = &hf->final;
			mlOutFtrPage(ge,cvt);
		}
		cprintf(ge, cvt," %s %s\n", tok[TEND], tok[TFINAL]);
	}

	cprintf(ge, cvt,"%s %s\n\n", tok[TEND], tok[TNHDRFTR]);

	/* make sure hfstr attrs are in sync with curr_atts */
	if ( cvt->hfstr.face_ix != cvt->curr_atts.face )
		cvt->curr_atts.face = cvt->hfstr.face_ix;
	if ( cvt->hfstr.pointsize != cvt->curr_atts.size )
		cvt->curr_atts.size = cvt->hfstr.pointsize;
	if ( cvt->hfstr.color_ix != cvt->curr_atts.fill_fg )
		cvt->curr_atts.fill_fg = cvt->hfstr.color_ix;
	if ( cvt->hfstr.bold != cvt->curr_atts.weight )
		cvt->curr_atts.weight = cvt->hfstr.bold;
	if ( cvt->hfstr.italic != cvt->curr_atts.slant )
		cvt->curr_atts.slant = cvt->hfstr.italic;
	if ( cvt->hfstr.strike_thru != cvt->curr_atts.strike )
		cvt->curr_atts.strike = cvt->hfstr.strike_thru;
	if ( cvt->hfstr.underline != cvt->curr_atts.ruling )
		cvt->curr_atts.ruling = cvt->hfstr.underline;
}

/****************************************************************************
 *
 * module:  m l O u t H d r s F t r s
 */
static void
mlOutHandoutHdrsFtrs(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	ge_hdrftr_main_type *hf;
	ge_print_info_type *prnt;

	prnt = ge->handout_print;
	if ( prnt == (ge_print_info_type *)0 )
		return;

	hf = &(ge->handout_print->hdrftr);
	if ( mlNeedHdrsFtrs(ge,cvt, hf) == FALSE )
		return;

	/* set defaults */
	memset(&cvt->hfstr,0,sizeof(ge_hdrftr_str_type));
	cvt->hfstr.pointsize = 120000;
	cvt->hfstr.color_ix = 1;

	/* write out the header */
	cprintf(ge, cvt,"%s", tok[THHDRFTR]);
	if (hf->use_first)
		cprintf(ge, cvt,"  %s %s", tok[TUSE],tok[TFIRST]);
	if (hf->use_even_odd)
		cprintf(ge, cvt,"  %s %s", tok[TUSE],tok[TEVENODD]);
	if (hf->use_final)
		cprintf(ge, cvt,"  %s %s", tok[TUSE],tok[TFINAL]);
	cprintf(ge, cvt,"\n");

	/* write out the special first page hdr/ftr info */
	if (hf->first.hdr_offset || hf->first.hdr_n_lines ||
		hf->first.ftr_offset || hf->first.ftr_n_lines)
	{
		cprintf(ge, cvt," %s\n", tok[TFIRST]);
		if (hf->first.hdr_offset || hf->first.hdr_n_lines)
		{
			cvt->hdr_ftr = &hf->first;
			mlOutHdrPage(ge,cvt);
		}
		if (hf->first.ftr_offset || hf->first.ftr_n_lines)
		{
			cvt->hdr_ftr = &hf->first;
			mlOutFtrPage(ge,cvt);
		}
		cprintf(ge, cvt," %s %s\n", tok[TEND], tok[TFIRST]);
	}

	/* write out the basic page hdr/ftr info (also used for odd pages) */
	if (hf->base.hdr_offset || hf->base.hdr_n_lines ||
		hf->base.ftr_offset || hf->base.ftr_n_lines)
	{
		cprintf(ge, cvt," %s\n", tok[TBASE]);
		if (hf->base.hdr_offset ||
			hf->base.hdr_n_lines)
		{
			cvt->hdr_ftr = &hf->base;
			mlOutHdrPage(ge,cvt);
		}
		if (hf->base.ftr_offset || hf->base.ftr_n_lines)
		{
			cvt->hdr_ftr = &hf->base;
			mlOutFtrPage(ge,cvt);
		}
		cprintf(ge, cvt," %s %s\n", tok[TEND], tok[TBASE]);
	}

	/* write out the special even page hdr/ftr info */
	if (hf->even.hdr_offset || hf->even.hdr_n_lines ||
		hf->even.ftr_offset || hf->even.ftr_n_lines)
	{
		cprintf(ge, cvt," %s\n", tok[TEVEN]);
		if (hf->even.hdr_offset || hf->even.hdr_n_lines)
		{
			cvt->hdr_ftr = &hf->even;
			mlOutHdrPage(ge,cvt);
		}
		if (hf->even.ftr_offset || hf->even.ftr_n_lines)
		{
			cvt->hdr_ftr = &hf->even;
			mlOutFtrPage(ge,cvt);
		}
		cprintf(ge, cvt," %s %s\n", tok[TEND], tok[TEVEN]);
	}

	/* write out the special final page hdr/ftr info */
	if (hf->final.hdr_offset ||
		hf->final.hdr_n_lines ||
		hf->final.ftr_offset ||
		hf->final.ftr_n_lines)
	{
		cprintf(ge, cvt," %s\n", tok[TFINAL]);
		if (hf->final.hdr_offset ||
			hf->final.hdr_n_lines)
		{
			cvt->hdr_ftr = &hf->final;
			mlOutHdrPage(ge,cvt);
		}
		if (hf->final.ftr_offset ||
			hf->final.ftr_n_lines)
		{
			cvt->hdr_ftr = &hf->final;
			mlOutFtrPage(ge,cvt);
		}
		cprintf(ge, cvt," %s %s\n", tok[TEND], tok[TFINAL]);
	}

	cprintf(ge, cvt,"%s %s\n\n", tok[TEND], tok[THHDRFTR]);

	/* make sure hfstr attrs are in sync with curr_atts */
	if ( cvt->hfstr.face_ix != cvt->curr_atts.face )
		cvt->curr_atts.face = cvt->hfstr.face_ix;
	if ( cvt->hfstr.pointsize != cvt->curr_atts.size )
		cvt->curr_atts.size = cvt->hfstr.pointsize;
	if ( cvt->hfstr.color_ix != cvt->curr_atts.fill_fg )
		cvt->curr_atts.fill_fg = cvt->hfstr.color_ix;
	if ( cvt->hfstr.bold != cvt->curr_atts.weight )
		cvt->curr_atts.weight = cvt->hfstr.bold;
	if ( cvt->hfstr.italic != cvt->curr_atts.slant )
		cvt->curr_atts.slant = cvt->hfstr.italic;
	if ( cvt->hfstr.strike_thru != cvt->curr_atts.strike )
		cvt->curr_atts.strike = cvt->hfstr.strike_thru;
	if ( cvt->hfstr.underline != cvt->curr_atts.ruling )
		cvt->curr_atts.ruling = cvt->hfstr.underline;
}

/****************************************************************************
 *
 * module:  m l O u t H d r s F t r s
 */
static void
mlOutPrintHdrsFtrs(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	ge_hdrftr_main_type *hf;

	hf = &(ge->print.hdrftr);
	if ( mlNeedHdrsFtrs(ge,cvt, hf) == FALSE )
		return;

	/* set defaults */
	memset(&cvt->hfstr,0,sizeof(ge_hdrftr_str_type));
	cvt->hfstr.pointsize = 120000;
	cvt->hfstr.color_ix = 1;

	/* write out the header */
	cprintf(ge, cvt,"%s", tok[THDRFTR]);
	if (hf->use_first)
		cprintf(ge, cvt,"  %s %s", tok[TUSE],tok[TFIRST]);
	if (hf->use_even_odd)
		cprintf(ge, cvt,"  %s %s", tok[TUSE],tok[TEVENODD]);
	if (hf->use_final)
		cprintf(ge, cvt,"  %s %s", tok[TUSE],tok[TFINAL]);
	cprintf(ge, cvt,"\n");

	/* write out the special first page hdr/ftr info */
	if (hf->first.hdr_offset ||
		hf->first.hdr_n_lines ||
		hf->first.ftr_offset ||
		hf->first.ftr_n_lines)
	{
		cprintf(ge, cvt," %s\n", tok[TFIRST]);
		if (hf->first.hdr_offset ||
			hf->first.hdr_n_lines)
		{
			cvt->hdr_ftr = &hf->first;
			mlOutHdrPage(ge,cvt);
		}
		if (hf->first.ftr_offset ||
			hf->first.ftr_n_lines)
		{
			cvt->hdr_ftr = &hf->first;
			mlOutFtrPage(ge,cvt);
		}
		cprintf(ge, cvt," %s %s\n", tok[TEND], tok[TFIRST]);
	}

	/* write out the basic page hdr/ftr info (also used for odd pages) */
	if (hf->base.hdr_offset ||
		hf->base.hdr_n_lines ||
		hf->base.ftr_offset ||
		hf->base.ftr_n_lines)
	{
		cprintf(ge, cvt," %s\n", tok[TBASE]);
		if (hf->base.hdr_offset ||
			hf->base.hdr_n_lines)
		{
			cvt->hdr_ftr = &hf->base;
			mlOutHdrPage(ge,cvt);
		}
		if (hf->base.ftr_offset ||
			hf->base.ftr_n_lines)
		{
			cvt->hdr_ftr = &hf->base;
			mlOutFtrPage(ge,cvt);
		}
		cprintf(ge, cvt," %s %s\n", tok[TEND], tok[TBASE]);
	}

	/* write out the special even page hdr/ftr info */
	if (hf->even.hdr_offset ||
		hf->even.hdr_n_lines ||
		hf->even.ftr_offset ||
		hf->even.ftr_n_lines)
	{
		cprintf(ge, cvt," %s\n", tok[TEVEN]);
		if (hf->even.hdr_offset ||
			hf->even.hdr_n_lines)
		{
			cvt->hdr_ftr = &hf->even;
			mlOutHdrPage(ge,cvt);
		}
		if (hf->even.ftr_offset ||
			hf->even.ftr_n_lines)
		{
			cvt->hdr_ftr = &hf->even;
			mlOutFtrPage(ge,cvt);
		}
		cprintf(ge, cvt," %s %s\n", tok[TEND], tok[TEVEN]);
	}

	/* write out the special final page hdr/ftr info */
	if (hf->final.hdr_offset ||
		hf->final.hdr_n_lines ||
		hf->final.ftr_offset ||
		hf->final.ftr_n_lines)
	{
		cprintf(ge, cvt," %s\n", tok[TFINAL]);
		if (hf->final.hdr_offset ||
			hf->final.hdr_n_lines)
		{
			cvt->hdr_ftr = &hf->final;
			mlOutHdrPage(ge,cvt);
		}
		if (hf->final.ftr_offset ||
			hf->final.ftr_n_lines)
		{
			cvt->hdr_ftr = &hf->final;
			mlOutFtrPage(ge,cvt);
		}
		cprintf(ge, cvt," %s %s\n", tok[TEND], tok[TFINAL]);
	}

	cprintf(ge, cvt,"%s %s\n\n", tok[TEND], tok[THDRFTR]);

	/* make sure hfstr attrs are in sync with curr_atts */
	if ( cvt->hfstr.face_ix != cvt->curr_atts.face )
		cvt->curr_atts.face = cvt->hfstr.face_ix;
	if ( cvt->hfstr.pointsize != cvt->curr_atts.size )
		cvt->curr_atts.size = cvt->hfstr.pointsize;
	if ( cvt->hfstr.color_ix != cvt->curr_atts.fill_fg )
		cvt->curr_atts.fill_fg = cvt->hfstr.color_ix;
	if ( cvt->hfstr.bold != cvt->curr_atts.weight )
		cvt->curr_atts.weight = cvt->hfstr.bold;
	if ( cvt->hfstr.italic != cvt->curr_atts.slant )
		cvt->curr_atts.slant = cvt->hfstr.italic;
	if ( cvt->hfstr.strike_thru != cvt->curr_atts.strike )
		cvt->curr_atts.strike = cvt->hfstr.strike_thru;
	if ( cvt->hfstr.underline != cvt->curr_atts.ruling )
		cvt->curr_atts.ruling = cvt->hfstr.underline;
}

/****************************************************************************
 *
 * module:  m l O u t H d r s F t r s
 */
static void
mlOutHdrsFtrs(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	mlOutPrintHdrsFtrs(ge,cvt);
	mlOutNotesHdrsFtrs(ge,cvt);
	mlOutHandoutHdrsFtrs(ge,cvt);
	mlOutOutlineHdrsFtrs(ge,cvt);

}

/****************************************************************************
 *
 * module:  m l O u t W i n d o w S i z e
 */
static void
mlOutWindowSize(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int wid, hyt;

	/* set the window width */
	wid = ge->state.drawn_area.x2 - ge->state.drawn_area.x1;
	if ( wid <= 0 )
		wid = 7.5 * MILS_P_INCH;

	/* set the window height */
	hyt = ge->state.drawn_area.y2 - ge->state.drawn_area.y1;
	if ( hyt <= 0 )
		hyt = 5.5 * MILS_P_INCH;

	cprintf(ge, cvt," %s <%d %d>\n", tok[TWIN_SIZE], wid, hyt);
}

/****************************************************************************
 *
 * module: m l O u t S l i d e S h o w I n f o
 */
static void
mlOutSlideShowInfo(ge, cvt, show_option)
graphics_main_type *ge;
cvt_global_type *cvt;
ge_slide_show *show_option;
{
	if ( show_option->full_screen == FALSE ||
	     show_option->all_pages == FALSE ||
	     show_option->at_end != 0 )
	{
		cprintf(ge, cvt," %s <%d %d %d %d %d>\n", tok[TSLIDE_SHOW],
					show_option->full_screen, show_option->all_pages,
					show_option->page1, show_option->pageN,
					show_option->at_end);
	}
}

/****************************************************************************
 *
 * module: m l O u t L a y e r
 */
static void
mlOutLayer(ge, cvt, layer)
graphics_main_type *ge;
cvt_global_type *cvt;
int layer;
{
	/* write the layer */
	if ( layer != cvt->curr_layer )
	{
		/* write the current indentation */
		mlOutIndent(ge, cvt);

		cprintf(ge, cvt,"%s %d\n", tok[TLAYERS], layer);
		cvt->curr_layer = layer;
		cvt->linepos = 0;
	}
}

/****************************************************************************
 *
 * module: 
 */
static void
mlOutZoomInfo(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	cprintf(ge, cvt," %s <%d %d %d %d %d>\n", tok[TREZ], ge->xform.zoom_to_fit,
										 ge->xform.x_res, ge->xform.y_res,
										 ge->state.win_orig.x1 - PAGE_ZEROX,
										 ge->state.win_orig.y1 - PAGE_ZEROY);
}

/****************************************************************************
 *
 * module:  m l O u t S e s s i o n I n f o
 */
static void
mlOutSessionInfo(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	struct ge_print_info_type *prnt;
	struct session *sess;

	if (cvt->as_1_1)
		return;

	/* begin */
	cprintf(ge, cvt,"%s\n", tok[TSESSION]);

	/* window info */
	mlOutWindowSize(ge, cvt);

	/* document info */
	sess = &ge->state;
	if ( sess->stickyDots )
		cprintf(ge, cvt," %s %s\n",tok[TSTICKY], tst[1]);
	if ( sess->autoGrid )
		cprintf(ge, cvt," %s %s\n",tok[TAUTOGRID], tst[1]);
	if ( sess->rulerOn || sess->pageOn || sess->readOut ||
	     sess->gridOn  || !sess->bordersOn || sess->guidesOn )
	{
		cprintf(ge, cvt, " %s <%d %d %d %d %d %d>\n", tok[TVIEW],
		        sess->rulerOn,
		        sess->pageOn,
		        sess->readOut,
		        sess->gridOn,
				sess->bordersOn,
		        sess->guidesOn);
	}
	if ( sess->im_hex || sess->im_raw )
	{
		if ( sess->im_hex && sess->im_raw )
			cprintf(ge, cvt," %s 3\n",tok[TIMAGE_FMT]);
		else if ( sess->im_hex )
			cprintf(ge, cvt," %s 2\n",tok[TIMAGE_FMT]);
		else if ( sess->im_raw )
			cprintf(ge, cvt," %s 1\n",tok[TIMAGE_FMT]);
		else
			cprintf(ge, cvt," %s 0\n",tok[TIMAGE_FMT]);
	}

	if (sess->dpi != 1000 || sess->gridSize != 250 || sess->unit_size != 1000)
	{
		cprintf(ge, cvt," %s %d %s %d %s <%d %d ",
			tok[TGRID_DPI], sess->dpi,
			tok[TGRID_FACTOR], sess->gridSize,
			tok[TUNITS], sess->unit_size, sess->unit_precision);
		if ( sess->unit_name == (char *)0 )
			cprintf(ge, cvt,"\"\">\n");
		else
			cprintf(ge, cvt,"\"%s\">\n", sess->unit_name);
	}

	if ((sess->line_1 != 28) || /* defaults */
	    (sess->line_2 != 42) ||
	    (sess->line_3 != 56))
	{
		cprintf(ge, cvt," %s <%d %d %d>\n",tok[TL_WIDTHS],
				sess->line_1, sess->line_2, sess->line_3);
	}

	/* zero point */
	cvt->linepos = 0;
	if ( ge->x_zero_pt != 60250 || ge->y_zero_pt != 58500 )
	{
		cvt->linepos += cprintf(ge, cvt," %s <%d %d>", tok[TPAGEZEROS],
		                             ge->x_zero_pt - 59750,
		                             ge->y_zero_pt - 58500);
	}
	if ( cvt->linepos > 0 )
	{
		cprintf(ge, cvt, "\n");
		cvt->linepos = 0;
	}

	/* base printable area - position and size */
	cvt->linepos = 0;
	prnt = &ge->print;
	if ( prnt->page_wid != 0 && prnt->page_hyt != 0 )
	{
		if ( prnt->page_x != 59750 || prnt->page_y != 58000 )
		{
			cvt->linepos += cprintf(ge, cvt," %s <%d %d>", tok[TPAGEXY],
			                             prnt->page_x - 59750,
			                             prnt->page_y - 58000);
		}
		if ( cvt->linepos > 0 )
		{
			cprintf(ge, cvt, "\n");
			cvt->linepos = 0;
		}
		if ( prnt->page_wid != 8500 )
		{
			cvt->linepos += cprintf(ge, cvt," %s %d",tok[TPAGEWID],
			                        prnt->page_wid);
		}
		if ( prnt->page_hyt != 11000 )
		{
			cvt->linepos += cprintf(ge, cvt," %s %d",tok[TPAGEHYT],
			                        prnt->page_hyt);
		}
		if ( cvt->linepos > 0 )
		{
			cprintf(ge, cvt, "\n");
			cvt->linepos = 0;
		}
	}

	/* paper type */
	if ( prnt->wid != 8500 )
		cvt->linepos += cprintf(ge, cvt," %s %d", tok[TPRINTWID], prnt->wid);
	if ( prnt->hyt != 11000 )
		cvt->linepos += cprintf(ge, cvt," %s %d", tok[TPRINTHYT], prnt->hyt);
	if ( prnt->landscape )
		cvt->linepos += cprintf(ge, cvt," %s %s", tok[TPRINTLAND], tst[1]);
	if ( cvt->linepos > 0 )
		cprintf(ge, cvt, "\n");
	cvt->linepos = 0;
	if ( prnt->guide_pt.x || prnt->guide_pt.y )
		cprintf(ge, cvt," %s <%d %d>\n", tok[TPRINTGUIDES],
		        prnt->guide_pt.x,
		        prnt->guide_pt.y);
	if ( prnt->margin.x1 != 500 || prnt->margin.y1 != 500 ||
	     prnt->margin.x2 != 500 || prnt->margin.y2 != 500 )
	{
		cprintf(ge, cvt," %s <%d %d %d %d>\n", tok[TPRINTMARG],
		                                   prnt->margin.x1,
		                                   prnt->margin.y1,
		                                   prnt->margin.x2,
		                                   prnt->margin.y2);
	}
	if ( prnt->use_content_border ||
	     prnt->use_crop_marks ||
	     prnt->use_regis_marks )
	{
		cprintf(ge, cvt, " %s <%d %d %d>\n", tok[TPRINTMARKS],
		                                 prnt->use_content_border,
		                                 prnt->use_crop_marks,
		                                 prnt->use_regis_marks);
	}
	if ( prnt->bleed )
		cprintf(ge, cvt," %s %d\n", tok[TPRINTBLEED], prnt->bleed);

	/* notes print setup info */
	prnt = ge->notes_print;
	if ( prnt )
	{
		/* printable area - position, size, and zero point */
		cvt->linepos = 0;
		if ( prnt->page_wid != 0 && prnt->page_hyt != 0 )
		{
			if ( prnt->page_x != 59750 || prnt->page_y != 58000 )
			{
				cvt->linepos += cprintf(ge, cvt," %s <%d %d>", tok[TNPAGEXY],
											 prnt->page_x - 59750,
											 prnt->page_y - 58000);
			}
			if ( cvt->linepos > 0 )
			{
				cprintf(ge, cvt, "\n");
				cvt->linepos = 0;
			}
			if ( prnt->page_wid != 8500 )
			{
				cvt->linepos += cprintf(ge, cvt," %s %d",tok[TNPAGEWID],
										prnt->page_wid);
			}
			if ( prnt->page_hyt != 11000 )
			{
				cvt->linepos += cprintf(ge, cvt," %s %d",tok[TNPAGEHYT],
										prnt->page_hyt);
			}
			if ( cvt->linepos > 0 )
			{
				cprintf(ge, cvt, "\n");
				cvt->linepos = 0;
			}
		}

		/* paper type */
		if ( prnt->wid != 8500 )
			cvt->linepos += cprintf(ge, cvt," %s %d", tok[TNPRINTWID], prnt->wid);
		if ( prnt->hyt != 11000 )
			cvt->linepos += cprintf(ge, cvt," %s %d", tok[TNPRINTHYT], prnt->hyt);
		if ( prnt->landscape )
			cvt->linepos += cprintf(ge, cvt," %s %s", tok[TNPRINTLAND], tst[1]);
		if ( cvt->linepos > 0 )
			cprintf(ge, cvt, "\n");
		cvt->linepos = 0;
		if ( prnt->guide_pt.x || prnt->guide_pt.y )
			cprintf(ge, cvt," %s <%d %d>\n", tok[TNPRINTGUIDES],
			        prnt->guide_pt.x,
			        prnt->guide_pt.y);
		if ( prnt->margin.x1 != 500 || prnt->margin.y1 != 500 ||
			 prnt->margin.x2 != 500 || prnt->margin.y2 != 500 )
		{
			cprintf(ge, cvt," %s <%d %d %d %d>\n", tok[TNPRINTMARG],
											   prnt->margin.x1,
											   prnt->margin.y1,
											   prnt->margin.x2,
											   prnt->margin.y2);
		}
		if ( prnt->use_content_border ||
			 prnt->use_crop_marks ||
			 prnt->use_regis_marks )
		{
			cprintf(ge, cvt, " %s <%d %d %d>\n", tok[TNPRINTMARKS],
											 prnt->use_content_border,
											 prnt->use_crop_marks,
											 prnt->use_regis_marks);
		}
		if ( prnt->bleed )
			cprintf(ge, cvt," %s %d\n", tok[TNPRINTBLEED], prnt->bleed);

	}

	/* handout print setup info */
	prnt = ge->handout_print;
	if ( prnt )
	{
		/* printable area - position, size, and zero point */
		cvt->linepos = 0;
		if ( prnt->page_wid != 0 && prnt->page_hyt != 0 )
		{
			if ( prnt->page_x != 59750 || prnt->page_y != 58000 )
			{
				cvt->linepos += cprintf(ge, cvt," %s <%d %d>", tok[THPAGEXY],
											 prnt->page_x - 59750,
											 prnt->page_y - 58000);
			}
			if ( cvt->linepos > 0 )
			{
				cprintf(ge, cvt, "\n");
				cvt->linepos = 0;
			}
			if ( prnt->page_wid != 8500 )
			{
				cvt->linepos += cprintf(ge, cvt," %s %d",tok[THPAGEWID],
										prnt->page_wid);
			}
			if ( prnt->page_hyt != 11000 )
			{
				cvt->linepos += cprintf(ge, cvt," %s %d",tok[THPAGEHYT],
										prnt->page_hyt);
			}
			if ( cvt->linepos > 0 )
			{
				cprintf(ge, cvt, "\n");
				cvt->linepos = 0;
			}
		}

		/* paper type */
		if ( prnt->wid != 8500 )
			cvt->linepos += cprintf(ge, cvt," %s %d", tok[THPRINTWID], prnt->wid);
		if ( prnt->hyt != 11000 )
			cvt->linepos += cprintf(ge, cvt," %s %d", tok[THPRINTHYT], prnt->hyt);
		if ( prnt->landscape )
			cvt->linepos += cprintf(ge, cvt," %s %s", tok[THPRINTLAND], tst[1]);
		if ( cvt->linepos > 0 )
			cprintf(ge, cvt, "\n");
		cvt->linepos = 0;
		if ( prnt->guide_pt.x || prnt->guide_pt.y )
			cprintf(ge, cvt," %s <%d %d>\n", tok[THPRINTGUIDES],
			        prnt->guide_pt.x,
			        prnt->guide_pt.y);
		if ( prnt->margin.x1 != 500 || prnt->margin.y1 != 500 ||
			 prnt->margin.x2 != 500 || prnt->margin.y2 != 500 )
		{
			cprintf(ge, cvt," %s <%d %d %d %d>\n", tok[THPRINTMARG],
											   prnt->margin.x1,
											   prnt->margin.y1,
											   prnt->margin.x2,
											   prnt->margin.y2);
		}
		if ( prnt->use_content_border ||
			 prnt->use_crop_marks ||
			 prnt->use_regis_marks )
		{
			cprintf(ge, cvt, " %s <%d %d %d>\n", tok[THPRINTMARKS],
											 prnt->use_content_border,
											 prnt->use_crop_marks,
											 prnt->use_regis_marks);
		}
		if ( prnt->bleed )
			cprintf(ge, cvt," %s %d\n", tok[THPRINTBLEED], prnt->bleed);

	}

	/* outline print setup info */
	prnt = ge->outline_print;
	if ( prnt )
	{
		/* printable area - position, size, and zero point */
		cvt->linepos = 0;
		if ( prnt->page_wid != 0 && prnt->page_hyt != 0 )
		{
			if ( prnt->page_x != 59750 || prnt->page_y != 58000 )
			{
				cvt->linepos += cprintf(ge, cvt," %s <%d %d>", tok[TOPAGEXY],
											 prnt->page_x - 59750,
											 prnt->page_y - 58000);
			}
			if ( cvt->linepos > 0 )
			{
				cprintf(ge, cvt, "\n");
				cvt->linepos = 0;
			}
			if ( prnt->page_wid != 8500 )
			{
				cvt->linepos += cprintf(ge, cvt," %s %d",tok[TOPAGEWID],
										prnt->page_wid);
			}
			if ( prnt->page_hyt != 11000 )
			{
				cvt->linepos += cprintf(ge, cvt," %s %d",tok[TOPAGEHYT],
										prnt->page_hyt);
			}
			if ( cvt->linepos > 0 )
			{
				cprintf(ge, cvt, "\n");
				cvt->linepos = 0;
			}
		}

		/* paper type */
		if ( prnt->wid != 8500 )
			cvt->linepos += cprintf(ge, cvt," %s %d", tok[TOPRINTWID], prnt->wid);
		if ( prnt->hyt != 11000 )
			cvt->linepos += cprintf(ge, cvt," %s %d", tok[TOPRINTHYT], prnt->hyt);
		if ( prnt->landscape )
			cvt->linepos += cprintf(ge, cvt," %s %s", tok[TOPRINTLAND], tst[1]);
		if ( cvt->linepos > 0 )
			cprintf(ge, cvt, "\n");
		cvt->linepos = 0;
		if ( prnt->guide_pt.x || prnt->guide_pt.y )
			cprintf(ge, cvt," %s <%d %d>\n", tok[TOPRINTGUIDES],
			        prnt->guide_pt.x,
			        prnt->guide_pt.y);
		if ( prnt->margin.x1 != 500 || prnt->margin.y1 != 500 ||
			 prnt->margin.x2 != 500 || prnt->margin.y2 != 500 )
		{
			cprintf(ge, cvt," %s <%d %d %d %d>\n", tok[TOPRINTMARG],
											   prnt->margin.x1,
											   prnt->margin.y1,
											   prnt->margin.x2,
											   prnt->margin.y2);
		}
		if ( prnt->use_content_border ||
			 prnt->use_crop_marks ||
			 prnt->use_regis_marks )
		{
			cprintf(ge, cvt, " %s <%d %d %d>\n", tok[TOPRINTMARKS],
											 prnt->use_content_border,
											 prnt->use_crop_marks,
											 prnt->use_regis_marks);
		}
		if ( prnt->bleed )
			cprintf(ge, cvt," %s %d\n", tok[TOPRINTBLEED], prnt->bleed);

	}

	/* slide show options */
	mlOutSlideShowInfo(ge, cvt, &sess->show_option);

	/* default slide stuff */
	if ( ge->state.wrap_slide_show )
		cprintf(ge, cvt, "%s TRUE\n", tok[TSLIDE_CYCLE]);
	if ( ge->state.default_slide_effect ||
	     ge->state.default_slide_delay != 5 ||
	     ge->state.default_slide_throttle != 1 )
	{
		cprintf(ge, cvt, " %s <%d %d %d>\n", tok[TSLIDE_STYLE],
		        ge->state.default_slide_effect,
		        ge->state.default_slide_delay,
		        ge->state.default_slide_throttle);
	}

	/* what page is active */
	if ( ge->page_ix != 0 )
		cprintf(ge, cvt, " %s %d\n", tok[TCURRENT_PAGE], ge->page_ix+1);

	/* default attributes */
	cvt->linepos = 0;
	cvt->indent = 1;
	mlOutAttrs(ge, &ge->def_attr, 0, cvt);
	mlOutLayer(ge, cvt, ge->state.layer);
	cvt->indent = 0;
	if ( ge->state.run_macro )
	{
		cprintf(ge, cvt, " %s ", tok[TRUN_MACRO]);
		mlOutString(ge, ge->state.n_mchars, ge->state.run_macro, cvt, FALSE);
		cprintf(ge, cvt, "\n");
	}
	/* zoom info */
	mlOutZoomInfo(ge, cvt);
		
	/* default tab size */
	if ( ge->state.def_tabsize != 1000 )
		cprintf(ge, cvt," %s <%d>\n", tok[TDEF_TABSIZE], ge->state.def_tabsize);
	/* done */
	cprintf(ge, cvt,"%s %s\n\n", tok[TEND], tok[TSESSION]);
}

/****************************************************************************
 *
 * module:	m l O u t L a y e r s
 */
static void
mlOutLayers(ge, cvt, page)
graphics_main_type *ge;
cvt_global_type *cvt;
ge_page_type *page;
{
	layer_type *l_pack;
	tx_char tbuff[400];
	char *def_name, str2[STRPATH+1];
	int i, j, len;

	if ( cvt->as_1_1 || page->n_layers == 0 )
		return;

	/* don't need to write the default layer */
	if ( page->n_layers == 1 )
	{
		def_name = XLT("Default");
		len = strlen(def_name);
		for ( i=0; i<len; i++ )
			tbuff[i] = (tx_char)def_name[i];
		tbuff[i] = (tx_char)0;
		l_pack = &page->layer[0];
		if ( l_pack->n_chars == len &&
		     tx_strncmp(tbuff, l_pack->name, len) == 0 &&
		     l_pack->locked == FALSE &&
		     l_pack->hidden == FALSE &&
		     l_pack->prints == TRUE )
			return;
	}

	/* note the start of the layers info section */
	cprintf(ge, cvt,"%s\n", tok[TLAYERS]);

	/* output info for each layer */
	for ( i=0; i<page->n_layers; i++ )
	{
		l_pack = &page->layer[i];

		/* print the layer name */
		cprintf(ge, cvt, "<");
		mlOutString(ge, l_pack->n_chars, l_pack->name, cvt, FALSE);

		/* locked and hidden tags */
		cprintf(ge, cvt," %d %d %d %d>\n",
		        l_pack->locked,
		        l_pack->hidden,
		        l_pack->prints,
		        l_pack->backgd);
	}

	/* note the end of the layers info section */
	cprintf(ge, cvt,"%s %s\n", tok[TEND], tok[TLAYERS]);
}

static int
mlNeedBuildInfo(ge, cvt, build, enable_build)
graphics_main_type *ge;
cvt_global_type *cvt;
ge_build_info_type *build;
int enable_build;
{
	/* default setting */
	if ( enable_build == FALSE &&
		 build->use_reverse_build == FALSE  &&
		 build->build_effect == 0  &&
		 build->build_rule == 0  &&
		 build->build_size == 0  &&
		 build->build_delay == (-1)  &&
		 build->use_inactive_color == FALSE  &&
		 build->inactive_color == 0  &&
		 build->build_throttle == 1  &&
	     build->baudio.play == FALSE &&
		 build->baudio.type == INSET_NONE &&
		 build->baudio.id   == 0 &&
	     build->baudio.volume == 50 )
		return(FALSE);
	return(TRUE);
}

/****************************************************************************
 *
 * module:
 */
static void
mlOutBuildInfo(ge, cvt, build, enable_build)
graphics_main_type *ge;
cvt_global_type *cvt;
ge_build_info_type *build;
int enable_build;
{
	if ( cvt->as_1_1 )
		return;

	cprintf(ge, cvt, "%s <%d %d %d %d %d %d %d %d %d %d %d %d %d %d>\n",
				tok[TBUILD_INFO],
		        enable_build,
		        build->use_reverse_build,
				build->build_effect,
		        build->build_rule,
		        build->build_size,
		        build->build_delay,
		        build->use_inactive_color,
		        build->inactive_color,
		        build->sequence,
		        build->build_throttle,
	            build->baudio.play,
				build->baudio.type,
				build->baudio.id,
				build->baudio.volume);
}

/****************************************************************************
 *
 * module:	m l O u t S l i d e I n f o S t a r t
 */
static int
mlOutSlideInfoStart(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt, "%s\n", tok[TSLIDE_INFO]);
	cvt->indent++;
}

/****************************************************************************
 *
 * module:	m l O u t S l i d e I n f o
 */
static int
mlOutSlideInfo(ge, cvt, page)
graphics_main_type *ge;
cvt_global_type *cvt;
ge_page_type *page;
{
	int have_slide_info, i;

	/* outline text */
	have_slide_info = FALSE;
	if ( page->n_ochars > 0 && page->outline_text )
	{
		if ( have_slide_info == FALSE )
		{
			mlOutSlideInfoStart(ge, cvt);
			have_slide_info = TRUE;
		}
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt, "%s ", tok[TSLIDE_OUTLINE]);
		mlOutString(ge, page->n_ochars, page->outline_text, cvt, FALSE);
	}

	/* slide info */
	if ( page->slide_effect != 0 ||
	     page->slide_delay  != 5 ||
	     page->slide_throttle != 1 )
	{
		if ( have_slide_info == FALSE )
		{
			mlOutSlideInfoStart(ge, cvt);
			have_slide_info = TRUE;
		}
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt, "%s <%d %d %d>\n", tok[TSLIDE_STYLE],
		        page->slide_effect,
		        page->slide_delay,
		        page->slide_throttle);
	}

	/* page layout ID */
	if ( page->layout_ix != 0 || page->object1 != 0 ||
	                             page->object2 != 0 ||
	                             page->object3 != 0 ||
	                             page->object4 != 0 )
	{
		if ( have_slide_info == FALSE )
		{
			mlOutSlideInfoStart(ge, cvt);
			have_slide_info = TRUE;
		}
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt, "%s <%d %d %d %d %d>\n", tok[TLAYOUT_ID],
		        page->layout_ix,
		        page->object1,
		        page->object2,
		        page->object3,
		        page->object4);
	}

	/* page background */
	if ( page->hide_objs )
	{
		if ( have_slide_info == FALSE )
		{
			mlOutSlideInfoStart(ge, cvt);
			have_slide_info = TRUE;
		}
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt, "%s %d\n", tok[THIDE_OBJECTS], page->hide_objs);
	}
	if ( page->hide_pg_backgd )
	{
		if ( have_slide_info == FALSE )
		{
			mlOutSlideInfoStart(ge, cvt);
			have_slide_info = TRUE;
		}
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt, "%s %d\n", tok[THIDE_BACKGD], page->hide_pg_backgd);
	}
	if ( page->use_backgd )
	{
		if ( have_slide_info == FALSE )
		{
			mlOutSlideInfoStart(ge, cvt);
			have_slide_info = TRUE;
		}
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt, "%s <%d %d %d %d>\n", tok[TPAGE_BACKGD],
		        page->c_scheme.page_pat,
		        page->c_scheme.page_type,
		        page->c_scheme.page_fg,
		        page->c_scheme.page_bg);
	}

	/* color scheme */
	if ( page->use_color )
	{
		if ( have_slide_info == FALSE )
		{
			mlOutSlideInfoStart(ge, cvt);
			have_slide_info = TRUE;
		}
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt, "%s <%d %d %d %d %d %d %d %d\n", tok[TC_SCHEME],
		        page->c_scheme.title_fg,
		        page->c_scheme.title_bg,
		        page->c_scheme.text_fg,
		        page->c_scheme.text_bg,
		        page->c_scheme.line_fg,
		        page->c_scheme.line_bg,
		        page->c_scheme.fill_fg,
		        page->c_scheme.fill_bg);
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt, "          %d %d %d %d %d %d %d %d\n",
		        page->c_scheme.accent_color[0],
		        page->c_scheme.accent_color[1],
		        page->c_scheme.accent_color[2],
		        page->c_scheme.accent_color[3],
		        page->c_scheme.accent_color[4],
		        page->c_scheme.accent_color[5],
		        page->c_scheme.accent_color[6],
		        page->c_scheme.accent_color[7]);
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt, "          %d %d %d %d %d %d %d %d",
		        page->c_scheme.accent_color[8],
		        page->c_scheme.accent_color[9],
		        page->c_scheme.accent_color[10],
		        page->c_scheme.accent_color[11],
		        page->c_scheme.accent_color[12],
		        page->c_scheme.accent_color[13],
		        page->c_scheme.accent_color[14],
		        page->c_scheme.shadow_color);
		if ( page->c_scheme.n_other_colors > 0 )
		{
			for ( i=0; i<page->c_scheme.n_other_colors; i++ )
			{
				if ( i%9 == 0 )
				{
					cprintf(ge, cvt, "\n");
					mlOutIndent(ge, cvt);
					cprintf(ge, cvt, "         ");
				}
				cprintf(ge, cvt, " %d", page->c_scheme.other_color[i]);
			}
		}
		cprintf(ge, cvt, ">\n");
	}
	/* write the page's build info */
	if ( mlNeedBuildInfo(ge, cvt, &page->build, page->use_slide_build))
	{
		if ( have_slide_info == FALSE )
		{
			mlOutSlideInfoStart(ge, cvt);
			have_slide_info = TRUE;
		}
		mlOutBuildInfo(ge, cvt, &page->build, page->use_slide_build);
	}

	if ( page->hide_slide )
	{
		if ( have_slide_info == FALSE )
		{
			mlOutSlideInfoStart(ge, cvt);
			have_slide_info = TRUE;
		}
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt, "%s %d\n", tok[THIDE_SLIDE], page->hide_slide);
	}

	/* make sure to end it if we have written anything */
	if ( have_slide_info )
	{
		cvt->indent--;
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt, "%s %s\n", tok[TEND], tok[TSLIDE_INFO]);
	}
	
}

/****************************************************************************
 *
 * module:  m l O u t T e m p l a t e I n f o
 */
static void
mlOutTemplateInfo(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	/* template info */
	if ( ge->template )
	{
		cprintf(ge, cvt, "%s\n", tok[TSLIDE_MASTER]);
		cvt->indent = 0;
		cprintf(ge, cvt, " %s <%d %d %d %d>\n", tok[TS_EXT1],
		        ge->template->slide_heading.x1 - PAGE_ZEROX,
		        ge->template->slide_heading.y1 - PAGE_ZEROY,
		        ge->template->slide_heading.x2 - PAGE_ZEROX,
		        ge->template->slide_heading.y2 - PAGE_ZEROY);
		cprintf(ge, cvt, " %s <%d %d %d %d>\n", tok[TS_EXT2],
		        ge->template->slide_content.x1 - PAGE_ZEROX,
		        ge->template->slide_content.y1 - PAGE_ZEROY,
		        ge->template->slide_content.x2 - PAGE_ZEROX,
		        ge->template->slide_content.y2 - PAGE_ZEROY);
		cprintf(ge, cvt," %s\n",tok[TT_ATTR0]);
		cvt->linepos = 0;
		if ( AttrCmp(ge, &ge->template->t0_attr, &cvt->curr_atts) )
		{
			cvt->indent = 2;
			mlOutAttrs(ge, &ge->template->t0_attr, 0, cvt);
			cvt->indent = 0;
		}
		cprintf(ge, cvt," %s %s\n",tok[TEND], tok[TT_ATTR0]);
		cprintf(ge, cvt," %s\n",tok[TT_ATTR1]);
		cvt->linepos = 0;
		if ( AttrCmp(ge, &ge->template->t1_attr, &cvt->curr_atts) )
		{
			cvt->indent = 2;
			mlOutAttrs(ge, &ge->template->t1_attr, 0, cvt);
			cvt->indent = 0;
		}
		cprintf(ge, cvt," %s %s\n",tok[TEND], tok[TT_ATTR1]);
		cprintf(ge, cvt," %s\n",tok[TS_ATTR0]);
		cvt->linepos = 0;
		if ( AttrCmp(ge, &ge->template->b0_attr, &cvt->curr_atts) )
		{
			cvt->indent = 2;
			mlOutAttrs(ge, &ge->template->b0_attr, 0, cvt);
			cvt->indent = 0;
		}
		cprintf(ge, cvt," %s %s\n",tok[TEND], tok[TS_ATTR0]);
		cprintf(ge, cvt," %s\n",tok[TS_ATTR1]);
		cvt->linepos = 0;
		if ( AttrCmp(ge, &ge->template->b1_attr, &cvt->curr_atts) )
		{
			cvt->indent = 2;
			mlOutAttrs(ge, &ge->template->b1_attr, 0, cvt);
			cvt->indent = 0;
		}
		cprintf(ge, cvt," %s %s\n",tok[TEND], tok[TS_ATTR1]);
		cprintf(ge, cvt," %s\n",tok[TS_ATTR2]);
		cvt->linepos = 0;
		if ( AttrCmp(ge, &ge->template->b2_attr, &cvt->curr_atts) )
		{
			cvt->indent = 2;
			mlOutAttrs(ge, &ge->template->b2_attr, 0, cvt);
			cvt->indent = 0;
		}
		cprintf(ge, cvt," %s %s\n",tok[TEND], tok[TS_ATTR2]);
		cprintf(ge, cvt," %s\n",tok[TS_ATTR3]);
		cvt->linepos = 0;
		if ( AttrCmp(ge, &ge->template->b3_attr, &cvt->curr_atts) )
		{
			cvt->indent = 2;
			mlOutAttrs(ge, &ge->template->b3_attr, 0, cvt);
			cvt->indent = 0;
		}
		cprintf(ge, cvt," %s %s\n",tok[TEND], tok[TS_ATTR3]);
		cprintf(ge, cvt," %s\n",tok[TS_ATTR4]);
		cvt->linepos = 0;
		if ( AttrCmp(ge, &ge->template->b4_attr, &cvt->curr_atts) )
		{
			cvt->indent = 2;
			mlOutAttrs(ge, &ge->template->b4_attr, 0, cvt);
			cvt->indent = 0;
		}
		cprintf(ge, cvt," %s %s\n",tok[TEND], tok[TS_ATTR4]);
		cprintf(ge, cvt," %s\n",tok[TS_ATTR5]);
		cvt->linepos = 0;
		if ( AttrCmp(ge, &ge->template->b5_attr, &cvt->curr_atts) )
		{
			cvt->indent = 2;
			mlOutAttrs(ge, &ge->template->b5_attr, 0, cvt);
			cvt->indent = 0;
		}
		cprintf(ge, cvt," %s %s\n",tok[TEND], tok[TS_ATTR5]);
		cvt->indent = 1;
		mlOutSlideInfo(ge, cvt, &ge->template->s_master_page);
		mlOutLayers   (ge, cvt, &ge->template->s_master_page);
		mlOutPart     (ge, cvt, &ge->template->s_master_page.t, FALSE);
		cvt->indent = 0;
		cprintf(ge, cvt, "%s %s\n", tok[TEND], tok[TSLIDE_MASTER]);

		if ( ge->template->t_master_page )
		{
			cprintf(ge, cvt, "%s\n", tok[TTITLE_MASTER]);
			cvt->indent = 0;
			cprintf(ge, cvt, " %s <%d %d %d %d>\n", tok[TS_EXT1],
		        		ge->template->title_heading.x1 - PAGE_ZEROX,
		        		ge->template->title_heading.y1 - PAGE_ZEROY,
		        		ge->template->title_heading.x2 - PAGE_ZEROX,
		        		ge->template->title_heading.y2 - PAGE_ZEROY);
			cprintf(ge, cvt, " %s <%d %d %d %d>\n", tok[TS_EXT2],
		        		ge->template->title_content.x1 - PAGE_ZEROX,
		        		ge->template->title_content.y1 - PAGE_ZEROY,
		        		ge->template->title_content.x2 - PAGE_ZEROX,
		        		ge->template->title_content.y2 - PAGE_ZEROY);
			cprintf(ge, cvt," %s\n",tok[TT_ATTR0]);
			cvt->linepos = 0;
			if ( AttrCmp(ge, &ge->template->ct0_attr, &cvt->curr_atts) )
			{
				cvt->indent = 2;
				mlOutAttrs(ge, &ge->template->ct0_attr, 0, cvt);
				cvt->indent = 0;
			}
			cprintf(ge, cvt," %s %s\n",tok[TEND], tok[TT_ATTR0]);
			cprintf(ge, cvt," %s\n",tok[TT_ATTR1]);
			cvt->linepos = 0;
			if ( AttrCmp(ge, &ge->template->ct1_attr, &cvt->curr_atts) )
			{
				cvt->indent = 2;
				mlOutAttrs(ge, &ge->template->ct1_attr, 0, cvt);
				cvt->indent = 0;
			}
			cprintf(ge, cvt," %s %s\n",tok[TEND], tok[TT_ATTR1]);
			cvt->linepos = 0;
		
			cvt->indent = 1;
			mlOutSlideInfo(ge, cvt, ge->template->t_master_page);
			mlOutLayers   (ge, cvt, ge->template->t_master_page);
			mlOutPart     (ge, cvt, &ge->template->t_master_page->t, FALSE);
			cvt->indent = 0;
			cprintf(ge, cvt, "%s %s\n", tok[TEND], tok[TTITLE_MASTER]);
		}

		cprintf(ge, cvt, "%s\n", tok[THANDOUT_MASTER]);
		cvt->indent = 1;
		mlOutSlideInfo(ge, cvt, &ge->template->h_master_page);
		mlOutLayers   (ge, cvt, &ge->template->h_master_page);
		mlOutPart     (ge, cvt, &ge->template->h_master_page.t, FALSE);
		cvt->indent = 0;
		cprintf(ge, cvt, "%s %s\n", tok[TEND], tok[THANDOUT_MASTER]);

		cprintf(ge, cvt, "%s\n", tok[TOUTLINE_MASTER]);
		cvt->indent = 1;
		mlOutSlideInfo(ge, cvt, &ge->template->o_master_page);
		mlOutLayers   (ge, cvt, &ge->template->o_master_page);
		mlOutPart     (ge, cvt, &ge->template->o_master_page.t, FALSE);
		cvt->indent = 0;
		cprintf(ge, cvt, "%s %s\n", tok[TEND], tok[TOUTLINE_MASTER]);

		cprintf(ge, cvt, "%s\n", tok[TNOTES_MASTER]);
		cprintf(ge, cvt, " %s <%d %d %d %d>\n", tok[TN_EXT1],
		        ge->template->notes_preview.x1 - PAGE_ZEROX,
		        ge->template->notes_preview.y1 - PAGE_ZEROY,
		        ge->template->notes_preview.x2 - PAGE_ZEROX,
		        ge->template->notes_preview.y2 - PAGE_ZEROY);
		cprintf(ge, cvt, " %s <%d %d %d %d>\n", tok[TN_EXT2],
		        ge->template->notes_content.x1 - PAGE_ZEROX,
		        ge->template->notes_content.y1 - PAGE_ZEROY,
		        ge->template->notes_content.x2 - PAGE_ZEROX,
		        ge->template->notes_content.y2 - PAGE_ZEROY);
		cprintf(ge, cvt,"%s\n",tok[TN_ATTR0]);
		cvt->linepos = 0;
		if ( AttrCmp(ge, &ge->template->n0_attr, &cvt->curr_atts) )
		{
			cvt->indent = 2;
			mlOutAttrs(ge, &ge->template->n0_attr, 0, cvt);
			cvt->indent = 0;
		}
		cprintf(ge, cvt,"%s %s\n",tok[TEND], tok[TN_ATTR0]);
		cprintf(ge, cvt,"%s\n",tok[TN_ATTR1]);
		cvt->linepos = 0;
		if ( AttrCmp(ge, &ge->template->n1_attr, &cvt->curr_atts) )
		{
			cvt->indent = 2;
			mlOutAttrs(ge, &ge->template->n1_attr, 0, cvt);
			cvt->indent = 0;
		}
		cprintf(ge, cvt,"%s %s\n",tok[TEND], tok[TN_ATTR1]);
		cprintf(ge, cvt,"%s\n",tok[TN_ATTR2]);
		cvt->linepos = 0;
		if ( AttrCmp(ge, &ge->template->n2_attr, &cvt->curr_atts) )
		{
			cvt->indent = 2;
			mlOutAttrs(ge, &ge->template->n2_attr, 0, cvt);
			cvt->indent = 0;
		}
		cprintf(ge, cvt,"%s %s\n",tok[TEND], tok[TN_ATTR2]);
		cprintf(ge, cvt,"%s\n",tok[TN_ATTR3]);
		cvt->linepos = 0;
		if ( AttrCmp(ge, &ge->template->n3_attr, &cvt->curr_atts) )
		{
			cvt->indent = 2;
			mlOutAttrs(ge, &ge->template->n3_attr, 0, cvt);
			cvt->indent = 0;
		}
		cprintf(ge, cvt,"%s %s\n",tok[TEND], tok[TN_ATTR3]);
		cprintf(ge, cvt,"%s\n",tok[TN_ATTR4]);
		cvt->linepos = 0;
		if ( AttrCmp(ge, &ge->template->n4_attr, &cvt->curr_atts) )
		{
			cvt->indent = 2;
			mlOutAttrs(ge, &ge->template->n4_attr, 0, cvt);
			cvt->indent = 0;
		}
		cprintf(ge, cvt,"%s %s\n",tok[TEND], tok[TN_ATTR4]);
		cprintf(ge, cvt,"%s\n",tok[TN_ATTR5]);
		cvt->linepos = 0;
		if ( AttrCmp(ge, &ge->template->n5_attr, &cvt->curr_atts) )
		{
			cvt->indent = 2;
			mlOutAttrs(ge, &ge->template->n5_attr, 0, cvt);
			cvt->indent = 0;
		}
		cprintf(ge, cvt,"%s %s\n",tok[TEND], tok[TN_ATTR5]);
		cvt->indent = 1;
		mlOutSlideInfo(ge, cvt, &ge->template->n_master_page);
		mlOutLayers   (ge, cvt, &ge->template->n_master_page);
		mlOutPart     (ge, cvt, &ge->template->n_master_page.t, FALSE);
		cvt->indent = 0;
		cprintf(ge, cvt, "%s %s\n\n", tok[TEND], tok[TNOTES_MASTER]);
	}
}

/****************************************************************************
 *
 * module:  mlOutNumberFormat
 */
static void
mlOutNumberFormat(ge,cvt,fmt)
graphics_main_type *ge;
cvt_global_type *cvt;
number_format *fmt;
{
    tx_char tbuff[500];
	int ii, len;

	/* format info */
	mlOutIndent(ge, cvt);
	cvt->linepos += cprintf(ge, cvt, "%s <", tok[TNUM_FMT]);
	cvt->linepos += cprintf(ge, cvt,"%d ", fmt->id);
	cvt->linepos += cprintf(ge, cvt,"%d ", fmt->aux_id);
	cvt->linepos += cprintf(ge, cvt,"%d ", fmt->display_format_errors);
	cvt->linepos += cprintf(ge, cvt,"%d ", fmt->trim_string);
	cvt->linepos += cprintf(ge, cvt,"%d ", fmt->units);
	cvt->linepos += cprintf(ge, cvt,"%d ", fmt->n_radix_places);

	/* the prefix and suffix strings */
	if ( !fmt->prefix )
		cvt->linepos += cprintf(ge, cvt,"\"\" ");
	else
	{
		len = MultiToWideStr(ge, fmt->prefix, tbuff );
		cvt->linepos += mlOutString(ge, len, tbuff, cvt, FALSE);
		cvt->linepos += cprintf(ge, cvt," ");
	}
	if ( !fmt->suffix )
		cvt->linepos += cprintf(ge, cvt,"\"\" ");
	else
	{
		len = MultiToWideStr(ge, fmt->suffix, tbuff );
		cvt->linepos += mlOutString(ge, len, tbuff, cvt, FALSE);
		cvt->linepos += cprintf(ge, cvt," ");
	}

	/* am and pm strings */
	if ( !fmt->am )
		cvt->linepos += cprintf(ge, cvt,"\"\" ");
	else
	{
		len = MultiToWideStr(ge, fmt->am, tbuff );
		cvt->linepos += mlOutString(ge, len, tbuff, cvt, FALSE);
		cvt->linepos += cprintf(ge, cvt," ");
	}
	if ( !fmt->pm )
		cvt->linepos += cprintf(ge, cvt,"\"\" ");
	else
	{
		len = MultiToWideStr(ge, fmt->pm, tbuff );
		cvt->linepos += mlOutString(ge, len, tbuff, cvt, FALSE);
		cvt->linepos += cprintf(ge, cvt," ");
	}

	/* the boolean strings */
	if ( !fmt->true )
		cvt->linepos += cprintf(ge, cvt,"\"\" ");
	else
	{
		len = MultiToWideStr(ge, fmt->true, tbuff );
		cvt->linepos += mlOutString(ge, len, tbuff, cvt, FALSE);
		cvt->linepos += cprintf(ge, cvt," ");
	}
	if ( !fmt->false )
		cvt->linepos += cprintf(ge, cvt,"\"\" ");
	else
	{
		len = MultiToWideStr(ge, fmt->false, tbuff );
		cvt->linepos += mlOutString(ge, len, tbuff, cvt, FALSE);
		cvt->linepos += cprintf(ge, cvt," ");
	}

	/* the separator characters */
	len = 1;
	tbuff[1] = (tx_char)0;
	tbuff[0] = (tx_char)fmt->t_sep;
	cvt->linepos += mlOutString(ge, len, tbuff, cvt, FALSE);
	cvt->linepos += cprintf(ge, cvt," ");
	tbuff[0] = (tx_char)fmt->d_sep;
	cvt->linepos += mlOutString(ge, len, tbuff, cvt, FALSE);
	cvt->linepos += cprintf(ge, cvt,">\n");

	/* done */
	cvt->linepos = 0;
}

/****************************************************************************
 *
 * module:  mlOutChartLegendType
 */
static void
mlOutChartLegendType (ge,cvt,legend)
graphics_main_type *ge;
cvt_global_type *cvt;
chart_legend_type *legend;
{
    tx_char tbuff[500];
    int len;

	if (!legend)
		return;

	/* the legend token and the optional legend title */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s ",tok[TLEGEND]);
	if (!legend->title_str)
		cprintf(ge, cvt,"\"\"");
	else
	{
        len = MultiToWideStr(ge,  legend->title_str, tbuff );
        mlOutString(ge,  len, tbuff, cvt, FALSE );
	}
	cprintf(ge, cvt,"\n");
	cvt->linepos = 0;

	/* the legend position */
	cvt->indent++;
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt, "%s <",   tok[TLEGALIGN]);
	cprintf(ge, cvt, "%d %d ", legend->h_align, legend->h_off);
	cprintf(ge, cvt, "%d %d ", legend->v_align, legend->v_off);
	cprintf(ge, cvt, "%d>\n",  legend->title_align);
	cvt->linepos = 0;

	/* the legend format */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt, "%s ",   tok[TLEGFMT]);
	cprintf(ge, cvt, "<%d",   legend->disabled);
	cprintf(ge, cvt, " %d",   legend->x_space_eating);
	cprintf(ge, cvt, " %d",   legend->y_space_eating);
	cprintf(ge, cvt, " %d",   legend->text_before_sample);
	cprintf(ge, cvt, " %d",   legend->arranged_by_row);
	cprintf(ge, cvt, " %d>\n",legend->max_per_row_or_column);
	cvt->linepos = 0;

	/* legend margins */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt, "%s ",   tok[TLEGMARG]);
	cprintf(ge, cvt, "<%d",   legend->t_margin);
	cprintf(ge, cvt, " %d",   legend->b_margin);
	cprintf(ge, cvt, " %d",   legend->l_margin);
	cprintf(ge, cvt, " %d",   legend->r_margin);
	cprintf(ge, cvt, " %d",   legend->row_margin);
	cprintf(ge, cvt, " %d ",  legend->column_margin);
	cprintf(ge, cvt, " %d>\n",legend->auto_margins);

	/* the attributes of the various parts of the legend */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n",tok[TLEGBOXATTR]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &legend->box_attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &legend->box_attr, 0, cvt);
		cvt->indent--;
	}

	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n",tok[TLEGLABATTR]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &legend->label_attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &legend->label_attr, 0, cvt);
		cvt->indent--;
	}

	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n",tok[TLEGTITATTR]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &legend->title_attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &legend->title_attr, 0, cvt);
		cvt->indent--;
	}

	/* done with legends */
	cvt->indent--;
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s %s\n",tok[TEND],tok[TLEGEND]);
	cvt->linepos = 0;
}

/****************************************************************************
 *
 * module:  mlOutChartDataType
 */
static void
mlOutChartDataType(ge,cvt,dat)
graphics_main_type *ge;
cvt_global_type *cvt;
chart_data_point_type *dat;
{
	tx_char tbuff[500];
	int len, ii;
	double mag;
	/* the data point token and the display type */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <",tok[TDAT]);
	switch(dat->display_flag)
	{
		default:
		case BG_PLOT: cprintf(ge, cvt, "%d ", BG_PLOT); break;
		case BG_SUBS: cprintf(ge, cvt, "%d ", BG_SUBS); break;
		case BG_SPAN: cprintf(ge, cvt, "%d ", BG_SPAN); break;
		case BG_GAP:  cprintf(ge, cvt, "%d ", BG_GAP);  break;
		case BG_NULL: cprintf(ge, cvt, "%d ", BG_NULL); break;
	}

	/* the data point label */
	if (!dat->label_str)
	cprintf(ge, cvt,"\"\"");
	else
	{
		len = MultiToWideStr(ge,  dat->label_str, tbuff );
		mlOutString(ge,  len, tbuff, cvt, FALSE );
	}

	/* the data label offsets */
	if ( dat->local_x )
		cprintf(ge, cvt," %d", dat->label_x_off);
	else
		cprintf(ge, cvt," \"\"");
	if ( dat->local_y )
		cprintf(ge, cvt," %d", dat->label_y_off);
	else
		cprintf(ge, cvt," \"\"");
	cprintf(ge, cvt, ">\n");
	cvt->linepos = 0;

	/* data point x position and auxiliary data */
	cvt->indent++;
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt, "%s <", tok[TX]);
	for ( ii=0; ii<dat->nx; ii++ )
	{
		if ( ii != 0 )
			cprintf(ge, cvt, " ");
		if ((fabs((double)dat->x[ii]) > 1.0e9 ||
			fabs((double)dat->x[ii]) < 1.0e-7) && dat->x[ii] )
			cprintf(ge, cvt,"%.7e ", dat->x[ii]);
		else
			cprintf(ge, cvt,"%f ",   dat->x[ii]);
	}

	cprintf(ge, cvt, ">\n");
	cvt->linepos = 0;

	/* data point y position and auxiliary data */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt, "%s <", tok[TY]);
	for ( ii=0; ii<dat->ny; ii++ )
	{
		if ( ii != 0 )
			cprintf(ge, cvt, " ");
		if ( (fabs((double)dat->y[ii]) > 1.0e9 ||
			 fabs((double)dat->y[ii]) < 1.0e-7) && dat->y[ii] )
			cprintf(ge, cvt,"%.7e ", dat->y[ii]);
		else
			cprintf(ge, cvt,"%f ",   dat->y[ii]);
	}
	cprintf(ge, cvt, ">\n");
	cvt->linepos = 0;

	/* data point z position and auxiliary data */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt, "%s <", tok[TZ]);
	for ( ii=0; ii<dat->nz; ii++ )
	{
		if ( ii != 0 )
			cprintf(ge, cvt, " ");
		if ( (fabs((double)dat->z[ii]) > 1.0e9 ||
			 fabs((double)dat->z[ii]) < 1.0e-7) && dat->z[ii] )
			cprintf(ge, cvt,"%.7e ", dat->z[ii]);
		else
			cprintf(ge, cvt,"%f ",   dat->z[ii]);
	}
	cprintf(ge, cvt, ">\n");
	cvt->linepos = 0;
	cvt->indent--;
}

/****************************************************************************
 *
 * module:  mlOutChartDataGroupType
 */
static void
mlOutChartDataGroupType(ge,cvt,grp)
graphics_main_type *ge;
cvt_global_type *cvt;
chart_data_group_type *grp;
{
	tx_char tbuff[500];
	int len, ii;

	/* the beginning of the data group and the group name */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <", tok[TDATGROUP]);
	len = MultiToWideStr(ge, grp->name, tbuff);
	mlOutString(ge, len, tbuff, cvt, FALSE);
	cprintf(ge, cvt, " %d>\n", grp->disabled);
	cvt->linepos = 0;

	/* the legend string */
	cvt->indent++;
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s ", tok[TLABEL]);
	if ( !grp->legend_str )
		cprintf(ge, cvt,"\"\"");
	else
	{
		len = MultiToWideStr(ge, grp->legend_str, tbuff);
		mlOutString(ge, len, tbuff, cvt, FALSE);
	}
	cprintf(ge, cvt,"\n");
	cvt->linepos = 0;

	/* data type (line, bar, etc...), stack id, and custom string */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <%d %d ", tok[TDATATYPE], grp->type, grp->stack_id);
	if ( !grp->part )
		cprintf(ge, cvt,"\"\"");
	else
	{
		len = MultiToWideStr(ge, grp->part, tbuff);
		mlOutString(ge, len, tbuff, cvt, FALSE);
	}
	cprintf(ge, cvt," %d %d", grp->new, grp->old_ix);
	cprintf(ge, cvt," %d", grp->disabled);
	cprintf(ge, cvt,">\n");
	cvt->linepos = 0;

	/* trend info,  if any.. handle exponential formats*/
	if ( grp->type == 15 )
	{
		mlOutIndent(ge, cvt);
		if ( (fabs((double)grp->trend_info.r_squared) > 1.0e7 ||
				  fabs((double)grp->trend_info.r_squared) < 1.0e-5 ) )
				cprintf(ge, cvt,"%s <%d %d %.7e %d ", tok[TTREND], 
						grp->trend_info.type,
						grp->trend_info.order,
						grp->trend_info.r_squared,
						grp->trend_info.n_correlations);
		else
			cprintf(ge, cvt,"%s <%d %d %f %d ", tok[TTREND], 
					grp->trend_info.type,
					grp->trend_info.order,
					grp->trend_info.r_squared,
					grp->trend_info.n_correlations);
		
		for( ii = 0; ii < grp->trend_info.n_correlations; ii++)
		{
			if ( (fabs((double)grp->trend_info.correlations[ii]) > 1.0e5 ||
				  fabs((double)grp->trend_info.correlations[ii]) < 1.0e-5 ) )
				cprintf(ge, cvt,"%.5e ", grp->trend_info.correlations[ii]);
			else
				cprintf(ge, cvt,"%f ",   grp->trend_info.correlations[ii]);

		}
		cprintf(ge, cvt,">\n");
		cvt->linepos = 0;
	}
	
	/* the axes this data is drawn against */	
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <%d %d %d>\n", tok[TAXES], grp->x_axis,
	                                   grp->y_axis,
	                                   grp->z_axis);
	cvt->linepos = 0;

	/* data label info */
	mlOutNumberFormat(ge,cvt,&grp->label_frmt);
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt, "%s <%d %d %d %d>\n", tok[TDATALABEL],
	                                   grp->data_label_type,
	                                   grp->label_x_off,
	                                   grp->label_y_off,
	                                   grp->label_align);
	cvt->linepos = 0;

	/* write each data point, calculating grp->n at read time */
	for (ii=0; ii<grp->n; ii++)
		mlOutChartDataType(ge,cvt,&grp->dat[ii]);

	/* write the data label and data attributes */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n", tok[TGRPLABATTR]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &grp->label_attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &grp->label_attr, 0, cvt);
		cvt->indent--;
	}

	/* write out enable buffers and number of buffers in a group */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <%d %d>\n", tok[TGRPBUF], grp->enable_bufs,
				                            grp->n_bufs);
	cvt->linepos = 0;
	
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n", tok[TGRPATTR]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &grp->attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &grp->attr, 0, cvt);
		cvt->indent--;
	}

	/* the end of the data group */
	cvt->indent--;
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s %s\n", tok[TEND], tok[TDATGROUP]);
	cvt->linepos = 0;
}

/****************************************************************************
 *
 * module:  mlOutChartAxisType
 */
static void
mlOutChartAxisType(ge,cvt,axis)
graphics_main_type *ge;
cvt_global_type *cvt;
chart_axis_type *axis;
{
    tx_char tbuff[500];
	int ii, len;
	double mag;

	/* type and space consumed (for additional axes) */
	cvt->indent++;
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <",   tok[TTIK_TYPE]);
	cprintf(ge, cvt,"%d ",    axis->type);
	cprintf(ge, cvt,"%d ",    axis->x_offset);
	cprintf(ge, cvt,"%d ",    axis->y_offset);
	cprintf(ge, cvt,"%d>\n",  axis->tik_margin);
	cvt->linepos = 0;

	/* tik and label alignment */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <",  tok[TTIK_ALIGN]);
	cprintf(ge, cvt,"%d ",   axis->label_align);
	cprintf(ge, cvt,"%d ",   axis->tik_align);
	cprintf(ge, cvt,"%d ",   axis->tight_label);
	cprintf(ge, cvt,"%d>\n", axis->tight_tiks);
	cvt->linepos = 0;

	/* axis format info */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <",  tok[TAXIS_FMT]);
	cprintf(ge, cvt,"%d ",   axis->hidden);
	cprintf(ge, cvt,"%d ",   axis->auto_max);
	cprintf(ge, cvt,"%d ",   axis->auto_min);
	cprintf(ge, cvt,"%d ",   axis->auto_bas);
	cprintf(ge, cvt,"%d ",   axis->auto_n_major_tiks);
	cprintf(ge, cvt,"%d ",   axis->auto_n_minor_tiks);
	cprintf(ge, cvt,"%d ",   axis->major_tik_format);
	cprintf(ge, cvt,"%d ",   axis->minor_tik_format);
	cprintf(ge, cvt,"%d ",   axis->pos);
	
	if ( (fabs((double)axis->pos_val) > 1.0e7 ||
		 fabs((double)axis->pos_val) < 1.0e-5) && axis->pos_val )
		cprintf(ge, cvt,"%.5e ", axis->pos_val);
	else
		cprintf(ge, cvt,"%f ",   axis->pos_val);

	if ( (fabs((double)axis->min_val) > 1.0e7 ||
		 fabs((double)axis->min_val) < 1.0e-9) && axis->min_val )
		cprintf(ge, cvt,"%.5e ", axis->min_val);
	else
		cprintf(ge, cvt,"%.9f ",   axis->min_val);

	if ( (fabs((double)axis->max_val) > 1.0e7 ||
		 fabs((double)axis->max_val) < 1.0e-9) && axis->max_val )
		cprintf(ge, cvt,"%.5e ", axis->max_val);
	else
		cprintf(ge, cvt,"%.9f ",   axis->max_val);

	if ( (fabs((double)axis->bas_val) > 1.0e7 ||
		 fabs((double)axis->bas_val) < 1.0e-9) && axis->bas_val)
		cprintf(ge, cvt,"%.5e ", axis->bas_val);
	else
		cprintf(ge, cvt,"%.9f ",   axis->bas_val);
	
	cprintf(ge, cvt,"%f ",   axis->first_delta);
	cprintf(ge, cvt,"%d ",   axis->invalid_tiks);
	cprintf(ge, cvt,"%d ",   axis->draw_behind);
	cprintf(ge, cvt,"%d ",   axis->use_tik_attr);	
	cprintf(ge, cvt,"%d>\n", axis->inverted);	
	
	cvt->linepos = 0;

	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s ",  tok[TAXIS_MAJOR]);
	cprintf(ge, cvt,"<%d %d %f %f>\n", axis->major_sz,
	                                   axis->n_maj_steps,
	                                   axis->maj_delta,
	                                   axis->maj_step);
	cvt->linepos = 0;
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s ",  tok[TAXIS_MINOR]);
	cprintf(ge, cvt,"<%d %d %f %f>\n", axis->minor_sz,
	                                   axis->n_min_steps,
	                                   axis->min_delta,
	                                   axis->min_step);
	cvt->linepos = 0;

	/* numerical format for tiks */
	mlOutNumberFormat(ge,cvt,&axis->tik_format);

	/* bar info */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <%f %f>\n", tok[TBAR_INFO],
	                                axis->bar_margin,
	                                axis->bar_overlap);
	cvt->linepos = 0;

	/* tik label filter (for catagory axes) */
	cprintf(ge, cvt,"  %s <%d %d %d>\n", tok[TLABEL_FILTER], 
			axis->filter_labels, axis->c_tiks[0], axis->c_tiks[1]);

	/* tik labels (for catagory axes) */
	for ( ii=0; ii<axis->n; ii++ )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt, "%s <", tok[TLABEL]);
		if ( !axis->tik_str[ii] )
			cprintf(ge, cvt,"\"\"");
		else
		{
			len = MultiToWideStr(ge, axis->tik_str[ii], tbuff);
			mlOutString(ge, len, tbuff, cvt, FALSE);
		}
		if ( axis->display )
			cprintf(ge, cvt, " %d", axis->display[ii]);
		else
			cprintf(ge, cvt, " 1");
		cprintf(ge, cvt,">\n");
		cvt->linepos = 0;
	}
 
	/* axis attribute info */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n", tok[TAXATTR]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &axis->attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &axis->attr, 0, cvt);
		cvt->indent--;
	}
 
 
	/* axis tik attribute info */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n", tok[TAXTIKATTR]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &axis->tik_attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &axis->tik_attr, 0, cvt);
		cvt->indent--;
	}

	
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n", tok[TAXLABATTR]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &axis->label_attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &axis->label_attr, 0, cvt);
		cvt->indent--;
	}
 
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n", tok[TAXTIKLABATTR]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &axis->tik_label_attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &axis->tik_label_attr, 0, cvt);
		cvt->indent--;
	}

	/* done */
	cvt->indent--;
}

/****************************************************************************
 *
 * module:  b g W r i t e E l e m
 */
static void
bgWriteElem(ge, cvt, str, indent)
graphics_main_type *ge;
cvt_global_type *cvt;
char *str;
int indent;
{
	int len;
	tx_char tbuff[STRPATH];
	
	if ( str[0] == '\0' )
		cprintf(ge, cvt, "\"\"");
	else
	{	
		len = MultiToWideStr(ge, str, tbuff );
		mlOutString(ge, len, tbuff, cvt, FALSE);
	}
}

/****************************************************************************
 *
 * module:  b g W r i t e I t e m
 */
static void
bgWriteItem(ge, cvt, chart, sheet, item, do_indent)
graphics_main_type *ge;
cvt_global_type *cvt;
chart_type *chart;
chart_sheet_info *sheet;
sheet_part_element *item;
int do_indent;
{
	sheet_part_element *n_item, *a_item;
	int i, j;
	
	if ( item->n == 1 && item->type == ISA_VALUE )
	{	
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt, "%s ", tok[TELEM]);
		bgWriteElem(ge, cvt, n_item->val, FALSE);
		cprintf(ge, cvt, "\n");
	}
	else
	{
		if ( item->n > 0 )
		{
			for ( i=0; i<item->n; i++ )
			{				
				n_item = item->item[i];
				if ( !n_item ) continue;
				if ( n_item->n >= 1 && n_item->type == ISA_ARRAY )
				{	
					mlOutIndent(ge, cvt);
					cprintf(ge, cvt, "%s <", tok[TARR]);
					a_item = n_item->item[0];
					bgWriteElem(ge, cvt, a_item->val, FALSE);
					for ( j=1; j<n_item->n; j++)
					{
						cprintf(ge, cvt, " ");
						a_item = n_item->item[j];
						bgWriteElem(ge, cvt, a_item->val, FALSE);
					}
					cprintf(ge, cvt, ">\n");
				}
				else
				{	
					mlOutIndent(ge, cvt);
					cprintf(ge, cvt, "%s ", tok[TELEM]);
					if ( n_item->type == ISA_VALUE )
						bgWriteElem(ge, cvt, n_item->val, FALSE);
					else
					{
						a_item = n_item->item[0];
						bgWriteElem(ge, cvt, a_item->val, FALSE);
					}
					cprintf(ge, cvt, "\n");
				}
			}
		}
	}
}

/****************************************************************************
 *
 * module:  b g W r i t e P a r t
 */
static void
bgWritePart(ge, cvt, chart, sheet, part)
graphics_main_type *ge;
cvt_global_type *cvt;
chart_type *chart;
chart_sheet_info *sheet;
sheet_part *part;
{
	sheet_part_element *item, *n_item;

	item = &part->item;
	if ( item->n == 1 && item->type == ISA_VALUE )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt, "%s ", tok[TELEM]);
		bgWriteElem(ge, cvt, item->val, FALSE);
		cprintf(ge, cvt, "\n");
	}
	else
		bgWriteItem(ge, cvt, chart, sheet, item, FALSE);
}

/****************************************************************************
 *
 * module:  b g W r i t e P a r t s: called in mlOutChartSheet
 */
static void
bgWriteParts(ge, cvt, chart)
graphics_main_type *ge;
cvt_global_type *cvt;
chart_type *chart;
{
	int i, n, len;
	sheet_part *part;
	tx_char tbuff[STRPATH];

	for( i=0; i<chart->sheet->n_parts; i++ ) 
	{
		part = chart->sheet->part[i];
		mlOutIndent(ge, cvt);
		if ( part->item.n >= 1 )			
			cprintf(ge, cvt, "%s <", tok[TSH_ARR]);
		else
			cprintf(ge, cvt, "%s <", tok[TSH_ELEM]);
		len = MultiToWideStr(ge, part->name, tbuff );
		mlOutString(ge, len, tbuff, cvt, FALSE);
		cprintf(ge, cvt, ">\n");
		cvt->indent++;
		bgWritePart(ge, cvt, chart, chart->sheet, part);
		cvt->indent--;
	}
}

/****************************************************************************
 *
 * module:  mlOutChartSheet
 */
static void
mlOutChartSheet(ge, cvt, chart)
graphics_main_type *ge;
cvt_global_type *cvt;
register chart_type *chart;
{
	chart_sheet_info *sheet;
	tx_char tbuff[STRPATH];
	int i, j, m, n, len;
	sheet_data *dat;
	
	sheet = chart->sheet;
	m = sheet->n_rows;
	n = sheet->n_cols;

	/* write the categories <-- first row of data */
	cprintf(ge, cvt, "%s <", tok[TROWDAT]);
	cprintf(ge, cvt, "\"\" ");
	for ( j=0; j<n; j++)
	{	
		cprintf(ge, cvt, " ");
		if ( (j+1) % NUM_ELEMS_IN_ROW == 0 )
		{
			cprintf(ge, cvt, "\n");
			mlOutIndent(ge, cvt);
		}
		if ( chart->col_sliced == 1 )
			len = MultiToWideStr(ge, sheet->legends[j], tbuff );
		else
			len = MultiToWideStr(ge, sheet->categories[j], tbuff );
		mlOutString(ge, len, tbuff, cvt, FALSE);
	}
	cprintf(ge, cvt, ">\n");
	for( i=1; i<=m; i++)
	{
		/* write the legend string and then any data in that row*/
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s <", tok[TROWDAT], sheet->legends[i-1]);
		if ( chart->col_sliced == 1 )
			len = MultiToWideStr(ge, sheet->categories[i-1], tbuff );
		else
			len = MultiToWideStr(ge, sheet->legends[i-1], tbuff );
		mlOutString(ge, len, tbuff, cvt, FALSE);
		cprintf(ge, cvt, " ");
		
		/* write the first cell => j=0 */
		dat = sheet->data[i-1][0];
		if (dat->isa_val == 2)
			cprintf(ge, cvt, "%lf", dat->u.val);
		else if ( dat->isa_val == 1)
		{		
			len = MultiToWideStr(ge, dat->u.str, tbuff );
			mlOutString(ge, len, tbuff, cvt, FALSE);
			cprintf(ge, cvt, " ");
		}
		else
			cprintf(ge, cvt, "\"\"");
		
		/* write out remaining cells */
		for( j=1; j<n; j++ )
		{
			cprintf(ge, cvt, " ");
			dat = sheet->data[i-1][j];
			if ( (j+1) % NUM_ELEMS_IN_ROW == 0 )
			{
				cprintf(ge, cvt, "\n");
				mlOutIndent(ge, cvt);
			}
			if (dat->isa_val == 2)
				cprintf(ge, cvt, "%lf", dat->u.val);
			else if ( dat->isa_val == 1)
			{		
				len = MultiToWideStr(ge, dat->u.str, tbuff );
				mlOutString(ge, len, tbuff, cvt, FALSE);
			}
			else
				cprintf(ge, cvt, "\"\"");
		}
		cprintf(ge, cvt, ">\n");
	}
	bgWriteParts(ge, cvt, chart);
	if ( sheet->n_col_sz )
	{	
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt, "%s <", tok[TCOLUMNS]);
		cprintf(ge, cvt, "%d", sheet->col_sz[0]);		
		for ( i=1; i<sheet->n_col_sz; i++ )
		{
			cprintf(ge, cvt, " ");
			if ( (i+1) % NUM_ELEMS_IN_ROW == 0 )
			{
				cprintf(ge, cvt, "\n");
				mlOutIndent(ge, cvt);
			}
			else
				cprintf(ge, cvt, "%d", sheet->col_sz[i]);
		}
		cprintf(ge, cvt, ">\n");
	}
}

/****************************************************************************
 *
 * module:  mlOutChartData
 */
static void
mlOutChartData(ge,cvt,chart)
graphics_main_type *ge;
cvt_global_type *cvt;
register chart_type *chart;
{
	register int ii, len;
	tx_char tbuff[500];

	/* chart token and name of chart */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s ", tok[TCHART]);
	len = MultiToWideStr(ge, chart->name, tbuff);
	mlOutString(ge, len, tbuff, cvt, FALSE);
	cprintf(ge, cvt,"\n");
	cvt->linepos = 0;
	cvt->indent++;

	/* chart type and orientation */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <%d %d %d>\n", tok[TTYPE], 
								chart->type, 
								chart->orientation,
								chart->old_orient);
	cvt->linepos = 0;

	/* title, subtitle, and footer strings */
	if ( chart->title )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s ", tok[TTITLE]);
		len = MultiToWideStr(ge, chart->title, tbuff );
		mlOutString(ge,  len, tbuff, cvt, FALSE );
		cprintf(ge, cvt,"\n");
		cvt->linepos=0;
	}
	if ( chart->subtitle )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s ", tok[TSUBTITLE]);
		len = MultiToWideStr(ge, chart->subtitle, tbuff );
		mlOutString(ge,  len, tbuff, cvt, FALSE );
		cprintf(ge, cvt,"\n");
		cvt->linepos=0;
	}
	if ( chart->footer )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s ", tok[TFOOT]);
		len = MultiToWideStr(ge,  chart->footer, tbuff );
		mlOutString(ge,  len, tbuff, cvt, FALSE );
		cprintf(ge, cvt,"\n");
		cvt->linepos=0;
	}
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <%d %d %d %d %d %d>\n", tok[TTITLE_ALIGN],
									    chart->title_align,
	                               	         chart->subtitle_align,
	                               		    chart->footer_align,
								   	    chart->title_expand,
	                               	         chart->subtitle_expand,
	                               	         chart->footer_expand);
	cvt->linepos=0;

	/* chart decorations (closing lines and grid lines) */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt, "%s <%d %d %d %d %d %d>\n", tok[TCLOSE], 
					   				     chart->close_top,
	                               	          chart->close_bot,
	                               	          chart->close_lft,
	                                             chart->close_rt,
									     chart->close_bg,
									     chart->close_fg);
	cvt->linepos=0;
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <%d %d %d %d %d %d>\n", tok[TGRIDS], 
								   chart->maj_x_grids,
 								   chart->min_x_grids,
	                        	             chart->maj_y_grids,
	                        	             chart->min_y_grids,
								   chart->maj_z_grids,
								   chart->min_z_grids);
	cvt->linepos=0;
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <%d %d %d %d %d %d %d %d>\n", tok[TOFFSET], 
										     chart->def_x_offset,
	                                                  chart->def_y_offset,
											chart->title_x_off,
											chart->title_y_off,		
											chart->subtitle_x_off,
											chart->subtitle_y_off,
											chart->footer_x_off,
											chart->footer_y_off);
	cvt->linepos=0;

	/* chart size and margins */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <%d %d %d %d>\n", tok[TEXTENT], 
								   chart->extent.x1,
	                                      chart->extent.y1,
	                                      chart->extent.x2,
	                                      chart->extent.y2);
	cvt->linepos=0;
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <%d %d %d %d>\n", tok[TMARGIN], chart->margin.x1,
	                                      chart->margin.y1,
	                                      chart->margin.x2,
	                                      chart->margin.y2);
	cvt->linepos=0;
 
	/* count of original axes and groups (as of last rendering) */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <%d %d %d %d %d>\n", tok[TCHART_COUNTS],
	                                         chart->old_n_x_axes,
	                                         chart->old_n_y_axes,
	                                         chart->old_n_z_axes,
	                                         chart->old_n_groups,
	                                         chart->old_type);
	cvt->linepos=0;

	/* default null treatment */
	mlOutIndent(ge, cvt);
	switch ( chart->null_format )
	{
		default:
		case 0:cprintf(ge, cvt,"%s %s\n",  tok[TNULL_FMT], tok[TPLOT]); break;
		case 1: cprintf(ge, cvt,"%s %s\n", tok[TNULL_FMT], tok[TSUBS]); break;
		case 2: cprintf(ge, cvt,"%s %s\n", tok[TNULL_FMT], tok[TSPAN]); break;
		case 3: cprintf(ge, cvt,"%s %s\n", tok[TNULL_FMT], tok[TGAP]); break;
	}
	cvt->linepos=0;

	/* 3D */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt, "%s <%d %d %d %d %d %d %d %d %d %d %d %d>\n",
	        tok[TTHREE_D],
	        chart->three_d.enabled, chart->old_three_d.enabled,
	        chart->three_d.pitch, chart->old_three_d.pitch,
	        chart->three_d.yaw, chart->old_three_d.yaw,
	        chart->three_d.projection, chart->old_three_d.projection,
	        chart->three_d.depth, chart->old_three_d.depth,
			chart->three_d.right_angles, chart->old_three_d.right_angles);
	cvt->linepos=0;

	/* chart type specific info */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <%d %f %f %d %d %d %d>\n",tok[TCHART_INFO],
										chart->info.stacked_strata, 	
										chart->info.bar_margin,
										chart->info.bar_overlap,	
										chart->info.start_angle,
										chart->info.def_type,
										chart->data_src.type,
										chart->data_src.ix);
	cvt->linepos=0;		

	
	/* chart type misc info */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s <%d %d>\n",tok[TMISC_CHART_INFO],
			chart->needs_redraw, chart->col_sliced);
								
	cvt->linepos=0;		


	/* Save chart sheet info if any */
	if ( chart->sheet )
	{	
		mlOutIndent(ge, cvt);
		mlOutChartSheet(ge, cvt, chart);
		cvt->linepos=0;
	}
	/* the axes */
	if ( chart->n_x_axes )
	{
		for ( ii=0; ii<chart->n_x_axes; ii++ )
		{
			mlOutIndent(ge, cvt);
			cprintf(ge, cvt,"%s ", tok[TX_AXIS]);
			if ( !chart->x_axis[ii].label_str )
				cprintf(ge, cvt,"\"\"");
			else
			{
				len = MultiToWideStr(ge, chart->x_axis[ii].label_str, tbuff );
				mlOutString(ge,  len, tbuff, cvt, FALSE );
			}
			cprintf(ge, cvt, "\n");
			cvt->linepos=0;
			mlOutChartAxisType(ge,cvt,&chart->x_axis[ii]);
			mlOutIndent(ge, cvt);
			cprintf(ge, cvt,"%s %s\n", tok[TEND], tok[TX_AXIS]);
			cvt->linepos=0;
		}
	}
	if ( chart->n_y_axes )
	{
		for ( ii=0; ii<chart->n_y_axes; ii++ )
		{
			mlOutIndent(ge, cvt);
			cprintf(ge, cvt,"%s ", tok[TY_AXIS]);
			if ( !chart->y_axis[ii].label_str )
				cprintf(ge, cvt,"\"\"");
			else
			{
				len = MultiToWideStr(ge, chart->y_axis[ii].label_str, tbuff );
				mlOutString(ge,  len, tbuff, cvt, FALSE );
			}
			cprintf(ge, cvt, "\n");
			cvt->linepos=0;
			mlOutChartAxisType(ge,cvt,&chart->y_axis[ii]);
			mlOutIndent(ge, cvt);
			cprintf(ge, cvt,"%s %s\n", tok[TEND], tok[TY_AXIS]);
			cvt->linepos=0;
		}
	}
	if ( chart->n_z_axes )
	{
		for ( ii=0; ii<chart->n_z_axes; ii++ )
		{
			mlOutIndent(ge, cvt);
			cprintf(ge, cvt,"%s ", tok[TZ_AXIS]);
			if ( !chart->z_axis[ii].label_str )
				cprintf(ge, cvt,"\"\"");
			else
			{
				len = MultiToWideStr(ge, chart->z_axis[ii].label_str, tbuff );
				mlOutString(ge,  len, tbuff, cvt, FALSE );
			}
			cprintf(ge, cvt, "\n");
			cvt->linepos=0;
			mlOutChartAxisType(ge,cvt,&chart->z_axis[ii]);
			mlOutIndent(ge, cvt);
			cprintf(ge, cvt,"%s %s\n", tok[TEND], tok[TZ_AXIS]);
			cvt->linepos=0;
		}
	}

	/* old_data type and related info such as old_ix*/
	for ( ii=0; ii<chart->old_n_groups; ii++ )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt, "%s <%d %d %d %d>\n", tok[TOLD_GRP], 
		        chart->old_grp[ii].n,
		        chart->old_grp[ii].type,
		        chart->old_grp[ii].stack_id,
		        chart->old_grp[ii].stack_ax);
	}
	cvt->linepos = 0;

	/* group data */
	for (ii=0; ii<chart->n_groups; ii++)
		mlOutChartDataGroupType(ge,cvt,&chart->grp[ii]);

	/* legend box */
	mlOutChartLegendType(ge,cvt,&chart->legend);

	/* write the default attributes */
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n", tok[TDEFATR]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &chart->def_attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &chart->def_attr, 0, cvt);
		cvt->indent--;
	}

	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n", tok[TMAJHRZ]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &chart->maj_x_attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &chart->maj_x_attr, 0, cvt);
		cvt->indent--;
	}

	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n", tok[TMINHRZ]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &chart->min_x_attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &chart->min_x_attr, 0, cvt);
		cvt->indent--;
	}

	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n", tok[TMAJVRT]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &chart->maj_y_attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &chart->maj_y_attr, 0, cvt);
		cvt->indent--;
	}

	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n", tok[TMINVRT]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &chart->min_y_attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &chart->min_y_attr, 0, cvt);
		cvt->indent--;
	}
	
	if ( chart->three_d.enabled && chart->n_z_axes > 0 )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s\n", tok[TMAJDPT]);
		cvt->linepos = 0;
		if ( AttrCmp(ge, &chart->maj_z_attr, &cvt->curr_atts) )
		{
			cvt->indent++;
			mlOutAttrs(ge, &chart->maj_z_attr, 0, cvt);
			cvt->indent--;
		}

		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s\n", tok[TMINDPT]);
		cvt->linepos = 0;
		if ( AttrCmp(ge, &chart->min_z_attr, &cvt->curr_atts) )
		{
			cvt->indent++;
			mlOutAttrs(ge, &chart->min_z_attr, 0, cvt);
			cvt->indent--;
		}
	}
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n", tok[TTITLEATR]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &chart->title_attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &chart->title_attr, 0, cvt);
		cvt->indent--;
	}

	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n", tok[TSUBTITLEATR]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &chart->subtitle_attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &chart->subtitle_attr, 0, cvt);
		cvt->indent--;
	}

	mlOutIndent(ge, cvt);
	cprintf(ge, cvt,"%s\n", tok[TFOOTATR]);
	cvt->linepos = 0;
	if ( AttrCmp(ge, &chart->footer_attr, &cvt->curr_atts) )
	{
		cvt->indent++;
		mlOutAttrs(ge, &chart->footer_attr, 0, cvt);
		cvt->indent--;
	}

	/* done */
	cprintf(ge, cvt,"%s %s\n\n", tok[TEND], tok[TCHART]);
	cvt->linepos = 0;
}

/****************************************************************************
 *
 * module:	C o n v e r t T e x t X f o r m
 */
static thing *
ConvertTextXform(ge, chart, t)
graphics_main_type *ge;
chart_type *chart;
thing *t;
{
	thing *child_t, *str_t, *new_t;
	vector extent;
	int i, size, x, y, max_wid, max_hyt, line_hyt;
	char *str;

	new_t = (thing *) 0;
	if ( !t )
		return((thing *) new_t);
	if ( t->type == GROUP )	
	{
		for ( i=0; i<t->data.group->num; i++ )
		{	
			new_t = (thing *) ConvertTextXform(ge, chart,
											   t->data.group->list[i]);
			if ( new_t )
			{	
				removeFromGroup(ge, t, i);
				tAddToGroup(ge, t, new_t, i);
				new_t = (thing *) 0;
			}
		}
	}
	else if ( t->type == TEXTBOX )
	{
		if ( t->n_paths == 0 && t->data.txtbox->trans.nPoints == 7 && 
			 t->attr.field_angle == 0 )
		{	
			/* create a bounder */
			t->n_paths = 1;
			size = sizeof(poly_type);
			t->paths = (poly_type*)gloc(ge, size);
			t->paths[0].nCtrls = 0;
			t->paths[0].controls = (TmPoint32 *)0;
			t->paths[0].nPoints = 5;
			size = 5 * sizeof(TmPoint32);
			t->paths[0].points = (TmPoint32 *)gloc(ge, size);
			
			line_hyt = t->attr.field_space;
			t->attr.l_margin = t->attr.r_margin = line_hyt/2;
			t->attr.t_margin = line_hyt;
			t->attr.b_margin = line_hyt/4;
			
			t->paths[0].points[0].x = t->data.txtbox->trans.points[1].x +
				t->data.txtbox->x_off - t->attr.l_margin;
			t->paths[0].points[0].y = t->data.txtbox->trans.points[1].y +
				t->data.txtbox->y_off - t->attr.t_margin;
			
			t->paths[0].points[1].x = t->data.txtbox->trans.points[2].x -
				t->data.txtbox->x_off + t->attr.r_margin + 30;
			t->paths[0].points[1].y = t->data.txtbox->trans.points[2].y +
				t->data.txtbox->y_off - t->attr.t_margin;
			
			t->paths[0].points[2].x = t->data.txtbox->trans.points[3].x -
				t->data.txtbox->x_off + t->attr.r_margin + 30;
			t->paths[0].points[2].y = t->data.txtbox->trans.points[3].y +
				t->attr.b_margin;
			
			t->paths[0].points[3].x = t->data.txtbox->trans.points[4].x +
				t->data.txtbox->x_off - t->attr.l_margin;
			t->paths[0].points[3].y = t->data.txtbox->trans.points[4].y +
				t->attr.b_margin;
			
			t->paths[0].points[4].x = t->paths[0].points[0].x;
			t->paths[0].points[4].y = t->paths[0].points[0].y;
		}
	}
	return((thing *)new_t);		
}

/****************************************************************************
 *
 * module:	m l C o n v e r t C h a r t T e x t X f o r m
 */
static void
mlConvertChartTextXform(ge, cvt, chart)
graphics_main_type *ge;
cvt_global_type *cvt;
chart_type *chart;
{
	thing *t, *child_t;
	int i;
	
	t = chart->t;
	if( !t )
		return;
	if ( t->type == GROUP )	for ( i=0; i<t->data.group->num; i++ )
	{
		child_t = t->data.group->list[i];
		if ( !child_t )
			continue;
		
		/* draw the part */
		ConvertTextXform(ge, chart, child_t);
	}
}

/****************************************************************************
 *
 * module:	m l O u t C h a r t
 */
static void
mlOutChart(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	chart_type *ge_chart, *chart;
	int i;

	ge_chart = (chart_type *)ge->chart;
	if ( !cvt->as_1_1 ) for ( i=0; i<ge->n_charts; i++ )
	{
		chart = (chart_type *)&ge_chart[i];
		if ( chart )
		{
			mlConvertChartTextXform(ge, cvt, chart);
			mlOutChartData(ge, cvt, chart);
		}
		cvt->indent--;
	}
}

/****************************************************************************
 *
 * module:	m l O u t O b j C o m m e n t
 */
static void
mlOutObjComment(ge, cvt, t)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
{
	int ii, needs_tag;
	ge_tag_type *tag;

	/* quick exit */
	if ( !t )
		return;

	/* Put any tags (comments) associated with an object */
	needs_tag = TRUE;
	if ( t->attr.use.tag )
	{
		tag = &ge->tag[t->attr.tag];
		if ( tag ) for ( ii=0; ii<tag->n_strs; ii++ )
		{
			/* indent */
			mlOutIndent(ge, cvt);
			/* comment */
			cvt->linepos = cprintf(ge, cvt, "#%s", tag->str[ii]);
#if (FOR_FILTER && (!TURBO)) /* rather helpful */
			if ( !strncmp(tag->str[ii], "#id", 3) )
				needs_tag = FALSE;
#endif
		}
	}

#if 0 /* rather helpful */
	if ( needs_tag )
	{
		mlOutIndent(ge, cvt);
		cvt->linepos = cprintf(ge, cvt, "#id %d", t->id);
	}
#endif
}

/****************************************************************************
 *
 * module:	m l O u t P a l e t
 */
static void
mlOutPalet(ge,cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	udo_type udoP;
	int ii;

	/* no parts in old rev saves (FIX!) */
	if ( cvt->as_1_1 )
		return;

	/* for each UDO */
	for ( ii=0; ii<ge->n_udos; ii++ )
	{
		/* write any comments */
		mlOutObjComment(ge, cvt, ge->udo[ii].t);

		/* write the part token and the part name */
		if ( cvt->linepos != 0 )
			mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s ", tok[TPART]);
		mlOutString(ge, ge->udo[ii].n_chars, ge->udo[ii].name, cvt, FALSE);

		/* write the object associated with the part */
		cprintf(ge, cvt, " ");
		mlOutPart(ge, cvt, ge->udo[ii].t, TRUE);
	}
	if ( ge->n_udos > 0 )
		cprintf(ge, cvt, "\n");
}

/****************************************************************************
 *
 * module:	m l O u t A u d i o I n f o
 */
static void
mlOutAudioInfo(ge, cvt, page)
graphics_main_type *ge;
cvt_global_type *cvt;
ge_page_type *page;
{
	if ( cvt->as_1_1 )
		return;

	/* default setting */
	if ( page->audio.play == FALSE &&
		 page->audio.type == INSET_NONE &&
		 page->audio.id   == 0 &&
	     page->audio.volume == 50 )
		return;

	cprintf(ge, cvt,"%s\n", tok[TAUDIO1]);
	cprintf(ge, cvt," <%d %d %d %d>\n", page->audio.play,
		                                page->audio.type,
		                                page->audio.id,
		                                page->audio.volume);
	cprintf(ge, cvt,"%s %s\n", tok[TEND], tok[TAUDIO1]);
}

/****************************************************************************
 *
 * module:	m l O u t P i c t
 */
static void
mlOutPict(ge, cvt, page)
graphics_main_type *ge;
cvt_global_type *cvt;
ge_page_type *page;
{
	thing *t;
	int i;

	/* note the start of the picture segment */
	cprintf(ge, cvt, "%s\n", tok[TPICTURE]);
	t = &page->t;
	if ( t && t->type != 0 && (t->type != GROUP || t->data.group->num > 0) )
	{
		if ( cvt->as_1_1 )
			cvt->curr_atts.face = cvt->curr_atts.size = INITIAL;

		/* write out all the parts in the document */
		if ( t->type != GROUP )
		{
			mlOutObjComment(ge, cvt, t);
			mlOutPart(ge, cvt, t, FALSE);
		}

		else for ( i=0; i<t->data.group->num; i++ )
		{
			mlOutObjComment(ge, cvt, t->data.group->list[i]);
			mlOutPart(ge, cvt, t->data.group->list[i], FALSE);
		}
	}

	/* note the end of the picture segment */
	cprintf(ge, cvt,"%s %s\n", tok[TEND], tok[TPICTURE]);
}

/****************************************************************************
 *
 * module:	m l O u t N o t e s
 */
static void
mlOutNotes(ge, cvt, page)
graphics_main_type *ge;
cvt_global_type *cvt;
ge_page_type *page;
{
	/* notes object */
	if ( page->notes_t )
	{
		cprintf(ge, cvt, "%s\n", tok[TNOTES]);
		cvt->indent++;
		mlOutPart(ge, cvt, page->notes_t, FALSE);
		cvt->indent--;
		cprintf(ge, cvt, "%s %s\n", tok[TEND], tok[TNOTES]);
	}
}

/****************************************************************************
 *
 * module:
 */
static void
mlOutPageHdrFtrInfo(ge, cvt, page)
graphics_main_type *ge;
cvt_global_type *cvt;
ge_page_type *page;
{
	ge_hdrftr_page_type *hf_page;

	if ( page->hf_page )
	{
		/* set defaults */
		memset(&cvt->hfstr,0,sizeof(ge_hdrftr_str_type));
		cvt->hfstr.pointsize = 120000;
		cvt->hfstr.color_ix = 1;
		hf_page = page->hf_page;

		/* write out the header */
		cprintf(ge, cvt,"%s", tok[TPHDRFTR]);
		cprintf(ge, cvt,"\n");

		/* write out the basic page hdr/ftr info */
		if ( hf_page->hdr_offset || hf_page->hdr_n_lines ||
		     hf_page->ftr_offset || hf_page->ftr_n_lines )
		{
			cprintf(ge, cvt," %s\n", tok[TBASE]);
			cvt->hdr_ftr = hf_page;
			if (hf_page->hdr_offset || hf_page->hdr_n_lines)
				mlOutHdrPage(ge,cvt);
			if (hf_page->ftr_offset || hf_page->ftr_n_lines)
				mlOutFtrPage(ge,cvt);
			cprintf(ge, cvt," %s %s\n", tok[TEND], tok[TBASE]);
		}
		cprintf(ge, cvt,"%s %s\n\n", tok[TEND], tok[TPHDRFTR]);

		/* make sure hfstr attrs are in sync with curr_atts */
		if ( cvt->hfstr.face_ix != cvt->curr_atts.face )
			cvt->curr_atts.face = cvt->hfstr.face_ix;
		if ( cvt->hfstr.pointsize != cvt->curr_atts.size )
			cvt->curr_atts.size = cvt->hfstr.pointsize;
		if ( cvt->hfstr.color_ix != cvt->curr_atts.fill_fg )
			cvt->curr_atts.fill_fg = cvt->hfstr.color_ix;
		if ( cvt->hfstr.bold != cvt->curr_atts.weight )
			cvt->curr_atts.weight = cvt->hfstr.bold;
		if ( cvt->hfstr.italic != cvt->curr_atts.slant )
			cvt->curr_atts.slant = cvt->hfstr.italic;
		if ( cvt->hfstr.strike_thru != cvt->curr_atts.strike )
			cvt->curr_atts.strike = cvt->hfstr.strike_thru;
		if ( cvt->hfstr.underline != cvt->curr_atts.ruling )
			cvt->curr_atts.ruling = cvt->hfstr.underline;
	}
}

/****************************************************************************
 *
 * module:	m l O u t P a g e s
 */
static void
mlOutPages(ge, cvt)
graphics_main_type *ge;
cvt_global_type *cvt;
{
	int i;
	for ( i=0; i<ge->n_pages; i++ )
	{
		/* write the hdrftrs for this page */
		mlOutPageHdrFtrInfo(ge, cvt, &ge->page[i]);

		/* write the slide info for this page */
		mlOutSlideInfo(ge, cvt, &ge->page[i]);

		/* write the page's layers */
		mlOutLayers(ge, cvt, &ge->page[i]);

		/* write the page's audio */
		mlOutAudioInfo(ge, cvt, &ge->page[i]);

		/* write the page's paint list */
		mlOutPict(ge, cvt, &ge->page[i]);

		/* write the speaker notes for this page */
		mlOutNotes(ge, cvt, &ge->page[i]);

		/* done */
		cprintf(ge, cvt, "\n");
	}
}

/****************************************************************************
 *
 * module:	m l O u t E n d e r
 */
static void
mlOutEnder(ge, cvt, header_offset)
graphics_main_type *ge;
cvt_global_type *cvt;
int header_offset;
{
	int encrypted, n_bytes, ix;

	if (!cvt->as_1_1)
	{
		encrypted = cvt->info.encrypted;
		AGetDefaultFileInfo(&cvt->info);
		cvt->info.docType = AX_DOC_TYPE_ASTERX_GRAPHICS;
		cvt->info.encrypted = encrypted;
		AFileEnderFromInfo(&cvt->info,cvt->buf);
		cprintf(ge, cvt,"%s\n",cvt->buf);
		if ( encrypted ) /* pad out to 64 bytes from header end */
		{
			n_bytes = cvt->n_bytes - header_offset;
			ix = n_bytes % 64;
			if ( ix != 0 )
			{
				n_bytes = 64 - ix;
				for ( ix=1; ix<n_bytes; ix++ )
					cprintf(ge, cvt, " ");
				cprintf(ge, cvt, "\n");
			}
		}
	}
}

/****************************************************************************
 *
 * module:	m l O u t O b j H d r - Write the object's header to file
 */
static void
mlOutCObjHdr(ge, cvt, t, no_indent)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
int no_indent;
{
	int x1, y1, x2, y2;

	/* write the object type */
	if ( no_indent == FALSE )
		mlOutIndent(ge, cvt);
	if ( t->type == TEXTBOX )
		cprintf(ge, cvt,"%s", tok[TCTXT]);
	else
		cprintf(ge, cvt,"%s", tok[TCRECT]);

	/* write the object's position */
	if ( t->extent.x1 <= t->extent.x2 )
	{
		x1 = t->extent.x1 - PAGE_ZEROX;
		x2 = t->extent.x2 - PAGE_ZEROX;
	}
	else
	{
		x2 = t->extent.x1 - PAGE_ZEROX;
		x1 = t->extent.x2 - PAGE_ZEROX;
	}
	if ( t->extent.y1 <= t->extent.y2 )
	{
		y1 = t->extent.y1 - PAGE_ZEROY;
		y2 = t->extent.y2 - PAGE_ZEROY;
	}
	else
	{
		y2 = t->extent.y1 - PAGE_ZEROY;
		y1 = t->extent.y2 - PAGE_ZEROY;
	}
	cprintf(ge, cvt, " %s (%d,%d)\n", tok[TCAT], x1, y1);

	/* update the current pos */
	cvt->dx = x1;
	cvt->dy = y1;
	
	/* write any persistant marks */
	cvt->linepos = 0;
}

/****************************************************************************
 *
 * module:	m l O u t O b j H d r - Write the object's header to file
 */
static void
mlOutObjHdr(ge, cvt, t, no_indent)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
int no_indent;
{
	int x1, y1, x2, y2;

	/* write the object type */
	if ( no_indent == FALSE )
		mlOutIndent(ge, cvt);
	if ( cvt->as_1_1 )
	{
		if ( t->type == ELL &&
		     (t->data.ell->n_sides > 0 || (!isRectForm(ge, t->paths[0].points))) )
			cprintf(ge, cvt,"%s", pnm2[POLY]);
		else if (cvt->as_1_1 && (t->type == TEXTBOX))
			cprintf(ge, cvt,"%s", pnm2[TEXT]);
		else
			cprintf(ge, cvt,"%s", pnm2[t->type]);
	}
	else
	{
		if ( t->type == ELL && t->data.ell->n_sides > 0 )
			cprintf(ge, cvt,"%s", pnm2[RPOL]);
		else
			cprintf(ge, cvt,"%s", pnm2[t->type]);
	}

	/* write the object's position */
	if ( t->extent.x1 <= t->extent.x2 )
	{
		x1 = t->extent.x1 - PAGE_ZEROX;
		x2 = t->extent.x2 - PAGE_ZEROX;
	}
	else
	{
		x2 = t->extent.x1 - PAGE_ZEROX;
		x1 = t->extent.x2 - PAGE_ZEROX;
	}
	if ( t->extent.y1 <= t->extent.y2 )
	{
		y1 = t->extent.y1 - PAGE_ZEROY;
		y2 = t->extent.y2 - PAGE_ZEROY;
	}
	else
	{
		y2 = t->extent.y1 - PAGE_ZEROY;
		y1 = t->extent.y2 - PAGE_ZEROY;
	}
	if ( !cvt->as_1_1 )
	{
		/* all non groups need a position */
		if ( t->type != GROUP )
			cprintf(ge, cvt, " %s (%d,%d)\n", tok[TAT], x1, y1);

		/* groups write any special info (chart, blend, extrusion, ...) */
		else if ( t->chart_ix )
			cprintf(ge, cvt, " %s <1 %d>\n", tok[TGRP_INFO], t->chart_ix-1);

		/* a group with no special info - just a return */
		else
			cprintf(ge, cvt, "\n");

		/* update the current pos */
		cvt->dx = x1;
		cvt->dy = y1;
	}
	else
	{
		cvt->indent++;
		mlOutIndent(ge, cvt);
		cvt->indent--;
		if ( t->type != LINE )
		{
			cprintf(ge, cvt,"%s (%d,%d) %s (%d,%d)\n",
			        tok[TORIGIN], x1, y1,
			        tok[TCORNER], x2, y2);
		}
		else
		{
			cprintf(ge, cvt,"%s (%d,%d) %s (%d,%d)\n",
			        tok[TORIGIN],	
			        t->paths[0].points[0].x-PAGE_ZEROX,
			        t->paths[0].points[0].y-PAGE_ZEROY,
			        tok[TCORNER],	
			        t->paths[0].points[1].x-PAGE_ZEROX,
			        t->paths[0].points[1].y-PAGE_ZEROY);
		}
	}
	
	/* write any persistant marks */
	cvt->linepos = 0;
	cvt->indent++;
	if ( t->attr.callback > 0 )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s %d\n", tok[TCBACK], t->attr.callback-1);
	}
	if ( t->flags.hidden )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s %s\n", tok[THIDDEN], tst[1]);
	}
	if ( t->flags.locked )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s %s\n", tok[TLOCKED], tst[1]);
	}
	if ( t->flags.fixed )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s %s\n", tok[TFIXED], tst[1]);
	}
	if ( t->flags.recolored )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s %s\n", tok[TRECOLOR], tst[1]);
	}
	if ( t->flags.templated )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s %s\n", tok[TTEMPLATED], tst[1]);
	}
	if ( t->flags.top_attr )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s %s\n", tok[TTOP_ATTR], tst[1]);
	}
	if ( t->flags.isa_title )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s %s\n", tok[TISA_TITLE], tst[1]);
	}
	if ( t->flags.isa_subtitle )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s %s\n", tok[TISA_SUBTITLE], tst[1]);
	}
	if ( t->flags.isa_footer )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s %s\n", tok[TISA_FOOTER], tst[1]);
	}
	if ( t->flags.isa_body )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s %s\n", tok[TISA_BODY], tst[1]);
	}
	if ( t->flags.isa_legend )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s %s\n", tok[TISA_LEGEND], tst[1]);
	}
	if ( t->flags.isa_axis )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s %s\n", tok[TISA_AXIS], tst[1]);
	}
	if ( t->flags.isa_axlab )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s %s\n", tok[TISA_AXLAB], tst[1]);
	}
	if ( t->flags.isa_datgrp )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s %s\n", tok[TISA_DATGRP], tst[1]);
	}
	if ( t->flags.isa_datlab )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s %s\n", tok[TISA_DATLAB], tst[1]);
	}
	if ( t->flags.isa_grid )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s %s\n", tok[TISA_GRID], tst[1]);
	}
	if ( t->t_class != 0 )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s %d\n", tok[TCLASS], t->t_class);		
	}	
}


/****************************************************************************
 *
 * module: m l O u t P n t s - write any Point infomation
 */
static void
mlOutPnts(ge, cvt, pathP, type_trans)
graphics_main_type *ge;
cvt_global_type *cvt;
poly_type *pathP;
int type_trans;
{
	int ii, x, y;

	/* get count of coord pairs and write them to file */
	if ( pathP->nPoints > 0 )
	{
		/* header */
		if ( cvt->as_1_1 )
			cvt->linepos = 0;
		else
		{
			cvt->linepos = mlOutIndent(ge, cvt);
			if ( type_trans )
				cvt->linepos += cprintf(ge, cvt,"%s ", tok[TT_PNTS]);
			else
				cvt->linepos += cprintf(ge, cvt,"%s ", tok[TPNTS]);
		}

		/* point list */
		for ( ii=0; ii<pathP->nPoints; ii++ )
		{
			x = pathP->points[ii].x - PAGE_ZEROX - cvt->dx;
			y = pathP->points[ii].y - PAGE_ZEROY - cvt->dy;
			if ( MAX_LINE < cvt->linepos )
			{
				cvt->linepos = 0;
				if ( type_trans )
				{
					cprintf(ge, cvt, "\n       ");
					cvt->linepos = 7 + mlOutIndent(ge, cvt);
				}
				else
				{
					cprintf(ge, cvt, "\n     ");
					cvt->linepos = 5 + mlOutIndent(ge, cvt);
				}
			}
			cvt->linepos += cprintf(ge, cvt,"(%d,%d)", x, y);
		}
		if ( cvt->linepos )
		{
			cputc(ge, cvt,'\n');
			cvt->linepos = 0;
		}
	}
}

/****************************************************************************
 *
 * module: m l O u t C t l s - write any control points
 */
static void
mlOutCtls(ge, cvt, pathP, type_trans)
graphics_main_type *ge;
cvt_global_type *cvt;
poly_type *pathP;
int type_trans;
{
	int ii, x, y;

	if ( cvt->as_1_1 )
		return;

	if ( pathP->nCtrls > 0 )
	{
		/* header */
		cvt->linepos = mlOutIndent(ge, cvt);
		if ( type_trans )
			cvt->linepos += cprintf(ge, cvt,"%s ", tok[TT_CTLS]);
		else
			cvt->linepos += cprintf(ge, cvt,"%s ", tok[TCTLS]);

		/* point list */
		for ( ii=0; ii<pathP->nCtrls; ii++ )
		{
			x = pathP->controls[ii].x - PAGE_ZEROX - cvt->dx;
			y = pathP->controls[ii].y - PAGE_ZEROY - cvt->dy;
			if ( MAX_LINE < cvt->linepos )
			{
				cvt->linepos = 0;
				if ( type_trans )
				{
					cprintf(ge, cvt, "\n       ");
					cvt->linepos = 7 + mlOutIndent(ge, cvt);
				}
				else
				{
					cprintf(ge, cvt, "\n     ");
					cvt->linepos = 5 + mlOutIndent(ge, cvt);
				}
			}
			cvt->linepos += cprintf(ge, cvt,"(%d,%d)", x, y);
		}
		if ( cvt->linepos )
		{
			cputc(ge, cvt,'\n');
			cvt->linepos = 0;
		}
	}
}

/****************************************************************************
 *
 * module: m l O u t T r a n s
 */
static void
mlOutTrans(ge, trans, cvt)
graphics_main_type *ge;
poly_type *trans;
cvt_global_type *cvt;
{
	mlOutPnts(ge, cvt, trans, TRUE);
	mlOutCtls(ge, cvt, trans, TRUE);
}

/****************************************************************************
 *
 * module: m l O u t P a t h
 */
static void
mlOutPath(ge, pathP, nPaths, cvt)
graphics_main_type *ge;
poly_type *pathP;
int nPaths;
cvt_global_type *cvt;
{
	int ii;
	for ( ii=0; ii<nPaths; ii++ )
	{
		mlOutPnts(ge, cvt, &pathP[ii], FALSE);
		mlOutCtls(ge, cvt, &pathP[ii], FALSE);
	}
}

/****************************************************************************
 *
 * module:	m l O u t C l i p M a s k
 */
static void
mlOutClipPath(ge, cvt, t, cinfo)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
clip_path *cinfo;
{
	if ( cvt->as_1_1 || cinfo == (clip_path *)0 )
		return;

	cprintf(ge,cvt,"%s <(%d,%d)(%d,%d)(%d,%d)(%d,%d)(%d,%d)>\n", tok[TCLIPPATH],
			cinfo->paths[0].points[0].x - PAGE_ZEROX,
			cinfo->paths[0].points[0].y - PAGE_ZEROY,
			cinfo->paths[0].points[1].x - PAGE_ZEROX,
			cinfo->paths[0].points[1].y - PAGE_ZEROY,
			cinfo->paths[0].points[2].x - PAGE_ZEROX,
			cinfo->paths[0].points[2].y - PAGE_ZEROY,
			cinfo->paths[0].points[3].x - PAGE_ZEROX,
			cinfo->paths[0].points[3].y - PAGE_ZEROY,
			cinfo->paths[0].points[4].x - PAGE_ZEROX,
			cinfo->paths[0].points[4].y - PAGE_ZEROY);
}

/****************************************************************************
 *
 * module: m l O u t P o l y
 */
static void
mlOutPoly(ge,t,cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	poly_type path;
	int npts;				/* number of points converted poly */
	TmPoint32 *newPointsP=NULL;	/* vertices of converted poly */

	if ( !cvt->as_1_1 )
		mlOutPath(ge, t->paths, t->n_paths,cvt);
	else
	{
		/* convert to a simple PATH */
		convertPath(ge, t->paths, &path);
		convertPols(1, &path, &npts, &newPointsP, FALSE);		
		gfree(ge, path.points);
		path.points = newPointsP;
		path.nPoints = npts;
		path.nCtrls = 0;
		mlOutPath(ge,&path,1,cvt);
		if (path.points)
			gfree(ge, path.points);
		if (path.controls)
			gfree(ge, path.controls);
	}
}

/****************************************************************************
 *
 * module:  m l O u t T e x t S t r i n g
 */
static void
mlOutTextString(ge,cvt,t)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
{
	mlOutIndent(ge, cvt);
	if ( t->data.str->n_chars > 0 )
		mlOutString(ge, t->data.str->n_chars, t->data.str->chars, cvt, FALSE);
	else
	{
		cputc(ge, cvt,'"');
		cputc(ge, cvt,'"');
	}
	cputc(ge, cvt,'\n');
	cvt->linepos = 0;
}


/****************************************************************************
 *
 * module:  m l O u t T e x t  - write the string(s)
 */
static void
mlOutCText(ge,t,cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	int ii;
	thing *txt_t;

	/* Write out the actual strings */
	for ( ii=0; ii<t->data.txtbox->n_strings; ii++ )
	{
		/* write the TEXT token */
		mlOutIndent(ge, cvt);
		txt_t = t->data.txtbox->strings[ii];
		cprintf(ge, cvt,"%s\n", tok[TCSTR]);
		cvt->linepos = 0;

		/* write the string */
		cvt->indent++;
		mlOutCAttrs(ge, &txt_t->attr, cvt, TRUE);
		mlOutTextString(ge, cvt, txt_t);
		cvt->indent--;
	}
}

/****************************************************************************
 *
 * module:
 */
static void
mlOutChildTextBox(ge,t,cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	thing_attributes attr;
	int i;

	/* don't write empty text strings */
	if ( t->n_paths <= 0 && t->data.txtbox->n_strings <= 0 )
		return;
	mlOutCObjHdr(ge, cvt, t, FALSE);

	/* text attrs */
	memcpy(&attr, &cvt->curr_atts, sizeof(thing_attributes));
	mlOutCAttrs(ge, &t->attr, cvt, TRUE);
	if (0 != t->data.txtbox->x_off || 0 != t->data.txtbox->y_off)
	{
		cvt_attr_indent(ge, cvt);
		cprintf(ge, cvt,"%s <%d %d>", tok[TCTXTXYOFF],
					t->data.txtbox->x_off, t->data.txtbox->y_off);
		mlOutNewline(ge, cvt);
	}

	/* draw the bounding polygon(s) */
	if ( 0 != t->data.txtbox->field_wid )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s %d\n", tok[TCH_SPACE], t->data.txtbox->field_wid);
	}

	/* draw the bounding polygon(s) */
	if ( 0 != t->data.txtbox->v_align )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s <%d>\n", tok[TV_ALIGN], t->data.txtbox->v_align);
	}
	/* subsitute text ? */
	if ( t->data.txtbox->sub_txt )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt,"%s <%d>\n", tok[TSUBTXT], 1);
	}

	if ( t->data.txtbox->n_tabs > 0 )
	{
		mlOutIndent(ge, cvt);
		cprintf(ge, cvt, "%s <%d %d", tok[TTABSTOPS],
									t->data.txtbox->def_tabsize,
									t->data.txtbox->n_tabs);
		for ( i=0; i<t->data.txtbox->n_tabs; i++ )
		{
			if ( i%10 == 0 && i != 0 )
			{
				cprintf(ge, cvt, "\n");
				mlOutIndent(ge, cvt);
				cprintf(ge, cvt, "         ");
			}
			cprintf(ge, cvt, " %d", t->data.txtbox->tabstops[i]);
		}
		cprintf(ge, cvt, ">\n");
	}
	/* the actual text */
	mlOutCText(ge,t,cvt);
	cvt->linepos = 0;
	memcpy(&cvt->curr_atts, &attr, sizeof(thing_attributes));
}

/****************************************************************************
 *
 * module:
 */
static void
mlOutChildRect(ge,t,cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	thing_attributes attr;

	mlOutCObjHdr(ge, cvt, t, FALSE);

	/* attrs */
	memcpy(&attr, &cvt->curr_atts, sizeof(thing_attributes));
	mlOutCAttrs(ge, &t->attr, cvt, FALSE);

	/* the path */
	mlOutPath(ge, t->paths, t->n_paths, cvt);
	memcpy(&cvt->curr_atts, &attr, sizeof(thing_attributes));
	
}

/****************************************************************************
 *
 * module: m l O u t E l l
 */
static void
mlOutEll(ge,t,cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	poly_type path;
	TmPoint32 pnts[16], ctls[16], *points, *controls;
	int sz, ii;
	thing newt;

	cvt->linepos = 0;
	if ( cvt->as_1_1 && t->data.ell->n_sides > 0 )
	{
		/* convert the regular poly to a path */
		getEllPath(ge, t,&path);
		mlOutPath(ge,&path,1,cvt);
		gfree(ge, path.points);
		return;
	}

	if ( cvt->as_1_1 && !isRectForm(ge, t->paths[0].points) )
	{
		path.points   = pnts;
		path.controls = ctls;

		ellToPath(ge, t->paths->points,
			t->data.ell->start_angle,
			t->data.ell->end_angle,
			FALSE, &path);

		/* change the ell path to a poly path */
		newt.paths = (poly_type*)gloc(ge, sizeof(poly_type));
		sz = (path.nPoints/2 + 1) * sizeof(TmPoint32);
		newt.paths->points = (TmPoint32*)gloc(ge, sz);
		sz = path.nCtrls * sizeof(TmPoint32);
		newt.paths->controls = (TmPoint32*)gloc(ge, sz);;
		newt.paths->nPoints = path.nPoints/2;
		newt.paths->nCtrls = path.nCtrls;
		for (ii=0; ii<newt.paths->nCtrls; ii++ )
			newt.paths->controls[ii] = ctls[ii];
		for (ii=0; ii<newt.paths->nPoints; ii++ )
			newt.paths->points[ii] = pnts[ii*2];

		/* close the path */
		newt.paths->points[ii] = pnts[0];
		newt.paths->nPoints++;

		mlOutPoly(ge,&newt,cvt);
		gfree(ge, newt.paths->points);
		gfree(ge, newt.paths->controls);
		gfree(ge, newt.paths);
		return;
	}

	/* some specific attributes */
	if ( t->data.ell->start_angle != cvt->startAngle )
	{
		cvt_attr_indent(ge, cvt);
		cvt->startAngle = t->data.ell->start_angle;
		cvt->linepos += cprintf(ge, cvt,"%s %d",tok[TSTART_ANGLE],cvt->startAngle);
		mlOutNewline(ge, cvt);
	}
	if ( t->data.ell->n_sides == 0 && t->data.ell->end_angle != cvt->endAngle )
	{
		cvt_attr_indent(ge, cvt);
		cvt->endAngle = t->data.ell->end_angle;
		cvt->linepos += cprintf(ge, cvt,"%s %d", tok[TEND_ANGLE], cvt->endAngle);
		mlOutNewline(ge, cvt);
	}
	if ( t->data.ell->n_sides > 0 && !cvt->as_1_1 )
	{
		cvt_attr_indent(ge, cvt);
		cvt->nsides = t->data.ell->n_sides;
		cvt->linepos += cprintf(ge, cvt,"%s %d", tok[TNSIDES], cvt->nsides);
		mlOutNewline(ge, cvt);
	}

	if ( t->data.ell->sshape_type != cvt->sshape_type ||
	     t->data.ell->sshape_pnt[0].x != cvt->sshape_pnt[0].x ||
	     t->data.ell->sshape_pnt[0].y != cvt->sshape_pnt[0].y ||
	     t->data.ell->sshape_pnt[1].x != cvt->sshape_pnt[1].x ||
	     t->data.ell->sshape_pnt[1].y != cvt->sshape_pnt[1].y ||
	     t->data.ell->sshape_pnt[2].x != cvt->sshape_pnt[2].x ||
	     t->data.ell->sshape_pnt[2].y != cvt->sshape_pnt[2].y ||
	     t->data.ell->sshape_pnt[3].x != cvt->sshape_pnt[3].x ||
	     t->data.ell->sshape_pnt[3].y != cvt->sshape_pnt[3].y ||
	     t->data.ell->sshape_pnt[4].x != cvt->sshape_pnt[4].x ||
	     t->data.ell->sshape_pnt[4].y != cvt->sshape_pnt[4].y ||
	     t->data.ell->sshape_pnt[5].x != cvt->sshape_pnt[5].x ||
	     t->data.ell->sshape_pnt[5].y != cvt->sshape_pnt[5].y )
	{
		cvt->sshape_type     = t->data.ell->sshape_type;
		cvt->sshape_pnt[0].x = t->data.ell->sshape_pnt[0].x;
		cvt->sshape_pnt[0].y = t->data.ell->sshape_pnt[0].y;
		cvt->sshape_pnt[1].x = t->data.ell->sshape_pnt[1].x;
		cvt->sshape_pnt[1].y = t->data.ell->sshape_pnt[1].y;
		cvt->sshape_pnt[2].x = t->data.ell->sshape_pnt[2].x;
		cvt->sshape_pnt[2].y = t->data.ell->sshape_pnt[2].y;
		cvt->sshape_pnt[3].x = t->data.ell->sshape_pnt[3].x;
		cvt->sshape_pnt[3].y = t->data.ell->sshape_pnt[3].y;
		cvt->sshape_pnt[4].x = t->data.ell->sshape_pnt[4].x;
		cvt->sshape_pnt[4].y = t->data.ell->sshape_pnt[4].y;
		cvt->sshape_pnt[5].x = t->data.ell->sshape_pnt[5].x;
		cvt->sshape_pnt[5].y = t->data.ell->sshape_pnt[5].y;

		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge,cvt,"%s <%d %d %d %d %d %d %d %d %d %d %d %d %d>",
		                        tok[TSSHAPE],
								cvt->sshape_type,
		                        cvt->sshape_pnt[0].x,
		                        cvt->sshape_pnt[0].y,
		                        cvt->sshape_pnt[1].x,
		                        cvt->sshape_pnt[1].y,
		                        cvt->sshape_pnt[2].x,
		                        cvt->sshape_pnt[2].y,
		                        cvt->sshape_pnt[3].x,
		                        cvt->sshape_pnt[3].y,
		                        cvt->sshape_pnt[4].x,
		                        cvt->sshape_pnt[4].y,
		                        cvt->sshape_pnt[5].x,
		                        cvt->sshape_pnt[5].y);
		mlOutNewline(ge, cvt);
	}
	if ( !cvt->as_1_1 )
		mlOutPath(ge, t->paths, t->n_paths,cvt);

	if ( t->data.ell->txt_t )
		mlOutChildTextBox(ge, t->data.ell->txt_t, cvt);
	
}
static void
mlOutRPoly(ge,t,cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	mlOutEll(ge,t,cvt);
}

/****************************************************************************
 *
 * module: m l O u t I n s e t
 */
static void
mlOutInset(ge,t,cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	tx_char tstr[STRPATH+1];
	int i, type, ix, len;

	type = t->data.inset->type;
	ix = t->data.inset->ix;
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt, "%s <%d %d %d %d>\n", tok[TINS_AREA],
	        t->data.inset->area.x1 - PAGE_ZEROX,
	        t->data.inset->area.y1 - PAGE_ZEROY,
	        t->data.inset->area.x2 - PAGE_ZEROX,
	        t->data.inset->area.y2 - PAGE_ZEROY);
	mlOutIndent(ge, cvt);
	cprintf(ge, cvt, "%s <%d %d", tok[TINS_REF], type, ix);
	if ( type == INSET_LINKED )
	{
		cprintf(ge, cvt, " %d %d %d",
		        t->data.inset->unsaved,
		        ge->link[ix].rt_inset.type,
		        ge->link[ix].rt_inset.icon_only);
		if ( ge->link[ix].rt_inset.exec == (char *)0 )
			cprintf(ge, cvt," \"\"");
		else
		{
			len = strlen(ge->link[ix].rt_inset.exec);
			if ( len <= 0 )
				cprintf(ge, cvt," \"\"");
			else
			{
				cprintf(ge, cvt, " ");
				for ( i=0; i<len; i++ )
					tstr[i] = (tx_char)ge->link[ix].rt_inset.exec[i];
				tstr[i] = (tx_char)0;
				mlOutString(ge, len, tstr, cvt, FALSE);
			}
		}
		if ( ge->link[ix].rt_inset.filt == (char *)0 )
			cprintf(ge, cvt," \"\"");
		else
		{
			len = strlen(ge->link[ix].rt_inset.filt);
			if ( len <= 0 )
				cprintf(ge, cvt," \"\"");
			else
			{
				cprintf(ge, cvt, " ");
				for ( i=0; i<len; i++ )
					tstr[i] = (tx_char)ge->link[ix].rt_inset.filt[i];
				tstr[i] = (tx_char)0;
				mlOutString(ge, len, tstr, cvt, FALSE);
			}
		}
	}
	else if ( type == INSET_EMBEDED )
	{
		cprintf(ge, cvt, " %d %d %d",
		        t->data.inset->unsaved,
		        ge->embed[ix].rt_inset.type,
		        ge->embed[ix].rt_inset.icon_only);
		if ( ge->embed[ix].rt_inset.exec == (char *)0 )
			cprintf(ge, cvt," \"\"");
		else
		{
			len = strlen(ge->embed[ix].rt_inset.exec);
			if ( len <= 0 )
				cprintf(ge, cvt," \"\"");
			else
			{
				cprintf(ge, cvt, " ");
				for ( i=0; i<len; i++ )
					tstr[i] = (tx_char)ge->embed[ix].rt_inset.exec[i];
				tstr[i] = (tx_char)0;
				mlOutString(ge, len, tstr, cvt, FALSE);
			}
		}
		if ( ge->embed[ix].rt_inset.filt == (char *)0 )
			cprintf(ge, cvt," \"\"");
		else
		{
			len = strlen(ge->embed[ix].rt_inset.filt);
			if ( len <= 0 )
				cprintf(ge, cvt," \"\"");
			else
			{
				cprintf(ge, cvt, " ");
				for ( i=0; i<len; i++ )
					tstr[i] = (tx_char)ge->embed[ix].rt_inset.filt[i];
				tstr[i] = (tx_char)0;
				mlOutString(ge, len, tstr, cvt, FALSE);
			}
		}
	}
	else
		cprintf(ge, cvt," \"\" \"\"");
	cprintf(ge, cvt, " %d %d %d %d %d",
	        t->data.inset->scale_mode,
	        t->data.inset->proportional,
	        t->data.inset->centered,
	        t->data.inset->x_zoom,
	        t->data.inset->y_zoom);
#if FOR_FILTER
	cprintf(ge, cvt, " \"\"");
#else
{
	extern char *AxTaskElfDataToInsetStr();
	char *str, *data;

	if ( type == INSET_LINKED )
		data = ge->link[ix].rt_inset.data;
	else if ( type == INSET_EMBEDED )
		data = ge->embed[ix].rt_inset.data;
	else
		data = (char *)0;
	if ( !data )
		cprintf(ge, cvt, " \"\"");
	else
	{
		str = AxTaskElfDataToInsetStr(ge->pid, data);
		len = strlen(str);
		cprintf(ge, cvt, " ");
		for ( i=0; i<len; i++ )
			tstr[i] = (tx_char)str[i];
		tstr[i] = (tx_char)0;
		mlOutString(ge, len, tstr, cvt, FALSE);
	}
}
#endif
	cprintf(ge, cvt, ">\n");
}

/****************************************************************************
 *
 * module: m l O u t T e m p
 */
static void
mlOutTemp(ge,t,cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	/* the template type */
	cvt_attr_indent(ge, cvt);
	cvt->temp_type = t->data.temp->type;
	cvt->temp_size = t->data.temp->size;
	cvt->temp_inactive = t->data.temp->inactive;
	cvt->temp_chart_ix = t->data.temp->ix;
	cvt->linepos += cprintf(ge, cvt,"%s <%d %d %d %d>\n", tok[TTEMPLATE_FLAGS],
	                        cvt->temp_type,
	                        cvt->temp_size,
	                        cvt->temp_inactive,
					    cvt->temp_chart_ix);

	/* the template area */
	cvt->linepos = 0;
	cvt_attr_indent(ge, cvt);
	cvt->temp_area = t->data.temp->area;
	cvt->linepos += cprintf(ge, cvt,"%s <%d %d %d %d>\n", tok[TT_POS],
	                        cvt->temp_area.x1 - PAGE_ZEROX,
	                        cvt->temp_area.y1 - PAGE_ZEROY,
	                        cvt->temp_area.x2 - PAGE_ZEROX,
	                        cvt->temp_area.y2 - PAGE_ZEROY);

	/* the template build info */
	if ( t->data.temp->build )
	{
		cvt->linepos = 0;
		cvt_attr_indent(ge, cvt);
		mlOutBuildInfo(ge, cvt, t->data.temp->build,
						   t->data.temp->enable_build);
	}

	/* the background rect */
	if ( t->data.temp->crect_t )
		mlOutChildRect(ge, t->data.temp->crect_t, cvt);

	/* the templated object (must be last!) */
	cvt->linepos = 0;
	mlOutPart(ge, cvt, t->data.temp->child_t, FALSE);

	
}


/****************************************************************************
 *
 * module: m l O u t R e c t
 */
static void
mlOutRect(ge,t,cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	if ( cvt->as_1_1 )
		return;

	if ((t->data.rect->x_corner_radius != cvt->x_corner_radius) ||
		(t->data.rect->y_corner_radius != cvt->y_corner_radius))
	{
		cvt_attr_indent(ge, cvt);
		cvt->x_corner_radius = t->data.rect->x_corner_radius;
		cvt->y_corner_radius = t->data.rect->y_corner_radius;
		cprintf(ge, cvt,"%s <%d %d>", tok[TXYRAD],
		        cvt->x_corner_radius,
		        cvt->y_corner_radius);
		mlOutNewline(ge, cvt);
	}
	if (t->data.rect->scale_corners != cvt->scale_corners)
	{
		cvt_attr_indent(ge, cvt);
		cvt->scale_corners = t->data.rect->scale_corners;
		cvt->linepos += cprintf(ge, cvt,"%s %d", tok[TSCALE_CORNERS],
		                        cvt->scale_corners);
		mlOutNewline(ge, cvt);
	}

	if ((t->data.rect->x_corner_numerator != cvt->x_corner_numerator) ||
		(t->data.rect->y_corner_numerator != cvt->y_corner_numerator))
	{
		cvt_attr_indent(ge, cvt);
		cvt->x_corner_numerator = t->data.rect->x_corner_numerator;
		cvt->y_corner_numerator = t->data.rect->y_corner_numerator;
		cvt->linepos += cprintf(ge, cvt,"%s <%d %d>", tok[TPOSITION],
		                        cvt->x_corner_numerator,
		                        cvt->y_corner_numerator);
		mlOutNewline(ge, cvt);
	}

	if ( t->data.rect->sshape_type != cvt->sshape_type ||
	     t->data.rect->sshape_pnt[0].x != cvt->sshape_pnt[0].x ||
	     t->data.rect->sshape_pnt[0].y != cvt->sshape_pnt[0].y ||
	     t->data.rect->sshape_pnt[1].x != cvt->sshape_pnt[1].x ||
	     t->data.rect->sshape_pnt[1].y != cvt->sshape_pnt[1].y ||
	     t->data.rect->sshape_pnt[2].x != cvt->sshape_pnt[2].x ||
	     t->data.rect->sshape_pnt[2].y != cvt->sshape_pnt[2].y ||
	     t->data.rect->sshape_pnt[3].x != cvt->sshape_pnt[3].x ||
	     t->data.rect->sshape_pnt[3].y != cvt->sshape_pnt[3].y ||
	     t->data.rect->sshape_pnt[4].x != cvt->sshape_pnt[4].x ||
	     t->data.rect->sshape_pnt[4].y != cvt->sshape_pnt[4].y ||
	     t->data.rect->sshape_pnt[5].x != cvt->sshape_pnt[5].x ||
	     t->data.rect->sshape_pnt[5].y != cvt->sshape_pnt[5].y )
	{
		cvt->sshape_type     = t->data.rect->sshape_type;
		cvt->sshape_pnt[0].x = t->data.rect->sshape_pnt[0].x;
		cvt->sshape_pnt[0].y = t->data.rect->sshape_pnt[0].y;
		cvt->sshape_pnt[1].x = t->data.rect->sshape_pnt[1].x;
		cvt->sshape_pnt[1].y = t->data.rect->sshape_pnt[1].y;
		cvt->sshape_pnt[2].x = t->data.rect->sshape_pnt[2].x;
		cvt->sshape_pnt[2].y = t->data.rect->sshape_pnt[2].y;
		cvt->sshape_pnt[3].x = t->data.rect->sshape_pnt[3].x;
		cvt->sshape_pnt[3].y = t->data.rect->sshape_pnt[3].y;
		cvt->sshape_pnt[4].x = t->data.rect->sshape_pnt[4].x;
		cvt->sshape_pnt[4].y = t->data.rect->sshape_pnt[4].y;
		cvt->sshape_pnt[5].x = t->data.rect->sshape_pnt[5].x;
		cvt->sshape_pnt[5].y = t->data.rect->sshape_pnt[5].y;

		cvt_attr_indent(ge, cvt);
		cvt->linepos += cprintf(ge, cvt,"%s <%d %d %d %d %d %d %d %d %d %d %d %d %d>",
		                        tok[TSSHAPE],
								cvt->sshape_type,
		                        cvt->sshape_pnt[0].x,
		                        cvt->sshape_pnt[0].y,
		                        cvt->sshape_pnt[1].x,
		                        cvt->sshape_pnt[1].y,
		                        cvt->sshape_pnt[2].x,
		                        cvt->sshape_pnt[2].y,
		                        cvt->sshape_pnt[3].x,
		                        cvt->sshape_pnt[3].y,
		                        cvt->sshape_pnt[4].x,
		                        cvt->sshape_pnt[4].y,
		                        cvt->sshape_pnt[5].x,
		                        cvt->sshape_pnt[5].y);
		mlOutNewline(ge, cvt);
	}
	mlOutPath(ge, t->paths, t->n_paths,cvt);
	if ( t->data.rect->txt_t )
		mlOutChildTextBox(ge, t->data.rect->txt_t, cvt);
}

/****************************************************************************
 *
 * module: m l O u t L i n e
 */
static void
mlOutLine(ge,t,cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	
	if (!cvt->as_1_1)
	{
		if ( t->data.line->sshape_type != cvt->lshape_type ||
			 t->data.line->end1_cpnt != cvt->end1_cpnt ||
			 t->data.line->end2_cpnt != cvt->end2_cpnt ||
			 t->data.line->end1_cpnt_dir != cvt->end1_cpnt_dir ||
			 t->data.line->end2_cpnt_dir != cvt->end2_cpnt_dir ||
			 t->data.line->sshape_pnt[0].x != cvt->lshape_pnt[0].x ||
			 t->data.line->sshape_pnt[0].y != cvt->lshape_pnt[0].y ||
			 t->data.line->sshape_pnt[1].x != cvt->lshape_pnt[1].x ||
			 t->data.line->sshape_pnt[1].y != cvt->lshape_pnt[1].y ||
			 t->data.line->sshape_pnt[2].x != cvt->lshape_pnt[2].x ||
			 t->data.line->sshape_pnt[2].y != cvt->lshape_pnt[2].y )
		{
#if 0
			cvt->lshape_type     = t->data.line->sshape_type;
			cvt->end1_cpnt       = t->data.line->end1_cpnt;
			cvt->end2_cpnt       = t->data.line->end2_cpnt;
			cvt->end1_cpnt_dir   = t->data.line->end1_cpnt_dir;
			cvt->end2_cpnt_dir   = t->data.line->end2_cpnt_dir;
			cvt->lshape_pnt[0].x = t->data.line->sshape_pnt[0].x;
			cvt->lshape_pnt[0].y = t->data.line->sshape_pnt[0].y;
			cvt->lshape_pnt[1].x = t->data.line->sshape_pnt[1].x;
			cvt->lshape_pnt[1].y = t->data.line->sshape_pnt[1].y;
			cvt->lshape_pnt[2].x = t->data.line->sshape_pnt[2].x;
			cvt->lshape_pnt[2].y = t->data.line->sshape_pnt[2].y;
			cvt->lshape_pnt[3].x = t->data.line->sshape_pnt[3].x;
			cvt->lshape_pnt[3].y = t->data.line->sshape_pnt[3].y;
			cvt->lshape_pnt[4].x = t->data.line->sshape_pnt[4].x;
			cvt->lshape_pnt[4].y = t->data.line->sshape_pnt[4].y;

			cvt->linepos += cprintf(ge, cvt,"%s <%d %d %d %d %d %d %d %d %d %d %d>",
									tok[TLSHAPE],
									cvt->lshape_type,
									cvt->end1_cpnt,
									cvt->end2_cpnt,
									cvt->end1_cpnt_dir,
									cvt->end2_cpnt_dir,
									cvt->lshape_pnt[0].x, 
									cvt->lshape_pnt[0].y,
									cvt->lshape_pnt[1].x, 
									cvt->lshape_pnt[1].y,
									cvt->lshape_pnt[2].x, 
									cvt->lshape_pnt[2].y);
			mlOutNewline(ge, cvt);
#endif
		}
		mlOutPath(ge, t->paths, t->n_paths,cvt);
	}
}

/****************************************************************************
 *
 * module: m l O u t S t r o k e
 */
static void
mlOutStroke(ge,t,cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	mlOutPoly(ge,t,cvt);
}

/****************************************************************************
 *
 * module:  m l O u t T e x t 2 1  - write the string(s) Rev 2.1 style
 */
static void
mlOutText21(ge,cvt,t)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
{
	poly_type path;
	thing *txt_t;
	tx_char *strs;
	int ii,nchars;

	if (t->data.txtbox->trans.nPoints > 0)
	{
		/* Text object. convert to old style */
		/* ref. cvti.c : mlTrans5ToTrans7()  */
		/* make as 1 string using first attrs */

		/* attributes here */

		/* get the total */
		for (ii=nchars=0; ii<t->data.txtbox->n_strings; ii++)
		{
			txt_t = t->data.txtbox->strings[ii];
			nchars += txt_t->data.str->n_chars;
		}

		/* cat the strings */
		strs = (tx_char*)gloc(ge, (nchars+1)*sizeof(tx_char));
		memset(strs, 0x0, (nchars+1)*sizeof(tx_char));
		for (ii=0; ii<t->data.txtbox->n_strings; ii++)
		{
			txt_t = t->data.txtbox->strings[ii];
			tx_strncat(strs,txt_t->data.str->chars,txt_t->data.str->n_chars);
		}
		mlOutString(ge, nchars,strs,cvt, FALSE);
		cputc(ge, cvt,'\n');

		path.nPoints = 5;
		path.points = (TmPoint32 *)gloc(ge, 5*sizeof(TmPoint32));
		path.controls = (TmPoint32 *)0;
		path.points[0] = t->data.txtbox->trans.points[1];
		path.points[1] = t->data.txtbox->trans.points[2];
		path.points[2] = t->data.txtbox->trans.points[3];
		path.points[3] = t->data.txtbox->trans.points[4];
		path.points[4] = t->data.txtbox->trans.points[5];
		mlOutPath(ge,&path,1,cvt);
		gfree(ge, path.points);
		gfree(ge, strs);
		return;
	}
}

/****************************************************************************
 *
 * module:  m l O u t T e x t  - write the string(s)
 */
static void
mlOutText(ge,t,cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	int ii;
	thing *txt_t;

	/* Write out the actual strings */
	for ( ii=0; ii<t->data.txtbox->n_strings; ii++ )
	{
		/* write the TEXT token */
		mlOutIndent(ge, cvt);
		txt_t = t->data.txtbox->strings[ii];
		cprintf(ge, cvt,"%s\n", pnm2[TEXT]);
		cvt->linepos = 0;

		/* write the string */
		cvt->indent++;
		mlOutAttrs(ge, &txt_t->attr, 0, cvt);
		mlOutPath(ge, txt_t->paths, txt_t->n_paths, cvt);
		mlOutTextString(ge, cvt,txt_t);
		cvt->indent--;
	}
}

/****************************************************************************
 *
 * module:	m l O u t B o x T e x t -  TEXTBOX
 */
static void
mlOutTextBox(ge,t,cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	int i;

	/* don't write empty text strings */
	if ( t->n_paths <= 0 && t->data.txtbox->n_strings <= 0 )
		return;
	if ( cvt->as_1_1 )
		mlOutText21(ge,cvt,t);
	else
	{
		/* text attrs */
		mlOutAttrs(ge, &t->attr, 0, cvt);
		mlOutLayer(ge, cvt, t->layer);
		if ((cvt->x_off != t->data.txtbox->x_off) ||
			(cvt->y_off != t->data.txtbox->y_off))
		{
			cvt_attr_indent(ge, cvt);
			cprintf(ge, cvt,"%s <%d %d>", tok[TTXTXYOFF],
				cvt->x_off = t->data.txtbox->x_off,
				cvt->y_off = t->data.txtbox->y_off);
			mlOutNewline(ge, cvt);
		}

		/* draw the bounding polygon(s) */
		mlOutPath(ge, t->paths, t->n_paths, cvt);
		mlOutTrans(ge, &t->data.txtbox->trans, cvt);
		if ( cvt->field_wid  != t->data.txtbox->field_wid )
		{
			mlOutIndent(ge, cvt);
			cprintf(ge, cvt,"%s %d\n", tok[TH_SPACE], t->data.txtbox->field_wid);
			cvt->field_wid = t->data.txtbox->field_wid;
		}
		if ( t->data.txtbox->n_tabs > 0 )
		{
			mlOutIndent(ge, cvt);
			cprintf(ge, cvt, "%s <%d %d", tok[TTABSTOPS],
			                            t->data.txtbox->def_tabsize,
			                            t->data.txtbox->n_tabs);
			for ( i=0; i<t->data.txtbox->n_tabs; i++ )
			{
				if ( i%10 == 0 && i != 0 )
				{
					cprintf(ge, cvt, "\n");
					mlOutIndent(ge, cvt);
					cprintf(ge, cvt, "         ");
				}
				cprintf(ge, cvt, " %d", t->data.txtbox->tabstops[i]);
			}
			cprintf(ge, cvt, ">\n");
		}

		/* the actual text */
		mlOutText(ge,t,cvt);
	}
}

/************************************************************************
 *
 * module:	m l M a p T o 1 2 5
 */
static void
mlMapTo125(ge,cvt, src_t, dst_t)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *src_t, *dst_t;
{
	unsigned char *srcP, *dstP, *outCmap;
	int size, ii, cmap_ix;
	unsigned int r,g,b;
	cmap_type *srcCmap;

	if ( src_t->type == IMAGE && dst_t->type == IMAGE )
	{
		srcP = src_t->data.img->info.data;
		dstP = dst_t->data.img->info.data;
		cmap_ix = src_t->data.img->cmapID;
		srcCmap = &ge->cmap[cmap_ix];

		/* start by making a map from the input colormap */
		outCmap = (unsigned char*)gloc(ge, srcCmap->n_colors);
		AxSet125PixVal(NULL, FALSE, 0, 0, 0, 16, TRUE);
		for (ii=srcCmap->n_colors; ii>=0; --ii)
		{
			if (srcCmap->color[ii].see_thru)
				outCmap[ii] = 124; /* set to white */
			else
			{
				ExtractRGBCmapEntry(ge, cvt, cmap_ix, ii, &r, &g, &b);
				AxSet125PixVal(outCmap+ii,FALSE,r,g,b,0,FALSE);
			}
		}
		size = src_t->data.img->info.hyt * src_t->data.img->info.data_byte_wid;
		for ( ii=size; ii>=0; --ii)
			dstP[ii] = outCmap[srcP[ii]];
		gfree(ge, outCmap);
	}
}

/************************************************************************
 *
 * module:	m l O u t I m a g e 2 1
 */
static void
mlOutImage21(ge, cvt, t)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
{
	char *epsP;
	int llen, (*compressFunc)();
	thing copyThing, *outThing;

	cprintf(ge, cvt," %s ",tok[TOUTLINE]);
	mlOutPath(ge, &t->data.img->trans, 1, cvt);

	/* if there is an EPS image, write it out. */
	if (t->data.img->info.eps_seg != NULL)
	{
		/* the width, height, size and a leading space */
		cprintf(ge, cvt," %s %d %d %d\n ", tok[TEPS],
			t->data.img->info.wid,
			t->data.img->info.hyt,
			t->data.img->info.eps_seg_size);

		llen = t->data.img->info.eps_seg_size;
		epsP = t->data.img->info.eps_seg;
		while ( *epsP != '\0')	
		{
			cputc(ge, cvt, *epsP);
			if (*epsP == '\n')
				cputc(ge, cvt, ' ');
			epsP++;
		}

		/* the end of the EPS segment */
		cprintf(ge, cvt,"%s %s\n",tok[TEND], tok[TEPS]);
	}

	cprintf(ge, cvt," FORMAT 1");
	cprintf(ge, cvt," %s %d", tok[TWIDTH], t->data.img->info.wid);
	cprintf(ge, cvt," %s %d", tok[THEIGHT], t->data.img->info.hyt);
	cprintf(ge, cvt," %s %d\n", tok[TDEPTH], t->data.img->info.dep);

	cprintf(ge, cvt," %s %s", tok[TCOMPRESSION],
		(t->data.img->info.dep == 8) ? cmp[COMP_RAW-1] : cmp[COMP_RUN-1]);
	cprintf(ge, cvt," %s %s", tok[TENCODING], enc[BIT6]);
	
	compressFunc = (t->data.img->info.dep == 8) ? WriteRAW : WriteRUN;

	llen = cprintf(ge, cvt,"\n %s ", tok[TDATA]);

	/* color images need to be remapped to 125 colors */
	if (t->data.img->info.dep == 8)
	{
		copyThing = *t;
		copyThing.data.img->info.data =
			(unsigned char*)gloc(ge, t->data.img->info.hyt*
			t->data.img->info.data_byte_wid);
		mlMapTo125(ge,cvt,t, &copyThing);
	}

	/* DO IT! */
	outThing = (t->data.img->info.dep == 8) ? &copyThing : t;
	(*compressFunc)(ge, cvt,llen,outThing->data.img->info.hyt,
		outThing->data.img->info.data_byte_wid,
		outThing->data.img->info.data,mailOutBit6, FALSE);

	if (t->data.img->info.dep == 8)
		gfree(ge, copyThing.data.img->info.data);

	cputc(ge, cvt,'\n');
	cputc(ge, cvt,'\n');
	return;
}

/************************************************************************
 *
 * module:	m l O u t I m a g e - ascii-ize an image
 */
static void
mlOutImage(ge, t, cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	char *epsP;
	int llen, ii, (*encodeFunc)(), (*compressFunc)();

	if (cvt->as_1_1)
	{
		mlOutImage21(ge, cvt, t);
		return;
	}

	mlOutPath(ge, t->paths, t->n_paths, cvt);
	mlOutTrans(ge, &t->data.img->trans, cvt);

	/* if there is an EPS image, write it out. */
	if (t->data.img->info.eps_seg != NULL)
	{
		/* the width, height, size and a leading space */
		cprintf(ge, cvt," %s %d %d %d\n ", tok[TEPS],
			t->data.img->info.wid,
			t->data.img->info.hyt,
			t->data.img->info.eps_seg_size);

		llen = t->data.img->info.eps_seg_size;
		epsP = t->data.img->info.eps_seg;
		while ( *epsP != '\0')
		{
			cputc(ge, cvt, *epsP);
			if (*epsP == '\n')
				cputc(ge, cvt, ' ');
			epsP++;
		}

		/* the end of the EPS segment */
		cprintf(ge, cvt,"%s %s\n",tok[TEND], tok[TEPS]);
	}

	cprintf(ge, cvt," %s %d", tok[TWIDTH], t->data.img->info.wid);
	cprintf(ge, cvt," %s %d", tok[THEIGHT], t->data.img->info.hyt);
	cprintf(ge, cvt," %s %d", tok[TDEPTH], t->data.img->info.dep);


	if ( cvt->compression != COMP_LZW)
		cprintf(ge, cvt," %s %s", tok[TCOMPRESSION], cmp[cvt->compression-1]);

	if ((cvt->encoding != BIT6) || cvt->as_1_1)
		cprintf(ge, cvt," %s %s", tok[TENCODING], enc[cvt->encoding]);
	
	encodeFunc = (cvt->encoding == BIT6) ? mailOutBit6 : mailOutHex;

	if (cvt->compression == COMP_RAW)
		compressFunc = WriteRAW;
	else if (cvt->compression == COMP_RUN)
		compressFunc = WriteRUN;
	else
		compressFunc = ReadWriteLZW;

	llen = cprintf(ge, cvt,"\n %s ", tok[TDATA]);

	/* DO IT! */
	(*compressFunc)(ge, cvt,llen,t->data.img->info.hyt,
		t->data.img->info.data_byte_wid,
		t->data.img->info.data,encodeFunc, FALSE);
	if ((t->data.img->info.dep == 8) && (t->data.img->info.mask))
	{
		llen = cprintf(ge, cvt,"\n %s ", tok[TMASK]);
		(*compressFunc)(ge, cvt,llen,t->data.img->info.hyt,
			t->data.img->info.mask_byte_wid,
			t->data.img->info.mask,encodeFunc,FALSE);
	}

	cputc(ge, cvt,'\n');
	return;
}

/****************************************************************************
 *
 * module:	m l O u t G r o u p - export a Grouped object
 */
static void
mlOutGroup(ge, t, cvt)
graphics_main_type *ge;
thing *t;
cvt_global_type *cvt;
{
	int ii;

	/* write the children */
	for ( ii=0; ii<t->data.group->num; ii++ )
	{
		mlOutObjComment(ge, cvt, t->data.group->list[ii]);
		mlOutPart(ge, cvt, t->data.group->list[ii], FALSE);
	}

	/* terminate the group */
	cvt->indent--;
	mlOutIndent(ge, cvt);
	cvt->indent++;
	cprintf(ge, cvt, "%s %s\n", tok[TEND], pnm2[GROUP]);
}

/****************************************************************************
 *
 * module:	m l O u t P a r t - top-level part 
 */
static void
mlOutPart(ge, cvt, t, no_indent)
graphics_main_type *ge;
cvt_global_type *cvt;
thing *t;
int no_indent;
{
	/* null objects became valid in rev3.0 */
	if ( !t )
	{
		if ( !cvt->as_1_1 )
		{
			if ( no_indent == FALSE )
				mlOutIndent(ge, cvt);
			cprintf(ge, cvt, "%s\n", pnm2[0]);
		}
	}
	else if ( t->type >= TEXTBOX && t->type <= TEMPLATE )
	{
		/* write the type, position, and any persistant marks */
		mlOutObjHdr(ge, cvt, t, no_indent);

		/* write out the attributes */
		if ( t->type == IMAGE && t->data.img->info.dep == 8 )
			mlOutAttrs(ge, &t->attr, t->data.img->cmapID, cvt);
		else if ( t->type != TEMPLATE )
			mlOutAttrs(ge, &t->attr, 0, cvt);
		mlOutLayer(ge, cvt, t->layer);

		/* write the type specific info */
		switch ( t->type )
		{
			case ELL:        mlOutEll    (ge,t,cvt); break;
			case RECT:       mlOutRect   (ge,t,cvt); break;
			case LINE:       mlOutLine   (ge,t,cvt); break;
			case STROK:      mlOutStroke (ge,t,cvt); break;
			case POLY:       mlOutPoly   (ge,t,cvt); break;
			case TEXTBOX:    mlOutTextBox(ge,t,cvt); break;
			case IMAGE:      mlOutImage  (ge,t,cvt); break;
			case GROUP:      mlOutGroup  (ge,t,cvt); break;
			case RPOL:       mlOutRPoly  (ge,t,cvt); break;
			case INSET:      mlOutInset  (ge,t,cvt); break;
			case TEMPLATE:   mlOutTemp   (ge,t,cvt); break;
		}
		mlOutClipPath(ge, cvt, t, t->clip);

		/* restore the indentation level (bumped in mlOutObjHdr) */
		cvt->indent--;
		cvt->dx = 0;
		cvt->dy = 0;
	}
}

/****************************************************************************
 *
 * module:	g r W r i t e
 */
int
grWrite(ge, cvt, outBuffer, outCount)
graphics_main_type *ge;
cvt_global_type *cvt;
char **outBuffer;
int *outCount;
{
	int offset;

	/* init */
	mlDumpTxtObjs(ge, ge->t);
	if ( outBuffer )
	{
		cvt->buffit = TRUE;
		cvt->outBuff = cvt->outBufPtr = (char*)gloc(ge, CVTBLKSIZE);
		cvt->outBufEndPtr = cvt->outBufPtr + CVTBLKSIZE;
	}
	else
	{
		cvt->fp = (FILE *)cvtGetFP(ge, cvt->fileName, "w+");
		if ( cvt->fp == (FILE *)0 )
			return(SGE_FILE_CREATE);
	}
	InitTokens(ge);

	/* generic file header (*START GRAPHICS...) */
	offset = mlOutHeader(ge, cvt);

	/* self contained segments */
	mlOutColormap(ge, cvt);
	mlOutFaces(ge, cvt);
	mlOutCallbacks(ge, cvt);

	/* fills may reference colors */
	mlOutNamedFills(ge, cvt);

	/* grads may reference colors */
	mlOutNamedGrads(ge, cvt);

	/* styles may reference colors, faces or callbacks */
	mlOutStyles(ge, cvt);

	/* headers/footers may reference colors and faces */
	mlOutHdrsFtrs(ge, cvt);

	/* default state may reference styles, colors, faces or callbacks */
	mlOutSessionInfo(ge, cvt);

	/* presentation stuff may reference styles, colors, faces or callbacks*/
	mlOutTemplateInfo(ge, cvt);

	/* charts may reference styles, colors, faces or callbacks */
	mlOutChart(ge, cvt);

	/* palet parts may reference styles, colors, faces or callbacks */
	mlOutPalet(ge, cvt);

	/* paint list may reference styles, colors, faces or callbacks */
	mlOutPages(ge, cvt);

	/* finish up with *END GRAPHICS */
	mlOutEnder(ge, cvt, offset);

	/* done */
	if ( !outBuffer )
		cvtCloseFile(ge, cvt);
	else
	{
		*outBuffer = cvt->outBuff;
		*outCount = cvt->outBufPtr - cvt->outBuff;
	}
	return(AOK);
}
