/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axtext.c
   Applix ASC version: 421
   Date: Wed Mar 24 21:14:29 1999
*/
#include <stdio.h>
#include <math.h>
#include <limits.h>
#include <aos.h>
#include <getstr5.h>
#include <atm.h>
#include <atmwin.h>
#include <aspeller.h>
#include <axtext.h>
#include <atxint.h>
#include <axhtab.h>
#include <axgfs.h>
#include <wchars.h>
#include <xim.h>
#include <ge.h>
#include <eldata.h>

#if !SCOUNIX
/* workaround for prototyper problem  on sco with stat, lstat, statlstat */
/* this needs to be fixed correctly                                      */
#include <axtext.p>
#endif

#if AXMS
#ifndef M_PI
#define  M_PI            3.14159265358979323846
#endif
#define hypot _hypot
extern Window RootWin;    /* for DefaultVisual macro on NT */
extern int Running_Windows95;
 
#define GRADIENT_FILL(fill) (fill == RECT_FILL_TYPE || \
                            fill == LINEAR_FILL_TYPE || \
                            fill == RADIAL_FILL_TYPE || \
                            fill == BILINEAR_FILL_TYPE)
#endif

#define BYTE_WIDTH(w)       ((((w) + 15) >> 3) & ~1)

#define WHITE 0
#define BLACK 1

    /* from atmattr.h */
#define TMS       3
#define TMS_X_D   6
#define TMS_O_D   7
#define TMIS_O_D 13

#define SPACECHAR (' ') /* %%%%%%% should really use txAttr->spaceChar */
#define WHITESPACE(x) (x == SPACECHAR)
#define HYPHEN ('-')
#define TABCHAR ('\t')
#define NEWLINE ('\n')
#define CR (0x000D)
#define NEWCHAR(c) ((c) == NEWLINE || (c) == CR)
#define WHITECHAR(c) ((c) == SPACECHAR || \
                      (c) == TABCHAR   || \
                      (c) == NEWLINE   || \
                      (c) == CR)
#define BLACKCHAR(c) (!(WHITECHAR(c)))
#define NONBREAKSPACE   (SPACECHAR + 0x80)  /* font convention */   
#define WORDBREAK(c) ((c < 0x30 && c != 0x2d ) || \
                      (c > 0x7b && c < 0x80)   || \
                      (c > 0x39 && c < 0x41)   || \
                      (c > 0x5a && c < 0x61 && c != 0x5f))

#define CHAR_IS_TABCHAR(cc) ( cc == TABCHAR )
#define CHAR_IS_NEWLINE(cc) ( cc == NEWLINE )
#define CHAR_IS_CR(cc) ( cc == CR )

#define NEWPARA(pc, ix) ((ix == 0) || CHAR_IS_NEWLINE((int)pc[ix].charcode))

#define CHAR_IS_DIMENSIONLESS(cc) (CHAR_IS_NEWLINE(cc) || \
                                   CHAR_IS_TABCHAR(cc) || \
                                   CHAR_IS_CR(cc))

/* BIG_MILS specifies a number considered too big for some formatting calcs.
 * any real text objects exceeding this size will probably cause trouble.
 */
#define BIG_MILS 1020000     /* 85 feet, in mils */


#define IS_SOLID_FILL(gcontext, fill) ( !fill || easy_text_fill(gcontext, (attr_type *)fill) )

#define TXtransformAtts (TXxscale | TXyscale |\
                         TXxshear | TXangle)

extern Display *Dpy;
extern Window AxTopWinNeverMapped;
extern int AxtLetterformPixmaps;
extern int TmAmSE; /* true if printing */
extern unsigned long AlisPixels[];
extern int TargetRes, ScreenRes, TargetRes;
extern short ColorDisplay;
extern GC Gc1;
extern unsigned long PlaneMask;
extern unsigned long FGpixel, BGpixel;

GC axTextGC, axTextStencilGC, axLiftGC;
GC PixClearGC, PixGC, axTextCursorGC, axInactiveTextCursorGC;

/* these tables hold emToMils vals for a variety of point sizes */
#define MTABLESIZE 1000
static int axtMilTable10_0[MTABLESIZE];     /* 10 pt @ 100% */
static int axtMilTable11_0[MTABLESIZE];     /* 11 pt @ 100% */
static int axtMilTable12_0[MTABLESIZE];     /* 12 pt @ 100%; 10 pt @ 120% */
static int axtMilTable13_2[MTABLESIZE];     /* 11 pt @ 120% (GS favorite!) */
static int axtMilTable14_0[MTABLESIZE];     /* 14 pt @ 100% */
static int axtMilTable14_4[MTABLESIZE];     /* 12 pt @ 120% (my favorite!) */

/* Given an integer in units of 10000 points, return a pointer to the
    proper conversion table (as above), or NULL if no table for that size */
/* Search in the 'most likely' order for best efficiency */
#define MTABLEPTR(p)    ( (p) == 144000 ? &axtMilTable14_4[0] : \
                            ((p) == 120000 ? &axtMilTable12_0[0] : \
                            ((p) == 132000 ? &axtMilTable13_2[0] : \
                            ((p) == 110000 ? &axtMilTable11_0[0] : \
                            ((p) == 140000 ? &axtMilTable14_0[0] : \
                            ((p) == 100000 ? &axtMilTable10_0[0] : NULL))))) )


static unsigned char msbfBitMask[] =
    {0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01};
static unsigned char lsbfBitMask[] =
    {0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80};

extern int qRound(), mToPix(), mToMils();
extern int initFonts();
extern void initFontCache(), freeErrString();
extern void unionRectangle();
#if FOR_FILTER
extern void cvtConvMilsToPix(), cvtConvPixToMils();
#else
extern void convMilsToPix(), convPixToMils();
#endif
extern char *getErrString();
extern char *axGetFontRecord(), *axGetFontFamily();
extern char *axGetFontRecordForScreen();
extern char *axGetLetterformImage(), *axGetScreenMetrics();
extern char *axGetLetterformPixmap();
extern char *axGetTargetMetrics(), *axGetCacheItem();
extern char *axGetCharacterCode(), *axGetFontRecordName();
extern char *axMakeCompoundWinName();
extern int axGetCache();
extern char *XLT();
extern char *getErrString(), *axtXLFDencodingName();
extern char *axtDirectScreenfont();
extern GC axCreateGC();
extern XCharStruct *getCharStruct();

#if BUILD_ORIENT
mbchar *sansserif_orient_face();
mbchar *serif_orient_face();
mbchar *serif_narrow_kana_face();
mbchar *sansserif_narrow_kana_face();
#endif

void forwardSXRT(), reverseSXRT();
void daxtext();
void liftCursor();
void freeTextAttrs();
void freeTextAttr();
void axtWriteTargetCtm();
void axDrawText(), axtFormScalableName();
void axtSetAlignment(), axtHiliteText(), axtLiftTextCursor();
void axtSetZoom(), axtGetBoxBounder();
void axtFreeFillPattern(), axtScaleInvR(), axtCopyColor();
void axtDefaultText(), axtInitGCs();
void axtAdjustRulingEndpoints();
void axtSetFieldWidth(), axtRemoveFieldWidth();
void axtBulletRect();
void axtToggleForegroundColor();

int milsToPix();
int setDefaultTextAttrs(), insChars(), axtGetAlignment();
int segCount();
int axtDecipointSize(), axtGetAngle();
int axCyanComponent(), axMagentaComponent();
int axYellowComponent(), axBlackComponent();
int axtPixToMils(), axtWellFormedXLFD();
int axtHasBullet(), axtHasOffset();

char *getOverallBounder(), *getMaxBounds();
char *constructForPrint(), *constructForDisplay();

axtext *axtCopy();

static void foldDefaults(), extractYscale(), extractXscale();
static void freeBounder();
static void setPosFromBounder();
static void updateTargetChar(), fieldDefaults();
static void updateScreenChars();
static void justifyLine(), doAlignment();
static void liftOld();
static void getRelativeOffset();
static void copyAxtext();
static void txAttDefaults();
static void copyInTextAttr(), copyOutTextAttr();
static void copyTxAttr();
static void liftText(), getTxBounderRect(), setLiftColor();
static void writeForwardCtm(), writeReverseCtm();
static void getTransform(), defaultTransform();
static void axtSetFgBg(), axtTransformChanged();
static void resetPerChars();
void getTriPoints(), getDiamondPoints(), getBarPoints();
static void switchCursor(), resetEditState();
void txtSetClip(), axUnSetClip(), intersectRects();
void axtCopyFillPattern();
void axtDrawText(), axtFreeIntercepts();
static void axtZoomFormattedText();
static void liftBullet();
void axtFreeIndent(), axtCopyIndent();
void axtFreeBullet(), axtCopyBullet();
void axtFreeDropShadow(), axtCopyDropShadow();
static void axtSnapshotText(), extendFieldBounder();
static void axtSnapshotFieldBounder();
static void axtAlignPos(), axtGetIndentLineVals();
static void axtIntSetFieldWidth(), axtIntSetRectBounder();
static void axtGetFieldXform(), axtTrimText();

static int axtEditPos();
static int setDrawMethod(), foundMatchingScreenfont();
static int getCursorPoints();
static int drawLetterforms();
static int qDrawText();
static int insertItems(), getOldIndex();
static int deleteItems();
static int attrChange(), needSplit(), checkJoin();
int doSplit(), affectsLetterforms();
static int defaultLineFormatter();
static int defaultGetIntercepts();
static int needLift();
static int nonrotated(), axtHypot();
static int wildCardsPresent(), pixelHeight();
static int samePointsize();
static int itemSize();
static int defaultPointsize();
int differingStrings(), differingBullets(), differingIndents();
int affectsFontName();
int useForegroundColorForBlack();
static int axtAcuteVectors();
static int firstLeftOffset();
static int axtGetMaxWidth();
static int axtNRindexAtPoint();

static txAttr *attrSegAtPos(), *inheritedAttr();

static intercept *axtCombineIntercepts();

axtIndent *axtNewDefaultIndent();
axtBullet *axtNewDefaultBullet();

static char *reFit();
#if AX_X11
static char *getClosestRasterName();
char *axtGetClosestRasterName();
#endif
static char *axtMeasuredText();
static char *formLetterName();
static char *axtDisplayableText();
static char *getAx2Bounder();
attr_type *axtNewFillPattern();
extern  Drawable AxGetDrawableFromWidget();

static poly *dexformedBounderCopy();

static XImage *scaleLetterform();
static GC getGC(), getPixmapGC();

#if !FOR_FILTER
extern char *TaskAlloc();
extern char *TaskRealloc();
extern void TaskFree();
#endif
extern int axtNeedBullet();

static void tossWchar(), tossTxAttr(), tossPerChar();
static void (*tossItem[])() = {
    tossWchar,
    tossTxAttr,
    tossPerChar,
};

char *
txloc(pid, bytesize)
int pid;
int bytesize;
{
#if FOR_FILTER
	 return(malloc(bytesize));
#else
	 return(TaskAlloc(pid, bytesize));
#endif
}
	
char *
retxloc(pid, buf, bytesize)
int	pid;
void *buf;
int bytesize;
{
#if FOR_FILTER
	return(realloc(buf, bytesize));
#else
	return(TaskRealloc(pid, buf, bytesize));
#endif
}

void
txfree(buf)
void *buf;
{
	int pid;

	if ( buf )
	{
#if FOR_FILTER
		free(buf);
#else
		pid = AxOwnerOfThisMemory(buf);
		TaskFree(pid, buf);
#endif
	}
}

/*
 * Contsruction / destruction routines
 */

/* : axtDefaultText - write hardcoded default attributes onto *tx
 */
void
axtDefaultText(tx)
axtext *tx; /* out: fill this in upto 'axtDefaultText' level */
{
	unsigned int jnk;
    txAttr *tDefault, tmpAttr;

    if (!tx)
        return;
    if (tx->textstr) {
        txfree(tx->textstr);
        tx->textstr = (tx_char *)NULL;
    }

    txAttDefaults(&tmpAttr, &jnk, tx->pid); /* hardcoded defaults */
    if (tx->userDefaults) {
            /* use existing user default */
        tDefault = tx->userDefaults;
    }
    else {
        tDefault = &tmpAttr; /* use hardcoded */
    }
    if (tx->atts) {
        txfree(tx->atts);
    }
        /* initial attribute info matches default */
    tx->atts = (txAttr *) txloc(tx->pid, sizeof(txAttr));
    tx->txAttrBuflen = 1;
    memset((char *)tx->atts, 0, sizeof(txAttr));
    copyTxAttr(&tx->atts[0], tDefault, tx->pid);

    tx->atts[0].nChars = 0;
    fieldDefaults(tx);
    tossTxAttr((char *) &tmpAttr);
    tx->whole.nChars = 0;
    tx->constructed = TRUE;
}

/* deallocate that constructed during axtDefaultText */
static void
dtext1(tx)
axtext *tx;
{
    int charcount;
    txAttr *txAp;
    
    if(tx) {
        if (tx->atts) {
            charcount = tx->whole.nChars;
            if (charcount) {
                int segix = 0;
                while(charcount > 0) {
                    txAp = &tx->atts[segix];
                    tossTxAttr((char *) txAp);
                    charcount -= txAp->nChars;
                    segix++;
                }
            }
            else {
                tossTxAttr((char *) &tx->atts[0]);
            }
            txfree(tx->atts);
            tx->atts = (txAttr *)NULL;
            tx->txAttrBuflen = 0;
        }
        if (tx->textstr) {
            txfree(tx->textstr);
            tx->textstr = (tx_char *) NULL;
            tx->wCharBuflen = 0;
        }
        if (tx->userDefaults) {
            tossTxAttr((char *) tx->userDefaults);
            txfree(tx->userDefaults);
            tx->userDefaults = (txAttr *) NULL;
        }
		if ( tx->whole.tabInfo.tabStops ) {
            txfree(tx->whole.tabInfo.tabStops);
            tx->whole.tabInfo.tabStops = (int *) NULL;
		}
        memset((char *)tx, 0, sizeof(axtext));
    }
}

/* : attrSegAtPos - returns a pointer to the txAttr instance
 * that applies to the textchar at position pos */
static txAttr *
attrSegAtPos(txt, pos, lpos)
axtext *txt;
int pos;
int *lpos; /* return index of pos within attr */
{
    int charcount, segix, charix, lastix, tpos;
    txAttr *txAp;
    txAttr *segp;

	*lpos = 0;
	if ( !txt )
	{
        return ((txAttr *)NULL);
	}

    charcount = txt->whole.nChars;
    txAp = &txt->atts[0];
	segp = txAp;
	segix = 0;
	charix = tpos = 0;
	if ( pos >= 0 )
	{
		if ( pos > charcount )
			pos = charcount;
		while ( charcount > 0 )
		{
			txAp = &txt->atts[segix++];
			segp = txAp;
			lastix = charix + txAp->nChars;
			if ( lastix > pos )
			{
				tpos = charix;
				charix = pos;
				*lpos = charix-tpos;
				break;
			}
			else
				charix = lastix;
			charcount -= txAp->nChars;
		}
	}
    return segp;
}

/* : axtMeasuredText - put target measurements onto *tx */
static char * /* NULL if ok */
axtMeasuredText(tx)
axtext *tx; /* out: fill this in upto 'axtMeasuredText' level */
{
    int charcount, i;
    int segix, charix, lastix, lpos, charsInSeg;
    txAttr *txAp;
    perChar *pc;
    char *errString;
    int lastfit;

    if (!tx)
        return (getErrString(XLT("The text pointer is null.."),
                    (char *)NULL));
    if (!tx->constructed) {
        axtDefaultText(tx);
    }
    if (!tx->items && tx->whole.nChars > 0) {
            /* get mem for per-char data */
        tx->items = (perChar *)txloc(tx->pid, tx->whole.nChars *
                            sizeof(perChar));
        tx->perCharBuflen = tx->whole.nChars;
        for(i=0; i<tx->whole.nChars; i++) {
            memset((char *)&tx->items[i], 0, sizeof(perChar));
        }
    }
        /* now go thru the attr seg structs */
    charcount = tx->whole.nChars - tx->editState.editPos;

#if BUILD_ORIENT && !FOR_FILTER
    CheckCharsAgainstAttrs( tx );
#endif

    txAp = attrSegAtPos(tx, tx->editState.editPos, &lpos);
    segix = txAp - tx->atts;
    charix = tx->editState.editPos;
    
    while(charcount > 0) {
        
        txAp = &tx->atts[segix++];
        charsInSeg = txAp->nChars - lpos;
        lpos = 0;
        
        if (!txAp->fontAttr.fontName) {
            if (errString = axGetFontRecordName(txAp->fontAttr.familyName,
                    txAp->fontAttr.weight, txAp->fontAttr.slant,
                    &txAp->fontAttr.fontName)) {
                return(errString);
            }
        }
            /* ensure uptodate refs to font table entry, target metrics */
        if (!txAp->fontAttr.fontName) {
            return (getErrString(XLT("The font family name is null.."),
                    (char *)NULL));
        }
        if (errString = axGetFontRecord(txAp->fontAttr.fontName,
                                        &txAp->fconx.font)) {
            return errString;
        }
        if (!txAp->fconx.font->targetMetricsNames[0]) {
            return (getErrString(XLT("The target metrics filename is null for font name"), txAp->fontAttr.fontName));
        }
        if (errString =
            axGetTargetMetrics(txAp->fconx.font->targetMetricsNames[0],
                               txAp->fconx.font,
                               &txAp->fconx.target)) {
            return errString;
        }
        txAp->fontAttr.underlinePosition =
                -txAp->fconx.target->underlinePosition;
        txAp->fontAttr.underlineThickness =
                txAp->fconx.target->underlineThickness;
            /* remember overall font bounding box */
        i = txAp->fconx.target->xfont->max_bounds.lbearing;
        txAp->fontAttr.bbox.origin.x = emToMils(i, txAp->fontAttr.ptsizeInt);
        i = -txAp->fconx.target->xfont->max_bounds.ascent;
        txAp->fontAttr.bbox.origin.y = emToMils(i, txAp->fontAttr.ptsizeInt);
        i = txAp->fconx.target->xfont->max_bounds.descent +
                    txAp->fconx.target->xfont->max_bounds.ascent;
        txAp->fontAttr.bbox.height = emToMils(i, txAp->fontAttr.ptsizeInt);
        i = txAp->fconx.target->xfont->max_bounds.rbearing -
                    txAp->fontAttr.bbox.origin.x + 1;
        txAp->fontAttr.bbox.width = emToMils(i, txAp->fontAttr.ptsizeInt);
            /* jot targ metrics for the particular chars in this seg */
        if (!txAp->tctm.known) {
                /* need to init the target ctm (tctm).  but first
                   test whether the transform is nonunity */
            if (txAp->transform.xscale == 10000 &&
                txAp->transform.yscale == 10000 &&
                !txAp->transform.xshear &&
                !txAp->transform.angle) {
                txAp->transform.unity = TRUE;
            }
            else {
                txAp->transform.unity = FALSE;
            }
            if (txAp->transform.angle || txAp->transform.xshear) {
                tx->editState.fullDisp |= COMPLEX_XFORM;
            }
            axtWriteTargetCtm(tx->whole.xZoom, tx->whole.yZoom,
                              &txAp->transform, &tx->whole.transform,
                              &txAp->tctm);  /* init txAp->tctm */
                /* need a zoomless version for formatting */
            axtWriteTargetCtm(1000, 1000,
                              &txAp->transform, &tx->whole.transform,
                              &txAp->fctm);
            txAp->tctm.known = TRUE;
        }
        for(lastix=charsInSeg+charix ;charix < lastix; charix++) {
            pc = &tx->items[charix];
            if(!pc->targetMetrics) {
                pc->charcode = tx->textstr[charix];
                    /* get the target metrics */
                updateTargetChar(tx, txAp, pc);
                pc->targetMetrics = TRUE;
            }
        }
        charcount -= charsInSeg;
    }
    tx->measured = TRUE;
    return( reFit(tx, &lastfit) );
}

/* deallocate that constructed during axtMeasuredText */
static void
dtext2(tx)
axtext *tx;
{
    if (!tx)
        return;
    if(tx->whole.bounder.paths) {
        freeBounder(&tx->whole.bounder);
        memset((void *) &tx->whole.bounder, 0, sizeof(poly));
    }
    if(tx->items) {
        txfree(tx->items);
        tx->items = (perChar *)NULL;
        tx->perCharBuflen = 0;
    }
    tx->measured = FALSE;
}

static void
axtMoveItems(tx)
axtext *tx;
{
    int ix;
    if (tx->editState.disp && tx->items) {
        ix = tx->whole.nChars - 1;
        while(ix >= 0) {
            tx->items[ix].moved = TRUE;
            ix--;
        }
    }
}

/* : axtDisplayableText - axtMeasuredText plus screenfont info */
static char *  /* NULL means ok */
axtDisplayableText(tx)
axtext *tx; /* out: fill this in upto 'axtDisplayableText' level */
{
    int charcount, segix, charix, itemsize, charsInSeg, lpos;
    int rasterError=0;
    char *errString;
    txAttr *txAp;
    fontTabEnt *ftab;

    if (!tx)
        return (getErrString(XLT("The text pointer is null.."),
                            (char *)NULL));
    if (!tx->measured) {
        if (errString = axtMeasuredText(tx)) {
            return errString;
        }
    }
    
    charcount = tx->whole.nChars - tx->editState.editPos;
    txAp = attrSegAtPos(tx, tx->editState.editPos, &lpos);
    segix = txAp - tx->atts;
    charix = tx->editState.editPos;
    
    while(charcount > 0) {

        txAp = &tx->atts[segix++];
        charsInSeg = txAp->nChars - lpos;
        lpos = 0;
        
        if (!txAp->fontAttr.fontName) {
            return (getErrString(XLT("The font family name is null.."),
                    (char *)NULL));
        }
        if (errString = axGetFontRecord(txAp->fontAttr.fontName,
                                        &txAp->fconx.font)) {
            return errString;
        }
        ftab = txAp->fconx.font;
        if (!ftab->xNamesKnown) {
            return (getErrString(XLT("No screenfont names for"),
                                 txAp->fontAttr.fontName));
        }
            /* ensure uptodate fconx.target reference
             */
        if (!txAp->fconx.font->targetMetricsNames[0]) {
            return (getErrString(XLT("Font metrics file name is null for font named"), txAp->fontAttr.fontName));
        }
        if (errString =
            axGetTargetMetrics(txAp->fconx.font->targetMetricsNames[0],
                               txAp->fconx.font,
                               &txAp->fconx.target)) {
            return errString;
        }
        if (!txAp->screenFontName) {
            float sy;
            char *txName;
            char *lsName = (char *) NULL;
            fontTabEnt *screenRecord, *notUsed;

            extractYscale(&txAp->tctm, &sy);   /* want max possible ptsize */
            
            if (ftab->XAlternate || ftab->largeXAlternate ||
                ftab->WindowsAlternate) {
                
                    /* use the alternate record's screenfont info
                     */
                if (errString =
                    axGetFontRecordForScreen(ftab->familyName,
                                             ftab->weight,
                                             ftab->slant,
                                             (int)(txAp->fontAttr.ptsize * sy),
                                             &notUsed,
                                             &screenRecord)) {
                    return errString;
                }
            }
            else {
                screenRecord = ftab;
            }
            
#if AX_X11   /*  {  */
            
                /* on X, find or form the name of the best xlfd to use
                 */
            if (screenRecord->scalable) {
                
                lsName = (char *) txloc(tx->pid, MAXLFD);
                axtFormScalableName(screenRecord->scalableName,
                                    (int)(txAp->fontAttr.ptsize * sy),
                                    ScreenRes, (axctm *) 0,
                                    lsName);
                txName = lsName;
            }
            else {
                txName = getClosestRasterName(screenRecord, txAp,
                                              &rasterError);
            }
            
#else

                /* on Win32, form "xlfd" as <ptsize>,<wt>,<slant>,<family>
                 */
            lsName = axMakeCompoundWinName(screenRecord->familyName, 
                                           (int)(txAp->fontAttr.ptsize * sy),
                                           txAp->fontAttr.weight,
                                           txAp->fontAttr.slant);
            txName = lsName;
            
#endif    /*  }  AX_X11 */
            itemsize = strlen(txName) + 1;
            txAp->screenFontName = (char *)txloc(tx->pid, itemsize);
            strcpy(txAp->screenFontName, txName);
                /* qDraw tells whether we can use straight-to-screen display
                 */
            txAp->qDraw = setDrawMethod(tx, txAp, screenRecord->scalable,
                                        rasterError, (double)sy);
            if (lsName) {
                txfree(lsName);
            }
        }
            /* load the metrics for the screenfont
             */
        if (!txAp->screenFontName) {
            return (getErrString(XLT("Screenfont name is null for font named"), txAp->fontAttr.fontName));
        }

        if (errString = axGetScreenMetrics(txAp->screenFontName,
                                           txAp->fconx.font,
                                           &txAp->fconx.screen)) {
            return errString;
        }
            /* write per-char screen metrics
             */
        updateScreenChars(txAp, tx->items, charix, charix + charsInSeg - 1);

        if (!txAp->qDraw) {
                /* some sort of transform is needed before display
                 */
            getLetterformInfo(txAp, tx->items, charix,
                              charix + charsInSeg - 1, tx->pid);
        }
        charcount -= charsInSeg;
        charix += charsInSeg;
    }
    tx->displayable = TRUE;
    return ((char *)NULL);
}

    /* deallocate that constructed during axtDisplayableText */
static void
dtext3(tx)
axtext *tx;
{
    int charcount;
    txAttr *txAp;
    perChar *pc;

    if (!tx)
        return;

    if (tx->atts) {
        charcount = tx->whole.nChars;
        if (charcount) {
            int segix = 0;
            while(charcount > 0) {
                txAp = &tx->atts[segix++];
                if (txAp->screenFontName) {
                    txfree((char *)txAp->screenFontName);
                    txAp->screenFontName = (char *)NULL;
                }
                charcount -= txAp->nChars;
            }
        }
        else {
            txAp = &tx->atts[0];
            if (txAp->screenFontName) {
                txfree((char *)txAp->screenFontName);
                txAp->screenFontName = (char *)NULL;
            }
        }
    }
    if (tx->items) {
        int charix;
        for(charix = 0; charix < tx->whole.nChars; charix++) {
            pc = &tx->items[charix];
            if (pc->ctm) {
                txfree((char *)pc->ctm);
                pc->ctm = (axform *)NULL;
            }
            if (pc->letterName) {
                txfree((char *)pc->letterName);
                pc->letterName = (char *)NULL;
            }
        }
    }
    if (tx->old) {
        daxtext(tx->old);
        txfree((char *) tx->old);
        tx->old = (axtext *)NULL;
    }
    tx->displayable = FALSE;

}

/*
 * Public routines
 */

/* : constructForPrint - wrapper for axtMeasuredText
 *   portion of construction
 */
char *  /* if null, OK; if not, caller must free error string */
constructForPrint(txt)
axtext *txt;
{
    if (txt->measured)
        return ((char *) NULL);
    else
        return (axtMeasuredText(txt));
}

/* : constructForDisplay - wrapper for axtDisplayableText
 *   portion of construction
 */
char *  /* if null, OK; if not, caller must free error string */
constructForDisplay(txt)
axtext *txt;
{
    if (txt->displayable)
        return ((char *) NULL);
    else
        return (axtDisplayableText(txt));
}

/* : daxtext - deconstruct an axtext instance */
void
daxtext(tx)
axtext *tx;
{
    if (tx) {
        dtext3(tx);
        dtext2(tx);
        dtext1(tx);
    }
}

/* : axDrawText - convenience wrapper
 */
void
axDrawText(w, textstr, len, x, y, atts, flags)
long w;
tx_char *textstr;
int len, x, y;       /* text length in chars, position in mils */
textAttrs *atts;     /* text attributes */
unsigned int flags;  /* flags for attributes */
{
    axtext txt;
    p32 pos;
    char *errStr, *dispTx();
    int insChars(), setDefaultTextAttrs();
    void daxtext(), setPos();

    memset((char *) &txt, 0, sizeof(txt));
    setDefaultTextAttrs(&txt, atts, flags);
    insChars(&txt, textstr, 0, len);
    pos.x = x;
    pos.y = y;
    setPos(&txt, &pos);
    if(errStr = dispTx(w, &txt, (r32 *) 0)) {
        freeErrString(errStr);
    }
    daxtext(&txt);
}

/* : setPos - set the window-relative position of *txt */
void
setPos(txt, pos)
axtext *txt;
p32 *pos; /* drawable-relative mil coords */
{
    if (!txt->whole.bounder.numPaths) {
        /* (the position of bounded text is determined by the
         * bounder itself, so accept position only if no bounder).
         */
        if (txt->editState.disp) {
            if(!txt->old) {
                /* first edit since last display -- save state.
                 * setPos is effectively an edit at pos 0
                 */
                axtSnapshotText(txt, 0);
            }
            axtMoveItems(txt);  /* mark all chars as moved */
            txt->editState.fieldBounder.width = 0;  /* turn off clear rect */
        }
        
        /* adjust fixedPos by the whole.pos displacement
         */
        txt->whole.fixedPos.x += pos->x - txt->whole.pos.x;
        txt->whole.fixedPos.y += pos->y - txt->whole.pos.y;

        txt->whole.pos = *pos;
        txt->measured = txt->displayable = FALSE;
    }
}

/* : axtGetPos - return the window-relative position of *txt */
void
axtGetPos(txt, pos)
axtext *txt;
p32 *pos;            /* out: drawable-relative mil coords */
{
    int formattedWidth;
    axform fieldXform;
    char *err;
    p32 left;

    if ( txt == (axtext *)0 )
        pos->x = pos->y = 0;
    else if ( !txt->whole.widthBounder || txt->whole.bounder.numPaths <= 0 )
        *pos = txt->whole.pos;
    else {
        if (!txt->measured) {
            if (err = axtMeasuredText(txt)) {
                pos->x = pos->y = 0;
                freeErrString(err);
                return;
            }
        }
        if ( txt->whole.bounder.numPaths <= 0 )
            *pos = txt->whole.pos;
        else {
            if (txt->whole.xZoom != 1000) {
                /* position contains zoom, so apply zoom to formattedWidth */
                formattedWidth = (int) (((float) txt->whole.formattedWidth *
                txt->whole.xZoom) / 1000.0);
            }
            else {
                formattedWidth = txt->whole.formattedWidth;
            }

            left.y = 0;
            switch(txt->whole.alignment) {
                case LEFT_ALIGN:
                case JUSTIFIED_ALIGN:
                    left.x = 0;
                break;
                
                case CENTER_ALIGN:
                    left.x = -(formattedWidth / 2);
                break;
                
                case RIGHT_ALIGN:
                    left.x = -formattedWidth;
                break;
            }
            
            if (left.x && !txt->whole.transform.unity) {
                axtGetFieldXform(txt, 1000, 1000, &fieldXform);
                forwardSXRT(left.x, 0, &left.x, &left.y, &fieldXform);
            }
            
            left.x += txt->whole.fixedPos.x;
            left.y += txt->whole.fixedPos.y;
            
            *pos = left;
        }
    }
}

/* : axtGetFixedPos - return the window-relative position of *txt */
void
axtGetFixedPos(txt, pos)
axtext *txt;
p32 *pos;            /* out: drawable-relative mil coords */
{
    char *err;

    if ( txt == (axtext *)0 )
        pos->x = pos->y = 0;
    else if ( !txt->whole.widthBounder )
        *pos = txt->whole.pos;
    else {
        if (!txt->measured) {
            if (err = axtMeasuredText(txt)) {
                pos->x = pos->y = 0;
                freeErrString(err);
                return;
            }
        }
        if ( txt->whole.bounder.numPaths <= 0 )
            *pos = txt->whole.pos;
        else {
            *pos = txt->whole.fixedPos;
        }
    }
}

/* : axtSetFixedPos - set the window-relative position of *txt */
void
axtSetFixedPos(txt, pos)
axtext *txt;
p32 *pos;            /* in: drawable-relative mil coords */
{
    if ( !txt->whole.widthBounder )
        txt->whole.fixedPos = txt->whole.pos;
    else {
        if ( txt->whole.bounder.numPaths <= 0 )
            txt->whole.fixedPos = txt->whole.pos;
        else {
            txt->whole.fixedPos = *pos;
            txt->measured = txt->displayable = FALSE;
            axtAlignPos(txt);
        }
    }
}

static int
axtNeedToIgnoreTabs(tx, ix)
axtext *tx;
int ix;
{
    txAttr *txAp;
    int ignore_tabs;

	txAp = axtAttrAtPos(tx, ix);
	ignore_tabs = FALSE;
    if ( txAp->indent && txAp->indent->bullet && txAp->indent->bullet->type)
    {
    	if ( ix == 0 )
    	{
    		ignore_tabs = TRUE;
    	}
    	else
    	{
    		if ( CHAR_IS_NEWLINE(tx->items[ix-1].charcode) )
				ignore_tabs = TRUE;
    	}
    }
    return ignore_tabs;
}

static int
axtGetTabWidth(tx, curr_x, do_zoom)
axtext *tx;
int curr_x;
int do_zoom;
{
	int i, accum_x, found, tstop;

#if 1
	if ( tx->whole.tabInfo.defTabSize <= 0 )
		return 0;
	if ( tx->whole.tabInfo.n_tabs )
	{
		accum_x = 0;
		found = FALSE;
#if 0
		do_zoom = FALSE;
#endif
		for ( i=0; i< tx->whole.tabInfo.n_tabs; i++ )
		{
			if ( curr_x < accum_x )
			{
				found = TRUE;
				break;
			}
			tstop = tx->whole.tabInfo.tabStops[i]; 
			if ( do_zoom )
				tstop =  (double)tstop * tx->whole.xZoom/1000.0;
			accum_x += tstop;
		}
		if ( found == FALSE )
		{
			tstop = tx->whole.tabInfo.defTabSize;
			if ( do_zoom )
				tstop =  (double)tstop * tx->whole.xZoom/1000.0;
			while ( curr_x >= accum_x )
				accum_x += tstop;
		}
	}
	else
	{
		accum_x = 0;
		tstop = tx->whole.tabInfo.defTabSize;
		if ( do_zoom )
			tstop =  (double)tstop * tx->whole.xZoom/1000.0;
		while ( curr_x >= accum_x )
			accum_x += tstop;
	}
	return accum_x - curr_x;
#else
	return 0;
#endif
}


/* : getRelaiveOffset - returns total positional offset over
 * text character range {ix1 - just before ix2} */
static void
getRelativeOffset(tx, ix1, ix2, off)
axtext *tx;
int ix1, ix2; /* compute offset from ix1 to ix2 */
p32 *off; /* return position of char relative to tx origin in mils */
{
    int i;
    p32 toff;
	txAttr *txAp;
    perChar *pc;

    toff.x = toff.y = 0;
    for(i=ix1; i<ix2; i++) {
		pc= &tx->items[i];
        toff.x += pc->advance.x + pc->offset.x + pc->kern.x;
        toff.y += pc->advance.y + pc->offset.y + pc->kern.y;
        
    }
    if (ix2 < tx->whole.nChars) {
		/* any offset on ix2 itself will affect its position */
		pc= &tx->items[ix2];
        toff.x += pc->offset.x + pc->kern.x + pc->sup.x;
        toff.y += pc->offset.y + pc->kern.y + pc->sup.y;
    }
    off->x = toff.x;
    off->y = toff.y;
}

void
liftCursor(w, txt, clip)
long w;
axtext *txt;
r32 *clip;
{
	graphics_main_type *gfx;
	TmRect32 java_clip, *c;
    int x_off, y_off, i, n;
	TmPoint32 *point;
    Drawable drw;
    XPoint *pt;
    GC gc;

	if ( txt )
	{
		pt = txt->editState.lastCursorPoints;
#if ALLOW_JAVA_GE && !FOR_FILTER

        gfx = (graphics_main_type *)txt->gContext;
        if ( gfx && gfx->state.java_display )
		{
			if ( clip == (r32 *)0 )
				c = (TmRect32 *)0;
			else
			{
				java_clip.x_position = clip->origin.x;
				java_clip.y_position = clip->origin.y;
				java_clip.x_size = clip->width;
				java_clip.y_size = clip->height;
				c = &java_clip;
			}
			n = txt->editState.lastCursorPointcount;
			point = (TmPoint32 *)txloc(txt->pid, n*sizeof(TmPoint32));
			for ( i=0; i<n; i++ )
			{
				point[i].x = pt[i].x;
				point[i].y = pt[i].y;
			}
			gJavaCursor(gfx, n, point, c);
			txfree(point);
		}
#endif
		/* get the drawable */
		drw = AxGetDrawableFromWidget((AXWidget)w, &x_off, &y_off);
		if ( txt->editState.cursorOn )
		{
			/* assume xor-transfers, lift by drawing */
			if ( txt->editState.lastCursorActive )
				gc = axTextCursorGC;
			else
				gc = axInactiveTextCursorGC;
			if ( txt->editState.lastCursorClip.height > 0 )
				txtSetClip(txt, gc, clip, &txt->editState.lastCursorClip);
			else
				txtSetClip(txt, gc, clip, (r32 *)0);
			if ( txt->editState.lastCursorPointcount == 2 )
				XDrawLine(Dpy, drw, gc, pt[0].x, pt[0].y, pt[1].x, pt[1].y);
			else
			{
				XFillPolygon(Dpy, drw, gc, pt,
				             txt->editState.lastCursorPointcount,
				             Convex, CoordModeOrigin);
			}
			axUnSetClip(gc, clip);
			txt->editState.cursorOn = FALSE;
		}
	}
}

void
axtLiftTextCursor(w, txt)
long w;
axtext *txt;
{
	liftCursor(w, txt, (r32 *)0);
}

void
axtDrawTextCursor(w, txt, clip, pos, type, active)
long w;
axtext *txt;
r32 *clip;
int pos; /* pos 0 puts cursor just before 1st char */
int type; /* i-bar or ol triangle */
int active;
{
	int x_off, y_off, free_bounder, lpos, i, pointcount;
	r32 bounderRect, cursorClip;
	p32 base_pt, relpt, cvec;
	graphics_main_type *gfx;
	axform fieldXform, ctm;
	TmRect32 java_clip, *c;
	TmPoint32 *point;
	XPoint pts[4];
	double xz, yz;
	Drawable drw;
	poly *boundp;
	txAttr *txAp;
	char *err;
	GC gc;

    if ( !txt || pos < 0 || pos > txt->whole.nChars )
		return;

	err = txt->displayable ? (char *)0 : axtDisplayableText(txt);
	if ( err != (char *)0 )
		freeErrString(err);
	else
	{
		/* get the base cursor pos (may need to set the text position) */
		defaultTransform(&fieldXform);
		if ( txt->whole.nChars == 0 && txt->whole.bounder.numPaths )
		{
			/* in the empty bounded text case, need a sane initial pos */
			if ( txt->whole.transform.unity &&
			     txt->whole.xZoom == 1000 &&
			     txt->whole.yZoom == 1000 )
			{
				boundp = &txt->whole.bounder;
				free_bounder = FALSE;
			}
			else
			{
				/* de-transform (field-level transform only) the bounder */
				axtGetFieldXform(txt, 1000, 1000, &fieldXform);
				boundp = dexformedBounderCopy(txt, &fieldXform);
				free_bounder = TRUE;
			}
			getTxBounderRect(boundp, &bounderRect);
			setPosFromBounder(txt, txt->whole.lineHeight,
			                  (int (*)()) NULL, boundp,
			                  &bounderRect, &fieldXform);
			xz = txt->whole.xZoom / 1000.0;
			yz = txt->whole.yZoom / 1000.0;
			txt->whole.pos.x = xz * txt->whole.pos.x;
			txt->whole.pos.y = yz * txt->whole.pos.y;
			txt->whole.fixedPos.x = xz * txt->whole.fixedPos.x;
			txt->whole.fixedPos.y = yz * txt->whole.fixedPos.y;
			if ( free_bounder )
			{
				freeBounder(boundp);
				txfree((char *)boundp);
			}
		}
		if ( !txt->whole.nChars && txt->whole.widthBounder )
			base_pt = txt->whole.fixedPos;
		else
    		base_pt = txt->whole.pos;

		/* lift the existing cursor */
		if ( txt->editState.cursorOn )
			axtLiftTextCursor(w, txt);
    
		/* draw the new cursor */
		if ( txt->items &&
		     ((pos < txt->whole.nChars && txt->items[pos].hidden) ||
		      (pos > 0 && txt->items[pos - 1].hidden)) )
		{
			/* don't draw if adjacent to hidden char */
			txt->editState.cursorOn = FALSE;
		}
		else /* not adjacent to hidden char */
		{
			/* get cursor size based on text size */
			txAp = (pos > 0) ? attrSegAtPos(txt, pos-1, &lpos) : &txt->atts[0];
			pointcount = getCursorPoints(txt, type, active, txAp, pts,
			                             &cursorClip);
			txt->editState.lastCursorPointcount = pointcount;

			/* add in whole-pos-relative offset */
			if ( txt->items )
			{
				if ( pos > 0 )
				{
					getRelativeOffset(txt, 0, pos - 1, &relpt);
					relpt.x += txt->items[pos - 1].advance.x;
					relpt.y += txt->items[pos - 1].advance.y;
					base_pt.x += relpt.x;
					base_pt.y += relpt.y;
				}
				else
				{
					base_pt.x += txt->items[0].offset.x;
					base_pt.y += txt->items[0].offset.y;
				}
			}
			else if ( txt->whole.nChars == 0 && txAp->indent )
			{
				/* indented with no text; offset by the indent distance */
				cvec.x = firstLeftOffset(txt, txAp);
				cvec.y = 0;
				axtWriteTargetCtm(txt->whole.xZoom, txt->whole.yZoom,
				                  &txAp->transform, &txt->whole.transform,
				                  &ctm);  /* init txAp->tctm */
				forwardSXRT(cvec.x, cvec.y, &cvec.x, &cvec.y, &ctm);
				base_pt.x += cvec.x;
				base_pt.y += cvec.y;
			}
			for ( i=0; i<4; i++ )
			{
				pts[i].x += (short)(milsToPix(base_pt.x, ScreenRes));
				pts[i].y += (short)(milsToPix(base_pt.y, ScreenRes));
				txt->editState.lastCursorPoints[i].x = pts[i].x;
				txt->editState.lastCursorPoints[i].y = pts[i].y;
			}

			/* draw the new cursor */
			gc = active ? axTextCursorGC : axInactiveTextCursorGC;
			if ( cursorClip.height <= 0 )
				txtSetClip(txt, gc, clip, (r32 *)0);
			else
			{
				cursorClip.origin.x += txt->whole.pos.x;
				cursorClip.origin.y += txt->whole.pos.y;
				txtSetClip(txt, gc, clip, &cursorClip);
			}

#if ALLOW_JAVA_GE && !FOR_FILTER
			gfx = (graphics_main_type *)txt->gContext;
	        if ( gfx && gfx->state.java_display )
			{
				if ( clip == (r32 *)0 )
					c = (TmRect32 *)0;
				else
				{
					java_clip.x_position = clip->origin.x;
					java_clip.y_position = clip->origin.y;
					java_clip.x_size = clip->width;
					java_clip.y_size = clip->height;
					c = &java_clip;
				}
				point = (TmPoint32 *)txloc(txt->pid,
				                           pointcount*sizeof(TmPoint32));
				for ( i=0; i<pointcount; i++ )
				{
					point[i].x = pts[i].x;
					point[i].y = pts[i].y;
				}
				gJavaCursor(gfx, pointcount, point, c);
				txfree(point);
			}
#endif
			drw = AxGetDrawableFromWidget((AXWidget)w, &x_off, &y_off);
			if ( drw )
			{
				if ( pointcount == 2 )
				{
					XDrawLine(Dpy, drw, gc,
					          pts[0].x, pts[0].y,
					          pts[1].x, pts[1].y);
				}
				else
				{
					XFillPolygon(Dpy, drw, gc,
					             pts, pointcount,
					             Convex, CoordModeOrigin);
				}
			}
			txt->editState.cursorOn = TRUE;
			axUnSetClip(gc, clip);
			txt->editState.cursorType = type;
			txt->editState.cursorActive = active;
			txt->editState.lastCursorActive = active;
			txt->editState.lastCursorClip = cursorClip;
	    }
	}
	txt->editState.cursorPos = pos;
}

void
axtClearEditState(txt) /* clear the editState */
axtext *txt;
{
    if (!txt)
        return;
    memset((void *) &txt->editState, 0, sizeof(eState));
}

void
deactivateCursor(txt) /* typically because we lost kb focus */
axtext *txt;
{
    if (!txt)
        return;
    txt->editState.cursorActive = FALSE;
}

void
activateCursor(txt) /* typically because we got kb focus */
axtext *txt;
{
    if (!txt)
        return;
    txt->editState.cursorActive = TRUE;
}

/* : pointAtIndex - given textchar index, return window-
 * relative location of the 'set position' of that character. */
void
pointAtIndex(txt, index, pos)
axtext *txt;
int index;
p32 *pos; /* return this; position of index, window-relative */
{
    char *err;

    if (!txt || !pos)
        return;

    if ( !txt->measured ) {
        if( err = axtMeasuredText(txt)) {
            freeErrString(err);
            txt->measured = txt->displayable = FALSE;
            pos->x = pos->y = 0;
            return;
        }
    }
    if (txt->whole.nChars) {
        getRelativeOffset(txt, 0, index, pos);
        pos->x += txt->whole.pos.x;
        pos->y += txt->whole.pos.y;
    }
    else {
        axtGetPos(txt, pos);
    }
}

/* : indexAtPoint - returns index of 1st textchar found whose
 * bounder contains pos
 */
int
indexAtPoint(txt, pos)
axtext *txt;
p32 *pos;
{
    int i;
    p32 currentpoint, cen, tmp, leastVec;
    int leastDistance, tdistance, leastIx;
    int rectangular;
    perChar *pc = NULL;
    char *err;

    if ( !txt )
        return 0;

    if ( !txt->measured ) {
        if( err = axtMeasuredText(txt)) {
            freeErrString(err);
            txt->measured = txt->displayable = FALSE;
            return 0;
        }
    }

    if (!txt->whole.transform.angle && !txt->whole.transform.xshear) {
        
        /* text is rectangular at field-level; also scan
         * attr segs to see whether any char-level angle occurs
         */
        txAttr *txAp;
        int scount, tix;
        
        rectangular = TRUE;
        scount = segCount(txt, -1, &txAp);
        for (tix = 0; tix < scount; tix++) {
            if (txt->atts[tix].transform.angle ||
                txt->atts[tix].transform.xshear) {
                
                rectangular = FALSE;
            }
        }
        if (rectangular) {
            return (axtNRindexAtPoint(txt, pos));
        }
    }
    
    /* use leastDistance to store the smallest
     * displacement between the passed point and the
     * center of a given letter's bounding box
     */
    leastDistance = MAXLRU;    /* big */
    leastIx = -1;
    leastVec.x = 1;    /* arbitrary init to unit vector */
    leastVec.y = 0;

    currentpoint = txt->whole.pos;
    for(i = 0; i < txt->whole.nChars; i++) {
        pc = &txt->items[i];
        currentpoint.x += pc->offset.x + pc->kern.x;
        currentpoint.y += pc->offset.y + pc->kern.y;

        if (!pc->hidden) {
            /* get the center of this letter's bbox
             */
            cen.x = (pc->bbox.p0.x + pc->bbox.p1.x + pc->bbox.p2.x +
                     pc->bbox.p3.x) / 4;
            cen.y = (pc->bbox.p0.y + pc->bbox.p1.y + pc->bbox.p2.y +
                     pc->bbox.p3.y) / 4;
            cen.x += (currentpoint.x + pc->sup.x);
            cen.y += (currentpoint.y + pc->sup.y);
            
            /* check whether this is the closest so far
             */
            tmp.x = pos->x - cen.x;
            tmp.y = pos->y - cen.y;
            tdistance = axtHypot(0, 0, tmp.x, tmp.y);
            if (tdistance < leastDistance) {
                leastDistance = tdistance;
                leastVec = tmp;
                leastIx = i;
            }
        }
            
        currentpoint.x += pc->advance.x;
        currentpoint.y += pc->advance.y;
    }

    if (leastIx >= 0) {
            /* determine whether the vector between the
             * closest character and the passed position has
             * any projection in the direction of that character's
             * advance; if so, report one index forward (even if
             * the closest character is the last character..)
             */
        if (axtAcuteVectors(&pc->advance, &leastVec)) {
            return (leastIx + 1);  /* NOTE can be one beyond end.. */
        }
        else {
            return (leastIx);
        }
    }
    else {
        return (0);
    }
}

/* : axtNRindexAtPoint - variant of indexAtPoint for the simpler
 * case in which no rotation or xshear occurs anywhere in the text.
 */
static int
axtNRindexAtPoint(txt, pos)
axtext *txt;
p32 *pos;
{
    int stripbot;    /* at line end, lowest character ink ycoord seen */
    int leftix;      /* at line end, greatest char index left of pos */
    int basey;       /* baseline ypos this line */
    int setx;        /* setpoint xpos */
    int lsetx;       /* setx of char at leftx */
    int ledgex;      /* setx of leftmost char on line */
    int ledge_ix;    /* index of leftmost char on line */
    int ix, cenx, charix, tmp;
    
    /* sanity */
    if ( txt->whole.nChars <= 0 )
        return(0);
    
    /* set the parameters of this line of the text field */
    stripbot = basey = txt->whole.pos.y;
    setx = ledgex = txt->whole.pos.x;
    ledge_ix = 0;

    /* set the info on the first character */
    lsetx = txt->whole.pos.x + txt->items[0].offset.x + txt->items[0].kern.x;
    leftix = 0;

    /* rip thru the text, finding the character to the left of the position */
    for ( ix=0; ix<txt->whole.nChars; ix++ )
    {
        /* account for offsets (indents, line wraps, kerns, ...) */
        setx += txt->items[ix].offset.x + txt->items[ix].kern.x;
        if ( txt->items[ix].offset.y )
        {
            if ( pos->y < stripbot ) /* must be on previous line */
            {
                /* bail out of loop */
                ix = txt->whole.nChars;
                continue;
            }
            else /* not on the previous line */
            {
                /* set the parameters of this line of the text field */
                basey += txt->items[ix].offset.y;
                stripbot = basey;
                ledgex = setx;
                ledge_ix = ix;
            }
        }

        /* if descender of character exceeds current bottom, expand bottom */
        tmp = basey + txt->items[ix].bbox.p2.y;
        if ( tmp > stripbot )
            stripbot = tmp;

        /* left edge is left of point, note it's index and pos */
        if ( setx + txt->items[ix].bbox.p0.x < pos->x )
        {
            lsetx = setx;
            leftix = ix;
        }

        /* advance the current position by the width of the character */
        if ( ix < txt->whole.nChars - 1 )
            setx += txt->items[ix].advance.x;
    }

    /* determine the index from the index in the last line examined */
    if ( pos->x < ledgex ) /* to the left of the first char on the last line */
        charix = ledge_ix ? ledge_ix+1 : ledge_ix;
    else
    {
        /* get the center point of the leftmost char */
        cenx = ((txt->items[leftix].bbox.p0.x +
                 txt->items[leftix].bbox.p1.x) / 2) + lsetx;
        charix = (pos->x < cenx) ? leftix : leftix+1;
    }
    return(charix);
}

int
axtGetShadowDisplayInfo(txt, txAp, shadow)
axtext *txt;
txAttr *txAp;
int shadow;
{
    int no_display, s_type;

    if ( !txAp->fillpattern )
        no_display = TRUE;
    else
    {
        s_type = txAp->fillpattern->s_type;
        if ( s_type == 0 )
            no_display = TRUE;
        else switch ( shadow )
        {
            default:
            case 0: /* text */
                no_display = TRUE;
            break;

            case 1: /* embossed shadow */
                if ( txt->draw_emboss == FALSE )
                    no_display = TRUE;
                else if ( s_type != 3 && s_type != 4 )
                    no_display = TRUE;
                else
                    no_display = FALSE;
            break;

            case 2: /* drop shadow */
                if ( txt->draw_shadow == TRUE )
                    no_display = FALSE;
                else
                    no_display = TRUE;
            break;
        }
    }
    return(no_display);
}

char *
axtSubDispTx(w, txt, clip, accum, shadow)
long w;
axtext *txt;
r32 *clip;
p32 accum;
int shadow;
{
    int charcount, segix, ix, charix, t_ix, atStart, lpos, charsInSeg;
    int lx, llen, xdelta, ydelta;
    int inText, txt_display, blt_disp, needs_redraw, full_display, no_display;
    int sdx, sdy, fillcolor, sfg, textless, startpos;
    int (*drawText)(), (*drawBull)();
    p32 prevSup, piecePos, rulept, tpt, rpt;
    txAttr *txAp;
    attr_type txt_attr;
    color_type new_color;
    perChar *pc;
    char *err;
    r32 bRect;

    /* init */
    charcount = txt->whole.nChars - txt->editState.editPos;
    txAp = attrSegAtPos(txt, txt->editState.editPos, &lpos);
    segix = txAp - txt->atts;
    ix = charix = txt->editState.editPos;
    prevSup.x = prevSup.y = 0;
    startpos = accum.x;
    atStart = TRUE;
    inText = TRUE;
    
    if ( txt->whole.nChars == 0 )
        textless = TRUE;   /* we're here only to draw a possible bullet */
    else
        textless = FALSE;
    
    /* draw characters */
    while ( charcount > 0 || textless )
    {
        /* get up-to-date screenfont reference */
        txAp = &txt->atts[segix++];
        charsInSeg = txAp->nChars - lpos;
        lpos = 0;
        
        if ( !textless )
        {
            if ( !txAp->screenFontName )
            {
                err = XLT("The screenfont name is null for font named");
                return(getErrString(err, txAp->fontAttr.fontName));
            }

            /* read screen metrics */
            err = axGetScreenMetrics(txAp->screenFontName,
                                     txAp->fconx.font,
                                     &txAp->fconx.screen);
            if ( err )
                return(err);
        }

        /* set segment starting position */
        lx = charix + charsInSeg;
        llen = ydelta = xdelta = 0;
        if ( atStart == TRUE )
            atStart = FALSE;
        else if ( txt->items )
        {
            accum.x += txt->items[charix].kern.x;
            accum.y += txt->items[charix].kern.y;
            accum.x += txt->items[charix].offset.x;
            accum.y += txt->items[charix].offset.y;
        }
        piecePos = rulept = accum;
        ix = charix;
        if ( txt->items )
        {
            piecePos.x += txt->items[ix].sup.x;
            piecePos.y += txt->items[ix].sup.y;
            rulept.x += txt->items[ix].sup.x;
            rulept.y += txt->items[ix].sup.y;
        }

        /* load drawing function */
        drawText = txAp->qDraw ? qDrawText : drawLetterforms;

        /* determine shadow offsets */
        memset(&txt_attr, 0, sizeof(attr_type));
        if ( txAp->fillpattern )
        {
            txt_attr.ffg = txAp->fillpattern->ffg;
            txt_attr.fbg = txAp->fillpattern->fbg;
            txt_attr.lfg = txAp->fillpattern->lfg;
            txt_attr.lbg = txAp->fillpattern->lbg;
            fillcolor = txAp->fillcolor;
        }
        switch(shadow)
        {
            case 2: /* dark shadow */
                no_display = axtGetShadowDisplayInfo(txt, txAp, shadow);
                if ( no_display )
                {
                    sdx = 0;
                    sdy = 0;
                }
                else
                {
                    sdx = txAp->fillpattern->s_xoff * txt->whole.xZoom / 1000.0;
                    sdy = txAp->fillpattern->s_yoff * txt->whole.yZoom / 1000.0;
                    txAp->fillpattern->ffg = txAp->fillpattern->sfg;
                    txAp->fillpattern->fbg = txAp->fillpattern->sfg;
                    txAp->fillpattern->lfg = txAp->fillpattern->sfg;
                    txAp->fillpattern->lbg = txAp->fillpattern->sfg;
                    sfg = 0;
                    sfg |= txAp->fillpattern->ffg.c << 24;
                    sfg |= txAp->fillpattern->ffg.m << 16;
                    sfg |= txAp->fillpattern->ffg.y << 8;
                    sfg |= txAp->fillpattern->ffg.k;
                    txAp->fillcolor = sfg;
                }
            break;
            case 1: /* embossed/sun shadow */
                no_display = axtGetShadowDisplayInfo(txt, txAp, shadow);
                if (txt->draw_emboss == FALSE )
                    no_display = TRUE;
                if ( no_display )
                {
                    sdx = 0;
                    sdy = 0;
                }
                else if ( txAp->fillpattern->s_type < 3 )
                {
                    no_display = TRUE;
                    sdx = 0;
                    sdy = 0;
                }
                else
                {
                    no_display = FALSE;

                    sdx = txAp->fillpattern->s_xoff * txt->whole.xZoom / 1000.0;
                    sdy = txAp->fillpattern->s_yoff * txt->whole.yZoom / 1000.0;
                    sdx *= (-1);
                    sdy *= (-1);
                    GetWashoutColor(&txAp->fillpattern->sfg, &new_color);
                    txAp->fillpattern->ffg.c = new_color.c;
                    txAp->fillpattern->ffg.m = new_color.m;
                    txAp->fillpattern->ffg.y = new_color.y;
                    txAp->fillpattern->ffg.k = new_color.k;

                    txAp->fillpattern->fbg = txAp->fillpattern->ffg;
                    txAp->fillpattern->lfg = txAp->fillpattern->ffg;
                    txAp->fillpattern->lbg = txAp->fillpattern->ffg;

                    sfg = 0;
                    sfg |= txAp->fillpattern->ffg.c << 24;
                    sfg |= txAp->fillpattern->ffg.m << 16;
                    sfg |= txAp->fillpattern->ffg.y << 8;
                    sfg |= txAp->fillpattern->ffg.k;
                    txAp->fillcolor = sfg;
                }
            break;
            default: /* no shadow */
                no_display = FALSE;
                sdx = 0;
                sdy = 0;
            break;
        }
        
        if ( textless )
        {
            if (txAp->indent && txAp->indent->bullet &&
                txAp->indent->bullet->type)
			{
                if ( !no_display && (!txt->old || txt->editState.fullDisp ||
                                     txAp->indent->bullet->changed) )
                {
                    /* no text -- only thing to display is the bullet
                     */
#if FOR_FILTER
					txAp->indent->bullet->cvt = txt->whole.cvt;
#endif
                    axtBulletRect(txt, txAp, &accum, &bRect);
                    drawBull = txAp->indent->bullet->bulletDraw;
                    bRect.origin.x = bRect.origin.x + sdx;
                    bRect.origin.y = bRect.origin.y + sdy;
                    drawBull(txAp->indent, txAp->fontAttr.ptsizeInt,
                             txAp->fillpattern, &bRect, shadow,
                             txAp->useFgPixel);
                    txAp->indent->bullet->changed = FALSE;
                    txt->editState.disp = TRUE;

                    /* load the clip into the relevant GC's */
                    txtSetClip(txt, axTextGC, clip, (r32 *) NULL);
                    txtSetClip(txt, axTextCursorGC, clip, (r32 *) NULL);
                    txtSetClip(txt, axTextStencilGC, clip, (r32 *) NULL);
                }
            }
            return ((char *) NULL);
        }

        /* draw the text associated with this attribute */
        for ( t_ix=charix; t_ix<lx; t_ix++ )
        {
            pc = &txt->items[t_ix];
            if ( t_ix > charix ) /* include kern and offset in accum */
            {
                accum.x += txt->items[t_ix].kern.x;
                accum.y += txt->items[t_ix].kern.y;
                accum.x += txt->items[t_ix].offset.x;
                accum.y += txt->items[t_ix].offset.y;
            }

            if ( pc->moved || pc->changed || pc->inserted || pc->damaged )
                needs_redraw = TRUE;
            else
                needs_redraw = FALSE;
            if ( !txt->old || txt->editState.fullDisp )
                full_display = TRUE;
            else
                full_display = FALSE;

            /* draw the bullet */
            if ( txAp->indent && axtHasBullet(txt, txAp, pc, t_ix) )
            {
                if ( pc->hidden || no_display )
                    blt_disp = FALSE;
                else if ( txAp->indent->bullet->changed )
                    blt_disp = TRUE;
                else if ( full_display )
                    blt_disp = TRUE;
                else
                    blt_disp = needs_redraw;
				/* don`t display an empty bullet if not in edit mode */
				if ( blt_disp && txt->editState.inEdit == FALSE &&
					axtNeedBullet(txt, txAp->indent, t_ix) == FALSE )
					blt_disp = FALSE;

                if ( blt_disp )
                {
#if FOR_FILTER
					txAp->indent->bullet->cvt = txt->whole.cvt;
#endif
                    axtBulletRect(txt, txAp, &accum, &bRect);
                    drawBull = txAp->indent->bullet->bulletDraw;
                    bRect.origin.x = bRect.origin.x + sdx;
                    bRect.origin.y = bRect.origin.y + sdy;
                    drawBull(txAp->indent,  txAp->fontAttr.ptsizeInt,
                             txAp->fillpattern, &bRect, shadow,
                             txAp->useFgPixel);
                    txAp->indent->bullet->changed = FALSE;

                    /* load the clip into the relevant GC's */
                    txtSetClip(txt, axTextGC, clip, (r32 *) NULL);
                    txtSetClip(txt, axTextCursorGC, clip, (r32 *) NULL);
                    txtSetClip(txt, axTextStencilGC, clip, (r32 *) NULL);
                }
            }
            /* draw the text */
            if ( pc->hidden || no_display )
                txt_display = FALSE;
            else if ( full_display )
                txt_display = TRUE;
            else
                txt_display = needs_redraw;
            if ( inText )
            {
                /* end of the last contiguous group */
                if ( !txt_display || (axtHasOffset(pc, &prevSup) && llen) )
                {
                    /* draw text */
                    tpt.x = piecePos.x + sdx;
                    tpt.y = piecePos.y + sdy;
                    rpt.x = rulept.x + sdx;
                    rpt.y = rulept.y + sdy;
                    drawText(Dpy, w, axTextGC, txt, txAp, ix, llen, &tpt);
                    if ( txAp->ruling || txAp->strike_thru )
                        drawRuling(txt, w, txAp, ix, llen, &tpt, &rpt, FALSE);

                    /* create new text segment */
                    if ( txt_display == FALSE )
                        inText = FALSE; /* end of contiguous group */
                    else
                    {
                        llen = 1;
                        ix = t_ix;
                        piecePos.x = accum.x + pc->sup.x;
                        piecePos.y = accum.y + pc->sup.y;
						if ( CHAR_IS_TABCHAR((int)pc->charcode) )
							piecePos.x += pc->advance.x;
                        rulept = piecePos;
                    }
                }
                else
                    llen++;
            }
            else if ( txt_display )
            {
                inText = TRUE; /* start of contiguous group */
                llen = 1;
                ix = t_ix;
                piecePos.x = accum.x + pc->sup.x;
                piecePos.y = accum.y + pc->sup.y;
				if ( CHAR_IS_TABCHAR((int)pc->charcode) )
					piecePos.x += pc->advance.x;
                rulept = piecePos;
            }

            /* advance the text position */
            prevSup = pc->sup;
            accum.x += pc->advance.x;
            accum.y += pc->advance.y;
            rulept.x = accum.x + pc->sup.x;
            rulept.y = accum.y + pc->sup.y;
        }

        /* any text left undrawn? - draw it now! */
        if ( inText && llen > 0 )
        {
            tpt.x = piecePos.x + sdx;
            tpt.y = piecePos.y + sdy;
            rpt.x = rulept.x + sdx;
            rpt.y = rulept.y + sdy;
            drawText(Dpy, w, axTextGC, txt, txAp, ix, llen, &tpt);
            if ( txAp->ruling || txAp->strike_thru )
                drawRuling(txt, w, txAp, ix, llen, &tpt, &rpt, FALSE);
        }
        charcount -= charsInSeg;
        charix += charsInSeg;
        if ( txAp->fillpattern )
        {
            txAp->fillpattern->ffg = txt_attr.ffg;
            txAp->fillpattern->fbg = txt_attr.fbg;
            txAp->fillpattern->lfg = txt_attr.lfg;
            txAp->fillpattern->lbg = txt_attr.lbg;
            txAp->fillcolor = fillcolor;
        }
    }
    txt->editState.disp = TRUE;   /* something is on-screen now */
    return((char *)0);
}


/* : dispTx - display an axtext instance */
char * /* NULL means ok */
dispTx(w, txt, clip)
long w;
axtext *txt;
r32 *clip;
{
	r32 *old_clip, newField;
	axtext *accum_txt;
	p32 accum;
	char *err;
    
#if BUILD_ORIENT && !FOR_FILTER
	CheckCharsAgainstAttrs(txt);
#endif

	/* display (or print) the characters, measuring as required */
	if ( TmAmSE ) /* we're printing */
	{
#if BUILD_KANJI && !FOR_FILTER
		CheckNarrowKanaAgainstAttrs(txt);
#endif
		/* if the text is unmeasured, format it (for printing) now */
		err = txt->measured ? (char *)0 : axtMeasuredText(txt);
		if ( err == (char *)0 )
		{
			/* draw shadows */
			if ( txt->draw_shadow ) /* draw dark shadow */
				seText(txt, clip, 2);
			if ( txt->draw_emboss ) /* draw sun shadow */
				seText(txt, clip, 1);

			/* draw text */
			if ( txt->shadow_only == FALSE )
				seText(txt, clip, 0);
		}
	}
	else /* not printing */
	{
		/* if the text is unmeasured, format it (for printing) now */
		err = txt->displayable ? (char *)0 : axtDisplayableText(txt);
		if ( err == (char *)0 )
		{
			/* capture the current clip, loading it into the relevant GC's */
			if ( clip )
			{
				txt->clipRect.origin = clip->origin;
				txt->clipRect.width  = clip->width;
				txt->clipRect.height = clip->height;
			}
			txtSetClip(txt, axTextGC, clip, (r32 *)0);
			txtSetClip(txt, axTextCursorGC, clip, (r32 *)0);
			txtSetClip(txt, axTextStencilGC, clip, (r32 *)0);

			/* init display start-point 'accum' */
			if ( !txt->old || txt->editState.editPos <= 0 )
				accum_txt = txt;
			else
				accum_txt = txt->old;
			if ( !accum_txt->items )
				accum = accum_txt->whole.pos;
			else
			{
				accum.x = accum_txt->whole.pos.x +
				          accum_txt->items[0].offset.x +
				          accum_txt->items[0].kern.x;
				accum.y = accum_txt->whole.pos.y +
				          accum_txt->items[0].offset.y +
				          accum_txt->items[0].kern.y;
			}

			/* lift the cursor */
			if ( txt->editState.inEdit )
				liftCursor(w, txt, (r32 *)0); /* ignore clip */
        
			/* lift old text */
			if ( txt->old )
			{
				/* compare current and last-displayed, lift non-current */
				if ( txt->old->clipRect.width <= 0 )
					old_clip = (r32 *)0;
				else if ( txt->old->clipRect.height <= 0 )
					old_clip = (r32 *)0;
				else
					old_clip = &txt->old->clipRect;
				txtSetClip(txt, axTextCursorGC, old_clip, (r32 *)0);
				txtSetClip(txt, axLiftGC, old_clip, (r32 *)0);
				liftOld(w, txt);
				axUnSetClip(axLiftGC, old_clip);
				txtSetClip(txt, axTextCursorGC, clip, (r32 *)0);
			}

			/* clear any portion of bounder that expanded since last display */
			if ( txt->editState.inEdit )
			{
				txtSetClip(txt, axLiftGC, clip, (r32 *) NULL);
				axtSnapshotFieldBounder(txt, &newField);
				if ( txt->editState.forceBGDraw )
					extendFieldBounder(txt, w, (r32 *)0);
				else
					extendFieldBounder(txt, w, &newField);
				txt->editState.fieldBounder = newField;
				axUnSetClip(axLiftGC, clip);
			}
    
			/* draw shadows */
			if ( txt->draw_shadow ) /* draw dark shadow */
				err = axtSubDispTx(w, txt, clip, accum, 2);
			if ( txt->draw_emboss && err == (char *)0 ) /* draw sun shadow */
				err = axtSubDispTx(w, txt, clip, accum, 1);

			/* draw text */
			if ( txt->shadow_only == FALSE && err == (char *)0 )
			{
				err = axtSubDispTx(w, txt, clip, accum, 0);
				if ( !err )
				{
					/* reset the per-character info that pertains to edits */
					resetPerChars(txt);
            
					/* toss the old copy of the text */
					if ( txt->old )
					{
						daxtext(txt->old); /* toss the copy */
						txfree((char *) txt->old);
						txt->old = (axtext *)NULL;
						txt->editState.editPos = 0;
					}

					/* draw cursor */
					if ( txt->editState.inEdit )
					{
						axtDrawTextCursor(w, txt, (r32 *)0,
						                  txt->editState.cursorPos,
						                  txt->editState.cursorType,
						                  txt->editState.cursorActive);
					}
				}
			}
    
			/* done */
			axUnSetClip(axTextStencilGC, clip);
			axUnSetClip(axTextCursorGC, clip);
			axUnSetClip(axTextGC, clip);
		}
	}
    return(err);
}

/* : getOverallBounder - return the 'ink bounder' for the given
 * range of text characters on *tx */
char * /* NULL means ok */
getOverallBounder(txt, sindex, len, bbox)
axtext *txt;
int sindex;
int len;
r32 *bbox; /* fill this in */
{
    int ix, tmp, endix, lpos;
    int segix, charix, charsInSeg;
    txAttr *txAp;
    perChar *pc;
    p32 curpoint;
    int minx, miny, maxx, maxy;
    int tx, ty, wid, ht;
    int anyVisibleChars = FALSE;
    char *errString;
    
    if (!txt)
        return (getErrString(XLT("The text pointer was null.."),
                        (char *)NULL));

    bbox->origin.x = bbox->origin.y = 0;
    bbox->width = bbox->height = 0;

    if (!txt->whole.nChars) {
        return ((char *)NULL);
    }

    if (!txt->measured) {
        if(errString = axtMeasuredText(txt)) {
            return errString;
        }
    }
    if (len == -1) {
        len = txt->whole.nChars - sindex;
    }
    
    txAp = attrSegAtPos(txt, sindex, &lpos);
    segix = txAp - txt->atts;
    charix = sindex;
    endix = sindex + len;

    curpoint.x = curpoint.y = 0;

    miny = minx = (INT_MAX >> 1);         /* very large */
    maxx = maxy = - (INT_MAX >> 1);
    
    while (charix < endix) {

        txAp = &txt->atts[segix++];
        charsInSeg = txAp->nChars - lpos;
        lpos = 0;
        
        for (ix = 0 ; ix < charsInSeg && charix < endix; ix++) {
            pc = &txt->items[charix];

            if ((charix > sindex) || !sindex) {
                    /* (these offsets are excluded for the 0th
                     *  char in the passed range, but not if that
                     *  char is the 0th in the entire text instance)
                     */
                curpoint.x += pc->offset.x;
                curpoint.y += pc->offset.y;
                curpoint.x += pc->kern.x;
                curpoint.y += pc->kern.y;
            }
            
            if (txAp->indent && axtHasBullet(txt, txAp, pc, charix)) {
                r32 bRect;
                p32 tpt;

                /* include dimensions of bullet
                 */
                tpt.x = curpoint.x + txt->whole.pos.x;
                tpt.y = curpoint.y + txt->whole.pos.y;
                axtBulletRect(txt, txAp, &tpt, &bRect);
                if (bRect.width && bRect.height) {
                    /* bRect's origin contains txt's overall origin.
                     * remove from this 0-relative measurement.
                     */
                    bRect.origin.x -= (txt->whole.pos.x);
                    bRect.origin.y -= (txt->whole.pos.y);
                    
                    if (bRect.origin.x < minx)
                        minx = bRect.origin.x;
                    if (bRect.origin.y < miny)
                        miny = bRect.origin.y;
                    if ((tmp = bRect.origin.x + bRect.width - 1) > maxx)
                        maxx = tmp;
                    if ((tmp = bRect.origin.y + bRect.height - 1) > maxy)
                        maxy = tmp;
                }
            }
            
            tx = pc->orig.x + pc->sup.x + curpoint.x;
            ty = pc->orig.y + pc->sup.y + curpoint.y;
            wid = pc->width;
            ht = pc->height;
            if (txAp->ruling || txAp->strike_thru) { 
                adjustRectForRuling(txAp, pc, &tx, &ty, &wid, &ht);
            }
            if (!pc->hidden) {
                anyVisibleChars = TRUE;
                if (tx < minx)
                    minx = tx;
                if (ty < miny)
                    miny = ty;
                if ((tmp = tx + wid - 1) > maxx)
                    maxx = tmp;
                if ((tmp = ty + ht - 1) > maxy)
                    maxy = tmp;
            }
            curpoint.x += pc->advance.x;
            curpoint.y += pc->advance.y;
            charix++;
        }
    }

    if (anyVisibleChars) {
        bbox->origin.x = minx;
        bbox->origin.y = miny;
        bbox->width = maxx - minx + 1;
        bbox->height = maxy - miny + 1;
    }

    return ((char *)NULL);
}

/*********** Note that this measure ignores any
 *********** transform that might be in effect!! */
char * /* NULL means ok */
getMaxBounds(tx, sindex, len, bbox)
axtext *tx;
int sindex;
int len;
r32 *bbox; /* fill this in */
{
    int tmp;
    txAttr *txAp, *first, *last;
    int minx, miny, maxx, maxy;
    char *errString;
    
    if (!tx)
        return (getErrString(XLT("The text pointer was null.."),
                        (char *)NULL));
    if (!tx->whole.nChars) {
        bbox->origin.x = bbox->origin.y = bbox->width = bbox->height = 0;
        return ((char *)NULL);
    }
    if (!tx->measured) {
        if(errString = axtMeasuredText(tx)) {
            return errString;
        }
    }
    first = attrSegAtPos(tx, sindex, &tmp);
    last = attrSegAtPos(tx, sindex+len, &tmp);
    minx = first->fontAttr.bbox.origin.x;
    miny = first->fontAttr.bbox.origin.y;
    maxx = minx + first->fontAttr.bbox.width - 1;
    maxy = miny + first->fontAttr.bbox.height - 1;
    for(txAp=first+1; txAp<=last; txAp++) {
        if (txAp->fontAttr.bbox.origin.x < minx)
            minx = txAp->fontAttr.bbox.origin.x;
        if (txAp->fontAttr.bbox.origin.y < miny)
            miny = txAp->fontAttr.bbox.origin.y;
        tmp = txAp->fontAttr.bbox.origin.x + txAp->fontAttr.bbox.width - 1;
        if (tmp > maxx)
            maxx = tmp;
        tmp = txAp->fontAttr.bbox.origin.y + txAp->fontAttr.bbox.height - 1;
        if (tmp > maxy)
            maxy = tmp;
    }
    bbox->origin.x = minx;
    bbox->origin.y = miny;
    bbox->width = maxx - minx + 1;
    bbox->height = maxy - miny + 1;
    return ((char *)NULL);
}

axtGetSize(txt, wid, hyt)
axtext *txt;
int *wid;
int *hyt;
{
    int i, n, xpos, minx, maxx, ypos, miny, maxy;
    perChar *pc;

    /* accumulate beginning and ending points of all characters */
    n = txt->whole.nChars;
    pc = txt->items;
    for ( ypos=miny=maxy=xpos=minx=maxx=i=0; i<n; i++ )
    {
        /* x setpoint */
        if ( !txt->whole.widthBounder || i != 0 )
		{
            xpos += pc[i].offset.x + pc[i].kern.x;

			/* y setpoint */
			ypos += pc[i].offset.y + pc[i].kern.y;
		}

        if ( xpos > maxx )
            maxx = xpos;
        if ( xpos < minx )
            minx = xpos;
        if ( ypos > maxy )
            maxy = ypos;
        if ( ypos < miny )
            miny = ypos;

        /* x end of character */
        xpos += pc[i].advance.x;

        /* y end of character */
        ypos += pc[i].advance.y;

        if ( xpos > maxx )
            maxx = xpos;
        if ( xpos < minx )
            minx = xpos;

        if ( ypos > maxy )
            maxy = ypos;
        if ( ypos < miny )
            miny = ypos;
    }

    /* done */
    *wid = maxx - minx;
    *hyt = maxy - miny;
}
int
axtGetSimpleOffset(txt, ix1, ix2, old_offset)
axtext *txt;
int ix1, ix2, old_offset;
{
    int i, new_offset;
    perChar *pc;

    new_offset = old_offset;
    pc = &txt->items[0];
    for ( i=ix1; i<ix2; i++ )
        new_offset += pc[i].offset.y;
    return(new_offset);
}

/* : insChars - inserts len# characters into *txt at index pos */
int
insChars(txt, str, pos, len)
axtext *txt;
tx_char *str;
int pos;
int len;
{
    int tmp;
    txAttr *txAp;
    perChar *pc = (perChar *)NULL;
    if (!txt)
        return FALSE;
    if (len == 0 || pos < 0 || pos > txt->whole.nChars) {
        return FALSE;
    }
    else if (pos < txt->editState.editPos) {
            /* optimization restriction -- see axtxint.h: editPos */
        return FALSE;
    }
    if (txt->editState.disp) {
        if (!txt->old) {
            /* first edit since last display -- save state.
             */
            int epos = pos;
            if (str && len == 1 && str[0] == TABCHAR) {
                /* a very cheap n quick test for tab insertion;
                 * axtEditPos should obviate this, but a safe hack anyway..
                 */
                epos = 0;
            }
            axtSnapshotText(txt, epos);
        }
    }
    if (!txt->constructed) {
        axtDefaultText(txt);
    }
        /* insert the tx_char into the textstr */
    if (!insertItems(TYPE_W_CHAR, pos, (char *)str, len,
                (char **)&txt->textstr, txt->whole.nChars,
                &txt->wCharBuflen, txt->pid)) {
        return FALSE;
    }
        /* insert empty perChars in position */
    tmp = sizeof(perChar) * len;
    pc = (perChar *)txloc(txt->pid, tmp);
    memset((char *) pc, 0, tmp);
    if (!insertItems(TYPE_PERCHAR, pos, (char *)pc, len,
            (char **)&txt->items, txt->whole.nChars,
            &txt->perCharBuflen, txt->pid)) {
        if (pc)
            txfree((char *)pc);
        return FALSE;
    }
    if (pc)
        txfree((char *)pc);
    if (txt->editState.disp) {
            /* on the current copy, mark the new chars as inserted */
        for(tmp=pos; tmp<pos+len; tmp++) {
            txt->items[tmp].inserted = TRUE;
        }
    }
            /* assign proper attr seg charcount */
    txAp = inheritedAttr(txt, pos);
    if (txAp) {
        txAp->nChars += len;
    }
    txt->whole.nChars += len;

    txt->editState.cursorPos += len;
    if (txt->editState.cursorPos >= txt->whole.nChars)
        txt->editState.cursorPos = txt->whole.nChars;
    
    txt->measured = txt->displayable = FALSE;
    
    return TRUE;
}

/* : delChars - deletes len# text characters from *txt at index pos */
int
delChars(txt, pos, len)
axtext *txt;
int pos;
int len;
{
    int ix, tmp, oldix;
    txAttr *txAp;

    if (!txt)
        return FALSE;
    if (len == 0 || pos < 0 || (pos+len) > txt->whole.nChars) {
        return FALSE;
    }
    else if (pos < txt->editState.editPos) {
            /* optimization restriction -- see axtxint.h: editPos */
        return FALSE;
    }
    if (!txt->items) {
        return FALSE;
    }
    if (txt->editState.disp) { 
        if (!txt->old) {
            int tpos, tlimit;
                /* first edit since last display -- save state
                 */
            axtSnapshotText(txt, pos);
            
                /* indicate deletion on old copy
                 */
            tpos = pos - txt->editState.editPos;
            tlimit = pos + len - txt->editState.editPos;
            for (ix = tpos; ix < tlimit; ix++) {
                txt->old->items[ix].deleted = TRUE;
            }
        }
        else {
            int fx = -1, lx = -2;
                /* indicate the deletion on the old copy - hard way
                   (since the deleted bit itself figures in finding
                   the old correspondent, collect sites first then
                   mark as deleted "atomically") */
            tmp = pos + len;
            for(ix = pos; ix < tmp; ix++) {
                oldix = getOldIndex(txt, txt->old, ix);
                if (ix == pos) {
                    fx = oldix;  /* 1st affected in range */
                }
                if (ix == tmp - 1) {
                    lx = oldix;  /* last affected in range */
                }
                if (oldix >= 0 && oldix < txt->old->whole.nChars) {
                        /* mark 'temp' member to collect deleted chars */
                    txt->old->items[oldix].temp = TRUE;
                }
            }
            for (ix = fx; ix <= lx; ix++) {
                    /* turn temp's into deleted's */
                if (txt->old->items[ix].temp) {
                    txt->old->items[ix].temp = 0; /* leave clean */
                    txt->old->items[ix].deleted = TRUE;
                }
            }
        }
    }
    if (!txt->constructed) {
        axtDefaultText(txt);
    }
    if (!deleteItems(TYPE_W_CHAR, pos, len, (char **)&txt->textstr,
                txt->whole.nChars, &txt->wCharBuflen, txt->pid)) {
        return FALSE;
    }
        /* delete the corresponding perChar struct */
    if (txt->items) {
/********* the number of perChars and tx_chars is equal in
         * this implementation so far -- may need to change to
         * handle location-dependent scripts, where perhaps one
         * perChar will refer to multiple letterforms.... */
        if (!deleteItems(TYPE_PERCHAR, pos, len, (char **)&txt->items,
                txt->whole.nChars, &txt->perCharBuflen, txt->pid)) {
            return FALSE;
        }
    }

    txAp = attrSegAtPos(txt, pos, &tmp);
    if (txAp) {
            /* assign new attr seg charcounts */
        int numLeft;
        numLeft = len;
        while (numLeft) {
            if ((txAp->nChars - tmp) < numLeft) {
                numLeft -= (txAp->nChars - tmp);
                txAp->nChars = tmp;
                txAp++;
            }
            else {
                txAp->nChars -= numLeft;
                break;
            }
            tmp = 0;  /* only 1st segment is affected by this offset */
        }
    }
    txt->whole.nChars -= len;
    txt->editState.cursorPos -= len;
    if ( txt->editState.cursorPos < 0 )
        txt->editState.cursorPos = 0;

        /* the delete may have brought two identical attr segs together */
    checkJoin(txt, pos-1, 2);

    txt->measured = txt->displayable = FALSE;
    return TRUE;
}

/* : setTextAttrs - applies one or more attrs to the *txt text character
 * range specified by index pos and length len */
int
setTextAttrs(txt, attr, flags, pos, len)
axtext *txt; /* all params in */
textAttrs *attr;
unsigned int flags; /* for example, TXpointsize | TXfontFamily | ... */
int pos, len;
{
    int fpos, lpos, tmp;
    int ok = TRUE;
    int firstSplit = FALSE;
    int lastSplit = FALSE;
    int firstChanged = FALSE;
    int lastChanged = FALSE;
    unsigned int tflags;
    txAttr *first, *last; /* of those affected */
    txAttr txDefault, *defp;
    txAttr tnew, newAttr, *tp;

    if (!txt || !attr)
        return FALSE;
    if (len==-1) {
        len=txt->whole.nChars - pos;
    }
    if (pos < 0 || pos + len > txt->whole.nChars) {
        return FALSE;
    }
    else if (pos < txt->editState.editPos) {
            /* optimization restriction -- see axtxint.h: editPos */
        return FALSE;
    }
    if (txt->editState.disp) { 
        if (!txt->old) {
            /* first edit since last display -- save state.
             */
            axtSnapshotText(txt, pos);
        }
        if (txt->items) {
                /* indicate on the current chars that their
                   appearance has changed from their on-screen image */
            for(tmp = pos; tmp < pos + len; tmp++) {
                txt->items[tmp].changed = TRUE;
            }
        }
    }

    txAttDefaults(&txDefault, &tflags, txt->pid); /* get hard defaults */
    if (txt->userDefaults) {
        defp = txt->userDefaults;
    }
    else {
        defp = &txDefault;
    }
        /* get private copy of public attr struct */
    memset((char *)&tnew, 0, sizeof(txAttr));
    memset((char *)&newAttr, 0, sizeof(txAttr));
    copyInTextAttr(attr, &tnew, defp, flags, txt->pid);
        /* complete newAttr by folding in txDefault info */
    foldDefaults(flags, &tnew, defp, &newAttr, txt->pid);
        /* put newAttr onto *txt */
    if (!txt->constructed) {
        axtDefaultText(txt);
    }
        /* first and last contain the endpoints of the range */
    first = attrSegAtPos(txt, pos, &fpos);
    last = attrSegAtPos(txt, pos+len-1, &lpos);
        /* before possibly splitting the first & last affected,
         * apply whatever changes to the attr segs in between. */
    for(tp=first; tp<=last; tp++) {
        if (tp!=first && tp!=last) {
            attrChange(tp, &newAttr, flags, TRUE, txt->pid); /* change if needed */
        }
    }
        /* now look at the endpoints */
    if (attrChange(last, &newAttr, flags, FALSE, txt->pid)) {
        if (needSplit(txt, pos+len-1, RIGHT_EDGE)) {
                /* split right-edge attr */
            if(!doSplit(txt, pos+len)) {
                ok = FALSE;
                goto allDone;
            }
            lastSplit = TRUE;
        }
        lastChanged = TRUE;
    }
    if (lastSplit) {
        /* reset first since above split may have readdressed things */
        first = attrSegAtPos(txt, pos, &fpos);
    }
    if (attrChange(first, &newAttr, flags, FALSE, txt->pid)) {
        if (needSplit(txt, pos, LEFT_EDGE)) {
            if(!doSplit(txt, pos)) {
                ok =  FALSE;
                goto allDone;
            }
            firstSplit = TRUE;
        }
        firstChanged = TRUE;
    }
    if (firstChanged || lastChanged) {
            /* apply the new attr info to the endpoint attrs */
        first = attrSegAtPos(txt, pos, &fpos);
        last = attrSegAtPos(txt, pos+len-1, &lpos);
        attrChange(first, &newAttr, flags, TRUE, txt->pid);
        if (last != first)
            attrChange(last, &newAttr, flags, TRUE, txt->pid);
    }
        /* scan the affected range for joins */
    if(!checkJoin(txt, pos, len)) {
        ok = FALSE;
        goto allDone;
    }
allDone:
    if (affectsLetterforms(flags)) {
        if (txt->items) {
            for(lpos=pos; lpos<pos+len; lpos++) {
                    /* mark perChar's in affected range */
                txt->items[lpos].targetMetrics = FALSE;
                txt->items[lpos].screenMetrics = FALSE;
                txt->items[lpos].letterNameKnown = FALSE;
            }
        }
        txt->measured = FALSE;
    }
    if (flags & TXindent) {
        txt->measured = FALSE;     /* this may move things */
    }
    txt->displayable = FALSE;
    tossTxAttr((char *) &newAttr);
    tossTxAttr((char *) &tnew);
    tossTxAttr((char *) &txDefault);
    
    return ok;
}

/* : setDefaultTextAttrs - establish a thorough txt->userDefaults
 * text attribute description on a new *txt instance */
int
setDefaultTextAttrs(txt, attr, flags)
axtext *txt; /* all params in */
textAttrs *attr;
unsigned int flags;
{
    txAttr tnew, newAttr, txDefault;
    int ix;
    unsigned int tflags;
    
    if (!txt || !attr)
        return FALSE;
    
	/* get hard defaults */
    txAttDefaults(&txDefault, &tflags, txt->pid);

	/* get private copy of public attr struct */
    memset((char *)&tnew, 0, sizeof(txAttr));
    memset((char *)&newAttr, 0, sizeof(txAttr));
    copyInTextAttr(attr, &tnew, &txDefault, flags, txt->pid);
    
    
        /* mix with user-supplied onto newAttr
         */
    foldDefaults(flags, &tnew, &txDefault, &newAttr, txt->pid);
    
        /* attach to *txt
         */
    if (txt->userDefaults) {
        tossTxAttr((char *) txt->userDefaults);
        txfree(txt->userDefaults);
    }
    txt->userDefaults = (txAttr *) txloc(txt->pid, sizeof(txAttr));
    memset((char *)txt->userDefaults, 0, sizeof(txAttr));
    copyTxAttr(txt->userDefaults, &newAttr, txt->pid);
    txt->userDefaultFlags = flags;
    if (txt->items) { /* probably not */
        for (ix = 0; ix < txt->whole.nChars; ix++) {
            txt->items[ix].targetMetrics = FALSE;
        }
    }
    txt->displayable = txt->measured = txt->constructed = FALSE;
    tossTxAttr((char *) &tnew);
    tossTxAttr((char *) &newAttr);
    tossTxAttr((char *) &txDefault);
    return TRUE;
}

/* : getDefaultTextAttr
 * report complete defaults onto *attr.  Note that
 * *attr should be in good shape -- either clear, or
 * with real data on it.  As this function copies onto
 * it, it will try to free any string pointers it finds. */
int
getDefaultTextAttr(txt, attr)
axtext *txt; /* in */
textAttrs *attr; /* out */
{
    unsigned int flags = (unsigned int) TXAllAtts;
    if (!txt || !attr)
        return FALSE;
    copyOutTextAttr(attr, txt->userDefaults, flags, txt->pid);
    return TRUE;
}
int
axtGetDefTextAttr(txt, attr)
axtext *txt;
textAttrs **attr; /* out */
{
    *attr = (textAttrs *) txloc(txt->pid, sizeof(textAttrs));
    memset((char *)*attr, 0, sizeof(textAttrs));
    if ( getDefaultTextAttr(txt, *attr) == FALSE ) {
        txfree((char *)*attr);
        *attr = (textAttrs *)NULL;
        return FALSE;
    }
    return TRUE;
}

/* : getTextAttrs
 * get textAttr array fully describing attr info for
 * selected range of *txt;  Note that caller must free
 * the resultant array, and the textAttrs have strings
 * attached.. */
int
getTextAttrs(txt, pos, len, attrs, attrLen)
axtext *txt;
int pos, len;
textAttrs **attrs; /* out */
int *attrLen;     /* out */
{
    unsigned int flags = (unsigned int) TXAllAtts;
    int lpos, segcount, firstix, lastix, i, j;
    txAttr *first, *last, txDefault, txFullDefault;
    if (!txt || !attrs || !attrLen) {
        *attrs = (textAttrs *)NULL;
        *attrLen = 0;
        return FALSE;
    }
    if (len==-1) {
        len=txt->whole.nChars - pos;
    }
    else if (len == 0) {
        len = 1;
    }
    if (txt->atts) {
        first = attrSegAtPos(txt, pos, &lpos);
        last = attrSegAtPos(txt, pos+len-1, &lpos);
        segcount = (last - first) + 1;
        *attrs = (textAttrs *) txloc(txt->pid, segcount * sizeof(textAttrs));
        memset((char *)*attrs, 0, segcount * sizeof(textAttrs));
        firstix = first - txt->atts;
        lastix = last - txt->atts;
        for(i=firstix, j=0; i<=lastix; i++, j++) {
            copyOutTextAttr(*attrs + j, &txt->atts[i], flags, txt->pid);
            (*attrs)[j].numChars = txt->atts[i].nChars;
        }
        *attrLen = segcount;
    }
    else {
            /* no explicit attrs; report defaults */
        memset((char *)&txDefault, 0, sizeof(txAttr));
                /* get hard defaults */
        txAttDefaults(&txDefault, &flags, txt->pid);
        if (txt->userDefaults) {
                /* fold in user defaults */
            memset((char *)&txFullDefault, 0, sizeof(txAttr));
            foldDefaults(txt->userDefaultFlags, txt->userDefaults,
                        &txDefault, &txFullDefault, txt->pid);
            *attrs = (textAttrs *) txloc(txt->pid,sizeof(textAttrs));
            memset((char *)*attrs, 0, sizeof(textAttrs));
            copyOutTextAttr(*attrs, &txFullDefault, flags, txt->pid);
            tossTxAttr((char *) &txFullDefault);
        }
        tossTxAttr((char *) &txDefault);
        *attrLen = 1;
    }
    return TRUE;
}

/* : getHardcodedDefaults - return internal hardcoded defaults
 * onto passed textAttrs instance.  the passed instance should
 * be clean or otherwise properly initialized -- copyOut will
 * free nonzero heap addrs that it finds there
 */
int
getHardcodedDefaults(attrs, pid)
textAttrs *attrs; /* out: internal defaults in textAttr form */
int pid;
{
    txAttr txDefault;
    unsigned int flags = (unsigned int) TXAllAtts;
    memset((char *)&txDefault, 0, sizeof(txAttr));
    txAttDefaults(&txDefault, &flags, pid);
    copyOutTextAttr(attrs, &txDefault, flags, pid);
    tossTxAttr((char *) &txDefault);
    return TRUE;
}

/* : getText
 * Return the complete textstr of *txt
 * Note that the text characters come back as tx_char's.. */
int
getText(txt, charcodes, len)
axtext *txt;
tx_char **charcodes; /* out -- caller must free */
int *len;           /* out -- number of tx_chars returned */
{
    int i, tlen;
    if (!txt || !charcodes || !len)
        return FALSE;
    tlen = txt->whole.nChars;
    if (tlen > 0) {
        *charcodes = (tx_char *) txloc(txt->pid, tlen * sizeof(tx_char));
        for (i=0; i<tlen; i++) {
            (*charcodes)[i] = txt->textstr[i];
        }
    }
    else
        *charcodes = (tx_char *)NULL;
    *len = tlen;
    return TRUE;
}

/* : getTextFromIndex
 * Return the partial textstr of *txt
 * Note that the text characters come back as tx_char's.. */
int
getTextFromIndex(txt, charcodes, index, len)
axtext *txt;
tx_char **charcodes; /* out -- caller must free */
int index;
int len;
{
    int i, j;

    *charcodes = (tx_char *)NULL;
    if (!txt || !charcodes || !len)
        return FALSE;

	if ( index < 0 || index > txt->whole.nChars ) 
		index = 0;
	if ( len <= 0 )
		len = txt->whole.nChars;

    *charcodes = (tx_char *) txloc(txt->pid, len * sizeof(tx_char));
    for (j=0, i=index; j<len; i++, j++)
		(*charcodes)[j] = txt->textstr[i];
    return TRUE;
}

/* : axtTextEditCleared - used when text-under-edit
 * has been lifted from screen by something other
 * than the axtext code.  This func lets the axtext
 * instance know that it's no longer on-screen
 */
void
axtTextEditCleared(txt)
axtext *txt;
{
    if (!txt || !txt->editState.inEdit)
        return;
    if (txt->old) {
        daxtext(txt->old); /* toss old state */
        txfree((char *) txt->old);
        txt->old = (axtext *) NULL;
    }
    txt->editState.cursorOn = FALSE;
    txt->editState.disp = FALSE;
    txt->editState.fieldBounder.width = 0;
}

void
axtClearOldText(txt)
axtext *txt;
{
    if (!txt )
        return;
    
    if (txt->old) {
        daxtext(txt->old); /* toss old state */
        txfree((char *) txt->old);
        txt->old = (axtext *) NULL;
    }
    txt->editState.cursorOn = FALSE;
    txt->editState.disp = FALSE;
    txt->editState.editPos = 0;
}

/* : axtStartEdit - enable text cursor display, etablish optional
 * background pixmap for text lifting.
 */
void
axtStartEdit(txt, background, backRect, color)
axtext *txt;       /* in */
Pixmap background; /* in: lift text by copying from this */
r32 *backRect;     /* in: expresses offset within passed pixmap */
color_type *color; /* in: use this for plain clears if no pixmap */
{
	unsigned long pel;
	if ( txt && txt->editState.inEdit == FALSE )
	{
		txt->editState.cursorType = CURSOR_TRIANGLE;
		txt->editState.cursorActive = ACTIVE;
		txt->editState.background = background;
		if ( backRect )
			txt->editState.backRect = *backRect; /* in screen pix */
		if ( color && color->see_thru == FALSE ) 
		{
			AxGetRealPixel(2, color->c,color->m,color->y,color->k, 255, &pel);
    		txt->editState.backgroundPixel = pel;
			txt->editState.javaPixel = 0xff000000 +
			 (((255 - (color->c + color->k))<<16)&0x00ff0000) +
			 (((255 - (color->m + color->k))<<8)&0x0000ff00) +
			  ((255 - (color->y + color->k))&0x000000ff);
		}
		else
		{
    		txt->editState.backgroundPixel = AlisPixels[WHITE];
			txt->editState.javaPixel = 0xffffffff;
		}
		txt->editState.inEdit = TRUE;
	}
}

void
freeTxString(textStr)
tx_char *textStr;
{
    if (textStr) {
        txfree((char *) textStr);
    }
}

void
freeTextAttr(atts)
textAttrs *atts;
{
	if (atts->fontFamily)
		txfree((char *) atts->fontFamily);
	if (atts->strokepattern)
		axtFreeFillPattern(atts->strokepattern);
	if (atts->fillpattern)
		axtFreeFillPattern(atts->fillpattern);
	if (atts->indent)
		axtFreeIndent(atts->indent);
	if (atts->shadow)
		axtFreeDropShadow(atts->shadow);
	if (atts->bullet)
		axtFreeBullet(atts->bullet);
}

/* : freeTextAttrs - free up an array of textAttr structs returned
 * by getTextAttrs */
void
freeTextAttrs(atts, len)
textAttrs *atts;
int len;
{
    int i;
    if (!atts)
        return;
    for (i=0; i< len; i++)
    {
		freeTextAttr(&atts[i]);
    }
    txfree((textAttrs *) atts);
}

/* : axtSetLineHeight
 */
void
axtSetLineHeight(tx, lineHeight)
axtext *tx;
int lineHeight;
{
    if (!tx)
        return;
    if (tx->whole.lineHeight != lineHeight) {
        if (tx->editState.disp) {
            if (!tx->old) {
                /* first edit since last display -- save state.
                 * setLineHeight is effectively an edit at pos 0.
                 */
                axtSnapshotText(tx, 0);
            }
            axtMoveItems(tx); /* lineheight change affects all chars */
        }
        tx->whole.lineHeight = lineHeight;
        tx->measured = tx->displayable = FALSE;
        
    }
}

void
axtSetZoom(tx, xZoom, yZoom)
axtext *tx;
int xZoom, yZoom;  /* where 100% zoom (unzoomed) is 1000 */
{
    int changed = FALSE;

    if (xZoom != tx->whole.xZoom) {
        changed = TRUE;
    }
    if (yZoom != tx->whole.yZoom) {
        changed = TRUE;
    }
    
    if (changed) {
        p32 txpos;
        
        if (tx->whole.widthBounder) {
            axtGetPos(tx, &txpos);     /* remember "app" position */
        }
        
        tx->whole.xZoom = xZoom;
        tx->whole.yZoom = yZoom;
        
        if (tx->whole.widthBounder) {
            setPos(tx, &txpos);        /* write "app" position */
            axtRemoveFieldWidth(tx);   /* remove whole.bounder */
            axtSetFieldWidth(tx, tx->whole.width);  /* request new one */
        }

#if BUILD_ORIENT
        /* Ashisuto added for BoxText inset inside in the Spreadsheet,
         * unity in Graphics, but resized in Spreadsheet. Should use
         * DrawLetterForm, not the Quick draw.
         * 1996.05.27
         */
        if (xZoom != yZoom)
            tx->whole.transform.unity = 0;
#endif

        axtTransformChanged(tx);  /* toss state re: old transform */
    }
}

void
axtGetZoom(tx, xZoom, yZoom)
axtext *tx;
int *xZoom, *yZoom;  /* out: (100% zoom (unzoomed) returned as 1000) */
{
    if (!tx || !xZoom || !yZoom)
        return;

    *xZoom = tx->whole.xZoom;
    *yZoom = tx->whole.yZoom;
}

void
axtZoomBounder(tx, bbox)
axtext *tx;
r32 *bbox;
{
    double zoom;
    int val;

    if ( tx->whole.xZoom != 1000 )
    {
        zoom = 1000.0 / tx->whole.xZoom;
        val = (bbox->origin.x + (int) bbox->width - 1) * zoom;
        bbox->origin.x = bbox->origin.x * zoom;
        bbox->width = val - bbox->origin.x + 1;
    }
    if ( tx->whole.yZoom != 1000 )
    {
        zoom = 1000.0 / tx->whole.yZoom;
        val = (bbox->origin.y + (int) bbox->height - 1) * zoom;
        bbox->origin.y = bbox->origin.y * zoom;
        bbox->height = val - bbox->origin.y + 1;
    }
}

void
axtSetFieldTransform(tx, atts, attflags)
axtext *tx;
textAttrs *atts;
unsigned int attflags;
{
    axformTerms tterms;
    p32 txpos;
    int changed = FALSE;
    
    if (!tx || !atts)
        return;
    
    tterms = tx->whole.transform;
    
    if (attflags & TXangle) {
        if (tx->whole.transform.angle != atts->angle) {
            changed = TRUE;
            tterms.angle = atts->angle;
        }
    }
    if (attflags & TXxscale) {
        if (tx->whole.transform.xscale != atts->xscale) {
            changed = TRUE;
            tterms.xscale = atts->xscale;
        }
    }
    if (attflags & TXyscale) {
        if (tx->whole.transform.yscale != atts->yscale) {
            changed = TRUE;
            tterms.yscale = atts->yscale;
        }
    }
    if (attflags & TXxshear) {
        if (tx->whole.transform.xshear != atts->xshear) {
            changed = TRUE;
            tterms.xshear = atts->xshear;
        }
    }
    
    if (!changed)
        return;

    if (tx->whole.widthBounder) {
        axtGetPos(tx, &txpos);     /* remember "app" position */
    }
    
    if (!(tterms.xscale && tterms.yscale)) {
        return;
    }
    else {
        tx->whole.transform = tterms;
    }
    
    if (!tx->whole.transform.angle &&
        !tx->whole.transform.xshear &&
        tx->whole.transform.xscale == 10000 &&
        tx->whole.transform.yscale == 10000) {
        tx->whole.transform.unity = TRUE;
    }
    else {
        tx->whole.transform.unity = FALSE;
    }
    if (tx->whole.transform.angle || tx->whole.transform.xshear) {
        tx->editState.fullDisp |= COMPLEX_XFORM;
    }
    
    if (tx->whole.widthBounder) {
        setPos(tx, &txpos);                       /* write "app" position */
        axtRemoveFieldWidth(tx);                  /* remove whole.bounder */
        axtSetFieldWidth(tx, tx->whole.width);    /* request new one */
    }
    
    if (changed) {
        axtTransformChanged(tx);  /* toss state re: old transform */
    }
}

    /* update state to indicate a completely new transform */
static void
axtTransformChanged(tx)
axtext *tx;
{
    int charcount;
    txAttr *txAp;
    
    if (tx->editState.disp) { 
        if (!tx->old) {
            /* first edit since last display -- save state.
             * xform-changed is effectively an edit at pos 0.
             */
            axtSnapshotText(tx, 0);
        }
    }
    if (tx->items) {
        charcount = tx->whole.nChars - 1;
        while(charcount >= 0) {
            tx->items[charcount].targetMetrics = FALSE;
            tx->items[charcount].screenMetrics = FALSE;
            tx->items[charcount].letterNameKnown = FALSE;
            if (tx->editState.disp)
                tx->items[charcount].changed = TRUE;
            charcount--;
        }
    }
    tx->measured = tx->displayable = FALSE;
    charcount = tx->whole.nChars;
    txAp = tx->atts;
    while(charcount) {
        txAp->tctm.known = FALSE; /* composite xform unknown now */
        if (txAp->screenFontName) {
            txfree((char *)txAp->screenFontName);
            txAp->screenFontName = (char *) NULL;
        }
        charcount -= txAp->nChars;
        txAp++;
    }
}

void
axtGetFieldTransform(tx, atts)
axtext *tx;
textAttrs *atts;
{
    if (!tx || !atts)
        return;
    atts->xscale = tx->whole.transform.xscale;
    atts->yscale = tx->whole.transform.yscale;
    atts->xshear = tx->whole.transform.xshear;
    atts->angle = tx->whole.transform.angle;
}

/* : axtSetBounder - associate a bounder polygon with an
 * axtext instance; when a bounder is present, the text
 * will be formatted to stay inside it */
void
axtSetBounder(tx, bounder, leftOffset, rightOffset,
    topOffset, botOffset, lineFormatter, getIntercepts)
axtext *tx;
poly *bounder;            /* copy in if nonnull;  leave existing
                             bounder alone if parameter is null.. */
int leftOffset, rightOffset; /* applied 'inward' to bounder intercepts */
int topOffset, botOffset; /* applied 'inward' to bounder maxy, miny */
int (*lineFormatter)();
int (*getIntercepts)();
{
    if (!tx)
        return;
    if (tx->editState.disp) {
        if (!tx->old) {
            /* first edit since last display -- save state.
             * setBounder is effectively an edit at pos 0.
             */
            axtSnapshotText(tx, 0);
        }
        axtMoveItems(tx); /* bounder change affects all chars */
    }
    if (bounder && tx->whole.bounder.numPaths) {
        freeBounder(&tx->whole.bounder);
        memset((void *) &tx->whole.bounder, 0, sizeof(poly));
    }
    tx->whole.topOffset = topOffset;
    tx->whole.botOffset = botOffset;
    tx->whole.leftOffset = leftOffset;
    tx->whole.rightOffset = rightOffset;
    if (bounder && bounder->numPaths) {

            /* copy in new bounder data; convert from spline
               to polyline while doing so */
        int ix, jx, miny, maxy, *n_pnts, size;
        poly_type *fixed_paths;
        TmPoint32 **pnt;

        fixed_paths = (poly_type *) txloc(tx->pid,
            bounder->numPaths * sizeof(poly_type));
        axtFixPaths(bounder->numPaths, bounder->paths, fixed_paths, tx->pid);

        n_pnts = (int *) txloc(tx->pid, bounder->numPaths * sizeof(int));
        pnt = (TmPoint32 **) txloc(tx->pid, bounder->numPaths *
                                    sizeof(TmPoint32 *));
        convertPols(bounder->numPaths, fixed_paths, n_pnts, pnt,
                    FALSE);    /* spline -> polyline */

        if (bounder->numPaths > 0 ) {
            for (ix = 0; ix < bounder->numPaths; ix++) {
                if ( fixed_paths[ix].points )
                    txfree(fixed_paths[ix].points);
                if ( fixed_paths[ix].controls )
                    txfree(fixed_paths[ix].controls);
            }
            txfree(fixed_paths);
        }

        /* make sure that the polys are closed */
        for (ix = 0; ix < bounder->numPaths; ix++) {
            if ( pnt[ix][0].x != pnt[ix][n_pnts[ix] - 1].x ||
                 pnt[ix][0].y != pnt[ix][n_pnts[ix] - 1].y ) {
                n_pnts[ix]++;
                size = n_pnts[ix] * sizeof(TmPoint32);
                if (!(pnt[ix]))
                  pnt[ix] = (TmPoint32 *) txloc(tx->pid, size);
                else
                  pnt[ix] = (TmPoint32 *) retxloc(tx->pid,pnt[ix], size);

                pnt[ix][n_pnts[ix] - 1].x = pnt[ix][0].x;
                pnt[ix][n_pnts[ix] - 1].y = pnt[ix][0].y;
            }
        }

            /* copy the poly info onto the text instance */
        tx->whole.bounder.numPaths = bounder->numPaths;
        tx->whole.bounder.paths = (poly_type *)
                txloc(tx->pid, bounder->numPaths * sizeof(poly_type));
        miny = MAXSHORT16;
        maxy = 0;
        for (ix = 0; ix < bounder->numPaths; ix++) {
            tx->whole.bounder.paths[ix].nPoints = n_pnts[ix];
            tx->whole.bounder.paths[ix].nCtrls = 0;
            if (n_pnts[ix]) {
                tx->whole.bounder.paths[ix].points = (TmPoint32 *)
                    txloc(tx->pid, n_pnts[ix] * sizeof(TmPoint32));
            }
            else {
                tx->whole.bounder.paths[ix].points = (TmPoint32 *) NULL;
            }
            tx->whole.bounder.paths[ix].controls = (TmPoint32 *) NULL;
            for (jx = 0; jx < n_pnts[ix]; jx++) {
                tx->whole.bounder.paths[ix].points[jx] = pnt[ix][jx];
                if (tx->whole.bounder.paths[ix].points[jx].y < miny)
                    miny = bounder->paths[ix].points[jx].y;
                if (tx->whole.bounder.paths[ix].points[jx].y > maxy)
                    maxy = bounder->paths[ix].points[jx].y;
            }
        }
        tx->whole.bounderMaxy = maxy;
        tx->whole.bounderMiny = miny;

        if (bounder->numPaths > 0 ) {
            for (ix = 0; ix < bounder->numPaths; ix++) {
                if (pnt[ix])
                    txfree((char *) pnt[ix]);
            }
            txfree((char *) n_pnts);
            txfree((char *) pnt);
        }
    }
    /* for the textbox, the bounder is always marked as
     * nonrectangular, regardless of how simple the shape.
     */
    tx->whole.bounder.rectangular = FALSE;
    
    tx->whole.getIntercepts = getIntercepts;
    tx->whole.lineFormatter = lineFormatter;
    tx->measured = tx->displayable = FALSE;
    
}

int
axtResetFieldWidth(tx, width)
axtext *tx;
int *width;
{
    int widthBounder;

    widthBounder =  tx->whole.widthBounder ? TRUE: FALSE;
    tx->whole.widthBounder = FALSE;
    *width = tx->whole.width;
    return (widthBounder);
}

axtUpdateBounder(tx, dx, dy)
axtext *tx;
int dx;
int dy;
{
	poly *bounder;
	int ix, jx;

	if (!tx)
		return;

	if ( tx->whole.widthBounder == FALSE && tx->whole.bounder.numPaths )
	{
        bounder = &tx->whole.bounder;
		for (ix = 0; ix < bounder->numPaths; ix++) {
			for (jx = 0; jx < bounder->paths[ix].nPoints; jx++) {
			    bounder->paths[ix].points[jx].x += dx;
			    bounder->paths[ix].points[jx].y += dy;
			}
			for (jx = 0; jx < bounder->paths[ix].nCtrls; jx++) {
			    bounder->paths[ix].controls[jx].x += dx;
			    bounder->paths[ix].controls[jx].y += dy;
			}
		}
		tx->measured = tx->displayable = FALSE;
	}
}
 
/* : axtSetFieldWidth - specify that the passed axtext instance's
 * text should be formatted to fit within the passed width.  This
 * function turns on the 'widthBounder' style of formatting; if the
 * passed width is 0, the bounding is unconstrained.
 */
void
axtSetFieldWidth(tx, width)
axtext *tx;
int width;
{
    if (!tx)
        return;
    
    tx->whole.widthBounder = TRUE;
    tx->whole.width = width;
}

/* : axtRemoveFieldWidth - shut off 'widthBounder'-style formatting
 */
void
axtRemoveFieldWidth(tx)
axtext *tx;
{
    if (!tx)
        return;
    tx->whole.widthBounder = FALSE;
    if (tx->whole.bounder.numPaths) {
        freeBounder(&tx->whole.bounder);
        memset((void *) &tx->whole.bounder, 0, sizeof(poly));
    }
}

/* : axtIntSetFieldWidth - sub called from reFit to
 * generate a text bounder corresponding to the width
 * set by axtSetFieldWidth
 */
static void
axtIntSetFieldWidth(tx, formattedWidth, fieldXform)
axtext *tx;                                /* in */
int formattedWidth;                        /* in */
axform *fieldXform;                        /* in */
{   
    int lineSpacing, mlen, leftx, rightx;
    int bounderHeight, bounderWidth;
    p32 pos;
    poly bounder;
    poly_type *paths;

    /* update value describing formatted width of text
     */
    tx->whole.formattedWidth = formattedWidth;
    
    if (tx->whole.bounder.numPaths) {
        return;                           /* already have a bounder */
    }
    
    if (!tx->whole.lineHeight) {
        lineSpacing = getDefaultLineHeight(tx);
    }
    else {
        lineSpacing = tx->whole.lineHeight;
    }
    
    /* implement field with by specifying a bounder that is
     * rectangular, with top edge one line-height above the
     * text's current position.
     */

    /* get the bounder position from the existing tx->whole.pos
     */
    pos = tx->whole.pos;
    if (tx->whole.xZoom != 1000 || tx->whole.yZoom != 1000) {
        /* the text position whole.pos contains any zoom factor,
         * but the text bounder never contains zoom.  But since
         * bounded text derives its position from its bounder,
         * it's necessary here to de-zoom the position..
         */
        pos.x = (int) (((float) pos.x * 1000.0) / (float) tx->whole.xZoom);
        pos.y = (int) (((float) pos.y * 1000.0) / (float) tx->whole.yZoom);
    }

    bounderHeight = bounderWidth = BIG_MILS;
    if (tx->whole.width) {
        bounderWidth = tx->whole.width;
    }
    
    leftx = pos.x;
    /* shift bounder position based on field alignment
     */
    switch (tx->whole.alignment) {
        
    case RIGHT_ALIGN:
        leftx += formattedWidth;
        leftx -= bounderWidth;
        break;
        
    case CENTER_ALIGN:
        leftx += formattedWidth / 2;
        leftx -= bounderWidth / 2;
        break;
        
    }
    rightx = leftx + bounderWidth;
    
    memset((char *) &bounder, 0, sizeof(bounder));
    bounder.rectangular = TRUE;
    bounder.numPaths = 1;
    paths = (poly_type *) txloc(tx->pid, sizeof(poly_type));
    memset((char *)paths, 0, sizeof(poly_type));
    paths[0].nPoints = 5;
    mlen = paths[0].nPoints * sizeof(TmPoint32);
    paths[0].points = (TmPoint32 *) txloc(tx->pid, mlen);
    memset((char *) paths[0].points, 0, mlen);
    paths[0].points[0].x = leftx;
    paths[0].points[0].y = pos.y - lineSpacing - 1;
    paths[0].points[4] = paths[0].points[0];
    paths[0].points[1].x = rightx;
    paths[0].points[1].y = paths[0].points[0].y;
    paths[0].points[2].x = rightx;
    paths[0].points[2].y = paths[0].points[0].y + bounderHeight;
    paths[0].points[3].x = leftx;
    paths[0].points[3].y = paths[0].points[2].y;
    bounder.paths = paths;
    
    if (!tx->whole.transform.unity) {
        /* transform the bounder
         */
        float xscale, yscale;
        int ix;
        int tmp = 0;

        extractYscale(fieldXform, &yscale);
        if (yscale > 500.0) {
            /* detransform the default bounderHeight to prevent
             * overflow during the coming forward-transform; do this
             * only for scale-ups -- neglect the scale-down case
             * (scale factor 500 < (INT_MAX / BIG_MILS))
             */
            bounderHeight = (int) ((float) bounderHeight / yscale);
            paths[0].points[3].y = bounderHeight; /* re-init bottom edge */
            paths[0].points[2].y = bounderHeight;
        }
        extractXscale(fieldXform, &xscale);
        if (xscale > 500.0) {
            /* similiarly to just above, but for horizontal direction
             * (scale factor 500 < (INT_MAX / BIG_MILS))
             * This is going to throw off positioning under non-left
             * alignment -- do we care?
             */
            bounderWidth = (int) ((float) bounderWidth / xscale);
            paths[0].points[1].x = paths[0].points[0].x + bounderWidth;
            paths[0].points[2].x = paths[0].points[1].x;
        }

        /* transform bounder
         */
        for (ix = 0; ix < 5; ix++) {
            paths[0].points[ix].x -= pos.x;
            paths[0].points[ix].y -= pos.y;
            forwardSXRT(paths[0].points[ix].x, paths[0].points[ix].y,
                        &paths[0].points[ix].x, &paths[0].points[ix].y,
                        fieldXform);
            paths[0].points[ix].x += pos.x;
            paths[0].points[ix].y += pos.y;
        }
    }
    
    axtIntSetRectBounder(tx, &bounder);
}

/* : axtIntSetRectBounder - sub called by axtIntSetFieldWidth.
 * This func assumes param 'bounder' contains one rectangular path.
 */
static void
axtIntSetRectBounder(tx, bounder)
axtext *tx;
poly *bounder;
{
    int miny, maxy, ix;
    
    miny = MAXSHORT16;
    maxy = 0;
    for (ix = 0; ix < bounder->paths[0].nPoints; ix++) {
        if (bounder->paths[0].points[ix].y < miny)
            miny = bounder->paths[0].points[ix].y;
        if (bounder->paths[0].points[ix].y > maxy)
            maxy = bounder->paths[0].points[ix].y;
    }
    tx->whole.bounderMiny = miny;
    tx->whole.bounderMaxy = maxy;
    
    if (tx->whole.bounder.numPaths) {
        freeBounder(&tx->whole.bounder);
    }
    tx->whole.bounder = *bounder;
    tx->whole.bounder.rectangular = TRUE;
}

/* : axtGetFieldWidth
 */
void
axtGetFieldWidth(tx, width)
axtext *tx;
int *width;
{
    *width = tx->whole.width;
}

void
axtSetAlignment(tx, align)
axtext *tx;
int align;  /* left, center, right.. */
{
    r32 bbox;
    p32 txpos;
    char *err;

    if (!tx)
        return;

    if (tx->whole.widthBounder) {
        axtGetPos(tx, &txpos);     /* remember "app" position */
    }
    
    if (align != tx->whole.alignment) {
        if (tx->editState.disp) { 
            if (!tx->old) {
                /* first edit since last display -- save state.
                 * setAlignment is effectively an edit at pos 0.
                 */
                axtSnapshotText(tx, 0);
            }
            axtMoveItems(tx); /* alignment change affects all chars */
        }
        if (!tx->whole.bounder.numPaths) {
            /* for unbounded text, align with respect to
             * a fixed point, e.g., for right-aligned text,
             * the fixed point is on the right edge of the
             * text;  start off by finding this point
             */
            switch(align) {
                
            case LEFT_ALIGN:
            case JUSTIFIED_ALIGN:
                tx->whole.fixedPos = tx->whole.pos;
                break;
                
            case RIGHT_ALIGN:
            case CENTER_ALIGN:
                tx->whole.fixedPos.y = tx->whole.pos.y;
                if(err = getOverallBounder(tx, 0, -1, &bbox)) {
                    freeErrString(err);
                    tx->whole.fixedPos.x = tx->whole.pos.x;
                    break;
                }
                if (align == CENTER_ALIGN) {
                    tx->whole.fixedPos.x = tx->whole.pos.x +
                        bbox.origin.x + (bbox.width / 2);
                }
                else {
                    tx->whole.fixedPos.x = tx->whole.pos.x +
                        bbox.origin.x + bbox.width;
                }
                break;
            }
        }
        tx->whole.alignment = align;

        if (tx->whole.widthBounder) {
            axtRemoveFieldWidth(tx);               /* remove whole.bounder */
            setPos(tx, &txpos);                    /* write "app" position */
            axtSetFieldWidth(tx, tx->whole.width); /* request new one */
        }
        
        tx->measured = tx->displayable = FALSE;
    }
}

int   /* return alignment value */
axtGetAlignment(tx)
axtext *tx;
{
    if (tx) {
        return(tx->whole.alignment);
    }
    else {
        return(LEFT_ALIGN);
    }
}

void
axtSetTabInfo(tx, defTabSize, n_tabs, tabStops)
axtext *tx;
int defTabSize;
int n_tabs;
int *tabStops;
{
    int sz;

    if (!tx)
        return;

	tx->whole.tabInfo.defTabSize = defTabSize;
	tx->whole.tabInfo.n_tabs     = n_tabs;
	if ( n_tabs > 0 && tabStops )
	{
		sz = n_tabs*sizeof(int);
		tx->whole.tabInfo.tabStops = (int *)txloc(tx->pid, sz);
		memcpy(tx->whole.tabInfo.tabStops, tabStops, sz);
	}
    tx->measured = tx->displayable = FALSE;
}

int /* FALSE if no cursor */
getCursorIndexAndPos(tx, indx, pos)
axtext *tx;
int *indx; /* out: cursor index */
p32 *pos; /* out: window-relative position of cursor in mils */
{
    
    if (!indx || !pos)
        return FALSE;
    
    if (tx->editState.inEdit) {
        if (tx->whole.nChars) {
            getRelativeOffset(tx, 0, tx->editState.cursorPos, pos);
            pos->x += tx->whole.pos.x;
            pos->y += tx->whole.pos.y;
            *indx = tx->editState.cursorPos;
        }
        else {
            axtGetPos(tx, pos);
            *indx = 0;
        }
        return TRUE;
    }
    else
        return FALSE;
}

    /* return the indices of the word, and line which enclose
     * refIndex */
int
getWordIndices(txt, refIndex, wLeft, wRight, lLeft, lRight)
axtext *txt;
int refIndex; /* in: reference index */
int *wLeft, *wRight; /* out: indices of ends of enclosing word */
int *lLeft, *lRight; /* out: indices of ends of enclosing line */
{
    int ix, break_char;
    tx_char code;
    perChar *pc;

    /* snaity check */
    if ( !txt || !txt->items || !txt->whole.nChars) {
        *wLeft = *wRight = *lLeft = *lRight = 0;
        return FALSE;
    }

    /* init */
    if ( refIndex >= txt->whole.nChars )
        refIndex = txt->whole.nChars - 1;
    if ( refIndex < 0 )
        refIndex = 0;
    code = txt->items[refIndex].charcode;
    break_char = WORDBREAK(code);

    /* look leftward for enclosing word index */
    *wLeft = 0;
    for ( ix=refIndex; ix>=0; ix-- ) {
        pc = &txt->items[ix];
        if ( break_char != WORDBREAK(pc->charcode) ) {
            *wLeft = ix + 1;
            break;
        }
    }

    /* look leftward for enclosing line index */
    *lLeft = 0;
    for ( ix=refIndex-1; ix>0; ix-- ) {
        pc = &txt->items[ix];
        if ( pc->offset.y ) {
            *lLeft = ix;
            break;
        }
    }

    /* look rightward for enclosing word index */
    *wRight = txt->whole.nChars - 1;
    for( ix=refIndex; ix<txt->whole.nChars; ix++ ) {
        pc = &txt->items[ix];
        if ( break_char != WORDBREAK(pc->charcode) ) {
            *wRight = ix - 1;
            break;
        }
    }

    /* look rightward for enclosing line index */
    *lRight = txt->whole.nChars - 1;
    for ( ix=refIndex; ix<txt->whole.nChars; ix++ ) {
        pc = &txt->items[ix];
        if ( pc->offset.y ) {
            if ( ix == *lLeft ) {
                *lRight = ix;
            } else {
                *lRight = ix - 1;
            }
            break;
        }
    }

    return TRUE;
}

int
axtTrimWordIndices(txt, ix1, ix2)
axtext *txt;
int *ix1;
int *ix2;
{
	int ix, wix1, wix2, modified;

	/* sanity check */
    if ( !txt || !txt->items || !txt->whole.nChars )
		return(FALSE);
	wix1 = *ix1;
	wix2 = *ix2;
	if ( wix1 == wix2 )
		return(FALSE);

	/* scan the end, from right to left */
	modified = FALSE;
	for ( ix=wix2; ix>wix1; ix-- )
	{
		if ( WHITECHAR(txt->items[ix-1].charcode) == FALSE )
			break;

		*ix2 = ix-1;
		if ( modified == FALSE )
			modified = TRUE;
	}

	/* scan the beginning, from left to right */
	wix2 = *ix2;
	if ( wix1 != wix2 )
	{
		for ( ix=wix1; ix<wix2; ix++ )
		{
			if ( WHITECHAR(txt->items[ix].charcode) == FALSE )
				break;

			*ix1 = ix+1;
			if ( modified == FALSE )
				modified = TRUE;
		}
	}
	return(modified);
}

/* : axtMinimumFormatRectSize - report the "actual"
 * formatting bounder for a rectangular bounded text
 * instance: this is the #lines times linespacing
 * for height, widest line's summed escapement for width
 * Computation is done with no transform, with the result
 * then transformed forward using zoom and field xforms.
 */
#define FMARGIN 1   /* fudge this many mils onto untransformed size */
int
axtMinimumFormatRectSize(txt, max_wid,max_hyt, p0,p1,p2,p3)
axtext *txt;            /* in  */
int max_wid, max_hyt;   /* in  */
p32 *p0, *p1, *p2, *p3; /* out */
{
    int i, ix, maxWid, numLines, twid, x1,y1,x2,y2, min_x,min_y;
    char *errString;
    axform xform;
    p32 p[4];

    /* if the text has not been measured, do it now */
    if (!txt->measured) {
        if(errString = axtMeasuredText(txt)) {
            txfree((char *) errString);
            p0->x = p0->y = 0;
            p1->x = p1->y = 0;
            p2->x = p2->y = 0;
            p3->x = p3->y = 0;
            return FALSE;
        }
    }

    /* assemble the metric info */
    if (!txt->whole.nChars) {
        p0->x = p0->y = 0;
        p1->x = p1->y = 0;
        p2->x = p2->y = 0;
        p3->x = p3->y = 0;

    }
    else {
        for(ix = 0, maxWid = 0, numLines = 0, twid = 0;
            ix < txt->whole.nChars; ix++) {
            if (txt->items[ix].offset.y) {
                if (twid > maxWid)
                    maxWid = twid;
                numLines++;
                twid = 0;
            }
            twid += txt->items[ix].padvance;
        }
        if (twid > maxWid)
            maxWid = twid;
        numLines++;

        /* set up a transform (field xform and zoom) */
        axtGetFieldXform(txt, txt->whole.xZoom, txt->whole.yZoom, &xform);

        /* transform */
        x1 = 0;
        y1 = 0;
        x2 = txt->whole.leftOffset + maxWid +
            txt->whole.rightOffset + FMARGIN;
#if 0
        y2 = (numLines * txt->whole.lineHeight) +
            txt->whole.topOffset + txt->whole.botOffset + FMARGIN;
#else
        y2 = (numLines * txt->whole.lineHeight) +
            txt->whole.botOffset + FMARGIN;
#endif
        forwardSXRT(x1,y1, &p[0].x,&p[0].y, &xform);
        forwardSXRT(x2,y1, &p[1].x,&p[1].y, &xform);
        forwardSXRT(x2,y2, &p[2].x,&p[2].y, &xform);
        forwardSXRT(x1,y2, &p[3].x,&p[3].y, &xform);

        /* force the vals to be zero relative */
        min_x = p[0].x;
        min_y = p[0].y;
        for ( i=1; i<4; i++ ) {
            if ( p[i].x < min_x )
                min_x = p[i].x;
            if ( p[i].y < min_y )
                min_y = p[i].y;
        }
        for ( i=0; i<4; i++ ) {
            p[i].x -= min_x;
            p[i].y -= min_y;
        }
    
        /* clip to fit the max width and height */
        for ( i=0; i<4; i++ ) {
            /* clip */
            if ( max_wid > 0 && p[i].x > max_wid )
                p[i].x = max_wid;
            if ( max_hyt > 0 && p[i].y > max_hyt )
                p[i].y = max_hyt;
        }

        *p0 = p[0];
        *p1 = p[1];
        *p2 = p[2];
        *p3 = p[3];
    }

    /* done */
    return TRUE;
}

void
axtLineBreaksToNewlines(txt)
axtext *txt;
{
    char *errString;
    int ix;
    tx_char nl;
    
    if (!txt->measured) {
        if(errString = axtMeasuredText(txt)) {
            txfree((char *) errString);
            return;
        }
    }
    nl = (tx_char) NEWLINE;
    for (ix = 0; ix < txt->whole.nChars; ix++) {
        if (((int) txt->items[ix].charcode != NEWLINE) &&
            txt->items[ix].offset.y) {
                /* note: this insert increments loop var
                   txt->whole.nChars */
            insChars(txt, &nl, ix, 1);
            ix++;    /* skip newline just inserted */
        }
    }
}

void
axtDeleteHiddenChars(txt)
axtext *txt;
{
    char *errString;
    int ix, hiddenIx, hiddenLen;
    
    if (!txt->measured) {
        if(errString = axtMeasuredText(txt)) {
            txfree((char *) errString);
            return;
        }
    }
        /* delete all hidden ranges of characters */
    while(txt->whole.nChars) {
        for (ix = 0, hiddenIx = -1, hiddenLen = 0;
             ix < txt->whole.nChars; ix++) {
            
            if (txt->items[ix].hidden) {
                if (hiddenIx == -1)
                    hiddenIx = ix;    /* start of a hidden range */
                hiddenLen++;
            }
            else if (hiddenIx != -1) {
                    /* end of a hidden range */
                delChars(txt, hiddenIx, hiddenLen);
                break;
            }
        }
        if (hiddenIx != -1) {
                /* the text ends in a hidden range */
            delChars(txt, hiddenIx, hiddenLen);
            break;
        }
        if (hiddenIx == -1) {
                /* no hidden chars left */
            break;
        }
    }
}

/*
 * : axtDirectScreenfont - return the xlfd name of the
 * screenfont that would be used by the axtext code to
 * draw text of the passed attributes.  A null xlfd return
 * means that the axtext code would resort to raster-scaling
 * to draw the text.  NOTE: unity zoom and transform are assumed.
 */
char *   /* caller must free errorString return if nonnull */
axtDirectScreenfont(family, pointsize, weight, slant, xlfd, freeXLFD,
                    targetRec)
char *family;              /* font family, e.g. Times */
double pointsize;           /* in points */
int weight;                /* bool */
int slant;                 /* bool */
char **xlfd;               /* out */
int *freeXLFD;             /* out: bool -- iff true, caller must free *xlfd */
fontTabEnt **targetRec;    /* out: return pointer to font rec of passed atts */
{
    fontTabEnt *screenRec;
    char *err, *txlfd;
    int rasterError;
    
    if (err = axGetFontRecordForScreen(family, weight, slant, (int) pointsize,
                                       targetRec, &screenRec)) {
        *xlfd = (char *) NULL;
        return (err);
    }
#if AX_X11
        /* on X, the font might not be scalable
         */
    if (screenRec->scalable) {
        txlfd = (char *) malloc(MAXLFD);
        axtFormScalableName(screenRec->scalableName, (int) pointsize,
                            ScreenRes, (axctm *) 0, txlfd);
        *xlfd = txlfd;
        *freeXLFD = TRUE;
    }
    else {
            /* get the screenfont name closest to the passed atts
             */
        txlfd = axtGetClosestRasterName(screenRec, pointsize, (double)1.0,
                                        &rasterError);
            /* test whether close enough
             */
        if (foundMatchingScreenfont(screenRec, pointsize, rasterError)) {
            *xlfd = txlfd;    /* would go direct using this screenfont */
        }
        else {
            *xlfd = (char *) NULL;   /* would use raster-scaling */
        }
        *freeXLFD = FALSE;           /* xlfd points onto a fontcache item */
    }
#else
        /* on Windows, we restrict ourselves to scalable screenfonts,
         * which means this func always returns some screenfont name.
         * call axMakeCompoundWinName passing family, pointsize, bold,
         * ital to produce the "xlfd"
         */
    *xlfd = axMakeCompoundWinName(screenRec->familyName, (int) pointsize,
                                  screenRec->weight, screenRec->slant);
    *freeXLFD = TRUE;
#endif
    return ((char *) NULL);
}

/* : return a copy of the passed instance.  Though copy is not
 * thorough, it will be measurable if passed a measurable instance.
 * caller must free non-null returned copy using daxtext().
 * Note that no copy is returned in the 0-chars case.
 */
axtext *
axtCopy(tx)
axtext *tx;
{
    axtext *tcopy = (axtext *) NULL;

    if (tx->whole.nChars) {
        copyAxtext(&tcopy, tx);
    }
    return (tcopy);
}

/*
 * Support routines
 */

static int
axtHypot(x1, y1, x2, y2)  /* this hypot version from "Graphics Gems" */
int x1, y1, x2, y2;
{
    int min;
    
    if ((x2 = x2 - x1) < 0)
        x2 = -x2;

    if ((y2 = y2 - y1) < 0)
        y2 = -y2;

    if (x2 > y2)
        min = y2;
    else
        min = x2;

    return(x2 + y2 - (min >> 1));
}

int
axtHasOffset(pc, prevSup)
perChar *pc;           /* in */
p32 *prevSup;          /* in: previous char's sub/sup offset */
{
    if (CHAR_IS_DIMENSIONLESS(pc->charcode)) {
        return TRUE;
    }
        
    if (pc->offset.x || pc->offset.y || pc->kern.x ||
            pc->kern.y) {
        
        return TRUE;       /* any nonzero value denotes offset */
    }
    else if ((pc->sup.x != prevSup->x) || (pc->sup.y != prevSup->y)) {
        return TRUE;       /* offset only when different from prev char */
    }
    else
        return FALSE;      /* good */
}

static void
fieldDefaults(tx)
axtext *tx;
{
    tx->whole.transform.xscale = 10000;
    tx->whole.transform.yscale = 10000;
    tx->whole.transform.xshear = 0;
    tx->whole.transform.angle = 0;
    tx->whole.xZoom = tx->whole.yZoom = 1000;
    tx->whole.transform.unity = TRUE;
}

    /* reset perChar info that pertains to edits */
static void
resetPerChars(txt)
axtext *txt;
{
    int lx;
    if (!txt || !txt->items)
        return;
    for (lx=txt->editState.editPos; lx<txt->whole.nChars; lx++) {
        txt->items[lx].changed =
        txt->items[lx].deleted = 
        txt->items[lx].damaged =
        txt->items[lx].moved = 
        txt->items[lx].inserted = FALSE;
    }
}

static void
resetEditState(est)
eState *est;
{
    if (est->background)
        XFreePixmap(Dpy, est->background);
    memset((char *)est, 0, sizeof(eState));
}

static int
drawRuling(tx, w, txAp, startix, len, ufrom, uto, lifting)
axtext *tx;
long w;
txAttr *txAp;       /* in: attr seg being underlined */
int startix, len;   /* in: start index, length in chars of the seg */
p32 *ufrom, *uto;   /* in: corresponding endpoints, on baseline */
int lifting;        /* in; bool: TRUE if lifting the ruling */
{
    p32 tfrom, tto, from, to, mapOrig;
    p32 lfrom, lto;
    p32 thickness, offset, strikeoffset;
    r32 mapRect;
    unsigned long fgCol, bgCol, planeMask;
    int strokewidth, patterned;
    Pixmap map;
    Drawable drw, screenDrw;
    GC tgc;
    int x_off, y_off;

    if (len == 0 )
        return FALSE;
    if (txAp->ruling == FALSE && txAp->strike_thru == FALSE )
        return FALSE;

        /* set drawable, get draw colors
         */
    screenDrw = drw = AxGetDrawableFromWidget((AXWidget)w, &x_off, &y_off);
    if (lifting) {
/* %%%%%%% note no treatment of background pixmap.... */
        setLiftColor(tx, axLiftGC);
    }
    else {
        axtSetFgBg(Dpy, txAp, &fgCol, &bgCol, &planeMask);
    }

        /* take copies of passed endpoints, adjust if needed
         */
    axtAdjustRulingEndpoints(tx, startix, len, ufrom, uto, &lfrom, &lto);
    
        /* get the metric info on underline
         */
    thickness.y = txAp->fontAttr.underlineThickness;
    offset.y = txAp->fontAttr.underlinePosition;
    if (thickness.y == 0)
        thickness.y = 50; /* most common in our core fontset */
    if (offset.y == 0)
        offset.y = 150;
    strokewidth = thickness.y;
    if (txAp->strike_thru ) {
        strikeoffset.y = strikeThruOffset(txAp);
    }
    else {
        strikeoffset.y = 0;
    }
    thickness.x = offset.x = strikeoffset.x = 0;
    if (!txAp->tctm.unity) {
        float dtmp;
            /* account for the transform
             */
        forwardSXRT(offset.x,offset.y,&offset.x,&offset.y,&txAp->tctm);
        forwardSXRT(thickness.x,thickness.y,&thickness.x,&thickness.y,
                    &txAp->tctm);
        forwardSXRT(strikeoffset.x,strikeoffset.y,&strikeoffset.x,
                    &strikeoffset.y,&txAp->tctm);
        extractYscale(&txAp->tctm, &dtmp);
        strokewidth = (int)((float)strokewidth * dtmp);
    }
    
        /* convert these numbers to screenpix
         */
    offset.x = mToPix(offset.x, txAp->fontAttr.ptsize, ScreenRes);
    offset.y = mToPix(offset.y, txAp->fontAttr.ptsize, ScreenRes);
    strikeoffset.x = milsToPix(strikeoffset.x, ScreenRes);
    strikeoffset.y = milsToPix(strikeoffset.y, ScreenRes);
    thickness.x = mToPix(thickness.x, txAp->fontAttr.ptsize, ScreenRes);
    thickness.y = mToPix(thickness.y, txAp->fontAttr.ptsize, ScreenRes);
    strokewidth = mToPix(strokewidth, txAp->fontAttr.ptsize, ScreenRes);
    if (strokewidth <= 0)
        strokewidth = 1;

    /* done with most geometry; do drawing now */

    if ( lifting ) {
        patterned = FALSE;
    }
    else if ( IS_SOLID_FILL(tx->gContext, txAp->fillpattern) == FALSE ) {
        patterned = TRUE;
    }
    else {
        patterned = FALSE;
    }
    

    if (patterned) {
            /* draw the ruling onto a pixmap whose size
             * contains the text being underlined
             */
        char *errString;
        int pixwid, pixht, tstart, tlen;
        p32 startpos;

            /* include an extra character in each direction if
             * possible.  ok for the pixmap to be 2 chars larger..
             */
        tstart = startix;
        tlen = len;
        if (startix) {
            tstart = startix - 1;
            tlen++;
        }
        if ((startix + len) < (tx->whole.nChars - 1)) {
            tlen++;
        }
        
        if (errString = getOverallBounder(tx, tstart, tlen, &mapRect)) {
            freeErrString(errString);
            return FALSE;
        }
        
        pixwid = milsToPix(mapRect.width, ScreenRes);
        pixht = milsToPix(mapRect.height, ScreenRes);
		if ( pixwid <= 0 || pixht <= 0 )
			return(FALSE);

        map = XCreatePixmap(Dpy, drw, pixwid, pixht, 1);
        XFillRectangle(Dpy, map, PixClearGC, 0, 0, pixwid, pixht);
        
            /* write screen pos at which to display the pixmap onto mapOrig
             */
        getRelativeOffset(tx, 0, tstart, &startpos);
                /* need to subtract any kern & offset on tstart'th char;
                 * these vectors are already present in the mapRect's origin
                 */
        startpos.x -= tx->items[tstart].kern.x;
        startpos.y -= tx->items[tstart].kern.y;
        startpos.x -= tx->items[tstart].offset.x;
        startpos.y -= tx->items[tstart].offset.y;
        mapOrig.x = startpos.x + mapRect.origin.x + tx->whole.pos.x;
        mapOrig.y = startpos.y + mapRect.origin.y + tx->whole.pos.y;
        
            /* init the pixmap-relative baseline endpoints (from, to)
             * from which the ruling endpoints will be derived; best
             * to think of this as the subtraction of the pixmap's origin
             * from the adjusted underline endpoints lfrom, lto
             */
        from.x = lfrom.x - mapOrig.x;
        from.y = lfrom.y - mapOrig.y;
        to.x = lto.x - mapOrig.x;
        to.y = lto.y - mapOrig.y;
        
        mapRect.width = pixwid;
        mapRect.height = pixht;
        mapOrig.x = milsToPix(mapOrig.x, ScreenRes);
        mapOrig.y = milsToPix(mapOrig.y, ScreenRes);
        
        drw = map;
        tgc = PixGC;
    }
    else {
        map = (Pixmap) 0;
        tgc = lifting ? axLiftGC : axTextGC;
        from = lfrom;
        to = lto;
    }
    
    XSetLineAttributes(Dpy, tgc, strokewidth, LineSolid,
                       CapButt, JoinRound);

    if (!(patterned || lifting)) {  /* patterned case uses (ready) PixGC */
        XSetState(Dpy, tgc, fgCol, ~AllPlanes, GXcopy, AllPlanes);
    }

    if (txAp->strike_thru ) {
        tfrom.x = milsToPix(from.x, ScreenRes) + strikeoffset.x;
        tfrom.y = milsToPix(from.y, ScreenRes) + strikeoffset.y;
        tto.x = milsToPix(to.x, ScreenRes) + strikeoffset.x;
        tto.y = milsToPix(to.y, ScreenRes) + strikeoffset.y;
        XDrawLine(Dpy, drw, tgc, tfrom.x, tfrom.y, tto.x, tto.y);
    }
    if (txAp->ruling & AXRULING_SINGLE || txAp->ruling & AXRULING_DOUBLE) {
        tfrom.x = milsToPix(from.x, ScreenRes) + offset.x;
        tfrom.y = milsToPix(from.y, ScreenRes) + offset.y;
        tto.x = milsToPix(to.x, ScreenRes) + offset.x;
        tto.y = milsToPix(to.y, ScreenRes) + offset.y;
        XDrawLine(Dpy, drw, tgc, tfrom.x, tfrom.y, tto.x, tto.y);
    }
    if (txAp->ruling & AXRULING_DOUBLE) {
        tfrom.x = tfrom.x + (3 * thickness.x);
        tfrom.y = tfrom.y + (3 * thickness.y);
        tto.x = tto.x + (3 * thickness.x);
        tto.y = tto.y + (3 * thickness.y);
        XDrawLine(Dpy, drw, tgc, tfrom.x, tfrom.y, tto.x, tto.y);
    }
    if (patterned) {
            /* draw results to screen using map and pattern info
             */
        axtDrawText(tx, screenDrw, mapOrig.x, mapOrig.y, mapRect.height,
                    mapRect.width, map,
                    txAp->fillpattern);
        XFreePixmap(Dpy, map);
    }
    return(TRUE);
}

int
strikeThruOffset(txAp)
txAttr *txAp;
{
    int tmp = -qRound(0.5 * (float) 
                      emToMils(txAp->fconx.target->xHeight,
                              txAp->fontAttr.ptsizeInt));

    if (tmp == 0)
        tmp = -1;
    return (tmp); /* in mils */
}

/* : reFit - reformats the text to keep it inside a
 * bounder -- if no bounder present, the text just
 * heads off to the right */
static char * /* NULL means ok */
reFit(tx, lastIndexThatFit)
axtext *tx;
int *lastIndexThatFit; /* return this */
{
    int leftIndex, ix, pairIx = 0;
    int numFit, pastBottom, tSpace;
    int lineSpacing, ypos, newix, free_bounder;
    int tLineHeight, lpos, vlap;
    int leftOffset = 0;
    int  rightOffset = 0;
    intercept *edges, *intermed;
    intercept iscan[3], baselineEdges;
    int doAdvance = TRUE;
    int firstLineFormatted = TRUE;
    int vertAdvance = TRUE;
    p32 prevpoint, leftpoint, dxleftpoint;
    p32 xLineSpacing, txLineHeight, ypoint;
    r32 bounderRect;
    poly *boundp;
    axform fieldXform;
    txAttr *txAp;
    char *errString;
    int (*line_formatter)();
    int (*get_intercepts)();

    if (!tx)
        return(getErrString
               (XLT("reFit: Text address is null.."), (char *)NULL));

    if (!tx->whole.bounder.numPaths && !tx->whole.widthBounder &&
        tx->whole.nChars) {
        
        /* no bounder; the only formatting that is needed is to
         * possibly adjust the position for alignment purposes
         */
        axtAlignPos(tx);
        *lastIndexThatFit = tx->whole.nChars - 1;
        return ((char *) NULL);
    }
    
    defaultTransform(&fieldXform);
    if (!tx->whole.transform.unity) {
        /* get a field-only transform
         */
        axtGetFieldXform(tx, 1000, 1000, &fieldXform);
    }

    if (tx->whole.widthBounder) {
        int maxwid = axtGetMaxWidth(tx, &fieldXform);

        axtIntSetFieldWidth(tx, maxwid, &fieldXform);
    }

    if (!tx->whole.nChars) {
        return ((char *) NULL);
    }

    /* A bounder is present; format the text to fit within
     */
    
    if (!tx->measured) {
        if(errString = axtMeasuredText(tx)) {
            return errString;
        }
        /* should return from here, since axtMeasured calls this func.
         * however, must plumb return of lastIndexThatFit all the way
         * up to dispTx first.  Not critical, this block is unlikely
         */
    }

    if (tx->whole.transform.unity) {
        boundp = &tx->whole.bounder;
        free_bounder = FALSE;
    }
    else {
            /* pass a de-transformed (field-level transform only)
               copy of the bounder to get_intercepts */
        boundp = dexformedBounderCopy(tx, &fieldXform);
        free_bounder = TRUE;
    }
    getTxBounderRect(boundp, &bounderRect);
    
    if (!tx->whole.lineFormatter) {
        line_formatter = defaultLineFormatter;
    }
    else {
        line_formatter = tx->whole.lineFormatter;
    }
    if (!tx->whole.getIntercepts) {
        get_intercepts = defaultGetIntercepts;
    }
    else {
        get_intercepts = tx->whole.getIntercepts;
    }

        /* init the default lineSpacing
         */
    if (!tx->whole.lineHeight) {
        lineSpacing = getDefaultLineHeight(tx);
    }
    else {
        lineSpacing = tx->whole.lineHeight;
    }
    if (!tx->whole.transform.unity) {
            /* get a forward-xformed lineSpacing
             */
        forwardSXRT(0, lineSpacing, &xLineSpacing.x,
                    &xLineSpacing.y, &fieldXform);
    }
        /* init the line-height vars with the default line spacing vals
         */
    tLineHeight = lineSpacing;
    txLineHeight = xLineSpacing;

    /* assign initial vals of leftpoint and prevpoint.
     * leftpoint is passed to line formatter as the position of
     * the 1st char to reformat; prevpoint is passed as the
     * already-formatted position of the character just previous;
     * it comes back as the pos of the last char formatted.
     */
    if ((tx->whole.xZoom == 1000 && tx->whole.yZoom == 1000) &&
        (!tx->whole.widthBounder || tx->whole.alignment == LEFT_ALIGN)) {

        if (tx->old) {
            leftIndex = tx->editState.editPos;
            if (leftIndex) {
                leftpoint = tx->old->whole.pos;
                if (tx->old->items) {
                    leftpoint.x += tx->old->items[0].offset.x +
                        tx->old->items[0].kern.x;
                    leftpoint.y += tx->old->items[0].offset.y +
                        tx->old->items[0].kern.y;
                }
            }
            else {
                leftpoint = tx->whole.pos;
                if (tx->items) {
                    leftpoint.x += tx->items[0].offset.x +
                        tx->items[0].kern.x;
                    leftpoint.y += tx->items[0].offset.y +
                        tx->items[0].kern.y;
                }
            }
        }
        else {
            /* assign tx->whole.pos as 1st valid setpoint within the bounder
             */
            setPosFromBounder(tx, lineSpacing, get_intercepts, boundp,
                              &bounderRect, &fieldXform);
            leftIndex = 0;
            leftpoint.x = tx->whole.pos.x;
            leftpoint.y = tx->whole.pos.y;
            tx->items[0].offset.x = tx->items[0].offset.y = 0;
        }
    }
    else {
        /* leftpoint, prevpoint are maintained in the unzoomed space
         * during the reformatting -- after linebreaks computed, the
         * text positions and offsets are zoomed forward.  This
         * requires formatting all the text, to ensure that unzoomed
         * position, offsets exist only during this function
         *
         * (we also punt to this method for the non-left-aligned widthBounder)
         */
        setPosFromBounder(tx, lineSpacing, get_intercepts, boundp,
                          &bounderRect, &fieldXform);
        leftIndex = 0;
        leftpoint.x = tx->whole.pos.x;
        leftpoint.y = tx->whole.pos.y;
        tx->items[0].offset.x = tx->items[0].offset.y = 0;
    }
    prevpoint.x = leftpoint.x - tx->items[leftIndex].offset.x;
    prevpoint.y = leftpoint.y - tx->items[leftIndex].offset.y;

        /* init the intercept-offset vars
         */
    txAp = attrSegAtPos(tx, leftIndex, &lpos);
    if (txAp->indent) {
        axtGetIndentLineVals(tx, txAp, &fieldXform,
                             leftIndex, &tSpace,
                             &rightOffset, &leftOffset);
        if (tSpace) {
            tLineHeight = tSpace;
        }
        if (!tx->whole.transform.unity) {
            /* get a forward-xformed lineSpacing
             */
            forwardSXRT(0, tLineHeight, &txLineHeight.x,
                        &txLineHeight.y, &fieldXform);
        }
    }

        /* clean the intercept info */
    memset((char *)iscan, 0, 3 * sizeof(intercept));
    memset((char *) &baselineEdges, 0, sizeof(intercept));
    edges = intermed = (intercept *) NULL;
    newix = 0;

    pastBottom = FALSE;
    vlap = -(INT_MAX >> 1);   /* this is for vertical overlap detection */
    
    while(TRUE) {
        
        if (tx->whole.transform.unity) {
            ypos = leftpoint.y;
        }
        else {
                /* reverse-xform leftpoint onto dxleftpoint;
                   this is used for getting intercepts and
                   comparing against bounder limits */
            reverseSXRT(leftpoint.x, leftpoint.y,
                        &dxleftpoint.x, &dxleftpoint.y,
                        &fieldXform);
            ypos = dxleftpoint.y;
        }

        if ((ypos >= (bounderRect.origin.y + (int)bounderRect.height -
                      tx->whole.botOffset)) ||
            (ypos - bounderRect.origin.y > BIG_MILS)) {
            
                /* we've advanced beyond bounder bottom, or beyond
                 * an arbitrarily chosen "big" distance; mark chars as hidden.
                 */
            for (ix = leftIndex; ix < tx->whole.nChars; ix++) {
                tx->items[ix].hidden = TRUE;
                if (tx->editState.disp)
                    tx->items[ix].changed = TRUE;  /* became hidden */
            }
            pastBottom = TRUE;
        }
        
        if (doAdvance && vertAdvance) {
            
            if (!pastBottom) {
                    /* On a new line; get new intercepts */

                if (tx->whole.transform.unity) {
                    ypoint = leftpoint;
                }
                else {
                        /* use de-xformed leftpoint for intercepts */
                    ypoint = dxleftpoint;
                }

                if (tx->whole.transform.unity &&
                    tx->whole.bounder.rectangular) {
                        /* simple case; can get by with just the
                           baseline intercepts info */
                    if(!(*get_intercepts)(tx, ypoint.y,
                                          &baselineEdges, boundp)) {
                        return(getErrString
                               (XLT("Couldn't find bounder edges.."),
                                (char *)0));
                    }
                        /* check whether too close to bottom
                         */
                    ypos = bounderRect.origin.y +
                        (int)bounderRect.height - tx->whole.botOffset;
                    if (ypoint.y > ypos) {
                        edges = (intercept *) NULL;  /* too close */
                    }
                    else {
                        edges = &baselineEdges;
                        
                        if (edges->paircount) {
                            /* add in intercept-offsets; only to 0th pair
                             */
                            edges->pairs[0].p0.x += leftOffset;
                            edges->pairs[0].p1.x += rightOffset;
                        }
                    }
                }
                else {
                        /* maintain intercepts for 3 y positions:
                           baseline, .5 above, .2 below */
                        /* init all three ypos' intercepts */
                    if(!(*get_intercepts)(tx, (int)(ypoint.y -
                                                    (tLineHeight * .5)),
                                          &iscan[0], boundp)) {
                        return(getErrString
                               (XLT("Couldn't find bounder edges.."),
                                (char *)0));
                    }
                    if (iscan[0].paircount) {
                        /* add in intercept-offsets; only to 0th pair
                         */
                        iscan[0].pairs[0].p0.x += leftOffset;
                        iscan[0].pairs[0].p1.x += rightOffset;
                    }
                    
                    if(!(*get_intercepts)(tx, ypoint.y,
                                          &iscan[1], boundp)) {
                        return(getErrString
                               (XLT("Couldn't find bounder edges.."),
                                (char *)0));
                    }
                    if (iscan[1].paircount) {
                        iscan[1].pairs[0].p0.x += leftOffset;
                        iscan[1].pairs[0].p1.x += rightOffset;
                    }
                    
                    if(!(*get_intercepts)(tx, (int)(ypoint.y +
                                                    (tLineHeight * .2)),
                                          &iscan[2], boundp)) {
                        return(getErrString
                               (XLT("Couldn't find bounder edges.."),
                                (char *)0));
                    }
                    if (iscan[2].paircount) {
                        iscan[2].pairs[0].p0.x += leftOffset;
                        iscan[2].pairs[0].p1.x += rightOffset;
                    }
                    
                        /* intersect the intercepts */
                    axtFreeIntercepts(edges);
                    axtFreeIntercepts(intermed);
                    intermed = axtCombineIntercepts(&iscan[1], &iscan[0],
															   tx->pid);
                    if (intermed) {
                        edges = axtCombineIntercepts(intermed, &iscan[2],
															   tx->pid);
                    }
                    else {
                        edges = (intercept *) NULL;
                    }
                    if (edges) {
                            /* assign the current baseline's ycoord to the
                               resultant intercept info */
                        for (ix = 0; ix < edges->paircount; ix++) {
                            edges->pairs[ix].p0.y =
                                edges->pairs[ix].p1.y = ypoint.y;
                        }
                    }
                    if (edges && !tx->whole.transform.unity) {
                            /* must forward-xform the intercepts coords */
                        for (ix = 0; ix < edges->paircount; ix++) {
                            forwardSXRT(edges->pairs[ix].p0.x,
                                        edges->pairs[ix].p0.y,
                                        &edges->pairs[ix].p0.x,
                                        &edges->pairs[ix].p0.y,
                                        &fieldXform);
                            forwardSXRT(edges->pairs[ix].p1.x,
                                        edges->pairs[ix].p1.y,
                                        &edges->pairs[ix].p1.x,
                                        &edges->pairs[ix].p1.y,
                                        &fieldXform);
                        }
                    }
                }
            }
            if (edges && edges->paircount > 0) {
                pairIx = 0;
                leftpoint = edges->pairs[pairIx].p0;
            }
        }

        /* format the line */
        if ( edges && edges->paircount != 0 ) {
            if (pairIx == edges->paircount - 1) {
                vertAdvance = TRUE; /* last segment at this ycoord */
            }
            else {
                vertAdvance = FALSE; /* more segs at this y coord */
            }

            /* fit the current line.  leftIndex is that of the 1st char at
             * which to start reformatting.  prevpoint is the set position
             * of the char just behind leftIndex (or of the 0th char if
             * leftIndex is zero). */
            doAdvance = (*line_formatter)(tx, leftIndex, 
                                          &edges->pairs[pairIx],
                                          &leftpoint, &prevpoint,
                                          &numFit, pastBottom,
                                          &fieldXform, &vlap);
        }
        else {
            vertAdvance = TRUE;  /* no segs at this y coord */
        }
        firstLineFormatted = FALSE;

		/* If there is a widthbounder and this char could not fit in that
		 *  width no point in trying to advance char height and try again.
		 *  IT NEVER FITS 
		 */
		if ( numFit == 0 && tx->whole.widthBounder )
			pastBottom = TRUE;

		/* the line formatter has set numFit to the #of chars that
		 * fit on the line just formatted.  prevpoint was set to (or
		 * left at) the position of the last char that was formatted
		 */
        
		/* advance leftIndex to the next, 'leftmost' character
		 */
        if ( edges && edges->paircount != 0 ) {
            if (leftIndex + numFit >= tx->whole.nChars)
                break; /* ran out of characters */
            leftIndex += numFit;
        }
        
        if (doAdvance) {
            if (vertAdvance) {
                
                /* move down a line.  But first check whether the
                 * char at the new leftIndex falls under a TXindent
                 * attribute, which would affect lineheight vars
                 */

                txAp = attrSegAtPos(tx, leftIndex, &lpos);
                if (txAp->indent) {
                    axtGetIndentLineVals(tx, txAp, &fieldXform,
                                         leftIndex, &tLineHeight,
                                         &rightOffset, &leftOffset);
                    if (!tLineHeight) {
                        tLineHeight = lineSpacing;
                    }
                    if (!tx->whole.transform.unity) {
                        /* get a forward-xformed lineSpacing
                         */
                        forwardSXRT(0, tLineHeight, &txLineHeight.x,
                                    &txLineHeight.y, &fieldXform);
                    }
                }
                else {
                    tLineHeight = lineSpacing;    /* use whole-field numbers */
                    txLineHeight = xLineSpacing;
                    leftOffset = rightOffset = 0;
                }
                
                /* move down a line
                 */
                if (tx->whole.transform.unity) {
                    leftpoint.y += tLineHeight;
                }
                else {
                    leftpoint.x += txLineHeight.x;
                    leftpoint.y += txLineHeight.y;
                }
            }
            else {
                /* this edge pair was filled, advance to the next
                 * edge pair found at this y coord
                 */
                pairIx++;
                leftpoint = edges->pairs[pairIx].p0;
            }
        }

        if (pastBottom) {
            break;  /* needed to format just one line past bottom */
        }
        
        if (tx->whole.transform.unity) {
            if (tLineHeight <= 0) {
                pastBottom = TRUE;
            }
        }
        else {
            if (!(txLineHeight.x || txLineHeight.y)) {
                pastBottom = TRUE;
            }
        }
    }
    
    if (tx->whole.xZoom != 1000 || tx->whole.yZoom != 1000) {
            /* zoom the position, and all the formatting offsets
             */
        axtZoomFormattedText(tx);
    }
    
    if (tx->whole.transform.unity &&
        tx->whole.bounder.rectangular) {
        if (baselineEdges.pairs)
            txfree((char *) baselineEdges.pairs);
    }
    else {
        axtFreeIntercepts(edges);
        axtFreeIntercepts(intermed);
        if (iscan[0].pairs)
            txfree((char *) iscan[0].pairs);
        if (iscan[1].pairs)
            txfree((char *) iscan[1].pairs);
        if (iscan[2].pairs)
            txfree((char *) iscan[2].pairs);
    }
    if (free_bounder) {
        freeBounder(boundp);    /* free bounder data */
        txfree((char *) boundp);  /* free base struct, too */
    }
    *lastIndexThatFit = leftIndex ? leftIndex - 1 : leftIndex;
    return ((char *)NULL);
}

/* : axtZoomFormattedText - zoom the position and all the
 * formatting offsets of the passed text.
 * (zoom is removed during formatting of wrapping text)
 */
static void
axtZoomFormattedText(tx)
axtext *tx;                         /* in */
{
    float xzoom, yzoom;
    int ix, numOffsets;
    int zlx;               /* index into array of zoomed left points */
    p32 *zlp;              /* zoomed left points */
    p32 tpt;
    perChar *pc;
    txAttr *txAp;

    /* state at call: formatting has been performed
     * in the unzoomed space.  the formatting information
     * is contained in the offsets found on the left-edge
     * characters of each line.
     *
     * process overview: remember the absolute positions
     * in unzoomed space of each left-edge character.
     * (a "left-edge" character has a nonzero formatting offset)
     * zoom these absolute positions, then rescan in
     * zoomed space to write new, zoomed offsets that
     * will hit these positions for display
     */
    
    xzoom = (float) (tx->whole.xZoom / 1000.0);
    yzoom = (float) (tx->whole.yZoom / 1000.0);
    
        /* scan once just to count offsets
         */
    for (ix = 0, numOffsets = 0; ix < tx->whole.nChars; ix++) {
        if (tx->items[ix].offset.x || tx->items[ix].offset.y) {
            numOffsets++;
        }
    }

    ix = numOffsets * sizeof(p32);
    zlp = (p32 *) txloc(tx->pid, ix);
    memset((char *) zlp, 0, ix);
    
        /* compute the left-edge positions, and
         * zoom them onto array zlp (zoomed left points)
         */
    for (ix = 0, zlx = 0, tpt = tx->whole.pos; ix < tx->whole.nChars; ix++) {
        pc = &tx->items[ix];
        if (pc->offset.x || pc->offset.y) {
            tpt.x += pc->offset.x;
            tpt.y += pc->offset.y;
            if (zlx < numOffsets) {                   /* should be true */
                    /* store each zoomed left-edge point
                     */
                zlp[zlx].x = (int)(tpt.x * xzoom);
                zlp[zlx].y = (int)(tpt.y * yzoom);
                zlx++;
            }
        }
        /* displacements w/o zoom */
        tpt.x += pc->fadvance.x + pc->fkern.x;
        tpt.y += pc->fadvance.y + pc->fkern.y;
    }

    /* position has been maintained unzoomed during formatting; zoom it back
     */
    tpt.x = (int) (tx->whole.fixedPos.x * xzoom);
    tpt.y = (int) (tx->whole.fixedPos.y * yzoom);
    tx->whole.fixedPos = tpt;
    tpt.x = (int) (tx->whole.pos.x * xzoom);
    tpt.y = (int) (tx->whole.pos.y * yzoom);
    tx->whole.pos = tpt;

    /* replace offsets with vectors which will hit
     * the zoomed left points
     */
    for (ix = 0, zlx = 0; ix < tx->whole.nChars; ix++) {
        pc = &tx->items[ix];
        
        if (pc->offset.x || pc->offset.y) {
            if (zlx < numOffsets) {                     /* should be true */
                pc->offset.x = zlp[zlx].x - tpt.x;
                pc->offset.y = zlp[zlx].y - tpt.y;
                zlx++;
            }
            tpt.x += pc->offset.x;
            tpt.y += pc->offset.y;
        }
        /* displacements contain zoom */
        tpt.x += pc->advance.x + pc->kern.x;
        tpt.y += pc->advance.y + pc->kern.y;
    }
    
    txfree(zlp);
}

/* : setPosFromBounder - Should really take whole.pos as
 * declared, but for now, when bounded text, assign it as
 * the 1st valid setpoint within the bounder.. */
static void
setPosFromBounder(tx, lineSpacing, interceptsProc,
                  bounder, bounderRect, fieldXform)
axtext *tx;
int lineSpacing;
int (*interceptsProc)();
poly *bounder;            /* bounding shape of text field */
r32 *bounderRect;         /* hv bounder of bounder */
axform *fieldXform;       /* field-level transform on text */
{
    p32 tpoint;
    intercept edges;
    int lineHeight, topy, unity, origy;
    int (*get_intercepts)();
    
    if (tx->whole.transform.unity && tx->whole.xZoom == 1000 &&
        tx->whole.yZoom == 1000) {
        unity = TRUE;
    }
    else {
        unity = FALSE;
    }

    if (!lineSpacing) {
        lineHeight = getDefaultLineHeight(tx);
    }
    else {
        lineHeight = lineSpacing;
    }
    
    if (tx->whole.widthBounder) {
        int bwid;
        
        tpoint.x = bounder->paths[0].points[0].x;
        tpoint.y = bounder->paths[0].points[0].y + lineHeight + 1;
        
        bwid = bounder->paths[0].points[1].x - bounder->paths[0].points[0].x;
        
        /* reassign fixedPos based on alignment value and bounder
         */
        tx->whole.fixedPos.y = tpoint.y;
        switch (tx->whole.alignment) {
            
        case JUSTIFIED_ALIGN:
        case LEFT_ALIGN:
            tx->whole.fixedPos.x = tpoint.x;
            break;
            
        case CENTER_ALIGN:
            tx->whole.fixedPos.x = tpoint.x + (bwid / 2);
            break;
            
        case RIGHT_ALIGN:
            tx->whole.fixedPos.x = tpoint.x + bwid;
            break;
            
        }
    }
    else {
        if (!interceptsProc) {
            get_intercepts = defaultGetIntercepts;
        }
        else {
            get_intercepts = interceptsProc;
        }
        if (bounderRect) {
            topy = bounderRect->origin.y;
        }
        else {
            topy = tx->whole.bounderMiny;
        }
        tpoint.x = 0;
        tpoint.y = topy + tx->whole.topOffset;
        origy = tpoint.y;
        
        memset((char *)&edges, 0, sizeof(intercept));
        while (TRUE) {
            if(!(*get_intercepts)(tx, tpoint.y, &edges, bounder)) {
                return;
            }
            if (edges.paircount) {
                tpoint = edges.pairs[0].p0;
                break;
            }
            else if (tpoint.y > tx->whole.bounderMaxy) {
                tpoint.x = 0; /* any suggestions? */
                break;
            }
            else {
                tpoint.y += lineHeight;
                if ((tpoint.y - origy) > BIG_MILS) {
                    /* we've wandered far enough looking for a starting point..
                     */
                    tpoint.x = 0;   /* any suggestions? */
                    break;
                }
            }
            if (lineHeight <= 0) {
                break;
            }
        }
        if (edges.pairs) {
            txfree((char *) edges.pairs);
        }
    }
    if (!unity) {
        /* the bounder is the detransformed version of
         * the real mccoy; so forward-transform the
         * position
         */
        forwardSXRT(tpoint.x, tpoint.y,
                    &tpoint.x, &tpoint.y, fieldXform);
        if (tx->whole.widthBounder) {
            forwardSXRT(tx->whole.fixedPos.x, tx->whole.fixedPos.y,
                        &tx->whole.fixedPos.x, &tx->whole.fixedPos.y,
                        fieldXform);
        }
    }
    tx->whole.pos = tpoint;
}

static int
axtFixPaths(n_paths, in_paths, out_paths, pid)
int n_paths;
poly_type *in_paths, *out_paths;
int pid;
{
    int i, n, n_pts, max;
    TmPoint32 *c, *p;

    for ( i=0; i<n_paths; i++ )
    {
        out_paths[i] = in_paths[i];
        if ( in_paths[i].nCtrls == 0 )
        {
            c = (TmPoint32 *)0;
            p = (TmPoint32 *)txloc(pid,in_paths[i].nPoints * sizeof(TmPoint32));
            for ( n=0; n<in_paths[i].nPoints; n++ )
                p[n] = in_paths[i].points[n];
        }
        else
        {
            n_pts = MAX(0, 2 * in_paths[i].nPoints - 2);
            p = (TmPoint32 *)txloc(pid, n_pts * sizeof(TmPoint32));
            c = (TmPoint32 *)txloc(pid, in_paths[i].nCtrls * sizeof(TmPoint32));
            max = in_paths[i].nPoints - 1;
            for ( n=0; n<max; n++ )
            {
                p[2*n+0] = in_paths[i].points[n+0];
                p[2*n+1] = in_paths[i].points[n+1];
            }
            for ( n=0; n<in_paths[i].nCtrls; n++ )
                c[n] = in_paths[i].controls[n];
            out_paths[i].nPoints = n_pts;
        }
        out_paths[i].points = p;
        out_paths[i].controls = c;
    }
}

/* The default case is to run off to the right forever,
 * in which case the edges returned are (0,y), (infinity, y).
 * If a bounder is present, it is assumed to be simple
 * (a rectangle whose points are ordered clockwize starting
 * from upper left corner.)  A single pointpair is returned. */
/* Note how if a bounder isn't there, then the whole.pos is
 * used; otherwise, the bounder plus offsets determine the intercepts */
static int
defaultGetIntercepts(tx, y, edges, bounder)
axtext *tx;
int y;
intercept *edges;
poly *bounder;
{
    int size, i, j, tmpx, swapped, ix, begin_ix;
    int n_segs, *at_x;
    TmPoint32 *seg1, *seg2;
    float dy, dx;

    if ( !edges )
        return FALSE;

    if ( edges->pairs )
        txfree((char *)edges->pairs);

    /* no bounder - create 'unbounded' intercepts */
    if (!bounder->numPaths) {
        edges->paircount = 1;
        edges->pairs = (pointPair *)txloc(tx->pid, sizeof(pointPair));
        edges->pairs[0].p0.x = tx->whole.pos.x;
        edges->pairs[0].p0.y = y;
        edges->pairs[0].p1.x = INFINITE;
        edges->pairs[0].p1.y = y;
    }
    else {
            /* find the bounder polyline segments that
               intersect a horizontal line at y */
        n_segs = 0;
        seg1 = (TmPoint32 *)0;
        seg2 = (TmPoint32 *)0;
        at_x = (int *)0;
        for ( i=0; i<bounder->numPaths; i++ ) {
            for ( j=1; j<bounder->paths[i].nPoints; j++ ) {
                /* the bottom point is in the segment, the top point isn't */
                if ( bounder->paths[i].points[j-1].y >= y &&
                     y > bounder->paths[i].points[j].y ) {
                    n_segs++;
                    size = n_segs * sizeof(TmPoint32);
                    if ( !seg1)
                      seg1 = (TmPoint32 *)txloc(tx->pid,size);
                    else
                      seg1 = (TmPoint32 *)retxloc(tx->pid, seg1, size);
                    if ( !seg2)
                      seg2 = (TmPoint32 *)txloc(tx->pid, size);
                    else
                      seg2 = (TmPoint32 *)retxloc(tx->pid, seg2, size);

                    seg1[n_segs-1] = bounder->paths[i].points[j-1];
                    seg2[n_segs-1] = bounder->paths[i].points[j];
                }
                else if ( bounder->paths[i].points[j].y >= y &&
                          y > bounder->paths[i].points[j-1].y ) {
                    n_segs++;
                    size = n_segs * sizeof(TmPoint32);
                    if ( !seg1)
                      seg1 = (TmPoint32 *)txloc(tx->pid, size);
                    else
                      seg1 = (TmPoint32 *)retxloc(tx->pid, seg1, size);
                    if ( !seg2)
                      seg2 = (TmPoint32 *)txloc(tx->pid, size);
                    else
                      seg2 = (TmPoint32 *)retxloc(tx->pid, seg2, size);
                    
                    seg1[n_segs-1] = bounder->paths[i].points[j];
                    seg2[n_segs-1] = bounder->paths[i].points[j-1];
                }
                /* note that horizontal lines are excluded */
            }
        }

        /* determine the point of intersection with line at y */
        at_x = (int *)txloc(tx->pid, n_segs * sizeof(int));;
        for ( i=0; i<n_segs; i++ ) {
            if ( seg1[i].x == seg2[i].x )
                at_x[i] = seg1[i].x;
            else {
                dy = (float)(seg1[i].y - seg2[i].y);
                dx = (float)(seg1[i].x - seg2[i].x);
                at_x[i] = seg1[i].x + (y - seg1[i].y) / (dy / dx);
            }
        }

        /* order the segments from left to right */
        if ( n_segs > 0 ) {
            begin_ix = n_segs - 1;
            while ( begin_ix > 0 ) {
                swapped = FALSE;
                for ( ix=0; ix<begin_ix; ix++ ) {
                    if ( at_x[begin_ix] < at_x[ix] ) {
                        swapped = TRUE;

                        tmpx = at_x[begin_ix];
                        at_x[begin_ix] = at_x[ix];
                        at_x[ix] = tmpx;

                        break;
                    }
                }
                if ( !swapped )
                    begin_ix--;
            }    
        }

        /* note the paired intersection points, accounting for offsets */
        edges->paircount = 0;
        edges->pairs = (pointPair *)0;
        for ( i=0; i<n_segs; i+=2 ) {
            at_x[i]   += tx->whole.leftOffset;
            at_x[i+1] -= tx->whole.rightOffset;
            if ( at_x[i] < at_x[i+1] ) {
                size = (edges->paircount+1) * sizeof(pointPair);
                if ( !edges->pairs )
                  edges->pairs = (pointPair *)txloc(tx->pid, size);
                else
                  edges->pairs = (pointPair *)retxloc(tx->pid, edges->pairs, size);
                edges->pairs[edges->paircount].p0.x = at_x[i];
                edges->pairs[edges->paircount].p0.y = y;
                edges->pairs[edges->paircount].p1.x = at_x[i+1];
                edges->pairs[edges->paircount].p1.y = y;
                edges->paircount++;
            }
        }
        
        if (tx->whole.alignment == JUSTIFIED_ALIGN &&
            tx->whole.widthBounder && !tx->whole.width) {
            
            /* right point of edges is out very far; pull in to
             * current formattedWidth (plus small fudge to prevent
             * hit-width linebreak)
             */
            if (edges && edges->paircount) {
                edges->pairs[0].p1.x = edges->pairs[0].p0.x +
                    tx->whole.formattedWidth + 50;
            }
        }

        /* done */
        if ( seg1 )
            txfree(seg1);
        if ( seg2 )
            txfree(seg2);
        if ( at_x )
            txfree(at_x);
    }
    return TRUE;
}
 

    /* Lay out a line of text 
     * simple left justification, no tab ruler, no hyphenation,
     * hardcoded notion of whitespace characters.. */

static int   /* return whether to advance to next intercept pair */
defaultLineFormatter(tx, leftIndex, edges, lpoint, ppoint, 
                     numFit, pastBottom, fieldXform, vlap)
axtext *tx;
int leftIndex;
pointPair *edges;   /* intercept info */
p32 *lpoint;        /* tentative position of 1st char on this line */
p32 *ppoint;        /* in/out; in: of last char prev line; out: this line */
int *numFit;        /* return #chars which could fit on this line */
int pastBottom;     /* in, bool: whether below bottom of bounder */
axform *fieldXform; /* in: field-level-only transform on text;
                       initialized only if nonunity field transform */
int *vlap;          /* in/out; in: bottom of ink of last line; out: this line;
                       this is used only for non-rotated/sheared text,
                       for vertical-overlap detection */
{
    perChar *pc;
    p32 tpt;
    int tmpx, ix, tab_wid, ignore_tabs;
    int inkbottom, tbot;
    int doAdvance = FALSE;
    int tnumfit = 0;

    if (!tx || !lpoint || !ppoint || !tx->items || !numFit) {
        return FALSE;
    }
    
    if (leftIndex > (tx->whole.nChars - 1))
        return FALSE;
    
    pc = &tx->items[leftIndex];
    tpt = *lpoint;
    inkbottom = *vlap;
    
        /* set 1st char offset to agree with lpoint
         */
    pc->offset.y = lpoint->y - ppoint->y;
    pc->offset.x = lpoint->x - ppoint->x;
    pc->linebreak = TRUE;
    pc->hanging = FALSE;
    ignore_tabs = FALSE;
    
        /* advance toward the right intercept
         */
    for (ix=leftIndex; ix < tx->whole.nChars; ix++) {
        if ( NEWCHAR((int)pc->charcode) ) {
            if (ix != leftIndex) {
                    /* if the newline isn't the 1st char on
                       the line, then it doesn't fit */
                doAdvance = TRUE;
                break;
            }
        }
        
        if (!tx->editState.fullDisp) {
                /* check whether the current character's top
                 * juts above the last line's bottom ink
                 */
            if ((pc->orig.y + lpoint->y) < *vlap) {
                tx->editState.fullDisp |= FORMAT_OVERLAP;  /* vert. overlap */
            }
        }
        tbot = lpoint->y + pc->orig.y + pc->height - 1;
        if (tbot > inkbottom) {
            inkbottom = tbot;
        }
         /* check whether current character fits */
        tab_wid = 0;
        if ( CHAR_IS_TABCHAR((int)pc->charcode) )
        {
        	if ( !ignore_tabs )
        		ignore_tabs = axtNeedToIgnoreTabs(tx, ix);
        	if ( !ignore_tabs )
            	tab_wid = axtGetTabWidth(tx, (tpt.x - tx->whole.pos.x), FALSE);
			pc->fadvance.x = tab_wid;
			pc->advance.x = tab_wid * tx->whole.xZoom / 1000.0;
		}
		else
        	ignore_tabs = FALSE;
        if (tx->whole.transform.unity) {
            tmpx = pc->fadvance.x + pc->fkern.y;
            if (tpt.x + tmpx > edges->p1.x) {
                doAdvance = TRUE;
                break;
            }
        }
        else {
                /* detransform before doing limit check
                 */
            p32 tadv, tcur, tlim;
            tadv.x = pc->fadvance.x + pc->fkern.x;
            tadv.y = pc->fadvance.y + pc->fkern.y;
            reverseSXRT(tadv.x, tadv.y, &tadv.x, &tadv.y, fieldXform);
            tcur = tpt;
            reverseSXRT(tcur.x, tcur.y, &tcur.x, &tcur.y, fieldXform);
            tlim = edges->p1;
            reverseSXRT(tlim.x, tlim.y, &tlim.x, &tlim.y, fieldXform);
            if (tcur.x + tadv.x > tlim.x) {
                doAdvance = TRUE;
                break;
            }
        }
        tnumfit++;
        tpt.x += pc->fadvance.x + pc->fkern.x;
        tpt.y += pc->fadvance.y + pc->fkern.y;
        if (pastBottom)
            pc->hidden = TRUE;
        else
            pc->hidden = FALSE;
        pc++;
        if (ix < tx->whole.nChars - 1) {
            pc->offset.x = pc->offset.y = 0;
            pc->hanging = FALSE;
            pc->linebreak = FALSE;
        }
    }

    if (tnumfit && doAdvance) {
        tnumfit = axtWordBreak(tx, leftIndex, tnumfit, TRUE, &tpt);
    }

    if (tnumfit) {
        if (tx->whole.bounder.numPaths || tx->whole.width) {
            /* handle field alignment
             */
            doAlignment(tx, edges, lpoint, &tpt, leftIndex, tnumfit);
        }
        /* report new prevpoint
         */
        *ppoint = tpt;
    }
    *numFit = tnumfit;
    *vlap = inkbottom;
    return doAdvance;
}

/* : axtWordBreak - given a 'raw' formatted line,
 * adjust based on wordbreak */
static int /* return the adjusted numfit */
axtWordBreak(tx, leftix, numfit, hyph, curpoint)
axtext *tx;
int leftix; /* index of leftmost char on line */
int numfit; /* raw format fit this many on the line */
int hyph; /* bool: whether to do hyphenation or just whole word break */
p32 *curpoint; /* in/out; in: current rightmost point, out: adjusted */
{
    perChar *pc;
    p32 tpt;
    int ix, tix;
    int numback = 0;
    /*int language = UPnumVal(PROF_LANGUAGE_);
    if (!language)
        language = ENGLISH;*/

    ix = leftix + numfit - 1; /* index of last char that fit */

    if ((ix >= tx->whole.nChars - 1) ||
        (BLACKCHAR(tx->items[ix].charcode) &&
         WHITECHAR(tx->items[ix + 1].charcode))) {
            /* the last char that fit is also the end of a word */
        return numfit;
    }
    
    pc = &tx->items[ix];
    for (tix = ix, tpt = *curpoint ; tix > leftix; tix--, pc--, numback++) {
#if BUILD_KANJI
        /* Ashisuto modified for breaking Japanese text, found this
         * necessary in TextBox of Graphics Component. Words is also
         * allow Japanese text to break if non-breakable punctuations
         * does not encountered.
         * 1996.11.14
         */
        if (pc->charcode == SPACECHAR || pc->charcode == TABCHAR ||
                (IS_ORIENT_CHAR(pc->charcode) &&
                 !IS_ORIENT_CHAR((pc+1)->charcode)) ||
                (!IS_ORIENT_CHAR(pc->charcode) &&
                 IS_ORIENT_CHAR((pc+1)->charcode)) ||
                (IS_ORIENT_CHAR(pc->charcode) &&
                 IsOKtoBreakOriental(pc->charcode, (pc+1)->charcode)))
#else
        if (pc->charcode == SPACECHAR || pc->charcode == TABCHAR)
#endif
        {
            *curpoint = tpt;
            numfit -= numback;
            break;
        }  
        else if ( hyph && pc->charcode == HYPHEN )
        {
            *curpoint = tpt;
            numfit -= numback;
            break;
        }

        tpt.x -= (pc->fadvance.x + pc->fkern.x);
        tpt.y -= (pc->fadvance.y + pc->fkern.y);
    }
    return numfit;
}

/* : doAlignment - do text alignment on the tentatively
 * formatted line.
 */
static void
doAlignment(tx, edges, lPoint, rPoint, leftIndex, numFit)
axtext *tx;
pointPair *edges;  /* bounder intercepts at this line */
p32 *lPoint;       /* in: setpoint of 1st char on line */
p32 *rPoint;       /* in/out: setpoint just after last char on line */
int leftIndex;     /* index of 1st char on line */
int numFit;        /* #chars that fit on the line */
{
    int lastix, talign, ix, hang;
    p32 width, tmp;
    
    if (tx->whole.bounder.numPaths || tx->whole.width) {
        hang = TRUE;
    }
    else {
        hang = FALSE;  /* don't hang whitespace for unconstrained case */
    }

        /* set 'width' to tentative line setwidth */
    width.x = rPoint->x - edges->p0.x;
    width.y = rPoint->y - edges->p0.y;

    lastix = leftIndex + numFit - 1; /* index of last char that fit */

    if (tx->whole.alignment == JUSTIFIED_ALIGN) {
        if (lastix >= tx->whole.nChars - 1) {
            talign = LEFT_ALIGN;  /* don't justify last line */
        }
        else if ((lastix < tx->whole.nChars - 2) &&
                 NEWCHAR((int)tx->items[lastix + 1].charcode) &&
                 NEWCHAR((int)tx->items[lastix + 2].charcode) ) {
                /* don't justify if the line precedes two consecutive
                   newlines -- a 'paragraph break' of sorts.. */
            talign = LEFT_ALIGN;
        }
        else {
            talign = tx->whole.alignment;
        }
    }
    else {
        talign = tx->whole.alignment;
    }

    switch(talign) {
        
    case LEFT_ALIGN:
        if (hang) {
            /* hang leading whitespace if not on the first line
             */
            if (leftIndex != 0) {
                /* the line is other than the first */
                for (ix = leftIndex; ix <= lastix; ix++) {
                    if (WHITESPACE(tx->items[ix].charcode)) {
                        tx->items[leftIndex].offset.x -=
                            tx->items[ix].fadvance.x;
                        tx->items[leftIndex].offset.y -=
                            tx->items[ix].fadvance.y;
                        tx->items[ix].hanging = TRUE;
                        
                        rPoint->x -= tx->items[ix].fadvance.x;
                        rPoint->y -= tx->items[ix].fadvance.y;
                    }
                    else
                        break;
                }
            }
            /* mark trailing spaces as "hanging", just to suppress uline
             */
            for (ix = lastix; ix >= leftIndex; ix--) {
                if (WHITESPACE(tx->items[ix].charcode)) {
                    tx->items[ix].hanging = TRUE;
                }
                else
                    break;
            }
        }
        break;
        
    case CENTER_ALIGN:
        if (hang) {
            /* exclude both leading and trailing whitespace */
            /* leading */
            for (ix = leftIndex; ix <= lastix; ix++) {
                if (WHITESPACE(tx->items[ix].charcode)) {
                    tx->items[leftIndex].offset.x -=
                        tx->items[ix].fadvance.x;
                    tx->items[leftIndex].offset.y -=
                        tx->items[ix].fadvance.y;
                    tx->items[ix].hanging = TRUE;
                    
                    rPoint->x -= tx->items[ix].fadvance.x;
                    rPoint->y -= tx->items[ix].fadvance.y;
                }
                else
                    break;
            }
            /* trailing */
            for (ix = lastix; ix >= leftIndex; ix--) {
                if (WHITESPACE(tx->items[ix].charcode)) {
                    width.x -= tx->items[ix].fadvance.x;
                    width.y -= tx->items[ix].fadvance.y;
                    tx->items[ix].hanging = TRUE;
                }
                else
                    break;
            }
        }
        
        tmp.x = edges->p1.x - edges->p0.x - width.x;
        tmp.y = edges->p1.y - edges->p0.y - width.y;
        tmp.x /= 2;  /* adjustment size */
        tmp.y /= 2;

        tx->items[leftIndex].offset.x += tmp.x;
        tx->items[leftIndex].offset.y += tmp.y;
        
        rPoint->x += tmp.x;
        rPoint->y += tmp.y;
        break;
        
    case RIGHT_ALIGN:
        if (hang) {
            /* hang trailing whitespace
             */
            for (ix = lastix; ix >= leftIndex; ix--) {
                if (WHITESPACE(tx->items[ix].charcode)) {
                    width.x -= tx->items[ix].fadvance.x;
                    width.y -= tx->items[ix].fadvance.y;
                    tx->items[ix].hanging = TRUE;
                }
                else
                    break;
            }
            /* mark leading spaces as "hanging", just to suppress uline
             */
            for (ix = leftIndex; ix <= lastix; ix++) {
                if (WHITESPACE(tx->items[ix].charcode)) {
                    tx->items[ix].hanging = TRUE;
                }
                else
                    break;
            }
        }
        tmp.x = edges->p1.x - edges->p0.x - width.x;
        tmp.y = edges->p1.y - edges->p0.y - width.y;

        tx->items[leftIndex].offset.x += tmp.x;
        tx->items[leftIndex].offset.y += tmp.y;

        rPoint->x += tmp.x;
        rPoint->y += tmp.y;
        break;
        
    case JUSTIFIED_ALIGN:
        tmp.x = edges->p1.x - rPoint->x;
        tmp.y = edges->p1.y - rPoint->y;
        justifyLine(tx, leftIndex, numFit, &tmp, &width);
        rPoint->x = lPoint->x + width.x;
        rPoint->y = lPoint->y + width.y;
        break;
    }
}

/* : axtGetMaxWidth - 'widthBounder' formatting util which
 * returns the maximum formatting linewidth in an axtext instance.
 */
static int
axtGetMaxWidth(tx, fieldXform)
axtext *tx;                /* in */
axform *fieldXform;        /* in */
{
    int lx, maxwid, twid, itmp, linebreak;
    int rightOffset, leftOffset, leftix;
    p32 tpt;
    perChar *pc;
    txAttr *txAp;
    
    if (!tx->whole.nChars) {
        return 0;
    }

    leftOffset = rightOffset = 0;
    if (tx->atts[0].indent) {
        axtGetIndentLineVals(tx, &tx->atts[0], fieldXform,
                             0, &itmp, &rightOffset, &leftOffset);
    }
    
    /* get widest line width, expressed with field xform removed.
     */
    tpt.x = tpt.y = 0;
    maxwid = leftOffset + rightOffset;
    linebreak = FALSE;
    for (lx = 0, twid = 0, leftix = 0; lx < tx->whole.nChars; lx++) {
        pc = &tx->items[lx];
        if (tx->whole.width && 
            (tpt.x + pc->fadvance.x + pc->fkern.x) > tx->whole.width) {
            
            linebreak = TRUE;   /* wider than constrained bounder */
        }
        if (linebreak || NEWCHAR((int)pc->charcode) ) {
            
            if (linebreak) {
                linebreak = FALSE;
                if (lx > leftix) {
                    itmp = axtWordBreak(tx, leftix, lx - leftix, TRUE, &tpt);
                    lx = leftix + itmp;    /* loop var */
                }
            }
            
            /* get untransformed line width.
             */
            reverseSXRT(tpt.x, tpt.y, &twid, &itmp, fieldXform);
            
            txAp = attrSegAtPos(tx, lx, &itmp);
            if (txAp->indent) {
                axtGetIndentLineVals(tx, txAp, fieldXform, lx,
                                     &itmp, &rightOffset, &leftOffset);
            }
            twid += rightOffset + leftOffset;
            
            if (twid > maxwid) {
                maxwid = twid;
            }
            tpt.x = tpt.y = 0;
            leftix = lx;
        }
        tpt.x += pc->fadvance.x + pc->fkern.x;
        tpt.y += pc->fadvance.y + pc->fkern.y;
    }
    reverseSXRT(tpt.x, tpt.y, &twid, &itmp, fieldXform);
    twid += rightOffset + leftOffset;
    if (twid > maxwid) {
        maxwid = twid;
    }
    if (tx->whole.width) {
        if (maxwid > tx->whole.width) {
            maxwid = tx->whole.width;
        }
    }
    
    return (maxwid);
}

/* : axtAlignPos - adjust overall position based on
 * alignment value.  Strangely enough, it's a width-only
 * adjustment, kept now for compatibility.
 */
static void
axtAlignPos(tx)
axtext *tx;
{
    r32 bbox;
    char *errString;
    
    switch(tx->whole.alignment) {
    case LEFT_ALIGN:
    case JUSTIFIED_ALIGN:
        break;
    case RIGHT_ALIGN:
        if (errString = getOverallBounder(tx, 0, -1, &bbox)) {
            freeErrString(errString);
        }
        else {
            tx->whole.pos.x = tx->whole.fixedPos.x - 
                              (bbox.origin.x + bbox.width);
        }
        break;
    case CENTER_ALIGN:
        if (errString = getOverallBounder(tx, 0, -1, &bbox)) {
            freeErrString(errString);
        }
        else {
                tx->whole.pos.x = tx->whole.fixedPos.x - 
                    (bbox.origin.x + (bbox.width / 2));
        }
        break;
    }
}

static void
justifyLine(txt, leftIndex, numFit, availWidth, totwid)
axtext *txt;
int leftIndex, numFit;
p32 *availWidth;
p32 *totwid;   /* out: report total width including hanging spaces */
{
    int ix, lastix, spCount, spx, hang;
    int lscanx, rscanx, foundBlack;
    p32 adjust, bigAdjustCount, smallAdjustCount;
    p32 totSpWid, inSmall;

    if (!txt || !numFit)
        return;
    
    if (txt->whole.bounder.numPaths || txt->whole.width) {
        hang = TRUE;
    }
    else {
        hang = FALSE;  /* don't hang whitespace in unconstrained case */
    }

        /* (smallCount * adjust) + (bigCount * (adjust + 1)
           will account for the error */

    /* first determine spacecount, total space width;
     * exclude leading whitespace while doing so
     */
    lastix = leftIndex + numFit - 1;
    totSpWid.x = totSpWid.y = spCount = 0;
    foundBlack = FALSE;
    for (ix = lscanx = leftIndex; ix <= lastix; ix++) {
        if (WHITESPACE(txt->items[ix].charcode)) {
            if (foundBlack) {
                spCount++;
                totSpWid.x += txt->items[ix].fadvance.x;
                totSpWid.y += txt->items[ix].fadvance.y;
            }
            else {
                if (hang) {
                    /* adjust offset to 'hang' the space
                     */
                    txt->items[leftIndex].offset.x -=
                        txt->items[ix].fadvance.x;
                    txt->items[leftIndex].offset.y -=
                        txt->items[ix].fadvance.y;
                    txt->items[ix].hanging = TRUE;
                }

                availWidth->x += txt->items[ix].fadvance.x;
                availWidth->y += txt->items[ix].fadvance.y;
            }
        }
        else if (!foundBlack) {
            foundBlack = TRUE;
            lscanx = ix;  /* skip leading white */
        }
    }
    
    if (hang) {
        /* exclude trailing white, too
         */
        for (ix = rscanx = lastix; ix > lscanx; ix--) {
            if (WHITESPACE(txt->items[ix].charcode)) {
                spCount--;
                totSpWid.x -= txt->items[ix].fadvance.x;
                totSpWid.y -= txt->items[ix].fadvance.y;
                availWidth->x += txt->items[ix].fadvance.x;
                availWidth->y += txt->items[ix].fadvance.y;
                txt->items[ix].hanging = TRUE;
            }
            else {
                rscanx = ix;
                break;
            }
        }
    }
    else {
        rscanx = lastix;
    }

    if (spCount) {
            /* set up adjustment size and counts */
        adjust.x = availWidth->x / spCount;
        bigAdjustCount.x = availWidth->x % spCount;
        smallAdjustCount.x = spCount - bigAdjustCount.x;
        inSmall.x = TRUE;
        spx = spCount;

            /* apply the adjustments */
        for (ix = lscanx; ix <= rscanx; ix++, spx-- ) {
            if (WHITESPACE(txt->items[ix].charcode)) {

                if (inSmall.x) {
                    if (smallAdjustCount.x == 0) {
                        inSmall.x = FALSE;
                        adjust.x++;
                    }
                    else {
                        smallAdjustCount.x--;
                    }
                }
                if (ix < rscanx) {
                        /* make this char wider by increasing
                           nextchar's xoffset */
                    txt->items[ix + 1].offset.x += adjust.x;
                }
            }
        }

        if (!txt->whole.transform.unity) {
                /* apply adjustments in the y direction, too */

            adjust.y = availWidth->y / spCount;
            bigAdjustCount.y = availWidth->y % spCount;
            smallAdjustCount.y = spCount - bigAdjustCount.y;
            inSmall.y = TRUE;
            spx = spCount;

            for (ix = lscanx; ix <= rscanx; ix++, spx-- ) {
                if (WHITESPACE(txt->items[ix].charcode)) {

                    if (inSmall.y) {
                        if (smallAdjustCount.y == 0) {
                            inSmall.y = FALSE;
                            adjust.y++;
                        }
                        else {
                            smallAdjustCount.y--;
                        }
                    }
                    if (ix < rscanx) {
                        /* make this char wider by increasing
                           nextchar's yoffset */
                        txt->items[ix + 1].offset.y += adjust.y;
                    }
                }
            }
        }
    }
        /* now report total width of all chars this line
           except any left-side hanging spaces */
    for (ix = lscanx, totwid->x = totwid->y = 0; ix <= lastix; ix++) {
        totwid->x += txt->items[ix].fadvance.x;
        totwid->y += txt->items[ix].fadvance.y;
        if (ix > leftIndex) {
            totwid->x += txt->items[ix].offset.x;
            totwid->y += txt->items[ix].offset.y;
        }
    }
}

/* : axtEditPos -- set index marking upstream-most effect of
 * an edit.  pos is passed as the position of an edit to *tx
 */
static int
axtEditPos(tx, pos)
axtext *tx;
int pos;
{
    perChar *pc;
    
    if (!tx || !tx->items || !tx->whole.nChars)
        return 0;
    
    if (pos < 0)
        return 0;
    
    if (pos >= tx->whole.nChars)
        pos = tx->whole.nChars - 1;
    
    if (tx->whole.bounder.numPaths) {
        int above = 0;
        int enough;
        
        /* set pos backward, ideally to the upstream-most
         * character on the line just above that containing 
         * the passed pos; the desired pos is the nearest
         * linestart which is upstream of any edit effects.
         * for nonleft alignment, we need to go an extra line
         * back.
         */
        if (tx->whole.widthBounder && tx->whole.alignment != LEFT_ALIGN) {
            pos = 0;     /* edits on nonleft align affect all chars */
        }
        else {
            if (tx->whole.alignment != LEFT_ALIGN)
                enough = 2;
            else
                enough = 1;
            
            pc = &tx->items[pos];
            while(pos > 0) {
                if (pc->offset.y) {
                    if (above == enough)
                        return pos;
                    else
                        above++;
                }
                pos--;
                pc--;
            }
        }
    }
    
    if (!tx->whole.bounder.numPaths) {
        if (tx->whole.alignment != LEFT_ALIGN) {
            pos = 0;     /* edits on nonleft align affect all chars */
        }
        if (pos > 0) /* prefer one back for simple strings; this */
            pos--;   /* helps with lift-damage repair */
    }
    if (tx->old) {
            /* an edit has already been done since last display;
             * use the currently passed pos only if upstream */
        if (pos > tx->editState.editPos)
            pos = tx->editState.editPos;
    }
    return pos;
}

/* : getOldIndex - given index on current copy,
 * return corresponding index on the old copy */
static int /* either the corr. index or -1 (i.e., 'no correspondent') */
getOldIndex(currtxt, oldtxt, currix)
axtext *currtxt;
axtext *oldtxt;
int currix;
{
        /* find the 'corresponding index' (corrx) on currtxt, then return
         * the real index of old's member which has the same corrx.. */
    int ix, corrx, ocorrx;
    if (!oldtxt || currtxt->items[currix].inserted ||
            currix < currtxt->editState.editPos) {
        return -1;
    }
    for (ix = currtxt->editState.editPos, corrx = -1; ix <= currix; ix++) {
        if (!currtxt->items[ix].inserted) {
            corrx++;
        }
    }
        /* now find the index on oldtxt with the same corresponding index */
    for (ix = 0, ocorrx = -1; ix < oldtxt->whole.nChars; ix++) {
        if (!oldtxt->items[ix].deleted) {
            ocorrx++;
            if (ocorrx >= corrx) {
                return ix;
            }
        }
    }
    return oldtxt->whole.nChars - 1;
}

static int /* either the corr. index or -1 (i.e., 'no correspondent') */
getNewIndex(currtxt, oldtxt, oldix)
axtext *currtxt;
axtext *oldtxt;
int oldix;
{
        /* find the 'corresponding index' (ocorrx) on oldtxt, then return
         * the real index of new's member which has the same ocorrx.. */
    int ix, corrx, ocorrx;
    if (!oldtxt || !currtxt || oldtxt->items[oldix].deleted) {
        return -1;
    }
    for (ix = 0, ocorrx = -1; ix <= oldix; ix++) {
        if (!oldtxt->items[ix].deleted) {
            ocorrx++;
        }
    }
        /* now find the index on currtxt with the same corresponding index */
    for (ix = currtxt->editState.editPos, corrx = -1;
                    ix < currtxt->whole.nChars; ix++) {
        if (!currtxt->items[ix].inserted) {
            corrx++;
            if (corrx >= ocorrx) {
                return ix;
            }
        }
    }
    return currtxt->whole.nChars - 1;
}

/* : liftOld - compare current and last-displayed states of the
 * passed text instance, and lift anything non-current.  This is
 * done in preparation for display, called from dispTx
 */
static void
liftOld(w, txt)
long w;                 /* in */
axtext *txt;            /* in */
{
    
    int pieceIx, inLiftage, liftlen, lift, liftPrev;
    int oldix, currix, currscan, corrx, ocorrx, lastcurrix;
    int charcount, charix, lx, hasCorrespondent;
    char *errString;
    p32 oldpoint, currentpoint, piecePos, rulept, prevSup;
    perChar *oldpc;
    txAttr *txAp, *pieceAp;
    
/* we scan the old copy marking and lifing those characters that
 * need it.  oldix and oldpoint refer to the current char on the
 * old copy during scan.
 */

    if (!txt || !txt->old )
        return;

    if (!txt->old->whole.nChars) {
        /* test whether to lift the 0th bullet if present
         */
        if (txt->old->atts && txt->old->atts[0].indent &&
            txt->old->atts[0].indent->bullet && 
            txt->old->atts[0].indent->bullet->type) {
            
            p32 oldpt, newpt;
            int liftb = FALSE;
            
            if (!txt->atts[0].indent->bullet ||
                !txt->atts[0].indent->bullet->type) {
                
                liftb = TRUE;          /* bullet gone on current */
            }
            else if (txt->atts[0].indent->bullet->changed) {
                liftb = TRUE;          /* modified on current */
            }
            else {
                /* compare old & new positions determining bullet pos.
                 * 0th bullet pos derives from whole.pos.x, 0th charpos.y
                 */
                oldpt = txt->old->whole.pos;
                newpt = txt->whole.pos;
                if (txt->items) {
                    newpt.y += txt->items[0].offset.y;
                }
                if ((newpt.x != oldpt.x) || (newpt.y != oldpt.y)) {
                    liftb = TRUE;
                }
            }
            if (liftb) {
                liftBullet(txt->old, w, &txt->old->atts[0], axLiftGC,
                           &txt->old->whole.pos);
            }
        }
        return;
    }
    
    if (!txt->old->displayable && txt->old->whole.nChars) {
        /* possibly need screenfont dimensions to do lifting
         */
        if(errString = axtDisplayableText(txt->old)) {
            freeErrString(errString);
            return;
        }
    }
    setLiftColor(txt, axLiftGC);
    
    /* init oldpoint, currentpoint.  These scan the
     * old-copy and new-copy set positions, respectively
     */
    oldpoint = txt->old->whole.pos;
    if (txt->editState.editPos) {
        currentpoint = oldpoint;
        if (txt->items) {
            currentpoint.x += txt->old->items[0].offset.x +
                txt->old->items[0].kern.x;
            currentpoint.y += txt->old->items[0].offset.y +
                txt->old->items[0].kern.y;
        }
    }
    else {
        currentpoint = txt->whole.pos;
        if (txt->items) {
            currentpoint.x += txt->items[0].offset.x + txt->items[0].kern.x;
            currentpoint.y += txt->items[0].offset.y + txt->items[0].kern.y;
        }
    }

    if (!txt->old->items) {
        return;
    }
    
    /* init vars describing the start of the current to-be-lifted
     * run of chars.  If "on", then the previous char (at least) is
     * to-be-lifted.  Init value is "off" (pieceIx == -1)
     */
    piecePos.x = piecePos.y = 0;
    rulept.x = rulept.y = 0;
    pieceIx = -1;
    inLiftage = FALSE;
    liftlen = 0;
    
    lift = FALSE;     /* bool "lift" says whether this char is to be lifted */
    liftPrev = FALSE; /* whether previous char was lifted */
    prevSup.x = prevSup.y = 0;    /* prev char's sub/sup offset */

    /* currix, currscan, corrx, ocorrix, along with currentpoint, support
     * peeking at the current text instance to test whether this char
     * should be lifted.
     */
    currix = currscan = txt->editState.editPos;
    corrx = ocorrx = -1;
    
    lastcurrix = txt->editState.editPos - 1;   /* used for damage-repair */
    
/* scan
 */
    
    txAp = pieceAp = txt->old->atts;
    charcount = txt->old->whole.nChars;
    charix = 0;
    
    while (charcount > 0) {
        
        /* per-attr-seg inits
         */
        lx = charix + txAp->nChars;
        if (txAp->qDraw) {
            /* need screen metrics to get lift dimensions
             */
            if (!txAp->screenFontName) {
                return;
            }
            if (errString = axGetScreenMetrics(txAp->screenFontName,
                                               txAp->fconx.font,
                                               &txAp->fconx.screen)) {
                freeErrString(errString);
                return;
            }
        }
        
        for (oldix = charix; oldix < lx; oldix++) {
            
            oldpc = &txt->old->items[oldix];

            oldpoint.x += txt->old->items[oldix].kern.x;
            oldpoint.y += txt->old->items[oldix].kern.y;
            oldpoint.x += txt->old->items[oldix].offset.x;
            oldpoint.y += txt->old->items[oldix].offset.y;
            
/* determine whether current char is to be lifted.
 */
            if (!oldpc->deleted) {
                ocorrx++;
                hasCorrespondent = TRUE;
            }
            else
                hasCorrespondent = FALSE;

            if (hasCorrespondent) {
                perChar *pc, *tpc;
                
                /* get currix and currentpoint in sync for comparison
                 */
                for ( ; currscan < txt->whole.nChars - 1; currscan++) {
                    if (!txt->items[currscan].inserted)
                        corrx++;
                    if (corrx >= ocorrx) {
                        break;
                    }
                }
                for ( ; currix < currscan; currix++) {
                        pc = &txt->items[currix];
                        currentpoint.x += pc->advance.x;
                        currentpoint.y += pc->advance.y;
                        if (currix < txt->whole.nChars - 1) {
                            tpc = &txt->items[currix+1];
                            currentpoint.x += tpc->offset.x + tpc->kern.x;
                            currentpoint.y += tpc->offset.y + tpc->kern.y;
                        }
                }
    
                if (txt->items[currix].moved) {
                    lift = TRUE;
                }
                else {
                    int hasMoved;
                    p32 tpt, otpt;
                        
                    tpt.x = currentpoint.x + txt->items[currix].sup.x;
                    tpt.y = currentpoint.y + txt->items[currix].sup.y;
                    otpt.x = oldpoint.x + oldpc->sup.x;
                    otpt.y = oldpoint.y + oldpc->sup.y;
                    lift = needLift(oldpc, &txt->items[currix],
                                    &otpt, &tpt, &hasMoved);
                    if (hasMoved) {
                        /* this character on the old copy has a
                         * correspondent on the current copy that has moved
                         */
                        if (currix >= 0 && currix < txt->whole.nChars) {
                            txt->items[currix].moved = TRUE;
                        }
                    }
                }
            }
            else {
                /* no current correspondent; lift unless hidden
                 */
                lift = !oldpc->hidden;
            }
            
            if (txt->items) {
                if (oldix && lift && !liftPrev) {
                    /* the previous character on the old copy has
                     * a correspondent on the current copy which
                     * may get lift-damaged; mark it
                     */
                    if (lastcurrix >= 0 && lastcurrix < txt->whole.nChars)
                        txt->items[lastcurrix].damaged = TRUE;
                }
                if (oldix && !lift && liftPrev) {
                    /* like above only left-edge damage to the current
                     * rather than right-edge damage to the prev
                     */
                    if (currix >= 0 && currix < txt->whole.nChars) {
                        txt->items[currix].damaged = TRUE;
                    }
                }
            }
            lastcurrix = currix;
            if (hasCorrespondent)
                currscan++;
            
            if (txAp->indent &&
                axtHasBullet(txt->old, txAp, oldpc,
                             txt->editState.editPos + oldix)) {
                
                /* determine whether at this text position
                 * there's a bullet which needs to be lifted
                 */
                int liftb = FALSE;
                int lpos;
                
                if (!hasCorrespondent) {
                    /* lift the bullet if the current-txt char
                     * associated with it is gone; but not for
                     * deletion of last (non-tab) character.
                     */
                    if (txt->whole.nChars > 0) {
                        liftb = TRUE;
                    }
                    else if (txt->old->items[0].charcode == TABCHAR) {
                        /* last char was deleted, but was a tab; this
                         * probably affected the position of any bullet
                         */
                        liftb = TRUE;
                    }
                }
                else if (txt->items[currix].moved) {
                    liftb = TRUE;   /* correspondent has moved */
                }
                else {
                    txAttr *newTxAp = attrSegAtPos(txt, currix, &lpos);
                    if (!axtHasBullet(txt, newTxAp,
                                      &txt->items[currix], currix)) {
                        
                        /* correspondent's att-seg no longer has bullet
                         */
                        liftb = TRUE;
                    }
                    else if (newTxAp->indent && 
                             newTxAp->indent->bullet &&
                             newTxAp->indent->bullet->changed) {
                        
                        liftb = TRUE;  /* has bullet, but changed */
                    }
                }
                if (liftb) {
                    liftBullet(txt->old, w, txAp, axLiftGC, &oldpoint);
                }
            }

/* call lifter as needed, based on these types of transitions:
 * 1) whether current character is 1st in its attribute-segment,
 * 2) whether current character has an position offset requiring a
 * separate lift call, and 3) whether current char is the 1st of a
 * to-be-lifted or not-to-be-lifted run of characters.
 */
            if ((oldix == charix) || axtHasOffset(oldpc, &prevSup)) {
                if (inLiftage && liftlen && (pieceIx >= 0)) {
                    liftText(txt->old, w, axLiftGC, pieceAp, pieceIx,
                             liftlen, piecePos.x, piecePos.y);
                    if (txAp->ruling || txAp->strike_thru) { 
                        drawRuling(txt->old, w, pieceAp, pieceIx, liftlen,
                                   &piecePos, &rulept, TRUE);
                    }
                }
                inLiftage = FALSE;
                liftlen = 0;
                
                /* piecePos, rulept, pieceIx should be off
                 */
                piecePos.x = piecePos.y = 0;
                rulept.x = rulept.y = 0;
                pieceIx = -1;
            }
            prevSup = oldpc->sup;
            if (lift) {
                if (inLiftage) {
                    liftlen++;
                }
                else {
                    /* init piecePos, pieceIx, rulepoint, describing this char
                     * as the start of a new to-be-lifted run
                     */
                    liftlen = 1;
                    pieceIx = oldix;
                    piecePos.x = oldpoint.x + oldpc->sup.x;
                    piecePos.y = oldpoint.y + oldpc->sup.y;
                    pieceAp = txAp;
                    rulept = piecePos;
                    inLiftage = TRUE;
                }
            }
            else {
                if (inLiftage) {
                    inLiftage = FALSE;
                    if (liftlen && (pieceIx >= 0)) {
                        liftText(txt->old, w, axLiftGC, pieceAp, pieceIx,
                                 liftlen, piecePos.x, piecePos.y);
                        if (txAp->ruling || txAp->strike_thru) { 
                            drawRuling(txt->old, w, pieceAp, pieceIx, liftlen,
                                       &piecePos, &rulept, TRUE);
                        }
                    }
                    /* turn off piecePos, rulept, pieceIx
                     */
                    piecePos.x = piecePos.y = 0;
                    rulept.x = rulept.y = 0;
                    pieceIx = -1;
                }
            }
            
            oldpoint.x += oldpc->advance.x;
            oldpoint.y += oldpc->advance.y;
        }
        
        charcount -= txAp->nChars;
        charix += txAp->nChars;
        if (charcount > 0) {
            txAp++;
        }
    }
    
    if (inLiftage) {
        if (liftlen && (pieceIx >= 0)) {
            liftText(txt->old, w, axLiftGC, pieceAp, pieceIx, liftlen,
                     piecePos.x, piecePos.y);
            if (txAp->ruling || txAp->strike_thru) { 
                drawRuling(txt->old, w, pieceAp, pieceIx, liftlen,
                           &piecePos, &rulept, TRUE);
            }
        }
    }
}       

void
tx_java_fillrect(gContext, rect, pat_id, fg,bg, x0,y0, clip)
void *gContext;              /* in */
TmRect32 *rect;
int pat_id;             /* fill pattern id */
int x0,y0;              /* tiling origin */
int fg,bg;              /* pattern fg & bg */
TmRect32 *clip;
{
#if ALLOW_JAVA_GE && !FOR_FILTER
	graphics_main_type *gfx;
	gfx = (graphics_main_type *)gContext;
	if ( gfx && gfx->state.java_display )
		gJavaFillrect(gfx, rect, pat_id, TMS_X_D, fg, bg, x0, y0, clip);
#endif
}

static void
liftBullet(tx, widget_handle, txAp, gc, pos)
axtext *tx;              /* in */
long widget_handle;      /* in: get drawable from this */
txAttr *txAp;            /* in: att-seg this bullet falls under */
GC gc;                   /* in */
p32 *pos;                /* in: setpos of char associated with bullet */
{
	int x_off, y_off;
	Drawable drw;
	TmRect32 r;
	r32 brect;

	drw = AxGetDrawableFromWidget((AXWidget)widget_handle, &x_off, &y_off);
	axtBulletRect(tx, txAp, pos, &brect);
	if (brect.width && brect.height)
	{
		r.x_position = milsToPix(brect.origin.x, ScreenRes);
		r.y_position = milsToPix(brect.origin.y, ScreenRes);
		r.x_size = milsToPix(brect.width, ScreenRes);
		r.y_size = milsToPix(brect.height, ScreenRes);
#if ALLOW_JAVA_GE && !FOR_FILTER
		{
			int fg;
			fg = tx->editState.javaPixel;
			tx_java_fillrect(tx->gContext, &r, 5, fg, fg, 0, 0, &r);
		}
#endif
		XFillRectangle(Dpy, drw, gc, r.x_position, r.y_position,
		               r.x_size, r.y_size);
    }
}

static int
needLift(oldpc, newpc, oldpos, newpos, hasMoved)
perChar *oldpc, *newpc;
p32 *oldpos, *newpos;
int *hasMoved; /* return; true == still there, but has moved */
{
    *hasMoved = FALSE;
    
    if (!oldpc->deleted) {
        *hasMoved = (oldpos->x != newpos->x) || (oldpos->y != newpos->y);
    }
    if (CHAR_IS_DIMENSIONLESS(oldpc->charcode)) {
        return FALSE; /* don't lift what was never drawn */
    }
    if (oldpc->hidden) {
        return FALSE; /* don't lift if it was invisible */
    }
    else if (oldpc->deleted) {
        return TRUE;  /* gone now */
    }
    else if (newpc->changed) {
        return TRUE;  /* looks different. */
    }
    else if (!(*hasMoved)) {
        return FALSE; /* still there, looks the same, and hasn't moved */
    }
    *hasMoved = TRUE;
    return TRUE; /* do the lift */
}

static void
liftText(txt, w, gc, txAp, in_startix, in_len, x, y)
axtext *txt;
long w;
GC gc;
txAttr *txAp;
int in_startix;
int in_len;
int x, y; /* mils */
{
    int ix, tx, ty, startix, len;
    int minx, miny, maxx, maxy;
    unsigned int wid, ht;
    p32 currentpoint;
    perChar *pc;
    Drawable drw;
    
    len = in_len;
    startix = in_startix;
    axtTrimText(txt, &startix, &len);

    if (!len)
        return;
    
    pc = &txt->items[startix];

    /* get the drawable */
    drw = AxGetDrawableFromWidget((AXWidget)w, &minx, &miny);
    
    for (ix = 0, currentpoint.x = x, currentpoint.y = y; ix < len; ix++) {
        
        if (txAp->qDraw) {
            /* use screenfont dimensions as much as possible
             */
            tx = milsToPix(currentpoint.x, ScreenRes);
            ty = milsToPix(currentpoint.y, ScreenRes);
            tx += pc->sbox.origin.x;
            ty += pc->sbox.origin.y;
            wid = pc->sbox.width;
            ht = pc->sbox.height;
            /* this extra xpos adjustment parallels adjustment in qDrawText.
             */
            tx -= pc->sbox.origin.x - milsToPix(pc->orig.x, ScreenRes);
        }
        else {
            tx = milsToPix(currentpoint.x + pc->orig.x, ScreenRes);
            ty = milsToPix(currentpoint.y + pc->orig.y, ScreenRes);
            wid = (unsigned int) milsToPix(pc->width, ScreenRes);
            ht = (unsigned int) milsToPix(pc->height, ScreenRes);
        }
        if (txAp->ruling || txAp->strike_thru) { 
            int ttx, tty, twid, tht;
            
            ttx = axtPixToMils(tx, ScreenRes);
            tty = axtPixToMils(ty, ScreenRes);
            twid = axtPixToMils(wid, ScreenRes);
            tht = axtPixToMils(ht, ScreenRes);
            adjustRectForRuling(txAp, pc, &ttx, &tty, &twid, &tht);
            tx = milsToPix(ttx, ScreenRes);
            ty = milsToPix(tty, ScreenRes);
            wid = milsToPix(twid, ScreenRes);
            ht = milsToPix(tht, ScreenRes);
        }

        if (!ix) {
            minx = tx;
            miny = ty;
            maxx = tx + wid;
            maxy = ty + ht;
        }
        else {
            if (tx < minx)
                minx = tx;
            if (ty < miny)
                miny = ty;
            if (tx + (int) wid > maxx)
                maxx = tx + wid;
            if (ty + (int) ht > maxy)
                maxy = ty + ht;
        }

        currentpoint.x += pc->advance.x;
        currentpoint.y += pc->advance.y;
        
        pc++;
        if (ix < len) {
            currentpoint.x += pc->offset.x + pc->kern.x;
            currentpoint.y += pc->offset.y + pc->kern.x;
        }
    }
    
    if (txAp->qDraw) {
        /* fudge the edges outward a bit;  this is because
         * the qDrawText function per-char positions text chars
         * to the screen, and since the lift pattern doesn't match
         * the set pattern in general, we can't exactly predict the
         * placement of the right/left edges of the lift range..
         */
        maxx++;
        minx--;
        maxy++;
        miny--;
    }
    else {
        /* for sizes above some wild guess, increase fudge
         * liberally; this is primarily for the largeXAlternate
         * case, where we're using an outline font, opened using
         * X11R6-style xform spec, but lifting using .afm/.tfm metrics.
         */
        if (maxx - minx > 30) {
            maxx += 3;
            minx -= 3;
        }
        if (maxy - miny > 30) {
            maxy += 3;
            miny -= 3;
        }
    }
    
#if ALLOW_JAVA_GE && !FOR_FILTER
	{
		TmRect32 r;
		int fg;

		r.x_position = minx;
		r.y_position = miny;
		r.x_size = maxx - minx;
		r.y_size = maxy - miny;
		fg = txt->editState.javaPixel;
		tx_java_fillrect(txt->gContext, &r, 5, fg, fg, 0, 0, &r);
	}
#endif
    if (txt->editState.background) {
        /* use the background pixmap to lift
         */
        XCopyArea(Dpy, txt->editState.background, drw, gc,
                  minx - txt->editState.backRect.origin.x,
                  miny - txt->editState.backRect.origin.y,
                  maxx - minx, maxy - miny, minx, miny);
    }
    else {
        /* lift by filling with background color
         */
        XFillRectangle(Dpy, drw, axLiftGC, minx, miny,
                       maxx - minx, maxy - miny);
    }
}

static void
setLiftColor(txt, gc)
axtext *txt;
GC gc;
{
    if (txt->editState.backgroundPixel) {
        XSetForeground(Dpy, gc, txt->editState.backgroundPixel);
    }
    else {
        /* the work area background color */
        XSetForeground(Dpy, gc, AlisPixels[WHITE]);
    }
}

static int
adjustRectForRuling(txAp, pc, tx, ty, wid, ht)
txAttr *txAp;  /* attr seg current char is in */
perChar *pc;   /* perChar struct for current char */
int *tx, *ty, *wid, *ht; /* out: mils; may be adjusted */
{
    int strikey, undery, tmp, lwid, lht;
    int xadvance;
    r32 lrect, refrect;
    b32 lbox, tbb;
    p32 tp;
    void reverseSXRT(), forwardSXRT();

    if (txAp->ruling == FALSE && txAp->strike_thru == FALSE )
        return FALSE;

            /* init undery, position of bottom edge of underline,
             * with 1/2 strokewidth extra. in milliems */
    undery = txAp->fontAttr.underlinePosition +
        txAp->fontAttr.underlineThickness;
    if (txAp->ruling & AXRULING_DOUBLE)
        undery += 3 * txAp->fontAttr.underlineThickness;

            /* init strikey, position of top edge of strikethru,
             * with 1/2 strokewidth extra, in mills */
    tmp = emToMils(txAp->fontAttr.underlineThickness,
                  txAp->fontAttr.ptsizeInt);
    strikey = strikeThruOffset(txAp) - tmp;

        /* get a copy of the unadjusted bounder to use as
           a reference; this copy is pc->bbox's bounder plus
           whatever sub/sup offset may be present */
    refrect.origin.x = pc->orig.x + pc->sup.x;
    refrect.origin.y = pc->orig.y + pc->sup.y;
    refrect.width = pc->width;
    refrect.height = pc->height;

        /* write lrect as the detransformed bounder */
    if (txAp->tctm.unity) {
        lrect.origin.x = pc->orig.x + pc->sup.x;
        lrect.origin.y = pc->orig.y + pc->sup.y;
        lrect.width = pc->width;
        lrect.height = pc->height;
        xadvance = pc->advance.x;
        refrect = lrect;
    }
    else {
            /* copy pc->bbox onto tbb, adding in sup offset */
        tbb = pc->bbox;
        tbb.p0.x += pc->sup.x;
        tbb.p0.y += pc->sup.y;
        tbb.p1.x += pc->sup.x;
        tbb.p1.y += pc->sup.y;
        tbb.p2.x += pc->sup.x;
        tbb.p2.y += pc->sup.y;
        tbb.p3.x += pc->sup.x;
        tbb.p3.y += pc->sup.y;
            /* detransform tbb onto lrect */
        reverseSXRT(tbb.p0.x, tbb.p0.y, &lrect.origin.x,
                    &lrect.origin.y, &txAp->tctm);
        reverseSXRT(tbb.p1.x, tbb.p1.y, &tp.x, &tp.y,
                    &txAp->tctm);
        lrect.width = tp.x - lrect.origin.x + 1;
        reverseSXRT(tbb.p2.x, tbb.p2.y, &tp.x, &tp.y,
                    &txAp->tctm);
        lrect.height = tp.y - lrect.origin.y + 1;
            /* get untransformed advance, too */
        reverseSXRT(pc->advance.x, pc->advance.y,
                    &xadvance, &tmp, &txAp->tctm);
    }

        /* do the adjustments to lrect */
    if (txAp->strike_thru ) {
        if ((tmp = lrect.origin.y - strikey) > 0) {
               /* bounder top is below strikey; pull up */
            lrect.origin.y -= tmp + 1;
            lrect.height += tmp + 1;
        }
    }
    if (txAp->ruling & AXRULING_SINGLE ||
        txAp->ruling & AXRULING_DOUBLE) {
        undery = emToMils(undery, txAp->fontAttr.ptsizeInt);
        if ((tmp = undery - (lrect.origin.y + lrect.height)) > 0) {
                /* bounder bot is below strikey; pull down */
            lrect.height += tmp + 1;
        }
    }
    if ((tmp = xadvance - (lrect.origin.x + lrect.width)) > 0) {
            /* make sure right edge of rect contains the advance */
        lrect.width += tmp;
    }
    if (lrect.origin.x > 0) {
            /* make sure left edge of rect contains set position */
        tmp = lrect.origin.x;
        lrect.origin.x = 0;
        lrect.width += tmp;
    }

    if (txAp->tctm.unity) {
            /* return lrect as the adjusted bounder */
        *tx += lrect.origin.x - refrect.origin.x;
        *ty += lrect.origin.y - refrect.origin.y;
        *wid += lrect.width - refrect.width;
        *ht += lrect.height - refrect.height;
    }
    else {
            /* forward-transform lrect onto lbox; return the
               hv-bounder of lbox as the adjusted bounder */
        forwardSXRT(lrect.origin.x, lrect.origin.y,
                    &lbox.p0.x, &lbox.p0.y, &txAp->tctm);
        forwardSXRT(lrect.origin.x + lrect.width - 1, lrect.origin.y,
                    &lbox.p1.x, &lbox.p1.y, &txAp->tctm);
        forwardSXRT(lrect.origin.x + lrect.width - 1,
                    lrect.origin.y + lrect.height - 1,
                    &lbox.p2.x, &lbox.p2.y, &txAp->tctm);
        forwardSXRT(lrect.origin.x, lrect.origin.y + lrect.height - 1,
                    &lbox.p3.x, &lbox.p3.y, &txAp->tctm);
        axtGetBoxBounder(&lbox, &lwid, &lht, &tp);

        *tx += tp.x - refrect.origin.x;
        *ty += tp.y - refrect.origin.y;
        *wid += lwid - refrect.width;
        *ht += lht - refrect.height;
    }
    return TRUE;
}

void
axHiliteRect(drw, txt, lastxform, lastpoint, minx, miny, maxx, maxy)
Drawable drw;
axtext *txt;
axform *lastxform;
p32 *lastpoint;
int minx, miny;
int maxx, maxy;
{
    XPoint hiRect[4];
    b32 bb;

    /* fudge the region outward a bit */
    minx -= 2;
    miny -= 2;
    maxx += 2;
    maxy += 2;

    lastpoint->x += txt->whole.pos.x;
    lastpoint->y += txt->whole.pos.y;

    /* transform the accumulated bounder, then do hilite fill */
    forwardSXRT(minx, miny, &bb.p0.x, &bb.p0.y, lastxform);
    forwardSXRT(maxx, miny, &bb.p1.x, &bb.p1.y, lastxform);
    forwardSXRT(maxx, maxy, &bb.p2.x, &bb.p2.y, lastxform);
    forwardSXRT(minx, maxy, &bb.p3.x, &bb.p3.y, lastxform);

    hiRect[0].x = (short) milsToPix(bb.p0.x + lastpoint->x, ScreenRes);
    hiRect[0].y = (short) milsToPix(bb.p0.y + lastpoint->y, ScreenRes);
    hiRect[1].x = (short) milsToPix(bb.p1.x + lastpoint->x, ScreenRes);
    hiRect[1].y = (short) milsToPix(bb.p1.y + lastpoint->y, ScreenRes);
    hiRect[2].x = (short) milsToPix(bb.p2.x + lastpoint->x, ScreenRes);
    hiRect[2].y = (short) milsToPix(bb.p2.y + lastpoint->y, ScreenRes);
    hiRect[3].x = (short) milsToPix(bb.p3.x + lastpoint->x, ScreenRes);
    hiRect[3].y = (short) milsToPix(bb.p3.y + lastpoint->y, ScreenRes);

#if ALLOW_JAVA_GE && !FOR_FILTER
	{
		TmPoint32 point[5];
		graphics_main_type *gfx;

        gfx = (graphics_main_type *)txt->gContext;
		if ( gfx && gfx->state.java_display )
		{
			point[0].x = hiRect[0].x;
			point[0].y = hiRect[0].y;
			point[1].x = hiRect[1].x;
			point[1].y = hiRect[1].y;
			point[2].x = hiRect[2].x;
			point[2].y = hiRect[2].y;
			point[3].x = hiRect[3].x;
			point[3].y = hiRect[3].y;
			point[4].x = hiRect[0].x;
			point[4].y = hiRect[0].y;
			gJavaCursor(gfx, 5, point, (TmRect32 *)0);
		}
	}
#endif
    XFillPolygon(Dpy, drw, axTextCursorGC, hiRect, 4, Convex, CoordModeOrigin);
}


void
axtHiliteText(w, txt, pos, len)
long w;
axtext *txt;
int pos, len;
{
    int ix, tlen, dxxpos;
    int x_off, y_off;
    int minx, miny, maxx, maxy, tmp;
    int tx, ty, wid, ht, charsInSeg, preChars;
    txAttr *txAp, *txAp1;
    p32 curpoint, lastpoint, dxsup, dxkern;
    b32 dxbox, bb;
    perChar *pc;
    axform lastxform;
    char *errString;
    int xformChanged = FALSE;
    Drawable drw;

    if (!txt || !txt->items)
        return;
    if (pos < 0 || pos + len > txt->whole.nChars || len < 0)
        return;

    /* get the drawable */
    drw = AxGetDrawableFromWidget((AXWidget)w, &x_off, &y_off);

    if (!txt->measured) {
        if(errString = axtMeasuredText(txt)) {
            freeErrString(errString);
            return;
        }
    }

    txAp = attrSegAtPos(txt, pos, &charsInSeg);
    charsInSeg = txAp->nChars - charsInSeg;

    /* get a count of total num of chars before this segment */
    txAp1 = &txt->atts[0];
    preChars = 0;
    while ( txAp1 != txAp )
    {
        preChars += txAp1->nChars;
        txAp1++;
    }

    getRelativeOffset(txt, 0, pos, &curpoint);
    curpoint.x -= txt->items[pos].sup.x;
    curpoint.y -= txt->items[pos].sup.y;
    lastpoint = curpoint;
    dxxpos = 0;
    lastxform = txAp->tctm;
    miny = minx = MAXLRU; /* very large */
    maxx = maxy = -MAXLRU;
    pc = &txt->items[pos];
    for (ix=0, tlen=0; ix<len; ix++, tlen++, pc++)
    {
        if (pc->hidden)
        {
            break;        /* range stops at first hidden char */
        }
        
        if (ix > 0)
        {
            /* (getRelative has included pos-th's offset, kern) */
            curpoint.x += pc->offset.x;
            curpoint.y += pc->offset.y;
            curpoint.x += pc->kern.x;
            curpoint.y += pc->kern.y;
        }

        if (pc->offset.y || pc->offset.x || xformChanged)
        {
                /* linebreak or transform change */
            if (tlen)
            {
                axHiliteRect(drw, txt, &lastxform, &lastpoint,
                             minx, miny, maxx, maxy);
            }
                /* do resets for possible next hilite seg */
            lastpoint = curpoint;
            tlen = 0;
            xformChanged = FALSE;
            lastxform = txAp->tctm;
            miny = minx = MAXLRU; /* very large */
            maxx = maxy = -MAXLRU;
            dxxpos = 0;
        }

        /* update the untransformed hilite dimensions */
                /* get untransformed character bounder */
        reverseSXRT(pc->bbox.p0.x, pc->bbox.p0.y,
                    &dxbox.p0.x, &dxbox.p0.y, &lastxform);
        reverseSXRT(pc->bbox.p1.x, pc->bbox.p1.y,
                    &dxbox.p1.x, &dxbox.p1.y, &lastxform);
        reverseSXRT(pc->bbox.p2.x, pc->bbox.p2.y,
                    &dxbox.p2.x, &dxbox.p2.y, &lastxform);
        reverseSXRT(pc->bbox.p3.x, pc->bbox.p3.y,
                    &dxbox.p3.x, &dxbox.p3.y, &lastxform);
        /* get untransformed sup and kern */
        reverseSXRT(pc->sup.x, pc->sup.y, 
                    &dxsup.x, &dxsup.y, &lastxform);
        reverseSXRT(pc->kern.x, pc->kern.y, 
                    &dxkern.x, &dxkern.y, &lastxform);
        tx = dxbox.p0.x + dxsup.x + dxxpos;
        ty = dxbox.p0.y + dxsup.y;
        wid = dxbox.p1.x - dxbox.p0.x + 1;
        ht = dxbox.p3.y - dxbox.p0.y + 1;
#if 0
        if (txAp->ruling) { 
                /* need a cheap-n-nasty adjust check for ruling */
        }
#endif
        if (tx < minx)
            minx = tx;
        if (ty < miny)
            miny = ty;
        if ((tmp = tx + wid - 1) > maxx)
            maxx = tmp;
        if ((tmp = ty + ht - 1) > maxy)
            maxy = tmp;
        dxxpos += pc->padvance;

        curpoint.x += pc->advance.x;
        curpoint.y += pc->advance.y;

        /* test for attr-seg boundary */
        charsInSeg--;
        if (charsInSeg <= 0)
        {
            preChars += txAp->nChars;
            /* if there are more chars go to the next segment */
            if ( preChars < txt->whole.nChars )
            {
                txAp++;
                charsInSeg = txAp->nChars;
                if (txAp->tctm.forward.ca != lastxform.forward.ca ||
                    txAp->tctm.forward.cb != lastxform.forward.cb ||
                    txAp->tctm.forward.cc != lastxform.forward.cc ||
                    txAp->tctm.forward.cd != lastxform.forward.cd ||
                    txAp->tctm.forward.ctx != lastxform.forward.ctx ||
                    txAp->tctm.forward.cty != lastxform.forward.cty)
                {
                    xformChanged = TRUE;
                }
            }
        }
    }
    if (tlen)
    {
        axHiliteRect(drw, txt, &lastxform, &lastpoint, minx, miny, maxx, maxy);
    }
}

/* per-char-position: set screenfont glyphs on screen
 * using target widths to determine screen placement.
 * The positions of each screen char should match
 * the spot that the target char would hit if it were
 * scaled onto the screen device.
 * Note that in order to work correctly, this call must
 * be given a text range which has a common, horizontal
 * baseline.  (y-component of character advance is ignored).
 */
static int
qDrawText(dpy, w, gc, txt, txAp, start_index, numChars, pt)
Display *dpy;
long w;
GC gc;
axtext *txt;
txAttr *txAp;
int start_index;
int numChars;  /* # perChar structs */
p32 *pt;
{
    int tx, ty; 
    int ix, delta, startix, nChars, x_off, y_off;
    perChar *pc, *tc;
    p32 currentpoint, screenpoint;
    unsigned long fgCol, bgCol, planeMask;
    Drawable drw;
#if BUILD_WIDE
    wchar *txstr;
#else
    mbchar *txstr;
#endif
    
    tx = pt->x;
    ty = pt->y;
    nChars = numChars;
    startix = start_index;
    axtTrimText(txt, &startix, &nChars);
    
    if (nChars == 0)
        return FALSE;

    /* take a string copy to get from tx_char to mb/wchar types
     */
#if BUILD_WIDE
    txstr = (wchar *) txloc(txt->pid, nChars * sizeof(wchar));
    for (ix = 0; ix < nChars; ix++) {
        txstr[ix] = (wchar) txt->textstr[ix + startix];
    }
#else
    txstr = (mbchar *) txloc(txt->pid, nChars * sizeof(mbchar));
    for (ix = 0; ix < nChars; ix++) {
        txstr[ix] = (mbchar) txt->textstr[ix + startix];
    }
#endif
    
    pc = &txt->items[startix];
    
    /* get the drawable, set font and pixel values
     */
    drw = AxGetDrawableFromWidget((AXWidget)w, &x_off, &y_off);
    axSetFont(Dpy, gc, txAp->fconx.screen->xfont);
    axtSetFgBg(dpy, txAp, &fgCol, &bgCol, &planeMask);
    
    /* sup for 0th char in tx, ty; remove during init of currentpoint
     */
    currentpoint.x = tx - pc[0].sup.x;
    currentpoint.y = ty - pc[0].sup.y;

    /* offset, kern also present in tx, ty; remove x component,
     * which is accumulated back in below loop
     */
    currentpoint.x -= (pc[0].offset.x + pc[0].kern.x);
    
    screenpoint.x = milsToPix(tx, ScreenRes);
    screenpoint.y = milsToPix(ty, ScreenRes);
    
    if ( IS_SOLID_FILL(txt->gContext, txAp->fillpattern) == FALSE ) {
        /* draw patterned style, one character at a time
         */
        int wid, hyt, xoff, yoff;
        Pixmap map;
        
        for (ix = 0, tc = pc; ix < nChars; ix++, tc++) {
            currentpoint.x += tc->offset.x + tc->kern.x;
            /* create a pixmap
             */
            wid = tc->sbox.width;
            hyt = tc->sbox.height;
			if ( wid > 0 && hyt > 0 )
			{
                map = XCreatePixmap(dpy, AxTopWinNeverMapped, wid, hyt, 1);
                XFillRectangle(dpy, map, PixClearGC, 0, 0, wid, hyt);
            
                /* draw the character onto the pixmap
                 */
                xoff = tc->sbox.origin.x * (-1);
                yoff = tc->sbox.origin.y * (-1);
                XSetForeground(dpy, PixGC, 1);
                axSetFont(dpy, PixGC, txAp->fconx.screen->xfont);
#if BUILD_WIDE
                axWDrawString(dpy, map, PixGC, txAp->fconx.screen->xfont,
                              xoff, yoff, &txstr[ix], 1);
#else
                axDrawString(dpy, map, PixGC, txAp->fconx.screen->xfont,
                             xoff, yoff, &txstr[ix], 1);
#endif
                /* Do the masked draw.  Note the additional xpos adjust for
                 * the difference between screen and target left sidebearings.
                 */
                xoff += tc->sbox.origin.x - milsToPix(tc->orig.x, ScreenRes);
                axtDrawText(txt, drw,
                            milsToPix(currentpoint.x + tc->sup.x, ScreenRes) -
                            xoff,
                            milsToPix(currentpoint.y + tc->sup.y, ScreenRes) -
                            yoff,
                            hyt, wid, map, txAp->fillpattern);
                XFreePixmap(Dpy, map);
			}
            currentpoint.x += tc->advance.x;
        }
    }
    else {
        /* draw directly, in segments as large as possible. A new segment
         * starts each time a delta is needed to match target & screen pos.
         */
        int lastscreenx, seglen, lastix, ttx;
        
        XSetState(dpy, gc, fgCol, bgCol, GXcopy, AllPlanes);
        lastscreenx = screenpoint.x;
        lastix = 0;
        for (ix = 0, seglen = 0, tc = pc; ix < nChars; ix++, seglen++, tc++) {
            currentpoint.x += tc->offset.x + tc->kern.x;
            
            /* Test for screen-target delta. Note the additional xpos adjust
             * for the difference between screen and target left sidebearings.
             */
            ttx = milsToPix(currentpoint.x + tc->orig.x, ScreenRes);
            delta = ttx - (screenpoint.x + tc->sbox.origin.x);
            if (delta) {
                if (seglen) {
#if BUILD_WIDE
                    axWDrawString(dpy, drw, gc, txAp->fconx.screen->xfont,
                                  lastscreenx, screenpoint.y,
                                  &txstr[lastix], seglen);
#else
                    axDrawString(dpy, drw, gc, txAp->fconx.screen->xfont,
                                 lastscreenx, screenpoint.y,
                                 &txstr[lastix], seglen);
#endif
                }
                seglen = 0;
                lastix = ix;
                lastscreenx = screenpoint.x + delta;
            }
            currentpoint.x += tc->advance.x;
            screenpoint.x += tc->sadvance + delta;
        }
        if (seglen) {
#if BUILD_WIDE
            axWDrawString(dpy, drw, gc, txAp->fconx.screen->xfont,
                          lastscreenx, screenpoint.y,
                          &txstr[lastix], seglen);
#else
            axDrawString(dpy, drw, gc, txAp->fconx.screen->xfont,
                         lastscreenx, screenpoint.y,
                         &txstr[lastix], seglen);
#endif
        }
    }
    txfree(txstr);
    return TRUE;
}

/*
 * My func to replacing FillStipple() for correspond NEC-EWS HACK!!
 * Ashisuto, 1997.12.11.
 */
#if !FOR_FILTER
drawStipple(dpy, drw, gc, pixmap, wid, hyt, xoff, yoff)
Display               *dpy;
Drawable      drw;
GC            gc;
Pixmap                pixmap;
int           wid, hyt, xoff, yoff;
{
      XImage                  *img;
      unsigned char   *maskPtr;
      int                     x, y, bit_offset, byte_offset;

      img = XGetImage(dpy, pixmap, 0, 0, wid, hyt, 1, XYPixmap);
      maskPtr = img->bitmap_bit_order==LSBFirst ? lsbfBitMask:msbfBitMask;
      for (y = 0; y < img->height; y++) {
              for (x = 0; x < img->width; x++) {
                      byte_offset = (y * img->bytes_per_line) + (x >> 3);
                      bit_offset = x & 7;
                      if (img->data[byte_offset] & maskPtr[bit_offset])
                              XDrawPoint(Dpy, drw, gc, xoff+x, yoff+y);
              }
      }
      XDestroyImage(img);
	  return TRUE;
}
#endif

/*
 * PutImage-based text drawing
 */
static int
drawLetterforms(dpy, w, gc, txt, txAp, start_index, numChars, pt)
Display *dpy;
long w;
GC gc;
axtext *txt;
txAttr *txAp;
int start_index;
int numChars;  /* # perChar structs */
p32 *pt;
{
    int tx, ty;
    int i, wid, ht, gfs;
    int x_off, y_off;
    int startix, nChars;
    unsigned long fgCol, bgCol, planeMask;
    perChar *pc, *tc;
    chSeg cS;
    axfs_ext *ext;
    axformed_lettername gfsLettername;
    p32 currentpoint, imagepoint;
    XImage *txImage;
    Pixmap txPixmap;
    char *errString;
    attr_type *attr;
    Drawable drw;

    nChars = numChars;
    startix = start_index;
    axtTrimText(txt, &startix, &nChars);
    
    if (nChars == 0)
        return FALSE;
    
    tx = pt->x;
    ty = pt->y;
    pc = &txt->items[startix];
    
    /* get the drawable
     */
    drw = AxGetDrawableFromWidget((AXWidget)w, &x_off, &y_off);

    axtSetFgBg(dpy, txAp, &fgCol, &bgCol, &planeMask);
    

    if (IS_AXGFS_FONT(txAp->fconx.screen->xfont)) {
        gfs = TRUE;
        ext = (axfs_ext *) txAp->fconx.screen->xfont->ext_data->private_data;
    }
    else {
        gfs = FALSE;
        axSetFont(Dpy, PixGC, txAp->fconx.screen->xfont);
    }
    
    /* offset, kern, sup for 0th character are in tx,ty
     */
    currentpoint.x = tx - (pc[0].offset.x + pc[0].kern.x + pc[0].sup.x);
    currentpoint.y = ty - (pc[0].offset.y + pc[0].kern.y + pc[0].sup.y);
    for (i=0; i<nChars; i++) {
        tc = &pc[i];
        currentpoint.x += tc->offset.x + tc->kern.x;
        currentpoint.y += tc->offset.y + tc->kern.y;
        
        /* lookup letterform
         */
        txPixmap = (Pixmap) NULL;
        txImage = (XImage *) NULL;
        
        if (gfs) {
            /* lookup gfs-supplied letterform
             */
            axletter *gletter;
            gfsLettername.charcode = (unsigned short) tc->charcode;
            gfsLettername.ctm = tc->ctm->forward;
            axhGetItemI(&gfsLettername, sizeof(axformed_lettername),
                        &ext->com.letters,
                        (char *) txAp->fconx.screen->xfont,
                        (char **) &gletter);
            if (!gletter || !gletter->bits || !gletter->bbox) {
                currentpoint.x += tc->advance.x;
                currentpoint.y += tc->advance.y;
                continue;
            }
            txPixmap = (Pixmap) gletter->bits;
            imagepoint.x =
                milsToPix(currentpoint.x + tc->sup.x, ScreenRes) +
                    gletter->bbox->origin.x;
            imagepoint.y =
                milsToPix(currentpoint.y + tc->sup.y, ScreenRes) +
                    gletter->bbox->origin.y;
            wid = gletter->bbox->width;
            ht = gletter->bbox->height;
        }
        else {
            /* lookup XServer-supplied letterform
             */
            cS.txAp = txAp;
            cS.pc = tc;
            if (tc->letterName) {
                if (AxtLetterformPixmaps) {
                    if (errString = axGetLetterformPixmap
                        (tc->letterName, &cS, &txPixmap)) {
                        freeErrString(errString);
                    }
                }
                else {
                    if (errString = axGetLetterformImage
                        (tc->letterName, &cS, &txImage)) {
                        freeErrString(errString);
                    }
                }
            }
            imagepoint.x = milsToPix(currentpoint.x +
                                     tc->orig.x + tc->sup.x,
                                     ScreenRes);
            imagepoint.y = milsToPix(currentpoint.y +
                                     tc->orig.y + tc->sup.y,
                                     ScreenRes);
            wid = milsToPix(tc->width, ScreenRes);
            ht = milsToPix(tc->height, ScreenRes);
        }
        
        if ( IS_SOLID_FILL(txt->gContext, txAp->fillpattern) == FALSE ) {
            /* text has special fills
             */
            attr = txAp->fillpattern;
            if (txPixmap) {
                /* letterform pixmap can be used directly
                 */
                axtDrawText(txt, drw, imagepoint.x, imagepoint.y, ht, wid,
                            txPixmap, attr);
            }
            else if (txImage) {
                /* need to create a mask from txImage
                 */
                Pixmap map;
                map = XCreatePixmap(dpy, drw, wid, ht, 1);
                XFillRectangle(dpy, map, PixClearGC, 0, 0, wid, ht);
                XPutImage(dpy, map, PixGC, txImage,
                          0, 0, 0, 0,
                          (unsigned int) wid, (unsigned int) ht);
                axtDrawText(txt, drw, imagepoint.x, imagepoint.y, ht, wid,
                            map, attr);
                XFreePixmap(Dpy, map);
            }
        }


        else if (txPixmap) {
             /* Pixmap-based draw */
             /* prepare GC for pixmap-style draw
              */
#if !FOR_FILTER
            extern Pixmap ScratchMap;
            if (getenv("EWS_XFILL_HACK") && (drw == ScratchMap))
			{
                /* Ashisuto added to correspond NEC-EWS 4800, 24 planes
                 * TrueColor Xlib bug. On that display, the Stipple-Fill
                 * is different from Solaris 24 planes TrueColor display,
                 * different from all 8-planes Pseudo-Color displays,
                 * and is even different from NEC-EWS itself when on 8-planes.
                 * The !!EWS_XFILL_HACK!! is take the Tile/Stipple origin
                 * as the absolute thing, not the relative thing.
                 * 1997.12.10
                 */
                XSetState(dpy, gc, fgCol, bgCol, GXcopy, planeMask);
                drawStipple(dpy, drw, gc, txPixmap, wid, ht,
			      				imagepoint.x, imagepoint.y);
            } else
#endif
            {
                XSetFillStyle(dpy, gc, FillStippled);
                XSetStipple(dpy, gc, txPixmap);
                XSetTSOrigin(dpy, gc, imagepoint.x, imagepoint.y);
                XSetState(dpy, gc, fgCol, bgCol, GXcopy, planeMask);
                XFillRectangle(dpy, drw, gc, imagepoint.x, imagepoint.y,
                               (unsigned int) wid, (unsigned int) ht);
                /* restore gc */
                XSetFillStyle(dpy, gc, FillSolid);
			}
        }
        else if (txImage) {
             /* XImage-based draw */
             /* stencil out the letterform foreground */
             XPutImage(dpy,drw,axTextStencilGC, txImage,
                          0, 0, imagepoint.x, imagepoint.y,
                          (unsigned int) wid, (unsigned int) ht);
             if (fgCol != 0) {
                /* now paint the letterform
                * foreground in color
                */
                XSetState(dpy,gc,fgCol,~AllPlanes,GXor,AllPlanes);
                XPutImage(dpy,drw,gc, txImage,
                          0, 0, imagepoint.x, imagepoint.y,
                          (unsigned int) wid, (unsigned int) ht);
            }
        }
        currentpoint.x += tc->advance.x;
        currentpoint.y += tc->advance.y;

#if AXMS
    if ( txImage && txImage->bitmap_handle )
        {
        MS_DeleteObject(txImage->bitmap_handle);
        txImage->bitmap_handle = NULL;
        }
#endif
    }
    return(TRUE);
}

static void
axtSetFgBg(dpy, txAp, fgCol, bgCol, planeMask)
Display *dpy;
txAttr *txAp;
unsigned long *fgCol, *bgCol, *planeMask;          /* out */
{

    /* lookup pixels; type 2 is CMYK, max value
     * assumed on txAp is 255
     */
    if (txAp->useFgPixel) {
        /* foreground pixel value stored directly on *txAp
         */
        *fgCol = txAp->fgPixel;
    }
    else if (isBlackCMYK(txAp->fillcolor) &&
             useForegroundColorForBlack(txAp->fillpattern)) {
        
        /* display black as the work area foreground color
         */
        *fgCol = (unsigned long) AlisPixels[BLACK];
    }
    else {
        AxGetRealPixel(2, axCyanComponent(txAp->fillcolor),
                       axMagentaComponent(txAp->fillcolor),
                       axYellowComponent(txAp->fillcolor),
                       axBlackComponent(txAp->fillcolor),
                       255, fgCol);
    }
    
    AxGetRealPixel(2, axCyanComponent(txAp->fillbcolor),
                   axMagentaComponent(txAp->fillbcolor),
                   axYellowComponent(txAp->fillbcolor),
                   axBlackComponent(txAp->fillbcolor),
                   255, bgCol);
    *planeMask = AllPlanes;
}

    /* the cmyk value is packed: 8bits each in order: c, m, y, k
       it's a hack; these vals should be independent integers */
int
axCyanComponent( cmyk )
int cmyk;
{
    return ((cmyk >> 24) & 0xff);
}

int
axMagentaComponent( cmyk )
int cmyk;
{
    return ((cmyk >> 16) & 0xff);
}

int
axYellowComponent( cmyk )
int cmyk;
{
    return ((cmyk >> 8) & 0xff);
}

int
axBlackComponent ( cmyk )
int cmyk;
{
    return (cmyk & 0xff);
}

static int
isWhiteCMYK( cmyk )
int cmyk;
{
    return ((axCyanComponent(cmyk) == 0) &&
            (axMagentaComponent(cmyk) == 0) &&
            (axYellowComponent(cmyk) == 0) &&
            (axBlackComponent(cmyk) == 0));
}

int
isBlackCMYK( cmyk )
int cmyk;
{
    return (axBlackComponent(cmyk) == 255);
}


/*
 * Beware unprotected access of cache objects fconx.target,
 * fconx.screen -- but, called only from axtDisplayableText
 * where those objects are looked up first.
 */
static int
getLetterformInfo(txAp, items, firstix, lastix, pid)
txAttr *txAp;
perChar *items;
int firstix, lastix; /* both within txAp's character range */
int pid;
{
    int tcharix;
    perChar *pc;
    b32 tbox;

    for(tcharix=firstix ;tcharix <= lastix; tcharix++) {
        pc = &items[tcharix];
        if (!pc->letterNameKnown) {
                /* convert bbox numbers from mils to screenpix */
            tbox.p0.x = milsToPix(pc->bbox.p0.x, ScreenRes);
            tbox.p0.y = milsToPix(pc->bbox.p0.y, ScreenRes);
            tbox.p1.x = milsToPix(pc->bbox.p1.x, ScreenRes);
            tbox.p1.y = milsToPix(pc->bbox.p1.y, ScreenRes);
            tbox.p2.x = milsToPix(pc->bbox.p2.x, ScreenRes);
            tbox.p2.y = milsToPix(pc->bbox.p2.y, ScreenRes);
            tbox.p3.x = milsToPix(pc->bbox.p3.x, ScreenRes);
            tbox.p3.y = milsToPix(pc->bbox.p3.y, ScreenRes);
            if (txAp->fconx.target->noBearings) {
                    /* substitute overall font numbers, so that the
                     * screen metrics for this char are equally
                     * handicapped */
                pc->sbox.origin.y = -txAp->fconx.screen->xfont->ascent;
                pc->sbox.height = txAp->fconx.screen->xfont->ascent +
                        txAp->fconx.screen->xfont->descent;
            }
            
            getTransform(&pc->sbox, &tbox, &txAp->transform,
                         IS_AXGFS_FONT(txAp->fconx.screen->xfont), pc, pid);
            
                /* form lookup name of letterform */
            if (pc->letterName)
                txfree((char *)pc->letterName);
            pc->letterName = formLetterName(txAp, pc, pid);
            pc->letterNameKnown = TRUE;
        }
    }
    return TRUE;
}

/* if the letterform cache has lots of items with these
 * numbers embedded in the names, something's wrong */
#define ILIMT (100000)
#define FLIMT (100000.0)

/*
 * Beware unprotected access of cache object fconx.font --
 * but, called only from axtDisplayableText (via getLetterFormInfo)
 * where that object is looked up first.
 */
static char * /* caller must free */
formLetterName(txAp, pc, pid)
txAttr *txAp;
perChar *pc;
int pid;
{
    int totalLen, tpoint, tmp;
    int dashcount;
    float dtmp;
    char scratch[512];
    char *xp, *txp;
    char *cp = scratch;
    /*
     * letterform or letterform-chunk name formation rules:
     *
     *  globalName charcode(s) pointsize[ (ctm elements) ca  cb  cc  cd ]
     *    char *    decint(s)    decint                  %f  %f  %f  %f
     *
     * spaces separate the fields.  the decints in the charcode string
     * are separated by periods if there are more than one charcode..
     * if any of the ctm matrix elements are nonunity, then all must
     * included either as some value or as 'u' for unity.
     */
    totalLen = strlen(txAp->fconx.font->globalName);
        /* charcode(s): just do single chars for now */
    sprintf(cp, " %d ", (int)pc->charcode);
    tmp = strlen(cp);
    cp += tmp;
    totalLen += tmp;

        /* pointsize */
#if 0
                             /* this is what we are trying to do */
    tpoint = axtDecipointSize(txAp->screenFontName);
    if (tpoint > ILIMT)
        tpoint = ILIMT;
    sprintf(cp, " %d", tpoint);
    tmp = strlen(cp);
    cp += tmp;
#else
                             /* but do it inline using a few lines from
                                axtDecipointSize; note we're not bothering
                                to check for a well formed name here */
    cp[0] = '0'; /* default */
    cp[1] = '\0';
    tmp =  1;    /* length of default */
    for (xp = txAp->screenFontName, dashcount = 0; *xp != '\0'; xp++) {
        if(*xp == '-') {
            dashcount++;
        }
        if(dashcount == 8) {
                /* we think next field is decipoint size; copy */
            xp++;
            txp = xp;
            tmp = 0;
            while(*txp != '-' && (cp - scratch) < 500) {
                *cp++ = *txp++;
                tmp++;
            }
            *cp = '\0';
            break;
        }
    }
#endif
    totalLen += tmp;

        /* transform terms */
    if (!pc->ctm->unity) {
        if (pc->ctm->reverse.ca != 1.0) {
            if ((dtmp = pc->ctm->reverse.ca) > FLIMT)
                dtmp = 0.0;
            sprintf(cp, " %.5f", dtmp);
            tmp = strlen(cp);
            totalLen += tmp;
            cp += tmp;
        }
        else {
            strcat(cp, " u");
            tmp = strlen(cp);
            totalLen += tmp;
            cp += tmp;
        }
        if (pc->ctm->reverse.cb != 0.0) {
            if ((dtmp = pc->ctm->reverse.cb) > FLIMT)
                dtmp = 0.0;
            sprintf(cp, " %.5f", dtmp);
            tmp = strlen(cp);
            totalLen += tmp;
            cp += tmp;
        }
        else {
            strcat(cp, " u");
            tmp = strlen(cp);
            totalLen += tmp;
            cp += tmp;
        }
        if (pc->ctm->reverse.cc != 0.0) {
            if ((dtmp = pc->ctm->reverse.cc) > FLIMT)
                dtmp = 0.0;
            sprintf(cp, " %.5f", dtmp);
            tmp = strlen(cp);
            totalLen += tmp;
            cp += tmp;
        }
        else {
            strcat(cp, " u");
            tmp = strlen(cp);
            totalLen += tmp;
            cp += tmp;
        }
        if (pc->ctm->reverse.cd != 1.0) {
            if ((dtmp = pc->ctm->reverse.cd) > FLIMT)
                dtmp = 0.0;
            sprintf(cp, " %.5f", dtmp);
            tmp = strlen(cp);
            totalLen += tmp;
            cp += tmp;
        }
        else {
            strcat(cp, " u");
            tmp = strlen(cp);
            totalLen += tmp;
            cp += tmp;
        }
    }
    cp = (char *)txloc(pid, totalLen+1);
    strcpy(cp, txAp->fconx.font->globalName);
    strcat(cp, scratch);
    return(cp);
}

/* : getTransform - compares source and destination
 * corners to determine the transform that will map
 * source onto destination; used for getting the
 * screenfont-to-target transform */
static void
getTransform(sbox, xbox, terms, gfs, pc, pid)
r32 *sbox;           /* in: source rect */
b32 *xbox;           /* in: transformed pointset */
axformTerms *terms;  /* in */
int gfs;             /* in: bool; whether for gallium-style access */
perChar *pc;         /* in/out: fill in ctm member */
int pid;
{
    axform *f;
    float xwid, sinAng, cosAng, hx, tx, ty, sx, sy;
    p32 tp0, tp3;
    int nrwid, nrht; /* bounding wid, ht of transformed pts */
    int tmp;

    if (!pc->ctm) {
        pc->ctm = (axform *)txloc(pid, sizeof(axform));
    }
    f = pc->ctm;
    nrwid = milsToPix(pc->width, ScreenRes);
    nrht = milsToPix(pc->height, ScreenRes);

    if (nonrotated(xbox)) {
        f->rotated = FALSE;
        sinAng = 0.0;
        cosAng = 1.0;
        hx = 0.0;
        if (sbox->width > 0) {
            sx = (float) nrwid /
                    (float) sbox->width;
            if (xbox->p1.x < xbox->p0.x)
                sx = -(sx);
        }
        else {
            sx = 1.0;
        }
        if (sbox->height > 0) {
            sy = (float) nrht / (float) sbox->height;
            if (xbox->p3.y < xbox->p0.y)
                sy = -(sy);
        }
        else {
            sy = 1.0;
        }
            /* write translation in screen pix units */
        tx = milsToPix(pc->bbox.p0.x - pc->orig.x, ScreenRes);
        ty = milsToPix(pc->bbox.p0.y - pc->orig.y, ScreenRes);
    }
    else {
            /* no yshear by definition, so xbox's width can be read
             * directly */
        f->rotated = TRUE;
        xwid = (float) hypot((double)(xbox->p1.x - xbox->p0.x),
                    (double)(xbox->p1.y - xbox->p0.y));
        if (sbox->width == 0) {
            sx = 1.0;
        }
        else {
            sx = (float)(xwid) / (float)(sbox->width);
        }
            /* translate xbox so that p0 is on the origin */
        xbox->p1.x -= xbox->p0.x;
        xbox->p1.y -= xbox->p0.y;
        xbox->p2.x -= xbox->p0.x;
        xbox->p2.y -= xbox->p0.y;
        xbox->p3.x -= xbox->p0.x;
        xbox->p3.y -= xbox->p0.y;
        xbox->p0.x = xbox->p0.y = 0;
            /* extract angle */
        if (xwid != 0.0) {
            sinAng = (float)(xbox->p1.y - xbox->p0.y) / xwid;
            cosAng = (float)(xbox->p1.x - xbox->p0.x) / xwid;
        }
        else {
            sinAng = 0.0;
            cosAng = 1.0;
        }
            /* derotate pts 0,3 and compare to get xshear and yscale */
        tp0.x = tp0.y = 0; /* derotate of 0,0 is 0,0 */
        axtScaleInvR(xbox->p3.x, xbox->p3.y, &tp3.x, &tp3.y, sinAng, cosAng);
        if ((tp3.y - tp0.y) == 0) {
            hx = 0.0;
        }
        else if ((tmp = tp3.x - tp0.x) != 0) {
            hx = (float)tmp / (float)(tp3.y - tp0.y);
        }
        else {
            hx = 0.0;
        }
        if (sbox->height == 0) {
            sy = 1.0;
        }
        else {
            sy = (float)(tp3.y - tp0.y) / (float) sbox->height;
        }
            /* write translation in screen pix units */
        tx = milsToPix(pc->bbox.p0.x - pc->orig.x, ScreenRes);
        ty = milsToPix(pc->bbox.p0.y - pc->orig.y, ScreenRes);
    }
    if (sx==1.0 && sy==1.0 && cosAng==1.0 && sinAng==0.0 &&
        hx==0.0) {
        f->unity = TRUE;
    }
    else {
        f->unity = FALSE;
    }
    
    if (gfs) {
        /* write elements of ctm for passage to gfs-rasterizer;
         * it's the forward transform that's of interest for this scaling;
         * This code should match that in writeForwardCtm, except that
         * sine terms and shears are negated compared to writeForwardCtm.
         * (This because our internal coord system is y-increasing-downwards).
         */
        tmp = (int) (sx * 100000);
        if (abs(tmp) > 1) {
            f->forward.ca = sx * cosAng;
            f->forward.cc = sy * ((-hx * cosAng) + sinAng);
            f->forward.ctx = (-hx * ((sy * cosAng) + (sx * sinAng)) / sx) -
                ((tx * cosAng) - (ty * sinAng)) / sx;
        }
        else {
            /* filter scale factors < 1/100000 to .001
             */
            f->forward.ca = f->forward.cc = 0.001; 
            f->forward.ctx = 0;
        }
        tmp = (int) (sy * 100000);
        if (abs(tmp) > 1) {
            f->forward.cb = -(sx * sinAng);
            f->forward.cd = sy * (cosAng + (hx * sinAng));
            f->forward.cty = -((tx * sinAng) + (ty * cosAng)) / sy;
        }
        else {
            f->forward.cb = f->forward.cd = 0.001;
            f->forward.cty = 0;
        }

        /* This is a tmp fix for a crash on alpha where formLetterName uses
         * reverse transformation matrix regardless of whether it is a gallium
         * font struct or xfont struct.           vamsi 1/29/96
         */
        tmp = (int) (sx * 100000);
        if (abs(tmp) > 1) {
            f->reverse.ca = (cosAng + (hx * sinAng)) / sx;
            f->reverse.cc = (sinAng - (hx * cosAng)) / sx;
            f->reverse.ctx = (hx * ((ty * cosAng) - (tx * sinAng))) / sx -
                ((tx * cosAng) + (ty * sinAng)) / sx;
        }
        else {
            /* filter scale factors < 1/100000 to .001
             */
            f->reverse.ca = f->reverse.cc = 0.001; 
            f->reverse.ctx = 0;
        }
        tmp = (int) (sy * 100000);
        if (abs(tmp) > 1) {
            f->reverse.cb = -(sinAng) / sy;
            f->reverse.cd = cosAng / sy;
            f->reverse.cty = ((tx * sinAng) - (ty * cosAng)) / sy;
        }
        else {
            f->reverse.cb = f->reverse.cd = 0.001;
            f->reverse.cty = 0;
        }

    }
    else {
        /* write elements of ctm for raster-scaling; it's the
         * reverse transform that's of interest for raster scaling;
         * this code to match that in writeReverseCtm
         */
        tmp = (int) (sx * 100000);
        if (abs(tmp) > 1) {
            f->reverse.ca = (cosAng + (hx * sinAng)) / sx;
            f->reverse.cc = (sinAng - (hx * cosAng)) / sx;
            f->reverse.ctx = (hx * ((ty * cosAng) - (tx * sinAng))) / sx -
                ((tx * cosAng) + (ty * sinAng)) / sx;
        }
        else {
            /* filter scale factors < 1/100000 to .001
             */
            f->reverse.ca = f->reverse.cc = 0.001; 
            f->reverse.ctx = 0;
        }
        tmp = (int) (sy * 100000);
        if (abs(tmp) > 1) {
            f->reverse.cb = -(sinAng) / sy;
            f->reverse.cd = cosAng / sy;
            f->reverse.cty = ((tx * sinAng) - (ty * cosAng)) / sy;
        }
        else {
            f->reverse.cb = f->reverse.cd = 0.001;
            f->reverse.cty = 0;
        }
    }
}

static void
extractYscale(targetCtm, yscale)
axform *targetCtm; /* composite forward transform */
float *yscale; /* out: report yscale component of composite xform */
{
    p32 inp, outp;
    if (targetCtm->unity) {
        *yscale = 1.0;
        return;
    }
    inp.x = 0;
    inp.y = 100000;
    forwardSXRT(inp.x, inp.y, &outp.x, &outp.y, targetCtm);
    *yscale = (float) hypot((double) outp.x, (double) outp.y) / 100000.0;
    return;
}

static void
extractXscale(targetCtm, xscale)
axform *targetCtm; /* composite forward transform */
float *xscale; /* out: report yscale component of composite xform */
{
    p32 inp, outp;
    if (targetCtm->unity) {
        *xscale = 1.0;
        return;
    }
    inp.x = 100000;
    inp.y = 0;
    forwardSXRT(inp.x, inp.y, &outp.x, &outp.y, targetCtm);
    *xscale = (float) hypot((double) outp.x, (double) outp.y) / 100000.0;
    return;
}

/* the bbox results from a max-font-height bbox thru some transform.
 * source of bbox is the ax2.0 graphic text data. */
int
getXformFromRev2BBox(txt, textstr, len, pointsize, bbox,
                baseLineOffset, tp, tflags, ht36, ascent36)
axtext *txt;
tx_char *textstr;
int len;
int pointsize; /* 'nominal', in points */
b32 *bbox; /* in */
p32 *baseLineOffset; /* out */
textAttrs *tp; /* in/out */
unsigned int *tflags; /* in/out */
int ht36, ascent36; /* in */
{
    r32 refBox;
    p32 tpt, tp0, tp3;
    int tmp, wid, ht, neg = 0;
    float dshear, dscale, dwid, ratio, sinAng, cosAng;
    char *errString;
        /* in this context, size is not yet known. do a reference
         * measurement at pointsize pt; set the *tp data for real later. */
    tp->pointsize = pointsize * 10000;
    *tflags |= TXpointsize;
    if(!setDefaultTextAttrs(txt, tp, *tflags)) {
        return FALSE;
    }
    if(!insChars(txt, textstr, 0, len)) {
        return FALSE;
    }
        /* get an ax2.0-bounder for the reference text */
    if(errString = getAx2Bounder(txt, 0, len, pointsize,
                                &refBox, &ratio, ht36, ascent36)) {
        txfree((char *)errString);
        return FALSE;
    }
    if (nonrotated(bbox)) {
        ht = bbox->p3.y - bbox->p0.y + 1;
        wid = bbox->p1.x - bbox->p0.x + 1;
    }
    else {
        dwid = (float) hypot((double)(bbox->p1.x - bbox->p0.x),
                    (double)(bbox->p1.y - bbox->p0.y));
            /* translate bbox so that p0 is on the origin */
        bbox->p1.x -= bbox->p0.x;
        bbox->p1.y -= bbox->p0.y;
        bbox->p2.x -= bbox->p0.x;
        bbox->p2.y -= bbox->p0.y;
        bbox->p3.x -= bbox->p0.x;
        bbox->p3.y -= bbox->p0.y;
        bbox->p0.x = bbox->p0.y = 0;
            /* extract angle */
        if (dwid != 0.0) {
            sinAng = (float)(bbox->p1.y - bbox->p0.y) / dwid;
            cosAng = (float)(bbox->p1.x - bbox->p0.x) / dwid;
        }
        else {
            sinAng = 0.0;
            cosAng = 1.0;
        }
            /* derotate pts 0,3 and compare to get xshear */
        tp0.x = tp0.y = 0; /* derotate of 0,0 is 0,0 */
        axtScaleInvR(bbox->p3.x, bbox->p3.y, &tp3.x, &tp3.y, sinAng, cosAng);
        if ((tp3.y - tp0.y) == 0) {
            dshear = 0.0;
        }
        else if ((tmp = tp3.x - tp0.x) != 0) {
            dshear = (float)tmp / (float)(tp3.y - tp0.y);
        }
        else {
            dshear = 0.0;
        }
        tp->xshear = (int) (dshear * 10000);
        tp->angle = axtGetAngle(sinAng, cosAng);
        *tflags |= TXxshear | TXangle;
        wid = (int)dwid;
        ht = tp3.y - tp0.y;
    }
    tp->pointsize = pointsize * 10000;
    if (refBox.width != wid || refBox.height != abs(ht)) {
#if 0
        neg = FALSE;
        if (wid < 0) {
            wid = (-wid);
            neg = TRUE;
        }
        dscale = (float)(refBox.height * wid) /
                (float)(refBox.width * abs(ht));
        tp->xscale = (int)(dscale * 10000);
        if (neg)
            tp->xscale = -tp->xscale;
        *tflags |= TXxscale;
        if (ht < 0) {
            ht = (-ht);
            tp->yscale = -10000;
            *tflags |= TXyscale;
        }
        tp->pointsize = (int)((float)(ht * pointsize) / refBox.height);
        tp->pointsize *= 10000;
#else
        neg = FALSE;
        if (wid < 0) {
            wid = (-wid);
            neg = TRUE;
        }
        dscale = (float) wid / (float) refBox.width;
        tp->xscale = (int)(dscale * 10000);
        if (neg)
            tp->xscale = -tp->xscale;
        *tflags |= TXxscale;

        neg = FALSE;
        if (ht < 0) {
            ht = (-ht);
            neg = TRUE;
        }
        dscale = (float) ht / (float) refBox.height;
        tp->yscale = (int)(dscale * 10000);
        if (neg)
            tp->yscale = -tp->yscale;
        *tflags |= TXyscale;
#endif
    }
        /* establish the xform -- note ax2.0 ge text is not polytext */
    if(!setTextAttrs(txt, tp, *tflags, 0, -1)) {
        return FALSE;
    }
        /* get baseline offset as the difference between the bbox
         * bounder origin (old-style text position is relative to this)
         * and the ascent-to-total-height-weighted average of bbox's
         * 'left edge' (p0 - p3);  new-style position is relative to
         * this 'left edge' baseline point */
    axtGetBoxBounder(bbox, &wid, &ht, &tp0); /* put bounder origin on tp0 */
        /* take weighted average of 'left' edge of bbox */
    tpt.x = (int)((bbox->p3.x - bbox->p0.x) * ratio) + bbox->p0.x;
    tpt.y = (int)((bbox->p3.y - bbox->p0.y) * ratio) + bbox->p0.y;
        /* apply this offset to old positions to get new */
    baseLineOffset->x = tpt.x - tp0.x;
    baseLineOffset->y = tpt.y - tp0.y;
    return TRUE;
}

/* : getAx2Bounder - compatibility measurement to be used
 * for the convert-read of version 2 document data, so that
 * the dimensions of text will match those used in version 2.
 * 
 * for backward-compatible measurements; rather than
 * an ink-bounder, return a bounder which matches a
 * (monotext) TmRInqTxExtent measurement; this means that the
 * definition of the bounder is different, as well as
 * the degree of error accumulation */
static char * /* NULL if ok */
getAx2Bounder(txt, pos, len, pointsize, bbox, ratio, ht36, ascent36)
axtext *txt;
int pos, len;
int pointsize; /* in points */
r32 *bbox; /* return this */
float *ratio; /* return the ratio of ascent to total height */
int ht36, ascent36; /* in; at 36pt, use fontmetrics table vals for ht */
{
    r32 cbox;
    perChar *pc;
    int wid = 0;
    int miny, maxy;
    int tmp, tlen, xadvance, lpos;
    fontMetrics *fm;
    txAttr *txAp;
    char *errString;

    if (!txt)
        return (getErrString(XLT("The text pointer was null.."),
                        (char *)NULL));
    if (!txt->whole.nChars) {
        bbox->origin.x = bbox->origin.y = bbox->width = bbox->height = 0;
        return ((char *)NULL);
    }
    if (!txt->measured) {
        if(errString = axtMeasuredText(txt)) {
            return errString;
        }
    }
        /* to best match the TmRInq numbers, go straight from
         * milliems to TargetRes pix, as is done when building
         * the fontmetrics data for ax2.0. */
    txAp = attrSegAtPos(txt, pos, &lpos);
    lpos = txAp->nChars - lpos; /* use lpos as #chars remaining in seg */
    fm = txAp->fconx.target;
    pc = &txt->items[pos];
    tlen = len;
    miny = -txAp->fconx.target->xfont->max_bounds.ascent;
    tmp = txAp->fconx.target->xfont->max_bounds.ascent +
        txAp->fconx.target->xfont->max_bounds.descent;
    maxy = miny + tmp - 1;
    maxy = mToPix(maxy,(float)pointsize, TargetRes);
    miny = mToPix(miny,(float)pointsize, TargetRes);
    while(--tlen > 0) {
        xadvance = WCharWid(pc->charcode, fm->xfont, &cbox); 
        wid += mToPix(xadvance, (float)pointsize, TargetRes);
        lpos--;
        pc++;
        if (lpos <= 0) {
            txAp++;
            lpos = txAp->nChars;
            fm = txAp->fconx.target;
            if (-txAp->fconx.target->xfont->max_bounds.ascent < miny)
                miny = -txAp->fconx.target->xfont->max_bounds.ascent;
            tmp = miny + txAp->fconx.target->xfont->max_bounds.ascent +
                txAp->fconx.target->xfont->max_bounds.descent;
            if (tmp > maxy)
                maxy = tmp;
        }
    }
    *ratio = (float)(-miny) / (float)(maxy - miny + 1);
        /* pc now looking at last char in range;
         * as in TmRInq, when adding in the last char's
         * contribution, use rbearing if it is larger
         * than the advance */
    xadvance = WCharWid(pc->charcode, fm->xfont, &cbox);

    if ((tmp = cbox.origin.x + cbox.width - xadvance) > 0) {
        wid += mToPix(xadvance + tmp, (float)pointsize, TargetRes);
    }
    else {
        wid += mToPix(xadvance, (float)pointsize, TargetRes);
    }
    if (0 && (pointsize == 36)) {
        /* 6,8,36pt screenfonts were untuned; the heights are
         * most affected (re: accented caps) -- the error is
         * significant at 36pt, so use the passed-in values
         * from the old fontmetrics table (which arrive in
         * 300dpi pix) */
        bbox->origin.x = 0; /* as in TmRInq */
        bbox->origin.y = axtPixToMils(-ascent36, 300);
        bbox->width = axtPixToMils(wid, TargetRes);
        bbox->height = axtPixToMils(ht36, 300);
        *ratio = (float)ascent36/(float)ht36;
    }
    else {
        bbox->origin.x = 0; /* as in TmRInq */
        bbox->origin.y = axtPixToMils(miny, TargetRes);
        bbox->width = axtPixToMils(wid, TargetRes);
        bbox->height = axtPixToMils(maxy - miny + 1, TargetRes);
    }
    bbox->width++;
    bbox->height++;
    return ((char *)NULL);
}

int /* return angle in 10000ths of a degree */
axtGetAngle(sinAng, cosAng)
float sinAng, cosAng;
{
    float arcsinAng;
    float angle;
    if (sinAng==0.0 && cosAng==1.0)
        return(0);

    arcsinAng = asin((double) -sinAng); /* sines negated thruout */
    if (arcsinAng > 0.0) {
        angle = cosAng > 0.0 ? arcsinAng : M_PI - arcsinAng;
    }
    else {
        angle = cosAng > 0.0 ? (2*M_PI) + arcsinAng : M_PI - arcsinAng;
    }
        /* convert from radians to degrees */
    angle = (angle * 180.0) / M_PI;
    return((int)(angle * 10000));
}

static int
nonrotated(b)
b32 *b;
{
    if (b->p0.y==b->p1.y && b->p0.y < b->p3.y &&
        b->p0.x==b->p3.x) {
        return TRUE;
    }
    else {
        return FALSE;
    }
}

#if AX_X11   /*  {  */

/* : getClosestRasterName - when stuck doing raster-scaling,
 * determine which size is best to work with.
 * Beware unprotected access of cache objects fconx.target,
 * fconx.screen, fconx.font -- but, called only from
 * axtDisplayableText where those objects are looked up first.
 */
static char *    /* do not free returned xlfd fontname */
getClosestRasterName(ftab, txAp, minError)
fontTabEnt *ftab;
txAttr *txAp;
int *minError;         /* return this */
{
    float yscale, dtmp;
    
        /* get absolute value of yscale */
    extractYscale(&txAp->tctm, &dtmp);
    if (dtmp < 0.0)
        yscale = (-dtmp);
    else
        yscale = dtmp;

    return( axtGetClosestRasterName(ftab, txAp->fontAttr.ptsize,
                                    yscale, minError) );
}

char *    /* do not free returned xlfd fontname */
axtGetClosestRasterName(ftab, pointsize, yscale, minError)
fontTabEnt *ftab;    /* in */
float pointsize;     /* in: requested font size in points */
float yscale;        /* in: vertical scale factor, 1.0 is unscaled */
int *minError;       /* out: return this pixel height error */
{
    int effectivePts, lowerIx, higherIx;
    int closestLower, closestHigher, ix;
    int screenPts, tmp, bestIndex, minPtError;
    
        /* get the 'effective' target pointsize in decipoints */
    effectivePts = (int) (pointsize * yscale * 10.0);
    if (effectivePts > 10000)
        effectivePts = 10000;
    
        /* find where this falls among the available screenfont
           nominal pointsizes */
    lowerIx = higherIx = -1;  /* indices */
    closestLower = closestHigher = MAXLRU; /* differences in ptsize */
    for (ix = 0; ix < ftab->xNameCount; ix++) {
        screenPts = axtDecipointSize(ftab->xNames[ix]);
        tmp = screenPts - effectivePts;
        if (yscale == 1.0 && tmp == 0) {
                /* an exact match at normal scaling; use this
                   size regardless of actual fit */
            *minError = 0; /* a lie, but this is the latch */
            return(ftab->xNames[ix]);
        }
        if (tmp >= 0) {
                /* screenfont is larger than requested size */
            if (tmp < closestHigher) {
                closestHigher = tmp;
                higherIx = ix;
            }
        }
        else {
            tmp = abs(tmp);
            if (tmp < closestLower) {
                closestLower = tmp;
                lowerIx = ix;
            }
        }
    }
#if 0 /* for using a milliem definition of tolerance */
        /* convert error numbers from decipoints to milliems */
    closestHigher = (closestHigher * 1000.0) / effectivePts;
    closestLower = (closestLower * 1000.0) / effectivePts;
#endif

    if (lowerIx != -1 && higherIx != -1) {
            /* in between two screenfont choices;  favor lower size,
               but take higher if it's close enough */
        if ((closestHigher < closestLower) &&
            (closestHigher <= ftab->rasterTolerance)) {
            bestIndex = higherIx;
            minPtError = closestHigher;
        }
        else {
            bestIndex = lowerIx;
            minPtError = closestLower;
        }
    }
    else if (lowerIx != -1) {
            /* requseted size is larger than all screenfonts */
        bestIndex = lowerIx;
        minPtError = closestLower;
    }
    else {
        bestIndex = higherIx;
        minPtError = closestHigher;
    }
        
    *minError = abs(minPtError);
    return(ftab->xNames[bestIndex]);
}

#endif   /*  }  AX_X11 */

/* Beware unprotected access of cache object fconx.screen,
 * but, called only from axtDisplayableText where that
 * object is looked up first.
 */
static void
updateScreenChars(txAp, items, firstix, lastix)
txAttr *txAp;
perChar *items;
int firstix, lastix; /* both inside txAp's character range */
{
    perChar *pc;
    int tcharix;
    
        /* collect screen metrics for each char of this attr seg
         */
    for(tcharix=firstix ;tcharix <= lastix; tcharix++) {
        pc = &items[tcharix];
        if (!pc->screenMetrics) {
            if (!(CHAR_IS_DIMENSIONLESS((int) pc->charcode))) {
                pc->sadvance = WCharWid(pc->charcode,
                                        txAp->fconx.screen->xfont,
                                        &pc->sbox);
            }
            pc->screenMetrics = TRUE;
        }
    }
}

/* Beware unprotected access of cache object fconx.font
 * but, called only from axtDisplayableText where those
 * objects are looked up first.
 */
/*
 * determine whether image-scaling
 * will be required for display.
 */
static int /* TRUE means use fast-path display */
setDrawMethod(tx, txAp, scalable, rasterError, yscale)
axtext *tx;             /* in */
txAttr *txAp;           /* in */
int scalable;           /* in: bool */
int rasterError;        /* in */
float yscale;           /* in */
{
    int maybeQDraw = FALSE;
    int qDraw = FALSE;
    fontTabEnt *ftab = txAp->fconx.font;

    if (!tx->whole.transform.angle && !txAp->transform.angle &&
        !tx->whole.transform.xshear && !txAp->transform.xshear &&
        (tx->whole.transform.xscale == tx->whole.transform.yscale) &&
        (tx->whole.transform.xscale > 0) && (txAp->transform.xscale > 0) &&
        (txAp->transform.xscale == txAp->transform.yscale) &&
        (tx->whole.xZoom == tx->whole.yZoom)) {
        
        maybeQDraw =  TRUE; /* no rotate or shear, square-scaled,
                               and solid-filled */
    }
    else {
        qDraw = FALSE; /* gotta do it the hard way */
    }
#if AX_X11  
    if (maybeQDraw) {
        if (scalable) {
            qDraw = TRUE;
        }
        else {
            qDraw = foundMatchingScreenfont(ftab,
                                            txAp->fontAttr.ptsize * yscale,
                                            rasterError);
        }
    }
#else
    qDraw = maybeQDraw;    /* always using scalable screenfont on win32 */
#endif
    return qDraw;
}

static int    /* return whether there's a screenfont close enough
                 to the desired pointsize */
foundMatchingScreenfont(ftab, pointsize, rasterError)
fontTabEnt *ftab;
float pointsize;
int rasterError;
{
    if (rasterError <= ftab->rasterTolerance &&
        ((int) pointsize >= ftab->toleranceLowCutoff) &&
        ((int) pointsize <= ftab->toleranceHighCutoff)) {
        
        return(TRUE); /* ignore the discrepancy */
    }
    return (FALSE);   /* none close enough for direct usage */
}
/*
 * Beware unprotected access of cache object fconx.target --
 * but, called only from axtMeasuredText where that object is looked
 * up first.
 */
/* get the target metrics for a character, along with
 * transformed metrics derived from the target numbers */
static void
updateTargetChar(tx, txAp, pc)
axtext *tx;
txAttr *txAp;
perChar *pc;
{
    r32 pbox;

    if (CHAR_IS_DIMENSIONLESS((int) pc->charcode)) {
        memset((char *) &pbox, 0, sizeof(r32));
        pc->advance.x = pc->advance.y = pc->padvance = 0;
        pc->width = pc->height = pc->orig.x = pc->orig.y = 0;
        return;
    }
    else {
        pc->padvance = WCharWid(pc->charcode, txAp->fconx.target->xfont,
                                &pbox);
    }

#if 0
    if (txAp->fontAttr.ptsize < 36.0 && ScreenRes <= 75) {
            /* for smaller sizes on coarse screens, report
               boxes which are whole-font-height; re: jitter */
        pbox.origin.y =
            -txAp->fconx.target->xfont->max_bounds.ascent;
        pbox.height =
            txAp->fconx.target->xfont->max_bounds.ascent +
            txAp->fconx.target->xfont->max_bounds.descent;
    }
#endif
        /* transform onto bbox */
    if (txAp->tctm.unity) {
        pc->bbox.p0 = pbox.origin;
        pc->bbox.p1.x = pbox.origin.x + pbox.width;
        pc->bbox.p1.y = pbox.origin.y;
        pc->bbox.p2.x = pc->bbox.p1.x;
        pc->bbox.p2.y = pbox.origin.y + pbox.height;
        pc->bbox.p3.x = pc->bbox.p0.x;
        pc->bbox.p3.y = pc->bbox.p2.y;
        pc->advance.x = pc->padvance;
        pc->advance.y = 0;
        pc->fadvance = pc->advance;
    }
    else {
        forwardSXRT(pbox.origin.x, pbox.origin.y,
                    &pc->bbox.p0.x, &pc->bbox.p0.y, &txAp->tctm);
        forwardSXRT(pbox.origin.x + pbox.width,
                    pbox.origin.y, &pc->bbox.p1.x,
                    &pc->bbox.p1.y, &txAp->tctm);
        forwardSXRT(pbox.origin.x + pbox.width,
                    pbox.origin.y + pbox.height,
                    &pc->bbox.p2.x, &pc->bbox.p2.y, &txAp->tctm);
        forwardSXRT(pbox.origin.x,
                    pbox.origin.y + pbox.height,
                    &pc->bbox.p3.x, &pc->bbox.p3.y, &txAp->tctm);
        forwardSXRT(pc->padvance, 0,
                    &pc->advance.x, &pc->advance.y, &txAp->tctm);
        forwardSXRT(pc->padvance, 0,
                    &pc->fadvance.x, &pc->fadvance.y, &txAp->fctm);
    }
        /* go from milliems to mills */
    pc->padvance = emToMils(pc->padvance, txAp->fontAttr.ptsizeInt);
    pc->bbox.p0.x = emToMils(pc->bbox.p0.x, txAp->fontAttr.ptsizeInt);
    pc->bbox.p0.y = emToMils(pc->bbox.p0.y, txAp->fontAttr.ptsizeInt);
    pc->bbox.p1.x = emToMils(pc->bbox.p1.x, txAp->fontAttr.ptsizeInt);
    pc->bbox.p1.y = emToMils(pc->bbox.p1.y, txAp->fontAttr.ptsizeInt);
    pc->bbox.p2.x = emToMils(pc->bbox.p2.x, txAp->fontAttr.ptsizeInt);
    pc->bbox.p2.y = emToMils(pc->bbox.p2.y, txAp->fontAttr.ptsizeInt);
    pc->bbox.p3.x = emToMils(pc->bbox.p3.x, txAp->fontAttr.ptsizeInt);
    pc->bbox.p3.y = emToMils(pc->bbox.p3.y, txAp->fontAttr.ptsizeInt);
    pc->advance.x = emToMils(pc->advance.x, txAp->fontAttr.ptsizeInt);
    pc->advance.y = emToMils(pc->advance.y, txAp->fontAttr.ptsizeInt);
    pc->fadvance.x = emToMils(pc->fadvance.x, txAp->fontAttr.ptsizeInt);
    pc->fadvance.y = emToMils(pc->fadvance.y, txAp->fontAttr.ptsizeInt);
    axtGetBoxBounder(&pc->bbox, &pc->width, &pc->height, &pc->orig);
        /* apply any kern and sup */
    if (txAp->tctm.unity) {
        pc->sup.x = txAp->supx;
        pc->sup.y = txAp->supy;
        pc->kern.x = txAp->kernx;
        pc->kern.y = txAp->kerny;
        pc->fkern = pc->kern;
    }
    else {
        forwardSXRT(txAp->supx, txAp->supy, &pc->sup.x, &pc->sup.y,
                    &txAp->tctm);
        forwardSXRT(txAp->kernx, txAp->kerny, &pc->kern.x, &pc->kern.y,
                    &txAp->tctm);
        forwardSXRT(txAp->kernx, txAp->kerny, &pc->fkern.x, &pc->fkern.y,
                    &txAp->fctm);
    }
}

int WCharWid(ch, xf, bbox)
tx_char ch;
XFontStruct *xf;
r32 *bbox;
{
    int tmp;
    XCharStruct *xc;

    xc = getCharStruct((int) ch, xf);

    if ( bbox != (r32 *)NULL ) {
        bbox->origin.x = xc->lbearing;
            /* write the bbox in y-increasing-downwards coords
             */
        bbox->origin.y =  -xc->ascent;
        if ((tmp = (int)(xc->rbearing - xc->lbearing)) > 0) {
                /* use the raster width if described
                 */
            bbox->width = tmp;
        }
        else {
                /* otherwise substitute the advance width as bbox width
                 */
            bbox->width = xc->width > 0 ? xc->width : 0;
        }
        
        if ((tmp = (int)(xc->ascent + xc->descent)) > 0) {
            bbox->height = tmp;
        }
        else {
            bbox->height = 0;
        }

            /* impose a minimum on bbox size..
             */
        if (!bbox->width) {
            bbox->width = xc->width ? xc->width : 5;
        }
        if (!bbox->height) {
            bbox->height = 5;
            bbox->origin.y = -bbox->height;
        }
    }
    return( xc->width );
}

/* : inheritedAttr - return a pointer to the txAttr struct
 * expressing the attribute that a new character should inherit.
 */
static txAttr *
inheritedAttr(txt, pos)
axtext *txt;
int pos;
{
    int tmp, ipos;

    if (!txt)
        return ((txAttr *) NULL);

    if (pos > 0) {
#if 0       
        if (WHITECHAR((int) txt->textstr[pos - 1])) {
            ipos = pos;       /* inherit from right */
        }
        else {
            ipos = pos - 1;   /* inherit from left */
        }
        return (attrSegAtPos(txt, ipos, &tmp));
#else
        return (attrSegAtPos(txt, pos - 1, &tmp));
#endif
    }
    else {
        return (attrSegAtPos(txt, pos, &tmp));
    }
}

int
emToMils(milliems, pointsize)
int milliems;
int pointsize;      /* 10000 times point size; e.g. 144000 for 14.4 pts */
{
    float f1, f2, f3;
    int     *mtable;

    /* Use precalculated conversion table for this point size if available */
    mtable = MTABLEPTR(pointsize);
    if (mtable) {
        int tmils;
        if (milliems < 0) {
            int abm;
            abm = -milliems;
            if (abm < MTABLESIZE) {
                tmils = mtable[abm];
                return (-tmils);
            }
        }
        else {
            if (milliems < MTABLESIZE) {
                tmils = mtable[milliems];
                return (tmils);
            }
        }
    }

    f1 = (float)(milliems);
    f2 = pointsize;
    f3 = f1 * f2;   /* now in units of 10 points */
    f3 /= 720000; /* now in 1000ths of an inch */
    return(qRound(f3));
}


/* Given a sequence of adjacent calls to mToMils (with a common pointsize),
    you can determine which conversion table to use (via MTABLEPTR) and
    substitute a sequence of the following function. This will eliminate
    the unnecessary integer comparisons as we look for the same conversion
    table for each mToMils call. */
int
mToMilsTable(milliems, pointsize, mtable)
int milliems;
int pointsize;      /* 10000 times point size; e.g. 144000 for 14.4 pts */
int *mtable;        /* Pointer to pre-calculated conversion table to use */
{
    float f1, f2, f3;
    int tmils;

    /* Use precalculated conversion table for this point size if available */
    if (mtable) {
        int tmils;
        if (milliems < 0) {
            int abm;
            abm = -milliems;
            if (abm < MTABLESIZE) {
                tmils = mtable[abm];
                return (-tmils);
            }
        }
        else {
            if (milliems < MTABLESIZE) {
                tmils = mtable[milliems];
                return (tmils);
            }
        }
    }

    f1 = (float)(milliems);
    f2 = pointsize;
    f3 = f1 * f2;   /* now in units of 10 points */
    f3 /= 720000; /* now in 1000ths of an inch */
    return(qRound(f3));
}

int
axtInitMilTables()
{
    int ix;
    for (ix = 0; ix < MTABLESIZE; ix++) {
        axtMilTable10_0[ix] = mToMils(ix, 10.0);
        axtMilTable11_0[ix] = mToMils(ix, 11.0);
        axtMilTable12_0[ix] = mToMils(ix, 12.0);
        axtMilTable13_2[ix] = mToMils(ix, 13.2);
        axtMilTable14_0[ix] = mToMils(ix, 14.0);
        axtMilTable14_4[ix] = mToMils(ix, 14.4);
    }
}

int
milsToPix(mills, devRes)
int mills;
int devRes;
{
    return(qRound((((float) mills * devRes) / 1000)));
}

int
axtPixToMils(pix, devRes)
int pix;
int devRes;
{
    int round_up = devRes - 1;
    if (pix < 0)
        round_up = -round_up;
    return((int)(((long) (pix * 1000) + round_up) / devRes));
}

static void
tossWchar(ch)
char *ch;
{
    /* here's one the compiler tosses for us */
}

/* tossTxAttr: free all dynamic memory from the
 * passed txAttr instance;  note that the txAttr
 * instance itself is not freed, only its appendages
 */
static void
tossTxAttr(tx)
char *tx;         /* in: treat as (txAttr *)  */
{
    txAttr *ttx = (txAttr *)tx;
    
    if (!ttx)
        return;
    
    if (ttx->strokepattern) {
        axtFreeFillPattern(ttx->strokepattern);
        ttx->strokepattern = (attr_type *)NULL;
    }
    if (ttx->fillpattern) {
        axtFreeFillPattern(ttx->fillpattern);
        ttx->fillpattern = (attr_type *)NULL;
    }
    if (ttx->screenFontName) {
        txfree((char *)ttx->screenFontName);
        ttx->screenFontName = (char *)NULL;
    }
    if (ttx->fontAttr.familyName) {
        txfree((char *)ttx->fontAttr.familyName);
        ttx->fontAttr.familyName = (char *)NULL;
    }
    if (ttx->fontAttr.fontName) {
        txfree((char *)ttx->fontAttr.fontName);
        ttx->fontAttr.fontName = (char *)NULL;
    }
    if (ttx->indent) {
        axtFreeIndent(ttx->indent);
        ttx->indent = (axtIndent *) NULL;
    }
}

/*
 * pc is looking at a good, dynamically allocated perChar instance.
 */
static void
tossPerChar(pc)
char *pc;
{
    perChar *tpc = (perChar *)pc;
    if (tpc->ctm) {
        txfree((char *)tpc->ctm);
        tpc->ctm = (axform *) NULL;
    }
    if (tpc->letterName) {
        txfree((char *)tpc->letterName);
        tpc->letterName = (char *) NULL;
    }
    /* don't toss *tpc itself; it's an array element */
}

/* insert just before pos, so "insert @ 0" means
 * put new item at front. */
static int
insertItems(itemType, pos, inData, inLen, curData, curLen, curBufLen, pid)
int itemType;
int pos;
char *inData;
int inLen;
char **curData; /* out */
int curLen;
int *curBufLen; /* out */
int pid;
{
    int i, j, dpos, itemsiz;
    char *to, *from;
    int newsize = curLen + inLen;
    if (pos < 0 || pos > curLen)
        return FALSE;
    itemsiz = itemSize(itemType);
    if (newsize >= *curBufLen) { /* note 1 itemsiz extra.. */
        j = newsize + DELTA;
        if (*curData) {
            *curData = (char *)retxloc(pid, (char *)*curData, j * itemsiz);
        }
        else {
            *curData = (char *)txloc(pid, j * itemsiz);
        }
        *curBufLen = j;
    }
        /* open up a space for the new items */
    dpos = newsize - 1;
    for(i=curLen-1; i>=pos; i--, dpos--) {
        to = (char *)(*curData) + (dpos * itemsiz);
        from = (char *)(*curData) + (i * itemsiz);
        memcpy(to, from, itemsiz);
    }

        /* copy in the new items */
    for(i=pos, j=0; j<inLen; i++, j++) {
        to = (char *)(*curData) + (i * itemsiz);
        from = (char *)(inData) + (j * itemsiz);
        memcpy(to, from, itemsiz);
    }
    return TRUE;
}

static int
deleteItems(itemType, pos, inLen, curData, curLen, curBufLen, pid)
int itemType;
int pos;
int inLen;
char **curData;
int curLen;
int *curBufLen;
int pid;
{
    int i, j, copylen, itemsiz, dpos;
    char *to, *from;
    int newsize = curLen - inLen;

    if (!(*curData))
        return TRUE;
    if (pos < 0 || (pos+inLen) > curLen)
        return FALSE;
    itemsiz = itemSize(itemType);
    copylen = curLen - inLen - pos;
    if (copylen == 0) {
            /* we're deleting from the end */
        dpos = curLen - 1;
        while (inLen--) {
            to = (char *)(*curData) + (dpos * itemsiz);
            (*tossItem[itemType])(to);
            dpos--;
        }
    }
    else {
        dpos = pos;
        for(i=pos+inLen, j=0; j<copylen; i++, j++, dpos++) {
            to = (char *)(*curData) + (dpos * itemsiz);
            from = (char *)(*curData) + (i * itemsiz);
            if (inLen-- > 0)
                (*tossItem[itemType])(to); /* as many as are deleted */
            memcpy(to, from, itemsiz);
        }
    }
    if (newsize > 0 && (curLen - newsize) > DELTA) {
        if ( !((char *)*curData) )
        *curData = (char *)txloc(pid,  newsize * itemsiz);
      else
        *curData = (char *)retxloc(pid,(char *)*curData,
                       newsize * itemsiz);
    }
    else if (newsize == 0) {
        if (*curData) {
            txfree(*curData);
            *curData = NULL;
        }
    }
    *curBufLen = newsize;
    return TRUE;
}

static int
itemSize(itemType)
int itemType;
{
    int itemsiz;
    switch(itemType) {
        case TYPE_W_CHAR:
            itemsiz = sizeof(tx_char);
            break;
        case TYPE_TXATTR:
            itemsiz = sizeof(txAttr);
            break;
        case TYPE_PERCHAR:
            itemsiz = sizeof(perChar);
            break;
      default:
            itemsiz = 0xdeadd0d0;
    }
    return itemsiz;
}

/* : foldDefaults - support for sparse attribute description.
 * combine the explicitly-specified (sparse) attr description
 * with underlying (thorough) default attr info to produce
 * a thorough description that expresses the overriding
 * explicit info */
static void
foldDefaults(mask, exp, dp, outp, pid)
unsigned int mask; /* which exp attrs to look at */
txAttr *exp; /* in: explicitly specified attrs */
txAttr *dp; /* in: defaults */
txAttr *outp; /* out: if explicit use exp, else use dp */
int pid;
{
    int tlen=0;
    axtIndent *indp = (axtIndent *) NULL;
    axtBullet *bulp = (axtBullet *) NULL;
    axtDropShadow *drop = (axtDropShadow *) NULL;
    char *cp=(char *)NULL;
    attr_type *tp = NULL;

    /* get the string propvals out of the way first */
    if (mask & TXfontFamily) {
        cp = exp->fontAttr.familyName;
    }
    else {
        cp = dp->fontAttr.familyName;
    }
    if (cp) {
        tlen = strlen(cp);
        if (outp->fontAttr.familyName) {
            txfree(outp->fontAttr.familyName);
        }
        outp->fontAttr.familyName = (char *)txloc(pid, tlen + 1);
        strcpy(outp->fontAttr.familyName, cp);
    }
    else
        outp->fontAttr.familyName = (char *)NULL;

    if (mask & TXstrokepattern) {
        tp = exp->strokepattern;
    }
    else {
        tp = dp->strokepattern;
    }
    if (tp) {
        if (outp->strokepattern) {
            axtFreeFillPattern(outp->strokepattern);
        }
        outp->strokepattern = axtNewFillPattern(pid);
        axtCopyFillPattern(tp, &outp->strokepattern, pid);
    }
    else
        outp->strokepattern = (attr_type *)NULL;

    if (mask & TXfillpattern) {
        tp = exp->fillpattern;
    }
    else {
        tp = dp->fillpattern;
    }
    if (tp) {
        if (outp->fillpattern) {
            axtFreeFillPattern(outp->fillpattern);
        }
        outp->fillpattern = axtNewFillPattern(pid);
        axtCopyFillPattern(tp, &outp->fillpattern, pid);
    }
    else
        outp->fillpattern = (attr_type *)NULL;

    /* do the rest of the txAttr struct */
    outp->fontAttr.weight = (mask & TXweight) ? 
            exp->fontAttr.weight : dp->fontAttr.weight;
    outp->fontAttr.slant = (mask & TXslant) ? 
            exp->fontAttr.slant : dp->fontAttr.slant;
    outp->fontAttr.ptsize = (mask & TXpointsize) ? 
            exp->fontAttr.ptsize : dp->fontAttr.ptsize;
    outp->fontAttr.ptsizeInt = (mask & TXpointsize) ? 
            exp->fontAttr.ptsizeInt : dp->fontAttr.ptsizeInt;
    outp->ruling = (mask & TXruling) ? exp->ruling : dp->ruling;
    outp->strike_thru = (mask & TXstrike) ? exp->strike_thru : dp->strike_thru;
    outp->strokecolor = (mask & TXstrokecolor) ? exp->strokecolor :
        dp->strokecolor;
    outp->strokebcolor = (mask & TXstrokebcolor) ? exp->strokebcolor :
        dp->strokebcolor;
    outp->fillcolor = (mask & TXfillcolor) ? exp->fillcolor :
        dp->fillcolor;
    outp->fillbcolor = (mask & TXfillbcolor) ? exp->fillbcolor :
        dp->fillbcolor;
    if (mask & TXfgPixel) {
        outp->fgPixel = exp->fgPixel;
        outp->useFgPixel = TRUE;
    }
    else {
        outp->fgPixel = (unsigned long) 0;
        outp->useFgPixel = FALSE;
    }
    outp->noRuleChars[0] = (mask & TXnoRuleChar) ? exp->noRuleChars[0] :
        dp->noRuleChars[0];
    outp->noRuleChars[1] = (mask & TXnoRuleChar) ? exp->noRuleChars[1] :
        dp->noRuleChars[1];
    outp->tctm = dp->tctm;
    if (mask & TXxscale) {
        outp->transform.xscale = exp->transform.xscale;
        outp->tctm.known = FALSE;
    }
    else {
        outp->transform.xscale = dp->transform.xscale;
    }
    if (mask & TXyscale) {
        outp->transform.yscale = exp->transform.yscale;
        outp->tctm.known = FALSE;
    }
    else {
        outp->transform.yscale = dp->transform.yscale;
    }
    if (mask & TXangle) {
        outp->transform.angle = exp->transform.angle;
        outp->tctm.known = FALSE;
    }
    else {
        outp->transform.angle = dp->transform.angle;
    }
    if (mask & TXxshear) {
        outp->transform.xshear = exp->transform.xshear;
        outp->tctm.known = FALSE;
    }
    else {
        outp->transform.xshear = dp->transform.xshear;
    }

    outp->supy = (mask & TXsupy) ? exp->supy : dp->supy;
    outp->supx = (mask & TXsupx) ? exp->supx : dp->supx;
    outp->kerny = (mask & TXkerny) ? exp->kerny : dp->kerny;
    outp->kernx = (mask & TXkernx) ? exp->kernx : dp->kernx;
    
    if (mask & TXindent) {
        indp = exp->indent;
    }
    else {
        indp = dp->indent;
    }
    if (indp) {
        if (outp->indent) {
            axtFreeIndent(outp->indent);
        }
        outp->indent = axtNewDefaultIndent(pid);
        axtCopyIndent(indp, outp->indent, pid);
    }
    else {
        outp->indent = (axtIndent *) NULL;
    }

    if (mask & TXbullet) {
        bulp = exp->indent ? exp->indent->bullet : (axtBullet *) NULL;
    }
    else {
        bulp = dp->indent ? dp->indent->bullet : (axtBullet *) NULL;
    }
    if (bulp) {
        if (outp->indent && outp->indent->bullet) {
            axtFreeBullet(outp->indent->bullet);
        }
        if (!outp->indent) {
            outp->indent = axtNewDefaultIndent(pid);
        }
        outp->indent->bullet = axtNewDefaultBullet(pid);
        axtCopyBullet(bulp, outp->indent->bullet, pid);
    }
    else if (outp->indent) {
        outp->indent->bullet = (axtBullet *)NULL;
    }

    if (mask & TXdropShadow) {
        drop = exp->dropShadow;
    }
    else {
        drop = dp->dropShadow;
    }
    if (drop) {
        if (outp->dropShadow) {
            axtFreeDropShadow(outp->dropShadow);
        }
        outp->dropShadow = (axtDropShadow *) txloc(pid, sizeof(axtDropShadow));
        axtCopyDropShadow(drop, outp->dropShadow, pid);
    }
    else
        outp->dropShadow = (axtDropShadow *) NULL;

}

/*
 * these are the hardcoded text attr defaults!
 */
static void
txAttDefaults(ap, flags, pid)
txAttr *ap;
unsigned int *flags; /* report which are actually set in here */
int pid;
{
    int len;
    if (!ap)
        return;

/* Ashisuto add for wp2rtf not crash.  1995.09.14 */
#if FOR_FILTER
    if (!axDefaultSerifFont)
#if AXMS
        axDefaultSerifFont = NXLT("Times New Roman");
#else
        axDefaultSerifFont = NXLT("Times");
#endif
#endif

    memset((char *)ap, 0, sizeof(txAttr));
    *flags = 0;
    len = strlen(NXLT(axDefaultSerifFont)) + 1;
    ap->fontAttr.familyName = (char *)txloc(pid, len);
    *flags |= TXfontFamily;
    strcpy(ap->fontAttr.familyName, NXLT(axDefaultSerifFont));
    ap->fontAttr.ptsize = (float) defaultPointsize();
    ap->fontAttr.ptsizeInt = defaultPointsize() * 10000;
    *flags |= TXpointsize;
    ap->strokecolor = 255;
    *flags |= TXstrokecolor;
    ap->strokebcolor = 0;
    *flags |= TXstrokebcolor;
    ap->fillcolor = 255;
    *flags |= TXfillcolor;
    ap->fillbcolor = 0;
    *flags |= TXfillbcolor;
    ap->noRuleChars[0] = ap->noRuleChars[1] = NO_CHAR;
    *flags |= TXnoRuleChar;
    ap->transform.unity = TRUE;
    ap->transform.angle = 0;
    ap->transform.xscale = 10000; /* i.e., 1.0 */
    ap->transform.yscale = 10000; /* i.e., 1.0 */
    ap->transform.xshear = 0;
    ap->tctm.known = FALSE;
    defaultTransform(&ap->tctm);
    *flags |= (TXangle & TXxscale & TXyscale & TXxshear);
}

static int
defaultPointsize()
{
    return 12; /* our hardcoded default! */
}

static void
defaultTransform(xt)
axform *xt;
{
    xt->unity = TRUE;
    xt->rotated = FALSE;
    xt->forward.ca = xt->forward.cd = 1.0;
    xt->forward.cb = xt->forward.cc = 0.0;
    xt->forward.ctx = xt->forward.cty = 0;
    xt->reverse = xt->forward;
}

static int
getDefaultLineHeight(tx)
axtext *tx;
{
    int pointsize;
    if (tx->userDefaults && (tx->userDefaultFlags && TXpointsize)) {
        pointsize = tx->userDefaults->fontAttr.ptsizeInt;
    }
    else {
        pointsize = 10000 * defaultPointsize();
    }
    return (emToMils(1200, pointsize));
}

void
axtInitGCs(drw)
Drawable drw;
{
    XGCValues gcv;
    unsigned long blackPix;


    axTextGC = getGC(drw, GXcopy, AlisPixels[BLACK],
                     AlisPixels[WHITE],
                     AllPlanes);
    axTextStencilGC = getGC(drw, GXand, ~AllPlanes, AllPlanes, AllPlanes);
    axLiftGC = getGC(drw, GXcopy, AlisPixels[WHITE],
                     AlisPixels[WHITE], AllPlanes);
    PixClearGC = getPixmapGC(drw, GXcopy, 0, 1, AllPlanes);
    PixGC = getPixmapGC(drw, GXcopy, 1, 0, AllPlanes);
    
    AxGetRealPixel(2, 0, 0, 0, 255, 255, &blackPix);
    axTextCursorGC = getGC(drw, GXxor, (AlisPixels[WHITE] ^ blackPix),
                           ~AllPlanes, AllPlanes);
    
    axInactiveTextCursorGC = getGC(drw, GXxor,
                                   (AlisPixels[WHITE] ^ AlisPixels[BLACK]),
                                   ~AllPlanes, AllPlanes);
}

/* SXRT (scale, xshear, rotate, translate):
 *
 * SxCos                   SxSin                  0
 *
 * Sy(HxCos - Sin)         Sy(Cos + HxSin)        0
 *
 * Tx                      Ty                     1
 *
 *
 * InvSXRT:
 *
 * (Cos + HxSin)/Sx        -Sin/Sy                 0
 *
 * (Sin - HxCos)/Sx        Cos/Sy                  0
 *
 * (a)                     (TxSin - TyCos)/Sy      1
 *
 * where:
 *
 * (a) = Hx(TyCos - TxSin)/Sx - (TxCos + TySin)/Sx
 */
/* (sin terms negated) */

void
forwardSXRT(srcx,srcy,dstxp,dstyp,xf)
int srcx,srcy,*dstxp,*dstyp;
axform *xf;
{
	float val;
    if (xf->unity) {
        *dstxp = srcx;
        *dstyp = srcy;
        return;
    }

	val = (srcx * xf->forward.ca) + (srcy * xf->forward.cc) + xf->forward.ctx;
    *dstxp = qRound(val);
                    
	val = (srcx * xf->forward.cb) + (srcy * xf->forward.cd) + xf->forward.cty;
    *dstyp = qRound(val);
}

void
reverseSXRT(srcx,srcy,dstxp,dstyp,xf)
int srcx,srcy,*dstxp,*dstyp;
axform *xf;
{
	float val;
    if (xf->unity) {
        *dstxp = srcx;
        *dstyp = srcy;
        return;
    }

	val = (srcx * xf->reverse.ca) + (srcy * xf->reverse.cc) + xf->reverse.ctx;
    *dstxp = qRound(val);
                    
	val = (srcx * xf->reverse.cb) + (srcy * xf->reverse.cd) + xf->reverse.cty;
    *dstyp = qRound(val);
}

    /* for angle (sinAng, cosAng), do reverse-rotate */
void
axtScaleInvR(srcx, srcy, dstxp, dstyp, sinAng, cosAng)
int srcx, srcy, *dstxp, *dstyp;
float sinAng, cosAng;
{
    *dstxp = (int)(srcx * cosAng + srcy * sinAng);
    *dstyp = (int)(srcy * cosAng - srcx * sinAng);
}

    /* convert transform terms (xscale, angle, etc..) into ctm elements
     * for the scale-shear-rotate-translate transform */
static void
writeForwardCtm(terms, tx, ty, ctm)
axformTerms *terms; /* in */
int tx, ty; /* in; additional translation */
axctm *ctm; /* out */
{
    float sx, sy, hx, cosAng, sinAng, dtmp;
        /* write terms as real numbers */
            /* angle */
    dtmp = (float) terms->angle;  /* in 10000ths of a degree */
    dtmp = (dtmp * M_PI) / 180.0;
    dtmp /= 10000; /* now in radians */
    
    sinAng = (float) -sin((double) dtmp); /* note sine terms negated */
    cosAng = (float) cos((double) dtmp);
            /* xscale, yscale */
    dtmp = (float) terms->xscale;
    sx = dtmp / 10000;
    dtmp = (float) terms->yscale;
    sy = dtmp / 10000;
            /* xshear */
    dtmp = (float) terms->xshear;
    hx = dtmp / 10000;
            /* write elements of ctm */
    ctm->ca = sx * cosAng;
    ctm->cb = sx * sinAng;
    ctm->cc = sy * ((hx * cosAng) - sinAng);
    ctm->cd = sy * (cosAng + (hx * sinAng));
    ctm->ctx = tx;
    ctm->cty = ty;
}

    /* convert transform terms (xscale, angle, etc..) into ctm elements
     * for the reverse of the scale-shear-rotate-translate transform */
static void
writeReverseCtm(terms, tx, ty, ctm)
axformTerms *terms; /* in */
int tx, ty; /* in; additional translation */
axctm *ctm; /* out */
{
    float sx, sy, hx, cosAng, sinAng, dtmp;
        /* write terms as real numbers */
            /* angle */
    dtmp = (float) terms->angle; /* in 10000ths of a degree */
    dtmp = (dtmp * M_PI) / 180.0;
    dtmp /= 10000; /* now in radians */
    sinAng = (float) -sin((double) dtmp); /* note sine terms negated */
    cosAng = (float) cos((double) dtmp);
            /* xscale, yscale */
    dtmp = (float) terms->xscale;
    sx = dtmp / 10000;
    dtmp = (float) terms->yscale;
    sy = dtmp / 10000;
            /* xshear */
    dtmp = (float) terms->xshear;
    hx = dtmp / 10000;
            /* write elements of ctm */
    if (terms->xscale != 0) {
        ctm->ca = (cosAng + (hx * sinAng)) / sx;
        ctm->cc = (sinAng - (hx * cosAng)) / sx;
        ctm->ctx = (hx * ((ty * cosAng) - (tx * sinAng))) / sx -
                ((tx * cosAng) + (ty * sinAng)) / sx;
    }
    else {
        ctm->ca = ctm->cc = 0.0;
        ctm->ctx = 0;
    }
    if (terms->yscale != 0) {
        ctm->cb = -(sinAng) / sy;
        ctm->cd = cosAng / sy;
        ctm->cty = ((tx * sinAng) - (ty * cosAng)) / sy;
    }
    else {
        ctm->cb = ctm->cd = 0.0;
        ctm->cty = 0;
    }
}

    /* the target ctm is the combination of the local, or
     * attr-seg-level ctm with the overall, or field-ctm;
     * data is transformed thru the local xform, and the
     * result is again transformed thru the field xform */
void
axtWriteTargetCtm(xZoom, yZoom, local, field, result)
int xZoom, yZoom;            /* in: zoom factor, 1000 is unzoomed */
axformTerms *local, *field;  /* in: local and overall transforms */
axform *result;              /* out: init combined xform */
{
    axctm lctm, fctm, tmp, zctm;

    if (local->unity && field->unity &&
        xZoom == 1000 && yZoom == 1000) {
        defaultTransform(result);
        return;
    }

    result->unity = FALSE;

    if (local->angle || field->angle) {
        result->rotated = TRUE;
    }
    else {
        result->rotated = FALSE;
    }

        /* write zoom ctm directly */
    zctm.ctx = zctm.cty = 0;
    zctm.cb = zctm.cc = 0.0;
    zctm.ca = (float) xZoom / 1000.0;
    zctm.cd = (float) yZoom / 1000.0;
    
    writeForwardCtm(local, 0, 0, &lctm);
    writeForwardCtm(field, 0, 0, &fctm);

        /* combine forward xform: local * field */
    tmp.ca = (lctm.ca * fctm.ca) + (lctm.cb * fctm.cc);
    tmp.cb = (lctm.ca * fctm.cb) + (lctm.cb * fctm.cd);
    tmp.cc = (lctm.cc * fctm.ca) + (lctm.cd * fctm.cc);
    tmp.cd = (lctm.cc * fctm.cb) + (lctm.cd * fctm.cd);
    tmp.ctx = (lctm.ctx * fctm.ca) +
        (lctm.cty * fctm.cc) + fctm.ctx;
    tmp.cty = (lctm.ctx * fctm.cb) +
        (lctm.cty * fctm.cd) + fctm.cty;

        /* combine this with zoom for resultant forward xform */
    result->forward.ca = (tmp.ca * zctm.ca) + (tmp.cb * zctm.cc);
    result->forward.cb = (tmp.ca * zctm.cb) + (tmp.cb * zctm.cd);
    result->forward.cc = (tmp.cc * zctm.ca) + (tmp.cd * zctm.cc);
    result->forward.cd = (tmp.cc * zctm.cb) + (tmp.cd * zctm.cd);
    result->forward.ctx = (tmp.ctx * zctm.ca) +
        (tmp.cty * zctm.cc) + zctm.ctx;
    result->forward.cty = (tmp.ctx * zctm.cb) +
        (tmp.cty * zctm.cd) + zctm.cty;


        /* do the reverse for the reverse xform:
           inverse-zoom * inverse-field * inverse-local */
    zctm.ca = 1.0 / zctm.ca;
    zctm.cd = 1.0 / zctm.cd;
    writeReverseCtm(local, 0, 0, &lctm);
    writeReverseCtm(field, 0, 0, &fctm);

        /* combine reverse xform: zoom * field */
    tmp.ca = (zctm.ca * fctm.ca) + (zctm.cb * fctm.cc);
    tmp.cb = (zctm.ca * fctm.cb) + (zctm.cb * fctm.cd);
    tmp.cc = (zctm.cc * fctm.ca) + (zctm.cd * fctm.cc);
    tmp.cd = (zctm.cc * fctm.cb) + (zctm.cd * fctm.cd);
    tmp.ctx = (zctm.ctx * fctm.ca) +
        (zctm.cty * fctm.cc) + fctm.ctx;
    tmp.cty = (zctm.ctx * fctm.cb) +
        (zctm.cty * fctm.cd) + fctm.cty;

        /* combine this with local for resultant reverse xform */
    result->reverse.ca = (tmp.ca * lctm.ca) + (tmp.cb * lctm.cc);
    result->reverse.cb = (tmp.ca * lctm.cb) + (tmp.cb * lctm.cd);
    result->reverse.cc = (tmp.cc * lctm.ca) + (tmp.cd * lctm.cc);
    result->reverse.cd = (tmp.cc * lctm.cb) + (tmp.cd * lctm.cd);
    result->reverse.ctx = (tmp.ctx * lctm.ca) +
        (tmp.cty * lctm.cc) + lctm.ctx;
    result->reverse.cty = (tmp.ctx * lctm.cb) +
        (tmp.cty * lctm.cd) + lctm.cty;
}

/* bbox is a set of 4 points representing a transformed
 * rectangular shape.  Fill in ht, wid, and pos as the
 * 'HV' box which bounds bbox */
void
axtGetBoxBounder(bbox, wid, ht, pos)
b32 *bbox; /* in */
int *wid, *ht; /* out */
p32 *pos; /* out */
{
    int minx, miny, maxx, maxy;
    
    minx = maxx = bbox->p0.x;
    miny = maxy = bbox->p0.y;
    if (bbox->p1.x > maxx)
        maxx = bbox->p1.x;
    if (bbox->p1.x < minx)
        minx = bbox->p1.x;
    if (bbox->p2.x > maxx)
        maxx = bbox->p2.x;
    if (bbox->p2.x < minx)
        minx = bbox->p2.x;
    if (bbox->p3.x > maxx)
        maxx = bbox->p3.x;
    if (bbox->p3.x < minx)
        minx = bbox->p3.x;

    if (bbox->p1.y > maxy)
        maxy = bbox->p1.y;
    if (bbox->p1.y < miny)
        miny = bbox->p1.y;
    if (bbox->p2.y > maxy)
        maxy = bbox->p2.y;
    if (bbox->p2.y < miny)
        miny = bbox->p2.y;
    if (bbox->p3.y > maxy)
        maxy = bbox->p3.y;
    if (bbox->p3.y < miny)
        miny = bbox->p3.y;
    pos->x = minx;
    pos->y = miny;
    *ht = maxy - miny + 1;
    *wid = maxx - minx + 1;
}

static void
copyTxAttr(to, from, pid)
txAttr *to, *from;
int pid;
{
    int len;
    
        /* (noncached) free-store appendages are:
         * fontAttr.familyName, fontAttr.fontName,
         * strokepattern, fillpattern, screenFontName.
         */
    if (to->fontAttr.familyName) {
        txfree((char *)to->fontAttr.familyName);
        to->fontAttr.familyName = (char *)NULL;
    }
    if (to->fontAttr.fontName) {
        txfree((char *)to->fontAttr.fontName);
        to->fontAttr.fontName = (char *)NULL;
    }
    if (to->strokepattern) {
        axtFreeFillPattern(to->strokepattern);
        to->strokepattern = (attr_type *)NULL;
    }
    if (to->fillpattern) {
        axtFreeFillPattern(to->fillpattern);
        to->fillpattern = (attr_type *)NULL;
    }
    if (to->screenFontName) {
        txfree((char *)to->screenFontName);
        to->screenFontName = (char *)NULL;
    }
    if (to->indent) {
        axtFreeIndent(to->indent);
        to->indent = (axtIndent *) NULL;
    }
    if (to->dropShadow) {
        axtFreeDropShadow(to->dropShadow);
        to->dropShadow = (axtDropShadow *) NULL;
    }

    *to = *from;

    if (from->fontAttr.familyName) {
        len = strlen(from->fontAttr.familyName) + 1;
        to->fontAttr.familyName = (char *)txloc(pid, len);
        strcpy((char *)to->fontAttr.familyName,
                (char *)from->fontAttr.familyName);
    }
    if (from->fontAttr.fontName) {
        len = strlen(from->fontAttr.fontName) + 1;
        to->fontAttr.fontName = (char *)txloc(pid, len);
        strcpy((char *)to->fontAttr.fontName,
                (char *)from->fontAttr.fontName);
    }
    if (from->strokepattern) {
        to->strokepattern = axtNewFillPattern(pid);
        axtCopyFillPattern(from->strokepattern, &to->strokepattern, pid);
    }
    if (from->fillpattern) {
        to->fillpattern = axtNewFillPattern(pid);
        axtCopyFillPattern(from->fillpattern, &to->fillpattern, pid);
    }
    if (from->screenFontName) {
        len = strlen(from->screenFontName) + 1;
        to->screenFontName = (char *)txloc(pid, len);
        strcpy((char *)to->screenFontName,
                (char *)from->screenFontName);
    }
    if (from->indent) {
        to->indent = axtNewDefaultIndent(pid);
        axtCopyIndent(from->indent, to->indent, pid);
    }
    if (from->dropShadow) {
        to->dropShadow = (axtDropShadow *) txloc(pid, sizeof(axtDropShadow));
        memset((char *) to->dropShadow, 0, sizeof (axtDropShadow));
        axtCopyDropShadow(from->dropShadow, to->dropShadow, pid);
    }
}

static void
copyPerChar(to, from, pid)
perChar *to, *from;
int pid;
{
    /* free-store appendages are: ctm, letterName */
    if (to->ctm) {
        txfree((char *)to->ctm);
        to->ctm = (axform *)NULL;
    }
    if (to->letterName) {
        txfree((char *)to->letterName);
        to->letterName = (char *)NULL;
    }
    *to = *from;
    if (from->ctm) {
        to->ctm = (axform *) txloc(pid, sizeof(axform));
        *to->ctm = *from->ctm;
    }
    if (from->letterName) {
        to->letterName = (char *)txloc(pid, strlen(from->letterName) + 1);
        strcpy(to->letterName, from->letterName);
    }
}

/* This routine is used to support interactive
 * display during edit. It doesn't do a complete
 * copy of the struct; in particular, it doesn't
 * copy the ->old member (typical use of this call
 * is to perform a copy onto ->old)
 * Doesn't copy whole.bounder, just nulls the pointlist ptr.
 * Doesn't copy userDefaults or userDefaultFlags;
 * Note that the copy describes an axtext
 * instance which represents that portion of the
 * original from editPos down to the end of the text.
 */
static void
copyAxtext(to, from)
axtext **to, *from;
{
    int i, j, charcount, segcount, fpos, lpos, segix;
    txAttr *firstSeg, *lastSeg;
    axtext *last;
    p32 oldOff;
    
        /* toss anything still on *to
         */
    last = *to;
    if (last) {
        daxtext(last);
    }
    else {
        last = (axtext *)txloc(from->pid, sizeof(axtext));
        memset((char *)last, 0, sizeof(axtext));
    }
    
    last->pid = from->pid;
    last->gContext = from->gContext;
    last->whole = from->whole;
    if (from->whole.widthBounder && from->whole.bounder.numPaths) {
            /* copy bounder (assumed to be single, rectangular path)
             */
        last->whole.bounder.numPaths = 1;
        last->whole.bounder.paths = (poly_type *) txloc(last->pid, sizeof(poly_type));
        memset((void *) last->whole.bounder.paths, 0, sizeof(poly_type));
        last->whole.bounder.paths[0].nPoints = 5;
        last->whole.bounder.paths[0].points =
            (TmPoint32 *) txloc(last->pid, 5 * sizeof(TmPoint32));
        for (i = 0; i < 5; i++) {
            last->whole.bounder.paths[0].points[i] =
                from->whole.bounder.paths[0].points[i];
        }
    }
    else {
        last->whole.bounder.paths = (poly_type *) NULL;
        last->whole.bounder.numPaths = 0;
    }
    charcount = from->whole.nChars - from->editState.editPos;
    last->whole.nChars = charcount;
    
    if (from->textstr && charcount) {
        last->textstr = (tx_char *)
                txloc(last->pid, charcount * sizeof(tx_char));
        for (i=from->editState.editPos, j=0; j<charcount; i++, j++) {
            last->textstr[j] = from->textstr[i];
        }
        last->wCharBuflen = charcount;
    }
    
        /* copy txAttr array
         */
    if (charcount) {
        firstSeg = attrSegAtPos(from, from->editState.editPos, &fpos);
        lastSeg = attrSegAtPos(from, from->whole.nChars-1, &lpos);
        segix = firstSeg - from->atts;
        segcount = lastSeg - firstSeg + 1;
        i = segcount * sizeof(txAttr);
        last->atts = (txAttr *)txloc(last->pid, i);
        last->txAttrBuflen = segcount;
        memset((char *)last->atts, 0, i);
        for (i=segix, j=0; j<segcount; i++, j++) {
            copyTxAttr(&last->atts[j], &from->atts[i], last->pid);
            if (j == 0) {
                /* editPos may be in the middle of the first txAttr copied
                 */
                last->atts[j].nChars = from->atts[i].nChars - fpos;
            }
        }
    }
    else if (from->atts) {
        last->atts = (txAttr *) txloc(last->pid, sizeof(txAttr));
        last->txAttrBuflen = 1;
        memset((char *)last->atts, 0, sizeof(txAttr));
        copyTxAttr(&last->atts[0], &from->atts[0], last->pid);
        last->atts[0].nChars = 0;
    }

    if (charcount) {
        /* copy perChar array
         */
        i = charcount * sizeof(perChar);
        last->perCharBuflen = charcount;
        if (i) {
            last->items = (perChar *) txloc(last->pid, i);
            memset((char *)(last)->items, 0, i);
            for (i=from->editState.editPos, j=0; j<charcount; i++, j++) {
                copyPerChar(&last->items[j], &from->items[i], last->pid);
                last->items[j].changed = last->items[j].deleted = 
                    last->items[j].damaged = last->items[j].moved = 
                        last->items[j].inserted = FALSE;
            }
        }
    }

        /* set last's position and 0th offset.  Position is taken
         * as equal to from's, with the offset to editPos stored as
         * last's 0th offset
         */
    last->whole.pos = from->whole.pos;
    if (from->whole.nChars) {
        getRelativeOffset(from, 0, from->editState.editPos, &oldOff);
        if (from->items) {
            oldOff.x -= from->items[0].sup.x;
            oldOff.y -= from->items[0].sup.y;
        }
        if (last->items) {
            last->items[0].offset = oldOff;
        }
    }
    
        /* for editState, copy only the background color/pix stuff,
         * and the fieldBounder.
         */
    memset((char *)&last->editState, 0, sizeof(eState));
    last->editState.background = from->editState.background;
    last->editState.backgroundPixel = from->editState.backgroundPixel;
    last->editState.javaPixel = from->editState.javaPixel;
    last->editState.backRect = from->editState.backRect;
    last->editState.fieldBounder = from->editState.fieldBounder;
    
        /* but copy the 'construction' state flags
         */
    last->constructed = from->constructed;
    last->measured = from->measured;
    last->displayable = from->displayable;
    
    *to = last;
}

/* test whether the txAttr will need to be split */
static int /* TRUE means split */
needSplit(txt, pos, whichEdge)
axtext *txt;
int pos; /* of edge of range affected by attr change */
int whichEdge; /* of range */
{
    int lpos;
    txAttr *thisAp;
    thisAp = attrSegAtPos(txt, pos, &lpos);
    if (lpos < 0) {
        return (FALSE);
    }
    if (whichEdge == LEFT_EDGE) {
        return (lpos != 0);
    }
    else { /* right edge */
        return (lpos != (thisAp->nChars - 1));
    }
}

/* split one attr seg into two copies of itself,
 * changing only the nChars members; 'pos' will
 * the leftmost position in the rightmost of the
 * resultant two segs */
static int
doSplit(txt, pos)
axtext *txt;
int pos;
{
    txAttr new, *tnew, *txAp, *tmpAp;
    int lpos, attrPos, nsegs;
    
        /* insert a copy of the affected attr */    
    memset((char *)&new, 0, sizeof(txAttr));
    txAp = attrSegAtPos(txt, pos, &lpos);
    attrPos = txAp - txt->atts;
    copyTxAttr(&new, txAp, txt->pid);
    tnew = &new;
    nsegs = segCount(txt, -1, &tmpAp);
    new.nChars = lpos;
    txAp->nChars -= lpos;
    if (!insertItems(TYPE_TXATTR, attrPos, (char *)tnew, 1,
                (char **)&txt->atts, nsegs, &txt->txAttrBuflen, txt->pid)) {
        tossTxAttr((char *) &new);
        return FALSE;
    }
    return TRUE;
}

/* compare txAttr range containing pos,len.  Join whenever
 * txAttr's are found that are adjacent & identical, or if
 * a txAttr is found whose character count has hit zero. */
static int
checkJoin(txt, pos, len)
axtext *txt;
int pos, len; /* check across attr range containing this char range */
{
    txAttr *first, *last, *this, *next, *txAp, *tmpAp;
    int fpos, lpos, attrIx, nsegs, empty;
    first = attrSegAtPos(txt, pos, &fpos);
    last = attrSegAtPos(txt, pos+len, &lpos);
    attrIx = first - txt->atts;
    txAp = first;
    while(txAp<last) {
        this = txAp;
        next = txAp + 1;
        if (!attrChange(this, next, (unsigned int)TXAllAtts, FALSE, txt->pid) ||
            this->nChars<=0) {
                /* identical to its neighbor, or empty; delete */
            empty = this->nChars <= 0;
            nsegs = segCount(txt, -1, &tmpAp);
            next->nChars += this->nChars;
            if (!deleteItems(TYPE_TXATTR, attrIx, 1, (char **)&txt->atts,
                        nsegs, &txt->txAttrBuflen, txt->pid )) {
                return FALSE;
            }
                /* delete can readdress the attr array, so.. */
            last = attrSegAtPos(txt, pos+len, &lpos); /* (loop variable..) */
                /* stay at same position in array until neighbor
                 * is different; ie, keep txAp at attrIx-th seg */
            segCount(txt, attrIx, &txAp); /* (loop variable..) */
            if (empty && attrIx) {
                    /* back up in case the seg just before the empty
                     * seg just deleted is identical to the next */
                attrIx--;
                txAp--;
            }
        }
        else {
            txAp++;
            attrIx++;
        }
    }
    return TRUE;
}

int
segCount(txt, ix, txAp)
axtext *txt;
int ix;        /* if >= 0, count only to ix */
txAttr **txAp; /* out: return & of ixth attr if ix >= 0 */
{
    txAttr *tmpAp;
    int charcount = txt->whole.nChars;
    int segix = 0;
    *txAp = (txAttr *)NULL;
    while(charcount > 0) {
        if (segix == ix) {
            *txAp = &txt->atts[segix++];
            break;
        }
        tmpAp = &txt->atts[segix++];
        charcount -= tmpAp->nChars;
    }
    return (segix);
}

/* public to private copy */
static void
copyInTextAttr(pub, newAttr, defP, flags, pid)
textAttrs *pub;
txAttr *newAttr;
txAttr *defP;
unsigned int flags;
int pid;
{

    if ((flags & TXfontFamily) && pub->fontFamily) {
        if (newAttr->fontAttr.familyName)
            txfree((char *)newAttr->fontAttr.familyName);
        newAttr->fontAttr.familyName = (char *)txloc(pid, 
            strlen(pub->fontFamily) + 1);
        strcpy((char *)newAttr->fontAttr.familyName,
            (char *)pub->fontFamily);
    }
    if ((flags & TXfillpattern) && pub->fillpattern) {
        if (newAttr->fillpattern)
            axtFreeFillPattern(newAttr->fillpattern);
        newAttr->fillpattern = axtNewFillPattern(pid);
        axtCopyFillPattern(pub->fillpattern, &newAttr->fillpattern, pid);
    }
    if ((flags & TXstrokepattern) && pub->strokepattern) {
        if (newAttr->strokepattern)
            axtFreeFillPattern(newAttr->strokepattern);
        newAttr->strokepattern = axtNewFillPattern(pid);
        axtCopyFillPattern(pub->strokepattern, &newAttr->strokepattern, pid);
    }
    if (flags & TXweight)
        newAttr->fontAttr.weight = pub->weight;
    if (flags & TXslant)
        newAttr->fontAttr.slant = pub->slant;
    if (flags & TXpointsize) {
        newAttr->fontAttr.ptsizeInt = pub->pointsize;
        newAttr->fontAttr.ptsize = (float) pub->pointsize / 10000.0;
    }
    if (flags & TXruling)
        newAttr->ruling = pub->ruling;
    if (flags & TXstrike)
        newAttr->strike_thru = pub->strike_thru;
    if (flags & TXsupx)
        newAttr->supx = pub->supx;
    if (flags & TXsupy)
        newAttr->supy = pub->supy;
    if (flags & TXkernx)
        newAttr->kernx = pub->kernx;
    if (flags & TXkerny)
        newAttr->kerny = pub->kerny;
    if (flags & TXnoRuleChar) {
        newAttr->noRuleChars[0] = pub->noRuleChars[0];
        newAttr->noRuleChars[1] = pub->noRuleChars[1];
    }
    defaultTransform(&newAttr->tctm);
    if (flags & TXangle) {
        newAttr->transform.angle = pub->angle;
        newAttr->tctm.known = FALSE;
    }
    if (flags & TXxscale) {
        newAttr->transform.xscale = pub->xscale;
        newAttr->tctm.known = FALSE;
    }
    if (flags & TXyscale) {
        newAttr->transform.yscale = pub->yscale;
        newAttr->tctm.known = FALSE;
    }
    if (flags & TXxshear) {
        newAttr->transform.xshear = pub->xshear;
        newAttr->tctm.known = FALSE;
    }
    if (flags & TXfillcolor)
        newAttr->fillcolor = pub->fillcolor;
    if (flags & TXfillbcolor)
        newAttr->fillbcolor = pub->fillbcolor;
    if (flags & TXfgPixel) {
        newAttr->fgPixel = pub->fgPixel;
        newAttr->useFgPixel = TRUE;
    }
    if (flags & TXstrokecolor)
        newAttr->strokecolor = pub->strokecolor;
    if (flags & TXstrokebcolor)
        newAttr->strokebcolor = pub->strokebcolor;
    
    if (flags & TXindent) {
        if (newAttr->indent) {
            axtFreeIndent(newAttr->indent);
            newAttr->indent = (axtIndent *) NULL;
        }
        if (pub->indent) {
            newAttr->indent = axtNewDefaultIndent(pid);
            axtCopyIndent(pub->indent, newAttr->indent, pid);
        }
    }
    if (flags & TXbullet) {
        if (newAttr->indent && newAttr->indent->bullet) {
            axtFreeBullet(newAttr->indent->bullet);
            newAttr->indent->bullet = (axtBullet *) NULL;
        }
        if (pub->bullet || (pub->indent && pub->indent->bullet)) {
            if (!newAttr->indent) {
                newAttr->indent = axtNewDefaultIndent(pid);
            }
            newAttr->indent->bullet = axtNewDefaultBullet(pid);
            if (pub->bullet) {
                axtCopyBullet(pub->bullet, newAttr->indent->bullet, pid);
            }
            else if (pub->indent && pub->indent->bullet) {
                axtCopyBullet(pub->indent->bullet, newAttr->indent->bullet, pid);
            }
        }
        if (newAttr->indent->bullet && newAttr->indent->bullet->bulletSize) {
            /* snapshot the bullet's size */
            int fsize;
            axtBullet *bb = newAttr->indent->bullet;
			if (flags & TXpointsize)
				fsize = newAttr->fontAttr.ptsizeInt;
			else
				fsize = defP->fontAttr.ptsizeInt;
            bb->bulletSize(newAttr->indent, fsize, 
                           &bb->bounder.origin.x, &bb->bounder.origin.y,
                           &bb->bounder.width, &bb->bounder.height);
        }
    }
    if (flags & TXdropShadow) {
        if (newAttr->dropShadow) {
            axtFreeDropShadow(newAttr->dropShadow);
            newAttr->dropShadow = (axtDropShadow *) NULL;
        }
        if (pub->shadow) {
            newAttr->dropShadow =
                (axtDropShadow *) txloc(pid, sizeof(axtDropShadow));
            memset((char *) newAttr->dropShadow, 0, sizeof(axtDropShadow));
            axtCopyDropShadow(pub->shadow, newAttr->dropShadow, pid);
        }
    }
}

/* private to public copy */
static void
copyOutTextAttr(pubAttr, attr, flags, pid)
textAttrs *pubAttr;
txAttr *attr;
unsigned int flags;
int pid;
{
    if ((flags & TXfontFamily) && attr->fontAttr.familyName) {
        if (pubAttr->fontFamily)
            txfree((char *)pubAttr->fontFamily);
        pubAttr->fontFamily = (char *) txloc(pid,
            strlen(attr->fontAttr.familyName) + 1);
        strcpy(pubAttr->fontFamily, attr->fontAttr.familyName);
    }
    if ((flags & TXfillpattern) && attr->fillpattern) {
        if (pubAttr->fillpattern)
            axtFreeFillPattern(pubAttr->fillpattern);
        pubAttr->fillpattern = axtNewFillPattern(pid);
        axtCopyFillPattern(attr->fillpattern, &pubAttr->fillpattern, pid);
    }
    if ((flags & TXstrokepattern) && attr->strokepattern) {
        if (pubAttr->strokepattern)
            axtFreeFillPattern(pubAttr->strokepattern);
        pubAttr->strokepattern = axtNewFillPattern(pid);
        axtCopyFillPattern(attr->strokepattern, &pubAttr->strokepattern, pid);
    }
    if (flags & TXweight) {
        pubAttr->weight = attr->fontAttr.weight;
    }
    if (flags & TXslant) {
        pubAttr->slant = attr->fontAttr.slant;
    }
    if (flags & TXpointsize) {
        pubAttr->pointsize = attr->fontAttr.ptsizeInt;
    }
    if (flags & TXruling) {
        pubAttr->ruling = attr->ruling;
    }
    if (flags & TXstrike) {
        pubAttr->strike_thru = attr->strike_thru;
    }
    if (flags & TXnoRuleChar) {
        pubAttr->noRuleChars[0] = attr->noRuleChars[0];
        pubAttr->noRuleChars[1] = attr->noRuleChars[1];
    }
    if (flags & TXangle) {
        pubAttr->angle = attr->transform.angle;
    }
    if (flags & TXxscale) {
        pubAttr->xscale = attr->transform.xscale;
    }
    if (flags & TXyscale) {
        pubAttr->yscale = attr->transform.yscale;
    }
    if (flags & TXxshear) {
        pubAttr->xshear = attr->transform.xshear;
    }
    if (flags & TXfillcolor) {
        pubAttr->fillcolor = attr->fillcolor;
    }
    if (flags & TXfillbcolor) {
        pubAttr->fillbcolor = attr->fillbcolor;
    }
    if (flags & TXfgPixel) {
        pubAttr->fgPixel = attr->fgPixel;
        pubAttr->useFgPixel = TRUE;
    }
    if (flags & TXstrokecolor) {
        pubAttr->strokecolor = attr->strokecolor;
    }
    if (flags & TXstrokebcolor) {
        pubAttr->strokebcolor = attr->strokebcolor;
    }
    if (flags & TXsupx) {
        pubAttr->supx = attr->supx;
    }
    if (flags & TXsupy) {
        pubAttr->supy = attr->supy;
    }
    if (flags & TXkernx) {
        pubAttr->kernx = attr->kernx;
    }
    if (flags & TXkerny) {
        pubAttr->kerny = attr->kerny;
    }
    if ((flags & TXindent) && attr->indent) {
        if (pubAttr->indent) {
            axtFreeIndent(pubAttr->indent);
        }
        pubAttr->indent = axtNewDefaultIndent(pid);
        axtCopyIndent(attr->indent, pubAttr->indent, pid);
    }
    if ((flags & TXbullet) && attr->indent && attr->indent->bullet) {
        if (pubAttr->bullet) {
            axtFreeBullet(pubAttr->bullet);
        }
        pubAttr->bullet = axtNewDefaultBullet(pid);
        axtCopyBullet(attr->indent->bullet, pubAttr->bullet, pid);
    }
    if ((flags & TXdropShadow) && attr->dropShadow) {
        if (pubAttr->shadow) {
            axtFreeDropShadow(pubAttr->shadow);
        }
        pubAttr->shadow =
                (axtDropShadow *) txloc(pid, sizeof(axtDropShadow));
        memset((char *) pubAttr->shadow, 0, sizeof(axtDropShadow));
        axtCopyDropShadow(attr->dropShadow, pubAttr->shadow, pid);
    }
}

/* compare two txAttr structs; possibly apply attr info */
static int /* true if different */
attrChange(txAp, newAp, flags, apply, pid)
txAttr *txAp;
txAttr *newAp;
unsigned int flags;
int apply; /* if true, apply changes from *newAp to *txAp */
int pid;
{
    int different = FALSE;

    if (flags & TXfontFamily) {
        if (differingStrings(newAp->fontAttr.familyName,
                        txAp->fontAttr.familyName)) {
            different = TRUE;
            if (apply) {
                if (txAp->fontAttr.familyName) {
                    txfree((char *)txAp->fontAttr.familyName);
                    txAp->fontAttr.familyName = (char *)NULL;
                }
                if (newAp->fontAttr.familyName) {
                    txAp->fontAttr.familyName =
                        (char *)txloc(pid, strlen(newAp->fontAttr.familyName)+1);
                    strcpy((char *)txAp->fontAttr.familyName,
                           (char *)newAp->fontAttr.familyName);
                }
            }
            else
                return TRUE;
        }
    }
    if (flags & TXstrokepattern) {
/* %%%%%%% need to compare fillpat data to prevent unneeded fragment'n */
        different = TRUE;  /* %%%%%% whether or not actually different */
        if (apply) {
            if (txAp->strokepattern)
                axtFreeFillPattern(txAp->strokepattern);
            if (newAp->strokepattern) {
                txAp->strokepattern = axtNewFillPattern(pid);
                axtCopyFillPattern(newAp->strokepattern,
                                   &txAp->strokepattern, pid);
            }
            else
                txAp->strokepattern = (attr_type *) NULL;
        }
        else
            return TRUE;
    }
    if (flags & TXfillpattern) {
/* %%%%%%% need to compare fillpat data to prevent unneeded fragment'n */
        different = TRUE;  /* %%%%%% whether or not actually different */
        if (apply) {
            if (txAp->fillpattern)
                axtFreeFillPattern(txAp->fillpattern);
            if (newAp->fillpattern) {
                txAp->fillpattern = axtNewFillPattern(pid);
                axtCopyFillPattern(newAp->fillpattern,
                                   &txAp->fillpattern, pid);
            }
            else
                txAp->fillpattern = (attr_type *) NULL;
        }
        else
            return TRUE;
    }
    if (flags & TXweight) {
        if (newAp->fontAttr.weight != txAp->fontAttr.weight) {
            different = TRUE;
            if (apply) {
                txAp->fontAttr.weight = newAp->fontAttr.weight;
            }
            else
                return TRUE;
        }
    }
    if (flags & TXslant) {
        if (newAp->fontAttr.slant != txAp->fontAttr.slant) {
            different = TRUE;
            if (apply) {
                txAp->fontAttr.slant = newAp->fontAttr.slant;
            }
            else
                return TRUE;
        }
    }
    if (flags & TXpointsize) {
        if (newAp->fontAttr.ptsize != txAp->fontAttr.ptsize) {
            different = TRUE;
            if (apply) {
                txAp->fontAttr.ptsize = newAp->fontAttr.ptsize;
                txAp->fontAttr.ptsizeInt = newAp->fontAttr.ptsizeInt;
            }
            else
                return TRUE;
        }
    }
    if (flags & TXruling) {
        if (newAp->ruling != txAp->ruling) {
            different = TRUE;
            if (apply) {
                txAp->ruling = newAp->ruling;
            }
            else
                return TRUE;
        }
    }
    if (flags & TXstrike) {
        if (newAp->strike_thru != txAp->strike_thru) {
            different = TRUE;
            if (apply) {
                txAp->strike_thru = newAp->strike_thru;
            }
            else
                return TRUE;
        }
    }
    if (flags & TXsupx) {
        if (newAp->supx != txAp->supx) {
            different = TRUE;
            if (apply) {
                txAp->supx = newAp->supx;
            }
            else
                return TRUE;
        }
    }
    if (flags & TXsupy) {
        if (newAp->supy != txAp->supy) {
            different = TRUE;
            if (apply) {
                txAp->supy = newAp->supy;
            }
            else
                return TRUE;
        }
    }
    if (flags & TXkernx) {
        if (newAp->kernx != txAp->kernx) {
            different = TRUE;
            if (apply) {
                txAp->kernx = newAp->kernx;
            }
            else
                return TRUE;
        }
    }
    if (flags & TXkerny) {
        if (newAp->kerny != txAp->kerny) {
            different = TRUE;
            if (apply) {
                txAp->kerny = newAp->kerny;
            }
            else
                return TRUE;
        }
    }
    if (flags & TXnoRuleChar) {
        if (newAp->noRuleChars[0] != NO_CHAR) {
            if (newAp->noRuleChars[0] != txAp->noRuleChars[0] &&
                newAp->noRuleChars[0] != txAp->noRuleChars[1]) {
                different = TRUE;
                if (apply) {
                    txAp->noRuleChars[0] = newAp->noRuleChars[0];
                    txAp->noRuleChars[1] = newAp->noRuleChars[1];
                }
                else
                    return TRUE;
            }
        }
        if (newAp->noRuleChars[1] != NO_CHAR) {
            if (newAp->noRuleChars[1] != txAp->noRuleChars[0] &&
                newAp->noRuleChars[1] != txAp->noRuleChars[1]) {
                different = TRUE;
                if (apply) {
                    txAp->noRuleChars[0] = newAp->noRuleChars[0];
                    txAp->noRuleChars[1] = newAp->noRuleChars[1];
                }
                else
                    return TRUE;
            }
        }
    }
    if (flags & TXangle) {
        if (newAp->transform.angle != txAp->transform.angle) {
            different = TRUE;
            if (apply) {
                txAp->transform.angle = newAp->transform.angle;
                txAp->tctm.known = FALSE;
            }
            else
                return TRUE;
        }
    }
    if (flags & TXxscale) {
        if (newAp->transform.xscale != txAp->transform.xscale) {
            different = TRUE;
            if (apply) {
                txAp->transform.xscale = newAp->transform.xscale;
                txAp->tctm.known = FALSE;
            }
            else
                return TRUE;
        }
    }
    if (flags & TXyscale) {
        if (newAp->transform.yscale != txAp->transform.yscale) {
            different = TRUE;
            if (apply) {
                txAp->transform.yscale = newAp->transform.yscale;
                txAp->tctm.known = FALSE;
            }
            else
                return TRUE;
        }
    }
    if (flags & TXxshear) {
        if (newAp->transform.xshear != txAp->transform.xshear) {
            different = TRUE;
            if (apply) {
                txAp->transform.xshear = newAp->transform.xshear;
                txAp->tctm.known = FALSE;
            }
            else
                return TRUE;
        }
    }
    if (flags & TXfillcolor) {
        if (newAp->fillcolor != txAp->fillcolor) {
            different = TRUE;
            if (apply) {
                txAp->fillcolor = newAp->fillcolor;
            }
            else
                return TRUE;
        }
    }
    if (flags & TXfillbcolor) {
        if (newAp->fillbcolor != txAp->fillbcolor) {
            different = TRUE;
            if (apply) {
                txAp->fillbcolor = newAp->fillbcolor;
            }
            else
                return TRUE;
        }
    }
    if (flags & TXfgPixel) {
        if ((newAp->useFgPixel != txAp->useFgPixel) ||
            (newAp->useFgPixel && (newAp->fgPixel != txAp->fgPixel))) {
                /* usage bits are different, or pixvals are different
                 */
            different = TRUE;
            if (apply) {
                txAp->fgPixel = newAp->fgPixel;
                txAp->useFgPixel = newAp->useFgPixel;
                if (!txAp->useFgPixel) {
                    txAp->fgPixel = (unsigned long) 0;
                }
            }
            else
                return TRUE;
        }
    }
    if (flags & TXstrokecolor) {
        if (newAp->strokecolor != txAp->strokecolor) {
            different = TRUE;
            if (apply) {
                txAp->strokecolor = newAp->strokecolor;
            }
            else
                return TRUE;
        }
    }
    if (flags & TXstrokebcolor) {
        if (newAp->strokebcolor != txAp->strokebcolor) {
            different = TRUE;
            if (apply) {
                txAp->strokebcolor = newAp->strokebcolor;
            }
            else
                return TRUE;
        }
    }
    if (flags & TXindent) {
        if (differingIndents(newAp->indent, txAp->indent)) {
            different = TRUE;
            if (apply) {
                if (newAp->indent) {
                    if (!txAp->indent) {
                        txAp->indent = axtNewDefaultIndent(pid);
                    }
                    axtCopyIndent(newAp->indent, txAp->indent, pid);
                }
                else if (txAp->indent) {
                    axtFreeIndent(txAp->indent);
                    txAp->indent = (axtIndent *) NULL;
                }
            }
            else
                return TRUE;
        }
    }
    if (flags & TXbullet) {
        axtBullet *bcur, *bnew;
        
        bnew = newAp->indent ? newAp->indent->bullet : (axtBullet *) NULL;
        bcur = txAp->indent ? txAp->indent->bullet : (axtBullet *) NULL;
        
        if (differingBullets(bnew, bcur)) {
            different = TRUE;
            if (apply) {
                if (newAp->indent && newAp->indent->bullet) {
                    if (!txAp->indent) {
                        txAp->indent = axtNewDefaultIndent(pid);
                    }
                    if (!txAp->indent->bullet) {
                        txAp->indent->bullet = axtNewDefaultBullet(pid);
                    }
                    axtCopyBullet(newAp->indent->bullet, txAp->indent->bullet, pid);
                    txAp->indent->bullet->changed = TRUE;
                }
                else if (txAp->indent && txAp->indent->bullet) {
                    axtFreeBullet(txAp->indent->bullet);
                    txAp->indent->bullet = (axtBullet *) NULL;
                }
            }
            else
                return TRUE;
        }
    }
    if (flags & TXdropShadow) {
/* %%%%%%% need to compare attr_type data to prevent unneeded fragment'n */
        different = TRUE;
        if (apply) {
            if (newAp->dropShadow) {
                if (!txAp->dropShadow) {
                    txAp->dropShadow =
                        (axtDropShadow *) txloc(pid, sizeof(axtDropShadow));
                    memset((char *) txAp->dropShadow, 0,
                           sizeof(axtDropShadow));
                }
                axtCopyDropShadow(newAp->dropShadow, txAp->dropShadow, pid);
            }
            else if (txAp->dropShadow) {
                axtFreeDropShadow(txAp->dropShadow);
                txAp->dropShadow = (axtDropShadow *) NULL;
            }
        }
        else
            return TRUE;
    }
    
    if (apply && affectsLetterforms(flags)) {
        if(txAp->screenFontName) {
            txfree((char *)txAp->screenFontName);
            txAp->screenFontName = (char *)NULL;
        }
    }
    if (apply && affectsFontName(flags)) {
        if(txAp->fontAttr.fontName) {
            txfree((char *)txAp->fontAttr.fontName);
            txAp->fontAttr.fontName = (char *)NULL;
        }
    }
    return different;
}

int
differingStrings(s1, s2)
char *s1, *s2;
{
    if (!s1 && !s2)
        return FALSE; /* they're the same: both useless */
    else if (!s1 || !s2)
        return TRUE; /* one's null, the other's not */
    else
        return (strcmp(s1, s2));
}

int
differingBullets(b1, b2)
axtBullet *b1, *b2;
{
    if (!b2 && !b2) {
        return FALSE;
    }
    else if (!b1 || !b2) {
        return TRUE;
    }
    else if (b1->bulletOffset.x != b2->bulletOffset.x) {
        return TRUE;
    }
    else if (b1->bulletOffset.y != b2->bulletOffset.y) {
        return TRUE;
    }
    else if (b1->type != b2->type) {
        return TRUE;
    }
    else if (b1->color != b2->color) {
        return TRUE;
    }
    else if (b1->scale != b2->scale) {
        return TRUE;
    }
    else if (b1->font != b2->font) {
        return TRUE;
    }
    else if (b1->glyph != b2->glyph) {
        return TRUE;
    }
    else if (b1->bulletDraw != b2->bulletDraw) {
        return TRUE;
    }
    else if (b1->bulletSize != b2->bulletSize) {
        return TRUE;
    }
    else
        return FALSE;
}

int
differingIndents(i1, i2)
axtIndent *i1, *i2;
{
    if (!i1 && !i2) {
        return FALSE;
    }
    else if (!i1 || !i2) {
        return TRUE;
    }
    else if (i1->firstLeft != i2->firstLeft) {
        return TRUE;
    }
    else if (i1->left != i2->left) {
        return TRUE;
    }
    else if (i1->right != i2->right) {
        return TRUE;
    }
    else if (i1->topLineHeight != i2->topLineHeight) {
        return TRUE;
    }
    else if (i1->midLineHeight != i2->midLineHeight) {
        return TRUE;
    }
    else if (i1->bottomLineOffset != i2->bottomLineOffset) {
        return TRUE;
    }
    else if (differingBullets(i1->bullet, i2->bullet)) {
        return TRUE;
    }
    else {
        return FALSE;
    }
}

int
affectsLetterforms(flags)
unsigned int flags;
{
    return (flags & (
            TXfontFamily | TXweight | TXslant | TXpointsize |
            TXangle | TXxscale | TXxshear | TXfillpattern |
            TXstrokepattern | TXyscale | TXdropShadow
        ));
}

int
affectsFontName(flags)
unsigned int flags;
{
    return (flags &
            (TXfontFamily | TXweight | TXslant | TXpointsize));
}

static int
getCursorPoints(txt, type, active, txAp, pts, clipRect)
axtext *txt;
int type, active;
txAttr *txAp;
XPoint *pts;          /* out: of which there are four */
r32 *clipRect;        /* out: clip to possibly use with display */
{
    int pointcount, i, dstx, dsty, tpointsize;
    axform ctm, *ctmp;
    float xscale, yscale;
    XPoint cpoints[4];

    if (!txAp->tctm.unity) {
            /* account for yscale in determining cursor size */
        extractYscale(&txAp->tctm, &yscale);
        tpointsize = (int) txAp->fontAttr.ptsize * yscale;
    }
    else {
        tpointsize = (int) txAp->fontAttr.ptsize;
    }
    if (type == CURSOR_TRIANGLE) {
        if (active) {
            pointcount = 3;
            getTriPoints(cpoints, tpointsize);
        }
        else {
            pointcount = 4;
            getDiamondPoints(cpoints, tpointsize);
        }
    }
    else {
        pointcount = 2;
        getBarPoints(txt, txAp, cpoints, clipRect);
    }
        /* suppress text-size clip */
    clipRect->origin.x = clipRect->origin.y =
        clipRect->height = clipRect->width = 0;

        /* now transform the cursor shape */
    if (!txAp->tctm.unity) {
            /* remove x and y scale factors,
               to keep size reasonable. (neglect translation) */
        memset((char *) &ctm, 0, sizeof(ctm));
        extractXscale(&txAp->tctm, &xscale);
        ctm.forward.ca = txAp->tctm.forward.ca / xscale;
        ctm.forward.cb = txAp->tctm.forward.cb / yscale;
        ctm.forward.cc = txAp->tctm.forward.cc / xscale;
        ctm.forward.cd = txAp->tctm.forward.cd / yscale;
        ctmp = &ctm;
    }
    else {
        ctmp = &txAp->tctm;
    }
    for (i=0; i<pointcount; i++) {
        forwardSXRT((int)cpoints[i].x, (int)cpoints[i].y,
                &dstx, &dsty, ctmp);
        pts[i].x = (short) dstx;
        pts[i].y = (short) dsty;
    }
    return pointcount;
}

static void
getBarPoints(txt, txAp, pts, clipRect)
axtext *txt;
txAttr *txAp;          /* in: containing text attr info */
XPoint *pts;           /* out */
r32 *clipRect;         /* out: clip rectangle to possibly use */
{
    int height, ascent;
    char *errString;
    r32 bbox;
   
        /* start with height derived from pointsize; guess
           an ascent of 80% height */
    height = mToPix(1000, txAp->fontAttr.ptsize, ScreenRes);
    ascent = (int)((height * 8.0) / 10.0);

    if (txt->whole.nChars != 0 && !txt->whole.bounder.numPaths) {
            /* for nonempty unbounded text, get the overall
               bounder for clipping purposes */
        if (errString = getOverallBounder(txt, 0, -1, &bbox)) {
            freeErrString(errString);
            clipRect->height = clipRect->width = 0;
            clipRect->origin.x = clipRect->origin.y = 0;
        }
        else {
            *clipRect = bbox;
        }
        if (!txAp->tctm.rotated) {
                /* in this hack, relax the horizontal dimension of
                   the clip; this is to ensure that cursors won't
                   clip out in the most common case.. */
            clipRect->origin.x -= 1000;
            clipRect->width += 2000;
            
        }
    }
    else {
        clipRect->origin.x = clipRect->origin.y =
            clipRect->width = clipRect->height = 0;
    }
    pts[0].x = pts[1].x = 0;
    pts[0].y = -ascent;
    pts[1].y = height - ascent;
}

void
getTriPoints(pts, pointsize)
XPoint *pts; /* out */
int pointsize;
{
    if (pointsize < 18) {
            /* Cursor whose height is 8. */
        pts[0].x = -3;
        pts[0].y = 5;
        pts[1].x = 0;
        pts[1].y = -2;
        pts[2].x = 3;
        pts[2].y = 5;
    }
    else if (pointsize < 48) {
            /* Cursor whose height is 10. */
        pts[0].x = -4;
        pts[0].y = 5;
        pts[1].x = 0;
        pts[1].y = -4;
        pts[2].x = 4;
        pts[2].y = 5;
    }
    else {
            /* Cursor whose height is 14. */
        pts[0].x = -6;
        pts[0].y = 7;
        pts[1].x = 0;
        pts[1].y = -6;
        pts[2].x = 6;
        pts[2].y = 7;
    }
}

void
getDiamondPoints(pts, pointsize)
XPoint *pts; /* out */
int pointsize;
{
    getTriPoints(pts, pointsize);
    switchCursor(pts, TRUE);
}

static void
switchCursor(pts, active)
XPoint *pts; /* out */
int active;
{
    XPoint tpt[4];
    int xtmp, ytmp, i;

    /*     active                inactive
     *       pt1                    pt1
     *                         pt0       pt2
     *   pt0      pt2               pt3
     */
    xtmp = (pts[2].x - pts[0].x) >> 1;
    tpt[1] = pts[1];
    if (active) {
        /* switch from active (black tri) to inactive
         * (grey diamond) */
        ytmp = (pts[0].y - pts[1].y) >> 1;
        tpt[0].x = pts[0].x;
        tpt[0].y = pts[0].y - ytmp;
        tpt[2].x = pts[2].x;
        tpt[2].y = pts[0].y - ytmp;
        tpt[3].x = pts[0].x + xtmp;
        tpt[3].y = pts[0].y;
    }
    else {
        /* switch from inactive (grey diamond) to
         * active (black tri) */
        tpt[0].x = pts[0].x;
        tpt[0].y = pts[3].y;
        tpt[2].x = pts[2].x;
        tpt[2].y = pts[3].y;
    }
    for(i=0; i<4; i++)
        pts[i] = tpt[i];
}

static GC
getGC(drw, function, foreground, background, planemask)
Drawable drw;
int function;
unsigned long foreground, background, planemask;
{
    XGCValues gcv;
    unsigned long mask;

    memset((char *)&gcv, 0, sizeof(XGCValues));
    mask = GCForeground | GCBackground | GCFunction | GCPlaneMask;
    gcv.function = function;
    gcv.foreground = foreground;
    gcv.background = background;
    gcv.plane_mask = planemask;
    return (axCreateGC(Dpy, drw, DefaultScreen(Dpy),
                       DisplayPlanes(Dpy, DefaultScreen(Dpy)),
                       (XFontStruct *) 0, mask, &gcv));
}

/* get a pixmap suitable for use in drawing on (other) 1-deep pixmaps */
static GC
getPixmapGC(drw, function, foreground, background, planemask)
Drawable drw;
int function;
unsigned long foreground, background, planemask;
{
    XGCValues gcv;
    unsigned long mask;
    Pixmap pix;
    GC tgc;

    memset((char *)&gcv, 0, sizeof(XGCValues));
    mask = GCForeground | GCBackground | GCFunction | GCPlaneMask;
    gcv.function = function;
    gcv.foreground = foreground;
    gcv.background = background;
    gcv.plane_mask = planemask;
        /* get 1-deep pixmap just for GC create reference */
    pix = XCreatePixmap(Dpy, drw, (unsigned int)1, (unsigned int)1,
                        (unsigned int)1);
    tgc = axCreateGC(Dpy, pix, DefaultScreen(Dpy), 1, (XFontStruct *) 0,
                     mask, &gcv);
    XFreePixmap(Dpy, pix);
    return tgc;
}

static void
freeBounder(bounder)
poly *bounder;
{
    int ix;
    if (!bounder)
        return;
    for (ix = 0; ix < bounder->numPaths; ix++) {
        txfree((char *) bounder->paths[ix].points);
        txfree((char *) bounder->paths[ix].controls);
    }
    txfree((char *) bounder->paths);
    bounder->paths = (poly_type *) NULL;
    bounder->numPaths = 0;
}

int /* TRUE if *ascent and *descent were written */
getTextHeight(atts, ascent, descent)
textAttrs *atts;
int *ascent; /* return ascent in mils */
int *descent; /* return descent in mils */
{
    char *errString, *fontName = NULL;
    fontTabEnt *fontRec = NULL;
    fontMetrics *metrics = NULL;

    if (!atts || !ascent || !descent)
        return FALSE;
        /* the atts struct is assumed to have:
         * fontFamily, weight, slant, pointsize */
    if (errString = axGetFontRecordName(atts->fontFamily, atts->weight,
                        atts->slant, &fontName)) {
        freeErrString(errString);
        return FALSE;
    }
    if (!fontName)
        return FALSE;
            /* use fontName to get the font record */
    if (errString = axGetFontRecord(fontName, &fontRec)) {
        if (fontName)
            txfree((char *) fontName);
        freeErrString(errString);
        return FALSE;
    }
    if (!fontRec->targetMetricsNames[0]) {
        if (fontName)
            txfree((char *) fontName);
        return FALSE;
    }
    if (errString = axGetTargetMetrics(fontRec->targetMetricsNames[0],
                                       fontRec,
                                       &metrics)) {
        if (fontName)
            txfree((char *) fontName);
        freeErrString(errString);
        return FALSE;
    }
    if (!metrics) {
        if (fontName)
            txfree((char *) fontName);
        return FALSE;
    }
    *ascent = emToMils(metrics->xfont->max_bounds.ascent,
                        atts->pointsize);
    *descent = emToMils(metrics->xfont->max_bounds.descent,
                        atts->pointsize);
    if (fontName)
        txfree((char *) fontName);
    return TRUE;
}

void
txtSetClip(txt, gc, clip, clip2)
axtext *txt;
GC gc;
r32 *clip;
r32 *clip2;  /* intersect this if present */
{
    XRectangle clipRect;
    r32 lclip;
    int tx, ty;
    int x1, x2, y1, y2;


    if (!clip && !clip2)
        return;

    if (clip && clip2) {
        intersectRects(clip, clip2, &lclip);
    }
    else if (clip) {
        lclip = *clip;
    }
    else {
        lclip = *clip2;
    }

#if FOR_FILTER
    cvtConvMilsToPix(txt->whole.cvt, lclip.origin.x, lclip.origin.y, &tx, &ty);
    clipRect.x = (short) tx - 1;  /* blatant fudge */
    clipRect.y = (short) ty - 1;
    cvtConvMilsToPix(txt->whole.cvt, lclip.width, lclip.height, &tx, &ty);
    clipRect.width = (short) tx + 2;    /* blatant fudge */
    clipRect.height = (short) ty + 2;
#else
    x1 = lclip.origin.x;
    y1 = lclip.origin.y;
    x2 = lclip.width  + x1 - 1;
    y2 = lclip.height + y1 - 1;
    convMilsToPix(x1, y1, &x1, &y1);
    convMilsToPix(x2, y2, &x2, &y2);
    clipRect.x =  x1;
    clipRect.y =  y1;
    clipRect.width = x2 - x1 + 1;
    clipRect.height = y2 - y1 + 1;
#endif

    XSetClipRectangles(Dpy, gc, 0, 0, &clipRect, 1, Unsorted);
}

void
axUnSetClip(gc, clip)
GC gc;
r32 *clip;
{
    if (!clip)
        return;
    XSetClipMask(Dpy, gc, None);
}

static void
intersectRects(r1, r2, out)
r32 *r1, *r2;
r32 *out;      /* out: init this */
{
    int tmp1, tmp2;
    int maxOx, minCx; /* limits in x direction; max origin, min corner */
    int maxOy, minCy; /* limits in y direction */

    if (r1->origin.x > r2->origin.x) {
        maxOx = r1->origin.x;
    }
    else {
        maxOx = r2->origin.x;
    }
    if ((tmp1 = r1->origin.x + r1->width - 1) <
        (tmp2 = r2->origin.x + r2->width - 1)) {
        minCx = tmp1;
    }
    else {
        minCx = tmp2;
    }
    if (r1->origin.y > r2->origin.y) {
        maxOy = r1->origin.y;
    }
    else {
        maxOy = r2->origin.y;
    }
    if ((tmp1 = r1->origin.y + r1->height - 1) <
        (tmp2 = r2->origin.y + r2->height - 1)) {
        minCy = tmp1;
    }
    else {
        minCy = tmp2;
    }
    if (maxOx < minCx && maxOy < minCy) {
        out->origin.x = maxOx;
        out->origin.y = maxOy;
        out->width = minCx - maxOx + 1;
        out->height = minCy - maxOy + 1;
    }
    else {
        out->origin.x = out->origin.y = 0;
        out->width = out->height = 0;
    }
}


/* : dexformedBounderCopy - take a de-transformed copy of the
 * text bounder (where the transform is the field-level xform only).
 * this should be called only when the field xform is nonunity.
 */
static poly *      /* write tx's bounder de-xformed thru field xform */
dexformedBounderCopy(tx, fieldXform)
axtext *tx;          /* in */
axform *fieldXform;  /* in */
{
    int ix, jx;
    poly *dexformed;

    /* remove the field transform from the bounder
       in order to find the intercepts in untransformed
       space; then restore the transform on the intercepts. */

    dexformed = (poly *) txloc(tx->pid, sizeof(poly));

        /* alloc a same-sized bounder copy */
    dexformed->numPaths = tx->whole.bounder.numPaths;
    dexformed->paths = (poly_type *) txloc(tx->pid, 
        dexformed->numPaths * sizeof(poly_type));
    for (ix = 0; ix < tx->whole.bounder.numPaths; ix++) {
        dexformed->paths[ix].nPoints =
            tx->whole.bounder.paths[ix].nPoints;
        if (dexformed->paths[ix].nPoints) {
            dexformed->paths[ix].points = (TmPoint32 *) txloc(tx->pid, 
                dexformed->paths[ix].nPoints * sizeof(TmPoint32));
        }
        else {
            dexformed->paths[ix].points = (TmPoint32 *) NULL;
        }
        
        dexformed->paths[ix].nCtrls =
            tx->whole.bounder.paths[ix].nCtrls;
        if (dexformed->paths[ix].nCtrls) {
            dexformed->paths[ix].controls = (TmPoint32 *) txloc(tx->pid, 
                dexformed->paths[ix].nCtrls * sizeof(TmPoint32));
        }
        else {
            dexformed->paths[ix].controls = (TmPoint32 *) NULL;
        }
    }
        /* do a dexformed copy of the bounder */
    for (ix = 0; ix < tx->whole.bounder.numPaths; ix++) {
        for (jx = 0; jx < dexformed->paths[ix].nPoints; jx++) {
            reverseSXRT(tx->whole.bounder.paths[ix].points[jx].x,
                        tx->whole.bounder.paths[ix].points[jx].y,
                        &dexformed->paths[ix].points[jx].x,
                        &dexformed->paths[ix].points[jx].y, fieldXform);
        }
        for (jx = 0; jx < dexformed->paths[ix].nCtrls; jx++) {
            reverseSXRT(tx->whole.bounder.paths[ix].controls[jx].x,
                        tx->whole.bounder.paths[ix].controls[jx].y,
                        &dexformed->paths[ix].controls[jx].x,
                        &dexformed->paths[ix].controls[jx].y, fieldXform);
        }
    }

    return(dexformed);
}

/* return rect describing min/max extent of bounder path */
static void
getTxBounderRect(bounder, rect)
poly *bounder;
r32 *rect;     /* out */
{
    int ix, px, minx, miny, maxx, maxy;

    minx = miny = MAXLRU;  /* very large */
    maxx = maxy = -MAXLRU;

    for(px = 0; px < bounder->numPaths; px++) {
        for (ix = 0; ix < bounder->paths[px].nPoints; ix++) {
            if (bounder->paths[px].points[ix].x < minx)
                minx = bounder->paths[px].points[ix].x;
            if (bounder->paths[px].points[ix].y < miny)
                miny = bounder->paths[px].points[ix].y;
            if (bounder->paths[px].points[ix].x > maxx)
                maxx = bounder->paths[px].points[ix].x;
            if (bounder->paths[px].points[ix].y > maxy)
                maxy = bounder->paths[px].points[ix].y;
        }
        for (ix = 0; ix < bounder->paths[px].nCtrls; ix++) {
            if (bounder->paths[px].controls[ix].x < minx)
                minx = bounder->paths[px].controls[ix].x;
            if (bounder->paths[px].controls[ix].y < miny)
                miny = bounder->paths[px].controls[ix].y;
            if (bounder->paths[px].controls[ix].x > maxx)
                maxx = bounder->paths[px].controls[ix].x;
            if (bounder->paths[px].controls[ix].y > maxy)
                maxy = bounder->paths[px].controls[ix].y;
        }
    }
    rect->origin.x = minx;
    rect->origin.y = miny;
    rect->width = maxx - minx + 1;
    rect->height = maxy - miny + 1;
}

int
useForegroundColorForBlack(fill)
attr_type *fill;
{
    attr_type *fillAtt = (attr_type *) fill;
    if (fillAtt) {
        return(fillAtt->use_ffg);
    }
    else {
        return (TRUE);
    }
}

    /* allocate clean memory for a fill pattern struct instance */
attr_type *
axtNewFillPattern(pid)
int pid;
{
    attr_type *attr;
    attr = (attr_type *)txloc(pid, sizeof(attr_type));
    memset((char *)attr, 0, sizeof(attr_type));
    return(attr);
}

void
axtCopyFillPattern(from, to, pid)
attr_type *from;
attr_type **to;
int pid;
{
    attr_type *attr_src, *attr_dst;
    int byte_wid, size;

    attr_src = (attr_type *)from;
    attr_dst = (attr_type *)*to;
    *attr_dst = *attr_src;

    /* fg fill color */
    if ( attr_src->ffg.name )
    {
        attr_dst->ffg.name = (char *)txloc(pid,
            strlen(attr_src->ffg.name) + 1);
        strcpy(attr_dst->ffg.name, attr_src->ffg.name);
    }

    /* bg fill color */
    if ( attr_src->fbg.name )
    {
        attr_dst->fbg.name = (char *)txloc(pid, 
            strlen(attr_src->fbg.name) + 1);
        strcpy(attr_dst->fbg.name, attr_src->fbg.name);
    }

    /* fill data */
    if ( attr_src->f_data && attr_src->f_dhyt && attr_src->f_dwid ) {
        if ( attr_src->f_ddep  == 8 )
            byte_wid = attr_src->f_dwid + attr_src->f_dwid%2;
        else
            byte_wid =  (attr_src->f_dwid+7)/8 + ((attr_src->f_dwid+7)/8)%2;
        size = attr_src->f_dhyt * byte_wid;
        attr_dst->f_data = (char *)txloc(pid, size);
        memcpy((char *)attr_dst->f_data, (char *)attr_src->f_data, size);
    }


    /* fg line color */
    if ( attr_src->lfg.name )
    {
        attr_dst->lfg.name = (char *)txloc(pid, 
            strlen(attr_src->lfg.name) + 1);
        strcpy(attr_dst->lfg.name, attr_src->lfg.name);
    }

    /* bg line color */
    if ( attr_src->lbg.name )
    {
        attr_dst->lbg.name = (char *)txloc(pid, 
            strlen(attr_src->lbg.name) + 1);
        strcpy(attr_dst->lbg.name, attr_src->lbg.name);
    }

    /* line fill data */
    if ( attr_src->l_data && attr_src->l_dhyt && attr_src->l_dwid ) {
        if ( attr_src->l_ddep  == 8 )
            byte_wid = attr_src->l_dwid + attr_src->l_dwid%2;
        else
            byte_wid =  (attr_src->l_dwid+7)/8 + ((attr_src->l_dwid+7)/8)%2;
        size = attr_src->l_dhyt * byte_wid;
        attr_dst->l_data = (char *)txloc(pid, size);
        memcpy((char *)attr_dst->l_data, (char *)attr_src->l_data, size);
    }

    /* shadow color */
    if ( attr_src->sfg.name )
    {
        attr_dst->sfg.name = (char *)txloc(pid, strlen(attr_src->sfg.name) + 1);
        strcpy(attr_dst->sfg.name, attr_src->sfg.name);
    }
}

void
axtDuplicateFillPattern(from, to, pid)
attr_type *from;
attr_type **to;
int pid;
{
    *to = axtNewFillPattern(pid);
    axtCopyFillPattern(from, to, pid);
}

void
axtFreeFillPattern(pat)
attr_type *pat;
{
    attr_type *attr;

    if (!pat)
        return;

    attr = (attr_type *)pat;
    if ( attr->f_data )
        txfree(attr->f_data);
    if ( attr->l_data )
        txfree(attr->l_data);

    if ( attr->ffg.name )
        txfree(attr->ffg.name);
    if ( attr->fbg.name )
        txfree(attr->fbg.name);
    if ( attr->lfg.name )
        txfree(attr->lfg.name);
    if ( attr->lbg.name )
        txfree(attr->lbg.name);
    if (attr->sfg.name)
        txfree(attr->sfg.name);
    
    txfree(attr);
}

void
axtCopyColor(fromcolor, tocolor, pid)
color_type *fromcolor, *tocolor;
int pid;
{
    if (tocolor->name) {
        txfree((char *) tocolor->name);
    }

    *tocolor = *fromcolor;

    if (fromcolor->name) {
        tocolor->name = (char *) txloc(pid, strlen(fromcolor->name) + 1);
        strcpy(tocolor->name, fromcolor->name);
    }
}

void
axtToggleForegroundColor(txt, fgpixel)
axtext *txt;          /* in */
int fgpixel;          /* in */
{
    int n_atts, i;
    txAttr *txAp;
        
    n_atts = segCount(txt, -1, &txAp);
    for ( i = 0; i<n_atts; i++ ) {
        txt->atts[i].useFgPixel  = !txt->atts[i].useFgPixel;
        txt->atts[i].fgPixel     = fgpixel;
    }
}

/* : milliemCharWid - get text character's horizontal dimensions from
 * XFontStruct-format printer metrics data; numbers returned in milliems */
int /* return setwidth in milliems */
milliemCharWid(textchar, xf, lbear, rbear)
tx_char textchar;
XFontStruct *xf;
int *lbear, *rbear;
{
    XCharStruct *xc;

    xc = getCharStruct((int) textchar, xf);
    *lbear = xc->lbearing;
    *rbear = xc->rbearing;
    return (xc->width);
}

/* : millemFontHeight - return overall font height and ascent
 * from XFontStruct-format printer metrics; numbers returned in milliems */
int
milliemFontHeight(fs, ascentP)
XFontStruct *fs;  /* in: printer metrics */
int *ascentP;     /* out: fill this in */
{
    int fontHeight;
    if (!fs)
        return 0;
    fontHeight = fs->max_bounds.ascent + fs->max_bounds.descent;
    if (ascentP) {
        *ascentP = fs->max_bounds.ascent;
    }
    return (fontHeight);
}

/* : axtAdjustRulingEndpoints - Adjust the endpoints of the
 * passed ruling endpoints corresponding to text {startix,len}
 * Adjustments are made to account for kerning, hanging spaces, etc..
 */
void
axtAdjustRulingEndpoints(tx, startix, len, ifrom, ito, ofrom, oto)
axtext *tx;     /* in */
int startix;    /* in: starting index of range to underline */
int len;        /* in: #chars to underline */
p32 *ifrom;     /* in: tentative upstream endpoint */
p32 *ito;       /* in: tentative downstream endpoint */
p32 *ofrom;     /* out: adjusted upstream endpoint */
p32 *oto;       /* out: adjusted downstream endpoint */
{
    int ix;
    p32 lpos;
    perChar *pc;
    
    if (!len) {
        *ofrom = *ifrom;
        *oto = *ito;
        return;
    }

        /* look at left endpoint first
         */
    lpos = *ifrom;
    pc = &tx->items[startix];
    
    if (pc->linebreak) {
            /* pc is 1st char on a line;  adjust forward
             * if needed to reject any left-side hanging chars
             */
        for (ix = startix; ix < startix + len; ix++) {
            pc = &tx->items[ix];
            if (!pc->hanging) {
                break;
            }
            lpos.x += (pc->kern.x + pc->advance.x);
            lpos.y += (pc->kern.y + pc->advance.y);
        }
    }
    else if (pc->offset.x || pc->offset.y) {
            /* pc has non-linebreak offset; adjust backward
             * by that amount
             */
        lpos.x -= pc->offset.x;
        lpos.y -= pc->offset.y;
    }
    else if (startix) {
            /* for a mid-line character, remove any kerning
             * on that char (*ifrom arrives containing that kern)
             */
        lpos.x -= pc->kern.x;
        lpos.y -= pc->kern.y;
    }
    *ofrom = lpos;
    
        /* now look at the right endpoint; sweep it backwards
         * for as far has right-side hanging chars are found
         */
    lpos = *ito;
    for (ix = startix + len - 1; ix >= startix; ix--) {
        pc = &tx->items[ix];
        if (!pc->hanging) {
            break;
        }
        lpos.x -= (pc->kern.x + pc->advance.x);
        lpos.y -= (pc->kern.y + pc->advance.y);
    }
    *oto = lpos;
}

/* : axtAcuteVectors - return whether the angle between
 * the passed vectors is acute
 */
static int
axtAcuteVectors(vec1, vec2)
p32 *vec1;
p32 *vec2;
{
    int v1Hyp, v2Hyp;
    int v1Ang, v2Ang;
    int diffAng;
        
    v1Hyp = axtHypot(0, 0, vec1->x, vec1->y);
    v2Hyp = axtHypot(0, 0, vec2->x, vec2->y);
    if (v1Hyp > 0) {
        v1Ang = axtGetAngle((float)((float) vec1->y /
                                     (float) v1Hyp),
                             (float)((float) vec1->x /
                                     (float) v1Hyp));
    }
    else {
        v1Ang = 0;
    }
    if (v2Hyp > 0) {
        v2Ang = axtGetAngle((float)((float) vec2->y /
                                     (float) v2Hyp),
                             (float)((float) vec2->x /
                                     (float) v2Hyp));
    }
    else {
        v2Ang = 0;
    }
    
    diffAng = abs(v1Ang - v2Ang);
    if (diffAng < 900000 || diffAng > 2700000) {
        return TRUE;
    }
    else {
        return FALSE;
    }
}   

int
axtTabHasWidth(txt, txAp, pc, pcIndex)
axtext *txt;                 /* in */
txAttr *txAp;                /* in: att seg containing pc */
perChar *pc;                 /* in: character within att seg */
int pcIndex;                 /* in: char's index within text */
{
    if (txAp->indent && txAp->indent->bullet && txAp->indent->bullet->type) {
        if (!txt->whole.nChars)
            return FALSE;
        if ((pc && CHAR_IS_TABCHAR((int) pc->charcode)) &&
             txt->items[pcIndex-1].charcode) {
            
               /* bullet points are defined for the 0th character (if
                * it falls under a TXindent attr) and for any newlines
                * under a TXindent attr.  Note that a bullet is hidden
                * if the associated character is hidden.
                */
            return (pc ? !pc->hidden : TRUE);
        }
    }
    return TRUE;
}

/* : axtHasBullet - return whether the character within the
 * passed attribute segment is associated with a bullet
 */
int
axtHasBullet(txt, txAp, pc, pcIndex)
axtext *txt;                 /* in */
txAttr *txAp;                /* in: att seg containing pc */
perChar *pc;                 /* in: character within att seg */
int pcIndex;                 /* in: char's index within text */
{
    if (txAp->indent && txAp->indent->bullet && txAp->indent->bullet->type) {
        if (!txt->whole.nChars) {
            return (TRUE);
        }
        if (!pcIndex || (pc && CHAR_IS_NEWLINE((int) pc->charcode))) {
            
               /* bullet points are defined for the 0th character (if
                * it falls under a TXindent attr) and for any newlines
                * under a TXindent attr.  Note that a bullet is hidden
                * if the associated character is hidden.
                */
            return (pc ? !pc->hidden : TRUE);
        }
    }
    return (FALSE);
}

/* : axtBulletRect - return window-relative footprint of bullet in mils
 */
void
axtBulletRect(tx, txAp, charpos, bRect)
axtext *tx;               /* in */
txAttr *txAp;             /* in: att-seg this bullet falls under */
p32 *charpos;             /* in: position of character assoc. with bullet */
r32 *bRect;               /* out: return window-relative footprint */
{
    int bx, by, bwidth, bheight;
    p32 bpoint, orig_bpoint;
    axtBullet *bb;
    axform fieldXform;
    
    if (!txAp->indent || !txAp->indent->bullet) {
        bRect->origin.x = bRect->origin.y =
            bRect->width = bRect->height = 0;
        return;
    }

    /* get bullet bounding rectangle
     */
    bb = txAp->indent->bullet;
    bx = bb->bounder.origin.x;
    by = bb->bounder.origin.y;
    bwidth = bb->bounder.width;
    bheight = bb->bounder.height;
    
    if (!bwidth || !bheight) {
        bRect->origin.x = bRect->origin.y =
            bRect->width = bRect->height = 0;
        return;
    }
    
    if (!tx->whole.transform.unity) {
        axtGetFieldXform(tx, tx->whole.xZoom, tx->whole.xZoom, &fieldXform);
    }
    
    /* init bpoint as the origin for the bullet rect.  this point
     * is taken from the passed position, plus indent and other offsets.
     */
    if (tx->whole.xZoom == 1000 && tx->whole.yZoom == 1000) {
        orig_bpoint = bpoint = *charpos;
    }
    else {
        /* character position contains zoom, while the indent offsets don't.
         * remove zoom while accumulating offsets.
         */
        bpoint.x = 
            (int)((float) charpos->x * 1000.0 / (float) tx->whole.xZoom);
        bpoint.y = 
            (int)((float) charpos->y * 1000.0 / (float) tx->whole.yZoom);
        orig_bpoint = bpoint;
    }
    
    if (tx->whole.nChars) {
        /* the first offset is the indent distance.  move leftward by
         * the difference between firstLeft and left indents.
         */
        bpoint.x -= firstLeftOffset(tx, txAp) - txAp->indent->firstLeft;
    }
    
    /* offset upwards in unzoomed space, to vertically center
     * the bullet on a "small" estimate of the current font's
     * xheight above baseline. our cheap n' nasty xheight estimate
     * is 1/3 pointsize....
     */
    bpoint.y -= emToMils(330, txAp->fontAttr.ptsizeInt) + (bheight / 2);
    
    /* finally, there's a bullet offset defined on axtBullet
     */
    bpoint.x += bb->bulletOffset.x;
    bpoint.y += bb->bulletOffset.y;
    
    if (!tx->whole.transform.unity) {
        p32 ptmp;
        
        ptmp.x = bpoint.x - orig_bpoint.x;
        ptmp.y = bpoint.y - orig_bpoint.y;
        forwardSXRT(ptmp.x, ptmp.y, &ptmp.x, &ptmp.y, &fieldXform);
        bpoint.x = orig_bpoint.x + ptmp.x;
        bpoint.y = orig_bpoint.y + ptmp.y;
    }

    if (tx->whole.xZoom != 1000 || tx->whole.yZoom != 1000) {
        /* restore zoom
         */
        bpoint.x = (int) (((float) bpoint.x * tx->whole.xZoom) / 1000.0);
        bpoint.y = (int) (((float) bpoint.y * tx->whole.yZoom) / 1000.0);
        bwidth = (int) (((float) bwidth * tx->whole.xZoom) / 1000.0);
        bheight = (int) (((float) bheight * tx->whole.yZoom) / 1000.0);
    }
    
    bRect->origin.x = bpoint.x;
    bRect->origin.y = bpoint.y;
    bRect->width = bwidth;
    bRect->height = bheight;
    
    /* fudge outward
     */
    bRect->origin.x -= 7;
    bRect->origin.y -= 7;
    bRect->height += 14;
    bRect->width += 14;
}

#if !FOR_FILTER   /*  {  */

#if BUILD_ORIENT   /* {  */
XFontStruct *FontSetFontStruct();
FontSet FindFontSet();

XFontStruct *VerifyFont( ch, xf )
tx_char ch;
XFontStruct *xf;
{
    FontSet font_set;
    int char_type;
    int junk;
#if BUILD_ORIENT
    static XFontStruct  *old_xf = NULL;
    static XFontStruct  *new_xf;

    if (((xf->max_byte1 > 0) && IS_ORIENT_CHAR(ch)) ||
            ((xf->max_byte1 == 0) && !IS_WIDE_CHAR(ch)))
        return(xf);
 
    /* fast route */
    if (old_xf == xf)
        return(new_xf);
 
    old_xf = xf;
#endif

    /* we need to find the correct font in this font set, based on the */
    /* pareticular character value                              */
    if ( IS_WIDE_CHAR( ch ) )
       char_type = WIDE_FONT;
    else if ( IS_NARROW_KANA(  ch ) )
       char_type = NARROW_KANA_FONT;
    else
       char_type = NARROW_FONT;

    font_set = FindFontSet( xf->fid, &junk, NULL );
#if BUILD_ORIENT
    new_xf = FontSetFontStruct(font_set, char_type);
    return(new_xf);
#else
    xf = FontSetFontStruct( font_set, char_type );
    return( xf );
#endif
}







CheckCharsAgainstAttrs(txt)
axtext *txt;
{
    txAttr  *txAp;
    int     charCount = 0;
    int     local_pos;
    wchar   *charP = txt->textstr;
    register int seg_len;
    textAttrs new_attrs;
    register int ix;
    int is_wide_font, is_wide_char;
    int RefFontIsWide;  /* FALSE if English, TRUE if Wide font */
    int RefFontIsSans;  /* FALSE if serif, TRUE if sans-serif */
    char RefFontFamily[100];

    /* Ashisuto added to avoid unknown crash when building help_tag.jpn
     * 1997.05.31
    if (txt->atts == NULL || txt->atts[0].fontAttr.familyName == NULL)
        return(0);
     */


    memset((char *)&new_attrs, 0, sizeof(textAttrs));
    while (charCount < txt->whole.nChars) {
        txAp = attrSegAtPos(txt, charCount, &local_pos);
        /**==================================================================*/
        /* remember the attributes that start the string, and use this face */
        /* to create "intelligent" defaults, and use this  when the string */
        /* returns to the character type of the first attribute segment  */
        /**==================================================================*/
        strcpy(RefFontFamily, txAp->fontAttr.familyName );
        RefFontIsWide = IsWideFontFace(RefFontFamily);
        RefFontIsSans = IsSansFontFace(RefFontFamily);
        is_wide_font = RefFontIsWide;
        seg_len = txAp->nChars;
#if BUILD_KANJI
    /* Is NarrowKana font appears here, we thought it has been checked before.
     * Because the NarrowKana font never appearing until do printing, or do
     * text width calculation. Do not check this segment twice. */
    if (!IsNarrowKanaFontFace(RefFontFamily))
#endif
        for (ix = 0; ix < seg_len; ix++) {
            is_wide_char = IS_ORIENT_CHAR(txt->textstr[charCount+ix]);
            /**--------------------------------------------------------------**/
            /* now, if the attribute are wrong, make them right.  We set the */
            /* "new" attributes onto the remainder of the string, since chars */
            /* tend to come in bunches.                                      */
            /**--------------------------------------------------------------**/
            if (is_wide_char != is_wide_font) {
                if (is_wide_char == RefFontIsWide)
                    new_attrs.fontFamily = RefFontFamily;
                else {
                    if (is_wide_char)
                        if (RefFontIsSans)
                            new_attrs.fontFamily = sansserif_orient_face();
                        else
                            new_attrs.fontFamily = serif_orient_face();
                    else
                        if (RefFontIsSans)
                            new_attrs.fontFamily = NXLT(axDefaultSansSerifFont);
                        else
                            new_attrs.fontFamily = NXLT(axDefaultSerifFont);
                }
                setTextAttrs(txt, &new_attrs, TXfontFamily, charCount + ix, 
                        seg_len - ix);
                is_wide_font = is_wide_char;
            }
        }
        charCount += seg_len;
    }
}




/* Ashisuto added for the new specification of Japanese font.
 * (KANJI and KATAKANA) We pretent until last to keep the Narrow Kana font as
 * a wide font, even in the function CheckCharsAgainstAttrs(). So that the
 * Narrow Kana be completely hidden from user, and also we got the simplicity
 * of our processing. But right before output texts, we had to modify the
 * font face at the end point. This is a special issue for Japanese.
 * 95.11.06
 */
#if BUILD_KANJI    /*  {  */
CheckNarrowKanaAgainstAttrs(txt)
axtext *txt;
{
    txAttr  *txAp;
    int     charCount = 0;
    int     local_pos;
    register int seg_len;
    textAttrs new_attrs;
    register int ix;
    int is_narrow_kana_font, is_narrow_kana_char;
    int RefFontIsNarrowKana;    /* FALSE if English, TRUE if Wide font */
    char WideFontFamily[100];
    char NarrowKanaFontFamily[100];

    memset((char *)&new_attrs, 0, sizeof(textAttrs));
    while (charCount < txt->whole.nChars) {
        txAp = attrSegAtPos(txt, charCount, &local_pos);
        seg_len = txAp->nChars;
        if (IsWideFontFace(txAp->fontAttr.familyName)) {
            strcpy(WideFontFamily, txAp->fontAttr.familyName);
            strcpy(NarrowKanaFontFamily, NXLT("Narrow"));
            strcat(NarrowKanaFontFamily, WideFontFamily);
            is_narrow_kana_font = FALSE;
            for (ix = 0; ix < seg_len; ix++) {
                is_narrow_kana_char = IS_NARROW_KANA(txt->textstr[charCount+ix]);
                if (is_narrow_kana_char != is_narrow_kana_font) {
                    if (is_narrow_kana_char)
                        new_attrs.fontFamily = NarrowKanaFontFamily;
                    else
                        new_attrs.fontFamily = WideFontFamily;
                    setTextAttrs(txt, &new_attrs, TXfontFamily,
                            charCount + ix, seg_len - ix);
                    is_narrow_kana_font = is_narrow_kana_char;
                }
            }
        }
        charCount += seg_len;
    }
}
#endif    /* }  BUILD_KANJI  */

#endif   /*  }  BUILD_ORIENT  */

#endif   /*  }  !FOR_FILTER  */


int
axtAttAtPos(txt, att, pos)
axtext *txt;
textAttrs *att;
int pos;
{
    unsigned int flags = (unsigned int)TXAllAtts;
    txAttr *txAp;
    int tmp;

    txAp = attrSegAtPos(txt, pos, &tmp);
    copyOutTextAttr(att, txAp, flags, txt->pid);
    return AOK;
}

txAttr *
axtAttrAtPos(txt, pos)
axtext *txt;
int pos;
{
    int tmp;
    return( attrSegAtPos(txt, pos, &tmp));
}

/*
 * : easy_text_fill
 * attr_type's pattern types:
 *    0 - built in fillpattern (of which #4 is solid fg)
 *    1 - user defined
 *    2 - linear gradation
 *  3,4 - circular / rectangular gradation
 *    5 - screen (of which #1000 is 100% foreground, 0% bg)
 */  
int
easy_text_fill(gContext, patt)
void *gContext;
attr_type *patt;
{
#if !FOR_FILTER
    /* NOTE that line patterns aren't currently supported, so
       we won't bother testing them in here.. */
#if ALLOW_JAVA_GE
	{
		graphics_main_type *gfx;
		gfx = (graphics_main_type *)gContext;
		if ( gfx && gfx->state.java_display )
			return(FALSE);
	}
#endif
    if ( patt->ffg.see_thru == FALSE )
    {
        if (patt->f_type == 0 && patt->f_pat == 4)
            return(TRUE);  /* fill pat is solid foreground */
        if (patt->f_type == 5 && patt->f_pat == 1000)
            return(TRUE);  /* fill pat is solid foreground */
        if (patt->f_pat == -1)
            return(TRUE);  /* no fill; NOTE this is treated as solid */
    }
#endif
    return(FALSE);
}

void
axtDrawText(txt, drw, x, y, hyt, wid, mask, attr)
axtext *txt;
Drawable drw;
int x, y, hyt, wid;
Drawable mask;
attr_type *attr;
{
#if FOR_FILTER /* { */
	XImage *img;
	if ( wid > 0 && hyt > 0 )
	{
		img = XGetImage(Dpy, mask, 0, 0, wid, hyt, 1, ZPixmap);
		if ( img )
		{
			digitizeFillText(txt->whole.cvt, img, x, y, hyt, wid, attr);
#if AXMS 
			/************** Windows HACK Alert!  *******************\
			* the XImage struct carries a copy of the 'map' Pixmap, *
			* so we can't delete it twice. Make XImage copy null.   *
			\************** Windows HACK Alert!  *******************/
			img->bitmap_handle = NULL;
#endif
			XDestroyImage(img);
		}
	}
#else /* } FOR_FILTER { */
	int f_type, f_pat, fill_flag, x_ix, y_ix, ix, sz, do_draw, mode;
	unsigned char *bits, bitmask;
	graphics_main_type *gfx;
	XImage *img;

	fill_flag = attr->fill_flag;
	f_type = attr->f_type;
	f_pat  = attr->f_pat;
	if ( fill_flag == FALSE )
	{
		if ( f_type == 0 && f_pat == -1 )
		{
			attr->fill_flag = TRUE;
			attr->f_type = 5;
			attr->f_pat = 1000;
		}
	}
	if ( check_for_fill(attr) )
	{
#if ALLOW_JAVA_GE
		gfx = (graphics_main_type *)txt->gContext;
		if ( gfx == (graphics_main_type *)0 || gfx->state.java_display == FALSE )
			do_draw = TRUE;
		else
		{
			img = XGetImage(Dpy, mask, 0, 0, wid, hyt, 1, ZPixmap);
			if ( img )
			{
				sz = ((wid+7)/8)*hyt;
				bits = (unsigned char *)txloc(txt->pid, sz);
				if ( bits )
				{
					memset((char *)bits, 0, sz);
					bitmask = 0x80;
					ix = 0;
					for ( y_ix=0; y_ix<hyt; y_ix++ )
					{
						for ( x_ix=0; x_ix<wid; x_ix++ )
						{
							if ( XGetPixel(img, x_ix, y_ix) ) /* slow read */
								bits[ix] |= bitmask;
							if ( bitmask != 0x01 )
								bitmask = bitmask >> 1;
							else
							{
								bitmask = 0x80;
								ix++;
							}
						}
						if ( bitmask != 0x80 )
						{
							bitmask = 0x80;
							ix++;
						}
					}
					mode = attr->mode;
					if ( txt->editState.inEdit )
						attr->mode = TMS_X_D;
					gJavaMaskBits(gfx, x,y, wid, hyt, bits, attr);
					attr->mode = mode;
					free(bits);
				}
#if AXMS 
				/************** Windows HACK Alert!  *******************\
				* the XImage struct carries a copy of the 'map' Pixmap, *
				* so we can't delete it twice. Make XImage copy null.   *
				\************** Windows HACK Alert!  *******************/
				img->bitmap_handle = NULL;
#endif
				XDestroyImage(img);
			}
			do_draw = TRUE;
		}
#else
		do_draw = TRUE;
#endif

#if AXMS
		if ( do_draw && Running_Windows95 && GRADIENT_FILL(attr->f_type) )
		{
			AxWinDrawGradientImage(drw, x, y, wid, hyt, attr, mask, FALSE);
			do_draw = FALSE;
		}
#endif

		/* render the glyph */
		if ( do_draw )
		{
			/* special modes and percents draw as bitmaps (display only) */
			if ( attr->mode != TMS )
			{
				attr->f_type = 0;
				attr->f_pat  = 4;
			}
			else if ( attr->f_type == 5 )
			{
				attr->f_type = 0;
				attr->f_pat = get_percent_ix(attr->f_pat);
			}

			/* fill thru the mask */
			XSetFunction(Dpy, Gc1, GXcopy);
			XSetClipOrigin(Dpy, Gc1, x, y);
			XSetClipMask(Dpy, Gc1, mask);
			XSetTSOrigin(Dpy, Gc1, x, y);
			doAttrFill(drw, x, y, wid, hyt, attr, FALSE);
		}
	}

	/* restore */
	attr->fill_flag = fill_flag;
	attr->f_type = f_type;
	attr->f_pat  = f_pat;
#endif /* } ! FOR FILTER */
}

/* : axtCombineIntercepts - intersect the x coordinates of
 *                          the passed intercept data to produce
 *                          an intercepts list which fits inside
 *                          both.  Caller must free returned list.
 */
static intercept *    /* return new list - caller must free */
axtCombineIntercepts(baseline, in2, pid)
intercept *baseline, *in2;   /* in: intercept data to intersect */
int pid;
{
    int ix, jx, kx, numNew;
    pointPair *ip1, *ip2;
    intercept *result;

        /* first count how many intercept pairs will
           result from the intersection */
    numNew = 0;
    for (ix = 0; ix < baseline->paircount; ix++) {
        for (jx = 0; jx < in2->paircount; jx++) {
            ip1 = &baseline->pairs[ix];
            ip2 = &in2->pairs[jx];
            if (ip1->p1.x > ip2->p0.x &&
                ip1->p0.x < ip2->p1.x)
                numNew++;
        }
    }

    if (!numNew)
        return ((intercept *) NULL);

    result = (intercept *)txloc(pid, sizeof(intercept));
    result->paircount = numNew;
    result->pairs = (pointPair *)txloc(pid, numNew * sizeof(pointPair));

    kx = 0;
    for (ix = 0; ix < baseline->paircount; ix++) {
        for (jx = 0; jx < in2->paircount; jx++) {
            ip1 = &baseline->pairs[ix];
            ip2 = &in2->pairs[jx];
            if (ip1->p1.x > ip2->p0.x &&
                ip1->p0.x < ip2->p1.x) {

                    /* result's left edge is the rightmost of
                       the two; similarly, its right edge is the
                       leftmost of the two.  Note that the ycoords
                       are assigned from ip1, which is assumed to
                       be looking at a baseline ycoord
                     */
                result->pairs[kx].p0.x = ip1->p0.x > ip2->p0.x ?
                    ip1->p0.x : ip2->p0.x;
                result->pairs[kx].p1.x = ip1->p1.x > ip2->p1.x ?
                    ip2->p1.x : ip1->p1.x;
                result->pairs[kx].p0.y = ip1->p0.y;
                result->pairs[kx].p1.y = ip1->p1.y;
                kx++;
            }
        }
    }
    return (result);
}

static void
axtFreeIntercepts(edges)
intercept *edges;
{
    if (edges) {
        if (edges->pairs)
            txfree((char *) edges->pairs);
        txfree((char *) edges);
    }
}

void
axtFreeIndent(indp)
axtIndent *indp;
{
    if (indp) {
        if (indp->bullet) {
            axtFreeBullet(indp->bullet);
        }
        txfree(indp);
    }
}

void
axtCopyIndent(in, out, pid)
axtIndent *in;
axtIndent *out;
int pid;
{
    if (in && out) {
        if (out->bullet) {
            axtFreeBullet(out->bullet);
            out->bullet = (axtBullet *) NULL;
        }
        
        *out = *in;
        
        if (in->bullet) {
            out->bullet = axtNewDefaultBullet(pid);
            axtCopyBullet(in->bullet, out->bullet, pid);
        }
    }
}

axtIndent *
axtNewDefaultIndent(pid)
int pid;
{
    axtIndent *newi = (axtIndent *)txloc(pid, sizeof(axtIndent));
    memset((char *) newi, 0, sizeof(axtIndent));
    return (newi);
}

void
axtCopyBullet(from, to, pid)
axtBullet *from, *to;
int pid;
{
    if (to->gAttr) {
        txfree(to->gAttr);
    }
    
    *to = *from;

    if (from->gAttr) {
        to->gAttr = (void *) txloc(pid, sizeof(thing_attributes));
        memcpy((void *) to->gAttr, (void *) from->gAttr,
               sizeof(thing_attributes));
    }
}

void
axtFreeBullet(inb)
axtBullet *inb;
{
    if (inb) {
        if (inb->gAttr) {
            txfree(inb->gAttr);
        }
        txfree(inb);
    }
}

axtBullet *
axtNewDefaultBullet(pid)
int pid;
{
    axtBullet *newb = (axtBullet *)txloc(pid, sizeof(axtBullet));
    memset((char *) newb, 0, sizeof(axtBullet));
    return (newb);
}

void
axtFreeDropShadow(shad)
axtDropShadow *shad;
{
    if (shad) {
        if (shad->attr) {
            axtFreeFillPattern(shad->attr);
        }
        txfree(shad);
    }
}

void
axtCopyDropShadow(in, out, pid)
axtDropShadow *in;
axtDropShadow *out;
int pid;
{
    if (in && out) {
        if (out->attr) {
            axtFreeFillPattern(out->attr);
        }
        
        *out = *in;
        
        if (in->attr) {
            axtDuplicateFillPattern(in->attr, &out->attr, pid);
        }
    }
}

/* : axtSnapshotText - take a copy of the current, in-edit text
 * instance in preparation for an edit of some sort
 */
static void
axtSnapshotText(current, editPos)
axtext *current;                  /* in */
int editPos;                      /* in: snap from here downstream */
{
        /* state that old and current will be the same from editPos upstream
         */
    current->editState.editPos = axtEditPos(current, editPos);
    
        /* copy the downstream portion of current onto txt->old
         */
    copyAxtext(&current->old, current);
}

static void
axtSnapshotFieldBounder(current, newBounder)
axtext *current;
r32 *newBounder;           /* out, in screenpix */
{   
    char *err;
    r32 trect;

        /* current has changed from editPos on down somehow; get the
         * bounder of this new material
         */
    if (err = getOverallBounder(current, current->editState.editPos,
                                -1, &trect)) {
        freeErrString(err);
    }
    
        /* add the text position to the bounder
         */
    if (current->old && current->editState.editPos) {
            /* use old's position (which contains any offset to editPos)
             */
        trect.origin.x += current->old->whole.pos.x;
        trect.origin.y += current->old->whole.pos.y;
        if (current->old->items) {
            trect.origin.x += current->old->items[0].offset.x +
                current->old->items[0].kern.x;
            trect.origin.y += current->old->items[0].offset.y +
                current->old->items[0].kern.y;
        }
    }
    else {
        trect.origin.x += current->whole.pos.x;
        trect.origin.y += current->whole.pos.y;
    }

        /* store fieldBounder in screenpix;
         */
    trect.origin.x = milsToPix(trect.origin.x, ScreenRes);
    trect.origin.y = milsToPix(trect.origin.y, ScreenRes);
    trect.width = milsToPix(trect.width, ScreenRes);
    trect.height = milsToPix(trect.height, ScreenRes);
    if (trect.width) {
        trect.origin.x--;     /* fudge outward */
        trect.origin.y--;
        trect.width += 2;
        trect.height += 2;
    }
    
        /* add it into current's field bounder
         */
    if (current->editState.fieldBounder.width) {
        unionRectangle(&current->editState.fieldBounder, &trect, newBounder);
    }
    else {
        *newBounder = trect;
    }
}

/* : extendFieldBounder - enlarge the field
 * bounder of text under edit, since the text in its
 * current state may have a larger footprint than last state.
 * The axtext member ->old is expected to be valid.
 */
static void
extendFieldBounder(txt, w, newField)
axtext *txt;            /* in: text under edit, w/ old fieldBounder val */
long w;                 /* in: widget handle from which to get drawable */
r32 *newField;          /* in: this is txt's new fieldBounder to extend to */
{
    int last_protruded, top_protruded, fx, fy, twid, thyt;
    r32 newf, *oldf;
    Drawable drw;
    char *err;
    
    if (!txt->whole.nChars)
        return;

    /* get drawing resources */
    drw = AxGetDrawableFromWidget((AXWidget)w, &fx, &fy);
    setLiftColor(txt, axLiftGC);

    /* fill the background */
    if ( txt->editState.fieldBounder.width )
    {
        /* get the old */
        oldf = &txt->editState.fieldBounder;

        if ( newField == (r32 *)0 ) /* no new field - fill the old one */
        {
#if ALLOW_JAVA_GE && !FOR_FILTER
			{
				TmRect32 r;
				int fg;

				r.x_position = oldf->origin.x;
				r.y_position = oldf->origin.y;
				r.x_size = oldf->width;
				r.y_size = oldf->height;
				fg = txt->editState.javaPixel;
				tx_java_fillrect(txt->gContext, &r, 5, fg, fg, 0, 0, &r);
			}
#endif
            XFillRectangle(Dpy, drw, axLiftGC,
                           oldf->origin.x,
                           oldf->origin.y,
                           oldf->width,
                           oldf->height);
        }

        else /* compare with the new bounder */
        {
            /* draw top part */
            unionRectangle(&txt->editState.fieldBounder, newField, &newf);
            if ( newf.origin.y >= oldf->origin.y )
                last_protruded = top_protruded = FALSE;
            else
            {
                last_protruded = top_protruded = TRUE;
                if ( newf.width )
                {
#if ALLOW_JAVA_GE && !FOR_FILTER
					{
						TmRect32 r;
						int fg;

						r.x_position = newf.origin.x;
						r.y_position = newf.origin.y;
						r.x_size = newf.width;
						r.y_size = oldf->origin.y - newf.origin.y;
						fg = txt->editState.javaPixel;
						tx_java_fillrect(txt->gContext, &r, 5, fg, fg, 0,0, &r);
					}
#endif
                    XFillRectangle(Dpy, drw, axLiftGC,
                                   newf.origin.x,
                                   newf.origin.y,
                                   newf.width,
                                   oldf->origin.y - newf.origin.y);
                }
            }

            /* draw right part */
            twid = newf.origin.x + newf.width - oldf->origin.x - oldf->width;
            if ( twid <= 0 )
                last_protruded = FALSE;
            else
            {
                fx = oldf->origin.x + oldf->width;
                fy = last_protruded ? oldf->origin.y : newf.origin.y;
                thyt = newf.origin.y + newf.height - fy;
                if ( thyt > 0 )
				{
#if ALLOW_JAVA_GE && !FOR_FILTER
					{
						TmRect32 r;

						int fg;
						r.x_position = fx;
						r.y_position = fy;
						r.x_size = twid;
						r.y_size = thyt;
						fg = txt->editState.javaPixel;
						tx_java_fillrect(txt->gContext, &r, 5, fg, fg, 0,0, &r);
					}
#endif
                    XFillRectangle(Dpy, drw, axLiftGC, fx,fy, twid,thyt);
				}
                last_protruded = TRUE;
            }
    
            /* draw bottom part */
            thyt = newf.origin.y + newf.height - oldf->origin.y - oldf->height;
            if ( thyt <= 0 )
                last_protruded = FALSE;
            else
            {
                fy = oldf->origin.y + oldf->height;
                fx = newf.origin.x;
                if ( last_protruded )
                    twid = oldf->origin.x + oldf->width - newf.origin.x;
                else
                    twid = newf.origin.x + newf.width - newf.origin.x;
                if ( twid > 0 )
				{
#if ALLOW_JAVA_GE && !FOR_FILTER
					{
						TmRect32 r;
						int fg;

						r.x_position = fx;
						r.y_position = fy;
						r.x_size = twid;
						r.y_size = thyt;
						fg = txt->editState.javaPixel;
						tx_java_fillrect(txt->gContext,&r,5, fg, fg, 0, 0, &r);
					}
#endif
                    XFillRectangle(Dpy, drw, axLiftGC, fx,fy, twid,thyt);
				}
                last_protruded = TRUE;
            }
    
            /* draw the left part */
            if ( newf.origin.x < oldf->origin.x )
            {
                fx = newf.origin.x;
                twid = oldf->origin.x - newf.origin.x;
                if ( top_protruded )
                    fy = oldf->origin.y;
                else
                    fy = newf.origin.y;
                if ( last_protruded )
                    thyt = oldf->origin.y + oldf->height - fy;
                else
                    thyt = newf.origin.y + newf.height - fy;
                if ( twid > 0 && thyt > 0 )
				{
#if ALLOW_JAVA_GE && !FOR_FILTER
					{
						TmRect32 r;
						int fg;

						r.x_position = fx;
						r.y_position = fy;
						r.x_size = twid;
						r.y_size = thyt;
						fg = txt->editState.javaPixel;
						tx_java_fillrect(txt->gContext, &r,5, fg, fg, 0, 0, &r);
					}
#endif
                    XFillRectangle(Dpy, drw, axLiftGC, fx,fy, twid,thyt);
				}
            }
        }
    }
    else /* nothing to extend from; just fill new field */
    {
        err = getOverallBounder(txt, 0, -1, &newf);
        if ( err )
            freeErrString(err);
        else
        {
            newf.origin.x += txt->whole.pos.x;
            newf.origin.y += txt->whole.pos.y;
            newf.origin.x = milsToPix(newf.origin.x, ScreenRes);
            newf.origin.y = milsToPix(newf.origin.y, ScreenRes);
            newf.width = milsToPix(newf.width, ScreenRes);
            newf.height = milsToPix(newf.height, ScreenRes);
            if ( newf.width )
            {
                newf.origin.x--;         /* fudge outward */
                newf.origin.y--;
                newf.width += 2;
                newf.height += 2;
#if ALLOW_JAVA_GE && !FOR_FILTER
				{
					TmRect32 r;
					int fg;

					r.x_position = newf.origin.x;
					r.y_position = newf.origin.y;
					r.x_size = newf.width;
					r.y_size = newf.height;
					fg = txt->editState.javaPixel;
					tx_java_fillrect(txt->gContext, &r, 5, fg, fg, 0, 0, &r);
				}
#endif
                XFillRectangle(Dpy, drw, axLiftGC, newf.origin.x,
                                                   newf.origin.y,
                                                   newf.width,
                                                   newf.height);
            }
        }
    }
}

void
axtGetFieldBounder(txt, bbox)
axtext *txt;     /* in */
r32 *bbox;       /* out, in mils */
{
    int ixtmp, iytmp;
    
    if (txt->editState.fieldBounder.width) {
#if FOR_FILTER
        cvtConvPixToMils(txt->whole.cvt,
		                 txt->editState.fieldBounder.origin.x,
                         txt->editState.fieldBounder.origin.y,
                         &bbox->origin.x, &bbox->origin.y);
        cvtConvPixToMils(txt->whole.cvt,
		                 txt->editState.fieldBounder.width,
                         txt->editState.fieldBounder.height,
                         &ixtmp, &iytmp);
#else
        convPixToMils(txt->editState.fieldBounder.origin.x,
                      txt->editState.fieldBounder.origin.y,
                      &bbox->origin.x, &bbox->origin.y);
        convPixToMils(txt->editState.fieldBounder.width,
                      txt->editState.fieldBounder.height,
                      &ixtmp, &iytmp);
#endif
        bbox->width = ixtmp;
        bbox->height = iytmp;
    }
    else {
        bbox->origin.x = bbox->origin.y = bbox->width = bbox->height = 0;
    }
}

/*
 * getLetterform, copyLetterform, freeLetterform: these
 * functions support the LETTERFORM cache item (scaled
 * letterform raster) which is used in drawLetterforms
 */
#define MAX_LETTERFORM_SIZE 50000    /* mils: 50" too big for X?? */

/* : getLetterform - create a scaled letterform image
 */
int
getLetterform(name, dp, sp, ap)
char *name;
char **dp; /* to be filled out as XImage* */
int *sp;
char *ap; /* arrives as pointer to chSeg struct */
{
    int tx, ty;
    unsigned int wid, ht;
    perChar *pc;
    chSeg *cS;
    XImage *charImage = (XImage *)NULL;
    XImage *transformedImage = (XImage *)NULL;
    Pixmap pix;
    XChar2b charcode;
    extern Display *Dpy;
    extern GC PixGC, PixClearGC;
    int milsToPix();

    cS = (chSeg *) ap;
    pc = cS->pc;

        /* filter out some no-draw cases */
    if (pc->sbox.width == 0 || pc->sbox.height == 0) {
            /* a newline is an example of such a case */
        *sp = 0;
        *dp = (char *) NULL;
        return TRUE;
    }
    if (pc->width > MAX_LETTERFORM_SIZE ||
        pc->height > MAX_LETTERFORM_SIZE) {
            /* don't abuse the poor x server */
        *sp = 0;
        *dp = (char *) NULL;
        return TRUE;
    }
    wid = (unsigned int) milsToPix(pc->width, ScreenRes);
    ht = (unsigned int) milsToPix(pc->height, ScreenRes);
    if (!wid || !ht) {
            /* the scaled letterform would be zero-area */
        *sp = 0;
        *dp = (char *) NULL;
        return TRUE;
    }

        /* get a pixmap to draw on */
    wid = (unsigned int) pc->sbox.width;
    ht = (unsigned int) pc->sbox.height;
    pix = XCreatePixmap(Dpy, AxTopWinNeverMapped, wid, ht, (unsigned int)1);
    XFillRectangle(Dpy, pix, PixClearGC, 0, 0, wid, ht);
        /* shoot the char */
#if BUILD_ORIENT && (SUNIS || LINUX)
/* Ashisuto added to be compatible with "bad cast!!" in wfntwrap.c so that
 * to get the same processing in xwrapr.c FOR SUNIS and Linux. Though
 * this change may also valid for other UNIX platforms.    1998.02.25
 */
 /* added the build orient condition since text was not displaying at all
  * at higher zoom levels - please see the comments in func: axDrawString16
  * please check with bobz before changing anything - Dec 12 96
  */
    memcpy(&charcode, &pc->charcode, 2);
#else
    charcode.byte1 = (unsigned char )((pc->charcode >> 8) & 0xff);
    charcode.byte2 = (unsigned char )(pc->charcode & 0xff);
#endif
    tx = -pc->sbox.origin.x;
    ty = -pc->sbox.origin.y;
#if BUILD_ORIENT
    axDrawString16(Dpy, pix, PixGC, NULL, tx, ty, &charcode, 1);
#else
    XDrawString16(Dpy, pix, PixGC, tx, ty, &charcode, 1);
#endif

        /* get the resultant image */
    charImage = XGetImage(Dpy, pix, 0, 0, wid, ht, 1, XYPixmap);
#if AX_X11
    /* don't free on Windows because the pix was added to the XImage *
     * and so will be free'd in the subsequent call to XDestroyImage.*/
    XFreePixmap(Dpy, pix);
#endif

    if (charImage) {
            /* apply transform */
        transformedImage = scaleLetterform(charImage, 0, 0, wid, ht, pc);
        XDestroyImage(charImage);
    }
    if (transformedImage) {
        transformedImage->format = XYBitmap;
            /* there could be extra stuff on ximage->obdata too,
             * but no way of knowing..  just use image size */
        *sp = sizeof(XImage) + (transformedImage->height *
                transformedImage->bytes_per_line);
    }
    else {
        *sp = 0;
    }
    if (AxtLetterformPixmaps) {
        if (transformedImage) {
                /* the item value is a pixmap depicting the
                   transformed image */          
            pix = XCreatePixmap(Dpy, AxTopWinNeverMapped, 
                                transformedImage->width,
                                transformedImage->height, 
                                (unsigned int)1);
            XFillRectangle(Dpy, pix, PixClearGC, 0, 0,
                           transformedImage->width,
                           transformedImage->height);
            XPutImage(Dpy, pix, PixGC, transformedImage,
                      (unsigned int) 0, (unsigned int) 0,
                      (unsigned int) 0, (unsigned int) 0,
                      (unsigned int) transformedImage->width,
                      (unsigned int) transformedImage->height);
            if (transformedImage->data) {
                    /* we allocated this image data; free it now
                       and null the pointer, or else XDestroy will
                       try to free it from Xlib's heap */
                txfree((char *) transformedImage->data);
                transformedImage->data = (char *) NULL;
            }
            XDestroyImage(transformedImage);
            *dp = (char *) pix;
        }
        else {
            *dp = (char *) NULL;
        }
    }
    else {
            /* the item value is the transformed image */
        *dp = (char *)transformedImage;
    }

    return TRUE;
}

int
copyLetterform(name, onto, sizep, from)
char *name;
char **onto;
int *sizep;
char *from;
{
    return FALSE; /* this copy op not defined yet */
}

int
freeLetterform(dp)
char **dp;
{
    XImage *charImage;
    Pixmap *charPix;
    
    if (*dp) {
        if (AxtLetterformPixmaps) {
            charPix = (Pixmap *) *dp;
            XFreePixmap(Dpy, (Pixmap)charPix);
        }
        else {
            charImage = (XImage *) *dp;
            if (charImage->data) {
                txfree((char *) charImage->data);
                charImage->data = (char *) NULL;
            }
            XDestroyImage(charImage);
        }
    }
    return( 0 );
}

/******************************************************************************
 * function: scaleLetterform -- adapted from bitblt.c's xyRotateImage
 */
static XImage *
scaleLetterform(xImage, srcXoff, srcYoff, srcWid, srcHyt, pc)
XImage *xImage;
int srcXoff, srcYoff;
int srcWid, srcHyt;
perChar *pc;
{
    int imByteWidth;
    int imHyt, imWid;
    int i;
    unsigned char *imageBuf;
    unsigned char *maskPtr;
    int srcX, dstX, srcY, dstY, tY;
    unsigned char *srcLine, *dstLine;
    XImage *newImage;
    int milsToPix();

    imWid = milsToPix(pc->width, ScreenRes);
    imHyt = milsToPix(pc->height, ScreenRes);
        /* allocate a buffer to hold the new image */
    imByteWidth = BYTE_WIDTH(imWid);
    i = imByteWidth * imHyt;
    if((imageBuf = (unsigned char *)malloc(i)) == NULL)
    {
        return( (XImage *)NULL ); /* failure */
    }
    memset((char *)imageBuf, 0, i);
    
    /* set up to read bits right to left or left to right (respectively)  */
    maskPtr = xImage->bitmap_bit_order==LSBFirst ? lsbfBitMask:msbfBitMask;

    /* scan thru the bounding box of the target image, transforming index *\
    \* values and looking up the transformed value in the source map.     */
    for(dstY=0; dstY<imHyt; dstY++)
    {
        for(dstX=0; dstX<imWid; dstX++)
        {
            reverseSXRT(dstX, dstY, &srcX, &srcY, pc->ctm);
            tY = srcY + srcYoff;
            if(srcX >= 0 && srcX < srcWid && tY >= 0 && tY < srcHyt)
            {
                srcLine = (unsigned char *)(xImage->data + (srcY+srcYoff) *
                            xImage->bytes_per_line);
                dstLine = imageBuf + dstY * imByteWidth;
                if(srcLine[(srcX+srcXoff)>>3] & maskPtr[(srcX+srcXoff)&7])
                    dstLine[dstX>>3] |= maskPtr[dstX&7];
            }
        }
    }

    /* plug the new image into an XImage struct */
    newImage = (XImage *)NULL;
    newImage = XCreateImage(Dpy, DefaultVisual(Dpy, DefaultScreen(Dpy)),
                    1, XYPixmap, 0, (char *)imageBuf, imWid, imHyt,
                    16, imByteWidth);
    return( (XImage *)newImage );
}

/* : firstLeftOffset - return the offset value for use
 * on the 1st line of text falling under a particular TXindent attr
 */
static int                     /* return offset */
firstLeftOffset(tx, txAp)
axtext *tx;                    /* in */
txAttr *txAp;                  /* in: att-seg this indent falls under */
{
    int leftOffset, bRight, itmp;

    if ( !txAp->indent->bullet || txAp->indent->bullet->type == 0 ||
        (!txAp->indent->bullet->bounder.width ||
         !txAp->indent->bullet->bounder.height)) {
            
        /* no bullet; use firstLeft directly
         */
        leftOffset = txAp->indent->firstLeft;
    }
    else {
        /* there's a bullet; start text at left rather
         * than firstLeft
         */
        leftOffset = txAp->indent->left;
        
        /* find right edge of bullet (which starts at firstLeft)
         */
        bRight = txAp->indent->firstLeft + txAp->indent->bullet->bounder.width;
        if ((itmp = bRight - txAp->indent->left) > 0) {
            leftOffset += itmp;    /* increase leftOffset to avoid bullet */
        }
    }
    return (leftOffset);
}

/* : axtGetIndentLineVals - report the linespacing and offset values for
 * the attr segment occurring at ix.
 */
static void
axtGetIndentLineVals(tx, curr_para_attr, xform, ix, space, right_off, left_off)
axtext *tx;             /* in */
txAttr *curr_para_attr; /* in */
axform *xform;          /* in */
int ix;                 /* in */

int *space;             /* out */
int *right_off;         /* out */
int *left_off;          /* out */
/* int *paraAlignment;   out: soon */
{
    txAttr *prev_para_attr;
    int i, jnk;

    /* normal text (ie. not at the beginning of a paragraph) */
    if ( NEWPARA(tx->items, ix) == FALSE )
    {
        /* get the left indent */
        *left_off = curr_para_attr->indent->left;

        /* get the baseline offset */
        *space = curr_para_attr->indent->midLineHeight;
    }
    else /* special treatment for beginning of a paragraph */
    {
        /* get the left indent */
        *left_off = firstLeftOffset(tx, curr_para_attr);

        /* get the baseline offset */
        if ( ix == 0 ) /* no previous paragraph */
            *space = curr_para_attr->indent->topLineHeight;
        else
        {
            /* find the previous paragraph attr */
            prev_para_attr = (txAttr *)0;
            for ( i=ix-1; i>=0; i-- )
            {
                if ( NEWPARA(tx->items, i) )
                {
                    prev_para_attr = attrSegAtPos(tx, i, &jnk);
                    break;
                }
            }
            if ( !prev_para_attr )
                prev_para_attr = attrSegAtPos(tx, 0, &jnk);

            /* line height plus previous para bottom offset */
            *space = curr_para_attr->indent->topLineHeight +
                     prev_para_attr->indent->bottomLineOffset;
        }
    }

    /* get the right indent */
    *right_off = curr_para_attr->indent->right;
}

static void
axtGetFieldXform(tx, xZoom, yZoom, fieldXform)
axtext *tx;              /* in */
int xZoom;               /* in */
int yZoom;               /* in */
axform *fieldXform;      /* out */
{
    axformTerms unityLocal;

    /* get a field-only transform
     */
    unityLocal.unity = TRUE;
    unityLocal.angle = 0;
    unityLocal.xscale = 10000;   /* i.e., 1.0 */
    unityLocal.yscale = 10000;
    unityLocal.xshear = 0;
    axtWriteTargetCtm(xZoom, yZoom, &unityLocal,
                      &tx->whole.transform, fieldXform);
    fieldXform->known = TRUE;
}

/* : axtTrimText - sub called by qDrawText, drawLetterforms.
 * reduce passed starting index and run length to trim out
 * non-graphic (CHAR_IS_DIMENSIONLESS) characters.  Work on
 * the ends of the range only; don't worry about the middle.
 */
static void
axtTrimText(tx, startix, len)
axtext *tx;                         /* in */
int *startix;                       /* in/out */
int *len;                           /* in/out */
{
    int ix, tix;
    int endix;

    tix = *startix;
    endix = *startix + *len - 1;
    
    /* trim upstream side
     */
    for (ix = tix; ix <= endix; ix++) {
        if (CHAR_IS_DIMENSIONLESS(tx->items[ix].charcode)) {
            (*startix)++;
            (*len)--;
        }
        else {
            break;
        }
    }
    
    if (*startix > endix) {
        *startix = endix;
        if (*len < 0)
            *len = 0;
        return;
    }
    
    /* trim downstream side
     */
    tix = *startix;
    for (ix = endix; ix >= tix; ix--) {
        if (CHAR_IS_DIMENSIONLESS(tx->items[ix].charcode)) {
            (*len)--;
        }
        else {
            break;
        }
    }
    
    if (*len < 0)
        *len = 0;
}

int
axtGetOffsets(tx, leftIndex, rightOffset, leftOffset)
axtext *tx;
int leftIndex;
int *rightOffset;
int *leftOffset;
{
	txAttr *txAp;
	axform fieldXform;
	int lpos, tSpace;
	/* init the intercept-offset vars */

	txAp = attrSegAtPos(tx, leftIndex, &lpos);
	if (txAp->indent) 
	{
		memset((void *) &fieldXform, 0, sizeof(axform));
		axtGetIndentLineVals(tx, txAp, &fieldXform,
							 leftIndex, &tSpace,
							 rightOffset, leftOffset);
	}
	return(*leftOffset - *rightOffset);
}


dbx_fixme_place()
{
  /* one place to put a breakpoint for all things to be fixed - jon*/
  int i;
  i = i + 1;
}

#if 0
static int
axtSimpleTextGeom(txt)
axtext *txt;
{
    int charcount, lineheight;
    txAttr *txAp;

    if (!txt || !txt->measured)
        return FALSE;

        /* a "simple" text instance is one which has
           no text larger than the lineheight, and
           no rotated text */

    if (txt->whole.bounder.numPaths) {
        lineheight = txt->whole.lineHeight;
    }
    else {
            /* a string; set lineheight huge so that
               it isn't a factor -- the pointsize test
               has to do with vertical overlap of text
               lines within a field; strings are exempt */
        lineheight = MAXLRU;
    }

    charcount = txt->whole.nChars;
    txAp = &txt->atts[0];
    while(charcount) {
        if (txAp->tctm.rotated) {
            return FALSE;
        }
        if ((int) txAp->fontAttr.ptsize > lineheight) {
            return FALSE;
        }
        charcount -= txAp->nChars;
        txAp++;
    }
    return TRUE;
}
#endif

#if 0
funSketchBounder(w, txt)
long w;
axtext *txt;
{
    r32 bounder;
    char *err;
    Drawable drw;
    int x_off, y_off;

    if (err = getOverallBounder(txt, 0, -1, &bounder)) {
        freeErrString(err);
    }
    
    /* get the drawable */
    drw = AxGetDrawableFromWidget((AXWidget)w, &x_off, &y_off);

    XDrawRectangle(Dpy, drw, axTextGC,
                   milsToPix(bounder.origin.x +
                             txt->whole.pos.x, ScreenRes),
                   milsToPix(bounder.origin.y +
                             txt->whole.pos.y, ScreenRes),
                   milsToPix(bounder.width, ScreenRes),
                   milsToPix(bounder.height, ScreenRes));
}
#endif
