/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: tmimage.c
   Applix ASC version: 199
   Date: Wed Feb 24 21:13:21 1999
*/
#include <stdio.h>
#include <math.h>
#include <aos.h>
#include <paripc.h>
#include <tmi.h>
#include <paratm.h>
#include <afile.h>
#include <paratmgr.h>
#include <parint.h>
#include <tmassert.h>
#include <getstr13.h>
#include <afileimg.h>
#include <basicwin.h>
#include <colors.h>
#include <xmpixel.h>

typedef struct fill_rect {
    int num_rect;
    XRectangle *rect;
} fill_rect;

#if 1
#include <tmimage.p>
#endif

extern unsigned long AlisPixels[], AxPixels[];
extern int AxNColorLevels;
extern int AxSlowDither;
extern unsigned long FGpixel, BGpixel;
extern short RastWid, RastHt, DpyDepth;
extern short NoColor;
extern short ColorDisplay;
extern short ToggleFG;
extern int AxUseGrayScale;
extern Visual *AxVisual;
extern Drawable HostWin;
extern Display *Dpy;
extern GC Gc1;
extern GC SunGC, ShadowGC;
extern Window RootWin, AxTopWinNeverMapped;
extern GC BackGC, ForeGC, RecessGC;
extern int UseWidgetColors;    /* color available for widgets and being used */
extern int axPreDefinedStdmap;

extern GC axCreateGC();

#if AXMS
#define hypot _hypot
#endif

unsigned char bitmask[] = {0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01};
unsigned char fmask[] = { 0, 0x80, 0xc0, 0xe0, 0xf0, 0xf8, 0xfc, 0xfe, 0xff };
int AxForceGreyscale = FALSE;
int AxRequestGreyscale = FALSE;
int NeedPixToIndexTable = TRUE;

extern unsigned long AxGetPixel32();
extern unsigned long AxGetPixel16();
extern unsigned long AxGetPixel24();
extern unsigned long AxGetPixel();

#if AX_X11    /* { */
typedef struct getpixel {
    int depth;
    unsigned long (*func)();
} getpixel;

getpixel GetPixelArray[] = {
    { 0,NULL},
    { 1,AxGetPixel},
    { 2,NULL},
    { 3,NULL},
    { 4,AxGetPixel},
    { 5,NULL},
    { 6,NULL},
    { 7,NULL},
    { 8,AxGetPixel},
    { 9,NULL},
    { 10,NULL},
    { 11,NULL},
    { 12,AxGetPixel},
    { 13,NULL},
    { 14,NULL},
    { 15,NULL},
    { 16,AxGetPixel16},
    { 17,NULL},
    { 18,NULL},
    { 19,NULL},
    { 20,NULL},
    { 21,NULL},
    { 22,NULL},
    { 23,NULL},
    { 24,AxGetPixel24},
    { 25,NULL},
    { 26,NULL},
    { 27,NULL},
    { 28,NULL},
    { 30,NULL},
    { 31,NULL},
    { 32,AxGetPixel32},
};
#endif    /* } AX_X11 */

/****************************************************************************\
*                             utility functions                              *
\****************************************************************************/

/*****************************************************************************
 *
 * function - A x A m I C o l o r
 */
int
AxAmIColor()
{
    return(DpyDepth != 1);
}

/*****************************************************************************
 *
 * function - h a s_ i m a g e_ d a t a
 */
int
has_image_data(imP)
ge_image_type *imP;
{
    if ( imP->data == (unsigned char *)0 || imP->wid == 0 || imP->hyt == 0 )
        return(FALSE);
    else
        return(TRUE);
}

/*****************************************************************************
 *
 * function - h a s_ s y s_ i m a g e_ d a t a
 */
int
has_sys_image_data(imP)
ge_image_type *imP;
{
    if ( imP->sys_data == (unsigned char *)0 )
        return(FALSE);
    else
        return(TRUE);
}

/*****************************************************************************
 *
 * function - h a s_ j a v a_ i m a g e_ d a t a
 */
int
has_java_image_data(imP)
ge_image_type *imP;
{
    if ( imP->java_data == (unsigned char *)0 )
        return(FALSE);
    else
        return(TRUE);
}

/*****************************************************************************
 *
 * function - e x t r a c t _t e r m s
 *
 */
extract_terms(xform, TrX,TrY, ScX,ScY, sinAng,cosAng, ShX, wid,hyt)
poly_type     *xform;
int           *TrX, *TrY;
double        *ScX, *ScY;
double        *sinAng, *cosAng;
double        *ShX;
int           *wid, *hyt;
{
    double width, delta_x, delta_y, orig_ScX, orig_ScY;
    int mwid,mhyt, i, min_x,min_y, max_x,max_y;
    TmPoint32 p1, p2, p3, p4;

    /* no trans means don't scale */
    if ( !xform )
    {
        *TrX = 0.0;
        *TrY = 0.0;
        *ScX = 1.0;
        *ScY = 1.0;
        *sinAng = 0.0;
        *cosAng = 1.0;
        *ShX = 0.0;
    }
    else
    {
        orig_ScX = *ScX;
        orig_ScY = *ScY;
        mwid = *wid;
        mhyt = *hyt;

        /* set the first outline point on the origin */
        min_x = max_x = xform->points[0].x;
        min_y = max_y = xform->points[0].y;

        /* for points[1..4], find the bounding box */
        for ( i=1; i<5; i++ )
        {
            if ( min_x > xform->points[i].x )
                min_x = xform->points[i].x;
            else if ( max_x < xform->points[i].x )
                max_x = xform->points[i].x;
            if ( min_y > xform->points[i].y )
                min_y = xform->points[i].y;
            else if ( max_y < xform->points[i].y )
                max_y = xform->points[i].y;
        }
        min_x = min_x * orig_ScX;
        max_x = max_x * orig_ScX;
        min_y = min_y * orig_ScY;
        max_y = max_y * orig_ScY;
        *wid = max_x - min_x + 1;
        *hyt = max_y - min_y + 1;
        p4.x = (xform->points[3].x - xform->points[0].x) * orig_ScX;
        p4.y = (xform->points[3].y - xform->points[0].y) * orig_ScY;
        p3.x = (xform->points[2].x - xform->points[0].x) * orig_ScX;
        p3.y = (xform->points[2].y - xform->points[0].y) * orig_ScY;
        p2.x = (xform->points[1].x - xform->points[0].x) * orig_ScX;
        p2.y = (xform->points[1].y - xform->points[0].y) * orig_ScY;
        p1.x = (xform->points[0].x - xform->points[0].x) * orig_ScX;
        p1.y = (xform->points[0].y - xform->points[0].y) * orig_ScY;

        /* determine the sin/cos, scale factors, and shear factor */
        width = fabs(hypot((double)(p1.x - p2.x), (double)(p1.y - p2.y)));
        if ( width == 0.0 ) /* an empty field? */
        {
            *ScX    = 1.0; *ScY    = 1.0;
            *sinAng = 0.0; *cosAng = 1.0;
            *ShX    = 0.0;
            *TrX    =   0; *TrY    =   0;
        }
        else
        {
            /* calc the angle info (note: y increases downward) */
            *sinAng = (double)(p2.y - p1.y) / width;
            *cosAng = (double)(p2.x - p1.x) / width;
 
            /* de-rotate last point and calc x shear relative to origin */
            delta_x = (double)p4.x * (*cosAng) + (double)p4.y * (*sinAng);
            delta_y = (double)p4.y * (*cosAng) - (double)p4.x * (*sinAng);
            if ( delta_y < 0.01 && delta_y > -0.01 )
                *ShX = 0.0;
            else
                *ShX = delta_x / delta_y;

            /* the offset of the origin in the image to be created */
            *TrX = (-1.0) * (min_x - (xform->points[0].x * orig_ScX));
            *TrY = (-1.0) * (min_y - (xform->points[0].y * orig_ScY));

            /* calc scaling */
            *ScX = mwid == 0 ? 1.0 : width   / (double)mwid;
            *ScY = mhyt == 0 ? 1.0 : delta_y / (double)mhyt;
        }
    }
}

/*****************************************************************************
 *
 * function - e x t r a c t _t r a n s
 */
extract_trans(xform, wid, hyt, trans)
poly_type *xform;
int wid, hyt;
ge_image_transform_type *trans;
{
    double ShX, ScX, ScY, sinAng, cosAng;
    int TrX, TrY;

    /* get translation, scaling, rotation, shearing from xform, wid, and hyt */
    ScY = trans->y_scale;
    ScX = trans->x_scale;
    extract_terms(xform, &TrX, &TrY,
                  &ScX, &ScY,
                  &sinAng, &cosAng,
                  &ShX,
                  &wid, &hyt);

    conv_i_p32(&wid, &hyt, &wid, &hyt, MILLS_TO_PIXEL);
    conv_i_p32(&TrX, &TrY, &TrX, &TrY, MILLS_TO_PIXEL);
    trans->wid = wid;
    trans->hyt = hyt;
    trans->x_scale = ScX;
    trans->y_scale = ScY;

    /* set up matrix that projects xform onto the source rect */
    if ( ScX == 0.0 )
    {
        trans->x_x = trans->x_y = 0.0;
        trans->x_t = 0;
    }
    else
    {
        trans->x_x = (cosAng + ShX*sinAng)/ScX;
        trans->x_y = (sinAng - ShX*cosAng)/ScX;
        trans->x_t = ShX*(TrY*cosAng - TrX*sinAng)/ScX -
                      (TrX*cosAng + TrY*sinAng)/ScX;
    }
    if ( ScY == 0.0 )
    {
        trans->y_x = trans->y_y = 0.0;
        trans->y_t = 0;
    }
    else
    {
        trans->y_x = (sinAng / ScY);
        trans->y_y = (cosAng / ScY);
        trans->y_t = (TrX*sinAng - TrY*cosAng)/ScY;
    }
}

/*****************************************************************************
 *
 * function - e x t r a c t _f a c t o r s
 */
extract_factors(imP, xform, trans)
ge_image_type *imP;
poly_type *xform;
ge_image_transform_type *trans;
{
	int wid, hyt;

    wid = imP->wid;
    hyt = imP->hyt;
    conv_i_p32(&wid,&hyt, &wid,&hyt, PIXEL_TO_MILLS);
	extract_trans(xform, wid, hyt, trans);
}

/*****************************************************************************
 *
 * function - 
 */
extract_transform_info(imP, xform, trans)
ge_image_type *imP;
poly_type *xform;
ge_image_transform_type *trans;
{
	int wid, hyt;
    double ShX, ScX, ScY, sinAng, cosAng;
    int TrX, TrY;

    wid = imP->wid;
    hyt = imP->hyt;
    conv_i_p32(&wid,&hyt, &wid,&hyt, PIXEL_TO_MILLS);

    /* get translation, scaling, rotation, shearing from xform, wid, and hyt */
    ScY = trans->y_scale;
    ScX = trans->x_scale;
    extract_terms(xform, &TrX, &TrY,
                  &ScX, &ScY,
                  &sinAng, &cosAng,
                  &ShX,
                  &wid, &hyt);

    conv_i_p32(&wid, &hyt, &wid, &hyt, MILLS_TO_PIXEL);
    conv_i_p32(&TrX, &TrY, &TrX, &TrY, MILLS_TO_PIXEL);
    trans->wid = wid;
    trans->hyt = hyt;
    trans->x_scale = ScX;
    trans->y_scale = ScY;
	trans->x_x = cosAng*ScX;
	trans->x_y = ScY*(ShX*cosAng - sinAng);
	trans->x_t = 0;
	trans->y_x = sinAng*ScX;
	trans->y_y = (cosAng + ShX*sinAng)* ScY;
	trans->y_t = 0;
}

/*****************************************************************************
 *
 * function - m u l t i p l y _p o i n t s
 */
multiply_points(x,y,tr)
int *x,*y;
ge_image_transform_type *tr;
{
    int nx, ny;

    ny = (*y);
    nx = (*x);

	*x = nx*tr->x_x + ny*tr->x_y + tr->x_t;
	*y = ny*tr->y_y + nx*tr->y_x + tr->y_t;
}

/*****************************************************************************
 *
 * function - m a t r i x _i n v e r s e
 * 
 *  Computes the inverse of the leading 3x3 submatrix of the 4x4 matrix inm 
 *  and places the inverse in aug. Uses Gaussian Elimination Technique  
 *  DOES NOT CHECK TO SEE IF MATRIX IS NOT INVERTIBLE!!
 */
matrix_inverse(inm,aug)
float inm[4][4];
float aug[3][3];
{
    float m[3][3], s;
    int i,j;

    /* Copy the input matrix, so we can muck about with it. */
    for ( i=0; i<3; i++ ) 
        for ( j=0; j<3; j++ )
            m[i][j] = inm[i][j];

    /* Make the augmented matrix the identity. */
    for ( i=0; i<3; i++ )
    {
        for ( j=0; j<3; j++ )
        {
            if ( i == j ) 
                aug[i][j] = 1;
            else
                aug[i][j] = 0;
        }
    }


    /* Make [0][0] be 1.0. */
    s = m[0][0];
    m[0][0] /= s;
    m[1][0] /= s;
    m[2][0] /= s;

    aug[0][0] /= s;

    /* add -m[0][1] * row0 to row1 to make m[0][1] = 0.0 */
    s = -m[0][1];
    m[0][1] += s * m[0][0];
    m[1][1] += s * m[1][0];
    m[2][1] += s * m[2][0];

    aug[0][1] += s * aug[0][0];

    /* Add -m[0][2] * row0 to row2 to make m[0][2] = 0.0 */
    s = -m[0][2];
    m[0][2] += s * m[0][0];
    m[1][2] += s * m[1][0];
    m[2][2] += s * m[2][0];
    aug[0][2] += s * aug[0][0];

    /* Make [1][1] be 1.0 */
    s = m[1][1];
    m[1][1] /= s;
    m[2][1] /= s;

    aug[0][1] /= s;
    aug[1][1] /= s;

    /* Add -m[1][2] * row1 to row2 to make m[1][2] = 0.0 */
    s = -m[1][2];
    m[1][2] += s * m[1][1];
    m[2][2] += s * m[2][1];

    aug[0][2] += s * aug[0][1];
    aug[1][2] += s * aug[1][1];

    /* Make [2][2] be 1.0. */
    s = m[2][2];
    m[2][2] /= s;

    aug[2][2] /= s;
    aug[0][2] /= s;
    aug[1][2] /= s;

    /* Make m[2][1] be 0.0 */
    s = -m[2][1];
    m[2][1] += s * m[2][2];

    aug[2][1] += s * aug[2][2];
    aug[0][1] += s * aug[0][2];
    aug[1][1] += s * aug[1][2];

    /* Make m[2][0] be 0.0 */
    s = -m[2][0];
    m[2][0] += s * m[2][2];

    aug[2][0] += s * aug[2][2];
    aug[0][0] += s * aug[0][2];
    aug[1][0] += s * aug[1][2];

    /* make m[1][0] be 0.0 */
    s = -m[1][0];
    m[1][0] += s * m[1][1];

    aug[0][0] += s * aug[0][1];
    aug[1][0] += s * aug[1][1];
    aug[2][0] += s * aug[2][1];
}

/*****************************************************************************
 *
 * function - c a l c u l a t e _i n v e r s e
 */
calculate_inverse(tr,inv_tr)
ge_image_transform_type *tr;
ge_image_transform_type *inv_tr;
{
    float tm[4][4];
    float aug[3][3];
    double val;

    /* Copy the "tranform" into a 4x4 matrix */
    tm[0][0] = tr->x_x;
    tm[0][1] = tr->x_y;
    tm[0][2] = 0;
    tm[0][3]= 0;

    tm[1][0] = tr->y_x;
    tm[1][1] = tr->y_y;
    tm[1][2] = 0;
    tm[1][3] = 0;

    tm[2][0] = tr->x_t;
    tm[2][1] = tr->y_t;
    tm[2][2] = 1.0;
    tm[2][3] = 1;

    tm[3][0] = 0;
    tm[3][1] = 0;
    tm[3][2] = 0;
    tm[3][3] = 1.0;

	/* First Check if the matrix is singular!!! */
	val = tm[0][0]*(tm[1][1]*tm[2][2] - tm[1][2]*tm[2][1]);
	val += tm[0][1]*(tm[1][0]*tm[2][2] - tm[1][2]*tm[2][0]);
	val += tm[0][2]*(tm[1][0]*tm[2][1] - tm[1][1]*tm[2][0]);
	if ( val <= 0 )
	{
		inv_tr->x_x = tr->x_x;
		inv_tr->x_y = tr->x_y;

		inv_tr->y_x = tr->y_x;
		inv_tr->y_y = tr->y_y;

		inv_tr->x_t = tr->x_t;
		inv_tr->y_t = tr->x_t;
	}
	else
	{
		/* invert it into aug */
		matrix_inverse(tm,aug);

		/* return it as a "transform" */
		inv_tr->x_x = aug[0][0];
		inv_tr->x_y = aug[0][1];

		inv_tr->y_x = aug[1][0];
		inv_tr->y_y = aug[1][1];

		inv_tr->x_t = aug[2][0];
		inv_tr->y_t = aug[2][1];
	}
}

/*****************************************************************************
 *
 * function - m o d i f y _t r a n s
 */
modify_trans(xform,imP,tx1,ty1,left,top,wid,hyt)
poly_type *xform;
ge_image_type *imP;
int tx1,ty1;
int left,top;
int wid,hyt;
{
    ge_image_transform_type tr, inv_tr;
    int i, min_x, min_y, max_x, max_y;
    TmPoint32 p0, p1, p2, p3;

    /* get the transform */
    tr.x_scale = 1.0;
    tr.y_scale = 1.0;
    extract_transform_info(imP, xform, &tr);

    /* convert height and width from pixels to mils */
    conv_i_p32(&left,&top,&left,&top, PIXEL_TO_MILLS);  
    conv_i_p32(&wid,&hyt,&wid,&hyt, PIXEL_TO_MILLS);

    /* get the untransformed boundary */
    p0.x = left;
    p0.y = top;
    p1.x = p0.x + wid;
    p1.y = p0.y;
    p2.x = p1.x;
    p2.y = p0.y + hyt;
    p3.x = p0.x;
    p3.y = p2.y;

    /* apply the transform */
    multiply_points(&p0.x,&p0.y,&tr);   
    multiply_points(&p1.x,&p1.y,&tr);   
    multiply_points(&p2.x,&p2.y,&tr);   
    multiply_points(&p3.x,&p3.y,&tr);   
    xform->points[0].x = p0.x;   
    xform->points[0].y = p0.y;
    xform->points[1].x = p1.x;
    xform->points[1].y = p1.y;
    xform->points[2].x = p2.x;
    xform->points[2].y = p2.y;
    xform->points[3].x = p3.x;
    xform->points[3].y = p3.y;
    xform->points[4].x = p0.x;
    xform->points[4].y = p0.y;
    
    /* find the min's and max's */
    min_x = max_x = xform->points[0].x;
    min_y = max_y = xform->points[0].y;
    for ( i=1; i<5; i++ )
    {
        if ( min_x > xform->points[i].x )
            min_x = xform->points[i].x;
        if ( min_y > xform->points[i].y )
            min_y = xform->points[i].y;
    }

    /* convert from local to global coordinates */
    xform->points[0].x -= min_x;
    xform->points[0].x += tx1;
    xform->points[0].y -= min_y;
    xform->points[0].y += ty1;

    xform->points[1].x -= min_x;
    xform->points[1].x += tx1;
    xform->points[1].y -= min_y;
    xform->points[1].y += ty1;

    xform->points[2].x -= min_x;
    xform->points[2].x += tx1;
    xform->points[2].y -= min_y;
    xform->points[2].y += ty1;

    xform->points[3].x -= min_x;
    xform->points[3].x += tx1;
    xform->points[3].y -= min_y;
    xform->points[3].y += ty1;

    xform->points[4].x = xform->points[0].x;
    xform->points[4].y = xform->points[0].y;
}

/*****************************************************************************
 *
 * function - c h e c k _b o u n d a r y (for fast scaling)
 */
check_boundary( imP, xform, trans )
ge_image_type *imP;
poly_type *xform;
ge_image_transform_type *trans;
{
    int mwid,mhyt, wid,hyt, i, min_x,min_y, max_x,max_y;
    double width, height, orig_ScX, orig_ScY, ScX,ScY;

    /* fast scaling cannot do Floyd-Steinberg dither */
    if ( imP->dep == 8  && ( AxNColorLevels == 1 || AxSlowDither ))
        return(FALSE);
    /* temp - untill the code gets fixed */
    if ( DpyDepth == 4 )
        return(FALSE);
    if ( DpyDepth == 16 )
        return(FALSE);
    if ( DpyDepth == 24 )
        return(FALSE);

    /* no transform means no scaling */
    if ( !xform )
        return(FALSE);

    /* check for vert flip */
	if ( trans->y_y < 0 )
        return(FALSE);
    /* check for rotation */
    if ( xform->points[0].x != xform->points[3].x )
        return(FALSE);
    if ( xform->points[0].y != xform->points[1].y )
        return(FALSE);

    return(TRUE);
}

/****************************************************************************/
morebits(dP, dbit, step)
register unsigned char *dP;
register int dbit, step;
{
    *dP++ |= 0xFF >> dbit;

    step -= 8 - dbit;
    while ( step >= 8 )
    {
        *dP++ = 0xFF;
        step -= 8;
    }
    *dP++ |= fmask[step];
}

/****************************************************************************
 *
 * module:
 */
flipbyte(dP, dst_wid)
unsigned char *dP;
int dst_wid;
{
    int xFlipByte, xFlipIndex, ix;

    xFlipIndex = dst_wid -1;
    ix = 0;
    while(xFlipIndex > ix )
    {
        xFlipByte      = dP[ix];
        dP[ix]          = dP[xFlipIndex];
        dP[xFlipIndex] = xFlipByte;
        xFlipIndex--;
        ix++;
    }
}

/****************************************************************************
 *
 * module:  f l i p B i t s
 */
flipBits(sp, dst_wid)
unsigned char *sp;
int dst_wid;
{
    register unsigned char *p0, *p1;
    register unsigned char m0, m1;
    register int i0, i1;

    i0 = 0;
    p0 = sp;
    i1 = dst_wid - 1;
    p1 = &sp[i1 >> 3];
    i1 &= 7;

    while (p0 < p1 || i0 < i1) {
        m0 = bitmask[i0];
        m1 = bitmask[i1];

        /* these casts will sometimes produce smaller code */
        if (((unsigned char) (*p0 & m0) == 0) !=
            ((unsigned char) (*p1 & m1) == 0)) {
            *p0 ^= m0;
            *p1 ^= m1;
        }

        if (7 < ++i0) {
            i0 = 0;
            ++p0;
        }
        if (--i1 < 0) {
            i1 = 7;
            --p1;
        }
    }
}

/****************************************************************************/
clip_mask(mask_rect,clip_rect, new_rect)
XRectangle *mask_rect;
XRectangle *clip_rect;
fill_rect *new_rect;
{
    int clip_rx,clip_ly,mask_rx,mask_ly;
    int num_of_rect;
    XRectangle rect[4];

    clip_rx = clip_rect->width +clip_rect->x;
    clip_ly = clip_rect->height +clip_rect->y;
    mask_rx = mask_rect->width +mask_rect->x;
    mask_ly = mask_rect->height +mask_rect->y;

    if( clip_rect->x <= mask_rect->x )
    {
        if ( mask_rx <= clip_rx ) /* 2 pts ?? */ 
        {
            if( clip_rect->y <= mask_rect->y ) /* 2 pts */
            {
                if( mask_ly > clip_ly ) /* 4pts */
                {
                    num_of_rect = 1;
                    new_rect->rect = (XRectangle *)malloc(num_of_rect *
                                sizeof(XRectangle ));
                    new_rect->rect[0].x = mask_rect->x;
                    new_rect->rect[0].y = clip_ly;
                    new_rect->rect[0].width = mask_rect->width;
                    new_rect->rect[0].height = mask_ly - clip_ly;
                }
                else        /* mask_ly <= clip_ly */
                    num_of_rect = 0;
            }
            else        /* clip_rect->y >mask_rect->y */ 
            {
                if( mask_ly <= clip_ly ) 
                {
                    if( mask_ly > clip_rect->y ) 
                    {
                        num_of_rect = 1;
                        new_rect->rect = (XRectangle *)malloc(num_of_rect *
                                    sizeof(XRectangle ));
                        new_rect->rect[0].x = mask_rect->x;
                        new_rect->rect[0].y = mask_rect->y;
                        new_rect->rect[0].width = mask_rect->width;
                        new_rect->rect[0].height = clip_rect->y - mask_rect->y;
                    }
                    else
                        num_of_rect = 0;
                }
                else        /* mask_ly > clip_ly */
                {
                    num_of_rect = 2;
                    new_rect->rect = (XRectangle *)malloc(num_of_rect *
                                sizeof(XRectangle ));
                    new_rect->rect[0].x = mask_rect->x;
                    new_rect->rect[0].y = mask_rect->y;
                    new_rect->rect[0].width = mask_rect->width;
                    new_rect->rect[0].height = clip_rect->y - mask_rect->y;

                    new_rect->rect[1].x = mask_rect->x;
                    new_rect->rect[1].y = clip_ly;
                    new_rect->rect[1].width = mask_rect->width;
                    new_rect->rect[1].height = mask_ly - clip_ly;
                }
            }
        }
        else            /* (mask_rx > clip_rx)  1 pt*/
        {
            if(clip_rect->y <= mask_rect->y )
            {
                if((mask_ly) >= clip_ly)
                {
                    if(clip_ly > mask_rect->y) 
                    {
                        num_of_rect = 2;
                        new_rect->rect = (XRectangle *)malloc(num_of_rect *
                                    sizeof(XRectangle ));
                        new_rect->rect[0].x = mask_rect->x;
                        new_rect->rect[0].y = clip_ly;
                        new_rect->rect[0].width = mask_rect->width;
                        new_rect->rect[0].height = mask_ly - clip_ly;

                        new_rect->rect[1].x = clip_rx;
                        new_rect->rect[1].y = mask_rect->y;
                        new_rect->rect[1].width = mask_rx - clip_rx;
                        new_rect->rect[1].height = clip_ly - mask_rect->y;
                    }
                    else
                        num_of_rect = 0;
                }
                else    /*(mask_ly) <clip_ly    */
                {
                    num_of_rect = 1;
                    new_rect->rect = (XRectangle *)malloc(num_of_rect *
                                sizeof(XRectangle ));
                    new_rect->rect[0].x = clip_rx;
                    new_rect->rect[0].y = mask_rect->y;
                    new_rect->rect[0].width = mask_rx - clip_rx;
                    new_rect->rect[0].height = mask_rect->height;
                }
            }
            else    /*(clip_rect->y > mask_rect->y)   */
            {
                if((mask_ly) >= clip_ly)
                {
                    num_of_rect = 3;
                    new_rect->rect = (XRectangle *)malloc(num_of_rect *
                                sizeof(XRectangle ));

                    new_rect->rect[0].x = mask_rect->x;
                    new_rect->rect[0].y = mask_rect->y;
                    new_rect->rect[0].width = mask_rect->width;
                    new_rect->rect[0].height = clip_rect->y -mask_rect->y;

                    new_rect->rect[1].x = clip_rx;
                    new_rect->rect[1].y = clip_rect->y;
                    new_rect->rect[1].width = mask_rx - clip_rx;
                    new_rect->rect[1].height = clip_ly - clip_rect->y;

                    new_rect->rect[2].x = mask_rect->x;
                    new_rect->rect[2].y = clip_ly;
                    new_rect->rect[2].width = mask_rect->width;
                    new_rect->rect[2].height = mask_ly - clip_ly;
                }
                else /*((mask_ly) <clip_ly)  */
                {
                    if(mask_ly > clip_rect->y)
                    {
                        num_of_rect = 2;
                        new_rect->rect = (XRectangle *)malloc(num_of_rect *
                                    sizeof(XRectangle ));
                        new_rect->rect[0].x = mask_rect->x;
                        new_rect->rect[0].y = mask_rect->y;
                        new_rect->rect[0].width = mask_rect->width;
                        new_rect->rect[0].height = clip_rect->y - mask_rect->y;

                        new_rect->rect[1].x = clip_rx;
                        new_rect->rect[1].y = clip_rect->y;
                        new_rect->rect[1].width = mask_rx - clip_rx;
                        new_rect->rect[1].height = mask_ly - clip_rect->y;
                    }
                    else 
                        num_of_rect =0;
                }
            }
        }
    }
    else /*if(clip_rect->x > mask_rect->x ) */
    {
        if(mask_rx >= clip_rx)
        {
            if(clip_rect->y <= mask_rect->y ) 
            {
                if(mask_ly <= clip_ly) 
                {
                    num_of_rect = 2;
                    new_rect->rect = (XRectangle *)malloc(num_of_rect *
                                sizeof(XRectangle ));
                    new_rect->rect[0].x = mask_rect->x;
                    new_rect->rect[0].y = mask_rect->y;
                    new_rect->rect[0].width = clip_rect->x - mask_rect->x;
                    new_rect->rect[0].height = mask_rect->height;

                    new_rect->rect[1].x = clip_rx;
                    new_rect->rect[1].y = mask_rect->y;
                    new_rect->rect[1].width = mask_rx -clip_rx;
                    new_rect->rect[1].height = mask_rect->height;
                }
                else        /* mask_ly > clip_ly */
                {
                    if(clip_ly > mask_rect->y)
                    {
                        num_of_rect = 3;
                        new_rect->rect = (XRectangle *)malloc(num_of_rect *
                                    sizeof(XRectangle ));

                        new_rect->rect[0].x = mask_rect->x;
                        new_rect->rect[0].y = clip_ly;
                        new_rect->rect[0].width = mask_rect->width;
                        new_rect->rect[0].height = mask_ly - clip_ly;
    
                        new_rect->rect[1].x = mask_rect->x;
                        new_rect->rect[1].y = mask_rect->y;
                        new_rect->rect[1].width = clip_rect->x -mask_rect->x;
                        new_rect->rect[1].height = clip_ly - mask_rect->y;

                        new_rect->rect[2].x = clip_rx;
                        new_rect->rect[2].y = mask_rect->y;
                        new_rect->rect[2].width = mask_rx - clip_rx;
                        new_rect->rect[2].height = clip_ly - mask_rect->y;
                    }
                    else
                        num_of_rect = 0;
                }
            }
            else        /* clip_rect->y >mask_rect->y */ 
            {
                if(mask_ly <= clip_ly) 
                {
                    if(mask_ly > clip_rect->y) 
                    {
                        num_of_rect = 3;
                        new_rect->rect = (XRectangle *)malloc(num_of_rect *
                                    sizeof(XRectangle ));
                        new_rect->rect[0].x = mask_rect->x;
                        new_rect->rect[0].y = mask_rect->y;
                        new_rect->rect[0].width = mask_rect->width;
                        new_rect->rect[0].height = clip_rect->y - mask_rect->y;
    
                        new_rect->rect[1].x = mask_rect->x;
                        new_rect->rect[1].y = clip_rect->y;
                        new_rect->rect[1].width = clip_rect->x -mask_rect->x;
                        new_rect->rect[1].height = mask_ly - clip_rect->y;

                        new_rect->rect[2].x = clip_rx;
                        new_rect->rect[2].y = clip_rect->y;
                        new_rect->rect[2].width = mask_rx -clip_rx;
                        new_rect->rect[2].height = mask_ly - clip_rect->y;
                    }
                    else
                        num_of_rect = 0;
                }
                else        /* mask_ly > clip_ly */
                {
                    num_of_rect = 4;
                    new_rect->rect = (XRectangle *)malloc(num_of_rect *
                                sizeof(XRectangle ));
                    new_rect->rect[0].x = mask_rect->x;
                    new_rect->rect[0].y = mask_rect->y;
                    new_rect->rect[0].width = mask_rect->width;
                    new_rect->rect[0].height = clip_rect->y - mask_rect->y;

                    new_rect->rect[1].x = mask_rect->x;
                    new_rect->rect[1].y = clip_ly;
                    new_rect->rect[1].width = mask_rect->width;
                    new_rect->rect[1].height = mask_ly - clip_ly;

                    new_rect->rect[2].x = mask_rect->x;
                    new_rect->rect[2].y = clip_rect->y;
                    new_rect->rect[2].width = clip_rect->x - mask_rect->x;
                    new_rect->rect[2].height = clip_ly - clip_rect->y;

                    new_rect->rect[3].x = clip_rx;
                    new_rect->rect[3].y = clip_rect->y;
                    new_rect->rect[3].width = mask_rx - clip_rx;
                    new_rect->rect[3].height = clip_ly - clip_rect->y;
                
                }
            }
        }
        else /*if(mask_rx < clip_rx) */
        {
            if(clip_rect->y <= mask_rect->y)
            {
                if(mask_ly >= clip_ly)
                {
                    if(clip_ly > mask_rect->y)
                    {
                        num_of_rect = 2;
                        new_rect->rect = (XRectangle *)malloc(num_of_rect *
                                    sizeof(XRectangle ));
                        new_rect->rect[0].x = mask_rect->x;
                        new_rect->rect[0].y = clip_ly;
                        new_rect->rect[0].width = mask_rect->width;
                        new_rect->rect[0].height = mask_ly - clip_ly;

                        new_rect->rect[1].x = mask_rect->x;
                        new_rect->rect[1].y = mask_rect->y;
                        new_rect->rect[1].width = clip_rect->x -mask_rect->x;
                        new_rect->rect[1].height = clip_ly -mask_rect->y;
                    }
                    else
                        num_of_rect = 0;
                }
                else /*if((mask_ly) < clip_ly) */
                {
                    num_of_rect = 1;
                    new_rect->rect = (XRectangle *)malloc(num_of_rect *
                                sizeof(XRectangle ));
                    new_rect->rect[0].x = mask_rect->x;
                    new_rect->rect[0].y = mask_rect->y;
                    new_rect->rect[0].width = clip_rect->x -mask_rect->x;
                    new_rect->rect[0].height = mask_rect->height;
                }
            }
            else /*if(clip_rect->y >mask_rect->y) */
            {
                if(mask_ly >= clip_ly)
                {
                    if(mask_rx > clip_rect->x)
                    {
                        num_of_rect = 3;
                        new_rect->rect = (XRectangle *)malloc(num_of_rect *
                                    sizeof(XRectangle ));
                        new_rect->rect[0].x = mask_rect->x;
                        new_rect->rect[0].y = mask_rect->y;
                        new_rect->rect[0].width = mask_rect->width;
                        new_rect->rect[0].height = clip_rect->y - mask_rect->y;
    
                        new_rect->rect[1].x = mask_rect->x;
                        new_rect->rect[1].y = clip_ly;
                        new_rect->rect[1].width = mask_rect->width;
                        new_rect->rect[1].height = mask_ly - clip_ly;
    
                        new_rect->rect[2].x = mask_rect->x;
                        new_rect->rect[2].y = clip_rect->y;
                        new_rect->rect[2].width = clip_rect->x - mask_rect->x;
                        new_rect->rect[2].height = clip_ly - clip_rect->y;
                    }
                    else
                        num_of_rect = 0;
                }
                else /* mask_ly <clip_ly */
                {
                    num_of_rect = 2;
                    new_rect->rect = (XRectangle *)malloc(num_of_rect *
                                sizeof(XRectangle ));
                    new_rect->rect[0].x = mask_rect->x;
                    new_rect->rect[0].y = mask_rect->y;
                    new_rect->rect[0].width = mask_rect->width;
                    new_rect->rect[0].height = clip_rect->y - mask_rect->y;

                    new_rect->rect[1].x = mask_rect->x;
                    new_rect->rect[1].y = clip_rect->y;
                    new_rect->rect[1].width = clip_rect->x -mask_rect->x;
                    new_rect->rect[1].height = mask_ly - clip_rect->y ;
                }
            }
        }
    }
    new_rect->num_rect = num_of_rect;
    return(num_of_rect);
}

/****************************************************************************/
unsigned char *
xMsame1(buf, new_wid, orig_wid, bwid, flip_flag)
unsigned char *buf;
int new_wid,orig_wid,bwid;
int flip_flag;
{
    unsigned char *temp_buf;

    /* nothing to be done - just copy the buffer and return */
    temp_buf = (unsigned char *)malloc(bwid);
    memcpy(temp_buf, buf, bwid);

    if ( flip_flag )
        flipBits(temp_buf, new_wid);
    /* done */
    return(temp_buf);
}

/****************************************************************************/
unsigned char *
xMcompress1(buf, new_wid, orig_wid, bwid, flip_flag)
unsigned char *buf;
int new_wid, orig_wid, bwid;
int flip_flag;
{
    register int step, accum, dst_wid, sbit, dbit, x_mstep, x_error;
    unsigned char *temp_buf, *sP, *dP;

    /* create a temp scanline */
    temp_buf = (unsigned char *)malloc(bwid);
    memset (temp_buf, 0, bwid);

    /* local init */
    x_mstep = orig_wid / new_wid;
    x_error = orig_wid - x_mstep * new_wid;
    sP = buf;
    dP = temp_buf;
    sbit = dbit = 0;
    dst_wid = orig_wid;
    accum = new_wid >> 1;

    /* fill out the temp scanline */
    while ( 0 < dst_wid )
    {
        /* get the number of rows to fold */
        step = x_mstep;
        accum -= x_error;
        if ( accum < 0 ) 
        {
            accum += new_wid;
            step++;
        }
        if ( dst_wid < step )
            step = dst_wid;

        /* reduce the remaining row count */
        dst_wid -= step;

        /* set a bit into the buffer */
        if ( *sP & bitmask[sbit] )
            *dP |= bitmask[dbit];

        /* increment the buffer pointers */
        sbit += step;
        while ( 8 <= sbit ) 
        {
            ++sP;
            sbit -= 8;
        }
        if ( 8 <= ++dbit )
        {
            ++dP;
            dbit = 0;
        }
    }

    if ( flip_flag )
        flipBits(temp_buf, new_wid);

    /* done */
    return (temp_buf);
}

/****************************************************************************/
unsigned char *
xMexpand1(buf, new_wid, orig_wid, bwid, flip_flag)
unsigned char *buf;
int new_wid, orig_wid, bwid;
int flip_flag;
{
    register int step, accum, dst_wid, sbit, dbit, x_mstep,x_error;
    unsigned char *temp_buf,*sP, *dP;

    /* create a temp scanline */
    temp_buf = (unsigned char *)malloc(bwid);
    memset(temp_buf, 0, bwid);

    dst_wid = new_wid;
    x_mstep = new_wid / orig_wid;
    x_error = new_wid - x_mstep * orig_wid;
    accum = orig_wid >> 1;
    sbit = dbit = 0;
    sP = buf;
    dP = temp_buf;

    while ( 0 < dst_wid )
    {
        /* get the number of rows to skip */
        step = x_mstep;
        accum -= x_error;
        if ( accum < 0 ) 
        {
            accum += orig_wid;
            ++step;
        }
        if ( dst_wid < step )
            step = dst_wid;

        /* reduce the width count */
        dst_wid -= step;

        if ( *sP & bitmask[sbit] )
        {
            if ( step <= 8 - dbit )
                *dP |= fmask[step] >> dbit;
            else
                morebits(dP, dbit, step);
        }

        /* increment the buffer pointers */
        if ( 8 <= ++sbit ) 
        {
            ++sP;
            sbit = 0;
        }
        dbit += step;
        while ( 8 <= dbit )
        {
            ++dP;
            dbit -= 8;
        }
    }

    if ( flip_flag )
        flipBits(temp_buf, new_wid);

    /* done */
    return (temp_buf);
}

/****************************************************************************/
unsigned char *
xCsame24(buf, xP, cmapPixel, new_wid, orig_wid, bwid, row_ix, flip)
unsigned char *buf;
XImage *xP;
unsigned long *cmapPixel;
int new_wid, orig_wid, bwid, row_ix;
int flip;
{
    unsigned char *sP;
    int col_ix; 

    sP = buf;
    for ( col_ix=0; col_ix<orig_wid; col_ix++ )
    {
        XPutPixel(xP, col_ix, row_ix, cmapPixel[*sP]);
        sP++;
    }

}

/****************************************************************************/
unsigned char *
xCcompress24(buf, xP,cmapPixel, new_wid, orig_wid, bwid, column, flip)
unsigned char *buf;
XImage *xP;
unsigned long *cmapPixel;
int new_wid,orig_wid,bwid, column;
int flip;
{
    register int ix, row_ix, col_ix, step, accum, dst_wid, x_mstep, x_error;
    unsigned char *sP;

    x_mstep = orig_wid / new_wid;
    x_error = orig_wid - x_mstep * new_wid;
    sP = buf;
    row_ix = 0;
    dst_wid = orig_wid;
    accum = new_wid >> 1;

    while ( 0 < dst_wid )
    {
        step = x_mstep;
        accum -= x_error;
        if (accum < 0) 
        {
            accum += new_wid;
            step++;
        }
        if ( dst_wid < step )
            step = dst_wid;

        dst_wid -= step;

        XPutPixel(xP, row_ix, column, cmapPixel[*sP]);

        sP += step;
        row_ix += step;
    }
}

/****************************************************************************/
unsigned char *
xCexpand24( buf, xP, cmapPixel,new_wid,orig_wid,bwid, column, flip )
unsigned char *buf;
XImage *xP;
unsigned long *cmapPixel;
int new_wid,orig_wid,bwid, column;
int flip;
{
    register int i, row_ix, step, accum, dst_wid, x_mstep, x_error;
    unsigned long ix;
    unsigned char *sP;

    dst_wid = new_wid;
    x_mstep = new_wid/orig_wid;
    x_error = new_wid - x_mstep*orig_wid;
    accum = orig_wid >> 1;
    sP = buf;

	row_ix = 0;
    while ( 0 < dst_wid )
    {
        step = x_mstep;
        accum -= x_error;
        if (accum < 0) 
        {
            accum += orig_wid;
            ++step;
        }
        if ( dst_wid < step )
            step = dst_wid;

        dst_wid -= step;

        ix = cmapPixel[*sP];
        for ( i=0; i<step; i++ )
        {
            XPutPixel(xP, row_ix, column,ix);
            row_ix++;
        }
        sP++;
    }
}

/****************************************************************************/
unsigned char *
xCsame8(buf, col_tab, new_wid, orig_wid, bwid, column, flip)
unsigned char *buf;
col_table *col_tab;
int new_wid, orig_wid, bwid, column;
int flip;
{
    unsigned char *temp_buf;
    unsigned char *sP, *dP;
    col_table *col;
    int row_ix, col_ix, ix;

    /* create a temp scaline */
    temp_buf = (unsigned char *)malloc(bwid);
    memset (temp_buf, 0, bwid);

    sP = buf;
    dP = temp_buf;
    col_ix = column%4;

    /* copy the src buffer and return */
    for ( row_ix=0; row_ix<orig_wid; row_ix++ )
    {
        ix = (row_ix%4)*4 + col_ix;
        col = &col_tab[*sP];
        *dP = AxPixels[DITHER(ix, col)];

        dP++;
        sP++;
    }
    dP = temp_buf;
    if ( flip )
        flipbyte(dP, new_wid);

    /* done */
    return (temp_buf);
}

/****************************************************************************/
unsigned char *
xCcompress8(buf, col_tab, new_wid, orig_wid, bwid, column, flip)
unsigned char *buf;
col_table *col_tab;
int new_wid,orig_wid,bwid, column;
int flip;
{
    register int ix, row_ix, col_ix, step, accum, dst_wid, x_mstep, x_error;
    unsigned char *temp_buf, *sP, *dP;
    col_table *col;
    int i, tmp;

    temp_buf = (unsigned char *)malloc(bwid);
    memset (temp_buf, 0, bwid);

    x_mstep = orig_wid / new_wid;
    x_error = orig_wid - x_mstep * new_wid;
    sP = buf;
    dP = temp_buf;
    row_ix = 0;
    col_ix = column%4;
    dst_wid = orig_wid;
    accum = new_wid >> 1;

    while ( 0 < dst_wid )
    {
        step = x_mstep;
        accum -= x_error;
        if (accum < 0) 
        {
            accum += new_wid;
            step++;
        }
        if ( dst_wid < step )
            step = dst_wid;

        dst_wid -= step;

        tmp = 0xFF;
        for(i=0; i<step; i++)
        {
            if (*sP < tmp)
                tmp = *sP;
            sP++;
            row_ix++;
        }
        ix = (row_ix%4)*4 + col_ix;
        col = &col_tab[tmp];
        *dP = AxPixels[DITHER(ix, col)];
        dP++;
    }

    dP = temp_buf;
    if ( flip )
        flipbyte(dP, new_wid);
    /* done */
    return (temp_buf);
}

/****************************************************************************/
unsigned char *
xCexpand8( buf,col_tab,new_wid,orig_wid,bwid, column, flip )
unsigned char *buf;
col_table *col_tab;
int new_wid,orig_wid,bwid, column;
int flip;
{
    register int i, ix, row_ix, col_ix, step, accum, dst_wid, x_mstep, x_error;
    unsigned char *temp_buf, *sP, *dP;
    col_table *col;

    temp_buf = (unsigned char *)malloc(bwid);
    memset(temp_buf, 0, bwid);

    dst_wid = new_wid;
    x_mstep = new_wid/orig_wid;
    x_error = new_wid - x_mstep*orig_wid;
    accum = orig_wid >> 1;
    sP = buf;
    dP = temp_buf;
    col_ix = column%4;

	row_ix = 0;
    while ( 0 < dst_wid )
    {
        step = x_mstep;
        accum -= x_error;
        if (accum < 0) 
        {
            accum += orig_wid;
            ++step;
        }
        if ( dst_wid < step )
            step = dst_wid;

        dst_wid -= step;

        col = &col_tab[*sP];
        for ( i=0; i<step; i++ )
        {
            ix = (row_ix%4)*4 + col_ix;
            *dP = AxPixels[DITHER(ix, col)];

            dP++;
            row_ix++;
        }
        sP++;
    }

    dP = temp_buf;
    if ( flip )
        flipbyte(dP, new_wid);
    /* done */
    return (temp_buf);
}

/****************************************************************************/
unsigned char *
xCsame4(buf, col_tab, new_wid, orig_wid, bwid, column, flip)
unsigned char *buf;
col_table *col_tab;
int new_wid, orig_wid, bwid, column;
int flip;
{
    unsigned char *temp_buf;
    unsigned char *sP, *dP;
    int row_ix, col_ix, ix;
    col_table *col;

    /* create a temp scaline */
    temp_buf = (unsigned char *)malloc(bwid);
    memset(temp_buf, 0, bwid);

    sP = buf;
    dP = temp_buf;
    col_ix = column%4;

    /* copy the src buffer and return */
    for ( row_ix=0; row_ix<orig_wid; row_ix++ )
    {
        ix = (row_ix%4)*4 + col_ix;
        col = &col_tab[*sP];
        *( dP+row_ix/2 )  = ( AxPixels[DITHER(ix, col)] & 0x0F);

        sP++;
    }

    /* done */
    return (temp_buf);
}

/****************************************************************************/
unsigned char *
xCcompress4(buf, col_tab, new_wid, orig_wid, bwid, column, flip)
unsigned char *buf;
col_table *col_tab;
int new_wid,orig_wid,bwid, column;
int flip;
{
    register int ix, row_ix, col_ix, step, accum, dst_wid, x_mstep, x_error;
    unsigned char *temp_buf, *sP, *dP;
    int tmp, i;
    col_table *col;

    temp_buf = (unsigned char *)malloc(bwid);
    memset(temp_buf, 0, bwid);

    x_mstep = orig_wid / new_wid;
    x_error = orig_wid - x_mstep * new_wid;
    sP = buf;
    dP = temp_buf;
    row_ix = 0;
    col_ix = column%4;
    dst_wid = orig_wid;
    accum = new_wid >> 1;

    while ( 0 < dst_wid )
    {
        step = x_mstep;
        accum -= x_error;
        if (accum < 0) 
        {
            accum += new_wid;
            step++;
        }
        if ( dst_wid < step )
            step = dst_wid;

        dst_wid -= step;

        tmp = 0xFF;
        for(i=0; i<step; i++)
        {
            if (*sP < tmp)
                tmp = *sP;
            sP++;
            row_ix++;
        }
    
        ix = (row_ix%4)*4 + col_ix;
        col = &col_tab[tmp];
        *(dP + row_ix/2) = (AxPixels[DITHER(ix, col)] & 0x0F);

    }

    /* done */
    return (temp_buf);
}

/****************************************************************************/
unsigned char *
xCexpand4( buf,col_tab,new_wid,orig_wid,bwid, column, flip )
unsigned char *buf;
col_table *col_tab;
int new_wid,orig_wid,bwid, column;
int flip;
{
    register int i, ix, row_ix, col_ix, step, accum, dst_wid, x_mstep, x_error;
    unsigned char *temp_buf, *sP, *dP;
    col_table *col;

    temp_buf = (unsigned char *)malloc(bwid);
    memset(temp_buf, 0, bwid);

    dst_wid = new_wid;
    x_mstep = new_wid/orig_wid;
    x_error = new_wid - x_mstep*orig_wid;
    accum = orig_wid >> 1;
    sP = buf;
    dP = temp_buf;
    col_ix = column%4;

	row_ix = 0;
    while ( 0 < dst_wid )
    {
        step = x_mstep;
        accum -= x_error;
        if (accum < 0) 
        {
            accum += orig_wid;
            ++step;
        }
        if ( dst_wid < step )
            step = dst_wid;

        dst_wid -= step;

        col = &col_tab[*sP];
        for ( i=0; i<step; i++ )
        {
            ix = (row_ix%4)*4 + col_ix;
            *(dP+row_ix / 2) = (AxPixels[DITHER(ix, col)] & 0x0F);
            row_ix++;
        }
        sP++;
    }

    /* done */
    return (temp_buf);
}

/****************************************************************************/
unsigned char *
xCsame1(buf, cmapPixel, new_wid, orig_wid, bwid, column, flip)
unsigned char *buf;
unsigned long *cmapPixel;
int new_wid, orig_wid, bwid, column;
int flip;
{
    unsigned char *temp_buf;
    unsigned char *sP, *dP;
    int row_ix, col_ix, ix, dbit, white_pel;

    /* create a temp scaline */
    temp_buf = (unsigned char *)malloc(bwid);
    memset(temp_buf, 0, bwid);

    dbit = 0;
    sP = buf;
    dP = temp_buf;
    white_pel = AxPixels[1];

    /* copy the src buffer and return */
    for ( row_ix=0; row_ix<orig_wid; row_ix++ )
    {
        if ( cmapPixel[*sP] != white_pel ) 
            *dP |= bitmask[dbit];

        dbit++;
        if ( dbit >= 8 )
        {
            dP++;
            dbit = 0;
        }
        sP++;
    }
    if ( flip )
        flipBits(temp_buf, new_wid);

    /* done */
    return (temp_buf);
}

/****************************************************************************/
unsigned char *
xCcompress1(buf,cmapPixel,new_wid,orig_wid,bwid,column,flip)
unsigned char *buf;
unsigned long *cmapPixel;
int new_wid,orig_wid,bwid,column;
int flip;
{
    register int ix, row_ix, col_ix, step, accum, dst_wid, x_mstep, x_error;
    unsigned char *temp_buf, *sP, *dP;
    int white_pel,dbit;

    temp_buf = (unsigned char *)malloc(bwid);
    memset(temp_buf, 0, bwid);

    x_mstep = orig_wid / new_wid;
    x_error = orig_wid - x_mstep * new_wid;
    sP = buf;
    dP = temp_buf;
    dbit = 0;
    row_ix = 0;
    dst_wid = orig_wid;
    accum = new_wid >> 1;
    white_pel = AxPixels[1];

    while ( 0 < dst_wid )
    {
        step = x_mstep;
        accum -= x_error;
        if (accum < 0) 
        {
            accum += new_wid;
            step++;
        }
        if ( dst_wid < step )
            step = dst_wid;

        dst_wid -= step;

        if ( cmapPixel[*sP] != white_pel ) 
            *dP |= bitmask[dbit];

        dbit++;
        if ( dbit >= 8 )
        {
            dP++;
            dbit = 0;
        }
        sP += step;
    }
    if ( flip )
        flipBits(temp_buf, new_wid);

    /* done */
    return (temp_buf);
}

/****************************************************************************/
unsigned char *
xCexpand1( buf,cmapPixel,new_wid,orig_wid,bwid, column, flip )
unsigned char *buf;
unsigned long *cmapPixel;
int new_wid,orig_wid,bwid, column;
int flip;
{
    register int i, ix, row_ix, step, dbit, accum, dst_wid, x_mstep, x_error;
    unsigned char *temp_buf, *sP, *dP;
    int white_pel, bitset;

    temp_buf = (unsigned char *)malloc(bwid);
    memset(temp_buf, 0, bwid);

    dst_wid = new_wid;
    x_mstep = new_wid/orig_wid;
    x_error = new_wid - x_mstep*orig_wid;
    accum = orig_wid >> 1;
    dbit = 0;
    sP = buf;
    dP = temp_buf;
    white_pel = AxPixels[1];

    while ( 0 < dst_wid )
    {
        step = x_mstep;
        accum -= x_error;
        if (accum < 0) 
        {
            accum += orig_wid;
            ++step;
        }
        if ( dst_wid < step )
            step = dst_wid;

        dst_wid -= step;

        bitset = ( cmapPixel[*sP] == white_pel ) ? FALSE: TRUE;
        if ( bitset )
        {
            for ( i=0; i<step; i++ )
            {
                *dP |= bitmask[dbit];
                dbit++;
            }
        }
        else
            dbit += step;
            
        while ( dbit >= 8 )
        {
            dP++;
            dbit -= 8;
        }
        sP++;
    }

    if ( flip )
        flipBits(temp_buf, new_wid);
    /* done */
    return (temp_buf);
}

/****************************************************************************/
dither_image( bwid,rBuf,gBuf,bBuf,col_ix,row_ix,row_max,col_max )
int  bwid;
int *rBuf;
int *gBuf;
int *bBuf;
int col_ix;
int row_ix,row_max,col_max;
{
    int line1_ix,line2_ix,wid,hyt,i,j,imax,jmax;
    int  *ths_err_r,*ths_err_g,*ths_err_b;
    int  *nxt_err_r,*nxt_err_g,*nxt_err_b;
    int  err_trm_r,err_trm_g,err_trm_b;
    int ix,r1,r2,r_pat;
    col_table col_tab;


    ix = col_ix%3;
    line1_ix = ( (ix+1)%3 ) *bwid;
    line2_ix = ( (ix+2)%3 ) *bwid;

    /* point to current line */
    ths_err_r = &( rBuf[line1_ix] );
    ths_err_g = &( gBuf[line1_ix] );
    ths_err_b = &( bBuf[line1_ix] );

    /* point to next line */
    nxt_err_r = &( rBuf[line2_ix] );
    nxt_err_g = &( gBuf[line2_ix] );
    nxt_err_b = &( bBuf[line2_ix] );

    find_nearest_percent(ths_err_r[row_ix], ths_err_g[row_ix],
                         ths_err_b[row_ix], &col_tab);

    err_trm_r = ( ths_err_r[row_ix] - col_tab.rfg );
    err_trm_g = ( ths_err_g[row_ix] - col_tab.gfg );
    err_trm_b = ( ths_err_b[row_ix] - col_tab.bfg );

    ths_err_r[row_ix] = col_tab.rfg;                 /* current pixel */
    ths_err_g[row_ix] = col_tab.gfg;                 /* current pixel */
    ths_err_b[row_ix] = col_tab.bfg;                 /* current pixel */

    if ( row_ix != row_max )
    {
        ths_err_r[row_ix+1] += ( err_trm_r * 7 )/16.0; /* right pixel */
        ths_err_g[row_ix+1] += ( err_trm_g * 7 )/16.0; /* right pixel */
        ths_err_b[row_ix+1] += ( err_trm_b * 7 )/16.0; /* right pixel */
    }
    if ( col_ix != col_max )
    {
        nxt_err_r[row_ix  ] += ( err_trm_r * 5 )/16.0; /* bottom pixel */
        nxt_err_g[row_ix  ] += ( err_trm_g * 5 )/16.0; /* bottom pixel */
        nxt_err_b[row_ix  ] += ( err_trm_b * 5 )/16.0; /* bottom pixel */
    }
    if (row_ix != 0 || col_ix != col_max )
    {
        nxt_err_r[row_ix-1] += ( err_trm_r * 3 )/16.0; /* bottom left  pixel */
        nxt_err_g[row_ix-1] += ( err_trm_g * 3 )/16.0; /* bottom left  pixel */
        nxt_err_b[row_ix-1] += ( err_trm_b * 3 )/16.0; /* bottom left  pixel */

    }
    if ( row_ix != row_max || col_ix != col_max )
    {
        nxt_err_r[row_ix+1] += ( err_trm_r  )/16.0; /* bottom right pixel */
        nxt_err_g[row_ix+1] += ( err_trm_g  )/16.0; /* bottom right pixel */
        nxt_err_b[row_ix+1] += ( err_trm_b  )/16.0; /* bottom right pixel */
    }
}

/*****************************************************************************
 *
 * create_image_mask - TmScaleImage utility function
 */
unsigned char *
create_data_mask(pid, imP, cmap, tr,xfunc,flip, flag )
int pid;
ge_image_type *imP;
cmap_type *cmap;
ge_image_transform_type *tr;
unsigned char *(*xfunc)();
int flip, flag;
{
    register int i, j, x, y, ix, dy, dx, line_ix,len;
    int orig_wid,orig_hyt,new_hyt,x_mstep,y_mstep,y_error;
    int src_bwid, src_wid, src_hyt;
    int dst_bwid, dst_wid, dst_hyt;
    unsigned char *src_data, *dst_data;
    unsigned char *buf,*buf1, *temp_data;
    unsigned char *temp_src, *final_src,*final_dst;
    register int step, accum, wid, hyt;
    unsigned short *sbit, *dbit;
    register unsigned char *sP, *dP;
    XImage *xP;

    orig_wid = imP->wid;
    orig_hyt = imP->hyt;

    /* get the width height and incrmt of the new, unmasked raster */

    dst_hyt = new_hyt =  tr->hyt;
    dst_wid =  tr->wid;

    /* don't do anything if there is nothing to do */
    if ( dst_wid == 0 || dst_hyt == 0 )
    {
        dst_data = (unsigned char *)0;
        xP = (XImage *)XCreateImage(Dpy, AxVisual,
                                    1, XYBitmap,
                                    0, ( char *)dst_data, 0, 0, 16, 0);
        if ( xP == (XImage *)NULL )
            return(( unsigned char *)0);
        xP->bitmap_unit = xP->bitmap_pad;
        xP->bytes_per_line = (dst_wid + 15)/16 * 2;
    }
    else
    {
        /* allocate a block of memory */
        dst_bwid = (dst_wid+7)/8 + ((dst_wid+7)/8)%2;
        dst_data = (unsigned char *)THIMhugeAlloc(pid, dst_hyt * dst_bwid);
        if ( dst_data == (unsigned char *)NULL )
            return( (unsigned char *)0 );
        xP = (XImage *)XCreateImage(Dpy, AxVisual, 1, XYBitmap, 0,
                                    ( char *)dst_data,
                                    dst_wid, dst_hyt,
                                    16, dst_bwid);
        if ( xP == (XImage *)NULL )
            return(( unsigned char *)0);
        xP->bitmap_unit = xP->bitmap_pad;
        xP->bytes_per_line = dst_bwid;

        /* override bit/byte order */
        xP->byte_order = MSBFirst;
        xP->bitmap_bit_order = MSBFirst;

        /* scan thru the dest space, getting the pel from the source image */
        src_wid = imP->wid;
        src_hyt = imP->hyt;
        dst_bwid = xP->bytes_per_line;
        dst_data = (unsigned char *)xP->data;

        if ( flag )
        {
            src_bwid = imP->data_byte_wid;
            src_data = imP->data;
        }
        else
        {
            src_bwid = imP->mask_byte_wid;
            src_data = imP->mask;
        }
        if ( tr->y_y < 0 ) /* vert flip */
        {
            if ( new_hyt == orig_hyt )
            {
                /* same */
                for ( i = orig_hyt; i> 0; i-- )
                {
                    final_src = (*xfunc)(&src_data[i*src_bwid],
                                        dst_wid, orig_wid,dst_bwid, flip );
                    memcpy(&dst_data[(orig_hyt - i)*dst_bwid], final_src, dst_bwid );
                    free(final_src);
                }
            }
            else if ( new_hyt < orig_hyt )
            {
                dst_hyt = orig_hyt; /* compress */
                y_mstep = orig_hyt/new_hyt;
                y_error = orig_hyt - y_mstep*new_hyt;
                accum   = new_hyt >> 1;
                j = 0;
                i = dst_hyt;
                while ( 0 < dst_hyt ) 
                {
                    /* get the number of scanlines to fold into 1 scanline */
                    step = y_mstep;
                    if ( (accum -= y_error ) < 0 )
                    {
                        accum += new_hyt;
                        ++step;
                    }
                    if ( dst_hyt < step )
                        step = dst_hyt;
    
                    /* decrement the remaining height available */
                    dst_hyt -= step;
    
                    /* copy the topmost source scanline */
                    temp_src = &src_data[i*src_bwid];
                    buf = (unsigned char *)malloc(src_bwid);
                    memcpy(buf, temp_src, src_bwid);
    
                    /* skip the rest of scanlines  */
                    i -= step;
    
                    /* do any x transform */
                    final_src = (*xfunc)( buf,dst_wid, orig_wid,dst_bwid,flip );
                    memcpy(&dst_data[j*dst_bwid], final_src, dst_bwid);
                    j++;
    
                    /* clean up */
                    free(final_src);
                    free(buf);
                }
            }
            else
            {
                dst_hyt = new_hyt; /* expand */
                y_mstep = new_hyt/orig_hyt;
                y_error = new_hyt - y_mstep*orig_hyt;
                accum   = orig_hyt >> 1;
                j = 0;
                i = dst_hyt;
                while ( 0 < dst_hyt )
                {
                    step = y_mstep;
                    if ( (accum -= y_error) < 0 ) 
                    {
                        accum += orig_hyt;
                        ++step;
                    }
                    if ( dst_hyt < step )
                        step = dst_hyt;
    
                    dst_hyt -= step;
    
                    final_src = (*xfunc)(&src_data[i*src_bwid],dst_wid,
                                        orig_wid,dst_bwid, flip );
    
                    while ( 0 <= --step ) 
                    {
                        memcpy(&dst_data[j*dst_bwid], final_src, dst_bwid);
                        j++;
                    }
                    i--;
    
                    free(final_src);
                }
            }
        }
        else
        {
            if ( new_hyt == orig_hyt )
            {
                /* same */
                for ( i=0; i<orig_hyt; i++ )
                {
                    final_src = (*xfunc)(&src_data[i*src_bwid],
                                        dst_wid, orig_wid,dst_bwid, flip );
                    memcpy(&dst_data[i*dst_bwid], final_src, dst_bwid );
                    free(final_src);
                }
            }
            else if ( new_hyt < orig_hyt )
            {
                dst_hyt = orig_hyt; /* compress */
                y_mstep = orig_hyt/new_hyt;
                y_error = orig_hyt - y_mstep*new_hyt;
                accum   = new_hyt >> 1;
                j = i = 0;
                while ( 0 < dst_hyt ) 
                {
                    /* get the number of scanlines to fold into 1 scanline */
                    step = y_mstep;
                    if ( (accum -= y_error ) < 0 )
                    {
                        accum += new_hyt;
                        ++step;
                    }
                    if ( dst_hyt < step )
                        step = dst_hyt;
        
                    /* decrement the remaining height available */
                    dst_hyt -= step;
        
                    /* copy the topmost source scanline */
                    temp_src = &src_data[i*src_bwid];
                    buf = (unsigned char *)malloc(src_bwid);
                    memcpy(buf, temp_src, src_bwid);
        
                    /* skip the rest of scanlines  */
                    i += step;
        
                    /* do any x transform */
                    final_src = (*xfunc)( buf,dst_wid, orig_wid,dst_bwid,flip );
                    memcpy(&dst_data[j*dst_bwid], final_src, dst_bwid);
                    j++;
        
                    /* clean up */
                    free(final_src);
                    free(buf);
                }
            }
            else
            {
                dst_hyt = new_hyt; /* expand */
                y_mstep = new_hyt/orig_hyt;
                y_error = new_hyt - y_mstep*orig_hyt;
                accum   = orig_hyt >> 1;
                j = i = 0;
                while ( 0 < dst_hyt )
                {
                    step = y_mstep;
                    if ( (accum -= y_error) < 0 ) 
                    {
                        accum += orig_hyt;
                        ++step;
                    }
                    if ( dst_hyt < step )
                        step = dst_hyt;
        
                    dst_hyt -= step;
        
                    final_src = (*xfunc)(&src_data[i*src_bwid],dst_wid,
                                        orig_wid,dst_bwid, flip );
        
                    while ( 0 <= --step ) 
                    {
                        memcpy(&dst_data[j*dst_bwid], final_src, dst_bwid);
                        j++;
                    }
                    i++;
        
                    free(final_src);
                }
            }
        }
    }
    return((unsigned char *)xP);
}

/*****************************************************************************
 *
 * tr_fst_sys_mono_image - TmScaleImage utility function
 */
unsigned char *
tr_fst_sys_mono_image(pid, imP, cmap, tr)
int pid;
ge_image_type *imP;
cmap_type *cmap;
ge_image_transform_type *tr;
{
    XImage *xP;
    unsigned char *(*xMfunc)();
    int flip;
    
    if ( tr->x_x < 0 ) 
        flip = TRUE;
    else
        flip = FALSE;

    if ( tr->wid == imP->wid )
        xMfunc = xMsame1;
    else if ( tr->wid < imP->wid )
        xMfunc = xMcompress1;
    else 
        xMfunc = xMexpand1;

    xP = (XImage *)0;
    if ( imP->data )
        xP = (XImage  *)create_data_mask(pid, imP, cmap, tr,xMfunc,flip, TRUE );
    return((unsigned char *)xP);
}

/*****************************************************************************
 *
 * tr_fst_sys_img_msk - TmScaleImage utility function
 */
unsigned char *
tr_fst_sys_img_msk(pid, imP, cmap, tr )
int pid;
ge_image_type *imP;
cmap_type *cmap;
ge_image_transform_type *tr;
{
    XImage *xP;
    unsigned char *(*xMfunc)();
    int flip;

    if ( tr->x_x < 0 ) 
        flip = TRUE;
    else
        flip = FALSE;

    if ( tr->wid == imP->wid )
        xMfunc = xMsame1;
    else if ( tr->wid < imP->wid )
        xMfunc = xMcompress1;
    else 
        xMfunc = xMexpand1;

    xP = (XImage *)0;
    if ( imP->mask )
        xP = (XImage  *)create_data_mask(pid, imP, cmap, tr,xMfunc,flip,FALSE );
    return((unsigned char *)xP);
}

/****************************************************************************/
unsigned char *
tr_fst_sys_col1_image(pid, imP, cmap, tr,xfunc, flip_flag )
int pid;
ge_image_type *imP;
cmap_type *cmap;
ge_image_transform_type *tr;
unsigned char *(*xfunc)();
int flip_flag;
{
    register int i, j,k, step, accum,sbit, dbit;
    int orig_wid,orig_hyt,new_hyt,y_mstep,y_error, r,g,b;
    int src_bwid, src_wid, src_hyt, dst_bwid, dst_wid, dst_hyt;
    unsigned long cmapPixels[256];
    unsigned long pel, white_pel;
    XImage *xP;
    unsigned char *src_data, *dst_data, *temp_src, *final_src, *buf;

    orig_hyt = imP->hyt;
    orig_wid = imP->wid;
    /* get the width and height of the new, unmasked raster */
    dst_hyt = new_hyt = tr->hyt;
    dst_wid = tr->wid;


    /* allocate a block of memory */
    dst_bwid = (dst_wid+7)/8 + ((dst_wid+7)/8)%2;
    dst_data = (unsigned char *)THIMhugeAlloc(pid, dst_hyt * dst_bwid);
    if ( dst_data == (unsigned char *)NULL )
        return ( (unsigned char *)0 );
    xP = (XImage *)XCreateImage(Dpy, AxVisual,
                                1, XYBitmap, 0, (char *)dst_data,
                                dst_wid, dst_hyt,
                                16, dst_bwid);
    if ( xP == (XImage *)NULL )
        return(( unsigned char *)0);
    xP->bitmap_unit = xP->bitmap_pad;
    xP->bytes_per_line = dst_bwid;

    /* override bit/byte order */
    xP->byte_order = MSBFirst;
    xP->bitmap_bit_order = MSBFirst;

    /* create an index to pixel table */
    white_pel = AxPixels[1];
    for ( i=0; i<cmap->n_colors && i<256; i++ ) 
    {
        if ( cmap->color[i].see_thru == TRUE )
            cmapPixels[i] = white_pel;
        else 
            AxGetRealPixel(2, cmap->color[i].c, cmap->color[i].m,
                           cmap->color[i].y, cmap->color[i].k,
                           255, &cmapPixels[i]);
    }

    /* Dither image for display */
    src_data = imP->data;
    src_bwid = imP->data_byte_wid;

    if ( new_hyt == orig_hyt )
    {
        dst_hyt = orig_hyt;
        for ( i=0;i<dst_hyt;i++ )
        {
            final_src = (*xfunc)(&src_data[i*src_bwid],cmapPixels,
                                    dst_wid, orig_wid, dst_bwid, i, flip_flag);
            memcpy(&dst_data[i*dst_bwid], final_src, dst_bwid );
            free(final_src);
        }
    }
    else if ( new_hyt < orig_hyt )
    {
        dst_hyt = orig_hyt; /* compress */
        y_mstep = orig_hyt/new_hyt;
        y_error = orig_hyt - y_mstep*new_hyt;
        accum   = new_hyt >> 1;
        j = i = 0;
        while ( 0 < dst_hyt ) 
        {
            /* get the number of scanlines to fold into 1 scanline */
            step = y_mstep;
            if ( (accum -= y_error ) < 0 )
            {
                accum += new_hyt;
                ++step;
            }
            if ( dst_hyt < step )
                step = dst_hyt;

            /* decrement the remaining height available */
            dst_hyt -= step;

            /* copy the topmost source scanline */
            temp_src = &src_data[i*src_bwid];
            buf = (unsigned char *)malloc(src_bwid);
            memcpy(buf, temp_src, src_bwid);
    
            /* skip the rest of scanlines  */
            i += step;

            /* do any x transform */
            final_src = (*xfunc)(&src_data[i*src_bwid],cmapPixels,
                                    dst_wid, orig_wid, dst_bwid, i, flip_flag);
            memcpy(&dst_data[j*dst_bwid], final_src, dst_bwid);
            j++;
    
            /* clean up */
            free(final_src);
            free(buf);
        }
    }
    else
    {
        dst_hyt = new_hyt; /* expand */
        y_mstep = new_hyt/orig_hyt;
        y_error = new_hyt - y_mstep*orig_hyt;
        accum   = orig_hyt >> 1;
        j = i = 0;
        while ( 0 < dst_hyt )
        {
            step = y_mstep;
            if ( (accum -= y_error) < 0 ) 
            {
                accum += orig_hyt;
                ++step;
            }
            if ( dst_hyt < step )
                step = dst_hyt;
    
            dst_hyt -= step;
    
            final_src = (*xfunc)(&src_data[i*src_bwid],cmapPixels,
                                    dst_wid, orig_wid, dst_bwid, i, flip_flag);
    
            while ( 0 <= --step ) 
            {
                memcpy(&dst_data[j*dst_bwid], final_src, dst_bwid);
                j++;
            }
            i++;
    
            free(final_src);
        }
    }
    return((unsigned char *)xP);
}

/****************************************************************************/
unsigned char *
tr_fst_sys_col4_image(pid, imP, cmap, tr,xfunc, flip_flag )
int pid;
ge_image_type *imP;
cmap_type *cmap;
ge_image_transform_type *tr;
unsigned char *(*xfunc)();
int flip_flag;
{
    register int i, j,k, step, accum,sbit, dbit;
    int orig_wid,orig_hyt,new_hyt,y_mstep,y_error, r,g,b;
    int src_bwid, src_wid, src_hyt, dst_bwid, dst_wid, dst_hyt;
    unsigned char *src_data, *dst_data, *temp_src, *final_src, *buf;
    col_table col_tab[256];
    color_type *tmp_color;
    XImage *xP;


    orig_wid = imP->wid;
    orig_hyt = imP->hyt;

    /* get the width height and incrmt of the new, unmasked raster */

    dst_hyt = new_hyt = tr->hyt;
    dst_wid = tr->wid;

    /* allocate a block of memory */
    dst_bwid = (dst_wid+1)/2 + ((dst_wid+1) / 2)%2;
    dst_data = (unsigned char *)THIMhugeAlloc(pid, dst_hyt * dst_bwid);
    if ( dst_data == (unsigned char *)NULL )
        return ( (unsigned char *)0 );

    xP = (XImage *)XCreateImage(Dpy, AxVisual,
                                8, ZPixmap,
                                0, ( char *)dst_data,
                                dst_wid, dst_hyt,
                                16, dst_bwid);
    if ( xP == (XImage *)NULL )
        return(( unsigned char *)0);

    xP->bitmap_unit = xP->bitmap_pad;
    xP->bytes_per_line = dst_bwid;

    /* override bit/byte order */
    xP->byte_order = MSBFirst;
    xP->bitmap_bit_order = MSBFirst;
    src_data = imP->data;
    src_bwid = imP->data_byte_wid;
    dst_bwid = xP->bytes_per_line;
    
    /* create an index to pixel table */
    for ( i=0; i<cmap->n_colors && i<256; i++ ) 
    {
        tmp_color = &cmap->color[i];
        if ( tmp_color->see_thru == TRUE )
        {
            memset((char *)&col_tab[i], 0, sizeof(col_table));
            col_tab[i].rfg  = 180;
            col_tab[i].gfg  = 30;
            col_tab[i].bfg  = 5;
        }
        else 
        {
            /* Convert cmyk to rgb */
            k = tmp_color->k;
            r = 255 - MIN(255, tmp_color->c + k);
            g = 255 - MIN(255, tmp_color->m + k);
            b = 255 - MIN(255, tmp_color->y + k);
            find_nearest_cmapindex(r,g,b,&col_tab[i],FALSE);
        }
    }   
    if ( new_hyt == orig_hyt )
    {
        dst_hyt = orig_hyt; /* same */
        for ( i=0; i<dst_hyt;i++ )
        {
            final_src = (*xfunc)(&src_data[i*src_bwid],col_tab,
                                     dst_wid, orig_wid, dst_bwid, i, flip_flag);
            memcpy(&dst_data[i*dst_bwid], final_src, dst_bwid );
            free(final_src);
        }
    }
    else if ( new_hyt < orig_hyt )
    {
        dst_hyt = orig_hyt; /* compress */
        y_mstep = orig_hyt/new_hyt;
        y_error = orig_hyt - y_mstep*new_hyt;
        accum   = new_hyt >> 1;
        j = i =0;
        while ( 0 < dst_hyt ) 
        {
            step = y_mstep;
            if ( (accum -= y_error ) < 0 )
            {
                accum += new_hyt;
                ++step;
            }
            if ( dst_hyt < step )
                step = dst_hyt;
            dst_hyt -= step;
            temp_src = &src_data[i*src_bwid];
            buf = (unsigned char *)malloc(src_bwid);
            memcpy(buf,temp_src,src_bwid);

            i += step;

            final_src = (*xfunc)(buf,col_tab,dst_wid,
                                    orig_wid, dst_bwid,i, flip_flag );
            memcpy(&dst_data[j*dst_bwid], final_src, dst_bwid );
            j++;
            /* clean up */
            free(final_src);
            free(buf);
        }
    }
    else
    {
        dst_hyt = new_hyt; /* expand */
        y_mstep = new_hyt/orig_hyt;
        y_error = new_hyt - y_mstep*orig_hyt;
        accum   = orig_hyt >> 1;
        j = i = 0;
        while ( 0 < dst_hyt )
        {
            step = y_mstep;
            if ( ( accum -= y_error ) < 0) 
            {
                accum += orig_hyt;
                ++step;
            }
            if ( dst_hyt < step )
                step = dst_hyt;
            dst_hyt -= step;
            temp_src = &src_data[i*src_bwid];
            final_src = (*xfunc)( temp_src,col_tab,dst_wid,
                                     orig_wid,dst_bwid,i, flip_flag );
            while ( 0 <= --step ) 
            {
                memcpy(&dst_data[j*dst_bwid], final_src, dst_bwid );
                j++;
            }
            i++;
            free(final_src);
        }
    }
    return((unsigned char *)xP);
}

/****************************************************************************/
unsigned char *
tr_fst_sys_col8_image(pid, imP, cmap, tr,xfunc, flip_flag )
int pid;
ge_image_type *imP;
cmap_type *cmap;
ge_image_transform_type *tr;
unsigned char *(*xfunc)();
int flip_flag;
{
    register int i, j,k, step, accum,sbit, dbit;
    int orig_wid,orig_hyt,new_hyt,y_mstep,y_error, r,g,b;
    int src_bwid, src_wid, src_hyt, dst_bwid, dst_wid, dst_hyt;
    unsigned char *src_data, *dst_data, *temp_src, *final_src, *buf;
    col_table col_tab[256];
    color_type *tmp_color;
    XImage *xP;
    int max, mult;

    orig_wid = imP->wid;
    orig_hyt = imP->hyt;

    /* get the width height and incrmt of the new, unmasked raster */

    dst_hyt = new_hyt = tr->hyt;
    dst_wid = tr->wid;

    /* allocate a block of memory */
    dst_bwid = dst_wid + dst_wid%2;
    dst_data = (unsigned char *)THIMhugeAlloc(pid, dst_hyt * dst_bwid);
    if ( dst_data == (unsigned char *)NULL )
        return ( (unsigned char *)0 );

    xP = (XImage *)XCreateImage(Dpy, AxVisual,
                                8, ZPixmap,
                                0, ( char *)dst_data,
                                dst_wid, dst_hyt,
                                16, dst_bwid);
    if ( xP == (XImage *)NULL )
        return(( unsigned char *)0);

    xP->bitmap_unit = xP->bitmap_pad;
    xP->bytes_per_line = dst_bwid;

    /* override bit/byte order */
    xP->byte_order = MSBFirst;
    xP->bitmap_bit_order = MSBFirst;
    src_data = imP->data;
    src_bwid = imP->data_byte_wid;
    dst_bwid = xP->bytes_per_line;
    
    /* create an index to pixel table */
    if ( axPreDefinedStdmap )
        axConvertStdIndexToPixel(col_tab, cmap);
    else /* use our own fixed position colormap */
    {
        /* create an index to pixel table */
        for ( i=0; i<cmap->n_colors && i<256; i++ ) 
        {
            tmp_color = &cmap->color[i];
            if ( tmp_color->see_thru == TRUE )
            {
                /* find index 215 - should be white */
                memset((char *)&col_tab[i], 0, sizeof(col_table));
                col_tab[i].rfg = 180;
                col_tab[i].gfg = 30;
                col_tab[i].bfg = 5;
            }
            else 
            {
                /* Convert cmyk to rgb */
                k = tmp_color->k;
                r = 255 - MIN(255, tmp_color->c + k);
                g = 255 - MIN(255, tmp_color->m + k);
                b = 255 - MIN(255, tmp_color->y + k);
                find_nearest_cmapindex(r,g,b, &col_tab[i], FALSE);
            }
        }   
    }

    if ( tr->y_y < 0 )  /* flip vert */
    {
        if ( new_hyt == orig_hyt )
        {
            dst_hyt = orig_hyt; /* same */
            for ( i=dst_hyt; i>0;i-- )
            {
                final_src = (*xfunc)(&src_data[i*src_bwid], col_tab,
                                    dst_wid, orig_wid, dst_bwid, i, flip_flag);
                memcpy(&dst_data[(dst_hyt -i)*dst_bwid], final_src, dst_bwid );
                free(final_src);
            }
        }
        else if ( new_hyt < orig_hyt )
        {
            dst_hyt = orig_hyt; /* compress */
            y_mstep = orig_hyt/new_hyt;
            y_error = orig_hyt - y_mstep*new_hyt;
            accum   = new_hyt >> 1;
            j = 0;
            i = dst_hyt;
            while ( 0 < dst_hyt ) 
            {
                step = y_mstep;
                if ( (accum -= y_error ) < 0 )
                {
                    accum += new_hyt;
                    ++step;
                }
                if ( dst_hyt < step )
                    step = dst_hyt;
                dst_hyt -= step;
                temp_src = &src_data[i*src_bwid];
                buf = (unsigned char *)malloc(src_bwid);
                memcpy(buf,temp_src,src_bwid);
    
                i -= step;
    
                final_src = (*xfunc)(buf,col_tab,dst_wid,
                                        orig_wid, dst_bwid,i, flip_flag );
                memcpy(&dst_data[j*dst_bwid], final_src, dst_bwid );
                j++;
                /* clean up */
                free(final_src);
                free(buf);
            }
        }
        else
        {
            dst_hyt = new_hyt; /* expand */
            y_mstep = new_hyt/orig_hyt;
            y_error = new_hyt - y_mstep*orig_hyt;
            accum   = orig_hyt >> 1;
            j = 0;
            i = dst_hyt;
            while ( 0 < dst_hyt )
            {
                step = y_mstep;
                if ( ( accum -= y_error ) < 0) 
                {
                    accum += orig_hyt;
                    ++step;
                }
                if ( dst_hyt < step )
                    step = dst_hyt;
                dst_hyt -= step;
                temp_src = &src_data[i*src_bwid];
                final_src = (*xfunc)( temp_src,col_tab,dst_wid,
                                        orig_wid,dst_bwid,i, flip_flag );
                while ( 0 <= --step ) 
                {
                    memcpy(&dst_data[j*dst_bwid], final_src, dst_bwid );
                    j++;
                }
                i--;
                free(final_src);
            }
        }
    }
    else
    {
        if ( new_hyt == orig_hyt )
        {
            dst_hyt = orig_hyt; /* same */
            for ( i=0; i<dst_hyt;i++ )
            {
                final_src = (*xfunc)(&src_data[i*src_bwid],col_tab,
                                    dst_wid, orig_wid, dst_bwid, i, flip_flag);
                memcpy(&dst_data[i*dst_bwid], final_src, dst_bwid );
                free(final_src);
            }
        }
        else if ( new_hyt < orig_hyt )
        {
            dst_hyt = orig_hyt; /* compress */
            y_mstep = orig_hyt/new_hyt;
            y_error = orig_hyt - y_mstep*new_hyt;
            accum   = new_hyt >> 1;
            j = i =0;
            while ( 0 < dst_hyt ) 
            {
                step = y_mstep;
                if ( (accum -= y_error ) < 0 )
                {
                    accum += new_hyt;
                    ++step;
                }
                if ( dst_hyt < step )
                    step = dst_hyt;
                dst_hyt -= step;
                temp_src = &src_data[i*src_bwid];
                buf = (unsigned char *)malloc(src_bwid);
                memcpy(buf,temp_src,src_bwid);
    
                i += step;
    
                final_src = (*xfunc)(buf,col_tab,dst_wid,
                                        orig_wid, dst_bwid,i, flip_flag );
                memcpy(&dst_data[j*dst_bwid], final_src, dst_bwid );
                j++;
                /* clean up */
                free(final_src);
                free(buf);
            }
        }
        else
        {
            dst_hyt = new_hyt; /* expand */
            y_mstep = new_hyt/orig_hyt;
            y_error = new_hyt - y_mstep*orig_hyt;
            accum   = orig_hyt >> 1;
            j = i = 0;
            while ( 0 < dst_hyt )
            {
                step = y_mstep;
                if ( ( accum -= y_error ) < 0) 
                {
                    accum += orig_hyt;
                    ++step;
                }
                if ( dst_hyt < step )
                    step = dst_hyt;
                dst_hyt -= step;
                temp_src = &src_data[i*src_bwid];
                final_src = (*xfunc)( temp_src,col_tab,dst_wid,
                                        orig_wid, dst_bwid, i,flip_flag);
                while ( 0 <= --step ) 
                {
                    memcpy(&dst_data[j*dst_bwid], final_src, dst_bwid );
                    j++;
                }
                i++;
                free(final_src);
            }
        }
    }
    return((unsigned char *)xP);
}

/****************************************************************************/
unsigned char *
tr_fst_sys_col24_image(pid, imP, cmap, tr,xfunc, flip_flag)
int pid;
ge_image_type *imP;
cmap_type *cmap;
ge_image_transform_type *tr;
unsigned char *(*xfunc)();
int flip_flag;
{
    register int i, j,k, step, accum,sbit, dbit;
    int orig_wid,orig_hyt,new_hyt,y_mstep,y_error, r,g,b;
    int src_bwid, src_wid, src_hyt, dst_bwid, dst_wid, dst_hyt;
    unsigned long cmapPixels[256];
    unsigned long pel, white_pel;
    XImage *xP;
    unsigned char *src_data, *dst_data, *temp_src, *buf;
    int rm,gm,bm, rs,gs,bs;
    color_type *tmp_color;

    orig_hyt = imP->hyt;
    orig_wid = imP->wid;
    /* get the width and height of the new, unmasked raster */
    dst_hyt = new_hyt = tr->hyt;
    dst_wid = tr->wid;

    white_pel = AxPixels[1];

    /* allocate a block of memory */
    dst_bwid = 4*dst_wid;
    dst_data = (unsigned char *)THIMhugeAlloc(pid, dst_hyt * dst_bwid);
    if ( dst_data == (unsigned char *)NULL )
        return ( (unsigned char *)0 );
    xP = (XImage *)XCreateImage(Dpy, AxVisual,
                                DpyDepth, ZPixmap, 0, (char *)dst_data,
                                dst_wid, dst_hyt,
                                32, dst_bwid);
    if ( xP == (XImage *)NULL )
        return(( unsigned char *)0);
    xP->bitmap_unit = xP->bitmap_pad;
    xP->bytes_per_line = dst_bwid;


    /* create an index to pixel table */
    rm = xP->red_mask;
    gm = xP->green_mask;
    bm = xP->blue_mask;
#if AX_X11
	if (0 == rm && 0 == gm && 0 == bm) {
		/* cheat, and read masks from display directly */
		Screen *ss = XScreenOfDisplay(Dpy, 0);
		Visual *vv = ss->root_visual;

		rm = vv->red_mask;
		gm = vv->green_mask;
		bm = vv->blue_mask;

		xP->red_mask = rm;
		xP->green_mask = gm;
		xP->blue_mask = bm;
		
	}
#endif
    rs = ExtractShift_from_mask(rm);
    gs = ExtractShift_from_mask(gm);
    bs = ExtractShift_from_mask(bm);
    
    for ( i=0; i<cmap->n_colors && i<256; i++ ) 
    {
        if ( cmap->color[i].see_thru == TRUE )
            cmapPixels[i] = white_pel;
        else 
        {
            r = 255 - MIN(255,(cmap->color[i].c+cmap->color[i].k));
            g = 255 - MIN(255,(cmap->color[i].m+cmap->color[i].k));
            b = 255 - MIN(255,(cmap->color[i].y+cmap->color[i].k));
            cmapPixels[i] = ((r << rs) & rm) | (( g << gs) & gm) |
                                                ((b << bs ) & bm);
        }
    }

    src_data = imP->data;
    src_bwid = imP->data_byte_wid;

    if ( tr->y_y < 0 )  /* flip vert */
    {
        if ( new_hyt == orig_hyt )
        {
            dst_hyt = orig_hyt;
            for ( i=dst_hyt;i>0;i-- )
                (*xfunc)(&src_data[i*src_bwid],xP,cmapPixels, 
                         dst_wid, orig_wid, dst_bwid, i, flip_flag);
        }
        else if ( new_hyt < orig_hyt )
        {
            dst_hyt = orig_hyt; /* compress */
            y_mstep = orig_hyt/new_hyt;
            y_error = orig_hyt - y_mstep*new_hyt;
            accum   = new_hyt >> 1;
            j = 0;
            i = dst_hyt;
            while ( 0 < dst_hyt ) 
            {
                /* get the number of scanlines to fold into 1 scanline */
                step = y_mstep;
                if ( (accum -= y_error ) < 0 )
                {
                    accum += new_hyt;
                    ++step;
                }
                if ( dst_hyt < step )
                    step = dst_hyt;

                /* decrement the remaining height available */
                dst_hyt -= step;

                temp_src = &src_data[i*src_bwid];
        
                /* skip the rest of scanlines  */
                i -= step;

                /* do any x transform */
                (*xfunc)(&src_data[i*src_bwid],xP,cmapPixels,
                         dst_wid, orig_wid, dst_bwid, i, flip_flag);
                j++;
            }
        }
        else
        {
            dst_hyt = new_hyt; /* expand */
            y_mstep = new_hyt/orig_hyt;
            y_error = new_hyt - y_mstep*orig_hyt;
            accum   = orig_hyt >> 1;
            j = 0;
            i = dst_hyt;
            while ( 0 < dst_hyt )
            {
                step = y_mstep;
                if ( (accum -= y_error) < 0 ) 
                {
                    accum += orig_hyt;
                    ++step;
                }
                if ( dst_hyt < step )
                    step = dst_hyt;
        
                dst_hyt -= step;
        
                (*xfunc)(&src_data[i*src_bwid],xP,cmapPixels,
                          dst_wid, orig_wid, dst_bwid, i, flip_flag);
                while ( 0 <= --step )
                {
                    memcpy(&dst_data[j*dst_bwid],
                           &dst_data[i*dst_bwid], dst_bwid );
                    j++;
                }
                i++;
        
                i--;
            }
        }
    }
    else
    {
        if ( new_hyt == orig_hyt )
        {
            dst_hyt = orig_hyt; /* same */
            for ( i=0;i<dst_hyt;i++ )
            {
                (*xfunc)(&src_data[i*src_bwid],xP,cmapPixels, 
                         dst_wid, orig_wid, dst_bwid, i, flip_flag);
            }
        }
        else if ( new_hyt < orig_hyt )
        {
            dst_hyt = orig_hyt; /* compress */
            y_mstep = orig_hyt/new_hyt;
            y_error = orig_hyt - y_mstep*new_hyt;
            accum   = new_hyt >> 1;
            j = i = 0;
            while ( 0 < dst_hyt ) 
            {
                /* get the number of scanlines to fold into 1 scanline */
                step = y_mstep;
                if ( (accum -= y_error ) < 0 )
                {
                    accum += new_hyt;
                    ++step;
                }
                if ( dst_hyt < step )
                    step = dst_hyt;

                /* decrement the remaining height available */
                dst_hyt -= step;

                temp_src = &src_data[i*src_bwid];
        
                /* skip the rest of scanlines  */
                i += step;

                /* do any x transform */
                (*xfunc)(&src_data[i*src_bwid],xP,cmapPixels,
                         dst_wid, orig_wid, dst_bwid, i, flip_flag);
                j++;
            }
        }
        else
        {
            dst_hyt = new_hyt; /* expand */
            y_mstep = new_hyt/orig_hyt;
            y_error = new_hyt - y_mstep*orig_hyt;
            accum   = orig_hyt >> 1;
            j = i = 0;
            while ( 0 < dst_hyt )
            {
                step = y_mstep;
                if ( (accum -= y_error) < 0 ) 
                {
                    accum += orig_hyt;
                    ++step;
                }
                if ( dst_hyt < step )
                    step = dst_hyt;
        
                dst_hyt -= step;
                (*xfunc)(&src_data[i*src_bwid],xP,cmapPixels,
                      dst_wid,orig_wid,dst_bwid,i,flip_flag);
                while ( 0 <= --step )
                {
                    (*xfunc)(&src_data[i*src_bwid], xP, cmapPixels,
                          dst_wid, orig_wid, dst_bwid, j, flip_flag);
                    j++;
                }
                i++;
            }
        }
    }
    return((unsigned char *)xP);
}

/*****************************************************************************
 *
 * tr_fst_sys_color_image - TmScaleImage utility function
 */
unsigned char *
tr_fst_sys_color_image(pid, imP, cmap, tr )
int pid;
ge_image_type *imP;
cmap_type *cmap;
ge_image_transform_type *tr;
{
    register int ix;
    XImage *xP;
    unsigned char *xBuf;
    unsigned char *(*xCfunc)();
    int flip;


    /* pad the cmap to be atleast 256 */
    if ( cmap->n_colors < 256 )
    {
        ix = cmap->n_colors;
        cmap->color = (color_type *)TaskRealloc( pid, cmap->color,
                                    256 * sizeof(color_type) );
        for ( ; ix<256; ix++ )
        {
            cmap->color[ix].see_thru = FALSE;
            cmap->color[ix].ink_type = 0;
            cmap->color[ix].c = 0;
            cmap->color[ix].m = 0;
            cmap->color[ix].y = 0;
            cmap->color[ix].k = 0;
            cmap->color[ix].name = (char *) TaskAlloc(pid, 9);
            sprintf(cmap->color[ix].name, "color%03d", ix);
        }
        cmap->n_colors = 256;
    }
    /* check for flip */
    if ( tr->x_x < 0 ) 
        flip = TRUE;
    else
        flip = FALSE;

    /* a zero size image is created as a XYBitmap with width and height 0 */
    if ( tr->wid == 0 || tr->hyt == 0 )
    {
        xBuf = (unsigned char *)0;
        xP = (XImage *)XCreateImage(Dpy, AxVisual,
                                    1, XYBitmap,
                                    0, xBuf,
                                    0, 0,
                                    8, 0);
        if ( xP == (XImage *)NULL )
            return(( unsigned char *)0);
        xP->bitmap_unit = xP->bitmap_pad;
        xP->bytes_per_line = (tr->wid + tr->wid%2);
    }

    /* create an 8 deep image */
    else if ( DpyDepth == 8 )
    {
        if ( tr->wid == imP->wid )
            xCfunc = xCsame8;
        else if ( tr->wid < imP->wid )
            xCfunc = xCcompress8;
        else 
            xCfunc = xCexpand8;
        xP = (XImage *)tr_fst_sys_col8_image( pid, imP, cmap, tr,xCfunc, flip );
    }
    
    /* create a 4 deep image  */
    else if ( DpyDepth == 4 ) 
    {
        if ( tr->wid == imP->wid )
            xCfunc = xCsame4;
        else if ( tr->wid < imP->wid )
            xCfunc = xCcompress4;
        else 
            xCfunc = xCexpand4;

        xP = (XImage *)tr_fst_sys_col4_image( pid, imP, cmap, tr,xCfunc, flip );
    }
    /* create a TrueColor/DirectColor DpyDepth image */
    else if ( DpyDepth > 8 )
    {
        if ( tr->wid == imP->wid )
            xCfunc = xCsame24;
        else if ( tr->wid < imP->wid )
            xCfunc = xCcompress24;
        else 
            xCfunc = xCexpand24;

        xP = (XImage *)tr_fst_sys_col24_image( pid, imP, cmap,tr,xCfunc, flip);
    }
    /* create a 1 deep image */
    else
    {
        if ( tr->wid == imP->wid )
            xCfunc = xCsame1;
        else if ( tr->wid < imP->wid )
            xCfunc = xCcompress1;
        else 
            xCfunc = xCexpand1;
        xP =  (XImage *)tr_fst_sys_col1_image( pid,imP, cmap,tr,xCfunc, flip);
    }
    return ( (unsigned char *)xP );
}

/*****************************************************************************
 *
 * tr_sys_image_mask - TmScaleImage utility function
 */
unsigned char *
tr_sys_image_mask(pid, imP, cmap, tr)
int pid;
ge_image_type *imP;
cmap_type *cmap;
ge_image_transform_type *tr;
{
    register int i, j, x, y, ix, line_ix;
    double x_xScale, x_yScale, x_Trans;
    double y_xScale, y_yScale, y_Trans;
    int src_bwid, src_wid, src_hyt;
    int dst_bwid, dst_wid, dst_hyt;
    unsigned char *src_data, *dst_data;
    XImage *xP;

    /* if there is no mask, do nothing */
    if ( imP->mask == (unsigned char *)0 )
        return((unsigned char *)0);

    /* get the width and height of the new, unmasked raster */
    dst_wid = tr->wid;
    dst_hyt = tr->hyt;

    /* don't do anything if there is nothing to do */
    if ( dst_wid == 0 || dst_hyt == 0 )
    {
        dst_data = (unsigned char *)0;
        xP = (XImage *)XCreateImage(Dpy, AxVisual,
                                    1, XYBitmap,
                                    0, dst_data,
                                    0, 0,
                                    16, 0);
        if ( xP == (XImage *)NULL )
            return(( unsigned char *)0);
        xP->bitmap_unit = xP->bitmap_pad;
        xP->bytes_per_line = (dst_wid + 15)/16 * 2;
    }
    else
    {
        /* allocate a block of memory */
        dst_bwid = (dst_wid+7)/8 + ((dst_wid+7)/8)%2;
        dst_data = (unsigned char *)THIMhugeAlloc(pid, dst_hyt * dst_bwid);
        if ( dst_data == (unsigned char *)NULL )
        {
            return((unsigned char *)0);
        }
        xP = (XImage *)XCreateImage(Dpy, AxVisual,
                                    1, XYBitmap,
                                    0, dst_data,
                                    dst_wid, dst_hyt,
                                    16, dst_bwid);
        if ( xP == (XImage *)NULL )
            return(( unsigned char *)0);
        xP->bitmap_unit = xP->bitmap_pad;
        xP->bytes_per_line = dst_bwid;

        /* override bit/byte order */
        xP->byte_order = MSBFirst;
        xP->bitmap_bit_order = MSBFirst;

        /* set up the transfrom factors */
        x_xScale = tr->x_x;
        x_yScale = tr->x_y;
        y_xScale = tr->y_x;
        y_yScale = tr->y_y;
        x_Trans  = tr->x_t;
        y_Trans  = tr->y_t;

        /* scan thru the dest space, getting the pel from the source image */
        src_bwid = imP->mask_byte_wid;
        dst_bwid = xP->bytes_per_line;
        src_data = (unsigned char *)imP->mask;
        dst_data = (unsigned char *)xP->data;
        src_wid = imP->wid;
        src_hyt = imP->hyt;
        for ( i=0; i<dst_hyt; i++ )
        {
            line_ix = i*dst_bwid;
            for (j=0; j<dst_wid; j++ )
            {
                /* get the index for the pixel in the source image */
                x = j*x_xScale + i*x_yScale + x_Trans;
                y = i*y_yScale - j*y_xScale + y_Trans;

                /* if this value is valid update the pixel */
                if ( x >= 0 && x < src_wid && y >= 0 && y < src_hyt )
                {
                    ix = y*src_bwid+(x>>3);
                    if ( src_data[ix] & bitmask[x&7] )
                        dst_data[line_ix+(j>>3)] |= bitmask[j&7];
                    else
                        dst_data[line_ix+(j>>3)] &= ~bitmask[j&7];
                }
                else
                    dst_data[line_ix+(j>>3)] &= ~bitmask[j&7];
            }
        }
    }
    return((unsigned char *)xP);
}

/*****************************************************************************
 *
 * tr_sys_mono_image - TmScaleImage utility function
 */
unsigned char *
tr_sys_mono_image(pid, imP, cmap, tr)
int pid;
ge_image_type *imP;
cmap_type *cmap;
ge_image_transform_type *tr;
{
    register int i, j, x, y, ix, line_ix;
    double x_xScale, x_yScale, x_Trans;
    double y_xScale, y_yScale, y_Trans;
    int src_bwid, src_wid, src_hyt;
    int dst_bwid, dst_wid, dst_hyt;
    unsigned char *src_data, *dst_data;
    XImage *xP;

    /* get the width and height of the new, unmasked raster */
    dst_wid = tr->wid;
    dst_hyt = tr->hyt;

    /* don't do anything if there is nothing to do */
    if ( dst_wid == 0 || dst_hyt == 0 )
    {
        dst_data = (unsigned char *)0;
        xP = (XImage *)XCreateImage(Dpy, AxVisual,
                                    1, XYBitmap,
                                    0, dst_data,
                                    0, 0,
                                    16, 0);
        if ( xP == (XImage *)NULL )
            return(( unsigned char *)0);
        xP->bitmap_unit = xP->bitmap_pad;
        xP->bytes_per_line = (dst_wid + 15)/16 * 2;
    }
    else
    {
        /* allocate a block of memory */
        dst_bwid = (dst_wid+7)/8 + ((dst_wid+7)/8)%2;
        dst_data = (unsigned char *)THIMhugeAlloc(pid, dst_hyt * dst_bwid);
        if ( dst_data == (unsigned char *)NULL )
            return( (unsigned char *)0 );
        xP = (XImage *)XCreateImage(Dpy, AxVisual,
                                    1, XYBitmap,
                                    0, dst_data,
                                    dst_wid, dst_hyt,
                                    16, dst_bwid);
        if ( xP == (XImage *)NULL )
            return(( unsigned char *)0);
        xP->bitmap_unit = xP->bitmap_pad;
        xP->bytes_per_line = dst_bwid;

        /* override bit/byte order */
        xP->byte_order = MSBFirst;
        xP->bitmap_bit_order = MSBFirst;

        /* set up the transfrom factors */
        x_xScale = tr->x_x;
        x_yScale = tr->x_y;
        y_xScale = tr->y_x;
        y_yScale = tr->y_y;
        x_Trans  = tr->x_t;
        y_Trans  = tr->y_t;

        /* scan thru the dest space, getting the pel from the source image */
        src_bwid = imP->data_byte_wid;
        dst_bwid = xP->bytes_per_line;
        src_data = (unsigned char *)imP->data;
        dst_data = (unsigned char *)xP->data;
        src_wid = imP->wid;
        src_hyt = imP->hyt;
        for ( i=0; i<dst_hyt; i++ )
        {
            line_ix = i*dst_bwid;
            for (j=0; j<dst_wid; j++ )
            {
                /* get the index for the pixel in the source image */
                x = j*x_xScale + i*x_yScale + x_Trans;
                y = i*y_yScale - j*y_xScale + y_Trans;

                /* if this value is valid update the pixel */
                if ( x >= 0 && x < src_wid && y >= 0 && y < src_hyt )
                {
                    ix = y*src_bwid+(x>>3);
                    if ( src_data[ix] & bitmask[x&7] )
                        dst_data[line_ix+(j>>3)] |= bitmask[j&7];
                    else
                        dst_data[line_ix+(j>>3)] &= ~bitmask[j&7];
                        
                }
                else
                    dst_data[line_ix+(j>>3)] &= ~bitmask[j&7];
            }
        }
    }
    return((unsigned char *)xP);
}

/***************************************************************************
function: axDitherColorImage - dither color image for mono display with
        Floyd Steinberg's algorithm
*/
static int
axDitherColorImage (pid,cmap, imP, outData)
int pid;
cmap_type *cmap;
ge_image_type *imP;
unsigned char **outData;
{
    register int i;
    short *inData;
    short *sP;
    int grayPix[256];
    int red, green, blue;
    unsigned char *iP;
    int wid;

    *outData = NULL;
    for (i=0; i < cmap->n_colors && i < 256; ++i) {
        /* cmyk to rgb */
        red = 255 - MIN(255, cmap->color[i].c+cmap->color[i].k);    /* red */
        green = 255 - MIN(255, cmap->color[i].m+cmap->color[i].k);  /* green */
        blue = 255 - MIN(255, cmap->color[i].y+cmap->color[i].k);   /* blue */

        /* compute color intensity */
        grayPix[i] = MONO(red,green,blue);
    }
    inData = (short *)THIMhugeAlloc(pid,imP->data_byte_wid * imP->hyt * sizeof(short));
    if (!inData)
        return(!AOK);

    wid = (imP->data_byte_wid+15)/16 * 2;   /* even boundary */
    *outData = (unsigned char *)THIMhugeAlloc(pid, wid*imP->hyt * sizeof(unsigned char));

    /* make up inData */
    sP = inData;
    iP = (unsigned char *)imP->data;
    for(i=imP->data_byte_wid * imP->hyt; i > 0; --i)
        *sP++ = grayPix[*iP++];

    FloydSteinbergDither(inData, imP->data_byte_wid, imP->hyt, 16, *outData);

    TaskFree(pid,(char *)inData);
    return(AOK);
}

/****************************************************************************/
unsigned char *
tr_sys_col1_image(pid, imP, cmap, tr)
int pid;
ge_image_type *imP;
cmap_type *cmap;
ge_image_transform_type *tr;
{
    register int i, j,k, x, y, ix, line_ix,line1_ix,line2_ix;
    int wid, hyt, bwid,jmax,imax;
    double x_xScale, x_yScale, x_Trans;
    double y_xScale, y_yScale, y_Trans;
    unsigned long cmapPixels[256];
    unsigned long pel, white_pel;
    XImage *xP;
    unsigned char *xBuf,*buf;
    unsigned char *ditheredData;

    /* get the width and height of the new, unmasked raster */
    wid = tr->wid;
    hyt = tr->hyt;

    /* set up the transfrom factors */
    x_xScale = tr->x_x;
    x_yScale = tr->x_y;
    y_xScale = tr->y_x;
    y_yScale = tr->y_y;
    x_Trans  = tr->x_t;
    y_Trans  = tr->y_t;

    /* create an index to pixel table */
    white_pel = AxPixels[1];
    for ( i=0; i<cmap->n_colors && i<256; i++ ) 
    {
        if ( cmap->color[i].see_thru == TRUE )
            cmapPixels[i] = white_pel;
        else 
        {
            AxGetRealPixel(2, cmap->color[i].c, cmap->color[i].m,
                              cmap->color[i].y, cmap->color[i].k,
                              255, &cmapPixels[i]);
        }
    }

    /* allocate a block of memory */
    bwid = (wid+7)/8 + ((wid+7)/8)%2;
    xBuf = (unsigned char *)THIMhugeAlloc(pid, hyt * bwid);
    if ( xBuf == (unsigned char *)NULL )
        return ( (unsigned char *)0 );
    xP = (XImage *)XCreateImage(Dpy, AxVisual,
                                1, XYBitmap, 0, xBuf, wid, hyt,
                                16, bwid);
    if ( xP == (XImage *)NULL )
        return(( unsigned char *)0);
    xP->bitmap_unit = xP->bitmap_pad;
    xP->bytes_per_line = bwid;

    /* override bit/byte order */
    xP->byte_order = MSBFirst;
    xP->bitmap_bit_order = MSBFirst;

    /* Dither image for display */
    if (axDitherColorImage(pid, cmap, imP, &ditheredData) == AOK)
    {
        /* scan thru the dest space, getting the pel from the source image */
        for ( i=0; i < hyt; ++i )
        {
            line_ix = i*bwid;
            for ( j=0; j < wid; ++j )
            {
                /* get the index for the pixel in the source image */
                x = j*x_xScale + i*x_yScale + x_Trans;
                y = i*y_yScale - j*y_xScale + y_Trans;
                /* if this value is valid update the pixel */
                if (x >= 0 && x < imP->wid && y >= 0 && y < imP->hyt)
                {
                    ix = ((imP->data_byte_wid+15)/16 * 2) * y + x/8;
                    if (ditheredData[ix] & bitmask[x&7])    /* bit set */
                        xBuf[line_ix+(j>>3)] |= bitmask[j&7];
                    else
                        xBuf[line_ix+(j>>3)] &= ~bitmask[j&7];
                }
                else
                    xBuf[line_ix+(j>>3)] &= ~bitmask[j&7];
            }
        }
        TaskFree(pid, (char *)ditheredData);
        return((unsigned char *)xP);
    }
    else    /* unable to dither ??? */
         return((unsigned char *)0);
}

/****************************************************************************/
unsigned char *
tr_sys_col4_image(pid, imP, cmap, tr)
int pid;
ge_image_type *imP;
cmap_type *cmap;
ge_image_transform_type *tr;
{
    register int i, j,k, x, y, ix, line_ix,line1_ix,line2_ix;
    int wid, hyt, bwid,jmax,imax;
    double x_xScale, x_yScale, x_Trans;
    double y_xScale, y_yScale, y_Trans;
    unsigned long cmapPixels[256];
    int r,g,b;
    unsigned long pel, white_pel, black_pel;
    XImage *xP;
    unsigned char *xBuf,*buf;
    int *tempBuf,*rBuf,*gBuf,*bBuf,*index;
    unsigned char *ditheredData;
    col_table *col,im_tab,col_tab[256];
    int red,green,blue,cindex,floyd_dither;
    color_type *tmp_color;

    /* get the width and height of the new, unmasked raster */
    wid = tr->wid;
    hyt = tr->hyt;

    /* set up the transfrom factors */
    x_xScale = tr->x_x;
    x_yScale = tr->x_y;
    y_xScale = tr->y_x;
    y_yScale = tr->y_y;
    x_Trans  = tr->x_t;
    y_Trans  = tr->y_t;

    floyd_dither = ( hyt <= 2 || axPreDefinedStdmap ) ? FALSE:AxSlowDither;
    /* create an index to pixel table */
    black_pel = AxPixels[0];
    white_pel = AxPixels[1];

    /* allocate a block of memory for 4bit-deep image */
    bwid = (wid+1)/2;
    bwid += (bwid%2);
    xBuf = (unsigned char *)THIMhugeAlloc(pid, hyt * bwid);
    if ( xBuf == (unsigned char *)NULL )
        return ( (unsigned char *)0 );
    tempBuf = (int *)malloc( 10 * bwid *sizeof(int));
    rBuf = &( tempBuf[0] );
    gBuf = &( tempBuf[3*bwid] );
    bBuf = &( tempBuf[6*bwid] );
    xP = (XImage *)XCreateImage(Dpy, AxVisual, 4, ZPixmap,
                                0, xBuf, wid, hyt, 16, bwid);
    if ( xP == (XImage *)NULL )
        return(( unsigned char *)0);
    xP->bitmap_unit = xP->bitmap_pad;
    xP->bytes_per_line = bwid;

    /* override bit/byte order */
    xP->byte_order = MSBFirst;
    xP->bitmap_bit_order = MSBFirst;

    /* dither if colorlevels == 1 */
    if ( AxNColorLevels == 1)
    {
        if ( axDitherColorImage(pid, cmap, imP, &ditheredData) == AOK )
        {
            for (i=0; i < hyt; ++i)
            {
                line_ix = i*bwid;
                for (j=0; j < wid; ++j)
                {
                    /* get the index for the pixel in the source image */
                    x = j*x_xScale + i*x_yScale + x_Trans;
                    y = i*y_yScale - j*y_xScale + y_Trans;
    
                    /* if this value is valid update the pixel */
                    if (x >= 0 && x < imP->wid && y >= 0 && y < imP->hyt) 
                    {
                        ix = ((imP->data_byte_wid+15)/16 * 2) * y + x/8;
                        if (ditheredData[ix] & bitmask[x&7])
                        {  /* bit set */
                            if (j%2)
                                xBuf[line_ix+j/2] |= black_pel;
                            else
                                xBuf[line_ix+j/2] = (black_pel << 4);
                        }
                        else
                        {
                            if (j%2)
                                xBuf[line_ix+j/2] |= white_pel;
                            else
                                xBuf[line_ix+j/2] = (white_pel << 4);
                        }
                    }
                    else 
                    {
                        if (j%2)
                            xBuf[line_ix+j/2] |= 0;
                        else
                            xBuf[line_ix+j/2] = 0;
                    }
                }
            }
            TaskFree(pid, (char *)ditheredData);
        }
        else
            return((unsigned char *)0);
    }
    else if ( floyd_dither ) /* this routine expects hyt >1 */
    {
        for ( i=0; i<2; i++ ) 
        { 
            line_ix = i*bwid;
            for (j=0; j<wid; j++ ) 
            {
                /* get the index for the pixel in the source image */
                x = j*x_xScale + i*x_yScale + x_Trans;
                y = i*y_yScale - j*y_xScale + y_Trans;

                /* if this value is valid update the pixel */
                if ( x >= 0 && x < imP->wid && y >= 0 && y < imP->hyt ) 
                {
                    /*get index of pixel locat from the color table */
                    tmp_color = &cmap->color[COL_IX(imP, x, y)];
                    k = tmp_color->k;
                    ix = line_ix+j;
                    rBuf[ix] = 255 - MIN(255, tmp_color->c + k);
                    gBuf[ix] = 255 - MIN(255, tmp_color->m + k);
                    bBuf[ix] = 255 - MIN(255, tmp_color->y + k);
                }
                else
                {
                    ix = line_ix+j;
                    rBuf[ix] = 0;
                    gBuf[ix] = 0;
                    bBuf[ix] = 0; 
                }
            }
        }
        jmax = wid-1;
        imax = hyt-1;
        for ( i=2; i<hyt; i++ ) 
        { 
            line_ix = i*bwid;
            for (j=0; j<wid; j++ ) 
            {
                /* get the index for the pixel in the source image */
                x = j*x_xScale + i*x_yScale + x_Trans;
                y = i*y_yScale - j*y_xScale + y_Trans;

                /* if this value is valid update the pixel */
                if ( x >= 0 && x < imP->wid && y >= 0 && y < imP->hyt ) 
                {
                    /*get index of pixel locat from the color table */
                    tmp_color = &cmap->color[COL_IX(imP, x, y)];
                    k = tmp_color->k;
                    line1_ix = (i%3)*bwid +j;
                    rBuf[line1_ix] = 255 - MIN(255, tmp_color->c + k);
                    gBuf[line1_ix] = 255 - MIN(255, tmp_color->m + k);
                    bBuf[line1_ix] = 255 - MIN(255, tmp_color->y + k);
                }
                else
                {
                    line1_ix = (i%3)*bwid +j;
                    rBuf[line1_ix] = 0;
                    gBuf[line1_ix] = 0;
                    bBuf[line1_ix] = 0;
                }
                k = (i%3);
                dither_image( bwid,rBuf,gBuf,bBuf,i,j,jmax,imax );
                line1_ix = ((k+1)%3)*bwid;
                ix = line1_ix+j;
                find_nearest_cmapindex(rBuf[ix],gBuf[ix],bBuf[ix],
                                       &im_tab,FALSE);
                line_ix = (i-2)*bwid;
                if (j%2)
                    xBuf[line_ix+j/2] = im_tab.fgcolor & 0x0F;
                else
                    xBuf[line_ix+j/2] = ((im_tab.fgcolor & 0x0F) << 4);
            }
        }
        /* last two scan lines */
        for ( i=hyt-2;i<hyt; i++ ) 
        { 
            for (j=0; j<wid; j++ ) 
            {
                if ( i == ( hyt-2 ) )
                    k = ((i-1)%3);
                else
                    k = (i%3);
                dither_image( bwid,rBuf,gBuf,bBuf,i,j,jmax,imax );
                line1_ix = k*bwid;
                ix = line1_ix+j;
                find_nearest_cmapindex(rBuf[ix],gBuf[ix],bBuf[ix],
                                       &im_tab,FALSE);
                line_ix = i*bwid;
                if (j%2)
                    xBuf[line_ix+j/2] = im_tab.fgcolor & 0x0F;
                else
                    xBuf[line_ix+j/2] = ( (im_tab.fgcolor & 0x0F) << 4 );
            }
        }
    }
    else  /* fast dithering -ordered matrix */
    {
        if ( axPreDefinedStdmap )
            axConvertStdIndexToPixel(col_tab, cmap);
        else for ( i=0; i<cmap->n_colors && i<256; i++ ) 
        {
            tmp_color = &cmap->color[i];
            if ( tmp_color->see_thru == TRUE )
            {
                memset((char *)&col_tab[i], 0, sizeof(col_table));
                col_tab[i].rfg  = 180;
                col_tab[i].gfg  = 30;
                col_tab[i].bfg  = 5;
            }
            else 
            {
                /* Convert cmyk to rgb */
                k = tmp_color->k;
                r = 255 - MIN(255, tmp_color->c + k);
                g = 255 - MIN(255, tmp_color->m + k);
                b = 255 - MIN(255, tmp_color->y + k);
                find_nearest_cmapindex(r,g,b,&col_tab[i],FALSE);
            }
        }
        for ( i=0; i<hyt; i++ ) 
        {
            line_ix = i*bwid;
            for (j=0; j<wid; j++ ) 
            {
                /* get the index for the pixel in the source image */
                x = j*x_xScale + i*x_yScale + x_Trans;
                y = i*y_yScale - j*y_xScale + y_Trans;

                /* if this value is valid update the pixel */
                if ( x >= 0 && x < imP->wid && y >= 0 && y < imP->hyt ) 
                {
                    /*get index of pixel locat from the color table */
                    col = &col_tab[COL_IX(imP, x, y)];

                    /*use 4x4 dither matrix  */
                    ix = (j%4)*4 + (i%4);
                    cindex = DITHER(ix,col);
                    if (j%2)
                        xBuf[line_ix+j/2] = (AxPixels[cindex] & 0x0F);
                    else
                        xBuf[line_ix+j/2] = ( (AxPixels[cindex] & 0x0F) << 4);
                }
                else 
                    xBuf[line_ix+j/2] = 0;
            }
        }
    }
    return((unsigned char *)xP);
}

/****************************************************************************/
unsigned char *
tr_sys_col8_image(pid, imP, cmap, tr)
int pid;
ge_image_type *imP;
cmap_type *cmap;
ge_image_transform_type *tr;
{
    register int i, j,k, x, y, ix, line_ix,line1_ix,line2_ix;
    int wid, hyt, bwid,jmax,imax;
    double x_xScale, x_yScale, x_Trans;
    double y_xScale, y_yScale, y_Trans;
    unsigned long cmapPixels[256];
    int r,g,b;
    unsigned long pel, white_pel, black_pel;
    XImage *xP;
    unsigned char *xBuf,*buf;
    int *tempBuf,*rBuf,*gBuf,*bBuf,*index;
    unsigned char *ditheredData;
    col_table *col,im_tab,col_tab[256];
    int red,green,blue,cindex,floyd_dither;
    color_type *tmp_color;

    black_pel = AxPixels[0];
    white_pel = AxPixels[1];

    /* get the width and height of the new, unmasked raster */
    wid = tr->wid;
    hyt = tr->hyt;

    /* set up the transfrom factors */
    x_xScale = tr->x_x;
    x_yScale = tr->x_y;
    y_xScale = tr->y_x;
    y_yScale = tr->y_y;
    x_Trans  = tr->x_t;
    y_Trans  = tr->y_t;

    floyd_dither = ( hyt <= 2 || axPreDefinedStdmap ) ? FALSE:AxSlowDither;

    /* allocate a block of memory */
    bwid = wid + wid%2;
    xBuf = (unsigned char *)THIMhugeAlloc(pid, hyt * bwid);
    if ( xBuf == (unsigned char *)NULL )
        return ( (unsigned char *)0 );

    tempBuf = (int *)malloc( 10 * bwid *sizeof(int));
    rBuf = &( tempBuf[0] );
    gBuf = &( tempBuf[3*bwid] );
    bBuf = &( tempBuf[6*bwid] );
    xP = (XImage *)XCreateImage(Dpy, AxVisual,
                                8, ZPixmap,
                                0, xBuf,
                                wid, hyt,
                                16, bwid);
    if ( xP == (XImage *)NULL )
        return(( unsigned char *)0);

    xP->bitmap_unit = xP->bitmap_pad;
    xP->bytes_per_line = bwid;

    /* override bit/byte order */
    xP->byte_order = MSBFirst;
    xP->bitmap_bit_order = MSBFirst;
    
    /* Dither image for display if AxNColorLevels == 1 */
    if ( AxNColorLevels == 1)
    {
         if (axDitherColorImage(pid, cmap, imP, &ditheredData) == AOK )
         {
            for ( i=0; i < hyt; ++i )
            {
                line_ix = i*bwid;
                for ( j=0; j < wid; ++j )
                {
                    /* get the index for the pixel in the source image */
                    x = j*x_xScale + i*x_yScale + x_Trans;
                    y = i*y_yScale - j*y_xScale + y_Trans;
    
                    /* if this value is valid update the pixel */
                    if ( x >= 0 && x < imP->wid && y >= 0 && y < imP->hyt )
                    {
                        ix = ((imP->data_byte_wid+15)/16 * 2) * y + x/8;
                        if (ditheredData[ix] & bitmask[x&7])    /* bit set */
                            xBuf[line_ix+j] = black_pel;
                        else
                            xBuf[line_ix+j] = white_pel;
                    }
                    else
                        xBuf[line_ix+j] = 0;
                }
            }
            TaskFree(pid, (char *)ditheredData);
        }
        else
            return((unsigned char *)0);
    }

    else if ( floyd_dither ) /* this routine expects hyt >1 */
    {
        for ( i=0; i<2; i++ ) 
        { 
            line_ix = i*bwid;
            for (j=0; j<wid; j++ ) 
            {
                /* get the index for the pixel in the source image */
                x = j*x_xScale + i*x_yScale + x_Trans;
                y = i*y_yScale - j*y_xScale + y_Trans;

                /* if this value is valid update the pixel */
                if ( x >= 0 && x < imP->wid && y >= 0 && y < imP->hyt ) 
                {
                    /*get index of pixel locat from the color table */
                    tmp_color = &cmap->color[COL_IX(imP, x, y)];
                    k = tmp_color->k;
                    ix = line_ix+j;
                    rBuf[ix] = 255 - MIN(255, tmp_color->c + k);
                    gBuf[ix] = 255 - MIN(255, tmp_color->m + k);
                    bBuf[ix] = 255 - MIN(255, tmp_color->y + k);
                }
                else
                {
                    ix = line_ix+j;
                    rBuf[ix] = 0;
                    gBuf[ix] = 0;
                    bBuf[ix] = 0; 
                }
            }
        }
        jmax = wid-1;
        imax = hyt-1;
        for ( i=2; i<hyt; i++ ) 
        { 
            line_ix = i*bwid;
            for (j=0; j<wid; j++ ) 
            {
                /* get the index for the pixel in the source image */
                x = j*x_xScale + i*x_yScale + x_Trans;
                y = i*y_yScale - j*y_xScale + y_Trans;

                /* if this value is valid update the pixel */
                if ( x >= 0 && x < imP->wid && y >= 0 && y < imP->hyt ) 
                {
                    /*get index of pixel locat from the color table */
                    tmp_color = &cmap->color[COL_IX(imP, x, y)];
                    k = tmp_color->k;
                    line1_ix = (i%3)*bwid +j;
                    rBuf[line1_ix] = 255 - MIN(255, tmp_color->c + k);
                    gBuf[line1_ix] = 255 - MIN(255, tmp_color->m + k);
                    bBuf[line1_ix] = 255 - MIN(255, tmp_color->y + k);
                }
                else
                {
                    line1_ix = (i%3)*bwid +j;
                    rBuf[line1_ix] = 0;
                    gBuf[line1_ix] = 0;
                    bBuf[line1_ix] = 0;
                }
                k = (i%3);
                dither_image( bwid,rBuf,gBuf,bBuf,i,j,jmax,imax );
                line1_ix = ((k+1)%3)*bwid;
                ix = line1_ix+j;
                find_nearest_cmapindex(rBuf[ix],gBuf[ix],bBuf[ix],
                                       &im_tab,FALSE);
                line_ix = (i-2)*bwid;
                xBuf[line_ix+j] = im_tab.fgcolor;
            }
        }
        /* last two scan lines */
        for ( i=hyt-2;i<hyt; i++ ) 
        { 
            for (j=0; j<wid; j++ ) 
            {
                if ( i == ( hyt-2 ) )
                    k = ((i-1)%3);
                else
                    k = (i%3);
                dither_image( bwid,rBuf,gBuf,bBuf,i,j,jmax,imax );
                line1_ix = k*bwid;
                ix = line1_ix+j;
                find_nearest_cmapindex(rBuf[ix],gBuf[ix],bBuf[ix],
                                       &im_tab,FALSE);
                line_ix = i*bwid;
                xBuf[line_ix+j] = im_tab.fgcolor;
            }
        }
    }
    else /* fast dithering */
    {
        /* create an index to pixel table */
        if ( axPreDefinedStdmap )
            axConvertStdIndexToPixel(col_tab, cmap);
        else for ( i=0; i<cmap->n_colors && i<256; i++ ) 
        {
            tmp_color = &cmap->color[i];
            if ( tmp_color->see_thru == TRUE )
            {
                memset((char *)&col_tab[i], 0, sizeof(col_table));
                col_tab[i].rfg  = 180;
                col_tab[i].gfg  = 30;
                col_tab[i].bfg  = 5;
            }
            else 
            {
                /* Convert cmyk to rgb */
                k = tmp_color->k;
                r = 255 - MIN(255, tmp_color->c + k);
                g = 255 - MIN(255, tmp_color->m + k);
                b = 255 - MIN(255, tmp_color->y + k);
                find_nearest_cmapindex(r,g,b,&col_tab[i],FALSE);
            }
        }
        for ( i=0; i<hyt; i++ ) 
        { 
            line_ix = i*bwid;
            buf = &xBuf[line_ix];
            for (j=0; j<wid; j++ ) 
            {
                /* get the index for the pixel in the source image */
                x = j*x_xScale + i*x_yScale + x_Trans;
                y = i*y_yScale - j*y_xScale + y_Trans;

                /* if this value is valid update the pixel */
                if ( x >= 0 && x < imP->wid && y >= 0 && y < imP->hyt ) 
                {
                    /*get index of pixel locat from the color table */
                    col = &col_tab[COL_IX(imP, x, y)];

                    /*use 4x4 dither matrix  */
                    ix = (j%4)*4 + (i%4);
                    cindex = DITHER(ix,col);
                    *buf = AxPixels[cindex];
                }
                else
                    *buf = 0;
                buf++;
            }
        }
    }
    /* free the  temp buffer */
    if ( tempBuf )
        free (tempBuf);
    return((unsigned char *)xP);
}

/************************************************************************
 * This function is for 16 bit images, e.g., DIBs on NT/Win95, which store
 * color values in 5-5-5 format (that is, 5 bits per R, G, B value).
*************************************************************************/
unsigned char *
tr_sys_col16_image(int pid, ge_image_type *imP, cmap_type *cmap,
                                    ge_image_transform_type *tr)
{
    register int i, j,k, x, y, ix, line1_ix,line2_ix;
	register int r_5, g_5, b_5;
    int wid, hyt, bwid,jmax,imax;
    double x_xScale, x_yScale, x_Trans;
    double y_xScale, y_yScale, y_Trans;
    unsigned long cmapPixels[256];
    int r,g,b, rm,gm,bm, rs, gs, bs;
    unsigned long pel;
    XImage *xP;
    unsigned char *xBuf,*buf;
    int *tempBuf,*rBuf,*gBuf,*bBuf,*index;
    unsigned char *ditheredData;
    col_table *col,im_tab,col_tab[256];
    int red,green,blue,cindex;
    color_type *tmp_color;

    /* get the width and height of the new, unmasked raster */
    wid = tr->wid;
    hyt = tr->hyt;

    /* set up the transfrom factors */
    x_xScale = tr->x_x;
    x_yScale = tr->x_y;
    y_xScale = tr->y_x;
    y_yScale = tr->y_y;
    x_Trans  = tr->x_t;
    y_Trans  = tr->y_t;

    /* allocate a block of memory */
    bwid = 2*(wid+wid%2);
    xBuf = (unsigned char *)THIMhugeAlloc(pid, hyt * bwid);
    if ( xBuf == (unsigned char *)NULL )
        return ( (unsigned char *)0 );
    xP = (XImage *)XCreateImage(Dpy, AxVisual, DpyDepth, ZPixmap,
                                    0, xBuf, wid, hyt, 32, bwid);
    if ( xP == (XImage *)NULL )
        return(( unsigned char *)0);
    xP->bitmap_unit = xP->bitmap_pad;
    xP->bytes_per_line = bwid;

    /* create an index to pixel table (assume 8 bit, contiguous masks) */
    rm = xP->red_mask;
    gm = xP->green_mask;
    bm = xP->blue_mask;
#if AX_X11
	if (0 == rm && 0 == gm && 0 == bm)
		{
		/* cheat, and read masks from display directly */
		Screen *ss = XScreenOfDisplay(Dpy, 0);
		Visual *vv = ss->root_visual;

		rm = vv->red_mask;
		gm = vv->green_mask;
		bm = vv->blue_mask;

		xP->red_mask = rm;
		xP->green_mask = gm;
		xP->blue_mask = bm;
	    }
#endif
	rs = ExtractShift_from_mask(rm);
	gs = ExtractShift_from_mask(gm);
	bs = ExtractShift_from_mask(bm);

    for ( i=0; i<cmap->n_colors; i++ ) 
        {
        if ( cmap->color[i].see_thru == TRUE )
            cmapPixels[i] = 0;
        else
            {
            r = 255 - MIN(255,(cmap->color[i].c+cmap->color[i].k));
            g = 255 - MIN(255,(cmap->color[i].m+cmap->color[i].k));
            b = 255 - MIN(255,(cmap->color[i].y+cmap->color[i].k));
#if AX_X11
			r_5 = ((double)r / 255.0) * (double) (rm >> rs);
			g_5 = ((double)g / 255.0) * (double) (gm >> gs);
			b_5 = ((double)b / 255.0) * (double) (bm >> bs);

			r = r_5 << rs;
			g = g_5 << gs;
			b = b_5 << bs;
#else
            normalize_rgb_to_555(r, g, b, &r, &g, &b);
#endif
            cmapPixels[i] = (r & rm) + (g & gm) + (b & bm);
            }
        }

    /* scan thru the dest space, getting the pel from the source image */
    for ( i=0; i < hyt; i++ )
        {
        for ( j=0; j < wid; j++ )
            {
            /* get the index for the pixel in the source image */
            x = j*x_xScale + i*x_yScale + x_Trans;
            y = i*y_yScale - j*y_xScale + y_Trans;

            /* if this value is valid update the pixel */
            if ( x >= 0 && x < imP->wid && y >= 0 && y < imP->hyt )
                {
                pel = cmapPixels[COL_IX(imP, x, y)];
                XPutPixel(xP, j, i, pel);
                }
            else
                XPutPixel(xP, j, i, 0L);
            }
        }
    return((unsigned char *)xP);
}

/****************************************************************************/
unsigned char *
tr_sys_col24_image(pid, imP, cmap, tr)
int pid;
ge_image_type *imP;
cmap_type *cmap;
ge_image_transform_type *tr;
{
    register int i, j,k, x, y, ix, line_ix,line1_ix,line2_ix;
    int wid, hyt, bwid,jmax,imax;
    double x_xScale, x_yScale, x_Trans;
    double y_xScale, y_yScale, y_Trans;
    unsigned long cmapPixels[256];
    int r,g,b, rm,gm,bm, rs,gs,bs;
    unsigned long pel;
    XImage *xP;
    unsigned char *xBuf,*buf;
    int *tempBuf,*rBuf,*gBuf,*bBuf,*index;
    unsigned char *ditheredData;
    col_table *col,im_tab,col_tab[256];
    int red,green,blue,cindex;
    color_type *tmp_color;

    /* get the width and height of the new, unmasked raster */
    wid = tr->wid;
    hyt = tr->hyt;

    /* set up the transfrom factors */
    x_xScale = tr->x_x;
    x_yScale = tr->x_y;
    y_xScale = tr->y_x;
    y_yScale = tr->y_y;
    x_Trans  = tr->x_t;
    y_Trans  = tr->y_t;

    /* allocate a block of memory */
    bwid = 4*wid;
    xBuf = (unsigned char *)THIMhugeAlloc(pid, hyt * bwid);
    if ( xBuf == (unsigned char *)NULL )
        return ( (unsigned char *)0 );
    xP = (XImage *)XCreateImage(Dpy, AxVisual, DpyDepth, ZPixmap,
                                0, xBuf, wid, hyt, 32, bwid);
    if ( xP == (XImage *)NULL )
        return(( unsigned char *)0);
    xP->bitmap_unit = xP->bitmap_pad;

    /* create an index to pixel table (assume 8 bit, contiguous masks) */
    rm = xP->red_mask;
    gm = xP->green_mask;
    bm = xP->blue_mask;
	
#if AX_X11
	if (0 == rm && 0 == gm && 0 == bm) {
		/* cheat, and read masks from display directly */
		Screen *ss = XScreenOfDisplay(Dpy, 0);
		Visual *vv = ss->root_visual;

		rm = vv->red_mask;
		gm = vv->green_mask;
		bm = vv->blue_mask;

		xP->red_mask = rm;
		xP->green_mask = gm;
		xP->blue_mask = bm;
		
	}
#endif
	
    rs = ExtractShift_from_mask(rm);
    gs = ExtractShift_from_mask(gm);
    bs = ExtractShift_from_mask(bm);
    for ( i=0; i<cmap->n_colors; i++ ) 
    {
        if ( cmap->color[i].see_thru == TRUE )
            cmapPixels[i] = 0;
        else
        {
            r = 255 - MIN(255,(cmap->color[i].c+cmap->color[i].k));
            g = 255 - MIN(255,(cmap->color[i].m+cmap->color[i].k));
            b = 255 - MIN(255,(cmap->color[i].y+cmap->color[i].k));
            cmapPixels[i] = ((r << rs) & rm) | (( g << gs) & gm) |
                                                ((b << bs ) & bm);
        }
    }

    /* scan thru the dest space, getting the pel from the source image */
    for ( i=0; i<hyt; i++ )
    {
        line_ix = i*bwid;
        for ( j=0; j<wid; j++ )
        {
            /* get the index for the pixel in the source image */
            x = j*x_xScale + i*x_yScale + x_Trans;
            y = i*y_yScale - j*y_xScale + y_Trans;

            /* if this value is valid update the pixel */
            if ( x >= 0 && x < imP->wid && y >= 0 && y < imP->hyt )
            {
                pel = cmapPixels[COL_IX(imP, x, y)];
                XPutPixel(xP, j, i, pel);
            }
            else
                XPutPixel(xP, j, i, 0L);
        }
    }
    return((unsigned char *)xP);
}

/*****************************************************************************
 *
 * tr_sys_color_image - TmScaleImage utility function
 */
unsigned char *
tr_sys_color_image(pid, imP, cmap, tr)
int pid;
ge_image_type *imP;
cmap_type *cmap;
ge_image_transform_type *tr;
{
    register int ix;
    XImage *xP;
    unsigned char *xBuf;

    /* pad the cmap to be atleast 256 */
    if ( cmap->n_colors < 256 )
    {
        ix = cmap->n_colors;
        cmap->color = (color_type *)TaskRealloc( pid, cmap->color,
                                    256 * sizeof(color_type) );
        for ( ; ix<256; ix++ )
        {
            cmap->color[ix].see_thru = FALSE;
            cmap->color[ix].ink_type = 0;
            cmap->color[ix].c = 0;
            cmap->color[ix].m = 0;
            cmap->color[ix].y = 0;
            cmap->color[ix].k = 0;
            cmap->color[ix].name = (char *) TaskAlloc(pid, 9);
            sprintf(cmap->color[ix].name, "color%03d", ix);
        }
        cmap->n_colors = 256;
    }
    /* a zero size image is created as a XYBitmap with width and height 0 */
    if ( tr->wid == 0 || tr->hyt == 0 )
    {
        xBuf = (unsigned char *)0;
        xP = (XImage *)XCreateImage(Dpy, AxVisual,
                                    1, XYBitmap,
                                    0, xBuf,
                                    0, 0,
                                    8, 0);
        if ( xP != (XImage *)NULL )
		{
			xP->bitmap_unit = xP->bitmap_pad;
			xP->bytes_per_line = (tr->wid + tr->wid%2);
		}
    }

    /* create a TrueColor/DirectColor DpyDepth image */
    else if ( DpyDepth > 8 ) {
        if ( DpyDepth == 16 )
            xP = (XImage *)tr_sys_col16_image(pid, imP, cmap, tr);
        else
			xP = (XImage *)tr_sys_col24_image( pid, imP, cmap, tr);
    }

    /* create an 8 deep image */
    else if ( DpyDepth == 8 )
            xP = (XImage *)tr_sys_col8_image( pid, imP, cmap, tr);
    
    /* create a 4 deep image (for now, a 1 deep image) */
    else if ( DpyDepth == 4 ) 
        xP = (XImage *)tr_sys_col4_image( pid, imP, cmap, tr);

    /* create a 1 deep image */
    else
        xP =  (XImage *)tr_sys_col1_image( pid, imP, cmap, tr);

    return ( (unsigned char *)xP );
}

/*****************************************************************************
 *
 * free_sys_image - TmScaleImage utility function
 */
free_sys_image(xp, pid)
unsigned char *xp;
int pid;
{
	XImage *x_im;
    if ( xp )
    {
		x_im = (XImage *)xp;
        x_im->height = x_im->width = 0;
        if ( x_im->data )
        {
            TaskFree(pid, x_im->data);
            x_im->data = (char *)0;
        }
        XDestroyImage(x_im);
    }
}

/*****************************************************************************
 *
 * free_java_image - TmScaleImage utility function
 */
void
free_java_image(java_img)
unsigned char *java_img;
{
    SAFEfree(java_img);
}

/*****************************************************************************
 *
 * tr_java_mono_image - TmScaleJavaImage function
 */
unsigned char *
tr_java_sub_mono_image(pid, imP, cmap, tr, use_mask)
int pid;
ge_image_type *imP;
cmap_type *cmap;
ge_image_transform_type *tr;
int use_mask;
{
    register int i, j, x, y, ix, line_ix;
    double x_xScale, x_yScale, x_Trans;
    double y_xScale, y_yScale, y_Trans;
    unsigned char *src_data, *dst_data;
    int src_bwid, src_wid, src_hyt;
    int dst_bwid, dst_wid, dst_hyt;

    /* get the width and height of the new, unmasked raster */
    dst_wid = tr->wid;
    dst_hyt = tr->hyt;

    /* allocate a block of memory */
    dst_bwid = (dst_wid+7)/8;
    dst_data = (unsigned char *)THIMhugeAlloc(pid, dst_hyt * dst_bwid);
    if ( dst_data != (unsigned char *)0 )
    {
        /* set up the transfrom factors */
        x_xScale = tr->x_x;
        x_yScale = tr->x_y;
        y_xScale = tr->y_x;
        y_yScale = tr->y_y;
        x_Trans  = tr->x_t;
        y_Trans  = tr->y_t;

        /* scan thru the dest space, getting the pel from the source image */
        if ( use_mask )
        {
            src_data = (unsigned char *)imP->mask;
            src_bwid = imP->mask_byte_wid;
        }
        else
        {
            src_data = (unsigned char *)imP->data;
            src_bwid = imP->data_byte_wid;
        }
        src_wid = imP->wid;
        src_hyt = imP->hyt;
        for ( i=0; i<dst_hyt; i++ )
        {
            line_ix = i*dst_bwid;
            for ( j=0; j<dst_wid; j++ )
            {
                /* get the index for the pixel in the source image */
                x = j*x_xScale + i*x_yScale + x_Trans;
                y = i*y_yScale - j*y_xScale + y_Trans;

                /* if this value is valid update the pixel */
                if ( x >= 0 && x < src_wid && y >= 0 && y < src_hyt )
                {
                    if ( src_data[y*src_bwid+(x>>3)] & bitmask[x&7] )
                        dst_data[line_ix+(j>>3)] |= bitmask[j&7];
                    else
                        dst_data[line_ix+(j>>3)] &= ~bitmask[j&7];
                }
                else
                    dst_data[line_ix+(j>>3)] &= ~bitmask[j&7];
            }
        }
    }
    return(dst_data);
}

/*****************************************************************************
 *
 * tr_java_mono_image - TmScaleJavaImage function
 */
unsigned char *
tr_java_mono_image(pid, imP, cmap, tr)
int pid;
ge_image_type *imP;
cmap_type *cmap;
ge_image_transform_type *tr;
{
	return(tr_java_sub_mono_image(pid, imP, cmap, tr, FALSE));
}

/*****************************************************************************
 *
 * tr_java_mask_image - TmScaleJavaImage function
 */
unsigned char *
tr_java_mask_image(pid, imP, cmap, tr)
int pid;
ge_image_type *imP;
cmap_type *cmap;
ge_image_transform_type *tr;
{
	return(tr_java_sub_mono_image(pid, imP, cmap, tr, TRUE));
}

/*****************************************************************************
 *
 * tr_java_color_image - TmScaleJavaImage function
 */
unsigned char *
tr_java_color_image(pid, imP, cmap, tr)
int pid;
ge_image_type *imP;
cmap_type *cmap;
ge_image_transform_type *tr;
{
    double x_xScale, x_yScale, x_Trans;
    double y_xScale, y_yScale, y_Trans;
    register int i, j, line_ix;
    int x, y, wid, hyt, bwid;
	unsigned char *buf, pel;

    /* get the width and height of the new, unmasked raster */
    wid = tr->wid;
    hyt = tr->hyt;

    /* set up the transform factors */
    x_xScale = tr->x_x;
    x_yScale = tr->x_y;
    y_xScale = tr->y_x;
    y_yScale = tr->y_y;
    x_Trans  = tr->x_t;
    y_Trans  = tr->y_t;

    /* allocate a block of memory */
    bwid = wid;
    buf = (unsigned char *)THIMhugeAlloc(pid, hyt*bwid);
    if ( buf != (unsigned char *)0 )
	{
        /* scan thru the dest space, getting the pel from the source image */
        for ( i=0; i<hyt; i++ )
        {
            line_ix = i*bwid;
            for ( j=0; j<wid; j++ )
            {
                /* get the index for the pixel in the source image */
                x = j*x_xScale + i*x_yScale + x_Trans;
                y = i*y_yScale - j*y_xScale + y_Trans;

                /* if this value is valid update the pixel */
                if ( x >= 0 && x < imP->wid && y >= 0 && y < imP->hyt )
					pel = imP->data[y*imP->data_byte_wid + x];
    			else
    				pel = (unsigned char)0;
    			buf[line_ix + j] = pel;
            }
        }
    }
    return(buf);
}

/*****************************************************************************
 *
 * ax_tm_fast_simage - TmScaleImage function
 */
ax_tm_fast_simage( pid, imP, xform, trns, cmap  )
int pid;
ge_image_type *imP;
poly_type *xform;
ge_image_transform_type *trns;
cmap_type *cmap;
{
#if AXMS || LINUX /* only temporary for linux, theoretically :( */
    return(FALSE);
#endif
    if ( check_boundary(imP, xform, trns) )
    {
        if ( imP->dep == 1 )
        {
            imP->sys_data = tr_fst_sys_mono_image (pid, imP,cmap,trns );
            imP->sys_mask = (unsigned char *)0;
        }
        else
        {
            imP->sys_data = tr_fst_sys_color_image(pid,imP,cmap,trns );
            if ( imP->sys_data == (unsigned char *)0 )
                imP->sys_mask = (unsigned char *)0;
            else
                imP->sys_mask = tr_fst_sys_img_msk( pid,imP,cmap,trns );
        }
        return(TRUE);
    }
    else
        return(FALSE);
}

no_image_data(imP)
ge_image_type *imP;
{
    if (imP->data == NULL || imP->wid == 0 || imP->hyt == 0 )
    {
        imP->sys_data = (unsigned char *)0;
        imP->sys_mask = (unsigned char *)0;
        return(TRUE);
    }
    return(FALSE);
}

/*****************************************************************************
 *
 * ax_tm_simage - AxScaleImage function
 */
ax_tm_simage(pid, imP, cmap, xform, x_scale, y_scale)
int pid;
ge_image_type *imP;
cmap_type *cmap;
poly_type *xform;
double x_scale;
double y_scale;
{
    ge_image_transform_type tr;
    int wid, hyt;

    /* get rid of any old image data */
    if ( imP->sys_data )
        free_sys_image((unsigned char *)imP->sys_data, pid);
    if ( imP->sys_mask )
        free_sys_image((unsigned char *)imP->sys_mask, pid);
    if ( has_image_data(imP) == FALSE )
    {
        imP->sys_data = (unsigned char *)0;
        imP->sys_mask = (unsigned char *)0;
    }
    else
    {
        /* calc the transformation factors */
        tr.x_scale = x_scale;
        tr.y_scale = y_scale;
        extract_factors(imP, xform, &tr);

        /* scale image */
        if ( !ax_tm_fast_simage(pid, imP, xform, &tr, cmap) )
        {
            /* create a scaled sys image and a scaled sys mask */
            if ( imP->dep == 1 )
            {
                imP->sys_data = tr_sys_mono_image (pid, imP,cmap,&tr);
                imP->sys_mask = (unsigned char *)0;
            }
            else
            {
                imP->sys_data = tr_sys_color_image(pid, imP, cmap, &tr);
                if ( imP->sys_data == (unsigned char *)0 )
                    imP->sys_mask = (unsigned char *)0;
                else
                    imP->sys_mask = tr_sys_image_mask(pid, imP, cmap, &tr);
            }
        }
    }
}

/*****************************************************************************
 *
 * ax_tm_simage - AxScaleJavaImage function
 */
ax_tm_sjimage(pid, imP, cmap, xform, x_scale, y_scale)
int pid;
ge_image_type *imP;
cmap_type *cmap;
poly_type *xform;
double x_scale;
double y_scale;
{
    ge_image_transform_type tr;
    int wid, hyt;

    /* get rid of any old image data */
    if ( imP->java_data )
        free_java_image(imP->java_data);
    if ( imP->java_mask )
        free_java_image(imP->java_mask);
    if ( has_image_data(imP) == FALSE )
    {
        imP->java_data = (unsigned char *)0;
        imP->java_mask = (unsigned char *)0;
    }
    else
    {
        /* calc the transformation factors */
        tr.x_scale = x_scale;
        tr.y_scale = y_scale;
        extract_factors(imP, xform, &tr);

        /* scale image */
        if ( imP->dep == 1 ) /* mono image */
        {
            imP->java_data = tr_java_mono_image(pid, imP, cmap, &tr);
            imP->java_mask = (unsigned char *)0;
        }
        else /* color image */
        {
            imP->java_data = tr_java_color_image(pid, imP, cmap, &tr);
            if ( imP->java_data )
            {
                imP->java_mask = tr_java_mask_image(pid, imP, cmap, &tr);
                if ( imP->java_mask == (unsigned char *)0 )
                {
                    free_java_image(imP->java_mask);
                    imP->java_data = (unsigned char *)0;
                }
            }
        }
    }
}

/****************************************************************************/
getColorLevel(ix)
int ix;
{
    if ( ix < 83 )
        return(0);
    if ( ix < 165 )
        return(1);
    if ( ix < 205 )
        return(2);
    if ( ix < 238 )
        return(3);
    if ( ix < 247 )
        return(4);
    return(5);
}

/****************************************************************************/
unsigned char getColorIndex(r, g, b)
int r, g, b;
{
    int base_index;

    /* rgb to get the index */
    r = getColorLevel(r);
    g = getColorLevel(g);
    b = getColorLevel(b);

    /* a grey (red, green, and blue are all the same) */
    if ( r == g && g == b )
    {
        if ( r == 0 )
            return(1); /* black at position 1 */
        else if ( r == 5 )
            return(2); /* white at position 2 */
        else
            return(2 + r); /* greys at 2 to 6 */
    }

    /* the '6' is for transparent and 6 greys which are placed at the start */
    base_index = r*36 + g*6 + b + 6;

    /* since the grey entries do not appear in the midst of the other color */
    /* entries, but rather are moved to the front of the colormap. we must  */
    /* adjust the index to account for these missing entires (see the code  */
    /* in "create_cmap()" ).  The funny numbers below are because an entry  */
    /* is skipped starting at 49, and every 42 entries after that.          */
    if ( base_index >= 217 )      /* r == g == b == 5 */
       base_index = base_index - 5;
    else if ( base_index >= 175 ) /* r == g == b == 4 */
       base_index = base_index - 4;
    else if ( base_index >= 133 ) /* r == g == b == 3 */
       base_index = base_index - 3;
    else if ( base_index >= 91  ) /* r == g == b == 2 */
       base_index = base_index - 2;
    else if ( base_index >= 49 )  /* r == g == b == 1 */
       base_index = base_index - 1;

    return(base_index);
}

/****************************************************************************/
unsigned char *convert24to8(pid, x_im, bytewid)
int pid;
XImage *x_im;
int *bytewid;
{
    register int y, x, bwid, src_wid, src_hyt;
    int r,g,b, rs,gs,bs, rm,gm,bm;
    unsigned char *imP, *dstP;
    unsigned long pel;
    unsigned long (*func)();

    /* allocate a 16 bit padded buffer */
    bwid = x_im->width + x_im->width%2;
    imP = (unsigned char *)TaskAlloc(pid, (x_im->height * bwid));

    /* set the mask's and shift's for reading the pixels */
    rm = x_im->red_mask;
    gm = x_im->green_mask;
    bm = x_im->blue_mask;

#if AX_X11
    if (0 == rm && 0 == gm && 0 == bm) {
		/* cheat, and read masks from display directly */
		Screen *ss = XScreenOfDisplay(Dpy, 0);
		Visual *vv = ss->root_visual;

		rm = vv->red_mask;
		gm = vv->green_mask;
		bm = vv->blue_mask;

		x_im->red_mask = rm;
		x_im->green_mask = gm;
		x_im->blue_mask = bm;
	}
#endif
    
    rs = ExtractShift_from_mask(rm);
    gs = ExtractShift_from_mask(gm);
    bs = ExtractShift_from_mask(bm);
    
    /* read the image */
    src_wid = x_im->width;
    src_hyt = x_im->height;

#if AXMS || LINUX
    func = AxGetPixel;
#else
    func = GetPixelArray[DpyDepth].func;
#endif
    for ( y=0; y<src_hyt; y++ )
    {
        /* point to the destination scanline... */
        dstP = &imP[bwid * y];
     
        /* read the source scanline into the dest scanline */
        for ( x=0; x<src_wid; x++ )
        {
            pel = (*func)(x_im, x, y);
            r = (pel & rm) >> rs;
            g = (pel & gm) >> gs;
            b = (pel & bm) >> bs;
            *dstP++ = getColorIndex(r, g, b);
        }
    }
    *bytewid = bwid;
    return ( imP );
}

/****************************************************************************/
unsigned char *
convert16to8(int pid, XImage *x_im, int *bytewid)
{
    register int y, x, bwid, src_wid, src_hyt;
    register int red_shift, green_shift, blue_shift;
    register int  r16, g16, b16;
    int r,g,b, red_mask,green_mask,blue_mask;
    unsigned char *imP, *dstP;
    unsigned long pel;
    unsigned long (*func)();
    double r_conversion, g_conversion, b_conversion;

    /* allocate a 16 bit padded buffer */
    bwid = x_im->width + x_im->width%2;
    imP = (unsigned char *)TaskAlloc(pid, (x_im->height * bwid));

    /* set the masks for reading the pixels */
    red_mask = x_im->red_mask;
    green_mask = x_im->green_mask;
    blue_mask = x_im->blue_mask;
    
#if AX_X11
    if (0 == red_mask && 0 == green_mask && 0 == blue_mask)
		{
		/* cheat, and read masks from display directly */
		Screen *ss = XScreenOfDisplay(Dpy, 0);
		Visual *vv = ss->root_visual;

		red_mask = vv->red_mask;
		green_mask = vv->green_mask;
		blue_mask = vv->blue_mask;

		x_im->red_mask = red_mask;
		x_im->green_mask = green_mask;
		x_im->blue_mask = blue_mask;
	    }
#endif
    
    /* get the shift values to use with these masks.  These may not always */
    /* be the same.  I have seen the masks come back from Windows as:      */
    /* 0xF800, 0x7E0, ox1F.  There are 6 bits allocated for green - jon    */
    red_shift = ExtractShift_from_mask( red_mask );
    green_shift = ExtractShift_from_mask( green_mask );
    blue_shift = ExtractShift_from_mask( blue_mask );

    /* read the image */
    src_wid = x_im->width;
    src_hyt = x_im->height;

    r_conversion = red_mask >> red_shift;
    g_conversion = green_mask >> green_shift;
    b_conversion = blue_mask >> blue_shift;

    for ( y=0; y < src_hyt; y++ )
        {
        /* point to the destination scanline... */
        dstP = &imP[bwid * y];
     
        /* read the source scanline into the dest scanline */
        for ( x=0; x < src_wid; x++ )
            {
            pel = XGetPixel(x_im, x, y);

            /* normalize to rgb values */
            r16 = (pel & red_mask) >> red_shift;
            g16 = (pel & green_mask) >> green_shift;
            b16 = (pel & blue_mask) >> blue_shift;

            r = (int) (((double) r16 * 255.0) / (double) r_conversion);
            g = (int) (((double) g16 * 255.0) / (double) g_conversion);
            b = (int) (((double) b16 * 255.0) / (double) b_conversion);

            *dstP++ = getColorIndex(r, g, b);
            }
        }
    *bytewid = bwid;
    return imP;
}

/****************************************************************************/
unsigned char *convert8to8(pid, x_im, bytewid, cmap)
int pid;
XImage *x_im;
int *bytewid;
cmap_type *cmap;
{
    register int y, x, src_wid, src_hyt, bwid;
    unsigned char *imP, *dstP;
    int i, n, ix, color_ix[256];
    unsigned long (*func)();

    /* allocate a 16 bit padded buffer */
    bwid = x_im->width + x_im->width%2;

    imP = (unsigned char *)TaskAlloc(pid, (x_im->height * bwid));

    /* create a pixel to colormap_index table */
    n = cmap->n_colors;
    for ( i=n-1; i>=0; i-- )
    {
#if AXMS
        axWinGetImageColormapIndex(cmap->color[i].c, cmap->color[i].m,
                            cmap->color[i].y, cmap->color[i].k, x_im, &ix);
 
        if ( cmap->color[i].see_thru )
            color_ix[AxPixels[ix]] = i;
        else
            color_ix[ix] = i;
#else
        axGetColormapIndex(2, cmap->color[i].c,
                              cmap->color[i].m,
                              cmap->color[i].y,
                              cmap->color[i].k,
                              255,
                              &ix);

        color_ix[AxPixels[ix]] = i;
#endif
    }

#if AXMS
    /** HACK ALERT!  Windows insists on giving us its own "static color" index
      * for Applix black when GetDIBits is called. So here, we explicitly set 
      * the index for the Windows black (0) to our black (1).    -donato
      */
    color_ix[0] = 1;    /* MS black (0) to Applix black (1) */
#endif

    /* read the image */
    src_wid = x_im->width;
    src_hyt = x_im->height;
    if ( DpyDepth == 8 )
    {
        for ( y=0; y<src_hyt; y++ )
        {
            /* point to the destination scanline... */
            dstP = &imP[bwid * y];

            /* read the source scanline into the dest scanline */
            for ( x=0; x<src_wid; x++ )
            {
                n = AXGETPIXEL08(x_im, x, y);
                if ( n == AlisPixels[0] )
                    *dstP++ = 0;
                else
                    *dstP++ = color_ix[n];
            }
        }
    }
    else if ( ( DpyDepth == 1 ) &&
            ( x_im->byte_order == x_im->bitmap_bit_order ))
    {
        if ( x_im->byte_order == MSBFirst )
        {
            for ( y=0; y<src_hyt; y++ )
            {
                /* point to the destination scanline... */
                dstP = &imP[bwid * y];
         
                /* read the source scanline into the dest scanline */
                for ( x=0; x<src_wid; x++ )
                    *dstP++ = color_ix[AXGETPIXEL01MSB(x_im, x, y)];
            }
        }
        else
        {
            for ( y=0; y<src_hyt; y++ )
            {
                /* point to the destination scanline... */
                dstP = &imP[bwid * y];
         
                /* read the source scanline into the dest scanline */
                for ( x=0; x<src_wid; x++ )
                    *dstP++ = color_ix[AXGETPIXEL01LSB(x_im, x, y)];
            }
        }
    }
    else
    {
#if AXMS || LINUX
        func = AxGetPixel;
#else
        func = GetPixelArray[DpyDepth].func;
#endif
        for ( y=0; y<src_hyt; y++ )
        {
            /* point to the destination scanline... */
            dstP = &imP[bwid * y];
    
            /* read the source scanline into the dest scanline */
            for ( x=0; x<src_wid; x++ )
                *dstP++ = color_ix[(*func)(x_im, x, y)];
        }
    }

    *bytewid = bwid;
    return ( imP );
}

/****************************************************************************/
unsigned char *pad_out_image(pid, x_im, bytewid, cmap)
int pid;
XImage *x_im;
int *bytewid;
cmap_type *cmap;
{
    unsigned char *imP;

    /* create a colormap */
    create_cmap(pid, cmap);

    if ( DpyDepth <= 8 ) /* pixel indices < 256 */
        imP = convert8to8(pid, x_im, bytewid, cmap);
    else if ( DpyDepth == 16 )
		imP = convert16to8(pid, x_im, bytewid);
	else
        imP = convert24to8(pid, x_im, bytewid);
    return ( imP );
}

/*****************************************************************************
 *
 * flatten_image - flatten image into 1 plane image (TmGetImage utility)
 */
unsigned char *flatten_image(pid, x_im, bytewid)
int pid;
XImage *x_im;
int *bytewid;
{
    unsigned char *imP, *dstP, bitmask1, outval;    
    register int i, j, size, hyt, wid, bwid;
    unsigned long (*func)();
    int pixel;


    pixel = AlisPixels[0];

    /* allocate a 16 bit scanline padded buffer */
    bwid = (x_im->width+7)/8 + ((x_im->width+7)/8)%2;
    size = x_im->height * bwid;
    imP = (unsigned char *)TaskAlloc(pid, size);

    /* for each scanline in the image... */
    hyt = x_im->height;
    wid = x_im->width;
    if ( DpyDepth == 8 )
    {
        for(i=0; i<hyt; i++)
        {
            /* point to the destination scanline... */
            dstP = &imP[bwid * i];
     
            /* some per scanline inits... */
            bitmask1 = 0x80;
            outval = 0;
     
            /* read the pixels from a given scanline */
            for(j=0; j<wid; j++)
            {
                /* all non-white pels are set to black */
                if ( pixel != AXGETPIXEL08(x_im, j, i) )
                    outval |= bitmask1;
     
                /* every 8 pels, update the image and clear the byte */
                if ( (bitmask1>>=1) == 0x00 ) 
                {
                    *dstP++ = outval;
                    outval = 0;
                    bitmask1 = 0x80;
                }
            }
    
            /* if bitmask is other than 0x80, last dest byte unfinished */
            if ( bitmask1 & 0x7f )
                *dstP++ = outval;
        }
    }
    else if ( ( DpyDepth == 1 ) &&
        ( x_im->byte_order == x_im->bitmap_bit_order ))
    {
        if ( x_im->byte_order == MSBFirst )
        {
            for(i=0; i<hyt; i++)
            {
                /* point to the destination scanline... */
                dstP = &imP[bwid * i];
         
                /* some per scanline inits... */
                bitmask1 = 0x80;
                outval = 0;
         
                /* read the pixels from a given scanline */
                for(j=0; j<wid; j++)
                {
                    /* all non-white pels are set to black */
                    if ( pixel != AXGETPIXEL01MSB(x_im, j, i) )
                        outval |= bitmask1;
         
                    /* every 8 pels, update the image and clear the byte */
                    if ( (bitmask1>>=1) == 0x00 ) 
                    {
                        *dstP++ = outval;
                        outval = 0;
                        bitmask1 = 0x80;
                    }
                }

                /* if bitmask is other than 0x80, last dest byte unfinished */
                if ( bitmask1 & 0x7f )
                    *dstP++ = outval;
            }
        }
        else
        {
            for(i=0; i<hyt; i++)
            {
                /* point to the destination scanline... */
                dstP = &imP[bwid * i];
         
                /* some per scanline inits... */
                bitmask1 = 0x80;
                outval = 0;
         
                /* read the pixels from a given scanline */
                for(j=0; j<wid; j++)
                {
                    /* all non-white pels are set to black */
                    if ( pixel != AXGETPIXEL01LSB(x_im, j, i) )
                        outval |= bitmask1;
         
                    /* every 8 pels, update the image and clear the byte */
                    if ( (bitmask1>>=1) == 0x00 ) 
                    {
                        *dstP++ = outval;
                        outval = 0;
                        bitmask1 = 0x80;
                    }
                }

                /* if bitmask is other than 0x80, last dest byte unfinished */
                if ( bitmask1 & 0x7f )
                    *dstP++ = outval;
            }
        }
    }
    else
    {
#if AXMS
        /* for 16,24 and 32 bit color, the values in the buffer are */
        /* not direct RGB values, so we can not compare it against */
        /* "AlisPixels[0]" (white) as the variable "pixel" is      */
        /* initialized at the top of this function.  Instead we    */
        /* create a white value by setting to 1 every bit in the   */
        /* appropriate color mask area. - jon                      */
        /* AlisPixels[0] is actually the background color, which
         * isn't necessarily white and is in RGB format, which isn't
         * necessarily the same as DIB format, so we need to get the
         * corresponding DIB color value.    -donato
         */
        pixel = (int)axmsRetrieveDIBValue(AlisPixels[0]);
#endif

#if AXMS || LINUX
        func = AxGetPixel;
#else
        func = GetPixelArray[DpyDepth].func;
#endif
        for(i=0; i<hyt; i++)
        {
            /* point to the destination scanline... */
            dstP = &imP[bwid * i];
     
            /* some per scanline inits... */
            bitmask1 = 0x80;
            outval = 0;
     
            /* read the pixels from a given scanline */
            for(j=0; j<wid; j++)
            {
                /* all non-white pels are set to black */
                
                if ( pixel != ( *func )(x_im, j, i) )
                    outval |= bitmask1;
     
                /* every 8 pels, update the image and clear the byte */
                if ( (bitmask1>>=1) == 0x00 ) 
                {
                    *dstP++ = outval;
                    outval = 0;
                    bitmask1 = 0x80;
                }
            }
    
            /* if bitmask is other than 0x80, last dest byte unfinished */
            if ( bitmask1 & 0x7f )
                *dstP++ = outval;
        }
    }

    *bytewid = bwid;
    return ( imP );
}

/*****************************************************************************
 *
 * ax_tm_create_image - create an empty image -TmCreateImage function
 */
ax_tm_create_image(w, pid, pnts, use_color, imP, cmap)
long w;
int pid;
TmPoint32 pnts[2];
int use_color;
ge_image_type *imP;
cmap_type *cmap;
{
    TmPoint32 p1, p2, points[5];
    ge_image_transform_type tr;
    int wid, hyt, size;
    poly_type xform;


    /* init transform factors */
    points[0].x = pnts[0].x;
    points[0].y = pnts[0].y;
    points[1].x = pnts[1].x;
    points[1].y = pnts[0].y;
    points[2].x = pnts[1].x;
    points[2].y = pnts[1].y;
    points[3].x = pnts[0].x;
    points[3].y = pnts[1].y;
    points[4].x = pnts[0].x;
    points[4].y = pnts[0].y;
    tr.x_scale     = 1.0;
    tr.y_scale     = 1.0;
    xform.nPoints  = 5;
    xform.nCtrls   = 0;
    xform.points   = points;
    xform.controls = (TmPoint32 *)0;

    /* convert units from mils to screen pixels */
    conv_i_p32(&pnts[0].x, &pnts[0].y, &p1.x, &p1.y, MILLS_TO_PIXEL);
    conv_i_p32(&pnts[1].x, &pnts[1].y, &p2.x, &p2.y, MILLS_TO_PIXEL);
	memset(imP, 0, sizeof(ge_image_type));
    imP->wid = wid = p2.x - p1.x;
    imP->hyt = hyt = p2.y - p1.y;
    imP->dep = use_color ? 8 : 1;
    imP->eps_seg = NULL;
    imP->eps_seg_size = 0;
    imP->java_data = (unsigned char *)0;
    imP->java_mask = (unsigned char *)0;

    /* create a colormap */
    cmap->n_colors = 0;
    cmap->color = (color_type *)0;
    if ( use_color )
    {
        /* get the colormap */
        create_cmap(pid, cmap);

        /* the image data */
        imP->data_byte_wid = wid + wid%2;
        size = hyt * imP->data_byte_wid;
        imP->data = (unsigned char *)TaskAlloc(pid, size);
        memset((char *)imP->data, 0, size);

        /* the mask image */
        imP->mask_byte_wid = (wid+7)/8 + ((wid+7)/8)%2;
        size = hyt * imP->mask_byte_wid;
        imP->mask = (unsigned char *)TaskAlloc(pid, size);
        memset((char *)imP->mask, 0, size);

        /* the system image and mask */
        extract_factors(imP, &xform, &tr);
        imP->sys_data = (unsigned char *)NULL;
        imP->sys_mask = (unsigned char *)NULL;
    }
    else
    {
        /* the image data */
        imP->data_byte_wid = (wid+7)/8 + ((wid+7)/8)%2;
        size = hyt * imP->data_byte_wid;
        imP->data = (unsigned char *)TaskAlloc(pid, size);
        memset((char *)imP->data, 0, size);

        /* the mask image */
        imP->mask = (unsigned char *)0;
        imP->mask_byte_wid = 0;

        /* the system image and mask */
        extract_factors(imP, &xform, &tr);
        imP->sys_data = (unsigned char *)NULL;
        imP->sys_mask = (unsigned char *)NULL;
    }
}

XImage *
JavaGetImage(pid, x, y, wid, hyt)
int pid;
int x, y, wid, hyt;
{
    return((XImage *)0);
}

/****************************************************************************/
unsigned char *pad_out_jimage(pid, wid, hyt, dat, cmap, dat_bwid)
int pid;
int wid, hyt;
unsigned char *dat;
cmap_type *cmap;
int *dat_bwid;
{
	register int ix, x, y, bwid;
	unsigned char *im, *dst;

	/* create a colormap */
	create_cmap(pid, cmap);

	/* allocate a 16 bit padded buffer */
	bwid = wid + wid%2;
	im = (unsigned char *)TaskAlloc(pid, (hyt*bwid));

	/* read the image */
	for ( ix=y=0; y<hyt; y++ )
	{
		/* point to the destination scanline... */
		dst = &im[bwid * y];

		/* read the source scanline into the dest scanline */
		for ( x=0; x<wid; x++ )
		{
			ix++; /* skip the alpha val */
			*dst++ = getColorIndex(dat[ix++], dat[ix++], dat[ix++]);
		}
	}
	*dat_bwid = bwid;
	return(im);
}

/****************************************************************************/
unsigned char *flatten_jimage(pid, wid, hyt, dat, dat_bwid)
int pid;
int wid, hyt;
unsigned char *dat;
int *dat_bwid;
{
	unsigned char *im, *dst, bitmask1, outval;    
	register int ix, x, y, bwid, a;

	/* allocate a 16 bit padded buffer */
    bwid = (wid+7)/8 + ((wid+7)/8)%2;
	im = (unsigned char *)TaskAlloc(pid, (hyt*bwid));

	/* read the image */
	for ( ix=y=0; y<hyt; y++ )
	{
		/* some per scanline inits... */
		dst = &im[bwid * y];
		bitmask1 = 0x80;
		outval = 0;

		/* read the source scanline into the dest scanline */
		for ( x=0; x<wid; x++ )
		{
			/* every 4th pixel is the alpha val */
			a = dat[4*(ix++)];
			if ( a )
				outval |= bitmask1;
    
			/* every 8 pels, update the image and clear the byte */
			if ( (bitmask1>>=1) == 0x00 ) 
			{
				*dst++ = outval;
				outval = 0;
				bitmask1 = 0x80;
			}
		}

		/* if bitmask is other than 0x80, last dest byte unfinished */
		if ( bitmask1 & 0x7f )
			*dst++ = outval;
	}
	*dat_bwid = bwid;
	return(im);
}

/*****************************************************************************
 *
 * java_gimage
 */
java_gimage(pid, use_color, wid, hyt, dat, imP, cmap)
int pid;
int use_color;
int wid, hyt;
unsigned char *dat;
ge_image_type *imP;
cmap_type *cmap;
{
	int bwid;

	memset(imP, 0, sizeof(ge_image_type));

	imP->wid = wid;
	imP->hyt = hyt;
	imP->dep = use_color ? 8 : 1;
	imP->eps_seg = NULL;
	imP->eps_seg_size = 0;

	/* create a colormap */
	cmap->n_colors = 0;
	cmap->color = (color_type *)0;

	/* convert the sys-image to a sys-independant image */
	if ( use_color )
	{
		/* the image data */
		imP->data = pad_out_jimage(pid, wid, hyt, dat, cmap, &bwid);
		imP->data_byte_wid = bwid;

		/* the mask image */
		imP->mask = flatten_jimage(pid, wid, hyt, dat, &bwid);
		imP->mask_byte_wid = bwid;

		/* the system image and mask */
		imP->sys_data = (unsigned char *)NULL;
		imP->sys_mask = (unsigned char *)NULL;
	}
	else
	{
		/* the image data */
		imP->data = flatten_jimage(pid, wid, hyt, dat, &bwid);
		imP->data_byte_wid = bwid;

		/* the mask image */
		imP->mask = (unsigned char *)NULL;
		imP->mask_byte_wid = 0;

		/* the system image and mask */
		imP->sys_data = (unsigned char *)NULL;
		imP->sys_mask = (unsigned char *)NULL;
	}
}

/*****************************************************************************
 *
 * ax_tm_gimage - TmGetImage function
 */
ax_tm_gimage(w, pid, pnts, use_color, imP, cmap)
long w;
int pid;
TmPoint32 pnts[2];
int use_color;
ge_image_type *imP;
cmap_type *cmap;
{
    TmPoint32 p1, p2, points[5];
    ge_image_transform_type tr;
    int wid, hyt;
    int x_off, y_off;
    poly_type xform;
    XImage *xP;
    Drawable xwin;

    /* init transform factors */
    points[0].x = pnts[0].x;
    points[0].y = pnts[0].y;
    points[1].x = pnts[1].x;
    points[1].y = pnts[0].y;
    points[2].x = pnts[1].x;
    points[2].y = pnts[1].y;
    points[3].x = pnts[0].x;
    points[3].y = pnts[1].y;
    points[4].x = pnts[0].x;
    points[4].y = pnts[0].y;
    tr.x_scale     = 1.0;
    tr.y_scale     = 1.0;
    xform.nPoints  = 5;
    xform.nCtrls   = 0;
    xform.points   = points;
    xform.controls = (TmPoint32 *)0;

    /* convert units from mils to screen pixels */
    conv_i_p32(&pnts[0].x, &pnts[0].y, &p1.x, &p1.y, MILLS_TO_PIXEL);
    conv_i_p32(&pnts[1].x, &pnts[1].y, &p2.x, &p2.y, MILLS_TO_PIXEL);
	memset(imP, 0, sizeof(ge_image_type));
    imP->wid = wid = p2.x - p1.x;
    imP->hyt = hyt = p2.y - p1.y;
    imP->dep = use_color ? 8 : 1;
    imP->eps_seg = NULL;
    imP->eps_seg_size = 0;

    /* always grab a full depth Ztype image */
#if ALLOW_JAVA_GE
    if ( TaskIsJavaBased(pid) && FALSE )
        xP = JavaGetImage(pid, p1.x, p1.y, wid, hyt);
    else
#endif
    {   
        xwin = (Drawable)AxGetDrawableFromWidget((AXWidget)w, &x_off, &y_off);
        xP = XGetImage(Dpy, xwin, p1.x, p1.y, wid, hyt, AllPlanes, ZPixmap);
#if AXMS
        GetWinImageData(xP);
#endif
    } 
    if ( xP == (XImage *)0 )
    {
        imP->data = (unsigned char *)0;
        imP->mask = (unsigned char *)0;
        imP->sys_data = (unsigned char *)0;
        imP->sys_mask = (unsigned char *)0;
    }
    else
    {
        /* create a colormap */
        cmap->n_colors = 0;
        cmap->color = (color_type *)0;

        /* convert the sys-image to a sys-independant image */
        if ( use_color )
        {
            /* the image data */
            imP->data = pad_out_image(pid, xP, &imP->data_byte_wid, cmap);

            /* the mask image */
            imP->mask = flatten_image(pid, xP, &imP->mask_byte_wid);

            /* the system image and mask */
            imP->sys_data = (unsigned char *)NULL;
            imP->sys_mask = (unsigned char *)NULL;
        }
        else
        {
            /* the image data */
            imP->data = flatten_image(pid, xP, &imP->data_byte_wid);

            /* the mask image */
            imP->mask = (unsigned char *)NULL;
            imP->mask_byte_wid = 0;

            /* the system image and mask */
            imP->sys_data = (unsigned char *)NULL;
            imP->sys_mask = (unsigned char *)NULL;
        }
        XDestroyImage(xP);
    }
}

/*****************************************************************************
 *
 * get_points_from_xform
 */
XPoint *
get_points_from_xform(outline, x1, y1, x_scale, y_scale, npoints)
poly_type *outline;
int x1, y1;
double x_scale, y_scale;
int *npoints;
{
    int i, x, y, minx, miny;
    XPoint *points;

    points = (XPoint *)malloc(outline->nPoints * sizeof(XPoint));
    *npoints = outline->nPoints;

    for ( i=0; i<(*npoints); i++ )
    {
        conv_i_p32(&outline->points[i].x, &outline->points[i].y,
                   &x, &y,
                   MILLS_TO_PIXEL);
        points[i].x = x;
        points[i].y = y;
        if ( i == 0 )
        {
            minx = x;
            miny = y;
        }
        else
        {
            if ( x < minx )
                minx = x;
            if ( y < miny )
                miny = y;
        }
    }

    /* zoom the size and set the position */
    for ( i=0; i<(*npoints); i++ )
    {
        points[i].x = (points[i].x - minx) * x_scale + x1;
        points[i].y = (points[i].y - miny) * y_scale + y1;
    }
    return(points);
}

/*****************************************************************************
 *
 * display_backup_image - TmPutImage utility function
 */
display_backup_image(xwin, x1, y1, outline, x_scale, y_scale)
Drawable xwin;
int x1, y1;
poly_type *outline;
double x_scale, y_scale;
{
    int i, x, y, minx, miny, npoints;
    XPoint *points;
    Pixmap map;

    if ( !outline )
        return;

    map = getBadImagePixmap();
    if ( !map )
        return;

    /* convert to pixels */
    points = (XPoint *)get_points_from_xform(outline, x1, y1,
                                             x_scale, y_scale, &npoints);

    /* draw */
    XSetState(Dpy, Gc1, AlisPixels[0], AlisPixels[1], GXcopy, AllPlanes);
    XSetFillStyle (Dpy, Gc1, FillOpaqueStippled);
    XSetStipple(Dpy, Gc1, map);
    XFillPolygon(Dpy, xwin, Gc1, points, npoints, Convex, CoordModeOrigin);

    /* done */
    free(points);
}

/*****************************************************************************
 *
 * display_mono_image - TmPutImage utility function
 */
display_mono_image(xwin, xform, x, y,x_scale,y_scale, xP, clip_rect, attr)
Drawable xwin;
poly_type *xform;
int x, y;
double x_scale, y_scale;
XImage *xP;
XRectangle *clip_rect;
attr_type *attr;
{
    unsigned long gc_mask, fg, bg;
    int invert, mode, ix, np;
    XGCValues gcv;
    Pixmap mask;
    GC gc;
    XRectangle mask_rect;
    int i,j,number;
    fill_rect  new_rect;
    XPoint *points;

    /* don't do anything if there is an image problem */
    if ( xP->width == 0 || xP->height == 0 )
        return;
    if ( attr->ffg.see_thru && attr->fbg.see_thru )
        return;

    /* try a pixmap first */
    mask = AxCreateQueryPixmap(Dpy, xwin, xP->width, xP->height, 1);
    if ( mask )
    {
        /* clear the mask */
        gc_mask = GCForeground | GCBackground;
        memset((char *)&gcv, 0, sizeof(XGCValues));
        gcv.foreground = 0;
        gcv.background = 1;
        gc = axCreateGC(Dpy, mask, DefaultScreen(Dpy), 1,
                        (XFontStruct *) 0, gc_mask, &gcv);
        XFillRectangle(Dpy, mask, gc, 0, 0, xP->width, xP->height);

        /* draw the image onto the mask */
        invert = attr->ffg.see_thru ? TRUE : FALSE;
        if ( !invert )
        {
            XSetForeground(Dpy, gc, 1);
            XSetBackground(Dpy, gc, 0);
        }
        XPutImage(Dpy, mask, gc, xP, 0, 0, 0, 0, xP->width, xP->height);

        /* set the pixel colors */
        if ( attr->use_ffg )
            fg = AlisPixels[TMA_BLACK];
        else
        {
            AxGetRealPixel(2, attr->ffg.c,
                                  attr->ffg.m,
                                  attr->ffg.y,
                                  attr->ffg.k, 255, &fg);
        }
        if ( attr->use_fbg )
            bg = AlisPixels[TMA_WHITE];
        else
        {
            AxGetRealPixel(2, attr->fbg.c,
                                  attr->fbg.m,
                                  attr->fbg.y,
                                  attr->fbg.k, 255, &bg);
        }
        if ( invert )
        {
            XSetForeground(Dpy, Gc1, bg);
            XSetBackground(Dpy, Gc1, fg);
        }
        else
        {
            XSetForeground(Dpy, Gc1, fg);
            XSetBackground(Dpy, Gc1, bg);
        }
        XSetFunction(Dpy, Gc1, GXcopy);

        /* prepare a stipple to use as a fill rect pattern */
        if ( attr->ffg.see_thru || attr->fbg.see_thru )
            XSetFillStyle(Dpy, Gc1, FillStippled);
        else
            XSetFillStyle(Dpy, Gc1, FillOpaqueStippled);
        XSetStipple(Dpy, Gc1, mask);
        XSetTSOrigin(Dpy, Gc1, x, y);

        /* draw a rect using the image as a stipple */
        XSetClipRectangles(Dpy, Gc1, 0, 0, clip_rect, 1, Unsorted);

        if ( xform )
        {
            points = (XPoint *)get_points_from_xform(xform, x, y,
                                                 x_scale, y_scale, &np);

            XFillPolygon(Dpy, xwin, Gc1, points, np, Convex, CoordModeOrigin);
            free(points);
        }
        else
            XFillRectangle(Dpy, xwin, Gc1, x, y, xP->width, xP->height);

        /* done */
        XFreePixmap   (Dpy, mask);
        XFreeGC       (Dpy, gc);
    }

    else /* then try image mode */
    {
        /* lay down the mask to mask out non-see thru parts */
        invert = attr->ffg.see_thru ? TRUE : FALSE;
        XSetClipRectangles(Dpy, Gc1, 0, 0, clip_rect, 1, Unsorted);

        if ( attr->ffg.see_thru || attr->fbg.see_thru )
        {
            mode = invert ? GXandInverted : GXand;
            XSetState(Dpy, Gc1, ~AllPlanes, AllPlanes, mode, AllPlanes);
            XPutImage(Dpy, xwin, Gc1, xP, 0, 0, x, y, xP->width, xP->height);
        }

        /* set the pixel colors */
        if ( attr->use_ffg )
            fg = AlisPixels[TMA_BLACK];
        else
        {
            AxGetRealPixel(2, attr->ffg.c,
                                  attr->ffg.m,
                                  attr->ffg.y,
                                  attr->ffg.k, 255, &fg);
        }
        if ( attr->use_fbg )
            bg = AlisPixels[TMA_WHITE];
        else
        {
            AxGetRealPixel(2, attr->fbg.c,
                                  attr->fbg.m,
                                  attr->fbg.y,
                                  attr->fbg.k, 255, &bg);
        }
        if ( attr->ffg.see_thru )
            XSetState(Dpy, Gc1, 0, bg, GXor, AllPlanes);
        else if ( attr->fbg.see_thru )
            XSetState(Dpy, Gc1, fg, 0, GXor, AllPlanes);
        else
            XSetState(Dpy, Gc1, fg, bg, GXcopy, AllPlanes);

        /* display the sys image */
        XPutImage(Dpy, xwin, Gc1, xP, 0, 0, x, y, xP->width, xP->height);
    }
}

/*****************************************************************************
 *
 * display_color_image - TmPutImage utility function
 */
display_color_image(xwin, x, y, xP, mP,clip_rect, attr)
Drawable xwin;
int x, y;
XImage *xP, *mP;
XRectangle *clip_rect;
attr_type *attr;
{
    unsigned long gc_mask;
    XGCValues gcv;
    Pixmap mask;
    GC gc;
    XRectangle mask_rect;
    int i,j,number;
    int src_x,src_y,dst_x,dst_y;
    fill_rect  new_rect;

    /* display the masked sys image */
    XSetClipMask(Dpy, Gc1, None);
    /* set the color defaults (for mono displays only) */
    XSetForeground(Dpy, Gc1, BlackPixel(Dpy, DefaultScreen(Dpy)));
    XSetBackground(Dpy, Gc1, WhitePixel(Dpy, DefaultScreen(Dpy)));

    /* try a pixmap first */
    mask = AxCreateQueryPixmap(Dpy, xwin, xP->width, xP->height, 1);
    if ( mask )
    {
        /* clear the mask */
        gc_mask = GCForeground | GCBackground;
        memset((char *)&gcv, 0, sizeof(XGCValues));
        gcv.foreground = 0;
        gcv.background = 0;
        gc = axCreateGC(Dpy, mask, DefaultScreen(Dpy), 1,
                        (XFontStruct *) 0, gc_mask, &gcv);
        XSetClipMask(Dpy, gc, None);
        XFillRectangle(Dpy, mask, gc, 0, 0, xP->width, xP->height);

        /* draw the image onto the mask */
        XSetForeground(Dpy, gc, 1);
        XSetBackground(Dpy, gc, 0);
        if ( mP )
            XPutImage(Dpy, mask, gc, mP, 0,0, 0,0, mP->width,mP->height);
        else
            XFillRectangle(Dpy, xwin, Gc1, 0,0, xP->width,xP->height);

        /* calc the intersection of clip and mask */
        XSetForeground(Dpy, gc, 0);
        mask_rect.x = x;
        mask_rect.y = y;
        mask_rect.width = xP->width;
        mask_rect.height = xP->height;
        new_rect.rect = (XRectangle *)NULL;
        number = clip_mask( &mask_rect, clip_rect, &new_rect );

        /* clear the clipped sections of the mask */
        switch( number )
        {
            case 1:
            case 2:
            case 3:
            case 4:
                    for(i=0;i< number;i++)
                    {
                        XFillRectangle( Dpy, mask, gc,
                            new_rect.rect[i].x-x, new_rect.rect[i].y-y,
                            new_rect.rect[i].width, 
                            new_rect.rect[i].height);
                    } 

            break;
            case 0:
            default:
            break;
        }
        if( new_rect.rect )
            free( new_rect.rect );

        /* load the mask and put the image thru it */
        XSetClipOrigin(Dpy, Gc1, x, y);
        XSetClipMask(Dpy, Gc1, mask);
        XSetFunction(Dpy, Gc1, GXcopy);
        src_x = 0;
        src_y = 0;
        dst_x = x;
        dst_y = y;
        if(x < 0 )
        {
            src_x = src_x - x;
            dst_x = dst_x - x;
        }
        if(y < 0 )
        {
            src_y = src_y - y;
            dst_y = dst_y - y;
        }
        XPutImage(Dpy, xwin, Gc1, xP, src_x,src_y, 
                    dst_x,dst_y, xP->width,xP->height);

        /* done */
        XFreePixmap(Dpy, mask);
        XFreeGC(Dpy, gc);
    }
    else /* try client side (image) mode masking */
    {
        /* lay down the mask to mask out non-see thru parts */
        XSetState(Dpy, Gc1, ~AllPlanes, AllPlanes, GXand, AllPlanes);
        if ( mP )
            XPutImage(Dpy, xwin, Gc1, mP, 0,0, x,y,mP->width,mP->height);
        else
            XFillRectangle(Dpy, xwin, Gc1, x,y, xP->width,xP->height);

        /* now OR the image into the masked out area */
        XSetFunction(Dpy, Gc1, GXor);
        XPutImage(Dpy, xwin, Gc1, xP, 0, 0, x, y, xP->width, xP->height);
    }
}
ax_display_image(xwin, x, y, xP, attr)
Drawable xwin;
int x, y;
XImage *xP;
attr_type *attr;
{
#if AXMS
	axmsFlipXImage(xP);
#endif;
    XPutImage(Dpy, xwin, Gc1, xP, 0, 0, x, y, xP->width, xP->height);
}

/****************************************************************************
 *
 * ax_tm_pimage - TmPutImage function
 */
ax_tm_pimage(w, pid, pnt, imP, attr, xform, x_scale, y_scale, cmap, clip)
long w;
int pid;
TmPoint32 *pnt;
ge_image_type *imP;
attr_type *attr;
poly_type *xform;
double x_scale,y_scale;
cmap_type *cmap;
TmRect32 *clip;
{
    int x1, y1, x2, y2, tmp_im, x_off, y_off;
    ge_image_transform_type tr;
    XRectangle c_rect;
    XImage *xP, *mP;
    Drawable xwin;

    /* nothing to do */
    if ( imP->dep == 1 && attr->ffg.see_thru && attr->fbg.see_thru )
        return;

    /* sanity tests */
    if ( has_image_data(imP) == FALSE )
        return;

    /* if we need them, create a scaled sys image and a scaled sys mask */
    tmp_im = (imP->sys_data == (unsigned char *)0) ? TRUE : FALSE;
    if ( tmp_im )
        ax_tm_simage(0, imP, cmap, xform, x_scale, y_scale);

    /* get the clipping rect */
    x1 = clip->x_position;
    y1 = clip->y_position;
    x2 = clip->x_size + x1 - 1;
    y2 = clip->y_size + y1 - 1;
    conv_i_p32(&x1, &y1, &x1, &y1, MILLS_TO_PIXEL);
    conv_i_p32(&x2, &y2, &x2, &y2, MILLS_TO_PIXEL);
    c_rect.x = x1;
    c_rect.y = y1;
    c_rect.width  = x2 - x1 + 1;
    c_rect.height = y2 - y1 + 1;

    /* get the images and their pixel positions */
    conv_i_p32(&pnt->x, &pnt->y, &x1, &y1, MILLS_TO_PIXEL);

    /* draw images on the server */
    if ( has_sys_image_data(imP) )
    {
    	/* display the image locally */
    	xP = (XImage *)imP->sys_data;
    	mP = (XImage *)imP->sys_mask;
    	xwin = (Drawable)AxGetDrawableFromWidget((AXWidget)w, &x_off, &y_off);
#if AXMS
    	display_win_image(xwin, pid, imP, cmap, x1, y1, xP, mP, &c_rect, attr);
#else
    	if ( !xP )
    		display_backup_image(xwin, x1, y1, xform, x_scale, y_scale);
    	else if ( xP->width > 0 && xP->height > 0 )
    	{
    		if ( imP->dep == 1 )
    			display_mono_image(xwin, xform, x1, y1,
    			                   x_scale, y_scale, xP, &c_rect, attr);
    		else
    			display_color_image(xwin, x1, y1, xP, mP, &c_rect, attr);
    	}
        WMUnsetClip();
#endif
    }

	/* if the image has no device specific data, free it now before it leaks */
    if ( tmp_im )
    {
        if ( imP->sys_data )
        {
            free_sys_image((unsigned char *)imP->sys_data, pid);
            imP->sys_data = (unsigned char *)0;
        }
        if ( imP->sys_mask )
        {
            free_sys_image((unsigned char *)imP->sys_mask, pid);
            imP->sys_mask = (unsigned char *)0;
        }
    }
}

/****************************************************************************
 *
 * do_ax_pimage - TmPutImage function
 */
do_ax_pimage(xwin, pid, pnt, imP, attr, xform, x_scale, y_scale, cmap)
Drawable xwin;
int pid;
TmPoint32 *pnt;
ge_image_type *imP;
attr_type *attr;
poly_type *xform;
double x_scale,y_scale;
cmap_type *cmap;
{
    int x1, y1, x2, y2, tmp_im, x_off, y_off;
    ge_image_transform_type tr;
    XImage *xP;

    /* nothing to do */
    if ( imP->dep == 1 && attr->ffg.see_thru && attr->fbg.see_thru )
        return;

    /* sanity tests */
    if ( has_image_data(imP) == FALSE )
        return;

    /* if we need them, create a scaled sys image and a scaled sys mask */
    tmp_im = (imP->sys_data == (unsigned char *)0) ? TRUE : FALSE;
    if ( tmp_im )
        ax_tm_simage(0, imP, cmap, xform, x_scale, y_scale);

    /* get the images and their pixel positions */
	conv_i_p32(&pnt->x, &pnt->y, &x1, &y1, MILLS_TO_PIXEL);

    /* draw images on the server */
    if ( has_sys_image_data(imP) )
    {
    	/* display the image locally */
    	xP = (XImage *)imP->sys_data;
    	if ( !xP )
    		display_backup_image(xwin, x1, y1, xform, x_scale, y_scale);
    	else if ( xP->width > 0 && xP->height > 0 )
    	{
    		if ( imP->dep == 1 )
    			display_mono_image(xwin, xform, x1, y1,
    			                   x_scale, y_scale, xP, NULL, attr);
    		else
    			ax_display_image(xwin, x1, y1, xP, attr);
    	}
    }

	/* if the image has no device specific data, free it now before it leaks */
    if ( tmp_im )
    {
        if ( imP->sys_data )
        {
            free_sys_image((unsigned char *)imP->sys_data, pid);
            imP->sys_data = (unsigned char *)0;
        }
        if ( imP->sys_mask )
        {
            free_sys_image((unsigned char *)imP->sys_mask, pid);
            imP->sys_mask = (unsigned char *)0;
        }
    }
}

/****************************************************************************\
*                                SS functions                                *
\****************************************************************************/


/****************************************************************************/
/*
*   ssTransPixmaps - read ss plates widget pixmaps and translate colors
*                   appropriately, ONLY done once during startup only
*/
ssTransPixmaps()
{
#if !AXMS
/* On NT/95 we don't use pixmaps to draw the tab buttons */

static int  firstTime=TRUE;
GC colorGC;
int left;


    if(firstTime)   {
        if(UseWidgetColors)
            colorGC = RecessGC;
        else
            colorGC = ForeGC;

        left = TRUE;

        /* left/right pixmaps for active folder tab */
        if(UseWidgetColors) {
            ssTransXMap(NXLT("ssfrltab"), SunGC, BackGC, BackGC, left);
            ssTransXMap(NXLT("ssfrrtab"), ShadowGC, BackGC, BackGC, !left);
        } else  {
            ssTransXMap(NXLT("ssbwltab"), SunGC, BackGC, BackGC, left);
            ssTransXMap(NXLT("ssbwrtab"), ShadowGC, BackGC, BackGC, !left);
        }

        /* left/right pixmaps for in-active folder tab */
        ssTransXMap(NXLT("ssbkltab"), SunGC, colorGC, BackGC, left);
        ssTransXMap(NXLT("ssbkrtab"), ShadowGC, colorGC, BackGC, !left);
    }
    firstTime = FALSE;
#endif
}

/*
*   ssTransXMap - translate colors in the named pixmap
*   Algorithm:
*       The pixmaps have a line going from bottom left to top right. All
*       the pixels to the left of the line are translated to intgc pixels,
*       all pixels on the line are translated to foregc pixels, and all pixels
*       to the right of the line are translated to backgc pixels. 
*       if left is false, the line goes from bottom right to top left.
*/
ssTransXMap(name, foregc, backgc, intgc, left)
char *name;
GC  foregc,     /* the sun/shadow gc */
    backgc,     /* the recess/back gc */
    intgc;      /* the gc used for the panel containing the tabs */
int left;       /* if TRUE the line goes from left/bottom to right/top, */
                /* else, the line goes from right/bottom to left/top */
{
pixcache_t *pxc, *Pixmapper();
Pixmap pix, pixmap;
XImage *ximg;
XGCValues values;
unsigned long fg, this_fg, bg, intfg, tmp, ref_fg;
int x,y,found;
int back_y_pos;


    pxc = Pixmapper(name);
    if(!pxc)
        return;
    XGetGCValues(Dpy, foregc, GCForeground, &values);
    fg = values.foreground;
    XGetGCValues(Dpy, backgc, GCForeground, &values);
    bg = values.foreground;
    XGetGCValues(Dpy, intgc, GCForeground, &values);
    intfg = values.foreground;

    /* get an image out of the source pixmap */
    ximg = XGetImage(Dpy, pxc->pix,
                0, 0, pxc->width, pxc->height,
                AllPlanes, ZPixmap);
#if AXMS
    if ( ximg->depth == 16 )
        {
        fg = Get16BitColor(fg);
        bg = Get16BitColor(bg);
        intfg = Get16BitColor(intfg);
        }

    /* clear windows bitmap handle so that XPutImage *
     * creates a new one, using the new data.        */
    ximg->bitmap_handle = (HBITMAP)NULL;

    /* on Windows all of the images are flipped over vertically, so the */
    /* pixel which corresponds to the X pixel is on the bottom row of the */
    /* image, not the top row                                          */
    back_y_pos = pxc->height - 1;
#else
    back_y_pos = 0;
#endif

    /* the bitmap contains pixels with two colors, one color for the */
    /* line in the bitmap and a different color for background */
    /* we extract the pixel for the background and use it a reference to */
    /* find out the pixels on the line, ref_fg is pixel used for background */
    /* in the bitmap */
    if(left)    
        ref_fg = XGetPixel(ximg, 0, back_y_pos );
    else    {
        /* we use the same algorithm as for left except that the colors */
        /* to the left and right of the line have to be swapped */
        ref_fg = XGetPixel(ximg, ximg->width-1, back_y_pos );
        /*swap bg and intfg */
        tmp = intfg;
        intfg = bg;
        bg = tmp;
    }

    /* for each scan line, go across and translate pixels */
    for(y=0; y<ximg->height; y++)   {
        found = FALSE;

        /* process the scan line */ 
        for(x=0; x<ximg->width; x++)    {

            /* get the pixel from image */
            this_fg = XGetPixel(ximg, x, y);
            
            if(this_fg != ref_fg) {
                /* found a pixel on the line, translated it to foregc pixel */
                XPutPixel(ximg, x, y, fg);
                if(!found)
                    found = TRUE;   /* we already translated pixel on the line*/
            } else if(found)    {
                /* translate this to backgc pixel */
                XPutPixel(ximg, x, y, bg);
            } else  {
                /* we haven't found a pixel on the line, translate it */
                /* to intermediate gc pixel */
                XPutPixel(ximg, x, y, intfg);
            }
        }
    }

#if AXMS
    /* write the modified image onto the pixmap */
    if ( ximg->depth == 16 || ximg->depth == 32 )
        {
        if ( ximg->bitmap_info )
           {
           /* we have an existing bitmap header, which will be used to */
           /* re-create the bitmap.  Make sure the header reflects the */
           /* default 5/5/5 bit encoding we stuffed in above           */
           ximg->bitmap_info->bmiHeader.biCompression = BI_RGB;
           }
        }
#endif
    XPutImage(Dpy, pxc->pix, foregc, ximg, 0,0, 0,0, ximg->width, ximg->height);
    /* destory the image */
    XDestroyImage(ximg);
}

/****************************************************************************
 *
 * update_image_from_cmap
 */
update_image_from_cmap(x_im, cmap, pid)
XImage *x_im;
cmap_type *cmap;
int pid;
{
    register int y, x, src_wid, src_hyt, bwid;
    int i, n, ix, sz, color_ix[256];
    char *imP, *dstP;
	unsigned char bitmask;
    int r,g,b, rs,gs,bs, rm,gm,bm;
    unsigned long pel;
    unsigned long (*func)();

    /* read the image */
    src_wid = x_im->width;
    src_hyt = x_im->height;

    if ( DpyDepth == 1 )
	{
		bwid = (x_im->width+ 7)/8;
		sz = x_im->height * bwid * sizeof(char);
		imP = (char *)TaskAlloc(pid, sz);
		memset(imP, 0, sz);
		bitmask = 0x80;
		ix = 0;
		pel = AlisPixels[0];

        for ( y=0; y<src_hyt; y++ )
        {
            dstP = &imP[bwid * y];
            for ( x=0; x<src_wid; x++ )
            {
                if ( pel != XGetPixel(x_im, x, y) )
					ix |= bitmask;
				if ( (bitmask>>=1) == 0x00 )
				{
					*dstP++ = ix;
					bitmask = 0x80;
					ix = 0;
				}
            }
			if ( bitmask & 0x7f )
				*dstP++ = ix;
        }
        free(x_im->data);
        x_im->data = imP;
        x_im->bytes_per_line = bwid;
		x_im->depth = 1;
	}
    else if ( DpyDepth == 8 )
    {
		/* create a pixel to colormap_index table */
		n = cmap->n_colors;
		bwid = x_im->width + x_im->width%2;
		sz = x_im->height * bwid * sizeof(char);
		imP = (char *)TaskAlloc(pid, sz);
		memset(imP, 0, sz);
        for ( i=0; i<n; i++ )
        {
#if AXMS
            axWinGetImageColormapIndex(cmap->color[i].c, cmap->color[i].m,
                            cmap->color[i].y, cmap->color[i].k, x_im, &ix);
            if ( cmap->color[i].see_thru )
                color_ix[AxPixels[ix]] = i;
            else
                color_ix[ix] = i;
#else
            axGetColormapIndex(2, cmap->color[i].c,
                                  cmap->color[i].m,
                                  cmap->color[i].y,
                                  cmap->color[i].k,
                                  255,
                                  &ix);
            color_ix[AxPixels[ix]] = i;
#endif
        }

        for ( y=0; y<src_hyt; y++ )
        {
            dstP = &imP[bwid * y];
            for ( x=0; x<src_wid; x++ )
            {
                pel = AXGETPIXEL08(x_im, x, y);
                if ( pel == AlisPixels[0] )
                    *dstP++ = 0;
                else if ( pel == AlisPixels[1] )
                    *dstP++ = 1;
                else
                    *dstP++ = color_ix[pel];
            }
        }
        free(x_im->data);
        x_im->data = imP;
        x_im->bytes_per_line = bwid;
    }
    else if ( DpyDepth > 8 )
    {
		bwid = x_im->width + x_im->width%2;
		sz = x_im->height * bwid * sizeof(char);
		imP = (char *)TaskAlloc(pid, sz);
		memset(imP, 0, sz);

        /* set the mask's and shift's for reading the pixels */
#if AXMS || LINUX
        rm = x_im->red_mask;
        gm = x_im->green_mask;
        bm = x_im->blue_mask;
        func = AxGetPixel;
#else
        if (24 == DpyDepth) {
            rm = 0x0000FF;
            gm = 0x00FF00;
            bm = 0xFF0000;
            func = NULL;
        }
		else if (16 == DpyDepth) {
			rm = x_im->red_mask;
			gm = x_im->green_mask;
			bm = x_im->blue_mask;
			func = NULL;
			if (0 == rm && 0 == gm && 0 == bm) {
				/* cheat, and read masks from display directly */
				Screen *ss = XScreenOfDisplay(Dpy, 0);
				Visual *vv = ss->root_visual;

				rm = vv->red_mask;
				gm = vv->green_mask;
				bm = vv->blue_mask;

				x_im->red_mask = rm;
				x_im->green_mask = gm;
				x_im->blue_mask = bm;
			}
		}
        else {
            rm = x_im->red_mask;
            gm = x_im->green_mask;
            bm = x_im->blue_mask;
            func = GetPixelArray[DpyDepth].func;
        }
#endif
        rs = ExtractShift_from_mask(rm);
        gs = ExtractShift_from_mask(gm);
        bs = ExtractShift_from_mask(bm); 

        /* read the image */
        src_wid = x_im->width;
        src_hyt = x_im->height;

        for ( y=0; y<src_hyt; y++ )
        {
            /* point to the destination scanline... */
            dstP = &imP[bwid * y];
         
            /* read the source scanline into the dest scanline */
            for ( x=0; x<src_wid; x++ )
            {
                pel = (NULL == func) ? XGetPixel(x_im,x,y) : (*func)(x_im,x,y);
                r = (pel & rm) >> rs;
                g = (pel & gm) >> gs;
                b = (pel & bm) >> bs;
#if AXMS
                if ( DpyDepth == 15 || DpyDepth == 16 ) {
                    /* normalize from 16-bit rgb vals to "normal" vals. */
                    r = ((double)r/(double)0x1F) * 255.0;
                    g = ((double)g/(double)0x1F) * 255.0;
                    b = ((double)b/(double)0x1F) * 255.0;
                }
#endif
                *dstP++ = getColorIndex(r, g, b);
            }
        }
        free(x_im->data);
        x_im->data = imP;
        x_im->bytes_per_line = bwid;
        x_im->depth = 8;
    }
}

/****************************************************************************
 *
 * AxCreateColorPixmap
 */
Pixmap
AxCreateColorPixmap(dat,wid,hyt,cmap)
char *dat;
int wid, hyt;
cmap_type *cmap;
{
    ge_image_type imP;
    ge_image_transform_type tr;
    XImage *ximage;
    GC gc;
    XGCValues gcv;
    Pixmap pix;

    memset(&imP, 0, sizeof(ge_image_type));
    imP.wid = wid;
    imP.hyt = hyt;
    imP.data = (unsigned char *)dat;
    imP.data_byte_wid = wid+wid%2;

    tr.x_scale = 1;
    tr.y_scale = 1;
    extract_factors(&imP, NULL, &tr);
    ximage = (XImage *)tr_sys_color_image(0, &imP, cmap, &tr);
    if ( ximage == (XImage *)0 )
        return (Pixmap) NULL;
    pix = (Pixmap)AxCreateQueryPixmap(Dpy, AxTopWinNeverMapped,
                                           wid, hyt, DpyDepth);
    gcv.foreground = 1;
    gcv.background = 0;
    if (! (gc = axCreateGC(Dpy, pix, DefaultScreen(Dpy),
                           DpyDepth, (XFontStruct *) 0,
                           GCForeground|GCBackground, &gcv)))
        return (Pixmap) NULL;
#if AXMS
	axmsFlipXImage(ximage);
#endif;
    XPutImage(Dpy, pix, gc, ximage, 0, 0, 0, 0, wid, hyt);

    /* done */
    XFreeGC(Dpy, gc);
    free_sys_image((unsigned char *)ximage, 0);

    return(pix);
}

/****************************************************************************
 *
 * AxCreateMonoPixmap
 */
Pixmap
AxCreateMonoPixmap(src_image, drw)
XImage *src_image;
Drawable drw;
{
    ge_image_transform_type tr;
    unsigned long gc_mask;
	ge_image_type imP;
    XImage *ximage;
    XGCValues gcv;
    Pixmap pix;
    GC gc;

	/* the mask image */
	memset(&imP, 0, sizeof(ge_image_type));
    imP.wid = src_image->width;
    imP.hyt = src_image->height;
	imP.data = flatten_image(0, src_image, &imP.data_byte_wid);

    tr.x_scale = 1;
    tr.y_scale = 1;
    extract_factors(&imP, NULL, &tr);
	ximage = (XImage *)tr_sys_mono_image(0, &imP, NULL, &tr);
    if ( ximage == (XImage *)0 )
        return (Pixmap) NULL;

#if AXMS
	axmsFlipXImage(ximage);
#endif;
    pix = (Pixmap)AxCreateQueryPixmap(Dpy, drw, imP.wid, imP.hyt, 1);
	/* clear the mask */
	gc_mask = GCForeground | GCBackground;
	memset((char *)&gcv, 0, sizeof(XGCValues));
	gcv.foreground = 0;
	gcv.background = 1;
	gc = axCreateGC(Dpy, pix, DefaultScreen(Dpy), 1,
					(XFontStruct *) 0, gc_mask, &gcv);
	XFillRectangle(Dpy, pix, gc, 0, 0, imP.wid, imP.hyt);

	/* draw the image onto the mask */
	XSetForeground(Dpy, gc, 1);
	XSetBackground(Dpy, gc, 0);
	XPutImage(Dpy, pix, gc, ximage, 0, 0, 0, 0, ximage->width, ximage->height);

    /* done */
    XFreeGC(Dpy, gc);
    free_sys_image((unsigned char *)ximage, 0);

    return(pix);
}

/****************************************************************************
 *
 * gJavaGetImageFromPixmap
 */
data_ptr
gJavaGetImageFromPixmap(wP, pid, x, y, wid, hyt)
data_ptr wP;
int pid, x, y;
int wid, hyt;
{
	extern data_ptr getJavaBitmapForMaskBits();
	int x_off, y_off;
	data_ptr dataP;
	Pixmap pmap;
	XImage *ximg;
	AXWidget w;

	w = (AXWidget)WidgetPointerFromHandleData(wP);
    pmap = AxGetDrawableFromWidget((AXWidget)w, &x_off, &y_off);

	ximg = XGetImage(Dpy, pmap, 0, 0, wid, hyt, AllPlanes, ZPixmap);
#if AXMS
	axmsFlipXImage(ximg);
	ximg->bitmap_handle = NULL;
#endif;
	dataP = getJavaBitmapForMaskBits(ximg, x, y);
	XDestroyImage(ximg);
	return(dataP);
}
