/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xmtable.c
   Applix ASC version: 220
   Date: Thu Jun  3 21:16:26 1999
*/
#include <basicwin.h>
#if AX_X11
#include <X11/keysym.h>
#endif
#include <ax_xm.h>
#include <aos.h>
#include <dwt2xm.h>
#include <xmlib.h>
#include <elxwins.h>
#include <wchars.h>
#include <thimble.h>
#include <ctype.h>
#include <elf_env.h>
#include <eldata.h>
#include <evreason.h>
#include <xmpixel.h>
#include <xmtable.h>
#include <axtext.h>
#include <tmkeys.h>

#define MIN_COL_DIM 15  /* FIXME, min column width for resize (in pix). */

#if AXMS
#define TBEVEL 1        /* bevel thickness for row/marker headers */
#else
#define TBEVEL 2
#endif

#ifdef ABS
#undef ABS
#endif
#define ABS(x) (((x) < 0) ? (-(x)) : (x))

#define ROW_MODE 0
#define COL_MODE 1

/* IS_NGC adapted from ATM's tmin1.c definition */
#define DEL ((unsigned) 0177)  /* ascii code */
#define IS_NGC(c) ((((unsigned) (c)) < (unsigned) ' ') || \
                   (((unsigned) (c)) == DEL))

/* Event mask used when the table widget window does not have focus.
   Prevent unnecessary MotionNotify events. */
#define TBL_NONFOCUS_MASK (TM_ALL_INPUTS_MASK)

/* Event mask used when the table widget window has focus.  Requests
   pointer motion. */
#define TBL_INFOCUS_MASK  ((TM_ALL_INPUTS_MASK) | PointerMotionMask)

/* Set `CLIP_DEBUG' to 1 in order to debug clipping rectangle problems.
   Note: Pulls in printf(); abort()s when it finds clipping errors!
   Use *only* for debugging. */
#define CLIP_DEBUG 0

/* Set `TBL_DEBUG' to 1 to pull in table-widget logging.  Prints
   logging to stdout.  Use *only* for debugging. */
#define TBL_DEBUG 0
#if TBL_DEBUG
#define TBL_LOG(m) TblLogMsg m
#else
#define TBL_LOG(m)
#endif


AXWidget_table TableArmed;

extern int WidgetNeedsUpdated;  /* Set when a widget needs repainted */
extern int AxCharAbove, AxCharHeight, AxCharBelow;
extern int AxBigFont;

extern int UseWidgetColors;

extern Display *Dpy;
extern short DpyDepth;
extern int AxButtonDown;

extern long WinMask;
extern XSetWindowAttributes WinAtt;
extern Pixmap UnknownPixmap;
extern AXWidget ClickWidget;
extern unsigned long BackWorkPixel;
extern Colormap WorkColormap;
extern GC ForeGC, BackGC;
extern GC SunGC, ShadowGC, DisabledGC, GrayLineGC;
extern GC RecessGC;
extern GC olClrSunButtonGC, olClrShadowButtonGC, olClrRecessButtonGC;
extern GC MonospaceGC;       /* monspace system font, forground mode */
extern GC MonospaceBackGC;   /* monspace system font, background mode */
extern GC MonospaceBoldGC, MonospaceBoldBackGC;
extern GC MonospaceItalGC, MonospaceItalBackGC;
extern GC MonospaceItalBoldGC, MonospaceItalBoldBackGC;
extern XFontStruct *AxDlgMainFont;  /* Bolded dialog font */
extern XFontStruct *MonospaceFont, *MonospaceBoldFont;
extern XFontStruct *MonospaceItalFont, *MonospaceItalBoldFont;
extern int DoubleMillisec;
extern unsigned long ForePixel;
extern unsigned long BackPixel;
extern short ColorDisplay;
extern short NoColor;
extern Window RootWin;
extern Cursor ThimCrosshairCursor;
extern Cursor ThimPointerCursor;
extern Cursor ThimLeftRightCursor;
extern unsigned long AlisPixels[];
extern unsigned long WidgetPixels[];


static char dit[] = {1, 1};
static char dah[] = {6, 3};

static struct {
    int markerXpos;  /* resize marker (vert line) xpos */
    int colXpos;     /* resize column xpos */
} colResize;

static struct {
    int clips_on; /* if TRUE, clip description stored in here is real */
    int x_origin;
    int y_origin;
    int ordering;
    XRectangle crect; /* always use 1 rectangle */
} tblSaveClip;

#if TBL_DEBUG
static void TblLogMsg(/* int va_alist */);
#endif

#include <xmtable.p>

/***********************************************************************
function: FreeTableInfo - Free up all elf data in `table_info'. Called
from AxXtDestroyWidget()
*/
void
FreeTableInfo(wt)
table_info *wt;
{
    if (!wt)
        return;
    
    AxTaskFreeElfData(0, wt->rows);
    AxTaskFreeElfData(0, wt->colText);
    AxTaskFreeElfData(0, wt->colWidth);
    AxTaskFreeElfData(0, wt->colType);
    AxTaskFreeElfData(0, wt->marker_pixmaps);
    AxTaskFreeElfData(0, wt->corner_pixmaps);
    AxTaskFreeElfData(0, wt->marker_strings);
    AxTaskFreeElfData(0, wt->headingShadow);
    AxFreeWColor(wt->headingColor);
    if (wt->fontFamily) {
        free(wt->fontFamily);
        wt->fontFamily = (char *) NULL;
    }
    return;
}

/* Should really be TaskAddNumToArray() */
static data_ptr
TblAddNumToArray(taskId, ap, ix, n)
int taskId;
data_ptr ap;
int ix;
int n;
{
    return(TaskWriteArray(taskId, ap, ix, MK_INTEGER(n)));
}

/* Should really be TaskAddStrToArray() */
static data_ptr
TblAddStrToArray(taskId, ap, ix, str)
int taskId;
data_ptr ap;
int ix;
char *str;
{
    return(TaskWriteArray(taskId, ap, ix, AxTaskCreateElfString(taskId, str)));
}

/*************************************************************************
function: XmCreateTable - Create rows/column table
*/
AXWidget_table
XmCreateTable(parentWidget, name, args, cnt)
register AXWidget parentWidget;
register char *name;
register Arg args[];
register int cnt;
{
    register AXWidget_table w;
    register table_info *tp = (table_info *) NULL;

    if (args[0].value) {
        w = (AXWidget_table)args[0].value;
        if (AXmNewFootprint((AXWidget)w, name, args, cnt)) {
            UnMapWidget((AXWidget)w);
            if (w->vScrollBar) {
                UnMapWidget((AXWidget)w->vScrollBar);
                xmSetWidgetDisplay((AXWidget)w->vScrollBar, FALSE);
            }
            if (w->hScrollBar) {
                UnMapWidget((AXWidget)w->hScrollBar);
                xmSetWidgetDisplay((AXWidget)w->hScrollBar, FALSE);
            }
        }
        else {
            if (w->vScrollBar) {
                xmSetWidgetDisplay((AXWidget)w->vScrollBar, TRUE);
			}
            if (w->hScrollBar) {
                xmSetWidgetDisplay((AXWidget)w->hScrollBar, TRUE);
			}
        }
        w->update_widget = TRUE;
        WidgetNeedsUpdated = TRUE;
    }
    else {
        w = (AXWidget_table)CrtWidget(W_TABLE, parentWidget, name);
        w->info.table = (dummy_info *)calloc(1, sizeof(table_info));
        tp = (table_info *) w->info.table;
        /* Do special one-time initializations here.  These are
           done *only* in the fresh widget-create case, after which,
           the table tracks their change. */
        tp->totalRowCount = -1;

        /* Init cursor */
        tp->cursor.row = tp->cursor.col = -1;

        w->last_item_selected = -1;
        w->value = -1; /* button press item */

    }
    
    TblInitFonts(w);           /* default init, prior to setVals */

    w->painter = PaintTable;
    AxXtSetValues((AXWidget)w, args, cnt);

    TblCreateGCs(w);
    
    if (!tp)
        tp = (table_info *) w->info.table;
    tp->headingHeight = TblComputeHeadingHeight(w);
    tp->rowHeight = tp->char_height + 4 + 2;


    tp->pointer_shape = None;
    tp->colToResize = -1;

    return(w);
}

/***********************************************************************
function: TblUnmapScrollBars - Unmap scrollbars
*/
static void
TblUnmapScrollBars(w, vscroll, hscroll)
AXWidget_table w;
int vscroll;
int hscroll;
{
    if (!w || w->type != W_TABLE)
        return;
#if CLIP_DEBUG
    if (tblSaveClip.clips_on) {
        printf("TblUnmapScrollBars... clips are on!!\n");
        abort();
    }
#endif
    if (vscroll && w->vScrollBar) {
        UnMapWidget((AXWidget)w->vScrollBar);
        xmSetWidgetDisplay((AXWidget)w->vScrollBar, FALSE);
    }
    if (hscroll && w->hScrollBar) {
        UnMapWidget((AXWidget)w->hScrollBar);
        xmSetWidgetDisplay((AXWidget)w->hScrollBar, FALSE);
    }
    return;
}

/***********************************************************************
function: TblVfyFontSpec - Verify font specification.
Return 0 if font spec is OK; -1 otherwise.
*/
/*ARGSUSED*/
static int
TblVfyFontSpec(fam, size, weight, slant)
char *fam;
int size;
int weight;
int slant;
{
    int number = 0;
    char **font_strs;
    int ix, found;

    if (!fam)
        return(-1);
    /* for now */
    
    font_strs = axListFontFamilies((char **) NULL, 0, &number);
    if (!font_strs || !number)
        return(-1);

       /* Scan returned string to
          see if it has `fam'. */
    for (ix = 0, found = FALSE; ix < number; ix++) {
        if (streq(fam, font_strs[ix])) {
            found = TRUE;
        }
    }
       
    for (ix = 0; ix < number; ix++) {
        THIMfree((char *) font_strs[ix]);
    }
    THIMfree(font_strs);
    
    if (found) {
        return (0);
    }
    else {
        return (-1);
    }
}

TblSetHeadingFont(w, nameData)
AXWidget_table w;        /* in */
data_ptr nameData;       /* in: array of font, size, weight, height */
{
    table_info *tp;

	if (!w || w->type != W_TABLE)
        return(-1);
    if (!(tp = (table_info *) w->info.table))
        return(-1);
	tp->headingFont = (XFontStruct *)lookupNamedFont(nameData);
	if (tp->headingFont) {
		tp->heading_char_above = tp->headingFont->ascent;
		tp->heading_char_below = tp->headingFont->descent;
		tp->heading_char_height = tp->headingFont->ascent +
			                      tp->headingFont->descent;
		tp->headingHeight = TblComputeHeadingHeight(w);

	}
	tp->numLines = TblComputeNumLines(w);
    return(0);
}

/***********************************************************************
function: TblSetFont - Set fonts on the table widget.
Note: Will repaint if table exists.
Returns 0 on success; -1 otherwise.
*/
int
TblSetFont(w, nameData)
AXWidget_table w;        /* in */
data_ptr nameData;       /* in: array of font, size, weight, height */
{
    table_info *tp;
    int asize, ix, font_size, font_weight, font_slant;
    data_ptr atmp;
    mbchar *font_family;

    if (!w || w->type != W_TABLE)
        return(-1);
    if (!(tp = (table_info *) w->info.table))
        return(-1);

    font_family = (mbchar *) NULL;
    font_size = font_weight = font_slant = 0;
    asize = AxArraySize(nameData);
    for (ix = 0; ix < asize; ix++) {
        atmp = AxArrayElement(nameData, ix);
        /* interpret as <attr type><attr val>
         */
        switch(AxIntFromArray(atmp, 0)) {
        case FONT_FAMILY:
            font_family = AxStrPtrFromArray(atmp, 1);
            break;
        case FONT_SIZE:
            font_size = AxIntFromArray(atmp, 1);
            break;
        case FONT_WEIGHT:
            font_weight = AxIntFromArray(atmp, 1);
            if (font_weight == -1)
                font_weight = 1;
            break;
        case FONT_SLANT:
            font_slant = AxIntFromArray(atmp, 1);
            if (font_slant == -1)
                font_slant = 1;
            break;
        }
    }

    if ( !font_family )
        return -1;

#if 0
    /* This fails on NT for non-system fonts. Also, I don't think
     * this check needs to be made on any platform,
     * so i've commented it out.  -donato
     */
    if (TblVfyFontSpec(font_family, font_size, font_weight,
                       font_slant) == -1) {
        return(-1);
    }
#endif

    tp->font = tp->headingFont = (XFontStruct *) NULL;
    if (tp->fontFamily)
        free(tp->fontFamily);
    tp->fontFamily = (char *) malloc(strlen(font_family) + 1);
    strcpy(tp->fontFamily, font_family);
    tp->fontSize = font_size;
    tp->fontWeight = font_weight;
    tp->fontSlant  = font_slant;
    TblInitFonts(w);
    tp->headingHeight = TblComputeHeadingHeight(w);
    tp->rowHeight = tp->char_height + 4 + 2;
    tp->numLines = TblComputeNumLines(w);
    TblUpdCursorCellInfo(tp);
    
    if (w->window) {
        WidgetNeedsUpdated = TRUE;
        w->update_widget = TRUE;
    }
    return(0);
}

/***********************************************************************
function: TblCreateGCs - Create GCs used by the table widget.
*/
static void
TblCreateGCs(w)
AXWidget_table w;
{
    table_info *tp;

    if ((!w) || (w->type != W_TABLE))
        return; /* invalid args */
    if (!(tp = (table_info *) w->info.table))
        return; /* invalid args. */

/* Create grid gc */
    if (!tp->grid_gc) {
        tp->grid_gc = axCreateGC(Dpy, RootWin, DefaultScreen(Dpy),
                                 DisplayPlanes(Dpy, DefaultScreen(Dpy)),
                                 (XFontStruct *) 0, 0, NULL);
        if (tp->grid_gc == (GC) NULL)
            return;
        XSetDashes(Dpy, tp->grid_gc, 0, dit, 2);
        XSetLineAttributes(Dpy, tp->grid_gc, 0, LineOnOffDash, CapRound,
                           JoinMiter);

        if (!UseWidgetColors /* !ColorDisplay || NoColor */) {
            XSetForeground(Dpy, tp->grid_gc, ForePixel);
            XSetBackground(Dpy, tp->grid_gc, BackPixel);
        } else {
            if (w->useWorkColors) {
                XSetForeground(Dpy, tp->grid_gc, AlisPixels[1]);
                XSetBackground(Dpy, tp->grid_gc, AlisPixels[0]);
            } else {
                XSetForeground(Dpy, tp->grid_gc, WidgetPixels[1]);
                XSetBackground(Dpy, tp->grid_gc, WidgetPixels[0]);
            }
        }
    }

/* Create mark gc */
    if (!tp->mark_gc) {
        XGCValues gcv;

        memset((char *) &gcv, 0, sizeof(XGCValues));
        if (!UseWidgetColors /* !ColorDisplay || NoColor */) {
            gcv.foreground = ForePixel^BackPixel;
            gcv.background = ForePixel^BackPixel;
        } else {
            if (w->useWorkColors) {
                gcv.foreground = AlisPixels[0]^AlisPixels[1];
                gcv.background = AlisPixels[0]^AlisPixels[1];
            } else {
                gcv.foreground = WidgetPixels[0]^WidgetPixels[1];
                gcv.background = WidgetPixels[0]^WidgetPixels[1];
            }
        }
        gcv.fill_style = FillSolid;
        gcv.function = GXxor;
        gcv.plane_mask = AllPlanes;
        gcv.line_width = 0;
        gcv.join_style = JoinRound; /* makes no diff */
        gcv.cap_style  = CapRound;  /* makes no diff */
        tp->mark_gc = axCreateGC(Dpy, RootWin, DefaultScreen(Dpy),
                                 DisplayPlanes(Dpy, DefaultScreen(Dpy)),
                                 (XFontStruct *) 0,
                                 (GCForeground | GCBackground | GCPlaneMask |
                                  GCFunction | GCLineWidth | GCJoinStyle |
                                  GCCapStyle), &gcv);
    }
    
    return;
}

/***********************************************************************
function: TblFreeGCs - Free all table related gcs.
return: 0 on success; -1 otherwise.
*/
int
TblFreeGCs(w)
AXWidget_table w;
{
    table_info *tp;

    if ((!w) || (w->type != W_TABLE))
        return(-1);
    if (!(tp = (table_info *) w->info.table))
        return(-1);

    if (tp->grid_gc)
        XFreeGC(Dpy, tp->grid_gc);
    tp->grid_gc = (GC) NULL;

    if (tp->mark_gc)
        XFreeGC(Dpy, tp->mark_gc);
    tp->mark_gc = (GC) NULL;

    return(0);
}

/***********************************************************************
function: TblInitFonts - [re] initialize table_info->headingFont, ->font
*/
static void
TblInitFonts(w)
AXWidget_table w;
{
    int ret;
    table_info *tp;
    
    WIDGET_PARANOID( w, W_TABLE );

    tp = (table_info *)w->info.table;
    if (!tp)
        return;
    
/* Set up table font. */
    if (!tp->font) {
        int scalable;

        ret = FALSE;
        if (tp->fontFamily) { /* using custom fonts */
            /* sets to TRUE if successful */
            ret = getXFontStruct(tp->fontFamily, tp->fontSize,
                                 tp->fontWeight, tp->fontSlant,
                                 &scalable, &tp->font);
        }
        if ((ret != TRUE) || (!tp->font)) {
            /* Use default font */
            tp->char_above  = AxCharAbove;
            tp->char_height = AxCharHeight;
            tp->char_below  = AxCharBelow;
            tp->font = AxDlgMainFont;
            w->font = (XFontStruct *) NULL;
        } else {
            /* We have a valid TableFont here */
            tp->char_above  = tp->font->max_bounds.ascent;
            tp->char_below  = tp->font->max_bounds.descent;
            tp->char_height = tp->char_above + tp->char_below;
            w->font = tp->font;
        }
    }

/* Set up heading font */
    if (!tp->headingFont) {
        int scalable;

        ret = FALSE;
        if (tp->fontFamily) {
            /* fontWeight must always be bold for headings */
            ret = getXFontStruct(tp->fontFamily, tp->fontSize, 1,
                                 0, &scalable, &tp->headingFont);
        }
        if ((ret != TRUE) || (!tp->headingFont)) {
            /* Fallback to using AxDlgMainFont for font */
            tp->heading_char_above  = AxCharAbove;
            tp->heading_char_below  = AxCharBelow;
            tp->heading_char_height = AxCharHeight;
            tp->headingFont = AxDlgMainFont;
        } else {
            /* Have a valid TableFont here. */
            tp->heading_char_above  = tp->headingFont->max_bounds.ascent;
            tp->heading_char_below  = tp->headingFont->max_bounds.descent;
            tp->heading_char_height = tp->heading_char_above +
                tp->heading_char_below;

        }
    }
    return;
}

/***********************************************************************
function: TblComputeNumLines - Compute the number of visible lines in
the table window.
*/

int
TblComputeNumLines(w)
AXWidget_table w;
{
    int ypos;
    int numLines;
    table_info *tp;

    if (!w || w->type != W_TABLE)
        return(0);
    if (!(tp = (table_info *) w->info.table))
        return(0);
    if (!tp->rowHeight)
        return(0);

    ypos = tp->headingHeight;
    numLines = 0;

    /*CONSTCOND*/
	if ( tp->rowcolcreate ) {
		while (1) {
			ypos += tp->rowHeight;
			if (ypos > tp->winHeight)
				break;
			numLines++;
		}
	}
	else {
		while (1) {
			ypos += tp->rowHeight;
			numLines++;
			if (ypos > tp->winHeight)
				break;
		}
	}
    return(numLines);
}

/***********************************************************************
function: TblComputeHeadingHeight - Return the heading height.
*/
int
TblComputeHeadingHeight(w)
register AXWidget_table w;
{
    register table_info *tp;
    register int ix;
    register int num_cols;
    register int nlines;
    register int heading_height = 0;
    int max_heading_lines = 1;

    if (!w || w->type != W_TABLE)
        return(0);
    if (!(tp = (table_info *) w->info.table))
        return(0);

    /* default heading height */
    heading_height = tp->heading_char_height + 9;

    if (tp->colText)
        num_cols = AxArraySize(tp->colText);
    else
        num_cols = 0;
    if (tp->colText && num_cols) {
        for (ix = 0; ix < num_cols; ix++) {
            nlines = TblNumLinesInHeading(AxStrFromArray(tp->colText, ix));
            if (nlines > max_heading_lines)
                max_heading_lines = nlines;
        }
    }
    if (max_heading_lines > 1)
        heading_height += (tp->heading_char_height * (max_heading_lines - 1));
    return(heading_height);
}

/***********************************************************************
function: TblNumLinesInHeading - Return the number of lines needed by
heading string STR.
Return 1 + (number of '\n's found in string).
*/
static int
TblNumLinesInHeading(str)
mbchar *str;
{
    mbchar *p;
    int num_lines = 1;

    if (!str)
        return(1);
    for (p = str; *p; p++) {
        switch (*p) {
        case '\n':
            num_lines++;
            break;
        default:
            break;
        }
    }
    return(num_lines);
}

/***********************************************************************
function: ElfbTblFuncs - Generic table builtin functions entry point.
WVAL - widget handle
COMMAND_CODEP  - table builtin command code. (currently hardcoded)
DP2..DP6 - interpreted data depending on command code.
*/
void
ElfbTblFuncs(wval, command_codeP, dP1, dP2, dP3, dP4, dP5)
data_ptr wval;
data_ptr command_codeP;
data_ptr dP1, dP2, dP3, dP4, dP5;
{
    register AXWidget_table w;
    table_info *tp = (table_info *) NULL;
    int command_code;
    int new_top_row;
    data_ptr retP = (data_ptr) NULL;
    int total_row_count;
    int start_row;
    int num_rows = 0;
    int ix;
    int goto_row;
    int goto_col;
    int goto_start_char;
    int goto_end_char;
    char  *insert_buf = (char *) NULL;
    wchar *insert_wbuf = (wchar *) NULL;
    int  num_bytes,  insert_num_chars;
    int elfkeyid;

    w = (AXWidget_table)WidgetPointerFromHandleData(wval);
    if ((!w) || (w->type != W_TABLE))
        return; /* invalid args */

    command_code = IntFromDataPtr(command_codeP);
    switch (command_code) {
        
    case 1: /* db-table-set-new-top-row */
        TBL_LOG(("db-table-set-new-top-row\n"));
        new_top_row = IntFromDataPtr(dP1);
        if (!(tp = (table_info *) w->info.table))
            return;
        TblScrollNewTopRow(w, new_top_row);
        break;
        
    case 2: /* db-table-get-disp-lines */
        TBL_LOG(("db-table-get-disp-lines\n"));
        if (!(tp = (table_info *) w->info.table))
            retP = MK_INTEGER(0);
        else
            retP = MK_INTEGER(tp->numLines);
        break;
        
    case 3: /* db-table-set-total-row-count */
        TBL_LOG(("db-table-set-total-row-count\n"));
        if (!(tp = (table_info *) w->info.table))
            return;
        total_row_count   = IntFromDataPtr(dP1);
        tp->totalRowCount = total_row_count;
        TblUpdateScrollBars(w); 
        break;
        
    case 4: /* db-table-get-top-row */
        TBL_LOG(("db-table-get-top-row\n"));
        if (!(tp = (table_info *) w->info.table))
            retP = MK_INTEGER(0); /* default */
        else
            retP = MK_INTEGER(tp->topRow);
        break;
        
    case 5: /* db-table-next-page */
        TBL_LOG(("db-table-next-page\n"));
        if (!(tp = (table_info *) w->info.table))
            return;
        new_top_row = tp->topRow + tp->numLines - 1;
        if (tp->totalRowCount != -1) {
            if ((new_top_row + tp->numLines) >= tp->totalRowCount)
                new_top_row = MIN(tp->topRow,(tp->totalRowCount -
                                              tp->numLines));
        }
        if (new_top_row == tp->topRow)
            return; /* nothing happens */
        TblScrollNewTopRow(w, new_top_row);
        break;
        
    case 6: /* db-table-previous-page */
        TBL_LOG(("db-table-previous-page\n"));
        if (!(tp = (table_info *) w->info.table))
            return;
        new_top_row = MAX(0, tp->topRow - 1);
        TblScrollNewTopRow(w, new_top_row);
        break;
        
    case 7: /* db-table-marker-pixmaps */
        TBL_LOG(("db-table-marker-pixmaps\n"));
        if (!(tp = (table_info *) w->info.table))
            return;
        start_row = IntFromDataPtr(dP1);
        
        TblCheckMarkerPixmaps(dP2);
        
        TblInsertRowMarkerPixmaps(w, start_row, dP2);
        
        if (AxIsArray(dP2))
            num_rows = AxArraySize(dP2);
        else
            num_rows = 0;
        if (num_rows == 0) {
            if (TblRowIsSelected(tp, start_row))
                TblDrawRowMarkerShadow(w, start_row, TRUE);
            else
                TblDrawRowMarkerShadow(w, start_row, FALSE);
            TblDrawRowMarkerContents(w, start_row);
        } else {
            for (ix = start_row; ix < (start_row + num_rows); ix++) {
                if (TblRowIsSelected(tp, ix))
                    TblDrawRowMarkerShadow(w, ix, TRUE);
                else
                    TblDrawRowMarkerShadow(w, ix, FALSE);
                TblDrawRowMarkerContents(w, ix);
            }
        }
        break;
        
    case 8: /* db-table-goto-cell */
        goto_row = IntFromDataPtr(dP1);
        goto_col = IntFromDataPtr(dP2);
        goto_start_char = IntFromDataPtr(dP3); /* 1 based */
        goto_end_char   = IntFromDataPtr(dP4); /* 1 based */
        TBL_LOG(("db-table-goto-cell: %d %d %d %d\n", goto_row, goto_col,
                 goto_start_char, goto_end_char));
        if ((goto_row == -1) && (goto_col == -1)) {
            TblCommitCursorCell(w);
        } else {
            if (TblGotoCellChar(w, goto_row, goto_col,
                                goto_start_char - 1, goto_end_char - 1) != 0)
                RingTheBell();
        }
        break;
        
    case 9: /* db-table-insert-text */
        TBL_LOG(("db-table-insert-text\n"));
        insert_buf = StrFromDataPtr(dP1);
        num_bytes = strlen(insert_buf);
        if ( num_bytes ) {
            insert_wbuf = (wchar *) malloc((num_bytes + 1) * sizeof(wchar));
            insert_num_chars = MbToWStr(insert_buf, insert_wbuf);
            TblInsertText(w, insert_wbuf, insert_num_chars);
        }
        if (insert_wbuf)
            free(insert_wbuf);
        break;
        
    case 10: /* db-table-send-key */
        TBL_LOG(("db-table-send-key\n"));
        elfkeyid = IntFromDataPtr(dP1);
        TblHandleElfKeyid(w, elfkeyid);
        break;
        
    case 11:
        TBL_LOG(("db-table-get-rows\n"));
        retP = AxXmGetTableRows(w, dP1, dP2);
        break;
        
    case 13: /* db-table-corner-pixmaps */
        TBL_LOG(("db-table-corner-pixmaps\n"));
        tp = (table_info *) w->info.table;
        if (!tp)
            return;
        TblInsertCornerPixmaps(w, dP1);
        TblDrawCornerBox(w);
        break;
        
    case 14:  /* return whether specified row is known by table */
        tp = (table_info *) w->info.table;
        if (tp) {
            ix = IntFromDataPtr(dP1);
            if (AxArrayElement(tp->rows, ix)) {
                retP = MK_INTEGER(-1);           /* true */
            } else {
                retP = MK_INTEGER(0);            /* false */
            }
        }
        else {
            retP = MK_INTEGER(0);
        }
        break;
    case 15:
        TBL_LOG(("db-table-marker-strings\n"));
        if (!(tp = (table_info *) w->info.table))
            return;
        start_row = IntFromDataPtr(dP1);
        
        TblCheckMarkerStrings(dP2);        
        TblInsertRowMarkerStrings(w, start_row, dP2);
        
        if (AxIsArray(dP2))
            num_rows = AxArraySize(dP2);
        else
            num_rows = 0;
        if (num_rows == 0) {
            if (TblRowIsSelected(tp, start_row))
                TblDrawRowMarkerShadow(w, start_row, TRUE);
            else
                TblDrawRowMarkerShadow(w, start_row, FALSE);
            TblDrawRowMarkerContents(w, start_row);
        } else {
            for (ix = start_row; ix < (start_row + num_rows); ix++) {
                if (TblRowIsSelected(tp, ix))
                    TblDrawRowMarkerShadow(w, ix, TRUE);
                else
                    TblDrawRowMarkerShadow(w, ix, FALSE);
                TblDrawRowMarkerContents(w, ix);
            }
        }
#if 0       
        /* might need to add data to keep rowmarkers in sync with wt->rows*/
        if ( tp->rowcolcreate ) {
            int n_rows, n_marks;
            n_marks = AxArraySize(tp->marker_strings);          
            n_rows  = AxArraySize(tp->rows);
            printf("n_markers = %d n_rows = %d \n", n_marks, n_rows);
            if ( n_marks > n_rows )
                TblAddData(w, TRUE, 0, n_marks - n_rows, 0);
        }
#endif      
        break;
    case 16: /* db-table-get-top-row */
        TBL_LOG(("db-table-get-left_col\n"));
        if (!(tp = (table_info *) w->info.table))
            retP = MK_INTEGER(0); /* default */
        else
            retP = MK_INTEGER(tp->leftCol);
        break;
    case 17:        
        TBL_LOG(("db-table-get-headings\n"));
        if (!(tp = (table_info *) w->info.table))
            retP = NULL;
        else
            retP = TblGetHeadings(w);
        break;
        
        break;
    default: /* unimpl */
        TBL_LOG(("bad command code: %d\n", command_code));
        break;
    }
    ElfRetData = retP;
    return;
}
        
/*************************************************************************
function: ElfbTableData - Set column headins, row data.
Note: Will set data only if DATA is non-null; Will set headings only
if HEADINGS is non-null; Will set marker pixmaps only if
MARKER_PIXMAPS is non-null.
The data dipper uses this call to change column headings by passing
null data and null marker_pixmaps.
*/
void ElfbTableData(wval, data, headings, marker_pixmaps)
data_ptr wval;
data_ptr data, headings;
data_ptr marker_pixmaps;
{
    register AXWidget_table w;
    register table_info *wt;
    int colCount;
    int ix;
    data_ptr colData;

    TBL_LOG(("ElfbTableData\n"));
    w = (AXWidget_table)WidgetPointerFromHandleData(wval);
    if ((!w) || (w->type != W_TABLE))
        return; /* invalid args */
    if (!(wt = (table_info *) w->info.table))
        return; /* invalid args */

    TblResetAll(w);

/* Set data */
    if (data) {
        TblCheckRowData(data);
        (void) TblClearAllSelections(w);
        AxTaskFreeElfData(0, wt->rows);
        wt->rows = AxTaskCopyElfData(0, data);
        if (AxArraySize(wt->rows) < wt->numLines) {
            wt->topRow = 0;    /* less than a screenful of rows */
        }
    }
	else if ( !data && wt->rowcolcreate )
	{
		/*
		 * Delete all the data: when the user passes null data, for presents
		 * only. 
		 */		
		int size, colSize;
        data_ptr row_data;
		size = AxArraySize(wt->rows);
		for ( ix = 0; ix < size; ix++)
		{   
			row_data = AxArrayElement(wt->rows, ix);
			colSize  = AxArraySize(row_data);			
			row_data = AxTaskCreateElfArray(0, colSize);                
			row_data = TaskWriteArray(0, row_data, colSize, NULL);
			wt->rows = TaskWriteArray(0, wt->rows, ix, row_data);
		}
        if (wt->cursor.wtext)
            free(wt->cursor.wtext);
        TblResetCellTextInfo(wt);
	}
	
/* Set marker pixmaps */
    if (marker_pixmaps) {
        TblCheckMarkerPixmaps(marker_pixmaps);
        wt->marker_pixmaps = AxTaskCopyElfData(0, marker_pixmaps);
    }

/* Set headings */
    if (headings) {
        AxTaskFreeElfData(0, wt->colText);
        AxTaskFreeElfData(0, wt->colWidth);
        AxTaskFreeElfData(0, wt->colType);

        colCount = ElfbArraySize(headings);
        wt->colText  = AxTaskCreateElfArray(0, 0);
        wt->colWidth = AxTaskCreateElfArray(0, 0);
        wt->colType  = AxTaskCreateElfArray(0, 0);
        wt->cursor.row = -1;
        wt->cursor.col = -1;
        for ( ix = 0; ix < colCount; ix++ ) {
            colData = ReadArray(AX(headings), ix);
            wt->colText = TblAddStrToArray(0, wt->colText, ix,
                                           AxStrFromArray(colData,0));
            wt->colWidth = TblAddNumToArray(0, wt->colWidth, ix,
                                            IntFromArray(colData,1));
            wt->colType = TblAddNumToArray(0, wt->colType, ix,
                                           IntFromArray(colData,2));
        }
        /* Recompute heading height */
        wt->headingHeight = TblComputeHeadingHeight(w);
        wt->numLines = TblComputeNumLines(w);
    }
    
    if( wt->rowcolcreate )
    {
		/* For Presents ONLY */
        int size, colSize, col_size, ix, jx;
        data_ptr row_data, new_row_data;
        
        size = AxArraySize(wt->rows);
        col_size = AxArraySize(ReadArray(AX(wt->rows), 0));
        colSize = MAX(col_size, AxArraySize(wt->colText)); 
        if ( col_size < colSize ) for ( ix = 0; ix < size; ix++)
        {
            new_row_data = AxTaskCopyElfData(0, ReadArray(AX(wt->rows), ix));

            for ( jx = col_size; jx< colSize; jx++)
                new_row_data = TaskWriteArray(0, new_row_data, jx, NULL);
            wt->rows = TaskWriteArray(0, wt->rows, ix, new_row_data);
        }
        if ( size < wt->totalRowCount ) /* PACK ADDITIONAL NULL DATA */
        {
            /* now, pad remaining rows with ALL null data */
            for ( ix = size; ix < wt->totalRowCount; ix++)
            {   
                row_data = AxTaskCreateElfArray(0, colSize);                
                row_data = TaskWriteArray(0, row_data, colSize, NULL);
                wt->rows = TaskWriteArray(0, wt->rows, ix, row_data);
            }
        }
    }
    WidgetNeedsUpdated = TRUE;
    w->update_widget = TRUE;
    TblUpdateScrollBars(w);
}

/***********************************************************************
function: ElfbTableNewData - Set new data in table starting from row
index START_ROW.
Elf level access:
db_table_set_new_data@(format dialog_box_ menu, wdef, data, start_row
*/
void
ElfbTableNewData(wval, data, start_row)
data_ptr wval;
data_ptr data;
data_ptr start_row; /* start row of data (0 based) */
{
    register AXWidget_table w;
    register table_info *tp;
    int row_idx;
    int num_new_rows;
    int old_max_rows;
    int new_max_rows;
    XRectangle crect[1];
    int row_ypos;
    int end_row_ypos;

    TBL_LOG(("ElfbTableNewData\n"));
    w = (AXWidget_table)WidgetPointerFromHandleData(wval);
    if ((!w) || (w->type != W_TABLE))
        return; /* invalid args */
    if (!(tp = (table_info *) w->info.table))
        return; /* invalid args */

    TblResetAll(w);
    TblClearAllSelections(w);
    old_max_rows = AxArraySize(tp->rows);
    row_idx = IntFromDataPtr(start_row);
    TblCheckRowData(data);
    num_new_rows = TblInsertNewData(w, data, row_idx);
	if (tp->rowcolcreate)
		tp->totalRowCount = num_new_rows;
    if (num_new_rows <= 0) {
        TblUpdateScrollBars(w);
        WidgetNeedsUpdated = TRUE;
        w->update_widget = TRUE;
        return;
    }

    new_max_rows = AxArraySize(tp->rows);
    if (new_max_rows != old_max_rows)
        TblUpdateScrollBars(w);

    if (new_max_rows < old_max_rows) {
        if (new_max_rows < tp->numLines) {
            tp->topRow = 0;       /* less than a screenful of rows */
        }
        TblUpdateScrollBars(w);
        WidgetNeedsUpdated = TRUE;
        w->update_widget = TRUE;
        return;
    }
    row_ypos = YPosOfCell(tp, row_idx, tp->leftCol);
    end_row_ypos = YPosOfCell(tp, (row_idx + num_new_rows), tp->leftCol);
    if ((row_ypos == -1) || (end_row_ypos == -1)) {
        TblUpdateScrollBars(w);
        WidgetNeedsUpdated = TRUE;
        w->update_widget = TRUE;
        return;
    }

    if (tp->cursor.col != -1) { /* FIXME, use `tp->doing_cell_editing' */
        TblLiftSelectionCursor(w);
    }
    
    crect[0].x = 0;
    crect[0].y = 0;
    crect[0].width  = tp->winWidth;
    crect[0].height = MIN((tp->winHeight - row_ypos),
                          (end_row_ypos - row_ypos));
    TblSetClipAllGCs(w, 0, row_ypos, crect, 1, YXBanded);

    PaintTableRegion(w, row_idx, row_idx + num_new_rows - 1,
                     tp->leftCol, -1);
    
    TblUnsetClipAllGCs(w);
    
    if (tp->cursor.col != -1) { /* FIXME, use `tp->doing_cell_editing' */
        TblPutSelectionCursor(w);
    }
    
    return;
}

/***********************************************************************
function: TblInsertNewData - Set new data in table starting from row
index START_ROW.
Return: number of new rows inserted.
*/
static int
TblInsertNewData(w, new_data, start_row)
register AXWidget_table w;     /* in */
register data_ptr new_data;    /* in */
int start_row;                 /* in: starting row of data (0 based) */
{
    register table_info *tp;
    register int ix;
    register int nx;
    int num_new_rows;
    data_ptr new_row_data;

    if (!w)
        return(0);
    if (!(tp = (table_info *) w->info.table))
        return(0);
    if (!new_data)
        return(0);
    if (start_row < 0)
        return(0);
    
    if ( (start_row != 0) && !tp->rows )
        tp->rows = AxTaskCreateElfArray(0, 0);

    num_new_rows = AxArraySize(new_data); 

    if (tp->rows) {
        for (ix = start_row, nx = 0; nx < num_new_rows; ix++, nx++) {
            new_row_data = AxTaskCopyElfData(0, ReadArray(AX(new_data), nx));
            tp->rows = TaskWriteArray(0, tp->rows, ix, new_row_data);
            if (tp->use_data_request_model && (tp->cursor.row == ix)) {
                /* transfer new cell text onto tp->cursor data */
                mbchar *txt;
                int len;
                if (txt = AxStrFromArray(new_row_data, tp->cursor.col)) {
                    len = MBstrlen(txt);
                    TblUpdCursorCellText(tp, txt, len);
                    tp->cursor.left_char_pos = 0;
                }
            }
        }
    }
    else {
        tp->rows = AxTaskCopyElfData(0, new_data);
    }

    TBL_LOG(("Inserted new data: cursor is %d, %d\n",
             tp->cursor.row, tp->cursor.col));
    return(num_new_rows);
}

/***********************************************************************
function: TblInsertRowMarkerPixmaps - Insert new row marker pixmaps
from row START_ROW.
NOTE:  Caller frees datum MARKER_PIXMAPS.
*/
static void
TblInsertRowMarkerPixmaps(w, start_row, marker_pixmaps)
AXWidget_table w;
int start_row;
data_ptr marker_pixmaps;
{
    table_info *tp;
    int ix, nx;
    int num_pixmap_rows;
    data_ptr pixmapList;

    if (!w)
        return;
    if (!(tp = (table_info *) w->info.table))
        return;
    if (start_row < 0)
        return;

    if (marker_pixmaps) {
        if (!AxIsArray(marker_pixmaps))
            return; /* silently return */
    }
            
    num_pixmap_rows = AxArraySize(marker_pixmaps);

    if (!tp->marker_pixmaps) {
        /* create one if necessary */
        if (num_pixmap_rows == 0)
            return; /* no need */
        tp->marker_pixmaps = AxTaskCreateElfArray(0,
                                                  start_row + num_pixmap_rows);
    }

    if (num_pixmap_rows == 0) {
        /* Caller wants to clear all pixmaps for this row */
        tp->marker_pixmaps = TaskWriteArray(0, tp->marker_pixmaps,
                                          start_row, NULL);
        return;
    }

    for (ix = start_row, nx = 0; nx < num_pixmap_rows; ix++, nx++) {
        pixmapList = AxTaskCopyElfData(0, ReadArray(AX(marker_pixmaps), nx));
        tp->marker_pixmaps = TaskWriteArray(0, tp->marker_pixmaps, ix,
                                          pixmapList);
    }
    return;
}


/***********************************************************************
function: TblInsertRowMarkerStrings - Insert new row marker strings
from row START_ROW.
NOTE:  Caller frees datum MARKER_STRINGS.
*/
static void
TblInsertRowMarkerStrings(w, start_row, marker_strings)
AXWidget_table w;
int start_row;
data_ptr marker_strings;
{
    table_info *tp;
    int ix, nx;
    int num_string_rows;
    data_ptr stringList;

    if (!w)
        return;
    if (!(tp = (table_info *) w->info.table))
        return;
    if (start_row < 0)
        return;

    if (marker_strings) {
        if (!AxIsArray(marker_strings))
            return; /* silently return */
    }
            
    num_string_rows = AxArraySize(marker_strings);

    if (!tp->marker_strings) {
        /* create one if necessary */
        if (num_string_rows == 0)
            return; /* no need */
        tp->marker_strings = AxTaskCreateElfArray(0,
                                                  start_row + num_string_rows);
    }

    if (num_string_rows == 0) {
        /* Caller wants to clear all strings for this row */
        tp->marker_strings = TaskWriteArray(0, tp->marker_strings,
                                          start_row, NULL);
        return;
    }

    for (ix = start_row, nx = 0; nx < num_string_rows; ix++, nx++) {
        stringList = AxTaskCopyElfData(0, ReadArray(AX(marker_strings), nx));
        tp->marker_strings = TaskWriteArray(0, tp->marker_strings, ix,
                                          stringList);
    }
    return;
}

/***********************************************************************
function: TblInsertCornerPixmaps - Insert corner pixmaps.  If
corner_pixmaps is NULL, clear existing pixmaps.
*/
static void
TblInsertCornerPixmaps(w, corner_pixmaps)
AXWidget_table w;
data_ptr corner_pixmaps;
{
    register table_info *tp;
    int num_pixmaps;

    if (!w)
        return;
    tp = (table_info *) w->info.table;
    if (!tp)
        return;

    if (!corner_pixmaps) {
        if (tp->corner_pixmaps)
            AxTaskFreeElfData(0, tp->corner_pixmaps);
        tp->corner_pixmaps = NULL;
        return;
    }
    if (!AxIsArray(corner_pixmaps))
        return; /* need an array of pixmaps -- silently return. */

    num_pixmaps = AxArraySize(corner_pixmaps);

    if (!num_pixmaps) { /* 0 length array */
        if (tp->corner_pixmaps)
            AxTaskFreeElfData(0, tp->corner_pixmaps);
        tp->corner_pixmaps = NULL;
        return;
    }       

    tp->corner_pixmaps = AxTaskCopyElfData(0, corner_pixmaps);
    return;
}

/***********************************************************************
function: ElfbClearTableData - Clear NUM_ROWS rows of data in table
starting at row index START_ROW.
Note: If NUM_ROWS is -1, clear out *all* rows in the table.
Elf level access:
db_table_clear_data@(format dialog_box_ menu, wdef, start_row, num_rows, send_expose)
*/
void
ElfbClearTableData(wval, start_row_datum, num_rows_datum, send_expose)
data_ptr wval;
data_ptr start_row_datum;
data_ptr num_rows_datum;
data_ptr send_expose;
{
    register AXWidget_table w;
    register table_info *tp;
    int start_row;
    int num_rows;

    TBL_LOG(("ElfbClearTableData\n"));
    w = (AXWidget_table)WidgetPointerFromHandleData(wval);
    if ((!w) || (w->type != W_TABLE))
        return; /* invalid args */
    if (!(tp = (table_info *) w->info.table))
        return; /* invalid args */

    start_row = IntFromDataPtr(start_row_datum);
    num_rows  = IntFromDataPtr(num_rows_datum);
    TblClearData(w, start_row, num_rows);
    if (num_rows != -1) {
        if ((start_row >= tp->topRow) &&
            ((start_row + num_rows) <= (tp->numLines + tp->topRow))) {
            TblUpdateScrollBars(w);
            WidgetNeedsUpdated = TRUE;
            w->update_widget = TRUE;
        }
    } else {
        /* All rows cleared -- keep display accurate. */
        TblUpdateScrollBars(w);
        WidgetNeedsUpdated = TRUE;
        w->update_widget = TRUE;
    }
    if (tp->use_data_request_model || IntFromDataPtr(send_expose)) {
        if (num_rows == -1) {
            start_row = tp->topRow;
            num_rows  = tp->numLines;
        }
        TblSendDataRequestEvent(w, start_row, num_rows);
    }
    return;
}

/***********************************************************************
function: TblClearData - Clear NUM_ROWS rows of data in the table
starting at row index START_ROW.
Note: If NUM_ROWS is -1, clear out *all* rows in the table.
*/
static void
TblClearData(w, start_row, num_rows)
register AXWidget_table w;
int start_row;
int num_rows;
{
    register table_info *tp;
    int max_num_rows;
    int ix, cx;

    if ((!w) || (w->type != W_TABLE))
        return; /* invalid args */
    if (!(tp = (table_info *) w->info.table))
        return; /* invalid args */
    if (start_row < 0)
        return; /* invalid args */

    if (!tp->rows)
        return;

    TblClearAllSelections(w);

    if (tp->use_data_request_model && tp->cursor.row != -1)
        goto clear_one_by_one;

    if (num_rows == -1) {
        /* clear whole table */
        AxTaskFreeElfData(0, tp->rows);
        tp->rows = (data_ptr) NULL;
        tp->topRow = 0; /* reset top row */
        /* keep row markers in sync */
        if (tp->marker_pixmaps)
            AxTaskFreeElfData(0, tp->marker_pixmaps);
        tp->marker_pixmaps = NULL;
        if (tp->marker_strings)
            AxTaskFreeElfData(0, tp->marker_strings);
        tp->marker_strings = NULL;
        return;
    }

clear_one_by_one:
    max_num_rows = AxArraySize(tp->rows);
    if (num_rows != -1) {
        if ((start_row + num_rows) > max_num_rows)
            num_rows = max_num_rows - start_row; /* clip */
    } else
        num_rows = max_num_rows;
    for (ix = start_row, cx = 0; cx < num_rows; ix++, cx++) {
        if (tp->use_data_request_model && ix == tp->cursor.row) {
            continue; /* leave it alone -- the dipper depends on it
                         being there */
        }
        tp->rows = TaskWriteArray(0, tp->rows, ix, NULL);
        /* keep row markers in sync */
        if (tp->marker_pixmaps)
            tp->marker_pixmaps = TaskWriteArray(0, tp->marker_pixmaps,
                                                ix, NULL);
        /* keep row markers in sync */
        if (tp->marker_strings)
            tp->marker_strings = TaskWriteArray(0, tp->marker_strings,
                                                ix, NULL);
    }
    return;
}

/***********************************************************************
function: AxXmSetTableSelections - Set selections in table.
SELECTIONS is an elf array of ints (0 based).  Clear all selections on
a null value for SELECTIONS.
returns: Always returns 0.
Elf level access:
db_table_set_selections@(format dialog_box_ menu, wdef, selections)
Elf Builtin Mapping:
`XDLG_TABLE_SET_SELECTIONS_' -> `ElfbSetTableSelections' ->
`AxXmSetTableSelections'.
Note: Does not send a selection change event here because caller is
explicitly setting selections.
*/
int
AxXmSetTableSelections(w, selections)
AXWidget_table w;
data_ptr selections;
{
    register table_info *tp;
    int num_selections;
    int ix;
    int row_num;
    
    if (!w || (w->type != W_TABLE))
        return(0);

    if (!(tp = (table_info *) w->info.table))
        return(0);

    if (!(num_selections = ElfbArraySize(selections))) {
        if (tp->cursor.col != -1) { /* FIXME, use `tp->doing_cell_editing' */
            TblCommitCursorCell(w);
        }
        TblClearAllSelections(w);
        return(0);
    }

    /* FIXME, RACE CONDITION CAN HAPPEN HERE!!
       Eg: If user calls `db_set_selections@()', and if there is a
       cell being edited, `TblCommitCursorCell()' will make a callback.
       I don't know what happens then.

       Safe for the time being, `TblCommitCursorCell()' does not
       callback. */
    if (tp->cursor.col != -1) { /* FIXME, use `tp->doing_cell_editing' */
        /* A cell was being edited */
        TblCommitCursorCell(w); /* commit changes; and lift cursor */
    }
    TblClearAllSelections(w);
    for (ix = 0; ix < num_selections; ix++) {
        row_num = AxIntFromArray(selections, ix);
        if (TblAddSelection(w, row_num, ROW_MODE) == 0)
            TblDrawRow(w, row_num, tp->leftCol);
    }
    return(0);
}

/***********************************************************************
function: AxXmGetTableSelections - Return an elf array of ints
containing selected row numbers (0 based).  Caller is responsible for
freeing the returned elf datum.
Elf level access:
`db_table_get_selections@(format dialog_box_ menu, wdef)'
Elf Builtin Mapping:
`XDLG_TABLE_GET_SELECTIONS_' -> `ElfbSetTableSelections' ->
`AxXmSetTableSelections'.
*/
data_ptr
AxXmGetTableSelections(w)
AXWidget_table w;
{
    table_info *tp;
    data_ptr dp;
    int ix;
    int num_selections;
    int tempint;

    if (!w || (w->type != W_TABLE))
        return((data_ptr) NULL);
    if (!(tp = (table_info *) w->info.table))
        return((data_ptr) NULL);

    dp = AxMakeArray(0);
    num_selections = IntsArraySize(tp->selected_rows);
    for (ix = 0; ix < num_selections; ix++) {
        tempint = ReadIntsArray(tp->selected_rows, ix);
        dp = ElfAddNumToArray(dp, ix, tempint);
    }
    return(dp);
}

int
AxXmTableIsEditable(w)
AXWidget_table w;
{
	table_info *tp = (table_info*)w->info.table;
	if ( tp && tp->allow_cell_editing )
		return TRUE;
	else
		return FALSE;
}

/***********************************************************************
function: AxXmAllowTableEditing - Allow cell editing if flag is TRUE;
turn off cell editing otherwise.
Elf level access:
`db_table_allow_editing@(format dialog_box_ menu, wdef, flag)'
Elf Builtin Mapping:
`XDLG_TABLE_ALLOW_EDITING_' -> `ElfbAllowTableEditing' ->
`AxXmAllowTableEditing'.
*/
void
AxXmAllowTableEditing(w, flag)
AXWidget_table w;
int flag;
{
    register table_info *tp;

    if ((!w) || (w->type != W_TABLE))
        return; /* invalid args */
    if (!(tp = (table_info *) w->info.table))
        return;
    if (tp->allow_cell_editing) /* cell editing was previously on. */
        if (tp->cursor.col != -1) /* FIXME, use `tp->doing_cell_editing' */
            TblCommitCursorCell(w);
    tp->allow_cell_editing = (flag) ? TRUE : FALSE;
    return;
}

/***********************************************************************
function: AxXmDataRequestModel - 
*/
void
AxXmDataRequestModel(w, flag)
AXWidget_table w;
int flag;
{
    table_info *tp;

    if ((!w) || (w->type != W_TABLE))
        return;
    if (!(tp = (table_info *) w->info.table))
        return;
    tp->use_data_request_model = (flag) ? TRUE : FALSE;
    return;
}

/***********************************************************************
function: AxXmGetTableData - Return a copy of table data.  Caller is
responsible for freeing the returned datum.
In presentations mode, we need the cursor to be displayed in that last edited
cell.
Elf level access:
`db_table_get_data@(format dialog_box_ menu, wdef)'
Elf Builtin Mapping:
`XDLG_TABLE_GET_DATA_' -> `ElfbGetTableData' -> `AxXmGetTableData'.
*/
data_ptr
AxXmGetTableData(w)
AXWidget_table w;
{
    table_info *tp;
    data_ptr dp;
	int row_ix, col_ix;

    if (!w || (w->type != W_TABLE))
        return((data_ptr) NULL);
    if (!(tp = (table_info *) w->info.table))
        return((data_ptr) NULL);

    if (tp->cursor.col != -1) { /* FIXME, use `tp->doing_cell_editing' */
        /* A cell was being edited */
		row_ix = tp->cursor.row;
		col_ix = tp->cursor.col;		
        TblCommitCursorCell(w); /* commit changes; and lift cursor */
		if ( tp->rowcolcreate && row_ix >= 0 && col_ix >= 0)
            TblGotoCell(w, row_ix, col_ix, -1, -1);
    }
    dp = AxTaskCopyElfData(THIMpid, tp->rows);
    return(dp);
}


/***********************************************************************
function:  TblGetHeadings - Return a copy of table headings.
Elf level access:
`db_table_get_headings@(format dialog_box_ menu, wdef)'
`XDLG_TABLE_FUNCS
*/
static data_ptr
TblGetHeadings(w)
AXWidget_table w;
{
    table_info *tp;
    data_ptr dp;

    if (!w || (w->type != W_TABLE))
        return(0);
    if (!(tp = (table_info *) w->info.table))
        return(0);

    dp = AxTaskCopyElfData(THIMpid, tp->colWidth);
    return(dp);
}

/***********************************************************************
function: AxXmGetTableRows - Return a copy of table data.  Caller is
responsible for freeing the returned datum.
This is a lot like AxXmGetTableData.
Elf level access:
`db_table_get_rows@(format dialog_box_ menu, wdef, start_row, count)'
*/
data_ptr
AxXmGetTableRows(w, start_rowP, countP)
AXWidget_table w;
data_ptr start_rowP, countP;
{
    int start_row, count, ix;
    table_info *tp;
    data_ptr dp;

    count = IntFromDataPtr(countP);
    start_row = IntFromDataPtr(start_rowP);
    if (!w || (w->type != W_TABLE))
        return((data_ptr) NULL);
    if (!(tp = (table_info *) w->info.table))
        return((data_ptr) NULL);

    if (tp->cursor.col != -1) { /* FIXME, use `tp->doing_cell_editing' */
        /* A cell was being edited */
        if (tp->use_data_request_model) {
            int row, col, char_pos, end_char_pos;
            /* The dipper want's to get table rows, but does not
               expect the widget to remove cursor... */
            row = tp->cursor.row;
            col = tp->cursor.col;
            char_pos = tp->cursor.char_pos;
            end_char_pos = tp->cursor.end_char_pos;
            TblCommitCursorCell(w); /* commit changes; and lift cursor */
            TblKeepCursorCell(w, row, col, char_pos, end_char_pos);
        } else
            TblCommitCursorCell(w); /* commit changes; and lift cursor */
    }

	/* Avoid error in ReadArray */
	if (!tp->rows)
		return((data_ptr) NULL);

    dp = AxTaskCreateElfArray(THIMpid, count);
    for (ix = 0; ix < count; ix++) {
        dp = TaskWriteArray(THIMpid, dp, ix,
                            AxTaskCopyElfData(THIMpid,
                                  ReadArray(AX(tp->rows), start_row + ix)));
    }
    return(dp);
}

/***********************************************************************
function: TableEvent - Handle all types of XEvents in the table
widget.
Return 1 if event was consumed; 0 otherwise.
*/
TableEvent(w, evp)
register AXWidget_table w;
register XEvent *evp;
{
    int event_consumed = FALSE;
    table_info *tp;

    if (!w || !evp)
        return(0); /* invalid args */
    if (w->type != W_TABLE)
        return(0); /* invalid args */
    if (!(tp = (table_info *) w->info.table))
        return(0); /* invalid args */

    switch (evp->xany.type) {
    case FocusIn:
#if AXMS
		/* update the caret since it gets destroyed on loss of focus. */
        tp->cursor.text_cursor_on = FALSE;
		TblPutSelectionCursor(w);
#endif
        break;
    case FocusOut:
        break;
    case Expose:
        /* TODO, set clip-rectangle... */
        TblExposeCB(w, evp);
        event_consumed = TRUE;
        break;
    case GraphicsExpose:
        TblGraphicsExposeCB(w, evp);
        event_consumed = TRUE;
        break;
    case ButtonPress:
        TblButtonPressCB(w, evp);
        event_consumed = TRUE;
        break;
    case ButtonRelease:
        TblButtonReleaseCB(w, evp);
        TableArmed = NULL;
        w->value = -1;
        event_consumed = TRUE;
        break;
    case MotionNotify:
        TblMotionNotifyCB(w, evp);
        event_consumed = TRUE;
        break;
    case KeyPress:
        if (AxButtonDown) {
            RingTheBell();
            return(1); /* event consumed!! */
        }
        if (!tp->allow_cell_editing)
            return(0); /* silently return. */
        if (tp->doing_column_resizing || tp->seeing_button1_motion) {
            RingTheBell();
            return(0);
        }
        event_consumed = TblKeyPressCB(w, evp, FALSE);
        /* TODO, if (event_consumed), and in ``cbreak'' mode, send
           TOUCHED_RETURN message here. */
        break;
    case KeyRelease:
        if (!tp->allow_cell_editing)
            return(0); /* silently return. */
        TblKeyReleaseCB(w, evp);
        event_consumed = FALSE; /* !! */
        break;
    case SelectionRequest:
        event_consumed = TblSelectionRequestCB(w, evp);
        break;
    case SelectionNotify:
        event_consumed = TblSelectionNotifyCB(w, evp);
        break;
    case SelectionClear:
        event_consumed = TblSelectionClearCB(w, evp);
        break;
    default:
        event_consumed = FALSE;
        break;
    }

    return(event_consumed);

ignore_event:
    return(1);
}

/***********************************************************************
function: TblExposeCB - Handle an exposure event in the table widget.
*/
static void
TblExposeCB(w, evp)
register AXWidget_table w;
register XEvent *evp;
{
    return;
}

/***********************************************************************
function: TblGraphicsExposeCB - 
*/
static void
TblGraphicsExposeCB(w, evp)
AXWidget_table w;
XEvent *evp;
{
    XExposeEvent exp;

    if (!w || !evp)
        return;
    if (w->type != W_TABLE)
        return;

    /* fake an expose event */
    memset((char *) &exp, 0, sizeof(XExposeEvent));
    exp.type = Expose;
    exp.x = evp->xgraphicsexpose.x;
    exp.y = evp->xgraphicsexpose.y;
    exp.width  = evp->xgraphicsexpose.width;
    exp.height = evp->xgraphicsexpose.height;
    PaintTable(w, &exp, FALSE, TRUE);
    return;
}

/***********************************************************************
function: TblRowIsSelected - Return 1 if ROW_NUM exists int the
`selected_rows' list; 0 otherwise.
*/
static int
TblRowIsSelected(tp, row_num)
register table_info *tp;
register int row_num;
{
    if (InIntsArray(tp->selected_rows, row_num))
        return(1);
    return(0);
}

/***********************************************************************
function: TblColIsSelected - Return TRUE if COL_NUM exists in the
`selected_columns' list; FALSE otherwise
*/
static int
TblColIsSelected(tp, col_num)
register table_info *tp;
register int col_num;
{
    if (InIntsArray(tp->selected_columns, col_num))
        return(TRUE);
    return(FALSE);
}

/***********************************************************************
function: TblAddSelection - Add IDX to `selected_{rows|columns}'
list.
NOTE: Caller must update row display on successful return of this function.
Return 0 on success; -1 on failure or if IDX already exists in the list.
*/
static int
TblAddSelection(w, idx, mode)
register AXWidget_table w;
register int idx;
int mode;
{
    register int ix;
    register int num_selections;
    register int tempint;
    register table_info *tp;
    long *new_selected = NULL;
    long *selected = NULL;

    if (!w)
        return(-1);
    if (!(tp = (table_info *) w->info.table))
        return(-1);

    if (mode == ROW_MODE) {
        if (!tp->selected_rows)
            tp->selected_rows = CreateIntsArray(0); /* sanity */
        selected = tp->selected_rows;
    } else if (mode == COL_MODE) {
        if (!tp->selected_columns)
            tp->selected_columns = CreateIntsArray(0); /* sanity */
        selected = tp->selected_columns;
    } else
        return(-1);

    if (idx < 0)
        return(-1);

    num_selections = IntsArraySize(selected);
    new_selected = CreateIntsArray(num_selections + 1);

    for (ix = 0; ix < num_selections; ix++) {
        tempint = ReadIntsArray(selected, ix);
        if (tempint == idx)
            goto bailout; /* already there */
        WriteIntsArray(new_selected, ix, tempint);
    }
    WriteIntsArray(new_selected, num_selections, idx);

    if (selected)
        free(selected);
    if (mode == ROW_MODE)
        tp->selected_rows = new_selected;
    else
        tp->selected_columns = new_selected;
    return(0);

bailout:
    if (new_selected)
        free(new_selected);
    return(-1);
}

/***********************************************************************
function: TblToggleSelection - Toggle the selection on IDX.
Note: Updates the display.
*/
static void
TblToggleSelection(w, idx, mode)
register AXWidget_table w;
register int idx;
int mode;
{
    register table_info *tp;
    long *selected = NULL;

    if (!w)
        return;
    if (!(tp = (table_info *) w->info.table))
        return;

    switch (mode) {
    case ROW_MODE:
        selected = tp->selected_rows;
        break;
    case COL_MODE:
        selected = tp->selected_columns;
        break;
    default:
        return;
        break;
    }

    if (InIntsArray(selected, idx)) {
        if (TblDeleteSelection(w, idx, mode) == 0) {
            if (mode == ROW_MODE)
                TblDrawRow(w, idx, tp->leftCol);
            else
                TblDrawColumn(w, idx);
        }
    } else {
        if (TblAddSelection(w, idx, mode) == 0) {
            if (mode == ROW_MODE)
                TblDrawRow(w, idx, tp->leftCol);
            else
                TblDrawColumn(w, idx);
        }
    }
    return;
}

/***********************************************************************
function: TblClearExcludingRange - Clear selections from all rows/cols that
do not fall in the range {START_IDX - END_IDX} inclusive.
Updates display.
*/
static void
TblClearExcludingRange(w, start_idx, end_idx, mode)
register AXWidget_table w;
int start_idx;
int end_idx;
int mode;
{
    table_info *tp;
    int upper_bound;
    int lower_bound;
    int ix;
    long tempint;
    int num_selections;
    long *tmp_selections = NULL;
    long *selected = NULL;
    
    if (!w)
        return;
    if (!(tp = (table_info *) w->info.table))
        return;

    switch (mode) {
    case ROW_MODE:
        if (!tp->selected_rows) {
            tp->selected_rows = CreateIntsArray(0); /* sanity */
            return;
        }
        selected = tp->selected_rows;
        break;
    case COL_MODE:
        if (!tp->selected_columns) {
            tp->selected_columns = CreateIntsArray(0); /* sanity */
            return;
        }
        selected = tp->selected_columns;
        break;
    default:
        break;
    }

    num_selections = IntsArraySize(selected);
    if (num_selections <= 0)
        return;
    if (start_idx == end_idx) {
        upper_bound = lower_bound = start_idx;
    } else if (start_idx > end_idx) {
        upper_bound = start_idx;
        lower_bound = end_idx;
    } else {
        upper_bound = end_idx;
        lower_bound = start_idx;
    }

    tmp_selections = CopyIntsArray(selected);
    for (ix = 0; ix < num_selections; ix++) {
        tempint = ReadIntsArray(tmp_selections, ix);
        if ((tempint < lower_bound) || (tempint > upper_bound)) {
            if (TblDeleteSelection(w, tempint, mode) == 0) {
                if (mode == ROW_MODE)
                    TblDrawRow(w, tempint, tp->leftCol);
                else
                    TblDrawColumn(w, tempint);
            }
        }
    }
    if (tmp_selections)
        free(tmp_selections);
    return;
}

/***********************************************************************
function: TblAddSelectionRange - Add range START_IDX to END_IDX to
the selections list if not already there.  Add inclusive.  Updates display.
Return: 0
*/
static int
TblAddSelectionRange(w, start_idx, end_idx, mode)
register AXWidget_table w;
register int start_idx; /* index */
register int end_idx; /* index */
int mode;
{
    register table_info *tp;
    register int upper_bound;
    register int lower_bound;
    register int ix;
    int max_idx;

    if (!w)
        return(0);
    if (!(tp = (table_info *) w->info.table))
        return(0);

    switch (mode) {
    case ROW_MODE:
        max_idx = AxArraySize(tp->rows);
        break;
    case COL_MODE:
        max_idx = AxArraySize(tp->colText);
        break;
    default:
        return(0);
        break;
    }

    if ((start_idx < 0) || (start_idx >= max_idx))
        return(0); /* invalid idx */
    if ((end_idx < 0) || (end_idx >= max_idx))
        return(0); /* invalid idx */

    if (start_idx == end_idx) {
        if (TblAddSelection(w, start_idx, mode) == 0) {
            if (mode == ROW_MODE)
                TblDrawRow(w, start_idx, tp->leftCol);
            else
                TblDrawColumn(w, start_idx);
        }
        return(0);
    }

    if (start_idx > end_idx) {
        upper_bound = start_idx;
        lower_bound = end_idx;
    } else {
        upper_bound = end_idx;
        lower_bound = start_idx;
    }

    for (ix = lower_bound; ix <= upper_bound; ix++) {
        if (TblAddSelection(w, ix, mode) == 0) {
            if (mode == ROW_MODE)
                TblDrawRow(w, ix, tp->leftCol);
            else
                TblDrawColumn(w, ix);
        }
    }
    return(0);
}

/***********************************************************************
function: TblDeleteSelection - Remove IDX from selected list.  Operate on
`selected_rows' if MODE is ROW_MODE; operate on `selected_columns' otherwise.
NOTE: Caller must update row display if return is successful.
Return: 0 if ROW_IDX was deleted; -1 if ROW_IDX doesn't exist or error.
*/
static int
TblDeleteSelection(w, idx, mode)
register AXWidget_table w;
register int idx;
int mode;
{
    register int ix;
    register int num_selections;
    register int tempint;
    register int nx;
    register table_info *tp;
    long *new_selected = NULL;
    long *selected =  NULL;
    int found = FALSE;

    if (!w)
        return(-1);
    if (!(tp = (table_info *) w->info.table))
        return(-1);

    if (mode == ROW_MODE) {
        if (!tp->selected_rows) {
            tp->selected_rows = CreateIntsArray(0); /* sanity */
            return(-1);
        }
        selected = tp->selected_rows;
    } else {
        if (!tp->selected_columns) {
            tp->selected_columns = CreateIntsArray(0); /* sanity */
            return(-1);
        }
        selected = tp->selected_columns;
    }

    num_selections = IntsArraySize(selected);
    if (num_selections <= 0)
        return(-1);
    if (num_selections == 1) {
        tempint = ReadIntsArray(selected, 0);
        if (tempint != idx)
            return(-1);
        if (selected)
            free(selected);
        if (mode == ROW_MODE)
            tp->selected_rows = CreateIntsArray(0);
        else
            tp->selected_columns = CreateIntsArray(0);
        return(0);
    }

    if (!InIntsArray(selected, idx))
        return(-1);
    new_selected = CreateIntsArray(num_selections - 1);
    for (ix = 0, nx = 0; ix < num_selections; ix++) {
        tempint = ReadIntsArray(selected, ix);
        if (tempint == idx) {
            found = TRUE;
            continue;
        }
        WriteIntsArray(new_selected, nx, tempint); nx++;
    }

    if (selected)
        free(selected);

    if (mode == ROW_MODE)
        tp->selected_rows = new_selected;
    else
        tp->selected_columns = new_selected;

    return(0);
}

/***********************************************************************
function: TblResetCellTextInfo - Zero out cursor text related info.
*/
static void
TblResetCellTextInfo(tp)
register table_info *tp;
{
    if (!tp)
        return;

    tp->cursor.wtext = NULL;
    tp->cursor.text_len = 0;
    tp->cursor.left_char_pos = 0;
    tp->cursor.char_pos = 0;
    tp->cursor.right_char_pos = 0;
    tp->cursor.end_char_pos = 0;
    return;
}

/***********************************************************************
function: TblUpdCursorCellInfo - (Re)compute cursor-cell dimensions/position.
*/
static void
TblUpdCursorCellInfo(tp)
register table_info *tp;
{
    register int xpos;
    register int ypos;

    if (!tp)
        return;
    if ((tp->cursor.row < 0) || (tp->cursor.col < 0))
        return;
    if (!tp->colWidth)
        return;
    xpos = XPosOfCell(tp, tp->cursor.row, tp->cursor.col);
    ypos = YPosOfCell(tp, tp->cursor.row, tp->cursor.col);
    /* Note: tp->cursor.cell_{xpos,ypos} == 0 means the cursor cell is
       not visible.  `TblDrawCursorCellText()' looks for this condition. */
    tp->cursor.cell_xpos = (xpos == -1) ? 0 : xpos;
    tp->cursor.cell_ypos = (ypos == -1) ? 0 : ypos;
    tp->cursor.cell_width = AxIntFromArray(tp->colWidth, tp->cursor.col);
    tp->cursor.cell_height = tp->rowHeight;
    
    if (tp->cursor.text_cursor_on) {
        TblComputeCursorPos(tp, &tp->cursor.cursor_x, &tp->cursor.cursor_y);
    }
    return;
}

/***********************************************************************
function: TblUpdCursorCellText - transfer cell text from row to tp->cursor
*/
static void
TblUpdCursorCellText(tp, txt, len)
table_info *tp;
mbchar *txt;
int len;
{
    if (len == 0) {
        tp->cursor.wtext = NULL;
        tp->cursor.text_len = 0;
    } else {
        tp->cursor.wtext = (wchar *) malloc(sizeof(wchar) * (len + 1));
        MbToWStr(txt, tp->cursor.wtext);
        tp->cursor.text_len = Wstrlen(tp->cursor.wtext);
    }
}

/***********************************************************************
function: XPosOfCell - Return the x position of a cell
Return -1 if cell is not visible.
Note: `TblUpdCursorCellInfo()' checks for this return value.
*/
/*ARGSUSED*/
static int
XPosOfCell(tp, row, col)
register table_info *tp;
int row;
register int col;
{
    register int xpos = -1;
    register int ix;

    if (!tp)
        return(-1);
    if (col < tp->leftCol)
        return(-1);
    for (xpos = tp->markerWidth, ix = tp->leftCol;
         ((xpos < tp->winWidth) && (ix < col)); ix++) {
        xpos += AxIntFromArray(tp->colWidth, ix);
    }
    return(xpos);
}

/***********************************************************************
function: YPosOfCell - Return the y position of a cell
Return -1 if cell is not visible.
Note: `TblUpdCursorCellInfo()' checks for this return value.
*/
/*ARGSUSED*/
static int
YPosOfCell(tp, row, col)
register table_info *tp;
register int row;
int col;
{
    register int ypos = -1;
    register int ix;

    if (!tp)
        return(-1);
    if (row < tp->topRow)
        return(-1); /* won't see it */

    for (ypos = tp->headingHeight, ix = tp->topRow;
         ((ypos < tp->winHeight) && (ix < row)); ix++)
        ypos += tp->rowHeight;
    return(ypos);
}

/*************************************
 * function: TblComputeCursorPos - compute reference position
 * of i-bar given tp->cursor.cell_xpos/ypos
 */
static void
TblComputeCursorPos(tp, xpos, ypos)
table_info *tp;                    /* in */
int *xpos, *ypos;                  /* out */
{
    int ix, cx, cy;
    wchar *wstr;
    
    cx = tp->cursor.cell_xpos;
    cy = tp->cursor.cell_ypos;
        
    if (!tp->cursor.wtext || !tp->cursor.text_len) {
        cx += 3;
    }
    else {
        /* xpos of the first char in this cell is (cx + 3)!! */
        for (ix = tp->cursor.left_char_pos, cx = cx + 3,
             wstr = &tp->cursor.wtext[tp->cursor.left_char_pos];
             ix < tp->cursor.char_pos; ix++, wstr++) {
            
            cx += AXWTextWidth(tp->font, wstr, 1);
        }
    }
    *xpos = cx;
    *ypos = cy;
}

/***********************************************************************
function: TblHighlightRow - Place a visual mark on row ROW_IDX.
*/
static int
TblHighlightRow(w, row_idx, col_idx)
register AXWidget_table w;
register int row_idx;               /* in: hilight this row index */
register int col_idx;               /* in: start from this column index */
{
    register table_info *tp = (table_info *) NULL;
    int xpos, ypos, ix, nCols;
    int width, height;

    if (!(tp = (table_info *) w->info.table))
        return(0);
    if (row_idx < 0)
        return(0); /* nothing to do */
    if (row_idx < tp->topRow)
        return(0);
    
    nCols = AxArraySize(tp->colWidth);
    height = tp->rowHeight - 4;
    ypos = tp->headingHeight + 2 + ((row_idx - tp->topRow) * tp->rowHeight);
    
    for (ix = tp->leftCol, xpos = tp->markerWidth; ix < col_idx; ix++) {
        xpos += AxIntFromArray(tp->colWidth, ix);
    }
    
    for (ix = col_idx; ix < nCols; ix++) {
        
        width = AxIntFromArray(tp->colWidth, ix);
        
        TblHilightFill(w, xpos + 1, ypos, width - 2, height,
                       TblRowIsSelected(tp, row_idx));
        
        if ((tp->cursor.col == ix) && (tp->cursor.row == row_idx)) {
            TblDrawCursorCellText(w);
            TblPutSelectionCursor(w);
            TblDrawEditCellBox(w, row_idx, ix);
        }
        else {
            r32 clip;
            
            clip.origin.x = xpos + 1;
            clip.origin.y = tp->headingHeight +
                ((row_idx - tp->topRow) * tp->rowHeight) + 1;
            clip.width = AxIntFromArray(tp->colWidth, ix) - 1;
            clip.height = tp->rowHeight - 1;
                
            TblDrawCellText(w, row_idx, ix, xpos, ypos - 2, &clip);
        }
        
        xpos += width;
    }

    return(0);
}

/***********************************************************************
function: TblHighlightColumn - Place a visual mark on column COL_IDX.
*/
static int
TblHighlightColumn(w, col_idx)
AXWidget_table w;
int col_idx;
{
    return(0); /* TODO */
}

/***********************************************************************
function: TblMakeCellVisible - Make cell at ROW,COL visible by
adjusting `topRow', `leftCol' and then issuing a PaintTable().
Return 0 if successful; -1 otherwise.
*/
static int
TblMakeCellVisible(w, row, col)
register AXWidget_table w;
int row;
int col;
{
    register table_info *tp;
    int xpos, ypos;
    int top_row, left_col;

    if (!w || (row < 0) || (col < 0))
        return(-1);
    if (!(tp = (table_info *) w->info.table))
        return(-1);

    top_row  = tp->topRow;
    left_col = tp->leftCol;
    xpos     = XPosOfCell(tp, row, col);
    ypos     = YPosOfCell(tp, row, col);
    if (row < tp->topRow) {
        if (tp->alis_style_scroll) { /* cool */
            top_row = row - (tp->numLines / 2) + 1;
            if (top_row < 0)
                top_row = 0;
        } else {
            top_row = row;
            if (top_row < 0)
                top_row = 0;
        }
        TblScrollNewTopRow(w, top_row);
        return(0);
    } else if ((row > (tp->topRow + tp->numLines - 1)) /* -1 accu-bot */
               || ((ypos + (tp->rowHeight / 2)) >= (tp->winHeight))) {
        if (tp->alis_style_scroll) { /* cool */
            top_row = row - (tp->numLines / 2) + 1;
            if (top_row < 0)
                top_row = 0;
        } else {
            /* accu-bot */
            top_row = row - tp->numLines + 1;
            if (top_row < 0)
                top_row = 0;
        }       
        TblScrollNewTopRow(w, top_row);
        return(0);
    }

    if ((col < tp->leftCol) || (xpos >= tp->winWidth)) {
        TblScrollNewLeftCol(w, col);
        return(0);
    }
    return(0); /* nothing happened */
}

static int
TblMakeCursorCharVisible(w, start_char_pos, end_char_pos)
AXWidget_table w;
int start_char_pos;
int end_char_pos;
{
    table_info *tp;
    wchar *wstrp = (wchar *) NULL;
    int this_char_pos;
    int start_char_xpos;
    int xleft;
    int xpos;

    if (!w)
        return(FALSE);
    if (!(tp = (table_info *) w->info.table))
        return(FALSE);
    if ((tp->cursor.col < 0) || (tp->cursor.row < 0))
        return(FALSE); /* no cursor */
    if (tp->leftCol > tp->cursor.col)
        return(FALSE); /* cell not visible */
    if ((tp->cursor.cell_xpos < tp->markerWidth) ||
        (tp->cursor.cell_ypos < tp->headingHeight))
        return(FALSE); /* absurd value -- won't see it. assert()? */

    if (!tp->cursor.wtext || !tp->cursor.text_len)
        return(FALSE); /* no text */
    if (start_char_pos > tp->cursor.text_len)
        return(FALSE); /* out of bounds */

    if (start_char_pos == tp->cursor.left_char_pos)
        return(FALSE); /* no change -- note: ignoring `end_char_pos' */
    if (start_char_pos < tp->cursor.left_char_pos) {
        tp->cursor.left_char_pos = start_char_pos; /* favour to caller */
        return(TRUE);
    }

/* Compute xpos of the end of `start_char_pos' into `start_char_xpos' */
    start_char_xpos = tp->cursor.cell_xpos + 3;
    for (this_char_pos = 0, wstrp = (wchar *) &tp->cursor.wtext[0];
         *wstrp; wstrp++, this_char_pos++) {
        start_char_xpos += AXWTextWidth(tp->font, wstrp, 1);
        if (this_char_pos == start_char_pos)
            break;
    }

    if (start_char_xpos <
        (tp->cursor.cell_xpos + tp->cursor.cell_width - 3)) {
        /* `start_char_pos' is wholly inside the cell */
        if (tp->cursor.left_char_pos != 0) {
            tp->cursor.left_char_pos = 0;
            return(TRUE);
        } else
            return(FALSE);
    }

/* `left_char_pos' needs to be shifted. */
    xleft = start_char_xpos - (tp->cursor.cell_width - 3);
    xpos = tp->cursor.cell_xpos + 3;
    for (this_char_pos = 0, wstrp = (wchar *) &tp->cursor.wtext[0];
         *wstrp; wstrp++, this_char_pos++) {
        if (xpos >= xleft)
            break;
        xpos += AXWTextWidth(tp->font, wstrp, 1);
    }
    tp->cursor.left_char_pos = this_char_pos;
    return(TRUE);
}

/***********************************************************************
function: TblDrawRow - Draw row.  For use after adding/removing a selection.
Returns: 0 on success; -1 otherwise.
*/
static int
TblDrawRow(w, row_num, col_num)
register AXWidget_table w;
register int row_num;
register int col_num;
{
    register table_info *tp;

    if ((row_num < 0) || (!w))
        return(-1);
    if (!(tp = (table_info *) w->info.table))
        return(-1);

    TblHighlightRow(w, row_num, col_num);
    if (TblRowIsSelected(tp, row_num))
        TblDrawRowMarkerShadow(w, row_num, TRUE);
    else
        TblDrawRowMarkerShadow(w, row_num, FALSE);
    TblDrawRowMarkerContents(w, row_num);
    return(0);
}

/***********************************************************************
function: TblDrawColumn - Draw column.  For use after adding/removing
a selection.
Note: Does not draw column material!! (just heading and hilighting)
Returns: 0 on success; -1 otherwise.
*/
static int
TblDrawColumn(w, col_num)
AXWidget_table w;
int col_num;
{
    table_info *tp;

    if ((col_num < 0) || (!w))
        return(-1);
    if (!(tp = (table_info *) w->info.table))
        return(-1);

    TblHighlightColumn(w, col_num);
    if (TblColIsSelected(tp, col_num))
        TblDrawHeadingShadow(w, col_num, TRUE);
    else
        TblDrawHeadingShadow(w, col_num, FALSE);
    TblDrawHeadingContents(w, col_num);
    return(0);
}

/***********************************************************************
function: TblClearAllSelections - Clear all selections; toggle hilighting
on them.  Update display.
Return 1 if selections changed (row(s) selections were cleared); 0 otherwise.
*/
static int
TblClearAllSelections(w)
register AXWidget_table w;
{
    register table_info *tp;
    register int ix;
    int tempint;
    int num_selections = 0;
    int row_selection_changed = FALSE;
    int col_selection_changed = FALSE;
    long *tmp_selections = NULL;

    if (!w)
        return(0);
    if (!(tp = (table_info *) w->info.table))
        return(0);

    if (!tp->selected_rows) {
        tp->selected_rows = CreateIntsArray(0);
        row_selection_changed = FALSE;
    } else {
        if ((num_selections = IntsArraySize(tp->selected_rows)) > 0) {
            tmp_selections = CopyIntsArray(tp->selected_rows);
            for (ix = 0; ix < num_selections; ix++) {
                tempint = ReadIntsArray(tmp_selections, ix);
                if (TblDeleteSelection(w, tempint, ROW_MODE) == 0)
                    TblDrawRow(w, tempint, tp->leftCol);
            }
            if (tp->selected_rows)
                free(tp->selected_rows);
            tp->selected_rows = CreateIntsArray(0);
            if (tmp_selections)
                free(tmp_selections);
            row_selection_changed = TRUE;
        } else
            row_selection_changed = FALSE;
    }

    if (!tp->selected_columns) {
        tp->selected_columns = CreateIntsArray(0);
        col_selection_changed = FALSE;
    } else {
        if ((num_selections = IntsArraySize(tp->selected_columns)) > 0) {
            tmp_selections = CopyIntsArray(tp->selected_columns);
            for (ix = 0; ix < num_selections; ix++) {
                tempint = ReadIntsArray(tmp_selections, ix);
                if (TblDeleteSelection(w, tempint, COL_MODE) == 0)
                    TblDrawColumn(w, tempint);
            }
            if (tp->selected_columns)
                free(tp->selected_columns);
            tp->selected_columns = CreateIntsArray(0);
            if (tmp_selections)
                free(tmp_selections);
            col_selection_changed = TRUE;
        } else
            col_selection_changed = FALSE;
    }

    return((row_selection_changed || col_selection_changed));
}

/***********************************************************************
function: TblTransferCursorCellData - Take text in the cursor cell,
and transfer it into the table data datum.
*/
static void
TblTransferCursorCellData(w)
register AXWidget_table w;
{
    register table_info *tp;
    data_ptr string_datum;
    mbchar *modified_cell_string = NULL;
    int   modified_cell_string_length;
    mbchar *table_cell_string;
    int row;
    int col;
    data_ptr this_row;
    data_ptr new_row_datum;

    if (!w)
        return;
    if (!(tp = (table_info *) w->info.table))
        return;
    if (!tp->rows)
        return;
    row = tp->cursor.row;
    col = tp->cursor.col;
    if ((row < 0) || (col < 0))
        return; /* safety net */

/* Safety net!! */
    if (!tp->allow_cell_editing) {
        /* Must not even get here!! */
        if (tp->cursor.wtext)
            free(tp->cursor.wtext);
        tp->cursor.row = -1;
        tp->cursor.col = -1;
        TblResetCellTextInfo(tp);
    }

    this_row  = ReadArray(AX(tp->rows), row);
    
    if (this_row == (data_ptr) NULL) {
        /* we're in this func in order to tranfer data from tp->cursor
         * onto the corresponding element of tp->rows.  But the nullness
         * of this row indicates that we don't know the actual data for it.
         */
/* %%%%% why are we here then? and should tp->cursor be reset?? */      
        return;
    }

    table_cell_string = (mbchar *)AxStrFromArray(this_row, col);

/* Form `modified_cell_string'. */
    if (!(tp->cursor.wtext) || (tp->cursor.text_len <= 0)) {
        /* the modified cell is empty */
        if (MBstrlen(table_cell_string) == 0) {
            /* Cell in table data is also empty.  Don't bother
               transfering contents.  There is nothing to modify. */
            return;
        }

        /* The cell in table data has contents.  Will therefore need
           to transfer an empty string. */
        modified_cell_string = (mbchar *)malloc(sizeof(mbchar));
        modified_cell_string[0] = 0;
        modified_cell_string_length = 0;
    } else {
        modified_cell_string = (mbchar *)calloc(1, 
                                 ((tp->cursor.text_len+1)*sizeof(wchar)));
        /* copy contents of `tp->cursor.wtext' over. */
        modified_cell_string_length = WnToMbStr(tp->cursor.wtext, 
                                 modified_cell_string, tp->cursor.text_len);
        modified_cell_string[modified_cell_string_length] = 0;
    }

/* Form `string_datum' with `modified_cell_string'. */
    string_datum = AxTaskCreateElfString(0, modified_cell_string);

    free(modified_cell_string);
    modified_cell_string = (mbchar *) NULL;

    new_row_datum = AxTaskCopyElfData(0, this_row);
    new_row_datum = TaskWriteArray(0, new_row_datum, col, string_datum);
    tp->rows = TaskWriteArray(0, tp->rows, row, new_row_datum);
    return;
}

/***********************************************************************
function: TblCommitCursorCell - Commit changes in cursor cell; lift
cursor; set cursor row/col to -1.
*/
static void
TblCommitCursorCell(w)
register AXWidget_table w;
{
    register table_info *tp;

    if (!w)
        return;
    if (!(tp = (table_info *) w->info.table))
        return;

    if ((tp->cursor.row >= 0) && (tp->cursor.col >= 0)) {
        TblDrawEditCellBox(w, tp->cursor.row, tp->cursor.col);
        TblLiftSelectionCursor(w);
        /* Redraw if necessary */
        if (tp->cursor.left_char_pos != 0) {
            tp->cursor.left_char_pos = 0;
            TblDrawCursorCellText(w);
        }
        TblTransferCursorCellData(w);
    }

    if (tp->cursor.wtext)
        free(tp->cursor.wtext);

    TblResetCellTextInfo(tp);
    tp->cursor.row = -1;
    tp->cursor.col = -1;
    tp->doing_cell_editing = FALSE;
    return;
}

/***********************************************************************
function: TblPointOnColumnBoundary - Check if point (x, y) is on any
column boundary.  Column boundary is the right edge of a column.
Return column number if so; -1 otherwise.
*/
/*ARGSUSED*/
static int
TblPointOnColumnBoundary(w, x, y)
register AXWidget_table w;
register int x;
register int y; /* unused */
{
    register table_info *tp;
    register int xpos;
    int col_num;
    int this_width;
    int tolerance = 2; /* pixel tolerance on each side of point. */

    if (!w)
        return(-1);
    if (!(tp = (table_info *) w->info.table))
        return(-1);

    for (col_num = (int) tp->leftCol, xpos = tp->markerWidth;
         xpos < (x + tolerance); col_num++) {
        this_width = AxIntFromArray(tp->colWidth, col_num);
        if (this_width <= 0) { /* safety net */
            xpos += 10;
            continue;
        } else
            xpos += this_width;
        if ((x >= (xpos - tolerance)) && (x <= (xpos + tolerance)))
            return(col_num);
    }

    return(-1);
}

/***********************************************************************
function: TblGotoCell - Put cursor in cell at (ROW_NUM, COL_NUM).
Always keep at most 1 cursor.
If X and Y are not -1, compute `cursorCharPos' so that the cursor is near
the X and Y values.
*/
int
TblGotoCell(w, row_num, col_num, x, y)
register AXWidget_table w;
register int row_num;
register int col_num;
int x, y;
{
    register table_info *tp;
    wchar *wstr_p;
    wchar *wstr;
    int xpos;
    int char_pos;
    int exact = TRUE;
    int default_char_pos;
    int max_char_pos;
    int text_len;
    int num_rows, num_cols;
    int xleft, xright;
    int selection_changed = FALSE;
    int from_row, from_col;
    int to_row, to_col;

    if ((row_num < 0) || (col_num < 0) || (!w))
        return(-1);
    if (!(tp = (table_info *) w->info.table))
        return(-1);
    if (!tp->rows)
        return(-1);
    if (!tp->colText)
        return(-1);

    if ((x == -1) && (y == -1))
        exact = FALSE;
    else
        exact = TRUE;

    num_rows = AxArraySize(tp->rows);
    num_cols = AxArraySize(tp->colText);

    if ((row_num >= num_rows) || (col_num >= num_cols))
        return(-1);

    from_row = tp->cursor.row;
    from_col = tp->cursor.col;
    to_row = row_num;
    to_col = col_num;
    if ((tp->cursor.row != row_num) || (tp->cursor.col != col_num)) {
        
        /* Fill in cursor-row and cursor-col *after* this call to
           `TblDupCellString ()'. */
        wstr = TblDupCellString(w, row_num, col_num);
        if (!wstr) {
            text_len = 0;
        }
        else {
            text_len = Wstrlen(wstr);
        }
        if (text_len == 0) {
            free(wstr);
            wstr = NULL;
        }

        /* Create new `tp->cursor' information. */
        selection_changed = TblClearAllSelections(w);
        TblCommitCursorCell(w); /* also lifts cursor */
        if (selection_changed)
            TblSendSelectionChangeEvent(w);

        TblMakeCellVisible(w, row_num, col_num);
        tp->cursor.row = row_num;
        tp->cursor.col = col_num;
        TblUpdCursorCellInfo(tp);

        tp->cursor.wtext = wstr;
        tp->cursor.text_len = text_len;
        tp->cursor.left_char_pos = 0; /* not needed */
    } else {
        /* Still in the same cell. */
        text_len = tp->cursor.text_len;
        wstr = tp->cursor.wtext;
        TblLiftSelectionCursor(w);
    }

    max_char_pos = text_len;
    if (exact)
        default_char_pos = max_char_pos;
    else
        default_char_pos = 0;

    xpos = tp->cursor.cell_xpos;
    if (exact == TRUE && text_len) { /* `wstr' is NULL when !text_len */
        char_pos = tp->cursor.left_char_pos;
        /* `wstr' formed above. */
        wstr_p = (wchar *) &wstr[tp->cursor.left_char_pos];
        xleft = xpos + 3; /* !! */
        while (x > xpos) {
            if (char_pos < max_char_pos) {
                xright = xleft + AXWTextWidth(tp->font, wstr_p, 1);
            } else {
                char_pos = max_char_pos;
                break;
            }
            if ((x <= xright) && (x >= xleft))
                break;
            xleft = xright;
            char_pos++;
            wstr_p++;
        }
        tp->cursor.char_pos = char_pos;
    } else
        tp->cursor.char_pos = default_char_pos;

    TblPutSelectionCursor(w); /* put cursor */
    TblDrawEditCellBox(w, row_num, col_num); /* nop */
    tp->doing_cell_editing = TRUE;
    if ((from_row != to_row) || (from_col != to_col))
        TblSendFocusChangeEvent(w, from_row, from_col, to_row, to_col);
    return(0);
}

/***********************************************************************
function: TblKeepCursorCell - Keep row_num,col_num as the cursor cell.
start_char_pos is a 0 based char index into cell.
Return 0 on success; -1 otherwise
Note: Currently ignores end-char-index (i.e. no way to hilight a bunch
of characters in a cell for pending delete...).

(NOTE:  This function is called by AxXmGetTableRows() ONLY so that the
 dipper works properly.  The dipper does not like loosing the cursor
 cell after it does a AxXmGetTableRows())

Functionality is identical to TblGotoCellChar() except that this
function does not try to make cell visible...
*/
static int
TblKeepCursorCell(w, row_num, col_num, start_char_pos, end_char_pos)
AXWidget_table w;
int row_num;
int col_num;
int start_char_pos;
int end_char_pos;
{
    table_info *tp;
    data_ptr rows;
    int selection_changed = FALSE;
    data_ptr this_row;
    mbchar *mbstr;
    int text_len;
    int num_rows;
    int num_cols;

    if ((row_num < 0) || (col_num < 0) || (!w))
        return(-1);
    if (!(tp = (table_info *) w->info.table))
        return(-1);
    if (!(rows = tp->rows))
        return(-1);
    if (!tp->colText)
        return(-1);

    num_rows = AxArraySize(tp->rows);
    num_cols = AxArraySize(tp->colText);
    if ((row_num >= num_rows) || (col_num >= num_cols))
        return(-1);

    tp->cursor.row = row_num;
    tp->cursor.col = col_num;
    TblUpdCursorCellInfo(tp);

    this_row = ReadArray(AX(rows), row_num);

    if (this_row) {
        mbstr = (mbchar *) AxStrFromArray(this_row, col_num); /* null-term */
        if (!mbstr)
            text_len = 0;
        else
            text_len = MbCharLen(mbstr); /* length in chars, not bytes */
    } else {
        mbstr = (mbchar *) NULL;
        text_len = 0;
    }
    
    TblUpdCursorCellText(tp, mbstr, text_len);
    
    tp->cursor.char_pos = start_char_pos;
    return(0);
}

/***********************************************************************
function: TblGotoCellChar - Put cursor in cell at (ROW_NUM, COL_NUM),
given a 0 based start-char-index and end-char-index into the string in cell.
Always keeps at most 1 cursor.
Return 0 on success; -1 otherwise
Note: Currently ignores end-char-index (i.e. no way to hilight a bunch
of characters in a cell for pending delete...).
*/
static int
TblGotoCellChar(w, row_num, col_num, start_char_pos, end_char_pos)
AXWidget_table w;
int row_num;
int col_num;
int start_char_pos;
int end_char_pos;
{
    table_info *tp;
    data_ptr rows;
    int selection_changed = FALSE;
    data_ptr this_row;
    mbchar *mbstr;
    int text_len;
    int num_rows;
    int num_cols;
    int left_char_shifted;
    int from_row, from_col, to_row, to_col;

    if ((row_num < 0) || (col_num < 0) || (!w))
        return(-1);
    if (!(tp = (table_info *) w->info.table))
        return(-1);
    if (!(rows = tp->rows))
        return(-1);
    if (!tp->colText)
        return(-1);

    num_rows = AxArraySize(tp->rows);
    num_cols = AxArraySize(tp->colText);
    
    if (col_num >= num_cols) {
        return (-1);              /* beyond row width */
    }
    
    if (!tp->use_data_request_model && (row_num >= num_rows)) {
        return (-1);              /* beyond known #rows */
    }
    
    if ((row_num < tp->topRow) ||
        (row_num > (tp->topRow + tp->numLines - 1))) {
        
        /* requested row_num is currently out of view.
         * scroll so as to put row_num about mid-window.
         */
        TblScrollNewTopRow(w, row_num - (tp->numLines / 2));
    }
    else if (row_num == (tp->topRow + tp->numLines - 1)) {
        /* request row_num is the current bottom visible row.
         */
        int ypos = YPosOfCell(tp, row_num, 0);
        if ((ypos + tp->rowHeight - 1) > tp->winHeight) {
            /* row is currently partially obscured.  scroll
             */
            TblScrollNewTopRow(w, row_num - (tp->numLines / 2));
        }
    }
    
    from_row = tp->cursor.row;
    from_col = tp->cursor.col;
    to_row = row_num;
    to_col = col_num;
    selection_changed = TblClearAllSelections(w);
    TblCommitCursorCell(w);
    if (selection_changed)
        TblSendSelectionChangeEvent(w);

    TblMakeCellVisible(w, row_num, col_num);
    tp->cursor.row = row_num;
    tp->cursor.col = col_num;
    TblUpdCursorCellInfo(tp);

    this_row = ReadArray(AX(rows), row_num);
    
    if (this_row) {
        mbstr = (mbchar *) AxStrFromArray(this_row, col_num); /* null-term */
        if (!mbstr)
            text_len = 0;
        else
            text_len = MbCharLen(mbstr); /* length in chars, not bytes */
        
    } else {
        mbstr = (mbchar *) NULL;
        text_len = 0;
    }

    TblUpdCursorCellText(tp, mbstr, text_len);
    
    tp->cursor.char_pos = start_char_pos;
/*  tp->cursor.end_char_pos = end_char_pos; */

    /* Note this must happen when there is a valid cursor cell.
       i.e.  all elements of `tp->cursor' are valid. */
    left_char_shifted = TblMakeCursorCharVisible(w, start_char_pos,
                                                 end_char_pos);
    if (left_char_shifted)
        TblDrawCursorCellText(w);

    TblPutSelectionCursor(w);

    TblDrawEditCellBox(w, row_num, col_num); /* nop */
    if ((from_row != to_row) || (from_col != to_col))
        TblSendFocusChangeEvent(w, from_row, from_col, to_row, to_col);
    return(0);
}

/***********************************************************************
function: TblDupCellString - Return the duplicate of the cell string
in ROW, COL.
Return: Allocated wstring of cell contents on success, NULL on failure.
Note:  Caller is responsible for freeing returned pointer.
*/
static wchar *
TblDupCellString(w, row, col)
AXWidget_table w;
int row;
int col;
{
    table_info *tp;
    wchar *wstr = (wchar *) NULL;
    mbchar *mbstr = (mbchar *) NULL;
    data_ptr rows;
    data_ptr this_row;
    int text_len;

    if (!w)
        return((wchar *) NULL);
    if (!(tp = (table_info *) w->info.table))
        return((wchar *) NULL);
    if (!(rows = tp->rows))
        return((wchar *) NULL);

    if ((tp->cursor.row == row) && (tp->cursor.col == col)) {
        /* This is the cursor cell.  Take (potentially) modified
           contents. */
        if (!tp->cursor.wtext || !tp->cursor.text_len) {
            wstr = (wchar *) malloc(sizeof(wchar));
            *wstr = 0;
            return(wstr);
        }
        text_len = tp->cursor.text_len;
        wstr = (wchar *) malloc(sizeof(wchar) * (text_len + 1));
        Wstrcpy(wstr, tp->cursor.wtext);
    } else {
        this_row = ReadArray(AX(rows), row);
        
        if (this_row) {
            mbstr = (mbchar *) AxStrFromArray(this_row, col);
            if (mbstr)
                text_len = MbCharLen(mbstr);
            else
                text_len = 0;
        } else {
            text_len = 0;
        }
        
        if (!text_len) {
            wstr = (wchar *) malloc(sizeof(wchar));
            *wstr = 0;
            return(wstr);
        }
        wstr = (wchar *) malloc(sizeof(wchar) * (text_len + 1));
        MbToWStr(mbstr, wstr);
    }
    return(wstr);
}

/***********************************************************************
function: TblFindCharIndexInCell - Given (X, Y) that lies inside (ROW, COL),
return the character index at under that co-ordinate.
Return -1 on error.
Note:  Follows identical method as in the `exact' case in TblGotoCell().
*/
static int
TblFindCharIndexInCell(w, row, col, x, y)
AXWidget_table w;
int row;
int col;
int x;
int y;
{
    table_info *tp;
    data_ptr rows;
    data_ptr this_row;
    wchar  *wstr   = (wchar *) NULL;
    wchar  *wstr_p = (wchar *) NULL;
    int xpos;
    int char_pos;
    int default_char_pos;
    int max_char_pos;
    int text_len;
    int num_rows;
    int num_cols;
    int xleft, xright;

    if (!w)
        return(-1);
    if (!(tp = (table_info *) w->info.table))
        return(-1);
    if ((row < 0) || (col < 0))
        return(-1);
    if (!(rows = tp->rows))
        return(-1);

    num_rows = AxArraySize(tp->rows);
    num_cols = AxArraySize(tp->colText);
    if ((row >= num_rows) || (col >= num_cols))
        return(-1);

    xpos = XPosOfCell(tp, row, col);
    this_row = ReadArray(AX(rows), row);

    wstr = TblDupCellString(w, row, col);
    if (!wstr)
        return(-1);
    wstr_p = wstr;
    text_len = Wstrlen(wstr);
    max_char_pos = text_len;

    if (x < (xpos + 3)) /* !! */
        default_char_pos = 0; /* no need to loop */
    else
        default_char_pos = max_char_pos;

    if (text_len && default_char_pos) {
        char_pos = 0;
        xleft = xpos + 3; /* !! */
        while (x > (xpos + 3)) { /* !! */
            if (char_pos < max_char_pos) {
                xright = xleft + AXWTextWidth(tp->font, wstr_p, 1);
            } else {
                char_pos = max_char_pos;
                break;
            }
            if ((x <= xright) && (x >= xleft))
                break;
            xleft = xright;
            char_pos++;
            wstr_p++;
        }
    } else
        char_pos = default_char_pos; /* 0 */

    if (wstr)
        free(wstr);

    return(char_pos);
}

/***********************************************************************
function: TblPutSelectionCursor - Put a cursor after
`tp->cursor.char_pos' and upto `tp->cursor.end_char_pos'.
*/
static int
TblPutSelectionCursor(w)
register AXWidget_table w;
{
    return (TblDrawSelectionCursor(w, FALSE));
}

/***********************************************************************
function: TblLiftSelectionCursor - Lift text editing cursor
*/
int
TblLiftSelectionCursor(w)
register AXWidget_table w;
{
    return (TblDrawSelectionCursor(w, TRUE));
}

/***********************************************************************
function: TblDrawSelectionCursor - do text cursor display func called by
          TblPutSelectionCursor, TblLiftSelectionCursor.
*/
static int
TblDrawSelectionCursor(w, lift)
register AXWidget_table w;        /* in */
register int lift;                /* in; bool: whether lifting cursor */
{
    int xpos, ypos;
    table_info *tp;
#if AXMS
	HWND hwnd;
	int caret_hyt;
#endif

    if (!w)
        return(-1);
    
    if (!(tp = (table_info *) w->info.table))
        return(-1);

/* Filter out redundant lift/draw attempts */
    if (lift) {
        if (!tp->cursor.text_cursor_on)
            return(-1);
        tp->cursor.text_cursor_on = FALSE;
    }
    else {
        if (tp->cursor.text_cursor_on)
            return(-1);
        
        tp->cursor.text_cursor_on = TRUE;
    }

/* Decide whether not to do drawing work */
    if ((tp->cursor.col < 0) || (tp->cursor.row < 0))
        return(-1);

    if (tp->leftCol > tp->cursor.col)
        return(-1); /* won't see it */
    
    if (tp->topRow > tp->cursor.row)
        return(-1); /* won't see it */
    
    if (tp->cursor.row > (tp->topRow + tp->numLines - 1))
        return(-1); /* won't see it */

    if ((tp->cursor.cell_xpos < tp->markerWidth) ||
        (tp->cursor.cell_ypos < tp->headingHeight))
        return(-1); /* absurd value.  won't see it */

/* Get draw position */
    if (!lift) {
        TblComputeCursorPos(tp, &tp->cursor.cursor_x, &tp->cursor.cursor_y);
    }
    xpos = tp->cursor.cursor_x;
    ypos = tp->cursor.cursor_y;

/* Draw cursor */
#if AXMS
	hwnd = GET_DRAWABLE_HANDLE(w->window);
	if ( lift )
		MS_HideCaret(hwnd);
	else
		{
		caret_hyt = (ypos + tp->rowHeight - 4) - (ypos + 2); 
		MS_CreateCaret(hwnd, (HBITMAP)NULL, 1, caret_hyt);
		MS_SetCaretPos(xpos, ypos + 2);
		MS_ShowCaret(hwnd);
		}
#else
    /* vert line */
    XDrawLine(Dpy, w->window, tp->mark_gc, xpos, ypos + 2, xpos,
              ypos + tp->rowHeight - 4);
    /* top horiz */
    XDrawLine(Dpy, w->window, tp->mark_gc, xpos - 2, ypos + 2,
              xpos + 2, ypos + 2);
    /* bottom horiz */
    XDrawLine(Dpy, w->window, tp->mark_gc,
              xpos - 2, ypos + tp->rowHeight - 4,
              xpos + 2, ypos + tp->rowHeight - 4);
#endif
    return(TRUE);
}

static int
TblCursorIntersectsRect(tp, x, y, width, height)
table_info *tp;                          /* in */
int x, y, width, height;                 /* in */
{
    r32 cursorRect, inRect;
    
    /* the magic numbers in the cursor rect must match those used
     * the i-bar draw code in TblDrawSelectionCursor
     */
    cursorRect.origin.x = tp->cursor.cursor_x - 2;
    cursorRect.origin.y = tp->cursor.cursor_y + 2;
    cursorRect.width = 4;
    cursorRect.height = tp->rowHeight - 2;
    
    inRect.origin.x = x;
    inRect.origin.y = y;
    inRect.width = width;
    inRect.height = height;
    
    return (rectanglesIntersect(&cursorRect, &inRect));
}

/***********************************************************************
function: TblSelectionRequestCB - Handle a SelectionRequest event in
the table widget.  Return 1 if event was consumed; 0 otherwise.
*/
static int
TblSelectionRequestCB(w, evp)
register AXWidget_table w;
register XEvent   *evp;
{
    int event_consumed = FALSE;
    return(event_consumed);
}

/***********************************************************************
function: TblSelectionNotifyCB - Handle a SelectionNotify event in the
table widget.  Return 1 if event was consumed; 0 otherwise.
*/
static int
TblSelectionNotifyCB(w, evp)
register AXWidget_table w;
register XEvent   *evp;
{
    int event_consumed = FALSE;
    return(event_consumed);
}

/***********************************************************************
function: TblSelectionClearCB - Handle a SelectionClear event in the
table widget.  Clear all selections. 
Return 1 if event was consumed; 0 otherwise.
*/
static int
TblSelectionClearCB(w, evp)
register AXWidget_table w;
register XEvent   *evp;
{
    int event_consumed = FALSE;
    if (!w)
        return(event_consumed);
    if (w->type != W_TABLE)
        return(event_consumed);

    TblClearAllSelections(w);
    event_consumed = TRUE;
    return(event_consumed);
}

/***********************************************************************
function: TblButtonPressCB - Handle a ButtonPress event in the table widget.
*/
static void
TblButtonPressCB(w, ev)
register AXWidget_table w;
register XEvent *ev;
{
    register table_info *tp = (table_info *) w->info.table;
    register int ix;
    register int num_rows_displayed = tp->numLines;
    int x = 0;
    int col_num;
    int selected_row_idx = -1;
    int selected_col_idx = -1;
    int max_rows;
    unsigned long time;
    int num_selected_rows;
    int num_selected_cols;

/*  w->last_click_time = ev->xbutton.time; */
    time = ev->xbutton.time;

    if (!w)
        return;
    if (!(tp = (table_info *) w->info.table))
        return;
    if (!tp->selected_rows)
        tp->selected_rows = CreateIntsArray(0); /* sanity */

#if LINUX /* mouse wheel scrolling */
	if (w->vScrollBar &&
		(AxButtonDown == Button4 || AxButtonDown == Button5)) {
		char *str = (char *)UPbyName(NXLT("axMouseWheelScroll"));
        int new_top_row;
		if (str && (atoi(str) != 0)) {
			if (AxButtonDown == Button4) { /* mouse scroll up */
				if (ThimIsShifted())
					new_top_row = tp->topRow - w->vScrollBar->scroll_max;
				else
					new_top_row = tp->topRow - 3;
			} else if (AxButtonDown == Button5) { /* mouse scroll down */
				if (ThimIsShifted())
					new_top_row = tp->topRow + w->vScrollBar->scroll_max;
				else
					new_top_row = tp->topRow + 3;
			}
		}
		/* TblScrollNewTopRow has sanity checks to make sure new_top_row is
		   is legit */
		TblScrollNewTopRow(w, new_top_row);
	}
	
	if (AxButtonDown != Button1)
		return;
#endif
	
/* find row in which button-press occured */
    for (ix = 0; ix < num_rows_displayed; ix++) {
        int top_y, bottom_y; /* top and bottom `y' of this row */

        top_y = tp->headingHeight + (ix * tp->rowHeight);
        bottom_y = top_y + tp->rowHeight;
        
        if ((ev->xbutton.y > top_y) && (ev->xbutton.y < bottom_y)) {
            selected_row_idx = ix + tp->topRow;
            break;
        }
    }

/* find column in which button press occured */
    for (col_num = (int) tp->leftCol, x = tp->markerWidth;
         x < ev->xbutton.x; col_num++) {
        int this_width = AxIntFromArray(tp->colWidth, col_num);

        if ((ev->xbutton.x > x) && (ev->xbutton.x < (x + this_width))) {
            selected_col_idx = col_num;
            break;
        }
        /* +10 to ensure an increasing `x'. */
        x += ((this_width <= 0) ? 10 : this_width);
    }

/* button press in the heading area */
    if (ev->xbutton.y < tp->headingHeight) {
        if (tp->colToResize != -1) { /* resizing columns */
            tp->doing_column_resizing = TRUE;
            colResize.colXpos = XPosOfCell(tp, 0, tp->colToResize);
            colResize.markerXpos = colResize.colXpos +
                AxIntFromArray(tp->colWidth, tp->colToResize);
            XDrawLine(Dpy, w->window, tp->mark_gc, colResize.markerXpos, 0,
                      colResize.markerXpos, tp->winHeight);
            return;
        } else if (selected_col_idx != -1) { /* marking columns */
#if 1 /* TODO, disable column marking (for now). */
            return;
#else
            w->value = selected_col_idx;
            TblCommitCursorCell(w);
            num_selected_cols = IntsArraySize(tp->selected_columns);
            if ((AxButtonDown == Button1) && (num_selected_cols == 1) &&
                (time - w->last_click_time <= DoubleMillisec) &&
                (selected_col_idx == w->last_item_selected)) {
                if (ThimIsControled()) {
                    w->last_click_time = time;
                    tp->doing_column_marking = FALSE;
                    return;
                }
                if (w->callback)
                    TblSendDoubleClickEvent(w, selected_col_idx, -1);
                tp->doing_column_marking = TRUE;
                w->last_click_time = time;
                return;
            }
            if (ThimIsControled()) {
                if (!w->multi_pick)
                    TblClearAllSelections(w);

                TblToggleSelection(w, selected_col_idx, COL_MODE);
            } else {
                TblClearAllSelections(w);
                if (TblAddSelection(w, selected_col_idx, COL_MODE) == 0)
                    TblHighlightColumn(w, selected_col_idx);
            }
            tp->doing_column_marking = TRUE;
            w->last_item_selected = selected_col_idx;
            w->last_click_time = time;
            return;
#endif
        }
        return; /* button press not in valid column */
    }

    max_rows = AxArraySize(tp->rows);

/* Need to have a valid row idx from this point on */
    if (selected_row_idx == -1)
        return;
	
	if ((w->flags & WFLG_INACTIVE) || (w->flags & WFLG_DLG_NOT_PICKABLE)) {
		return;
	}

/* If button-press is in the row-marker area, hilight the row and
   add to `selected_rows'.  If row is already higlighted, remove row 
   selection, and (un)highlight the row.  If table is display-only,
   the button press need not be in the row-marker area. */
    if (!tp->allow_cell_editing ||
        (tp->markerWidth && (ev->xbutton.x < tp->markerWidth))) {
        if (selected_row_idx >= max_rows)
            return;

        w->value = selected_row_idx;

        TblCommitCursorCell(w);

        /* Process left-btn-double-click */
        num_selected_rows = IntsArraySize(tp->selected_rows);
        if ((AxButtonDown == Button1) && (num_selected_rows == 1) &&
            (time - w->last_click_time <= DoubleMillisec) &&
            (selected_row_idx == w->last_item_selected)) {
            if (ThimIsControled()) {
                w->last_click_time = time;
                tp->doing_row_marking = FALSE;
                return;
            }
            if (w->callback)
                TblSendDoubleClickEvent(w, selected_row_idx, -1);
            tp->doing_row_marking = FALSE;
            w->last_click_time = time;
            return;
        }
        if (ThimIsControled()) {
            if (!w->multi_pick) {
                if (w->last_item_selected == selected_row_idx)
                    TblToggleSelection(w, selected_row_idx, ROW_MODE);
                else {
                    TblClearAllSelections(w);
                    TblToggleSelection(w, selected_row_idx, ROW_MODE);
                }
            } else
                TblToggleSelection(w, selected_row_idx, ROW_MODE);
        } else {
            TblClearAllSelections(w);
            if (TblAddSelection(w, selected_row_idx, ROW_MODE) == 0)
                TblDrawRow(w, selected_row_idx, tp->leftCol);
        }
        tp->doing_row_marking = TRUE;
#if 0
		/* This causes the bug where dbl clicking on a selected row does not
		send a double_click_event. It is to do with setting the 
		last_item_selected to -1. (mahul) */
        if (!w->multi_pick && (w->last_item_selected == selected_row_idx))
            w->last_item_selected = -1;  /* selection was deleted */
        else
#endif
        w->last_item_selected = selected_row_idx;
        w->last_click_time = time;
        return;
    }

/* Stop here if this is a display-only table. */
    if (!tp->allow_cell_editing)
        return;

/* Need to have a valid column idx from this point on */
    if (selected_col_idx == -1)
        return;

    if (w->minorEventNotify) {
        int start_char_pos;
        start_char_pos = TblFindCharIndexInCell(w, selected_row_idx,
                                                selected_col_idx,
                                                ev->xbutton.x,
                                                ev->xbutton.y);
        if (start_char_pos == -1) {
            RingTheBell();
            return; /* error */
        }
        if (!tp->doing_column_marking) {
            if ((AxButtonDown == Button1) &&
                (time - w->last_click_time <= DoubleMillisec) &&
                (selected_row_idx == w->last_item_selected)) {
                if (w->callback)
                    TblSendDoubleClickEvent(w, selected_row_idx,
                                            selected_col_idx);
                w->last_click_time = time;
                return;
            }
        }
        w->last_item_selected = selected_row_idx;
        TblSendButtonPressEvent(w, selected_row_idx, selected_col_idx,
                                start_char_pos + 1,  /* 1 based */
                                start_char_pos + 1); /* 1 based */
        w->last_click_time = time;
        if ( tp->rowcolcreate )
            TblGotoCell(w, selected_row_idx, selected_col_idx,
                        ev->xbutton.x, ev->xbutton.y);
    } else {
        TblGotoCell(w, selected_row_idx, selected_col_idx,
                    ev->xbutton.x, ev->xbutton.y);
    }
    return;
}

/***********************************************************************
function: TblButtonReleaseCB - Handle a ButtonRelease event in the
table widget.
*/
static void
TblButtonReleaseCB(w, evp)
register AXWidget_table w;
register XEvent *evp;
{
    register table_info *tp;
    int new_column_width;
    int need_repaint = FALSE;

    if (!w)
        return;
    if (!(tp = (table_info *) w->info.table))
        return;

    if (tp->doing_column_resizing && tp->colToResize != -1) {
        new_column_width = colResize.markerXpos - colResize.colXpos;
        tp->colWidth = TblAddNumToArray(0, tp->colWidth, tp->colToResize,
                                        new_column_width);
/*      if (tp->colToResize == tp->cursor.col) */
        TblUpdCursorCellInfo(tp);

        if (w->callback)
            TblSendColumnResizeEvent(w, tp->colToResize, new_column_width);
        need_repaint = TRUE;
    } else if (tp->doing_row_marking && w->callback) {
        TblSendSelectionChangeEvent(w);
    }

    TblResetAll(w);

    if (need_repaint)
        PaintTable(w, NULL, FALSE, FALSE);
    return;
}

/***********************************************************************
function: TblResetAll - Reset all row_marking, column_marking...
states and put up appropriate cursor.
*/
static void
TblResetAll(w)
AXWidget_table w;
{
    table_info *tp;

    if (!w || w->type != W_TABLE)
        return;
    if (!(tp = (table_info *) w->info.table))
        return;

    tp->doing_column_resizing = FALSE;
    tp->seeing_button1_motion = FALSE;
    tp->doing_row_marking = FALSE;
    tp->doing_column_marking = FALSE;
    tp->colToResize = -1;
    colResize.markerXpos = 0;
    colResize.colXpos = -1;
    AXDefineAppCursor((AXWidget)w, NULL);
    tp->pointer_shape = ThimPointerCursor;
    return;
}

/***********************************************************************
function: TblMotionNotifyCB - Handle a MotionNotify event in the table
widget.
*/
static void
TblMotionNotifyCB(w, ev)
register AXWidget_table w;
XEvent *ev;
{
    register table_info *tp = (table_info *) w->info.table;
    register int ix;
    register int num_rows_displayed;
    register int x = ev->xmotion.x;
    register int y = ev->xmotion.y;
    int row_idx = -1;
    int col_idx = -1;
    int col;
    int max_rows;
    Cursor pointer = None;

    if (AxButtonDown == Button1)
        tp->seeing_button1_motion = TRUE;
    else
        tp->seeing_button1_motion = FALSE;

    max_rows = AxArraySize(tp->rows);
    if (max_rows == 0)
        return; /* empty table */

    if (!tp->seeing_button1_motion) { /* Just pointer motion */
        if ((x < tp->markerWidth) && (y < tp->winHeight) &&
            (y > tp->headingHeight)) { /* in row marker region. */
            tp->colToResize = -1;
            pointer = ThimPointerCursor;
        } else if ((y < tp->headingHeight) && (x > tp->markerWidth) &&
                   (x < tp->winWidth)) { /* in column heading */
            if (tp->allow_column_resizing) {
                col = TblPointOnColumnBoundary(w, x, y);
                if (col != -1) {
                    tp->colToResize = col;
                    pointer = ThimLeftRightCursor;
                } else {
                    tp->colToResize = -1;
                    pointer = ThimPointerCursor;
                }
            } else {
                tp->colToResize = -1;
                pointer =  ThimPointerCursor;
            }
        } else { /* in cell region or top left corner box */
            tp->colToResize = -1;
            pointer = ThimPointerCursor;
        }

        /* Put up appropriate pointer shape. */
        if (pointer == None)
            pointer = ThimPointerCursor; /* safety net */

        if (tp->pointer_shape != pointer) {
            if (pointer == ThimPointerCursor)
                AXDefineAppCursor((AXWidget)w, NULL);
            else
                AXDefineAppCursor((AXWidget)w, pointer);
            tp->pointer_shape = pointer;
        }
        return;
    }

/* At this point, handle Button1 Motion.  *Only* one of the following
   conditions can be true:
   `doing_column_resizing', `doing_row_marking'. */

    if (tp->doing_column_resizing) {
        /* compute new marker x-pos;
           if new x-pos is ok, lift old marker; put new marker, update
           positions... return */
        if (x <= colResize.colXpos + MIN_COL_DIM)
            return; /* won't let column-width go below `MIN_COL_DIM'. */
        if (x >= tp->winWidth - 4) /* FIXME, hardcoded */
            return; /* won't let marker cross window boundary (for now). */

        /* lift old marker */
        XDrawLine(Dpy, w->window, tp->mark_gc, colResize.markerXpos, 0,
                  colResize.markerXpos, tp->winHeight);
        /* put new marker */
        colResize.markerXpos = x;
        XDrawLine(Dpy, w->window, tp->mark_gc, colResize.markerXpos, 0,
                  colResize.markerXpos, tp->winHeight);
        return;
    }
        
/* Marking rows or columns. */
    if (!tp->doing_row_marking && !tp->doing_column_marking)
        return; /* trap out */
    if (w->value == -1)
        return; /* trap out */

    if (!w->multi_pick)
        return;

/* Marking rows */
    if (tp->doing_row_marking) {
        if (y >= tp->winHeight - 3) {
            if (tp->use_data_request_model)
                return;
            w->armed = XmCR_INCREMENT;
            TableArmed = w;
            TableAutoScroll(w);
            return;
        } else if (y < tp->headingHeight) {
            if (tp->use_data_request_model)
                return;
            w->armed = XmCR_DECREMENT;
            TableArmed = w;
            TableAutoScroll(w);
            return;
        }
        TableArmed = NULL;

        /* find row in which motion is occuring in. */
        for (ix = 0, num_rows_displayed = tp->numLines;
             ix < num_rows_displayed; ix++) {
            int top_y, bottom_y; /* top and bottom `y' of this row */

            top_y = tp->headingHeight + (ix * tp->rowHeight);
            bottom_y = top_y + tp->rowHeight;

            if ((y > top_y) && (y < bottom_y)) {
                row_idx = ix + tp->topRow;
                break;
            }
        }
        if (row_idx == -1)
            return; /* Motion not on any of the visible rows. */

        if (row_idx >= max_rows)
            return; /* won't mark excess rows in the bottom */

        if (w->last_item_selected == row_idx)
            return;

        TblClearExcludingRange(w, w->value, row_idx, ROW_MODE);
        TblAddSelectionRange(w, w->value, row_idx, ROW_MODE);

        w->last_item_selected = row_idx;
        return;
    }

/* Marking columns */
    if (tp->doing_column_marking) {
        if (x >= tp->winWidth) {
            w->armed = XmCR_INCREMENT;
            TableArmed = w;
            TableAutoScroll(w);
            return;
        } else if (x < tp->markerWidth) {
            w->armed = XmCR_DECREMENT;
            TableArmed = w;
            TableAutoScroll(w);
            return;
        }
        TableArmed = NULL;

        /* find column in which motion is occuring in */
        /* TODO */
        if (col_idx == -1)
            return; /* Motion not on any of the visible columns. */

        if (w->last_item_selected == col_idx)
            return;

        TblClearExcludingRange(w, w->value, col_idx, COL_MODE);
        TblAddSelectionRange(w, w->value, col_idx, COL_MODE);
        w->last_item_selected = col_idx;
        return;
    }

    return;
}

/***********************************************************************
function: TblPrintableBuffer - Return TRUE if all characters in buffer are
printable and "valid"; FALSE otherwise.
*/
static int
TblPrintableBuffer(w, wbuf, num_chars)
AXWidget_table w;
wchar *wbuf;
int   num_chars;
{
    register int ix;
    register wchar *wbufp;
    int real_len;

    if (!w || w->type != W_TABLE)
        return(FALSE);

    for (ix = 0, wbufp = wbuf, real_len = 0; ix < num_chars; ix++) {
        if (!IS_NGC(*wbufp)) {
            /* char is printable */
            if (w->validChars && !strchr(w->validChars, *wbufp)) {
                /* invalid char */
                RingTheBell();
                return(FALSE);
            }
            wbufp++;
            real_len++;
        }
    }
    if (!real_len)
        return(FALSE);

    return(TRUE);
}

extern KeySym AxEuroDollarKeysym;							/* method 1 */
extern int AxEuroDollarModifiers;							/* method 1 */
extern int AxEuroDollarKeyState, AxEuroDollarKeyCode ;		/* method 2 */

/***********************************************************************
function: TblKeyPressCB - Handle a KeyPress event in the table widget.
Return 1 if key-press event was consumed, 0 otherwise.
*/
int
TblKeyPressCB(w, ev, in_menu_bar)
register AXWidget_table w;
register XEvent *ev;
int in_menu_bar;
{
    int num_chars = 0;
    wchar *wbuf, buf[100];
    KeySym ksym;
    table_info *tp;
    int changed = FALSE;
    int event_consumed = FALSE;
	int euro = FALSE;
	
    if (!w || !ev)
        return(0);
    if (ev->xany.type != KeyPress)
        return(0);
    if (w->type != W_TABLE)
        return(0);
    
	if (AxEuroDollarKeyCode && (AxEuroDollarKeyCode == ev->xkey.keycode) && 
								(AxEuroDollarKeyState == ev->xkey.state)) {
		/* we have found the euro dollar key sequence, return back EuroDollar
			ascii */
		buf[0] = EURO_SYMBOL_8BIT_VALUE;
		buf[1] = 0;
		wbuf = buf;
		num_chars = 1;
		ksym = 0;
		euro = TRUE;
	} else {
    	num_chars = AXLookupString16(ev, &wbuf, &ksym);
	}

	/* High-level Euro keysym - hardware-independent */
	if (AxEuroDollarKeysym && (AxEuroDollarKeysym == ksym)) {
		int mods = AxDecodeModifiers(ev->xkey.state);
		if (mods == AxEuroDollarModifiers) {
			buf[0] = EURO_SYMBOL_8BIT_VALUE; 
			buf[1] = 0;
			wbuf = buf;
			num_chars = 1;
			ksym = 0;
			euro = TRUE;
		}
	}

	/* If this keystroke is or might be in the menubar, and it wasn't
		a Euro, reject it */
	if (in_menu_bar && !euro)
    	return(FALSE);

    if ((num_chars == 0) && ((ksym == 0) || AxKeysymIsModifier(ksym)))
      return(0); /* ignore shift/control... */

    tp = (table_info *) w->info.table;
    if (!tp)
        return(0);
    if ( tp->rowcolcreate == FALSE )
    {   
        switch (ksym) {
        case XK_Return:
            if (w->minorEventNotify) {
                TblSendKeyPressEvent(w, ELFK_RETURN, FALSE, 0, 0,
                                     tp->cursor.row, tp->cursor.col);
                event_consumed = TRUE;
            } else
                event_consumed = FALSE;
            break;
        case XK_Down:
            if (w->minorEventNotify)
                TblSendKeyPressEvent(w, ELFK_DOWN, FALSE, 0, 0,
                                     tp->cursor.row, tp->cursor.col);
            else
                TblForwardRow(w);
            event_consumed = TRUE;
            break;
        case XK_Up:
            if (w->minorEventNotify)
                TblSendKeyPressEvent(w, ELFK_UP, FALSE, 0, 0,
                                     tp->cursor.row, tp->cursor.col);
            else
                TblPreviousRow(w);
            event_consumed = TRUE;
            break;
        case XK_Tab:
            /* ThimIsShifted() does'nt say it's shifted.  Modifiers are
               set only on ButtonPress!! */
            if (ev->xkey.state & ShiftMask) {
                if (w->minorEventNotify)
                    TblSendKeyPressEvent(w, ELFK_BACKTAB, FALSE, 0, 0,
                                         tp->cursor.row, tp->cursor.col);
                else
                    TblBackwardColumn(w);
            } else {
                if (w->minorEventNotify)
                    TblSendKeyPressEvent(w, ELFK_TAB, FALSE, 0, 0,
                                         tp->cursor.row, tp->cursor.col);
                else
                    TblForwardColumn(w);
            }
            event_consumed = TRUE;
            break;
        case XK_Left:
            if (w->minorEventNotify)
                TblSendKeyPressEvent(w, ELFK_LEFT, FALSE, 0, 0,
                                     tp->cursor.row, tp->cursor.col);
            else
                TblBackwardChar(w);
            event_consumed = TRUE;
            break;
        case XK_Right:
            if (w->minorEventNotify)
                TblSendKeyPressEvent(w, ELFK_RIGHT, FALSE, 0, 0,
                                     tp->cursor.row, tp->cursor.col);
            else
                TblForwardChar(w);
            event_consumed = TRUE;
            break;
        case XK_BackSpace:
            if (w->minorEventNotify)
                TblSendKeyPressEvent(w, ELFK_DELETE, FALSE, 0, 0,
                                     tp->cursor.row, tp->cursor.col);
            else {
                TblBackwardCharErase(w);
                TblSendTypingEvent(w, tp->cursor.row, tp->cursor.col);
            }
            event_consumed = TRUE;
            break;
        case XK_Delete:
            if (w->minorEventNotify)
                TblSendKeyPressEvent(w, ELFK_DELETE, FALSE, 0, 0,
                                     tp->cursor.row, tp->cursor.col);
            else {
                TblEraseChar(w);
                TblSendTypingEvent(w, tp->cursor.row, tp->cursor.col);
            }
            event_consumed = TRUE;
            break;
        case XK_Linefeed:
            if (w->minorEventNotify)
                TblSendKeyPressEvent(w, ELFK_LINEFEED, FALSE, 0, 0,
                                     tp->cursor.row, tp->cursor.col);
            event_consumed = TRUE;
            break;
        case XK_Escape:
            if (w->minorEventNotify) {
                TblSendKeyPressEvent(w, ELFK_ESCAPE, FALSE, 0, 0,
                                     tp->cursor.row, tp->cursor.col);
                event_consumed = TRUE;
            } else
                event_consumed = FALSE;
            break;
        case XK_Clear:
        case XK_Pause:
        case XK_Num_Lock:
        case XK_Shift_L:
        case XK_Shift_R:
        case XK_Control_L:
        case XK_Control_R:
        case XK_Caps_Lock:
        case XK_Shift_Lock:
        case XK_Meta_L:
        case XK_Meta_R:
        case XK_Alt_L:
        case XK_Alt_R:
        case XK_Super_L:
        case XK_Super_R:
        case XK_Hyper_L:
        case XK_Hyper_R:
            event_consumed = FALSE; /* toss these up to parent. */
            break;
        default:
            if ((ev->xkey.state & ControlMask) &&
					!((num_chars == 1) && (wbuf[0] == EURO_SYMBOL_8BIT_VALUE)) )
                return(FALSE); /* dump back controlled keys */
            if (!TblPrintableBuffer(w, wbuf, num_chars)) {
                changed = FALSE;
            } else {
                if (w->minorEventNotify) {
                    TblSendKeyPressEvent(w, ELFK_PRINTABLE, TRUE, wbuf, 
                                         num_chars, tp->cursor.row, 
                                         tp->cursor.col);
                    changed = TRUE;
                } else {
                    changed = TblInsertText(w, wbuf, num_chars);
                    TblSendTypingEvent(w, tp->cursor.row, tp->cursor.col);
                }
            }
            
            if (changed)
                event_consumed = TRUE;
            else
                event_consumed = FALSE;
            break;
        }
    }
    else  { /* if in presentation mode ie.,  rowcolcreate*/
        switch (ksym) {
        case XK_Return:
            if (w->minorEventNotify) {
                TblSendKeyPressEvent(w, ELFK_RETURN, FALSE, 0, 0,
                                     tp->cursor.row, tp->cursor.col);
                event_consumed = TRUE;
            } else
                event_consumed = FALSE;
            break;
        case XK_Down:
			TblForwardRow(w);
            event_consumed = TRUE;
            break;
        case XK_Up:
            TblPreviousRow(w);
            event_consumed = TRUE;
            break;
        case XK_Tab:
            /* ThimIsShifted() does'nt say it's shifted.  Modifiers are
               set only on ButtonPress!! */
            if (ev->xkey.state & ShiftMask) {
                TblSendKeyPressEvent(w, ELFK_BACKTAB, FALSE, 0, 0,
                                     tp->cursor.row, tp->cursor.col);
                TblBackwardColumn(w);
            } else {
                TblSendKeyPressEvent(w, ELFK_TAB, FALSE, 0, 0,
                                     tp->cursor.row, tp->cursor.col);
                TblForwardColumn(w);
            }
            event_consumed = TRUE;
            break;
        case XK_Left:
            TblBackwardChar(w);
            event_consumed = TRUE;
            break;
        case XK_Right:
            TblForwardChar(w);
            event_consumed = TRUE;
            break;
        case XK_BackSpace:
            TblBackwardCharErase(w);
            event_consumed = TRUE;
            break;
        case XK_Delete:
            TblEraseChar(w);
            event_consumed = TRUE;
            break;
        case XK_Linefeed:
            TblSendKeyPressEvent(w, ELFK_LINEFEED, FALSE, 0, 0,
                                 tp->cursor.row, tp->cursor.col);
            TblEraseChar(w);
            event_consumed = TRUE;
            break;
        case XK_Escape:
            if (w->minorEventNotify) {
                TblSendKeyPressEvent(w, ELFK_ESCAPE, FALSE, 0, 0,
                                     tp->cursor.row, tp->cursor.col);
                event_consumed = TRUE;
            } else
                event_consumed = FALSE;
            break;
        case XK_Clear:
        case XK_Pause:
        case XK_Num_Lock:
        case XK_Shift_L:
        case XK_Shift_R:
        case XK_Control_L:
        case XK_Control_R:
        case XK_Caps_Lock:
        case XK_Shift_Lock:
        case XK_Meta_L:
        case XK_Meta_R:
        case XK_Alt_L:
        case XK_Alt_R:
        case XK_Super_L:
        case XK_Super_R:
        case XK_Hyper_L:
        case XK_Hyper_R:
            event_consumed = FALSE; /* toss these up to parent. */
            break;
        default:
            if ((ev->xkey.state & ControlMask) &&
					!((num_chars == 1) && (wbuf[0] == EURO_SYMBOL_8BIT_VALUE)) )
                return(FALSE); /* dump back controlled keys */
            if (!TblPrintableBuffer(w, wbuf, num_chars)) {
                changed = FALSE;
            } else {
                TblSendKeyPressEvent(w, ELFK_PRINTABLE, TRUE, wbuf, 
                                     num_chars, tp->cursor.row, 
                                     tp->cursor.col);
                changed = TblInsertText(w, wbuf, num_chars);
            }           
            TblTransferCursorCellData(w);
            if (changed)
                event_consumed = TRUE;
            else
                event_consumed = FALSE;
            break;

        }               
    }

    return(event_consumed);
}

/***********************************************************************
function: TblKeyReleaseCB - Handle a KeyRelease event in the table widget.
*/
static void
TblKeyReleaseCB(w, evp)
register AXWidget_table w;
register XEvent *evp;
{
    return;
}

/***********************************************************************
function: TblHandleElfKeyid - Handle default action for a ELFK_XXX key.
*/
static void
TblHandleElfKeyid(w, elfkeyid)
AXWidget_table w;
int elfkeyid;
{
    switch (elfkeyid) {
    case ELFK_RETURN:
        TblForwardColumn(w);
        break;
    case ELFK_DOWN:
        TblForwardRow(w);
        break;
    case ELFK_UP:
        TblPreviousRow(w);
        break;
    case ELFK_TAB:
        TblForwardColumn(w);
        break;
    case ELFK_BACKTAB:
        TblBackwardColumn(w);
        break;
    case ELFK_LEFT:
        TblBackwardChar(w);
        break;
    case ELFK_RIGHT:
        TblForwardChar(w);
        break;
    case ELFK_LINEFEED: /* FIXME, bizzare */
/*      TblEraseChar(w); */
        break;
    case ELFK_DELETE:
        TblBackwardCharErase(w);
        break;
    case ELFK_ESCAPE:
        break;
    default:
        RingTheBell();
        break;
    }
    return;
}

/***********************************************************************
function: PaintTableBorder - Draw a border around the whole table.
*/
void
PaintTableBorder(w)
register AXWidget_table w;
{
    int thickness, style;

    if (!w)
        return; /* invalid args */

    style = BEVEL_NONE;
    if (w->thickness < 0) {
        thickness = -w->thickness;
        style = BEVEL_IN;
    }
    else if (w->thickness) {
        thickness = w->thickness;
        style = BEVEL_OUT;
    }
    if (style != BEVEL_NONE) {
        AxDrawBevel((AXWidget)w, WIN_OF(w), w->xpos,w->ypos,w->width,w->height,
                    thickness, style);
    }
    return;
}

/***********************************************************************
function: TblDrawCornerBox - Draw top left corner box.
*/
static void
TblDrawCornerBox(w)
register AXWidget_table w;
{
    int thisWidth;
    int thisHeight;
    int xpos;
    int ypos;
    register table_info *tp;
    int ix, num_pixmaps;
    mbchar *pixmap_name;
    pixcache_t *pixent;
    int clips_were_on = FALSE;
    int pix_separator = 2; /* 2 pixel separation */

    if (!w)
        return;
    tp = (table_info *) w->info.table;
    thisWidth  = tp->markerWidth;
    thisHeight = tp->headingHeight;

    if ((thisWidth <= 0) || (thisHeight <= 0))
        return;

    xpos = 0;
    /* top left corner box */
    AxDrawBevel((AXWidget)w, w->window, xpos, 0, thisWidth, thisHeight,
				TBEVEL, BEVEL_OUT);
    xmColorFillRectangle((AXWidget)w, w->window, BackGC, xpos + TBEVEL, TBEVEL,
                         thisWidth - 2 * TBEVEL, thisHeight - 2 * TBEVEL);
    if (tp->corner_pixmaps) {
        num_pixmaps = AxArraySize(tp->corner_pixmaps);
        xpos = 2;
        ypos = 0;
        for (ix = 0; ix < num_pixmaps; ix++) {
            pixmap_name = (mbchar *) AxStrFromArray(tp->corner_pixmaps, ix);
            if (!pixmap_name)
                continue;
            if (!MBstrlen(pixmap_name))
                continue;
            pixent = Pixmapper(pixmap_name);
            if (!pixent)
                continue;
            clips_were_on = FALSE;
            if (pixent->clip_mask) {
                if (tblSaveClip.clips_on) {
                    TblUnsetClipAllGCs(w);
                    clips_were_on = TRUE;
                }
                XSetClipOrigin(Dpy, ForeGC, xpos, ypos + 2);
                XSetClipMask(Dpy, ForeGC, pixent->clip_mask);
            }
            XCopyArea(Dpy, pixent->pix, w->window, ForeGC, 0, 0,
                      pixent->width, pixent->height, xpos, ypos + 2);
            if (pixent->clip_mask)
                XSetClipMask(Dpy, ForeGC, None);
            if (clips_were_on)
                TblRestoreSavedClips(w);
            xpos += (pixent->width + pix_separator);
        }
    }
    return;
}

/***********************************************************************
function: TblDrawRowMarkers - Draw row markers
*/
static void
TblDrawRowMarkers(w)
register AXWidget_table w;
{
    register table_info *wt;
    register int rowNum;
    register int thisWidth;
    register int thisHeight;
    register int xpos;
    register int ypos;
    int max_rows;

    if (!w)
        return;
    wt = (table_info *) w->info.table;
    ypos = wt->headingHeight;
    xpos = 0;
    thisHeight = wt->rowHeight;
    thisWidth = wt->markerWidth;
    max_rows = AxArraySize(wt->rows);
    for (rowNum = wt->topRow + 1; ypos < wt->winHeight; rowNum++) {
        TblDrawRowMarkerShadow(w, rowNum - 1, FALSE);
        if ((rowNum <= max_rows) /* && (wt->showRowNumbers) */) {
            TblDrawRowMarkerContents(w, rowNum - 1);
        }
        ypos += wt->rowHeight;
    }
    if (w->flags & WFLG_IN_EDIT) {
        int thickness = w->thickness;

        if (thickness < 0)
            thickness = -thickness;
        XCopyArea(Dpy, w->window, w->parent->window, ForeGC, 0, 0,
                  wt->winWidth, wt->winHeight,
                  w->xpos + thickness, w->ypos + thickness);
    }
    return;
}

/***********************************************************************
function: TblDrawRowMarkersRegion - Draw all visible row markers starting
at START_ROW and ending at END_ROW (both inclusive)
Note: START_ROW must be <= END_ROW
*/
static void
TblDrawRowMarkersRegion(w, start_row, end_row)
AXWidget_table w;
int start_row;
int end_row;
{
    table_info *tp;
    int lower_bound;
    int upper_bound;
    int ix;
    int ypos;

    if (!w)
        return;
    if (!(tp = (table_info *) w->info.table))
        return;

    if (tp->markerWidth <= 0)
        return;

/* Compute lower bound */
    if (start_row == -1)
        lower_bound = tp->topRow;
    else {
        lower_bound = start_row;
        if (start_row < tp->topRow)
            lower_bound = tp->topRow;
    }

/* Compute upper bound */
    if (end_row == -1)
        upper_bound = tp->topRow + tp->numLines;
    else
        upper_bound = MIN(end_row, (tp->topRow + tp->numLines));

/* Bound check */
    if (lower_bound > upper_bound) { /* bad args */
        TblDrawRowMarkers(w); /* sanity */
        return;
    }

/* Draw row markers */
    for (ix = tp->topRow, ypos = tp->headingHeight;
         ypos < tp->winHeight; ix++) {
        if ((ix >= lower_bound) && (ix <= upper_bound)) {
            TblDrawRowMarkerShadow(w, ix, FALSE);
            TblDrawRowMarkerContents(w, ix);
        }
        if (ix >= upper_bound)
            break;
        ypos += tp->rowHeight;
    }

/* Dlg editor */
    if (w->flags & WFLG_IN_EDIT) {
        int thickness = w->thickness;

        if (thickness < 0)
            thickness = -thickness;
        XCopyArea(Dpy, w->window, w->parent->window, ForeGC, 0, 0,
                  tp->winWidth, tp->winHeight,
                  w->xpos + thickness, w->ypos + thickness);
    }

    return;
}

/***********************************************************************
function: TblDrawRowMarkerContents - Draw row marker contents.
FIXME, rowHeight assumption.
*/
static void
TblDrawRowMarkerContents(w, row_num)
register AXWidget_table w;
register int row_num;
{
    table_info *tp;
    int text_x, itmp;
    int text_width, avail_width;
    int ypos;
    data_ptr pixmapList = (data_ptr) NULL;
    data_ptr stringList = (data_ptr) NULL;
    mbchar *pixmapName, *stringName;
    pixcache_t *pixent;
    int pix_separator = 2; /* 2 pixel separation */
    char text_string[STRNORMAL + 1];
    int  ix, copy_wid, copy_ht;
    int num_pixmaps;
    int clips_were_on = FALSE;
    XFontStruct *saved_font;

    if (!w)
        return;
    if (!(tp = (table_info *) w->info.table))
        return;
    if ((row_num < 0) || (row_num < tp->topRow))
        return;

    if (row_num == 0)
        ypos = tp->headingHeight;
    else
        ypos = tp->headingHeight + ((row_num - tp->topRow) * tp->rowHeight);

    if (tp->showRowNumbers && (row_num < AxArraySize(tp->rows))) {
        sprintf(text_string, "%d", row_num + 1); /* make 1 based */
        avail_width = tp->markerWidth - 4;
        text_width  = BasicStrlen(text_string);
        if (text_width < avail_width) {
            text_x = (((avail_width / 2) + 2) - (text_width / 2));
            if ((text_x < 2) || (text_x >= (avail_width + 2)))
                text_x = 2;
        } else
            text_x = 2;

        saved_font = w->font;
        w->font = tp->headingFont;
        xmDrawString((AXWidget)w, w->window, ForeGC, text_x,
                     ypos + tp->heading_char_above + 3, text_string,
                     strlen(text_string), tp->headingShadow,
                     tp->headingColor, w->widgetColor);
        w->font = saved_font;
    }
    else if (tp->marker_pixmaps) {
        text_x = 2;
        pixmapList = ReadArray(AX(tp->marker_pixmaps), row_num);
        if (pixmapList && ((num_pixmaps = AxArraySize(pixmapList)) > 0)) {
            for (ix = 0; ix < num_pixmaps; ix++) {
                pixmapName = (mbchar *) AxStrFromArray(pixmapList, ix);
                if (!pixmapName)
                    continue;
                if (!MBstrlen(pixmapName))
                    continue;

                if (tblSaveClip.clips_on) {
                    /* This is necessary because the pixmapper uses
                       some of the (global) gcs used by the table.
                       *Need* to unset clips (on all gcs for safety).
                       The pixmapper will return garbage if clips are
                       set on any of the GCs it uses. */
/*                  TblUnsetClipAllGCs(w); */
                    pixent = Pixmapper(pixmapName);
/*                  TblRestoreSavedClips(w); */
                } else
                    pixent = Pixmapper(pixmapName);
                if (!pixent)
                    continue;

                clips_were_on = FALSE;
                if (pixent->clip_mask) {
                    if (tblSaveClip.clips_on) {
                        TblUnsetClipAllGCs(w);
                        clips_were_on = TRUE;
                    }
                    XSetClipOrigin(Dpy, ForeGC, text_x, ypos + 2);
                    XSetClipMask(Dpy, ForeGC, pixent->clip_mask);
                }
                
                /* constrain copy width, height to fit within marker flat
                 */
                copy_wid = pixent->width;
                if ((itmp = (text_x + copy_wid + 2 - tp->markerWidth)) > 0) {
                    copy_wid -= itmp;
                }
                copy_ht = pixent->height;
                if ((itmp = (copy_ht + 4 - tp->rowHeight)) > 0) {
                    copy_ht -= itmp;
                }

                if ((copy_wid > 0) && (copy_wid <= pixent->width) &&
                    (copy_ht > 0) && (copy_ht <= pixent->height)) {
                    
                    /* draw the marker pixmap
                     */
                    XCopyArea(Dpy, pixent->pix, w->window, ForeGC, 0, 0,
                              copy_wid, copy_ht, text_x, ypos + 2);
                }
                
                if (pixent->clip_mask)
                    XSetClipMask(Dpy, ForeGC, None);
                if (clips_were_on)
                    TblRestoreSavedClips(w);

                text_x += (pixent->width + pix_separator);
            }
        }
    }
    else if (tp->marker_strings) {
        text_x = 2;
        stringName =(mbchar *)AxStrFromArray(tp->marker_strings, row_num);
        if ( stringName && MBstrlen(stringName) ) {
            avail_width = tp->markerWidth - 4;
            text_width  = BasicStrlen(stringName);
            if (text_width < avail_width) {
                text_x = (((avail_width / 2) + 2) - (text_width / 2));
                if ((text_x < 2) || (text_x >= (avail_width + 2)))
                    text_x = 2;
            } else
                text_x = 2;
            
            saved_font = w->font;
            w->font = tp->headingFont;
            xmDrawString((AXWidget)w, w->window, ForeGC, text_x,
                         ypos + tp->heading_char_above + 3, stringName,
                         strlen(stringName), tp->headingShadow,
                         tp->headingColor, w->widgetColor);
            w->font = saved_font;
        }
    }
    return;
}

/***********************************************************************
function: TblDrawHeadingContents - Draw text in heading of column COL_NUM.
*/
static void
TblDrawHeadingContents(w, col_num)
AXWidget_table w;
int col_num;
{
    return; /* TODO */
}

/***********************************************************************
function: TblDrawEditCellBox - Draw a box around a cell.
*/
static void
TblDrawEditCellBox(w, row, col)
AXWidget_table w;
int row;
int col;
{
    return;
}

/***********************************************************************
function: TblDrawRowMarkerShadow - Draw shadow around the row-marker
on row ROW_NUM.  If PUSHED is TRUE, draw the "pushed" shadow; "normal"
shadow otherwise.
*/
static void
TblDrawRowMarkerShadow(w, row_num, pushed)
register AXWidget_table w;
register int row_num;
int pushed;
{
    int x, y;
    int width, height;
    table_info *tp;

    if (!w)
        return;
    if (!(tp = (table_info *) w->info.table))
        return;
    if ((row_num < 0) || (row_num < tp->topRow))
        return;
    if (tp->markerWidth == 0)
        return; /* no marker */

    height = tp->rowHeight;
    width  = tp->markerWidth;
    x = 0;
/*  x = tp->markerWidth + 2; */

    if (row_num <= 0)
        y = tp->headingHeight;
    else
        y = tp->headingHeight + ((row_num - tp->topRow) * tp->rowHeight);

    if (pushed == TRUE) {
        AxDrawBevel((AXWidget)w,w->window, x,y,width,height, TBEVEL,BEVEL_IN);
        if (UseWidgetColors)
            XFillRectangle(Dpy, w->window, RecessGC, x + TBEVEL, y + TBEVEL,
                           width - 2 * TBEVEL, height - 2 * TBEVEL);
    } else {
        AxDrawBevel((AXWidget)w,w->window, x,y,width,height, TBEVEL,BEVEL_OUT);
        xmColorFillRectangle((AXWidget)w,w->window, BackGC, x+TBEVEL, y+TBEVEL,
                             width - 2 * TBEVEL, height - 2 * TBEVEL);
    }
    return;
}

/***********************************************************************
function: TblDrawHeadingShadow - Draw shadow around the heading on
column COL_NUM.  If PUSHED is TRUE, draw the "pushed" shadow; "normal"
shadow otherwise
*/
static void
TblDrawHeadingShadow(w, col_num, pushed)
AXWidget_table w;
int col_num;
int pushed;
{
    return; /* TODO */
}

/***********************************************************************
function: TblDrawCellText - Draw text in cell ROW_NUM,COL_NUM.
Compute CELL_XPOS, CELL_YPOS if they are -1.
*/
static void
TblDrawCellText(w, row_num, col_num, cell_xpos, cell_ypos, clip)
AXWidget_table w;
int row_num;
int col_num;
int cell_xpos;
int cell_ypos;
r32 *clip;              /* cell clip dimensions (use if text long enough) */
{
    table_info *tp;
    wchar *wstr;
    wchar *wstr_p;
    int text_len;
    int visible_len;
    int ix, right_ink;
    int xleft, xright;
    int clip_cell = FALSE;
    XCharStruct *xc;

    if (!w || w->type != W_TABLE)
        return;

    if (!(tp = (table_info *) w->info.table))
        return;
    
    if (AxIntFromArray(tp->colWidth, col_num) <= 0)
        return;                /* no column width */

    if (cell_xpos == -1) {
        cell_xpos = XPosOfCell(tp, row_num, col_num);
        if (cell_xpos == -1)
            return; /* cell not visible. */
    }
    if (cell_ypos == -1) {
        cell_ypos = YPosOfCell(tp, row_num, col_num);
        if (cell_ypos == -1)
            return; /* cell not visible */
    }

    wstr = TblDupCellString(w, row_num, col_num);
    if (!wstr)
        return;

    if (w->useWorkColors) {
        AxSetWidgetWorkColors((AXWidget)w);
    }
    
	wstr_p = wstr;
    text_len = Wstrlen(wstr);
    xright = cell_xpos + AxIntFromArray(tp->colWidth, col_num) - 1;
    xleft  = cell_xpos + 3;
    
#if BUILD_WIDE
    /* we're using an escapement-only measurement for wide text,
     * so we're uncertain whether the ink really fits; always clip.
     */
    clip_cell = TRUE;
    right_ink = 0;
#endif

    /* measure how many characters will fit in the cell
     */
    for (ix = 0, visible_len = 0; ix < text_len; ix++) {
#if BUILD_WIDE
        xleft += AXWTextWidth(tp->font, wstr_p, 1);
#else
        xc = getCharStruct((int) *wstr_p, tp->font);
        xleft += xc->width;
        if ((int) xc->rbearing > (int) xc->width) {
            right_ink = (int) xc->rbearing - (int) xc->width;
        }
        else {
            right_ink = 0;
        }
        if (!ix && (xc->lbearing < 0)) {
            clip_cell = TRUE;    /* 0th char juts on the left */
        }
#endif
        if (xleft + right_ink > xright) {
            clip_cell = TRUE;
            if (visible_len < text_len)
                visible_len++;    /* overshoot just a bit (will clip) */
            break;
        }
        wstr_p++;
        visible_len++;
    }

    /* draw */
    if (TblRowIsSelected(tp, row_num) || TblColIsSelected(tp, col_num)) {
        r32 *cr = (r32 *) NULL;
        
        if (clip_cell) {
            cr = clip;
        }
        TblDrawHilitText(w, cell_xpos + 3,
                         cell_ypos + tp->char_above + 2 + 1,
                         wstr, visible_len, cr);
    }
    else {
        if (clip_cell && clip) {
            TblClip(ForeGC, clip->origin.x, clip->origin.y,
                    clip->width, clip->height);
        }
        
        xmDrawString16((AXWidget)w, w->window, ForeGC, cell_xpos + 3,
                       cell_ypos + tp->char_above + 2 + 1, wstr,
                       visible_len, w->textShadow, w->textColor,
                       (wcolor_t *) NULL);
        
        if (clip_cell && clip) {
            TblRestoreClip(ForeGC);
        }
    }
    
    if (w->useWorkColors) {
        AxUnsetWidgetWorkColors((AXWidget)w);
    }
	
    if (wstr)
        free(wstr);
    return;
}

/***********************************************************************
function: TblDrawCursorCellText - Draw text contained in the cursor-cell.
*/
static void
TblDrawCursorCellText(w)
register AXWidget_table w;
{
    register table_info *tp;
    int xpos;
    int ypos;
    wchar *wstr;
    wchar *wstr_p;
    int str_len;
    int text_len;
    int char_width;
    int xmax;
#if BUILD_ORIENT
    int under_start, under_len;
    int highlight_start, highlight_char_len, highlight_len;
    int ix;
    wchar *wcharP;
    AXWidget_main main_w;
#endif
    int clip_cell = FALSE;
    int right_ink;
    XCharStruct *xc;

    if (!w)
        return;
    if (w->type != W_TABLE)
        return;
    if (!(tp = (table_info *) w->info.table))
        return;

    if ((tp->cursor.cell_xpos < tp->markerWidth) ||
        (tp->cursor.cell_ypos < tp->headingHeight))
        return; /* absurd value.  won't see */

/* Clear cell area */
    if (w->useWorkColors) {
        AxSetWidgetWorkColors((AXWidget)w);
    }
    xpos = tp->cursor.cell_xpos;
    ypos = tp->cursor.cell_ypos;
    XFillRectangle(Dpy, w->window, BackGC, tp->cursor.cell_xpos + 1,
                   tp->cursor.cell_ypos + 1,
                   tp->cursor.cell_width - 1,
                   tp->cursor.cell_height - 1);
    tp->cursor.text_cursor_on = FALSE;  /* just got erased */

/* Empty cell */
    if (!tp->cursor.wtext) {
		if (w->useWorkColors) {
			AxUnsetWidgetWorkColors((AXWidget)w);
		}
        return; /* nothing to draw */
	}

/* Compute `text_len' (number of chars to draw) */
    wstr = tp->cursor.wtext;
    str_len = tp->cursor.text_len;
    if (tp->cursor.left_char_pos >= str_len) {
        tp->cursor.left_char_pos = str_len - 1;
    }
    wstr_p = (wchar *) &wstr[tp->cursor.left_char_pos];
    xmax = tp->cursor.cell_xpos + tp->cursor.cell_width;
    
#if BUILD_WIDE
    /* we're using an escapement-only measurement for wide text,
     * so we're uncertain whether the ink really fits; always clip.
     */
    clip_cell = TRUE;
    right_ink = 0;
#endif

    for (xpos = xpos + 3, text_len = 0;
         (text_len < tp->cursor.text_len) && *wstr_p; wstr_p++) {

#if BUILD_WIDE
        char_width = AXWTextWidth(tp->font, wstr_p, 1);
#else
        xc = getCharStruct((int) *wstr_p, tp->font);
        char_width = xc->width;
        if ((int) xc->rbearing > (int) xc->width) {
            right_ink = (int) xc->rbearing - (int) xc->width;
        }
        else {
            right_ink = 0;
        }
        if (!text_len && (xc->lbearing < 0)) {
            clip_cell = TRUE;    /* 0th char juts on the left */
        }
#endif
        if ((xpos + char_width + right_ink) >= xmax) {
            clip_cell = TRUE;
            if (text_len < tp->cursor.text_len)
                text_len++;    /* overshoot just a bit (will clip) */
            break;
        }
        text_len++;
        xpos += char_width;
    }

/* Draw cell contents. */
    xpos = tp->cursor.cell_xpos;
    wstr_p = (wchar *) &wstr[tp->cursor.left_char_pos];
    
    if (TblRowIsSelected(tp, tp->cursor.row) ||
        TblColIsSelected(tp, tp->cursor.col)) {

        r32 cellrect, *cr;
        
        cr = (r32 *) NULL;
        if (clip_cell) {
            cellrect.origin.x = tp->cursor.cell_xpos + 1;
            cellrect.origin.y = tp->cursor.cell_ypos + 1;
            cellrect.width = AxIntFromArray(tp->colWidth, tp->cursor.col) - 1;
            cellrect.height = tp->rowHeight - 1;
            cr = &cellrect;
        }

        TblDrawHilitText(w, xpos + 3,
                         ypos + tp->char_above + 2 + 1,
                         wstr_p, text_len, cr);
    }
    else {
        if (clip_cell) {
            TblClip(ForeGC, tp->cursor.cell_xpos + 1,
                    tp->cursor.cell_ypos + 1,
                    AxIntFromArray(tp->colWidth, tp->cursor.col) - 1,
                    tp->rowHeight - 1);
        }
        
        xmDrawString16((AXWidget)w, w->window, ForeGC, xpos + 3,
                       ypos + tp->char_above + 2 + 1, wstr_p, text_len,
                       w->textShadow, w->textColor, (wcolor_t *) NULL);
        
        if (clip_cell) {
            TblRestoreClip(ForeGC);
        }
    }
    
#if BUILD_ORIENT           /*  {   */
    /* underline and highlight any active Asian input area */
    main_w = TheMainWidget( w );
    if ((main_w->conversion_start >= 0) || (main_w->conv_phrase_start >= 0))
       {
       /* some active input in progress */
       under_start = highlight_start = -1;
       under_len = highlight_char_len = highlight_len = 0;

       xpos = tp->cursor.cell_xpos + 3;
       wcharP = tp->cursor.wtext + tp->cursor.left_char_pos;
       ix = tp->cursor.left_char_pos-1;
       while (*wcharP != (wchar)0) 
            {
            ix++;
            if ( (main_w->conversion_start <= ix) && (under_start < 0) )
                under_start = xpos;
            if ( (main_w->conv_phrase_start <= ix) && (highlight_start < 0) )
                highlight_start = xpos;
 
            char_width = AXWTextWidth(tp->font, wcharP, 1);
            if ( (xpos + char_width) > xmax )
                break;
            xpos += char_width;
 
            if ( (ix >= main_w->conversion_start) && 
                 (ix <= main_w->conversion_end))
                under_len = under_len + char_width;
            if ( (ix >= main_w->conv_phrase_start) && 
                 (ix <= main_w->conv_phrase_end))
                {
                highlight_char_len++;
                highlight_len = highlight_len + char_width;
                }
            wcharP++;
            }

       if ( under_start > 0 )
            {
            /* underline */
            ypos = tp->cursor.cell_ypos + tp->char_above + 3;
            XDrawLine(Dpy, w->window, ForeGC, under_start, ypos, 
                      (under_start+under_len), ypos );
            }
       if ( highlight_start > 0 )
            {
            /* we are to highlight a conversion zone, so fill it to */
            /*   foreground, then draw the text in background. */
            XFillRectangle(Dpy, w->window, ForeGC,
                  highlight_start, tp->cursor.cell_ypos + 2, 
                  highlight_len, tp->cursor.cell_height - 4 );
            wcharP = &tp->cursor.wtext[ main_w->conv_phrase_start ];
            
            /* keep font, but drop color and shadow during this */
            xmDrawString16(w, w->window, BackGC, highlight_start, ypos,
                           wcharP, highlight_char_len, NULL, NULL, NULL);
                  
            }
       }
#endif      /*  }  BUILD_ORIENT */

    if (w->useWorkColors) {
        AxUnsetWidgetWorkColors((AXWidget)w);
    }
	
    return;
}

/***********************************************************************
function: TblDrawColumnHeadings - Draw column headings.
*/
static void
TblDrawColumnHeadings(w)
register AXWidget_table w;
{
    int xpos;
    int numCols;
    int colNum;
    int thisHeight, thisWidth;
    mbchar *mbstr;
    table_info *wt;
    int text_x;

    if (!w)
        return;
    if (!(wt = (table_info *) w->info.table))
        return;
    xpos = wt->markerWidth;
    numCols = AxArraySize(wt->colText);
    thisHeight = wt->headingHeight;
    for (colNum = wt->leftCol;
         (xpos < wt->winWidth) && (colNum < numCols); colNum++) {
        thisWidth = AxIntFromArray(wt->colWidth, colNum);
        AxDrawBevel((AXWidget)w, w->window, xpos, 0, thisWidth, thisHeight,
                    TBEVEL, BEVEL_OUT);
        xmColorFillRectangle((AXWidget)w,w->window, BackGC, xpos+TBEVEL,TBEVEL,
                             thisWidth - 2 * TBEVEL, thisHeight - 2 * TBEVEL);
        mbstr = (mbchar *) AxStrFromArray(wt->colText, colNum);
        text_x = xpos + 3;     /* default left justified position */
        TblDrawHeading(w, text_x, wt->heading_char_above + 5, mbstr);
        xpos += thisWidth;
    }
    if (xpos < wt->winWidth) {  /* blank button excess area */
        thisWidth = wt->winWidth - xpos;
        AxDrawBevel((AXWidget)w, w->window, xpos, 0, thisWidth, thisHeight,
                    TBEVEL, BEVEL_OUT);
        xmColorFillRectangle((AXWidget)w,w->window, BackGC, xpos+TBEVEL,TBEVEL,
                             thisWidth - 2 * TBEVEL, thisHeight - 2 * TBEVEL);
    }
    if (w->flags & WFLG_IN_EDIT) {
        int thickness = w->thickness;

        if (thickness < 0)
            thickness = -thickness;
        XCopyArea(Dpy, w->window, w->parent->window, ForeGC, 0, 0,
                  wt->winWidth, wt->winHeight,
                  w->xpos + thickness, w->ypos + thickness);
    }
    return;
}

/***********************************************************************
function: TblDrawHeading - Actually paint heading string.  Handle
multi-line... 
*/
static void
TblDrawHeading(w, xpos, ypos, mbstr)
AXWidget_table w;
int xpos;
int ypos;
mbchar *mbstr;
{
    int first;
    mbchar *tmp_str = (mbchar *) NULL;
    mbchar *p;
    table_info *tp;
    int len;
    XFontStruct *saved_font;

    if (!w || w->type != W_TABLE)
        return;
    if (!(tp = (table_info *) w->info.table))
        return;
    if (!mbstr)
        return;
    /* MBstrtok() steps on it's first arg. */
    tmp_str = (mbchar *) malloc(MBstrlen(mbstr) + 1);
    MBstrcpy(tmp_str, mbstr);
    for (first = TRUE;
         (p = MBstrtok(first ? (char *)tmp_str : (char *)NULL, "\n"));
         first = FALSE) {
        if (!first)
            ypos += tp->heading_char_height;
        len = MBstrlen(p);
        if (len) {
            saved_font = w->font;
            w->font = tp->headingFont;
            xmDrawString((AXWidget)w, w->window, ForeGC, xpos, ypos, p, len,
                         tp->headingShadow, tp->headingColor,
                         w->widgetColor);
            w->font = saved_font;
        }
    }
    if (tmp_str)
        free(tmp_str);
    return;
}

/*************************************************************************
function: PaintTable - Respond to an exposure
*/
static void
PaintTable(w, exP, regen, layered)
register AXWidget_table w;
XExposeEvent *exP;  /* NULL or exposure event */
int regen;      /* what does this flag do? */
int layered;    /* what does this mean */
{
    register table_info *wt = (table_info *) w->info.table;
    data_ptr rows = wt->rows;
    int numRows = AxArraySize(rows);    /* size of data (not display) */
    int numCols = AxArraySize(wt->colText);
    int colNum, xpos, ypos, rowNum;
    int thisWidth;                      /* width of this column */
    int thisHeight;
    int realize_table = FALSE;
    int dims_changed = TRUE;
    int ix;

    if ( !w->display || !w->parent )
        /* dialog not yet fully created */
        return;

#if AX_X11
    if ( !WIN_OF(w) )
        return;
#endif

    if ( (w->parent->type == W_DIALOG) &&
                !((AXWidget_dialog) w->parent)->exposed )
        return;

#if AXMS
    if ( w->parent->winid == (HWND)NULL )
        /* dialog not yet fully created */
        return;
#endif

    if (layered) {
        if (w->flags & WFLG_IN_EDIT) {
            if (dePaintWidget((AXWidget)w)) {
                return;
            }
        }
        else {
            if (topWidget((AXWidget)w) == (AXWidget)w) {
                paintCompositeWidget((AXWidget)w, exP, regen);
                return;
            }
        }
    }
    
    if (!w->mapped)
        w->mapped = TRUE;

    if (!w->window) {
        RealizeTable(w, layered);   /* create windows, scrolls... */
        realize_table = TRUE;
#if AXMS
    /* If we are here the table was just created, which means we'll get
     * an expose from Windows. Instead of continuing on, we return
     * and let the future expose do the painting; otherwise, we'll get
     * a double paint.
     */
        if ( !regen )
            return;
#endif
    }

    /* Recreate scrollbars if dimensions of scrollbars changed.  Don't
       bother if realizing table (gets taken care of...) */
    if (!realize_table &&
        (wt->vscroll_dim_changed || wt->hscroll_dim_changed)) {
        TblUnmapScrollBars(w,
                           wt->vscroll_dim_changed ? TRUE : FALSE,
                           wt->hscroll_dim_changed ? TRUE : FALSE);
        TblCreateScrollBars(w, layered);
        dims_changed = TRUE;
    } else
        dims_changed = FALSE;

    wt->vscroll_dim_changed = FALSE;
    wt->hscroll_dim_changed = FALSE;

    if (exP) {
        XRectangle crect[1];

        crect[0].x = 0;
        crect[0].y = 0;
        crect[0].width  = exP->width;
        crect[0].height = exP->height;
        TblSetClipAllGCs(w, exP->x, exP->y, crect, 1, YXBanded);
        
        if (wt->cursor.text_cursor_on) {
            if (TblCursorIntersectsRect(wt, exP->x, exP->y,
                                        exP->width, exP->height)) {
                wt->cursor.text_cursor_on = FALSE;  /* assume cleared by exp */
            }
        }
    }

    /* all buttons/borders are done with BackGC, therefore unset work
       colors */
    AxUnsetWidgetWorkColors((AXWidget)w);
    PaintTableBorder(w);
    TblDrawCornerBox(w);
    TblDrawRowMarkers(w);
    TblDrawColumnHeadings(w);

/* If in data-request-model, send the first data-expose here. */
    if ((realize_table == TRUE) && wt->use_data_request_model) {
        TblSendDataRequestEvent(w, wt->topRow, wt->numLines);
        goto painter_leave; /* No need to draw rows.  There aren't any. */
    }

    AxSetWidgetWorkColors((AXWidget)w);
    XFillRectangle(Dpy, w->window, BackGC, wt->markerWidth,
                   wt->headingHeight, wt->winWidth - wt->markerWidth,
                   wt->winHeight - wt->headingHeight);
    wt->cursor.text_cursor_on = FALSE;
    AxUnsetWidgetWorkColors((AXWidget)w);

/* Apply Gridding */
    if (!wt->hide_vertical_grid || !wt->hide_horizontal_grid) {
        int bottom_y, right_x, bottom_row, right_col, total_rows;

        total_rows = TblTotalRows(wt);
        if ((total_rows - wt->topRow) < wt->numLines) {
            bottom_y = YPosOfCell(wt, total_rows - 1, 0);
            if (bottom_y == -1) {
                bottom_y = wt->headingHeight;
            }
            else {
                bottom_y += wt->rowHeight - 1;
            }
            bottom_row = total_rows;
        }
        else {
            bottom_y = wt->winHeight - 1;
            bottom_row = wt->topRow + wt->numLines;
        }
        
        for (ix = right_col = wt->leftCol, right_x = wt->markerWidth;
             ix < numCols; ix++) {
            
            thisWidth = AxIntFromArray(wt->colWidth, ix);
            if (right_x + thisWidth > wt->winWidth - 1) {
                right_x = wt->winWidth - 1;
                break;
            }
            right_x += thisWidth;
            right_col++;
        }
        
        if (!wt->hide_vertical_grid) {
            xpos = wt->markerWidth;
            ypos = wt->headingHeight;
            for (colNum = wt->leftCol;
                 (xpos < wt->winWidth) && (colNum <= right_col); colNum++) {
                
                thisWidth = AxIntFromArray(wt->colWidth, colNum);
                if (ypos != bottom_y) {
                    XDrawLine(Dpy, w->window, wt->grid_gc,
                              xpos, ypos, xpos, bottom_y);
                }
                xpos += thisWidth;
            }
        }

        if (!wt->hide_horizontal_grid) {
            xpos = wt->markerWidth;
            ypos = wt->headingHeight;
            for (rowNum = wt->topRow;
                 (ypos < wt->winHeight) && (rowNum <= bottom_row); rowNum++) {

                if (xpos != right_x) {
                    XDrawLine(Dpy, w->window, wt->grid_gc,
                              xpos, ypos, right_x, ypos);
                }
                ypos += wt->rowHeight;
            }
        }
    }

/* Draw in material */
    ypos = wt->headingHeight;
    thisHeight = wt->rowHeight;
    for (rowNum = wt->topRow; (ypos < wt->winHeight) && (rowNum < numRows);
         rowNum++) {
        
        if (TblRowIsSelected(wt, rowNum)) {
            TblHighlightRow(w, rowNum, wt->leftCol);
        }
        else {
            xpos = wt->markerWidth;
            for (colNum = wt->leftCol;
                 (xpos < wt->winWidth) && (colNum < numCols); colNum++) {
                
                thisWidth = AxIntFromArray(wt->colWidth, colNum);
				if ( !wt->rowcolcreate )
				{	
					if (thisWidth > 0) {
						if ((wt->cursor.col == colNum) &&
							(wt->cursor.row == rowNum)) {
							TblDrawCursorCellText(w);
							TblPutSelectionCursor(w);
							TblDrawEditCellBox(w, rowNum, colNum);
						}
						else {
							r32 clip;
							
							clip.origin.x = xpos + 1;
							clip.origin.y = ypos + 1;
							clip.width = thisWidth - 1;
							clip.height = thisHeight - 1;
							TblDrawCellText(w,rowNum,colNum,xpos,ypos,&clip);
						}
					}
				}
				else {	
					if (thisWidth > 0) {
						r32 clip;
						
						clip.origin.x = xpos + 1;
						clip.origin.y = ypos + 1;
						clip.width = thisWidth - 1;
						clip.height = thisHeight - 1;
						TblDrawCellText(w, rowNum, colNum, xpos, ypos, &clip);
					}
				}

                xpos += thisWidth;
            }
        }
        ypos += thisHeight;
    }

/* Highlight selected rows */
    if (IntsArraySize(wt->selected_rows) > 0) {
        for (rowNum = wt->topRow, ix = 0; ix < wt->numLines;
             ix++, rowNum++) {
            if (TblRowIsSelected(wt, rowNum)) {
                TblDrawRowMarkerShadow(w, rowNum, TRUE);
                TblDrawRowMarkerContents(w, rowNum);
            }
        }
    }
    
    if (w->graphic)
        AxRenderInsetGraphic((AXWidget)w, w->graphic);

painter_leave:
    TblUnsetClipAllGCs(w);

    if (regen || dims_changed) /* for first repaint */
        TblUpdateScrollBars(w);

    return;
}

/*************************************************************************
function: PaintTableRegion - Paint the visible region of the table
starting at row START_ROW and column START_COL and ending at END_ROW
and END_COL, inclusive.
Note:
  If either START_ROW or START_COL are <0, paint full table.
  If END_ROW is <0, paint all visible rows starting at START_ROW.
  If END_COL is <0, paint all visible columns starting at START_COL.

  Also updates scrollbars.
*/
static void
PaintTableRegion(w, start_row, end_row, start_col, end_col)
register AXWidget_table w;
int start_row;
int end_row;
int start_col;
int end_col;
{
    register table_info *tp;
    data_ptr rows;
    int numRows; /* size of data not display */
    int numCols;
    int colNum, xpos, ypos, width, height, rowNum;
    int thisWidth;      /* width of this column */
    int thisHeight;
    int ix;

    if (!w)
        return;
    if (w->type != W_TABLE)
        return;
    if (!(tp = (table_info *) w->info.table))
        return;
    if (!w->display)
        return;

    rows = tp->rows;
    numRows = AxArraySize(rows);
    numCols = AxArraySize(tp->colText);

    if ((start_row < 0) || (start_col < 0)) {
        /* trigger full-paint
         */
        WidgetNeedsUpdated = TRUE;
        w->update_widget = TRUE;
        return;
    }
    if ((start_row <= tp->topRow) &&
        (end_row >= (tp->topRow + tp->numLines - 1))) {
        
        /* range to paint covers fully visible data; trigger full-paint
         */
        WidgetNeedsUpdated = TRUE;
        w->update_widget = TRUE;
        return;
    }

    if (end_row < 0)
        end_row = tp->topRow + tp->numLines - 1;
    if (end_col < 0)
        end_col = numCols;

    AxUnsetWidgetWorkColors((AXWidget)w);
    TblDrawRowMarkersRegion(w, start_row, end_row);
    
    /* compute dimensions of region being repainted
     */
    xpos = XPosOfCell(tp, start_row, start_col);
    if (xpos == -1) {
        xpos = tp->markerWidth;
    }
    ypos = YPosOfCell(tp, start_row, start_col);
    if (ypos == -1) {
        ypos = tp->headingHeight;
    }
    ix = XPosOfCell(tp, end_row + 1, end_col + 1);
    if (ix == -1) {
        ix = tp->winWidth;
    }
    width = ix - xpos + 1;
    ix = YPosOfCell(tp, end_row + 1, end_col + 1);
    if (ix == -1) {
        ix = tp->winHeight;
    }
    height = ix - ypos + 1;

    /* do fill of this region
     */
    AxSetWidgetWorkColors((AXWidget)w);
    XFillRectangle(Dpy, w->window, BackGC, xpos, ypos, width, height);
    if ((tp->cursor.row >= start_row && tp->cursor.row <= end_row) &&
        (tp->cursor.col >= start_col && tp->cursor.col <= end_col)) {
        
        tp->cursor.text_cursor_on = FALSE;
    }
    AxUnsetWidgetWorkColors((AXWidget)w);
    
    /* Apply Gridding */
    if (!tp->hide_vertical_grid || !tp->hide_horizontal_grid) {
        int bottom_y, right_x, bottom_row, right_col, total_rows;
        
        total_rows = TblTotalRows(tp);
        if ((total_rows - tp->topRow) < tp->numLines) {
            bottom_y = YPosOfCell(tp, total_rows - 1, 0);
            if (bottom_y == -1) {
                bottom_y = tp->headingHeight;
            }
            else {
                bottom_y += tp->rowHeight - 1;
            }
            bottom_row = total_rows;
        }
        else {
            bottom_y = tp->winHeight - 1;
            bottom_row = tp->topRow + total_rows;
        }
        
        for (ix = right_col = tp->leftCol, right_x = tp->markerWidth;
             ix < numCols; ix++) {
            
            thisWidth = AxIntFromArray(tp->colWidth, ix);
            if (right_x + thisWidth > tp->winWidth - 1) {
                right_x = tp->winWidth - 1;
                break;
            }
            right_x += thisWidth;
            right_col++;
        }

        if (!tp->hide_vertical_grid) {
            xpos = tp->markerWidth;
            ypos = tp->headingHeight;
            for (colNum = tp->leftCol;
                 (xpos < tp->winWidth) && (colNum <= right_col);
                 colNum++) {
                
                thisWidth = AxIntFromArray(tp->colWidth, colNum);
                if ((colNum >= start_col) && (colNum <= end_col)) {
                    if (ypos != bottom_y) {
                        XDrawLine(Dpy, w->window, tp->grid_gc,
                                  xpos, ypos, xpos, bottom_y);
                    }
                }
                xpos += thisWidth;
            }
        }

        if (!tp->hide_horizontal_grid) {
            xpos = tp->markerWidth;
            ypos = tp->headingHeight;
            for (rowNum = tp->topRow;
                 (ypos < tp->winHeight) && (rowNum <= bottom_row);
                 rowNum++) {
                
                if ((rowNum >= start_row) && (rowNum <= end_row)) {
                    if (xpos != right_x) {
                        XDrawLine(Dpy, w->window, tp->grid_gc,
                                  xpos, ypos, right_x, ypos);
                    }
                }
                ypos += tp->rowHeight;
            }
        }
    }

/* Draw in material */
    ypos = tp->headingHeight;
    thisHeight = tp->rowHeight;
    for (rowNum = tp->topRow; (ypos < tp->winHeight) && (rowNum < numRows);
         rowNum++) {
        
        if (TblRowIsSelected(tp, rowNum)) {
            TblHighlightRow(w, rowNum, start_col);
        }
        else {
            xpos = tp->markerWidth;
            if ((rowNum >= start_row) && (rowNum <= end_row)) {
                for (colNum = tp->leftCol; (xpos < tp->winWidth) &&
                     (colNum < numCols); colNum++) {
                    thisWidth = AxIntFromArray(tp->colWidth, colNum);
                    if (thisWidth > 0) {
                        if ((colNum >= start_col) && (colNum <= end_col)) {
                            if ((tp->cursor.col == colNum) &&
                                (tp->cursor.row == rowNum)) {
                                TblDrawCursorCellText(w);
                                TblDrawEditCellBox(w, rowNum, colNum);
                            }
                            else {
                                r32 clip;
                                
                                clip.origin.x = xpos + 1;
                                clip.origin.y = ypos + 1;
                                clip.width = thisWidth - 1;
                                clip.height = thisHeight - 1;
                                
                                TblDrawCellText(w, rowNum, colNum,
                                                xpos, ypos, &clip);
                            }
                        }
                    }
                    xpos += thisWidth;
                }
            }
        }
        ypos += thisHeight;
    }

/* Highlight selected rows. */
    if (IntsArraySize(tp->selected_rows) > 0) {
        for (rowNum = tp->topRow, ix = 0; ix < tp->numLines;
             ix++, rowNum++) {
            if ((rowNum >= start_row) && (rowNum <= end_row)) {
                if (TblRowIsSelected(tp, rowNum)) {
                    TblDrawRowMarkerShadow(w, rowNum, TRUE);
                    TblDrawRowMarkerContents(w, rowNum);
                }
            }
        }
    }

    TblUpdateScrollBars(w);
    return;
}

/***********************************************************************
function: TblUpdateScrollBars - Get scroll bars in sync.
*/
static void
TblUpdateScrollBars(w)
register AXWidget_table w;
{
    register table_info *wt;
    int numRows, leftColPos, colNum, totalWidth, numCols;

    if (!w)
        return;
    if (!(wt = (table_info *) w->info.table))
        return;

#if CLIP_DEBUG
    if (tblSaveClip.clips_on) {
        printf("TblUpdateScrollBars... clips are on!!\n");
        abort();
    }
#endif
    
    numRows = TblTotalRows(wt);
    
    /* determine total number of columns */
    if (wt->colText)
        numCols = AxArraySize(wt->colText);
    else
        numCols = 0;

    /* adjust vertical scrollbar */
    if (w->vScrollBar) {
        if ( wt->rowcolcreate ){
            w->vScrollBar->scroll_per = wt->winHeight;
            w->vScrollBar->scroll_max = numRows * wt->rowHeight;
            w->vScrollBar->scroll_pos = wt->topRow * wt->rowHeight;
            if ( w->vScrollBar->scroll_per >= w->vScrollBar->scroll_max )
                w->vScrollBar->scroll_per = w->vScrollBar->scroll_max - 1;
        }
        else {  
			if ((numRows - wt->topRow) < wt->numLines) {
				/* there's some empty space at the bottom of the window */
				int extraRows = wt->numLines - (numRows - wt->topRow);
				if (extraRows > 0) {
					numRows += extraRows;
				}
			}
            w->vScrollBar->scroll_max = numRows * 10;
            w->vScrollBar->scroll_per = MAX(1, (wt->numLines - 1)) * 10;
            w->vScrollBar->scroll_pos = wt->topRow * 10;
        }
        PaintScrollBar(w->vScrollBar, 0, FALSE, FALSE);
    }
        
    /* adjust horizontal scrollbar */
    if (w->hScrollBar) {
        int rightOfLastColumn;
        leftColPos = 0;
        for ( colNum = 0; colNum < wt->leftCol; colNum++ )
            leftColPos += AxIntFromArray(wt->colWidth, colNum);
        totalWidth = leftColPos;
        for ( colNum = wt->leftCol; colNum < numCols; colNum++ ) {
            totalWidth += AxIntFromArray(wt->colWidth, colNum);
        }
        rightOfLastColumn = (wt->winWidth -
                             AxIntFromArray(wt->colWidth, numCols - 1));
        if ( wt->rowcolcreate ) {
            w->hScrollBar->scroll_pos = leftColPos;
            w->hScrollBar->scroll_max = totalWidth;
            w->hScrollBar->scroll_per = wt->winWidth;
            if ( w->hScrollBar->scroll_per >= w->hScrollBar->scroll_max )
                w->hScrollBar->scroll_per = w->hScrollBar->scroll_max - 1;
        }
        else
        {   
            w->hScrollBar->scroll_per = wt->winWidth * 10;
            w->hScrollBar->scroll_max = (totalWidth + rightOfLastColumn) * 10;
            w->hScrollBar->scroll_pos = leftColPos * 10;
        }
        PaintScrollBar(w->hScrollBar, 0, FALSE, FALSE);
    }
    return;
}

/*************************************************************************
function: TableComputeExposedCols -
*/
static int
TableComputeExposedCols(w)
AXWidget_table w;
{
    register table_info *wt = (table_info *) w->info.table;
    int xpos, thisWidth, colNum, numCols, num_exposed_cols;
    
    xpos = wt->markerWidth;
    num_exposed_cols = 0;
    numCols = AxArraySize(wt->colText);
    for (colNum = wt->leftCol; 
         (xpos < wt->winWidth) && (colNum < numCols); colNum++) {
        thisWidth = AxIntFromArray(wt->colWidth, colNum);
        xpos += thisWidth;
        num_exposed_cols++;
    }
    return(num_exposed_cols);
}

/*************************************************************************
function: TableBarChanged - Callback routine for Tables
*/
static void
TableBarChanged(widget, w, data)
Widget widget;
AXWidget_table w;
XmScrollBarCallbackStruct *data;
{
    register table_info *wt = (table_info *) w->info.table;
    int new_top_row;
    int new_left_col, new_cols, new_rows, exposed_cols, valid_col;
    int valid_row;

    if ( !wt->rowcolcreate ) {
        if (w->vScrollBar == (AXWidget_scroll )widget) {
            int numRows = TblTotalRows(wt);
            
            switch ( data->reason ) {
            case XmCR_INCREMENT:
                new_top_row = MIN(wt->topRow + 1, numRows);
                if ((new_top_row + wt->numLines - 1) > numRows)
                    return;
            break;
            case XmCR_DECREMENT:
                new_top_row = MAX(0, wt->topRow - 1);
                break;
            case XmCR_PAGE_INCREMENT:
                new_top_row = wt->topRow + wt->numLines - 1;
                if (new_top_row == wt->topRow)
                    return; /* nothing happens */
                break;
            case XmCR_PAGE_DECREMENT:
                new_top_row = MAX(0, wt->topRow - (wt->numLines - 1));
                if (new_top_row == wt->topRow)
                    return; /* nothing happens */
                break;
            case XmCR_VALUE_CHANGED:
                if (wt->topRow == data->value/10)
                    return;
                new_top_row = data->value / 10;
                if ((new_top_row + wt->numLines - 1) > numRows) {
					/* there's empty space at the bottom of the window */
					if (new_top_row > wt->topRow) {
						/* we're trying to scroll the material upwards,
						 * but its bottom edge is already fully in view */
						return;
					}
                    new_top_row = MAX(0, new_top_row - 1);
				}
                if (new_top_row == wt->topRow)
                    return; /* nothing happens */
                break;
            default:
                return;
            }
            TblScrollNewTopRow(w, new_top_row);
        }
        else if (w->hScrollBar == (AXWidget_scroll ) widget) {
            int numCols = AxArraySize(wt->colText);
            int leftColPos, colNum;
            
            switch ( data->reason ) {
            case XmCR_INCREMENT:
            case XmCR_PAGE_INCREMENT:
                new_left_col = MIN(wt->leftCol + 1, numCols - 1);
                break;
            case XmCR_DECREMENT:
            case XmCR_PAGE_DECREMENT:
                new_left_col = MAX(0, wt->leftCol - 1);
                break;
            case XmCR_VALUE_CHANGED:
                leftColPos = 0;
                for ( colNum = 0; colNum < numCols; colNum++ ) {
                    if (data->value <= leftColPos * 10)
                        break;
                    leftColPos += AxIntFromArray(wt->colWidth, colNum);
                }
                if (wt->leftCol == colNum)
                    return;
                new_left_col = colNum;
                break;
            default:
                return;
            }
            TblScrollNewLeftCol(w, new_left_col);
        }
    }
    else {
        /* presentations mode */
        new_cols = 0;
        if (w->vScrollBar == (AXWidget_scroll )widget) {
            int numRows = TblTotalRows(wt);

            switch ( data->reason ) {
            case XmCR_INCREMENT:
                new_top_row = MIN(wt->topRow + 1, numRows);
                TblRowColCreate(w, 1, new_cols, 0);
                break;
            case XmCR_DECREMENT:
                new_top_row = MAX(0, wt->topRow - 1);
                valid_row = TblGetLastValidRow(w);                     
                if ( new_top_row + wt->numLines > wt->totalRowCount-1); {
                    new_rows = 1;
                    if ( new_top_row + wt->numLines > valid_row)
                        TblDeleteRowOfData(w, new_rows);
                }
                break;
            case XmCR_PAGE_INCREMENT:
                new_top_row = wt->topRow + wt->numLines - 1;
                TblRowColCreate(w, 1, new_cols, 0);
                break;
            case XmCR_PAGE_DECREMENT:
                new_top_row = MAX(0, wt->topRow - (wt->numLines - 1));
                break;
            case XmCR_VALUE_CHANGED:
                new_top_row = data->value / wt->rowHeight;
                if ((new_top_row + wt->numLines - 1) > numRows)
                    new_top_row = MAX(wt->topRow,(numRows -
                                                  wt->numLines));
                if ( new_top_row < wt->topRow ) {
                    valid_row = MAX(TblGetLastValidRow(w), 
                                    new_top_row+wt->numLines);
                    new_rows = wt->topRow-new_top_row; 
                    if ( valid_row+1 < numRows-new_rows ) {    
                        TblDeleteRowOfData(w, new_rows); 
                    }
                }
                break;

            default:
                break;
            }
            if (new_top_row != wt->topRow)
                TblScrollNewTopRow(w, new_top_row);
        }
        else if (w->hScrollBar == (AXWidget_scroll ) widget) {
            int numCols = AxArraySize(wt->colText);
            int leftColPos, colNum;
            exposed_cols = TableComputeExposedCols(w);            
            switch ( data->reason ) {
            case XmCR_INCREMENT:
            case XmCR_PAGE_INCREMENT:
                new_left_col = MIN(wt->leftCol + 1, numCols - 1);
                if ( new_left_col + exposed_cols < numCols )
                    new_cols = new_left_col + exposed_cols - numCols;
                else
                    new_cols = 1;
                TblRowColCreate(w, 1, new_cols, exposed_cols);
                break;
            case XmCR_DECREMENT:
            case XmCR_PAGE_DECREMENT:
                new_left_col = MAX(0, wt->leftCol - 1);
                valid_col    = MAX(TblGetLastValidCellPos(w), exposed_cols);
                if ( valid_col+1 < AxArraySize(wt->colText) ) {   
                    new_cols = 1;
                    TblDeleteColumnOfData(w, new_cols); 
                }
                break;
            case XmCR_VALUE_CHANGED:
                leftColPos = 0;
                for ( colNum = 0; colNum < numCols; colNum++ ) {
                    if ((data->value-1) <= leftColPos)
                        break;
                    leftColPos += AxIntFromArray(wt->colWidth, colNum);
                }
                new_left_col = colNum;
                if ( new_left_col < wt->leftCol ) {
                    valid_col = MAX(TblGetLastValidCellPos(w), 
                                    wt->leftCol + exposed_cols);
                    if ( valid_col+1 < AxArraySize(wt->colText) ) {   
                        new_cols = wt->leftCol - new_left_col;
                        TblDeleteColumnOfData(w, new_cols); 
                    }
                }
                break;
            default:
                break;
            }
            TblScrollNewLeftCol(w, new_left_col);
        }
    }
    return;
}

/* Returns the last row which has data. */
static int
TblGetLastValidRow(w)
AXWidget_table w;
{
    register table_info *wt = (table_info *) w->info.table;
    data_ptr rowData, colData;
    int n_rows, n_cols, i, j;
    int row_pos=0;
    n_rows = AxArraySize(wt->rows);
    n_cols = AxArraySize(wt->colText);

    for ( i = n_rows; i > 0; i-- ){
        rowData = ReadArray(AX(wt->rows), i);
        for(j = n_cols; j >= 0; j--){
            colData = AxArrayElement(rowData, j);
            if(AxIsNumber(colData) || AxIsString(colData)){
                if ( i > row_pos )
                    row_pos = i;
            }
        }
    }
    return(row_pos);
}



static int
TblGetLastValidCellPos(w)
AXWidget_table w;
{
    register table_info *wt = (table_info *) w->info.table;
    data_ptr rowData, colData;
    int n_rows, n_cols, i, j;
    int col_pos=0;
    n_rows = AxArraySize(wt->rows);
    n_cols = AxArraySize(wt->colText);

    for ( i = 0; i < n_rows; i++ ){
        rowData = ReadArray(AX(wt->rows), i);
        for(j = n_cols; j > 0; j--){
            colData = AxArrayElement(rowData, j);
            if(AxIsNumber(colData) || AxIsString(colData)){
                if ( j > col_pos )
                    col_pos = j;
            }
        }
    }
    return(col_pos+1);
}


/*************************************************************************
function: TblDeleteColumnOfData- Delete a column of data if possible, 
ONLY in presentations mode
*/
static void
TblDeleteColumnOfData(w, n_cols)
AXWidget_table w;
int n_cols;
{
    register table_info *wt = (table_info *) w->info.table;
    data_ptr rowData, n_rowData;
    int colsize, size, ix, sz;

    colsize = AxArraySize(wt->colText);
 
    /* remove n_cols number of columns from each row of data */
    if ( n_cols > 0 ) {         
        size = AxArraySize(wt->rows);
        for ( ix = 0; ix < size; ix++) {
            rowData    = ReadArray(AX(wt->rows), ix);
            n_rowData  = AxTaskCopyElfData(0, rowData);
            sz         = AxArraySize(n_rowData);
            n_rowData  = (data_ptr)TaskTruncArray(0, n_rowData, sz - n_cols);
            wt->rows   = TaskWriteArray(0, wt->rows, ix, n_rowData);
        }

        /* Now free the coltext , colWidth, colType*/
        sz            = AxArraySize(wt->colText);
        wt->colText   = (data_ptr)TaskTruncArray(0,wt->colText, sz-n_cols); 
        sz            = AxArraySize(wt->colWidth);
        wt->colWidth  = (data_ptr)TaskTruncArray(0,wt->colWidth, sz-n_cols); 
        sz            = AxArraySize(wt->colType);
        wt->colType   = (data_ptr)TaskTruncArray(0,wt->colType, sz-n_cols);
        wt->right_col = sz - n_cols;        
        sz            = AxArraySize(ReadArray(AX(wt->rows), 1));
        size          = AxArraySize(n_rowData);
    }
}

/*************************************************************************
function: TblDeleteRowOfData- Delete a row of data if possible, 
ONLY in presentations mode
*/
static void
TblDeleteRowOfData(w, n_rows)
AXWidget_table w;
int n_rows;
{
    register table_info *wt = (table_info *) w->info.table;
    int size, ix;
    data_ptr rowData, n_rowData, rows;  

    /* remove n_rows number of columns from each row of data */
    if ( n_rows > 0 ) {         
        size = AxArraySize(wt->rows);
        rows = AxTaskCopyElfData(0, wt->rows);
        AxTaskFreeElfData(0, wt->rows);
        wt->rows = AxTaskCreateElfArray(0, 0);
        for ( ix = 0; ix < size-n_rows; ix++) {
            rowData    = ReadArray(AX(rows), ix);
            n_rowData  = AxTaskCopyElfData(0, rowData);
            wt->rows       = TaskWriteArray(0, wt->rows, ix, n_rowData);
        }
        wt->totalRowCount = size - n_rows;
    }
}



/*************************************************************************
function: TblRowColCreate - 
if you donot want to create a col during vscroll set exposed cols to 0
if you donot want to create a row during hscroll set n_rows to 0
*/
static void
TblRowColCreate(w, n_rows, n_cols, exposed_cols)
AXWidget_table w;
int n_rows, n_cols;
int exposed_cols;
{
    register table_info *wt = (table_info *) w->info.table;
    TblEventData data1;
    int colsize, hit_bottom, hit_right_corner;

    colsize = AxArraySize(wt->colText);
    hit_right_corner = ((wt->leftCol+exposed_cols) > colsize -1 ? TRUE : FALSE);
    hit_bottom = (wt->topRow+wt->numLines > wt->totalRowCount-1 ? TRUE : FALSE);
    if ( hit_bottom || hit_right_corner ) {         
        if ( n_rows < 0 )
            n_rows = 1;
        if ( n_cols < 0 )
            n_cols = 1;
        colsize = TblAddData(w, hit_bottom, hit_right_corner, n_rows, n_cols);
        data1.rowcolcreate.type      = EV_TBL_ROWCOLCREATE_;
        data1.rowcolcreate.totalCol  = colsize;
        data1.rowcolcreate.totalRow  = wt->totalRowCount;
        AxCallback(w->callback, (AXWidget)w, w->callbacktag, (caddr_t)&data1);
    }

}

/*************************************************************************
function: TblAddData - 
*/
static int
TblAddData(w, addrows, addcols, n_rows, n_cols)
register AXWidget_table w;
int addrows;
int addcols;
int n_rows, n_cols;
{
    register table_info *wt = (table_info *) w->info.table;
    data_ptr new_row_data, row_data;
    int i, jx, size, colSize, col_size;
    
    if (addrows) {
        wt->totalRowCount += n_rows;
        size = AxArraySize(wt->rows);
        colSize = AxArraySize(ReadArray(AX(wt->rows), 0));
        row_data = AxTaskCreateElfArray(0, colSize);                
        for ( i = 0; i < colSize; i++ )
            row_data = TaskWriteArray(0, row_data, i, NULL);
        wt->rows = TaskWriteArray(0, wt->rows, size, row_data);
    }
    if (addcols) {
        size = AxArraySize(wt->rows);
        row_data = ReadArray(AX(wt->rows), 0);
        
        if ( n_cols > 1)
            colSize = wt->leftCol + n_cols;
        else
            colSize = AxArraySize(row_data) + 1;
            
        for ( i = 0; i < size; i++ ) {  
            row_data = ReadArray(AX(wt->rows), i);
            col_size = AxArraySize(row_data);
            new_row_data  = AxTaskCopyElfData(0, row_data);
            for ( jx=col_size; jx < colSize; jx++)
                new_row_data = TaskWriteArray(0, new_row_data, jx, NULL);
            wt->rows = TaskWriteArray(0, wt->rows, i, new_row_data);
        }
    }
    return(colSize);
}

/*************************************************************************
function: RealizeTable - Create window & scroll bars
*/
static void RealizeTable(w, layered)
register AXWidget_table w;
int layered;    /* what does this mean */
{
    register table_info *wt = (table_info *) w->info.table;
    int x, y, win_xpos, win_ypos, width, height;
    unsigned long mask;
    XSetWindowAttributes atts;
    int thickness = w->thickness;
#if AXMS
    HDC tmp_dc;
    HBITMAP old_hbm;
    Pixmap save_pixmap = NULL;
#endif

    if (w->thickness < 0)
        thickness = -thickness;

    x = w->xpos;
    y = w->ypos;
    win_xpos = x + thickness;
    win_ypos = y + thickness;
    width = w->width;
    height = w->height;
    wt->winWidth = width - 2*thickness;
    wt->winHeight = height - 2*thickness;
    wt->lineHeight =  MAX(wt->char_height, wt->heading_char_height) + 4 + 2;
    wt->numLines = TblComputeNumLines(w);

    mask = WinMask;
    atts = WinAtt;
    if (w->useWorkColors) {
        mask |= CWBackPixel | CWColormap;
        atts.background_pixel = BackWorkPixel;
        atts.colormap = WorkColormap;
    }
    if (w->flags & WFLG_IN_EDIT) { /* dialog editor */
        w->window = AxCreateQueryPixmap(Dpy, w->parent->window,
                                        wt->winWidth, wt->winHeight,
                                        DpyDepth);
#if AXMS
        /* we save the pixmap handle, and replace it with an HDC so that */
        /* all the lower levels can draw correctly */
        save_pixmap = w->window;
        tmp_dc = MS_CreateCompatibleDC(NULL);
        old_hbm = MS_SelectObject(tmp_dc, save_pixmap);
        w->window = AxPackageHandle(tmp_dc, AX_MEMDC_TYPE);
#endif

        XFillRectangle(Dpy, w->window, BackGC, 0, 0,
                       wt->winWidth, wt->winHeight);
    }
    else {
#if AXMS
        if ( !w->winid )
            WinCreateTableWindow(w);
#else        
        w->window = AxCrtWindow(Dpy, WIN_OF(w),
                                win_xpos, win_ypos, wt->winWidth,
                                wt->winHeight, 0, CopyFromParent, InputOutput,
                                CopyFromParent, mask, &atts);
        XSelectInput(Dpy, w->window, TBL_INFOCUS_MASK);
        XMapWindow(Dpy, w->window);
        AXDefineDefaultCursor((AXWidget)w, ThimPointerCursor);
#endif
        AddToWindowList((AXWidget)w);
    }

    TblCreateGCs(w);

    AxSetWidgetWorkColors((AXWidget)w);
/* clear gap between border and window
   (should be AFTER GCs changed for work area colors) */
    XFillRectangle(Dpy, w->window, BackGC, 0, 0, wt->winWidth, wt->winHeight);
    AxUnsetWidgetWorkColors((AXWidget)w);

    TblCreateScrollBars(w, layered);

    if (w->flags & WFLG_IN_EDIT) {
        XCopyArea(Dpy, w->window, w->parent->window, ForeGC, 0, 0,
                  wt->winWidth, wt->winHeight, win_xpos, win_ypos);
    }

#if AXMS
    /* put back the pixmap, and destroy the HDC.  We leave the pixmap */
    /* in the widget because we need to get rid of it later when the  */
    /* widget gets destroyed                                          */
    if (w->flags & WFLG_IN_EDIT)    /* dialog editor */
        {
        /* make sure pixmap is unselected */
        MS_SelectObject(tmp_dc, old_hbm);
        MS_DeleteDC(tmp_dc);
        REMOVE_DRAWABLE(w->window);
        w->window = save_pixmap;
        }
#endif
}

/***********************************************************************
function: TblCreateScrollBars - 
*/
static void
TblCreateScrollBars(w, layered)
AXWidget_table w;
int layered;
{
    table_info *tp;
    int vscroll_origin = 0;
    int vscroll_length = 0;
    int hscroll_origin = 0;
    int hscroll_length = 0;

    if (!w)
        return;

    tp = (table_info *) w->info.table;
    if (tp) {
        vscroll_origin = tp->vScrollOrigin;
        vscroll_length = tp->vScrollLength;
        hscroll_origin = tp->hScrollOrigin;
        hscroll_length = tp->hScrollLength;
    }

#if CLIP_DEBUG
    if (tblSaveClip.clips_on) {
        printf("TblCreateScrollBars... clips are on!!\n");
        abort();
    }
#endif
    /* NOTE: scrollbar parent MUST be dialog to ensure proper
     *       behavior when task of listbox is pended */
    if (!w->hide_vscroll) {
        Arg args[10];
        int n = 0;
        int offset = 2;
        static XtCallbackRec callback_list[2];

        if (w->thickness == 0)
            offset = 0;
        CALLBACKER(callback_list, TableBarChanged, w);
        XtSetArg(args[n], XmNx, w->xpos+w->width+offset); n++;
        XtSetArg(args[n],XmNy,w->ypos + vscroll_origin); n++;
        if (vscroll_length) {
            XtSetArg(args[n],XmNheight, vscroll_length); n++;
        } else {
            XtSetArg(args[n],XmNheight,w->height); n++;
        }
        XtSetArg(args[n],XmNorientation,XmVERTICAL); n++;
        XtSetArg(args[n],XmNvalueChangedCallback,callback_list); n++;
        if (w->vScrollBar)
            AxXtSetValues((AXWidget)w->vScrollBar, args, n);
        else
            w->vScrollBar = (AXWidget_scroll)AXmCreateScrollBar((AXWidget)w->parent,
                                "x",args,n);

        if (w->event_table) {
                /* give the newly created scrollbar an event table
                 * which matches that of the list widget */
            copyEventHandlerList(w->event_table,
                                 &w->vScrollBar->event_table);
        }
            /* dlg editor support */
        XtSetArg(args[0], ZXmNcompositeLinkUp, w);
        XtSetArg(args[1], ZXmNinEdit, w->flags & WFLG_IN_EDIT);
        AxXtSetValues((AXWidget)w->vScrollBar, args, 2);
        XtSetArg(args[0], ZXmNcompositeLinkDown, w->vScrollBar);
        AxXtSetValues((AXWidget)w, args, 1);

        xmSetWidgetDisplay((AXWidget)w->vScrollBar, TRUE);
        w->vScrollBar->scroll_per = 100;
        w->vScrollBar->scroll_max = 100;
        w->vScrollBar->scroll_pos = 0;
        PaintScrollBar(w->vScrollBar,0,FALSE,layered);
    }

    if (w->list_hscroll) {
        Arg args[10];
        int n = 0;
        static XtCallbackRec hcallback_list[2];
        int offset = 2;

        if (w->thickness == 0)
            offset = 0;
        CALLBACKER(hcallback_list,TableBarChanged,w);
        XtSetArg(args[n], XmNx, w->xpos + hscroll_origin); n++;
        XtSetArg(args[n],XmNy,w->ypos+w->height+offset); n++;
        if (hscroll_length) {
            XtSetArg(args[n],XmNwidth, hscroll_length); n++;
        } else {
            XtSetArg(args[n],XmNwidth,w->width); n++;
        }
        XtSetArg(args[n],XmNorientation,XmHORIZONTAL); n++;
        XtSetArg(args[n],XmNvalueChangedCallback,hcallback_list); n++;
        if (w->hScrollBar)
            AxXtSetValues((AXWidget)w->hScrollBar, args, n);
        else
            w->hScrollBar = (AXWidget_scroll)AXmCreateScrollBar((AXWidget)w->parent,
                                                          "y",args,n);

        if (w->event_table) {
                /* give the newly created scrollbar an event table
                 * which matches that of the list widget */
            copyEventHandlerList(w->event_table,
                                 &w->hScrollBar->event_table);
        }
            /* dlg editor support */
        XtSetArg(args[0], ZXmNcompositeLinkUp, w);
        XtSetArg(args[1], ZXmNinEdit, w->flags & WFLG_IN_EDIT);
        AxXtSetValues((AXWidget)w->hScrollBar, args, 2);
        XtSetArg(args[0], ZXmNcompositeLinkDown, w->hScrollBar);
        AxXtSetValues((AXWidget)w, args, 1);
        xmSetWidgetDisplay((AXWidget)w->hScrollBar, TRUE);
        w->hScrollBar->scroll_per = 100;
        w->hScrollBar->scroll_max = 100;
        w->hScrollBar->scroll_pos = 0;
        PaintScrollBar(w->hScrollBar,0,FALSE,layered);
    }

    /* Propagate `buttonUpScroll' value here.  This is done here because
       we need valid handles to the scrollbar widgets of this table.  This
       attribute might not have been set at the AxXtSetValues() layer because
       the table might not have scrollbars at that point. */
    if (w->hScrollBar)
        w->hScrollBar->buttonUpScroll = w->buttonUpScroll;
    if (w->vScrollBar)
        w->vScrollBar->buttonUpScroll = w->buttonUpScroll;

    if (tp) {
        tp->vscroll_dim_changed = FALSE;
        tp->hscroll_dim_changed = FALSE;
    }
    return;
}

/***********************************************************************
function: TblSetClipAllGCs - Set clip rectangle(s) on all table
related drawing GCs
*/
static void
TblSetClipAllGCs(w, clip_x_origin, clip_y_origin, rectangles, n,
                 ordering)
register AXWidget_table w;
int clip_x_origin;
int clip_y_origin;
XRectangle *rectangles;
int n;
int ordering;
{
    register table_info *tp;

    if (!w)
        return;
    if (!(tp = (table_info *) w->info.table))
        return;

#if CLIP_DEBUG
    if (tblSaveClip.clips_on) {
        printf("TblSetClipAllGCs: Clips are already on! aborting...\n");
        abort();
    }
#endif

    /* Save this clip description. */
    tblSaveClip.clips_on = TRUE;
    tblSaveClip.x_origin = clip_x_origin;
    tblSaveClip.y_origin = clip_y_origin;
    tblSaveClip.ordering = ordering;
    memcpy((char *) &tblSaveClip.crect, (char *) rectangles,
           sizeof(XRectangle)); /* note: assumes 1 rectangle passed */

    XSetClipRectangles(Dpy, BackGC, clip_x_origin, clip_y_origin,
                       rectangles, n, ordering);
    XSetClipRectangles(Dpy, ForeGC, clip_x_origin, clip_y_origin,
                       rectangles, n, ordering);
    XSetClipRectangles(Dpy, DisabledGC, clip_x_origin, clip_y_origin,
                       rectangles, n, ordering);
    XSetClipRectangles(Dpy, SunGC, clip_x_origin, clip_y_origin,
                       rectangles, n, ordering);
    XSetClipRectangles(Dpy, ShadowGC, clip_x_origin, clip_y_origin,
                       rectangles, n, ordering);
    XSetClipRectangles(Dpy, RecessGC, clip_x_origin, clip_y_origin,
                       rectangles, n, ordering);
    XSetClipRectangles(Dpy, tp->grid_gc, clip_x_origin, clip_y_origin,
                       rectangles, n, ordering);
    XSetClipRectangles(Dpy, tp->mark_gc, clip_x_origin, clip_y_origin,
                       rectangles, n, ordering);
    return;
}

/***********************************************************************
function: TblUnsetClipAllGCs - Unset clips from all table drawing
related GCs.
*/
static void
TblUnsetClipAllGCs(w)
register AXWidget_table w;
{
    register table_info *tp;

    if (!w)
        return;
    if (!(tp = (table_info *) w->info.table))
        return;

    tblSaveClip.clips_on = FALSE;

    XSetClipMask(Dpy, BackGC, None);
    XSetClipMask(Dpy, ForeGC, None);
    XSetClipMask(Dpy, DisabledGC, None);
    XSetClipMask(Dpy, SunGC, None);
    XSetClipMask(Dpy, ShadowGC, None);
    XSetClipMask(Dpy, RecessGC, None);
    XSetClipMask(Dpy, tp->grid_gc, None);
    XSetClipMask(Dpy, tp->mark_gc, None);
    return;
}

/***********************************************************************
function: TblRestoreSavedClips - Restore previously saved clips from
`tblSaveClip'.
*/
static void
TblRestoreSavedClips(w)
AXWidget_table w;
{
    if (!w)
        return;
    TblSetClipAllGCs(w, tblSaveClip.x_origin, tblSaveClip.y_origin,
                     &tblSaveClip.crect, 1, tblSaveClip.ordering);
    return;
}

/******************************************************************
function: TblClip - set clip on the passed gc to the passed dimensions.
If there is already a clip in effect, then intersect the passed
dimensions.  Return FALSE if this intersection is null.  Return
TRUE if either there was no clip in effect, or the intersection is
non-null.
*/
static int
TblClip(gc, origx, origy, width, height)
GC gc;
int origx, origy, width, height;
{
    XRectangle crect;
    int cornx, corny;
    
    cornx = origx + width - 1;
    corny = origy + height - 1;
    
    if (tblSaveClip.clips_on) {
        /* intersect passed cliprect with current
         */
        int minCx, minCy, maxOx, maxOy;
        int torigx, torigy, tcornx, tcorny;

        torigx = (int) tblSaveClip.crect.x + tblSaveClip.x_origin;
        torigy = (int) tblSaveClip.crect.y + tblSaveClip.y_origin;
        tcornx = (int) (torigx + (int) tblSaveClip.crect.width - 1);
        tcorny = (int) (torigy + (int) tblSaveClip.crect.height - 1);
        
        minCx = cornx < tcornx ? cornx : tcornx;
        minCy = corny < tcorny ? corny : tcorny;
        maxOx = origx > torigx ? origx : torigx;
        maxOy = origy > torigy ? origy : torigy;
        
        if ((minCx > maxOx) && (minCy > maxOy)) {
            crect.x = (short) maxOx;
            crect.y = (short) maxOy;
            crect.width = (unsigned short) (minCx - maxOx + 1);
            crect.height = (unsigned short) (minCy - maxOy + 1);
        }
        else {
            return FALSE;        /* intersection is null */
        }
    }
    else {
        crect.x = (short) origx;
        crect.y = (short) origy;
        crect.width = (unsigned short) (cornx - origx + 1);
        crect.height = (unsigned short) (corny - origy + 1);
    }
    
    XSetClipRectangles(Dpy, gc, 0, 0, &crect, 1, Unsorted);
    
    return TRUE;
}

static void
TblRestoreClip(gc)
GC gc;
{
    if (tblSaveClip.clips_on) {
        XSetClipRectangles(Dpy, gc, tblSaveClip.x_origin,
                           tblSaveClip.y_origin, &tblSaveClip.crect,
                           1, tblSaveClip.ordering);
    }
    else {
        XSetClipMask(Dpy, gc, None);
    }
}

/***********************************************************************
function: TableAutoScroll - Auto scroll the list box, and do row selecting.
*/
void
TableAutoScroll(w)
register AXWidget_table w;
{
    register table_info *tp;
    int new_top_row;

    if (!w)
        return;
    if (w->type != W_TABLE)
        return;
    if (!w->multi_pick)
        return;
    if (!(tp = (table_info *) w->info.table))
        return;

    /* No auto scroll if using data request model.
       FIXME, allow horizontal auto-scroll for data request model. */
    if (tp->use_data_request_model)
        return;

    if (!tp->doing_row_marking && !tp->doing_column_marking)
        return;
    
    if (tp->numLines > TblTotalRows(tp))
        return;                       /* all rows are fully visible */

    if ((w->armed != XmCR_INCREMENT) && (w->armed != XmCR_DECREMENT))
        return;

    if (w->armed == XmCR_INCREMENT) {
        if (tp->doing_row_marking) {
            new_top_row = tp->topRow + 1;
            TblScrollNewTopRow(w, new_top_row);
            if (new_top_row != tp->topRow)
                return;
            TblAddSelectionRange(w, w->value,
                                 tp->topRow + tp->numLines - 1/*- 3*/,
                                 ROW_MODE);
            w->last_item_selected = tp->topRow + tp->numLines - 1 /*- 3 */;
        }
#if 0 /* FIXME, no range checking... */
        else if (tp->doing_column_marking) {
            new_left_col = tp->leftCol + 1;
            TblScrollNewLeftCol(w, new_left_col);
            TblAddSelectionRange(w, w->value, new_left_col, COL_MODE);
            w->last_item_selected = new_left_col;
        }
#endif

    } else { /* XmCR_DECREMENT */
        if (tp->doing_row_marking) {
            new_top_row = tp->topRow - 1;
            TblScrollNewTopRow(w, new_top_row);
            if (new_top_row != tp->topRow)
                return;
            TblAddSelectionRange(w, w->value, tp->topRow, ROW_MODE);
            w->last_item_selected = tp->topRow;
        }
#if 0 /* FIXME, no range checking... */
        else if (tp->doing_column_marking) {
            new_left_col = tp->leftCol - 1;
            TblScrollNewLeftCol(w, new_left_col);
            TblAddSelectionRange(w, w->value, new_left_col, COL_MODE);
            w->last_item_selected = new_left_col;
        }
#endif
    }
    return;
}

/***********************************************************************
function: TblScrollNewTopRow - Handle a table scroll so as to make
NEW_TOP_ROW the new top row.
*/
static int
TblScrollNewTopRow(w, new_top_row)
register AXWidget_table w;
int new_top_row;
{
    register table_info *tp;
    int max_num_rows;
    int src_x, src_y, dst_x, dst_y;
    unsigned int width, height;
    int clip_x_origin, clip_y_origin;
    XRectangle crect[1];
    int start_row = -1;
    int start_col = -1;
    int end_row = -1;
    int end_col = -1;
    int num_exposed_rows;

    if (!w)
        return(-1);
    if (!(tp = (table_info *) w->info.table))
        return(-1);
    if (new_top_row < 0)
        new_top_row = 0; /* clip */
    
    max_num_rows = TblTotalRows(tp);

    new_top_row = MIN(new_top_row, max_num_rows); /* clip if necessary */
    if (new_top_row == tp->topRow)
        return(0); /* nothing happens */

    if (tp->cursor.col != -1) { /* FIXME, use `tp->doing_cell_editing' */
        TblLiftSelectionCursor(w);
    }
    
    num_exposed_rows = ABS(new_top_row - tp->topRow);
    if (num_exposed_rows > tp->numLines)
        num_exposed_rows = tp->numLines;
    else
        num_exposed_rows++; /* KLUDGE, to account for partially visible
                               last row -- last row fix */
    /*  if (num_exposed_rows >= (tp->numLines - 1)) */
    if (num_exposed_rows >= tp->numLines) { /* last row fix */
        /* results in a page scroll */
        if (tp->use_data_request_model) {
			data_ptr top_row_data;
            /* trap data request model */
			/* hold onto the top_row_data */
			top_row_data = AxXmGetTableRows(w, 	AxMakeIntData(new_top_row),
												AxMakeIntData(1));
            TblClearData(w, 0, -1); /* clear whole table */
			/* set the top_row_data back */
			TblInsertNewData(w, top_row_data, new_top_row);
			AxTaskFreeElfData(THIMpid, top_row_data);
            tp->topRow = new_top_row;
            PaintTable(w, NULL, FALSE, FALSE);
            TblUpdateScrollBars(w);
            TblSendDataRequestEvent(w, new_top_row, num_exposed_rows);
            return(0);
        }
        tp->topRow = new_top_row;
        TblUpdCursorCellInfo(tp);
        TblUpdateScrollBars(w);
        PaintTable(w, NULL, FALSE, FALSE);
        return(0);
    }

    if (new_top_row > tp->topRow) { /* shift up */
        src_y = YPosOfCell(tp, new_top_row, 0);
        src_x = 0;
        dst_x = src_x;
        dst_y = tp->headingHeight;
        width  = tp->winWidth;
        height = tp->winHeight - src_y;

        AxSetWidgetWorkColors((AXWidget)w);
#if AXMS
        Overlapped_XCopyArea( w->window, BackGC, src_x, src_y, width, 
           height, dst_x, dst_y, w->parent->parent->winid, w->winid, FALSE );
#else
        TblXCopyArea(Dpy, w->window, w->window, BackGC, src_x, src_y, width,
                     height, dst_x, dst_y);
#endif
        clip_x_origin = src_x;
        clip_y_origin = dst_y + height;
        crect[0].x = 0;
        crect[0].y = 0;
        crect[0].width  = width;
        crect[0].height = (tp->winHeight - dst_y);
        AxUnsetWidgetWorkColors((AXWidget)w);
        start_col = tp->leftCol;
        end_col   = -1;
        start_row = new_top_row + tp->numLines - num_exposed_rows;
        end_row   = -1;
    } else { /* shift down */
        src_x = 0;
        dst_x = src_x;
        src_y = tp->headingHeight;
        dst_y = src_y + ((tp->topRow - new_top_row) * tp->rowHeight);
        width  = tp->winWidth;
        height = tp->winHeight - dst_y;

        AxSetWidgetWorkColors((AXWidget)w);
#if AXMS
        Overlapped_XCopyArea( w->window, BackGC, src_x, src_y, width, 
           height, dst_x, dst_y, w->parent->parent->winid, w->winid, FALSE );
#else
        TblXCopyArea(Dpy, w->window, w->window, BackGC, src_x, src_y,
                     width, height, dst_x, dst_y);
#endif
        clip_x_origin = src_x;
        clip_y_origin = src_y;
        crect[0].x = 0;
        crect[0].y = 0;
        crect[0].width  = width;
        crect[0].height = (dst_y - tp->headingHeight);
        AxUnsetWidgetWorkColors((AXWidget)w);
        start_row = new_top_row;
        end_row   = new_top_row + num_exposed_rows;
        start_col = tp->leftCol;
        end_col   = -1;
    }

    /* FIXME, Clip out excess data if using data request model here! */

    tp->topRow = new_top_row;
    TblUpdateScrollBars(w);
    TblUpdCursorCellInfo(tp);

    /* Fill in contents of exposed region */
    TblSetClipAllGCs(w, clip_x_origin, clip_y_origin, crect, 1, YXBanded);

    if (tp->use_data_request_model) {
        TblClearData(w, start_row, num_exposed_rows);
    }

    PaintTableRegion(w, start_row, end_row, start_col, end_col);

    if (tp->use_data_request_model)
        TblSendDataRequestEvent(w, start_row, num_exposed_rows);

    TblUnsetClipAllGCs(w);
    
    if (tp->cursor.col != -1) { /* FIXME, use `tp->doing_cell_editing' */
        TblPutSelectionCursor(w);
    }
    
    return(0);
}

/***********************************************************************
function: TblScrollNewLeftCol - Handle table scroll so as to make
NEW_LEFT_COL the new left column.
Return: 0 on success; -1 on otherwise.
*/
static int
TblScrollNewLeftCol(w, new_left_col)
register AXWidget_table w;
int new_left_col;
{
    register table_info *tp;
    
    if (!w)
        return(-1);
    
    if (w->type != W_TABLE)
        return(-1);
    if (!(tp = (table_info *) w->info.table))
        return(-1);

    if (new_left_col < 0)
        new_left_col = 0; /* clip */

    if (new_left_col == tp->leftCol)
        return(0); /* nothing happens */
    TblLiftSelectionCursor(w);
    tp->leftCol = new_left_col;
    TblUpdCursorCellInfo(tp);
    PaintTable(w, NULL, FALSE, FALSE);
    TblUpdateScrollBars(w);
    return(0);
}

/***********************************************************************
function: TblForwardColumn - Goto the next column in the currently
edited row.
Note: Expect's a valid AXWidget_table pointer, W.
*/
static void
TblForwardColumn(w)
register AXWidget_table w;
{
    register table_info *tp = (table_info *) w->info.table;
    int max_cols;
    int row_idx, col_idx;

    if (!tp)
        return;
    max_cols = AxArraySize(tp->colText);

    if ((tp->cursor.row < 0) || (tp->cursor.col < 0)) {
        if (TblGotoCell(w, 0, 0, -1, -1) != 0) {
            RingTheBell();
        }
    } else {
        row_idx = tp->cursor.row;
        col_idx = tp->cursor.col + 1;
        if (col_idx >= max_cols) {
            if ( tp->rowcolcreate ) {
                if ( XPosOfCell(tp, row_idx, col_idx) >= tp->winWidth) {
                    int exposed_cols = TableComputeExposedCols(w);
                    tp->leftCol = col_idx;
                    TblRowColCreate(w, 0, exposed_cols, exposed_cols);
                }
            }
            return;
        } else {
            if (TblGotoCell(w, row_idx, col_idx, -1, -1) != 0) {
                RingTheBell();
                return;
            }
        }
    }
    return;
}

/***********************************************************************
function: TblBackwardColumn - Goto the previous column.
Note: Expect's valid AXWidget_table pointer, W.
*/
static void
TblBackwardColumn(w)
AXWidget_table w;
{
    register table_info *tp = (table_info *) w->info.table;
    int row_idx, col_idx;

    if ((tp->cursor.col < 0) || (tp->cursor.row < 0)) {
        if (TblGotoCell(w, 0, 0, -1, -1) != 0)
            RingTheBell();
        return;
    }
    row_idx = tp->cursor.row;
    col_idx = tp->cursor.col - 1;
    if (col_idx < 0) {
        RingTheBell();
        return;
    }
    if (TblGotoCell(w, row_idx, col_idx, -1, -1) != 0) {
        RingTheBell();
        return;
    }
    return;
}

/***********************************************************************
function: TblForwardRow - Goto the next row, holding the column
constant.
Note: Expect's a valid AXWidget_table pointer, W.
*/
static void
TblForwardRow(w)
register AXWidget_table w;
{
    register table_info *tp = (table_info *) w->info.table;
    int max_rows;
    int row_idx, col_idx, topRow;

    if (!tp)
        return; /* invalid args. */
    max_rows = AxArraySize(tp->rows);

    if ((tp->cursor.row < 0) || (tp->cursor.col < 0)) {
        if (TblGotoCell(w, 0, 0, -1, -1) != 0) {
            RingTheBell();
            return;
        }
    } else {
        row_idx = tp->cursor.row + 1;
        col_idx = tp->cursor.col;
        if (row_idx >= max_rows) {
            if ( tp->rowcolcreate ) {
				topRow = MAX(0, tp->topRow + 1);
				tp->topRow = topRow;
				TblRowColCreate(w, 1, 0, 0);
				TblUpdateScrollBars(w);
				TblScrollNewTopRow(w, tp->topRow);
				PaintTable(w, NULL, FALSE, FALSE); 
            }
			if (TblGotoCell(w, row_idx, col_idx, -1, -1) != 0) {
				RingTheBell();
				return;  /* can't go any further. */
			}
			return;  /* can't go any further. */
        }
		else if ( tp->cursor.row+1 < tp->topRow + tp->numLines ) {
			if ( tp->rowcolcreate ) {
				TblUpdateScrollBars(w);
				TblScrollNewTopRow(w, tp->topRow);
				PaintTable(w, NULL, FALSE, FALSE); 
			}
		}
			
        if (TblGotoCell(w, row_idx, col_idx, -1, -1) != 0) {
            RingTheBell();
            return;
        }
    }
    return;
}

/***********************************************************************
function: TblPreviousRow - Goto the previous row, holding the column
constant.
Note: Expect's a valid AXWidget_table pointer, W.
*/
static void
TblPreviousRow(w)
register AXWidget_table w;
{
    register table_info *tp = (table_info *) w->info.table;
    int row_idx, col_idx;
	int top_row, valid_row;

    if (!tp)
        return; /* invalid args. */
    if ((tp->cursor.row < 0) || (tp->cursor.col < 0)) {
        if (TblGotoCell(w, 0, 0, -1, -1) != 0) {
            RingTheBell();
            return;
        }
    }
    row_idx = tp->cursor.row - 1;
    col_idx = tp->cursor.col;
    if (row_idx < 0)
        return;
	if ( tp->rowcolcreate && (row_idx == tp->topRow-1 ) )
	{
		top_row = tp->topRow - 1;
		valid_row = TblGetLastValidRow(w); 
		if ( top_row + tp->numLines > tp->totalRowCount-1); {
			if ( top_row + tp->numLines > valid_row)
				TblDeleteRowOfData(w, 1);
		}
		tp->topRow = top_row;
		TblUpdateScrollBars(w);
		TblScrollNewTopRow(w, tp->topRow);
		PaintTable(w, NULL, FALSE, FALSE); 		
	}
    if (TblGotoCell(w, row_idx, col_idx, -1, -1) != 0) {
        RingTheBell();
        return;
    }
    return;
}

/***********************************************************************
function: TblBackwardChar - Move cursor back by one character.
*/
static void
TblBackwardChar(w)
register AXWidget_table w;
{
    register table_info *tp;

    if (!w)
        return;
    if (w->type != W_TABLE)
        return;
    if (!(tp = (table_info *) w->info.table))
        return;
    if (!tp->cursor.wtext)
        return;

    if ((tp->cursor.char_pos - 1) >= 0) {
        TblLiftSelectionCursor(w);
        tp->cursor.char_pos--;
        if (TblMakeCursorCharVisible(w, tp->cursor.char_pos, 0))
            TblDrawCursorCellText(w);
        TblPutSelectionCursor(w); /* put a new one. */
    } else
        RingTheBell();

    return;
}

/***********************************************************************
function: TblForwardChar - Move cursor forward by one character.
*/
static void
TblForwardChar(w)
register AXWidget_table w;
{
    register table_info *tp;
    int max_char_pos;

    if (!(tp = (table_info *) w->info.table))
        return;
    if (!(tp->cursor.wtext))
        return;

    max_char_pos = tp->cursor.text_len;

    if ((tp->cursor.char_pos + 1) > max_char_pos) {
        RingTheBell(); /* reached limit on max chars in this cell. */
    } else {
        TblLiftSelectionCursor(w);
        tp->cursor.char_pos++;
        if (TblMakeCursorCharVisible(w, tp->cursor.char_pos, 0))
            TblDrawCursorCellText(w);
        TblPutSelectionCursor(w);
    }
    return;
}

/***********************************************************************
function: TblInsertText - Insert text at cursor.  Return TRUE if any text
change happened; FALSE otherwise.
*/
static int
TblInsertText(w, wbuf, num_chars)
register AXWidget_table w;
register wchar *wbuf;
register int num_chars;
{
    wchar *tmpwbuf = (wchar *) NULL;
    table_info *tp;
    int text_len = 0;
    int tmp_idx = 0;
    int new_size;

    if (!w)
        return(FALSE);
    if (!(tp = (table_info *) w->info.table))
        return(FALSE);
    if (num_chars <= 0)
        return(FALSE);
    if (!wbuf)
        return(FALSE);

    if ((tp->cursor.row < 0) || (tp->cursor.col < 0)) {
        RingTheBell(); /* cursor not inside any cell. */
        return(FALSE);
    }
    if ((tp->cursor.row < tp->topRow) ||
        (tp->cursor.col < tp->leftCol)) {
        RingTheBell(); /* cursor not visible!! */
        return(FALSE);
    }

/* TODO, check if we can take in `real_len' more characters.  If not so,
   beep and return. */

    TblLiftSelectionCursor(w);

    if (tp->cursor.wtext)
        text_len = tp->cursor.text_len;
    else
        text_len = 0;

    new_size = text_len + num_chars;
    tmpwbuf = (wchar *)malloc((new_size + 1)*sizeof(wchar));
    if (!tmpwbuf) {
        RingTheBell();
        goto bad_ret;
    }

    /* Insert text into `tmpwbuf'. */
    /* re-written - jon */
    if (tp->cursor.wtext == (wchar *) NULL) {
        /* no current text, the surff to be insert is the whole new string  */
        Wstrncpy(tmpwbuf, wbuf, num_chars);
        tmpwbuf[num_chars] = 0;
    } else {
        /* there is already some text, insert it */
        Wstrncpy(tmpwbuf, tp->cursor.wtext, tp->cursor.char_pos);
        Wstrncpy(&tmpwbuf[tp->cursor.char_pos], wbuf, num_chars);
        Wstrcpy(&tmpwbuf[(tp->cursor.char_pos+num_chars)],
                           (tp->cursor.wtext+tp->cursor.char_pos));
    }
    tmpwbuf[new_size] = 0;

    /* Replace `cursorText' with `tmpwbuf'. */
    if (tp->cursor.wtext)
        free(tp->cursor.wtext); /* is valid pointer */
    tp->cursor.wtext = tmpwbuf;

    /* Compute new cursor position. */
    tp->cursor.char_pos += num_chars;
    tp->cursor.text_len += num_chars;

    /* Make cursor visible */
    (void) TblMakeCursorCharVisible(w, tp->cursor.char_pos, 0);
    /* Draw contents of cell. */
    TblDrawCursorCellText(w);

    TblUpdCursorCellInfo(tp);
    TblPutSelectionCursor(w);
    return(TRUE);

bad_ret:
    TblLiftSelectionCursor(w);
    return(FALSE);
}

/***********************************************************************
function: TblBackwardCharErase - Erase the character left of the cursor.
*/
void
TblBackwardCharErase(w)
register AXWidget_table w;
{
    register table_info *tp;
    register int ix;
    register int cnt;
    wchar *new_text_buf = (wchar *) NULL;
    int new_text_len;
    int new_char_pos;

    if (!w)
        return;
    if (!(tp = (table_info *) w->info.table))
        return;
    
    if ((tp->cursor.row < 0) || (tp->cursor.col < 0)) {
        RingTheBell(); /* cursor not inside any cell. */
        return;
    }
    if (tp->cursor.row < tp->topRow) {
        RingTheBell(); /* cursor not visible!! */
        return;
    }

    if ((!tp->cursor.wtext) || (tp->cursor.text_len <= 0) ||
        (tp->cursor.char_pos == 0)) {
        RingTheBell();
        return;
    }

    if ((tp->cursor.char_pos - 1) < 0) {
        RingTheBell();
        return; /* at the beginning of line. */
    }

    TblLiftSelectionCursor(w);

    new_char_pos = tp->cursor.char_pos - 1;
    new_text_len = tp->cursor.text_len - 1;

    if (new_text_len <= 0) {
        if (tp->cursor.wtext)
            free(tp->cursor.wtext);
        TblResetCellTextInfo(tp);
    } else {
        new_text_buf = (wchar *)malloc((new_text_len+1)*sizeof(wchar));
        if (!new_text_buf) {
            RingTheBell();
            goto bad_ret;
        }
        for (ix = 0, cnt = 0; ix < tp->cursor.text_len; ix++) {
            if (ix == new_char_pos)
                continue;
            new_text_buf[cnt] = tp->cursor.wtext[ix];
            cnt++;
        }
        new_text_buf[cnt] = 0;
        if (tp->cursor.wtext)
            free(tp->cursor.wtext); /* is valid pointer */
        tp->cursor.wtext = new_text_buf;
        tp->cursor.char_pos = new_char_pos;
        tp->cursor.text_len = new_text_len;
    }
    (void) TblMakeCursorCharVisible(w, tp->cursor.char_pos, 0);
    TblDrawCursorCellText(w);
    TblPutSelectionCursor(w);
    return;

bad_ret:
    TblPutSelectionCursor(w);
    return;
}

/***********************************************************************
function: TblEraseChar - Erase character at cursor.
*/
void
TblEraseChar(w)
register AXWidget_table w;
{
    register table_info *tp;
    register int ix;
    register int cnt;
    int new_char_pos;
    int new_text_len;
    wchar *new_text_buf = (wchar *) NULL;

    if (!w)
        return;
    if (!(tp = (table_info *) w->info.table))
        return;
    if ((tp->cursor.row < 0) || (tp->cursor.col < 0))
        goto bailout;
    if (tp->cursor.row < tp->topRow)
        goto bailout;
    if ((!tp->cursor.wtext) || (tp->cursor.text_len <= 0))
        goto bailout;

    if (tp->cursor.char_pos == tp->cursor.text_len)
        goto bailout; /* cursor at end of line. */

    TblLiftSelectionCursor(w);

    new_char_pos = tp->cursor.char_pos;
    new_text_len = tp->cursor.text_len - 1;
    if (new_text_len <= 0) {
        if (tp->cursor.wtext)
            free(tp->cursor.wtext);
        TblResetCellTextInfo(tp);
    } else {
        new_text_buf = (wchar *)malloc((new_text_len+1)*sizeof(wchar));
        if (!new_text_buf)
            goto bailout;
        for (ix = 0, cnt = 0; ix < tp->cursor.text_len; ix++) {
            if (ix == new_char_pos)
                continue;
            new_text_buf[cnt] = tp->cursor.wtext[ix];
            cnt++;
        }
        new_text_buf[new_text_len] = 0;
        if (tp->cursor.wtext)
            free(tp->cursor.wtext);
        tp->cursor.wtext = new_text_buf;
        tp->cursor.char_pos = new_char_pos;
        tp->cursor.text_len = new_text_len;
    }
    (void) TblMakeCursorCharVisible(w, tp->cursor.char_pos, 0);
    TblDrawCursorCellText(w);
    TblPutSelectionCursor(w);
    return;

bailout:
    TblPutSelectionCursor(w);
    RingTheBell();
    return;
}

/***********************************************************************
function: TblSendSelectionChangeEvent - Send a selection change event
to widget owner.  Message contains list of currently selected rows in
the table.
Note:
Selections can change when:
- Button Up happens during row marking.
- A GotoCell takes place.
*/
static void
TblSendSelectionChangeEvent(w)
AXWidget_table w;
{
    register table_info *tp;
    register int ix;
    register int tempint;
    register int num_selections;
    int *row_list = (int *) NULL;
    TblEventData data;

    if (!w)
        return;
    if (!w->callback)
        return;
    if (!(tp = (table_info *) w->info.table))
        return;

    data.selection_change.type = EV_TBL_SELECTION_CHANGE_;
    num_selections = IntsArraySize(tp->selected_rows);
    if (num_selections) {
        row_list = (int *) malloc(num_selections * sizeof(int));
        for (ix = 0; ix < num_selections; ix++) {
            tempint = ReadIntsArray(tp->selected_rows, ix);
            row_list[ix] = tempint;
        }
    }
    data.selection_change.num_rows = num_selections;
    data.selection_change.row_list = row_list;
    AxCallback(w->callback, (AXWidget)w, w->callbacktag, (caddr_t)&data);
    if (row_list)
        free(row_list);
    return;
}

/***********************************************************************
function: TblSendDoubleClickEvent - Send a double click event to
widget owner.  ROW_NUM is the row to report the double click on.
*/
static void
TblSendDoubleClickEvent(w, row_num, col_num)
AXWidget_table w;
int row_num;
int col_num;
{
    TblEventData data;

    if (!w)
        return;
    if (!w->callback)
        return;

    data.double_click.type = EV_TBL_DBLCLICK_;
    data.double_click.row  = row_num;
    data.double_click.col = col_num;
    AxCallback(w->callback, (AXWidget)w, w->callbacktag, (caddr_t)&data);
    return;
}

/***********************************************************************
function: TblSendColumnResizeEvent - Send a column resize event to
widget owner.  COL is the column number that got resized, COL_WIDTH is
the new width of the column.
*/
static void
TblSendColumnResizeEvent(w, col, col_width)
AXWidget_table w;
int col;
int col_width;
{
    TblEventData data;
    if (!w)
        return;
    if (!w->callback)
        return;

    data.column_resize.type = EV_TBL_COLUMN_RESIZE_;
    data.column_resize.column = col;
    data.column_resize.width  = col_width;
    AxCallback(w->callback, (AXWidget)w, w->callbacktag, (caddr_t)&data);
    return;
}

/***********************************************************************
function: TblSendDataRequestEvent - Send a data-request event to
widget owner. Caller requests NUM_ROWS number of rows of data starting
at row-index START_ROW.
*/
static void
TblSendDataRequestEvent(w, start_row, num_rows)
AXWidget_table w;
int start_row;
int num_rows;
{
    TblEventData data;
    if (!w)
        return;
    if (!w->callback)
        return;
    data.request_data.type = EV_TBL_REQUEST_DATA_;
    data.request_data.starting_row = start_row;
    data.request_data.num_rows = num_rows;
    AxCallback(w->callback, (AXWidget)w, w->callbacktag, (caddr_t)&data);
    return;
}

/***********************************************************************
function: TblSendKeyPressEvent - Send a KeyPress event up to the
widget owner.
ELFKEYID  - The elf key id as defined in evreason.h
PRINTABLE - if TRUE, key is printable; FALSE means non-printable
WBUF      - If PRINTABLE, this holds the character(s) typed.
NUM_CHARS - Number of characters contained in WBUF.
*/
static void
TblSendKeyPressEvent(w, elfkeyid, printable, wbuf, num_chars, row, col)
AXWidget_table w;
int elfkeyid;
int printable;
wchar *wbuf;
int  num_chars;
int row;
int col;
{
    TblEventData data;
    wchar *tmp_wbuf = (wchar *) NULL;

    if (!w)
        return;
    if (!w->callback)
        return;
    data.keypress.type = EV_TBL_KEYPRESS_;
    data.keypress.elfkeyid = elfkeyid;
    data.keypress.printable = printable;
    data.keypress.row = row;
    data.keypress.col = col;
    if (wbuf && num_chars) {
        tmp_wbuf = (wchar *) calloc(1, sizeof(wchar) * num_chars);
        Wstrncpy(tmp_wbuf, wbuf, num_chars);
        data.keypress.wbuf = tmp_wbuf;
        data.keypress.num_chars = num_chars;
    } else {
        data.keypress.wbuf = (wchar *) NULL;
        data.keypress.num_chars = 0;
    }
    AxCallback(w->callback, (AXWidget)w, w->callbacktag, (caddr_t)&data);
    if (tmp_wbuf)
        free(tmp_wbuf);
    return;
}

/***********************************************************************
function: TblSendButtonPressEvent - Send a button-press event to the
widget owner.  Only ButtonPress events in the "white" area are reported.
*/
static void
TblSendButtonPressEvent(w, row, col, start_char_pos, end_char_pos)
AXWidget_table w;
int row;
int col;
int start_char_pos; /* 1 based */
int end_char_pos;   /* 1 based */
{
    TblEventData data;

    if (!w)
        return;
    if (!w->callback)
        return;
    data.buttonpress.type = EV_TBL_BUTTONPRESS_;
    data.buttonpress.row  = row;
    data.buttonpress.col  = col;
    data.buttonpress.start_char_pos = start_char_pos;
    data.buttonpress.end_char_pos   = end_char_pos;
    AxCallback(w->callback, (AXWidget)w, w->callbacktag, (caddr_t)&data);
    return;
}

/***********************************************************************
function: TblSendEditBeginEvent - Send a edit-begin request to widget
owner.  Notify widget owner that the first KeyPress has occurred in
the cell at ROW,COL.
*/
static void
TblSendEditBeginEvent(w, row, col)
AXWidget_table w;
int row;
int col;
{
    TblEventData data;
    if (!w)
        return;
    if (!w->callback)
        return;
    data.edit_begin.type = EV_TBL_EDIT_BEGIN_;
    data.edit_begin.row  = row;
    data.edit_begin.col  = col;
    AxCallback(w->callback, (AXWidget)w, w->callbacktag, (caddr_t)&data);
    return;
}

/***********************************************************************
function: TblSendTypingEvent - Send a typing request to widget owner.
Notify widget owner the row and col where typing is occuring.

This event is sent only if db_ctrl_typing_return@() is true.
*/
static void
TblSendTypingEvent(w, row, col)
AXWidget_table w;
int row;
int col;
{
    TblEventData data;

    if (!w)
        return;
    if (!w->callback)
        return;
    if (!(w->flags & WFLG_TOUCHED_RETURN))
        return;
    if (row == -1 || col == -1)
        return;
    data.cell_typing.type = EV_TYPING_;
    data.cell_typing.row = row;
    data.cell_typing.col = col;
    AxCallback(w->callback, (AXWidget)w, w->callbacktag, (caddr_t)&data);
    return;
}

/***********************************************************************
function: TblSendFocusChangeEvent - Send a focus-change event to the widget
owner.  This tells focus has changed from row,col to row,col
*/
static void
TblSendFocusChangeEvent(w, from_row, from_col, to_row, to_col)
AXWidget_table w;
int from_row;
int from_col;
int to_row;
int to_col;
{
    TblEventData data;

    if (!w)
        return;
    if (!w->callback)
        return;
    if (!(w->flags & WFLG_ACTIVE_RETURN))
        return;
    data.focus_change.type = EV_TBL_FOCUS_CHANGE_;
    data.focus_change.from_row = from_row;
    data.focus_change.from_col = from_col;
    data.focus_change.to_row = to_row;
    data.focus_change.to_col = to_col;
    AxCallback(w->callback, (AXWidget)w, w->callbacktag, (caddr_t)&data);
    return;
}

/* Set a bunch of generic table attributes */
void SetTableAttributes( in_w, att_code, value )
AXWidget in_w;
int att_code;
int value;
{
    AXWidget_table w;
    table_info *tp;

    if ( in_w->type != W_TABLE )
        return;

    w = (AXWidget_table)in_w;
    tp = (table_info *)w->info.table;
    if ( tp == (table_info *)NULL )
        return;

    /* we have a table widget, and there is an info block, so */
    /* we can set the value */
    switch( att_code )
        {
        case XXmNmarkerWidth:
                tp->markerWidth = value;
            break;
        case XXmNallowHiddenColumns:
                tp->allow_hidden_columns = value ? 1 : 0;
            break;
        case XXmNallowColumnResizing:
                tp->allow_column_resizing = value ? 1 : 0;
            break;
        case XXmNsuppressRowMarkers:
                tp->suppress_row_markers = value ? 1 : 0;
            break;
        case XXmNallowColumnReordering:
                tp->allow_column_reordering = value ? 1 : 0;
            break;
        case XXmNhideVerticalGrid:
                tp->hide_vertical_grid = value ? 1 : 0;
            break;
        case XXmNhideHorizontalGrid:
                tp->hide_horizontal_grid = value ? 1 : 0;
            break;
        case XXmNuseDataRequestModel:
                tp->use_data_request_model = value ? 1 : 0;
            break;
        case XXmNalisStyleScroll:
                tp->alis_style_scroll =  value ? 1 : 0;
            break;
        case XXmNRowColCreate:
                tp->rowcolcreate =  value ? 1 : 0;
            break;
        case XXmNvscrollOrigin:
            tp->vScrollOrigin = value;
            tp->vscroll_dim_changed = TRUE;
            WidgetNeedsUpdated = TRUE;
            w->update_widget = TRUE;
            if (w->vScrollBar)
                w->vScrollBar->update_widget = TRUE;
            break;
        case XXmNvscrollLength:
            tp->vScrollLength = value;
            tp->vscroll_dim_changed = TRUE;
            WidgetNeedsUpdated = TRUE;
            if (w->vScrollBar)
                w->vScrollBar->update_widget = TRUE;
            w->update_widget = TRUE;
            break;
        case XXmNhscrollOrigin:
            tp->hScrollOrigin = value;
            tp->hscroll_dim_changed = TRUE;
            WidgetNeedsUpdated = TRUE;
            if (w->hScrollBar)
                w->hScrollBar->update_widget = TRUE;
            w->update_widget = TRUE;
            break;
        case XXmNhscrollLength:
            tp->hScrollLength = value;
            tp->hscroll_dim_changed = TRUE;
            WidgetNeedsUpdated = TRUE;
            if (w->hScrollBar)
                w->hScrollBar->update_widget = TRUE;
            w->update_widget = TRUE;
            break;
        case XXmNshowRowNumbers:
            if (value == BTRUE)
                tp->showRowNumbers = TRUE;
            else
                tp->showRowNumbers = FALSE;
            break;
        }
    return;
}

/***********************************************************************
function: TblXCopyArea - XCopyArea() wrapper for the table widget
*/
static void
TblXCopyArea(dpy, src_drawable, dst_drawable, gc, src_x, src_y, width, height,
             dst_x, dst_y)
register Display *dpy;
Drawable src_drawable, dst_drawable;
GC  gc;
int src_x, src_y;
unsigned int width, height;
int  dst_x, dst_y;
{
    XGCValues    orig_values;
    unsigned int valuemask;
    int changed = FALSE;

    valuemask = GCGraphicsExposures;
    XGetGCValues(dpy, gc, valuemask, &orig_values);
    if (orig_values.graphics_exposures == FALSE) {
        XSetGraphicsExposures(dpy, gc, TRUE);
        changed = TRUE;
    }
    XCopyArea(dpy, src_drawable, dst_drawable, gc, src_x, src_y,
              width, height, dst_x, dst_y);
    if (changed)
        XSetGraphicsExposures(dpy, gc, orig_values.graphics_exposures);
    return;
}

/************************
function: TblTotalRows - return # rows currently in table data.
                         this is often different from # rows on display
*/
static int
TblTotalRows(tp)
table_info *tp;
{
    if (!tp)
        return 0;
    
    if (tp->totalRowCount != -1) {
        return (tp->totalRowCount);
    }
    else {
        if (tp->rows)
            return (AxArraySize(tp->rows));
        else
            return 0;
    }
}

static void
TblHilightFill(w, x, y, width, height, on)
AXWidget_table w;                         /* in */
int x, y, width, height;                  /* in: fill dimensions */
int on;                                   /* in: bool; true if hilit */
{
    GC tgc;
    
    if (w->useWorkColors) {
        AxSetWidgetWorkColors((AXWidget)w);
    }
    
    tgc = on ? ForeGC : BackGC;
    
    XFillRectangle(Dpy, w->window, tgc, x, y, width, height);
    
    if (w->useWorkColors) {
        AxUnsetWidgetWorkColors((AXWidget)w);
    }
}

static void
TblDrawHilitText(w, xpos, ypos, str, len, clip)
AXWidget_table w;
int xpos, ypos;
wchar *str;
int len;
r32 *clip;
{
    if (w->useWorkColors) {
        AxSetWidgetWorkColors((AXWidget)w);
    }
    
    if (clip) {
        TblClip(BackGC, clip->origin.x, clip->origin.y,
                clip->width, clip->height);
    }
    
    xmDrawString16((AXWidget)w, w->window, BackGC, xpos, ypos, str, len,
                   (data_ptr) NULL, (wcolor_t *) NULL, (wcolor_t *) NULL);
    
    if (clip) {
        TblRestoreClip(BackGC);
    }
    
    if (w->useWorkColors) {
        AxUnsetWidgetWorkColors((AXWidget)w);
    }
}

/* : TblCheckMarkerPixmaps - check the datatyping of the passed
 * elf array.  This is done inside elf-builtin call so that a
 * valid error throw can occur here; this array will be copied
 * onto table_info_t and read at paint time (task 0) -- mustn't
 * allow error throw to occur there.
 *
 * expected data: an array num_rows in size; each array element
 * is an array of strings num_cols in size.
 */
static void
TblCheckMarkerPixmaps(dp)
data_ptr dp;
{
    if (AxIsArray(dp)) {
        /* spin thru the data for possible error throw
         */
        char *ctmp;
        int numPix, rx, px, asize;
        data_ptr dtmp;
        
        asize = AxArraySize(dp);
        for (rx = 0; rx < asize; rx++) {
            dtmp = AxArrayElement(dp, rx);
            numPix = AxArraySize(dtmp);
            for (px = 0; px < numPix; px++) {
                ctmp = AxStrPtrFromArray(dtmp, px);
            }
        }
    }
}


/* : expected data: an array num_rows in size; each element is a string
 */
static void
TblCheckCornerPixmaps(dp)
data_ptr dp;
{
    int asize, ix;
    char *ctmp;
    
    /* spin thru the data for possible error throw
     */
    asize = AxArraySize(dp);
    for (ix = 0; ix < asize; ix++) {
        ctmp = AxStrPtrFromArray(dp, ix);
    }
}

/* This is similar to TblCheckMarkerPixmaps
 */
static void
TblCheckMarkerStrings(dp)
data_ptr dp;
{
    if (AxIsArray(dp)) {
        /* spin thru the data for possible error throw         */
        char *ctmp;
        int rx, asize;
        
        asize = AxArraySize(dp);
        for (rx = 0; rx < asize; rx++) {
            ctmp = AxStrPtrFromArray(dp, rx);
        }
    }
}

/* : TblCheckRowData - check that passed row data is in the expected
 * format: an array num_rows in size; each element is an array of strings
 */
static void
TblCheckRowData(dp)
data_ptr dp;
{
    int asize, rowWid, ix, jx;
    char *ctmp;
    data_ptr dtmp;

    /* spin thru the data for possible error throw.
     * don't worry about whether each row is the same #elements wide,
     * only that each element is of string type.
     */
    asize = AxArraySize(dp);
    for (ix = 0; ix < asize; ix++) {
        dtmp = AxArrayElement(dp, ix);
        rowWid = AxArraySize(dtmp);
        for (jx = 0; jx < rowWid; jx++) {
            ctmp = AxStrPtrFromArray(dtmp, jx);
        }
    }
}

#if TBL_DEBUG
/***********************************************************************
function: TblLogMsg - Log a debugging message to stdout
*/
static void
TblLogMsg(va_alist)
va_dcl
{
    va_list argp;
    char *fmt;

    va_start(argp);
    fmt = va_arg(argp, char *); /*LOK pa*/
    vfprintf(stdout, fmt, argp);
    va_end(argp);
    fflush(stdout);
}
#endif



#if BUILD_ORIENT    /*  {  */

/***********************************************************************
function: TblReplaceText - replace the current oriental input text with
                           the new text
*/
void
TblReplaceText( w, new_wtext, add_len )
register AXWidget_table w;
wchar *new_wtext;
int add_len;
{
    register table_info *tp;
    register int ix;
    wchar *new_text_buf = (wchar *) NULL;
    int new_text_len;
    int new_char_pos;
    int change_len;
    wchar *start_replace;
    int i;
    int change_start;
    AXWidget_main main_w;

    if (!w)
        return;
    if (!(tp = (table_info *) w->info.table))
        return;
    
    if ((tp->cursor.row < 0) || (tp->cursor.col < 0)) {
        RingTheBell(); /* cursor not inside any cell. */
        return;
    }
    if (tp->cursor.row < tp->topRow) {
        RingTheBell(); /* cursor not visible!! */
        return;
    }

    TblLiftSelectionCursor(w);

    main_w = TheMainWidget( w );
    change_len = main_w->conv_change_end - main_w->conv_change_start;
    new_text_len = tp->cursor.text_len - change_len + add_len; 
    new_char_pos = main_w->conv_change_start + add_len;

    if ( change_len == add_len )
       {
       /* we are doing a character for character replacement, take the */
       /* fast route.  This type of operation happens fairly frequently */
       /* during the keyboard entry process                     */
       start_replace = tp->cursor.wtext + main_w->conv_change_start;
       for ( i=0; i<add_len; i++ )
          { *start_replace++ = *new_wtext++; }
       /* get the display updated */
       TblDrawCursorCellText(w);
       TblPutSelectionCursor(w);
       return;
       }

    /* Now we start the harder case, where the text we are to put in is, */
    /* not the same length as the current conversion area, which we will */
    /* take out                                                  */
    if ( main_w->conv_change_start < 0 )
       /* nothing to change */
       change_start = tp->cursor.char_pos;
    else
       change_start = main_w->conv_change_start;

    new_text_buf = (wchar *)malloc((new_text_len+1)*sizeof(wchar));
    if ( tp->cursor.wtext != (wchar *)NULL )
        {
        Wstrncpy( new_text_buf, tp->cursor.wtext, change_start );
        Wstrcat( new_text_buf, new_wtext );
        Wstrcat( new_text_buf, (tp->cursor.wtext+change_start+change_len));
        free(tp->cursor.wtext); /* is valid pointer */
        }
    else
        /* no current text, a new buffer */
        Wstrcpy( new_text_buf, new_wtext );

    tp->cursor.wtext = new_text_buf;
    tp->cursor.char_pos = change_start + add_len;
    tp->cursor.text_len = new_text_len;

    (void) TblMakeCursorCharVisible(w, tp->cursor.char_pos, 0);
    TblDrawCursorCellText(w);
    TblPutSelectionCursor(w);
    return;
}



#endif /*  }  BUILD_ORIENT    */




/* GNU Emacs settings - This (including the "\n^L") MUST be the last thing
 *                      in this file!
 * Local variables:
 * mode: c
 * tab-width: 4
 * end:
 */
