/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: pf01prim.c
   Applix ASC version: 139
   Date: Fri May 14 21:13:46 1999
*/
#include    <stdio.h>
#if AX_X11
#if !M_88K
#include    <varargs.h>
#endif
#endif

#include    <math.h>
#include    <aos.h>
#include    <wchars.h>
#include    <string.h>
#include    <pf.h>
#include    <paratm.h>
#include    <devsetup.h>
#include    <pfint.h>
#include    <sff.h>
#include    <fsrts.h>
#include    <pfextra.h>
#include    <pfpat.h>
#include    <getstr5.h>
#include    <getstr2.h>
#include    <axtext.h>
#include    <axbitmap.h>
#include    <revstamp.h>
#include    <axhtab.h>
#include    <axgfs.h>
#include    <pfsoft.h>

/*-----------------------------------------------------------------------------
    Types
-----------------------------------------------------------------------------*/

typedef struct {
    double x, y;
} dpoint;

typedef struct epi {
    TmPoint32 BBox[2];
    char *Title;
    char *Extensions;
    int LanguageLevel;
} eps_info;

typedef struct
{
    char    *psFontName;            /* e.g., Palatino-BoldItalic */
    char    *encodingName;          /* e.g., ISOLatin1Encoding */
    int     fontNumber;             /* taken from font counter D01FontNumber */
    int     numSizes;
    int     *sizes;                 /* varsized array of sizes */
    int     gallium ;
    unsigned int eightBit : 1;
    unsigned int wideChar : 1;      /* as in tx_char */
} printFontName;

struct PfCur_font
{
    int  number;
    int  size;
} Curfont;


/* Hey, future Hacker; don't use this struct as a general purpose
   color equivalence tool! (Only use what you've set...)
   */
typedef struct
{
    int gray_c;
    int cyan;
    int magenta;
    int yellow;
    int black;
    int red;
    int blue;
    int green;
} cmap_cache_entry;



#include    <pf01prim.p>


/*-----------------------------------------------------------------------------
    define
-----------------------------------------------------------------------------*/


#ifndef TRUE
#define FALSE   0
#define TRUE    1
#endif

/* size (in mils) of line style elements */
#define LONG_DASH           100
#define LONG_SPACE          70
#define SHORT_DASH          30
#define DOT                 10
#define DOT_SPACE           15  

#if 1
#define PF_MEM_CHECK
#else
#define PF_MEM_CHECK ThimMemCheck()
#endif


/* We can generate large numbers of points which break level 1 jobs
   with limitcheck.  The real answer is to partition the Line into
   several shorter lines (each less than the limit). How do we cope
   with background fills? clipping?
   */
#define MAX_PATH_POINTS     1450    /* arbitrary limit for PS lev. 1 */
#define MARK_MARGIN         125
#define SCREEN_RESOLUTION   75
#define PATTERN_RESOLUTION  300
#define RESOLUTION          1000
#define MINIMUM_LINE_WIDTH  4
#define NUMFDES             44
#define XTRA_XOFF           0      /* hardwired xoffset for the laserwriter */
#define MAX_PS_LINE         255
#define MAX_PS_STRING       65536   /* maximum string size */
#define COL_MODEL_GRAY      1


/* This should be Papertop - (y), until then... */
#if 1
#define Y(y)                (PF.pagetop - (y))
#else
static  int Y_func();
#define                     Y(y) Y_func(y)
#endif

#define Y_mil(y)            (PF.PfPageMilHeight - (y))
#define ABS(x)              (((x) > 0) ? (x) : -(x))
#define SIGN(x)             (((x) >= 0) ? 1 : -1)
#define COL_MODEL_CMYK      4
#define COL_MODEL_RGB       3

#define EllFactor           (.275)  /* for intermediate control points on ellipses */


/*-----------------------------------------------------------------------------
    local data
-----------------------------------------------------------------------------*/
static  char    NameOfThisSourceFile[] = __FILE__ ; /* for the assert code */

static  char    PSbox_needed = FALSE;
static  char    PScirc_needed = FALSE;
static  char    Pf_has_landscape = FALSE;
static  char    PScharpath_needed = FALSE;

static  char    *AxSymbolEncoding = NXLT("AxSymbolEncoding");

static  char    pf_color_model = COL_MODEL_CMYK;

static  int  numColorComponents = COL_MODEL_CMYK;   /* # components in dest (PS) image */
static  int  bitsPerColorComponent = 4;       /* # bits per color per pixel in dest */
static  int  bitsPerGrayComponent = 4;         /* # bits per pixel for grayscale image */

static	int	 banner_font_size = 14 ;
static  int  which_line = 0;
static  int  total_lines = 0;
static  int  which_text = 0;
static  int  total_texts = 0;
static  int  max_path_points = MAXINT;
static  int  duplex ;                   /* -1: none
                                                0: simplex
                                                1: duplex, long-edge
                                                2: duplex, short-edge
                                            */
static  int  duplex_page ;
static  int  input_tray ;
static  int  output_tray ;
static  int  page_started ;

static  struct {                            /* clipping at GS level */
        int  x1 ;
        int  y1 ;
        int  x2 ;
        int  y2 ;
        } current_clip[5] ;

PSinserts PfPsInsertData;          /* hooks for extra postscript */
static  char    *InsertSpec;
static  FILE    *PPDfile ;
static  axhtab  printFontEncodings;
static  axhtab  printFontNames;


/*-----------------------------------------------------------------------------
    external data
-----------------------------------------------------------------------------*/
#define TWICE_MAX_PATTERN      2048
char    pf70PatBuf[TWICE_MAX_PATTERN];
char    *pf70PatScannedPtr;
char    *pf70PatInputPtr;
int     pf70PatOffset;

extern  int  PfJobToPSPrinter;     /* if TRUE: job to printer, not file */
extern  int  TargetRes;
extern  char    *printer_fills[];

char    **PfEpsTitles;                /* Where do these go away?... */


/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
static int Y_func(y)
     int y;
{
    register int tmp_y;

    tmp_y = PF.pagetop - y;
#if 0
    if (tmp_y > PF.pagetop)
        tmp_y = PF.pagetop;
    else
        if (tmp_y < 0)
            tmp_y = 0;
#endif
    return(tmp_y);
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
static void PS_GS()
{
    int n ;

    put(NXLT("GS\n"));
    n = ++PF.ClipGSaveLevel ;
    current_clip[n] = current_clip[n-1] ;

}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
static void PS_GR()
{
    put(NXLT("GR\n"));
    PF.ClipGSaveLevel-- ;
}

/*-----------------------------------------------------------------------------
    varargs function for formatting and outputing postscript
    This should be the one function used for writing any postscript text.
-----------------------------------------------------------------------------*/
static void put(va_alist)
va_dcl
{
    va_list argp;
    char *format;

    va_start(argp);
    format = va_arg(argp, char *);
    if (format == (char *)NULL)
    {
            /* last put call for the printjob; flush */
        if (PfPsInsertData.searchPatterns.numPatterns && pf70PatOffset > 0)
        {
            PfOutCString(pf70PatBuf, pf70PatOffset);
        }
        return;
    }

    vsprintf(pf70PatInputPtr, format, argp);

    va_end(argp);
    if (PfPsInsertData.searchPatterns.numPatterns)
    {
            /* There are some text patterns to search for
             * in the formatted output.  Filter it through
             * prSearch before calling the lowlevel output. */
        char *fp, *cp;
        int fplen, cplen, adv;
        int prSearch();
        searchPat *gp = &PfPsInsertData.searchPatterns;

        while (adv = prSearch(gp, pf70PatScannedPtr,
                             &fp, &fplen,
                             &cp, &cplen))
        {
            pf70PatScannedPtr += adv;
            if (fp)
                PfOutCString(fp, fplen);
            if (cp)
                PfOutCString(cp, cplen);
        }
        if (pf70PatOffset = gp->offset)
        {
                /* not enough data left to judge whether a match
                 * will occur; save remainder for next put */
            memcpy(pf70PatBuf, pf70PatScannedPtr, gp->offset);
        }
        pf70PatScannedPtr = pf70PatBuf;
        pf70PatInputPtr = pf70PatBuf + pf70PatOffset;
    }
    else
    {
        PfOutString(pf70PatInputPtr);
    }
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
static void put_string(ttext)
    char *ttext ;
{
    char *s = ttext ;
    char *t ;
    int i ;

    while (strlen(s) > 80)
    {
        for (i = 76 ; i <= 79 ; i++)
            if (s[i] == '\\')
                break ;
        if (i == 80)
        {
            put("%.80s\\\n", s) ;
            s += 80 ;
        }
    else
        /* can't break between \ and its escaped char */
        {
            if (*s == '\\' && *(s+1) == '\\')
                t = strchr(s+2, '\\') ;
            else if (*s == '\\')
                t = strchr(s+1, '\\') ;
            else
                t = strchr(s, '\\') ;

            if (strlen(t) < 2)
                break ;
            *t = 0 ;
            put("%s\\\n", s) ;
            s = t ;
            *s = '\\' ;
        }
    }
    if (strlen(s) > 0)
        put("%s", s) ;
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
static void put_X_text(ttext, d)
    char *ttext ;
    int d ;
{
    put(NXLT("(")) ;
    put_string(ttext) ;
    put(NXLT(") %d X\n"), d) ;
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
static void put_Xc_text(ttext)
    char *ttext ;
{
    put(NXLT("(")) ;
    put_string(ttext) ;
    put(NXLT(") Xc\n")) ;
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
static void put_ST_text(ttext)
    char *ttext ;
{
    put(NXLT("(")) ;
    put_string(ttext) ;
    put(NXLT(") ST\n")) ;
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
static void pf70InitPatBuf()
{
    pf70PatScannedPtr = pf70PatBuf;
    pf70PatInputPtr = pf70PatBuf;
    pf70PatOffset = 0;
}



/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
static void D01BitmapFill(sattr, boundingBox, id, pres,
                            data_hyt, data_wid, data_dep, eoflag)
    sub_attr_type *sattr;
    TmPoint32 boundingBox[];
    int id, pres, data_hyt, data_wid, data_dep;
    int eoflag;
{
    double scaled_wid, scaled_hyt;
    int num_pat_x, num_pat_y, startOffset, h, invert ;
    int bb_width, bb_height;
    int startingX, startingY ;
    double intens;

    if (data_hyt == 0 || data_wid == 0)
        goto leave_DoBitmapFill;

    /* don't do anything if solid fill when color is transparent */
    if ( sattr->type == BUILTIN_FILL_TYPE &&
         id == 0 &&
         sattr->mode == TMIS_O_D )
        goto leave_DoBitmapFill;

    bb_width = ABS(boundingBox[1].x-boundingBox[0].x);
    bb_height = ABS(boundingBox[1].y-boundingBox[0].y);
    if ((bb_width == 0) || (bb_height == 0))
        goto leave_DoBitmapFill;

    if ((id == 0 && sattr->mode == TMS_O_D) ||
        (id == 4 && sattr->mode == TMS_A_D) ||
        (sattr->mode == TMD))
        goto leave_DoBitmapFill;

    intens = IntensFromCMYKColor(sattr->fg.c,
                                 sattr->fg.m,
                                 sattr->fg.y,
                                 sattr->fg.k, 255);
#if 1
    if ((sattr->type == BUILTIN_FILL_TYPE) &&
        ((intens * pfGetPermilFill(id)) < 10) &&
        (sattr->mode == TMS_O_D))
        goto leave_DoBitmapFill;
#else
    if ((intens < 0.01) && (sattr->mode == TMS_O_D))
        goto leave_DoBitmapFill;
#endif

    scaled_wid = ((double)data_wid * RESOLUTION) / pres ;
    scaled_hyt = ((double)data_hyt * RESOLUTION) / pres ;

    /* special case #4 */
    if (( id == 4 ) && (sattr->mode != TMS_O_D))        /* solid fg fill */
    {
        if ( sattr->use_f_black )
            D01SetColor(TMA_BLACK);
        else
            D01SetCMYKColor(sattr->fg.c,
                            sattr->fg.m,
                            sattr->fg.y,
                            sattr->fg.k, 255);

        if (eoflag)
            put(NXLT("eofill\n")); /* was "eoclip eofill" */
        else
            put(NXLT("fill\n"));

        goto leave_DoBitmapFill;
    }

    /* set a background color */

    if ( sattr->use_b_white )
        D01SetColor(TMA_WHITE);
    else
        D01SetCMYKColor(sattr->bg.c,
                        sattr->bg.m,
                        sattr->bg.y,
                        sattr->bg.k, 255);

    /* fill the shape with the above color (if not transparent) */
    if ( eoflag )
    {
        if ( sattr->mode == TMS_O_D || sattr->mode == TMIS_O_D )
            put(NXLT("eoclip\n"));
        else
            put(NXLT("eoclip eofill\n"));
    }
    else
    {
        if ( sattr->mode == TMS_O_D || sattr->mode == TMIS_O_D )
            put(NXLT("clip\n"));
        else
            put(NXLT("clip fill\n"));
    }

    /* set the foreground color to tile over the background color */
    if (( sattr->mode == TMIS_O_D ) ||
        ( sattr->mode == TMIS))
    {
        invert = FALSE ;
        if ( sattr->use_b_white )
            D01SetColor(TMA_WHITE);
        else
            D01SetCMYKColor(sattr->bg.c,
                            sattr->bg.m,
                            sattr->bg.y,
                            sattr->bg.k, 255);
    }
    else
    {
        invert = TRUE ;
        if ( sattr->use_f_black )
            D01SetColor(TMA_BLACK);
        else
            D01SetCMYKColor(sattr->fg.c,
                            sattr->fg.m,
                            sattr->fg.y,
                            sattr->fg.k, 255);
    }

    /* determine the number and position of patterns to fill the boundaries */

    /* set the number of patterns to be tiled into shape */
    num_pat_x = (int)(bb_width/scaled_wid + 1);
    num_pat_y = (int)(bb_height/scaled_hyt + 1);

#if 0
    /* position the first tile */
    startingX = (double)(sattr->xoff + boundingBox[0].x);
    startingY = (double)(sattr->yoff + boundingBox[0].y);

    startingX = (startingX * pres) / RESOLUTION ;
    startingY = (startingY * pres) / RESOLUTION ;

    /* set starting position for fill pattern */
    if ( sattr->fill_abs )
    {
        startOffset = ((int)startingX) % data_wid;
        if ( startOffset != 0 )
            startingX -= (double)startOffset;
            num_pat_x++;
        }
        startOffset = ((int)startingY) % data_hyt;
        if ( startOffset != 0 )
        {
            startingY -= (double)startOffset;
            num_pat_y++;
        }
    }
    else
    {
        if ( sattr->xoff != 0 )
        {
            startingX -= (double)sattr->xoff;
            num_pat_x++;
        }
        if ( sattr->yoff != 0 )
        {
            startingY -= (double)sattr->yoff;
            num_pat_y++;
        }
    }
#else
    /* position the first tile */
    startingX = sattr->xoff + boundingBox[0].x ;
    startingY = sattr->yoff + boundingBox[0].y ;

    startingX = (startingX * pres) / RESOLUTION ;
    startingY = (startingY * pres) / RESOLUTION ;

    if ( sattr->fill_abs )
    {
        startOffset = startingX % data_wid;
        if ( startOffset != 0 )
        {
            startingX -= startOffset;
            num_pat_x++;
        }

        startOffset = startingY % data_hyt;
        if ( startOffset != 0 )
        {
            startingY -= startOffset;
            num_pat_y++;
        }
    }
    else
    {
        if ( sattr->xoff != 0 )
        {
            startingX -= sattr->xoff;
            num_pat_x++;
        }
        if ( sattr->yoff != 0 )
        {
            startingY -= sattr->yoff;
            num_pat_y++;
        }
    }
#endif

    /* otherwise tile the pattern over the background */

#if 0
    dots_to_pts = 72.0 / PATTERN_RESOLUTION ;
    scaleX = scaled_wid * dots_to_pts;
    scaleY = scaled_hyt * num_pat_y * dots_to_pts;
#endif
    h = data_hyt * num_pat_y ;

    /* translate and scale to put (unit sqaure at origin) to where we
       want the images to start stamping */
#if 0
    put(NXLT("/Tx %.2f def\n"),
            (startingX * dots_to_pts) / scaleX);
    put(NXLT("%.2f %.2f scale\n"),
        (double)scaled_wid, (double)(scaled_hyt * num_pat_y));
#endif
    put(NXLT("/Tx %.5f def\n"), startingX / (double)data_wid) ;
    put(NXLT("%d %d mul %d div %d %d mul %d div scale\n"),
        data_wid, RESOLUTION, pres,
        h, RESOLUTION, pres) ;

    put(NXLT("%d {\n"), num_pat_x);
    put(NXLT("    GS\n"));
#if 0
    put(NXLT("    Tx %.2f translate\n"), (startingY * dots_to_pts) / scaleY);
#endif
    put(NXLT("    Tx %.5f translate\n"), startingY / (double)h);
    /* See the cheat? if we specify more data (in the y) than is in
       the source then PS repeats the pattern "for free" hence,
       data_hyt * num_pat_y */

    /* mono image? */
    if (data_dep == 1)
    {
        put(NXLT("    %d %d %s"), data_wid, h, invert ? NXLT("true") : NXLT("false"));
        put(NXLT(" [%d 0 0 %d 0 %d]"), data_wid, -h, h) ;
        put(NXLT(" {p%02d} imagemask\n"), id);
    }
    else
    /* color image */
    {
        if (PF.PfColordev && pf_color_model != COL_MODEL_GRAY)
        {
            put(NXLT("    %d %d %d"), data_wid, h, bitsPerColorComponent);
            put(NXLT(" [%d 0 0 %d 0 %d]"), data_wid, -h, h) ;
            put(NXLT(" {pc%02d} {pg%02d} SELECTBUF COLORIMAGE\n"), id, id);
        }
        else
        {
            put(NXLT("    %d %d %d"), data_wid, h, bitsPerColorComponent);
            put(NXLT(" [%d 0 0 %d 0 %d]"), data_wid, -h, h) ;
            put(NXLT(" {pg%02d} image\n"), id);
        }
    }
    put(NXLT("    /Tx Tx %d add def\n"), 1);
    put(NXLT("    GR\n"));
    put(NXLT("} repeat\n"));

leave_DoBitmapFill:
    return;
}


/*-----------------------------------------------------------------------------
    D01BuiltInPatternFill
-----------------------------------------------------------------------------*/
static void D01BuiltInPatternFill(sattr, boundingBox, eoflag)
    sub_attr_type *sattr;
    TmPoint32 boundingBox[];
    int eoflag;
{
    int id = sattr->pat;

    if (id == -1)
        return ;

    D01BitmapFill(sattr,
                    boundingBox,
                    id,
                    PATTERN_RESOLUTION,
                    pattern[id].hyt,
                    pattern[id].wid,
                    1,
                    eoflag);
}


/*-----------------------------------------------------------------------------
    D01UserDefPatternFill
-----------------------------------------------------------------------------*/
static void D01UserDefPatternFill(sattr, boundingBox, eoflag)
    sub_attr_type *sattr;
    TmPoint32 boundingBox[];
    int eoflag;
{
    int id;

    id = pfFindPattern(sattr->pat) ;


    if (id != -1)
        D01BitmapFill(sattr,
                     boundingBox,
                     id,
                     SCREEN_RESOLUTION,
                     pattern[id].hyt,
                     pattern[id].wid,
                     pattern[id].dep,
                     eoflag);
}


/*-----------------------------------------------------------------------------
    D01RadialGradientFill
-----------------------------------------------------------------------------*/
static void D01RadialGradientFill(sattr, boundingBox, eoflag)
    sub_attr_type *sattr;
    TmPoint32 boundingBox[];
    int eoflag;
{
    int cx, cy;
    int tmp_fc, tmp_fy, tmp_fm, tmp_fg;
    int tmp_bc, tmp_by, tmp_bm, tmp_bg;
    int x_pos, y_pos, wid, hyt, xoff, yoff;
    double radius, delta;
    double cyanInc = 0.0;
    double magnInc = 0.0;
    double yellInc = 0.0;
    double blckInc = 0.0;
    double grayInc = 0.0;
    int colorSteps;


    /* set clipping for the shape */
    if (sattr->use_f_black)
    {
        put(NXLT("0 GRAY\n"));
    }
    else
    {
        D01SetCMYKColor(sattr->fg.c,
                        sattr->fg.m,
                        sattr->fg.y,
                        sattr->fg.k, 255);
    }

    if ( eoflag )
    {
        put(NXLT("eoclip\n"));
        put(NXLT("eofill\n"));
    }
    else
    {
        put(NXLT("clip\n"));
        put(NXLT("fill\n"));
    }


    x_pos = boundingBox[0].x;
    y_pos = boundingBox[0].y;
    wid = ABS(boundingBox[1].x - boundingBox[0].x);
    hyt = ABS(boundingBox[1].y - boundingBox[0].y);
    xoff = sattr->xoff ;
    yoff = sattr->yoff ;

    if ( (hyt == 0) || (wid == 0) )
        return;

    /* compute (cx, cy), center point */
    cx = wid / 2 ;
    cy = hyt / 2 ;

#if 0
    if (xoff > 0)
    {
        if (xoff > 100)
            xoff = 100 ;
        cx += (xoff*cx)/100 ;
    }
    else if (xoff < 0)
    {
        if (xoff < -100)
            xoff = -100 ;
        cx -= (-xoff*cx)/100 ;
    }

    if (yoff > 0)
    {
        if (yoff > 100)
            yoff = 100 ;
        cy += (yoff*cy)/100 ;
    }
    else if (yoff < 0)
    {
        if (yoff < -100)
            yoff = -100 ;
        cy -= (-yoff*cy)/100 ;
    }
#else
    if (xoff > 0)
    {
        if (xoff > 100)
            xoff = 100 ;
        cx += (xoff*cx)/100 ;
    }
    else if (xoff < 0)
    {
        if (xoff < -100)
            xoff = -100 ;
        cx -= (-xoff*cx)/100 ;
    }

    if (yoff > 0)
    {
        if (yoff > 100)
            yoff = 100 ;
        cy += (yoff*cy)/100 ;
    }
    else if (yoff < 0)
    {
        if (yoff < -100)
            yoff = -100 ;
        cy -= (-yoff*cy)/100 ;
    }
#endif

    /* compute radius = distance to furthest corner */
    if (xoff >= 0)
    {
        if (yoff >= 0)
            radius = sqrt((double)(cx*cx + cy*cy)) ;
        else
            radius = sqrt((double)((cx*cx) + (hyt-cy)*(hyt-cy))) ;
    }
    else
    {
        if (yoff >= 0)
            radius = sqrt((double)((wid-cx)*(wid-cx) + (cy*cy))) ;
        else
            radius = sqrt((double)((wid-cx)*(wid-cx) + (hyt-cy)*(hyt-cy))) ;
    }

    cx = x_pos + cx ;
    cy = y_pos + (hyt - cy) ;

    /* remove black component from color step calcs */
    tmp_fc = MIN ( sattr->fg.c + sattr->fg.k, 255 );
    tmp_fm = MIN ( sattr->fg.m + sattr->fg.k, 255 );
    tmp_fy = MIN ( sattr->fg.y + sattr->fg.k, 255 );
    tmp_bc = MIN ( sattr->bg.c + sattr->bg.k, 255 );
    tmp_bm = MIN ( sattr->bg.m + sattr->bg.k, 255 );
    tmp_by = MIN ( sattr->bg.y + sattr->bg.k, 255 );
    tmp_fg = (double)(25500 - 30*tmp_fc - 59*tmp_fm - 11*tmp_fy) / 100.0;
    tmp_bg = (double)(25500 - 30*tmp_bc - 59*tmp_bm - 11*tmp_by) / 100.0;


    if (PF.PfColordev)
    {
        colorSteps = ABS ( tmp_fc - tmp_bc );
        colorSteps = MAX ( ABS ( tmp_fm - tmp_bm ), colorSteps );
        colorSteps = MAX ( ABS ( tmp_fy - tmp_by ), colorSteps );
        if (colorSteps < 1)
            colorSteps = 1;

        cyanInc = ((double)sattr->bg.c - (double)sattr->fg.c) / colorSteps;
        magnInc = ((double)sattr->bg.m - (double)sattr->fg.m) / colorSteps;
        yellInc = ((double)sattr->bg.y - (double)sattr->fg.y) / colorSteps;
        blckInc = ((double)sattr->bg.k - (double)sattr->fg.k) / colorSteps;
    }
    else
    {
        colorSteps = ABS ( tmp_fg - tmp_bg );
        if (colorSteps < 1)
            colorSteps = 1;

        grayInc = (double)(tmp_bg - tmp_fg) / colorSteps;
        /* "gray" is whiteness, hence bg - fg */
    }

    delta = radius/colorSteps;

    /* fill with a gradient from fg color to bg color */
    /* the real way to do gray shading is in the Orange Book
       pages 259-.  Increment the tone in single point
       steps... */

    if ( radius > 0.0 )
    {
        put(NXLT("0 1 %d\n"), colorSteps);
        put(NXLT("{\n"));
        put(NXLT("    /loopix exch def\n"));
        put(NXLT("    n\n"));
        if (PF.PfColordev)
        {
            put(NXLT("    GET_CMYK\n"));    /* get current col,
                                                incr it */
            put(NXLT("    %f add 4 1 roll\n"), blckInc/255); /* cyan pushed first */
            put(NXLT("    %f add 4 1 roll\n"), yellInc/255);
            put(NXLT("    %f add 4 1 roll\n"), magnInc/255);
            put(NXLT("    %f add 4 1 roll\n"), cyanInc/255);
            put(NXLT("    %d %d %2f %2f loopix mul sub\n"), cx, cy, radius, delta);
            put(NXLT("    COLORCIRC\n"));
        }
        else
        {
            put(NXLT("    currentgray\n"));  /* => g */
            put(NXLT("    %f add\n"), grayInc/255);
            put(NXLT("    %d %d %.2f %.2f loopix mul sub\n"), cx, cy, radius, delta);
            put(NXLT("    GRAYCIRC\n"));
        }
        put(NXLT("} for\n"));
    }
}


/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
static void D01RectGradientFill(sattr, boundingBox, eoflag)
    sub_attr_type *sattr;
    TmPoint32 boundingBox[];
    int eoflag;
{
    int tmp_fc, tmp_fy, tmp_fm, tmp_fg;
    int tmp_bc, tmp_by, tmp_bm, tmp_bg;
    int wid, hyt;
    int xoff, yoff, xc, yc ;

    double fore_gray_f, back_gray_f;
    double cyanInc = 0.0;
    double magnInc = 0.0;
    double yellInc = 0.0;
    double blckInc = 0.0;
    double grayInc = 0.0;
    double deltax, deltay, deltaw, deltah ;
    int colorSteps;
    double angle, sina, cosa ;

    wid = boundingBox[3].x ;
    hyt = boundingBox[3].y ;

    if ( (hyt == 0) || (wid == 0) )
        return;
    
    angle = ((double)sattr->ang * M_PI) / 180.0 ;
    sina = sin(angle) ;
    cosa = cos(angle) ;

    xc = wid/2 ;
    yc = hyt/2 ;
    xoff = sattr->xoff ;
    yoff = sattr->yoff ;

    if (xoff > 0)
        xc = xc + ((xoff * wid) / 200) ;
    else if (xoff < 0)
        xc = xc - ((-xoff * wid) / 200) ;

    if (yoff > 0)
        yc = yc - ((yoff * hyt) / 200) ;
    else if (yoff < 0)
        yc = yc + ((-yoff * hyt) / 200) ;


    if (sattr->use_f_black)
    {
        put(NXLT("0 GRAY\n"));
    }
    else
    {
        D01SetCMYKColor(sattr->fg.c,
                        sattr->fg.m,
                        sattr->fg.y,
                        sattr->fg.k, 255);
    }

    /* set clipping for the shape */
    if ( eoflag )
    {
        put(NXLT("eoclip\n"));
        put(NXLT("eofill\n"));
    }
    else
    {
        put(NXLT("clip\n"));
        put(NXLT("fill\n"));
    }



    /* remove black component from color step calcs */
    tmp_fc = MIN ( sattr->fg.c + sattr->fg.k, 255 );
    tmp_fm = MIN ( sattr->fg.m + sattr->fg.k, 255 );
    tmp_fy = MIN ( sattr->fg.y + sattr->fg.k, 255 );
    tmp_bc = MIN ( sattr->bg.c + sattr->bg.k, 255 );
    tmp_bm = MIN ( sattr->bg.m + sattr->bg.k, 255 );
    tmp_by = MIN ( sattr->bg.y + sattr->bg.k, 255 );
    tmp_fg = (double)(25500 - 30*tmp_fc - 59*tmp_fm - 11*tmp_fy) / 100.0;
    tmp_bg = (double)(25500 - 30*tmp_bc - 59*tmp_bm - 11*tmp_by) / 100.0;

    fore_gray_f = (double)(25500 - 30*tmp_fc - 59*tmp_fm - 11*tmp_fy) / 25500.0;
    back_gray_f = (double)(25500 - 30*tmp_bc - 59*tmp_bm - 11*tmp_by) / 25500.0;

    if (PF.PfColordev)
    {
        colorSteps = ABS ( tmp_fc - tmp_bc );
        colorSteps = MAX ( ABS ( tmp_fm - tmp_bm ), colorSteps );
        colorSteps = MAX ( ABS ( tmp_fy - tmp_by ), colorSteps );
        if (colorSteps < 1)
            colorSteps = 1;

        cyanInc = ((double)sattr->bg.c - (double)sattr->fg.c) / colorSteps;
        magnInc = ((double)sattr->bg.m - (double)sattr->fg.m) / colorSteps;
        yellInc = ((double)sattr->bg.y - (double)sattr->fg.y) / colorSteps;
        blckInc = ((double)sattr->bg.k - (double)sattr->fg.k) / colorSteps;
    }
    else
    {
        colorSteps = ABS ( tmp_fg - tmp_bg );
        if (colorSteps < 1)
            colorSteps = 1;

        grayInc = (double)(tmp_bg - tmp_fg) / colorSteps;
    }

    deltax = (double)xc/colorSteps ;
    deltay = (double)yc/colorSteps;
    deltaw = (double)(wid)/colorSteps;
    deltah = (double)(hyt)/colorSteps;

    grayInc = back_gray_f - fore_gray_f;

    /* translate and rotate so that LLC of rotated box is (0,0) */
    put(NXLT("[%f %f %f %f %d %d] concat\n"),
            cosa, sina, -sina, cosa,
            boundingBox[2].x,   
            boundingBox[2].y) ;

    /* fill with a gradient from fg color to bg color */
    put(NXLT("0 1 %d\n"), colorSteps);
    put(NXLT("{\n"));
    put(NXLT("    /loopix exch def\n"));
    put(NXLT("    n\n"));
    put(NXLT("    %f loopix mul\n"), deltax);
    put(NXLT("    %f loopix mul M\n"), deltay);
    put(NXLT("    %d %f loopix mul sub\n"), wid, deltaw);
    put(NXLT("    %d %f loopix mul sub\n"), hyt, deltah);
    if (PF.PfColordev)
    {
        put(NXLT("    GET_CMYK\n"));
        put(NXLT("    %f add 4 1 roll\n"), blckInc/255);
        put(NXLT("    %f add 4 1 roll\n"), yellInc/255);
        put(NXLT("    %f add 4 1 roll\n"), magnInc/255);
        put(NXLT("    %f add 4 1 roll\n"), cyanInc/255);
        put(NXLT("    SET_CMYK\n"));
    }
    else
        put(NXLT("    currentgray %f add setgray\n"), grayInc/255);
    put(NXLT("    BOX fill\n"));
    put(NXLT("} for\n"));
}


/*-----------------------------------------------------------------------------
    D01LinearGradientFill
-----------------------------------------------------------------------------*/
static void D01LinearGradientFill(sattr, boundingBox, eoflag)
    sub_attr_type *sattr;
    TmPoint32 boundingBox[];
    int eoflag;
{
    int tmp_fc, tmp_fy, tmp_fm, tmp_fk, tmp_fg;
    int tmp_bc, tmp_by, tmp_bm, tmp_bk, tmp_bg;
    int x_pos, y_pos, hyt, wid;
    int colorSteps;
    int yoff, y0, y1 ;
    double dhyt;
    double cyanInc = 0.0;
    double magnInc = 0.0;
    double yellInc = 0.0;
    double blckInc = 0.0;
    double grayInc = 0.0;
    double angle, sina, cosa ;

    x_pos = y_pos = 0 ;
    wid = boundingBox[3].x ;
    hyt = boundingBox[3].y ;
    
    angle = ((double)sattr->ang * M_PI) / 180.0 ;
    sina = sin(angle) ;
    cosa = cos(angle) ;

    if ( (hyt == 0) || (wid == 0) )
        return;

    if (sattr->use_f_black)
        put(NXLT("0 GRAY\n"));
    else
    {
        D01SetCMYKColor(sattr->fg.c,
                        sattr->fg.m,
                        sattr->fg.y,
                        sattr->fg.k, 255);
    }

    /* set clipping for the shape */
    if ( eoflag )
        put(NXLT("eoclip\n"));
    else
        put(NXLT("clip\n"));

    /* remove black component from color step calcs */
    tmp_fc = MIN ( sattr->fg.c + sattr->fg.k, 255 );
    tmp_fm = MIN ( sattr->fg.m + sattr->fg.k, 255 );
    tmp_fy = MIN ( sattr->fg.y + sattr->fg.k, 255 );
    tmp_fk = 0;
    tmp_bc = MIN ( sattr->bg.c + sattr->bg.k, 255 );
    tmp_bm = MIN ( sattr->bg.m + sattr->bg.k, 255 );
    tmp_by = MIN ( sattr->bg.y + sattr->bg.k, 255 );
    tmp_bk = 0;
    tmp_fg = (double)(25500 - 30*tmp_fc - 59*tmp_fm - 11*tmp_fy) / 100.0;
    tmp_bg = (double)(25500 - 30*tmp_bc - 59*tmp_bm - 11*tmp_by) / 100.0;


    if (PF.PfColordev)
    {
        colorSteps = ABS ( tmp_fc - tmp_bc );
        colorSteps = MAX ( ABS ( tmp_fm - tmp_bm ), colorSteps );
        colorSteps = MAX ( ABS ( tmp_fy - tmp_by ), colorSteps );
        if (colorSteps < 1)
            colorSteps = 1;

        cyanInc = ((double)sattr->bg.c - (double)sattr->fg.c) / colorSteps;
        magnInc = ((double)sattr->bg.m - (double)sattr->fg.m) / colorSteps;
        yellInc = ((double)sattr->bg.y - (double)sattr->fg.y) / colorSteps;
        blckInc = ((double)sattr->bg.k - (double)sattr->fg.k) / colorSteps;
    }
    else
    {
        colorSteps = ABS ( tmp_fg - tmp_bg );
        if (colorSteps < 1)
            colorSteps = 1;

        grayInc = (double)(tmp_bg - tmp_fg) / colorSteps;
    }

    /* check for offset % */
    y0 = y1 = 0 ;
    yoff = sattr->yoff ;
    if (yoff > 0)
    {
        if (yoff > 100)
            yoff = 100 ;
        y1 = (yoff * hyt) / 100 ;
        hyt -= y1 ;
    }
    else if (yoff < 0)
    {
        if (yoff < -100)
            yoff = -100 ;
        y0 = (-yoff * hyt) / 100 ;
        hyt -= y0 ;
    }
    dhyt = (double)hyt/colorSteps;


    /* translate and rotate so that LLC of rotated box is (0,0) */
    put(NXLT("[%f %f %f %f %d %d] concat\n"),
            cosa, sina, -sina, cosa,
            boundingBox[2].x,   
            boundingBox[2].y) ;


    /* bottom slab */
    if (y1 > 0)
    {
        put(NXLT("n 0 0 M\n"), x_pos, y_pos) ;
        put(NXLT("%d %d BOX fill\n"), wid, y1);
        y_pos += y1 ;
    }

    /* gradient, bottom to top */
    if (dhyt > 0)
    {
        put(NXLT("0 1 %d\n{\n"), colorSteps-1);
        put(NXLT("    n %f mul %d add 0 exch M\n"), dhyt, y_pos);
        put(NXLT("    %d %f BOX fill\n"), wid, dhyt);
        if (PF.PfColordev)
        {
            put(NXLT("    GET_CMYK\n"));
            put(NXLT("    %f add 4 1 roll\n"), blckInc/255);
            put(NXLT("    %f add 4 1 roll\n"), yellInc/255);
            put(NXLT("    %f add 4 1 roll\n"), magnInc/255);
            put(NXLT("    %f add 4 1 roll\n"), cyanInc/255);
            put(NXLT("    SET_CMYK\n"));
        }
        else
        {
            put(NXLT("    currentgray %f add setgray\n"), grayInc/255);
        }
        put(NXLT("} for\n"));
    }

    /* top slab */
    if (y0 > 0)
    {
        D01SetCMYKColor(sattr->bg.c,
                        sattr->bg.m,
                        sattr->bg.y,
                        sattr->bg.k, 255);
        put(NXLT("n 0 %d M\n"), y_pos + hyt) ;
        put(NXLT("%d %d BOX fill\n"), wid, y0);
    }
}


/*-----------------------------------------------------------------------------
    D01BilinearGradientFill

-----------------------------------------------------------------------------*/
static void D01BilinearGradientFill(sattr, boundingBox, eoflag)
    sub_attr_type *sattr;
    TmPoint32 boundingBox[];
    int eoflag;
{
    int tmp_fc, tmp_fy, tmp_fm, tmp_fk, tmp_fg;
    int tmp_bc, tmp_by, tmp_bm, tmp_bk, tmp_bg;
    int x_pos, y_pos, hyt, wid;
    int colorSteps;
    int cpos ;
    int yoff ;
    double dhyt;
    double cyanInc = 0.0;
    double magnInc = 0.0;
    double yellInc = 0.0;
    double blckInc = 0.0;
    double grayInc = 0.0;
    double angle, sina, cosa ;

    x_pos = y_pos = 0 ;
    wid = boundingBox[3].x ;
    hyt = boundingBox[3].y ;
    
    angle = ((double)sattr->ang * M_PI) / 180.0 ;
    sina = sin(angle) ;
    cosa = cos(angle) ;

    if ( (hyt == 0) || (wid == 0) )
        return;

    /* set clipping for the shape */
    if ( eoflag )
        put(NXLT("eoclip\n"));
    else
        put(NXLT("clip\n"));

    /* remove black component from color step calcs */
    tmp_fc = MIN ( sattr->fg.c + sattr->fg.k, 255 );
    tmp_fm = MIN ( sattr->fg.m + sattr->fg.k, 255 );
    tmp_fy = MIN ( sattr->fg.y + sattr->fg.k, 255 );
    tmp_fk = 0;
    tmp_bc = MIN ( sattr->bg.c + sattr->bg.k, 255 );
    tmp_bm = MIN ( sattr->bg.m + sattr->bg.k, 255 );
    tmp_by = MIN ( sattr->bg.y + sattr->bg.k, 255 );
    tmp_bk = 0;
    tmp_fg = (double)(25500 - 30*tmp_fc - 59*tmp_fm - 11*tmp_fy) / 100.0;
    tmp_bg = (double)(25500 - 30*tmp_bc - 59*tmp_bm - 11*tmp_by) / 100.0;


    if (PF.PfColordev)
    {
        colorSteps = ABS ( tmp_fc - tmp_bc );
        colorSteps = MAX ( ABS ( tmp_fm - tmp_bm ), colorSteps );
        colorSteps = MAX ( ABS ( tmp_fy - tmp_by ), colorSteps );
        if (colorSteps < 1)
            colorSteps = 1;
        if (colorSteps & 1)
            colorSteps++ ;

        cyanInc = ((double)sattr->bg.c - (double)sattr->fg.c) / colorSteps;
        magnInc = ((double)sattr->bg.m - (double)sattr->fg.m) / colorSteps;
        yellInc = ((double)sattr->bg.y - (double)sattr->fg.y) / colorSteps;
        blckInc = ((double)sattr->bg.k - (double)sattr->fg.k) / colorSteps;
    }
    else
    {
        colorSteps = ABS ( tmp_fg - tmp_bg );
        if (colorSteps < 1)
            colorSteps = 1;

        grayInc = (double)(tmp_bg - tmp_fg) / colorSteps;
    }

    /* translate and rotate so that LLC of rotated box is (0,0) */
    put(NXLT("[%f %f %f %f %d %d] concat\n"),
            cosa, sina, -sina, cosa,
            boundingBox[2].x,   
            boundingBox[2].y) ;

    yoff = sattr->yoff ;
    if (yoff > 0)
    {
        if (yoff > 100)
            yoff = 100 ;
    }
    else if (yoff < 0)
    {
        if (yoff < -100)
            yoff = -100 ;
    }

    cpos = hyt*(yoff-100)/-200 ;

    /* box of bg -> fg gradient from 0 up to cpos */
    if (sattr->use_f_black)
        put(NXLT("1 GRAY\n"));
    else
        D01SetCMYKColor(sattr->bg.c,
                        sattr->bg.m,
                        sattr->bg.y,
                        sattr->bg.k, 255);

    dhyt = (double)cpos/colorSteps;
    if (dhyt > 0.0)
    {
            
        put(NXLT("0 1 %d\n"), colorSteps-1);
        put(NXLT("{\n"));
        put(NXLT("    n  %f mul %d add 0 exch M\n"), dhyt, 0);
        put(NXLT("    %d %f BOX fill \n"), wid, dhyt);
        if (PF.PfColordev)
        {
            put(NXLT("    GET_CMYK\n"));    /* get current color, incr it */
            put(NXLT("    %f sub 4 1 roll\n"), blckInc/255); /* cyan pushed first */
            put(NXLT("    %f sub 4 1 roll\n"), yellInc/255);
            put(NXLT("    %f sub 4 1 roll\n"), magnInc/255);
            put(NXLT("    %f sub 4 1 roll\n"), cyanInc/255);
            put(NXLT("    SET_CMYK\n"));
        }
        else
            put(NXLT("    currentgray %f sub setgray\n"), grayInc/255);
        put(NXLT("} for\n"));
    }

    /* box of fg -> bg gradient from cpos up to hyt */
    if (sattr->use_f_black)
        put(NXLT("0 GRAY\n"));
    else
    {
        D01SetCMYKColor(sattr->fg.c,
                        sattr->fg.m,
                        sattr->fg.y,
                        sattr->fg.k, 255);
    }

    dhyt = (double)(hyt-cpos)/colorSteps;
    if (dhyt > 0.0)
    {
        put(NXLT("0 1 %d\n"), colorSteps-1);
        put(NXLT("{\n"));
        put(NXLT("    n %f mul %d add 0 exch M\n"), dhyt, cpos);
        put(NXLT("    %d %f BOX fill\n"), wid, dhyt);
        if (PF.PfColordev)
        {
            put(NXLT("    GET_CMYK\n"));    /* get current color, incr it */
            put(NXLT("    %f add 4 1 roll\n"), blckInc/255); /* cyan pushed first */
            put(NXLT("    %f add 4 1 roll\n"), yellInc/255);
            put(NXLT("    %f add 4 1 roll\n"), magnInc/255);
            put(NXLT("    %f add 4 1 roll\n"), cyanInc/255);
            put(NXLT("    SET_CMYK\n"));
        }
        else
            put(NXLT("    currentgray %f add setgray\n"), grayInc/255);
        put(NXLT("} for\n"));
    }

}


/* what is the inversion (TMIS, TMIS_O_D etc.) of 25% gray? 75%gray?
 What about color? What is 90% of <c,m,y,k> ? */

/*-----------------------------------------------------------------------------
    D01GrayFill
-----------------------------------------------------------------------------*/
static void D01GrayFill(sattr, boundingBox, eoflag)
    sub_attr_type *sattr;
    TmPoint32 boundingBox[];
    int eoflag;
{
    /* This is a type 5 fill with pat representing the grayness */
    int bb_height, bb_width;
    int gray_factor;
    int c_diff, m_diff, y_diff, k_diff;

    bb_width = ABS(boundingBox[1].x - boundingBox[0].x);
    bb_height = ABS(boundingBox[1].y - boundingBox[0].y);
    if ((bb_width == 0) || (bb_height == 0))
        return;

    /* crude. Can we just "lighten" a color like this? */
    /* I suspect that this only works with (both fg and bg) undercolor
       removed */
    if ( sattr->use_f_black )
    {
        sattr->fg.c = 0;
        sattr->fg.m = 0;
        sattr->fg.y = 0;
        sattr->fg.k = 255;
    }
    if ( sattr->use_b_white )
    {
        sattr->bg.c = 0;
        sattr->bg.m = 0;
        sattr->bg.y = 0;
        sattr->bg.k = 0;
    }

    c_diff = sattr->bg.c - sattr->fg.c;
    m_diff = sattr->bg.m - sattr->fg.m;
    y_diff = sattr->bg.y - sattr->fg.y;
    k_diff = sattr->bg.k - sattr->fg.k;

    gray_factor = 1000 - sattr->pat;
    D01SetCMYKColor(sattr->fg.c + (c_diff * gray_factor)/1000,
                    sattr->fg.m + (m_diff * gray_factor)/1000,
                    sattr->fg.y + (y_diff * gray_factor)/1000,
                    sattr->fg.k + (k_diff * gray_factor)/1000, 255);

    

    /* fill the shape with the above color (fix for transparent) */
    /* If transparent then we _MUST_ emulate the fill with an
       imagemask and the moire be dammed! */
    if ( eoflag )
    {
        put(NXLT("eofill\n")); /* was "eoclip eofill" */
    }
    else
    {
        put(NXLT("fill\n")); /* was "clip fill" */
    }
}


/*-----------------------------------------------------------------------------
    D01FillPattern

    fills a bounded region. Any clipping is setup prior to
    calling the function.
-----------------------------------------------------------------------------*/
static void D01FillPattern(attr, boundingBox)
    attr_type *attr;
    TmPoint32 boundingBox[];
{
    sub_attr_type sattr;
    int id = 0;

    /* is this all transparent? */
    if (attr->ffg.see_thru && attr->fbg.see_thru)
        return ;

    /* start */
    PS_GS();

    /* extract the attributes */
    sattr.mode = attr->mode;
    sattr.type = attr->f_type;
    sattr.pat = attr->f_pat;
    sattr.ang = attr->f_ang;
    sattr.data = attr->f_data;
    sattr.dhyt = attr->f_dhyt;
    sattr.dwid = attr->f_dwid;
    sattr.ddep = attr->f_ddep;
    sattr.fill_abs = attr->fill_abs;
    sattr.xoff = attr->f_xoff;
    sattr.yoff = attr->f_yoff;
    sattr.fg = attr->ffg;
    sattr.bg = attr->fbg;
    sattr.use_f_black = attr->use_ffg;
    sattr.use_b_white = attr->use_fbg;


    /* compatability hacks... */
    if ((attr->fbg.see_thru)&&
                                /* ignore the 100% fills */
        !((attr->f_type == GRAY_FILL_TYPE && attr->f_pat == 1000) ||
          (attr->f_type == BUILTIN_FILL_TYPE && attr->f_pat == 4)))
    {
        sattr.mode = TMS_O_D;
        if (attr->ffg.see_thru)
            /* Absurd! fg+bg see thru?... */
            goto leave_D01FillPattern;
    }

    if (attr->ffg.see_thru)
    {
        sattr.mode = TMIS_O_D;
        /* flip the sense of the colors */
#if 0
        sattr.fg = attr->fbg;
        sattr.bg = attr->ffg;
#endif
    }

    /* gray fill with transparent bg becomes pattern fill */
    if ((sattr.type == GRAY_FILL_TYPE) && (sattr.pat != 1000)
        && (sattr.mode == TMS_O_D))
    {
        sattr.pat = snap_black(&sattr) ;
        sattr.type = BUILTIN_FILL_TYPE;
    }

    if ((sattr.mode != TMS_O_D) && (attr->f_type == BUILTIN_FILL_TYPE))
    {
        switch (attr->f_pat)
        {
        case 0:
        case 12:
        case 1:
        case 10:
        case 2:
        case 21:
        case 3:
        case 13:
        case 4:
            sattr.type = GRAY_FILL_TYPE;
            sattr.pat = pfGetPermilFill(attr->f_pat);
        }
    }
    /* now do the fill */
    switch ( sattr.type )
    {
    case BUILTIN_FILL_TYPE:
        D01BuiltInPatternFill(&sattr, boundingBox, TRUE);
        break;
    case USERDEF_FILL_TYPE:
        D01UserDefPatternFill(&sattr, boundingBox, TRUE);
        break;
    case LINEAR_FILL_TYPE:
        D01LinearGradientFill(&sattr, boundingBox, TRUE);
        break;
    case BILINEAR_FILL_TYPE:
        D01BilinearGradientFill(&sattr, boundingBox, TRUE);
        break;
    case RADIAL_FILL_TYPE:
        D01RadialGradientFill(&sattr, boundingBox, TRUE);
        break;
    case RECT_FILL_TYPE:
        D01RectGradientFill(&sattr, boundingBox, TRUE);
        break;
    case GRAY_FILL_TYPE:
        D01GrayFill(&sattr, boundingBox, TRUE);
        break;
    default:
        break;
    }

    /* finish */
leave_D01FillPattern:
    PS_GR();
    return;
}

/*-----------------------------------------------------------------------------
    returns a bultin-in  pattern based on gray fill and color,
        and sets color to black if it's gray
-----------------------------------------------------------------------------*/
static int snap_black(sattr)
    sub_attr_type *sattr;
{
    int k = sattr->fg.k ;

    if (sattr->fg.c == 0 && 
        sattr->fg.m == 0 &&
        sattr->fg.y == 0 &&
        sattr->fg.k > 0)
    {
        sattr->fg.k = 255 ;
        return getIXfromPercent((int)((sattr->pat * k)/255)) ;
    }
    else
        return getIXfromPercent(sattr->pat) ;
}


/*-----------------------------------------------------------------------------
    D01FillStroke

    tiles a fill pattern into a region. Any clipping is setup
    prior to calling the function (isRect should then be set to FALSE).
-----------------------------------------------------------------------------*/
static void D01FillStroke(attr, boundingBox)
    attr_type *attr;
    TmPoint32 boundingBox[];
{
    sub_attr_type sattr;
    int width = attr->l_wid ;


    /* set the line width */
    if (width < MINIMUM_LINE_WIDTH)
        width = MINIMUM_LINE_WIDTH ;

    if ( width != PF.cur_thick )
    {
        put(NXLT("%d setlinewidth\n"), width);
        PF.cur_thick = width ;
    }

    /* !! for line caps it is assumed that this routine is called in
     * a GS/GR context so that the linecap is reset to round on exit */
    switch ( attr->l_cap ) {
      case TM_CAP_ROUND:
        /* do nothing, this is default */
        break;
      case TM_CAP_BUTT:
        put(NXLT("0 setlinecap\n"));
        break;
      case TM_CAP_PROJECTING:
        put(NXLT("2 setlinecap\n"));
        break;
      default:
        /* garbage?? do nothing, use default (round cap) */
        break;
    }

    /* set the line style */
    if ( attr->l_style != PF.cur_style )
    {
        switch (attr->l_style)
        {
        case TM_SOLID:
            put(NXLT("1 setlinejoin "));
            put(NXLT("[] 0 setdash\n"));
            break;

        case TM_DASHED:
            put(NXLT("2 setlinejoin "));
            put(NXLT("0 setlinecap\n"));
            put(NXLT("[%d %d] 0 setdash\n"),
#if BUILD_KANJI
                        LONG_DASH*width/16,
                        LONG_SPACE*width/16);
#else
                        LONG_DASH,
                        LONG_SPACE);
#endif
            break;

        case TM_DOTTED:
            put(NXLT("2 setlinejoin "));
            put(NXLT("0 setlinecap\n"));
            put(NXLT("[%d %d] 0 setdash\n"),
#if BUILD_KANJI
                        LONG_DASH*width/16,
                        LONG_SPACE*width/16);
#else
                        SHORT_DASH,
                        LONG_SPACE);
#endif
            break;

        case TM_DADOT:
            put(NXLT("2 setlinejoin "));
            put(NXLT("0 setlinecap\n"));
            put(NXLT("[%d %d %d %d] 0 setdash\n"),
#if BUILD_KANJI
                        LONG_DASH*width/16,
                        LONG_SPACE*width/16,
                        SHORT_DASH*width/16,
                        LONG_SPACE*width/16);
#else
                        LONG_DASH,
                        LONG_SPACE,
                        SHORT_DASH,
                        LONG_SPACE);
#endif 
            break;

        case TM_THIN_DOTTED:
            put(NXLT("2 setlinejoin "));
            put(NXLT("0 setlinecap\n"));
            put(NXLT("[%d %d] 0 setdash\n"),
#if BUILD_KANJI
                        DOT*width/16,
                        DOT_SPACE*width/16);
#else
                        DOT,
                        DOT_SPACE);
#endif
            break;

        case 5:
            put(NXLT("2 setlinejoin "));
            put(NXLT("0 setlinecap\n"));
            put(NXLT("[%d %d %d %d %d %d] 0 setdash\n"),
#if BUILD_KANJI
                        LONG_DASH*width/16,
                        LONG_SPACE*width/16,
                        SHORT_DASH*width/16,
                        LONG_SPACE*width/16,
                        SHORT_DASH*width/16,
                        LONG_SPACE*width/16);
#else
                        LONG_DASH,
                        LONG_SPACE,
                        SHORT_DASH,
                        LONG_SPACE,
                        SHORT_DASH,
                        LONG_SPACE) ;
#endif
            break;

        case 6:
            put(NXLT("2 setlinejoin "));
            put(NXLT("0 setlinecap\n"));
            put(NXLT("[%d %d] 0 setdash\n"),
#if BUILD_KANJI
                        2*LONG_DASH*width/16,
                        LONG_SPACE*width/16);
#else

                        2*LONG_DASH,
                        LONG_SPACE);
#endif
            break;
        }
        PF.cur_style = attr->l_style;
    }

    /* simple cases - solid line */
    if ( attr->l_type == BUILTIN_FILL_TYPE && attr->l_pat == 4) 
    {
        /* don't do anything if fg color is transparent */
        if ( attr->lfg.see_thru == FALSE )
        {
            if ( attr->use_lfg )
                D01SetColor(TMA_BLACK);
            else
                D01SetCMYKColor(attr->lfg.c,
                                attr->lfg.m,
                                attr->lfg.y,
                                attr->lfg.k,255);
            put(NXLT("S\n"));
        }
    }
    else if ( attr->l_type == BUILTIN_FILL_TYPE && attr->l_pat == 0 )
    {
        /* don't do anything if bg color is transparent */
        if ( attr->lbg.see_thru == FALSE )
        {
            if ( attr->use_lbg )
                D01SetColor(TMA_WHITE);
            else
                D01SetCMYKColor(attr->lbg.c,
                                attr->lbg.m,
                                attr->lbg.y,
                                attr->lbg.k,255);
            put(NXLT("S\n"));
        }
    }
    else if ( attr->l_type == GRAY_FILL_TYPE)
    {
        /* don't do anything if bg color is transparent */
        if ( attr->lfg.see_thru == FALSE )
        {
            int c_diff, m_diff, y_diff, k_diff, gray_factor;
            /* I suspect that this only works with (both fg and bg) undercolor
               removed */
            sattr.use_f_black = attr->use_lfg;
            sattr.use_b_white = attr->use_lbg;
#if 0
            if (sattr.use_f_black)
            {
                sattr.bc = 0;
                sattr.bm = 0;
                sattr.by = 0;
                sattr.bb = 255;
            }
            else
            {
                sattr.bc = attr->lbg.c;
                sattr.bm = attr->lbg.m;
                sattr.by = attr->lbg.y;
                sattr.bb = attr->lbg.k;
            }
            if (sattr.use_b_white)
            {
                sattr.fc = 0;
                sattr.fm = 0;
                sattr.fy = 0;
                sattr.fb = 0;
            }
            else
            {
                sattr.fc = attr->lfg.c;
                sattr.fm = attr->lfg.m;
                sattr.fy = attr->lfg.y;
                sattr.fb = attr->lfg.k;
            }
#else
            if (sattr.use_b_white)
            {
                sattr.bg.c = 0;
                sattr.bg.m = 0;
                sattr.bg.y = 0;
                sattr.bg.k = 0;
            }
            else
            {
                sattr.bg = attr->lbg;
            }
            if (sattr.use_f_black)
            {
                sattr.fg.c = 0;
                sattr.fg.m = 0;
                sattr.fg.y = 0;
                sattr.fg.k = 255;
            }
            else
            {
                sattr.fg = attr->lfg;
            }
#endif
            c_diff = sattr.bg.c - sattr.fg.c;
            m_diff = sattr.bg.m - sattr.fg.m;
            y_diff = sattr.bg.y - sattr.fg.y;
            k_diff = sattr.bg.k - sattr.fg.k;

            gray_factor = 1000 - attr->l_pat;
            D01SetCMYKColor(sattr.fg.c + (c_diff * gray_factor)/1000,
                            sattr.fg.m + (m_diff * gray_factor)/1000,
                            sattr.fg.y + (y_diff * gray_factor)/1000,
                            sattr.fg.k + (k_diff * gray_factor)/1000, 255);
            put(NXLT("S\n"));
        }
    }
    else                        /* fill pattern or gradient */
    {
        put(NXLT("SP\n"));

        /* extract the attributes */
        sattr.mode = attr->mode;
        sattr.type = attr->l_type;
        sattr.pat = attr->l_pat;
        sattr.ang = attr->l_ang;
        sattr.data = attr->l_data;
        sattr.dhyt = attr->l_dhyt;
        sattr.dwid = attr->l_dwid;
        sattr.ddep = attr->l_ddep;
        sattr.fill_abs = attr->fill_abs;
        sattr.xoff = attr->l_xoff;
        sattr.yoff = attr->l_yoff;
        sattr.fg = attr->lfg;
        sattr.bg = attr->lbg;
        sattr.use_f_black = attr->use_lfg;
        sattr.use_b_white = attr->use_lbg;

        /* compatability hacks... */
        if (attr->lbg.see_thru)
            sattr.mode = TMS_O_D;

        if (attr->l_type == BUILTIN_FILL_TYPE)
        {
            switch (attr->l_pat)
            {
            case 0:          /* 0% */
            case 12:
            case 1:          /* 25% */
            case 10:
            case 2:          /* 50% */
            case 21:
            case 3:          /* 75% */
            case 13:
            case 4:          /* 100% */
                sattr.type = GRAY_FILL_TYPE;
                sattr.pat = pfGetPermilFill(attr->l_pat);
                break;
            }
        }
        /* now do the fill */
        switch ( attr->l_type )
        {
          case BUILTIN_FILL_TYPE:
            D01BuiltInPatternFill ( &sattr, boundingBox, FALSE );
            break;
          case USERDEF_FILL_TYPE:
            D01UserDefPatternFill ( &sattr, boundingBox, FALSE );
            break;
          case LINEAR_FILL_TYPE:
            D01LinearGradientFill ( &sattr, boundingBox, FALSE );
            break;
          case BILINEAR_FILL_TYPE:
            D01BilinearGradientFill ( &sattr, boundingBox, FALSE );
            break;
          case RADIAL_FILL_TYPE:
            D01RadialGradientFill ( &sattr, boundingBox, FALSE );
            break;
          case RECT_FILL_TYPE:
            D01RectGradientFill ( &sattr, boundingBox, FALSE );
            break;
          case GRAY_FILL_TYPE:
            D01GrayFill ( &sattr, boundingBox, FALSE );
            break;
        }
    }
    return;
}


/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
static void D01LoadBitmapPattern(Bitmap_p, pat_id)
    bitmap_struct *Bitmap_p;
    int pat_id;
{
    register int i;
    register int j;
    register int bit_count;
    int byte_count, bytewidth;
    int src_data_byte_index;
    unsigned char out_byte;

    /* format the data and write it to the print ready file */
    out_byte = 0;
    bit_count = byte_count = 0;
    bytewidth = Bitmap_p->image.bytes_per_line;
    put(NXLT("/p%02d <"), pat_id);
    for ( i=0; i<Bitmap_p->image.height; i++ )
    {
        for ( j=0; j<Bitmap_p->image.width; j++ )
        {
            /* set a bit (l to r) for each set source bit */
            src_data_byte_index = i*bytewidth + j/8;
            if ( Bitmap_p->image.data[src_data_byte_index] & (0x80 >> (j%8)) )
            {
                out_byte |= 0x80 >> bit_count; /* bit swapped..?.. */
            }

            /* write out the byte every 8 bits */
            bit_count++;
            if ( bit_count == 8 )
            {
                put(NXLT("%02x"), out_byte);
                out_byte = bit_count = 0;
                byte_count++;
                if ( byte_count == 32 )
                {
                    byte_count = 0;
                    put(NXLT("\n      "));
                }
            }
        }

        /* postscript rows are byte aligned */
        if ( bit_count != 0 )
        {
            put(NXLT("%02x"), out_byte);
            out_byte = bit_count = 0;
            byte_count++;
            if ( byte_count == 32 )
            {
                byte_count = 0;
                put(NXLT("\n      "));
            }
        }
    }
    put(NXLT("> def\n"));

}

/*-----------------------------------------------------------------------------
    D01LoadMonoPattern

    Load a mono-colored bit fill pattern for use with the imagemask operator
-----------------------------------------------------------------------------*/
static void D01LoadMonoPattern(hyt, wid, data, id)
    int hyt, wid ;
    char *data;
    int id ;
{
    register int i;
    register int j;
    register int bit_count;
    int byte_count, bytewidth;
    int sx;
    unsigned char out_byte;

    /* format the data and write it to the print ready file */
    out_byte = 0;
    bit_count = byte_count = 0;
    bytewidth = 2*((wid+15)/16);
    put(NXLT("/p%02d <"), id);
    for (i=0; i<hyt; i++)
    {
        for (j=0; j<wid; j++)
        {
            /* set a bit (l to r) for each set source bit */
            sx = i*bytewidth + j/8;
            if ( data[sx] & (0x80 >>(j%8)) )
                out_byte |= 0x80 >> bit_count; /* bit swapped..?.. */

            /* write out the byte every 8 bits */
            bit_count++;
            if ( bit_count == 8 )
            {
                put(NXLT("%02x"), out_byte);
                out_byte = bit_count = 0;
                byte_count++;
                if (byte_count == 32)
                {
                    byte_count = 0 ;
                    put(NXLT("\n      "));
                }
            }
        }

        /* postscript rows are byte aligned */
        if ( bit_count != 0 )
        {
            put(NXLT("%02x"), out_byte);
            out_byte = bit_count = 0;
            byte_count++;
            if ( byte_count == 32 )
            {
                byte_count = 0;
                put(NXLT("\n      "));
            }
        }
    }
    put(NXLT("> def\n"));
}

/*-----------------------------------------------------------------------------
    D01LoadColorPattern

    Load a color bit fill pattern for use with the colorimage operator
-----------------------------------------------------------------------------*/
static void D01LoadColorPattern(hyt, wid, dep, data, cmap, id)
    int hyt, wid, dep ;
    char *data;
    cmap_type *cmap ;
    int id ;
{
    char *colorImagebuf, *grayImagebuf;
    char *datap;
    int col, line;
    int color_x_bytes, gray_x_bytes, bx_size;
    int len;
    int x_size, y_size;
    cmap_cache_entry *cmap_cache;
    int i ;


    x_size = wid;
    y_size = hyt;
    color_x_bytes = ((x_size*numColorComponents*bitsPerColorComponent) + 7) >> 3; 
    gray_x_bytes = ((x_size*bitsPerGrayComponent) + 7) >> 3;
    bx_size = x_size + x_size % 2;

    if (pf_color_model == COL_MODEL_GRAY)
    {
        put(NXLT("/pg%02d <"), id);
        GenerateGrayImage(data, bx_size, x_size, y_size, cmap) ;
        put(NXLT("> def\n"));
        return ;
    }

    cmap_cache =
        (cmap_cache_entry *)calloc(cmap->n_colors, sizeof(cmap_cache_entry));
    colorImagebuf = malloc(color_x_bytes);
    grayImagebuf = malloc(gray_x_bytes);


    for (i=0; i <cmap->n_colors; i++)
        cmap_cache[i].gray_c = -1;

    if (PF.PfColordev)
    {
        /* generate the color data */
        put(NXLT("/pc%02d <"), id);
        if (y_size * color_x_bytes >=  MAX_PS_STRING)
        {
#if 0
            Abandon(XLT("Custom Bitmap Fill exceeds PostScript limit")) ;
#elseif 0
            put(XLT("\n%% Custom Bitmap Fill exceeds PostScript limit\n")) ;
            goto end_color_def ;
#else
	        y_size =  MAX_PS_STRING / color_x_bytes ;
#endif
        }

        for (line = y_size, datap = data ;
                line > 0 ;
                line--, datap += bx_size)
        {

            memset(colorImagebuf, 0, color_x_bytes);
            memset(grayImagebuf, 0, gray_x_bytes);

            switch (pf_color_model)
            {
            case COL_MODEL_CMYK :
                GenerateCMYKGrayImages(datap, bx_size,
                        colorImagebuf, grayImagebuf,
                        cmap, cmap_cache) ;
                break ;
                    
            case COL_MODEL_RGB :
                GenerateRGBGrayImages(datap, bx_size,
                        colorImagebuf, grayImagebuf,
                        cmap, cmap_cache) ;
                break ;
            }

            /* write one row of color bytes */
            for (len = col = 0; col < color_x_bytes; )
            {
                put("%02x", colorImagebuf[col++] & 0xFF) ;
                if (len == 78)
                {
                    put("\n");
                    len = 0;
                }
                else
                    len += 2;
            }
            if (len)
                put("\n");

        }
end_color_def:
        put(NXLT("> def\n"));
    }

    /* generate the gray data */
    put(NXLT("/pg%02d <"), id);

    if (y_size * gray_x_bytes >=  MAX_PS_STRING)
    {
#if 0
        Abandon(XLT("Custom Bitmap Fill exceeds PostScript limit")) ;
#elseif 0
        put(XLT("\n%% Custom Bitmap Fill exceeds PostScript limit\n")) ;
        goto end_gray_def ;
#else
		y_size =  MAX_PS_STRING / gray_x_bytes ;
#endif
    }

    for (line = y_size, datap = data ;
            line > 0 ;
            line--, datap += bx_size)
    {

        memset(colorImagebuf, 0, color_x_bytes);
        memset(grayImagebuf, 0, gray_x_bytes);

        switch (pf_color_model)
        {
        default:
        case COL_MODEL_CMYK :
            GenerateCMYKGrayImages(datap, bx_size,
                    colorImagebuf, grayImagebuf,
                    cmap, cmap_cache) ;
            break ;
                    
        case COL_MODEL_RGB :
            GenerateRGBGrayImages(datap, bx_size,
                    colorImagebuf, grayImagebuf,
                    cmap, cmap_cache) ;
            break ;
        }


        /* write gray bytes */
        for (len = col = 0; col < gray_x_bytes; )
        {
            put("%02x", grayImagebuf[col++] & 0xFF) ;
            if (len == 78)
            {
                put("\n");
                len = 0;
            }
            else
                len += 2;
        }
        if (len)
            put("\n");
    }

end_gray_def:
    put(NXLT("> def\n"));

    /* let my buffers go, oh Lord */
    free(colorImagebuf);
    free(grayImagebuf);
    free(cmap_cache);
}

/*-----------------------------------------------------------------------------
    D01PreLoadPatterns
-----------------------------------------------------------------------------*/
static void D01PreLoadPatterns()
{
    int pat, bytesize = 0, bytewidth;
    unsigned char *ip;
    int ii;
    bitmap_struct Bitmap;


    /* load the built in patterns used in this doc */
    for ( pat=0; pat<MAX_PATS; pat++ )
    {
        /* don't send unused patterns to the print ready file */
#if 0
        if ( pat_needed[pat] && pat != 0 && pat != 4 )
#else
        if ( pattern[pat].needed )
#endif
        {

            memset((char *)&Bitmap, '\0', sizeof(bitmap_struct));
            if (AxIMtoXImage(printer_fills[pat], printer_fills[pat], &Bitmap) == AOK)
            {
                pattern[pat].wid = Bitmap.image.width;
                pattern[pat].hyt = Bitmap.image.height;
                bytewidth = Bitmap.image.bytes_per_line;
                bytesize = bytewidth * Bitmap.image.height;
                
                if (Bitmap.image.bitmap_bit_order != MSBFirst)
                {
                    ip = (unsigned char*)Bitmap.image.data;
                    for (ii = bytesize; ii>0; ii--)
                    {
                        *ip = BitSpin(*ip);
                        ip++;
                    }
                }
            }
            else
            {
                put(NXLT("/p%02d <ff00> def\n"), pat);
            }
#if 0
            if (Bitmap.image.bitmap_bit_order == MSBFirst)
            {
                ip = (unsigned char*)Bitmap.image.data;
                for (ii = bytesize; ii>0; ii--)
                {
                    *ip = BitSpin(*ip);
                    ip++;
                }
            }
#endif
            D01LoadBitmapPattern(&Bitmap, pat);
        }
    }

    /* now load the user defined patterns used in this doc */
    for (pat = MAX_PATS ; pat < 2*MAX_PATS ; pat++)
    {
        if (!pattern[pat].needed)
            continue ;

        /* format the data and write it to the print ready file */
        if (pattern[pat].dep == 1)
            D01LoadMonoPattern(pattern[pat].hyt,
                    pattern[pat].wid,
                    pattern[pat].data,
                    pat);
        else
            D01LoadColorPattern(pattern[pat].hyt,
                    pattern[pat].wid,
                    pattern[pat].dep,
                    pattern[pat].data,
                    &pattern[pat].cmap,
                    pat);
    }
    return;
}


/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
int D01LoadNamedFont(fontName, pointsize)
    char *fontName;
    int pointsize;   /* in 10000ths of a point */
{
    printFontName *fn = (printFontName *) 0;

    /* get the printFontName record for fontName */
    axhGetItem(fontName, &printFontNames, (char *) 0, (char **) &fn);
    if (!fn)
        return FALSE;
    
    if (fn->fontNumber == Curfont.number && pointsize == Curfont.size)
        return TRUE;
    
    put(NXLT("AXFont%dp%d SETFONT\n"), fn->fontNumber, pointsize);
    Curfont.number = fn->fontNumber;
    Curfont.size = pointsize;
    return TRUE;
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
static int RecordNamedFont(fontName, pointSize, eightBit, wideChar)
    char *fontName;
    int pointSize;  /* in 10000ths of a point */
    int eightBit, wideChar;  /* bools */
{
    int ix;
    printFontName *pName = NULL;
    printFontName *newName = NULL;
    int  *tEight = NULL;
    int found = FALSE;
    
    /* look up the printFontName and encoding records for fontName */
    axhGetItem(fontName, &printFontNames, (char *) 0, (char **) &pName);
    if (!pName)
        return FALSE;
    
    /* make sure the font's encoding is recorded */
    axhGetItem(pName->encodingName, &printFontEncodings,
               (char *) 0, (char **) &tEight);
    if (!tEight)
        return FALSE;
    
    if (wideChar)
        pName->wideChar = TRUE; /* update printFontName record */
    if (eightBit)
    {
        if (tEight)
            *tEight = TRUE; /* update encoding record */

        pName->eightBit = TRUE; /* update printFontName record */
    }

    /* check whether to add a new pointsize to the record */
    for (ix = 0; ix < pName->numSizes; ix++)
    {
        if (pointSize == pName->sizes[ix])
            found = TRUE;
    }
    if (!found)
    {
        pName->numSizes++;
        pName->sizes = (int *) realloc((char *) pName->sizes,
                        pName->numSizes * sizeof(int));
        pName->sizes[ pName->numSizes - 1 ] = pointSize;
    }

    /* set global flag indicating that at least one font is used */
    PF.fontNeeded = TRUE;
    return TRUE;
}

/*-----------------------------------------------------------------------------
    D01PreLoadFonts
-----------------------------------------------------------------------------*/
static void D01PreLoadFonts()
{
    int *eight, ix, jx;
    printFontName *pName;
    char **encodingNames = (char **) NULL;
    char **fontNames = (char **) NULL;
    char *errString;
    int numEncodingNames = 0;
    int numFontNames = 0;
    int reEncodeDefined = FALSE;
    int symbolEncoding;
    fontTabEnt *ftab ;
    char *fontName ;

    if (!PF.fontNeeded)
        return;

    D01SoftDownLoad() ;

    /* get lists of the font names and encoding names
     * collected during the prescan
     */
    axhListNames(&printFontNames, &fontNames, &numFontNames);
    axhListNames(&printFontEncodings, &encodingNames, &numEncodingNames);

    /* check for Type 1 download fonts */
    for (ix = 0; ix < numFontNames; ix++)
    {
        errString = axGetFontRecord(fontNames[ix], &ftab);
        if (errString)
        {
            freeErrString(errString) ;
            ftab = NULL ;
        }
        if (ftab && ftab->Type1FontFileName)
        {
            /* download this font, if you can */
            Type1DownLoad(ftab->Type1FontFileName);
        }
    }
        
    /* go thru the encodings and download vectors as needed.
     * rules: 1) send vector for any !latin1, !symbol, !dingbats
     *      2) if latin1 or dingbats, send vector if eightbit
     */
    for (ix = 0; ix < numEncodingNames; ix++)
    {
        if (strcmp(encodingNames[ix], AxSymbolEncoding) == 0)
            continue; /* no vector needed for symbol */
        if ((strcmp(encodingNames[ix], NXLT("ISOLatin1Encoding")) == 0) ||
            (strcmp(encodingNames[ix], NXLT("AxDingbatsEncoding")) == 0))
        {
            /* must look up whether eightbit characters were used */
            eight = (int *) 0;
            axhGetItem(encodingNames[ix], &printFontEncodings,
                       (char *) 0, (char **) &eight);
            if (!eight)
                return;
            else
            {
                if (*eight)
                {
                    /* eightbit chars from this encoding were used */
                    if (!reEncodeDefined)
                    {
                        defineReEncode();
                        reEncodeDefined = TRUE;
                    }
                    downloadEncodingVector(encodingNames[ix]);
                }
            }
        }
        else
        {
            if (!reEncodeDefined)
            {
                defineReEncode();
                reEncodeDefined = TRUE;
            }
            downloadEncodingVector(encodingNames[ix]);
        }
    }

    /* now to establish the font dictionary keys
     * to be used in this print job
     */
    put(NXLT("/DEFINEFONTS {\n"));
    for (ix = 0; ix < numFontNames; ix ++)
    {
        pName = (printFontName *) 0;
        axhGetItem(fontNames[ix], &printFontNames, (char *) 0,
                   (char **) &pName);
        if (!pName)
            continue ;
        else
        {
#if BUILD_ORIENT
            if (pName->wideChar || !pName->encodingName[0])
            {                 
                /* Considering as Kanji code */
                free( pName->encodingName );
                pName->encodingName =
                    (char *)malloc( strlen(AxSymbolEncoding) +1 );
                strcpy(pName->encodingName, AxSymbolEncoding);
            }
#endif
            symbolEncoding = !strcmp(pName->encodingName, AxSymbolEncoding);
            for (jx = 0; jx < pName->numSizes; jx++)
            {
                if (pName->gallium)
                    fontName = PSfontName(pName->psFontName,
                                          pName->sizes[jx]) ;
                else
                    fontName = pName->psFontName ;
                if (pName->eightBit && !symbolEncoding)
                {
                    put(NXLT("      /%s /E%s ev%s ReEncodeSmall\n"),
                        fontName,
                        fontName, pName->encodingName);
                    put(NXLT("      /AXFont%dp%d /E%s FINDFONT "),
                        pName->fontNumber,
                        pName->sizes[jx], fontName);
                }
                else
                {
                    put(NXLT("      /AXFont%dp%d /%s FINDFONT "),
                        pName->fontNumber,
                        pName->sizes[jx], fontName);
                }
                put(NXLT("%.3f POINTSCALEFONT def\n"),
                    (double) pName->sizes[jx] / 10000.0);
                if (pName->gallium)
                    free(fontName) ;
            }
        }
    }
    put(NXLT("} def\n"));
    axhFreeListNames(fontNames);
    axhFreeListNames(encodingNames);
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
static void defineReEncode()
{
    /*  newcn - new codes and names
     *  newfn - new font name
     *  basefn - base font name
     *  basefd - base font dictionary
     *  newf - new font
     */
    put(NXLT("/ReEncodeSmall {\n"));
    put(NXLT("      AXredict begin\n"));
    put(NXLT("      /newcn exch def\n"));
    put(NXLT("      /newfn exch def\n"));
    put(NXLT("      /basefn exch def\n"));
    put(NXLT("      /basefd basefn FINDFONT def\n"));
    put(NXLT("      /newf basefd maxlength dict def\n"));
    put(NXLT("      basefd\n"));
    put(NXLT("      {\n"));
    put(NXLT("      exch dup /FID ne\n"));
    put(NXLT("      {\n"));
    put(NXLT("        dup dup /Encoding eq exch /FontBBox eq or\n"));
    put(NXLT("        { exch dup length array copy newf 3 1 roll put }\n"));
    put(NXLT("        { exch newf 3 1 roll put }\n"));
    put(NXLT("        ifelse\n"));
    put(NXLT("      }\n"));
    put(NXLT("      { pop pop }\n"));
    put(NXLT("      ifelse\n"));
    put(NXLT("      } forall\n"));
    put(NXLT("      newf /FontName newfn put\n"));
    put(NXLT("      newcn aload pop\n"));
    put(NXLT("      newcn length 2 idiv\n"));
    put(NXLT("      { newf /Encoding get 3 1 roll put }\n"));
    put(NXLT("      repeat\n"));
    put(NXLT("      newfn newf definefont pop\n"));
    put(NXLT("      end\n"));
    put(NXLT("} def\n"));
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
static int downloadEncodingVector(eName)
    char *eName; /* e.g., "ISOLatin1Encoding" */
{
    int numChars, *charCode;
    char *cp, *eptr, *errString;
    char *eList = (char *) NULL;
    
        /* get a list of all the charnames in the encoding */
    if (!dumpCachelistNames(eName, &eList, &numChars))
        return FALSE; /* no list */
    put(NXLT("/ev%s [\n"), eName);
    eptr = eList;
    while (cp = getNextName(&eptr))
    {
            /* look up the encoding of the character name */
        errString = axGetCharacterCode(eName, cp, &charCode);
        if (errString)
        {
            freeErrString(errString);   /* lookup failed */
        }
        else
        {
            if (*charCode && *charCode >= 0)
                put(NXLT("%d /%s\n"), *charCode, cp);
        }
    }
    put(NXLT("] def\n"));
    freeCacheNameList(eList);
    return TRUE;
}

/*-----------------------------------------------------------------------------
 * getNextName - used on newline-separated name lists.
 * this function is destructive to the passed list (*cp);
 * it changes the newlines to nulls 
-----------------------------------------------------------------------------*/
static char *getNextName(cp)
    char **cp;
{
    register char *tp, *np;
    np = tp = *cp;
    if ( np == (char *)0 )
        return((char *) NULL);   /* no list */
    if (*np == '\0')
        return ((char *) NULL); /* end of list */

    while(*tp != '\0' && *tp != '\n')
        tp++;
    if (*tp == '\n')
        *tp = '\0'; /* null-stamp */
    *cp = tp + 1; /* position at start of next name */
    
    return (np);
}

#if AXMS                       /*  {  */
static int download_to_postscript( fontTabEnt *ftab )
{
    char *p;

    if ( ftab->PostScriptPrintName != (char *)NULL )
        /* substitute some other font */
        return( FALSE );
    p = ftab->targetMetricsNames[0];
    if ( !p || (strlen(p) < 4) )
        return( FALSE );

    p = p + strlen(p) - 4;
    if ( strcmp(p, ".win") != 0 )
        /* not a special truetype metrics file name as we construct them */
        return( FALSE );

    return( TRUE );
}
#endif                            /*  }  */


/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
static void D01PreScanText()
{
    int segCount, width, eightBit, wideChar;
    register int piececount;
    register int txlen;
    register int ix;
    int underPos, underThick, strikePos, xZoom, yZoom;
    int sdx, sdy, sfg, no_prnt, draw_shadow, shadow_only, s_type, draw_emboss;
    p32 pos;
    tx_char *textstr;
    textAttrs *txProps, fieldProps;
    attr_type *fill_attr;
    char *fontName, *errString;
    fontTabEnt *ftab, *alttab;
    unsigned int f_type, f_pat;
    char *mapped_fontName;
    r32 clip ;
    PF_soft *soft ;
    axctm unit_ctm ;
    color_type new_color;

    unit_ctm.ca = unit_ctm.cd = 1.0;
    unit_ctm.cb = unit_ctm.cc = 0.0;
    unit_ctm.ctx = unit_ctm.cty = 0;

    s_type = InInt();
    InTxFieldProps(&fieldProps, &xZoom, &yZoom, &draw_shadow, &shadow_only, &draw_emboss);

    /* clipping rectangle */
    InBlock((char *)&clip, sizeof(clip)) ;

    segCount = InInt(); /* # mono-attr segments in this text */
    while (segCount--)
    {
        piececount = InInt(); /* # 'default-positioned' pieces in seg */

        /* read attr info for this seg */
        InTextAttrs(&txProps, &fontName, &underPos, &underThick,
                    &strikePos);


        if ( s_type == 2 ) /* dark shadow */
        {
            no_prnt = axtGetShadowPrintInfo(txProps, draw_shadow, shadow_only, s_type);
            if ( no_prnt )
            {
                sdx = 0;
                sdy = 0;
            }
            else
            {
                sdx = txProps->fillpattern->s_xoff;
                sdy = txProps->fillpattern->s_yoff;
                txProps->fillpattern->ffg = txProps->fillpattern->sfg;
                txProps->fillpattern->fbg = txProps->fillpattern->sfg;
                txProps->fillpattern->lfg = txProps->fillpattern->sfg;
                txProps->fillpattern->lbg = txProps->fillpattern->sfg;
                sfg = 0;
                sfg |= txProps->fillpattern->ffg.c << 24;
                sfg |= txProps->fillpattern->ffg.m << 16;
                sfg |= txProps->fillpattern->ffg.y << 8;
                sfg |= txProps->fillpattern->ffg.k;
                txProps->fillcolor = sfg;
            }
        }
        else if ( s_type == 1 ) /* emboss shadow */
        {
            no_prnt = axtGetShadowPrintInfo(txProps, draw_shadow, shadow_only, s_type);
            if ( no_prnt )
            {
                sdx = 0;
                sdy = 0;
            }
            else if ( txProps->fillpattern->s_type < 3 )
            {
                sdx = 0;
                sdy = 0;
                no_prnt = TRUE;
            }
            else
            {
                sdx = (-1)*txProps->fillpattern->s_xoff;
                sdy = (-1)*txProps->fillpattern->s_yoff;

                GetWashoutColor(&txProps->fillpattern->sfg, &new_color);
                txProps->fillpattern->sfg.c = new_color.c;
                txProps->fillpattern->sfg.m = new_color.m;
                txProps->fillpattern->sfg.y = new_color.y;
                txProps->fillpattern->sfg.k = new_color.k;

                txProps->fillpattern->ffg = txProps->fillpattern->sfg;
                txProps->fillpattern->fbg = txProps->fillpattern->sfg;
                txProps->fillpattern->lfg = txProps->fillpattern->sfg;
                txProps->fillpattern->lbg = txProps->fillpattern->sfg;
                sfg = 0;
                sfg |= txProps->fillpattern->ffg.c << 24;
                sfg |= txProps->fillpattern->ffg.m << 16;
                sfg |= txProps->fillpattern->ffg.y << 8;
                sfg |= txProps->fillpattern->ffg.k;
                txProps->fillcolor = sfg;
            }
        }
        else /* normal text */
        {
            sdx = 0;
            sdy = 0;
            no_prnt = FALSE;
        }
    
        /* Is this font patterned? */
        /* register a need for the "equivalent" bitmap pattern of 
        a gray fill in the case that we must imagemask for it */
        if (txProps->fillpattern != NULL)
        {
            fill_attr = txProps->fillpattern;
            f_type = fill_attr->f_type;
            f_pat = fill_attr->f_pat;

            if (f_type == GRAY_FILL_TYPE)
                pattern[getIXfromPercent(f_pat)].needed = TRUE;

            if (f_type == BUILTIN_FILL_TYPE)
            {
                if (f_pat > MAX_PATS)   /* dirty hack */
                    f_pat = 4;        /* solid 100%... */
                pattern[f_pat].needed = TRUE;
            }

            if (f_type == LINEAR_FILL_TYPE ||
                f_type == BILINEAR_FILL_TYPE ||
                f_type == RECT_FILL_TYPE)
                PSbox_needed = TRUE;
            else if ( f_type == RADIAL_FILL_TYPE )
                PScirc_needed = TRUE;

            if (fill_attr->f_data && fill_attr->f_dhyt && fill_attr->f_dwid)
                pfSetPattern(fill_attr->f_type,
                                fill_attr->f_pat,
                                fill_attr->f_dhyt,
                                fill_attr->f_dwid,
                                fill_attr->f_ddep,
                                fill_attr->f_data,
                                fill_attr->f_cmap);

            PScharpath_needed = TRUE;
        }

        /* see whether this font declares an alternate for
         * use in printing (PostScriptAlternate)
         */
        if (errString = axGetFontRecord(fontName, &ftab))
        {
            freeErrString(errString);
            ftab = (fontTabEnt *) NULL;
            soft = NULL ;
        }
        else
        {
#if AXMS
            if ( download_to_postscript( ftab ) )
#else 
            if (ftab->gallium)
#endif
            {
                /* we will be down-loading this font as individual character */
                /* rasters as part of the print job                          */
                txProps->pointsize = (int)((double)txProps->pointsize/10000.0 *
                        fabs((double)fieldProps.yscale)) ;
                soft = pfSoftFontRegister(ftab, txProps->pointsize, &unit_ctm) ;
            }
            else
                soft = NULL ;
        }
        eightBit = wideChar = FALSE;
        while (piececount--)
        {
            pos.x = InInt();
            pos.y = InInt();
            width = InInt();
            txlen = InTxCharString(&textstr);
            if (txProps->ruling || txProps->strike_thru) {
                InInt();
                InInt();
                InInt();
                InInt();
            }
            for (ix = 0; ix < txlen; ix++)
            {
                if ( textstr[ix] > 0x7f && !soft)
                    eightBit = TRUE;

                if ( IS_ORIENT_CHAR( textstr[ix] ) )
                    wideChar = TRUE;

                if (soft)
                    pfSoftCharRegister(soft, textstr[ix]) ;
            }
            free(textstr);
            textstr = NULL;
        }
        if (ftab)
        {
            if(!ftab->PostScriptAlternate)
            {
                RecordNamedFont(ftab->globalName,
                                txProps->pointsize,
                                eightBit, wideChar);
            }
            else
            {
                /* get the alternate's record */
                if (errString = axGetFontRecord(ftab->PostScriptAlternate,
                                                &alttab))
                {
                    freeErrString(errString);
                }
                else
                {
                    RecordNamedFont(alttab->globalName,
                                    txProps->pointsize,
                                    eightBit, wideChar);
                }
            }
        }


        free(fontName);
        fontName = NULL;
        pfFreeTextAttrs(txProps, 1);
    }

    PF.fontNeeded = TRUE;
}


/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
static void D01PreScan()
{
    int error_check;
    long mark;
    int y;
    register int i;
    register int j;
    register int x;
    int ix;
    TmTextDescriptor defTextDes;
    int nPolys, nPts, nCtrls, fill_flag;
    int wid, f_type, f_pat, l_type, l_pat;
    register int hyt;
    register int byte_wid;
    int dep;
    int orient = 0;
    unsigned char *data;
    attr_type attr;
    TmRect32 clip ;
    char buf[512];
	int pw, ph ;

    PF.PfSEByteAlignment = 0;  /* to keep track of se file byte alignment */

    Pf_has_landscape = FALSE;

    mark = InMark();

    defTextDes = 0;

    /* mark all the builtin corresponding to gray fills as needed,
       'cause we can't tell for sure we'll need them */

    pattern[12].needed = TRUE ;
    pattern[1].needed = TRUE ;
    pattern[10].needed = TRUE ;
    pattern[2].needed = TRUE ;
    pattern[21].needed = TRUE ;
    pattern[3].needed = TRUE ;
    pattern[13].needed = TRUE ;

    PF.PfJob_done = FALSE;
    while (!PF.PfJob_done)
    {
        switch (InWord())
        {
        case SfCreate:
            if (PF.PfReverse)
                pf_markpage() ;
            PF.PfNumDocPages++;
            orient = InWord();
            pw = InInt();    /* page width */
            ph = InInt();    /* page height */
            InInt();    /* left margin */
            InInt();    /* use crops */
            InInt();    /* use regis */
            InInt();    /* use bounder */
            InInt();    /* bounder x */
            InInt();    /* bounder y */
            InInt();    /* bounder width */
            InInt();    /* bounder height */
            InInt();    /* clip x */
            InInt();    /* clip y */
            InInt();    /* clip width */
            InInt();    /* clip height */
            InInt();    /* input tray */
            InInt();    /* output tray */
            InInt();    /* duplex */


            /* this document has a landscape page? */
            if (orient == 1)
                Pf_has_landscape = TRUE;

            /* save size of 1st page for possible banner page */
			if (PF.PfNumDocPages == 1)
			{
				PF.PfPageMilWidth = pw ;
				PF.PfPageMilHeight = ph ;
			}
            break;

        case SfTerm:
            if (PF.PfReverse)
                pf_markpage() ;
            PF.PfJob_done = TRUE;
            break;

        case SfPolyline:
            /* setup what patterns are needed */
            total_lines++;
            nPolys = InWord ();

            /* pass the points */
            for ( i=0; i<nPolys; i++ )
            {
                /* read in the poly points */
                nPts = InWord ();

                for ( j=0; j<nPts; j++ )
                {
                    InInt ();
                    InInt ();
                }
            
                /* read in the control points */
                nCtrls = InWord ();
                for ( j=0; j<nCtrls; j++ )
                {
                    InInt ();
                    InInt ();
                }
            }

            /* read the attributes */
            memset((char *)&attr, 0, sizeof(attr_type));
            pf_read_poly_attr(&attr);

            /* clipping rectangle */
            InBlock((char *)&clip, sizeof(clip)) ;

            nPolys = InWord ();

			if ( nPolys > 0 )
			{
				/* pass the points */
				for ( i=0; i<nPolys; i++ )
				{
					/* read in the poly points */
					nPts = InWord ();

					for ( j=0; j<nPts; j++ )
					{
						InInt ();
						InInt ();
					}
				
					/* read in the control points */
					nCtrls = InWord ();
					for ( j=0; j<nCtrls; j++ )
					{
						InInt ();
						InInt ();
					}
				}
			}

            if (attr.fill_flag)
            {
                switch (attr.f_type)
                {
                case GRAY_FILL_TYPE:
                    pattern[getIXfromPercent(attr.f_pat)].needed = TRUE;
                    break ;

                case BUILTIN_FILL_TYPE:
                    pattern[attr.f_pat].needed = TRUE;
                    break ;

                case LINEAR_FILL_TYPE:
                case BILINEAR_FILL_TYPE:
                case RECT_FILL_TYPE:
                    PSbox_needed = TRUE;
                    break ;

                case RADIAL_FILL_TYPE:
                    PScirc_needed = TRUE;
                    break ;

                case USERDEF_FILL_TYPE:
                    pfSetPattern(attr.f_type,
                                attr.f_pat,
                                attr.f_dhyt,
                                attr.f_dwid,
                                attr.f_ddep,
                                attr.f_data,
                                attr.f_cmap);
                    break ;
                }
            }
            if (attr.line_flag)
            {
                switch (attr.l_type)
                {
                case GRAY_FILL_TYPE:
                    pattern[getIXfromPercent(attr.l_pat)].needed = TRUE;
                    break ;

                case BUILTIN_FILL_TYPE:
                    pattern[attr.l_pat].needed = TRUE;
                    break ;

                case LINEAR_FILL_TYPE:
                case BILINEAR_FILL_TYPE:
                case RECT_FILL_TYPE:
                    PSbox_needed = TRUE;
                    break ;

                case RADIAL_FILL_TYPE:
                    PScirc_needed = TRUE;
                    break ;

                case USERDEF_FILL_TYPE:
                    pfSetPattern(attr.l_type,
                                attr.l_pat,
                                attr.l_dhyt,
                                attr.l_dwid,
                                attr.l_ddep,
                                attr.l_data,
                                attr.l_cmap);
                    break ;
                }
            }
            if (attr.f_data)
                free(attr.f_data);

            if (attr.l_data)
                free(attr.l_data);

            break;

        case SfArc:
#if 0
            /* obsolete .... */
            /* read the ellipse data */
            InWord (); InWord (); InWord ();
            InWord (); InWord (); InWord ();
 
            /* read in the fill attributes */
            fill_flag = InWord();
            InWord();
            f_type = InWord();
            f_pat = InWord();
            if ( f_type == 0 ) /* built in fill */
                pattern[f_pat].needed = TRUE;
            if (f_type == LINEAR_FILL_TYPE || f_type == RECT_FILL_TYPE)
              PSbox_needed = TRUE;
            else if (f_type == RADIAL_FILL_TYPE)
            PScirc_needed = TRUE;
            InWord(); InWord(); InWord();

            /* read any passed in fill fill pattern */
            hyt = InWord();
            wid = InWord();
            byte_wid = (wid+7) / 8;
            data = (unsigned char *)malloc(hyt * byte_wid);
            for ( i=0; i<hyt; i++ )
                for (j=0; j<byte_wid; j++ )
                    data[i*byte_wid + j] = (unsigned char)InByte();
            InAlign();
            pfAddFillToList(f_type, hyt, wid, 1, data, NULL);
            free(data);
            data = NULL;
       
            /* read the line attributes */
            InWord(); InWord(); InWord();
            l_type = InWord ();
            l_pat = InWord ();
            if ( l_type == 0 ) /* built in fill */
                pat_needed[l_pat] = TRUE;
            if (l_type == LINEAR_FILL_TYPE || l_type == RECT_FILL_TYPE)
                PSbox_needed = TRUE;
            else if (l_type == RADIAL_FILL_TYPE)
                PScirc_needed = TRUE;
            InWord(); InWord(); InWord();
 
            /* read any passed in line fill pattern */
            hyt = InWord();
            wid = InWord();
            byte_wid = (wid+7) / 8;
            data = (unsigned char *)malloc(hyt * byte_wid);
            for ( i=0; i<hyt; i++ )
                for (j=0; j<byte_wid; j++ )
                    data[i*byte_wid + j] = (unsigned char)InByte();
            InAlign();
            pfAddFillToList(f_type, hyt, wid, 1, data, NULL);
            free(data);
            data = NULL;
 
            /* read the color attributes and their overrides */
            InWord (); InWord (); InWord (); InWord ();
            InWord (); InWord (); InWord (); InWord ();
            InWord (); InWord (); InWord (); InWord ();
            InWord (); InWord (); InWord (); InWord ();
            InWord (); InWord (); InWord (); InWord ();
#endif
            break;

        case SfSetWinAttr:
            InWord ();  /* line style */
            InWord ();  /* line thickness */
            defTextDes = InLong (); /* text descriptor */
            InWord ();  /* transfer mode */
            InWord ();  /* text alignment */
            if (PF.PfSe_version >= 4)
            {
                InWord();   /* foreground color */
                InWord();   /* background color */
            }
            InWord ();  /* clip rect */
            InWord ();
            InWord ();
            InWord ();
            break;

        case SfGMode:
            InWord ();  /* On off flag */
            InWord ();  /* graphics rectangle */
            InWord ();
            InWord ();
            InWord ();
            break;

        case SfImage:
            InWord ();  /* pixel flag */
            InWord ();  /* image mode */
            InByte();   /* attr->ffg.c */
            InByte();   /* attr->ffg.m */
            InByte();   /* attr->ffg.y */
            InByte();   /* attr->ffg.k */
            InByte();   /* attr->fbg.c */
            InByte();   /* attr->fbg.m */
            InByte();   /* attr->fbg.y */
            InByte();   /* attr->fbg.k */
            InInt ();   /* xpos */
            InInt ();   /* ypos */
            x = InInt ();   /* xsize */
            i = InInt ();   /* ysize */
            dep = InWord ();    /* image_depth */

            InLong (); InLong ();   /* the xform (five points). Should be ints? */
            InLong (); InLong ();
            InLong (); InLong ();
            InLong (); InLong ();
            InLong (); InLong ();
#if BYTE_WID_IN_SE
            j = InWord();
#else
            j = EVEN_BYTES_FROM_BITS(x * dep);
#endif

            /* there MUST be a better way of reading an arbitrary
               number of bytes... */
            for (y=0; y<i; y++)
                    for (x=0; x<j; x++)
                        InByte();
            /* MASK_IN_STREAM */
            InAlign();
            j = InWord();   /* BYTE wid; not bit! */
            if (j != 0)
            {
                for (y=0; y<i; y++)
                    for (x=0; x<j; x++)
                        InByte();
            }

            i = InLong();   /* cmap->n_colors */
            for (y=0; y<i; y++)
            {
                InInt();
                InByte();   /* c */
                InByte();   /* m */
                InByte();   /* y */
                InByte();   /* k */
            }

            /* clipping rectangle */
            InBlock((char *)&clip, sizeof(clip)) ;

            /* Now for the EPS data */
            i = InLong();
            if (i > 0)
            {
                PF.PfNumEPSIncluded++;
                for (y=0; y<i; y++)
                    InByte();
                if (i & 1)  /*always write an even # of bytes */
                    InByte();
            }
            break;

        case SfTabs:
            pf_tabs();  /* no harm in setting them */
            break;

        case SfScaledTxt:
            InWord(); /* on/off */
            InWord(); /* imageBox[0].x */
            InWord(); /* imageBox[0].y */
            InWord(); /* imageBox[1].x */
            InWord(); /* imageBox[1].y */
            InWord(); /* imageBox[2].x */
            InWord(); /* imageBox[2].y */
            InWord(); /* imageBox[3].x */
            InWord(); /* imageBox[3].y */
            break;

        case SfInclude:
            /* PreScanEPSInclude(); */  /* obsolete */
            break;

        case SfText:
            total_texts++;
            D01PreScanText();
            break;

        case SfSave:
        case SfRestore:
            break;

        case SfClip:
            InBlock((char *)&clip, sizeof(clip)) ;
            break;

        case SfPrControl:
            InInt() ;
            i = InInt() ;
            InBlock(buf, i) ;
            InAlign();
            break;
            
        default:
            break;
        }

        error_check = InWord ();
        if (error_check!=SfEndCall)
        {
            Debugf ("Expected end of call, found %d\n", error_check);
            PFabort ("File out of synch - didn't read end of call");
        }

    }
    InGotoMark(mark);
    PF.PfJob_done = FALSE;

    return;
}  /* D01PreScan() */



/*-----------------------------------------------------------------------------
 *   We scan the document to find out what fonts we need, and
 *   whether they are eightbit.  We then create the fonts
 *   in the PostScript prolog.
 *   Note that we check the banner explicitly for eightbit -
 *   the banner isn't in the input file.
 *   We also record all patterns used, and put them in
 *   the prolog, too.
-----------------------------------------------------------------------------*/
int D01StartJob()
{
    int i, tkn;
    int ix, namelen, len, numFontNames;
    char *prof_str;
    typeFuncs tfuncs;
    char **pFontNames;
    printFontName *pn;
	char *banner_font ;

    which_line = 0;
    total_lines = 0;
    which_text = 0;
    total_texts = 0;
    current_clip[0].x1 = 0 ;
    current_clip[0].y1 = 0 ;
    current_clip[0].x2 = 0 ;
    current_clip[0].y2 = 0 ;

    duplex = -1 ;
    duplex_page = 0 ;
    input_tray = -1 ;
    output_tray = -1 ;
    page_started = FALSE ;

    PF.PfXPixOffset += XTRA_XOFF;

    PF.eightbit = FALSE;
    PF.ClipGSaveLevel = 0;
    PF.fontNeeded = FALSE;
    PF.PfNumEPSIncluded = 0;
    PF.PfNumEPSAddedFonts = 0;
    PF.PfNumEPSProcessed = 0;
    PSbox_needed = FALSE;
    PScirc_needed = FALSE;
    PScharpath_needed = FALSE;

    PF.left = 0;
    PF.right = 0;
    PF.top = 0;
    PF.bottom = 0;
    tkn = 0;

    PF.cur_thick = MINIMUM_LINE_WIDTH;        /* setup below in DOSETUP */
    PF.cur_style = TM_SOLID;                    /* setup below in DOSETUP */

    prof_str = GetProfileByStr(NXLT("axPrintMaxPathPoints"));
        
    if ((prof_str == NULL) || (*prof_str == '\0'))
        max_path_points = MAXINT ;                /* default behavior */
    else if (strcasecmp(prof_str, NXLT("null")) == 0)
        max_path_points = MAX_PATH_POINTS;                /* default behavior */
    else
    {
        max_path_points = atoi(prof_str);
        if (max_path_points == 0)
            max_path_points = MAXINT;
    }

    prof_str = GetProfileByCode(PROF_PRINT_COLOR_MODEL_);
    if (prof_str != NULL)
    {
        pf_color_model = atoi(prof_str);
        switch (pf_color_model)
        {
        default:
        case COL_MODEL_CMYK :
            numColorComponents = 4;              
            break;
        case COL_MODEL_RGB :
            numColorComponents = 3;
            break;
        case COL_MODEL_GRAY :
            numColorComponents = 1;
            break;
        }
    }
    else
    {
        numColorComponents = 4;
    }

    prof_str = GetProfileByCode(PROF_PRINT_BITS_);
    if (prof_str != NULL)
    {
        bitsPerColorComponent = atoi(prof_str);
        if (bitsPerColorComponent == 8)
        {
            bitsPerGrayComponent = 8;
        }
        else
        {
            bitsPerColorComponent = 4;
            bitsPerGrayComponent = 4;
        }
    }
    else
    {
        bitsPerColorComponent = 4;
        bitsPerGrayComponent = 4;
    }

    /* initialize font resource accounting */

    /* init list of print font names */
    Curfont.number = Curfont.size = -1;
    axhInitTable(&printFontNames, 31, getPrintFontName, freePrintFontName);

    /* init list of print font encoding structs */
    axhInitTable(&printFontEncodings, 7, getPrintFontEncoding,
                 freePrintFontEncoding);

    /* Unconditionally register need for registration font (18pt Times) */
    RecordNamedFont(NXLT("Helvetica"), 180000, FALSE, FALSE);

    /* Conditionally register need for the banner font */
    if (PF.PfBanner)
	{
		banner_font = NXLT("Courier") ;
        if (!RecordNamedFont(banner_font, banner_font_size * 10000, FALSE, FALSE))
		{
			banner_font = NXLT("Helvetica") ;
	        RecordNamedFont(banner_font, banner_font_size * 10000, FALSE, FALSE) ;
		}
	}

    /* initialize pattern-matching support which put() uses */
    pf70InitPatBuf();

    pfClearPattern() ;
    D01PreScan();

    /* find PPD file if one specified */
    PPDfile = NULL ;
    if (PF.pi.PPDfile && *PF.pi.PPDfile != 0)
    {
        PPDfile = OpenAxLocal(PF.pi.PPDfile) ;
        initPPDkeys() ;
    }

    PF.PfNumTotalPages = PF.PfNumDocPages;
    if (PF.PfBanner)
        PF.PfNumTotalPages++;
#if 0
    if ( PfJobToPSPrinter && (PF.PfNumDocPages == 1) && (PF.PfCopies > 1) ) {
        /* force usage of #copies with showpage instead
         * of multiple shell commands for one page docs */
        PF.PfPrintMultiCopiesInLang = TRUE;
    }
    if ( PF.PfPrintMultiCopiesInLang > 1)
        PF.PfBanner = FALSE;
#endif

    if (PF.PfBanner)
    {
        int loclen;
        int banner_eightbit;

        banner_eightbit = FALSE;

        loclen = strlen(PF.PfDoc_name);
        for(i=0; i<loclen; i++)
            if (PF.PfDoc_name[i] & 128)
            {
                banner_eightbit = TRUE;
                goto record_it;
            }

#if 0
        loclen = strlen(PF.PfCab_name);
        for(i=0; i<loclen; i++)
            if (PF.PfCab_name[i] & 128)
            {
                banner_eightbit = TRUE;
                goto record_it;
            }
#endif

        loclen = strlen(PF.PfUser_name);
        for(i=0; i<loclen; i++)
            if (PF.PfUser_name[i] & 128)
            {
                banner_eightbit = TRUE;
                goto record_it;
            }

        loclen = strlen(PF.PfHost_name);
        for(i=0; i<loclen; i++)
            if (PF.PfHost_name[i] & 128)
            {
                banner_eightbit = TRUE;
                goto record_it;
            }

        loclen = strlen(PF.PfDoc_date);
        for(i=0; i<loclen; i++)
            if (PF.PfDoc_date[i] & 128)
            {
                banner_eightbit = TRUE;
                goto record_it;
            }
record_it:
        /* update as eightbit */
        RecordNamedFont(banner_font, banner_font_size * 10000, banner_eightbit, FALSE);
    }

    PF.pagetop = PF.PfPagePixHeight;

    if (InsertSpec = openPSinsertsFile(&len))
    {
        /* there's some extra postscript code to possibly
         * insert.  setup the PSinserts global with this data. */
        initPSinserts(InsertSpec, len, &PfPsInsertData, PF.PfPrinterName);
        freeInsertsFile(InsertSpec);
        InsertSpec = ((char *)NULL);
    }

    put(NXLT("%%!PS-Adobe-3.0\n"));
    D01AddGeneralHeaderComments();

    if ( PF.PfPrintMultiCopiesInLang > 1)
        put(NXLT("%%%%Requirements: numcopies(%d)\n"), PF.PfPrintMultiCopiesInLang);

    if ( PF.fontNeeded )
    {
        put(NXLT("%%%%DocumentNeededResources: font"));
        len = 31;       /* length of "%%DocumentNeededResources: font" */

        pFontNames = (char **) 0;
        axhListNames(&printFontNames, &pFontNames, &numFontNames);    
        for (ix = 0; ix < numFontNames; ix++)
        {
            pn = (printFontName *) 0;
            axhGetItem(pFontNames[ix], &printFontNames,
                       (char *) 0, (char **) &pn);
            if (pn)
            {
                namelen = strlen(pn->psFontName) + 1;   /* add space */
                if ( (len + namelen) > 78 )
                {
                    put(NXLT("\n%%%%+ font"));  /* use new line */
                    len = 8;  /* length of "%%+ font" */
                }
                put(NXLT(" %s"), pn->psFontName);
                len += namelen;
            }
        }
        put(NXLT("\n"));
        axhFreeListNames(pFontNames);
    }

    put(NXLT("%%%%EndComments\n"));

    put(NXLT("%%%%BeginProlog\n"));
    checkInsert(&PfPsInsertData, InsertAtPrologTop);
#if 0
    put(NXLT("systemdict /setglobal known {false setglobal} if\n"));
#endif
    put(NXLT("/bd { bind def } bind def\n"));
    put(NXLT("/n  { newpath } bd\n"));
    put(NXLT("/L  { lineto } bd\n"));
    put(NXLT("/M  { moveto } bd\n"));
    put(NXLT("/C  { curveto } bd\n"));
    put(NXLT("/RL { rlineto } bd\n"));
    put(NXLT("/MR { rmoveto } bd\n"));
    put(NXLT("/ST { show } bd\n"));
    put(NXLT("/S  { stroke } bd\n"));
    put(NXLT("/SP { strokepath } bd\n"));
    put(NXLT("/GS { gsave } bd\n"));
    put(NXLT("/GR { grestore } bd\n"));
    put(NXLT("/GRAY { setgray } bd\n"));
    if (PSbox_needed)
    {
        put(NXLT("/BOX {\n"));                /* width height BOX */
        put(NXLT("    /Boxheight exch def\n"));
        put(NXLT("    /Boxwidth exch def\n"));
        put(NXLT("    Boxwidth 0 rlineto\n"));  /* along the bottom */
        put(NXLT("    0 Boxheight rlineto\n")); /* up the rhs */
        put(NXLT("    Boxwidth neg 0 rlineto\n"));  /* along the top */
        put(NXLT("    closepath\n"));
        put(NXLT("} bd\n"));
    }
    if (PScirc_needed)
    {
        put(NXLT("/CIRC\n"));   /* x y r CIRC => */
        put(NXLT("{\n"));
        put(NXLT("    0 360 arc\n"));
        put(NXLT("    closepath\n"));
        put(NXLT("} bd\n"));
        if (PF.PfColordev)
        {
            put(NXLT("/COLORCIRC\n"));  /* c m y k x y r COLORCIRC => */
            put(NXLT("{\n"));
            put(NXLT("    CIRC\n"));
            put(NXLT("    SET_CMYK\n"));
            put(NXLT("    fill\n"));
            put(NXLT("} bd\n"));
        }
        else
        {
            put(NXLT("/GRAYCIRC\n"));   /* g x y r GRAYCIRC => */
            put(NXLT("{\n"));
            put(NXLT("    CIRC\n"));
            put(NXLT("    setgray\n"));
            put(NXLT("    fill\n"));
            put(NXLT("} bd\n"));
        }
    }
    /* should really separate these into Xscale/Yscale */
    put(NXLT("/AXscale { 72 %d div } bd\n"), xmiltopix(1000));
    put(NXLT("/DOSETUP {\n"));
    put(NXLT("      AXscale dup scale\n"));
    put(NXLT("      1 setlinecap\n"));
    put(NXLT("      %d setlinewidth\n"), MINIMUM_LINE_WIDTH);
    put(NXLT("      1 setlinejoin\n"));
    put(NXLT("      4 setmiterlimit\n"));
    put(NXLT("      [] 0 setdash\n"));
    put(NXLT("} bd\n"));
    put(NXLT("/DOCLIPBOX {\n"));
    put(NXLT("    n 4 2 roll 2 copy M 4 1 roll exch 2 copy L exch pop 2 copy L pop\n")) ;
    put(NXLT("    exch L closepath clip\n")) ;
    put(NXLT("} bd\n"));

    if ( Pf_has_landscape == TRUE )
    { /* this doc has landscape pages */
        put(NXLT("/DOLANDSCAPE {\n"));
        put(NXLT("      90 rotate\n"));
        put(NXLT("      0 exch -1 mul translate\n"));   /* pagetop /DOLANDSCAPE => */
        put(NXLT("} bd\n"));
        put(NXLT("/UNDOLANDSCAPE {\n"));
        put(NXLT("      0 exch translate\n"));  /* -ve off */
        put(NXLT("      -90 rotate\n"));
        put(NXLT("} bd\n"));
    }

    put(NXLT("/AXredict 14 dict def\n"));
    put(NXLT("/X {\n"));
    put(NXLT("      AXredict begin\n"));
    put(NXLT("      exch /str exch def\n"));
    put(NXLT("      str stringwidth pop sub\n"));
    put(NXLT("      str length div 0 str ashow\n"));
    put(NXLT("      end\n"));
    put(NXLT("} bd\n"));

    if (PScharpath_needed )
    {
        put(NXLT("/Xc {\n"));
        put(NXLT("     AXredict begin\n"));
        put(NXLT("     true charpath\n"));
        put(NXLT("     end\n"));
        put(NXLT("} bd\n"));
    }
    /* Font selects: if the desired face isn't
     * on the printer, default to (standard) Times.
     */
    put(NXLT("/FINDFONT {\n"));
    put(NXLT("    { findfont } stopped\n"));
    put(NXLT("    { /Times-Roman findfont } if\n"));
    put(NXLT("} bd\n"));

    put(NXLT("/POINTSCALEFONT { AXscale div scalefont } bd\n"));
    put(NXLT("/DOTSCALEFONT { scalefont } bd\n"));
    put(NXLT("/SETFONT { setfont } bd\n"));

    /* define EPSF setup procs if included */
    if ( PF.PfNumEPSIncluded )
    {
        put(NXLT("/BEGINEPSF {\n"));
        put(NXLT("      save /AXEPSFSave exch def\n"));
        put(NXLT("      /epsfDictCount countdictstack def\n"));
        put(NXLT("      /epsfOpCount count 1 sub def\n"));
        put(NXLT("      userdict begin\n"));
        put(NXLT("      /showpage {} def\n"));
        put(NXLT("      0 setgray 0 setlinecap\n"));
        put(NXLT("      0 setlinejoin 10 setmiterlimit\n"));
        put(NXLT("      [] 0 setdash 1 setlinewidth n\n"));
        put(NXLT("} bd\n"));
        put(NXLT("/ENDEPSF {\n"));
        put(NXLT("      count epsfOpCount sub {pop} repeat\n"));
        put(NXLT("      countdictstack epsfDictCount sub {end} repeat\n"));
        put(NXLT("      AXEPSFSave restore\n"));
        put(NXLT("} bd\n"));
    }


    /* Load pattern definitions.  If pattern font is to be
     * downloaded, do in document setup. */
    D01PreLoadPatterns();

    D01PreLoadFonts();

    checkInsert(&PfPsInsertData, InsertAtPrologBottom);
    put(NXLT("%%%%EndProlog\n"));

    if ( PF.fontNeeded || PF.PfColordev || PF.PfPrintMultiCopiesInLang )
    {
        put(NXLT("%%%%BeginSetup\n"));
        checkInsert(&PfPsInsertData, InsertAtSetupTop);
            /* any adobe printer description file (ppd) data will be
             * included here in the document setup */
        checkPPD(&PfPsInsertData);
        if ( PF.fontNeeded )
        {
            numFontNames = 0;
            axhListNames(&printFontNames, &pFontNames, &numFontNames);    
            for (ix = 0; ix < numFontNames; ix++)
            {
                pn = (printFontName *) 0;
                axhGetItem(pFontNames[ix], &printFontNames, (char *) 0, (char **)&pn);
                if (pn)
                {
                    put(NXLT("%%%%IncludeResource: font %s\n"),
                        pn->psFontName);
                }
            }
            put(NXLT("\n"));
            axhFreeListNames(pFontNames);
            put(NXLT("DEFINEFONTS\n"));
        }

        if(PF.PfColordev)
        {
            /*
            *  Check if enhanced color operators are known to the printer,
            *  for example: "setcmykcolor".  If not, define "setcmykcolor" to
            *  an equivalent gray value.  Also define the halftone/color 
            *  screen.
            */
            put(NXLT("systemdict /setcmykcolor known\n"));
            put(NXLT("{\n"));
            put(NXLT("       /SET_CMYK { setcmykcolor } bd\n"));
            put(NXLT("}\n"));
            put(NXLT("{\n"));
            put(NXLT("       /SET_CMYK {\n"));
            put(NXLT("       exch .2 mul add\n"));
            put(NXLT("       exch .4 mul add\n"));
            put(NXLT("       exch .3 mul add\n"));
            put(NXLT("       dup 1 gt\n"));
            put(NXLT("       {pop 1} {} ifelse\n"));
            put(NXLT("       1 exch sub setgray\n"));
            put(NXLT("       } bd\n"));
            put(NXLT("}\n"));
            put(NXLT("ifelse\n"));

            put(NXLT("systemdict /colorimage known\n"));
            put(NXLT("{\n"));
            put(NXLT("       /GET_CMYK { currentcmykcolor } bd\n}\n{\n"));
            put(NXLT("       /GET_CMYK {\n"));
            put(NXLT("       0 0 0 \n"));
            put(NXLT("       1 currentgray sub\n"));
            put(NXLT("       } bd\n}\nifelse\n"));

            put(NXLT("systemdict /colorimage known\n"));
            switch (pf_color_model)
            {
                default:
                case COL_MODEL_CMYK:
                    put(NXLT("{\n   /COLORIMAGE { false 4 colorimage } bd\n"));
                    break;
                case COL_MODEL_RGB:
                    put(NXLT("{\n   /COLORIMAGE { false 3 colorimage } bd\n"));
                    break;
                case COL_MODEL_GRAY:
                    put(NXLT("{\n   /COLORIMAGE { false 1 colorimage } bd\n"));
                    break;
            }
            put(NXLT("       /SELECTBUF { pop } bd\n}\n{\n"));
            put(NXLT("       /COLORIMAGE { image } bd\n"));
            put(NXLT("       /SELECTBUF { exch pop } bd\n"));
            put(NXLT("}\nifelse\n"));
        }

        checkInsert(&PfPsInsertData, InsertAtSetupBottom);

        put(NXLT("%%%%EndSetup\n"));
    }

    /* print the banner page */
    if (PF.PfBanner)
    {
        StdBanner(banner_font, banner_font_size, 500);
        D01EndPage() ;
    }

    if (PF.pi.duplex >= 0)
        select_plex(PF.pi.duplex) ;

    return(0);
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
int D01EndJob()
{

    put(NXLT("%%%%Trailer\n"));
    checkInsert(&PfPsInsertData, InsertAtTrailerTop);
    checkInsert(&PfPsInsertData, InsertAtTrailerBottom);
    put(NXLT("%%%%EOF\n"));

    /* add ctrl-D to end PS job */
    put(NXLT("%c"), 4);

    if (PfPsInsertData.numInserts)
    {
        put((char *)NULL);
        freePSinserts(&PfPsInsertData);
    }

    /* reset for next job */
    axhDeleteTable(&printFontNames, (char *)NULL);
    memset((void *) &printFontNames, 0, sizeof(axhtab));
    axhDeleteTable(&printFontEncodings, (char *)NULL);
    memset((void *) &printFontEncodings, 0, sizeof(axhtab));
    PF.D70FontNumber = 0;
    pfResetPattern();
    pf70InitPatBuf();
    if (PPDfile)
        fclose(PPDfile) ;

    PF_MEM_CHECK;
    return(0);
}


/*-----------------------------------------------------------------------------
 *  Save state, print page and erase, restore state.
-----------------------------------------------------------------------------*/
int D01NewPage(firstPage, clip)
    int firstPage; /* bool: true for 1st physical page of printed output */
    TmRect32 *clip ;
{

    /* finish up the previous page (if there is one) */
    if ( page_started )
        D01EndPage() ;

    /* set up for next page */
    if ( PF.PfPageNum <= PF.PfNumDocPages )
    {
        /* Page comment is for utility programs. */
        if ( PF.PfPageNum == 0 )            /* this is the banner page */
            put(NXLT("%%%%Page: banner %d\n"), PF.PfPageNum);
        else
            put(NXLT("%%%%Page: %d %d\n"), PF.PfPageNum, PF.PfPageNum);
        put(NXLT("%%%%BeginPageSetup\n"));
        checkInsert(&PfPsInsertData, InsertAtPageSetupTop);
        if (PF.PfPageNum == 1)
        {
            checkInsert(&PfPsInsertData, InsertAtFirstPageSetupTop);
            if ( PF.PfPrintMultiCopiesInLang > 1)
                put(NXLT("/#copies %d def\n"), PF.PfPrintMultiCopiesInLang);
        }
        /* select source for this page */
        select_source_tray() ;

        /* select page size */
        if ( PF.PfLandscape )
            select_size(PF.PfPageMilHeight, PF.PfPageMilWidth) ;
        else
            select_size(PF.PfPageMilWidth, PF.PfPageMilHeight) ;

        /* select destination for this page */
        select_destination() ;

        /* select plex */
        if (PF.pi.duplex == -1 && PF.PfPageNum > 0)
            select_plex(PF.duplex) ;

        put(NXLT("save /AXPageSave exch def\nDOSETUP\n"));

        if ( PF.PfLandscape )
            put(NXLT("%d DOLANDSCAPE\n"), PF.pagetop);

        if (clip != NULL && clip->x_size != 0 && clip->y_size != 0)
            clip_box(clip->x_position,
                    clip->y_position,
                    clip->x_position + clip->x_size,
                    clip->y_position + clip->y_size) ;
        else
            clip_box(0, 0, PF.PfPageMilWidth, PF.PfPageMilHeight) ;

        checkInsert(&PfPsInsertData, InsertAtPageSetupBottom);
        if (PF.PfPageNum == 1)
            checkInsert(&PfPsInsertData, InsertAtFirstPageSetupBottom);
        put(NXLT("%%%%EndPageSetup\n"));
        page_started = TRUE ;
    }

    PF.ClipGSaveLevel = 0;                /* just making sure */

    PF.left = 0;
    PF.right = 0;
    PF.top = 0;
    PF.bottom = 0;

    PF.cur_thick = MINIMUM_LINE_WIDTH;    /* this is setup in DOSETUP */
    PF.cur_style = TM_SOLID;                /* this is setup in DOSETUP */

    Curfont.number = Curfont.size = -1;

    /* reset the appclip to the page clip */
    PF.PfPage.appClip = PF.PfPage.inner;

    /* set the new page top */
    PF.PfPagePixHeight = ymiltogpix(PF.PfPageMilHeight);
    PF.PfPagePixWidth = xmiltogpix(PF.PfPageMilWidth);
    PF.pagetop = PF.PfPagePixHeight;

    return(0);
}

/*-----------------------------------------------------------------------------
    EndPage
-----------------------------------------------------------------------------*/
static void D01EndPage()

{
    while ( PF.ClipGSaveLevel > 0 ) /* should loop only once */
    {
        PS_GR();
        }
    put(NXLT("%%%%PageTrailer\n"));
    checkInsert(&PfPsInsertData, InsertAtPageTrailerTop);
    if ( PF.PfPageNum == 2 )
        checkInsert(&PfPsInsertData, InsertAtFirstPageTrailerTop);
    put(NXLT("AXPageSave restore\n"));

    put(NXLT("showpage\n"));
    if ( PF.PfPageNum > 0 )
        checkInsert(&PfPsInsertData, InsertAtPageTrailerBottom);
    if ( PF.PfPageNum == 2 )
        checkInsert(&PfPsInsertData, InsertAtFirstPageTrailerBottom);
    page_started = FALSE ;
}

/*-----------------------------------------------------------------------------
    select_source_tray()
-----------------------------------------------------------------------------*/
static void select_source_tray()
{
    char *s ;

    if (PF.PfPageNum == 1 && PF.pi.source_1st)
        InsertPSFragment(PF.pi.source_1st) ;

    else if (PF.PfPageNum == PF.PfNumDocPages && PF.pi.source_last)
        InsertPSFragment(PF.pi.source_last) ;

    else if (PF.pi.source_all)
        InsertPSFragment(PF.pi.source_all) ;

    else if (input_tray != PF.inputTray)
    {
        s = PS_source_tray(PF.inputTray) ;
        if (s)
            InsertPSFragment(s) ;
        input_tray = PF.inputTray ;
    }
}

/*-----------------------------------------------------------------------------
    PS_source_tray
-----------------------------------------------------------------------------*/
static char *PS_source_tray(n)
    int n ;
{
    int k ;
    char *p = NULL ;

    if (PF.pi.source)
    {
        k = AxArraySize(PF.pi.source) ;
        if (n < k)
            p = AxStrFromArray(PF.pi.source, n) ;
    }
    return p ;
}

/*-----------------------------------------------------------------------------
    select_size
----------------------------------------------------------------------------*/
static void select_size(w, h)
    int w, h ;
{
    int i, k ;
    data_ptr z ;
    int wz, hz1 ;
    char *p ;

#if 0
    if (w == page_width && h == page_height)
        return ;

    page_width = w ;
    page_height = h ;
#endif

    if (duplex > 0)
        return ;

    if (PF.pi.size)
    {
        k = AxArraySize(PF.pi.size) ;
        for (i = 0 ; i < k ; i++)
        {
            z = AxArrayFromArray(PF.pi.size, i) ;
            wz = AxIntFromArray(z, 0) ;
            hz1 = AxIntFromArray(z, 1) ;
            if (quite_close(w, wz) && quite_close(h, hz1))
            {
                p = AxStrFromArray(z, 2) ;
                if (p)
                    InsertPSFragment(p) ;
                return ;
            }
        }
    }
}

/*-----------------------------------------------------------------------------
    quite_close
-----------------------------------------------------------------------------*/
static int quite_close(a, b)
    int a, b ;
{
    return a >= b-4 && a <= b+4 ;
}


/*-----------------------------------------------------------------------------
    select_destination
-----------------------------------------------------------------------------*/
static void select_destination()
{
    char *s ;

    if (PF.PfPageNum == 1 && PF.pi.dest_1st)
        InsertPSFragment(PF.pi.dest_1st) ;

    else if (PF.PfPageNum == PF.PfNumDocPages && PF.pi.dest_last)
        InsertPSFragment(PF.pi.dest_last) ;

    else if (PF.pi.dest_all)
        InsertPSFragment(PF.pi.dest_all) ;

    else if (output_tray != PF.outputTray)
    {
        s = PS_dest_tray(PF.outputTray) ;
        if (s)
            InsertPSFragment(s) ;
        output_tray = PF.outputTray ;
    }
}

/*-----------------------------------------------------------------------------
    PS_dest_tray
-----------------------------------------------------------------------------*/
static char *PS_dest_tray(n)
    int n ;
{
    int k ;
    char *p = NULL ;

    if (PF.pi.dest)
    {
        k = AxArraySize(PF.pi.dest) ;
        if (n < k)
            p = AxStrFromArray(PF.pi.dest, n) ;
    }
    return p ;
}

/*-----------------------------------------------------------------------------
    select_plex
-----------------------------------------------------------------------------*/
static void select_plex(p)
    int p ;
{
    int k ;
    char *s ;

    if (p != duplex)
    {
        if (p >= 0)
        {
            if (PF.pi.duplex_code)
            {
                k = AxArraySize(PF.pi.duplex_code) ;
                if (p < k)
                {
                    s = AxStrFromArray(PF.pi.duplex_code, p) ;
                    InsertPSFragment(s) ;
                }
            }
        }
        duplex = p ;
        duplex_page = 0 ;
    }
}

/*-----------------------------------------------------------------------------
    D01Printmarks
-----------------------------------------------------------------------------*/
int D01Printmarks(use_crops, use_regis, use_bounds, bounds)
    int use_crops, use_regis, use_bounds;
    TmRect32 *bounds;
{
    int i, edge_x1, edge_y1, edge_x2, edge_y2, page_wid, page_hyt, margin;
    TmRect32 pf_bounds;

    if ( use_crops || use_regis || use_bounds )
    {
        /* save the current printer state */
        put(NXLT("%%Printer marks\nGS\n"));

        /* set the line color to black */
        D01SetColor(TMA_BLACK);

        pf_bounds.x_position = xmiltopix(bounds->x_position);
        pf_bounds.y_position = ymiltopix(bounds->y_position);
        pf_bounds.x_size = xmiltopix(bounds->x_size);
        pf_bounds.y_size = ymiltopix(bounds->y_size);
        page_wid = xmiltopix(PF.PfPageMilWidth);
        page_hyt = ymiltopix(PF.PfPageMilHeight);
        margin = xmiltopix(MARK_MARGIN);

        if ( use_bounds )
        {
            edge_x1 = pf_bounds.x_position;
            edge_y1 = pf_bounds.y_position;
            edge_x2 = edge_x1 + pf_bounds.x_size;
            edge_y2 = edge_y1 + pf_bounds.y_size;

            put(NXLT("n "));
            put(NXLT("%4d %4d M "), edge_x1, Y(edge_y1));
            put(NXLT("%4d %4d L "), edge_x2, Y(edge_y1));
            put(NXLT("%4d %4d L "), edge_x2, Y(edge_y2));
            put(NXLT("%4d %4d L "), edge_x1, Y(edge_y2));
            put(NXLT("%4d %4d L "), edge_x1, Y(edge_y1));
            put(NXLT("S\n"));
        }

        if ( use_crops )
        {
            /* top crops */
            if ( bounds->y_position > 250 )
            {
                edge_x1 = pf_bounds.x_position;
                edge_y1 = margin;
                edge_x2 = pf_bounds.x_position + pf_bounds.x_size;  
                edge_y2 = pf_bounds.y_position - margin;

                /* left */
                put(NXLT("n "));
                put(NXLT("%4d %4d M "), edge_x1, Y(edge_y1));
                put(NXLT("%4d %4d L "), edge_x1, Y(edge_y2));
                put(NXLT("S\n"));

                /* right */
                put(NXLT("n "));
                put(NXLT("%4d %4d M "), edge_x2, Y(edge_y1));
                put(NXLT("%4d %4d L "), edge_x2, Y(edge_y2));
                put(NXLT("S\n"));
            }

            /* right crops */
            if ( PF.PfPageMilWidth - (bounds->x_position+bounds->x_size) > 250 )
            {
                edge_x1 = pf_bounds.x_position + pf_bounds.x_size + margin;
                edge_y1 = pf_bounds.y_position;
                edge_x2 = page_wid - margin;
                edge_y2 = pf_bounds.y_position + pf_bounds.y_size;
 
                /* top */
                put(NXLT("n "));
                put(NXLT("%4d %4d M "), edge_x1, Y(edge_y1));
                put(NXLT("%4d %4d L "), edge_x2, Y(edge_y1));
                put(NXLT("S\n"));
 
                /* right */
                put(NXLT("n "));
                put(NXLT("%4d %4d M "), edge_x1, Y(edge_y2));
                put(NXLT("%4d %4d L "), edge_x2, Y(edge_y2));
                put(NXLT("S\n"));
            }
 
            /* bottom crops */
            if ( PF.PfPageMilHeight - (bounds->y_position+bounds->y_size) > 250 )
            {
                edge_x1 = pf_bounds.x_position;
                edge_y1 = pf_bounds.y_position + pf_bounds.y_size + margin;
                edge_x2 = pf_bounds.x_position + pf_bounds.x_size;
                edge_y2 = page_hyt - margin;
 
                /* right */
                put(NXLT("n "));
                put(NXLT("%4d %4d M "), edge_x2, Y(edge_y1));
                put(NXLT("%4d %4d L "), edge_x2, Y(edge_y2));
                put(NXLT("S\n"));
 
                /* left */
                put(NXLT("n "));
                put(NXLT("%4d %4d M "), edge_x1, Y(edge_y1));
                put(NXLT("%4d %4d L "), edge_x1, Y(edge_y2));
                put(NXLT("S\n"));
            }
 
            /* left crops */
            if ( bounds->x_position > 250 )
            {
                edge_x1 = margin;
                edge_y1 = pf_bounds.y_position;
                edge_x2 = pf_bounds.x_position - margin;
                edge_y2 = pf_bounds.y_position + pf_bounds.y_size;
 
                /* bottom */
                put(NXLT("n "));
                put(NXLT("%4d %4d M "), edge_x1, Y(edge_y2));
                put(NXLT("%4d %4d L "), edge_x2, Y(edge_y2));
                put(NXLT("S\n"));
 
                /* top */
                put(NXLT("n "));
                put(NXLT("%4d %4d M "), edge_x1, Y(edge_y1));
                put(NXLT("%4d %4d L "), edge_x2, Y(edge_y1));
                put(NXLT("S\n"));
            }
        }

        if ( use_regis )
        {
            /* left mark */
            if ( bounds->x_position > 250 )
            {
                edge_x1 = pf_bounds.x_position / 2;
                edge_y1 = page_hyt / 2;

                /* draw the registration circle */
                put(NXLT("n "));
                put(NXLT("%4d %4d M "), edge_x1+margin, edge_y1);
                put(NXLT("%4d %4d %d 0 360 arc "), edge_x1, edge_y1, margin);
                put(NXLT("S\n"));
                put(NXLT("n "));
                put(NXLT("%4d %4d M "), edge_x1 - (margin+10), edge_y1);
                put(NXLT("%4d %4d L S\n"), edge_x1 + (margin+10), edge_y1);
                put(NXLT("n "));
                put(NXLT("%4d %4d M "), edge_x1, edge_y1 - (margin+10));
                put(NXLT("%4d %4d L S\n"), edge_x1, edge_y1 + (margin+10));
                put(NXLT("n "));
                put(NXLT("%4d %4d M "), (int)(edge_x1+margin/2.0), edge_y1);
                put(NXLT("%4d %4d %d 0 360 arc "), edge_x1, edge_y1, (int)(margin/2.0));
                put(NXLT("fill\n"));
                put(NXLT("1 GRAY\n"));
                put(NXLT("n "));
                put(NXLT("%4d %4d M "), (int)(edge_x1 - margin/2.0), edge_y1);
                put(NXLT("%4d %4d L S\n"), (int)(edge_x1 + margin/2.0), edge_y1);
                put(NXLT("n "));
                put(NXLT("%4d %4d M "), edge_x1, (int)(edge_y1 - margin/2.0));
                put(NXLT("%4d %4d L S\n"), edge_x1, (int)(edge_y1 + margin/2.0));
            }

            /* right mark */
            if ( PF.PfPageMilWidth - (bounds->x_position+bounds->x_size) > 250 )
            {
                edge_x2 = (page_wid - (pf_bounds.x_position+pf_bounds.x_size))/2.0;
                edge_x1 = pf_bounds.x_position + pf_bounds.x_size + edge_x2;
                edge_y1 = page_hyt / 2;

                /* draw the registration circle */
                put(NXLT("0 GRAY\n"));
                put(NXLT("n "));
                put(NXLT("%4d %4d M "), edge_x1+margin, edge_y1);
                put(NXLT("%4d %4d %d 0 360 arc "), edge_x1, edge_y1, margin);
                put(NXLT("S\n"));
                put(NXLT("n "));
                put(NXLT("%4d %4d M "), edge_x1 - (margin+10), edge_y1);
                put(NXLT("%4d %4d L S\n"), edge_x1 + (margin+10), edge_y1);
                put(NXLT("n "));
                put(NXLT("%4d %4d M "), edge_x1, edge_y1 - (margin+10));
                put(NXLT("%4d %4d L S\n"), edge_x1, edge_y1 + (margin+10));
                put(NXLT("n "));
                put(NXLT("%4d %4d M "), (int)(edge_x1 + margin/2.0), edge_y1);
                put(NXLT("%4d %4d %d 0 360 arc "), edge_x1, edge_y1, (int)(margin/2.0));
                put(NXLT("fill\n"));
                put(NXLT("1 GRAY\n"));
                put(NXLT("n "));
                put(NXLT("%4d %4d M "), (int)(edge_x1 - margin/2.0), edge_y1);
                put(NXLT("%4d %4d L S\n"), (int)(edge_x1 + margin/2.0), edge_y1);
                put(NXLT("n "));
                put(NXLT("%4d %4d M "), edge_x1, (int)(edge_y1 - margin/2.0));
                put(NXLT("%4d %4d L S\n"), edge_x1, (int)(edge_y1 + margin/2.0));

                /* draw the color boxes */
                margin = margin * 2;
                edge_x1 -= margin/2;
                edge_y1 += margin * 2;
                for ( i=0; i<10; i++ )
                {
                    put(NXLT("n "));
                    put(NXLT("%4d %4d M "), edge_x1, edge_y1);
                    put(NXLT("%d %d RL "), margin, 0);
                    put(NXLT("%d %d RL "), 0, margin);
                    put(NXLT("%d %d RL "), (-1)*margin, 0);
                    put(NXLT("%d %d RL "), 0, (-1)*margin);
                    put(NXLT("GS %d.%d GRAY fill GR 0 GRAY S\n"), i==9?1:0, i);
                    edge_y1 += margin;
                }

                /* draw the registration string */
				put(NXLT("0 GRAY\n"));
                edge_x2 = PF.PfPageMilWidth - (bounds->x_position+bounds->x_size);
                edge_x1 = bounds->x_position+bounds->x_size + edge_x2/2.0 - 120;
                edge_y1 = PF.PfPageMilHeight / 2.0 + 3*MARK_MARGIN;
          		pf_display_reg_string(NXLT("Helvetica"), edge_x1, edge_y1);

            }
        }

        put(NXLT("GR\n"));
    }
    return(0);
}



/*-----------------------------------------------------------------------------
    D01milSetPos
-----------------------------------------------------------------------------*/
int D01milSetPos(new_x_mil, new_y_mil)
    int new_x_mil, new_y_mil;
{
    int new_x, new_y;
    double new_x_f, new_y_f;


    new_x_f = xmiltopix(new_x_mil);
    new_y_f = ymiltopix(new_y_mil);
    new_x = (int)new_x_f;
    new_y = (int)new_y_f;

    PF.PfHardX = new_x;
    PF.PfHardY = new_y;

    put(NXLT("%.2f %.2f M\n"), new_x_f, Y_mil(new_y_f));    /* Y should still work... */

    return(0);
}

/*-----------------------------------------------------------------------------
    D01SetPos
-----------------------------------------------------------------------------*/
int D01SetPos(new_x, new_y)
    int new_x, new_y;
{
    PF.PfHardX = new_x;
    PF.PfHardY = new_y;

    put(NXLT("%d %d M\n"), new_x, Y(new_y));

    return(0);
}


/*-----------------------------------------------------------------------------
    D01AxText
-----------------------------------------------------------------------------*/
int D01AxText(textstr, txlen, pos, width, fontName,
          underPos, underThick, strikePos, txProps,
          fieldProps, xZoom, yZoom, clip)
    tx_char *textstr;
    int txlen;
    p32 *pos;               /* mils */
    int width;              /* mils */
    char *fontName;
    int underPos;           /* in milliems */
    int underThick;         /* in milliems */
    int strikePos;          /* in mils */
    textAttrs *txProps;     /* properties of this segment of text */
    textAttrs *fieldProps;  /* props of the overall field of text */
    int xZoom, yZoom;       /* zoom factor relative to ref val 100 */
    TmRect32 *clip;         /* mils */
{
    int tx, ty, unity;
    int pointsize ;
    textAttrs txprops, fldprops, *txp, *fld;
    axform xform, rulxform;
    char *errString;
    fontTabEnt *ftab, *altRec;
    int xformed = FALSE;


    if (txlen <= 0 || width <= 0)
        return FALSE;

    if (pfIsTransparent(txProps->fillpattern))
		return TRUE ;

    tx = xmiltopix(pos->x);
    ty = ymiltopix(pos->y);

    D01SetCMYKColor(axCyanComponent(txProps->fillcolor),
                    axMagentaComponent(txProps->fillcolor),
                    axYellowComponent(txProps->fillcolor),
                    axBlackComponent(txProps->fillcolor), 255);

    /* look up the font record; select the font */
    if (errString = axGetFontRecord(fontName, &ftab))
    {
        freeErrString(errString);
        return FALSE;
    }

    pointsize = txProps->pointsize ;
    if (ftab->gallium)
    {
        pointsize = (int)((double)pointsize/10000.0 * fabs((double)fieldProps->yscale)) ;
        width = (int)((double)width/10000.0 * fabs((double)fieldProps->yscale)) ;
    }


    if (!ftab->PostScriptAlternate)
    {
        /* the record is considered "usable" to PostScript */
        altRec = (fontTabEnt *) NULL;  /* no need to use alternate font */
        D01LoadNamedFont(ftab->globalName, pointsize);
    }
    else
    {
        /* the record is considered "unusable" to postscript;
         * use the alternate's record instead */
        if (errString = axGetFontRecord(ftab->PostScriptAlternate,
                                        &altRec))
        {
            freeErrString(errString);
            return FALSE;
        }
        D01LoadNamedFont(altRec->globalName, pointsize);
    }

    /* the transform terms arrive in a form relative to our internal,
       y-increasing-downwards coord sys.  Postscript's sys is normal,
       so account for the differences on the transform terms before
       converting to ctm form.  (that is, negate sines and shears) */
    if (txProps)
    {
        txprops = *txProps;
        txprops.pointsize = pointsize ;
        txprops.angle = -txprops.angle;
        txprops.xshear = -txprops.xshear;
        txp = &txprops;
    }
    else
    {
        txp = (textAttrs *) NULL;
    }

    if (fieldProps)
    {
        fldprops = *fieldProps;
        fldprops.angle = -fldprops.angle;
        fldprops.xshear = -fldprops.xshear;
        fld = &fldprops;
        if (ftab->gallium)
        {
            fldprops.xscale = SIGN(fldprops.xscale) *
                (int)fabs(((double)fldprops.xscale * 10000.0) / (double)fldprops.yscale) ;
            fldprops.yscale = SIGN(fldprops.yscale) * 10000 ;
        }
    }
    else
    {
        fld = (textAttrs *) NULL;
    }
    
    /* should test for zero scale factors..
     * (to guard against a concat of  [0 0 0 0 0 0]) ....
     */
    if (txp != NULL)
    {
        if (txp->xscale == 0)
            txp->xscale = 10000;
        if (txp->yscale == 0)
            txp->yscale = 10000;
    }
    if (fld != NULL)
    {
        if (fld->xscale == 0)
            fld->xscale = 10000;
        if (fld->yscale == 0)
            fld->yscale = 10000;
    }

    unity = pfGetTextSegTransform(txp, fld, xZoom, yZoom, &xform);
    if (txProps->ruling || txProps->strike_thru)
    {
            /* get an unmirrored xform for use in drawing underlines;
               underlines won't be drawn under a modified ctm */
        pfGetTextSegTransform(txProps, fieldProps, xZoom, yZoom,
                              &rulxform);
    }

    PS_GS() ;

    D01DrawClip(clip) ;

    put(NXLT("n\n"));

    D01SetPos(tx, ty);

    if (!unity)
    {
        put(NXLT ("[%.4f %.4f %.4f %.4f %d %d] concat\n"),
            xform.forward.ca,  xform.forward.cb,
            xform.forward.cc,  xform.forward.cd,
            xform.forward.ctx, xform.forward.cty);
    }

    /* set the text */
    if (pfEasyTextFill(txProps->fillpattern))
        D01PutPlainTextString(
                        textstr, txlen, width, pos->x, pos->y,
                        fontName, txp, fld, &xform,
                        ftab, altRec);
    else
        D01PutFilledTextString(
                        textstr, txlen, width, pos->x, pos->y,
                        fontName, txp, fld, &xform,
                        ftab, altRec);

    PS_GR();

    return(TRUE);
}


/*-----------------------------------------------------------------------------
    D01PutPlainTextString

    little fill sophistication, but will color the text.
-----------------------------------------------------------------------------*/
static void D01PutPlainTextString(textstr, txlen, width, xpos, ypos,
                      fontName, txProps, fieldProps, xform,
                      ftab, altRec)

    tx_char *textstr;
    int txlen;
    int width;  /* mils */
    int xpos, ypos; /* mils */
    char *fontName;
    textAttrs *txProps;
    textAttrs *fieldProps;
    axform *xform;
    fontTabEnt *ftab;   /* font record for app-specified attrs */
    fontTabEnt *altRec; /* if nonnull, use this alternate record */
{
    tx_char *tp;
    int tmpx, ypix;
    register int ix;
    register int seglen;
    register int milx;
    char *ttext = NULL;
    attr_type *attr;
    sub_attr_type sattr;

    attr = txProps->fillpattern;

    if ( !pfEasyTextFill(attr) && (attr->f_type == GRAY_FILL_TYPE))
    {
        /* don't do anything if bg color is transparent */
        if ( attr->ffg.see_thru == FALSE )
        {
            if (attr->use_ffg == FALSE)
            {
                int c_diff, m_diff, y_diff, k_diff, gray_factor;
                /* I suspect that this only works with (both fg
                and bg) undercolor removed */

                sattr.use_f_black = attr->use_lfg;
                sattr.use_b_white = attr->use_lbg;
                if (sattr.use_f_black)
                {
                    sattr.bg.c = 0;
                    sattr.bg.m = 0;
                    sattr.bg.y = 0;
                    sattr.bg.k = 255;
                }
                else
                {
                    sattr.bg.c = attr->fbg.c;
                    sattr.bg.m = attr->fbg.m;
                    sattr.bg.y = attr->fbg.y;
                    sattr.bg.k = attr->fbg.k;
                }
                if (sattr.use_b_white)
                {
                    sattr.fg.c = 0;
                    sattr.fg.m = 0;
                    sattr.fg.y = 0;
                    sattr.fg.k = 0;
                }
                else
                {
                    sattr.fg.c = attr->ffg.c;
                    sattr.fg.m = attr->ffg.m;
                    sattr.fg.y = attr->ffg.y;
                    sattr.fg.k = attr->ffg.k;
                }
                
                c_diff = sattr.bg.c - sattr.fg.c;
                m_diff = sattr.bg.m - sattr.fg.m;
                y_diff = sattr.bg.y - sattr.fg.y;
                k_diff = sattr.bg.k - sattr.fg.k;
                
                gray_factor = 1000 - attr->l_pat;
                D01SetCMYKColor(sattr.fg.c + (c_diff * gray_factor)/1000,
                                sattr.fg.m + (m_diff * gray_factor)/1000,
                                sattr.fg.y + (y_diff * gray_factor)/1000,
                                sattr.fg.k + (k_diff * gray_factor)/1000, 255);
            }
            else
            {
                put(NXLT("0 GRAY\n"));  /* black ink */
            }
        }
    }
    if (altRec == (fontTabEnt *) NULL)
    {
        /* we assume that the font is available on the
        output device (whether or not true); output directly */
        ttext = PSescapeText(textstr, txlen);   /* convert 8bit, etc. */
        put_X_text(ttext, xmiltopix(width));

        if (ttext)
            free(ttext);
        return;
    }
    else
    {
        /* this is the print-bash code; set the text in
        a substitue font, adjusting layout to reflect
        the ideal font's metrics */
        axtext alt;   /* represents passed info, but in substitue font */
        int *deltas ;

        /* get the deltas array describing layout adjustments */
        memset((char *) &alt, 0, sizeof(axtext));
        deltas = pfGetTextLayoutDeltas(ftab, textstr, txlen, width,
                                       txProps, altRec, &alt);
        if (!deltas)
        {
            daxtext(&alt);
            return;
        }

        /* use the deltas array to set the text */
        tp = textstr;
        milx = xpos;
        seglen = 0;
        ypix = ymiltopix(ypos);
        for (ix = 0; ix < txlen; ix++)
        {
            seglen++;
            if (deltas[ix])
            {
                /* set text segment, adjust position */
                ttext = PSescapeText(tp, seglen) ;
                put_ST_text(ttext);
                free(ttext);
                tp += seglen;
                seglen = 0;

                milx += alt.items[ix].padvance;
                milx += deltas[ix];
                tmpx = xmiltopix(deltas[ix]);
                put(NXLT("%d 0 MR\n"), tmpx);
            }
            else
            {
                milx += alt.items[ix].padvance;
            }
        }
        if (seglen > 0)
        {
            /* set the caboose */
            ttext = PSescapeText(tp, seglen) ;
            put_ST_text(ttext);
            free(ttext);
        }
        if (deltas)
            free(deltas);

        daxtext(&alt);
    }
}

/*-----------------------------------------------------------------------------
    D01PutFilledTextString
-----------------------------------------------------------------------------*/
static void D01PutFilledTextString(textstr, txlen, width, xpos, ypos,
                       fontName, txProps, fieldProps, xform,
                       ftab, altRec)
    tx_char *textstr;
    int txlen;
    int width;  /* mils */
    int xpos, ypos; /* mils */
    char *fontName;
    textAttrs *txProps;
    textAttrs *fieldProps;
    axform *xform;
    fontTabEnt *ftab;   /* font record for app-specified attrs */
    fontTabEnt *altRec; /* if nonnull, use this alternate record */
{
    tx_char *tp;
    int tmpx, ypix;
    register int ix;
    register int seglen;
    register int milx, lastx;
    char *ttext = NULL;
    attr_type *attr;
    TmPoint32 boundingBox[4];
	int c_diff, m_diff, y_diff, k_diff, gray_factor;

    int doFill = TRUE;

    /* "local" fill & stroke attributes */
    attr = txProps->fillpattern;

	/* we can't gradient fill download gallium fonts, so change them to solid */
	if (ftab->gallium)
	{
		switch (attr->f_type)
		{
	    case LINEAR_FILL_TYPE:
	    case BILINEAR_FILL_TYPE:
	    case RADIAL_FILL_TYPE:
	    case RECT_FILL_TYPE:
	    case BUILTIN_FILL_TYPE:
	    case USERDEF_FILL_TYPE:
			attr->f_type = GRAY_FILL_TYPE ;
			attr->f_pat = 500;
			break ;
		}
	}

    /* test whether fill can be avoided */
    if (attr->f_type == GRAY_FILL_TYPE)  /* easy */
    {
        /* don't do anything if fg color is transparent */
        if ( attr->ffg.see_thru == FALSE )
        {
            if (attr->use_ffg == FALSE)
            {
                /* I suspect that this only works with (both fg
                and bg) undercolor removed */
                c_diff = attr->fbg.c - attr->ffg.c;
                m_diff = attr->fbg.m - attr->ffg.m;
                y_diff = attr->fbg.y - attr->ffg.y;
                k_diff = attr->fbg.k - attr->ffg.k;
                gray_factor = 1000 - attr->f_pat;
                D01SetCMYKColor(attr->ffg.c + (c_diff * gray_factor)/1000,
                                attr->ffg.m + (m_diff * gray_factor)/1000,
                                attr->ffg.y + (y_diff * gray_factor)/1000,
                                attr->ffg.k + (k_diff * gray_factor)/1000,
                                255);
            }
            else
            {
                put(NXLT("0 GRAY\n"));  /* black ink */
            }
        }
        doFill = FALSE;
    }
    else
    {
        if (attr->f_type == BUILTIN_FILL_TYPE)
        {
            /* Yes, this has happened */
            if ((unsigned int)(attr->f_pat) > MAX_PATS) /* dirty hack */
            {
                attr->f_pat = 4;    /* solid 100%... */
                doFill = FALSE;
            }
            if (attr->f_pat == 4)   /* also easy */
            {
                if (attr->use_ffg == FALSE)
                {
                    D01SetCMYKColor(attr->ffg.c,
                                    attr->ffg.m,
                                    attr->ffg.y,
                                    attr->ffg.k, 255);
                }
                else
                {
                    put(NXLT("0 GRAY\n"));  /* black ink */
                }
                doFill = FALSE;
            }
        }
    }

    /* set the text */
    if (altRec == (fontTabEnt *) NULL)
    {
        /* the font is .afm-based; use it directly */
        ttext = PSescapeText(textstr, txlen);   /* convert 8bit, etc. */

        if (doFill)
        {
            /* This is bogus! charpath lays out the text as if it were
               shown with "show"... per character basis?... */
            put_Xc_text(ttext);

            TextFillBounder(ttext, txlen, txProps, fieldProps, xform,
                                altRec, xpos, ypos, boundingBox);
            D01FillPattern(attr, boundingBox);
        }
        else
        {
            put_X_text(ttext, xmiltopix(width));
        }

        if (ttext)
            free(ttext);

        return;
    }
    else
    {
        /* this is the print-bash code; set the text in
            a substitue font, adjusting layout to reflect
            the ideal font's metrics */
        axtext alt;   /* represents passed info, but in substitue font */
        int *deltas = (int *) NULL;

            /* get the deltas array describing layout adjustments */
        memset((char *) &alt, 0, sizeof(axtext));
        deltas = pfGetTextLayoutDeltas(ftab, textstr, txlen, width,
                                       txProps, altRec, &alt);
        if (!deltas)
        {
            daxtext(&alt);
            return;
        }
            /* use the deltas array to set the text */
        tp = textstr;
        milx = lastx = xpos;
        seglen = 0;
        ypix = ymiltopix(ypos);
        for (ix = 0; ix < txlen; ix++)
        {
            seglen++;
            if (deltas[ix])
            {
                    /* set text segment, adjust position */
                if(ttext = PSescapeText(tp, seglen))
                {
                    if (doFill)
                        put_Xc_text(ttext);
                    else
                        put_ST_text(ttext);
                    free(ttext);
                    ttext = NULL;
                }
                tp += seglen;
                seglen = 0;

                milx += alt.items[ix].padvance;
                milx += deltas[ix];
                lastx = milx;
                tmpx = xmiltopix(deltas[ix]);
                put(NXLT("%d 0 MR\n"), tmpx);
            }
            else
            {
                milx += alt.items[ix].padvance;
            }
        }
        if (seglen > 0)
        {
            /* set the caboose */
            if(ttext = PSescapeText(tp, seglen))
            {
                if (doFill)
                    put_Xc_text(ttext);
                else
                    put_ST_text(ttext);
                free(ttext);
                ttext = NULL;
            }
        }

        if(doFill && (ttext = PSescapeText(textstr, txlen)))
        {
            TextFillBounder(ttext, txlen, txProps,
                                fieldProps, xform, (fontTabEnt *) NULL,
                                xpos, ypos, boundingBox);
                /* the bounder just produced isn't correct in
                   general; it's off by whatever differences
                   there are between the alternate and ideal
                   letterforms; only for the setpoints of the
                   first and last letterforms will there be
                   any agreement between alternate and ideal;
                   this can be handled by measuring each
                   letterform in the alternate font, and
                   taking the union of the lot, but just fudge
                   liberally.. */

            boundingBox[0].x -= 300;
            boundingBox[0].y -= 300;
            boundingBox[1].x += 300;
            boundingBox[1].y += 300;   /* sigh */

            D01FillPattern(attr, boundingBox);
            free(ttext);
            ttext = NULL;
        }

        if (deltas)
        {
            free(deltas);
            deltas = NULL;
        }
        daxtext(&alt);
    }
}


/*-----------------------------------------------------------------------------
    TextFillBounder

    a function for obtaining a bounding
    box for use in filling a path generated by charpath;     the text
    is assumed to be untransformed at this point (a hack, but currently
    don't have support for doing our own charpath.. so, untransformed
    text is filled, then projected thru a transform; this process will
    transform the fill.....)

-----------------------------------------------------------------------------*/
static void TextFillBounder(textstr, txlen, txProps, fieldProps,
                    xform, altRec, xpos, ypos, bbox)
    char *textstr;
    int txlen;
    textAttrs *txProps;
    textAttrs *fieldProps;
    axform *xform;
    fontTabEnt *altRec;
    int xpos, ypos;
    TmPoint32 *bbox;       /* out: report bounder in printer pix */
{
    textAttrs txatt, *tattr;
    tx_char *insText = (tx_char *) NULL;
    int ix, tmp;
    axtext txt;
    char *err;
    r32 tbb;

    txProps->angle = -txProps->angle ;
    txProps->xshear = -txProps->xshear ;
    fieldProps->angle = -fieldProps->angle ;
    fieldProps->xshear = -fieldProps->xshear ;

    txatt.fontFamily = (char *) NULL;
    txatt.fillpattern = txatt.strokepattern = (attr_type *) NULL;

    if (altRec && altRec->familyName)
    {
            /* copy txProps, and switch the fontFamily name */
        txatt = *txProps;
        txatt.fontFamily = (char *) malloc
            (strlen(txProps->fontFamily) + 1);
        strcpy(txatt.fontFamily, altRec->familyName);

        if (txProps->fillpattern)
        {
            txatt.fillpattern = (attr_type *) malloc(sizeof(attr_type));
            memset((char *) txatt.fillpattern, 0, sizeof(attr_type));
            pfCopyFillPattern(txProps->fillpattern,
                              &txatt.fillpattern);
        }
        if (txProps->strokepattern)
        {
            txatt.strokepattern = (attr_type *) malloc(sizeof(attr_type));
            memset((char *) txatt.strokepattern, 0, sizeof(attr_type));
            pfCopyFillPattern(txProps->strokepattern,
                              &txatt.strokepattern);
        }
        tattr = &txatt;
    }
    else
    {
        tattr = txProps;
    }

        /* create a text instance and take its bounder */
    memset((char *) &txt, 0, sizeof(axtext));
    setDefaultTextAttrs(&txt, tattr, TXPrintAtts);
    insText = (tx_char *) malloc(txlen * sizeof(tx_char));
    for (ix = 0; ix < txlen; ix++)
    {
        insText[ix] = (tx_char) textstr[ix];
    }
    insChars(&txt, insText, 0, txlen);
    axtSetFieldTransform(&txt, fieldProps, TXPrintAtts);
    if(err = getOverallBounder(&txt, 0, -1, &tbb))
    {
        freeErrString(err);
        tbb.origin.x = tbb.origin.y = tbb.width = tbb.height = 0;
    }


    /* add position to the bounder, and convert */
    bbox[0].x = xpos + tbb.origin.x;
    bbox[1].x = xpos + tbb.origin.x + tbb.width - 1;
    bbox[0].y = ypos + tbb.origin.y;
    bbox[1].y = ypos + tbb.origin.y + tbb.height - 1;

    bbox[0].x = xmiltopix(bbox[0].x) - 30;
    bbox[0].y = ymiltopix(bbox[0].y) - 30;
    bbox[1].x = xmiltopix(bbox[1].x) + 30;
    bbox[1].y = ymiltopix(bbox[1].y) + 30;

    tmp = bbox[0].y;
    bbox[0].y = Y(bbox[1].y);
    bbox[1].y = Y(tmp);

    if (!xform->unity)
    {
    /* undo the forward transform under which
       charpath was executed; that path will remain
       for filling under the previous ctm (i.e., its
       approximation produced here) */
        put(NXLT ("[%.4f %.4f %.4f %.4f %d %d] concat\n"),
            xform->reverse.ca,  xform->reverse.cb,
            xform->reverse.cc,  xform->reverse.cd,
            xform->reverse.ctx, xform->reverse.cty);
    }

    if (txatt.fontFamily)
        free(txatt.fontFamily);
    if (txatt.fillpattern)
        pfFreeFillPattern(txatt.fillpattern);
    if (txatt.strokepattern)
        pfFreeFillPattern(txatt.strokepattern);
    if (insText)
        free(insText);

    /* don't know how to calculate rotated bounding box, yet */
    bbox[2] = bbox[0] ;
    bbox[3].x = bbox[1].x - bbox[0].x ;
    bbox[3].y = bbox[1].y - bbox[0].y ;
}

/*************************************************************************
 *
 * Used by ClrBox, ClrImageBox
 * 15-Dec-93  steve  Funtionality change: we now expect the color to have been
 *               set before this call...
 * 
 * 
************************************************************************/
int D01ClrFill()
{
    int fg_color, bg_color, temp_color;

    fg_color = PF.PfWin_fg_color;
    bg_color = PF.PfWin_bg_color;
    
    if (PF.PfPage.trans_mode==TMIS ||
        PF.PfPage.trans_mode==TMIS_O_D ||
        PF.PfPage.trans_mode==TMIS_A_D)
    {
        temp_color = fg_color;
        fg_color = bg_color;
        bg_color = temp_color;
    }

    put(NXLT("fill\n"));

    return(0);
}

/*************************************************************************
 *
 * 
 *  parameters are in "User Space pixels"
 * 
************************************************************************/
int D01ClrBox (x, y, w, h)
    int x, y, w, h;
{

    put(NXLT("GS\n"));
    put(NXLT("n\n"));
    put(NXLT("%d %d M\n"), x, Y(y));
    put(NXLT("0 %d RL "), -h);
    put(NXLT("%d 0 RL "), w);
    put(NXLT("0 %d RL\n"), h);
    put(NXLT("closepath\n"));
    D01ClrFill();
    put(NXLT("GR\n"));

    return(0);
}

/*************************************************************************
 * 
 * A variation of D01ClrBox for use when the PfPage.imageBox
 * is defined.
 *
 * 
************************************************************************/
int D01ClrImageBox(box, xpos, ypos, ascent, uwidth, uheight)
    TmPoint32 *box; /* there's four points */
    int xpos, ypos; /* text object's app-specified position */
    int ascent;  /* text object's untranformed ascent */
    int uwidth, uheight; /* object's untransformed bbox dimensions */
{
    TmPoint32 tpt[4];

    tpt[0].x = xpos + box[0].x;
    tpt[0].y = ypos - ascent + box[0].y;
    tpt[1].x = xpos + box[1].x;
    tpt[1].y = ypos - ascent + box[1].y;
    tpt[2].x = xpos + box[2].x;
    tpt[2].y = ypos - ascent + box[2].y;
    tpt[3].x = xpos + box[3].x;
    tpt[3].y = ypos - ascent + box[3].y;
    /* now do the clear */
    put(NXLT("n\n"));
    put(NXLT("%d %d M\n"), tpt[0].x, Y(tpt[0].y));
    put(NXLT("%d %d L ") , tpt[1].x, Y(tpt[1].y));
    put(NXLT("%d %d L "),  tpt[2].x, Y(tpt[2].y));
    put(NXLT("%d %d L\n"), tpt[3].x, Y(tpt[3].y));
    put(NXLT("closepath\n"));
    D01ClrFill();

    return(0);
}

/*-----------------------------------------------------------------------------
    D01Line
-----------------------------------------------------------------------------*/
int D01Line(nPolys, attr, pol, clip)
    int nPolys;
    attr_type *attr;
    poly_type *pol;
    TmRect32 *clip ;
{
    int i, j, nPoints;
    TmPoint32 boundingBox[4];
    int x_val, y_val;

    /* start */
    PS_GS();
    D01DrawClip(clip) ;

    D01GetBoundingBox(nPolys, attr, pol, boundingBox);

    /* write out the path */
    put(NXLT("n\n"));
    for ( i=0; i<nPolys; i++ )
    {
        /* move to the start of the poly */
        put(NXLT("%d %d M\n"), pol[i].points[0].x, Y(pol[i].points[0].y));

        if ( pol[i].nCtrls == 0 ) /* draw a simple poly */
        {
            x_val = pol[i].points[0].x;
            y_val = pol[i].points[0].y;

            for ( j=1; (j<pol[i].nPoints) && (j<max_path_points); j++ ) 
            {
                    /* output the point */
#if 0
                pol[i].points[j].x = (pol[i].points[j].x < 0) ? 0 : pol[i].points[j].x;
                pol[i].points[j].y = (pol[i].points[j].y < 0) ? 0 : pol[i].points[j].y;
#endif

#if 0
                if ((ABS(x_val - pol[i].points[j].x) > 1) ||
                    (ABS(y_val - pol[i].points[j].y) > 1))
                {
                    x_val = pol[i].points[j].x;
                    y_val = pol[i].points[j].y;
                    
                    put(NXLT("%d %d L "), pol[i].points[j].x, Y(pol[i].points[j].y));
                }
#else
                put(NXLT("%d %d L "), pol[i].points[j].x, Y(pol[i].points[j].y));
#endif
                /* add newlines to keep the output neatly formatted */
                if ( j%5 == 0 )
                    put(NXLT("\n"));
            }

            if ( attr->l_style == TM_SOLID &&
                 pol[i].nPoints > 2 &&
                 pol[i].points[0].x == pol[i].points[pol[i].nPoints-1].x &&
                 pol[i].points[0].y == pol[i].points[pol[i].nPoints-1].y )
            {
                /* wrap around corner to use adobe's corner filling */
                put(NXLT("%d %d L\n"), pol[i].points[1].x, Y(pol[i].points[1].y));
            }
            else if ( (j-1)%5 != 0 )
                put(NXLT("\n")); /* formatting */
        }
        else /* draw a curve */
        {
            /* link individual beziers into a single curve */
            nPoints = 2 + (pol[i].nPoints - 2) / 2;

            /* connect other points to path */
            for ( j=1; j<nPoints; j++ ) 
            {
                /* output the points */
                put(NXLT("%d %d %d %d %d %d C\n"),
                     pol[i].controls[(2*j)-2].x, Y(pol[i].controls[(2*j)-2].y),
                     pol[i].controls[(2*j)-1].x, Y(pol[i].controls[(2*j)-1].y),
                     pol[i].points[j*2-1].x, Y(pol[i].points[j*2-1].y));
            }
        }
    }

    /* fill the path according to pattern info passed in */
    if ( attr->fill_flag )
        D01FillPattern(attr, boundingBox);

    /* stroke the path according to line info passed in */
    if ( attr->l_wid )
        D01FillStroke(attr, boundingBox);

    /* finish */
    /* We are before a GR, therefore we must trash the current width
       and line style values... Is there no way we can set these
       outside of our GS/GR so they persist for others?  Or we could
       keep track of whether we have changed cur_thick/cur_style
       WITHIN THIS GS/GR BRACKET.  This is just like keeping a stack
       of these values for the current GS/GR level... Hm...
       */
    PF.cur_thick = MINIMUM_LINE_WIDTH;
    PF.cur_style = TM_SOLID;

    PS_GR();

    return(0);
}


#define TX(x, y)    ((x-orgx)*cosa - (y-orgy)*sina)
#define TY(x, y)    ((x-orgx)*sina + (y-orgy)*cosa)

/*************************************************************************

 *  accumulate an enclosing bounder for a poly_type instance into box
 * 

        box[0] = LLC of rect
        box[1] = URC of rect
        box[2] = LLC of rotated rect
        box[3] = width, height pf rotated rect
************************************************************************/
static void D01GetBoundingBox(nPolys, attr, pol, box)
    int nPolys;
    attr_type *attr;
    poly_type *pol;
    TmPoint32 *box;
{
    register int ix, jx, half_line_width;
    double minx, miny, maxx, maxy, xr, yr ;
    double angle, sina, cosa ;
    double orgx, orgy ;
        

    /* init boundary */
    box[0].x = box[1].x = (int) pol[0].points[0].x;
    box[0].y = box[1].y = (int) (Y(pol[0].points[0].y));

    /* scan poly info */
    for (ix = 0; ix < nPolys; ix++)
    {
        /* scan points */
        for (jx = 0; jx < pol[ix].nPoints; jx++)
        {
            if (pol[ix].points[jx].x < box[0].x)
                box[0].x = pol[ix].points[jx].x;
            else if (pol[ix].points[jx].x > box[1].x)
                box[1].x = pol[ix].points[jx].x;
            if (Y(pol[ix].points[jx].y) < box[0].y)
                box[0].y = Y(pol[ix].points[jx].y);
            else if (Y(pol[ix].points[jx].y) > box[1].y)
                box[1].y = Y(pol[ix].points[jx].y);
        }

        /* this is the case for round ends, need to do square, butt */
        half_line_width = attr->l_wid/2;
        box[0].x -= half_line_width;
        box[0].y -= half_line_width;
        box[1].x += half_line_width;
        box[1].y += half_line_width;

        /* scan control points also */
        for (jx = 0; jx < pol[ix].nCtrls; jx++)
        {
            if (pol[ix].controls[jx].x < box[0].x)
                box[0].x = pol[ix].controls[jx].x;
            else if (pol[ix].controls[jx].x > box[1].x)
                box[1].x = pol[ix].controls[jx].x;
            if (Y(pol[ix].controls[jx].y) < box[0].y)
                box[0].y = Y(pol[ix].controls[jx].y);
            else if (Y(pol[ix].controls[jx].y) > box[1].y)
                box[1].y = Y(pol[ix].controls[jx].y);
        }
    }
    /* now compute the rotated box */

    if (attr->f_ang == 0)
    {
        box[2] = box[0] ;
        box[3].x = ABS(box[1].x - box[0].x) ;
        box[3].y = ABS(box[1].y - box[0].y) ;
        return ;
    }      

    angle = ((double)(-attr->f_ang) * M_PI) / 180.0 ;
    sina = sin(angle) ;
    cosa = cos(angle) ;
    orgx = box[0].x ;
    orgy = box[0].y ;

    /* init boundary */
    minx = maxx = TX(pol[0].points[0].x, Y(pol[0].points[0].y)) ;
    miny = maxy = TY(pol[0].points[0].x, Y(pol[0].points[0].y)) ;

    /* scan poly info */
    for (ix = 0; ix < nPolys; ix++)
    {
        /* scan points */
        for (jx = 0; jx < pol[ix].nPoints; jx++)
        {
            xr = TX(pol[ix].points[jx].x, Y(pol[ix].points[jx].y)) ;
            yr = TY(pol[ix].points[jx].x, Y(pol[ix].points[jx].y)) ;
            if (xr < minx)
                minx = xr ;
            else if (xr > maxx)
                maxx = xr ;
            if (yr < miny)
                miny = yr ;
            else if (yr > maxy)
                maxy = yr ;

        }

        /* this is the case for round ends, need to do square, butt */
        half_line_width = attr->l_wid/2;
        minx -= half_line_width;
        miny -= half_line_width;
        maxx += half_line_width;
        maxy += half_line_width;

        /* scan control points also */
        for (jx = 0; jx < pol[ix].nCtrls; jx++)
        {
            xr = TX(pol[ix].controls[jx].x, Y(pol[ix].controls[jx].y)) ;
            yr = TY(pol[ix].controls[jx].x, Y(pol[ix].controls[jx].y)) ;
            if (xr < minx)
                minx = xr ;
            else if (xr > maxx)
                maxx = xr ;
            if (yr < miny)
                miny = yr ;
            else if (yr > maxy)
                maxy = yr ;
        }
    }

    /* now rotate back */
    angle = ((double)(attr->f_ang) * M_PI) / 180.0 ;
    sina = sin(angle) ;
    cosa = cos(angle) ;
    orgx = orgy = 0 ;

    box[2].x = (int)TX(minx, miny) + box[0].x; 
    box[2].y = (int)TY(minx, miny) + box[0].y ;
    box[3].x = maxx - minx ;
    box[3].y = maxy - miny ;

}


/*************************************************************************
 *
 * 
 * 
 * 
************************************************************************/
static xMirror(pt, outpt, center)
    dpoint *pt, *outpt;
    double center;
{
    outpt->x = center + (center - pt->x);
    outpt->y = pt->y;
}

/*************************************************************************
 *
 * 
 * 
 * 
************************************************************************/
static yMirror(pt, outpt, center)
    dpoint *pt, *outpt;
    double center;
{
    outpt->y = center + (center - pt->y);
    outpt->x = pt->x;
}

/*************************************************************************
 *
 * 
 * 
 * 
************************************************************************/
static void
ellipsePath(bbox, startAng, arcAng, toBeFilled)
TmRect32 *bbox;
int startAng, arcAng;
int toBeFilled; /* bool */
{
    double xAxis, yAxis, centerX, centerY;

    dpoint p0, p1, p2, p3;
    dpoint b0, b1, b2, b3;

    xAxis = bbox->x_size;
    yAxis = bbox->y_size;
    centerX = bbox->x_position + (bbox->x_size / 2);
    centerY = bbox->y_position + (bbox->y_size / 2);

    /* if it's a circle, use the arc call */
    if ( xAxis == yAxis )
    {
        /* if filled, moveto center to create 'wedge edge' in the path */
        if ( toBeFilled && arcAng < 3600 )
            put(NXLT("%.3f %.3f M\n"), centerX, Y(centerY));

        /* now the rest of the arc */
        put(NXLT("%.3f %.3f %.3f %d %d arc\n"),
            centerX, Y(centerY),
            (double)(centerX - bbox->x_position),
            startAng/10, (startAng + arcAng)/10);

        /* if filled, do the other 'wedge edge' */
        if (toBeFilled && arcAng < 3600)
            put(NXLT("%.3f %.3f L\n"), centerX, Y(centerY));
    }
    else
    {
        /* curveto takes exactly 4 control points, (where
         * the 1st is taken as the current position).  For
         * an ellipse, this means 4 curveto's are needed,
         * one for each quadrant of the ellipse.  The 1st
         * and 4th ctl points for each quad fall on the
         * ellipse's path; the intermediate (2nd & 3rd) ctl
         * points fall on the bounding box of the ellipse.
         * "EllFactor" is used to place the intermediate
         * ctl points along the bounding box edge at a
         * position = (EllFactor)(axis) from the center of
         * the edge.   This code determines the ctl point
         * placements for the first quad, then mirrors them
         * around to get the positions for the other quads */

        /* start with upper left quad */
        p0.x = bbox->x_position;
        p0.y = centerY;

        /* next ctl point has same x, but is higher by (EllFactor)yAxis */
        p1.x = p0.x;
        p1.y = p0.y - (EllFactor * yAxis);

        /* etc.. */
        p2.x = centerX - (EllFactor * xAxis);
        p2.y = bbox->y_position;
        p3.x = centerX;
        p3.y = p2.y;
        put(NXLT("%.3f %.3f moveto\n"), p0.x, Y(p0.y));
        put(NXLT("%.3f %.3f %.3f %.3f %.3f %.3f C\n"),
            p1.x, Y(p1.y),
            p2.x, Y(p2.y),
            p3.x, Y(p3.y));

        /* now do upper right quad */
        b0 = p3;
        xMirror(&p2, &b1, centerX);
        xMirror(&p1, &b2, centerX);
        xMirror(&p0, &b3, centerX);
        put(NXLT("%.3f %.3f %.3f %.3f %.3f %.3f C\n"),
            b1.x, Y(b1.y),
            b2.x, Y(b2.y),
            b3.x, Y(b3.y));

        /* now do lower right quad */
        p0 = b3;
        yMirror(&b2, &p1, centerY);
        yMirror(&b1, &p2, centerY);
        yMirror(&b0, &p3, centerY);
        put(NXLT("%.3f %.3f %.3f %.3f %.3f %.3f C\n"),
            p1.x, Y(p1.y),
            p2.x, Y(p2.y),
            p3.x, Y(p3.y));

        /* lower left quad, now */
        b0 = p3;
        xMirror(&p2, &b1, centerX);
        xMirror(&p1, &b2, centerX);
        xMirror(&p0, &b3, centerX);
        put(NXLT("%.3f %.3f %.3f %.3f %.3f %.3f C\n"),
            b1.x, Y(b1.y),
            b2.x, Y(b2.y),
            b3.x, Y(b3.y));
    }
}

/*************************************************************************
 *
 * 
 * 
 * 
************************************************************************/
int D01Ellipse(bbox, start_angle, arc_angle, attr)
    TmRect32 *bbox;
    int start_angle;
    int arc_angle;
    attr_type *attr;
{
    TmPoint32 boundingBox[4];

    /* start */
    PS_GS();

    /* create the path */
    ellipsePath(bbox, start_angle, arc_angle, attr->fill_flag);

    /* init boundary */
    boundingBox[0].x = bbox->x_position;
    boundingBox[0].y = bbox->y_position;
    boundingBox[1].x = bbox->x_position + bbox->x_size;
    boundingBox[1].y = bbox->y_position + bbox->y_size;

    /* fill the path according to pattern info passed in */
    if ( attr->fill_flag )
        D01FillPattern(attr, boundingBox);


    /* stroke the path according to line info passed in */
    if ( attr->l_wid )
        D01FillStroke(attr, boundingBox);


    /* finish */
    PS_GR();
}

/*************************************************************************
 *
 * 
 * 
 * 
************************************************************************/
int D01Save()
{

    put(NXLT("GS\n"));
    return(0);

}

/*************************************************************************
 *
 * 
 * 
 * 
************************************************************************/
int D01Restore() 
{
    
    put(NXLT("GR\n"));

    /* NOW TRASH THE CACHED FONT INFO!!! (LINESTYLE TOO?) */
    Curfont.number = -1;
    Curfont.size = -1;
    PF.cur_thick = MINIMUM_LINE_WIDTH;  /* default page setup */
    PF.cur_style = TM_SOLID;            /* default page setup */
    return(0);
}


/*************************************************************************
 *
 * 
 * 
 * 
************************************************************************/
static BadChar(indx, c)
    int indx;
    char c;
{
    Debug2f("BadChar: Bad char index %d; char code = %d\n", indx, c);
}



/*************************************************************************
 *
 * 
 * draw a clip box for (x1, y1) to (x2, y2)
 * 
 * 
************************************************************************/
static void clip_box(x1, y1, x2, y2)
    int x1, y1, x2, y2 ;
{
    int n = PF.ClipGSaveLevel ;

    if (x1 != current_clip[n].x1 ||
        y1 != current_clip[n].y1 ||
        x2 != current_clip[n].x2 ||
        y2 != current_clip[n].y2)
    {
        put(NXLT("%d %d %d %d DOCLIPBOX\n"), x1, Y(y1), x2, Y(y2)) ;
        current_clip[n].x1 = x1 ;
        current_clip[n].y1 = y1 ;
        current_clip[n].x2 = x2 ;
        current_clip[n].y2 = y2 ;
    }
}


/*************************************************************************
 *
 * 
 * 
 * 
************************************************************************/
int
D01PutImage(x_pos_mil, y_pos_mil, imgP, mode, xform, cmap, attr, clip)
    int x_pos_mil;
    int y_pos_mil;
    ge_image_type *imgP;
    int mode;
    poly_type *xform;
    cmap_type *cmap;
    attr_type *attr;
    TmRect32 *clip ;
{
    int height_mil, width_mil;
    int height_pix, width_pix;
    int x_pos_pix, y_pos_pix;
    int transx_mil, transy_mil;      /* x,y translation */
    int transx_pix, transy_pix;
    double scalex, scaley;            /* x,y scale factors */
    double cosAng, sinAng;            /* sin, cos of rotation angle */
    double xShear;                    /* x shear */
    double a1, a2, a3, a4, a5, a6;    /* terms for concat matrix */
    int im_fg_color;                    /* fg color for 1-plane images */
    int fill, close, pat, pointcount;
    int boxwid=0, boxht=0;            /* width & height of bounding box */
    TmPoint image_box[5];              /* corners of bounding box */

    if ((imgP->eps_seg_size != 0) && (imgP->eps_seg != NULL))
    {
        D01EPSInclude(x_pos_mil, y_pos_mil, imgP->eps_seg,
                      imgP->eps_seg_size, xform);
        return(0);
    }

    scalex = 1.0;
    scaley = 1.0;

    /* any units returned by extract_terms() should be in mils */

    extract_terms_for_device(imgP, xform,
                  &transx_mil, &transy_mil, &scalex, &scaley,
                  &sinAng, &cosAng, &xShear, &width_mil, &height_mil, 1);

    width_pix  = xmiltopix(width_mil);
    height_pix = ymiltopix(height_mil);
    transx_pix = xmiltopix(transx_mil);
    transy_pix = ymiltopix(transy_mil);
    x_pos_pix  = xmiltopix(x_pos_mil);
    y_pos_pix  = ymiltopix(y_pos_mil);

	/* no width, no height, no service */
	if (width_pix == 0 || height_pix == 0)
		return 0 ;

    /* do all this in our own graphics context */
    PS_GS();

    D01DrawClip(clip) ;

    /* Recompute translation:
     * The translation terms returned by pfGetTextXform are
     * for translation required to translate top left corner
     * of untransformed image from origin of coordiante system
     * to its transformed position.  Add absolute position of
     * origin to get actual translation. */
    transx_pix = x_pos_pix + transx_pix;
    transy_pix = Y(y_pos_pix + transy_pix);

    /* Adjust scale:
     * The scale returned by pfGetTextXform is to scale original
     * image (x_size by y_size) to the existing size.  Adjust to
     * scale unit square image (returned by "image" operator) to
     * the original size. */
    scalex = scalex * (double) imgP->wid;
    scaley = scaley * (double) imgP->hyt;

    /*
     * Compute concat matrix terms:
     *
     * The order of transformations to transform unit square image
     * at PS origin to existing shape and pos is 
     * Scale-Shear-Rotate-Translate (SXRT).  The final CTM is
     * [Scale] x [Shear] x [Rotate] x [Translate] x CTM. 
     * The result matrix from the first four terms is computed
     * below and passed to the "concat" operator which replaces
     * the CTM with [our matrix] x CTM.
     * NOTE: The matrix in the "image" operator produces an image
     *     of unit size, with the image origin at the PS page
     *     origin, and having the same orientation as the original.
     */
    a1 = scalex * cosAng + 0.0 ;
    a2 = scalex * sinAng ;

    /* sheer insanity to get rid of -0.0 */
    if (a2 != 0.0)
        a2 = -a2 ;

    a3 = -(xShear * scaley * cosAng) + (scaley * sinAng);
    a4 = (xShear * scaley * sinAng) + (scaley * cosAng);
    a5 = (double) transx_pix;
    a6 = (double) transy_pix;

    /* do concat */
    put(NXLT("[%.4f %.4f %.4f %.4f %.4f %.4f] concat\n"), a1, a2, a3, a4, a5, a6);

    if (imgP->dep == 1)
    {
        if ( (mode != TMS_O_D) && (mode != TMIS_O_D) )
        {
            PF.PfPage.trans_mode = mode;
            /* opaque 1-plane image: fill with background color */
            D01SetCMYKColor(attr->fbg.c,
                            attr->fbg.m,
                            attr->fbg.y,
                            attr->fbg.k, 255);
            /* under a transform; unit square will be scaled */
            D01ClrBox(0, PF.pagetop, 1, 1);

        }
        if ( mode == TMIS_O_D )
            im_fg_color = PF.PfWin_bg_color;
        else
            im_fg_color = PF.PfWin_fg_color;

        /* set foreground color */
        D01SetCMYKColor(attr->ffg.c,
                        attr->ffg.m,
                        attr->ffg.y,
                        attr->ffg.k, 255);

        DrawMonoImage(x_pos_mil, y_pos_mil, mode, imgP);
    }

    else if (!ImageTransparent(x_pos_mil, y_pos_mil, mode, imgP, cmap))
    {
        
        /* color image */
        if (PF.PfColordev)
        {
            switch (pf_color_model)
            {
              default:
            case COL_MODEL_CMYK:
                DrawColorImageAsCMYK(x_pos_mil, y_pos_mil, mode, imgP, cmap);
                break;
            case COL_MODEL_RGB:
                DrawColorImageAsRGB(x_pos_mil, y_pos_mil, mode, imgP, cmap);
                break;
            case COL_MODEL_GRAY:
                DrawColorImageAsGray(x_pos_mil, y_pos_mil, mode, imgP, cmap);
                break;

            }
        }
        else
        {
            DrawColorImageAsGray(x_pos_mil, y_pos_mil, mode, imgP, cmap);
        }
    }
    PF_MEM_CHECK;

    PS_GR();

    /*  Draw box around image */
    fill = close = pat = 0;
    pointcount = 5;
    image_box[0].x = x_pos_mil;
    image_box[0].y = y_pos_mil;
    image_box[1].x = x_pos_mil;
    image_box[1].y = y_pos_mil + boxht;
    image_box[2].x = x_pos_mil + boxwid;
    image_box[2].y = y_pos_mil + boxht;
    image_box[3].x = x_pos_mil + boxwid;
    image_box[3].y = y_pos_mil;
    image_box[4].x = image_box[0].x;
    image_box[4].y = image_box[0].y;

    return(0);
}


/*************************************************************************
 * 
    ImageTransparent()
 *
 *  If this image has any tranparent pixels return FALSE,
 *  Otherwise generate the code to draw it and return TRUE
 * 
************************************************************************/
static int ImageTransparent(x_pos, y_pos, mode, imgP, cmap)
    int x_pos, y_pos, mode;
    ge_image_type *imgP;
    cmap_type *cmap;
{
    int i, j, k ;
    unsigned char *dP ;
    int color_used[256] ;
    unsigned char x ;
    int trans = FALSE ;
    unsigned short a;
    int len;
    int x_size, y_size;
    int data_width_bytes, image_width_bytes;
    int bit, bitcount ;

    for (k = 0 ; k < 256 ; k++)
        color_used[k] = FALSE ;

    x_size = imgP->wid;
    y_size = imgP->hyt;
    data_width_bytes = imgP->data_byte_wid;

    /* determine which colors are used, and if any are transparent */
     ;
    for (i = 0, dP = imgP->data ; i < y_size ; i++, dP += data_width_bytes)
        for (j = 0 ; j < x_size ; j++)
        {
            x = dP[j] ;
            if (cmap->color[x].see_thru)
                trans = TRUE ;
            else
                color_used[x] = TRUE ;
        }
    
    if (!trans)
        return FALSE ;

    /* an image mask for each color used */
    image_width_bytes = (x_size + 7) >> 3;
    for (k = 0 ; k < 255 ; k++)
    {
        if (!color_used[k])
            continue ;

        D01SetCMYKColor(cmap->color[k].c,
                        cmap->color[k].m,
                        cmap->color[k].y,
                        cmap->color[k].k, 255);


        put(NXLT("/AXimagebuf %d string def\n"), image_width_bytes);
        
#if 0
        if(mode == TMIS_O_D)
            put(NXLT("%d %d true "), x_size, y_size);
        else
            put(NXLT("%d %d false "), x_size, y_size);
#else
        put(NXLT("%d %d true "), x_size, y_size);
#endif
    
        put(NXLT("[%d 0 0 -%d 0 0] \n"), x_size, y_size);

        put(NXLT("{currentfile AXimagebuf readhexstring pop} "));
    
        put(NXLT("imagemask\n"));

        dP = imgP->data;
        for (i = 0, dP = imgP->data ; i < y_size ; i++, dP += data_width_bytes)
        {
            len = 0;
            for (j = 0, a = 0, bitcount = 0 ; j < x_size; j++)
            {
                if (dP[j] == k)
                    bit = 1 ;
                else
                    bit = 0 ;
                a = (a << 1) | bit ;
                bitcount++ ;
                if (bitcount == 8)
                {
                    put("%02x", a);
                    len += 2;
                    if (len == 78)
                    {
                        put(NXLT("\n"));
                        len = 0;
                    }
                    a = bitcount = 0 ;
                }
            }
            if (bitcount != 0)
            {
                a = a << (8-bitcount) ;
                put("%02x", a) ;
            }
            put(NXLT("\n"));
        }
    }
    return TRUE;
}

/*************************************************************************
 *
 * 
    DrawMonoImage()
 * 
 * 
************************************************************************/
static void DrawMonoImage(x_pos, y_pos, mode, imgP)
int x_pos, y_pos, mode;
ge_image_type *imgP;
{
    register unsigned char *scan_line_data;
    unsigned char *datap;
    int scan_line, data_width_bytes;
    register int col;
    register int image_width_bytes;
    unsigned short a;
    int len;
    int x_size, y_size;

    x_size = imgP->wid;
    y_size = imgP->hyt;
    datap = imgP->data;
    data_width_bytes = imgP->data_byte_wid;
    image_width_bytes = (x_size + 7)>>3; /* convert bits to bytes, round up */ 

    put(NXLT("/AXimagebuf %d string def\n"), image_width_bytes);
        
    if(mode == TMIS_O_D)
        put(NXLT("%d %d true "), x_size, y_size);
    else
        put(NXLT("%d %d false "), x_size, y_size);
    
    put(NXLT("[%d 0 0 -%d 0 0] \n"), x_size, y_size);

    put(NXLT("{currentfile AXimagebuf readhexstring pop} "));
    
    put(NXLT("imagemask\n"));

    scan_line_data = datap;
    for( scan_line = y_size; scan_line--; )
    {
        len = 0;
        for ( col = 0; col < image_width_bytes; )
        {
            if (PF.PfWin_fg_color == TMA_WHITE && PF.PfWin_bg_color == TMA_BLACK)
            {
                /* inverted image */
                a = scan_line_data[col++] & 255;
            }
            else {
                a = (~scan_line_data[col++]) & 255;
            }
            put("%02x", a);
            len += 2;
            if ( (len == 78) && (col < data_width_bytes) )
            {
                put(NXLT("\n"));
                len = 0;
            }
        }
        put(NXLT("\n"));

        /* get next scan line */
        scan_line_data += data_width_bytes;
    }
}


/*************************************************************************
 *
 * 
    DrawColorImageAsRGB()
 * 
 * 
************************************************************************/
static void DrawColorImageAsRGB(x_pos, y_pos, mode, imgP, cmap)
    int x_pos, y_pos, mode;
    ge_image_type *imgP;
    cmap_type *cmap;
{
    register char *colorImagebuf, *grayImagebuf;
    char *datap;
    unsigned short a;
    int col, line;
    int c_lshift, g_lshift;
    int color_x_bytes, gray_x_bytes, bx_size;
    unsigned char red, green, blue, gray_c;
    unsigned int cyan, magenta, yellow, black;
    int len;
    int x_size, y_size;
    cmap_cache_entry *cmap_cache;
    
    /* 
     * To prevent color image jobs from dying on b&w printers,
     * the data for both color & gray images is sent down.
     *
     * One byte in the source data corresponds to one pixel, and
     * is a value calculated as (blue + green*5 + red*25).
     *
     * For printing as color:
     * The source color pixel values are first 
     * mapped to actual values using the array pf_Ax64Values[3][5].  The 
     * new pixel values are in the range 0-15, and hence four bits are 
     * used to represent each pixel in PostScript.  The data is 
     * supplied in the form RRRRGGGG BBBBRRRR GGGGBBBB RRRRGGGG .....
     * (with the multiproc argument to colorimage set to false).  
     *
     * For printing as grayscale:
     * The data is formatted as four bits/pixel, allowing the image call
     * to render a sixteen-level grayscale image.  These four bits are 
     * defined using the NTSC conversion from RGB to gray.
     */

    x_size = imgP->wid;
    y_size = imgP->hyt;
    color_x_bytes = ((x_size*numColorComponents*bitsPerColorComponent) + 7)>>3; 

    /* #pixels*3colors*2bits/8 rounded to one extra pixel */
    gray_x_bytes = ((x_size*bitsPerGrayComponent) + 7) >> 3;

    bx_size = x_size + x_size%2;

    put(NXLT("/AXcolorImagebuf %d string def "), color_x_bytes);
    put(NXLT("/AXgrayImagebuf %d string def\n"), gray_x_bytes);
    put(NXLT("%d %d %d "), x_size, y_size, bitsPerColorComponent);

    put(NXLT("[%d 0 0 -%d 0 0] \n"), x_size, y_size); /* "always" xformed */
    put(NXLT("{ currentfile AXcolorImagebuf readhexstring pop currentfile\n"));
    put(NXLT("AXgrayImagebuf readhexstring pop SELECTBUF } COLORIMAGE\n"));

    colorImagebuf = (char *)malloc(color_x_bytes);
    grayImagebuf = (char *)malloc(gray_x_bytes);
    cmap_cache =
        (cmap_cache_entry *)calloc(cmap->n_colors, sizeof(cmap_cache_entry));

    for (line=0; line <cmap->n_colors; line++)
        cmap_cache[line].gray_c = -1;

    for( line = y_size, datap = (char *)imgP->data ;
            line > 0 ;
            line--, datap += bx_size )
    {

        /* transfer data from 2 src bytes to 3 dest bytes */
        memset(colorImagebuf, 0, color_x_bytes);
        memset(grayImagebuf, 0, gray_x_bytes);

        GenerateRGBGrayImages(datap, bx_size,
                    colorImagebuf, grayImagebuf,
                    cmap, cmap_cache) ;

        /* write row with color dest bytes */
        len = 0;
        for ( col = 0; col < color_x_bytes; )
        {
            a = colorImagebuf[col++] & 0xFF;
            put("%02x", a);
            if ( (len == 78) && (col < color_x_bytes) )
            {
                put("\n");
                len = 0;
            }
            else
                len += 2;
        }
        put("\n");

        /* write row with gray dest bytes */
        len = 0;
        for ( col =0; col < gray_x_bytes; )
        {
            a = grayImagebuf[col++] & 0xFF;
            put("%02x", a);
            len += 2;
            if ( (len == 78) && (col < gray_x_bytes) ) {
                put("\n");
                len = 0;
            }
        }
        put("\n");
    }
    free(colorImagebuf);
    free(grayImagebuf);
    free(cmap_cache);
}



/******************************************************************************
*
    GenerateRGBGrayImages()

*   Generate parallel rows of RGB color and gray bytes from
*   the color image row in buf
*
******************************************************************************/
static void GenerateRGBGrayImages(buf, width, colorImagebuf, grayImagebuf, cmap, cmap_cache)
    char *buf, *colorImagebuf, *grayImagebuf;
    int width ;
    cmap_type *cmap;
    cmap_cache_entry *cmap_cache;
{
    int sx, cx, gx, ix ;
    register int c_lshift, g_lshift;
    int cyan, magenta, yellow, black;
    unsigned char red, green, blue, gray_c;

    c_lshift = bitsPerColorComponent;
    g_lshift = bitsPerGrayComponent;

    for (sx=cx=gx=0 ; sx < width ; sx++)
    {
        ix = buf[sx] & 0xFF;

        if (cmap_cache[ix].gray_c == -1)
        {
            cyan =  cmap->color[ix].c;
            magenta = cmap->color[ix].m;
            yellow =  cmap->color[ix].y;
            black =   cmap->color[ix].k;

            /* calculate gray */
            red =   255 - MIN((int)(cyan+black), 255);
            green = 255 - MIN((int)(magenta+black), 255);
            blue = 255 - MIN((int)(yellow+black), 255);

            if (bitsPerColorComponent == 4)
            {
                gray_c = ((30*red)+(59*green)+(11*blue))/((unsigned)1600);
                cmap_cache[ix].gray_c = MIN(15,gray_c);

                /* hack each color component down to 4 bits */
                red = (red + 7) / (unsigned)16;
                cmap_cache[ix].red = MIN(15,red);
                    
                green = (green + 7) / (unsigned)16;
                cmap_cache[ix].green = MIN(15,green);

                blue = (blue + 7) / (unsigned)16;
                cmap_cache[ix].blue = MIN(15,blue);
            }
            else
            {   /* assumed 8 bits... */
                cmap_cache[ix].gray_c =
                    ((30*red) + (59*green) + (11*blue))/(unsigned)100;
                cmap_cache[ix].red = red;
                cmap_cache[ix].green = green;
                cmap_cache[ix].blue = blue;
            }
        }

        /* stuff 'em into the color buffer */
        if (bitsPerColorComponent == 4)
        {
            colorImagebuf[cx] |= (cmap_cache[ix].red << c_lshift);
            c_lshift -= bitsPerColorComponent;
            if (c_lshift < 0)
            {
                c_lshift = bitsPerColorComponent;
                cx++;
            }

            colorImagebuf[cx] |= (cmap_cache[ix].green << c_lshift);
            c_lshift -= bitsPerColorComponent;
            if (c_lshift < 0)
            {
                c_lshift = bitsPerColorComponent;
                cx++;
            }

            colorImagebuf[cx] |= (cmap_cache[ix].blue << c_lshift);
            c_lshift -= bitsPerColorComponent;
            if (c_lshift < 0)
            {
                c_lshift = bitsPerColorComponent;
                cx++;
            }

        }
        else
        {
            colorImagebuf[cx] |= cmap_cache[ix].red;
            cx++;

            colorImagebuf[cx] |= cmap_cache[ix].green;
            cx++;

            colorImagebuf[cx] |= cmap_cache[ix].blue;
            cx++;
        }

        if (bitsPerGrayComponent == 4)
        {
            grayImagebuf[gx] |= (cmap_cache[ix].gray_c << g_lshift);
            g_lshift -= bitsPerGrayComponent;
            if (g_lshift < 0)
            {
                g_lshift = bitsPerColorComponent;
                gx++;
            }
        }
        else
        {
            grayImagebuf[gx] |= (cmap_cache[ix].gray_c);
            gx++;
        }

    }
}
/*************************************************************************
 *
    DrawColorImageAsCMYK()
 * 
************************************************************************/
static void DrawColorImageAsCMYK(x_pos, y_pos, mode, imgP, cmap)
    int x_pos, y_pos, mode;
    ge_image_type *imgP;
    cmap_type *cmap;
{
    char *colorImagebuf, *grayImagebuf;
    char *datap;
    int col, line;
    int color_x_bytes, gray_x_bytes, bx_size;
    int len;
    int x_size, y_size;
    cmap_cache_entry *cmap_cache;

    /* 
     * To prevent color image jobs from dying on b&w printers,
     * the data for both color & gray images is sent down.
     *
     * One byte in the source data corresponds to one pixel, and
     * is a value calculated as (blue + green*5 + red*25).
     *
     * For printing as color:
     *
     * For printing as grayscale:
     * The data is formatted as four bits/pixel, allowing the image call
     * to render a sixteen-level grayscale image.  These four bits are 
     * defined using the NTSC conversion from RGB to gray.
     */

    x_size = imgP->wid;
    y_size = imgP->hyt;
    color_x_bytes = ((x_size*numColorComponents*bitsPerColorComponent) + 7)>>3; 

    gray_x_bytes = ((x_size*bitsPerGrayComponent) + 7)>>3;

    bx_size = x_size + x_size%2;

    put(NXLT("/AXcolorImagebuf %d string def\n"), color_x_bytes);
    put(NXLT("/AXgrayImagebuf %d string def\n"), gray_x_bytes);
    put(NXLT("%d %d %d "), x_size, y_size, bitsPerColorComponent);
    put(NXLT("[%d 0 0 -%d 0 0] \n"), x_size, y_size);   /* "always" xformed */
    put(NXLT("{ currentfile AXcolorImagebuf readhexstring pop\n"));
    put(NXLT("  currentfile AXgrayImagebuf readhexstring pop SELECTBUF } COLORIMAGE\n"));

    colorImagebuf = malloc(color_x_bytes);
    grayImagebuf = malloc(gray_x_bytes);

    cmap_cache =
        (cmap_cache_entry *)calloc(cmap->n_colors, sizeof(cmap_cache_entry));

    for (line=0; line <cmap->n_colors; line++)
        cmap_cache[line].gray_c = -1;

    for (line = y_size, datap = (char *)imgP->data ;
            line > 0 ;
            line--, datap += bx_size)
    {

        /* transfer data from 2 source bytes to 4 dest bytes */
        memset(colorImagebuf, 0, color_x_bytes);
        memset(grayImagebuf, 0, gray_x_bytes);

        GenerateCMYKGrayImages(datap, bx_size,
                    colorImagebuf, grayImagebuf,
                    cmap, cmap_cache) ;

        /* write color bytes */
        for (len = col = 0; col < color_x_bytes; )
        {
            put("%02x", colorImagebuf[col++] & 0xFF) ;
            if (len == 78)
            {
                put("\n");
                len = 0;
            }
            else
                len += 2;
        }
        if (len)
            put("\n");

        /* write gray bytes */
        for (len = col =0; col < gray_x_bytes; )
        {
            put("%02x", grayImagebuf[col++] & 0xFF) ;
            if (len == 78)
            {
                put("\n");
                len = 0;
            }
            else
                len += 2;
        }
        if (len)
            put("\n");
    }

    /* let my buffers go, oh Lord */
    free(colorImagebuf);
    free(grayImagebuf);
    free(cmap_cache);
}

/******************************************************************************
*

    GenerateCMYKGrayImages()
*
*   Generate parallel rows of CMYK color and gray bytes from
*   the color image row in buf
*
******************************************************************************/
static void GenerateCMYKGrayImages(buf, width, colorImagebuf, grayImagebuf, cmap, cmap_cache)
    char *buf, *colorImagebuf, *grayImagebuf;
    int width ;
    cmap_type *cmap;
    cmap_cache_entry *cmap_cache;
{
    int sx, cx, gx, ix ;
    register int c_lshift, g_lshift;
    int cyan, magenta, yellow, black;
    unsigned char red, green, blue, gray_c;
    double gray_f;

    c_lshift = bitsPerColorComponent;
    g_lshift = bitsPerGrayComponent;
    for (sx = cx = gx = 0 ; sx < width ; sx++)
    {
        ix = (unsigned int)(buf[sx] & 0xFF);

        if (cmap_cache[ix].gray_c == -1)
        {
            if (cmap->color[ix].see_thru)
            {
                cyan =  0;
                magenta = 0;
                yellow =  0;
                black =   0;
            }
            else
            {
                cyan =  cmap->color[ix].c;
                magenta = cmap->color[ix].m;
                yellow =  cmap->color[ix].y;
                black =   cmap->color[ix].k;
            }

            /* calculate gray */
            red =   255 - MIN((int)(cyan+black), 255);
            green = 255 - MIN((int)(magenta+black), 255);
            blue =  255 - MIN((int)(yellow+black), 255);

            if (bitsPerColorComponent == 4)
            {

                gray_c = ((30*red)+(59*green)+(11*blue))/((unsigned)1600);
                cmap_cache[ix].gray_c = MIN(15,gray_c);

                gray_f = 15.0 -
                        (15.0 * IntensFromCMYKColor(cyan, magenta, yellow, black, 255));
                gray_c = (unsigned char)gray_f;
                cmap_cache[ix].gray_c = MIN(15,gray_c);

                /* hack each color component down to 4 bits */
                cyan = (cyan +7) / 16;
                cmap_cache[ix].cyan = MIN(15,cyan);

                magenta = (magenta +7) / 16;
                cmap_cache[ix].magenta = MIN(15,magenta);

                yellow = (yellow +7) / 16;
                cmap_cache[ix].yellow = MIN(15,yellow);

                black = (black +7) / 16;
                cmap_cache[ix].black = MIN(15,black);
            }
            else
            {   /* assumed 8 bits... */

                cmap_cache[ix].gray_c =
                        ((30*red) + (59*green) + (11*blue))/((unsigned)100);
                cmap_cache[ix].cyan = cyan;
                cmap_cache[ix].magenta = magenta;
                cmap_cache[ix].yellow = yellow;
                cmap_cache[ix].black = black;
            }
        }

        /* stuff 'em into the color buffer */
        if (bitsPerColorComponent == 4)
        {
            colorImagebuf[cx] |=
                    (cmap_cache[ix].cyan << c_lshift);
            c_lshift -= bitsPerColorComponent;
            if (c_lshift < 0)
            {
                c_lshift = bitsPerColorComponent;
                cx++;
            }

            colorImagebuf[cx] |=
                    (cmap_cache[ix].magenta << c_lshift);
            c_lshift -= bitsPerColorComponent;
            if (c_lshift < 0)
            {
                c_lshift = bitsPerColorComponent;
                cx++;
            }

            colorImagebuf[cx] |=
                    (cmap_cache[ix].yellow << c_lshift);
            c_lshift -= bitsPerColorComponent;
            if (c_lshift < 0)
            {
                c_lshift = bitsPerColorComponent;
                cx++;
            }

            colorImagebuf[cx] |=
                    (cmap_cache[ix].black << c_lshift);
            c_lshift -= bitsPerColorComponent;
            if (c_lshift < 0)
            {
                c_lshift = bitsPerColorComponent;
                cx++;
            }
        }
        else
        {
            colorImagebuf[cx] |= cmap_cache[ix].cyan;
            cx++;

            colorImagebuf[cx] |= cmap_cache[ix].magenta;
            cx++;

            colorImagebuf[cx] |= cmap_cache[ix].yellow;
            cx++;

            colorImagebuf[cx] |= cmap_cache[ix].black;
            cx++;
        }

        if (bitsPerGrayComponent == 4)
        {
            grayImagebuf[gx] |= (cmap_cache[ix].gray_c << g_lshift);
            g_lshift -= bitsPerGrayComponent;
            if (g_lshift < 0)
            {
                g_lshift = bitsPerColorComponent;
                gx++;
            }
        }
        else
        {
            grayImagebuf[gx] |= (cmap_cache[ix].gray_c);
                gx++;
        }
    }
}

/*************************************************************************
 *
 * 
    DrawColorImageAsGray()
 * 
 * 
************************************************************************/
static void DrawColorImageAsGray(x_pos, y_pos, mode, imgP, cmap)
int x_pos,y_pos,mode;
ge_image_type *imgP;
cmap_type *cmap;
{
    int x_bytes, bx_size;
    int x_size, y_size;

    x_size = imgP->wid;
    y_size = imgP->hyt;
    x_bytes = ((x_size*bitsPerGrayComponent) + 7)>>3;

    bx_size = x_size + x_size%2;

    put(NXLT("/AXimagebuf %d string def\n"), x_bytes);
    put(NXLT("%d %d %d "), x_size, y_size, bitsPerGrayComponent);

    put(NXLT("[%d 0 0 -%d 0 0] \n"), x_size, y_size);
    put(NXLT("{currentfile AXimagebuf readhexstring pop} image\n"));

    GenerateGrayImage((char *)imgP->data, bx_size, x_size, y_size, cmap) ;

}


/*************************************************************************
 *
 * 
    GenerateGrayImage()
 * 
 * 
************************************************************************/
static void GenerateGrayImage(datap, bx_size, x_size, y_size, cmap)
    char *datap;
    int bx_size;
    int x_size, y_size;
    cmap_type *cmap ;
{
    char *out_image_line;
    unsigned short a;
    short *FSbuffer, *FS;
    int sx, line, col, len, x_bytes ;

    x_bytes = ((x_size*bitsPerGrayComponent) + 7)>>3;
    out_image_line = malloc(x_bytes);
    FSbuffer = (short *)malloc(bx_size * y_size * sizeof(short));

    ColorToGray(datap, x_size, y_size, cmap, FSbuffer, 16);

    /* FSbuffer now contain 0..255 values, each in a short, must
       quantize into bitsPerGrayComponent */
    for( line = 0 ; line < y_size; line++ )
    {
        FS = &FSbuffer[line*bx_size] ;
        for(sx = 0, col=0; col<x_bytes; sx++, col++)
        {
            if (bitsPerGrayComponent == 4)
            {
                /* Take only the high nibble from FSBuffer[] but pack
                   it into *alternate* nibbles of out_image_line[] */
                out_image_line[col] = (((uint8)(FS[sx++])) >> 4) << 4;
                out_image_line[col] |= ((uint8)(FS[sx])) >> 4;
            }
            else    /* assume 8 bits for now... */
                out_image_line[col] = ((uint8)(FS[sx]));
        }

        /* write */
        len = 0;
        for (col=0; col<x_bytes; col++)
        {
            a = out_image_line[col] & 0xFF;
            put("%02x", a);
            len += 2;
            if ( (len == 78) && (col < x_bytes) ) {
                put(NXLT("\n"));
                len = 0;
            }
        }
        put("\n");
    }
    free(FSbuffer);
    free(out_image_line);
}


/*************************************************************************
 *
    This tries to kind of emulate strtok() but we're hardwired to \n
     as a token separator.
 * 
************************************************************************/
static int local_fgets(source, target, target_size)
     char *source;
     char *target;
     int target_size;
{
    int strindex;
    char ch;
    static char *current_buffer;

    if (source != NULL)
        current_buffer = source;
    for (strindex=1; strindex<target_size; strindex++)
    {
        ch = *current_buffer++;
        if (ch == '\n')
            break;
        if (ch < ' ' && ch != '\t')
            return 0 ;
        *target++ = ch;
    }
    *target = '\0';
    return(strindex);
}

/*************************************************************************
 *
 * 
    comment_match()

 * lifted from epsi.c, thanx JimR
 * 
 * 
************************************************************************/
static char *comment_match(s1, s2)
     register char *s1;
     register char *s2;
{
    int ch;
 
    /*
     * compare 2 words in upper case; the 2nd word is already caps
     * a space in s2 will match any amount of whitespace in s1
     * return
     */
    if (*s1 != '%') /* must be a comment */
        return(NULL);
    while (*s1 && !isalnum(*s1))    /* skip to first "word" */
        s1++;

    while (*s2) {
        if (*s2 == ' ') {
            while (isspace(*s1))
                ++s1;
        }
        else {
            ch = *s1++;
            if (islower(ch))
                ch = toupper(ch);
            if (ch != *s2)
                return NULL;
        }
        ++s2;
    }
    return s1;

}

/*************************************************************************
 *
 * 
    GetBoundingBoxFromEPS()
 * 
 * 
************************************************************************/
static void GetBoundingBoxFromEPS(eps_block, eps_block_len, BBox)
    char *eps_block;
    int eps_block_len;
    TmPoint32 *BBox;
{
    char buf[MAX_PS_LINE+1];
    char *ap;
    char *paren_found;
    double llxflt, llyflt, urxflt, uryflt;
    int just_got;
    int bytes_remaining = eps_block_len;

    BBox[0].x = 0;
    BBox[0].y = 0;
    BBox[1].x = 0;
    BBox[1].y = 0;

    if ((eps_block == NULL) || (eps_block_len == 0))
        return;

    just_got = local_fgets(eps_block, buf, MAX_PS_LINE);

    while ((just_got) && (bytes_remaining > 0))
    {
        if (ap = comment_match(buf, NXLT("BOUNDINGBOX:")))
        {
            paren_found = strchr((char *)ap, '(');
            if (paren_found == NULL)
            {
                /* absence of a paren char is taken to mean
                 * that real bbox numbers are supplied on this
                 * line rather than the (atend) notation
                 */
                sscanf(ap, "%lf %lf %lf %lf",
                       &llxflt, &llyflt, &urxflt, &uryflt);
                BBox[0].x = (int)llxflt;
                BBox[0].y = (int)llyflt;
                BBox[1].x = (int)urxflt;
                BBox[1].y = (int)uryflt;
                return;
            }
#if 0
            else
            {
                paren_found++;
                if (strncmp(paren_found, NXLT("atend"), 5) == 0)
                    break;       /* skip! */
            }
#endif
        }
        just_got = local_fgets(NULL, buf, MAX_PS_LINE);
        bytes_remaining -= just_got;
    }

    return;
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
static void
pfGetInfoFromEPS(eps_block, eps_block_len, eps_info_block)
char *eps_block;
int eps_block_len;
eps_info *eps_info_block;
{
}

/*-----------------------------------------------------------------------------
 
    How to include an EPS:
     - Do the setup
     - translate to new origin
     - scale to fit placement box
     - translate to final position
     - include the bytes
     - Restore everything

-----------------------------------------------------------------------------*/
int D01EPSInclude(x_pos, y_pos, eps_block, eps_block_len, xform)
    int x_pos;
    int y_pos;
    char *eps_block;
    int eps_block_len;
    poly_type *xform;   /* the points defining the target of the image/EPS */
{
    int ix, tx, ty;
    int deltax, deltay ;
    char *ixp ;
    TmPoint32 EPSBoundingBox[2];
    TmPoint32 EPSBoundingBox_mils[2];
    r32 referenceCell;
    b32 xformedCell;
    axform ctm;
    axformTerms terms, fieldTerms;

    PS_GS();
    GetBoundingBoxFromEPS(eps_block, eps_block_len, EPSBoundingBox);

    if ((EPSBoundingBox[1].x - EPSBoundingBox[0].x == 0) ||
        (EPSBoundingBox[1].y - EPSBoundingBox[0].y == 0)) {
        return (0);
    }

    /* the eps bounding box is in the default postscript resolution
     * of 72dpi; convert to 1000dpi
     */
    EPSBoundingBox_mils[0].x = (EPSBoundingBox[0].x * 1000) / 72;
    EPSBoundingBox_mils[0].y = (EPSBoundingBox[0].y * 1000) / 72;
    EPSBoundingBox_mils[1].x = (EPSBoundingBox[1].x * 1000) / 72;
    EPSBoundingBox_mils[1].y = (EPSBoundingBox[1].y * 1000) / 72;

    /* the referenceCell is just the eps bounding box */
    referenceCell.origin.x = referenceCell.origin.y = 0;
    referenceCell.width = abs(EPSBoundingBox[1].x -
                              EPSBoundingBox[0].x);
    referenceCell.height = abs(EPSBoundingBox[1].y -
                               EPSBoundingBox[0].y);
    
    /* the xformedCell is this same box, but however transformed
       by the including application (typically the graphics editor) */
    xformedCell.p0.x = xform->points[0].x;
    xformedCell.p0.y = xform->points[0].y;
    xformedCell.p1.x = xform->points[1].x;
    xformedCell.p1.y = xform->points[1].y;
    xformedCell.p2.x = xform->points[2].x;
    xformedCell.p2.y = xform->points[2].y;
    xformedCell.p3.x = xform->points[3].x;
    xformedCell.p3.y = xform->points[3].y;

    /* compare xformed and reference cells to get transform terms */
    pfGetTransformTermsFromBox(&referenceCell, &xformedCell, &terms);
    
    /* the transform terms arrive in a form relative to our internal,
       y-increasing-downwards coord sys.  Postscript's sys is normal,
       so account for the differences on the transform terms before
       converting to ctm form.  (that is, negate sines and shears) */
    terms.angle = -terms.angle;
    terms.xshear = -terms.xshear;
    
    /* convert the transform terms to ctm form */
        /* (need to specify a field transform of unity,
            and x and y zoom factors of 100% --> 1000 (also unity)) */
    fieldTerms.unity = TRUE;
    fieldTerms.angle = 0;
    fieldTerms.xscale = 10000;
    fieldTerms.yscale = 10000;
    fieldTerms.xshear = 0;   /* these values mean 'no field transform' */
    axtWriteTargetCtm(1000, 1000, &terms, &fieldTerms, &ctm);
    
    /* translate the coord system origin to the upper left
     * corner of the including app's bounder placement
     */

    /* first, find the offset between xform->points[0] and
     * the upperleft of the rectangular bounder of xform->points
    */
    tx = xform->points[0].x;
    ty = xform->points[0].y;
    for (ix = 1; ix < 4; ix++) {
        if (xform->points[ix].x < tx)
            tx = xform->points[ix].x;
        if (xform->points[ix].y < ty)
            ty = xform->points[ix].y;
    }
    tx = xform->points[0].x - tx;
    ty = xform->points[0].y - ty;
	deltax = abs(xform->points[2].x - xform->points[0].x) ;
	deltay = abs(xform->points[2].y - xform->points[0].y) ;
	/* clip to stay here */
#if 1 /* may need to revist this code */
	clip_box(x_pos, y_pos, x_pos + deltax, y_pos + deltay) ;
#endif

    put(NXLT("BEGINEPSF\n"));  /* initialize state for eps include */

    put(NXLT("%d %d translate\n"), x_pos + tx, Y(y_pos + ty));
    
    /* apply the transform */
    put(NXLT ("[%.4f %.4f %.4f %.4f %d %d] concat\n"),
        ctm.forward.ca,  ctm.forward.cb,
        ctm.forward.cc,  ctm.forward.cd, 0, 0);

    /* now translate by the negative of the eps boundingbox upper left */
    put(NXLT("%d %d translate\n"), -EPSBoundingBox[0].x,
        -EPSBoundingBox[1].y);

    /* now output the eps data and reset state */
    
#if NOTYET
    if ( PfEpsTitles[PF.PfNumEPSProcessed] )
        put(NXLT("%%%%BeginDocument:%s\n"),
            PfEpsTitles[PF.PfNumEPSProcessed]);
    else
#endif
        put(NXLT("%%%%BeginDocument: <NoName>\n"));

    /* filter out any cnt-D in EPS data */
    for (ixp = eps_block ; ixp != NULL ; )
    {
        ixp = strchr(ixp, (char)4) ;
        if (ixp != NULL)
            strcpy(ixp, ixp+1) ;
    }

    /* put it all out into the stream */
    PfOutString(eps_block);

    put(NXLT("%%%%EndDocument\n"));
    put(NXLT("ENDEPSF\n"));   /* reset state */

    PF.PfNumEPSProcessed++;
    PS_GR();

    return(0);
}


/*************************************************************************
 *
 * 
 * 
 * 
************************************************************************/
void
D01SetCMYKColor(c, m, y, k, max)
int c, m, y, k, max;
{
    int tmp_c, tmp_m, tmp_y, tmp_k;
    double gray_f;

    if ( PF.PfColordev )
    {
        /* should there be an option somewhere to reduce the k? */
        put(NXLT("%.4f %.4f %.4f %.4f SET_CMYK\n"),
            (double)c/(double)max,
            (double)m/(double)max,
            (double)y/(double)max,
            (double)k/(double)max);
    }
    else if ( c || m || y || k )
    {
        tmp_c = MIN(c+k, max);
        tmp_m = MIN(m+k, max);
        tmp_y = MIN(y+k, max);
        tmp_k = 0;
        gray_f = (double)((double)(max*100) - (30*tmp_c) - (59*tmp_m) - (11*tmp_y))/(double)(max*100);
        put(NXLT("%.4f GRAY\n"), gray_f);
    }
    else
        put(NXLT("1 GRAY\n")); /* white */
}

/*************************************************************************
 *
 * 
 * 
 * 
************************************************************************/
/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
int
D01SetColor(color_request)
short color_request;
{
    if (PF.PfColordev)
        switch(color_request)
        {
         case TMA_WHITE:
            put(NXLT("1 GRAY\n"));
            break;
         case TMA_RED:
            put(NXLT("0.0 0.87 0.87 0.0 SET_CMYK\n"));
            break;
         case TMA_LBLUE:
            put(NXLT("1.0 0.0 0.0 0.0 SET_CMYK\n"));
            break;
         case TMA_GREEN:
            put(NXLT("0.74 0.0 0.63 0.0 SET_CMYK\n"));
            break;
         case TMA_PURPLE:
            put(NXLT("0.44 0.81 0.0 0.0 SET_CMYK\n"));
            break;
         case TMA_BLUE:
            put(NXLT("1.0 0.3 0.0 0.0 SET_CMYK\n"));
            break;
         case TMA_YELLOW:
            put(NXLT("0.0 0.0 1.0 0.0 SET_CMYK\n"));
            break;
         case TMA_DGRAY:
            put(NXLT("0.0 0.0 0.0 0.5 SET_CMYK\n"));
            break;
         case TMA_OLIVE:
            put(NXLT("0.64 0.5 0.63 0.0 SET_CMYK\n"));
            break;
         case TMA_LGRAY:
            put(NXLT("0.0 0.0 0.0 0.25 SET_CMYK\n"));
            break;
         case TMA_DRED:
            put(NXLT("0.0 0.74 0.81 0.37 SET_CMYK\n"));
            break;
         case TMA_ORANGE:
            put(NXLT("0.0 0.38 0.69 0.0 SET_CMYK\n"));
            break;
         case TMA_DBLUE:
            put(NXLT("1.0 0.72 0.0 0.11 SET_CMYK\n"));
            break;
         case TMA_BLACK:
         default:
            put(NXLT("0.0 0.0 0.0 1.0 SET_CMYK\n"));
            break;
        }
    else
        switch(color_request)
        {
        case TMA_WHITE:
            put(NXLT("1 GRAY\n"));
            break;
            
          default:
            put(NXLT("0 GRAY\n"));
            break;
        }
    return(0);
}


/*************************************************************************
 *
 * 
 * 
 * 
************************************************************************/
/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
static int
D197SetColor(color_request)
short color_request;
{
 Debugf("Color %d requested.\n", color_request);
 switch(color_request)
 {
  case TMA_WHITE:
        put(NXLT("1 GRAY\n"));
        break;
  case TMA_RED:
        put(NXLT("0.0 0.87 0.87 0.0 SET_CMYK\n"));
        break;
  case TMA_LBLUE:
        put(NXLT("1.0 0.0 0.0 0.0 SET_CMYK\n"));
        break;
  case TMA_GREEN:
        put(NXLT("0.74 0.0 0.63 0.0 SET_CMYK\n"));
        break;
  case TMA_PURPLE:
        put(NXLT("0.44 0.81 0.0 0.0 SET_CMYK\n"));
        break;
  case TMA_BLUE:
        put(NXLT("1.0 0.3 0.0 0.0 SET_CMYK\n"));
        break;
  case TMA_YELLOW:
        put(NXLT("0.0 0.0 1.0 0.0 SET_CMYK\n"));
        break;
  case TMA_DGRAY:
        put(NXLT("0.0 0.0 0.0 0.5 SET_CMYK\n"));
        break;
  case TMA_OLIVE:
        put(NXLT("0.64 0.5 0.63 0.0 SET_CMYK\n"));
        break;
  case TMA_LGRAY:
        put(NXLT("0.0 0.0 0.0 0.25 SET_CMYK\n"));
        break;
  case TMA_DRED:
        put(NXLT("0.0 0.74 0.81 0.37 SET_CMYK\n"));
        break;
  case TMA_ORANGE:
        put(NXLT("0.0 0.38 0.69 0.0 SET_CMYK\n"));
        break;
  case TMA_DBLUE:
        put(NXLT("1.0 0.72 0.0 0.11 SET_CMYK\n"));
        break;
  case TMA_BLACK:
  default:
        put(NXLT("0.0 0.0 0.0 1.0 SET_CMYK\n"));
        break;
 }
 return(0);
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
static char *PFstrstr(longstr,shortstr)
char *longstr, *shortstr;
{
    int len = strlen(shortstr);

    /* Worst possible way to implement this... */

    while ( *longstr )
    {
        if ( 0 == strncmp(longstr,shortstr,len) )
            return(longstr);
        longstr++;
    }
    return(NULL);
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
static void InsertPSFragment(s)
    char *s ;
{
    char *insertText ;

    if (*s != '*')
    {
        put(NXLT("%s\n"), s);
        return ;
    }

    if (PPDfile == NULL)
        return ;

    if (insertText = doPPDInvocation(PPDfile, s))
    {
        put(NXLT("%s\n"), insertText);
        freeInsertText(insertText);
    }

}

/*-----------------------------------------------------------------------------
 *
 * the PSinserts global may specify some extra PS code
 * to include in the output (for paper handling, e.g.).
 * check whether there are any postscript fragments of
 * the specified type; output any that are found.
 *
-----------------------------------------------------------------------------*/
static void checkInsert(ins, type)
    PSinserts *ins;
    int type;
{
    patData **inserts;
    int n, ix;
    if (!ins || !ins->numInserts)
        return;
    if (inserts = getInserts(ins, type, &n))
    {
        for(ix = 0; ix < n; ix++)
        {
            if ((inserts[ix])->text[0] == '~')
            {
                FILE *textFile;
                char linebuf[256];
                    /* the remainder of the text string is to
                     * be interpreted as the name of a file which
                     * contains the actual insert text */
                textFile = fopen(&((inserts[ix])->text[1]), "r");
                if (textFile == NULL)
                    continue;
                while(fgets(linebuf, 256, textFile))
                {
                    /* use file contents as is */
                    put(NXLT("%s"), linebuf);
                }
                fclose(textFile);
            }
            else
            {
                put(NXLT("%s\n"), (inserts[ix])->text);
            }
        }
        freeInsertList((char *)inserts);
    }
}

/*-----------------------------------------------------------------------------
    
    
    The PSinserts global may specify that some PS frags
    should be looked up in an adobe printer description
    file (ppd). Do the lookup and any output.
 *
    
-----------------------------------------------------------------------------*/
static void checkPPD(ins)
    PSinserts *ins;
{
    patData **insData;
    char *insertText;
    int numInserts, i;
    FILE *ppdFile = (FILE *)NULL;

    if (!ins || !ins->numInserts)
        return;

    if (insData = getInserts(ins, PPDFileName, &numInserts))
    {
        /* open 1st ppd file found */
        ppdFile = fopen((insData[0])->text, "r");
        freeInsertList((char *)insData);
        if (ppdFile == NULL)
            return;
    }
    else
    {
        return;
    }

    /* now look for particular ppd features */
    if (insData = getInserts(ins, PPDInvocation, &numInserts))
    {
        for (i = 0; i < numInserts; i++)
        {
            if (insertText = doPPDInvocation(ppdFile, (insData[i])->text))
            {
                /* stuff the text */
                put(NXLT("%%%%BeginFeature: %s\n"), (insData[i])->text);
                put(NXLT("%s\n"), insertText);
                put(NXLT("%%%%EndFeature\n"));
                freeInsertText(insertText);
            }
        }
        freeInsertList((char *)insData);
    }

    if (ppdFile)
        fclose(ppdFile);
}

/*-----------------------------------------------------------------------------
  [get/free]PrintFont[Encoding/Name] are used for listing the job's
  font and encoding names, in the tables printFontNames and printFontEncodings
-----------------------------------------------------------------------------*/
static char * getPrintFontName( name, notUsed )
    char *name; /* e.g., Times-Roman */
    char *notUsed;
{
    int len;
    printFontName *pName;
    fontTabEnt *ft;
    fontMetrics *fm = 0;
    char *metricsFileName;
    char *errString;
    char *actualName;
	char *encodingNameString = 0;

        /* get the font record for key 'name' */
    if (!name)
        return FALSE;

    if (errString = axGetFontRecord(name, &ft))
    {
        freeErrString(errString);
        return FALSE;
    }
	if (ft->encodingFileName)
	{
		/* prefer the encoding name string from the encoding file, if present;
		 * this is needed in case the metrics file is not .enc/.afm -- this
		 * matters because only .enc/.afm files supply the letterform names
		 * needed for a postscript encoding vector download.
		 */
		encodingNameString = axGetEncodingNameFromFile(ft->encodingFileName);
	}
	
	    /* get the target metrics named in the record
		 */
	if (!ft->targetMetricsNames)
		return FALSE;
	
	if (!(metricsFileName = ft->targetMetricsNames[0]))
	{
		/* must switch font record to that of alternate here */
		return FALSE; /* for now */
	}

    if (errString = axGetTargetMetrics(metricsFileName,
                                       ft, &fm))
    {
        freeErrString(errString);
        return FALSE;
    }
	
	if (!encodingNameString)
	{
		/* use the target metrics-file-based encoding name as a backup
		 */
		encodingNameString = fm->encodingName;
	}

    /* now alloc a printFontName instance and fill it out */
    pName = (printFontName *) malloc (sizeof(printFontName));
    actualName = D01PrintName(ft);
    len = strlen(actualName) + 1;
    pName->psFontName = (char *) malloc (len);
    strcpy(pName->psFontName, actualName);
    
    if (encodingNameString) {
        len = strlen(encodingNameString) + 1;
        pName->encodingName = (char *)malloc(len);
        strcpy(pName->encodingName, encodingNameString);
    }
    else {
#if BUILD_KANJI
        if (strncmp(metricsFileName, "BBox", 4) == 0) {
            len = 1;
            pName->encodingName = (char *)malloc(len);
            pName->encodingName[0] = '\0';
        } else {
            len = strlen(NXLT("ISOLatin1Encoding")) + 1;
            pName->encodingName = (char *)malloc(len);
            strcpy(pName->encodingName, NXLT("ISOLatin1Encoding"));
        }
#else
        /* this is a fallback, often lousy */
        len = strlen(NXLT("ISOLatin1Encoding")) + 1;
        pName->encodingName = (char *)malloc(len);
        strcpy(pName->encodingName, NXLT("ISOLatin1Encoding"));
#endif
    }

    /* assign a font number for the job */
    pName->fontNumber = PF.D70FontNumber++;
    pName->eightBit = pName->wideChar = FALSE;
    pName->gallium = ft->gallium ;
#if AXMS
    if ( ft->PostScriptPrintName == (char *)NULL )
        /* we must be down-loading this font, set the gallium bit to get */
        /* upper levels to output correct font names                     */
        pName->gallium = TRUE;
#endif
    pName->sizes = (int *) NULL;
    pName->numSizes = 0;
    return ((char *) pName);
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
static void freePrintFontName(np, notUsed)
    char *np;
    char *notUsed;
{
    printFontName *fn;
    
    if (np)
    {

        fn = (printFontName *) np;
        if (fn->psFontName)
        {
            free(fn->psFontName);
            fn->psFontName = NULL;
        }
        if (fn->encodingName)
        {
            free(fn->encodingName);
            fn->encodingName = NULL;
        }
        if (fn->sizes)
        {
            free(fn->sizes);
            fn->sizes = NULL;
        }
        free(fn);
    }
}

/*-----------------------------------------------------------------------------

    a list of encoding names is used to collect
    a print job's set of encoding vector names

-----------------------------------------------------------------------------*/
static char * getPrintFontEncoding( name, notUsed )
    char *name;          /* e.g., ISOLatin1Encoding */
    char *notUsed;
{
    int *tEight;

    tEight = (int *) malloc (sizeof(int));
    *tEight = FALSE;
    return ((char *) tEight);
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
static void freePrintFontEncoding(np, notUsed)
    char *np;
    char *notUsed;
{
    if (np)
        free(np);
}

/*-----------------------------------------------------------------------------

    D01SetClip - sets the clipping in the print graphics context

-----------------------------------------------------------------------------*/
int D01SetClip(clipRect, x, y, last_x, last_y)
    AxRectangle *clipRect;        /* current app clip */
    int x, y, last_x, last_y;      /* new appclip requested */
{
    /* set app clip */
    if ( (x != clipRect->origin.x) ||
         (y != clipRect->origin.y) ||
         (last_x != clipRect->corner.x) ||
         (last_y != clipRect->corner.y) )
    {
        PF.PfPage.appClip.origin.x = x;
        PF.PfPage.appClip.origin.y = y;
        PF.PfPage.appClip.corner.x = last_x;
        PF.PfPage.appClip.corner.y = last_y;

        PF.left = x;
        PF.top = y;
        PF.right = last_x;
        PF.bottom = last_y;

        while ( PF.ClipGSaveLevel > 0 ) /* should this be a while() ? */
        {
            PS_GR();

        }
        PS_GS();

        Curfont.number = Curfont.size = -1;
        PF.cur_thick = MINIMUM_LINE_WIDTH;        /* default page setup */
        PF.cur_style = TM_SOLID;                    /* default page setup */

        clip_box(PF.left, PF.top, PF.right, PF.bottom) ;


        return (TRUE);
    }
    return (FALSE);
}

/*-----------------------------------------------------------------------------

    D01DrawClip - sets the clipping in the print graphics context
    
-----------------------------------------------------------------------------*/
int D01DrawClip(clip)
    TmRect32 *clip ;
{
    if (clip != NULL && clip->x_size != 0 && clip->y_size != 0)
        clip_box(clip->x_position,
                clip->y_position,
                clip->x_position + clip->x_size,
                clip->y_position + clip->y_size) ;

    return (0);
}

/*-----------------------------------------------------------------------------
    This will set landscape if necessary for the first page
-----------------------------------------------------------------------------*/
int D01Landscape()
{
    if ( PF.PfLandscape )
        put(NXLT("%d DOLANDSCAPE\n"), PF.pagetop);

    return(0);
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
static void D01AddGeneralHeaderComments()
{
    char *s, *t ;

    s = GetProfileByStr(NXLT("axPrintHeaderComments"));

    put(NXLT("%%%%Creator: Applixware\n"));
    while (s != NULL)
    {
        s = strchr(s, '<');
        if (s == NULL)
            break ; 
        s++ ;
        switch (*s)
        {
        case 'T':
            put(NXLT("%%%%Title: %s\n"), PF.PfDoc_name);
            break;
        case 'D':
            put(NXLT("%%%%CreationDate: %s\n"), PF.PfDoc_date);
            break;
        case 'U':
            put(NXLT("%%%%For: %s\n"), PF.PfUser_name);
            break;
        case 'H':
            put(NXLT("%%%%Host: %s\n"), PF.PfHost_name);
            break;
        case 'V':
            put(NXLT("%%%%Revision: %.3f (%s)\n"),
                (double)BUILD_REVISION, BUILD_VERSION_STR);
            break;
        default:
            t = strchr(s, '>') ;
            if (t != NULL)
            {              
                *t = 0 ;
                put(NXLT("%%%%: %s\n"), s);
                *t = '>' ;
                s = t + 1 ;
            }
            break;
        }
    }
    put(NXLT("%%%%Pages: %d\n"), PF.PfNumTotalPages);
}

/*-----------------------------------------------------------------------------
    This scans the string to be printed looking for characters that must
    be either escaped or octally converted. It will perform the 
    appropriate action for each 
-----------------------------------------------------------------------------*/
static char *PSescapeText(instr, txlen)
    register tx_char *instr;
    register int txlen;
{

    int malloc_len = 0;
    register int tx;
    register int out_x;
    register char *outstr = NULL;

    /* scan the string looking for special chars that will
       require the escape char. These will not be accounted for in the
       txlen value */
    malloc_len = txlen * 8;   /* prepare for worst case: all 16-bit chars*/

#if 0
 /*** the following is taken care of by the "worst case" computation above - jon */
    for (tx = 0; tx < txlen; tx++)
    {
        if ((instr[tx] == '(') ||
            (instr[tx] == ')') ||
            (instr[tx] == '\\'))
        {
            malloc_len = malloc_len + 1;
        }
        else if ((char)instr[tx] < SPACE || (char)instr[tx] > '~')
        {
            malloc_len = malloc_len + 4; /* for the octal conversion */
        }
    }
#endif

    outstr = (char *) malloc (malloc_len + 1);
    for (tx = 0, out_x=0; tx < txlen; tx++, out_x++)
    {
        /* look for special chars */

        if ((instr[tx] >= SPACE) && (instr[tx] <= '~'))
        {

            switch (instr[tx])
            {
                case '(':
                case ')':
                case '\\':
                    outstr[out_x++] = '\\';
                    outstr[out_x]=instr[tx];
                    break;
                default:
                    outstr[out_x] = instr[tx];
                    break;
            } /* switch */
        }
        else
        {
            /* chars outside the normal printing range must be
               converted */
            if (instr[tx] == NONBREAK_SPACE)
                outstr[out_x] = SPACE;
            else if ( IS_ORIENT_CHAR(instr[tx]) )
            {
#if BUILD_KANJI
             /* Since the PostScript printer don't use any shift code,
              * we must throw away here the shift code '0x8e' in
              * Narrow Kana */
             if ((instr[tx] >> 8) != 0x8e)
                 {
                 /* output a 16 bit char as 2 octal bytes */
                 outstr[out_x++] = '\\';
                 outstr[out_x++] = (char)('0' + ((instr[tx] >> 14) & 0x03));
                 outstr[out_x++] = (char)('0' + ((instr[tx] >> 11) & 0x07));
                 outstr[out_x++] = (char)('0' + ((instr[tx] >> 8) & 0x07));
                 }
             else
                 {
                 /* don't output the first byte in instr[tx] */
                 ;
                 }
#else
                    /* output a 16 bit char as 2 octal bytes */
                outstr[out_x++] = '\\';
                outstr[out_x++] = (char)('0' + ((instr[tx] >> 14) & 0x03));
                outstr[out_x++] = (char)('0' + ((instr[tx] >> 11) & 0x07));
                outstr[out_x++] = (char)('0' + ((instr[tx] >> 8) & 0x07));
#endif
                outstr[out_x++] = '\\';
                outstr[out_x++] = (char)('0' + ((instr[tx] >> 6) & 0x03));
                outstr[out_x++] = (char)('0' + ((instr[tx] >> 3) & 0x07));
                outstr[out_x] = (char)('0' + (instr[tx] & 0x07));
            }
            else
            {
                /* an 8-bit char, between 128 and 256 in value */
                outstr[out_x++] = '\\';
                outstr[out_x++] = (char)('0' + (((int)instr[tx] >> 6) & 0x07));
                outstr[out_x++] = (char)('0' + (((int)instr[tx] >> 3) & 0x07));
                outstr[out_x] = (char)('0' + ((int)instr[tx] & 0x07));
            }
        }
    }
    outstr[out_x] = '\0';

    return(outstr);
}


/*-----------------------------------------------------------------------------
    Type1DownLoad

    download the Type 1 PostScript font
-----------------------------------------------------------------------------*/
static void Type1DownLoad(filename)
    char *filename ;
{
    FILE *fp = axOpenFontFile(filename) ;
    int dbyte, seglen, i, j ;

    if (fp == NULL)
        PFabort1(XLT("Unable to find font file %s"), filename) ;

    while (TRUE)
    {
        if (getc(fp) != 128)
            goto read_error ;
        
        switch (getc(fp))
        {
        case 1:
            seglen = segmentLen(fp) ;
            if (seglen < 0)
                goto read_error ;
            for (i = 0 ; i < seglen ; i++)
            {
                dbyte = getc(fp) ;
                if (dbyte == EOF)
                    goto read_error ;
                if (dbyte == 0x0d)
                    dbyte = 0x0a ;
                put(NXLT("%c"), dbyte) ;
            }
            continue ;
        case 2:
            seglen = segmentLen(fp) ;
            if (seglen < 0)
                goto read_error ;
            for (i = 0, j = 0 ; i < seglen ; i++, j++)
            {
                dbyte = getc(fp) ;
                if (dbyte == EOF)
                    goto read_error ;
                put(NXLT("%c"), hex(dbyte>>4)) ;
                put(NXLT("%c"), hex(dbyte)) ;
                if (j == 40)
                {
                    put("\n") ;
                    j = 0 ;
                }
            }
            continue ;
        case 3:
            break ;
        default:
read_error:
            fclose(fp) ;
            PFabort1(XLT("Read error on font file %s"), filename) ;
        }
        break ;
    }
    fclose(fp) ;
    return ;
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
static int segmentLen(fp)
    FILE *fp ;
{
    int b, i, k = 0 ;

    for (i = 0 ; i < 4 ; i++)
    {
        b = getc(fp) ;
        if (b == EOF)
            return -1 ;
        k += b << (8*i) ;
    }
    return k ;
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
static int hex(d)
    int d ;
{
    d &= 0xf ;
    if (d <= 9)
        d += '0' ;
    else
        d += 'A' - 10 ;

    return d ;
}

/*-----------------------------------------------------------------------------
    D01PrintName - a safety wrapper for making a
    <fontTabEnt>->PostScriptPrintName reference
-----------------------------------------------------------------------------*/
static char *D01PrintName(ftab)
    fontTabEnt *ftab;
{
    if (!ftab)
        return (NXLT("Times-Roman"));
    if (ftab->PostScriptPrintName)
        return (ftab->PostScriptPrintName);
#if AX_X11
    if (ftab->gallium)
        return (ftab->globalName);
#else
    /* we will now down-load any true-type which does not have a postscript */
    /* equivalent                                                           */
    return( ftab->globalName );
#endif
    return (NXLT("Times-Roman"));
}

/*-----------------------------------------------------------------------------
    D01SoftDownLoad
-----------------------------------------------------------------------------*/
static void D01SoftDownLoad()
{
    int ix, jx, kx, k, c ;
    fontTabEnt *ftab ;
    char *err;
    PF_soft *s ;
    PF_soft *sl = PF.slist ;
    axpfs_ext *x ;
    fontMetrics *fm ;
    axletter *al;
    plettername pl;
    XCharStruct *xc ;
    int ccount, cw, ch ;
    char *fontname ;
    float pointsize ;

    for (ix = 0 ; ix < PF.scount ; ix++)
    {
        s = &sl[ix] ;
        ftab = (fontTabEnt *)s->ftab ;

        pointsize = s->pointsize/10000.0 ;
        ccount = 0 ;
        for (jx = 0 ; jx < 256 ; jx++)
            { 
            if (s->needed[jx])
                ccount++ ;  
            }

        err = axGetTargetMetrics(ftab->targetMetricsNames[0], ftab, &fm);
        if (err)
        {
            freeErrString(err);
            return ;
        }

#if AXMS
        /* if the soft-font structures have not bee set up yet, do so now */
        if ( !fm->xfont->ext_data )
            axgfsAddFontExtension( fm->xfont, TargetRes, AXGFS_TARGET_POINTSIZE, 
                                                ftab->targetMetricsNames[0], ftab );
#endif
        x = (axpfs_ext *) fm->xfont->ext_data->private_data;
        fontname = PSfontName(ftab->globalName, s->pointsize) ;

        put(NXLT("\n"));

        put(NXLT("9 dict dup begin\n"));
        put(NXLT(" /FontType 3 def\n"));
        put(NXLT(" /FontMatrix [1 0 0 1 0 0] def\n"));
#if 0
        put(NXLT(" /FontBBox [%f %f %f %f] def\n"), 
#if 0
                pointsize * mToPix(fm->xfont->max_bounds.width, pointsize, 300)/300.0,
                pointsize * mToPix(fm->xfont->max_bounds.ascent + fm->xfont->max_bounds.descent, pointsize, 300)/300.0,
#endif
                fm->xfont->max_bounds.width/1000.0,
                (fm->xfont->max_bounds.ascent + fm->xfont->max_bounds.descent)/1000.0,
                0.0, 0.0) ;
#endif
        put(NXLT(" /FontBBox [0 0 0 0] def\n")) ;
        put(NXLT(" /Encoding 256 array def\n"));
        put(NXLT(" 0 1 255 {Encoding exch /.notdef put} for\n"));
        put(NXLT(" Encoding\n"));

        for (jx = 0 ; jx < 256 ; jx++)
            if (s->needed[jx])
                put(NXLT(" dup %d /c%d put\n"), jx, jx);
            
        put(NXLT(" pop\n"));

        put(NXLT(" /BuildChar\n"));
        put(NXLT("  { 0 begin\n"));
        put(NXLT("     /char exch def\n"));
        put(NXLT("     /fontdict exch def\n"));
        put(NXLT("     /charname fontdict /Encoding get char get def\n"));
        put(NXLT("     /charinfo fontdict /CharData get charname get def\n"));
        put(NXLT("     /wx charinfo 0 get def\n"));
        put(NXLT("     /charbbox charinfo 1 4 getinterval def\n"));
        put(NXLT("     wx 0 charbbox aload pop setcachedevice\n"));
        put(NXLT("     charinfo 5 get charinfo 6 get true\n"));
        put(NXLT("     fontdict /imagemaskmatrix get\n"));
        put(NXLT("       dup 4 charinfo 7 get put\n"));
        put(NXLT("       dup 5 charinfo 8 get put\n"));
        put(NXLT("     charinfo 9 1 getinterval cvx\n"));
        put(NXLT("     imagemask\n"));
        put(NXLT("   end\n"));
        put(NXLT("   } def\n"));
        put(NXLT(" /BuildChar load 0 6 dict put\n"));
        put(NXLT(" /imagemaskmatrix [%f 0 0 -%f 0 0] def\n"),
                     300.0*pointsize/72.0,
                     300.0*pointsize/72.0);

        put(NXLT("  /CharData %d dict def\n"), ccount+1);
        put(NXLT("  CharData begin\n"));

        for (jx = 0 ; jx < 256 ; jx++)
        {
            if (s->needed[jx])
            {
				memset((void *) &pl, 0, sizeof(pl));
                pl.charcode = (unsigned short) jx;
                pl.pfid.pointsize = s->pointsize / 10000 ;
                pl.pfid.ctm = s->ctm ;
                axhGetItemI(&pl, sizeof(plettername), &x->com.letters,
                            (char *) fm->xfont, (char **)&al);
                if (al->bits == NULL)
                    cw = ch = 0 ;
                else
                {
                    cw = al->bbox->width ;
                    ch = al->bbox->height ;
                }
                k = ch * ((cw + 7) / 8) ;
                xc = getCharStruct(jx, fm->xfont) ;
                put(NXLT("  /c%d [%f %f %f %f %f %d %d %f %f\n"),
                        jx,
                        xc->width/1000.0,
                        xc->lbearing/1000.0,
                        -xc->descent/1000.0,
                        xc->rbearing/1000.0,
                        xc->ascent/1000.0,
                        cw > 0 ? cw : 1,
                        ch > 0 ? ch : 1,
                        -(float)mToPix(xc->lbearing, pointsize, 300) - 0.5,
                        ((float)ch - (float)mToPix(xc->descent, pointsize, 300)) - 0.5 ) ;
                if (k > MAX_PS_STRING)
                {
                    put(NXLT("%%%% character exceeds PostScript string limit\n")) ;
                    put(NXLT("  < > ] def \n")) ;
                }                  
                else
                {
                    put(NXLT("  <")) ;
                    for (kx = 0 ; kx < k ; kx++)
                    {
                        c = ((char *)al->bits)[kx] ;
                        put(NXLT("%02x"), c & 0xff) ;
                        if (((kx+1) % 40) == 0)
                            put(NXLT("\n   ")) ;
                    }                  
                    put(NXLT("> ] def\n")) ;
                }                  
            }
        }
            
        put(NXLT("    /.notdef [.24 0 0 0 0 1 1 0 0 <>] def\n"));
        put(NXLT("  end\n"));
#if 0
        put(NXLT("  /UniqueID %d def\n"), s->id);
#endif
        put(NXLT("end\n"));
        put(NXLT("/%s exch definefont pop\n"), fontname);

        free(fontname) ;    
    }

#if 0
        /* down load font header */
    PCL_font_header.size = sizeof(PCL_font_header) ;
    PCL_font_header.type = 2 ;
    PCL_font_header.style_MSB = 0 ;
    PCL_font_header.reserved_1 = 0 ;
    PCL_font_header.baseline = mToPix(fm->xfont->max_bounds.ascent,
                        ((float)s->pointsize)/10000.0, 300) ;
    PCL_font_header.cell_width = mToPix(fm->xfont->max_bounds.width,
                        ((float)s->pointsize)/10000.0, 300) ;
    PCL_font_header.cell_height = mToPix(fm->xfont->max_bounds.ascent +
                        fm->xfont->max_bounds.descent,
                        ((float)s->pointsize)/10000.0, 300) ;
    PCL_font_header.orientation = 0 ;
    PCL_font_header.spacing = 1 ;
    PCL_font_header.symbol_set = 14 ;
    PCL_font_header.pitch = 0 ;
    PCL_font_header.height = 0 ;
    PCL_font_header.x_height = 0 ;
    PCL_font_header.width_type = 0 ;
    PCL_font_header.style_LSB = 0 ;
    PCL_font_header.stroke_weight = 0 ;
    PCL_font_header.typeface_LSB = 0 ;
    PCL_font_header.typeface_MSB = 0 ;
    PCL_font_header.serif_style = 0;
    PCL_font_header.quality = 2 ;
    PCL_font_header.placement = 0 ;
    PCL_font_header.underline_position = 0 ;
    PCL_font_header.underline_thickness = 0 ;
    PCL_font_header.text_height = 0 ;
    PCL_font_header.text_width = 0 ;
    PCL_font_header.first_code = 0;
    PCL_font_header.Last_code = 255 ;
    PCL_font_header.pitch_extended = 0 ;
    PCL_font_header.height_extended = 0 ;
    PCL_font_header.cap_height = 0 ;
    PCL_font_header.font_number = 0 ;

    /* get the width of a space char to use as the pitch */
    xc = getCharStruct(' ', fm->xfont) ;
    PCL_font_header.pitch = 4*mToPix(xc->width,
                            ((float)s->pointsize)/10000.0, 300) ;

    /* mail the font header */
    PCL_command(NXLT("*c%dD"), s->id) ;
    PCL_command(NXLT(")s%dW"), sizeof(PCL_font_header)) ;
    PfOutCString(&PCL_font_header, sizeof(PCL_font_header)) ;

    /* find all the characters we need */
    x = (axpfs_ext *) fm->xfont->ext_data->private_data;
    for (jx = 0 ; jx < 256 ; jx++)
    {
        if (s->needed[jx])
        {
            pl.charcode = (unsigned short) jx;
            pl.pfid.pointsize = s->pointsize / 10000;
            pl.pfid.ctm = s->ctm;
            axhGetItemI(&pl, sizeof(plettername), &x->com.letters,
                        (char *) fm->xfont, &al);

            PCL_command(NXLT("*c%dE"), jx) ;

            /* character descriptor */
            xc = getCharStruct(jx, fm->xfont) ;
            PCLchar.format = 4 ;
            PCLchar.continuation = 0 ;
            PCLchar.size = 14 ;
            PCLchar.class = 1 ;
            PCLchar.orientation = 0 ;
            PCLchar.reserved_1 = 0 ;
            if (al == NULL || al->bits == NULL)
            {
                PCLchar.left_offset = 0 ;
                PCLchar.top_offset = 0 ;
                PCLchar.width = 0 ;
                PCLchar.height = 0 ;
                PCLchar.delta_x = 0 ;
            }
            else
            {
                PCLchar.left_offset = al->bbox->origin.x ;
                PCLchar.top_offset = -al->bbox->origin.y ;
                PCLchar.width = al->bbox->width ;
                PCLchar.height = al->bbox->height ;
                    PCLchar.delta_x = 4*mToPix(xc->width,
                            ((float)s->pointsize)/10000.0, 300) ;
            }
            k = PCLchar.height * ((PCLchar.width + 7) / 8) ;
            PCL_command(NXLT("(s%dW"), k + sizeof(PCL_char)) ;
            PfOutCString(&PCLchar, sizeof(PCL_char)) ;
            if (al)
                PfOutCString(al->bits, k) ;
        }
    }
#endif
}

/*-----------------------------------------------------------------------------
    PSfontName
-----------------------------------------------------------------------------*/
static char *PSfontName(s, psize)
    char *s ;
    int psize ;
{
    char p[100] ;
    char *z ;

    sprintf(p, "%s-%d", s, psize) ;
    while (z = strchr(p, ' '))
        *z = '-' ;

    z = malloc(strlen(p)+1) ;
    strcpy(z, p) ;
    return z ;
}
/*-----------------------------------------------------------------------------
        END
-----------------------------------------------------------------------------*/


